/* -*- mode: c++; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */

/*
 Copyright (C) 2003, 2004 Ferdinando Ametrano
 Copyright (C) 2006 Richard Gould
 Copyright (C) 2007 Mark Joshi

 This file is part of QuantLib, a free-software/open-source library
 for financial quantitative analysts and developers - http://quantlib.org/

 QuantLib is free software: you can redistribute it and/or modify it
 under the terms of the QuantLib license.  You should have received a
 copy of the license along with this program; if not, please email
 <quantlib-dev@lists.sf.net>. The license is also available online at
 <http://quantlib.org/license.shtml>.

 This program is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the license for more details.
*/

#include <ql/math/randomnumbers/sobolrsg.hpp>
#include <ql/math/randomnumbers/primitivepolynomials.h>
#include <ql/math/randomnumbers/mt19937uniformrng.hpp>
#include <ql/errors.hpp>

#include <cmath>
namespace QuantLib {

    namespace {

        // number of dimensions in the alternative primitive polynomials
        const unsigned int maxAltDegree = 52;

        static const long AltPrimitivePolynomialDegree01[]=
        {
            0, /* x+1 (1)(1) */
            -1
        };

        static const long AltPrimitivePolynomialDegree02[]=
        {
            1, /* x^2+x+1 (1)1(1) */
            -1
        };

        static const long AltPrimitivePolynomialDegree03[]=
        {
            1, /* x^3    +x+1 (1)01(1) */
            2, /* x^3+x^2  +1 (1)10(1) */
            -1
        };

        static const long AltPrimitivePolynomialDegree04[]=
        {
            1, /* x^4+       +x+1 (1)001(1) */
            4, /* x^4+x^3+     +1 (1)100(1) */
            -1
        };


        static const long AltPrimitivePolynomialDegree05[]=
        {
            2,  /* x^5        +x^2  +1 (1)0010(1) */
            13, /* x^5+x^4+x^3    +x+1 (1)1101(1) */
            7,  /* x^5    +x^3+x^2+x+1 (1)0111(1) */
            14, /* x^5+x^4+x^3+x^2  +1 (1)1110(1) */
            11, /* x^5+x^4    +x^2+x+1 (1)1011(1) */
            4,  /* x^5    +x^3      +1 (1)0100(1) */
            -1
        };

        static const long AltPrimitivePolynomialDegree06[]=
        {
            1,  /* x^6                +x+1 (1)00001(1) */
            16, /* x^6+x^5              +1 (1)10000(1) */
            13, /* x^6    +x^4+x^3    +x+1 (1)01101(1) */
            22, /* x^6+x^5    +x^3+x^2  +1 (1)10110(1) */
            19, /* x^6            +x^2+x+1 (1)10011(1) */
            25, /* x^6+x^5+x^4        +x+1 (1)11001(1) */
            -1
        };


        static const long AltPrimitivePolynomialDegree07[]=
        {
            1,  /* x^7                    +x+1 (1)000001(1) */
            32, /* x^7+x^6                  +1 (1)100000(1) */
            4,  /* x^7            +x^3      +1 (1)000100(1) */
            8,  /* x^7        +x^4          +1 (1)001000(1) */
            7,  /* x^7            +x^3+x^2+x+1 (1)000111(1) */
            56, /* x^7+x^6+x^5+x^4          +1 (1)111000(1) */
            14, /* x^7        +x^4+x^3+x^2  +1 (1)001110(1) */
            28, /* x^7    +x^5+x^4+x^3      +1 (1)011100(1) */
            19, /* x^7    +x^5        +x^2+x+1 (1)010011(1) */
            50, /* x^7+x^6+x^5        +x^2  +1 (1)110010(1) */
            21, /* x^7    +x^5    +x^3    +x+1 (1)010101(1) */
            42, /* x^7+x^6    +x^4    +x^2  +1 (1)101010(1) */
            31, /* x^7    +x^5+x^4+x^3+x^2+x+1 (1)011111(1) */
            62, /* x^7+x^6+x^5+x^4+x^3+x^2  +1 (1)111110(1) */
            37, /* x^7+x^6        +x^3    +x+1 (1)100101(1) */
            41, /* x^7+x^6    +x^4        +x+1 (1)101001(1) */
            55, /* x^7+x^6+x^5    +x^3+x^2+x+1 (1)110111(1) */
            59, /* x^7+x^6+x^5+x^4    +x^2+x+1 (1)111011(1) */
            -1
        };

        static const long AltPrimitivePolynomialDegree08[]=
        {
            14,
            56,
            21,
            22,
            38,
            47,
            49,
            50,
            52,
            67,
            70,
            84,
            97,
            103,
            115,
            122,
            -1
        };



#define N_ALT_MAX_DEGREE 8

        const long *const AltPrimitivePolynomials[N_ALT_MAX_DEGREE]=
        {
            AltPrimitivePolynomialDegree01,
            AltPrimitivePolynomialDegree02,
            AltPrimitivePolynomialDegree03,
            AltPrimitivePolynomialDegree04,
            AltPrimitivePolynomialDegree05,
            AltPrimitivePolynomialDegree06,
            AltPrimitivePolynomialDegree07,
            AltPrimitivePolynomialDegree08
        };

        /* Sobol' Levitan coefficients of the free direction integers as given
           by Bratley, P., Fox, B.L. (1988)
        */
        const unsigned long dim02SLinitializers[] = {
            1UL, 0UL };
        const unsigned long dim03SLinitializers[] = {
            1UL, 1UL, 0UL };
        const unsigned long dim04SLinitializers[] = {
            1UL, 3UL, 7UL, 0UL };
        const unsigned long dim05SLinitializers[] = {
            1UL, 1UL, 5UL, 0UL };
        const unsigned long dim06SLinitializers[] = {
            1UL, 3UL, 1UL, 1UL, 0UL };
        const unsigned long dim07SLinitializers[] = {
            1UL, 1UL, 3UL, 7UL, 0UL };
        const unsigned long dim08SLinitializers[] = {
            1UL, 3UL, 3UL, 9UL, 9UL, 0UL };
        const unsigned long dim09SLinitializers[] = {
            1UL, 3UL, 7UL, 13UL, 3UL, 0UL };
        const unsigned long dim10SLinitializers[] = {
            1UL, 1UL, 5UL, 11UL, 27UL, 0UL };
        const unsigned long dim11SLinitializers[] = {
            1UL, 3UL, 5UL, 1UL, 15UL, 0UL };
        const unsigned long dim12SLinitializers[] = {
            1UL, 1UL, 7UL, 3UL, 29UL, 0UL };
        const unsigned long dim13SLinitializers[] = {
            1UL, 3UL, 7UL, 7UL, 21UL, 0UL };
        const unsigned long dim14SLinitializers[] = {
            1UL, 1UL, 1UL, 9UL, 23UL, 37UL, 0UL };
        const unsigned long dim15SLinitializers[] = {
            1UL, 3UL, 3UL, 5UL, 19UL, 33UL, 0UL };
        const unsigned long dim16SLinitializers[] = {
            1UL, 1UL, 3UL, 13UL, 11UL, 7UL, 0UL };
        const unsigned long dim17SLinitializers[] = {
            1UL, 1UL, 7UL, 13UL, 25UL, 5UL, 0UL };
        const unsigned long dim18SLinitializers[] = {
            1UL, 3UL, 5UL, 11UL, 7UL, 11UL, 0UL };
        const unsigned long dim19SLinitializers[] = {
            1UL, 1UL, 1UL, 3UL, 13UL, 39UL, 0UL };
        const unsigned long dim20SLinitializers[] = {
            1UL, 3UL, 1UL, 15UL, 17UL, 63UL, 13UL, 0UL };
        const unsigned long dim21SLinitializers[] = {
            1UL, 1UL, 5UL, 5UL, 1UL, 27UL, 33UL, 0UL };
        const unsigned long dim22SLinitializers[] = {
            1UL, 3UL, 3UL, 3UL, 25UL, 17UL, 115UL, 0UL };
        const unsigned long dim23SLinitializers[] = {
            1UL, 1UL, 3UL, 15UL, 29UL, 15UL, 41UL, 0UL };
        const unsigned long dim24SLinitializers[] = {
            1UL, 3UL, 1UL, 7UL, 3UL, 23UL, 79UL, 0UL };
        const unsigned long dim25SLinitializers[] = {
            1UL, 3UL, 7UL, 9UL, 31UL, 29UL, 17UL, 0UL };
        const unsigned long dim26SLinitializers[] = {
            1UL, 1UL, 5UL, 13UL, 11UL, 3UL, 29UL, 0UL };
        const unsigned long dim27SLinitializers[] = {
            1UL, 3UL, 1UL, 9UL, 5UL, 21UL, 119UL, 0UL };
        const unsigned long dim28SLinitializers[] = {
            1UL, 1UL, 3UL, 1UL, 23UL, 13UL, 75UL, 0UL };
        const unsigned long dim29SLinitializers[] = {
            1UL, 3UL, 3UL, 11UL, 27UL, 31UL, 73UL, 0UL };
        const unsigned long dim30SLinitializers[] = {
            1UL, 1UL, 7UL, 7UL, 19UL, 25UL, 105UL, 0UL };
        const unsigned long dim31SLinitializers[] = {
            1UL, 3UL, 5UL, 5UL, 21UL, 9UL, 7UL, 0UL };
        const unsigned long dim32SLinitializers[] = {
            1UL, 1UL, 1UL, 15UL, 5UL, 49UL, 59UL, 0UL };
        const unsigned long dim33SLinitializers[] = {
            1UL, 1UL, 1UL, 1UL, 1UL, 33UL, 65UL, 0UL };
        const unsigned long dim34SLinitializers[] = {
            1UL, 3UL, 5UL, 15UL, 17UL, 19UL, 21UL, 0UL };
        const unsigned long dim35SLinitializers[] = {
            1UL, 1UL, 7UL, 11UL, 13UL, 29UL, 3UL, 0UL };
        const unsigned long dim36SLinitializers[] = {
            1UL, 3UL, 7UL, 5UL, 7UL, 11UL, 113UL, 0UL };
        const unsigned long dim37SLinitializers[] = {
            1UL, 1UL, 5UL, 3UL, 15UL, 19UL, 61UL, 0UL };
        const unsigned long dim38SLinitializers[] = {
            1UL, 3UL, 1UL, 1UL, 9UL, 27UL, 89UL, 7UL, 0UL };
        const unsigned long dim39SLinitializers[] = {
            1UL, 1UL, 3UL, 7UL, 31UL, 15UL, 45UL, 23UL, 0UL };
        const unsigned long dim40SLinitializers[] = {
            1UL, 3UL, 3UL, 9UL, 9UL, 25UL, 107UL, 39UL, 0UL };

        const unsigned long * const SLinitializers[39] = {
            dim02SLinitializers,
            dim03SLinitializers,
            dim04SLinitializers,
            dim05SLinitializers,
            dim06SLinitializers,
            dim07SLinitializers,
            dim08SLinitializers,
            dim09SLinitializers,
            dim10SLinitializers,
            dim11SLinitializers,
            dim12SLinitializers,
            dim13SLinitializers,
            dim14SLinitializers,
            dim15SLinitializers,
            dim16SLinitializers,
            dim17SLinitializers,
            dim18SLinitializers,
            dim19SLinitializers,
            dim20SLinitializers,
            dim21SLinitializers,
            dim22SLinitializers,
            dim23SLinitializers,
            dim24SLinitializers,
            dim25SLinitializers,
            dim26SLinitializers,
            dim27SLinitializers,
            dim28SLinitializers,
            dim29SLinitializers,
            dim30SLinitializers,
            dim31SLinitializers,
            dim32SLinitializers,
            dim33SLinitializers,
            dim34SLinitializers,
            dim35SLinitializers,
            dim36SLinitializers,
            dim37SLinitializers,
            dim38SLinitializers,
            dim39SLinitializers,
            dim40SLinitializers
        };

        /* coefficients of the free direction integers as given in
           "Monte Carlo Methods in Finance", by Peter Jckel, section 8.3
        */
        const unsigned long dim09initializers[] = {
            1UL, 3UL, 7UL, 7UL, 21UL, 0UL };
        const unsigned long dim10initializers[] = {
            1UL, 1UL, 5UL, 11UL, 27UL, 0UL };
        const unsigned long dim11initializers[] = {
            1UL, 1UL, 7UL, 3UL, 29UL, 0UL };
        const unsigned long dim12initializers[] = {
            1UL, 3UL, 7UL, 13UL, 3UL, 0UL };
        const unsigned long dim13initializers[] = {
            1UL, 3UL, 5UL, 1UL, 15UL, 0UL };
        const unsigned long dim14initializers[] = {
            1UL, 1UL, 1UL, 9UL, 23UL, 37UL, 0UL };
        const unsigned long dim15initializers[] = {
            1UL, 1UL, 3UL, 13UL, 11UL, 7UL, 0UL };
        const unsigned long dim16initializers[] = {
            1UL, 3UL, 3UL, 5UL, 19UL, 33UL, 0UL };

        const unsigned long dim17initializers[] = {
            1UL, 1UL, 7UL, 13UL, 25UL, 5UL, 0UL };
        const unsigned long dim18initializers[] = {
            1UL, 1UL, 1UL, 3UL, 13UL, 39UL, 0UL };
        const unsigned long dim19initializers[] = {
            1UL, 3UL, 5UL, 11UL, 7UL, 11UL, 0UL };

        const unsigned long dim20initializers[] = {
            1UL, 3UL, 1UL, 7UL, 3UL, 23UL, 79UL, 0UL };
        const unsigned long dim21initializers[] = {
            1UL, 3UL, 1UL, 15UL, 17UL, 63UL, 13UL, 0UL };
        const unsigned long dim22initializers[] = {
            1UL, 3UL, 3UL, 3UL, 25UL, 17UL, 115UL, 0UL };
        const unsigned long dim23initializers[] = {
            1UL, 3UL, 7UL, 9UL, 31UL, 29UL, 17UL, 0UL };
        const unsigned long dim24initializers[] = {
            1UL, 1UL, 3UL, 15UL, 29UL, 15UL, 41UL, 0UL };
        const unsigned long dim25initializers[] = {
            1UL, 3UL, 1UL, 9UL, 5UL, 21UL, 119UL, 0UL };
        const unsigned long dim26initializers[] = {
            1UL, 1UL, 5UL, 5UL, 1UL, 27UL, 33UL, 0UL };
        const unsigned long dim27initializers[] = {
            1UL, 1UL, 3UL, 1UL, 23UL, 13UL, 75UL, 0UL };
        const unsigned long dim28initializers[] = {
            1UL, 1UL, 7UL, 7UL, 19UL, 25UL, 105UL, 0UL };
        const unsigned long dim29initializers[] = {
            1UL, 3UL, 5UL, 5UL, 21UL, 9UL, 7UL, 0UL };
        const unsigned long dim30initializers[] = {
            1UL, 1UL, 1UL, 15UL, 5UL, 49UL, 59UL, 0UL };
        const unsigned long dim31initializers[] = {
            1UL, 3UL, 5UL, 15UL, 17UL, 19UL, 21UL, 0UL };
        const unsigned long dim32initializers[] = {
            1UL, 1UL, 7UL, 11UL, 13UL, 29UL, 3UL, 0UL };

        const unsigned long * const initializers[31] = {
            dim02SLinitializers,
            dim03SLinitializers,
            dim04SLinitializers,
            dim05SLinitializers,
            dim06SLinitializers,
            dim07SLinitializers,
            dim08SLinitializers,
            dim09initializers,
            dim10initializers,
            dim11initializers,
            dim12initializers,
            dim13initializers,
            dim14initializers,
            dim15initializers,
            dim16initializers,
            dim17initializers,
            dim18initializers,
            dim19initializers,
            dim20initializers,
            dim21initializers,
            dim22initializers,
            dim23initializers,
            dim24initializers,
            dim25initializers,
            dim26initializers,
            dim27initializers,
            dim28initializers,
            dim29initializers,
            dim30initializers,
            dim31initializers,
            dim32initializers
        };

        /* Lemieux coefficients of the free direction integers as given
           in Christiane Lemieux, private communication, September 2004
        */
        const unsigned long dim041Linitializers[] = {
            1UL,1UL,3UL,13UL,7UL,35UL,61UL,91UL,0UL};
        const unsigned long dim042Linitializers[] = {
            1UL,1UL,7UL,11UL,5UL,35UL,55UL,75UL,0UL};
        const unsigned long dim043Linitializers[] = {
            1UL,3UL,5UL,5UL,11UL,23UL,29UL,139UL,0UL};
        const unsigned long dim044Linitializers[] = {
            1UL,1UL,1UL,7UL,11UL,15UL,17UL,81UL,0UL};
        const unsigned long dim045Linitializers[] = {
            1UL,1UL,7UL,9UL,5UL,57UL,79UL,103UL,0UL};
        const unsigned long dim046Linitializers[] = {
            1UL,1UL,7UL,13UL,19UL,5UL,5UL,185UL,0UL};
        const unsigned long dim047Linitializers[] = {
            1UL,3UL,1UL,3UL,13UL,57UL,97UL,131UL,0UL};
        const unsigned long dim048Linitializers[] = {
            1UL,1UL,5UL,5UL,21UL,25UL,125UL,197UL,0UL};
        const unsigned long dim049Linitializers[] = {
            1UL,3UL,3UL,9UL,31UL,11UL,103UL,201UL,0UL};
        const unsigned long dim050Linitializers[] = {
            1UL,1UL,5UL,3UL,7UL,25UL,51UL,121UL,0UL};
        const unsigned long dim051Linitializers[] = {
            1UL,3UL,7UL,15UL,19UL,53UL,73UL,189UL,0UL};
        const unsigned long dim052Linitializers[] = {
            1UL,1UL,1UL,15UL,19UL,55UL,27UL,183UL,0UL};
        const unsigned long dim053Linitializers[] = {
            1UL,1UL,7UL,13UL,3UL,29UL,109UL,69UL,0UL};
        const unsigned long dim054Linitializers[] = {
            1UL,1UL,5UL,15UL,15UL,23UL,15UL,1UL,57UL,0UL};
        const unsigned long dim055Linitializers[] = {
            1UL,3UL,1UL,3UL,23UL,55UL,43UL,143UL,397UL,0UL};
        const unsigned long dim056Linitializers[] = {
            1UL,1UL,3UL,11UL,29UL,9UL,35UL,131UL,411UL,0UL};
        const unsigned long dim057Linitializers[] = {
            1UL,3UL,1UL,7UL,27UL,39UL,103UL,199UL,277UL,0UL};
        const unsigned long dim058Linitializers[] = {
            1UL,3UL,7UL,3UL,19UL,55UL,127UL,67UL,449UL,0UL};
        const unsigned long dim059Linitializers[] = {
            1UL,3UL,7UL,3UL,5UL,29UL,45UL,85UL,3UL,0UL};
        const unsigned long dim060Linitializers[] = {
            1UL,3UL,5UL,5UL,13UL,23UL,75UL,245UL,453UL,0UL};
        const unsigned long dim061Linitializers[] = {
            1UL,3UL,1UL,15UL,21UL,47UL,3UL,77UL,165UL,0UL};
        const unsigned long dim062Linitializers[] = {
            1UL,1UL,7UL,9UL,15UL,5UL,117UL,73UL,473UL,0UL};
        const unsigned long dim063Linitializers[] = {
            1UL,3UL,1UL,9UL,1UL,21UL,13UL,173UL,313UL,0UL};
        const unsigned long dim064Linitializers[] = {
            1UL,1UL,7UL,3UL,11UL,45UL,63UL,77UL,49UL,0UL};
        const unsigned long dim065Linitializers[] = {
            1UL,1UL,1UL,1UL,1UL,25UL,123UL,39UL,259UL,0UL};
        const unsigned long dim066Linitializers[] = {
            1UL,1UL,1UL,5UL,23UL,11UL,59UL,11UL,203UL,0UL};
        const unsigned long dim067Linitializers[] = {
            1UL,3UL,3UL,15UL,21UL,1UL,73UL,71UL,421UL,0UL};
        const unsigned long dim068Linitializers[] = {
            1UL,1UL,5UL,11UL,15UL,31UL,115UL,95UL,217UL,0UL};
        const unsigned long dim069Linitializers[] = {
            1UL,1UL,3UL,3UL,7UL,53UL,37UL,43UL,439UL,0UL};
        const unsigned long dim070Linitializers[] = {
            1UL,1UL,1UL,1UL,27UL,53UL,69UL,159UL,321UL,0UL};
        const unsigned long dim071Linitializers[] = {
            1UL,1UL,5UL,15UL,29UL,17UL,19UL,43UL,449UL,0UL};
        const unsigned long dim072Linitializers[] = {
            1UL,1UL,3UL,9UL,1UL,55UL,121UL,205UL,255UL,0UL};
        const unsigned long dim073Linitializers[] = {
            1UL,1UL,3UL,11UL,9UL,47UL,107UL,11UL,417UL,0UL};
        const unsigned long dim074Linitializers[] = {
            1UL,1UL,1UL,5UL,17UL,25UL,21UL,83UL,95UL,0UL};
        const unsigned long dim075Linitializers[] = {
            1UL,3UL,5UL,13UL,31UL,25UL,61UL,157UL,407UL,0UL};
        const unsigned long dim076Linitializers[] = {
            1UL,1UL,7UL,9UL,25UL,33UL,41UL,35UL,17UL,0UL};
        const unsigned long dim077Linitializers[] = {
            1UL,3UL,7UL,15UL,13UL,39UL,61UL,187UL,461UL,0UL};
        const unsigned long dim078Linitializers[] = {
            1UL,3UL,7UL,13UL,5UL,57UL,23UL,177UL,435UL,0UL};
        const unsigned long dim079Linitializers[] = {
            1UL,1UL,3UL,15UL,11UL,27UL,115UL,5UL,337UL,0UL};
        const unsigned long dim080Linitializers[] = {
            1UL,3UL,7UL,3UL,15UL,63UL,61UL,171UL,339UL,0UL};
        const unsigned long dim081Linitializers[] = {
            1UL,3UL,3UL,13UL,15UL,61UL,59UL,47UL,1UL,0UL};
        const unsigned long dim082Linitializers[] = {
            1UL,1UL,5UL,15UL,13UL,5UL,39UL,83UL,329UL,0UL};
        const unsigned long dim083Linitializers[] = {
            1UL,1UL,5UL,5UL,5UL,27UL,25UL,39UL,301UL,0UL};
        const unsigned long dim084Linitializers[] = {
            1UL,1UL,5UL,11UL,31UL,41UL,35UL,233UL,27UL,0UL};
        const unsigned long dim085Linitializers[] = {
            1UL,3UL,5UL,15UL,7UL,37UL,119UL,171UL,419UL,0UL};
        const unsigned long dim086Linitializers[] = {
            1UL,3UL,5UL,5UL,3UL,29UL,21UL,189UL,417UL,0UL};
        const unsigned long dim087Linitializers[] = {
            1UL,1UL,1UL,1UL,21UL,41UL,117UL,119UL,351UL,0UL};
        const unsigned long dim088Linitializers[] = {
            1UL,1UL,3UL,1UL,7UL,27UL,87UL,19UL,213UL,0UL};
        const unsigned long dim089Linitializers[] = {
            1UL,1UL,1UL,1UL,17UL,7UL,97UL,217UL,477UL,0UL};
        const unsigned long dim090Linitializers[] = {
            1UL,1UL,7UL,1UL,29UL,61UL,103UL,231UL,269UL,0UL};
        const unsigned long dim091Linitializers[] = {
            1UL,1UL,7UL,13UL,9UL,27UL,107UL,207UL,311UL,0UL};
        const unsigned long dim092Linitializers[] = {
            1UL,1UL,7UL,5UL,25UL,21UL,107UL,179UL,423UL,0UL};
        const unsigned long dim093Linitializers[] = {
            1UL,3UL,5UL,11UL,7UL,1UL,17UL,245UL,281UL,0UL};
        const unsigned long dim094Linitializers[] = {
            1UL,3UL,5UL,9UL,1UL,5UL,53UL,59UL,125UL,0UL};
        const unsigned long dim095Linitializers[] = {
            1UL,1UL,7UL,1UL,31UL,57UL,71UL,245UL,125UL,0UL};
        const unsigned long dim096Linitializers[] = {
            1UL,1UL,7UL,5UL,5UL,57UL,53UL,253UL,441UL,0UL};
        const unsigned long dim097Linitializers[] = {
            1UL,3UL,1UL,13UL,19UL,35UL,119UL,235UL,381UL,0UL};
        const unsigned long dim098Linitializers[] = {
            1UL,3UL,1UL,7UL,19UL,59UL,115UL,33UL,361UL,0UL};
        const unsigned long dim099Linitializers[] = {
            1UL,1UL,3UL,5UL,13UL,1UL,49UL,143UL,501UL,0UL};
        const unsigned long dim100Linitializers[] = {
            1UL,1UL,3UL,5UL,1UL,63UL,101UL,85UL,189UL,0UL};
        const unsigned long dim101Linitializers[] = {
            1UL,1UL,5UL,11UL,27UL,63UL,13UL,131UL,5UL,0UL};
        const unsigned long dim102Linitializers[] = {
            1UL,1UL,5UL,7UL,15UL,45UL,75UL,59UL,455UL,585UL,0UL};
        const unsigned long dim103Linitializers[] = {
            1UL,3UL,1UL,3UL,7UL,7UL,111UL,23UL,119UL,959UL,0UL};
        const unsigned long dim104Linitializers[] = {
            1UL,3UL,3UL,9UL,11UL,41UL,109UL,163UL,161UL,879UL,0UL};
        const unsigned long dim105Linitializers[] = {
            1UL,3UL,5UL,1UL,21UL,41UL,121UL,183UL,315UL,219UL,0UL};
        const unsigned long dim106Linitializers[] = {
            1UL,1UL,3UL,9UL,15UL,3UL,9UL,223UL,441UL,929UL,0UL};
        const unsigned long dim107Linitializers[] = {
            1UL,1UL,7UL,9UL,3UL,5UL,93UL,57UL,253UL,457UL,0UL};
        const unsigned long dim108Linitializers[] = {
            1UL,1UL,7UL,13UL,15UL,29UL,83UL,21UL,35UL,45UL,0UL};
        const unsigned long dim109Linitializers[] = {
            1UL,1UL,3UL,7UL,13UL,61UL,119UL,219UL,85UL,505UL,0UL};
        const unsigned long dim110Linitializers[] = {
            1UL,1UL,3UL,3UL,17UL,13UL,35UL,197UL,291UL,109UL,0UL};
        const unsigned long dim111Linitializers[] = {
            1UL,1UL,3UL,3UL,5UL,1UL,113UL,103UL,217UL,253UL,0UL};
        const unsigned long dim112Linitializers[] = {
            1UL,1UL,7UL,1UL,15UL,39UL,63UL,223UL,17UL,9UL,0UL};
        const unsigned long dim113Linitializers[] = {
            1UL,3UL,7UL,1UL,17UL,29UL,67UL,103UL,495UL,383UL,0UL};
        const unsigned long dim114Linitializers[] = {
            1UL,3UL,3UL,15UL,31UL,59UL,75UL,165UL,51UL,913UL,0UL};
        const unsigned long dim115Linitializers[] = {
            1UL,3UL,7UL,9UL,5UL,27UL,79UL,219UL,233UL,37UL,0UL};
        const unsigned long dim116Linitializers[] = {
            1UL,3UL,5UL,15UL,1UL,11UL,15UL,211UL,417UL,811UL,0UL};
        const unsigned long dim117Linitializers[] = {
            1UL,3UL,5UL,3UL,29UL,27UL,39UL,137UL,407UL,231UL,0UL};
        const unsigned long dim118Linitializers[] = {
            1UL,1UL,3UL,5UL,29UL,43UL,125UL,135UL,109UL,67UL,0UL};
        const unsigned long dim119Linitializers[] = {
            1UL,1UL,1UL,5UL,11UL,39UL,107UL,159UL,323UL,381UL,0UL};
        const unsigned long dim120Linitializers[] = {
            1UL,1UL,1UL,1UL,9UL,11UL,33UL,55UL,169UL,253UL,0UL};
        const unsigned long dim121Linitializers[] = {
            1UL,3UL,5UL,5UL,11UL,53UL,63UL,101UL,251UL,897UL,0UL};
        const unsigned long dim122Linitializers[] = {
            1UL,3UL,7UL,1UL,25UL,15UL,83UL,119UL,53UL,157UL,0UL};
        const unsigned long dim123Linitializers[] = {
            1UL,3UL,5UL,13UL,5UL,5UL,3UL,195UL,111UL,451UL,0UL};
        const unsigned long dim124Linitializers[] = {
            1UL,3UL,1UL,15UL,11UL,1UL,19UL,11UL,307UL,777UL,0UL};
        const unsigned long dim125Linitializers[] = {
            1UL,3UL,7UL,11UL,5UL,5UL,17UL,231UL,345UL,981UL,0UL};
        const unsigned long dim126Linitializers[] = {
            1UL,1UL,3UL,3UL,1UL,33UL,83UL,201UL,57UL,475UL,0UL};
        const unsigned long dim127Linitializers[] = {
            1UL,3UL,7UL,7UL,17UL,13UL,35UL,175UL,499UL,809UL,0UL};
        const unsigned long dim128Linitializers[] = {
            1UL,1UL,5UL,3UL,3UL,17UL,103UL,119UL,499UL,865UL,0UL};
        const unsigned long dim129Linitializers[] = {
            1UL,1UL,1UL,11UL,27UL,25UL,37UL,121UL,401UL,11UL,0UL};
        const unsigned long dim130Linitializers[] = {
            1UL,1UL,1UL,11UL,9UL,25UL,25UL,241UL,403UL,3UL,0UL};
        const unsigned long dim131Linitializers[] = {
            1UL,1UL,1UL,1UL,11UL,1UL,39UL,163UL,231UL,573UL,0UL};
        const unsigned long dim132Linitializers[] = {
            1UL,1UL,1UL,13UL,13UL,21UL,75UL,185UL,99UL,545UL,0UL};
        const unsigned long dim133Linitializers[] = {
            1UL,1UL,1UL,15UL,3UL,63UL,69UL,11UL,173UL,315UL,0UL};
        const unsigned long dim134Linitializers[] = {
            1UL,3UL,5UL,15UL,11UL,3UL,95UL,49UL,123UL,765UL,0UL};
        const unsigned long dim135Linitializers[] = {
            1UL,1UL,1UL,15UL,3UL,63UL,77UL,31UL,425UL,711UL,0UL};
        const unsigned long dim136Linitializers[] = {
            1UL,1UL,7UL,15UL,1UL,37UL,119UL,145UL,489UL,583UL,0UL};
        const unsigned long dim137Linitializers[] = {
            1UL,3UL,5UL,15UL,3UL,49UL,117UL,211UL,165UL,323UL,0UL};
        const unsigned long dim138Linitializers[] = {
            1UL,3UL,7UL,1UL,27UL,63UL,77UL,201UL,225UL,803UL,0UL};
        const unsigned long dim139Linitializers[] = {
            1UL,1UL,1UL,11UL,23UL,35UL,67UL,21UL,469UL,357UL,0UL};
        const unsigned long dim140Linitializers[] = {
            1UL,1UL,7UL,7UL,9UL,7UL,25UL,237UL,237UL,571UL,0UL};
        const unsigned long dim141Linitializers[] = {
            1UL,1UL,3UL,15UL,29UL,5UL,107UL,109UL,241UL,47UL,0UL};
        const unsigned long dim142Linitializers[] = {
            1UL,3UL,5UL,11UL,27UL,63UL,29UL,13UL,203UL,675UL,0UL};
        const unsigned long dim143Linitializers[] = {
            1UL,1UL,3UL,9UL,9UL,11UL,103UL,179UL,449UL,263UL,0UL};
        const unsigned long dim144Linitializers[] = {
            1UL,3UL,5UL,11UL,29UL,63UL,53UL,151UL,259UL,223UL,0UL};
        const unsigned long dim145Linitializers[] = {
            1UL,1UL,3UL,7UL,9UL,25UL,5UL,197UL,237UL,163UL,0UL};
        const unsigned long dim146Linitializers[] = {
            1UL,3UL,7UL,13UL,5UL,57UL,67UL,193UL,147UL,241UL,0UL};
        const unsigned long dim147Linitializers[] = {
            1UL,1UL,5UL,15UL,15UL,33UL,17UL,67UL,161UL,341UL,0UL};
        const unsigned long dim148Linitializers[] = {
            1UL,1UL,3UL,13UL,17UL,43UL,21UL,197UL,441UL,985UL,0UL};
        const unsigned long dim149Linitializers[] = {
            1UL,3UL,1UL,5UL,15UL,33UL,33UL,193UL,305UL,829UL,0UL};
        const unsigned long dim150Linitializers[] = {
            1UL,1UL,1UL,13UL,19UL,27UL,71UL,187UL,477UL,239UL,0UL};
        const unsigned long dim151Linitializers[] = {
            1UL,1UL,1UL,9UL,9UL,17UL,41UL,177UL,229UL,983UL,0UL};
        const unsigned long dim152Linitializers[] = {
            1UL,3UL,5UL,9UL,15UL,45UL,97UL,205UL,43UL,767UL,0UL};
        const unsigned long dim153Linitializers[] = {
            1UL,1UL,1UL,9UL,31UL,31UL,77UL,159UL,395UL,809UL,0UL};
        const unsigned long dim154Linitializers[] = {
            1UL,3UL,3UL,3UL,29UL,19UL,73UL,123UL,165UL,307UL,0UL};
        const unsigned long dim155Linitializers[] = {
            1UL,3UL,1UL,7UL,5UL,11UL,77UL,227UL,355UL,403UL,0UL};
        const unsigned long dim156Linitializers[] = {
            1UL,3UL,5UL,5UL,25UL,31UL,1UL,215UL,451UL,195UL,0UL};
        const unsigned long dim157Linitializers[] = {
            1UL,3UL,7UL,15UL,29UL,37UL,101UL,241UL,17UL,633UL,0UL};
        const unsigned long dim158Linitializers[] = {
            1UL,1UL,5UL,1UL,11UL,3UL,107UL,137UL,489UL,5UL,0UL};
        const unsigned long dim159Linitializers[] = {
            1UL,1UL,1UL,7UL,19UL,19UL,75UL,85UL,471UL,355UL,0UL};
        const unsigned long dim160Linitializers[] = {
            1UL,1UL,3UL,3UL,9UL,13UL,113UL,167UL,13UL,27UL,0UL};
        const unsigned long dim161Linitializers[] = {
            1UL,3UL,5UL,11UL,21UL,3UL,89UL,205UL,377UL,307UL,0UL};
        const unsigned long dim162Linitializers[] = {
            1UL,1UL,1UL,9UL,31UL,61UL,65UL,9UL,391UL,141UL,867UL,0UL};
        const unsigned long dim163Linitializers[] = {
            1UL,1UL,1UL,9UL,19UL,19UL,61UL,227UL,241UL,55UL,161UL,0UL};
        const unsigned long dim164Linitializers[] = {
            1UL,1UL,1UL,11UL,1UL,19UL,7UL,233UL,463UL,171UL,1941UL,0UL};
        const unsigned long dim165Linitializers[] = {
            1UL,1UL,5UL,7UL,25UL,13UL,103UL,75UL,19UL,1021UL,1063UL,0UL};
        const unsigned long dim166Linitializers[] = {
            1UL,1UL,1UL,15UL,17UL,17UL,79UL,63UL,391UL,403UL,1221UL,0UL};
        const unsigned long dim167Linitializers[] = {
            1UL,3UL,3UL,11UL,29UL,25UL,29UL,107UL,335UL,475UL,963UL,0UL};
        const unsigned long dim168Linitializers[] = {
            1UL,3UL,5UL,1UL,31UL,33UL,49UL,43UL,155UL,9UL,1285UL,0UL};
        const unsigned long dim169Linitializers[] = {
            1UL,1UL,5UL,5UL,15UL,47UL,39UL,161UL,357UL,863UL,1039UL,0UL};
        const unsigned long dim170Linitializers[] = {
            1UL,3UL,7UL,15UL,1UL,39UL,47UL,109UL,427UL,393UL,1103UL,0UL};
        const unsigned long dim171Linitializers[] = {
            1UL,1UL,1UL,9UL,9UL,29UL,121UL,233UL,157UL,99UL,701UL,0UL};
        const unsigned long dim172Linitializers[] = {
            1UL,1UL,1UL,7UL,1UL,29UL,75UL,121UL,439UL,109UL,993UL,0UL};
        const unsigned long dim173Linitializers[] = {
            1UL,1UL,1UL,9UL,5UL,1UL,39UL,59UL,89UL,157UL,1865UL,0UL};
        const unsigned long dim174Linitializers[] = {
            1UL,1UL,5UL,1UL,3UL,37UL,89UL,93UL,143UL,533UL,175UL,0UL};
        const unsigned long dim175Linitializers[] = {
            1UL,1UL,3UL,5UL,7UL,33UL,35UL,173UL,159UL,135UL,241UL,0UL};
        const unsigned long dim176Linitializers[] = {
            1UL,1UL,1UL,15UL,17UL,37UL,79UL,131UL,43UL,891UL,229UL,0UL};
        const unsigned long dim177Linitializers[] = {
            1UL,1UL,1UL,1UL,1UL,35UL,121UL,177UL,397UL,1017UL,583UL,0UL};
        const unsigned long dim178Linitializers[] = {
            1UL,1UL,3UL,15UL,31UL,21UL,43UL,67UL,467UL,923UL,1473UL,0UL};
        const unsigned long dim179Linitializers[] = {
            1UL,1UL,1UL,7UL,1UL,33UL,77UL,111UL,125UL,771UL,1975UL,0UL};
        const unsigned long dim180Linitializers[] = {
            1UL,3UL,7UL,13UL,1UL,51UL,113UL,139UL,245UL,573UL,503UL,0UL};
        const unsigned long dim181Linitializers[] = {
            1UL,3UL,1UL,9UL,21UL,49UL,15UL,157UL,49UL,483UL,291UL,0UL};
        const unsigned long dim182Linitializers[] = {
            1UL,1UL,1UL,1UL,29UL,35UL,17UL,65UL,403UL,485UL,1603UL,0UL};
        const unsigned long dim183Linitializers[] = {
            1UL,1UL,1UL,7UL,19UL,1UL,37UL,129UL,203UL,321UL,1809UL,0UL};
        const unsigned long dim184Linitializers[] = {
            1UL,3UL,7UL,15UL,15UL,9UL,5UL,77UL,29UL,485UL,581UL,0UL};
        const unsigned long dim185Linitializers[] = {
            1UL,1UL,3UL,5UL,15UL,49UL,97UL,105UL,309UL,875UL,1581UL,0UL};
        const unsigned long dim186Linitializers[] = {
            1UL,3UL,5UL,1UL,5UL,19UL,63UL,35UL,165UL,399UL,1489UL,0UL};
        const unsigned long dim187Linitializers[] = {
            1UL,3UL,5UL,3UL,23UL,5UL,79UL,137UL,115UL,599UL,1127UL,0UL};
        const unsigned long dim188Linitializers[] = {
            1UL,1UL,7UL,5UL,3UL,61UL,27UL,177UL,257UL,91UL,841UL,0UL};
        const unsigned long dim189Linitializers[] = {
            1UL,1UL,3UL,5UL,9UL,31UL,91UL,209UL,409UL,661UL,159UL,0UL};
        const unsigned long dim190Linitializers[] = {
            1UL,3UL,1UL,15UL,23UL,39UL,23UL,195UL,245UL,203UL,947UL,0UL};
        const unsigned long dim191Linitializers[] = {
            1UL,1UL,3UL,1UL,15UL,59UL,67UL,95UL,155UL,461UL,147UL,0UL};
        const unsigned long dim192Linitializers[] = {
            1UL,3UL,7UL,5UL,23UL,25UL,87UL,11UL,51UL,449UL,1631UL,0UL};
        const unsigned long dim193Linitializers[] = {
            1UL,1UL,1UL,1UL,17UL,57UL,7UL,197UL,409UL,609UL,135UL,0UL};
        const unsigned long dim194Linitializers[] = {
            1UL,1UL,1UL,9UL,1UL,61UL,115UL,113UL,495UL,895UL,1595UL,0UL};
        const unsigned long dim195Linitializers[] = {
            1UL,3UL,7UL,15UL,9UL,47UL,121UL,211UL,379UL,985UL,1755UL,0UL};
        const unsigned long dim196Linitializers[] = {
            1UL,3UL,1UL,3UL,7UL,57UL,27UL,231UL,339UL,325UL,1023UL,0UL};
        const unsigned long dim197Linitializers[] = {
            1UL,1UL,1UL,1UL,19UL,63UL,63UL,239UL,31UL,643UL,373UL,0UL};
        const unsigned long dim198Linitializers[] = {
            1UL,3UL,1UL,11UL,19UL,9UL,7UL,171UL,21UL,691UL,215UL,0UL};
        const unsigned long dim199Linitializers[] = {
            1UL,1UL,5UL,13UL,11UL,57UL,39UL,211UL,241UL,893UL,555UL,0UL};
        const unsigned long dim200Linitializers[] = {
            1UL,1UL,7UL,5UL,29UL,21UL,45UL,59UL,509UL,223UL,491UL,0UL};
        const unsigned long dim201Linitializers[] = {
            1UL,1UL,7UL,9UL,15UL,61UL,97UL,75UL,127UL,779UL,839UL,0UL};
        const unsigned long dim202Linitializers[] = {
            1UL,1UL,7UL,15UL,17UL,33UL,75UL,237UL,191UL,925UL,681UL,0UL};
        const unsigned long dim203Linitializers[] = {
            1UL,3UL,5UL,7UL,27UL,57UL,123UL,111UL,101UL,371UL,1129UL,0UL};
        const unsigned long dim204Linitializers[] = {
            1UL,3UL,5UL,5UL,29UL,45UL,59UL,127UL,229UL,967UL,2027UL,0UL};
        const unsigned long dim205Linitializers[] = {
            1UL,1UL,1UL,1UL,17UL,7UL,23UL,199UL,241UL,455UL,135UL,0UL};
        const unsigned long dim206Linitializers[] = {
            1UL,1UL,7UL,15UL,27UL,29UL,105UL,171UL,337UL,503UL,1817UL,0UL};
        const unsigned long dim207Linitializers[] = {
            1UL,1UL,3UL,7UL,21UL,35UL,61UL,71UL,405UL,647UL,2045UL,0UL};
        const unsigned long dim208Linitializers[] = {
            1UL,1UL,1UL,1UL,1UL,15UL,65UL,167UL,501UL,79UL,737UL,0UL};
        const unsigned long dim209Linitializers[] = {
            1UL,1UL,5UL,1UL,3UL,49UL,27UL,189UL,341UL,615UL,1287UL,0UL};
        const unsigned long dim210Linitializers[] = {
            1UL,1UL,1UL,9UL,1UL,7UL,31UL,159UL,503UL,327UL,1613UL,0UL};
        const unsigned long dim211Linitializers[] = {
            1UL,3UL,3UL,3UL,3UL,23UL,99UL,115UL,323UL,997UL,987UL,0UL};
        const unsigned long dim212Linitializers[] = {
            1UL,1UL,1UL,9UL,19UL,33UL,93UL,247UL,509UL,453UL,891UL,0UL};
        const unsigned long dim213Linitializers[] = {
            1UL,1UL,3UL,1UL,13UL,19UL,35UL,153UL,161UL,633UL,445UL,0UL};
        const unsigned long dim214Linitializers[] = {
            1UL,3UL,5UL,15UL,31UL,5UL,87UL,197UL,183UL,783UL,1823UL,0UL};
        const unsigned long dim215Linitializers[] = {
            1UL,1UL,7UL,5UL,19UL,63UL,69UL,221UL,129UL,231UL,1195UL,0UL};
        const unsigned long dim216Linitializers[] = {
            1UL,1UL,5UL,5UL,13UL,23UL,19UL,231UL,245UL,917UL,379UL,0UL};
        const unsigned long dim217Linitializers[] = {
            1UL,3UL,1UL,15UL,19UL,43UL,27UL,223UL,171UL,413UL,125UL,0UL};
        const unsigned long dim218Linitializers[] = {
            1UL,1UL,1UL,9UL,1UL,59UL,21UL,15UL,509UL,207UL,589UL,0UL};
        const unsigned long dim219Linitializers[] = {
            1UL,3UL,5UL,3UL,19UL,31UL,113UL,19UL,23UL,733UL,499UL,0UL};
        const unsigned long dim220Linitializers[] = {
            1UL,1UL,7UL,1UL,19UL,51UL,101UL,165UL,47UL,925UL,1093UL,0UL};
        const unsigned long dim221Linitializers[] = {
            1UL,3UL,3UL,9UL,15UL,21UL,43UL,243UL,237UL,461UL,1361UL,0UL};
        const unsigned long dim222Linitializers[] = {
            1UL,1UL,1UL,9UL,17UL,15UL,75UL,75UL,113UL,715UL,1419UL,0UL};
        const unsigned long dim223Linitializers[] = {
            1UL,1UL,7UL,13UL,17UL,1UL,99UL,15UL,347UL,721UL,1405UL,0UL};
        const unsigned long dim224Linitializers[] = {
            1UL,1UL,7UL,15UL,7UL,27UL,23UL,183UL,39UL,59UL,571UL,0UL};
        const unsigned long dim225Linitializers[] = {
            1UL,3UL,5UL,9UL,7UL,43UL,35UL,165UL,463UL,567UL,859UL,0UL};
        const unsigned long dim226Linitializers[] = {
            1UL,3UL,3UL,11UL,15UL,19UL,17UL,129UL,311UL,343UL,15UL,0UL};
        const unsigned long dim227Linitializers[] = {
            1UL,1UL,1UL,15UL,31UL,59UL,63UL,39UL,347UL,359UL,105UL,0UL};
        const unsigned long dim228Linitializers[] = {
            1UL,1UL,1UL,15UL,5UL,43UL,87UL,241UL,109UL,61UL,685UL,0UL};
        const unsigned long dim229Linitializers[] = {
            1UL,1UL,7UL,7UL,9UL,39UL,121UL,127UL,369UL,579UL,853UL,0UL};
        const unsigned long dim230Linitializers[] = {
            1UL,1UL,1UL,1UL,17UL,15UL,15UL,95UL,325UL,627UL,299UL,0UL};
        const unsigned long dim231Linitializers[] = {
            1UL,1UL,3UL,13UL,31UL,53UL,85UL,111UL,289UL,811UL,1635UL,0UL};
        const unsigned long dim232Linitializers[] = {
            1UL,3UL,7UL,1UL,19UL,29UL,75UL,185UL,153UL,573UL,653UL,0UL};
        const unsigned long dim233Linitializers[] = {
            1UL,3UL,7UL,1UL,29UL,31UL,55UL,91UL,249UL,247UL,1015UL,0UL};
        const unsigned long dim234Linitializers[] = {
            1UL,3UL,5UL,7UL,1UL,49UL,113UL,139UL,257UL,127UL,307UL,0UL};
        const unsigned long dim235Linitializers[] = {
            1UL,3UL,5UL,9UL,15UL,15UL,123UL,105UL,105UL,225UL,1893UL,0UL};
        const unsigned long dim236Linitializers[] = {
            1UL,3UL,3UL,1UL,15UL,5UL,105UL,249UL,73UL,709UL,1557UL,0UL};
        const unsigned long dim237Linitializers[] = {
            1UL,1UL,1UL,9UL,17UL,31UL,113UL,73UL,65UL,701UL,1439UL,0UL};
        const unsigned long dim238Linitializers[] = {
            1UL,3UL,5UL,15UL,13UL,21UL,117UL,131UL,243UL,859UL,323UL,0UL};
        const unsigned long dim239Linitializers[] = {
            1UL,1UL,1UL,9UL,19UL,15UL,69UL,149UL,89UL,681UL,515UL,0UL};
        const unsigned long dim240Linitializers[] = {
            1UL,1UL,1UL,5UL,29UL,13UL,21UL,97UL,301UL,27UL,967UL,0UL};
        const unsigned long dim241Linitializers[] = {
            1UL,1UL,3UL,3UL,15UL,45UL,107UL,227UL,495UL,769UL,1935UL,0UL};
        const unsigned long dim242Linitializers[] = {
            1UL,1UL,1UL,11UL,5UL,27UL,41UL,173UL,261UL,703UL,1349UL,0UL};
        const unsigned long dim243Linitializers[] = {
            1UL,3UL,3UL,3UL,11UL,35UL,97UL,43UL,501UL,563UL,1331UL,0UL};
        const unsigned long dim244Linitializers[] = {
            1UL,1UL,1UL,7UL,1UL,17UL,87UL,17UL,429UL,245UL,1941UL,0UL};
        const unsigned long dim245Linitializers[] = {
            1UL,1UL,7UL,15UL,29UL,13UL,1UL,175UL,425UL,233UL,797UL,0UL};
        const unsigned long dim246Linitializers[] = {
            1UL,1UL,3UL,11UL,21UL,57UL,49UL,49UL,163UL,685UL,701UL,0UL};
        const unsigned long dim247Linitializers[] = {
            1UL,3UL,3UL,7UL,11UL,45UL,107UL,111UL,379UL,703UL,1403UL,0UL};
        const unsigned long dim248Linitializers[] = {
            1UL,1UL,7UL,3UL,21UL,7UL,117UL,49UL,469UL,37UL,775UL,0UL};
        const unsigned long dim249Linitializers[] = {
            1UL,1UL,5UL,15UL,31UL,63UL,101UL,77UL,507UL,489UL,1955UL,0UL};
        const unsigned long dim250Linitializers[] = {
            1UL,3UL,3UL,11UL,19UL,21UL,101UL,255UL,203UL,673UL,665UL,0UL};
        const unsigned long dim251Linitializers[] = {
            1UL,3UL,3UL,15UL,17UL,47UL,125UL,187UL,271UL,899UL,2003UL,0UL};
        const unsigned long dim252Linitializers[] = {
            1UL,1UL,7UL,7UL,1UL,35UL,13UL,235UL,5UL,337UL,905UL,0UL};
        const unsigned long dim253Linitializers[] = {
            1UL,3UL,1UL,15UL,1UL,43UL,1UL,27UL,37UL,695UL,1429UL,0UL};
        const unsigned long dim254Linitializers[] = {
            1UL,3UL,1UL,11UL,21UL,27UL,93UL,161UL,299UL,665UL,495UL,0UL};
        const unsigned long dim255Linitializers[] = {
            1UL,3UL,3UL,15UL,3UL,1UL,81UL,111UL,105UL,547UL,897UL,0UL};
        const unsigned long dim256Linitializers[] = {
            1UL,3UL,5UL,1UL,3UL,53UL,97UL,253UL,401UL,827UL,1467UL,0UL};
        const unsigned long dim257Linitializers[] = {
            1UL,1UL,1UL,5UL,19UL,59UL,105UL,125UL,271UL,351UL,719UL,0UL};
        const unsigned long dim258Linitializers[] = {
            1UL,3UL,5UL,13UL,7UL,11UL,91UL,41UL,441UL,759UL,1827UL,0UL};
        const unsigned long dim259Linitializers[] = {
            1UL,3UL,7UL,11UL,29UL,61UL,61UL,23UL,307UL,863UL,363UL,0UL};
        const unsigned long dim260Linitializers[] = {
            1UL,1UL,7UL,1UL,15UL,35UL,29UL,133UL,415UL,473UL,1737UL,0UL};
        const unsigned long dim261Linitializers[] = {
            1UL,1UL,1UL,13UL,7UL,33UL,35UL,225UL,117UL,681UL,1545UL,0UL};
        const unsigned long dim262Linitializers[] = {
            1UL,1UL,1UL,3UL,5UL,41UL,83UL,247UL,13UL,373UL,1091UL,0UL};
        const unsigned long dim263Linitializers[] = {
            1UL,3UL,1UL,13UL,25UL,61UL,71UL,217UL,233UL,313UL,547UL,0UL};
        const unsigned long dim264Linitializers[] = {
            1UL,3UL,1UL,7UL,3UL,29UL,3UL,49UL,93UL,465UL,15UL,0UL};
        const unsigned long dim265Linitializers[] = {
            1UL,1UL,1UL,9UL,17UL,61UL,99UL,163UL,129UL,485UL,1087UL,0UL};
        const unsigned long dim266Linitializers[] = {
            1UL,1UL,1UL,9UL,9UL,33UL,31UL,163UL,145UL,649UL,253UL,0UL};
        const unsigned long dim267Linitializers[] = {
            1UL,1UL,1UL,1UL,17UL,63UL,43UL,235UL,287UL,111UL,567UL,0UL};
        const unsigned long dim268Linitializers[] = {
            1UL,3UL,5UL,13UL,29UL,7UL,11UL,69UL,153UL,127UL,449UL,0UL};
        const unsigned long dim269Linitializers[] = {
            1UL,1UL,5UL,9UL,11UL,21UL,15UL,189UL,431UL,493UL,1219UL,0UL};
        const unsigned long dim270Linitializers[] = {
            1UL,1UL,1UL,15UL,19UL,5UL,47UL,91UL,399UL,293UL,1743UL,0UL};
        const unsigned long dim271Linitializers[] = {
            1UL,3UL,3UL,11UL,29UL,53UL,53UL,225UL,409UL,303UL,333UL,0UL};
        const unsigned long dim272Linitializers[] = {
            1UL,1UL,1UL,15UL,31UL,31UL,21UL,81UL,147UL,287UL,1753UL,0UL};
        const unsigned long dim273Linitializers[] = {
            1UL,3UL,5UL,5UL,5UL,63UL,35UL,125UL,41UL,687UL,1793UL,0UL};
        const unsigned long dim274Linitializers[] = {
            1UL,1UL,1UL,9UL,19UL,59UL,107UL,219UL,455UL,971UL,297UL,0UL};
        const unsigned long dim275Linitializers[] = {
            1UL,1UL,3UL,5UL,3UL,51UL,121UL,31UL,245UL,105UL,1311UL,0UL};
        const unsigned long dim276Linitializers[] = {
            1UL,3UL,1UL,5UL,5UL,57UL,75UL,107UL,161UL,431UL,1693UL,0UL};
        const unsigned long dim277Linitializers[] = {
            1UL,3UL,1UL,3UL,19UL,53UL,27UL,31UL,191UL,565UL,1015UL,0UL};
        const unsigned long dim278Linitializers[] = {
            1UL,3UL,5UL,13UL,9UL,41UL,35UL,249UL,287UL,49UL,123UL,0UL};
        const unsigned long dim279Linitializers[] = {
            1UL,1UL,5UL,7UL,27UL,17UL,21UL,3UL,151UL,885UL,1165UL,0UL};
        const unsigned long dim280Linitializers[] = {
            1UL,1UL,7UL,1UL,15UL,17UL,65UL,139UL,427UL,339UL,1171UL,0UL};
        const unsigned long dim281Linitializers[] = {
            1UL,1UL,1UL,5UL,23UL,5UL,9UL,89UL,321UL,907UL,391UL,0UL};
        const unsigned long dim282Linitializers[] = {
            1UL,1UL,7UL,9UL,15UL,1UL,77UL,71UL,87UL,701UL,917UL,0UL};
        const unsigned long dim283Linitializers[] = {
            1UL,1UL,7UL,1UL,17UL,37UL,115UL,127UL,469UL,779UL,1543UL,0UL};
        const unsigned long dim284Linitializers[] = {
            1UL,3UL,7UL,3UL,5UL,61UL,15UL,37UL,301UL,951UL,1437UL,0UL};
        const unsigned long dim285Linitializers[] = {
            1UL,1UL,1UL,13UL,9UL,51UL,127UL,145UL,229UL,55UL,1567UL,0UL};
        const unsigned long dim286Linitializers[] = {
            1UL,3UL,7UL,15UL,19UL,47UL,53UL,153UL,295UL,47UL,1337UL,0UL};
        const unsigned long dim287Linitializers[] = {
            1UL,3UL,3UL,5UL,11UL,31UL,29UL,133UL,327UL,287UL,507UL,0UL};
        const unsigned long dim288Linitializers[] = {
            1UL,1UL,7UL,7UL,25UL,31UL,37UL,199UL,25UL,927UL,1317UL,0UL};
        const unsigned long dim289Linitializers[] = {
            1UL,1UL,7UL,9UL,3UL,39UL,127UL,167UL,345UL,467UL,759UL,0UL};
        const unsigned long dim290Linitializers[] = {
            1UL,1UL,1UL,1UL,31UL,21UL,15UL,101UL,293UL,787UL,1025UL,0UL};
        const unsigned long dim291Linitializers[] = {
            1UL,1UL,5UL,3UL,11UL,41UL,105UL,109UL,149UL,837UL,1813UL,0UL};
        const unsigned long dim292Linitializers[] = {
            1UL,1UL,3UL,5UL,29UL,13UL,19UL,97UL,309UL,901UL,753UL,0UL};
        const unsigned long dim293Linitializers[] = {
            1UL,1UL,7UL,1UL,19UL,17UL,31UL,39UL,173UL,361UL,1177UL,0UL};
        const unsigned long dim294Linitializers[] = {
            1UL,3UL,3UL,3UL,3UL,41UL,81UL,7UL,341UL,491UL,43UL,0UL};
        const unsigned long dim295Linitializers[] = {
            1UL,1UL,7UL,7UL,31UL,35UL,29UL,77UL,11UL,335UL,1275UL,0UL};
        const unsigned long dim296Linitializers[] = {
            1UL,3UL,3UL,15UL,17UL,45UL,19UL,63UL,151UL,849UL,129UL,0UL};
        const unsigned long dim297Linitializers[] = {
            1UL,1UL,7UL,5UL,7UL,13UL,47UL,73UL,79UL,31UL,499UL,0UL};
        const unsigned long dim298Linitializers[] = {
            1UL,3UL,1UL,11UL,1UL,41UL,59UL,151UL,247UL,115UL,1295UL,0UL};
        const unsigned long dim299Linitializers[] = {
            1UL,1UL,1UL,9UL,31UL,37UL,73UL,23UL,295UL,483UL,179UL,0UL};
        const unsigned long dim300Linitializers[] = {
            1UL,3UL,1UL,15UL,13UL,63UL,81UL,27UL,169UL,825UL,2037UL,0UL};
        const unsigned long dim301Linitializers[] = {
            1UL,3UL,5UL,15UL,7UL,11UL,73UL,1UL,451UL,101UL,2039UL,0UL};
        const unsigned long dim302Linitializers[] = {
            1UL,3UL,5UL,3UL,13UL,53UL,31UL,137UL,173UL,319UL,1521UL,0UL};
        const unsigned long dim303Linitializers[] = {
            1UL,3UL,1UL,3UL,29UL,1UL,73UL,227UL,377UL,337UL,1189UL,0UL};
        const unsigned long dim304Linitializers[] = {
            1UL,3UL,3UL,13UL,27UL,9UL,31UL,101UL,229UL,165UL,1983UL,0UL};
        const unsigned long dim305Linitializers[] = {
            1UL,3UL,1UL,13UL,13UL,19UL,19UL,111UL,319UL,421UL,223UL,0UL};
        const unsigned long dim306Linitializers[] = {
            1UL,1UL,7UL,15UL,25UL,37UL,61UL,55UL,359UL,255UL,1955UL,0UL};
        const unsigned long dim307Linitializers[] = {
            1UL,1UL,5UL,13UL,17UL,43UL,49UL,215UL,383UL,915UL,51UL,0UL};
        const unsigned long dim308Linitializers[] = {
            1UL,1UL,3UL,1UL,3UL,7UL,13UL,119UL,155UL,585UL,967UL,0UL};
        const unsigned long dim309Linitializers[] = {
            1UL,3UL,1UL,13UL,1UL,63UL,125UL,21UL,103UL,287UL,457UL,0UL};
        const unsigned long dim310Linitializers[] = {
            1UL,1UL,7UL,1UL,31UL,17UL,125UL,137UL,345UL,379UL,1925UL,0UL};
        const unsigned long dim311Linitializers[] = {
            1UL,1UL,3UL,5UL,5UL,25UL,119UL,153UL,455UL,271UL,2023UL,0UL};
        const unsigned long dim312Linitializers[] = {
            1UL,1UL,7UL,9UL,9UL,37UL,115UL,47UL,5UL,255UL,917UL,0UL};
        const unsigned long dim313Linitializers[] = {
            1UL,3UL,5UL,3UL,31UL,21UL,75UL,203UL,489UL,593UL,1UL,0UL};
        const unsigned long dim314Linitializers[] = {
            1UL,3UL,7UL,15UL,19UL,63UL,123UL,153UL,135UL,977UL,1875UL,0UL};
        const unsigned long dim315Linitializers[] = {
            1UL,1UL,1UL,1UL,5UL,59UL,31UL,25UL,127UL,209UL,745UL,0UL};
        const unsigned long dim316Linitializers[] = {
            1UL,1UL,1UL,1UL,19UL,45UL,67UL,159UL,301UL,199UL,535UL,0UL};
        const unsigned long dim317Linitializers[] = {
            1UL,1UL,7UL,1UL,31UL,17UL,19UL,225UL,369UL,125UL,421UL,0UL};
        const unsigned long dim318Linitializers[] = {
            1UL,3UL,3UL,11UL,7UL,59UL,115UL,197UL,459UL,469UL,1055UL,0UL};
        const unsigned long dim319Linitializers[] = {
            1UL,3UL,1UL,3UL,27UL,45UL,35UL,131UL,349UL,101UL,411UL,0UL};
        const unsigned long dim320Linitializers[] = {
            1UL,3UL,7UL,11UL,9UL,3UL,67UL,145UL,299UL,253UL,1339UL,0UL};
        const unsigned long dim321Linitializers[] = {
            1UL,3UL,3UL,11UL,9UL,37UL,123UL,229UL,273UL,269UL,515UL,0UL};
        const unsigned long dim322Linitializers[] = {
            1UL,3UL,7UL,15UL,11UL,25UL,75UL,5UL,367UL,217UL,951UL,0UL};
        const unsigned long dim323Linitializers[] = {
            1UL,1UL,3UL,7UL,9UL,23UL,63UL,237UL,385UL,159UL,1273UL,0UL};
        const unsigned long dim324Linitializers[] = {
            1UL,1UL,5UL,11UL,23UL,5UL,55UL,193UL,109UL,865UL,663UL,0UL};
        const unsigned long dim325Linitializers[] = {
            1UL,1UL,7UL,15UL,1UL,57UL,17UL,141UL,51UL,217UL,1259UL,0UL};
        const unsigned long dim326Linitializers[] = {
            1UL,1UL,3UL,3UL,15UL,7UL,89UL,233UL,71UL,329UL,203UL,0UL};
        const unsigned long dim327Linitializers[] = {
            1UL,3UL,7UL,11UL,11UL,1UL,19UL,155UL,89UL,437UL,573UL,0UL};
        const unsigned long dim328Linitializers[] = {
            1UL,3UL,1UL,9UL,27UL,61UL,47UL,109UL,161UL,913UL,1681UL,0UL};
        const unsigned long dim329Linitializers[] = {
            1UL,1UL,7UL,15UL,1UL,33UL,19UL,15UL,23UL,913UL,989UL,0UL};
        const unsigned long dim330Linitializers[] = {
            1UL,3UL,1UL,1UL,25UL,39UL,119UL,193UL,13UL,571UL,157UL,0UL};
        const unsigned long dim331Linitializers[] = {
            1UL,1UL,7UL,13UL,9UL,55UL,59UL,147UL,361UL,935UL,515UL,0UL};
        const unsigned long dim332Linitializers[] = {
            1UL,1UL,1UL,9UL,7UL,59UL,67UL,117UL,71UL,855UL,1493UL,0UL};
        const unsigned long dim333Linitializers[] = {
            1UL,3UL,1UL,3UL,13UL,19UL,57UL,141UL,305UL,275UL,1079UL,0UL};
        const unsigned long dim334Linitializers[] = {
            1UL,1UL,1UL,9UL,17UL,61UL,33UL,7UL,43UL,931UL,781UL,0UL};
        const unsigned long dim335Linitializers[] = {
            1UL,1UL,3UL,1UL,11UL,17UL,21UL,97UL,295UL,277UL,1721UL,0UL};
        const unsigned long dim336Linitializers[] = {
            1UL,3UL,1UL,13UL,15UL,43UL,11UL,241UL,147UL,391UL,1641UL,0UL};
        const unsigned long dim337Linitializers[] = {
            1UL,1UL,1UL,1UL,1UL,19UL,37UL,21UL,255UL,263UL,1571UL,0UL};
        const unsigned long dim338Linitializers[] = {
            1UL,1UL,3UL,3UL,23UL,59UL,89UL,17UL,475UL,303UL,757UL,543UL,0UL};
        const unsigned long dim339Linitializers[] = {
           1UL,3UL,3UL,9UL,11UL,55UL,35UL,159UL,139UL,203UL,1531UL,1825UL,0UL};
        const unsigned long dim340Linitializers[] = {
            1UL,1UL,5UL,3UL,17UL,53UL,51UL,241UL,269UL,949UL,1373UL,325UL,0UL};
        const unsigned long dim341Linitializers[] = {
            1UL,3UL,7UL,7UL,5UL,29UL,91UL,149UL,239UL,193UL,1951UL,2675UL,0UL};
        const unsigned long dim342Linitializers[] = {
            1UL,3UL,5UL,1UL,27UL,33UL,69UL,11UL,51UL,371UL,833UL,2685UL,0UL};
        const unsigned long dim343Linitializers[] = {
            1UL,1UL,1UL,15UL,1UL,17UL,35UL,57UL,171UL,1007UL,449UL,367UL,0UL};
        const unsigned long dim344Linitializers[] = {
            1UL,1UL,1UL,7UL,25UL,61UL,73UL,219UL,379UL,53UL,589UL,4065UL,0UL};
        const unsigned long dim345Linitializers[] = {
            1UL,3UL,5UL,13UL,21UL,29UL,45UL,19UL,163UL,169UL,147UL,597UL,0UL};
        const unsigned long dim346Linitializers[] = {
            1UL,1UL,5UL,11UL,21UL,27UL,7UL,17UL,237UL,591UL,255UL,1235UL,0UL};
        const unsigned long dim347Linitializers[] = {
            1UL,1UL,7UL,7UL,17UL,41UL,69UL,237UL,397UL,173UL,1229UL,2341UL,0UL};
        const unsigned long dim348Linitializers[] = {
            1UL,1UL,3UL,1UL,1UL,33UL,125UL,47UL,11UL,783UL,1323UL,2469UL,0UL};
        const unsigned long dim349Linitializers[] = {
            1UL,3UL,1UL,11UL,3UL,39UL,35UL,133UL,153UL,55UL,1171UL,3165UL,0UL};
        const unsigned long dim350Linitializers[] = {
           1UL,1UL,5UL,11UL,27UL,23UL,103UL,245UL,375UL,753UL,477UL,2165UL,0UL};
        const unsigned long dim351Linitializers[] = {
          1UL,3UL,1UL,15UL,15UL,49UL,127UL,223UL,387UL,771UL,1719UL,1465UL,0UL};
        const unsigned long dim352Linitializers[] = {
            1UL,1UL,1UL,9UL,11UL,9UL,17UL,185UL,239UL,899UL,1273UL,3961UL,0UL};
        const unsigned long dim353Linitializers[] = {
            1UL,1UL,3UL,13UL,11UL,51UL,73UL,81UL,389UL,647UL,1767UL,1215UL,0UL};
        const unsigned long dim354Linitializers[] = {
            1UL,3UL,5UL,15UL,19UL,9UL,69UL,35UL,349UL,977UL,1603UL,1435UL,0UL};
        const unsigned long dim355Linitializers[] = {
            1UL,1UL,1UL,1UL,19UL,59UL,123UL,37UL,41UL,961UL,181UL,1275UL,0UL};
        const unsigned long dim356Linitializers[] = {
            1UL,1UL,1UL,1UL,31UL,29UL,37UL,71UL,205UL,947UL,115UL,3017UL,0UL};
        const unsigned long dim357Linitializers[] = {
            1UL,1UL,7UL,15UL,5UL,37UL,101UL,169UL,221UL,245UL,687UL,195UL,0UL};
        const unsigned long dim358Linitializers[] = {
            1UL,1UL,1UL,1UL,19UL,9UL,125UL,157UL,119UL,283UL,1721UL,743UL,0UL};
        const unsigned long dim359Linitializers[] = {
            1UL,1UL,7UL,3UL,1UL,7UL,61UL,71UL,119UL,257UL,1227UL,2893UL,0UL};
        const unsigned long dim360Linitializers[] = {
            1UL,3UL,3UL,3UL,25UL,41UL,25UL,225UL,31UL,57UL,925UL,2139UL,0UL};


        const unsigned long * const Linitializers[359] = {
            dim02SLinitializers,
            dim03SLinitializers,
            dim04SLinitializers,
            dim05SLinitializers,
            dim06SLinitializers,
            dim07SLinitializers,
            dim08SLinitializers,
            dim09SLinitializers,
            dim10SLinitializers,
            dim11SLinitializers,
            dim12SLinitializers,
            dim13SLinitializers,
            dim14SLinitializers,
            dim15SLinitializers,
            dim16SLinitializers,
            dim17SLinitializers,
            dim18SLinitializers,
            dim19SLinitializers,
            dim20SLinitializers,
            dim21SLinitializers,
            dim22SLinitializers,
            dim23SLinitializers,
            dim24SLinitializers,
            dim25SLinitializers,
            dim26SLinitializers,
            dim27SLinitializers,
            dim28SLinitializers,
            dim29SLinitializers,
            dim30SLinitializers,
            dim31SLinitializers,
            dim32SLinitializers,
            dim33SLinitializers,
            dim34SLinitializers,
            dim35SLinitializers,
            dim36SLinitializers,
            dim37SLinitializers,
            dim38SLinitializers,
            dim39SLinitializers,
            dim40SLinitializers,
            dim041Linitializers,
            dim042Linitializers,
            dim043Linitializers,
            dim044Linitializers,
            dim045Linitializers,
            dim046Linitializers,
            dim047Linitializers,
            dim048Linitializers,
            dim049Linitializers,
            dim050Linitializers,
            dim051Linitializers,
            dim052Linitializers,
            dim053Linitializers,
            dim054Linitializers,
            dim055Linitializers,
            dim056Linitializers,
            dim057Linitializers,
            dim058Linitializers,
            dim059Linitializers,
            dim060Linitializers,
            dim061Linitializers,
            dim062Linitializers,
            dim063Linitializers,
            dim064Linitializers,
            dim065Linitializers,
            dim066Linitializers,
            dim067Linitializers,
            dim068Linitializers,
            dim069Linitializers,
            dim070Linitializers,
            dim071Linitializers,
            dim072Linitializers,
            dim073Linitializers,
            dim074Linitializers,
            dim075Linitializers,
            dim076Linitializers,
            dim077Linitializers,
            dim078Linitializers,
            dim079Linitializers,
            dim080Linitializers,
            dim081Linitializers,
            dim082Linitializers,
            dim083Linitializers,
            dim084Linitializers,
            dim085Linitializers,
            dim086Linitializers,
            dim087Linitializers,
            dim088Linitializers,
            dim089Linitializers,
            dim090Linitializers,
            dim091Linitializers,
            dim092Linitializers,
            dim093Linitializers,
            dim094Linitializers,
            dim095Linitializers,
            dim096Linitializers,
            dim097Linitializers,
            dim098Linitializers,
            dim099Linitializers,
            dim100Linitializers,
            dim101Linitializers,
            dim102Linitializers,
            dim103Linitializers,
            dim104Linitializers,
            dim105Linitializers,
            dim106Linitializers,
            dim107Linitializers,
            dim108Linitializers,
            dim109Linitializers,
            dim110Linitializers,
            dim111Linitializers,
            dim112Linitializers,
            dim113Linitializers,
            dim114Linitializers,
            dim115Linitializers,
            dim116Linitializers,
            dim117Linitializers,
            dim118Linitializers,
            dim119Linitializers,
            dim120Linitializers,
            dim121Linitializers,
            dim122Linitializers,
            dim123Linitializers,
            dim124Linitializers,
            dim125Linitializers,
            dim126Linitializers,
            dim127Linitializers,
            dim128Linitializers,
            dim129Linitializers,
            dim130Linitializers,
            dim131Linitializers,
            dim132Linitializers,
            dim133Linitializers,
            dim134Linitializers,
            dim135Linitializers,
            dim136Linitializers,
            dim137Linitializers,
            dim138Linitializers,
            dim139Linitializers,
            dim140Linitializers,
            dim141Linitializers,
            dim142Linitializers,
            dim143Linitializers,
            dim144Linitializers,
            dim145Linitializers,
            dim146Linitializers,
            dim147Linitializers,
            dim148Linitializers,
            dim149Linitializers,
            dim150Linitializers,
            dim151Linitializers,
            dim152Linitializers,
            dim153Linitializers,
            dim154Linitializers,
            dim155Linitializers,
            dim156Linitializers,
            dim157Linitializers,
            dim158Linitializers,
            dim159Linitializers,
            dim160Linitializers,
            dim161Linitializers,
            dim162Linitializers,
            dim163Linitializers,
            dim164Linitializers,
            dim165Linitializers,
            dim166Linitializers,
            dim167Linitializers,
            dim168Linitializers,
            dim169Linitializers,
            dim170Linitializers,
            dim171Linitializers,
            dim172Linitializers,
            dim173Linitializers,
            dim174Linitializers,
            dim175Linitializers,
            dim176Linitializers,
            dim177Linitializers,
            dim178Linitializers,
            dim179Linitializers,
            dim180Linitializers,
            dim181Linitializers,
            dim182Linitializers,
            dim183Linitializers,
            dim184Linitializers,
            dim185Linitializers,
            dim186Linitializers,
            dim187Linitializers,
            dim188Linitializers,
            dim189Linitializers,
            dim190Linitializers,
            dim191Linitializers,
            dim192Linitializers,
            dim193Linitializers,
            dim194Linitializers,
            dim195Linitializers,
            dim196Linitializers,
            dim197Linitializers,
            dim198Linitializers,
            dim199Linitializers,
            dim200Linitializers,
            dim201Linitializers,
            dim202Linitializers,
            dim203Linitializers,
            dim204Linitializers,
            dim205Linitializers,
            dim206Linitializers,
            dim207Linitializers,
            dim208Linitializers,
            dim209Linitializers,
            dim210Linitializers,
            dim211Linitializers,
            dim212Linitializers,
            dim213Linitializers,
            dim214Linitializers,
            dim215Linitializers,
            dim216Linitializers,
            dim217Linitializers,
            dim218Linitializers,
            dim219Linitializers,
            dim220Linitializers,
            dim221Linitializers,
            dim222Linitializers,
            dim223Linitializers,
            dim224Linitializers,
            dim225Linitializers,
            dim226Linitializers,
            dim227Linitializers,
            dim228Linitializers,
            dim229Linitializers,
            dim230Linitializers,
            dim231Linitializers,
            dim232Linitializers,
            dim233Linitializers,
            dim234Linitializers,
            dim235Linitializers,
            dim236Linitializers,
            dim237Linitializers,
            dim238Linitializers,
            dim239Linitializers,
            dim240Linitializers,
            dim241Linitializers,
            dim242Linitializers,
            dim243Linitializers,
            dim244Linitializers,
            dim245Linitializers,
            dim246Linitializers,
            dim247Linitializers,
            dim248Linitializers,
            dim249Linitializers,
            dim250Linitializers,
            dim251Linitializers,
            dim252Linitializers,
            dim253Linitializers,
            dim254Linitializers,
            dim255Linitializers,
            dim256Linitializers,
            dim257Linitializers,
            dim258Linitializers,
            dim259Linitializers,
            dim260Linitializers,
            dim261Linitializers,
            dim262Linitializers,
            dim263Linitializers,
            dim264Linitializers,
            dim265Linitializers,
            dim266Linitializers,
            dim267Linitializers,
            dim268Linitializers,
            dim269Linitializers,
            dim270Linitializers,
            dim271Linitializers,
            dim272Linitializers,
            dim273Linitializers,
            dim274Linitializers,
            dim275Linitializers,
            dim276Linitializers,
            dim277Linitializers,
            dim278Linitializers,
            dim279Linitializers,
            dim280Linitializers,
            dim281Linitializers,
            dim282Linitializers,
            dim283Linitializers,
            dim284Linitializers,
            dim285Linitializers,
            dim286Linitializers,
            dim287Linitializers,
            dim288Linitializers,
            dim289Linitializers,
            dim290Linitializers,
            dim291Linitializers,
            dim292Linitializers,
            dim293Linitializers,
            dim294Linitializers,
            dim295Linitializers,
            dim296Linitializers,
            dim297Linitializers,
            dim298Linitializers,
            dim299Linitializers,
            dim300Linitializers,
            dim301Linitializers,
            dim302Linitializers,
            dim303Linitializers,
            dim304Linitializers,
            dim305Linitializers,
            dim306Linitializers,
            dim307Linitializers,
            dim308Linitializers,
            dim309Linitializers,
            dim310Linitializers,
            dim311Linitializers,
            dim312Linitializers,
            dim313Linitializers,
            dim314Linitializers,
            dim315Linitializers,
            dim316Linitializers,
            dim317Linitializers,
            dim318Linitializers,
            dim319Linitializers,
            dim320Linitializers,
            dim321Linitializers,
            dim322Linitializers,
            dim323Linitializers,
            dim324Linitializers,
            dim325Linitializers,
            dim326Linitializers,
            dim327Linitializers,
            dim328Linitializers,
            dim329Linitializers,
            dim330Linitializers,
            dim331Linitializers,
            dim332Linitializers,
            dim333Linitializers,
            dim334Linitializers,
            dim335Linitializers,
            dim336Linitializers,
            dim337Linitializers,
            dim338Linitializers,
            dim339Linitializers,
            dim340Linitializers,
            dim341Linitializers,
            dim342Linitializers,
            dim343Linitializers,
            dim344Linitializers,
            dim345Linitializers,
            dim346Linitializers,
            dim347Linitializers,
            dim348Linitializers,
            dim349Linitializers,
            dim350Linitializers,
            dim351Linitializers,
            dim352Linitializers,
            dim353Linitializers,
            dim354Linitializers,
            dim355Linitializers,
            dim356Linitializers,
            dim357Linitializers,
            dim358Linitializers,
            dim359Linitializers,
            dim360Linitializers
        };

            const unsigned long dim1KuoInit[]   =   {   1   ,0 };
            const unsigned long dim2KuoInit[]   =   {   1   ,   1   ,0 };
            const unsigned long dim3KuoInit[]   =   {   1   ,   1   ,   1   ,0 };
            const unsigned long dim4KuoInit[]   =   {   1   ,   3   ,   1   ,0 };
            const unsigned long dim5KuoInit[]   =   {   1   ,   1   ,   7   ,   13  ,0 };
            const unsigned long dim6KuoInit[]   =   {   1   ,   1   ,   3   ,   7   ,0 };
            const unsigned long dim7KuoInit[]   =   {   1   ,   3   ,   1   ,   7   ,   21  ,0 };
            const unsigned long dim8KuoInit[]   =   {   1   ,   3   ,   1   ,   3   ,   9   ,0 };
            const unsigned long dim9KuoInit[]   =   {   1   ,   1   ,   5   ,   9   ,   13  ,0 };
            const unsigned long dim10KuoInit[]  =   {   1   ,   1   ,   3   ,   9   ,   13  ,0 };
            const unsigned long dim11KuoInit[]  =   {   1   ,   1   ,   5   ,   3   ,   7   ,0 };
            const unsigned long dim12KuoInit[]  =   {   1   ,   1   ,   5   ,   7   ,   11  ,0 };
            const unsigned long dim13KuoInit[]  =   {   1   ,   3   ,   3   ,   13  ,   15  ,   43  ,0 };
            const unsigned long dim14KuoInit[]  =   {   1   ,   3   ,   5   ,   11  ,   25  ,   45  ,0 };
            const unsigned long dim15KuoInit[]  =   {   1   ,   1   ,   3   ,   11  ,   3   ,   45  ,0 };
            const unsigned long dim16KuoInit[]  =   {   1   ,   3   ,   5   ,   1   ,   9   ,   21  ,0 };
            const unsigned long dim17KuoInit[]  =   {   1   ,   1   ,   3   ,   13  ,   9   ,   9   ,0 };
            const unsigned long dim18KuoInit[]  =   {   1   ,   3   ,   5   ,   7   ,   17  ,   53  ,0 };
            const unsigned long dim19KuoInit[]  =   {   1   ,   3   ,   1   ,   7   ,   11  ,   51  ,   115 ,0 };
            const unsigned long dim20KuoInit[]  =   {   1   ,   1   ,   7   ,   9   ,   25  ,   35  ,   11  ,0 };
            const unsigned long dim21KuoInit[]  =   {   1   ,   3   ,   1   ,   1   ,   31  ,   5   ,   1   ,0 };
            const unsigned long dim22KuoInit[]  =   {   1   ,   1   ,   5   ,   9   ,   11  ,   1   ,   121 ,0 };
            const unsigned long dim23KuoInit[]  =   {   1   ,   1   ,   1   ,   15  ,   11  ,   59  ,   21  ,0 };
            const unsigned long dim24KuoInit[]  =   {   1   ,   3   ,   1   ,   3   ,   17  ,   49  ,   51  ,0 };
            const unsigned long dim25KuoInit[]  =   {   1   ,   1   ,   5   ,   7   ,   15  ,   25  ,   13  ,0 };
            const unsigned long dim26KuoInit[]  =   {   1   ,   3   ,   7   ,   7   ,   1   ,   45  ,   7   ,0 };
            const unsigned long dim27KuoInit[]  =   {   1   ,   1   ,   5   ,   7   ,   21  ,   21  ,   37  ,0 };
            const unsigned long dim28KuoInit[]  =   {   1   ,   3   ,   1   ,   13  ,   9   ,   49  ,   23  ,0 };
            const unsigned long dim29KuoInit[]  =   {   1   ,   1   ,   1   ,   3   ,   3   ,   35  ,   123 ,0 };
            const unsigned long dim30KuoInit[]  =   {   1   ,   1   ,   7   ,   5   ,   15  ,   47  ,   117 ,0 };
            const unsigned long dim31KuoInit[]  =   {   1   ,   1   ,   3   ,   13  ,   9   ,   23  ,   33  ,0 };
            const unsigned long dim32KuoInit[]  =   {   1   ,   1   ,   1   ,   13  ,   25  ,   23  ,   63  ,0 };
            const unsigned long dim33KuoInit[]  =   {   1   ,   1   ,   5   ,   3   ,   5   ,   13  ,   91  ,0 };
            const unsigned long dim34KuoInit[]  =   {   1   ,   1   ,   5   ,   5   ,   23  ,   7   ,   101 ,0 };
            const unsigned long dim35KuoInit[]  =   {   1   ,   3   ,   1   ,   11  ,   9   ,   61  ,   127 ,0 };
            const unsigned long dim36KuoInit[]  =   {   1   ,   3   ,   1   ,   9   ,   1   ,   57  ,   93  ,0 };
            const unsigned long dim37KuoInit[]  =   {   1   ,   1   ,   1   ,   1   ,   31  ,   31  ,   25  ,   231 ,0 };
            const unsigned long dim38KuoInit[]  =   {   1   ,   3   ,   7   ,   7   ,   29  ,   17  ,   59  ,   215 ,0 };
            const unsigned long dim39KuoInit[]  =   {   1   ,   1   ,   3   ,   15  ,   5   ,   63  ,   117 ,   217 ,0 };
            const unsigned long dim40KuoInit[]  =   {   1   ,   3   ,   1   ,   1   ,   29  ,   5   ,   111 ,   51  ,0 };
            const unsigned long dim41KuoInit[]  =   {   1   ,   1   ,   7   ,   1   ,   17  ,   37  ,   11  ,   53  ,0 };
            const unsigned long dim42KuoInit[]  =   {   1   ,   3   ,   7   ,   15  ,   29  ,   5   ,   3   ,   67  ,0 };
            const unsigned long dim43KuoInit[]  =   {   1   ,   3   ,   7   ,   11  ,   23  ,   27  ,   35  ,   143 ,0 };
            const unsigned long dim44KuoInit[]  =   {   1   ,   1   ,   5   ,   11  ,   11  ,   33  ,   103 ,   179 ,0 };
            const unsigned long dim45KuoInit[]  =   {   1   ,   3   ,   5   ,   7   ,   13  ,   45  ,   87  ,   143 ,0 };
            const unsigned long dim46KuoInit[]  =   {   1   ,   3   ,   1   ,   13  ,   17  ,   17  ,   49  ,   249 ,0 };
            const unsigned long dim47KuoInit[]  =   {   1   ,   1   ,   1   ,   3   ,   17  ,   29  ,   75  ,   143 ,0 };
            const unsigned long dim48KuoInit[]  =   {   1   ,   1   ,   1   ,   3   ,   19  ,   55  ,   65  ,   109 ,0 };
            const unsigned long dim49KuoInit[]  =   {   1   ,   1   ,   7   ,   3   ,   21  ,   29  ,   13  ,   191 ,0 };
            const unsigned long dim50KuoInit[]  =   {   1   ,   1   ,   3   ,   7   ,   7   ,   5   ,   97  ,   179 ,0 };
            const unsigned long dim51KuoInit[]  =   {   1   ,   3   ,   5   ,   13  ,   1   ,   17  ,   113 ,   149 ,0 };
            const unsigned long dim52KuoInit[]  =   {   1   ,   3   ,   3   ,   11  ,   25  ,   13  ,   105 ,   75  ,0 };
            const unsigned long dim53KuoInit[]  =   {   1   ,   3   ,   1   ,   13  ,   29  ,   33  ,   71  ,   117 ,   77  ,0 };
            const unsigned long dim54KuoInit[]  =   {   1   ,   1   ,   1   ,   1   ,   31  ,   41  ,   9   ,   245 ,   205 ,0 };
            const unsigned long dim55KuoInit[]  =   {   1   ,   1   ,   1   ,   15  ,   25  ,   23  ,   111 ,   105 ,   95  ,0 };
            const unsigned long dim56KuoInit[]  =   {   1   ,   1   ,   3   ,   5   ,   3   ,   5   ,   81  ,   251 ,   221 ,0 };
            const unsigned long dim57KuoInit[]  =   {   1   ,   3   ,   5   ,   5   ,   9   ,   53  ,   37  ,   41  ,   509 ,0 };
            const unsigned long dim58KuoInit[]  =   {   1   ,   1   ,   5   ,   15  ,   21  ,   25  ,   77  ,   225 ,   333 ,0 };
            const unsigned long dim59KuoInit[]  =   {   1   ,   3   ,   5   ,   13  ,   3   ,   17  ,   1   ,   101 ,   397 ,0 };
            const unsigned long dim60KuoInit[]  =   {   1   ,   3   ,   5   ,   11  ,   27  ,   53  ,   115 ,   87  ,   47  ,0 };
            const unsigned long dim61KuoInit[]  =   {   1   ,   1   ,   3   ,   11  ,   31  ,   39  ,   21  ,   233 ,   9   ,0 };
            const unsigned long dim62KuoInit[]  =   {   1   ,   1   ,   3   ,   15  ,   17  ,   19  ,   73  ,   147 ,   351 ,0 };
            const unsigned long dim63KuoInit[]  =   {   1   ,   1   ,   3   ,   3   ,   29  ,   17  ,   1   ,   105 ,   293 ,0 };
            const unsigned long dim64KuoInit[]  =   {   1   ,   1   ,   3   ,   15  ,   9   ,   21  ,   103 ,   239 ,   433 ,0 };
            const unsigned long dim65KuoInit[]      =   {   1   ,   1   ,   1   ,   15  ,   17  ,   35  ,   49  ,   7   ,   435 ,0 };
            const unsigned long dim66KuoInit[]      =   {   1   ,   3   ,   7   ,   3   ,   9   ,   51  ,   91  ,   177 ,   255 ,0 };
            const unsigned long dim67KuoInit[]      =   {   1   ,   1   ,   5   ,   5   ,   31  ,   9   ,   39  ,   209 ,   511 ,0 };
            const unsigned long dim68KuoInit[]      =   {   1   ,   1   ,   7   ,   3   ,   11  ,   11  ,   19  ,   227 ,   343 ,0 };
            const unsigned long dim69KuoInit[]      =   {   1   ,   3   ,   7   ,   3   ,   17  ,   13  ,   105 ,   151 ,   225 ,0 };
            const unsigned long dim70KuoInit[]      =   {   1   ,   1   ,   7   ,   11  ,   29  ,   35  ,   105 ,   11  ,   497 ,0 };
            const unsigned long dim71KuoInit[]      =   {   1   ,   3   ,   7   ,   15  ,   5   ,   13  ,   31  ,   201 ,   63  ,0 };
            const unsigned long dim72KuoInit[]      =   {   1   ,   1   ,   5   ,   7   ,   17  ,   31  ,   9   ,   63  ,   167 ,0 };
            const unsigned long dim73KuoInit[]      =   {   1   ,   1   ,   1   ,   13  ,   1   ,   51  ,   9   ,   115 ,   217 ,0 };
            const unsigned long dim74KuoInit[]      =   {   1   ,   1   ,   7   ,   15  ,   23  ,   49  ,   125 ,   163 ,   39  ,0 };
            const unsigned long dim75KuoInit[]      =   {   1   ,   1   ,   3   ,   13  ,   11  ,   45  ,   77  ,   149 ,   173 ,0 };
            const unsigned long dim76KuoInit[]      =   {   1   ,   3   ,   3   ,   3   ,   7   ,   27  ,   43  ,   27  ,   361 ,0 };
            const unsigned long dim77KuoInit[]      =   {   1   ,   3   ,   3   ,   5   ,   5   ,   37  ,   9   ,   51  ,   149 ,0 };
            const unsigned long dim78KuoInit[]      =   {   1   ,   1   ,   1   ,   15  ,   13  ,   13  ,   49  ,   251 ,   385 ,0 };
            const unsigned long dim79KuoInit[]      =   {   1   ,   1   ,   3   ,   7   ,   1   ,   57  ,   27  ,   25  ,   335 ,0 };
            const unsigned long dim80KuoInit[]      =   {   1   ,   3   ,   3   ,   9   ,   31  ,   55  ,   41  ,   143 ,   97  ,0 };
            const unsigned long dim81KuoInit[]      =   {   1   ,   3   ,   3   ,   11  ,   31  ,   33  ,   65  ,   57  ,   113 ,0 };
            const unsigned long dim82KuoInit[]      =   {   1   ,   1   ,   5   ,   13  ,   19  ,   63  ,   5   ,   71  ,   317 ,0 };
            const unsigned long dim83KuoInit[]      =   {   1   ,   1   ,   1   ,   5   ,   29  ,   45  ,   35  ,   107 ,   113 ,0 };
            const unsigned long dim84KuoInit[]      =   {   1   ,   1   ,   3   ,   7   ,   3   ,   31  ,   81  ,   57  ,   439 ,0 };
            const unsigned long dim85KuoInit[]      =   {   1   ,   3   ,   5   ,   5   ,   1   ,   7   ,   3   ,   21  ,   319 ,0 };
            const unsigned long dim86KuoInit[]      =   {   1   ,   3   ,   1   ,   15  ,   27  ,   49  ,   25  ,   247 ,   455 ,0 };
            const unsigned long dim87KuoInit[]      =   {   1   ,   1   ,   3   ,   5   ,   5   ,   43  ,   105 ,   207 ,   271 ,0 };
            const unsigned long dim88KuoInit[]      =   {   1   ,   3   ,   3   ,   7   ,   29  ,   63  ,   25  ,   239 ,   165 ,0 };
            const unsigned long dim89KuoInit[]      =   {   1   ,   3   ,   7   ,   1   ,   19  ,   23  ,   87  ,   23  ,   161 ,0 };
            const unsigned long dim90KuoInit[]      =   {   1   ,   3   ,   7   ,   5   ,   27  ,   1   ,   15  ,   113 ,   85  ,0 };
            const unsigned long dim91KuoInit[]      =   {   1   ,   1   ,   1   ,   13  ,   23  ,   39  ,   103 ,   59  ,   105 ,0 };
            const unsigned long dim92KuoInit[]      =   {   1   ,   3   ,   1   ,   15  ,   9   ,   35  ,   65  ,   225 ,   385 ,0 };
            const unsigned long dim93KuoInit[]      =   {   1   ,   1   ,   5   ,   9   ,   5   ,   63  ,   109 ,   153 ,   25  ,0 };
            const unsigned long dim94KuoInit[]      =   {   1   ,   3   ,   7   ,   11  ,   5   ,   41  ,   103 ,   103 ,   245 ,0 };
            const unsigned long dim95KuoInit[]      =   {   1   ,   1   ,   1   ,   3   ,   25  ,   5   ,   71  ,   203 ,   351 ,0 };
            const unsigned long dim96KuoInit[]      =   {   1   ,   1   ,   5   ,   5   ,   27  ,   49  ,   111 ,   253 ,   265 ,0 };
            const unsigned long dim97KuoInit[]      =   {   1   ,   1   ,   1   ,   11  ,   31  ,   63  ,   91  ,   179 ,   359 ,0 };
            const unsigned long dim98KuoInit[]      =   {   1   ,   3   ,   7   ,   1   ,   9   ,   41  ,   117 ,   69  ,   141 ,0 };
            const unsigned long dim99KuoInit[]      =   {   1   ,   3   ,   1   ,   5   ,   11  ,   25  ,   7   ,   73  ,   477 ,0 };
            const unsigned long dim100KuoInit[]     =   {   1   ,   1   ,   3   ,   1   ,   31  ,   7   ,   51  ,   67  ,   97  ,0 };
            const unsigned long dim101KuoInit[]     =   {   1   ,   3   ,   3   ,   15  ,   31  ,   61  ,   27  ,   137 ,   319 ,   605 ,0 };
            const unsigned long dim102KuoInit[]     =   {   1   ,   3   ,   5   ,   7   ,   29  ,   45  ,   31  ,   211 ,   287 ,   13  ,0 };
            const unsigned long dim103KuoInit[]     =   {   1   ,   3   ,   1   ,   7   ,   25  ,   61  ,   121 ,   91  ,   215 ,   439 ,0 };
            const unsigned long dim104KuoInit[]     =   {   1   ,   1   ,   3   ,   11  ,   31  ,   51  ,   31  ,   151 ,   421 ,   195 ,0 };
            const unsigned long dim105KuoInit[]     =   {   1   ,   1   ,   5   ,   13  ,   17  ,   23  ,   47  ,   9   ,   33  ,   1003    ,0 };
            const unsigned long dim106KuoInit[]     =   {   1   ,   1   ,   1   ,   3   ,   23  ,   5   ,   109 ,   23  ,   125 ,   773 ,0 };
            const unsigned long dim107KuoInit[]     =   {   1   ,   3   ,   5   ,   9   ,   7   ,   41  ,   33  ,   81  ,   277 ,   623 ,0 };
            const unsigned long dim108KuoInit[]     =   {   1   ,   1   ,   1   ,   11  ,   9   ,   33  ,   51  ,   31  ,   181 ,   985 ,0 };
            const unsigned long dim109KuoInit[]     =   {   1   ,   1   ,   7   ,   11  ,   3   ,   37  ,   3   ,   107 ,   405 ,   639 ,0 };
            const unsigned long dim110KuoInit[]     =   {   1   ,   1   ,   5   ,   15  ,   17  ,   7   ,   35  ,   221 ,   255 ,   753 ,0 };
            const unsigned long dim111KuoInit[]     =   {   1   ,   1   ,   7   ,   3   ,   31  ,   11  ,   113 ,   105 ,   275 ,   683 ,0 };
            const unsigned long dim112KuoInit[]     =   {   1   ,   1   ,   3   ,   11  ,   3   ,   39  ,   57  ,   237 ,   189 ,   561 ,0 };
            const unsigned long dim113KuoInit[]     =   {   1   ,   3   ,   3   ,   7   ,   17  ,   37  ,   55  ,   97  ,   499 ,   511 ,0 };
            const unsigned long dim114KuoInit[]     =   {   1   ,   1   ,   7   ,   11  ,   31  ,   27  ,   21  ,   241 ,   127 ,   279 ,0 };
            const unsigned long dim115KuoInit[]     =   {   1   ,   3   ,   1   ,   7   ,   1   ,   19  ,   95  ,   3   ,   511 ,   143 ,0 };
            const unsigned long dim116KuoInit[]     =   {   1   ,   3   ,   5   ,   3   ,   21  ,   39  ,   115 ,   155 ,   67  ,   607 ,0 };
            const unsigned long dim117KuoInit[]     =   {   1   ,   1   ,   7   ,   7   ,   25  ,   37  ,   33  ,   149 ,   89  ,   997 ,0 };
            const unsigned long dim118KuoInit[]     =   {   1   ,   3   ,   1   ,   7   ,   3   ,   7   ,   113 ,   165 ,   235 ,   943 ,0 };
            const unsigned long dim119KuoInit[]     =   {   1   ,   3   ,   7   ,   9   ,   29  ,   5   ,   93  ,   255 ,   323 ,   555 ,0 };
            const unsigned long dim120KuoInit[]     =   {   1   ,   1   ,   1   ,   3   ,   13  ,   45  ,   89  ,   87  ,   29  ,   23  ,0 };
            const unsigned long dim121KuoInit[]     =   {   1   ,   3   ,   3   ,   11  ,   11  ,   29  ,   75  ,   91  ,   211 ,   459 ,0 };
            const unsigned long dim122KuoInit[]     =   {   1   ,   3   ,   5   ,   15  ,   11  ,   43  ,   105 ,   1   ,   227 ,   711 ,0 };
            const unsigned long dim123KuoInit[]     =   {   1   ,   1   ,   5   ,   5   ,   13  ,   33  ,   29  ,   217 ,   511 ,   241 ,0 };
            const unsigned long dim124KuoInit[]     =   {   1   ,   3   ,   7   ,   9   ,   1   ,   47  ,   125 ,   151 ,   375 ,   815 ,0 };
            const unsigned long dim125KuoInit[]     =   {   1   ,   1   ,   1   ,   3   ,   25  ,   29  ,   5   ,   39  ,   133 ,   585 ,0 };
            const unsigned long dim126KuoInit[]     =   {   1   ,   1   ,   5   ,   9   ,   9   ,   31  ,   79  ,   227 ,   395 ,   517 ,0 };
            const unsigned long dim127KuoInit[]     =   {   1   ,   3   ,   3   ,   3   ,   15  ,   3   ,   47  ,   159 ,   233 ,   339 ,0 };
            const unsigned long dim128KuoInit[]     =   {   1   ,   3   ,   5   ,   3   ,   15  ,   13  ,   79  ,   109 ,   187 ,   249 ,0 };
            const unsigned long dim129KuoInit[]     =   {   1   ,   3   ,   1   ,   15  ,   23  ,   39  ,   85  ,   157 ,   323 ,   535 ,0 };
            const unsigned long dim130KuoInit[]     =   {   1   ,   1   ,   1   ,   5   ,   11  ,   25  ,   5   ,   153 ,   19  ,   779 ,0 };
            const unsigned long dim131KuoInit[]     =   {   1   ,   3   ,   5   ,   1   ,   9   ,   43  ,   27  ,   179 ,   407 ,   167 ,0 };
            const unsigned long dim132KuoInit[]     =   {   1   ,   3   ,   5   ,   5   ,   25  ,   47  ,   21  ,   23  ,   57  ,   415 ,0 };
            const unsigned long dim133KuoInit[]     =   {   1   ,   1   ,   1   ,   7   ,   3   ,   29  ,   45  ,   195 ,   349 ,   267 ,0 };
            const unsigned long dim134KuoInit[]     =   {   1   ,   3   ,   5   ,   3   ,   19  ,   13  ,   61  ,   243 ,   125 ,   961 ,0 };
            const unsigned long dim135KuoInit[]     =   {   1   ,   3   ,   7   ,   3   ,   29  ,   49  ,   7   ,   197 ,   387 ,   747 ,0 };
            const unsigned long dim136KuoInit[]     =   {   1   ,   3   ,   7   ,   11  ,   31  ,   31  ,   109 ,   87  ,   345 ,   807 ,0 };
            const unsigned long dim137KuoInit[]     =   {   1   ,   3   ,   1   ,   3   ,   21  ,   57  ,   43  ,   207 ,   221 ,   85  ,0 };
            const unsigned long dim138KuoInit[]     =   {   1   ,   1   ,   3   ,   13  ,   19  ,   19  ,   41  ,   19  ,   509 ,   419 ,0 };
            const unsigned long dim139KuoInit[]     =   {   1   ,   1   ,   3   ,   3   ,   23  ,   17  ,   85  ,   99  ,   421 ,   875 ,0 };
            const unsigned long dim140KuoInit[]     =   {   1   ,   3   ,   1   ,   15  ,   27  ,   43  ,   89  ,   151 ,   39  ,   643 ,0 };
            const unsigned long dim141KuoInit[]     =   {   1   ,   1   ,   1   ,   5   ,   3   ,   59  ,   25  ,   79  ,   319 ,   959 ,0 };
            const unsigned long dim142KuoInit[]     =   {   1   ,   3   ,   5   ,   9   ,   3   ,   27  ,   71  ,   209 ,   195 ,   569 ,0 };
            const unsigned long dim143KuoInit[]     =   {   1   ,   3   ,   3   ,   3   ,   11  ,   47  ,   67  ,   1   ,   221 ,   119 ,0 };
            const unsigned long dim144KuoInit[]     =   {   1   ,   1   ,   5   ,   11  ,   11  ,   39  ,   95  ,   29  ,   7   ,   957 ,0 };
            const unsigned long dim145KuoInit[]     =   {   1   ,   1   ,   7   ,   13  ,   13  ,   57  ,   47  ,   71  ,   165 ,   393 ,0 };
            const unsigned long dim146KuoInit[]     =   {   1   ,   3   ,   5   ,   3   ,   5   ,   47  ,   127 ,   209 ,   377 ,   53  ,0 };
            const unsigned long dim147KuoInit[]     =   {   1   ,   3   ,   1   ,   5   ,   13  ,   51  ,   69  ,   165 ,   477 ,   961 ,0 };
            const unsigned long dim148KuoInit[]     =   {   1   ,   3   ,   5   ,   9   ,   25  ,   37  ,   5   ,   133 ,   117 ,   29  ,0 };
            const unsigned long dim149KuoInit[]     =   {   1   ,   1   ,   1   ,   9   ,   7   ,   21  ,   95  ,   53  ,   205 ,   805 ,0 };
            const unsigned long dim150KuoInit[]     =   {   1   ,   1   ,   5   ,   11  ,   25  ,   39  ,   25  ,   15  ,   497 ,   525 ,0 };
            const unsigned long dim151KuoInit[]     =   {   1   ,   3   ,   5   ,   5   ,   15  ,   37  ,   33  ,   255 ,   97  ,   341 ,0 };
            const unsigned long dim152KuoInit[]     =   {   1   ,   3   ,   1   ,   13  ,   31  ,   63  ,   35  ,   203 ,   415 ,   903 ,0 };
            const unsigned long dim153KuoInit[]     =   {   1   ,   1   ,   5   ,   7   ,   13  ,   59  ,   55  ,   17  ,   103 ,   221 ,0 };
            const unsigned long dim154KuoInit[]     =   {   1   ,   1   ,   5   ,   3   ,   7   ,   37  ,   113 ,   139 ,   269 ,   987 ,0 };
            const unsigned long dim155KuoInit[]     =   {   1   ,   3   ,   3   ,   7   ,   21  ,   3   ,   47  ,   175 ,   327 ,   633 ,0 };
            const unsigned long dim156KuoInit[]     =   {   1   ,   1   ,   1   ,   3   ,   13  ,   61  ,   19  ,   209 ,   39  ,   695 ,0 };
            const unsigned long dim157KuoInit[]     =   {   1   ,   1   ,   1   ,   9   ,   31  ,   61  ,   21  ,   75  ,   445 ,   647 ,0 };
            const unsigned long dim158KuoInit[]     =   {   1   ,   1   ,   5   ,   3   ,   21  ,   1   ,   65  ,   211 ,   405 ,   629 ,0 };
            const unsigned long dim159KuoInit[]     =   {   1   ,   3   ,   7   ,   15  ,   9   ,   25  ,   119 ,   13  ,   465 ,   865 ,0 };
            const unsigned long dim160KuoInit[]     =   {   1   ,   3   ,   1   ,   15  ,   29  ,   19  ,   103 ,   39  ,   37  ,   289 ,0 };
            const unsigned long dim161KuoInit[]     =   {   1   ,   1   ,   1   ,   5   ,   25  ,   23  ,   21  ,   111 ,   455 ,   665 ,   1609    ,0 };
            const unsigned long dim162KuoInit[]     =   {   1   ,   3   ,   3   ,   13  ,   27  ,   29  ,   45  ,   191 ,   385 ,   413 ,   1875    ,0 };
            const unsigned long dim163KuoInit[]     =   {   1   ,   3   ,   7   ,   7   ,   19  ,   17  ,   47  ,   231 ,   337 ,   781 ,   1235    ,0 };
            const unsigned long dim164KuoInit[]     =   {   1   ,   1   ,   5   ,   7   ,   7   ,   25  ,   99  ,   239 ,   501 ,   29  ,   159 ,0 };
            const unsigned long dim165KuoInit[]     =   {   1   ,   1   ,   1   ,   3   ,   5   ,   19  ,   101 ,   147 ,   225 ,   353 ,   1117    ,0 };
            const unsigned long dim166KuoInit[]     =   {   1   ,   3   ,   7   ,   11  ,   17  ,   57  ,   45  ,   73  ,   219 ,   853 ,   1379    ,0 };
            const unsigned long dim167KuoInit[]     =   {   1   ,   3   ,   3   ,   7   ,   11  ,   47  ,   23  ,   67  ,   157 ,   283 ,   867 ,0 };
            const unsigned long dim168KuoInit[]     =   {   1   ,   1   ,   7   ,   7   ,   23  ,   55  ,   71  ,   155 ,   237 ,   927 ,   929 ,0 };
            const unsigned long dim169KuoInit[]     =   {   1   ,   3   ,   3   ,   15  ,   11  ,   19  ,   59  ,   147 ,   373 ,   619 ,   53  ,0 };
            const unsigned long dim170KuoInit[]     =   {   1   ,   1   ,   3   ,   15  ,   13  ,   23  ,   11  ,   33  ,   45  ,   121 ,   241 ,0 };
            const unsigned long dim171KuoInit[]     =   {   1   ,   1   ,   1   ,   9   ,   13  ,   29  ,   3   ,   131 ,   373 ,   91  ,   1207    ,0 };
            const unsigned long dim172KuoInit[]     =   {   1   ,   1   ,   1   ,   15  ,   21  ,   29  ,   59  ,   197 ,   375 ,   637 ,   1275    ,0 };
            const unsigned long dim173KuoInit[]     =   {   1   ,   1   ,   3   ,   11  ,   27  ,   31  ,   87  ,   95  ,   119 ,   249 ,   1913    ,0 };
            const unsigned long dim174KuoInit[]     =   {   1   ,   3   ,   1   ,   5   ,   25  ,   51  ,   1   ,   241 ,   53  ,   883 ,   61  ,0 };
            const unsigned long dim175KuoInit[]     =   {   1   ,   3   ,   1   ,   5   ,   23  ,   37  ,   83  ,   45  ,   287 ,   911 ,   717 ,0 };
            const unsigned long dim176KuoInit[]     =   {   1   ,   3   ,   1   ,   3   ,   9   ,   39  ,   105 ,   129 ,   475 ,   399 ,   1783    ,0 };
            const unsigned long dim177KuoInit[]     =   {   1   ,   3   ,   5   ,   5   ,   23  ,   11  ,   119 ,   205 ,   317 ,   797 ,   605 ,0 };
            const unsigned long dim178KuoInit[]     =   {   1   ,   1   ,   7   ,   5   ,   13  ,   57  ,   19  ,   225 ,   53  ,   987 ,   1463    ,0 };
            const unsigned long dim179KuoInit[]     =   {   1   ,   3   ,   5   ,   11  ,   3   ,   59  ,   1   ,   201 ,   97  ,   1011    ,   1059    ,0 };
            const unsigned long dim180KuoInit[]     =   {   1   ,   3   ,   3   ,   9   ,   29  ,   51  ,   39  ,   199 ,   243 ,   877 ,   883 ,0 };
            const unsigned long dim181KuoInit[]     =   {   1   ,   1   ,   3   ,   5   ,   29  ,   13  ,   97  ,   15  ,   387 ,   435 ,   701 ,0 };
            const unsigned long dim182KuoInit[]     =   {   1   ,   3   ,   3   ,   11  ,   21  ,   41  ,   3   ,   119 ,   385 ,   503 ,   2015    ,0 };
            const unsigned long dim183KuoInit[]     =   {   1   ,   3   ,   3   ,   1   ,   17  ,   9   ,   49  ,   13  ,   309 ,   29  ,   1501    ,0 };
            const unsigned long dim184KuoInit[]     =   {   1   ,   3   ,   1   ,   3   ,   13  ,   29  ,   91  ,   51  ,   225 ,   989 ,   1701    ,0 };
            const unsigned long dim185KuoInit[]     =   {   1   ,   1   ,   7   ,   13  ,   1   ,   55  ,   59  ,   243 ,   399 ,   65  ,   1725    ,0 };
            const unsigned long dim186KuoInit[]     =   {   1   ,   3   ,   7   ,   7   ,   19  ,   29  ,   97  ,   161 ,   195 ,   257 ,   1595    ,0 };
            const unsigned long dim187KuoInit[]     =   {   1   ,   3   ,   7   ,   15  ,   29  ,   49  ,   49  ,   23  ,   157 ,   363 ,   1355    ,0 };
            const unsigned long dim188KuoInit[]     =   {   1   ,   3   ,   1   ,   13  ,   23  ,   39  ,   45  ,   73  ,   147 ,   323 ,   1965    ,0 };
            const unsigned long dim189KuoInit[]     =   {   1   ,   3   ,   1   ,   15  ,   17  ,   27  ,   17  ,   19  ,   371 ,   387 ,   395 ,0 };
            const unsigned long dim190KuoInit[]     =   {   1   ,   1   ,   1   ,   9   ,   19  ,   39  ,   31  ,   3   ,   291 ,   379 ,   605 ,0 };
            const unsigned long dim191KuoInit[]     =   {   1   ,   1   ,   7   ,   1   ,   15  ,   17  ,   47  ,   133 ,   485 ,   145 ,   629 ,0 };
            const unsigned long dim192KuoInit[]     =   {   1   ,   1   ,   1   ,   13  ,   11  ,   11  ,   117 ,   233 ,   85  ,   1   ,   1357    ,0 };
            const unsigned long dim193KuoInit[]     =   {   1   ,   1   ,   7   ,   15  ,   19  ,   51  ,   115 ,   19  ,   333 ,   683 ,   917 ,0 };
            const unsigned long dim194KuoInit[]     =   {   1   ,   3   ,   7   ,   9   ,   27  ,   47  ,   17  ,   209 ,   461 ,   261 ,   761 ,0 };
            const unsigned long dim195KuoInit[]     =   {   1   ,   1   ,   3   ,   1   ,   27  ,   15  ,   43  ,   141 ,   291 ,   363 ,   605 ,0 };
            const unsigned long dim196KuoInit[]     =   {   1   ,   3   ,   7   ,   5   ,   1   ,   3   ,   43  ,   75  ,   115 ,   653 ,   847 ,0 };
            const unsigned long dim197KuoInit[]     =   {   1   ,   1   ,   3   ,   1   ,   11  ,   51  ,   69  ,   119 ,   195 ,   979 ,   1543    ,0 };
            const unsigned long dim198KuoInit[]     =   {   1   ,   1   ,   7   ,   3   ,   11  ,   29  ,   29  ,   59  ,   339 ,   273 ,   855 ,0 };
            const unsigned long dim199KuoInit[]     =   {   1   ,   1   ,   1   ,   1   ,   11  ,   15  ,   59  ,   97  ,   267 ,   647 ,   2037    ,0 };
            const unsigned long dim200KuoInit[]     =   {   1   ,   1   ,   5   ,   3   ,   25  ,   43  ,   41  ,   195 ,   457 ,   155 ,   1707    ,0 };
            const unsigned long dim201KuoInit[]     =   {   1   ,   3   ,   3   ,   15  ,   11  ,   47  ,   43  ,   15  ,   505 ,   761 ,   1259    ,0 };
            const unsigned long dim202KuoInit[]     =   {   1   ,   3   ,   5   ,   15  ,   27  ,   29  ,   55  ,   109 ,   187 ,   545 ,   1177    ,0 };
            const unsigned long dim203KuoInit[]     =   {   1   ,   3   ,   1   ,   5   ,   29  ,   25  ,   89  ,   11  ,   75  ,   771 ,   1437    ,0 };
            const unsigned long dim204KuoInit[]     =   {   1   ,   3   ,   5   ,   11  ,   19  ,   17  ,   27  ,   59  ,   231 ,   35  ,   391 ,0 };
            const unsigned long dim205KuoInit[]     =   {   1   ,   1   ,   7   ,   9   ,   3   ,   17  ,   31  ,   81  ,   317 ,   973 ,   1913    ,0 };
            const unsigned long dim206KuoInit[]     =   {   1   ,   1   ,   1   ,   7   ,   3   ,   21  ,   113 ,   41  ,   335 ,   789 ,   1195    ,0 };
            const unsigned long dim207KuoInit[]     =   {   1   ,   1   ,   7   ,   1   ,   13  ,   61  ,   41  ,   117 ,   301 ,   707 ,   693 ,0 };
            const unsigned long dim208KuoInit[]     =   {   1   ,   1   ,   1   ,   7   ,   3   ,   1   ,   79  ,   145 ,   137 ,   307 ,   205 ,0 };
            const unsigned long dim209KuoInit[]     =   {   1   ,   3   ,   7   ,   1   ,   27  ,   49  ,   83  ,   127 ,   265 ,   669 ,   87  ,0 };
            const unsigned long dim210KuoInit[]     =   {   1   ,   1   ,   3   ,   15  ,   21  ,   31  ,   115 ,   115 ,   343 ,   55  ,   1049    ,0 };
            const unsigned long dim211KuoInit[]     =   {   1   ,   1   ,   5   ,   9   ,   7   ,   25  ,   43  ,   255 ,   363 ,   123 ,   299 ,0 };
            const unsigned long dim212KuoInit[]     =   {   1   ,   3   ,   5   ,   13  ,   1   ,   9   ,   79  ,   245 ,   91  ,   233 ,   1703    ,0 };
            const unsigned long dim213KuoInit[]     =   {   1   ,   3   ,   5   ,   1   ,   17  ,   39  ,   81  ,   147 ,   147 ,   335 ,   819 ,0 };
            const unsigned long dim214KuoInit[]     =   {   1   ,   3   ,   7   ,   3   ,   31  ,   63  ,   55  ,   53  ,   389 ,   307 ,   451 ,0 };
            const unsigned long dim215KuoInit[]     =   {   1   ,   1   ,   7   ,   7   ,   23  ,   41  ,   125 ,   147 ,   323 ,   7   ,   1211    ,0 };
            const unsigned long dim216KuoInit[]     =   {   1   ,   1   ,   1   ,   13  ,   23  ,   9   ,   35  ,   177 ,   165 ,   315 ,   1885    ,0 };
            const unsigned long dim217KuoInit[]     =   {   1   ,   3   ,   3   ,   7   ,   11  ,   39  ,   99  ,   197 ,   405 ,   377 ,   555 ,0 };
            const unsigned long dim218KuoInit[]     =   {   1   ,   3   ,   3   ,   11  ,   15  ,   39  ,   125 ,   59  ,   323 ,   377 ,   325 ,0 };
            const unsigned long dim219KuoInit[]     =   {   1   ,   3   ,   5   ,   9   ,   7   ,   45  ,   75  ,   143 ,   265 ,   161 ,   1701    ,0 };
            const unsigned long dim220KuoInit[]     =   {   1   ,   1   ,   7   ,   5   ,   31  ,   21  ,   35  ,   1   ,   511 ,   585 ,   923 ,0 };
            const unsigned long dim221KuoInit[]     =   {   1   ,   3   ,   1   ,   5   ,   3   ,   31  ,   65  ,   33  ,   199 ,   205 ,   1673    ,0 };
            const unsigned long dim222KuoInit[]     =   {   1   ,   3   ,   7   ,   9   ,   17  ,   61  ,   123 ,   33  ,   405 ,   235 ,   645 ,0 };
            const unsigned long dim223KuoInit[]     =   {   1   ,   3   ,   3   ,   15  ,   27  ,   1   ,   35  ,   165 ,   509 ,   681 ,   621 ,0 };
            const unsigned long dim224KuoInit[]     =   {   1   ,   3   ,   5   ,   15  ,   19  ,   13  ,   49  ,   85  ,   5   ,   661 ,   805 ,0 };
            const unsigned long dim225KuoInit[]     =   {   1   ,   3   ,   7   ,   3   ,   3   ,   51  ,   77  ,   167 ,   461 ,   731 ,   1681    ,0 };
            const unsigned long dim226KuoInit[]     =   {   1   ,   1   ,   3   ,   5   ,   21  ,   51  ,   125 ,   65  ,   41  ,   613 ,   1009    ,0 };
            const unsigned long dim227KuoInit[]     =   {   1   ,   3   ,   7   ,   11  ,   17  ,   39  ,   127 ,   31  ,   423 ,   87  ,   1413    ,0 };
            const unsigned long dim228KuoInit[]     =   {   1   ,   3   ,   1   ,   5   ,   11  ,   41  ,   25  ,   49  ,   169 ,   637 ,   1599    ,0 };
            const unsigned long dim229KuoInit[]     =   {   1   ,   3   ,   5   ,   13  ,   31  ,   27  ,   75  ,   65  ,   391 ,   857 ,   1491    ,0 };
            const unsigned long dim230KuoInit[]     =   {   1   ,   1   ,   1   ,   13  ,   5   ,   61  ,   65  ,   183 ,   373 ,   161 ,   235 ,0 };
            const unsigned long dim231KuoInit[]     =   {   1   ,   3   ,   1   ,   1   ,   27  ,   49  ,   47  ,   55  ,   503 ,   683 ,   301 ,0 };
            const unsigned long dim232KuoInit[]     =   {   1   ,   1   ,   7   ,   11  ,   17  ,   31  ,   113 ,   255 ,   281 ,   341 ,   2033    ,0 };
            const unsigned long dim233KuoInit[]     =   {   1   ,   3   ,   3   ,   13  ,   27  ,   27  ,   27  ,   225 ,   161 ,   187 ,   485 ,0 };
            const unsigned long dim234KuoInit[]     =   {   1   ,   1   ,   1   ,   3   ,   21  ,   27  ,   105 ,   215 ,   241 ,   211 ,   899 ,0 };
            const unsigned long dim235KuoInit[]     =   {   1   ,   3   ,   5   ,   5   ,   11  ,   37  ,   1   ,   7   ,   335 ,   221 ,   757 ,0 };
            const unsigned long dim236KuoInit[]     =   {   1   ,   1   ,   5   ,   7   ,   21  ,   55  ,   61  ,   99  ,   23  ,   957 ,   35  ,0 };
            const unsigned long dim237KuoInit[]     =   {   1   ,   3   ,   5   ,   5   ,   13  ,   53  ,   31  ,   121 ,   111 ,   47  ,   1491    ,0 };
            const unsigned long dim238KuoInit[]     =   {   1   ,   1   ,   5   ,   13  ,   17  ,   5   ,   11  ,   179 ,   357 ,   1003    ,   1179    ,0 };
            const unsigned long dim239KuoInit[]     =   {   1   ,   1   ,   5   ,   13  ,   29  ,   9   ,   3   ,   91  ,   471 ,   369 ,   323 ,0 };
            const unsigned long dim240KuoInit[]     =   {   1   ,   3   ,   7   ,   5   ,   7   ,   25  ,   109 ,   43  ,   133 ,   261 ,   1623    ,0 };
            const unsigned long dim241KuoInit[]     =   {   1   ,   3   ,   7   ,   15  ,   13  ,   33  ,   125 ,   81  ,   491 ,   331 ,   1243    ,0 };
            const unsigned long dim242KuoInit[]     =   {   1   ,   3   ,   3   ,   11  ,   15  ,   21  ,   57  ,   63  ,   55  ,   341 ,   1717    ,0 };
            const unsigned long dim243KuoInit[]     =   {   1   ,   3   ,   1   ,   1   ,   31  ,   13  ,   67  ,   89  ,   459 ,   879 ,   863 ,0 };
            const unsigned long dim244KuoInit[]     =   {   1   ,   3   ,   7   ,   15  ,   7   ,   21  ,   117 ,   229 ,   467 ,   227 ,   1463    ,0 };
            const unsigned long dim245KuoInit[]     =   {   1   ,   1   ,   1   ,   7   ,   23  ,   43  ,   53  ,   79  ,   3   ,   905 ,   1049    ,0 };
            const unsigned long dim246KuoInit[]     =   {   1   ,   3   ,   7   ,   1   ,   19  ,   1   ,   61  ,   15  ,   377 ,   579 ,   145 ,0 };
            const unsigned long dim247KuoInit[]     =   {   1   ,   3   ,   1   ,   15  ,   27  ,   25  ,   61  ,   171 ,   425 ,   631 ,   1343    ,0 };
            const unsigned long dim248KuoInit[]     =   {   1   ,   3   ,   7   ,   5   ,   15  ,   31  ,   69  ,   5   ,   217 ,   809 ,   647 ,0 };
            const unsigned long dim249KuoInit[]     =   {   1   ,   1   ,   5   ,   3   ,   5   ,   35  ,   101 ,   21  ,   33  ,   757 ,   559 ,0 };
            const unsigned long dim250KuoInit[]     =   {   1   ,   3   ,   5   ,   13  ,   3   ,   7   ,   101 ,   93  ,   31  ,   315 ,   1897    ,0 };
            const unsigned long dim251KuoInit[]     =   {   1   ,   1   ,   3   ,   1   ,   21  ,   31  ,   37  ,   127 ,   211 ,   429 ,   1019    ,0 };
            const unsigned long dim252KuoInit[]     =   {   1   ,   1   ,   3   ,   9   ,   11  ,   27  ,   113 ,   203 ,   115 ,   765 ,   759 ,0 };
            const unsigned long dim253KuoInit[]     =   {   1   ,   1   ,   3   ,   13  ,   7   ,   21  ,   89  ,   99  ,   469 ,   729 ,   683 ,0 };
            const unsigned long dim254KuoInit[]     =   {   1   ,   1   ,   1   ,   5   ,   27  ,   17  ,   43  ,   177 ,   225 ,   405 ,   1929    ,0 };
            const unsigned long dim255KuoInit[]     =   {   1   ,   3   ,   1   ,   7   ,   21  ,   55  ,   69  ,   41  ,   309 ,   487 ,   681 ,0 };
            const unsigned long dim256KuoInit[]     =   {   1   ,   3   ,   5   ,   11  ,   19  ,   21  ,   61  ,   163 ,   337 ,   61  ,   301 ,0 };
            const unsigned long dim257KuoInit[]     =   {   1   ,   3   ,   1   ,   7   ,   21  ,   61  ,   67  ,   219 ,   281 ,   21  ,   1271    ,0 };
            const unsigned long dim258KuoInit[]     =   {   1   ,   1   ,   5   ,   11  ,   5   ,   25  ,   51  ,   207 ,   235 ,   289 ,   395 ,0 };
            const unsigned long dim259KuoInit[]     =   {   1   ,   1   ,   3   ,   3   ,   17  ,   61  ,   5   ,   179 ,   457 ,   941 ,   1577    ,0 };
            const unsigned long dim260KuoInit[]     =   {   1   ,   1   ,   7   ,   7   ,   17  ,   23  ,   127 ,   35  ,   407 ,   545 ,   169 ,0 };
            const unsigned long dim261KuoInit[]     =   {   1   ,   3   ,   7   ,   15  ,   21  ,   19  ,   123 ,   145 ,   97  ,   263 ,   1881    ,0 };
            const unsigned long dim262KuoInit[]     =   {   1   ,   3   ,   7   ,   11  ,   25  ,   49  ,   117 ,   165 ,   79  ,   137 ,   1311    ,0 };
            const unsigned long dim263KuoInit[]     =   {   1   ,   3   ,   1   ,   11  ,   3   ,   17  ,   9   ,   177 ,   367 ,   863 ,   1255    ,0 };
            const unsigned long dim264KuoInit[]     =   {   1   ,   1   ,   5   ,   13  ,   9   ,   1   ,   31  ,   3   ,   69  ,   1009    ,   1091    ,0 };
            const unsigned long dim265KuoInit[]     =   {   1   ,   3   ,   7   ,   15  ,   25  ,   41  ,   13  ,   3   ,   319 ,   131 ,   493 ,0 };
            const unsigned long dim266KuoInit[]     =   {   1   ,   1   ,   1   ,   15  ,   9   ,   45  ,   25  ,   21  ,   405 ,   61  ,   1167    ,0 };
            const unsigned long dim267KuoInit[]     =   {   1   ,   1   ,   1   ,   1   ,   31  ,   3   ,   95  ,   113 ,   475 ,   209 ,   1473    ,0 };
            const unsigned long dim268KuoInit[]     =   {   1   ,   1   ,   5   ,   5   ,   27  ,   9   ,   95  ,   121 ,   193 ,   481 ,   1785    ,0 };
            const unsigned long dim269KuoInit[]     =   {   1   ,   1   ,   3   ,   13  ,   1   ,   13  ,   57  ,   167 ,   273 ,   491 ,   867 ,0 };
            const unsigned long dim270KuoInit[]     =   {   1   ,   1   ,   7   ,   1   ,   9   ,   27  ,   123 ,   65  ,   407 ,   609 ,   1979    ,0 };
            const unsigned long dim271KuoInit[]     =   {   1   ,   3   ,   1   ,   15  ,   5   ,   5   ,   89  ,   17  ,   207 ,   249 ,   1725    ,0 };
            const unsigned long dim272KuoInit[]     =   {   1   ,   3   ,   5   ,   11  ,   11  ,   15  ,   91  ,   33  ,   113 ,   531 ,   831 ,0 };
            const unsigned long dim273KuoInit[]     =   {   1   ,   1   ,   5   ,   9   ,   27  ,   51  ,   17  ,   141 ,   215 ,   171 ,   455 ,0 };
            const unsigned long dim274KuoInit[]     =   {   1   ,   1   ,   5   ,   9   ,   11  ,   25  ,   115 ,   5   ,   173 ,   209 ,   1455    ,0 };
            const unsigned long dim275KuoInit[]     =   {   1   ,   1   ,   7   ,   7   ,   13  ,   51  ,   21  ,   79  ,   357 ,   423 ,   1855    ,0 };
            const unsigned long dim276KuoInit[]     =   {   1   ,   1   ,   7   ,   1   ,   13  ,   37  ,   123 ,   27  ,   477 ,   1023    ,   1235    ,0 };
            const unsigned long dim277KuoInit[]     =   {   1   ,   1   ,   7   ,   13  ,   25  ,   43  ,   123 ,   41  ,   199 ,   779 ,   1691    ,0 };
            const unsigned long dim278KuoInit[]     =   {   1   ,   3   ,   3   ,   9   ,   7   ,   27  ,   81  ,   147 ,   465 ,   135 ,   829 ,0 };
            const unsigned long dim279KuoInit[]     =   {   1   ,   1   ,   7   ,   11  ,   17  ,   5   ,   25  ,   177 ,   377 ,   629 ,   915 ,0 };
            const unsigned long dim280KuoInit[]     =   {   1   ,   1   ,   3   ,   15  ,   11  ,   19  ,   45  ,   229 ,   409 ,   187 ,   507 ,0 };
            const unsigned long dim281KuoInit[]     =   {   1   ,   1   ,   7   ,   5   ,   7   ,   49  ,   21  ,   49  ,   395 ,   185 ,   391 ,0 };
            const unsigned long dim282KuoInit[]     =   {   1   ,   3   ,   1   ,   1   ,   1   ,   37  ,   123 ,   255 ,   223 ,   473 ,   1011    ,0 };
            const unsigned long dim283KuoInit[]     =   {   1   ,   3   ,   3   ,   13  ,   3   ,   7   ,   113 ,   183 ,   457 ,   561 ,   1427    ,0 };
            const unsigned long dim284KuoInit[]     =   {   1   ,   1   ,   1   ,   11  ,   27  ,   39  ,   125 ,   241 ,   279 ,   905 ,   1907    ,0 };
            const unsigned long dim285KuoInit[]     =   {   1   ,   3   ,   1   ,   15  ,   19  ,   19  ,   9   ,   233 ,   39  ,   701 ,   939 ,0 };
            const unsigned long dim286KuoInit[]     =   {   1   ,   3   ,   5   ,   11  ,   15  ,   51  ,   1   ,   25  ,   461 ,   671 ,   611 ,0 };
            const unsigned long dim287KuoInit[]     =   {   1   ,   3   ,   1   ,   3   ,   23  ,   39  ,   23  ,   191 ,   219 ,   911 ,   1247    ,0 };
            const unsigned long dim288KuoInit[]     =   {   1   ,   3   ,   1   ,   3   ,   1   ,   27  ,   91  ,   97  ,   55  ,   863 ,   805 ,0 };
            const unsigned long dim289KuoInit[]     =   {   1   ,   3   ,   7   ,   7   ,   5   ,   63  ,   81  ,   33  ,   383 ,   645 ,   841 ,0 };
            const unsigned long dim290KuoInit[]     =   {   1   ,   3   ,   7   ,   5   ,   27  ,   39  ,   33  ,   25  ,   185 ,   799 ,   1109    ,0 };
            const unsigned long dim291KuoInit[]     =   {   1   ,   3   ,   1   ,   11  ,   11  ,   61  ,   109 ,   209 ,   345 ,   465 ,   935 ,0 };
            const unsigned long dim292KuoInit[]     =   {   1   ,   3   ,   5   ,   13  ,   19  ,   55  ,   117 ,   237 ,   455 ,   953 ,   905 ,0 };
            const unsigned long dim293KuoInit[]     =   {   1   ,   3   ,   7   ,   9   ,   9   ,   25  ,   77  ,   37  ,   111 ,   357 ,   131 ,0 };
            const unsigned long dim294KuoInit[]     =   {   1   ,   1   ,   3   ,   13  ,   1   ,   51  ,   27  ,   235 ,   349 ,   517 ,   653 ,0 };
            const unsigned long dim295KuoInit[]     =   {   1   ,   3   ,   7   ,   3   ,   27  ,   43  ,   23  ,   129 ,   455 ,   431 ,   1999    ,0 };
            const unsigned long dim296KuoInit[]     =   {   1   ,   3   ,   7   ,   13  ,   11  ,   3   ,   19  ,   55  ,   261 ,   217 ,   535 ,0 };
            const unsigned long dim297KuoInit[]     =   {   1   ,   3   ,   7   ,   7   ,   23  ,   53  ,   33  ,   237 ,   73  ,   421 ,   1439    ,0 };
            const unsigned long dim298KuoInit[]     =   {   1   ,   1   ,   5   ,   5   ,   19  ,   7   ,   9   ,   201 ,   339 ,   33  ,   2019    ,0 };
            const unsigned long dim299KuoInit[]     =   {   1   ,   3   ,   5   ,   7   ,   1   ,   3   ,   9   ,   79  ,   265 ,   125 ,   1063    ,0 };
            const unsigned long dim300KuoInit[]     =   {   1   ,   1   ,   7   ,   11  ,   31  ,   61  ,   83  ,   71  ,   93  ,   393 ,   851 ,0 };
            const unsigned long dim301KuoInit[]     =   {   1   ,   3   ,   1   ,   13  ,   13  ,   45  ,   65  ,   89  ,   87  ,   657 ,   1635    ,0 };
            const unsigned long dim302KuoInit[]     =   {   1   ,   3   ,   5   ,   7   ,   21  ,   3   ,   27  ,   251 ,   287 ,   305 ,   1143    ,0 };
            const unsigned long dim303KuoInit[]     =   {   1   ,   1   ,   3   ,   13  ,   5   ,   45  ,   35  ,   89  ,   61  ,   347 ,   1349    ,0 };
            const unsigned long dim304KuoInit[]     =   {   1   ,   1   ,   7   ,   1   ,   7   ,   61  ,   19  ,   33  ,   11  ,   845 ,   839 ,0 };
            const unsigned long dim305KuoInit[]     =   {   1   ,   3   ,   5   ,   15  ,   3   ,   1   ,   127 ,   245 ,   423 ,   57  ,   865 ,0 };
            const unsigned long dim306KuoInit[]     =   {   1   ,   3   ,   5   ,   7   ,   31  ,   33  ,   53  ,   109 ,   397 ,   705 ,   307 ,0 };
            const unsigned long dim307KuoInit[]     =   {   1   ,   3   ,   7   ,   7   ,   11  ,   51  ,   53  ,   75  ,   325 ,   61  ,   2035    ,0 };
            const unsigned long dim308KuoInit[]     =   {   1   ,   3   ,   3   ,   9   ,   23  ,   37  ,   63  ,   115 ,   377 ,   741 ,   1181    ,0 };
            const unsigned long dim309KuoInit[]     =   {   1   ,   1   ,   1   ,   13  ,   13  ,   7   ,   77  ,   171 ,   129 ,   615 ,   157 ,0 };
            const unsigned long dim310KuoInit[]     =   {   1   ,   1   ,   5   ,   5   ,   9   ,   41  ,   59  ,   215 ,   275 ,   969 ,   31  ,0 };
            const unsigned long dim311KuoInit[]     =   {   1   ,   1   ,   3   ,   1   ,   27  ,   1   ,   33  ,   111 ,   459 ,   503 ,   1703    ,0 };
            const unsigned long dim312KuoInit[]     =   {   1   ,   3   ,   3   ,   9   ,   11  ,   41  ,   27  ,   213 ,   67  ,   203 ,   497 ,0 };
            const unsigned long dim313KuoInit[]     =   {   1   ,   3   ,   3   ,   3   ,   1   ,   35  ,   39  ,   69  ,   457 ,   765 ,   903 ,0 };
            const unsigned long dim314KuoInit[]     =   {   1   ,   3   ,   5   ,   11  ,   11  ,   31  ,   31  ,   195 ,   481 ,   831 ,   1727    ,0 };
            const unsigned long dim315KuoInit[]     =   {   1   ,   3   ,   1   ,   1   ,   9   ,   39  ,   11  ,   179 ,   241 ,   945 ,   439 ,0 };
            const unsigned long dim316KuoInit[]     =   {   1   ,   1   ,   5   ,   13  ,   1   ,   29  ,   37  ,   113 ,   193 ,   229 ,   1639    ,0 };
            const unsigned long dim317KuoInit[]     =   {   1   ,   3   ,   5   ,   7   ,   9   ,   27  ,   75  ,   105 ,   449 ,   631 ,   2025    ,0 };
            const unsigned long dim318KuoInit[]     =   {   1   ,   3   ,   7   ,   11  ,   7   ,   9   ,   73  ,   201 ,   273 ,   915 ,   729 ,0 };
            const unsigned long dim319KuoInit[]     =   {   1   ,   3   ,   3   ,   15  ,   3   ,   41  ,   61  ,   5   ,   475 ,   193 ,   699 ,0 };
            const unsigned long dim320KuoInit[]     =   {   1   ,   1   ,   3   ,   3   ,   13  ,   57  ,   51  ,   201 ,   313 ,   877 ,   507 ,0 };
            const unsigned long dim321KuoInit[]     =   {   1   ,   1   ,   7   ,   9   ,   7   ,   3   ,   73  ,   103 ,   113 ,   719 ,   1747    ,0 };
            const unsigned long dim322KuoInit[]     =   {   1   ,   1   ,   7   ,   1   ,   1   ,   25  ,   77  ,   249 ,   185 ,   583 ,   123 ,0 };
            const unsigned long dim323KuoInit[]     =   {   1   ,   1   ,   5   ,   11  ,   25  ,   49  ,   17  ,   45  ,   161 ,   647 ,   1685    ,0 };
            const unsigned long dim324KuoInit[]     =   {   1   ,   3   ,   1   ,   15  ,   15  ,   49  ,   35  ,   59  ,   47  ,   523 ,   1961    ,0 };
            const unsigned long dim325KuoInit[]     =   {   1   ,   1   ,   5   ,   5   ,   19  ,   11  ,   105 ,   33  ,   333 ,   451 ,   45  ,0 };
            const unsigned long dim326KuoInit[]     =   {   1   ,   3   ,   1   ,   7   ,   21  ,   1   ,   1   ,   15  ,   465 ,   731 ,   447 ,0 };
            const unsigned long dim327KuoInit[]     =   {   1   ,   3   ,   5   ,   7   ,   3   ,   9   ,   31  ,   205 ,   479 ,   921 ,   1705    ,0 };
            const unsigned long dim328KuoInit[]     =   {   1   ,   1   ,   7   ,   13  ,   7   ,   29  ,   55  ,   1   ,   293 ,   135 ,   1459    ,0 };
            const unsigned long dim329KuoInit[]     =   {   1   ,   1   ,   7   ,   5   ,   19  ,   17  ,   7   ,   217 ,   133 ,   959 ,   811 ,0 };
            const unsigned long dim330KuoInit[]     =   {   1   ,   1   ,   1   ,   15  ,   3   ,   37  ,   55  ,   223 ,   253 ,   535 ,   1039    ,0 };
            const unsigned long dim331KuoInit[]     =   {   1   ,   3   ,   7   ,   3   ,   19  ,   47  ,   125 ,   117 ,   473 ,   257 ,   1215    ,0 };
            const unsigned long dim332KuoInit[]     =   {   1   ,   1   ,   3   ,   5   ,   5   ,   1   ,   87  ,   95  ,   189 ,   431 ,   1355    ,0 };
            const unsigned long dim333KuoInit[]     =   {   1   ,   1   ,   5   ,   11  ,   17  ,   5   ,   63  ,   15  ,   153 ,   325 ,   259 ,0 };
            const unsigned long dim334KuoInit[]     =   {   1   ,   1   ,   3   ,   11  ,   9   ,   41  ,   101 ,   21  ,   161 ,   653 ,   1091    ,0 };
            const unsigned long dim335KuoInit[]     =   {   1   ,   3   ,   3   ,   13  ,   5   ,   19  ,   37  ,   223 ,   469 ,   953 ,   15  ,0 };
            const unsigned long dim336KuoInit[]     =   {   1   ,   1   ,   1   ,   5   ,   3   ,   33  ,   7   ,   63  ,   111 ,   685 ,   381 ,0 };
            const unsigned long dim337KuoInit[]     =   {   1   ,   1   ,   3   ,   11  ,   31  ,   27  ,   49  ,   119 ,   277 ,   303 ,   1719    ,   909 ,0 };
            const unsigned long dim338KuoInit[]     =   {   1   ,   3   ,   1   ,   11  ,   27  ,   13  ,   9   ,   107 ,   251 ,   759 ,   1641    ,   2177    ,0 };
            const unsigned long dim339KuoInit[]     =   {   1   ,   3   ,   3   ,   3   ,   3   ,   57  ,   33  ,   169 ,   239 ,   357 ,   1519    ,   997 ,0 };
            const unsigned long dim340KuoInit[]     =   {   1   ,   3   ,   3   ,   3   ,   11  ,   29  ,   53  ,   109 ,   411 ,   775 ,   1647    ,   819 ,0 };
            const unsigned long dim341KuoInit[]     =   {   1   ,   1   ,   5   ,   13  ,   9   ,   3   ,   15  ,   219 ,   483 ,   513 ,   1319    ,   965 ,0 };
            const unsigned long dim342KuoInit[]     =   {   1   ,   1   ,   1   ,   5   ,   3   ,   41  ,   9   ,   53  ,   303 ,   349 ,   1071    ,   681 ,0 };
            const unsigned long dim343KuoInit[]     =   {   1   ,   3   ,   5   ,   15  ,   27  ,   15  ,   107 ,   211 ,   317 ,   283 ,   1317    ,   2661    ,0 };
            const unsigned long dim344KuoInit[]     =   {   1   ,   3   ,   3   ,   1   ,   13  ,   41  ,   29  ,   17  ,   91  ,   985 ,   327 ,   1637    ,0 };
            const unsigned long dim345KuoInit[]     =   {   1   ,   1   ,   3   ,   15  ,   27  ,   3   ,   127 ,   233 ,   177 ,   877 ,   1467    ,   3951    ,0 };
            const unsigned long dim346KuoInit[]     =   {   1   ,   3   ,   1   ,   5   ,   1   ,   51  ,   5   ,   195 ,   501 ,   859 ,   905 ,   3163    ,0 };
            const unsigned long dim347KuoInit[]     =   {   1   ,   1   ,   1   ,   5   ,   1   ,   25  ,   5   ,   205 ,   65  ,   435 ,   1645    ,   4013    ,0 };
            const unsigned long dim348KuoInit[]     =   {   1   ,   3   ,   5   ,   13  ,   29  ,   57  ,   101 ,   59  ,   289 ,   701 ,   1363    ,   961 ,0 };
            const unsigned long dim349KuoInit[]     =   {   1   ,   3   ,   7   ,   5   ,   1   ,   61  ,   33  ,   117 ,   311 ,   395 ,   1531    ,   2491    ,0 };
            const unsigned long dim350KuoInit[]     =   {   1   ,   1   ,   1   ,   1   ,   5   ,   35  ,   15  ,   59  ,   201 ,   195 ,   1423    ,   2195    ,0 };
            const unsigned long dim351KuoInit[]     =   {   1   ,   1   ,   7   ,   7   ,   13  ,   29  ,   37  ,   109 ,   481 ,   573 ,   1347    ,   3335    ,0 };
            const unsigned long dim352KuoInit[]     =   {   1   ,   1   ,   5   ,   5   ,   25  ,   1   ,   63  ,   53  ,   419 ,   475 ,   1527    ,   3327    ,0 };
            const unsigned long dim353KuoInit[]     =   {   1   ,   1   ,   3   ,   1   ,   27  ,   55  ,   75  ,   225 ,   379 ,   471 ,   211 ,   1655    ,0 };
            const unsigned long dim354KuoInit[]     =   {   1   ,   1   ,   7   ,   15  ,   13  ,   61  ,   111 ,   95  ,   457 ,   801 ,   1865    ,   2529    ,0 };
            const unsigned long dim355KuoInit[]     =   {   1   ,   3   ,   5   ,   5   ,   21  ,   3   ,   31  ,   97  ,   3   ,   495 ,   1953    ,   145 ,0 };
            const unsigned long dim356KuoInit[]     =   {   1   ,   3   ,   3   ,   5   ,   13  ,   27  ,   31  ,   143 ,   133 ,   357 ,   1975    ,   1077    ,0 };
            const unsigned long dim357KuoInit[]     =   {   1   ,   3   ,   3   ,   15  ,   19  ,   19  ,   23  ,   221 ,   9   ,   783 ,   1403    ,   4003    ,0 };
            const unsigned long dim358KuoInit[]     =   {   1   ,   3   ,   7   ,   3   ,   5   ,   53  ,   63  ,   35  ,   209 ,   221 ,   1517    ,   1915    ,0 };
            const unsigned long dim359KuoInit[]     =   {   1   ,   3   ,   3   ,   11  ,   27  ,   21  ,   53  ,   95  ,   359 ,   925 ,   1035    ,   3093    ,0 };
            const unsigned long dim360KuoInit[]     =   {   1   ,   1   ,   3   ,   9   ,   9   ,   13  ,   25  ,   125 ,   101 ,   119 ,   1361    ,   3503    ,0 };
            const unsigned long dim361KuoInit[]     =   {   1   ,   3   ,   7   ,   11  ,   19  ,   43  ,   125 ,   95  ,   113 ,   373 ,   655 ,   1001    ,0 };
            const unsigned long dim362KuoInit[]     =   {   1   ,   1   ,   3   ,   15  ,   3   ,   29  ,   125 ,   5   ,   287 ,   859 ,   567 ,   3451    ,0 };
            const unsigned long dim363KuoInit[]     =   {   1   ,   3   ,   1   ,   1   ,   7   ,   21  ,   67  ,   249 ,   487 ,   663 ,   2041    ,   799 ,0 };
            const unsigned long dim364KuoInit[]     =   {   1   ,   1   ,   3   ,   9   ,   9   ,   29  ,   11  ,   229 ,   303 ,   821 ,   1187    ,   287 ,0 };
            const unsigned long dim365KuoInit[]     =   {   1   ,   1   ,   5   ,   11  ,   25  ,   19  ,   77  ,   93  ,   209 ,   155 ,   1717    ,   4085    ,0 };
            const unsigned long dim366KuoInit[]     =   {   1   ,   1   ,   3   ,   3   ,   23  ,   39  ,   5   ,   29  ,   295 ,   757 ,   1185    ,   817 ,0 };
            const unsigned long dim367KuoInit[]     =   {   1   ,   3   ,   7   ,   7   ,   27  ,   25  ,   33  ,   73  ,   489 ,   543 ,   2011    ,   447 ,0 };
            const unsigned long dim368KuoInit[]     =   {   1   ,   3   ,   1   ,   5   ,   21  ,   59  ,   107 ,   109 ,   415 ,   599 ,   325 ,   1457    ,0 };
            const unsigned long dim369KuoInit[]     =   {   1   ,   3   ,   1   ,   5   ,   27  ,   61  ,   45  ,   129 ,   37  ,   923 ,   285 ,   2077    ,0 };
            const unsigned long dim370KuoInit[]     =   {   1   ,   1   ,   1   ,   3   ,   11  ,   11  ,   17  ,   157 ,   289 ,   575 ,   749 ,   2463    ,0 };
            const unsigned long dim371KuoInit[]     =   {   1   ,   1   ,   3   ,   7   ,   25  ,   1   ,   19  ,   177 ,   361 ,   831 ,   911 ,   1833    ,0 };
            const unsigned long dim372KuoInit[]     =   {   1   ,   3   ,   7   ,   7   ,   25  ,   27  ,   11  ,   105 ,   489 ,   399 ,   483 ,   1287    ,0 };
            const unsigned long dim373KuoInit[]     =   {   1   ,   3   ,   7   ,   11  ,   3   ,   49  ,   27  ,   63  ,   33  ,   991 ,   341 ,   2239    ,0 };
            const unsigned long dim374KuoInit[]     =   {   1   ,   3   ,   1   ,   5   ,   3   ,   49  ,   23  ,   23  ,   157 ,   145 ,   1709    ,   1459    ,0 };
            const unsigned long dim375KuoInit[]     =   {   1   ,   1   ,   5   ,   9   ,   29  ,   29  ,   5   ,   239 ,   201 ,   503 ,   229 ,   1943    ,0 };
            const unsigned long dim376KuoInit[]     =   {   1   ,   3   ,   7   ,   3   ,   5   ,   39  ,   43  ,   127 ,   239 ,   679 ,   1225    ,   2123    ,0 };
            const unsigned long dim377KuoInit[]     =   {   1   ,   1   ,   7   ,   11  ,   13  ,   17  ,   125 ,   253 ,   223 ,   67  ,   1319    ,   4019    ,0 };
            const unsigned long dim378KuoInit[]     =   {   1   ,   3   ,   3   ,   3   ,   7   ,   57  ,   71  ,   131 ,   205 ,   953 ,   915 ,   1349    ,0 };
            const unsigned long dim379KuoInit[]     =   {   1   ,   1   ,   1   ,   7   ,   1   ,   59  ,   77  ,   221 ,   227 ,   67  ,   1853    ,   2577    ,0 };
            const unsigned long dim380KuoInit[]     =   {   1   ,   1   ,   7   ,   11  ,   31  ,   15  ,   103 ,   51  ,   505 ,   403 ,   1947    ,   1915    ,0 };
            const unsigned long dim381KuoInit[]     =   {   1   ,   3   ,   3   ,   3   ,   15  ,   7   ,   41  ,   179 ,   265 ,   787 ,   1249    ,   3663    ,0 };
            const unsigned long dim382KuoInit[]     =   {   1   ,   3   ,   1   ,   11  ,   27  ,   53  ,   123 ,   95  ,   427 ,   877 ,   471 ,   2167    ,0 };
            const unsigned long dim383KuoInit[]     =   {   1   ,   1   ,   1   ,   9   ,   25  ,   7   ,   31  ,   47  ,   21  ,   299 ,   1397    ,   1119    ,0 };
            const unsigned long dim384KuoInit[]     =   {   1   ,   3   ,   3   ,   9   ,   15  ,   49  ,   9   ,   255 ,   19  ,   995 ,   837 ,   2323    ,0 };
            const unsigned long dim385KuoInit[]     =   {   1   ,   1   ,   7   ,   3   ,   31  ,   45  ,   35  ,   93  ,   179 ,   561 ,   1737    ,   3969    ,0 };
            const unsigned long dim386KuoInit[]     =   {   1   ,   1   ,   1   ,   5   ,   7   ,   59  ,   27  ,   71  ,   9   ,   411 ,   59  ,   1783    ,0 };
            const unsigned long dim387KuoInit[]     =   {   1   ,   3   ,   3   ,   7   ,   27  ,   17  ,   43  ,   37  ,   505 ,   893 ,   709 ,   3163    ,0 };
            const unsigned long dim388KuoInit[]     =   {   1   ,   1   ,   5   ,   7   ,   7   ,   59  ,   47  ,   119 ,   39  ,   153 ,   259 ,   1409    ,0 };
            const unsigned long dim389KuoInit[]     =   {   1   ,   3   ,   5   ,   1   ,   21  ,   29  ,   23  ,   153 ,   309 ,   137 ,   1015    ,   1527    ,0 };
            const unsigned long dim390KuoInit[]     =   {   1   ,   1   ,   5   ,   15  ,   15  ,   27  ,   41  ,   15  ,   135 ,   707 ,   481 ,   2429    ,0 };
            const unsigned long dim391KuoInit[]     =   {   1   ,   1   ,   7   ,   13  ,   1   ,   17  ,   95  ,   131 ,   497 ,   873 ,   193 ,   1683    ,0 };
            const unsigned long dim392KuoInit[]     =   {   1   ,   1   ,   1   ,   13  ,   29  ,   39  ,   55  ,   51  ,   163 ,   581 ,   1867    ,   2219    ,0 };
            const unsigned long dim393KuoInit[]     =   {   1   ,   1   ,   7   ,   3   ,   21  ,   29  ,   9   ,   111 ,   503 ,   7   ,   89  ,   3431    ,0 };
            const unsigned long dim394KuoInit[]     =   {   1   ,   3   ,   7   ,   9   ,   13  ,   45  ,   61  ,   239 ,   61  ,   969 ,   77  ,   4053    ,0 };
            const unsigned long dim395KuoInit[]     =   {   1   ,   1   ,   7   ,   11  ,   13  ,   29  ,   79  ,   39  ,   361 ,   635 ,   1867    ,   3051    ,0 };
            const unsigned long dim396KuoInit[]     =   {   1   ,   3   ,   7   ,   3   ,   29  ,   11  ,   81  ,   193 ,   469 ,   365 ,   17  ,   3657    ,0 };
            const unsigned long dim397KuoInit[]     =   {   1   ,   3   ,   3   ,   15  ,   13  ,   45  ,   69  ,   181 ,   95  ,   315 ,   1025    ,   1935    ,0 };
            const unsigned long dim398KuoInit[]     =   {   1   ,   1   ,   5   ,   15  ,   31  ,   61  ,   97  ,   117 ,   207 ,   409 ,   69  ,   187 ,0 };
            const unsigned long dim399KuoInit[]     =   {   1   ,   1   ,   5   ,   9   ,   27  ,   59  ,   123 ,   157 ,   409 ,   999 ,   905 ,   1097    ,0 };
            const unsigned long dim400KuoInit[]     =   {   1   ,   1   ,   7   ,   1   ,   1   ,   51  ,   99  ,   143 ,   255 ,   817 ,   179 ,   3317    ,0 };
            const unsigned long dim401KuoInit[]     =   {   1   ,   1   ,   1   ,   1   ,   23  ,   15  ,   117 ,   139 ,   245 ,   865 ,   1723    ,   3501    ,0 };
            const unsigned long dim402KuoInit[]     =   {   1   ,   1   ,   5   ,   15  ,   25  ,   53  ,   9   ,   207 ,   291 ,   137 ,   291 ,   653 ,0 };
            const unsigned long dim403KuoInit[]     =   {   1   ,   1   ,   3   ,   9   ,   9   ,   63  ,   55  ,   147 ,   199 ,   455 ,   2047    ,   2903    ,0 };
            const unsigned long dim404KuoInit[]     =   {   1   ,   1   ,   5   ,   15  ,   29  ,   55  ,   85  ,   13  ,   117 ,   479 ,   1371    ,   2939    ,0 };
            const unsigned long dim405KuoInit[]     =   {   1   ,   3   ,   3   ,   5   ,   17  ,   15  ,   33  ,   83  ,   359 ,   469 ,   463 ,   491 ,0 };
            const unsigned long dim406KuoInit[]     =   {   1   ,   1   ,   3   ,   5   ,   23  ,   39  ,   35  ,   83  ,   387 ,   891 ,   789 ,   1469    ,0 };
            const unsigned long dim407KuoInit[]     =   {   1   ,   3   ,   7   ,   9   ,   27  ,   49  ,   41  ,   49  ,   207 ,   39  ,   25  ,   2355    ,0 };
            const unsigned long dim408KuoInit[]     =   {   1   ,   3   ,   5   ,   1   ,   5   ,   7   ,   29  ,   233 ,   169 ,   827 ,   1631    ,   2263    ,0 };
            const unsigned long dim409KuoInit[]     =   {   1   ,   1   ,   1   ,   1   ,   1   ,   13  ,   97  ,   109 ,   255 ,   449 ,   345 ,   2533    ,0 };
            const unsigned long dim410KuoInit[]     =   {   1   ,   3   ,   1   ,   5   ,   3   ,   15  ,   77  ,   217 ,   237 ,   57  ,   1711    ,   725 ,0 };
            const unsigned long dim411KuoInit[]     =   {   1   ,   3   ,   3   ,   7   ,   9   ,   49  ,   93  ,   213 ,   487 ,   811 ,   1659    ,   3019    ,0 };
            const unsigned long dim412KuoInit[]     =   {   1   ,   1   ,   1   ,   7   ,   7   ,   9   ,   57  ,   223 ,   263 ,   9   ,   443 ,   1103    ,0 };
            const unsigned long dim413KuoInit[]     =   {   1   ,   1   ,   5   ,   11  ,   21  ,   11  ,   83  ,   239 ,   347 ,   489 ,   983 ,   677 ,0 };
            const unsigned long dim414KuoInit[]     =   {   1   ,   1   ,   1   ,   11  ,   9   ,   29  ,   25  ,   105 ,   391 ,   1001    ,   1807    ,   1041    ,0 };
            const unsigned long dim415KuoInit[]     =   {   1   ,   3   ,   3   ,   1   ,   11  ,   21  ,   89  ,   85  ,   259 ,   237 ,   1489    ,   3825    ,0 };
            const unsigned long dim416KuoInit[]     =   {   1   ,   3   ,   5   ,   5   ,   1   ,   21  ,   53  ,   127 ,   269 ,   741 ,   529 ,   1649    ,0 };
            const unsigned long dim417KuoInit[]     =   {   1   ,   3   ,   1   ,   7   ,   3   ,   35  ,   71  ,   65  ,   21  ,   985 ,   111 ,   2199    ,0 };
            const unsigned long dim418KuoInit[]     =   {   1   ,   1   ,   5   ,   13  ,   3   ,   39  ,   33  ,   151 ,   95  ,   655 ,   1805    ,   1545    ,0 };
            const unsigned long dim419KuoInit[]     =   {   1   ,   3   ,   1   ,   11  ,   17  ,   53  ,   63  ,   171 ,   275 ,   649 ,   1791    ,   3309    ,0 };
            const unsigned long dim420KuoInit[]     =   {   1   ,   1   ,   7   ,   1   ,   3   ,   5   ,   63  ,   241 ,   77  ,   677 ,   1713    ,   1643    ,0 };
            const unsigned long dim421KuoInit[]     =   {   1   ,   3   ,   3   ,   3   ,   9   ,   25  ,   7   ,   33  ,   277 ,   837 ,   1235    ,   3289    ,0 };
            const unsigned long dim422KuoInit[]     =   {   1   ,   3   ,   3   ,   1   ,   27  ,   13  ,   97  ,   79  ,   145 ,   197 ,   1237    ,   1689    ,0 };
            const unsigned long dim423KuoInit[]     =   {   1   ,   3   ,   5   ,   7   ,   27  ,   45  ,   71  ,   165 ,   405 ,   251 ,   383 ,   737 ,0 };
            const unsigned long dim424KuoInit[]     =   {   1   ,   3   ,   7   ,   1   ,   13  ,   21  ,   107 ,   7   ,   207 ,   329 ,   1671    ,   2053    ,0 };
            const unsigned long dim425KuoInit[]     =   {   1   ,   3   ,   7   ,   1   ,   31  ,   33  ,   25  ,   223 ,   37  ,   693 ,   1831    ,   2249    ,0 };
            const unsigned long dim426KuoInit[]     =   {   1   ,   3   ,   3   ,   15  ,   5   ,   53  ,   111 ,   11  ,   317 ,   819 ,   1345    ,   401 ,0 };
            const unsigned long dim427KuoInit[]     =   {   1   ,   3   ,   1   ,   9   ,   7   ,   61  ,   35  ,   127 ,   453 ,   565 ,   37  ,   17  ,0 };
            const unsigned long dim428KuoInit[]     =   {   1   ,   3   ,   1   ,   1   ,   15  ,   3   ,   37  ,   133 ,   339 ,   181 ,   377 ,   2339    ,0 };
            const unsigned long dim429KuoInit[]     =   {   1   ,   3   ,   7   ,   3   ,   27  ,   57  ,   27  ,   23  ,   45  ,   219 ,   455 ,   2821    ,0 };
            const unsigned long dim430KuoInit[]     =   {   1   ,   3   ,   7   ,   9   ,   3   ,   43  ,   79  ,   67  ,   15  ,   797 ,   129 ,   1233    ,0 };
            const unsigned long dim431KuoInit[]     =   {   1   ,   3   ,   5   ,   11  ,   5   ,   45  ,   123 ,   161 ,   355 ,   447 ,   1165    ,   2437    ,0 };
            const unsigned long dim432KuoInit[]     =   {   1   ,   3   ,   7   ,   1   ,   3   ,   59  ,   65  ,   109 ,   253 ,   447 ,   649 ,   379 ,0 };
            const unsigned long dim433KuoInit[]     =   {   1   ,   3   ,   1   ,   5   ,   21  ,   23  ,   45  ,   47  ,   213 ,   813 ,   1475    ,   1271    ,0 };
            const unsigned long dim434KuoInit[]     =   {   1   ,   1   ,   1   ,   7   ,   9   ,   31  ,   91  ,   47  ,   161 ,   535 ,   1529    ,   881 ,0 };
            const unsigned long dim435KuoInit[]     =   {   1   ,   1   ,   5   ,   13  ,   1   ,   25  ,   85  ,   107 ,   59  ,   487 ,   221 ,   51  ,0 };
            const unsigned long dim436KuoInit[]     =   {   1   ,   1   ,   1   ,   1   ,   1   ,   55  ,   93  ,   95  ,   267 ,   157 ,   1255    ,   2115    ,0 };
            const unsigned long dim437KuoInit[]     =   {   1   ,   3   ,   1   ,   11  ,   23  ,   55  ,   29  ,   187 ,   261 ,   139 ,   515 ,   2627    ,0 };
            const unsigned long dim438KuoInit[]     =   {   1   ,   3   ,   3   ,   13  ,   31  ,   3   ,   121 ,   245 ,   475 ,   67  ,   195 ,   979 ,0 };
            const unsigned long dim439KuoInit[]     =   {   1   ,   1   ,   3   ,   7   ,   25  ,   13  ,   79  ,   129 ,   475 ,   177 ,   1519    ,   1069    ,0 };
            const unsigned long dim440KuoInit[]     =   {   1   ,   3   ,   3   ,   9   ,   17  ,   23  ,   83  ,   191 ,   133 ,   901 ,   1161    ,   1657    ,0 };
            const unsigned long dim441KuoInit[]     =   {   1   ,   3   ,   7   ,   15  ,   23  ,   61  ,   111 ,   135 ,   107 ,   27  ,   583 ,   225 ,0 };
            const unsigned long dim442KuoInit[]     =   {   1   ,   1   ,   5   ,   1   ,   17  ,   11  ,   31  ,   197 ,   111 ,   953 ,   1227    ,   3737    ,0 };
            const unsigned long dim443KuoInit[]     =   {   1   ,   1   ,   3   ,   9   ,   19  ,   25  ,   83  ,   97  ,   489 ,   781 ,   1591    ,   2061    ,0 };
            const unsigned long dim444KuoInit[]     =   {   1   ,   1   ,   1   ,   1   ,   9   ,   61  ,   31  ,   93  ,   39  ,   53  ,   1613    ,   3371    ,0 };
            const unsigned long dim445KuoInit[]     =   {   1   ,   1   ,   1   ,   7   ,   3   ,   5   ,   103 ,   247 ,   255 ,   333 ,   31  ,   3087    ,0 };
            const unsigned long dim446KuoInit[]     =   {   1   ,   1   ,   7   ,   7   ,   17  ,   35  ,   51  ,   153 ,   363 ,   897 ,   395 ,   1153    ,0 };
            const unsigned long dim447KuoInit[]     =   {   1   ,   3   ,   7   ,   9   ,   3   ,   59  ,   127 ,   177 ,   191 ,   255 ,   1543    ,   503 ,0 };
            const unsigned long dim448KuoInit[]     =   {   1   ,   1   ,   7   ,   3   ,   1   ,   7   ,   81  ,   199 ,   411 ,   565 ,   1717    ,   615 ,0 };
            const unsigned long dim449KuoInit[]     =   {   1   ,   1   ,   1   ,   11  ,   17  ,   35  ,   35  ,   245 ,   507 ,   439 ,   523 ,   2343    ,0 };
            const unsigned long dim450KuoInit[]     =   {   1   ,   1   ,   7   ,   15  ,   13  ,   61  ,   31  ,   99  ,   241 ,   605 ,   1575    ,   1321    ,0 };
            const unsigned long dim451KuoInit[]     =   {   1   ,   3   ,   1   ,   5   ,   15  ,   31  ,   45  ,   89  ,   193 ,   905 ,   691 ,   3917    ,0 };
            const unsigned long dim452KuoInit[]     =   {   1   ,   3   ,   3   ,   9   ,   29  ,   31  ,   101 ,   151 ,   187 ,   199 ,   1543    ,   3951    ,0 };
            const unsigned long dim453KuoInit[]     =   {   1   ,   1   ,   7   ,   9   ,   21  ,   45  ,   71  ,   19  ,   395 ,   175 ,   503 ,   3759    ,0 };
            const unsigned long dim454KuoInit[]     =   {   1   ,   3   ,   7   ,   7   ,   17  ,   31  ,   47  ,   7   ,   135 ,   719 ,   567 ,   1447    ,0 };
            const unsigned long dim455KuoInit[]     =   {   1   ,   3   ,   3   ,   9   ,   1   ,   39  ,   1   ,   53  ,   5   ,   553 ,   371 ,   3833    ,0 };
            const unsigned long dim456KuoInit[]     =   {   1   ,   1   ,   1   ,   11  ,   1   ,   33  ,   97  ,   209 ,   377 ,   187 ,   1683    ,   143 ,0 };
            const unsigned long dim457KuoInit[]     =   {   1   ,   1   ,   1   ,   3   ,   29  ,   33  ,   111 ,   35  ,   343 ,   511 ,   273 ,   445 ,0 };
            const unsigned long dim458KuoInit[]     =   {   1   ,   1   ,   1   ,   15  ,   15  ,   63  ,   73  ,   1   ,   111 ,   775 ,   665 ,   2737    ,0 };
            const unsigned long dim459KuoInit[]     =   {   1   ,   1   ,   1   ,   9   ,   29  ,   23  ,   127 ,   77  ,   33  ,   131 ,   1055    ,   2351    ,0 };
            const unsigned long dim460KuoInit[]     =   {   1   ,   3   ,   5   ,   7   ,   31  ,   53  ,   117 ,   57  ,   287 ,   221 ,   1099    ,   1147    ,0 };
            const unsigned long dim461KuoInit[]     =   {   1   ,   3   ,   7   ,   15  ,   3   ,   17  ,   15  ,   177 ,   189 ,   379 ,   179 ,   1095    ,0 };
            const unsigned long dim462KuoInit[]     =   {   1   ,   3   ,   1   ,   15  ,   25  ,   19  ,   43  ,   113 ,   123 ,   123 ,   877 ,   1869    ,0 };
            const unsigned long dim463KuoInit[]     =   {   1   ,   3   ,   1   ,   1   ,   13  ,   63  ,   1   ,   83  ,   197 ,   215 ,   1005    ,   3531    ,0 };
            const unsigned long dim464KuoInit[]     =   {   1   ,   3   ,   7   ,   1   ,   27  ,   25  ,   111 ,   199 ,   239 ,   909 ,   485 ,   3399    ,0 };
            const unsigned long dim465KuoInit[]     =   {   1   ,   1   ,   7   ,   3   ,   19  ,   57  ,   43  ,   75  ,   163 ,   111 ,   359 ,   1987    ,0 };
            const unsigned long dim466KuoInit[]     =   {   1   ,   1   ,   3   ,   7   ,   11  ,   63  ,   79  ,   235 ,   197 ,   757 ,   293 ,   3795    ,0 };
            const unsigned long dim467KuoInit[]     =   {   1   ,   3   ,   5   ,   9   ,   13  ,   55  ,   49  ,   29  ,   133 ,   949 ,   1185    ,   905 ,0 };
            const unsigned long dim468KuoInit[]     =   {   1   ,   1   ,   7   ,   1   ,   15  ,   51  ,   13  ,   223 ,   243 ,   831 ,   845 ,   2663    ,0 };
            const unsigned long dim469KuoInit[]     =   {   1   ,   1   ,   3   ,   11  ,   25  ,   21  ,   65  ,   57  ,   483 ,   665 ,   741 ,   2713    ,0 };
            const unsigned long dim470KuoInit[]     =   {   1   ,   3   ,   5   ,   3   ,   21  ,   3   ,   81  ,   57  ,   133 ,   553 ,   683 ,   425 ,0 };
            const unsigned long dim471KuoInit[]     =   {   1   ,   1   ,   5   ,   11  ,   31  ,   29  ,   59  ,   189 ,   399 ,   751 ,   651 ,   4093    ,0 };
            const unsigned long dim472KuoInit[]     =   {   1   ,   3   ,   3   ,   7   ,   13  ,   59  ,   19  ,   217 ,   267 ,   355 ,   173 ,   4051    ,0 };
            const unsigned long dim473KuoInit[]     =   {   1   ,   1   ,   5   ,   9   ,   23  ,   49  ,   57  ,   145 ,   77  ,   591 ,   1913    ,   779 ,0 };
            const unsigned long dim474KuoInit[]     =   {   1   ,   3   ,   7   ,   3   ,   25  ,   1   ,   83  ,   103 ,   393 ,   121 ,   1415    ,   1503    ,0 };
            const unsigned long dim475KuoInit[]     =   {   1   ,   1   ,   7   ,   9   ,   23  ,   41  ,   35  ,   157 ,   63  ,   293 ,   1911    ,   2303    ,0 };
            const unsigned long dim476KuoInit[]     =   {   1   ,   3   ,   5   ,   15  ,   7   ,   3   ,   9   ,   1   ,   201 ,   913 ,   1275    ,   3495    ,0 };
            const unsigned long dim477KuoInit[]     =   {   1   ,   1   ,   7   ,   11  ,   15  ,   5   ,   107 ,   57  ,   53  ,   451 ,   351 ,   2621    ,0 };
            const unsigned long dim478KuoInit[]     =   {   1   ,   3   ,   7   ,   15  ,   23  ,   25  ,   97  ,   29  ,   111 ,   531 ,   1979    ,   2115    ,0 };
            const unsigned long dim479KuoInit[]     =   {   1   ,   1   ,   1   ,   1   ,   19  ,   11  ,   49  ,   229 ,   355 ,   195 ,   2033    ,   1701    ,0 };
            const unsigned long dim480KuoInit[]     =   {   1   ,   3   ,   3   ,   3   ,   15  ,   35  ,   17  ,   13  ,   425 ,   285 ,   965 ,   159 ,0 };
            const unsigned long dim481KuoInit[]     =   {   1   ,   3   ,   1   ,   11  ,   11  ,   11  ,   1   ,   145 ,   365 ,   471 ,   1527    ,   2309    ,   5863    ,0 };
            const unsigned long dim482KuoInit[]     =   {   1   ,   1   ,   3   ,   9   ,   21  ,   59  ,   35  ,   123 ,   91  ,   741 ,   527 ,   2061    ,   5505    ,0 };
            const unsigned long dim483KuoInit[]     =   {   1   ,   3   ,   7   ,   9   ,   9   ,   37  ,   75  ,   171 ,   75  ,   943 ,   443 ,   201 ,   4861    ,0 };
            const unsigned long dim484KuoInit[]     =   {   1   ,   1   ,   7   ,   13  ,   23  ,   49  ,   67  ,   103 ,   453 ,   643 ,   1117    ,   1621    ,   7135    ,0 };
            const unsigned long dim485KuoInit[]     =   {   1   ,   3   ,   1   ,   7   ,   15  ,   43  ,   113 ,   17  ,   277 ,   301 ,   349 ,   1155    ,   5285    ,0 };
            const unsigned long dim486KuoInit[]     =   {   1   ,   3   ,   5   ,   1   ,   25  ,   61  ,   103 ,   81  ,   141 ,   347 ,   1467    ,   1535    ,   421 ,0 };
            const unsigned long dim487KuoInit[]     =   {   1   ,   3   ,   5   ,   13  ,   29  ,   45  ,   43  ,   31  ,   403 ,   711 ,   455 ,   3387    ,   3521    ,0 };
            const unsigned long dim488KuoInit[]     =   {   1   ,   3   ,   3   ,   7   ,   7   ,   41  ,   71  ,   251 ,   147 ,   241 ,   1129    ,   2093    ,   7683    ,0 };
            const unsigned long dim489KuoInit[]     =   {   1   ,   1   ,   1   ,   5   ,   23  ,   61  ,   57  ,   97  ,   177 ,   387 ,   1557    ,   3587    ,   751 ,0 };
            const unsigned long dim490KuoInit[]     =   {   1   ,   3   ,   1   ,   13  ,   11  ,   33  ,   47  ,   147 ,   321 ,   661 ,   53  ,   2437    ,   3087    ,0 };
            const unsigned long dim491KuoInit[]     =   {   1   ,   3   ,   1   ,   3   ,   17  ,   21  ,   79  ,   91  ,   77  ,   1011    ,   745 ,   3551    ,   1857    ,0 };
            const unsigned long dim492KuoInit[]     =   {   1   ,   1   ,   3   ,   15  ,   19  ,   27  ,   63  ,   231 ,   267 ,   879 ,   1959    ,   3019    ,   1655    ,0 };
            const unsigned long dim493KuoInit[]     =   {   1   ,   1   ,   7   ,   15  ,   27  ,   55  ,   21  ,   37  ,   417 ,   69  ,   327 ,   2709    ,   6335    ,0 };
            const unsigned long dim494KuoInit[]     =   {   1   ,   3   ,   1   ,   15  ,   1   ,   9   ,   27  ,   89  ,   205 ,   355 ,   373 ,   2191    ,   311 ,0 };
            const unsigned long dim495KuoInit[]     =   {   1   ,   1   ,   3   ,   5   ,   15  ,   51  ,   85  ,   233 ,   21  ,   971 ,   1823    ,   1843    ,   1417    ,0 };
            const unsigned long dim496KuoInit[]     =   {   1   ,   3   ,   5   ,   13  ,   31  ,   57  ,   3   ,   199 ,   211 ,   401 ,   1559    ,   3601    ,   4569    ,0 };
            const unsigned long dim497KuoInit[]     =   {   1   ,   3   ,   3   ,   1   ,   27  ,   63  ,   93  ,   123 ,   223 ,   601 ,   1023    ,   2501    ,   7105    ,0 };
            const unsigned long dim498KuoInit[]     =   {   1   ,   1   ,   1   ,   7   ,   3   ,   31  ,   5   ,   51  ,   453 ,   535 ,   255 ,   735 ,   749 ,0 };
            const unsigned long dim499KuoInit[]     =   {   1   ,   3   ,   5   ,   9   ,   23  ,   63  ,   109 ,   163 ,   221 ,   845 ,   911 ,   933 ,   361 ,0 };
            const unsigned long dim500KuoInit[]     =   {   1   ,   1   ,   1   ,   13  ,   13  ,   17  ,   19  ,   191 ,   227 ,   711 ,   1523    ,   1635    ,   6397    ,0 };
            const unsigned long dim501KuoInit[]     =   {   1   ,   3   ,   7   ,   7   ,   21  ,   21  ,   23  ,   91  ,   439 ,   227 ,   1831    ,   1475    ,   705 ,0 };
            const unsigned long dim502KuoInit[]     =   {   1   ,   3   ,   7   ,   7   ,   13  ,   39  ,   99  ,   199 ,   149 ,   1005    ,   149 ,   445 ,   3367    ,0 };
            const unsigned long dim503KuoInit[]     =   {   1   ,   3   ,   1   ,   5   ,   25  ,   61  ,   99  ,   187 ,   33  ,   619 ,   1833    ,   2177    ,   7631    ,0 };
            const unsigned long dim504KuoInit[]     =   {   1   ,   1   ,   1   ,   13  ,   15  ,   45  ,   1   ,   225 ,   129 ,   113 ,   2009    ,   1223    ,   49  ,0 };
            const unsigned long dim505KuoInit[]     =   {   1   ,   3   ,   1   ,   9   ,   7   ,   43  ,   57  ,   193 ,   409 ,   265 ,   1151    ,   2635    ,   6005    ,0 };
            const unsigned long dim506KuoInit[]     =   {   1   ,   3   ,   5   ,   5   ,   13  ,   13  ,   99  ,   163 ,   1   ,   629 ,   87  ,   925 ,   137 ,0 };
            const unsigned long dim507KuoInit[]     =   {   1   ,   3   ,   7   ,   9   ,   23  ,   43  ,   43  ,   1   ,   505 ,   179 ,   335 ,   2033    ,   8065    ,0 };
            const unsigned long dim508KuoInit[]     =   {   1   ,   3   ,   5   ,   1   ,   7   ,   55  ,   99  ,   159 ,   19  ,   883 ,   1503    ,   801 ,   1433    ,0 };
            const unsigned long dim509KuoInit[]     =   {   1   ,   3   ,   7   ,   1   ,   1   ,   55  ,   53  ,   241 ,   95  ,   315 ,   723 ,   3571    ,   4685    ,0 };
            const unsigned long dim510KuoInit[]     =   {   1   ,   3   ,   5   ,   15  ,   29  ,   43  ,   97  ,   159 ,   427 ,   87  ,   1995    ,   1849    ,   6771    ,0 };
            const unsigned long dim511KuoInit[]     =   {   1   ,   1   ,   1   ,   7   ,   5   ,   55  ,   55  ,   203 ,   119 ,   185 ,   1333    ,   2107    ,   4271    ,0 };
            const unsigned long dim512KuoInit[]     =   {   1   ,   1   ,   3   ,   15  ,   5   ,   7   ,   19  ,   185 ,   75  ,   601 ,   2015    ,   2175    ,   4259    ,0 };
            const unsigned long dim513KuoInit[]     =   {   1   ,   3   ,   3   ,   15  ,   3   ,   49  ,   33  ,   115 ,   413 ,   321 ,   1787    ,   469 ,   3721    ,0 };
            const unsigned long dim514KuoInit[]     =   {   1   ,   3   ,   1   ,   9   ,   25  ,   63  ,   123 ,   17  ,   355 ,   903 ,   1269    ,   3737    ,   5333    ,0 };
            const unsigned long dim515KuoInit[]     =   {   1   ,   1   ,   3   ,   7   ,   25  ,   41  ,   113 ,   239 ,   373 ,   533 ,   1823    ,   3513    ,   7745    ,0 };
            const unsigned long dim516KuoInit[]     =   {   1   ,   1   ,   1   ,   11  ,   13  ,   7   ,   7   ,   43  ,   145 ,   607 ,   2043    ,   2397    ,   4051    ,0 };
            const unsigned long dim517KuoInit[]     =   {   1   ,   1   ,   3   ,   3   ,   9   ,   49  ,   5   ,   173 ,   63  ,   157 ,   637 ,   2307    ,   6603    ,0 };
            const unsigned long dim518KuoInit[]     =   {   1   ,   3   ,   1   ,   3   ,   23  ,   33  ,   89  ,   139 ,   207 ,   389 ,   1389    ,   3385    ,   2169    ,0 };
            const unsigned long dim519KuoInit[]     =   {   1   ,   1   ,   7   ,   3   ,   7   ,   7   ,   19  ,   7   ,   217 ,   189 ,   739 ,   2475    ,   7925    ,0 };
            const unsigned long dim520KuoInit[]     =   {   1   ,   1   ,   7   ,   5   ,   23  ,   37  ,   71  ,   31  ,   145 ,   29  ,   537 ,   959 ,   2227    ,0 };
            const unsigned long dim521KuoInit[]     =   {   1   ,   1   ,   7   ,   11  ,   7   ,   55  ,   39  ,   167 ,   347 ,   293 ,   577 ,   3077    ,   531 ,0 };
            const unsigned long dim522KuoInit[]     =   {   1   ,   3   ,   1   ,   5   ,   19  ,   1   ,   119 ,   133 ,   487 ,   867 ,   1331    ,   1759    ,   3683    ,0 };
            const unsigned long dim523KuoInit[]     =   {   1   ,   3   ,   5   ,   13  ,   27  ,   31  ,   77  ,   71  ,   465 ,   133 ,   105 ,   477 ,   6301    ,0 };
            const unsigned long dim524KuoInit[]     =   {   1   ,   3   ,   5   ,   15  ,   25  ,   49  ,   35  ,   75  ,   57  ,   663 ,   195 ,   1041    ,   5697    ,0 };
            const unsigned long dim525KuoInit[]     =   {   1   ,   3   ,   7   ,   9   ,   9   ,   7   ,   47  ,   233 ,   397 ,   155 ,   439 ,   2487    ,   4697    ,0 };
            const unsigned long dim526KuoInit[]     =   {   1   ,   1   ,   5   ,   9   ,   5   ,   41  ,   125 ,   9   ,   477 ,   567 ,   2033    ,   769 ,   1553    ,0 };
            const unsigned long dim527KuoInit[]     =   {   1   ,   1   ,   3   ,   7   ,   7   ,   25  ,   95  ,   45  ,   211 ,   453 ,   483 ,   3965    ,   1865    ,0 };
            const unsigned long dim528KuoInit[]     =   {   1   ,   3   ,   7   ,   5   ,   17  ,   41  ,   109 ,   59  ,   385 ,   1023    ,   1527    ,   2445    ,   6303    ,0 };
            const unsigned long dim529KuoInit[]     =   {   1   ,   3   ,   3   ,   9   ,   7   ,   29  ,   91  ,   245 ,   189 ,   251 ,   1237    ,   2825    ,   7919    ,0 };
            const unsigned long dim530KuoInit[]     =   {   1   ,   3   ,   7   ,   1   ,   25  ,   43  ,   27  ,   11  ,   63  ,   641 ,   1765    ,   1683    ,   2445    ,0 };
            const unsigned long dim531KuoInit[]     =   {   1   ,   1   ,   1   ,   7   ,   3   ,   59  ,   19  ,   81  ,   387 ,   207 ,   31  ,   1583    ,   607 ,0 };
            const unsigned long dim532KuoInit[]     =   {   1   ,   1   ,   3   ,   1   ,   3   ,   33  ,   99  ,   135 ,   267 ,   355 ,   585 ,   4017    ,   2385    ,0 };
            const unsigned long dim533KuoInit[]     =   {   1   ,   1   ,   7   ,   15  ,   1   ,   49  ,   35  ,   211 ,   359 ,   379 ,   1173    ,   2709    ,   3945    ,0 };
            const unsigned long dim534KuoInit[]     =   {   1   ,   1   ,   5   ,   7   ,   29  ,   23  ,   99  ,   219 ,   269 ,   977 ,   1175    ,   2305    ,   5633    ,0 };
            const unsigned long dim535KuoInit[]     =   {   1   ,   3   ,   3   ,   15  ,   17  ,   9   ,   115 ,   21  ,   503 ,   329 ,   1725    ,   289 ,   4667    ,0 };
            const unsigned long dim536KuoInit[]     =   {   1   ,   1   ,   7   ,   1   ,   9   ,   1   ,   31  ,   145 ,   209 ,   833 ,   1527    ,   4021    ,   2177    ,0 };
            const unsigned long dim537KuoInit[]     =   {   1   ,   3   ,   3   ,   1   ,   25  ,   13  ,   37  ,   143 ,   217 ,   891 ,   1289    ,   3635    ,   6931    ,0 };
            const unsigned long dim538KuoInit[]     =   {   1   ,   1   ,   3   ,   9   ,   9   ,   49  ,   33  ,   123 ,   435 ,   993 ,   389 ,   2491    ,   2097    ,0 };
            const unsigned long dim539KuoInit[]     =   {   1   ,   3   ,   5   ,   9   ,   11  ,   53  ,   121 ,   25  ,   31  ,   589 ,   1939    ,   3981    ,   621 ,0 };
            const unsigned long dim540KuoInit[]     =   {   1   ,   1   ,   1   ,   1   ,   27  ,   53  ,   9   ,   217 ,   135 ,   605 ,   1779    ,   3053    ,   2865    ,0 };
            const unsigned long dim541KuoInit[]     =   {   1   ,   1   ,   7   ,   5   ,   3   ,   53  ,   9   ,   15  ,   275 ,   613 ,   1395    ,   2287    ,   1111    ,0 };
            const unsigned long dim542KuoInit[]     =   {   1   ,   1   ,   3   ,   1   ,   13  ,   17  ,   3   ,   103 ,   195 ,   509 ,   493 ,   2395    ,   4669    ,0 };
            const unsigned long dim543KuoInit[]     =   {   1   ,   3   ,   5   ,   1   ,   13  ,   17  ,   59  ,   3   ,   117 ,   39  ,   1699    ,   2339    ,   5203    ,0 };
            const unsigned long dim544KuoInit[]     =   {   1   ,   3   ,   5   ,   3   ,   29  ,   23  ,   59  ,   251 ,   139 ,   625 ,   935 ,   785 ,   7547    ,0 };
            const unsigned long dim545KuoInit[]     =   {   1   ,   3   ,   1   ,   13  ,   23  ,   47  ,   5   ,   91  ,   283 ,   59  ,   919 ,   2427    ,   1631    ,0 };
            const unsigned long dim546KuoInit[]     =   {   1   ,   1   ,   3   ,   9   ,   15  ,   35  ,   67  ,   131 ,   183 ,   891 ,   345 ,   3741    ,   3175    ,0 };
            const unsigned long dim547KuoInit[]     =   {   1   ,   3   ,   1   ,   3   ,   23  ,   11  ,   85  ,   135 ,   269 ,   347 ,   713 ,   2947    ,   1617    ,0 };
            const unsigned long dim548KuoInit[]     =   {   1   ,   3   ,   1   ,   3   ,   27  ,   33  ,   77  ,   189 ,   243 ,   893 ,   283 ,   885 ,   1999    ,0 };
            const unsigned long dim549KuoInit[]     =   {   1   ,   1   ,   5   ,   1   ,   31  ,   39  ,   99  ,   21  ,   437 ,   979 ,   919 ,   3663    ,   6689    ,0 };
            const unsigned long dim550KuoInit[]     =   {   1   ,   1   ,   1   ,   1   ,   17  ,   27  ,   55  ,   161 ,   205 ,   583 ,   383 ,   1681    ,   6445    ,0 };
            const unsigned long dim551KuoInit[]     =   {   1   ,   3   ,   7   ,   1   ,   31  ,   59  ,   111 ,   129 ,   121 ,   555 ,   653 ,   2787    ,   3899    ,0 };
            const unsigned long dim552KuoInit[]     =   {   1   ,   3   ,   1   ,   9   ,   11  ,   59  ,   51  ,   115 ,   349 ,   809 ,   513 ,   1585    ,   31  ,0 };
            const unsigned long dim553KuoInit[]     =   {   1   ,   1   ,   1   ,   3   ,   29  ,   17  ,   3   ,   173 ,   111 ,   671 ,   1289    ,   397 ,   799 ,0 };
            const unsigned long dim554KuoInit[]     =   {   1   ,   3   ,   1   ,   7   ,   17  ,   55  ,   85  ,   49  ,   23  ,   915 ,   1929    ,   685 ,   7021    ,0 };
            const unsigned long dim555KuoInit[]     =   {   1   ,   3   ,   7   ,   11  ,   3   ,   41  ,   123 ,   227 ,   111 ,   805 ,   1567    ,   751 ,   7927    ,0 };
            const unsigned long dim556KuoInit[]     =   {   1   ,   1   ,   7   ,   3   ,   15  ,   45  ,   117 ,   39  ,   491 ,   401 ,   1775    ,   2197    ,   4303    ,0 };
            const unsigned long dim557KuoInit[]     =   {   1   ,   1   ,   7   ,   9   ,   13  ,   23  ,   63  ,   245 ,   163 ,   595 ,   1271    ,   2143    ,   8105    ,0 };
            const unsigned long dim558KuoInit[]     =   {   1   ,   1   ,   3   ,   1   ,   31  ,   57  ,   45  ,   35  ,   327 ,   89  ,   1845    ,   1075    ,   4649    ,0 };
            const unsigned long dim559KuoInit[]     =   {   1   ,   3   ,   7   ,   7   ,   11  ,   31  ,   51  ,   209 ,   103 ,   215 ,   1693    ,   3427    ,   677 ,0 };
            const unsigned long dim560KuoInit[]     =   {   1   ,   1   ,   5   ,   13  ,   3   ,   7   ,   83  ,   199 ,   325 ,   461 ,   1185    ,   1997    ,   2327    ,0 };
            const unsigned long dim561KuoInit[]     =   {   1   ,   1   ,   3   ,   15  ,   23  ,   13  ,   23  ,   39  ,   209 ,   267 ,   1995    ,   3999    ,   6277    ,0 };
            const unsigned long dim562KuoInit[]     =   {   1   ,   1   ,   7   ,   1   ,   5   ,   37  ,   99  ,   117 ,   509 ,   303 ,   655 ,   935 ,   6321    ,0 };
            const unsigned long dim563KuoInit[]     =   {   1   ,   1   ,   1   ,   15  ,   23  ,   39  ,   123 ,   11  ,   445 ,   897 ,   1327    ,   2819    ,   1899    ,0 };
            const unsigned long dim564KuoInit[]     =   {   1   ,   1   ,   5   ,   5   ,   7   ,   7   ,   7   ,   61  ,   349 ,   865 ,   707 ,   2097    ,   1217    ,0 };
            const unsigned long dim565KuoInit[]     =   {   1   ,   1   ,   5   ,   9   ,   1   ,   63  ,   35  ,   187 ,   207 ,   53  ,   1917    ,   621 ,   2113    ,0 };
            const unsigned long dim566KuoInit[]     =   {   1   ,   1   ,   1   ,   15  ,   21  ,   61  ,   101 ,   169 ,   129 ,   813 ,   453 ,   323 ,   7283    ,0 };
            const unsigned long dim567KuoInit[]     =   {   1   ,   1   ,   7   ,   7   ,   27  ,   1   ,   27  ,   141 ,   33  ,   307 ,   2023    ,   2561    ,   8111    ,0 };
            const unsigned long dim568KuoInit[]     =   {   1   ,   1   ,   5   ,   13  ,   3   ,   9   ,   3   ,   87  ,   451 ,   691 ,   973 ,   539 ,   4393    ,0 };
            const unsigned long dim569KuoInit[]     =   {   1   ,   1   ,   5   ,   1   ,   23  ,   23  ,   9   ,   101 ,   349 ,   299 ,   583 ,   2037    ,   5247    ,0 };
            const unsigned long dim570KuoInit[]     =   {   1   ,   1   ,   5   ,   11  ,   15  ,   39  ,   99  ,   163 ,   395 ,   123 ,   1067    ,   2735    ,   4407    ,0 };
            const unsigned long dim571KuoInit[]     =   {   1   ,   3   ,   5   ,   11  ,   9   ,   61  ,   51  ,   237 ,   483 ,   501 ,   1767    ,   1419    ,   5047    ,0 };
            const unsigned long dim572KuoInit[]     =   {   1   ,   1   ,   7   ,   9   ,   27  ,   31  ,   57  ,   103 ,   291 ,   537 ,   1759    ,   1861    ,   3815    ,0 };
            const unsigned long dim573KuoInit[]     =   {   1   ,   1   ,   5   ,   15  ,   31  ,   5   ,   107 ,   103 ,   159 ,   939 ,   1331    ,   1219    ,   2045    ,0 };
            const unsigned long dim574KuoInit[]     =   {   1   ,   3   ,   5   ,   15  ,   7   ,   45  ,   99  ,   175 ,   131 ,   967 ,   593 ,   3191    ,   5403    ,0 };
            const unsigned long dim575KuoInit[]     =   {   1   ,   3   ,   7   ,   11  ,   19  ,   43  ,   57  ,   129 ,   285 ,   943 ,   1507    ,   3979    ,   2807    ,0 };
            const unsigned long dim576KuoInit[]     =   {   1   ,   1   ,   5   ,   3   ,   11  ,   7   ,   117 ,   241 ,   393 ,   1023    ,   95  ,   535 ,   7235    ,0 };
            const unsigned long dim577KuoInit[]     =   {   1   ,   3   ,   1   ,   5   ,   17  ,   5   ,   111 ,   165 ,   343 ,   65  ,   1393    ,   2855    ,   3801    ,0 };
            const unsigned long dim578KuoInit[]     =   {   1   ,   1   ,   5   ,   5   ,   27  ,   1   ,   87  ,   9   ,   177 ,   455 ,   1915    ,   1557    ,   7175    ,0 };
            const unsigned long dim579KuoInit[]     =   {   1   ,   1   ,   5   ,   11  ,   31  ,   33  ,   105 ,   255 ,   281 ,   303 ,   1831    ,   1575    ,   1179    ,0 };
            const unsigned long dim580KuoInit[]     =   {   1   ,   3   ,   3   ,   9   ,   7   ,   37  ,   3   ,   85  ,   161 ,   491 ,   299 ,   4003    ,   7429    ,0 };
            const unsigned long dim581KuoInit[]     =   {   1   ,   1   ,   5   ,   9   ,   1   ,   19  ,   23  ,   175 ,   497 ,   613 ,   2043    ,   1467    ,   2379    ,0 };
            const unsigned long dim582KuoInit[]     =   {   1   ,   3   ,   7   ,   13  ,   5   ,   57  ,   93  ,   17  ,   383 ,   801 ,   1465    ,   3151    ,   4717    ,0 };
            const unsigned long dim583KuoInit[]     =   {   1   ,   3   ,   7   ,   5   ,   3   ,   19  ,   87  ,   25  ,   95  ,   643 ,   79  ,   123 ,   7175    ,0 };
            const unsigned long dim584KuoInit[]     =   {   1   ,   1   ,   1   ,   5   ,   3   ,   27  ,   49  ,   155 ,   45  ,   655 ,   1079    ,   3243    ,   8147    ,0 };
            const unsigned long dim585KuoInit[]     =   {   1   ,   3   ,   1   ,   11  ,   13  ,   57  ,   107 ,   37  ,   147 ,   119 ,   887 ,   1317    ,   6687    ,0 };
            const unsigned long dim586KuoInit[]     =   {   1   ,   3   ,   3   ,   13  ,   29  ,   23  ,   105 ,   137 ,   339 ,   509 ,   1543    ,   2339    ,   2167    ,0 };
            const unsigned long dim587KuoInit[]     =   {   1   ,   1   ,   5   ,   3   ,   29  ,   5   ,   89  ,   65  ,   191 ,   33  ,   755 ,   737 ,   3495    ,0 };
            const unsigned long dim588KuoInit[]     =   {   1   ,   1   ,   5   ,   1   ,   19  ,   27  ,   123 ,   67  ,   323 ,   731 ,   1753    ,   1773    ,   1731    ,0 };
            const unsigned long dim589KuoInit[]     =   {   1   ,   1   ,   5   ,   11  ,   23  ,   35  ,   21  ,   1   ,   107 ,   509 ,   1617    ,   1565    ,   1617    ,0 };
            const unsigned long dim590KuoInit[]     =   {   1   ,   1   ,   7   ,   15  ,   31  ,   7   ,   31  ,   241 ,   479 ,   457 ,   1353    ,   1531    ,   5939    ,0 };
            const unsigned long dim591KuoInit[]     =   {   1   ,   3   ,   1   ,   1   ,   31  ,   19  ,   71  ,   127 ,   379 ,   151 ,   739 ,   893 ,   7857    ,0 };
            const unsigned long dim592KuoInit[]     =   {   1   ,   1   ,   1   ,   3   ,   31  ,   41  ,   99  ,   55  ,   229 ,   585 ,   1881    ,   2267    ,   3555    ,0 };
            const unsigned long dim593KuoInit[]     =   {   1   ,   3   ,   3   ,   1   ,   25  ,   3   ,   55  ,   95  ,   409 ,   429 ,   527 ,   611 ,   6551    ,0 };
            const unsigned long dim594KuoInit[]     =   {   1   ,   3   ,   5   ,   3   ,   3   ,   59  ,   81  ,   83  ,   131 ,   981 ,   67  ,   295 ,   3313    ,0 };
            const unsigned long dim595KuoInit[]     =   {   1   ,   1   ,   1   ,   13  ,   23  ,   57  ,   47  ,   127 ,   307 ,   621 ,   865 ,   2589    ,   4033    ,0 };
            const unsigned long dim596KuoInit[]     =   {   1   ,   1   ,   7   ,   11  ,   7   ,   17  ,   15  ,   207 ,   89  ,   77  ,   1923    ,   3649    ,   119 ,0 };
            const unsigned long dim597KuoInit[]     =   {   1   ,   3   ,   5   ,   3   ,   23  ,   5   ,   87  ,   247 ,   233 ,   43  ,   475 ,   999 ,   6999    ,0 };
            const unsigned long dim598KuoInit[]     =   {   1   ,   1   ,   7   ,   13  ,   19  ,   49  ,   53  ,   21  ,   105 ,   181 ,   1569    ,   3151    ,   5089    ,0 };
            const unsigned long dim599KuoInit[]     =   {   1   ,   1   ,   5   ,   7   ,   11  ,   53  ,   119 ,   11  ,   301 ,   647 ,   759 ,   2849    ,   4541    ,0 };
            const unsigned long dim600KuoInit[]     =   {   1   ,   1   ,   7   ,   11  ,   31  ,   31  ,   83  ,   107 ,   373 ,   41  ,   921 ,   3935    ,   2859    ,0 };
            const unsigned long dim601KuoInit[]     =   {   1   ,   3   ,   5   ,   13  ,   13  ,   57  ,   121 ,   133 ,   355 ,   153 ,   607 ,   3199    ,   8097    ,0 };
            const unsigned long dim602KuoInit[]     =   {   1   ,   1   ,   5   ,   9   ,   5   ,   25  ,   113 ,   21  ,   17  ,   643 ,   415 ,   2705    ,   903 ,0 };
            const unsigned long dim603KuoInit[]     =   {   1   ,   1   ,   5   ,   1   ,   19  ,   3   ,   119 ,   199 ,   241 ,   957 ,   1203    ,   1291    ,   6355    ,0 };
            const unsigned long dim604KuoInit[]     =   {   1   ,   1   ,   3   ,   13  ,   19  ,   1   ,   11  ,   73  ,   387 ,   97  ,   783 ,   3875    ,   4573    ,0 };
            const unsigned long dim605KuoInit[]     =   {   1   ,   1   ,   7   ,   3   ,   27  ,   11  ,   99  ,   193 ,   311 ,   889 ,   535 ,   801 ,   4467    ,0 };
            const unsigned long dim606KuoInit[]     =   {   1   ,   1   ,   1   ,   1   ,   3   ,   53  ,   61  ,   245 ,   497 ,   127 ,   1213    ,   1147    ,   2225    ,0 };
            const unsigned long dim607KuoInit[]     =   {   1   ,   3   ,   5   ,   1   ,   9   ,   49  ,   21  ,   229 ,   481 ,   23  ,   47  ,   3959    ,   2991    ,0 };
            const unsigned long dim608KuoInit[]     =   {   1   ,   1   ,   5   ,   1   ,   9   ,   21  ,   123 ,   107 ,   359 ,   759 ,   993 ,   695 ,   5609    ,0 };
            const unsigned long dim609KuoInit[]     =   {   1   ,   3   ,   5   ,   11  ,   25  ,   55  ,   73  ,   245 ,   373 ,   525 ,   1399    ,   1727    ,   1763    ,0 };
            const unsigned long dim610KuoInit[]     =   {   1   ,   1   ,   5   ,   3   ,   21  ,   41  ,   83  ,   161 ,   43  ,   563 ,   1655    ,   3539    ,   7941    ,0 };
            const unsigned long dim611KuoInit[]     =   {   1   ,   3   ,   1   ,   13  ,   29  ,   59  ,   25  ,   119 ,   199 ,   375 ,   1507    ,   2095    ,   3281    ,0 };
            const unsigned long dim612KuoInit[]     =   {   1   ,   3   ,   3   ,   5   ,   21  ,   23  ,   63  ,   85  ,   115 ,   77  ,   745 ,   771 ,   3319    ,0 };
            const unsigned long dim613KuoInit[]     =   {   1   ,   1   ,   7   ,   3   ,   3   ,   37  ,   11  ,   233 ,   173 ,   51  ,   1723    ,   1441    ,   6529    ,0 };
            const unsigned long dim614KuoInit[]     =   {   1   ,   1   ,   1   ,   11  ,   1   ,   55  ,   113 ,   25  ,   369 ,   1023    ,   1691    ,   1797    ,   159 ,0 };
            const unsigned long dim615KuoInit[]     =   {   1   ,   3   ,   1   ,   15  ,   13  ,   43  ,   63  ,   113 ,   55  ,   351 ,   1459    ,   3117    ,   3499    ,0 };
            const unsigned long dim616KuoInit[]     =   {   1   ,   3   ,   7   ,   1   ,   19  ,   61  ,   17  ,   243 ,   203 ,   349 ,   1069    ,   105 ,   6015    ,0 };
            const unsigned long dim617KuoInit[]     =   {   1   ,   1   ,   5   ,   1   ,   23  ,   49  ,   7   ,   129 ,   99  ,   377 ,   1989    ,   531 ,   6315    ,0 };
            const unsigned long dim618KuoInit[]     =   {   1   ,   1   ,   7   ,   7   ,   11  ,   23  ,   107 ,   63  ,   163 ,   351 ,   1385    ,   127 ,   7439    ,0 };
            const unsigned long dim619KuoInit[]     =   {   1   ,   3   ,   5   ,   15  ,   15  ,   43  ,   81  ,   229 ,   21  ,   179 ,   1357    ,   3813    ,   7345    ,0 };
            const unsigned long dim620KuoInit[]     =   {   1   ,   1   ,   1   ,   15  ,   23  ,   19  ,   89  ,   251 ,   161 ,   443 ,   983 ,   1059    ,   1241    ,0 };
            const unsigned long dim621KuoInit[]     =   {   1   ,   3   ,   3   ,   9   ,   7   ,   17  ,   11  ,   239 ,   325 ,   117 ,   893 ,   917 ,   6327    ,0 };
            const unsigned long dim622KuoInit[]     =   {   1   ,   1   ,   7   ,   1   ,   31  ,   33  ,   23  ,   241 ,   255 ,   991 ,   903 ,   3973    ,   3935    ,0 };
            const unsigned long dim623KuoInit[]     =   {   1   ,   3   ,   7   ,   3   ,   11  ,   17  ,   61  ,   219 ,   461 ,   437 ,   1843    ,   1845    ,   771 ,0 };
            const unsigned long dim624KuoInit[]     =   {   1   ,   3   ,   7   ,   5   ,   7   ,   61  ,   79  ,   225 ,   215 ,   411 ,   885 ,   3949    ,   2757    ,0 };
            const unsigned long dim625KuoInit[]     =   {   1   ,   3   ,   7   ,   3   ,   3   ,   15  ,   25  ,   49  ,   351 ,   671 ,   1573    ,   3643    ,   2383    ,0 };
            const unsigned long dim626KuoInit[]     =   {   1   ,   1   ,   7   ,   1   ,   29  ,   1   ,   121 ,   61  ,   49  ,   833 ,   1363    ,   427 ,   3237    ,0 };
            const unsigned long dim627KuoInit[]     =   {   1   ,   3   ,   3   ,   5   ,   11  ,   9   ,   79  ,   113 ,   341 ,   25  ,   1797    ,   1641    ,   6899    ,0 };
            const unsigned long dim628KuoInit[]     =   {   1   ,   1   ,   5   ,   9   ,   11  ,   15  ,   71  ,   169 ,   269 ,   359 ,   1991    ,   3673    ,   601 ,0 };
            const unsigned long dim629KuoInit[]     =   {   1   ,   1   ,   1   ,   11  ,   31  ,   7   ,   5   ,   165 ,   23  ,   917 ,   617 ,   4059    ,   2599    ,0 };
            const unsigned long dim630KuoInit[]     =   {   1   ,   1   ,   7   ,   9   ,   19  ,   25  ,   119 ,   103 ,   115 ,   247 ,   1265    ,   859 ,   5543    ,0 };
            const unsigned long dim631KuoInit[]     =   {   1   ,   1   ,   7   ,   7   ,   15  ,   27  ,   45  ,   29  ,   67  ,   989 ,   753 ,   3215    ,   6281    ,0 };
            const unsigned long dim632KuoInit[]     =   {   1   ,   3   ,   1   ,   7   ,   11  ,   15  ,   35  ,   39  ,   61  ,   483 ,   1427    ,   193 ,   3469    ,0 };
            const unsigned long dim633KuoInit[]     =   {   1   ,   3   ,   5   ,   9   ,   21  ,   41  ,   21  ,   223 ,   233 ,   341 ,   177 ,   1167    ,   3101    ,0 };
            const unsigned long dim634KuoInit[]     =   {   1   ,   3   ,   3   ,   5   ,   19  ,   25  ,   53  ,   247 ,   29  ,   909 ,   2047    ,   1683    ,   4433    ,0 };
            const unsigned long dim635KuoInit[]     =   {   1   ,   3   ,   3   ,   9   ,   31  ,   7   ,   71  ,   85  ,   501 ,   447 ,   113 ,   3465    ,   3445    ,0 };
            const unsigned long dim636KuoInit[]     =   {   1   ,   1   ,   3   ,   15  ,   19  ,   9   ,   127 ,   69  ,   429 ,   451 ,   531 ,   1319    ,   507 ,0 };
            const unsigned long dim637KuoInit[]     =   {   1   ,   3   ,   1   ,   3   ,   29  ,   1   ,   53  ,   63  ,   267 ,   293 ,   333 ,   2749    ,   1785    ,0 };
            const unsigned long dim638KuoInit[]     =   {   1   ,   3   ,   1   ,   7   ,   5   ,   43  ,   71  ,   165 ,   325 ,   309 ,   921 ,   2277    ,   6423    ,0 };
            const unsigned long dim639KuoInit[]     =   {   1   ,   1   ,   7   ,   5   ,   23  ,   31  ,   63  ,   35  ,   221 ,   733 ,   1901    ,   153 ,   3577    ,0 };
            const unsigned long dim640KuoInit[]     =   {   1   ,   3   ,   7   ,   11  ,   7   ,   5   ,   39  ,   141 ,   293 ,   237 ,   501 ,   1759    ,   7763    ,0 };
            const unsigned long dim641KuoInit[]     =   {   1   ,   1   ,   3   ,   15  ,   1   ,   49  ,   5   ,   7   ,   239 ,   1021    ,   1215    ,   3857    ,   4637    ,0 };
            const unsigned long dim642KuoInit[]     =   {   1   ,   1   ,   5   ,   1   ,   1   ,   31  ,   77  ,   243 ,   155 ,   505 ,   855 ,   3117    ,   815 ,0 };
            const unsigned long dim643KuoInit[]     =   {   1   ,   1   ,   7   ,   3   ,   3   ,   33  ,   55  ,   249 ,   185 ,   443 ,   543 ,   1333    ,   5041    ,0 };
            const unsigned long dim644KuoInit[]     =   {   1   ,   3   ,   5   ,   11  ,   13  ,   13  ,   83  ,   55  ,   443 ,   723 ,   135 ,   3073    ,   4215    ,0 };
            const unsigned long dim645KuoInit[]     =   {   1   ,   3   ,   5   ,   9   ,   15  ,   5   ,   13  ,   209 ,   313 ,   891 ,   431 ,   4009    ,   7531    ,0 };
            const unsigned long dim646KuoInit[]     =   {   1   ,   1   ,   5   ,   15  ,   13  ,   37  ,   91  ,   117 ,   321 ,   791 ,   2045    ,   4073    ,   6787    ,0 };
            const unsigned long dim647KuoInit[]     =   {   1   ,   1   ,   3   ,   7   ,   1   ,   43  ,   61  ,   121 ,   393 ,   797 ,   1347    ,   491 ,   341 ,0 };
            const unsigned long dim648KuoInit[]     =   {   1   ,   1   ,   7   ,   3   ,   1   ,   31  ,   63  ,   205 ,   43  ,   359 ,   325 ,   3329    ,   4961    ,0 };
            const unsigned long dim649KuoInit[]     =   {   1   ,   1   ,   7   ,   11  ,   27  ,   21  ,   83  ,   151 ,   97  ,   45  ,   1839    ,   3221    ,   2575    ,0 };
            const unsigned long dim650KuoInit[]     =   {   1   ,   3   ,   7   ,   3   ,   13  ,   7   ,   97  ,   89  ,   277 ,   351 ,   1493    ,   1161    ,   6629    ,0 };
            const unsigned long dim651KuoInit[]     =   {   1   ,   3   ,   7   ,   11  ,   27  ,   17  ,   19  ,   211 ,   373 ,   73  ,   1325    ,   183 ,   1073    ,0 };
            const unsigned long dim652KuoInit[]     =   {   1   ,   3   ,   3   ,   13  ,   9   ,   23  ,   111 ,   235 ,   153 ,   857 ,   1669    ,   3423    ,   3817    ,0 };
            const unsigned long dim653KuoInit[]     =   {   1   ,   1   ,   1   ,   5   ,   25  ,   63  ,   99  ,   11  ,   283 ,   755 ,   999 ,   2795    ,   7795    ,0 };
            const unsigned long dim654KuoInit[]     =   {   1   ,   1   ,   5   ,   11  ,   3   ,   53  ,   127 ,   73  ,   61  ,   91  ,   1099    ,   921 ,   1787    ,0 };
            const unsigned long dim655KuoInit[]     =   {   1   ,   3   ,   5   ,   1   ,   9   ,   39  ,   9   ,   181 ,   287 ,   545 ,   1343    ,   2587    ,   961 ,0 };
            const unsigned long dim656KuoInit[]     =   {   1   ,   3   ,   7   ,   5   ,   9   ,   47  ,   7   ,   113 ,   69  ,   531 ,   753 ,   2225    ,   5705    ,0 };
            const unsigned long dim657KuoInit[]     =   {   1   ,   1   ,   3   ,   13  ,   29  ,   39  ,   71  ,   223 ,   179 ,   871 ,   1743    ,   2645    ,   7355    ,0 };
            const unsigned long dim658KuoInit[]     =   {   1   ,   1   ,   7   ,   5   ,   1   ,   41  ,   87  ,   251 ,   303 ,   271 ,   791 ,   2825    ,   2723    ,0 };
            const unsigned long dim659KuoInit[]     =   {   1   ,   3   ,   1   ,   11  ,   31  ,   15  ,   119 ,   127 ,   451 ,   161 ,   1909    ,   3591    ,   2811    ,0 };
            const unsigned long dim660KuoInit[]     =   {   1   ,   1   ,   3   ,   13  ,   31  ,   27  ,   111 ,   119 ,   423 ,   419 ,   1659    ,   941 ,   6123    ,0 };
            const unsigned long dim661KuoInit[]     =   {   1   ,   1   ,   1   ,   7   ,   27  ,   39  ,   27  ,   61  ,   417 ,   629 ,   1411    ,   1489    ,   5915    ,0 };
            const unsigned long dim662KuoInit[]     =   {   1   ,   3   ,   3   ,   1   ,   29  ,   37  ,   111 ,   13  ,   129 ,   831 ,   1423    ,   3859    ,   7009    ,0 };
            const unsigned long dim663KuoInit[]     =   {   1   ,   1   ,   1   ,   3   ,   3   ,   51  ,   67  ,   45  ,   289 ,   715 ,   183 ,   357 ,   3661    ,0 };
            const unsigned long dim664KuoInit[]     =   {   1   ,   3   ,   3   ,   1   ,   23  ,   49  ,   55  ,   99  ,   77  ,   795 ,   429 ,   1815    ,   1867    ,0 };
            const unsigned long dim665KuoInit[]     =   {   1   ,   1   ,   5   ,   7   ,   1   ,   37  ,   45  ,   27  ,   57  ,   937 ,   1583    ,   899 ,   819 ,0 };
            const unsigned long dim666KuoInit[]     =   {   1   ,   3   ,   5   ,   7   ,   7   ,   37  ,   1   ,   109 ,   217 ,   203 ,   207 ,   2243    ,   2337    ,0 };
            const unsigned long dim667KuoInit[]     =   {   1   ,   1   ,   3   ,   15  ,   27  ,   51  ,   119 ,   45  ,   289 ,   395 ,   95  ,   2425    ,   7421    ,0 };
            const unsigned long dim668KuoInit[]     =   {   1   ,   3   ,   1   ,   9   ,   27  ,   61  ,   41  ,   227 ,   279 ,   31  ,   1105    ,   153 ,   5045    ,0 };
            const unsigned long dim669KuoInit[]     =   {   1   ,   1   ,   7   ,   15  ,   31  ,   43  ,   109 ,   67  ,   347 ,   611 ,   1811    ,   2969    ,   3545    ,0 };
            const unsigned long dim670KuoInit[]     =   {   1   ,   3   ,   1   ,   7   ,   17  ,   15  ,   103 ,   35  ,   407 ,   935 ,   1159    ,   2357    ,   6559    ,0 };
            const unsigned long dim671KuoInit[]     =   {   1   ,   1   ,   1   ,   7   ,   5   ,   7   ,   101 ,   43  ,   301 ,   189 ,   641 ,   3015    ,   6955    ,0 };
            const unsigned long dim672KuoInit[]     =   {   1   ,   3   ,   1   ,   1   ,   3   ,   41  ,   61  ,   43  ,   213 ,   595 ,   955 ,   377 ,   7527    ,0 };
            const unsigned long dim673KuoInit[]     =   {   1   ,   3   ,   3   ,   15  ,   19  ,   45  ,   59  ,   57  ,   425 ,   683 ,   1187    ,   221 ,   4265    ,0 };
            const unsigned long dim674KuoInit[]     =   {   1   ,   3   ,   5   ,   3   ,   27  ,   47  ,   85  ,   73  ,   67  ,   509 ,   555 ,   3887    ,   1883    ,0 };
            const unsigned long dim675KuoInit[]     =   {   1   ,   3   ,   5   ,   13  ,   19  ,   45  ,   83  ,   191 ,   117 ,   723 ,   701 ,   2783    ,   389 ,0 };
            const unsigned long dim676KuoInit[]     =   {   1   ,   1   ,   1   ,   3   ,   17  ,   21  ,   125 ,   13  ,   387 ,   613 ,   1647    ,   2555    ,   2435    ,0 };
            const unsigned long dim677KuoInit[]     =   {   1   ,   3   ,   3   ,   11  ,   21  ,   53  ,   105 ,   5   ,   493 ,   167 ,   1195    ,   2815    ,   5761    ,0 };
            const unsigned long dim678KuoInit[]     =   {   1   ,   1   ,   1   ,   7   ,   27  ,   17  ,   107 ,   83  ,   31  ,   787 ,   1541    ,   3761    ,   4083    ,0 };
            const unsigned long dim679KuoInit[]     =   {   1   ,   3   ,   5   ,   11  ,   5   ,   15  ,   17  ,   219 ,   273 ,   207 ,   119 ,   1693    ,   373 ,0 };
            const unsigned long dim680KuoInit[]     =   {   1   ,   1   ,   5   ,   15  ,   7   ,   11  ,   9   ,   103 ,   253 ,   377 ,   1667    ,   1159    ,   5669    ,0 };
            const unsigned long dim681KuoInit[]     =   {   1   ,   3   ,   7   ,   5   ,   17  ,   55  ,   23  ,   77  ,   341 ,   419 ,   1165    ,   1625    ,   6609    ,0 };
            const unsigned long dim682KuoInit[]     =   {   1   ,   3   ,   1   ,   15  ,   29  ,   7   ,   101 ,   183 ,   181 ,   111 ,   751 ,   247 ,   4205    ,0 };
            const unsigned long dim683KuoInit[]     =   {   1   ,   1   ,   1   ,   11  ,   3   ,   5   ,   91  ,   43  ,   247 ,   661 ,   621 ,   2791    ,   4245    ,0 };
            const unsigned long dim684KuoInit[]     =   {   1   ,   1   ,   1   ,   11  ,   11  ,   63  ,   25  ,   239 ,   39  ,   311 ,   1603    ,   1179    ,   4959    ,0 };
            const unsigned long dim685KuoInit[]     =   {   1   ,   3   ,   1   ,   7   ,   23  ,   61  ,   49  ,   99  ,   287 ,   251 ,   1053    ,   3819    ,   5575    ,0 };
            const unsigned long dim686KuoInit[]     =   {   1   ,   1   ,   1   ,   7   ,   21  ,   17  ,   33  ,   163 ,   503 ,   537 ,   1159    ,   2473    ,   8113    ,0 };
            const unsigned long dim687KuoInit[]     =   {   1   ,   1   ,   5   ,   13  ,   23  ,   57  ,   57  ,   89  ,   503 ,   299 ,   1319    ,   2177    ,   8155    ,0 };
            const unsigned long dim688KuoInit[]     =   {   1   ,   3   ,   3   ,   5   ,   5   ,   5   ,   61  ,   103 ,   247 ,   339 ,   1137    ,   3717    ,   747 ,0 };
            const unsigned long dim689KuoInit[]     =   {   1   ,   3   ,   1   ,   7   ,   17  ,   55  ,   41  ,   155 ,   363 ,   873 ,   59  ,   2093    ,   4255    ,0 };
            const unsigned long dim690KuoInit[]     =   {   1   ,   3   ,   3   ,   7   ,   9   ,   25  ,   93  ,   87  ,   479 ,   477 ,   2047    ,   321 ,   2277    ,0 };
            const unsigned long dim691KuoInit[]     =   {   1   ,   1   ,   3   ,   11  ,   7   ,   37  ,   41  ,   237 ,   121 ,   187 ,   929 ,   1823    ,   313 ,0 };
            const unsigned long dim692KuoInit[]     =   {   1   ,   1   ,   1   ,   11  ,   29  ,   41  ,   79  ,   11  ,   295 ,   715 ,   979 ,   2649    ,   915 ,0 };
            const unsigned long dim693KuoInit[]     =   {   1   ,   1   ,   3   ,   13  ,   5   ,   9   ,   55  ,   159 ,   343 ,   887 ,   235 ,   53  ,   593 ,0 };
            const unsigned long dim694KuoInit[]     =   {   1   ,   1   ,   1   ,   3   ,   13  ,   23  ,   105 ,   125 ,   405 ,   521 ,   897 ,   559 ,   1561    ,0 };
            const unsigned long dim695KuoInit[]     =   {   1   ,   1   ,   3   ,   9   ,   23  ,   33  ,   121 ,   183 ,   181 ,   141 ,   1103    ,   821 ,   7193    ,0 };
            const unsigned long dim696KuoInit[]     =   {   1   ,   1   ,   3   ,   9   ,   23  ,   1   ,   113 ,   63  ,   59  ,   783 ,   1525    ,   115 ,   3193    ,0 };
            const unsigned long dim697KuoInit[]     =   {   1   ,   3   ,   1   ,   9   ,   23  ,   29  ,   83  ,   237 ,   83  ,   477 ,   1163    ,   2951    ,   3523    ,0 };
            const unsigned long dim698KuoInit[]     =   {   1   ,   3   ,   5   ,   1   ,   21  ,   25  ,   55  ,   79  ,   39  ,   45  ,   1509    ,   2127    ,   5259    ,0 };
            const unsigned long dim699KuoInit[]     =   {   1   ,   1   ,   7   ,   9   ,   31  ,   61  ,   83  ,   145 ,   89  ,   723 ,   393 ,   1543    ,   7765    ,0 };
            const unsigned long dim700KuoInit[]     =   {   1   ,   1   ,   5   ,   15  ,   21  ,   17  ,   73  ,   189 ,   473 ,   1023    ,   1627    ,   2305    ,   7127    ,0 };
            const unsigned long dim701KuoInit[]     =   {   1   ,   3   ,   1   ,   7   ,   7   ,   23  ,   61  ,   203 ,   373 ,   941 ,   975 ,   3225    ,   7659    ,0 };
            const unsigned long dim702KuoInit[]     =   {   1   ,   3   ,   1   ,   15  ,   31  ,   19  ,   99  ,   197 ,   113 ,   131 ,   263 ,   2427    ,   2085    ,0 };
            const unsigned long dim703KuoInit[]     =   {   1   ,   1   ,   1   ,   13  ,   25  ,   27  ,   3   ,   139 ,   351 ,   267 ,   1583    ,   913 ,   5497    ,0 };
            const unsigned long dim704KuoInit[]     =   {   1   ,   3   ,   5   ,   11  ,   7   ,   1   ,   65  ,   85  ,   271 ,   151 ,   1003    ,   889 ,   6127    ,0 };
            const unsigned long dim705KuoInit[]     =   {   1   ,   3   ,   7   ,   5   ,   21  ,   27  ,   39  ,   147 ,   139 ,   81  ,   1377    ,   1969    ,   6171    ,0 };
            const unsigned long dim706KuoInit[]     =   {   1   ,   3   ,   5   ,   13  ,   1   ,   11  ,   101 ,   27  ,   285 ,   939 ,   1055    ,   3237    ,   7063    ,0 };
            const unsigned long dim707KuoInit[]     =   {   1   ,   3   ,   3   ,   7   ,   23  ,   43  ,   67  ,   89  ,   201 ,   393 ,   277 ,   3531    ,   4003    ,0 };
            const unsigned long dim708KuoInit[]     =   {   1   ,   3   ,   7   ,   9   ,   3   ,   13  ,   103 ,   241 ,   161 ,   111 ,   61  ,   4001    ,   6291    ,0 };
            const unsigned long dim709KuoInit[]     =   {   1   ,   1   ,   5   ,   7   ,   5   ,   27  ,   57  ,   87  ,   313 ,   471 ,   227 ,   1395    ,   2821    ,0 };
            const unsigned long dim710KuoInit[]     =   {   1   ,   3   ,   1   ,   5   ,   11  ,   35  ,   11  ,   157 ,   257 ,   679 ,   245 ,   329 ,   2423    ,0 };
            const unsigned long dim711KuoInit[]     =   {   1   ,   1   ,   7   ,   11  ,   15  ,   5   ,   29  ,   229 ,   365 ,   911 ,   477 ,   219 ,   4851    ,0 };
            const unsigned long dim712KuoInit[]     =   {   1   ,   3   ,   1   ,   9   ,   5   ,   57  ,   87  ,   19  ,   93  ,   403 ,   1729    ,   1843    ,   7301    ,0 };
            const unsigned long dim713KuoInit[]     =   {   1   ,   1   ,   3   ,   3   ,   27  ,   29  ,   11  ,   99  ,   149 ,   917 ,   709 ,   3761    ,   6947    ,0 };
            const unsigned long dim714KuoInit[]     =   {   1   ,   1   ,   5   ,   1   ,   19  ,   63  ,   69  ,   71  ,   209 ,   423 ,   2015    ,   3905    ,   2077    ,0 };
            const unsigned long dim715KuoInit[]     =   {   1   ,   3   ,   7   ,   7   ,   7   ,   3   ,   9   ,   63  ,   137 ,   781 ,   1825    ,   1331    ,   2987    ,0 };
            const unsigned long dim716KuoInit[]     =   {   1   ,   3   ,   7   ,   11  ,   15  ,   23  ,   7   ,   205 ,   27  ,   619 ,   225 ,   297 ,   8179    ,0 };
            const unsigned long dim717KuoInit[]     =   {   1   ,   3   ,   3   ,   13  ,   11  ,   15  ,   11  ,   185 ,   343 ,   269 ,   1415    ,   843 ,   4001    ,0 };
            const unsigned long dim718KuoInit[]     =   {   1   ,   3   ,   3   ,   1   ,   15  ,   47  ,   101 ,   49  ,   25  ,   933 ,   903 ,   1927    ,   6641    ,0 };
            const unsigned long dim719KuoInit[]     =   {   1   ,   3   ,   3   ,   13  ,   7   ,   45  ,   29  ,   51  ,   143 ,   549 ,   119 ,   235 ,   7345    ,0 };
            const unsigned long dim720KuoInit[]     =   {   1   ,   1   ,   3   ,   9   ,   29  ,   45  ,   53  ,   133 ,   173 ,   543 ,   545 ,   2015    ,   149 ,0 };
            const unsigned long dim721KuoInit[]     =   {   1   ,   1   ,   3   ,   11  ,   3   ,   3   ,   109 ,   193 ,   349 ,   243 ,   1381    ,   3547    ,   4387    ,0 };
            const unsigned long dim722KuoInit[]     =   {   1   ,   3   ,   7   ,   9   ,   7   ,   3   ,   37  ,   121 ,   79  ,   59  ,   1453    ,   3273    ,   477 ,0 };
            const unsigned long dim723KuoInit[]     =   {   1   ,   3   ,   5   ,   15  ,   19  ,   15  ,   83  ,   5   ,   203 ,   603 ,   373 ,   955 ,   1587    ,0 };
            const unsigned long dim724KuoInit[]     =   {   1   ,   1   ,   5   ,   11  ,   25  ,   41  ,   89  ,   47  ,   313 ,   443 ,   1799    ,   1827    ,   5881    ,0 };
            const unsigned long dim725KuoInit[]     =   {   1   ,   1   ,   3   ,   5   ,   9   ,   7   ,   15  ,   87  ,   285 ,   509 ,   289 ,   855 ,   3131    ,0 };
            const unsigned long dim726KuoInit[]     =   {   1   ,   1   ,   3   ,   11  ,   13  ,   57  ,   41  ,   235 ,   357 ,   213 ,   1867    ,   1961    ,   2759    ,0 };
            const unsigned long dim727KuoInit[]     =   {   1   ,   3   ,   5   ,   9   ,   25  ,   15  ,   111 ,   197 ,   479 ,   539 ,   1833    ,   105 ,   7245    ,0 };
            const unsigned long dim728KuoInit[]     =   {   1   ,   1   ,   5   ,   5   ,   23  ,   47  ,   101 ,   97  ,   97  ,   261 ,   1301    ,   3895    ,   5425    ,0 };
            const unsigned long dim729KuoInit[]     =   {   1   ,   1   ,   7   ,   11  ,   13  ,   49  ,   39  ,   233 ,   7   ,   965 ,   2015    ,   1341    ,   941 ,0 };
            const unsigned long dim730KuoInit[]     =   {   1   ,   1   ,   5   ,   7   ,   19  ,   63  ,   91  ,   77  ,   469 ,   367 ,   1849    ,   1213    ,   7611    ,0 };
            const unsigned long dim731KuoInit[]     =   {   1   ,   3   ,   3   ,   15  ,   17  ,   57  ,   27  ,   107 ,   133 ,   431 ,   445 ,   2885    ,   5963    ,0 };
            const unsigned long dim732KuoInit[]     =   {   1   ,   1   ,   3   ,   1   ,   27  ,   37  ,   119 ,   241 ,   403 ,   239 ,   833 ,   3057    ,   5557    ,0 };
            const unsigned long dim733KuoInit[]     =   {   1   ,   3   ,   3   ,   13  ,   11  ,   47  ,   35  ,   41  ,   243 ,   143 ,   1923    ,   1883    ,   2543    ,0 };
            const unsigned long dim734KuoInit[]     =   {   1   ,   1   ,   5   ,   5   ,   3   ,   15  ,   45  ,   205 ,   179 ,   515 ,   161 ,   895 ,   3679    ,0 };
            const unsigned long dim735KuoInit[]     =   {   1   ,   3   ,   1   ,   1   ,   5   ,   39  ,   99  ,   221 ,   115 ,   983 ,   1913    ,   665 ,   7835    ,0 };
            const unsigned long dim736KuoInit[]     =   {   1   ,   3   ,   5   ,   15  ,   5   ,   5   ,   111 ,   45  ,   373 ,   143 ,   1107    ,   1413    ,   2935    ,0 };
            const unsigned long dim737KuoInit[]     =   {   1   ,   3   ,   7   ,   11  ,   17  ,   53  ,   77  ,   159 ,   319 ,   279 ,   835 ,   139 ,   3633    ,0 };
            const unsigned long dim738KuoInit[]     =   {   1   ,   1   ,   7   ,   15  ,   5   ,   57  ,   105 ,   139 ,   491 ,   409 ,   51  ,   2875    ,   427 ,0 };
            const unsigned long dim739KuoInit[]     =   {   1   ,   1   ,   1   ,   9   ,   9   ,   9   ,   93  ,   45  ,   265 ,   307 ,   589 ,   557 ,   3971    ,0 };
            const unsigned long dim740KuoInit[]     =   {   1   ,   3   ,   5   ,   3   ,   23  ,   45  ,   5   ,   9   ,   123 ,   841 ,   679 ,   2213    ,   4507    ,0 };
            const unsigned long dim741KuoInit[]     =   {   1   ,   3   ,   5   ,   3   ,   27  ,   55  ,   55  ,   19  ,   405 ,   43  ,   1897    ,   3515    ,   683 ,0 };
            const unsigned long dim742KuoInit[]     =   {   1   ,   1   ,   7   ,   15  ,   23  ,   47  ,   121 ,   105 ,   341 ,   707 ,   799 ,   1843    ,   7011    ,0 };
            const unsigned long dim743KuoInit[]     =   {   1   ,   1   ,   3   ,   13  ,   23  ,   51  ,   71  ,   161 ,   415 ,   819 ,   729 ,   297 ,   4461    ,0 };
            const unsigned long dim744KuoInit[]     =   {   1   ,   3   ,   5   ,   15  ,   13  ,   39  ,   43  ,   103 ,   251 ,   583 ,   775 ,   877 ,   4947    ,0 };
            const unsigned long dim745KuoInit[]     =   {   1   ,   3   ,   7   ,   5   ,   19  ,   61  ,   121 ,   49  ,   75  ,   963 ,   297 ,   3751    ,   911 ,0 };
            const unsigned long dim746KuoInit[]     =   {   1   ,   3   ,   5   ,   7   ,   15  ,   51  ,   23  ,   23  ,   453 ,   675 ,   1575    ,   473 ,   3763    ,0 };
            const unsigned long dim747KuoInit[]     =   {   1   ,   1   ,   7   ,   11  ,   9   ,   37  ,   69  ,   137 ,   251 ,   89  ,   1823    ,   1289    ,   2983    ,0 };
            const unsigned long dim748KuoInit[]     =   {   1   ,   3   ,   3   ,   5   ,   11  ,   39  ,   39  ,   231 ,   427 ,   633 ,   1453    ,   579 ,   5823    ,0 };
            const unsigned long dim749KuoInit[]     =   {   1   ,   3   ,   7   ,   11  ,   31  ,   57  ,   87  ,   81  ,   161 ,   929 ,   1293    ,   3215    ,   4803    ,0 };
            const unsigned long dim750KuoInit[]     =   {   1   ,   3   ,   5   ,   5   ,   27  ,   33  ,   11  ,   151 ,   25  ,   1017    ,   41  ,   61  ,   4381    ,0 };
            const unsigned long dim751KuoInit[]     =   {   1   ,   3   ,   1   ,   15  ,   23  ,   25  ,   7   ,   21  ,   231 ,   407 ,   817 ,   2089    ,   839 ,0 };
            const unsigned long dim752KuoInit[]     =   {   1   ,   1   ,   7   ,   1   ,   5   ,   51  ,   117 ,   59  ,   251 ,   861 ,   727 ,   481 ,   2207    ,0 };
            const unsigned long dim753KuoInit[]     =   {   1   ,   1   ,   5   ,   3   ,   21  ,   19  ,   97  ,   199 ,   135 ,   203 ,   1177    ,   2751    ,   2863    ,0 };
            const unsigned long dim754KuoInit[]     =   {   1   ,   3   ,   7   ,   3   ,   27  ,   3   ,   5   ,   109 ,   17  ,   933 ,   911 ,   921 ,   1173    ,0 };
            const unsigned long dim755KuoInit[]     =   {   1   ,   3   ,   7   ,   1   ,   7   ,   1   ,   121 ,   205 ,   381 ,   971 ,   255 ,   3313    ,   5479    ,0 };
            const unsigned long dim756KuoInit[]     =   {   1   ,   3   ,   1   ,   1   ,   17  ,   59  ,   3   ,   5   ,   271 ,   149 ,   111 ,   3707    ,   7893    ,0 };
            const unsigned long dim757KuoInit[]     =   {   1   ,   1   ,   1   ,   1   ,   17  ,   61  ,   47  ,   117 ,   187 ,   839 ,   435 ,   399 ,   3173    ,0 };
            const unsigned long dim758KuoInit[]     =   {   1   ,   3   ,   3   ,   7   ,   5   ,   9   ,   119 ,   191 ,   197 ,   791 ,   375 ,   3361    ,   6677    ,0 };
            const unsigned long dim759KuoInit[]     =   {   1   ,   1   ,   5   ,   3   ,   21  ,   49  ,   69  ,   249 ,   39  ,   69  ,   673 ,   109 ,   7051    ,0 };
            const unsigned long dim760KuoInit[]     =   {   1   ,   1   ,   3   ,   11  ,   21  ,   37  ,   121 ,   93  ,   261 ,   5   ,   1269    ,   767 ,   2801    ,0 };
            const unsigned long dim761KuoInit[]     =   {   1   ,   1   ,   5   ,   9   ,   31  ,   49  ,   9   ,   87  ,   375 ,   773 ,   1517    ,   3057    ,   2169    ,0 };
            const unsigned long dim762KuoInit[]     =   {   1   ,   1   ,   5   ,   3   ,   13  ,   33  ,   31  ,   87  ,   353 ,   991 ,   1809    ,   859 ,   6117    ,0 };
            const unsigned long dim763KuoInit[]     =   {   1   ,   1   ,   5   ,   11  ,   3   ,   21  ,   103 ,   41  ,   129 ,   137 ,   1567    ,   3483    ,   5317    ,0 };
            const unsigned long dim764KuoInit[]     =   {   1   ,   3   ,   5   ,   1   ,   3   ,   31  ,   5   ,   15  ,   109 ,   723 ,   1439    ,   2543    ,   2377    ,0 };
            const unsigned long dim765KuoInit[]     =   {   1   ,   3   ,   3   ,   13  ,   11  ,   17  ,   9   ,   243 ,   233 ,   543 ,   903 ,   2775    ,   6877    ,0 };
            const unsigned long dim766KuoInit[]     =   {   1   ,   1   ,   7   ,   1   ,   23  ,   45  ,   61  ,   137 ,   207 ,   457 ,   245 ,   3833    ,   6533    ,0 };
            const unsigned long dim767KuoInit[]     =   {   1   ,   3   ,   1   ,   1   ,   31  ,   7   ,   71  ,   153 ,   501 ,   749 ,   1477    ,   621 ,   6577    ,0 };
            const unsigned long dim768KuoInit[]     =   {   1   ,   3   ,   1   ,   13  ,   3   ,   45  ,   3   ,   233 ,   85  ,   847 ,   1031    ,   141 ,   5029    ,0 };
            const unsigned long dim769KuoInit[]     =   {   1   ,   3   ,   1   ,   13  ,   17  ,   35  ,   49  ,   157 ,   243 ,   175 ,   905 ,   539 ,   8059    ,0 };
            const unsigned long dim770KuoInit[]     =   {   1   ,   3   ,   1   ,   11  ,   19  ,   45  ,   31  ,   49  ,   399 ,   341 ,   2047    ,   1527    ,   5993    ,0 };
            const unsigned long dim771KuoInit[]     =   {   1   ,   3   ,   1   ,   3   ,   5   ,   5   ,   33  ,   163 ,   239 ,   607 ,   1367    ,   2747    ,   1513    ,0 };
            const unsigned long dim772KuoInit[]     =   {   1   ,   3   ,   3   ,   11  ,   9   ,   29  ,   33  ,   255 ,   29  ,   955 ,   367 ,   1709    ,   4993    ,0 };
            const unsigned long dim773KuoInit[]     =   {   1   ,   3   ,   7   ,   5   ,   19  ,   35  ,   11  ,   47  ,   109 ,   235 ,   1125    ,   4019    ,   5395    ,0 };
            const unsigned long dim774KuoInit[]     =   {   1   ,   1   ,   7   ,   1   ,   19  ,   29  ,   43  ,   1   ,   199 ,   479 ,   1025    ,   1065    ,   2203    ,0 };
            const unsigned long dim775KuoInit[]     =   {   1   ,   3   ,   1   ,   9   ,   17  ,   13  ,   45  ,   167 ,   125 ,   869 ,   1667    ,   1083    ,   4015    ,0 };
            const unsigned long dim776KuoInit[]     =   {   1   ,   1   ,   5   ,   11  ,   27  ,   45  ,   115 ,   199 ,   375 ,   761 ,   1641    ,   3761    ,   5509    ,0 };
            const unsigned long dim777KuoInit[]     =   {   1   ,   1   ,   7   ,   15  ,   9   ,   47  ,   47  ,   133 ,   401 ,   555 ,   1077    ,   1381    ,   1903    ,0 };
            const unsigned long dim778KuoInit[]     =   {   1   ,   3   ,   3   ,   15  ,   3   ,   45  ,   19  ,   49  ,   377 ,   135 ,   239 ,   3605    ,   1053    ,0 };
            const unsigned long dim779KuoInit[]     =   {   1   ,   3   ,   1   ,   11  ,   17  ,   5   ,   83  ,   101 ,   247 ,   267 ,   1337    ,   2649    ,   513 ,0 };
            const unsigned long dim780KuoInit[]     =   {   1   ,   3   ,   1   ,   11  ,   13  ,   41  ,   41  ,   113 ,   243 ,   463 ,   1077    ,   565 ,   2871    ,0 };
            const unsigned long dim781KuoInit[]     =   {   1   ,   3   ,   5   ,   11  ,   19  ,   63  ,   81  ,   51  ,   139 ,   921 ,   2021    ,   417 ,   1305    ,0 };
            const unsigned long dim782KuoInit[]     =   {   1   ,   3   ,   3   ,   1   ,   23  ,   61  ,   13  ,   123 ,   135 ,   521 ,   1963    ,   3085    ,   4395    ,0 };
            const unsigned long dim783KuoInit[]     =   {   1   ,   3   ,   1   ,   1   ,   15  ,   61  ,   17  ,   245 ,   479 ,   77  ,   287 ,   3043    ,   2647    ,0 };
            const unsigned long dim784KuoInit[]     =   {   1   ,   1   ,   7   ,   15  ,   29  ,   5   ,   125 ,   123 ,   149 ,   919 ,   1779    ,   3763    ,   1479    ,0 };
            const unsigned long dim785KuoInit[]     =   {   1   ,   1   ,   3   ,   1   ,   19  ,   49  ,   27  ,   127 ,   253 ,   279 ,   739 ,   147 ,   6835    ,0 };
            const unsigned long dim786KuoInit[]     =   {   1   ,   3   ,   5   ,   11  ,   21  ,   27  ,   33  ,   85  ,   419 ,   441 ,   1341    ,   289 ,   2759    ,0 };
            const unsigned long dim787KuoInit[]     =   {   1   ,   3   ,   1   ,   3   ,   3   ,   9   ,   17  ,   99  ,   375 ,   251 ,   1823    ,   2433    ,   1525    ,0 };
            const unsigned long dim788KuoInit[]     =   {   1   ,   3   ,   1   ,   3   ,   5   ,   55  ,   25  ,   225 ,   473 ,   825 ,   1061    ,   2987    ,   4397    ,0 };
            const unsigned long dim789KuoInit[]     =   {   1   ,   1   ,   7   ,   15  ,   23  ,   59  ,   9   ,   163 ,   105 ,   911 ,   905 ,   3715    ,   4917    ,0 };
            const unsigned long dim790KuoInit[]     =   {   1   ,   3   ,   1   ,   11  ,   21  ,   53  ,   127 ,   3   ,   507 ,   467 ,   547 ,   3003    ,   6939    ,0 };
            const unsigned long dim791KuoInit[]     =   {   1   ,   1   ,   1   ,   13  ,   15  ,   7   ,   1   ,   7   ,   101 ,   645 ,   1469    ,   843 ,   2843    ,0 };
            const unsigned long dim792KuoInit[]     =   {   1   ,   3   ,   3   ,   9   ,   11  ,   35  ,   95  ,   53  ,   291 ,   459 ,   669 ,   1163    ,   4893    ,0 };
            const unsigned long dim793KuoInit[]     =   {   1   ,   1   ,   3   ,   3   ,   25  ,   29  ,   29  ,   223 ,   21  ,   553 ,   177 ,   3347    ,   3575    ,0 };
            const unsigned long dim794KuoInit[]     =   {   1   ,   1   ,   5   ,   1   ,   23  ,   17  ,   21  ,   57  ,   383 ,   427 ,   397 ,   1947    ,   2627    ,0 };
            const unsigned long dim795KuoInit[]     =   {   1   ,   3   ,   7   ,   1   ,   21  ,   53  ,   79  ,   59  ,   133 ,   461 ,   1703    ,   3071    ,   4679    ,0 };
            const unsigned long dim796KuoInit[]     =   {   1   ,   3   ,   5   ,   1   ,   27  ,   31  ,   83  ,   159 ,   505 ,   121 ,   1969    ,   1073    ,   2117    ,0 };
            const unsigned long dim797KuoInit[]     =   {   1   ,   1   ,   3   ,   1   ,   19  ,   25  ,   43  ,   115 ,   191 ,   845 ,   1923    ,   903 ,   2319    ,0 };
            const unsigned long dim798KuoInit[]     =   {   1   ,   1   ,   3   ,   9   ,   27  ,   37  ,   31  ,   111 ,   437 ,   925 ,   501 ,   883 ,   3485    ,0 };
            const unsigned long dim799KuoInit[]     =   {   1   ,   3   ,   5   ,   5   ,   15  ,   25  ,   51  ,   131 ,   149 ,   811 ,   1827    ,   1519    ,   1435    ,0 };
            const unsigned long dim800KuoInit[]     =   {   1   ,   1   ,   7   ,   13  ,   3   ,   39  ,   103 ,   107 ,   473 ,   553 ,   1763    ,   399 ,   2137    ,0 };
            const unsigned long dim801KuoInit[]     =   {   1   ,   3   ,   7   ,   13  ,   27  ,   25  ,   49  ,   29  ,   497 ,   863 ,   1715    ,   99  ,   1521    ,0 };
            const unsigned long dim802KuoInit[]     =   {   1   ,   1   ,   1   ,   9   ,   27  ,   37  ,   67  ,   59  ,   11  ,   369 ,   893 ,   2249    ,   5303    ,0 };
            const unsigned long dim803KuoInit[]     =   {   1   ,   3   ,   3   ,   13  ,   19  ,   9   ,   93  ,   29  ,   247 ,   35  ,   833 ,   1037    ,   7137    ,0 };
            const unsigned long dim804KuoInit[]     =   {   1   ,   3   ,   7   ,   3   ,   25  ,   17  ,   63  ,   239 ,   395 ,   367 ,   1935    ,   2095    ,   5191    ,0 };
            const unsigned long dim805KuoInit[]     =   {   1   ,   3   ,   5   ,   15  ,   23  ,   51  ,   13  ,   101 ,   477 ,   511 ,   1951    ,   2505    ,   4155    ,0 };
            const unsigned long dim806KuoInit[]     =   {   1   ,   3   ,   3   ,   15  ,   3   ,   55  ,   15  ,   173 ,   21  ,   699 ,   1617    ,   1573    ,   2645    ,0 };
            const unsigned long dim807KuoInit[]     =   {   1   ,   1   ,   3   ,   11  ,   7   ,   55  ,   69  ,   165 ,   255 ,   397 ,   2033    ,   9   ,   843 ,0 };
            const unsigned long dim808KuoInit[]     =   {   1   ,   3   ,   1   ,   13  ,   5   ,   3   ,   37  ,   99  ,   165 ,   971 ,   555 ,   2379    ,   7879    ,0 };
            const unsigned long dim809KuoInit[]     =   {   1   ,   1   ,   5   ,   3   ,   19  ,   37  ,   23  ,   197 ,   145 ,   315 ,   981 ,   161 ,   323 ,0 };
            const unsigned long dim810KuoInit[]     =   {   1   ,   1   ,   7   ,   7   ,   17  ,   37  ,   27  ,   189 ,   371 ,   145 ,   1097    ,   2801    ,   5469    ,0 };
            const unsigned long dim811KuoInit[]     =   {   1   ,   1   ,   3   ,   9   ,   23  ,   15  ,   39  ,   7   ,   25  ,   221 ,   757 ,   1393    ,   4629    ,0 };
            const unsigned long dim812KuoInit[]     =   {   1   ,   1   ,   1   ,   1   ,   13  ,   31  ,   29  ,   197 ,   237 ,   173 ,   337 ,   743 ,   1459    ,0 };
            const unsigned long dim813KuoInit[]     =   {   1   ,   1   ,   5   ,   15  ,   17  ,   7   ,   127 ,   29  ,   307 ,   965 ,   1567    ,   2771    ,   6053    ,0 };
            const unsigned long dim814KuoInit[]     =   {   1   ,   3   ,   1   ,   1   ,   13  ,   57  ,   41  ,   59  ,   77  ,   67  ,   1985    ,   3443    ,   7495    ,0 };
            const unsigned long dim815KuoInit[]     =   {   1   ,   3   ,   7   ,   15  ,   17  ,   21  ,   57  ,   143 ,   179 ,   99  ,   217 ,   1763    ,   873 ,0 };
            const unsigned long dim816KuoInit[]     =   {   1   ,   1   ,   7   ,   1   ,   21  ,   17  ,   123 ,   39  ,   95  ,   775 ,   1865    ,   341 ,   6167    ,0 };
            const unsigned long dim817KuoInit[]     =   {   1   ,   3   ,   7   ,   3   ,   31  ,   33  ,   33  ,   163 ,   281 ,   765 ,   1809    ,   1469    ,   7011    ,0 };
            const unsigned long dim818KuoInit[]     =   {   1   ,   3   ,   7   ,   7   ,   25  ,   35  ,   45  ,   109 ,   221 ,   25  ,   757 ,   597 ,   3057    ,0 };
            const unsigned long dim819KuoInit[]     =   {   1   ,   3   ,   5   ,   3   ,   27  ,   25  ,   97  ,   21  ,   351 ,   841 ,   1339    ,   565 ,   7181    ,0 };
            const unsigned long dim820KuoInit[]     =   {   1   ,   1   ,   1   ,   15  ,   23  ,   61  ,   97  ,   57  ,   503 ,   379 ,   25  ,   1197    ,   6957    ,0 };
            const unsigned long dim821KuoInit[]     =   {   1   ,   3   ,   3   ,   3   ,   13  ,   49  ,   127 ,   89  ,   329 ,   503 ,   1957    ,   3569    ,   7013    ,0 };
            const unsigned long dim822KuoInit[]     =   {   1   ,   1   ,   7   ,   7   ,   7   ,   15  ,   43  ,   61  ,   227 ,   859 ,   279 ,   3309    ,   1941    ,0 };
            const unsigned long dim823KuoInit[]     =   {   1   ,   3   ,   1   ,   9   ,   19  ,   1   ,   111 ,   45  ,   315 ,   339 ,   1049    ,   121 ,   2711    ,0 };
            const unsigned long dim824KuoInit[]     =   {   1   ,   1   ,   3   ,   11  ,   7   ,   33  ,   121 ,   253 ,   267 ,   845 ,   1677    ,   569 ,   3743    ,0 };
            const unsigned long dim825KuoInit[]     =   {   1   ,   3   ,   7   ,   9   ,   9   ,   53  ,   113 ,   11  ,   177 ,   165 ,   1707    ,   2027    ,   2099    ,0 };
            const unsigned long dim826KuoInit[]     =   {   1   ,   1   ,   5   ,   7   ,   19  ,   41  ,   33  ,   129 ,   389 ,   249 ,   1945    ,   1393    ,   5941    ,0 };
            const unsigned long dim827KuoInit[]     =   {   1   ,   1   ,   1   ,   1   ,   27  ,   37  ,   63  ,   171 ,   301 ,   179 ,   223 ,   3775    ,   1189    ,0 };
            const unsigned long dim828KuoInit[]     =   {   1   ,   3   ,   3   ,   5   ,   5   ,   33  ,   101 ,   37  ,   505 ,   561 ,   1681    ,   567 ,   2197    ,0 };
            const unsigned long dim829KuoInit[]     =   {   1   ,   3   ,   7   ,   13  ,   25  ,   23  ,   29  ,   27  ,   325 ,   757 ,   5   ,   487 ,   3241    ,0 };
            const unsigned long dim830KuoInit[]     =   {   1   ,   3   ,   3   ,   7   ,   27  ,   13  ,   85  ,   35  ,   171 ,   565 ,   969 ,   773 ,   7173    ,0 };
            const unsigned long dim831KuoInit[]     =   {   1   ,   3   ,   5   ,   5   ,   11  ,   55  ,   33  ,   243 ,   401 ,   939 ,   73  ,   775 ,   1159    ,0 };
            const unsigned long dim832KuoInit[]     =   {   1   ,   1   ,   3   ,   13  ,   13  ,   63  ,   9   ,   161 ,   343 ,   547 ,   1659    ,   1005    ,   8041    ,0 };
            const unsigned long dim833KuoInit[]     =   {   1   ,   3   ,   5   ,   7   ,   15  ,   27  ,   27  ,   51  ,   475 ,   957 ,   445 ,   3745    ,   5113    ,0 };
            const unsigned long dim834KuoInit[]     =   {   1   ,   1   ,   3   ,   7   ,   15  ,   49  ,   63  ,   181 ,   15  ,   1   ,   1915    ,   633 ,   1039    ,0 };
            const unsigned long dim835KuoInit[]     =   {   1   ,   3   ,   7   ,   1   ,   1   ,   11  ,   15  ,   221 ,   375 ,   59  ,   1471    ,   113 ,   5661    ,0 };
            const unsigned long dim836KuoInit[]     =   {   1   ,   3   ,   5   ,   5   ,   11  ,   41  ,   115 ,   87  ,   263 ,   173 ,   1249    ,   2235    ,   6863    ,0 };
            const unsigned long dim837KuoInit[]     =   {   1   ,   1   ,   5   ,   11  ,   19  ,   1   ,   61  ,   25  ,   239 ,   5   ,   1869    ,   2269    ,   3911    ,0 };
            const unsigned long dim838KuoInit[]     =   {   1   ,   3   ,   5   ,   7   ,   23  ,   59  ,   99  ,   47  ,   439 ,   105 ,   455 ,   547 ,   2447    ,0 };
            const unsigned long dim839KuoInit[]     =   {   1   ,   1   ,   7   ,   15  ,   25  ,   15  ,   117 ,   185 ,   83  ,   243 ,   1099    ,   2613    ,   7721    ,0 };
            const unsigned long dim840KuoInit[]     =   {   1   ,   3   ,   1   ,   9   ,   5   ,   49  ,   27  ,   237 ,   239 ,   177 ,   1507    ,   2059    ,   1731    ,0 };
            const unsigned long dim841KuoInit[]     =   {   1   ,   3   ,   3   ,   11  ,   29  ,   47  ,   119 ,   243 ,   475 ,   443 ,   111 ,   787 ,   2597    ,0 };
            const unsigned long dim842KuoInit[]     =   {   1   ,   1   ,   7   ,   1   ,   25  ,   39  ,   95  ,   107 ,   391 ,   149 ,   397 ,   3755    ,   2735    ,0 };
            const unsigned long dim843KuoInit[]     =   {   1   ,   3   ,   7   ,   11  ,   13  ,   49  ,   37  ,   135 ,   99  ,   473 ,   1961    ,   1233    ,   7359    ,0 };
            const unsigned long dim844KuoInit[]     =   {   1   ,   3   ,   5   ,   11  ,   29  ,   59  ,   123 ,   53  ,   371 ,   573 ,   1099    ,   3035    ,   3217    ,0 };
            const unsigned long dim845KuoInit[]     =   {   1   ,   1   ,   1   ,   11  ,   25  ,   47  ,   105 ,   147 ,   317 ,   69  ,   1201    ,   3781    ,   2457    ,0 };
            const unsigned long dim846KuoInit[]     =   {   1   ,   1   ,   5   ,   9   ,   3   ,   31  ,   25  ,   127 ,   153 ,   303 ,   1367    ,   3583    ,   5229    ,0 };
            const unsigned long dim847KuoInit[]     =   {   1   ,   1   ,   3   ,   13  ,   25  ,   33  ,   11  ,   101 ,   85  ,   741 ,   1191    ,   1769    ,   4119    ,0 };
            const unsigned long dim848KuoInit[]     =   {   1   ,   1   ,   3   ,   13  ,   19  ,   59  ,   81  ,   71  ,   255 ,   277 ,   991 ,   2183    ,   3257    ,0 };
            const unsigned long dim849KuoInit[]     =   {   1   ,   3   ,   1   ,   3   ,   23  ,   27  ,   89  ,   237 ,   249 ,   837 ,   759 ,   43  ,   5571    ,0 };
            const unsigned long dim850KuoInit[]     =   {   1   ,   1   ,   1   ,   9   ,   15  ,   35  ,   45  ,   213 ,   423 ,   159 ,   353 ,   3123    ,   6399    ,0 };
            const unsigned long dim851KuoInit[]     =   {   1   ,   1   ,   5   ,   9   ,   15  ,   45  ,   83  ,   65  ,   115 ,   513 ,   361 ,   1877    ,   4307    ,0 };
            const unsigned long dim852KuoInit[]     =   {   1   ,   1   ,   7   ,   11  ,   17  ,   61  ,   15  ,   143 ,   491 ,   253 ,   443 ,   3521    ,   7611    ,0 };
            const unsigned long dim853KuoInit[]     =   {   1   ,   1   ,   3   ,   11  ,   11  ,   49  ,   117 ,   117 ,   279 ,   715 ,   1365    ,   689 ,   4339    ,0 };
            const unsigned long dim854KuoInit[]     =   {   1   ,   1   ,   3   ,   13  ,   3   ,   53  ,   123 ,   89  ,   91  ,   471 ,   1723    ,   1811    ,   1285    ,0 };
            const unsigned long dim855KuoInit[]     =   {   1   ,   3   ,   7   ,   13  ,   7   ,   5   ,   99  ,   235 ,   465 ,   915 ,   1813    ,   3137    ,   7645    ,0 };
            const unsigned long dim856KuoInit[]     =   {   1   ,   3   ,   3   ,   13  ,   5   ,   45  ,   37  ,   99  ,   459 ,   541 ,   225 ,   2417    ,   6385    ,0 };
            const unsigned long dim857KuoInit[]     =   {   1   ,   1   ,   1   ,   15  ,   19  ,   39  ,   109 ,   21  ,   117 ,   589 ,   1261    ,   3695    ,   4251    ,0 };
            const unsigned long dim858KuoInit[]     =   {   1   ,   1   ,   7   ,   11  ,   23  ,   43  ,   35  ,   43  ,   135 ,   509 ,   1331    ,   1567    ,   7085    ,0 };
            const unsigned long dim859KuoInit[]     =   {   1   ,   1   ,   5   ,   5   ,   23  ,   3   ,   61  ,   151 ,   171 ,   295 ,   1377    ,   2165    ,   2125    ,0 };
            const unsigned long dim860KuoInit[]     =   {   1   ,   3   ,   3   ,   5   ,   3   ,   57  ,   17  ,   73  ,   71  ,   55  ,   1373    ,   2067    ,   3249    ,0 };
            const unsigned long dim861KuoInit[]     =   {   1   ,   1   ,   5   ,   9   ,   25  ,   39  ,   65  ,   75  ,   141 ,   849 ,   237 ,   3719    ,   6705    ,0 };
            const unsigned long dim862KuoInit[]     =   {   1   ,   1   ,   7   ,   1   ,   17  ,   17  ,   23  ,   91  ,   333 ,   749 ,   657 ,   3267    ,   3343    ,0 };
            const unsigned long dim863KuoInit[]     =   {   1   ,   3   ,   1   ,   5   ,   5   ,   27  ,   15  ,   107 ,   345 ,   405 ,   111 ,   345 ,   5505    ,0 };
            const unsigned long dim864KuoInit[]     =   {   1   ,   3   ,   3   ,   13  ,   1   ,   9   ,   19  ,   63  ,   427 ,   995 ,   1257    ,   923 ,   267 ,0 };
            const unsigned long dim865KuoInit[]     =   {   1   ,   1   ,   1   ,   7   ,   21  ,   19  ,   65  ,   249 ,   401 ,   729 ,   1403    ,   1579    ,   3443    ,0 };
            const unsigned long dim866KuoInit[]     =   {   1   ,   3   ,   5   ,   1   ,   3   ,   43  ,   81  ,   11  ,   473 ,   1011    ,   1755    ,   117 ,   2881    ,0 };
            const unsigned long dim867KuoInit[]     =   {   1   ,   1   ,   7   ,   13  ,   3   ,   63  ,   95  ,   11  ,   23  ,   895 ,   1591    ,   2623    ,   6887    ,0 };
            const unsigned long dim868KuoInit[]     =   {   1   ,   3   ,   1   ,   3   ,   17  ,   11  ,   39  ,   165 ,   367 ,   363 ,   1207    ,   2555    ,   5437    ,0 };
            const unsigned long dim869KuoInit[]     =   {   1   ,   1   ,   1   ,   7   ,   17  ,   9   ,   19  ,   175 ,   7   ,   535 ,   1779    ,   3341    ,   4825    ,0 };
            const unsigned long dim870KuoInit[]     =   {   1   ,   1   ,   7   ,   3   ,   27  ,   7   ,   19  ,   217 ,   479 ,   143 ,   1189    ,   2809    ,   339 ,0 };
            const unsigned long dim871KuoInit[]     =   {   1   ,   3   ,   7   ,   9   ,   25  ,   41  ,   9   ,   39  ,   391 ,   331 ,   1427    ,   2321    ,   1673    ,0 };
            const unsigned long dim872KuoInit[]     =   {   1   ,   3   ,   1   ,   13  ,   7   ,   17  ,   11  ,   27  ,   125 ,   799 ,   1325    ,   517 ,   1043    ,0 };
            const unsigned long dim873KuoInit[]     =   {   1   ,   1   ,   3   ,   15  ,   29  ,   11  ,   73  ,   11  ,   289 ,   1   ,   839 ,   3311    ,   7535    ,0 };
            const unsigned long dim874KuoInit[]     =   {   1   ,   1   ,   1   ,   11  ,   11  ,   59  ,   25  ,   47  ,   153 ,   297 ,   679 ,   1441    ,   7105    ,0 };
            const unsigned long dim875KuoInit[]     =   {   1   ,   3   ,   5   ,   7   ,   25  ,   29  ,   13  ,   227 ,   379 ,   247 ,   449 ,   679 ,   6581    ,0 };
            const unsigned long dim876KuoInit[]     =   {   1   ,   1   ,   1   ,   5   ,   9   ,   59  ,   81  ,   195 ,   289 ,   705 ,   183 ,   1   ,   3703    ,0 };
            const unsigned long dim877KuoInit[]     =   {   1   ,   3   ,   3   ,   9   ,   31  ,   1   ,   71  ,   155 ,   307 ,   467 ,   1699    ,   269 ,   6077    ,0 };
            const unsigned long dim878KuoInit[]     =   {   1   ,   3   ,   1   ,   11  ,   31  ,   21  ,   59  ,   225 ,   385 ,   623 ,   117 ,   1435    ,   3195    ,0 };
            const unsigned long dim879KuoInit[]     =   {   1   ,   1   ,   7   ,   9   ,   3   ,   43  ,   97  ,   15  ,   145 ,   279 ,   1785    ,   545 ,   3977    ,0 };
            const unsigned long dim880KuoInit[]     =   {   1   ,   1   ,   5   ,   9   ,   25  ,   27  ,   113 ,   193 ,   31  ,   255 ,   691 ,   3315    ,   7005    ,0 };
            const unsigned long dim881KuoInit[]     =   {   1   ,   3   ,   7   ,   5   ,   9   ,   19  ,   69  ,   87  ,   379 ,   931 ,   1567    ,   1391    ,   4765    ,0 };
            const unsigned long dim882KuoInit[]     =   {   1   ,   1   ,   1   ,   15  ,   15  ,   23  ,   127 ,   69  ,   113 ,   801 ,   1923    ,   2847    ,   1073    ,0 };
            const unsigned long dim883KuoInit[]     =   {   1   ,   1   ,   1   ,   1   ,   21  ,   51  ,   65  ,   133 ,   465 ,   739 ,   2029    ,   2243    ,   4023    ,0 };
            const unsigned long dim884KuoInit[]     =   {   1   ,   1   ,   5   ,   11  ,   9   ,   41  ,   71  ,   217 ,   317 ,   855 ,   275 ,   1965    ,   3151    ,0 };
            const unsigned long dim885KuoInit[]     =   {   1   ,   1   ,   7   ,   5   ,   19  ,   3   ,   81  ,   127 ,   423 ,   415 ,   565 ,   1941    ,   5079    ,0 };
            const unsigned long dim886KuoInit[]     =   {   1   ,   1   ,   3   ,   1   ,   19  ,   19  ,   117 ,   169 ,   353 ,   159 ,   1235    ,   845 ,   2381    ,0 };
            const unsigned long dim887KuoInit[]     =   {   1   ,   3   ,   1   ,   15  ,   21  ,   27  ,   77  ,   5   ,   113 ,   201 ,   311 ,   2263    ,   7855    ,0 };
            const unsigned long dim888KuoInit[]     =   {   1   ,   3   ,   5   ,   9   ,   5   ,   21  ,   41  ,   119 ,   447 ,   983 ,   1703    ,   2707    ,   2789    ,0 };
            const unsigned long dim889KuoInit[]     =   {   1   ,   1   ,   5   ,   3   ,   5   ,   19  ,   9   ,   131 ,   269 ,   999 ,   1259    ,   3843    ,   7753    ,0 };
            const unsigned long dim890KuoInit[]     =   {   1   ,   3   ,   7   ,   11  ,   25  ,   27  ,   91  ,   187 ,   177 ,   921 ,   163 ,   1141    ,   4903    ,0 };
            const unsigned long dim891KuoInit[]     =   {   1   ,   3   ,   1   ,   7   ,   15  ,   21  ,   67  ,   155 ,   335 ,   149 ,   2011    ,   2483    ,   7015    ,0 };
            const unsigned long dim892KuoInit[]     =   {   1   ,   1   ,   1   ,   7   ,   3   ,   53  ,   29  ,   119 ,   21  ,   933 ,   595 ,   3141    ,   7925    ,0 };
            const unsigned long dim893KuoInit[]     =   {   1   ,   1   ,   1   ,   9   ,   21  ,   33  ,   55  ,   87  ,   451 ,   297 ,   1615    ,   697 ,   2299    ,0 };
            const unsigned long dim894KuoInit[]     =   {   1   ,   3   ,   1   ,   9   ,   9   ,   5   ,   3   ,   113 ,   197 ,   137 ,   1105    ,   1949    ,   4651    ,0 };
            const unsigned long dim895KuoInit[]     =   {   1   ,   1   ,   7   ,   3   ,   7   ,   27  ,   109 ,   99  ,   121 ,   463 ,   621 ,   887 ,   5847    ,0 };
            const unsigned long dim896KuoInit[]     =   {   1   ,   3   ,   7   ,   7   ,   31  ,   61  ,   21  ,   69  ,   443 ,   81  ,   151 ,   3601    ,   699 ,0 };
            const unsigned long dim897KuoInit[]     =   {   1   ,   1   ,   7   ,   11  ,   7   ,   63  ,   97  ,   241 ,   87  ,   961 ,   1279    ,   1715    ,   3183    ,0 };
            const unsigned long dim898KuoInit[]     =   {   1   ,   1   ,   3   ,   5   ,   7   ,   41  ,   5   ,   123 ,   189 ,   283 ,   1775    ,   49  ,   6107    ,0 };
            const unsigned long dim899KuoInit[]     =   {   1   ,   1   ,   7   ,   15  ,   7   ,   19  ,   71  ,   117 ,   195 ,   67  ,   1583    ,   3371    ,   777 ,0 };
            const unsigned long dim900KuoInit[]     =   {   1   ,   1   ,   1   ,   15  ,   19  ,   43  ,   81  ,   103 ,   173 ,   209 ,   447 ,   1735    ,   5023    ,0 };
            const unsigned long dim901KuoInit[]     =   {   1   ,   1   ,   5   ,   15  ,   15  ,   9   ,   99  ,   185 ,   451 ,   605 ,   1429    ,   1257    ,   6825    ,0 };
            const unsigned long dim902KuoInit[]     =   {   1   ,   3   ,   1   ,   11  ,   11  ,   21  ,   79  ,   13  ,   23  ,   607 ,   1601    ,   3603    ,   147 ,0 };
            const unsigned long dim903KuoInit[]     =   {   1   ,   1   ,   3   ,   1   ,   27  ,   51  ,   83  ,   191 ,   59  ,   281 ,   1853    ,   1441    ,   5359    ,0 };
            const unsigned long dim904KuoInit[]     =   {   1   ,   1   ,   1   ,   11  ,   31  ,   17  ,   117 ,   239 ,   385 ,   465 ,   1787    ,   2645    ,   6047    ,0 };
            const unsigned long dim905KuoInit[]     =   {   1   ,   1   ,   7   ,   5   ,   17  ,   59  ,   111 ,   143 ,   419 ,   163 ,   763 ,   2327    ,   1741    ,0 };
            const unsigned long dim906KuoInit[]     =   {   1   ,   1   ,   1   ,   3   ,   3   ,   43  ,   3   ,   203 ,   1   ,   373 ,   1829    ,   3493    ,   2347    ,0 };
            const unsigned long dim907KuoInit[]     =   {   1   ,   3   ,   7   ,   13  ,   21  ,   17  ,   89  ,   147 ,   51  ,   665 ,   397 ,   2475    ,   3393    ,0 };
            const unsigned long dim908KuoInit[]     =   {   1   ,   3   ,   7   ,   15  ,   19  ,   33  ,   1   ,   211 ,   155 ,   355 ,   1137    ,   1633    ,   5025    ,0 };
            const unsigned long dim909KuoInit[]     =   {   1   ,   3   ,   3   ,   9   ,   7   ,   63  ,   127 ,   59  ,   205 ,   645 ,   2047    ,   1187    ,   2711    ,0 };
            const unsigned long dim910KuoInit[]     =   {   1   ,   3   ,   3   ,   5   ,   7   ,   59  ,   125 ,   35  ,   285 ,   811 ,   1489    ,   3269    ,   2737    ,0 };
            const unsigned long dim911KuoInit[]     =   {   1   ,   1   ,   7   ,   3   ,   15  ,   39  ,   75  ,   229 ,   213 ,   23  ,   1825    ,   2269    ,   6739    ,0 };
            const unsigned long dim912KuoInit[]     =   {   1   ,   1   ,   3   ,   7   ,   1   ,   5   ,   73  ,   173 ,   305 ,   681 ,   1225    ,   2475    ,   6981    ,0 };
            const unsigned long dim913KuoInit[]     =   {   1   ,   1   ,   5   ,   3   ,   21  ,   3   ,   27  ,   241 ,   475 ,   543 ,   1065    ,   2731    ,   1739    ,0 };
            const unsigned long dim914KuoInit[]     =   {   1   ,   3   ,   7   ,   13  ,   19  ,   35  ,   57  ,   19  ,   307 ,   113 ,   451 ,   1897    ,   2135    ,0 };
            const unsigned long dim915KuoInit[]     =   {   1   ,   1   ,   3   ,   11  ,   31  ,   13  ,   63  ,   15  ,   477 ,   389 ,   829 ,   2167    ,   3317    ,0 };
            const unsigned long dim916KuoInit[]     =   {   1   ,   1   ,   7   ,   11  ,   9   ,   39  ,   1   ,   41  ,   99  ,   295 ,   1009    ,   1411    ,   3423    ,0 };
            const unsigned long dim917KuoInit[]     =   {   1   ,   1   ,   5   ,   9   ,   15  ,   17  ,   93  ,   171 ,   361 ,   51  ,   455 ,   2413    ,   7295    ,0 };
            const unsigned long dim918KuoInit[]     =   {   1   ,   1   ,   5   ,   13  ,   7   ,   27  ,   127 ,   153 ,   163 ,   635 ,   1065    ,   2243    ,   3813    ,0 };
            const unsigned long dim919KuoInit[]     =   {   1   ,   1   ,   7   ,   9   ,   15  ,   45  ,   7   ,   241 ,   365 ,   447 ,   1435    ,   3473    ,   715 ,0 };
            const unsigned long dim920KuoInit[]     =   {   1   ,   1   ,   5   ,   9   ,   9   ,   11  ,   21  ,   21  ,   255 ,   717 ,   1579    ,   217 ,   6921    ,0 };
            const unsigned long dim921KuoInit[]     =   {   1   ,   1   ,   5   ,   15  ,   25  ,   21  ,   31  ,   47  ,   455 ,   631 ,   971 ,   3105    ,   2219    ,0 };
            const unsigned long dim922KuoInit[]     =   {   1   ,   3   ,   1   ,   7   ,   9   ,   1   ,   75  ,   85  ,   351 ,   555 ,   1379    ,   1361    ,   6163    ,0 };
            const unsigned long dim923KuoInit[]     =   {   1   ,   3   ,   5   ,   7   ,   9   ,   63  ,   19  ,   99  ,   335 ,   173 ,   1385    ,   3011    ,   5787    ,0 };
            const unsigned long dim924KuoInit[]     =   {   1   ,   3   ,   5   ,   1   ,   23  ,   35  ,   79  ,   163 ,   97  ,   213 ,   1281    ,   1739    ,   477 ,0 };
            const unsigned long dim925KuoInit[]     =   {   1   ,   1   ,   5   ,   5   ,   9   ,   13  ,   41  ,   167 ,   349 ,   959 ,   925 ,   1383    ,   883 ,0 };
            const unsigned long dim926KuoInit[]     =   {   1   ,   3   ,   7   ,   7   ,   17  ,   23  ,   69  ,   241 ,   447 ,   31  ,   767 ,   2981    ,   253 ,0 };
            const unsigned long dim927KuoInit[]     =   {   1   ,   3   ,   7   ,   15  ,   31  ,   15  ,   3   ,   107 ,   225 ,   207 ,   429 ,   959 ,   2627    ,0 };
            const unsigned long dim928KuoInit[]     =   {   1   ,   3   ,   3   ,   11  ,   29  ,   25  ,   109 ,   41  ,   261 ,   585 ,   1049    ,   3013    ,   4361    ,0 };
            const unsigned long dim929KuoInit[]     =   {   1   ,   1   ,   7   ,   11  ,   29  ,   63  ,   117 ,   147 ,   149 ,   427 ,   1371    ,   2171    ,   7797    ,0 };
            const unsigned long dim930KuoInit[]     =   {   1   ,   3   ,   5   ,   3   ,   13  ,   33  ,   103 ,   33  ,   227 ,   785 ,   1277    ,   27  ,   6713    ,0 };
            const unsigned long dim931KuoInit[]     =   {   1   ,   3   ,   1   ,   3   ,   19  ,   3   ,   121 ,   123 ,   313 ,   283 ,   1231    ,   3331    ,   7271    ,0 };
            const unsigned long dim932KuoInit[]     =   {   1   ,   1   ,   7   ,   3   ,   21  ,   57  ,   53  ,   9   ,   275 ,   661 ,   1665    ,   1557    ,   3813    ,0 };
            const unsigned long dim933KuoInit[]     =   {   1   ,   3   ,   5   ,   1   ,   19  ,   63  ,   69  ,   149 ,   373 ,   667 ,   975 ,   3483    ,   7605    ,0 };
            const unsigned long dim934KuoInit[]     =   {   1   ,   3   ,   5   ,   3   ,   3   ,   53  ,   37  ,   247 ,   293 ,   491 ,   657 ,   3123    ,   1165    ,0 };
            const unsigned long dim935KuoInit[]     =   {   1   ,   1   ,   1   ,   13  ,   27  ,   41  ,   23  ,   185 ,   405 ,   215 ,   455 ,   433 ,   7775    ,0 };
            const unsigned long dim936KuoInit[]     =   {   1   ,   1   ,   3   ,   13  ,   23  ,   61  ,   63  ,   61  ,   195 ,   83  ,   33  ,   1903    ,   269 ,0 };
            const unsigned long dim937KuoInit[]     =   {   1   ,   1   ,   3   ,   7   ,   5   ,   7   ,   1   ,   161 ,   295 ,   539 ,   1371    ,   3911    ,   4095    ,0 };
            const unsigned long dim938KuoInit[]     =   {   1   ,   1   ,   7   ,   9   ,   7   ,   7   ,   63  ,   115 ,   323 ,   95  ,   1965    ,   137 ,   7147    ,0 };
            const unsigned long dim939KuoInit[]     =   {   1   ,   3   ,   7   ,   11  ,   19  ,   55  ,   107 ,   71  ,   169 ,   937 ,   939 ,   113 ,   6187    ,0 };
            const unsigned long dim940KuoInit[]     =   {   1   ,   1   ,   3   ,   7   ,   13  ,   17  ,   55  ,   5   ,   167 ,   337 ,   2043    ,   1485    ,   5069    ,0 };
            const unsigned long dim941KuoInit[]     =   {   1   ,   1   ,   7   ,   9   ,   21  ,   3   ,   107 ,   125 ,   253 ,   363 ,   49  ,   469 ,   2873    ,0 };
            const unsigned long dim942KuoInit[]     =   {   1   ,   1   ,   7   ,   9   ,   9   ,   31  ,   91  ,   65  ,   315 ,   973 ,   653 ,   1023    ,   6479    ,0 };
            const unsigned long dim943KuoInit[]     =   {   1   ,   3   ,   5   ,   11  ,   1   ,   5   ,   19  ,   109 ,   143 ,   335 ,   1035    ,   3893    ,   4573    ,0 };
            const unsigned long dim944KuoInit[]     =   {   1   ,   3   ,   3   ,   13  ,   27  ,   53  ,   37  ,   177 ,   407 ,   315 ,   1943    ,   3111    ,   1385    ,0 };
            const unsigned long dim945KuoInit[]     =   {   1   ,   3   ,   5   ,   15  ,   27  ,   47  ,   57  ,   105 ,   131 ,   31  ,   551 ,   3733    ,   5157    ,0 };
            const unsigned long dim946KuoInit[]     =   {   1   ,   3   ,   5   ,   9   ,   15  ,   31  ,   41  ,   11  ,   93  ,   895 ,   587 ,   421 ,   885 ,0 };
            const unsigned long dim947KuoInit[]     =   {   1   ,   3   ,   7   ,   1   ,   5   ,   35  ,   107 ,   189 ,   287 ,   457 ,   645 ,   519 ,   7459    ,0 };
            const unsigned long dim948KuoInit[]     =   {   1   ,   1   ,   7   ,   5   ,   23  ,   7   ,   79  ,   45  ,   73  ,   711 ,   575 ,   3571    ,   1453    ,0 };
            const unsigned long dim949KuoInit[]     =   {   1   ,   1   ,   3   ,   9   ,   9   ,   7   ,   89  ,   197 ,   215 ,   111 ,   1099    ,   3915    ,   3709    ,0 };
            const unsigned long dim950KuoInit[]     =   {   1   ,   3   ,   1   ,   7   ,   25  ,   1   ,   83  ,   55  ,   449 ,   227 ,   1563    ,   1103    ,   4129    ,0 };
            const unsigned long dim951KuoInit[]     =   {   1   ,   1   ,   1   ,   9   ,   23  ,   25  ,   103 ,   183 ,   379 ,   541 ,   673 ,   2711    ,   7409    ,0 };
            const unsigned long dim952KuoInit[]     =   {   1   ,   3   ,   1   ,   11  ,   27  ,   21  ,   83  ,   61  ,   65  ,   867 ,   1755    ,   3815    ,   7235    ,0 };
            const unsigned long dim953KuoInit[]     =   {   1   ,   1   ,   3   ,   15  ,   15  ,   55  ,   13  ,   185 ,   145 ,   955 ,   785 ,   189 ,   6653    ,0 };
            const unsigned long dim954KuoInit[]     =   {   1   ,   3   ,   5   ,   15  ,   5   ,   19  ,   83  ,   157 ,   429 ,   57  ,   1293    ,   1551    ,   581 ,0 };
            const unsigned long dim955KuoInit[]     =   {   1   ,   3   ,   3   ,   11  ,   25  ,   19  ,   117 ,   73  ,   503 ,   1015    ,   203 ,   2689    ,   6697    ,0 };
            const unsigned long dim956KuoInit[]     =   {   1   ,   1   ,   7   ,   1   ,   27  ,   31  ,   111 ,   193 ,   403 ,   267 ,   233 ,   3031    ,   5451    ,0 };
            const unsigned long dim957KuoInit[]     =   {   1   ,   3   ,   7   ,   5   ,   9   ,   7   ,   73  ,   221 ,   27  ,   483 ,   185 ,   1805    ,   1813    ,0 };
            const unsigned long dim958KuoInit[]     =   {   1   ,   3   ,   3   ,   5   ,   19  ,   27  ,   103 ,   197 ,   245 ,   191 ,   301 ,   531 ,   3677    ,0 };
            const unsigned long dim959KuoInit[]     =   {   1   ,   1   ,   5   ,   1   ,   9   ,   49  ,   5   ,   55  ,   81  ,   907 ,   313 ,   3379    ,   2573    ,0 };
            const unsigned long dim960KuoInit[]     =   {   1   ,   3   ,   5   ,   3   ,   13  ,   11  ,   75  ,   145 ,   143 ,   991 ,   201 ,   3769    ,   5041    ,0 };
            const unsigned long dim961KuoInit[]     =   {   1   ,   3   ,   1   ,   5   ,   15  ,   45  ,   89  ,   17  ,   209 ,   91  ,   1635    ,   3025    ,   3885    ,0 };
            const unsigned long dim962KuoInit[]     =   {   1   ,   3   ,   3   ,   5   ,   15  ,   37  ,   79  ,   163 ,   455 ,   477 ,   69  ,   3617    ,   1263    ,0 };
            const unsigned long dim963KuoInit[]     =   {   1   ,   1   ,   3   ,   11  ,   15  ,   47  ,   77  ,   73  ,   199 ,   493 ,   207 ,   2109    ,   4991    ,0 };
            const unsigned long dim964KuoInit[]     =   {   1   ,   1   ,   5   ,   15  ,   29  ,   23  ,   35  ,   177 ,   483 ,   151 ,   1497    ,   631 ,   1647    ,0 };
            const unsigned long dim965KuoInit[]     =   {   1   ,   3   ,   7   ,   5   ,   29  ,   45  ,   127 ,   199 ,   491 ,   965 ,   371 ,   3549    ,   2757    ,0 };
            const unsigned long dim966KuoInit[]     =   {   1   ,   1   ,   7   ,   5   ,   21  ,   11  ,   27  ,   217 ,   407 ,   493 ,   1979    ,   359 ,   6187    ,0 };
            const unsigned long dim967KuoInit[]     =   {   1   ,   3   ,   5   ,   7   ,   31  ,   61  ,   113 ,   197 ,   109 ,   781 ,   1859    ,   991 ,   3617    ,0 };
            const unsigned long dim968KuoInit[]     =   {   1   ,   3   ,   1   ,   5   ,   17  ,   43  ,   75  ,   169 ,   413 ,   257 ,   1751    ,   4047    ,   1951    ,0 };
            const unsigned long dim969KuoInit[]     =   {   1   ,   3   ,   1   ,   13  ,   17  ,   49  ,   95  ,   223 ,   439 ,   587 ,   427 ,   89  ,   7723    ,0 };
            const unsigned long dim970KuoInit[]     =   {   1   ,   3   ,   5   ,   1   ,   7   ,   53  ,   1   ,   79  ,   111 ,   721 ,   169 ,   4087    ,   5255    ,0 };
            const unsigned long dim971KuoInit[]     =   {   1   ,   1   ,   7   ,   15  ,   31  ,   7   ,   111 ,   163 ,   491 ,   559 ,   259 ,   3057    ,   5669    ,0 };
            const unsigned long dim972KuoInit[]     =   {   1   ,   3   ,   3   ,   9   ,   25  ,   41  ,   45  ,   101 ,   145 ,   769 ,   1313    ,   3101    ,   405 ,0 };
            const unsigned long dim973KuoInit[]     =   {   1   ,   3   ,   3   ,   11  ,   3   ,   39  ,   67  ,   201 ,   171 ,   79  ,   1727    ,   2331    ,   8161    ,0 };
            const unsigned long dim974KuoInit[]     =   {   1   ,   3   ,   5   ,   5   ,   31  ,   29  ,   13  ,   111 ,   449 ,   579 ,   1133    ,   1321    ,   2447    ,0 };
            const unsigned long dim975KuoInit[]     =   {   1   ,   3   ,   3   ,   15  ,   21  ,   39  ,   57  ,   203 ,   39  ,   733 ,   1773    ,   2867    ,   6751    ,0 };
            const unsigned long dim976KuoInit[]     =   {   1   ,   3   ,   5   ,   3   ,   19  ,   33  ,   73  ,   219 ,   159 ,   901 ,   919 ,   1445    ,   1387    ,0 };
            const unsigned long dim977KuoInit[]     =   {   1   ,   3   ,   1   ,   7   ,   9   ,   33  ,   113 ,   221 ,   315 ,   503 ,   1795    ,   577 ,   1605    ,0 };
            const unsigned long dim978KuoInit[]     =   {   1   ,   1   ,   7   ,   11  ,   31  ,   47  ,   51  ,   53  ,   159 ,   731 ,   1101    ,   1895    ,   2061    ,0 };
            const unsigned long dim979KuoInit[]     =   {   1   ,   3   ,   5   ,   7   ,   29  ,   11  ,   107 ,   39  ,   89  ,   357 ,   13  ,   805 ,   5527    ,0 };
            const unsigned long dim980KuoInit[]     =   {   1   ,   3   ,   1   ,   9   ,   13  ,   15  ,   59  ,   95  ,   467 ,   875 ,   217 ,   3345    ,   1579    ,0 };
            const unsigned long dim981KuoInit[]     =   {   1   ,   3   ,   1   ,   15  ,   29  ,   33  ,   11  ,   121 ,   243 ,   429 ,   1017    ,   1511    ,   6289    ,0 };
            const unsigned long dim982KuoInit[]     =   {   1   ,   3   ,   7   ,   1   ,   27  ,   41  ,   125 ,   121 ,   265 ,   695 ,   155 ,   3933    ,   661 ,0 };
            const unsigned long dim983KuoInit[]     =   {   1   ,   1   ,   3   ,   7   ,   5   ,   3   ,   95  ,   213 ,   385 ,   183 ,   1693    ,   2685    ,   5647    ,0 };
            const unsigned long dim984KuoInit[]     =   {   1   ,   1   ,   5   ,   5   ,   23  ,   29  ,   21  ,   193 ,   327 ,   965 ,   49  ,   2157    ,   1539    ,0 };
            const unsigned long dim985KuoInit[]     =   {   1   ,   3   ,   1   ,   7   ,   15  ,   15  ,   103 ,   191 ,   249 ,   347 ,   283 ,   2459    ,   5957    ,0 };
            const unsigned long dim986KuoInit[]     =   {   1   ,   1   ,   7   ,   5   ,   21  ,   43  ,   29  ,   205 ,   125 ,   63  ,   1645    ,   3357    ,   3333    ,0 };
            const unsigned long dim987KuoInit[]     =   {   1   ,   3   ,   1   ,   5   ,   11  ,   21  ,   109 ,   213 ,   123 ,   587 ,   2047    ,   343 ,   6669    ,0 };
            const unsigned long dim988KuoInit[]     =   {   1   ,   1   ,   5   ,   5   ,   25  ,   63  ,   75  ,   237 ,   421 ,   883 ,   801 ,   2403    ,   4723    ,0 };
            const unsigned long dim989KuoInit[]     =   {   1   ,   3   ,   5   ,   11  ,   13  ,   63  ,   87  ,   107 ,   411 ,   3   ,   139 ,   3097    ,   6109    ,0 };
            const unsigned long dim990KuoInit[]     =   {   1   ,   3   ,   7   ,   9   ,   19  ,   31  ,   33  ,   23  ,   499 ,   105 ,   471 ,   2311    ,   439 ,0 };
            const unsigned long dim991KuoInit[]     =   {   1   ,   3   ,   3   ,   13  ,   13  ,   39  ,   19  ,   253 ,   53  ,   315 ,   1381    ,   3875    ,   5741    ,0 };
            const unsigned long dim992KuoInit[]     =   {   1   ,   1   ,   1   ,   7   ,   19  ,   9   ,   47  ,   169 ,   265 ,   1005    ,   159 ,   2957    ,   2987    ,0 };
            const unsigned long dim993KuoInit[]     =   {   1   ,   3   ,   3   ,   3   ,   1   ,   35  ,   107 ,   77  ,   47  ,   89  ,   1901    ,   1719    ,   2873    ,0 };
            const unsigned long dim994KuoInit[]     =   {   1   ,   1   ,   5   ,   7   ,   7   ,   51  ,   87  ,   91  ,   45  ,   991 ,   1369    ,   3759    ,   3511    ,0 };
            const unsigned long dim995KuoInit[]     =   {   1   ,   3   ,   7   ,   15  ,   17  ,   51  ,   93  ,   43  ,   71  ,   213 ,   1637    ,   2617    ,   1767    ,0 };
            const unsigned long dim996KuoInit[]     =   {   1   ,   1   ,   5   ,   15  ,   29  ,   63  ,   81  ,   15  ,   409 ,   779 ,   1419    ,   497 ,   463 ,0 };
            const unsigned long dim997KuoInit[]     =   {   1   ,   1   ,   7   ,   1   ,   1   ,   31  ,   21  ,   167 ,   445 ,   893 ,   1687    ,   603 ,   3211    ,0 };
            const unsigned long dim998KuoInit[]     =   {   1   ,   3   ,   7   ,   15  ,   19  ,   3   ,   41  ,   53  ,   175 ,   971 ,   1873    ,   43  ,   4795    ,0 };
            const unsigned long dim999KuoInit[]     =   {   1   ,   3   ,   5   ,   15  ,   21  ,   59  ,   77  ,   67  ,   495 ,   547 ,   411 ,   2373    ,   5187    ,0 };
            const unsigned long dim1000KuoInit[]        =   {   1   ,   1   ,   5   ,   1   ,   27  ,   45  ,   119 ,   201 ,   237 ,   499 ,   261 ,   3331    ,   3973    ,0 };
            const unsigned long dim1001KuoInit[]        =   {   1   ,   3   ,   5   ,   9   ,   17  ,   31  ,   33  ,   15  ,   237 ,   135 ,   1751    ,   297 ,   6331    ,0 };
            const unsigned long dim1002KuoInit[]        =   {   1   ,   3   ,   7   ,   9   ,   15  ,   1   ,   11  ,   135 ,   355 ,   617 ,   805 ,   1201    ,   457 ,0 };
            const unsigned long dim1003KuoInit[]        =   {   1   ,   1   ,   7   ,   5   ,   19  ,   55  ,   21  ,   121 ,   205 ,   999 ,   1719    ,   2815    ,   6753    ,0 };
            const unsigned long dim1004KuoInit[]        =   {   1   ,   1   ,   5   ,   3   ,   9   ,   5   ,   7   ,   231 ,   323 ,   953 ,   1811    ,   2847    ,   2059    ,0 };
            const unsigned long dim1005KuoInit[]        =   {   1   ,   3   ,   3   ,   9   ,   17  ,   15  ,   85  ,   3   ,   259 ,   989 ,   287 ,   3995    ,   2677    ,0 };
            const unsigned long dim1006KuoInit[]        =   {   1   ,   1   ,   3   ,   11  ,   13  ,   61  ,   69  ,   143 ,   167 ,   187 ,   481 ,   2359    ,   1913    ,0 };
            const unsigned long dim1007KuoInit[]        =   {   1   ,   1   ,   5   ,   13  ,   29  ,   3   ,   29  ,   247 ,   407 ,   733 ,   677 ,   99  ,   1625    ,0 };
            const unsigned long dim1008KuoInit[]        =   {   1   ,   1   ,   7   ,   13  ,   7   ,   17  ,   67  ,   115 ,   299 ,   765 ,   1885    ,   2769    ,   1619    ,0 };
            const unsigned long dim1009KuoInit[]        =   {   1   ,   1   ,   3   ,   3   ,   17  ,   7   ,   27  ,   61  ,   433 ,   323 ,   1091    ,   1309    ,   7239    ,0 };
            const unsigned long dim1010KuoInit[]        =   {   1   ,   1   ,   7   ,   3   ,   19  ,   63  ,   41  ,   157 ,   43  ,   743 ,   675 ,   2149    ,   6729    ,0 };
            const unsigned long dim1011KuoInit[]        =   {   1   ,   1   ,   5   ,   1   ,   21  ,   35  ,   45  ,   145 ,   235 ,   185 ,   1563    ,   1117    ,   3669    ,0 };
            const unsigned long dim1012KuoInit[]        =   {   1   ,   3   ,   1   ,   9   ,   9   ,   37  ,   105 ,   61  ,   259 ,   935 ,   793 ,   823 ,   1765    ,0 };
            const unsigned long dim1013KuoInit[]        =   {   1   ,   3   ,   3   ,   11  ,   25  ,   41  ,   35  ,   11  ,   389 ,   399 ,   339 ,   859 ,   4299    ,0 };
            const unsigned long dim1014KuoInit[]        =   {   1   ,   3   ,   5   ,   13  ,   7   ,   29  ,   53  ,   91  ,   185 ,   163 ,   1689    ,   983 ,   4545    ,0 };
            const unsigned long dim1015KuoInit[]        =   {   1   ,   3   ,   7   ,   11  ,   19  ,   35  ,   45  ,   141 ,   445 ,   469 ,   1827    ,   3523    ,   377 ,0 };
            const unsigned long dim1016KuoInit[]        =   {   1   ,   3   ,   5   ,   13  ,   27  ,   29  ,   35  ,   199 ,   73  ,   163 ,   1591    ,   1021    ,   2867    ,0 };
            const unsigned long dim1017KuoInit[]        =   {   1   ,   1   ,   1   ,   15  ,   7   ,   61  ,   95  ,   95  ,   147 ,   959 ,   971 ,   649 ,   5047    ,0 };
            const unsigned long dim1018KuoInit[]        =   {   1   ,   1   ,   1   ,   15  ,   11  ,   23  ,   49  ,   231 ,   359 ,   677 ,   1401    ,   2889    ,   3799    ,0 };
            const unsigned long dim1019KuoInit[]        =   {   1   ,   3   ,   7   ,   11  ,   11  ,   15  ,   119 ,   103 ,   403 ,   983 ,   399 ,   321 ,   437 ,0 };
            const unsigned long dim1020KuoInit[]        =   {   1   ,   3   ,   7   ,   1   ,   29  ,   13  ,   83  ,   55  ,   175 ,   835 ,   1637    ,   209 ,   1923    ,0 };
            const unsigned long dim1021KuoInit[]        =   {   1   ,   1   ,   5   ,   13  ,   25  ,   37  ,   5   ,   239 ,   227 ,   229 ,   243 ,   1837    ,   4821    ,0 };
            const unsigned long dim1022KuoInit[]        =   {   1   ,   1   ,   7   ,   9   ,   9   ,   59  ,   29  ,   135 ,   227 ,   473 ,   1759    ,   3711    ,   4113    ,0 };
            const unsigned long dim1023KuoInit[]        =   {   1   ,   3   ,   3   ,   3   ,   15  ,   3   ,   53  ,   225 ,   165 ,   375 ,   537 ,   331 ,   5085    ,0 };
            const unsigned long dim1024KuoInit[]        =   {   1   ,   1   ,   1   ,   5   ,   15  ,   41  ,   57  ,   165 ,   45  ,   137 ,   931 ,   1015    ,   6451    ,0 };
            const unsigned long dim1025KuoInit[]        =   {   1   ,   3   ,   5   ,   7   ,   5   ,   47  ,   35  ,   249 ,   89  ,   111 ,   1275    ,   529 ,   3297    ,0 };
            const unsigned long dim1026KuoInit[]        =   {   1   ,   3   ,   1   ,   1   ,   27  ,   21  ,   65  ,   193 ,   381 ,   861 ,   59  ,   1457    ,   2447    ,0 };
            const unsigned long dim1027KuoInit[]        =   {   1   ,   1   ,   7   ,   15  ,   17  ,   25  ,   75  ,   171 ,   263 ,   401 ,   1369    ,   2609    ,   6265    ,0 };
            const unsigned long dim1028KuoInit[]        =   {   1   ,   1   ,   3   ,   7   ,   17  ,   61  ,   75  ,   77  ,   487 ,   7   ,   1565    ,   527 ,   7213    ,0 };
            const unsigned long dim1029KuoInit[]        =   {   1   ,   3   ,   5   ,   9   ,   5   ,   45  ,   47  ,   27  ,   163 ,   43  ,   219 ,   2119    ,   1793    ,0 };
            const unsigned long dim1030KuoInit[]        =   {   1   ,   1   ,   3   ,   3   ,   27  ,   9   ,   89  ,   151 ,   487 ,   925 ,   1257    ,   1065    ,   3413    ,0 };
            const unsigned long dim1031KuoInit[]        =   {   1   ,   1   ,   5   ,   3   ,   11  ,   27  ,   9   ,   161 ,   303 ,   457 ,   1367    ,   2319    ,   6145    ,0 };
            const unsigned long dim1032KuoInit[]        =   {   1   ,   3   ,   1   ,   13  ,   23  ,   43  ,   115 ,   227 ,   319 ,   469 ,   1651    ,   1693    ,   2095    ,0 };
            const unsigned long dim1033KuoInit[]        =   {   1   ,   1   ,   1   ,   11  ,   23  ,   5   ,   91  ,   103 ,   393 ,   787 ,   1557    ,   925 ,   7189    ,0 };
            const unsigned long dim1034KuoInit[]        =   {   1   ,   3   ,   5   ,   7   ,   31  ,   5   ,   51  ,   171 ,   375 ,   309 ,   199 ,   1833    ,   2741    ,0 };
            const unsigned long dim1035KuoInit[]        =   {   1   ,   1   ,   5   ,   13  ,   27  ,   59  ,   15  ,   255 ,   205 ,   807 ,   1407    ,   161 ,   6955    ,0 };
            const unsigned long dim1036KuoInit[]        =   {   1   ,   3   ,   5   ,   5   ,   5   ,   35  ,   7   ,   167 ,   265 ,   67  ,   1133    ,   699 ,   6225    ,0 };
            const unsigned long dim1037KuoInit[]        =   {   1   ,   1   ,   1   ,   7   ,   19  ,   31  ,   5   ,   243 ,   279 ,   505 ,   303 ,   2303    ,   5367    ,0 };
            const unsigned long dim1038KuoInit[]        =   {   1   ,   3   ,   1   ,   3   ,   19  ,   47  ,   105 ,   159 ,   309 ,   783 ,   1845    ,   2603    ,   4663    ,0 };
            const unsigned long dim1039KuoInit[]        =   {   1   ,   1   ,   7   ,   13  ,   13  ,   35  ,   101 ,   123 ,   99  ,   559 ,   1681    ,   2751    ,   1965    ,0 };
            const unsigned long dim1040KuoInit[]        =   {   1   ,   3   ,   7   ,   5   ,   1   ,   63  ,   29  ,   245 ,   105 ,   469 ,   939 ,   721 ,   6213    ,0 };
            const unsigned long dim1041KuoInit[]        =   {   1   ,   3   ,   7   ,   15  ,   11  ,   45  ,   13  ,   121 ,   81  ,   15  ,   1503    ,   2203    ,   7467    ,0 };
            const unsigned long dim1042KuoInit[]        =   {   1   ,   1   ,   7   ,   11  ,   29  ,   25  ,   59  ,   93  ,   5   ,   551 ,   1799    ,   2251    ,   115 ,0 };
            const unsigned long dim1043KuoInit[]        =   {   1   ,   1   ,   1   ,   7   ,   17  ,   37  ,   81  ,   117 ,   183 ,   301 ,   1085    ,   3925    ,   697 ,0 };
            const unsigned long dim1044KuoInit[]        =   {   1   ,   3   ,   7   ,   7   ,   13  ,   13  ,   5   ,   207 ,   443 ,   723 ,   897 ,   3481    ,   3377    ,0 };
            const unsigned long dim1045KuoInit[]        =   {   1   ,   3   ,   7   ,   13  ,   9   ,   41  ,   25  ,   161 ,   283 ,   3   ,   1515    ,   2445    ,   2179    ,0 };
            const unsigned long dim1046KuoInit[]        =   {   1   ,   1   ,   1   ,   1   ,   29  ,   17  ,   73  ,   129 ,   499 ,   457 ,   103 ,   2287    ,   525 ,0 };
            const unsigned long dim1047KuoInit[]        =   {   1   ,   3   ,   7   ,   11  ,   9   ,   13  ,   95  ,   31  ,   411 ,   185 ,   931 ,   119 ,   711 ,0 };
            const unsigned long dim1048KuoInit[]        =   {   1   ,   1   ,   3   ,   13  ,   19  ,   53  ,   49  ,   209 ,   217 ,   593 ,   733 ,   4057    ,   2853    ,0 };
            const unsigned long dim1049KuoInit[]        =   {   1   ,   3   ,   5   ,   1   ,   7   ,   63  ,   95  ,   45  ,   209 ,   641 ,   1767    ,   3001    ,   4089    ,0 };
            const unsigned long dim1050KuoInit[]        =   {   1   ,   1   ,   5   ,   9   ,   9   ,   25  ,   69  ,   85  ,   345 ,   475 ,   613 ,   3975    ,   2179    ,0 };
            const unsigned long dim1051KuoInit[]        =   {   1   ,   1   ,   1   ,   9   ,   19  ,   9   ,   47  ,   51  ,   111 ,   7   ,   627 ,   1635    ,   3857    ,0 };
            const unsigned long dim1052KuoInit[]        =   {   1   ,   3   ,   7   ,   1   ,   25  ,   37  ,   27  ,   253 ,   111 ,   369 ,   235 ,   2887    ,   5649    ,0 };
            const unsigned long dim1053KuoInit[]        =   {   1   ,   1   ,   7   ,   3   ,   15  ,   5   ,   13  ,   181 ,   471 ,   825 ,   121 ,   3007    ,   777 ,0 };
            const unsigned long dim1054KuoInit[]        =   {   1   ,   1   ,   1   ,   9   ,   27  ,   41  ,   59  ,   9   ,   391 ,   779 ,   915 ,   3965    ,   4431    ,0 };
            const unsigned long dim1055KuoInit[]        =   {   1   ,   3   ,   1   ,   1   ,   25  ,   27  ,   75  ,   123 ,   385 ,   923 ,   141 ,   3771    ,   2227    ,0 };
            const unsigned long dim1056KuoInit[]        =   {   1   ,   3   ,   3   ,   15  ,   17  ,   29  ,   85  ,   253 ,   261 ,   517 ,   87  ,   621 ,   7947    ,0 };
            const unsigned long dim1057KuoInit[]        =   {   1   ,   3   ,   7   ,   5   ,   13  ,   15  ,   83  ,   49  ,   399 ,   875 ,   845 ,   1731    ,   7071    ,0 };
            const unsigned long dim1058KuoInit[]        =   {   1   ,   1   ,   3   ,   5   ,   21  ,   49  ,   93  ,   121 ,   271 ,   921 ,   183 ,   1533    ,   1609    ,0 };
            const unsigned long dim1059KuoInit[]        =   {   1   ,   3   ,   7   ,   7   ,   9   ,   51  ,   89  ,   135 ,   389 ,   151 ,   1813    ,   3673    ,   2099    ,0 };
            const unsigned long dim1060KuoInit[]        =   {   1   ,   1   ,   1   ,   11  ,   25  ,   63  ,   127 ,   239 ,   329 ,   455 ,   403 ,   363 ,   2689    ,0 };
            const unsigned long dim1061KuoInit[]        =   {   1   ,   3   ,   5   ,   13  ,   1   ,   45  ,   65  ,   93  ,   371 ,   577 ,   669 ,   2433    ,   3507    ,0 };
            const unsigned long dim1062KuoInit[]        =   {   1   ,   3   ,   5   ,   9   ,   9   ,   15  ,   13  ,   85  ,   67  ,   809 ,   1133    ,   129 ,   3869    ,0 };
            const unsigned long dim1063KuoInit[]        =   {   1   ,   3   ,   7   ,   15  ,   3   ,   47  ,   31  ,   79  ,   395 ,   227 ,   677 ,   3399    ,   8173    ,0 };
            const unsigned long dim1064KuoInit[]        =   {   1   ,   3   ,   1   ,   11  ,   7   ,   19  ,   81  ,   27  ,   393 ,   49  ,   299 ,   879 ,   5401    ,0 };
            const unsigned long dim1065KuoInit[]        =   {   1   ,   1   ,   1   ,   1   ,   3   ,   25  ,   105 ,   67  ,   147 ,   23  ,   749 ,   2677    ,   2799    ,0 };
            const unsigned long dim1066KuoInit[]        =   {   1   ,   1   ,   3   ,   15  ,   3   ,   49  ,   55  ,   131 ,   249 ,   183 ,   399 ,   4065    ,   4575    ,0 };
            const unsigned long dim1067KuoInit[]        =   {   1   ,   3   ,   7   ,   13  ,   21  ,   1   ,   105 ,   167 ,   379 ,   567 ,   1843    ,   3019    ,   7479    ,0 };
            const unsigned long dim1068KuoInit[]        =   {   1   ,   1   ,   7   ,   3   ,   7   ,   15  ,   65  ,   213 ,   191 ,   735 ,   1087    ,   1223    ,   7827    ,0 };
            const unsigned long dim1069KuoInit[]        =   {   1   ,   1   ,   3   ,   3   ,   13  ,   35  ,   79  ,   55  ,   169 ,   513 ,   639 ,   1153    ,   1129    ,0 };
            const unsigned long dim1070KuoInit[]        =   {   1   ,   3   ,   5   ,   15  ,   1   ,   37  ,   49  ,   209 ,   355 ,   533 ,   1073    ,   3249    ,   4309    ,0 };
            const unsigned long dim1071KuoInit[]        =   {   1   ,   1   ,   3   ,   15  ,   23  ,   45  ,   69  ,   195 ,   155 ,   267 ,   291 ,   1773    ,   3575    ,0 };
            const unsigned long dim1072KuoInit[]        =   {   1   ,   3   ,   1   ,   13  ,   25  ,   3   ,   51  ,   175 ,   505 ,   467 ,   1909    ,   1929    ,   4909    ,0 };
            const unsigned long dim1073KuoInit[]        =   {   1   ,   3   ,   3   ,   1   ,   31  ,   51  ,   99  ,   119 ,   505 ,   57  ,   1987    ,   1303    ,   4273    ,0 };
            const unsigned long dim1074KuoInit[]        =   {   1   ,   3   ,   7   ,   9   ,   19  ,   49  ,   61  ,   147 ,   29  ,   1001    ,   633 ,   963 ,   5601    ,0 };
            const unsigned long dim1075KuoInit[]        =   {   1   ,   3   ,   5   ,   7   ,   1   ,   13  ,   101 ,   133 ,   31  ,   263 ,   2033    ,   575 ,   983 ,0 };
            const unsigned long dim1076KuoInit[]        =   {   1   ,   1   ,   1   ,   15  ,   21  ,   29  ,   101 ,   143 ,   263 ,   601 ,   1239    ,   1187    ,   3045    ,0 };
            const unsigned long dim1077KuoInit[]        =   {   1   ,   3   ,   3   ,   15  ,   5   ,   49  ,   7   ,   165 ,   411 ,   517 ,   619 ,   1517    ,   1839    ,0 };
            const unsigned long dim1078KuoInit[]        =   {   1   ,   3   ,   5   ,   15  ,   29  ,   45  ,   71  ,   245 ,   67  ,   469 ,   1793    ,   3353    ,   7799    ,0 };
            const unsigned long dim1079KuoInit[]        =   {   1   ,   1   ,   5   ,   7   ,   15  ,   49  ,   5   ,   183 ,   345 ,   585 ,   1877    ,   2205    ,   7963    ,0 };
            const unsigned long dim1080KuoInit[]        =   {   1   ,   3   ,   1   ,   1   ,   7   ,   39  ,   85  ,   183 ,   131 ,   527 ,   1731    ,   1899    ,   3813    ,0 };
            const unsigned long dim1081KuoInit[]        =   {   1   ,   1   ,   7   ,   9   ,   29  ,   23  ,   77  ,   233 ,   77  ,   699 ,   1963    ,   171 ,   6557    ,0 };
            const unsigned long dim1082KuoInit[]        =   {   1   ,   3   ,   5   ,   15  ,   5   ,   3   ,   79  ,   199 ,   279 ,   279 ,   399 ,   769 ,   1661    ,0 };
            const unsigned long dim1083KuoInit[]        =   {   1   ,   3   ,   7   ,   1   ,   17  ,   61  ,   61  ,   227 ,   289 ,   541 ,   379 ,   3155    ,   5791    ,0 };
            const unsigned long dim1084KuoInit[]        =   {   1   ,   3   ,   3   ,   7   ,   31  ,   37  ,   73  ,   215 ,   47  ,   895 ,   1511    ,   305 ,   5247    ,0 };
            const unsigned long dim1085KuoInit[]        =   {   1   ,   1   ,   7   ,   5   ,   31  ,   47  ,   123 ,   191 ,   175 ,   875 ,   501 ,   2203    ,   1781    ,0 };
            const unsigned long dim1086KuoInit[]        =   {   1   ,   3   ,   3   ,   3   ,   13  ,   5   ,   43  ,   145 ,   183 ,   705 ,   1143    ,   3113    ,   1959    ,0 };
            const unsigned long dim1087KuoInit[]        =   {   1   ,   3   ,   7   ,   1   ,   23  ,   53  ,   93  ,   133 ,   13  ,   987 ,   1711    ,   2241    ,   7721    ,0 };
            const unsigned long dim1088KuoInit[]        =   {   1   ,   3   ,   1   ,   7   ,   3   ,   19  ,   123 ,   65  ,   421 ,   891 ,   159 ,   1633    ,   733 ,0 };
            const unsigned long dim1089KuoInit[]        =   {   1   ,   3   ,   3   ,   3   ,   23  ,   59  ,   127 ,   109 ,   315 ,   939 ,   1177    ,   1237    ,   1805    ,0 };
            const unsigned long dim1090KuoInit[]        =   {   1   ,   3   ,   5   ,   15  ,   13  ,   19  ,   39  ,   97  ,   437 ,   949 ,   539 ,   1171    ,   151 ,0 };
            const unsigned long dim1091KuoInit[]        =   {   1   ,   3   ,   5   ,   9   ,   13  ,   45  ,   107 ,   159 ,   363 ,   727 ,   509 ,   3051    ,   4071    ,0 };
            const unsigned long dim1092KuoInit[]        =   {   1   ,   3   ,   3   ,   1   ,   15  ,   11  ,   93  ,   199 ,   137 ,   947 ,   479 ,   2127    ,   1377    ,0 };
            const unsigned long dim1093KuoInit[]        =   {   1   ,   3   ,   7   ,   15  ,   13  ,   5   ,   67  ,   227 ,   69  ,   795 ,   299 ,   1845    ,   4765    ,0 };
            const unsigned long dim1094KuoInit[]        =   {   1   ,   1   ,   5   ,   7   ,   31  ,   15  ,   11  ,   255 ,   361 ,   599 ,   1235    ,   1785    ,   1611    ,0 };
            const unsigned long dim1095KuoInit[]        =   {   1   ,   3   ,   5   ,   15  ,   29  ,   21  ,   85  ,   49  ,   101 ,   785 ,   1071    ,   1333    ,   1691    ,0 };
            const unsigned long dim1096KuoInit[]        =   {   1   ,   3   ,   1   ,   1   ,   19  ,   11  ,   23  ,   157 ,   201 ,   67  ,   789 ,   999 ,   4281    ,0 };
            const unsigned long dim1097KuoInit[]        =   {   1   ,   1   ,   7   ,   3   ,   7   ,   61  ,   109 ,   253 ,   455 ,   863 ,   81  ,   3185    ,   7105    ,0 };
            const unsigned long dim1098KuoInit[]        =   {   1   ,   1   ,   1   ,   1   ,   11  ,   29  ,   51  ,   43  ,   211 ,   103 ,   1333    ,   3425    ,   6669    ,0 };
            const unsigned long dim1099KuoInit[]        =   {   1   ,   3   ,   3   ,   3   ,   19  ,   15  ,   25  ,   21  ,   9   ,   563 ,   1737    ,   159 ,   1827    ,0 };
            const unsigned long dim1100KuoInit[]        =   {   1   ,   3   ,   5   ,   11  ,   9   ,   35  ,   19  ,   145 ,   159 ,   345 ,   475 ,   1453    ,   7807    ,0 };
            const unsigned long dim1101KuoInit[]        =   {   1   ,   3   ,   7   ,   9   ,   19  ,   41  ,   61  ,   63  ,   77  ,   3   ,   997 ,   223 ,   5493    ,0 };
            const unsigned long dim1102KuoInit[]        =   {   1   ,   3   ,   7   ,   15  ,   25  ,   17  ,   87  ,   99  ,   441 ,   973 ,   713 ,   631 ,   4039    ,0 };
            const unsigned long dim1103KuoInit[]        =   {   1   ,   3   ,   5   ,   15  ,   17  ,   19  ,   89  ,   11  ,   321 ,   143 ,   11  ,   915 ,   3359    ,0 };
            const unsigned long dim1104KuoInit[]        =   {   1   ,   3   ,   7   ,   1   ,   11  ,   55  ,   43  ,   3   ,   1   ,   893 ,   1845    ,   1237    ,   693 ,0 };
            const unsigned long dim1105KuoInit[]        =   {   1   ,   3   ,   7   ,   13  ,   25  ,   21  ,   93  ,   43  ,   127 ,   197 ,   1349    ,   1727    ,   1133    ,0 };
            const unsigned long dim1106KuoInit[]        =   {   1   ,   3   ,   7   ,   15  ,   29  ,   57  ,   19  ,   35  ,   37  ,   81  ,   1599    ,   3469    ,   4327    ,0 };
            const unsigned long dim1107KuoInit[]        =   {   1   ,   1   ,   7   ,   9   ,   15  ,   47  ,   119 ,   203 ,   301 ,   835 ,   1221    ,   2771    ,   2023    ,0 };
            const unsigned long dim1108KuoInit[]        =   {   1   ,   3   ,   5   ,   9   ,   25  ,   43  ,   109 ,   51  ,   431 ,   189 ,   255 ,   2595    ,   2447    ,0 };
            const unsigned long dim1109KuoInit[]        =   {   1   ,   1   ,   5   ,   5   ,   13  ,   3   ,   113 ,   109 ,   321 ,   837 ,   159 ,   2617    ,   2277    ,0 };
            const unsigned long dim1110KuoInit[]        =   {   1   ,   1   ,   1   ,   11  ,   29  ,   43  ,   33  ,   173 ,   47  ,   395 ,   1113    ,   1711    ,   7909    ,0 };
            const unsigned long dim1111KuoInit[]        =   {   1   ,   1   ,   3   ,   15  ,   31  ,   45  ,   37  ,   95  ,   13  ,   449 ,   1377    ,   2117    ,   6607    ,   11053   ,0 };
            const unsigned long dim1112KuoInit[]        =   {   1   ,   1   ,   1   ,   9   ,   21  ,   7   ,   127 ,   71  ,   11  ,   271 ,   65  ,   3681    ,   4209    ,   3183    ,0 };
            const unsigned long dim1113KuoInit[]        =   {   1   ,   3   ,   3   ,   15  ,   19  ,   27  ,   99  ,   243 ,   371 ,   77  ,   629 ,   3989    ,   3105    ,   6287    ,0 };
            const unsigned long dim1114KuoInit[]        =   {   1   ,   3   ,   1   ,   15  ,   7   ,   35  ,   115 ,   237 ,   403 ,   273 ,   697 ,   117 ,   6887    ,   3243    ,0 };
            const unsigned long dim1115KuoInit[]        =   {   1   ,   1   ,   5   ,   9   ,   29  ,   1   ,   117 ,   101 ,   483 ,   909 ,   867 ,   2051    ,   7081    ,   14905   ,0 };
            const unsigned long dim1116KuoInit[]        =   {   1   ,   1   ,   7   ,   13  ,   21  ,   49  ,   107 ,   101 ,   121 ,   979 ,   309 ,   1871    ,   7339    ,   4213    ,0 };
            const unsigned long dim1117KuoInit[]        =   {   1   ,   1   ,   7   ,   13  ,   9   ,   25  ,   13  ,   61  ,   409 ,   469 ,   411 ,   2041    ,   7685    ,   15501   ,0 };
            const unsigned long dim1118KuoInit[]        =   {   1   ,   1   ,   7   ,   5   ,   1   ,   29  ,   109 ,   125 ,   63  ,   829 ,   175 ,   957 ,   4781    ,   12473   ,0 };
            const unsigned long dim1119KuoInit[]        =   {   1   ,   3   ,   3   ,   1   ,   11  ,   53  ,   17  ,   171 ,   393 ,   165 ,   51  ,   737 ,   1697    ,   4137    ,0 };
            const unsigned long dim1120KuoInit[]        =   {   1   ,   3   ,   7   ,   9   ,   15  ,   45  ,   49  ,   219 ,   105 ,   121 ,   549 ,   991 ,   1337    ,   11311   ,0 };
            const unsigned long dim1121KuoInit[]        =   {   1   ,   3   ,   3   ,   5   ,   27  ,   9   ,   39  ,   207 ,   85  ,   185 ,   901 ,   3077    ,   5295    ,   3933    ,0 };
            const unsigned long dim1122KuoInit[]        =   {   1   ,   1   ,   5   ,   7   ,   23  ,   29  ,   95  ,   63  ,   125 ,   447 ,   573 ,   2053    ,   4621    ,   10135   ,0 };
            const unsigned long dim1123KuoInit[]        =   {   1   ,   1   ,   1   ,   7   ,   23  ,   45  ,   3   ,   109 ,   335 ,   339 ,   1453    ,   3121    ,   37  ,   9155    ,0 };
            const unsigned long dim1124KuoInit[]        =   {   1   ,   3   ,   1   ,   13  ,   9   ,   25  ,   27  ,   67  ,   35  ,   573 ,   215 ,   1969    ,   7083    ,   15729   ,0 };
            const unsigned long dim1125KuoInit[]        =   {   1   ,   3   ,   1   ,   9   ,   21  ,   3   ,   21  ,   143 ,   131 ,   613 ,   299 ,   1127    ,   5647    ,   15007   ,0 };
            const unsigned long dim1126KuoInit[]        =   {   1   ,   3   ,   7   ,   9   ,   29  ,   47  ,   53  ,   223 ,   409 ,   633 ,   535 ,   2423    ,   5763    ,   11633   ,0 };
            const unsigned long dim1127KuoInit[]        =   {   1   ,   3   ,   1   ,   9   ,   1   ,   21  ,   5   ,   1   ,   359 ,   547 ,   285 ,   779 ,   2091    ,   11785   ,0 };
            const unsigned long dim1128KuoInit[]        =   {   1   ,   3   ,   7   ,   7   ,   1   ,   31  ,   61  ,   245 ,   331 ,   45  ,   1559    ,   3223    ,   7577    ,   2157    ,0 };
            const unsigned long dim1129KuoInit[]        =   {   1   ,   1   ,   1   ,   7   ,   3   ,   27  ,   37  ,   191 ,   229 ,   205 ,   1535    ,   887 ,   701 ,   10867   ,0 };
            const unsigned long dim1130KuoInit[]        =   {   1   ,   3   ,   3   ,   15  ,   21  ,   53  ,   61  ,   177 ,   485 ,   575 ,   681 ,   1447    ,   1427    ,   11279   ,0 };
            const unsigned long dim1131KuoInit[]        =   {   1   ,   1   ,   1   ,   3   ,   17  ,   3   ,   5   ,   49  ,   311 ,   507 ,   995 ,   3971    ,   471 ,   3045    ,0 };
            const unsigned long dim1132KuoInit[]        =   {   1   ,   1   ,   3   ,   9   ,   7   ,   45  ,   73  ,   183 ,   419 ,   635 ,   523 ,   1793    ,   7549    ,   9727    ,0 };
            const unsigned long dim1133KuoInit[]        =   {   1   ,   1   ,   7   ,   3   ,   19  ,   33  ,   109 ,   91  ,   19  ,   825 ,   1629    ,   2111    ,   2673    ,   6677    ,0 };
            const unsigned long dim1134KuoInit[]        =   {   1   ,   1   ,   3   ,   9   ,   31  ,   13  ,   71  ,   199 ,   113 ,   77  ,   371 ,   3117    ,   7799    ,   15919   ,0 };
            const unsigned long dim1135KuoInit[]        =   {   1   ,   1   ,   7   ,   5   ,   27  ,   41  ,   91  ,   183 ,   75  ,   159 ,   911 ,   2081    ,   7065    ,   16313   ,0 };
            const unsigned long dim1136KuoInit[]        =   {   1   ,   3   ,   1   ,   15  ,   17  ,   29  ,   65  ,   9   ,   399 ,   259 ,   671 ,   525 ,   1393    ,   13367   ,0 };
            const unsigned long dim1137KuoInit[]        =   {   1   ,   3   ,   3   ,   13  ,   1   ,   29  ,   101 ,   187 ,   231 ,   413 ,   1245    ,   269 ,   201 ,   1673    ,0 };
            const unsigned long dim1138KuoInit[]        =   {   1   ,   1   ,   7   ,   3   ,   27  ,   47  ,   11  ,   149 ,   49  ,   309 ,   1923    ,   1187    ,   3165    ,   7927    ,0 };
            const unsigned long dim1139KuoInit[]        =   {   1   ,   1   ,   1   ,   5   ,   31  ,   57  ,   29  ,   203 ,   11  ,   305 ,   1771    ,   2997    ,   377 ,   5777    ,0 };
            const unsigned long dim1140KuoInit[]        =   {   1   ,   1   ,   7   ,   5   ,   1   ,   29  ,   29  ,   205 ,   93  ,   821 ,   1583    ,   2991    ,   4137    ,   14967   ,0 };
            const unsigned long dim1141KuoInit[]        =   {   1   ,   1   ,   3   ,   3   ,   7   ,   57  ,   13  ,   179 ,   261 ,   547 ,   219 ,   1979    ,   4951    ,   2529    ,0 };
            const unsigned long dim1142KuoInit[]        =   {   1   ,   3   ,   1   ,   5   ,   3   ,   51  ,   89  ,   7   ,   121 ,   611 ,   1655    ,   3639    ,   1991    ,   9911    ,0 };
            const unsigned long dim1143KuoInit[]        =   {   1   ,   3   ,   3   ,   15  ,   1   ,   21  ,   9   ,   173 ,   93  ,   579 ,   1685    ,   767 ,   5183    ,   3965    ,0 };
            const unsigned long dim1144KuoInit[]        =   {   1   ,   1   ,   3   ,   7   ,   15  ,   47  ,   55  ,   113 ,   27  ,   941 ,   1633    ,   3791    ,   1705    ,   16293   ,0 };
            const unsigned long dim1145KuoInit[]        =   {   1   ,   3   ,   3   ,   1   ,   7   ,   41  ,   11  ,   229 ,   367 ,   997 ,   543 ,   933 ,   4249    ,   7049    ,0 };
            const unsigned long dim1146KuoInit[]        =   {   1   ,   1   ,   1   ,   13  ,   7   ,   7   ,   25  ,   87  ,   213 ,   277 ,   627 ,   721 ,   6161    ,   14877   ,0 };
            const unsigned long dim1147KuoInit[]        =   {   1   ,   3   ,   5   ,   7   ,   25  ,   61  ,   47  ,   103 ,   199 ,   983 ,   543 ,   1935    ,   4663    ,   3597    ,0 };
            const unsigned long dim1148KuoInit[]        =   {   1   ,   3   ,   1   ,   15  ,   15  ,   49  ,   125 ,   27  ,   153 ,   521 ,   351 ,   1129    ,   5897    ,   11325   ,0 };
            const unsigned long dim1149KuoInit[]        =   {   1   ,   3   ,   1   ,   13  ,   13  ,   35  ,   81  ,   53  ,   217 ,   739 ,   963 ,   3441    ,   6843    ,   3505    ,0 };
            const unsigned long dim1150KuoInit[]        =   {   1   ,   3   ,   7   ,   1   ,   17  ,   35  ,   119 ,   187 ,   475 ,   919 ,   1325    ,   3269    ,   677 ,   617 ,0 };
            const unsigned long dim1151KuoInit[]        =   {   1   ,   1   ,   7   ,   15  ,   7   ,   15  ,   91  ,   99  ,   259 ,   37  ,   1461    ,   1145    ,   6973    ,   15643   ,0 };
            const unsigned long dim1152KuoInit[]        =   {   1   ,   1   ,   5   ,   15  ,   13  ,   11  ,   27  ,   75  ,   9   ,   611 ,   1331    ,   2057    ,   863 ,   6245    ,0 };
            const unsigned long dim1153KuoInit[]        =   {   1   ,   1   ,   1   ,   15  ,   3   ,   25  ,   117 ,   45  ,   141 ,   839 ,   1201    ,   629 ,   6289    ,   12231   ,0 };
            const unsigned long dim1154KuoInit[]        =   {   1   ,   1   ,   3   ,   13  ,   17  ,   61  ,   23  ,   243 ,   349 ,   153 ,   1747    ,   3107    ,   7763    ,   10067   ,0 };
            const unsigned long dim1155KuoInit[]        =   {   1   ,   1   ,   1   ,   1   ,   21  ,   25  ,   53  ,   67  ,   97  ,   851 ,   1921    ,   2247    ,   5769    ,   5779    ,0 };
            const unsigned long dim1156KuoInit[]        =   {   1   ,   3   ,   3   ,   3   ,   9   ,   37  ,   93  ,   247 ,   287 ,   609 ,   1643    ,   427 ,   4269    ,   15923   ,0 };
            const unsigned long dim1157KuoInit[]        =   {   1   ,   3   ,   3   ,   15  ,   7   ,   35  ,   111 ,   105 ,   357 ,   539 ,   319 ,   993 ,   6155    ,   6587    ,0 };
            const unsigned long dim1158KuoInit[]        =   {   1   ,   1   ,   3   ,   1   ,   13  ,   33  ,   99  ,   73  ,   89  ,   231 ,   1537    ,   309 ,   3837    ,   6669    ,0 };
            const unsigned long dim1159KuoInit[]        =   {   1   ,   3   ,   1   ,   1   ,   9   ,   25  ,   105 ,   197 ,   399 ,   259 ,   1381    ,   1631    ,   3551    ,   4593    ,0 };
            const unsigned long dim1160KuoInit[]        =   {   1   ,   3   ,   1   ,   3   ,   7   ,   63  ,   113 ,   65  ,   109 ,   529 ,   1451    ,   1169    ,   3321    ,   6811    ,0 };
            const unsigned long dim1161KuoInit[]        =   {   1   ,   1   ,   1   ,   5   ,   11  ,   55  ,   109 ,   27  ,   65  ,   199 ,   199 ,   2745    ,   2009    ,   14853   ,0 };
            const unsigned long dim1162KuoInit[]        =   {   1   ,   3   ,   1   ,   9   ,   13  ,   31  ,   109 ,   119 ,   431 ,   363 ,   1191    ,   703 ,   2617    ,   6817    ,0 };
            const unsigned long dim1163KuoInit[]        =   {   1   ,   1   ,   3   ,   11  ,   27  ,   1   ,   107 ,   11  ,   457 ,   495 ,   215 ,   1459    ,   4137    ,   13007   ,0 };
            const unsigned long dim1164KuoInit[]        =   {   1   ,   1   ,   7   ,   3   ,   7   ,   29  ,   11  ,   19  ,   159 ,   821 ,   1881    ,   3277    ,   1303    ,   7145    ,0 };
            const unsigned long dim1165KuoInit[]        =   {   1   ,   3   ,   3   ,   11  ,   25  ,   11  ,   125 ,   241 ,   197 ,   863 ,   423 ,   2107    ,   5885    ,   14279   ,0 };
            const unsigned long dim1166KuoInit[]        =   {   1   ,   3   ,   3   ,   1   ,   7   ,   29  ,   97  ,   249 ,   333 ,   945 ,   1081    ,   2405    ,   109 ,   2869    ,0 };
            const unsigned long dim1167KuoInit[]        =   {   1   ,   3   ,   3   ,   3   ,   27  ,   57  ,   89  ,   243 ,   481 ,   431 ,   269 ,   3371    ,   3667    ,   13293   ,0 };
            const unsigned long dim1168KuoInit[]        =   {   1   ,   3   ,   7   ,   7   ,   23  ,   3   ,   91  ,   57  ,   79  ,   165 ,   629 ,   2771    ,   7927    ,   2689    ,0 };
            const unsigned long dim1169KuoInit[]        =   {   1   ,   1   ,   1   ,   11  ,   17  ,   51  ,   81  ,   147 ,   19  ,   501 ,   861 ,   159 ,   2117    ,   6777    ,0 };
            const unsigned long dim1170KuoInit[]        =   {   1   ,   3   ,   3   ,   15  ,   7   ,   9   ,   51  ,   197 ,   97  ,   983 ,   133 ,   1819    ,   4449    ,   7821    ,0 };
            const unsigned long dim1171KuoInit[]        =   {   1   ,   3   ,   3   ,   7   ,   21  ,   55  ,   61  ,   171 ,   461 ,   741 ,   1901    ,   2195    ,   5477    ,   1681    ,0 };
            const unsigned long dim1172KuoInit[]        =   {   1   ,   1   ,   3   ,   1   ,   13  ,   37  ,   73  ,   151 ,   479 ,   561 ,   1483    ,   2679    ,   295 ,   7325    ,0 };
            const unsigned long dim1173KuoInit[]        =   {   1   ,   3   ,   3   ,   9   ,   11  ,   17  ,   127 ,   115 ,   371 ,   979 ,   1391    ,   2189    ,   2805    ,   7471    ,0 };
            const unsigned long dim1174KuoInit[]        =   {   1   ,   1   ,   3   ,   13  ,   31  ,   25  ,   65  ,   249 ,   91  ,   649 ,   1103    ,   493 ,   6341    ,   10313   ,0 };
            const unsigned long dim1175KuoInit[]        =   {   1   ,   1   ,   1   ,   3   ,   7   ,   31  ,   85  ,   201 ,   277 ,   771 ,   1007    ,   69  ,   3357    ,   15059   ,0 };
            const unsigned long dim1176KuoInit[]        =   {   1   ,   1   ,   1   ,   7   ,   17  ,   17  ,   9   ,   229 ,   269 ,   591 ,   945 ,   3003    ,   6723    ,   16101   ,0 };
            const unsigned long dim1177KuoInit[]        =   {   1   ,   1   ,   3   ,   5   ,   1   ,   29  ,   119 ,   181 ,   269 ,   269 ,   1781    ,   2935    ,   6265    ,   13989   ,0 };
            const unsigned long dim1178KuoInit[]        =   {   1   ,   3   ,   1   ,   11  ,   13  ,   51  ,   117 ,   141 ,   467 ,   521 ,   449 ,   797 ,   3447    ,   11693   ,0 };
            const unsigned long dim1179KuoInit[]        =   {   1   ,   1   ,   5   ,   15  ,   3   ,   57  ,   61  ,   209 ,   423 ,   213 ,   1663    ,   3735    ,   5229    ,   11809   ,0 };
            const unsigned long dim1180KuoInit[]        =   {   1   ,   3   ,   7   ,   9   ,   13  ,   51  ,   113 ,   127 ,   229 ,   793 ,   905 ,   3271    ,   7501    ,   10801   ,0 };
            const unsigned long dim1181KuoInit[]        =   {   1   ,   3   ,   7   ,   7   ,   5   ,   19  ,   43  ,   87  ,   9   ,   981 ,   221 ,   3925    ,   2095    ,   7185    ,0 };
            const unsigned long dim1182KuoInit[]        =   {   1   ,   1   ,   7   ,   15  ,   23  ,   35  ,   113 ,   175 ,   383 ,   699 ,   1617    ,   3771    ,   5847    ,   9097    ,0 };
            const unsigned long dim1183KuoInit[]        =   {   1   ,   3   ,   1   ,   7   ,   25  ,   23  ,   27  ,   77  ,   491 ,   211 ,   1785    ,   199 ,   4865    ,   10147   ,0 };
            const unsigned long dim1184KuoInit[]        =   {   1   ,   3   ,   1   ,   11  ,   25  ,   39  ,   15  ,   15  ,   191 ,   889 ,   457 ,   2647    ,   4555    ,   4799    ,0 };
            const unsigned long dim1185KuoInit[]        =   {   1   ,   1   ,   7   ,   5   ,   15  ,   1   ,   117 ,   169 ,   245 ,   917 ,   1025    ,   3805    ,   1455    ,   3051    ,0 };
            const unsigned long dim1186KuoInit[]        =   {   1   ,   3   ,   7   ,   5   ,   7   ,   15  ,   43  ,   13  ,   359 ,   685 ,   75  ,   271 ,   555 ,   3595    ,0 };
            const unsigned long dim1187KuoInit[]        =   {   1   ,   3   ,   7   ,   5   ,   19  ,   9   ,   29  ,   15  ,   139 ,   13  ,   753 ,   205 ,   677 ,   5945    ,0 };
            const unsigned long dim1188KuoInit[]        =   {   1   ,   3   ,   1   ,   9   ,   5   ,   21  ,   99  ,   77  ,   287 ,   663 ,   589 ,   229 ,   5255    ,   4177    ,0 };
            const unsigned long dim1189KuoInit[]        =   {   1   ,   1   ,   1   ,   9   ,   15  ,   47  ,   21  ,   149 ,   461 ,   871 ,   577 ,   3393    ,   6879    ,   6353    ,0 };
            const unsigned long dim1190KuoInit[]        =   {   1   ,   3   ,   3   ,   3   ,   21  ,   33  ,   29  ,   25  ,   195 ,   1   ,   1817    ,   3437    ,   4265    ,   16121   ,0 };
            const unsigned long dim1191KuoInit[]        =   {   1   ,   3   ,   1   ,   7   ,   27  ,   31  ,   93  ,   141 ,   471 ,   735 ,   1737    ,   2997    ,   3529    ,   12285   ,0 };
            const unsigned long dim1192KuoInit[]        =   {   1   ,   1   ,   3   ,   15  ,   15  ,   49  ,   105 ,   157 ,   165 ,   593 ,   651 ,   3331    ,   4333    ,   8607    ,0 };
            const unsigned long dim1193KuoInit[]        =   {   1   ,   3   ,   7   ,   1   ,   5   ,   35  ,   73  ,   53  ,   461 ,   41  ,   1407    ,   357 ,   3537    ,   1123    ,0 };
            const unsigned long dim1194KuoInit[]        =   {   1   ,   1   ,   7   ,   9   ,   15  ,   39  ,   83  ,   23  ,   501 ,   3   ,   743 ,   3819    ,   5491    ,   15317   ,0 };
            const unsigned long dim1195KuoInit[]        =   {   1   ,   1   ,   7   ,   15  ,   31  ,   29  ,   125 ,   173 ,   465 ,   347 ,   1515    ,   2871    ,   3861    ,   9353    ,0 };
            const unsigned long dim1196KuoInit[]        =   {   1   ,   1   ,   7   ,   5   ,   9   ,   31  ,   121 ,   221 ,   491 ,   579 ,   313 ,   3549    ,   711 ,   12501   ,0 };
            const unsigned long dim1197KuoInit[]        =   {   1   ,   1   ,   7   ,   15  ,   7   ,   47  ,   15  ,   175 ,   201 ,   399 ,   609 ,   749 ,   3587    ,   10091   ,0 };
            const unsigned long dim1198KuoInit[]        =   {   1   ,   1   ,   5   ,   11  ,   1   ,   39  ,   43  ,   69  ,   429 ,   87  ,   2017    ,   1961    ,   5679    ,   3179    ,0 };
            const unsigned long dim1199KuoInit[]        =   {   1   ,   1   ,   3   ,   3   ,   29  ,   51  ,   27  ,   13  ,   377 ,   99  ,   1291    ,   247 ,   2657    ,   4793    ,0 };
            const unsigned long dim1200KuoInit[]        =   {   1   ,   1   ,   1   ,   1   ,   23  ,   63  ,   119 ,   33  ,   405 ,   689 ,   517 ,   1331    ,   1047    ,   403 ,0 };
            const unsigned long dim1201KuoInit[]        =   {   1   ,   1   ,   1   ,   5   ,   9   ,   27  ,   121 ,   49  ,   317 ,   181 ,   623 ,   1789    ,   5867    ,   11009   ,0 };
            const unsigned long dim1202KuoInit[]        =   {   1   ,   1   ,   7   ,   15  ,   15  ,   63  ,   113 ,   129 ,   299 ,   889 ,   361 ,   2339    ,   3893    ,   3953    ,0 };
            const unsigned long dim1203KuoInit[]        =   {   1   ,   1   ,   1   ,   15  ,   9   ,   51  ,   73  ,   77  ,   511 ,   745 ,   639 ,   1261    ,   439 ,   13409   ,0 };
            const unsigned long dim1204KuoInit[]        =   {   1   ,   3   ,   5   ,   3   ,   29  ,   15  ,   37  ,   31  ,   47  ,   901 ,   1383    ,   1979    ,   7341    ,   13411   ,0 };
            const unsigned long dim1205KuoInit[]        =   {   1   ,   3   ,   7   ,   5   ,   31  ,   59  ,   87  ,   35  ,   123 ,   943 ,   1181    ,   3877    ,   4821    ,   9231    ,0 };
            const unsigned long dim1206KuoInit[]        =   {   1   ,   3   ,   5   ,   13  ,   3   ,   47  ,   123 ,   89  ,   41  ,   883 ,   241 ,   2055    ,   1723    ,   6747    ,0 };
            const unsigned long dim1207KuoInit[]        =   {   1   ,   1   ,   7   ,   5   ,   21  ,   29  ,   89  ,   101 ,   423 ,   279 ,   29  ,   2089    ,   6233    ,   7343    ,0 };
            const unsigned long dim1208KuoInit[]        =   {   1   ,   1   ,   7   ,   5   ,   29  ,   25  ,   23  ,   179 ,   13  ,   707 ,   1603    ,   585 ,   4159    ,   10767   ,0 };
            const unsigned long dim1209KuoInit[]        =   {   1   ,   1   ,   1   ,   1   ,   15  ,   11  ,   55  ,   159 ,   279 ,   613 ,   1445    ,   613 ,   499 ,   7095    ,0 };
            const unsigned long dim1210KuoInit[]        =   {   1   ,   1   ,   1   ,   5   ,   15  ,   61  ,   75  ,   171 ,   173 ,   765 ,   1763    ,   2193    ,   7897    ,   4559    ,0 };
            const unsigned long dim1211KuoInit[]        =   {   1   ,   1   ,   1   ,   9   ,   9   ,   57  ,   127 ,   113 ,   7   ,   269 ,   791 ,   619 ,   2923    ,   6309    ,0 };
            const unsigned long dim1212KuoInit[]        =   {   1   ,   1   ,   3   ,   13  ,   21  ,   9   ,   99  ,   1   ,   263 ,   69  ,   1879    ,   4079    ,   33  ,   7313    ,0 };
            const unsigned long dim1213KuoInit[]        =   {   1   ,   1   ,   7   ,   9   ,   17  ,   1   ,   69  ,   93  ,   147 ,   461 ,   1777    ,   3375    ,   4067    ,   15307   ,0 };
            const unsigned long dim1214KuoInit[]        =   {   1   ,   3   ,   7   ,   11  ,   11  ,   7   ,   77  ,   159 ,   125 ,   159 ,   625 ,   1339    ,   2301    ,   11525   ,0 };
            const unsigned long dim1215KuoInit[]        =   {   1   ,   1   ,   7   ,   9   ,   1   ,   5   ,   23  ,   115 ,   25  ,   755 ,   1523    ,   2767    ,   7667    ,   7129    ,0 };
            const unsigned long dim1216KuoInit[]        =   {   1   ,   1   ,   5   ,   5   ,   23  ,   31  ,   41  ,   3   ,   249 ,   69  ,   1821    ,   155 ,   585 ,   7905    ,0 };
            const unsigned long dim1217KuoInit[]        =   {   1   ,   3   ,   7   ,   11  ,   21  ,   35  ,   7   ,   161 ,   409 ,   771 ,   967 ,   2429    ,   1   ,   5391    ,0 };
            const unsigned long dim1218KuoInit[]        =   {   1   ,   1   ,   1   ,   9   ,   9   ,   3   ,   13  ,   125 ,   87  ,   115 ,   1013    ,   2103    ,   4769    ,   7389    ,0 };
            const unsigned long dim1219KuoInit[]        =   {   1   ,   3   ,   3   ,   7   ,   27  ,   25  ,   85  ,   119 ,   115 ,   775 ,   189 ,   3577    ,   5369    ,   5213    ,0 };
            const unsigned long dim1220KuoInit[]        =   {   1   ,   3   ,   7   ,   3   ,   27  ,   49  ,   81  ,   23  ,   223 ,   523 ,   1939    ,   431 ,   6247    ,   10509   ,0 };
            const unsigned long dim1221KuoInit[]        =   {   1   ,   3   ,   3   ,   3   ,   31  ,   7   ,   119 ,   235 ,   315 ,   431 ,   1669    ,   1223    ,   5785    ,   6507    ,0 };
            const unsigned long dim1222KuoInit[]        =   {   1   ,   1   ,   5   ,   9   ,   27  ,   23  ,   35  ,   141 ,   327 ,   395 ,   869 ,   177 ,   3997    ,   11191   ,0 };
            const unsigned long dim1223KuoInit[]        =   {   1   ,   1   ,   7   ,   9   ,   21  ,   15  ,   35  ,   163 ,   273 ,   93  ,   835 ,   3339    ,   6109    ,   2617    ,0 };
            const unsigned long dim1224KuoInit[]        =   {   1   ,   3   ,   5   ,   9   ,   1   ,   29  ,   107 ,   161 ,   301 ,   27  ,   1391    ,   211 ,   5263    ,   10899   ,0 };
            const unsigned long dim1225KuoInit[]        =   {   1   ,   3   ,   7   ,   7   ,   13  ,   61  ,   91  ,   179 ,   227 ,   193 ,   1909    ,   3759    ,   5875    ,   6799    ,0 };
            const unsigned long dim1226KuoInit[]        =   {   1   ,   3   ,   1   ,   9   ,   17  ,   23  ,   57  ,   117 ,   201 ,   459 ,   1841    ,   2371    ,   2641    ,   13927   ,0 };
            const unsigned long dim1227KuoInit[]        =   {   1   ,   3   ,   7   ,   11  ,   1   ,   1   ,   13  ,   139 ,   313 ,   1   ,   1551    ,   2693    ,   2785    ,   6959    ,0 };
            const unsigned long dim1228KuoInit[]        =   {   1   ,   3   ,   3   ,   13  ,   1   ,   29  ,   117 ,   67  ,   147 ,   821 ,   1935    ,   2913    ,   7911    ,   4829    ,0 };
            const unsigned long dim1229KuoInit[]        =   {   1   ,   1   ,   7   ,   7   ,   19  ,   17  ,   57  ,   33  ,   93  ,   819 ,   1035    ,   3493    ,   1831    ,   5345    ,0 };
            const unsigned long dim1230KuoInit[]        =   {   1   ,   3   ,   5   ,   13  ,   9   ,   13  ,   117 ,   219 ,   415 ,   743 ,   381 ,   2343    ,   3865    ,   3915    ,0 };
            const unsigned long dim1231KuoInit[]        =   {   1   ,   1   ,   5   ,   15  ,   3   ,   25  ,   67  ,   197 ,   59  ,   419 ,   461 ,   1951    ,   4657    ,   9263    ,0 };
            const unsigned long dim1232KuoInit[]        =   {   1   ,   1   ,   7   ,   15  ,   31  ,   53  ,   33  ,   213 ,   169 ,   231 ,   615 ,   441 ,   7055    ,   16079   ,0 };
            const unsigned long dim1233KuoInit[]        =   {   1   ,   1   ,   1   ,   15  ,   5   ,   27  ,   57  ,   79  ,   373 ,   713 ,   583 ,   3773    ,   4239    ,   14231   ,0 };
            const unsigned long dim1234KuoInit[]        =   {   1   ,   1   ,   7   ,   15  ,   21  ,   55  ,   119 ,   219 ,   371 ,   321 ,   185 ,   1103    ,   7783    ,   8705    ,0 };
            const unsigned long dim1235KuoInit[]        =   {   1   ,   3   ,   5   ,   15  ,   3   ,   11  ,   69  ,   127 ,   27  ,   223 ,   867 ,   2249    ,   597 ,   10051   ,0 };
            const unsigned long dim1236KuoInit[]        =   {   1   ,   3   ,   5   ,   7   ,   5   ,   13  ,   99  ,   115 ,   21  ,   659 ,   1287    ,   1941    ,   6505    ,   1091    ,0 };
            const unsigned long dim1237KuoInit[]        =   {   1   ,   1   ,   7   ,   1   ,   17  ,   57  ,   57  ,   155 ,   373 ,   729 ,   1521    ,   3741    ,   4935    ,   13867   ,0 };
            const unsigned long dim1238KuoInit[]        =   {   1   ,   1   ,   7   ,   1   ,   19  ,   21  ,   39  ,   47  ,   451 ,   67  ,   463 ,   1849    ,   5311    ,   5831    ,0 };
            const unsigned long dim1239KuoInit[]        =   {   1   ,   3   ,   7   ,   7   ,   29  ,   11  ,   59  ,   161 ,   29  ,   107 ,   1413    ,   3915    ,   6863    ,   5095    ,0 };
            const unsigned long dim1240KuoInit[]        =   {   1   ,   1   ,   7   ,   15  ,   13  ,   37  ,   29  ,   91  ,   83  ,   663 ,   1629    ,   3005    ,   655 ,   7601    ,0 };
            const unsigned long dim1241KuoInit[]        =   {   1   ,   1   ,   3   ,   3   ,   31  ,   47  ,   29  ,   31  ,   509 ,   319 ,   929 ,   1481    ,   6255    ,   14003   ,0 };
            const unsigned long dim1242KuoInit[]        =   {   1   ,   3   ,   3   ,   3   ,   5   ,   7   ,   97  ,   253 ,   163 ,   801 ,   283 ,   813 ,   6277    ,   3201    ,0 };
            const unsigned long dim1243KuoInit[]        =   {   1   ,   3   ,   3   ,   15  ,   5   ,   59  ,   49  ,   47  ,   259 ,   837 ,   2003    ,   1489    ,   3425    ,   6863    ,0 };
            const unsigned long dim1244KuoInit[]        =   {   1   ,   3   ,   5   ,   13  ,   25  ,   55  ,   37  ,   139 ,   347 ,   95  ,   259 ,   3175    ,   1043    ,   1395    ,0 };
            const unsigned long dim1245KuoInit[]        =   {   1   ,   3   ,   7   ,   5   ,   31  ,   37  ,   73  ,   7   ,   1   ,   883 ,   1641    ,   1941    ,   6885    ,   23  ,0 };
            const unsigned long dim1246KuoInit[]        =   {   1   ,   1   ,   3   ,   9   ,   13  ,   45  ,   85  ,   67  ,   55  ,   11  ,   1717    ,   3461    ,   3189    ,   15989   ,0 };
            const unsigned long dim1247KuoInit[]        =   {   1   ,   1   ,   1   ,   13  ,   27  ,   15  ,   93  ,   39  ,   329 ,   43  ,   175 ,   2253    ,   1193    ,   13257   ,0 };
            const unsigned long dim1248KuoInit[]        =   {   1   ,   3   ,   3   ,   1   ,   25  ,   41  ,   111 ,   211 ,   99  ,   267 ,   965 ,   1311    ,   2661    ,   1981    ,0 };
            const unsigned long dim1249KuoInit[]        =   {   1   ,   3   ,   7   ,   15  ,   31  ,   15  ,   37  ,   73  ,   239 ,   505 ,   983 ,   3299    ,   2891    ,   8247    ,0 };
            const unsigned long dim1250KuoInit[]        =   {   1   ,   1   ,   5   ,   3   ,   19  ,   5   ,   45  ,   55  ,   251 ,   125 ,   977 ,   955 ,   1169    ,   8495    ,0 };
            const unsigned long dim1251KuoInit[]        =   {   1   ,   3   ,   1   ,   3   ,   19  ,   31  ,   121 ,   119 ,   185 ,   43  ,   847 ,   3939    ,   2511    ,   10889   ,0 };
            const unsigned long dim1252KuoInit[]        =   {   1   ,   3   ,   1   ,   7   ,   7   ,   17  ,   81  ,   37  ,   425 ,   663 ,   1715    ,   789 ,   661 ,   9089    ,0 };
            const unsigned long dim1253KuoInit[]        =   {   1   ,   3   ,   3   ,   11  ,   1   ,   3   ,   17  ,   247 ,   311 ,   369 ,   211 ,   97  ,   6355    ,   16297   ,0 };
            const unsigned long dim1254KuoInit[]        =   {   1   ,   3   ,   7   ,   13  ,   21  ,   53  ,   33  ,   71  ,   73  ,   629 ,   1321    ,   3181    ,   6627    ,   161 ,0 };
            const unsigned long dim1255KuoInit[]        =   {   1   ,   1   ,   3   ,   5   ,   21  ,   25  ,   21  ,   231 ,   291 ,   585 ,   111 ,   1651    ,   1031    ,   13705   ,0 };
            const unsigned long dim1256KuoInit[]        =   {   1   ,   3   ,   3   ,   9   ,   5   ,   31  ,   59  ,   51  ,   245 ,   465 ,   37  ,   3425    ,   4251    ,   1121    ,0 };
            const unsigned long dim1257KuoInit[]        =   {   1   ,   3   ,   7   ,   11  ,   31  ,   45  ,   125 ,   245 ,   335 ,   973 ,   49  ,   1557    ,   4847    ,   10739   ,0 };
            const unsigned long dim1258KuoInit[]        =   {   1   ,   1   ,   1   ,   3   ,   21  ,   39  ,   127 ,   67  ,   101 ,   9   ,   1099    ,   1159    ,   2483    ,   4283    ,0 };
            const unsigned long dim1259KuoInit[]        =   {   1   ,   1   ,   7   ,   13  ,   29  ,   53  ,   49  ,   65  ,   5   ,   775 ,   433 ,   351 ,   493 ,   7233    ,0 };
            const unsigned long dim1260KuoInit[]        =   {   1   ,   3   ,   5   ,   5   ,   27  ,   13  ,   73  ,   83  ,   417 ,   31  ,   1891    ,   1809    ,   1103    ,   4555    ,0 };
            const unsigned long dim1261KuoInit[]        =   {   1   ,   1   ,   5   ,   9   ,   15  ,   1   ,   25  ,   63  ,   191 ,   251 ,   1759    ,   429 ,   2477    ,   12577   ,0 };
            const unsigned long dim1262KuoInit[]        =   {   1   ,   1   ,   3   ,   5   ,   3   ,   21  ,   61  ,   245 ,   191 ,   217 ,   845 ,   2121    ,   2133    ,   6045    ,0 };
            const unsigned long dim1263KuoInit[]        =   {   1   ,   3   ,   3   ,   5   ,   23  ,   33  ,   85  ,   15  ,   57  ,   645 ,   1109    ,   3411    ,   2895    ,   6885    ,0 };
            const unsigned long dim1264KuoInit[]        =   {   1   ,   3   ,   3   ,   11  ,   13  ,   41  ,   13  ,   59  ,   45  ,   199 ,   1145    ,   2637    ,   6741    ,   1273    ,0 };
            const unsigned long dim1265KuoInit[]        =   {   1   ,   3   ,   3   ,   5   ,   27  ,   33  ,   121 ,   21  ,   75  ,   687 ,   349 ,   2097    ,   1195    ,   4227    ,0 };
            const unsigned long dim1266KuoInit[]        =   {   1   ,   1   ,   7   ,   9   ,   25  ,   51  ,   71  ,   85  ,   281 ,   825 ,   1335    ,   277 ,   797 ,   407 ,0 };
            const unsigned long dim1267KuoInit[]        =   {   1   ,   3   ,   3   ,   15  ,   21  ,   31  ,   19  ,   11  ,   39  ,   821 ,   1623    ,   97  ,   1573    ,   13297   ,0 };
            const unsigned long dim1268KuoInit[]        =   {   1   ,   1   ,   3   ,   1   ,   25  ,   39  ,   67  ,   113 ,   109 ,   719 ,   1427    ,   2539    ,   5885    ,   13537   ,0 };
            const unsigned long dim1269KuoInit[]        =   {   1   ,   1   ,   5   ,   3   ,   15  ,   53  ,   11  ,   139 ,   393 ,   365 ,   709 ,   1603    ,   4929    ,   2651    ,0 };
            const unsigned long dim1270KuoInit[]        =   {   1   ,   3   ,   7   ,   5   ,   3   ,   11  ,   19  ,   41  ,   123 ,   7   ,   993 ,   3627    ,   2027    ,   4009    ,0 };
            const unsigned long dim1271KuoInit[]        =   {   1   ,   1   ,   3   ,   7   ,   3   ,   17  ,   63  ,   129 ,   323 ,   303 ,   1669    ,   1023    ,   1259    ,   289 ,0 };
            const unsigned long dim1272KuoInit[]        =   {   1   ,   3   ,   7   ,   15  ,   15  ,   33  ,   69  ,   229 ,   291 ,   939 ,   969 ,   2497    ,   4323    ,   16299   ,0 };
            const unsigned long dim1273KuoInit[]        =   {   1   ,   3   ,   5   ,   5   ,   17  ,   35  ,   79  ,   33  ,   471 ,   857 ,   975 ,   297 ,   7219    ,   12903   ,0 };
            const unsigned long dim1274KuoInit[]        =   {   1   ,   1   ,   7   ,   7   ,   21  ,   5   ,   65  ,   53  ,   11  ,   265 ,   1447    ,   1405    ,   3351    ,   2661    ,0 };
            const unsigned long dim1275KuoInit[]        =   {   1   ,   3   ,   3   ,   3   ,   1   ,   21  ,   123 ,   123 ,   361 ,   71  ,   1233    ,   2673    ,   7643    ,   9519    ,0 };
            const unsigned long dim1276KuoInit[]        =   {   1   ,   1   ,   5   ,   7   ,   27  ,   41  ,   117 ,   239 ,   125 ,   183 ,   1375    ,   649 ,   5267    ,   1735    ,0 };
            const unsigned long dim1277KuoInit[]        =   {   1   ,   1   ,   1   ,   11  ,   7   ,   33  ,   59  ,   7   ,   37  ,   705 ,   1055    ,   1661    ,   6213    ,   2083    ,0 };
            const unsigned long dim1278KuoInit[]        =   {   1   ,   3   ,   1   ,   9   ,   11  ,   61  ,   91  ,   143 ,   113 ,   229 ,   809 ,   431 ,   1971    ,   12453   ,0 };
            const unsigned long dim1279KuoInit[]        =   {   1   ,   1   ,   3   ,   9   ,   17  ,   23  ,   97  ,   249 ,   109 ,   669 ,   1643    ,   3305    ,   6141    ,   1793    ,0 };
            const unsigned long dim1280KuoInit[]        =   {   1   ,   1   ,   5   ,   13  ,   3   ,   39  ,   37  ,   107 ,   123 ,   637 ,   869 ,   1969    ,   4195    ,   9431    ,0 };
            const unsigned long dim1281KuoInit[]        =   {   1   ,   1   ,   7   ,   13  ,   7   ,   25  ,   97  ,   63  ,   395 ,   1011    ,   161 ,   827 ,   4001    ,   15159   ,0 };
            const unsigned long dim1282KuoInit[]        =   {   1   ,   1   ,   3   ,   7   ,   11  ,   59  ,   11  ,   49  ,   25  ,   691 ,   73  ,   1171    ,   2937    ,   13877   ,0 };
            const unsigned long dim1283KuoInit[]        =   {   1   ,   3   ,   7   ,   15  ,   1   ,   19  ,   119 ,   129 ,   63  ,   451 ,   1593    ,   1207    ,   67  ,   3621    ,0 };
            const unsigned long dim1284KuoInit[]        =   {   1   ,   1   ,   1   ,   1   ,   25  ,   27  ,   45  ,   23  ,   111 ,   191 ,   1205    ,   4079    ,   1821    ,   5571    ,0 };
            const unsigned long dim1285KuoInit[]        =   {   1   ,   1   ,   7   ,   5   ,   17  ,   37  ,   37  ,   249 ,   81  ,   909 ,   153 ,   209 ,   7919    ,   10763   ,0 };
            const unsigned long dim1286KuoInit[]        =   {   1   ,   3   ,   5   ,   5   ,   19  ,   9   ,   101 ,   193 ,   431 ,   63  ,   1309    ,   2693    ,   2413    ,   14513   ,0 };
            const unsigned long dim1287KuoInit[]        =   {   1   ,   3   ,   5   ,   7   ,   13  ,   5   ,   5   ,   203 ,   313 ,   723 ,   1637    ,   2725    ,   4197    ,   3767    ,0 };
            const unsigned long dim1288KuoInit[]        =   {   1   ,   3   ,   3   ,   11  ,   1   ,   53  ,   33  ,   245 ,   181 ,   1001    ,   1057    ,   2701    ,   5861    ,   4199    ,0 };
            const unsigned long dim1289KuoInit[]        =   {   1   ,   1   ,   5   ,   3   ,   31  ,   35  ,   19  ,   253 ,   403 ,   429 ,   1667    ,   1671    ,   7491    ,   1023    ,0 };
            const unsigned long dim1290KuoInit[]        =   {   1   ,   3   ,   5   ,   7   ,   15  ,   51  ,   41  ,   253 ,   247 ,   883 ,   1887    ,   1067    ,   3227    ,   13259   ,0 };
            const unsigned long dim1291KuoInit[]        =   {   1   ,   3   ,   7   ,   15  ,   5   ,   63  ,   9   ,   97  ,   87  ,   11  ,   1873    ,   2529    ,   833 ,   5583    ,0 };
            const unsigned long dim1292KuoInit[]        =   {   1   ,   1   ,   5   ,   7   ,   17  ,   59  ,   23  ,   225 ,   91  ,   189 ,   923 ,   3177    ,   599 ,   4315    ,0 };
            const unsigned long dim1293KuoInit[]        =   {   1   ,   3   ,   3   ,   11  ,   15  ,   59  ,   85  ,   75  ,   281 ,   935 ,   1219    ,   3121    ,   231 ,   13839   ,0 };
            const unsigned long dim1294KuoInit[]        =   {   1   ,   1   ,   5   ,   9   ,   15  ,   61  ,   105 ,   27  ,   179 ,   145 ,   827 ,   2985    ,   6907    ,   14961   ,0 };
            const unsigned long dim1295KuoInit[]        =   {   1   ,   1   ,   1   ,   1   ,   1   ,   63  ,   7   ,   3   ,   449 ,   387 ,   483 ,   603 ,   6607    ,   13241   ,0 };
            const unsigned long dim1296KuoInit[]        =   {   1   ,   1   ,   7   ,   1   ,   9   ,   37  ,   73  ,   255 ,   331 ,   797 ,   679 ,   579 ,   3015    ,   2517    ,0 };
            const unsigned long dim1297KuoInit[]        =   {   1   ,   1   ,   3   ,   5   ,   23  ,   15  ,   97  ,   159 ,   9   ,   469 ,   881 ,   1935    ,   1871    ,   7943    ,0 };
            const unsigned long dim1298KuoInit[]        =   {   1   ,   1   ,   7   ,   15  ,   31  ,   33  ,   103 ,   247 ,   7   ,   217 ,   397 ,   3761    ,   3071    ,   15357   ,0 };
            const unsigned long dim1299KuoInit[]        =   {   1   ,   3   ,   3   ,   9   ,   23  ,   25  ,   61  ,   61  ,   495 ,   195 ,   1615    ,   923 ,   6583    ,   11169   ,0 };
            const unsigned long dim1300KuoInit[]        =   {   1   ,   1   ,   3   ,   9   ,   25  ,   7   ,   67  ,   105 ,   423 ,   519 ,   327 ,   2789    ,   7527    ,   13465   ,0 };
            const unsigned long dim1301KuoInit[]        =   {   1   ,   1   ,   3   ,   11  ,   29  ,   57  ,   21  ,   47  ,   71  ,   771 ,   1255    ,   1469    ,   737 ,   11533   ,0 };
            const unsigned long dim1302KuoInit[]        =   {   1   ,   1   ,   5   ,   11  ,   23  ,   31  ,   103 ,   145 ,   19  ,   931 ,   963 ,   2627    ,   1419    ,   14451   ,0 };
            const unsigned long dim1303KuoInit[]        =   {   1   ,   3   ,   7   ,   1   ,   15  ,   45  ,   37  ,   57  ,   223 ,   151 ,   979 ,   2569    ,   2487    ,   2877    ,0 };
            const unsigned long dim1304KuoInit[]        =   {   1   ,   1   ,   3   ,   7   ,   13  ,   39  ,   83  ,   127 ,   341 ,   687 ,   1575    ,   3193    ,   2845    ,   725 ,0 };
            const unsigned long dim1305KuoInit[]        =   {   1   ,   1   ,   1   ,   5   ,   23  ,   39  ,   71  ,   205 ,   53  ,   477 ,   1961    ,   3967    ,   7169    ,   5443    ,0 };
            const unsigned long dim1306KuoInit[]        =   {   1   ,   1   ,   5   ,   5   ,   23  ,   17  ,   107 ,   151 ,   291 ,   101 ,   1917    ,   3429    ,   6279    ,   1781    ,0 };
            const unsigned long dim1307KuoInit[]        =   {   1   ,   3   ,   3   ,   11  ,   23  ,   45  ,   11  ,   137 ,   243 ,   257 ,   495 ,   3267    ,   1373    ,   15477   ,0 };
            const unsigned long dim1308KuoInit[]        =   {   1   ,   1   ,   7   ,   5   ,   23  ,   13  ,   37  ,   139 ,   473 ,   215 ,   1281    ,   989 ,   2187    ,   1551    ,0 };
            const unsigned long dim1309KuoInit[]        =   {   1   ,   1   ,   7   ,   7   ,   5   ,   21  ,   113 ,   195 ,   247 ,   493 ,   513 ,   1775    ,   1885    ,   5131    ,0 };
            const unsigned long dim1310KuoInit[]        =   {   1   ,   1   ,   7   ,   15  ,   29  ,   31  ,   9   ,   59  ,   329 ,   11  ,   1739    ,   1657    ,   1745    ,   8969    ,0 };
            const unsigned long dim1311KuoInit[]        =   {   1   ,   1   ,   3   ,   7   ,   19  ,   29  ,   121 ,   171 ,   33  ,   393 ,   1463    ,   1817    ,   393 ,   2509    ,0 };
            const unsigned long dim1312KuoInit[]        =   {   1   ,   1   ,   5   ,   11  ,   23  ,   55  ,   103 ,   169 ,   313 ,   293 ,   1791    ,   2727    ,   2311    ,   14519   ,0 };
            const unsigned long dim1313KuoInit[]        =   {   1   ,   3   ,   7   ,   11  ,   13  ,   21  ,   95  ,   205 ,   87  ,   431 ,   667 ,   367 ,   6879    ,   9843    ,0 };
            const unsigned long dim1314KuoInit[]        =   {   1   ,   3   ,   1   ,   3   ,   3   ,   53  ,   75  ,   43  ,   115 ,   561 ,   795 ,   613 ,   4491    ,   913 ,0 };
            const unsigned long dim1315KuoInit[]        =   {   1   ,   1   ,   5   ,   5   ,   25  ,   37  ,   71  ,   13  ,   499 ,   105 ,   9   ,   3759    ,   4997    ,   14747   ,0 };
            const unsigned long dim1316KuoInit[]        =   {   1   ,   3   ,   3   ,   13  ,   25  ,   63  ,   27  ,   163 ,   359 ,   431 ,   5   ,   3371    ,   6051    ,   1099    ,0 };
            const unsigned long dim1317KuoInit[]        =   {   1   ,   1   ,   3   ,   7   ,   5   ,   45  ,   19  ,   217 ,   423 ,   171 ,   933 ,   3441    ,   7451    ,   11831   ,0 };
            const unsigned long dim1318KuoInit[]        =   {   1   ,   3   ,   7   ,   7   ,   7   ,   53  ,   81  ,   67  ,   465 ,   713 ,   1761    ,   3945    ,   3051    ,   8139    ,0 };
            const unsigned long dim1319KuoInit[]        =   {   1   ,   1   ,   5   ,   7   ,   23  ,   55  ,   85  ,   205 ,   265 ,   99  ,   353 ,   4061    ,   3385    ,   4273    ,0 };
            const unsigned long dim1320KuoInit[]        =   {   1   ,   3   ,   7   ,   7   ,   9   ,   19  ,   7   ,   81  ,   9   ,   373 ,   837 ,   1397    ,   3945    ,   14325   ,0 };
            const unsigned long dim1321KuoInit[]        =   {   1   ,   1   ,   5   ,   7   ,   7   ,   47  ,   61  ,   195 ,   101 ,   931 ,   1667    ,   3277    ,   7701    ,   14697   ,0 };
            const unsigned long dim1322KuoInit[]        =   {   1   ,   3   ,   7   ,   7   ,   11  ,   57  ,   43  ,   55  ,   399 ,   783 ,   1647    ,   269 ,   6055    ,   9935    ,0 };
            const unsigned long dim1323KuoInit[]        =   {   1   ,   1   ,   5   ,   3   ,   5   ,   57  ,   87  ,   193 ,   15  ,   515 ,   1663    ,   495 ,   4725    ,   7221    ,0 };
            const unsigned long dim1324KuoInit[]        =   {   1   ,   1   ,   5   ,   3   ,   23  ,   23  ,   87  ,   231 ,   133 ,   123 ,   1477    ,   1811    ,   7383    ,   16339   ,0 };
            const unsigned long dim1325KuoInit[]        =   {   1   ,   3   ,   5   ,   13  ,   31  ,   57  ,   105 ,   237 ,   375 ,   129 ,   193 ,   1293    ,   2445    ,   13383   ,0 };
            const unsigned long dim1326KuoInit[]        =   {   1   ,   3   ,   7   ,   13  ,   19  ,   51  ,   91  ,   27  ,   309 ,   421 ,   647 ,   3777    ,   6779    ,   2807    ,0 };
            const unsigned long dim1327KuoInit[]        =   {   1   ,   3   ,   5   ,   15  ,   5   ,   43  ,   125 ,   37  ,   269 ,   155 ,   1733    ,   447 ,   7883    ,   6085    ,0 };
            const unsigned long dim1328KuoInit[]        =   {   1   ,   3   ,   7   ,   7   ,   27  ,   63  ,   91  ,   7   ,   117 ,   369 ,   1019    ,   847 ,   6747    ,   5057    ,0 };
            const unsigned long dim1329KuoInit[]        =   {   1   ,   3   ,   5   ,   3   ,   31  ,   19  ,   87  ,   125 ,   259 ,   355 ,   161 ,   1421    ,   2295    ,   9683    ,0 };
            const unsigned long dim1330KuoInit[]        =   {   1   ,   3   ,   3   ,   13  ,   25  ,   37  ,   63  ,   169 ,   477 ,   633 ,   231 ,   2039    ,   3469    ,   2531    ,0 };
            const unsigned long dim1331KuoInit[]        =   {   1   ,   3   ,   1   ,   1   ,   25  ,   1   ,   81  ,   211 ,   237 ,   233 ,   1743    ,   961 ,   645 ,   2605    ,0 };
            const unsigned long dim1332KuoInit[]        =   {   1   ,   3   ,   1   ,   11  ,   3   ,   25  ,   3   ,   111 ,   123 ,   417 ,   1799    ,   1419    ,   2501    ,   5765    ,0 };
            const unsigned long dim1333KuoInit[]        =   {   1   ,   1   ,   3   ,   15  ,   11  ,   17  ,   27  ,   95  ,   295 ,   315 ,   987 ,   481 ,   6627    ,   16171   ,0 };
            const unsigned long dim1334KuoInit[]        =   {   1   ,   3   ,   5   ,   15  ,   15  ,   61  ,   123 ,   5   ,   369 ,   161 ,   897 ,   855 ,   625 ,   9975    ,0 };
            const unsigned long dim1335KuoInit[]        =   {   1   ,   1   ,   5   ,   3   ,   9   ,   45  ,   59  ,   53  ,   43  ,   117 ,   789 ,   21  ,   7799    ,   11061   ,0 };
            const unsigned long dim1336KuoInit[]        =   {   1   ,   1   ,   3   ,   7   ,   3   ,   17  ,   113 ,   159 ,   115 ,   965 ,   1541    ,   1037    ,   4723    ,   10151   ,0 };
            const unsigned long dim1337KuoInit[]        =   {   1   ,   3   ,   1   ,   9   ,   7   ,   15  ,   19  ,   161 ,   123 ,   421 ,   109 ,   433 ,   7387    ,   6047    ,0 };
            const unsigned long dim1338KuoInit[]        =   {   1   ,   3   ,   1   ,   13  ,   19  ,   49  ,   31  ,   215 ,   25  ,   289 ,   1995    ,   1135    ,   3719    ,   8453    ,0 };
            const unsigned long dim1339KuoInit[]        =   {   1   ,   3   ,   3   ,   9   ,   3   ,   33  ,   15  ,   191 ,   83  ,   963 ,   199 ,   2199    ,   3273    ,   16165   ,0 };
            const unsigned long dim1340KuoInit[]        =   {   1   ,   3   ,   3   ,   11  ,   17  ,   63  ,   71  ,   49  ,   427 ,   633 ,   735 ,   2009    ,   5049    ,   13649   ,0 };
            const unsigned long dim1341KuoInit[]        =   {   1   ,   1   ,   3   ,   7   ,   25  ,   41  ,   79  ,   161 ,   293 ,   967 ,   969 ,   3253    ,   269 ,   6971    ,0 };
            const unsigned long dim1342KuoInit[]        =   {   1   ,   1   ,   5   ,   7   ,   25  ,   9   ,   43  ,   237 ,   283 ,   135 ,   669 ,   2193    ,   4193    ,   4787    ,0 };
            const unsigned long dim1343KuoInit[]        =   {   1   ,   3   ,   3   ,   7   ,   5   ,   19  ,   83  ,   89  ,   443 ,   443 ,   1511    ,   2309    ,   5483    ,   7583    ,0 };
            const unsigned long dim1344KuoInit[]        =   {   1   ,   3   ,   3   ,   15  ,   21  ,   29  ,   113 ,   127 ,   405 ,   465 ,   1135    ,   2401    ,   7715    ,   13061   ,0 };
            const unsigned long dim1345KuoInit[]        =   {   1   ,   3   ,   7   ,   7   ,   23  ,   19  ,   17  ,   59  ,   483 ,   519 ,   1439    ,   3839    ,   1691    ,   5   ,0 };
            const unsigned long dim1346KuoInit[]        =   {   1   ,   1   ,   1   ,   13  ,   23  ,   43  ,   115 ,   103 ,   155 ,   769 ,   455 ,   17  ,   7791    ,   3333    ,0 };
            const unsigned long dim1347KuoInit[]        =   {   1   ,   1   ,   7   ,   9   ,   3   ,   25  ,   31  ,   7   ,   125 ,   357 ,   759 ,   1203    ,   2107    ,   5361    ,0 };
            const unsigned long dim1348KuoInit[]        =   {   1   ,   1   ,   5   ,   11  ,   27  ,   31  ,   23  ,   73  ,   77  ,   449 ,   661 ,   2337    ,   941 ,   9421    ,0 };
            const unsigned long dim1349KuoInit[]        =   {   1   ,   1   ,   7   ,   15  ,   27  ,   17  ,   35  ,   15  ,   405 ,   317 ,   1977    ,   2449    ,   237 ,   7741    ,0 };
            const unsigned long dim1350KuoInit[]        =   {   1   ,   1   ,   7   ,   3   ,   1   ,   15  ,   63  ,   251 ,   61  ,   123 ,   855 ,   773 ,   4291    ,   15581   ,0 };
            const unsigned long dim1351KuoInit[]        =   {   1   ,   1   ,   5   ,   9   ,   15  ,   49  ,   101 ,   215 ,   239 ,   31  ,   1193    ,   3839    ,   1913    ,   15471   ,0 };
            const unsigned long dim1352KuoInit[]        =   {   1   ,   3   ,   3   ,   5   ,   29  ,   1   ,   109 ,   211 ,   97  ,   1017    ,   1477    ,   2131    ,   6539    ,   4891    ,0 };
            const unsigned long dim1353KuoInit[]        =   {   1   ,   1   ,   3   ,   11  ,   15  ,   63  ,   67  ,   175 ,   209 ,   105 ,   1427    ,   769 ,   7021    ,   287 ,0 };
            const unsigned long dim1354KuoInit[]        =   {   1   ,   3   ,   3   ,   7   ,   19  ,   41  ,   109 ,   165 ,   307 ,   935 ,   379 ,   395 ,   2533    ,   6291    ,0 };
            const unsigned long dim1355KuoInit[]        =   {   1   ,   3   ,   5   ,   13  ,   17  ,   3   ,   69  ,   79  ,   437 ,   861 ,   1115    ,   1351    ,   4951    ,   1753    ,0 };
            const unsigned long dim1356KuoInit[]        =   {   1   ,   3   ,   1   ,   7   ,   19  ,   15  ,   83  ,   85  ,   431 ,   207 ,   481 ,   711 ,   6715    ,   13439   ,0 };
            const unsigned long dim1357KuoInit[]        =   {   1   ,   1   ,   1   ,   5   ,   3   ,   13  ,   127 ,   27  ,   223 ,   297 ,   985 ,   227 ,   6665    ,   13587   ,0 };
            const unsigned long dim1358KuoInit[]        =   {   1   ,   3   ,   7   ,   7   ,   7   ,   61  ,   113 ,   73  ,   443 ,   397 ,   717 ,   1151    ,   3047    ,   3247    ,0 };
            const unsigned long dim1359KuoInit[]        =   {   1   ,   1   ,   1   ,   9   ,   1   ,   1   ,   103 ,   239 ,   351 ,   305 ,   1345    ,   3543    ,   4321    ,   9147    ,0 };
            const unsigned long dim1360KuoInit[]        =   {   1   ,   1   ,   3   ,   1   ,   11  ,   53  ,   57  ,   85  ,   95  ,   707 ,   1817    ,   249 ,   357 ,   11769   ,0 };
            const unsigned long dim1361KuoInit[]        =   {   1   ,   3   ,   1   ,   15  ,   15  ,   25  ,   7   ,   169 ,   499 ,   717 ,   915 ,   365 ,   3131    ,   12487   ,0 };
            const unsigned long dim1362KuoInit[]        =   {   1   ,   1   ,   3   ,   11  ,   15  ,   47  ,   109 ,   9   ,   171 ,   67  ,   1473    ,   1779    ,   5567    ,   3799    ,0 };
            const unsigned long dim1363KuoInit[]        =   {   1   ,   3   ,   5   ,   15  ,   23  ,   21  ,   31  ,   59  ,   1   ,   505 ,   1157    ,   1209    ,   7661    ,   11903   ,0 };
            const unsigned long dim1364KuoInit[]        =   {   1   ,   3   ,   1   ,   13  ,   11  ,   5   ,   75  ,   53  ,   481 ,   433 ,   915 ,   2881    ,   5299    ,   2835    ,0 };
            const unsigned long dim1365KuoInit[]        =   {   1   ,   3   ,   7   ,   5   ,   23  ,   13  ,   127 ,   229 ,   157 ,   643 ,   847 ,   3613    ,   1999    ,   3807    ,0 };
            const unsigned long dim1366KuoInit[]        =   {   1   ,   3   ,   5   ,   11  ,   13  ,   39  ,   79  ,   11  ,   131 ,   603 ,   1013    ,   1197    ,   7583    ,   8195    ,0 };
            const unsigned long dim1367KuoInit[]        =   {   1   ,   3   ,   3   ,   7   ,   27  ,   29  ,   99  ,   19  ,   503 ,   455 ,   1453    ,   2233    ,   2479    ,   8489    ,0 };
            const unsigned long dim1368KuoInit[]        =   {   1   ,   3   ,   5   ,   1   ,   21  ,   37  ,   27  ,   207 ,   405 ,   87  ,   319 ,   3323    ,   4809    ,   12599   ,0 };
            const unsigned long dim1369KuoInit[]        =   {   1   ,   3   ,   7   ,   15  ,   29  ,   3   ,   121 ,   249 ,   31  ,   601 ,   865 ,   2819    ,   2633    ,   15541   ,0 };
            const unsigned long dim1370KuoInit[]        =   {   1   ,   3   ,   5   ,   1   ,   25  ,   53  ,   121 ,   57  ,   75  ,   327 ,   2047    ,   253 ,   7945    ,   14209   ,0 };
            const unsigned long dim1371KuoInit[]        =   {   1   ,   1   ,   7   ,   11  ,   21  ,   3   ,   13  ,   91  ,   417 ,   825 ,   1381    ,   3653    ,   4187    ,   4685    ,0 };
            const unsigned long dim1372KuoInit[]        =   {   1   ,   3   ,   3   ,   9   ,   11  ,   17  ,   87  ,   75  ,   333 ,   871 ,   1679    ,   2943    ,   4803    ,   8147    ,0 };
            const unsigned long dim1373KuoInit[]        =   {   1   ,   1   ,   1   ,   9   ,   13  ,   1   ,   105 ,   95  ,   347 ,   543 ,   1435    ,   2337    ,   5971    ,   7605    ,0 };
            const unsigned long dim1374KuoInit[]        =   {   1   ,   3   ,   5   ,   15  ,   19  ,   51  ,   35  ,   91  ,   187 ,   833 ,   1921    ,   573 ,   3605    ,   8627    ,0 };
            const unsigned long dim1375KuoInit[]        =   {   1   ,   3   ,   5   ,   7   ,   25  ,   7   ,   115 ,   93  ,   87  ,   13  ,   1323    ,   3821    ,   587 ,   5079    ,0 };
            const unsigned long dim1376KuoInit[]        =   {   1   ,   3   ,   5   ,   3   ,   25  ,   25  ,   89  ,   185 ,   473 ,   909 ,   1245    ,   1589    ,   63  ,   10765   ,0 };
            const unsigned long dim1377KuoInit[]        =   {   1   ,   3   ,   5   ,   5   ,   15  ,   3   ,   9   ,   149 ,   237 ,   107 ,   879 ,   2487    ,   7995    ,   12713   ,0 };
            const unsigned long dim1378KuoInit[]        =   {   1   ,   3   ,   3   ,   3   ,   19  ,   55  ,   43  ,   221 ,   289 ,   129 ,   1123    ,   3411    ,   703 ,   9585    ,0 };
            const unsigned long dim1379KuoInit[]        =   {   1   ,   1   ,   5   ,   3   ,   7   ,   63  ,   13  ,   243 ,   235 ,   253 ,   357 ,   563 ,   519 ,   3471    ,0 };
            const unsigned long dim1380KuoInit[]        =   {   1   ,   1   ,   3   ,   13  ,   9   ,   59  ,   67  ,   61  ,   321 ,   149 ,   365 ,   2645    ,   2309    ,   3303    ,0 };
            const unsigned long dim1381KuoInit[]        =   {   1   ,   1   ,   5   ,   5   ,   13  ,   7   ,   53  ,   175 ,   207 ,   507 ,   1159    ,   3869    ,   8139    ,   13899   ,0 };
            const unsigned long dim1382KuoInit[]        =   {   1   ,   3   ,   5   ,   3   ,   23  ,   9   ,   123 ,   37  ,   325 ,   41  ,   679 ,   513 ,   4379    ,   3493    ,0 };
            const unsigned long dim1383KuoInit[]        =   {   1   ,   1   ,   7   ,   3   ,   29  ,   9   ,   117 ,   203 ,   461 ,   831 ,   1757    ,   2815    ,   1215    ,   1725    ,0 };
            const unsigned long dim1384KuoInit[]        =   {   1   ,   1   ,   7   ,   7   ,   23  ,   45  ,   75  ,   159 ,   395 ,   315 ,   709 ,   887 ,   6935    ,   2307    ,0 };
            const unsigned long dim1385KuoInit[]        =   {   1   ,   1   ,   7   ,   7   ,   5   ,   51  ,   29  ,   141 ,   215 ,   783 ,   1161    ,   3887    ,   8103    ,   7461    ,0 };
            const unsigned long dim1386KuoInit[]        =   {   1   ,   1   ,   7   ,   3   ,   21  ,   5   ,   59  ,   123 ,   283 ,   921 ,   225 ,   453 ,   2125    ,   559 ,0 };
            const unsigned long dim1387KuoInit[]        =   {   1   ,   3   ,   7   ,   13  ,   21  ,   19  ,   121 ,   201 ,   367 ,   215 ,   1727    ,   1743    ,   4875    ,   2819    ,0 };
            const unsigned long dim1388KuoInit[]        =   {   1   ,   3   ,   3   ,   13  ,   15  ,   53  ,   57  ,   35  ,   419 ,   595 ,   1581    ,   2881    ,   3473    ,   7201    ,0 };
            const unsigned long dim1389KuoInit[]        =   {   1   ,   3   ,   5   ,   5   ,   19  ,   17  ,   113 ,   175 ,   195 ,   829 ,   115 ,   491 ,   6439    ,   4085    ,0 };
            const unsigned long dim1390KuoInit[]        =   {   1   ,   1   ,   5   ,   9   ,   7   ,   9   ,   111 ,   23  ,   459 ,   335 ,   1529    ,   1519    ,   2855    ,   11923   ,0 };
            const unsigned long dim1391KuoInit[]        =   {   1   ,   3   ,   7   ,   9   ,   13  ,   41  ,   19  ,   229 ,   73  ,   865 ,   1379    ,   1315    ,   1403    ,   10489   ,0 };
            const unsigned long dim1392KuoInit[]        =   {   1   ,   1   ,   1   ,   7   ,   17  ,   35  ,   71  ,   199 ,   195 ,   89  ,   855 ,   355 ,   5087    ,   13173   ,0 };
            const unsigned long dim1393KuoInit[]        =   {   1   ,   3   ,   5   ,   7   ,   3   ,   45  ,   77  ,   109 ,   105 ,   919 ,   21  ,   2749    ,   5401    ,   11701   ,0 };
            const unsigned long dim1394KuoInit[]        =   {   1   ,   3   ,   1   ,   15  ,   5   ,   9   ,   29  ,   213 ,   489 ,   455 ,   791 ,   215 ,   1011    ,   2501    ,0 };
            const unsigned long dim1395KuoInit[]        =   {   1   ,   3   ,   1   ,   9   ,   17  ,   7   ,   61  ,   99  ,   381 ,   669 ,   1787    ,   1443    ,   6135    ,   3273    ,0 };
            const unsigned long dim1396KuoInit[]        =   {   1   ,   3   ,   1   ,   3   ,   27  ,   47  ,   31  ,   153 ,   365 ,   271 ,   1499    ,   1229    ,   5791    ,   2251    ,0 };
            const unsigned long dim1397KuoInit[]        =   {   1   ,   1   ,   1   ,   3   ,   9   ,   31  ,   109 ,   53  ,   293 ,   135 ,   1329    ,   2795    ,   4335    ,   14817   ,0 };
            const unsigned long dim1398KuoInit[]        =   {   1   ,   3   ,   1   ,   15  ,   19  ,   7   ,   43  ,   223 ,   483 ,   237 ,   803 ,   1713    ,   7969    ,   9681    ,0 };
            const unsigned long dim1399KuoInit[]        =   {   1   ,   1   ,   7   ,   1   ,   23  ,   45  ,   67  ,   87  ,   57  ,   261 ,   955 ,   2917    ,   549 ,   4725    ,0 };
            const unsigned long dim1400KuoInit[]        =   {   1   ,   1   ,   1   ,   7   ,   13  ,   43  ,   63  ,   153 ,   395 ,   313 ,   1219    ,   655 ,   6489    ,   2129    ,0 };
            const unsigned long dim1401KuoInit[]        =   {   1   ,   3   ,   3   ,   3   ,   7   ,   51  ,   35  ,   233 ,   289 ,   715 ,   1227    ,   2041    ,   3639    ,   1465    ,0 };
            const unsigned long dim1402KuoInit[]        =   {   1   ,   3   ,   1   ,   7   ,   3   ,   37  ,   97  ,   149 ,   51  ,   313 ,   1883    ,   1411    ,   3319    ,   2599    ,0 };
            const unsigned long dim1403KuoInit[]        =   {   1   ,   1   ,   7   ,   13  ,   25  ,   13  ,   89  ,   245 ,   251 ,   841 ,   1367    ,   4039    ,   4389    ,   7129    ,0 };
            const unsigned long dim1404KuoInit[]        =   {   1   ,   3   ,   5   ,   15  ,   3   ,   55  ,   113 ,   97  ,   159 ,   621 ,   251 ,   2539    ,   7017    ,   1039    ,0 };
            const unsigned long dim1405KuoInit[]        =   {   1   ,   3   ,   3   ,   9   ,   27  ,   31  ,   53  ,   61  ,   165 ,   753 ,   1733    ,   671 ,   665 ,   4893    ,0 };
            const unsigned long dim1406KuoInit[]        =   {   1   ,   1   ,   7   ,   5   ,   25  ,   11  ,   7   ,   189 ,   457 ,   5   ,   97  ,   2399    ,   3089    ,   4811    ,0 };
            const unsigned long dim1407KuoInit[]        =   {   1   ,   1   ,   3   ,   13  ,   29  ,   19  ,   73  ,   105 ,   187 ,   871 ,   499 ,   443 ,   6807    ,   3753    ,0 };
            const unsigned long dim1408KuoInit[]        =   {   1   ,   3   ,   1   ,   5   ,   9   ,   15  ,   53  ,   83  ,   91  ,   763 ,   1585    ,   1675    ,   5137    ,   10483   ,0 };
            const unsigned long dim1409KuoInit[]        =   {   1   ,   1   ,   7   ,   5   ,   25  ,   51  ,   43  ,   47  ,   151 ,   13  ,   1117    ,   1465    ,   2497    ,   7677    ,0 };
            const unsigned long dim1410KuoInit[]        =   {   1   ,   1   ,   1   ,   3   ,   15  ,   15  ,   43  ,   53  ,   161 ,   689 ,   2027    ,   437 ,   3599    ,   6837    ,0 };
            const unsigned long dim1411KuoInit[]        =   {   1   ,   3   ,   3   ,   9   ,   21  ,   5   ,   81  ,   99  ,   249 ,   785 ,   763 ,   1203    ,   1541    ,   6497    ,0 };
            const unsigned long dim1412KuoInit[]        =   {   1   ,   1   ,   5   ,   7   ,   27  ,   7   ,   121 ,   91  ,   271 ,   461 ,   1531    ,   3255    ,   847 ,   14503   ,0 };
            const unsigned long dim1413KuoInit[]        =   {   1   ,   1   ,   3   ,   15  ,   9   ,   53  ,   125 ,   107 ,   321 ,   579 ,   813 ,   2295    ,   2521    ,   6773    ,0 };
            const unsigned long dim1414KuoInit[]        =   {   1   ,   3   ,   3   ,   1   ,   15  ,   55  ,   71  ,   61  ,   307 ,   917 ,   217 ,   1133    ,   7255    ,   16083   ,0 };
            const unsigned long dim1415KuoInit[]        =   {   1   ,   3   ,   1   ,   1   ,   3   ,   7   ,   57  ,   27  ,   243 ,   989 ,   223 ,   3857    ,   2357    ,   11315   ,0 };
            const unsigned long dim1416KuoInit[]        =   {   1   ,   3   ,   5   ,   5   ,   23  ,   41  ,   55  ,   75  ,   241 ,   829 ,   1865    ,   909 ,   4509    ,   5163    ,0 };
            const unsigned long dim1417KuoInit[]        =   {   1   ,   1   ,   1   ,   13  ,   27  ,   5   ,   117 ,   79  ,   171 ,   135 ,   623 ,   2515    ,   3661    ,   941 ,0 };
            const unsigned long dim1418KuoInit[]        =   {   1   ,   3   ,   5   ,   13  ,   17  ,   29  ,   9   ,   147 ,   73  ,   321 ,   895 ,   3263    ,   6897    ,   8551    ,0 };
            const unsigned long dim1419KuoInit[]        =   {   1   ,   1   ,   5   ,   11  ,   21  ,   45  ,   57  ,   153 ,   83  ,   971 ,   403 ,   617 ,   545 ,   6489    ,0 };
            const unsigned long dim1420KuoInit[]        =   {   1   ,   1   ,   5   ,   9   ,   31  ,   37  ,   19  ,   241 ,   89  ,   583 ,   1365    ,   2649    ,   7011    ,   6183    ,0 };
            const unsigned long dim1421KuoInit[]        =   {   1   ,   3   ,   5   ,   9   ,   3   ,   5   ,   45  ,   237 ,   351 ,   811 ,   445 ,   557 ,   4103    ,   11185   ,0 };
            const unsigned long dim1422KuoInit[]        =   {   1   ,   3   ,   3   ,   13  ,   25  ,   23  ,   21  ,   249 ,   209 ,   775 ,   1397    ,   2501    ,   4535    ,   4957    ,0 };
            const unsigned long dim1423KuoInit[]        =   {   1   ,   1   ,   1   ,   13  ,   7   ,   43  ,   55  ,   179 ,   239 ,   43  ,   225 ,   2827    ,   955 ,   665 ,0 };
            const unsigned long dim1424KuoInit[]        =   {   1   ,   1   ,   3   ,   5   ,   15  ,   35  ,   77  ,   153 ,   429 ,   35  ,   539 ,   3543    ,   1971    ,   2865    ,0 };
            const unsigned long dim1425KuoInit[]        =   {   1   ,   3   ,   7   ,   3   ,   27  ,   35  ,   87  ,   51  ,   281 ,   247 ,   547 ,   2185    ,   2593    ,   1301    ,0 };
            const unsigned long dim1426KuoInit[]        =   {   1   ,   3   ,   5   ,   11  ,   19  ,   11  ,   79  ,   241 ,   467 ,   515 ,   531 ,   3285    ,   1575    ,   7869    ,0 };
            const unsigned long dim1427KuoInit[]        =   {   1   ,   3   ,   1   ,   13  ,   25  ,   15  ,   109 ,   231 ,   349 ,   955 ,   2019    ,   3475    ,   2927    ,   14383   ,0 };
            const unsigned long dim1428KuoInit[]        =   {   1   ,   1   ,   7   ,   1   ,   17  ,   43  ,   43  ,   131 ,   485 ,   179 ,   901 ,   483 ,   4879    ,   15189   ,0 };
            const unsigned long dim1429KuoInit[]        =   {   1   ,   1   ,   5   ,   11  ,   7   ,   39  ,   111 ,   27  ,   29  ,   79  ,   261 ,   77  ,   231 ,   8577    ,0 };
            const unsigned long dim1430KuoInit[]        =   {   1   ,   3   ,   1   ,   11  ,   25  ,   15  ,   25  ,   41  ,   457 ,   677 ,   1033    ,   161 ,   3215    ,   105 ,0 };
            const unsigned long dim1431KuoInit[]        =   {   1   ,   3   ,   1   ,   15  ,   3   ,   57  ,   125 ,   41  ,   325 ,   467 ,   1325    ,   1287    ,   51  ,   9463    ,0 };
            const unsigned long dim1432KuoInit[]        =   {   1   ,   1   ,   1   ,   9   ,   1   ,   33  ,   39  ,   47  ,   313 ,   305 ,   2023    ,   1131    ,   4167    ,   4519    ,0 };
            const unsigned long dim1433KuoInit[]        =   {   1   ,   1   ,   5   ,   15  ,   1   ,   47  ,   123 ,   151 ,   263 ,   143 ,   91  ,   1673    ,   4175    ,   4575    ,0 };
            const unsigned long dim1434KuoInit[]        =   {   1   ,   1   ,   7   ,   3   ,   7   ,   41  ,   95  ,   27  ,   183 ,   437 ,   587 ,   3821    ,   8035    ,   7627    ,0 };
            const unsigned long dim1435KuoInit[]        =   {   1   ,   1   ,   3   ,   5   ,   17  ,   53  ,   95  ,   73  ,   335 ,   497 ,   643 ,   3309    ,   7139    ,   10811   ,0 };
            const unsigned long dim1436KuoInit[]        =   {   1   ,   3   ,   1   ,   13  ,   7   ,   29  ,   83  ,   15  ,   29  ,   881 ,   881 ,   1253    ,   7267    ,   4221    ,0 };
            const unsigned long dim1437KuoInit[]        =   {   1   ,   3   ,   7   ,   13  ,   15  ,   47  ,   71  ,   63  ,   493 ,   365 ,   1209    ,   2503    ,   3963    ,   5069    ,0 };
            const unsigned long dim1438KuoInit[]        =   {   1   ,   1   ,   5   ,   5   ,   23  ,   55  ,   107 ,   255 ,   17  ,   561 ,   1131    ,   497 ,   5731    ,   13287   ,0 };
            const unsigned long dim1439KuoInit[]        =   {   1   ,   3   ,   5   ,   11  ,   5   ,   59  ,   25  ,   131 ,   375 ,   165 ,   1051    ,   3129    ,   6517    ,   6317    ,0 };
            const unsigned long dim1440KuoInit[]        =   {   1   ,   3   ,   5   ,   9   ,   17  ,   29  ,   57  ,   215 ,   77  ,   197 ,   1775    ,   3959    ,   185 ,   11431   ,0 };
            const unsigned long dim1441KuoInit[]        =   {   1   ,   1   ,   1   ,   11  ,   11  ,   45  ,   29  ,   245 ,   193 ,   747 ,   1449    ,   2185    ,   7751    ,   10009   ,0 };
            const unsigned long dim1442KuoInit[]        =   {   1   ,   1   ,   1   ,   5   ,   27  ,   19  ,   39  ,   161 ,   355 ,   455 ,   1625    ,   1283    ,   1547    ,   8707    ,0 };
            const unsigned long dim1443KuoInit[]        =   {   1   ,   1   ,   1   ,   11  ,   7   ,   15  ,   35  ,   193 ,   489 ,   619 ,   515 ,   199 ,   5439    ,   15711   ,0 };
            const unsigned long dim1444KuoInit[]        =   {   1   ,   1   ,   7   ,   1   ,   29  ,   21  ,   69  ,   13  ,   343 ,   581 ,   495 ,   3963    ,   911 ,   3995    ,0 };
            const unsigned long dim1445KuoInit[]        =   {   1   ,   1   ,   1   ,   7   ,   21  ,   5   ,   9   ,   31  ,   291 ,   825 ,   7   ,   2621    ,   1149    ,   2709    ,0 };
            const unsigned long dim1446KuoInit[]        =   {   1   ,   3   ,   3   ,   13  ,   7   ,   35  ,   69  ,   43  ,   397 ,   641 ,   89  ,   3325    ,   5567    ,   3487    ,0 };
            const unsigned long dim1447KuoInit[]        =   {   1   ,   3   ,   5   ,   5   ,   1   ,   17  ,   11  ,   97  ,   309 ,   371 ,   145 ,   3255    ,   5161    ,   7787    ,0 };
            const unsigned long dim1448KuoInit[]        =   {   1   ,   3   ,   7   ,   7   ,   27  ,   1   ,   1   ,   235 ,   47  ,   251 ,   1525    ,   2661    ,   3971    ,   11319   ,0 };
            const unsigned long dim1449KuoInit[]        =   {   1   ,   3   ,   5   ,   3   ,   11  ,   21  ,   69  ,   97  ,   13  ,   931 ,   1955    ,   3497    ,   3963    ,   519 ,0 };
            const unsigned long dim1450KuoInit[]        =   {   1   ,   1   ,   5   ,   9   ,   21  ,   15  ,   11  ,   201 ,   175 ,   855 ,   1867    ,   3941    ,   6505    ,   3451    ,0 };
            const unsigned long dim1451KuoInit[]        =   {   1   ,   1   ,   7   ,   13  ,   25  ,   45  ,   27  ,   91  ,   149 ,   827 ,   1791    ,   139 ,   6721    ,   9   ,0 };
            const unsigned long dim1452KuoInit[]        =   {   1   ,   3   ,   3   ,   7   ,   19  ,   13  ,   95  ,   161 ,   91  ,   333 ,   203 ,   3195    ,   517 ,   3911    ,0 };
            const unsigned long dim1453KuoInit[]        =   {   1   ,   1   ,   5   ,   1   ,   3   ,   7   ,   113 ,   195 ,   501 ,   203 ,   193 ,   3585    ,   3217    ,   2009    ,0 };
            const unsigned long dim1454KuoInit[]        =   {   1   ,   1   ,   5   ,   13  ,   21  ,   29  ,   9   ,   209 ,   381 ,   245 ,   1317    ,   3577    ,   3327    ,   9253    ,0 };
            const unsigned long dim1455KuoInit[]        =   {   1   ,   3   ,   3   ,   15  ,   1   ,   5   ,   75  ,   215 ,   349 ,   143 ,   1607    ,   2817    ,   8185    ,   13991   ,0 };
            const unsigned long dim1456KuoInit[]        =   {   1   ,   3   ,   3   ,   11  ,   23  ,   29  ,   73  ,   15  ,   159 ,   425 ,   1169    ,   2989    ,   5835    ,   12687   ,0 };
            const unsigned long dim1457KuoInit[]        =   {   1   ,   1   ,   1   ,   11  ,   13  ,   13  ,   79  ,   97  ,   351 ,   455 ,   1465    ,   2013    ,   5133    ,   1747    ,0 };
            const unsigned long dim1458KuoInit[]        =   {   1   ,   3   ,   3   ,   13  ,   3   ,   29  ,   53  ,   37  ,   357 ,   55  ,   195 ,   43  ,   3903    ,   7013    ,0 };
            const unsigned long dim1459KuoInit[]        =   {   1   ,   3   ,   5   ,   13  ,   25  ,   43  ,   19  ,   233 ,   17  ,   31  ,   1727    ,   557 ,   2581    ,   847 ,0 };
            const unsigned long dim1460KuoInit[]        =   {   1   ,   1   ,   7   ,   5   ,   21  ,   27  ,   7   ,   27  ,   487 ,   961 ,   1095    ,   3813    ,   7967    ,   15283   ,0 };
            const unsigned long dim1461KuoInit[]        =   {   1   ,   3   ,   7   ,   7   ,   27  ,   49  ,   15  ,   199 ,   413 ,   295 ,   441 ,   2023    ,   2453    ,   2243    ,0 };
            const unsigned long dim1462KuoInit[]        =   {   1   ,   3   ,   7   ,   9   ,   19  ,   13  ,   99  ,   85  ,   227 ,   61  ,   933 ,   3897    ,   2399    ,   455 ,0 };
            const unsigned long dim1463KuoInit[]        =   {   1   ,   3   ,   1   ,   15  ,   13  ,   45  ,   123 ,   13  ,   39  ,   431 ,   1573    ,   2771    ,   2273    ,   12613   ,0 };
            const unsigned long dim1464KuoInit[]        =   {   1   ,   1   ,   3   ,   13  ,   1   ,   29  ,   127 ,   59  ,   159 ,   1005    ,   483 ,   3031    ,   5029    ,   1763    ,0 };
            const unsigned long dim1465KuoInit[]        =   {   1   ,   1   ,   7   ,   9   ,   17  ,   27  ,   87  ,   25  ,   313 ,   305 ,   1611    ,   693 ,   3701    ,   573 ,0 };
            const unsigned long dim1466KuoInit[]        =   {   1   ,   3   ,   7   ,   13  ,   5   ,   61  ,   93  ,   69  ,   357 ,   681 ,   291 ,   1251    ,   3889    ,   3687    ,0 };
            const unsigned long dim1467KuoInit[]        =   {   1   ,   1   ,   3   ,   15  ,   9   ,   31  ,   121 ,   213 ,   115 ,   231 ,   349 ,   2141    ,   4443    ,   12319   ,0 };
            const unsigned long dim1468KuoInit[]        =   {   1   ,   1   ,   1   ,   5   ,   1   ,   37  ,   59  ,   21  ,   399 ,   751 ,   841 ,   33  ,   7195    ,   8403    ,0 };
            const unsigned long dim1469KuoInit[]        =   {   1   ,   3   ,   3   ,   13  ,   5   ,   41  ,   111 ,   21  ,   135 ,   945 ,   1915    ,   2351    ,   6429    ,   10993   ,0 };
            const unsigned long dim1470KuoInit[]        =   {   1   ,   1   ,   1   ,   1   ,   29  ,   43  ,   45  ,   13  ,   395 ,   407 ,   47  ,   1007    ,   7525    ,   6707    ,0 };
            const unsigned long dim1471KuoInit[]        =   {   1   ,   3   ,   5   ,   15  ,   23  ,   3   ,   49  ,   149 ,   391 ,   689 ,   147 ,   1173    ,   3793    ,   4249    ,0 };
            const unsigned long dim1472KuoInit[]        =   {   1   ,   3   ,   1   ,   13  ,   25  ,   35  ,   109 ,   147 ,   323 ,   161 ,   257 ,   2023    ,   5693    ,   16199   ,0 };
            const unsigned long dim1473KuoInit[]        =   {   1   ,   3   ,   3   ,   9   ,   17  ,   3   ,   9   ,   191 ,   19  ,   405 ,   1841    ,   1669    ,   4691    ,   11907   ,0 };
            const unsigned long dim1474KuoInit[]        =   {   1   ,   1   ,   3   ,   5   ,   13  ,   35  ,   45  ,   27  ,   127 ,   329 ,   931 ,   1   ,   2521    ,   15307   ,0 };
            const unsigned long dim1475KuoInit[]        =   {   1   ,   1   ,   5   ,   3   ,   11  ,   25  ,   31  ,   233 ,   1   ,   769 ,   433 ,   3785    ,   3631    ,   8573    ,0 };
            const unsigned long dim1476KuoInit[]        =   {   1   ,   1   ,   3   ,   1   ,   23  ,   57  ,   67  ,   227 ,   393 ,   469 ,   1579    ,   2235    ,   493 ,   2477    ,0 };
            const unsigned long dim1477KuoInit[]        =   {   1   ,   3   ,   7   ,   3   ,   9   ,   19  ,   69  ,   123 ,   205 ,   553 ,   1871    ,   3807    ,   1563    ,   2771    ,0 };
            const unsigned long dim1478KuoInit[]        =   {   1   ,   3   ,   3   ,   11  ,   11  ,   33  ,   55  ,   27  ,   51  ,   903 ,   1681    ,   2613    ,   2519    ,   9529    ,0 };
            const unsigned long dim1479KuoInit[]        =   {   1   ,   1   ,   7   ,   11  ,   29  ,   55  ,   31  ,   185 ,   85  ,   929 ,   233 ,   2619    ,   2989    ,   191 ,0 };
            const unsigned long dim1480KuoInit[]        =   {   1   ,   1   ,   1   ,   3   ,   1   ,   47  ,   47  ,   225 ,   387 ,   717 ,   815 ,   853 ,   819 ,   5645    ,0 };
            const unsigned long dim1481KuoInit[]        =   {   1   ,   1   ,   5   ,   3   ,   1   ,   55  ,   89  ,   59  ,   39  ,   687 ,   1395    ,   535 ,   4977    ,   8155    ,0 };
            const unsigned long dim1482KuoInit[]        =   {   1   ,   3   ,   7   ,   1   ,   11  ,   57  ,   107 ,   137 ,   407 ,   155 ,   313 ,   2205    ,   3095    ,   15179   ,0 };
            const unsigned long dim1483KuoInit[]        =   {   1   ,   1   ,   7   ,   13  ,   29  ,   5   ,   121 ,   173 ,   21  ,   247 ,   1663    ,   1937    ,   5113    ,   7439    ,0 };
            const unsigned long dim1484KuoInit[]        =   {   1   ,   3   ,   1   ,   7   ,   1   ,   39  ,   31  ,   129 ,   109 ,   447 ,   1269    ,   3133    ,   4933    ,   11893   ,0 };
            const unsigned long dim1485KuoInit[]        =   {   1   ,   3   ,   3   ,   15  ,   9   ,   7   ,   21  ,   157 ,   397 ,   465 ,   1057    ,   2355    ,   4663    ,   13955   ,0 };
            const unsigned long dim1486KuoInit[]        =   {   1   ,   1   ,   7   ,   3   ,   1   ,   35  ,   71  ,   175 ,   439 ,   1019    ,   1479    ,   2519    ,   5119    ,   15245   ,0 };
            const unsigned long dim1487KuoInit[]        =   {   1   ,   1   ,   7   ,   3   ,   23  ,   49  ,   55  ,   91  ,   269 ,   407 ,   393 ,   1495    ,   6153    ,   15703   ,0 };
            const unsigned long dim1488KuoInit[]        =   {   1   ,   1   ,   3   ,   7   ,   25  ,   3   ,   45  ,   87  ,   141 ,   693 ,   555 ,   3629    ,   2723    ,   4357    ,0 };
            const unsigned long dim1489KuoInit[]        =   {   1   ,   3   ,   3   ,   7   ,   15  ,   27  ,   103 ,   161 ,   423 ,   785 ,   1171    ,   2335    ,   1845    ,   14373   ,0 };
            const unsigned long dim1490KuoInit[]        =   {   1   ,   3   ,   5   ,   5   ,   1   ,   31  ,   117 ,   35  ,   181 ,   263 ,   1337    ,   3239    ,   617 ,   11499   ,0 };
            const unsigned long dim1491KuoInit[]        =   {   1   ,   1   ,   1   ,   9   ,   15  ,   17  ,   103 ,   201 ,   79  ,   211 ,   2043    ,   3733    ,   7593    ,   6523    ,0 };
            const unsigned long dim1492KuoInit[]        =   {   1   ,   3   ,   5   ,   5   ,   3   ,   59  ,   11  ,   163 ,   155 ,   215 ,   1621    ,   2615    ,   2915    ,   2459    ,0 };
            const unsigned long dim1493KuoInit[]        =   {   1   ,   3   ,   7   ,   11  ,   11  ,   11  ,   93  ,   239 ,   367 ,   553 ,   1341    ,   129 ,   7651    ,   2841    ,0 };
            const unsigned long dim1494KuoInit[]        =   {   1   ,   3   ,   7   ,   7   ,   7   ,   21  ,   45  ,   99  ,   63  ,   601 ,   1877    ,   765 ,   5831    ,   1825    ,0 };
            const unsigned long dim1495KuoInit[]        =   {   1   ,   1   ,   7   ,   3   ,   21  ,   27  ,   71  ,   53  ,   185 ,   675 ,   891 ,   657 ,   5343    ,   3583    ,0 };
            const unsigned long dim1496KuoInit[]        =   {   1   ,   3   ,   5   ,   9   ,   17  ,   57  ,   95  ,   57  ,   427 ,   329 ,   681 ,   2663    ,   3183    ,   7429    ,0 };
            const unsigned long dim1497KuoInit[]        =   {   1   ,   3   ,   7   ,   3   ,   17  ,   9   ,   93  ,   105 ,   409 ,   975 ,   205 ,   429 ,   4473    ,   10269   ,0 };
            const unsigned long dim1498KuoInit[]        =   {   1   ,   3   ,   1   ,   1   ,   11  ,   51  ,   25  ,   97  ,   333 ,   71  ,   1513    ,   1273    ,   1235    ,   5023    ,0 };
            const unsigned long dim1499KuoInit[]        =   {   1   ,   1   ,   7   ,   9   ,   17  ,   59  ,   75  ,   51  ,   277 ,   377 ,   3   ,   3357    ,   2595    ,   7939    ,0 };
            const unsigned long dim1500KuoInit[]        =   {   1   ,   3   ,   3   ,   1   ,   7   ,   47  ,   11  ,   241 ,   235 ,   761 ,   283 ,   3277    ,   2121    ,   69  ,0 };
            const unsigned long dim1501KuoInit[]        =   {   1   ,   3   ,   1   ,   3   ,   5   ,   5   ,   57  ,   5   ,   99  ,   869 ,   707 ,   2379    ,   5439    ,   4981    ,0 };
            const unsigned long dim1502KuoInit[]        =   {   1   ,   3   ,   7   ,   9   ,   5   ,   9   ,   115 ,   155 ,   191 ,   905 ,   1173    ,   975 ,   6825    ,   8257    ,0 };
            const unsigned long dim1503KuoInit[]        =   {   1   ,   1   ,   3   ,   13  ,   17  ,   13  ,   63  ,   143 ,   305 ,   577 ,   1895    ,   1943    ,   2335    ,   8957    ,0 };
            const unsigned long dim1504KuoInit[]        =   {   1   ,   3   ,   7   ,   3   ,   15  ,   19  ,   89  ,   91  ,   497 ,   145 ,   1419    ,   1449    ,   5305    ,   1379    ,0 };
            const unsigned long dim1505KuoInit[]        =   {   1   ,   3   ,   1   ,   13  ,   9   ,   51  ,   103 ,   169 ,   265 ,   347 ,   697 ,   3735    ,   6987    ,   2565    ,0 };
            const unsigned long dim1506KuoInit[]        =   {   1   ,   3   ,   3   ,   5   ,   1   ,   19  ,   83  ,   211 ,   247 ,   81  ,   131 ,   2243    ,   7563    ,   14221   ,0 };
            const unsigned long dim1507KuoInit[]        =   {   1   ,   3   ,   3   ,   11  ,   27  ,   1   ,   97  ,   85  ,   265 ,   423 ,   173 ,   1873    ,   3889    ,   1845    ,0 };
            const unsigned long dim1508KuoInit[]        =   {   1   ,   1   ,   1   ,   11  ,   15  ,   25  ,   25  ,   155 ,   111 ,   327 ,   1089    ,   335 ,   7635    ,   6467    ,0 };
            const unsigned long dim1509KuoInit[]        =   {   1   ,   3   ,   5   ,   7   ,   15  ,   33  ,   103 ,   231 ,   455 ,   681 ,   547 ,   3389    ,   4279    ,   6037    ,0 };
            const unsigned long dim1510KuoInit[]        =   {   1   ,   3   ,   3   ,   7   ,   3   ,   11  ,   11  ,   165 ,   113 ,   499 ,   977 ,   3805    ,   2733    ,   15643   ,0 };
            const unsigned long dim1511KuoInit[]        =   {   1   ,   1   ,   3   ,   7   ,   23  ,   49  ,   39  ,   165 ,   323 ,   313 ,   1731    ,   2977    ,   4795    ,   12387   ,0 };
            const unsigned long dim1512KuoInit[]        =   {   1   ,   1   ,   5   ,   15  ,   27  ,   15  ,   43  ,   149 ,   71  ,   19  ,   1025    ,   2585    ,   5975    ,   12473   ,0 };
            const unsigned long dim1513KuoInit[]        =   {   1   ,   3   ,   7   ,   15  ,   15  ,   21  ,   33  ,   29  ,   351 ,   909 ,   813 ,   3901    ,   3235    ,   3719    ,0 };
            const unsigned long dim1514KuoInit[]        =   {   1   ,   3   ,   3   ,   1   ,   29  ,   63  ,   55  ,   173 ,   405 ,   841 ,   871 ,   213 ,   4747    ,   11611   ,0 };
            const unsigned long dim1515KuoInit[]        =   {   1   ,   1   ,   1   ,   7   ,   15  ,   11  ,   83  ,   21  ,   123 ,   929 ,   1137    ,   3843    ,   5919    ,   151 ,0 };
            const unsigned long dim1516KuoInit[]        =   {   1   ,   1   ,   3   ,   1   ,   29  ,   53  ,   111 ,   133 ,   285 ,   499 ,   2011    ,   121 ,   2853    ,   7587    ,0 };
            const unsigned long dim1517KuoInit[]        =   {   1   ,   3   ,   7   ,   1   ,   15  ,   33  ,   87  ,   139 ,   429 ,   97  ,   375 ,   1273    ,   2745    ,   12045   ,0 };
            const unsigned long dim1518KuoInit[]        =   {   1   ,   1   ,   5   ,   13  ,   17  ,   55  ,   43  ,   3   ,   465 ,   1   ,   1065    ,   647 ,   2497    ,   14775   ,0 };
            const unsigned long dim1519KuoInit[]        =   {   1   ,   1   ,   5   ,   5   ,   15  ,   41  ,   39  ,   37  ,   397 ,   621 ,   1735    ,   1009    ,   1289    ,   2375    ,0 };
            const unsigned long dim1520KuoInit[]        =   {   1   ,   3   ,   5   ,   11  ,   1   ,   45  ,   73  ,   207 ,   375 ,   825 ,   697 ,   3765    ,   7181    ,   4953    ,0 };
            const unsigned long dim1521KuoInit[]        =   {   1   ,   3   ,   7   ,   7   ,   23  ,   15  ,   75  ,   11  ,   137 ,   1017    ,   203 ,   1325    ,   6661    ,   8569    ,0 };
            const unsigned long dim1522KuoInit[]        =   {   1   ,   1   ,   1   ,   15  ,   23  ,   9   ,   15  ,   157 ,   179 ,   699 ,   1493    ,   2161    ,   3159    ,   2709    ,0 };
            const unsigned long dim1523KuoInit[]        =   {   1   ,   3   ,   7   ,   3   ,   15  ,   51  ,   95  ,   101 ,   467 ,   303 ,   1867    ,   3819    ,   7593    ,   6133    ,0 };
            const unsigned long dim1524KuoInit[]        =   {   1   ,   3   ,   7   ,   13  ,   25  ,   27  ,   87  ,   11  ,   113 ,   861 ,   1393    ,   1521    ,   2109    ,   11899   ,0 };
            const unsigned long dim1525KuoInit[]        =   {   1   ,   3   ,   5   ,   11  ,   25  ,   53  ,   27  ,   249 ,   193 ,   783 ,   609 ,   3069    ,   3901    ,   391 ,0 };
            const unsigned long dim1526KuoInit[]        =   {   1   ,   3   ,   5   ,   11  ,   19  ,   13  ,   55  ,   147 ,   325 ,   131 ,   659 ,   1277    ,   6127    ,   1033    ,0 };
            const unsigned long dim1527KuoInit[]        =   {   1   ,   3   ,   1   ,   7   ,   31  ,   13  ,   1   ,   141 ,   377 ,   159 ,   1945    ,   2031    ,   8039    ,   13851   ,0 };
            const unsigned long dim1528KuoInit[]        =   {   1   ,   3   ,   5   ,   7   ,   27  ,   17  ,   79  ,   47  ,   421 ,   517 ,   751 ,   3049    ,   6109    ,   11027   ,0 };
            const unsigned long dim1529KuoInit[]        =   {   1   ,   3   ,   3   ,   11  ,   23  ,   41  ,   89  ,   225 ,   367 ,   109 ,   1295    ,   2691    ,   4677    ,   9207    ,0 };
            const unsigned long dim1530KuoInit[]        =   {   1   ,   1   ,   7   ,   7   ,   7   ,   27  ,   89  ,   221 ,   267 ,   537 ,   2003    ,   3357    ,   8113    ,   8173    ,0 };
            const unsigned long dim1531KuoInit[]        =   {   1   ,   1   ,   7   ,   15  ,   19  ,   23  ,   21  ,   173 ,   159 ,   535 ,   1685    ,   2259    ,   6515    ,   15853   ,0 };
            const unsigned long dim1532KuoInit[]        =   {   1   ,   3   ,   3   ,   15  ,   13  ,   21  ,   105 ,   173 ,   173 ,   13  ,   423 ,   2259    ,   4145    ,   6543    ,0 };
            const unsigned long dim1533KuoInit[]        =   {   1   ,   1   ,   3   ,   11  ,   13  ,   41  ,   35  ,   239 ,   307 ,   893 ,   73  ,   1325    ,   7975    ,   8395    ,0 };
            const unsigned long dim1534KuoInit[]        =   {   1   ,   1   ,   7   ,   11  ,   11  ,   11  ,   115 ,   195 ,   407 ,   543 ,   801 ,   1449    ,   8121    ,   9627    ,0 };
            const unsigned long dim1535KuoInit[]        =   {   1   ,   1   ,   1   ,   1   ,   13  ,   33  ,   41  ,   207 ,   347 ,   49  ,   897 ,   557 ,   5917    ,   13163   ,0 };
            const unsigned long dim1536KuoInit[]        =   {   1   ,   1   ,   7   ,   13  ,   7   ,   9   ,   51  ,   241 ,   47  ,   479 ,   1399    ,   3229    ,   5201    ,   537 ,0 };
            const unsigned long dim1537KuoInit[]        =   {   1   ,   3   ,   7   ,   3   ,   9   ,   63  ,   31  ,   109 ,   381 ,   631 ,   521 ,   139 ,   4229    ,   1639    ,0 };
            const unsigned long dim1538KuoInit[]        =   {   1   ,   3   ,   3   ,   7   ,   1   ,   23  ,   15  ,   99  ,   395 ,   991 ,   627 ,   571 ,   833 ,   6631    ,0 };
            const unsigned long dim1539KuoInit[]        =   {   1   ,   1   ,   7   ,   3   ,   31  ,   9   ,   99  ,   145 ,   237 ,   181 ,   1911    ,   2985    ,   2415    ,   5363    ,0 };
            const unsigned long dim1540KuoInit[]        =   {   1   ,   3   ,   5   ,   3   ,   27  ,   1   ,   111 ,   133 ,   307 ,   211 ,   1119    ,   2289    ,   5333    ,   14747   ,0 };
            const unsigned long dim1541KuoInit[]        =   {   1   ,   3   ,   7   ,   5   ,   9   ,   27  ,   5   ,   101 ,   113 ,   505 ,   931 ,   677 ,   6225    ,   16279   ,0 };
            const unsigned long dim1542KuoInit[]        =   {   1   ,   1   ,   5   ,   9   ,   19  ,   47  ,   57  ,   3   ,   281 ,   207 ,   621 ,   939 ,   839 ,   4089    ,0 };
            const unsigned long dim1543KuoInit[]        =   {   1   ,   1   ,   5   ,   11  ,   25  ,   61  ,   117 ,   139 ,   5   ,   325 ,   745 ,   3423    ,   31  ,   1451    ,0 };
            const unsigned long dim1544KuoInit[]        =   {   1   ,   3   ,   5   ,   5   ,   3   ,   9   ,   121 ,   185 ,   497 ,   203 ,   367 ,   3049    ,   4827    ,   3363    ,0 };
            const unsigned long dim1545KuoInit[]        =   {   1   ,   3   ,   7   ,   5   ,   1   ,   45  ,   91  ,   199 ,   183 ,   577 ,   625 ,   3641    ,   963 ,   9379    ,0 };
            const unsigned long dim1546KuoInit[]        =   {   1   ,   3   ,   5   ,   9   ,   5   ,   53  ,   17  ,   205 ,   331 ,   329 ,   669 ,   2403    ,   3027    ,   297 ,0 };
            const unsigned long dim1547KuoInit[]        =   {   1   ,   3   ,   5   ,   9   ,   19  ,   31  ,   43  ,   249 ,   361 ,   923 ,   635 ,   3461    ,   3901    ,   439 ,0 };
            const unsigned long dim1548KuoInit[]        =   {   1   ,   3   ,   7   ,   15  ,   21  ,   25  ,   115 ,   115 ,   457 ,   607 ,   699 ,   2649    ,   3003    ,   1599    ,0 };
            const unsigned long dim1549KuoInit[]        =   {   1   ,   1   ,   5   ,   15  ,   21  ,   35  ,   61  ,   97  ,   131 ,   521 ,   1045    ,   2401    ,   2885    ,   11795   ,0 };
            const unsigned long dim1550KuoInit[]        =   {   1   ,   1   ,   1   ,   11  ,   19  ,   13  ,   119 ,   117 ,   31  ,   575 ,   427 ,   3689    ,   3801    ,   12267   ,0 };
            const unsigned long dim1551KuoInit[]        =   {   1   ,   1   ,   3   ,   7   ,   27  ,   45  ,   15  ,   235 ,   281 ,   777 ,   1181    ,   2445    ,   7947    ,   2081    ,0 };
            const unsigned long dim1552KuoInit[]        =   {   1   ,   1   ,   5   ,   3   ,   31  ,   57  ,   57  ,   219 ,   135 ,   717 ,   1369    ,   2221    ,   4397    ,   16195   ,0 };
            const unsigned long dim1553KuoInit[]        =   {   1   ,   3   ,   5   ,   13  ,   11  ,   17  ,   7   ,   189 ,   97  ,   7   ,   397 ,   1557    ,   7181    ,   16205   ,0 };
            const unsigned long dim1554KuoInit[]        =   {   1   ,   1   ,   7   ,   15  ,   13  ,   25  ,   79  ,   241 ,   143 ,   959 ,   529 ,   3899    ,   5411    ,   8769    ,0 };
            const unsigned long dim1555KuoInit[]        =   {   1   ,   1   ,   7   ,   1   ,   3   ,   17  ,   109 ,   25  ,   389 ,   463 ,   1675    ,   627 ,   6383    ,   5007    ,0 };
            const unsigned long dim1556KuoInit[]        =   {   1   ,   3   ,   1   ,   13  ,   19  ,   3   ,   13  ,   3   ,   293 ,   755 ,   1827    ,   755 ,   3325    ,   13447   ,0 };
            const unsigned long dim1557KuoInit[]        =   {   1   ,   1   ,   7   ,   15  ,   17  ,   5   ,   87  ,   17  ,   423 ,   901 ,   777 ,   919 ,   2655    ,   1423    ,0 };
            const unsigned long dim1558KuoInit[]        =   {   1   ,   1   ,   3   ,   13  ,   19  ,   41  ,   89  ,   9   ,   475 ,   479 ,   981 ,   1031    ,   7207    ,   3605    ,0 };
            const unsigned long dim1559KuoInit[]        =   {   1   ,   3   ,   5   ,   7   ,   9   ,   41  ,   111 ,   201 ,   389 ,   139 ,   1853    ,   1773    ,   233 ,   3695    ,0 };
            const unsigned long dim1560KuoInit[]        =   {   1   ,   1   ,   3   ,   1   ,   3   ,   47  ,   83  ,   15  ,   49  ,   867 ,   1513    ,   3203    ,   3081    ,   8429    ,0 };
            const unsigned long dim1561KuoInit[]        =   {   1   ,   1   ,   5   ,   11  ,   11  ,   33  ,   67  ,   225 ,   303 ,   825 ,   1711    ,   1521    ,   1569    ,   5233    ,0 };
            const unsigned long dim1562KuoInit[]        =   {   1   ,   3   ,   3   ,   11  ,   27  ,   41  ,   37  ,   247 ,   211 ,   143 ,   855 ,   3755    ,   569 ,   10155   ,0 };
            const unsigned long dim1563KuoInit[]        =   {   1   ,   1   ,   7   ,   3   ,   21  ,   57  ,   7   ,   83  ,   133 ,   791 ,   565 ,   2217    ,   7621    ,   14141   ,0 };
            const unsigned long dim1564KuoInit[]        =   {   1   ,   1   ,   5   ,   13  ,   3   ,   57  ,   7   ,   29  ,   131 ,   935 ,   1555    ,   1103    ,   7395    ,   4079    ,0 };
            const unsigned long dim1565KuoInit[]        =   {   1   ,   1   ,   7   ,   3   ,   13  ,   9   ,   13  ,   71  ,   345 ,   1003    ,   1519    ,   1757    ,   539 ,   9207    ,0 };
            const unsigned long dim1566KuoInit[]        =   {   1   ,   3   ,   1   ,   3   ,   1   ,   11  ,   13  ,   143 ,   365 ,   981 ,   1691    ,   2501    ,   7155    ,   4189    ,0 };
            const unsigned long dim1567KuoInit[]        =   {   1   ,   3   ,   3   ,   1   ,   23  ,   37  ,   33  ,   229 ,   411 ,   329 ,   915 ,   2373    ,   2525    ,   12319   ,0 };
            const unsigned long dim1568KuoInit[]        =   {   1   ,   1   ,   3   ,   3   ,   5   ,   45  ,   77  ,   23  ,   317 ,   311 ,   425 ,   1329    ,   3619    ,   11941   ,0 };
            const unsigned long dim1569KuoInit[]        =   {   1   ,   3   ,   1   ,   1   ,   19  ,   17  ,   39  ,   37  ,   405 ,   275 ,   983 ,   995 ,   4925    ,   1005    ,0 };
            const unsigned long dim1570KuoInit[]        =   {   1   ,   1   ,   1   ,   5   ,   31  ,   1   ,   71  ,   137 ,   481 ,   489 ,   1381    ,   3919    ,   7841    ,   14843   ,0 };
            const unsigned long dim1571KuoInit[]        =   {   1   ,   1   ,   7   ,   15  ,   13  ,   37  ,   11  ,   201 ,   43  ,   311 ,   921 ,   2755    ,   2997    ,   4753    ,0 };
            const unsigned long dim1572KuoInit[]        =   {   1   ,   1   ,   3   ,   7   ,   27  ,   35  ,   87  ,   203 ,   197 ,   27  ,   1839    ,   2847    ,   3261    ,   3139    ,0 };
            const unsigned long dim1573KuoInit[]        =   {   1   ,   1   ,   7   ,   1   ,   25  ,   5   ,   9   ,   9   ,   497 ,   125 ,   1451    ,   1381    ,   1975    ,   10493   ,0 };
            const unsigned long dim1574KuoInit[]        =   {   1   ,   1   ,   1   ,   11  ,   11  ,   25  ,   3   ,   9   ,   39  ,   989 ,   975 ,   1825    ,   7089    ,   10475   ,0 };
            const unsigned long dim1575KuoInit[]        =   {   1   ,   1   ,   5   ,   13  ,   25  ,   33  ,   69  ,   55  ,   179 ,   163 ,   1233    ,   1601    ,   1893    ,   8225    ,0 };
            const unsigned long dim1576KuoInit[]        =   {   1   ,   1   ,   3   ,   15  ,   23  ,   11  ,   117 ,   71  ,   151 ,   437 ,   371 ,   1543    ,   2417    ,   13195   ,0 };
            const unsigned long dim1577KuoInit[]        =   {   1   ,   3   ,   7   ,   9   ,   25  ,   33  ,   63  ,   215 ,   193 ,   415 ,   1771    ,   715 ,   4011    ,   11739   ,0 };
            const unsigned long dim1578KuoInit[]        =   {   1   ,   3   ,   3   ,   13  ,   17  ,   1   ,   105 ,   169 ,   5   ,   259 ,   169 ,   1207    ,   5425    ,   6091    ,0 };
            const unsigned long dim1579KuoInit[]        =   {   1   ,   1   ,   5   ,   11  ,   29  ,   43  ,   25  ,   245 ,   33  ,   995 ,   1797    ,   279 ,   6443    ,   14507   ,0 };
            const unsigned long dim1580KuoInit[]        =   {   1   ,   3   ,   7   ,   7   ,   5   ,   29  ,   117 ,   141 ,   409 ,   367 ,   1223    ,   1499    ,   5651    ,   3337    ,0 };
            const unsigned long dim1581KuoInit[]        =   {   1   ,   1   ,   3   ,   9   ,   15  ,   21  ,   21  ,   27  ,   23  ,   693 ,   1431    ,   2055    ,   2959    ,   1347    ,0 };
            const unsigned long dim1582KuoInit[]        =   {   1   ,   1   ,   1   ,   9   ,   7   ,   27  ,   67  ,   127 ,   63  ,   213 ,   763 ,   1221    ,   2319    ,   5313    ,0 };
            const unsigned long dim1583KuoInit[]        =   {   1   ,   1   ,   3   ,   11  ,   21  ,   33  ,   43  ,   243 ,   113 ,   223 ,   1095    ,   2433    ,   1253    ,   16303   ,0 };
            const unsigned long dim1584KuoInit[]        =   {   1   ,   1   ,   5   ,   15  ,   21  ,   61  ,   39  ,   147 ,   263 ,   943 ,   321 ,   2599    ,   3513    ,   6595    ,0 };
            const unsigned long dim1585KuoInit[]        =   {   1   ,   1   ,   7   ,   3   ,   9   ,   25  ,   19  ,   73  ,   253 ,   889 ,   1699    ,   1709    ,   6451    ,   2939    ,0 };
            const unsigned long dim1586KuoInit[]        =   {   1   ,   3   ,   3   ,   15  ,   29  ,   29  ,   53  ,   133 ,   349 ,   553 ,   1677    ,   3949    ,   1701    ,   1187    ,0 };
            const unsigned long dim1587KuoInit[]        =   {   1   ,   3   ,   7   ,   9   ,   27  ,   5   ,   17  ,   15  ,   185 ,   583 ,   1117    ,   2869    ,   6929    ,   8117    ,0 };
            const unsigned long dim1588KuoInit[]        =   {   1   ,   1   ,   5   ,   1   ,   3   ,   15  ,   117 ,   73  ,   321 ,   245 ,   537 ,   2689    ,   4883    ,   5993    ,0 };
            const unsigned long dim1589KuoInit[]        =   {   1   ,   3   ,   7   ,   15  ,   31  ,   57  ,   103 ,   65  ,   89  ,   323 ,   1341    ,   1951    ,   4999    ,   12663   ,0 };
            const unsigned long dim1590KuoInit[]        =   {   1   ,   3   ,   5   ,   5   ,   7   ,   23  ,   77  ,   133 ,   413 ,   157 ,   1281    ,   3779    ,   2081    ,   12201   ,0 };
            const unsigned long dim1591KuoInit[]        =   {   1   ,   3   ,   7   ,   13  ,   15  ,   49  ,   119 ,   45  ,   27  ,   621 ,   2017    ,   1351    ,   7647    ,   8845    ,0 };
            const unsigned long dim1592KuoInit[]        =   {   1   ,   1   ,   7   ,   15  ,   7   ,   51  ,   77  ,   181 ,   25  ,   769 ,   1159    ,   1505    ,   5219    ,   647 ,0 };
            const unsigned long dim1593KuoInit[]        =   {   1   ,   1   ,   3   ,   5   ,   3   ,   35  ,   69  ,   43  ,   327 ,   383 ,   755 ,   253 ,   1137    ,   16133   ,0 };
            const unsigned long dim1594KuoInit[]        =   {   1   ,   1   ,   7   ,   9   ,   13  ,   63  ,   69  ,   215 ,   357 ,   873 ,   477 ,   1243    ,   5475    ,   5335    ,0 };
            const unsigned long dim1595KuoInit[]        =   {   1   ,   3   ,   1   ,   9   ,   27  ,   13  ,   115 ,   197 ,   131 ,   651 ,   741 ,   2539    ,   265 ,   15757   ,0 };
            const unsigned long dim1596KuoInit[]        =   {   1   ,   1   ,   5   ,   11  ,   13  ,   47  ,   45  ,   235 ,   103 ,   597 ,   523 ,   3129    ,   5113    ,   7031    ,0 };
            const unsigned long dim1597KuoInit[]        =   {   1   ,   1   ,   1   ,   13  ,   31  ,   63  ,   25  ,   75  ,   343 ,   537 ,   83  ,   3165    ,   3247    ,   2589    ,0 };
            const unsigned long dim1598KuoInit[]        =   {   1   ,   3   ,   5   ,   5   ,   11  ,   13  ,   29  ,   79  ,   233 ,   901 ,   627 ,   1691    ,   2659    ,   10251   ,0 };
            const unsigned long dim1599KuoInit[]        =   {   1   ,   3   ,   7   ,   1   ,   11  ,   63  ,   123 ,   63  ,   441 ,   931 ,   573 ,   4083    ,   4483    ,   5685    ,0 };
            const unsigned long dim1600KuoInit[]        =   {   1   ,   3   ,   7   ,   1   ,   23  ,   51  ,   107 ,   113 ,   233 ,   921 ,   1761    ,   207 ,   5083    ,   9165    ,0 };
            const unsigned long dim1601KuoInit[]        =   {   1   ,   3   ,   3   ,   15  ,   13  ,   23  ,   5   ,   151 ,   487 ,   157 ,   23  ,   3967    ,   6241    ,   487 ,0 };
            const unsigned long dim1602KuoInit[]        =   {   1   ,   1   ,   7   ,   1   ,   7   ,   39  ,   95  ,   185 ,   379 ,   225 ,   1685    ,   513 ,   2463    ,   11033   ,0 };
            const unsigned long dim1603KuoInit[]        =   {   1   ,   1   ,   5   ,   11  ,   19  ,   43  ,   55  ,   69  ,   433 ,   933 ,   811 ,   3717    ,   5555    ,   15391   ,0 };
            const unsigned long dim1604KuoInit[]        =   {   1   ,   1   ,   7   ,   7   ,   21  ,   17  ,   19  ,   149 ,   185 ,   475 ,   1817    ,   913 ,   4463    ,   12837   ,0 };
            const unsigned long dim1605KuoInit[]        =   {   1   ,   1   ,   7   ,   5   ,   27  ,   63  ,   5   ,   131 ,   217 ,   271 ,   889 ,   3905    ,   2881    ,   419 ,0 };
            const unsigned long dim1606KuoInit[]        =   {   1   ,   3   ,   7   ,   11  ,   5   ,   55  ,   103 ,   135 ,   59  ,   775 ,   1105    ,   2017    ,   493 ,   12747   ,0 };
            const unsigned long dim1607KuoInit[]        =   {   1   ,   1   ,   3   ,   5   ,   9   ,   25  ,   65  ,   9   ,   61  ,   709 ,   1925    ,   1769    ,   3273    ,   1627    ,0 };
            const unsigned long dim1608KuoInit[]        =   {   1   ,   3   ,   1   ,   11  ,   7   ,   35  ,   37  ,   223 ,   105 ,   55  ,   293 ,   2103    ,   7531    ,   2795    ,0 };
            const unsigned long dim1609KuoInit[]        =   {   1   ,   3   ,   7   ,   13  ,   17  ,   23  ,   35  ,   163 ,   177 ,   797 ,   1539    ,   3843    ,   6761    ,   11817   ,0 };
            const unsigned long dim1610KuoInit[]        =   {   1   ,   1   ,   3   ,   15  ,   11  ,   19  ,   55  ,   161 ,   33  ,   701 ,   601 ,   1055    ,   4949    ,   10741   ,0 };
            const unsigned long dim1611KuoInit[]        =   {   1   ,   1   ,   7   ,   3   ,   11  ,   45  ,   71  ,   199 ,   233 ,   917 ,   1447    ,   3651    ,   6021    ,   4745    ,0 };
            const unsigned long dim1612KuoInit[]        =   {   1   ,   1   ,   3   ,   11  ,   21  ,   17  ,   39  ,   141 ,   17  ,   827 ,   355 ,   3337    ,   6897    ,   9005    ,0 };
            const unsigned long dim1613KuoInit[]        =   {   1   ,   3   ,   3   ,   1   ,   11  ,   61  ,   67  ,   103 ,   377 ,   351 ,   1863    ,   207 ,   7333    ,   4481    ,0 };
            const unsigned long dim1614KuoInit[]        =   {   1   ,   1   ,   1   ,   1   ,   7   ,   33  ,   103 ,   121 ,   361 ,   233 ,   195 ,   985 ,   4743    ,   8715    ,0 };
            const unsigned long dim1615KuoInit[]        =   {   1   ,   3   ,   1   ,   9   ,   17  ,   47  ,   127 ,   129 ,   93  ,   217 ,   1049    ,   689 ,   5595    ,   13625   ,0 };
            const unsigned long dim1616KuoInit[]        =   {   1   ,   1   ,   5   ,   11  ,   13  ,   33  ,   59  ,   29  ,   439 ,   447 ,   21  ,   621 ,   3731    ,   9371    ,0 };
            const unsigned long dim1617KuoInit[]        =   {   1   ,   3   ,   1   ,   7   ,   15  ,   31  ,   41  ,   217 ,   275 ,   97  ,   143 ,   3779    ,   5453    ,   13167   ,0 };
            const unsigned long dim1618KuoInit[]        =   {   1   ,   3   ,   7   ,   1   ,   11  ,   31  ,   7   ,   249 ,   193 ,   907 ,   1275    ,   691 ,   5701    ,   6215    ,0 };
            const unsigned long dim1619KuoInit[]        =   {   1   ,   3   ,   5   ,   1   ,   11  ,   63  ,   7   ,   209 ,   227 ,   959 ,   1313    ,   1107    ,   2281    ,   5567    ,0 };
            const unsigned long dim1620KuoInit[]        =   {   1   ,   3   ,   7   ,   9   ,   25  ,   5   ,   103 ,   193 ,   137 ,   345 ,   1953    ,   369 ,   2201    ,   9435    ,0 };
            const unsigned long dim1621KuoInit[]        =   {   1   ,   1   ,   3   ,   1   ,   5   ,   19  ,   11  ,   191 ,   101 ,   861 ,   1665    ,   1439    ,   6889    ,   2467    ,0 };
            const unsigned long dim1622KuoInit[]        =   {   1   ,   1   ,   7   ,   5   ,   19  ,   47  ,   13  ,   123 ,   363 ,   659 ,   1011    ,   2961    ,   2231    ,   11523   ,0 };
            const unsigned long dim1623KuoInit[]        =   {   1   ,   3   ,   1   ,   1   ,   9   ,   19  ,   7   ,   1   ,   351 ,   461 ,   1323    ,   1125    ,   3063    ,   9589    ,0 };
            const unsigned long dim1624KuoInit[]        =   {   1   ,   1   ,   5   ,   11  ,   1   ,   25  ,   79  ,   171 ,   503 ,   319 ,   729 ,   2133    ,   1979    ,   13769   ,0 };
            const unsigned long dim1625KuoInit[]        =   {   1   ,   3   ,   1   ,   3   ,   17  ,   45  ,   55  ,   109 ,   143 ,   541 ,   65  ,   431 ,   2949    ,   11501   ,0 };
            const unsigned long dim1626KuoInit[]        =   {   1   ,   3   ,   7   ,   7   ,   7   ,   7   ,   85  ,   81  ,   317 ,   245 ,   1683    ,   1009    ,   3853    ,   12597   ,0 };
            const unsigned long dim1627KuoInit[]        =   {   1   ,   1   ,   1   ,   7   ,   9   ,   25  ,   85  ,   19  ,   59  ,   5   ,   1919    ,   577 ,   7723    ,   9365    ,0 };
            const unsigned long dim1628KuoInit[]        =   {   1   ,   1   ,   7   ,   11  ,   13  ,   27  ,   35  ,   163 ,   209 ,   491 ,   1759    ,   383 ,   5697    ,   13131   ,0 };
            const unsigned long dim1629KuoInit[]        =   {   1   ,   1   ,   7   ,   13  ,   19  ,   57  ,   121 ,   61  ,   143 ,   735 ,   721 ,   3405    ,   6939    ,   4691    ,0 };
            const unsigned long dim1630KuoInit[]        =   {   1   ,   1   ,   7   ,   15  ,   21  ,   23  ,   55  ,   203 ,   145 ,   673 ,   419 ,   3023    ,   6439    ,   6361    ,0 };
            const unsigned long dim1631KuoInit[]        =   {   1   ,   1   ,   5   ,   15  ,   5   ,   21  ,   11  ,   153 ,   223 ,   229 ,   721 ,   715 ,   7981    ,   15459   ,0 };
            const unsigned long dim1632KuoInit[]        =   {   1   ,   3   ,   5   ,   7   ,   21  ,   59  ,   47  ,   15  ,   213 ,   549 ,   465 ,   1353    ,   2611    ,   11989   ,0 };
            const unsigned long dim1633KuoInit[]        =   {   1   ,   3   ,   5   ,   7   ,   13  ,   51  ,   65  ,   11  ,   295 ,   63  ,   1887    ,   3347    ,   3061    ,   2219    ,0 };
            const unsigned long dim1634KuoInit[]        =   {   1   ,   1   ,   5   ,   3   ,   29  ,   29  ,   83  ,   175 ,   507 ,   341 ,   2031    ,   3737    ,   1741    ,   7615    ,0 };
            const unsigned long dim1635KuoInit[]        =   {   1   ,   3   ,   7   ,   13  ,   11  ,   55  ,   59  ,   213 ,   197 ,   851 ,   1723    ,   2805    ,   2461    ,   8071    ,0 };
            const unsigned long dim1636KuoInit[]        =   {   1   ,   1   ,   3   ,   15  ,   5   ,   37  ,   55  ,   237 ,   319 ,   345 ,   1677    ,   1791    ,   5063    ,   4301    ,0 };
            const unsigned long dim1637KuoInit[]        =   {   1   ,   3   ,   1   ,   11  ,   25  ,   13  ,   87  ,   85  ,   293 ,   97  ,   557 ,   3583    ,   7077    ,   6433    ,0 };
            const unsigned long dim1638KuoInit[]        =   {   1   ,   3   ,   3   ,   1   ,   31  ,   9   ,   63  ,   29  ,   193 ,   351 ,   1071    ,   3587    ,   4695    ,   1771    ,0 };
            const unsigned long dim1639KuoInit[]        =   {   1   ,   3   ,   7   ,   11  ,   11  ,   3   ,   45  ,   119 ,   215 ,   367 ,   677 ,   3405    ,   7177    ,   14607   ,0 };
            const unsigned long dim1640KuoInit[]        =   {   1   ,   3   ,   7   ,   3   ,   3   ,   23  ,   33  ,   33  ,   223 ,   209 ,   1033    ,   3167    ,   4189    ,   2195    ,0 };
            const unsigned long dim1641KuoInit[]        =   {   1   ,   1   ,   5   ,   1   ,   3   ,   53  ,   21  ,   151 ,   475 ,   987 ,   1671    ,   393 ,   7783    ,   2561    ,0 };
            const unsigned long dim1642KuoInit[]        =   {   1   ,   3   ,   7   ,   13  ,   15  ,   5   ,   21  ,   247 ,   225 ,   861 ,   1707    ,   3237    ,   5793    ,   14557   ,0 };
            const unsigned long dim1643KuoInit[]        =   {   1   ,   3   ,   5   ,   9   ,   13  ,   3   ,   37  ,   49  ,   119 ,   33  ,   805 ,   681 ,   7591    ,   1707    ,0 };
            const unsigned long dim1644KuoInit[]        =   {   1   ,   1   ,   3   ,   15  ,   27  ,   55  ,   99  ,   17  ,   361 ,   773 ,   1415    ,   3547    ,   5427    ,   4579    ,0 };
            const unsigned long dim1645KuoInit[]        =   {   1   ,   3   ,   5   ,   9   ,   17  ,   55  ,   89  ,   111 ,   189 ,   719 ,   1691    ,   1565    ,   941 ,   15481   ,0 };
            const unsigned long dim1646KuoInit[]        =   {   1   ,   1   ,   1   ,   11  ,   17  ,   3   ,   77  ,   57  ,   71  ,   739 ,   869 ,   165 ,   6411    ,   1565    ,0 };
            const unsigned long dim1647KuoInit[]        =   {   1   ,   3   ,   3   ,   11  ,   31  ,   47  ,   87  ,   105 ,   241 ,   999 ,   245 ,   3695    ,   5151    ,   9087    ,0 };
            const unsigned long dim1648KuoInit[]        =   {   1   ,   3   ,   1   ,   15  ,   27  ,   59  ,   15  ,   45  ,   413 ,   291 ,   87  ,   1735    ,   779 ,   5977    ,0 };
            const unsigned long dim1649KuoInit[]        =   {   1   ,   3   ,   7   ,   1   ,   9   ,   19  ,   7   ,   183 ,   133 ,   833 ,   1347    ,   3559    ,   7745    ,   11071   ,0 };
            const unsigned long dim1650KuoInit[]        =   {   1   ,   1   ,   3   ,   13  ,   23  ,   59  ,   13  ,   227 ,   407 ,   661 ,   217 ,   299 ,   6033    ,   14519   ,0 };
            const unsigned long dim1651KuoInit[]        =   {   1   ,   1   ,   5   ,   5   ,   27  ,   33  ,   105 ,   15  ,   141 ,   555 ,   397 ,   3287    ,   5859    ,   9553    ,0 };
            const unsigned long dim1652KuoInit[]        =   {   1   ,   3   ,   3   ,   9   ,   27  ,   59  ,   11  ,   47  ,   379 ,   275 ,   889 ,   3069    ,   4135    ,   15735   ,0 };
            const unsigned long dim1653KuoInit[]        =   {   1   ,   1   ,   5   ,   13  ,   27  ,   47  ,   107 ,   165 ,   487 ,   463 ,   425 ,   3949    ,   3213    ,   4161    ,0 };
            const unsigned long dim1654KuoInit[]        =   {   1   ,   3   ,   5   ,   7   ,   27  ,   55  ,   17  ,   3   ,   159 ,   181 ,   393 ,   2441    ,   2531    ,   895 ,0 };
            const unsigned long dim1655KuoInit[]        =   {   1   ,   3   ,   7   ,   7   ,   31  ,   3   ,   95  ,   81  ,   181 ,   367 ,   1689    ,   2351    ,   193 ,   7165    ,0 };
            const unsigned long dim1656KuoInit[]        =   {   1   ,   3   ,   7   ,   15  ,   1   ,   17  ,   103 ,   181 ,   29  ,   851 ,   15  ,   1447    ,   435 ,   14059   ,0 };
            const unsigned long dim1657KuoInit[]        =   {   1   ,   1   ,   1   ,   15  ,   21  ,   31  ,   83  ,   41  ,   237 ,   437 ,   509 ,   3757    ,   5609    ,   15859   ,0 };
            const unsigned long dim1658KuoInit[]        =   {   1   ,   1   ,   7   ,   15  ,   5   ,   35  ,   101 ,   37  ,   245 ,   487 ,   449 ,   2089    ,   27  ,   13667   ,0 };
            const unsigned long dim1659KuoInit[]        =   {   1   ,   1   ,   7   ,   15  ,   29  ,   49  ,   89  ,   161 ,   271 ,   455 ,   1599    ,   3843    ,   1673    ,   3683    ,0 };
            const unsigned long dim1660KuoInit[]        =   {   1   ,   1   ,   7   ,   1   ,   1   ,   25  ,   37  ,   211 ,   369 ,   207 ,   243 ,   2113    ,   2707    ,   7951    ,0 };
            const unsigned long dim1661KuoInit[]        =   {   1   ,   1   ,   7   ,   5   ,   19  ,   21  ,   25  ,   161 ,   97  ,   505 ,   273 ,   1443    ,   4461    ,   1407    ,0 };
            const unsigned long dim1662KuoInit[]        =   {   1   ,   1   ,   5   ,   15  ,   17  ,   57  ,   115 ,   175 ,   511 ,   389 ,   1333    ,   2459    ,   5561    ,   4409    ,0 };
            const unsigned long dim1663KuoInit[]        =   {   1   ,   1   ,   5   ,   9   ,   29  ,   45  ,   45  ,   121 ,   159 ,   51  ,   865 ,   531 ,   1387    ,   3517    ,0 };
            const unsigned long dim1664KuoInit[]        =   {   1   ,   3   ,   5   ,   15  ,   25  ,   17  ,   59  ,   161 ,   193 ,   495 ,   611 ,   1009    ,   5035    ,   15325   ,0 };
            const unsigned long dim1665KuoInit[]        =   {   1   ,   1   ,   1   ,   7   ,   7   ,   53  ,   99  ,   241 ,   305 ,   989 ,   175 ,   2841    ,   2891    ,   4975    ,0 };
            const unsigned long dim1666KuoInit[]        =   {   1   ,   1   ,   7   ,   1   ,   31  ,   45  ,   85  ,   251 ,   339 ,   731 ,   513 ,   2557    ,   3119    ,   15961   ,0 };
            const unsigned long dim1667KuoInit[]        =   {   1   ,   3   ,   7   ,   15  ,   11  ,   37  ,   73  ,   93  ,   83  ,   931 ,   1561    ,   791 ,   3263    ,   10789   ,0 };
            const unsigned long dim1668KuoInit[]        =   {   1   ,   3   ,   1   ,   11  ,   27  ,   25  ,   29  ,   71  ,   199 ,   673 ,   173 ,   2267    ,   1277    ,   4373    ,0 };
            const unsigned long dim1669KuoInit[]        =   {   1   ,   1   ,   5   ,   5   ,   7   ,   19  ,   79  ,   143 ,   63  ,   139 ,   977 ,   845 ,   5125    ,   5635    ,0 };
            const unsigned long dim1670KuoInit[]        =   {   1   ,   1   ,   3   ,   15  ,   3   ,   27  ,   117 ,   219 ,   393 ,   457 ,   1001    ,   1235    ,   6409    ,   10883   ,0 };
            const unsigned long dim1671KuoInit[]        =   {   1   ,   1   ,   3   ,   3   ,   19  ,   31  ,   61  ,   225 ,   391 ,   719 ,   1919    ,   1843    ,   4883    ,   10499   ,0 };
            const unsigned long dim1672KuoInit[]        =   {   1   ,   3   ,   1   ,   3   ,   13  ,   63  ,   73  ,   103 ,   15  ,   777 ,   1019    ,   2677    ,   7317    ,   4349    ,0 };
            const unsigned long dim1673KuoInit[]        =   {   1   ,   3   ,   5   ,   15  ,   5   ,   7   ,   31  ,   33  ,   237 ,   1021    ,   2045    ,   2283    ,   7045    ,   13829   ,0 };
            const unsigned long dim1674KuoInit[]        =   {   1   ,   1   ,   3   ,   13  ,   11  ,   43  ,   47  ,   239 ,   5   ,   293 ,   2015    ,   3613    ,   1551    ,   3203    ,0 };
            const unsigned long dim1675KuoInit[]        =   {   1   ,   3   ,   1   ,   11  ,   11  ,   11  ,   67  ,   201 ,   287 ,   161 ,   193 ,   1815    ,   1493    ,   2863    ,0 };
            const unsigned long dim1676KuoInit[]        =   {   1   ,   3   ,   5   ,   13  ,   11  ,   31  ,   119 ,   57  ,   305 ,   1009    ,   915 ,   2855    ,   3595    ,   11653   ,0 };
            const unsigned long dim1677KuoInit[]        =   {   1   ,   1   ,   1   ,   9   ,   17  ,   7   ,   59  ,   27  ,   189 ,   227 ,   763 ,   2843    ,   7335    ,   13559   ,0 };
            const unsigned long dim1678KuoInit[]        =   {   1   ,   1   ,   5   ,   13  ,   3   ,   1   ,   17  ,   129 ,   83  ,   763 ,   443 ,   895 ,   1573    ,   7343    ,0 };
            const unsigned long dim1679KuoInit[]        =   {   1   ,   3   ,   7   ,   3   ,   5   ,   43  ,   31  ,   95  ,   503 ,   815 ,   1725    ,   2319    ,   2323    ,   12543   ,0 };
            const unsigned long dim1680KuoInit[]        =   {   1   ,   1   ,   1   ,   9   ,   29  ,   27  ,   103 ,   43  ,   39  ,   289 ,   277 ,   453 ,   4297    ,   1009    ,0 };
            const unsigned long dim1681KuoInit[]        =   {   1   ,   1   ,   7   ,   5   ,   31  ,   61  ,   105 ,   79  ,   259 ,   141 ,   49  ,   925 ,   2793    ,   13247   ,0 };
            const unsigned long dim1682KuoInit[]        =   {   1   ,   1   ,   1   ,   5   ,   15  ,   23  ,   59  ,   55  ,   497 ,   561 ,   343 ,   3633    ,   2097    ,   2035    ,0 };
            const unsigned long dim1683KuoInit[]        =   {   1   ,   1   ,   3   ,   13  ,   17  ,   23  ,   13  ,   23  ,   15  ,   519 ,   105 ,   3719    ,   3639    ,   6387    ,0 };
            const unsigned long dim1684KuoInit[]        =   {   1   ,   1   ,   1   ,   11  ,   1   ,   43  ,   97  ,   73  ,   295 ,   841 ,   1033    ,   11  ,   3027    ,   4073    ,0 };
            const unsigned long dim1685KuoInit[]        =   {   1   ,   3   ,   1   ,   3   ,   21  ,   15  ,   67  ,   95  ,   425 ,   979 ,   1641    ,   3059    ,   4195    ,   14291   ,0 };
            const unsigned long dim1686KuoInit[]        =   {   1   ,   3   ,   5   ,   3   ,   9   ,   23  ,   83  ,   203 ,   479 ,   999 ,   1469    ,   2227    ,   7995    ,   4053    ,0 };
            const unsigned long dim1687KuoInit[]        =   {   1   ,   1   ,   5   ,   7   ,   1   ,   13  ,   17  ,   191 ,   297 ,   397 ,   329 ,   1551    ,   6885    ,   2413    ,0 };
            const unsigned long dim1688KuoInit[]        =   {   1   ,   1   ,   7   ,   13  ,   31  ,   21  ,   123 ,   59  ,   79  ,   647 ,   475 ,   347 ,   667 ,   9871    ,0 };
            const unsigned long dim1689KuoInit[]        =   {   1   ,   1   ,   7   ,   15  ,   25  ,   43  ,   35  ,   117 ,   323 ,   805 ,   1543    ,   2631    ,   2919    ,   1377    ,0 };
            const unsigned long dim1690KuoInit[]        =   {   1   ,   1   ,   3   ,   11  ,   17  ,   55  ,   45  ,   213 ,   507 ,   641 ,   1947    ,   2439    ,   4749    ,   6865    ,0 };
            const unsigned long dim1691KuoInit[]        =   {   1   ,   3   ,   7   ,   5   ,   13  ,   31  ,   51  ,   31  ,   167 ,   431 ,   1999    ,   1041    ,   3613    ,   16289   ,0 };
            const unsigned long dim1692KuoInit[]        =   {   1   ,   3   ,   7   ,   9   ,   1   ,   15  ,   19  ,   191 ,   181 ,   247 ,   11  ,   2305    ,   5451    ,   1719    ,0 };
            const unsigned long dim1693KuoInit[]        =   {   1   ,   3   ,   1   ,   11  ,   25  ,   39  ,   13  ,   123 ,   93  ,   527 ,   1359    ,   1985    ,   6789    ,   2045    ,0 };
            const unsigned long dim1694KuoInit[]        =   {   1   ,   1   ,   1   ,   7   ,   25  ,   57  ,   113 ,   91  ,   311 ,   661 ,   1461    ,   2801    ,   5281    ,   13423   ,0 };
            const unsigned long dim1695KuoInit[]        =   {   1   ,   1   ,   1   ,   11  ,   31  ,   9   ,   119 ,   55  ,   91  ,   965 ,   1973    ,   2901    ,   1147    ,   9341    ,0 };
            const unsigned long dim1696KuoInit[]        =   {   1   ,   1   ,   7   ,   15  ,   9   ,   59  ,   83  ,   21  ,   49  ,   921 ,   749 ,   1249    ,   3577    ,   14341   ,0 };
            const unsigned long dim1697KuoInit[]        =   {   1   ,   3   ,   5   ,   13  ,   25  ,   63  ,   65  ,   135 ,   337 ,   743 ,   1669    ,   2519    ,   341 ,   16285   ,0 };
            const unsigned long dim1698KuoInit[]        =   {   1   ,   3   ,   3   ,   7   ,   23  ,   9   ,   97  ,   205 ,   87  ,   289 ,   1195    ,   2843    ,   7419    ,   4183    ,0 };
            const unsigned long dim1699KuoInit[]        =   {   1   ,   3   ,   1   ,   9   ,   3   ,   33  ,   117 ,   243 ,   161 ,   977 ,   109 ,   3333    ,   4639    ,   2299    ,0 };
            const unsigned long dim1700KuoInit[]        =   {   1   ,   1   ,   3   ,   7   ,   1   ,   53  ,   79  ,   73  ,   71  ,   953 ,   349 ,   2517    ,   6281    ,   2625    ,0 };
            const unsigned long dim1701KuoInit[]        =   {   1   ,   1   ,   5   ,   9   ,   21  ,   59  ,   31  ,   253 ,   265 ,   403 ,   603 ,   2185    ,   7277    ,   6057    ,0 };
            const unsigned long dim1702KuoInit[]        =   {   1   ,   1   ,   1   ,   13  ,   27  ,   35  ,   63  ,   109 ,   483 ,   205 ,   1245    ,   2599    ,   1697    ,   11381   ,0 };
            const unsigned long dim1703KuoInit[]        =   {   1   ,   1   ,   7   ,   15  ,   25  ,   29  ,   107 ,   205 ,   411 ,   627 ,   745 ,   1585    ,   1175    ,   3077    ,0 };
            const unsigned long dim1704KuoInit[]        =   {   1   ,   3   ,   5   ,   3   ,   25  ,   15  ,   121 ,   97  ,   283 ,   809 ,   753 ,   2577    ,   1099    ,   12155   ,0 };
            const unsigned long dim1705KuoInit[]        =   {   1   ,   1   ,   7   ,   15  ,   29  ,   21  ,   55  ,   53  ,   101 ,   69  ,   657 ,   2727    ,   931 ,   4873    ,0 };
            const unsigned long dim1706KuoInit[]        =   {   1   ,   3   ,   7   ,   15  ,   5   ,   57  ,   9   ,   11  ,   269 ,   261 ,   1681    ,   1025    ,   4001    ,   16111   ,0 };
            const unsigned long dim1707KuoInit[]        =   {   1   ,   1   ,   3   ,   7   ,   9   ,   51  ,   29  ,   77  ,   313 ,   199 ,   1305    ,   2665    ,   5971    ,   7525    ,0 };
            const unsigned long dim1708KuoInit[]        =   {   1   ,   1   ,   3   ,   15  ,   3   ,   29  ,   99  ,   255 ,   331 ,   925 ,   1973    ,   3825    ,   3549    ,   8537    ,0 };
            const unsigned long dim1709KuoInit[]        =   {   1   ,   1   ,   3   ,   1   ,   27  ,   33  ,   19  ,   225 ,   209 ,   895 ,   47  ,   4063    ,   1873    ,   15631   ,0 };
            const unsigned long dim1710KuoInit[]        =   {   1   ,   3   ,   1   ,   9   ,   11  ,   57  ,   121 ,   205 ,   89  ,   565 ,   1387    ,   1119    ,   7375    ,   6153    ,0 };
            const unsigned long dim1711KuoInit[]        =   {   1   ,   1   ,   3   ,   15  ,   1   ,   27  ,   9   ,   177 ,   443 ,   437 ,   1487    ,   2119    ,   5299    ,   16251   ,0 };
            const unsigned long dim1712KuoInit[]        =   {   1   ,   3   ,   7   ,   5   ,   17  ,   33  ,   79  ,   153 ,   327 ,   719 ,   1587    ,   407 ,   1137    ,   9607    ,0 };
            const unsigned long dim1713KuoInit[]        =   {   1   ,   1   ,   3   ,   15  ,   7   ,   31  ,   109 ,   251 ,   501 ,   553 ,   785 ,   3047    ,   8141    ,   6305    ,0 };
            const unsigned long dim1714KuoInit[]        =   {   1   ,   1   ,   3   ,   9   ,   11  ,   63  ,   65  ,   135 ,   257 ,   313 ,   191 ,   1073    ,   6821    ,   10629   ,0 };
            const unsigned long dim1715KuoInit[]        =   {   1   ,   1   ,   7   ,   1   ,   19  ,   55  ,   7   ,   69  ,   503 ,   131 ,   1133    ,   3325    ,   347 ,   9255    ,0 };
            const unsigned long dim1716KuoInit[]        =   {   1   ,   3   ,   5   ,   3   ,   11  ,   55  ,   115 ,   171 ,   261 ,   581 ,   1169    ,   3609    ,   307 ,   14055   ,0 };
            const unsigned long dim1717KuoInit[]        =   {   1   ,   3   ,   7   ,   15  ,   29  ,   3   ,   33  ,   229 ,   431 ,   969 ,   281 ,   1617    ,   3685    ,   13583   ,0 };
            const unsigned long dim1718KuoInit[]        =   {   1   ,   1   ,   7   ,   11  ,   9   ,   23  ,   5   ,   147 ,   305 ,   651 ,   1419    ,   3225    ,   7029    ,   15331   ,0 };
            const unsigned long dim1719KuoInit[]        =   {   1   ,   3   ,   7   ,   15  ,   23  ,   61  ,   111 ,   221 ,   145 ,   809 ,   1127    ,   2287    ,   6023    ,   6103    ,0 };
            const unsigned long dim1720KuoInit[]        =   {   1   ,   3   ,   1   ,   7   ,   1   ,   49  ,   93  ,   251 ,   239 ,   617 ,   401 ,   3419    ,   8123    ,   4855    ,0 };
            const unsigned long dim1721KuoInit[]        =   {   1   ,   3   ,   5   ,   5   ,   15  ,   9   ,   69  ,   5   ,   509 ,   83  ,   405 ,   3451    ,   4401    ,   3321    ,0 };
            const unsigned long dim1722KuoInit[]        =   {   1   ,   1   ,   3   ,   9   ,   29  ,   39  ,   53  ,   155 ,   73  ,   207 ,   499 ,   3179    ,   6099    ,   1377    ,0 };
            const unsigned long dim1723KuoInit[]        =   {   1   ,   1   ,   1   ,   11  ,   9   ,   43  ,   101 ,   211 ,   439 ,   865 ,   621 ,   3271    ,   4699    ,   5329    ,0 };
            const unsigned long dim1724KuoInit[]        =   {   1   ,   3   ,   5   ,   9   ,   21  ,   59  ,   105 ,   89  ,   5   ,   493 ,   227 ,   1177    ,   6829    ,   8659    ,0 };
            const unsigned long dim1725KuoInit[]        =   {   1   ,   1   ,   1   ,   3   ,   27  ,   19  ,   49  ,   209 ,   337 ,   961 ,   93  ,   81  ,   4607    ,   13579   ,0 };
            const unsigned long dim1726KuoInit[]        =   {   1   ,   3   ,   7   ,   1   ,   31  ,   53  ,   77  ,   213 ,   3   ,   555 ,   1409    ,   3917    ,   23  ,   7837    ,0 };
            const unsigned long dim1727KuoInit[]        =   {   1   ,   3   ,   1   ,   1   ,   21  ,   37  ,   121 ,   17  ,   295 ,   367 ,   1105    ,   3053    ,   1453    ,   8869    ,0 };
            const unsigned long dim1728KuoInit[]        =   {   1   ,   3   ,   5   ,   5   ,   31  ,   23  ,   13  ,   145 ,   459 ,   609 ,   77  ,   197 ,   4457    ,   5145    ,0 };
            const unsigned long dim1729KuoInit[]        =   {   1   ,   1   ,   3   ,   7   ,   27  ,   23  ,   83  ,   165 ,   25  ,   491 ,   1487    ,   2335    ,   7019    ,   499 ,0 };
            const unsigned long dim1730KuoInit[]        =   {   1   ,   1   ,   7   ,   15  ,   29  ,   53  ,   87  ,   47  ,   21  ,   233 ,   253 ,   2529    ,   1651    ,   16091   ,0 };
            const unsigned long dim1731KuoInit[]        =   {   1   ,   3   ,   5   ,   11  ,   7   ,   55  ,   99  ,   193 ,   395 ,   63  ,   817 ,   2109    ,   1167    ,   5739    ,0 };
            const unsigned long dim1732KuoInit[]        =   {   1   ,   1   ,   7   ,   15  ,   13  ,   21  ,   61  ,   31  ,   45  ,   723 ,   103 ,   1347    ,   4259    ,   13185   ,0 };
            const unsigned long dim1733KuoInit[]        =   {   1   ,   1   ,   5   ,   9   ,   7   ,   51  ,   43  ,   21  ,   411 ,   887 ,   1303    ,   3389    ,   6653    ,   9265    ,0 };
            const unsigned long dim1734KuoInit[]        =   {   1   ,   1   ,   3   ,   7   ,   3   ,   7   ,   43  ,   197 ,   151 ,   977 ,   717 ,   2293    ,   3135    ,   5223    ,0 };
            const unsigned long dim1735KuoInit[]        =   {   1   ,   1   ,   1   ,   1   ,   23  ,   49  ,   23  ,   105 ,   279 ,   401 ,   1309    ,   3485    ,   4305    ,   13181   ,0 };
            const unsigned long dim1736KuoInit[]        =   {   1   ,   3   ,   1   ,   15  ,   23  ,   15  ,   89  ,   171 ,   237 ,   513 ,   369 ,   3731    ,   4753    ,   4655    ,0 };
            const unsigned long dim1737KuoInit[]        =   {   1   ,   1   ,   7   ,   11  ,   3   ,   35  ,   49  ,   211 ,   353 ,   843 ,   705 ,   2237    ,   7843    ,   9009    ,0 };
            const unsigned long dim1738KuoInit[]        =   {   1   ,   3   ,   3   ,   9   ,   29  ,   25  ,   63  ,   29  ,   369 ,   777 ,   477 ,   547 ,   6575    ,   13769   ,0 };
            const unsigned long dim1739KuoInit[]        =   {   1   ,   1   ,   3   ,   1   ,   3   ,   5   ,   101 ,   107 ,   11  ,   817 ,   413 ,   3241    ,   6275    ,   4035    ,0 };
            const unsigned long dim1740KuoInit[]        =   {   1   ,   1   ,   1   ,   3   ,   21  ,   17  ,   25  ,   113 ,   183 ,   159 ,   307 ,   447 ,   7881    ,   7399    ,0 };
            const unsigned long dim1741KuoInit[]        =   {   1   ,   1   ,   7   ,   15  ,   7   ,   45  ,   39  ,   39  ,   111 ,   191 ,   151 ,   267 ,   7133    ,   5557    ,0 };
            const unsigned long dim1742KuoInit[]        =   {   1   ,   3   ,   7   ,   5   ,   27  ,   39  ,   29  ,   59  ,   141 ,   835 ,   1307    ,   1007    ,   2101    ,   7675    ,0 };
            const unsigned long dim1743KuoInit[]        =   {   1   ,   3   ,   1   ,   11  ,   9   ,   15  ,   65  ,   47  ,   371 ,   385 ,   683 ,   3495    ,   711 ,   3963    ,0 };
            const unsigned long dim1744KuoInit[]        =   {   1   ,   1   ,   7   ,   3   ,   5   ,   59  ,   81  ,   9   ,   23  ,   885 ,   291 ,   2349    ,   5495    ,   12985   ,0 };
            const unsigned long dim1745KuoInit[]        =   {   1   ,   3   ,   5   ,   1   ,   21  ,   49  ,   99  ,   101 ,   447 ,   289 ,   659 ,   1061    ,   5771    ,   7591    ,0 };
            const unsigned long dim1746KuoInit[]        =   {   1   ,   3   ,   5   ,   7   ,   5   ,   11  ,   13  ,   255 ,   7   ,   775 ,   1919    ,   1965    ,   7069    ,   10379   ,0 };
            const unsigned long dim1747KuoInit[]        =   {   1   ,   1   ,   7   ,   15  ,   29  ,   23  ,   61  ,   183 ,   117 ,   817 ,   113 ,   1665    ,   2349    ,   305 ,0 };
            const unsigned long dim1748KuoInit[]        =   {   1   ,   3   ,   7   ,   13  ,   25  ,   9   ,   97  ,   161 ,   361 ,   33  ,   1781    ,   3717    ,   6319    ,   13199   ,0 };
            const unsigned long dim1749KuoInit[]        =   {   1   ,   3   ,   5   ,   13  ,   21  ,   63  ,   85  ,   225 ,   447 ,   169 ,   1601    ,   2359    ,   8087    ,   2609    ,0 };
            const unsigned long dim1750KuoInit[]        =   {   1   ,   1   ,   3   ,   9   ,   1   ,   11  ,   1   ,   173 ,   345 ,   217 ,   1131    ,   493 ,   4613    ,   9223    ,0 };
            const unsigned long dim1751KuoInit[]        =   {   1   ,   3   ,   7   ,   3   ,   5   ,   51  ,   55  ,   235 ,   129 ,   451 ,   1191    ,   2065    ,   6781    ,   9135    ,0 };
            const unsigned long dim1752KuoInit[]        =   {   1   ,   3   ,   7   ,   1   ,   27  ,   41  ,   123 ,   49  ,   495 ,   999 ,   1731    ,   1597    ,   7471    ,   1129    ,0 };
            const unsigned long dim1753KuoInit[]        =   {   1   ,   1   ,   5   ,   3   ,   13  ,   3   ,   123 ,   83  ,   199 ,   325 ,   1521    ,   1023    ,   2703    ,   10915   ,0 };
            const unsigned long dim1754KuoInit[]        =   {   1   ,   1   ,   3   ,   1   ,   15  ,   19  ,   45  ,   223 ,   17  ,   643 ,   661 ,   633 ,   6475    ,   10153   ,0 };
            const unsigned long dim1755KuoInit[]        =   {   1   ,   1   ,   7   ,   9   ,   17  ,   21  ,   101 ,   137 ,   39  ,   975 ,   1003    ,   1419    ,   353 ,   11971   ,0 };
            const unsigned long dim1756KuoInit[]        =   {   1   ,   3   ,   5   ,   3   ,   11  ,   37  ,   75  ,   87  ,   49  ,   535 ,   801 ,   1689    ,   8125    ,   12419   ,0 };
            const unsigned long dim1757KuoInit[]        =   {   1   ,   3   ,   3   ,   3   ,   19  ,   11  ,   99  ,   97  ,   285 ,   641 ,   1485    ,   257 ,   5417    ,   8669    ,0 };
            const unsigned long dim1758KuoInit[]        =   {   1   ,   1   ,   1   ,   1   ,   21  ,   53  ,   19  ,   61  ,   223 ,   917 ,   931 ,   1779    ,   5859    ,   3111    ,0 };
            const unsigned long dim1759KuoInit[]        =   {   1   ,   3   ,   1   ,   13  ,   25  ,   17  ,   55  ,   227 ,   61  ,   953 ,   255 ,   3939    ,   2523    ,   2119    ,0 };
            const unsigned long dim1760KuoInit[]        =   {   1   ,   1   ,   7   ,   11  ,   23  ,   43  ,   111 ,   21  ,   263 ,   399 ,   723 ,   1067    ,   3633    ,   4165    ,0 };
            const unsigned long dim1761KuoInit[]        =   {   1   ,   3   ,   7   ,   3   ,   31  ,   29  ,   69  ,   135 ,   105 ,   839 ,   1077    ,   1479    ,   643 ,   1063    ,0 };
            const unsigned long dim1762KuoInit[]        =   {   1   ,   3   ,   1   ,   11  ,   23  ,   29  ,   51  ,   235 ,   287 ,   941 ,   393 ,   2241    ,   1601    ,   4629    ,0 };
            const unsigned long dim1763KuoInit[]        =   {   1   ,   3   ,   3   ,   1   ,   1   ,   7   ,   47  ,   119 ,   369 ,   789 ,   1943    ,   3417    ,   4073    ,   7495    ,0 };
            const unsigned long dim1764KuoInit[]        =   {   1   ,   1   ,   7   ,   7   ,   21  ,   37  ,   25  ,   217 ,   267 ,   275 ,   903 ,   865 ,   6251    ,   1323    ,0 };
            const unsigned long dim1765KuoInit[]        =   {   1   ,   3   ,   7   ,   15  ,   13  ,   61  ,   77  ,   181 ,   479 ,   1003    ,   335 ,   2917    ,   1133    ,   14321   ,0 };
            const unsigned long dim1766KuoInit[]        =   {   1   ,   1   ,   7   ,   7   ,   9   ,   53  ,   47  ,   197 ,   257 ,   823 ,   421 ,   2465    ,   4333    ,   3841    ,0 };
            const unsigned long dim1767KuoInit[]        =   {   1   ,   3   ,   3   ,   9   ,   23  ,   17  ,   83  ,   131 ,   427 ,   25  ,   319 ,   517 ,   4109    ,   15841   ,0 };
            const unsigned long dim1768KuoInit[]        =   {   1   ,   1   ,   7   ,   7   ,   7   ,   25  ,   105 ,   63  ,   117 ,   173 ,   909 ,   2731    ,   877 ,   1095    ,0 };
            const unsigned long dim1769KuoInit[]        =   {   1   ,   1   ,   3   ,   9   ,   29  ,   17  ,   125 ,   45  ,   269 ,   681 ,   1389    ,   283 ,   7217    ,   15315   ,0 };
            const unsigned long dim1770KuoInit[]        =   {   1   ,   1   ,   7   ,   3   ,   3   ,   59  ,   29  ,   177 ,   191 ,   803 ,   809 ,   131 ,   857 ,   16213   ,0 };
            const unsigned long dim1771KuoInit[]        =   {   1   ,   1   ,   1   ,   15  ,   15  ,   21  ,   41  ,   177 ,   245 ,   543 ,   761 ,   2423    ,   1381    ,   1125    ,0 };
            const unsigned long dim1772KuoInit[]        =   {   1   ,   3   ,   5   ,   13  ,   31  ,   5   ,   121 ,   183 ,   477 ,   143 ,   559 ,   765 ,   703 ,   12255   ,0 };
            const unsigned long dim1773KuoInit[]        =   {   1   ,   1   ,   1   ,   15  ,   21  ,   23  ,   127 ,   227 ,   331 ,   963 ,   333 ,   2101    ,   2215    ,   10107   ,0 };
            const unsigned long dim1774KuoInit[]        =   {   1   ,   3   ,   7   ,   3   ,   1   ,   47  ,   53  ,   189 ,   331 ,   587 ,   1417    ,   731 ,   4317    ,   11629   ,0 };
            const unsigned long dim1775KuoInit[]        =   {   1   ,   3   ,   7   ,   13  ,   11  ,   21  ,   45  ,   109 ,   455 ,   347 ,   1483    ,   2589    ,   6933    ,   8343    ,0 };
            const unsigned long dim1776KuoInit[]        =   {   1   ,   1   ,   1   ,   11  ,   5   ,   31  ,   45  ,   85  ,   253 ,   729 ,   111 ,   3315    ,   3183    ,   12847   ,0 };
            const unsigned long dim1777KuoInit[]        =   {   1   ,   3   ,   1   ,   1   ,   25  ,   5   ,   3   ,   219 ,   263 ,   517 ,   247 ,   2303    ,   2709    ,   11973   ,0 };
            const unsigned long dim1778KuoInit[]        =   {   1   ,   1   ,   5   ,   11  ,   21  ,   39  ,   115 ,   231 ,   5   ,   451 ,   1503    ,   1527    ,   4071    ,   5879    ,0 };
            const unsigned long dim1779KuoInit[]        =   {   1   ,   3   ,   5   ,   15  ,   23  ,   31  ,   15  ,   61  ,   313 ,   83  ,   49  ,   1377    ,   7029    ,   14187   ,0 };
            const unsigned long dim1780KuoInit[]        =   {   1   ,   3   ,   7   ,   9   ,   29  ,   7   ,   117 ,   111 ,   493 ,   159 ,   513 ,   3841    ,   7011    ,   14723   ,0 };
            const unsigned long dim1781KuoInit[]        =   {   1   ,   1   ,   5   ,   13  ,   21  ,   15  ,   3   ,   139 ,   463 ,   1023    ,   1301    ,   243 ,   5105    ,   2259    ,0 };
            const unsigned long dim1782KuoInit[]        =   {   1   ,   1   ,   7   ,   13  ,   5   ,   3   ,   119 ,   111 ,   447 ,   75  ,   1161    ,   1917    ,   345 ,   557 ,0 };
            const unsigned long dim1783KuoInit[]        =   {   1   ,   1   ,   5   ,   5   ,   25  ,   21  ,   87  ,   93  ,   237 ,   921 ,   475 ,   2555    ,   4105    ,   5691    ,0 };
            const unsigned long dim1784KuoInit[]        =   {   1   ,   1   ,   3   ,   15  ,   29  ,   51  ,   5   ,   73  ,   471 ,   421 ,   861 ,   3833    ,   2531    ,   1059    ,0 };
            const unsigned long dim1785KuoInit[]        =   {   1   ,   3   ,   7   ,   9   ,   3   ,   11  ,   101 ,   35  ,   271 ,   981 ,   1661    ,   1629    ,   7321    ,   1335    ,0 };
            const unsigned long dim1786KuoInit[]        =   {   1   ,   1   ,   5   ,   9   ,   9   ,   37  ,   9   ,   127 ,   45  ,   215 ,   611 ,   1193    ,   659 ,   5975    ,0 };
            const unsigned long dim1787KuoInit[]        =   {   1   ,   1   ,   3   ,   9   ,   31  ,   31  ,   21  ,   167 ,   415 ,   513 ,   1917    ,   829 ,   1161    ,   3585    ,0 };
            const unsigned long dim1788KuoInit[]        =   {   1   ,   1   ,   7   ,   1   ,   19  ,   23  ,   49  ,   55  ,   169 ,   351 ,   1389    ,   3251    ,   5001    ,   13445   ,0 };
            const unsigned long dim1789KuoInit[]        =   {   1   ,   3   ,   1   ,   15  ,   3   ,   3   ,   113 ,   193 ,   391 ,   851 ,   681 ,   2961    ,   7239    ,   12837   ,0 };
            const unsigned long dim1790KuoInit[]        =   {   1   ,   3   ,   5   ,   15  ,   5   ,   43  ,   71  ,   1   ,   347 ,   635 ,   1123    ,   3399    ,   2087    ,   8209    ,0 };
            const unsigned long dim1791KuoInit[]        =   {   1   ,   1   ,   3   ,   15  ,   23  ,   7   ,   59  ,   115 ,   139 ,   311 ,   893 ,   2421    ,   405 ,   6541    ,0 };
            const unsigned long dim1792KuoInit[]        =   {   1   ,   3   ,   3   ,   3   ,   25  ,   23  ,   127 ,   135 ,   199 ,   661 ,   7   ,   3279    ,   91  ,   11879   ,0 };
            const unsigned long dim1793KuoInit[]        =   {   1   ,   1   ,   3   ,   3   ,   19  ,   37  ,   33  ,   159 ,   3   ,   425 ,   379 ,   1321    ,   369 ,   13125   ,0 };
            const unsigned long dim1794KuoInit[]        =   {   1   ,   3   ,   5   ,   15  ,   7   ,   17  ,   47  ,   47  ,   167 ,   909 ,   403 ,   95  ,   7861    ,   231 ,0 };
            const unsigned long dim1795KuoInit[]        =   {   1   ,   1   ,   5   ,   1   ,   15  ,   5   ,   69  ,   179 ,   229 ,   597 ,   1743    ,   467 ,   1859    ,   2059    ,0 };
            const unsigned long dim1796KuoInit[]        =   {   1   ,   1   ,   7   ,   15  ,   5   ,   63  ,   109 ,   191 ,   271 ,   281 ,   907 ,   489 ,   5863    ,   12697   ,0 };
            const unsigned long dim1797KuoInit[]        =   {   1   ,   1   ,   3   ,   3   ,   31  ,   1   ,   67  ,   81  ,   3   ,   755 ,   1839    ,   1481    ,   4691    ,   14549   ,0 };
            const unsigned long dim1798KuoInit[]        =   {   1   ,   3   ,   3   ,   15  ,   17  ,   13  ,   59  ,   23  ,   85  ,   951 ,   1259    ,   363 ,   5479    ,   10705   ,0 };
            const unsigned long dim1799KuoInit[]        =   {   1   ,   1   ,   3   ,   15  ,   11  ,   61  ,   111 ,   109 ,   361 ,   483 ,   1691    ,   1039    ,   3777    ,   14613   ,0 };
            const unsigned long dim1800KuoInit[]        =   {   1   ,   1   ,   3   ,   5   ,   29  ,   55  ,   101 ,   55  ,   407 ,   61  ,   607 ,   829 ,   8011    ,   14021   ,0 };
            const unsigned long dim1801KuoInit[]        =   {   1   ,   1   ,   1   ,   13  ,   1   ,   39  ,   29  ,   19  ,   389 ,   85  ,   313 ,   1123    ,   1303    ,   11535   ,0 };
            const unsigned long dim1802KuoInit[]        =   {   1   ,   1   ,   1   ,   5   ,   5   ,   45  ,   47  ,   139 ,   165 ,   937 ,   1677    ,   2973    ,   501 ,   4387    ,0 };
            const unsigned long dim1803KuoInit[]        =   {   1   ,   3   ,   5   ,   9   ,   5   ,   55  ,   89  ,   235 ,   29  ,   75  ,   1839    ,   801 ,   5753    ,   4409    ,0 };
            const unsigned long dim1804KuoInit[]        =   {   1   ,   1   ,   3   ,   5   ,   7   ,   41  ,   7   ,   185 ,   297 ,   661 ,   1955    ,   2193    ,   6001    ,   12349   ,0 };
            const unsigned long dim1805KuoInit[]        =   {   1   ,   1   ,   7   ,   15  ,   27  ,   47  ,   7   ,   7   ,   183 ,   725 ,   1511    ,   3021    ,   6091    ,   10567   ,0 };
            const unsigned long dim1806KuoInit[]        =   {   1   ,   3   ,   3   ,   13  ,   21  ,   7   ,   79  ,   133 ,   307 ,   759 ,   369 ,   2779    ,   7629    ,   16191   ,0 };
            const unsigned long dim1807KuoInit[]        =   {   1   ,   1   ,   5   ,   9   ,   5   ,   47  ,   15  ,   141 ,   293 ,   339 ,   1471    ,   127 ,   5805    ,   5543    ,0 };
            const unsigned long dim1808KuoInit[]        =   {   1   ,   3   ,   1   ,   9   ,   31  ,   7   ,   125 ,   27  ,   511 ,   549 ,   619 ,   743 ,   3471    ,   8595    ,0 };
            const unsigned long dim1809KuoInit[]        =   {   1   ,   1   ,   7   ,   5   ,   15  ,   9   ,   25  ,   35  ,   161 ,   691 ,   659 ,   909 ,   969 ,   9285    ,0 };
            const unsigned long dim1810KuoInit[]        =   {   1   ,   1   ,   7   ,   11  ,   31  ,   17  ,   59  ,   255 ,   281 ,   7   ,   1075    ,   2561    ,   6883    ,   8193    ,0 };
            const unsigned long dim1811KuoInit[]        =   {   1   ,   1   ,   7   ,   1   ,   9   ,   15  ,   83  ,   253 ,   415 ,   601 ,   417 ,   3999    ,   7513    ,   3663    ,0 };
            const unsigned long dim1812KuoInit[]        =   {   1   ,   3   ,   5   ,   7   ,   11  ,   25  ,   109 ,   135 ,   41  ,   5   ,   697 ,   1703    ,   8175    ,   11369   ,0 };
            const unsigned long dim1813KuoInit[]        =   {   1   ,   1   ,   7   ,   15  ,   7   ,   39  ,   1   ,   253 ,   373 ,   273 ,   857 ,   401 ,   1609    ,   2391    ,0 };
            const unsigned long dim1814KuoInit[]        =   {   1   ,   1   ,   5   ,   1   ,   1   ,   61  ,   85  ,   213 ,   183 ,   695 ,   523 ,   1237    ,   5485    ,   6843    ,0 };
            const unsigned long dim1815KuoInit[]        =   {   1   ,   1   ,   7   ,   9   ,   17  ,   41  ,   125 ,   209 ,   285 ,   823 ,   857 ,   31  ,   839 ,   10847   ,0 };
            const unsigned long dim1816KuoInit[]        =   {   1   ,   3   ,   5   ,   9   ,   31  ,   23  ,   65  ,   75  ,   243 ,   471 ,   669 ,   539 ,   3949    ,   14913   ,0 };
            const unsigned long dim1817KuoInit[]        =   {   1   ,   1   ,   7   ,   3   ,   25  ,   23  ,   109 ,   19  ,   89  ,   983 ,   193 ,   1433    ,   6401    ,   663 ,0 };
            const unsigned long dim1818KuoInit[]        =   {   1   ,   3   ,   1   ,   15  ,   25  ,   9   ,   93  ,   13  ,   133 ,   561 ,   1343    ,   4049    ,   6221    ,   3137    ,0 };
            const unsigned long dim1819KuoInit[]        =   {   1   ,   1   ,   3   ,   3   ,   31  ,   47  ,   105 ,   225 ,   231 ,   75  ,   187 ,   963 ,   3339    ,   9705    ,0 };
            const unsigned long dim1820KuoInit[]        =   {   1   ,   3   ,   7   ,   1   ,   9   ,   31  ,   21  ,   221 ,   309 ,   791 ,   2011    ,   241 ,   5507    ,   11579   ,0 };
            const unsigned long dim1821KuoInit[]        =   {   1   ,   3   ,   7   ,   9   ,   1   ,   19  ,   53  ,   239 ,   305 ,   151 ,   53  ,   873 ,   5235    ,   13357   ,0 };
            const unsigned long dim1822KuoInit[]        =   {   1   ,   1   ,   5   ,   15  ,   13  ,   27  ,   23  ,   255 ,   75  ,   435 ,   209 ,   2001    ,   2529    ,   5363    ,0 };
            const unsigned long dim1823KuoInit[]        =   {   1   ,   1   ,   7   ,   5   ,   25  ,   53  ,   71  ,   203 ,   259 ,   403 ,   1535    ,   3381    ,   6791    ,   16061   ,0 };
            const unsigned long dim1824KuoInit[]        =   {   1   ,   3   ,   5   ,   3   ,   29  ,   5   ,   93  ,   5   ,   179 ,   813 ,   1569    ,   2875    ,   7855    ,   13863   ,0 };
            const unsigned long dim1825KuoInit[]        =   {   1   ,   1   ,   3   ,   7   ,   1   ,   13  ,   97  ,   125 ,   479 ,   209 ,   1659    ,   2997    ,   3405    ,   9707    ,0 };
            const unsigned long dim1826KuoInit[]        =   {   1   ,   3   ,   1   ,   9   ,   5   ,   43  ,   15  ,   29  ,   317 ,   317 ,   1317    ,   3177    ,   519 ,   8115    ,0 };
            const unsigned long dim1827KuoInit[]        =   {   1   ,   1   ,   1   ,   5   ,   1   ,   55  ,   49  ,   145 ,   415 ,   785 ,   1489    ,   1217    ,   3829    ,   10207   ,0 };
            const unsigned long dim1828KuoInit[]        =   {   1   ,   3   ,   3   ,   9   ,   17  ,   5   ,   99  ,   235 ,   277 ,   921 ,   71  ,   343 ,   2935    ,   12495   ,0 };
            const unsigned long dim1829KuoInit[]        =   {   1   ,   1   ,   3   ,   15  ,   25  ,   11  ,   11  ,   69  ,   209 ,   543 ,   17  ,   1489    ,   8119    ,   3557    ,0 };
            const unsigned long dim1830KuoInit[]        =   {   1   ,   1   ,   7   ,   9   ,   29  ,   43  ,   125 ,   87  ,   11  ,   947 ,   1709    ,   17  ,   3467    ,   3339    ,0 };
            const unsigned long dim1831KuoInit[]        =   {   1   ,   3   ,   5   ,   15  ,   23  ,   47  ,   113 ,   185 ,   75  ,   523 ,   1253    ,   859 ,   6359    ,   15563   ,0 };
            const unsigned long dim1832KuoInit[]        =   {   1   ,   1   ,   7   ,   11  ,   1   ,   25  ,   25  ,   23  ,   267 ,   275 ,   1751    ,   3237    ,   3967    ,   9943    ,0 };
            const unsigned long dim1833KuoInit[]        =   {   1   ,   1   ,   1   ,   13  ,   7   ,   35  ,   43  ,   151 ,   489 ,   97  ,   143 ,   257 ,   771 ,   8181    ,0 };
            const unsigned long dim1834KuoInit[]        =   {   1   ,   1   ,   3   ,   7   ,   19  ,   11  ,   17  ,   23  ,   425 ,   869 ,   525 ,   95  ,   2217    ,   3931    ,0 };
            const unsigned long dim1835KuoInit[]        =   {   1   ,   3   ,   7   ,   11  ,   3   ,   19  ,   77  ,   219 ,   501 ,   23  ,   773 ,   2421    ,   3713    ,   6209    ,0 };
            const unsigned long dim1836KuoInit[]        =   {   1   ,   1   ,   3   ,   1   ,   7   ,   1   ,   33  ,   35  ,   229 ,   329 ,   201 ,   291 ,   4165    ,   16059   ,0 };
            const unsigned long dim1837KuoInit[]        =   {   1   ,   3   ,   5   ,   9   ,   17  ,   23  ,   107 ,   23  ,   3   ,   673 ,   609 ,   2931    ,   5989    ,   3133    ,0 };
            const unsigned long dim1838KuoInit[]        =   {   1   ,   3   ,   3   ,   3   ,   17  ,   59  ,   31  ,   63  ,   111 ,   287 ,   407 ,   769 ,   2381    ,   2203    ,0 };
            const unsigned long dim1839KuoInit[]        =   {   1   ,   1   ,   1   ,   15  ,   3   ,   51  ,   105 ,   57  ,   313 ,   847 ,   261 ,   2697    ,   749 ,   15419   ,0 };
            const unsigned long dim1840KuoInit[]        =   {   1   ,   3   ,   1   ,   11  ,   3   ,   21  ,   65  ,   139 ,   363 ,   21  ,   1889    ,   1007    ,   6969    ,   3609    ,0 };
            const unsigned long dim1841KuoInit[]        =   {   1   ,   3   ,   7   ,   15  ,   17  ,   51  ,   73  ,   121 ,   481 ,   339 ,   1409    ,   131 ,   2005    ,   11847   ,0 };
            const unsigned long dim1842KuoInit[]        =   {   1   ,   3   ,   5   ,   3   ,   31  ,   13  ,   107 ,   253 ,   125 ,   973 ,   1027    ,   3473    ,   6883    ,   15061   ,0 };
            const unsigned long dim1843KuoInit[]        =   {   1   ,   3   ,   1   ,   5   ,   25  ,   17  ,   11  ,   255 ,   209 ,   35  ,   1987    ,   2275    ,   2327    ,   16079   ,0 };
            const unsigned long dim1844KuoInit[]        =   {   1   ,   3   ,   7   ,   5   ,   3   ,   1   ,   25  ,   93  ,   397 ,   467 ,   1895    ,   3199    ,   1861    ,   10943   ,0 };
            const unsigned long dim1845KuoInit[]        =   {   1   ,   3   ,   7   ,   1   ,   23  ,   27  ,   13  ,   211 ,   299 ,   117 ,   1689    ,   457 ,   6335    ,   11921   ,0 };
            const unsigned long dim1846KuoInit[]        =   {   1   ,   1   ,   5   ,   5   ,   9   ,   59  ,   49  ,   177 ,   81  ,   913 ,   429 ,   2705    ,   339 ,   9347    ,0 };
            const unsigned long dim1847KuoInit[]        =   {   1   ,   3   ,   1   ,   15  ,   1   ,   61  ,   81  ,   143 ,   455 ,   311 ,   381 ,   235 ,   2537    ,   12511   ,0 };
            const unsigned long dim1848KuoInit[]        =   {   1   ,   3   ,   5   ,   11  ,   7   ,   33  ,   103 ,   137 ,   377 ,   717 ,   851 ,   2939    ,   303 ,   11175   ,0 };
            const unsigned long dim1849KuoInit[]        =   {   1   ,   3   ,   3   ,   13  ,   25  ,   59  ,   15  ,   43  ,   71  ,   767 ,   47  ,   3425    ,   7015    ,   14431   ,0 };
            const unsigned long dim1850KuoInit[]        =   {   1   ,   3   ,   3   ,   9   ,   25  ,   37  ,   37  ,   39  ,   279 ,   491 ,   1483    ,   2633    ,   3791    ,   14311   ,0 };
            const unsigned long dim1851KuoInit[]        =   {   1   ,   1   ,   1   ,   13  ,   29  ,   63  ,   83  ,   89  ,   339 ,   923 ,   455 ,   965 ,   7339    ,   16383   ,0 };
            const unsigned long dim1852KuoInit[]        =   {   1   ,   1   ,   3   ,   1   ,   9   ,   33  ,   7   ,   63  ,   385 ,   717 ,   1449    ,   2705    ,   6233    ,   10753   ,0 };
            const unsigned long dim1853KuoInit[]        =   {   1   ,   3   ,   1   ,   11  ,   15  ,   13  ,   11  ,   181 ,   271 ,   749 ,   1825    ,   875 ,   2497    ,   12999   ,0 };
            const unsigned long dim1854KuoInit[]        =   {   1   ,   3   ,   3   ,   7   ,   13  ,   49  ,   35  ,   65  ,   467 ,   259 ,   429 ,   549 ,   5481    ,   11343   ,0 };
            const unsigned long dim1855KuoInit[]        =   {   1   ,   3   ,   7   ,   15  ,   17  ,   55  ,   125 ,   217 ,   91  ,   495 ,   133 ,   1985    ,   4871    ,   3415    ,0 };
            const unsigned long dim1856KuoInit[]        =   {   1   ,   3   ,   3   ,   13  ,   11  ,   33  ,   29  ,   53  ,   253 ,   935 ,   1919    ,   2607    ,   3205    ,   819 ,0 };
            const unsigned long dim1857KuoInit[]        =   {   1   ,   1   ,   1   ,   3   ,   25  ,   23  ,   127 ,   151 ,   39  ,   583 ,   985 ,   715 ,   765 ,   16343   ,0 };
            const unsigned long dim1858KuoInit[]        =   {   1   ,   3   ,   5   ,   3   ,   7   ,   21  ,   127 ,   179 ,   239 ,   353 ,   221 ,   531 ,   5147    ,   6699    ,0 };
            const unsigned long dim1859KuoInit[]        =   {   1   ,   1   ,   7   ,   9   ,   7   ,   63  ,   91  ,   99  ,   495 ,   111 ,   1295    ,   2609    ,   6389    ,   7719    ,0 };
            const unsigned long dim1860KuoInit[]        =   {   1   ,   3   ,   5   ,   7   ,   9   ,   27  ,   123 ,   171 ,   155 ,   683 ,   1209    ,   2081    ,   2823    ,   1007    ,0 };
            const unsigned long dim1861KuoInit[]        =   {   1   ,   3   ,   3   ,   15  ,   25  ,   45  ,   91  ,   93  ,   329 ,   961 ,   1843    ,   2637    ,   4333    ,   8359    ,0 };
            const unsigned long dim1862KuoInit[]        =   {   1   ,   3   ,   1   ,   5   ,   9   ,   5   ,   73  ,   221 ,   71  ,   951 ,   1175    ,   2489    ,   4565    ,   7221    ,0 };
            const unsigned long dim1863KuoInit[]        =   {   1   ,   3   ,   1   ,   3   ,   31  ,   33  ,   67  ,   127 ,   427 ,   487 ,   239 ,   787 ,   1731    ,   11867   ,0 };
            const unsigned long dim1864KuoInit[]        =   {   1   ,   1   ,   1   ,   9   ,   17  ,   23  ,   5   ,   201 ,   175 ,   661 ,   1381    ,   3017    ,   4237    ,   14403   ,0 };
            const unsigned long dim1865KuoInit[]        =   {   1   ,   3   ,   7   ,   7   ,   31  ,   7   ,   39  ,   141 ,   299 ,   237 ,   281 ,   3057    ,   5553    ,   11339   ,0 };
            const unsigned long dim1866KuoInit[]        =   {   1   ,   1   ,   1   ,   15  ,   25  ,   49  ,   51  ,   135 ,   363 ,   377 ,   1767    ,   4069    ,   3001    ,   7111    ,0 };
            const unsigned long dim1867KuoInit[]        =   {   1   ,   3   ,   7   ,   11  ,   29  ,   7   ,   91  ,   145 ,   263 ,   581 ,   35  ,   3101    ,   6227    ,   12007   ,   30807   ,0 };
            const unsigned long dim1868KuoInit[]        =   {   1   ,   1   ,   1   ,   13  ,   11  ,   41  ,   61  ,   135 ,   61  ,   501 ,   189 ,   1923    ,   129 ,   14563   ,   12247   ,0 };
            const unsigned long dim1869KuoInit[]        =   {   1   ,   1   ,   3   ,   5   ,   17  ,   45  ,   53  ,   5   ,   199 ,   165 ,   859 ,   771 ,   4553    ,   14723   ,   17293   ,0 };
            const unsigned long dim1870KuoInit[]        =   {   1   ,   1   ,   3   ,   15  ,   27  ,   57  ,   41  ,   3   ,   437 ,   543 ,   1291    ,   1083    ,   1311    ,   4407    ,   9107    ,0 };
            const unsigned long dim1871KuoInit[]        =   {   1   ,   1   ,   3   ,   11  ,   19  ,   23  ,   17  ,   251 ,   267 ,   575 ,   1675    ,   775 ,   1281    ,   12389   ,   11781   ,0 };
            const unsigned long dim1872KuoInit[]        =   {   1   ,   3   ,   3   ,   9   ,   19  ,   41  ,   69  ,   45  ,   113 ,   683 ,   1843    ,   1217    ,   1075    ,   7141    ,   11259   ,0 };
            const unsigned long dim1873KuoInit[]        =   {   1   ,   3   ,   5   ,   5   ,   21  ,   45  ,   75  ,   199 ,   227 ,   717 ,   267 ,   3211    ,   5893    ,   6559    ,   16445   ,0 };
            const unsigned long dim1874KuoInit[]        =   {   1   ,   1   ,   3   ,   15  ,   25  ,   19  ,   13  ,   3   ,   477 ,   511 ,   1607    ,   61  ,   5937    ,   6471    ,   4561    ,0 };
            const unsigned long dim1875KuoInit[]        =   {   1   ,   1   ,   7   ,   9   ,   25  ,   9   ,   59  ,   209 ,   181 ,   671 ,   1473    ,   1863    ,   6835    ,   5915    ,   21573   ,0 };
            const unsigned long dim1876KuoInit[]        =   {   1   ,   1   ,   7   ,   5   ,   31  ,   43  ,   91  ,   191 ,   505 ,   979 ,   1431    ,   1347    ,   1445    ,   9759    ,   31653   ,0 };
            const unsigned long dim1877KuoInit[]        =   {   1   ,   1   ,   5   ,   3   ,   11  ,   63  ,   35  ,   1   ,   313 ,   863 ,   577 ,   3039    ,   5205    ,   5089    ,   20009   ,0 };
            const unsigned long dim1878KuoInit[]        =   {   1   ,   3   ,   3   ,   7   ,   1   ,   27  ,   79  ,   21  ,   179 ,   187 ,   917 ,   1831    ,   1737    ,   2879    ,   5801    ,0 };
            const unsigned long dim1879KuoInit[]        =   {   1   ,   1   ,   1   ,   3   ,   19  ,   51  ,   77  ,   201 ,   499 ,   345 ,   617 ,   33  ,   543 ,   6925    ,   5945    ,0 };
            const unsigned long dim1880KuoInit[]        =   {   1   ,   3   ,   1   ,   1   ,   19  ,   49  ,   67  ,   51  ,   181 ,   1015    ,   1159    ,   2109    ,   879 ,   10257   ,   23207   ,0 };
            const unsigned long dim1881KuoInit[]        =   {   1   ,   1   ,   7   ,   7   ,   19  ,   29  ,   7   ,   111 ,   365 ,   373 ,   1309    ,   3633    ,   5951    ,   2597    ,   27947   ,0 };
            const unsigned long dim1882KuoInit[]        =   {   1   ,   3   ,   5   ,   13  ,   21  ,   27  ,   85  ,   71  ,   149 ,   665 ,   1499    ,   4025    ,   3939    ,   8947    ,   15811   ,0 };
            const unsigned long dim1883KuoInit[]        =   {   1   ,   3   ,   7   ,   5   ,   25  ,   47  ,   77  ,   99  ,   105 ,   739 ,   1077    ,   3145    ,   7859    ,   1925    ,   16765   ,0 };
            const unsigned long dim1884KuoInit[]        =   {   1   ,   1   ,   3   ,   11  ,   25  ,   9   ,   113 ,   105 ,   309 ,   397 ,   1687    ,   1171    ,   455 ,   3183    ,   12487   ,0 };
            const unsigned long dim1885KuoInit[]        =   {   1   ,   1   ,   7   ,   15  ,   7   ,   9   ,   51  ,   239 ,   469 ,   415 ,   1895    ,   89  ,   4243    ,   967 ,   7113    ,0 };
            const unsigned long dim1886KuoInit[]        =   {   1   ,   1   ,   5   ,   1   ,   23  ,   41  ,   39  ,   247 ,   115 ,   541 ,   1461    ,   2907    ,   2175    ,   3133    ,   3171    ,0 };
            const unsigned long dim1887KuoInit[]        =   {   1   ,   3   ,   7   ,   11  ,   7   ,   53  ,   43  ,   187 ,   213 ,   549 ,   947 ,   3695    ,   3931    ,   699 ,   21299   ,0 };
            const unsigned long dim1888KuoInit[]        =   {   1   ,   1   ,   1   ,   7   ,   15  ,   9   ,   9   ,   3   ,   175 ,   231 ,   1529    ,   337 ,   5333    ,   6945    ,   1359    ,0 };
            const unsigned long dim1889KuoInit[]        =   {   1   ,   1   ,   3   ,   3   ,   13  ,   25  ,   123 ,   231 ,   389 ,   351 ,   1695    ,   2959    ,   1037    ,   2103    ,   327 ,0 };
            const unsigned long dim1890KuoInit[]        =   {   1   ,   3   ,   5   ,   11  ,   9   ,   31  ,   19  ,   149 ,   1   ,   477 ,   1819    ,   3487    ,   3299    ,   8835    ,   10351   ,0 };
            const unsigned long dim1891KuoInit[]        =   {   1   ,   1   ,   5   ,   7   ,   9   ,   37  ,   123 ,   93  ,   267 ,   481 ,   901 ,   3927    ,   7295    ,   13251   ,   22551   ,0 };
            const unsigned long dim1892KuoInit[]        =   {   1   ,   1   ,   7   ,   11  ,   21  ,   13  ,   127 ,   173 ,   459 ,   169 ,   1511    ,   1649    ,   5431    ,   8637    ,   3601    ,0 };
            const unsigned long dim1893KuoInit[]        =   {   1   ,   1   ,   1   ,   9   ,   29  ,   3   ,   101 ,   135 ,   159 ,   295 ,   1767    ,   3089    ,   3753    ,   4969    ,   21523   ,0 };
            const unsigned long dim1894KuoInit[]        =   {   1   ,   3   ,   5   ,   3   ,   3   ,   9   ,   45  ,   143 ,   301 ,   307 ,   191 ,   3337    ,   6721    ,   12095   ,   7017    ,0 };
            const unsigned long dim1895KuoInit[]        =   {   1   ,   3   ,   7   ,   13  ,   17  ,   25  ,   55  ,   37  ,   379 ,   725 ,   1079    ,   311 ,   2853    ,   15097   ,   31857   ,0 };
            const unsigned long dim1896KuoInit[]        =   {   1   ,   3   ,   1   ,   13  ,   19  ,   57  ,   39  ,   99  ,   39  ,   283 ,   425 ,   2447    ,   2077    ,   815 ,   7639    ,0 };
            const unsigned long dim1897KuoInit[]        =   {   1   ,   3   ,   3   ,   15  ,   11  ,   5   ,   17  ,   83  ,   231 ,   535 ,   35  ,   2773    ,   4127    ,   9951    ,   5773    ,0 };
            const unsigned long dim1898KuoInit[]        =   {   1   ,   3   ,   3   ,   15  ,   7   ,   47  ,   87  ,   119 ,   291 ,   523 ,   1325    ,   1757    ,   2377    ,   6481    ,   7901    ,0 };
            const unsigned long dim1899KuoInit[]        =   {   1   ,   3   ,   7   ,   15  ,   25  ,   21  ,   29  ,   87  ,   241 ,   241 ,   1273    ,   2813    ,   2011    ,   8549    ,   7253    ,0 };
            const unsigned long dim1900KuoInit[]        =   {   1   ,   1   ,   1   ,   7   ,   19  ,   35  ,   119 ,   227 ,   483 ,   1015    ,   685 ,   1149    ,   4823    ,   4113    ,   7169    ,0 };
            const unsigned long dim1901KuoInit[]        =   {   1   ,   3   ,   3   ,   5   ,   11  ,   59  ,   125 ,   99  ,   255 ,   149 ,   829 ,   2869    ,   7927    ,   6779    ,   31949   ,0 };
            const unsigned long dim1902KuoInit[]        =   {   1   ,   3   ,   7   ,   3   ,   3   ,   49  ,   21  ,   111 ,   219 ,   33  ,   91  ,   1585    ,   303 ,   14931   ,   11777   ,0 };
            const unsigned long dim1903KuoInit[]        =   {   1   ,   3   ,   3   ,   11  ,   3   ,   47  ,   27  ,   227 ,   477 ,   891 ,   1281    ,   1927    ,   7325    ,   14075   ,   26633   ,0 };
            const unsigned long dim1904KuoInit[]        =   {   1   ,   1   ,   5   ,   5   ,   5   ,   29  ,   25  ,   23  ,   223 ,   65  ,   1511    ,   2563    ,   7825    ,   7339    ,   32095   ,0 };
            const unsigned long dim1905KuoInit[]        =   {   1   ,   1   ,   5   ,   7   ,   3   ,   29  ,   49  ,   83  ,   317 ,   513 ,   755 ,   1339    ,   4829    ,   6213    ,   29055   ,0 };
            const unsigned long dim1906KuoInit[]        =   {   1   ,   1   ,   3   ,   5   ,   23  ,   23  ,   97  ,   27  ,   379 ,   183 ,   933 ,   3431    ,   797 ,   9655    ,   32061   ,0 };
            const unsigned long dim1907KuoInit[]        =   {   1   ,   1   ,   7   ,   1   ,   29  ,   61  ,   31  ,   75  ,   331 ,   487 ,   229 ,   2319    ,   1385    ,   14697   ,   17081   ,0 };
            const unsigned long dim1908KuoInit[]        =   {   1   ,   1   ,   1   ,   5   ,   19  ,   1   ,   43  ,   27  ,   433 ,   751 ,   1283    ,   559 ,   7941    ,   7385    ,   13809   ,0 };
            const unsigned long dim1909KuoInit[]        =   {   1   ,   1   ,   3   ,   15  ,   19  ,   37  ,   97  ,   219 ,   217 ,   39  ,   1969    ,   1909    ,   2895    ,   749 ,   27607   ,0 };
            const unsigned long dim1910KuoInit[]        =   {   1   ,   3   ,   5   ,   9   ,   25  ,   17  ,   79  ,   213 ,   321 ,   73  ,   533 ,   3773    ,   6989    ,   10183   ,   20179   ,0 };
            const unsigned long dim1911KuoInit[]        =   {   1   ,   1   ,   5   ,   7   ,   15  ,   51  ,   71  ,   161 ,   107 ,   531 ,   461 ,   15  ,   2887    ,   13271   ,   24273   ,0 };
            const unsigned long dim1912KuoInit[]        =   {   1   ,   3   ,   5   ,   7   ,   29  ,   21  ,   119 ,   53  ,   207 ,   471 ,   987 ,   1771    ,   7259    ,   4377    ,   15405   ,0 };
            const unsigned long dim1913KuoInit[]        =   {   1   ,   1   ,   3   ,   11  ,   3   ,   5   ,   7   ,   189 ,   231 ,   181 ,   751 ,   1   ,   225 ,   12575   ,   4853    ,0 };
            const unsigned long dim1914KuoInit[]        =   {   1   ,   1   ,   1   ,   13  ,   1   ,   51  ,   35  ,   239 ,   145 ,   827 ,   1169    ,   3671    ,   7585    ,   4593    ,   4433    ,0 };
            const unsigned long dim1915KuoInit[]        =   {   1   ,   1   ,   3   ,   13  ,   15  ,   55  ,   43  ,   157 ,   13  ,   517 ,   1669    ,   25  ,   1175    ,   3971    ,   11165   ,0 };
            const unsigned long dim1916KuoInit[]        =   {   1   ,   3   ,   7   ,   15  ,   15  ,   37  ,   33  ,   155 ,   243 ,   375 ,   947 ,   1869    ,   2693    ,   5191    ,   9865    ,0 };
            const unsigned long dim1917KuoInit[]        =   {   1   ,   3   ,   7   ,   1   ,   15  ,   57  ,   73  ,   15  ,   389 ,   341 ,   207 ,   3249    ,   439 ,   5975    ,   23461   ,0 };
            const unsigned long dim1918KuoInit[]        =   {   1   ,   3   ,   3   ,   11  ,   17  ,   3   ,   35  ,   245 ,   349 ,   915 ,   1221    ,   1495    ,   5525    ,   10223   ,   29109   ,0 };
            const unsigned long dim1919KuoInit[]        =   {   1   ,   1   ,   7   ,   9   ,   19  ,   17  ,   39  ,   49  ,   159 ,   163 ,   419 ,   1117    ,   7947    ,   7177    ,   10945   ,0 };
            const unsigned long dim1920KuoInit[]        =   {   1   ,   3   ,   7   ,   5   ,   23  ,   35  ,   31  ,   33  ,   325 ,   565 ,   1951    ,   745 ,   6867    ,   13121   ,   25831   ,0 };
            const unsigned long dim1921KuoInit[]        =   {   1   ,   3   ,   7   ,   7   ,   27  ,   11  ,   89  ,   191 ,   233 ,   57  ,   1621    ,   2167    ,   7155    ,   15613   ,   14111   ,0 };
            const unsigned long dim1922KuoInit[]        =   {   1   ,   1   ,   7   ,   15  ,   31  ,   49  ,   3   ,   245 ,   51  ,   173 ,   1809    ,   379 ,   8085    ,   5125    ,   10279   ,0 };
            const unsigned long dim1923KuoInit[]        =   {   1   ,   1   ,   7   ,   7   ,   29  ,   59  ,   37  ,   197 ,   377 ,   5   ,   1851    ,   3685    ,   3537    ,   1897    ,   24257   ,0 };
            const unsigned long dim1924KuoInit[]        =   {   1   ,   1   ,   1   ,   11  ,   13  ,   9   ,   11  ,   19  ,   377 ,   435 ,   869 ,   497 ,   8181    ,   3411    ,   4415    ,0 };
            const unsigned long dim1925KuoInit[]        =   {   1   ,   1   ,   5   ,   7   ,   7   ,   21  ,   51  ,   115 ,   435 ,   9   ,   1099    ,   987 ,   1633    ,   7689    ,   19131   ,0 };
            const unsigned long dim1926KuoInit[]        =   {   1   ,   3   ,   3   ,   5   ,   17  ,   59  ,   93  ,   131 ,   17  ,   637 ,   211 ,   2759    ,   3059    ,   4691    ,   10225   ,0 };
            const unsigned long dim1927KuoInit[]        =   {   1   ,   3   ,   3   ,   13  ,   5   ,   35  ,   55  ,   49  ,   425 ,   395 ,   615 ,   1505    ,   1477    ,   6689    ,   28435   ,0 };
            const unsigned long dim1928KuoInit[]        =   {   1   ,   1   ,   3   ,   5   ,   27  ,   1   ,   43  ,   53  ,   375 ,   657 ,   1307    ,   1869    ,   415 ,   1143    ,   16961   ,0 };
            const unsigned long dim1929KuoInit[]        =   {   1   ,   1   ,   7   ,   5   ,   13  ,   27  ,   111 ,   113 ,   459 ,   149 ,   139 ,   79  ,   5031    ,   15165   ,   443 ,0 };
            const unsigned long dim1930KuoInit[]        =   {   1   ,   3   ,   1   ,   11  ,   9   ,   45  ,   35  ,   217 ,   259 ,   523 ,   601 ,   499 ,   7683    ,   1447    ,   5267    ,0 };
            const unsigned long dim1931KuoInit[]        =   {   1   ,   3   ,   1   ,   7   ,   3   ,   51  ,   101 ,   191 ,   469 ,   435 ,   883 ,   343 ,   1897    ,   5245    ,   1037    ,0 };
            const unsigned long dim1932KuoInit[]        =   {   1   ,   3   ,   5   ,   1   ,   19  ,   21  ,   109 ,   105 ,   363 ,   409 ,   1651    ,   3739    ,   339 ,   12915   ,   31603   ,0 };
            const unsigned long dim1933KuoInit[]        =   {   1   ,   3   ,   5   ,   13  ,   9   ,   7   ,   65  ,   193 ,   447 ,   829 ,   1123    ,   2431    ,   3357    ,   6687    ,   26903   ,0 };
            const unsigned long dim1934KuoInit[]        =   {   1   ,   3   ,   5   ,   5   ,   11  ,   9   ,   69  ,   147 ,   331 ,   657 ,   1589    ,   3417    ,   5805    ,   8673    ,   2517    ,0 };
            const unsigned long dim1935KuoInit[]        =   {   1   ,   3   ,   1   ,   5   ,   23  ,   45  ,   59  ,   53  ,   285 ,   711 ,   1447    ,   927 ,   8005    ,   3513    ,   18147   ,0 };
            const unsigned long dim1936KuoInit[]        =   {   1   ,   3   ,   5   ,   9   ,   9   ,   15  ,   17  ,   85  ,   165 ,   923 ,   915 ,   3263    ,   1103    ,   9445    ,   15371   ,0 };
            const unsigned long dim1937KuoInit[]        =   {   1   ,   1   ,   5   ,   13  ,   13  ,   13  ,   33  ,   249 ,   501 ,   495 ,   1771    ,   1305    ,   2821    ,   9405    ,   7193    ,0 };
            const unsigned long dim1938KuoInit[]        =   {   1   ,   1   ,   3   ,   5   ,   9   ,   11  ,   23  ,   5   ,   359 ,   977 ,   613 ,   2163    ,   4511    ,   7901    ,   1609    ,0 };
            const unsigned long dim1939KuoInit[]        =   {   1   ,   1   ,   3   ,   5   ,   29  ,   29  ,   73  ,   69  ,   155 ,   863 ,   243 ,   2705    ,   4005    ,   11249   ,   13983   ,0 };
            const unsigned long dim1940KuoInit[]        =   {   1   ,   3   ,   7   ,   9   ,   9   ,   15  ,   75  ,   171 ,   213 ,   931 ,   1613    ,   3801    ,   4617    ,   3929    ,   24787   ,0 };
            const unsigned long dim1941KuoInit[]        =   {   1   ,   3   ,   5   ,   11  ,   19  ,   35  ,   107 ,   251 ,   121 ,   455 ,   2027    ,   2323    ,   4297    ,   7949    ,   12043   ,0 };
            const unsigned long dim1942KuoInit[]        =   {   1   ,   3   ,   3   ,   7   ,   3   ,   59  ,   13  ,   151 ,   161 ,   615 ,   1427    ,   2469    ,   3211    ,   6601    ,   26829   ,0 };
            const unsigned long dim1943KuoInit[]        =   {   1   ,   1   ,   1   ,   9   ,   31  ,   1   ,   41  ,   247 ,   225 ,   65  ,   1751    ,   3557    ,   935 ,   16115   ,   3009    ,0 };
            const unsigned long dim1944KuoInit[]        =   {   1   ,   1   ,   7   ,   1   ,   31  ,   3   ,   31  ,   39  ,   279 ,   213 ,   1141    ,   371 ,   7529    ,   12485   ,   6585    ,0 };
            const unsigned long dim1945KuoInit[]        =   {   1   ,   3   ,   3   ,   3   ,   13  ,   31  ,   121 ,   185 ,   171 ,   991 ,   147 ,   1913    ,   475 ,   15155   ,   2427    ,0 };
            const unsigned long dim1946KuoInit[]        =   {   1   ,   3   ,   5   ,   5   ,   7   ,   13  ,   59  ,   127 ,   175 ,   717 ,   759 ,   3949    ,   4229    ,   2091    ,   20267   ,0 };
            const unsigned long dim1947KuoInit[]        =   {   1   ,   3   ,   1   ,   15  ,   21  ,   33  ,   45  ,   57  ,   1   ,   463 ,   1217    ,   1499    ,   269 ,   11159   ,   17175   ,0 };
            const unsigned long dim1948KuoInit[]        =   {   1   ,   1   ,   1   ,   9   ,   21  ,   29  ,   51  ,   63  ,   299 ,   775 ,   239 ,   575 ,   1033    ,   3627    ,   12017   ,0 };
            const unsigned long dim1949KuoInit[]        =   {   1   ,   3   ,   3   ,   3   ,   23  ,   35  ,   63  ,   197 ,   341 ,   775 ,   675 ,   1835    ,   4155    ,   7439    ,   1797    ,0 };
            const unsigned long dim1950KuoInit[]        =   {   1   ,   3   ,   7   ,   5   ,   23  ,   57  ,   71  ,   181 ,   1   ,   95  ,   337 ,   1971    ,   5721    ,   15013   ,   3387    ,0 };
            const unsigned long dim1951KuoInit[]        =   {   1   ,   3   ,   1   ,   5   ,   31  ,   21  ,   33  ,   203 ,   463 ,   3   ,   1261    ,   555 ,   3847    ,   5653    ,   18155   ,0 };
            const unsigned long dim1952KuoInit[]        =   {   1   ,   1   ,   3   ,   9   ,   27  ,   57  ,   57  ,   119 ,   213 ,   723 ,   1695    ,   2233    ,   1605    ,   8169    ,   28485   ,0 };
            const unsigned long dim1953KuoInit[]        =   {   1   ,   1   ,   7   ,   11  ,   17  ,   35  ,   119 ,   167 ,   391 ,   111 ,   229 ,   695 ,   4209    ,   5457    ,   6729    ,0 };
            const unsigned long dim1954KuoInit[]        =   {   1   ,   1   ,   7   ,   11  ,   5   ,   23  ,   5   ,   223 ,   89  ,   447 ,   2015    ,   3403    ,   1335    ,   15235   ,   20897   ,0 };
            const unsigned long dim1955KuoInit[]        =   {   1   ,   1   ,   1   ,   7   ,   1   ,   63  ,   55  ,   233 ,   351 ,   323 ,   497 ,   1247    ,   2455    ,   2337    ,   17523   ,0 };
            const unsigned long dim1956KuoInit[]        =   {   1   ,   3   ,   1   ,   11  ,   7   ,   21  ,   127 ,   207 ,   197 ,   479 ,   611 ,   3525    ,   2093    ,   2591    ,   21085   ,0 };
            const unsigned long dim1957KuoInit[]        =   {   1   ,   3   ,   5   ,   5   ,   1   ,   37  ,   107 ,   199 ,   29  ,   995 ,   1857    ,   133 ,   1423    ,   9967    ,   11367   ,0 };
            const unsigned long dim1958KuoInit[]        =   {   1   ,   1   ,   7   ,   11  ,   25  ,   25  ,   17  ,   155 ,   221 ,   539 ,   1895    ,   1909    ,   2921    ,   13107   ,   21681   ,0 };
            const unsigned long dim1959KuoInit[]        =   {   1   ,   1   ,   1   ,   13  ,   25  ,   59  ,   45  ,   41  ,   401 ,   657 ,   2043    ,   1317    ,   8065    ,   11897   ,   4767    ,0 };
            const unsigned long dim1960KuoInit[]        =   {   1   ,   1   ,   7   ,   15  ,   21  ,   29  ,   43  ,   125 ,   407 ,   851 ,   1421    ,   1101    ,   811 ,   9577    ,   27  ,0 };
            const unsigned long dim1961KuoInit[]        =   {   1   ,   3   ,   5   ,   7   ,   25  ,   49  ,   65  ,   175 ,   319 ,   807 ,   1691    ,   357 ,   4015    ,   1059    ,   599 ,0 };
            const unsigned long dim1962KuoInit[]        =   {   1   ,   1   ,   7   ,   13  ,   23  ,   5   ,   65  ,   79  ,   43  ,   665 ,   1931    ,   2143    ,   4577    ,   5905    ,   14917   ,0 };
            const unsigned long dim1963KuoInit[]        =   {   1   ,   3   ,   1   ,   7   ,   5   ,   41  ,   121 ,   251 ,   395 ,   951 ,   795 ,   1589    ,   4027    ,   13671   ,   18787   ,0 };
            const unsigned long dim1964KuoInit[]        =   {   1   ,   3   ,   7   ,   11  ,   3   ,   51  ,   57  ,   177 ,   63  ,   33  ,   1085    ,   1357    ,   5039    ,   2583    ,   2805    ,0 };
            const unsigned long dim1965KuoInit[]        =   {   1   ,   1   ,   1   ,   1   ,   29  ,   53  ,   117 ,   225 ,   373 ,   1   ,   901 ,   3871    ,   2815    ,   9871    ,   7737    ,0 };
            const unsigned long dim1966KuoInit[]        =   {   1   ,   3   ,   7   ,   7   ,   31  ,   61  ,   89  ,   253 ,   291 ,   829 ,   9   ,   1131    ,   4951    ,   14509   ,   23491   ,0 };
            const unsigned long dim1967KuoInit[]        =   {   1   ,   1   ,   7   ,   5   ,   25  ,   43  ,   35  ,   63  ,   425 ,   757 ,   217 ,   821 ,   1453    ,   8413    ,   7745    ,0 };
            const unsigned long dim1968KuoInit[]        =   {   1   ,   1   ,   5   ,   5   ,   7   ,   1   ,   41  ,   233 ,   235 ,   599 ,   833 ,   3637    ,   7005    ,   789 ,   27709   ,0 };
            const unsigned long dim1969KuoInit[]        =   {   1   ,   1   ,   3   ,   13  ,   7   ,   57  ,   69  ,   115 ,   417 ,   99  ,   1435    ,   3119    ,   1655    ,   3515    ,   7273    ,0 };
            const unsigned long dim1970KuoInit[]        =   {   1   ,   1   ,   3   ,   15  ,   9   ,   13  ,   33  ,   57  ,   179 ,   847 ,   289 ,   2213    ,   4635    ,   2349    ,   20203   ,0 };
            const unsigned long dim1971KuoInit[]        =   {   1   ,   1   ,   3   ,   9   ,   1   ,   39  ,   63  ,   147 ,   277 ,   389 ,   575 ,   423 ,   4943    ,   4327    ,   29067   ,0 };
            const unsigned long dim1972KuoInit[]        =   {   1   ,   1   ,   3   ,   13  ,   9   ,   9   ,   69  ,   167 ,   119 ,   591 ,   1679    ,   3543    ,   4517    ,   9263    ,   17957   ,0 };
            const unsigned long dim1973KuoInit[]        =   {   1   ,   3   ,   5   ,   13  ,   25  ,   37  ,   127 ,   159 ,   391 ,   29  ,   329 ,   3477    ,   6513    ,   11971   ,   18791   ,0 };
            const unsigned long dim1974KuoInit[]        =   {   1   ,   1   ,   7   ,   5   ,   25  ,   29  ,   49  ,   135 ,   387 ,   921 ,   391 ,   665 ,   6315    ,   16183   ,   553 ,0 };
            const unsigned long dim1975KuoInit[]        =   {   1   ,   1   ,   7   ,   3   ,   27  ,   43  ,   43  ,   253 ,   9   ,   285 ,   1073    ,   2627    ,   7417    ,   3223    ,   2649    ,0 };
            const unsigned long dim1976KuoInit[]        =   {   1   ,   1   ,   5   ,   3   ,   13  ,   57  ,   119 ,   37  ,   475 ,   471 ,   1145    ,   2709    ,   2115    ,   6507    ,   12393   ,0 };
            const unsigned long dim1977KuoInit[]        =   {   1   ,   1   ,   3   ,   1   ,   17  ,   29  ,   117 ,   189 ,   319 ,   533 ,   1299    ,   1987    ,   4971    ,   15719   ,   19395   ,0 };
            const unsigned long dim1978KuoInit[]        =   {   1   ,   1   ,   1   ,   15  ,   11  ,   63  ,   77  ,   245 ,   57  ,   9   ,   1285    ,   2069    ,   579 ,   5105    ,   13797   ,0 };
            const unsigned long dim1979KuoInit[]        =   {   1   ,   3   ,   1   ,   11  ,   19  ,   49  ,   1   ,   241 ,   89  ,   519 ,   591 ,   3875    ,   1585    ,   7239    ,   25867   ,0 };
            const unsigned long dim1980KuoInit[]        =   {   1   ,   1   ,   5   ,   11  ,   27  ,   47  ,   1   ,   143 ,   209 ,   695 ,   539 ,   3781    ,   6317    ,   14075   ,   27849   ,0 };
            const unsigned long dim1981KuoInit[]        =   {   1   ,   1   ,   3   ,   5   ,   17  ,   3   ,   17  ,   201 ,   105 ,   771 ,   497 ,   735 ,   3517    ,   5207    ,   20897   ,0 };
            const unsigned long dim1982KuoInit[]        =   {   1   ,   1   ,   3   ,   15  ,   3   ,   3   ,   103 ,   35  ,   421 ,   235 ,   1527    ,   3139    ,   2693    ,   9739    ,   63  ,0 };
            const unsigned long dim1983KuoInit[]        =   {   1   ,   3   ,   1   ,   13  ,   21  ,   39  ,   101 ,   213 ,   205 ,   853 ,   1455    ,   39  ,   7917    ,   7893    ,   4627    ,0 };
            const unsigned long dim1984KuoInit[]        =   {   1   ,   3   ,   7   ,   5   ,   13  ,   5   ,   9   ,   239 ,   129 ,   355 ,   809 ,   3413    ,   7535    ,   7063    ,   29149   ,0 };
            const unsigned long dim1985KuoInit[]        =   {   1   ,   1   ,   5   ,   3   ,   19  ,   37  ,   125 ,   55  ,   495 ,   597 ,   735 ,   1729    ,   619 ,   12881   ,   22913   ,0 };
            const unsigned long dim1986KuoInit[]        =   {   1   ,   1   ,   5   ,   11  ,   27  ,   31  ,   87  ,   65  ,   413 ,   421 ,   575 ,   2185    ,   4217    ,   12223   ,   22487   ,0 };
            const unsigned long dim1987KuoInit[]        =   {   1   ,   3   ,   3   ,   7   ,   23  ,   51  ,   115 ,   221 ,   369 ,   27  ,   229 ,   613 ,   3739    ,   1425    ,   10927   ,0 };
            const unsigned long dim1988KuoInit[]        =   {   1   ,   1   ,   1   ,   13  ,   23  ,   49  ,   73  ,   223 ,   139 ,   511 ,   899 ,   483 ,   7747    ,   7947    ,   32357   ,0 };
            const unsigned long dim1989KuoInit[]        =   {   1   ,   3   ,   5   ,   3   ,   7   ,   31  ,   9   ,   119 ,   417 ,   723 ,   2033    ,   663 ,   5927    ,   1787    ,   12841   ,0 };
            const unsigned long dim1990KuoInit[]        =   {   1   ,   1   ,   5   ,   9   ,   7   ,   5   ,   55  ,   201 ,   245 ,   221 ,   1447    ,   405 ,   447 ,   2447    ,   24331   ,0 };
            const unsigned long dim1991KuoInit[]        =   {   1   ,   3   ,   5   ,   11  ,   9   ,   47  ,   33  ,   175 ,   187 ,   513 ,   1613    ,   2413    ,   4207    ,   6941    ,   2341    ,0 };
            const unsigned long dim1992KuoInit[]        =   {   1   ,   1   ,   5   ,   9   ,   19  ,   33  ,   33  ,   181 ,   505 ,   923 ,   1923    ,   2471    ,   6499    ,   11243   ,   29653   ,0 };
            const unsigned long dim1993KuoInit[]        =   {   1   ,   1   ,   5   ,   15  ,   1   ,   49  ,   21  ,   15  ,   443 ,   565 ,   1955    ,   3299    ,   139 ,   14389   ,   28011   ,0 };
            const unsigned long dim1994KuoInit[]        =   {   1   ,   1   ,   3   ,   9   ,   17  ,   27  ,   29  ,   65  ,   509 ,   867 ,   707 ,   2425    ,   7279    ,   15271   ,   18239   ,0 };
            const unsigned long dim1995KuoInit[]        =   {   1   ,   1   ,   7   ,   15  ,   3   ,   1   ,   103 ,   79  ,   49  ,   909 ,   529 ,   1207    ,   231 ,   11181   ,   23319   ,0 };
            const unsigned long dim1996KuoInit[]        =   {   1   ,   1   ,   5   ,   1   ,   13  ,   23  ,   33  ,   29  ,   189 ,   491 ,   129 ,   501 ,   2635    ,   2071    ,   18057   ,0 };
            const unsigned long dim1997KuoInit[]        =   {   1   ,   1   ,   7   ,   7   ,   19  ,   1   ,   95  ,   185 ,   67  ,   251 ,   1017    ,   105 ,   3683    ,   9797    ,   24413   ,0 };
            const unsigned long dim1998KuoInit[]        =   {   1   ,   3   ,   5   ,   9   ,   13  ,   35  ,   79  ,   213 ,   223 ,   1013    ,   1205    ,   1777    ,   1415    ,   5023    ,   15901   ,0 };
            const unsigned long dim1999KuoInit[]        =   {   1   ,   1   ,   1   ,   5   ,   15  ,   13  ,   91  ,   57  ,   447 ,   897 ,   935 ,   1543    ,   5461    ,   5091    ,   12697   ,0 };
            const unsigned long dim2000KuoInit[]        =   {   1   ,   1   ,   1   ,   11  ,   9   ,   61  ,   81  ,   147 ,   55  ,   121 ,   449 ,   2833    ,   2383    ,   12065   ,   19435   ,0 };
            const unsigned long dim2001KuoInit[]        =   {   1   ,   3   ,   1   ,   3   ,   17  ,   27  ,   77  ,   203 ,   415 ,   489 ,   665 ,   21  ,   6453    ,   5069    ,   29629   ,0 };
            const unsigned long dim2002KuoInit[]        =   {   1   ,   1   ,   1   ,   9   ,   7   ,   3   ,   33  ,   209 ,   85  ,   851 ,   367 ,   783 ,   309 ,   14821   ,   1217    ,0 };
            const unsigned long dim2003KuoInit[]        =   {   1   ,   1   ,   1   ,   15  ,   5   ,   39  ,   15  ,   225 ,   241 ,   477 ,   999 ,   841 ,   4637    ,   5383    ,   18081   ,0 };
            const unsigned long dim2004KuoInit[]        =   {   1   ,   1   ,   5   ,   5   ,   19  ,   49  ,   101 ,   115 ,   87  ,   781 ,   815 ,   2405    ,   2579    ,   15043   ,   31067   ,0 };
            const unsigned long dim2005KuoInit[]        =   {   1   ,   1   ,   3   ,   5   ,   3   ,   57  ,   123 ,   191 ,   305 ,   999 ,   1091    ,   733 ,   3219    ,   891 ,   30495   ,0 };
            const unsigned long dim2006KuoInit[]        =   {   1   ,   1   ,   5   ,   3   ,   27  ,   17  ,   95  ,   121 ,   39  ,   41  ,   171 ,   97  ,   7363    ,   11279   ,   28039   ,0 };
            const unsigned long dim2007KuoInit[]        =   {   1   ,   3   ,   7   ,   5   ,   21  ,   9   ,   21  ,   147 ,   269 ,   475 ,   1825    ,   1459    ,   6337    ,   3259    ,   27731   ,0 };
            const unsigned long dim2008KuoInit[]        =   {   1   ,   1   ,   3   ,   5   ,   9   ,   17  ,   107 ,   171 ,   253 ,   71  ,   629 ,   1671    ,   2587    ,   11429   ,   8985    ,0 };
            const unsigned long dim2009KuoInit[]        =   {   1   ,   3   ,   3   ,   9   ,   25  ,   1   ,   63  ,   15  ,   173 ,   79  ,   1465    ,   2647    ,   6885    ,   13043   ,   24703   ,0 };
            const unsigned long dim2010KuoInit[]        =   {   1   ,   3   ,   1   ,   9   ,   1   ,   13  ,   3   ,   217 ,   49  ,   77  ,   1635    ,   3767    ,   173 ,   12697   ,   11837   ,0 };
            const unsigned long dim2011KuoInit[]        =   {   1   ,   3   ,   5   ,   13  ,   31  ,   7   ,   11  ,   181 ,   509 ,   601 ,   1267    ,   1857    ,   7871    ,   1703    ,   6091    ,0 };
            const unsigned long dim2012KuoInit[]        =   {   1   ,   1   ,   1   ,   11  ,   29  ,   61  ,   117 ,   151 ,   487 ,   185 ,   1899    ,   575 ,   7341    ,   3055    ,   26513   ,0 };
            const unsigned long dim2013KuoInit[]        =   {   1   ,   3   ,   7   ,   15  ,   1   ,   49  ,   87  ,   23  ,   483 ,   603 ,   1197    ,   2589    ,   1705    ,   5759    ,   19791   ,0 };
            const unsigned long dim2014KuoInit[]        =   {   1   ,   3   ,   3   ,   11  ,   1   ,   37  ,   71  ,   115 ,   59  ,   167 ,   1907    ,   2225    ,   5673    ,   14571   ,   19389   ,0 };
            const unsigned long dim2015KuoInit[]        =   {   1   ,   1   ,   3   ,   7   ,   15  ,   57  ,   33  ,   143 ,   45  ,   321 ,   779 ,   543 ,   3465    ,   4769    ,   19753   ,0 };
            const unsigned long dim2016KuoInit[]        =   {   1   ,   3   ,   5   ,   5   ,   31  ,   21  ,   21  ,   141 ,   197 ,   397 ,   1041    ,   317 ,   109 ,   9395    ,   7141    ,0 };
            const unsigned long dim2017KuoInit[]        =   {   1   ,   3   ,   7   ,   11  ,   1   ,   61  ,   45  ,   215 ,   405 ,   365 ,   1801    ,   2767    ,   6381    ,   12975   ,   41  ,0 };
            const unsigned long dim2018KuoInit[]        =   {   1   ,   3   ,   7   ,   7   ,   23  ,   13  ,   37  ,   193 ,   305 ,   883 ,   1427    ,   3859    ,   2049    ,   16065   ,   2471    ,0 };
            const unsigned long dim2019KuoInit[]        =   {   1   ,   3   ,   3   ,   3   ,   31  ,   3   ,   9   ,   169 ,   433 ,   187 ,   545 ,   985 ,   7699    ,   1183    ,   27773   ,0 };
            const unsigned long dim2020KuoInit[]        =   {   1   ,   3   ,   5   ,   15  ,   11  ,   15  ,   9   ,   171 ,   89  ,   711 ,   163 ,   3637    ,   5725    ,   5079    ,   13373   ,0 };
            const unsigned long dim2021KuoInit[]        =   {   1   ,   1   ,   7   ,   1   ,   1   ,   29  ,   123 ,   235 ,   229 ,   667 ,   1155    ,   3387    ,   129 ,   13889   ,   24547   ,0 };
            const unsigned long dim2022KuoInit[]        =   {   1   ,   3   ,   3   ,   7   ,   27  ,   15  ,   75  ,   117 ,   335 ,   821 ,   609 ,   3843    ,   79  ,   6787    ,   5605    ,0 };
            const unsigned long dim2023KuoInit[]        =   {   1   ,   1   ,   3   ,   15  ,   27  ,   47  ,   13  ,   163 ,   205 ,   773 ,   541 ,   1155    ,   4341    ,   14155   ,   6643    ,0 };
            const unsigned long dim2024KuoInit[]        =   {   1   ,   3   ,   3   ,   11  ,   31  ,   1   ,   67  ,   35  ,   443 ,   723 ,   1089    ,   2605    ,   4193    ,   6133    ,   31127   ,0 };
            const unsigned long dim2025KuoInit[]        =   {   1   ,   1   ,   1   ,   15  ,   27  ,   57  ,   11  ,   153 ,   199 ,   219 ,   1987    ,   2517    ,   1455    ,   6051    ,   9615    ,0 };
            const unsigned long dim2026KuoInit[]        =   {   1   ,   3   ,   5   ,   9   ,   5   ,   7   ,   79  ,   187 ,   93  ,   679 ,   1607    ,   799 ,   2753    ,   6513    ,   11371   ,0 };
            const unsigned long dim2027KuoInit[]        =   {   1   ,   3   ,   1   ,   15  ,   31  ,   55  ,   5   ,   175 ,   261 ,   545 ,   995 ,   2225    ,   2217    ,   2037    ,   14785   ,0 };
            const unsigned long dim2028KuoInit[]        =   {   1   ,   3   ,   5   ,   9   ,   29  ,   5   ,   101 ,   223 ,   241 ,   829 ,   1625    ,   3417    ,   1335    ,   10705   ,   531 ,0 };
            const unsigned long dim2029KuoInit[]        =   {   1   ,   3   ,   3   ,   1   ,   19  ,   15  ,   5   ,   203 ,   97  ,   165 ,   99  ,   2371    ,   3267    ,   13169   ,   22229   ,0 };
            const unsigned long dim2030KuoInit[]        =   {   1   ,   1   ,   1   ,   7   ,   11  ,   63  ,   113 ,   255 ,   145 ,   589 ,   1149    ,   683 ,   2619    ,   5569    ,   5933    ,0 };
            const unsigned long dim2031KuoInit[]        =   {   1   ,   1   ,   5   ,   9   ,   23  ,   41  ,   21  ,   217 ,   503 ,   975 ,   329 ,   3433    ,   6299    ,   5337    ,   199 ,0 };
            const unsigned long dim2032KuoInit[]        =   {   1   ,   1   ,   7   ,   13  ,   3   ,   39  ,   109 ,   223 ,   241 ,   521 ,   1071    ,   589 ,   1199    ,   13059   ,   24921   ,0 };
            const unsigned long dim2033KuoInit[]        =   {   1   ,   1   ,   3   ,   7   ,   11  ,   23  ,   27  ,   15  ,   107 ,   679 ,   87  ,   3259    ,   4369    ,   7365    ,   16293   ,0 };
            const unsigned long dim2034KuoInit[]        =   {   1   ,   3   ,   3   ,   11  ,   11  ,   5   ,   75  ,   111 ,   21  ,   587 ,   1707    ,   1915    ,   6617    ,   11143   ,   24389   ,0 };
            const unsigned long dim2035KuoInit[]        =   {   1   ,   1   ,   5   ,   9   ,   5   ,   59  ,   77  ,   3   ,   43  ,   583 ,   1601    ,   2183    ,   6209    ,   5409    ,   18175   ,0 };
            const unsigned long dim2036KuoInit[]        =   {   1   ,   1   ,   1   ,   1   ,   15  ,   1   ,   55  ,   95  ,   437 ,   731 ,   1559    ,   2421    ,   7441    ,   6567    ,   1549    ,0 };
            const unsigned long dim2037KuoInit[]        =   {   1   ,   3   ,   3   ,   11  ,   17  ,   21  ,   13  ,   61  ,   153 ,   81  ,   141 ,   3095    ,   5485    ,   14461   ,   8495    ,0 };
            const unsigned long dim2038KuoInit[]        =   {   1   ,   1   ,   1   ,   13  ,   25  ,   61  ,   87  ,   195 ,   243 ,   47  ,   103 ,   3451    ,   5583    ,   1611    ,   11939   ,0 };
            const unsigned long dim2039KuoInit[]        =   {   1   ,   1   ,   5   ,   7   ,   31  ,   53  ,   93  ,   101 ,   145 ,   917 ,   1669    ,   867 ,   6949    ,   16333   ,   12577   ,0 };
            const unsigned long dim2040KuoInit[]        =   {   1   ,   1   ,   5   ,   1   ,   5   ,   41  ,   35  ,   75  ,   331 ,   671 ,   1623    ,   1573    ,   4275    ,   9031    ,   9047    ,0 };
            const unsigned long dim2041KuoInit[]        =   {   1   ,   3   ,   5   ,   13  ,   15  ,   39  ,   29  ,   183 ,   477 ,   787 ,   1333    ,   2499    ,   2189    ,   8387    ,   2783    ,0 };
            const unsigned long dim2042KuoInit[]        =   {   1   ,   1   ,   3   ,   7   ,   19  ,   33  ,   117 ,   201 ,   177 ,   595 ,   69  ,   3341    ,   1841    ,   11091   ,   16745   ,0 };
            const unsigned long dim2043KuoInit[]        =   {   1   ,   3   ,   7   ,   11  ,   29  ,   59  ,   69  ,   119 ,   433 ,   595 ,   221 ,   1245    ,   7805    ,   1761    ,   27657   ,0 };
            const unsigned long dim2044KuoInit[]        =   {   1   ,   3   ,   5   ,   3   ,   23  ,   19  ,   75  ,   189 ,   279 ,   867 ,   113 ,   3817    ,   6093    ,   12237   ,   20297   ,0 };
            const unsigned long dim2045KuoInit[]        =   {   1   ,   1   ,   1   ,   15  ,   21  ,   39  ,   15  ,   249 ,   9   ,   997 ,   51  ,   1317    ,   1083    ,   13683   ,   5759    ,0 };
            const unsigned long dim2046KuoInit[]        =   {   1   ,   1   ,   5   ,   15  ,   29  ,   29  ,   25  ,   93  ,   465 ,   549 ,   283 ,   1673    ,   1745    ,   3913    ,   7931    ,0 };
            const unsigned long dim2047KuoInit[]        =   {   1   ,   3   ,   5   ,   1   ,   5   ,   23  ,   123 ,   217 ,   211 ,   215 ,   1999    ,   1229    ,   6969    ,   13481   ,   14795   ,0 };
            const unsigned long dim2048KuoInit[]        =   {   1   ,   1   ,   3   ,   5   ,   17  ,   35  ,   107 ,   163 ,   103 ,   807 ,   463 ,   2999    ,   4889    ,   12863   ,   24619   ,0 };
            const unsigned long dim2049KuoInit[]        =   {   1   ,   3   ,   1   ,   15  ,   5   ,   17  ,   121 ,   209 ,   49  ,   871 ,   1931    ,   345 ,   6887    ,   1357    ,   14693   ,0 };
            const unsigned long dim2050KuoInit[]        =   {   1   ,   1   ,   1   ,   1   ,   23  ,   63  ,   53  ,   35  ,   511 ,   261 ,   11  ,   283 ,   3283    ,   1963    ,   11703   ,0 };
            const unsigned long dim2051KuoInit[]        =   {   1   ,   3   ,   3   ,   13  ,   15  ,   41  ,   19  ,   39  ,   445 ,   21  ,   513 ,   2957    ,   2307    ,   11451   ,   6539    ,0 };
            const unsigned long dim2052KuoInit[]        =   {   1   ,   1   ,   3   ,   11  ,   19  ,   45  ,   7   ,   155 ,   387 ,   287 ,   347 ,   1557    ,   7407    ,   10885   ,   31829   ,0 };
            const unsigned long dim2053KuoInit[]        =   {   1   ,   3   ,   5   ,   13  ,   9   ,   53  ,   61  ,   99  ,   355 ,   449 ,   1353    ,   3135    ,   2329    ,   733 ,   22443   ,0 };
            const unsigned long dim2054KuoInit[]        =   {   1   ,   1   ,   3   ,   5   ,   27  ,   47  ,   43  ,   217 ,   113 ,   385 ,   903 ,   3015    ,   4169    ,   10705   ,   11351   ,0 };
            const unsigned long dim2055KuoInit[]        =   {   1   ,   3   ,   7   ,   11  ,   9   ,   33  ,   7   ,   167 ,   473 ,   465 ,   357 ,   2387    ,   2901    ,   12995   ,   16871   ,0 };
            const unsigned long dim2056KuoInit[]        =   {   1   ,   3   ,   7   ,   3   ,   29  ,   35  ,   85  ,   201 ,   311 ,   581 ,   491 ,   2977    ,   7433    ,   14615   ,   29995   ,0 };
            const unsigned long dim2057KuoInit[]        =   {   1   ,   3   ,   1   ,   15  ,   5   ,   5   ,   121 ,   17  ,   93  ,   847 ,   835 ,   3885    ,   2437    ,   2237    ,   15669   ,0 };
            const unsigned long dim2058KuoInit[]        =   {   1   ,   3   ,   7   ,   5   ,   29  ,   39  ,   57  ,   241 ,   247 ,   847 ,   311 ,   3115    ,   7605    ,   3117    ,   30677   ,0 };
            const unsigned long dim2059KuoInit[]        =   {   1   ,   3   ,   1   ,   13  ,   21  ,   27  ,   123 ,   77  ,   511 ,   319 ,   1699    ,   709 ,   7021    ,   14377   ,   9601    ,0 };
            const unsigned long dim2060KuoInit[]        =   {   1   ,   3   ,   3   ,   5   ,   31  ,   1   ,   49  ,   249 ,   271 ,   595 ,   1069    ,   2307    ,   567 ,   13145   ,   8779    ,0 };
            const unsigned long dim2061KuoInit[]        =   {   1   ,   3   ,   3   ,   3   ,   29  ,   25  ,   55  ,   63  ,   289 ,   617 ,   1209    ,   3949    ,   481 ,   13495   ,   16659   ,0 };
            const unsigned long dim2062KuoInit[]        =   {   1   ,   3   ,   3   ,   3   ,   5   ,   19  ,   47  ,   125 ,   323 ,   917 ,   1875    ,   1683    ,   7451    ,   16145   ,   3029    ,0 };
            const unsigned long dim2063KuoInit[]        =   {   1   ,   1   ,   7   ,   5   ,   27  ,   33  ,   3   ,   133 ,   477 ,   457 ,   803 ,   1141    ,   3965    ,   9127    ,   21173   ,0 };
            const unsigned long dim2064KuoInit[]        =   {   1   ,   3   ,   5   ,   7   ,   19  ,   61  ,   17  ,   27  ,   297 ,   995 ,   1119    ,   2555    ,   6741    ,   3233    ,   13735   ,0 };
            const unsigned long dim2065KuoInit[]        =   {   1   ,   3   ,   7   ,   11  ,   11  ,   21  ,   21  ,   143 ,   319 ,   143 ,   675 ,   187 ,   6909    ,   7145    ,   27029   ,0 };
            const unsigned long dim2066KuoInit[]        =   {   1   ,   3   ,   3   ,   1   ,   29  ,   5   ,   1   ,   65  ,   123 ,   801 ,   933 ,   3105    ,   1163    ,   15703   ,   9567    ,0 };
            const unsigned long dim2067KuoInit[]        =   {   1   ,   1   ,   3   ,   11  ,   3   ,   15  ,   11  ,   109 ,   117 ,   1011    ,   1407    ,   3287    ,   5799    ,   147 ,   32641   ,0 };
            const unsigned long dim2068KuoInit[]        =   {   1   ,   1   ,   1   ,   5   ,   29  ,   25  ,   63  ,   39  ,   139 ,   379 ,   1887    ,   417 ,   2965    ,   4409    ,   19149   ,0 };
            const unsigned long dim2069KuoInit[]        =   {   1   ,   1   ,   1   ,   13  ,   31  ,   25  ,   123 ,   187 ,   311 ,   397 ,   1467    ,   15  ,   4253    ,   14565   ,   32609   ,0 };
            const unsigned long dim2070KuoInit[]        =   {   1   ,   3   ,   5   ,   7   ,   25  ,   5   ,   13  ,   25  ,   489 ,   837 ,   865 ,   527 ,   4965    ,   2567    ,   13505   ,0 };
            const unsigned long dim2071KuoInit[]        =   {   1   ,   1   ,   3   ,   9   ,   3   ,   15  ,   59  ,   115 ,   339 ,   459 ,   2011    ,   3311    ,   4067    ,   3437    ,   21435   ,0 };
            const unsigned long dim2072KuoInit[]        =   {   1   ,   1   ,   5   ,   3   ,   17  ,   35  ,   49  ,   11  ,   217 ,   401 ,   385 ,   569 ,   1727    ,   9927    ,   27229   ,0 };
            const unsigned long dim2073KuoInit[]        =   {   1   ,   3   ,   1   ,   5   ,   27  ,   43  ,   21  ,   11  ,   459 ,   555 ,   237 ,   87  ,   7531    ,   16117   ,   23625   ,0 };
            const unsigned long dim2074KuoInit[]        =   {   1   ,   3   ,   3   ,   5   ,   15  ,   57  ,   93  ,   35  ,   31  ,   257 ,   77  ,   3437    ,   3665    ,   3137    ,   2477    ,0 };
            const unsigned long dim2075KuoInit[]        =   {   1   ,   3   ,   5   ,   9   ,   1   ,   43  ,   91  ,   55  ,   53  ,   265 ,   1495    ,   551 ,   4729    ,   7839    ,   23475   ,0 };
            const unsigned long dim2076KuoInit[]        =   {   1   ,   1   ,   7   ,   11  ,   11  ,   45  ,   9   ,   27  ,   485 ,   397 ,   925 ,   3601    ,   4419    ,   4563    ,   6443    ,0 };
            const unsigned long dim2077KuoInit[]        =   {   1   ,   3   ,   5   ,   3   ,   15  ,   35  ,   103 ,   239 ,   193 ,   831 ,   559 ,   1299    ,   4939    ,   8051    ,   6899    ,0 };
            const unsigned long dim2078KuoInit[]        =   {   1   ,   3   ,   3   ,   11  ,   17  ,   47  ,   77  ,   13  ,   227 ,   935 ,   1973    ,   1789    ,   4357    ,   5135    ,   27643   ,0 };
            const unsigned long dim2079KuoInit[]        =   {   1   ,   1   ,   3   ,   7   ,   11  ,   51  ,   109 ,   143 ,   101 ,   131 ,   565 ,   3503    ,   5649    ,   10007   ,   6283    ,0 };
            const unsigned long dim2080KuoInit[]        =   {   1   ,   1   ,   3   ,   15  ,   19  ,   5   ,   15  ,   241 ,   399 ,   341 ,   943 ,   587 ,   4757    ,   13941   ,   7571    ,0 };
            const unsigned long dim2081KuoInit[]        =   {   1   ,   3   ,   3   ,   3   ,   21  ,   9   ,   119 ,   119 ,   445 ,   349 ,   1517    ,   17  ,   3593    ,   11983   ,   9713    ,0 };
            const unsigned long dim2082KuoInit[]        =   {   1   ,   1   ,   3   ,   3   ,   23  ,   55  ,   21  ,   251 ,   191 ,   585 ,   1619    ,   431 ,   4161    ,   5667    ,   17703   ,0 };
            const unsigned long dim2083KuoInit[]        =   {   1   ,   3   ,   3   ,   3   ,   1   ,   57  ,   65  ,   231 ,   67  ,   531 ,   1403    ,   3883    ,   4463    ,   12795   ,   4427    ,0 };
            const unsigned long dim2084KuoInit[]        =   {   1   ,   1   ,   5   ,   15  ,   19  ,   7   ,   9   ,   163 ,   487 ,   455 ,   1663    ,   3141    ,   5975    ,   11001   ,   24781   ,0 };
            const unsigned long dim2085KuoInit[]        =   {   1   ,   1   ,   5   ,   1   ,   29  ,   19  ,   45  ,   235 ,   371 ,   935 ,   1757    ,   187 ,   7729    ,   6241    ,   27171   ,0 };
            const unsigned long dim2086KuoInit[]        =   {   1   ,   1   ,   1   ,   1   ,   15  ,   39  ,   39  ,   171 ,   145 ,   413 ,   1545    ,   3733    ,   1815    ,   9443    ,   32705   ,0 };
            const unsigned long dim2087KuoInit[]        =   {   1   ,   1   ,   3   ,   9   ,   31  ,   1   ,   87  ,   11  ,   219 ,   913 ,   131 ,   3625    ,   7927    ,   9945    ,   2595    ,0 };
            const unsigned long dim2088KuoInit[]        =   {   1   ,   3   ,   3   ,   13  ,   21  ,   15  ,   39  ,   29  ,   267 ,   275 ,   973 ,   701 ,   1057    ,   13641   ,   22149   ,0 };
            const unsigned long dim2089KuoInit[]        =   {   1   ,   3   ,   3   ,   5   ,   3   ,   33  ,   81  ,   29  ,   85  ,   515 ,   1719    ,   3967    ,   5145    ,   12921   ,   8303    ,0 };
            const unsigned long dim2090KuoInit[]        =   {   1   ,   3   ,   7   ,   11  ,   17  ,   23  ,   113 ,   129 ,   395 ,   153 ,   427 ,   1113    ,   8157    ,   9087    ,   20471   ,0 };
            const unsigned long dim2091KuoInit[]        =   {   1   ,   1   ,   5   ,   9   ,   23  ,   5   ,   119 ,   67  ,   215 ,   581 ,   1249    ,   3949    ,   731 ,   10387   ,   27979   ,0 };
            const unsigned long dim2092KuoInit[]        =   {   1   ,   1   ,   5   ,   5   ,   5   ,   53  ,   7   ,   39  ,   51  ,   519 ,   225 ,   119 ,   4335    ,   11723   ,   27971   ,0 };
            const unsigned long dim2093KuoInit[]        =   {   1   ,   3   ,   5   ,   3   ,   5   ,   49  ,   105 ,   29  ,   443 ,   457 ,   191 ,   2679    ,   5803    ,   12017   ,   2205    ,0 };
            const unsigned long dim2094KuoInit[]        =   {   1   ,   1   ,   3   ,   1   ,   23  ,   31  ,   11  ,   199 ,   479 ,   231 ,   431 ,   1989    ,   705 ,   6557    ,   27763   ,0 };
            const unsigned long dim2095KuoInit[]        =   {   1   ,   1   ,   3   ,   1   ,   27  ,   9   ,   39  ,   105 ,   149 ,   1017    ,   79  ,   1575    ,   6365    ,   839 ,   27601   ,0 };
            const unsigned long dim2096KuoInit[]        =   {   1   ,   1   ,   3   ,   7   ,   1   ,   29  ,   61  ,   129 ,   253 ,   541 ,   1953    ,   351 ,   2895    ,   14385   ,   23685   ,0 };
            const unsigned long dim2097KuoInit[]        =   {   1   ,   3   ,   1   ,   7   ,   19  ,   15  ,   55  ,   133 ,   195 ,   881 ,   311 ,   2021    ,   7935    ,   4513    ,   31065   ,0 };
            const unsigned long dim2098KuoInit[]        =   {   1   ,   1   ,   5   ,   13  ,   23  ,   55  ,   53  ,   169 ,   79  ,   707 ,   1415    ,   3999    ,   299 ,   6689    ,   10219   ,0 };
            const unsigned long dim2099KuoInit[]        =   {   1   ,   3   ,   3   ,   7   ,   7   ,   47  ,   77  ,   209 ,   467 ,   39  ,   291 ,   3769    ,   1295    ,   12407   ,   2205    ,0 };
            const unsigned long dim2100KuoInit[]        =   {   1   ,   1   ,   3   ,   9   ,   13  ,   3   ,   63  ,   241 ,   251 ,   939 ,   1391    ,   2053    ,   719 ,   13739   ,   12167   ,0 };
            const unsigned long dim2101KuoInit[]        =   {   1   ,   1   ,   7   ,   9   ,   21  ,   31  ,   37  ,   11  ,   187 ,   201 ,   1425    ,   249 ,   2475    ,   12421   ,   14443   ,0 };
            const unsigned long dim2102KuoInit[]        =   {   1   ,   1   ,   3   ,   9   ,   31  ,   51  ,   53  ,   119 ,   19  ,   13  ,   1513    ,   1115    ,   1265    ,   15707   ,   22795   ,0 };
            const unsigned long dim2103KuoInit[]        =   {   1   ,   3   ,   5   ,   9   ,   23  ,   13  ,   69  ,   247 ,   103 ,   773 ,   997 ,   703 ,   2921    ,   3165    ,   3027    ,0 };
            const unsigned long dim2104KuoInit[]        =   {   1   ,   3   ,   7   ,   7   ,   15  ,   27  ,   81  ,   221 ,   259 ,   405 ,   391 ,   2635    ,   1747    ,   7743    ,   28833   ,0 };
            const unsigned long dim2105KuoInit[]        =   {   1   ,   1   ,   3   ,   9   ,   29  ,   47  ,   23  ,   127 ,   49  ,   81  ,   543 ,   919 ,   5517    ,   9573    ,   1523    ,0 };
            const unsigned long dim2106KuoInit[]        =   {   1   ,   3   ,   3   ,   9   ,   11  ,   37  ,   123 ,   199 ,   343 ,   697 ,   293 ,   3123    ,   4601    ,   14209   ,   30425   ,0 };
            const unsigned long dim2107KuoInit[]        =   {   1   ,   1   ,   5   ,   7   ,   3   ,   31  ,   127 ,   13  ,   191 ,   509 ,   1999    ,   3379    ,   2365    ,   10307   ,   26693   ,0 };
            const unsigned long dim2108KuoInit[]        =   {   1   ,   3   ,   7   ,   3   ,   3   ,   41  ,   71  ,   153 ,   119 ,   9   ,   1601    ,   3839    ,   369 ,   16241   ,   19115   ,0 };
            const unsigned long dim2109KuoInit[]        =   {   1   ,   1   ,   5   ,   5   ,   3   ,   17  ,   7   ,   65  ,   191 ,   143 ,   127 ,   1973    ,   6885    ,   13615   ,   26289   ,0 };
            const unsigned long dim2110KuoInit[]        =   {   1   ,   1   ,   3   ,   5   ,   13  ,   5   ,   87  ,   17  ,   445 ,   975 ,   1527    ,   253 ,   5983    ,   7095    ,   21271   ,0 };
            const unsigned long dim2111KuoInit[]        =   {   1   ,   3   ,   1   ,   3   ,   7   ,   49  ,   29  ,   19  ,   37  ,   695 ,   603 ,   3153    ,   4065    ,   5421    ,   3285    ,0 };
            const unsigned long dim2112KuoInit[]        =   {   1   ,   1   ,   5   ,   13  ,   25  ,   47  ,   43  ,   35  ,   205 ,   629 ,   1255    ,   2843    ,   7533    ,   751 ,   10119   ,0 };
            const unsigned long dim2113KuoInit[]        =   {   1   ,   3   ,   1   ,   11  ,   13  ,   27  ,   127 ,   221 ,   67  ,   523 ,   217 ,   1493    ,   1793    ,   10871   ,   14289   ,0 };
            const unsigned long dim2114KuoInit[]        =   {   1   ,   1   ,   1   ,   5   ,   21  ,   47  ,   45  ,   73  ,   479 ,   653 ,   1553    ,   695 ,   3131    ,   8773    ,   20303   ,0 };
            const unsigned long dim2115KuoInit[]        =   {   1   ,   3   ,   1   ,   1   ,   19  ,   31  ,   25  ,   207 ,   361 ,   961 ,   1595    ,   1807    ,   3649    ,   1029    ,   31579   ,0 };
            const unsigned long dim2116KuoInit[]        =   {   1   ,   1   ,   1   ,   1   ,   27  ,   23  ,   109 ,   111 ,   477 ,   233 ,   1305    ,   51  ,   4517    ,   14867   ,   4075    ,0 };
            const unsigned long dim2117KuoInit[]        =   {   1   ,   1   ,   7   ,   11  ,   21  ,   5   ,   51  ,   9   ,   131 ,   671 ,   191 ,   4045    ,   7773    ,   5037    ,   23229   ,0 };
            const unsigned long dim2118KuoInit[]        =   {   1   ,   1   ,   5   ,   7   ,   3   ,   57  ,   75  ,   141 ,   439 ,   269 ,   99  ,   1595    ,   3597    ,   10183   ,   19849   ,0 };
            const unsigned long dim2119KuoInit[]        =   {   1   ,   1   ,   3   ,   3   ,   17  ,   61  ,   81  ,   75  ,   471 ,   1023    ,   1045    ,   849 ,   3741    ,   12803   ,   10735   ,0 };
            const unsigned long dim2120KuoInit[]        =   {   1   ,   1   ,   7   ,   11  ,   3   ,   33  ,   79  ,   113 ,   335 ,   245 ,   1807    ,   1927    ,   7381    ,   6405    ,   20035   ,0 };
            const unsigned long dim2121KuoInit[]        =   {   1   ,   3   ,   1   ,   11  ,   5   ,   37  ,   31  ,   177 ,   233 ,   937 ,   1655    ,   511 ,   1109    ,   5083    ,   31325   ,0 };
            const unsigned long dim2122KuoInit[]        =   {   1   ,   3   ,   5   ,   15  ,   15  ,   35  ,   125 ,   105 ,   313 ,   553 ,   259 ,   1433    ,   327 ,   2079    ,   5279    ,0 };
            const unsigned long dim2123KuoInit[]        =   {   1   ,   3   ,   1   ,   1   ,   25  ,   57  ,   121 ,   189 ,   409 ,   459 ,   749 ,   3159    ,   6389    ,   4141    ,   16463   ,0 };
            const unsigned long dim2124KuoInit[]        =   {   1   ,   1   ,   7   ,   15  ,   7   ,   21  ,   113 ,   139 ,   241 ,   617 ,   1385    ,   1387    ,   145 ,   6605    ,   2057    ,0 };
            const unsigned long dim2125KuoInit[]        =   {   1   ,   3   ,   3   ,   11  ,   17  ,   33  ,   89  ,   17  ,   341 ,   369 ,   719 ,   2481    ,   7589    ,   16283   ,   28157   ,0 };
            const unsigned long dim2126KuoInit[]        =   {   1   ,   3   ,   7   ,   1   ,   11  ,   17  ,   31  ,   135 ,   333 ,   335 ,   197 ,   2401    ,   7637    ,   703 ,   3001    ,0 };
            const unsigned long dim2127KuoInit[]        =   {   1   ,   3   ,   5   ,   11  ,   17  ,   31  ,   33  ,   249 ,   315 ,   475 ,   1743    ,   2429    ,   4923    ,   2779    ,   25677   ,0 };
            const unsigned long dim2128KuoInit[]        =   {   1   ,   1   ,   5   ,   5   ,   11  ,   29  ,   111 ,   103 ,   143 ,   779 ,   1437    ,   3759    ,   5827    ,   3129    ,   26961   ,0 };
            const unsigned long dim2129KuoInit[]        =   {   1   ,   3   ,   5   ,   3   ,   25  ,   1   ,   109 ,   117 ,   159 ,   53  ,   1291    ,   1787    ,   7851    ,   14363   ,   15271   ,0 };
            const unsigned long dim2130KuoInit[]        =   {   1   ,   1   ,   1   ,   9   ,   9   ,   1   ,   31  ,   131 ,   163 ,   723 ,   1023    ,   27  ,   853 ,   2623    ,   16859   ,0 };
            const unsigned long dim2131KuoInit[]        =   {   1   ,   1   ,   3   ,   11  ,   23  ,   13  ,   91  ,   87  ,   141 ,   111 ,   1677    ,   1483    ,   1461    ,   15981   ,   29417   ,0 };
            const unsigned long dim2132KuoInit[]        =   {   1   ,   3   ,   7   ,   9   ,   17  ,   17  ,   115 ,   43  ,   397 ,   683 ,   2017    ,   985 ,   3703    ,   16135   ,   3661    ,0 };
            const unsigned long dim2133KuoInit[]        =   {   1   ,   1   ,   7   ,   13  ,   17  ,   45  ,   71  ,   149 ,   317 ,   559 ,   279 ,   2441    ,   2323    ,   16071   ,   29709   ,0 };
            const unsigned long dim2134KuoInit[]        =   {   1   ,   3   ,   5   ,   11  ,   31  ,   33  ,   105 ,   127 ,   375 ,   1001    ,   115 ,   1429    ,   7003    ,   13369   ,   23281   ,0 };
            const unsigned long dim2135KuoInit[]        =   {   1   ,   1   ,   5   ,   7   ,   21  ,   7   ,   27  ,   105 ,   99  ,   479 ,   849 ,   1361    ,   4935    ,   14831   ,   29  ,0 };
            const unsigned long dim2136KuoInit[]        =   {   1   ,   3   ,   7   ,   9   ,   15  ,   53  ,   63  ,   109 ,   171 ,   323 ,   289 ,   2937    ,   3177    ,   4615    ,   10711   ,0 };
            const unsigned long dim2137KuoInit[]        =   {   1   ,   1   ,   1   ,   1   ,   27  ,   41  ,   29  ,   101 ,   209 ,   673 ,   1597    ,   3743    ,   827 ,   4025    ,   27267   ,0 };
            const unsigned long dim2138KuoInit[]        =   {   1   ,   3   ,   7   ,   13  ,   25  ,   17  ,   19  ,   219 ,   233 ,   769 ,   909 ,   3585    ,   55  ,   13349   ,   13685   ,0 };
            const unsigned long dim2139KuoInit[]        =   {   1   ,   1   ,   1   ,   15  ,   17  ,   57  ,   63  ,   181 ,   339 ,   319 ,   655 ,   699 ,   1633    ,   8181    ,   23039   ,0 };
            const unsigned long dim2140KuoInit[]        =   {   1   ,   1   ,   3   ,   1   ,   25  ,   11  ,   61  ,   195 ,   311 ,   723 ,   365 ,   2791    ,   1281    ,   9113    ,   26741   ,0 };
            const unsigned long dim2141KuoInit[]        =   {   1   ,   1   ,   7   ,   11  ,   19  ,   43  ,   65  ,   37  ,   89  ,   847 ,   951 ,   851 ,   4441    ,   6681    ,   917 ,0 };
            const unsigned long dim2142KuoInit[]        =   {   1   ,   3   ,   5   ,   13  ,   1   ,   11  ,   55  ,   29  ,   419 ,   769 ,   815 ,   1487    ,   2151    ,   7981    ,   19509   ,0 };
            const unsigned long dim2143KuoInit[]        =   {   1   ,   1   ,   5   ,   13  ,   11  ,   55  ,   127 ,   19  ,   419 ,   679 ,   119 ,   1417    ,   1717    ,   5849    ,   6365    ,0 };
            const unsigned long dim2144KuoInit[]        =   {   1   ,   1   ,   7   ,   5   ,   29  ,   5   ,   9   ,   241 ,   405 ,   253 ,   1015    ,   2027    ,   7695    ,   12003   ,   659 ,0 };
            const unsigned long dim2145KuoInit[]        =   {   1   ,   3   ,   7   ,   5   ,   27  ,   63  ,   87  ,   221 ,   219 ,   355 ,   941 ,   3901    ,   6535    ,   4715    ,   3571    ,0 };
            const unsigned long dim2146KuoInit[]        =   {   1   ,   3   ,   3   ,   9   ,   13  ,   1   ,   113 ,   21  ,   71  ,   353 ,   1063    ,   1447    ,   2157    ,   5195    ,   21311   ,0 };
            const unsigned long dim2147KuoInit[]        =   {   1   ,   1   ,   3   ,   15  ,   29  ,   1   ,   7   ,   223 ,   43  ,   605 ,   1063    ,   67  ,   2529    ,   1125    ,   16933   ,0 };
            const unsigned long dim2148KuoInit[]        =   {   1   ,   3   ,   1   ,   15  ,   29  ,   37  ,   51  ,   239 ,   117 ,   351 ,   1953    ,   1933    ,   6371    ,   13467   ,   8265    ,0 };
            const unsigned long dim2149KuoInit[]        =   {   1   ,   1   ,   3   ,   5   ,   7   ,   25  ,   19  ,   221 ,   409 ,   121 ,   89  ,   219 ,   105 ,   4213    ,   1903    ,0 };
            const unsigned long dim2150KuoInit[]        =   {   1   ,   3   ,   7   ,   1   ,   13  ,   3   ,   79  ,   123 ,   423 ,   589 ,   1085    ,   3387    ,   6117    ,   8749    ,   8947    ,0 };
            const unsigned long dim2151KuoInit[]        =   {   1   ,   3   ,   7   ,   9   ,   21  ,   49  ,   1   ,   39  ,   461 ,   115 ,   1131    ,   719 ,   3709    ,   1495    ,   23871   ,0 };
            const unsigned long dim2152KuoInit[]        =   {   1   ,   1   ,   7   ,   5   ,   23  ,   41  ,   107 ,   225 ,   499 ,   721 ,   1977    ,   2063    ,   1825    ,   9589    ,   22137   ,0 };
            const unsigned long dim2153KuoInit[]        =   {   1   ,   1   ,   5   ,   7   ,   15  ,   53  ,   33  ,   145 ,   179 ,   109 ,   425 ,   2729    ,   7637    ,   6805    ,   26223   ,0 };
            const unsigned long dim2154KuoInit[]        =   {   1   ,   1   ,   5   ,   13  ,   1   ,   1   ,   31  ,   117 ,   25  ,   427 ,   681 ,   2991    ,   7681    ,   15927   ,   20913   ,0 };
            const unsigned long dim2155KuoInit[]        =   {   1   ,   1   ,   5   ,   13  ,   31  ,   53  ,   23  ,   71  ,   397 ,   661 ,   131 ,   2815    ,   2351    ,   15881   ,   10321   ,0 };
            const unsigned long dim2156KuoInit[]        =   {   1   ,   3   ,   1   ,   15  ,   19  ,   41  ,   39  ,   57  ,   349 ,   761 ,   1999    ,   1151    ,   201 ,   1025    ,   18821   ,0 };
            const unsigned long dim2157KuoInit[]        =   {   1   ,   1   ,   3   ,   3   ,   17  ,   43  ,   73  ,   163 ,   187 ,   913 ,   31  ,   81  ,   4829    ,   6993    ,   32175   ,0 };
            const unsigned long dim2158KuoInit[]        =   {   1   ,   3   ,   1   ,   15  ,   19  ,   41  ,   79  ,   13  ,   343 ,   415 ,   863 ,   1943    ,   2231    ,   11353   ,   15631   ,0 };
            const unsigned long dim2159KuoInit[]        =   {   1   ,   3   ,   1   ,   15  ,   29  ,   35  ,   101 ,   23  ,   111 ,   937 ,   833 ,   2369    ,   3503    ,   6605    ,   14279   ,0 };
            const unsigned long dim2160KuoInit[]        =   {   1   ,   3   ,   7   ,   15  ,   27  ,   61  ,   43  ,   211 ,   259 ,   583 ,   1529    ,   1013    ,   8005    ,   7717    ,   1085    ,0 };
            const unsigned long dim2161KuoInit[]        =   {   1   ,   1   ,   7   ,   15  ,   23  ,   39  ,   17  ,   201 ,   465 ,   289 ,   1051    ,   33  ,   7403    ,   7971    ,   24905   ,0 };
            const unsigned long dim2162KuoInit[]        =   {   1   ,   1   ,   5   ,   13  ,   23  ,   41  ,   5   ,   239 ,   9   ,   23  ,   1353    ,   2005    ,   2513    ,   8707    ,   3893    ,0 };
            const unsigned long dim2163KuoInit[]        =   {   1   ,   3   ,   1   ,   7   ,   7   ,   41  ,   73  ,   177 ,   351 ,   297 ,   1057    ,   3867    ,   2481    ,   15751   ,   25231   ,0 };
            const unsigned long dim2164KuoInit[]        =   {   1   ,   1   ,   3   ,   1   ,   17  ,   59  ,   45  ,   171 ,   367 ,   575 ,   1085    ,   953 ,   3467    ,   8435    ,   6345    ,0 };
            const unsigned long dim2165KuoInit[]        =   {   1   ,   3   ,   1   ,   9   ,   5   ,   21  ,   125 ,   137 ,   415 ,   101 ,   121 ,   3499    ,   6671    ,   15541   ,   14081   ,0 };
            const unsigned long dim2166KuoInit[]        =   {   1   ,   3   ,   5   ,   15  ,   15  ,   19  ,   111 ,   235 ,   53  ,   119 ,   1307    ,   583 ,   1695    ,   211 ,   5567    ,0 };
            const unsigned long dim2167KuoInit[]        =   {   1   ,   1   ,   5   ,   9   ,   7   ,   13  ,   119 ,   109 ,   273 ,   677 ,   2001    ,   3307    ,   7999    ,   11769   ,   10069   ,0 };
            const unsigned long dim2168KuoInit[]        =   {   1   ,   1   ,   1   ,   15  ,   11  ,   15  ,   57  ,   233 ,   77  ,   815 ,   31  ,   669 ,   6365    ,   5085    ,   13583   ,0 };
            const unsigned long dim2169KuoInit[]        =   {   1   ,   3   ,   7   ,   1   ,   31  ,   9   ,   101 ,   33  ,   13  ,   699 ,   637 ,   1159    ,   5101    ,   13805   ,   32651   ,0 };
            const unsigned long dim2170KuoInit[]        =   {   1   ,   1   ,   7   ,   13  ,   15  ,   9   ,   125 ,   71  ,   241 ,   147 ,   529 ,   1863    ,   1551    ,   4981    ,   30389   ,0 };
            const unsigned long dim2171KuoInit[]        =   {   1   ,   1   ,   7   ,   13  ,   11  ,   35  ,   11  ,   5   ,   473 ,   55  ,   1589    ,   2835    ,   3771    ,   11433   ,   11099   ,0 };
            const unsigned long dim2172KuoInit[]        =   {   1   ,   1   ,   7   ,   15  ,   11  ,   11  ,   35  ,   207 ,   267 ,   289 ,   999 ,   3183    ,   3827    ,   6617    ,   29015   ,0 };
            const unsigned long dim2173KuoInit[]        =   {   1   ,   1   ,   3   ,   1   ,   3   ,   11  ,   119 ,   223 ,   255 ,   319 ,   1823    ,   1301    ,   4401    ,   12827   ,   889 ,0 };
            const unsigned long dim2174KuoInit[]        =   {   1   ,   3   ,   5   ,   1   ,   1   ,   5   ,   5   ,   109 ,   27  ,   963 ,   727 ,   335 ,   6295    ,   9775    ,   32681   ,0 };
            const unsigned long dim2175KuoInit[]        =   {   1   ,   3   ,   7   ,   15  ,   29  ,   47  ,   55  ,   253 ,   281 ,   303 ,   857 ,   903 ,   2977    ,   14777   ,   31689   ,0 };
            const unsigned long dim2176KuoInit[]        =   {   1   ,   3   ,   1   ,   7   ,   17  ,   49  ,   105 ,   155 ,   467 ,   467 ,   1287    ,   1281    ,   3429    ,   3195    ,   15297   ,0 };
            const unsigned long dim2177KuoInit[]        =   {   1   ,   3   ,   1   ,   1   ,   27  ,   31  ,   17  ,   201 ,   23  ,   897 ,   1991    ,   2661    ,   645 ,   16127   ,   30857   ,0 };
            const unsigned long dim2178KuoInit[]        =   {   1   ,   3   ,   3   ,   11  ,   9   ,   15  ,   85  ,   183 ,   31  ,   179 ,   779 ,   171 ,   5989    ,   7253    ,   15077   ,0 };
            const unsigned long dim2179KuoInit[]        =   {   1   ,   1   ,   1   ,   5   ,   19  ,   23  ,   37  ,   39  ,   281 ,   707 ,   17  ,   2707    ,   289 ,   12081   ,   4381    ,0 };
            const unsigned long dim2180KuoInit[]        =   {   1   ,   1   ,   7   ,   9   ,   21  ,   27  ,   115 ,   179 ,   99  ,   277 ,   331 ,   3407    ,   2057    ,   13893   ,   11501   ,0 };
            const unsigned long dim2181KuoInit[]        =   {   1   ,   3   ,   1   ,   9   ,   31  ,   51  ,   51  ,   25  ,   319 ,   319 ,   27  ,   2161    ,   89  ,   3421    ,   18273   ,0 };
            const unsigned long dim2182KuoInit[]        =   {   1   ,   1   ,   7   ,   15  ,   13  ,   21  ,   11  ,   189 ,   63  ,   15  ,   363 ,   1951    ,   5751    ,   5797    ,   22123   ,0 };
            const unsigned long dim2183KuoInit[]        =   {   1   ,   1   ,   3   ,   1   ,   25  ,   47  ,   99  ,   77  ,   77  ,   595 ,   1473    ,   1653    ,   6751    ,   5005    ,   29051   ,0 };
            const unsigned long dim2184KuoInit[]        =   {   1   ,   1   ,   3   ,   7   ,   25  ,   47  ,   31  ,   13  ,   311 ,   269 ,   1095    ,   1359    ,   305 ,   6811    ,   16455   ,0 };
            const unsigned long dim2185KuoInit[]        =   {   1   ,   1   ,   5   ,   3   ,   13  ,   63  ,   87  ,   13  ,   221 ,   459 ,   1177    ,   891 ,   5781    ,   13803   ,   18795   ,0 };
            const unsigned long dim2186KuoInit[]        =   {   1   ,   1   ,   7   ,   3   ,   23  ,   13  ,   25  ,   115 ,   47  ,   903 ,   1551    ,   2887    ,   229 ,   16067   ,   9511    ,0 };
            const unsigned long dim2187KuoInit[]        =   {   1   ,   1   ,   7   ,   15  ,   23  ,   15  ,   7   ,   65  ,   27  ,   677 ,   1503    ,   125 ,   7739    ,   14655   ,   29577   ,0 };
            const unsigned long dim2188KuoInit[]        =   {   1   ,   1   ,   1   ,   5   ,   9   ,   23  ,   93  ,   187 ,   205 ,   915 ,   599 ,   2105    ,   7767    ,   2411    ,   15045   ,0 };
            const unsigned long dim2189KuoInit[]        =   {   1   ,   3   ,   5   ,   1   ,   21  ,   45  ,   17  ,   57  ,   171 ,   491 ,   1465    ,   2013    ,   3253    ,   5517    ,   19113   ,0 };
            const unsigned long dim2190KuoInit[]        =   {   1   ,   1   ,   7   ,   5   ,   5   ,   25  ,   87  ,   247 ,   301 ,   385 ,   1655    ,   1351    ,   6307    ,   14695   ,   29745   ,0 };
            const unsigned long dim2191KuoInit[]        =   {   1   ,   3   ,   7   ,   3   ,   31  ,   25  ,   113 ,   155 ,   123 ,   755 ,   649 ,   1809    ,   1703    ,   875 ,   5235    ,0 };
            const unsigned long dim2192KuoInit[]        =   {   1   ,   1   ,   1   ,   7   ,   7   ,   43  ,   121 ,   47  ,   499 ,   63  ,   449 ,   2039    ,   835 ,   2883    ,   2287    ,0 };
            const unsigned long dim2193KuoInit[]        =   {   1   ,   3   ,   7   ,   1   ,   9   ,   11  ,   99  ,   169 ,   305 ,   469 ,   157 ,   2043    ,   4851    ,   3367    ,   8689    ,0 };
            const unsigned long dim2194KuoInit[]        =   {   1   ,   1   ,   1   ,   11  ,   17  ,   15  ,   41  ,   209 ,   377 ,   81  ,   335 ,   1685    ,   143 ,   1513    ,   28465   ,0 };
            const unsigned long dim2195KuoInit[]        =   {   1   ,   3   ,   1   ,   15  ,   9   ,   45  ,   83  ,   123 ,   347 ,   437 ,   783 ,   3561    ,   6003    ,   9329    ,   4729    ,0 };
            const unsigned long dim2196KuoInit[]        =   {   1   ,   1   ,   1   ,   1   ,   13  ,   33  ,   109 ,   217 ,   403 ,   539 ,   1393    ,   1069    ,   235 ,   11261   ,   13423   ,0 };
            const unsigned long dim2197KuoInit[]        =   {   1   ,   1   ,   5   ,   13  ,   27  ,   1   ,   81  ,   139 ,   421 ,   301 ,   55  ,   1883    ,   1525    ,   6029    ,   1879    ,0 };
            const unsigned long dim2198KuoInit[]        =   {   1   ,   1   ,   1   ,   15  ,   15  ,   55  ,   49  ,   249 ,   487 ,   373 ,   737 ,   2133    ,   6883    ,   1603    ,   27607   ,0 };
            const unsigned long dim2199KuoInit[]        =   {   1   ,   1   ,   1   ,   13  ,   25  ,   17  ,   17  ,   15  ,   119 ,   571 ,   1053    ,   2153    ,   5221    ,   2681    ,   24595   ,0 };
            const unsigned long dim2200KuoInit[]        =   {   1   ,   3   ,   1   ,   5   ,   3   ,   63  ,   31  ,   9   ,   263 ,   977 ,   575 ,   1225    ,   4127    ,   15649   ,   1281    ,0 };
            const unsigned long dim2201KuoInit[]        =   {   1   ,   1   ,   7   ,   9   ,   19  ,   1   ,   37  ,   109 ,   13  ,   191 ,   1483    ,   465 ,   1321    ,   13697   ,   26065   ,0 };
            const unsigned long dim2202KuoInit[]        =   {   1   ,   1   ,   3   ,   9   ,   17  ,   15  ,   103 ,   97  ,   237 ,   861 ,   1251    ,   2069    ,   4459    ,   9999    ,   7369    ,0 };
            const unsigned long dim2203KuoInit[]        =   {   1   ,   3   ,   7   ,   9   ,   15  ,   33  ,   55  ,   5   ,   355 ,   519 ,   445 ,   4033    ,   627 ,   2375    ,   31013   ,0 };
            const unsigned long dim2204KuoInit[]        =   {   1   ,   1   ,   1   ,   3   ,   23  ,   51  ,   65  ,   177 ,   225 ,   63  ,   987 ,   2921    ,   823 ,   1781    ,   22923   ,0 };
            const unsigned long dim2205KuoInit[]        =   {   1   ,   3   ,   5   ,   7   ,   27  ,   33  ,   111 ,   47  ,   465 ,   379 ,   1575    ,   1793    ,   513 ,   14033   ,   27409   ,0 };
            const unsigned long dim2206KuoInit[]        =   {   1   ,   3   ,   5   ,   7   ,   1   ,   57  ,   31  ,   195 ,   393 ,   257 ,   1489    ,   2589    ,   5339    ,   7487    ,   29023   ,0 };
            const unsigned long dim2207KuoInit[]        =   {   1   ,   3   ,   5   ,   11  ,   15  ,   27  ,   39  ,   89  ,   387 ,   753 ,   1213    ,   2417    ,   6287    ,   129 ,   12295   ,0 };
            const unsigned long dim2208KuoInit[]        =   {   1   ,   1   ,   3   ,   11  ,   11  ,   37  ,   117 ,   55  ,   55  ,   677 ,   1057    ,   3971    ,   5017    ,   12941   ,   3001    ,0 };
            const unsigned long dim2209KuoInit[]        =   {   1   ,   3   ,   1   ,   7   ,   21  ,   33  ,   103 ,   19  ,   427 ,   637 ,   393 ,   2245    ,   2853    ,   15927   ,   14633   ,0 };
            const unsigned long dim2210KuoInit[]        =   {   1   ,   1   ,   5   ,   7   ,   31  ,   11  ,   75  ,   191 ,   367 ,   659 ,   121 ,   2947    ,   515 ,   10625   ,   19753   ,0 };
            const unsigned long dim2211KuoInit[]        =   {   1   ,   1   ,   5   ,   5   ,   15  ,   63  ,   115 ,   119 ,   85  ,   527 ,   875 ,   67  ,   475 ,   2207    ,   4101    ,0 };
            const unsigned long dim2212KuoInit[]        =   {   1   ,   3   ,   3   ,   9   ,   23  ,   37  ,   53  ,   187 ,   37  ,   535 ,   509 ,   1765    ,   6577    ,   13249   ,   3875    ,0 };
            const unsigned long dim2213KuoInit[]        =   {   1   ,   3   ,   3   ,   5   ,   13  ,   49  ,   23  ,   193 ,   109 ,   295 ,   309 ,   2457    ,   4193    ,   9927    ,   16313   ,0 };
            const unsigned long dim2214KuoInit[]        =   {   1   ,   3   ,   3   ,   5   ,   5   ,   15  ,   33  ,   69  ,   225 ,   231 ,   131 ,   3135    ,   3597    ,   2989    ,   12365   ,0 };
            const unsigned long dim2215KuoInit[]        =   {   1   ,   3   ,   7   ,   1   ,   27  ,   33  ,   9   ,   139 ,   507 ,   789 ,   663 ,   1693    ,   949 ,   9149    ,   29529   ,0 };
            const unsigned long dim2216KuoInit[]        =   {   1   ,   1   ,   7   ,   7   ,   11  ,   1   ,   57  ,   159 ,   511 ,   929 ,   1337    ,   2181    ,   5573    ,   425 ,   26935   ,0 };
            const unsigned long dim2217KuoInit[]        =   {   1   ,   1   ,   5   ,   9   ,   15  ,   53  ,   57  ,   81  ,   105 ,   389 ,   1563    ,   1   ,   669 ,   8403    ,   15747   ,0 };
            const unsigned long dim2218KuoInit[]        =   {   1   ,   3   ,   3   ,   5   ,   19  ,   21  ,   29  ,   133 ,   203 ,   689 ,   2011    ,   1825    ,   3365    ,   1957    ,   21393   ,0 };
            const unsigned long dim2219KuoInit[]        =   {   1   ,   1   ,   3   ,   15  ,   7   ,   11  ,   73  ,   189 ,   63  ,   81  ,   1429    ,   3403    ,   4249    ,   16265   ,   22423   ,0 };
            const unsigned long dim2220KuoInit[]        =   {   1   ,   1   ,   5   ,   1   ,   31  ,   47  ,   75  ,   199 ,   343 ,   957 ,   691 ,   939 ,   2165    ,   10523   ,   3577    ,0 };
            const unsigned long dim2221KuoInit[]        =   {   1   ,   1   ,   7   ,   13  ,   3   ,   27  ,   5   ,   131 ,   437 ,   405 ,   1809    ,   2637    ,   5735    ,   12897   ,   22905   ,0 };
            const unsigned long dim2222KuoInit[]        =   {   1   ,   1   ,   5   ,   1   ,   17  ,   17  ,   85  ,   161 ,   205 ,   81  ,   1125    ,   1783    ,   2241    ,   6321    ,   10891   ,0 };
            const unsigned long dim2223KuoInit[]        =   {   1   ,   3   ,   7   ,   15  ,   27  ,   7   ,   119 ,   61  ,   383 ,   171 ,   1495    ,   1237    ,   729 ,   1977    ,   11779   ,0 };
            const unsigned long dim2224KuoInit[]        =   {   1   ,   1   ,   5   ,   7   ,   21  ,   3   ,   63  ,   167 ,   493 ,   117 ,   855 ,   143 ,   2579    ,   7527    ,   7395    ,0 };
            const unsigned long dim2225KuoInit[]        =   {   1   ,   1   ,   1   ,   9   ,   31  ,   61  ,   15  ,   31  ,   445 ,   397 ,   1369    ,   1047    ,   4751    ,   10375   ,   30319   ,0 };
            const unsigned long dim2226KuoInit[]        =   {   1   ,   3   ,   1   ,   15  ,   7   ,   13  ,   87  ,   99  ,   333 ,   433 ,   287 ,   2861    ,   2085    ,   14691   ,   14801   ,0 };
            const unsigned long dim2227KuoInit[]        =   {   1   ,   3   ,   3   ,   7   ,   27  ,   45  ,   41  ,   69  ,   391 ,   259 ,   899 ,   1019    ,   2999    ,   6487    ,   19831   ,0 };
            const unsigned long dim2228KuoInit[]        =   {   1   ,   3   ,   5   ,   11  ,   19  ,   33  ,   49  ,   87  ,   89  ,   649 ,   1209    ,   43  ,   8035    ,   12687   ,   24783   ,0 };
            const unsigned long dim2229KuoInit[]        =   {   1   ,   3   ,   1   ,   9   ,   13  ,   11  ,   55  ,   5   ,   315 ,   347 ,   1361    ,   41  ,   807 ,   11551   ,   499 ,0 };
            const unsigned long dim2230KuoInit[]        =   {   1   ,   1   ,   7   ,   1   ,   25  ,   15  ,   11  ,   19  ,   233 ,   269 ,   135 ,   2577    ,   4385    ,   8109    ,   13161   ,0 };
            const unsigned long dim2231KuoInit[]        =   {   1   ,   3   ,   1   ,   9   ,   19  ,   13  ,   127 ,   79  ,   49  ,   525 ,   1499    ,   1209    ,   3747    ,   8311    ,   15061   ,0 };
            const unsigned long dim2232KuoInit[]        =   {   1   ,   3   ,   1   ,   15  ,   9   ,   47  ,   59  ,   81  ,   491 ,   401 ,   1691    ,   543 ,   7129    ,   13617   ,   17031   ,0 };
            const unsigned long dim2233KuoInit[]        =   {   1   ,   3   ,   3   ,   5   ,   15  ,   51  ,   41  ,   85  ,   245 ,   91  ,   1293    ,   2041    ,   7205    ,   12709   ,   17421   ,0 };
            const unsigned long dim2234KuoInit[]        =   {   1   ,   3   ,   1   ,   5   ,   3   ,   19  ,   29  ,   149 ,   51  ,   133 ,   953 ,   2385    ,   7471    ,   14829   ,   8197    ,0 };
            const unsigned long dim2235KuoInit[]        =   {   1   ,   3   ,   1   ,   11  ,   31  ,   15  ,   5   ,   139 ,   299 ,   595 ,   1585    ,   2115    ,   5647    ,   12125   ,   1265    ,0 };
            const unsigned long dim2236KuoInit[]        =   {   1   ,   3   ,   7   ,   9   ,   7   ,   5   ,   93  ,   187 ,   307 ,   329 ,   671 ,   3983    ,   817 ,   3965    ,   21561   ,0 };
            const unsigned long dim2237KuoInit[]        =   {   1   ,   3   ,   7   ,   1   ,   21  ,   57  ,   67  ,   159 ,   467 ,   65  ,   1365    ,   2797    ,   7051    ,   12603   ,   10429   ,0 };
            const unsigned long dim2238KuoInit[]        =   {   1   ,   3   ,   1   ,   7   ,   27  ,   59  ,   113 ,   199 ,   13  ,   393 ,   1605    ,   3215    ,   5233    ,   12749   ,   32365   ,0 };
            const unsigned long dim2239KuoInit[]        =   {   1   ,   3   ,   1   ,   1   ,   1   ,   13  ,   15  ,   217 ,   317 ,   945 ,   1713    ,   1951    ,   5585    ,   1529    ,   4969    ,0 };
            const unsigned long dim2240KuoInit[]        =   {   1   ,   1   ,   7   ,   15  ,   3   ,   59  ,   23  ,   255 ,   71  ,   539 ,   21  ,   2603    ,   5145    ,   295 ,   2681    ,0 };
            const unsigned long dim2241KuoInit[]        =   {   1   ,   1   ,   1   ,   13  ,   11  ,   41  ,   85  ,   145 ,   209 ,   1001    ,   659 ,   951 ,   4039    ,   1409    ,   4059    ,0 };
            const unsigned long dim2242KuoInit[]        =   {   1   ,   1   ,   7   ,   13  ,   15  ,   41  ,   15  ,   121 ,   485 ,   515 ,   325 ,   3673    ,   2883    ,   8119    ,   29183   ,0 };
            const unsigned long dim2243KuoInit[]        =   {   1   ,   3   ,   3   ,   13  ,   21  ,   39  ,   31  ,   163 ,   13  ,   219 ,   411 ,   2061    ,   1999    ,   2999    ,   24835   ,0 };
            const unsigned long dim2244KuoInit[]        =   {   1   ,   3   ,   7   ,   7   ,   21  ,   49  ,   9   ,   75  ,   401 ,   339 ,   191 ,   2679    ,   7483    ,   1361    ,   21399   ,0 };
            const unsigned long dim2245KuoInit[]        =   {   1   ,   3   ,   1   ,   13  ,   5   ,   29  ,   9   ,   77  ,   459 ,   247 ,   1777    ,   3873    ,   6747    ,   7541    ,   12015   ,0 };
            const unsigned long dim2246KuoInit[]        =   {   1   ,   1   ,   5   ,   9   ,   23  ,   27  ,   55  ,   51  ,   389 ,   63  ,   1163    ,   255 ,   6049    ,   1737    ,   12403   ,0 };
            const unsigned long dim2247KuoInit[]        =   {   1   ,   1   ,   7   ,   3   ,   23  ,   25  ,   81  ,   247 ,   261 ,   445 ,   1297    ,   1945    ,   2769    ,   1287    ,   24403   ,0 };
            const unsigned long dim2248KuoInit[]        =   {   1   ,   1   ,   5   ,   15  ,   19  ,   13  ,   127 ,   105 ,   475 ,   859 ,   683 ,   941 ,   2911    ,   14571   ,   22769   ,0 };
            const unsigned long dim2249KuoInit[]        =   {   1   ,   1   ,   1   ,   13  ,   31  ,   25  ,   55  ,   63  ,   355 ,   525 ,   397 ,   3213    ,   2267    ,   2731    ,   32145   ,0 };
            const unsigned long dim2250KuoInit[]        =   {   1   ,   1   ,   5   ,   15  ,   3   ,   21  ,   45  ,   229 ,   83  ,   661 ,   149 ,   1615    ,   5309    ,   12233   ,   23101   ,0 };
            const unsigned long dim2251KuoInit[]        =   {   1   ,   1   ,   7   ,   1   ,   15  ,   41  ,   91  ,   17  ,   119 ,   65  ,   1921    ,   97  ,   6521    ,   8969    ,   5419    ,0 };
            const unsigned long dim2252KuoInit[]        =   {   1   ,   1   ,   7   ,   11  ,   31  ,   35  ,   73  ,   123 ,   331 ,   83  ,   125 ,   3001    ,   5485    ,   1693    ,   24327   ,0 };
            const unsigned long dim2253KuoInit[]        =   {   1   ,   3   ,   5   ,   15  ,   29  ,   37  ,   63  ,   229 ,   143 ,   529 ,   961 ,   2845    ,   8003    ,   7741    ,   23393   ,0 };
            const unsigned long dim2254KuoInit[]        =   {   1   ,   3   ,   1   ,   15  ,   25  ,   27  ,   91  ,   171 ,   361 ,   575 ,   1977    ,   2141    ,   89  ,   7097    ,   8115    ,0 };
            const unsigned long dim2255KuoInit[]        =   {   1   ,   3   ,   7   ,   15  ,   13  ,   7   ,   51  ,   253 ,   17  ,   991 ,   595 ,   1819    ,   7091    ,   8619    ,   1147    ,0 };
            const unsigned long dim2256KuoInit[]        =   {   1   ,   3   ,   3   ,   7   ,   13  ,   31  ,   41  ,   11  ,   57  ,   185 ,   1353    ,   907 ,   6585    ,   8597    ,   10239   ,0 };
            const unsigned long dim2257KuoInit[]        =   {   1   ,   3   ,   1   ,   11  ,   31  ,   1   ,   119 ,   165 ,   105 ,   725 ,   493 ,   3279    ,   7117    ,   15141   ,   5957    ,0 };
            const unsigned long dim2258KuoInit[]        =   {   1   ,   3   ,   3   ,   13  ,   29  ,   1   ,   35  ,   95  ,   477 ,   847 ,   733 ,   2599    ,   581 ,   6487    ,   10879   ,0 };
            const unsigned long dim2259KuoInit[]        =   {   1   ,   3   ,   7   ,   3   ,   15  ,   23  ,   115 ,   177 ,   409 ,   711 ,   1179    ,   685 ,   4071    ,   15129   ,   2951    ,0 };
            const unsigned long dim2260KuoInit[]        =   {   1   ,   1   ,   3   ,   11  ,   19  ,   5   ,   69  ,   33  ,   41  ,   501 ,   135 ,   1063    ,   4857    ,   15537   ,   29247   ,0 };
            const unsigned long dim2261KuoInit[]        =   {   1   ,   3   ,   5   ,   13  ,   1   ,   25  ,   123 ,   1   ,   431 ,   399 ,   483 ,   3159    ,   7359    ,   6169    ,   22525   ,0 };
            const unsigned long dim2262KuoInit[]        =   {   1   ,   3   ,   5   ,   7   ,   1   ,   9   ,   125 ,   131 ,   99  ,   911 ,   421 ,   1403    ,   993 ,   14155   ,   20863   ,0 };
            const unsigned long dim2263KuoInit[]        =   {   1   ,   3   ,   1   ,   3   ,   1   ,   43  ,   93  ,   9   ,   345 ,   483 ,   1809    ,   2379    ,   653 ,   14851   ,   20619   ,0 };
            const unsigned long dim2264KuoInit[]        =   {   1   ,   3   ,   7   ,   5   ,   15  ,   37  ,   33  ,   185 ,   173 ,   397 ,   2021    ,   613 ,   37  ,   7185    ,   15399   ,0 };
            const unsigned long dim2265KuoInit[]        =   {   1   ,   3   ,   5   ,   5   ,   17  ,   55  ,   119 ,   207 ,   301 ,   461 ,   1853    ,   2471    ,   3611    ,   1149    ,   3321    ,0 };
            const unsigned long dim2266KuoInit[]        =   {   1   ,   3   ,   3   ,   3   ,   25  ,   57  ,   51  ,   135 ,   337 ,   935 ,   619 ,   1165    ,   1613    ,   8711    ,   2329    ,0 };
            const unsigned long dim2267KuoInit[]        =   {   1   ,   3   ,   1   ,   15  ,   27  ,   25  ,   63  ,   127 ,   169 ,   221 ,   1103    ,   1087    ,   1729    ,   3967    ,   6323    ,0 };
            const unsigned long dim2268KuoInit[]        =   {   1   ,   1   ,   3   ,   11  ,   13  ,   29  ,   105 ,   189 ,   321 ,   103 ,   2047    ,   923 ,   4443    ,   2175    ,   12411   ,0 };
            const unsigned long dim2269KuoInit[]        =   {   1   ,   1   ,   7   ,   9   ,   7   ,   3   ,   3   ,   61  ,   479 ,   533 ,   227 ,   3137    ,   6065    ,   3621    ,   3583    ,0 };
            const unsigned long dim2270KuoInit[]        =   {   1   ,   1   ,   5   ,   13  ,   11  ,   31  ,   23  ,   147 ,   67  ,   823 ,   663 ,   3717    ,   6747    ,   10935   ,   21931   ,0 };
            const unsigned long dim2271KuoInit[]        =   {   1   ,   1   ,   5   ,   13  ,   29  ,   9   ,   75  ,   31  ,   349 ,   603 ,   1071    ,   1767    ,   1811    ,   15921   ,   25681   ,0 };
            const unsigned long dim2272KuoInit[]        =   {   1   ,   3   ,   7   ,   13  ,   29  ,   31  ,   57  ,   15  ,   405 ,   263 ,   715 ,   445 ,   1825    ,   15057   ,   3449    ,0 };
            const unsigned long dim2273KuoInit[]        =   {   1   ,   3   ,   5   ,   5   ,   5   ,   21  ,   61  ,   137 ,   15  ,   935 ,   897 ,   125 ,   6651    ,   6813    ,   27863   ,0 };
            const unsigned long dim2274KuoInit[]        =   {   1   ,   1   ,   3   ,   13  ,   9   ,   53  ,   5   ,   91  ,   93  ,   179 ,   489 ,   3615    ,   5911    ,   7501    ,   14205   ,0 };
            const unsigned long dim2275KuoInit[]        =   {   1   ,   1   ,   5   ,   13  ,   27  ,   49  ,   35  ,   93  ,   371 ,   519 ,   589 ,   1563    ,   2609    ,   14055   ,   28071   ,0 };
            const unsigned long dim2276KuoInit[]        =   {   1   ,   1   ,   5   ,   7   ,   29  ,   41  ,   23  ,   135 ,   491 ,   745 ,   1327    ,   4065    ,   6327    ,   6375    ,   24901   ,0 };
            const unsigned long dim2277KuoInit[]        =   {   1   ,   3   ,   3   ,   5   ,   23  ,   47  ,   113 ,   85  ,   453 ,   321 ,   597 ,   1817    ,   7579    ,   11633   ,   25569   ,0 };
            const unsigned long dim2278KuoInit[]        =   {   1   ,   3   ,   1   ,   13  ,   13  ,   55  ,   127 ,   47  ,   159 ,   885 ,   235 ,   199 ,   3133    ,   9793    ,   25877   ,0 };
            const unsigned long dim2279KuoInit[]        =   {   1   ,   1   ,   3   ,   9   ,   17  ,   21  ,   27  ,   237 ,   413 ,   999 ,   451 ,   3279    ,   1607    ,   9625    ,   16943   ,0 };
            const unsigned long dim2280KuoInit[]        =   {   1   ,   3   ,   1   ,   1   ,   21  ,   61  ,   29  ,   143 ,   33  ,   235 ,   1591    ,   2379    ,   3573    ,   1399    ,   16215   ,0 };
            const unsigned long dim2281KuoInit[]        =   {   1   ,   1   ,   7   ,   1   ,   29  ,   23  ,   95  ,   207 ,   87  ,   199 ,   1389    ,   3587    ,   71  ,   14401   ,   28431   ,0 };
            const unsigned long dim2282KuoInit[]        =   {   1   ,   1   ,   5   ,   13  ,   23  ,   43  ,   109 ,   119 ,   151 ,   451 ,   23  ,   2709    ,   3125    ,   11541   ,   3923    ,0 };
            const unsigned long dim2283KuoInit[]        =   {   1   ,   3   ,   7   ,   3   ,   11  ,   55  ,   5   ,   227 ,   171 ,   465 ,   1685    ,   333 ,   2605    ,   15055   ,   8333    ,0 };
            const unsigned long dim2284KuoInit[]        =   {   1   ,   3   ,   5   ,   3   ,   21  ,   11  ,   57  ,   37  ,   67  ,   667 ,   521 ,   2961    ,   6145    ,   12527   ,   17433   ,0 };
            const unsigned long dim2285KuoInit[]        =   {   1   ,   1   ,   7   ,   11  ,   9   ,   39  ,   103 ,   83  ,   313 ,   827 ,   721 ,   2965    ,   3231    ,   10121   ,   19991   ,0 };
            const unsigned long dim2286KuoInit[]        =   {   1   ,   1   ,   5   ,   7   ,   23  ,   9   ,   31  ,   39  ,   7   ,   767 ,   153 ,   2969    ,   5539    ,   6581    ,   30515   ,0 };
            const unsigned long dim2287KuoInit[]        =   {   1   ,   1   ,   3   ,   13  ,   3   ,   43  ,   49  ,   143 ,   3   ,   227 ,   99  ,   597 ,   7207    ,   4737    ,   2259    ,0 };
            const unsigned long dim2288KuoInit[]        =   {   1   ,   3   ,   3   ,   3   ,   3   ,   57  ,   69  ,   243 ,   199 ,   803 ,   657 ,   1209    ,   1343    ,   11673   ,   21543   ,0 };
            const unsigned long dim2289KuoInit[]        =   {   1   ,   1   ,   1   ,   1   ,   17  ,   27  ,   127 ,   107 ,   171 ,   391 ,   1737    ,   1247    ,   1389    ,   14387   ,   1671    ,0 };
            const unsigned long dim2290KuoInit[]        =   {   1   ,   3   ,   3   ,   9   ,   23  ,   47  ,   97  ,   139 ,   435 ,   943 ,   1111    ,   2723    ,   1021    ,   1155    ,   10241   ,0 };
            const unsigned long dim2291KuoInit[]        =   {   1   ,   3   ,   3   ,   13  ,   15  ,   15  ,   101 ,   205 ,   229 ,   639 ,   1077    ,   1613    ,   3035    ,   12897   ,   14303   ,0 };
            const unsigned long dim2292KuoInit[]        =   {   1   ,   1   ,   1   ,   9   ,   17  ,   43  ,   3   ,   81  ,   149 ,   931 ,   731 ,   4005    ,   7397    ,   8707    ,   27927   ,0 };
            const unsigned long dim2293KuoInit[]        =   {   1   ,   3   ,   7   ,   15  ,   19  ,   59  ,   97  ,   97  ,   347 ,   489 ,   1577    ,   3927    ,   3431    ,   4051    ,   8489    ,0 };
            const unsigned long dim2294KuoInit[]        =   {   1   ,   1   ,   5   ,   3   ,   29  ,   49  ,   19  ,   151 ,   27  ,   701 ,   541 ,   481 ,   101 ,   10701   ,   31909   ,0 };
            const unsigned long dim2295KuoInit[]        =   {   1   ,   1   ,   1   ,   9   ,   11  ,   47  ,   59  ,   45  ,   439 ,   757 ,   1579    ,   1681    ,   6701    ,   11707   ,   10715   ,0 };
            const unsigned long dim2296KuoInit[]        =   {   1   ,   3   ,   3   ,   11  ,   27  ,   55  ,   1   ,   35  ,   491 ,   767 ,   1247    ,   1143    ,   5079    ,   8029    ,   10267   ,0 };
            const unsigned long dim2297KuoInit[]        =   {   1   ,   3   ,   3   ,   1   ,   15  ,   5   ,   53  ,   215 ,   249 ,   659 ,   329 ,   1959    ,   249 ,   11775   ,   4815    ,0 };
            const unsigned long dim2298KuoInit[]        =   {   1   ,   3   ,   7   ,   15  ,   19  ,   17  ,   41  ,   29  ,   157 ,   995 ,   235 ,   1397    ,   3983    ,   11739   ,   7523    ,0 };
            const unsigned long dim2299KuoInit[]        =   {   1   ,   3   ,   5   ,   5   ,   13  ,   3   ,   55  ,   211 ,   59  ,   717 ,   1389    ,   3219    ,   2279    ,   9583    ,   16981   ,0 };
            const unsigned long dim2300KuoInit[]        =   {   1   ,   3   ,   5   ,   11  ,   31  ,   41  ,   95  ,   217 ,   375 ,   255 ,   1433    ,   1595    ,   5795    ,   11893   ,   16665   ,0 };
            const unsigned long dim2301KuoInit[]        =   {   1   ,   3   ,   1   ,   13  ,   15  ,   45  ,   19  ,   125 ,   387 ,   1017    ,   1075    ,   3469    ,   7873    ,   9135    ,   28215   ,0 };
            const unsigned long dim2302KuoInit[]        =   {   1   ,   3   ,   7   ,   5   ,   23  ,   41  ,   35  ,   171 ,   159 ,   545 ,   2043    ,   887 ,   8167    ,   8977    ,   30985   ,0 };
            const unsigned long dim2303KuoInit[]        =   {   1   ,   3   ,   1   ,   3   ,   21  ,   17  ,   31  ,   171 ,   45  ,   979 ,   757 ,   3839    ,   4221    ,   8659    ,   7005    ,0 };
            const unsigned long dim2304KuoInit[]        =   {   1   ,   3   ,   3   ,   1   ,   3   ,   23  ,   81  ,   157 ,   385 ,   233 ,   527 ,   2357    ,   7299    ,   10977   ,   16285   ,0 };
            const unsigned long dim2305KuoInit[]        =   {   1   ,   1   ,   7   ,   13  ,   23  ,   27  ,   31  ,   211 ,   451 ,   211 ,   45  ,   2239    ,   981 ,   14279   ,   14515   ,0 };
            const unsigned long dim2306KuoInit[]        =   {   1   ,   3   ,   1   ,   5   ,   13  ,   61  ,   17  ,   197 ,   455 ,   457 ,   2023    ,   1809    ,   4243    ,   3269    ,   7159    ,0 };
            const unsigned long dim2307KuoInit[]        =   {   1   ,   1   ,   5   ,   5   ,   9   ,   61  ,   49  ,   111 ,   281 ,   827 ,   1155    ,   3107    ,   5761    ,   12043   ,   3061    ,0 };
            const unsigned long dim2308KuoInit[]        =   {   1   ,   1   ,   1   ,   5   ,   19  ,   29  ,   63  ,   209 ,   207 ,   503 ,   1563    ,   2539    ,   5115    ,   12939   ,   22977   ,0 };
            const unsigned long dim2309KuoInit[]        =   {   1   ,   3   ,   7   ,   5   ,   31  ,   33  ,   91  ,   45  ,   399 ,   263 ,   53  ,   2723    ,   7909    ,   7125    ,   14795   ,0 };
            const unsigned long dim2310KuoInit[]        =   {   1   ,   1   ,   7   ,   11  ,   9   ,   5   ,   119 ,   91  ,   87  ,   847 ,   85  ,   4025    ,   1891    ,   791 ,   30447   ,0 };
            const unsigned long dim2311KuoInit[]        =   {   1   ,   1   ,   1   ,   1   ,   25  ,   59  ,   101 ,   85  ,   159 ,   491 ,   635 ,   1047    ,   6221    ,   11579   ,   13943   ,0 };
            const unsigned long dim2312KuoInit[]        =   {   1   ,   3   ,   7   ,   3   ,   27  ,   55  ,   19  ,   29  ,   479 ,   299 ,   1705    ,   3505    ,   6811    ,   14851   ,   6827    ,0 };
            const unsigned long dim2313KuoInit[]        =   {   1   ,   1   ,   1   ,   9   ,   23  ,   19  ,   47  ,   235 ,   81  ,   35  ,   1297    ,   2029    ,   1303    ,   6877    ,   30813   ,0 };
            const unsigned long dim2314KuoInit[]        =   {   1   ,   3   ,   1   ,   11  ,   21  ,   37  ,   49  ,   209 ,   93  ,   439 ,   2023    ,   437 ,   6405    ,   2407    ,   10475   ,0 };
            const unsigned long dim2315KuoInit[]        =   {   1   ,   1   ,   3   ,   3   ,   5   ,   51  ,   11  ,   127 ,   339 ,   419 ,   1495    ,   43  ,   5707    ,   3905    ,   331 ,0 };
            const unsigned long dim2316KuoInit[]        =   {   1   ,   1   ,   5   ,   15  ,   27  ,   7   ,   73  ,   137 ,   111 ,   283 ,   1017    ,   1487    ,   7359    ,   4921    ,   3285    ,0 };
            const unsigned long dim2317KuoInit[]        =   {   1   ,   1   ,   1   ,   9   ,   3   ,   23  ,   123 ,   75  ,   317 ,   517 ,   109 ,   2307    ,   1095    ,   10309   ,   27651   ,0 };
            const unsigned long dim2318KuoInit[]        =   {   1   ,   3   ,   1   ,   1   ,   21  ,   55  ,   5   ,   115 ,   483 ,   379 ,   1935    ,   3483    ,   2485    ,   1991    ,   2403    ,0 };
            const unsigned long dim2319KuoInit[]        =   {   1   ,   3   ,   5   ,   9   ,   27  ,   57  ,   1   ,   227 ,   235 ,   373 ,   241 ,   1115    ,   3037    ,   6841    ,   28711   ,0 };
            const unsigned long dim2320KuoInit[]        =   {   1   ,   3   ,   1   ,   13  ,   21  ,   27  ,   19  ,   217 ,   449 ,   727 ,   475 ,   515 ,   7933    ,   4401    ,   5455    ,0 };
            const unsigned long dim2321KuoInit[]        =   {   1   ,   3   ,   5   ,   13  ,   27  ,   3   ,   117 ,   67  ,   293 ,   625 ,   1463    ,   3713    ,   1271    ,   11873   ,   24239   ,0 };
            const unsigned long dim2322KuoInit[]        =   {   1   ,   1   ,   3   ,   1   ,   7   ,   25  ,   5   ,   187 ,   377 ,   15  ,   1973    ,   597 ,   725 ,   8809    ,   2419    ,0 };
            const unsigned long dim2323KuoInit[]        =   {   1   ,   1   ,   7   ,   13  ,   7   ,   9   ,   89  ,   41  ,   443 ,   921 ,   17  ,   3675    ,   715 ,   7147    ,   25345   ,0 };
            const unsigned long dim2324KuoInit[]        =   {   1   ,   1   ,   5   ,   1   ,   31  ,   27  ,   127 ,   139 ,   101 ,   349 ,   1551    ,   1703    ,   5981    ,   3259    ,   31203   ,0 };
            const unsigned long dim2325KuoInit[]        =   {   1   ,   1   ,   1   ,   5   ,   31  ,   59  ,   1   ,   149 ,   111 ,   947 ,   1919    ,   2841    ,   5383    ,   6973    ,   19565   ,0 };
            const unsigned long dim2326KuoInit[]        =   {   1   ,   3   ,   3   ,   11  ,   5   ,   57  ,   65  ,   197 ,   413 ,   741 ,   1737    ,   1333    ,   5101    ,   14375   ,   28969   ,0 };
            const unsigned long dim2327KuoInit[]        =   {   1   ,   1   ,   1   ,   3   ,   5   ,   21  ,   31  ,   187 ,   51  ,   267 ,   539 ,   1305    ,   4085    ,   14735   ,   9441    ,0 };
            const unsigned long dim2328KuoInit[]        =   {   1   ,   1   ,   7   ,   1   ,   9   ,   27  ,   123 ,   13  ,   383 ,   739 ,   649 ,   1253    ,   4697    ,   9571    ,   16983   ,0 };
            const unsigned long dim2329KuoInit[]        =   {   1   ,   3   ,   1   ,   11  ,   23  ,   3   ,   31  ,   153 ,   185 ,   969 ,   329 ,   3919    ,   3905    ,   14321   ,   16535   ,0 };
            const unsigned long dim2330KuoInit[]        =   {   1   ,   1   ,   1   ,   5   ,   15  ,   31  ,   97  ,   243 ,   117 ,   269 ,   1545    ,   1495    ,   167 ,   14281   ,   1369    ,0 };
            const unsigned long dim2331KuoInit[]        =   {   1   ,   1   ,   1   ,   1   ,   7   ,   47  ,   9   ,   187 ,   233 ,   629 ,   1981    ,   2327    ,   7473    ,   14641   ,   31829   ,0 };
            const unsigned long dim2332KuoInit[]        =   {   1   ,   1   ,   1   ,   11  ,   11  ,   43  ,   117 ,   193 ,   395 ,   501 ,   1213    ,   3499    ,   3369    ,   817 ,   22069   ,0 };
            const unsigned long dim2333KuoInit[]        =   {   1   ,   3   ,   7   ,   1   ,   13  ,   63  ,   67  ,   67  ,   119 ,   371 ,   1887    ,   1281    ,   129 ,   14277   ,   5213    ,0 };
            const unsigned long dim2334KuoInit[]        =   {   1   ,   3   ,   1   ,   11  ,   11  ,   57  ,   47  ,   149 ,   119 ,   547 ,   389 ,   1977    ,   1709    ,   2335    ,   7023    ,0 };
            const unsigned long dim2335KuoInit[]        =   {   1   ,   1   ,   7   ,   7   ,   15  ,   41  ,   19  ,   57  ,   279 ,   737 ,   29  ,   2015    ,   1103    ,   4933    ,   26145   ,0 };
            const unsigned long dim2336KuoInit[]        =   {   1   ,   3   ,   1   ,   1   ,   23  ,   61  ,   17  ,   1   ,   309 ,   637 ,   809 ,   3443    ,   135 ,   6851    ,   27321   ,0 };
            const unsigned long dim2337KuoInit[]        =   {   1   ,   3   ,   1   ,   3   ,   21  ,   63  ,   81  ,   131 ,   495 ,   209 ,   1901    ,   2921    ,   1021    ,   7207    ,   22053   ,0 };
            const unsigned long dim2338KuoInit[]        =   {   1   ,   1   ,   7   ,   9   ,   7   ,   63  ,   1   ,   119 ,   329 ,   553 ,   1405    ,   97  ,   4751    ,   6955    ,   20049   ,0 };
            const unsigned long dim2339KuoInit[]        =   {   1   ,   3   ,   7   ,   15  ,   1   ,   59  ,   99  ,   201 ,   487 ,   271 ,   2035    ,   3261    ,   4953    ,   15257   ,   11835   ,0 };
            const unsigned long dim2340KuoInit[]        =   {   1   ,   1   ,   5   ,   11  ,   27  ,   21  ,   27  ,   227 ,   331 ,   779 ,   1003    ,   2291    ,   7139    ,   14125   ,   14371   ,0 };
            const unsigned long dim2341KuoInit[]        =   {   1   ,   1   ,   3   ,   13  ,   15  ,   43  ,   23  ,   183 ,   463 ,   273 ,   1707    ,   2001    ,   3219    ,   14481   ,   751 ,0 };
            const unsigned long dim2342KuoInit[]        =   {   1   ,   1   ,   5   ,   5   ,   23  ,   31  ,   51  ,   39  ,   221 ,   687 ,   2003    ,   1251    ,   3541    ,   6347    ,   551 ,0 };
            const unsigned long dim2343KuoInit[]        =   {   1   ,   1   ,   7   ,   3   ,   17  ,   51  ,   3   ,   57  ,   131 ,   13  ,   2027    ,   3301    ,   4223    ,   6499    ,   30913   ,0 };
            const unsigned long dim2344KuoInit[]        =   {   1   ,   1   ,   7   ,   13  ,   11  ,   47  ,   53  ,   205 ,   59  ,   315 ,   1595    ,   3747    ,   2623    ,   7317    ,   8309    ,0 };
            const unsigned long dim2345KuoInit[]        =   {   1   ,   3   ,   1   ,   11  ,   13  ,   45  ,   83  ,   131 ,   193 ,   739 ,   61  ,   1657    ,   3789    ,   9859    ,   3755    ,0 };
            const unsigned long dim2346KuoInit[]        =   {   1   ,   3   ,   7   ,   11  ,   25  ,   7   ,   123 ,   103 ,   307 ,   463 ,   1113    ,   2245    ,   2819    ,   12983   ,   18077   ,0 };
            const unsigned long dim2347KuoInit[]        =   {   1   ,   1   ,   1   ,   1   ,   7   ,   23  ,   103 ,   183 ,   409 ,   657 ,   583 ,   583 ,   5449    ,   7397    ,   13085   ,0 };
            const unsigned long dim2348KuoInit[]        =   {   1   ,   3   ,   3   ,   5   ,   3   ,   61  ,   31  ,   147 ,   31  ,   449 ,   1579    ,   4021    ,   1163    ,   9719    ,   27079   ,0 };
            const unsigned long dim2349KuoInit[]        =   {   1   ,   3   ,   1   ,   3   ,   9   ,   53  ,   125 ,   77  ,   99  ,   863 ,   1719    ,   51  ,   4183    ,   13009   ,   791 ,0 };
            const unsigned long dim2350KuoInit[]        =   {   1   ,   3   ,   3   ,   11  ,   27  ,   33  ,   43  ,   5   ,   173 ,   775 ,   593 ,   3701    ,   3323    ,   6481    ,   25765   ,0 };
            const unsigned long dim2351KuoInit[]        =   {   1   ,   1   ,   7   ,   15  ,   11  ,   3   ,   109 ,   221 ,   375 ,   883 ,   903 ,   2361    ,   255 ,   12851   ,   16697   ,0 };
            const unsigned long dim2352KuoInit[]        =   {   1   ,   1   ,   5   ,   9   ,   3   ,   61  ,   127 ,   89  ,   213 ,   71  ,   303 ,   2525    ,   3691    ,   2945    ,   12359   ,0 };
            const unsigned long dim2353KuoInit[]        =   {   1   ,   3   ,   1   ,   13  ,   19  ,   15  ,   7   ,   231 ,   159 ,   275 ,   1161    ,   819 ,   3003    ,   12707   ,   20699   ,0 };
            const unsigned long dim2354KuoInit[]        =   {   1   ,   1   ,   3   ,   1   ,   31  ,   51  ,   125 ,   21  ,   27  ,   427 ,   941 ,   237 ,   4343    ,   9555    ,   30511   ,0 };
            const unsigned long dim2355KuoInit[]        =   {   1   ,   3   ,   3   ,   9   ,   17  ,   23  ,   105 ,   21  ,   471 ,   357 ,   773 ,   1185    ,   2655    ,   2317    ,   9969    ,0 };
            const unsigned long dim2356KuoInit[]        =   {   1   ,   1   ,   5   ,   7   ,   15  ,   55  ,   87  ,   249 ,   111 ,   319 ,   985 ,   2305    ,   1641    ,   12655   ,   7321    ,0 };
            const unsigned long dim2357KuoInit[]        =   {   1   ,   3   ,   5   ,   3   ,   23  ,   49  ,   85  ,   31  ,   299 ,   391 ,   549 ,   3831    ,   1413    ,   13803   ,   30599   ,0 };
            const unsigned long dim2358KuoInit[]        =   {   1   ,   1   ,   3   ,   3   ,   31  ,   39  ,   43  ,   253 ,   67  ,   321 ,   463 ,   2327    ,   6831    ,   10461   ,   18077   ,0 };
            const unsigned long dim2359KuoInit[]        =   {   1   ,   3   ,   1   ,   9   ,   13  ,   13  ,   127 ,   175 ,   425 ,   567 ,   1663    ,   195 ,   1763    ,   13271   ,   1739    ,0 };
            const unsigned long dim2360KuoInit[]        =   {   1   ,   1   ,   7   ,   9   ,   9   ,   33  ,   11  ,   167 ,   87  ,   901 ,   1625    ,   2055    ,   2475    ,   7241    ,   29717   ,0 };
            const unsigned long dim2361KuoInit[]        =   {   1   ,   1   ,   1   ,   1   ,   15  ,   63  ,   127 ,   29  ,   227 ,   817 ,   637 ,   1873    ,   4885    ,   16299   ,   31243   ,0 };
            const unsigned long dim2362KuoInit[]        =   {   1   ,   3   ,   7   ,   3   ,   17  ,   27  ,   81  ,   237 ,   283 ,   161 ,   1905    ,   1551    ,   6171    ,   4519    ,   11047   ,0 };
            const unsigned long dim2363KuoInit[]        =   {   1   ,   1   ,   7   ,   9   ,   25  ,   23  ,   25  ,   1   ,   461 ,   623 ,   493 ,   827 ,   5795    ,   16065   ,   8401    ,0 };
            const unsigned long dim2364KuoInit[]        =   {   1   ,   3   ,   5   ,   11  ,   29  ,   29  ,   91  ,   133 ,   91  ,   565 ,   1821    ,   269 ,   1731    ,   1499    ,   7975    ,0 };
            const unsigned long dim2365KuoInit[]        =   {   1   ,   1   ,   1   ,   13  ,   3   ,   41  ,   99  ,   23  ,   239 ,   945 ,   219 ,   1127    ,   4919    ,   9713    ,   3321    ,0 };
            const unsigned long dim2366KuoInit[]        =   {   1   ,   3   ,   1   ,   11  ,   25  ,   59  ,   59  ,   127 ,   83  ,   643 ,   881 ,   1671    ,   3521    ,   459 ,   15479   ,0 };
            const unsigned long dim2367KuoInit[]        =   {   1   ,   1   ,   5   ,   11  ,   15  ,   23  ,   111 ,   225 ,   145 ,   703 ,   1643    ,   419 ,   4853    ,   7195    ,   20879   ,0 };
            const unsigned long dim2368KuoInit[]        =   {   1   ,   3   ,   7   ,   15  ,   5   ,   31  ,   111 ,   151 ,   489 ,   629 ,   473 ,   1507    ,   117 ,   15305   ,   11011   ,0 };
            const unsigned long dim2369KuoInit[]        =   {   1   ,   1   ,   3   ,   3   ,   1   ,   55  ,   55  ,   223 ,   489 ,   911 ,   1721    ,   3533    ,   5893    ,   2891    ,   8929    ,0 };
            const unsigned long dim2370KuoInit[]        =   {   1   ,   1   ,   5   ,   13  ,   7   ,   43  ,   113 ,   99  ,   19  ,   681 ,   1765    ,   1983    ,   5477    ,   7531    ,   26507   ,0 };
            const unsigned long dim2371KuoInit[]        =   {   1   ,   1   ,   3   ,   1   ,   9   ,   7   ,   77  ,   231 ,   345 ,   745 ,   285 ,   3213    ,   2611    ,   4449    ,   24621   ,0 };
            const unsigned long dim2372KuoInit[]        =   {   1   ,   1   ,   5   ,   1   ,   21  ,   21  ,   3   ,   11  ,   131 ,   327 ,   1893    ,   33  ,   6947    ,   337 ,   8877    ,0 };
            const unsigned long dim2373KuoInit[]        =   {   1   ,   1   ,   5   ,   13  ,   5   ,   11  ,   41  ,   7   ,   447 ,   625 ,   1231    ,   3621    ,   3785    ,   12351   ,   13829   ,0 };
            const unsigned long dim2374KuoInit[]        =   {   1   ,   3   ,   1   ,   13  ,   29  ,   41  ,   11  ,   63  ,   357 ,   723 ,   1657    ,   91  ,   1373    ,   6383    ,   8327    ,0 };
            const unsigned long dim2375KuoInit[]        =   {   1   ,   3   ,   3   ,   15  ,   23  ,   51  ,   85  ,   113 ,   103 ,   499 ,   1669    ,   3685    ,   63  ,   2853    ,   20859   ,0 };
            const unsigned long dim2376KuoInit[]        =   {   1   ,   1   ,   5   ,   11  ,   31  ,   39  ,   15  ,   103 ,   365 ,   231 ,   443 ,   3213    ,   6951    ,   1715    ,   32301   ,0 };
            const unsigned long dim2377KuoInit[]        =   {   1   ,   3   ,   7   ,   5   ,   19  ,   23  ,   13  ,   127 ,   475 ,   281 ,   1095    ,   473 ,   4955    ,   2217    ,   10427   ,0 };
            const unsigned long dim2378KuoInit[]        =   {   1   ,   1   ,   5   ,   13  ,   17  ,   41  ,   47  ,   121 ,   131 ,   69  ,   1747    ,   3679    ,   2765    ,   4349    ,   32041   ,0 };
            const unsigned long dim2379KuoInit[]        =   {   1   ,   1   ,   7   ,   9   ,   3   ,   47  ,   11  ,   127 ,   321 ,   669 ,   1283    ,   797 ,   6393    ,   12351   ,   5161    ,0 };
            const unsigned long dim2380KuoInit[]        =   {   1   ,   3   ,   7   ,   15  ,   21  ,   37  ,   73  ,   15  ,   393 ,   173 ,   1461    ,   1437    ,   2367    ,   7171    ,   15123   ,0 };
            const unsigned long dim2381KuoInit[]        =   {   1   ,   3   ,   5   ,   15  ,   7   ,   57  ,   29  ,   83  ,   241 ,   221 ,   1607    ,   1291    ,   4839    ,   11293   ,   1837    ,0 };
            const unsigned long dim2382KuoInit[]        =   {   1   ,   3   ,   5   ,   7   ,   1   ,   31  ,   59  ,   143 ,   487 ,   317 ,   535 ,   1505    ,   3705    ,   9947    ,   6149    ,0 };
            const unsigned long dim2383KuoInit[]        =   {   1   ,   3   ,   1   ,   3   ,   29  ,   31  ,   69  ,   63  ,   183 ,   173 ,   1965    ,   3331    ,   4473    ,   993 ,   4937    ,0 };
            const unsigned long dim2384KuoInit[]        =   {   1   ,   1   ,   1   ,   15  ,   13  ,   13  ,   55  ,   43  ,   283 ,   897 ,   1019    ,   2761    ,   3237    ,   14661   ,   12309   ,0 };
            const unsigned long dim2385KuoInit[]        =   {   1   ,   3   ,   1   ,   15  ,   9   ,   15  ,   1   ,   19  ,   227 ,   777 ,   1713    ,   3471    ,   2917    ,   1117    ,   16397   ,0 };
            const unsigned long dim2386KuoInit[]        =   {   1   ,   1   ,   1   ,   1   ,   21  ,   19  ,   49  ,   3   ,   477 ,   659 ,   61  ,   2373    ,   6363    ,   13771   ,   5675    ,0 };
            const unsigned long dim2387KuoInit[]        =   {   1   ,   3   ,   3   ,   3   ,   13  ,   23  ,   43  ,   253 ,   483 ,   975 ,   1629    ,   19  ,   6339    ,   13963   ,   6065    ,0 };
            const unsigned long dim2388KuoInit[]        =   {   1   ,   3   ,   7   ,   3   ,   9   ,   31  ,   65  ,   1   ,   311 ,   67  ,   293 ,   667 ,   3685    ,   3589    ,   25251   ,0 };
            const unsigned long dim2389KuoInit[]        =   {   1   ,   1   ,   3   ,   1   ,   29  ,   31  ,   47  ,   85  ,   431 ,   131 ,   1495    ,   1271    ,   6121    ,   7183    ,   20477   ,0 };
            const unsigned long dim2390KuoInit[]        =   {   1   ,   1   ,   7   ,   5   ,   13  ,   37  ,   83  ,   201 ,   277 ,   643 ,   19  ,   2519    ,   2589    ,   2471    ,   19727   ,0 };
            const unsigned long dim2391KuoInit[]        =   {   1   ,   3   ,   5   ,   13  ,   27  ,   13  ,   97  ,   183 ,   463 ,   639 ,   87  ,   181 ,   747 ,   3715    ,   16207   ,0 };
            const unsigned long dim2392KuoInit[]        =   {   1   ,   1   ,   7   ,   3   ,   9   ,   29  ,   21  ,   191 ,   485 ,   211 ,   541 ,   69  ,   511 ,   9923    ,   24543   ,0 };
            const unsigned long dim2393KuoInit[]        =   {   1   ,   3   ,   5   ,   9   ,   19  ,   43  ,   85  ,   189 ,   47  ,   723 ,   1211    ,   1259    ,   1645    ,   2983    ,   27195   ,0 };
            const unsigned long dim2394KuoInit[]        =   {   1   ,   1   ,   7   ,   15  ,   17  ,   25  ,   103 ,   83  ,   389 ,   229 ,   1457    ,   4045    ,   161 ,   15471   ,   19929   ,0 };
            const unsigned long dim2395KuoInit[]        =   {   1   ,   1   ,   3   ,   9   ,   5   ,   27  ,   55  ,   231 ,   245 ,   723 ,   479 ,   3761    ,   787 ,   12939   ,   30227   ,0 };
            const unsigned long dim2396KuoInit[]        =   {   1   ,   3   ,   3   ,   1   ,   19  ,   39  ,   63  ,   27  ,   137 ,   399 ,   1353    ,   1143    ,   5835    ,   2013    ,   4107    ,0 };
            const unsigned long dim2397KuoInit[]        =   {   1   ,   3   ,   5   ,   7   ,   5   ,   39  ,   45  ,   207 ,   7   ,   131 ,   961 ,   1715    ,   3761    ,   11501   ,   9803    ,0 };
            const unsigned long dim2398KuoInit[]        =   {   1   ,   3   ,   1   ,   7   ,   17  ,   11  ,   23  ,   45  ,   417 ,   289 ,   1545    ,   2585    ,   1345    ,   15529   ,   8365    ,0 };
            const unsigned long dim2399KuoInit[]        =   {   1   ,   1   ,   1   ,   7   ,   21  ,   27  ,   97  ,   23  ,   461 ,   983 ,   997 ,   1321    ,   2551    ,   5485    ,   23565   ,0 };
            const unsigned long dim2400KuoInit[]        =   {   1   ,   1   ,   5   ,   5   ,   23  ,   19  ,   77  ,   103 ,   171 ,   553 ,   1767    ,   3255    ,   7955    ,   249 ,   32681   ,0 };
            const unsigned long dim2401KuoInit[]        =   {   1   ,   1   ,   7   ,   1   ,   23  ,   43  ,   73  ,   15  ,   381 ,   245 ,   1389    ,   3693    ,   1975    ,   13629   ,   2079    ,0 };
            const unsigned long dim2402KuoInit[]        =   {   1   ,   1   ,   7   ,   11  ,   23  ,   61  ,   79  ,   39  ,   345 ,   541 ,   875 ,   1375    ,   123 ,   15305   ,   13731   ,0 };
            const unsigned long dim2403KuoInit[]        =   {   1   ,   1   ,   3   ,   15  ,   3   ,   17  ,   19  ,   11  ,   79  ,   415 ,   1791    ,   447 ,   6529    ,   11243   ,   10427   ,0 };
            const unsigned long dim2404KuoInit[]        =   {   1   ,   1   ,   7   ,   15  ,   9   ,   31  ,   121 ,   247 ,   403 ,   935 ,   705 ,   2921    ,   3769    ,   625 ,   22905   ,0 };
            const unsigned long dim2405KuoInit[]        =   {   1   ,   3   ,   5   ,   15  ,   19  ,   9   ,   5   ,   47  ,   135 ,   585 ,   1155    ,   1517    ,   4961    ,   15725   ,   7585    ,0 };
            const unsigned long dim2406KuoInit[]        =   {   1   ,   3   ,   3   ,   7   ,   19  ,   57  ,   37  ,   69  ,   117 ,   485 ,   115 ,   2709    ,   1909    ,   12781   ,   19323   ,0 };
            const unsigned long dim2407KuoInit[]        =   {   1   ,   1   ,   1   ,   5   ,   7   ,   1   ,   37  ,   39  ,   141 ,   1017    ,   993 ,   191 ,   6173    ,   8083    ,   9733    ,0 };
            const unsigned long dim2408KuoInit[]        =   {   1   ,   1   ,   7   ,   3   ,   3   ,   21  ,   17  ,   55  ,   129 ,   353 ,   673 ,   3233    ,   6289    ,   10955   ,   19357   ,0 };
            const unsigned long dim2409KuoInit[]        =   {   1   ,   3   ,   7   ,   11  ,   31  ,   31  ,   31  ,   105 ,   187 ,   839 ,   733 ,   1795    ,   5075    ,   1053    ,   613 ,0 };
            const unsigned long dim2410KuoInit[]        =   {   1   ,   3   ,   3   ,   1   ,   9   ,   9   ,   23  ,   139 ,   417 ,   431 ,   1943    ,   3067    ,   267 ,   4401    ,   9847    ,0 };
            const unsigned long dim2411KuoInit[]        =   {   1   ,   3   ,   5   ,   11  ,   23  ,   61  ,   61  ,   119 ,   383 ,   869 ,   503 ,   195 ,   1935    ,   4617    ,   21393   ,0 };
            const unsigned long dim2412KuoInit[]        =   {   1   ,   1   ,   1   ,   5   ,   7   ,   15  ,   51  ,   53  ,   427 ,   499 ,   1829    ,   2533    ,   3029    ,   12485   ,   28425   ,0 };
            const unsigned long dim2413KuoInit[]        =   {   1   ,   3   ,   3   ,   7   ,   9   ,   11  ,   55  ,   205 ,   255 ,   619 ,   1323    ,   1983    ,   15  ,   4237    ,   20983   ,0 };
            const unsigned long dim2414KuoInit[]        =   {   1   ,   1   ,   3   ,   15  ,   15  ,   37  ,   109 ,   247 ,   179 ,   1005    ,   1111    ,   77  ,   4533    ,   12849   ,   25601   ,0 };
            const unsigned long dim2415KuoInit[]        =   {   1   ,   1   ,   5   ,   3   ,   15  ,   25  ,   59  ,   117 ,   55  ,   755 ,   681 ,   649 ,   6041    ,   1695    ,   13719   ,0 };
            const unsigned long dim2416KuoInit[]        =   {   1   ,   3   ,   1   ,   3   ,   15  ,   17  ,   117 ,   113 ,   197 ,   701 ,   773 ,   1097    ,   1267    ,   16093   ,   21555   ,0 };
            const unsigned long dim2417KuoInit[]        =   {   1   ,   3   ,   1   ,   5   ,   9   ,   57  ,   39  ,   171 ,   449 ,   507 ,   29  ,   3581    ,   1929    ,   12715   ,   24575   ,0 };
            const unsigned long dim2418KuoInit[]        =   {   1   ,   3   ,   5   ,   5   ,   23  ,   1   ,   67  ,   153 ,   375 ,   333 ,   379 ,   3163    ,   8121    ,   13465   ,   27641   ,0 };
            const unsigned long dim2419KuoInit[]        =   {   1   ,   3   ,   5   ,   5   ,   31  ,   27  ,   83  ,   243 ,   415 ,   997 ,   1787    ,   4021    ,   5955    ,   1861    ,   10511   ,0 };
            const unsigned long dim2420KuoInit[]        =   {   1   ,   3   ,   1   ,   3   ,   13  ,   47  ,   69  ,   169 ,   185 ,   645 ,   1857    ,   3487    ,   1557    ,   9589    ,   7971    ,0 };
            const unsigned long dim2421KuoInit[]        =   {   1   ,   3   ,   3   ,   15  ,   3   ,   1   ,   93  ,   97  ,   71  ,   433 ,   1371    ,   2495    ,   8171    ,   9101    ,   32435   ,0 };
            const unsigned long dim2422KuoInit[]        =   {   1   ,   3   ,   7   ,   11  ,   3   ,   29  ,   37  ,   167 ,   169 ,   15  ,   1575    ,   1331    ,   7163    ,   1555    ,   32439   ,0 };
            const unsigned long dim2423KuoInit[]        =   {   1   ,   1   ,   5   ,   1   ,   29  ,   11  ,   83  ,   135 ,   367 ,   37  ,   1455    ,   3939    ,   3383    ,   5291    ,   14269   ,0 };
            const unsigned long dim2424KuoInit[]        =   {   1   ,   1   ,   5   ,   15  ,   7   ,   63  ,   17  ,   45  ,   259 ,   57  ,   1933    ,   3101    ,   7871    ,   12633   ,   15789   ,0 };
            const unsigned long dim2425KuoInit[]        =   {   1   ,   1   ,   1   ,   7   ,   19  ,   37  ,   89  ,   159 ,   487 ,   101 ,   1551    ,   3703    ,   4323    ,   9153    ,   6155    ,0 };
            const unsigned long dim2426KuoInit[]        =   {   1   ,   3   ,   3   ,   13  ,   15  ,   49  ,   81  ,   51  ,   267 ,   761 ,   863 ,   2239    ,   7683    ,   15491   ,   28655   ,0 };
            const unsigned long dim2427KuoInit[]        =   {   1   ,   3   ,   1   ,   1   ,   25  ,   53  ,   113 ,   117 ,   11  ,   1003    ,   539 ,   3603    ,   1007    ,   11329   ,   15099   ,0 };
            const unsigned long dim2428KuoInit[]        =   {   1   ,   3   ,   5   ,   7   ,   9   ,   27  ,   125 ,   231 ,   301 ,   461 ,   2031    ,   183 ,   6361    ,   10811   ,   29425   ,0 };
            const unsigned long dim2429KuoInit[]        =   {   1   ,   1   ,   3   ,   1   ,   3   ,   37  ,   65  ,   127 ,   487 ,   611 ,   867 ,   2361    ,   5243    ,   11829   ,   31297   ,0 };
            const unsigned long dim2430KuoInit[]        =   {   1   ,   3   ,   3   ,   15  ,   9   ,   39  ,   31  ,   219 ,   77  ,   97  ,   1061    ,   435 ,   2449    ,   14751   ,   20273   ,0 };
            const unsigned long dim2431KuoInit[]        =   {   1   ,   1   ,   5   ,   11  ,   1   ,   23  ,   41  ,   239 ,   183 ,   287 ,   969 ,   1325    ,   3553    ,   6751    ,   14679   ,0 };
            const unsigned long dim2432KuoInit[]        =   {   1   ,   3   ,   7   ,   1   ,   17  ,   61  ,   47  ,   207 ,   509 ,   601 ,   1539    ,   345 ,   3723    ,   5415    ,   11667   ,0 };
            const unsigned long dim2433KuoInit[]        =   {   1   ,   1   ,   7   ,   7   ,   23  ,   33  ,   25  ,   165 ,   345 ,   569 ,   983 ,   2857    ,   1563    ,   1737    ,   19285   ,0 };
            const unsigned long dim2434KuoInit[]        =   {   1   ,   1   ,   1   ,   7   ,   3   ,   43  ,   107 ,   229 ,   223 ,   583 ,   1495    ,   3487    ,   7497    ,   10367   ,   20575   ,0 };
            const unsigned long dim2435KuoInit[]        =   {   1   ,   3   ,   3   ,   13  ,   21  ,   49  ,   17  ,   195 ,   115 ,   797 ,   1489    ,   1587    ,   7847    ,   11013   ,   14547   ,0 };
            const unsigned long dim2436KuoInit[]        =   {   1   ,   3   ,   5   ,   11  ,   9   ,   29  ,   19  ,   245 ,   153 ,   779 ,   1311    ,   425 ,   5809    ,   4457    ,   6905    ,0 };
            const unsigned long dim2437KuoInit[]        =   {   1   ,   1   ,   5   ,   3   ,   21  ,   33  ,   13  ,   193 ,   343 ,   971 ,   1315    ,   1989    ,   1969    ,   9227    ,   8881    ,0 };
            const unsigned long dim2438KuoInit[]        =   {   1   ,   3   ,   7   ,   15  ,   3   ,   59  ,   87  ,   205 ,   261 ,   567 ,   557 ,   1143    ,   6923    ,   9451    ,   20239   ,0 };
            const unsigned long dim2439KuoInit[]        =   {   1   ,   1   ,   1   ,   15  ,   29  ,   35  ,   127 ,   123 ,   173 ,   807 ,   1279    ,   2413    ,   1899    ,   11657   ,   22257   ,0 };
            const unsigned long dim2440KuoInit[]        =   {   1   ,   3   ,   5   ,   3   ,   23  ,   35  ,   125 ,   129 ,   239 ,   951 ,   499 ,   1609    ,   4643    ,   339 ,   26701   ,0 };
            const unsigned long dim2441KuoInit[]        =   {   1   ,   3   ,   3   ,   13  ,   9   ,   15  ,   23  ,   55  ,   105 ,   81  ,   591 ,   947 ,   3761    ,   7351    ,   25697   ,0 };
            const unsigned long dim2442KuoInit[]        =   {   1   ,   1   ,   7   ,   13  ,   23  ,   17  ,   127 ,   149 ,   227 ,   1005    ,   839 ,   3391    ,   5761    ,   10431   ,   31899   ,0 };
            const unsigned long dim2443KuoInit[]        =   {   1   ,   3   ,   1   ,   1   ,   23  ,   7   ,   11  ,   3   ,   77  ,   457 ,   375 ,   3823    ,   5447    ,   8339    ,   18847   ,0 };
            const unsigned long dim2444KuoInit[]        =   {   1   ,   3   ,   7   ,   3   ,   29  ,   1   ,   19  ,   39  ,   23  ,   285 ,   569 ,   2979    ,   7029    ,   11311   ,   18269   ,0 };
            const unsigned long dim2445KuoInit[]        =   {   1   ,   1   ,   5   ,   9   ,   21  ,   19  ,   85  ,   211 ,   437 ,   185 ,   597 ,   787 ,   4555    ,   2853    ,   18493   ,0 };
            const unsigned long dim2446KuoInit[]        =   {   1   ,   3   ,   3   ,   9   ,   11  ,   27  ,   25  ,   245 ,   55  ,   869 ,   323 ,   3999    ,   2639    ,   9385    ,   26823   ,0 };
            const unsigned long dim2447KuoInit[]        =   {   1   ,   3   ,   1   ,   7   ,   19  ,   61  ,   51  ,   111 ,   115 ,   721 ,   893 ,   23  ,   6375    ,   5101    ,   25567   ,0 };
            const unsigned long dim2448KuoInit[]        =   {   1   ,   3   ,   5   ,   9   ,   9   ,   13  ,   9   ,   241 ,   297 ,   391 ,   1529    ,   3843    ,   737 ,   5897    ,   23469   ,0 };
            const unsigned long dim2449KuoInit[]        =   {   1   ,   3   ,   7   ,   3   ,   31  ,   53  ,   43  ,   219 ,   81  ,   507 ,   1627    ,   3503    ,   817 ,   6533    ,   31633   ,0 };
            const unsigned long dim2450KuoInit[]        =   {   1   ,   1   ,   7   ,   1   ,   13  ,   59  ,   13  ,   185 ,   387 ,   869 ,   1013    ,   2291    ,   4945    ,   11493   ,   7785    ,0 };
            const unsigned long dim2451KuoInit[]        =   {   1   ,   3   ,   5   ,   9   ,   31  ,   35  ,   11  ,   225 ,   235 ,   893 ,   1749    ,   2393    ,   5317    ,   10069   ,   8569    ,0 };
            const unsigned long dim2452KuoInit[]        =   {   1   ,   1   ,   1   ,   3   ,   5   ,   15  ,   93  ,   25  ,   503 ,   205 ,   65  ,   2563    ,   7971    ,   2917    ,   19687   ,0 };
            const unsigned long dim2453KuoInit[]        =   {   1   ,   3   ,   3   ,   13  ,   1   ,   37  ,   127 ,   63  ,   339 ,   121 ,   1895    ,   1943    ,   7527    ,   15869   ,   20533   ,0 };
            const unsigned long dim2454KuoInit[]        =   {   1   ,   1   ,   5   ,   7   ,   19  ,   25  ,   87  ,   27  ,   279 ,   419 ,   1261    ,   613 ,   7687    ,   5301    ,   6761    ,0 };
            const unsigned long dim2455KuoInit[]        =   {   1   ,   1   ,   1   ,   1   ,   13  ,   33  ,   59  ,   235 ,   471 ,   811 ,   695 ,   2727    ,   5615    ,   6269    ,   13469   ,0 };
            const unsigned long dim2456KuoInit[]        =   {   1   ,   3   ,   1   ,   15  ,   17  ,   23  ,   123 ,   71  ,   15  ,   1005    ,   973 ,   411 ,   2627    ,   1967    ,   25597   ,0 };
            const unsigned long dim2457KuoInit[]        =   {   1   ,   3   ,   5   ,   1   ,   3   ,   61  ,   71  ,   217 ,   377 ,   613 ,   1819    ,   141 ,   6265    ,   9223    ,   20635   ,0 };
            const unsigned long dim2458KuoInit[]        =   {   1   ,   1   ,   5   ,   11  ,   15  ,   1   ,   39  ,   121 ,   83  ,   57  ,   1133    ,   3165    ,   7675    ,   13715   ,   11383   ,0 };
            const unsigned long dim2459KuoInit[]        =   {   1   ,   3   ,   7   ,   15  ,   9   ,   3   ,   97  ,   223 ,   493 ,   77  ,   159 ,   2921    ,   5475    ,   9289    ,   20925   ,0 };
            const unsigned long dim2460KuoInit[]        =   {   1   ,   1   ,   3   ,   7   ,   29  ,   13  ,   87  ,   219 ,   327 ,   879 ,   883 ,   1637    ,   3451    ,   13939   ,   8203    ,0 };
            const unsigned long dim2461KuoInit[]        =   {   1   ,   3   ,   7   ,   15  ,   25  ,   21  ,   95  ,   129 ,   343 ,   9   ,   1415    ,   2965    ,   747 ,   7297    ,   24293   ,0 };
            const unsigned long dim2462KuoInit[]        =   {   1   ,   1   ,   1   ,   3   ,   23  ,   47  ,   57  ,   31  ,   359 ,   1017    ,   769 ,   2761    ,   1933    ,   6091    ,   6475    ,0 };
            const unsigned long dim2463KuoInit[]        =   {   1   ,   1   ,   5   ,   9   ,   27  ,   45  ,   125 ,   67  ,   145 ,   581 ,   157 ,   2153    ,   2727    ,   8025    ,   28165   ,0 };
            const unsigned long dim2464KuoInit[]        =   {   1   ,   3   ,   5   ,   9   ,   7   ,   53  ,   27  ,   161 ,   343 ,   139 ,   1651    ,   1275    ,   2981    ,   9027    ,   31175   ,0 };
            const unsigned long dim2465KuoInit[]        =   {   1   ,   3   ,   3   ,   11  ,   7   ,   15  ,   81  ,   229 ,   115 ,   849 ,   459 ,   1651    ,   3867    ,   7029    ,   6741    ,0 };
            const unsigned long dim2466KuoInit[]        =   {   1   ,   1   ,   3   ,   5   ,   31  ,   45  ,   17  ,   107 ,   467 ,   401 ,   1473    ,   1953    ,   3779    ,   11237   ,   31257   ,0 };
            const unsigned long dim2467KuoInit[]        =   {   1   ,   3   ,   3   ,   11  ,   3   ,   27  ,   45  ,   217 ,   133 ,   515 ,   1487    ,   1427    ,   1999    ,   12979   ,   3543    ,0 };
            const unsigned long dim2468KuoInit[]        =   {   1   ,   3   ,   7   ,   3   ,   17  ,   31  ,   123 ,   205 ,   263 ,   963 ,   1697    ,   2453    ,   1249    ,   603 ,   8581    ,0 };
            const unsigned long dim2469KuoInit[]        =   {   1   ,   3   ,   1   ,   15  ,   23  ,   9   ,   23  ,   251 ,   489 ,   393 ,   1103    ,   3273    ,   5703    ,   3033    ,   22631   ,0 };
            const unsigned long dim2470KuoInit[]        =   {   1   ,   3   ,   3   ,   13  ,   23  ,   29  ,   49  ,   233 ,   461 ,   795 ,   991 ,   3695    ,   4715    ,   7211    ,   3695    ,0 };
            const unsigned long dim2471KuoInit[]        =   {   1   ,   3   ,   3   ,   15  ,   21  ,   11  ,   33  ,   69  ,   347 ,   535 ,   25  ,   3465    ,   2825    ,   9345    ,   4485    ,0 };
            const unsigned long dim2472KuoInit[]        =   {   1   ,   3   ,   7   ,   3   ,   23  ,   47  ,   65  ,   209 ,   357 ,   363 ,   1979    ,   289 ,   35  ,   4459    ,   17929   ,0 };
            const unsigned long dim2473KuoInit[]        =   {   1   ,   3   ,   1   ,   11  ,   19  ,   17  ,   125 ,   231 ,   221 ,   953 ,   2041    ,   2359    ,   5925    ,   15693   ,   30859   ,0 };
            const unsigned long dim2474KuoInit[]        =   {   1   ,   1   ,   3   ,   13  ,   31  ,   11  ,   11  ,   215 ,   335 ,   153 ,   1459    ,   2303    ,   3471    ,   7415    ,   3783    ,0 };
            const unsigned long dim2475KuoInit[]        =   {   1   ,   1   ,   7   ,   5   ,   11  ,   55  ,   47  ,   43  ,   355 ,   469 ,   963 ,   383 ,   667 ,   2233    ,   27759   ,0 };
            const unsigned long dim2476KuoInit[]        =   {   1   ,   3   ,   1   ,   9   ,   23  ,   29  ,   27  ,   135 ,   359 ,   111 ,   387 ,   3021    ,   817 ,   4211    ,   17243   ,0 };
            const unsigned long dim2477KuoInit[]        =   {   1   ,   1   ,   3   ,   9   ,   25  ,   27  ,   63  ,   89  ,   195 ,   359 ,   413 ,   3509    ,   3831    ,   4785    ,   30097   ,0 };
            const unsigned long dim2478KuoInit[]        =   {   1   ,   1   ,   5   ,   5   ,   11  ,   29  ,   29  ,   25  ,   445 ,   643 ,   261 ,   3901    ,   3019    ,   14213   ,   16115   ,0 };
            const unsigned long dim2479KuoInit[]        =   {   1   ,   3   ,   3   ,   15  ,   21  ,   31  ,   7   ,   255 ,   207 ,   535 ,   1033    ,   1017    ,   3791    ,   11455   ,   31701   ,0 };
            const unsigned long dim2480KuoInit[]        =   {   1   ,   3   ,   5   ,   13  ,   15  ,   37  ,   79  ,   205 ,   439 ,   609 ,   1433    ,   3333    ,   6583    ,   9689    ,   14979   ,0 };
            const unsigned long dim2481KuoInit[]        =   {   1   ,   1   ,   7   ,   3   ,   31  ,   27  ,   53  ,   101 ,   427 ,   773 ,   1877    ,   1715    ,   6177    ,   11851   ,   17989   ,0 };
            const unsigned long dim2482KuoInit[]        =   {   1   ,   3   ,   1   ,   3   ,   21  ,   11  ,   101 ,   203 ,   49  ,   543 ,   1979    ,   1469    ,   6701    ,   9963    ,   7793    ,0 };
            const unsigned long dim2483KuoInit[]        =   {   1   ,   1   ,   3   ,   1   ,   15  ,   47  ,   123 ,   141 ,   433 ,   783 ,   211 ,   1489    ,   4983    ,   6417    ,   29767   ,0 };
            const unsigned long dim2484KuoInit[]        =   {   1   ,   1   ,   5   ,   3   ,   1   ,   41  ,   109 ,   13  ,   465 ,   959 ,   431 ,   543 ,   4681    ,   8573    ,   7889    ,0 };
            const unsigned long dim2485KuoInit[]        =   {   1   ,   1   ,   1   ,   3   ,   15  ,   43  ,   15  ,   63  ,   145 ,   859 ,   939 ,   629 ,   7077    ,   5617    ,   16779   ,0 };
            const unsigned long dim2486KuoInit[]        =   {   1   ,   3   ,   3   ,   1   ,   23  ,   63  ,   67  ,   139 ,   219 ,   257 ,   943 ,   1501    ,   827 ,   10855   ,   10035   ,0 };
            const unsigned long dim2487KuoInit[]        =   {   1   ,   3   ,   1   ,   15  ,   9   ,   47  ,   27  ,   119 ,   231 ,   767 ,   999 ,   3553    ,   7683    ,   5399    ,   11553   ,0 };
            const unsigned long dim2488KuoInit[]        =   {   1   ,   3   ,   3   ,   11  ,   15  ,   13  ,   35  ,   139 ,   463 ,   985 ,   553 ,   3997    ,   2935    ,   7707    ,   2511    ,0 };
            const unsigned long dim2489KuoInit[]        =   {   1   ,   1   ,   3   ,   11  ,   23  ,   53  ,   103 ,   1   ,   409 ,   853 ,   737 ,   2077    ,   523 ,   1839    ,   32145   ,0 };
            const unsigned long dim2490KuoInit[]        =   {   1   ,   3   ,   1   ,   15  ,   23  ,   7   ,   77  ,   185 ,   371 ,   447 ,   499 ,   569 ,   6647    ,   501 ,   27649   ,0 };
            const unsigned long dim2491KuoInit[]        =   {   1   ,   1   ,   3   ,   5   ,   19  ,   31  ,   119 ,   157 ,   385 ,   427 ,   1411    ,   4031    ,   4199    ,   14567   ,   8133    ,0 };
            const unsigned long dim2492KuoInit[]        =   {   1   ,   3   ,   7   ,   5   ,   31  ,   3   ,   123 ,   149 ,   209 ,   467 ,   1619    ,   879 ,   5949    ,   7295    ,   25165   ,0 };
            const unsigned long dim2493KuoInit[]        =   {   1   ,   3   ,   1   ,   15  ,   31  ,   47  ,   111 ,   129 ,   359 ,   385 ,   463 ,   3363    ,   6823    ,   10751   ,   4473    ,0 };
            const unsigned long dim2494KuoInit[]        =   {   1   ,   3   ,   3   ,   7   ,   5   ,   39  ,   71  ,   45  ,   439 ,   161 ,   597 ,   2093    ,   1725    ,   11285   ,   18149   ,0 };
            const unsigned long dim2495KuoInit[]        =   {   1   ,   3   ,   3   ,   7   ,   21  ,   55  ,   115 ,   199 ,   339 ,   425 ,   71  ,   305 ,   1161    ,   3359    ,   16835   ,0 };
            const unsigned long dim2496KuoInit[]        =   {   1   ,   1   ,   7   ,   7   ,   1   ,   53  ,   51  ,   31  ,   57  ,   581 ,   147 ,   903 ,   7269    ,   1461    ,   7417    ,0 };
            const unsigned long dim2497KuoInit[]        =   {   1   ,   3   ,   7   ,   3   ,   17  ,   31  ,   65  ,   195 ,   47  ,   575 ,   1877    ,   1603    ,   3043    ,   5385    ,   15177   ,0 };
            const unsigned long dim2498KuoInit[]        =   {   1   ,   1   ,   3   ,   11  ,   15  ,   7   ,   121 ,   51  ,   67  ,   391 ,   1553    ,   1817    ,   6503    ,   2907    ,   3611    ,0 };
            const unsigned long dim2499KuoInit[]        =   {   1   ,   1   ,   1   ,   9   ,   13  ,   51  ,   95  ,   19  ,   315 ,   549 ,   1417    ,   2073    ,   485 ,   5257    ,   32419   ,0 };
            const unsigned long dim2500KuoInit[]        =   {   1   ,   3   ,   1   ,   9   ,   31  ,   63  ,   107 ,   103 ,   173 ,   681 ,   211 ,   3283    ,   2413    ,   2073    ,   24497   ,0 };
            const unsigned long dim2501KuoInit[]        =   {   1   ,   3   ,   1   ,   7   ,   3   ,   41  ,   49  ,   81  ,   289 ,   177 ,   699 ,   1305    ,   7867    ,   4699    ,   27669   ,0 };
            const unsigned long dim2502KuoInit[]        =   {   1   ,   1   ,   5   ,   3   ,   7   ,   15  ,   73  ,   57  ,   319 ,   959 ,   701 ,   2071    ,   1587    ,   4701    ,   2595    ,0 };
            const unsigned long dim2503KuoInit[]        =   {   1   ,   1   ,   5   ,   7   ,   7   ,   1   ,   29  ,   129 ,   361 ,   619 ,   1405    ,   2185    ,   6631    ,   3505    ,   391 ,0 };
            const unsigned long dim2504KuoInit[]        =   {   1   ,   3   ,   3   ,   5   ,   19  ,   9   ,   31  ,   63  ,   149 ,   181 ,   299 ,   2755    ,   1511    ,   5745    ,   20323   ,0 };
            const unsigned long dim2505KuoInit[]        =   {   1   ,   1   ,   3   ,   11  ,   23  ,   5   ,   29  ,   175 ,   119 ,   25  ,   319 ,   835 ,   1339    ,   5429    ,   7175    ,0 };
            const unsigned long dim2506KuoInit[]        =   {   1   ,   1   ,   3   ,   3   ,   31  ,   57  ,   55  ,   181 ,   473 ,   229 ,   309 ,   1567    ,   5063    ,   9745    ,   3713    ,0 };
            const unsigned long dim2507KuoInit[]        =   {   1   ,   3   ,   3   ,   13  ,   1   ,   15  ,   23  ,   175 ,   71  ,   53  ,   1749    ,   743 ,   285 ,   49  ,   7389    ,0 };
            const unsigned long dim2508KuoInit[]        =   {   1   ,   3   ,   1   ,   5   ,   7   ,   39  ,   75  ,   177 ,   259 ,   975 ,   257 ,   1497    ,   1565    ,   6973    ,   3827    ,0 };
            const unsigned long dim2509KuoInit[]        =   {   1   ,   3   ,   7   ,   13  ,   1   ,   39  ,   43  ,   213 ,   291 ,   941 ,   613 ,   3041    ,   1859    ,   10755   ,   29831   ,0 };
            const unsigned long dim2510KuoInit[]        =   {   1   ,   3   ,   3   ,   9   ,   17  ,   43  ,   15  ,   135 ,   473 ,   989 ,   429 ,   3287    ,   3699    ,   14349   ,   16311   ,0 };
            const unsigned long dim2511KuoInit[]        =   {   1   ,   3   ,   1   ,   15  ,   1   ,   51  ,   59  ,   203 ,   383 ,   731 ,   171 ,   1259    ,   5833    ,   12029   ,   18099   ,0 };
            const unsigned long dim2512KuoInit[]        =   {   1   ,   3   ,   1   ,   3   ,   19  ,   63  ,   125 ,   57  ,   429 ,   93  ,   1179    ,   2707    ,   449 ,   2213    ,   4759    ,0 };
            const unsigned long dim2513KuoInit[]        =   {   1   ,   1   ,   7   ,   3   ,   19  ,   11  ,   107 ,   127 ,   359 ,   759 ,   503 ,   761 ,   6817    ,   15343   ,   13777   ,0 };
            const unsigned long dim2514KuoInit[]        =   {   1   ,   1   ,   1   ,   7   ,   21  ,   49  ,   67  ,   179 ,   241 ,   403 ,   769 ,   3341    ,   1845    ,   2855    ,   3825    ,0 };
            const unsigned long dim2515KuoInit[]        =   {   1   ,   3   ,   3   ,   3   ,   7   ,   29  ,   71  ,   35  ,   389 ,   297 ,   673 ,   1353    ,   6781    ,   5633    ,   12459   ,0 };
            const unsigned long dim2516KuoInit[]        =   {   1   ,   1   ,   1   ,   15  ,   13  ,   15  ,   11  ,   57  ,   143 ,   35  ,   809 ,   2295    ,   6497    ,   455 ,   23689   ,0 };
            const unsigned long dim2517KuoInit[]        =   {   1   ,   3   ,   7   ,   15  ,   23  ,   39  ,   39  ,   211 ,   335 ,   537 ,   1569    ,   3989    ,   5777    ,   12951   ,   22841   ,0 };
            const unsigned long dim2518KuoInit[]        =   {   1   ,   1   ,   7   ,   3   ,   23  ,   17  ,   107 ,   237 ,   241 ,   533 ,   57  ,   3421    ,   263 ,   889 ,   28841   ,0 };
            const unsigned long dim2519KuoInit[]        =   {   1   ,   1   ,   5   ,   1   ,   11  ,   29  ,   17  ,   141 ,   431 ,   809 ,   1227    ,   2835    ,   2955    ,   11823   ,   23061   ,0 };
            const unsigned long dim2520KuoInit[]        =   {   1   ,   1   ,   3   ,   9   ,   5   ,   51  ,   73  ,   43  ,   75  ,   985 ,   1665    ,   1369    ,   519 ,   6379    ,   8665    ,0 };
            const unsigned long dim2521KuoInit[]        =   {   1   ,   3   ,   3   ,   13  ,   17  ,   21  ,   9   ,   173 ,   333 ,   689 ,   277 ,   1755    ,   5173    ,   7651    ,   30597   ,0 };
            const unsigned long dim2522KuoInit[]        =   {   1   ,   3   ,   5   ,   9   ,   3   ,   59  ,   59  ,   221 ,   177 ,   819 ,   1483    ,   1945    ,   4725    ,   4821    ,   3709    ,0 };
            const unsigned long dim2523KuoInit[]        =   {   1   ,   1   ,   7   ,   7   ,   19  ,   35  ,   93  ,   47  ,   49  ,   861 ,   561 ,   753 ,   8047    ,   419 ,   4335    ,0 };
            const unsigned long dim2524KuoInit[]        =   {   1   ,   3   ,   3   ,   15  ,   31  ,   3   ,   97  ,   177 ,   443 ,   55  ,   1541    ,   2627    ,   6129    ,   4165    ,   3079    ,0 };
            const unsigned long dim2525KuoInit[]        =   {   1   ,   1   ,   1   ,   9   ,   13  ,   13  ,   13  ,   29  ,   309 ,   679 ,   1893    ,   3629    ,   513 ,   16333   ,   25371   ,0 };
            const unsigned long dim2526KuoInit[]        =   {   1   ,   3   ,   1   ,   5   ,   3   ,   23  ,   85  ,   169 ,   143 ,   537 ,   1387    ,   3997    ,   3659    ,   12107   ,   11689   ,0 };
            const unsigned long dim2527KuoInit[]        =   {   1   ,   1   ,   1   ,   5   ,   9   ,   7   ,   11  ,   59  ,   325 ,   855 ,   1853    ,   855 ,   5743    ,   7625    ,   10381   ,0 };
            const unsigned long dim2528KuoInit[]        =   {   1   ,   3   ,   3   ,   3   ,   5   ,   37  ,   109 ,   101 ,   243 ,   301 ,   1849    ,   1633    ,   4451    ,   8859    ,   22017   ,0 };
            const unsigned long dim2529KuoInit[]        =   {   1   ,   3   ,   5   ,   1   ,   7   ,   1   ,   9   ,   1   ,   411 ,   941 ,   153 ,   2709    ,   389 ,   9403    ,   24103   ,0 };
            const unsigned long dim2530KuoInit[]        =   {   1   ,   1   ,   1   ,   15  ,   17  ,   19  ,   105 ,   127 ,   271 ,   543 ,   449 ,   1385    ,   7387    ,   8335    ,   741 ,0 };
            const unsigned long dim2531KuoInit[]        =   {   1   ,   3   ,   7   ,   5   ,   23  ,   21  ,   43  ,   17  ,   357 ,   499 ,   649 ,   1121    ,   7655    ,   15623   ,   12975   ,0 };
            const unsigned long dim2532KuoInit[]        =   {   1   ,   1   ,   5   ,   5   ,   1   ,   57  ,   37  ,   91  ,   159 ,   1013    ,   653 ,   3371    ,   4857    ,   441 ,   17729   ,0 };
            const unsigned long dim2533KuoInit[]        =   {   1   ,   3   ,   7   ,   15  ,   15  ,   31  ,   71  ,   77  ,   113 ,   47  ,   1597    ,   43  ,   1139    ,   3191    ,   31775   ,0 };
            const unsigned long dim2534KuoInit[]        =   {   1   ,   1   ,   1   ,   13  ,   19  ,   39  ,   93  ,   15  ,   433 ,   867 ,   1091    ,   2333    ,   6113    ,   13005   ,   13309   ,0 };
            const unsigned long dim2535KuoInit[]        =   {   1   ,   1   ,   1   ,   9   ,   5   ,   57  ,   21  ,   7   ,   267 ,   79  ,   1829    ,   43  ,   211 ,   2795    ,   18739   ,0 };
            const unsigned long dim2536KuoInit[]        =   {   1   ,   1   ,   5   ,   13  ,   25  ,   59  ,   93  ,   213 ,   23  ,   281 ,   625 ,   439 ,   1449    ,   2259    ,   31713   ,0 };
            const unsigned long dim2537KuoInit[]        =   {   1   ,   1   ,   1   ,   3   ,   9   ,   13  ,   41  ,   171 ,   365 ,   907 ,   395 ,   3465    ,   6433    ,   33  ,   17191   ,0 };
            const unsigned long dim2538KuoInit[]        =   {   1   ,   1   ,   1   ,   5   ,   3   ,   41  ,   45  ,   55  ,   413 ,   579 ,   1215    ,   3983    ,   1201    ,   3019    ,   16401   ,0 };
            const unsigned long dim2539KuoInit[]        =   {   1   ,   1   ,   3   ,   3   ,   9   ,   25  ,   111 ,   33  ,   359 ,   427 ,   63  ,   4005    ,   2753    ,   15063   ,   1925    ,0 };
            const unsigned long dim2540KuoInit[]        =   {   1   ,   1   ,   5   ,   3   ,   21  ,   55  ,   15  ,   5   ,   121 ,   867 ,   1645    ,   2387    ,   1167    ,   13865   ,   15899   ,0 };
            const unsigned long dim2541KuoInit[]        =   {   1   ,   3   ,   5   ,   15  ,   13  ,   49  ,   9   ,   173 ,   291 ,   517 ,   143 ,   399 ,   7545    ,   7225    ,   29753   ,0 };
            const unsigned long dim2542KuoInit[]        =   {   1   ,   1   ,   3   ,   5   ,   29  ,   23  ,   105 ,   165 ,   35  ,   325 ,   983 ,   215 ,   2537    ,   8661    ,   1715    ,0 };
            const unsigned long dim2543KuoInit[]        =   {   1   ,   3   ,   7   ,   7   ,   19  ,   1   ,   91  ,   77  ,   311 ,   715 ,   347 ,   1705    ,   1045    ,   15003   ,   10961   ,0 };
            const unsigned long dim2544KuoInit[]        =   {   1   ,   1   ,   1   ,   15  ,   5   ,   3   ,   73  ,   159 ,   113 ,   201 ,   257 ,   181 ,   551 ,   1189    ,   30719   ,0 };
            const unsigned long dim2545KuoInit[]        =   {   1   ,   1   ,   1   ,   3   ,   15  ,   57  ,   69  ,   85  ,   133 ,   735 ,   429 ,   2087    ,   1453    ,   14277   ,   15245   ,0 };
            const unsigned long dim2546KuoInit[]        =   {   1   ,   3   ,   1   ,   1   ,   27  ,   5   ,   27  ,   247 ,   401 ,   841 ,   1723    ,   3717    ,   3545    ,   3123    ,   27441   ,0 };
            const unsigned long dim2547KuoInit[]        =   {   1   ,   3   ,   1   ,   15  ,   5   ,   11  ,   97  ,   117 ,   331 ,   843 ,   1101    ,   2977    ,   5123    ,   5761    ,   2019    ,0 };
            const unsigned long dim2548KuoInit[]        =   {   1   ,   3   ,   1   ,   5   ,   31  ,   25  ,   53  ,   191 ,   265 ,   961 ,   1695    ,   1847    ,   6377    ,   13261   ,   8853    ,0 };
            const unsigned long dim2549KuoInit[]        =   {   1   ,   3   ,   1   ,   5   ,   5   ,   29  ,   85  ,   3   ,   127 ,   407 ,   1511    ,   549 ,   4015    ,   15327   ,   9167    ,0 };
            const unsigned long dim2550KuoInit[]        =   {   1   ,   3   ,   5   ,   3   ,   9   ,   11  ,   1   ,   217 ,   53  ,   201 ,   1237    ,   2663    ,   7429    ,   7349    ,   5619    ,0 };
            const unsigned long dim2551KuoInit[]        =   {   1   ,   3   ,   3   ,   15  ,   13  ,   63  ,   21  ,   173 ,   59  ,   881 ,   351 ,   1285    ,   7155    ,   6665    ,   25117   ,0 };
            const unsigned long dim2552KuoInit[]        =   {   1   ,   1   ,   3   ,   3   ,   3   ,   43  ,   103 ,   227 ,   397 ,   75  ,   1211    ,   2959    ,   3507    ,   3105    ,   21739   ,0 };
            const unsigned long dim2553KuoInit[]        =   {   1   ,   1   ,   3   ,   11  ,   1   ,   1   ,   5   ,   3   ,   41  ,   243 ,   1317    ,   2205    ,   4079    ,   7447    ,   22493   ,0 };
            const unsigned long dim2554KuoInit[]        =   {   1   ,   3   ,   1   ,   3   ,   7   ,   39  ,   23  ,   171 ,   263 ,   375 ,   805 ,   2309    ,   7123    ,   4935    ,   4517    ,0 };
            const unsigned long dim2555KuoInit[]        =   {   1   ,   1   ,   5   ,   15  ,   5   ,   47  ,   35  ,   15  ,   131 ,   875 ,   1771    ,   3001    ,   951 ,   14269   ,   6395    ,0 };
            const unsigned long dim2556KuoInit[]        =   {   1   ,   1   ,   3   ,   9   ,   27  ,   1   ,   59  ,   247 ,   295 ,   229 ,   309 ,   213 ,   6661    ,   10415   ,   17855   ,0 };
            const unsigned long dim2557KuoInit[]        =   {   1   ,   3   ,   5   ,   1   ,   15  ,   61  ,   47  ,   175 ,   279 ,   221 ,   1665    ,   3475    ,   4171    ,   15515   ,   30835   ,0 };
            const unsigned long dim2558KuoInit[]        =   {   1   ,   3   ,   3   ,   15  ,   29  ,   21  ,   61  ,   123 ,   57  ,   839 ,   399 ,   3069    ,   4617    ,   7051    ,   3245    ,0 };
            const unsigned long dim2559KuoInit[]        =   {   1   ,   3   ,   7   ,   11  ,   27  ,   57  ,   83  ,   163 ,   499 ,   901 ,   855 ,   2049    ,   2993    ,   10323   ,   3391    ,0 };
            const unsigned long dim2560KuoInit[]        =   {   1   ,   1   ,   7   ,   5   ,   17  ,   5   ,   23  ,   51  ,   445 ,   639 ,   811 ,   139 ,   2863    ,   9447    ,   25971   ,0 };
            const unsigned long dim2561KuoInit[]        =   {   1   ,   3   ,   3   ,   13  ,   31  ,   43  ,   75  ,   251 ,   9   ,   239 ,   1573    ,   871 ,   2211    ,   15057   ,   879 ,0 };
            const unsigned long dim2562KuoInit[]        =   {   1   ,   1   ,   5   ,   11  ,   23  ,   39  ,   69  ,   79  ,   335 ,   729 ,   1807    ,   1497    ,   2363    ,   8713    ,   3987    ,0 };
            const unsigned long dim2563KuoInit[]        =   {   1   ,   3   ,   5   ,   11  ,   27  ,   9   ,   1   ,   5   ,   207 ,   217 ,   1097    ,   2737    ,   6815    ,   9969    ,   21731   ,0 };
            const unsigned long dim2564KuoInit[]        =   {   1   ,   3   ,   3   ,   15  ,   9   ,   33  ,   71  ,   85  ,   15  ,   323 ,   1955    ,   1299    ,   4439    ,   6413    ,   23055   ,0 };
            const unsigned long dim2565KuoInit[]        =   {   1   ,   3   ,   5   ,   13  ,   15  ,   11  ,   75  ,   201 ,   5   ,   69  ,   1067    ,   2287    ,   8099    ,   11293   ,   22547   ,0 };
            const unsigned long dim2566KuoInit[]        =   {   1   ,   3   ,   1   ,   7   ,   19  ,   1   ,   63  ,   195 ,   463 ,   463 ,   1741    ,   915 ,   1289    ,   11275   ,   16383   ,0 };
            const unsigned long dim2567KuoInit[]        =   {   1   ,   1   ,   7   ,   3   ,   3   ,   53  ,   51  ,   95  ,   389 ,   385 ,   1621    ,   955 ,   137 ,   8523    ,   24087   ,0 };
            const unsigned long dim2568KuoInit[]        =   {   1   ,   3   ,   5   ,   11  ,   25  ,   27  ,   49  ,   27  ,   459 ,   979 ,   277 ,   2917    ,   6639    ,   3471    ,   20407   ,0 };
            const unsigned long dim2569KuoInit[]        =   {   1   ,   1   ,   5   ,   3   ,   19  ,   25  ,   75  ,   47  ,   395 ,   269 ,   1445    ,   287 ,   4563    ,   8817    ,   10645   ,0 };
            const unsigned long dim2570KuoInit[]        =   {   1   ,   1   ,   1   ,   15  ,   5   ,   9   ,   59  ,   177 ,   463 ,   683 ,   1917    ,   1189    ,   2423    ,   6967    ,   433 ,0 };
            const unsigned long dim2571KuoInit[]        =   {   1   ,   3   ,   5   ,   7   ,   11  ,   53  ,   101 ,   81  ,   89  ,   419 ,   1605    ,   3719    ,   5449    ,   4303    ,   11235   ,0 };
            const unsigned long dim2572KuoInit[]        =   {   1   ,   1   ,   7   ,   15  ,   7   ,   55  ,   5   ,   133 ,   273 ,   757 ,   313 ,   3451    ,   7837    ,   14543   ,   1607    ,0 };
            const unsigned long dim2573KuoInit[]        =   {   1   ,   3   ,   7   ,   11  ,   9   ,   13  ,   103 ,   95  ,   301 ,   637 ,   1951    ,   2577    ,   2571    ,   7881    ,   27057   ,0 };
            const unsigned long dim2574KuoInit[]        =   {   1   ,   3   ,   3   ,   9   ,   25  ,   17  ,   119 ,   11  ,   7   ,   141 ,   1219    ,   3091    ,   1737    ,   16355   ,   2807    ,0 };
            const unsigned long dim2575KuoInit[]        =   {   1   ,   1   ,   1   ,   1   ,   3   ,   55  ,   59  ,   193 ,   23  ,   643 ,   159 ,   3393    ,   2671    ,   14737   ,   21193   ,0 };
            const unsigned long dim2576KuoInit[]        =   {   1   ,   3   ,   5   ,   9   ,   9   ,   11  ,   109 ,   227 ,   333 ,   861 ,   915 ,   3239    ,   6199    ,   9555    ,   28563   ,0 };
            const unsigned long dim2577KuoInit[]        =   {   1   ,   3   ,   3   ,   9   ,   1   ,   33  ,   51  ,   169 ,   291 ,   159 ,   1983    ,   173 ,   4133    ,   11477   ,   31525   ,0 };
            const unsigned long dim2578KuoInit[]        =   {   1   ,   3   ,   5   ,   5   ,   31  ,   39  ,   47  ,   195 ,   421 ,   949 ,   311 ,   1537    ,   8009    ,   6273    ,   10893   ,0 };
            const unsigned long dim2579KuoInit[]        =   {   1   ,   1   ,   1   ,   1   ,   17  ,   19  ,   37  ,   211 ,   245 ,   553 ,   769 ,   2131    ,   2819    ,   16119   ,   25851   ,0 };
            const unsigned long dim2580KuoInit[]        =   {   1   ,   3   ,   3   ,   13  ,   7   ,   3   ,   63  ,   141 ,   465 ,   367 ,   949 ,   3997    ,   5217    ,   3425    ,   31161   ,0 };
            const unsigned long dim2581KuoInit[]        =   {   1   ,   3   ,   3   ,   5   ,   19  ,   35  ,   91  ,   115 ,   153 ,   127 ,   231 ,   3723    ,   7911    ,   4421    ,   2271    ,0 };
            const unsigned long dim2582KuoInit[]        =   {   1   ,   3   ,   5   ,   9   ,   17  ,   57  ,   49  ,   151 ,   263 ,   89  ,   1635    ,   2991    ,   7011    ,   13243   ,   1473    ,0 };
            const unsigned long dim2583KuoInit[]        =   {   1   ,   1   ,   1   ,   15  ,   21  ,   59  ,   53  ,   179 ,   113 ,   839 ,   649 ,   2083    ,   7015    ,   5507    ,   9539    ,0 };
            const unsigned long dim2584KuoInit[]        =   {   1   ,   1   ,   1   ,   1   ,   13  ,   49  ,   97  ,   81  ,   255 ,   973 ,   275 ,   2005    ,   5493    ,   11019   ,   14427   ,0 };
            const unsigned long dim2585KuoInit[]        =   {   1   ,   1   ,   7   ,   5   ,   5   ,   53  ,   27  ,   197 ,   269 ,   699 ,   285 ,   3725    ,   7681    ,   3767    ,   11881   ,0 };
            const unsigned long dim2586KuoInit[]        =   {   1   ,   1   ,   1   ,   13  ,   15  ,   15  ,   11  ,   245 ,   419 ,   559 ,   727 ,   1545    ,   1821    ,   5655    ,   26971   ,0 };
            const unsigned long dim2587KuoInit[]        =   {   1   ,   1   ,   3   ,   3   ,   5   ,   1   ,   77  ,   211 ,   139 ,   949 ,   1945    ,   3327    ,   6699    ,   6227    ,   9769    ,0 };
            const unsigned long dim2588KuoInit[]        =   {   1   ,   3   ,   1   ,   1   ,   3   ,   9   ,   113 ,   17  ,   283 ,   939 ,   1941    ,   19  ,   7431    ,   14503   ,   24541   ,0 };
            const unsigned long dim2589KuoInit[]        =   {   1   ,   1   ,   1   ,   3   ,   1   ,   57  ,   17  ,   85  ,   371 ,   491 ,   775 ,   2093    ,   5615    ,   4977    ,   30995   ,0 };
            const unsigned long dim2590KuoInit[]        =   {   1   ,   1   ,   3   ,   3   ,   1   ,   61  ,   75  ,   163 ,   307 ,   527 ,   1821    ,   1291    ,   5761    ,   9983    ,   4647    ,0 };
            const unsigned long dim2591KuoInit[]        =   {   1   ,   1   ,   1   ,   13  ,   7   ,   61  ,   9   ,   203 ,   255 ,   293 ,   1683    ,   1081    ,   3179    ,   13283   ,   13759   ,0 };
            const unsigned long dim2592KuoInit[]        =   {   1   ,   1   ,   5   ,   5   ,   5   ,   25  ,   23  ,   97  ,   37  ,   277 ,   1787    ,   2709    ,   6201    ,   8041    ,   2301    ,0 };
            const unsigned long dim2593KuoInit[]        =   {   1   ,   3   ,   1   ,   1   ,   21  ,   57  ,   33  ,   13  ,   307 ,   427 ,   1577    ,   2043    ,   8131    ,   11723   ,   14353   ,0 };
            const unsigned long dim2594KuoInit[]        =   {   1   ,   1   ,   3   ,   5   ,   19  ,   39  ,   63  ,   11  ,   139 ,   173 ,   517 ,   1597    ,   3871    ,   15533   ,   4123    ,0 };
            const unsigned long dim2595KuoInit[]        =   {   1   ,   3   ,   1   ,   11  ,   19  ,   43  ,   123 ,   127 ,   237 ,   599 ,   893 ,   3645    ,   947 ,   12565   ,   9367    ,0 };
            const unsigned long dim2596KuoInit[]        =   {   1   ,   1   ,   7   ,   11  ,   27  ,   9   ,   119 ,   251 ,   499 ,   777 ,   1667    ,   1699    ,   3431    ,   2727    ,   7097    ,0 };
            const unsigned long dim2597KuoInit[]        =   {   1   ,   1   ,   1   ,   11  ,   7   ,   25  ,   95  ,   143 ,   319 ,   659 ,   1567    ,   1391    ,   5149    ,   15177   ,   6583    ,0 };
            const unsigned long dim2598KuoInit[]        =   {   1   ,   3   ,   1   ,   13  ,   7   ,   39  ,   125 ,   219 ,   405 ,   739 ,   1267    ,   3065    ,   7529    ,   3195    ,   21889   ,0 };
            const unsigned long dim2599KuoInit[]        =   {   1   ,   3   ,   3   ,   9   ,   27  ,   13  ,   87  ,   129 ,   451 ,   997 ,   103 ,   1667    ,   545 ,   6341    ,   3105    ,0 };
            const unsigned long dim2600KuoInit[]        =   {   1   ,   1   ,   5   ,   3   ,   11  ,   43  ,   95  ,   205 ,   445 ,   919 ,   1353    ,   2435    ,   269 ,   3221    ,   8983    ,0 };
            const unsigned long dim2601KuoInit[]        =   {   1   ,   1   ,   3   ,   1   ,   5   ,   47  ,   15  ,   127 ,   47  ,   89  ,   1409    ,   1935    ,   2047    ,   10629   ,   4867    ,0 };
            const unsigned long dim2602KuoInit[]        =   {   1   ,   1   ,   5   ,   9   ,   21  ,   51  ,   31  ,   175 ,   337 ,   643 ,   1233    ,   1049    ,   569 ,   3383    ,   18771   ,0 };
            const unsigned long dim2603KuoInit[]        =   {   1   ,   1   ,   7   ,   11  ,   7   ,   7   ,   29  ,   139 ,   65  ,   621 ,   363 ,   513 ,   4149    ,   5225    ,   11595   ,0 };
            const unsigned long dim2604KuoInit[]        =   {   1   ,   1   ,   7   ,   7   ,   11  ,   31  ,   43  ,   223 ,   249 ,   661 ,   123 ,   2725    ,   5179    ,   13237   ,   10483   ,0 };
            const unsigned long dim2605KuoInit[]        =   {   1   ,   1   ,   5   ,   3   ,   5   ,   35  ,   121 ,   193 ,   33  ,   343 ,   413 ,   1879    ,   8013    ,   11161   ,   7945    ,0 };
            const unsigned long dim2606KuoInit[]        =   {   1   ,   3   ,   7   ,   5   ,   27  ,   43  ,   123 ,   25  ,   45  ,   279 ,   979 ,   1121    ,   1041    ,   3757    ,   22217   ,0 };
            const unsigned long dim2607KuoInit[]        =   {   1   ,   3   ,   5   ,   1   ,   29  ,   45  ,   29  ,   215 ,   39  ,   679 ,   799 ,   1491    ,   885 ,   15475   ,   29269   ,0 };
            const unsigned long dim2608KuoInit[]        =   {   1   ,   1   ,   3   ,   1   ,   1   ,   5   ,   49  ,   181 ,   441 ,   691 ,   639 ,   859 ,   2785    ,   3237    ,   1703    ,0 };
            const unsigned long dim2609KuoInit[]        =   {   1   ,   1   ,   1   ,   9   ,   19  ,   1   ,   41  ,   131 ,   469 ,   999 ,   1665    ,   2309    ,   4447    ,   13851   ,   7063    ,0 };
            const unsigned long dim2610KuoInit[]        =   {   1   ,   3   ,   5   ,   3   ,   21  ,   53  ,   51  ,   155 ,   41  ,   43  ,   127 ,   1653    ,   6237    ,   10481   ,   22557   ,0 };
            const unsigned long dim2611KuoInit[]        =   {   1   ,   3   ,   5   ,   7   ,   31  ,   21  ,   39  ,   177 ,   283 ,   535 ,   771 ,   3783    ,   3033    ,   10971   ,   8691    ,0 };
            const unsigned long dim2612KuoInit[]        =   {   1   ,   1   ,   3   ,   1   ,   25  ,   57  ,   13  ,   139 ,   169 ,   717 ,   795 ,   2337    ,   7347    ,   6951    ,   20509   ,0 };
            const unsigned long dim2613KuoInit[]        =   {   1   ,   1   ,   5   ,   11  ,   1   ,   29  ,   27  ,   21  ,   259 ,   889 ,   83  ,   3295    ,   3699    ,   1523    ,   6227    ,0 };
            const unsigned long dim2614KuoInit[]        =   {   1   ,   3   ,   5   ,   3   ,   9   ,   37  ,   35  ,   95  ,   217 ,   335 ,   865 ,   3093    ,   3077    ,   9831    ,   12173   ,0 };
            const unsigned long dim2615KuoInit[]        =   {   1   ,   1   ,   7   ,   13  ,   31  ,   7   ,   69  ,   193 ,   459 ,   967 ,   481 ,   3271    ,   677 ,   13131   ,   22665   ,0 };
            const unsigned long dim2616KuoInit[]        =   {   1   ,   3   ,   3   ,   9   ,   13  ,   31  ,   43  ,   155 ,   465 ,   303 ,   1153    ,   1591    ,   7793    ,   921 ,   12805   ,0 };
            const unsigned long dim2617KuoInit[]        =   {   1   ,   1   ,   5   ,   5   ,   23  ,   59  ,   3   ,   151 ,   391 ,   729 ,   913 ,   899 ,   1793    ,   10173   ,   19681   ,0 };
            const unsigned long dim2618KuoInit[]        =   {   1   ,   3   ,   1   ,   5   ,   31  ,   21  ,   19  ,   219 ,   311 ,   1013    ,   227 ,   1813    ,   3103    ,   14191   ,   22995   ,0 };
            const unsigned long dim2619KuoInit[]        =   {   1   ,   1   ,   3   ,   5   ,   3   ,   59  ,   11  ,   191 ,   433 ,   63  ,   465 ,   3505    ,   6801    ,   3047    ,   16913   ,0 };
            const unsigned long dim2620KuoInit[]        =   {   1   ,   1   ,   7   ,   9   ,   19  ,   3   ,   95  ,   203 ,   1   ,   679 ,   1719    ,   269 ,   1747    ,   9465    ,   24445   ,0 };
            const unsigned long dim2621KuoInit[]        =   {   1   ,   1   ,   7   ,   9   ,   9   ,   17  ,   79  ,   45  ,   179 ,   863 ,   801 ,   2825    ,   5345    ,   637 ,   9729    ,0 };
            const unsigned long dim2622KuoInit[]        =   {   1   ,   3   ,   7   ,   11  ,   11  ,   59  ,   45  ,   107 ,   45  ,   891 ,   629 ,   443 ,   805 ,   15277   ,   1961    ,0 };
            const unsigned long dim2623KuoInit[]        =   {   1   ,   1   ,   3   ,   5   ,   9   ,   49  ,   123 ,   183 ,   13  ,   867 ,   261 ,   3883    ,   2479    ,   8221    ,   3327    ,0 };
            const unsigned long dim2624KuoInit[]        =   {   1   ,   1   ,   5   ,   11  ,   19  ,   19  ,   95  ,   85  ,   351 ,   861 ,   1499    ,   3697    ,   7369    ,   10757   ,   16445   ,0 };
            const unsigned long dim2625KuoInit[]        =   {   1   ,   3   ,   5   ,   5   ,   27  ,   13  ,   119 ,   243 ,   437 ,   655 ,   955 ,   1919    ,   3643    ,   12389   ,   6865    ,0 };
            const unsigned long dim2626KuoInit[]        =   {   1   ,   3   ,   3   ,   1   ,   23  ,   15  ,   43  ,   65  ,   87  ,   833 ,   503 ,   999 ,   703 ,   4411    ,   3343    ,0 };
            const unsigned long dim2627KuoInit[]        =   {   1   ,   3   ,   7   ,   1   ,   31  ,   13  ,   67  ,   199 ,   261 ,   257 ,   1767    ,   3983    ,   823 ,   149 ,   32353   ,0 };
            const unsigned long dim2628KuoInit[]        =   {   1   ,   3   ,   1   ,   15  ,   15  ,   17  ,   69  ,   237 ,   441 ,   657 ,   1125    ,   3377    ,   5971    ,   15477   ,   31363   ,0 };
            const unsigned long dim2629KuoInit[]        =   {   1   ,   3   ,   1   ,   13  ,   3   ,   55  ,   67  ,   93  ,   315 ,   1015    ,   61  ,   35  ,   3059    ,   15503   ,   16811   ,0 };
            const unsigned long dim2630KuoInit[]        =   {   1   ,   1   ,   3   ,   9   ,   1   ,   51  ,   91  ,   213 ,   3   ,   37  ,   849 ,   2329    ,   2237    ,   6189    ,   625 ,0 };
            const unsigned long dim2631KuoInit[]        =   {   1   ,   1   ,   5   ,   7   ,   25  ,   3   ,   57  ,   103 ,   401 ,   357 ,   461 ,   2677    ,   6059    ,   291 ,   11461   ,0 };
            const unsigned long dim2632KuoInit[]        =   {   1   ,   1   ,   5   ,   7   ,   23  ,   29  ,   119 ,   61  ,   121 ,   535 ,   251 ,   565 ,   7307    ,   1085    ,   8429    ,0 };
            const unsigned long dim2633KuoInit[]        =   {   1   ,   1   ,   3   ,   13  ,   23  ,   19  ,   63  ,   155 ,   5   ,   617 ,   9   ,   715 ,   5399    ,   1747    ,   17633   ,0 };
            const unsigned long dim2634KuoInit[]        =   {   1   ,   1   ,   1   ,   13  ,   3   ,   57  ,   13  ,   161 ,   123 ,   389 ,   1695    ,   4083    ,   1957    ,   11359   ,   18961   ,0 };
            const unsigned long dim2635KuoInit[]        =   {   1   ,   1   ,   3   ,   15  ,   27  ,   25  ,   65  ,   15  ,   205 ,   695 ,   995 ,   1471    ,   1131    ,   11435   ,   22107   ,0 };
            const unsigned long dim2636KuoInit[]        =   {   1   ,   3   ,   3   ,   11  ,   31  ,   29  ,   109 ,   121 ,   81  ,   15  ,   1087    ,   299 ,   3495    ,   14137   ,   19073   ,0 };
            const unsigned long dim2637KuoInit[]        =   {   1   ,   3   ,   3   ,   15  ,   23  ,   23  ,   117 ,   55  ,   215 ,   287 ,   1421    ,   731 ,   279 ,   15497   ,   15545   ,0 };
            const unsigned long dim2638KuoInit[]        =   {   1   ,   3   ,   5   ,   15  ,   1   ,   31  ,   95  ,   249 ,   403 ,   239 ,   771 ,   1279    ,   1863    ,   12893   ,   21043   ,0 };
            const unsigned long dim2639KuoInit[]        =   {   1   ,   3   ,   1   ,   9   ,   21  ,   23  ,   123 ,   169 ,   69  ,   591 ,   621 ,   1697    ,   5867    ,   5193    ,   16573   ,0 };
            const unsigned long dim2640KuoInit[]        =   {   1   ,   3   ,   3   ,   5   ,   29  ,   19  ,   105 ,   177 ,   143 ,   141 ,   1455    ,   3119    ,   347 ,   8797    ,   3925    ,0 };
            const unsigned long dim2641KuoInit[]        =   {   1   ,   3   ,   3   ,   15  ,   9   ,   11  ,   41  ,   169 ,   141 ,   555 ,   1761    ,   3441    ,   2245    ,   14205   ,   9769    ,0 };
            const unsigned long dim2642KuoInit[]        =   {   1   ,   1   ,   1   ,   5   ,   9   ,   33  ,   125 ,   203 ,   333 ,   461 ,   551 ,   569 ,   1305    ,   10213   ,   24549   ,0 };
            const unsigned long dim2643KuoInit[]        =   {   1   ,   1   ,   7   ,   13  ,   21  ,   9   ,   99  ,   43  ,   395 ,   667 ,   121 ,   1225    ,   7835    ,   8327    ,   17717   ,0 };
            const unsigned long dim2644KuoInit[]        =   {   1   ,   3   ,   3   ,   7   ,   21  ,   9   ,   31  ,   59  ,   321 ,   505 ,   1357    ,   3213    ,   4219    ,   10637   ,   11127   ,0 };
            const unsigned long dim2645KuoInit[]        =   {   1   ,   1   ,   1   ,   7   ,   17  ,   49  ,   77  ,   243 ,   189 ,   115 ,   1339    ,   2023    ,   7431    ,   5087    ,   24495   ,0 };
            const unsigned long dim2646KuoInit[]        =   {   1   ,   1   ,   1   ,   11  ,   1   ,   39  ,   41  ,   51  ,   315 ,   383 ,   237 ,   3877    ,   5731    ,   11  ,   21307   ,0 };
            const unsigned long dim2647KuoInit[]        =   {   1   ,   1   ,   7   ,   5   ,   5   ,   3   ,   17  ,   227 ,   483 ,   395 ,   1251    ,   1265    ,   3965    ,   15203   ,   17163   ,0 };
            const unsigned long dim2648KuoInit[]        =   {   1   ,   3   ,   7   ,   7   ,   23  ,   41  ,   53  ,   45  ,   43  ,   989 ,   629 ,   1557    ,   2891    ,   4613    ,   3647    ,0 };
            const unsigned long dim2649KuoInit[]        =   {   1   ,   3   ,   3   ,   5   ,   15  ,   39  ,   35  ,   127 ,   267 ,   45  ,   1299    ,   2251    ,   7363    ,   13595   ,   31061   ,0 };
            const unsigned long dim2650KuoInit[]        =   {   1   ,   3   ,   7   ,   15  ,   23  ,   45  ,   99  ,   153 ,   419 ,   727 ,   1077    ,   3407    ,   4861    ,   3697    ,   14479   ,0 };
            const unsigned long dim2651KuoInit[]        =   {   1   ,   3   ,   5   ,   15  ,   19  ,   21  ,   123 ,   173 ,   295 ,   47  ,   1977    ,   1185    ,   7403    ,   11901   ,   29753   ,0 };
            const unsigned long dim2652KuoInit[]        =   {   1   ,   1   ,   3   ,   13  ,   21  ,   53  ,   83  ,   151 ,   309 ,   293 ,   439 ,   2381    ,   7965    ,   3449    ,   8861    ,0 };
            const unsigned long dim2653KuoInit[]        =   {   1   ,   1   ,   1   ,   3   ,   31  ,   49  ,   61  ,   81  ,   327 ,   173 ,   111 ,   3893    ,   2621    ,   1409    ,   13255   ,0 };
            const unsigned long dim2654KuoInit[]        =   {   1   ,   3   ,   5   ,   5   ,   3   ,   13  ,   75  ,   71  ,   271 ,   689 ,   369 ,   2719    ,   759 ,   8691    ,   24611   ,0 };
            const unsigned long dim2655KuoInit[]        =   {   1   ,   1   ,   1   ,   5   ,   21  ,   7   ,   69  ,   17  ,   347 ,   699 ,   1581    ,   1095    ,   1727    ,   8213    ,   5033    ,0 };
            const unsigned long dim2656KuoInit[]        =   {   1   ,   1   ,   1   ,   1   ,   23  ,   41  ,   85  ,   233 ,   43  ,   863 ,   553 ,   2625    ,   2323    ,   7471    ,   12457   ,0 };
            const unsigned long dim2657KuoInit[]        =   {   1   ,   1   ,   3   ,   15  ,   3   ,   7   ,   99  ,   27  ,   173 ,   539 ,   113 ,   2309    ,   1771    ,   11709   ,   29703   ,0 };
            const unsigned long dim2658KuoInit[]        =   {   1   ,   3   ,   1   ,   1   ,   7   ,   43  ,   73  ,   91  ,   151 ,   961 ,   251 ,   2089    ,   4475    ,   2745    ,   6077    ,0 };
            const unsigned long dim2659KuoInit[]        =   {   1   ,   1   ,   3   ,   7   ,   15  ,   61  ,   91  ,   139 ,   161 ,   613 ,   603 ,   1587    ,   3995    ,   9241    ,   24467   ,0 };
            const unsigned long dim2660KuoInit[]        =   {   1   ,   1   ,   1   ,   3   ,   1   ,   21  ,   7   ,   157 ,   311 ,   451 ,   137 ,   3289    ,   5583    ,   5107    ,   24471   ,0 };
            const unsigned long dim2661KuoInit[]        =   {   1   ,   1   ,   7   ,   9   ,   15  ,   3   ,   7   ,   19  ,   375 ,   265 ,   1781    ,   335 ,   233 ,   3569    ,   21407   ,0 };
            const unsigned long dim2662KuoInit[]        =   {   1   ,   1   ,   5   ,   13  ,   7   ,   53  ,   81  ,   79  ,   355 ,   89  ,   1233    ,   267 ,   6523    ,   8819    ,   30595   ,0 };
            const unsigned long dim2663KuoInit[]        =   {   1   ,   1   ,   7   ,   7   ,   1   ,   59  ,   65  ,   93  ,   251 ,   565 ,   1575    ,   2709    ,   7845    ,   13355   ,   6043    ,0 };
            const unsigned long dim2664KuoInit[]        =   {   1   ,   1   ,   5   ,   7   ,   9   ,   21  ,   71  ,   97  ,   383 ,   189 ,   3   ,   1841    ,   1403    ,   9649    ,   25053   ,0 };
            const unsigned long dim2665KuoInit[]        =   {   1   ,   3   ,   5   ,   9   ,   25  ,   47  ,   41  ,   63  ,   177 ,   551 ,   1389    ,   1639    ,   4209    ,   15787   ,   26499   ,0 };
            const unsigned long dim2666KuoInit[]        =   {   1   ,   1   ,   3   ,   15  ,   29  ,   53  ,   119 ,   3   ,   33  ,   677 ,   207 ,   1239    ,   453 ,   6281    ,   26701   ,0 };
            const unsigned long dim2667KuoInit[]        =   {   1   ,   3   ,   5   ,   13  ,   7   ,   37  ,   87  ,   241 ,   477 ,   765 ,   1511    ,   2659    ,   1043    ,   313 ,   27957   ,0 };
            const unsigned long dim2668KuoInit[]        =   {   1   ,   3   ,   5   ,   3   ,   1   ,   31  ,   101 ,   27  ,   401 ,   547 ,   617 ,   3605    ,   4899    ,   5093    ,   19933   ,0 };
            const unsigned long dim2669KuoInit[]        =   {   1   ,   1   ,   7   ,   13  ,   7   ,   45  ,   127 ,   191 ,   439 ,   383 ,   745 ,   3143    ,   2083    ,   2941    ,   31981   ,0 };
            const unsigned long dim2670KuoInit[]        =   {   1   ,   3   ,   3   ,   13  ,   7   ,   15  ,   31  ,   203 ,   75  ,   297 ,   603 ,   2945    ,   1823    ,   12359   ,   29245   ,0 };
            const unsigned long dim2671KuoInit[]        =   {   1   ,   3   ,   7   ,   15  ,   1   ,   57  ,   35  ,   255 ,   77  ,   483 ,   1051    ,   1999    ,   4975    ,   16211   ,   22903   ,0 };
            const unsigned long dim2672KuoInit[]        =   {   1   ,   1   ,   5   ,   5   ,   23  ,   53  ,   13  ,   251 ,   181 ,   1011    ,   983 ,   3735    ,   4221    ,   5557    ,   25777   ,0 };
            const unsigned long dim2673KuoInit[]        =   {   1   ,   1   ,   1   ,   11  ,   13  ,   51  ,   21  ,   155 ,   293 ,   79  ,   1015    ,   2653    ,   2959    ,   9913    ,   17751   ,0 };
            const unsigned long dim2674KuoInit[]        =   {   1   ,   1   ,   5   ,   3   ,   17  ,   33  ,   27  ,   125 ,   437 ,   663 ,   343 ,   1799    ,   2129    ,   7993    ,   1373    ,0 };
            const unsigned long dim2675KuoInit[]        =   {   1   ,   3   ,   3   ,   3   ,   9   ,   17  ,   31  ,   123 ,   453 ,   853 ,   1221    ,   2863    ,   4637    ,   8239    ,   27233   ,0 };
            const unsigned long dim2676KuoInit[]        =   {   1   ,   1   ,   5   ,   1   ,   7   ,   19  ,   43  ,   77  ,   19  ,   559 ,   1397    ,   1631    ,   3187    ,   1445    ,   25437   ,0 };
            const unsigned long dim2677KuoInit[]        =   {   1   ,   1   ,   3   ,   5   ,   31  ,   21  ,   71  ,   127 ,   223 ,   127 ,   717 ,   3163    ,   2079    ,   2399    ,   16415   ,0 };
            const unsigned long dim2678KuoInit[]        =   {   1   ,   1   ,   1   ,   13  ,   17  ,   13  ,   99  ,   227 ,   127 ,   625 ,   1065    ,   1779    ,   7275    ,   11865   ,   17397   ,0 };
            const unsigned long dim2679KuoInit[]        =   {   1   ,   3   ,   3   ,   11  ,   31  ,   43  ,   27  ,   231 ,   103 ,   873 ,   949 ,   1179    ,   3149    ,   2251    ,   22993   ,0 };
            const unsigned long dim2680KuoInit[]        =   {   1   ,   3   ,   1   ,   11  ,   25  ,   11  ,   89  ,   149 ,   71  ,   171 ,   1985    ,   3779    ,   901 ,   16307   ,   5663    ,0 };
            const unsigned long dim2681KuoInit[]        =   {   1   ,   1   ,   3   ,   5   ,   1   ,   25  ,   125 ,   241 ,   167 ,   381 ,   407 ,   2801    ,   635 ,   313 ,   10199   ,0 };
            const unsigned long dim2682KuoInit[]        =   {   1   ,   1   ,   3   ,   15  ,   29  ,   31  ,   101 ,   121 ,   251 ,   685 ,   677 ,   1683    ,   3499    ,   16193   ,   7139    ,0 };
            const unsigned long dim2683KuoInit[]        =   {   1   ,   3   ,   5   ,   3   ,   7   ,   13  ,   27  ,   101 ,   7   ,   159 ,   2005    ,   3297    ,   2775    ,   11193   ,   17385   ,0 };
            const unsigned long dim2684KuoInit[]        =   {   1   ,   1   ,   5   ,   5   ,   25  ,   31  ,   119 ,   43  ,   317 ,   371 ,   1505    ,   2631    ,   1021    ,   1297    ,   22791   ,0 };
            const unsigned long dim2685KuoInit[]        =   {   1   ,   3   ,   5   ,   7   ,   17  ,   57  ,   63  ,   229 ,   97  ,   599 ,   1485    ,   1113    ,   403 ,   77  ,   25541   ,0 };
            const unsigned long dim2686KuoInit[]        =   {   1   ,   1   ,   7   ,   13  ,   11  ,   43  ,   29  ,   85  ,   335 ,   943 ,   1605    ,   1757    ,   2787    ,   3313    ,   14991   ,0 };
            const unsigned long dim2687KuoInit[]        =   {   1   ,   3   ,   7   ,   13  ,   17  ,   43  ,   5   ,   231 ,   441 ,   883 ,   465 ,   841 ,   2785    ,   8193    ,   7951    ,0 };
            const unsigned long dim2688KuoInit[]        =   {   1   ,   3   ,   1   ,   3   ,   27  ,   33  ,   1   ,   215 ,   295 ,   381 ,   283 ,   219 ,   3419    ,   3343    ,   17629   ,0 };
            const unsigned long dim2689KuoInit[]        =   {   1   ,   1   ,   5   ,   9   ,   5   ,   49  ,   107 ,   179 ,   281 ,   223 ,   1845    ,   2505    ,   2281    ,   12051   ,   12609   ,0 };
            const unsigned long dim2690KuoInit[]        =   {   1   ,   1   ,   7   ,   13  ,   21  ,   13  ,   109 ,   183 ,   385 ,   293 ,   773 ,   3975    ,   4737    ,   8143    ,   1335    ,0 };
            const unsigned long dim2691KuoInit[]        =   {   1   ,   1   ,   1   ,   13  ,   13  ,   47  ,   23  ,   27  ,   271 ,   957 ,   1937    ,   4091    ,   3563    ,   4055    ,   17595   ,0 };
            const unsigned long dim2692KuoInit[]        =   {   1   ,   3   ,   3   ,   15  ,   31  ,   45  ,   43  ,   7   ,   125 ,   549 ,   169 ,   1567    ,   5405    ,   3919    ,   21795   ,0 };
            const unsigned long dim2693KuoInit[]        =   {   1   ,   3   ,   5   ,   7   ,   5   ,   53  ,   121 ,   107 ,   217 ,   745 ,   861 ,   481 ,   5503    ,   3723    ,   12215   ,0 };
            const unsigned long dim2694KuoInit[]        =   {   1   ,   3   ,   7   ,   3   ,   23  ,   27  ,   91  ,   175 ,   477 ,   289 ,   817 ,   2483    ,   6071    ,   5429    ,   12923   ,0 };
            const unsigned long dim2695KuoInit[]        =   {   1   ,   1   ,   3   ,   15  ,   7   ,   63  ,   29  ,   157 ,   323 ,   851 ,   1941    ,   1927    ,   5165    ,   9875    ,   11809   ,0 };
            const unsigned long dim2696KuoInit[]        =   {   1   ,   3   ,   3   ,   1   ,   13  ,   17  ,   93  ,   187 ,   329 ,   187 ,   1417    ,   1477    ,   7891    ,   11093   ,   24705   ,0 };
            const unsigned long dim2697KuoInit[]        =   {   1   ,   3   ,   3   ,   13  ,   5   ,   41  ,   119 ,   135 ,   255 ,   47  ,   803 ,   3941    ,   1647    ,   7777    ,   17515   ,0 };
            const unsigned long dim2698KuoInit[]        =   {   1   ,   1   ,   5   ,   9   ,   25  ,   11  ,   15  ,   231 ,   81  ,   267 ,   1633    ,   407 ,   3319    ,   12033   ,   9651    ,0 };
            const unsigned long dim2699KuoInit[]        =   {   1   ,   3   ,   1   ,   1   ,   31  ,   43  ,   3   ,   101 ,   435 ,   869 ,   159 ,   441 ,   613 ,   10283   ,   6873    ,0 };
            const unsigned long dim2700KuoInit[]        =   {   1   ,   1   ,   5   ,   1   ,   11  ,   55  ,   7   ,   65  ,   357 ,   595 ,   1543    ,   3957    ,   6075    ,   7265    ,   27063   ,0 };
            const unsigned long dim2701KuoInit[]        =   {   1   ,   3   ,   3   ,   11  ,   23  ,   23  ,   41  ,   205 ,   197 ,   863 ,   695 ,   971 ,   39  ,   12389   ,   2983    ,0 };
            const unsigned long dim2702KuoInit[]        =   {   1   ,   1   ,   3   ,   15  ,   7   ,   39  ,   39  ,   25  ,   173 ,   277 ,   999 ,   3797    ,   3051    ,   189 ,   13559   ,0 };
            const unsigned long dim2703KuoInit[]        =   {   1   ,   1   ,   7   ,   5   ,   3   ,   7   ,   47  ,   79  ,   411 ,   109 ,   55  ,   3803    ,   6617    ,   14317   ,   9533    ,0 };
            const unsigned long dim2704KuoInit[]        =   {   1   ,   1   ,   3   ,   9   ,   31  ,   17  ,   53  ,   97  ,   495 ,   995 ,   1735    ,   1055    ,   6805    ,   6865    ,   23321   ,0 };
            const unsigned long dim2705KuoInit[]        =   {   1   ,   3   ,   7   ,   13  ,   3   ,   63  ,   37  ,   185 ,   311 ,   995 ,   1819    ,   2633    ,   2593    ,   2107    ,   15755   ,0 };
            const unsigned long dim2706KuoInit[]        =   {   1   ,   3   ,   7   ,   11  ,   25  ,   37  ,   79  ,   179 ,   435 ,   167 ,   147 ,   2159    ,   5909    ,   10427   ,   12487   ,0 };
            const unsigned long dim2707KuoInit[]        =   {   1   ,   3   ,   5   ,   7   ,   31  ,   9   ,   69  ,   93  ,   477 ,   283 ,   479 ,   1895    ,   5693    ,   4809    ,   3157    ,0 };
            const unsigned long dim2708KuoInit[]        =   {   1   ,   1   ,   3   ,   11  ,   27  ,   21  ,   11  ,   3   ,   157 ,   1011    ,   1473    ,   2393    ,   259 ,   4155    ,   28375   ,0 };
            const unsigned long dim2709KuoInit[]        =   {   1   ,   3   ,   1   ,   9   ,   7   ,   11  ,   73  ,   215 ,   289 ,   933 ,   1369    ,   2911    ,   3251    ,   727 ,   19237   ,0 };
            const unsigned long dim2710KuoInit[]        =   {   1   ,   1   ,   1   ,   15  ,   29  ,   19  ,   77  ,   59  ,   213 ,   1005    ,   2009    ,   1899    ,   5987    ,   7429    ,   5607    ,0 };
            const unsigned long dim2711KuoInit[]        =   {   1   ,   1   ,   1   ,   13  ,   7   ,   53  ,   79  ,   151 ,   443 ,   911 ,   1777    ,   1223    ,   663 ,   2099    ,   17297   ,0 };
            const unsigned long dim2712KuoInit[]        =   {   1   ,   1   ,   1   ,   9   ,   11  ,   7   ,   81  ,   175 ,   265 ,   879 ,   1867    ,   539 ,   2353    ,   5667    ,   4973    ,0 };
            const unsigned long dim2713KuoInit[]        =   {   1   ,   1   ,   1   ,   9   ,   23  ,   63  ,   35  ,   57  ,   139 ,   653 ,   1837    ,   681 ,   6993    ,   643 ,   21771   ,0 };
            const unsigned long dim2714KuoInit[]        =   {   1   ,   3   ,   7   ,   9   ,   11  ,   13  ,   123 ,   199 ,   405 ,   373 ,   1817    ,   3331    ,   5585    ,   11721   ,   9355    ,0 };
            const unsigned long dim2715KuoInit[]        =   {   1   ,   1   ,   7   ,   7   ,   27  ,   29  ,   79  ,   121 ,   399 ,   85  ,   531 ,   507 ,   4361    ,   6167    ,   29317   ,0 };
            const unsigned long dim2716KuoInit[]        =   {   1   ,   1   ,   7   ,   11  ,   13  ,   19  ,   75  ,   245 ,   197 ,   731 ,   233 ,   487 ,   2245    ,   11413   ,   31865   ,0 };
            const unsigned long dim2717KuoInit[]        =   {   1   ,   1   ,   7   ,   5   ,   27  ,   43  ,   109 ,   135 ,   91  ,   475 ,   209 ,   865 ,   2239    ,   8065    ,   23295   ,0 };
            const unsigned long dim2718KuoInit[]        =   {   1   ,   1   ,   1   ,   9   ,   13  ,   27  ,   101 ,   121 ,   23  ,   319 ,   873 ,   251 ,   299 ,   7145    ,   26225   ,0 };
            const unsigned long dim2719KuoInit[]        =   {   1   ,   1   ,   7   ,   1   ,   31  ,   51  ,   85  ,   91  ,   299 ,   545 ,   1557    ,   2447    ,   4055    ,   3911    ,   12805   ,0 };
            const unsigned long dim2720KuoInit[]        =   {   1   ,   3   ,   1   ,   9   ,   31  ,   63  ,   123 ,   191 ,   241 ,   7   ,   347 ,   23  ,   5135    ,   7353    ,   30865   ,0 };
            const unsigned long dim2721KuoInit[]        =   {   1   ,   1   ,   5   ,   1   ,   3   ,   13  ,   81  ,   249 ,   65  ,   853 ,   893 ,   219 ,   6555    ,   4941    ,   21669   ,0 };
            const unsigned long dim2722KuoInit[]        =   {   1   ,   3   ,   5   ,   13  ,   13  ,   27  ,   13  ,   203 ,   205 ,   459 ,   883 ,   1723    ,   2721    ,   9173    ,   9643    ,0 };
            const unsigned long dim2723KuoInit[]        =   {   1   ,   1   ,   5   ,   5   ,   11  ,   59  ,   109 ,   25  ,   107 ,   1007    ,   357 ,   2877    ,   5075    ,   3251    ,   29541   ,0 };
            const unsigned long dim2724KuoInit[]        =   {   1   ,   1   ,   1   ,   15  ,   19  ,   37  ,   99  ,   183 ,   267 ,   571 ,   49  ,   975 ,   2441    ,   4459    ,   15481   ,0 };
            const unsigned long dim2725KuoInit[]        =   {   1   ,   3   ,   1   ,   1   ,   21  ,   7   ,   89  ,   233 ,   209 ,   151 ,   1919    ,   31  ,   4363    ,   9781    ,   4683    ,0 };
            const unsigned long dim2726KuoInit[]        =   {   1   ,   1   ,   3   ,   9   ,   15  ,   35  ,   101 ,   29  ,   27  ,   737 ,   1577    ,   2611    ,   3727    ,   2245    ,   31889   ,0 };
            const unsigned long dim2727KuoInit[]        =   {   1   ,   3   ,   5   ,   9   ,   19  ,   19  ,   125 ,   229 ,   313 ,   239 ,   67  ,   1373    ,   6459    ,   1339    ,   19909   ,0 };
            const unsigned long dim2728KuoInit[]        =   {   1   ,   3   ,   3   ,   13  ,   19  ,   35  ,   17  ,   31  ,   183 ,   391 ,   1515    ,   2357    ,   2673    ,   10055   ,   943 ,0 };
            const unsigned long dim2729KuoInit[]        =   {   1   ,   1   ,   3   ,   11  ,   31  ,   17  ,   37  ,   239 ,   79  ,   869 ,   629 ,   2755    ,   7431    ,   127 ,   19155   ,0 };
            const unsigned long dim2730KuoInit[]        =   {   1   ,   3   ,   5   ,   3   ,   9   ,   13  ,   91  ,   27  ,   163 ,   455 ,   2007    ,   2479    ,   2387    ,   1557    ,   437 ,0 };
            const unsigned long dim2731KuoInit[]        =   {   1   ,   1   ,   1   ,   13  ,   15  ,   63  ,   11  ,   105 ,   51  ,   673 ,   997 ,   2487    ,   35  ,   8655    ,   21481   ,0 };
            const unsigned long dim2732KuoInit[]        =   {   1   ,   3   ,   5   ,   11  ,   1   ,   3   ,   27  ,   117 ,   195 ,   233 ,   269 ,   2163    ,   2083    ,   8907    ,   28005   ,0 };
            const unsigned long dim2733KuoInit[]        =   {   1   ,   1   ,   1   ,   9   ,   9   ,   17  ,   63  ,   211 ,   87  ,   107 ,   1183    ,   2705    ,   6811    ,   9959    ,   13937   ,0 };
            const unsigned long dim2734KuoInit[]        =   {   1   ,   1   ,   1   ,   1   ,   1   ,   3   ,   57  ,   231 ,   209 ,   949 ,   895 ,   2773    ,   237 ,   8943    ,   22453   ,0 };
            const unsigned long dim2735KuoInit[]        =   {   1   ,   3   ,   7   ,   11  ,   7   ,   53  ,   1   ,   247 ,   279 ,   411 ,   1247    ,   2121    ,   3127    ,   16279   ,   277 ,0 };
            const unsigned long dim2736KuoInit[]        =   {   1   ,   3   ,   5   ,   5   ,   15  ,   35  ,   15  ,   15  ,   485 ,   763 ,   1631    ,   1717    ,   191 ,   6609    ,   32383   ,0 };
            const unsigned long dim2737KuoInit[]        =   {   1   ,   1   ,   5   ,   11  ,   27  ,   5   ,   73  ,   143 ,   127 ,   477 ,   535 ,   3815    ,   3135    ,   9865    ,   5375    ,0 };
            const unsigned long dim2738KuoInit[]        =   {   1   ,   1   ,   7   ,   9   ,   21  ,   31  ,   83  ,   63  ,   313 ,   775 ,   875 ,   4053    ,   905 ,   14255   ,   31537   ,0 };
            const unsigned long dim2739KuoInit[]        =   {   1   ,   1   ,   1   ,   3   ,   7   ,   3   ,   107 ,   231 ,   333 ,   935 ,   1625    ,   2527    ,   2293    ,   10407   ,   4795    ,0 };
            const unsigned long dim2740KuoInit[]        =   {   1   ,   3   ,   5   ,   7   ,   25  ,   19  ,   123 ,   193 ,   353 ,   353 ,   1403    ,   3007    ,   677 ,   15333   ,   1263    ,0 };
            const unsigned long dim2741KuoInit[]        =   {   1   ,   1   ,   3   ,   9   ,   1   ,   39  ,   79  ,   251 ,   59  ,   973 ,   1707    ,   2153    ,   7445    ,   9083    ,   6675    ,0 };
            const unsigned long dim2742KuoInit[]        =   {   1   ,   1   ,   7   ,   3   ,   7   ,   55  ,   21  ,   101 ,   203 ,   735 ,   1789    ,   3119    ,   6633    ,   3643    ,   1947    ,0 };
            const unsigned long dim2743KuoInit[]        =   {   1   ,   1   ,   5   ,   9   ,   13  ,   49  ,   103 ,   141 ,   181 ,   893 ,   385 ,   1541    ,   5887    ,   10687   ,   16183   ,0 };
            const unsigned long dim2744KuoInit[]        =   {   1   ,   3   ,   3   ,   7   ,   17  ,   59  ,   35  ,   119 ,   467 ,   77  ,   1591    ,   469 ,   4991    ,   8113    ,   20231   ,0 };
            const unsigned long dim2745KuoInit[]        =   {   1   ,   3   ,   1   ,   15  ,   19  ,   41  ,   5   ,   71  ,   247 ,   955 ,   565 ,   1931    ,   2895    ,   13189   ,   28461   ,0 };
            const unsigned long dim2746KuoInit[]        =   {   1   ,   3   ,   5   ,   9   ,   19  ,   21  ,   123 ,   39  ,   175 ,   35  ,   673 ,   107 ,   3157    ,   1495    ,   16341   ,0 };
            const unsigned long dim2747KuoInit[]        =   {   1   ,   1   ,   3   ,   9   ,   31  ,   25  ,   9   ,   17  ,   463 ,   585 ,   1701    ,   3325    ,   6071    ,   2719    ,   4479    ,0 };
            const unsigned long dim2748KuoInit[]        =   {   1   ,   3   ,   3   ,   15  ,   25  ,   51  ,   63  ,   241 ,   289 ,   787 ,   1109    ,   453 ,   5119    ,   139 ,   377 ,0 };
            const unsigned long dim2749KuoInit[]        =   {   1   ,   3   ,   5   ,   7   ,   29  ,   3   ,   53  ,   119 ,   219 ,   367 ,   1785    ,   777 ,   5255    ,   10013   ,   27979   ,0 };
            const unsigned long dim2750KuoInit[]        =   {   1   ,   1   ,   1   ,   3   ,   1   ,   45  ,   3   ,   239 ,   429 ,   595 ,   1591    ,   3111    ,   2373    ,   4121    ,   12671   ,0 };
            const unsigned long dim2751KuoInit[]        =   {   1   ,   3   ,   3   ,   9   ,   5   ,   49  ,   11  ,   145 ,   167 ,   19  ,   155 ,   2017    ,   3593    ,   15275   ,   22319   ,0 };
            const unsigned long dim2752KuoInit[]        =   {   1   ,   3   ,   1   ,   3   ,   11  ,   15  ,   17  ,   241 ,   293 ,   345 ,   1663    ,   301 ,   627 ,   12161   ,   19939   ,0 };
            const unsigned long dim2753KuoInit[]        =   {   1   ,   3   ,   5   ,   1   ,   3   ,   19  ,   41  ,   75  ,   121 ,   611 ,   1741    ,   4055    ,   5567    ,   773 ,   767 ,0 };
            const unsigned long dim2754KuoInit[]        =   {   1   ,   1   ,   1   ,   3   ,   5   ,   49  ,   127 ,   49  ,   93  ,   819 ,   1391    ,   2173    ,   4021    ,   11953   ,   28465   ,0 };
            const unsigned long dim2755KuoInit[]        =   {   1   ,   1   ,   3   ,   5   ,   9   ,   59  ,   63  ,   85  ,   249 ,   239 ,   521 ,   3333    ,   6301    ,   15283   ,   10645   ,0 };
            const unsigned long dim2756KuoInit[]        =   {   1   ,   1   ,   1   ,   7   ,   9   ,   53  ,   107 ,   201 ,   151 ,   79  ,   1561    ,   3497    ,   4473    ,   4369    ,   6231    ,0 };
            const unsigned long dim2757KuoInit[]        =   {   1   ,   1   ,   1   ,   9   ,   11  ,   61  ,   57  ,   69  ,   215 ,   637 ,   1087    ,   2827    ,   2495    ,   7471    ,   19127   ,0 };
            const unsigned long dim2758KuoInit[]        =   {   1   ,   1   ,   1   ,   13  ,   3   ,   25  ,   87  ,   131 ,   171 ,   577 ,   957 ,   989 ,   3391    ,   3803    ,   27229   ,0 };
            const unsigned long dim2759KuoInit[]        =   {   1   ,   1   ,   5   ,   5   ,   1   ,   25  ,   83  ,   71  ,   161 ,   477 ,   573 ,   3709    ,   4393    ,   4607    ,   32373   ,0 };
            const unsigned long dim2760KuoInit[]        =   {   1   ,   1   ,   7   ,   5   ,   13  ,   21  ,   19  ,   71  ,   207 ,   433 ,   849 ,   2685    ,   4455    ,   463 ,   25163   ,0 };
            const unsigned long dim2761KuoInit[]        =   {   1   ,   1   ,   7   ,   3   ,   5   ,   19  ,   35  ,   41  ,   487 ,   3   ,   865 ,   25  ,   7925    ,   9015    ,   29397   ,0 };
            const unsigned long dim2762KuoInit[]        =   {   1   ,   1   ,   7   ,   3   ,   7   ,   41  ,   83  ,   27  ,   79  ,   501 ,   267 ,   2015    ,   8109    ,   8877    ,   4029    ,0 };
            const unsigned long dim2763KuoInit[]        =   {   1   ,   3   ,   5   ,   1   ,   1   ,   31  ,   71  ,   233 ,   167 ,   185 ,   33  ,   1249    ,   3755    ,   2805    ,   29559   ,0 };
            const unsigned long dim2764KuoInit[]        =   {   1   ,   1   ,   1   ,   7   ,   11  ,   21  ,   111 ,   105 ,   469 ,   527 ,   711 ,   3551    ,   521 ,   10593   ,   22701   ,0 };
            const unsigned long dim2765KuoInit[]        =   {   1   ,   1   ,   1   ,   1   ,   29  ,   1   ,   79  ,   109 ,   279 ,   643 ,   1263    ,   2405    ,   3717    ,   11395   ,   4325    ,0 };
            const unsigned long dim2766KuoInit[]        =   {   1   ,   1   ,   1   ,   1   ,   29  ,   47  ,   59  ,   1   ,   195 ,   399 ,   1963    ,   265 ,   317 ,   15117   ,   9935    ,0 };
            const unsigned long dim2767KuoInit[]        =   {   1   ,   1   ,   7   ,   5   ,   21  ,   43  ,   93  ,   113 ,   273 ,   67  ,   281 ,   477 ,   6309    ,   3649    ,   20147   ,0 };
            const unsigned long dim2768KuoInit[]        =   {   1   ,   1   ,   3   ,   15  ,   29  ,   61  ,   69  ,   151 ,   185 ,   809 ,   993 ,   2383    ,   7263    ,   901 ,   18607   ,0 };
            const unsigned long dim2769KuoInit[]        =   {   1   ,   1   ,   3   ,   9   ,   27  ,   33  ,   25  ,   103 ,   387 ,   131 ,   1555    ,   301 ,   1471    ,   11331   ,   15091   ,0 };
            const unsigned long dim2770KuoInit[]        =   {   1   ,   1   ,   7   ,   7   ,   19  ,   55  ,   115 ,   121 ,   99  ,   371 ,   1563    ,   895 ,   6207    ,   93  ,   8549    ,0 };
            const unsigned long dim2771KuoInit[]        =   {   1   ,   1   ,   1   ,   9   ,   19  ,   23  ,   13  ,   181 ,   281 ,   479 ,   1   ,   1515    ,   3853    ,   8601    ,   19241   ,0 };
            const unsigned long dim2772KuoInit[]        =   {   1   ,   3   ,   1   ,   15  ,   29  ,   63  ,   113 ,   253 ,   105 ,   421 ,   1679    ,   2113    ,   4671    ,   7923    ,   7671    ,0 };
            const unsigned long dim2773KuoInit[]        =   {   1   ,   3   ,   1   ,   3   ,   17  ,   9   ,   15  ,   241 ,   339 ,   619 ,   417 ,   3257    ,   5917    ,   9981    ,   19033   ,0 };
            const unsigned long dim2774KuoInit[]        =   {   1   ,   3   ,   3   ,   13  ,   21  ,   11  ,   97  ,   201 ,   257 ,   307 ,   133 ,   1793    ,   193 ,   8609    ,   27771   ,0 };
            const unsigned long dim2775KuoInit[]        =   {   1   ,   1   ,   7   ,   3   ,   25  ,   55  ,   51  ,   97  ,   453 ,   37  ,   639 ,   3451    ,   6587    ,   3835    ,   13229   ,0 };
            const unsigned long dim2776KuoInit[]        =   {   1   ,   3   ,   1   ,   1   ,   3   ,   21  ,   123 ,   81  ,   447 ,   963 ,   449 ,   1329    ,   4887    ,   1141    ,   3213    ,0 };
            const unsigned long dim2777KuoInit[]        =   {   1   ,   1   ,   3   ,   1   ,   5   ,   33  ,   107 ,   89  ,   67  ,   229 ,   2025    ,   1925    ,   5369    ,   14049   ,   10461   ,0 };
            const unsigned long dim2778KuoInit[]        =   {   1   ,   1   ,   7   ,   3   ,   23  ,   17  ,   75  ,   79  ,   51  ,   25  ,   1541    ,   3475    ,   6129    ,   13391   ,   31091   ,0 };
            const unsigned long dim2779KuoInit[]        =   {   1   ,   3   ,   3   ,   13  ,   17  ,   27  ,   5   ,   57  ,   147 ,   549 ,   721 ,   2477    ,   893 ,   6413    ,   24053   ,0 };
            const unsigned long dim2780KuoInit[]        =   {   1   ,   3   ,   1   ,   15  ,   11  ,   29  ,   119 ,   149 ,   185 ,   713 ,   1603    ,   3671    ,   3409    ,   2689    ,   7933    ,0 };
            const unsigned long dim2781KuoInit[]        =   {   1   ,   1   ,   7   ,   15  ,   9   ,   33  ,   97  ,   85  ,   185 ,   225 ,   1747    ,   1677    ,   4133    ,   11421   ,   26591   ,0 };
            const unsigned long dim2782KuoInit[]        =   {   1   ,   1   ,   5   ,   15  ,   29  ,   43  ,   75  ,   251 ,   157 ,   861 ,   517 ,   3341    ,   4447    ,   14803   ,   18465   ,0 };
            const unsigned long dim2783KuoInit[]        =   {   1   ,   3   ,   5   ,   15  ,   11  ,   37  ,   57  ,   189 ,   423 ,   881 ,   1199    ,   1263    ,   2691    ,   6969    ,   21803   ,0 };
            const unsigned long dim2784KuoInit[]        =   {   1   ,   1   ,   3   ,   3   ,   25  ,   47  ,   15  ,   107 ,   119 ,   885 ,   1149    ,   673 ,   4459    ,   9997    ,   4519    ,0 };
            const unsigned long dim2785KuoInit[]        =   {   1   ,   3   ,   7   ,   13  ,   23  ,   7   ,   37  ,   161 ,   17  ,   499 ,   1675    ,   2309    ,   4159    ,   13771   ,   17741   ,0 };
            const unsigned long dim2786KuoInit[]        =   {   1   ,   3   ,   1   ,   13  ,   7   ,   51  ,   111 ,   163 ,   313 ,   835 ,   1903    ,   2201    ,   321 ,   14655   ,   16845   ,0 };
            const unsigned long dim2787KuoInit[]        =   {   1   ,   1   ,   3   ,   3   ,   7   ,   41  ,   23  ,   87  ,   101 ,   371 ,   257 ,   1679    ,   7449    ,   12967   ,   31  ,0 };
            const unsigned long dim2788KuoInit[]        =   {   1   ,   3   ,   3   ,   7   ,   29  ,   37  ,   57  ,   163 ,   507 ,   1   ,   691 ,   2933    ,   4687    ,   10621   ,   32351   ,0 };
            const unsigned long dim2789KuoInit[]        =   {   1   ,   3   ,   7   ,   3   ,   31  ,   63  ,   73  ,   67  ,   503 ,   289 ,   1533    ,   1203    ,   7931    ,   11399   ,   23463   ,0 };
            const unsigned long dim2790KuoInit[]        =   {   1   ,   1   ,   5   ,   7   ,   23  ,   39  ,   19  ,   73  ,   467 ,   65  ,   791 ,   2017    ,   1763    ,   11921   ,   31643   ,0 };
            const unsigned long dim2791KuoInit[]        =   {   1   ,   3   ,   3   ,   1   ,   19  ,   21  ,   53  ,   61  ,   401 ,   429 ,   1095    ,   297 ,   7931    ,   13085   ,   2013    ,0 };
            const unsigned long dim2792KuoInit[]        =   {   1   ,   1   ,   5   ,   11  ,   27  ,   41  ,   23  ,   225 ,   299 ,   943 ,   823 ,   3507    ,   4333    ,   817 ,   26137   ,0 };
            const unsigned long dim2793KuoInit[]        =   {   1   ,   1   ,   1   ,   7   ,   27  ,   29  ,   31  ,   251 ,   499 ,   151 ,   295 ,   1731    ,   7661    ,   11825   ,   29587   ,0 };
            const unsigned long dim2794KuoInit[]        =   {   1   ,   1   ,   3   ,   7   ,   17  ,   43  ,   17  ,   5   ,   25  ,   813 ,   1875    ,   2941    ,   1407    ,   11669   ,   16753   ,0 };
            const unsigned long dim2795KuoInit[]        =   {   1   ,   3   ,   1   ,   1   ,   17  ,   59  ,   93  ,   173 ,   335 ,   933 ,   369 ,   3421    ,   2769    ,   16361   ,   11891   ,0 };
            const unsigned long dim2796KuoInit[]        =   {   1   ,   1   ,   1   ,   3   ,   27  ,   5   ,   63  ,   7   ,   415 ,   533 ,   179 ,   851 ,   1623    ,   16375   ,   1271    ,0 };
            const unsigned long dim2797KuoInit[]        =   {   1   ,   1   ,   5   ,   1   ,   17  ,   45  ,   125 ,   93  ,   43  ,   311 ,   153 ,   1225    ,   4891    ,   5621    ,   14355   ,0 };
            const unsigned long dim2798KuoInit[]        =   {   1   ,   3   ,   7   ,   5   ,   13  ,   63  ,   103 ,   205 ,   119 ,   953 ,   1087    ,   703 ,   3729    ,   10249   ,   19085   ,0 };
            const unsigned long dim2799KuoInit[]        =   {   1   ,   3   ,   5   ,   3   ,   11  ,   51  ,   17  ,   97  ,   323 ,   657 ,   269 ,   467 ,   2747    ,   13603   ,   22803   ,0 };
            const unsigned long dim2800KuoInit[]        =   {   1   ,   3   ,   3   ,   15  ,   3   ,   5   ,   35  ,   23  ,   433 ,   259 ,   1857    ,   2313    ,   7709    ,   2297    ,   31109   ,0 };
            const unsigned long dim2801KuoInit[]        =   {   1   ,   3   ,   1   ,   15  ,   19  ,   11  ,   51  ,   251 ,   167 ,   741 ,   1705    ,   3763    ,   967 ,   2593    ,   28229   ,0 };
            const unsigned long dim2802KuoInit[]        =   {   1   ,   3   ,   3   ,   13  ,   29  ,   47  ,   47  ,   183 ,   305 ,   223 ,   381 ,   1753    ,   2285    ,   9113    ,   21685   ,0 };
            const unsigned long dim2803KuoInit[]        =   {   1   ,   3   ,   7   ,   13  ,   31  ,   17  ,   105 ,   25  ,   49  ,   679 ,   1501    ,   3887    ,   5007    ,   4275    ,   31161   ,0 };
            const unsigned long dim2804KuoInit[]        =   {   1   ,   3   ,   7   ,   5   ,   27  ,   59  ,   17  ,   183 ,   477 ,   191 ,   1111    ,   383 ,   5519    ,   7435    ,   20341   ,0 };
            const unsigned long dim2805KuoInit[]        =   {   1   ,   1   ,   7   ,   13  ,   25  ,   15  ,   67  ,   57  ,   5   ,   955 ,   731 ,   3785    ,   2055    ,   13017   ,   27843   ,0 };
            const unsigned long dim2806KuoInit[]        =   {   1   ,   1   ,   5   ,   1   ,   27  ,   45  ,   67  ,   93  ,   311 ,   635 ,   1893    ,   1753    ,   7307    ,   2765    ,   6263    ,0 };
            const unsigned long dim2807KuoInit[]        =   {   1   ,   1   ,   3   ,   1   ,   13  ,   53  ,   109 ,   5   ,   3   ,   705 ,   1745    ,   2233    ,   3425    ,   6125    ,   28839   ,0 };
            const unsigned long dim2808KuoInit[]        =   {   1   ,   3   ,   5   ,   5   ,   5   ,   39  ,   5   ,   251 ,   431 ,   85  ,   599 ,   3459    ,   3211    ,   15611   ,   28739   ,0 };
            const unsigned long dim2809KuoInit[]        =   {   1   ,   1   ,   7   ,   7   ,   25  ,   5   ,   25  ,   31  ,   283 ,   909 ,   399 ,   1335    ,   6973    ,   15791   ,   28351   ,0 };
            const unsigned long dim2810KuoInit[]        =   {   1   ,   1   ,   5   ,   5   ,   3   ,   29  ,   47  ,   227 ,   463 ,   815 ,   1089    ,   2217    ,   4911    ,   14743   ,   17843   ,0 };
            const unsigned long dim2811KuoInit[]        =   {   1   ,   1   ,   7   ,   3   ,   29  ,   39  ,   125 ,   165 ,   367 ,   461 ,   569 ,   3441    ,   1293    ,   1577    ,   10567   ,0 };
            const unsigned long dim2812KuoInit[]        =   {   1   ,   1   ,   1   ,   11  ,   5   ,   11  ,   19  ,   129 ,   401 ,   111 ,   2019    ,   899 ,   2079    ,   7363    ,   32675   ,0 };
            const unsigned long dim2813KuoInit[]        =   {   1   ,   3   ,   1   ,   7   ,   13  ,   27  ,   47  ,   183 ,   465 ,   421 ,   665 ,   1279    ,   5629    ,   7559    ,   21265   ,0 };
            const unsigned long dim2814KuoInit[]        =   {   1   ,   1   ,   7   ,   1   ,   11  ,   19  ,   75  ,   1   ,   137 ,   893 ,   1709    ,   385 ,   2221    ,   15185   ,   10923   ,0 };
            const unsigned long dim2815KuoInit[]        =   {   1   ,   3   ,   5   ,   5   ,   9   ,   1   ,   115 ,   203 ,   133 ,   259 ,   915 ,   2477    ,   3615    ,   1749    ,   7643    ,0 };
            const unsigned long dim2816KuoInit[]        =   {   1   ,   3   ,   3   ,   5   ,   9   ,   57  ,   21  ,   113 ,   453 ,   739 ,   1857    ,   685 ,   5727    ,   10625   ,   23469   ,0 };
            const unsigned long dim2817KuoInit[]        =   {   1   ,   1   ,   5   ,   5   ,   13  ,   45  ,   37  ,   25  ,   495 ,   723 ,   1805    ,   1631    ,   4889    ,   12885   ,   16895   ,0 };
            const unsigned long dim2818KuoInit[]        =   {   1   ,   1   ,   5   ,   7   ,   17  ,   59  ,   111 ,   29  ,   143 ,   863 ,   1757    ,   1621    ,   3995    ,   7315    ,   30271   ,0 };
            const unsigned long dim2819KuoInit[]        =   {   1   ,   1   ,   1   ,   1   ,   5   ,   31  ,   81  ,   161 ,   275 ,   623 ,   2017    ,   967 ,   2395    ,   14923   ,   23931   ,0 };
            const unsigned long dim2820KuoInit[]        =   {   1   ,   3   ,   5   ,   11  ,   31  ,   33  ,   95  ,   135 ,   477 ,   221 ,   1603    ,   231 ,   5957    ,   12461   ,   2623    ,0 };
            const unsigned long dim2821KuoInit[]        =   {   1   ,   3   ,   5   ,   11  ,   7   ,   3   ,   53  ,   43  ,   351 ,   557 ,   1981    ,   2127    ,   909 ,   8943    ,   21035   ,0 };
            const unsigned long dim2822KuoInit[]        =   {   1   ,   3   ,   5   ,   9   ,   31  ,   11  ,   13  ,   231 ,   145 ,   321 ,   567 ,   1919    ,   7101    ,   12089   ,   8331    ,0 };
            const unsigned long dim2823KuoInit[]        =   {   1   ,   1   ,   7   ,   1   ,   13  ,   27  ,   101 ,   229 ,   89  ,   395 ,   349 ,   1785    ,   6441    ,   8241    ,   30735   ,0 };
            const unsigned long dim2824KuoInit[]        =   {   1   ,   1   ,   7   ,   9   ,   21  ,   57  ,   123 ,   77  ,   251 ,   347 ,   1367    ,   2205    ,   4017    ,   16057   ,   9247    ,0 };
            const unsigned long dim2825KuoInit[]        =   {   1   ,   3   ,   3   ,   11  ,   5   ,   1   ,   47  ,   57  ,   69  ,   381 ,   1613    ,   1525    ,   5715    ,   3677    ,   13015   ,0 };
            const unsigned long dim2826KuoInit[]        =   {   1   ,   3   ,   7   ,   9   ,   31  ,   33  ,   15  ,   255 ,   149 ,   1021    ,   1149    ,   3609    ,   1739    ,   11867   ,   3093    ,0 };
            const unsigned long dim2827KuoInit[]        =   {   1   ,   1   ,   7   ,   15  ,   27  ,   25  ,   57  ,   5   ,   137 ,   1005    ,   343 ,   2567    ,   1025    ,   8615    ,   6357    ,0 };
            const unsigned long dim2828KuoInit[]        =   {   1   ,   3   ,   5   ,   1   ,   29  ,   57  ,   85  ,   65  ,   85  ,   789 ,   1179    ,   2523    ,   7589    ,   14657   ,   23637   ,0 };
            const unsigned long dim2829KuoInit[]        =   {   1   ,   3   ,   1   ,   11  ,   3   ,   3   ,   107 ,   37  ,   125 ,   423 ,   225 ,   2991    ,   4515    ,   5057    ,   31181   ,0 };
            const unsigned long dim2830KuoInit[]        =   {   1   ,   1   ,   5   ,   9   ,   31  ,   5   ,   111 ,   241 ,   301 ,   37  ,   943 ,   555 ,   7051    ,   11627   ,   7125    ,0 };
            const unsigned long dim2831KuoInit[]        =   {   1   ,   3   ,   3   ,   15  ,   17  ,   31  ,   119 ,   185 ,   187 ,   621 ,   395 ,   3595    ,   1055    ,   7993    ,   28441   ,0 };
            const unsigned long dim2832KuoInit[]        =   {   1   ,   1   ,   5   ,   1   ,   31  ,   25  ,   113 ,   223 ,   493 ,   919 ,   1455    ,   31  ,   4743    ,   14435   ,   31627   ,0 };
            const unsigned long dim2833KuoInit[]        =   {   1   ,   1   ,   1   ,   1   ,   27  ,   59  ,   45  ,   75  ,   343 ,   701 ,   1737    ,   161 ,   6919    ,   13537   ,   26835   ,0 };
            const unsigned long dim2834KuoInit[]        =   {   1   ,   3   ,   1   ,   9   ,   31  ,   45  ,   89  ,   215 ,   137 ,   237 ,   1135    ,   3239    ,   3345    ,   12063   ,   14599   ,0 };
            const unsigned long dim2835KuoInit[]        =   {   1   ,   3   ,   5   ,   15  ,   5   ,   51  ,   85  ,   73  ,   353 ,   943 ,   1981    ,   127 ,   2605    ,   5673    ,   10037   ,0 };
            const unsigned long dim2836KuoInit[]        =   {   1   ,   1   ,   7   ,   1   ,   3   ,   49  ,   109 ,   105 ,   483 ,   499 ,   1581    ,   793 ,   5857    ,   5509    ,   4651    ,0 };
            const unsigned long dim2837KuoInit[]        =   {   1   ,   3   ,   5   ,   9   ,   23  ,   5   ,   109 ,   83  ,   7   ,   155 ,   187 ,   2205    ,   4199    ,   5487    ,   31925   ,0 };
            const unsigned long dim2838KuoInit[]        =   {   1   ,   1   ,   7   ,   5   ,   13  ,   11  ,   75  ,   53  ,   125 ,   139 ,   1241    ,   203 ,   7361    ,   13783   ,   4317    ,0 };
            const unsigned long dim2839KuoInit[]        =   {   1   ,   3   ,   7   ,   5   ,   25  ,   57  ,   105 ,   197 ,   207 ,   261 ,   1251    ,   4057    ,   3877    ,   13613   ,   7829    ,0 };
            const unsigned long dim2840KuoInit[]        =   {   1   ,   1   ,   5   ,   13  ,   31  ,   21  ,   89  ,   73  ,   163 ,   941 ,   1683    ,   2239    ,   1893    ,   16067   ,   609 ,0 };
            const unsigned long dim2841KuoInit[]        =   {   1   ,   1   ,   1   ,   3   ,   5   ,   9   ,   41  ,   149 ,   43  ,   333 ,   1459    ,   277 ,   5687    ,   14201   ,   26879   ,0 };
            const unsigned long dim2842KuoInit[]        =   {   1   ,   1   ,   3   ,   15  ,   21  ,   37  ,   67  ,   227 ,   443 ,   767 ,   1109    ,   2151    ,   7875    ,   15519   ,   13741   ,0 };
            const unsigned long dim2843KuoInit[]        =   {   1   ,   3   ,   1   ,   5   ,   31  ,   15  ,   107 ,   43  ,   343 ,   27  ,   1321    ,   3009    ,   5751    ,   2931    ,   9733    ,0 };
            const unsigned long dim2844KuoInit[]        =   {   1   ,   1   ,   1   ,   11  ,   5   ,   27  ,   49  ,   33  ,   439 ,   555 ,   97  ,   59  ,   4043    ,   12063   ,   27227   ,0 };
            const unsigned long dim2845KuoInit[]        =   {   1   ,   3   ,   7   ,   7   ,   27  ,   9   ,   55  ,   69  ,   485 ,   75  ,   1255    ,   101 ,   7193    ,   9427    ,   31125   ,0 };
            const unsigned long dim2846KuoInit[]        =   {   1   ,   3   ,   7   ,   3   ,   15  ,   23  ,   51  ,   145 ,   501 ,   875 ,   1611    ,   1321    ,   6809    ,   5051    ,   17135   ,0 };
            const unsigned long dim2847KuoInit[]        =   {   1   ,   3   ,   3   ,   13  ,   5   ,   45  ,   35  ,   115 ,   3   ,   639 ,   1705    ,   207 ,   4235    ,   3329    ,   14481   ,0 };
            const unsigned long dim2848KuoInit[]        =   {   1   ,   3   ,   3   ,   1   ,   5   ,   15  ,   17  ,   115 ,   175 ,   583 ,   1649    ,   2195    ,   3827    ,   16227   ,   21401   ,0 };
            const unsigned long dim2849KuoInit[]        =   {   1   ,   3   ,   3   ,   7   ,   1   ,   3   ,   107 ,   167 ,   195 ,   953 ,   743 ,   1375    ,   701 ,   1155    ,   1027    ,0 };
            const unsigned long dim2850KuoInit[]        =   {   1   ,   1   ,   7   ,   1   ,   31  ,   31  ,   67  ,   133 ,   219 ,   307 ,   1031    ,   3321    ,   4855    ,   11285   ,   16987   ,0 };
            const unsigned long dim2851KuoInit[]        =   {   1   ,   1   ,   7   ,   1   ,   25  ,   9   ,   41  ,   105 ,   89  ,   289 ,   1851    ,   2245    ,   1647    ,   2925    ,   8637    ,0 };
            const unsigned long dim2852KuoInit[]        =   {   1   ,   3   ,   7   ,   15  ,   29  ,   29  ,   107 ,   21  ,   101 ,   419 ,   1207    ,   3301    ,   1217    ,   14157   ,   17373   ,0 };
            const unsigned long dim2853KuoInit[]        =   {   1   ,   1   ,   5   ,   11  ,   9   ,   41  ,   49  ,   165 ,   353 ,   215 ,   883 ,   2031    ,   7363    ,   13221   ,   8329    ,0 };
            const unsigned long dim2854KuoInit[]        =   {   1   ,   1   ,   5   ,   15  ,   17  ,   7   ,   67  ,   217 ,   87  ,   217 ,   2013    ,   3809    ,   4983    ,   5341    ,   30353   ,0 };
            const unsigned long dim2855KuoInit[]        =   {   1   ,   1   ,   7   ,   9   ,   1   ,   37  ,   97  ,   109 ,   485 ,   777 ,   171 ,   1391    ,   781 ,   5019    ,   19883   ,0 };
            const unsigned long dim2856KuoInit[]        =   {   1   ,   3   ,   3   ,   13  ,   7   ,   17  ,   27  ,   69  ,   265 ,   663 ,   569 ,   2909    ,   7409    ,   14813   ,   5849    ,0 };
            const unsigned long dim2857KuoInit[]        =   {   1   ,   3   ,   5   ,   13  ,   11  ,   29  ,   27  ,   105 ,   215 ,   59  ,   1251    ,   2003    ,   2201    ,   7221    ,   7549    ,0 };
            const unsigned long dim2858KuoInit[]        =   {   1   ,   1   ,   1   ,   3   ,   15  ,   7   ,   79  ,   237 ,   327 ,   839 ,   1167    ,   3211    ,   275 ,   16031   ,   25691   ,0 };
            const unsigned long dim2859KuoInit[]        =   {   1   ,   1   ,   5   ,   3   ,   3   ,   33  ,   117 ,   113 ,   361 ,   381 ,   1093    ,   1791    ,   2399    ,   1245    ,   23275   ,0 };
            const unsigned long dim2860KuoInit[]        =   {   1   ,   1   ,   1   ,   3   ,   7   ,   49  ,   77  ,   5   ,   217 ,   995 ,   1127    ,   3891    ,   81  ,   13331   ,   12479   ,0 };
            const unsigned long dim2861KuoInit[]        =   {   1   ,   1   ,   7   ,   9   ,   9   ,   57  ,   119 ,   227 ,   41  ,   337 ,   183 ,   3201    ,   3627    ,   16263   ,   28153   ,0 };
            const unsigned long dim2862KuoInit[]        =   {   1   ,   3   ,   1   ,   13  ,   11  ,   9   ,   81  ,   99  ,   247 ,   157 ,   1421    ,   1087    ,   151 ,   6529    ,   13537   ,0 };
            const unsigned long dim2863KuoInit[]        =   {   1   ,   3   ,   3   ,   5   ,   31  ,   43  ,   113 ,   143 ,   163 ,   583 ,   2031    ,   3087    ,   7565    ,   4987    ,   17131   ,0 };
            const unsigned long dim2864KuoInit[]        =   {   1   ,   1   ,   3   ,   1   ,   9   ,   1   ,   121 ,   93  ,   113 ,   849 ,   679 ,   1403    ,   6815    ,   15605   ,   29785   ,0 };
            const unsigned long dim2865KuoInit[]        =   {   1   ,   3   ,   7   ,   11  ,   27  ,   63  ,   53  ,   83  ,   397 ,   935 ,   23  ,   125 ,   5813    ,   15765   ,   2869    ,0 };
            const unsigned long dim2866KuoInit[]        =   {   1   ,   1   ,   3   ,   3   ,   9   ,   11  ,   65  ,   77  ,   387 ,   129 ,   1039    ,   1583    ,   2235    ,   3851    ,   30603   ,0 };
            const unsigned long dim2867KuoInit[]        =   {   1   ,   1   ,   7   ,   3   ,   5   ,   27  ,   45  ,   233 ,   351 ,   827 ,   215 ,   1209    ,   409 ,   2185    ,   3827    ,0 };
            const unsigned long dim2868KuoInit[]        =   {   1   ,   1   ,   3   ,   3   ,   31  ,   23  ,   35  ,   83  ,   271 ,   65  ,   553 ,   3189    ,   2435    ,   12865   ,   28271   ,0 };
            const unsigned long dim2869KuoInit[]        =   {   1   ,   1   ,   5   ,   1   ,   29  ,   51  ,   35  ,   105 ,   477 ,   1001    ,   1859    ,   2063    ,   6895    ,   1855    ,   13931   ,0 };
            const unsigned long dim2870KuoInit[]        =   {   1   ,   3   ,   3   ,   11  ,   17  ,   63  ,   125 ,   165 ,   37  ,   671 ,   2001    ,   1237    ,   7355    ,   339 ,   24771   ,0 };
            const unsigned long dim2871KuoInit[]        =   {   1   ,   3   ,   3   ,   5   ,   13  ,   5   ,   9   ,   201 ,   183 ,   39  ,   119 ,   3889    ,   949 ,   7889    ,   25951   ,0 };
            const unsigned long dim2872KuoInit[]        =   {   1   ,   1   ,   3   ,   5   ,   13  ,   63  ,   71  ,   117 ,   209 ,   955 ,   1175    ,   317 ,   4261    ,   2341    ,   29819   ,0 };
            const unsigned long dim2873KuoInit[]        =   {   1   ,   3   ,   5   ,   5   ,   9   ,   59  ,   109 ,   137 ,   353 ,   355 ,   1227    ,   3183    ,   1553    ,   15039   ,   27979   ,0 };
            const unsigned long dim2874KuoInit[]        =   {   1   ,   3   ,   1   ,   5   ,   25  ,   57  ,   87  ,   231 ,   259 ,   333 ,   975 ,   3413    ,   6035    ,   14783   ,   22309   ,0 };
            const unsigned long dim2875KuoInit[]        =   {   1   ,   3   ,   5   ,   11  ,   19  ,   37  ,   55  ,   229 ,   423 ,   761 ,   1669    ,   3023    ,   2517    ,   2961    ,   12043   ,0 };
            const unsigned long dim2876KuoInit[]        =   {   1   ,   1   ,   3   ,   13  ,   23  ,   5   ,   87  ,   193 ,   113 ,   577 ,   269 ,   983 ,   5453    ,   7827    ,   28197   ,0 };
            const unsigned long dim2877KuoInit[]        =   {   1   ,   1   ,   7   ,   13  ,   1   ,   49  ,   89  ,   107 ,   155 ,   767 ,   1547    ,   2139    ,   5951    ,   3143    ,   12735   ,0 };
            const unsigned long dim2878KuoInit[]        =   {   1   ,   3   ,   1   ,   13  ,   13  ,   57  ,   53  ,   243 ,   231 ,   131 ,   1505    ,   1255    ,   7831    ,   15615   ,   28755   ,0 };
            const unsigned long dim2879KuoInit[]        =   {   1   ,   1   ,   5   ,   11  ,   21  ,   21  ,   37  ,   65  ,   327 ,   353 ,   1625    ,   543 ,   669 ,   1139    ,   20745   ,0 };
            const unsigned long dim2880KuoInit[]        =   {   1   ,   3   ,   1   ,   1   ,   5   ,   33  ,   87  ,   211 ,   21  ,   581 ,   397 ,   2691    ,   1069    ,   10449   ,   17727   ,0 };
            const unsigned long dim2881KuoInit[]        =   {   1   ,   1   ,   7   ,   7   ,   1   ,   5   ,   115 ,   143 ,   99  ,   523 ,   51  ,   939 ,   4635    ,   11491   ,   32731   ,0 };
            const unsigned long dim2882KuoInit[]        =   {   1   ,   3   ,   3   ,   15  ,   1   ,   19  ,   119 ,   193 ,   127 ,   897 ,   1337    ,   3977    ,   5027    ,   5913    ,   9329    ,0 };
            const unsigned long dim2883KuoInit[]        =   {   1   ,   1   ,   5   ,   15  ,   15  ,   1   ,   93  ,   209 ,   427 ,   625 ,   1581    ,   2639    ,   1291    ,   11177   ,   26943   ,0 };
            const unsigned long dim2884KuoInit[]        =   {   1   ,   1   ,   3   ,   13  ,   27  ,   9   ,   47  ,   243 ,   211 ,   21  ,   1159    ,   3157    ,   6729    ,   219 ,   19671   ,0 };
            const unsigned long dim2885KuoInit[]        =   {   1   ,   3   ,   1   ,   15  ,   19  ,   27  ,   39  ,   237 ,   39  ,   389 ,   1201    ,   2237    ,   7623    ,   12041   ,   18491   ,0 };
            const unsigned long dim2886KuoInit[]        =   {   1   ,   1   ,   3   ,   5   ,   25  ,   43  ,   47  ,   101 ,   319 ,   749 ,   659 ,   2319    ,   1221    ,   3887    ,   29927   ,0 };
            const unsigned long dim2887KuoInit[]        =   {   1   ,   3   ,   3   ,   9   ,   1   ,   55  ,   57  ,   179 ,   163 ,   695 ,   611 ,   2511    ,   1261    ,   15561   ,   29535   ,0 };
            const unsigned long dim2888KuoInit[]        =   {   1   ,   1   ,   1   ,   13  ,   7   ,   31  ,   79  ,   21  ,   143 ,   993 ,   377 ,   791 ,   187 ,   6485    ,   6827    ,0 };
            const unsigned long dim2889KuoInit[]        =   {   1   ,   3   ,   5   ,   1   ,   11  ,   57  ,   39  ,   159 ,   373 ,   179 ,   1907    ,   2475    ,   1355    ,   9885    ,   18683   ,0 };
            const unsigned long dim2890KuoInit[]        =   {   1   ,   1   ,   1   ,   3   ,   7   ,   5   ,   83  ,   151 ,   267 ,   377 ,   1463    ,   365 ,   4725    ,   2319    ,   18957   ,0 };
            const unsigned long dim2891KuoInit[]        =   {   1   ,   3   ,   7   ,   3   ,   21  ,   51  ,   115 ,   119 ,   73  ,   561 ,   1513    ,   403 ,   4889    ,   2329    ,   20141   ,0 };
            const unsigned long dim2892KuoInit[]        =   {   1   ,   1   ,   1   ,   15  ,   21  ,   47  ,   31  ,   61  ,   171 ,   529 ,   185 ,   717 ,   7065    ,   4273    ,   16643   ,0 };
            const unsigned long dim2893KuoInit[]        =   {   1   ,   3   ,   5   ,   9   ,   3   ,   55  ,   41  ,   213 ,   489 ,   379 ,   1951    ,   1957    ,   7721    ,   6415    ,   27049   ,0 };
            const unsigned long dim2894KuoInit[]        =   {   1   ,   3   ,   1   ,   13  ,   19  ,   43  ,   29  ,   203 ,   139 ,   137 ,   463 ,   1447    ,   2585    ,   14381   ,   29073   ,0 };
            const unsigned long dim2895KuoInit[]        =   {   1   ,   1   ,   1   ,   1   ,   19  ,   37  ,   111 ,   119 ,   411 ,   671 ,   1135    ,   2891    ,   7929    ,   7465    ,   21699   ,0 };
            const unsigned long dim2896KuoInit[]        =   {   1   ,   1   ,   5   ,   13  ,   21  ,   31  ,   83  ,   227 ,   435 ,   61  ,   969 ,   2171    ,   789 ,   3415    ,   18293   ,0 };
            const unsigned long dim2897KuoInit[]        =   {   1   ,   3   ,   3   ,   15  ,   31  ,   33  ,   87  ,   227 ,   369 ,   341 ,   1515    ,   25  ,   2891    ,   16199   ,   22179   ,0 };
            const unsigned long dim2898KuoInit[]        =   {   1   ,   3   ,   3   ,   1   ,   25  ,   49  ,   111 ,   167 ,   59  ,   355 ,   1687    ,   1799    ,   5963    ,   6161    ,   32157   ,0 };
            const unsigned long dim2899KuoInit[]        =   {   1   ,   1   ,   1   ,   5   ,   7   ,   57  ,   95  ,   171 ,   417 ,   277 ,   47  ,   847 ,   3889    ,   8645    ,   3367    ,0 };
            const unsigned long dim2900KuoInit[]        =   {   1   ,   1   ,   7   ,   3   ,   25  ,   51  ,   99  ,   251 ,   463 ,   833 ,   491 ,   3595    ,   2883    ,   8183    ,   28937   ,0 };
            const unsigned long dim2901KuoInit[]        =   {   1   ,   1   ,   3   ,   7   ,   1   ,   5   ,   73  ,   151 ,   397 ,   521 ,   905 ,   603 ,   2315    ,   15231   ,   17955   ,0 };
            const unsigned long dim2902KuoInit[]        =   {   1   ,   1   ,   1   ,   13  ,   23  ,   59  ,   9   ,   187 ,   163 ,   727 ,   1909    ,   1135    ,   6135    ,   11563   ,   2047    ,0 };
            const unsigned long dim2903KuoInit[]        =   {   1   ,   1   ,   3   ,   3   ,   9   ,   1   ,   121 ,   207 ,   55  ,   947 ,   873 ,   419 ,   2799    ,   3285    ,   10237   ,0 };
            const unsigned long dim2904KuoInit[]        =   {   1   ,   1   ,   1   ,   3   ,   21  ,   27  ,   27  ,   241 ,   509 ,   295 ,   1509    ,   3527    ,   4269    ,   13579   ,   29125   ,0 };
            const unsigned long dim2905KuoInit[]        =   {   1   ,   3   ,   1   ,   5   ,   7   ,   21  ,   35  ,   67  ,   251 ,   883 ,   1117    ,   1311    ,   1251    ,   10401   ,   10615   ,0 };
            const unsigned long dim2906KuoInit[]        =   {   1   ,   3   ,   7   ,   3   ,   21  ,   13  ,   23  ,   75  ,   401 ,   279 ,   1557    ,   1419    ,   7067    ,   2563    ,   22285   ,0 };
            const unsigned long dim2907KuoInit[]        =   {   1   ,   3   ,   1   ,   7   ,   23  ,   15  ,   11  ,   189 ,   315 ,   353 ,   1377    ,   3551    ,   4939    ,   14529   ,   12459   ,0 };
            const unsigned long dim2908KuoInit[]        =   {   1   ,   3   ,   1   ,   3   ,   21  ,   41  ,   111 ,   67  ,   311 ,   95  ,   1919    ,   165 ,   2097    ,   12965   ,   12517   ,0 };
            const unsigned long dim2909KuoInit[]        =   {   1   ,   3   ,   3   ,   11  ,   29  ,   61  ,   33  ,   35  ,   209 ,   541 ,   149 ,   3073    ,   8187    ,   15087   ,   17617   ,0 };
            const unsigned long dim2910KuoInit[]        =   {   1   ,   1   ,   7   ,   5   ,   9   ,   53  ,   73  ,   145 ,   205 ,   157 ,   1737    ,   1737    ,   6471    ,   12327   ,   11605   ,0 };
            const unsigned long dim2911KuoInit[]        =   {   1   ,   1   ,   5   ,   5   ,   15  ,   35  ,   105 ,   99  ,   467 ,   421 ,   483 ,   3191    ,   2321    ,   8953    ,   22045   ,0 };
            const unsigned long dim2912KuoInit[]        =   {   1   ,   3   ,   3   ,   9   ,   1   ,   35  ,   65  ,   249 ,   37  ,   535 ,   1301    ,   2141    ,   1629    ,   9709    ,   995 ,0 };
            const unsigned long dim2913KuoInit[]        =   {   1   ,   3   ,   1   ,   13  ,   9   ,   27  ,   75  ,   115 ,   121 ,   731 ,   1567    ,   2341    ,   6657    ,   14959   ,   6781    ,0 };
            const unsigned long dim2914KuoInit[]        =   {   1   ,   1   ,   1   ,   5   ,   13  ,   33  ,   89  ,   215 ,   189 ,   693 ,   357 ,   3383    ,   6345    ,   14263   ,   23973   ,0 };
            const unsigned long dim2915KuoInit[]        =   {   1   ,   3   ,   1   ,   9   ,   21  ,   5   ,   59  ,   11  ,   21  ,   703 ,   1217    ,   2623    ,   6429    ,   479 ,   28659   ,0 };
            const unsigned long dim2916KuoInit[]        =   {   1   ,   3   ,   3   ,   9   ,   11  ,   15  ,   113 ,   89  ,   453 ,   233 ,   1323    ,   3859    ,   1841    ,   2753    ,   2839    ,0 };
            const unsigned long dim2917KuoInit[]        =   {   1   ,   1   ,   5   ,   15  ,   19  ,   1   ,   119 ,   83  ,   205 ,   861 ,   787 ,   1039    ,   5257    ,   8945    ,   27271   ,0 };
            const unsigned long dim2918KuoInit[]        =   {   1   ,   3   ,   1   ,   15  ,   9   ,   37  ,   7   ,   83  ,   51  ,   611 ,   1121    ,   3185    ,   1603    ,   5821    ,   14533   ,0 };
            const unsigned long dim2919KuoInit[]        =   {   1   ,   1   ,   7   ,   7   ,   27  ,   61  ,   115 ,   191 ,   357 ,   93  ,   1647    ,   3249    ,   2479    ,   2423    ,   649 ,0 };
            const unsigned long dim2920KuoInit[]        =   {   1   ,   1   ,   7   ,   11  ,   25  ,   17  ,   13  ,   213 ,   37  ,   571 ,   1179    ,   3897    ,   3817    ,   7597    ,   26239   ,0 };
            const unsigned long dim2921KuoInit[]        =   {   1   ,   1   ,   5   ,   5   ,   27  ,   57  ,   73  ,   209 ,   433 ,   779 ,   75  ,   2135    ,   8183    ,   7343    ,   23751   ,0 };
            const unsigned long dim2922KuoInit[]        =   {   1   ,   1   ,   7   ,   13  ,   27  ,   51  ,   31  ,   33  ,   195 ,   443 ,   1639    ,   1415    ,   6203    ,   4373    ,   31271   ,0 };
            const unsigned long dim2923KuoInit[]        =   {   1   ,   1   ,   7   ,   11  ,   9   ,   9   ,   75  ,   103 ,   237 ,   521 ,   1653    ,   1209    ,   5981    ,   5961    ,   15229   ,0 };
            const unsigned long dim2924KuoInit[]        =   {   1   ,   3   ,   3   ,   11  ,   15  ,   29  ,   1   ,   201 ,   171 ,   81  ,   729 ,   589 ,   441 ,   2453    ,   25209   ,0 };
            const unsigned long dim2925KuoInit[]        =   {   1   ,   3   ,   1   ,   1   ,   21  ,   11  ,   95  ,   225 ,   213 ,   981 ,   1479    ,   3259    ,   4415    ,   8393    ,   17647   ,0 };
            const unsigned long dim2926KuoInit[]        =   {   1   ,   1   ,   7   ,   1   ,   7   ,   11  ,   95  ,   135 ,   329 ,   483 ,   1711    ,   3223    ,   4841    ,   11083   ,   4033    ,0 };
            const unsigned long dim2927KuoInit[]        =   {   1   ,   1   ,   3   ,   5   ,   23  ,   37  ,   77  ,   39  ,   363 ,   555 ,   1451    ,   99  ,   6963    ,   13279   ,   24155   ,0 };
            const unsigned long dim2928KuoInit[]        =   {   1   ,   3   ,   7   ,   11  ,   11  ,   17  ,   31  ,   29  ,   357 ,   177 ,   411 ,   1439    ,   815 ,   7017    ,   25527   ,0 };
            const unsigned long dim2929KuoInit[]        =   {   1   ,   3   ,   3   ,   15  ,   25  ,   7   ,   105 ,   9   ,   301 ,   209 ,   917 ,   1369    ,   2941    ,   11599   ,   25847   ,0 };
            const unsigned long dim2930KuoInit[]        =   {   1   ,   1   ,   7   ,   7   ,   11  ,   33  ,   121 ,   9   ,   465 ,   945 ,   201 ,   3929    ,   5521    ,   4787    ,   10713   ,0 };
            const unsigned long dim2931KuoInit[]        =   {   1   ,   1   ,   7   ,   9   ,   27  ,   3   ,   77  ,   11  ,   103 ,   181 ,   1199    ,   2833    ,   6259    ,   3879    ,   13217   ,0 };
            const unsigned long dim2932KuoInit[]        =   {   1   ,   3   ,   3   ,   11  ,   11  ,   9   ,   5   ,   217 ,   471 ,   973 ,   293 ,   3693    ,   5457    ,   11359   ,   30043   ,0 };
            const unsigned long dim2933KuoInit[]        =   {   1   ,   1   ,   3   ,   15  ,   27  ,   25  ,   7   ,   65  ,   95  ,   883 ,   277 ,   3147    ,   6203    ,   12201   ,   23203   ,0 };
            const unsigned long dim2934KuoInit[]        =   {   1   ,   1   ,   3   ,   9   ,   3   ,   19  ,   51  ,   55  ,   381 ,   879 ,   845 ,   1605    ,   2433    ,   4781    ,   5075    ,0 };
            const unsigned long dim2935KuoInit[]        =   {   1   ,   1   ,   3   ,   15  ,   17  ,   49  ,   21  ,   81  ,   209 ,   285 ,   1483    ,   1897    ,   5591    ,   131 ,   30459   ,0 };
            const unsigned long dim2936KuoInit[]        =   {   1   ,   3   ,   1   ,   9   ,   17  ,   19  ,   47  ,   97  ,   325 ,   963 ,   1665    ,   3361    ,   369 ,   11023   ,   23701   ,0 };
            const unsigned long dim2937KuoInit[]        =   {   1   ,   3   ,   1   ,   13  ,   3   ,   19  ,   57  ,   7   ,   57  ,   383 ,   1945    ,   3357    ,   2553    ,   5161    ,   22363   ,0 };
            const unsigned long dim2938KuoInit[]        =   {   1   ,   1   ,   3   ,   15  ,   15  ,   37  ,   41  ,   123 ,   453 ,   879 ,   337 ,   251 ,   7137    ,   4117    ,   30879   ,0 };
            const unsigned long dim2939KuoInit[]        =   {   1   ,   3   ,   5   ,   3   ,   13  ,   45  ,   55  ,   151 ,   357 ,   429 ,   671 ,   1183    ,   3047    ,   4991    ,   15273   ,0 };
            const unsigned long dim2940KuoInit[]        =   {   1   ,   1   ,   3   ,   3   ,   15  ,   15  ,   65  ,   131 ,   505 ,   339 ,   869 ,   2983    ,   4451    ,   2375    ,   26745   ,0 };
            const unsigned long dim2941KuoInit[]        =   {   1   ,   3   ,   5   ,   9   ,   15  ,   55  ,   23  ,   83  ,   255 ,   395 ,   15  ,   1175    ,   849 ,   5959    ,   17355   ,0 };
            const unsigned long dim2942KuoInit[]        =   {   1   ,   1   ,   3   ,   7   ,   9   ,   3   ,   57  ,   95  ,   7   ,   267 ,   801 ,   3843    ,   3139    ,   12565   ,   20737   ,0 };
            const unsigned long dim2943KuoInit[]        =   {   1   ,   3   ,   7   ,   7   ,   9   ,   41  ,   121 ,   77  ,   489 ,   975 ,   215 ,   325 ,   7505    ,   15517   ,   8833    ,0 };
            const unsigned long dim2944KuoInit[]        =   {   1   ,   3   ,   3   ,   1   ,   9   ,   43  ,   105 ,   199 ,   339 ,   335 ,   1719    ,   1999    ,   4153    ,   9665    ,   15457   ,0 };
            const unsigned long dim2945KuoInit[]        =   {   1   ,   3   ,   1   ,   13  ,   1   ,   35  ,   47  ,   157 ,   3   ,   93  ,   1399    ,   471 ,   6337    ,   2789    ,   32727   ,0 };
            const unsigned long dim2946KuoInit[]        =   {   1   ,   3   ,   5   ,   1   ,   29  ,   9   ,   101 ,   195 ,   371 ,   573 ,   1361    ,   521 ,   5737    ,   4845    ,   4871    ,0 };
            const unsigned long dim2947KuoInit[]        =   {   1   ,   1   ,   7   ,   9   ,   9   ,   11  ,   43  ,   129 ,   295 ,   1015    ,   665 ,   451 ,   1249    ,   7733    ,   8861    ,0 };
            const unsigned long dim2948KuoInit[]        =   {   1   ,   1   ,   1   ,   11  ,   19  ,   31  ,   127 ,   171 ,   341 ,   865 ,   1391    ,   955 ,   5585    ,   1481    ,   1311    ,0 };
            const unsigned long dim2949KuoInit[]        =   {   1   ,   3   ,   1   ,   11  ,   19  ,   43  ,   113 ,   83  ,   285 ,   865 ,   815 ,   807 ,   779 ,   6685    ,   22315   ,0 };
            const unsigned long dim2950KuoInit[]        =   {   1   ,   1   ,   1   ,   11  ,   1   ,   19  ,   99  ,   19  ,   455 ,   155 ,   637 ,   2153    ,   3013    ,   4791    ,   20619   ,0 };
            const unsigned long dim2951KuoInit[]        =   {   1   ,   1   ,   1   ,   1   ,   13  ,   39  ,   95  ,   229 ,   35  ,   515 ,   1945    ,   3631    ,   3359    ,   7913    ,   8081    ,0 };
            const unsigned long dim2952KuoInit[]        =   {   1   ,   3   ,   7   ,   13  ,   1   ,   5   ,   25  ,   227 ,   487 ,   269 ,   1213    ,   303 ,   4041    ,   15847   ,   1125    ,0 };
            const unsigned long dim2953KuoInit[]        =   {   1   ,   1   ,   3   ,   7   ,   27  ,   53  ,   121 ,   143 ,   1   ,   977 ,   1963    ,   807 ,   253 ,   7649    ,   16047   ,0 };
            const unsigned long dim2954KuoInit[]        =   {   1   ,   3   ,   5   ,   9   ,   29  ,   17  ,   97  ,   93  ,   83  ,   739 ,   1221    ,   963 ,   6849    ,   4307    ,   17277   ,0 };
            const unsigned long dim2955KuoInit[]        =   {   1   ,   1   ,   7   ,   1   ,   27  ,   55  ,   71  ,   177 ,   63  ,   11  ,   1781    ,   3341    ,   2797    ,   8035    ,   6135    ,0 };
            const unsigned long dim2956KuoInit[]        =   {   1   ,   1   ,   1   ,   11  ,   15  ,   43  ,   59  ,   107 ,   485 ,   995 ,   1571    ,   1527    ,   2195    ,   4883    ,   32563   ,0 };
            const unsigned long dim2957KuoInit[]        =   {   1   ,   3   ,   5   ,   15  ,   7   ,   45  ,   55  ,   211 ,   425 ,   771 ,   621 ,   1141    ,   5277    ,   13575   ,   203 ,0 };
            const unsigned long dim2958KuoInit[]        =   {   1   ,   1   ,   7   ,   1   ,   21  ,   51  ,   81  ,   41  ,   71  ,   551 ,   839 ,   3315    ,   5075    ,   3651    ,   6649    ,0 };
            const unsigned long dim2959KuoInit[]        =   {   1   ,   1   ,   1   ,   9   ,   17  ,   45  ,   51  ,   211 ,   349 ,   329 ,   987 ,   3153    ,   7121    ,   16101   ,   1475    ,0 };
            const unsigned long dim2960KuoInit[]        =   {   1   ,   3   ,   3   ,   9   ,   5   ,   5   ,   99  ,   17  ,   245 ,   827 ,   231 ,   949 ,   2577    ,   12925   ,   2055    ,0 };
            const unsigned long dim2961KuoInit[]        =   {   1   ,   1   ,   5   ,   1   ,   19  ,   59  ,   75  ,   97  ,   303 ,   649 ,   901 ,   3987    ,   1209    ,   14483   ,   25855   ,0 };
            const unsigned long dim2962KuoInit[]        =   {   1   ,   3   ,   7   ,   15  ,   25  ,   51  ,   5   ,   69  ,   199 ,   845 ,   1847    ,   1491    ,   6183    ,   3359    ,   11809   ,0 };
            const unsigned long dim2963KuoInit[]        =   {   1   ,   1   ,   5   ,   5   ,   19  ,   43  ,   17  ,   113 ,   215 ,   337 ,   583 ,   2199    ,   6375    ,   499 ,   11283   ,0 };
            const unsigned long dim2964KuoInit[]        =   {   1   ,   1   ,   5   ,   15  ,   5   ,   3   ,   45  ,   219 ,   375 ,   269 ,   1251    ,   2711    ,   7897    ,   3379    ,   11887   ,0 };
            const unsigned long dim2965KuoInit[]        =   {   1   ,   1   ,   7   ,   13  ,   3   ,   59  ,   17  ,   41  ,   275 ,   1007    ,   711 ,   1689    ,   3727    ,   14237   ,   32057   ,0 };
            const unsigned long dim2966KuoInit[]        =   {   1   ,   3   ,   5   ,   7   ,   5   ,   27  ,   73  ,   167 ,   295 ,   581 ,   233 ,   585 ,   7207    ,   12709   ,   3847    ,0 };
            const unsigned long dim2967KuoInit[]        =   {   1   ,   1   ,   7   ,   5   ,   9   ,   17  ,   65  ,   15  ,   439 ,   385 ,   455 ,   1073    ,   5727    ,   2679    ,   3689    ,0 };
            const unsigned long dim2968KuoInit[]        =   {   1   ,   1   ,   7   ,   5   ,   7   ,   63  ,   111 ,   161 ,   85  ,   295 ,   755 ,   2805    ,   3031    ,   4833    ,   11149   ,0 };
            const unsigned long dim2969KuoInit[]        =   {   1   ,   1   ,   5   ,   7   ,   13  ,   53  ,   63  ,   47  ,   163 ,   275 ,   289 ,   1377    ,   639 ,   15367   ,   32517   ,0 };
            const unsigned long dim2970KuoInit[]        =   {   1   ,   1   ,   7   ,   9   ,   21  ,   43  ,   123 ,   221 ,   451 ,   901 ,   1339    ,   3367    ,   2531    ,   9319    ,   4461    ,0 };
            const unsigned long dim2971KuoInit[]        =   {   1   ,   3   ,   3   ,   9   ,   3   ,   53  ,   87  ,   117 ,   471 ,   349 ,   1487    ,   2375    ,   7399    ,   8561    ,   19611   ,0 };
            const unsigned long dim2972KuoInit[]        =   {   1   ,   3   ,   1   ,   7   ,   29  ,   55  ,   53  ,   197 ,   301 ,   701 ,   357 ,   3651    ,   2233    ,   6365    ,   705 ,0 };
            const unsigned long dim2973KuoInit[]        =   {   1   ,   3   ,   5   ,   3   ,   17  ,   19  ,   23  ,   17  ,   433 ,   979 ,   1377    ,   213 ,   3993    ,   3773    ,   24369   ,0 };
            const unsigned long dim2974KuoInit[]        =   {   1   ,   3   ,   5   ,   5   ,   17  ,   9   ,   105 ,   149 ,   145 ,   751 ,   1173    ,   1831    ,   7005    ,   3559    ,   10913   ,0 };
            const unsigned long dim2975KuoInit[]        =   {   1   ,   3   ,   1   ,   9   ,   23  ,   47  ,   83  ,   213 ,   135 ,   223 ,   937 ,   3607    ,   7375    ,   4731    ,   14307   ,0 };
            const unsigned long dim2976KuoInit[]        =   {   1   ,   1   ,   7   ,   3   ,   23  ,   21  ,   71  ,   241 ,   399 ,   199 ,   289 ,   2049    ,   7959    ,   14897   ,   18661   ,0 };
            const unsigned long dim2977KuoInit[]        =   {   1   ,   3   ,   3   ,   9   ,   13  ,   45  ,   125 ,   217 ,   325 ,   893 ,   1523    ,   2477    ,   7829    ,   15737   ,   2127    ,0 };
            const unsigned long dim2978KuoInit[]        =   {   1   ,   1   ,   5   ,   1   ,   21  ,   51  ,   117 ,   203 ,   253 ,   713 ,   545 ,   191 ,   729 ,   8403    ,   8381    ,0 };
            const unsigned long dim2979KuoInit[]        =   {   1   ,   3   ,   7   ,   7   ,   11  ,   23  ,   3   ,   193 ,   421 ,   341 ,   1309    ,   2457    ,   5805    ,   7457    ,   26857   ,0 };
            const unsigned long dim2980KuoInit[]        =   {   1   ,   3   ,   7   ,   15  ,   1   ,   17  ,   27  ,   159 ,   127 ,   713 ,   1465    ,   3551    ,   7409    ,   7957    ,   2317    ,0 };
            const unsigned long dim2981KuoInit[]        =   {   1   ,   3   ,   1   ,   15  ,   7   ,   33  ,   17  ,   1   ,   237 ,   199 ,   527 ,   3361    ,   3785    ,   13111   ,   19119   ,0 };
            const unsigned long dim2982KuoInit[]        =   {   1   ,   3   ,   3   ,   1   ,   23  ,   37  ,   91  ,   31  ,   35  ,   969 ,   661 ,   85  ,   5917    ,   10895   ,   11049   ,0 };
            const unsigned long dim2983KuoInit[]        =   {   1   ,   3   ,   1   ,   7   ,   23  ,   53  ,   101 ,   181 ,   381 ,   159 ,   1967    ,   3427    ,   4499    ,   6877    ,   12155   ,0 };
            const unsigned long dim2984KuoInit[]        =   {   1   ,   3   ,   3   ,   13  ,   9   ,   57  ,   77  ,   27  ,   223 ,   101 ,   1709    ,   3023    ,   6655    ,   6221    ,   1027    ,0 };
            const unsigned long dim2985KuoInit[]        =   {   1   ,   3   ,   5   ,   7   ,   19  ,   47  ,   73  ,   175 ,   103 ,   35  ,   1039    ,   343 ,   6757    ,   11521   ,   24495   ,0 };
            const unsigned long dim2986KuoInit[]        =   {   1   ,   3   ,   1   ,   9   ,   17  ,   19  ,   9   ,   205 ,   279 ,   373 ,   173 ,   3641    ,   3945    ,   10831   ,   15621   ,0 };
            const unsigned long dim2987KuoInit[]        =   {   1   ,   1   ,   3   ,   13  ,   29  ,   63  ,   71  ,   217 ,   367 ,   381 ,   647 ,   9   ,   5521    ,   5209    ,   24275   ,0 };
            const unsigned long dim2988KuoInit[]        =   {   1   ,   3   ,   7   ,   7   ,   23  ,   15  ,   7   ,   203 ,   153 ,   655 ,   2047    ,   287 ,   973 ,   6757    ,   29207   ,0 };
            const unsigned long dim2989KuoInit[]        =   {   1   ,   1   ,   1   ,   9   ,   17  ,   33  ,   63  ,   97  ,   49  ,   31  ,   1989    ,   569 ,   3321    ,   4017    ,   6437    ,0 };
            const unsigned long dim2990KuoInit[]        =   {   1   ,   1   ,   7   ,   11  ,   15  ,   15  ,   57  ,   255 ,   381 ,   115 ,   573 ,   3903    ,   7141    ,   1569    ,   11789   ,0 };
            const unsigned long dim2991KuoInit[]        =   {   1   ,   1   ,   5   ,   13  ,   9   ,   45  ,   69  ,   111 ,   91  ,   389 ,   1347    ,   3351    ,   3023    ,   517 ,   28373   ,0 };
            const unsigned long dim2992KuoInit[]        =   {   1   ,   3   ,   3   ,   15  ,   9   ,   11  ,   81  ,   205 ,   299 ,   903 ,   661 ,   3377    ,   5355    ,   9151    ,   17597   ,0 };
            const unsigned long dim2993KuoInit[]        =   {   1   ,   1   ,   3   ,   9   ,   19  ,   55  ,   45  ,   123 ,   285 ,   161 ,   1591    ,   1971    ,   5471    ,   8221    ,   2855    ,0 };
            const unsigned long dim2994KuoInit[]        =   {   1   ,   3   ,   5   ,   13  ,   23  ,   25  ,   87  ,   197 ,   303 ,   487 ,   395 ,   3137    ,   4395    ,   5861    ,   5847    ,0 };
            const unsigned long dim2995KuoInit[]        =   {   1   ,   3   ,   7   ,   5   ,   23  ,   9   ,   107 ,   13  ,   127 ,   971 ,   1701    ,   2387    ,   6195    ,   5041    ,   24857   ,0 };
            const unsigned long dim2996KuoInit[]        =   {   1   ,   1   ,   3   ,   3   ,   19  ,   63  ,   91  ,   205 ,   207 ,   897 ,   295 ,   2093    ,   1223    ,   14599   ,   13803   ,0 };
            const unsigned long dim2997KuoInit[]        =   {   1   ,   1   ,   5   ,   1   ,   1   ,   25  ,   93  ,   33  ,   393 ,   221 ,   1845    ,   2517    ,   1723    ,   13881   ,   15777   ,0 };
            const unsigned long dim2998KuoInit[]        =   {   1   ,   3   ,   5   ,   3   ,   1   ,   21  ,   115 ,   103 ,   377 ,   789 ,   541 ,   3423    ,   3111    ,   8077    ,   25089   ,0 };
            const unsigned long dim2999KuoInit[]        =   {   1   ,   1   ,   1   ,   11  ,   31  ,   55  ,   95  ,   53  ,   141 ,   697 ,   1085    ,   3031    ,   653 ,   4145    ,   28121   ,0 };
            const unsigned long dim3000KuoInit[]        =   {   1   ,   3   ,   1   ,   1   ,   7   ,   39  ,   55  ,   57  ,   447 ,   257 ,   101 ,   2307    ,   5623    ,   8505    ,   995 ,0 };
            const unsigned long dim3001KuoInit[]        =   {   1   ,   3   ,   5   ,   11  ,   25  ,   37  ,   11  ,   203 ,   361 ,   281 ,   701 ,   3643    ,   957 ,   14541   ,   8109    ,0 };
            const unsigned long dim3002KuoInit[]        =   {   1   ,   3   ,   1   ,   3   ,   21  ,   49  ,   57  ,   225 ,   21  ,   523 ,   1275    ,   3743    ,   6259    ,   13625   ,   26721   ,0 };
            const unsigned long dim3003KuoInit[]        =   {   1   ,   1   ,   1   ,   1   ,   13  ,   53  ,   9   ,   167 ,   117 ,   971 ,   875 ,   2659    ,   7589    ,   13575   ,   11659   ,0 };
            const unsigned long dim3004KuoInit[]        =   {   1   ,   1   ,   7   ,   13  ,   21  ,   1   ,   35  ,   201 ,   261 ,   533 ,   1069    ,   1459    ,   5433    ,   7211    ,   3535    ,0 };
            const unsigned long dim3005KuoInit[]        =   {   1   ,   1   ,   3   ,   13  ,   23  ,   7   ,   41  ,   137 ,   341 ,   767 ,   1823    ,   3745    ,   1415    ,   4811    ,   21669   ,0 };
            const unsigned long dim3006KuoInit[]        =   {   1   ,   3   ,   3   ,   5   ,   5   ,   11  ,   71  ,   117 ,   451 ,   1   ,   381 ,   967 ,   7977    ,   10741   ,   20141   ,0 };
            const unsigned long dim3007KuoInit[]        =   {   1   ,   3   ,   1   ,   3   ,   21  ,   1   ,   65  ,   37  ,   451 ,   29  ,   327 ,   2471    ,   7573    ,   5051    ,   23925   ,0 };
            const unsigned long dim3008KuoInit[]        =   {   1   ,   3   ,   3   ,   11  ,   17  ,   61  ,   97  ,   121 ,   327 ,   929 ,   1165    ,   185 ,   2343    ,   1479    ,   3817    ,0 };
            const unsigned long dim3009KuoInit[]        =   {   1   ,   1   ,   3   ,   7   ,   27  ,   55  ,   87  ,   53  ,   389 ,   577 ,   1051    ,   2413    ,   609 ,   12651   ,   15695   ,0 };
            const unsigned long dim3010KuoInit[]        =   {   1   ,   3   ,   5   ,   13  ,   1   ,   41  ,   25  ,   149 ,   175 ,   667 ,   1219    ,   2297    ,   6599    ,   1667    ,   19647   ,0 };
            const unsigned long dim3011KuoInit[]        =   {   1   ,   1   ,   3   ,   11  ,   17  ,   63  ,   83  ,   37  ,   425 ,   675 ,   1111    ,   499 ,   7727    ,   7491    ,   9759    ,0 };
            const unsigned long dim3012KuoInit[]        =   {   1   ,   1   ,   3   ,   1   ,   21  ,   23  ,   17  ,   125 ,   485 ,   643 ,   341 ,   1159    ,   7073    ,   14405   ,   22409   ,0 };
            const unsigned long dim3013KuoInit[]        =   {   1   ,   3   ,   7   ,   15  ,   15  ,   53  ,   23  ,   17  ,   365 ,   101 ,   685 ,   1391    ,   5123    ,   14961   ,   31179   ,0 };
            const unsigned long dim3014KuoInit[]        =   {   1   ,   3   ,   3   ,   3   ,   23  ,   11  ,   101 ,   139 ,   401 ,   779 ,   1723    ,   3421    ,   623 ,   3443    ,   17179   ,0 };
            const unsigned long dim3015KuoInit[]        =   {   1   ,   1   ,   1   ,   7   ,   13  ,   19  ,   93  ,   129 ,   127 ,   389 ,   1361    ,   2341    ,   6491    ,   15063   ,   8199    ,0 };
            const unsigned long dim3016KuoInit[]        =   {   1   ,   1   ,   7   ,   1   ,   31  ,   1   ,   55  ,   169 ,   373 ,   69  ,   539 ,   513 ,   6329    ,   393 ,   21963   ,0 };
            const unsigned long dim3017KuoInit[]        =   {   1   ,   1   ,   7   ,   1   ,   5   ,   39  ,   73  ,   87  ,   401 ,   923 ,   1747    ,   565 ,   2227    ,   14509   ,   7601    ,0 };
            const unsigned long dim3018KuoInit[]        =   {   1   ,   3   ,   3   ,   1   ,   25  ,   3   ,   77  ,   207 ,   149 ,   591 ,   1165    ,   3235    ,   7615    ,   12237   ,   20849   ,0 };
            const unsigned long dim3019KuoInit[]        =   {   1   ,   3   ,   1   ,   11  ,   21  ,   41  ,   83  ,   109 ,   265 ,   349 ,   485 ,   1411    ,   2705    ,   13017   ,   25205   ,0 };
            const unsigned long dim3020KuoInit[]        =   {   1   ,   1   ,   7   ,   7   ,   31  ,   27  ,   55  ,   23  ,   129 ,   369 ,   1785    ,   2031    ,   1993    ,   123 ,   20063   ,0 };
            const unsigned long dim3021KuoInit[]        =   {   1   ,   3   ,   5   ,   9   ,   9   ,   45  ,   75  ,   37  ,   69  ,   957 ,   2047    ,   2515    ,   5327    ,   15345   ,   2147    ,0 };
            const unsigned long dim3022KuoInit[]        =   {   1   ,   3   ,   1   ,   7   ,   5   ,   57  ,   91  ,   19  ,   499 ,   173 ,   765 ,   3653    ,   6397    ,   1509    ,   11223   ,0 };
            const unsigned long dim3023KuoInit[]        =   {   1   ,   3   ,   5   ,   15  ,   1   ,   15  ,   69  ,   107 ,   225 ,   233 ,   1015    ,   807 ,   4265    ,   8823    ,   22047   ,0 };
            const unsigned long dim3024KuoInit[]        =   {   1   ,   1   ,   7   ,   13  ,   15  ,   55  ,   15  ,   83  ,   49  ,   711 ,   935 ,   1865    ,   29  ,   607 ,   8113    ,0 };
            const unsigned long dim3025KuoInit[]        =   {   1   ,   3   ,   1   ,   15  ,   13  ,   15  ,   127 ,   193 ,   219 ,   653 ,   1787    ,   905 ,   3311    ,   11595   ,   21193   ,0 };
            const unsigned long dim3026KuoInit[]        =   {   1   ,   3   ,   5   ,   11  ,   21  ,   51  ,   67  ,   97  ,   371 ,   367 ,   1763    ,   2953    ,   3701    ,   15983   ,   18631   ,0 };
            const unsigned long dim3027KuoInit[]        =   {   1   ,   1   ,   5   ,   9   ,   13  ,   1   ,   83  ,   15  ,   305 ,   939 ,   1073    ,   1577    ,   2073    ,   5335    ,   17173   ,0 };
            const unsigned long dim3028KuoInit[]        =   {   1   ,   3   ,   5   ,   7   ,   19  ,   7   ,   27  ,   213 ,   357 ,   681 ,   1969    ,   4007    ,   7887    ,   11191   ,   19395   ,0 };
            const unsigned long dim3029KuoInit[]        =   {   1   ,   1   ,   5   ,   3   ,   19  ,   29  ,   119 ,   239 ,   123 ,   559 ,   1613    ,   1563    ,   8047    ,   5753    ,   26719   ,0 };
            const unsigned long dim3030KuoInit[]        =   {   1   ,   3   ,   7   ,   5   ,   3   ,   3   ,   127 ,   241 ,   333 ,   453 ,   1407    ,   2687    ,   4603    ,   3637    ,   23041   ,0 };
            const unsigned long dim3031KuoInit[]        =   {   1   ,   1   ,   1   ,   15  ,   17  ,   9   ,   7   ,   31  ,   145 ,   613 ,   1851    ,   1981    ,   3921    ,   6469    ,   32397   ,0 };
            const unsigned long dim3032KuoInit[]        =   {   1   ,   1   ,   5   ,   3   ,   9   ,   39  ,   113 ,   79  ,   165 ,   359 ,   1573    ,   2311    ,   651 ,   7047    ,   15961   ,0 };
            const unsigned long dim3033KuoInit[]        =   {   1   ,   3   ,   7   ,   7   ,   11  ,   57  ,   59  ,   153 ,   363 ,   279 ,   77  ,   3085    ,   5971    ,   29  ,   22087   ,0 };
            const unsigned long dim3034KuoInit[]        =   {   1   ,   1   ,   7   ,   11  ,   25  ,   47  ,   49  ,   229 ,   147 ,   777 ,   1177    ,   1117    ,   3783    ,   3681    ,   1465    ,0 };
            const unsigned long dim3035KuoInit[]        =   {   1   ,   3   ,   5   ,   11  ,   3   ,   45  ,   11  ,   139 ,   393 ,   9   ,   471 ,   1643    ,   4963    ,   5837    ,   17023   ,0 };
            const unsigned long dim3036KuoInit[]        =   {   1   ,   3   ,   1   ,   7   ,   5   ,   41  ,   123 ,   113 ,   223 ,   237 ,   239 ,   1099    ,   6065    ,   11105   ,   10259   ,0 };
            const unsigned long dim3037KuoInit[]        =   {   1   ,   1   ,   3   ,   11  ,   15  ,   57  ,   41  ,   173 ,   169 ,   585 ,   201 ,   2133    ,   2629    ,   12861   ,   16559   ,0 };
            const unsigned long dim3038KuoInit[]        =   {   1   ,   1   ,   7   ,   5   ,   9   ,   27  ,   3   ,   157 ,   279 ,   393 ,   1849    ,   437 ,   5549    ,   10009   ,   11783   ,0 };
            const unsigned long dim3039KuoInit[]        =   {   1   ,   1   ,   7   ,   9   ,   17  ,   23  ,   67  ,   143 ,   1   ,   203 ,   153 ,   1699    ,   1333    ,   16001   ,   11755   ,0 };
            const unsigned long dim3040KuoInit[]        =   {   1   ,   3   ,   5   ,   15  ,   17  ,   61  ,   59  ,   235 ,   67  ,   393 ,   1597    ,   3523    ,   2049    ,   11887   ,   7737    ,0 };
            const unsigned long dim3041KuoInit[]        =   {   1   ,   3   ,   5   ,   13  ,   21  ,   45  ,   5   ,   221 ,   299 ,   887 ,   699 ,   959 ,   1143    ,   4879    ,   8767    ,0 };
            const unsigned long dim3042KuoInit[]        =   {   1   ,   3   ,   7   ,   15  ,   11  ,   35  ,   83  ,   237 ,   167 ,   1   ,   1429    ,   2877    ,   4611    ,   14881   ,   4453    ,0 };
            const unsigned long dim3043KuoInit[]        =   {   1   ,   3   ,   5   ,   3   ,   7   ,   55  ,   21  ,   237 ,   9   ,   743 ,   1707    ,   1229    ,   6827    ,   407 ,   777 ,0 };
            const unsigned long dim3044KuoInit[]        =   {   1   ,   3   ,   7   ,   3   ,   15  ,   39  ,   3   ,   127 ,   163 ,   747 ,   1773    ,   1569    ,   2377    ,   15023   ,   32249   ,0 };
            const unsigned long dim3045KuoInit[]        =   {   1   ,   1   ,   7   ,   7   ,   19  ,   33  ,   1   ,   55  ,   319 ,   585 ,   413 ,   1241    ,   95  ,   7163    ,   29935   ,0 };
            const unsigned long dim3046KuoInit[]        =   {   1   ,   3   ,   1   ,   1   ,   7   ,   51  ,   77  ,   197 ,   481 ,   517 ,   1441    ,   1779    ,   1405    ,   3633    ,   32545   ,0 };
            const unsigned long dim3047KuoInit[]        =   {   1   ,   3   ,   7   ,   5   ,   13  ,   51  ,   69  ,   35  ,   131 ,   115 ,   1759    ,   3991    ,   8065    ,   10311   ,   12749   ,0 };
            const unsigned long dim3048KuoInit[]        =   {   1   ,   3   ,   3   ,   5   ,   5   ,   35  ,   37  ,   181 ,   371 ,   543 ,   1389    ,   815 ,   3261    ,   11997   ,   29665   ,0 };
            const unsigned long dim3049KuoInit[]        =   {   1   ,   3   ,   7   ,   11  ,   31  ,   47  ,   33  ,   147 ,   445 ,   663 ,   63  ,   1033    ,   6081    ,   15523   ,   28477   ,0 };
            const unsigned long dim3050KuoInit[]        =   {   1   ,   1   ,   7   ,   11  ,   11  ,   9   ,   1   ,   39  ,   357 ,   873 ,   1829    ,   3245    ,   3219    ,   601 ,   1335    ,0 };
            const unsigned long dim3051KuoInit[]        =   {   1   ,   1   ,   5   ,   7   ,   5   ,   5   ,   33  ,   145 ,   465 ,   5   ,   1959    ,   3573    ,   3791    ,   8689    ,   14513   ,0 };
            const unsigned long dim3052KuoInit[]        =   {   1   ,   1   ,   1   ,   13  ,   31  ,   37  ,   103 ,   255 ,   87  ,   903 ,   1325    ,   1423    ,   8031    ,   14845   ,   12261   ,0 };
            const unsigned long dim3053KuoInit[]        =   {   1   ,   3   ,   3   ,   13  ,   31  ,   25  ,   41  ,   45  ,   261 ,   225 ,   1711    ,   641 ,   2331    ,   10825   ,   21539   ,0 };
            const unsigned long dim3054KuoInit[]        =   {   1   ,   3   ,   1   ,   9   ,   9   ,   33  ,   23  ,   135 ,   165 ,   363 ,   1435    ,   3019    ,   4907    ,   9791    ,   1337    ,0 };
            const unsigned long dim3055KuoInit[]        =   {   1   ,   3   ,   3   ,   3   ,   13  ,   23  ,   109 ,   161 ,   107 ,   829 ,   1635    ,   791 ,   2757    ,   15265   ,   8771    ,0 };
            const unsigned long dim3056KuoInit[]        =   {   1   ,   1   ,   1   ,   11  ,   7   ,   17  ,   45  ,   167 ,   69  ,   437 ,   1051    ,   1443    ,   875 ,   2461    ,   28471   ,0 };
            const unsigned long dim3057KuoInit[]        =   {   1   ,   3   ,   5   ,   11  ,   23  ,   43  ,   55  ,   69  ,   365 ,   497 ,   1529    ,   1101    ,   211 ,   8145    ,   10579   ,0 };
            const unsigned long dim3058KuoInit[]        =   {   1   ,   1   ,   5   ,   3   ,   19  ,   9   ,   91  ,   91  ,   453 ,   875 ,   1815    ,   1455    ,   7969    ,   5621    ,   4773    ,0 };
            const unsigned long dim3059KuoInit[]        =   {   1   ,   1   ,   5   ,   1   ,   17  ,   63  ,   19  ,   193 ,   335 ,   571 ,   703 ,   919 ,   773 ,   1789    ,   17839   ,0 };
            const unsigned long dim3060KuoInit[]        =   {   1   ,   1   ,   5   ,   7   ,   1   ,   57  ,   11  ,   171 ,   11  ,   833 ,   191 ,   1719    ,   7649    ,   11005   ,   20171   ,0 };
            const unsigned long dim3061KuoInit[]        =   {   1   ,   1   ,   7   ,   15  ,   25  ,   55  ,   63  ,   71  ,   127 ,   445 ,   1151    ,   2709    ,   6633    ,   4959    ,   13511   ,0 };
            const unsigned long dim3062KuoInit[]        =   {   1   ,   1   ,   3   ,   1   ,   3   ,   41  ,   43  ,   59  ,   211 ,   123 ,   987 ,   3701    ,   6391    ,   6755    ,   3069    ,0 };
            const unsigned long dim3063KuoInit[]        =   {   1   ,   1   ,   3   ,   11  ,   5   ,   3   ,   93  ,   29  ,   233 ,   365 ,   117 ,   1483    ,   4917    ,   1949    ,   23697   ,0 };
            const unsigned long dim3064KuoInit[]        =   {   1   ,   3   ,   5   ,   1   ,   15  ,   53  ,   21  ,   9   ,   23  ,   701 ,   1337    ,   2591    ,   7719    ,   14063   ,   28453   ,0 };
            const unsigned long dim3065KuoInit[]        =   {   1   ,   1   ,   1   ,   11  ,   21  ,   3   ,   127 ,   179 ,   409 ,   261 ,   1683    ,   3197    ,   1999    ,   10009   ,   25031   ,0 };
            const unsigned long dim3066KuoInit[]        =   {   1   ,   3   ,   5   ,   13  ,   31  ,   57  ,   9   ,   91  ,   31  ,   125 ,   1831    ,   1943    ,   5931    ,   15015   ,   22269   ,0 };
            const unsigned long dim3067KuoInit[]        =   {   1   ,   3   ,   5   ,   9   ,   7   ,   7   ,   83  ,   113 ,   203 ,   679 ,   1279    ,   3403    ,   7069    ,   8617    ,   27827   ,0 };
            const unsigned long dim3068KuoInit[]        =   {   1   ,   3   ,   7   ,   5   ,   29  ,   47  ,   93  ,   59  ,   315 ,   107 ,   807 ,   2861    ,   2345    ,   11787   ,   6543    ,0 };
            const unsigned long dim3069KuoInit[]        =   {   1   ,   3   ,   1   ,   7   ,   21  ,   25  ,   3   ,   35  ,   345 ,   833 ,   2029    ,   281 ,   2389    ,   11529   ,   20647   ,0 };
            const unsigned long dim3070KuoInit[]        =   {   1   ,   1   ,   1   ,   13  ,   19  ,   47  ,   109 ,   21  ,   251 ,   625 ,   1699    ,   2169    ,   7293    ,   14035   ,   24277   ,0 };
            const unsigned long dim3071KuoInit[]        =   {   1   ,   3   ,   5   ,   13  ,   15  ,   7   ,   51  ,   125 ,   423 ,   981 ,   437 ,   1021    ,   5209    ,   11447   ,   26181   ,0 };
            const unsigned long dim3072KuoInit[]        =   {   1   ,   3   ,   1   ,   1   ,   15  ,   41  ,   75  ,   227 ,   13  ,   85  ,   675 ,   3993    ,   8163    ,   8095    ,   23801   ,0 };
            const unsigned long dim3073KuoInit[]        =   {   1   ,   3   ,   3   ,   9   ,   25  ,   1   ,   85  ,   143 ,   309 ,   815 ,   1271    ,   1331    ,   6565    ,   369 ,   7975    ,0 };
            const unsigned long dim3074KuoInit[]        =   {   1   ,   1   ,   7   ,   1   ,   21  ,   45  ,   113 ,   153 ,   283 ,   117 ,   1481    ,   3527    ,   3371    ,   15801   ,   22337   ,0 };
            const unsigned long dim3075KuoInit[]        =   {   1   ,   3   ,   5   ,   1   ,   17  ,   47  ,   107 ,   25  ,   373 ,   549 ,   1387    ,   1247    ,   2307    ,   6469    ,   24327   ,0 };
            const unsigned long dim3076KuoInit[]        =   {   1   ,   1   ,   5   ,   15  ,   9   ,   57  ,   43  ,   179 ,   157 ,   119 ,   547 ,   3665    ,   1495    ,   13405   ,   2391    ,0 };
            const unsigned long dim3077KuoInit[]        =   {   1   ,   3   ,   1   ,   1   ,   7   ,   57  ,   123 ,   189 ,   281 ,   87  ,   1559    ,   2863    ,   8017    ,   10279   ,   32005   ,0 };
            const unsigned long dim3078KuoInit[]        =   {   1   ,   1   ,   5   ,   1   ,   29  ,   21  ,   51  ,   123 ,   351 ,   683 ,   1505    ,   681 ,   49  ,   2929    ,   7591    ,0 };
            const unsigned long dim3079KuoInit[]        =   {   1   ,   1   ,   3   ,   7   ,   17  ,   63  ,   119 ,   209 ,   441 ,   717 ,   567 ,   1283    ,   1007    ,   14661   ,   32383   ,0 };
            const unsigned long dim3080KuoInit[]        =   {   1   ,   3   ,   1   ,   7   ,   27  ,   41  ,   41  ,   157 ,   487 ,   257 ,   1267    ,   165 ,   1023    ,   16065   ,   16431   ,0 };
            const unsigned long dim3081KuoInit[]        =   {   1   ,   1   ,   3   ,   9   ,   17  ,   53  ,   51  ,   175 ,   367 ,   453 ,   1445    ,   1889    ,   4567    ,   1781    ,   6543    ,0 };
            const unsigned long dim3082KuoInit[]        =   {   1   ,   1   ,   7   ,   1   ,   27  ,   47  ,   95  ,   51  ,   483 ,   581 ,   1617    ,   685 ,   3017    ,   4889    ,   6185    ,0 };
            const unsigned long dim3083KuoInit[]        =   {   1   ,   3   ,   3   ,   9   ,   29  ,   23  ,   9   ,   23  ,   165 ,   3   ,   279 ,   861 ,   1007    ,   9607    ,   24999   ,0 };
            const unsigned long dim3084KuoInit[]        =   {   1   ,   1   ,   1   ,   5   ,   11  ,   59  ,   55  ,   135 ,   399 ,   729 ,   683 ,   2119    ,   3781    ,   15525   ,   1125    ,0 };
            const unsigned long dim3085KuoInit[]        =   {   1   ,   3   ,   3   ,   15  ,   25  ,   61  ,   123 ,   153 ,   267 ,   823 ,   951 ,   4069    ,   8017    ,   4899    ,   6535    ,0 };
            const unsigned long dim3086KuoInit[]        =   {   1   ,   1   ,   7   ,   11  ,   5   ,   43  ,   5   ,   123 ,   57  ,   229 ,   129 ,   131 ,   7491    ,   2105    ,   16411   ,0 };
            const unsigned long dim3087KuoInit[]        =   {   1   ,   1   ,   5   ,   11  ,   9   ,   37  ,   123 ,   211 ,   331 ,   623 ,   1469    ,   1155    ,   6327    ,   12591   ,   18971   ,0 };
            const unsigned long dim3088KuoInit[]        =   {   1   ,   1   ,   1   ,   11  ,   17  ,   35  ,   59  ,   137 ,   451 ,   829 ,   365 ,   305 ,   5857    ,   2095    ,   14597   ,0 };
            const unsigned long dim3089KuoInit[]        =   {   1   ,   3   ,   7   ,   7   ,   15  ,   11  ,   23  ,   201 ,   247 ,   883 ,   1103    ,   4035    ,   5915    ,   4275    ,   12511   ,0 };
            const unsigned long dim3090KuoInit[]        =   {   1   ,   3   ,   7   ,   1   ,   27  ,   19  ,   87  ,   189 ,   91  ,   113 ,   237 ,   2073    ,   419 ,   1339    ,   16705   ,0 };
            const unsigned long dim3091KuoInit[]        =   {   1   ,   3   ,   7   ,   7   ,   15  ,   59  ,   99  ,   237 ,   167 ,   327 ,   1507    ,   771 ,   3743    ,   1723    ,   28873   ,0 };
            const unsigned long dim3092KuoInit[]        =   {   1   ,   3   ,   5   ,   15  ,   27  ,   19  ,   35  ,   175 ,   403 ,   1   ,   617 ,   2425    ,   2021    ,   12547   ,   23655   ,0 };
            const unsigned long dim3093KuoInit[]        =   {   1   ,   1   ,   7   ,   1   ,   7   ,   1   ,   45  ,   169 ,   43  ,   501 ,   1653    ,   593 ,   7219    ,   5501    ,   3671    ,0 };
            const unsigned long dim3094KuoInit[]        =   {   1   ,   1   ,   5   ,   3   ,   23  ,   53  ,   117 ,   165 ,   329 ,   177 ,   89  ,   2269    ,   7897    ,   11225   ,   25807   ,0 };
            const unsigned long dim3095KuoInit[]        =   {   1   ,   1   ,   5   ,   11  ,   25  ,   41  ,   91  ,   43  ,   107 ,   699 ,   1745    ,   3117    ,   7497    ,   6579    ,   21803   ,0 };
            const unsigned long dim3096KuoInit[]        =   {   1   ,   3   ,   3   ,   13  ,   19  ,   9   ,   103 ,   65  ,   413 ,   459 ,   1751    ,   4025    ,   6695    ,   9589    ,   1443    ,0 };
            const unsigned long dim3097KuoInit[]        =   {   1   ,   3   ,   5   ,   15  ,   31  ,   31  ,   23  ,   251 ,   343 ,   809 ,   283 ,   1501    ,   445 ,   7945    ,   26993   ,0 };
            const unsigned long dim3098KuoInit[]        =   {   1   ,   3   ,   7   ,   11  ,   1   ,   11  ,   113 ,   235 ,   127 ,   45  ,   109 ,   2421    ,   3955    ,   1017    ,   21497   ,0 };
            const unsigned long dim3099KuoInit[]        =   {   1   ,   3   ,   1   ,   13  ,   5   ,   57  ,   47  ,   227 ,   447 ,   499 ,   1755    ,   3527    ,   1625    ,   13341   ,   22301   ,0 };
            const unsigned long dim3100KuoInit[]        =   {   1   ,   3   ,   1   ,   9   ,   13  ,   9   ,   113 ,   101 ,   377 ,   39  ,   991 ,   3791    ,   7351    ,   12841   ,   2911    ,0 };
            const unsigned long dim3101KuoInit[]        =   {   1   ,   1   ,   3   ,   1   ,   1   ,   31  ,   111 ,   85  ,   247 ,   387 ,   1907    ,   3061    ,   569 ,   5845    ,   6729    ,0 };
            const unsigned long dim3102KuoInit[]        =   {   1   ,   3   ,   7   ,   15  ,   31  ,   17  ,   77  ,   33  ,   109 ,   661 ,   67  ,   3725    ,   955 ,   9475    ,   14345   ,0 };
            const unsigned long dim3103KuoInit[]        =   {   1   ,   1   ,   7   ,   15  ,   11  ,   23  ,   63  ,   23  ,   67  ,   843 ,   1747    ,   2441    ,   2063    ,   8575    ,   11921   ,0 };
            const unsigned long dim3104KuoInit[]        =   {   1   ,   3   ,   5   ,   13  ,   25  ,   19  ,   43  ,   177 ,   67  ,   57  ,   1127    ,   3817    ,   3097    ,   6143    ,   27757   ,0 };
            const unsigned long dim3105KuoInit[]        =   {   1   ,   1   ,   3   ,   1   ,   19  ,   25  ,   65  ,   135 ,   143 ,   863 ,   1013    ,   1285    ,   6133    ,   13491   ,   5093    ,0 };
            const unsigned long dim3106KuoInit[]        =   {   1   ,   3   ,   5   ,   15  ,   7   ,   17  ,   61  ,   199 ,   39  ,   767 ,   295 ,   127 ,   335 ,   10789   ,   20473   ,0 };
            const unsigned long dim3107KuoInit[]        =   {   1   ,   3   ,   3   ,   15  ,   15  ,   51  ,   87  ,   75  ,   439 ,   71  ,   1371    ,   1911    ,   1145    ,   4201    ,   26883   ,0 };
            const unsigned long dim3108KuoInit[]        =   {   1   ,   1   ,   3   ,   11  ,   3   ,   15  ,   79  ,   237 ,   43  ,   375 ,   1925    ,   1517    ,   6415    ,   303 ,   2217    ,0 };
            const unsigned long dim3109KuoInit[]        =   {   1   ,   3   ,   1   ,   3   ,   21  ,   23  ,   19  ,   83  ,   159 ,   249 ,   1297    ,   1853    ,   2609    ,   13635   ,   29119   ,0 };
            const unsigned long dim3110KuoInit[]        =   {   1   ,   1   ,   3   ,   9   ,   29  ,   53  ,   125 ,   73  ,   299 ,   45  ,   557 ,   3641    ,   1593    ,   10585   ,   22545   ,0 };
            const unsigned long dim3111KuoInit[]        =   {   1   ,   1   ,   3   ,   9   ,   13  ,   45  ,   53  ,   245 ,   401 ,   635 ,   785 ,   2841    ,   3033    ,   12377   ,   1235    ,0 };
            const unsigned long dim3112KuoInit[]        =   {   1   ,   1   ,   3   ,   11  ,   23  ,   35  ,   71  ,   207 ,   211 ,   901 ,   351 ,   1473    ,   6683    ,   981 ,   17449   ,0 };
            const unsigned long dim3113KuoInit[]        =   {   1   ,   1   ,   3   ,   9   ,   17  ,   13  ,   125 ,   213 ,   109 ,   631 ,   677 ,   2141    ,   1037    ,   6119    ,   25945   ,0 };
            const unsigned long dim3114KuoInit[]        =   {   1   ,   3   ,   7   ,   1   ,   29  ,   3   ,   53  ,   23  ,   45  ,   347 ,   27  ,   1537    ,   7833    ,   16315   ,   12337   ,0 };
            const unsigned long dim3115KuoInit[]        =   {   1   ,   3   ,   1   ,   1   ,   31  ,   23  ,   15  ,   241 ,   315 ,   977 ,   99  ,   3067    ,   3227    ,   6153    ,   9751    ,0 };
            const unsigned long dim3116KuoInit[]        =   {   1   ,   1   ,   7   ,   7   ,   13  ,   33  ,   5   ,   5   ,   29  ,   547 ,   1105    ,   1135    ,   7263    ,   7421    ,   9545    ,0 };
            const unsigned long dim3117KuoInit[]        =   {   1   ,   3   ,   7   ,   13  ,   19  ,   25  ,   53  ,   239 ,   65  ,   417 ,   1681    ,   659 ,   133 ,   4113    ,   16735   ,0 };
            const unsigned long dim3118KuoInit[]        =   {   1   ,   3   ,   7   ,   5   ,   3   ,   55  ,   57  ,   33  ,   357 ,   885 ,   941 ,   3471    ,   6863    ,   777 ,   19111   ,0 };
            const unsigned long dim3119KuoInit[]        =   {   1   ,   1   ,   1   ,   9   ,   27  ,   47  ,   65  ,   161 ,   115 ,   577 ,   1951    ,   363 ,   2045    ,   2497    ,   6931    ,0 };
            const unsigned long dim3120KuoInit[]        =   {   1   ,   3   ,   3   ,   11  ,   29  ,   37  ,   21  ,   235 ,   233 ,   833 ,   81  ,   2875    ,   4759    ,   8435    ,   16017   ,0 };
            const unsigned long dim3121KuoInit[]        =   {   1   ,   1   ,   3   ,   7   ,   9   ,   57  ,   89  ,   21  ,   291 ,   445 ,   1833    ,   735 ,   3935    ,   14943   ,   30987   ,0 };
            const unsigned long dim3122KuoInit[]        =   {   1   ,   1   ,   5   ,   13  ,   23  ,   9   ,   37  ,   71  ,   175 ,   899 ,   1897    ,   707 ,   6201    ,   15543   ,   13801   ,0 };
            const unsigned long dim3123KuoInit[]        =   {   1   ,   3   ,   7   ,   1   ,   13  ,   25  ,   17  ,   209 ,   455 ,   969 ,   755 ,   397 ,   4103    ,   3159    ,   28665   ,0 };
            const unsigned long dim3124KuoInit[]        =   {   1   ,   1   ,   1   ,   5   ,   15  ,   7   ,   17  ,   249 ,   299 ,   423 ,   83  ,   1203    ,   7397    ,   15657   ,   15661   ,0 };
            const unsigned long dim3125KuoInit[]        =   {   1   ,   1   ,   7   ,   7   ,   19  ,   19  ,   21  ,   231 ,   443 ,   729 ,   947 ,   1451    ,   3399    ,   5691    ,   11943   ,0 };
            const unsigned long dim3126KuoInit[]        =   {   1   ,   1   ,   5   ,   7   ,   15  ,   11  ,   89  ,   211 ,   459 ,   3   ,   741 ,   105 ,   1147    ,   927 ,   31805   ,0 };
            const unsigned long dim3127KuoInit[]        =   {   1   ,   1   ,   7   ,   9   ,   7   ,   1   ,   61  ,   145 ,   291 ,   1021    ,   221 ,   821 ,   4999    ,   3165    ,   22549   ,0 };
            const unsigned long dim3128KuoInit[]        =   {   1   ,   1   ,   1   ,   7   ,   5   ,   41  ,   101 ,   137 ,   143 ,   249 ,   87  ,   4045    ,   2301    ,   3833    ,   9841    ,0 };
            const unsigned long dim3129KuoInit[]        =   {   1   ,   1   ,   5   ,   5   ,   19  ,   61  ,   111 ,   109 ,   87  ,   65  ,   971 ,   3943    ,   6559    ,   9647    ,   3049    ,0 };
            const unsigned long dim3130KuoInit[]        =   {   1   ,   1   ,   1   ,   13  ,   9   ,   11  ,   41  ,   61  ,   145 ,   27  ,   1287    ,   3537    ,   1943    ,   10635   ,   32557   ,0 };
            const unsigned long dim3131KuoInit[]        =   {   1   ,   3   ,   5   ,   7   ,   17  ,   35  ,   29  ,   177 ,   5   ,   131 ,   465 ,   1235    ,   1383    ,   14253   ,   17595   ,0 };
            const unsigned long dim3132KuoInit[]        =   {   1   ,   1   ,   3   ,   7   ,   15  ,   63  ,   21  ,   119 ,   195 ,   627 ,   1635    ,   2369    ,   5549    ,   7641    ,   29481   ,0 };
            const unsigned long dim3133KuoInit[]        =   {   1   ,   1   ,   7   ,   13  ,   25  ,   3   ,   73  ,   233 ,   115 ,   749 ,   707 ,   1939    ,   3667    ,   1899    ,   22345   ,0 };
            const unsigned long dim3134KuoInit[]        =   {   1   ,   1   ,   5   ,   9   ,   15  ,   33  ,   17  ,   91  ,   507 ,   289 ,   1929    ,   2001    ,   629 ,   15559   ,   12725   ,0 };
            const unsigned long dim3135KuoInit[]        =   {   1   ,   1   ,   1   ,   13  ,   7   ,   7   ,   91  ,   171 ,   95  ,   257 ,   1611    ,   3527    ,   2769    ,   5779    ,   7783    ,0 };
            const unsigned long dim3136KuoInit[]        =   {   1   ,   1   ,   1   ,   13  ,   27  ,   27  ,   55  ,   113 ,   171 ,   61  ,   1229    ,   3681    ,   4487    ,   9815    ,   20357   ,0 };
            const unsigned long dim3137KuoInit[]        =   {   1   ,   3   ,   7   ,   15  ,   1   ,   21  ,   57  ,   67  ,   185 ,   189 ,   1673    ,   3637    ,   2987    ,   4619    ,   6339    ,0 };
            const unsigned long dim3138KuoInit[]        =   {   1   ,   3   ,   7   ,   3   ,   9   ,   45  ,   71  ,   183 ,   107 ,   145 ,   821 ,   1115    ,   6067    ,   8301    ,   7565    ,0 };
            const unsigned long dim3139KuoInit[]        =   {   1   ,   3   ,   5   ,   1   ,   21  ,   29  ,   85  ,   9   ,   427 ,   553 ,   887 ,   2463    ,   4023    ,   9107    ,   25405   ,0 };
            const unsigned long dim3140KuoInit[]        =   {   1   ,   1   ,   5   ,   13  ,   25  ,   45  ,   89  ,   235 ,   349 ,   671 ,   1667    ,   2023    ,   2385    ,   1831    ,   515 ,0 };
            const unsigned long dim3141KuoInit[]        =   {   1   ,   1   ,   1   ,   11  ,   17  ,   37  ,   95  ,   77  ,   445 ,   169 ,   829 ,   9   ,   2589    ,   12809   ,   32469   ,0 };
            const unsigned long dim3142KuoInit[]        =   {   1   ,   1   ,   5   ,   5   ,   23  ,   31  ,   117 ,   81  ,   275 ,   861 ,   1439    ,   4041    ,   3337    ,   14945   ,   15921   ,0 };
            const unsigned long dim3143KuoInit[]        =   {   1   ,   1   ,   3   ,   13  ,   9   ,   21  ,   121 ,   23  ,   133 ,   563 ,   1889    ,   1939    ,   4427    ,   5501    ,   4609    ,0 };
            const unsigned long dim3144KuoInit[]        =   {   1   ,   1   ,   7   ,   7   ,   17  ,   25  ,   7   ,   245 ,   319 ,   697 ,   1561    ,   3209    ,   2369    ,   16045   ,   31799   ,0 };
            const unsigned long dim3145KuoInit[]        =   {   1   ,   1   ,   7   ,   5   ,   13  ,   19  ,   83  ,   165 ,   457 ,   1005    ,   743 ,   4049    ,   2563    ,   14487   ,   23699   ,0 };
            const unsigned long dim3146KuoInit[]        =   {   1   ,   1   ,   7   ,   1   ,   5   ,   25  ,   65  ,   185 ,   27  ,   235 ,   1771    ,   3505    ,   6807    ,   14715   ,   19841   ,0 };
            const unsigned long dim3147KuoInit[]        =   {   1   ,   1   ,   7   ,   11  ,   15  ,   15  ,   19  ,   111 ,   99  ,   397 ,   315 ,   2587    ,   7335    ,   4371    ,   11753   ,0 };
            const unsigned long dim3148KuoInit[]        =   {   1   ,   1   ,   7   ,   9   ,   1   ,   21  ,   55  ,   255 ,   105 ,   865 ,   179 ,   2127    ,   4485    ,   1803    ,   19503   ,0 };
            const unsigned long dim3149KuoInit[]        =   {   1   ,   3   ,   7   ,   13  ,   27  ,   23  ,   9   ,   9   ,   323 ,   421 ,   671 ,   3323    ,   5263    ,   2963    ,   16065   ,0 };
            const unsigned long dim3150KuoInit[]        =   {   1   ,   1   ,   5   ,   7   ,   17  ,   47  ,   79  ,   255 ,   259 ,   913 ,   1691    ,   1553    ,   2443    ,   14167   ,   24499   ,0 };
            const unsigned long dim3151KuoInit[]        =   {   1   ,   3   ,   3   ,   3   ,   21  ,   7   ,   17  ,   71  ,   57  ,   715 ,   2023    ,   2133    ,   5439    ,   12483   ,   32547   ,0 };
            const unsigned long dim3152KuoInit[]        =   {   1   ,   3   ,   3   ,   15  ,   1   ,   37  ,   23  ,   235 ,   35  ,   739 ,   1767    ,   3531    ,   597 ,   13259   ,   7869    ,0 };
            const unsigned long dim3153KuoInit[]        =   {   1   ,   3   ,   3   ,   9   ,   7   ,   5   ,   29  ,   129 ,   223 ,   991 ,   31  ,   375 ,   1917    ,   1157    ,   9723    ,0 };
            const unsigned long dim3154KuoInit[]        =   {   1   ,   3   ,   5   ,   5   ,   29  ,   33  ,   33  ,   73  ,   339 ,   21  ,   2037    ,   1127    ,   6385    ,   16335   ,   25825   ,0 };
            const unsigned long dim3155KuoInit[]        =   {   1   ,   3   ,   7   ,   7   ,   7   ,   49  ,   103 ,   235 ,   503 ,   687 ,   277 ,   2639    ,   5997    ,   11797   ,   27141   ,0 };
            const unsigned long dim3156KuoInit[]        =   {   1   ,   1   ,   1   ,   13  ,   27  ,   31  ,   103 ,   145 ,   245 ,   493 ,   867 ,   2075    ,   5369    ,   12551   ,   4907    ,0 };
            const unsigned long dim3157KuoInit[]        =   {   1   ,   3   ,   1   ,   5   ,   19  ,   57  ,   101 ,   165 ,   61  ,   201 ,   1299    ,   3081    ,   8173    ,   12835   ,   7915    ,0 };
            const unsigned long dim3158KuoInit[]        =   {   1   ,   3   ,   5   ,   3   ,   29  ,   23  ,   49  ,   41  ,   473 ,   273 ,   1493    ,   1915    ,   6669    ,   14263   ,   13385   ,0 };
            const unsigned long dim3159KuoInit[]        =   {   1   ,   1   ,   5   ,   9   ,   21  ,   27  ,   49  ,   73  ,   281 ,   1017    ,   1905    ,   1005    ,   631 ,   5963    ,   5729    ,0 };
            const unsigned long dim3160KuoInit[]        =   {   1   ,   3   ,   3   ,   5   ,   9   ,   25  ,   1   ,   155 ,   275 ,   675 ,   831 ,   109 ,   5349    ,   9607    ,   22431   ,0 };
            const unsigned long dim3161KuoInit[]        =   {   1   ,   1   ,   1   ,   3   ,   11  ,   33  ,   43  ,   63  ,   293 ,   425 ,   1525    ,   4091    ,   1865    ,   13187   ,   25191   ,0 };
            const unsigned long dim3162KuoInit[]        =   {   1   ,   1   ,   1   ,   7   ,   27  ,   25  ,   89  ,   39  ,   11  ,   961 ,   1601    ,   785 ,   3317    ,   13121   ,   13647   ,0 };
            const unsigned long dim3163KuoInit[]        =   {   1   ,   1   ,   7   ,   1   ,   31  ,   59  ,   25  ,   75  ,   317 ,   305 ,   1585    ,   3649    ,   7099    ,   1819    ,   10977   ,0 };
            const unsigned long dim3164KuoInit[]        =   {   1   ,   3   ,   7   ,   5   ,   7   ,   29  ,   85  ,   189 ,   137 ,   783 ,   395 ,   1887    ,   6253    ,   4043    ,   32279   ,0 };
            const unsigned long dim3165KuoInit[]        =   {   1   ,   1   ,   3   ,   7   ,   15  ,   47  ,   53  ,   125 ,   325 ,   579 ,   1783    ,   3393    ,   485 ,   13923   ,   4947    ,0 };
            const unsigned long dim3166KuoInit[]        =   {   1   ,   3   ,   5   ,   1   ,   19  ,   53  ,   67  ,   47  ,   503 ,   875 ,   1717    ,   857 ,   5151    ,   12279   ,   21249   ,0 };
            const unsigned long dim3167KuoInit[]        =   {   1   ,   1   ,   5   ,   11  ,   1   ,   53  ,   49  ,   211 ,   139 ,   955 ,   1129    ,   1967    ,   6833    ,   2923    ,   13251   ,0 };
            const unsigned long dim3168KuoInit[]        =   {   1   ,   3   ,   3   ,   9   ,   15  ,   53  ,   103 ,   71  ,   51  ,   379 ,   1565    ,   2017    ,   6013    ,   13147   ,   1227    ,0 };
            const unsigned long dim3169KuoInit[]        =   {   1   ,   3   ,   3   ,   1   ,   17  ,   25  ,   105 ,   177 ,   275 ,   969 ,   1373    ,   2471    ,   6959    ,   11107   ,   23657   ,0 };
            const unsigned long dim3170KuoInit[]        =   {   1   ,   3   ,   1   ,   13  ,   15  ,   3   ,   19  ,   145 ,   309 ,   1021    ,   985 ,   2385    ,   5727    ,   10357   ,   4915    ,0 };
            const unsigned long dim3171KuoInit[]        =   {   1   ,   3   ,   1   ,   9   ,   27  ,   33  ,   63  ,   115 ,   369 ,   613 ,   277 ,   751 ,   1959    ,   5709    ,   1851    ,0 };
            const unsigned long dim3172KuoInit[]        =   {   1   ,   1   ,   7   ,   7   ,   11  ,   31  ,   95  ,   131 ,   383 ,   821 ,   725 ,   1381    ,   3953    ,   14945   ,   29101   ,0 };
            const unsigned long dim3173KuoInit[]        =   {   1   ,   1   ,   1   ,   11  ,   1   ,   59  ,   81  ,   89  ,   325 ,   27  ,   1865    ,   987 ,   3877    ,   12509   ,   4129    ,0 };
            const unsigned long dim3174KuoInit[]        =   {   1   ,   3   ,   5   ,   7   ,   3   ,   33  ,   63  ,   141 ,   429 ,   847 ,   1355    ,   1451    ,   4741    ,   5427    ,   16907   ,0 };
            const unsigned long dim3175KuoInit[]        =   {   1   ,   3   ,   5   ,   9   ,   25  ,   45  ,   27  ,   249 ,   453 ,   673 ,   1519    ,   905 ,   5235    ,   10057   ,   24965   ,0 };
            const unsigned long dim3176KuoInit[]        =   {   1   ,   1   ,   5   ,   3   ,   27  ,   43  ,   91  ,   61  ,   121 ,   485 ,   1047    ,   405 ,   8159    ,   8363    ,   19695   ,0 };
            const unsigned long dim3177KuoInit[]        =   {   1   ,   1   ,   3   ,   9   ,   17  ,   5   ,   107 ,   139 ,   293 ,   685 ,   521 ,   253 ,   3309    ,   15335   ,   4967    ,0 };
            const unsigned long dim3178KuoInit[]        =   {   1   ,   1   ,   7   ,   13  ,   7   ,   43  ,   63  ,   119 ,   225 ,   269 ,   1489    ,   2983    ,   1041    ,   15149   ,   21039   ,0 };
            const unsigned long dim3179KuoInit[]        =   {   1   ,   3   ,   1   ,   5   ,   5   ,   1   ,   123 ,   255 ,   469 ,   139 ,   1861    ,   385 ,   3123    ,   14739   ,   23049   ,0 };
            const unsigned long dim3180KuoInit[]        =   {   1   ,   3   ,   5   ,   9   ,   25  ,   47  ,   107 ,   55  ,   435 ,   935 ,   1751    ,   555 ,   973 ,   10123   ,   31869   ,0 };
            const unsigned long dim3181KuoInit[]        =   {   1   ,   3   ,   7   ,   5   ,   21  ,   25  ,   113 ,   169 ,   313 ,   607 ,   631 ,   3409    ,   6539    ,   5781    ,   12681   ,0 };
            const unsigned long dim3182KuoInit[]        =   {   1   ,   1   ,   3   ,   15  ,   7   ,   13  ,   1   ,   245 ,   57  ,   495 ,   905 ,   1627    ,   7107    ,   13939   ,   22091   ,0 };
            const unsigned long dim3183KuoInit[]        =   {   1   ,   1   ,   1   ,   5   ,   7   ,   55  ,   121 ,   137 ,   93  ,   265 ,   1263    ,   2251    ,   7101    ,   3095    ,   13779   ,0 };
            const unsigned long dim3184KuoInit[]        =   {   1   ,   1   ,   5   ,   13  ,   21  ,   11  ,   37  ,   193 ,   481 ,   551 ,   1659    ,   563 ,   4405    ,   9541    ,   8495    ,0 };
            const unsigned long dim3185KuoInit[]        =   {   1   ,   1   ,   3   ,   15  ,   5   ,   7   ,   23  ,   19  ,   147 ,   779 ,   1089    ,   723 ,   3837    ,   16291   ,   23911   ,0 };
            const unsigned long dim3186KuoInit[]        =   {   1   ,   1   ,   1   ,   11  ,   27  ,   17  ,   57  ,   191 ,   337 ,   361 ,   1261    ,   3605    ,   8051    ,   8909    ,   14951   ,0 };
            const unsigned long dim3187KuoInit[]        =   {   1   ,   3   ,   3   ,   9   ,   25  ,   23  ,   39  ,   109 ,   107 ,   605 ,   373 ,   649 ,   5747    ,   2037    ,   24933   ,0 };
            const unsigned long dim3188KuoInit[]        =   {   1   ,   3   ,   5   ,   7   ,   9   ,   27  ,   91  ,   53  ,   153 ,   497 ,   941 ,   2999    ,   3721    ,   11115   ,   17589   ,0 };
            const unsigned long dim3189KuoInit[]        =   {   1   ,   3   ,   1   ,   7   ,   31  ,   7   ,   101 ,   115 ,   419 ,   479 ,   1097    ,   2253    ,   4365    ,   14285   ,   26469   ,0 };
            const unsigned long dim3190KuoInit[]        =   {   1   ,   1   ,   1   ,   15  ,   13  ,   37  ,   31  ,   231 ,   195 ,   539 ,   941 ,   3723    ,   2483    ,   2603    ,   18897   ,0 };
            const unsigned long dim3191KuoInit[]        =   {   1   ,   3   ,   5   ,   13  ,   7   ,   31  ,   19  ,   99  ,   197 ,   833 ,   79  ,   3899    ,   7045    ,   6913    ,   17959   ,0 };
            const unsigned long dim3192KuoInit[]        =   {   1   ,   1   ,   3   ,   5   ,   1   ,   51  ,   83  ,   135 ,   307 ,   121 ,   187 ,   2135    ,   2927    ,   6183    ,   20327   ,0 };
            const unsigned long dim3193KuoInit[]        =   {   1   ,   3   ,   5   ,   11  ,   7   ,   31  ,   47  ,   123 ,   409 ,   239 ,   1787    ,   3063    ,   6685    ,   7487    ,   31243   ,0 };
            const unsigned long dim3194KuoInit[]        =   {   1   ,   3   ,   7   ,   9   ,   9   ,   31  ,   57  ,   5   ,   493 ,   771 ,   1941    ,   199 ,   6389    ,   681 ,   19379   ,0 };
            const unsigned long dim3195KuoInit[]        =   {   1   ,   1   ,   5   ,   3   ,   1   ,   57  ,   63  ,   35  ,   369 ,   267 ,   923 ,   2019    ,   6487    ,   15555   ,   6993    ,0 };
            const unsigned long dim3196KuoInit[]        =   {   1   ,   3   ,   7   ,   15  ,   15  ,   59  ,   123 ,   179 ,   5   ,   35  ,   1061    ,   125 ,   1025    ,   5907    ,   29671   ,0 };
            const unsigned long dim3197KuoInit[]        =   {   1   ,   3   ,   1   ,   11  ,   21  ,   63  ,   71  ,   237 ,   333 ,   893 ,   1139    ,   1969    ,   2007    ,   10017   ,   27597   ,0 };
            const unsigned long dim3198KuoInit[]        =   {   1   ,   1   ,   1   ,   13  ,   15  ,   29  ,   103 ,   241 ,   25  ,   135 ,   1147    ,   1263    ,   4793    ,   16109   ,   22447   ,0 };
            const unsigned long dim3199KuoInit[]        =   {   1   ,   1   ,   3   ,   7   ,   11  ,   21  ,   85  ,   7   ,   123 ,   69  ,   1379    ,   3243    ,   7907    ,   15859   ,   19993   ,0 };
            const unsigned long dim3200KuoInit[]        =   {   1   ,   3   ,   5   ,   3   ,   17  ,   49  ,   125 ,   211 ,   165 ,   307 ,   1331    ,   2287    ,   1127    ,   4031    ,   13719   ,0 };
            const unsigned long dim3201KuoInit[]        =   {   1   ,   1   ,   7   ,   1   ,   23  ,   23  ,   21  ,   159 ,   387 ,   443 ,   291 ,   2843    ,   4413    ,   9035    ,   28887   ,0 };
            const unsigned long dim3202KuoInit[]        =   {   1   ,   1   ,   7   ,   15  ,   9   ,   25  ,   113 ,   185 ,   285 ,   211 ,   317 ,   3721    ,   6921    ,   4921    ,   2353    ,0 };
            const unsigned long dim3203KuoInit[]        =   {   1   ,   1   ,   3   ,   5   ,   21  ,   9   ,   53  ,   21  ,   337 ,   51  ,   429 ,   4035    ,   4697    ,   3695    ,   22589   ,0 };
            const unsigned long dim3204KuoInit[]        =   {   1   ,   1   ,   5   ,   5   ,   21  ,   63  ,   79  ,   213 ,   471 ,   183 ,   299 ,   409 ,   6125    ,   3653    ,   9785    ,0 };
            const unsigned long dim3205KuoInit[]        =   {   1   ,   3   ,   3   ,   3   ,   27  ,   27  ,   9   ,   73  ,   129 ,   391 ,   1877    ,   4023    ,   1631    ,   7101    ,   1491    ,0 };
            const unsigned long dim3206KuoInit[]        =   {   1   ,   3   ,   5   ,   3   ,   21  ,   7   ,   93  ,   51  ,   469 ,   3   ,   1343    ,   3827    ,   6623    ,   10149   ,   29233   ,0 };
            const unsigned long dim3207KuoInit[]        =   {   1   ,   1   ,   1   ,   9   ,   31  ,   25  ,   35  ,   249 ,   469 ,   393 ,   1315    ,   1247    ,   275 ,   15841   ,   23877   ,0 };
            const unsigned long dim3208KuoInit[]        =   {   1   ,   3   ,   5   ,   11  ,   25  ,   47  ,   105 ,   45  ,   79  ,   131 ,   499 ,   1411    ,   2077    ,   9809    ,   14917   ,0 };
            const unsigned long dim3209KuoInit[]        =   {   1   ,   3   ,   3   ,   3   ,   13  ,   15  ,   117 ,   27  ,   499 ,   885 ,   301 ,   935 ,   3443    ,   3845    ,   26873   ,0 };
            const unsigned long dim3210KuoInit[]        =   {   1   ,   3   ,   3   ,   1   ,   1   ,   47  ,   21  ,   97  ,   245 ,   95  ,   899 ,   993 ,   2493    ,   5249    ,   32695   ,0 };
            const unsigned long dim3211KuoInit[]        =   {   1   ,   3   ,   3   ,   11  ,   21  ,   39  ,   53  ,   15  ,   53  ,   687 ,   1269    ,   1673    ,   2015    ,   3205    ,   21447   ,0 };
            const unsigned long dim3212KuoInit[]        =   {   1   ,   1   ,   5   ,   11  ,   1   ,   39  ,   117 ,   85  ,   277 ,   635 ,   1459    ,   3977    ,   2449    ,   13341   ,   30607   ,0 };
            const unsigned long dim3213KuoInit[]        =   {   1   ,   1   ,   1   ,   13  ,   27  ,   45  ,   7   ,   255 ,   281 ,   807 ,   533 ,   2173    ,   4561    ,   2501    ,   5767    ,0 };
            const unsigned long dim3214KuoInit[]        =   {   1   ,   3   ,   5   ,   1   ,   21  ,   17  ,   65  ,   123 ,   239 ,   431 ,   1003    ,   241 ,   8115    ,   11633   ,   27709   ,0 };
            const unsigned long dim3215KuoInit[]        =   {   1   ,   3   ,   1   ,   1   ,   23  ,   39  ,   81  ,   35  ,   33  ,   905 ,   2019    ,   3327    ,   1197    ,   6891    ,   5521    ,0 };
            const unsigned long dim3216KuoInit[]        =   {   1   ,   3   ,   3   ,   11  ,   1   ,   51  ,   51  ,   153 ,   23  ,   685 ,   1663    ,   3167    ,   5803    ,   11111   ,   3371    ,0 };
            const unsigned long dim3217KuoInit[]        =   {   1   ,   3   ,   3   ,   11  ,   31  ,   29  ,   7   ,   173 ,   27  ,   129 ,   341 ,   2971    ,   4907    ,   13481   ,   19619   ,0 };
            const unsigned long dim3218KuoInit[]        =   {   1   ,   3   ,   3   ,   9   ,   29  ,   29  ,   115 ,   3   ,   45  ,   765 ,   385 ,   3775    ,   5349    ,   8809    ,   11019   ,0 };
            const unsigned long dim3219KuoInit[]        =   {   1   ,   1   ,   5   ,   7   ,   15  ,   19  ,   21  ,   29  ,   377 ,   915 ,   1835    ,   517 ,   3763    ,   5439    ,   7649    ,0 };
            const unsigned long dim3220KuoInit[]        =   {   1   ,   3   ,   7   ,   13  ,   25  ,   27  ,   33  ,   225 ,   135 ,   869 ,   1383    ,   3817    ,   2981    ,   8185    ,   5689    ,0 };
            const unsigned long dim3221KuoInit[]        =   {   1   ,   1   ,   5   ,   9   ,   1   ,   27  ,   19  ,   55  ,   39  ,   451 ,   365 ,   2905    ,   4311    ,   16005   ,   4537    ,0 };
            const unsigned long dim3222KuoInit[]        =   {   1   ,   1   ,   7   ,   15  ,   19  ,   1   ,   99  ,   177 ,   163 ,   877 ,   511 ,   1977    ,   3291    ,   7739    ,   6383    ,0 };
            const unsigned long dim3223KuoInit[]        =   {   1   ,   1   ,   7   ,   15  ,   1   ,   29  ,   27  ,   1   ,   411 ,   425 ,   2027    ,   1739    ,   5137    ,   107 ,   3515    ,0 };
            const unsigned long dim3224KuoInit[]        =   {   1   ,   3   ,   7   ,   5   ,   5   ,   29  ,   15  ,   55  ,   423 ,   671 ,   1145    ,   2403    ,   5429    ,   11507   ,   24747   ,0 };
            const unsigned long dim3225KuoInit[]        =   {   1   ,   1   ,   1   ,   13  ,   9   ,   61  ,   93  ,   215 ,   447 ,   969 ,   853 ,   487 ,   5735    ,   2153    ,   27857   ,0 };
            const unsigned long dim3226KuoInit[]        =   {   1   ,   3   ,   1   ,   3   ,   29  ,   41  ,   47  ,   245 ,   197 ,   261 ,   77  ,   805 ,   965 ,   14089   ,   29199   ,0 };
            const unsigned long dim3227KuoInit[]        =   {   1   ,   1   ,   5   ,   11  ,   3   ,   49  ,   53  ,   7   ,   91  ,   123 ,   413 ,   123 ,   7611    ,   11493   ,   16417   ,0 };
            const unsigned long dim3228KuoInit[]        =   {   1   ,   3   ,   7   ,   9   ,   23  ,   59  ,   105 ,   191 ,   13  ,   421 ,   1541    ,   2045    ,   2821    ,   9769    ,   19893   ,0 };
            const unsigned long dim3229KuoInit[]        =   {   1   ,   3   ,   5   ,   13  ,   3   ,   19  ,   37  ,   229 ,   465 ,   1015    ,   1871    ,   1101    ,   7515    ,   8547    ,   7853    ,0 };
            const unsigned long dim3230KuoInit[]        =   {   1   ,   3   ,   5   ,   13  ,   9   ,   61  ,   113 ,   209 ,   503 ,   441 ,   905 ,   2693    ,   1827    ,   9551    ,   26081   ,0 };
            const unsigned long dim3231KuoInit[]        =   {   1   ,   1   ,   1   ,   7   ,   31  ,   27  ,   123 ,   37  ,   415 ,   729 ,   921 ,   453 ,   673 ,   14393   ,   9139    ,0 };
            const unsigned long dim3232KuoInit[]        =   {   1   ,   1   ,   5   ,   3   ,   13  ,   9   ,   79  ,   243 ,   211 ,   89  ,   295 ,   3663    ,   8165    ,   12099   ,   1413    ,0 };
            const unsigned long dim3233KuoInit[]        =   {   1   ,   1   ,   5   ,   7   ,   23  ,   33  ,   33  ,   129 ,   51  ,   965 ,   1145    ,   713 ,   8143    ,   10045   ,   5291    ,0 };
            const unsigned long dim3234KuoInit[]        =   {   1   ,   1   ,   3   ,   3   ,   11  ,   45  ,   77  ,   223 ,   423 ,   1021    ,   343 ,   3103    ,   119 ,   3371    ,   24053   ,0 };
            const unsigned long dim3235KuoInit[]        =   {   1   ,   1   ,   7   ,   7   ,   7   ,   3   ,   65  ,   31  ,   433 ,   277 ,   1541    ,   2001    ,   121 ,   13543   ,   6723    ,0 };
            const unsigned long dim3236KuoInit[]        =   {   1   ,   1   ,   7   ,   1   ,   1   ,   15  ,   83  ,   85  ,   351 ,   351 ,   1063    ,   3989    ,   5739    ,   11939   ,   28009   ,0 };
            const unsigned long dim3237KuoInit[]        =   {   1   ,   1   ,   5   ,   11  ,   25  ,   57  ,   73  ,   127 ,   25  ,   67  ,   1789    ,   199 ,   6373    ,   10945   ,   17839   ,0 };
            const unsigned long dim3238KuoInit[]        =   {   1   ,   3   ,   5   ,   11  ,   17  ,   15  ,   5   ,   191 ,   209 ,   455 ,   89  ,   2611    ,   7557    ,   16167   ,   25993   ,0 };
            const unsigned long dim3239KuoInit[]        =   {   1   ,   1   ,   7   ,   7   ,   29  ,   21  ,   79  ,   81  ,   257 ,   373 ,   557 ,   2363    ,   5497    ,   8343    ,   2011    ,0 };
            const unsigned long dim3240KuoInit[]        =   {   1   ,   1   ,   5   ,   11  ,   25  ,   63  ,   83  ,   31  ,   477 ,   201 ,   2013    ,   3349    ,   3517    ,   9527    ,   14883   ,0 };
            const unsigned long dim3241KuoInit[]        =   {   1   ,   3   ,   7   ,   13  ,   21  ,   13  ,   41  ,   151 ,   359 ,   829 ,   661 ,   1959    ,   6109    ,   13757   ,   17677   ,0 };
            const unsigned long dim3242KuoInit[]        =   {   1   ,   3   ,   7   ,   1   ,   11  ,   15  ,   119 ,   25  ,   351 ,   135 ,   515 ,   3727    ,   5701    ,   5283    ,   18771   ,0 };
            const unsigned long dim3243KuoInit[]        =   {   1   ,   1   ,   1   ,   3   ,   3   ,   59  ,   55  ,   117 ,   269 ,   161 ,   645 ,   3459    ,   8173    ,   7193    ,   10275   ,0 };
            const unsigned long dim3244KuoInit[]        =   {   1   ,   3   ,   5   ,   7   ,   21  ,   25  ,   53  ,   211 ,   237 ,   775 ,   43  ,   3613    ,   5555    ,   1689    ,   1781    ,0 };
            const unsigned long dim3245KuoInit[]        =   {   1   ,   1   ,   1   ,   7   ,   31  ,   43  ,   71  ,   201 ,   371 ,   457 ,   1757    ,   2529    ,   7067    ,   2987    ,   3945    ,0 };
            const unsigned long dim3246KuoInit[]        =   {   1   ,   1   ,   5   ,   7   ,   15  ,   47  ,   29  ,   83  ,   259 ,   813 ,   881 ,   1901    ,   6635    ,   8515    ,   20385   ,0 };
            const unsigned long dim3247KuoInit[]        =   {   1   ,   1   ,   5   ,   5   ,   3   ,   45  ,   55  ,   43  ,   17  ,   179 ,   1801    ,   49  ,   8121    ,   9333    ,   12701   ,0 };
            const unsigned long dim3248KuoInit[]        =   {   1   ,   1   ,   3   ,   9   ,   21  ,   9   ,   55  ,   59  ,   275 ,   775 ,   1567    ,   1557    ,   3465    ,   5727    ,   12881   ,0 };
            const unsigned long dim3249KuoInit[]        =   {   1   ,   1   ,   1   ,   5   ,   21  ,   63  ,   81  ,   27  ,   139 ,   859 ,   1211    ,   3787    ,   1179    ,   4045    ,   26225   ,0 };
            const unsigned long dim3250KuoInit[]        =   {   1   ,   3   ,   7   ,   3   ,   7   ,   5   ,   99  ,   125 ,   499 ,   437 ,   1141    ,   3557    ,   5349    ,   2701    ,   25741   ,0 };
            const unsigned long dim3251KuoInit[]        =   {   1   ,   1   ,   5   ,   1   ,   7   ,   59  ,   65  ,   117 ,   201 ,   351 ,   1063    ,   3083    ,   6637    ,   5969    ,   23037   ,0 };
            const unsigned long dim3252KuoInit[]        =   {   1   ,   3   ,   7   ,   7   ,   25  ,   45  ,   115 ,   31  ,   351 ,   203 ,   1791    ,   2931    ,   7127    ,   12117   ,   10781   ,0 };
            const unsigned long dim3253KuoInit[]        =   {   1   ,   1   ,   3   ,   3   ,   29  ,   59  ,   85  ,   101 ,   147 ,   219 ,   1293    ,   1645    ,   7653    ,   649 ,   32221   ,0 };
            const unsigned long dim3254KuoInit[]        =   {   1   ,   1   ,   7   ,   5   ,   29  ,   51  ,   127 ,   139 ,   205 ,   77  ,   859 ,   2277    ,   2333    ,   105 ,   9053    ,0 };
            const unsigned long dim3255KuoInit[]        =   {   1   ,   3   ,   5   ,   3   ,   29  ,   23  ,   119 ,   71  ,   263 ,   551 ,   1095    ,   359 ,   6223    ,   7665    ,   1095    ,0 };
            const unsigned long dim3256KuoInit[]        =   {   1   ,   3   ,   1   ,   11  ,   13  ,   47  ,   11  ,   207 ,   507 ,   139 ,   1383    ,   1339    ,   4111    ,   11685   ,   12607   ,0 };
            const unsigned long dim3257KuoInit[]        =   {   1   ,   3   ,   7   ,   3   ,   25  ,   9   ,   87  ,   117 ,   21  ,   289 ,   351 ,   225 ,   4045    ,   5275    ,   531 ,0 };
            const unsigned long dim3258KuoInit[]        =   {   1   ,   3   ,   7   ,   3   ,   29  ,   55  ,   31  ,   57  ,   83  ,   423 ,   201 ,   4055    ,   1977    ,   8891    ,   17961   ,0 };
            const unsigned long dim3259KuoInit[]        =   {   1   ,   1   ,   1   ,   3   ,   9   ,   7   ,   45  ,   237 ,   413 ,   53  ,   1063    ,   23  ,   6615    ,   16359   ,   4743    ,0 };
            const unsigned long dim3260KuoInit[]        =   {   1   ,   3   ,   1   ,   3   ,   21  ,   57  ,   63  ,   221 ,   111 ,   95  ,   647 ,   3419    ,   1919    ,   923 ,   8357    ,0 };
            const unsigned long dim3261KuoInit[]        =   {   1   ,   1   ,   1   ,   3   ,   27  ,   25  ,   119 ,   237 ,   397 ,   73  ,   41  ,   17  ,   6839    ,   15695   ,   30365   ,0 };
            const unsigned long dim3262KuoInit[]        =   {   1   ,   3   ,   1   ,   1   ,   23  ,   59  ,   107 ,   121 ,   267 ,   163 ,   181 ,   2289    ,   167 ,   7077    ,   12895   ,0 };
            const unsigned long dim3263KuoInit[]        =   {   1   ,   1   ,   5   ,   5   ,   15  ,   45  ,   45  ,   179 ,   195 ,   95  ,   43  ,   377 ,   629 ,   11983   ,   23439   ,0 };
            const unsigned long dim3264KuoInit[]        =   {   1   ,   3   ,   3   ,   9   ,   17  ,   43  ,   67  ,   143 ,   223 ,   243 ,   209 ,   3   ,   7651    ,   12607   ,   3927    ,0 };
            const unsigned long dim3265KuoInit[]        =   {   1   ,   3   ,   1   ,   11  ,   19  ,   21  ,   59  ,   75  ,   353 ,   509 ,   1215    ,   3223    ,   2373    ,   3873    ,   20391   ,0 };
            const unsigned long dim3266KuoInit[]        =   {   1   ,   1   ,   7   ,   3   ,   7   ,   9   ,   53  ,   89  ,   407 ,   987 ,   1165    ,   2473    ,   4295    ,   2531    ,   5615    ,0 };
            const unsigned long dim3267KuoInit[]        =   {   1   ,   1   ,   3   ,   11  ,   23  ,   7   ,   111 ,   147 ,   49  ,   627 ,   1769    ,   3497    ,   2807    ,   4743    ,   26491   ,0 };
            const unsigned long dim3268KuoInit[]        =   {   1   ,   3   ,   3   ,   13  ,   29  ,   29  ,   83  ,   57  ,   377 ,   581 ,   1699    ,   1689    ,   2297    ,   1377    ,   24551   ,0 };
            const unsigned long dim3269KuoInit[]        =   {   1   ,   3   ,   5   ,   15  ,   25  ,   33  ,   41  ,   187 ,   195 ,   647 ,   73  ,   619 ,   5585    ,   8379    ,   26079   ,0 };
            const unsigned long dim3270KuoInit[]        =   {   1   ,   1   ,   3   ,   9   ,   3   ,   47  ,   105 ,   83  ,   501 ,   843 ,   1867    ,   1855    ,   6347    ,   3269    ,   26155   ,0 };
            const unsigned long dim3271KuoInit[]        =   {   1   ,   3   ,   3   ,   15  ,   29  ,   33  ,   117 ,   141 ,   171 ,   801 ,   1489    ,   1   ,   4081    ,   15529   ,   4081    ,0 };
            const unsigned long dim3272KuoInit[]        =   {   1   ,   1   ,   3   ,   13  ,   21  ,   43  ,   13  ,   75  ,   75  ,   443 ,   1443    ,   3227    ,   907 ,   11821   ,   24209   ,0 };
            const unsigned long dim3273KuoInit[]        =   {   1   ,   1   ,   3   ,   3   ,   31  ,   31  ,   111 ,   193 ,   275 ,   545 ,   371 ,   3071    ,   3491    ,   6661    ,   1113    ,0 };
            const unsigned long dim3274KuoInit[]        =   {   1   ,   1   ,   7   ,   3   ,   11  ,   39  ,   17  ,   107 ,   303 ,   201 ,   321 ,   3083    ,   6839    ,   4381    ,   3449    ,0 };
            const unsigned long dim3275KuoInit[]        =   {   1   ,   1   ,   7   ,   15  ,   31  ,   21  ,   7   ,   143 ,   211 ,   369 ,   369 ,   323 ,   3811    ,   5131    ,   11425   ,0 };
            const unsigned long dim3276KuoInit[]        =   {   1   ,   1   ,   1   ,   3   ,   9   ,   27  ,   23  ,   21  ,   349 ,   607 ,   529 ,   1161    ,   5753    ,   6015    ,   20489   ,0 };
            const unsigned long dim3277KuoInit[]        =   {   1   ,   3   ,   7   ,   11  ,   23  ,   17  ,   103 ,   97  ,   227 ,   933 ,   2011    ,   1705    ,   5185    ,   299 ,   5673    ,0 };
            const unsigned long dim3278KuoInit[]        =   {   1   ,   3   ,   5   ,   9   ,   7   ,   45  ,   63  ,   43  ,   339 ,   299 ,   1617    ,   1393    ,   4507    ,   13513   ,   29039   ,0 };
            const unsigned long dim3279KuoInit[]        =   {   1   ,   1   ,   5   ,   5   ,   17  ,   11  ,   49  ,   159 ,   51  ,   449 ,   497 ,   2627    ,   1541    ,   15711   ,   16591   ,0 };
            const unsigned long dim3280KuoInit[]        =   {   1   ,   1   ,   7   ,   3   ,   23  ,   19  ,   53  ,   79  ,   133 ,   279 ,   327 ,   2847    ,   5825    ,   14429   ,   26491   ,0 };
            const unsigned long dim3281KuoInit[]        =   {   1   ,   3   ,   1   ,   3   ,   17  ,   3   ,   33  ,   55  ,   263 ,   663 ,   1253    ,   1509    ,   4487    ,   1811    ,   32123   ,0 };
            const unsigned long dim3282KuoInit[]        =   {   1   ,   3   ,   7   ,   9   ,   17  ,   7   ,   23  ,   87  ,   221 ,   183 ,   1577    ,   1373    ,   2717    ,   3323    ,   24573   ,0 };
            const unsigned long dim3283KuoInit[]        =   {   1   ,   3   ,   3   ,   9   ,   5   ,   51  ,   45  ,   11  ,   11  ,   135 ,   505 ,   221 ,   3847    ,   13487   ,   23843   ,0 };
            const unsigned long dim3284KuoInit[]        =   {   1   ,   1   ,   5   ,   13  ,   15  ,   61  ,   85  ,   53  ,   321 ,   433 ,   61  ,   3377    ,   3993    ,   15325   ,   8641    ,0 };
            const unsigned long dim3285KuoInit[]        =   {   1   ,   3   ,   7   ,   13  ,   27  ,   35  ,   57  ,   43  ,   509 ,   595 ,   2033    ,   77  ,   7533    ,   5027    ,   15567   ,0 };
            const unsigned long dim3286KuoInit[]        =   {   1   ,   1   ,   1   ,   1   ,   31  ,   51  ,   69  ,   157 ,   105 ,   797 ,   249 ,   2033    ,   2457    ,   10773   ,   31537   ,0 };
            const unsigned long dim3287KuoInit[]        =   {   1   ,   3   ,   3   ,   13  ,   23  ,   25  ,   75  ,   11  ,   145 ,   361 ,   819 ,   3689    ,   5751    ,   6395    ,   30685   ,0 };
            const unsigned long dim3288KuoInit[]        =   {   1   ,   1   ,   7   ,   11  ,   13  ,   25  ,   13  ,   175 ,   11  ,   685 ,   835 ,   2641    ,   2037    ,   2037    ,   2807    ,0 };
            const unsigned long dim3289KuoInit[]        =   {   1   ,   1   ,   7   ,   11  ,   21  ,   27  ,   115 ,   91  ,   21  ,   105 ,   1817    ,   2057    ,   2797    ,   3825    ,   23761   ,0 };
            const unsigned long dim3290KuoInit[]        =   {   1   ,   3   ,   3   ,   5   ,   5   ,   49  ,   51  ,   57  ,   425 ,   959 ,   473 ,   575 ,   6851    ,   3683    ,   27429   ,0 };
            const unsigned long dim3291KuoInit[]        =   {   1   ,   1   ,   7   ,   5   ,   29  ,   1   ,   65  ,   13  ,   447 ,   467 ,   1193    ,   4021    ,   275 ,   13453   ,   29749   ,0 };
            const unsigned long dim3292KuoInit[]        =   {   1   ,   3   ,   5   ,   11  ,   27  ,   53  ,   71  ,   63  ,   125 ,   545 ,   1673    ,   3917    ,   6625    ,   12805   ,   2905    ,0 };
            const unsigned long dim3293KuoInit[]        =   {   1   ,   3   ,   3   ,   1   ,   7   ,   1   ,   45  ,   65  ,   23  ,   237 ,   1573    ,   3445    ,   6307    ,   5015    ,   11263   ,0 };
            const unsigned long dim3294KuoInit[]        =   {   1   ,   1   ,   5   ,   1   ,   7   ,   29  ,   67  ,   35  ,   307 ,   675 ,   727 ,   2789    ,   2353    ,   13241   ,   16621   ,0 };
            const unsigned long dim3295KuoInit[]        =   {   1   ,   3   ,   1   ,   15  ,   13  ,   5   ,   115 ,   75  ,   21  ,   745 ,   2027    ,   265 ,   3583    ,   1365    ,   397 ,0 };
            const unsigned long dim3296KuoInit[]        =   {   1   ,   1   ,   5   ,   7   ,   9   ,   31  ,   115 ,   25  ,   441 ,   433 ,   1665    ,   2499    ,   5765    ,   4001    ,   22105   ,0 };
            const unsigned long dim3297KuoInit[]        =   {   1   ,   3   ,   7   ,   9   ,   23  ,   21  ,   71  ,   165 ,   353 ,   101 ,   1335    ,   3411    ,   1415    ,   12641   ,   21961   ,0 };
            const unsigned long dim3298KuoInit[]        =   {   1   ,   3   ,   7   ,   1   ,   3   ,   3   ,   21  ,   157 ,   85  ,   117 ,   469 ,   1573    ,   2475    ,   6471    ,   15975   ,0 };
            const unsigned long dim3299KuoInit[]        =   {   1   ,   1   ,   7   ,   9   ,   31  ,   35  ,   127 ,   221 ,   447 ,   697 ,   1393    ,   701 ,   7291    ,   4279    ,   2897    ,0 };
            const unsigned long dim3300KuoInit[]        =   {   1   ,   3   ,   5   ,   5   ,   21  ,   13  ,   15  ,   15  ,   395 ,   519 ,   1985    ,   2331    ,   5483    ,   13905   ,   31549   ,0 };
            const unsigned long dim3301KuoInit[]        =   {   1   ,   1   ,   7   ,   5   ,   1   ,   43  ,   57  ,   131 ,   7   ,   893 ,   1631    ,   1581    ,   6691    ,   6249    ,   4975    ,0 };
            const unsigned long dim3302KuoInit[]        =   {   1   ,   3   ,   3   ,   9   ,   29  ,   33  ,   51  ,   177 ,   423 ,   459 ,   1293    ,   2467    ,   3877    ,   15653   ,   2727    ,0 };
            const unsigned long dim3303KuoInit[]        =   {   1   ,   3   ,   5   ,   1   ,   5   ,   45  ,   11  ,   27  ,   311 ,   645 ,   283 ,   65  ,   5913    ,   12811   ,   20041   ,0 };
            const unsigned long dim3304KuoInit[]        =   {   1   ,   3   ,   7   ,   5   ,   19  ,   3   ,   39  ,   99  ,   269 ,   429 ,   1901    ,   2213    ,   6755    ,   1205    ,   16073   ,0 };
            const unsigned long dim3305KuoInit[]        =   {   1   ,   3   ,   5   ,   13  ,   7   ,   37  ,   109 ,   113 ,   395 ,   981 ,   1991    ,   845 ,   2473    ,   3801    ,   24405   ,0 };
            const unsigned long dim3306KuoInit[]        =   {   1   ,   1   ,   1   ,   15  ,   9   ,   31  ,   117 ,   213 ,   141 ,   867 ,   387 ,   1985    ,   1985    ,   12683   ,   20501   ,0 };
            const unsigned long dim3307KuoInit[]        =   {   1   ,   1   ,   1   ,   15  ,   11  ,   3   ,   25  ,   31  ,   375 ,   437 ,   645 ,   335 ,   207 ,   14915   ,   2577    ,0 };
            const unsigned long dim3308KuoInit[]        =   {   1   ,   1   ,   7   ,   7   ,   19  ,   61  ,   87  ,   107 ,   53  ,   777 ,   31  ,   3335    ,   157 ,   2927    ,   25905   ,0 };
            const unsigned long dim3309KuoInit[]        =   {   1   ,   1   ,   7   ,   9   ,   13  ,   13  ,   21  ,   151 ,   51  ,   501 ,   213 ,   2177    ,   2821    ,   3081    ,   5955    ,0 };
            const unsigned long dim3310KuoInit[]        =   {   1   ,   1   ,   5   ,   13  ,   19  ,   33  ,   5   ,   209 ,   283 ,   613 ,   281 ,   715 ,   297 ,   15033   ,   20489   ,0 };
            const unsigned long dim3311KuoInit[]        =   {   1   ,   3   ,   5   ,   5   ,   31  ,   7   ,   31  ,   49  ,   5   ,   41  ,   939 ,   2359    ,   3663    ,   10151   ,   20457   ,0 };
            const unsigned long dim3312KuoInit[]        =   {   1   ,   1   ,   7   ,   1   ,   7   ,   61  ,   69  ,   255 ,   17  ,   637 ,   863 ,   3351    ,   6033    ,   5005    ,   23107   ,0 };
            const unsigned long dim3313KuoInit[]        =   {   1   ,   1   ,   5   ,   3   ,   5   ,   61  ,   15  ,   87  ,   111 ,   383 ,   1577    ,   3515    ,   789 ,   4219    ,   2527    ,0 };
            const unsigned long dim3314KuoInit[]        =   {   1   ,   1   ,   5   ,   11  ,   19  ,   9   ,   37  ,   61  ,   329 ,   811 ,   1829    ,   3109    ,   397 ,   2065    ,   32001   ,0 };
            const unsigned long dim3315KuoInit[]        =   {   1   ,   3   ,   1   ,   9   ,   7   ,   13  ,   55  ,   7   ,   463 ,   307 ,   1551    ,   3161    ,   649 ,   14073   ,   20839   ,0 };
            const unsigned long dim3316KuoInit[]        =   {   1   ,   3   ,   1   ,   9   ,   1   ,   43  ,   71  ,   87  ,   149 ,   557 ,   857 ,   2739    ,   6269    ,   3297    ,   26067   ,0 };
            const unsigned long dim3317KuoInit[]        =   {   1   ,   3   ,   1   ,   5   ,   5   ,   29  ,   45  ,   159 ,   125 ,   389 ,   627 ,   983 ,   755 ,   15585   ,   12555   ,0 };
            const unsigned long dim3318KuoInit[]        =   {   1   ,   1   ,   3   ,   7   ,   9   ,   53  ,   79  ,   29  ,   423 ,   423 ,   1493    ,   61  ,   1631    ,   5157    ,   29845   ,0 };
            const unsigned long dim3319KuoInit[]        =   {   1   ,   3   ,   5   ,   9   ,   23  ,   27  ,   57  ,   235 ,   57  ,   833 ,   371 ,   1797    ,   4541    ,   12491   ,   23127   ,0 };
            const unsigned long dim3320KuoInit[]        =   {   1   ,   3   ,   3   ,   1   ,   17  ,   1   ,   79  ,   87  ,   247 ,   987 ,   651 ,   3167    ,   4423    ,   831 ,   16811   ,0 };
            const unsigned long dim3321KuoInit[]        =   {   1   ,   3   ,   3   ,   5   ,   15  ,   23  ,   51  ,   167 ,   421 ,   493 ,   1877    ,   1079    ,   4147    ,   13485   ,   28619   ,0 };
            const unsigned long dim3322KuoInit[]        =   {   1   ,   1   ,   7   ,   13  ,   19  ,   15  ,   69  ,   31  ,   225 ,   597 ,   1477    ,   2363    ,   6969    ,   13221   ,   2561    ,0 };
            const unsigned long dim3323KuoInit[]        =   {   1   ,   3   ,   1   ,   7   ,   19  ,   61  ,   75  ,   199 ,   103 ,   823 ,   375 ,   1301    ,   4349    ,   135 ,   5701    ,0 };
            const unsigned long dim3324KuoInit[]        =   {   1   ,   3   ,   7   ,   11  ,   21  ,   25  ,   85  ,   249 ,   91  ,   885 ,   469 ,   1809    ,   4351    ,   14393   ,   19607   ,0 };
            const unsigned long dim3325KuoInit[]        =   {   1   ,   3   ,   5   ,   1   ,   25  ,   19  ,   71  ,   111 ,   427 ,   523 ,   1691    ,   4085    ,   7587    ,   1949    ,   10471   ,0 };
            const unsigned long dim3326KuoInit[]        =   {   1   ,   3   ,   1   ,   11  ,   27  ,   59  ,   13  ,   117 ,   155 ,   369 ,   1767    ,   2951    ,   6109    ,   13867   ,   13887   ,0 };
            const unsigned long dim3327KuoInit[]        =   {   1   ,   3   ,   3   ,   3   ,   31  ,   53  ,   35  ,   101 ,   295 ,   653 ,   337 ,   2281    ,   7563    ,   8847    ,   15235   ,0 };
            const unsigned long dim3328KuoInit[]        =   {   1   ,   1   ,   7   ,   13  ,   3   ,   17  ,   63  ,   251 ,   293 ,   861 ,   1539    ,   2409    ,   663 ,   1637    ,   22309   ,0 };
            const unsigned long dim3329KuoInit[]        =   {   1   ,   3   ,   7   ,   9   ,   13  ,   27  ,   41  ,   81  ,   483 ,   901 ,   1963    ,   1855    ,   799 ,   10933   ,   14395   ,0 };
            const unsigned long dim3330KuoInit[]        =   {   1   ,   1   ,   3   ,   5   ,   19  ,   23  ,   97  ,   179 ,   93  ,   657 ,   1361    ,   4019    ,   7577    ,   355 ,   4881    ,0 };
            const unsigned long dim3331KuoInit[]        =   {   1   ,   3   ,   3   ,   15  ,   23  ,   25  ,   101 ,   141 ,   91  ,   37  ,   1165    ,   3509    ,   2445    ,   7675    ,   30803   ,0 };
            const unsigned long dim3332KuoInit[]        =   {   1   ,   1   ,   1   ,   5   ,   13  ,   53  ,   87  ,   197 ,   321 ,   297 ,   277 ,   2159    ,   5457    ,   6463    ,   1797    ,0 };
            const unsigned long dim3333KuoInit[]        =   {   1   ,   3   ,   3   ,   7   ,   3   ,   3   ,   117 ,   5   ,   95  ,   527 ,   755 ,   2865    ,   6359    ,   4859    ,   11395   ,0 };
            const unsigned long dim3334KuoInit[]        =   {   1   ,   3   ,   3   ,   13  ,   1   ,   49  ,   43  ,   247 ,   369 ,   351 ,   1277    ,   1639    ,   3133    ,   4193    ,   26645   ,0 };
            const unsigned long dim3335KuoInit[]        =   {   1   ,   1   ,   1   ,   9   ,   27  ,   21  ,   59  ,   191 ,   365 ,   95  ,   337 ,   3279    ,   4455    ,   8235    ,   12457   ,0 };
            const unsigned long dim3336KuoInit[]        =   {   1   ,   1   ,   7   ,   5   ,   9   ,   1   ,   7   ,   23  ,   117 ,   711 ,   1027    ,   613 ,   4729    ,   14415   ,   15705   ,0 };
            const unsigned long dim3337KuoInit[]        =   {   1   ,   3   ,   1   ,   3   ,   29  ,   13  ,   43  ,   3   ,   117 ,   763 ,   1787    ,   2109    ,   4949    ,   7579    ,   24083   ,0 };
            const unsigned long dim3338KuoInit[]        =   {   1   ,   1   ,   5   ,   11  ,   7   ,   61  ,   107 ,   113 ,   127 ,   495 ,   1167    ,   2871    ,   7187    ,   13235   ,   2037    ,0 };
            const unsigned long dim3339KuoInit[]        =   {   1   ,   1   ,   5   ,   5   ,   13  ,   33  ,   23  ,   69  ,   5   ,   137 ,   157 ,   1107    ,   1025    ,   1907    ,   2279    ,0 };
            const unsigned long dim3340KuoInit[]        =   {   1   ,   1   ,   1   ,   9   ,   11  ,   63  ,   27  ,   13  ,   295 ,   951 ,   2027    ,   3625    ,   1447    ,   8939    ,   3965    ,0 };
            const unsigned long dim3341KuoInit[]        =   {   1   ,   3   ,   3   ,   13  ,   29  ,   31  ,   67  ,   221 ,   413 ,   13  ,   1643    ,   2847    ,   4889    ,   3733    ,   25713   ,0 };
            const unsigned long dim3342KuoInit[]        =   {   1   ,   3   ,   5   ,   11  ,   11  ,   17  ,   37  ,   37  ,   159 ,   835 ,   1775    ,   2059    ,   2995    ,   6141    ,   12277   ,0 };
            const unsigned long dim3343KuoInit[]        =   {   1   ,   3   ,   3   ,   7   ,   11  ,   25  ,   107 ,   239 ,   361 ,   835 ,   1943    ,   1375    ,   7077    ,   13843   ,   5685    ,0 };
            const unsigned long dim3344KuoInit[]        =   {   1   ,   1   ,   3   ,   11  ,   23  ,   27  ,   3   ,   191 ,   339 ,   675 ,   1181    ,   2005    ,   245 ,   7305    ,   5831    ,0 };
            const unsigned long dim3345KuoInit[]        =   {   1   ,   3   ,   5   ,   9   ,   15  ,   17  ,   91  ,   15  ,   255 ,   201 ,   535 ,   3063    ,   4351    ,   12527   ,   15981   ,0 };
            const unsigned long dim3346KuoInit[]        =   {   1   ,   3   ,   5   ,   9   ,   7   ,   47  ,   111 ,   245 ,   223 ,   781 ,   367 ,   559 ,   1433    ,   9631    ,   7207    ,0 };
            const unsigned long dim3347KuoInit[]        =   {   1   ,   3   ,   7   ,   5   ,   23  ,   5   ,   111 ,   237 ,   371 ,   541 ,   663 ,   693 ,   6591    ,   13089   ,   11381   ,0 };
            const unsigned long dim3348KuoInit[]        =   {   1   ,   1   ,   3   ,   9   ,   25  ,   5   ,   99  ,   17  ,   401 ,   787 ,   971 ,   1035    ,   1143    ,   12041   ,   22455   ,0 };
            const unsigned long dim3349KuoInit[]        =   {   1   ,   1   ,   5   ,   1   ,   3   ,   5   ,   125 ,   193 ,   69  ,   939 ,   37  ,   2843    ,   3139    ,   7967    ,   15099   ,0 };
            const unsigned long dim3350KuoInit[]        =   {   1   ,   1   ,   5   ,   3   ,   11  ,   39  ,   35  ,   103 ,   203 ,   255 ,   727 ,   2513    ,   1981    ,   12855   ,   20887   ,0 };
            const unsigned long dim3351KuoInit[]        =   {   1   ,   3   ,   3   ,   15  ,   27  ,   41  ,   41  ,   57  ,   275 ,   497 ,   727 ,   901 ,   5653    ,   12917   ,   20783   ,0 };
            const unsigned long dim3352KuoInit[]        =   {   1   ,   3   ,   1   ,   5   ,   17  ,   13  ,   101 ,   95  ,   465 ,   155 ,   973 ,   1911    ,   2719    ,   157 ,   19475   ,0 };
            const unsigned long dim3353KuoInit[]        =   {   1   ,   3   ,   3   ,   5   ,   3   ,   61  ,   31  ,   225 ,   351 ,   931 ,   855 ,   3871    ,   1835    ,   11133   ,   25203   ,0 };
            const unsigned long dim3354KuoInit[]        =   {   1   ,   3   ,   7   ,   7   ,   25  ,   1   ,   91  ,   193 ,   455 ,   873 ,   1255    ,   1657    ,   4143    ,   433 ,   22789   ,0 };
            const unsigned long dim3355KuoInit[]        =   {   1   ,   3   ,   7   ,   9   ,   31  ,   19  ,   35  ,   81  ,   175 ,   885 ,   1007    ,   3443    ,   2063    ,   3179    ,   28563   ,0 };
            const unsigned long dim3356KuoInit[]        =   {   1   ,   1   ,   3   ,   9   ,   1   ,   25  ,   109 ,   249 ,   341 ,   393 ,   497 ,   1181    ,   2943    ,   9021    ,   9985    ,0 };
            const unsigned long dim3357KuoInit[]        =   {   1   ,   1   ,   7   ,   13  ,   15  ,   49  ,   27  ,   145 ,   351 ,   119 ,   257 ,   151 ,   4287    ,   1709    ,   2585    ,0 };
            const unsigned long dim3358KuoInit[]        =   {   1   ,   1   ,   7   ,   7   ,   3   ,   19  ,   105 ,   51  ,   443 ,   693 ,   29  ,   2657    ,   6967    ,   4797    ,   14565   ,0 };
            const unsigned long dim3359KuoInit[]        =   {   1   ,   3   ,   3   ,   1   ,   25  ,   53  ,   65  ,   169 ,   461 ,   275 ,   1191    ,   731 ,   6295    ,   15083   ,   28445   ,0 };
            const unsigned long dim3360KuoInit[]        =   {   1   ,   3   ,   1   ,   1   ,   11  ,   39  ,   21  ,   3   ,   411 ,   45  ,   1325    ,   3279    ,   3153    ,   289 ,   30973   ,0 };
            const unsigned long dim3361KuoInit[]        =   {   1   ,   1   ,   1   ,   11  ,   29  ,   57  ,   107 ,   245 ,   279 ,   677 ,   599 ,   1145    ,   5135    ,   6131    ,   12667   ,0 };
            const unsigned long dim3362KuoInit[]        =   {   1   ,   3   ,   1   ,   15  ,   27  ,   7   ,   99  ,   215 ,   17  ,   171 ,   977 ,   2065    ,   1357    ,   8705    ,   29641   ,0 };
            const unsigned long dim3363KuoInit[]        =   {   1   ,   1   ,   3   ,   9   ,   31  ,   23  ,   53  ,   9   ,   177 ,   645 ,   99  ,   3175    ,   3543    ,   13637   ,   9923    ,0 };
            const unsigned long dim3364KuoInit[]        =   {   1   ,   1   ,   1   ,   13  ,   5   ,   45  ,   69  ,   39  ,   207 ,   651 ,   1507    ,   2969    ,   2743    ,   4025    ,   9983    ,0 };
            const unsigned long dim3365KuoInit[]        =   {   1   ,   1   ,   5   ,   13  ,   13  ,   57  ,   27  ,   215 ,   51  ,   927 ,   1973    ,   3127    ,   3655    ,   10653   ,   5649    ,0 };
            const unsigned long dim3366KuoInit[]        =   {   1   ,   3   ,   3   ,   7   ,   21  ,   7   ,   119 ,   163 ,   461 ,   859 ,   253 ,   2817    ,   8001    ,   393 ,   20171   ,0 };
            const unsigned long dim3367KuoInit[]        =   {   1   ,   3   ,   3   ,   3   ,   25  ,   7   ,   101 ,   109 ,   57  ,   185 ,   2005    ,   3167    ,   3539    ,   2465    ,   783 ,0 };
            const unsigned long dim3368KuoInit[]        =   {   1   ,   3   ,   1   ,   3   ,   31  ,   43  ,   69  ,   189 ,   471 ,   669 ,   339 ,   1125    ,   4273    ,   10417   ,   20129   ,0 };
            const unsigned long dim3369KuoInit[]        =   {   1   ,   1   ,   3   ,   7   ,   19  ,   37  ,   115 ,   111 ,   439 ,   19  ,   425 ,   995 ,   3983    ,   15523   ,   25679   ,0 };
            const unsigned long dim3370KuoInit[]        =   {   1   ,   3   ,   5   ,   9   ,   5   ,   1   ,   37  ,   105 ,   359 ,   423 ,   545 ,   1099    ,   6809    ,   8531    ,   25155   ,0 };
            const unsigned long dim3371KuoInit[]        =   {   1   ,   1   ,   3   ,   3   ,   9   ,   37  ,   37  ,   149 ,   457 ,   1017    ,   25  ,   3641    ,   2673    ,   1501    ,   28811   ,0 };
            const unsigned long dim3372KuoInit[]        =   {   1   ,   3   ,   5   ,   11  ,   25  ,   31  ,   57  ,   233 ,   125 ,   755 ,   1807    ,   337 ,   81  ,   14709   ,   4061    ,0 };
            const unsigned long dim3373KuoInit[]        =   {   1   ,   1   ,   1   ,   7   ,   25  ,   7   ,   37  ,   35  ,   437 ,   889 ,   965 ,   3997    ,   4493    ,   1357    ,   26383   ,0 };
            const unsigned long dim3374KuoInit[]        =   {   1   ,   1   ,   1   ,   3   ,   9   ,   37  ,   105 ,   127 ,   283 ,   141 ,   1451    ,   2275    ,   3261    ,   8135    ,   429 ,0 };
            const unsigned long dim3375KuoInit[]        =   {   1   ,   1   ,   1   ,   5   ,   3   ,   19  ,   9   ,   233 ,   293 ,   827 ,   103 ,   1963    ,   2877    ,   7701    ,   17431   ,0 };
            const unsigned long dim3376KuoInit[]        =   {   1   ,   3   ,   7   ,   11  ,   15  ,   17  ,   87  ,   243 ,   141 ,   721 ,   829 ,   3569    ,   4081    ,   6039    ,   21523   ,0 };
            const unsigned long dim3377KuoInit[]        =   {   1   ,   3   ,   3   ,   11  ,   17  ,   21  ,   31  ,   169 ,   425 ,   755 ,   1391    ,   3433    ,   3275    ,   15657   ,   9085    ,0 };
            const unsigned long dim3378KuoInit[]        =   {   1   ,   3   ,   1   ,   7   ,   25  ,   23  ,   49  ,   227 ,   103 ,   5   ,   1995    ,   3803    ,   1109    ,   6015    ,   2433    ,0 };
            const unsigned long dim3379KuoInit[]        =   {   1   ,   1   ,   7   ,   7   ,   19  ,   39  ,   91  ,   219 ,   403 ,   475 ,   1629    ,   117 ,   2317    ,   3991    ,   23467   ,0 };
            const unsigned long dim3380KuoInit[]        =   {   1   ,   1   ,   7   ,   5   ,   19  ,   63  ,   79  ,   9   ,   19  ,   747 ,   1045    ,   875 ,   5809    ,   949 ,   10943   ,0 };
            const unsigned long dim3381KuoInit[]        =   {   1   ,   1   ,   3   ,   11  ,   3   ,   29  ,   83  ,   103 ,   481 ,   565 ,   451 ,   1353    ,   5417    ,   2747    ,   141 ,0 };
            const unsigned long dim3382KuoInit[]        =   {   1   ,   1   ,   1   ,   11  ,   17  ,   43  ,   75  ,   51  ,   335 ,   257 ,   1077    ,   517 ,   4029    ,   13761   ,   8653    ,0 };
            const unsigned long dim3383KuoInit[]        =   {   1   ,   1   ,   1   ,   1   ,   13  ,   51  ,   41  ,   195 ,   331 ,   155 ,   1711    ,   2787    ,   6163    ,   12641   ,   20731   ,0 };
            const unsigned long dim3384KuoInit[]        =   {   1   ,   1   ,   1   ,   11  ,   5   ,   37  ,   117 ,   97  ,   133 ,   793 ,   491 ,   2185    ,   3671    ,   12257   ,   6239    ,0 };
            const unsigned long dim3385KuoInit[]        =   {   1   ,   1   ,   1   ,   5   ,   9   ,   33  ,   69  ,   153 ,   29  ,   443 ,   921 ,   3679    ,   7385    ,   7229    ,   16443   ,0 };
            const unsigned long dim3386KuoInit[]        =   {   1   ,   3   ,   3   ,   9   ,   19  ,   19  ,   87  ,   147 ,   113 ,   577 ,   505 ,   343 ,   7385    ,   3247    ,   6411    ,0 };
            const unsigned long dim3387KuoInit[]        =   {   1   ,   1   ,   7   ,   9   ,   15  ,   33  ,   83  ,   69  ,   27  ,   519 ,   115 ,   4059    ,   5281    ,   7995    ,   25537   ,0 };
            const unsigned long dim3388KuoInit[]        =   {   1   ,   1   ,   5   ,   15  ,   21  ,   25  ,   37  ,   35  ,   263 ,   531 ,   197 ,   27  ,   7871    ,   7383    ,   28467   ,0 };
            const unsigned long dim3389KuoInit[]        =   {   1   ,   3   ,   5   ,   1   ,   1   ,   49  ,   87  ,   173 ,   37  ,   633 ,   65  ,   2449    ,   149 ,   7145    ,   6461    ,0 };
            const unsigned long dim3390KuoInit[]        =   {   1   ,   1   ,   3   ,   7   ,   17  ,   35  ,   91  ,   45  ,   257 ,   351 ,   1569    ,   3757    ,   5203    ,   8071    ,   32211   ,0 };
            const unsigned long dim3391KuoInit[]        =   {   1   ,   1   ,   3   ,   9   ,   9   ,   39  ,   51  ,   33  ,   77  ,   663 ,   287 ,   1127    ,   4549    ,   6321    ,   16199   ,0 };
            const unsigned long dim3392KuoInit[]        =   {   1   ,   1   ,   5   ,   9   ,   3   ,   3   ,   123 ,   35  ,   75  ,   253 ,   1661    ,   3191    ,   1749    ,   14421   ,   15767   ,0 };
            const unsigned long dim3393KuoInit[]        =   {   1   ,   1   ,   1   ,   15  ,   21  ,   27  ,   69  ,   233 ,   425 ,   89  ,   1069    ,   603 ,   5439    ,   5597    ,   24183   ,0 };
            const unsigned long dim3394KuoInit[]        =   {   1   ,   1   ,   7   ,   15  ,   19  ,   57  ,   127 ,   83  ,   445 ,   619 ,   1541    ,   2695    ,   1657    ,   10787   ,   27703   ,0 };
            const unsigned long dim3395KuoInit[]        =   {   1   ,   3   ,   1   ,   3   ,   31  ,   31  ,   29  ,   61  ,   245 ,   449 ,   145 ,   3619    ,   5267    ,   9403    ,   9637    ,0 };
            const unsigned long dim3396KuoInit[]        =   {   1   ,   1   ,   5   ,   9   ,   29  ,   61  ,   33  ,   245 ,   341 ,   991 ,   639 ,   3391    ,   2971    ,   857 ,   26191   ,0 };
            const unsigned long dim3397KuoInit[]        =   {   1   ,   1   ,   1   ,   5   ,   9   ,   7   ,   57  ,   33  ,   79  ,   489 ,   379 ,   3907    ,   3759    ,   4243    ,   9811    ,0 };
            const unsigned long dim3398KuoInit[]        =   {   1   ,   3   ,   5   ,   15  ,   7   ,   61  ,   7   ,   119 ,   51  ,   927 ,   1505    ,   2997    ,   723 ,   8987    ,   29007   ,0 };
            const unsigned long dim3399KuoInit[]        =   {   1   ,   1   ,   3   ,   1   ,   31  ,   55  ,   63  ,   15  ,   421 ,   423 ,   1041    ,   3967    ,   2471    ,   13273   ,   18587   ,0 };
            const unsigned long dim3400KuoInit[]        =   {   1   ,   3   ,   3   ,   13  ,   9   ,   49  ,   111 ,   211 ,   87  ,   887 ,   1867    ,   19  ,   5329    ,   12411   ,   18537   ,0 };
            const unsigned long dim3401KuoInit[]        =   {   1   ,   3   ,   1   ,   1   ,   27  ,   35  ,   89  ,   81  ,   119 ,   635 ,   63  ,   2185    ,   7001    ,   11283   ,   11883   ,0 };
            const unsigned long dim3402KuoInit[]        =   {   1   ,   1   ,   3   ,   9   ,   31  ,   61  ,   55  ,   237 ,   327 ,   537 ,   123 ,   3589    ,   3019    ,   705 ,   16701   ,0 };
            const unsigned long dim3403KuoInit[]        =   {   1   ,   1   ,   7   ,   1   ,   7   ,   13  ,   103 ,   101 ,   473 ,   7   ,   381 ,   3643    ,   1227    ,   11335   ,   615 ,0 };
            const unsigned long dim3404KuoInit[]        =   {   1   ,   3   ,   5   ,   13  ,   1   ,   9   ,   127 ,   65  ,   479 ,   623 ,   469 ,   3403    ,   2423    ,   4269    ,   15195   ,0 };
            const unsigned long dim3405KuoInit[]        =   {   1   ,   1   ,   7   ,   15  ,   3   ,   23  ,   125 ,   195 ,   267 ,   923 ,   961 ,   799 ,   783 ,   893 ,   22403   ,0 };
            const unsigned long dim3406KuoInit[]        =   {   1   ,   3   ,   1   ,   3   ,   21  ,   7   ,   21  ,   151 ,   373 ,   377 ,   1685    ,   3731    ,   6019    ,   1059    ,   32457   ,0 };
            const unsigned long dim3407KuoInit[]        =   {   1   ,   3   ,   3   ,   11  ,   9   ,   31  ,   55  ,   207 ,   109 ,   945 ,   1073    ,   3545    ,   4191    ,   12965   ,   20863   ,0 };
            const unsigned long dim3408KuoInit[]        =   {   1   ,   1   ,   1   ,   5   ,   27  ,   45  ,   15  ,   67  ,   101 ,   207 ,   77  ,   3309    ,   747 ,   13857   ,   25737   ,0 };
            const unsigned long dim3409KuoInit[]        =   {   1   ,   1   ,   1   ,   5   ,   1   ,   31  ,   15  ,   235 ,   57  ,   449 ,   411 ,   3739    ,   5093    ,   12963   ,   30649   ,0 };
            const unsigned long dim3410KuoInit[]        =   {   1   ,   3   ,   7   ,   5   ,   5   ,   33  ,   87  ,   13  ,   57  ,   727 ,   1041    ,   3369    ,   6047    ,   1007    ,   4587    ,0 };
            const unsigned long dim3411KuoInit[]        =   {   1   ,   1   ,   3   ,   9   ,   29  ,   43  ,   11  ,   251 ,   99  ,   413 ,   1943    ,   3461    ,   993 ,   13983   ,   8157    ,0 };
            const unsigned long dim3412KuoInit[]        =   {   1   ,   1   ,   5   ,   15  ,   7   ,   27  ,   105 ,   9   ,   69  ,   949 ,   245 ,   2537    ,   5941    ,   10961   ,   17115   ,0 };
            const unsigned long dim3413KuoInit[]        =   {   1   ,   1   ,   7   ,   5   ,   17  ,   41  ,   63  ,   167 ,   45  ,   391 ,   895 ,   2565    ,   2681    ,   10855   ,   6919    ,0 };
            const unsigned long dim3414KuoInit[]        =   {   1   ,   3   ,   7   ,   3   ,   7   ,   53  ,   15  ,   207 ,   467 ,   131 ,   761 ,   189 ,   4977    ,   4345    ,   7539    ,0 };
            const unsigned long dim3415KuoInit[]        =   {   1   ,   1   ,   5   ,   5   ,   23  ,   17  ,   127 ,   3   ,   455 ,   787 ,   383 ,   2381    ,   6407    ,   3253    ,   1787    ,0 };
            const unsigned long dim3416KuoInit[]        =   {   1   ,   3   ,   5   ,   15  ,   5   ,   7   ,   37  ,   75  ,   351 ,   141 ,   93  ,   2631    ,   609 ,   11455   ,   20603   ,0 };
            const unsigned long dim3417KuoInit[]        =   {   1   ,   3   ,   7   ,   9   ,   5   ,   33  ,   19  ,   15  ,   101 ,   1005    ,   7   ,   1341    ,   4577    ,   11627   ,   17061   ,0 };
            const unsigned long dim3418KuoInit[]        =   {   1   ,   3   ,   3   ,   9   ,   5   ,   7   ,   87  ,   181 ,   197 ,   423 ,   361 ,   545 ,   2459    ,   7419    ,   14457   ,0 };
            const unsigned long dim3419KuoInit[]        =   {   1   ,   1   ,   7   ,   7   ,   25  ,   37  ,   93  ,   85  ,   345 ,   459 ,   465 ,   3377    ,   795 ,   13439   ,   4765    ,0 };
            const unsigned long dim3420KuoInit[]        =   {   1   ,   1   ,   7   ,   9   ,   25  ,   55  ,   109 ,   63  ,   381 ,   329 ,   1787    ,   3163    ,   279 ,   4647    ,   8723    ,0 };
            const unsigned long dim3421KuoInit[]        =   {   1   ,   3   ,   3   ,   13  ,   23  ,   33  ,   85  ,   41  ,   151 ,   319 ,   265 ,   259 ,   29  ,   4217    ,   21025   ,0 };
            const unsigned long dim3422KuoInit[]        =   {   1   ,   1   ,   5   ,   9   ,   31  ,   17  ,   93  ,   27  ,   455 ,   1015    ,   81  ,   621 ,   6511    ,   7617    ,   6775    ,0 };
            const unsigned long dim3423KuoInit[]        =   {   1   ,   1   ,   1   ,   11  ,   25  ,   21  ,   73  ,   161 ,   267 ,   911 ,   25  ,   1385    ,   4073    ,   14423   ,   31473   ,0 };
            const unsigned long dim3424KuoInit[]        =   {   1   ,   3   ,   5   ,   11  ,   13  ,   33  ,   95  ,   113 ,   371 ,   291 ,   1055    ,   3779    ,   991 ,   9699    ,   9495    ,0 };
            const unsigned long dim3425KuoInit[]        =   {   1   ,   1   ,   1   ,   7   ,   31  ,   55  ,   111 ,   215 ,   399 ,   721 ,   1443    ,   763 ,   3119    ,   11075   ,   15489   ,0 };
            const unsigned long dim3426KuoInit[]        =   {   1   ,   1   ,   1   ,   3   ,   17  ,   57  ,   119 ,   171 ,   299 ,   299 ,   1919    ,   3103    ,   5691    ,   7927    ,   18875   ,0 };
            const unsigned long dim3427KuoInit[]        =   {   1   ,   1   ,   5   ,   7   ,   13  ,   41  ,   35  ,   209 ,   383 ,   381 ,   901 ,   3341    ,   3421    ,   11501   ,   2187    ,0 };
            const unsigned long dim3428KuoInit[]        =   {   1   ,   3   ,   1   ,   7   ,   1   ,   3   ,   83  ,   105 ,   495 ,   297 ,   413 ,   2941    ,   4309    ,   9829    ,   32115   ,0 };
            const unsigned long dim3429KuoInit[]        =   {   1   ,   3   ,   5   ,   3   ,   3   ,   7   ,   81  ,   9   ,   55  ,   19  ,   225 ,   2915    ,   6255    ,   10967   ,   8767    ,0 };
            const unsigned long dim3430KuoInit[]        =   {   1   ,   3   ,   1   ,   1   ,   23  ,   59  ,   21  ,   127 ,   7   ,   703 ,   1189    ,   2375    ,   6273    ,   1655    ,   10115   ,0 };
            const unsigned long dim3431KuoInit[]        =   {   1   ,   1   ,   1   ,   15  ,   19  ,   35  ,   39  ,   185 ,   143 ,   163 ,   1491    ,   201 ,   2473    ,   661 ,   12827   ,0 };
            const unsigned long dim3432KuoInit[]        =   {   1   ,   1   ,   1   ,   5   ,   29  ,   49  ,   5   ,   27  ,   131 ,   1005    ,   991 ,   1159    ,   5437    ,   5039    ,   4037    ,0 };
            const unsigned long dim3433KuoInit[]        =   {   1   ,   3   ,   7   ,   3   ,   11  ,   41  ,   123 ,   43  ,   11  ,   443 ,   109 ,   3995    ,   5377    ,   2907    ,   30035   ,0 };
            const unsigned long dim3434KuoInit[]        =   {   1   ,   3   ,   1   ,   1   ,   31  ,   59  ,   53  ,   219 ,   185 ,   31  ,   641 ,   1695    ,   5817    ,   9001    ,   28827   ,0 };
            const unsigned long dim3435KuoInit[]        =   {   1   ,   1   ,   1   ,   3   ,   13  ,   37  ,   37  ,   119 ,   123 ,   57  ,   1241    ,   551 ,   5825    ,   11551   ,   23521   ,0 };
            const unsigned long dim3436KuoInit[]        =   {   1   ,   1   ,   3   ,   15  ,   13  ,   37  ,   33  ,   61  ,   209 ,   375 ,   335 ,   1223    ,   4023    ,   12743   ,   8375    ,0 };
            const unsigned long dim3437KuoInit[]        =   {   1   ,   3   ,   3   ,   3   ,   31  ,   7   ,   103 ,   43  ,   451 ,   517 ,   1025    ,   2853    ,   5943    ,   13655   ,   21409   ,0 };
            const unsigned long dim3438KuoInit[]        =   {   1   ,   1   ,   1   ,   13  ,   11  ,   37  ,   11  ,   167 ,   271 ,   763 ,   85  ,   3727    ,   641 ,   2841    ,   23803   ,0 };
            const unsigned long dim3439KuoInit[]        =   {   1   ,   1   ,   7   ,   5   ,   3   ,   19  ,   117 ,   225 ,   275 ,   327 ,   1263    ,   1597    ,   1439    ,   8531    ,   18063   ,0 };
            const unsigned long dim3440KuoInit[]        =   {   1   ,   3   ,   5   ,   3   ,   9   ,   45  ,   57  ,   169 ,   453 ,   949 ,   585 ,   1365    ,   4263    ,   3461    ,   7191    ,0 };
            const unsigned long dim3441KuoInit[]        =   {   1   ,   3   ,   7   ,   9   ,   25  ,   51  ,   97  ,   255 ,   457 ,   293 ,   1027    ,   1291    ,   6685    ,   2617    ,   3055    ,0 };
            const unsigned long dim3442KuoInit[]        =   {   1   ,   3   ,   7   ,   5   ,   27  ,   33  ,   47  ,   65  ,   43  ,   545 ,   123 ,   1093    ,   5563    ,   3013    ,   18451   ,0 };
            const unsigned long dim3443KuoInit[]        =   {   1   ,   1   ,   3   ,   11  ,   9   ,   25  ,   71  ,   241 ,   495 ,   561 ,   1565    ,   703 ,   809 ,   10961   ,   3001    ,0 };
            const unsigned long dim3444KuoInit[]        =   {   1   ,   3   ,   5   ,   1   ,   1   ,   47  ,   97  ,   233 ,   259 ,   39  ,   543 ,   303 ,   2705    ,   3683    ,   32123   ,0 };
            const unsigned long dim3445KuoInit[]        =   {   1   ,   1   ,   5   ,   7   ,   1   ,   35  ,   91  ,   43  ,   417 ,   299 ,   1635    ,   2163    ,   7959    ,   9015    ,   5661    ,0 };
            const unsigned long dim3446KuoInit[]        =   {   1   ,   1   ,   3   ,   11  ,   1   ,   63  ,   49  ,   209 ,   125 ,   241 ,   987 ,   3173    ,   631 ,   9767    ,   8467    ,0 };
            const unsigned long dim3447KuoInit[]        =   {   1   ,   1   ,   1   ,   5   ,   25  ,   49  ,   43  ,   117 ,   65  ,   81  ,   313 ,   1917    ,   6087    ,   1503    ,   3383    ,0 };
            const unsigned long dim3448KuoInit[]        =   {   1   ,   3   ,   5   ,   1   ,   23  ,   53  ,   43  ,   29  ,   91  ,   727 ,   1157    ,   2985    ,   3851    ,   15897   ,   17577   ,0 };
            const unsigned long dim3449KuoInit[]        =   {   1   ,   1   ,   3   ,   15  ,   15  ,   27  ,   35  ,   129 ,   161 ,   679 ,   1001    ,   2681    ,   5287    ,   10367   ,   21017   ,0 };
            const unsigned long dim3450KuoInit[]        =   {   1   ,   3   ,   3   ,   1   ,   23  ,   51  ,   127 ,   119 ,   467 ,   161 ,   1229    ,   2969    ,   6321    ,   1141    ,   7827    ,0 };
            const unsigned long dim3451KuoInit[]        =   {   1   ,   3   ,   5   ,   15  ,   19  ,   45  ,   33  ,   173 ,   205 ,   815 ,   783 ,   1323    ,   69  ,   5191    ,   23795   ,0 };
            const unsigned long dim3452KuoInit[]        =   {   1   ,   3   ,   5   ,   9   ,   19  ,   1   ,   111 ,   253 ,   463 ,   247 ,   1877    ,   1831    ,   8037    ,   12769   ,   8111    ,0 };
            const unsigned long dim3453KuoInit[]        =   {   1   ,   3   ,   1   ,   1   ,   21  ,   7   ,   19  ,   241 ,   297 ,   953 ,   1759    ,   2187    ,   1587    ,   14857   ,   13245   ,0 };
            const unsigned long dim3454KuoInit[]        =   {   1   ,   1   ,   1   ,   15  ,   27  ,   21  ,   111 ,   37  ,   333 ,   485 ,   127 ,   1781    ,   545 ,   251 ,   2757    ,0 };
            const unsigned long dim3455KuoInit[]        =   {   1   ,   3   ,   1   ,   1   ,   3   ,   7   ,   21  ,   69  ,   171 ,   225 ,   1505    ,   2725    ,   923 ,   9451    ,   29443   ,0 };
            const unsigned long dim3456KuoInit[]        =   {   1   ,   3   ,   7   ,   7   ,   13  ,   57  ,   97  ,   175 ,   429 ,   529 ,   677 ,   1501    ,   4717    ,   12705   ,   31473   ,0 };
            const unsigned long dim3457KuoInit[]        =   {   1   ,   3   ,   7   ,   11  ,   29  ,   7   ,   57  ,   115 ,   55  ,   167 ,   903 ,   3855    ,   8051    ,   15567   ,   31733   ,0 };
            const unsigned long dim3458KuoInit[]        =   {   1   ,   1   ,   1   ,   1   ,   1   ,   1   ,   1   ,   1   ,   1   ,   1   ,   1   ,   1   ,   1   ,   1   ,   1   ,0 };
            const unsigned long dim3459KuoInit[]        =   {   1   ,   1   ,   1   ,   13  ,   11  ,   17  ,   73  ,   193 ,   151 ,   927 ,   1375    ,   4037    ,   6855    ,   10481   ,   8973    ,0 };
            const unsigned long dim3460KuoInit[]        =   {   1   ,   3   ,   7   ,   1   ,   25  ,   7   ,   115 ,   39  ,   469 ,   903 ,   655 ,   299 ,   6007    ,   6629    ,   11871   ,0 };
            const unsigned long dim3461KuoInit[]        =   {   1   ,   3   ,   7   ,   5   ,   15  ,   49  ,   79  ,   115 ,   55  ,   841 ,   1091    ,   1641    ,   5693    ,   3607    ,   13699   ,0 };
            const unsigned long dim3462KuoInit[]        =   {   1   ,   1   ,   1   ,   15  ,   5   ,   33  ,   99  ,   239 ,   157 ,   133 ,   2023    ,   683 ,   3419    ,   10993   ,   32187   ,0 };
            const unsigned long dim3463KuoInit[]        =   {   1   ,   1   ,   7   ,   1   ,   3   ,   13  ,   95  ,   233 ,   341 ,   243 ,   2043    ,   653 ,   4673    ,   1441    ,   19087   ,0 };
            const unsigned long dim3464KuoInit[]        =   {   1   ,   3   ,   3   ,   3   ,   29  ,   51  ,   117 ,   199 ,   49  ,   467 ,   461 ,   2465    ,   7541    ,   6815    ,   28711   ,0 };
            const unsigned long dim3465KuoInit[]        =   {   1   ,   3   ,   1   ,   1   ,   25  ,   47  ,   31  ,   79  ,   9   ,   39  ,   733 ,   3567    ,   5965    ,   6627    ,   1109    ,0 };
            const unsigned long dim3466KuoInit[]        =   {   1   ,   3   ,   1   ,   13  ,   9   ,   53  ,   27  ,   101 ,   341 ,   369 ,   1659    ,   1645    ,   4465    ,   10175   ,   4565    ,0 };
            const unsigned long dim3467KuoInit[]        =   {   1   ,   3   ,   5   ,   7   ,   13  ,   61  ,   67  ,   125 ,   427 ,   105 ,   133 ,   2303    ,   5595    ,   4527    ,   23151   ,0 };
            const unsigned long dim3468KuoInit[]        =   {   1   ,   1   ,   5   ,   7   ,   19  ,   29  ,   87  ,   167 ,   141 ,   657 ,   947 ,   367 ,   2617    ,   9287    ,   629 ,0 };
            const unsigned long dim3469KuoInit[]        =   {   1   ,   3   ,   7   ,   5   ,   21  ,   25  ,   35  ,   21  ,   391 ,   1019    ,   991 ,   3343    ,   1221    ,   1591    ,   18095   ,0 };
            const unsigned long dim3470KuoInit[]        =   {   1   ,   3   ,   7   ,   9   ,   31  ,   49  ,   99  ,   185 ,   405 ,   261 ,   935 ,   3447    ,   4081    ,   5537    ,   26329   ,0 };
            const unsigned long dim3471KuoInit[]        =   {   1   ,   1   ,   3   ,   7   ,   25  ,   51  ,   107 ,   83  ,   449 ,   733 ,   597 ,   1023    ,   7457    ,   14343   ,   7119    ,0 };
            const unsigned long dim3472KuoInit[]        =   {   1   ,   1   ,   1   ,   5   ,   23  ,   21  ,   35  ,   83  ,   185 ,   531 ,   1283    ,   2235    ,   7775    ,   15895   ,   32651   ,0 };
            const unsigned long dim3473KuoInit[]        =   {   1   ,   1   ,   5   ,   15  ,   27  ,   13  ,   23  ,   227 ,   491 ,   701 ,   341 ,   2109    ,   8045    ,   1529    ,   10589   ,0 };
            const unsigned long dim3474KuoInit[]        =   {   1   ,   1   ,   5   ,   5   ,   21  ,   51  ,   31  ,   101 ,   31  ,   861 ,   401 ,   1519    ,   5349    ,   2623    ,   22667   ,0 };
            const unsigned long dim3475KuoInit[]        =   {   1   ,   3   ,   5   ,   5   ,   1   ,   41  ,   43  ,   25  ,   491 ,   329 ,   537 ,   1357    ,   5567    ,   3951    ,   15293   ,0 };
            const unsigned long dim3476KuoInit[]        =   {   1   ,   3   ,   7   ,   3   ,   15  ,   39  ,   113 ,   105 ,   51  ,   483 ,   1417    ,   2971    ,   3   ,   9673    ,   4331    ,0 };
            const unsigned long dim3477KuoInit[]        =   {   1   ,   3   ,   5   ,   13  ,   17  ,   3   ,   19  ,   55  ,   443 ,   967 ,   327 ,   3463    ,   5427    ,   195 ,   7495    ,0 };
            const unsigned long dim3478KuoInit[]        =   {   1   ,   1   ,   3   ,   3   ,   15  ,   15  ,   11  ,   157 ,   311 ,   967 ,   89  ,   3691    ,   7307    ,   12309   ,   1915    ,0 };
            const unsigned long dim3479KuoInit[]        =   {   1   ,   3   ,   5   ,   7   ,   25  ,   7   ,   125 ,   163 ,   231 ,   177 ,   1497    ,   3119    ,   3281    ,   15033   ,   9457    ,0 };
            const unsigned long dim3480KuoInit[]        =   {   1   ,   3   ,   3   ,   9   ,   25  ,   51  ,   51  ,   243 ,   427 ,   831 ,   709 ,   3741    ,   445 ,   14275   ,   32323   ,0 };
            const unsigned long dim3481KuoInit[]        =   {   1   ,   3   ,   3   ,   13  ,   1   ,   41  ,   123 ,   237 ,   401 ,   19  ,   1161    ,   985 ,   5119    ,   14879   ,   5785    ,0 };
            const unsigned long dim3482KuoInit[]        =   {   1   ,   3   ,   1   ,   3   ,   3   ,   11  ,   121 ,   209 ,   507 ,   271 ,   1045    ,   2547    ,   1019    ,   7107    ,   31823   ,0 };
            const unsigned long dim3483KuoInit[]        =   {   1   ,   1   ,   5   ,   3   ,   21  ,   3   ,   33  ,   73  ,   79  ,   607 ,   47  ,   3045    ,   2471    ,   2023    ,   8115    ,0 };
            const unsigned long dim3484KuoInit[]        =   {   1   ,   3   ,   7   ,   13  ,   19  ,   49  ,   111 ,   185 ,   209 ,   683 ,   1459    ,   3827    ,   6591    ,   13139   ,   2057    ,0 };
            const unsigned long dim3485KuoInit[]        =   {   1   ,   3   ,   7   ,   9   ,   17  ,   63  ,   77  ,   217 ,   161 ,   287 ,   335 ,   3947    ,   5255    ,   12297   ,   16829   ,0 };
            const unsigned long dim3486KuoInit[]        =   {   1   ,   3   ,   1   ,   11  ,   17  ,   17  ,   31  ,   231 ,   483 ,   437 ,   431 ,   957 ,   6499    ,   12947   ,   29751   ,0 };
            const unsigned long dim3487KuoInit[]        =   {   1   ,   3   ,   7   ,   15  ,   5   ,   57  ,   119 ,   43  ,   385 ,   743 ,   249 ,   3479    ,   7751    ,   6321    ,   31747   ,0 };
            const unsigned long dim3488KuoInit[]        =   {   1   ,   1   ,   5   ,   3   ,   5   ,   1   ,   113 ,   87  ,   439 ,   507 ,   1517    ,   2081    ,   5947    ,   10967   ,   10219   ,0 };
            const unsigned long dim3489KuoInit[]        =   {   1   ,   3   ,   1   ,   13  ,   23  ,   29  ,   39  ,   33  ,   129 ,   493 ,   1119    ,   2177    ,   4135    ,   5051    ,   20241   ,0 };
            const unsigned long dim3490KuoInit[]        =   {   1   ,   3   ,   7   ,   13  ,   19  ,   13  ,   71  ,   163 ,   107 ,   285 ,   1999    ,   3371    ,   3555    ,   16265   ,   2825    ,0 };
            const unsigned long dim3491KuoInit[]        =   {   1   ,   3   ,   7   ,   11  ,   5   ,   19  ,   67  ,   201 ,   203 ,   619 ,   927 ,   1769    ,   5655    ,   4539    ,   1719    ,0 };
            const unsigned long dim3492KuoInit[]        =   {   1   ,   3   ,   5   ,   7   ,   15  ,   43  ,   107 ,   91  ,   315 ,   281 ,   525 ,   3243    ,   7751    ,   8991    ,   29195   ,0 };
            const unsigned long dim3493KuoInit[]        =   {   1   ,   1   ,   7   ,   7   ,   21  ,   43  ,   35  ,   81  ,   199 ,   51  ,   935 ,   357 ,   1723    ,   8183    ,   6209    ,0 };
            const unsigned long dim3494KuoInit[]        =   {   1   ,   3   ,   5   ,   5   ,   9   ,   31  ,   75  ,   71  ,   375 ,   181 ,   1503    ,   3479    ,   2255    ,   1371    ,   16127   ,0 };
            const unsigned long dim3495KuoInit[]        =   {   1   ,   1   ,   7   ,   11  ,   29  ,   61  ,   57  ,   145 ,   149 ,   673 ,   1953    ,   2111    ,   3525    ,   1007    ,   15927   ,0 };
            const unsigned long dim3496KuoInit[]        =   {   1   ,   1   ,   7   ,   15  ,   25  ,   25  ,   103 ,   175 ,   273 ,   193 ,   1215    ,   3059    ,   7805    ,   1455    ,   1627    ,0 };
            const unsigned long dim3497KuoInit[]        =   {   1   ,   3   ,   1   ,   1   ,   23  ,   45  ,   61  ,   227 ,   487 ,   89  ,   739 ,   1867    ,   3643    ,   9879    ,   17307   ,0 };
            const unsigned long dim3498KuoInit[]        =   {   1   ,   1   ,   5   ,   1   ,   23  ,   5   ,   117 ,   25  ,   325 ,   725 ,   1999    ,   3453    ,   2353    ,   5467    ,   2065    ,0 };
            const unsigned long dim3499KuoInit[]        =   {   1   ,   1   ,   5   ,   5   ,   25  ,   3   ,   95  ,   243 ,   175 ,   291 ,   821 ,   3531    ,   4873    ,   9975    ,   30437   ,0 };
            const unsigned long dim3500KuoInit[]        =   {   1   ,   3   ,   3   ,   11  ,   5   ,   41  ,   121 ,   23  ,   15  ,   195 ,   1141    ,   171 ,   2245    ,   8043    ,   2201    ,0 };
            const unsigned long dim3501KuoInit[]        =   {   1   ,   3   ,   7   ,   1   ,   29  ,   1   ,   41  ,   233 ,   217 ,   501 ,   913 ,   3589    ,   1153    ,   4601    ,   23147   ,0 };
            const unsigned long dim3502KuoInit[]        =   {   1   ,   1   ,   5   ,   15  ,   1   ,   41  ,   71  ,   255 ,   283 ,   125 ,   845 ,   1261    ,   7519    ,   15623   ,   13301   ,0 };
            const unsigned long dim3503KuoInit[]        =   {   1   ,   3   ,   3   ,   15  ,   11  ,   3   ,   73  ,   161 ,   203 ,   775 ,   905 ,   411 ,   4139    ,   8289    ,   8581    ,0 };
            const unsigned long dim3504KuoInit[]        =   {   1   ,   1   ,   7   ,   13  ,   7   ,   57  ,   17  ,   115 ,   273 ,   949 ,   913 ,   3267    ,   6949    ,   6785    ,   30449   ,0 };
            const unsigned long dim3505KuoInit[]        =   {   1   ,   1   ,   7   ,   5   ,   3   ,   49  ,   55  ,   191 ,   505 ,   807 ,   287 ,   3595    ,   2261    ,   13751   ,   15339   ,0 };
            const unsigned long dim3506KuoInit[]        =   {   1   ,   1   ,   5   ,   11  ,   23  ,   13  ,   83  ,   145 ,   387 ,   37  ,   341 ,   111 ,   3309    ,   4481    ,   21679   ,0 };
            const unsigned long dim3507KuoInit[]        =   {   1   ,   1   ,   7   ,   15  ,   3   ,   51  ,   13  ,   247 ,   419 ,   555 ,   431 ,   187 ,   2753    ,   9071    ,   25189   ,0 };
            const unsigned long dim3508KuoInit[]        =   {   1   ,   3   ,   7   ,   3   ,   29  ,   33  ,   119 ,   63  ,   363 ,   197 ,   1901    ,   2231    ,   1453    ,   3901    ,   13713   ,0 };
            const unsigned long dim3509KuoInit[]        =   {   1   ,   1   ,   5   ,   7   ,   9   ,   15  ,   91  ,   107 ,   397 ,   739 ,   1419    ,   481 ,   5713    ,   14089   ,   11699   ,0 };
            const unsigned long dim3510KuoInit[]        =   {   1   ,   3   ,   3   ,   7   ,   9   ,   37  ,   47  ,   183 ,   45  ,   135 ,   903 ,   2769    ,   5929    ,   8175    ,   8661    ,0 };
            const unsigned long dim3511KuoInit[]        =   {   1   ,   1   ,   3   ,   9   ,   7   ,   63  ,   117 ,   81  ,   385 ,   459 ,   1455    ,   4089    ,   395 ,   7569    ,   18089   ,0 };
            const unsigned long dim3512KuoInit[]        =   {   1   ,   1   ,   1   ,   3   ,   15  ,   29  ,   71  ,   47  ,   423 ,   177 ,   1481    ,   1087    ,   3703    ,   5857    ,   9631    ,0 };
            const unsigned long dim3513KuoInit[]        =   {   1   ,   3   ,   1   ,   7   ,   23  ,   45  ,   127 ,   91  ,   457 ,   81  ,   681 ,   257 ,   361 ,   11739   ,   1503    ,0 };
            const unsigned long dim3514KuoInit[]        =   {   1   ,   3   ,   3   ,   9   ,   29  ,   29  ,   107 ,   125 ,   55  ,   759 ,   1287    ,   1937    ,   353 ,   1275    ,   14349   ,0 };
            const unsigned long dim3515KuoInit[]        =   {   1   ,   3   ,   5   ,   11  ,   15  ,   53  ,   35  ,   1   ,   419 ,   255 ,   1891    ,   7   ,   3005    ,   1493    ,   165 ,0 };
            const unsigned long dim3516KuoInit[]        =   {   1   ,   3   ,   5   ,   1   ,   5   ,   19  ,   45  ,   109 ,   479 ,   309 ,   277 ,   2749    ,   4859    ,   4301    ,   17799   ,0 };
            const unsigned long dim3517KuoInit[]        =   {   1   ,   3   ,   7   ,   15  ,   21  ,   57  ,   39  ,   19  ,   433 ,   77  ,   1999    ,   3467    ,   2097    ,   6853    ,   1763    ,0 };
            const unsigned long dim3518KuoInit[]        =   {   1   ,   1   ,   5   ,   5   ,   7   ,   39  ,   109 ,   53  ,   1   ,   543 ,   623 ,   573 ,   4039    ,   1367    ,   23177   ,0 };
            const unsigned long dim3519KuoInit[]        =   {   1   ,   3   ,   1   ,   11  ,   1   ,   31  ,   3   ,   203 ,   409 ,   409 ,   1471    ,   1245    ,   3881    ,   14467   ,   22731   ,0 };
            const unsigned long dim3520KuoInit[]        =   {   1   ,   3   ,   1   ,   11  ,   19  ,   41  ,   39  ,   95  ,   187 ,   823 ,   973 ,   3055    ,   3803    ,   4229    ,   29363   ,0 };
            const unsigned long dim3521KuoInit[]        =   {   1   ,   1   ,   5   ,   15  ,   29  ,   41  ,   69  ,   121 ,   385 ,   717 ,   1601    ,   3061    ,   6523    ,   13423   ,   12559   ,0 };
            const unsigned long dim3522KuoInit[]        =   {   1   ,   3   ,   7   ,   1   ,   5   ,   45  ,   85  ,   153 ,   337 ,   587 ,   595 ,   1183    ,   3799    ,   12775   ,   19537   ,0 };
            const unsigned long dim3523KuoInit[]        =   {   1   ,   3   ,   7   ,   15  ,   11  ,   45  ,   97  ,   37  ,   167 ,   227 ,   1003    ,   2991    ,   2291    ,   15865   ,   6645    ,0 };
            const unsigned long dim3524KuoInit[]        =   {   1   ,   3   ,   3   ,   7   ,   31  ,   41  ,   123 ,   225 ,   275 ,   733 ,   1611    ,   1599    ,   7345    ,   297 ,   22551   ,0 };
            const unsigned long dim3525KuoInit[]        =   {   1   ,   1   ,   5   ,   3   ,   7   ,   45  ,   45  ,   155 ,   41  ,   581 ,   635 ,   4065    ,   3993    ,   7233    ,   27097   ,0 };
            const unsigned long dim3526KuoInit[]        =   {   1   ,   1   ,   7   ,   11  ,   9   ,   39  ,   17  ,   19  ,   335 ,   819 ,   1845    ,   1199    ,   591 ,   7743    ,   26797   ,0 };
            const unsigned long dim3527KuoInit[]        =   {   1   ,   3   ,   7   ,   9   ,   29  ,   33  ,   35  ,   19  ,   287 ,   935 ,   1463    ,   3639    ,   5537    ,   2751    ,   20073   ,0 };
            const unsigned long dim3528KuoInit[]        =   {   1   ,   3   ,   5   ,   5   ,   3   ,   45  ,   59  ,   117 ,   431 ,   419 ,   1325    ,   983 ,   6065    ,   3501    ,   15765   ,0 };
            const unsigned long dim3529KuoInit[]        =   {   1   ,   1   ,   1   ,   9   ,   27  ,   7   ,   123 ,   205 ,   155 ,   59  ,   1139    ,   1147    ,   4845    ,   15131   ,   10155   ,0 };
            const unsigned long dim3530KuoInit[]        =   {   1   ,   3   ,   7   ,   5   ,   23  ,   37  ,   97  ,   33  ,   337 ,   837 ,   1379    ,   2459    ,   469 ,   10615   ,   16475   ,0 };
            const unsigned long dim3531KuoInit[]        =   {   1   ,   3   ,   1   ,   7   ,   1   ,   23  ,   27  ,   63  ,   503 ,   635 ,   935 ,   1767    ,   6887    ,   6675    ,   10973   ,0 };
            const unsigned long dim3532KuoInit[]        =   {   1   ,   3   ,   1   ,   9   ,   13  ,   29  ,   15  ,   109 ,   185 ,   33  ,   1039    ,   959 ,   1745    ,   10587   ,   9269    ,0 };
            const unsigned long dim3533KuoInit[]        =   {   1   ,   3   ,   3   ,   1   ,   5   ,   29  ,   89  ,   63  ,   341 ,   773 ,   1851    ,   775 ,   1717    ,   6041    ,   11901   ,0 };
            const unsigned long dim3534KuoInit[]        =   {   1   ,   1   ,   7   ,   9   ,   23  ,   47  ,   21  ,   111 ,   289 ,   997 ,   1559    ,   501 ,   5913    ,   4737    ,   4623    ,0 };
            const unsigned long dim3535KuoInit[]        =   {   1   ,   1   ,   1   ,   15  ,   3   ,   3   ,   87  ,   197 ,   391 ,   121 ,   627 ,   3705    ,   4093    ,   3933    ,   13787   ,0 };
            const unsigned long dim3536KuoInit[]        =   {   1   ,   3   ,   3   ,   15  ,   31  ,   9   ,   117 ,   47  ,   373 ,   697 ,   131 ,   3933    ,   3135    ,   9155    ,   3937    ,0 };
            const unsigned long dim3537KuoInit[]        =   {   1   ,   1   ,   7   ,   3   ,   11  ,   39  ,   5   ,   27  ,   253 ,   301 ,   1151    ,   3771    ,   1581    ,   9811    ,   17617   ,0 };
            const unsigned long dim3538KuoInit[]        =   {   1   ,   3   ,   7   ,   15  ,   25  ,   59  ,   97  ,   229 ,   291 ,   349 ,   1153    ,   1859    ,   3357    ,   7919    ,   2733    ,0 };
            const unsigned long dim3539KuoInit[]        =   {   1   ,   1   ,   5   ,   13  ,   5   ,   7   ,   37  ,   177 ,   177 ,   969 ,   29  ,   4017    ,   6791    ,   2823    ,   10607   ,0 };
            const unsigned long dim3540KuoInit[]        =   {   1   ,   1   ,   7   ,   13  ,   27  ,   17  ,   25  ,   137 ,   325 ,   949 ,   1519    ,   3711    ,   4399    ,   9501    ,   2989    ,0 };
            const unsigned long dim3541KuoInit[]        =   {   1   ,   1   ,   5   ,   5   ,   29  ,   27  ,   75  ,   195 ,   417 ,   945 ,   1851    ,   2887    ,   2379    ,   15229   ,   4805    ,0 };
            const unsigned long dim3542KuoInit[]        =   {   1   ,   1   ,   3   ,   11  ,   31  ,   45  ,   105 ,   87  ,   241 ,   153 ,   1941    ,   3459    ,   4809    ,   9823    ,   19419   ,0 };
            const unsigned long dim3543KuoInit[]        =   {   1   ,   1   ,   1   ,   1   ,   1   ,   59  ,   75  ,   51  ,   309 ,   671 ,   1039    ,   2999    ,   1025    ,   15823   ,   4835    ,0 };
            const unsigned long dim3544KuoInit[]        =   {   1   ,   3   ,   7   ,   13  ,   19  ,   37  ,   41  ,   85  ,   11  ,   713 ,   1047    ,   989 ,   2941    ,   6423    ,   7983    ,0 };
            const unsigned long dim3545KuoInit[]        =   {   1   ,   3   ,   5   ,   13  ,   1   ,   55  ,   103 ,   207 ,   275 ,   431 ,   563 ,   101 ,   691 ,   1997    ,   17227   ,0 };
            const unsigned long dim3546KuoInit[]        =   {   1   ,   3   ,   1   ,   3   ,   19  ,   15  ,   67  ,   197 ,   125 ,   377 ,   367 ,   601 ,   4983    ,   1699    ,   29315   ,0 };
            const unsigned long dim3547KuoInit[]        =   {   1   ,   1   ,   5   ,   3   ,   23  ,   19  ,   19  ,   133 ,   83  ,   625 ,   757 ,   1349    ,   5821    ,   2483    ,   26653   ,0 };
            const unsigned long dim3548KuoInit[]        =   {   1   ,   1   ,   3   ,   3   ,   19  ,   17  ,   19  ,   83  ,   235 ,   657 ,   733 ,   1773    ,   7011    ,   3893    ,   8495    ,0 };
            const unsigned long dim3549KuoInit[]        =   {   1   ,   1   ,   3   ,   15  ,   19  ,   33  ,   37  ,   69  ,   385 ,   835 ,   1025    ,   1585    ,   6685    ,   10627   ,   19203   ,0 };
            const unsigned long dim3550KuoInit[]        =   {   1   ,   1   ,   5   ,   11  ,   9   ,   31  ,   65  ,   177 ,   153 ,   781 ,   1183    ,   1215    ,   6271    ,   13099   ,   12449   ,0 };
            const unsigned long dim3551KuoInit[]        =   {   1   ,   1   ,   7   ,   9   ,   13  ,   49  ,   41  ,   57  ,   443 ,   633 ,   1953    ,   2423    ,   789 ,   5591    ,   20891   ,0 };
            const unsigned long dim3552KuoInit[]        =   {   1   ,   1   ,   1   ,   15  ,   17  ,   21  ,   69  ,   177 ,   57  ,   797 ,   13  ,   3987    ,   6597    ,   9935    ,   26559   ,0 };
            const unsigned long dim3553KuoInit[]        =   {   1   ,   3   ,   3   ,   1   ,   21  ,   27  ,   17  ,   207 ,   139 ,   43  ,   1935    ,   3285    ,   4251    ,   4411    ,   24469   ,0 };
            const unsigned long dim3554KuoInit[]        =   {   1   ,   3   ,   3   ,   11  ,   11  ,   47  ,   63  ,   243 ,   19  ,   801 ,   971 ,   1925    ,   5981    ,   13055   ,   12185   ,0 };
            const unsigned long dim3555KuoInit[]        =   {   1   ,   3   ,   5   ,   5   ,   9   ,   13  ,   49  ,   251 ,   315 ,   889 ,   1613    ,   1007    ,   7849    ,   3121    ,   4279    ,0 };
            const unsigned long dim3556KuoInit[]        =   {   1   ,   3   ,   1   ,   15  ,   7   ,   37  ,   95  ,   253 ,   65  ,   177 ,   1551    ,   2307    ,   2325    ,   9975    ,   18277   ,0 };
            const unsigned long dim3557KuoInit[]        =   {   1   ,   3   ,   1   ,   15  ,   7   ,   53  ,   45  ,   5   ,   447 ,   1005    ,   279 ,   2141    ,   4271    ,   4707    ,   11225   ,0 };
            const unsigned long dim3558KuoInit[]        =   {   1   ,   1   ,   3   ,   13  ,   31  ,   11  ,   15  ,   133 ,   313 ,   249 ,   1971    ,   2905    ,   819 ,   11603   ,   25679   ,0 };
            const unsigned long dim3559KuoInit[]        =   {   1   ,   3   ,   3   ,   15  ,   21  ,   49  ,   107 ,   229 ,   101 ,   449 ,   245 ,   1401    ,   391 ,   3749    ,   14555   ,0 };
            const unsigned long dim3560KuoInit[]        =   {   1   ,   1   ,   1   ,   15  ,   7   ,   39  ,   69  ,   223 ,   37  ,   771 ,   1469    ,   2085    ,   1075    ,   14485   ,   10019   ,0 };
            const unsigned long dim3561KuoInit[]        =   {   1   ,   3   ,   1   ,   13  ,   3   ,   27  ,   37  ,   59  ,   69  ,   675 ,   807 ,   2711    ,   3867    ,   9737    ,   21913   ,0 };
            const unsigned long dim3562KuoInit[]        =   {   1   ,   1   ,   3   ,   9   ,   25  ,   45  ,   109 ,   113 ,   267 ,   867 ,   771 ,   1647    ,   2063    ,   14063   ,   27737   ,0 };
            const unsigned long dim3563KuoInit[]        =   {   1   ,   1   ,   7   ,   15  ,   23  ,   51  ,   71  ,   33  ,   103 ,   71  ,   1177    ,   2353    ,   4741    ,   6907    ,   11121   ,0 };
            const unsigned long dim3564KuoInit[]        =   {   1   ,   3   ,   3   ,   7   ,   21  ,   31  ,   77  ,   255 ,   341 ,   249 ,   1263    ,   505 ,   1845    ,   12867   ,   11393   ,0 };
            const unsigned long dim3565KuoInit[]        =   {   1   ,   3   ,   1   ,   1   ,   3   ,   1   ,   5   ,   215 ,   155 ,   107 ,   1283    ,   3501    ,   8079    ,   13921   ,   22415   ,0 };
            const unsigned long dim3566KuoInit[]        =   {   1   ,   1   ,   7   ,   9   ,   31  ,   21  ,   21  ,   29  ,   341 ,   13  ,   155 ,   2161    ,   517 ,   13135   ,   15893   ,0 };
            const unsigned long dim3567KuoInit[]        =   {   1   ,   1   ,   7   ,   5   ,   21  ,   33  ,   23  ,   131 ,   367 ,   939 ,   1139    ,   865 ,   6015    ,   6945    ,   2301    ,0 };
            const unsigned long dim3568KuoInit[]        =   {   1   ,   3   ,   7   ,   3   ,   1   ,   25  ,   109 ,   183 ,   267 ,   757 ,   371 ,   1459    ,   3671    ,   2925    ,   10345   ,0 };
            const unsigned long dim3569KuoInit[]        =   {   1   ,   1   ,   5   ,   15  ,   29  ,   57  ,   95  ,   169 ,   301 ,   645 ,   5   ,   225 ,   5711    ,   305 ,   16515   ,0 };
            const unsigned long dim3570KuoInit[]        =   {   1   ,   3   ,   7   ,   9   ,   27  ,   47  ,   95  ,   121 ,   203 ,   341 ,   657 ,   2467    ,   7619    ,   9587    ,   7981    ,0 };
            const unsigned long dim3571KuoInit[]        =   {   1   ,   3   ,   1   ,   5   ,   27  ,   11  ,   103 ,   67  ,   19  ,   289 ,   187 ,   3857    ,   8117    ,   12931   ,   11279   ,0 };
            const unsigned long dim3572KuoInit[]        =   {   1   ,   1   ,   5   ,   1   ,   31  ,   9   ,   57  ,   67  ,   507 ,   405 ,   181 ,   2301    ,   715 ,   12165   ,   18415   ,0 };
            const unsigned long dim3573KuoInit[]        =   {   1   ,   1   ,   7   ,   1   ,   11  ,   3   ,   35  ,   215 ,   177 ,   679 ,   735 ,   1569    ,   6231    ,   5141    ,   28291   ,0 };
            const unsigned long dim3574KuoInit[]        =   {   1   ,   3   ,   1   ,   9   ,   7   ,   7   ,   101 ,   247 ,   103 ,   373 ,   1485    ,   2515    ,   3981    ,   12029   ,   5845    ,0 };
            const unsigned long dim3575KuoInit[]        =   {   1   ,   1   ,   1   ,   9   ,   29  ,   27  ,   117 ,   111 ,   107 ,   737 ,   629 ,   1331    ,   7495    ,   9813    ,   24925   ,0 };
            const unsigned long dim3576KuoInit[]        =   {   1   ,   1   ,   1   ,   13  ,   17  ,   11  ,   95  ,   11  ,   1   ,   525 ,   1677    ,   2673    ,   2353    ,   7211    ,   1075    ,0 };
            const unsigned long dim3577KuoInit[]        =   {   1   ,   3   ,   3   ,   7   ,   9   ,   29  ,   9   ,   125 ,   249 ,   815 ,   105 ,   1649    ,   2789    ,   6041    ,   15287   ,0 };
            const unsigned long dim3578KuoInit[]        =   {   1   ,   3   ,   3   ,   3   ,   25  ,   53  ,   87  ,   99  ,   427 ,   255 ,   285 ,   2339    ,   5003    ,   10395   ,   20969   ,0 };
            const unsigned long dim3579KuoInit[]        =   {   1   ,   3   ,   1   ,   5   ,   25  ,   33  ,   97  ,   149 ,   207 ,   925 ,   645 ,   1867    ,   4591    ,   10823   ,   16603   ,0 };
            const unsigned long dim3580KuoInit[]        =   {   1   ,   3   ,   3   ,   13  ,   1   ,   33  ,   71  ,   49  ,   417 ,   535 ,   923 ,   525 ,   811 ,   7443    ,   12971   ,0 };
            const unsigned long dim3581KuoInit[]        =   {   1   ,   3   ,   5   ,   3   ,   5   ,   9   ,   39  ,   121 ,   87  ,   883 ,   1957    ,   1375    ,   4109    ,   12691   ,   15557   ,0 };
            const unsigned long dim3582KuoInit[]        =   {   1   ,   1   ,   7   ,   9   ,   23  ,   11  ,   59  ,   41  ,   117 ,   243 ,   1953    ,   3967    ,   7303    ,   10767   ,   1345    ,0 };
            const unsigned long dim3583KuoInit[]        =   {   1   ,   3   ,   1   ,   7   ,   29  ,   59  ,   19  ,   177 ,   405 ,   197 ,   59  ,   855 ,   5009    ,   11887   ,   6361    ,0 };
            const unsigned long dim3584KuoInit[]        =   {   1   ,   3   ,   7   ,   7   ,   19  ,   59  ,   71  ,   199 ,   207 ,   757 ,   1147    ,   1857    ,   3787    ,   16081   ,   29557   ,0 };
            const unsigned long dim3585KuoInit[]        =   {   1   ,   1   ,   5   ,   9   ,   25  ,   13  ,   37  ,   27  ,   511 ,   983 ,   1087    ,   371 ,   4953    ,   8987    ,   3155    ,0 };
            const unsigned long dim3586KuoInit[]        =   {   1   ,   3   ,   7   ,   15  ,   21  ,   43  ,   75  ,   227 ,   441 ,   523 ,   551 ,   3517    ,   77  ,   5869    ,   24607   ,0 };
            const unsigned long dim3587KuoInit[]        =   {   1   ,   1   ,   5   ,   15  ,   1   ,   33  ,   25  ,   217 ,   277 ,   73  ,   105 ,   1697    ,   1545    ,   10205   ,   4275    ,0 };
            const unsigned long dim3588KuoInit[]        =   {   1   ,   1   ,   5   ,   7   ,   3   ,   53  ,   99  ,   19  ,   121 ,   427 ,   585 ,   2471    ,   103 ,   3771    ,   23525   ,0 };
            const unsigned long dim3589KuoInit[]        =   {   1   ,   3   ,   5   ,   3   ,   15  ,   17  ,   35  ,   205 ,   9   ,   941 ,   1281    ,   1339    ,   7421    ,   3341    ,   4393    ,0 };
            const unsigned long dim3590KuoInit[]        =   {   1   ,   1   ,   1   ,   13  ,   27  ,   13  ,   35  ,   99  ,   87  ,   157 ,   951 ,   2739    ,   1993    ,   6107    ,   29023   ,0 };
            const unsigned long dim3591KuoInit[]        =   {   1   ,   3   ,   5   ,   15  ,   15  ,   5   ,   83  ,   5   ,   283 ,   393 ,   1273    ,   2251    ,   4181    ,   3851    ,   6845    ,0 };
            const unsigned long dim3592KuoInit[]        =   {   1   ,   1   ,   5   ,   3   ,   1   ,   9   ,   63  ,   97  ,   83  ,   93  ,   293 ,   2087    ,   2259    ,   10313   ,   1037    ,0 };
            const unsigned long dim3593KuoInit[]        =   {   1   ,   1   ,   1   ,   5   ,   11  ,   19  ,   39  ,   19  ,   327 ,   329 ,   189 ,   3851    ,   4839    ,   1669    ,   25805   ,0 };
            const unsigned long dim3594KuoInit[]        =   {   1   ,   3   ,   3   ,   7   ,   21  ,   1   ,   41  ,   233 ,   497 ,   265 ,   1041    ,   4003    ,   8029    ,   15  ,   14437   ,0 };
            const unsigned long dim3595KuoInit[]        =   {   1   ,   3   ,   7   ,   5   ,   7   ,   15  ,   123 ,   215 ,   403 ,   931 ,   277 ,   3501    ,   3667    ,   13313   ,   23977   ,0 };
            const unsigned long dim3596KuoInit[]        =   {   1   ,   1   ,   3   ,   9   ,   5   ,   23  ,   99  ,   161 ,   25  ,   403 ,   1695    ,   2493    ,   199 ,   15745   ,   9167    ,0 };
            const unsigned long dim3597KuoInit[]        =   {   1   ,   1   ,   1   ,   15  ,   1   ,   25  ,   93  ,   31  ,   227 ,   833 ,   1891    ,   1171    ,   4551    ,   3329    ,   30593   ,0 };
            const unsigned long dim3598KuoInit[]        =   {   1   ,   3   ,   1   ,   11  ,   29  ,   51  ,   109 ,   27  ,   475 ,   511 ,   1763    ,   2285    ,   3191    ,   7943    ,   8373    ,0 };
            const unsigned long dim3599KuoInit[]        =   {   1   ,   1   ,   5   ,   3   ,   21  ,   63  ,   89  ,   9   ,   429 ,   613 ,   141 ,   4011    ,   6563    ,   5871    ,   30701   ,0 };
            const unsigned long dim3600KuoInit[]        =   {   1   ,   1   ,   5   ,   1   ,   31  ,   49  ,   87  ,   15  ,   481 ,   237 ,   133 ,   3267    ,   7385    ,   4733    ,   16853   ,0 };
            const unsigned long dim3601KuoInit[]        =   {   1   ,   3   ,   7   ,   9   ,   1   ,   63  ,   95  ,   17  ,   311 ,   29  ,   1543    ,   2971    ,   695 ,   3641    ,   28155   ,0 };
            const unsigned long dim3602KuoInit[]        =   {   1   ,   3   ,   3   ,   7   ,   27  ,   1   ,   59  ,   181 ,   321 ,   431 ,   143 ,   213 ,   1471    ,   4707    ,   22897   ,0 };
            const unsigned long dim3603KuoInit[]        =   {   1   ,   1   ,   3   ,   11  ,   19  ,   3   ,   95  ,   1   ,   407 ,   755 ,   1581    ,   3435    ,   2025    ,   7781    ,   5573    ,0 };
            const unsigned long dim3604KuoInit[]        =   {   1   ,   1   ,   1   ,   1   ,   19  ,   15  ,   119 ,   169 ,   109 ,   1003    ,   403 ,   1633    ,   4963    ,   9785    ,   18483   ,0 };
            const unsigned long dim3605KuoInit[]        =   {   1   ,   3   ,   7   ,   1   ,   23  ,   5   ,   77  ,   79  ,   429 ,   921 ,   1913    ,   2075    ,   7923    ,   5515    ,   19257   ,0 };
            const unsigned long dim3606KuoInit[]        =   {   1   ,   3   ,   1   ,   11  ,   1   ,   45  ,   99  ,   249 ,   385 ,   427 ,   1201    ,   1363    ,   335 ,   12561   ,   18753   ,0 };
            const unsigned long dim3607KuoInit[]        =   {   1   ,   1   ,   1   ,   11  ,   27  ,   41  ,   9   ,   13  ,   267 ,   987 ,   1023    ,   2723    ,   3873    ,   15911   ,   3165    ,0 };
            const unsigned long dim3608KuoInit[]        =   {   1   ,   3   ,   5   ,   1   ,   17  ,   7   ,   117 ,   241 ,   509 ,   757 ,   1301    ,   3935    ,   3361    ,   13745   ,   27239   ,0 };
            const unsigned long dim3609KuoInit[]        =   {   1   ,   1   ,   5   ,   15  ,   21  ,   33  ,   55  ,   245 ,   1   ,   263 ,   437 ,   2279    ,   1201    ,   2043    ,   20473   ,0 };
            const unsigned long dim3610KuoInit[]        =   {   1   ,   3   ,   7   ,   5   ,   29  ,   9   ,   31  ,   165 ,   299 ,   923 ,   847 ,   769 ,   5683    ,   1805    ,   13433   ,0 };
            const unsigned long dim3611KuoInit[]        =   {   1   ,   1   ,   1   ,   11  ,   29  ,   13  ,   49  ,   173 ,   269 ,   35  ,   1057    ,   973 ,   1853    ,   12677   ,   13565   ,0 };
            const unsigned long dim3612KuoInit[]        =   {   1   ,   3   ,   1   ,   11  ,   11  ,   21  ,   19  ,   3   ,   127 ,   133 ,   193 ,   2665    ,   7187    ,   15425   ,   23805   ,0 };
            const unsigned long dim3613KuoInit[]        =   {   1   ,   3   ,   5   ,   7   ,   13  ,   61  ,   95  ,   17  ,   197 ,   381 ,   865 ,   1241    ,   685 ,   9525    ,   2503    ,0 };
            const unsigned long dim3614KuoInit[]        =   {   1   ,   1   ,   5   ,   7   ,   13  ,   19  ,   83  ,   141 ,   179 ,   201 ,   463 ,   2083    ,   3617    ,   1945    ,   20371   ,0 };
            const unsigned long dim3615KuoInit[]        =   {   1   ,   3   ,   5   ,   13  ,   9   ,   39  ,   47  ,   3   ,   13  ,   739 ,   1501    ,   1007    ,   6963    ,   15819   ,   29739   ,0 };
            const unsigned long dim3616KuoInit[]        =   {   1   ,   1   ,   5   ,   7   ,   3   ,   17  ,   105 ,   27  ,   165 ,   515 ,   949 ,   81  ,   947 ,   9615    ,   24403   ,0 };
            const unsigned long dim3617KuoInit[]        =   {   1   ,   3   ,   7   ,   5   ,   23  ,   43  ,   81  ,   141 ,   181 ,   1007    ,   1343    ,   1919    ,   5971    ,   8563    ,   17091   ,0 };
            const unsigned long dim3618KuoInit[]        =   {   1   ,   3   ,   3   ,   11  ,   1   ,   25  ,   19  ,   117 ,   225 ,   347 ,   785 ,   791 ,   7321    ,   15647   ,   25491   ,0 };
            const unsigned long dim3619KuoInit[]        =   {   1   ,   1   ,   5   ,   5   ,   11  ,   61  ,   25  ,   205 ,   145 ,   695 ,   713 ,   1787    ,   8087    ,   5935    ,   5105    ,0 };
            const unsigned long dim3620KuoInit[]        =   {   1   ,   1   ,   7   ,   1   ,   15  ,   17  ,   25  ,   141 ,   75  ,   657 ,   1641    ,   575 ,   3387    ,   6093    ,   15565   ,0 };
            const unsigned long dim3621KuoInit[]        =   {   1   ,   3   ,   7   ,   11  ,   31  ,   49  ,   113 ,   37  ,   321 ,   547 ,   819 ,   2079    ,   7231    ,   7497    ,   29877   ,0 };
            const unsigned long dim3622KuoInit[]        =   {   1   ,   1   ,   3   ,   11  ,   31  ,   19  ,   39  ,   99  ,   271 ,   527 ,   1551    ,   2491    ,   5515    ,   5541    ,   6131    ,0 };
            const unsigned long dim3623KuoInit[]        =   {   1   ,   1   ,   1   ,   7   ,   23  ,   35  ,   31  ,   35  ,   315 ,   911 ,   1133    ,   3509    ,   1381    ,   15869   ,   23211   ,0 };
            const unsigned long dim3624KuoInit[]        =   {   1   ,   3   ,   3   ,   5   ,   1   ,   33  ,   127 ,   175 ,   311 ,   497 ,   187 ,   2983    ,   6553    ,   9653    ,   5711    ,0 };
            const unsigned long dim3625KuoInit[]        =   {   1   ,   1   ,   3   ,   13  ,   19  ,   3   ,   65  ,   91  ,   265 ,   199 ,   2035    ,   1277    ,   4575    ,   5517    ,   15247   ,0 };
            const unsigned long dim3626KuoInit[]        =   {   1   ,   1   ,   5   ,   15  ,   31  ,   39  ,   71  ,   167 ,   337 ,   245 ,   399 ,   2681    ,   1257    ,   15117   ,   24145   ,0 };
            const unsigned long dim3627KuoInit[]        =   {   1   ,   1   ,   1   ,   11  ,   7   ,   19  ,   113 ,   133 ,   83  ,   851 ,   1543    ,   1699    ,   4527    ,   3167    ,   10793   ,0 };
            const unsigned long dim3628KuoInit[]        =   {   1   ,   3   ,   3   ,   3   ,   31  ,   55  ,   35  ,   161 ,   133 ,   83  ,   469 ,   3011    ,   4513    ,   4677    ,   2363    ,0 };
            const unsigned long dim3629KuoInit[]        =   {   1   ,   1   ,   3   ,   1   ,   29  ,   51  ,   29  ,   163 ,   95  ,   221 ,   973 ,   2819    ,   3103    ,   9969    ,   10347   ,0 };
            const unsigned long dim3630KuoInit[]        =   {   1   ,   1   ,   5   ,   7   ,   31  ,   5   ,   73  ,   159 ,   339 ,   523 ,   399 ,   4041    ,   4341    ,   11527   ,   505 ,0 };
            const unsigned long dim3631KuoInit[]        =   {   1   ,   1   ,   3   ,   7   ,   15  ,   55  ,   91  ,   117 ,   233 ,   497 ,   485 ,   1187    ,   5185    ,   13127   ,   14393   ,0 };
            const unsigned long dim3632KuoInit[]        =   {   1   ,   3   ,   7   ,   9   ,   27  ,   19  ,   23  ,   197 ,   99  ,   779 ,   423 ,   2735    ,   3513    ,   8227    ,   31127   ,0 };
            const unsigned long dim3633KuoInit[]        =   {   1   ,   3   ,   3   ,   1   ,   3   ,   61  ,   89  ,   137 ,   189 ,   325 ,   1797    ,   2287    ,   5033    ,   7351    ,   1131    ,0 };
            const unsigned long dim3634KuoInit[]        =   {   1   ,   3   ,   3   ,   15  ,   19  ,   9   ,   83  ,   87  ,   451 ,   317 ,   667 ,   3379    ,   2225    ,   6437    ,   31515   ,0 };
            const unsigned long dim3635KuoInit[]        =   {   1   ,   1   ,   1   ,   15  ,   13  ,   31  ,   121 ,   131 ,   473 ,   717 ,   1507    ,   1483    ,   4819    ,   14375   ,   24667   ,0 };
            const unsigned long dim3636KuoInit[]        =   {   1   ,   1   ,   1   ,   5   ,   5   ,   55  ,   101 ,   111 ,   89  ,   481 ,   2031    ,   3043    ,   5727    ,   1253    ,   11951   ,0 };
            const unsigned long dim3637KuoInit[]        =   {   1   ,   3   ,   1   ,   1   ,   3   ,   13  ,   67  ,   129 ,   311 ,   125 ,   121 ,   2833    ,   3555    ,   7761    ,   27343   ,0 };
            const unsigned long dim3638KuoInit[]        =   {   1   ,   3   ,   7   ,   7   ,   29  ,   37  ,   63  ,   59  ,   5   ,   691 ,   1625    ,   97  ,   6113    ,   14795   ,   31471   ,0 };
            const unsigned long dim3639KuoInit[]        =   {   1   ,   3   ,   3   ,   3   ,   25  ,   7   ,   65  ,   207 ,   371 ,   29  ,   1627    ,   1785    ,   5203    ,   5153    ,   1461    ,0 };
            const unsigned long dim3640KuoInit[]        =   {   1   ,   3   ,   3   ,   11  ,   1   ,   37  ,   121 ,   249 ,   381 ,   309 ,   1545    ,   4039    ,   5657    ,   6783    ,   1889    ,0 };
            const unsigned long dim3641KuoInit[]        =   {   1   ,   1   ,   1   ,   9   ,   29  ,   19  ,   51  ,   213 ,   185 ,   481 ,   1929    ,   2475    ,   567 ,   10045   ,   32619   ,0 };
            const unsigned long dim3642KuoInit[]        =   {   1   ,   1   ,   3   ,   7   ,   13  ,   13  ,   17  ,   47  ,   55  ,   521 ,   473 ,   495 ,   7807    ,   1645    ,   17713   ,0 };
            const unsigned long dim3643KuoInit[]        =   {   1   ,   1   ,   1   ,   3   ,   5   ,   63  ,   119 ,   167 ,   357 ,   231 ,   1915    ,   3605    ,   3579    ,   3041    ,   603 ,0 };
            const unsigned long dim3644KuoInit[]        =   {   1   ,   3   ,   3   ,   7   ,   31  ,   5   ,   93  ,   193 ,   175 ,   801 ,   1963    ,   1413    ,   1641    ,   4073    ,   16017   ,0 };
            const unsigned long dim3645KuoInit[]        =   {   1   ,   3   ,   3   ,   11  ,   11  ,   49  ,   53  ,   57  ,   1   ,   799 ,   301 ,   2611    ,   2975    ,   7595    ,   25891   ,0 };
            const unsigned long dim3646KuoInit[]        =   {   1   ,   3   ,   1   ,   3   ,   19  ,   41  ,   63  ,   173 ,   257 ,   903 ,   1977    ,   3799    ,   3985    ,   14807   ,   24903   ,0 };
            const unsigned long dim3647KuoInit[]        =   {   1   ,   3   ,   3   ,   11  ,   17  ,   53  ,   61  ,   115 ,   453 ,   395 ,   993 ,   3365    ,   4071    ,   4737    ,   29469   ,0 };
            const unsigned long dim3648KuoInit[]        =   {   1   ,   3   ,   1   ,   13  ,   3   ,   55  ,   109 ,   149 ,   313 ,   621 ,   345 ,   4037    ,   677 ,   3787    ,   6701    ,0 };
            const unsigned long dim3649KuoInit[]        =   {   1   ,   3   ,   3   ,   9   ,   23  ,   11  ,   63  ,   21  ,   47  ,   949 ,   47  ,   3055    ,   3403    ,   7581    ,   12853   ,0 };
            const unsigned long dim3650KuoInit[]        =   {   1   ,   1   ,   1   ,   9   ,   7   ,   21  ,   15  ,   203 ,   167 ,   677 ,   49  ,   2777    ,   7817    ,   13207   ,   31485   ,0 };
            const unsigned long dim3651KuoInit[]        =   {   1   ,   1   ,   1   ,   3   ,   29  ,   41  ,   37  ,   61  ,   343 ,   279 ,   127 ,   925 ,   7091    ,   14477   ,   24149   ,0 };
            const unsigned long dim3652KuoInit[]        =   {   1   ,   1   ,   1   ,   11  ,   3   ,   33  ,   35  ,   237 ,   133 ,   653 ,   1491    ,   2183    ,   3565    ,   14773   ,   4521    ,0 };
            const unsigned long dim3653KuoInit[]        =   {   1   ,   1   ,   5   ,   3   ,   31  ,   43  ,   13  ,   79  ,   465 ,   785 ,   1913    ,   1133    ,   2701    ,   9007    ,   15695   ,0 };
            const unsigned long dim3654KuoInit[]        =   {   1   ,   3   ,   3   ,   1   ,   23  ,   43  ,   5   ,   155 ,   369 ,   131 ,   671 ,   1109    ,   7951    ,   1929    ,   23965   ,0 };
            const unsigned long dim3655KuoInit[]        =   {   1   ,   1   ,   5   ,   7   ,   9   ,   15  ,   23  ,   59  ,   511 ,   931 ,   259 ,   1953    ,   1197    ,   9441    ,   25545   ,0 };
            const unsigned long dim3656KuoInit[]        =   {   1   ,   1   ,   7   ,   7   ,   27  ,   25  ,   51  ,   81  ,   479 ,   491 ,   203 ,   4009    ,   4621    ,   11153   ,   12049   ,0 };
            const unsigned long dim3657KuoInit[]        =   {   1   ,   3   ,   5   ,   15  ,   25  ,   11  ,   77  ,   253 ,   249 ,   761 ,   707 ,   161 ,   7519    ,   5381    ,   1549    ,0 };
            const unsigned long dim3658KuoInit[]        =   {   1   ,   1   ,   1   ,   13  ,   11  ,   39  ,   59  ,   191 ,   443 ,   471 ,   53  ,   1639    ,   505 ,   1281    ,   1919    ,0 };
            const unsigned long dim3659KuoInit[]        =   {   1   ,   1   ,   3   ,   7   ,   29  ,   47  ,   121 ,   135 ,   411 ,   95  ,   1255    ,   635 ,   835 ,   7905    ,   18435   ,0 };
            const unsigned long dim3660KuoInit[]        =   {   1   ,   1   ,   3   ,   15  ,   21  ,   57  ,   65  ,   75  ,   333 ,   61  ,   1861    ,   3903    ,   6051    ,   8311    ,   2791    ,0 };
            const unsigned long dim3661KuoInit[]        =   {   1   ,   3   ,   3   ,   7   ,   7   ,   49  ,   121 ,   233 ,   305 ,   697 ,   525 ,   3921    ,   5747    ,   2951    ,   24321   ,0 };
            const unsigned long dim3662KuoInit[]        =   {   1   ,   1   ,   3   ,   3   ,   9   ,   23  ,   53  ,   121 ,   187 ,   35  ,   1973    ,   829 ,   575 ,   4003    ,   32159   ,0 };
            const unsigned long dim3663KuoInit[]        =   {   1   ,   1   ,   3   ,   1   ,   3   ,   57  ,   25  ,   99  ,   383 ,   97  ,   1751    ,   1799    ,   2001    ,   8335    ,   20349   ,0 };
            const unsigned long dim3664KuoInit[]        =   {   1   ,   1   ,   7   ,   9   ,   11  ,   17  ,   25  ,   107 ,   483 ,   183 ,   559 ,   2531    ,   6509    ,   9449    ,   1793    ,0 };
            const unsigned long dim3665KuoInit[]        =   {   1   ,   3   ,   3   ,   1   ,   3   ,   43  ,   89  ,   57  ,   15  ,   269 ,   521 ,   1119    ,   3573    ,   9801    ,   30943   ,0 };
            const unsigned long dim3666KuoInit[]        =   {   1   ,   3   ,   7   ,   3   ,   7   ,   41  ,   101 ,   177 ,   47  ,   47  ,   2009    ,   2127    ,   5113    ,   13541   ,   4133    ,0 };
            const unsigned long dim3667KuoInit[]        =   {   1   ,   1   ,   3   ,   7   ,   23  ,   63  ,   67  ,   73  ,   335 ,   471 ,   1137    ,   3101    ,   5539    ,   16203   ,   15395   ,   12455   ,0 };
            const unsigned long dim3668KuoInit[]        =   {   1   ,   3   ,   1   ,   11  ,   5   ,   9   ,   43  ,   123 ,   91  ,   23  ,   1199    ,   287 ,   5343    ,   1937    ,   12115   ,   4749    ,0 };
            const unsigned long dim3669KuoInit[]        =   {   1   ,   1   ,   7   ,   11  ,   27  ,   3   ,   29  ,   251 ,   413 ,   147 ,   403 ,   2065    ,   2695    ,   11657   ,   15519   ,   49391   ,0 };
            const unsigned long dim3670KuoInit[]        =   {   1   ,   3   ,   7   ,   11  ,   25  ,   7   ,   125 ,   137 ,   479 ,   753 ,   309 ,   1075    ,   7991    ,   2629    ,   4039    ,   6749    ,0 };
            const unsigned long dim3671KuoInit[]        =   {   1   ,   1   ,   7   ,   1   ,   3   ,   27  ,   69  ,   227 ,   503 ,   915 ,   569 ,   115 ,   4299    ,   10799   ,   13745   ,   7973    ,0 };
            const unsigned long dim3672KuoInit[]        =   {   1   ,   3   ,   1   ,   11  ,   17  ,   49  ,   33  ,   161 ,   429 ,   153 ,   1179    ,   4005    ,   6507    ,   10507   ,   1779    ,   5331    ,0 };
            const unsigned long dim3673KuoInit[]        =   {   1   ,   3   ,   1   ,   13  ,   25  ,   45  ,   117 ,   47  ,   253 ,   851 ,   627 ,   593 ,   7555    ,   3677    ,   30695   ,   38851   ,0 };
            const unsigned long dim3674KuoInit[]        =   {   1   ,   3   ,   1   ,   11  ,   25  ,   19  ,   95  ,   191 ,   505 ,   683 ,   67  ,   2375    ,   995 ,   8647    ,   4853    ,   17483   ,0 };
            const unsigned long dim3675KuoInit[]        =   {   1   ,   3   ,   3   ,   15  ,   7   ,   57  ,   101 ,   151 ,   477 ,   343 ,   879 ,   2161    ,   1585    ,   1903    ,   6459    ,   45285   ,0 };
            const unsigned long dim3676KuoInit[]        =   {   1   ,   1   ,   5   ,   3   ,   7   ,   7   ,   31  ,   219 ,   137 ,   107 ,   119 ,   2959    ,   7453    ,   9215    ,   20463   ,   61165   ,0 };
            const unsigned long dim3677KuoInit[]        =   {   1   ,   1   ,   3   ,   15  ,   25  ,   39  ,   41  ,   165 ,   421 ,   567 ,   1095    ,   4055    ,   7321    ,   831 ,   81  ,   50377   ,0 };
            const unsigned long dim3678KuoInit[]        =   {   1   ,   1   ,   3   ,   1   ,   21  ,   23  ,   65  ,   31  ,   331 ,   875 ,   1449    ,   653 ,   1975    ,   783 ,   29021   ,   47225   ,0 };
            const unsigned long dim3679KuoInit[]        =   {   1   ,   1   ,   7   ,   1   ,   21  ,   21  ,   39  ,   39  ,   289 ,   129 ,   345 ,   2983    ,   2419    ,   133 ,   2519    ,   33307   ,0 };
            const unsigned long dim3680KuoInit[]        =   {   1   ,   3   ,   5   ,   1   ,   27  ,   23  ,   53  ,   63  ,   315 ,   93  ,   335 ,   3239    ,   6749    ,   12985   ,   5863    ,   49565   ,0 };
            const unsigned long dim3681KuoInit[]        =   {   1   ,   3   ,   7   ,   11  ,   17  ,   23  ,   89  ,   253 ,   329 ,   837 ,   589 ,   1135    ,   4099    ,   3123    ,   25567   ,   21017   ,0 };
            const unsigned long dim3682KuoInit[]        =   {   1   ,   3   ,   7   ,   3   ,   21  ,   17  ,   111 ,   21  ,   73  ,   901 ,   2017    ,   3711    ,   159 ,   6703    ,   32209   ,   34411   ,0 };
            const unsigned long dim3683KuoInit[]        =   {   1   ,   1   ,   1   ,   9   ,   19  ,   59  ,   1   ,   31  ,   273 ,   683 ,   1155    ,   2543    ,   477 ,   8479    ,   25235   ,   61595   ,0 };
            const unsigned long dim3684KuoInit[]        =   {   1   ,   1   ,   7   ,   1   ,   13  ,   17  ,   97  ,   49  ,   221 ,   63  ,   197 ,   465 ,   1801    ,   5389    ,   10395   ,   54157   ,0 };
            const unsigned long dim3685KuoInit[]        =   {   1   ,   1   ,   5   ,   9   ,   11  ,   59  ,   79  ,   139 ,   249 ,   185 ,   1131    ,   507 ,   3325    ,   2713    ,   27293   ,   21539   ,0 };
            const unsigned long dim3686KuoInit[]        =   {   1   ,   1   ,   3   ,   5   ,   9   ,   45  ,   103 ,   145 ,   327 ,   675 ,   1609    ,   1703    ,   761 ,   4947    ,   187 ,   53903   ,0 };
            const unsigned long dim3687KuoInit[]        =   {   1   ,   1   ,   1   ,   9   ,   19  ,   13  ,   77  ,   143 ,   303 ,   119 ,   1985    ,   1347    ,   6661    ,   6081    ,   9723    ,   52129   ,0 };
            const unsigned long dim3688KuoInit[]        =   {   1   ,   3   ,   3   ,   3   ,   27  ,   27  ,   15  ,   163 ,   139 ,   159 ,   1921    ,   3549    ,   6477    ,   11423   ,   16055   ,   32399   ,0 };
            const unsigned long dim3689KuoInit[]        =   {   1   ,   3   ,   1   ,   13  ,   19  ,   59  ,   5   ,   73  ,   481 ,   945 ,   187 ,   1223    ,   1771    ,   12961   ,   13853   ,   14513   ,0 };
            const unsigned long dim3690KuoInit[]        =   {   1   ,   1   ,   7   ,   11  ,   5   ,   63  ,   33  ,   185 ,   477 ,   857 ,   1057    ,   2509    ,   7341    ,   7669    ,   13829   ,   51965   ,0 };
            const unsigned long dim3691KuoInit[]        =   {   1   ,   1   ,   3   ,   5   ,   23  ,   17  ,   7   ,   87  ,   101 ,   817 ,   241 ,   853 ,   4067    ,   6409    ,   28217   ,   54395   ,0 };
            const unsigned long dim3692KuoInit[]        =   {   1   ,   3   ,   3   ,   3   ,   21  ,   55  ,   47  ,   159 ,   407 ,   675 ,   1805    ,   2079    ,   1113    ,   149 ,   26071   ,   30829   ,0 };
            const unsigned long dim3693KuoInit[]        =   {   1   ,   1   ,   3   ,   13  ,   7   ,   3   ,   89  ,   161 ,   95  ,   333 ,   525 ,   3979    ,   6831    ,   8007    ,   7063    ,   45299   ,0 };
            const unsigned long dim3694KuoInit[]        =   {   1   ,   1   ,   7   ,   5   ,   17  ,   37  ,   95  ,   129 ,   215 ,   313 ,   1395    ,   2415    ,   6191    ,   6409    ,   29717   ,   4199    ,0 };
            const unsigned long dim3695KuoInit[]        =   {   1   ,   1   ,   1   ,   7   ,   7   ,   45  ,   75  ,   175 ,   407 ,   763 ,   1037    ,   2477    ,   3167    ,   6991    ,   23905   ,   14707   ,0 };
            const unsigned long dim3696KuoInit[]        =   {   1   ,   3   ,   7   ,   7   ,   23  ,   23  ,   103 ,   217 ,   479 ,   779 ,   941 ,   3173    ,   3639    ,   12833   ,   7133    ,   39269   ,0 };
            const unsigned long dim3697KuoInit[]        =   {   1   ,   1   ,   1   ,   5   ,   9   ,   55  ,   19  ,   147 ,   287 ,   291 ,   1751    ,   2475    ,   6639    ,   5933    ,   27837   ,   33713   ,0 };
            const unsigned long dim3698KuoInit[]        =   {   1   ,   1   ,   5   ,   5   ,   11  ,   39  ,   65  ,   109 ,   95  ,   637 ,   1897    ,   2395    ,   6975    ,   12431   ,   25097   ,   15649   ,0 };
            const unsigned long dim3699KuoInit[]        =   {   1   ,   1   ,   3   ,   7   ,   25  ,   37  ,   27  ,   53  ,   373 ,   881 ,   1221    ,   2127    ,   1775    ,   2025    ,   10249   ,   55325   ,0 };
            const unsigned long dim3700KuoInit[]        =   {   1   ,   1   ,   7   ,   15  ,   5   ,   37  ,   1   ,   99  ,   249 ,   209 ,   1177    ,   2953    ,   4455    ,   12125   ,   7679    ,   25543   ,0 };
            const unsigned long dim3701KuoInit[]        =   {   1   ,   3   ,   3   ,   13  ,   9   ,   19  ,   97  ,   11  ,   331 ,   107 ,   1475    ,   769 ,   1749    ,   9375    ,   19477   ,   47559   ,0 };
            const unsigned long dim3702KuoInit[]        =   {   1   ,   3   ,   7   ,   15  ,   25  ,   57  ,   69  ,   45  ,   21  ,   455 ,   1841    ,   1101    ,   2931    ,   2845    ,   21925   ,   43731   ,0 };
            const unsigned long dim3703KuoInit[]        =   {   1   ,   1   ,   3   ,   11  ,   27  ,   33  ,   31  ,   33  ,   9   ,   825 ,   883 ,   2259    ,   7317    ,   5507    ,   3931    ,   61075   ,0 };
            const unsigned long dim3704KuoInit[]        =   {   1   ,   3   ,   7   ,   5   ,   7   ,   21  ,   97  ,   203 ,   449 ,   671 ,   1037    ,   3985    ,   5655    ,   11099   ,   21981   ,   50149   ,0 };
            const unsigned long dim3705KuoInit[]        =   {   1   ,   1   ,   1   ,   1   ,   5   ,   19  ,   109 ,   247 ,   73  ,   981 ,   1981    ,   3613    ,   6575    ,   3493    ,   9553    ,   259 ,0 };
            const unsigned long dim3706KuoInit[]        =   {   1   ,   1   ,   7   ,   9   ,   3   ,   55  ,   13  ,   175 ,   285 ,   865 ,   1817    ,   911 ,   6393    ,   4843    ,   31569   ,   30115   ,0 };
            const unsigned long dim3707KuoInit[]        =   {   1   ,   1   ,   3   ,   15  ,   21  ,   51  ,   85  ,   239 ,   171 ,   741 ,   881 ,   749 ,   4021    ,   2049    ,   25495   ,   62913   ,0 };
            const unsigned long dim3708KuoInit[]        =   {   1   ,   3   ,   1   ,   1   ,   19  ,   9   ,   111 ,   119 ,   171 ,   903 ,   567 ,   1513    ,   8179    ,   10553   ,   5315    ,   63291   ,0 };
            const unsigned long dim3709KuoInit[]        =   {   1   ,   3   ,   5   ,   1   ,   1   ,   61  ,   21  ,   197 ,   165 ,   371 ,   3   ,   2085    ,   669 ,   4497    ,   27467   ,   52995   ,0 };
            const unsigned long dim3710KuoInit[]        =   {   1   ,   3   ,   5   ,   13  ,   29  ,   37  ,   95  ,   247 ,   161 ,   621 ,   83  ,   2597    ,   6779    ,   13747   ,   18873   ,   6173    ,0 };
            const unsigned long dim3711KuoInit[]        =   {   1   ,   1   ,   3   ,   5   ,   21  ,   5   ,   41  ,   137 ,   77  ,   829 ,   689 ,   563 ,   5331    ,   8061    ,   21371   ,   58565   ,0 };
            const unsigned long dim3712KuoInit[]        =   {   1   ,   3   ,   7   ,   13  ,   1   ,   19  ,   11  ,   111 ,   35  ,   39  ,   1513    ,   1241    ,   3739    ,   16191   ,   25097   ,   25947   ,0 };
            const unsigned long dim3713KuoInit[]        =   {   1   ,   1   ,   5   ,   1   ,   27  ,   45  ,   11  ,   29  ,   153 ,   321 ,   1657    ,   1419    ,   6983    ,   11935   ,   8295    ,   24313   ,0 };
            const unsigned long dim3714KuoInit[]        =   {   1   ,   3   ,   3   ,   11  ,   1   ,   57  ,   109 ,   139 ,   53  ,   883 ,   81  ,   333 ,   5133    ,   13891   ,   7775    ,   13401   ,0 };
            const unsigned long dim3715KuoInit[]        =   {   1   ,   3   ,   5   ,   15  ,   5   ,   35  ,   121 ,   15  ,   73  ,   313 ,   1903    ,   437 ,   1567    ,   11905   ,   7679    ,   14265   ,0 };
            const unsigned long dim3716KuoInit[]        =   {   1   ,   3   ,   7   ,   3   ,   31  ,   63  ,   69  ,   135 ,   75  ,   489 ,   1271    ,   585 ,   7361    ,   15965   ,   3099    ,   37309   ,0 };
            const unsigned long dim3717KuoInit[]        =   {   1   ,   1   ,   7   ,   11  ,   15  ,   21  ,   97  ,   73  ,   115 ,   795 ,   421 ,   69  ,   7795    ,   9895    ,   1755    ,   44375   ,0 };
            const unsigned long dim3718KuoInit[]        =   {   1   ,   3   ,   5   ,   9   ,   17  ,   13  ,   49  ,   61  ,   395 ,   637 ,   1737    ,   3103    ,   5197    ,   7799    ,   25425   ,   56313   ,0 };
            const unsigned long dim3719KuoInit[]        =   {   1   ,   3   ,   1   ,   9   ,   17  ,   19  ,   87  ,   131 ,   315 ,   865 ,   813 ,   163 ,   1045    ,   12573   ,   12657   ,   43715   ,0 };
            const unsigned long dim3720KuoInit[]        =   {   1   ,   3   ,   7   ,   3   ,   27  ,   23  ,   9   ,   185 ,   291 ,   421 ,   603 ,   1339    ,   1085    ,   15449   ,   1543    ,   5815    ,0 };
            const unsigned long dim3721KuoInit[]        =   {   1   ,   3   ,   1   ,   5   ,   29  ,   35  ,   15  ,   75  ,   135 ,   911 ,   1153    ,   3067    ,   3475    ,   11653   ,   27485   ,   8459    ,0 };
            const unsigned long dim3722KuoInit[]        =   {   1   ,   3   ,   5   ,   7   ,   3   ,   9   ,   43  ,   29  ,   77  ,   551 ,   973 ,   2165    ,   1499    ,   9185    ,   27025   ,   31605   ,0 };
            const unsigned long dim3723KuoInit[]        =   {   1   ,   1   ,   1   ,   9   ,   23  ,   49  ,   13  ,   229 ,   295 ,   497 ,   187 ,   1995    ,   1883    ,   9109    ,   2689    ,   37275   ,0 };
            const unsigned long dim3724KuoInit[]        =   {   1   ,   1   ,   7   ,   15  ,   11  ,   1   ,   61  ,   183 ,   129 ,   869 ,   1271    ,   3439    ,   7743    ,   5831    ,   12829   ,   1341    ,0 };
            const unsigned long dim3725KuoInit[]        =   {   1   ,   3   ,   7   ,   1   ,   17  ,   45  ,   31  ,   29  ,   441 ,   533 ,   1479    ,   3519    ,   2083    ,   149 ,   22119   ,   37615   ,0 };
            const unsigned long dim3726KuoInit[]        =   {   1   ,   1   ,   3   ,   7   ,   31  ,   39  ,   101 ,   209 ,   237 ,   211 ,   1015    ,   1339    ,   857 ,   3833    ,   16875   ,   38147   ,0 };
            const unsigned long dim3727KuoInit[]        =   {   1   ,   3   ,   7   ,   13  ,   29  ,   23  ,   9   ,   79  ,   71  ,   363 ,   487 ,   595 ,   259 ,   12425   ,   28181   ,   20717   ,0 };
            const unsigned long dim3728KuoInit[]        =   {   1   ,   3   ,   3   ,   11  ,   1   ,   21  ,   97  ,   39  ,   101 ,   1   ,   121 ,   3083    ,   109 ,   7889    ,   29005   ,   57081   ,0 };
            const unsigned long dim3729KuoInit[]        =   {   1   ,   3   ,   3   ,   15  ,   19  ,   19  ,   91  ,   17  ,   281 ,   121 ,   1653    ,   3099    ,   181 ,   15911   ,   29553   ,   10047   ,0 };
            const unsigned long dim3730KuoInit[]        =   {   1   ,   1   ,   3   ,   9   ,   27  ,   39  ,   69  ,   57  ,   25  ,   595 ,   723 ,   1779    ,   7537    ,   9397    ,   5563    ,   8197    ,0 };
            const unsigned long dim3731KuoInit[]        =   {   1   ,   1   ,   1   ,   1   ,   11  ,   61  ,   127 ,   81  ,   439 ,   723 ,   1147    ,   3639    ,   1207    ,   5127    ,   10209   ,   33629   ,0 };
            const unsigned long dim3732KuoInit[]        =   {   1   ,   3   ,   1   ,   9   ,   27  ,   55  ,   69  ,   147 ,   477 ,   663 ,   833 ,   3113    ,   2585    ,   9497    ,   679 ,   62759   ,0 };
            const unsigned long dim3733KuoInit[]        =   {   1   ,   3   ,   1   ,   15  ,   9   ,   51  ,   67  ,   191 ,   105 ,   3   ,   1439    ,   1915    ,   3059    ,   3001    ,   4035    ,   551 ,0 };
            const unsigned long dim3734KuoInit[]        =   {   1   ,   1   ,   3   ,   15  ,   1   ,   3   ,   11  ,   171 ,   189 ,   659 ,   1279    ,   3995    ,   6729    ,   6187    ,   22213   ,   22709   ,0 };
            const unsigned long dim3735KuoInit[]        =   {   1   ,   1   ,   7   ,   11  ,   27  ,   31  ,   3   ,   183 ,   209 ,   931 ,   99  ,   1901    ,   113 ,   15779   ,   11525   ,   29949   ,0 };
            const unsigned long dim3736KuoInit[]        =   {   1   ,   3   ,   7   ,   13  ,   11  ,   51  ,   1   ,   173 ,   271 ,   785 ,   1875    ,   2115    ,   1977    ,   189 ,   7819    ,   8133    ,0 };
            const unsigned long dim3737KuoInit[]        =   {   1   ,   3   ,   7   ,   7   ,   21  ,   9   ,   5   ,   253 ,   495 ,   415 ,   3   ,   1979    ,   2331    ,   12229   ,   29901   ,   7115    ,0 };
            const unsigned long dim3738KuoInit[]        =   {   1   ,   3   ,   1   ,   15  ,   21  ,   11  ,   105 ,   81  ,   269 ,   347 ,   901 ,   81  ,   7077    ,   11813   ,   18739   ,   57959   ,0 };
            const unsigned long dim3739KuoInit[]        =   {   1   ,   1   ,   1   ,   1   ,   19  ,   57  ,   17  ,   87  ,   325 ,   129 ,   1907    ,   1641    ,   6247    ,   12073   ,   31629   ,   28503   ,0 };
            const unsigned long dim3740KuoInit[]        =   {   1   ,   1   ,   5   ,   15  ,   29  ,   27  ,   81  ,   11  ,   383 ,   881 ,   1385    ,   3787    ,   503 ,   1369    ,   25031   ,   42049   ,0 };
            const unsigned long dim3741KuoInit[]        =   {   1   ,   3   ,   7   ,   15  ,   27  ,   15  ,   13  ,   13  ,   279 ,   785 ,   163 ,   1907    ,   4523    ,   12133   ,   1497    ,   18845   ,0 };
            const unsigned long dim3742KuoInit[]        =   {   1   ,   3   ,   7   ,   9   ,   13  ,   37  ,   51  ,   83  ,   309 ,   645 ,   383 ,   1607    ,   2087    ,   14287   ,   14375   ,   39619   ,0 };
            const unsigned long dim3743KuoInit[]        =   {   1   ,   3   ,   7   ,   5   ,   11  ,   13  ,   21  ,   45  ,   95  ,   667 ,   1597    ,   1823    ,   5887    ,   10909   ,   26279   ,   29845   ,0 };
            const unsigned long dim3744KuoInit[]        =   {   1   ,   3   ,   1   ,   1   ,   25  ,   61  ,   21  ,   91  ,   81  ,   173 ,   1271    ,   339 ,   2185    ,   16033   ,   3687    ,   53797   ,0 };
            const unsigned long dim3745KuoInit[]        =   {   1   ,   3   ,   3   ,   15  ,   19  ,   27  ,   59  ,   153 ,   453 ,   523 ,   85  ,   3367    ,   4439    ,   13605   ,   6219    ,   25593   ,0 };
            const unsigned long dim3746KuoInit[]        =   {   1   ,   3   ,   5   ,   3   ,   27  ,   57  ,   127 ,   225 ,   199 ,   823 ,   431 ,   2911    ,   4421    ,   14063   ,   1467    ,   46787   ,0 };
            const unsigned long dim3747KuoInit[]        =   {   1   ,   3   ,   5   ,   1   ,   31  ,   33  ,   17  ,   73  ,   357 ,   199 ,   677 ,   2015    ,   1213    ,   8503    ,   4785    ,   54835   ,0 };
            const unsigned long dim3748KuoInit[]        =   {   1   ,   1   ,   1   ,   11  ,   15  ,   13  ,   65  ,   1   ,   311 ,   469 ,   1307    ,   2605    ,   5281    ,   15665   ,   6349    ,   61753   ,0 };
            const unsigned long dim3749KuoInit[]        =   {   1   ,   1   ,   1   ,   3   ,   19  ,   33  ,   49  ,   141 ,   229 ,   925 ,   1059    ,   1295    ,   3093    ,   6207    ,   7607    ,   63987   ,0 };
            const unsigned long dim3750KuoInit[]        =   {   1   ,   1   ,   3   ,   1   ,   21  ,   63  ,   17  ,   131 ,   27  ,   295 ,   1969    ,   3259    ,   2011    ,   11005   ,   32249   ,   57641   ,0 };
            const unsigned long dim3751KuoInit[]        =   {   1   ,   3   ,   7   ,   15  ,   25  ,   27  ,   11  ,   105 ,   99  ,   923 ,   1003    ,   2765    ,   49  ,   10367   ,   22119   ,   27199   ,0 };
            const unsigned long dim3752KuoInit[]        =   {   1   ,   1   ,   1   ,   3   ,   31  ,   11  ,   57  ,   253 ,   31  ,   259 ,   1687    ,   1135    ,   1675    ,   12051   ,   3543    ,   19087   ,0 };
            const unsigned long dim3753KuoInit[]        =   {   1   ,   3   ,   5   ,   9   ,   9   ,   53  ,   119 ,   215 ,   213 ,   67  ,   343 ,   2743    ,   7303    ,   14225   ,   13379   ,   51001   ,0 };
            const unsigned long dim3754KuoInit[]        =   {   1   ,   3   ,   7   ,   9   ,   1   ,   45  ,   11  ,   125 ,   209 ,   459 ,   1037    ,   673 ,   2497    ,   669 ,   11251   ,   19387   ,0 };
            const unsigned long dim3755KuoInit[]        =   {   1   ,   1   ,   1   ,   15  ,   21  ,   47  ,   63  ,   45  ,   349 ,   561 ,   1823    ,   2005    ,   4173    ,   5595    ,   3789    ,   34133   ,0 };
            const unsigned long dim3756KuoInit[]        =   {   1   ,   1   ,   3   ,   3   ,   5   ,   7   ,   83  ,   63  ,   305 ,   679 ,   879 ,   3613    ,   7627    ,   9849    ,   3987    ,   28795   ,0 };
            const unsigned long dim3757KuoInit[]        =   {   1   ,   3   ,   7   ,   11  ,   7   ,   51  ,   119 ,   1   ,   21  ,   213 ,   1399    ,   3715    ,   7775    ,   1003    ,   6749    ,   5547    ,0 };
            const unsigned long dim3758KuoInit[]        =   {   1   ,   1   ,   7   ,   9   ,   21  ,   53  ,   51  ,   137 ,   379 ,   439 ,   1497    ,   3455    ,   2509    ,   2539    ,   3583    ,   5107    ,0 };
            const unsigned long dim3759KuoInit[]        =   {   1   ,   3   ,   5   ,   3   ,   9   ,   13  ,   73  ,   237 ,   65  ,   53  ,   663 ,   3895    ,   7579    ,   10027   ,   23969   ,   34767   ,0 };
            const unsigned long dim3760KuoInit[]        =   {   1   ,   3   ,   7   ,   3   ,   13  ,   57  ,   125 ,   245 ,   243 ,   561 ,   91  ,   863 ,   1925    ,   13553   ,   4113    ,   31077   ,0 };
            const unsigned long dim3761KuoInit[]        =   {   1   ,   1   ,   5   ,   7   ,   9   ,   31  ,   9   ,   83  ,   355 ,   249 ,   1741    ,   3593    ,   425 ,   6377    ,   10607   ,   48275   ,0 };
            const unsigned long dim3762KuoInit[]        =   {   1   ,   3   ,   7   ,   3   ,   19  ,   17  ,   123 ,   75  ,   413 ,   981 ,   877 ,   3489    ,   5907    ,   6141    ,   2729    ,   49219   ,0 };
            const unsigned long dim3763KuoInit[]        =   {   1   ,   3   ,   1   ,   3   ,   31  ,   39  ,   97  ,   233 ,   99  ,   17  ,   105 ,   2911    ,   347 ,   4069    ,   32535   ,   47179   ,0 };
            const unsigned long dim3764KuoInit[]        =   {   1   ,   1   ,   5   ,   7   ,   1   ,   51  ,   1   ,   59  ,   197 ,   487 ,   1763    ,   2449    ,   4249    ,   55  ,   11885   ,   57597   ,0 };
            const unsigned long dim3765KuoInit[]        =   {   1   ,   3   ,   5   ,   3   ,   25  ,   49  ,   43  ,   101 ,   353 ,   759 ,   1295    ,   3441    ,   1521    ,   2879    ,   9815    ,   27091   ,0 };
            const unsigned long dim3766KuoInit[]        =   {   1   ,   3   ,   5   ,   11  ,   1   ,   7   ,   69  ,   205 ,   271 ,   511 ,   1157    ,   3597    ,   2857    ,   10997   ,   15525   ,   6807    ,0 };
            const unsigned long dim3767KuoInit[]        =   {   1   ,   1   ,   1   ,   15  ,   11  ,   1   ,   45  ,   213 ,   463 ,   389 ,   1761    ,   3005    ,   3359    ,   12829   ,   18211   ,   2107    ,0 };
            const unsigned long dim3768KuoInit[]        =   {   1   ,   1   ,   7   ,   13  ,   27  ,   59  ,   115 ,   203 ,   321 ,   711 ,   1489    ,   2893    ,   1427    ,   4599    ,   17105   ,   10063   ,0 };
            const unsigned long dim3769KuoInit[]        =   {   1   ,   3   ,   1   ,   13  ,   5   ,   5   ,   53  ,   221 ,   97  ,   535 ,   1967    ,   1951    ,   4077    ,   327 ,   18243   ,   43285   ,0 };
            const unsigned long dim3770KuoInit[]        =   {   1   ,   3   ,   5   ,   13  ,   9   ,   61  ,   59  ,   255 ,   143 ,   21  ,   1859    ,   3509    ,   4295    ,   5107    ,   11117   ,   29431   ,0 };
            const unsigned long dim3771KuoInit[]        =   {   1   ,   3   ,   3   ,   13  ,   19  ,   13  ,   33  ,   169 ,   17  ,   773 ,   1693    ,   147 ,   5143    ,   2357    ,   2295    ,   50187   ,0 };
            const unsigned long dim3772KuoInit[]        =   {   1   ,   1   ,   5   ,   11  ,   11  ,   19  ,   79  ,   151 ,   441 ,   825 ,   665 ,   2947    ,   5391    ,   8695    ,   155 ,   26751   ,0 };
            const unsigned long dim3773KuoInit[]        =   {   1   ,   1   ,   7   ,   15  ,   13  ,   31  ,   77  ,   61  ,   249 ,   341 ,   1737    ,   3567    ,   6337    ,   14457   ,   3799    ,   20283   ,0 };
            const unsigned long dim3774KuoInit[]        =   {   1   ,   1   ,   7   ,   11  ,   13  ,   5   ,   83  ,   59  ,   185 ,   325 ,   2029    ,   1329    ,   1125    ,   10799   ,   27107   ,   2817    ,0 };
            const unsigned long dim3775KuoInit[]        =   {   1   ,   3   ,   5   ,   7   ,   17  ,   43  ,   21  ,   113 ,   57  ,   503 ,   1647    ,   393 ,   4525    ,   8385    ,   27153   ,   37175   ,0 };
            const unsigned long dim3776KuoInit[]        =   {   1   ,   3   ,   1   ,   15  ,   23  ,   57  ,   3   ,   177 ,   419 ,   65  ,   1895    ,   1107    ,   7705    ,   7915    ,   20677   ,   27931   ,0 };
            const unsigned long dim3777KuoInit[]        =   {   1   ,   3   ,   7   ,   13  ,   13  ,   63  ,   53  ,   237 ,   237 ,   161 ,   265 ,   1053    ,   161 ,   11635   ,   1455    ,   55551   ,0 };
            const unsigned long dim3778KuoInit[]        =   {   1   ,   3   ,   7   ,   5   ,   15  ,   41  ,   7   ,   77  ,   93  ,   737 ,   1195    ,   1381    ,   399 ,   7829    ,   6251    ,   34883   ,0 };
            const unsigned long dim3779KuoInit[]        =   {   1   ,   1   ,   5   ,   7   ,   23  ,   63  ,   119 ,   35  ,   115 ,   829 ,   441 ,   4059    ,   4755    ,   11331   ,   18071   ,   6911    ,0 };
            const unsigned long dim3780KuoInit[]        =   {   1   ,   3   ,   3   ,   7   ,   21  ,   33  ,   49  ,   47  ,   329 ,   341 ,   45  ,   3093    ,   5937    ,   6803    ,   31681   ,   9575    ,0 };
            const unsigned long dim3781KuoInit[]        =   {   1   ,   1   ,   1   ,   13  ,   13  ,   19  ,   121 ,   71  ,   231 ,   915 ,   981 ,   93  ,   5879    ,   15645   ,   8555    ,   58411   ,0 };
            const unsigned long dim3782KuoInit[]        =   {   1   ,   3   ,   7   ,   5   ,   31  ,   45  ,   71  ,   189 ,   337 ,   955 ,   1263    ,   2271    ,   1615    ,   15137   ,   29933   ,   16585   ,0 };
            const unsigned long dim3783KuoInit[]        =   {   1   ,   1   ,   7   ,   15  ,   29  ,   49  ,   85  ,   255 ,   143 ,   585 ,   1099    ,   1159    ,   1519    ,   15593   ,   6177    ,   16083   ,0 };
            const unsigned long dim3784KuoInit[]        =   {   1   ,   3   ,   5   ,   1   ,   13  ,   59  ,   79  ,   67  ,   341 ,   815 ,   703 ,   3857    ,   4711    ,   15155   ,   15997   ,   31699   ,0 };
            const unsigned long dim3785KuoInit[]        =   {   1   ,   3   ,   3   ,   7   ,   11  ,   51  ,   77  ,   231 ,   241 ,   261 ,   1607    ,   4031    ,   3779    ,   4289    ,   22039   ,   47177   ,0 };
            const unsigned long dim3786KuoInit[]        =   {   1   ,   3   ,   7   ,   1   ,   21  ,   23  ,   31  ,   233 ,   225 ,   883 ,   715 ,   777 ,   2433    ,   3665    ,   26695   ,   60997   ,0 };
            const unsigned long dim3787KuoInit[]        =   {   1   ,   3   ,   1   ,   1   ,   27  ,   17  ,   55  ,   33  ,   133 ,   899 ,   1443    ,   283 ,   6839    ,   9497    ,   22399   ,   60527   ,0 };
            const unsigned long dim3788KuoInit[]        =   {   1   ,   1   ,   3   ,   3   ,   5   ,   63  ,   91  ,   5   ,   275 ,   219 ,   1679    ,   2385    ,   4727    ,   9995    ,   14079   ,   24695   ,0 };
            const unsigned long dim3789KuoInit[]        =   {   1   ,   3   ,   7   ,   15  ,   9   ,   3   ,   77  ,   215 ,   265 ,   317 ,   1655    ,   1027    ,   5029    ,   6867    ,   9465    ,   64257   ,0 };
            const unsigned long dim3790KuoInit[]        =   {   1   ,   3   ,   1   ,   3   ,   23  ,   7   ,   67  ,   63  ,   503 ,   269 ,   85  ,   3669    ,   5941    ,   13833   ,   25857   ,   34221   ,0 };
            const unsigned long dim3791KuoInit[]        =   {   1   ,   1   ,   1   ,   7   ,   13  ,   31  ,   83  ,   113 ,   389 ,   923 ,   1173    ,   2457    ,   997 ,   9761    ,   17033   ,   50521   ,0 };
            const unsigned long dim3792KuoInit[]        =   {   1   ,   3   ,   1   ,   1   ,   15  ,   31  ,   93  ,   243 ,   221 ,   331 ,   1961    ,   3251    ,   6459    ,   3341    ,   28535   ,   27909   ,0 };
            const unsigned long dim3793KuoInit[]        =   {   1   ,   3   ,   1   ,   15  ,   27  ,   1   ,   75  ,   169 ,   277 ,   921 ,   5   ,   1391    ,   5665    ,   14659   ,   6895    ,   37927   ,0 };
            const unsigned long dim3794KuoInit[]        =   {   1   ,   3   ,   1   ,   9   ,   17  ,   63  ,   93  ,   35  ,   225 ,   109 ,   977 ,   2623    ,   3779    ,   8011    ,   11643   ,   57789   ,0 };
            const unsigned long dim3795KuoInit[]        =   {   1   ,   1   ,   5   ,   3   ,   17  ,   5   ,   63  ,   109 ,   449 ,   695 ,   1021    ,   2749    ,   3981    ,   2073    ,   24537   ,   43643   ,0 };
            const unsigned long dim3796KuoInit[]        =   {   1   ,   1   ,   5   ,   1   ,   1   ,   19  ,   15  ,   249 ,   3   ,   431 ,   743 ,   3813    ,   4013    ,   4027    ,   1943    ,   23569   ,0 };
            const unsigned long dim3797KuoInit[]        =   {   1   ,   3   ,   7   ,   1   ,   9   ,   49  ,   93  ,   59  ,   425 ,   325 ,   1925    ,   495 ,   6611    ,   2425    ,   19847   ,   58683   ,0 };
            const unsigned long dim3798KuoInit[]        =   {   1   ,   3   ,   5   ,   5   ,   7   ,   21  ,   71  ,   51  ,   335 ,   67  ,   373 ,   1805    ,   1805    ,   4069    ,   6585    ,   56579   ,0 };
            const unsigned long dim3799KuoInit[]        =   {   1   ,   3   ,   1   ,   7   ,   25  ,   27  ,   49  ,   131 ,   491 ,   105 ,   1467    ,   665 ,   6777    ,   15467   ,   13597   ,   49141   ,0 };
            const unsigned long dim3800KuoInit[]        =   {   1   ,   3   ,   5   ,   13  ,   9   ,   9   ,   37  ,   159 ,   19  ,   91  ,   1963    ,   3133    ,   5147    ,   8655    ,   11153   ,   44491   ,0 };
            const unsigned long dim3801KuoInit[]        =   {   1   ,   3   ,   5   ,   7   ,   13  ,   55  ,   41  ,   67  ,   155 ,   853 ,   1651    ,   3511    ,   4391    ,   6359    ,   9129    ,   32359   ,0 };
            const unsigned long dim3802KuoInit[]        =   {   1   ,   3   ,   5   ,   5   ,   9   ,   3   ,   27  ,   101 ,   467 ,   185 ,   1071    ,   1477    ,   1523    ,   9535    ,   15433   ,   7493    ,0 };
            const unsigned long dim3803KuoInit[]        =   {   1   ,   1   ,   1   ,   3   ,   15  ,   19  ,   3   ,   97  ,   241 ,   995 ,   451 ,   1651    ,   241 ,   1063    ,   2521    ,   65531   ,0 };
            const unsigned long dim3804KuoInit[]        =   {   1   ,   3   ,   3   ,   5   ,   23  ,   13  ,   117 ,   65  ,   119 ,   871 ,   1245    ,   3577    ,   7303    ,   9433    ,   13701   ,   21321   ,0 };
            const unsigned long dim3805KuoInit[]        =   {   1   ,   1   ,   1   ,   11  ,   25  ,   33  ,   121 ,   203 ,   325 ,   775 ,   1835    ,   1385    ,   2541    ,   3399    ,   275 ,   54311   ,0 };
            const unsigned long dim3806KuoInit[]        =   {   1   ,   1   ,   3   ,   3   ,   13  ,   41  ,   5   ,   63  ,   415 ,   771 ,   375 ,   3933    ,   5239    ,   1731    ,   7675    ,   11855   ,0 };
            const unsigned long dim3807KuoInit[]        =   {   1   ,   1   ,   5   ,   7   ,   23  ,   49  ,   27  ,   119 ,   231 ,   897 ,   1995    ,   3847    ,   25  ,   7505    ,   2823    ,   10115   ,0 };
            const unsigned long dim3808KuoInit[]        =   {   1   ,   3   ,   5   ,   1   ,   5   ,   41  ,   41  ,   129 ,   339 ,   861 ,   147 ,   3299    ,   1123    ,   2175    ,   27111   ,   2067    ,0 };
            const unsigned long dim3809KuoInit[]        =   {   1   ,   1   ,   3   ,   5   ,   5   ,   39  ,   71  ,   53  ,   483 ,   585 ,   1167    ,   2037    ,   3125    ,   499 ,   29491   ,   43467   ,0 };
            const unsigned long dim3810KuoInit[]        =   {   1   ,   1   ,   1   ,   3   ,   25  ,   9   ,   85  ,   197 ,   271 ,   457 ,   1159    ,   3805    ,   2423    ,   11777   ,   4383    ,   57495   ,0 };
            const unsigned long dim3811KuoInit[]        =   {   1   ,   3   ,   7   ,   9   ,   25  ,   5   ,   33  ,   255 ,   325 ,   643 ,   499 ,   3233    ,   4477    ,   14189   ,   12183   ,   7667    ,0 };
            const unsigned long dim3812KuoInit[]        =   {   1   ,   3   ,   3   ,   5   ,   1   ,   63  ,   3   ,   51  ,   173 ,   1011    ,   1423    ,   3453    ,   2471    ,   7801    ,   29989   ,   8347    ,0 };
            const unsigned long dim3813KuoInit[]        =   {   1   ,   1   ,   7   ,   11  ,   17  ,   39  ,   23  ,   13  ,   413 ,   503 ,   321 ,   3699    ,   6441    ,   11953   ,   24253   ,   2561    ,0 };
            const unsigned long dim3814KuoInit[]        =   {   1   ,   1   ,   1   ,   3   ,   31  ,   45  ,   21  ,   215 ,   55  ,   555 ,   771 ,   2703    ,   8113    ,   3279    ,   6053    ,   48617   ,0 };
            const unsigned long dim3815KuoInit[]        =   {   1   ,   1   ,   7   ,   9   ,   25  ,   25  ,   33  ,   101 ,   111 ,   643 ,   1277    ,   3833    ,   7431    ,   14395   ,   20849   ,   14455   ,0 };
            const unsigned long dim3816KuoInit[]        =   {   1   ,   3   ,   7   ,   11  ,   21  ,   15  ,   95  ,   179 ,   437 ,   747 ,   1131    ,   83  ,   415 ,   12605   ,   29245   ,   28021   ,0 };
            const unsigned long dim3817KuoInit[]        =   {   1   ,   3   ,   7   ,   5   ,   13  ,   59  ,   9   ,   217 ,   287 ,   929 ,   691 ,   3201    ,   4915    ,   3301    ,   15371   ,   17063   ,0 };
            const unsigned long dim3818KuoInit[]        =   {   1   ,   1   ,   1   ,   7   ,   31  ,   3   ,   27  ,   229 ,   151 ,   321 ,   801 ,   1271    ,   1195    ,   11221   ,   16903   ,   27919   ,0 };
            const unsigned long dim3819KuoInit[]        =   {   1   ,   1   ,   5   ,   9   ,   5   ,   41  ,   105 ,   133 ,   265 ,   167 ,   505 ,   3939    ,   771 ,   1291    ,   1109    ,   52601   ,0 };
            const unsigned long dim3820KuoInit[]        =   {   1   ,   3   ,   5   ,   11  ,   21  ,   23  ,   23  ,   229 ,   197 ,   623 ,   1973    ,   901 ,   147 ,   15751   ,   29429   ,   13159   ,0 };
            const unsigned long dim3821KuoInit[]        =   {   1   ,   1   ,   5   ,   11  ,   25  ,   53  ,   79  ,   123 ,   245 ,   939 ,   1889    ,   4073    ,   417 ,   13625   ,   14691   ,   56965   ,0 };
            const unsigned long dim3822KuoInit[]        =   {   1   ,   1   ,   1   ,   11  ,   1   ,   39  ,   51  ,   231 ,   367 ,   233 ,   1453    ,   3095    ,   6071    ,   7521    ,   469 ,   8407    ,0 };
            const unsigned long dim3823KuoInit[]        =   {   1   ,   3   ,   5   ,   1   ,   9   ,   17  ,   109 ,   135 ,   85  ,   151 ,   1217    ,   1643    ,   4455    ,   3923    ,   11681   ,   32807   ,0 };
            const unsigned long dim3824KuoInit[]        =   {   1   ,   1   ,   1   ,   13  ,   11  ,   39  ,   27  ,   205 ,   85  ,   1019    ,   1759    ,   1705    ,   4585    ,   8311    ,   29123   ,   44045   ,0 };
            const unsigned long dim3825KuoInit[]        =   {   1   ,   3   ,   1   ,   3   ,   9   ,   33  ,   65  ,   253 ,   319 ,   65  ,   851 ,   3357    ,   2827    ,   10107   ,   10197   ,   19111   ,0 };
            const unsigned long dim3826KuoInit[]        =   {   1   ,   3   ,   7   ,   3   ,   3   ,   53  ,   67  ,   99  ,   501 ,   603 ,   1409    ,   359 ,   7395    ,   12827   ,   10867   ,   17473   ,0 };
            const unsigned long dim3827KuoInit[]        =   {   1   ,   1   ,   7   ,   5   ,   5   ,   35  ,   97  ,   49  ,   365 ,   971 ,   1707    ,   1305    ,   6801    ,   14451   ,   32473   ,   57269   ,0 };
            const unsigned long dim3828KuoInit[]        =   {   1   ,   1   ,   5   ,   11  ,   25  ,   57  ,   115 ,   195 ,   47  ,   923 ,   1515    ,   2043    ,   803 ,   6371    ,   25341   ,   63263   ,0 };
            const unsigned long dim3829KuoInit[]        =   {   1   ,   1   ,   5   ,   3   ,   27  ,   53  ,   83  ,   91  ,   367 ,   695 ,   1809    ,   927 ,   6815    ,   1229    ,   27359   ,   28937   ,0 };
            const unsigned long dim3830KuoInit[]        =   {   1   ,   1   ,   3   ,   7   ,   3   ,   3   ,   17  ,   203 ,   95  ,   577 ,   1531    ,   1645    ,   7271    ,   5541    ,   30905   ,   38825   ,0 };
            const unsigned long dim3831KuoInit[]        =   {   1   ,   1   ,   7   ,   9   ,   7   ,   57  ,   97  ,   185 ,   75  ,   681 ,   1197    ,   3713    ,   6429    ,   4723    ,   26275   ,   23495   ,0 };
            const unsigned long dim3832KuoInit[]        =   {   1   ,   3   ,   1   ,   3   ,   5   ,   25  ,   71  ,   171 ,   143 ,   603 ,   1349    ,   1723    ,   7493    ,   8417    ,   24171   ,   13421   ,0 };
            const unsigned long dim3833KuoInit[]        =   {   1   ,   3   ,   1   ,   15  ,   23  ,   57  ,   33  ,   247 ,   409 ,   511 ,   29  ,   1403    ,   5831    ,   1749    ,   28441   ,   4117    ,0 };
            const unsigned long dim3834KuoInit[]        =   {   1   ,   1   ,   3   ,   15  ,   31  ,   11  ,   105 ,   239 ,   431 ,   127 ,   343 ,   961 ,   6449    ,   1677    ,   24937   ,   15859   ,0 };
            const unsigned long dim3835KuoInit[]        =   {   1   ,   1   ,   5   ,   3   ,   27  ,   57  ,   61  ,   245 ,   425 ,   417 ,   301 ,   521 ,   2425    ,   2547    ,   11349   ,   47931   ,0 };
            const unsigned long dim3836KuoInit[]        =   {   1   ,   3   ,   5   ,   3   ,   15  ,   19  ,   19  ,   127 ,   151 ,   913 ,   1061    ,   1795    ,   2519    ,   6297    ,   2519    ,   56677   ,0 };
            const unsigned long dim3837KuoInit[]        =   {   1   ,   1   ,   3   ,   9   ,   25  ,   3   ,   67  ,   53  ,   305 ,   609 ,   1067    ,   1003    ,   6933    ,   14009   ,   24475   ,   23183   ,0 };
            const unsigned long dim3838KuoInit[]        =   {   1   ,   1   ,   5   ,   7   ,   21  ,   35  ,   89  ,   95  ,   399 ,   341 ,   1337    ,   2815    ,   3605    ,   3209    ,   9881    ,   6399    ,0 };
            const unsigned long dim3839KuoInit[]        =   {   1   ,   1   ,   7   ,   1   ,   29  ,   63  ,   23  ,   219 ,   145 ,   583 ,   985 ,   2577    ,   3951    ,   15199   ,   26981   ,   2755    ,0 };
            const unsigned long dim3840KuoInit[]        =   {   1   ,   3   ,   7   ,   1   ,   13  ,   45  ,   105 ,   139 ,   103 ,   83  ,   819 ,   2655    ,   39  ,   11661   ,   11057   ,   38189   ,0 };
            const unsigned long dim3841KuoInit[]        =   {   1   ,   3   ,   1   ,   15  ,   23  ,   41  ,   79  ,   237 ,   353 ,   67  ,   1551    ,   2345    ,   1169    ,   4623    ,   1391    ,   18607   ,0 };
            const unsigned long dim3842KuoInit[]        =   {   1   ,   1   ,   1   ,   13  ,   29  ,   31  ,   43  ,   65  ,   25  ,   387 ,   1109    ,   2515    ,   3967    ,   2221    ,   27321   ,   20511   ,0 };
            const unsigned long dim3843KuoInit[]        =   {   1   ,   1   ,   1   ,   3   ,   17  ,   45  ,   35  ,   211 ,   31  ,   45  ,   1791    ,   2225    ,   6995    ,   15241   ,   24555   ,   45993   ,0 };
            const unsigned long dim3844KuoInit[]        =   {   1   ,   1   ,   5   ,   5   ,   31  ,   25  ,   71  ,   57  ,   87  ,   531 ,   117 ,   267 ,   2415    ,   2213    ,   603 ,   23369   ,0 };
            const unsigned long dim3845KuoInit[]        =   {   1   ,   1   ,   1   ,   5   ,   23  ,   53  ,   39  ,   3   ,   93  ,   697 ,   25  ,   465 ,   1517    ,   6061    ,   11759   ,   34863   ,0 };
            const unsigned long dim3846KuoInit[]        =   {   1   ,   1   ,   1   ,   13  ,   29  ,   29  ,   41  ,   27  ,   499 ,   285 ,   355 ,   3915    ,   1683    ,   11933   ,   24685   ,   59013   ,0 };
            const unsigned long dim3847KuoInit[]        =   {   1   ,   3   ,   1   ,   3   ,   21  ,   21  ,   25  ,   251 ,   109 ,   469 ,   1841    ,   889 ,   6541    ,   12961   ,   9449    ,   48755   ,0 };
            const unsigned long dim3848KuoInit[]        =   {   1   ,   3   ,   1   ,   1   ,   21  ,   59  ,   41  ,   159 ,   159 ,   105 ,   999 ,   2535    ,   3251    ,   12213   ,   24005   ,   17145   ,0 };
            const unsigned long dim3849KuoInit[]        =   {   1   ,   1   ,   7   ,   9   ,   1   ,   39  ,   43  ,   235 ,   447 ,   767 ,   871 ,   453 ,   7285    ,   3387    ,   14625   ,   40099   ,0 };
            const unsigned long dim3850KuoInit[]        =   {   1   ,   1   ,   3   ,   3   ,   5   ,   11  ,   15  ,   205 ,   183 ,   309 ,   1747    ,   1089    ,   3141    ,   9129    ,   25149   ,   63109   ,0 };
            const unsigned long dim3851KuoInit[]        =   {   1   ,   3   ,   7   ,   15  ,   23  ,   15  ,   7   ,   157 ,   45  ,   263 ,   1383    ,   735 ,   205 ,   16021   ,   13233   ,   23425   ,0 };
            const unsigned long dim3852KuoInit[]        =   {   1   ,   1   ,   7   ,   13  ,   11  ,   31  ,   53  ,   23  ,   357 ,   953 ,   611 ,   2579    ,   475 ,   15183   ,   12755   ,   10347   ,0 };
            const unsigned long dim3853KuoInit[]        =   {   1   ,   1   ,   1   ,   5   ,   7   ,   27  ,   41  ,   215 ,   247 ,   1001    ,   1369    ,   3105    ,   3195    ,   13417   ,   6967    ,   52501   ,0 };
            const unsigned long dim3854KuoInit[]        =   {   1   ,   3   ,   7   ,   3   ,   3   ,   61  ,   35  ,   175 ,   415 ,   835 ,   1381    ,   3951    ,   6857    ,   10273   ,   26657   ,   5239    ,0 };
            const unsigned long dim3855KuoInit[]        =   {   1   ,   3   ,   7   ,   5   ,   1   ,   27  ,   3   ,   251 ,   499 ,   637 ,   1101    ,   3817    ,   4823    ,   3065    ,   29919   ,   4151    ,0 };
            const unsigned long dim3856KuoInit[]        =   {   1   ,   3   ,   7   ,   13  ,   19  ,   39  ,   1   ,   141 ,   353 ,   853 ,   295 ,   1201    ,   6895    ,   3965    ,   6417    ,   47379   ,0 };
            const unsigned long dim3857KuoInit[]        =   {   1   ,   1   ,   7   ,   13  ,   11  ,   9   ,   99  ,   225 ,   161 ,   349 ,   421 ,   2803    ,   4461    ,   2319    ,   2495    ,   62429   ,0 };
            const unsigned long dim3858KuoInit[]        =   {   1   ,   3   ,   1   ,   13  ,   31  ,   1   ,   47  ,   115 ,   299 ,   331 ,   1917    ,   2547    ,   969 ,   4841    ,   8237    ,   2405    ,0 };
            const unsigned long dim3859KuoInit[]        =   {   1   ,   1   ,   1   ,   9   ,   9   ,   19  ,   77  ,   35  ,   225 ,   211 ,   1875    ,   3531    ,   2753    ,   12679   ,   6863    ,   42143   ,0 };
            const unsigned long dim3860KuoInit[]        =   {   1   ,   1   ,   1   ,   3   ,   29  ,   25  ,   105 ,   19  ,   269 ,   863 ,   811 ,   2741    ,   143 ,   6353    ,   25109   ,   62251   ,0 };
            const unsigned long dim3861KuoInit[]        =   {   1   ,   3   ,   5   ,   3   ,   27  ,   51  ,   11  ,   61  ,   245 ,   517 ,   891 ,   2081    ,   4523    ,   11703   ,   31477   ,   12113   ,0 };
            const unsigned long dim3862KuoInit[]        =   {   1   ,   1   ,   1   ,   1   ,   27  ,   29  ,   25  ,   47  ,   143 ,   489 ,   1475    ,   1139    ,   1855    ,   8031    ,   21341   ,   23619   ,0 };
            const unsigned long dim3863KuoInit[]        =   {   1   ,   1   ,   5   ,   1   ,   3   ,   45  ,   103 ,   59  ,   123 ,   831 ,   1883    ,   149 ,   5843    ,   10275   ,   3121    ,   21411   ,0 };
            const unsigned long dim3864KuoInit[]        =   {   1   ,   1   ,   7   ,   1   ,   21  ,   11  ,   1   ,   35  ,   285 ,   501 ,   1407    ,   1307    ,   1371    ,   15057   ,   14091   ,   52279   ,0 };
            const unsigned long dim3865KuoInit[]        =   {   1   ,   1   ,   3   ,   13  ,   3   ,   5   ,   9   ,   167 ,   139 ,   255 ,   1561    ,   2183    ,   1079    ,   2055    ,   24779   ,   19737   ,0 };
            const unsigned long dim3866KuoInit[]        =   {   1   ,   1   ,   7   ,   13  ,   31  ,   29  ,   59  ,   157 ,   311 ,   447 ,   21  ,   2177    ,   3679    ,   7519    ,   987 ,   44221   ,0 };
            const unsigned long dim3867KuoInit[]        =   {   1   ,   3   ,   3   ,   1   ,   11  ,   31  ,   95  ,   23  ,   457 ,   733 ,   1413    ,   273 ,   333 ,   14391   ,   16557   ,   35227   ,0 };
            const unsigned long dim3868KuoInit[]        =   {   1   ,   1   ,   3   ,   9   ,   5   ,   3   ,   103 ,   59  ,   21  ,   475 ,   1455    ,   2995    ,   7651    ,   4893    ,   21739   ,   21379   ,0 };
            const unsigned long dim3869KuoInit[]        =   {   1   ,   3   ,   5   ,   1   ,   9   ,   63  ,   55  ,   127 ,   283 ,   603 ,   1477    ,   331 ,   2115    ,   15965   ,   22433   ,   17311   ,0 };
            const unsigned long dim3870KuoInit[]        =   {   1   ,   1   ,   3   ,   3   ,   5   ,   61  ,   105 ,   175 ,   245 ,   431 ,   479 ,   821 ,   3835    ,   3585    ,   16533   ,   63233   ,0 };
            const unsigned long dim3871KuoInit[]        =   {   1   ,   1   ,   7   ,   1   ,   31  ,   57  ,   9   ,   11  ,   145 ,   519 ,   1915    ,   2809    ,   2943    ,   13885   ,   31223   ,   32897   ,0 };
            const unsigned long dim3872KuoInit[]        =   {   1   ,   3   ,   1   ,   15  ,   5   ,   39  ,   67  ,   179 ,   285 ,   91  ,   1429    ,   3571    ,   1925    ,   9737    ,   28647   ,   8189    ,0 };
            const unsigned long dim3873KuoInit[]        =   {   1   ,   3   ,   7   ,   11  ,   21  ,   25  ,   51  ,   173 ,   151 ,   283 ,   765 ,   2717    ,   3541    ,   10317   ,   9827    ,   24445   ,0 };
            const unsigned long dim3874KuoInit[]        =   {   1   ,   3   ,   1   ,   13  ,   31  ,   25  ,   121 ,   71  ,   347 ,   129 ,   159 ,   1845    ,   475 ,   12777   ,   19607   ,   7163    ,0 };
            const unsigned long dim3875KuoInit[]        =   {   1   ,   3   ,   1   ,   9   ,   29  ,   37  ,   59  ,   117 ,   375 ,   411 ,   1031    ,   3781    ,   3119    ,   6851    ,   30759   ,   46181   ,0 };
            const unsigned long dim3876KuoInit[]        =   {   1   ,   3   ,   1   ,   7   ,   5   ,   41  ,   81  ,   141 ,   377 ,   817 ,   1007    ,   2211    ,   7303    ,   5833    ,   6117    ,   52265   ,0 };
            const unsigned long dim3877KuoInit[]        =   {   1   ,   3   ,   5   ,   3   ,   7   ,   63  ,   83  ,   181 ,   169 ,   19  ,   1589    ,   3793    ,   7003    ,   12145   ,   25625   ,   35731   ,0 };
            const unsigned long dim3878KuoInit[]        =   {   1   ,   1   ,   1   ,   5   ,   21  ,   39  ,   15  ,   225 ,   231 ,   497 ,   917 ,   559 ,   6515    ,   13637   ,   17721   ,   4725    ,0 };
            const unsigned long dim3879KuoInit[]        =   {   1   ,   3   ,   1   ,   15  ,   17  ,   19  ,   93  ,   65  ,   281 ,   775 ,   613 ,   2453    ,   1969    ,   2499    ,   5373    ,   28207   ,0 };
            const unsigned long dim3880KuoInit[]        =   {   1   ,   1   ,   5   ,   11  ,   9   ,   37  ,   103 ,   211 ,   357 ,   409 ,   457 ,   1527    ,   3289    ,   14293   ,   7681    ,   58069   ,0 };
            const unsigned long dim3881KuoInit[]        =   {   1   ,   3   ,   5   ,   9   ,   1   ,   29  ,   83  ,   1   ,   35  ,   655 ,   491 ,   1447    ,   725 ,   7807    ,   13311   ,   345 ,0 };
            const unsigned long dim3882KuoInit[]        =   {   1   ,   3   ,   1   ,   7   ,   19  ,   59  ,   103 ,   243 ,   157 ,   209 ,   399 ,   1113    ,   1595    ,   15075   ,   22475   ,   42835   ,0 };
            const unsigned long dim3883KuoInit[]        =   {   1   ,   3   ,   5   ,   5   ,   27  ,   53  ,   69  ,   73  ,   413 ,   443 ,   559 ,   427 ,   2841    ,   13107   ,   8909    ,   35265   ,0 };
            const unsigned long dim3884KuoInit[]        =   {   1   ,   3   ,   3   ,   7   ,   21  ,   39  ,   99  ,   75  ,   341 ,   917 ,   11  ,   1247    ,   247 ,   14285   ,   24531   ,   58523   ,0 };
            const unsigned long dim3885KuoInit[]        =   {   1   ,   1   ,   7   ,   9   ,   29  ,   1   ,   55  ,   143 ,   363 ,   527 ,   103 ,   3735    ,   6373    ,   7837    ,   24451   ,   36999   ,0 };
            const unsigned long dim3886KuoInit[]        =   {   1   ,   1   ,   3   ,   3   ,   3   ,   51  ,   71  ,   103 ,   173 ,   397 ,   1683    ,   489 ,   6515    ,   2531    ,   12517   ,   63165   ,0 };
            const unsigned long dim3887KuoInit[]        =   {   1   ,   3   ,   7   ,   3   ,   27  ,   15  ,   31  ,   63  ,   323 ,   903 ,   1025    ,   2761    ,   1813    ,   2839    ,   2065    ,   20013   ,0 };
            const unsigned long dim3888KuoInit[]        =   {   1   ,   1   ,   1   ,   13  ,   19  ,   37  ,   49  ,   207 ,   57  ,   45  ,   1517    ,   2715    ,   6315    ,   10499   ,   27195   ,   56175   ,0 };
            const unsigned long dim3889KuoInit[]        =   {   1   ,   3   ,   5   ,   11  ,   13  ,   47  ,   91  ,   81  ,   459 ,   801 ,   1111    ,   3023    ,   5901    ,   12735   ,   14233   ,   24383   ,0 };
            const unsigned long dim3890KuoInit[]        =   {   1   ,   3   ,   1   ,   5   ,   19  ,   39  ,   11  ,   69  ,   333 ,   17  ,   1339    ,   2991    ,   4411    ,   2749    ,   15423   ,   13805   ,0 };
            const unsigned long dim3891KuoInit[]        =   {   1   ,   3   ,   5   ,   9   ,   19  ,   3   ,   109 ,   45  ,   461 ,   103 ,   811 ,   3089    ,   893 ,   16121   ,   1337    ,   31929   ,0 };
            const unsigned long dim3892KuoInit[]        =   {   1   ,   3   ,   7   ,   11  ,   3   ,   21  ,   125 ,   241 ,   355 ,   519 ,   1551    ,   3273    ,   2413    ,   3371    ,   11435   ,   2277    ,0 };
            const unsigned long dim3893KuoInit[]        =   {   1   ,   1   ,   5   ,   9   ,   7   ,   21  ,   93  ,   79  ,   25  ,   265 ,   1045    ,   2011    ,   4883    ,   8631    ,   22461   ,   31117   ,0 };
            const unsigned long dim3894KuoInit[]        =   {   1   ,   1   ,   3   ,   9   ,   13  ,   35  ,   61  ,   203 ,   371 ,   185 ,   1445    ,   2017    ,   363 ,   13603   ,   23097   ,   22617   ,0 };
            const unsigned long dim3895KuoInit[]        =   {   1   ,   1   ,   1   ,   1   ,   29  ,   33  ,   9   ,   139 ,   73  ,   119 ,   755 ,   2117    ,   3949    ,   3403    ,   30007   ,   58147   ,0 };
            const unsigned long dim3896KuoInit[]        =   {   1   ,   1   ,   3   ,   5   ,   9   ,   29  ,   85  ,   1   ,   349 ,   573 ,   1191    ,   133 ,   235 ,   1557    ,   18147   ,   5839    ,0 };
            const unsigned long dim3897KuoInit[]        =   {   1   ,   3   ,   3   ,   5   ,   15  ,   45  ,   5   ,   95  ,   359 ,   223 ,   37  ,   547 ,   3407    ,   4927    ,   32417   ,   49205   ,0 };
            const unsigned long dim3898KuoInit[]        =   {   1   ,   1   ,   7   ,   11  ,   21  ,   57  ,   127 ,   95  ,   355 ,   903 ,   1555    ,   9   ,   1249    ,   2463    ,   7921    ,   20633   ,0 };
            const unsigned long dim3899KuoInit[]        =   {   1   ,   3   ,   3   ,   11  ,   29  ,   35  ,   95  ,   237 ,   43  ,   289 ,   1547    ,   3951    ,   5489    ,   2699    ,   347 ,   14405   ,0 };
            const unsigned long dim3900KuoInit[]        =   {   1   ,   3   ,   3   ,   11  ,   11  ,   23  ,   33  ,   171 ,   37  ,   989 ,   917 ,   139 ,   5497    ,   617 ,   25555   ,   27031   ,0 };
            const unsigned long dim3901KuoInit[]        =   {   1   ,   1   ,   5   ,   1   ,   21  ,   55  ,   63  ,   191 ,   93  ,   207 ,   1799    ,   2143    ,   4147    ,   6853    ,   18129   ,   18287   ,0 };
            const unsigned long dim3902KuoInit[]        =   {   1   ,   3   ,   7   ,   15  ,   23  ,   39  ,   59  ,   131 ,   51  ,   113 ,   941 ,   2599    ,   5593    ,   11311   ,   30561   ,   22307   ,0 };
            const unsigned long dim3903KuoInit[]        =   {   1   ,   3   ,   7   ,   5   ,   21  ,   49  ,   93  ,   5   ,   399 ,   173 ,   1017    ,   3551    ,   937 ,   13653   ,   21411   ,   56835   ,0 };
            const unsigned long dim3904KuoInit[]        =   {   1   ,   3   ,   1   ,   1   ,   15  ,   47  ,   57  ,   19  ,   93  ,   931 ,   1803    ,   2947    ,   1291    ,   10979   ,   22069   ,   47485   ,0 };
            const unsigned long dim3905KuoInit[]        =   {   1   ,   3   ,   7   ,   11  ,   31  ,   19  ,   75  ,   195 ,   23  ,   235 ,   137 ,   2521    ,   4109    ,   233 ,   26961   ,   19185   ,0 };
            const unsigned long dim3906KuoInit[]        =   {   1   ,   3   ,   3   ,   15  ,   9   ,   17  ,   87  ,   239 ,   233 ,   939 ,   1643    ,   1189    ,   7875    ,   5199    ,   26603   ,   52223   ,0 };
            const unsigned long dim3907KuoInit[]        =   {   1   ,   3   ,   5   ,   7   ,   3   ,   31  ,   21  ,   255 ,   41  ,   95  ,   1267    ,   1873    ,   5535    ,   8419    ,   19791   ,   14811   ,0 };
            const unsigned long dim3908KuoInit[]        =   {   1   ,   1   ,   7   ,   15  ,   7   ,   51  ,   27  ,   35  ,   17  ,   1015    ,   661 ,   2679    ,   1269    ,   15729   ,   17329   ,   33555   ,0 };
            const unsigned long dim3909KuoInit[]        =   {   1   ,   1   ,   5   ,   1   ,   13  ,   19  ,   39  ,   105 ,   173 ,   255 ,   1873    ,   1527    ,   253 ,   3517    ,   19867   ,   56765   ,0 };
            const unsigned long dim3910KuoInit[]        =   {   1   ,   1   ,   7   ,   1   ,   15  ,   13  ,   85  ,   27  ,   491 ,   827 ,   1151    ,   3591    ,   7819    ,   10085   ,   11307   ,   45059   ,0 };
            const unsigned long dim3911KuoInit[]        =   {   1   ,   3   ,   5   ,   15  ,   21  ,   17  ,   23  ,   207 ,   223 ,   721 ,   655 ,   243 ,   4289    ,   4973    ,   29709   ,   9203    ,0 };
            const unsigned long dim3912KuoInit[]        =   {   1   ,   1   ,   5   ,   11  ,   27  ,   51  ,   3   ,   199 ,   277 ,   121 ,   1637    ,   3973    ,   861 ,   7215    ,   32577   ,   4391    ,0 };
            const unsigned long dim3913KuoInit[]        =   {   1   ,   1   ,   7   ,   9   ,   21  ,   27  ,   33  ,   199 ,   29  ,   917 ,   99  ,   2755    ,   821 ,   16003   ,   8325    ,   55587   ,0 };
            const unsigned long dim3914KuoInit[]        =   {   1   ,   1   ,   7   ,   1   ,   19  ,   17  ,   61  ,   39  ,   377 ,   739 ,   13  ,   3435    ,   5081    ,   10591   ,   11957   ,   30049   ,0 };
            const unsigned long dim3915KuoInit[]        =   {   1   ,   1   ,   5   ,   1   ,   25  ,   13  ,   117 ,   5   ,   479 ,   579 ,   907 ,   3135    ,   1011    ,   3389    ,   135 ,   41711   ,0 };
            const unsigned long dim3916KuoInit[]        =   {   1   ,   3   ,   5   ,   5   ,   19  ,   15  ,   89  ,   169 ,   321 ,   981 ,   365 ,   333 ,   4913    ,   9577    ,   13835   ,   23439   ,0 };
            const unsigned long dim3917KuoInit[]        =   {   1   ,   1   ,   5   ,   11  ,   31  ,   27  ,   103 ,   25  ,   129 ,   207 ,   259 ,   2033    ,   7931    ,   8357    ,   17451   ,   60583   ,0 };
            const unsigned long dim3918KuoInit[]        =   {   1   ,   1   ,   7   ,   11  ,   15  ,   47  ,   59  ,   151 ,   291 ,   669 ,   663 ,   111 ,   4809    ,   14179   ,   22313   ,   51969   ,0 };
            const unsigned long dim3919KuoInit[]        =   {   1   ,   3   ,   3   ,   5   ,   1   ,   5   ,   87  ,   25  ,   179 ,   641 ,   1133    ,   629 ,   2281    ,   6213    ,   14281   ,   54147   ,0 };
            const unsigned long dim3920KuoInit[]        =   {   1   ,   1   ,   5   ,   15  ,   29  ,   33  ,   89  ,   63  ,   331 ,   379 ,   453 ,   1903    ,   7319    ,   15063   ,   31383   ,   23099   ,0 };
            const unsigned long dim3921KuoInit[]        =   {   1   ,   3   ,   1   ,   15  ,   23  ,   7   ,   13  ,   71  ,   367 ,   53  ,   983 ,   3615    ,   5601    ,   2569    ,   18173   ,   40563   ,0 };
            const unsigned long dim3922KuoInit[]        =   {   1   ,   3   ,   1   ,   15  ,   23  ,   41  ,   35  ,   215 ,   401 ,   807 ,   1809    ,   3919    ,   5651    ,   1249    ,   9699    ,   14275   ,0 };
            const unsigned long dim3923KuoInit[]        =   {   1   ,   1   ,   3   ,   1   ,   15  ,   55  ,   7   ,   235 ,   171 ,   911 ,   1351    ,   2863    ,   5735    ,   8577    ,   7489    ,   42047   ,0 };
            const unsigned long dim3924KuoInit[]        =   {   1   ,   3   ,   1   ,   5   ,   29  ,   35  ,   11  ,   211 ,   133 ,   507 ,   1127    ,   2443    ,   7895    ,   10519   ,   13035   ,   26165   ,0 };
            const unsigned long dim3925KuoInit[]        =   {   1   ,   1   ,   1   ,   9   ,   21  ,   39  ,   107 ,   55  ,   235 ,   409 ,   39  ,   1681    ,   2699    ,   3143    ,   847 ,   35635   ,0 };
            const unsigned long dim3926KuoInit[]        =   {   1   ,   3   ,   7   ,   3   ,   27  ,   57  ,   79  ,   173 ,   273 ,   919 ,   1917    ,   4005    ,   1299    ,   14743   ,   24093   ,   64047   ,0 };
            const unsigned long dim3927KuoInit[]        =   {   1   ,   1   ,   7   ,   11  ,   13  ,   53  ,   45  ,   195 ,   39  ,   379 ,   1699    ,   2401    ,   7131    ,   7205    ,   24419   ,   32161   ,0 };
            const unsigned long dim3928KuoInit[]        =   {   1   ,   3   ,   3   ,   15  ,   31  ,   31  ,   35  ,   199 ,   421 ,   367 ,   1517    ,   1559    ,   3627    ,   10281   ,   27963   ,   31115   ,0 };
            const unsigned long dim3929KuoInit[]        =   {   1   ,   1   ,   5   ,   15  ,   5   ,   59  ,   89  ,   195 ,   263 ,   853 ,   1043    ,   3699    ,   2543    ,   9893    ,   727 ,   62533   ,0 };
            const unsigned long dim3930KuoInit[]        =   {   1   ,   1   ,   7   ,   5   ,   29  ,   49  ,   101 ,   219 ,   135 ,   643 ,   321 ,   83  ,   1651    ,   2347    ,   1519    ,   21683   ,0 };
            const unsigned long dim3931KuoInit[]        =   {   1   ,   1   ,   3   ,   5   ,   11  ,   17  ,   3   ,   75  ,   255 ,   359 ,   1231    ,   2333    ,   6745    ,   2393    ,   4213    ,   7689    ,0 };
            const unsigned long dim3932KuoInit[]        =   {   1   ,   1   ,   5   ,   13  ,   23  ,   27  ,   53  ,   51  ,   307 ,   87  ,   1409    ,   3555    ,   8139    ,   8597    ,   27357   ,   28489   ,0 };
            const unsigned long dim3933KuoInit[]        =   {   1   ,   3   ,   3   ,   1   ,   29  ,   31  ,   53  ,   175 ,   245 ,   815 ,   461 ,   1681    ,   2255    ,   761 ,   18061   ,   13273   ,0 };
            const unsigned long dim3934KuoInit[]        =   {   1   ,   3   ,   3   ,   7   ,   21  ,   41  ,   81  ,   243 ,   9   ,   973 ,   1411    ,   2327    ,   4575    ,   2659    ,   22903   ,   27605   ,0 };
            const unsigned long dim3935KuoInit[]        =   {   1   ,   1   ,   3   ,   7   ,   1   ,   3   ,   91  ,   203 ,   149 ,   101 ,   1745    ,   1427    ,   4107    ,   11967   ,   13545   ,   19249   ,0 };
            const unsigned long dim3936KuoInit[]        =   {   1   ,   3   ,   1   ,   1   ,   3   ,   13  ,   47  ,   13  ,   291 ,   813 ,   97  ,   2741    ,   2847    ,   2845    ,   6437    ,   1497    ,0 };
            const unsigned long dim3937KuoInit[]        =   {   1   ,   3   ,   3   ,   1   ,   15  ,   7   ,   121 ,   173 ,   333 ,   977 ,   357 ,   3439    ,   8023    ,   8741    ,   205 ,   11447   ,0 };
            const unsigned long dim3938KuoInit[]        =   {   1   ,   1   ,   3   ,   13  ,   17  ,   57  ,   79  ,   3   ,   455 ,   673 ,   1877    ,   2019    ,   4809    ,   10635   ,   19927   ,   38703   ,0 };
            const unsigned long dim3939KuoInit[]        =   {   1   ,   1   ,   1   ,   15  ,   17  ,   23  ,   65  ,   157 ,   265 ,   177 ,   697 ,   3281    ,   409 ,   471 ,   113 ,   16349   ,0 };
            const unsigned long dim3940KuoInit[]        =   {   1   ,   3   ,   3   ,   11  ,   19  ,   57  ,   125 ,   55  ,   343 ,   177 ,   621 ,   1827    ,   983 ,   2491    ,   10389   ,   49879   ,0 };
            const unsigned long dim3941KuoInit[]        =   {   1   ,   1   ,   5   ,   1   ,   23  ,   21  ,   125 ,   13  ,   155 ,   459 ,   1677    ,   2585    ,   6703    ,   945 ,   5341    ,   54821   ,0 };
            const unsigned long dim3942KuoInit[]        =   {   1   ,   3   ,   3   ,   1   ,   1   ,   59  ,   75  ,   137 ,   261 ,   563 ,   1145    ,   759 ,   2363    ,   12949   ,   16677   ,   43535   ,0 };
            const unsigned long dim3943KuoInit[]        =   {   1   ,   3   ,   1   ,   1   ,   31  ,   13  ,   29  ,   59  ,   263 ,   879 ,   49  ,   1365    ,   4977    ,   6981    ,   32683   ,   25063   ,0 };
            const unsigned long dim3944KuoInit[]        =   {   1   ,   3   ,   3   ,   11  ,   27  ,   63  ,   41  ,   101 ,   353 ,   611 ,   349 ,   1947    ,   6473    ,   3099    ,   28855   ,   27945   ,0 };
            const unsigned long dim3945KuoInit[]        =   {   1   ,   1   ,   5   ,   3   ,   13  ,   49  ,   59  ,   5   ,   401 ,   679 ,   2045    ,   2267    ,   5879    ,   12809   ,   23767   ,   47855   ,0 };
            const unsigned long dim3946KuoInit[]        =   {   1   ,   1   ,   5   ,   11  ,   17  ,   1   ,   25  ,   235 ,   67  ,   155 ,   629 ,   25  ,   2565    ,   12261   ,   28847   ,   19987   ,0 };
            const unsigned long dim3947KuoInit[]        =   {   1   ,   3   ,   1   ,   1   ,   23  ,   39  ,   117 ,   147 ,   155 ,   797 ,   1487    ,   921 ,   1205    ,   9947    ,   14603   ,   21335   ,0 };
            const unsigned long dim3948KuoInit[]        =   {   1   ,   1   ,   3   ,   13  ,   21  ,   45  ,   49  ,   175 ,   447 ,   303 ,   531 ,   2923    ,   3887    ,   16273   ,   5121    ,   38965   ,0 };
            const unsigned long dim3949KuoInit[]        =   {   1   ,   1   ,   7   ,   11  ,   5   ,   21  ,   53  ,   201 ,   1   ,   207 ,   181 ,   1263    ,   833 ,   5435    ,   28363   ,   62913   ,0 };
            const unsigned long dim3950KuoInit[]        =   {   1   ,   1   ,   3   ,   5   ,   11  ,   31  ,   1   ,   63  ,   343 ,   221 ,   23  ,   1315    ,   7639    ,   9983    ,   9167    ,   63489   ,0 };
            const unsigned long dim3951KuoInit[]        =   {   1   ,   3   ,   3   ,   1   ,   21  ,   43  ,   7   ,   161 ,   471 ,   749 ,   1547    ,   1645    ,   1557    ,   9719    ,   14017   ,   43281   ,0 };
            const unsigned long dim3952KuoInit[]        =   {   1   ,   1   ,   1   ,   7   ,   3   ,   23  ,   49  ,   29  ,   259 ,   367 ,   1929    ,   3925    ,   4197    ,   8353    ,   28995   ,   41225   ,0 };
            const unsigned long dim3953KuoInit[]        =   {   1   ,   3   ,   1   ,   9   ,   5   ,   31  ,   33  ,   195 ,   33  ,   171 ,   1021    ,   127 ,   5275    ,   10751   ,   12483   ,   7097    ,0 };
            const unsigned long dim3954KuoInit[]        =   {   1   ,   3   ,   5   ,   3   ,   19  ,   21  ,   3   ,   113 ,   217 ,   953 ,   419 ,   1689    ,   8011    ,   805 ,   20691   ,   39159   ,0 };
            const unsigned long dim3955KuoInit[]        =   {   1   ,   1   ,   5   ,   1   ,   27  ,   35  ,   83  ,   173 ,   299 ,   687 ,   493 ,   3965    ,   5963    ,   1307    ,   16379   ,   34297   ,0 };
            const unsigned long dim3956KuoInit[]        =   {   1   ,   3   ,   7   ,   3   ,   11  ,   63  ,   103 ,   199 ,   207 ,   25  ,   1471    ,   2047    ,   7271    ,   7735    ,   4313    ,   45205   ,0 };
            const unsigned long dim3957KuoInit[]        =   {   1   ,   3   ,   1   ,   7   ,   9   ,   23  ,   33  ,   207 ,   213 ,   125 ,   1871    ,   2073    ,   53  ,   9953    ,   1327    ,   41943   ,0 };
            const unsigned long dim3958KuoInit[]        =   {   1   ,   3   ,   7   ,   3   ,   19  ,   55  ,   3   ,   159 ,   147 ,   649 ,   1161    ,   2245    ,   4303    ,   9971    ,   29633   ,   56535   ,0 };
            const unsigned long dim3959KuoInit[]        =   {   1   ,   3   ,   3   ,   5   ,   1   ,   1   ,   57  ,   101 ,   265 ,   859 ,   137 ,   763 ,   6657    ,   8659    ,   21037   ,   22757   ,0 };
            const unsigned long dim3960KuoInit[]        =   {   1   ,   1   ,   1   ,   7   ,   11  ,   23  ,   39  ,   129 ,   49  ,   987 ,   2013    ,   4071    ,   1289    ,   6471    ,   6231    ,   23987   ,0 };
            const unsigned long dim3961KuoInit[]        =   {   1   ,   3   ,   7   ,   3   ,   29  ,   7   ,   87  ,   253 ,   111 ,   55  ,   1719    ,   321 ,   3579    ,   4489    ,   28323   ,   35091   ,0 };
            const unsigned long dim3962KuoInit[]        =   {   1   ,   1   ,   5   ,   13  ,   17  ,   1   ,   39  ,   73  ,   455 ,   31  ,   1215    ,   2683    ,   6193    ,   15725   ,   2737    ,   24927   ,0 };
            const unsigned long dim3963KuoInit[]        =   {   1   ,   3   ,   1   ,   7   ,   15  ,   63  ,   69  ,   125 ,   127 ,   337 ,   129 ,   2667    ,   4723    ,   8489    ,   6129    ,   14973   ,0 };
            const unsigned long dim3964KuoInit[]        =   {   1   ,   3   ,   1   ,   9   ,   21  ,   37  ,   109 ,   91  ,   499 ,   767 ,   1095    ,   2731    ,   6129    ,   2167    ,   21311   ,   43503   ,0 };
            const unsigned long dim3965KuoInit[]        =   {   1   ,   3   ,   7   ,   7   ,   29  ,   11  ,   87  ,   101 ,   19  ,   699 ,   1601    ,   2341    ,   3623    ,   1799    ,   2253    ,   40705   ,0 };
            const unsigned long dim3966KuoInit[]        =   {   1   ,   1   ,   7   ,   1   ,   27  ,   25  ,   111 ,   91  ,   49  ,   3   ,   13  ,   3851    ,   3735    ,   13387   ,   4501    ,   64303   ,0 };
            const unsigned long dim3967KuoInit[]        =   {   1   ,   1   ,   3   ,   13  ,   21  ,   63  ,   7   ,   65  ,   241 ,   925 ,   705 ,   3745    ,   827 ,   7227    ,   1869    ,   60375   ,0 };
            const unsigned long dim3968KuoInit[]        =   {   1   ,   3   ,   1   ,   7   ,   31  ,   45  ,   5   ,   51  ,   45  ,   595 ,   1069    ,   3139    ,   5843    ,   2581    ,   22573   ,   53297   ,0 };
            const unsigned long dim3969KuoInit[]        =   {   1   ,   1   ,   5   ,   3   ,   5   ,   9   ,   89  ,   91  ,   349 ,   241 ,   369 ,   279 ,   617 ,   4479    ,   12761   ,   14473   ,0 };
            const unsigned long dim3970KuoInit[]        =   {   1   ,   3   ,   1   ,   1   ,   13  ,   45  ,   91  ,   29  ,   219 ,   341 ,   1831    ,   1993    ,   3661    ,   1855    ,   7243    ,   3477    ,0 };
            const unsigned long dim3971KuoInit[]        =   {   1   ,   3   ,   3   ,   13  ,   15  ,   17  ,   27  ,   123 ,   223 ,   127 ,   111 ,   2375    ,   3919    ,   7569    ,   4419    ,   24957   ,0 };
            const unsigned long dim3972KuoInit[]        =   {   1   ,   3   ,   7   ,   1   ,   17  ,   1   ,   49  ,   111 ,   103 ,   67  ,   75  ,   3109    ,   1783    ,   13291   ,   5409    ,   33123   ,0 };
            const unsigned long dim3973KuoInit[]        =   {   1   ,   3   ,   5   ,   3   ,   1   ,   51  ,   107 ,   89  ,   233 ,   51  ,   745 ,   2005    ,   5485    ,   7537    ,   13353   ,   34579   ,0 };
            const unsigned long dim3974KuoInit[]        =   {   1   ,   1   ,   7   ,   3   ,   9   ,   3   ,   25  ,   95  ,   413 ,   461 ,   1597    ,   767 ,   3183    ,   10285   ,   10579   ,   26581   ,0 };
            const unsigned long dim3975KuoInit[]        =   {   1   ,   1   ,   5   ,   7   ,   25  ,   3   ,   27  ,   93  ,   433 ,   59  ,   701 ,   1825    ,   6375    ,   6477    ,   6743    ,   42101   ,0 };
            const unsigned long dim3976KuoInit[]        =   {   1   ,   3   ,   1   ,   3   ,   19  ,   45  ,   95  ,   207 ,   267 ,   623 ,   1523    ,   1541    ,   3955    ,   4681    ,   2365    ,   16661   ,0 };
            const unsigned long dim3977KuoInit[]        =   {   1   ,   1   ,   3   ,   13  ,   19  ,   35  ,   83  ,   151 ,   159 ,   825 ,   205 ,   2215    ,   4767    ,   797 ,   12821   ,   14951   ,0 };
            const unsigned long dim3978KuoInit[]        =   {   1   ,   1   ,   5   ,   13  ,   15  ,   5   ,   99  ,   141 ,   21  ,   537 ,   657 ,   2055    ,   1571    ,   12725   ,   4227    ,   21061   ,0 };
            const unsigned long dim3979KuoInit[]        =   {   1   ,   1   ,   1   ,   13  ,   1   ,   39  ,   85  ,   229 ,   9   ,   721 ,   1117    ,   519 ,   1763    ,   13211   ,   32533   ,   41393   ,0 };
            const unsigned long dim3980KuoInit[]        =   {   1   ,   1   ,   3   ,   15  ,   23  ,   11  ,   125 ,   59  ,   439 ,   887 ,   1273    ,   3707    ,   7737    ,   4115    ,   7013    ,   47453   ,0 };
            const unsigned long dim3981KuoInit[]        =   {   1   ,   1   ,   1   ,   5   ,   9   ,   25  ,   53  ,   203 ,   9   ,   189 ,   471 ,   637 ,   5983    ,   1693    ,   8425    ,   52181   ,0 };
            const unsigned long dim3982KuoInit[]        =   {   1   ,   1   ,   1   ,   1   ,   29  ,   61  ,   33  ,   87  ,   153 ,   625 ,   539 ,   3595    ,   3161    ,   3929    ,   22825   ,   50113   ,0 };
            const unsigned long dim3983KuoInit[]        =   {   1   ,   3   ,   7   ,   7   ,   1   ,   31  ,   105 ,   169 ,   475 ,   631 ,   1511    ,   3687    ,   5321    ,   3859    ,   19329   ,   9571    ,0 };
            const unsigned long dim3984KuoInit[]        =   {   1   ,   3   ,   7   ,   1   ,   9   ,   13  ,   111 ,   79  ,   245 ,   351 ,   69  ,   59  ,   5561    ,   4713    ,   22789   ,   14929   ,0 };
            const unsigned long dim3985KuoInit[]        =   {   1   ,   1   ,   3   ,   1   ,   1   ,   55  ,   87  ,   147 ,   37  ,   587 ,   479 ,   3175    ,   5437    ,   12267   ,   7057    ,   57899   ,0 };
            const unsigned long dim3986KuoInit[]        =   {   1   ,   1   ,   5   ,   11  ,   15  ,   55  ,   87  ,   53  ,   359 ,   491 ,   343 ,   621 ,   1055    ,   8697    ,   28993   ,   20473   ,0 };
            const unsigned long dim3987KuoInit[]        =   {   1   ,   3   ,   3   ,   15  ,   13  ,   11  ,   63  ,   179 ,   293 ,   639 ,   809 ,   1365    ,   7043    ,   13269   ,   5423    ,   36465   ,0 };
            const unsigned long dim3988KuoInit[]        =   {   1   ,   1   ,   5   ,   11  ,   31  ,   49  ,   27  ,   71  ,   187 ,   377 ,   283 ,   2407    ,   2987    ,   8565    ,   15159   ,   44501   ,0 };
            const unsigned long dim3989KuoInit[]        =   {   1   ,   3   ,   1   ,   1   ,   27  ,   5   ,   79  ,   211 ,   315 ,   645 ,   951 ,   4025    ,   1647    ,   2893    ,   23829   ,   23941   ,0 };
            const unsigned long dim3990KuoInit[]        =   {   1   ,   1   ,   7   ,   15  ,   15  ,   29  ,   81  ,   133 ,   345 ,   573 ,   1605    ,   2693    ,   5293    ,   13579   ,   21793   ,   30981   ,0 };
            const unsigned long dim3991KuoInit[]        =   {   1   ,   3   ,   7   ,   5   ,   31  ,   19  ,   109 ,   141 ,   389 ,   473 ,   1359    ,   1695    ,   6761    ,   3701    ,   27583   ,   47907   ,0 };
            const unsigned long dim3992KuoInit[]        =   {   1   ,   1   ,   3   ,   5   ,   7   ,   43  ,   51  ,   139 ,   81  ,   111 ,   417 ,   2627    ,   2297    ,   12493   ,   24401   ,   42525   ,0 };
            const unsigned long dim3993KuoInit[]        =   {   1   ,   3   ,   7   ,   3   ,   21  ,   9   ,   65  ,   141 ,   119 ,   971 ,   589 ,   2635    ,   6365    ,   1925    ,   15853   ,   47123   ,0 };
            const unsigned long dim3994KuoInit[]        =   {   1   ,   1   ,   3   ,   1   ,   25  ,   15  ,   93  ,   151 ,   443 ,   325 ,   457 ,   3779    ,   7669    ,   2063    ,   20005   ,   6513    ,0 };
            const unsigned long dim3995KuoInit[]        =   {   1   ,   1   ,   5   ,   11  ,   7   ,   13  ,   127 ,   55  ,   183 ,   663 ,   437 ,   3475    ,   6391    ,   8423    ,   8473    ,   41971   ,0 };
            const unsigned long dim3996KuoInit[]        =   {   1   ,   1   ,   5   ,   9   ,   7   ,   15  ,   113 ,   225 ,   259 ,   85  ,   945 ,   645 ,   5129    ,   3657    ,   931 ,   30429   ,0 };
            const unsigned long dim3997KuoInit[]        =   {   1   ,   3   ,   3   ,   15  ,   9   ,   63  ,   23  ,   117 ,   205 ,   295 ,   1325    ,   1349    ,   5521    ,   9399    ,   7907    ,   26373   ,0 };
            const unsigned long dim3998KuoInit[]        =   {   1   ,   3   ,   7   ,   5   ,   17  ,   23  ,   47  ,   75  ,   97  ,   35  ,   1363    ,   793 ,   2811    ,   3491    ,   18173   ,   8353    ,0 };
            const unsigned long dim3999KuoInit[]        =   {   1   ,   3   ,   3   ,   11  ,   25  ,   27  ,   107 ,   9   ,   331 ,   1017    ,   615 ,   1123    ,   1303    ,   5705    ,   14013   ,   39207   ,0 };
            const unsigned long dim4000KuoInit[]        =   {   1   ,   3   ,   1   ,   7   ,   3   ,   17  ,   101 ,   17  ,   461 ,   781 ,   345 ,   2687    ,   5867    ,   1367    ,   17245   ,   18929   ,0 };
            const unsigned long dim4001KuoInit[]        =   {   1   ,   1   ,   1   ,   13  ,   3   ,   53  ,   55  ,   37  ,   257 ,   101 ,   195 ,   1193    ,   5419    ,   4243    ,   6079    ,   32461   ,0 };
            const unsigned long dim4002KuoInit[]        =   {   1   ,   1   ,   1   ,   13  ,   29  ,   21  ,   51  ,   123 ,   145 ,   105 ,   639 ,   2389    ,   6985    ,   2519    ,   12351   ,   45859   ,0 };
            const unsigned long dim4003KuoInit[]        =   {   1   ,   1   ,   3   ,   13  ,   31  ,   17  ,   75  ,   121 ,   11  ,   945 ,   1917    ,   601 ,   3765    ,   14215   ,   19075   ,   26317   ,0 };
            const unsigned long dim4004KuoInit[]        =   {   1   ,   3   ,   7   ,   13  ,   21  ,   3   ,   57  ,   17  ,   97  ,   567 ,   839 ,   135 ,   2441    ,   4613    ,   10511   ,   21927   ,0 };
            const unsigned long dim4005KuoInit[]        =   {   1   ,   3   ,   5   ,   1   ,   21  ,   27  ,   1   ,   109 ,   381 ,   241 ,   1701    ,   1255    ,   1469    ,   8415    ,   18789   ,   1769    ,0 };
            const unsigned long dim4006KuoInit[]        =   {   1   ,   3   ,   1   ,   11  ,   11  ,   45  ,   77  ,   209 ,   55  ,   371 ,   1323    ,   2899    ,   7387    ,   10317   ,   32695   ,   52199   ,0 };
            const unsigned long dim4007KuoInit[]        =   {   1   ,   3   ,   5   ,   9   ,   15  ,   59  ,   45  ,   97  ,   369 ,   247 ,   1621    ,   2037    ,   4569    ,   16043   ,   2225    ,   39055   ,0 };
            const unsigned long dim4008KuoInit[]        =   {   1   ,   1   ,   7   ,   1   ,   21  ,   43  ,   85  ,   195 ,   99  ,   85  ,   1759    ,   547 ,   4371    ,   14467   ,   17551   ,   52185   ,0 };
            const unsigned long dim4009KuoInit[]        =   {   1   ,   1   ,   3   ,   1   ,   17  ,   53  ,   79  ,   203 ,   49  ,   289 ,   1547    ,   1173    ,   1445    ,   11593   ,   29927   ,   63479   ,0 };
            const unsigned long dim4010KuoInit[]        =   {   1   ,   3   ,   3   ,   5   ,   3   ,   61  ,   31  ,   71  ,   321 ,   923 ,   1337    ,   3465    ,   3305    ,   10841   ,   31921   ,   3057    ,0 };
            const unsigned long dim4011KuoInit[]        =   {   1   ,   3   ,   5   ,   11  ,   31  ,   1   ,   39  ,   201 ,   57  ,   913 ,   1953    ,   193 ,   4185    ,   5895    ,   27027   ,   10889   ,0 };
            const unsigned long dim4012KuoInit[]        =   {   1   ,   3   ,   1   ,   9   ,   7   ,   9   ,   89  ,   145 ,   445 ,   643 ,   279 ,   1983    ,   491 ,   11825   ,   1857    ,   11201   ,0 };
            const unsigned long dim4013KuoInit[]        =   {   1   ,   3   ,   1   ,   5   ,   9   ,   39  ,   21  ,   245 ,   403 ,   393 ,   779 ,   1419    ,   967 ,   14309   ,   6789    ,   54477   ,0 };
            const unsigned long dim4014KuoInit[]        =   {   1   ,   3   ,   5   ,   7   ,   31  ,   21  ,   65  ,   211 ,   447 ,   263 ,   665 ,   3311    ,   405 ,   1775    ,   31121   ,   54025   ,0 };
            const unsigned long dim4015KuoInit[]        =   {   1   ,   3   ,   5   ,   11  ,   13  ,   9   ,   95  ,   243 ,   139 ,   733 ,   813 ,   3855    ,   2731    ,   11651   ,   2175    ,   9655    ,0 };
            const unsigned long dim4016KuoInit[]        =   {   1   ,   1   ,   1   ,   1   ,   21  ,   5   ,   5   ,   241 ,   415 ,   241 ,   1223    ,   3237    ,   2265    ,   3761    ,   26883   ,   37875   ,0 };
            const unsigned long dim4017KuoInit[]        =   {   1   ,   1   ,   1   ,   13  ,   25  ,   45  ,   97  ,   73  ,   353 ,   393 ,   1217    ,   563 ,   5813    ,   2791    ,   16477   ,   4401    ,0 };
            const unsigned long dim4018KuoInit[]        =   {   1   ,   3   ,   5   ,   13  ,   17  ,   61  ,   1   ,   247 ,   439 ,   1013    ,   501 ,   881 ,   353 ,   4689    ,   18733   ,   46731   ,0 };
            const unsigned long dim4019KuoInit[]        =   {   1   ,   3   ,   1   ,   11  ,   11  ,   29  ,   87  ,   53  ,   461 ,   751 ,   1503    ,   2851    ,   2891    ,   15231   ,   1513    ,   12573   ,0 };
            const unsigned long dim4020KuoInit[]        =   {   1   ,   3   ,   1   ,   3   ,   1   ,   13  ,   95  ,   239 ,   151 ,   159 ,   55  ,   1203    ,   1331    ,   10429   ,   19159   ,   18431   ,0 };
            const unsigned long dim4021KuoInit[]        =   {   1   ,   3   ,   3   ,   15  ,   11  ,   43  ,   81  ,   133 ,   449 ,   523 ,   667 ,   2759    ,   6829    ,   15129   ,   19829   ,   31745   ,0 };
            const unsigned long dim4022KuoInit[]        =   {   1   ,   1   ,   1   ,   13  ,   25  ,   43  ,   115 ,   171 ,   475 ,   553 ,   1615    ,   1421    ,   2363    ,   13091   ,   27505   ,   21247   ,0 };
            const unsigned long dim4023KuoInit[]        =   {   1   ,   3   ,   5   ,   13  ,   1   ,   23  ,   15  ,   25  ,   205 ,   797 ,   1013    ,   1903    ,   1595    ,   16373   ,   29983   ,   23009   ,0 };
            const unsigned long dim4024KuoInit[]        =   {   1   ,   1   ,   1   ,   11  ,   3   ,   27  ,   79  ,   179 ,   147 ,   905 ,   1893    ,   2459    ,   3063    ,   7447    ,   16293   ,   14331   ,0 };
            const unsigned long dim4025KuoInit[]        =   {   1   ,   3   ,   7   ,   1   ,   19  ,   49  ,   41  ,   19  ,   97  ,   559 ,   1213    ,   847 ,   3715    ,   13963   ,   20273   ,   25565   ,0 };
            const unsigned long dim4026KuoInit[]        =   {   1   ,   3   ,   7   ,   3   ,   27  ,   21  ,   75  ,   1   ,   157 ,   715 ,   41  ,   2383    ,   4609    ,   9875    ,   21509   ,   57199   ,0 };
            const unsigned long dim4027KuoInit[]        =   {   1   ,   1   ,   1   ,   9   ,   5   ,   51  ,   93  ,   115 ,   495 ,   687 ,   1805    ,   1369    ,   1525    ,   15841   ,   8897    ,   21023   ,0 };
            const unsigned long dim4028KuoInit[]        =   {   1   ,   1   ,   3   ,   11  ,   21  ,   41  ,   93  ,   107 ,   391 ,   567 ,   223 ,   2331    ,   6009    ,   9405    ,   13137   ,   54727   ,0 };
            const unsigned long dim4029KuoInit[]        =   {   1   ,   3   ,   1   ,   15  ,   5   ,   59  ,   119 ,   165 ,   485 ,   765 ,   243 ,   2589    ,   4631    ,   4045    ,   26867   ,   55329   ,0 };
            const unsigned long dim4030KuoInit[]        =   {   1   ,   1   ,   7   ,   9   ,   17  ,   45  ,   35  ,   219 ,   137 ,   315 ,   293 ,   1137    ,   6675    ,   10693   ,   17455   ,   35603   ,0 };
            const unsigned long dim4031KuoInit[]        =   {   1   ,   3   ,   7   ,   13  ,   11  ,   43  ,   65  ,   205 ,   351 ,   155 ,   1371    ,   3069    ,   1345    ,   859 ,   21385   ,   25923   ,0 };
            const unsigned long dim4032KuoInit[]        =   {   1   ,   1   ,   3   ,   3   ,   11  ,   21  ,   39  ,   195 ,   349 ,   333 ,   687 ,   1155    ,   1491    ,   10453   ,   17615   ,   64801   ,0 };
            const unsigned long dim4033KuoInit[]        =   {   1   ,   3   ,   1   ,   9   ,   17  ,   45  ,   39  ,   137 ,   25  ,   105 ,   441 ,   2437    ,   7715    ,   8501    ,   3999    ,   15913   ,0 };
            const unsigned long dim4034KuoInit[]        =   {   1   ,   1   ,   1   ,   9   ,   31  ,   21  ,   93  ,   31  ,   63  ,   635 ,   331 ,   149 ,   1683    ,   6183    ,   19333   ,   34947   ,0 };
            const unsigned long dim4035KuoInit[]        =   {   1   ,   3   ,   3   ,   5   ,   23  ,   11  ,   15  ,   239 ,   365 ,   983 ,   389 ,   945 ,   1765    ,   6417    ,   4471    ,   41431   ,0 };
            const unsigned long dim4036KuoInit[]        =   {   1   ,   3   ,   1   ,   5   ,   5   ,   49  ,   49  ,   27  ,   417 ,   605 ,   1495    ,   4021    ,   1775    ,   3691    ,   21803   ,   60999   ,0 };
            const unsigned long dim4037KuoInit[]        =   {   1   ,   3   ,   3   ,   5   ,   23  ,   21  ,   33  ,   205 ,   481 ,   105 ,   359 ,   1407    ,   4007    ,   35  ,   10789   ,   27721   ,0 };
            const unsigned long dim4038KuoInit[]        =   {   1   ,   1   ,   5   ,   3   ,   5   ,   33  ,   51  ,   61  ,   29  ,   337 ,   1523    ,   955 ,   4923    ,   1977    ,   16505   ,   21249   ,0 };
            const unsigned long dim4039KuoInit[]        =   {   1   ,   3   ,   1   ,   11  ,   17  ,   19  ,   39  ,   7   ,   443 ,   557 ,   611 ,   4037    ,   5873    ,   15161   ,   449 ,   41747   ,0 };
            const unsigned long dim4040KuoInit[]        =   {   1   ,   1   ,   7   ,   11  ,   5   ,   53  ,   91  ,   227 ,   217 ,   689 ,   7   ,   775 ,   8139    ,   10997   ,   29539   ,   60185   ,0 };
            const unsigned long dim4041KuoInit[]        =   {   1   ,   1   ,   5   ,   1   ,   7   ,   59  ,   117 ,   101 ,   145 ,   779 ,   1359    ,   3375    ,   2493    ,   13075   ,   11475   ,   49507   ,0 };
            const unsigned long dim4042KuoInit[]        =   {   1   ,   1   ,   1   ,   3   ,   13  ,   31  ,   39  ,   219 ,   181 ,   203 ,   19  ,   659 ,   533 ,   2541    ,   7285    ,   29353   ,0 };
            const unsigned long dim4043KuoInit[]        =   {   1   ,   3   ,   3   ,   1   ,   29  ,   33  ,   1   ,   125 ,   283 ,   61  ,   303 ,   1613    ,   4309    ,   1891    ,   7107    ,   25977   ,0 };
            const unsigned long dim4044KuoInit[]        =   {   1   ,   3   ,   5   ,   9   ,   21  ,   59  ,   41  ,   63  ,   123 ,   715 ,   893 ,   3509    ,   6785    ,   11009   ,   17749   ,   18349   ,0 };
            const unsigned long dim4045KuoInit[]        =   {   1   ,   3   ,   1   ,   1   ,   25  ,   23  ,   15  ,   1   ,   147 ,   709 ,   957 ,   1101    ,   2733    ,   14069   ,   18905   ,   12467   ,0 };
            const unsigned long dim4046KuoInit[]        =   {   1   ,   3   ,   3   ,   13  ,   9   ,   63  ,   109 ,   39  ,   123 ,   919 ,   427 ,   435 ,   5241    ,   14169   ,   31179   ,   15887   ,0 };
            const unsigned long dim4047KuoInit[]        =   {   1   ,   1   ,   7   ,   15  ,   13  ,   25  ,   29  ,   187 ,   433 ,   633 ,   1203    ,   113 ,   4577    ,   13349   ,   18407   ,   62825   ,0 };
            const unsigned long dim4048KuoInit[]        =   {   1   ,   1   ,   5   ,   1   ,   15  ,   33  ,   53  ,   209 ,   415 ,   711 ,   1543    ,   1529    ,   4991    ,   9545    ,   19655   ,   40427   ,0 };
            const unsigned long dim4049KuoInit[]        =   {   1   ,   1   ,   5   ,   5   ,   23  ,   59  ,   27  ,   23  ,   7   ,   841 ,   1715    ,   43  ,   4987    ,   2073    ,   13765   ,   36105   ,0 };
            const unsigned long dim4050KuoInit[]        =   {   1   ,   3   ,   3   ,   15  ,   31  ,   5   ,   51  ,   183 ,   319 ,   443 ,   621 ,   349 ,   2835    ,   10847   ,   19059   ,   51689   ,0 };
            const unsigned long dim4051KuoInit[]        =   {   1   ,   3   ,   1   ,   9   ,   5   ,   3   ,   21  ,   81  ,   357 ,   609 ,   2019    ,   2613    ,   333 ,   5699    ,   2485    ,   52105   ,0 };
            const unsigned long dim4052KuoInit[]        =   {   1   ,   1   ,   3   ,   15  ,   3   ,   11  ,   73  ,   121 ,   263 ,   833 ,   65  ,   347 ,   4821    ,   4631    ,   29929   ,   42157   ,0 };
            const unsigned long dim4053KuoInit[]        =   {   1   ,   1   ,   7   ,   3   ,   1   ,   63  ,   7   ,   191 ,   47  ,   405 ,   43  ,   1499    ,   5539    ,   12607   ,   26095   ,   20759   ,0 };
            const unsigned long dim4054KuoInit[]        =   {   1   ,   3   ,   3   ,   5   ,   15  ,   21  ,   109 ,   3   ,   15  ,   57  ,   1669    ,   4033    ,   5251    ,   4465    ,   28057   ,   64943   ,0 };
            const unsigned long dim4055KuoInit[]        =   {   1   ,   1   ,   3   ,   9   ,   7   ,   61  ,   31  ,   17  ,   331 ,   671 ,   1955    ,   1035    ,   7183    ,   6559    ,   19299   ,   10225   ,0 };
            const unsigned long dim4056KuoInit[]        =   {   1   ,   1   ,   5   ,   11  ,   31  ,   15  ,   97  ,   3   ,   361 ,   541 ,   323 ,   1463    ,   6867    ,   12717   ,   9593    ,   15407   ,0 };
            const unsigned long dim4057KuoInit[]        =   {   1   ,   1   ,   1   ,   5   ,   29  ,   39  ,   71  ,   27  ,   305 ,   589 ,   1441    ,   2349    ,   185 ,   447 ,   24223   ,   53591   ,0 };
            const unsigned long dim4058KuoInit[]        =   {   1   ,   1   ,   7   ,   15  ,   15  ,   59  ,   81  ,   77  ,   371 ,   891 ,   31  ,   457 ,   3073    ,   8427    ,   8225    ,   53157   ,0 };
            const unsigned long dim4059KuoInit[]        =   {   1   ,   3   ,   3   ,   11  ,   23  ,   61  ,   71  ,   85  ,   39  ,   77  ,   1583    ,   933 ,   6147    ,   2271    ,   24761   ,   15835   ,0 };
            const unsigned long dim4060KuoInit[]        =   {   1   ,   3   ,   7   ,   11  ,   25  ,   55  ,   41  ,   131 ,   275 ,   603 ,   43  ,   761 ,   5071    ,   14149   ,   4001    ,   6545    ,0 };
            const unsigned long dim4061KuoInit[]        =   {   1   ,   1   ,   5   ,   3   ,   5   ,   1   ,   25  ,   217 ,   445 ,   197 ,   863 ,   3557    ,   585 ,   13599   ,   12329   ,   17331   ,0 };
            const unsigned long dim4062KuoInit[]        =   {   1   ,   3   ,   3   ,   13  ,   9   ,   21  ,   21  ,   89  ,   121 ,   1017    ,   1559    ,   83  ,   5225    ,   13247   ,   12003   ,   10049   ,0 };
            const unsigned long dim4063KuoInit[]        =   {   1   ,   3   ,   3   ,   9   ,   31  ,   15  ,   71  ,   45  ,   161 ,   895 ,   1431    ,   1277    ,   2547    ,   1943    ,   17631   ,   59133   ,0 };
            const unsigned long dim4064KuoInit[]        =   {   1   ,   1   ,   1   ,   9   ,   11  ,   17  ,   29  ,   157 ,   355 ,   89  ,   401 ,   983 ,   8175    ,   11687   ,   20087   ,   7541    ,0 };
            const unsigned long dim4065KuoInit[]        =   {   1   ,   3   ,   7   ,   5   ,   15  ,   45  ,   23  ,   141 ,   365 ,   593 ,   1137    ,   873 ,   2903    ,   8861    ,   25805   ,   22601   ,0 };
            const unsigned long dim4066KuoInit[]        =   {   1   ,   1   ,   3   ,   11  ,   27  ,   17  ,   97  ,   3   ,   359 ,   531 ,   329 ,   437 ,   3483    ,   8653    ,   15953   ,   48333   ,0 };
            const unsigned long dim4067KuoInit[]        =   {   1   ,   3   ,   5   ,   3   ,   9   ,   43  ,   79  ,   223 ,   149 ,   177 ,   1075    ,   545 ,   7235    ,   4271    ,   20153   ,   65481   ,0 };
            const unsigned long dim4068KuoInit[]        =   {   1   ,   3   ,   1   ,   13  ,   5   ,   55  ,   29  ,   51  ,   497 ,   677 ,   167 ,   391 ,   1371    ,   8753    ,   23771   ,   48167   ,0 };
            const unsigned long dim4069KuoInit[]        =   {   1   ,   1   ,   5   ,   3   ,   7   ,   17  ,   7   ,   227 ,   21  ,   521 ,   221 ,   3529    ,   5641    ,   4139    ,   14839   ,   49863   ,0 };
            const unsigned long dim4070KuoInit[]        =   {   1   ,   1   ,   3   ,   9   ,   7   ,   41  ,   17  ,   151 ,   401 ,   407 ,   689 ,   507 ,   6683    ,   12515   ,   26895   ,   60863   ,0 };
            const unsigned long dim4071KuoInit[]        =   {   1   ,   1   ,   3   ,   7   ,   3   ,   53  ,   27  ,   185 ,   119 ,   891 ,   2025    ,   607 ,   6961    ,   4263    ,   7795    ,   34157   ,0 };
            const unsigned long dim4072KuoInit[]        =   {   1   ,   1   ,   3   ,   3   ,   25  ,   19  ,   15  ,   131 ,   323 ,   27  ,   1749    ,   2429    ,   2207    ,   9207    ,   18333   ,   25857   ,0 };
            const unsigned long dim4073KuoInit[]        =   {   1   ,   1   ,   5   ,   5   ,   21  ,   57  ,   61  ,   93  ,   305 ,   497 ,   1465    ,   3639    ,   7695    ,   3903    ,   31201   ,   45699   ,0 };
            const unsigned long dim4074KuoInit[]        =   {   1   ,   1   ,   5   ,   11  ,   25  ,   35  ,   93  ,   55  ,   415 ,   397 ,   1375    ,   809 ,   6205    ,   9761    ,   23129   ,   28757   ,0 };
            const unsigned long dim4075KuoInit[]        =   {   1   ,   3   ,   5   ,   9   ,   13  ,   21  ,   1   ,   167 ,   65  ,   689 ,   525 ,   2839    ,   4069    ,   10123   ,   24403   ,   4121    ,0 };
            const unsigned long dim4076KuoInit[]        =   {   1   ,   1   ,   3   ,   13  ,   11  ,   21  ,   67  ,   159 ,   131 ,   579 ,   1397    ,   251 ,   3933    ,   14737   ,   22047   ,   4173    ,0 };
            const unsigned long dim4077KuoInit[]        =   {   1   ,   1   ,   3   ,   13  ,   9   ,   47  ,   71  ,   81  ,   141 ,   607 ,   1879    ,   3057    ,   6493    ,   1177    ,   27261   ,   59313   ,0 };
            const unsigned long dim4078KuoInit[]        =   {   1   ,   1   ,   3   ,   11  ,   25  ,   45  ,   35  ,   193 ,   251 ,   223 ,   1725    ,   1533    ,   5799    ,   12417   ,   29765   ,   50203   ,0 };
            const unsigned long dim4079KuoInit[]        =   {   1   ,   3   ,   1   ,   13  ,   27  ,   1   ,   41  ,   171 ,   293 ,   65  ,   607 ,   2843    ,   2955    ,   8783    ,   21249   ,   41695   ,0 };
            const unsigned long dim4080KuoInit[]        =   {   1   ,   3   ,   1   ,   3   ,   3   ,   57  ,   39  ,   31  ,   507 ,   687 ,   2009    ,   2881    ,   4595    ,   9275    ,   9773    ,   31489   ,0 };
            const unsigned long dim4081KuoInit[]        =   {   1   ,   3   ,   1   ,   7   ,   17  ,   11  ,   47  ,   233 ,   193 ,   395 ,   1879    ,   1983    ,   4811    ,   1221    ,   27279   ,   52089   ,0 };
            const unsigned long dim4082KuoInit[]        =   {   1   ,   3   ,   1   ,   1   ,   13  ,   57  ,   51  ,   253 ,   381 ,   779 ,   897 ,   2893    ,   4373    ,   6599    ,   7533    ,   30359   ,0 };
            const unsigned long dim4083KuoInit[]        =   {   1   ,   3   ,   3   ,   3   ,   5   ,   35  ,   1   ,   117 ,   81  ,   35  ,   2005    ,   2133    ,   4493    ,   9773    ,   24265   ,   45945   ,0 };
            const unsigned long dim4084KuoInit[]        =   {   1   ,   1   ,   3   ,   1   ,   21  ,   57  ,   67  ,   157 ,   111 ,   857 ,   899 ,   2081    ,   2097    ,   16005   ,   20535   ,   2955    ,0 };
            const unsigned long dim4085KuoInit[]        =   {   1   ,   1   ,   1   ,   11  ,   19  ,   45  ,   109 ,   83  ,   375 ,   1003    ,   79  ,   3801    ,   3433    ,   4429    ,   32385   ,   31551   ,0 };
            const unsigned long dim4086KuoInit[]        =   {   1   ,   1   ,   3   ,   3   ,   1   ,   15  ,   115 ,   47  ,   511 ,   533 ,   1711    ,   2433    ,   3027    ,   13511   ,   23743   ,   23805   ,0 };
            const unsigned long dim4087KuoInit[]        =   {   1   ,   1   ,   7   ,   15  ,   31  ,   11  ,   107 ,   215 ,   431 ,   935 ,   1033    ,   213 ,   7913    ,   557 ,   22063   ,   7947    ,0 };
            const unsigned long dim4088KuoInit[]        =   {   1   ,   3   ,   3   ,   1   ,   23  ,   25  ,   109 ,   183 ,   445 ,   477 ,   1297    ,   2269    ,   3593    ,   5657    ,   995 ,   19821   ,0 };
            const unsigned long dim4089KuoInit[]        =   {   1   ,   3   ,   3   ,   9   ,   5   ,   19  ,   93  ,   137 ,   155 ,   5   ,   1905    ,   1551    ,   6321    ,   11213   ,   6305    ,   5063    ,0 };
            const unsigned long dim4090KuoInit[]        =   {   1   ,   1   ,   3   ,   15  ,   9   ,   5   ,   123 ,   203 ,   379 ,   159 ,   1871    ,   2377    ,   665 ,   4877    ,   17663   ,   60371   ,0 };
            const unsigned long dim4091KuoInit[]        =   {   1   ,   1   ,   1   ,   15  ,   3   ,   7   ,   109 ,   15  ,   109 ,   117 ,   79  ,   3061    ,   667 ,   6963    ,   30565   ,   25745   ,0 };
            const unsigned long dim4092KuoInit[]        =   {   1   ,   3   ,   5   ,   1   ,   15  ,   7   ,   105 ,   139 ,   29  ,   923 ,   1449    ,   1349    ,   2951    ,   7443    ,   10243   ,   51739   ,0 };
            const unsigned long dim4093KuoInit[]        =   {   1   ,   1   ,   7   ,   5   ,   5   ,   47  ,   91  ,   77  ,   437 ,   483 ,   1309    ,   1559    ,   6491    ,   13469   ,   9461    ,   40993   ,0 };
            const unsigned long dim4094KuoInit[]        =   {   1   ,   1   ,   7   ,   5   ,   5   ,   35  ,   57  ,   139 ,   207 ,   267 ,   707 ,   1887    ,   3447    ,   3959    ,   8169    ,   14511   ,0 };
            const unsigned long dim4095KuoInit[]        =   {   1   ,   1   ,   7   ,   15  ,   23  ,   9   ,   57  ,   183 ,   437 ,   595 ,   1317    ,   3847    ,   2515    ,   13415   ,   26651   ,   41885   ,0 };
            const unsigned long dim4096KuoInit[]        =   {   1   ,   3   ,   3   ,   3   ,   21  ,   49  ,   71  ,   129 ,   15  ,   183 ,   315 ,   57  ,   1739    ,   15119   ,   15293   ,   8489    ,0 };
            const unsigned long dim4097KuoInit[]        =   {   1   ,   3   ,   1   ,   3   ,   7   ,   63  ,   91  ,   241 ,   355 ,   921 ,   1827    ,   1   ,   5713    ,   3233    ,   21901   ,   48009   ,0 };
            const unsigned long dim4098KuoInit[]        =   {   1   ,   3   ,   7   ,   13  ,   31  ,   9   ,   103 ,   209 ,   289 ,   575 ,   1543    ,   653 ,   7189    ,   1843    ,   5205    ,   27325   ,0 };
            const unsigned long dim4099KuoInit[]        =   {   1   ,   1   ,   7   ,   7   ,   15  ,   63  ,   7   ,   71  ,   95  ,   861 ,   857 ,   701 ,   6843    ,   10609   ,   11177   ,   21043   ,0 };
            const unsigned long dim4100KuoInit[]        =   {   1   ,   3   ,   3   ,   9   ,   25  ,   31  ,   85  ,   255 ,   247 ,   773 ,   1503    ,   2379    ,   7231    ,   6013    ,   16193   ,   33855   ,0 };
            const unsigned long dim4101KuoInit[]        =   {   1   ,   1   ,   3   ,   11  ,   5   ,   27  ,   125 ,   159 ,   381 ,   665 ,   1059    ,   2527    ,   1423    ,   3055    ,   8363    ,   57053   ,0 };
            const unsigned long dim4102KuoInit[]        =   {   1   ,   1   ,   5   ,   3   ,   25  ,   57  ,   107 ,   89  ,   499 ,   761 ,   909 ,   1487    ,   1569    ,   9353    ,   27477   ,   39373   ,0 };
            const unsigned long dim4103KuoInit[]        =   {   1   ,   3   ,   5   ,   9   ,   1   ,   57  ,   81  ,   27  ,   41  ,   371 ,   1111    ,   2437    ,   3669    ,   6033    ,   28601   ,   38985   ,0 };
            const unsigned long dim4104KuoInit[]        =   {   1   ,   3   ,   5   ,   5   ,   25  ,   63  ,   85  ,   171 ,   187 ,   387 ,   1971    ,   579 ,   2779    ,   7667    ,   25423   ,   61005   ,0 };
            const unsigned long dim4105KuoInit[]        =   {   1   ,   1   ,   5   ,   1   ,   25  ,   37  ,   87  ,   93  ,   387 ,   95  ,   65  ,   3345    ,   6581    ,   13325   ,   9683    ,   25399   ,0 };
            const unsigned long dim4106KuoInit[]        =   {   1   ,   3   ,   5   ,   11  ,   9   ,   33  ,   81  ,   69  ,   483 ,   79  ,   123 ,   3241    ,   2599    ,   7545    ,   17775   ,   34933   ,0 };
            const unsigned long dim4107KuoInit[]        =   {   1   ,   1   ,   7   ,   1   ,   9   ,   41  ,   27  ,   191 ,   385 ,   439 ,   347 ,   1279    ,   4491    ,   8883    ,   31599   ,   27167   ,0 };
            const unsigned long dim4108KuoInit[]        =   {   1   ,   3   ,   5   ,   13  ,   1   ,   57  ,   51  ,   81  ,   161 ,   15  ,   7   ,   3695    ,   1777    ,   14947   ,   27853   ,   22383   ,0 };
            const unsigned long dim4109KuoInit[]        =   {   1   ,   1   ,   5   ,   5   ,   15  ,   21  ,   79  ,   159 ,   85  ,   677 ,   401 ,   665 ,   3547    ,   13659   ,   18035   ,   38181   ,0 };
            const unsigned long dim4110KuoInit[]        =   {   1   ,   3   ,   3   ,   15  ,   11  ,   55  ,   29  ,   219 ,   45  ,   683 ,   673 ,   2653    ,   6841    ,   7809    ,   8969    ,   17773   ,0 };
            const unsigned long dim4111KuoInit[]        =   {   1   ,   3   ,   5   ,   13  ,   25  ,   61  ,   19  ,   67  ,   469 ,   265 ,   1835    ,   2733    ,   4537    ,   8113    ,   27089   ,   59191   ,0 };
            const unsigned long dim4112KuoInit[]        =   {   1   ,   1   ,   5   ,   11  ,   9   ,   37  ,   99  ,   31  ,   229 ,   635 ,   1615    ,   3073    ,   2161    ,   13917   ,   22197   ,   4331    ,0 };
            const unsigned long dim4113KuoInit[]        =   {   1   ,   3   ,   1   ,   11  ,   13  ,   59  ,   119 ,   65  ,   55  ,   113 ,   189 ,   57  ,   903 ,   11119   ,   3835    ,   30549   ,0 };
            const unsigned long dim4114KuoInit[]        =   {   1   ,   1   ,   1   ,   5   ,   23  ,   25  ,   29  ,   167 ,   199 ,   669 ,   691 ,   3047    ,   7433    ,   15043   ,   22471   ,   39473   ,0 };
            const unsigned long dim4115KuoInit[]        =   {   1   ,   1   ,   5   ,   13  ,   13  ,   15  ,   121 ,   127 ,   343 ,   293 ,   511 ,   1873    ,   4535    ,   6395    ,   10161   ,   52113   ,0 };
            const unsigned long dim4116KuoInit[]        =   {   1   ,   3   ,   1   ,   11  ,   13  ,   7   ,   95  ,   151 ,   187 ,   221 ,   843 ,   2433    ,   5083    ,   15319   ,   5763    ,   46801   ,0 };
            const unsigned long dim4117KuoInit[]        =   {   1   ,   3   ,   5   ,   3   ,   3   ,   57  ,   109 ,   65  ,   203 ,   583 ,   209 ,   2395    ,   5615    ,   14949   ,   10819   ,   5109    ,0 };
            const unsigned long dim4118KuoInit[]        =   {   1   ,   3   ,   7   ,   15  ,   23  ,   41  ,   65  ,   17  ,   351 ,   89  ,   1451    ,   219 ,   5863    ,   15197   ,   22741   ,   16599   ,0 };
            const unsigned long dim4119KuoInit[]        =   {   1   ,   1   ,   3   ,   13  ,   23  ,   49  ,   51  ,   9   ,   365 ,   595 ,   1619    ,   933 ,   2231    ,   15219   ,   105 ,   1657    ,0 };
            const unsigned long dim4120KuoInit[]        =   {   1   ,   3   ,   5   ,   7   ,   27  ,   33  ,   39  ,   243 ,   257 ,   299 ,   479 ,   1999    ,   1781    ,   9267    ,   3631    ,   21449   ,0 };
            const unsigned long dim4121KuoInit[]        =   {   1   ,   3   ,   1   ,   3   ,   13  ,   1   ,   83  ,   9   ,   373 ,   99  ,   1873    ,   3959    ,   1009    ,   31  ,   29329   ,   1489    ,0 };
            const unsigned long dim4122KuoInit[]        =   {   1   ,   1   ,   7   ,   3   ,   15  ,   21  ,   99  ,   25  ,   265 ,   461 ,   467 ,   2475    ,   8149    ,   5685    ,   22679   ,   9787    ,0 };
            const unsigned long dim4123KuoInit[]        =   {   1   ,   1   ,   3   ,   1   ,   5   ,   43  ,   109 ,   55  ,   451 ,   11  ,   1631    ,   999 ,   1181    ,   3377    ,   4807    ,   14675   ,0 };
            const unsigned long dim4124KuoInit[]        =   {   1   ,   3   ,   7   ,   11  ,   5   ,   43  ,   27  ,   253 ,   73  ,   697 ,   81  ,   3739    ,   4267    ,   3801    ,   2913    ,   33117   ,0 };
            const unsigned long dim4125KuoInit[]        =   {   1   ,   3   ,   1   ,   5   ,   11  ,   55  ,   97  ,   3   ,   3   ,   949 ,   155 ,   3575    ,   941 ,   3727    ,   4047    ,   26071   ,0 };
            const unsigned long dim4126KuoInit[]        =   {   1   ,   1   ,   5   ,   3   ,   17  ,   35  ,   3   ,   9   ,   185 ,   1   ,   421 ,   2831    ,   957 ,   4505    ,   6421    ,   16279   ,0 };
            const unsigned long dim4127KuoInit[]        =   {   1   ,   1   ,   3   ,   9   ,   21  ,   43  ,   115 ,   183 ,   231 ,   13  ,   1711    ,   3403    ,   5447    ,   2761    ,   27591   ,   431 ,0 };
            const unsigned long dim4128KuoInit[]        =   {   1   ,   3   ,   3   ,   15  ,   1   ,   25  ,   111 ,   157 ,   429 ,   981 ,   1117    ,   2767    ,   3107    ,   2203    ,   15335   ,   62597   ,0 };
            const unsigned long dim4129KuoInit[]        =   {   1   ,   1   ,   1   ,   9   ,   5   ,   11  ,   59  ,   163 ,   37  ,   101 ,   63  ,   2829    ,   5583    ,   12695   ,   21683   ,   62563   ,0 };
            const unsigned long dim4130KuoInit[]        =   {   1   ,   3   ,   3   ,   15  ,   7   ,   29  ,   39  ,   241 ,   333 ,   321 ,   931 ,   3811    ,   1981    ,   229 ,   18877   ,   6597    ,0 };
            const unsigned long dim4131KuoInit[]        =   {   1   ,   3   ,   3   ,   13  ,   31  ,   3   ,   83  ,   21  ,   253 ,   117 ,   757 ,   2335    ,   291 ,   3723    ,   24961   ,   41985   ,0 };
            const unsigned long dim4132KuoInit[]        =   {   1   ,   1   ,   1   ,   1   ,   23  ,   37  ,   71  ,   227 ,   211 ,   671 ,   757 ,   597 ,   4015    ,   7297    ,   26275   ,   46051   ,0 };
            const unsigned long dim4133KuoInit[]        =   {   1   ,   3   ,   3   ,   11  ,   9   ,   15  ,   23  ,   143 ,   207 ,   235 ,   1215    ,   1401    ,   7487    ,   13397   ,   30443   ,   43735   ,0 };
            const unsigned long dim4134KuoInit[]        =   {   1   ,   1   ,   7   ,   7   ,   23  ,   5   ,   57  ,   169 ,   3   ,   609 ,   1077    ,   109 ,   6029    ,   14891   ,   6641    ,   26525   ,0 };
            const unsigned long dim4135KuoInit[]        =   {   1   ,   3   ,   7   ,   13  ,   7   ,   45  ,   99  ,   187 ,   43  ,   751 ,   1189    ,   1787    ,   4903    ,   10605   ,   15799   ,   61769   ,0 };
            const unsigned long dim4136KuoInit[]        =   {   1   ,   3   ,   1   ,   5   ,   31  ,   51  ,   95  ,   127 ,   149 ,   557 ,   1027    ,   3377    ,   1461    ,   15217   ,   16955   ,   27339   ,0 };
            const unsigned long dim4137KuoInit[]        =   {   1   ,   3   ,   5   ,   11  ,   31  ,   29  ,   11  ,   143 ,   47  ,   709 ,   1413    ,   343 ,   5421    ,   3567    ,   5383    ,   62347   ,0 };
            const unsigned long dim4138KuoInit[]        =   {   1   ,   3   ,   3   ,   11  ,   13  ,   3   ,   33  ,   19  ,   79  ,   751 ,   1979    ,   785 ,   6339    ,   4133    ,   31633   ,   1963    ,0 };
            const unsigned long dim4139KuoInit[]        =   {   1   ,   3   ,   3   ,   9   ,   15  ,   21  ,   117 ,   13  ,   105 ,   519 ,   1575    ,   3041    ,   5465    ,   7167    ,   17443   ,   15781   ,0 };
            const unsigned long dim4140KuoInit[]        =   {   1   ,   1   ,   1   ,   7   ,   29  ,   25  ,   51  ,   237 ,   11  ,   431 ,   1597    ,   751 ,   4115    ,   12345   ,   9451    ,   1417    ,0 };
            const unsigned long dim4141KuoInit[]        =   {   1   ,   3   ,   5   ,   13  ,   1   ,   43  ,   61  ,   141 ,   279 ,   541 ,   149 ,   3783    ,   1503    ,   7597    ,   25469   ,   21143   ,0 };
            const unsigned long dim4142KuoInit[]        =   {   1   ,   1   ,   3   ,   13  ,   23  ,   13  ,   21  ,   53  ,   49  ,   491 ,   135 ,   3079    ,   2593    ,   10611   ,   24467   ,   18627   ,0 };
            const unsigned long dim4143KuoInit[]        =   {   1   ,   3   ,   3   ,   5   ,   19  ,   31  ,   59  ,   139 ,   137 ,   839 ,   799 ,   1081    ,   251 ,   2685    ,   15993   ,   15909   ,0 };
            const unsigned long dim4144KuoInit[]        =   {   1   ,   3   ,   5   ,   1   ,   3   ,   1   ,   33  ,   131 ,   259 ,   137 ,   651 ,   3879    ,   3089    ,   6957    ,   24589   ,   8945    ,0 };
            const unsigned long dim4145KuoInit[]        =   {   1   ,   1   ,   3   ,   1   ,   23  ,   13  ,   127 ,   255 ,   325 ,   451 ,   275 ,   1425    ,   3649    ,   8149    ,   28913   ,   16481   ,0 };
            const unsigned long dim4146KuoInit[]        =   {   1   ,   1   ,   5   ,   3   ,   19  ,   35  ,   55  ,   183 ,   383 ,   419 ,   551 ,   961 ,   5413    ,   3779    ,   10935   ,   30523   ,0 };
            const unsigned long dim4147KuoInit[]        =   {   1   ,   1   ,   7   ,   3   ,   19  ,   35  ,   125 ,   165 ,   21  ,   23  ,   1119    ,   4061    ,   8057    ,   5129    ,   6035    ,   29907   ,0 };
            const unsigned long dim4148KuoInit[]        =   {   1   ,   1   ,   1   ,   11  ,   31  ,   5   ,   9   ,   165 ,   189 ,   117 ,   1371    ,   3171    ,   3723    ,   9051    ,   11071   ,   1669    ,0 };
            const unsigned long dim4149KuoInit[]        =   {   1   ,   1   ,   1   ,   7   ,   23  ,   37  ,   61  ,   241 ,   253 ,   79  ,   241 ,   2629    ,   6791    ,   2305    ,   17085   ,   17163   ,0 };
            const unsigned long dim4150KuoInit[]        =   {   1   ,   1   ,   7   ,   11  ,   25  ,   49  ,   109 ,   139 ,   271 ,   325 ,   1673    ,   1827    ,   319 ,   10729   ,   15739   ,   19971   ,0 };
            const unsigned long dim4151KuoInit[]        =   {   1   ,   3   ,   3   ,   9   ,   15  ,   3   ,   21  ,   153 ,   229 ,   781 ,   2029    ,   3169    ,   2191    ,   7861    ,   26817   ,   44695   ,0 };
            const unsigned long dim4152KuoInit[]        =   {   1   ,   1   ,   1   ,   11  ,   9   ,   45  ,   93  ,   117 ,   361 ,   319 ,   1747    ,   2379    ,   259 ,   15451   ,   2431    ,   29633   ,0 };
            const unsigned long dim4153KuoInit[]        =   {   1   ,   1   ,   3   ,   1   ,   21  ,   3   ,   87  ,   157 ,   389 ,   373 ,   1235    ,   1347    ,   6573    ,   3447    ,   1569    ,   42287   ,0 };
            const unsigned long dim4154KuoInit[]        =   {   1   ,   3   ,   5   ,   1   ,   3   ,   35  ,   27  ,   49  ,   283 ,   281 ,   733 ,   1557    ,   3589    ,   4047    ,   4589    ,   981 ,0 };
            const unsigned long dim4155KuoInit[]        =   {   1   ,   3   ,   5   ,   5   ,   11  ,   15  ,   69  ,   209 ,   319 ,   601 ,   1559    ,   35  ,   185 ,   8659    ,   26943   ,   10885   ,0 };
            const unsigned long dim4156KuoInit[]        =   {   1   ,   3   ,   3   ,   15  ,   15  ,   25  ,   105 ,   241 ,   153 ,   813 ,   481 ,   2269    ,   5677    ,   11121   ,   21781   ,   29427   ,0 };
            const unsigned long dim4157KuoInit[]        =   {   1   ,   1   ,   1   ,   13  ,   1   ,   41  ,   123 ,   235 ,   487 ,   339 ,   1815    ,   241 ,   5673    ,   9319    ,   6031    ,   3785    ,0 };
            const unsigned long dim4158KuoInit[]        =   {   1   ,   1   ,   3   ,   9   ,   3   ,   15  ,   111 ,   207 ,   281 ,   7   ,   1089    ,   709 ,   8001    ,   14277   ,   27351   ,   21503   ,0 };
            const unsigned long dim4159KuoInit[]        =   {   1   ,   3   ,   1   ,   11  ,   29  ,   11  ,   69  ,   111 ,   471 ,   549 ,   1731    ,   115 ,   1067    ,   8417    ,   19377   ,   18729   ,0 };
            const unsigned long dim4160KuoInit[]        =   {   1   ,   1   ,   7   ,   1   ,   29  ,   11  ,   31  ,   251 ,   437 ,   217 ,   1815    ,   1229    ,   5735    ,   11813   ,   20831   ,   27237   ,0 };
            const unsigned long dim4161KuoInit[]        =   {   1   ,   1   ,   3   ,   9   ,   25  ,   63  ,   115 ,   45  ,   71  ,   359 ,   1313    ,   543 ,   1837    ,   5359    ,   13635   ,   47057   ,0 };
            const unsigned long dim4162KuoInit[]        =   {   1   ,   3   ,   1   ,   13  ,   5   ,   21  ,   73  ,   47  ,   13  ,   553 ,   471 ,   465 ,   591 ,   1943    ,   12549   ,   20245   ,0 };
            const unsigned long dim4163KuoInit[]        =   {   1   ,   3   ,   5   ,   1   ,   5   ,   35  ,   1   ,   5   ,   403 ,   797 ,   987 ,   1227    ,   2847    ,   931 ,   7417    ,   7551    ,0 };
            const unsigned long dim4164KuoInit[]        =   {   1   ,   3   ,   3   ,   11  ,   15  ,   13  ,   7   ,   201 ,   145 ,   475 ,   1557    ,   2871    ,   3161    ,   3489    ,   7893    ,   1745    ,0 };
            const unsigned long dim4165KuoInit[]        =   {   1   ,   1   ,   1   ,   7   ,   15  ,   27  ,   9   ,   129 ,   83  ,   31  ,   1351    ,   647 ,   6217    ,   15201   ,   21019   ,   39925   ,0 };
            const unsigned long dim4166KuoInit[]        =   {   1   ,   1   ,   7   ,   3   ,   19  ,   39  ,   99  ,   253 ,   491 ,   455 ,   405 ,   131 ,   4755    ,   8395    ,   26647   ,   65287   ,0 };
            const unsigned long dim4167KuoInit[]        =   {   1   ,   3   ,   5   ,   13  ,   13  ,   13  ,   15  ,   49  ,   397 ,   237 ,   1491    ,   1881    ,   4233    ,   15801   ,   25591   ,   47061   ,0 };
            const unsigned long dim4168KuoInit[]        =   {   1   ,   1   ,   1   ,   15  ,   11  ,   11  ,   11  ,   161 ,   225 ,   257 ,   231 ,   3303    ,   5145    ,   5217    ,   21527   ,   50231   ,0 };
            const unsigned long dim4169KuoInit[]        =   {   1   ,   3   ,   1   ,   13  ,   21  ,   51  ,   11  ,   161 ,   55  ,   945 ,   1835    ,   1493    ,   5069    ,   9443    ,   31181   ,   2157    ,0 };
            const unsigned long dim4170KuoInit[]        =   {   1   ,   3   ,   1   ,   15  ,   29  ,   45  ,   75  ,   141 ,   475 ,   43  ,   219 ,   2689    ,   7579    ,   523 ,   27125   ,   48615   ,0 };
            const unsigned long dim4171KuoInit[]        =   {   1   ,   3   ,   7   ,   13  ,   5   ,   7   ,   125 ,   13  ,   73  ,   231 ,   1397    ,   295 ,   3197    ,   4023    ,   2167    ,   42101   ,0 };
            const unsigned long dim4172KuoInit[]        =   {   1   ,   1   ,   5   ,   11  ,   5   ,   15  ,   99  ,   241 ,   1   ,   49  ,   1635    ,   2467    ,   4939    ,   9411    ,   145 ,   62117   ,0 };
            const unsigned long dim4173KuoInit[]        =   {   1   ,   3   ,   7   ,   1   ,   3   ,   43  ,   91  ,   179 ,   379 ,   997 ,   1839    ,   3463    ,   4967    ,   12773   ,   26273   ,   52999   ,0 };
            const unsigned long dim4174KuoInit[]        =   {   1   ,   1   ,   1   ,   5   ,   7   ,   45  ,   127 ,   213 ,   63  ,   93  ,   1735    ,   623 ,   153 ,   7039    ,   4233    ,   49775   ,0 };
            const unsigned long dim4175KuoInit[]        =   {   1   ,   3   ,   1   ,   9   ,   19  ,   57  ,   27  ,   225 ,   225 ,   361 ,   1349    ,   829 ,   1441    ,   3557    ,   7595    ,   48405   ,0 };
            const unsigned long dim4176KuoInit[]        =   {   1   ,   1   ,   7   ,   13  ,   27  ,   19  ,   119 ,   115 ,   15  ,   929 ,   345 ,   2199    ,   8101    ,   5787    ,   3105    ,   58313   ,0 };
            const unsigned long dim4177KuoInit[]        =   {   1   ,   3   ,   3   ,   9   ,   23  ,   41  ,   81  ,   19  ,   445 ,   11  ,   771 ,   3767    ,   6781    ,   6969    ,   28385   ,   8473    ,0 };
            const unsigned long dim4178KuoInit[]        =   {   1   ,   1   ,   1   ,   3   ,   1   ,   49  ,   5   ,   149 ,   383 ,   567 ,   1663    ,   1277    ,   733 ,   7651    ,   4835    ,   56429   ,0 };
            const unsigned long dim4179KuoInit[]        =   {   1   ,   1   ,   7   ,   1   ,   29  ,   53  ,   33  ,   135 ,   491 ,   957 ,   1215    ,   687 ,   1025    ,   607 ,   459 ,   58433   ,0 };
            const unsigned long dim4180KuoInit[]        =   {   1   ,   1   ,   1   ,   5   ,   17  ,   29  ,   67  ,   149 ,   93  ,   169 ,   851 ,   2311    ,   4971    ,   13403   ,   15023   ,   31645   ,0 };
            const unsigned long dim4181KuoInit[]        =   {   1   ,   1   ,   7   ,   15  ,   25  ,   61  ,   127 ,   73  ,   379 ,   279 ,   1407    ,   285 ,   5697    ,   11141   ,   28961   ,   35551   ,0 };
            const unsigned long dim4182KuoInit[]        =   {   1   ,   3   ,   1   ,   3   ,   17  ,   17  ,   123 ,   113 ,   349 ,   203 ,   685 ,   687 ,   6957    ,   13033   ,   25147   ,   61541   ,0 };
            const unsigned long dim4183KuoInit[]        =   {   1   ,   3   ,   7   ,   3   ,   29  ,   47  ,   67  ,   161 ,   163 ,   427 ,   241 ,   2767    ,   6855    ,   4975    ,   19081   ,   4613    ,0 };
            const unsigned long dim4184KuoInit[]        =   {   1   ,   1   ,   1   ,   3   ,   7   ,   51  ,   81  ,   57  ,   107 ,   171 ,   1671    ,   2373    ,   341 ,   4071    ,   12877   ,   55775   ,0 };
            const unsigned long dim4185KuoInit[]        =   {   1   ,   1   ,   7   ,   15  ,   29  ,   63  ,   111 ,   255 ,   189 ,   33  ,   1613    ,   113 ,   4347    ,   549 ,   9793    ,   52611   ,0 };
            const unsigned long dim4186KuoInit[]        =   {   1   ,   3   ,   3   ,   3   ,   17  ,   11  ,   41  ,   103 ,   509 ,   387 ,   1341    ,   663 ,   611 ,   12977   ,   6435    ,   40351   ,0 };
            const unsigned long dim4187KuoInit[]        =   {   1   ,   1   ,   1   ,   5   ,   5   ,   39  ,   13  ,   229 ,   473 ,   159 ,   585 ,   2905    ,   3467    ,   8331    ,   11539   ,   57525   ,0 };
            const unsigned long dim4188KuoInit[]        =   {   1   ,   3   ,   1   ,   7   ,   13  ,   25  ,   87  ,   115 ,   199 ,   611 ,   635 ,   2653    ,   5559    ,   7959    ,   9545    ,   14829   ,0 };
            const unsigned long dim4189KuoInit[]        =   {   1   ,   1   ,   3   ,   7   ,   5   ,   57  ,   53  ,   155 ,   57  ,   301 ,   263 ,   1597    ,   7805    ,   8327    ,   20457   ,   6761    ,0 };
            const unsigned long dim4190KuoInit[]        =   {   1   ,   3   ,   3   ,   5   ,   17  ,   39  ,   105 ,   177 ,   317 ,   681 ,   679 ,   691 ,   7751    ,   959 ,   2989    ,   31533   ,0 };
            const unsigned long dim4191KuoInit[]        =   {   1   ,   1   ,   3   ,   1   ,   23  ,   15  ,   65  ,   223 ,   379 ,   523 ,   1285    ,   3175    ,   1821    ,   15655   ,   25183   ,   9415    ,0 };
            const unsigned long dim4192KuoInit[]        =   {   1   ,   3   ,   3   ,   1   ,   11  ,   15  ,   49  ,   33  ,   243 ,   755 ,   2027    ,   3115    ,   4665    ,   12511   ,   8653    ,   35519   ,0 };
            const unsigned long dim4193KuoInit[]        =   {   1   ,   3   ,   3   ,   7   ,   5   ,   45  ,   45  ,   183 ,   409 ,   107 ,   1277    ,   2975    ,   2291    ,   13839   ,   18513   ,   62993   ,0 };
            const unsigned long dim4194KuoInit[]        =   {   1   ,   3   ,   5   ,   3   ,   21  ,   39  ,   47  ,   175 ,   197 ,   737 ,   707 ,   2769    ,   171 ,   12341   ,   22361   ,   16647   ,0 };
            const unsigned long dim4195KuoInit[]        =   {   1   ,   3   ,   3   ,   1   ,   3   ,   35  ,   127 ,   23  ,   349 ,   477 ,   1347    ,   1939    ,   1199    ,   14441   ,   9795    ,   48051   ,0 };
            const unsigned long dim4196KuoInit[]        =   {   1   ,   3   ,   1   ,   15  ,   23  ,   59  ,   69  ,   227 ,   429 ,   359 ,   507 ,   3615    ,   2331    ,   5741    ,   23905   ,   977 ,0 };
            const unsigned long dim4197KuoInit[]        =   {   1   ,   3   ,   3   ,   1   ,   7   ,   5   ,   111 ,   17  ,   425 ,   697 ,   1957    ,   1771    ,   3495    ,   14373   ,   8385    ,   59295   ,0 };
            const unsigned long dim4198KuoInit[]        =   {   1   ,   3   ,   3   ,   1   ,   1   ,   25  ,   113 ,   237 ,   393 ,   23  ,   575 ,   1117    ,   665 ,   3803    ,   28291   ,   60053   ,0 };
            const unsigned long dim4199KuoInit[]        =   {   1   ,   1   ,   7   ,   7   ,   5   ,   33  ,   43  ,   121 ,   109 ,   201 ,   877 ,   3203    ,   7967    ,   11671   ,   13397   ,   9063    ,0 };
            const unsigned long dim4200KuoInit[]        =   {   1   ,   1   ,   1   ,   9   ,   31  ,   57  ,   31  ,   75  ,   429 ,   471 ,   683 ,   3593    ,   4277    ,   10315   ,   1913    ,   54527   ,0 };
            const unsigned long dim4201KuoInit[]        =   {   1   ,   1   ,   1   ,   3   ,   29  ,   17  ,   69  ,   75  ,   251 ,   595 ,   1127    ,   3015    ,   5621    ,   389 ,   2245    ,   43007   ,0 };
            const unsigned long dim4202KuoInit[]        =   {   1   ,   1   ,   1   ,   5   ,   7   ,   41  ,   127 ,   57  ,   321 ,   971 ,   1217    ,   3731    ,   4667    ,   5919    ,   29871   ,   60487   ,0 };
            const unsigned long dim4203KuoInit[]        =   {   1   ,   3   ,   3   ,   7   ,   21  ,   9   ,   83  ,   109 ,   489 ,   689 ,   1601    ,   2979    ,   405 ,   3131    ,   19409   ,   20739   ,0 };
            const unsigned long dim4204KuoInit[]        =   {   1   ,   3   ,   7   ,   15  ,   31  ,   45  ,   77  ,   129 ,   269 ,   473 ,   427 ,   2751    ,   6841    ,   5957    ,   25297   ,   2301    ,0 };
            const unsigned long dim4205KuoInit[]        =   {   1   ,   3   ,   7   ,   3   ,   27  ,   3   ,   107 ,   99  ,   37  ,   421 ,   1853    ,   2683    ,   7305    ,   467 ,   32481   ,   1085    ,0 };
            const unsigned long dim4206KuoInit[]        =   {   1   ,   3   ,   7   ,   15  ,   25  ,   3   ,   79  ,   7   ,   47  ,   299 ,   731 ,   1321    ,   1917    ,   4315    ,   17177   ,   26499   ,0 };
            const unsigned long dim4207KuoInit[]        =   {   1   ,   3   ,   7   ,   15  ,   5   ,   1   ,   19  ,   163 ,   335 ,   13  ,   1855    ,   3805    ,   6707    ,   2133    ,   19671   ,   53653   ,0 };
            const unsigned long dim4208KuoInit[]        =   {   1   ,   1   ,   3   ,   7   ,   3   ,   5   ,   77  ,   69  ,   31  ,   959 ,   1815    ,   669 ,   5965    ,   2925    ,   16285   ,   42043   ,0 };
            const unsigned long dim4209KuoInit[]        =   {   1   ,   3   ,   5   ,   15  ,   25  ,   5   ,   35  ,   175 ,   55  ,   765 ,   597 ,   703 ,   3141    ,   9825    ,   29067   ,   32817   ,0 };
            const unsigned long dim4210KuoInit[]        =   {   1   ,   1   ,   3   ,   5   ,   1   ,   57  ,   119 ,   85  ,   167 ,   299 ,   561 ,   1045    ,   4255    ,   14663   ,   27787   ,   38785   ,0 };
            const unsigned long dim4211KuoInit[]        =   {   1   ,   3   ,   5   ,   13  ,   5   ,   41  ,   97  ,   31  ,   55  ,   773 ,   707 ,   1083    ,   359 ,   15955   ,   11203   ,   50129   ,0 };
            const unsigned long dim4212KuoInit[]        =   {   1   ,   3   ,   7   ,   5   ,   9   ,   35  ,   101 ,   231 ,   23  ,   301 ,   683 ,   3075    ,   3221    ,   7763    ,   5585    ,   35205   ,0 };
            const unsigned long dim4213KuoInit[]        =   {   1   ,   3   ,   1   ,   13  ,   23  ,   29  ,   83  ,   227 ,   111 ,   307 ,   837 ,   465 ,   5933    ,   5265    ,   4507    ,   8719    ,0 };
            const unsigned long dim4214KuoInit[]        =   {   1   ,   3   ,   3   ,   11  ,   21  ,   9   ,   49  ,   157 ,   271 ,   809 ,   1467    ,   1337    ,   5893    ,   13681   ,   11261   ,   56709   ,0 };
            const unsigned long dim4215KuoInit[]        =   {   1   ,   3   ,   7   ,   1   ,   29  ,   33  ,   89  ,   39  ,   131 ,   339 ,   359 ,   3087    ,   5167    ,   5659    ,   629 ,   53865   ,0 };
            const unsigned long dim4216KuoInit[]        =   {   1   ,   3   ,   1   ,   15  ,   11  ,   23  ,   59  ,   105 ,   201 ,   17  ,   923 ,   613 ,   1665    ,   8931    ,   28457   ,   63933   ,0 };
            const unsigned long dim4217KuoInit[]        =   {   1   ,   1   ,   5   ,   13  ,   17  ,   33  ,   67  ,   73  ,   367 ,   263 ,   1377    ,   393 ,   1561    ,   3409    ,   24453   ,   61637   ,0 };
            const unsigned long dim4218KuoInit[]        =   {   1   ,   1   ,   5   ,   1   ,   29  ,   9   ,   31  ,   51  ,   485 ,   775 ,   741 ,   1427    ,   1857    ,   10243   ,   18101   ,   29795   ,0 };
            const unsigned long dim4219KuoInit[]        =   {   1   ,   1   ,   1   ,   5   ,   5   ,   39  ,   29  ,   137 ,   171 ,   837 ,   347 ,   1271    ,   7365    ,   3783    ,   22731   ,   18131   ,0 };
            const unsigned long dim4220KuoInit[]        =   {   1   ,   1   ,   5   ,   11  ,   1   ,   45  ,   89  ,   175 ,   11  ,   859 ,   355 ,   1033    ,   4767    ,   11495   ,   28761   ,   11335   ,0 };
            const unsigned long dim4221KuoInit[]        =   {   1   ,   1   ,   7   ,   1   ,   25  ,   7   ,   85  ,   229 ,   243 ,   259 ,   383 ,   1163    ,   4833    ,   11283   ,   31573   ,   30891   ,0 };
            const unsigned long dim4222KuoInit[]        =   {   1   ,   3   ,   7   ,   5   ,   11  ,   41  ,   111 ,   87  ,   203 ,   177 ,   1113    ,   415 ,   3613    ,   1327    ,   20103   ,   61295   ,0 };
            const unsigned long dim4223KuoInit[]        =   {   1   ,   3   ,   3   ,   9   ,   15  ,   53  ,   25  ,   245 ,   335 ,   31  ,   1335    ,   3339    ,   2337    ,   5719    ,   3333    ,   51325   ,0 };
            const unsigned long dim4224KuoInit[]        =   {   1   ,   1   ,   5   ,   3   ,   5   ,   29  ,   117 ,   11  ,   239 ,   135 ,   1501    ,   2707    ,   2775    ,   2135    ,   11477   ,   44363   ,0 };
            const unsigned long dim4225KuoInit[]        =   {   1   ,   1   ,   3   ,   1   ,   23  ,   33  ,   3   ,   219 ,   301 ,   643 ,   1555    ,   1   ,   709 ,   11  ,   30649   ,   19729   ,0 };
            const unsigned long dim4226KuoInit[]        =   {   1   ,   1   ,   5   ,   7   ,   31  ,   9   ,   45  ,   185 ,   203 ,   741 ,   1851    ,   981 ,   6437    ,   1559    ,   27245   ,   36797   ,0 };
            const unsigned long dim4227KuoInit[]        =   {   1   ,   1   ,   3   ,   7   ,   1   ,   41  ,   11  ,   141 ,   53  ,   533 ,   927 ,   2733    ,   1049    ,   16191   ,   1241    ,   49695   ,0 };
            const unsigned long dim4228KuoInit[]        =   {   1   ,   3   ,   7   ,   1   ,   31  ,   21  ,   105 ,   181 ,   487 ,   1009    ,   135 ,   3649    ,   6459    ,   3397    ,   1805    ,   29793   ,0 };
            const unsigned long dim4229KuoInit[]        =   {   1   ,   1   ,   3   ,   1   ,   1   ,   29  ,   65  ,   127 ,   461 ,   297 ,   1151    ,   2403    ,   4875    ,   343 ,   24293   ,   47203   ,0 };
            const unsigned long dim4230KuoInit[]        =   {   1   ,   1   ,   7   ,   3   ,   9   ,   1   ,   23  ,   139 ,   337 ,   591 ,   1557    ,   1993    ,   6927    ,   5791    ,   29199   ,   2991    ,0 };
            const unsigned long dim4231KuoInit[]        =   {   1   ,   1   ,   7   ,   9   ,   7   ,   1   ,   81  ,   129 ,   469 ,   215 ,   695 ,   2691    ,   6165    ,   14741   ,   24339   ,   58681   ,0 };
            const unsigned long dim4232KuoInit[]        =   {   1   ,   3   ,   5   ,   13  ,   29  ,   59  ,   113 ,   143 ,   15  ,   399 ,   1141    ,   975 ,   1285    ,   12773   ,   593 ,   15481   ,0 };
            const unsigned long dim4233KuoInit[]        =   {   1   ,   1   ,   3   ,   11  ,   19  ,   41  ,   85  ,   95  ,   479 ,   311 ,   1721    ,   1089    ,   7199    ,   15683   ,   15909   ,   40911   ,0 };
            const unsigned long dim4234KuoInit[]        =   {   1   ,   1   ,   1   ,   3   ,   25  ,   55  ,   23  ,   195 ,   413 ,   623 ,   411 ,   279 ,   3381    ,   15313   ,   19323   ,   22829   ,0 };
            const unsigned long dim4235KuoInit[]        =   {   1   ,   3   ,   5   ,   15  ,   13  ,   1   ,   55  ,   91  ,   493 ,   967 ,   1571    ,   2887    ,   2791    ,   1331    ,   20565   ,   61267   ,0 };
            const unsigned long dim4236KuoInit[]        =   {   1   ,   3   ,   5   ,   1   ,   1   ,   17  ,   127 ,   253 ,   249 ,   559 ,   739 ,   2921    ,   1315    ,   13699   ,   22501   ,   9699    ,0 };
            const unsigned long dim4237KuoInit[]        =   {   1   ,   3   ,   7   ,   13  ,   19  ,   53  ,   31  ,   103 ,   307 ,   821 ,   661 ,   591 ,   5379    ,   6401    ,   10109   ,   7325    ,0 };
            const unsigned long dim4238KuoInit[]        =   {   1   ,   1   ,   1   ,   5   ,   15  ,   35  ,   59  ,   149 ,   443 ,   413 ,   139 ,   1851    ,   3679    ,   10721   ,   31393   ,   37329   ,0 };
            const unsigned long dim4239KuoInit[]        =   {   1   ,   1   ,   3   ,   3   ,   15  ,   47  ,   43  ,   241 ,   323 ,   357 ,   1567    ,   3183    ,   6869    ,   4615    ,   5617    ,   28691   ,0 };
            const unsigned long dim4240KuoInit[]        =   {   1   ,   3   ,   3   ,   1   ,   25  ,   11  ,   55  ,   55  ,   105 ,   277 ,   1317    ,   3631    ,   5351    ,   12747   ,   1685    ,   61239   ,0 };
            const unsigned long dim4241KuoInit[]        =   {   1   ,   1   ,   7   ,   7   ,   31  ,   19  ,   87  ,   119 ,   409 ,   817 ,   97  ,   3549    ,   3517    ,   13901   ,   11707   ,   9501    ,0 };
            const unsigned long dim4242KuoInit[]        =   {   1   ,   1   ,   7   ,   11  ,   13  ,   17  ,   55  ,   195 ,   457 ,   223 ,   837 ,   453 ,   2953    ,   2269    ,   10149   ,   48941   ,0 };
            const unsigned long dim4243KuoInit[]        =   {   1   ,   3   ,   3   ,   13  ,   5   ,   31  ,   63  ,   117 ,   419 ,   329 ,   1759    ,   1167    ,   7967    ,   8975    ,   11247   ,   37159   ,0 };
            const unsigned long dim4244KuoInit[]        =   {   1   ,   1   ,   7   ,   1   ,   17  ,   19  ,   41  ,   61  ,   349 ,   299 ,   1449    ,   2329    ,   471 ,   5119    ,   29595   ,   43335   ,0 };
            const unsigned long dim4245KuoInit[]        =   {   1   ,   1   ,   1   ,   13  ,   13  ,   63  ,   99  ,   239 ,   239 ,   903 ,   55  ,   1347    ,   969 ,   9409    ,   11603   ,   28553   ,0 };
            const unsigned long dim4246KuoInit[]        =   {   1   ,   3   ,   1   ,   7   ,   25  ,   59  ,   23  ,   241 ,   361 ,   1005    ,   771 ,   923 ,   3315    ,   4201    ,   32695   ,   54151   ,0 };
            const unsigned long dim4247KuoInit[]        =   {   1   ,   3   ,   3   ,   5   ,   19  ,   1   ,   113 ,   103 ,   469 ,   591 ,   319 ,   2297    ,   1033    ,   9273    ,   20529   ,   58301   ,0 };
            const unsigned long dim4248KuoInit[]        =   {   1   ,   3   ,   7   ,   11  ,   23  ,   1   ,   77  ,   161 ,   321 ,   855 ,   1987    ,   3735    ,   4533    ,   12061   ,   19305   ,   31679   ,0 };
            const unsigned long dim4249KuoInit[]        =   {   1   ,   1   ,   3   ,   7   ,   27  ,   15  ,   73  ,   83  ,   113 ,   455 ,   1855    ,   2677    ,   3243    ,   14961   ,   27147   ,   21865   ,0 };
            const unsigned long dim4250KuoInit[]        =   {   1   ,   3   ,   5   ,   11  ,   15  ,   33  ,   61  ,   87  ,   237 ,   775 ,   1765    ,   2783    ,   3051    ,   3617    ,   32519   ,   43991   ,0 };
            const unsigned long dim4251KuoInit[]        =   {   1   ,   3   ,   5   ,   1   ,   9   ,   57  ,   21  ,   177 ,   499 ,   605 ,   391 ,   381 ,   2027    ,   9221    ,   11213   ,   10365   ,0 };
            const unsigned long dim4252KuoInit[]        =   {   1   ,   1   ,   7   ,   5   ,   31  ,   29  ,   57  ,   159 ,   413 ,   321 ,   245 ,   2007    ,   2817    ,   5173    ,   23227   ,   19885   ,0 };
            const unsigned long dim4253KuoInit[]        =   {   1   ,   1   ,   7   ,   7   ,   29  ,   43  ,   113 ,   173 ,   505 ,   449 ,   409 ,   981 ,   2257    ,   13261   ,   85  ,   6477    ,0 };
            const unsigned long dim4254KuoInit[]        =   {   1   ,   1   ,   5   ,   13  ,   9   ,   51  ,   1   ,   127 ,   65  ,   963 ,   1541    ,   1321    ,   7849    ,   6481    ,   1583    ,   14737   ,0 };
            const unsigned long dim4255KuoInit[]        =   {   1   ,   1   ,   3   ,   3   ,   1   ,   45  ,   11  ,   157 ,   375 ,   439 ,   1673    ,   271 ,   3853    ,   8795    ,   23287   ,   45365   ,0 };
            const unsigned long dim4256KuoInit[]        =   {   1   ,   1   ,   3   ,   11  ,   17  ,   7   ,   43  ,   3   ,   3   ,   1005    ,   523 ,   2553    ,   2457    ,   259 ,   20999   ,   23611   ,0 };
            const unsigned long dim4257KuoInit[]        =   {   1   ,   3   ,   3   ,   11  ,   17  ,   33  ,   97  ,   227 ,   19  ,   19  ,   2039    ,   3181    ,   1393    ,   1427    ,   215 ,   49915   ,0 };
            const unsigned long dim4258KuoInit[]        =   {   1   ,   1   ,   5   ,   9   ,   19  ,   1   ,   83  ,   95  ,   7   ,   571 ,   655 ,   1281    ,   3841    ,   3925    ,   24603   ,   38413   ,0 };
            const unsigned long dim4259KuoInit[]        =   {   1   ,   3   ,   1   ,   3   ,   19  ,   47  ,   43  ,   105 ,   431 ,   921 ,   927 ,   3713    ,   8181    ,   4177    ,   9411    ,   1481    ,0 };
            const unsigned long dim4260KuoInit[]        =   {   1   ,   3   ,   1   ,   9   ,   3   ,   29  ,   11  ,   121 ,   505 ,   629 ,   959 ,   523 ,   257 ,   3089    ,   15075   ,   14087   ,0 };
            const unsigned long dim4261KuoInit[]        =   {   1   ,   3   ,   5   ,   13  ,   9   ,   3   ,   5   ,   167 ,   7   ,   705 ,   1737    ,   583 ,   2297    ,   13073   ,   27617   ,   22083   ,0 };
            const unsigned long dim4262KuoInit[]        =   {   1   ,   3   ,   3   ,   13  ,   9   ,   51  ,   33  ,   191 ,   263 ,   961 ,   1019    ,   2771    ,   3487    ,   7471    ,   10939   ,   23831   ,0 };
            const unsigned long dim4263KuoInit[]        =   {   1   ,   1   ,   5   ,   5   ,   25  ,   19  ,   75  ,   115 ,   221 ,   93  ,   1135    ,   257 ,   5597    ,   14715   ,   2385    ,   46659   ,0 };
            const unsigned long dim4264KuoInit[]        =   {   1   ,   1   ,   3   ,   5   ,   31  ,   13  ,   63  ,   231 ,   463 ,   501 ,   1437    ,   1777    ,   6161    ,   9885    ,   10543   ,   26031   ,0 };
            const unsigned long dim4265KuoInit[]        =   {   1   ,   1   ,   5   ,   5   ,   5   ,   1   ,   51  ,   141 ,   311 ,   929 ,   261 ,   2959    ,   59  ,   4163    ,   9855    ,   54825   ,0 };
            const unsigned long dim4266KuoInit[]        =   {   1   ,   3   ,   1   ,   9   ,   1   ,   57  ,   31  ,   21  ,   395 ,   211 ,   89  ,   3703    ,   7565    ,   2095    ,   16273   ,   52105   ,0 };
            const unsigned long dim4267KuoInit[]        =   {   1   ,   1   ,   3   ,   11  ,   25  ,   57  ,   27  ,   163 ,   459 ,   849 ,   955 ,   2437    ,   7115    ,   8029    ,   165 ,   12007   ,0 };
            const unsigned long dim4268KuoInit[]        =   {   1   ,   1   ,   3   ,   7   ,   29  ,   57  ,   49  ,   71  ,   389 ,   483 ,   991 ,   629 ,   1517    ,   2481    ,   4881    ,   39431   ,0 };
            const unsigned long dim4269KuoInit[]        =   {   1   ,   3   ,   5   ,   1   ,   1   ,   11  ,   29  ,   81  ,   323 ,   581 ,   1515    ,   905 ,   2813    ,   12221   ,   20181   ,   9855    ,0 };
            const unsigned long dim4270KuoInit[]        =   {   1   ,   3   ,   1   ,   5   ,   27  ,   3   ,   41  ,   65  ,   369 ,   925 ,   1449    ,   2711    ,   3333    ,   6263    ,   10639   ,   43537   ,0 };
            const unsigned long dim4271KuoInit[]        =   {   1   ,   1   ,   1   ,   15  ,   17  ,   5   ,   31  ,   181 ,   77  ,   999 ,   1011    ,   2631    ,   1761    ,   10139   ,   24369   ,   39093   ,0 };
            const unsigned long dim4272KuoInit[]        =   {   1   ,   3   ,   3   ,   11  ,   11  ,   23  ,   65  ,   165 ,   41  ,   591 ,   1045    ,   263 ,   1707    ,   4419    ,   24005   ,   47489   ,0 };
            const unsigned long dim4273KuoInit[]        =   {   1   ,   1   ,   7   ,   13  ,   11  ,   43  ,   65  ,   161 ,   229 ,   35  ,   637 ,   3693    ,   721 ,   6863    ,   12001   ,   48067   ,0 };
            const unsigned long dim4274KuoInit[]        =   {   1   ,   3   ,   3   ,   15  ,   9   ,   39  ,   27  ,   65  ,   407 ,   103 ,   1437    ,   3403    ,   7355    ,   2009    ,   5979    ,   37959   ,0 };
            const unsigned long dim4275KuoInit[]        =   {   1   ,   1   ,   5   ,   5   ,   13  ,   23  ,   65  ,   145 ,   395 ,   603 ,   1711    ,   1529    ,   6309    ,   14631   ,   4615    ,   24459   ,0 };
            const unsigned long dim4276KuoInit[]        =   {   1   ,   3   ,   7   ,   13  ,   25  ,   27  ,   109 ,   91  ,   205 ,   53  ,   199 ,   1215    ,   4371    ,   4821    ,   11793   ,   20719   ,0 };
            const unsigned long dim4277KuoInit[]        =   {   1   ,   1   ,   5   ,   5   ,   9   ,   9   ,   115 ,   77  ,   223 ,   381 ,   1035    ,   1961    ,   5285    ,   12465   ,   20531   ,   51677   ,0 };
            const unsigned long dim4278KuoInit[]        =   {   1   ,   3   ,   3   ,   9   ,   13  ,   31  ,   123 ,   123 ,   325 ,   719 ,   1683    ,   1253    ,   4241    ,   1443    ,   7507    ,   8253    ,0 };
            const unsigned long dim4279KuoInit[]        =   {   1   ,   3   ,   3   ,   7   ,   5   ,   1   ,   71  ,   145 ,   299 ,   1021    ,   677 ,   1733    ,   2791    ,   12337   ,   24395   ,   64145   ,0 };
            const unsigned long dim4280KuoInit[]        =   {   1   ,   1   ,   7   ,   11  ,   21  ,   31  ,   113 ,   79  ,   1   ,   29  ,   1171    ,   4081    ,   7599    ,   11661   ,   4305    ,   11383   ,0 };
            const unsigned long dim4281KuoInit[]        =   {   1   ,   3   ,   3   ,   7   ,   27  ,   9   ,   97  ,   117 ,   205 ,   391 ,   1571    ,   1801    ,   6085    ,   15867   ,   8987    ,   29925   ,0 };
            const unsigned long dim4282KuoInit[]        =   {   1   ,   3   ,   3   ,   3   ,   5   ,   35  ,   27  ,   143 ,   503 ,   533 ,   1293    ,   3737    ,   3011    ,   4457    ,   18381   ,   45513   ,0 };
            const unsigned long dim4283KuoInit[]        =   {   1   ,   3   ,   1   ,   1   ,   21  ,   55  ,   113 ,   127 ,   365 ,   265 ,   1211    ,   841 ,   807 ,   3901    ,   28993   ,   28467   ,0 };
            const unsigned long dim4284KuoInit[]        =   {   1   ,   1   ,   1   ,   7   ,   21  ,   17  ,   101 ,   217 ,   413 ,   273 ,   53  ,   2415    ,   655 ,   9939    ,   11233   ,   31631   ,0 };
            const unsigned long dim4285KuoInit[]        =   {   1   ,   3   ,   7   ,   13  ,   11  ,   11  ,   111 ,   63  ,   301 ,   343 ,   1757    ,   2615    ,   1683    ,   11143   ,   16405   ,   16671   ,0 };
            const unsigned long dim4286KuoInit[]        =   {   1   ,   1   ,   5   ,   15  ,   17  ,   19  ,   83  ,   195 ,   475 ,   189 ,   1881    ,   699 ,   4853    ,   7849    ,   15063   ,   5517    ,0 };
            const unsigned long dim4287KuoInit[]        =   {   1   ,   1   ,   5   ,   5   ,   27  ,   25  ,   19  ,   239 ,   153 ,   725 ,   709 ,   3717    ,   4801    ,   1433    ,   17789   ,   34549   ,0 };
            const unsigned long dim4288KuoInit[]        =   {   1   ,   3   ,   1   ,   9   ,   9   ,   5   ,   47  ,   215 ,   123 ,   161 ,   583 ,   917 ,   5373    ,   6839    ,   20747   ,   47179   ,0 };
            const unsigned long dim4289KuoInit[]        =   {   1   ,   1   ,   1   ,   15  ,   1   ,   25  ,   31  ,   217 ,   185 ,   331 ,   261 ,   3877    ,   2339    ,   11557   ,   22779   ,   58509   ,0 };
            const unsigned long dim4290KuoInit[]        =   {   1   ,   1   ,   7   ,   15  ,   11  ,   1   ,   13  ,   53  ,   319 ,   237 ,   763 ,   2411    ,   7953    ,   2117    ,   24947   ,   26643   ,0 };
            const unsigned long dim4291KuoInit[]        =   {   1   ,   3   ,   5   ,   5   ,   17  ,   53  ,   9   ,   253 ,   219 ,   715 ,   1923    ,   823 ,   1495    ,   1975    ,   3739    ,   61343   ,0 };
            const unsigned long dim4292KuoInit[]        =   {   1   ,   1   ,   1   ,   11  ,   27  ,   31  ,   121 ,   253 ,   319 ,   1001    ,   125 ,   567 ,   215 ,   6879    ,   22581   ,   38495   ,0 };
            const unsigned long dim4293KuoInit[]        =   {   1   ,   1   ,   1   ,   11  ,   23  ,   17  ,   83  ,   65  ,   279 ,   725 ,   1611    ,   3911    ,   509 ,   1729    ,   30025   ,   10791   ,0 };
            const unsigned long dim4294KuoInit[]        =   {   1   ,   3   ,   1   ,   11  ,   19  ,   17  ,   67  ,   101 ,   129 ,   607 ,   557 ,   3475    ,   1445    ,   8789    ,   1815    ,   18171   ,0 };
            const unsigned long dim4295KuoInit[]        =   {   1   ,   3   ,   7   ,   7   ,   23  ,   57  ,   69  ,   127 ,   147 ,   743 ,   549 ,   3269    ,   6193    ,   12667   ,   17081   ,   6899    ,0 };
            const unsigned long dim4296KuoInit[]        =   {   1   ,   3   ,   1   ,   9   ,   9   ,   29  ,   83  ,   175 ,   389 ,   289 ,   367 ,   1259    ,   6631    ,   15675   ,   20975   ,   54067   ,0 };
            const unsigned long dim4297KuoInit[]        =   {   1   ,   1   ,   7   ,   1   ,   27  ,   11  ,   31  ,   209 ,   159 ,   615 ,   1813    ,   3235    ,   7487    ,   7103    ,   5857    ,   19435   ,0 };
            const unsigned long dim4298KuoInit[]        =   {   1   ,   3   ,   7   ,   7   ,   17  ,   3   ,   13  ,   255 ,   405 ,   925 ,   55  ,   3539    ,   3351    ,   359 ,   7633    ,   38209   ,0 };
            const unsigned long dim4299KuoInit[]        =   {   1   ,   3   ,   5   ,   5   ,   5   ,   51  ,   123 ,   61  ,   485 ,   785 ,   1975    ,   121 ,   7841    ,   7253    ,   11769   ,   7389    ,0 };
            const unsigned long dim4300KuoInit[]        =   {   1   ,   1   ,   1   ,   9   ,   1   ,   3   ,   51  ,   169 ,   163 ,   875 ,   35  ,   1573    ,   7931    ,   13611   ,   14697   ,   64843   ,0 };
            const unsigned long dim4301KuoInit[]        =   {   1   ,   1   ,   5   ,   13  ,   11  ,   35  ,   47  ,   135 ,   475 ,   893 ,   39  ,   2763    ,   2671    ,   7417    ,   22905   ,   60593   ,0 };
            const unsigned long dim4302KuoInit[]        =   {   1   ,   3   ,   5   ,   3   ,   27  ,   27  ,   85  ,   219 ,   341 ,   509 ,   1131    ,   3607    ,   1307    ,   15265   ,   28589   ,   61113   ,0 };
            const unsigned long dim4303KuoInit[]        =   {   1   ,   1   ,   5   ,   11  ,   11  ,   37  ,   53  ,   67  ,   177 ,   951 ,   237 ,   3865    ,   2233    ,   1405    ,   28373   ,   47767   ,0 };
            const unsigned long dim4304KuoInit[]        =   {   1   ,   3   ,   7   ,   9   ,   17  ,   49  ,   35  ,   167 ,   67  ,   397 ,   927 ,   3305    ,   2791    ,   3609    ,   3111    ,   4555    ,0 };
            const unsigned long dim4305KuoInit[]        =   {   1   ,   1   ,   5   ,   13  ,   13  ,   41  ,   123 ,   89  ,   65  ,   931 ,   1959    ,   1751    ,   2765    ,   8361    ,   23653   ,   45677   ,0 };
            const unsigned long dim4306KuoInit[]        =   {   1   ,   1   ,   1   ,   3   ,   25  ,   37  ,   91  ,   225 ,   181 ,   275 ,   1515    ,   1355    ,   7971    ,   16081   ,   13715   ,   63963   ,0 };
            const unsigned long dim4307KuoInit[]        =   {   1   ,   3   ,   3   ,   3   ,   7   ,   11  ,   125 ,   181 ,   127 ,   855 ,   1277    ,   3587    ,   1351    ,   4907    ,   16985   ,   357 ,0 };
            const unsigned long dim4308KuoInit[]        =   {   1   ,   3   ,   5   ,   13  ,   13  ,   57  ,   23  ,   99  ,   115 ,   923 ,   1395    ,   3083    ,   3939    ,   475 ,   12385   ,   30113   ,0 };
            const unsigned long dim4309KuoInit[]        =   {   1   ,   3   ,   1   ,   1   ,   1   ,   19  ,   27  ,   73  ,   91  ,   155 ,   1145    ,   467 ,   5655    ,   341 ,   8875    ,   61583   ,0 };
            const unsigned long dim4310KuoInit[]        =   {   1   ,   3   ,   1   ,   1   ,   17  ,   33  ,   113 ,   219 ,   99  ,   817 ,   1731    ,   3379    ,   5729    ,   13615   ,   19047   ,   14143   ,0 };
            const unsigned long dim4311KuoInit[]        =   {   1   ,   1   ,   3   ,   5   ,   29  ,   7   ,   35  ,   111 ,   405 ,   207 ,   1811    ,   3701    ,   6101    ,   7147    ,   13825   ,   52081   ,0 };
            const unsigned long dim4312KuoInit[]        =   {   1   ,   1   ,   7   ,   9   ,   19  ,   23  ,   89  ,   95  ,   503 ,   311 ,   1615    ,   371 ,   5853    ,   4635    ,   31911   ,   65345   ,0 };
            const unsigned long dim4313KuoInit[]        =   {   1   ,   1   ,   7   ,   5   ,   17  ,   41  ,   49  ,   143 ,   445 ,   855 ,   1311    ,   3961    ,   3559    ,   5483    ,   31609   ,   34147   ,0 };
            const unsigned long dim4314KuoInit[]        =   {   1   ,   3   ,   1   ,   15  ,   19  ,   49  ,   67  ,   141 ,   293 ,   757 ,   1905    ,   2513    ,   4405    ,   41  ,   1191    ,   15259   ,0 };
            const unsigned long dim4315KuoInit[]        =   {   1   ,   1   ,   3   ,   11  ,   11  ,   39  ,   113 ,   73  ,   23  ,   117 ,   1315    ,   725 ,   7429    ,   5483    ,   26213   ,   9331    ,0 };
            const unsigned long dim4316KuoInit[]        =   {   1   ,   1   ,   3   ,   3   ,   3   ,   9   ,   119 ,   35  ,   309 ,   515 ,   751 ,   323 ,   2521    ,   12093   ,   1123    ,   43651   ,0 };
            const unsigned long dim4317KuoInit[]        =   {   1   ,   1   ,   3   ,   3   ,   29  ,   43  ,   49  ,   53  ,   481 ,   269 ,   915 ,   1845    ,   3341    ,   511 ,   23845   ,   34623   ,0 };
            const unsigned long dim4318KuoInit[]        =   {   1   ,   1   ,   3   ,   1   ,   23  ,   23  ,   71  ,   225 ,   269 ,   713 ,   21  ,   1051    ,   7635    ,   14245   ,   3947    ,   41757   ,0 };
            const unsigned long dim4319KuoInit[]        =   {   1   ,   3   ,   5   ,   11  ,   5   ,   19  ,   39  ,   99  ,   309 ,   531 ,   1645    ,   787 ,   6913    ,   8159    ,   1359    ,   61973   ,0 };
            const unsigned long dim4320KuoInit[]        =   {   1   ,   1   ,   5   ,   5   ,   5   ,   61  ,   107 ,   91  ,   197 ,   109 ,   369 ,   1053    ,   1995    ,   11551   ,   4751    ,   4981    ,0 };
            const unsigned long dim4321KuoInit[]        =   {   1   ,   3   ,   5   ,   11  ,   29  ,   31  ,   103 ,   21  ,   393 ,   63  ,   1465    ,   2813    ,   995 ,   67  ,   11997   ,   34887   ,0 };
            const unsigned long dim4322KuoInit[]        =   {   1   ,   3   ,   1   ,   7   ,   9   ,   49  ,   111 ,   135 ,   83  ,   79  ,   353 ,   2703    ,   5027    ,   10287   ,   3753    ,   60785   ,0 };
            const unsigned long dim4323KuoInit[]        =   {   1   ,   1   ,   7   ,   15  ,   17  ,   17  ,   51  ,   69  ,   465 ,   25  ,   1231    ,   995 ,   4643    ,   3405    ,   2237    ,   54357   ,0 };
            const unsigned long dim4324KuoInit[]        =   {   1   ,   1   ,   1   ,   1   ,   5   ,   39  ,   31  ,   187 ,   459 ,   197 ,   905 ,   279 ,   2279    ,   6585    ,   7073    ,   1897    ,0 };
            const unsigned long dim4325KuoInit[]        =   {   1   ,   3   ,   3   ,   15  ,   13  ,   17  ,   105 ,   29  ,   183 ,   193 ,   541 ,   977 ,   2247    ,   5173    ,   10511   ,   59969   ,0 };
            const unsigned long dim4326KuoInit[]        =   {   1   ,   1   ,   1   ,   15  ,   3   ,   11  ,   89  ,   1   ,   411 ,   503 ,   267 ,   2875    ,   2585    ,   9391    ,   19395   ,   30107   ,0 };
            const unsigned long dim4327KuoInit[]        =   {   1   ,   1   ,   7   ,   9   ,   7   ,   19  ,   45  ,   251 ,   143 ,   729 ,   1825    ,   2571    ,   6277    ,   12759   ,   5057    ,   58455   ,0 };
            const unsigned long dim4328KuoInit[]        =   {   1   ,   3   ,   3   ,   9   ,   7   ,   49  ,   119 ,   3   ,   127 ,   1015    ,   937 ,   2585    ,   1229    ,   16289   ,   30087   ,   45069   ,0 };
            const unsigned long dim4329KuoInit[]        =   {   1   ,   3   ,   5   ,   13  ,   15  ,   39  ,   39  ,   13  ,   315 ,   533 ,   143 ,   2107    ,   805 ,   1487    ,   8049    ,   3643    ,0 };
            const unsigned long dim4330KuoInit[]        =   {   1   ,   3   ,   1   ,   15  ,   23  ,   35  ,   127 ,   143 ,   247 ,   491 ,   837 ,   2263    ,   1375    ,   8283    ,   24527   ,   37967   ,0 };
            const unsigned long dim4331KuoInit[]        =   {   1   ,   3   ,   7   ,   11  ,   13  ,   9   ,   51  ,   171 ,   421 ,   161 ,   1777    ,   3679    ,   4821    ,   10113   ,   28459   ,   55339   ,0 };
            const unsigned long dim4332KuoInit[]        =   {   1   ,   3   ,   3   ,   1   ,   25  ,   61  ,   53  ,   145 ,   81  ,   509 ,   2001    ,   315 ,   7775    ,   7741    ,   9239    ,   10041   ,0 };
            const unsigned long dim4333KuoInit[]        =   {   1   ,   1   ,   3   ,   3   ,   11  ,   31  ,   21  ,   91  ,   425 ,   415 ,   523 ,   3649    ,   6595    ,   9409    ,   26631   ,   8761    ,0 };
            const unsigned long dim4334KuoInit[]        =   {   1   ,   3   ,   7   ,   7   ,   13  ,   1   ,   51  ,   53  ,   123 ,   125 ,   1055    ,   3951    ,   6621    ,   9207    ,   25991   ,   36435   ,0 };
            const unsigned long dim4335KuoInit[]        =   {   1   ,   3   ,   5   ,   3   ,   21  ,   63  ,   57  ,   35  ,   197 ,   601 ,   665 ,   3223    ,   817 ,   3467    ,   2667    ,   2811    ,0 };
            const unsigned long dim4336KuoInit[]        =   {   1   ,   3   ,   1   ,   11  ,   7   ,   29  ,   51  ,   29  ,   383 ,   969 ,   557 ,   795 ,   7535    ,   513 ,   1329    ,   57567   ,0 };
            const unsigned long dim4337KuoInit[]        =   {   1   ,   1   ,   1   ,   13  ,   11  ,   59  ,   85  ,   183 ,   163 ,   445 ,   1595    ,   133 ,   4277    ,   3323    ,   6833    ,   50359   ,0 };
            const unsigned long dim4338KuoInit[]        =   {   1   ,   3   ,   1   ,   13  ,   9   ,   11  ,   65  ,   209 ,   509 ,   65  ,   1299    ,   2225    ,   7727    ,   3851    ,   17661   ,   39777   ,0 };
            const unsigned long dim4339KuoInit[]        =   {   1   ,   1   ,   5   ,   5   ,   13  ,   31  ,   15  ,   189 ,   43  ,   435 ,   707 ,   83  ,   6713    ,   13721   ,   24013   ,   47037   ,0 };
            const unsigned long dim4340KuoInit[]        =   {   1   ,   1   ,   1   ,   9   ,   29  ,   49  ,   101 ,   205 ,   7   ,   539 ,   1809    ,   4095    ,   7581    ,   8187    ,   16475   ,   10923   ,0 };
            const unsigned long dim4341KuoInit[]        =   {   1   ,   3   ,   3   ,   7   ,   27  ,   39  ,   65  ,   117 ,   183 ,   439 ,   997 ,   1389    ,   2945    ,   13303   ,   10621   ,   63085   ,0 };
            const unsigned long dim4342KuoInit[]        =   {   1   ,   1   ,   3   ,   15  ,   11  ,   21  ,   121 ,   179 ,   37  ,   993 ,   1299    ,   2741    ,   5797    ,   10677   ,   32519   ,   45061   ,0 };
            const unsigned long dim4343KuoInit[]        =   {   1   ,   1   ,   7   ,   1   ,   23  ,   47  ,   107 ,   131 ,   145 ,   679 ,   1519    ,   1147    ,   887 ,   2017    ,   31191   ,   4047    ,0 };
            const unsigned long dim4344KuoInit[]        =   {   1   ,   1   ,   3   ,   11  ,   1   ,   51  ,   1   ,   105 ,   245 ,   269 ,   1067    ,   3551    ,   6323    ,   3729    ,   16351   ,   62965   ,0 };
            const unsigned long dim4345KuoInit[]        =   {   1   ,   1   ,   3   ,   3   ,   17  ,   11  ,   53  ,   163 ,   387 ,   601 ,   75  ,   3219    ,   5683    ,   783 ,   28143   ,   20695   ,0 };
            const unsigned long dim4346KuoInit[]        =   {   1   ,   1   ,   5   ,   15  ,   21  ,   21  ,   33  ,   171 ,   73  ,   157 ,   49  ,   3443    ,   4057    ,   8631    ,   17555   ,   8959    ,0 };
            const unsigned long dim4347KuoInit[]        =   {   1   ,   1   ,   5   ,   9   ,   3   ,   35  ,   13  ,   211 ,   181 ,   425 ,   1367    ,   2217    ,   5949    ,   2741    ,   8141    ,   45769   ,0 };
            const unsigned long dim4348KuoInit[]        =   {   1   ,   3   ,   1   ,   11  ,   27  ,   43  ,   117 ,   3   ,   331 ,   553 ,   1847    ,   2085    ,   7125    ,   10827   ,   10583   ,   60169   ,0 };
            const unsigned long dim4349KuoInit[]        =   {   1   ,   1   ,   3   ,   1   ,   15  ,   35  ,   81  ,   15  ,   259 ,   89  ,   1071    ,   3807    ,   3149    ,   3309    ,   17791   ,   63869   ,0 };
            const unsigned long dim4350KuoInit[]        =   {   1   ,   3   ,   3   ,   5   ,   19  ,   7   ,   113 ,   131 ,   399 ,   691 ,   57  ,   1837    ,   3613    ,   1755    ,   521 ,   64867   ,0 };
            const unsigned long dim4351KuoInit[]        =   {   1   ,   3   ,   1   ,   11  ,   25  ,   53  ,   11  ,   241 ,   187 ,   237 ,   1591    ,   4059    ,   6021    ,   8451    ,   14305   ,   30965   ,0 };
            const unsigned long dim4352KuoInit[]        =   {   1   ,   1   ,   7   ,   11  ,   27  ,   15  ,   9   ,   209 ,   43  ,   137 ,   1351    ,   2553    ,   2841    ,   6375    ,   20291   ,   2523    ,0 };
            const unsigned long dim4353KuoInit[]        =   {   1   ,   3   ,   3   ,   5   ,   27  ,   37  ,   83  ,   61  ,   365 ,   739 ,   69  ,   725 ,   6745    ,   7353    ,   2621    ,   57747   ,0 };
            const unsigned long dim4354KuoInit[]        =   {   1   ,   1   ,   1   ,   1   ,   17  ,   9   ,   53  ,   93  ,   103 ,   467 ,   1543    ,   2353    ,   4901    ,   13269   ,   21635   ,   63425   ,0 };
            const unsigned long dim4355KuoInit[]        =   {   1   ,   3   ,   1   ,   9   ,   15  ,   19  ,   61  ,   127 ,   423 ,   749 ,   809 ,   515 ,   7819    ,   1209    ,   22879   ,   54069   ,0 };
            const unsigned long dim4356KuoInit[]        =   {   1   ,   1   ,   3   ,   7   ,   11  ,   23  ,   93  ,   5   ,   91  ,   613 ,   1529    ,   2621    ,   2689    ,   2251    ,   16049   ,   60773   ,0 };
            const unsigned long dim4357KuoInit[]        =   {   1   ,   1   ,   5   ,   5   ,   15  ,   27  ,   57  ,   177 ,   477 ,   585 ,   249 ,   3673    ,   2651    ,   10137   ,   30395   ,   57475   ,0 };
            const unsigned long dim4358KuoInit[]        =   {   1   ,   3   ,   1   ,   5   ,   31  ,   39  ,   25  ,   185 ,   233 ,   915 ,   1335    ,   2889    ,   1135    ,   12515   ,   27385   ,   54785   ,0 };
            const unsigned long dim4359KuoInit[]        =   {   1   ,   3   ,   5   ,   15  ,   3   ,   19  ,   13  ,   105 ,   465 ,   11  ,   799 ,   1699    ,   6495    ,   14307   ,   18071   ,   55405   ,0 };
            const unsigned long dim4360KuoInit[]        =   {   1   ,   1   ,   1   ,   15  ,   31  ,   49  ,   27  ,   209 ,   503 ,   979 ,   877 ,   3121    ,   3545    ,   12875   ,   799 ,   44427   ,0 };
            const unsigned long dim4361KuoInit[]        =   {   1   ,   1   ,   5   ,   3   ,   23  ,   17  ,   113 ,   109 ,   301 ,   587 ,   1865    ,   1623    ,   453 ,   14449   ,   26901   ,   34687   ,0 };
            const unsigned long dim4362KuoInit[]        =   {   1   ,   3   ,   7   ,   11  ,   1   ,   29  ,   25  ,   113 ,   147 ,   53  ,   27  ,   2971    ,   5663    ,   12013   ,   1281    ,   65173   ,0 };
            const unsigned long dim4363KuoInit[]        =   {   1   ,   3   ,   1   ,   11  ,   13  ,   47  ,   13  ,   209 ,   339 ,   393 ,   1389    ,   2205    ,   537 ,   10911   ,   20093   ,   61499   ,0 };
            const unsigned long dim4364KuoInit[]        =   {   1   ,   1   ,   7   ,   9   ,   21  ,   45  ,   75  ,   43  ,   425 ,   89  ,   1319    ,   2245    ,   6519    ,   12819   ,   30653   ,   18927   ,0 };
            const unsigned long dim4365KuoInit[]        =   {   1   ,   1   ,   3   ,   11  ,   1   ,   31  ,   55  ,   29  ,   337 ,   733 ,   1303    ,   1721    ,   7163    ,   13807   ,   5865    ,   20009   ,0 };
            const unsigned long dim4366KuoInit[]        =   {   1   ,   1   ,   3   ,   15  ,   1   ,   5   ,   101 ,   211 ,   263 ,   865 ,   59  ,   2473    ,   3857    ,   9909    ,   21353   ,   8807    ,0 };
            const unsigned long dim4367KuoInit[]        =   {   1   ,   1   ,   3   ,   9   ,   25  ,   29  ,   5   ,   157 ,   17  ,   985 ,   369 ,   1903    ,   2431    ,   15395   ,   20245   ,   51253   ,0 };
            const unsigned long dim4368KuoInit[]        =   {   1   ,   3   ,   3   ,   11  ,   7   ,   31  ,   45  ,   209 ,   287 ,   13  ,   591 ,   3907    ,   7501    ,   10003   ,   21815   ,   43213   ,0 };
            const unsigned long dim4369KuoInit[]        =   {   1   ,   3   ,   7   ,   5   ,   27  ,   25  ,   73  ,   191 ,   361 ,   731 ,   757 ,   1907    ,   9   ,   16357   ,   4241    ,   40187   ,0 };
            const unsigned long dim4370KuoInit[]        =   {   1   ,   1   ,   7   ,   1   ,   5   ,   55  ,   121 ,   89  ,   39  ,   531 ,   493 ,   3453    ,   5709    ,   8215    ,   21831   ,   5661    ,0 };
            const unsigned long dim4371KuoInit[]        =   {   1   ,   3   ,   7   ,   11  ,   27  ,   59  ,   49  ,   251 ,   17  ,   223 ,   875 ,   1133    ,   3161    ,   10273   ,   26463   ,   45987   ,0 };
            const unsigned long dim4372KuoInit[]        =   {   1   ,   3   ,   5   ,   9   ,   17  ,   39  ,   117 ,   153 ,   243 ,   449 ,   1261    ,   3917    ,   1359    ,   4525    ,   22119   ,   1781    ,0 };
            const unsigned long dim4373KuoInit[]        =   {   1   ,   1   ,   3   ,   7   ,   15  ,   15  ,   19  ,   151 ,   411 ,   693 ,   445 ,   2525    ,   2867    ,   12809   ,   1487    ,   7019    ,0 };
            const unsigned long dim4374KuoInit[]        =   {   1   ,   1   ,   7   ,   7   ,   15  ,   57  ,   61  ,   243 ,   427 ,   529 ,   1051    ,   3391    ,   6259    ,   8791    ,   3405    ,   54757   ,0 };
            const unsigned long dim4375KuoInit[]        =   {   1   ,   3   ,   1   ,   5   ,   13  ,   41  ,   105 ,   73  ,   485 ,   881 ,   1943    ,   1419    ,   5761    ,   11807   ,   16497   ,   12763   ,0 };
            const unsigned long dim4376KuoInit[]        =   {   1   ,   3   ,   3   ,   15  ,   7   ,   47  ,   61  ,   99  ,   473 ,   487 ,   1973    ,   1675    ,   3163    ,   4033    ,   5323    ,   31313   ,0 };
            const unsigned long dim4377KuoInit[]        =   {   1   ,   1   ,   1   ,   3   ,   21  ,   33  ,   23  ,   129 ,   469 ,   743 ,   769 ,   1561    ,   235 ,   13713   ,   6827    ,   61273   ,0 };
            const unsigned long dim4378KuoInit[]        =   {   1   ,   3   ,   1   ,   13  ,   11  ,   19  ,   59  ,   37  ,   431 ,   5   ,   1333    ,   2381    ,   5707    ,   2499    ,   9339    ,   35803   ,0 };
            const unsigned long dim4379KuoInit[]        =   {   1   ,   1   ,   5   ,   1   ,   17  ,   11  ,   77  ,   137 ,   111 ,   567 ,   1469    ,   543 ,   5633    ,   1927    ,   26491   ,   37955   ,0 };
            const unsigned long dim4380KuoInit[]        =   {   1   ,   3   ,   5   ,   15  ,   21  ,   9   ,   57  ,   35  ,   291 ,   371 ,   65  ,   2891    ,   841 ,   11859   ,   16685   ,   5641    ,0 };
            const unsigned long dim4381KuoInit[]        =   {   1   ,   3   ,   7   ,   1   ,   3   ,   35  ,   121 ,   153 ,   353 ,   633 ,   995 ,   121 ,   7317    ,   273 ,   23329   ,   32591   ,0 };
            const unsigned long dim4382KuoInit[]        =   {   1   ,   3   ,   5   ,   1   ,   25  ,   19  ,   35  ,   221 ,   295 ,   543 ,   1269    ,   3901    ,   2109    ,   4463    ,   4275    ,   89  ,0 };
            const unsigned long dim4383KuoInit[]        =   {   1   ,   3   ,   7   ,   13  ,   1   ,   13  ,   85  ,   97  ,   155 ,   731 ,   1017    ,   3713    ,   1613    ,   805 ,   8927    ,   32165   ,0 };
            const unsigned long dim4384KuoInit[]        =   {   1   ,   3   ,   3   ,   13  ,   17  ,   1   ,   21  ,   235 ,   103 ,   875 ,   1849    ,   2669    ,   3193    ,   9053    ,   16123   ,   29481   ,0 };
            const unsigned long dim4385KuoInit[]        =   {   1   ,   3   ,   7   ,   9   ,   23  ,   21  ,   37  ,   251 ,   441 ,   437 ,   629 ,   1009    ,   5663    ,   14447   ,   23437   ,   32521   ,0 };
            const unsigned long dim4386KuoInit[]        =   {   1   ,   1   ,   7   ,   5   ,   9   ,   5   ,   7   ,   123 ,   483 ,   433 ,   867 ,   149 ,   1975    ,   5497    ,   7887    ,   25779   ,0 };
            const unsigned long dim4387KuoInit[]        =   {   1   ,   3   ,   3   ,   3   ,   31  ,   17  ,   81  ,   167 ,   399 ,   1   ,   1777    ,   833 ,   8035    ,   8597    ,   28371   ,   19293   ,0 };
            const unsigned long dim4388KuoInit[]        =   {   1   ,   3   ,   3   ,   11  ,   9   ,   45  ,   125 ,   121 ,   147 ,   613 ,   599 ,   111 ,   7583    ,   3189    ,   13185   ,   35603   ,0 };
            const unsigned long dim4389KuoInit[]        =   {   1   ,   1   ,   1   ,   15  ,   31  ,   51  ,   9   ,   191 ,   55  ,   837 ,   1773    ,   2781    ,   6351    ,   1909    ,   5407    ,   7533    ,0 };
            const unsigned long dim4390KuoInit[]        =   {   1   ,   1   ,   3   ,   13  ,   13  ,   35  ,   71  ,   25  ,   465 ,   377 ,   613 ,   2487    ,   6297    ,   8299    ,   27541   ,   17779   ,0 };
            const unsigned long dim4391KuoInit[]        =   {   1   ,   1   ,   3   ,   15  ,   21  ,   59  ,   43  ,   115 ,   369 ,   315 ,   315 ,   3451    ,   5785    ,   6353    ,   22609   ,   42793   ,0 };
            const unsigned long dim4392KuoInit[]        =   {   1   ,   1   ,   5   ,   15  ,   13  ,   3   ,   127 ,   87  ,   211 ,   981 ,   629 ,   2683    ,   5683    ,   15475   ,   20431   ,   32977   ,0 };
            const unsigned long dim4393KuoInit[]        =   {   1   ,   1   ,   3   ,   7   ,   31  ,   3   ,   23  ,   7   ,   117 ,   1001    ,   431 ,   2215    ,   5817    ,   16283   ,   12781   ,   63883   ,0 };
            const unsigned long dim4394KuoInit[]        =   {   1   ,   1   ,   3   ,   1   ,   27  ,   11  ,   43  ,   195 ,   329 ,   375 ,   655 ,   3815    ,   5699    ,   11259   ,   6733    ,   26015   ,0 };
            const unsigned long dim4395KuoInit[]        =   {   1   ,   3   ,   3   ,   11  ,   19  ,   3   ,   63  ,   187 ,   47  ,   891 ,   1463    ,   3259    ,   1479    ,   15109   ,   25387   ,   43753   ,0 };
            const unsigned long dim4396KuoInit[]        =   {   1   ,   1   ,   5   ,   11  ,   31  ,   43  ,   59  ,   149 ,   205 ,   447 ,   1787    ,   627 ,   6227    ,   9951    ,   19069   ,   50193   ,0 };
            const unsigned long dim4397KuoInit[]        =   {   1   ,   3   ,   3   ,   7   ,   7   ,   1   ,   41  ,   33  ,   395 ,   139 ,   689 ,   1597    ,   529 ,   6861    ,   2025    ,   19569   ,0 };
            const unsigned long dim4398KuoInit[]        =   {   1   ,   1   ,   1   ,   7   ,   17  ,   63  ,   17  ,   73  ,   459 ,   1007    ,   1563    ,   879 ,   1441    ,   525 ,   32679   ,   56409   ,0 };
            const unsigned long dim4399KuoInit[]        =   {   1   ,   3   ,   3   ,   13  ,   3   ,   5   ,   23  ,   239 ,   7   ,   807 ,   1999    ,   1427    ,   6647    ,   2249    ,   17687   ,   5519    ,0 };
            const unsigned long dim4400KuoInit[]        =   {   1   ,   3   ,   5   ,   1   ,   7   ,   15  ,   7   ,   243 ,   347 ,   415 ,   1519    ,   2267    ,   3119    ,   3145    ,   27799   ,   24453   ,0 };
            const unsigned long dim4401KuoInit[]        =   {   1   ,   3   ,   3   ,   3   ,   5   ,   1   ,   93  ,   215 ,   475 ,   209 ,   361 ,   951 ,   147 ,   1401    ,   6851    ,   59257   ,0 };
            const unsigned long dim4402KuoInit[]        =   {   1   ,   1   ,   7   ,   3   ,   29  ,   39  ,   67  ,   195 ,   287 ,   587 ,   315 ,   1747    ,   7169    ,   11967   ,   4893    ,   797 ,0 };
            const unsigned long dim4403KuoInit[]        =   {   1   ,   3   ,   5   ,   5   ,   13  ,   49  ,   47  ,   97  ,   105 ,   203 ,   1237    ,   805 ,   3595    ,   9225    ,   23105   ,   52679   ,0 };
            const unsigned long dim4404KuoInit[]        =   {   1   ,   1   ,   5   ,   1   ,   23  ,   41  ,   39  ,   85  ,   7   ,   237 ,   1545    ,   2591    ,   1771    ,   2291    ,   20185   ,   46995   ,0 };
            const unsigned long dim4405KuoInit[]        =   {   1   ,   1   ,   1   ,   3   ,   15  ,   5   ,   43  ,   211 ,   109 ,   599 ,   77  ,   703 ,   5749    ,   15909   ,   12619   ,   41635   ,0 };
            const unsigned long dim4406KuoInit[]        =   {   1   ,   3   ,   1   ,   9   ,   19  ,   9   ,   49  ,   211 ,   149 ,   883 ,   1617    ,   497 ,   7689    ,   4253    ,   18737   ,   25293   ,0 };
            const unsigned long dim4407KuoInit[]        =   {   1   ,   1   ,   7   ,   11  ,   25  ,   3   ,   57  ,   83  ,   181 ,   801 ,   1515    ,   3821    ,   5287    ,   15243   ,   2347    ,   42333   ,0 };
            const unsigned long dim4408KuoInit[]        =   {   1   ,   3   ,   7   ,   15  ,   15  ,   25  ,   107 ,   79  ,   249 ,   721 ,   1237    ,   1587    ,   3469    ,   10319   ,   6421    ,   59103   ,0 };
            const unsigned long dim4409KuoInit[]        =   {   1   ,   1   ,   3   ,   3   ,   17  ,   21  ,   81  ,   35  ,   173 ,   101 ,   1879    ,   3301    ,   4409    ,   8563    ,   27981   ,   64269   ,0 };
            const unsigned long dim4410KuoInit[]        =   {   1   ,   1   ,   7   ,   3   ,   29  ,   17  ,   103 ,   17  ,   443 ,   821 ,   1133    ,   73  ,   2587    ,   16119   ,   17385   ,   2293    ,0 };
            const unsigned long dim4411KuoInit[]        =   {   1   ,   3   ,   7   ,   13  ,   9   ,   3   ,   41  ,   125 ,   149 ,   403 ,   537 ,   695 ,   3885    ,   15769   ,   11315   ,   53979   ,0 };
            const unsigned long dim4412KuoInit[]        =   {   1   ,   3   ,   5   ,   7   ,   11  ,   47  ,   57  ,   105 ,   433 ,   1009    ,   1749    ,   2453    ,   2083    ,   1543    ,   25941   ,   58037   ,0 };
            const unsigned long dim4413KuoInit[]        =   {   1   ,   1   ,   1   ,   3   ,   19  ,   63  ,   45  ,   251 ,   265 ,   591 ,   1235    ,   1349    ,   6069    ,   14139   ,   4299    ,   32719   ,0 };
            const unsigned long dim4414KuoInit[]        =   {   1   ,   3   ,   5   ,   9   ,   1   ,   49  ,   125 ,   217 ,   247 ,   915 ,   1289    ,   3589    ,   7   ,   5285    ,   32443   ,   56439   ,0 };
            const unsigned long dim4415KuoInit[]        =   {   1   ,   3   ,   5   ,   3   ,   31  ,   31  ,   67  ,   55  ,   203 ,   487 ,   983 ,   1841    ,   1587    ,   14881   ,   29187   ,   59293   ,0 };
            const unsigned long dim4416KuoInit[]        =   {   1   ,   1   ,   1   ,   13  ,   1   ,   59  ,   121 ,   57  ,   99  ,   857 ,   1085    ,   3505    ,   3615    ,   13725   ,   10363   ,   16841   ,0 };
            const unsigned long dim4417KuoInit[]        =   {   1   ,   1   ,   5   ,   3   ,   19  ,   55  ,   43  ,   95  ,   141 ,   217 ,   1087    ,   1129    ,   3417    ,   515 ,   23815   ,   13135   ,0 };
            const unsigned long dim4418KuoInit[]        =   {   1   ,   1   ,   1   ,   15  ,   9   ,   63  ,   5   ,   139 ,   327 ,   743 ,   1321    ,   4055    ,   4337    ,   15115   ,   8635    ,   53571   ,0 };
            const unsigned long dim4419KuoInit[]        =   {   1   ,   1   ,   5   ,   11  ,   31  ,   47  ,   105 ,   91  ,   493 ,   77  ,   1767    ,   1647    ,   4287    ,   5455    ,   1875    ,   6669    ,0 };
            const unsigned long dim4420KuoInit[]        =   {   1   ,   3   ,   1   ,   1   ,   25  ,   47  ,   19  ,   243 ,   131 ,   387 ,   69  ,   2057    ,   7151    ,   8497    ,   10975   ,   64383   ,0 };
            const unsigned long dim4421KuoInit[]        =   {   1   ,   3   ,   3   ,   1   ,   13  ,   19  ,   93  ,   35  ,   63  ,   881 ,   1787    ,   331 ,   57  ,   12875   ,   5689    ,   56411   ,0 };
            const unsigned long dim4422KuoInit[]        =   {   1   ,   3   ,   5   ,   1   ,   1   ,   33  ,   39  ,   193 ,   17  ,   291 ,   1507    ,   1719    ,   3507    ,   12683   ,   16467   ,   21453   ,0 };
            const unsigned long dim4423KuoInit[]        =   {   1   ,   1   ,   7   ,   7   ,   15  ,   41  ,   113 ,   23  ,   499 ,   641 ,   937 ,   2505    ,   4093    ,   4897    ,   2493    ,   28043   ,0 };
            const unsigned long dim4424KuoInit[]        =   {   1   ,   3   ,   1   ,   15  ,   13  ,   11  ,   59  ,   175 ,   299 ,   233 ,   189 ,   675 ,   2623    ,   15695   ,   2895    ,   29213   ,0 };
            const unsigned long dim4425KuoInit[]        =   {   1   ,   1   ,   3   ,   3   ,   25  ,   33  ,   113 ,   81  ,   485 ,   615 ,   1451    ,   1331    ,   607 ,   13451   ,   5853    ,   5051    ,0 };
            const unsigned long dim4426KuoInit[]        =   {   1   ,   1   ,   7   ,   15  ,   13  ,   55  ,   107 ,   43  ,   343 ,   383 ,   63  ,   2761    ,   5543    ,   10249   ,   2903    ,   28567   ,0 };
            const unsigned long dim4427KuoInit[]        =   {   1   ,   1   ,   7   ,   13  ,   7   ,   23  ,   13  ,   199 ,   109 ,   87  ,   389 ,   561 ,   2403    ,   12411   ,   18247   ,   40239   ,0 };
            const unsigned long dim4428KuoInit[]        =   {   1   ,   1   ,   1   ,   13  ,   21  ,   31  ,   3   ,   109 ,   97  ,   803 ,   991 ,   847 ,   6517    ,   3389    ,   15347   ,   18295   ,0 };
            const unsigned long dim4429KuoInit[]        =   {   1   ,   1   ,   1   ,   1   ,   27  ,   41  ,   115 ,   59  ,   189 ,   105 ,   63  ,   3169    ,   2293    ,   6051    ,   26017   ,   49947   ,0 };
            const unsigned long dim4430KuoInit[]        =   {   1   ,   3   ,   5   ,   13  ,   19  ,   27  ,   47  ,   13  ,   9   ,   135 ,   637 ,   2167    ,   2865    ,   13353   ,   1725    ,   41671   ,0 };
            const unsigned long dim4431KuoInit[]        =   {   1   ,   1   ,   5   ,   7   ,   17  ,   61  ,   37  ,   247 ,   447 ,   633 ,   1761    ,   2075    ,   6385    ,   12393   ,   31483   ,   1857    ,0 };
            const unsigned long dim4432KuoInit[]        =   {   1   ,   3   ,   5   ,   13  ,   17  ,   17  ,   103 ,   47  ,   391 ,   725 ,   1379    ,   3939    ,   6467    ,   6723    ,   4229    ,   50221   ,0 };
            const unsigned long dim4433KuoInit[]        =   {   1   ,   1   ,   5   ,   5   ,   1   ,   61  ,   13  ,   157 ,   379 ,   935 ,   105 ,   419 ,   147 ,   6369    ,   15413   ,   38553   ,0 };
            const unsigned long dim4434KuoInit[]        =   {   1   ,   1   ,   5   ,   1   ,   13  ,   49  ,   63  ,   165 ,   207 ,   741 ,   1285    ,   3247    ,   4053    ,   1647    ,   6571    ,   61075   ,0 };
            const unsigned long dim4435KuoInit[]        =   {   1   ,   1   ,   5   ,   15  ,   29  ,   41  ,   51  ,   5   ,   123 ,   45  ,   537 ,   2887    ,   1329    ,   9447    ,   22285   ,   45247   ,0 };
            const unsigned long dim4436KuoInit[]        =   {   1   ,   3   ,   1   ,   3   ,   29  ,   51  ,   65  ,   91  ,   153 ,   673 ,   1759    ,   401 ,   2969    ,   6631    ,   16965   ,   49027   ,0 };
            const unsigned long dim4437KuoInit[]        =   {   1   ,   3   ,   3   ,   15  ,   29  ,   61  ,   31  ,   85  ,   13  ,   427 ,   1833    ,   2145    ,   4229    ,   14543   ,   22741   ,   24913   ,0 };
            const unsigned long dim4438KuoInit[]        =   {   1   ,   3   ,   5   ,   3   ,   21  ,   9   ,   45  ,   253 ,   89  ,   397 ,   965 ,   685 ,   4013    ,   11731   ,   2947    ,   11995   ,0 };
            const unsigned long dim4439KuoInit[]        =   {   1   ,   3   ,   3   ,   11  ,   27  ,   45  ,   101 ,   151 ,   21  ,   563 ,   203 ,   445 ,   6863    ,   2151    ,   4999    ,   8839    ,0 };
            const unsigned long dim4440KuoInit[]        =   {   1   ,   3   ,   5   ,   9   ,   25  ,   57  ,   15  ,   29  ,   3   ,   127 ,   871 ,   1851    ,   7825    ,   1593    ,   11143   ,   24507   ,0 };
            const unsigned long dim4441KuoInit[]        =   {   1   ,   1   ,   1   ,   7   ,   17  ,   27  ,   49  ,   159 ,   425 ,   773 ,   1841    ,   1693    ,   7829    ,   2905    ,   16347   ,   5581    ,0 };
            const unsigned long dim4442KuoInit[]        =   {   1   ,   3   ,   5   ,   7   ,   13  ,   29  ,   81  ,   77  ,   349 ,   747 ,   119 ,   1195    ,   5   ,   8481    ,   1543    ,   50033   ,0 };
            const unsigned long dim4443KuoInit[]        =   {   1   ,   1   ,   7   ,   15  ,   1   ,   25  ,   33  ,   13  ,   9   ,   565 ,   555 ,   699 ,   4691    ,   1749    ,   9531    ,   30133   ,0 };
            const unsigned long dim4444KuoInit[]        =   {   1   ,   1   ,   5   ,   5   ,   13  ,   43  ,   17  ,   31  ,   199 ,   381 ,   393 ,   3511    ,   6331    ,   10377   ,   31467   ,   25305   ,0 };
            const unsigned long dim4445KuoInit[]        =   {   1   ,   3   ,   3   ,   13  ,   1   ,   49  ,   23  ,   241 ,   499 ,   791 ,   1851    ,   321 ,   6057    ,   2141    ,   8825    ,   1829    ,0 };
            const unsigned long dim4446KuoInit[]        =   {   1   ,   1   ,   1   ,   1   ,   13  ,   23  ,   77  ,   103 ,   123 ,   247 ,   1655    ,   1403    ,   6059    ,   14505   ,   1579    ,   48169   ,0 };
            const unsigned long dim4447KuoInit[]        =   {   1   ,   1   ,   7   ,   13  ,   9   ,   23  ,   125 ,   15  ,   479 ,   447 ,   327 ,   3743    ,   1199    ,   5015    ,   19009   ,   50929   ,0 };
            const unsigned long dim4448KuoInit[]        =   {   1   ,   1   ,   7   ,   5   ,   5   ,   55  ,   39  ,   131 ,   229 ,   607 ,   1591    ,   769 ,   2899    ,   9681    ,   11789   ,   62135   ,0 };
            const unsigned long dim4449KuoInit[]        =   {   1   ,   3   ,   3   ,   13  ,   17  ,   59  ,   65  ,   117 ,   299 ,   85  ,   1187    ,   3249    ,   295 ,   15603   ,   28909   ,   60317   ,0 };
            const unsigned long dim4450KuoInit[]        =   {   1   ,   1   ,   5   ,   13  ,   13  ,   31  ,   7   ,   5   ,   65  ,   785 ,   985 ,   1637    ,   3015    ,   14261   ,   15089   ,   37337   ,0 };
            const unsigned long dim4451KuoInit[]        =   {   1   ,   3   ,   3   ,   3   ,   17  ,   49  ,   123 ,   151 ,   41  ,   419 ,   391 ,   1265    ,   6225    ,   2037    ,   8409    ,   13437   ,0 };
            const unsigned long dim4452KuoInit[]        =   {   1   ,   1   ,   5   ,   13  ,   1   ,   23  ,   111 ,   21  ,   381 ,   343 ,   519 ,   1071    ,   3889    ,   14897   ,   9611    ,   10657   ,0 };
            const unsigned long dim4453KuoInit[]        =   {   1   ,   3   ,   5   ,   13  ,   25  ,   15  ,   79  ,   215 ,   99  ,   783 ,   107 ,   3845    ,   443 ,   1115    ,   24657   ,   54263   ,0 };
            const unsigned long dim4454KuoInit[]        =   {   1   ,   3   ,   3   ,   7   ,   7   ,   29  ,   15  ,   213 ,   165 ,   911 ,   351 ,   4025    ,   7657    ,   2079    ,   21525   ,   60751   ,0 };
            const unsigned long dim4455KuoInit[]        =   {   1   ,   1   ,   3   ,   3   ,   15  ,   19  ,   107 ,   181 ,   177 ,   279 ,   803 ,   2825    ,   4609    ,   9955    ,   7725    ,   44161   ,0 };
            const unsigned long dim4456KuoInit[]        =   {   1   ,   1   ,   3   ,   7   ,   21  ,   9   ,   69  ,   107 ,   43  ,   239 ,   397 ,   2567    ,   1375    ,   12483   ,   19783   ,   27889   ,0 };
            const unsigned long dim4457KuoInit[]        =   {   1   ,   3   ,   7   ,   5   ,   7   ,   39  ,   39  ,   71  ,   397 ,   829 ,   373 ,   1869    ,   6229    ,   2299    ,   19737   ,   58479   ,0 };
            const unsigned long dim4458KuoInit[]        =   {   1   ,   1   ,   3   ,   7   ,   13  ,   53  ,   39  ,   41  ,   105 ,   547 ,   1687    ,   153 ,   4067    ,   16317   ,   15  ,   52949   ,0 };
            const unsigned long dim4459KuoInit[]        =   {   1   ,   3   ,   7   ,   11  ,   9   ,   19  ,   97  ,   45  ,   225 ,   977 ,   545 ,   3691    ,   5315    ,   6239    ,   26833   ,   28115   ,0 };
            const unsigned long dim4460KuoInit[]        =   {   1   ,   3   ,   7   ,   3   ,   23  ,   11  ,   17  ,   13  ,   437 ,   145 ,   1153    ,   1193    ,   835 ,   5095    ,   5535    ,   17033   ,0 };
            const unsigned long dim4461KuoInit[]        =   {   1   ,   3   ,   1   ,   3   ,   7   ,   13  ,   17  ,   197 ,   289 ,   875 ,   97  ,   1721    ,   2369    ,   8351    ,   10813   ,   10129   ,0 };
            const unsigned long dim4462KuoInit[]        =   {   1   ,   3   ,   7   ,   13  ,   7   ,   59  ,   11  ,   37  ,   77  ,   541 ,   837 ,   2943    ,   423 ,   6093    ,   32495   ,   59203   ,0 };
            const unsigned long dim4463KuoInit[]        =   {   1   ,   1   ,   5   ,   9   ,   13  ,   17  ,   117 ,   239 ,   499 ,   819 ,   1585    ,   1755    ,   219 ,   4681    ,   10771   ,   14713   ,0 };
            const unsigned long dim4464KuoInit[]        =   {   1   ,   3   ,   1   ,   9   ,   15  ,   21  ,   3   ,   123 ,   303 ,   223 ,   1579    ,   1959    ,   7767    ,   6079    ,   2085    ,   7683    ,0 };
            const unsigned long dim4465KuoInit[]        =   {   1   ,   1   ,   7   ,   15  ,   13  ,   37  ,   35  ,   37  ,   479 ,   341 ,   1315    ,   3339    ,   3059    ,   251 ,   31629   ,   32757   ,0 };
            const unsigned long dim4466KuoInit[]        =   {   1   ,   1   ,   5   ,   11  ,   19  ,   23  ,   55  ,   109 ,   443 ,   77  ,   577 ,   1963    ,   3769    ,   1783    ,   5059    ,   35905   ,0 };
            const unsigned long dim4467KuoInit[]        =   {   1   ,   1   ,   5   ,   15  ,   17  ,   45  ,   53  ,   139 ,   261 ,   147 ,   1117    ,   2255    ,   1389    ,   6505    ,   21785   ,   52763   ,0 };
            const unsigned long dim4468KuoInit[]        =   {   1   ,   1   ,   7   ,   15  ,   21  ,   61  ,   97  ,   147 ,   397 ,   231 ,   775 ,   3297    ,   4883    ,   7059    ,   16769   ,   22765   ,0 };
            const unsigned long dim4469KuoInit[]        =   {   1   ,   3   ,   7   ,   13  ,   13  ,   29  ,   67  ,   117 ,   381 ,   505 ,   1437    ,   939 ,   4551    ,   8767    ,   7355    ,   10507   ,0 };
            const unsigned long dim4470KuoInit[]        =   {   1   ,   3   ,   3   ,   7   ,   5   ,   57  ,   115 ,   205 ,   159 ,   587 ,   169 ,   3787    ,   1271    ,   9583    ,   12205   ,   63993   ,0 };
            const unsigned long dim4471KuoInit[]        =   {   1   ,   3   ,   3   ,   13  ,   9   ,   59  ,   51  ,   165 ,   159 ,   391 ,   421 ,   2179    ,   2943    ,   505 ,   19315   ,   16175   ,0 };
            const unsigned long dim4472KuoInit[]        =   {   1   ,   3   ,   5   ,   13  ,   21  ,   53  ,   95  ,   131 ,   275 ,   421 ,   435 ,   2655    ,   4999    ,   14225   ,   26417   ,   57465   ,0 };
            const unsigned long dim4473KuoInit[]        =   {   1   ,   1   ,   1   ,   5   ,   29  ,   39  ,   115 ,   183 ,   507 ,   849 ,   1935    ,   2849    ,   6301    ,   11937   ,   19519   ,   37019   ,0 };
            const unsigned long dim4474KuoInit[]        =   {   1   ,   1   ,   5   ,   1   ,   13  ,   15  ,   37  ,   25  ,   113 ,   885 ,   1265    ,   2287    ,   4535    ,   12265   ,   10433   ,   56641   ,0 };
            const unsigned long dim4475KuoInit[]        =   {   1   ,   3   ,   1   ,   9   ,   29  ,   11  ,   67  ,   19  ,   11  ,   347 ,   1659    ,   3251    ,   3179    ,   2583    ,   10127   ,   21549   ,0 };
            const unsigned long dim4476KuoInit[]        =   {   1   ,   3   ,   7   ,   7   ,   13  ,   27  ,   101 ,   101 ,   197 ,   381 ,   987 ,   1699    ,   2913    ,   5179    ,   3423    ,   30389   ,0 };
            const unsigned long dim4477KuoInit[]        =   {   1   ,   1   ,   7   ,   13  ,   29  ,   39  ,   87  ,   65  ,   265 ,   891 ,   973 ,   1487    ,   6687    ,   7875    ,   13389   ,   39141   ,0 };
            const unsigned long dim4478KuoInit[]        =   {   1   ,   1   ,   1   ,   3   ,   7   ,   13  ,   101 ,   223 ,   181 ,   143 ,   1291    ,   2803    ,   251 ,   12541   ,   20713   ,   18137   ,0 };
            const unsigned long dim4479KuoInit[]        =   {   1   ,   1   ,   1   ,   15  ,   3   ,   9   ,   103 ,   137 ,   121 ,   563 ,   1079    ,   3119    ,   5619    ,   2175    ,   27223   ,   25289   ,0 };
            const unsigned long dim4480KuoInit[]        =   {   1   ,   3   ,   3   ,   15  ,   7   ,   13  ,   109 ,   201 ,   385 ,   347 ,   319 ,   883 ,   13  ,   3683    ,   19615   ,   9781    ,0 };
            const unsigned long dim4481KuoInit[]        =   {   1   ,   1   ,   1   ,   5   ,   13  ,   15  ,   127 ,   161 ,   353 ,   567 ,   555 ,   2287    ,   1791    ,   9535    ,   26227   ,   15335   ,0 };
            const unsigned long dim4482KuoInit[]        =   {   1   ,   1   ,   5   ,   9   ,   3   ,   63  ,   11  ,   83  ,   235 ,   741 ,   977 ,   2265    ,   449 ,   10293   ,   20451   ,   25143   ,0 };
            const unsigned long dim4483KuoInit[]        =   {   1   ,   1   ,   3   ,   3   ,   23  ,   25  ,   119 ,   1   ,   381 ,   463 ,   67  ,   257 ,   7087    ,   10553   ,   10781   ,   60003   ,0 };
            const unsigned long dim4484KuoInit[]        =   {   1   ,   3   ,   7   ,   15  ,   17  ,   11  ,   81  ,   17  ,   363 ,   163 ,   1765    ,   285 ,   4239    ,   14777   ,   15081   ,   54151   ,0 };
            const unsigned long dim4485KuoInit[]        =   {   1   ,   3   ,   3   ,   13  ,   17  ,   9   ,   73  ,   237 ,   273 ,   439 ,   1419    ,   1539    ,   667 ,   13833   ,   31783   ,   49459   ,0 };
            const unsigned long dim4486KuoInit[]        =   {   1   ,   1   ,   3   ,   3   ,   15  ,   11  ,   85  ,   3   ,   481 ,   169 ,   1443    ,   3403    ,   1157    ,   595 ,   6851    ,   59343   ,0 };
            const unsigned long dim4487KuoInit[]        =   {   1   ,   3   ,   7   ,   1   ,   23  ,   21  ,   13  ,   251 ,   403 ,   859 ,   175 ,   1373    ,   3711    ,   1233    ,   16879   ,   4659    ,0 };
            const unsigned long dim4488KuoInit[]        =   {   1   ,   1   ,   5   ,   7   ,   21  ,   7   ,   15  ,   165 ,   265 ,   117 ,   385 ,   2661    ,   1481    ,   1709    ,   19843   ,   29639   ,0 };
            const unsigned long dim4489KuoInit[]        =   {   1   ,   1   ,   1   ,   11  ,   3   ,   7   ,   81  ,   3   ,   429 ,   997 ,   1759    ,   285 ,   375 ,   3487    ,   5257    ,   14203   ,0 };
            const unsigned long dim4490KuoInit[]        =   {   1   ,   3   ,   7   ,   1   ,   15  ,   17  ,   93  ,   69  ,   303 ,   809 ,   929 ,   3025    ,   6333    ,   4525    ,   6441    ,   32265   ,0 };
            const unsigned long dim4491KuoInit[]        =   {   1   ,   3   ,   1   ,   5   ,   27  ,   41  ,   119 ,   25  ,   461 ,   347 ,   369 ,   489 ,   377 ,   3899    ,   12711   ,   24475   ,0 };
            const unsigned long dim4492KuoInit[]        =   {   1   ,   3   ,   7   ,   5   ,   29  ,   59  ,   101 ,   159 ,   305 ,   337 ,   1371    ,   891 ,   6013    ,   2987    ,   27201   ,   61055   ,0 };
            const unsigned long dim4493KuoInit[]        =   {   1   ,   1   ,   1   ,   3   ,   19  ,   5   ,   3   ,   219 ,   71  ,   213 ,   137 ,   1925    ,   2365    ,   583 ,   11591   ,   9495    ,0 };
            const unsigned long dim4494KuoInit[]        =   {   1   ,   3   ,   7   ,   3   ,   31  ,   31  ,   85  ,   115 ,   169 ,   499 ,   1669    ,   499 ,   1987    ,   2505    ,   435 ,   6377    ,0 };
            const unsigned long dim4495KuoInit[]        =   {   1   ,   1   ,   3   ,   5   ,   3   ,   33  ,   33  ,   41  ,   199 ,   187 ,   1023    ,   771 ,   3447    ,   4355    ,   29353   ,   42293   ,0 };
            const unsigned long dim4496KuoInit[]        =   {   1   ,   1   ,   5   ,   7   ,   1   ,   41  ,   17  ,   231 ,   305 ,   975 ,   975 ,   1697    ,   7267    ,   9567    ,   9975    ,   51313   ,0 };
            const unsigned long dim4497KuoInit[]        =   {   1   ,   3   ,   5   ,   3   ,   9   ,   41  ,   27  ,   29  ,   285 ,   469 ,   1845    ,   665 ,   3007    ,   4199    ,   3367    ,   50849   ,0 };
            const unsigned long dim4498KuoInit[]        =   {   1   ,   1   ,   5   ,   9   ,   29  ,   5   ,   93  ,   63  ,   227 ,   751 ,   1591    ,   649 ,   6219    ,   13371   ,   3473    ,   34033   ,0 };
            const unsigned long dim4499KuoInit[]        =   {   1   ,   3   ,   3   ,   1   ,   13  ,   5   ,   27  ,   239 ,   89  ,   893 ,   1869    ,   3789    ,   7569    ,   525 ,   32019   ,   17501   ,0 };
            const unsigned long dim4500KuoInit[]        =   {   1   ,   1   ,   5   ,   11  ,   9   ,   5   ,   51  ,   149 ,   351 ,   81  ,   1895    ,   3011    ,   2521    ,   4185    ,   1277    ,   65397   ,0 };
            const unsigned long dim4501KuoInit[]        =   {   1   ,   3   ,   5   ,   3   ,   21  ,   13  ,   121 ,   213 ,   67  ,   825 ,   1493    ,   2933    ,   6189    ,   14135   ,   7021    ,   47435   ,0 };
            const unsigned long dim4502KuoInit[]        =   {   1   ,   3   ,   7   ,   15  ,   11  ,   51  ,   63  ,   163 ,   371 ,   363 ,   1885    ,   1535    ,   629 ,   2939    ,   12547   ,   24779   ,0 };
            const unsigned long dim4503KuoInit[]        =   {   1   ,   3   ,   7   ,   3   ,   3   ,   3   ,   49  ,   47  ,   409 ,   963 ,   1499    ,   1107    ,   8063    ,   11137   ,   21251   ,   4695    ,0 };
            const unsigned long dim4504KuoInit[]        =   {   1   ,   1   ,   3   ,   1   ,   31  ,   17  ,   61  ,   149 ,   263 ,   503 ,   219 ,   3419    ,   4525    ,   14689   ,   30841   ,   60799   ,0 };
            const unsigned long dim4505KuoInit[]        =   {   1   ,   3   ,   3   ,   5   ,   31  ,   25  ,   5   ,   247 ,   219 ,   79  ,   1605    ,   3523    ,   4815    ,   12907   ,   22017   ,   16145   ,0 };
            const unsigned long dim4506KuoInit[]        =   {   1   ,   1   ,   7   ,   11  ,   21  ,   51  ,   47  ,   85  ,   37  ,   461 ,   1299    ,   2007    ,   2093    ,   10707   ,   17997   ,   4221    ,0 };
            const unsigned long dim4507KuoInit[]        =   {   1   ,   3   ,   3   ,   11  ,   11  ,   59  ,   63  ,   191 ,   359 ,   859 ,   1265    ,   853 ,   6733    ,   4605    ,   27559   ,   32557   ,0 };
            const unsigned long dim4508KuoInit[]        =   {   1   ,   3   ,   7   ,   11  ,   29  ,   61  ,   119 ,   73  ,   73  ,   259 ,   1939    ,   59  ,   5177    ,   12375   ,   22143   ,   28073   ,0 };
            const unsigned long dim4509KuoInit[]        =   {   1   ,   3   ,   7   ,   7   ,   21  ,   29  ,   19  ,   33  ,   59  ,   539 ,   281 ,   2989    ,   2239    ,   6671    ,   31629   ,   16065   ,0 };
            const unsigned long dim4510KuoInit[]        =   {   1   ,   3   ,   5   ,   7   ,   17  ,   27  ,   39  ,   57  ,   483 ,   463 ,   51  ,   3771    ,   5641    ,   5875    ,   2369    ,   21619   ,0 };
            const unsigned long dim4511KuoInit[]        =   {   1   ,   3   ,   1   ,   1   ,   15  ,   9   ,   121 ,   217 ,   107 ,   615 ,   1413    ,   2083    ,   231 ,   9177    ,   12275   ,   20069   ,0 };
            const unsigned long dim4512KuoInit[]        =   {   1   ,   3   ,   5   ,   15  ,   17  ,   55  ,   85  ,   27  ,   477 ,   687 ,   1295    ,   3471    ,   6951    ,   2645    ,   30051   ,   36149   ,0 };
            const unsigned long dim4513KuoInit[]        =   {   1   ,   3   ,   3   ,   1   ,   1   ,   27  ,   9   ,   183 ,   39  ,   559 ,   1639    ,   5   ,   1177    ,   16011   ,   1447    ,   42897   ,0 };
            const unsigned long dim4514KuoInit[]        =   {   1   ,   1   ,   1   ,   5   ,   21  ,   9   ,   75  ,   217 ,   311 ,   679 ,   1913    ,   2623    ,   3853    ,   10889   ,   22501   ,   40089   ,0 };
            const unsigned long dim4515KuoInit[]        =   {   1   ,   3   ,   5   ,   7   ,   17  ,   59  ,   23  ,   199 ,   147 ,   339 ,   1027    ,   1471    ,   6539    ,   7963    ,   30713   ,   1939    ,0 };
            const unsigned long dim4516KuoInit[]        =   {   1   ,   1   ,   3   ,   11  ,   23  ,   51  ,   111 ,   175 ,   433 ,   803 ,   1577    ,   4091    ,   2197    ,   10003   ,   2597    ,   53553   ,0 };
            const unsigned long dim4517KuoInit[]        =   {   1   ,   3   ,   5   ,   1   ,   3   ,   11  ,   125 ,   11  ,   281 ,   969 ,   1943    ,   1003    ,   2189    ,   3551    ,   25831   ,   64261   ,0 };
            const unsigned long dim4518KuoInit[]        =   {   1   ,   3   ,   7   ,   1   ,   17  ,   19  ,   111 ,   235 ,   511 ,   723 ,   1779    ,   1041    ,   4067    ,   3827    ,   17285   ,   63075   ,0 };
            const unsigned long dim4519KuoInit[]        =   {   1   ,   1   ,   3   ,   15  ,   3   ,   63  ,   7   ,   165 ,   153 ,   829 ,   221 ,   1485    ,   4319    ,   9865    ,   937 ,   65171   ,0 };
            const unsigned long dim4520KuoInit[]        =   {   1   ,   1   ,   7   ,   1   ,   21  ,   21  ,   89  ,   233 ,   417 ,   15  ,   367 ,   311 ,   3575    ,   11285   ,   5245    ,   3947    ,0 };
            const unsigned long dim4521KuoInit[]        =   {   1   ,   3   ,   7   ,   9   ,   27  ,   55  ,   3   ,   71  ,   227 ,   397 ,   561 ,   4075    ,   5495    ,   16071   ,   11495   ,   48733   ,0 };
            const unsigned long dim4522KuoInit[]        =   {   1   ,   3   ,   3   ,   3   ,   3   ,   39  ,   57  ,   219 ,   237 ,   151 ,   1305    ,   2485    ,   5045    ,   5065    ,   21041   ,   20055   ,0 };
            const unsigned long dim4523KuoInit[]        =   {   1   ,   1   ,   3   ,   11  ,   15  ,   33  ,   75  ,   241 ,   335 ,   207 ,   1695    ,   119 ,   839 ,   12449   ,   28467   ,   7445    ,0 };
            const unsigned long dim4524KuoInit[]        =   {   1   ,   1   ,   7   ,   3   ,   21  ,   3   ,   43  ,   137 ,   321 ,   763 ,   575 ,   2317    ,   8073    ,   1225    ,   8257    ,   11639   ,0 };
            const unsigned long dim4525KuoInit[]        =   {   1   ,   3   ,   5   ,   11  ,   27  ,   21  ,   17  ,   115 ,   265 ,   1023    ,   1403    ,   853 ,   757 ,   1799    ,   25505   ,   49207   ,0 };
            const unsigned long dim4526KuoInit[]        =   {   1   ,   1   ,   7   ,   5   ,   9   ,   45  ,   57  ,   77  ,   127 ,   447 ,   1097    ,   1729    ,   2361    ,   1599    ,   15727   ,   49413   ,0 };
            const unsigned long dim4527KuoInit[]        =   {   1   ,   1   ,   1   ,   15  ,   21  ,   41  ,   13  ,   41  ,   183 ,   151 ,   441 ,   3077    ,   4205    ,   125 ,   23767   ,   44667   ,0 };
            const unsigned long dim4528KuoInit[]        =   {   1   ,   3   ,   3   ,   9   ,   3   ,   37  ,   5   ,   203 ,   39  ,   49  ,   1793    ,   3523    ,   7541    ,   23  ,   22527   ,   35999   ,0 };
            const unsigned long dim4529KuoInit[]        =   {   1   ,   3   ,   7   ,   13  ,   13  ,   5   ,   127 ,   169 ,   465 ,   759 ,   393 ,   1543    ,   3327    ,   6223    ,   18675   ,   36075   ,0 };
            const unsigned long dim4530KuoInit[]        =   {   1   ,   3   ,   5   ,   15  ,   17  ,   57  ,   83  ,   237 ,   307 ,   495 ,   135 ,   1461    ,   4353    ,   15251   ,   10005   ,   12769   ,0 };
            const unsigned long dim4531KuoInit[]        =   {   1   ,   1   ,   7   ,   13  ,   27  ,   49  ,   57  ,   19  ,   463 ,   97  ,   217 ,   2825    ,   7729    ,   2375    ,   8339    ,   17563   ,0 };
            const unsigned long dim4532KuoInit[]        =   {   1   ,   3   ,   7   ,   1   ,   31  ,   13  ,   17  ,   245 ,   329 ,   1021    ,   619 ,   3149    ,   7089    ,   4381    ,   14079   ,   35659   ,0 };
            const unsigned long dim4533KuoInit[]        =   {   1   ,   3   ,   5   ,   15  ,   31  ,   19  ,   15  ,   217 ,   115 ,   159 ,   281 ,   2233    ,   4315    ,   7429    ,   25527   ,   44671   ,0 };
            const unsigned long dim4534KuoInit[]        =   {   1   ,   1   ,   7   ,   1   ,   11  ,   23  ,   7   ,   181 ,   43  ,   271 ,   607 ,   389 ,   357 ,   3003    ,   15899   ,   50849   ,0 };
            const unsigned long dim4535KuoInit[]        =   {   1   ,   1   ,   7   ,   15  ,   11  ,   15  ,   41  ,   247 ,   37  ,   163 ,   665 ,   3059    ,   5151    ,   15335   ,   13167   ,   18541   ,0 };
            const unsigned long dim4536KuoInit[]        =   {   1   ,   1   ,   7   ,   15  ,   25  ,   33  ,   33  ,   135 ,   425 ,   785 ,   1557    ,   793 ,   7153    ,   13473   ,   8833    ,   21335   ,0 };
            const unsigned long dim4537KuoInit[]        =   {   1   ,   3   ,   3   ,   1   ,   17  ,   27  ,   7   ,   75  ,   205 ,   71  ,   581 ,   733 ,   5009    ,   1029    ,   27537   ,   41905   ,0 };
            const unsigned long dim4538KuoInit[]        =   {   1   ,   3   ,   3   ,   3   ,   5   ,   63  ,   121 ,   3   ,   381 ,   607 ,   1099    ,   1571    ,   1077    ,   15491   ,   13785   ,   31537   ,0 };
            const unsigned long dim4539KuoInit[]        =   {   1   ,   1   ,   5   ,   7   ,   17  ,   15  ,   87  ,   81  ,   313 ,   683 ,   31  ,   3301    ,   1683    ,   12667   ,   17867   ,   38405   ,0 };
            const unsigned long dim4540KuoInit[]        =   {   1   ,   1   ,   1   ,   13  ,   3   ,   23  ,   51  ,   95  ,   409 ,   711 ,   381 ,   245 ,   7   ,   15679   ,   14281   ,   39789   ,0 };
            const unsigned long dim4541KuoInit[]        =   {   1   ,   3   ,   1   ,   1   ,   31  ,   35  ,   97  ,   35  ,   121 ,   45  ,   1397    ,   4009    ,   2829    ,   3273    ,   2725    ,   913 ,0 };
            const unsigned long dim4542KuoInit[]        =   {   1   ,   3   ,   1   ,   11  ,   29  ,   29  ,   111 ,   153 ,   499 ,   371 ,   1767    ,   3063    ,   6175    ,   8205    ,   3485    ,   8951    ,0 };
            const unsigned long dim4543KuoInit[]        =   {   1   ,   3   ,   5   ,   5   ,   27  ,   9   ,   113 ,   141 ,   295 ,   901 ,   1367    ,   791 ,   1645    ,   5993    ,   6139    ,   12853   ,0 };
            const unsigned long dim4544KuoInit[]        =   {   1   ,   1   ,   1   ,   7   ,   27  ,   17  ,   113 ,   187 ,   293 ,   517 ,   1327    ,   579 ,   1191    ,   12039   ,   13973   ,   51129   ,0 };
            const unsigned long dim4545KuoInit[]        =   {   1   ,   1   ,   7   ,   3   ,   11  ,   9   ,   55  ,   71  ,   47  ,   787 ,   1311    ,   2567    ,   7555    ,   2825    ,   11485   ,   16787   ,0 };
            const unsigned long dim4546KuoInit[]        =   {   1   ,   3   ,   1   ,   7   ,   29  ,   33  ,   115 ,   115 ,   69  ,   155 ,   1185    ,   2653    ,   7127    ,   1449    ,   8875    ,   34819   ,0 };
            const unsigned long dim4547KuoInit[]        =   {   1   ,   3   ,   5   ,   15  ,   9   ,   61  ,   61  ,   161 ,   29  ,   33  ,   1155    ,   613 ,   7233    ,   2811    ,   4837    ,   27781   ,0 };
            const unsigned long dim4548KuoInit[]        =   {   1   ,   1   ,   5   ,   3   ,   25  ,   57  ,   85  ,   121 ,   505 ,   45  ,   1143    ,   3237    ,   6403    ,   2377    ,   30791   ,   26637   ,0 };
            const unsigned long dim4549KuoInit[]        =   {   1   ,   3   ,   3   ,   5   ,   31  ,   39  ,   63  ,   255 ,   153 ,   417 ,   865 ,   441 ,   7649    ,   16321   ,   29147   ,   11089   ,0 };
            const unsigned long dim4550KuoInit[]        =   {   1   ,   1   ,   7   ,   5   ,   25  ,   19  ,   43  ,   45  ,   457 ,   353 ,   25  ,   1135    ,   4663    ,   11419   ,   25373   ,   8411    ,0 };
            const unsigned long dim4551KuoInit[]        =   {   1   ,   3   ,   7   ,   11  ,   11  ,   47  ,   101 ,   255 ,   437 ,   937 ,   263 ,   1171    ,   7893    ,   277 ,   4413    ,   20197   ,0 };
            const unsigned long dim4552KuoInit[]        =   {   1   ,   1   ,   5   ,   13  ,   21  ,   61  ,   1   ,   31  ,   227 ,   345 ,   1535    ,   1597    ,   7061    ,   11595   ,   12055   ,   29611   ,0 };
            const unsigned long dim4553KuoInit[]        =   {   1   ,   1   ,   3   ,   1   ,   19  ,   49  ,   83  ,   61  ,   305 ,   611 ,   399 ,   1397    ,   4753    ,   2309    ,   15097   ,   13439   ,0 };
            const unsigned long dim4554KuoInit[]        =   {   1   ,   1   ,   3   ,   9   ,   27  ,   37  ,   77  ,   165 ,   273 ,   85  ,   483 ,   299 ,   5073    ,   1255    ,   4073    ,   56333   ,0 };
            const unsigned long dim4555KuoInit[]        =   {   1   ,   1   ,   7   ,   3   ,   5   ,   11  ,   127 ,   157 ,   489 ,   653 ,   165 ,   2071    ,   4819    ,   121 ,   32755   ,   21623   ,0 };
            const unsigned long dim4556KuoInit[]        =   {   1   ,   3   ,   5   ,   11  ,   23  ,   5   ,   111 ,   25  ,   173 ,   661 ,   353 ,   1167    ,   2625    ,   8021    ,   18731   ,   42705   ,0 };
            const unsigned long dim4557KuoInit[]        =   {   1   ,   3   ,   5   ,   15  ,   9   ,   33  ,   31  ,   137 ,   283 ,   925 ,   1641    ,   115 ,   7047    ,   11365   ,   13097   ,   4711    ,0 };
            const unsigned long dim4558KuoInit[]        =   {   1   ,   3   ,   7   ,   13  ,   13  ,   13  ,   29  ,   135 ,   273 ,   325 ,   1085    ,   1683    ,   1365    ,   9931    ,   8557    ,   3855    ,0 };
            const unsigned long dim4559KuoInit[]        =   {   1   ,   3   ,   5   ,   13  ,   17  ,   27  ,   103 ,   183 ,   237 ,   343 ,   907 ,   2945    ,   3039    ,   12873   ,   28635   ,   14791   ,0 };
            const unsigned long dim4560KuoInit[]        =   {   1   ,   3   ,   3   ,   13  ,   5   ,   51  ,   13  ,   57  ,   131 ,   1021    ,   1995    ,   921 ,   6911    ,   14381   ,   17113   ,   56783   ,0 };
            const unsigned long dim4561KuoInit[]        =   {   1   ,   3   ,   5   ,   5   ,   5   ,   55  ,   89  ,   29  ,   409 ,   687 ,   1151    ,   3423    ,   1893    ,   12153   ,   8167    ,   14849   ,0 };
            const unsigned long dim4562KuoInit[]        =   {   1   ,   3   ,   5   ,   1   ,   1   ,   53  ,   41  ,   53  ,   81  ,   797 ,   1665    ,   85  ,   8131    ,   3853    ,   17815   ,   7771    ,0 };
            const unsigned long dim4563KuoInit[]        =   {   1   ,   1   ,   7   ,   3   ,   31  ,   41  ,   37  ,   107 ,   319 ,   421 ,   279 ,   2061    ,   1981    ,   6917    ,   25545   ,   46053   ,0 };
            const unsigned long dim4564KuoInit[]        =   {   1   ,   3   ,   3   ,   11  ,   13  ,   25  ,   85  ,   203 ,   399 ,   863 ,   1577    ,   1445    ,   8093    ,   1845    ,   16307   ,   61075   ,0 };
            const unsigned long dim4565KuoInit[]        =   {   1   ,   1   ,   5   ,   5   ,   1   ,   35  ,   117 ,   201 ,   271 ,   901 ,   1939    ,   2291    ,   343 ,   13591   ,   28371   ,   61429   ,0 };
            const unsigned long dim4566KuoInit[]        =   {   1   ,   3   ,   3   ,   7   ,   21  ,   55  ,   85  ,   123 ,   183 ,   787 ,   1121    ,   447 ,   5761    ,   6925    ,   22407   ,   53149   ,0 };
            const unsigned long dim4567KuoInit[]        =   {   1   ,   3   ,   1   ,   13  ,   31  ,   25  ,   63  ,   247 ,   109 ,   741 ,   231 ,   2621    ,   4939    ,   11155   ,   21625   ,   30433   ,0 };
            const unsigned long dim4568KuoInit[]        =   {   1   ,   1   ,   5   ,   15  ,   3   ,   49  ,   63  ,   147 ,   117 ,   703 ,   1317    ,   2803    ,   4655    ,   5981    ,   14473   ,   52739   ,0 };
            const unsigned long dim4569KuoInit[]        =   {   1   ,   1   ,   7   ,   1   ,   27  ,   45  ,   117 ,   55  ,   391 ,   275 ,   1793    ,   1279    ,   4799    ,   15509   ,   17173   ,   25103   ,0 };
            const unsigned long dim4570KuoInit[]        =   {   1   ,   3   ,   3   ,   1   ,   27  ,   15  ,   35  ,   43  ,   209 ,   869 ,   1833    ,   3625    ,   5973    ,   8965    ,   1443    ,   59053   ,0 };
            const unsigned long dim4571KuoInit[]        =   {   1   ,   1   ,   1   ,   1   ,   1   ,   19  ,   93  ,   63  ,   407 ,   735 ,   1821    ,   2875    ,   1263    ,   1461    ,   7811    ,   17257   ,0 };
            const unsigned long dim4572KuoInit[]        =   {   1   ,   1   ,   7   ,   5   ,   15  ,   35  ,   3   ,   89  ,   355 ,   983 ,   1183    ,   2709    ,   127 ,   10671   ,   31223   ,   26689   ,0 };
            const unsigned long dim4573KuoInit[]        =   {   1   ,   3   ,   3   ,   3   ,   15  ,   29  ,   105 ,   239 ,   293 ,   57  ,   515 ,   363 ,   5227    ,   14693   ,   18437   ,   13107   ,0 };
            const unsigned long dim4574KuoInit[]        =   {   1   ,   1   ,   5   ,   1   ,   29  ,   61  ,   65  ,   77  ,   23  ,   311 ,   1581    ,   215 ,   1409    ,   2575    ,   13915   ,   63179   ,0 };
            const unsigned long dim4575KuoInit[]        =   {   1   ,   3   ,   1   ,   5   ,   27  ,   17  ,   113 ,   255 ,   347 ,   493 ,   1795    ,   817 ,   7483    ,   10611   ,   9941    ,   56425   ,0 };
            const unsigned long dim4576KuoInit[]        =   {   1   ,   3   ,   1   ,   11  ,   1   ,   29  ,   79  ,   153 ,   183 ,   975 ,   1127    ,   881 ,   1319    ,   11481   ,   16807   ,   23167   ,0 };
            const unsigned long dim4577KuoInit[]        =   {   1   ,   1   ,   5   ,   15  ,   7   ,   31  ,   107 ,   245 ,   59  ,   579 ,   423 ,   2165    ,   5137    ,   14735   ,   13187   ,   28781   ,0 };
            const unsigned long dim4578KuoInit[]        =   {   1   ,   3   ,   1   ,   13  ,   13  ,   55  ,   63  ,   155 ,   63  ,   581 ,   1083    ,   2249    ,   7397    ,   2081    ,   23727   ,   23981   ,0 };
            const unsigned long dim4579KuoInit[]        =   {   1   ,   3   ,   7   ,   9   ,   23  ,   29  ,   93  ,   215 ,   361 ,   71  ,   119 ,   3215    ,   3529    ,   14249   ,   29593   ,   27305   ,0 };
            const unsigned long dim4580KuoInit[]        =   {   1   ,   1   ,   7   ,   9   ,   5   ,   63  ,   45  ,   181 ,   463 ,   21  ,   229 ,   4069    ,   4889    ,   2141    ,   25623   ,   6191    ,0 };
            const unsigned long dim4581KuoInit[]        =   {   1   ,   3   ,   1   ,   1   ,   17  ,   31  ,   105 ,   153 ,   301 ,   433 ,   517 ,   585 ,   285 ,   12805   ,   30173   ,   21363   ,0 };
            const unsigned long dim4582KuoInit[]        =   {   1   ,   3   ,   3   ,   15  ,   29  ,   23  ,   117 ,   19  ,   415 ,   389 ,   881 ,   2067    ,   583 ,   11383   ,   24951   ,   34745   ,0 };
            const unsigned long dim4583KuoInit[]        =   {   1   ,   1   ,   3   ,   13  ,   27  ,   55  ,   37  ,   5   ,   337 ,   943 ,   1851    ,   3227    ,   3881    ,   7365    ,   10853   ,   23325   ,0 };
            const unsigned long dim4584KuoInit[]        =   {   1   ,   3   ,   1   ,   11  ,   25  ,   5   ,   39  ,   189 ,   503 ,   745 ,   1349    ,   3881    ,   5037    ,   13855   ,   19537   ,   22329   ,0 };
            const unsigned long dim4585KuoInit[]        =   {   1   ,   3   ,   7   ,   9   ,   7   ,   33  ,   63  ,   61  ,   39  ,   825 ,   53  ,   1471    ,   3589    ,   13473   ,   29241   ,   40829   ,0 };
            const unsigned long dim4586KuoInit[]        =   {   1   ,   1   ,   7   ,   1   ,   1   ,   37  ,   111 ,   251 ,   427 ,   427 ,   937 ,   3921    ,   8033    ,   269 ,   14261   ,   37345   ,0 };
            const unsigned long dim4587KuoInit[]        =   {   1   ,   1   ,   7   ,   7   ,   31  ,   59  ,   111 ,   205 ,   285 ,   113 ,   951 ,   2867    ,   3769    ,   5237    ,   693 ,   42097   ,0 };
            const unsigned long dim4588KuoInit[]        =   {   1   ,   1   ,   1   ,   15  ,   29  ,   43  ,   15  ,   253 ,   383 ,   783 ,   1407    ,   3851    ,   4333    ,   11607   ,   12777   ,   55499   ,0 };
            const unsigned long dim4589KuoInit[]        =   {   1   ,   1   ,   3   ,   15  ,   15  ,   27  ,   75  ,   57  ,   87  ,   185 ,   1079    ,   2721    ,   915 ,   13633   ,   26439   ,   10697   ,0 };
            const unsigned long dim4590KuoInit[]        =   {   1   ,   1   ,   7   ,   13  ,   17  ,   63  ,   43  ,   65  ,   5   ,   699 ,   1157    ,   2271    ,   837 ,   687 ,   29551   ,   22109   ,0 };
            const unsigned long dim4591KuoInit[]        =   {   1   ,   3   ,   7   ,   3   ,   5   ,   47  ,   103 ,   143 ,   281 ,   81  ,   283 ,   3073    ,   3741    ,   11525   ,   4499    ,   41325   ,0 };
            const unsigned long dim4592KuoInit[]        =   {   1   ,   3   ,   7   ,   11  ,   13  ,   19  ,   63  ,   215 ,   215 ,   21  ,   1443    ,   2485    ,   2661    ,   8123    ,   11739   ,   48729   ,0 };
            const unsigned long dim4593KuoInit[]        =   {   1   ,   3   ,   3   ,   3   ,   27  ,   13  ,   111 ,   153 ,   17  ,   715 ,   1453    ,   1827    ,   5201    ,   1139    ,   19531   ,   18853   ,0 };
            const unsigned long dim4594KuoInit[]        =   {   1   ,   3   ,   7   ,   15  ,   7   ,   61  ,   109 ,   213 ,   383 ,   483 ,   301 ,   1791    ,   559 ,   14205   ,   28441   ,   13179   ,0 };
            const unsigned long dim4595KuoInit[]        =   {   1   ,   1   ,   7   ,   7   ,   23  ,   55  ,   119 ,   45  ,   467 ,   775 ,   951 ,   2533    ,   6703    ,   15791   ,   12891   ,   54513   ,0 };
            const unsigned long dim4596KuoInit[]        =   {   1   ,   1   ,   7   ,   15  ,   19  ,   49  ,   71  ,   127 ,   455 ,   881 ,   329 ,   663 ,   4997    ,   8755    ,   17517   ,   31507   ,0 };
            const unsigned long dim4597KuoInit[]        =   {   1   ,   1   ,   5   ,   9   ,   9   ,   13  ,   75  ,   161 ,   109 ,   631 ,   153 ,   15  ,   1181    ,   14299   ,   11815   ,   32017   ,0 };
            const unsigned long dim4598KuoInit[]        =   {   1   ,   3   ,   5   ,   15  ,   9   ,   63  ,   101 ,   85  ,   487 ,   213 ,   1743    ,   2575    ,   6979    ,   14421   ,   18051   ,   14505   ,0 };
            const unsigned long dim4599KuoInit[]        =   {   1   ,   1   ,   5   ,   7   ,   5   ,   5   ,   109 ,   245 ,   427 ,   765 ,   2013    ,   2929    ,   6737    ,   14561   ,   32227   ,   60605   ,0 };
            const unsigned long dim4600KuoInit[]        =   {   1   ,   3   ,   1   ,   3   ,   1   ,   7   ,   17  ,   169 ,   27  ,   925 ,   1711    ,   101 ,   2065    ,   4857    ,   863 ,   38371   ,0 };
            const unsigned long dim4601KuoInit[]        =   {   1   ,   1   ,   7   ,   5   ,   5   ,   25  ,   119 ,   145 ,   27  ,   377 ,   1535    ,   441 ,   5181    ,   9789    ,   26757   ,   46977   ,0 };
            const unsigned long dim4602KuoInit[]        =   {   1   ,   1   ,   7   ,   5   ,   31  ,   1   ,   79  ,   223 ,   199 ,   409 ,   2033    ,   1781    ,   5269    ,   2631    ,   10627   ,   41785   ,0 };
            const unsigned long dim4603KuoInit[]        =   {   1   ,   3   ,   5   ,   3   ,   29  ,   7   ,   91  ,   123 ,   407 ,   19  ,   847 ,   4019    ,   1029    ,   15763   ,   30265   ,   49287   ,0 };
            const unsigned long dim4604KuoInit[]        =   {   1   ,   1   ,   3   ,   9   ,   1   ,   37  ,   33  ,   53  ,   49  ,   949 ,   1737    ,   229 ,   3419    ,   5125    ,   16365   ,   48867   ,0 };
            const unsigned long dim4605KuoInit[]        =   {   1   ,   3   ,   7   ,   9   ,   25  ,   17  ,   13  ,   31  ,   27  ,   53  ,   287 ,   4021    ,   7687    ,   14511   ,   12627   ,   54409   ,0 };
            const unsigned long dim4606KuoInit[]        =   {   1   ,   1   ,   3   ,   9   ,   21  ,   29  ,   1   ,   193 ,   479 ,   895 ,   605 ,   2013    ,   6571    ,   14073   ,   28185   ,   58845   ,0 };
            const unsigned long dim4607KuoInit[]        =   {   1   ,   3   ,   5   ,   11  ,   3   ,   21  ,   111 ,   215 ,   259 ,   941 ,   1729    ,   645 ,   695 ,   3295    ,   23825   ,   1009    ,0 };
            const unsigned long dim4608KuoInit[]        =   {   1   ,   3   ,   3   ,   9   ,   25  ,   17  ,   103 ,   187 ,   153 ,   287 ,   911 ,   1265    ,   41  ,   1687    ,   22443   ,   21117   ,0 };
            const unsigned long dim4609KuoInit[]        =   {   1   ,   3   ,   5   ,   5   ,   31  ,   25  ,   3   ,   107 ,   271 ,   531 ,   1413    ,   2813    ,   5075    ,   2127    ,   18647   ,   5257    ,0 };
            const unsigned long dim4610KuoInit[]        =   {   1   ,   1   ,   1   ,   7   ,   5   ,   25  ,   71  ,   225 ,   391 ,   877 ,   1713    ,   1841    ,   3239    ,   9063    ,   11445   ,   48693   ,0 };
            const unsigned long dim4611KuoInit[]        =   {   1   ,   1   ,   1   ,   13  ,   21  ,   25  ,   15  ,   157 ,   59  ,   37  ,   1349    ,   413 ,   6401    ,   3533    ,   6019    ,   40373   ,0 };
            const unsigned long dim4612KuoInit[]        =   {   1   ,   1   ,   5   ,   5   ,   5   ,   3   ,   31  ,   185 ,   59  ,   933 ,   1415    ,   901 ,   6075    ,   4839    ,   32477   ,   43085   ,0 };
            const unsigned long dim4613KuoInit[]        =   {   1   ,   3   ,   1   ,   5   ,   11  ,   7   ,   81  ,   109 ,   167 ,   277 ,   1803    ,   39  ,   7421    ,   15573   ,   2233    ,   8193    ,0 };
            const unsigned long dim4614KuoInit[]        =   {   1   ,   3   ,   1   ,   1   ,   13  ,   15  ,   47  ,   109 ,   279 ,   515 ,   1393    ,   1785    ,   4297    ,   813 ,   19177   ,   5951    ,0 };
            const unsigned long dim4615KuoInit[]        =   {   1   ,   3   ,   3   ,   1   ,   21  ,   23  ,   3   ,   27  ,   317 ,   677 ,   1775    ,   1109    ,   93  ,   9007    ,   8767    ,   42181   ,0 };
            const unsigned long dim4616KuoInit[]        =   {   1   ,   3   ,   5   ,   13  ,   23  ,   15  ,   49  ,   9   ,   369 ,   293 ,   529 ,   2115    ,   4781    ,   1955    ,   21897   ,   22871   ,0 };
            const unsigned long dim4617KuoInit[]        =   {   1   ,   1   ,   7   ,   9   ,   5   ,   39  ,   85  ,   85  ,   405 ,   837 ,   1891    ,   27  ,   509 ,   14643   ,   10497   ,   34021   ,0 };
            const unsigned long dim4618KuoInit[]        =   {   1   ,   1   ,   7   ,   9   ,   7   ,   47  ,   49  ,   87  ,   481 ,   1003    ,   741 ,   329 ,   2951    ,   5317    ,   21601   ,   20485   ,0 };
            const unsigned long dim4619KuoInit[]        =   {   1   ,   3   ,   5   ,   11  ,   27  ,   61  ,   45  ,   251 ,   489 ,   623 ,   1673    ,   891 ,   5429    ,   4457    ,   23147   ,   5201    ,0 };
            const unsigned long dim4620KuoInit[]        =   {   1   ,   3   ,   3   ,   7   ,   25  ,   27  ,   3   ,   233 ,   193 ,   253 ,   483 ,   1881    ,   6753    ,   7843    ,   3975    ,   23413   ,0 };
            const unsigned long dim4621KuoInit[]        =   {   1   ,   3   ,   1   ,   15  ,   27  ,   31  ,   51  ,   213 ,   463 ,   995 ,   1127    ,   3589    ,   1409    ,   15743   ,   22755   ,   39913   ,0 };
            const unsigned long dim4622KuoInit[]        =   {   1   ,   3   ,   1   ,   13  ,   21  ,   17  ,   35  ,   157 ,   233 ,   899 ,   1003    ,   201 ,   3425    ,   6703    ,   26585   ,   18389   ,0 };
            const unsigned long dim4623KuoInit[]        =   {   1   ,   3   ,   1   ,   9   ,   11  ,   23  ,   69  ,   93  ,   361 ,   177 ,   1071    ,   3523    ,   4117    ,   6107    ,   3701    ,   64213   ,0 };
            const unsigned long dim4624KuoInit[]        =   {   1   ,   3   ,   5   ,   1   ,   19  ,   1   ,   35  ,   95  ,   385 ,   853 ,   1635    ,   1815    ,   5307    ,   3155    ,   11347   ,   39289   ,0 };
            const unsigned long dim4625KuoInit[]        =   {   1   ,   1   ,   7   ,   9   ,   3   ,   41  ,   37  ,   161 ,   293 ,   1005    ,   1915    ,   2867    ,   841 ,   195 ,   4247    ,   8621    ,0 };
            const unsigned long dim4626KuoInit[]        =   {   1   ,   3   ,   7   ,   11  ,   19  ,   43  ,   69  ,   95  ,   455 ,   669 ,   217 ,   129 ,   1275    ,   8431    ,   22519   ,   39261   ,0 };
            const unsigned long dim4627KuoInit[]        =   {   1   ,   3   ,   7   ,   13  ,   31  ,   49  ,   105 ,   61  ,   437 ,   923 ,   1579    ,   1695    ,   7129    ,   10555   ,   3177    ,   6667    ,0 };
            const unsigned long dim4628KuoInit[]        =   {   1   ,   1   ,   5   ,   7   ,   3   ,   47  ,   29  ,   159 ,   207 ,   969 ,   433 ,   3541    ,   689 ,   1255    ,   1359    ,   38349   ,0 };
            const unsigned long dim4629KuoInit[]        =   {   1   ,   1   ,   3   ,   9   ,   9   ,   31  ,   23  ,   37  ,   435 ,   729 ,   121 ,   2483    ,   8009    ,   14229   ,   6961    ,   40681   ,0 };
            const unsigned long dim4630KuoInit[]        =   {   1   ,   3   ,   1   ,   7   ,   19  ,   33  ,   111 ,   131 ,   155 ,   641 ,   795 ,   3753    ,   4381    ,   11135   ,   8381    ,   62553   ,0 };
            const unsigned long dim4631KuoInit[]        =   {   1   ,   3   ,   7   ,   13  ,   29  ,   37  ,   15  ,   245 ,   341 ,   1023    ,   1123    ,   3317    ,   2409    ,   8957    ,   4607    ,   53947   ,0 };
            const unsigned long dim4632KuoInit[]        =   {   1   ,   1   ,   7   ,   15  ,   7   ,   19  ,   105 ,   187 ,   407 ,   339 ,   607 ,   11  ,   1679    ,   13701   ,   2385    ,   6203    ,0 };
            const unsigned long dim4633KuoInit[]        =   {   1   ,   3   ,   3   ,   9   ,   7   ,   63  ,   7   ,   173 ,   475 ,   87  ,   807 ,   1467    ,   791 ,   1949    ,   20839   ,   54179   ,0 };
            const unsigned long dim4634KuoInit[]        =   {   1   ,   3   ,   3   ,   3   ,   25  ,   25  ,   111 ,   155 ,   249 ,   55  ,   1455    ,   2503    ,   6869    ,   14463   ,   10519   ,   56761   ,0 };
            const unsigned long dim4635KuoInit[]        =   {   1   ,   3   ,   1   ,   9   ,   9   ,   31  ,   89  ,   235 ,   499 ,   61  ,   747 ,   487 ,   6767    ,   4777    ,   12513   ,   43677   ,0 };
            const unsigned long dim4636KuoInit[]        =   {   1   ,   3   ,   3   ,   5   ,   25  ,   47  ,   13  ,   3   ,   313 ,   541 ,   1323    ,   3551    ,   7971    ,   1453    ,   1531    ,   38915   ,0 };
            const unsigned long dim4637KuoInit[]        =   {   1   ,   1   ,   1   ,   5   ,   7   ,   63  ,   125 ,   243 ,   199 ,   875 ,   791 ,   103 ,   3997    ,   16199   ,   20615   ,   18359   ,0 };
            const unsigned long dim4638KuoInit[]        =   {   1   ,   1   ,   7   ,   1   ,   21  ,   19  ,   13  ,   9   ,   117 ,   625 ,   1333    ,   561 ,   889 ,   10961   ,   11313   ,   48139   ,0 };
            const unsigned long dim4639KuoInit[]        =   {   1   ,   3   ,   5   ,   9   ,   19  ,   57  ,   121 ,   163 ,   333 ,   681 ,   1065    ,   885 ,   5961    ,   6385    ,   20135   ,   19733   ,0 };
            const unsigned long dim4640KuoInit[]        =   {   1   ,   1   ,   7   ,   13  ,   27  ,   11  ,   113 ,   187 ,   91  ,   83  ,   591 ,   517 ,   7379    ,   12823   ,   31815   ,   14721   ,0 };
            const unsigned long dim4641KuoInit[]        =   {   1   ,   3   ,   5   ,   7   ,   27  ,   21  ,   47  ,   203 ,   411 ,   191 ,   881 ,   2597    ,   6107    ,   135 ,   14525   ,   4059    ,0 };
            const unsigned long dim4642KuoInit[]        =   {   1   ,   3   ,   5   ,   7   ,   29  ,   3   ,   61  ,   231 ,   259 ,   425 ,   1953    ,   803 ,   7921    ,   2321    ,   28919   ,   41349   ,0 };
            const unsigned long dim4643KuoInit[]        =   {   1   ,   3   ,   1   ,   1   ,   27  ,   7   ,   87  ,   197 ,   287 ,   345 ,   457 ,   1813    ,   2701    ,   2095    ,   32705   ,   52579   ,0 };
            const unsigned long dim4644KuoInit[]        =   {   1   ,   3   ,   3   ,   9   ,   13  ,   59  ,   87  ,   105 ,   49  ,   387 ,   73  ,   2983    ,   6453    ,   10977   ,   8851    ,   61031   ,0 };
            const unsigned long dim4645KuoInit[]        =   {   1   ,   3   ,   5   ,   7   ,   1   ,   53  ,   121 ,   129 ,   307 ,   251 ,   1405    ,   3093    ,   4487    ,   1953    ,   2681    ,   4391    ,0 };
            const unsigned long dim4646KuoInit[]        =   {   1   ,   1   ,   3   ,   1   ,   1   ,   29  ,   119 ,   125 ,   95  ,   425 ,   385 ,   1577    ,   3387    ,   14709   ,   25143   ,   4317    ,0 };
            const unsigned long dim4647KuoInit[]        =   {   1   ,   3   ,   1   ,   3   ,   27  ,   59  ,   53  ,   203 ,   75  ,   731 ,   1019    ,   3179    ,   4533    ,   3217    ,   30801   ,   53439   ,0 };
            const unsigned long dim4648KuoInit[]        =   {   1   ,   1   ,   3   ,   11  ,   29  ,   53  ,   93  ,   251 ,   441 ,   75  ,   363 ,   2405    ,   6263    ,   12121   ,   11963   ,   13219   ,0 };
            const unsigned long dim4649KuoInit[]        =   {   1   ,   3   ,   1   ,   13  ,   3   ,   7   ,   57  ,   169 ,   387 ,   543 ,   1027    ,   1169    ,   5089    ,   5077    ,   17933   ,   8131    ,0 };
            const unsigned long dim4650KuoInit[]        =   {   1   ,   1   ,   3   ,   5   ,   17  ,   17  ,   123 ,   161 ,   21  ,   641 ,   795 ,   2415    ,   2743    ,   1489    ,   25137   ,   19393   ,0 };
            const unsigned long dim4651KuoInit[]        =   {   1   ,   3   ,   3   ,   1   ,   1   ,   47  ,   37  ,   95  ,   315 ,   37  ,   1481    ,   2189    ,   687 ,   2859    ,   5339    ,   26907   ,0 };
            const unsigned long dim4652KuoInit[]        =   {   1   ,   3   ,   7   ,   15  ,   23  ,   51  ,   71  ,   203 ,   237 ,   777 ,   417 ,   2767    ,   5243    ,   1453    ,   17269   ,   27679   ,0 };
            const unsigned long dim4653KuoInit[]        =   {   1   ,   3   ,   1   ,   15  ,   17  ,   41  ,   29  ,   45  ,   363 ,   77  ,   1071    ,   3419    ,   6453    ,   13897   ,   25275   ,   26379   ,0 };
            const unsigned long dim4654KuoInit[]        =   {   1   ,   3   ,   7   ,   11  ,   21  ,   5   ,   89  ,   133 ,   339 ,   971 ,   1581    ,   377 ,   6563    ,   13073   ,   24521   ,   41107   ,0 };
            const unsigned long dim4655KuoInit[]        =   {   1   ,   1   ,   7   ,   15  ,   1   ,   27  ,   57  ,   249 ,   511 ,   833 ,   269 ,   2771    ,   4127    ,   12567   ,   27359   ,   22105   ,0 };
            const unsigned long dim4656KuoInit[]        =   {   1   ,   1   ,   3   ,   5   ,   31  ,   47  ,   63  ,   221 ,   25  ,   869 ,   1947    ,   3075    ,   6857    ,   13507   ,   16601   ,   29621   ,0 };
            const unsigned long dim4657KuoInit[]        =   {   1   ,   1   ,   3   ,   7   ,   17  ,   23  ,   109 ,   169 ,   211 ,   407 ,   1147    ,   2453    ,   2281    ,   10763   ,   11539   ,   49127   ,0 };
            const unsigned long dim4658KuoInit[]        =   {   1   ,   1   ,   3   ,   11  ,   27  ,   25  ,   111 ,   75  ,   465 ,   67  ,   1641    ,   3623    ,   7957    ,   14307   ,   17469   ,   52229   ,0 };
            const unsigned long dim4659KuoInit[]        =   {   1   ,   3   ,   1   ,   5   ,   23  ,   31  ,   49  ,   131 ,   351 ,   311 ,   1937    ,   2181    ,   7485    ,   10883   ,   20843   ,   33447   ,0 };
            const unsigned long dim4660KuoInit[]        =   {   1   ,   3   ,   1   ,   3   ,   5   ,   47  ,   81  ,   47  ,   305 ,   567 ,   219 ,   2495    ,   1207    ,   14359   ,   4683    ,   53717   ,0 };
            const unsigned long dim4661KuoInit[]        =   {   1   ,   1   ,   3   ,   3   ,   29  ,   43  ,   107 ,   81  ,   227 ,   795 ,   1261    ,   3887    ,   6177    ,   6219    ,   2557    ,   53203   ,0 };
            const unsigned long dim4662KuoInit[]        =   {   1   ,   3   ,   1   ,   9   ,   25  ,   47  ,   87  ,   177 ,   511 ,   899 ,   1967    ,   601 ,   3659    ,   3961    ,   27301   ,   9497    ,0 };
            const unsigned long dim4663KuoInit[]        =   {   1   ,   1   ,   5   ,   11  ,   7   ,   19  ,   69  ,   95  ,   319 ,   811 ,   1539    ,   931 ,   6609    ,   4325    ,   4639    ,   60125   ,0 };
            const unsigned long dim4664KuoInit[]        =   {   1   ,   3   ,   3   ,   13  ,   31  ,   25  ,   65  ,   137 ,   67  ,   903 ,   1519    ,   3   ,   81  ,   9991    ,   24985   ,   12571   ,0 };
            const unsigned long dim4665KuoInit[]        =   {   1   ,   3   ,   1   ,   15  ,   13  ,   45  ,   87  ,   95  ,   185 ,   531 ,   591 ,   3649    ,   2289    ,   6353    ,   4305    ,   12801   ,0 };
            const unsigned long dim4666KuoInit[]        =   {   1   ,   3   ,   7   ,   11  ,   13  ,   49  ,   79  ,   205 ,   9   ,   457 ,   151 ,   3363    ,   1865    ,   12337   ,   3601    ,   23781   ,0 };
            const unsigned long dim4667KuoInit[]        =   {   1   ,   3   ,   3   ,   7   ,   21  ,   7   ,   117 ,   177 ,   327 ,   547 ,   385 ,   301 ,   5595    ,   14389   ,   8243    ,   3133    ,0 };
            const unsigned long dim4668KuoInit[]        =   {   1   ,   3   ,   1   ,   3   ,   17  ,   39  ,   61  ,   45  ,   489 ,   769 ,   1893    ,   1633    ,   2645    ,   10849   ,   3827    ,   41365   ,0 };
            const unsigned long dim4669KuoInit[]        =   {   1   ,   1   ,   1   ,   11  ,   11  ,   43  ,   93  ,   81  ,   497 ,   731 ,   1049    ,   91  ,   7979    ,   4707    ,   8757    ,   54837   ,0 };
            const unsigned long dim4670KuoInit[]        =   {   1   ,   1   ,   3   ,   7   ,   1   ,   63  ,   127 ,   173 ,   137 ,   291 ,   473 ,   3095    ,   6291    ,   14415   ,   20763   ,   45687   ,0 };
            const unsigned long dim4671KuoInit[]        =   {   1   ,   1   ,   5   ,   13  ,   29  ,   59  ,   73  ,   47  ,   171 ,   1001    ,   537 ,   2009    ,   5057    ,   15013   ,   22195   ,   56067   ,0 };
            const unsigned long dim4672KuoInit[]        =   {   1   ,   1   ,   3   ,   5   ,   21  ,   35  ,   121 ,   169 ,   377 ,   389 ,   1209    ,   415 ,   2641    ,   10423   ,   11327   ,   64157   ,0 };
            const unsigned long dim4673KuoInit[]        =   {   1   ,   1   ,   7   ,   1   ,   23  ,   43  ,   59  ,   205 ,   39  ,   455 ,   633 ,   1965    ,   7955    ,   305 ,   12245   ,   45229   ,0 };
            const unsigned long dim4674KuoInit[]        =   {   1   ,   3   ,   1   ,   15  ,   15  ,   27  ,   49  ,   77  ,   407 ,   549 ,   1159    ,   1701    ,   7331    ,   6421    ,   1799    ,   43539   ,0 };
            const unsigned long dim4675KuoInit[]        =   {   1   ,   3   ,   5   ,   3   ,   5   ,   33  ,   41  ,   127 ,   141 ,   149 ,   1957    ,   2823    ,   1303    ,   7339    ,   25253   ,   32329   ,0 };
            const unsigned long dim4676KuoInit[]        =   {   1   ,   1   ,   5   ,   15  ,   5   ,   37  ,   119 ,   111 ,   329 ,   781 ,   983 ,   2021    ,   1191    ,   1579    ,   4191    ,   24065   ,0 };
            const unsigned long dim4677KuoInit[]        =   {   1   ,   3   ,   7   ,   9   ,   7   ,   11  ,   25  ,   111 ,   301 ,   819 ,   1073    ,   2311    ,   2357    ,   6385    ,   17015   ,   8051    ,0 };
            const unsigned long dim4678KuoInit[]        =   {   1   ,   1   ,   1   ,   9   ,   29  ,   63  ,   91  ,   101 ,   71  ,   679 ,   267 ,   2909    ,   5359    ,   2305    ,   2523    ,   32423   ,0 };
            const unsigned long dim4679KuoInit[]        =   {   1   ,   3   ,   1   ,   5   ,   11  ,   45  ,   107 ,   95  ,   487 ,   605 ,   1647    ,   1977    ,   5305    ,   8855    ,   9057    ,   7809    ,0 };
            const unsigned long dim4680KuoInit[]        =   {   1   ,   1   ,   1   ,   1   ,   5   ,   45  ,   51  ,   51  ,   97  ,   775 ,   1559    ,   935 ,   7421    ,   2823    ,   11415   ,   49193   ,0 };
            const unsigned long dim4681KuoInit[]        =   {   1   ,   3   ,   3   ,   5   ,   23  ,   63  ,   99  ,   189 ,   175 ,   77  ,   675 ,   4021    ,   899 ,   11315   ,   18627   ,   46175   ,0 };
            const unsigned long dim4682KuoInit[]        =   {   1   ,   1   ,   3   ,   13  ,   9   ,   31  ,   67  ,   235 ,   409 ,   541 ,   2013    ,   2715    ,   2121    ,   3695    ,   11953   ,   5745    ,0 };
            const unsigned long dim4683KuoInit[]        =   {   1   ,   3   ,   5   ,   1   ,   13  ,   3   ,   15  ,   111 ,   455 ,   849 ,   231 ,   2711    ,   6831    ,   7813    ,   17173   ,   26019   ,0 };
            const unsigned long dim4684KuoInit[]        =   {   1   ,   1   ,   1   ,   15  ,   19  ,   3   ,   21  ,   135 ,   275 ,   891 ,   61  ,   5   ,   2901    ,   15727   ,   18015   ,   55741   ,0 };
            const unsigned long dim4685KuoInit[]        =   {   1   ,   3   ,   1   ,   5   ,   5   ,   31  ,   5   ,   147 ,   353 ,   199 ,   229 ,   3821    ,   537 ,   11667   ,   649 ,   33923   ,0 };
            const unsigned long dim4686KuoInit[]        =   {   1   ,   1   ,   7   ,   13  ,   3   ,   9   ,   75  ,   217 ,   423 ,   487 ,   723 ,   4091    ,   7075    ,   819 ,   2755    ,   13593   ,0 };
            const unsigned long dim4687KuoInit[]        =   {   1   ,   1   ,   1   ,   9   ,   25  ,   45  ,   69  ,   41  ,   183 ,   547 ,   69  ,   3591    ,   1331    ,   5701    ,   25891   ,   26553   ,0 };
            const unsigned long dim4688KuoInit[]        =   {   1   ,   3   ,   3   ,   13  ,   29  ,   51  ,   79  ,   153 ,   145 ,   923 ,   1345    ,   3113    ,   93  ,   10617   ,   12865   ,   27341   ,0 };
            const unsigned long dim4689KuoInit[]        =   {   1   ,   3   ,   3   ,   13  ,   13  ,   61  ,   67  ,   115 ,   295 ,   767 ,   1077    ,   3959    ,   7741    ,   16145   ,   3571    ,   38941   ,0 };
            const unsigned long dim4690KuoInit[]        =   {   1   ,   1   ,   3   ,   3   ,   1   ,   11  ,   25  ,   33  ,   481 ,   901 ,   135 ,   2093    ,   3531    ,   12137   ,   4521    ,   42301   ,0 };
            const unsigned long dim4691KuoInit[]        =   {   1   ,   1   ,   3   ,   1   ,   9   ,   59  ,   105 ,   71  ,   57  ,   131 ,   1491    ,   3259    ,   4125    ,   10279   ,   7537    ,   38827   ,0 };
            const unsigned long dim4692KuoInit[]        =   {   1   ,   3   ,   7   ,   13  ,   7   ,   53  ,   83  ,   43  ,   243 ,   423 ,   43  ,   2021    ,   4269    ,   1793    ,   9969    ,   55429   ,0 };
            const unsigned long dim4693KuoInit[]        =   {   1   ,   1   ,   3   ,   11  ,   17  ,   57  ,   73  ,   55  ,   369 ,   975 ,   603 ,   1893    ,   3299    ,   12177   ,   8965    ,   17159   ,0 };
            const unsigned long dim4694KuoInit[]        =   {   1   ,   3   ,   5   ,   11  ,   23  ,   11  ,   61  ,   11  ,   119 ,   905 ,   1841    ,   2597    ,   691 ,   6237    ,   2285    ,   33503   ,0 };
            const unsigned long dim4695KuoInit[]        =   {   1   ,   3   ,   5   ,   3   ,   21  ,   17  ,   35  ,   169 ,   511 ,   1009    ,   1287    ,   3967    ,   1543    ,   6983    ,   6895    ,   24959   ,0 };
            const unsigned long dim4696KuoInit[]        =   {   1   ,   1   ,   1   ,   13  ,   23  ,   9   ,   21  ,   47  ,   119 ,   323 ,   1255    ,   3601    ,   3157    ,   14041   ,   32065   ,   4759    ,0 };
            const unsigned long dim4697KuoInit[]        =   {   1   ,   1   ,   7   ,   15  ,   9   ,   1   ,   103 ,   43  ,   485 ,   281 ,   1651    ,   1957    ,   4215    ,   2105    ,   24455   ,   51001   ,0 };
            const unsigned long dim4698KuoInit[]        =   {   1   ,   1   ,   5   ,   7   ,   11  ,   11  ,   59  ,   31  ,   351 ,   275 ,   573 ,   2481    ,   2545    ,   15767   ,   4319    ,   63375   ,0 };
            const unsigned long dim4699KuoInit[]        =   {   1   ,   3   ,   1   ,   1   ,   1   ,   41  ,   41  ,   13  ,   33  ,   693 ,   1035    ,   905 ,   6129    ,   11131   ,   23571   ,   28789   ,0 };
            const unsigned long dim4700KuoInit[]        =   {   1   ,   1   ,   7   ,   11  ,   9   ,   39  ,   119 ,   31  ,   109 ,   837 ,   1905    ,   763 ,   3737    ,   16031   ,   8751    ,   17307   ,0 };
            const unsigned long dim4701KuoInit[]        =   {   1   ,   3   ,   3   ,   7   ,   17  ,   49  ,   77  ,   199 ,   33  ,   185 ,   1489    ,   445 ,   5495    ,   13693   ,   21453   ,   21039   ,0 };
            const unsigned long dim4702KuoInit[]        =   {   1   ,   3   ,   5   ,   15  ,   5   ,   9   ,   113 ,   179 ,   315 ,   85  ,   1559    ,   1505    ,   593 ,   259 ,   10393   ,   39025   ,0 };
            const unsigned long dim4703KuoInit[]        =   {   1   ,   1   ,   1   ,   5   ,   31  ,   33  ,   41  ,   45  ,   313 ,   513 ,   861 ,   2697    ,   4767    ,   10737   ,   25019   ,   12715   ,0 };
            const unsigned long dim4704KuoInit[]        =   {   1   ,   3   ,   3   ,   1   ,   19  ,   21  ,   19  ,   19  ,   127 ,   1009    ,   859 ,   829 ,   6887    ,   5407    ,   28491   ,   61079   ,0 };
            const unsigned long dim4705KuoInit[]        =   {   1   ,   1   ,   3   ,   15  ,   15  ,   35  ,   69  ,   255 ,   367 ,   889 ,   49  ,   3237    ,   3269    ,   1559    ,   26337   ,   16967   ,0 };
            const unsigned long dim4706KuoInit[]        =   {   1   ,   1   ,   5   ,   3   ,   17  ,   5   ,   125 ,   171 ,   505 ,   981 ,   773 ,   2973    ,   2905    ,   1537    ,   17729   ,   52289   ,0 };
            const unsigned long dim4707KuoInit[]        =   {   1   ,   1   ,   3   ,   3   ,   29  ,   5   ,   57  ,   247 ,   149 ,   707 ,   673 ,   429 ,   7583    ,   12135   ,   5405    ,   313 ,0 };
            const unsigned long dim4708KuoInit[]        =   {   1   ,   1   ,   3   ,   13  ,   31  ,   35  ,   117 ,   69  ,   413 ,   723 ,   645 ,   3989    ,   6461    ,   77  ,   29713   ,   32381   ,0 };
            const unsigned long dim4709KuoInit[]        =   {   1   ,   3   ,   3   ,   13  ,   11  ,   59  ,   5   ,   53  ,   507 ,   355 ,   335 ,   4027    ,   3111    ,   8385    ,   22369   ,   42907   ,0 };
            const unsigned long dim4710KuoInit[]        =   {   1   ,   1   ,   5   ,   9   ,   1   ,   47  ,   87  ,   13  ,   397 ,   693 ,   1745    ,   2907    ,   3919    ,   14843   ,   869 ,   24161   ,0 };
            const unsigned long dim4711KuoInit[]        =   {   1   ,   3   ,   7   ,   13  ,   11  ,   25  ,   25  ,   153 ,   501 ,   305 ,   941 ,   2379    ,   2587    ,   8225    ,   27951   ,   14809   ,0 };
            const unsigned long dim4712KuoInit[]        =   {   1   ,   1   ,   7   ,   3   ,   31  ,   7   ,   9   ,   47  ,   129 ,   611 ,   1367    ,   1575    ,   4421    ,   2277    ,   8113    ,   34229   ,0 };
            const unsigned long dim4713KuoInit[]        =   {   1   ,   1   ,   5   ,   11  ,   7   ,   41  ,   65  ,   5   ,   289 ,   799 ,   1143    ,   291 ,   243 ,   12957   ,   2729    ,   14589   ,0 };
            const unsigned long dim4714KuoInit[]        =   {   1   ,   3   ,   3   ,   7   ,   15  ,   29  ,   49  ,   243 ,   191 ,   565 ,   271 ,   2485    ,   1507    ,   13479   ,   3541    ,   10039   ,0 };
            const unsigned long dim4715KuoInit[]        =   {   1   ,   3   ,   5   ,   3   ,   25  ,   29  ,   113 ,   157 ,   357 ,   891 ,   199 ,   3113    ,   7133    ,   11047   ,   27517   ,   26795   ,0 };
            const unsigned long dim4716KuoInit[]        =   {   1   ,   1   ,   3   ,   15  ,   27  ,   23  ,   5   ,   207 ,   183 ,   85  ,   247 ,   15  ,   5559    ,   8005    ,   32485   ,   26403   ,0 };
            const unsigned long dim4717KuoInit[]        =   {   1   ,   1   ,   3   ,   9   ,   13  ,   9   ,   87  ,   9   ,   237 ,   875 ,   1621    ,   2955    ,   4105    ,   9645    ,   9613    ,   31235   ,0 };
            const unsigned long dim4718KuoInit[]        =   {   1   ,   1   ,   7   ,   11  ,   17  ,   3   ,   77  ,   241 ,   75  ,   433 ,   1481    ,   1901    ,   6249    ,   12411   ,   18975   ,   18911   ,0 };
            const unsigned long dim4719KuoInit[]        =   {   1   ,   3   ,   5   ,   15  ,   3   ,   43  ,   99  ,   197 ,   407 ,   161 ,   1103    ,   587 ,   6695    ,   255 ,   15571   ,   7523    ,0 };
            const unsigned long dim4720KuoInit[]        =   {   1   ,   1   ,   1   ,   13  ,   1   ,   21  ,   9   ,   155 ,   447 ,   785 ,   1573    ,   2497    ,   1515    ,   6031    ,   23509   ,   38199   ,0 };
            const unsigned long dim4721KuoInit[]        =   {   1   ,   1   ,   5   ,   15  ,   23  ,   43  ,   125 ,   187 ,   57  ,   259 ,   941 ,   1213    ,   2303    ,   1127    ,   31819   ,   53539   ,0 };
            const unsigned long dim4722KuoInit[]        =   {   1   ,   3   ,   3   ,   9   ,   23  ,   19  ,   29  ,   83  ,   431 ,   79  ,   781 ,   1887    ,   4857    ,   2079    ,   11521   ,   64925   ,0 };
            const unsigned long dim4723KuoInit[]        =   {   1   ,   3   ,   5   ,   5   ,   23  ,   11  ,   33  ,   107 ,   127 ,   173 ,   223 ,   491 ,   1529    ,   14655   ,   18857   ,   25247   ,0 };
            const unsigned long dim4724KuoInit[]        =   {   1   ,   3   ,   3   ,   9   ,   17  ,   53  ,   119 ,   159 ,   503 ,   153 ,   697 ,   657 ,   6207    ,   7659    ,   23899   ,   10347   ,0 };
            const unsigned long dim4725KuoInit[]        =   {   1   ,   3   ,   1   ,   15  ,   13  ,   37  ,   117 ,   219 ,   341 ,   983 ,   193 ,   1641    ,   7811    ,   11961   ,   12553   ,   9467    ,0 };
            const unsigned long dim4726KuoInit[]        =   {   1   ,   1   ,   7   ,   5   ,   27  ,   11  ,   35  ,   45  ,   89  ,   535 ,   965 ,   1269    ,   1243    ,   3987    ,   4629    ,   45335   ,0 };
            const unsigned long dim4727KuoInit[]        =   {   1   ,   1   ,   7   ,   1   ,   17  ,   25  ,   79  ,   133 ,   287 ,   7   ,   269 ,   697 ,   2537    ,   241 ,   15747   ,   49429   ,0 };
            const unsigned long dim4728KuoInit[]        =   {   1   ,   3   ,   3   ,   3   ,   27  ,   63  ,   23  ,   229 ,   3   ,   625 ,   453 ,   3037    ,   5747    ,   13525   ,   28835   ,   13411   ,0 };
            const unsigned long dim4729KuoInit[]        =   {   1   ,   3   ,   5   ,   5   ,   25  ,   57  ,   101 ,   13  ,   81  ,   917 ,   1849    ,   3931    ,   3319    ,   1353    ,   8253    ,   28897   ,0 };
            const unsigned long dim4730KuoInit[]        =   {   1   ,   1   ,   3   ,   15  ,   3   ,   41  ,   89  ,   165 ,   1   ,   935 ,   1859    ,   2881    ,   4351    ,   6737    ,   12361   ,   27821   ,0 };
            const unsigned long dim4731KuoInit[]        =   {   1   ,   3   ,   5   ,   3   ,   9   ,   45  ,   89  ,   251 ,   77  ,   899 ,   777 ,   4063    ,   8111    ,   5953    ,   26645   ,   62567   ,0 };
            const unsigned long dim4732KuoInit[]        =   {   1   ,   3   ,   5   ,   1   ,   25  ,   7   ,   15  ,   189 ,   157 ,   751 ,   627 ,   629 ,   2875    ,   11143   ,   23841   ,   58751   ,0 };
            const unsigned long dim4733KuoInit[]        =   {   1   ,   1   ,   1   ,   9   ,   7   ,   43  ,   95  ,   171 ,   311 ,   809 ,   971 ,   1317    ,   3521    ,   15227   ,   18155   ,   11409   ,0 };
            const unsigned long dim4734KuoInit[]        =   {   1   ,   3   ,   3   ,   7   ,   29  ,   7   ,   69  ,   1   ,   235 ,   699 ,   971 ,   3149    ,   71  ,   3609    ,   859 ,   17059   ,0 };
            const unsigned long dim4735KuoInit[]        =   {   1   ,   3   ,   1   ,   1   ,   3   ,   43  ,   3   ,   181 ,   465 ,   843 ,   1391    ,   3313    ,   8115    ,   1835    ,   32365   ,   31303   ,0 };
            const unsigned long dim4736KuoInit[]        =   {   1   ,   1   ,   5   ,   1   ,   27  ,   49  ,   15  ,   37  ,   257 ,   365 ,   1577    ,   3037    ,   2767    ,   8183    ,   173 ,   43089   ,0 };
            const unsigned long dim4737KuoInit[]        =   {   1   ,   1   ,   1   ,   7   ,   7   ,   29  ,   81  ,   33  ,   299 ,   827 ,   221 ,   2049    ,   1163    ,   13023   ,   6755    ,   53987   ,0 };
            const unsigned long dim4738KuoInit[]        =   {   1   ,   3   ,   1   ,   11  ,   19  ,   17  ,   63  ,   19  ,   469 ,   925 ,   37  ,   3291    ,   2699    ,   10083   ,   32197   ,   53749   ,0 };
            const unsigned long dim4739KuoInit[]        =   {   1   ,   3   ,   5   ,   15  ,   1   ,   23  ,   71  ,   173 ,   441 ,   109 ,   391 ,   3611    ,   7515    ,   5095    ,   26379   ,   55999   ,0 };
            const unsigned long dim4740KuoInit[]        =   {   1   ,   3   ,   3   ,   13  ,   29  ,   13  ,   23  ,   57  ,   329 ,   285 ,   1595    ,   2681    ,   5427    ,   3251    ,   7025    ,   3921    ,0 };
            const unsigned long dim4741KuoInit[]        =   {   1   ,   1   ,   1   ,   13  ,   7   ,   45  ,   121 ,   19  ,   19  ,   317 ,   1619    ,   2245    ,   7151    ,   5797    ,   31369   ,   7765    ,0 };
            const unsigned long dim4742KuoInit[]        =   {   1   ,   3   ,   7   ,   13  ,   9   ,   17  ,   11  ,   127 ,   337 ,   271 ,   2005    ,   787 ,   1017    ,   3075    ,   29453   ,   62275   ,0 };
            const unsigned long dim4743KuoInit[]        =   {   1   ,   3   ,   3   ,   13  ,   9   ,   51  ,   77  ,   31  ,   395 ,   175 ,   247 ,   1021    ,   4343    ,   3829    ,   959 ,   12069   ,0 };
            const unsigned long dim4744KuoInit[]        =   {   1   ,   3   ,   1   ,   9   ,   21  ,   33  ,   3   ,   129 ,   327 ,   13  ,   1943    ,   3305    ,   6173    ,   6903    ,   32539   ,   9525    ,0 };
            const unsigned long dim4745KuoInit[]        =   {   1   ,   1   ,   3   ,   15  ,   23  ,   5   ,   49  ,   125 ,   445 ,   849 ,   547 ,   2911    ,   7611    ,   8137    ,   28701   ,   6883    ,0 };
            const unsigned long dim4746KuoInit[]        =   {   1   ,   1   ,   3   ,   13  ,   11  ,   15  ,   21  ,   35  ,   303 ,   401 ,   187 ,   1981    ,   7919    ,   6179    ,   15643   ,   10327   ,0 };
            const unsigned long dim4747KuoInit[]        =   {   1   ,   1   ,   5   ,   7   ,   9   ,   45  ,   95  ,   205 ,   137 ,   425 ,   353 ,   673 ,   1337    ,   4357    ,   12521   ,   55229   ,0 };
            const unsigned long dim4748KuoInit[]        =   {   1   ,   3   ,   5   ,   5   ,   29  ,   47  ,   93  ,   213 ,   357 ,   821 ,   467 ,   395 ,   5079    ,   10341   ,   24117   ,   59757   ,0 };
            const unsigned long dim4749KuoInit[]        =   {   1   ,   3   ,   7   ,   5   ,   29  ,   39  ,   25  ,   85  ,   437 ,   873 ,   795 ,   91  ,   1979    ,   15503   ,   12157   ,   6803    ,0 };
            const unsigned long dim4750KuoInit[]        =   {   1   ,   3   ,   7   ,   1   ,   31  ,   21  ,   31  ,   245 ,   79  ,   661 ,   323 ,   1843    ,   2225    ,   15383   ,   17507   ,   30667   ,0 };
            const unsigned long dim4751KuoInit[]        =   {   1   ,   3   ,   5   ,   11  ,   1   ,   11  ,   119 ,   151 ,   489 ,   303 ,   835 ,   319 ,   2071    ,   1497    ,   10343   ,   40641   ,0 };
            const unsigned long dim4752KuoInit[]        =   {   1   ,   1   ,   1   ,   13  ,   7   ,   7   ,   125 ,   237 ,   103 ,   65  ,   1917    ,   2203    ,   5895    ,   8651    ,   7331    ,   27591   ,0 };
            const unsigned long dim4753KuoInit[]        =   {   1   ,   3   ,   3   ,   15  ,   21  ,   17  ,   73  ,   21  ,   485 ,   985 ,   253 ,   1877    ,   7423    ,   6569    ,   25971   ,   57821   ,0 };
            const unsigned long dim4754KuoInit[]        =   {   1   ,   3   ,   5   ,   13  ,   3   ,   59  ,   9   ,   77  ,   99  ,   851 ,   343 ,   2655    ,   367 ,   7253    ,   16357   ,   61563   ,0 };
            const unsigned long dim4755KuoInit[]        =   {   1   ,   3   ,   7   ,   13  ,   11  ,   1   ,   15  ,   227 ,   223 ,   317 ,   1111    ,   2287    ,   4035    ,   6803    ,   16787   ,   27797   ,0 };
            const unsigned long dim4756KuoInit[]        =   {   1   ,   3   ,   3   ,   5   ,   17  ,   49  ,   103 ,   191 ,   245 ,   473 ,   2007    ,   3297    ,   6487    ,   7371    ,   30497   ,   40255   ,0 };
            const unsigned long dim4757KuoInit[]        =   {   1   ,   3   ,   5   ,   13  ,   1   ,   61  ,   39  ,   1   ,   501 ,   665 ,   1085    ,   3933    ,   855 ,   883 ,   26949   ,   48761   ,0 };
            const unsigned long dim4758KuoInit[]        =   {   1   ,   3   ,   3   ,   13  ,   29  ,   39  ,   5   ,   43  ,   39  ,   529 ,   1605    ,   2235    ,   5081    ,   7467    ,   20727   ,   48611   ,0 };
            const unsigned long dim4759KuoInit[]        =   {   1   ,   1   ,   3   ,   11  ,   19  ,   61  ,   89  ,   15  ,   149 ,   639 ,   337 ,   2953    ,   2229    ,   11161   ,   21981   ,   26003   ,0 };
            const unsigned long dim4760KuoInit[]        =   {   1   ,   1   ,   3   ,   5   ,   29  ,   47  ,   71  ,   65  ,   107 ,   485 ,   57  ,   2255    ,   4383    ,   12383   ,   18161   ,   40233   ,0 };
            const unsigned long dim4761KuoInit[]        =   {   1   ,   1   ,   7   ,   3   ,   21  ,   41  ,   75  ,   133 ,   141 ,   343 ,   633 ,   2043    ,   2441    ,   5479    ,   2493    ,   45707   ,0 };
            const unsigned long dim4762KuoInit[]        =   {   1   ,   1   ,   7   ,   15  ,   21  ,   37  ,   19  ,   73  ,   361 ,   195 ,   1007    ,   1265    ,   4075    ,   1383    ,   11809   ,   64265   ,0 };
            const unsigned long dim4763KuoInit[]        =   {   1   ,   1   ,   5   ,   11  ,   19  ,   63  ,   21  ,   163 ,   349 ,   961 ,   1221    ,   2265    ,   4801    ,   14049   ,   8655    ,   32823   ,0 };
            const unsigned long dim4764KuoInit[]        =   {   1   ,   1   ,   5   ,   1   ,   9   ,   1   ,   101 ,   37  ,   473 ,   293 ,   1191    ,   791 ,   5327    ,   13739   ,   21761   ,   5755    ,0 };
            const unsigned long dim4765KuoInit[]        =   {   1   ,   1   ,   1   ,   15  ,   23  ,   57  ,   63  ,   195 ,   451 ,   627 ,   1725    ,   2405    ,   6943    ,   6855    ,   31563   ,   44699   ,0 };
            const unsigned long dim4766KuoInit[]        =   {   1   ,   3   ,   7   ,   1   ,   31  ,   41  ,   51  ,   177 ,   95  ,   961 ,   1407    ,   2847    ,   7   ,   8975    ,   4017    ,   46229   ,0 };
            const unsigned long dim4767KuoInit[]        =   {   1   ,   1   ,   3   ,   13  ,   19  ,   9   ,   67  ,   183 ,   255 ,   1   ,   1769    ,   2585    ,   2157    ,   9747    ,   32047   ,   13063   ,0 };
            const unsigned long dim4768KuoInit[]        =   {   1   ,   3   ,   3   ,   1   ,   21  ,   39  ,   31  ,   117 ,   397 ,   441 ,   1443    ,   3849    ,   5993    ,   13981   ,   23579   ,   40637   ,0 };
            const unsigned long dim4769KuoInit[]        =   {   1   ,   1   ,   7   ,   1   ,   31  ,   23  ,   61  ,   203 ,   111 ,   273 ,   1693    ,   3031    ,   4125    ,   1527    ,   30525   ,   17465   ,0 };
            const unsigned long dim4770KuoInit[]        =   {   1   ,   3   ,   5   ,   9   ,   19  ,   39  ,   127 ,   51  ,   419 ,   205 ,   425 ,   1939    ,   3703    ,   12009   ,   6291    ,   48639   ,0 };
            const unsigned long dim4771KuoInit[]        =   {   1   ,   3   ,   5   ,   13  ,   15  ,   27  ,   63  ,   239 ,   483 ,   181 ,   1621    ,   3789    ,   1929    ,   11115   ,   29385   ,   53075   ,0 };
            const unsigned long dim4772KuoInit[]        =   {   1   ,   1   ,   1   ,   9   ,   13  ,   9   ,   61  ,   15  ,   221 ,   163 ,   1645    ,   419 ,   639 ,   15081   ,   2097    ,   42133   ,0 };
            const unsigned long dim4773KuoInit[]        =   {   1   ,   3   ,   1   ,   7   ,   7   ,   31  ,   27  ,   141 ,   215 ,   645 ,   1293    ,   2237    ,   7129    ,   12619   ,   14563   ,   46081   ,0 };
            const unsigned long dim4774KuoInit[]        =   {   1   ,   3   ,   1   ,   13  ,   3   ,   41  ,   85  ,   37  ,   295 ,   815 ,   1417    ,   3083    ,   5873    ,   2037    ,   29909   ,   3817    ,0 };
            const unsigned long dim4775KuoInit[]        =   {   1   ,   3   ,   3   ,   9   ,   11  ,   49  ,   77  ,   103 ,   189 ,   319 ,   439 ,   3547    ,   173 ,   1105    ,   12125   ,   49989   ,0 };
            const unsigned long dim4776KuoInit[]        =   {   1   ,   3   ,   1   ,   15  ,   21  ,   9   ,   5   ,   37  ,   87  ,   905 ,   145 ,   1151    ,   2213    ,   4157    ,   16171   ,   56121   ,0 };
            const unsigned long dim4777KuoInit[]        =   {   1   ,   3   ,   5   ,   3   ,   15  ,   21  ,   91  ,   103 ,   367 ,   423 ,   101 ,   327 ,   3819    ,   3141    ,   3515    ,   11433   ,0 };
            const unsigned long dim4778KuoInit[]        =   {   1   ,   3   ,   1   ,   13  ,   13  ,   11  ,   61  ,   175 ,   265 ,   1011    ,   1825    ,   937 ,   983 ,   9575    ,   6053    ,   19633   ,0 };
            const unsigned long dim4779KuoInit[]        =   {   1   ,   3   ,   3   ,   9   ,   25  ,   47  ,   55  ,   233 ,   397 ,   325 ,   1217    ,   1691    ,   2751    ,   5339    ,   14269   ,   28791   ,0 };
            const unsigned long dim4780KuoInit[]        =   {   1   ,   1   ,   7   ,   15  ,   15  ,   51  ,   29  ,   215 ,   47  ,   713 ,   1879    ,   1981    ,   3087    ,   9945    ,   8959    ,   59367   ,0 };
            const unsigned long dim4781KuoInit[]        =   {   1   ,   1   ,   1   ,   9   ,   1   ,   33  ,   97  ,   7   ,   247 ,   893 ,   655 ,   3137    ,   6593    ,   7261    ,   11829   ,   22101   ,0 };
            const unsigned long dim4782KuoInit[]        =   {   1   ,   1   ,   1   ,   7   ,   11  ,   23  ,   75  ,   55  ,   313 ,   533 ,   529 ,   1929    ,   3647    ,   3317    ,   2987    ,   11917   ,0 };
            const unsigned long dim4783KuoInit[]        =   {   1   ,   1   ,   1   ,   11  ,   17  ,   35  ,   105 ,   221 ,   211 ,   545 ,   1595    ,   1949    ,   1895    ,   39  ,   7163    ,   33303   ,0 };
            const unsigned long dim4784KuoInit[]        =   {   1   ,   1   ,   1   ,   9   ,   21  ,   29  ,   73  ,   235 ,   189 ,   11  ,   1747    ,   1023    ,   3987    ,   10337   ,   1689    ,   64043   ,0 };
            const unsigned long dim4785KuoInit[]        =   {   1   ,   3   ,   1   ,   5   ,   11  ,   17  ,   115 ,   57  ,   465 ,   899 ,   1421    ,   1329    ,   6099    ,   8081    ,   11297   ,   13217   ,0 };
            const unsigned long dim4786KuoInit[]        =   {   1   ,   3   ,   3   ,   9   ,   5   ,   23  ,   53  ,   23  ,   11  ,   177 ,   817 ,   3003    ,   5073    ,   12071   ,   11263   ,   43159   ,0 };
            const unsigned long dim4787KuoInit[]        =   {   1   ,   1   ,   3   ,   13  ,   1   ,   63  ,   127 ,   41  ,   303 ,   681 ,   1559    ,   2385    ,   6265    ,   2417    ,   7063    ,   57965   ,0 };
            const unsigned long dim4788KuoInit[]        =   {   1   ,   3   ,   1   ,   1   ,   3   ,   21  ,   21  ,   37  ,   485 ,   771 ,   2043    ,   2065    ,   7047    ,   8867    ,   7019    ,   31437   ,0 };
            const unsigned long dim4789KuoInit[]        =   {   1   ,   3   ,   1   ,   11  ,   27  ,   49  ,   77  ,   89  ,   391 ,   623 ,   1675    ,   1919    ,   3575    ,   2187    ,   22801   ,   11059   ,0 };
            const unsigned long dim4790KuoInit[]        =   {   1   ,   1   ,   3   ,   5   ,   9   ,   57  ,   51  ,   255 ,   287 ,   851 ,   1999    ,   811 ,   7769    ,   11185   ,   19279   ,   65491   ,0 };
            const unsigned long dim4791KuoInit[]        =   {   1   ,   3   ,   1   ,   15  ,   31  ,   43  ,   113 ,   11  ,   461 ,   95  ,   841 ,   2937    ,   3969    ,   5901    ,   32681   ,   9231    ,0 };
            const unsigned long dim4792KuoInit[]        =   {   1   ,   3   ,   3   ,   5   ,   31  ,   47  ,   21  ,   187 ,   171 ,   249 ,   787 ,   459 ,   6897    ,   12131   ,   14843   ,   11805   ,0 };
            const unsigned long dim4793KuoInit[]        =   {   1   ,   1   ,   7   ,   13  ,   27  ,   61  ,   91  ,   77  ,   9   ,   393 ,   1231    ,   1901    ,   291 ,   1487    ,   22139   ,   30569   ,0 };
            const unsigned long dim4794KuoInit[]        =   {   1   ,   1   ,   1   ,   1   ,   27  ,   33  ,   55  ,   151 ,   89  ,   603 ,   551 ,   3767    ,   695 ,   15225   ,   9665    ,   60459   ,0 };
            const unsigned long dim4795KuoInit[]        =   {   1   ,   1   ,   7   ,   13  ,   31  ,   19  ,   83  ,   27  ,   381 ,   105 ,   481 ,   3095    ,   137 ,   10433   ,   12967   ,   19555   ,0 };
            const unsigned long dim4796KuoInit[]        =   {   1   ,   1   ,   1   ,   11  ,   15  ,   1   ,   3   ,   121 ,   107 ,   435 ,   1571    ,   3205    ,   5213    ,   3689    ,   31365   ,   14353   ,0 };
            const unsigned long dim4797KuoInit[]        =   {   1   ,   3   ,   7   ,   11  ,   23  ,   61  ,   93  ,   59  ,   179 ,   361 ,   965 ,   413 ,   7757    ,   15779   ,   1119    ,   34673   ,0 };
            const unsigned long dim4798KuoInit[]        =   {   1   ,   1   ,   3   ,   13  ,   29  ,   63  ,   79  ,   135 ,   431 ,   723 ,   727 ,   2685    ,   1775    ,   9027    ,   27051   ,   42581   ,0 };
            const unsigned long dim4799KuoInit[]        =   {   1   ,   3   ,   5   ,   9   ,   27  ,   19  ,   17  ,   141 ,   367 ,   1021    ,   651 ,   2549    ,   4003    ,   15737   ,   22063   ,   12443   ,0 };
            const unsigned long dim4800KuoInit[]        =   {   1   ,   1   ,   1   ,   11  ,   23  ,   27  ,   35  ,   225 ,   399 ,   173 ,   451 ,   1701    ,   1093    ,   5639    ,   1855    ,   1817    ,0 };
            const unsigned long dim4801KuoInit[]        =   {   1   ,   3   ,   7   ,   11  ,   23  ,   63  ,   119 ,   13  ,   147 ,   997 ,   53  ,   3127    ,   101 ,   6633    ,   14377   ,   61843   ,0 };
            const unsigned long dim4802KuoInit[]        =   {   1   ,   1   ,   1   ,   15  ,   23  ,   21  ,   65  ,   213 ,   311 ,   789 ,   1395    ,   1239    ,   241 ,   15041   ,   14683   ,   27805   ,0 };
            const unsigned long dim4803KuoInit[]        =   {   1   ,   3   ,   3   ,   9   ,   7   ,   29  ,   83  ,   153 ,   231 ,   337 ,   153 ,   3097    ,   6137    ,   11003   ,   31091   ,   57005   ,0 };
            const unsigned long dim4804KuoInit[]        =   {   1   ,   3   ,   7   ,   9   ,   19  ,   33  ,   91  ,   235 ,   123 ,   747 ,   1331    ,   2307    ,   3791    ,   2063    ,   23613   ,   3589    ,0 };
            const unsigned long dim4805KuoInit[]        =   {   1   ,   3   ,   3   ,   15  ,   9   ,   25  ,   95  ,   237 ,   381 ,   349 ,   1339    ,   3187    ,   2387    ,   3487    ,   4149    ,   39343   ,0 };
            const unsigned long dim4806KuoInit[]        =   {   1   ,   1   ,   1   ,   13  ,   5   ,   3   ,   43  ,   193 ,   431 ,   853 ,   1419    ,   905 ,   2395    ,   11459   ,   26109   ,   31277   ,0 };
            const unsigned long dim4807KuoInit[]        =   {   1   ,   3   ,   1   ,   15  ,   13  ,   29  ,   13  ,   253 ,   475 ,   81  ,   1199    ,   841 ,   7385    ,   2609    ,   10445   ,   11151   ,0 };
            const unsigned long dim4808KuoInit[]        =   {   1   ,   1   ,   7   ,   15  ,   29  ,   11  ,   81  ,   91  ,   287 ,   615 ,   949 ,   2541    ,   5059    ,   12567   ,   9547    ,   45647   ,0 };
            const unsigned long dim4809KuoInit[]        =   {   1   ,   3   ,   7   ,   13  ,   9   ,   27  ,   123 ,   165 ,   249 ,   457 ,   607 ,   1203    ,   987 ,   4477    ,   21205   ,   44547   ,0 };
            const unsigned long dim4810KuoInit[]        =   {   1   ,   1   ,   3   ,   15  ,   3   ,   1   ,   37  ,   43  ,   445 ,   205 ,   1849    ,   3461    ,   7667    ,   209 ,   8773    ,   1507    ,0 };
            const unsigned long dim4811KuoInit[]        =   {   1   ,   3   ,   1   ,   3   ,   27  ,   43  ,   75  ,   199 ,   185 ,   249 ,   725 ,   301 ,   2181    ,   6859    ,   16825   ,   3471    ,0 };
            const unsigned long dim4812KuoInit[]        =   {   1   ,   3   ,   1   ,   9   ,   29  ,   5   ,   55  ,   5   ,   345 ,   325 ,   1853    ,   3925    ,   1739    ,   14991   ,   27845   ,   6931    ,0 };
            const unsigned long dim4813KuoInit[]        =   {   1   ,   1   ,   3   ,   5   ,   31  ,   15  ,   33  ,   47  ,   201 ,   649 ,   415 ,   639 ,   391 ,   12065   ,   17343   ,   3063    ,0 };
            const unsigned long dim4814KuoInit[]        =   {   1   ,   1   ,   7   ,   11  ,   19  ,   37  ,   39  ,   77  ,   367 ,   29  ,   1251    ,   2637    ,   5601    ,   10661   ,   17647   ,   22131   ,0 };
            const unsigned long dim4815KuoInit[]        =   {   1   ,   1   ,   7   ,   15  ,   29  ,   51  ,   13  ,   129 ,   279 ,   427 ,   1535    ,   2185    ,   2651    ,   8251    ,   14857   ,   61467   ,0 };
            const unsigned long dim4816KuoInit[]        =   {   1   ,   1   ,   7   ,   5   ,   25  ,   55  ,   45  ,   105 ,   213 ,   31  ,   735 ,   3795    ,   6627    ,   14299   ,   3155    ,   3659    ,0 };
            const unsigned long dim4817KuoInit[]        =   {   1   ,   1   ,   1   ,   9   ,   21  ,   27  ,   21  ,   85  ,   315 ,   793 ,   1663    ,   437 ,   4113    ,   4639    ,   11321   ,   13633   ,0 };
            const unsigned long dim4818KuoInit[]        =   {   1   ,   3   ,   5   ,   9   ,   23  ,   17  ,   49  ,   35  ,   451 ,   799 ,   1451    ,   539 ,   4549    ,   43  ,   1901    ,   21155   ,0 };
            const unsigned long dim4819KuoInit[]        =   {   1   ,   3   ,   7   ,   9   ,   23  ,   53  ,   25  ,   71  ,   371 ,   1005    ,   129 ,   157 ,   2977    ,   15823   ,   30393   ,   55629   ,0 };
            const unsigned long dim4820KuoInit[]        =   {   1   ,   1   ,   7   ,   7   ,   9   ,   27  ,   79  ,   101 ,   147 ,   511 ,   199 ,   2843    ,   101 ,   14975   ,   3455    ,   843 ,0 };
            const unsigned long dim4821KuoInit[]        =   {   1   ,   3   ,   1   ,   7   ,   11  ,   33  ,   43  ,   7   ,   445 ,   501 ,   1071    ,   1851    ,   5763    ,   5233    ,   16121   ,   23357   ,0 };
            const unsigned long dim4822KuoInit[]        =   {   1   ,   1   ,   1   ,   13  ,   11  ,   21  ,   75  ,   153 ,   421 ,   311 ,   1709    ,   3745    ,   2971    ,   11219   ,   23107   ,   40057   ,0 };
            const unsigned long dim4823KuoInit[]        =   {   1   ,   1   ,   3   ,   15  ,   17  ,   39  ,   17  ,   165 ,   103 ,   31  ,   575 ,   3945    ,   1921    ,   11305   ,   13417   ,   20603   ,0 };
            const unsigned long dim4824KuoInit[]        =   {   1   ,   3   ,   3   ,   3   ,   5   ,   49  ,   37  ,   61  ,   177 ,   595 ,   1133    ,   1883    ,   819 ,   5137    ,   10423   ,   42009   ,0 };
            const unsigned long dim4825KuoInit[]        =   {   1   ,   1   ,   3   ,   5   ,   5   ,   15  ,   89  ,   191 ,   97  ,   1017    ,   205 ,   4037    ,   3431    ,   13877   ,   29867   ,   26041   ,0 };
            const unsigned long dim4826KuoInit[]        =   {   1   ,   3   ,   3   ,   7   ,   9   ,   9   ,   109 ,   157 ,   357 ,   815 ,   1539    ,   3507    ,   1379    ,   1105    ,   14387   ,   46689   ,0 };
            const unsigned long dim4827KuoInit[]        =   {   1   ,   3   ,   3   ,   9   ,   3   ,   11  ,   77  ,   161 ,   473 ,   619 ,   741 ,   2945    ,   5123    ,   3325    ,   8679    ,   36167   ,0 };
            const unsigned long dim4828KuoInit[]        =   {   1   ,   1   ,   1   ,   9   ,   9   ,   53  ,   3   ,   77  ,   319 ,   223 ,   875 ,   425 ,   1369    ,   509 ,   25895   ,   32803   ,0 };
            const unsigned long dim4829KuoInit[]        =   {   1   ,   1   ,   3   ,   13  ,   31  ,   15  ,   113 ,   133 ,   473 ,   29  ,   1009    ,   1385    ,   871 ,   15785   ,   1015    ,   42627   ,0 };
            const unsigned long dim4830KuoInit[]        =   {   1   ,   3   ,   7   ,   13  ,   5   ,   53  ,   123 ,   7   ,   465 ,   485 ,   1413    ,   1663    ,   6541    ,   13051   ,   12251   ,   36151   ,0 };
            const unsigned long dim4831KuoInit[]        =   {   1   ,   1   ,   1   ,   13  ,   9   ,   21  ,   123 ,   223 ,   483 ,   849 ,   489 ,   3905    ,   4697    ,   7583    ,   5373    ,   62397   ,0 };
            const unsigned long dim4832KuoInit[]        =   {   1   ,   1   ,   7   ,   9   ,   29  ,   53  ,   85  ,   151 ,   511 ,   917 ,   1719    ,   937 ,   2035    ,   6385    ,   23529   ,   62213   ,0 };
            const unsigned long dim4833KuoInit[]        =   {   1   ,   1   ,   3   ,   15  ,   11  ,   47  ,   111 ,   79  ,   377 ,   585 ,   475 ,   3201    ,   7025    ,   1483    ,   8815    ,   47235   ,0 };
            const unsigned long dim4834KuoInit[]        =   {   1   ,   1   ,   3   ,   13  ,   13  ,   5   ,   39  ,   31  ,   71  ,   39  ,   941 ,   2215    ,   2195    ,   2827    ,   24353   ,   18435   ,0 };
            const unsigned long dim4835KuoInit[]        =   {   1   ,   1   ,   7   ,   15  ,   13  ,   31  ,   31  ,   15  ,   421 ,   81  ,   895 ,   1345    ,   7049    ,   4373    ,   28005   ,   41781   ,0 };
            const unsigned long dim4836KuoInit[]        =   {   1   ,   1   ,   3   ,   15  ,   17  ,   57  ,   21  ,   175 ,   453 ,   755 ,   1295    ,   353 ,   5257    ,   15901   ,   15643   ,   31725   ,0 };
            const unsigned long dim4837KuoInit[]        =   {   1   ,   1   ,   1   ,   9   ,   17  ,   17  ,   71  ,   209 ,   155 ,   511 ,   1461    ,   3313    ,   7515    ,   8407    ,   9561    ,   48169   ,0 };
            const unsigned long dim4838KuoInit[]        =   {   1   ,   3   ,   7   ,   15  ,   23  ,   59  ,   89  ,   39  ,   85  ,   857 ,   1285    ,   3099    ,   6497    ,   5501    ,   21321   ,   34319   ,0 };
            const unsigned long dim4839KuoInit[]        =   {   1   ,   3   ,   5   ,   5   ,   27  ,   13  ,   41  ,   1   ,   191 ,   275 ,   1567    ,   895 ,   13  ,   673 ,   10079   ,   55075   ,0 };
            const unsigned long dim4840KuoInit[]        =   {   1   ,   1   ,   7   ,   15  ,   9   ,   33  ,   33  ,   35  ,   55  ,   597 ,   365 ,   3205    ,   6427    ,   837 ,   16011   ,   23279   ,0 };
            const unsigned long dim4841KuoInit[]        =   {   1   ,   3   ,   5   ,   11  ,   23  ,   23  ,   9   ,   79  ,   383 ,   303 ,   1715    ,   1651    ,   7507    ,   14553   ,   30715   ,   56659   ,0 };
            const unsigned long dim4842KuoInit[]        =   {   1   ,   3   ,   1   ,   11  ,   1   ,   61  ,   31  ,   85  ,   181 ,   593 ,   1783    ,   901 ,   6069    ,   1521    ,   13275   ,   18057   ,0 };
            const unsigned long dim4843KuoInit[]        =   {   1   ,   3   ,   1   ,   13  ,   23  ,   59  ,   111 ,   221 ,   47  ,   599 ,   193 ,   1547    ,   4077    ,   15271   ,   19315   ,   1115    ,0 };
            const unsigned long dim4844KuoInit[]        =   {   1   ,   3   ,   7   ,   7   ,   23  ,   53  ,   115 ,   69  ,   19  ,   409 ,   1041    ,   3395    ,   6429    ,   9263    ,   21637   ,   35533   ,0 };
            const unsigned long dim4845KuoInit[]        =   {   1   ,   1   ,   7   ,   1   ,   31  ,   45  ,   43  ,   251 ,   163 ,   907 ,   761 ,   2009    ,   3807    ,   1529    ,   12001   ,   38823   ,0 };
            const unsigned long dim4846KuoInit[]        =   {   1   ,   3   ,   5   ,   3   ,   3   ,   59  ,   87  ,   247 ,   179 ,   931 ,   709 ,   53  ,   6283    ,   13723   ,   15407   ,   22417   ,0 };
            const unsigned long dim4847KuoInit[]        =   {   1   ,   1   ,   1   ,   5   ,   27  ,   17  ,   21  ,   183 ,   393 ,   177 ,   1187    ,   19  ,   5409    ,   2539    ,   17401   ,   29287   ,0 };
            const unsigned long dim4848KuoInit[]        =   {   1   ,   1   ,   5   ,   15  ,   3   ,   51  ,   91  ,   95  ,   389 ,   855 ,   1453    ,   2815    ,   7507    ,   5449    ,   12459   ,   59111   ,0 };
            const unsigned long dim4849KuoInit[]        =   {   1   ,   1   ,   5   ,   3   ,   7   ,   53  ,   101 ,   255 ,   395 ,   345 ,   915 ,   1649    ,   6641    ,   7771    ,   20747   ,   17067   ,0 };
            const unsigned long dim4850KuoInit[]        =   {   1   ,   3   ,   1   ,   7   ,   5   ,   35  ,   69  ,   63  ,   249 ,   189 ,   1937    ,   965 ,   3463    ,   13615   ,   12449   ,   11891   ,0 };
            const unsigned long dim4851KuoInit[]        =   {   1   ,   3   ,   3   ,   9   ,   5   ,   59  ,   57  ,   187 ,   403 ,   825 ,   1291    ,   2995    ,   6365    ,   15995   ,   6871    ,   25785   ,0 };
            const unsigned long dim4852KuoInit[]        =   {   1   ,   3   ,   5   ,   15  ,   17  ,   23  ,   117 ,   103 ,   131 ,   601 ,   1259    ,   1331    ,   3039    ,   4761    ,   8105    ,   14521   ,0 };
            const unsigned long dim4853KuoInit[]        =   {   1   ,   1   ,   3   ,   15  ,   27  ,   51  ,   97  ,   221 ,   345 ,   615 ,   487 ,   2535    ,   139 ,   16233   ,   16271   ,   713 ,0 };
            const unsigned long dim4854KuoInit[]        =   {   1   ,   3   ,   7   ,   13  ,   25  ,   49  ,   99  ,   71  ,   353 ,   1015    ,   1889    ,   689 ,   1925    ,   6523    ,   30995   ,   23057   ,0 };
            const unsigned long dim4855KuoInit[]        =   {   1   ,   3   ,   5   ,   9   ,   21  ,   55  ,   11  ,   29  ,   477 ,   581 ,   1719    ,   2603    ,   1757    ,   3995    ,   27025   ,   62833   ,0 };
            const unsigned long dim4856KuoInit[]        =   {   1   ,   3   ,   1   ,   9   ,   9   ,   39  ,   9   ,   121 ,   459 ,   991 ,   2021    ,   1469    ,   2399    ,   4441    ,   11829   ,   36865   ,0 };
            const unsigned long dim4857KuoInit[]        =   {   1   ,   3   ,   1   ,   5   ,   9   ,   13  ,   115 ,   27  ,   487 ,   307 ,   453 ,   351 ,   885 ,   12539   ,   39  ,   33887   ,0 };
            const unsigned long dim4858KuoInit[]        =   {   1   ,   3   ,   1   ,   7   ,   23  ,   35  ,   87  ,   131 ,   129 ,   1   ,   2015    ,   3921    ,   1309    ,   7915    ,   16661   ,   45945   ,0 };
            const unsigned long dim4859KuoInit[]        =   {   1   ,   3   ,   1   ,   3   ,   1   ,   15  ,   93  ,   255 ,   285 ,   915 ,   563 ,   527 ,   7763    ,   4581    ,   11161   ,   64619   ,0 };
            const unsigned long dim4860KuoInit[]        =   {   1   ,   3   ,   3   ,   9   ,   31  ,   37  ,   87  ,   81  ,   313 ,   647 ,   1065    ,   863 ,   6121    ,   1385    ,   26171   ,   56323   ,0 };
            const unsigned long dim4861KuoInit[]        =   {   1   ,   1   ,   3   ,   7   ,   25  ,   49  ,   23  ,   107 ,   507 ,   511 ,   951 ,   669 ,   67  ,   14319   ,   17669   ,   32497   ,0 };
            const unsigned long dim4862KuoInit[]        =   {   1   ,   3   ,   1   ,   13  ,   9   ,   33  ,   93  ,   171 ,   457 ,   687 ,   703 ,   1073    ,   3699    ,   7029    ,   28983   ,   14501   ,0 };
            const unsigned long dim4863KuoInit[]        =   {   1   ,   1   ,   7   ,   13  ,   29  ,   3   ,   19  ,   173 ,   131 ,   285 ,   469 ,   1223    ,   7779    ,   15367   ,   643 ,   63715   ,0 };
            const unsigned long dim4864KuoInit[]        =   {   1   ,   3   ,   7   ,   5   ,   5   ,   37  ,   21  ,   223 ,   227 ,   939 ,   1205    ,   3167    ,   5535    ,   11543   ,   7411    ,   30841   ,0 };
            const unsigned long dim4865KuoInit[]        =   {   1   ,   3   ,   5   ,   11  ,   31  ,   37  ,   75  ,   39  ,   171 ,   135 ,   1487    ,   3503    ,   7939    ,   77  ,   19745   ,   2551    ,0 };
            const unsigned long dim4866KuoInit[]        =   {   1   ,   3   ,   5   ,   13  ,   25  ,   33  ,   79  ,   213 ,   289 ,   27  ,   1435    ,   1839    ,   3089    ,   8173    ,   27555   ,   20599   ,0 };
            const unsigned long dim4867KuoInit[]        =   {   1   ,   1   ,   3   ,   13  ,   19  ,   31  ,   67  ,   9   ,   229 ,   591 ,   561 ,   3797    ,   1833    ,   3311    ,   16053   ,   61183   ,0 };
            const unsigned long dim4868KuoInit[]        =   {   1   ,   1   ,   3   ,   3   ,   27  ,   27  ,   79  ,   95  ,   7   ,   1023    ,   517 ,   3627    ,   4117    ,   8165    ,   17961   ,   39011   ,0 };
            const unsigned long dim4869KuoInit[]        =   {   1   ,   3   ,   5   ,   3   ,   3   ,   43  ,   37  ,   231 ,   389 ,   367 ,   517 ,   3117    ,   2643    ,   2103    ,   127 ,   42669   ,0 };
            const unsigned long dim4870KuoInit[]        =   {   1   ,   1   ,   1   ,   11  ,   11  ,   49  ,   35  ,   71  ,   295 ,   87  ,   1653    ,   1899    ,   1177    ,   16071   ,   8429    ,   54751   ,0 };
            const unsigned long dim4871KuoInit[]        =   {   1   ,   1   ,   7   ,   13  ,   5   ,   47  ,   7   ,   229 ,   261 ,   837 ,   645 ,   1581    ,   159 ,   73  ,   10233   ,   18557   ,0 };
            const unsigned long dim4872KuoInit[]        =   {   1   ,   3   ,   1   ,   9   ,   11  ,   45  ,   25  ,   3   ,   271 ,   1005    ,   1193    ,   507 ,   269 ,   14689   ,   7601    ,   56781   ,0 };
            const unsigned long dim4873KuoInit[]        =   {   1   ,   1   ,   5   ,   11  ,   13  ,   61  ,   107 ,   29  ,   155 ,   415 ,   1435    ,   1979    ,   7339    ,   1839    ,   21097   ,   5851    ,0 };
            const unsigned long dim4874KuoInit[]        =   {   1   ,   1   ,   5   ,   9   ,   9   ,   27  ,   83  ,   163 ,   15  ,   539 ,   765 ,   463 ,   7493    ,   1413    ,   15597   ,   42819   ,0 };
            const unsigned long dim4875KuoInit[]        =   {   1   ,   3   ,   1   ,   9   ,   21  ,   7   ,   91  ,   213 ,   463 ,   827 ,   1847    ,   2825    ,   2115    ,   14549   ,   29923   ,   61355   ,0 };
            const unsigned long dim4876KuoInit[]        =   {   1   ,   1   ,   3   ,   15  ,   1   ,   55  ,   69  ,   85  ,   7   ,   847 ,   1935    ,   3225    ,   693 ,   12069   ,   26303   ,   19143   ,0 };
            const unsigned long dim4877KuoInit[]        =   {   1   ,   3   ,   1   ,   11  ,   15  ,   47  ,   53  ,   181 ,   327 ,   493 ,   483 ,   999 ,   4421    ,   6707    ,   27933   ,   11533   ,0 };
            const unsigned long dim4878KuoInit[]        =   {   1   ,   3   ,   3   ,   3   ,   23  ,   7   ,   7   ,   63  ,   235 ,   493 ,   937 ,   2935    ,   2869    ,   14243   ,   30975   ,   12881   ,0 };
            const unsigned long dim4879KuoInit[]        =   {   1   ,   1   ,   3   ,   9   ,   9   ,   7   ,   113 ,   203 ,   297 ,   183 ,   1687    ,   3211    ,   8131    ,   7033    ,   15007   ,   1307    ,0 };
            const unsigned long dim4880KuoInit[]        =   {   1   ,   3   ,   1   ,   9   ,   9   ,   39  ,   99  ,   5   ,   61  ,   435 ,   427 ,   2719    ,   7741    ,   4267    ,   2423    ,   37573   ,0 };
            const unsigned long dim4881KuoInit[]        =   {   1   ,   3   ,   7   ,   3   ,   17  ,   1   ,   105 ,   13  ,   89  ,   693 ,   1835    ,   2455    ,   1135    ,   15545   ,   2447    ,   25673   ,0 };
            const unsigned long dim4882KuoInit[]        =   {   1   ,   3   ,   5   ,   5   ,   15  ,   47  ,   103 ,   167 ,   21  ,   791 ,   1297    ,   3523    ,   4427    ,   7013    ,   1845    ,   44461   ,0 };
            const unsigned long dim4883KuoInit[]        =   {   1   ,   1   ,   5   ,   7   ,   11  ,   17  ,   55  ,   247 ,   317 ,   229 ,   1465    ,   3781    ,   2397    ,   10541   ,   8943    ,   29155   ,0 };
            const unsigned long dim4884KuoInit[]        =   {   1   ,   1   ,   5   ,   3   ,   17  ,   17  ,   97  ,   141 ,   153 ,   883 ,   449 ,   1163    ,   3165    ,   125 ,   4689    ,   44461   ,0 };
            const unsigned long dim4885KuoInit[]        =   {   1   ,   3   ,   7   ,   1   ,   17  ,   13  ,   9   ,   235 ,   87  ,   251 ,   57  ,   2769    ,   4845    ,   5733    ,   4305    ,   49489   ,0 };
            const unsigned long dim4886KuoInit[]        =   {   1   ,   1   ,   7   ,   1   ,   9   ,   41  ,   79  ,   97  ,   201 ,   789 ,   37  ,   2593    ,   8023    ,   139 ,   30623   ,   7667    ,0 };
            const unsigned long dim4887KuoInit[]        =   {   1   ,   1   ,   7   ,   7   ,   7   ,   27  ,   9   ,   99  ,   199 ,   163 ,   1175    ,   729 ,   2731    ,   4391    ,   27597   ,   23087   ,0 };
            const unsigned long dim4888KuoInit[]        =   {   1   ,   3   ,   5   ,   5   ,   29  ,   39  ,   5   ,   129 ,   395 ,   1007    ,   1751    ,   3951    ,   5879    ,   13105   ,   26223   ,   31361   ,0 };
            const unsigned long dim4889KuoInit[]        =   {   1   ,   3   ,   5   ,   3   ,   3   ,   55  ,   89  ,   253 ,   207 ,   553 ,   1407    ,   2277    ,   6879    ,   16027   ,   23477   ,   43633   ,0 };
            const unsigned long dim4890KuoInit[]        =   {   1   ,   1   ,   1   ,   3   ,   13  ,   21  ,   9   ,   149 ,   31  ,   359 ,   1631    ,   2199    ,   2937    ,   15663   ,   5899    ,   59959   ,0 };
            const unsigned long dim4891KuoInit[]        =   {   1   ,   1   ,   1   ,   7   ,   25  ,   63  ,   87  ,   13  ,   179 ,   237 ,   359 ,   3841    ,   6467    ,   5621    ,   29459   ,   31339   ,0 };
            const unsigned long dim4892KuoInit[]        =   {   1   ,   1   ,   1   ,   15  ,   25  ,   45  ,   77  ,   17  ,   419 ,   759 ,   1045    ,   1285    ,   6739    ,   11035   ,   11401   ,   44821   ,0 };
            const unsigned long dim4893KuoInit[]        =   {   1   ,   3   ,   7   ,   1   ,   31  ,   47  ,   17  ,   237 ,   69  ,   389 ,   1251    ,   1205    ,   833 ,   2313    ,   17005   ,   32683   ,0 };
            const unsigned long dim4894KuoInit[]        =   {   1   ,   1   ,   7   ,   9   ,   13  ,   19  ,   51  ,   69  ,   261 ,   85  ,   215 ,   2215    ,   1643    ,   12985   ,   10363   ,   17123   ,0 };
            const unsigned long dim4895KuoInit[]        =   {   1   ,   3   ,   1   ,   3   ,   3   ,   51  ,   41  ,   173 ,   161 ,   633 ,   1727    ,   3089    ,   7791    ,   7367    ,   32641   ,   14361   ,0 };
            const unsigned long dim4896KuoInit[]        =   {   1   ,   3   ,   5   ,   3   ,   21  ,   49  ,   19  ,   243 ,   427 ,   589 ,   1699    ,   2187    ,   4177    ,   927 ,   6201    ,   46319   ,0 };
            const unsigned long dim4897KuoInit[]        =   {   1   ,   1   ,   3   ,   11  ,   1   ,   59  ,   49  ,   235 ,   345 ,   731 ,   793 ,   1893    ,   4735    ,   10647   ,   31169   ,   12051   ,0 };
            const unsigned long dim4898KuoInit[]        =   {   1   ,   1   ,   7   ,   15  ,   23  ,   51  ,   67  ,   151 ,   253 ,   793 ,   1673    ,   3255    ,   4919    ,   11309   ,   24903   ,   9083    ,0 };
            const unsigned long dim4899KuoInit[]        =   {   1   ,   3   ,   3   ,   9   ,   7   ,   59  ,   41  ,   225 ,   397 ,   671 ,   513 ,   1351    ,   2835    ,   15789   ,   17679   ,   28641   ,0 };
            const unsigned long dim4900KuoInit[]        =   {   1   ,   3   ,   1   ,   7   ,   17  ,   39  ,   49  ,   103 ,   181 ,   375 ,   1843    ,   1183    ,   4219    ,   2931    ,   18385   ,   1203    ,0 };
            const unsigned long dim4901KuoInit[]        =   {   1   ,   1   ,   1   ,   3   ,   29  ,   19  ,   113 ,   55  ,   81  ,   69  ,   1541    ,   3249    ,   643 ,   841 ,   6053    ,   63571   ,0 };
            const unsigned long dim4902KuoInit[]        =   {   1   ,   1   ,   7   ,   5   ,   1   ,   3   ,   39  ,   237 ,   283 ,   283 ,   813 ,   3539    ,   5899    ,   8489    ,   18393   ,   22371   ,0 };
            const unsigned long dim4903KuoInit[]        =   {   1   ,   3   ,   7   ,   13  ,   5   ,   53  ,   15  ,   129 ,   95  ,   181 ,   657 ,   3735    ,   7517    ,   2435    ,   17245   ,   4075    ,0 };
            const unsigned long dim4904KuoInit[]        =   {   1   ,   1   ,   5   ,   3   ,   9   ,   37  ,   73  ,   95  ,   175 ,   989 ,   1443    ,   1745    ,   8117    ,   5867    ,   29729   ,   28441   ,0 };
            const unsigned long dim4905KuoInit[]        =   {   1   ,   3   ,   3   ,   11  ,   5   ,   41  ,   25  ,   3   ,   177 ,   735 ,   187 ,   357 ,   5365    ,   8161    ,   2071    ,   41651   ,0 };
            const unsigned long dim4906KuoInit[]        =   {   1   ,   1   ,   1   ,   15  ,   29  ,   47  ,   27  ,   233 ,   177 ,   339 ,   1839    ,   1761    ,   5483    ,   7203    ,   32411   ,   59859   ,0 };
            const unsigned long dim4907KuoInit[]        =   {   1   ,   3   ,   7   ,   9   ,   13  ,   59  ,   25  ,   17  ,   157 ,   247 ,   697 ,   4011    ,   4539    ,   4253    ,   1381    ,   63009   ,0 };
            const unsigned long dim4908KuoInit[]        =   {   1   ,   1   ,   5   ,   13  ,   19  ,   21  ,   19  ,   161 ,   325 ,   389 ,   1067    ,   525 ,   529 ,   277 ,   13775   ,   52265   ,0 };
            const unsigned long dim4909KuoInit[]        =   {   1   ,   1   ,   5   ,   13  ,   21  ,   17  ,   77  ,   191 ,   329 ,   627 ,   1857    ,   1705    ,   2207    ,   5329    ,   29833   ,   1523    ,0 };
            const unsigned long dim4910KuoInit[]        =   {   1   ,   3   ,   3   ,   1   ,   21  ,   3   ,   107 ,   85  ,   29  ,   887 ,   793 ,   1089    ,   1361    ,   10361   ,   15869   ,   1659    ,0 };
            const unsigned long dim4911KuoInit[]        =   {   1   ,   1   ,   1   ,   15  ,   11  ,   47  ,   85  ,   137 ,   373 ,   777 ,   875 ,   2735    ,   4381    ,   8831    ,   5891    ,   44809   ,0 };
            const unsigned long dim4912KuoInit[]        =   {   1   ,   1   ,   1   ,   11  ,   19  ,   9   ,   77  ,   59  ,   497 ,   643 ,   51  ,   3115    ,   7267    ,   11927   ,   28119   ,   28151   ,0 };
            const unsigned long dim4913KuoInit[]        =   {   1   ,   3   ,   1   ,   3   ,   13  ,   27  ,   55  ,   201 ,   319 ,   167 ,   1435    ,   2197    ,   5049    ,   5923    ,   12759   ,   62383   ,0 };
            const unsigned long dim4914KuoInit[]        =   {   1   ,   3   ,   1   ,   3   ,   23  ,   43  ,   45  ,   193 ,   221 ,   647 ,   1933    ,   2915    ,   6461    ,   4927    ,   22105   ,   54557   ,0 };
            const unsigned long dim4915KuoInit[]        =   {   1   ,   1   ,   5   ,   11  ,   5   ,   23  ,   67  ,   59  ,   79  ,   507 ,   925 ,   3015    ,   317 ,   3887    ,   29967   ,   17205   ,0 };
            const unsigned long dim4916KuoInit[]        =   {   1   ,   3   ,   7   ,   7   ,   19  ,   39  ,   51  ,   159 ,   485 ,   315 ,   1739    ,   2617    ,   8017    ,   11655   ,   1141    ,   26055   ,0 };
            const unsigned long dim4917KuoInit[]        =   {   1   ,   1   ,   3   ,   1   ,   27  ,   53  ,   119 ,   113 ,   135 ,   85  ,   2023    ,   1659    ,   307 ,   8189    ,   28339   ,   53297   ,0 };
            const unsigned long dim4918KuoInit[]        =   {   1   ,   3   ,   3   ,   5   ,   9   ,   61  ,   57  ,   225 ,   137 ,   381 ,   1045    ,   1253    ,   1185    ,   9885    ,   30153   ,   29355   ,0 };
            const unsigned long dim4919KuoInit[]        =   {   1   ,   1   ,   3   ,   5   ,   19  ,   9   ,   87  ,   87  ,   393 ,   421 ,   1941    ,   57  ,   2483    ,   10855   ,   26741   ,   28091   ,0 };
            const unsigned long dim4920KuoInit[]        =   {   1   ,   1   ,   7   ,   5   ,   3   ,   35  ,   35  ,   137 ,   123 ,   327 ,   933 ,   3429    ,   281 ,   1705    ,   17123   ,   29337   ,0 };
            const unsigned long dim4921KuoInit[]        =   {   1   ,   3   ,   7   ,   13  ,   9   ,   61  ,   111 ,   207 ,   45  ,   57  ,   1093    ,   2223    ,   575 ,   11959   ,   21147   ,   30273   ,0 };
            const unsigned long dim4922KuoInit[]        =   {   1   ,   1   ,   5   ,   5   ,   31  ,   37  ,   123 ,   5   ,   277 ,   273 ,   525 ,   1773    ,   2685    ,   11607   ,   32079   ,   25199   ,0 };
            const unsigned long dim4923KuoInit[]        =   {   1   ,   1   ,   7   ,   11  ,   5   ,   29  ,   121 ,   193 ,   437 ,   417 ,   1521    ,   3057    ,   4681    ,   2773    ,   5391    ,   59747   ,0 };
            const unsigned long dim4924KuoInit[]        =   {   1   ,   3   ,   7   ,   11  ,   27  ,   3   ,   103 ,   17  ,   395 ,   695 ,   1731    ,   4055    ,   2727    ,   2171    ,   26893   ,   7047    ,0 };
            const unsigned long dim4925KuoInit[]        =   {   1   ,   3   ,   7   ,   5   ,   23  ,   9   ,   25  ,   167 ,   241 ,   755 ,   1789    ,   1977    ,   6325    ,   5425    ,   23247   ,   64181   ,0 };


            const unsigned long * const Kuoinitializers[4925]
            =
            {
                dim1KuoInit,
                dim2KuoInit,
                dim3KuoInit,
                dim4KuoInit,
                dim5KuoInit,
                dim6KuoInit,
                dim7KuoInit,
                dim8KuoInit,
                dim9KuoInit,
                dim10KuoInit,
                dim11KuoInit,
                dim12KuoInit,
                dim13KuoInit,
                dim14KuoInit,
                dim15KuoInit,
                dim16KuoInit,
                dim17KuoInit,
                dim18KuoInit,
                dim19KuoInit,
                dim20KuoInit,
                dim21KuoInit,
                dim22KuoInit,
                dim23KuoInit,
                dim24KuoInit,
                dim25KuoInit,
                dim26KuoInit,
                dim27KuoInit,
                dim28KuoInit,
                dim29KuoInit,
                dim30KuoInit,
                dim31KuoInit,
                dim32KuoInit,
                dim33KuoInit,
                dim34KuoInit,
                dim35KuoInit,
                dim36KuoInit,
                dim37KuoInit,
                dim38KuoInit,
                dim39KuoInit,
                dim40KuoInit,
                dim41KuoInit,
                dim42KuoInit,
                dim43KuoInit,
                dim44KuoInit,
                dim45KuoInit,
                dim46KuoInit,
                dim47KuoInit,
                dim48KuoInit,
                dim49KuoInit,
                dim50KuoInit,
                dim51KuoInit,
                dim52KuoInit,
                dim53KuoInit,
                dim54KuoInit,
                dim55KuoInit,
                dim56KuoInit,
                dim57KuoInit,
                dim58KuoInit,
                dim59KuoInit,
                dim60KuoInit,
                dim61KuoInit,
                dim62KuoInit,
                dim63KuoInit,
                dim64KuoInit,
                dim65KuoInit,
                dim66KuoInit,
                dim67KuoInit,
                dim68KuoInit,
                dim69KuoInit,
                dim70KuoInit,
                dim71KuoInit,
                dim72KuoInit,
                dim73KuoInit,
                dim74KuoInit,
                dim75KuoInit,
                dim76KuoInit,
                dim77KuoInit,
                dim78KuoInit,
                dim79KuoInit,
                dim80KuoInit,
                dim81KuoInit,
                dim82KuoInit,
                dim83KuoInit,
                dim84KuoInit,
                dim85KuoInit,
                dim86KuoInit,
                dim87KuoInit,
                dim88KuoInit,
                dim89KuoInit,
                dim90KuoInit,
                dim91KuoInit,
                dim92KuoInit,
                dim93KuoInit,
                dim94KuoInit,
                dim95KuoInit,
                dim96KuoInit,
                dim97KuoInit,
                dim98KuoInit,
                dim99KuoInit,
                dim100KuoInit,
                dim101KuoInit,
                dim102KuoInit,
                dim103KuoInit,
                dim104KuoInit,
                dim105KuoInit,
                dim106KuoInit,
                dim107KuoInit,
                dim108KuoInit,
                dim109KuoInit,
                dim110KuoInit,
                dim111KuoInit,
                dim112KuoInit,
                dim113KuoInit,
                dim114KuoInit,
                dim115KuoInit,
                dim116KuoInit,
                dim117KuoInit,
                dim118KuoInit,
                dim119KuoInit,
                dim120KuoInit,
                dim121KuoInit,
                dim122KuoInit,
                dim123KuoInit,
                dim124KuoInit,
                dim125KuoInit,
                dim126KuoInit,
                dim127KuoInit,
                dim128KuoInit,
                dim129KuoInit,
                dim130KuoInit,
                dim131KuoInit,
                dim132KuoInit,
                dim133KuoInit,
                dim134KuoInit,
                dim135KuoInit,
                dim136KuoInit,
                dim137KuoInit,
                dim138KuoInit,
                dim139KuoInit,
                dim140KuoInit,
                dim141KuoInit,
                dim142KuoInit,
                dim143KuoInit,
                dim144KuoInit,
                dim145KuoInit,
                dim146KuoInit,
                dim147KuoInit,
                dim148KuoInit,
                dim149KuoInit,
                dim150KuoInit,
                dim151KuoInit,
                dim152KuoInit,
                dim153KuoInit,
                dim154KuoInit,
                dim155KuoInit,
                dim156KuoInit,
                dim157KuoInit,
                dim158KuoInit,
                dim159KuoInit,
                dim160KuoInit,
                dim161KuoInit,
                dim162KuoInit,
                dim163KuoInit,
                dim164KuoInit,
                dim165KuoInit,
                dim166KuoInit,
                dim167KuoInit,
                dim168KuoInit,
                dim169KuoInit,
                dim170KuoInit,
                dim171KuoInit,
                dim172KuoInit,
                dim173KuoInit,
                dim174KuoInit,
                dim175KuoInit,
                dim176KuoInit,
                dim177KuoInit,
                dim178KuoInit,
                dim179KuoInit,
                dim180KuoInit,
                dim181KuoInit,
                dim182KuoInit,
                dim183KuoInit,
                dim184KuoInit,
                dim185KuoInit,
                dim186KuoInit,
                dim187KuoInit,
                dim188KuoInit,
                dim189KuoInit,
                dim190KuoInit,
                dim191KuoInit,
                dim192KuoInit,
                dim193KuoInit,
                dim194KuoInit,
                dim195KuoInit,
                dim196KuoInit,
                dim197KuoInit,
                dim198KuoInit,
                dim199KuoInit,
                dim200KuoInit,
                dim201KuoInit,
                dim202KuoInit,
                dim203KuoInit,
                dim204KuoInit,
                dim205KuoInit,
                dim206KuoInit,
                dim207KuoInit,
                dim208KuoInit,
                dim209KuoInit,
                dim210KuoInit,
                dim211KuoInit,
                dim212KuoInit,
                dim213KuoInit,
                dim214KuoInit,
                dim215KuoInit,
                dim216KuoInit,
                dim217KuoInit,
                dim218KuoInit,
                dim219KuoInit,
                dim220KuoInit,
                dim221KuoInit,
                dim222KuoInit,
                dim223KuoInit,
                dim224KuoInit,
                dim225KuoInit,
                dim226KuoInit,
                dim227KuoInit,
                dim228KuoInit,
                dim229KuoInit,
                dim230KuoInit,
                dim231KuoInit,
                dim232KuoInit,
                dim233KuoInit,
                dim234KuoInit,
                dim235KuoInit,
                dim236KuoInit,
                dim237KuoInit,
                dim238KuoInit,
                dim239KuoInit,
                dim240KuoInit,
                dim241KuoInit,
                dim242KuoInit,
                dim243KuoInit,
                dim244KuoInit,
                dim245KuoInit,
                dim246KuoInit,
                dim247KuoInit,
                dim248KuoInit,
                dim249KuoInit,
                dim250KuoInit,
                dim251KuoInit,
                dim252KuoInit,
                dim253KuoInit,
                dim254KuoInit,
                dim255KuoInit,
                dim256KuoInit,
                dim257KuoInit,
                dim258KuoInit,
                dim259KuoInit,
                dim260KuoInit,
                dim261KuoInit,
                dim262KuoInit,
                dim263KuoInit,
                dim264KuoInit,
                dim265KuoInit,
                dim266KuoInit,
                dim267KuoInit,
                dim268KuoInit,
                dim269KuoInit,
                dim270KuoInit,
                dim271KuoInit,
                dim272KuoInit,
                dim273KuoInit,
                dim274KuoInit,
                dim275KuoInit,
                dim276KuoInit,
                dim277KuoInit,
                dim278KuoInit,
                dim279KuoInit,
                dim280KuoInit,
                dim281KuoInit,
                dim282KuoInit,
                dim283KuoInit,
                dim284KuoInit,
                dim285KuoInit,
                dim286KuoInit,
                dim287KuoInit,
                dim288KuoInit,
                dim289KuoInit,
                dim290KuoInit,
                dim291KuoInit,
                dim292KuoInit,
                dim293KuoInit,
                dim294KuoInit,
                dim295KuoInit,
                dim296KuoInit,
                dim297KuoInit,
                dim298KuoInit,
                dim299KuoInit,
                dim300KuoInit,
                dim301KuoInit,
                dim302KuoInit,
                dim303KuoInit,
                dim304KuoInit,
                dim305KuoInit,
                dim306KuoInit,
                dim307KuoInit,
                dim308KuoInit,
                dim309KuoInit,
                dim310KuoInit,
                dim311KuoInit,
                dim312KuoInit,
                dim313KuoInit,
                dim314KuoInit,
                dim315KuoInit,
                dim316KuoInit,
                dim317KuoInit,
                dim318KuoInit,
                dim319KuoInit,
                dim320KuoInit,
                dim321KuoInit,
                dim322KuoInit,
                dim323KuoInit,
                dim324KuoInit,
                dim325KuoInit,
                dim326KuoInit,
                dim327KuoInit,
                dim328KuoInit,
                dim329KuoInit,
                dim330KuoInit,
                dim331KuoInit,
                dim332KuoInit,
                dim333KuoInit,
                dim334KuoInit,
                dim335KuoInit,
                dim336KuoInit,
                dim337KuoInit,
                dim338KuoInit,
                dim339KuoInit,
                dim340KuoInit,
                dim341KuoInit,
                dim342KuoInit,
                dim343KuoInit,
                dim344KuoInit,
                dim345KuoInit,
                dim346KuoInit,
                dim347KuoInit,
                dim348KuoInit,
                dim349KuoInit,
                dim350KuoInit,
                dim351KuoInit,
                dim352KuoInit,
                dim353KuoInit,
                dim354KuoInit,
                dim355KuoInit,
                dim356KuoInit,
                dim357KuoInit,
                dim358KuoInit,
                dim359KuoInit,
                dim360KuoInit,
                dim361KuoInit,
                dim362KuoInit,
                dim363KuoInit,
                dim364KuoInit,
                dim365KuoInit,
                dim366KuoInit,
                dim367KuoInit,
                dim368KuoInit,
                dim369KuoInit,
                dim370KuoInit,
                dim371KuoInit,
                dim372KuoInit,
                dim373KuoInit,
                dim374KuoInit,
                dim375KuoInit,
                dim376KuoInit,
                dim377KuoInit,
                dim378KuoInit,
                dim379KuoInit,
                dim380KuoInit,
                dim381KuoInit,
                dim382KuoInit,
                dim383KuoInit,
                dim384KuoInit,
                dim385KuoInit,
                dim386KuoInit,
                dim387KuoInit,
                dim388KuoInit,
                dim389KuoInit,
                dim390KuoInit,
                dim391KuoInit,
                dim392KuoInit,
                dim393KuoInit,
                dim394KuoInit,
                dim395KuoInit,
                dim396KuoInit,
                dim397KuoInit,
                dim398KuoInit,
                dim399KuoInit,
                dim400KuoInit,
                dim401KuoInit,
                dim402KuoInit,
                dim403KuoInit,
                dim404KuoInit,
                dim405KuoInit,
                dim406KuoInit,
                dim407KuoInit,
                dim408KuoInit,
                dim409KuoInit,
                dim410KuoInit,
                dim411KuoInit,
                dim412KuoInit,
                dim413KuoInit,
                dim414KuoInit,
                dim415KuoInit,
                dim416KuoInit,
                dim417KuoInit,
                dim418KuoInit,
                dim419KuoInit,
                dim420KuoInit,
                dim421KuoInit,
                dim422KuoInit,
                dim423KuoInit,
                dim424KuoInit,
                dim425KuoInit,
                dim426KuoInit,
                dim427KuoInit,
                dim428KuoInit,
                dim429KuoInit,
                dim430KuoInit,
                dim431KuoInit,
                dim432KuoInit,
                dim433KuoInit,
                dim434KuoInit,
                dim435KuoInit,
                dim436KuoInit,
                dim437KuoInit,
                dim438KuoInit,
                dim439KuoInit,
                dim440KuoInit,
                dim441KuoInit,
                dim442KuoInit,
                dim443KuoInit,
                dim444KuoInit,
                dim445KuoInit,
                dim446KuoInit,
                dim447KuoInit,
                dim448KuoInit,
                dim449KuoInit,
                dim450KuoInit,
                dim451KuoInit,
                dim452KuoInit,
                dim453KuoInit,
                dim454KuoInit,
                dim455KuoInit,
                dim456KuoInit,
                dim457KuoInit,
                dim458KuoInit,
                dim459KuoInit,
                dim460KuoInit,
                dim461KuoInit,
                dim462KuoInit,
                dim463KuoInit,
                dim464KuoInit,
                dim465KuoInit,
                dim466KuoInit,
                dim467KuoInit,
                dim468KuoInit,
                dim469KuoInit,
                dim470KuoInit,
                dim471KuoInit,
                dim472KuoInit,
                dim473KuoInit,
                dim474KuoInit,
                dim475KuoInit,
                dim476KuoInit,
                dim477KuoInit,
                dim478KuoInit,
                dim479KuoInit,
                dim480KuoInit,
                dim481KuoInit,
                dim482KuoInit,
                dim483KuoInit,
                dim484KuoInit,
                dim485KuoInit,
                dim486KuoInit,
                dim487KuoInit,
                dim488KuoInit,
                dim489KuoInit,
                dim490KuoInit,
                dim491KuoInit,
                dim492KuoInit,
                dim493KuoInit,
                dim494KuoInit,
                dim495KuoInit,
                dim496KuoInit,
                dim497KuoInit,
                dim498KuoInit,
                dim499KuoInit,
                dim500KuoInit,
                dim501KuoInit,
                dim502KuoInit,
                dim503KuoInit,
                dim504KuoInit,
                dim505KuoInit,
                dim506KuoInit,
                dim507KuoInit,
                dim508KuoInit,
                dim509KuoInit,
                dim510KuoInit,
                dim511KuoInit,
                dim512KuoInit,
                dim513KuoInit,
                dim514KuoInit,
                dim515KuoInit,
                dim516KuoInit,
                dim517KuoInit,
                dim518KuoInit,
                dim519KuoInit,
                dim520KuoInit,
                dim521KuoInit,
                dim522KuoInit,
                dim523KuoInit,
                dim524KuoInit,
                dim525KuoInit,
                dim526KuoInit,
                dim527KuoInit,
                dim528KuoInit,
                dim529KuoInit,
                dim530KuoInit,
                dim531KuoInit,
                dim532KuoInit,
                dim533KuoInit,
                dim534KuoInit,
                dim535KuoInit,
                dim536KuoInit,
                dim537KuoInit,
                dim538KuoInit,
                dim539KuoInit,
                dim540KuoInit,
                dim541KuoInit,
                dim542KuoInit,
                dim543KuoInit,
                dim544KuoInit,
                dim545KuoInit,
                dim546KuoInit,
                dim547KuoInit,
                dim548KuoInit,
                dim549KuoInit,
                dim550KuoInit,
                dim551KuoInit,
                dim552KuoInit,
                dim553KuoInit,
                dim554KuoInit,
                dim555KuoInit,
                dim556KuoInit,
                dim557KuoInit,
                dim558KuoInit,
                dim559KuoInit,
                dim560KuoInit,
                dim561KuoInit,
                dim562KuoInit,
                dim563KuoInit,
                dim564KuoInit,
                dim565KuoInit,
                dim566KuoInit,
                dim567KuoInit,
                dim568KuoInit,
                dim569KuoInit,
                dim570KuoInit,
                dim571KuoInit,
                dim572KuoInit,
                dim573KuoInit,
                dim574KuoInit,
                dim575KuoInit,
                dim576KuoInit,
                dim577KuoInit,
                dim578KuoInit,
                dim579KuoInit,
                dim580KuoInit,
                dim581KuoInit,
                dim582KuoInit,
                dim583KuoInit,
                dim584KuoInit,
                dim585KuoInit,
                dim586KuoInit,
                dim587KuoInit,
                dim588KuoInit,
                dim589KuoInit,
                dim590KuoInit,
                dim591KuoInit,
                dim592KuoInit,
                dim593KuoInit,
                dim594KuoInit,
                dim595KuoInit,
                dim596KuoInit,
                dim597KuoInit,
                dim598KuoInit,
                dim599KuoInit,
                dim600KuoInit,
                dim601KuoInit,
                dim602KuoInit,
                dim603KuoInit,
                dim604KuoInit,
                dim605KuoInit,
                dim606KuoInit,
                dim607KuoInit,
                dim608KuoInit,
                dim609KuoInit,
                dim610KuoInit,
                dim611KuoInit,
                dim612KuoInit,
                dim613KuoInit,
                dim614KuoInit,
                dim615KuoInit,
                dim616KuoInit,
                dim617KuoInit,
                dim618KuoInit,
                dim619KuoInit,
                dim620KuoInit,
                dim621KuoInit,
                dim622KuoInit,
                dim623KuoInit,
                dim624KuoInit,
                dim625KuoInit,
                dim626KuoInit,
                dim627KuoInit,
                dim628KuoInit,
                dim629KuoInit,
                dim630KuoInit,
                dim631KuoInit,
                dim632KuoInit,
                dim633KuoInit,
                dim634KuoInit,
                dim635KuoInit,
                dim636KuoInit,
                dim637KuoInit,
                dim638KuoInit,
                dim639KuoInit,
                dim640KuoInit,
                dim641KuoInit,
                dim642KuoInit,
                dim643KuoInit,
                dim644KuoInit,
                dim645KuoInit,
                dim646KuoInit,
                dim647KuoInit,
                dim648KuoInit,
                dim649KuoInit,
                dim650KuoInit,
                dim651KuoInit,
                dim652KuoInit,
                dim653KuoInit,
                dim654KuoInit,
                dim655KuoInit,
                dim656KuoInit,
                dim657KuoInit,
                dim658KuoInit,
                dim659KuoInit,
                dim660KuoInit,
                dim661KuoInit,
                dim662KuoInit,
                dim663KuoInit,
                dim664KuoInit,
                dim665KuoInit,
                dim666KuoInit,
                dim667KuoInit,
                dim668KuoInit,
                dim669KuoInit,
                dim670KuoInit,
                dim671KuoInit,
                dim672KuoInit,
                dim673KuoInit,
                dim674KuoInit,
                dim675KuoInit,
                dim676KuoInit,
                dim677KuoInit,
                dim678KuoInit,
                dim679KuoInit,
                dim680KuoInit,
                dim681KuoInit,
                dim682KuoInit,
                dim683KuoInit,
                dim684KuoInit,
                dim685KuoInit,
                dim686KuoInit,
                dim687KuoInit,
                dim688KuoInit,
                dim689KuoInit,
                dim690KuoInit,
                dim691KuoInit,
                dim692KuoInit,
                dim693KuoInit,
                dim694KuoInit,
                dim695KuoInit,
                dim696KuoInit,
                dim697KuoInit,
                dim698KuoInit,
                dim699KuoInit,
                dim700KuoInit,
                dim701KuoInit,
                dim702KuoInit,
                dim703KuoInit,
                dim704KuoInit,
                dim705KuoInit,
                dim706KuoInit,
                dim707KuoInit,
                dim708KuoInit,
                dim709KuoInit,
                dim710KuoInit,
                dim711KuoInit,
                dim712KuoInit,
                dim713KuoInit,
                dim714KuoInit,
                dim715KuoInit,
                dim716KuoInit,
                dim717KuoInit,
                dim718KuoInit,
                dim719KuoInit,
                dim720KuoInit,
                dim721KuoInit,
                dim722KuoInit,
                dim723KuoInit,
                dim724KuoInit,
                dim725KuoInit,
                dim726KuoInit,
                dim727KuoInit,
                dim728KuoInit,
                dim729KuoInit,
                dim730KuoInit,
                dim731KuoInit,
                dim732KuoInit,
                dim733KuoInit,
                dim734KuoInit,
                dim735KuoInit,
                dim736KuoInit,
                dim737KuoInit,
                dim738KuoInit,
                dim739KuoInit,
                dim740KuoInit,
                dim741KuoInit,
                dim742KuoInit,
                dim743KuoInit,
                dim744KuoInit,
                dim745KuoInit,
                dim746KuoInit,
                dim747KuoInit,
                dim748KuoInit,
                dim749KuoInit,
                dim750KuoInit,
                dim751KuoInit,
                dim752KuoInit,
                dim753KuoInit,
                dim754KuoInit,
                dim755KuoInit,
                dim756KuoInit,
                dim757KuoInit,
                dim758KuoInit,
                dim759KuoInit,
                dim760KuoInit,
                dim761KuoInit,
                dim762KuoInit,
                dim763KuoInit,
                dim764KuoInit,
                dim765KuoInit,
                dim766KuoInit,
                dim767KuoInit,
                dim768KuoInit,
                dim769KuoInit,
                dim770KuoInit,
                dim771KuoInit,
                dim772KuoInit,
                dim773KuoInit,
                dim774KuoInit,
                dim775KuoInit,
                dim776KuoInit,
                dim777KuoInit,
                dim778KuoInit,
                dim779KuoInit,
                dim780KuoInit,
                dim781KuoInit,
                dim782KuoInit,
                dim783KuoInit,
                dim784KuoInit,
                dim785KuoInit,
                dim786KuoInit,
                dim787KuoInit,
                dim788KuoInit,
                dim789KuoInit,
                dim790KuoInit,
                dim791KuoInit,
                dim792KuoInit,
                dim793KuoInit,
                dim794KuoInit,
                dim795KuoInit,
                dim796KuoInit,
                dim797KuoInit,
                dim798KuoInit,
                dim799KuoInit,
                dim800KuoInit,
                dim801KuoInit,
                dim802KuoInit,
                dim803KuoInit,
                dim804KuoInit,
                dim805KuoInit,
                dim806KuoInit,
                dim807KuoInit,
                dim808KuoInit,
                dim809KuoInit,
                dim810KuoInit,
                dim811KuoInit,
                dim812KuoInit,
                dim813KuoInit,
                dim814KuoInit,
                dim815KuoInit,
                dim816KuoInit,
                dim817KuoInit,
                dim818KuoInit,
                dim819KuoInit,
                dim820KuoInit,
                dim821KuoInit,
                dim822KuoInit,
                dim823KuoInit,
                dim824KuoInit,
                dim825KuoInit,
                dim826KuoInit,
                dim827KuoInit,
                dim828KuoInit,
                dim829KuoInit,
                dim830KuoInit,
                dim831KuoInit,
                dim832KuoInit,
                dim833KuoInit,
                dim834KuoInit,
                dim835KuoInit,
                dim836KuoInit,
                dim837KuoInit,
                dim838KuoInit,
                dim839KuoInit,
                dim840KuoInit,
                dim841KuoInit,
                dim842KuoInit,
                dim843KuoInit,
                dim844KuoInit,
                dim845KuoInit,
                dim846KuoInit,
                dim847KuoInit,
                dim848KuoInit,
                dim849KuoInit,
                dim850KuoInit,
                dim851KuoInit,
                dim852KuoInit,
                dim853KuoInit,
                dim854KuoInit,
                dim855KuoInit,
                dim856KuoInit,
                dim857KuoInit,
                dim858KuoInit,
                dim859KuoInit,
                dim860KuoInit,
                dim861KuoInit,
                dim862KuoInit,
                dim863KuoInit,
                dim864KuoInit,
                dim865KuoInit,
                dim866KuoInit,
                dim867KuoInit,
                dim868KuoInit,
                dim869KuoInit,
                dim870KuoInit,
                dim871KuoInit,
                dim872KuoInit,
                dim873KuoInit,
                dim874KuoInit,
                dim875KuoInit,
                dim876KuoInit,
                dim877KuoInit,
                dim878KuoInit,
                dim879KuoInit,
                dim880KuoInit,
                dim881KuoInit,
                dim882KuoInit,
                dim883KuoInit,
                dim884KuoInit,
                dim885KuoInit,
                dim886KuoInit,
                dim887KuoInit,
                dim888KuoInit,
                dim889KuoInit,
                dim890KuoInit,
                dim891KuoInit,
                dim892KuoInit,
                dim893KuoInit,
                dim894KuoInit,
                dim895KuoInit,
                dim896KuoInit,
                dim897KuoInit,
                dim898KuoInit,
                dim899KuoInit,
                dim900KuoInit,
                dim901KuoInit,
                dim902KuoInit,
                dim903KuoInit,
                dim904KuoInit,
                dim905KuoInit,
                dim906KuoInit,
                dim907KuoInit,
                dim908KuoInit,
                dim909KuoInit,
                dim910KuoInit,
                dim911KuoInit,
                dim912KuoInit,
                dim913KuoInit,
                dim914KuoInit,
                dim915KuoInit,
                dim916KuoInit,
                dim917KuoInit,
                dim918KuoInit,
                dim919KuoInit,
                dim920KuoInit,
                dim921KuoInit,
                dim922KuoInit,
                dim923KuoInit,
                dim924KuoInit,
                dim925KuoInit,
                dim926KuoInit,
                dim927KuoInit,
                dim928KuoInit,
                dim929KuoInit,
                dim930KuoInit,
                dim931KuoInit,
                dim932KuoInit,
                dim933KuoInit,
                dim934KuoInit,
                dim935KuoInit,
                dim936KuoInit,
                dim937KuoInit,
                dim938KuoInit,
                dim939KuoInit,
                dim940KuoInit,
                dim941KuoInit,
                dim942KuoInit,
                dim943KuoInit,
                dim944KuoInit,
                dim945KuoInit,
                dim946KuoInit,
                dim947KuoInit,
                dim948KuoInit,
                dim949KuoInit,
                dim950KuoInit,
                dim951KuoInit,
                dim952KuoInit,
                dim953KuoInit,
                dim954KuoInit,
                dim955KuoInit,
                dim956KuoInit,
                dim957KuoInit,
                dim958KuoInit,
                dim959KuoInit,
                dim960KuoInit,
                dim961KuoInit,
                dim962KuoInit,
                dim963KuoInit,
                dim964KuoInit,
                dim965KuoInit,
                dim966KuoInit,
                dim967KuoInit,
                dim968KuoInit,
                dim969KuoInit,
                dim970KuoInit,
                dim971KuoInit,
                dim972KuoInit,
                dim973KuoInit,
                dim974KuoInit,
                dim975KuoInit,
                dim976KuoInit,
                dim977KuoInit,
                dim978KuoInit,
                dim979KuoInit,
                dim980KuoInit,
                dim981KuoInit,
                dim982KuoInit,
                dim983KuoInit,
                dim984KuoInit,
                dim985KuoInit,
                dim986KuoInit,
                dim987KuoInit,
                dim988KuoInit,
                dim989KuoInit,
                dim990KuoInit,
                dim991KuoInit,
                dim992KuoInit,
                dim993KuoInit,
                dim994KuoInit,
                dim995KuoInit,
                dim996KuoInit,
                dim997KuoInit,
                dim998KuoInit,
                dim999KuoInit,
                dim1000KuoInit,
                dim1001KuoInit,
                dim1002KuoInit,
                dim1003KuoInit,
                dim1004KuoInit,
                dim1005KuoInit,
                dim1006KuoInit,
                dim1007KuoInit,
                dim1008KuoInit,
                dim1009KuoInit,
                dim1010KuoInit,
                dim1011KuoInit,
                dim1012KuoInit,
                dim1013KuoInit,
                dim1014KuoInit,
                dim1015KuoInit,
                dim1016KuoInit,
                dim1017KuoInit,
                dim1018KuoInit,
                dim1019KuoInit,
                dim1020KuoInit,
                dim1021KuoInit,
                dim1022KuoInit,
                dim1023KuoInit,
                dim1024KuoInit,
                dim1025KuoInit,
                dim1026KuoInit,
                dim1027KuoInit,
                dim1028KuoInit,
                dim1029KuoInit,
                dim1030KuoInit,
                dim1031KuoInit,
                dim1032KuoInit,
                dim1033KuoInit,
                dim1034KuoInit,
                dim1035KuoInit,
                dim1036KuoInit,
                dim1037KuoInit,
                dim1038KuoInit,
                dim1039KuoInit,
                dim1040KuoInit,
                dim1041KuoInit,
                dim1042KuoInit,
                dim1043KuoInit,
                dim1044KuoInit,
                dim1045KuoInit,
                dim1046KuoInit,
                dim1047KuoInit,
                dim1048KuoInit,
                dim1049KuoInit,
                dim1050KuoInit,
                dim1051KuoInit,
                dim1052KuoInit,
                dim1053KuoInit,
                dim1054KuoInit,
                dim1055KuoInit,
                dim1056KuoInit,
                dim1057KuoInit,
                dim1058KuoInit,
                dim1059KuoInit,
                dim1060KuoInit,
                dim1061KuoInit,
                dim1062KuoInit,
                dim1063KuoInit,
                dim1064KuoInit,
                dim1065KuoInit,
                dim1066KuoInit,
                dim1067KuoInit,
                dim1068KuoInit,
                dim1069KuoInit,
                dim1070KuoInit,
                dim1071KuoInit,
                dim1072KuoInit,
                dim1073KuoInit,
                dim1074KuoInit,
                dim1075KuoInit,
                dim1076KuoInit,
                dim1077KuoInit,
                dim1078KuoInit,
                dim1079KuoInit,
                dim1080KuoInit,
                dim1081KuoInit,
                dim1082KuoInit,
                dim1083KuoInit,
                dim1084KuoInit,
                dim1085KuoInit,
                dim1086KuoInit,
                dim1087KuoInit,
                dim1088KuoInit,
                dim1089KuoInit,
                dim1090KuoInit,
                dim1091KuoInit,
                dim1092KuoInit,
                dim1093KuoInit,
                dim1094KuoInit,
                dim1095KuoInit,
                dim1096KuoInit,
                dim1097KuoInit,
                dim1098KuoInit,
                dim1099KuoInit,
                dim1100KuoInit,
                dim1101KuoInit,
                dim1102KuoInit,
                dim1103KuoInit,
                dim1104KuoInit,
                dim1105KuoInit,
                dim1106KuoInit,
                dim1107KuoInit,
                dim1108KuoInit,
                dim1109KuoInit,
                dim1110KuoInit,
                dim1111KuoInit,
                dim1112KuoInit,
                dim1113KuoInit,
                dim1114KuoInit,
                dim1115KuoInit,
                dim1116KuoInit,
                dim1117KuoInit,
                dim1118KuoInit,
                dim1119KuoInit,
                dim1120KuoInit,
                dim1121KuoInit,
                dim1122KuoInit,
                dim1123KuoInit,
                dim1124KuoInit,
                dim1125KuoInit,
                dim1126KuoInit,
                dim1127KuoInit,
                dim1128KuoInit,
                dim1129KuoInit,
                dim1130KuoInit,
                dim1131KuoInit,
                dim1132KuoInit,
                dim1133KuoInit,
                dim1134KuoInit,
                dim1135KuoInit,
                dim1136KuoInit,
                dim1137KuoInit,
                dim1138KuoInit,
                dim1139KuoInit,
                dim1140KuoInit,
                dim1141KuoInit,
                dim1142KuoInit,
                dim1143KuoInit,
                dim1144KuoInit,
                dim1145KuoInit,
                dim1146KuoInit,
                dim1147KuoInit,
                dim1148KuoInit,
                dim1149KuoInit,
                dim1150KuoInit,
                dim1151KuoInit,
                dim1152KuoInit,
                dim1153KuoInit,
                dim1154KuoInit,
                dim1155KuoInit,
                dim1156KuoInit,
                dim1157KuoInit,
                dim1158KuoInit,
                dim1159KuoInit,
                dim1160KuoInit,
                dim1161KuoInit,
                dim1162KuoInit,
                dim1163KuoInit,
                dim1164KuoInit,
                dim1165KuoInit,
                dim1166KuoInit,
                dim1167KuoInit,
                dim1168KuoInit,
                dim1169KuoInit,
                dim1170KuoInit,
                dim1171KuoInit,
                dim1172KuoInit,
                dim1173KuoInit,
                dim1174KuoInit,
                dim1175KuoInit,
                dim1176KuoInit,
                dim1177KuoInit,
                dim1178KuoInit,
                dim1179KuoInit,
                dim1180KuoInit,
                dim1181KuoInit,
                dim1182KuoInit,
                dim1183KuoInit,
                dim1184KuoInit,
                dim1185KuoInit,
                dim1186KuoInit,
                dim1187KuoInit,
                dim1188KuoInit,
                dim1189KuoInit,
                dim1190KuoInit,
                dim1191KuoInit,
                dim1192KuoInit,
                dim1193KuoInit,
                dim1194KuoInit,
                dim1195KuoInit,
                dim1196KuoInit,
                dim1197KuoInit,
                dim1198KuoInit,
                dim1199KuoInit,
                dim1200KuoInit,
                dim1201KuoInit,
                dim1202KuoInit,
                dim1203KuoInit,
                dim1204KuoInit,
                dim1205KuoInit,
                dim1206KuoInit,
                dim1207KuoInit,
                dim1208KuoInit,
                dim1209KuoInit,
                dim1210KuoInit,
                dim1211KuoInit,
                dim1212KuoInit,
                dim1213KuoInit,
                dim1214KuoInit,
                dim1215KuoInit,
                dim1216KuoInit,
                dim1217KuoInit,
                dim1218KuoInit,
                dim1219KuoInit,
                dim1220KuoInit,
                dim1221KuoInit,
                dim1222KuoInit,
                dim1223KuoInit,
                dim1224KuoInit,
                dim1225KuoInit,
                dim1226KuoInit,
                dim1227KuoInit,
                dim1228KuoInit,
                dim1229KuoInit,
                dim1230KuoInit,
                dim1231KuoInit,
                dim1232KuoInit,
                dim1233KuoInit,
                dim1234KuoInit,
                dim1235KuoInit,
                dim1236KuoInit,
                dim1237KuoInit,
                dim1238KuoInit,
                dim1239KuoInit,
                dim1240KuoInit,
                dim1241KuoInit,
                dim1242KuoInit,
                dim1243KuoInit,
                dim1244KuoInit,
                dim1245KuoInit,
                dim1246KuoInit,
                dim1247KuoInit,
                dim1248KuoInit,
                dim1249KuoInit,
                dim1250KuoInit,
                dim1251KuoInit,
                dim1252KuoInit,
                dim1253KuoInit,
                dim1254KuoInit,
                dim1255KuoInit,
                dim1256KuoInit,
                dim1257KuoInit,
                dim1258KuoInit,
                dim1259KuoInit,
                dim1260KuoInit,
                dim1261KuoInit,
                dim1262KuoInit,
                dim1263KuoInit,
                dim1264KuoInit,
                dim1265KuoInit,
                dim1266KuoInit,
                dim1267KuoInit,
                dim1268KuoInit,
                dim1269KuoInit,
                dim1270KuoInit,
                dim1271KuoInit,
                dim1272KuoInit,
                dim1273KuoInit,
                dim1274KuoInit,
                dim1275KuoInit,
                dim1276KuoInit,
                dim1277KuoInit,
                dim1278KuoInit,
                dim1279KuoInit,
                dim1280KuoInit,
                dim1281KuoInit,
                dim1282KuoInit,
                dim1283KuoInit,
                dim1284KuoInit,
                dim1285KuoInit,
                dim1286KuoInit,
                dim1287KuoInit,
                dim1288KuoInit,
                dim1289KuoInit,
                dim1290KuoInit,
                dim1291KuoInit,
                dim1292KuoInit,
                dim1293KuoInit,
                dim1294KuoInit,
                dim1295KuoInit,
                dim1296KuoInit,
                dim1297KuoInit,
                dim1298KuoInit,
                dim1299KuoInit,
                dim1300KuoInit,
                dim1301KuoInit,
                dim1302KuoInit,
                dim1303KuoInit,
                dim1304KuoInit,
                dim1305KuoInit,
                dim1306KuoInit,
                dim1307KuoInit,
                dim1308KuoInit,
                dim1309KuoInit,
                dim1310KuoInit,
                dim1311KuoInit,
                dim1312KuoInit,
                dim1313KuoInit,
                dim1314KuoInit,
                dim1315KuoInit,
                dim1316KuoInit,
                dim1317KuoInit,
                dim1318KuoInit,
                dim1319KuoInit,
                dim1320KuoInit,
                dim1321KuoInit,
                dim1322KuoInit,
                dim1323KuoInit,
                dim1324KuoInit,
                dim1325KuoInit,
                dim1326KuoInit,
                dim1327KuoInit,
                dim1328KuoInit,
                dim1329KuoInit,
                dim1330KuoInit,
                dim1331KuoInit,
                dim1332KuoInit,
                dim1333KuoInit,
                dim1334KuoInit,
                dim1335KuoInit,
                dim1336KuoInit,
                dim1337KuoInit,
                dim1338KuoInit,
                dim1339KuoInit,
                dim1340KuoInit,
                dim1341KuoInit,
                dim1342KuoInit,
                dim1343KuoInit,
                dim1344KuoInit,
                dim1345KuoInit,
                dim1346KuoInit,
                dim1347KuoInit,
                dim1348KuoInit,
                dim1349KuoInit,
                dim1350KuoInit,
                dim1351KuoInit,
                dim1352KuoInit,
                dim1353KuoInit,
                dim1354KuoInit,
                dim1355KuoInit,
                dim1356KuoInit,
                dim1357KuoInit,
                dim1358KuoInit,
                dim1359KuoInit,
                dim1360KuoInit,
                dim1361KuoInit,
                dim1362KuoInit,
                dim1363KuoInit,
                dim1364KuoInit,
                dim1365KuoInit,
                dim1366KuoInit,
                dim1367KuoInit,
                dim1368KuoInit,
                dim1369KuoInit,
                dim1370KuoInit,
                dim1371KuoInit,
                dim1372KuoInit,
                dim1373KuoInit,
                dim1374KuoInit,
                dim1375KuoInit,
                dim1376KuoInit,
                dim1377KuoInit,
                dim1378KuoInit,
                dim1379KuoInit,
                dim1380KuoInit,
                dim1381KuoInit,
                dim1382KuoInit,
                dim1383KuoInit,
                dim1384KuoInit,
                dim1385KuoInit,
                dim1386KuoInit,
                dim1387KuoInit,
                dim1388KuoInit,
                dim1389KuoInit,
                dim1390KuoInit,
                dim1391KuoInit,
                dim1392KuoInit,
                dim1393KuoInit,
                dim1394KuoInit,
                dim1395KuoInit,
                dim1396KuoInit,
                dim1397KuoInit,
                dim1398KuoInit,
                dim1399KuoInit,
                dim1400KuoInit,
                dim1401KuoInit,
                dim1402KuoInit,
                dim1403KuoInit,
                dim1404KuoInit,
                dim1405KuoInit,
                dim1406KuoInit,
                dim1407KuoInit,
                dim1408KuoInit,
                dim1409KuoInit,
                dim1410KuoInit,
                dim1411KuoInit,
                dim1412KuoInit,
                dim1413KuoInit,
                dim1414KuoInit,
                dim1415KuoInit,
                dim1416KuoInit,
                dim1417KuoInit,
                dim1418KuoInit,
                dim1419KuoInit,
                dim1420KuoInit,
                dim1421KuoInit,
                dim1422KuoInit,
                dim1423KuoInit,
                dim1424KuoInit,
                dim1425KuoInit,
                dim1426KuoInit,
                dim1427KuoInit,
                dim1428KuoInit,
                dim1429KuoInit,
                dim1430KuoInit,
                dim1431KuoInit,
                dim1432KuoInit,
                dim1433KuoInit,
                dim1434KuoInit,
                dim1435KuoInit,
                dim1436KuoInit,
                dim1437KuoInit,
                dim1438KuoInit,
                dim1439KuoInit,
                dim1440KuoInit,
                dim1441KuoInit,
                dim1442KuoInit,
                dim1443KuoInit,
                dim1444KuoInit,
                dim1445KuoInit,
                dim1446KuoInit,
                dim1447KuoInit,
                dim1448KuoInit,
                dim1449KuoInit,
                dim1450KuoInit,
                dim1451KuoInit,
                dim1452KuoInit,
                dim1453KuoInit,
                dim1454KuoInit,
                dim1455KuoInit,
                dim1456KuoInit,
                dim1457KuoInit,
                dim1458KuoInit,
                dim1459KuoInit,
                dim1460KuoInit,
                dim1461KuoInit,
                dim1462KuoInit,
                dim1463KuoInit,
                dim1464KuoInit,
                dim1465KuoInit,
                dim1466KuoInit,
                dim1467KuoInit,
                dim1468KuoInit,
                dim1469KuoInit,
                dim1470KuoInit,
                dim1471KuoInit,
                dim1472KuoInit,
                dim1473KuoInit,
                dim1474KuoInit,
                dim1475KuoInit,
                dim1476KuoInit,
                dim1477KuoInit,
                dim1478KuoInit,
                dim1479KuoInit,
                dim1480KuoInit,
                dim1481KuoInit,
                dim1482KuoInit,
                dim1483KuoInit,
                dim1484KuoInit,
                dim1485KuoInit,
                dim1486KuoInit,
                dim1487KuoInit,
                dim1488KuoInit,
                dim1489KuoInit,
                dim1490KuoInit,
                dim1491KuoInit,
                dim1492KuoInit,
                dim1493KuoInit,
                dim1494KuoInit,
                dim1495KuoInit,
                dim1496KuoInit,
                dim1497KuoInit,
                dim1498KuoInit,
                dim1499KuoInit,
                dim1500KuoInit,
                dim1501KuoInit,
                dim1502KuoInit,
                dim1503KuoInit,
                dim1504KuoInit,
                dim1505KuoInit,
                dim1506KuoInit,
                dim1507KuoInit,
                dim1508KuoInit,
                dim1509KuoInit,
                dim1510KuoInit,
                dim1511KuoInit,
                dim1512KuoInit,
                dim1513KuoInit,
                dim1514KuoInit,
                dim1515KuoInit,
                dim1516KuoInit,
                dim1517KuoInit,
                dim1518KuoInit,
                dim1519KuoInit,
                dim1520KuoInit,
                dim1521KuoInit,
                dim1522KuoInit,
                dim1523KuoInit,
                dim1524KuoInit,
                dim1525KuoInit,
                dim1526KuoInit,
                dim1527KuoInit,
                dim1528KuoInit,
                dim1529KuoInit,
                dim1530KuoInit,
                dim1531KuoInit,
                dim1532KuoInit,
                dim1533KuoInit,
                dim1534KuoInit,
                dim1535KuoInit,
                dim1536KuoInit,
                dim1537KuoInit,
                dim1538KuoInit,
                dim1539KuoInit,
                dim1540KuoInit,
                dim1541KuoInit,
                dim1542KuoInit,
                dim1543KuoInit,
                dim1544KuoInit,
                dim1545KuoInit,
                dim1546KuoInit,
                dim1547KuoInit,
                dim1548KuoInit,
                dim1549KuoInit,
                dim1550KuoInit,
                dim1551KuoInit,
                dim1552KuoInit,
                dim1553KuoInit,
                dim1554KuoInit,
                dim1555KuoInit,
                dim1556KuoInit,
                dim1557KuoInit,
                dim1558KuoInit,
                dim1559KuoInit,
                dim1560KuoInit,
                dim1561KuoInit,
                dim1562KuoInit,
                dim1563KuoInit,
                dim1564KuoInit,
                dim1565KuoInit,
                dim1566KuoInit,
                dim1567KuoInit,
                dim1568KuoInit,
                dim1569KuoInit,
                dim1570KuoInit,
                dim1571KuoInit,
                dim1572KuoInit,
                dim1573KuoInit,
                dim1574KuoInit,
                dim1575KuoInit,
                dim1576KuoInit,
                dim1577KuoInit,
                dim1578KuoInit,
                dim1579KuoInit,
                dim1580KuoInit,
                dim1581KuoInit,
                dim1582KuoInit,
                dim1583KuoInit,
                dim1584KuoInit,
                dim1585KuoInit,
                dim1586KuoInit,
                dim1587KuoInit,
                dim1588KuoInit,
                dim1589KuoInit,
                dim1590KuoInit,
                dim1591KuoInit,
                dim1592KuoInit,
                dim1593KuoInit,
                dim1594KuoInit,
                dim1595KuoInit,
                dim1596KuoInit,
                dim1597KuoInit,
                dim1598KuoInit,
                dim1599KuoInit,
                dim1600KuoInit,
                dim1601KuoInit,
                dim1602KuoInit,
                dim1603KuoInit,
                dim1604KuoInit,
                dim1605KuoInit,
                dim1606KuoInit,
                dim1607KuoInit,
                dim1608KuoInit,
                dim1609KuoInit,
                dim1610KuoInit,
                dim1611KuoInit,
                dim1612KuoInit,
                dim1613KuoInit,
                dim1614KuoInit,
                dim1615KuoInit,
                dim1616KuoInit,
                dim1617KuoInit,
                dim1618KuoInit,
                dim1619KuoInit,
                dim1620KuoInit,
                dim1621KuoInit,
                dim1622KuoInit,
                dim1623KuoInit,
                dim1624KuoInit,
                dim1625KuoInit,
                dim1626KuoInit,
                dim1627KuoInit,
                dim1628KuoInit,
                dim1629KuoInit,
                dim1630KuoInit,
                dim1631KuoInit,
                dim1632KuoInit,
                dim1633KuoInit,
                dim1634KuoInit,
                dim1635KuoInit,
                dim1636KuoInit,
                dim1637KuoInit,
                dim1638KuoInit,
                dim1639KuoInit,
                dim1640KuoInit,
                dim1641KuoInit,
                dim1642KuoInit,
                dim1643KuoInit,
                dim1644KuoInit,
                dim1645KuoInit,
                dim1646KuoInit,
                dim1647KuoInit,
                dim1648KuoInit,
                dim1649KuoInit,
                dim1650KuoInit,
                dim1651KuoInit,
                dim1652KuoInit,
                dim1653KuoInit,
                dim1654KuoInit,
                dim1655KuoInit,
                dim1656KuoInit,
                dim1657KuoInit,
                dim1658KuoInit,
                dim1659KuoInit,
                dim1660KuoInit,
                dim1661KuoInit,
                dim1662KuoInit,
                dim1663KuoInit,
                dim1664KuoInit,
                dim1665KuoInit,
                dim1666KuoInit,
                dim1667KuoInit,
                dim1668KuoInit,
                dim1669KuoInit,
                dim1670KuoInit,
                dim1671KuoInit,
                dim1672KuoInit,
                dim1673KuoInit,
                dim1674KuoInit,
                dim1675KuoInit,
                dim1676KuoInit,
                dim1677KuoInit,
                dim1678KuoInit,
                dim1679KuoInit,
                dim1680KuoInit,
                dim1681KuoInit,
                dim1682KuoInit,
                dim1683KuoInit,
                dim1684KuoInit,
                dim1685KuoInit,
                dim1686KuoInit,
                dim1687KuoInit,
                dim1688KuoInit,
                dim1689KuoInit,
                dim1690KuoInit,
                dim1691KuoInit,
                dim1692KuoInit,
                dim1693KuoInit,
                dim1694KuoInit,
                dim1695KuoInit,
                dim1696KuoInit,
                dim1697KuoInit,
                dim1698KuoInit,
                dim1699KuoInit,
                dim1700KuoInit,
                dim1701KuoInit,
                dim1702KuoInit,
                dim1703KuoInit,
                dim1704KuoInit,
                dim1705KuoInit,
                dim1706KuoInit,
                dim1707KuoInit,
                dim1708KuoInit,
                dim1709KuoInit,
                dim1710KuoInit,
                dim1711KuoInit,
                dim1712KuoInit,
                dim1713KuoInit,
                dim1714KuoInit,
                dim1715KuoInit,
                dim1716KuoInit,
                dim1717KuoInit,
                dim1718KuoInit,
                dim1719KuoInit,
                dim1720KuoInit,
                dim1721KuoInit,
                dim1722KuoInit,
                dim1723KuoInit,
                dim1724KuoInit,
                dim1725KuoInit,
                dim1726KuoInit,
                dim1727KuoInit,
                dim1728KuoInit,
                dim1729KuoInit,
                dim1730KuoInit,
                dim1731KuoInit,
                dim1732KuoInit,
                dim1733KuoInit,
                dim1734KuoInit,
                dim1735KuoInit,
                dim1736KuoInit,
                dim1737KuoInit,
                dim1738KuoInit,
                dim1739KuoInit,
                dim1740KuoInit,
                dim1741KuoInit,
                dim1742KuoInit,
                dim1743KuoInit,
                dim1744KuoInit,
                dim1745KuoInit,
                dim1746KuoInit,
                dim1747KuoInit,
                dim1748KuoInit,
                dim1749KuoInit,
                dim1750KuoInit,
                dim1751KuoInit,
                dim1752KuoInit,
                dim1753KuoInit,
                dim1754KuoInit,
                dim1755KuoInit,
                dim1756KuoInit,
                dim1757KuoInit,
                dim1758KuoInit,
                dim1759KuoInit,
                dim1760KuoInit,
                dim1761KuoInit,
                dim1762KuoInit,
                dim1763KuoInit,
                dim1764KuoInit,
                dim1765KuoInit,
                dim1766KuoInit,
                dim1767KuoInit,
                dim1768KuoInit,
                dim1769KuoInit,
                dim1770KuoInit,
                dim1771KuoInit,
                dim1772KuoInit,
                dim1773KuoInit,
                dim1774KuoInit,
                dim1775KuoInit,
                dim1776KuoInit,
                dim1777KuoInit,
                dim1778KuoInit,
                dim1779KuoInit,
                dim1780KuoInit,
                dim1781KuoInit,
                dim1782KuoInit,
                dim1783KuoInit,
                dim1784KuoInit,
                dim1785KuoInit,
                dim1786KuoInit,
                dim1787KuoInit,
                dim1788KuoInit,
                dim1789KuoInit,
                dim1790KuoInit,
                dim1791KuoInit,
                dim1792KuoInit,
                dim1793KuoInit,
                dim1794KuoInit,
                dim1795KuoInit,
                dim1796KuoInit,
                dim1797KuoInit,
                dim1798KuoInit,
                dim1799KuoInit,
                dim1800KuoInit,
                dim1801KuoInit,
                dim1802KuoInit,
                dim1803KuoInit,
                dim1804KuoInit,
                dim1805KuoInit,
                dim1806KuoInit,
                dim1807KuoInit,
                dim1808KuoInit,
                dim1809KuoInit,
                dim1810KuoInit,
                dim1811KuoInit,
                dim1812KuoInit,
                dim1813KuoInit,
                dim1814KuoInit,
                dim1815KuoInit,
                dim1816KuoInit,
                dim1817KuoInit,
                dim1818KuoInit,
                dim1819KuoInit,
                dim1820KuoInit,
                dim1821KuoInit,
                dim1822KuoInit,
                dim1823KuoInit,
                dim1824KuoInit,
                dim1825KuoInit,
                dim1826KuoInit,
                dim1827KuoInit,
                dim1828KuoInit,
                dim1829KuoInit,
                dim1830KuoInit,
                dim1831KuoInit,
                dim1832KuoInit,
                dim1833KuoInit,
                dim1834KuoInit,
                dim1835KuoInit,
                dim1836KuoInit,
                dim1837KuoInit,
                dim1838KuoInit,
                dim1839KuoInit,
                dim1840KuoInit,
                dim1841KuoInit,
                dim1842KuoInit,
                dim1843KuoInit,
                dim1844KuoInit,
                dim1845KuoInit,
                dim1846KuoInit,
                dim1847KuoInit,
                dim1848KuoInit,
                dim1849KuoInit,
                dim1850KuoInit,
                dim1851KuoInit,
                dim1852KuoInit,
                dim1853KuoInit,
                dim1854KuoInit,
                dim1855KuoInit,
                dim1856KuoInit,
                dim1857KuoInit,
                dim1858KuoInit,
                dim1859KuoInit,
                dim1860KuoInit,
                dim1861KuoInit,
                dim1862KuoInit,
                dim1863KuoInit,
                dim1864KuoInit,
                dim1865KuoInit,
                dim1866KuoInit,
                dim1867KuoInit,
                dim1868KuoInit,
                dim1869KuoInit,
                dim1870KuoInit,
                dim1871KuoInit,
                dim1872KuoInit,
                dim1873KuoInit,
                dim1874KuoInit,
                dim1875KuoInit,
                dim1876KuoInit,
                dim1877KuoInit,
                dim1878KuoInit,
                dim1879KuoInit,
                dim1880KuoInit,
                dim1881KuoInit,
                dim1882KuoInit,
                dim1883KuoInit,
                dim1884KuoInit,
                dim1885KuoInit,
                dim1886KuoInit,
                dim1887KuoInit,
                dim1888KuoInit,
                dim1889KuoInit,
                dim1890KuoInit,
                dim1891KuoInit,
                dim1892KuoInit,
                dim1893KuoInit,
                dim1894KuoInit,
                dim1895KuoInit,
                dim1896KuoInit,
                dim1897KuoInit,
                dim1898KuoInit,
                dim1899KuoInit,
                dim1900KuoInit,
                dim1901KuoInit,
                dim1902KuoInit,
                dim1903KuoInit,
                dim1904KuoInit,
                dim1905KuoInit,
                dim1906KuoInit,
                dim1907KuoInit,
                dim1908KuoInit,
                dim1909KuoInit,
                dim1910KuoInit,
                dim1911KuoInit,
                dim1912KuoInit,
                dim1913KuoInit,
                dim1914KuoInit,
                dim1915KuoInit,
                dim1916KuoInit,
                dim1917KuoInit,
                dim1918KuoInit,
                dim1919KuoInit,
                dim1920KuoInit,
                dim1921KuoInit,
                dim1922KuoInit,
                dim1923KuoInit,
                dim1924KuoInit,
                dim1925KuoInit,
                dim1926KuoInit,
                dim1927KuoInit,
                dim1928KuoInit,
                dim1929KuoInit,
                dim1930KuoInit,
                dim1931KuoInit,
                dim1932KuoInit,
                dim1933KuoInit,
                dim1934KuoInit,
                dim1935KuoInit,
                dim1936KuoInit,
                dim1937KuoInit,
                dim1938KuoInit,
                dim1939KuoInit,
                dim1940KuoInit,
                dim1941KuoInit,
                dim1942KuoInit,
                dim1943KuoInit,
                dim1944KuoInit,
                dim1945KuoInit,
                dim1946KuoInit,
                dim1947KuoInit,
                dim1948KuoInit,
                dim1949KuoInit,
                dim1950KuoInit,
                dim1951KuoInit,
                dim1952KuoInit,
                dim1953KuoInit,
                dim1954KuoInit,
                dim1955KuoInit,
                dim1956KuoInit,
                dim1957KuoInit,
                dim1958KuoInit,
                dim1959KuoInit,
                dim1960KuoInit,
                dim1961KuoInit,
                dim1962KuoInit,
                dim1963KuoInit,
                dim1964KuoInit,
                dim1965KuoInit,
                dim1966KuoInit,
                dim1967KuoInit,
                dim1968KuoInit,
                dim1969KuoInit,
                dim1970KuoInit,
                dim1971KuoInit,
                dim1972KuoInit,
                dim1973KuoInit,
                dim1974KuoInit,
                dim1975KuoInit,
                dim1976KuoInit,
                dim1977KuoInit,
                dim1978KuoInit,
                dim1979KuoInit,
                dim1980KuoInit,
                dim1981KuoInit,
                dim1982KuoInit,
                dim1983KuoInit,
                dim1984KuoInit,
                dim1985KuoInit,
                dim1986KuoInit,
                dim1987KuoInit,
                dim1988KuoInit,
                dim1989KuoInit,
                dim1990KuoInit,
                dim1991KuoInit,
                dim1992KuoInit,
                dim1993KuoInit,
                dim1994KuoInit,
                dim1995KuoInit,
                dim1996KuoInit,
                dim1997KuoInit,
                dim1998KuoInit,
                dim1999KuoInit,
                dim2000KuoInit,
                dim2001KuoInit,
                dim2002KuoInit,
                dim2003KuoInit,
                dim2004KuoInit,
                dim2005KuoInit,
                dim2006KuoInit,
                dim2007KuoInit,
                dim2008KuoInit,
                dim2009KuoInit,
                dim2010KuoInit,
                dim2011KuoInit,
                dim2012KuoInit,
                dim2013KuoInit,
                dim2014KuoInit,
                dim2015KuoInit,
                dim2016KuoInit,
                dim2017KuoInit,
                dim2018KuoInit,
                dim2019KuoInit,
                dim2020KuoInit,
                dim2021KuoInit,
                dim2022KuoInit,
                dim2023KuoInit,
                dim2024KuoInit,
                dim2025KuoInit,
                dim2026KuoInit,
                dim2027KuoInit,
                dim2028KuoInit,
                dim2029KuoInit,
                dim2030KuoInit,
                dim2031KuoInit,
                dim2032KuoInit,
                dim2033KuoInit,
                dim2034KuoInit,
                dim2035KuoInit,
                dim2036KuoInit,
                dim2037KuoInit,
                dim2038KuoInit,
                dim2039KuoInit,
                dim2040KuoInit,
                dim2041KuoInit,
                dim2042KuoInit,
                dim2043KuoInit,
                dim2044KuoInit,
                dim2045KuoInit,
                dim2046KuoInit,
                dim2047KuoInit,
                dim2048KuoInit,
                dim2049KuoInit,
                dim2050KuoInit,
                dim2051KuoInit,
                dim2052KuoInit,
                dim2053KuoInit,
                dim2054KuoInit,
                dim2055KuoInit,
                dim2056KuoInit,
                dim2057KuoInit,
                dim2058KuoInit,
                dim2059KuoInit,
                dim2060KuoInit,
                dim2061KuoInit,
                dim2062KuoInit,
                dim2063KuoInit,
                dim2064KuoInit,
                dim2065KuoInit,
                dim2066KuoInit,
                dim2067KuoInit,
                dim2068KuoInit,
                dim2069KuoInit,
                dim2070KuoInit,
                dim2071KuoInit,
                dim2072KuoInit,
                dim2073KuoInit,
                dim2074KuoInit,
                dim2075KuoInit,
                dim2076KuoInit,
                dim2077KuoInit,
                dim2078KuoInit,
                dim2079KuoInit,
                dim2080KuoInit,
                dim2081KuoInit,
                dim2082KuoInit,
                dim2083KuoInit,
                dim2084KuoInit,
                dim2085KuoInit,
                dim2086KuoInit,
                dim2087KuoInit,
                dim2088KuoInit,
                dim2089KuoInit,
                dim2090KuoInit,
                dim2091KuoInit,
                dim2092KuoInit,
                dim2093KuoInit,
                dim2094KuoInit,
                dim2095KuoInit,
                dim2096KuoInit,
                dim2097KuoInit,
                dim2098KuoInit,
                dim2099KuoInit,
                dim2100KuoInit,
                dim2101KuoInit,
                dim2102KuoInit,
                dim2103KuoInit,
                dim2104KuoInit,
                dim2105KuoInit,
                dim2106KuoInit,
                dim2107KuoInit,
                dim2108KuoInit,
                dim2109KuoInit,
                dim2110KuoInit,
                dim2111KuoInit,
                dim2112KuoInit,
                dim2113KuoInit,
                dim2114KuoInit,
                dim2115KuoInit,
                dim2116KuoInit,
                dim2117KuoInit,
                dim2118KuoInit,
                dim2119KuoInit,
                dim2120KuoInit,
                dim2121KuoInit,
                dim2122KuoInit,
                dim2123KuoInit,
                dim2124KuoInit,
                dim2125KuoInit,
                dim2126KuoInit,
                dim2127KuoInit,
                dim2128KuoInit,
                dim2129KuoInit,
                dim2130KuoInit,
                dim2131KuoInit,
                dim2132KuoInit,
                dim2133KuoInit,
                dim2134KuoInit,
                dim2135KuoInit,
                dim2136KuoInit,
                dim2137KuoInit,
                dim2138KuoInit,
                dim2139KuoInit,
                dim2140KuoInit,
                dim2141KuoInit,
                dim2142KuoInit,
                dim2143KuoInit,
                dim2144KuoInit,
                dim2145KuoInit,
                dim2146KuoInit,
                dim2147KuoInit,
                dim2148KuoInit,
                dim2149KuoInit,
                dim2150KuoInit,
                dim2151KuoInit,
                dim2152KuoInit,
                dim2153KuoInit,
                dim2154KuoInit,
                dim2155KuoInit,
                dim2156KuoInit,
                dim2157KuoInit,
                dim2158KuoInit,
                dim2159KuoInit,
                dim2160KuoInit,
                dim2161KuoInit,
                dim2162KuoInit,
                dim2163KuoInit,
                dim2164KuoInit,
                dim2165KuoInit,
                dim2166KuoInit,
                dim2167KuoInit,
                dim2168KuoInit,
                dim2169KuoInit,
                dim2170KuoInit,
                dim2171KuoInit,
                dim2172KuoInit,
                dim2173KuoInit,
                dim2174KuoInit,
                dim2175KuoInit,
                dim2176KuoInit,
                dim2177KuoInit,
                dim2178KuoInit,
                dim2179KuoInit,
                dim2180KuoInit,
                dim2181KuoInit,
                dim2182KuoInit,
                dim2183KuoInit,
                dim2184KuoInit,
                dim2185KuoInit,
                dim2186KuoInit,
                dim2187KuoInit,
                dim2188KuoInit,
                dim2189KuoInit,
                dim2190KuoInit,
                dim2191KuoInit,
                dim2192KuoInit,
                dim2193KuoInit,
                dim2194KuoInit,
                dim2195KuoInit,
                dim2196KuoInit,
                dim2197KuoInit,
                dim2198KuoInit,
                dim2199KuoInit,
                dim2200KuoInit,
                dim2201KuoInit,
                dim2202KuoInit,
                dim2203KuoInit,
                dim2204KuoInit,
                dim2205KuoInit,
                dim2206KuoInit,
                dim2207KuoInit,
                dim2208KuoInit,
                dim2209KuoInit,
                dim2210KuoInit,
                dim2211KuoInit,
                dim2212KuoInit,
                dim2213KuoInit,
                dim2214KuoInit,
                dim2215KuoInit,
                dim2216KuoInit,
                dim2217KuoInit,
                dim2218KuoInit,
                dim2219KuoInit,
                dim2220KuoInit,
                dim2221KuoInit,
                dim2222KuoInit,
                dim2223KuoInit,
                dim2224KuoInit,
                dim2225KuoInit,
                dim2226KuoInit,
                dim2227KuoInit,
                dim2228KuoInit,
                dim2229KuoInit,
                dim2230KuoInit,
                dim2231KuoInit,
                dim2232KuoInit,
                dim2233KuoInit,
                dim2234KuoInit,
                dim2235KuoInit,
                dim2236KuoInit,
                dim2237KuoInit,
                dim2238KuoInit,
                dim2239KuoInit,
                dim2240KuoInit,
                dim2241KuoInit,
                dim2242KuoInit,
                dim2243KuoInit,
                dim2244KuoInit,
                dim2245KuoInit,
                dim2246KuoInit,
                dim2247KuoInit,
                dim2248KuoInit,
                dim2249KuoInit,
                dim2250KuoInit,
                dim2251KuoInit,
                dim2252KuoInit,
                dim2253KuoInit,
                dim2254KuoInit,
                dim2255KuoInit,
                dim2256KuoInit,
                dim2257KuoInit,
                dim2258KuoInit,
                dim2259KuoInit,
                dim2260KuoInit,
                dim2261KuoInit,
                dim2262KuoInit,
                dim2263KuoInit,
                dim2264KuoInit,
                dim2265KuoInit,
                dim2266KuoInit,
                dim2267KuoInit,
                dim2268KuoInit,
                dim2269KuoInit,
                dim2270KuoInit,
                dim2271KuoInit,
                dim2272KuoInit,
                dim2273KuoInit,
                dim2274KuoInit,
                dim2275KuoInit,
                dim2276KuoInit,
                dim2277KuoInit,
                dim2278KuoInit,
                dim2279KuoInit,
                dim2280KuoInit,
                dim2281KuoInit,
                dim2282KuoInit,
                dim2283KuoInit,
                dim2284KuoInit,
                dim2285KuoInit,
                dim2286KuoInit,
                dim2287KuoInit,
                dim2288KuoInit,
                dim2289KuoInit,
                dim2290KuoInit,
                dim2291KuoInit,
                dim2292KuoInit,
                dim2293KuoInit,
                dim2294KuoInit,
                dim2295KuoInit,
                dim2296KuoInit,
                dim2297KuoInit,
                dim2298KuoInit,
                dim2299KuoInit,
                dim2300KuoInit,
                dim2301KuoInit,
                dim2302KuoInit,
                dim2303KuoInit,
                dim2304KuoInit,
                dim2305KuoInit,
                dim2306KuoInit,
                dim2307KuoInit,
                dim2308KuoInit,
                dim2309KuoInit,
                dim2310KuoInit,
                dim2311KuoInit,
                dim2312KuoInit,
                dim2313KuoInit,
                dim2314KuoInit,
                dim2315KuoInit,
                dim2316KuoInit,
                dim2317KuoInit,
                dim2318KuoInit,
                dim2319KuoInit,
                dim2320KuoInit,
                dim2321KuoInit,
                dim2322KuoInit,
                dim2323KuoInit,
                dim2324KuoInit,
                dim2325KuoInit,
                dim2326KuoInit,
                dim2327KuoInit,
                dim2328KuoInit,
                dim2329KuoInit,
                dim2330KuoInit,
                dim2331KuoInit,
                dim2332KuoInit,
                dim2333KuoInit,
                dim2334KuoInit,
                dim2335KuoInit,
                dim2336KuoInit,
                dim2337KuoInit,
                dim2338KuoInit,
                dim2339KuoInit,
                dim2340KuoInit,
                dim2341KuoInit,
                dim2342KuoInit,
                dim2343KuoInit,
                dim2344KuoInit,
                dim2345KuoInit,
                dim2346KuoInit,
                dim2347KuoInit,
                dim2348KuoInit,
                dim2349KuoInit,
                dim2350KuoInit,
                dim2351KuoInit,
                dim2352KuoInit,
                dim2353KuoInit,
                dim2354KuoInit,
                dim2355KuoInit,
                dim2356KuoInit,
                dim2357KuoInit,
                dim2358KuoInit,
                dim2359KuoInit,
                dim2360KuoInit,
                dim2361KuoInit,
                dim2362KuoInit,
                dim2363KuoInit,
                dim2364KuoInit,
                dim2365KuoInit,
                dim2366KuoInit,
                dim2367KuoInit,
                dim2368KuoInit,
                dim2369KuoInit,
                dim2370KuoInit,
                dim2371KuoInit,
                dim2372KuoInit,
                dim2373KuoInit,
                dim2374KuoInit,
                dim2375KuoInit,
                dim2376KuoInit,
                dim2377KuoInit,
                dim2378KuoInit,
                dim2379KuoInit,
                dim2380KuoInit,
                dim2381KuoInit,
                dim2382KuoInit,
                dim2383KuoInit,
                dim2384KuoInit,
                dim2385KuoInit,
                dim2386KuoInit,
                dim2387KuoInit,
                dim2388KuoInit,
                dim2389KuoInit,
                dim2390KuoInit,
                dim2391KuoInit,
                dim2392KuoInit,
                dim2393KuoInit,
                dim2394KuoInit,
                dim2395KuoInit,
                dim2396KuoInit,
                dim2397KuoInit,
                dim2398KuoInit,
                dim2399KuoInit,
                dim2400KuoInit,
                dim2401KuoInit,
                dim2402KuoInit,
                dim2403KuoInit,
                dim2404KuoInit,
                dim2405KuoInit,
                dim2406KuoInit,
                dim2407KuoInit,
                dim2408KuoInit,
                dim2409KuoInit,
                dim2410KuoInit,
                dim2411KuoInit,
                dim2412KuoInit,
                dim2413KuoInit,
                dim2414KuoInit,
                dim2415KuoInit,
                dim2416KuoInit,
                dim2417KuoInit,
                dim2418KuoInit,
                dim2419KuoInit,
                dim2420KuoInit,
                dim2421KuoInit,
                dim2422KuoInit,
                dim2423KuoInit,
                dim2424KuoInit,
                dim2425KuoInit,
                dim2426KuoInit,
                dim2427KuoInit,
                dim2428KuoInit,
                dim2429KuoInit,
                dim2430KuoInit,
                dim2431KuoInit,
                dim2432KuoInit,
                dim2433KuoInit,
                dim2434KuoInit,
                dim2435KuoInit,
                dim2436KuoInit,
                dim2437KuoInit,
                dim2438KuoInit,
                dim2439KuoInit,
                dim2440KuoInit,
                dim2441KuoInit,
                dim2442KuoInit,
                dim2443KuoInit,
                dim2444KuoInit,
                dim2445KuoInit,
                dim2446KuoInit,
                dim2447KuoInit,
                dim2448KuoInit,
                dim2449KuoInit,
                dim2450KuoInit,
                dim2451KuoInit,
                dim2452KuoInit,
                dim2453KuoInit,
                dim2454KuoInit,
                dim2455KuoInit,
                dim2456KuoInit,
                dim2457KuoInit,
                dim2458KuoInit,
                dim2459KuoInit,
                dim2460KuoInit,
                dim2461KuoInit,
                dim2462KuoInit,
                dim2463KuoInit,
                dim2464KuoInit,
                dim2465KuoInit,
                dim2466KuoInit,
                dim2467KuoInit,
                dim2468KuoInit,
                dim2469KuoInit,
                dim2470KuoInit,
                dim2471KuoInit,
                dim2472KuoInit,
                dim2473KuoInit,
                dim2474KuoInit,
                dim2475KuoInit,
                dim2476KuoInit,
                dim2477KuoInit,
                dim2478KuoInit,
                dim2479KuoInit,
                dim2480KuoInit,
                dim2481KuoInit,
                dim2482KuoInit,
                dim2483KuoInit,
                dim2484KuoInit,
                dim2485KuoInit,
                dim2486KuoInit,
                dim2487KuoInit,
                dim2488KuoInit,
                dim2489KuoInit,
                dim2490KuoInit,
                dim2491KuoInit,
                dim2492KuoInit,
                dim2493KuoInit,
                dim2494KuoInit,
                dim2495KuoInit,
                dim2496KuoInit,
                dim2497KuoInit,
                dim2498KuoInit,
                dim2499KuoInit,
                dim2500KuoInit,
                dim2501KuoInit,
                dim2502KuoInit,
                dim2503KuoInit,
                dim2504KuoInit,
                dim2505KuoInit,
                dim2506KuoInit,
                dim2507KuoInit,
                dim2508KuoInit,
                dim2509KuoInit,
                dim2510KuoInit,
                dim2511KuoInit,
                dim2512KuoInit,
                dim2513KuoInit,
                dim2514KuoInit,
                dim2515KuoInit,
                dim2516KuoInit,
                dim2517KuoInit,
                dim2518KuoInit,
                dim2519KuoInit,
                dim2520KuoInit,
                dim2521KuoInit,
                dim2522KuoInit,
                dim2523KuoInit,
                dim2524KuoInit,
                dim2525KuoInit,
                dim2526KuoInit,
                dim2527KuoInit,
                dim2528KuoInit,
                dim2529KuoInit,
                dim2530KuoInit,
                dim2531KuoInit,
                dim2532KuoInit,
                dim2533KuoInit,
                dim2534KuoInit,
                dim2535KuoInit,
                dim2536KuoInit,
                dim2537KuoInit,
                dim2538KuoInit,
                dim2539KuoInit,
                dim2540KuoInit,
                dim2541KuoInit,
                dim2542KuoInit,
                dim2543KuoInit,
                dim2544KuoInit,
                dim2545KuoInit,
                dim2546KuoInit,
                dim2547KuoInit,
                dim2548KuoInit,
                dim2549KuoInit,
                dim2550KuoInit,
                dim2551KuoInit,
                dim2552KuoInit,
                dim2553KuoInit,
                dim2554KuoInit,
                dim2555KuoInit,
                dim2556KuoInit,
                dim2557KuoInit,
                dim2558KuoInit,
                dim2559KuoInit,
                dim2560KuoInit,
                dim2561KuoInit,
                dim2562KuoInit,
                dim2563KuoInit,
                dim2564KuoInit,
                dim2565KuoInit,
                dim2566KuoInit,
                dim2567KuoInit,
                dim2568KuoInit,
                dim2569KuoInit,
                dim2570KuoInit,
                dim2571KuoInit,
                dim2572KuoInit,
                dim2573KuoInit,
                dim2574KuoInit,
                dim2575KuoInit,
                dim2576KuoInit,
                dim2577KuoInit,
                dim2578KuoInit,
                dim2579KuoInit,
                dim2580KuoInit,
                dim2581KuoInit,
                dim2582KuoInit,
                dim2583KuoInit,
                dim2584KuoInit,
                dim2585KuoInit,
                dim2586KuoInit,
                dim2587KuoInit,
                dim2588KuoInit,
                dim2589KuoInit,
                dim2590KuoInit,
                dim2591KuoInit,
                dim2592KuoInit,
                dim2593KuoInit,
                dim2594KuoInit,
                dim2595KuoInit,
                dim2596KuoInit,
                dim2597KuoInit,
                dim2598KuoInit,
                dim2599KuoInit,
                dim2600KuoInit,
                dim2601KuoInit,
                dim2602KuoInit,
                dim2603KuoInit,
                dim2604KuoInit,
                dim2605KuoInit,
                dim2606KuoInit,
                dim2607KuoInit,
                dim2608KuoInit,
                dim2609KuoInit,
                dim2610KuoInit,
                dim2611KuoInit,
                dim2612KuoInit,
                dim2613KuoInit,
                dim2614KuoInit,
                dim2615KuoInit,
                dim2616KuoInit,
                dim2617KuoInit,
                dim2618KuoInit,
                dim2619KuoInit,
                dim2620KuoInit,
                dim2621KuoInit,
                dim2622KuoInit,
                dim2623KuoInit,
                dim2624KuoInit,
                dim2625KuoInit,
                dim2626KuoInit,
                dim2627KuoInit,
                dim2628KuoInit,
                dim2629KuoInit,
                dim2630KuoInit,
                dim2631KuoInit,
                dim2632KuoInit,
                dim2633KuoInit,
                dim2634KuoInit,
                dim2635KuoInit,
                dim2636KuoInit,
                dim2637KuoInit,
                dim2638KuoInit,
                dim2639KuoInit,
                dim2640KuoInit,
                dim2641KuoInit,
                dim2642KuoInit,
                dim2643KuoInit,
                dim2644KuoInit,
                dim2645KuoInit,
                dim2646KuoInit,
                dim2647KuoInit,
                dim2648KuoInit,
                dim2649KuoInit,
                dim2650KuoInit,
                dim2651KuoInit,
                dim2652KuoInit,
                dim2653KuoInit,
                dim2654KuoInit,
                dim2655KuoInit,
                dim2656KuoInit,
                dim2657KuoInit,
                dim2658KuoInit,
                dim2659KuoInit,
                dim2660KuoInit,
                dim2661KuoInit,
                dim2662KuoInit,
                dim2663KuoInit,
                dim2664KuoInit,
                dim2665KuoInit,
                dim2666KuoInit,
                dim2667KuoInit,
                dim2668KuoInit,
                dim2669KuoInit,
                dim2670KuoInit,
                dim2671KuoInit,
                dim2672KuoInit,
                dim2673KuoInit,
                dim2674KuoInit,
                dim2675KuoInit,
                dim2676KuoInit,
                dim2677KuoInit,
                dim2678KuoInit,
                dim2679KuoInit,
                dim2680KuoInit,
                dim2681KuoInit,
                dim2682KuoInit,
                dim2683KuoInit,
                dim2684KuoInit,
                dim2685KuoInit,
                dim2686KuoInit,
                dim2687KuoInit,
                dim2688KuoInit,
                dim2689KuoInit,
                dim2690KuoInit,
                dim2691KuoInit,
                dim2692KuoInit,
                dim2693KuoInit,
                dim2694KuoInit,
                dim2695KuoInit,
                dim2696KuoInit,
                dim2697KuoInit,
                dim2698KuoInit,
                dim2699KuoInit,
                dim2700KuoInit,
                dim2701KuoInit,
                dim2702KuoInit,
                dim2703KuoInit,
                dim2704KuoInit,
                dim2705KuoInit,
                dim2706KuoInit,
                dim2707KuoInit,
                dim2708KuoInit,
                dim2709KuoInit,
                dim2710KuoInit,
                dim2711KuoInit,
                dim2712KuoInit,
                dim2713KuoInit,
                dim2714KuoInit,
                dim2715KuoInit,
                dim2716KuoInit,
                dim2717KuoInit,
                dim2718KuoInit,
                dim2719KuoInit,
                dim2720KuoInit,
                dim2721KuoInit,
                dim2722KuoInit,
                dim2723KuoInit,
                dim2724KuoInit,
                dim2725KuoInit,
                dim2726KuoInit,
                dim2727KuoInit,
                dim2728KuoInit,
                dim2729KuoInit,
                dim2730KuoInit,
                dim2731KuoInit,
                dim2732KuoInit,
                dim2733KuoInit,
                dim2734KuoInit,
                dim2735KuoInit,
                dim2736KuoInit,
                dim2737KuoInit,
                dim2738KuoInit,
                dim2739KuoInit,
                dim2740KuoInit,
                dim2741KuoInit,
                dim2742KuoInit,
                dim2743KuoInit,
                dim2744KuoInit,
                dim2745KuoInit,
                dim2746KuoInit,
                dim2747KuoInit,
                dim2748KuoInit,
                dim2749KuoInit,
                dim2750KuoInit,
                dim2751KuoInit,
                dim2752KuoInit,
                dim2753KuoInit,
                dim2754KuoInit,
                dim2755KuoInit,
                dim2756KuoInit,
                dim2757KuoInit,
                dim2758KuoInit,
                dim2759KuoInit,
                dim2760KuoInit,
                dim2761KuoInit,
                dim2762KuoInit,
                dim2763KuoInit,
                dim2764KuoInit,
                dim2765KuoInit,
                dim2766KuoInit,
                dim2767KuoInit,
                dim2768KuoInit,
                dim2769KuoInit,
                dim2770KuoInit,
                dim2771KuoInit,
                dim2772KuoInit,
                dim2773KuoInit,
                dim2774KuoInit,
                dim2775KuoInit,
                dim2776KuoInit,
                dim2777KuoInit,
                dim2778KuoInit,
                dim2779KuoInit,
                dim2780KuoInit,
                dim2781KuoInit,
                dim2782KuoInit,
                dim2783KuoInit,
                dim2784KuoInit,
                dim2785KuoInit,
                dim2786KuoInit,
                dim2787KuoInit,
                dim2788KuoInit,
                dim2789KuoInit,
                dim2790KuoInit,
                dim2791KuoInit,
                dim2792KuoInit,
                dim2793KuoInit,
                dim2794KuoInit,
                dim2795KuoInit,
                dim2796KuoInit,
                dim2797KuoInit,
                dim2798KuoInit,
                dim2799KuoInit,
                dim2800KuoInit,
                dim2801KuoInit,
                dim2802KuoInit,
                dim2803KuoInit,
                dim2804KuoInit,
                dim2805KuoInit,
                dim2806KuoInit,
                dim2807KuoInit,
                dim2808KuoInit,
                dim2809KuoInit,
                dim2810KuoInit,
                dim2811KuoInit,
                dim2812KuoInit,
                dim2813KuoInit,
                dim2814KuoInit,
                dim2815KuoInit,
                dim2816KuoInit,
                dim2817KuoInit,
                dim2818KuoInit,
                dim2819KuoInit,
                dim2820KuoInit,
                dim2821KuoInit,
                dim2822KuoInit,
                dim2823KuoInit,
                dim2824KuoInit,
                dim2825KuoInit,
                dim2826KuoInit,
                dim2827KuoInit,
                dim2828KuoInit,
                dim2829KuoInit,
                dim2830KuoInit,
                dim2831KuoInit,
                dim2832KuoInit,
                dim2833KuoInit,
                dim2834KuoInit,
                dim2835KuoInit,
                dim2836KuoInit,
                dim2837KuoInit,
                dim2838KuoInit,
                dim2839KuoInit,
                dim2840KuoInit,
                dim2841KuoInit,
                dim2842KuoInit,
                dim2843KuoInit,
                dim2844KuoInit,
                dim2845KuoInit,
                dim2846KuoInit,
                dim2847KuoInit,
                dim2848KuoInit,
                dim2849KuoInit,
                dim2850KuoInit,
                dim2851KuoInit,
                dim2852KuoInit,
                dim2853KuoInit,
                dim2854KuoInit,
                dim2855KuoInit,
                dim2856KuoInit,
                dim2857KuoInit,
                dim2858KuoInit,
                dim2859KuoInit,
                dim2860KuoInit,
                dim2861KuoInit,
                dim2862KuoInit,
                dim2863KuoInit,
                dim2864KuoInit,
                dim2865KuoInit,
                dim2866KuoInit,
                dim2867KuoInit,
                dim2868KuoInit,
                dim2869KuoInit,
                dim2870KuoInit,
                dim2871KuoInit,
                dim2872KuoInit,
                dim2873KuoInit,
                dim2874KuoInit,
                dim2875KuoInit,
                dim2876KuoInit,
                dim2877KuoInit,
                dim2878KuoInit,
                dim2879KuoInit,
                dim2880KuoInit,
                dim2881KuoInit,
                dim2882KuoInit,
                dim2883KuoInit,
                dim2884KuoInit,
                dim2885KuoInit,
                dim2886KuoInit,
                dim2887KuoInit,
                dim2888KuoInit,
                dim2889KuoInit,
                dim2890KuoInit,
                dim2891KuoInit,
                dim2892KuoInit,
                dim2893KuoInit,
                dim2894KuoInit,
                dim2895KuoInit,
                dim2896KuoInit,
                dim2897KuoInit,
                dim2898KuoInit,
                dim2899KuoInit,
                dim2900KuoInit,
                dim2901KuoInit,
                dim2902KuoInit,
                dim2903KuoInit,
                dim2904KuoInit,
                dim2905KuoInit,
                dim2906KuoInit,
                dim2907KuoInit,
                dim2908KuoInit,
                dim2909KuoInit,
                dim2910KuoInit,
                dim2911KuoInit,
                dim2912KuoInit,
                dim2913KuoInit,
                dim2914KuoInit,
                dim2915KuoInit,
                dim2916KuoInit,
                dim2917KuoInit,
                dim2918KuoInit,
                dim2919KuoInit,
                dim2920KuoInit,
                dim2921KuoInit,
                dim2922KuoInit,
                dim2923KuoInit,
                dim2924KuoInit,
                dim2925KuoInit,
                dim2926KuoInit,
                dim2927KuoInit,
                dim2928KuoInit,
                dim2929KuoInit,
                dim2930KuoInit,
                dim2931KuoInit,
                dim2932KuoInit,
                dim2933KuoInit,
                dim2934KuoInit,
                dim2935KuoInit,
                dim2936KuoInit,
                dim2937KuoInit,
                dim2938KuoInit,
                dim2939KuoInit,
                dim2940KuoInit,
                dim2941KuoInit,
                dim2942KuoInit,
                dim2943KuoInit,
                dim2944KuoInit,
                dim2945KuoInit,
                dim2946KuoInit,
                dim2947KuoInit,
                dim2948KuoInit,
                dim2949KuoInit,
                dim2950KuoInit,
                dim2951KuoInit,
                dim2952KuoInit,
                dim2953KuoInit,
                dim2954KuoInit,
                dim2955KuoInit,
                dim2956KuoInit,
                dim2957KuoInit,
                dim2958KuoInit,
                dim2959KuoInit,
                dim2960KuoInit,
                dim2961KuoInit,
                dim2962KuoInit,
                dim2963KuoInit,
                dim2964KuoInit,
                dim2965KuoInit,
                dim2966KuoInit,
                dim2967KuoInit,
                dim2968KuoInit,
                dim2969KuoInit,
                dim2970KuoInit,
                dim2971KuoInit,
                dim2972KuoInit,
                dim2973KuoInit,
                dim2974KuoInit,
                dim2975KuoInit,
                dim2976KuoInit,
                dim2977KuoInit,
                dim2978KuoInit,
                dim2979KuoInit,
                dim2980KuoInit,
                dim2981KuoInit,
                dim2982KuoInit,
                dim2983KuoInit,
                dim2984KuoInit,
                dim2985KuoInit,
                dim2986KuoInit,
                dim2987KuoInit,
                dim2988KuoInit,
                dim2989KuoInit,
                dim2990KuoInit,
                dim2991KuoInit,
                dim2992KuoInit,
                dim2993KuoInit,
                dim2994KuoInit,
                dim2995KuoInit,
                dim2996KuoInit,
                dim2997KuoInit,
                dim2998KuoInit,
                dim2999KuoInit,
                dim3000KuoInit,
                dim3001KuoInit,
                dim3002KuoInit,
                dim3003KuoInit,
                dim3004KuoInit,
                dim3005KuoInit,
                dim3006KuoInit,
                dim3007KuoInit,
                dim3008KuoInit,
                dim3009KuoInit,
                dim3010KuoInit,
                dim3011KuoInit,
                dim3012KuoInit,
                dim3013KuoInit,
                dim3014KuoInit,
                dim3015KuoInit,
                dim3016KuoInit,
                dim3017KuoInit,
                dim3018KuoInit,
                dim3019KuoInit,
                dim3020KuoInit,
                dim3021KuoInit,
                dim3022KuoInit,
                dim3023KuoInit,
                dim3024KuoInit,
                dim3025KuoInit,
                dim3026KuoInit,
                dim3027KuoInit,
                dim3028KuoInit,
                dim3029KuoInit,
                dim3030KuoInit,
                dim3031KuoInit,
                dim3032KuoInit,
                dim3033KuoInit,
                dim3034KuoInit,
                dim3035KuoInit,
                dim3036KuoInit,
                dim3037KuoInit,
                dim3038KuoInit,
                dim3039KuoInit,
                dim3040KuoInit,
                dim3041KuoInit,
                dim3042KuoInit,
                dim3043KuoInit,
                dim3044KuoInit,
                dim3045KuoInit,
                dim3046KuoInit,
                dim3047KuoInit,
                dim3048KuoInit,
                dim3049KuoInit,
                dim3050KuoInit,
                dim3051KuoInit,
                dim3052KuoInit,
                dim3053KuoInit,
                dim3054KuoInit,
                dim3055KuoInit,
                dim3056KuoInit,
                dim3057KuoInit,
                dim3058KuoInit,
                dim3059KuoInit,
                dim3060KuoInit,
                dim3061KuoInit,
                dim3062KuoInit,
                dim3063KuoInit,
                dim3064KuoInit,
                dim3065KuoInit,
                dim3066KuoInit,
                dim3067KuoInit,
                dim3068KuoInit,
                dim3069KuoInit,
                dim3070KuoInit,
                dim3071KuoInit,
                dim3072KuoInit,
                dim3073KuoInit,
                dim3074KuoInit,
                dim3075KuoInit,
                dim3076KuoInit,
                dim3077KuoInit,
                dim3078KuoInit,
                dim3079KuoInit,
                dim3080KuoInit,
                dim3081KuoInit,
                dim3082KuoInit,
                dim3083KuoInit,
                dim3084KuoInit,
                dim3085KuoInit,
                dim3086KuoInit,
                dim3087KuoInit,
                dim3088KuoInit,
                dim3089KuoInit,
                dim3090KuoInit,
                dim3091KuoInit,
                dim3092KuoInit,
                dim3093KuoInit,
                dim3094KuoInit,
                dim3095KuoInit,
                dim3096KuoInit,
                dim3097KuoInit,
                dim3098KuoInit,
                dim3099KuoInit,
                dim3100KuoInit,
                dim3101KuoInit,
                dim3102KuoInit,
                dim3103KuoInit,
                dim3104KuoInit,
                dim3105KuoInit,
                dim3106KuoInit,
                dim3107KuoInit,
                dim3108KuoInit,
                dim3109KuoInit,
                dim3110KuoInit,
                dim3111KuoInit,
                dim3112KuoInit,
                dim3113KuoInit,
                dim3114KuoInit,
                dim3115KuoInit,
                dim3116KuoInit,
                dim3117KuoInit,
                dim3118KuoInit,
                dim3119KuoInit,
                dim3120KuoInit,
                dim3121KuoInit,
                dim3122KuoInit,
                dim3123KuoInit,
                dim3124KuoInit,
                dim3125KuoInit,
                dim3126KuoInit,
                dim3127KuoInit,
                dim3128KuoInit,
                dim3129KuoInit,
                dim3130KuoInit,
                dim3131KuoInit,
                dim3132KuoInit,
                dim3133KuoInit,
                dim3134KuoInit,
                dim3135KuoInit,
                dim3136KuoInit,
                dim3137KuoInit,
                dim3138KuoInit,
                dim3139KuoInit,
                dim3140KuoInit,
                dim3141KuoInit,
                dim3142KuoInit,
                dim3143KuoInit,
                dim3144KuoInit,
                dim3145KuoInit,
                dim3146KuoInit,
                dim3147KuoInit,
                dim3148KuoInit,
                dim3149KuoInit,
                dim3150KuoInit,
                dim3151KuoInit,
                dim3152KuoInit,
                dim3153KuoInit,
                dim3154KuoInit,
                dim3155KuoInit,
                dim3156KuoInit,
                dim3157KuoInit,
                dim3158KuoInit,
                dim3159KuoInit,
                dim3160KuoInit,
                dim3161KuoInit,
                dim3162KuoInit,
                dim3163KuoInit,
                dim3164KuoInit,
                dim3165KuoInit,
                dim3166KuoInit,
                dim3167KuoInit,
                dim3168KuoInit,
                dim3169KuoInit,
                dim3170KuoInit,
                dim3171KuoInit,
                dim3172KuoInit,
                dim3173KuoInit,
                dim3174KuoInit,
                dim3175KuoInit,
                dim3176KuoInit,
                dim3177KuoInit,
                dim3178KuoInit,
                dim3179KuoInit,
                dim3180KuoInit,
                dim3181KuoInit,
                dim3182KuoInit,
                dim3183KuoInit,
                dim3184KuoInit,
                dim3185KuoInit,
                dim3186KuoInit,
                dim3187KuoInit,
                dim3188KuoInit,
                dim3189KuoInit,
                dim3190KuoInit,
                dim3191KuoInit,
                dim3192KuoInit,
                dim3193KuoInit,
                dim3194KuoInit,
                dim3195KuoInit,
                dim3196KuoInit,
                dim3197KuoInit,
                dim3198KuoInit,
                dim3199KuoInit,
                dim3200KuoInit,
                dim3201KuoInit,
                dim3202KuoInit,
                dim3203KuoInit,
                dim3204KuoInit,
                dim3205KuoInit,
                dim3206KuoInit,
                dim3207KuoInit,
                dim3208KuoInit,
                dim3209KuoInit,
                dim3210KuoInit,
                dim3211KuoInit,
                dim3212KuoInit,
                dim3213KuoInit,
                dim3214KuoInit,
                dim3215KuoInit,
                dim3216KuoInit,
                dim3217KuoInit,
                dim3218KuoInit,
                dim3219KuoInit,
                dim3220KuoInit,
                dim3221KuoInit,
                dim3222KuoInit,
                dim3223KuoInit,
                dim3224KuoInit,
                dim3225KuoInit,
                dim3226KuoInit,
                dim3227KuoInit,
                dim3228KuoInit,
                dim3229KuoInit,
                dim3230KuoInit,
                dim3231KuoInit,
                dim3232KuoInit,
                dim3233KuoInit,
                dim3234KuoInit,
                dim3235KuoInit,
                dim3236KuoInit,
                dim3237KuoInit,
                dim3238KuoInit,
                dim3239KuoInit,
                dim3240KuoInit,
                dim3241KuoInit,
                dim3242KuoInit,
                dim3243KuoInit,
                dim3244KuoInit,
                dim3245KuoInit,
                dim3246KuoInit,
                dim3247KuoInit,
                dim3248KuoInit,
                dim3249KuoInit,
                dim3250KuoInit,
                dim3251KuoInit,
                dim3252KuoInit,
                dim3253KuoInit,
                dim3254KuoInit,
                dim3255KuoInit,
                dim3256KuoInit,
                dim3257KuoInit,
                dim3258KuoInit,
                dim3259KuoInit,
                dim3260KuoInit,
                dim3261KuoInit,
                dim3262KuoInit,
                dim3263KuoInit,
                dim3264KuoInit,
                dim3265KuoInit,
                dim3266KuoInit,
                dim3267KuoInit,
                dim3268KuoInit,
                dim3269KuoInit,
                dim3270KuoInit,
                dim3271KuoInit,
                dim3272KuoInit,
                dim3273KuoInit,
                dim3274KuoInit,
                dim3275KuoInit,
                dim3276KuoInit,
                dim3277KuoInit,
                dim3278KuoInit,
                dim3279KuoInit,
                dim3280KuoInit,
                dim3281KuoInit,
                dim3282KuoInit,
                dim3283KuoInit,
                dim3284KuoInit,
                dim3285KuoInit,
                dim3286KuoInit,
                dim3287KuoInit,
                dim3288KuoInit,
                dim3289KuoInit,
                dim3290KuoInit,
                dim3291KuoInit,
                dim3292KuoInit,
                dim3293KuoInit,
                dim3294KuoInit,
                dim3295KuoInit,
                dim3296KuoInit,
                dim3297KuoInit,
                dim3298KuoInit,
                dim3299KuoInit,
                dim3300KuoInit,
                dim3301KuoInit,
                dim3302KuoInit,
                dim3303KuoInit,
                dim3304KuoInit,
                dim3305KuoInit,
                dim3306KuoInit,
                dim3307KuoInit,
                dim3308KuoInit,
                dim3309KuoInit,
                dim3310KuoInit,
                dim3311KuoInit,
                dim3312KuoInit,
                dim3313KuoInit,
                dim3314KuoInit,
                dim3315KuoInit,
                dim3316KuoInit,
                dim3317KuoInit,
                dim3318KuoInit,
                dim3319KuoInit,
                dim3320KuoInit,
                dim3321KuoInit,
                dim3322KuoInit,
                dim3323KuoInit,
                dim3324KuoInit,
                dim3325KuoInit,
                dim3326KuoInit,
                dim3327KuoInit,
                dim3328KuoInit,
                dim3329KuoInit,
                dim3330KuoInit,
                dim3331KuoInit,
                dim3332KuoInit,
                dim3333KuoInit,
                dim3334KuoInit,
                dim3335KuoInit,
                dim3336KuoInit,
                dim3337KuoInit,
                dim3338KuoInit,
                dim3339KuoInit,
                dim3340KuoInit,
                dim3341KuoInit,
                dim3342KuoInit,
                dim3343KuoInit,
                dim3344KuoInit,
                dim3345KuoInit,
                dim3346KuoInit,
                dim3347KuoInit,
                dim3348KuoInit,
                dim3349KuoInit,
                dim3350KuoInit,
                dim3351KuoInit,
                dim3352KuoInit,
                dim3353KuoInit,
                dim3354KuoInit,
                dim3355KuoInit,
                dim3356KuoInit,
                dim3357KuoInit,
                dim3358KuoInit,
                dim3359KuoInit,
                dim3360KuoInit,
                dim3361KuoInit,
                dim3362KuoInit,
                dim3363KuoInit,
                dim3364KuoInit,
                dim3365KuoInit,
                dim3366KuoInit,
                dim3367KuoInit,
                dim3368KuoInit,
                dim3369KuoInit,
                dim3370KuoInit,
                dim3371KuoInit,
                dim3372KuoInit,
                dim3373KuoInit,
                dim3374KuoInit,
                dim3375KuoInit,
                dim3376KuoInit,
                dim3377KuoInit,
                dim3378KuoInit,
                dim3379KuoInit,
                dim3380KuoInit,
                dim3381KuoInit,
                dim3382KuoInit,
                dim3383KuoInit,
                dim3384KuoInit,
                dim3385KuoInit,
                dim3386KuoInit,
                dim3387KuoInit,
                dim3388KuoInit,
                dim3389KuoInit,
                dim3390KuoInit,
                dim3391KuoInit,
                dim3392KuoInit,
                dim3393KuoInit,
                dim3394KuoInit,
                dim3395KuoInit,
                dim3396KuoInit,
                dim3397KuoInit,
                dim3398KuoInit,
                dim3399KuoInit,
                dim3400KuoInit,
                dim3401KuoInit,
                dim3402KuoInit,
                dim3403KuoInit,
                dim3404KuoInit,
                dim3405KuoInit,
                dim3406KuoInit,
                dim3407KuoInit,
                dim3408KuoInit,
                dim3409KuoInit,
                dim3410KuoInit,
                dim3411KuoInit,
                dim3412KuoInit,
                dim3413KuoInit,
                dim3414KuoInit,
                dim3415KuoInit,
                dim3416KuoInit,
                dim3417KuoInit,
                dim3418KuoInit,
                dim3419KuoInit,
                dim3420KuoInit,
                dim3421KuoInit,
                dim3422KuoInit,
                dim3423KuoInit,
                dim3424KuoInit,
                dim3425KuoInit,
                dim3426KuoInit,
                dim3427KuoInit,
                dim3428KuoInit,
                dim3429KuoInit,
                dim3430KuoInit,
                dim3431KuoInit,
                dim3432KuoInit,
                dim3433KuoInit,
                dim3434KuoInit,
                dim3435KuoInit,
                dim3436KuoInit,
                dim3437KuoInit,
                dim3438KuoInit,
                dim3439KuoInit,
                dim3440KuoInit,
                dim3441KuoInit,
                dim3442KuoInit,
                dim3443KuoInit,
                dim3444KuoInit,
                dim3445KuoInit,
                dim3446KuoInit,
                dim3447KuoInit,
                dim3448KuoInit,
                dim3449KuoInit,
                dim3450KuoInit,
                dim3451KuoInit,
                dim3452KuoInit,
                dim3453KuoInit,
                dim3454KuoInit,
                dim3455KuoInit,
                dim3456KuoInit,
                dim3457KuoInit,
                dim3458KuoInit,
                dim3459KuoInit,
                dim3460KuoInit,
                dim3461KuoInit,
                dim3462KuoInit,
                dim3463KuoInit,
                dim3464KuoInit,
                dim3465KuoInit,
                dim3466KuoInit,
                dim3467KuoInit,
                dim3468KuoInit,
                dim3469KuoInit,
                dim3470KuoInit,
                dim3471KuoInit,
                dim3472KuoInit,
                dim3473KuoInit,
                dim3474KuoInit,
                dim3475KuoInit,
                dim3476KuoInit,
                dim3477KuoInit,
                dim3478KuoInit,
                dim3479KuoInit,
                dim3480KuoInit,
                dim3481KuoInit,
                dim3482KuoInit,
                dim3483KuoInit,
                dim3484KuoInit,
                dim3485KuoInit,
                dim3486KuoInit,
                dim3487KuoInit,
                dim3488KuoInit,
                dim3489KuoInit,
                dim3490KuoInit,
                dim3491KuoInit,
                dim3492KuoInit,
                dim3493KuoInit,
                dim3494KuoInit,
                dim3495KuoInit,
                dim3496KuoInit,
                dim3497KuoInit,
                dim3498KuoInit,
                dim3499KuoInit,
                dim3500KuoInit,
                dim3501KuoInit,
                dim3502KuoInit,
                dim3503KuoInit,
                dim3504KuoInit,
                dim3505KuoInit,
                dim3506KuoInit,
                dim3507KuoInit,
                dim3508KuoInit,
                dim3509KuoInit,
                dim3510KuoInit,
                dim3511KuoInit,
                dim3512KuoInit,
                dim3513KuoInit,
                dim3514KuoInit,
                dim3515KuoInit,
                dim3516KuoInit,
                dim3517KuoInit,
                dim3518KuoInit,
                dim3519KuoInit,
                dim3520KuoInit,
                dim3521KuoInit,
                dim3522KuoInit,
                dim3523KuoInit,
                dim3524KuoInit,
                dim3525KuoInit,
                dim3526KuoInit,
                dim3527KuoInit,
                dim3528KuoInit,
                dim3529KuoInit,
                dim3530KuoInit,
                dim3531KuoInit,
                dim3532KuoInit,
                dim3533KuoInit,
                dim3534KuoInit,
                dim3535KuoInit,
                dim3536KuoInit,
                dim3537KuoInit,
                dim3538KuoInit,
                dim3539KuoInit,
                dim3540KuoInit,
                dim3541KuoInit,
                dim3542KuoInit,
                dim3543KuoInit,
                dim3544KuoInit,
                dim3545KuoInit,
                dim3546KuoInit,
                dim3547KuoInit,
                dim3548KuoInit,
                dim3549KuoInit,
                dim3550KuoInit,
                dim3551KuoInit,
                dim3552KuoInit,
                dim3553KuoInit,
                dim3554KuoInit,
                dim3555KuoInit,
                dim3556KuoInit,
                dim3557KuoInit,
                dim3558KuoInit,
                dim3559KuoInit,
                dim3560KuoInit,
                dim3561KuoInit,
                dim3562KuoInit,
                dim3563KuoInit,
                dim3564KuoInit,
                dim3565KuoInit,
                dim3566KuoInit,
                dim3567KuoInit,
                dim3568KuoInit,
                dim3569KuoInit,
                dim3570KuoInit,
                dim3571KuoInit,
                dim3572KuoInit,
                dim3573KuoInit,
                dim3574KuoInit,
                dim3575KuoInit,
                dim3576KuoInit,
                dim3577KuoInit,
                dim3578KuoInit,
                dim3579KuoInit,
                dim3580KuoInit,
                dim3581KuoInit,
                dim3582KuoInit,
                dim3583KuoInit,
                dim3584KuoInit,
                dim3585KuoInit,
                dim3586KuoInit,
                dim3587KuoInit,
                dim3588KuoInit,
                dim3589KuoInit,
                dim3590KuoInit,
                dim3591KuoInit,
                dim3592KuoInit,
                dim3593KuoInit,
                dim3594KuoInit,
                dim3595KuoInit,
                dim3596KuoInit,
                dim3597KuoInit,
                dim3598KuoInit,
                dim3599KuoInit,
                dim3600KuoInit,
                dim3601KuoInit,
                dim3602KuoInit,
                dim3603KuoInit,
                dim3604KuoInit,
                dim3605KuoInit,
                dim3606KuoInit,
                dim3607KuoInit,
                dim3608KuoInit,
                dim3609KuoInit,
                dim3610KuoInit,
                dim3611KuoInit,
                dim3612KuoInit,
                dim3613KuoInit,
                dim3614KuoInit,
                dim3615KuoInit,
                dim3616KuoInit,
                dim3617KuoInit,
                dim3618KuoInit,
                dim3619KuoInit,
                dim3620KuoInit,
                dim3621KuoInit,
                dim3622KuoInit,
                dim3623KuoInit,
                dim3624KuoInit,
                dim3625KuoInit,
                dim3626KuoInit,
                dim3627KuoInit,
                dim3628KuoInit,
                dim3629KuoInit,
                dim3630KuoInit,
                dim3631KuoInit,
                dim3632KuoInit,
                dim3633KuoInit,
                dim3634KuoInit,
                dim3635KuoInit,
                dim3636KuoInit,
                dim3637KuoInit,
                dim3638KuoInit,
                dim3639KuoInit,
                dim3640KuoInit,
                dim3641KuoInit,
                dim3642KuoInit,
                dim3643KuoInit,
                dim3644KuoInit,
                dim3645KuoInit,
                dim3646KuoInit,
                dim3647KuoInit,
                dim3648KuoInit,
                dim3649KuoInit,
                dim3650KuoInit,
                dim3651KuoInit,
                dim3652KuoInit,
                dim3653KuoInit,
                dim3654KuoInit,
                dim3655KuoInit,
                dim3656KuoInit,
                dim3657KuoInit,
                dim3658KuoInit,
                dim3659KuoInit,
                dim3660KuoInit,
                dim3661KuoInit,
                dim3662KuoInit,
                dim3663KuoInit,
                dim3664KuoInit,
                dim3665KuoInit,
                dim3666KuoInit,
                dim3667KuoInit,
                dim3668KuoInit,
                dim3669KuoInit,
                dim3670KuoInit,
                dim3671KuoInit,
                dim3672KuoInit,
                dim3673KuoInit,
                dim3674KuoInit,
                dim3675KuoInit,
                dim3676KuoInit,
                dim3677KuoInit,
                dim3678KuoInit,
                dim3679KuoInit,
                dim3680KuoInit,
                dim3681KuoInit,
                dim3682KuoInit,
                dim3683KuoInit,
                dim3684KuoInit,
                dim3685KuoInit,
                dim3686KuoInit,
                dim3687KuoInit,
                dim3688KuoInit,
                dim3689KuoInit,
                dim3690KuoInit,
                dim3691KuoInit,
                dim3692KuoInit,
                dim3693KuoInit,
                dim3694KuoInit,
                dim3695KuoInit,
                dim3696KuoInit,
                dim3697KuoInit,
                dim3698KuoInit,
                dim3699KuoInit,
                dim3700KuoInit,
                dim3701KuoInit,
                dim3702KuoInit,
                dim3703KuoInit,
                dim3704KuoInit,
                dim3705KuoInit,
                dim3706KuoInit,
                dim3707KuoInit,
                dim3708KuoInit,
                dim3709KuoInit,
                dim3710KuoInit,
                dim3711KuoInit,
                dim3712KuoInit,
                dim3713KuoInit,
                dim3714KuoInit,
                dim3715KuoInit,
                dim3716KuoInit,
                dim3717KuoInit,
                dim3718KuoInit,
                dim3719KuoInit,
                dim3720KuoInit,
                dim3721KuoInit,
                dim3722KuoInit,
                dim3723KuoInit,
                dim3724KuoInit,
                dim3725KuoInit,
                dim3726KuoInit,
                dim3727KuoInit,
                dim3728KuoInit,
                dim3729KuoInit,
                dim3730KuoInit,
                dim3731KuoInit,
                dim3732KuoInit,
                dim3733KuoInit,
                dim3734KuoInit,
                dim3735KuoInit,
                dim3736KuoInit,
                dim3737KuoInit,
                dim3738KuoInit,
                dim3739KuoInit,
                dim3740KuoInit,
                dim3741KuoInit,
                dim3742KuoInit,
                dim3743KuoInit,
                dim3744KuoInit,
                dim3745KuoInit,
                dim3746KuoInit,
                dim3747KuoInit,
                dim3748KuoInit,
                dim3749KuoInit,
                dim3750KuoInit,
                dim3751KuoInit,
                dim3752KuoInit,
                dim3753KuoInit,
                dim3754KuoInit,
                dim3755KuoInit,
                dim3756KuoInit,
                dim3757KuoInit,
                dim3758KuoInit,
                dim3759KuoInit,
                dim3760KuoInit,
                dim3761KuoInit,
                dim3762KuoInit,
                dim3763KuoInit,
                dim3764KuoInit,
                dim3765KuoInit,
                dim3766KuoInit,
                dim3767KuoInit,
                dim3768KuoInit,
                dim3769KuoInit,
                dim3770KuoInit,
                dim3771KuoInit,
                dim3772KuoInit,
                dim3773KuoInit,
                dim3774KuoInit,
                dim3775KuoInit,
                dim3776KuoInit,
                dim3777KuoInit,
                dim3778KuoInit,
                dim3779KuoInit,
                dim3780KuoInit,
                dim3781KuoInit,
                dim3782KuoInit,
                dim3783KuoInit,
                dim3784KuoInit,
                dim3785KuoInit,
                dim3786KuoInit,
                dim3787KuoInit,
                dim3788KuoInit,
                dim3789KuoInit,
                dim3790KuoInit,
                dim3791KuoInit,
                dim3792KuoInit,
                dim3793KuoInit,
                dim3794KuoInit,
                dim3795KuoInit,
                dim3796KuoInit,
                dim3797KuoInit,
                dim3798KuoInit,
                dim3799KuoInit,
                dim3800KuoInit,
                dim3801KuoInit,
                dim3802KuoInit,
                dim3803KuoInit,
                dim3804KuoInit,
                dim3805KuoInit,
                dim3806KuoInit,
                dim3807KuoInit,
                dim3808KuoInit,
                dim3809KuoInit,
                dim3810KuoInit,
                dim3811KuoInit,
                dim3812KuoInit,
                dim3813KuoInit,
                dim3814KuoInit,
                dim3815KuoInit,
                dim3816KuoInit,
                dim3817KuoInit,
                dim3818KuoInit,
                dim3819KuoInit,
                dim3820KuoInit,
                dim3821KuoInit,
                dim3822KuoInit,
                dim3823KuoInit,
                dim3824KuoInit,
                dim3825KuoInit,
                dim3826KuoInit,
                dim3827KuoInit,
                dim3828KuoInit,
                dim3829KuoInit,
                dim3830KuoInit,
                dim3831KuoInit,
                dim3832KuoInit,
                dim3833KuoInit,
                dim3834KuoInit,
                dim3835KuoInit,
                dim3836KuoInit,
                dim3837KuoInit,
                dim3838KuoInit,
                dim3839KuoInit,
                dim3840KuoInit,
                dim3841KuoInit,
                dim3842KuoInit,
                dim3843KuoInit,
                dim3844KuoInit,
                dim3845KuoInit,
                dim3846KuoInit,
                dim3847KuoInit,
                dim3848KuoInit,
                dim3849KuoInit,
                dim3850KuoInit,
                dim3851KuoInit,
                dim3852KuoInit,
                dim3853KuoInit,
                dim3854KuoInit,
                dim3855KuoInit,
                dim3856KuoInit,
                dim3857KuoInit,
                dim3858KuoInit,
                dim3859KuoInit,
                dim3860KuoInit,
                dim3861KuoInit,
                dim3862KuoInit,
                dim3863KuoInit,
                dim3864KuoInit,
                dim3865KuoInit,
                dim3866KuoInit,
                dim3867KuoInit,
                dim3868KuoInit,
                dim3869KuoInit,
                dim3870KuoInit,
                dim3871KuoInit,
                dim3872KuoInit,
                dim3873KuoInit,
                dim3874KuoInit,
                dim3875KuoInit,
                dim3876KuoInit,
                dim3877KuoInit,
                dim3878KuoInit,
                dim3879KuoInit,
                dim3880KuoInit,
                dim3881KuoInit,
                dim3882KuoInit,
                dim3883KuoInit,
                dim3884KuoInit,
                dim3885KuoInit,
                dim3886KuoInit,
                dim3887KuoInit,
                dim3888KuoInit,
                dim3889KuoInit,
                dim3890KuoInit,
                dim3891KuoInit,
                dim3892KuoInit,
                dim3893KuoInit,
                dim3894KuoInit,
                dim3895KuoInit,
                dim3896KuoInit,
                dim3897KuoInit,
                dim3898KuoInit,
                dim3899KuoInit,
                dim3900KuoInit,
                dim3901KuoInit,
                dim3902KuoInit,
                dim3903KuoInit,
                dim3904KuoInit,
                dim3905KuoInit,
                dim3906KuoInit,
                dim3907KuoInit,
                dim3908KuoInit,
                dim3909KuoInit,
                dim3910KuoInit,
                dim3911KuoInit,
                dim3912KuoInit,
                dim3913KuoInit,
                dim3914KuoInit,
                dim3915KuoInit,
                dim3916KuoInit,
                dim3917KuoInit,
                dim3918KuoInit,
                dim3919KuoInit,
                dim3920KuoInit,
                dim3921KuoInit,
                dim3922KuoInit,
                dim3923KuoInit,
                dim3924KuoInit,
                dim3925KuoInit,
                dim3926KuoInit,
                dim3927KuoInit,
                dim3928KuoInit,
                dim3929KuoInit,
                dim3930KuoInit,
                dim3931KuoInit,
                dim3932KuoInit,
                dim3933KuoInit,
                dim3934KuoInit,
                dim3935KuoInit,
                dim3936KuoInit,
                dim3937KuoInit,
                dim3938KuoInit,
                dim3939KuoInit,
                dim3940KuoInit,
                dim3941KuoInit,
                dim3942KuoInit,
                dim3943KuoInit,
                dim3944KuoInit,
                dim3945KuoInit,
                dim3946KuoInit,
                dim3947KuoInit,
                dim3948KuoInit,
                dim3949KuoInit,
                dim3950KuoInit,
                dim3951KuoInit,
                dim3952KuoInit,
                dim3953KuoInit,
                dim3954KuoInit,
                dim3955KuoInit,
                dim3956KuoInit,
                dim3957KuoInit,
                dim3958KuoInit,
                dim3959KuoInit,
                dim3960KuoInit,
                dim3961KuoInit,
                dim3962KuoInit,
                dim3963KuoInit,
                dim3964KuoInit,
                dim3965KuoInit,
                dim3966KuoInit,
                dim3967KuoInit,
                dim3968KuoInit,
                dim3969KuoInit,
                dim3970KuoInit,
                dim3971KuoInit,
                dim3972KuoInit,
                dim3973KuoInit,
                dim3974KuoInit,
                dim3975KuoInit,
                dim3976KuoInit,
                dim3977KuoInit,
                dim3978KuoInit,
                dim3979KuoInit,
                dim3980KuoInit,
                dim3981KuoInit,
                dim3982KuoInit,
                dim3983KuoInit,
                dim3984KuoInit,
                dim3985KuoInit,
                dim3986KuoInit,
                dim3987KuoInit,
                dim3988KuoInit,
                dim3989KuoInit,
                dim3990KuoInit,
                dim3991KuoInit,
                dim3992KuoInit,
                dim3993KuoInit,
                dim3994KuoInit,
                dim3995KuoInit,
                dim3996KuoInit,
                dim3997KuoInit,
                dim3998KuoInit,
                dim3999KuoInit,
                dim4000KuoInit,
                dim4001KuoInit,
                dim4002KuoInit,
                dim4003KuoInit,
                dim4004KuoInit,
                dim4005KuoInit,
                dim4006KuoInit,
                dim4007KuoInit,
                dim4008KuoInit,
                dim4009KuoInit,
                dim4010KuoInit,
                dim4011KuoInit,
                dim4012KuoInit,
                dim4013KuoInit,
                dim4014KuoInit,
                dim4015KuoInit,
                dim4016KuoInit,
                dim4017KuoInit,
                dim4018KuoInit,
                dim4019KuoInit,
                dim4020KuoInit,
                dim4021KuoInit,
                dim4022KuoInit,
                dim4023KuoInit,
                dim4024KuoInit,
                dim4025KuoInit,
                dim4026KuoInit,
                dim4027KuoInit,
                dim4028KuoInit,
                dim4029KuoInit,
                dim4030KuoInit,
                dim4031KuoInit,
                dim4032KuoInit,
                dim4033KuoInit,
                dim4034KuoInit,
                dim4035KuoInit,
                dim4036KuoInit,
                dim4037KuoInit,
                dim4038KuoInit,
                dim4039KuoInit,
                dim4040KuoInit,
                dim4041KuoInit,
                dim4042KuoInit,
                dim4043KuoInit,
                dim4044KuoInit,
                dim4045KuoInit,
                dim4046KuoInit,
                dim4047KuoInit,
                dim4048KuoInit,
                dim4049KuoInit,
                dim4050KuoInit,
                dim4051KuoInit,
                dim4052KuoInit,
                dim4053KuoInit,
                dim4054KuoInit,
                dim4055KuoInit,
                dim4056KuoInit,
                dim4057KuoInit,
                dim4058KuoInit,
                dim4059KuoInit,
                dim4060KuoInit,
                dim4061KuoInit,
                dim4062KuoInit,
                dim4063KuoInit,
                dim4064KuoInit,
                dim4065KuoInit,
                dim4066KuoInit,
                dim4067KuoInit,
                dim4068KuoInit,
                dim4069KuoInit,
                dim4070KuoInit,
                dim4071KuoInit,
                dim4072KuoInit,
                dim4073KuoInit,
                dim4074KuoInit,
                dim4075KuoInit,
                dim4076KuoInit,
                dim4077KuoInit,
                dim4078KuoInit,
                dim4079KuoInit,
                dim4080KuoInit,
                dim4081KuoInit,
                dim4082KuoInit,
                dim4083KuoInit,
                dim4084KuoInit,
                dim4085KuoInit,
                dim4086KuoInit,
                dim4087KuoInit,
                dim4088KuoInit,
                dim4089KuoInit,
                dim4090KuoInit,
                dim4091KuoInit,
                dim4092KuoInit,
                dim4093KuoInit,
                dim4094KuoInit,
                dim4095KuoInit,
                dim4096KuoInit,
                dim4097KuoInit,
                dim4098KuoInit,
                dim4099KuoInit,
                dim4100KuoInit,
                dim4101KuoInit,
                dim4102KuoInit,
                dim4103KuoInit,
                dim4104KuoInit,
                dim4105KuoInit,
                dim4106KuoInit,
                dim4107KuoInit,
                dim4108KuoInit,
                dim4109KuoInit,
                dim4110KuoInit,
                dim4111KuoInit,
                dim4112KuoInit,
                dim4113KuoInit,
                dim4114KuoInit,
                dim4115KuoInit,
                dim4116KuoInit,
                dim4117KuoInit,
                dim4118KuoInit,
                dim4119KuoInit,
                dim4120KuoInit,
                dim4121KuoInit,
                dim4122KuoInit,
                dim4123KuoInit,
                dim4124KuoInit,
                dim4125KuoInit,
                dim4126KuoInit,
                dim4127KuoInit,
                dim4128KuoInit,
                dim4129KuoInit,
                dim4130KuoInit,
                dim4131KuoInit,
                dim4132KuoInit,
                dim4133KuoInit,
                dim4134KuoInit,
                dim4135KuoInit,
                dim4136KuoInit,
                dim4137KuoInit,
                dim4138KuoInit,
                dim4139KuoInit,
                dim4140KuoInit,
                dim4141KuoInit,
                dim4142KuoInit,
                dim4143KuoInit,
                dim4144KuoInit,
                dim4145KuoInit,
                dim4146KuoInit,
                dim4147KuoInit,
                dim4148KuoInit,
                dim4149KuoInit,
                dim4150KuoInit,
                dim4151KuoInit,
                dim4152KuoInit,
                dim4153KuoInit,
                dim4154KuoInit,
                dim4155KuoInit,
                dim4156KuoInit,
                dim4157KuoInit,
                dim4158KuoInit,
                dim4159KuoInit,
                dim4160KuoInit,
                dim4161KuoInit,
                dim4162KuoInit,
                dim4163KuoInit,
                dim4164KuoInit,
                dim4165KuoInit,
                dim4166KuoInit,
                dim4167KuoInit,
                dim4168KuoInit,
                dim4169KuoInit,
                dim4170KuoInit,
                dim4171KuoInit,
                dim4172KuoInit,
                dim4173KuoInit,
                dim4174KuoInit,
                dim4175KuoInit,
                dim4176KuoInit,
                dim4177KuoInit,
                dim4178KuoInit,
                dim4179KuoInit,
                dim4180KuoInit,
                dim4181KuoInit,
                dim4182KuoInit,
                dim4183KuoInit,
                dim4184KuoInit,
                dim4185KuoInit,
                dim4186KuoInit,
                dim4187KuoInit,
                dim4188KuoInit,
                dim4189KuoInit,
                dim4190KuoInit,
                dim4191KuoInit,
                dim4192KuoInit,
                dim4193KuoInit,
                dim4194KuoInit,
                dim4195KuoInit,
                dim4196KuoInit,
                dim4197KuoInit,
                dim4198KuoInit,
                dim4199KuoInit,
                dim4200KuoInit,
                dim4201KuoInit,
                dim4202KuoInit,
                dim4203KuoInit,
                dim4204KuoInit,
                dim4205KuoInit,
                dim4206KuoInit,
                dim4207KuoInit,
                dim4208KuoInit,
                dim4209KuoInit,
                dim4210KuoInit,
                dim4211KuoInit,
                dim4212KuoInit,
                dim4213KuoInit,
                dim4214KuoInit,
                dim4215KuoInit,
                dim4216KuoInit,
                dim4217KuoInit,
                dim4218KuoInit,
                dim4219KuoInit,
                dim4220KuoInit,
                dim4221KuoInit,
                dim4222KuoInit,
                dim4223KuoInit,
                dim4224KuoInit,
                dim4225KuoInit,
                dim4226KuoInit,
                dim4227KuoInit,
                dim4228KuoInit,
                dim4229KuoInit,
                dim4230KuoInit,
                dim4231KuoInit,
                dim4232KuoInit,
                dim4233KuoInit,
                dim4234KuoInit,
                dim4235KuoInit,
                dim4236KuoInit,
                dim4237KuoInit,
                dim4238KuoInit,
                dim4239KuoInit,
                dim4240KuoInit,
                dim4241KuoInit,
                dim4242KuoInit,
                dim4243KuoInit,
                dim4244KuoInit,
                dim4245KuoInit,
                dim4246KuoInit,
                dim4247KuoInit,
                dim4248KuoInit,
                dim4249KuoInit,
                dim4250KuoInit,
                dim4251KuoInit,
                dim4252KuoInit,
                dim4253KuoInit,
                dim4254KuoInit,
                dim4255KuoInit,
                dim4256KuoInit,
                dim4257KuoInit,
                dim4258KuoInit,
                dim4259KuoInit,
                dim4260KuoInit,
                dim4261KuoInit,
                dim4262KuoInit,
                dim4263KuoInit,
                dim4264KuoInit,
                dim4265KuoInit,
                dim4266KuoInit,
                dim4267KuoInit,
                dim4268KuoInit,
                dim4269KuoInit,
                dim4270KuoInit,
                dim4271KuoInit,
                dim4272KuoInit,
                dim4273KuoInit,
                dim4274KuoInit,
                dim4275KuoInit,
                dim4276KuoInit,
                dim4277KuoInit,
                dim4278KuoInit,
                dim4279KuoInit,
                dim4280KuoInit,
                dim4281KuoInit,
                dim4282KuoInit,
                dim4283KuoInit,
                dim4284KuoInit,
                dim4285KuoInit,
                dim4286KuoInit,
                dim4287KuoInit,
                dim4288KuoInit,
                dim4289KuoInit,
                dim4290KuoInit,
                dim4291KuoInit,
                dim4292KuoInit,
                dim4293KuoInit,
                dim4294KuoInit,
                dim4295KuoInit,
                dim4296KuoInit,
                dim4297KuoInit,
                dim4298KuoInit,
                dim4299KuoInit,
                dim4300KuoInit,
                dim4301KuoInit,
                dim4302KuoInit,
                dim4303KuoInit,
                dim4304KuoInit,
                dim4305KuoInit,
                dim4306KuoInit,
                dim4307KuoInit,
                dim4308KuoInit,
                dim4309KuoInit,
                dim4310KuoInit,
                dim4311KuoInit,
                dim4312KuoInit,
                dim4313KuoInit,
                dim4314KuoInit,
                dim4315KuoInit,
                dim4316KuoInit,
                dim4317KuoInit,
                dim4318KuoInit,
                dim4319KuoInit,
                dim4320KuoInit,
                dim4321KuoInit,
                dim4322KuoInit,
                dim4323KuoInit,
                dim4324KuoInit,
                dim4325KuoInit,
                dim4326KuoInit,
                dim4327KuoInit,
                dim4328KuoInit,
                dim4329KuoInit,
                dim4330KuoInit,
                dim4331KuoInit,
                dim4332KuoInit,
                dim4333KuoInit,
                dim4334KuoInit,
                dim4335KuoInit,
                dim4336KuoInit,
                dim4337KuoInit,
                dim4338KuoInit,
                dim4339KuoInit,
                dim4340KuoInit,
                dim4341KuoInit,
                dim4342KuoInit,
                dim4343KuoInit,
                dim4344KuoInit,
                dim4345KuoInit,
                dim4346KuoInit,
                dim4347KuoInit,
                dim4348KuoInit,
                dim4349KuoInit,
                dim4350KuoInit,
                dim4351KuoInit,
                dim4352KuoInit,
                dim4353KuoInit,
                dim4354KuoInit,
                dim4355KuoInit,
                dim4356KuoInit,
                dim4357KuoInit,
                dim4358KuoInit,
                dim4359KuoInit,
                dim4360KuoInit,
                dim4361KuoInit,
                dim4362KuoInit,
                dim4363KuoInit,
                dim4364KuoInit,
                dim4365KuoInit,
                dim4366KuoInit,
                dim4367KuoInit,
                dim4368KuoInit,
                dim4369KuoInit,
                dim4370KuoInit,
                dim4371KuoInit,
                dim4372KuoInit,
                dim4373KuoInit,
                dim4374KuoInit,
                dim4375KuoInit,
                dim4376KuoInit,
                dim4377KuoInit,
                dim4378KuoInit,
                dim4379KuoInit,
                dim4380KuoInit,
                dim4381KuoInit,
                dim4382KuoInit,
                dim4383KuoInit,
                dim4384KuoInit,
                dim4385KuoInit,
                dim4386KuoInit,
                dim4387KuoInit,
                dim4388KuoInit,
                dim4389KuoInit,
                dim4390KuoInit,
                dim4391KuoInit,
                dim4392KuoInit,
                dim4393KuoInit,
                dim4394KuoInit,
                dim4395KuoInit,
                dim4396KuoInit,
                dim4397KuoInit,
                dim4398KuoInit,
                dim4399KuoInit,
                dim4400KuoInit,
                dim4401KuoInit,
                dim4402KuoInit,
                dim4403KuoInit,
                dim4404KuoInit,
                dim4405KuoInit,
                dim4406KuoInit,
                dim4407KuoInit,
                dim4408KuoInit,
                dim4409KuoInit,
                dim4410KuoInit,
                dim4411KuoInit,
                dim4412KuoInit,
                dim4413KuoInit,
                dim4414KuoInit,
                dim4415KuoInit,
                dim4416KuoInit,
                dim4417KuoInit,
                dim4418KuoInit,
                dim4419KuoInit,
                dim4420KuoInit,
                dim4421KuoInit,
                dim4422KuoInit,
                dim4423KuoInit,
                dim4424KuoInit,
                dim4425KuoInit,
                dim4426KuoInit,
                dim4427KuoInit,
                dim4428KuoInit,
                dim4429KuoInit,
                dim4430KuoInit,
                dim4431KuoInit,
                dim4432KuoInit,
                dim4433KuoInit,
                dim4434KuoInit,
                dim4435KuoInit,
                dim4436KuoInit,
                dim4437KuoInit,
                dim4438KuoInit,
                dim4439KuoInit,
                dim4440KuoInit,
                dim4441KuoInit,
                dim4442KuoInit,
                dim4443KuoInit,
                dim4444KuoInit,
                dim4445KuoInit,
                dim4446KuoInit,
                dim4447KuoInit,
                dim4448KuoInit,
                dim4449KuoInit,
                dim4450KuoInit,
                dim4451KuoInit,
                dim4452KuoInit,
                dim4453KuoInit,
                dim4454KuoInit,
                dim4455KuoInit,
                dim4456KuoInit,
                dim4457KuoInit,
                dim4458KuoInit,
                dim4459KuoInit,
                dim4460KuoInit,
                dim4461KuoInit,
                dim4462KuoInit,
                dim4463KuoInit,
                dim4464KuoInit,
                dim4465KuoInit,
                dim4466KuoInit,
                dim4467KuoInit,
                dim4468KuoInit,
                dim4469KuoInit,
                dim4470KuoInit,
                dim4471KuoInit,
                dim4472KuoInit,
                dim4473KuoInit,
                dim4474KuoInit,
                dim4475KuoInit,
                dim4476KuoInit,
                dim4477KuoInit,
                dim4478KuoInit,
                dim4479KuoInit,
                dim4480KuoInit,
                dim4481KuoInit,
                dim4482KuoInit,
                dim4483KuoInit,
                dim4484KuoInit,
                dim4485KuoInit,
                dim4486KuoInit,
                dim4487KuoInit,
                dim4488KuoInit,
                dim4489KuoInit,
                dim4490KuoInit,
                dim4491KuoInit,
                dim4492KuoInit,
                dim4493KuoInit,
                dim4494KuoInit,
                dim4495KuoInit,
                dim4496KuoInit,
                dim4497KuoInit,
                dim4498KuoInit,
                dim4499KuoInit,
                dim4500KuoInit,
                dim4501KuoInit,
                dim4502KuoInit,
                dim4503KuoInit,
                dim4504KuoInit,
                dim4505KuoInit,
                dim4506KuoInit,
                dim4507KuoInit,
                dim4508KuoInit,
                dim4509KuoInit,
                dim4510KuoInit,
                dim4511KuoInit,
                dim4512KuoInit,
                dim4513KuoInit,
                dim4514KuoInit,
                dim4515KuoInit,
                dim4516KuoInit,
                dim4517KuoInit,
                dim4518KuoInit,
                dim4519KuoInit,
                dim4520KuoInit,
                dim4521KuoInit,
                dim4522KuoInit,
                dim4523KuoInit,
                dim4524KuoInit,
                dim4525KuoInit,
                dim4526KuoInit,
                dim4527KuoInit,
                dim4528KuoInit,
                dim4529KuoInit,
                dim4530KuoInit,
                dim4531KuoInit,
                dim4532KuoInit,
                dim4533KuoInit,
                dim4534KuoInit,
                dim4535KuoInit,
                dim4536KuoInit,
                dim4537KuoInit,
                dim4538KuoInit,
                dim4539KuoInit,
                dim4540KuoInit,
                dim4541KuoInit,
                dim4542KuoInit,
                dim4543KuoInit,
                dim4544KuoInit,
                dim4545KuoInit,
                dim4546KuoInit,
                dim4547KuoInit,
                dim4548KuoInit,
                dim4549KuoInit,
                dim4550KuoInit,
                dim4551KuoInit,
                dim4552KuoInit,
                dim4553KuoInit,
                dim4554KuoInit,
                dim4555KuoInit,
                dim4556KuoInit,
                dim4557KuoInit,
                dim4558KuoInit,
                dim4559KuoInit,
                dim4560KuoInit,
                dim4561KuoInit,
                dim4562KuoInit,
                dim4563KuoInit,
                dim4564KuoInit,
                dim4565KuoInit,
                dim4566KuoInit,
                dim4567KuoInit,
                dim4568KuoInit,
                dim4569KuoInit,
                dim4570KuoInit,
                dim4571KuoInit,
                dim4572KuoInit,
                dim4573KuoInit,
                dim4574KuoInit,
                dim4575KuoInit,
                dim4576KuoInit,
                dim4577KuoInit,
                dim4578KuoInit,
                dim4579KuoInit,
                dim4580KuoInit,
                dim4581KuoInit,
                dim4582KuoInit,
                dim4583KuoInit,
                dim4584KuoInit,
                dim4585KuoInit,
                dim4586KuoInit,
                dim4587KuoInit,
                dim4588KuoInit,
                dim4589KuoInit,
                dim4590KuoInit,
                dim4591KuoInit,
                dim4592KuoInit,
                dim4593KuoInit,
                dim4594KuoInit,
                dim4595KuoInit,
                dim4596KuoInit,
                dim4597KuoInit,
                dim4598KuoInit,
                dim4599KuoInit,
                dim4600KuoInit,
                dim4601KuoInit,
                dim4602KuoInit,
                dim4603KuoInit,
                dim4604KuoInit,
                dim4605KuoInit,
                dim4606KuoInit,
                dim4607KuoInit,
                dim4608KuoInit,
                dim4609KuoInit,
                dim4610KuoInit,
                dim4611KuoInit,
                dim4612KuoInit,
                dim4613KuoInit,
                dim4614KuoInit,
                dim4615KuoInit,
                dim4616KuoInit,
                dim4617KuoInit,
                dim4618KuoInit,
                dim4619KuoInit,
                dim4620KuoInit,
                dim4621KuoInit,
                dim4622KuoInit,
                dim4623KuoInit,
                dim4624KuoInit,
                dim4625KuoInit,
                dim4626KuoInit,
                dim4627KuoInit,
                dim4628KuoInit,
                dim4629KuoInit,
                dim4630KuoInit,
                dim4631KuoInit,
                dim4632KuoInit,
                dim4633KuoInit,
                dim4634KuoInit,
                dim4635KuoInit,
                dim4636KuoInit,
                dim4637KuoInit,
                dim4638KuoInit,
                dim4639KuoInit,
                dim4640KuoInit,
                dim4641KuoInit,
                dim4642KuoInit,
                dim4643KuoInit,
                dim4644KuoInit,
                dim4645KuoInit,
                dim4646KuoInit,
                dim4647KuoInit,
                dim4648KuoInit,
                dim4649KuoInit,
                dim4650KuoInit,
                dim4651KuoInit,
                dim4652KuoInit,
                dim4653KuoInit,
                dim4654KuoInit,
                dim4655KuoInit,
                dim4656KuoInit,
                dim4657KuoInit,
                dim4658KuoInit,
                dim4659KuoInit,
                dim4660KuoInit,
                dim4661KuoInit,
                dim4662KuoInit,
                dim4663KuoInit,
                dim4664KuoInit,
                dim4665KuoInit,
                dim4666KuoInit,
                dim4667KuoInit,
                dim4668KuoInit,
                dim4669KuoInit,
                dim4670KuoInit,
                dim4671KuoInit,
                dim4672KuoInit,
                dim4673KuoInit,
                dim4674KuoInit,
                dim4675KuoInit,
                dim4676KuoInit,
                dim4677KuoInit,
                dim4678KuoInit,
                dim4679KuoInit,
                dim4680KuoInit,
                dim4681KuoInit,
                dim4682KuoInit,
                dim4683KuoInit,
                dim4684KuoInit,
                dim4685KuoInit,
                dim4686KuoInit,
                dim4687KuoInit,
                dim4688KuoInit,
                dim4689KuoInit,
                dim4690KuoInit,
                dim4691KuoInit,
                dim4692KuoInit,
                dim4693KuoInit,
                dim4694KuoInit,
                dim4695KuoInit,
                dim4696KuoInit,
                dim4697KuoInit,
                dim4698KuoInit,
                dim4699KuoInit,
                dim4700KuoInit,
                dim4701KuoInit,
                dim4702KuoInit,
                dim4703KuoInit,
                dim4704KuoInit,
                dim4705KuoInit,
                dim4706KuoInit,
                dim4707KuoInit,
                dim4708KuoInit,
                dim4709KuoInit,
                dim4710KuoInit,
                dim4711KuoInit,
                dim4712KuoInit,
                dim4713KuoInit,
                dim4714KuoInit,
                dim4715KuoInit,
                dim4716KuoInit,
                dim4717KuoInit,
                dim4718KuoInit,
                dim4719KuoInit,
                dim4720KuoInit,
                dim4721KuoInit,
                dim4722KuoInit,
                dim4723KuoInit,
                dim4724KuoInit,
                dim4725KuoInit,
                dim4726KuoInit,
                dim4727KuoInit,
                dim4728KuoInit,
                dim4729KuoInit,
                dim4730KuoInit,
                dim4731KuoInit,
                dim4732KuoInit,
                dim4733KuoInit,
                dim4734KuoInit,
                dim4735KuoInit,
                dim4736KuoInit,
                dim4737KuoInit,
                dim4738KuoInit,
                dim4739KuoInit,
                dim4740KuoInit,
                dim4741KuoInit,
                dim4742KuoInit,
                dim4743KuoInit,
                dim4744KuoInit,
                dim4745KuoInit,
                dim4746KuoInit,
                dim4747KuoInit,
                dim4748KuoInit,
                dim4749KuoInit,
                dim4750KuoInit,
                dim4751KuoInit,
                dim4752KuoInit,
                dim4753KuoInit,
                dim4754KuoInit,
                dim4755KuoInit,
                dim4756KuoInit,
                dim4757KuoInit,
                dim4758KuoInit,
                dim4759KuoInit,
                dim4760KuoInit,
                dim4761KuoInit,
                dim4762KuoInit,
                dim4763KuoInit,
                dim4764KuoInit,
                dim4765KuoInit,
                dim4766KuoInit,
                dim4767KuoInit,
                dim4768KuoInit,
                dim4769KuoInit,
                dim4770KuoInit,
                dim4771KuoInit,
                dim4772KuoInit,
                dim4773KuoInit,
                dim4774KuoInit,
                dim4775KuoInit,
                dim4776KuoInit,
                dim4777KuoInit,
                dim4778KuoInit,
                dim4779KuoInit,
                dim4780KuoInit,
                dim4781KuoInit,
                dim4782KuoInit,
                dim4783KuoInit,
                dim4784KuoInit,
                dim4785KuoInit,
                dim4786KuoInit,
                dim4787KuoInit,
                dim4788KuoInit,
                dim4789KuoInit,
                dim4790KuoInit,
                dim4791KuoInit,
                dim4792KuoInit,
                dim4793KuoInit,
                dim4794KuoInit,
                dim4795KuoInit,
                dim4796KuoInit,
                dim4797KuoInit,
                dim4798KuoInit,
                dim4799KuoInit,
                dim4800KuoInit,
                dim4801KuoInit,
                dim4802KuoInit,
                dim4803KuoInit,
                dim4804KuoInit,
                dim4805KuoInit,
                dim4806KuoInit,
                dim4807KuoInit,
                dim4808KuoInit,
                dim4809KuoInit,
                dim4810KuoInit,
                dim4811KuoInit,
                dim4812KuoInit,
                dim4813KuoInit,
                dim4814KuoInit,
                dim4815KuoInit,
                dim4816KuoInit,
                dim4817KuoInit,
                dim4818KuoInit,
                dim4819KuoInit,
                dim4820KuoInit,
                dim4821KuoInit,
                dim4822KuoInit,
                dim4823KuoInit,
                dim4824KuoInit,
                dim4825KuoInit,
                dim4826KuoInit,
                dim4827KuoInit,
                dim4828KuoInit,
                dim4829KuoInit,
                dim4830KuoInit,
                dim4831KuoInit,
                dim4832KuoInit,
                dim4833KuoInit,
                dim4834KuoInit,
                dim4835KuoInit,
                dim4836KuoInit,
                dim4837KuoInit,
                dim4838KuoInit,
                dim4839KuoInit,
                dim4840KuoInit,
                dim4841KuoInit,
                dim4842KuoInit,
                dim4843KuoInit,
                dim4844KuoInit,
                dim4845KuoInit,
                dim4846KuoInit,
                dim4847KuoInit,
                dim4848KuoInit,
                dim4849KuoInit,
                dim4850KuoInit,
                dim4851KuoInit,
                dim4852KuoInit,
                dim4853KuoInit,
                dim4854KuoInit,
                dim4855KuoInit,
                dim4856KuoInit,
                dim4857KuoInit,
                dim4858KuoInit,
                dim4859KuoInit,
                dim4860KuoInit,
                dim4861KuoInit,
                dim4862KuoInit,
                dim4863KuoInit,
                dim4864KuoInit,
                dim4865KuoInit,
                dim4866KuoInit,
                dim4867KuoInit,
                dim4868KuoInit,
                dim4869KuoInit,
                dim4870KuoInit,
                dim4871KuoInit,
                dim4872KuoInit,
                dim4873KuoInit,
                dim4874KuoInit,
                dim4875KuoInit,
                dim4876KuoInit,
                dim4877KuoInit,
                dim4878KuoInit,
                dim4879KuoInit,
                dim4880KuoInit,
                dim4881KuoInit,
                dim4882KuoInit,
                dim4883KuoInit,
                dim4884KuoInit,
                dim4885KuoInit,
                dim4886KuoInit,
                dim4887KuoInit,
                dim4888KuoInit,
                dim4889KuoInit,
                dim4890KuoInit,
                dim4891KuoInit,
                dim4892KuoInit,
                dim4893KuoInit,
                dim4894KuoInit,
                dim4895KuoInit,
                dim4896KuoInit,
                dim4897KuoInit,
                dim4898KuoInit,
                dim4899KuoInit,
                dim4900KuoInit,
                dim4901KuoInit,
                dim4902KuoInit,
                dim4903KuoInit,
                dim4904KuoInit,
                dim4905KuoInit,
                dim4906KuoInit,
                dim4907KuoInit,
                dim4908KuoInit,
                dim4909KuoInit,
                dim4910KuoInit,
                dim4911KuoInit,
                dim4912KuoInit,
                dim4913KuoInit,
                dim4914KuoInit,
                dim4915KuoInit,
                dim4916KuoInit,
                dim4917KuoInit,
                dim4918KuoInit,
                dim4919KuoInit,
                dim4920KuoInit,
                dim4921KuoInit,
                dim4922KuoInit,
                dim4923KuoInit,
                dim4924KuoInit,
                dim4925KuoInit,

            };

            const unsigned long dim1Kuo2Init[]  =   {   1   ,0 };
            const unsigned long dim2Kuo2Init[]  =   {   1   ,   1   ,0 };
            const unsigned long dim3Kuo2Init[]  =   {   1   ,   1   ,   1   ,0 };
            const unsigned long dim4Kuo2Init[]  =   {   1   ,   3   ,   1   ,0 };
            const unsigned long dim5Kuo2Init[]  =   {   1   ,   3   ,   7   ,   1   ,0 };
            const unsigned long dim6Kuo2Init[]  =   {   1   ,   1   ,   3   ,   7   ,0 };
            const unsigned long dim7Kuo2Init[]  =   {   1   ,   3   ,   1   ,   7   ,   23  ,0 };
            const unsigned long dim8Kuo2Init[]  =   {   1   ,   1   ,   5   ,   1   ,   27  ,0 };
            const unsigned long dim9Kuo2Init[]  =   {   1   ,   3   ,   5   ,   3   ,   25  ,0 };
            const unsigned long dim10Kuo2Init[] =   {   1   ,   1   ,   5   ,   9   ,   21  ,0 };
            const unsigned long dim11Kuo2Init[] =   {   1   ,   1   ,   3   ,   13  ,   27  ,0 };
            const unsigned long dim12Kuo2Init[] =   {   1   ,   3   ,   7   ,   11  ,   3   ,0 };
            const unsigned long dim13Kuo2Init[] =   {   1   ,   3   ,   1   ,   13  ,   31  ,   47  ,0 };
            const unsigned long dim14Kuo2Init[] =   {   1   ,   3   ,   3   ,   5   ,   13  ,   17  ,0 };
            const unsigned long dim15Kuo2Init[] =   {   1   ,   1   ,   7   ,   3   ,   13  ,   51  ,0 };
            const unsigned long dim16Kuo2Init[] =   {   1   ,   3   ,   3   ,   11  ,   7   ,   63  ,0 };
            const unsigned long dim17Kuo2Init[] =   {   1   ,   1   ,   1   ,   7   ,   3   ,   53  ,0 };
            const unsigned long dim18Kuo2Init[] =   {   1   ,   3   ,   5   ,   11  ,   3   ,   43  ,0 };
            const unsigned long dim19Kuo2Init[] =   {   1   ,   1   ,   7   ,   5   ,   19  ,   59  ,   59  ,0 };
            const unsigned long dim20Kuo2Init[] =   {   1   ,   1   ,   7   ,   9   ,   15  ,   25  ,   63  ,0 };
            const unsigned long dim21Kuo2Init[] =   {   1   ,   3   ,   7   ,   7   ,   27  ,   5   ,   17  ,0 };
            const unsigned long dim22Kuo2Init[] =   {   1   ,   3   ,   1   ,   5   ,   7   ,   51  ,   69  ,0 };
            const unsigned long dim23Kuo2Init[] =   {   1   ,   3   ,   3   ,   11  ,   23  ,   5   ,   35  ,0 };
            const unsigned long dim24Kuo2Init[] =   {   1   ,   1   ,   5   ,   1   ,   3   ,   17  ,   59  ,0 };
            const unsigned long dim25Kuo2Init[] =   {   1   ,   3   ,   7   ,   7   ,   3   ,   51  ,   57  ,0 };
            const unsigned long dim26Kuo2Init[] =   {   1   ,   3   ,   7   ,   11  ,   17  ,   1   ,   17  ,0 };
            const unsigned long dim27Kuo2Init[] =   {   1   ,   1   ,   1   ,   11  ,   25  ,   45  ,   3   ,0 };
            const unsigned long dim28Kuo2Init[] =   {   1   ,   1   ,   5   ,   15  ,   23  ,   53  ,   79  ,0 };
            const unsigned long dim29Kuo2Init[] =   {   1   ,   3   ,   3   ,   3   ,   19  ,   31  ,   39  ,0 };
            const unsigned long dim30Kuo2Init[] =   {   1   ,   1   ,   1   ,   9   ,   9   ,   21  ,   23  ,0 };
            const unsigned long dim31Kuo2Init[] =   {   1   ,   1   ,   3   ,   5   ,   31  ,   13  ,   97  ,0 };
            const unsigned long dim32Kuo2Init[] =   {   1   ,   1   ,   1   ,   15  ,   9   ,   39  ,   3   ,0 };
            const unsigned long dim33Kuo2Init[] =   {   1   ,   3   ,   1   ,   1   ,   27  ,   55  ,   37  ,0 };
            const unsigned long dim34Kuo2Init[] =   {   1   ,   3   ,   7   ,   11  ,   21  ,   49  ,   17  ,0 };
            const unsigned long dim35Kuo2Init[] =   {   1   ,   3   ,   3   ,   9   ,   31  ,   37  ,   95  ,0 };
            const unsigned long dim36Kuo2Init[] =   {   1   ,   1   ,   1   ,   1   ,   7   ,   35  ,   95  ,0 };
            const unsigned long dim37Kuo2Init[] =   {   1   ,   1   ,   5   ,   15  ,   15  ,   61  ,   11  ,   211 ,0 };
            const unsigned long dim38Kuo2Init[] =   {   1   ,   1   ,   7   ,   7   ,   29  ,   33  ,   127 ,   171 ,0 };
            const unsigned long dim39Kuo2Init[] =   {   1   ,   3   ,   5   ,   9   ,   17  ,   27  ,   85  ,   95  ,0 };
            const unsigned long dim40Kuo2Init[] =   {   1   ,   3   ,   7   ,   1   ,   25  ,   43  ,   35  ,   21  ,0 };
            const unsigned long dim41Kuo2Init[] =   {   1   ,   3   ,   5   ,   15  ,   9   ,   17  ,   43  ,   253 ,0 };
            const unsigned long dim42Kuo2Init[] =   {   1   ,   1   ,   5   ,   11  ,   27  ,   45  ,   37  ,   147 ,0 };
            const unsigned long dim43Kuo2Init[] =   {   1   ,   1   ,   7   ,   5   ,   7   ,   35  ,   29  ,   191 ,0 };
            const unsigned long dim44Kuo2Init[] =   {   1   ,   3   ,   1   ,   1   ,   15  ,   33  ,   21  ,   125 ,0 };
            const unsigned long dim45Kuo2Init[] =   {   1   ,   1   ,   1   ,   7   ,   31  ,   41  ,   33  ,   11  ,0 };
            const unsigned long dim46Kuo2Init[] =   {   1   ,   1   ,   7   ,   7   ,   27  ,   9   ,   111 ,   209 ,0 };
            const unsigned long dim47Kuo2Init[] =   {   1   ,   3   ,   7   ,   15  ,   1   ,   57  ,   63  ,   43  ,0 };
            const unsigned long dim48Kuo2Init[] =   {   1   ,   3   ,   3   ,   13  ,   1   ,   51  ,   5   ,   51  ,0 };
            const unsigned long dim49Kuo2Init[] =   {   1   ,   3   ,   5   ,   3   ,   27  ,   17  ,   103 ,   247 ,0 };
            const unsigned long dim50Kuo2Init[] =   {   1   ,   3   ,   7   ,   7   ,   11  ,   27  ,   51  ,   103 ,0 };
            const unsigned long dim51Kuo2Init[] =   {   1   ,   3   ,   3   ,   1   ,   23  ,   21  ,   65  ,   135 ,0 };
            const unsigned long dim52Kuo2Init[] =   {   1   ,   1   ,   7   ,   9   ,   1   ,   7   ,   71  ,   251 ,0 };
            const unsigned long dim53Kuo2Init[] =   {   1   ,   3   ,   1   ,   5   ,   11  ,   31  ,   43  ,   129 ,   225 ,0 };
            const unsigned long dim54Kuo2Init[] =   {   1   ,   1   ,   1   ,   5   ,   15  ,   15  ,   127 ,   225 ,   439 ,0 };
            const unsigned long dim55Kuo2Init[] =   {   1   ,   1   ,   3   ,   13  ,   27  ,   45  ,   103 ,   65  ,   313 ,0 };
            const unsigned long dim56Kuo2Init[] =   {   1   ,   3   ,   3   ,   15  ,   13  ,   31  ,   27  ,   231 ,   183 ,0 };
            const unsigned long dim57Kuo2Init[] =   {   1   ,   1   ,   3   ,   7   ,   23  ,   5   ,   15  ,   75  ,   159 ,0 };
            const unsigned long dim58Kuo2Init[] =   {   1   ,   1   ,   5   ,   15  ,   21  ,   3   ,   99  ,   23  ,   465 ,0 };
            const unsigned long dim59Kuo2Init[] =   {   1   ,   3   ,   7   ,   5   ,   23  ,   47  ,   119 ,   103 ,   271 ,0 };
            const unsigned long dim60Kuo2Init[] =   {   1   ,   3   ,   1   ,   11  ,   25  ,   13  ,   83  ,   5   ,   21  ,0 };
            const unsigned long dim61Kuo2Init[] =   {   1   ,   3   ,   5   ,   3   ,   7   ,   41  ,   123 ,   87  ,   243 ,0 };
            const unsigned long dim62Kuo2Init[] =   {   1   ,   1   ,   3   ,   11  ,   3   ,   7   ,   91  ,   227 ,   349 ,0 };
            const unsigned long dim63Kuo2Init[] =   {   1   ,   1   ,   1   ,   9   ,   29  ,   23  ,   31  ,   175 ,   145 ,0 };
            const unsigned long dim64Kuo2Init[] =   {   1   ,   1   ,   7   ,   7   ,   13  ,   25  ,   67  ,   59  ,   109 ,0 };
            const unsigned long dim65Kuo2Init[] =   {   1   ,   3   ,   7   ,   11  ,   25  ,   49  ,   23  ,   55  ,   135 ,0 };
            const unsigned long dim66Kuo2Init[] =   {   1   ,   3   ,   5   ,   7   ,   27  ,   61  ,   5   ,   193 ,   375 ,0 };
            const unsigned long dim67Kuo2Init[]     =   {   1   ,   1   ,   5   ,   9   ,   23  ,   27  ,   41  ,   239 ,   383 ,0 };
            const unsigned long dim68Kuo2Init[]     =   {   1   ,   1   ,   3   ,   15  ,   3   ,   43  ,   81  ,   77  ,   91  ,0 };
            const unsigned long dim69Kuo2Init[]     =   {   1   ,   1   ,   1   ,   13  ,   29  ,   17  ,   23  ,   105 ,   393 ,0 };
            const unsigned long dim70Kuo2Init[]     =   {   1   ,   3   ,   3   ,   3   ,   27  ,   41  ,   89  ,   79  ,   295 ,0 };
            const unsigned long dim71Kuo2Init[]     =   {   1   ,   3   ,   3   ,   5   ,   27  ,   63  ,   29  ,   189 ,   505 ,0 };
            const unsigned long dim72Kuo2Init[]     =   {   1   ,   1   ,   3   ,   7   ,   23  ,   63  ,   117 ,   75  ,   165 ,0 };
            const unsigned long dim73Kuo2Init[]     =   {   1   ,   1   ,   5   ,   13  ,   1   ,   47  ,   123 ,   159 ,   485 ,0 };
            const unsigned long dim74Kuo2Init[]     =   {   1   ,   3   ,   5   ,   1   ,   23  ,   27  ,   125 ,   45  ,   351 ,0 };
            const unsigned long dim75Kuo2Init[]     =   {   1   ,   3   ,   1   ,   5   ,   11  ,   33  ,   75  ,   17  ,   85  ,0 };
            const unsigned long dim76Kuo2Init[]     =   {   1   ,   3   ,   7   ,   9   ,   23  ,   31  ,   33  ,   91  ,   131 ,0 };
            const unsigned long dim77Kuo2Init[]     =   {   1   ,   1   ,   3   ,   7   ,   23  ,   31  ,   89  ,   243 ,   229 ,0 };
            const unsigned long dim78Kuo2Init[]     =   {   1   ,   1   ,   7   ,   15  ,   3   ,   9   ,   97  ,   59  ,   159 ,0 };
            const unsigned long dim79Kuo2Init[]     =   {   1   ,   3   ,   3   ,   9   ,   23  ,   1   ,   105 ,   217 ,   491 ,0 };
            const unsigned long dim80Kuo2Init[]     =   {   1   ,   3   ,   7   ,   13  ,   19  ,   3   ,   95  ,   149 ,   169 ,0 };
            const unsigned long dim81Kuo2Init[]     =   {   1   ,   1   ,   5   ,   11  ,   29  ,   57  ,   7   ,   69  ,   33  ,0 };
            const unsigned long dim82Kuo2Init[]     =   {   1   ,   1   ,   7   ,   7   ,   23  ,   1   ,   17  ,   3   ,   325 ,0 };
            const unsigned long dim83Kuo2Init[]     =   {   1   ,   3   ,   1   ,   7   ,   23  ,   7   ,   49  ,   251 ,   501 ,0 };
            const unsigned long dim84Kuo2Init[]     =   {   1   ,   1   ,   5   ,   9   ,   13  ,   59  ,   51  ,   15  ,   231 ,0 };
            const unsigned long dim85Kuo2Init[]     =   {   1   ,   1   ,   1   ,   7   ,   13  ,   43  ,   21  ,   227 ,   165 ,0 };
            const unsigned long dim86Kuo2Init[]     =   {   1   ,   1   ,   5   ,   1   ,   17  ,   59  ,   85  ,   13  ,   81  ,0 };
            const unsigned long dim87Kuo2Init[]     =   {   1   ,   1   ,   5   ,   11  ,   25  ,   39  ,   109 ,   149 ,   89  ,0 };
            const unsigned long dim88Kuo2Init[]     =   {   1   ,   3   ,   1   ,   13  ,   3   ,   47  ,   85  ,   1   ,   241 ,0 };
            const unsigned long dim89Kuo2Init[]     =   {   1   ,   1   ,   7   ,   7   ,   29  ,   25  ,   71  ,   13  ,   221 ,0 };
            const unsigned long dim90Kuo2Init[]     =   {   1   ,   1   ,   3   ,   5   ,   1   ,   3   ,   17  ,   241 ,   227 ,0 };
            const unsigned long dim91Kuo2Init[]     =   {   1   ,   3   ,   1   ,   7   ,   3   ,   23  ,   127 ,   209 ,   139 ,0 };
            const unsigned long dim92Kuo2Init[]     =   {   1   ,   1   ,   3   ,   5   ,   31  ,   39  ,   55  ,   193 ,   27  ,0 };
            const unsigned long dim93Kuo2Init[]     =   {   1   ,   3   ,   5   ,   3   ,   5   ,   19  ,   71  ,   161 ,   255 ,0 };
            const unsigned long dim94Kuo2Init[]     =   {   1   ,   3   ,   3   ,   7   ,   17  ,   27  ,   105 ,   13  ,   131 ,0 };
            const unsigned long dim95Kuo2Init[]     =   {   1   ,   1   ,   3   ,   1   ,   23  ,   7   ,   43  ,   171 ,   503 ,0 };
            const unsigned long dim96Kuo2Init[]     =   {   1   ,   3   ,   5   ,   7   ,   7   ,   11  ,   9   ,   243 ,   173 ,0 };
            const unsigned long dim97Kuo2Init[]     =   {   1   ,   1   ,   7   ,   13  ,   23  ,   3   ,   119 ,   243 ,   189 ,0 };
            const unsigned long dim98Kuo2Init[]     =   {   1   ,   1   ,   3   ,   11  ,   17  ,   55  ,   53  ,   159 ,   339 ,0 };
            const unsigned long dim99Kuo2Init[]     =   {   1   ,   3   ,   3   ,   11  ,   5   ,   17  ,   69  ,   103 ,   453 ,0 };
            const unsigned long dim100Kuo2Init[]        =   {   1   ,   3   ,   7   ,   7   ,   9   ,   41  ,   31  ,   165 ,   117 ,0 };
            const unsigned long dim101Kuo2Init[]        =   {   1   ,   1   ,   5   ,   3   ,   5   ,   39  ,   79  ,   141 ,   205 ,   483 ,0 };
            const unsigned long dim102Kuo2Init[]        =   {   1   ,   3   ,   1   ,   3   ,   3   ,   59  ,   57  ,   189 ,   279 ,   769 ,0 };
            const unsigned long dim103Kuo2Init[]        =   {   1   ,   1   ,   3   ,   13  ,   25  ,   21  ,   107 ,   61  ,   329 ,   695 ,0 };
            const unsigned long dim104Kuo2Init[]        =   {   1   ,   3   ,   5   ,   15  ,   3   ,   61  ,   71  ,   161 ,   55  ,   757 ,0 };
            const unsigned long dim105Kuo2Init[]        =   {   1   ,   1   ,   1   ,   11  ,   9   ,   17  ,   121 ,   41  ,   177 ,   261 ,0 };
            const unsigned long dim106Kuo2Init[]        =   {   1   ,   1   ,   5   ,   3   ,   25  ,   49  ,   55  ,   119 ,   171 ,   213 ,0 };
            const unsigned long dim107Kuo2Init[]        =   {   1   ,   3   ,   3   ,   13  ,   21  ,   35  ,   73  ,   33  ,   43  ,   673 ,0 };
            const unsigned long dim108Kuo2Init[]        =   {   1   ,   3   ,   5   ,   7   ,   25  ,   5   ,   7   ,   93  ,   485 ,   635 ,0 };
            const unsigned long dim109Kuo2Init[]        =   {   1   ,   1   ,   5   ,   1   ,   5   ,   55  ,   17  ,   33  ,   67  ,   553 ,0 };
            const unsigned long dim110Kuo2Init[]        =   {   1   ,   1   ,   7   ,   1   ,   15  ,   31  ,   67  ,   219 ,   241 ,   233 ,0 };
            const unsigned long dim111Kuo2Init[]        =   {   1   ,   1   ,   3   ,   7   ,   9   ,   33  ,   77  ,   115 ,   401 ,   617 ,0 };
            const unsigned long dim112Kuo2Init[]        =   {   1   ,   1   ,   3   ,   3   ,   1   ,   11  ,   79  ,   49  ,   327 ,   287 ,0 };
            const unsigned long dim113Kuo2Init[]        =   {   1   ,   3   ,   5   ,   15  ,   19  ,   49  ,   115 ,   73  ,   193 ,   615 ,0 };
            const unsigned long dim114Kuo2Init[]        =   {   1   ,   3   ,   5   ,   13  ,   15  ,   41  ,   17  ,   133 ,   369 ,   783 ,0 };
            const unsigned long dim115Kuo2Init[]        =   {   1   ,   1   ,   7   ,   13  ,   9   ,   15  ,   59  ,   43  ,   129 ,   337 ,0 };
            const unsigned long dim116Kuo2Init[]        =   {   1   ,   1   ,   3   ,   11  ,   13  ,   41  ,   35  ,   111 ,   385 ,   243 ,0 };
            const unsigned long dim117Kuo2Init[]        =   {   1   ,   3   ,   5   ,   11  ,   7   ,   25  ,   79  ,   9   ,   5   ,   159 ,0 };
            const unsigned long dim118Kuo2Init[]        =   {   1   ,   3   ,   3   ,   1   ,   7   ,   17  ,   31  ,   3   ,   27  ,   829 ,0 };
            const unsigned long dim119Kuo2Init[]        =   {   1   ,   3   ,   3   ,   15  ,   27  ,   33  ,   13  ,   29  ,   223 ,   545 ,0 };
            const unsigned long dim120Kuo2Init[]        =   {   1   ,   1   ,   5   ,   1   ,   7   ,   7   ,   31  ,   103 ,   307 ,   315 ,0 };
            const unsigned long dim121Kuo2Init[]        =   {   1   ,   3   ,   7   ,   9   ,   15  ,   9   ,   61  ,   75  ,   99  ,   173 ,0 };
            const unsigned long dim122Kuo2Init[]        =   {   1   ,   3   ,   3   ,   15  ,   21  ,   31  ,   49  ,   45  ,   301 ,   593 ,0 };
            const unsigned long dim123Kuo2Init[]        =   {   1   ,   3   ,   1   ,   11  ,   3   ,   1   ,   37  ,   109 ,   413 ,   517 ,0 };
            const unsigned long dim124Kuo2Init[]        =   {   1   ,   3   ,   1   ,   15  ,   1   ,   47  ,   45  ,   189 ,   331 ,   3   ,0 };
            const unsigned long dim125Kuo2Init[]        =   {   1   ,   3   ,   7   ,   5   ,   3   ,   47  ,   107 ,   141 ,   257 ,   189 ,0 };
            const unsigned long dim126Kuo2Init[]        =   {   1   ,   1   ,   1   ,   11  ,   5   ,   31  ,   49  ,   35  ,   139 ,   895 ,0 };
            const unsigned long dim127Kuo2Init[]        =   {   1   ,   3   ,   3   ,   7   ,   17  ,   23  ,   47  ,   229 ,   225 ,   393 ,0 };
            const unsigned long dim128Kuo2Init[]        =   {   1   ,   1   ,   1   ,   11  ,   17  ,   39  ,   125 ,   223 ,   251 ,   95  ,0 };
            const unsigned long dim129Kuo2Init[]        =   {   1   ,   3   ,   7   ,   9   ,   31  ,   1   ,   29  ,   81  ,   301 ,   309 ,0 };
            const unsigned long dim130Kuo2Init[]        =   {   1   ,   1   ,   1   ,   5   ,   29  ,   3   ,   5   ,   7   ,   481 ,   989 ,0 };
            const unsigned long dim131Kuo2Init[]        =   {   1   ,   1   ,   1   ,   5   ,   19  ,   29  ,   103 ,   103 ,   385 ,   541 ,0 };
            const unsigned long dim132Kuo2Init[]        =   {   1   ,   3   ,   7   ,   13  ,   31  ,   53  ,   27  ,   51  ,   305 ,   345 ,0 };
            const unsigned long dim133Kuo2Init[]        =   {   1   ,   3   ,   5   ,   5   ,   19  ,   51  ,   41  ,   15  ,   31  ,   761 ,0 };
            const unsigned long dim134Kuo2Init[]        =   {   1   ,   1   ,   5   ,   3   ,   25  ,   21  ,   93  ,   205 ,   129 ,   179 ,0 };
            const unsigned long dim135Kuo2Init[]        =   {   1   ,   3   ,   7   ,   11  ,   5   ,   31  ,   29  ,   143 ,   103 ,   227 ,0 };
            const unsigned long dim136Kuo2Init[]        =   {   1   ,   3   ,   1   ,   11  ,   29  ,   47  ,   97  ,   217 ,   13  ,   897 ,0 };
            const unsigned long dim137Kuo2Init[]        =   {   1   ,   1   ,   3   ,   13  ,   15  ,   39  ,   85  ,   181 ,   471 ,   853 ,0 };
            const unsigned long dim138Kuo2Init[]        =   {   1   ,   3   ,   1   ,   11  ,   11  ,   21  ,   5   ,   233 ,   415 ,   897 ,0 };
            const unsigned long dim139Kuo2Init[]        =   {   1   ,   1   ,   5   ,   7   ,   3   ,   19  ,   103 ,   27  ,   353 ,   775 ,0 };
            const unsigned long dim140Kuo2Init[]        =   {   1   ,   3   ,   3   ,   1   ,   21  ,   57  ,   123 ,   27  ,   61  ,   719 ,0 };
            const unsigned long dim141Kuo2Init[]        =   {   1   ,   1   ,   3   ,   13  ,   5   ,   57  ,   87  ,   77  ,   301 ,   633 ,0 };
            const unsigned long dim142Kuo2Init[]        =   {   1   ,   3   ,   7   ,   5   ,   21  ,   59  ,   19  ,   177 ,   147 ,   357 ,0 };
            const unsigned long dim143Kuo2Init[]        =   {   1   ,   1   ,   5   ,   1   ,   5   ,   57  ,   43  ,   249 ,   475 ,   189 ,0 };
            const unsigned long dim144Kuo2Init[]        =   {   1   ,   3   ,   7   ,   9   ,   15  ,   49  ,   29  ,   75  ,   75  ,   839 ,0 };
            const unsigned long dim145Kuo2Init[]        =   {   1   ,   1   ,   3   ,   7   ,   23  ,   47  ,   17  ,   83  ,   275 ,   917 ,0 };
            const unsigned long dim146Kuo2Init[]        =   {   1   ,   3   ,   7   ,   1   ,   23  ,   17  ,   35  ,   237 ,   357 ,   363 ,0 };
            const unsigned long dim147Kuo2Init[]        =   {   1   ,   1   ,   3   ,   1   ,   17  ,   35  ,   85  ,   215 ,   405 ,   469 ,0 };
            const unsigned long dim148Kuo2Init[]        =   {   1   ,   3   ,   1   ,   9   ,   21  ,   21  ,   103 ,   205 ,   79  ,   431 ,0 };
            const unsigned long dim149Kuo2Init[]        =   {   1   ,   1   ,   3   ,   1   ,   7   ,   13  ,   127 ,   233 ,   93  ,   215 ,0 };
            const unsigned long dim150Kuo2Init[]        =   {   1   ,   3   ,   1   ,   11  ,   13  ,   19  ,   87  ,   219 ,   49  ,   133 ,0 };
            const unsigned long dim151Kuo2Init[]        =   {   1   ,   3   ,   7   ,   3   ,   27  ,   51  ,   125 ,   235 ,   159 ,   981 ,0 };
            const unsigned long dim152Kuo2Init[]        =   {   1   ,   1   ,   3   ,   7   ,   29  ,   37  ,   3   ,   189 ,   419 ,   329 ,0 };
            const unsigned long dim153Kuo2Init[]        =   {   1   ,   1   ,   5   ,   13  ,   9   ,   61  ,   125 ,   29  ,   111 ,   751 ,0 };
            const unsigned long dim154Kuo2Init[]        =   {   1   ,   1   ,   3   ,   5   ,   5   ,   43  ,   59  ,   15  ,   445 ,   803 ,0 };
            const unsigned long dim155Kuo2Init[]        =   {   1   ,   1   ,   7   ,   15  ,   25  ,   55  ,   37  ,   231 ,   287 ,   97  ,0 };
            const unsigned long dim156Kuo2Init[]        =   {   1   ,   1   ,   1   ,   1   ,   13  ,   5   ,   97  ,   25  ,   53  ,   579 ,0 };
            const unsigned long dim157Kuo2Init[]        =   {   1   ,   3   ,   5   ,   13  ,   27  ,   63  ,   17  ,   197 ,   285 ,   741 ,0 };
            const unsigned long dim158Kuo2Init[]        =   {   1   ,   1   ,   5   ,   3   ,   7   ,   31  ,   13  ,   167 ,   235 ,   965 ,0 };
            const unsigned long dim159Kuo2Init[]        =   {   1   ,   3   ,   5   ,   7   ,   3   ,   35  ,   97  ,   131 ,   89  ,   313 ,0 };
            const unsigned long dim160Kuo2Init[]        =   {   1   ,   1   ,   3   ,   13  ,   23  ,   3   ,   91  ,   155 ,   249 ,   829 ,0 };
            const unsigned long dim161Kuo2Init[]        =   {   1   ,   3   ,   7   ,   13  ,   1   ,   41  ,   51  ,   233 ,   233 ,   163 ,   1357    ,0 };
            const unsigned long dim162Kuo2Init[]        =   {   1   ,   3   ,   1   ,   3   ,   13  ,   3   ,   55  ,   9   ,   511 ,   441 ,   811 ,0 };
            const unsigned long dim163Kuo2Init[]        =   {   1   ,   3   ,   5   ,   5   ,   11  ,   57  ,   45  ,   65  ,   397 ,   199 ,   1195    ,0 };
            const unsigned long dim164Kuo2Init[]        =   {   1   ,   3   ,   5   ,   11  ,   11  ,   15  ,   95  ,   163 ,   319 ,   727 ,   1793    ,0 };
            const unsigned long dim165Kuo2Init[]        =   {   1   ,   1   ,   5   ,   7   ,   31  ,   47  ,   49  ,   177 ,   229 ,   719 ,   241 ,0 };
            const unsigned long dim166Kuo2Init[]        =   {   1   ,   3   ,   1   ,   13  ,   9   ,   25  ,   69  ,   67  ,   151 ,   369 ,   1205    ,0 };
            const unsigned long dim167Kuo2Init[]        =   {   1   ,   1   ,   7   ,   3   ,   9   ,   39  ,   99  ,   109 ,   111 ,   787 ,   181 ,0 };
            const unsigned long dim168Kuo2Init[]        =   {   1   ,   1   ,   3   ,   11  ,   23  ,   1   ,   49  ,   111 ,   45  ,   293 ,   1603    ,0 };
            const unsigned long dim169Kuo2Init[]        =   {   1   ,   1   ,   1   ,   3   ,   5   ,   45  ,   11  ,   197 ,   357 ,   109 ,   1291    ,0 };
            const unsigned long dim170Kuo2Init[]        =   {   1   ,   3   ,   5   ,   11  ,   5   ,   25  ,   91  ,   91  ,   339 ,   559 ,   311 ,0 };
            const unsigned long dim171Kuo2Init[]        =   {   1   ,   3   ,   3   ,   3   ,   21  ,   37  ,   69  ,   129 ,   205 ,   333 ,   165 ,0 };
            const unsigned long dim172Kuo2Init[]        =   {   1   ,   1   ,   1   ,   15  ,   9   ,   19  ,   115 ,   153 ,   23  ,   381 ,   1727    ,0 };
            const unsigned long dim173Kuo2Init[]        =   {   1   ,   1   ,   3   ,   1   ,   3   ,   35  ,   77  ,   217 ,   133 ,   943 ,   169 ,0 };
            const unsigned long dim174Kuo2Init[]        =   {   1   ,   1   ,   5   ,   9   ,   1   ,   39  ,   67  ,   217 ,   143 ,   529 ,   769 ,0 };
            const unsigned long dim175Kuo2Init[]        =   {   1   ,   1   ,   3   ,   3   ,   9   ,   9   ,   9   ,   203 ,   425 ,   601 ,   1389    ,0 };
            const unsigned long dim176Kuo2Init[]        =   {   1   ,   3   ,   7   ,   3   ,   15  ,   19  ,   109 ,   31  ,   417 ,   127 ,   1109    ,0 };
            const unsigned long dim177Kuo2Init[]        =   {   1   ,   3   ,   1   ,   15  ,   1   ,   37  ,   75  ,   199 ,   419 ,   917 ,   1929    ,0 };
            const unsigned long dim178Kuo2Init[]        =   {   1   ,   3   ,   3   ,   9   ,   21  ,   37  ,   23  ,   165 ,   67  ,   39  ,   1685    ,0 };
            const unsigned long dim179Kuo2Init[]        =   {   1   ,   3   ,   3   ,   7   ,   3   ,   25  ,   75  ,   67  ,   451 ,   889 ,   1015    ,0 };
            const unsigned long dim180Kuo2Init[]        =   {   1   ,   3   ,   1   ,   1   ,   27  ,   37  ,   39  ,   7   ,   411 ,   391 ,   463 ,0 };
            const unsigned long dim181Kuo2Init[]        =   {   1   ,   1   ,   7   ,   15  ,   25  ,   19  ,   55  ,   131 ,   245 ,   887 ,   971 ,0 };
            const unsigned long dim182Kuo2Init[]        =   {   1   ,   3   ,   1   ,   9   ,   19  ,   9   ,   9   ,   31  ,   143 ,   647 ,   39  ,0 };
            const unsigned long dim183Kuo2Init[]        =   {   1   ,   3   ,   7   ,   1   ,   27  ,   7   ,   107 ,   179 ,   327 ,   947 ,   953 ,0 };
            const unsigned long dim184Kuo2Init[]        =   {   1   ,   3   ,   7   ,   13  ,   23  ,   31  ,   65  ,   29  ,   157 ,   67  ,   2009    ,0 };
            const unsigned long dim185Kuo2Init[]        =   {   1   ,   3   ,   5   ,   13  ,   19  ,   31  ,   19  ,   249 ,   385 ,   209 ,   281 ,0 };
            const unsigned long dim186Kuo2Init[]        =   {   1   ,   1   ,   7   ,   15  ,   3   ,   63  ,   1   ,   97  ,   349 ,   1007    ,   1463    ,0 };
            const unsigned long dim187Kuo2Init[]        =   {   1   ,   3   ,   3   ,   11  ,   25  ,   3   ,   21  ,   195 ,   213 ,   1013    ,   637 ,0 };
            const unsigned long dim188Kuo2Init[]        =   {   1   ,   3   ,   1   ,   1   ,   27  ,   7   ,   95  ,   215 ,   473 ,   1023    ,   231 ,0 };
            const unsigned long dim189Kuo2Init[]        =   {   1   ,   3   ,   3   ,   11  ,   15  ,   15  ,   45  ,   55  ,   63  ,   693 ,   845 ,0 };
            const unsigned long dim190Kuo2Init[]        =   {   1   ,   1   ,   3   ,   1   ,   27  ,   3   ,   63  ,   139 ,   139 ,   729 ,   861 ,0 };
            const unsigned long dim191Kuo2Init[]        =   {   1   ,   1   ,   5   ,   3   ,   31  ,   37  ,   125 ,   175 ,   311 ,   951 ,   1745    ,0 };
            const unsigned long dim192Kuo2Init[]        =   {   1   ,   1   ,   1   ,   11  ,   1   ,   19  ,   7   ,   203 ,   301 ,   403 ,   1293    ,0 };
            const unsigned long dim193Kuo2Init[]        =   {   1   ,   3   ,   1   ,   5   ,   13  ,   39  ,   27  ,   155 ,   27  ,   725 ,   993 ,0 };
            const unsigned long dim194Kuo2Init[]        =   {   1   ,   3   ,   5   ,   1   ,   5   ,   57  ,   19  ,   225 ,   407 ,   713 ,   1687    ,0 };
            const unsigned long dim195Kuo2Init[]        =   {   1   ,   1   ,   1   ,   1   ,   29  ,   23  ,   83  ,   191 ,   469 ,   105 ,   311 ,0 };
            const unsigned long dim196Kuo2Init[]        =   {   1   ,   1   ,   1   ,   11  ,   5   ,   17  ,   111 ,   5   ,   395 ,   183 ,   1167    ,0 };
            const unsigned long dim197Kuo2Init[]        =   {   1   ,   3   ,   1   ,   9   ,   1   ,   41  ,   3   ,   49  ,   253 ,   735 ,   693 ,0 };
            const unsigned long dim198Kuo2Init[]        =   {   1   ,   1   ,   5   ,   3   ,   21  ,   11  ,   43  ,   33  ,   225 ,   291 ,   1581    ,0 };
            const unsigned long dim199Kuo2Init[]        =   {   1   ,   1   ,   1   ,   1   ,   5   ,   29  ,   43  ,   85  ,   465 ,   879 ,   1161    ,0 };
            const unsigned long dim200Kuo2Init[]        =   {   1   ,   3   ,   1   ,   7   ,   11  ,   43  ,   61  ,   41  ,   287 ,   655 ,   1417    ,0 };
            const unsigned long dim201Kuo2Init[]        =   {   1   ,   3   ,   1   ,   15  ,   21  ,   13  ,   85  ,   69  ,   353 ,   1005    ,   1109    ,0 };
            const unsigned long dim202Kuo2Init[]        =   {   1   ,   1   ,   7   ,   1   ,   3   ,   29  ,   115 ,   227 ,   473 ,   877 ,   443 ,0 };
            const unsigned long dim203Kuo2Init[]        =   {   1   ,   3   ,   3   ,   13  ,   13  ,   57  ,   95  ,   91  ,   473 ,   667 ,   1551    ,0 };
            const unsigned long dim204Kuo2Init[]        =   {   1   ,   3   ,   3   ,   3   ,   13  ,   5   ,   45  ,   219 ,   371 ,   843 ,   95  ,0 };
            const unsigned long dim205Kuo2Init[]        =   {   1   ,   3   ,   7   ,   1   ,   17  ,   45  ,   93  ,   63  ,   235 ,   37  ,   1081    ,0 };
            const unsigned long dim206Kuo2Init[]        =   {   1   ,   3   ,   3   ,   1   ,   17  ,   21  ,   5   ,   91  ,   433 ,   321 ,   645 ,0 };
            const unsigned long dim207Kuo2Init[]        =   {   1   ,   1   ,   7   ,   11  ,   9   ,   5   ,   33  ,   189 ,   205 ,   51  ,   259 ,0 };
            const unsigned long dim208Kuo2Init[]        =   {   1   ,   1   ,   7   ,   3   ,   17  ,   9   ,   11  ,   235 ,   477 ,   419 ,   721 ,0 };
            const unsigned long dim209Kuo2Init[]        =   {   1   ,   1   ,   1   ,   15  ,   21  ,   49  ,   25  ,   103 ,   25  ,   273 ,   1159    ,0 };
            const unsigned long dim210Kuo2Init[]        =   {   1   ,   1   ,   1   ,   15  ,   29  ,   7   ,   97  ,   47  ,   219 ,   979 ,   961 ,0 };
            const unsigned long dim211Kuo2Init[]        =   {   1   ,   1   ,   7   ,   13  ,   3   ,   9   ,   103 ,   179 ,   471 ,   841 ,   1265    ,0 };
            const unsigned long dim212Kuo2Init[]        =   {   1   ,   1   ,   7   ,   15  ,   11  ,   33  ,   57  ,   181 ,   125 ,   307 ,   535 ,0 };
            const unsigned long dim213Kuo2Init[]        =   {   1   ,   1   ,   7   ,   15  ,   21  ,   27  ,   15  ,   65  ,   235 ,   45  ,   1169    ,0 };
            const unsigned long dim214Kuo2Init[]        =   {   1   ,   3   ,   1   ,   5   ,   9   ,   59  ,   123 ,   25  ,   391 ,   253 ,   1801    ,0 };
            const unsigned long dim215Kuo2Init[]        =   {   1   ,   3   ,   1   ,   1   ,   13  ,   17  ,   85  ,   207 ,   417 ,   179 ,   951 ,0 };
            const unsigned long dim216Kuo2Init[]        =   {   1   ,   1   ,   1   ,   15  ,   27  ,   13  ,   51  ,   169 ,   351 ,   351 ,   1255    ,0 };
            const unsigned long dim217Kuo2Init[]        =   {   1   ,   3   ,   7   ,   13  ,   21  ,   63  ,   93  ,   27  ,   277 ,   999 ,   1381    ,0 };
            const unsigned long dim218Kuo2Init[]        =   {   1   ,   1   ,   5   ,   9   ,   21  ,   11  ,   99  ,   199 ,   281 ,   281 ,   373 ,0 };
            const unsigned long dim219Kuo2Init[]        =   {   1   ,   1   ,   5   ,   1   ,   1   ,   41  ,   17  ,   43  ,   411 ,   315 ,   1953    ,0 };
            const unsigned long dim220Kuo2Init[]        =   {   1   ,   1   ,   5   ,   11  ,   25  ,   49  ,   7   ,   61  ,   359 ,   989 ,   1165    ,0 };
            const unsigned long dim221Kuo2Init[]        =   {   1   ,   1   ,   5   ,   1   ,   21  ,   23  ,   77  ,   51  ,   501 ,   579 ,   1969    ,0 };
            const unsigned long dim222Kuo2Init[]        =   {   1   ,   3   ,   7   ,   3   ,   15  ,   43  ,   89  ,   107 ,   331 ,   723 ,   771 ,0 };
            const unsigned long dim223Kuo2Init[]        =   {   1   ,   1   ,   3   ,   15  ,   3   ,   57  ,   41  ,   241 ,   183 ,   579 ,   1577    ,0 };
            const unsigned long dim224Kuo2Init[]        =   {   1   ,   3   ,   7   ,   3   ,   29  ,   31  ,   31  ,   121 ,   87  ,   1009    ,   235 ,0 };
            const unsigned long dim225Kuo2Init[]        =   {   1   ,   3   ,   3   ,   1   ,   25  ,   11  ,   67  ,   33  ,   275 ,   297 ,   1143    ,0 };
            const unsigned long dim226Kuo2Init[]        =   {   1   ,   1   ,   7   ,   13  ,   17  ,   29  ,   99  ,   39  ,   111 ,   29  ,   881 ,0 };
            const unsigned long dim227Kuo2Init[]        =   {   1   ,   1   ,   7   ,   1   ,   7   ,   25  ,   27  ,   107 ,   169 ,   77  ,   1037    ,0 };
            const unsigned long dim228Kuo2Init[]        =   {   1   ,   3   ,   1   ,   15  ,   17  ,   59  ,   105 ,   87  ,   295 ,   253 ,   1847    ,0 };
            const unsigned long dim229Kuo2Init[]        =   {   1   ,   1   ,   7   ,   11  ,   25  ,   57  ,   75  ,   129 ,   379 ,   95  ,   381 ,0 };
            const unsigned long dim230Kuo2Init[]        =   {   1   ,   1   ,   7   ,   11  ,   19  ,   51  ,   69  ,   23  ,   443 ,   89  ,   1245    ,0 };
            const unsigned long dim231Kuo2Init[]        =   {   1   ,   3   ,   7   ,   15  ,   3   ,   43  ,   71  ,   173 ,   435 ,   125 ,   1039    ,0 };
            const unsigned long dim232Kuo2Init[]        =   {   1   ,   1   ,   1   ,   13  ,   27  ,   37  ,   101 ,   97  ,   201 ,   187 ,   243 ,0 };
            const unsigned long dim233Kuo2Init[]        =   {   1   ,   1   ,   3   ,   15  ,   13  ,   37  ,   11  ,   81  ,   233 ,   701 ,   1629    ,0 };
            const unsigned long dim234Kuo2Init[]        =   {   1   ,   1   ,   3   ,   11  ,   21  ,   43  ,   37  ,   233 ,   129 ,   281 ,   511 ,0 };
            const unsigned long dim235Kuo2Init[]        =   {   1   ,   3   ,   7   ,   3   ,   17  ,   7   ,   89  ,   19  ,   59  ,   339 ,   337 ,0 };
            const unsigned long dim236Kuo2Init[]        =   {   1   ,   3   ,   7   ,   13  ,   27  ,   31  ,   97  ,   197 ,   485 ,   367 ,   1169    ,0 };
            const unsigned long dim237Kuo2Init[]        =   {   1   ,   1   ,   7   ,   13  ,   9   ,   5   ,   93  ,   93  ,   289 ,   79  ,   57  ,0 };
            const unsigned long dim238Kuo2Init[]        =   {   1   ,   3   ,   5   ,   13  ,   23  ,   61  ,   99  ,   201 ,   115 ,   583 ,   1271    ,0 };
            const unsigned long dim239Kuo2Init[]        =   {   1   ,   3   ,   3   ,   15  ,   9   ,   55  ,   1   ,   197 ,   149 ,   919 ,   805 ,0 };
            const unsigned long dim240Kuo2Init[]        =   {   1   ,   1   ,   1   ,   5   ,   29  ,   43  ,   119 ,   55  ,   397 ,   973 ,   1843    ,0 };
            const unsigned long dim241Kuo2Init[]        =   {   1   ,   3   ,   1   ,   1   ,   5   ,   55  ,   99  ,   223 ,   323 ,   543 ,   1851    ,0 };
            const unsigned long dim242Kuo2Init[]        =   {   1   ,   1   ,   1   ,   15  ,   1   ,   31  ,   17  ,   107 ,   303 ,   675 ,   1093    ,0 };
            const unsigned long dim243Kuo2Init[]        =   {   1   ,   1   ,   5   ,   11  ,   9   ,   19  ,   61  ,   123 ,   411 ,   65  ,   317 ,0 };
            const unsigned long dim244Kuo2Init[]        =   {   1   ,   3   ,   5   ,   5   ,   9   ,   15  ,   35  ,   181 ,   467 ,   289 ,   85  ,0 };
            const unsigned long dim245Kuo2Init[]        =   {   1   ,   1   ,   1   ,   13  ,   19  ,   59  ,   75  ,   167 ,   507 ,   249 ,   259 ,0 };
            const unsigned long dim246Kuo2Init[]        =   {   1   ,   3   ,   5   ,   3   ,   13  ,   7   ,   53  ,   87  ,   97  ,   667 ,   1493    ,0 };
            const unsigned long dim247Kuo2Init[]        =   {   1   ,   1   ,   7   ,   1   ,   17  ,   35  ,   109 ,   21  ,   437 ,   225 ,   155 ,0 };
            const unsigned long dim248Kuo2Init[]        =   {   1   ,   3   ,   7   ,   11  ,   3   ,   25  ,   51  ,   29  ,   499 ,   987 ,   223 ,0 };
            const unsigned long dim249Kuo2Init[]        =   {   1   ,   1   ,   7   ,   7   ,   23  ,   57  ,   13  ,   23  ,   335 ,   627 ,   887 ,0 };
            const unsigned long dim250Kuo2Init[]        =   {   1   ,   1   ,   7   ,   3   ,   15  ,   45  ,   89  ,   217 ,   393 ,   235 ,   969 ,0 };
            const unsigned long dim251Kuo2Init[]        =   {   1   ,   3   ,   3   ,   7   ,   19  ,   47  ,   1   ,   193 ,   229 ,   45  ,   1617    ,0 };
            const unsigned long dim252Kuo2Init[]        =   {   1   ,   3   ,   1   ,   9   ,   29  ,   63  ,   15  ,   57  ,   509 ,   147 ,   323 ,0 };
            const unsigned long dim253Kuo2Init[]        =   {   1   ,   3   ,   5   ,   9   ,   9   ,   33  ,   77  ,   83  ,   109 ,   335 ,   481 ,0 };
            const unsigned long dim254Kuo2Init[]        =   {   1   ,   1   ,   3   ,   1   ,   7   ,   47  ,   69  ,   3   ,   23  ,   243 ,   1787    ,0 };
            const unsigned long dim255Kuo2Init[]        =   {   1   ,   3   ,   7   ,   7   ,   7   ,   13  ,   77  ,   211 ,   349 ,   383 ,   1793    ,0 };
            const unsigned long dim256Kuo2Init[]        =   {   1   ,   1   ,   5   ,   11  ,   9   ,   9   ,   1   ,   173 ,   263 ,   965 ,   993 ,0 };
            const unsigned long dim257Kuo2Init[]        =   {   1   ,   3   ,   3   ,   1   ,   11  ,   57  ,   85  ,   1   ,   293 ,   921 ,   1711    ,0 };
            const unsigned long dim258Kuo2Init[]        =   {   1   ,   3   ,   3   ,   13  ,   29  ,   25  ,   13  ,   143 ,   373 ,   575 ,   1985    ,0 };
            const unsigned long dim259Kuo2Init[]        =   {   1   ,   3   ,   1   ,   13  ,   1   ,   9   ,   71  ,   135 ,   117 ,   989 ,   1331    ,0 };
            const unsigned long dim260Kuo2Init[]        =   {   1   ,   3   ,   3   ,   13  ,   29  ,   21  ,   47  ,   225 ,   199 ,   107 ,   117 ,0 };
            const unsigned long dim261Kuo2Init[]        =   {   1   ,   1   ,   7   ,   13  ,   25  ,   5   ,   57  ,   135 ,   397 ,   411 ,   1275    ,0 };
            const unsigned long dim262Kuo2Init[]        =   {   1   ,   1   ,   3   ,   1   ,   11  ,   19  ,   39  ,   9   ,   165 ,   813 ,   871 ,0 };
            const unsigned long dim263Kuo2Init[]        =   {   1   ,   3   ,   1   ,   11  ,   29  ,   1   ,   71  ,   35  ,   361 ,   331 ,   773 ,0 };
            const unsigned long dim264Kuo2Init[]        =   {   1   ,   1   ,   3   ,   9   ,   23  ,   23  ,   73  ,   7   ,   505 ,   183 ,   563 ,0 };
            const unsigned long dim265Kuo2Init[]        =   {   1   ,   3   ,   5   ,   9   ,   25  ,   35  ,   123 ,   111 ,   221 ,   1   ,   873 ,0 };
            const unsigned long dim266Kuo2Init[]        =   {   1   ,   1   ,   3   ,   3   ,   15  ,   43  ,   53  ,   71  ,   339 ,   321 ,   1973    ,0 };
            const unsigned long dim267Kuo2Init[]        =   {   1   ,   1   ,   5   ,   13  ,   13  ,   57  ,   27  ,   151 ,   337 ,   739 ,   1205    ,0 };
            const unsigned long dim268Kuo2Init[]        =   {   1   ,   3   ,   5   ,   13  ,   23  ,   3   ,   111 ,   101 ,   37  ,   523 ,   659 ,0 };
            const unsigned long dim269Kuo2Init[]        =   {   1   ,   3   ,   5   ,   13  ,   23  ,   53  ,   53  ,   91  ,   505 ,   779 ,   759 ,0 };
            const unsigned long dim270Kuo2Init[]        =   {   1   ,   1   ,   5   ,   7   ,   21  ,   43  ,   91  ,   213 ,   163 ,   29  ,   137 ,0 };
            const unsigned long dim271Kuo2Init[]        =   {   1   ,   1   ,   1   ,   3   ,   21  ,   37  ,   97  ,   241 ,   77  ,   169 ,   1255    ,0 };
            const unsigned long dim272Kuo2Init[]        =   {   1   ,   3   ,   3   ,   1   ,   5   ,   31  ,   119 ,   127 ,   137 ,   915 ,   1705    ,0 };
            const unsigned long dim273Kuo2Init[]        =   {   1   ,   3   ,   3   ,   13  ,   5   ,   47  ,   117 ,   171 ,   67  ,   285 ,   225 ,0 };
            const unsigned long dim274Kuo2Init[]        =   {   1   ,   1   ,   7   ,   9   ,   3   ,   35  ,   25  ,   25  ,   147 ,   537 ,   1703    ,0 };
            const unsigned long dim275Kuo2Init[]        =   {   1   ,   3   ,   5   ,   3   ,   27  ,   41  ,   57  ,   71  ,   383 ,   763 ,   431 ,0 };
            const unsigned long dim276Kuo2Init[]        =   {   1   ,   3   ,   3   ,   1   ,   17  ,   9   ,   105 ,   119 ,   243 ,   459 ,   881 ,0 };
            const unsigned long dim277Kuo2Init[]        =   {   1   ,   3   ,   5   ,   15  ,   13  ,   45  ,   49  ,   41  ,   117 ,   473 ,   1701    ,0 };
            const unsigned long dim278Kuo2Init[]        =   {   1   ,   1   ,   3   ,   11  ,   13  ,   5   ,   77  ,   17  ,   11  ,   1019    ,   353 ,0 };
            const unsigned long dim279Kuo2Init[]        =   {   1   ,   3   ,   1   ,   11  ,   31  ,   35  ,   87  ,   195 ,   371 ,   169 ,   1075    ,0 };
            const unsigned long dim280Kuo2Init[]        =   {   1   ,   3   ,   5   ,   1   ,   1   ,   1   ,   65  ,   177 ,   35  ,   497 ,   103 ,0 };
            const unsigned long dim281Kuo2Init[]        =   {   1   ,   1   ,   3   ,   1   ,   31  ,   21  ,   77  ,   153 ,   139 ,   663 ,   443 ,0 };
            const unsigned long dim282Kuo2Init[]        =   {   1   ,   1   ,   1   ,   1   ,   31  ,   25  ,   21  ,   87  ,   345 ,   13  ,   301 ,0 };
            const unsigned long dim283Kuo2Init[]        =   {   1   ,   1   ,   1   ,   7   ,   7   ,   1   ,   95  ,   5   ,   429 ,   551 ,   1011    ,0 };
            const unsigned long dim284Kuo2Init[]        =   {   1   ,   3   ,   5   ,   15  ,   23  ,   63  ,   119 ,   185 ,   463 ,   371 ,   567 ,0 };
            const unsigned long dim285Kuo2Init[]        =   {   1   ,   1   ,   5   ,   5   ,   21  ,   21  ,   33  ,   141 ,   181 ,   885 ,   717 ,0 };
            const unsigned long dim286Kuo2Init[]        =   {   1   ,   1   ,   1   ,   5   ,   11  ,   35  ,   37  ,   231 ,   67  ,   951 ,   1721    ,0 };
            const unsigned long dim287Kuo2Init[]        =   {   1   ,   3   ,   7   ,   9   ,   13  ,   29  ,   113 ,   55  ,   27  ,   625 ,   285 ,0 };
            const unsigned long dim288Kuo2Init[]        =   {   1   ,   1   ,   7   ,   5   ,   19  ,   45  ,   5   ,   233 ,   133 ,   401 ,   99  ,0 };
            const unsigned long dim289Kuo2Init[]        =   {   1   ,   1   ,   5   ,   1   ,   7   ,   49  ,   127 ,   173 ,   187 ,   513 ,   171 ,0 };
            const unsigned long dim290Kuo2Init[]        =   {   1   ,   1   ,   7   ,   3   ,   17  ,   59  ,   51  ,   81  ,   197 ,   1021    ,   763 ,0 };
            const unsigned long dim291Kuo2Init[]        =   {   1   ,   1   ,   3   ,   1   ,   9   ,   51  ,   31  ,   149 ,   397 ,   929 ,   81  ,0 };
            const unsigned long dim292Kuo2Init[]        =   {   1   ,   1   ,   7   ,   9   ,   29  ,   29  ,   21  ,   191 ,   107 ,   9   ,   807 ,0 };
            const unsigned long dim293Kuo2Init[]        =   {   1   ,   1   ,   5   ,   3   ,   15  ,   3   ,   125 ,   105 ,   225 ,   571 ,   33  ,0 };
            const unsigned long dim294Kuo2Init[]        =   {   1   ,   1   ,   5   ,   15  ,   31  ,   61  ,   113 ,   7   ,   59  ,   829 ,   1259    ,0 };
            const unsigned long dim295Kuo2Init[]        =   {   1   ,   3   ,   1   ,   5   ,   29  ,   27  ,   33  ,   145 ,   211 ,   837 ,   749 ,0 };
            const unsigned long dim296Kuo2Init[]        =   {   1   ,   1   ,   1   ,   1   ,   27  ,   15  ,   25  ,   197 ,   51  ,   71  ,   813 ,0 };
            const unsigned long dim297Kuo2Init[]        =   {   1   ,   1   ,   1   ,   3   ,   27  ,   59  ,   1   ,   163 ,   467 ,   991 ,   1489    ,0 };
            const unsigned long dim298Kuo2Init[]        =   {   1   ,   3   ,   3   ,   13  ,   1   ,   27  ,   105 ,   225 ,   403 ,   803 ,   409 ,0 };
            const unsigned long dim299Kuo2Init[]        =   {   1   ,   1   ,   5   ,   3   ,   7   ,   31  ,   99  ,   55  ,   275 ,   487 ,   983 ,0 };
            const unsigned long dim300Kuo2Init[]        =   {   1   ,   1   ,   1   ,   5   ,   27  ,   63  ,   107 ,   161 ,   491 ,   887 ,   825 ,0 };
            const unsigned long dim301Kuo2Init[]        =   {   1   ,   1   ,   5   ,   15  ,   15  ,   19  ,   87  ,   165 ,   55  ,   477 ,   1641    ,0 };
            const unsigned long dim302Kuo2Init[]        =   {   1   ,   3   ,   5   ,   7   ,   21  ,   49  ,   105 ,   151 ,   251 ,   385 ,   1265    ,0 };
            const unsigned long dim303Kuo2Init[]        =   {   1   ,   1   ,   5   ,   3   ,   5   ,   43  ,   47  ,   213 ,   387 ,   523 ,   1547    ,0 };
            const unsigned long dim304Kuo2Init[]        =   {   1   ,   3   ,   1   ,   11  ,   15  ,   55  ,   93  ,   167 ,   377 ,   201 ,   1031    ,0 };
            const unsigned long dim305Kuo2Init[]        =   {   1   ,   1   ,   7   ,   7   ,   9   ,   27  ,   35  ,   251 ,   425 ,   27  ,   1527    ,0 };
            const unsigned long dim306Kuo2Init[]        =   {   1   ,   1   ,   7   ,   7   ,   15  ,   55  ,   123 ,   21  ,   53  ,   77  ,   345 ,0 };
            const unsigned long dim307Kuo2Init[]        =   {   1   ,   1   ,   5   ,   15  ,   13  ,   11  ,   75  ,   101 ,   75  ,   217 ,   219 ,0 };
            const unsigned long dim308Kuo2Init[]        =   {   1   ,   1   ,   3   ,   9   ,   1   ,   29  ,   117 ,   209 ,   301 ,   803 ,   1487    ,0 };
            const unsigned long dim309Kuo2Init[]        =   {   1   ,   3   ,   5   ,   3   ,   31  ,   57  ,   63  ,   171 ,   59  ,   591 ,   1323    ,0 };
            const unsigned long dim310Kuo2Init[]        =   {   1   ,   3   ,   7   ,   11  ,   19  ,   27  ,   111 ,   87  ,   439 ,   1023    ,   691 ,0 };
            const unsigned long dim311Kuo2Init[]        =   {   1   ,   3   ,   7   ,   13  ,   15  ,   57  ,   11  ,   249 ,   383 ,   291 ,   1221    ,0 };
            const unsigned long dim312Kuo2Init[]        =   {   1   ,   1   ,   3   ,   13  ,   23  ,   43  ,   31  ,   75  ,   293 ,   701 ,   1065    ,0 };
            const unsigned long dim313Kuo2Init[]        =   {   1   ,   1   ,   5   ,   5   ,   9   ,   53  ,   115 ,   235 ,   389 ,   907 ,   1883    ,0 };
            const unsigned long dim314Kuo2Init[]        =   {   1   ,   1   ,   1   ,   11  ,   17  ,   31  ,   111 ,   187 ,   483 ,   323 ,   307 ,0 };
            const unsigned long dim315Kuo2Init[]        =   {   1   ,   1   ,   3   ,   1   ,   23  ,   33  ,   25  ,   27  ,   267 ,   231 ,   467 ,0 };
            const unsigned long dim316Kuo2Init[]        =   {   1   ,   3   ,   1   ,   1   ,   29  ,   15  ,   1   ,   151 ,   369 ,   943 ,   1293    ,0 };
            const unsigned long dim317Kuo2Init[]        =   {   1   ,   1   ,   1   ,   15  ,   31  ,   9   ,   1   ,   223 ,   7   ,   99  ,   1725    ,0 };
            const unsigned long dim318Kuo2Init[]        =   {   1   ,   1   ,   1   ,   1   ,   13  ,   3   ,   103 ,   177 ,   61  ,   939 ,   1321    ,0 };
            const unsigned long dim319Kuo2Init[]        =   {   1   ,   3   ,   3   ,   7   ,   29  ,   45  ,   63  ,   119 ,   467 ,   1011    ,   1687    ,0 };
            const unsigned long dim320Kuo2Init[]        =   {   1   ,   3   ,   5   ,   13  ,   13  ,   51  ,   69  ,   113 ,   231 ,   807 ,   863 ,0 };
            const unsigned long dim321Kuo2Init[]        =   {   1   ,   1   ,   1   ,   13  ,   13  ,   57  ,   121 ,   219 ,   77  ,   959 ,   1883    ,0 };
            const unsigned long dim322Kuo2Init[]        =   {   1   ,   1   ,   3   ,   13  ,   25  ,   19  ,   83  ,   139 ,   31  ,   3   ,   367 ,0 };
            const unsigned long dim323Kuo2Init[]        =   {   1   ,   3   ,   7   ,   13  ,   19  ,   49  ,   45  ,   47  ,   197 ,   253 ,   1769    ,0 };
            const unsigned long dim324Kuo2Init[]        =   {   1   ,   3   ,   1   ,   5   ,   13  ,   59  ,   65  ,   23  ,   31  ,   127 ,   1693    ,0 };
            const unsigned long dim325Kuo2Init[]        =   {   1   ,   1   ,   7   ,   5   ,   15  ,   33  ,   85  ,   59  ,   483 ,   87  ,   1109    ,0 };
            const unsigned long dim326Kuo2Init[]        =   {   1   ,   1   ,   1   ,   13  ,   29  ,   39  ,   87  ,   109 ,   223 ,   927 ,   723 ,0 };
            const unsigned long dim327Kuo2Init[]        =   {   1   ,   1   ,   7   ,   11  ,   29  ,   19  ,   13  ,   73  ,   479 ,   989 ,   767 ,0 };
            const unsigned long dim328Kuo2Init[]        =   {   1   ,   3   ,   1   ,   15  ,   27  ,   33  ,   119 ,   237 ,   117 ,   181 ,   1719    ,0 };
            const unsigned long dim329Kuo2Init[]        =   {   1   ,   1   ,   7   ,   9   ,   1   ,   55  ,   115 ,   91  ,   355 ,   21  ,   531 ,0 };
            const unsigned long dim330Kuo2Init[]        =   {   1   ,   1   ,   5   ,   7   ,   29  ,   23  ,   41  ,   227 ,   455 ,   587 ,   523 ,0 };
            const unsigned long dim331Kuo2Init[]        =   {   1   ,   3   ,   5   ,   9   ,   31  ,   21  ,   105 ,   139 ,   69  ,   907 ,   1195    ,0 };
            const unsigned long dim332Kuo2Init[]        =   {   1   ,   1   ,   1   ,   13  ,   11  ,   63  ,   33  ,   161 ,   133 ,   493 ,   1055    ,0 };
            const unsigned long dim333Kuo2Init[]        =   {   1   ,   1   ,   7   ,   15  ,   19  ,   37  ,   37  ,   145 ,   321 ,   781 ,   575 ,0 };
            const unsigned long dim334Kuo2Init[]        =   {   1   ,   1   ,   5   ,   5   ,   29  ,   17  ,   109 ,   193 ,   271 ,   831 ,   915 ,0 };
            const unsigned long dim335Kuo2Init[]        =   {   1   ,   3   ,   3   ,   11  ,   17  ,   63  ,   57  ,   245 ,   347 ,   647 ,   321 ,0 };
            const unsigned long dim336Kuo2Init[]        =   {   1   ,   3   ,   5   ,   7   ,   15  ,   15  ,   45  ,   11  ,   383 ,   231 ,   267 ,0 };
            const unsigned long dim337Kuo2Init[]        =   {   1   ,   3   ,   5   ,   13  ,   15  ,   45  ,   67  ,   135 ,   135 ,   247 ,   1261    ,   205 ,0 };
            const unsigned long dim338Kuo2Init[]        =   {   1   ,   1   ,   7   ,   5   ,   13  ,   37  ,   63  ,   179 ,   43  ,   679 ,   1127    ,   779 ,0 };
            const unsigned long dim339Kuo2Init[]        =   {   1   ,   3   ,   7   ,   15  ,   7   ,   11  ,   105 ,   207 ,   93  ,   127 ,   907 ,   3931    ,0 };
            const unsigned long dim340Kuo2Init[]        =   {   1   ,   3   ,   7   ,   3   ,   29  ,   45  ,   53  ,   91  ,   97  ,   175 ,   1847    ,   1119    ,0 };
            const unsigned long dim341Kuo2Init[]        =   {   1   ,   3   ,   7   ,   13  ,   29  ,   25  ,   49  ,   245 ,   39  ,   651 ,   1353    ,   481 ,0 };
            const unsigned long dim342Kuo2Init[]        =   {   1   ,   1   ,   5   ,   13  ,   13  ,   49  ,   119 ,   85  ,   101 ,   361 ,   1163    ,   2693    ,0 };
            const unsigned long dim343Kuo2Init[]        =   {   1   ,   1   ,   7   ,   11  ,   3   ,   13  ,   23  ,   95  ,   169 ,   1005    ,   1911    ,   849 ,0 };
            const unsigned long dim344Kuo2Init[]        =   {   1   ,   3   ,   7   ,   15  ,   1   ,   25  ,   45  ,   111 ,   3   ,   683 ,   1383    ,   1387    ,0 };
            const unsigned long dim345Kuo2Init[]        =   {   1   ,   3   ,   1   ,   3   ,   19  ,   37  ,   97  ,   183 ,   27  ,   971 ,   853 ,   255 ,0 };
            const unsigned long dim346Kuo2Init[]        =   {   1   ,   3   ,   5   ,   7   ,   7   ,   61  ,   19  ,   163 ,   161 ,   921 ,   1561    ,   841 ,0 };
            const unsigned long dim347Kuo2Init[]        =   {   1   ,   1   ,   3   ,   7   ,   3   ,   17  ,   43  ,   185 ,   231 ,   513 ,   1499    ,   695 ,0 };
            const unsigned long dim348Kuo2Init[]        =   {   1   ,   3   ,   1   ,   13  ,   27  ,   17  ,   123 ,   155 ,   13  ,   873 ,   21  ,   1687    ,0 };
            const unsigned long dim349Kuo2Init[]        =   {   1   ,   1   ,   5   ,   11  ,   7   ,   15  ,   83  ,   103 ,   489 ,   529 ,   1331    ,   3783    ,0 };
            const unsigned long dim350Kuo2Init[]        =   {   1   ,   1   ,   7   ,   13  ,   11  ,   45  ,   11  ,   105 ,   341 ,   39  ,   1697    ,   2701    ,0 };
            const unsigned long dim351Kuo2Init[]        =   {   1   ,   1   ,   1   ,   15  ,   1   ,   37  ,   119 ,   179 ,   471 ,   343 ,   927 ,   2469    ,0 };
            const unsigned long dim352Kuo2Init[]        =   {   1   ,   1   ,   3   ,   9   ,   29  ,   57  ,   59  ,   133 ,   309 ,   587 ,   1775    ,   1811    ,0 };
            const unsigned long dim353Kuo2Init[]        =   {   1   ,   1   ,   5   ,   15  ,   7   ,   51  ,   25  ,   151 ,   173 ,   961 ,   1449    ,   2215    ,0 };
            const unsigned long dim354Kuo2Init[]        =   {   1   ,   3   ,   1   ,   5   ,   21  ,   49  ,   113 ,   199 ,   433 ,   563 ,   849 ,   2741    ,0 };
            const unsigned long dim355Kuo2Init[]        =   {   1   ,   3   ,   7   ,   9   ,   19  ,   7   ,   3   ,   95  ,   259 ,   919 ,   1939    ,   2747    ,0 };
            const unsigned long dim356Kuo2Init[]        =   {   1   ,   3   ,   5   ,   1   ,   19  ,   35  ,   83  ,   95  ,   359 ,   85  ,   677 ,   805 ,0 };
            const unsigned long dim357Kuo2Init[]        =   {   1   ,   3   ,   1   ,   9   ,   19  ,   37  ,   63  ,   43  ,   251 ,   301 ,   1217    ,   3361    ,0 };
            const unsigned long dim358Kuo2Init[]        =   {   1   ,   1   ,   3   ,   7   ,   5   ,   33  ,   45  ,   15  ,   191 ,   49  ,   1511    ,   157 ,0 };
            const unsigned long dim359Kuo2Init[]        =   {   1   ,   3   ,   3   ,   3   ,   7   ,   3   ,   1   ,   241 ,   239 ,   229 ,   103 ,   215 ,0 };
            const unsigned long dim360Kuo2Init[]        =   {   1   ,   3   ,   3   ,   3   ,   29  ,   27  ,   31  ,   61  ,   331 ,   847 ,   1699    ,   923 ,0 };
            const unsigned long dim361Kuo2Init[]        =   {   1   ,   3   ,   1   ,   5   ,   11  ,   15  ,   81  ,   97  ,   411 ,   471 ,   1521    ,   73  ,0 };
            const unsigned long dim362Kuo2Init[]        =   {   1   ,   1   ,   7   ,   7   ,   7   ,   17  ,   95  ,   245 ,   415 ,   31  ,   259 ,   1835    ,0 };
            const unsigned long dim363Kuo2Init[]        =   {   1   ,   1   ,   1   ,   1   ,   23  ,   55  ,   27  ,   201 ,   21  ,   651 ,   1965    ,   1723    ,0 };
            const unsigned long dim364Kuo2Init[]        =   {   1   ,   1   ,   7   ,   11  ,   19  ,   43  ,   33  ,   7   ,   1   ,   635 ,   1651    ,   1041    ,0 };
            const unsigned long dim365Kuo2Init[]        =   {   1   ,   1   ,   1   ,   15  ,   29  ,   11  ,   31  ,   11  ,   29  ,   935 ,   715 ,   1115    ,0 };
            const unsigned long dim366Kuo2Init[]        =   {   1   ,   1   ,   3   ,   3   ,   23  ,   51  ,   115 ,   55  ,   45  ,   135 ,   913 ,   3181    ,0 };
            const unsigned long dim367Kuo2Init[]        =   {   1   ,   3   ,   7   ,   5   ,   15  ,   23  ,   39  ,   145 ,   51  ,   243 ,   1705    ,   3889    ,0 };
            const unsigned long dim368Kuo2Init[]        =   {   1   ,   1   ,   7   ,   5   ,   27  ,   43  ,   63  ,   15  ,   365 ,   299 ,   1413    ,   1553    ,0 };
            const unsigned long dim369Kuo2Init[]        =   {   1   ,   3   ,   1   ,   1   ,   19  ,   41  ,   123 ,   123 ,   331 ,   487 ,   2033    ,   1849    ,0 };
            const unsigned long dim370Kuo2Init[]        =   {   1   ,   3   ,   5   ,   13  ,   21  ,   27  ,   91  ,   225 ,   387 ,   529 ,   1529    ,   1855    ,0 };
            const unsigned long dim371Kuo2Init[]        =   {   1   ,   3   ,   5   ,   13  ,   23  ,   51  ,   27  ,   213 ,   55  ,   619 ,   773 ,   2037    ,0 };
            const unsigned long dim372Kuo2Init[]        =   {   1   ,   3   ,   3   ,   1   ,   11  ,   9   ,   125 ,   205 ,   255 ,   813 ,   925 ,   2031    ,0 };
            const unsigned long dim373Kuo2Init[]        =   {   1   ,   3   ,   5   ,   9   ,   23  ,   45  ,   43  ,   197 ,   475 ,   487 ,   79  ,   3649    ,0 };
            const unsigned long dim374Kuo2Init[]        =   {   1   ,   1   ,   1   ,   1   ,   5   ,   53  ,   97  ,   243 ,   433 ,   75  ,   467 ,   1213    ,0 };
            const unsigned long dim375Kuo2Init[]        =   {   1   ,   1   ,   5   ,   7   ,   5   ,   53  ,   15  ,   91  ,   405 ,   791 ,   1293    ,   1517    ,0 };
            const unsigned long dim376Kuo2Init[]        =   {   1   ,   1   ,   1   ,   3   ,   17  ,   27  ,   19  ,   81  ,   227 ,   935 ,   1511    ,   3709    ,0 };
            const unsigned long dim377Kuo2Init[]        =   {   1   ,   1   ,   5   ,   11  ,   17  ,   41  ,   125 ,   131 ,   295 ,   867 ,   1291    ,   2833    ,0 };
            const unsigned long dim378Kuo2Init[]        =   {   1   ,   3   ,   7   ,   13  ,   3   ,   11  ,   63  ,   159 ,   341 ,   143 ,   1779    ,   1441    ,0 };
            const unsigned long dim379Kuo2Init[]        =   {   1   ,   1   ,   7   ,   3   ,   29  ,   3   ,   103 ,   7   ,   145 ,   589 ,   723 ,   663 ,0 };
            const unsigned long dim380Kuo2Init[]        =   {   1   ,   1   ,   7   ,   15  ,   5   ,   25  ,   105 ,   129 ,   277 ,   163 ,   349 ,   219 ,0 };
            const unsigned long dim381Kuo2Init[]        =   {   1   ,   3   ,   1   ,   9   ,   21  ,   27  ,   127 ,   167 ,   285 ,   135 ,   1493    ,   3825    ,0 };
            const unsigned long dim382Kuo2Init[]        =   {   1   ,   3   ,   1   ,   1   ,   7   ,   41  ,   95  ,   255 ,   273 ,   75  ,   383 ,   15  ,0 };
            const unsigned long dim383Kuo2Init[]        =   {   1   ,   3   ,   5   ,   7   ,   21  ,   19  ,   67  ,   19  ,   465 ,   899 ,   1515    ,   425 ,0 };
            const unsigned long dim384Kuo2Init[]        =   {   1   ,   1   ,   5   ,   15  ,   17  ,   45  ,   31  ,   227 ,   229 ,   271 ,   1169    ,   3915    ,0 };
            const unsigned long dim385Kuo2Init[]        =   {   1   ,   3   ,   7   ,   11  ,   7   ,   23  ,   117 ,   195 ,   435 ,   171 ,   777 ,   3473    ,0 };
            const unsigned long dim386Kuo2Init[]        =   {   1   ,   1   ,   3   ,   3   ,   27  ,   7   ,   99  ,   11  ,   361 ,   129 ,   717 ,   1241    ,0 };
            const unsigned long dim387Kuo2Init[]        =   {   1   ,   1   ,   5   ,   9   ,   27  ,   1   ,   11  ,   115 ,   15  ,   561 ,   1989    ,   3857    ,0 };
            const unsigned long dim388Kuo2Init[]        =   {   1   ,   1   ,   1   ,   5   ,   29  ,   55  ,   23  ,   83  ,   45  ,   427 ,   987 ,   3627    ,0 };
            const unsigned long dim389Kuo2Init[]        =   {   1   ,   1   ,   1   ,   9   ,   5   ,   59  ,   119 ,   77  ,   297 ,   129 ,   1675    ,   2621    ,0 };
            const unsigned long dim390Kuo2Init[]        =   {   1   ,   1   ,   5   ,   11  ,   27  ,   49  ,   69  ,   115 ,   453 ,   863 ,   811 ,   535 ,0 };
            const unsigned long dim391Kuo2Init[]        =   {   1   ,   1   ,   5   ,   9   ,   21  ,   43  ,   109 ,   207 ,   5   ,   987 ,   1747    ,   47  ,0 };
            const unsigned long dim392Kuo2Init[]        =   {   1   ,   3   ,   5   ,   5   ,   27  ,   31  ,   67  ,   133 ,   153 ,   137 ,   1167    ,   325 ,0 };
            const unsigned long dim393Kuo2Init[]        =   {   1   ,   1   ,   3   ,   15  ,   9   ,   7   ,   51  ,   221 ,   317 ,   917 ,   1589    ,   1701    ,0 };
            const unsigned long dim394Kuo2Init[]        =   {   1   ,   1   ,   3   ,   11  ,   7   ,   21  ,   3   ,   235 ,   237 ,   853 ,   1883    ,   2733    ,0 };
            const unsigned long dim395Kuo2Init[]        =   {   1   ,   3   ,   7   ,   1   ,   9   ,   59  ,   69  ,   71  ,   191 ,   507 ,   1387    ,   2643    ,0 };
            const unsigned long dim396Kuo2Init[]        =   {   1   ,   1   ,   1   ,   7   ,   13  ,   17  ,   59  ,   115 ,   401 ,   373 ,   431 ,   2895    ,0 };
            const unsigned long dim397Kuo2Init[]        =   {   1   ,   3   ,   7   ,   7   ,   31  ,   23  ,   51  ,   173 ,   425 ,   823 ,   1375    ,   1675    ,0 };
            const unsigned long dim398Kuo2Init[]        =   {   1   ,   3   ,   1   ,   13  ,   1   ,   57  ,   57  ,   203 ,   275 ,   825 ,   969 ,   1357    ,0 };
            const unsigned long dim399Kuo2Init[]        =   {   1   ,   3   ,   1   ,   1   ,   31  ,   39  ,   125 ,   49  ,   11  ,   255 ,   373 ,   805 ,0 };
            const unsigned long dim400Kuo2Init[]        =   {   1   ,   3   ,   7   ,   1   ,   7   ,   59  ,   43  ,   143 ,   447 ,   73  ,   1979    ,   3401    ,0 };
            const unsigned long dim401Kuo2Init[]        =   {   1   ,   3   ,   7   ,   15  ,   29  ,   31  ,   11  ,   49  ,   241 ,   157 ,   1147    ,   1049    ,0 };
            const unsigned long dim402Kuo2Init[]        =   {   1   ,   1   ,   3   ,   11  ,   19  ,   39  ,   41  ,   145 ,   219 ,   811 ,   425 ,   3561    ,0 };
            const unsigned long dim403Kuo2Init[]        =   {   1   ,   1   ,   3   ,   13  ,   23  ,   49  ,   123 ,   3   ,   423 ,   661 ,   329 ,   3883    ,0 };
            const unsigned long dim404Kuo2Init[]        =   {   1   ,   3   ,   1   ,   15  ,   21  ,   21  ,   5   ,   105 ,   19  ,   171 ,   1187    ,   1451    ,0 };
            const unsigned long dim405Kuo2Init[]        =   {   1   ,   1   ,   7   ,   3   ,   27  ,   7   ,   31  ,   25  ,   339 ,   107 ,   827 ,   2813    ,0 };
            const unsigned long dim406Kuo2Init[]        =   {   1   ,   1   ,   3   ,   3   ,   7   ,   9   ,   5   ,   241 ,   419 ,   19  ,   303 ,   1327    ,0 };
            const unsigned long dim407Kuo2Init[]        =   {   1   ,   3   ,   3   ,   5   ,   1   ,   17  ,   105 ,   143 ,   345 ,   953 ,   857 ,   2241    ,0 };
            const unsigned long dim408Kuo2Init[]        =   {   1   ,   3   ,   7   ,   5   ,   11  ,   63  ,   39  ,   49  ,   3   ,   601 ,   1491    ,   3095    ,0 };
            const unsigned long dim409Kuo2Init[]        =   {   1   ,   1   ,   5   ,   1   ,   27  ,   35  ,   25  ,   231 ,   265 ,   27  ,   83  ,   511 ,0 };
            const unsigned long dim410Kuo2Init[]        =   {   1   ,   3   ,   5   ,   3   ,   29  ,   1   ,   119 ,   39  ,   81  ,   551 ,   1709    ,   2105    ,0 };
            const unsigned long dim411Kuo2Init[]        =   {   1   ,   3   ,   1   ,   5   ,   29  ,   31  ,   99  ,   93  ,   391 ,   691 ,   411 ,   637 ,0 };
            const unsigned long dim412Kuo2Init[]        =   {   1   ,   3   ,   5   ,   1   ,   21  ,   53  ,   81  ,   171 ,   303 ,   987 ,   1235    ,   1559    ,0 };
            const unsigned long dim413Kuo2Init[]        =   {   1   ,   3   ,   3   ,   5   ,   7   ,   63  ,   51  ,   89  ,   257 ,   489 ,   275 ,   2309    ,0 };
            const unsigned long dim414Kuo2Init[]        =   {   1   ,   3   ,   1   ,   5   ,   5   ,   33  ,   25  ,   55  ,   241 ,   259 ,   1213    ,   2409    ,0 };
            const unsigned long dim415Kuo2Init[]        =   {   1   ,   3   ,   7   ,   3   ,   11  ,   57  ,   85  ,   209 ,   329 ,   611 ,   45  ,   1183    ,0 };
            const unsigned long dim416Kuo2Init[]        =   {   1   ,   3   ,   3   ,   15  ,   3   ,   15  ,   31  ,   27  ,   363 ,   291 ,   1063    ,   229 ,0 };
            const unsigned long dim417Kuo2Init[]        =   {   1   ,   3   ,   1   ,   1   ,   1   ,   3   ,   47  ,   189 ,   287 ,   337 ,   879 ,   1881    ,0 };
            const unsigned long dim418Kuo2Init[]        =   {   1   ,   1   ,   3   ,   5   ,   9   ,   11  ,   65  ,   237 ,   499 ,   431 ,   1677    ,   2503    ,0 };
            const unsigned long dim419Kuo2Init[]        =   {   1   ,   3   ,   1   ,   15  ,   3   ,   17  ,   101 ,   23  ,   367 ,   643 ,   1485    ,   1625    ,0 };
            const unsigned long dim420Kuo2Init[]        =   {   1   ,   3   ,   3   ,   13  ,   9   ,   3   ,   87  ,   11  ,   137 ,   835 ,   1095    ,   891 ,0 };
            const unsigned long dim421Kuo2Init[]        =   {   1   ,   1   ,   5   ,   7   ,   15  ,   3   ,   49  ,   253 ,   29  ,   159 ,   315 ,   2133    ,0 };
            const unsigned long dim422Kuo2Init[]        =   {   1   ,   1   ,   3   ,   15  ,   11  ,   11  ,   17  ,   37  ,   287 ,   35  ,   813 ,   2715    ,0 };
            const unsigned long dim423Kuo2Init[]        =   {   1   ,   1   ,   1   ,   9   ,   3   ,   51  ,   57  ,   205 ,   385 ,   157 ,   811 ,   627 ,0 };
            const unsigned long dim424Kuo2Init[]        =   {   1   ,   3   ,   7   ,   15  ,   13  ,   33  ,   65  ,   117 ,   339 ,   805 ,   19  ,   3775    ,0 };
            const unsigned long dim425Kuo2Init[]        =   {   1   ,   3   ,   5   ,   5   ,   9   ,   5   ,   25  ,   177 ,   469 ,   599 ,   1527    ,   3799    ,0 };
            const unsigned long dim426Kuo2Init[]        =   {   1   ,   1   ,   3   ,   9   ,   1   ,   45  ,   65  ,   69  ,   121 ,   263 ,   1355    ,   1205    ,0 };
            const unsigned long dim427Kuo2Init[]        =   {   1   ,   1   ,   3   ,   3   ,   3   ,   33  ,   121 ,   163 ,   95  ,   137 ,   1393    ,   4017    ,0 };
            const unsigned long dim428Kuo2Init[]        =   {   1   ,   3   ,   7   ,   7   ,   19  ,   27  ,   17  ,   239 ,   253 ,   305 ,   1929    ,   1725    ,0 };
            const unsigned long dim429Kuo2Init[]        =   {   1   ,   3   ,   5   ,   7   ,   5   ,   19  ,   115 ,   95  ,   93  ,   87  ,   1379    ,   3295    ,0 };
            const unsigned long dim430Kuo2Init[]        =   {   1   ,   3   ,   3   ,   3   ,   5   ,   51  ,   7   ,   171 ,   153 ,   329 ,   1627    ,   3503    ,0 };
            const unsigned long dim431Kuo2Init[]        =   {   1   ,   1   ,   5   ,   9   ,   27  ,   15  ,   91  ,   99  ,   433 ,   689 ,   1551    ,   1827    ,0 };
            const unsigned long dim432Kuo2Init[]        =   {   1   ,   3   ,   3   ,   11  ,   17  ,   35  ,   29  ,   203 ,   171 ,   513 ,   771 ,   2011    ,0 };
            const unsigned long dim433Kuo2Init[]        =   {   1   ,   3   ,   5   ,   11  ,   29  ,   57  ,   69  ,   239 ,   87  ,   477 ,   187 ,   4079    ,0 };
            const unsigned long dim434Kuo2Init[]        =   {   1   ,   1   ,   1   ,   11  ,   9   ,   25  ,   37  ,   41  ,   113 ,   339 ,   1347    ,   3475    ,0 };
            const unsigned long dim435Kuo2Init[]        =   {   1   ,   3   ,   1   ,   15  ,   17  ,   17  ,   53  ,   17  ,   327 ,   123 ,   915 ,   655 ,0 };
            const unsigned long dim436Kuo2Init[]        =   {   1   ,   1   ,   5   ,   13  ,   29  ,   3   ,   85  ,   173 ,   139 ,   51  ,   929 ,   3407    ,0 };
            const unsigned long dim437Kuo2Init[]        =   {   1   ,   3   ,   5   ,   3   ,   25  ,   13  ,   19  ,   91  ,   285 ,   309 ,   105 ,   2331    ,0 };
            const unsigned long dim438Kuo2Init[]        =   {   1   ,   3   ,   7   ,   7   ,   13  ,   23  ,   103 ,   7   ,   435 ,   571 ,   731 ,   431 ,0 };
            const unsigned long dim439Kuo2Init[]        =   {   1   ,   1   ,   3   ,   11  ,   23  ,   3   ,   29  ,   25  ,   369 ,   941 ,   2031    ,   2311    ,0 };
            const unsigned long dim440Kuo2Init[]        =   {   1   ,   3   ,   5   ,   3   ,   11  ,   49  ,   69  ,   61  ,   267 ,   241 ,   1221    ,   1533    ,0 };
            const unsigned long dim441Kuo2Init[]        =   {   1   ,   3   ,   3   ,   5   ,   17  ,   43  ,   27  ,   191 ,   201 ,   653 ,   933 ,   2571    ,0 };
            const unsigned long dim442Kuo2Init[]        =   {   1   ,   1   ,   1   ,   13  ,   1   ,   7   ,   117 ,   179 ,   61  ,   1003    ,   379 ,   2359    ,0 };
            const unsigned long dim443Kuo2Init[]        =   {   1   ,   3   ,   5   ,   15  ,   9   ,   53  ,   23  ,   217 ,   237 ,   647 ,   377 ,   677 ,0 };
            const unsigned long dim444Kuo2Init[]        =   {   1   ,   1   ,   1   ,   1   ,   15  ,   55  ,   73  ,   1   ,   119 ,   679 ,   203 ,   309 ,0 };
            const unsigned long dim445Kuo2Init[]        =   {   1   ,   1   ,   5   ,   15  ,   11  ,   13  ,   15  ,   211 ,   105 ,   655 ,   739 ,   1439    ,0 };
            const unsigned long dim446Kuo2Init[]        =   {   1   ,   3   ,   5   ,   5   ,   9   ,   41  ,   87  ,   237 ,   65  ,   837 ,   1127    ,   2539    ,0 };
            const unsigned long dim447Kuo2Init[]        =   {   1   ,   1   ,   3   ,   9   ,   3   ,   57  ,   23  ,   55  ,   359 ,   909 ,   969 ,   1437    ,0 };
            const unsigned long dim448Kuo2Init[]        =   {   1   ,   1   ,   1   ,   9   ,   23  ,   57  ,   23  ,   127 ,   139 ,   961 ,   1985    ,   2785    ,0 };
            const unsigned long dim449Kuo2Init[]        =   {   1   ,   3   ,   7   ,   1   ,   31  ,   7   ,   65  ,   215 ,   305 ,   245 ,   971 ,   139 ,0 };
            const unsigned long dim450Kuo2Init[]        =   {   1   ,   3   ,   3   ,   11  ,   3   ,   3   ,   29  ,   173 ,   189 ,   807 ,   219 ,   2807    ,0 };
            const unsigned long dim451Kuo2Init[]        =   {   1   ,   3   ,   5   ,   7   ,   25  ,   39  ,   7   ,   157 ,   397 ,   773 ,   461 ,   3051    ,0 };
            const unsigned long dim452Kuo2Init[]        =   {   1   ,   3   ,   7   ,   7   ,   9   ,   7   ,   33  ,   19  ,   127 ,   177 ,   189 ,   1247    ,0 };
            const unsigned long dim453Kuo2Init[]        =   {   1   ,   3   ,   5   ,   3   ,   5   ,   13  ,   77  ,   219 ,   339 ,   969 ,   1193    ,   2545    ,0 };
            const unsigned long dim454Kuo2Init[]        =   {   1   ,   3   ,   3   ,   13  ,   29  ,   45  ,   123 ,   9   ,   335 ,   985 ,   1923    ,   2449    ,0 };
            const unsigned long dim455Kuo2Init[]        =   {   1   ,   3   ,   1   ,   9   ,   7   ,   43  ,   29  ,   83  ,   153 ,   381 ,   1849    ,   2563    ,0 };
            const unsigned long dim456Kuo2Init[]        =   {   1   ,   3   ,   3   ,   1   ,   29  ,   31  ,   107 ,   151 ,   171 ,   137 ,   1749    ,   1541    ,0 };
            const unsigned long dim457Kuo2Init[]        =   {   1   ,   3   ,   7   ,   1   ,   13  ,   53  ,   27  ,   119 ,   295 ,   879 ,   1113    ,   1293    ,0 };
            const unsigned long dim458Kuo2Init[]        =   {   1   ,   1   ,   7   ,   9   ,   19  ,   25  ,   127 ,   215 ,   271 ,   521 ,   1093    ,   1595    ,0 };
            const unsigned long dim459Kuo2Init[]        =   {   1   ,   1   ,   5   ,   11  ,   11  ,   9   ,   35  ,   55  ,   447 ,   513 ,   1415    ,   2427    ,0 };
            const unsigned long dim460Kuo2Init[]        =   {   1   ,   3   ,   1   ,   5   ,   19  ,   19  ,   19  ,   111 ,   73  ,   895 ,   1879    ,   3679    ,0 };
            const unsigned long dim461Kuo2Init[]        =   {   1   ,   1   ,   7   ,   13  ,   13  ,   55  ,   65  ,   131 ,   461 ,   775 ,   1817    ,   1795    ,0 };
            const unsigned long dim462Kuo2Init[]        =   {   1   ,   3   ,   7   ,   9   ,   31  ,   41  ,   119 ,   75  ,   419 ,   319 ,   1463    ,   3491    ,0 };
            const unsigned long dim463Kuo2Init[]        =   {   1   ,   1   ,   7   ,   7   ,   27  ,   33  ,   71  ,   223 ,   381 ,   181 ,   541 ,   1695    ,0 };
            const unsigned long dim464Kuo2Init[]        =   {   1   ,   3   ,   7   ,   5   ,   29  ,   53  ,   109 ,   109 ,   35  ,   607 ,   559 ,   4045    ,0 };
            const unsigned long dim465Kuo2Init[]        =   {   1   ,   3   ,   1   ,   13  ,   11  ,   9   ,   9   ,   91  ,   85  ,   41  ,   1155    ,   3989    ,0 };
            const unsigned long dim466Kuo2Init[]        =   {   1   ,   1   ,   3   ,   15  ,   23  ,   13  ,   3   ,   243 ,   443 ,   671 ,   1461    ,   3329    ,0 };
            const unsigned long dim467Kuo2Init[]        =   {   1   ,   1   ,   1   ,   1   ,   27  ,   25  ,   37  ,   153 ,   299 ,   51  ,   371 ,   1269    ,0 };
            const unsigned long dim468Kuo2Init[]        =   {   1   ,   3   ,   7   ,   11  ,   13  ,   51  ,   85  ,   177 ,   489 ,   867 ,   779 ,   3423    ,0 };
            const unsigned long dim469Kuo2Init[]        =   {   1   ,   1   ,   7   ,   1   ,   25  ,   51  ,   105 ,   117 ,   105 ,   743 ,   67  ,   2889    ,0 };
            const unsigned long dim470Kuo2Init[]        =   {   1   ,   1   ,   3   ,   3   ,   29  ,   31  ,   65  ,   11  ,   87  ,   647 ,   193 ,   1229    ,0 };
            const unsigned long dim471Kuo2Init[]        =   {   1   ,   3   ,   5   ,   9   ,   19  ,   19  ,   19  ,   193 ,   441 ,   947 ,   1385    ,   1321    ,0 };
            const unsigned long dim472Kuo2Init[]        =   {   1   ,   1   ,   1   ,   9   ,   9   ,   31  ,   109 ,   77  ,   449 ,   775 ,   187 ,   2609    ,0 };
            const unsigned long dim473Kuo2Init[]        =   {   1   ,   1   ,   1   ,   13  ,   31  ,   57  ,   121 ,   251 ,   9   ,   673 ,   473 ,   755 ,0 };
            const unsigned long dim474Kuo2Init[]        =   {   1   ,   3   ,   3   ,   7   ,   25  ,   21  ,   93  ,   183 ,   277 ,   167 ,   1305    ,   897 ,0 };
            const unsigned long dim475Kuo2Init[]        =   {   1   ,   1   ,   5   ,   5   ,   29  ,   15  ,   29  ,   67  ,   193 ,   525 ,   851 ,   2287    ,0 };
            const unsigned long dim476Kuo2Init[]        =   {   1   ,   1   ,   5   ,   9   ,   7   ,   53  ,   117 ,   203 ,   89  ,   221 ,   235 ,   2783    ,0 };
            const unsigned long dim477Kuo2Init[]        =   {   1   ,   3   ,   5   ,   13  ,   7   ,   25  ,   69  ,   25  ,   329 ,   151 ,   1721    ,   1279    ,0 };
            const unsigned long dim478Kuo2Init[]        =   {   1   ,   1   ,   1   ,   7   ,   15  ,   9   ,   119 ,   75  ,   95  ,   297 ,   1149    ,   1385    ,0 };
            const unsigned long dim479Kuo2Init[]        =   {   1   ,   3   ,   3   ,   15  ,   23  ,   1   ,   109 ,   81  ,   249 ,   189 ,   737 ,   1809    ,0 };
            const unsigned long dim480Kuo2Init[]        =   {   1   ,   1   ,   7   ,   5   ,   29  ,   53  ,   77  ,   79  ,   339 ,   625 ,   1131    ,   2539    ,0 };
            const unsigned long dim481Kuo2Init[]        =   {   1   ,   1   ,   7   ,   15  ,   23  ,   35  ,   39  ,   25  ,   287 ,   321 ,   151 ,   193 ,   6637    ,0 };
            const unsigned long dim482Kuo2Init[]        =   {   1   ,   3   ,   5   ,   9   ,   15  ,   31  ,   31  ,   111 ,   393 ,   97  ,   407 ,   1087    ,   1413    ,0 };
            const unsigned long dim483Kuo2Init[]        =   {   1   ,   1   ,   1   ,   13  ,   21  ,   63  ,   71  ,   193 ,   443 ,   541 ,   1989    ,   2979    ,   7087    ,0 };
            const unsigned long dim484Kuo2Init[]        =   {   1   ,   3   ,   3   ,   1   ,   11  ,   41  ,   19  ,   219 ,   497 ,   273 ,   1293    ,   2233    ,   2967    ,0 };
            const unsigned long dim485Kuo2Init[]        =   {   1   ,   1   ,   5   ,   7   ,   23  ,   15  ,   119 ,   245 ,   223 ,   773 ,   1261    ,   1623    ,   8067    ,0 };
            const unsigned long dim486Kuo2Init[]        =   {   1   ,   1   ,   1   ,   15  ,   3   ,   39  ,   39  ,   11  ,   141 ,   347 ,   1473    ,   3503    ,   3595    ,0 };
            const unsigned long dim487Kuo2Init[]        =   {   1   ,   1   ,   1   ,   5   ,   27  ,   15  ,   81  ,   159 ,   279 ,   907 ,   143 ,   3195    ,   7127    ,0 };
            const unsigned long dim488Kuo2Init[]        =   {   1   ,   1   ,   1   ,   9   ,   27  ,   29  ,   109 ,   29  ,   147 ,   255 ,   1851    ,   677 ,   2883    ,0 };
            const unsigned long dim489Kuo2Init[]        =   {   1   ,   1   ,   5   ,   1   ,   1   ,   23  ,   3   ,   107 ,   9   ,   757 ,   319 ,   949 ,   3887    ,0 };
            const unsigned long dim490Kuo2Init[]        =   {   1   ,   1   ,   7   ,   1   ,   13  ,   41  ,   21  ,   9   ,   53  ,   179 ,   1273    ,   17  ,   5195    ,0 };
            const unsigned long dim491Kuo2Init[]        =   {   1   ,   3   ,   3   ,   15  ,   9   ,   41  ,   121 ,   13  ,   245 ,   415 ,   2041    ,   3813    ,   6807    ,0 };
            const unsigned long dim492Kuo2Init[]        =   {   1   ,   3   ,   5   ,   13  ,   7   ,   5   ,   43  ,   137 ,   191 ,   855 ,   735 ,   3711    ,   2459    ,0 };
            const unsigned long dim493Kuo2Init[]        =   {   1   ,   3   ,   3   ,   13  ,   15  ,   53  ,   13  ,   245 ,   217 ,   159 ,   271 ,   261 ,   7383    ,0 };
            const unsigned long dim494Kuo2Init[]        =   {   1   ,   3   ,   1   ,   9   ,   1   ,   39  ,   79  ,   73  ,   349 ,   367 ,   51  ,   155 ,   53  ,0 };
            const unsigned long dim495Kuo2Init[]        =   {   1   ,   1   ,   7   ,   5   ,   11  ,   19  ,   99  ,   131 ,   217 ,   729 ,   1481    ,   4015    ,   6109    ,0 };
            const unsigned long dim496Kuo2Init[]        =   {   1   ,   3   ,   1   ,   3   ,   9   ,   7   ,   55  ,   119 ,   285 ,   803 ,   427 ,   3755    ,   3153    ,0 };
            const unsigned long dim497Kuo2Init[]        =   {   1   ,   1   ,   7   ,   13  ,   25  ,   3   ,   111 ,   51  ,   1   ,   995 ,   657 ,   1551    ,   3633    ,0 };
            const unsigned long dim498Kuo2Init[]        =   {   1   ,   3   ,   3   ,   1   ,   29  ,   49  ,   15  ,   219 ,   295 ,   335 ,   191 ,   611 ,   1151    ,0 };
            const unsigned long dim499Kuo2Init[]        =   {   1   ,   3   ,   3   ,   3   ,   5   ,   37  ,   115 ,   199 ,   143 ,   11  ,   1487    ,   3841    ,   4669    ,0 };
            const unsigned long dim500Kuo2Init[]        =   {   1   ,   1   ,   3   ,   5   ,   23  ,   33  ,   67  ,   51  ,   433 ,   351 ,   1467    ,   2935    ,   4135    ,0 };
            const unsigned long dim501Kuo2Init[]        =   {   1   ,   3   ,   3   ,   15  ,   3   ,   57  ,   55  ,   7   ,   487 ,   983 ,   1481    ,   799 ,   6141    ,0 };
            const unsigned long dim502Kuo2Init[]        =   {   1   ,   3   ,   5   ,   3   ,   23  ,   1   ,   37  ,   117 ,   131 ,   55  ,   449 ,   637 ,   693 ,0 };
            const unsigned long dim503Kuo2Init[]        =   {   1   ,   3   ,   1   ,   9   ,   11  ,   7   ,   43  ,   149 ,   463 ,   349 ,   1773    ,   643 ,   765 ,0 };
            const unsigned long dim504Kuo2Init[]        =   {   1   ,   3   ,   1   ,   1   ,   25  ,   39  ,   1   ,   33  ,   209 ,   131 ,   157 ,   2695    ,   623 ,0 };
            const unsigned long dim505Kuo2Init[]        =   {   1   ,   1   ,   1   ,   7   ,   15  ,   41  ,   105 ,   51  ,   465 ,   453 ,   1091    ,   905 ,   4831    ,0 };
            const unsigned long dim506Kuo2Init[]        =   {   1   ,   3   ,   7   ,   9   ,   1   ,   39  ,   45  ,   117 ,   481 ,   481 ,   87  ,   1869    ,   7063    ,0 };
            const unsigned long dim507Kuo2Init[]        =   {   1   ,   1   ,   7   ,   1   ,   27  ,   1   ,   87  ,   75  ,   295 ,   287 ,   163 ,   911 ,   1725    ,0 };
            const unsigned long dim508Kuo2Init[]        =   {   1   ,   3   ,   3   ,   9   ,   31  ,   61  ,   75  ,   53  ,   63  ,   179 ,   1941    ,   555 ,   1303    ,0 };
            const unsigned long dim509Kuo2Init[]        =   {   1   ,   3   ,   3   ,   9   ,   31  ,   37  ,   9   ,   51  ,   381 ,   639 ,   595 ,   731 ,   7531    ,0 };
            const unsigned long dim510Kuo2Init[]        =   {   1   ,   3   ,   1   ,   1   ,   3   ,   25  ,   59  ,   19  ,   187 ,   235 ,   173 ,   893 ,   4845    ,0 };
            const unsigned long dim511Kuo2Init[]        =   {   1   ,   1   ,   7   ,   7   ,   1   ,   23  ,   9   ,   9   ,   493 ,   643 ,   53  ,   179 ,   6993    ,0 };
            const unsigned long dim512Kuo2Init[]        =   {   1   ,   1   ,   1   ,   1   ,   25  ,   55  ,   81  ,   63  ,   67  ,   967 ,   957 ,   1025    ,   8033    ,0 };
            const unsigned long dim513Kuo2Init[]        =   {   1   ,   3   ,   1   ,   7   ,   5   ,   43  ,   67  ,   127 ,   103 ,   241 ,   77  ,   43  ,   5173    ,0 };
            const unsigned long dim514Kuo2Init[]        =   {   1   ,   1   ,   3   ,   7   ,   9   ,   25  ,   17  ,   231 ,   315 ,   699 ,   787 ,   591 ,   7257    ,0 };
            const unsigned long dim515Kuo2Init[]        =   {   1   ,   3   ,   1   ,   15  ,   21  ,   19  ,   7   ,   179 ,   335 ,   701 ,   323 ,   121 ,   4733    ,0 };
            const unsigned long dim516Kuo2Init[]        =   {   1   ,   1   ,   3   ,   7   ,   11  ,   59  ,   75  ,   147 ,   391 ,   535 ,   1841    ,   3025    ,   5621    ,0 };
            const unsigned long dim517Kuo2Init[]        =   {   1   ,   1   ,   3   ,   7   ,   21  ,   25  ,   101 ,   71  ,   195 ,   495 ,   1241    ,   2229    ,   2865    ,0 };
            const unsigned long dim518Kuo2Init[]        =   {   1   ,   3   ,   1   ,   5   ,   19  ,   55  ,   93  ,   41  ,   287 ,   787 ,   929 ,   3161    ,   4247    ,0 };
            const unsigned long dim519Kuo2Init[]        =   {   1   ,   3   ,   3   ,   11  ,   25  ,   31  ,   79  ,   225 ,   189 ,   187 ,   595 ,   465 ,   7169    ,0 };
            const unsigned long dim520Kuo2Init[]        =   {   1   ,   1   ,   3   ,   13  ,   25  ,   37  ,   7   ,   87  ,   511 ,   451 ,   1681    ,   1911    ,   5269    ,0 };
            const unsigned long dim521Kuo2Init[]        =   {   1   ,   3   ,   3   ,   13  ,   21  ,   31  ,   117 ,   73  ,   391 ,   127 ,   1201    ,   189 ,   6621    ,0 };
            const unsigned long dim522Kuo2Init[]        =   {   1   ,   3   ,   1   ,   7   ,   9   ,   49  ,   77  ,   91  ,   197 ,   889 ,   119 ,   1335    ,   4833    ,0 };
            const unsigned long dim523Kuo2Init[]        =   {   1   ,   1   ,   7   ,   3   ,   11  ,   61  ,   103 ,   19  ,   345 ,   937 ,   1393    ,   963 ,   2283    ,0 };
            const unsigned long dim524Kuo2Init[]        =   {   1   ,   1   ,   1   ,   9   ,   3   ,   3   ,   113 ,   189 ,   383 ,   727 ,   781 ,   3515    ,   5045    ,0 };
            const unsigned long dim525Kuo2Init[]        =   {   1   ,   1   ,   7   ,   15  ,   27  ,   45  ,   65  ,   137 ,   5   ,   185 ,   1561    ,   2063    ,   761 ,0 };
            const unsigned long dim526Kuo2Init[]        =   {   1   ,   1   ,   3   ,   9   ,   17  ,   5   ,   71  ,   209 ,   387 ,   693 ,   773 ,   2105    ,   1265    ,0 };
            const unsigned long dim527Kuo2Init[]        =   {   1   ,   1   ,   3   ,   11  ,   9   ,   49  ,   53  ,   53  ,   7   ,   5   ,   543 ,   1605    ,   265 ,0 };
            const unsigned long dim528Kuo2Init[]        =   {   1   ,   3   ,   3   ,   3   ,   23  ,   21  ,   27  ,   187 ,   495 ,   101 ,   517 ,   853 ,   2801    ,0 };
            const unsigned long dim529Kuo2Init[]        =   {   1   ,   3   ,   5   ,   1   ,   7   ,   61  ,   47  ,   113 ,   375 ,   307 ,   255 ,   3341    ,   3663    ,0 };
            const unsigned long dim530Kuo2Init[]        =   {   1   ,   1   ,   1   ,   1   ,   19  ,   9   ,   33  ,   235 ,   413 ,   999 ,   1889    ,   3633    ,   745 ,0 };
            const unsigned long dim531Kuo2Init[]        =   {   1   ,   1   ,   1   ,   7   ,   27  ,   27  ,   19  ,   57  ,   1   ,   639 ,   831 ,   3481    ,   395 ,0 };
            const unsigned long dim532Kuo2Init[]        =   {   1   ,   3   ,   3   ,   5   ,   19  ,   55  ,   27  ,   215 ,   319 ,   855 ,   1597    ,   1539    ,   4359    ,0 };
            const unsigned long dim533Kuo2Init[]        =   {   1   ,   3   ,   7   ,   11  ,   3   ,   35  ,   5   ,   3   ,   341 ,   795 ,   315 ,   935 ,   2417    ,0 };
            const unsigned long dim534Kuo2Init[]        =   {   1   ,   1   ,   1   ,   3   ,   25  ,   55  ,   53  ,   143 ,   361 ,   203 ,   863 ,   2491    ,   4701    ,0 };
            const unsigned long dim535Kuo2Init[]        =   {   1   ,   1   ,   5   ,   9   ,   19  ,   9   ,   29  ,   153 ,   387 ,   373 ,   647 ,   2825    ,   1391    ,0 };
            const unsigned long dim536Kuo2Init[]        =   {   1   ,   1   ,   3   ,   9   ,   1   ,   37  ,   103 ,   17  ,   347 ,   577 ,   181 ,   2863    ,   6525    ,0 };
            const unsigned long dim537Kuo2Init[]        =   {   1   ,   1   ,   1   ,   3   ,   17  ,   19  ,   41  ,   61  ,   49  ,   75  ,   1243    ,   33  ,   7725    ,0 };
            const unsigned long dim538Kuo2Init[]        =   {   1   ,   3   ,   1   ,   9   ,   7   ,   23  ,   107 ,   227 ,   225 ,   139 ,   451 ,   1303    ,   5657    ,0 };
            const unsigned long dim539Kuo2Init[]        =   {   1   ,   3   ,   5   ,   11  ,   27  ,   37  ,   95  ,   195 ,   261 ,   827 ,   817 ,   2555    ,   8189    ,0 };
            const unsigned long dim540Kuo2Init[]        =   {   1   ,   3   ,   1   ,   11  ,   13  ,   17  ,   73  ,   113 ,   287 ,   73  ,   977 ,   2815    ,   4281    ,0 };
            const unsigned long dim541Kuo2Init[]        =   {   1   ,   1   ,   5   ,   7   ,   23  ,   7   ,   103 ,   41  ,   237 ,   745 ,   1645    ,   2041    ,   6063    ,0 };
            const unsigned long dim542Kuo2Init[]        =   {   1   ,   1   ,   3   ,   15  ,   17  ,   1   ,   63  ,   115 ,   303 ,   175 ,   1511    ,   3027    ,   1265    ,0 };
            const unsigned long dim543Kuo2Init[]        =   {   1   ,   3   ,   5   ,   3   ,   29  ,   57  ,   1   ,   51  ,   127 ,   539 ,   1171    ,   3567    ,   6409    ,0 };
            const unsigned long dim544Kuo2Init[]        =   {   1   ,   1   ,   1   ,   9   ,   29  ,   63  ,   85  ,   127 ,   375 ,   453 ,   539 ,   563 ,   5709    ,0 };
            const unsigned long dim545Kuo2Init[]        =   {   1   ,   1   ,   1   ,   7   ,   29  ,   31  ,   109 ,   159 ,   15  ,   621 ,   1593    ,   3793    ,   4503    ,0 };
            const unsigned long dim546Kuo2Init[]        =   {   1   ,   1   ,   3   ,   3   ,   17  ,   27  ,   59  ,   9   ,   263 ,   705 ,   909 ,   643 ,   4967    ,0 };
            const unsigned long dim547Kuo2Init[]        =   {   1   ,   3   ,   5   ,   7   ,   23  ,   13  ,   37  ,   179 ,   439 ,   77  ,   1055    ,   3129    ,   371 ,0 };
            const unsigned long dim548Kuo2Init[]        =   {   1   ,   3   ,   1   ,   11  ,   25  ,   1   ,   123 ,   109 ,   27  ,   393 ,   1729    ,   3597    ,   5197    ,0 };
            const unsigned long dim549Kuo2Init[]        =   {   1   ,   3   ,   5   ,   11  ,   15  ,   9   ,   25  ,   19  ,   1   ,   927 ,   39  ,   925 ,   87  ,0 };
            const unsigned long dim550Kuo2Init[]        =   {   1   ,   3   ,   5   ,   1   ,   9   ,   7   ,   85  ,   95  ,   497 ,   487 ,   311 ,   2109    ,   259 ,0 };
            const unsigned long dim551Kuo2Init[]        =   {   1   ,   3   ,   5   ,   1   ,   5   ,   59  ,   75  ,   243 ,   179 ,   975 ,   187 ,   415 ,   1161    ,0 };
            const unsigned long dim552Kuo2Init[]        =   {   1   ,   3   ,   3   ,   13  ,   11  ,   61  ,   109 ,   129 ,   71  ,   917 ,   523 ,   1189    ,   4857    ,0 };
            const unsigned long dim553Kuo2Init[]        =   {   1   ,   3   ,   1   ,   1   ,   3   ,   57  ,   15  ,   79  ,   265 ,   839 ,   559 ,   729 ,   2737    ,0 };
            const unsigned long dim554Kuo2Init[]        =   {   1   ,   1   ,   7   ,   9   ,   21  ,   45  ,   69  ,   245 ,   331 ,   197 ,   1947    ,   3145    ,   6927    ,0 };
            const unsigned long dim555Kuo2Init[]        =   {   1   ,   3   ,   7   ,   7   ,   31  ,   41  ,   73  ,   59  ,   93  ,   565 ,   1885    ,   3803    ,   7013    ,0 };
            const unsigned long dim556Kuo2Init[]        =   {   1   ,   1   ,   1   ,   13  ,   23  ,   19  ,   107 ,   77  ,   449 ,   965 ,   99  ,   1215    ,   4699    ,0 };
            const unsigned long dim557Kuo2Init[]        =   {   1   ,   1   ,   3   ,   13  ,   11  ,   3   ,   37  ,   209 ,   221 ,   197 ,   1913    ,   631 ,   1621    ,0 };
            const unsigned long dim558Kuo2Init[]        =   {   1   ,   1   ,   5   ,   9   ,   27  ,   51  ,   61  ,   249 ,   269 ,   507 ,   1005    ,   2471    ,   4621    ,0 };
            const unsigned long dim559Kuo2Init[]        =   {   1   ,   3   ,   5   ,   1   ,   3   ,   49  ,   69  ,   227 ,   289 ,   261 ,   1103    ,   389 ,   4883    ,0 };
            const unsigned long dim560Kuo2Init[]        =   {   1   ,   1   ,   5   ,   3   ,   29  ,   35  ,   43  ,   167 ,   451 ,   1017    ,   865 ,   1507    ,   7387    ,0 };
            const unsigned long dim561Kuo2Init[]        =   {   1   ,   1   ,   3   ,   15  ,   13  ,   1   ,   53  ,   9   ,   23  ,   581 ,   627 ,   3823    ,   6767    ,0 };
            const unsigned long dim562Kuo2Init[]        =   {   1   ,   3   ,   1   ,   7   ,   3   ,   21  ,   115 ,   73  ,   353 ,   725 ,   1971    ,   1623    ,   7359    ,0 };
            const unsigned long dim563Kuo2Init[]        =   {   1   ,   1   ,   3   ,   5   ,   17  ,   15  ,   37  ,   219 ,   187 ,   699 ,   1463    ,   1563    ,   2785    ,0 };
            const unsigned long dim564Kuo2Init[]        =   {   1   ,   3   ,   7   ,   13  ,   11  ,   43  ,   9   ,   49  ,   125 ,   681 ,   499 ,   2127    ,   6613    ,0 };
            const unsigned long dim565Kuo2Init[]        =   {   1   ,   3   ,   5   ,   11  ,   5   ,   21  ,   115 ,   227 ,   305 ,   247 ,   1947    ,   853 ,   5357    ,0 };
            const unsigned long dim566Kuo2Init[]        =   {   1   ,   3   ,   1   ,   9   ,   29  ,   37  ,   1   ,   181 ,   427 ,   435 ,   1757    ,   1535    ,   4383    ,0 };
            const unsigned long dim567Kuo2Init[]        =   {   1   ,   3   ,   1   ,   15  ,   15  ,   41  ,   123 ,   7   ,   367 ,   767 ,   777 ,   3455    ,   7739    ,0 };
            const unsigned long dim568Kuo2Init[]        =   {   1   ,   1   ,   5   ,   9   ,   13  ,   3   ,   5   ,   79  ,   3   ,   121 ,   1113    ,   1731    ,   501 ,0 };
            const unsigned long dim569Kuo2Init[]        =   {   1   ,   1   ,   3   ,   13  ,   1   ,   13  ,   127 ,   215 ,   329 ,   381 ,   205 ,   2391    ,   581 ,0 };
            const unsigned long dim570Kuo2Init[]        =   {   1   ,   1   ,   3   ,   7   ,   7   ,   19  ,   9   ,   23  ,   463 ,   455 ,   1921    ,   763 ,   5077    ,0 };
            const unsigned long dim571Kuo2Init[]        =   {   1   ,   1   ,   3   ,   3   ,   13  ,   23  ,   65  ,   7   ,   337 ,   603 ,   1243    ,   2161    ,   5883    ,0 };
            const unsigned long dim572Kuo2Init[]        =   {   1   ,   3   ,   5   ,   1   ,   15  ,   23  ,   73  ,   151 ,   243 ,   847 ,   1189    ,   1663    ,   2011    ,0 };
            const unsigned long dim573Kuo2Init[]        =   {   1   ,   1   ,   7   ,   7   ,   11  ,   17  ,   91  ,   219 ,   127 ,   317 ,   1807    ,   1369    ,   7433    ,0 };
            const unsigned long dim574Kuo2Init[]        =   {   1   ,   3   ,   5   ,   9   ,   31  ,   49  ,   65  ,   47  ,   11  ,   541 ,   55  ,   2743    ,   4987    ,0 };
            const unsigned long dim575Kuo2Init[]        =   {   1   ,   3   ,   7   ,   13  ,   23  ,   23  ,   51  ,   129 ,   111 ,   319 ,   1195    ,   1593    ,   4407    ,0 };
            const unsigned long dim576Kuo2Init[]        =   {   1   ,   3   ,   1   ,   7   ,   21  ,   25  ,   35  ,   187 ,   409 ,   981 ,   293 ,   873 ,   7447    ,0 };
            const unsigned long dim577Kuo2Init[]        =   {   1   ,   3   ,   1   ,   9   ,   3   ,   11  ,   25  ,   75  ,   347 ,   907 ,   1645    ,   2997    ,   5953    ,0 };
            const unsigned long dim578Kuo2Init[]        =   {   1   ,   3   ,   7   ,   7   ,   25  ,   49  ,   99  ,   37  ,   45  ,   81  ,   95  ,   2637    ,   1337    ,0 };
            const unsigned long dim579Kuo2Init[]        =   {   1   ,   3   ,   5   ,   3   ,   29  ,   35  ,   85  ,   167 ,   503 ,   257 ,   1101    ,   2647    ,   1469    ,0 };
            const unsigned long dim580Kuo2Init[]        =   {   1   ,   3   ,   5   ,   1   ,   27  ,   17  ,   93  ,   197 ,   485 ,   899 ,   1201    ,   3633    ,   1289    ,0 };
            const unsigned long dim581Kuo2Init[]        =   {   1   ,   3   ,   1   ,   7   ,   5   ,   17  ,   53  ,   167 ,   353 ,   1023    ,   1063    ,   2221    ,   3635    ,0 };
            const unsigned long dim582Kuo2Init[]        =   {   1   ,   1   ,   5   ,   11  ,   9   ,   43  ,   123 ,   241 ,   77  ,   657 ,   1789    ,   235 ,   2105    ,0 };
            const unsigned long dim583Kuo2Init[]        =   {   1   ,   1   ,   5   ,   9   ,   9   ,   53  ,   109 ,   185 ,   443 ,   803 ,   115 ,   2079    ,   7839    ,0 };
            const unsigned long dim584Kuo2Init[]        =   {   1   ,   1   ,   1   ,   13  ,   29  ,   29  ,   35  ,   235 ,   253 ,   529 ,   153 ,   1613    ,   6723    ,0 };
            const unsigned long dim585Kuo2Init[]        =   {   1   ,   3   ,   1   ,   13  ,   11  ,   31  ,   81  ,   177 ,   135 ,   659 ,   605 ,   1253    ,   5077    ,0 };
            const unsigned long dim586Kuo2Init[]        =   {   1   ,   3   ,   5   ,   3   ,   25  ,   53  ,   87  ,   183 ,   285 ,   639 ,   657 ,   2307    ,   5477    ,0 };
            const unsigned long dim587Kuo2Init[]        =   {   1   ,   1   ,   7   ,   7   ,   9   ,   35  ,   31  ,   83  ,   165 ,   735 ,   1445    ,   3935    ,   515 ,0 };
            const unsigned long dim588Kuo2Init[]        =   {   1   ,   3   ,   5   ,   1   ,   13  ,   59  ,   99  ,   191 ,   187 ,   375 ,   959 ,   1145    ,   1487    ,0 };
            const unsigned long dim589Kuo2Init[]        =   {   1   ,   3   ,   7   ,   3   ,   1   ,   1   ,   127 ,   127 ,   101 ,   583 ,   121 ,   2879    ,   63  ,0 };
            const unsigned long dim590Kuo2Init[]        =   {   1   ,   3   ,   7   ,   11  ,   13  ,   1   ,   125 ,   81  ,   31  ,   177 ,   107 ,   2339    ,   755 ,0 };
            const unsigned long dim591Kuo2Init[]        =   {   1   ,   1   ,   3   ,   13  ,   7   ,   33  ,   9   ,   247 ,   395 ,   255 ,   1817    ,   1509    ,   5401    ,0 };
            const unsigned long dim592Kuo2Init[]        =   {   1   ,   1   ,   7   ,   7   ,   27  ,   19  ,   87  ,   167 ,   425 ,   1019    ,   1419    ,   4025    ,   5483    ,0 };
            const unsigned long dim593Kuo2Init[]        =   {   1   ,   3   ,   7   ,   7   ,   5   ,   49  ,   97  ,   65  ,   355 ,   207 ,   323 ,   2367    ,   171 ,0 };
            const unsigned long dim594Kuo2Init[]        =   {   1   ,   3   ,   7   ,   5   ,   23  ,   5   ,   125 ,   131 ,   115 ,   585 ,   1533    ,   3489    ,   507 ,0 };
            const unsigned long dim595Kuo2Init[]        =   {   1   ,   3   ,   3   ,   15  ,   29  ,   51  ,   79  ,   239 ,   289 ,   909 ,   1339    ,   857 ,   1661    ,0 };
            const unsigned long dim596Kuo2Init[]        =   {   1   ,   3   ,   5   ,   11  ,   23  ,   1   ,   65  ,   123 ,   439 ,   349 ,   1593    ,   1913    ,   433 ,0 };
            const unsigned long dim597Kuo2Init[]        =   {   1   ,   1   ,   7   ,   5   ,   27  ,   45  ,   59  ,   33  ,   375 ,   553 ,   1955    ,   731 ,   4335    ,0 };
            const unsigned long dim598Kuo2Init[]        =   {   1   ,   1   ,   3   ,   7   ,   25  ,   63  ,   51  ,   183 ,   67  ,   783 ,   885 ,   3385    ,   7109    ,0 };
            const unsigned long dim599Kuo2Init[]        =   {   1   ,   3   ,   7   ,   9   ,   3   ,   43  ,   105 ,   145 ,   123 ,   503 ,   203 ,   1029    ,   6713    ,0 };
            const unsigned long dim600Kuo2Init[]        =   {   1   ,   3   ,   7   ,   7   ,   11  ,   49  ,   113 ,   69  ,   489 ,   649 ,   1393    ,   1475    ,   6663    ,0 };
            const unsigned long dim601Kuo2Init[]        =   {   1   ,   1   ,   5   ,   11  ,   13  ,   5   ,   105 ,   21  ,   451 ,   137 ,   2029    ,   1459    ,   7981    ,0 };
            const unsigned long dim602Kuo2Init[]        =   {   1   ,   1   ,   7   ,   5   ,   7   ,   17  ,   3   ,   213 ,   255 ,   867 ,   89  ,   2437    ,   1151    ,0 };
            const unsigned long dim603Kuo2Init[]        =   {   1   ,   1   ,   7   ,   9   ,   19  ,   23  ,   115 ,   161 ,   257 ,   419 ,   1133    ,   1291    ,   243 ,0 };
            const unsigned long dim604Kuo2Init[]        =   {   1   ,   3   ,   5   ,   7   ,   27  ,   45  ,   23  ,   157 ,   371 ,   219 ,   1001    ,   343 ,   2419    ,0 };
            const unsigned long dim605Kuo2Init[]        =   {   1   ,   3   ,   1   ,   1   ,   21  ,   59  ,   91  ,   153 ,   19  ,   469 ,   1923    ,   2439    ,   3837    ,0 };
            const unsigned long dim606Kuo2Init[]        =   {   1   ,   3   ,   7   ,   3   ,   23  ,   29  ,   57  ,   201 ,   209 ,   249 ,   63  ,   2929    ,   2719    ,0 };
            const unsigned long dim607Kuo2Init[]        =   {   1   ,   1   ,   5   ,   3   ,   11  ,   3   ,   75  ,   91  ,   157 ,   557 ,   1775    ,   387 ,   803 ,0 };
            const unsigned long dim608Kuo2Init[]        =   {   1   ,   3   ,   3   ,   3   ,   27  ,   57  ,   69  ,   135 ,   317 ,   897 ,   1417    ,   1853    ,   747 ,0 };
            const unsigned long dim609Kuo2Init[]        =   {   1   ,   1   ,   7   ,   11  ,   15  ,   19  ,   115 ,   219 ,   57  ,   925 ,   1869    ,   609 ,   1631    ,0 };
            const unsigned long dim610Kuo2Init[]        =   {   1   ,   1   ,   1   ,   11  ,   29  ,   21  ,   45  ,   19  ,   91  ,   271 ,   971 ,   2569    ,   423 ,0 };
            const unsigned long dim611Kuo2Init[]        =   {   1   ,   3   ,   3   ,   13  ,   5   ,   9   ,   5   ,   101 ,   147 ,   777 ,   1647    ,   457 ,   4053    ,0 };
            const unsigned long dim612Kuo2Init[]        =   {   1   ,   1   ,   5   ,   9   ,   9   ,   25  ,   13  ,   175 ,   53  ,   467 ,   1555    ,   2969    ,   1669    ,0 };
            const unsigned long dim613Kuo2Init[]        =   {   1   ,   1   ,   1   ,   7   ,   21  ,   49  ,   1   ,   49  ,   317 ,   505 ,   1495    ,   1053    ,   2293    ,0 };
            const unsigned long dim614Kuo2Init[]        =   {   1   ,   1   ,   1   ,   3   ,   15  ,   43  ,   91  ,   207 ,   211 ,   725 ,   2039    ,   803 ,   4029    ,0 };
            const unsigned long dim615Kuo2Init[]        =   {   1   ,   1   ,   7   ,   13  ,   17  ,   45  ,   39  ,   125 ,   129 ,   169 ,   47  ,   2105    ,   1749    ,0 };
            const unsigned long dim616Kuo2Init[]        =   {   1   ,   3   ,   7   ,   11  ,   19  ,   13  ,   5   ,   105 ,   203 ,   611 ,   147 ,   2291    ,   1823    ,0 };
            const unsigned long dim617Kuo2Init[]        =   {   1   ,   1   ,   5   ,   1   ,   5   ,   15  ,   15  ,   35  ,   33  ,   467 ,   59  ,   2435    ,   3141    ,0 };
            const unsigned long dim618Kuo2Init[]        =   {   1   ,   3   ,   1   ,   5   ,   29  ,   43  ,   35  ,   3   ,   173 ,   619 ,   1595    ,   347 ,   2513    ,0 };
            const unsigned long dim619Kuo2Init[]        =   {   1   ,   3   ,   3   ,   1   ,   23  ,   37  ,   3   ,   123 ,   301 ,   403 ,   1603    ,   2345    ,   1031    ,0 };
            const unsigned long dim620Kuo2Init[]        =   {   1   ,   3   ,   5   ,   9   ,   23  ,   55  ,   23  ,   19  ,   497 ,   501 ,   99  ,   1925    ,   6243    ,0 };
            const unsigned long dim621Kuo2Init[]        =   {   1   ,   3   ,   5   ,   1   ,   21  ,   37  ,   83  ,   59  ,   285 ,   107 ,   1909    ,   3647    ,   8041    ,0 };
            const unsigned long dim622Kuo2Init[]        =   {   1   ,   3   ,   5   ,   1   ,   7   ,   11  ,   91  ,   1   ,   357 ,   643 ,   1319    ,   2895    ,   2193    ,0 };
            const unsigned long dim623Kuo2Init[]        =   {   1   ,   1   ,   5   ,   9   ,   27  ,   29  ,   25  ,   125 ,   433 ,   525 ,   1511    ,   585 ,   5263    ,0 };
            const unsigned long dim624Kuo2Init[]        =   {   1   ,   3   ,   1   ,   11  ,   13  ,   49  ,   79  ,   37  ,   165 ,   923 ,   749 ,   2207    ,   4673    ,0 };
            const unsigned long dim625Kuo2Init[]        =   {   1   ,   1   ,   5   ,   9   ,   29  ,   37  ,   51  ,   3   ,   133 ,   859 ,   153 ,   1199    ,   2269    ,0 };
            const unsigned long dim626Kuo2Init[]        =   {   1   ,   1   ,   7   ,   5   ,   9   ,   57  ,   101 ,   231 ,   383 ,   245 ,   1997    ,   2661    ,   6069    ,0 };
            const unsigned long dim627Kuo2Init[]        =   {   1   ,   1   ,   7   ,   7   ,   25  ,   61  ,   25  ,   185 ,   247 ,   509 ,   1953    ,   3309    ,   4879    ,0 };
            const unsigned long dim628Kuo2Init[]        =   {   1   ,   1   ,   3   ,   7   ,   19  ,   61  ,   95  ,   105 ,   355 ,   219 ,   103 ,   249 ,   579 ,0 };
            const unsigned long dim629Kuo2Init[]        =   {   1   ,   3   ,   7   ,   15  ,   7   ,   3   ,   51  ,   143 ,   273 ,   203 ,   353 ,   1587    ,   2239    ,0 };
            const unsigned long dim630Kuo2Init[]        =   {   1   ,   3   ,   3   ,   9   ,   27  ,   47  ,   123 ,   185 ,   337 ,   183 ,   983 ,   2313    ,   815 ,0 };
            const unsigned long dim631Kuo2Init[]        =   {   1   ,   1   ,   5   ,   15  ,   19  ,   19  ,   3   ,   199 ,   247 ,   151 ,   1911    ,   301 ,   5311    ,0 };
            const unsigned long dim632Kuo2Init[]        =   {   1   ,   1   ,   1   ,   3   ,   7   ,   63  ,   61  ,   151 ,   297 ,   541 ,   1457    ,   2945    ,   5703    ,0 };
            const unsigned long dim633Kuo2Init[]        =   {   1   ,   3   ,   7   ,   5   ,   31  ,   1   ,   109 ,   179 ,   197 ,   109 ,   695 ,   2429    ,   3483    ,0 };
            const unsigned long dim634Kuo2Init[]        =   {   1   ,   3   ,   7   ,   1   ,   27  ,   25  ,   27  ,   113 ,   455 ,   655 ,   1813    ,   2341    ,   4785    ,0 };
            const unsigned long dim635Kuo2Init[]        =   {   1   ,   1   ,   5   ,   15  ,   9   ,   47  ,   105 ,   15  ,   75  ,   901 ,   1539    ,   2599    ,   1013    ,0 };
            const unsigned long dim636Kuo2Init[]        =   {   1   ,   3   ,   1   ,   11  ,   17  ,   17  ,   29  ,   219 ,   121 ,   383 ,   1575    ,   3153    ,   6151    ,0 };
            const unsigned long dim637Kuo2Init[]        =   {   1   ,   1   ,   1   ,   3   ,   29  ,   51  ,   35  ,   95  ,   89  ,   279 ,   1255    ,   2967    ,   455 ,0 };
            const unsigned long dim638Kuo2Init[]        =   {   1   ,   3   ,   5   ,   15  ,   3   ,   3   ,   15  ,   181 ,   91  ,   169 ,   1703    ,   777 ,   6539    ,0 };
            const unsigned long dim639Kuo2Init[]        =   {   1   ,   1   ,   3   ,   13  ,   15  ,   29  ,   115 ,   225 ,   179 ,   271 ,   737 ,   2425    ,   3149    ,0 };
            const unsigned long dim640Kuo2Init[]        =   {   1   ,   3   ,   7   ,   3   ,   13  ,   29  ,   1   ,   169 ,   35  ,   145 ,   725 ,   3679    ,   2593    ,0 };
            const unsigned long dim641Kuo2Init[]        =   {   1   ,   1   ,   7   ,   9   ,   11  ,   1   ,   51  ,   65  ,   283 ,   791 ,   449 ,   1387    ,   7347    ,0 };
            const unsigned long dim642Kuo2Init[]        =   {   1   ,   3   ,   1   ,   13  ,   21  ,   23  ,   79  ,   207 ,   425 ,   839 ,   737 ,   243 ,   6761    ,0 };
            const unsigned long dim643Kuo2Init[]        =   {   1   ,   3   ,   3   ,   5   ,   11  ,   25  ,   93  ,   119 ,   245 ,   539 ,   405 ,   1925    ,   1595    ,0 };
            const unsigned long dim644Kuo2Init[]        =   {   1   ,   3   ,   7   ,   15  ,   5   ,   19  ,   97  ,   225 ,   471 ,   801 ,   379 ,   923 ,   1961    ,0 };
            const unsigned long dim645Kuo2Init[]        =   {   1   ,   3   ,   7   ,   15  ,   25  ,   49  ,   41  ,   237 ,   11  ,   983 ,   1889    ,   3947    ,   2579    ,0 };
            const unsigned long dim646Kuo2Init[]        =   {   1   ,   3   ,   5   ,   13  ,   3   ,   39  ,   51  ,   131 ,   477 ,   699 ,   1101    ,   1289    ,   7683    ,0 };
            const unsigned long dim647Kuo2Init[]        =   {   1   ,   1   ,   7   ,   1   ,   23  ,   39  ,   99  ,   167 ,   53  ,   333 ,   1059    ,   2947    ,   2097    ,0 };
            const unsigned long dim648Kuo2Init[]        =   {   1   ,   1   ,   5   ,   15  ,   9   ,   55  ,   33  ,   121 ,   485 ,   767 ,   793 ,   1015    ,   7657    ,0 };
            const unsigned long dim649Kuo2Init[]        =   {   1   ,   1   ,   7   ,   13  ,   15  ,   57  ,   5   ,   39  ,   197 ,   67  ,   1471    ,   841 ,   1797    ,0 };
            const unsigned long dim650Kuo2Init[]        =   {   1   ,   3   ,   5   ,   3   ,   15  ,   17  ,   11  ,   139 ,   21  ,   827 ,   465 ,   1341    ,   1115    ,0 };
            const unsigned long dim651Kuo2Init[]        =   {   1   ,   3   ,   7   ,   11  ,   7   ,   3   ,   29  ,   107 ,   79  ,   19  ,   493 ,   1705    ,   4631    ,0 };
            const unsigned long dim652Kuo2Init[]        =   {   1   ,   1   ,   7   ,   1   ,   9   ,   61  ,   79  ,   23  ,   83  ,   863 ,   1189    ,   2653    ,   2139    ,0 };
            const unsigned long dim653Kuo2Init[]        =   {   1   ,   1   ,   7   ,   7   ,   7   ,   39  ,   105 ,   147 ,   293 ,   799 ,   781 ,   1085    ,   1405    ,0 };
            const unsigned long dim654Kuo2Init[]        =   {   1   ,   3   ,   3   ,   13  ,   27  ,   55  ,   89  ,   73  ,   255 ,   845 ,   485 ,   739 ,   4937    ,0 };
            const unsigned long dim655Kuo2Init[]        =   {   1   ,   3   ,   1   ,   5   ,   19  ,   41  ,   93  ,   125 ,   203 ,   789 ,   1447    ,   189 ,   4743    ,0 };
            const unsigned long dim656Kuo2Init[]        =   {   1   ,   3   ,   3   ,   11  ,   9   ,   9   ,   51  ,   237 ,   353 ,   313 ,   101 ,   873 ,   5605    ,0 };
            const unsigned long dim657Kuo2Init[]        =   {   1   ,   3   ,   1   ,   5   ,   1   ,   33  ,   67  ,   29  ,   361 ,   945 ,   5   ,   305 ,   1461    ,0 };
            const unsigned long dim658Kuo2Init[]        =   {   1   ,   1   ,   7   ,   3   ,   5   ,   31  ,   47  ,   145 ,   369 ,   419 ,   415 ,   2645    ,   4665    ,0 };
            const unsigned long dim659Kuo2Init[]        =   {   1   ,   1   ,   1   ,   1   ,   9   ,   35  ,   3   ,   67  ,   233 ,   991 ,   1683    ,   1145    ,   6283    ,0 };
            const unsigned long dim660Kuo2Init[]        =   {   1   ,   1   ,   1   ,   7   ,   13  ,   43  ,   95  ,   141 ,   415 ,   219 ,   1877    ,   2763    ,   3975    ,0 };
            const unsigned long dim661Kuo2Init[]        =   {   1   ,   1   ,   3   ,   9   ,   25  ,   1   ,   125 ,   139 ,   31  ,   791 ,   2033    ,   2477    ,   297 ,0 };
            const unsigned long dim662Kuo2Init[]        =   {   1   ,   1   ,   3   ,   7   ,   13  ,   53  ,   29  ,   53  ,   341 ,   905 ,   1953    ,   593 ,   3121    ,0 };
            const unsigned long dim663Kuo2Init[]        =   {   1   ,   3   ,   3   ,   5   ,   21  ,   45  ,   97  ,   231 ,   377 ,   117 ,   1449    ,   1159    ,   1831    ,0 };
            const unsigned long dim664Kuo2Init[]        =   {   1   ,   3   ,   7   ,   3   ,   13  ,   25  ,   33  ,   15  ,   479 ,   141 ,   1101    ,   1611    ,   2145    ,0 };
            const unsigned long dim665Kuo2Init[]        =   {   1   ,   3   ,   5   ,   13  ,   19  ,   33  ,   85  ,   215 ,   107 ,   273 ,   1361    ,   1103    ,   3461    ,0 };
            const unsigned long dim666Kuo2Init[]        =   {   1   ,   3   ,   5   ,   9   ,   7   ,   29  ,   91  ,   55  ,   239 ,   971 ,   1389    ,   301 ,   6011    ,0 };
            const unsigned long dim667Kuo2Init[]        =   {   1   ,   3   ,   3   ,   3   ,   31  ,   39  ,   73  ,   131 ,   53  ,   583 ,   1387    ,   3827    ,   5621    ,0 };
            const unsigned long dim668Kuo2Init[]        =   {   1   ,   1   ,   7   ,   1   ,   21  ,   47  ,   39  ,   99  ,   95  ,   915 ,   1281    ,   675 ,   4147    ,0 };
            const unsigned long dim669Kuo2Init[]        =   {   1   ,   3   ,   7   ,   13  ,   19  ,   17  ,   121 ,   151 ,   163 ,   447 ,   1741    ,   2563    ,   957 ,0 };
            const unsigned long dim670Kuo2Init[]        =   {   1   ,   3   ,   1   ,   11  ,   19  ,   19  ,   9   ,   31  ,   105 ,   829 ,   1957    ,   3327    ,   2923    ,0 };
            const unsigned long dim671Kuo2Init[]        =   {   1   ,   1   ,   1   ,   1   ,   29  ,   49  ,   67  ,   135 ,   165 ,   203 ,   691 ,   4043    ,   4973    ,0 };
            const unsigned long dim672Kuo2Init[]        =   {   1   ,   3   ,   3   ,   15  ,   17  ,   57  ,   53  ,   123 ,   311 ,   169 ,   1705    ,   3853    ,   3243    ,0 };
            const unsigned long dim673Kuo2Init[]        =   {   1   ,   1   ,   7   ,   5   ,   11  ,   51  ,   39  ,   215 ,   209 ,   353 ,   1857    ,   369 ,   1793    ,0 };
            const unsigned long dim674Kuo2Init[]        =   {   1   ,   1   ,   1   ,   3   ,   1   ,   55  ,   19  ,   145 ,   283 ,   471 ,   929 ,   3207    ,   5819    ,0 };
            const unsigned long dim675Kuo2Init[]        =   {   1   ,   1   ,   3   ,   13  ,   21  ,   23  ,   49  ,   241 ,   53  ,   153 ,   591 ,   949 ,   859 ,0 };
            const unsigned long dim676Kuo2Init[]        =   {   1   ,   3   ,   1   ,   5   ,   21  ,   13  ,   105 ,   147 ,   29  ,   407 ,   589 ,   1659    ,   5265    ,0 };
            const unsigned long dim677Kuo2Init[]        =   {   1   ,   1   ,   1   ,   9   ,   19  ,   27  ,   83  ,   175 ,   329 ,   459 ,   1401    ,   2195    ,   5997    ,0 };
            const unsigned long dim678Kuo2Init[]        =   {   1   ,   1   ,   5   ,   1   ,   29  ,   39  ,   33  ,   73  ,   363 ,   143 ,   617 ,   1507    ,   5871    ,0 };
            const unsigned long dim679Kuo2Init[]        =   {   1   ,   1   ,   1   ,   7   ,   29  ,   23  ,   19  ,   51  ,   331 ,   299 ,   1883    ,   1973    ,   1969    ,0 };
            const unsigned long dim680Kuo2Init[]        =   {   1   ,   3   ,   7   ,   13  ,   1   ,   49  ,   25  ,   43  ,   17  ,   297 ,   683 ,   1189    ,   6049    ,0 };
            const unsigned long dim681Kuo2Init[]        =   {   1   ,   3   ,   1   ,   15  ,   23  ,   25  ,   29  ,   231 ,   425 ,   501 ,   1787    ,   3223    ,   5789    ,0 };
            const unsigned long dim682Kuo2Init[]        =   {   1   ,   3   ,   7   ,   1   ,   1   ,   3   ,   23  ,   231 ,   377 ,   485 ,   1469    ,   1089    ,   1539    ,0 };
            const unsigned long dim683Kuo2Init[]        =   {   1   ,   1   ,   5   ,   5   ,   29  ,   25  ,   93  ,   51  ,   249 ,   25  ,   1355    ,   23  ,   2961    ,0 };
            const unsigned long dim684Kuo2Init[]        =   {   1   ,   3   ,   5   ,   13  ,   11  ,   55  ,   117 ,   221 ,   137 ,   941 ,   1127    ,   735 ,   4835    ,0 };
            const unsigned long dim685Kuo2Init[]        =   {   1   ,   3   ,   3   ,   15  ,   5   ,   53  ,   35  ,   245 ,   31  ,   985 ,   2035    ,   163 ,   5223    ,0 };
            const unsigned long dim686Kuo2Init[]        =   {   1   ,   1   ,   7   ,   7   ,   27  ,   25  ,   17  ,   113 ,   187 ,   283 ,   723 ,   2703    ,   4873    ,0 };
            const unsigned long dim687Kuo2Init[]        =   {   1   ,   1   ,   1   ,   1   ,   3   ,   61  ,   81  ,   125 ,   335 ,   997 ,   975 ,   2371    ,   1829    ,0 };
            const unsigned long dim688Kuo2Init[]        =   {   1   ,   3   ,   1   ,   15  ,   21  ,   9   ,   67  ,   55  ,   109 ,   657 ,   837 ,   1737    ,   5147    ,0 };
            const unsigned long dim689Kuo2Init[]        =   {   1   ,   1   ,   3   ,   3   ,   19  ,   41  ,   117 ,   39  ,   441 ,   985 ,   955 ,   1467    ,   7297    ,0 };
            const unsigned long dim690Kuo2Init[]        =   {   1   ,   1   ,   7   ,   7   ,   23  ,   1   ,   11  ,   11  ,   221 ,   539 ,   1807    ,   1887    ,   7921    ,0 };
            const unsigned long dim691Kuo2Init[]        =   {   1   ,   3   ,   3   ,   1   ,   27  ,   45  ,   73  ,   39  ,   217 ,   233 ,   2015    ,   1209    ,   7101    ,0 };
            const unsigned long dim692Kuo2Init[]        =   {   1   ,   3   ,   1   ,   11  ,   19  ,   57  ,   85  ,   45  ,   395 ,   23  ,   405 ,   2723    ,   653 ,0 };
            const unsigned long dim693Kuo2Init[]        =   {   1   ,   1   ,   5   ,   15  ,   3   ,   43  ,   95  ,   149 ,   395 ,   359 ,   1067    ,   3571    ,   6219    ,0 };
            const unsigned long dim694Kuo2Init[]        =   {   1   ,   1   ,   5   ,   5   ,   15  ,   3   ,   33  ,   181 ,   427 ,   565 ,   103 ,   261 ,   2375    ,0 };
            const unsigned long dim695Kuo2Init[]        =   {   1   ,   3   ,   3   ,   15  ,   1   ,   3   ,   89  ,   239 ,   461 ,   675 ,   1149    ,   819 ,   3481    ,0 };
            const unsigned long dim696Kuo2Init[]        =   {   1   ,   1   ,   5   ,   9   ,   23  ,   57  ,   1   ,   147 ,   245 ,   941 ,   1425    ,   2403    ,   7915    ,0 };
            const unsigned long dim697Kuo2Init[]        =   {   1   ,   3   ,   1   ,   13  ,   3   ,   25  ,   43  ,   205 ,   295 ,   799 ,   1845    ,   2695    ,   3183    ,0 };
            const unsigned long dim698Kuo2Init[]        =   {   1   ,   3   ,   1   ,   15  ,   11  ,   19  ,   111 ,   83  ,   479 ,   927 ,   709 ,   2739    ,   1233    ,0 };
            const unsigned long dim699Kuo2Init[]        =   {   1   ,   3   ,   7   ,   13  ,   25  ,   41  ,   49  ,   39  ,   63  ,   265 ,   717 ,   3843    ,   3091    ,0 };
            const unsigned long dim700Kuo2Init[]        =   {   1   ,   3   ,   1   ,   5   ,   27  ,   11  ,   55  ,   31  ,   103 ,   263 ,   759 ,   1969    ,   1301    ,0 };
            const unsigned long dim701Kuo2Init[]        =   {   1   ,   1   ,   7   ,   15  ,   15  ,   33  ,   61  ,   93  ,   177 ,   645 ,   387 ,   3999    ,   4511    ,0 };
            const unsigned long dim702Kuo2Init[]        =   {   1   ,   1   ,   7   ,   9   ,   5   ,   43  ,   87  ,   125 ,   443 ,   825 ,   1423    ,   1575    ,   2661    ,0 };
            const unsigned long dim703Kuo2Init[]        =   {   1   ,   3   ,   1   ,   5   ,   3   ,   59  ,   59  ,   91  ,   63  ,   805 ,   1461    ,   4009    ,   4235    ,0 };
            const unsigned long dim704Kuo2Init[]        =   {   1   ,   1   ,   3   ,   7   ,   15  ,   13  ,   95  ,   179 ,   479 ,   735 ,   1075    ,   4089    ,   4807    ,0 };
            const unsigned long dim705Kuo2Init[]        =   {   1   ,   3   ,   1   ,   3   ,   23  ,   63  ,   63  ,   77  ,   299 ,   939 ,   661 ,   1007    ,   3299    ,0 };
            const unsigned long dim706Kuo2Init[]        =   {   1   ,   1   ,   5   ,   3   ,   7   ,   49  ,   1   ,   217 ,   87  ,   561 ,   275 ,   3137    ,   4355    ,0 };
            const unsigned long dim707Kuo2Init[]        =   {   1   ,   3   ,   5   ,   9   ,   21  ,   25  ,   49  ,   199 ,   207 ,   513 ,   231 ,   3053    ,   8023    ,0 };
            const unsigned long dim708Kuo2Init[]        =   {   1   ,   1   ,   5   ,   1   ,   19  ,   39  ,   79  ,   225 ,   213 ,   719 ,   577 ,   3885    ,   7151    ,0 };
            const unsigned long dim709Kuo2Init[]        =   {   1   ,   3   ,   1   ,   7   ,   9   ,   7   ,   9   ,   63  ,   185 ,   877 ,   253 ,   1401    ,   7665    ,0 };
            const unsigned long dim710Kuo2Init[]        =   {   1   ,   1   ,   7   ,   15  ,   7   ,   9   ,   33  ,   79  ,   241 ,   727 ,   681 ,   2195    ,   5285    ,0 };
            const unsigned long dim711Kuo2Init[]        =   {   1   ,   3   ,   3   ,   9   ,   21  ,   27  ,   9   ,   17  ,   277 ,   365 ,   1497    ,   2323    ,   1215    ,0 };
            const unsigned long dim712Kuo2Init[]        =   {   1   ,   1   ,   3   ,   3   ,   29  ,   23  ,   63  ,   209 ,   487 ,   243 ,   1137    ,   45  ,   3949    ,0 };
            const unsigned long dim713Kuo2Init[]        =   {   1   ,   1   ,   1   ,   15  ,   1   ,   33  ,   29  ,   245 ,   25  ,   497 ,   809 ,   4017    ,   2149    ,0 };
            const unsigned long dim714Kuo2Init[]        =   {   1   ,   3   ,   3   ,   3   ,   25  ,   11  ,   117 ,   129 ,   23  ,   547 ,   175 ,   463 ,   2731    ,0 };
            const unsigned long dim715Kuo2Init[]        =   {   1   ,   3   ,   3   ,   3   ,   5   ,   47  ,   69  ,   197 ,   131 ,   101 ,   1003    ,   1573    ,   3629    ,0 };
            const unsigned long dim716Kuo2Init[]        =   {   1   ,   1   ,   1   ,   9   ,   7   ,   33  ,   63  ,   53  ,   159 ,   1009    ,   963 ,   203 ,   3395    ,0 };
            const unsigned long dim717Kuo2Init[]        =   {   1   ,   3   ,   1   ,   1   ,   9   ,   11  ,   95  ,   73  ,   359 ,   357 ,   421 ,   1169    ,   7317    ,0 };
            const unsigned long dim718Kuo2Init[]        =   {   1   ,   3   ,   5   ,   15  ,   27  ,   57  ,   105 ,   109 ,   457 ,   161 ,   901 ,   2239    ,   2143    ,0 };
            const unsigned long dim719Kuo2Init[]        =   {   1   ,   1   ,   3   ,   7   ,   7   ,   61  ,   11  ,   211 ,   7   ,   515 ,   895 ,   1341    ,   2603    ,0 };
            const unsigned long dim720Kuo2Init[]        =   {   1   ,   3   ,   3   ,   3   ,   7   ,   15  ,   67  ,   237 ,   5   ,   345 ,   1631    ,   923 ,   4269    ,0 };
            const unsigned long dim721Kuo2Init[]        =   {   1   ,   1   ,   1   ,   15  ,   15  ,   27  ,   73  ,   7   ,   359 ,   49  ,   381 ,   2415    ,   1979    ,0 };
            const unsigned long dim722Kuo2Init[]        =   {   1   ,   3   ,   7   ,   7   ,   15  ,   33  ,   1   ,   39  ,   19  ,   171 ,   93  ,   1815    ,   2675    ,0 };
            const unsigned long dim723Kuo2Init[]        =   {   1   ,   3   ,   7   ,   9   ,   13  ,   37  ,   85  ,   75  ,   395 ,   767 ,   1855    ,   3413    ,   1845    ,0 };
            const unsigned long dim724Kuo2Init[]        =   {   1   ,   3   ,   5   ,   15  ,   7   ,   59  ,   33  ,   235 ,   455 ,   21  ,   237 ,   3103    ,   6065    ,0 };
            const unsigned long dim725Kuo2Init[]        =   {   1   ,   1   ,   3   ,   3   ,   29  ,   9   ,   67  ,   173 ,   303 ,   9   ,   1899    ,   3799    ,   63  ,0 };
            const unsigned long dim726Kuo2Init[]        =   {   1   ,   1   ,   3   ,   5   ,   15  ,   53  ,   13  ,   149 ,   45  ,   223 ,   823 ,   1351    ,   4517    ,0 };
            const unsigned long dim727Kuo2Init[]        =   {   1   ,   3   ,   3   ,   15  ,   21  ,   59  ,   81  ,   185 ,   245 ,   589 ,   1149    ,   3351    ,   4879    ,0 };
            const unsigned long dim728Kuo2Init[]        =   {   1   ,   1   ,   1   ,   1   ,   11  ,   29  ,   17  ,   57  ,   199 ,   941 ,   1497    ,   227 ,   6271    ,0 };
            const unsigned long dim729Kuo2Init[]        =   {   1   ,   3   ,   5   ,   7   ,   13  ,   35  ,   97  ,   247 ,   439 ,   687 ,   1821    ,   1583    ,   5919    ,0 };
            const unsigned long dim730Kuo2Init[]        =   {   1   ,   1   ,   3   ,   7   ,   27  ,   25  ,   123 ,   129 ,   15  ,   497 ,   1491    ,   171 ,   6369    ,0 };
            const unsigned long dim731Kuo2Init[]        =   {   1   ,   3   ,   1   ,   1   ,   25  ,   27  ,   63  ,   27  ,   261 ,   707 ,   579 ,   2599    ,   5925    ,0 };
            const unsigned long dim732Kuo2Init[]        =   {   1   ,   1   ,   3   ,   9   ,   9   ,   39  ,   63  ,   141 ,   293 ,   601 ,   573 ,   445 ,   1003    ,0 };
            const unsigned long dim733Kuo2Init[]        =   {   1   ,   3   ,   3   ,   15  ,   7   ,   25  ,   21  ,   189 ,   137 ,   13  ,   1923    ,   1019    ,   7219    ,0 };
            const unsigned long dim734Kuo2Init[]        =   {   1   ,   3   ,   5   ,   9   ,   27  ,   45  ,   13  ,   251 ,   403 ,   781 ,   2009    ,   239 ,   4777    ,0 };
            const unsigned long dim735Kuo2Init[]        =   {   1   ,   1   ,   5   ,   3   ,   21  ,   35  ,   93  ,   89  ,   249 ,   473 ,   1241    ,   2485    ,   291 ,0 };
            const unsigned long dim736Kuo2Init[]        =   {   1   ,   3   ,   7   ,   15  ,   5   ,   53  ,   45  ,   213 ,   41  ,   393 ,   529 ,   599 ,   4993    ,0 };
            const unsigned long dim737Kuo2Init[]        =   {   1   ,   3   ,   7   ,   5   ,   29  ,   29  ,   9   ,   163 ,   27  ,   617 ,   2015    ,   117 ,   6065    ,0 };
            const unsigned long dim738Kuo2Init[]        =   {   1   ,   1   ,   7   ,   3   ,   9   ,   19  ,   31  ,   67  ,   395 ,   1003    ,   1907    ,   1165    ,   3375    ,0 };
            const unsigned long dim739Kuo2Init[]        =   {   1   ,   1   ,   7   ,   1   ,   1   ,   37  ,   41  ,   131 ,   397 ,   181 ,   481 ,   2691    ,   6519    ,0 };
            const unsigned long dim740Kuo2Init[]        =   {   1   ,   3   ,   5   ,   3   ,   9   ,   31  ,   127 ,   145 ,   435 ,   149 ,   597 ,   3611    ,   7715    ,0 };
            const unsigned long dim741Kuo2Init[]        =   {   1   ,   1   ,   3   ,   9   ,   27  ,   33  ,   37  ,   165 ,   167 ,   699 ,   1313    ,   3599    ,   827 ,0 };
            const unsigned long dim742Kuo2Init[]        =   {   1   ,   1   ,   1   ,   9   ,   15  ,   3   ,   21  ,   123 ,   319 ,   1009    ,   901 ,   819 ,   823 ,0 };
            const unsigned long dim743Kuo2Init[]        =   {   1   ,   1   ,   7   ,   9   ,   29  ,   27  ,   101 ,   33  ,   341 ,   449 ,   607 ,   2807    ,   5553    ,0 };
            const unsigned long dim744Kuo2Init[]        =   {   1   ,   3   ,   5   ,   3   ,   21  ,   25  ,   41  ,   105 ,   255 ,   339 ,   1971    ,   175 ,   47  ,0 };
            const unsigned long dim745Kuo2Init[]        =   {   1   ,   1   ,   3   ,   5   ,   5   ,   3   ,   21  ,   73  ,   111 ,   293 ,   1283    ,   2531    ,   7609    ,0 };
            const unsigned long dim746Kuo2Init[]        =   {   1   ,   3   ,   7   ,   3   ,   29  ,   47  ,   83  ,   165 ,   249 ,   281 ,   175 ,   1043    ,   4143    ,0 };
            const unsigned long dim747Kuo2Init[]        =   {   1   ,   3   ,   1   ,   15  ,   27  ,   59  ,   3   ,   101 ,   343 ,   813 ,   1559    ,   553 ,   7801    ,0 };
            const unsigned long dim748Kuo2Init[]        =   {   1   ,   1   ,   7   ,   15  ,   1   ,   9   ,   77  ,   205 ,   151 ,   255 ,   701 ,   2653    ,   2079    ,0 };
            const unsigned long dim749Kuo2Init[]        =   {   1   ,   3   ,   3   ,   3   ,   13  ,   39  ,   37  ,   161 ,   499 ,   657 ,   1557    ,   3969    ,   3489    ,0 };
            const unsigned long dim750Kuo2Init[]        =   {   1   ,   1   ,   1   ,   15  ,   9   ,   27  ,   41  ,   253 ,   119 ,   583 ,   73  ,   3255    ,   6067    ,0 };
            const unsigned long dim751Kuo2Init[]        =   {   1   ,   1   ,   7   ,   7   ,   27  ,   43  ,   53  ,   99  ,   245 ,   169 ,   659 ,   81  ,   2943    ,0 };
            const unsigned long dim752Kuo2Init[]        =   {   1   ,   1   ,   1   ,   11  ,   27  ,   13  ,   41  ,   217 ,   493 ,   145 ,   835 ,   3745    ,   97  ,0 };
            const unsigned long dim753Kuo2Init[]        =   {   1   ,   3   ,   1   ,   9   ,   7   ,   13  ,   117 ,   113 ,   197 ,   503 ,   953 ,   2433    ,   7663    ,0 };
            const unsigned long dim754Kuo2Init[]        =   {   1   ,   3   ,   5   ,   5   ,   9   ,   47  ,   101 ,   47  ,   187 ,   469 ,   69  ,   4031    ,   6231    ,0 };
            const unsigned long dim755Kuo2Init[]        =   {   1   ,   1   ,   1   ,   3   ,   27  ,   27  ,   47  ,   219 ,   11  ,   201 ,   1267    ,   2703    ,   6711    ,0 };
            const unsigned long dim756Kuo2Init[]        =   {   1   ,   3   ,   7   ,   7   ,   27  ,   27  ,   97  ,   101 ,   7   ,   541 ,   105 ,   3821    ,   7533    ,0 };
            const unsigned long dim757Kuo2Init[]        =   {   1   ,   3   ,   5   ,   13  ,   21  ,   17  ,   99  ,   213 ,   347 ,   595 ,   519 ,   1113    ,   5427    ,0 };
            const unsigned long dim758Kuo2Init[]        =   {   1   ,   1   ,   7   ,   13  ,   31  ,   1   ,   25  ,   111 ,   387 ,   27  ,   99  ,   119 ,   3265    ,0 };
            const unsigned long dim759Kuo2Init[]        =   {   1   ,   1   ,   7   ,   15  ,   11  ,   43  ,   19  ,   177 ,   11  ,   943 ,   3   ,   623 ,   4223    ,0 };
            const unsigned long dim760Kuo2Init[]        =   {   1   ,   3   ,   7   ,   3   ,   31  ,   59  ,   21  ,   241 ,   375 ,   703 ,   1465    ,   2375    ,   2853    ,0 };
            const unsigned long dim761Kuo2Init[]        =   {   1   ,   1   ,   7   ,   15  ,   1   ,   49  ,   51  ,   161 ,   397 ,   583 ,   913 ,   985 ,   5519    ,0 };
            const unsigned long dim762Kuo2Init[]        =   {   1   ,   3   ,   1   ,   3   ,   13  ,   37  ,   23  ,   165 ,   223 ,   427 ,   129 ,   1553    ,   4145    ,0 };
            const unsigned long dim763Kuo2Init[]        =   {   1   ,   3   ,   3   ,   3   ,   1   ,   23  ,   47  ,   237 ,   371 ,   315 ,   1631    ,   2037    ,   61  ,0 };
            const unsigned long dim764Kuo2Init[]        =   {   1   ,   3   ,   5   ,   15  ,   23  ,   59  ,   71  ,   237 ,   487 ,   443 ,   1475    ,   3065    ,   2453    ,0 };
            const unsigned long dim765Kuo2Init[]        =   {   1   ,   1   ,   3   ,   13  ,   25  ,   57  ,   71  ,   3   ,   247 ,   145 ,   2011    ,   553 ,   7565    ,0 };
            const unsigned long dim766Kuo2Init[]        =   {   1   ,   1   ,   1   ,   9   ,   21  ,   1   ,   61  ,   91  ,   3   ,   807 ,   1795    ,   2665    ,   473 ,0 };
            const unsigned long dim767Kuo2Init[]        =   {   1   ,   3   ,   7   ,   11  ,   9   ,   7   ,   123 ,   233 ,   249 ,   825 ,   1505    ,   3599    ,   6653    ,0 };
            const unsigned long dim768Kuo2Init[]        =   {   1   ,   1   ,   3   ,   15  ,   11  ,   37  ,   119 ,   63  ,   233 ,   541 ,   1373    ,   765 ,   4913    ,0 };
            const unsigned long dim769Kuo2Init[]        =   {   1   ,   1   ,   1   ,   9   ,   5   ,   15  ,   77  ,   65  ,   321 ,   613 ,   1857    ,   3699    ,   1   ,0 };
            const unsigned long dim770Kuo2Init[]        =   {   1   ,   3   ,   1   ,   15  ,   13  ,   37  ,   123 ,   9   ,   105 ,   121 ,   893 ,   1009    ,   7871    ,0 };
            const unsigned long dim771Kuo2Init[]        =   {   1   ,   3   ,   5   ,   1   ,   23  ,   61  ,   29  ,   223 ,   209 ,   605 ,   409 ,   875 ,   5233    ,0 };
            const unsigned long dim772Kuo2Init[]        =   {   1   ,   1   ,   3   ,   11  ,   5   ,   61  ,   9   ,   85  ,   361 ,   613 ,   1555    ,   2323    ,   5119    ,0 };
            const unsigned long dim773Kuo2Init[]        =   {   1   ,   1   ,   1   ,   1   ,   17  ,   49  ,   67  ,   191 ,   247 ,   873 ,   1283    ,   171 ,   1689    ,0 };
            const unsigned long dim774Kuo2Init[]        =   {   1   ,   3   ,   3   ,   13  ,   17  ,   17  ,   45  ,   73  ,   175 ,   243 ,   97  ,   315 ,   1429    ,0 };
            const unsigned long dim775Kuo2Init[]        =   {   1   ,   3   ,   5   ,   5   ,   13  ,   1   ,   113 ,   87  ,   241 ,   199 ,   145 ,   3823    ,   5965    ,0 };
            const unsigned long dim776Kuo2Init[]        =   {   1   ,   3   ,   3   ,   7   ,   7   ,   47  ,   115 ,   139 ,   239 ,   547 ,   109 ,   2251    ,   6987    ,0 };
            const unsigned long dim777Kuo2Init[]        =   {   1   ,   3   ,   1   ,   11  ,   3   ,   59  ,   111 ,   225 ,   385 ,   887 ,   425 ,   1879    ,   4535    ,0 };
            const unsigned long dim778Kuo2Init[]        =   {   1   ,   1   ,   3   ,   9   ,   27  ,   35  ,   55  ,   125 ,   379 ,   529 ,   191 ,   1657    ,   2835    ,0 };
            const unsigned long dim779Kuo2Init[]        =   {   1   ,   1   ,   7   ,   11  ,   25  ,   11  ,   85  ,   21  ,   65  ,   377 ,   1989    ,   1963    ,   6985    ,0 };
            const unsigned long dim780Kuo2Init[]        =   {   1   ,   3   ,   7   ,   11  ,   15  ,   59  ,   91  ,   51  ,   145 ,   627 ,   1617    ,   3029    ,   3621    ,0 };
            const unsigned long dim781Kuo2Init[]        =   {   1   ,   3   ,   7   ,   11  ,   29  ,   9   ,   15  ,   189 ,   495 ,   997 ,   467 ,   343 ,   7615    ,0 };
            const unsigned long dim782Kuo2Init[]        =   {   1   ,   1   ,   7   ,   7   ,   17  ,   37  ,   89  ,   31  ,   355 ,   323 ,   1361    ,   3785    ,   643 ,0 };
            const unsigned long dim783Kuo2Init[]        =   {   1   ,   1   ,   7   ,   1   ,   27  ,   19  ,   15  ,   5   ,   383 ,   1023    ,   923 ,   3283    ,   2997    ,0 };
            const unsigned long dim784Kuo2Init[]        =   {   1   ,   3   ,   7   ,   15  ,   11  ,   41  ,   117 ,   255 ,   407 ,   117 ,   495 ,   99  ,   7687    ,0 };
            const unsigned long dim785Kuo2Init[]        =   {   1   ,   1   ,   1   ,   9   ,   9   ,   29  ,   109 ,   103 ,   81  ,   237 ,   1547    ,   3083    ,   2449    ,0 };
            const unsigned long dim786Kuo2Init[]        =   {   1   ,   3   ,   5   ,   1   ,   19  ,   57  ,   43  ,   215 ,   251 ,   417 ,   1719    ,   3767    ,   1543    ,0 };
            const unsigned long dim787Kuo2Init[]        =   {   1   ,   3   ,   7   ,   11  ,   15  ,   9   ,   1   ,   171 ,   439 ,   637 ,   1709    ,   1483    ,   4111    ,0 };
            const unsigned long dim788Kuo2Init[]        =   {   1   ,   3   ,   1   ,   5   ,   25  ,   31  ,   101 ,   101 ,   301 ,   169 ,   803 ,   2119    ,   5035    ,0 };
            const unsigned long dim789Kuo2Init[]        =   {   1   ,   1   ,   7   ,   5   ,   21  ,   57  ,   59  ,   149 ,   31  ,   1023    ,   1063    ,   2465    ,   1517    ,0 };
            const unsigned long dim790Kuo2Init[]        =   {   1   ,   1   ,   5   ,   5   ,   7   ,   29  ,   101 ,   19  ,   49  ,   187 ,   715 ,   843 ,   1553    ,0 };
            const unsigned long dim791Kuo2Init[]        =   {   1   ,   3   ,   3   ,   5   ,   21  ,   43  ,   27  ,   203 ,   125 ,   629 ,   1395    ,   15  ,   7647    ,0 };
            const unsigned long dim792Kuo2Init[]        =   {   1   ,   1   ,   3   ,   11  ,   27  ,   49  ,   69  ,   135 ,   447 ,   503 ,   1589    ,   209 ,   791 ,0 };
            const unsigned long dim793Kuo2Init[]        =   {   1   ,   3   ,   1   ,   15  ,   17  ,   57  ,   27  ,   55  ,   237 ,   223 ,   949 ,   2119    ,   3221    ,0 };
            const unsigned long dim794Kuo2Init[]        =   {   1   ,   3   ,   3   ,   9   ,   3   ,   37  ,   43  ,   49  ,   41  ,   899 ,   1657    ,   2463    ,   1303    ,0 };
            const unsigned long dim795Kuo2Init[]        =   {   1   ,   1   ,   7   ,   13  ,   27  ,   27  ,   19  ,   93  ,   139 ,   745 ,   889 ,   951 ,   6925    ,0 };
            const unsigned long dim796Kuo2Init[]        =   {   1   ,   3   ,   1   ,   13  ,   15  ,   33  ,   97  ,   217 ,   333 ,   295 ,   683 ,   397 ,   3889    ,0 };
            const unsigned long dim797Kuo2Init[]        =   {   1   ,   3   ,   5   ,   15  ,   17  ,   57  ,   47  ,   173 ,   15  ,   77  ,   863 ,   1441    ,   6851    ,0 };
            const unsigned long dim798Kuo2Init[]        =   {   1   ,   3   ,   5   ,   7   ,   3   ,   43  ,   21  ,   37  ,   349 ,   1021    ,   725 ,   1985    ,   481 ,0 };
            const unsigned long dim799Kuo2Init[]        =   {   1   ,   1   ,   1   ,   13  ,   9   ,   45  ,   85  ,   177 ,   15  ,   915 ,   187 ,   15  ,   6621    ,0 };
            const unsigned long dim800Kuo2Init[]        =   {   1   ,   3   ,   5   ,   11  ,   5   ,   29  ,   113 ,   95  ,   461 ,   607 ,   1637    ,   1897    ,   721 ,0 };
            const unsigned long dim801Kuo2Init[]        =   {   1   ,   3   ,   1   ,   11  ,   13  ,   23  ,   55  ,   169 ,   29  ,   803 ,   1433    ,   2113    ,   7255    ,0 };
            const unsigned long dim802Kuo2Init[]        =   {   1   ,   1   ,   5   ,   3   ,   31  ,   1   ,   45  ,   251 ,   129 ,   267 ,   1489    ,   427 ,   4993    ,0 };
            const unsigned long dim803Kuo2Init[]        =   {   1   ,   1   ,   1   ,   9   ,   9   ,   11  ,   19  ,   31  ,   493 ,   195 ,   1545    ,   2681    ,   2187    ,0 };
            const unsigned long dim804Kuo2Init[]        =   {   1   ,   3   ,   1   ,   7   ,   3   ,   51  ,   121 ,   149 ,   325 ,   921 ,   241 ,   999 ,   6039    ,0 };
            const unsigned long dim805Kuo2Init[]        =   {   1   ,   3   ,   3   ,   15  ,   17  ,   37  ,   41  ,   157 ,   209 ,   587 ,   1351    ,   2353    ,   1197    ,0 };
            const unsigned long dim806Kuo2Init[]        =   {   1   ,   1   ,   5   ,   13  ,   13  ,   55  ,   45  ,   3   ,   475 ,   9   ,   11  ,   3175    ,   2185    ,0 };
            const unsigned long dim807Kuo2Init[]        =   {   1   ,   3   ,   5   ,   13  ,   31  ,   33  ,   63  ,   149 ,   463 ,   493 ,   1483    ,   1457    ,   6051    ,0 };
            const unsigned long dim808Kuo2Init[]        =   {   1   ,   1   ,   1   ,   3   ,   3   ,   11  ,   119 ,   177 ,   207 ,   631 ,   105 ,   263 ,   2533    ,0 };
            const unsigned long dim809Kuo2Init[]        =   {   1   ,   1   ,   1   ,   15  ,   23  ,   27  ,   73  ,   77  ,   143 ,   523 ,   1883    ,   3289    ,   1685    ,0 };
            const unsigned long dim810Kuo2Init[]        =   {   1   ,   3   ,   1   ,   5   ,   21  ,   31  ,   25  ,   183 ,   71  ,   679 ,   1697    ,   3473    ,   4233    ,0 };
            const unsigned long dim811Kuo2Init[]        =   {   1   ,   1   ,   7   ,   15  ,   9   ,   13  ,   103 ,   125 ,   223 ,   83  ,   1255    ,   2755    ,   1447    ,0 };
            const unsigned long dim812Kuo2Init[]        =   {   1   ,   3   ,   5   ,   15  ,   15  ,   51  ,   69  ,   101 ,   369 ,   463 ,   1515    ,   3485    ,   1871    ,0 };
            const unsigned long dim813Kuo2Init[]        =   {   1   ,   3   ,   5   ,   7   ,   21  ,   31  ,   55  ,   223 ,   477 ,   995 ,   147 ,   387 ,   7917    ,0 };
            const unsigned long dim814Kuo2Init[]        =   {   1   ,   3   ,   5   ,   5   ,   7   ,   31  ,   73  ,   243 ,   283 ,   949 ,   705 ,   449 ,   3309    ,0 };
            const unsigned long dim815Kuo2Init[]        =   {   1   ,   1   ,   5   ,   9   ,   15  ,   5   ,   25  ,   219 ,   379 ,   1013    ,   1683    ,   703 ,   4855    ,0 };
            const unsigned long dim816Kuo2Init[]        =   {   1   ,   1   ,   7   ,   15  ,   25  ,   15  ,   99  ,   217 ,   265 ,   851 ,   149 ,   1557    ,   8107    ,0 };
            const unsigned long dim817Kuo2Init[]        =   {   1   ,   3   ,   3   ,   5   ,   1   ,   59  ,   35  ,   119 ,   245 ,   9   ,   765 ,   833 ,   1707    ,0 };
            const unsigned long dim818Kuo2Init[]        =   {   1   ,   1   ,   1   ,   13  ,   23  ,   49  ,   25  ,   153 ,   215 ,   365 ,   1943    ,   2975    ,   1301    ,0 };
            const unsigned long dim819Kuo2Init[]        =   {   1   ,   3   ,   5   ,   7   ,   15  ,   45  ,   25  ,   215 ,   225 ,   431 ,   1161    ,   2389    ,   1969    ,0 };
            const unsigned long dim820Kuo2Init[]        =   {   1   ,   3   ,   5   ,   9   ,   19  ,   31  ,   121 ,   189 ,   331 ,   23  ,   419 ,   3171    ,   6241    ,0 };
            const unsigned long dim821Kuo2Init[]        =   {   1   ,   1   ,   5   ,   5   ,   17  ,   39  ,   69  ,   215 ,   307 ,   827 ,   557 ,   3251    ,   3211    ,0 };
            const unsigned long dim822Kuo2Init[]        =   {   1   ,   1   ,   3   ,   11  ,   31  ,   27  ,   23  ,   119 ,   463 ,   605 ,   453 ,   1573    ,   1981    ,0 };
            const unsigned long dim823Kuo2Init[]        =   {   1   ,   1   ,   5   ,   11  ,   1   ,   15  ,   37  ,   111 ,   481 ,   331 ,   1481    ,   1013    ,   3259    ,0 };
            const unsigned long dim824Kuo2Init[]        =   {   1   ,   1   ,   1   ,   11  ,   21  ,   39  ,   25  ,   183 ,   351 ,   553 ,   3   ,   3969    ,   2551    ,0 };
            const unsigned long dim825Kuo2Init[]        =   {   1   ,   3   ,   7   ,   1   ,   19  ,   5   ,   3   ,   55  ,   483 ,   373 ,   1529    ,   535 ,   4215    ,0 };
            const unsigned long dim826Kuo2Init[]        =   {   1   ,   1   ,   3   ,   1   ,   25  ,   61  ,   89  ,   197 ,   465 ,   493 ,   661 ,   2361    ,   4211    ,0 };
            const unsigned long dim827Kuo2Init[]        =   {   1   ,   3   ,   1   ,   9   ,   29  ,   61  ,   59  ,   175 ,   389 ,   973 ,   1763    ,   4077    ,   6211    ,0 };
            const unsigned long dim828Kuo2Init[]        =   {   1   ,   3   ,   1   ,   7   ,   11  ,   41  ,   117 ,   75  ,   291 ,   125 ,   273 ,   5   ,   8169    ,0 };
            const unsigned long dim829Kuo2Init[]        =   {   1   ,   1   ,   5   ,   11  ,   21  ,   53  ,   121 ,   31  ,   451 ,   631 ,   1901    ,   4059    ,   6801    ,0 };
            const unsigned long dim830Kuo2Init[]        =   {   1   ,   3   ,   7   ,   3   ,   13  ,   33  ,   51  ,   247 ,   409 ,   881 ,   935 ,   1927    ,   1559    ,0 };
            const unsigned long dim831Kuo2Init[]        =   {   1   ,   1   ,   3   ,   13  ,   15  ,   51  ,   105 ,   209 ,   159 ,   333 ,   325 ,   3699    ,   5461    ,0 };
            const unsigned long dim832Kuo2Init[]        =   {   1   ,   3   ,   3   ,   5   ,   17  ,   7   ,   47  ,   57  ,   59  ,   673 ,   611 ,   253 ,   1531    ,0 };
            const unsigned long dim833Kuo2Init[]        =   {   1   ,   1   ,   7   ,   1   ,   19  ,   27  ,   15  ,   189 ,   319 ,   649 ,   1593    ,   875 ,   8187    ,0 };
            const unsigned long dim834Kuo2Init[]        =   {   1   ,   1   ,   1   ,   3   ,   15  ,   17  ,   81  ,   183 ,   409 ,   433 ,   679 ,   1881    ,   7827    ,0 };
            const unsigned long dim835Kuo2Init[]        =   {   1   ,   1   ,   3   ,   9   ,   23  ,   37  ,   83  ,   235 ,   225 ,   599 ,   1979    ,   2871    ,   1353    ,0 };
            const unsigned long dim836Kuo2Init[]        =   {   1   ,   1   ,   7   ,   7   ,   27  ,   11  ,   93  ,   115 ,   339 ,   233 ,   1733    ,   2071    ,   3187    ,0 };
            const unsigned long dim837Kuo2Init[]        =   {   1   ,   1   ,   1   ,   7   ,   27  ,   47  ,   61  ,   73  ,   361 ,   621 ,   1907    ,   3103    ,   1873    ,0 };
            const unsigned long dim838Kuo2Init[]        =   {   1   ,   1   ,   1   ,   9   ,   9   ,   13  ,   67  ,   161 ,   21  ,   953 ,   1101    ,   2255    ,   4841    ,0 };
            const unsigned long dim839Kuo2Init[]        =   {   1   ,   3   ,   7   ,   7   ,   29  ,   39  ,   31  ,   73  ,   351 ,   401 ,   441 ,   4085    ,   2829    ,0 };
            const unsigned long dim840Kuo2Init[]        =   {   1   ,   1   ,   5   ,   15  ,   9   ,   23  ,   127 ,   111 ,   309 ,   955 ,   623 ,   1995    ,   1961    ,0 };
            const unsigned long dim841Kuo2Init[]        =   {   1   ,   3   ,   5   ,   9   ,   3   ,   3   ,   27  ,   91  ,   287 ,   73  ,   1449    ,   1347    ,   8069    ,0 };
            const unsigned long dim842Kuo2Init[]        =   {   1   ,   1   ,   5   ,   3   ,   9   ,   49  ,   87  ,   3   ,   59  ,   193 ,   1999    ,   2295    ,   2247    ,0 };
            const unsigned long dim843Kuo2Init[]        =   {   1   ,   3   ,   3   ,   3   ,   21  ,   21  ,   97  ,   47  ,   195 ,   65  ,   1281    ,   1411    ,   6503    ,0 };
            const unsigned long dim844Kuo2Init[]        =   {   1   ,   1   ,   7   ,   13  ,   1   ,   11  ,   85  ,   195 ,   429 ,   71  ,   1053    ,   3623    ,   2921    ,0 };
            const unsigned long dim845Kuo2Init[]        =   {   1   ,   3   ,   1   ,   5   ,   17  ,   15  ,   105 ,   155 ,   291 ,   889 ,   1997    ,   3629    ,   7025    ,0 };
            const unsigned long dim846Kuo2Init[]        =   {   1   ,   1   ,   3   ,   7   ,   9   ,   3   ,   79  ,   41  ,   21  ,   819 ,   1665    ,   4021    ,   7057    ,0 };
            const unsigned long dim847Kuo2Init[]        =   {   1   ,   1   ,   3   ,   1   ,   1   ,   23  ,   73  ,   5   ,   75  ,   435 ,   889 ,   2577    ,   697 ,0 };
            const unsigned long dim848Kuo2Init[]        =   {   1   ,   3   ,   7   ,   1   ,   1   ,   5   ,   77  ,   183 ,   109 ,   159 ,   1463    ,   2689    ,   7799    ,0 };
            const unsigned long dim849Kuo2Init[]        =   {   1   ,   1   ,   1   ,   13  ,   29  ,   55  ,   123 ,   247 ,   509 ,   909 ,   247 ,   3053    ,   7301    ,0 };
            const unsigned long dim850Kuo2Init[]        =   {   1   ,   1   ,   5   ,   13  ,   7   ,   23  ,   43  ,   177 ,   59  ,   643 ,   1517    ,   1869    ,   3269    ,0 };
            const unsigned long dim851Kuo2Init[]        =   {   1   ,   3   ,   1   ,   9   ,   7   ,   43  ,   3   ,   13  ,   387 ,   893 ,   1013    ,   7   ,   7481    ,0 };
            const unsigned long dim852Kuo2Init[]        =   {   1   ,   1   ,   1   ,   1   ,   9   ,   35  ,   111 ,   199 ,   77  ,   375 ,   981 ,   3475    ,   2339    ,0 };
            const unsigned long dim853Kuo2Init[]        =   {   1   ,   3   ,   5   ,   7   ,   5   ,   13  ,   27  ,   5   ,   113 ,   65  ,   1785    ,   1517    ,   8027    ,0 };
            const unsigned long dim854Kuo2Init[]        =   {   1   ,   3   ,   7   ,   5   ,   25  ,   47  ,   15  ,   7   ,   45  ,   441 ,   7   ,   1601    ,   6543    ,0 };
            const unsigned long dim855Kuo2Init[]        =   {   1   ,   1   ,   7   ,   3   ,   17  ,   59  ,   71  ,   93  ,   441 ,   763 ,   769 ,   1977    ,   287 ,0 };
            const unsigned long dim856Kuo2Init[]        =   {   1   ,   1   ,   7   ,   11  ,   29  ,   23  ,   61  ,   199 ,   469 ,   757 ,   1473    ,   2625    ,   5909    ,0 };
            const unsigned long dim857Kuo2Init[]        =   {   1   ,   1   ,   5   ,   11  ,   27  ,   47  ,   41  ,   91  ,   457 ,   1003    ,   217 ,   401 ,   8113    ,0 };
            const unsigned long dim858Kuo2Init[]        =   {   1   ,   1   ,   5   ,   7   ,   1   ,   33  ,   43  ,   45  ,   169 ,   485 ,   191 ,   615 ,   5141    ,0 };
            const unsigned long dim859Kuo2Init[]        =   {   1   ,   3   ,   1   ,   5   ,   5   ,   53  ,   19  ,   185 ,   189 ,   681 ,   1685    ,   661 ,   6825    ,0 };
            const unsigned long dim860Kuo2Init[]        =   {   1   ,   3   ,   3   ,   1   ,   1   ,   7   ,   47  ,   27  ,   275 ,   63  ,   1029    ,   449 ,   4665    ,0 };
            const unsigned long dim861Kuo2Init[]        =   {   1   ,   3   ,   3   ,   1   ,   3   ,   31  ,   1   ,   77  ,   275 ,   871 ,   1031    ,   1025    ,   4683    ,0 };
            const unsigned long dim862Kuo2Init[]        =   {   1   ,   1   ,   3   ,   3   ,   9   ,   35  ,   39  ,   43  ,   363 ,   129 ,   431 ,   1351    ,   8037    ,0 };
            const unsigned long dim863Kuo2Init[]        =   {   1   ,   1   ,   5   ,   3   ,   3   ,   39  ,   29  ,   99  ,   441 ,   819 ,   175 ,   3503    ,   4911    ,0 };
            const unsigned long dim864Kuo2Init[]        =   {   1   ,   3   ,   5   ,   13  ,   9   ,   17  ,   87  ,   5   ,   311 ,   307 ,   659 ,   2795    ,   4055    ,0 };
            const unsigned long dim865Kuo2Init[]        =   {   1   ,   1   ,   3   ,   13  ,   13  ,   51  ,   69  ,   99  ,   65  ,   257 ,   87  ,   1407    ,   3157    ,0 };
            const unsigned long dim866Kuo2Init[]        =   {   1   ,   1   ,   5   ,   9   ,   19  ,   41  ,   97  ,   231 ,   85  ,   743 ,   637 ,   755 ,   6209    ,0 };
            const unsigned long dim867Kuo2Init[]        =   {   1   ,   1   ,   5   ,   15  ,   21  ,   63  ,   81  ,   35  ,   477 ,   39  ,   873 ,   3679    ,   6477    ,0 };
            const unsigned long dim868Kuo2Init[]        =   {   1   ,   1   ,   7   ,   13  ,   23  ,   37  ,   115 ,   87  ,   171 ,   33  ,   1779    ,   3169    ,   3205    ,0 };
            const unsigned long dim869Kuo2Init[]        =   {   1   ,   3   ,   7   ,   11  ,   9   ,   41  ,   111 ,   9   ,   5   ,   441 ,   1677    ,   311 ,   1019    ,0 };
            const unsigned long dim870Kuo2Init[]        =   {   1   ,   1   ,   5   ,   15  ,   7   ,   23  ,   7   ,   181 ,   461 ,   357 ,   187 ,   3461    ,   6689    ,0 };
            const unsigned long dim871Kuo2Init[]        =   {   1   ,   1   ,   3   ,   11  ,   15  ,   37  ,   101 ,   125 ,   271 ,   205 ,   57  ,   3361    ,   3965    ,0 };
            const unsigned long dim872Kuo2Init[]        =   {   1   ,   3   ,   3   ,   9   ,   25  ,   41  ,   97  ,   133 ,   39  ,   383 ,   669 ,   4007    ,   25  ,0 };
            const unsigned long dim873Kuo2Init[]        =   {   1   ,   3   ,   7   ,   13  ,   17  ,   13  ,   79  ,   61  ,   145 ,   567 ,   1079    ,   287 ,   467 ,0 };
            const unsigned long dim874Kuo2Init[]        =   {   1   ,   1   ,   5   ,   1   ,   7   ,   47  ,   97  ,   131 ,   119 ,   1013    ,   1991    ,   741 ,   27  ,0 };
            const unsigned long dim875Kuo2Init[]        =   {   1   ,   3   ,   5   ,   9   ,   23  ,   21  ,   71  ,   209 ,   239 ,   801 ,   133 ,   2891    ,   1141    ,0 };
            const unsigned long dim876Kuo2Init[]        =   {   1   ,   1   ,   7   ,   9   ,   17  ,   11  ,   109 ,   251 ,   127 ,   447 ,   1257    ,   3831    ,   1511    ,0 };
            const unsigned long dim877Kuo2Init[]        =   {   1   ,   3   ,   1   ,   13  ,   21  ,   9   ,   69  ,   241 ,   61  ,   521 ,   1281    ,   3499    ,   4001    ,0 };
            const unsigned long dim878Kuo2Init[]        =   {   1   ,   1   ,   5   ,   15  ,   5   ,   23  ,   117 ,   111 ,   159 ,   861 ,   295 ,   3403    ,   981 ,0 };
            const unsigned long dim879Kuo2Init[]        =   {   1   ,   3   ,   3   ,   5   ,   3   ,   37  ,   85  ,   129 ,   281 ,   261 ,   1753    ,   1399    ,   43  ,0 };
            const unsigned long dim880Kuo2Init[]        =   {   1   ,   1   ,   5   ,   5   ,   3   ,   49  ,   67  ,   221 ,   3   ,   143 ,   617 ,   2579    ,   373 ,0 };
            const unsigned long dim881Kuo2Init[]        =   {   1   ,   3   ,   5   ,   11  ,   29  ,   49  ,   59  ,   47  ,   21  ,   725 ,   655 ,   3467    ,   2375    ,0 };
            const unsigned long dim882Kuo2Init[]        =   {   1   ,   3   ,   3   ,   13  ,   15  ,   43  ,   19  ,   185 ,   411 ,   299 ,   1351    ,   3543    ,   5429    ,0 };
            const unsigned long dim883Kuo2Init[]        =   {   1   ,   1   ,   5   ,   11  ,   19  ,   63  ,   19  ,   167 ,   439 ,   627 ,   1729    ,   245 ,   193 ,0 };
            const unsigned long dim884Kuo2Init[]        =   {   1   ,   1   ,   3   ,   1   ,   5   ,   5   ,   115 ,   127 ,   295 ,   321 ,   1167    ,   3187    ,   6165    ,0 };
            const unsigned long dim885Kuo2Init[]        =   {   1   ,   3   ,   1   ,   9   ,   9   ,   51  ,   87  ,   205 ,   211 ,   427 ,   1817    ,   3169    ,   1469    ,0 };
            const unsigned long dim886Kuo2Init[]        =   {   1   ,   3   ,   1   ,   15  ,   27  ,   57  ,   15  ,   199 ,   87  ,   361 ,   583 ,   2467    ,   1367    ,0 };
            const unsigned long dim887Kuo2Init[]        =   {   1   ,   1   ,   1   ,   11  ,   27  ,   7   ,   53  ,   173 ,   505 ,   839 ,   929 ,   3977    ,   7667    ,0 };
            const unsigned long dim888Kuo2Init[]        =   {   1   ,   3   ,   3   ,   13  ,   7   ,   37  ,   51  ,   109 ,   249 ,   233 ,   899 ,   3991    ,   6667    ,0 };
            const unsigned long dim889Kuo2Init[]        =   {   1   ,   3   ,   1   ,   11  ,   17  ,   23  ,   105 ,   213 ,   425 ,   175 ,   869 ,   3185    ,   6301    ,0 };
            const unsigned long dim890Kuo2Init[]        =   {   1   ,   1   ,   5   ,   5   ,   17  ,   47  ,   5   ,   135 ,   497 ,   227 ,   1135    ,   689 ,   5047    ,0 };
            const unsigned long dim891Kuo2Init[]        =   {   1   ,   3   ,   3   ,   11  ,   5   ,   49  ,   29  ,   177 ,   325 ,   727 ,   1719    ,   1805    ,   3361    ,0 };
            const unsigned long dim892Kuo2Init[]        =   {   1   ,   1   ,   3   ,   3   ,   11  ,   7   ,   87  ,   169 ,   27  ,   137 ,   1835    ,   1377    ,   4915    ,0 };
            const unsigned long dim893Kuo2Init[]        =   {   1   ,   3   ,   5   ,   5   ,   21  ,   53  ,   99  ,   137 ,   133 ,   599 ,   479 ,   2859    ,   6731    ,0 };
            const unsigned long dim894Kuo2Init[]        =   {   1   ,   1   ,   5   ,   5   ,   15  ,   15  ,   31  ,   239 ,   183 ,   167 ,   1799    ,   1533    ,   309 ,0 };
            const unsigned long dim895Kuo2Init[]        =   {   1   ,   3   ,   5   ,   15  ,   23  ,   51  ,   85  ,   133 ,   51  ,   991 ,   899 ,   3777    ,   2231    ,0 };
            const unsigned long dim896Kuo2Init[]        =   {   1   ,   1   ,   3   ,   1   ,   15  ,   19  ,   29  ,   5   ,   415 ,   531 ,   253 ,   49  ,   7197    ,0 };
            const unsigned long dim897Kuo2Init[]        =   {   1   ,   1   ,   5   ,   15  ,   21  ,   23  ,   41  ,   213 ,   189 ,   579 ,   1181    ,   3425    ,   8037    ,0 };
            const unsigned long dim898Kuo2Init[]        =   {   1   ,   1   ,   5   ,   13  ,   9   ,   63  ,   21  ,   195 ,   71  ,   137 ,   1283    ,   2895    ,   3301    ,0 };
            const unsigned long dim899Kuo2Init[]        =   {   1   ,   1   ,   7   ,   1   ,   11  ,   5   ,   103 ,   183 ,   403 ,   177 ,   1561    ,   201 ,   2843    ,0 };
            const unsigned long dim900Kuo2Init[]        =   {   1   ,   3   ,   3   ,   9   ,   19  ,   15  ,   85  ,   123 ,   273 ,   891 ,   493 ,   3761    ,   557 ,0 };
            const unsigned long dim901Kuo2Init[]        =   {   1   ,   3   ,   7   ,   9   ,   9   ,   23  ,   123 ,   243 ,   71  ,   137 ,   1937    ,   3199    ,   6865    ,0 };
            const unsigned long dim902Kuo2Init[]        =   {   1   ,   1   ,   1   ,   15  ,   15  ,   55  ,   119 ,   203 ,   449 ,   969 ,   181 ,   2573    ,   3261    ,0 };
            const unsigned long dim903Kuo2Init[]        =   {   1   ,   1   ,   1   ,   9   ,   29  ,   9   ,   87  ,   117 ,   85  ,   369 ,   911 ,   1203    ,   7287    ,0 };
            const unsigned long dim904Kuo2Init[]        =   {   1   ,   1   ,   1   ,   7   ,   1   ,   35  ,   9   ,   85  ,   253 ,   123 ,   67  ,   2089    ,   4693    ,0 };
            const unsigned long dim905Kuo2Init[]        =   {   1   ,   3   ,   5   ,   15  ,   11  ,   13  ,   43  ,   97  ,   407 ,   1011    ,   1839    ,   525 ,   6809    ,0 };
            const unsigned long dim906Kuo2Init[]        =   {   1   ,   3   ,   3   ,   7   ,   11  ,   49  ,   5   ,   29  ,   299 ,   35  ,   1883    ,   763 ,   3099    ,0 };
            const unsigned long dim907Kuo2Init[]        =   {   1   ,   3   ,   3   ,   13  ,   31  ,   5   ,   61  ,   79  ,   409 ,   1001    ,   1057    ,   1673    ,   4689    ,0 };
            const unsigned long dim908Kuo2Init[]        =   {   1   ,   1   ,   3   ,   7   ,   23  ,   53  ,   23  ,   243 ,   335 ,   847 ,   1011    ,   659 ,   3739    ,0 };
            const unsigned long dim909Kuo2Init[]        =   {   1   ,   3   ,   5   ,   1   ,   15  ,   55  ,   15  ,   161 ,   417 ,   225 ,   227 ,   3933    ,   5625    ,0 };
            const unsigned long dim910Kuo2Init[]        =   {   1   ,   3   ,   7   ,   11  ,   25  ,   1   ,   75  ,   33  ,   401 ,   777 ,   1999    ,   1649    ,   1269    ,0 };
            const unsigned long dim911Kuo2Init[]        =   {   1   ,   3   ,   5   ,   5   ,   3   ,   1   ,   13  ,   99  ,   447 ,   579 ,   225 ,   3851    ,   5379    ,0 };
            const unsigned long dim912Kuo2Init[]        =   {   1   ,   1   ,   3   ,   13  ,   19  ,   33  ,   39  ,   93  ,   133 ,   889 ,   113 ,   3983    ,   1115    ,0 };
            const unsigned long dim913Kuo2Init[]        =   {   1   ,   3   ,   5   ,   13  ,   1   ,   47  ,   113 ,   231 ,   461 ,   653 ,   1965    ,   849 ,   4513    ,0 };
            const unsigned long dim914Kuo2Init[]        =   {   1   ,   1   ,   3   ,   1   ,   23  ,   7   ,   105 ,   115 ,   463 ,   385 ,   1355    ,   1417    ,   6341    ,0 };
            const unsigned long dim915Kuo2Init[]        =   {   1   ,   1   ,   5   ,   7   ,   9   ,   27  ,   123 ,   193 ,   5   ,   735 ,   733 ,   2385    ,   6569    ,0 };
            const unsigned long dim916Kuo2Init[]        =   {   1   ,   3   ,   1   ,   11  ,   13  ,   51  ,   53  ,   15  ,   289 ,   433 ,   951 ,   3657    ,   5239    ,0 };
            const unsigned long dim917Kuo2Init[]        =   {   1   ,   1   ,   7   ,   3   ,   7   ,   59  ,   69  ,   57  ,   471 ,   247 ,   259 ,   2617    ,   47  ,0 };
            const unsigned long dim918Kuo2Init[]        =   {   1   ,   3   ,   7   ,   13  ,   5   ,   41  ,   45  ,   67  ,   363 ,   295 ,   1715    ,   3567    ,   2457    ,0 };
            const unsigned long dim919Kuo2Init[]        =   {   1   ,   1   ,   5   ,   3   ,   21  ,   19  ,   79  ,   37  ,   273 ,   613 ,   475 ,   3989    ,   3779    ,0 };
            const unsigned long dim920Kuo2Init[]        =   {   1   ,   1   ,   5   ,   3   ,   21  ,   11  ,   53  ,   117 ,   367 ,   295 ,   1027    ,   2579    ,   2717    ,0 };
            const unsigned long dim921Kuo2Init[]        =   {   1   ,   1   ,   5   ,   7   ,   29  ,   1   ,   47  ,   185 ,   377 ,   267 ,   1375    ,   801 ,   5967    ,0 };
            const unsigned long dim922Kuo2Init[]        =   {   1   ,   3   ,   3   ,   3   ,   15  ,   5   ,   51  ,   183 ,   249 ,   789 ,   785 ,   1459    ,   7865    ,0 };
            const unsigned long dim923Kuo2Init[]        =   {   1   ,   3   ,   5   ,   5   ,   31  ,   17  ,   89  ,   9   ,   247 ,   67  ,   1259    ,   3099    ,   2805    ,0 };
            const unsigned long dim924Kuo2Init[]        =   {   1   ,   3   ,   3   ,   11  ,   31  ,   51  ,   55  ,   37  ,   201 ,   779 ,   1235    ,   1905    ,   1995    ,0 };
            const unsigned long dim925Kuo2Init[]        =   {   1   ,   3   ,   1   ,   1   ,   27  ,   31  ,   25  ,   37  ,   443 ,   399 ,   1911    ,   3951    ,   4311    ,0 };
            const unsigned long dim926Kuo2Init[]        =   {   1   ,   3   ,   3   ,   11  ,   11  ,   13  ,   69  ,   209 ,   295 ,   431 ,   573 ,   1781    ,   7767    ,0 };
            const unsigned long dim927Kuo2Init[]        =   {   1   ,   3   ,   1   ,   3   ,   23  ,   55  ,   57  ,   29  ,   137 ,   407 ,   1339    ,   1411    ,   123 ,0 };
            const unsigned long dim928Kuo2Init[]        =   {   1   ,   3   ,   3   ,   13  ,   5   ,   15  ,   7   ,   29  ,   353 ,   143 ,   133 ,   2351    ,   1555    ,0 };
            const unsigned long dim929Kuo2Init[]        =   {   1   ,   1   ,   7   ,   9   ,   17  ,   53  ,   61  ,   197 ,   245 ,   327 ,   147 ,   3231    ,   1825    ,0 };
            const unsigned long dim930Kuo2Init[]        =   {   1   ,   3   ,   5   ,   3   ,   27  ,   55  ,   31  ,   15  ,   429 ,   703 ,   1863    ,   3697    ,   2169    ,0 };
            const unsigned long dim931Kuo2Init[]        =   {   1   ,   3   ,   3   ,   1   ,   19  ,   45  ,   61  ,   179 ,   471 ,   219 ,   1127    ,   3521    ,   8065    ,0 };
            const unsigned long dim932Kuo2Init[]        =   {   1   ,   3   ,   5   ,   5   ,   17  ,   21  ,   77  ,   117 ,   331 ,   979 ,   175 ,   3155    ,   1589    ,0 };
            const unsigned long dim933Kuo2Init[]        =   {   1   ,   3   ,   1   ,   9   ,   13  ,   41  ,   25  ,   221 ,   115 ,   333 ,   1851    ,   543 ,   5479    ,0 };
            const unsigned long dim934Kuo2Init[]        =   {   1   ,   3   ,   3   ,   13  ,   25  ,   35  ,   113 ,   107 ,   21  ,   85  ,   815 ,   3013    ,   8067    ,0 };
            const unsigned long dim935Kuo2Init[]        =   {   1   ,   1   ,   3   ,   9   ,   23  ,   37  ,   55  ,   193 ,   301 ,   463 ,   1361    ,   3455    ,   5745    ,0 };
            const unsigned long dim936Kuo2Init[]        =   {   1   ,   3   ,   5   ,   7   ,   3   ,   63  ,   73  ,   83  ,   35  ,   709 ,   719 ,   1133    ,   6711    ,0 };
            const unsigned long dim937Kuo2Init[]        =   {   1   ,   1   ,   5   ,   7   ,   11  ,   7   ,   67  ,   199 ,   285 ,   527 ,   1181    ,   39  ,   5639    ,0 };
            const unsigned long dim938Kuo2Init[]        =   {   1   ,   3   ,   3   ,   3   ,   17  ,   13  ,   33  ,   213 ,   97  ,   499 ,   1651    ,   3369    ,   6001    ,0 };
            const unsigned long dim939Kuo2Init[]        =   {   1   ,   1   ,   3   ,   1   ,   15  ,   37  ,   67  ,   31  ,   225 ,   205 ,   1009    ,   2579    ,   331 ,0 };
            const unsigned long dim940Kuo2Init[]        =   {   1   ,   1   ,   7   ,   1   ,   29  ,   3   ,   91  ,   153 ,   395 ,   89  ,   2019    ,   1297    ,   7035    ,0 };
            const unsigned long dim941Kuo2Init[]        =   {   1   ,   3   ,   3   ,   5   ,   29  ,   31  ,   65  ,   137 ,   289 ,   351 ,   677 ,   3305    ,   5073    ,0 };
            const unsigned long dim942Kuo2Init[]        =   {   1   ,   1   ,   3   ,   13  ,   15  ,   21  ,   43  ,   227 ,   399 ,   921 ,   123 ,   3199    ,   6925    ,0 };
            const unsigned long dim943Kuo2Init[]        =   {   1   ,   1   ,   7   ,   13  ,   17  ,   11  ,   107 ,   251 ,   55  ,   525 ,   869 ,   209 ,   6209    ,0 };
            const unsigned long dim944Kuo2Init[]        =   {   1   ,   1   ,   3   ,   11  ,   29  ,   39  ,   123 ,   171 ,   223 ,   441 ,   525 ,   2583    ,   71  ,0 };
            const unsigned long dim945Kuo2Init[]        =   {   1   ,   3   ,   7   ,   9   ,   5   ,   3   ,   3   ,   205 ,   119 ,   631 ,   1997    ,   3105    ,   6483    ,0 };
            const unsigned long dim946Kuo2Init[]        =   {   1   ,   1   ,   7   ,   7   ,   23  ,   37  ,   79  ,   245 ,   401 ,   241 ,   1471    ,   2941    ,   4389    ,0 };
            const unsigned long dim947Kuo2Init[]        =   {   1   ,   3   ,   1   ,   13  ,   25  ,   7   ,   43  ,   215 ,   155 ,   805 ,   1497    ,   387 ,   2045    ,0 };
            const unsigned long dim948Kuo2Init[]        =   {   1   ,   1   ,   3   ,   3   ,   7   ,   37  ,   115 ,   39  ,   139 ,   555 ,   1325    ,   2351    ,   6045    ,0 };
            const unsigned long dim949Kuo2Init[]        =   {   1   ,   1   ,   5   ,   7   ,   3   ,   53  ,   35  ,   103 ,   505 ,   361 ,   457 ,   2495    ,   1325    ,0 };
            const unsigned long dim950Kuo2Init[]        =   {   1   ,   3   ,   7   ,   13  ,   21  ,   33  ,   39  ,   81  ,   465 ,   601 ,   1131    ,   2977    ,   1409    ,0 };
            const unsigned long dim951Kuo2Init[]        =   {   1   ,   1   ,   5   ,   5   ,   25  ,   1   ,   21  ,   241 ,   255 ,   663 ,   1509    ,   2455    ,   5705    ,0 };
            const unsigned long dim952Kuo2Init[]        =   {   1   ,   1   ,   7   ,   13  ,   1   ,   47  ,   91  ,   195 ,   195 ,   503 ,   1931    ,   1681    ,   5131    ,0 };
            const unsigned long dim953Kuo2Init[]        =   {   1   ,   1   ,   3   ,   9   ,   27  ,   37  ,   103 ,   21  ,   237 ,   677 ,   1949    ,   357 ,   2781    ,0 };
            const unsigned long dim954Kuo2Init[]        =   {   1   ,   3   ,   7   ,   9   ,   29  ,   51  ,   107 ,   49  ,   251 ,   433 ,   1235    ,   4041    ,   2971    ,0 };
            const unsigned long dim955Kuo2Init[]        =   {   1   ,   3   ,   3   ,   1   ,   19  ,   37  ,   113 ,   75  ,   365 ,   145 ,   93  ,   1185    ,   1871    ,0 };
            const unsigned long dim956Kuo2Init[]        =   {   1   ,   1   ,   5   ,   5   ,   25  ,   15  ,   113 ,   239 ,   27  ,   321 ,   1351    ,   1441    ,   3333    ,0 };
            const unsigned long dim957Kuo2Init[]        =   {   1   ,   1   ,   5   ,   13  ,   31  ,   35  ,   103 ,   83  ,   111 ,   573 ,   59  ,   3753    ,   5667    ,0 };
            const unsigned long dim958Kuo2Init[]        =   {   1   ,   1   ,   5   ,   13  ,   1   ,   43  ,   75  ,   115 ,   397 ,   187 ,   1669    ,   3837    ,   5219    ,0 };
            const unsigned long dim959Kuo2Init[]        =   {   1   ,   1   ,   3   ,   1   ,   23  ,   13  ,   65  ,   145 ,   271 ,   909 ,   1517    ,   1659    ,   7975    ,0 };
            const unsigned long dim960Kuo2Init[]        =   {   1   ,   1   ,   7   ,   15  ,   21  ,   23  ,   1   ,   33  ,   37  ,   313 ,   687 ,   2375    ,   555 ,0 };
            const unsigned long dim961Kuo2Init[]        =   {   1   ,   1   ,   7   ,   13  ,   1   ,   35  ,   11  ,   157 ,   295 ,   403 ,   1295    ,   3301    ,   4805    ,0 };
            const unsigned long dim962Kuo2Init[]        =   {   1   ,   3   ,   1   ,   13  ,   9   ,   27  ,   3   ,   111 ,   133 ,   395 ,   981 ,   2399    ,   7009    ,0 };
            const unsigned long dim963Kuo2Init[]        =   {   1   ,   1   ,   7   ,   15  ,   23  ,   11  ,   75  ,   31  ,   5   ,   599 ,   1589    ,   2243    ,   3263    ,0 };
            const unsigned long dim964Kuo2Init[]        =   {   1   ,   1   ,   3   ,   9   ,   7   ,   49  ,   31  ,   5   ,   457 ,   93  ,   221 ,   2463    ,   2287    ,0 };
            const unsigned long dim965Kuo2Init[]        =   {   1   ,   1   ,   5   ,   3   ,   9   ,   31  ,   49  ,   5   ,   481 ,   145 ,   53  ,   2477    ,   5569    ,0 };
            const unsigned long dim966Kuo2Init[]        =   {   1   ,   1   ,   1   ,   1   ,   19  ,   47  ,   59  ,   83  ,   19  ,   887 ,   335 ,   2305    ,   6265    ,0 };
            const unsigned long dim967Kuo2Init[]        =   {   1   ,   3   ,   7   ,   15  ,   27  ,   15  ,   93  ,   37  ,   213 ,   17  ,   311 ,   3159    ,   6911    ,0 };
            const unsigned long dim968Kuo2Init[]        =   {   1   ,   1   ,   3   ,   9   ,   19  ,   47  ,   93  ,   31  ,   489 ,   731 ,   1905    ,   1675    ,   1   ,0 };
            const unsigned long dim969Kuo2Init[]        =   {   1   ,   1   ,   1   ,   7   ,   9   ,   3   ,   17  ,   185 ,   203 ,   963 ,   521 ,   2797    ,   7511    ,0 };
            const unsigned long dim970Kuo2Init[]        =   {   1   ,   3   ,   7   ,   11  ,   31  ,   27  ,   47  ,   147 ,   467 ,   505 ,   1991    ,   2247    ,   4207    ,0 };
            const unsigned long dim971Kuo2Init[]        =   {   1   ,   3   ,   1   ,   5   ,   9   ,   11  ,   89  ,   199 ,   359 ,   723 ,   1203    ,   1511    ,   7915    ,0 };
            const unsigned long dim972Kuo2Init[]        =   {   1   ,   3   ,   5   ,   9   ,   27  ,   63  ,   93  ,   19  ,   305 ,   641 ,   997 ,   1725    ,   5681    ,0 };
            const unsigned long dim973Kuo2Init[]        =   {   1   ,   3   ,   3   ,   3   ,   13  ,   19  ,   9   ,   43  ,   99  ,   111 ,   1353    ,   935 ,   7765    ,0 };
            const unsigned long dim974Kuo2Init[]        =   {   1   ,   3   ,   1   ,   9   ,   5   ,   39  ,   109 ,   79  ,   497 ,   351 ,   1345    ,   3019    ,   5909    ,0 };
            const unsigned long dim975Kuo2Init[]        =   {   1   ,   1   ,   3   ,   1   ,   13  ,   29  ,   11  ,   199 ,   31  ,   443 ,   1587    ,   3397    ,   7045    ,0 };
            const unsigned long dim976Kuo2Init[]        =   {   1   ,   1   ,   1   ,   11  ,   7   ,   9   ,   49  ,   107 ,   153 ,   207 ,   887 ,   2135    ,   1389    ,0 };
            const unsigned long dim977Kuo2Init[]        =   {   1   ,   1   ,   1   ,   3   ,   29  ,   53  ,   111 ,   143 ,   421 ,   521 ,   1099    ,   1709    ,   7027    ,0 };
            const unsigned long dim978Kuo2Init[]        =   {   1   ,   1   ,   5   ,   5   ,   5   ,   19  ,   127 ,   99  ,   501 ,   819 ,   1729    ,   1701    ,   5829    ,0 };
            const unsigned long dim979Kuo2Init[]        =   {   1   ,   3   ,   7   ,   5   ,   7   ,   15  ,   43  ,   49  ,   223 ,   793 ,   889 ,   793 ,   271 ,0 };
            const unsigned long dim980Kuo2Init[]        =   {   1   ,   3   ,   5   ,   11  ,   27  ,   27  ,   55  ,   69  ,   185 ,   511 ,   951 ,   2821    ,   5357    ,0 };
            const unsigned long dim981Kuo2Init[]        =   {   1   ,   3   ,   7   ,   13  ,   17  ,   19  ,   31  ,   9   ,   401 ,   397 ,   829 ,   1063    ,   693 ,0 };
            const unsigned long dim982Kuo2Init[]        =   {   1   ,   1   ,   7   ,   9   ,   21  ,   13  ,   75  ,   13  ,   327 ,   929 ,   955 ,   1421    ,   2809    ,0 };
            const unsigned long dim983Kuo2Init[]        =   {   1   ,   1   ,   3   ,   15  ,   17  ,   5   ,   115 ,   231 ,   29  ,   23  ,   1743    ,   779 ,   5083    ,0 };
            const unsigned long dim984Kuo2Init[]        =   {   1   ,   1   ,   7   ,   11  ,   9   ,   51  ,   93  ,   3   ,   155 ,   111 ,   243 ,   3889    ,   6733    ,0 };
            const unsigned long dim985Kuo2Init[]        =   {   1   ,   1   ,   5   ,   9   ,   29  ,   63  ,   119 ,   145 ,   391 ,   1021    ,   503 ,   555 ,   4657    ,0 };
            const unsigned long dim986Kuo2Init[]        =   {   1   ,   1   ,   1   ,   15  ,   15  ,   41  ,   69  ,   131 ,   365 ,   195 ,   1809    ,   2773    ,   4067    ,0 };
            const unsigned long dim987Kuo2Init[]        =   {   1   ,   1   ,   1   ,   15  ,   5   ,   3   ,   119 ,   63  ,   243 ,   759 ,   2037    ,   3443    ,   5771    ,0 };
            const unsigned long dim988Kuo2Init[]        =   {   1   ,   3   ,   5   ,   5   ,   1   ,   1   ,   21  ,   163 ,   481 ,   781 ,   769 ,   3031    ,   2707    ,0 };
            const unsigned long dim989Kuo2Init[]        =   {   1   ,   1   ,   5   ,   7   ,   15  ,   29  ,   3   ,   119 ,   13  ,   977 ,   1759    ,   2327    ,   3961    ,0 };
            const unsigned long dim990Kuo2Init[]        =   {   1   ,   1   ,   3   ,   1   ,   9   ,   25  ,   55  ,   131 ,   335 ,   289 ,   1049    ,   3589    ,   3131    ,0 };
            const unsigned long dim991Kuo2Init[]        =   {   1   ,   1   ,   7   ,   1   ,   27  ,   55  ,   113 ,   97  ,   249 ,   817 ,   1497    ,   3773    ,   6183    ,0 };
            const unsigned long dim992Kuo2Init[]        =   {   1   ,   3   ,   3   ,   7   ,   3   ,   15  ,   11  ,   157 ,   201 ,   145 ,   1741    ,   721 ,   4945    ,0 };
            const unsigned long dim993Kuo2Init[]        =   {   1   ,   1   ,   7   ,   3   ,   19  ,   9   ,   123 ,   151 ,   169 ,   841 ,   65  ,   2463    ,   5357    ,0 };
            const unsigned long dim994Kuo2Init[]        =   {   1   ,   3   ,   1   ,   11  ,   5   ,   31  ,   123 ,   69  ,   93  ,   541 ,   131 ,   1469    ,   5791    ,0 };
            const unsigned long dim995Kuo2Init[]        =   {   1   ,   3   ,   5   ,   1   ,   25  ,   23  ,   115 ,   69  ,   333 ,   457 ,   17  ,   1155    ,   523 ,0 };
            const unsigned long dim996Kuo2Init[]        =   {   1   ,   3   ,   5   ,   11  ,   11  ,   51  ,   21  ,   145 ,   209 ,   557 ,   1109    ,   2293    ,   4631    ,0 };
            const unsigned long dim997Kuo2Init[]        =   {   1   ,   1   ,   7   ,   5   ,   11  ,   21  ,   107 ,   41  ,   409 ,   485 ,   1745    ,   3471    ,   1699    ,0 };
            const unsigned long dim998Kuo2Init[]        =   {   1   ,   1   ,   3   ,   7   ,   5   ,   47  ,   65  ,   79  ,   299 ,   191 ,   381 ,   3803    ,   6499    ,0 };
            const unsigned long dim999Kuo2Init[]        =   {   1   ,   3   ,   3   ,   7   ,   19  ,   11  ,   73  ,   69  ,   461 ,   777 ,   1871    ,   47  ,   1281    ,0 };
            const unsigned long dim1000Kuo2Init[]       =   {   1   ,   3   ,   5   ,   7   ,   13  ,   39  ,   63  ,   103 ,   103 ,   845 ,   1023    ,   3241    ,   1291    ,0 };
            const unsigned long dim1001Kuo2Init[]       =   {   1   ,   1   ,   5   ,   13  ,   3   ,   11  ,   103 ,   135 ,   403 ,   801 ,   847 ,   371 ,   6243    ,0 };
            const unsigned long dim1002Kuo2Init[]       =   {   1   ,   3   ,   1   ,   5   ,   21  ,   57  ,   19  ,   205 ,   33  ,   399 ,   777 ,   2981    ,   2365    ,0 };
            const unsigned long dim1003Kuo2Init[]       =   {   1   ,   1   ,   1   ,   5   ,   17  ,   11  ,   105 ,   23  ,   243 ,   581 ,   1391    ,   1509    ,   8009    ,0 };
            const unsigned long dim1004Kuo2Init[]       =   {   1   ,   3   ,   3   ,   3   ,   11  ,   13  ,   85  ,   107 ,   181 ,   89  ,   1323    ,   3059    ,   5097    ,0 };
            const unsigned long dim1005Kuo2Init[]       =   {   1   ,   1   ,   1   ,   15  ,   1   ,   29  ,   57  ,   221 ,   389 ,   127 ,   841 ,   1727    ,   6217    ,0 };
            const unsigned long dim1006Kuo2Init[]       =   {   1   ,   3   ,   1   ,   9   ,   25  ,   51  ,   89  ,   67  ,   257 ,   921 ,   621 ,   1523    ,   3449    ,0 };
            const unsigned long dim1007Kuo2Init[]       =   {   1   ,   3   ,   5   ,   5   ,   25  ,   15  ,   25  ,   47  ,   29  ,   775 ,   1925    ,   3819    ,   8063    ,0 };
            const unsigned long dim1008Kuo2Init[]       =   {   1   ,   1   ,   1   ,   5   ,   25  ,   55  ,   89  ,   75  ,   47  ,   135 ,   1155    ,   1923    ,   2241    ,0 };
            const unsigned long dim1009Kuo2Init[]       =   {   1   ,   3   ,   5   ,   9   ,   29  ,   5   ,   55  ,   43  ,   355 ,   879 ,   127 ,   3565    ,   2583    ,0 };
            const unsigned long dim1010Kuo2Init[]       =   {   1   ,   1   ,   3   ,   13  ,   23  ,   47  ,   27  ,   63  ,   217 ,   119 ,   357 ,   423 ,   5897    ,0 };
            const unsigned long dim1011Kuo2Init[]       =   {   1   ,   3   ,   5   ,   1   ,   23  ,   1   ,   89  ,   255 ,   381 ,   175 ,   1069    ,   3809    ,   1743    ,0 };
            const unsigned long dim1012Kuo2Init[]       =   {   1   ,   1   ,   7   ,   3   ,   7   ,   35  ,   47  ,   159 ,   347 ,   543 ,   187 ,   1103    ,   5019    ,0 };
            const unsigned long dim1013Kuo2Init[]       =   {   1   ,   1   ,   1   ,   7   ,   19  ,   3   ,   89  ,   15  ,   487 ,   541 ,   625 ,   1711    ,   3713    ,0 };
            const unsigned long dim1014Kuo2Init[]       =   {   1   ,   1   ,   5   ,   15  ,   17  ,   61  ,   13  ,   45  ,   359 ,   107 ,   1487    ,   121 ,   6919    ,0 };
            const unsigned long dim1015Kuo2Init[]       =   {   1   ,   3   ,   3   ,   3   ,   1   ,   5   ,   99  ,   255 ,   147 ,   737 ,   1675    ,   4069    ,   7685    ,0 };
            const unsigned long dim1016Kuo2Init[]       =   {   1   ,   3   ,   7   ,   1   ,   5   ,   33  ,   39  ,   217 ,   255 ,   833 ,   661 ,   3411    ,   7273    ,0 };
            const unsigned long dim1017Kuo2Init[]       =   {   1   ,   3   ,   3   ,   5   ,   9   ,   41  ,   25  ,   223 ,   319 ,   45  ,   687 ,   3861    ,   2603    ,0 };
            const unsigned long dim1018Kuo2Init[]       =   {   1   ,   1   ,   3   ,   15  ,   15  ,   43  ,   61  ,   163 ,   45  ,   835 ,   545 ,   561 ,   4895    ,0 };
            const unsigned long dim1019Kuo2Init[]       =   {   1   ,   3   ,   7   ,   3   ,   21  ,   61  ,   65  ,   119 ,   337 ,   609 ,   1185    ,   2863    ,   1431    ,0 };
            const unsigned long dim1020Kuo2Init[]       =   {   1   ,   1   ,   3   ,   3   ,   29  ,   49  ,   51  ,   173 ,   107 ,   21  ,   1655    ,   1333    ,   4703    ,0 };
            const unsigned long dim1021Kuo2Init[]       =   {   1   ,   3   ,   1   ,   3   ,   27  ,   23  ,   111 ,   77  ,   3   ,   789 ,   943 ,   3089    ,   2865    ,0 };
            const unsigned long dim1022Kuo2Init[]       =   {   1   ,   3   ,   1   ,   1   ,   5   ,   7   ,   85  ,   209 ,   317 ,   241 ,   1159    ,   2193    ,   229 ,0 };
            const unsigned long dim1023Kuo2Init[]       =   {   1   ,   1   ,   3   ,   13  ,   23  ,   27  ,   115 ,   111 ,   265 ,   427 ,   1457    ,   2281    ,   2235    ,0 };
            const unsigned long dim1024Kuo2Init[]       =   {   1   ,   3   ,   1   ,   13  ,   5   ,   51  ,   21  ,   125 ,   335 ,   359 ,   1189    ,   1095    ,   6843    ,0 };
            const unsigned long dim1025Kuo2Init[]       =   {   1   ,   3   ,   5   ,   7   ,   5   ,   17  ,   93  ,   43  ,   297 ,   109 ,   1721    ,   3991    ,   6861    ,0 };
            const unsigned long dim1026Kuo2Init[]       =   {   1   ,   3   ,   7   ,   7   ,   23  ,   45  ,   61  ,   149 ,   89  ,   1009    ,   937 ,   2477    ,   1617    ,0 };
            const unsigned long dim1027Kuo2Init[]       =   {   1   ,   1   ,   3   ,   1   ,   23  ,   19  ,   21  ,   29  ,   481 ,   771 ,   261 ,   2377    ,   2323    ,0 };
            const unsigned long dim1028Kuo2Init[]       =   {   1   ,   1   ,   3   ,   7   ,   29  ,   21  ,   99  ,   159 ,   219 ,   983 ,   1169    ,   667 ,   3073    ,0 };
            const unsigned long dim1029Kuo2Init[]       =   {   1   ,   3   ,   5   ,   9   ,   7   ,   61  ,   69  ,   197 ,   29  ,   3   ,   1767    ,   2113    ,   689 ,0 };
            const unsigned long dim1030Kuo2Init[]       =   {   1   ,   1   ,   3   ,   7   ,   9   ,   37  ,   45  ,   11  ,   405 ,   243 ,   661 ,   2545    ,   5185    ,0 };
            const unsigned long dim1031Kuo2Init[]       =   {   1   ,   1   ,   7   ,   3   ,   7   ,   23  ,   13  ,   193 ,   91  ,   551 ,   1963    ,   2663    ,   7605    ,0 };
            const unsigned long dim1032Kuo2Init[]       =   {   1   ,   1   ,   1   ,   1   ,   27  ,   17  ,   79  ,   251 ,   103 ,   33  ,   45  ,   2517    ,   3611    ,0 };
            const unsigned long dim1033Kuo2Init[]       =   {   1   ,   3   ,   3   ,   1   ,   19  ,   57  ,   15  ,   175 ,   77  ,   299 ,   1805    ,   3331    ,   1365    ,0 };
            const unsigned long dim1034Kuo2Init[]       =   {   1   ,   1   ,   3   ,   5   ,   31  ,   59  ,   51  ,   85  ,   359 ,   197 ,   679 ,   3017    ,   7929    ,0 };
            const unsigned long dim1035Kuo2Init[]       =   {   1   ,   3   ,   7   ,   13  ,   15  ,   19  ,   71  ,   31  ,   69  ,   353 ,   489 ,   1167    ,   337 ,0 };
            const unsigned long dim1036Kuo2Init[]       =   {   1   ,   1   ,   3   ,   13  ,   5   ,   33  ,   35  ,   33  ,   345 ,   63  ,   1775    ,   3475    ,   1375    ,0 };
            const unsigned long dim1037Kuo2Init[]       =   {   1   ,   1   ,   3   ,   1   ,   29  ,   53  ,   109 ,   3   ,   191 ,   1011    ,   1043    ,   3057    ,   4757    ,0 };
            const unsigned long dim1038Kuo2Init[]       =   {   1   ,   1   ,   5   ,   11  ,   3   ,   21  ,   101 ,   137 ,   155 ,   959 ,   683 ,   1407    ,   1239    ,0 };
            const unsigned long dim1039Kuo2Init[]       =   {   1   ,   3   ,   5   ,   13  ,   3   ,   9   ,   117 ,   81  ,   143 ,   753 ,   561 ,   871 ,   2417    ,0 };
            const unsigned long dim1040Kuo2Init[]       =   {   1   ,   1   ,   7   ,   7   ,   25  ,   11  ,   33  ,   173 ,   321 ,   487 ,   33  ,   551 ,   6613    ,0 };
            const unsigned long dim1041Kuo2Init[]       =   {   1   ,   3   ,   5   ,   7   ,   3   ,   29  ,   9   ,   133 ,   251 ,   561 ,   975 ,   3153    ,   1115    ,0 };
            const unsigned long dim1042Kuo2Init[]       =   {   1   ,   3   ,   5   ,   3   ,   1   ,   11  ,   89  ,   145 ,   33  ,   967 ,   365 ,   3661    ,   2917    ,0 };
            const unsigned long dim1043Kuo2Init[]       =   {   1   ,   3   ,   7   ,   5   ,   21  ,   55  ,   49  ,   139 ,   507 ,   405 ,   99  ,   2819    ,   6541    ,0 };
            const unsigned long dim1044Kuo2Init[]       =   {   1   ,   1   ,   7   ,   1   ,   13  ,   41  ,   77  ,   53  ,   185 ,   995 ,   2045    ,   1659    ,   1081    ,0 };
            const unsigned long dim1045Kuo2Init[]       =   {   1   ,   3   ,   1   ,   3   ,   5   ,   29  ,   25  ,   111 ,   173 ,   793 ,   953 ,   2825    ,   673 ,0 };
            const unsigned long dim1046Kuo2Init[]       =   {   1   ,   3   ,   1   ,   3   ,   13  ,   3   ,   73  ,   39  ,   143 ,   603 ,   253 ,   2879    ,   3827    ,0 };
            const unsigned long dim1047Kuo2Init[]       =   {   1   ,   3   ,   1   ,   11  ,   3   ,   35  ,   93  ,   237 ,   401 ,   1005    ,   1047    ,   3065    ,   6285    ,0 };
            const unsigned long dim1048Kuo2Init[]       =   {   1   ,   3   ,   1   ,   13  ,   7   ,   7   ,   17  ,   65  ,   107 ,   971 ,   999 ,   623 ,   6491    ,0 };
            const unsigned long dim1049Kuo2Init[]       =   {   1   ,   3   ,   3   ,   11  ,   31  ,   1   ,   83  ,   111 ,   255 ,   969 ,   871 ,   3459    ,   2419    ,0 };
            const unsigned long dim1050Kuo2Init[]       =   {   1   ,   3   ,   1   ,   7   ,   3   ,   49  ,   99  ,   27  ,   253 ,   937 ,   1237    ,   343 ,   6133    ,0 };
            const unsigned long dim1051Kuo2Init[]       =   {   1   ,   1   ,   5   ,   11  ,   31  ,   7   ,   83  ,   75  ,   77  ,   583 ,   847 ,   2263    ,   1925    ,0 };
            const unsigned long dim1052Kuo2Init[]       =   {   1   ,   1   ,   3   ,   11  ,   3   ,   37  ,   119 ,   229 ,   411 ,   953 ,   1087    ,   3005    ,   1521    ,0 };
            const unsigned long dim1053Kuo2Init[]       =   {   1   ,   3   ,   7   ,   13  ,   5   ,   53  ,   125 ,   207 ,   297 ,   465 ,   1989    ,   1769    ,   5113    ,0 };
            const unsigned long dim1054Kuo2Init[]       =   {   1   ,   3   ,   5   ,   11  ,   13  ,   13  ,   9   ,   105 ,   303 ,   641 ,   997 ,   929 ,   393 ,0 };
            const unsigned long dim1055Kuo2Init[]       =   {   1   ,   3   ,   5   ,   9   ,   21  ,   21  ,   125 ,   49  ,   509 ,   993 ,   317 ,   2367    ,   2775    ,0 };
            const unsigned long dim1056Kuo2Init[]       =   {   1   ,   3   ,   3   ,   9   ,   15  ,   1   ,   91  ,   101 ,   339 ,   625 ,   1521    ,   2285    ,   1847    ,0 };
            const unsigned long dim1057Kuo2Init[]       =   {   1   ,   1   ,   5   ,   15  ,   15  ,   31  ,   57  ,   253 ,   407 ,   715 ,   1143    ,   3707    ,   5105    ,0 };
            const unsigned long dim1058Kuo2Init[]       =   {   1   ,   3   ,   1   ,   5   ,   21  ,   1   ,   77  ,   139 ,   499 ,   777 ,   349 ,   1647    ,   1389    ,0 };
            const unsigned long dim1059Kuo2Init[]       =   {   1   ,   3   ,   5   ,   5   ,   11  ,   55  ,   29  ,   209 ,   491 ,   823 ,   1961    ,   3833    ,   5825    ,0 };
            const unsigned long dim1060Kuo2Init[]       =   {   1   ,   1   ,   7   ,   11  ,   3   ,   21  ,   43  ,   147 ,   83  ,   245 ,   139 ,   3585    ,   975 ,0 };
            const unsigned long dim1061Kuo2Init[]       =   {   1   ,   1   ,   1   ,   3   ,   19  ,   9   ,   19  ,   251 ,   163 ,   85  ,   239 ,   2573    ,   1079    ,0 };
            const unsigned long dim1062Kuo2Init[]       =   {   1   ,   1   ,   5   ,   13  ,   27  ,   35  ,   27  ,   191 ,   279 ,   589 ,   1737    ,   2759    ,   1691    ,0 };
            const unsigned long dim1063Kuo2Init[]       =   {   1   ,   1   ,   7   ,   3   ,   17  ,   5   ,   47  ,   173 ,   183 ,   713 ,   667 ,   3683    ,   2167    ,0 };
            const unsigned long dim1064Kuo2Init[]       =   {   1   ,   3   ,   5   ,   13  ,   25  ,   41  ,   29  ,   125 ,   173 ,   325 ,   1341    ,   481 ,   365 ,0 };
            const unsigned long dim1065Kuo2Init[]       =   {   1   ,   3   ,   1   ,   5   ,   27  ,   61  ,   23  ,   11  ,   265 ,   255 ,   801 ,   921 ,   7145    ,0 };
            const unsigned long dim1066Kuo2Init[]       =   {   1   ,   3   ,   5   ,   5   ,   27  ,   49  ,   27  ,   245 ,   253 ,   787 ,   1451    ,   2349    ,   2125    ,0 };
            const unsigned long dim1067Kuo2Init[]       =   {   1   ,   1   ,   7   ,   7   ,   13  ,   31  ,   61  ,   215 ,   339 ,   315 ,   377 ,   1121    ,   2667    ,0 };
            const unsigned long dim1068Kuo2Init[]       =   {   1   ,   3   ,   5   ,   9   ,   31  ,   63  ,   27  ,   115 ,   367 ,   939 ,   437 ,   1771    ,   1219    ,0 };
            const unsigned long dim1069Kuo2Init[]       =   {   1   ,   1   ,   3   ,   5   ,   3   ,   7   ,   33  ,   147 ,   177 ,   249 ,   649 ,   2505    ,   4705    ,0 };
            const unsigned long dim1070Kuo2Init[]       =   {   1   ,   3   ,   1   ,   9   ,   11  ,   31  ,   23  ,   237 ,   223 ,   311 ,   493 ,   1255    ,   2121    ,0 };
            const unsigned long dim1071Kuo2Init[]       =   {   1   ,   3   ,   5   ,   7   ,   5   ,   43  ,   51  ,   1   ,   425 ,   115 ,   1487    ,   373 ,   3765    ,0 };
            const unsigned long dim1072Kuo2Init[]       =   {   1   ,   3   ,   3   ,   9   ,   3   ,   11  ,   125 ,   31  ,   375 ,   73  ,   673 ,   1933    ,   7763    ,0 };
            const unsigned long dim1073Kuo2Init[]       =   {   1   ,   1   ,   3   ,   7   ,   31  ,   43  ,   67  ,   177 ,   33  ,   693 ,   185 ,   137 ,   4869    ,0 };
            const unsigned long dim1074Kuo2Init[]       =   {   1   ,   1   ,   1   ,   9   ,   21  ,   53  ,   111 ,   41  ,   249 ,   845 ,   193 ,   2519    ,   4745    ,0 };
            const unsigned long dim1075Kuo2Init[]       =   {   1   ,   3   ,   5   ,   3   ,   11  ,   15  ,   11  ,   249 ,   325 ,   561 ,   411 ,   71  ,   6973    ,0 };
            const unsigned long dim1076Kuo2Init[]       =   {   1   ,   1   ,   7   ,   5   ,   5   ,   13  ,   31  ,   113 ,   149 ,   475 ,   869 ,   627 ,   2789    ,0 };
            const unsigned long dim1077Kuo2Init[]       =   {   1   ,   1   ,   1   ,   5   ,   7   ,   23  ,   49  ,   65  ,   201 ,   111 ,   1061    ,   393 ,   2101    ,0 };
            const unsigned long dim1078Kuo2Init[]       =   {   1   ,   1   ,   3   ,   5   ,   23  ,   47  ,   49  ,   111 ,   315 ,   227 ,   1985    ,   3579    ,   875 ,0 };
            const unsigned long dim1079Kuo2Init[]       =   {   1   ,   3   ,   5   ,   5   ,   21  ,   17  ,   101 ,   135 ,   475 ,   505 ,   51  ,   3203    ,   1735    ,0 };
            const unsigned long dim1080Kuo2Init[]       =   {   1   ,   1   ,   7   ,   15  ,   19  ,   41  ,   53  ,   197 ,   117 ,   315 ,   1663    ,   143 ,   1081    ,0 };
            const unsigned long dim1081Kuo2Init[]       =   {   1   ,   1   ,   3   ,   7   ,   5   ,   29  ,   87  ,   39  ,   235 ,   675 ,   1101    ,   1257    ,   4045    ,0 };
            const unsigned long dim1082Kuo2Init[]       =   {   1   ,   3   ,   5   ,   15  ,   25  ,   49  ,   63  ,   125 ,   447 ,   181 ,   357 ,   3017    ,   6861    ,0 };
            const unsigned long dim1083Kuo2Init[]       =   {   1   ,   1   ,   5   ,   15  ,   31  ,   21  ,   61  ,   197 ,   133 ,   85  ,   913 ,   3031    ,   855 ,0 };
            const unsigned long dim1084Kuo2Init[]       =   {   1   ,   1   ,   5   ,   7   ,   31  ,   45  ,   111 ,   243 ,   277 ,   569 ,   1311    ,   125 ,   3417    ,0 };
            const unsigned long dim1085Kuo2Init[]       =   {   1   ,   1   ,   3   ,   11  ,   9   ,   49  ,   127 ,   151 ,   407 ,   407 ,   915 ,   1831    ,   4093    ,0 };
            const unsigned long dim1086Kuo2Init[]       =   {   1   ,   3   ,   5   ,   11  ,   3   ,   47  ,   69  ,   97  ,   315 ,   993 ,   845 ,   67  ,   39  ,0 };
            const unsigned long dim1087Kuo2Init[]       =   {   1   ,   3   ,   3   ,   11  ,   1   ,   5   ,   105 ,   115 ,   293 ,   187 ,   1369    ,   3039    ,   6833    ,0 };
            const unsigned long dim1088Kuo2Init[]       =   {   1   ,   3   ,   3   ,   15  ,   23  ,   19  ,   15  ,   59  ,   139 ,   523 ,   1599    ,   1773    ,   8059    ,0 };
            const unsigned long dim1089Kuo2Init[]       =   {   1   ,   1   ,   1   ,   13  ,   3   ,   9   ,   39  ,   205 ,   115 ,   35  ,   1737    ,   2713    ,   3263    ,0 };
            const unsigned long dim1090Kuo2Init[]       =   {   1   ,   1   ,   3   ,   7   ,   11  ,   41  ,   71  ,   61  ,   329 ,   11  ,   1069    ,   2565    ,   6359    ,0 };
            const unsigned long dim1091Kuo2Init[]       =   {   1   ,   3   ,   7   ,   9   ,   1   ,   57  ,   5   ,   17  ,   117 ,   37  ,   827 ,   3449    ,   1297    ,0 };
            const unsigned long dim1092Kuo2Init[]       =   {   1   ,   1   ,   5   ,   1   ,   29  ,   49  ,   3   ,   7   ,   145 ,   183 ,   1777    ,   2351    ,   6675    ,0 };
            const unsigned long dim1093Kuo2Init[]       =   {   1   ,   1   ,   5   ,   7   ,   5   ,   57  ,   97  ,   61  ,   197 ,   713 ,   695 ,   2385    ,   1917    ,0 };
            const unsigned long dim1094Kuo2Init[]       =   {   1   ,   3   ,   1   ,   11  ,   21  ,   49  ,   41  ,   133 ,   307 ,   473 ,   1983    ,   3693    ,   3241    ,0 };
            const unsigned long dim1095Kuo2Init[]       =   {   1   ,   1   ,   5   ,   1   ,   3   ,   59  ,   47  ,   141 ,   49  ,   531 ,   1237    ,   3085    ,   5651    ,0 };
            const unsigned long dim1096Kuo2Init[]       =   {   1   ,   1   ,   5   ,   1   ,   17  ,   21  ,   3   ,   101 ,   169 ,   475 ,   473 ,   255 ,   6637    ,0 };
            const unsigned long dim1097Kuo2Init[]       =   {   1   ,   3   ,   5   ,   1   ,   19  ,   25  ,   71  ,   251 ,   327 ,   715 ,   1945    ,   1123    ,   4911    ,0 };
            const unsigned long dim1098Kuo2Init[]       =   {   1   ,   1   ,   5   ,   3   ,   15  ,   55  ,   127 ,   221 ,   163 ,   281 ,   1391    ,   4083    ,   7995    ,0 };
            const unsigned long dim1099Kuo2Init[]       =   {   1   ,   1   ,   3   ,   9   ,   27  ,   7   ,   65  ,   143 ,   433 ,   599 ,   1971    ,   2169    ,   813 ,0 };
            const unsigned long dim1100Kuo2Init[]       =   {   1   ,   3   ,   3   ,   1   ,   9   ,   49  ,   107 ,   103 ,   381 ,   1005    ,   705 ,   2385    ,   5007    ,0 };
            const unsigned long dim1101Kuo2Init[]       =   {   1   ,   3   ,   5   ,   7   ,   15  ,   45  ,   41  ,   175 ,   137 ,   1019    ,   1869    ,   2337    ,   3433    ,0 };
            const unsigned long dim1102Kuo2Init[]       =   {   1   ,   3   ,   7   ,   15  ,   15  ,   51  ,   89  ,   193 ,   325 ,   391 ,   197 ,   2607    ,   6717    ,0 };
            const unsigned long dim1103Kuo2Init[]       =   {   1   ,   3   ,   7   ,   1   ,   15  ,   11  ,   113 ,   245 ,   255 ,   481 ,   237 ,   3205    ,   4551    ,0 };
            const unsigned long dim1104Kuo2Init[]       =   {   1   ,   3   ,   3   ,   1   ,   11  ,   19  ,   93  ,   97  ,   67  ,   793 ,   1165    ,   989 ,   4643    ,0 };
            const unsigned long dim1105Kuo2Init[]       =   {   1   ,   1   ,   1   ,   15  ,   17  ,   51  ,   63  ,   37  ,   353 ,   745 ,   1439    ,   2599    ,   2771    ,0 };
            const unsigned long dim1106Kuo2Init[]       =   {   1   ,   1   ,   7   ,   9   ,   7   ,   3   ,   87  ,   245 ,   257 ,   399 ,   311 ,   3063    ,   3801    ,0 };
            const unsigned long dim1107Kuo2Init[]       =   {   1   ,   1   ,   5   ,   11  ,   13  ,   3   ,   83  ,   39  ,   165 ,   827 ,   1581    ,   2491    ,   2067    ,0 };
            const unsigned long dim1108Kuo2Init[]       =   {   1   ,   3   ,   3   ,   3   ,   21  ,   3   ,   21  ,   95  ,   505 ,   971 ,   1599    ,   2479    ,   4155    ,0 };
            const unsigned long dim1109Kuo2Init[]       =   {   1   ,   3   ,   1   ,   11  ,   11  ,   57  ,   115 ,   83  ,   473 ,   457 ,   2015    ,   1855    ,   1851    ,0 };
            const unsigned long dim1110Kuo2Init[]       =   {   1   ,   3   ,   7   ,   11  ,   29  ,   7   ,   73  ,   203 ,   445 ,   771 ,   1225    ,   3587    ,   245 ,0 };
            const unsigned long dim1111Kuo2Init[]       =   {   1   ,   3   ,   5   ,   1   ,   13  ,   53  ,   79  ,   245 ,   347 ,   277 ,   779 ,   653 ,   6213    ,   7661    ,0 };
            const unsigned long dim1112Kuo2Init[]       =   {   1   ,   3   ,   5   ,   3   ,   17  ,   43  ,   31  ,   251 ,   177 ,   337 ,   157 ,   3171    ,   3813    ,   7787    ,0 };
            const unsigned long dim1113Kuo2Init[]       =   {   1   ,   3   ,   5   ,   9   ,   23  ,   9   ,   7   ,   45  ,   65  ,   517 ,   1449    ,   1943    ,   4995    ,   2403    ,0 };
            const unsigned long dim1114Kuo2Init[]       =   {   1   ,   3   ,   3   ,   9   ,   19  ,   45  ,   71  ,   91  ,   491 ,   113 ,   1041    ,   717 ,   681 ,   13331   ,0 };
            const unsigned long dim1115Kuo2Init[]       =   {   1   ,   3   ,   5   ,   9   ,   15  ,   35  ,   19  ,   25  ,   119 ,   677 ,   1355    ,   3065    ,   6325    ,   7821    ,0 };
            const unsigned long dim1116Kuo2Init[]       =   {   1   ,   3   ,   7   ,   5   ,   27  ,   63  ,   41  ,   117 ,   271 ,   327 ,   1731    ,   3501    ,   4489    ,   11119   ,0 };
            const unsigned long dim1117Kuo2Init[]       =   {   1   ,   1   ,   3   ,   5   ,   11  ,   35  ,   51  ,   79  ,   347 ,   905 ,   857 ,   3647    ,   4773    ,   15111   ,0 };
            const unsigned long dim1118Kuo2Init[]       =   {   1   ,   3   ,   7   ,   3   ,   21  ,   3   ,   19  ,   147 ,   429 ,   275 ,   1813    ,   1937    ,   4489    ,   7993    ,0 };
            const unsigned long dim1119Kuo2Init[]       =   {   1   ,   3   ,   7   ,   9   ,   27  ,   59  ,   77  ,   179 ,   209 ,   897 ,   807 ,   1121    ,   7575    ,   9769    ,0 };
            const unsigned long dim1120Kuo2Init[]       =   {   1   ,   3   ,   3   ,   9   ,   15  ,   39  ,   81  ,   183 ,   213 ,   419 ,   125 ,   1943    ,   5621    ,   9073    ,0 };
            const unsigned long dim1121Kuo2Init[]       =   {   1   ,   3   ,   1   ,   7   ,   11  ,   41  ,   111 ,   169 ,   31  ,   807 ,   1277    ,   3293    ,   3233    ,   409 ,0 };
            const unsigned long dim1122Kuo2Init[]       =   {   1   ,   1   ,   3   ,   11  ,   31  ,   11  ,   39  ,   69  ,   429 ,   185 ,   1523    ,   469 ,   5453    ,   12453   ,0 };
            const unsigned long dim1123Kuo2Init[]       =   {   1   ,   1   ,   5   ,   13  ,   15  ,   59  ,   5   ,   179 ,   457 ,   803 ,   1503    ,   633 ,   7591    ,   8861    ,0 };
            const unsigned long dim1124Kuo2Init[]       =   {   1   ,   3   ,   5   ,   13  ,   19  ,   47  ,   15  ,   37  ,   271 ,   361 ,   1403    ,   2043    ,   6671    ,   4171    ,0 };
            const unsigned long dim1125Kuo2Init[]       =   {   1   ,   3   ,   3   ,   15  ,   3   ,   37  ,   107 ,   29  ,   505 ,   465 ,   955 ,   2431    ,   5899    ,   11363   ,0 };
            const unsigned long dim1126Kuo2Init[]       =   {   1   ,   1   ,   3   ,   9   ,   11  ,   49  ,   71  ,   191 ,   197 ,   365 ,   757 ,   309 ,   3881    ,   9427    ,0 };
            const unsigned long dim1127Kuo2Init[]       =   {   1   ,   1   ,   7   ,   13  ,   11  ,   9   ,   3   ,   109 ,   403 ,   403 ,   867 ,   325 ,   329 ,   13675   ,0 };
            const unsigned long dim1128Kuo2Init[]       =   {   1   ,   3   ,   3   ,   7   ,   13  ,   61  ,   33  ,   125 ,   431 ,   797 ,   565 ,   4065    ,   6375    ,   10343   ,0 };
            const unsigned long dim1129Kuo2Init[]       =   {   1   ,   1   ,   1   ,   1   ,   23  ,   59  ,   45  ,   103 ,   231 ,   829 ,   1197    ,   1547    ,   2319    ,   13541   ,0 };
            const unsigned long dim1130Kuo2Init[]       =   {   1   ,   1   ,   7   ,   13  ,   17  ,   59  ,   119 ,   51  ,   305 ,   153 ,   571 ,   1323    ,   6427    ,   12291   ,0 };
            const unsigned long dim1131Kuo2Init[]       =   {   1   ,   3   ,   3   ,   9   ,   21  ,   17  ,   27  ,   223 ,   311 ,   1015    ,   1925    ,   1497    ,   6407    ,   15517   ,0 };
            const unsigned long dim1132Kuo2Init[]       =   {   1   ,   3   ,   1   ,   15  ,   13  ,   63  ,   99  ,   163 ,   35  ,   583 ,   787 ,   2679    ,   4991    ,   5933    ,0 };
            const unsigned long dim1133Kuo2Init[]       =   {   1   ,   1   ,   3   ,   7   ,   31  ,   43  ,   77  ,   141 ,   153 ,   455 ,   745 ,   1377    ,   7363    ,   7301    ,0 };
            const unsigned long dim1134Kuo2Init[]       =   {   1   ,   1   ,   3   ,   15  ,   23  ,   49  ,   1   ,   149 ,   393 ,   483 ,   1577    ,   1885    ,   7483    ,   9769    ,0 };
            const unsigned long dim1135Kuo2Init[]       =   {   1   ,   1   ,   7   ,   11  ,   23  ,   5   ,   103 ,   143 ,   35  ,   899 ,   2001    ,   2643    ,   2533    ,   7827    ,0 };
            const unsigned long dim1136Kuo2Init[]       =   {   1   ,   3   ,   5   ,   5   ,   13  ,   49  ,   121 ,   113 ,   75  ,   271 ,   1111    ,   535 ,   1255    ,   6977    ,0 };
            const unsigned long dim1137Kuo2Init[]       =   {   1   ,   1   ,   1   ,   3   ,   25  ,   53  ,   11  ,   151 ,   107 ,   985 ,   1933    ,   2791    ,   4101    ,   14515   ,0 };
            const unsigned long dim1138Kuo2Init[]       =   {   1   ,   3   ,   1   ,   7   ,   29  ,   39  ,   73  ,   103 ,   139 ,   65  ,   573 ,   3909    ,   2645    ,   9603    ,0 };
            const unsigned long dim1139Kuo2Init[]       =   {   1   ,   3   ,   7   ,   9   ,   3   ,   31  ,   45  ,   79  ,   195 ,   749 ,   65  ,   4041    ,   3915    ,   3137    ,0 };
            const unsigned long dim1140Kuo2Init[]       =   {   1   ,   3   ,   3   ,   13  ,   5   ,   37  ,   87  ,   211 ,   61  ,   237 ,   1903    ,   755 ,   4077    ,   14791   ,0 };
            const unsigned long dim1141Kuo2Init[]       =   {   1   ,   1   ,   7   ,   5   ,   3   ,   63  ,   63  ,   45  ,   357 ,   887 ,   1511    ,   2503    ,   6669    ,   3163    ,0 };
            const unsigned long dim1142Kuo2Init[]       =   {   1   ,   3   ,   1   ,   5   ,   17  ,   29  ,   113 ,   117 ,   487 ,   369 ,   1337    ,   3129    ,   3517    ,   2107    ,0 };
            const unsigned long dim1143Kuo2Init[]       =   {   1   ,   1   ,   7   ,   1   ,   7   ,   13  ,   15  ,   203 ,   421 ,   267 ,   1049    ,   1645    ,   2047    ,   11733   ,0 };
            const unsigned long dim1144Kuo2Init[]       =   {   1   ,   1   ,   1   ,   5   ,   17  ,   27  ,   5   ,   101 ,   109 ,   651 ,   601 ,   673 ,   67  ,   8539    ,0 };
            const unsigned long dim1145Kuo2Init[]       =   {   1   ,   1   ,   5   ,   13  ,   9   ,   53  ,   85  ,   165 ,   197 ,   979 ,   925 ,   419 ,   4631    ,   15269   ,0 };
            const unsigned long dim1146Kuo2Init[]       =   {   1   ,   3   ,   7   ,   15  ,   5   ,   5   ,   61  ,   205 ,   257 ,   279 ,   81  ,   4059    ,   2375    ,   12821   ,0 };
            const unsigned long dim1147Kuo2Init[]       =   {   1   ,   3   ,   1   ,   11  ,   27  ,   41  ,   39  ,   41  ,   75  ,   1017    ,   529 ,   2709    ,   3385    ,   3879    ,0 };
            const unsigned long dim1148Kuo2Init[]       =   {   1   ,   3   ,   1   ,   13  ,   25  ,   63  ,   125 ,   137 ,   27  ,   839 ,   617 ,   2129    ,   711 ,   3955    ,0 };
            const unsigned long dim1149Kuo2Init[]       =   {   1   ,   3   ,   5   ,   5   ,   27  ,   19  ,   27  ,   51  ,   369 ,   593 ,   605 ,   2471    ,   6995    ,   5957    ,0 };
            const unsigned long dim1150Kuo2Init[]       =   {   1   ,   3   ,   3   ,   5   ,   9   ,   11  ,   63  ,   157 ,   447 ,   713 ,   1145    ,   2013    ,   3061    ,   13763   ,0 };
            const unsigned long dim1151Kuo2Init[]       =   {   1   ,   1   ,   5   ,   1   ,   15  ,   51  ,   49  ,   185 ,   337 ,   109 ,   1335    ,   841 ,   7005    ,   8969    ,0 };
            const unsigned long dim1152Kuo2Init[]       =   {   1   ,   1   ,   5   ,   15  ,   17  ,   25  ,   31  ,   135 ,   43  ,   469 ,   1895    ,   3369    ,   6981    ,   8515    ,0 };
            const unsigned long dim1153Kuo2Init[]       =   {   1   ,   1   ,   5   ,   9   ,   21  ,   11  ,   117 ,   79  ,   39  ,   63  ,   1355    ,   2309    ,   4409    ,   6463    ,0 };
            const unsigned long dim1154Kuo2Init[]       =   {   1   ,   3   ,   3   ,   15  ,   11  ,   55  ,   55  ,   75  ,   251 ,   209 ,   1715    ,   697 ,   2199    ,   2155    ,0 };
            const unsigned long dim1155Kuo2Init[]       =   {   1   ,   3   ,   5   ,   13  ,   31  ,   29  ,   55  ,   23  ,   391 ,   1011    ,   1701    ,   787 ,   8165    ,   1315    ,0 };
            const unsigned long dim1156Kuo2Init[]       =   {   1   ,   1   ,   7   ,   7   ,   23  ,   55  ,   125 ,   15  ,   77  ,   107 ,   2043    ,   1501    ,   727 ,   13693   ,0 };
            const unsigned long dim1157Kuo2Init[]       =   {   1   ,   1   ,   7   ,   7   ,   27  ,   37  ,   47  ,   247 ,   227 ,   595 ,   175 ,   3459    ,   395 ,   9903    ,0 };
            const unsigned long dim1158Kuo2Init[]       =   {   1   ,   3   ,   5   ,   9   ,   7   ,   7   ,   61  ,   81  ,   405 ,   117 ,   51  ,   3715    ,   2757    ,   2797    ,0 };
            const unsigned long dim1159Kuo2Init[]       =   {   1   ,   1   ,   1   ,   15  ,   15  ,   25  ,   9   ,   109 ,   497 ,   395 ,   777 ,   1187    ,   7451    ,   15959   ,0 };
            const unsigned long dim1160Kuo2Init[]       =   {   1   ,   3   ,   3   ,   7   ,   7   ,   23  ,   69  ,   27  ,   97  ,   145 ,   1165    ,   993 ,   235 ,   1785    ,0 };
            const unsigned long dim1161Kuo2Init[]       =   {   1   ,   1   ,   3   ,   15  ,   17  ,   49  ,   73  ,   27  ,   223 ,   83  ,   295 ,   1425    ,   4817    ,   7325    ,0 };
            const unsigned long dim1162Kuo2Init[]       =   {   1   ,   3   ,   7   ,   5   ,   25  ,   13  ,   69  ,   119 ,   483 ,   635 ,   1105    ,   2547    ,   457 ,   5607    ,0 };
            const unsigned long dim1163Kuo2Init[]       =   {   1   ,   1   ,   7   ,   9   ,   13  ,   29  ,   69  ,   7   ,   223 ,   95  ,   1255    ,   459 ,   4633    ,   12609   ,0 };
            const unsigned long dim1164Kuo2Init[]       =   {   1   ,   1   ,   1   ,   1   ,   15  ,   57  ,   105 ,   161 ,   429 ,   615 ,   201 ,   2601    ,   3169    ,   3163    ,0 };
            const unsigned long dim1165Kuo2Init[]       =   {   1   ,   3   ,   3   ,   1   ,   5   ,   5   ,   127 ,   129 ,   457 ,   59  ,   459 ,   3731    ,   2069    ,   9147    ,0 };
            const unsigned long dim1166Kuo2Init[]       =   {   1   ,   1   ,   1   ,   7   ,   25  ,   45  ,   101 ,   111 ,   119 ,   739 ,   157 ,   3967    ,   6427    ,   2845    ,0 };
            const unsigned long dim1167Kuo2Init[]       =   {   1   ,   3   ,   1   ,   9   ,   15  ,   15  ,   105 ,   147 ,   119 ,   815 ,   1263    ,   2201    ,   5413    ,   15255   ,0 };
            const unsigned long dim1168Kuo2Init[]       =   {   1   ,   3   ,   1   ,   5   ,   15  ,   47  ,   87  ,   131 ,   451 ,   805 ,   1043    ,   2917    ,   5495    ,   5103    ,0 };
            const unsigned long dim1169Kuo2Init[]       =   {   1   ,   1   ,   3   ,   11  ,   29  ,   35  ,   33  ,   135 ,   509 ,   425 ,   723 ,   1947    ,   7237    ,   8941    ,0 };
            const unsigned long dim1170Kuo2Init[]       =   {   1   ,   1   ,   3   ,   7   ,   5   ,   39  ,   83  ,   83  ,   353 ,   685 ,   195 ,   385 ,   2431    ,   8753    ,0 };
            const unsigned long dim1171Kuo2Init[]       =   {   1   ,   3   ,   3   ,   7   ,   25  ,   45  ,   3   ,   129 ,   413 ,   1005    ,   625 ,   3503    ,   5473    ,   3577    ,0 };
            const unsigned long dim1172Kuo2Init[]       =   {   1   ,   1   ,   3   ,   13  ,   7   ,   37  ,   29  ,   163 ,   397 ,   671 ,   1043    ,   3581    ,   465 ,   9521    ,0 };
            const unsigned long dim1173Kuo2Init[]       =   {   1   ,   1   ,   3   ,   5   ,   31  ,   21  ,   111 ,   95  ,   385 ,   787 ,   1205    ,   795 ,   4779    ,   11421   ,0 };
            const unsigned long dim1174Kuo2Init[]       =   {   1   ,   1   ,   1   ,   3   ,   25  ,   15  ,   51  ,   149 ,   423 ,   907 ,   863 ,   1535    ,   5809    ,   1069    ,0 };
            const unsigned long dim1175Kuo2Init[]       =   {   1   ,   3   ,   5   ,   5   ,   19  ,   57  ,   51  ,   49  ,   39  ,   349 ,   1453    ,   3091    ,   4861    ,   8771    ,0 };
            const unsigned long dim1176Kuo2Init[]       =   {   1   ,   1   ,   5   ,   15  ,   1   ,   45  ,   93  ,   85  ,   511 ,   607 ,   793 ,   1023    ,   1525    ,   535 ,0 };
            const unsigned long dim1177Kuo2Init[]       =   {   1   ,   3   ,   3   ,   11  ,   7   ,   53  ,   83  ,   139 ,   173 ,   947 ,   1381    ,   433 ,   5821    ,   4173    ,0 };
            const unsigned long dim1178Kuo2Init[]       =   {   1   ,   3   ,   5   ,   9   ,   27  ,   17  ,   17  ,   237 ,   427 ,   295 ,   929 ,   1539    ,   2373    ,   6671    ,0 };
            const unsigned long dim1179Kuo2Init[]       =   {   1   ,   1   ,   1   ,   1   ,   9   ,   53  ,   87  ,   129 ,   443 ,   729 ,   1067    ,   359 ,   789 ,   7637    ,0 };
            const unsigned long dim1180Kuo2Init[]       =   {   1   ,   1   ,   1   ,   7   ,   7   ,   23  ,   71  ,   115 ,   343 ,   207 ,   889 ,   3561    ,   475 ,   3437    ,0 };
            const unsigned long dim1181Kuo2Init[]       =   {   1   ,   3   ,   5   ,   9   ,   17  ,   31  ,   85  ,   135 ,   155 ,   269 ,   575 ,   537 ,   5305    ,   979 ,0 };
            const unsigned long dim1182Kuo2Init[]       =   {   1   ,   3   ,   1   ,   5   ,   19  ,   41  ,   61  ,   103 ,   49  ,   93  ,   733 ,   1791    ,   3577    ,   1763    ,0 };
            const unsigned long dim1183Kuo2Init[]       =   {   1   ,   1   ,   3   ,   3   ,   27  ,   23  ,   3   ,   143 ,   341 ,   847 ,   1859    ,   995 ,   5277    ,   5025    ,0 };
            const unsigned long dim1184Kuo2Init[]       =   {   1   ,   3   ,   5   ,   13  ,   7   ,   29  ,   37  ,   201 ,   359 ,   255 ,   1543    ,   3735    ,   1217    ,   7251    ,0 };
            const unsigned long dim1185Kuo2Init[]       =   {   1   ,   1   ,   7   ,   9   ,   27  ,   45  ,   15  ,   187 ,   181 ,   571 ,   1345    ,   1473    ,   519 ,   7499    ,0 };
            const unsigned long dim1186Kuo2Init[]       =   {   1   ,   1   ,   7   ,   7   ,   11  ,   11  ,   75  ,   179 ,   455 ,   495 ,   283 ,   2995    ,   1515    ,   5497    ,0 };
            const unsigned long dim1187Kuo2Init[]       =   {   1   ,   1   ,   7   ,   3   ,   11  ,   19  ,   91  ,   165 ,   175 ,   967 ,   1719    ,   739 ,   103 ,   15259   ,0 };
            const unsigned long dim1188Kuo2Init[]       =   {   1   ,   3   ,   3   ,   9   ,   7   ,   5   ,   121 ,   65  ,   83  ,   687 ,   837 ,   2303    ,   1953    ,   15419   ,0 };
            const unsigned long dim1189Kuo2Init[]       =   {   1   ,   3   ,   1   ,   13  ,   31  ,   39  ,   43  ,   69  ,   129 ,   815 ,   1419    ,   3179    ,   595 ,   14471   ,0 };
            const unsigned long dim1190Kuo2Init[]       =   {   1   ,   1   ,   5   ,   9   ,   9   ,   33  ,   75  ,   141 ,   271 ,   259 ,   91  ,   553 ,   7521    ,   9591    ,0 };
            const unsigned long dim1191Kuo2Init[]       =   {   1   ,   1   ,   5   ,   1   ,   27  ,   55  ,   125 ,   41  ,   103 ,   161 ,   9   ,   1339    ,   3205    ,   973 ,0 };
            const unsigned long dim1192Kuo2Init[]       =   {   1   ,   1   ,   3   ,   3   ,   13  ,   15  ,   93  ,   179 ,   435 ,   667 ,   1365    ,   169 ,   4685    ,   16175   ,0 };
            const unsigned long dim1193Kuo2Init[]       =   {   1   ,   1   ,   5   ,   9   ,   17  ,   55  ,   45  ,   237 ,   291 ,   837 ,   1019    ,   2983    ,   1679    ,   13583   ,0 };
            const unsigned long dim1194Kuo2Init[]       =   {   1   ,   3   ,   1   ,   3   ,   3   ,   37  ,   37  ,   239 ,   265 ,   117 ,   1239    ,   529 ,   4855    ,   11423   ,0 };
            const unsigned long dim1195Kuo2Init[]       =   {   1   ,   3   ,   5   ,   3   ,   11  ,   11  ,   121 ,   81  ,   55  ,   35  ,   1551    ,   775 ,   6973    ,   13525   ,0 };
            const unsigned long dim1196Kuo2Init[]       =   {   1   ,   3   ,   3   ,   11  ,   15  ,   19  ,   49  ,   117 ,   269 ,   147 ,   1891    ,   2975    ,   1779    ,   9327    ,0 };
            const unsigned long dim1197Kuo2Init[]       =   {   1   ,   1   ,   1   ,   7   ,   9   ,   39  ,   111 ,   131 ,   179 ,   629 ,   1807    ,   3327    ,   1535    ,   2845    ,0 };
            const unsigned long dim1198Kuo2Init[]       =   {   1   ,   3   ,   7   ,   13  ,   11  ,   59  ,   125 ,   41  ,   13  ,   637 ,   1679    ,   487 ,   4655    ,   8045    ,0 };
            const unsigned long dim1199Kuo2Init[]       =   {   1   ,   1   ,   3   ,   7   ,   17  ,   15  ,   63  ,   105 ,   271 ,   205 ,   283 ,   3427    ,   1921    ,   12729   ,0 };
            const unsigned long dim1200Kuo2Init[]       =   {   1   ,   1   ,   3   ,   5   ,   13  ,   57  ,   23  ,   141 ,   261 ,   829 ,   1149    ,   945 ,   4191    ,   13277   ,0 };
            const unsigned long dim1201Kuo2Init[]       =   {   1   ,   1   ,   5   ,   7   ,   13  ,   59  ,   79  ,   221 ,   473 ,   803 ,   2027    ,   437 ,   7823    ,   14225   ,0 };
            const unsigned long dim1202Kuo2Init[]       =   {   1   ,   3   ,   3   ,   9   ,   15  ,   21  ,   53  ,   69  ,   13  ,   479 ,   193 ,   1031    ,   4911    ,   12011   ,0 };
            const unsigned long dim1203Kuo2Init[]       =   {   1   ,   1   ,   5   ,   7   ,   25  ,   7   ,   99  ,   81  ,   21  ,   247 ,   1019    ,   2537    ,   3647    ,   8211    ,0 };
            const unsigned long dim1204Kuo2Init[]       =   {   1   ,   3   ,   7   ,   15  ,   27  ,   47  ,   105 ,   141 ,   49  ,   65  ,   525 ,   1717    ,   1339    ,   9079    ,0 };
            const unsigned long dim1205Kuo2Init[]       =   {   1   ,   1   ,   7   ,   3   ,   27  ,   1   ,   31  ,   135 ,   175 ,   997 ,   1517    ,   1139    ,   7975    ,   14941   ,0 };
            const unsigned long dim1206Kuo2Init[]       =   {   1   ,   1   ,   1   ,   5   ,   9   ,   5   ,   77  ,   175 ,   333 ,   975 ,   873 ,   523 ,   2967    ,   9463    ,0 };
            const unsigned long dim1207Kuo2Init[]       =   {   1   ,   3   ,   5   ,   9   ,   5   ,   61  ,   17  ,   141 ,   405 ,   591 ,   711 ,   1077    ,   2875    ,   8827    ,0 };
            const unsigned long dim1208Kuo2Init[]       =   {   1   ,   3   ,   1   ,   11  ,   9   ,   13  ,   29  ,   9   ,   443 ,   717 ,   2019    ,   1031    ,   6459    ,   6673    ,0 };
            const unsigned long dim1209Kuo2Init[]       =   {   1   ,   1   ,   3   ,   5   ,   7   ,   51  ,   45  ,   163 ,   113 ,   705 ,   1411    ,   3861    ,   527 ,   6119    ,0 };
            const unsigned long dim1210Kuo2Init[]       =   {   1   ,   1   ,   1   ,   5   ,   9   ,   43  ,   57  ,   79  ,   221 ,   131 ,   1543    ,   2243    ,   5277    ,   6249    ,0 };
            const unsigned long dim1211Kuo2Init[]       =   {   1   ,   1   ,   1   ,   7   ,   15  ,   63  ,   89  ,   11  ,   211 ,   245 ,   89  ,   219 ,   241 ,   545 ,0 };
            const unsigned long dim1212Kuo2Init[]       =   {   1   ,   3   ,   3   ,   7   ,   31  ,   21  ,   97  ,   43  ,   457 ,   941 ,   363 ,   3275    ,   4731    ,   3167    ,0 };
            const unsigned long dim1213Kuo2Init[]       =   {   1   ,   1   ,   7   ,   9   ,   31  ,   45  ,   15  ,   119 ,   283 ,   645 ,   1239    ,   3835    ,   5317    ,   4643    ,0 };
            const unsigned long dim1214Kuo2Init[]       =   {   1   ,   1   ,   5   ,   3   ,   23  ,   31  ,   49  ,   67  ,   487 ,   935 ,   1899    ,   2545    ,   8135    ,   12609   ,0 };
            const unsigned long dim1215Kuo2Init[]       =   {   1   ,   3   ,   7   ,   9   ,   21  ,   63  ,   15  ,   233 ,   463 ,   763 ,   327 ,   2585    ,   4399    ,   14643   ,0 };
            const unsigned long dim1216Kuo2Init[]       =   {   1   ,   1   ,   7   ,   9   ,   13  ,   5   ,   95  ,   11  ,   239 ,   507 ,   955 ,   569 ,   6443    ,   3009    ,0 };
            const unsigned long dim1217Kuo2Init[]       =   {   1   ,   3   ,   5   ,   9   ,   31  ,   27  ,   99  ,   251 ,   487 ,   383 ,   1117    ,   2231    ,   6541    ,   10409   ,0 };
            const unsigned long dim1218Kuo2Init[]       =   {   1   ,   3   ,   3   ,   15  ,   23  ,   37  ,   81  ,   149 ,   127 ,   341 ,   1749    ,   207 ,   2257    ,   2659    ,0 };
            const unsigned long dim1219Kuo2Init[]       =   {   1   ,   1   ,   7   ,   9   ,   15  ,   7   ,   59  ,   63  ,   187 ,   719 ,   1295    ,   3509    ,   2293    ,   4959    ,0 };
            const unsigned long dim1220Kuo2Init[]       =   {   1   ,   1   ,   5   ,   13  ,   3   ,   49  ,   27  ,   17  ,   261 ,   463 ,   1719    ,   793 ,   4365    ,   4041    ,0 };
            const unsigned long dim1221Kuo2Init[]       =   {   1   ,   3   ,   3   ,   7   ,   7   ,   37  ,   19  ,   133 ,   197 ,   789 ,   383 ,   1005    ,   7353    ,   2431    ,0 };
            const unsigned long dim1222Kuo2Init[]       =   {   1   ,   1   ,   1   ,   13  ,   21  ,   11  ,   27  ,   247 ,   309 ,   103 ,   1707    ,   2555    ,   2731    ,   11755   ,0 };
            const unsigned long dim1223Kuo2Init[]       =   {   1   ,   3   ,   1   ,   15  ,   13  ,   45  ,   119 ,   233 ,   387 ,   781 ,   851 ,   3207    ,   5251    ,   9741    ,0 };
            const unsigned long dim1224Kuo2Init[]       =   {   1   ,   3   ,   7   ,   15  ,   3   ,   17  ,   13  ,   217 ,   245 ,   267 ,   55  ,   3065    ,   3733    ,   2577    ,0 };
            const unsigned long dim1225Kuo2Init[]       =   {   1   ,   1   ,   7   ,   13  ,   11  ,   25  ,   7   ,   219 ,   209 ,   295 ,   1725    ,   359 ,   5489    ,   1057    ,0 };
            const unsigned long dim1226Kuo2Init[]       =   {   1   ,   3   ,   1   ,   3   ,   7   ,   1   ,   93  ,   33  ,   485 ,   977 ,   1113    ,   3049    ,   7015    ,   14275   ,0 };
            const unsigned long dim1227Kuo2Init[]       =   {   1   ,   3   ,   3   ,   13  ,   15  ,   9   ,   115 ,   113 ,   221 ,   1   ,   875 ,   1485    ,   3523    ,   6817    ,0 };
            const unsigned long dim1228Kuo2Init[]       =   {   1   ,   1   ,   5   ,   15  ,   3   ,   41  ,   113 ,   5   ,   7   ,   1023    ,   349 ,   1901    ,   1223    ,   11355   ,0 };
            const unsigned long dim1229Kuo2Init[]       =   {   1   ,   3   ,   7   ,   13  ,   27  ,   59  ,   113 ,   123 ,   81  ,   795 ,   803 ,   955 ,   7101    ,   7715    ,0 };
            const unsigned long dim1230Kuo2Init[]       =   {   1   ,   1   ,   1   ,   15  ,   15  ,   19  ,   123 ,   183 ,   341 ,   9   ,   943 ,   549 ,   1261    ,   8305    ,0 };
            const unsigned long dim1231Kuo2Init[]       =   {   1   ,   1   ,   3   ,   5   ,   25  ,   45  ,   109 ,   239 ,   399 ,   405 ,   1859    ,   3721    ,   2023    ,   12543   ,0 };
            const unsigned long dim1232Kuo2Init[]       =   {   1   ,   3   ,   1   ,   11  ,   11  ,   51  ,   119 ,   83  ,   189 ,   119 ,   51  ,   3601    ,   1589    ,   3821    ,0 };
            const unsigned long dim1233Kuo2Init[]       =   {   1   ,   3   ,   3   ,   3   ,   9   ,   41  ,   13  ,   91  ,   329 ,   625 ,   1837    ,   2539    ,   5603    ,   2135    ,0 };
            const unsigned long dim1234Kuo2Init[]       =   {   1   ,   3   ,   5   ,   7   ,   5   ,   59  ,   125 ,   133 ,   321 ,   759 ,   237 ,   3299    ,   4663    ,   15767   ,0 };
            const unsigned long dim1235Kuo2Init[]       =   {   1   ,   1   ,   7   ,   7   ,   29  ,   3   ,   45  ,   37  ,   69  ,   437 ,   1321    ,   3545    ,   373 ,   3349    ,0 };
            const unsigned long dim1236Kuo2Init[]       =   {   1   ,   1   ,   1   ,   13  ,   13  ,   47  ,   87  ,   33  ,   93  ,   639 ,   673 ,   1675    ,   5997    ,   9049    ,0 };
            const unsigned long dim1237Kuo2Init[]       =   {   1   ,   3   ,   1   ,   3   ,   31  ,   27  ,   33  ,   175 ,   71  ,   509 ,   229 ,   2959    ,   7929    ,   7551    ,0 };
            const unsigned long dim1238Kuo2Init[]       =   {   1   ,   1   ,   3   ,   7   ,   17  ,   41  ,   101 ,   119 ,   51  ,   297 ,   595 ,   41  ,   7919    ,   6173    ,0 };
            const unsigned long dim1239Kuo2Init[]       =   {   1   ,   3   ,   5   ,   11  ,   11  ,   63  ,   17  ,   177 ,   349 ,   327 ,   531 ,   783 ,   4025    ,   3367    ,0 };
            const unsigned long dim1240Kuo2Init[]       =   {   1   ,   3   ,   1   ,   13  ,   11  ,   31  ,   97  ,   137 ,   117 ,   981 ,   3   ,   2873    ,   6399    ,   5055    ,0 };
            const unsigned long dim1241Kuo2Init[]       =   {   1   ,   3   ,   7   ,   11  ,   9   ,   63  ,   53  ,   237 ,   255 ,   161 ,   1709    ,   3591    ,   2569    ,   8687    ,0 };
            const unsigned long dim1242Kuo2Init[]       =   {   1   ,   3   ,   3   ,   11  ,   23  ,   9   ,   97  ,   143 ,   403 ,   597 ,   1651    ,   933 ,   1351    ,   3675    ,0 };
            const unsigned long dim1243Kuo2Init[]       =   {   1   ,   1   ,   1   ,   9   ,   23  ,   11  ,   93  ,   103 ,   245 ,   797 ,   723 ,   4029    ,   4313    ,   14217   ,0 };
            const unsigned long dim1244Kuo2Init[]       =   {   1   ,   1   ,   1   ,   3   ,   13  ,   51  ,   77  ,   237 ,   109 ,   933 ,   449 ,   429 ,   6991    ,   13475   ,0 };
            const unsigned long dim1245Kuo2Init[]       =   {   1   ,   3   ,   1   ,   1   ,   29  ,   63  ,   65  ,   63  ,   129 ,   693 ,   1441    ,   3877    ,   3015    ,   7817    ,0 };
            const unsigned long dim1246Kuo2Init[]       =   {   1   ,   1   ,   7   ,   3   ,   7   ,   57  ,   63  ,   129 ,   15  ,   257 ,   739 ,   3395    ,   1473    ,   12735   ,0 };
            const unsigned long dim1247Kuo2Init[]       =   {   1   ,   1   ,   1   ,   9   ,   23  ,   21  ,   123 ,   81  ,   297 ,   899 ,   1719    ,   3523    ,   4895    ,   12503   ,0 };
            const unsigned long dim1248Kuo2Init[]       =   {   1   ,   1   ,   3   ,   1   ,   9   ,   3   ,   61  ,   7   ,   251 ,   1003    ,   289 ,   1321    ,   7041    ,   6811    ,0 };
            const unsigned long dim1249Kuo2Init[]       =   {   1   ,   3   ,   5   ,   9   ,   3   ,   9   ,   87  ,   13  ,   363 ,   713 ,   673 ,   717 ,   1449    ,   6869    ,0 };
            const unsigned long dim1250Kuo2Init[]       =   {   1   ,   1   ,   3   ,   9   ,   25  ,   53  ,   115 ,   111 ,   499 ,   425 ,   1037    ,   3751    ,   5767    ,   10047   ,0 };
            const unsigned long dim1251Kuo2Init[]       =   {   1   ,   3   ,   7   ,   11  ,   23  ,   43  ,   75  ,   101 ,   105 ,   115 ,   887 ,   3747    ,   149 ,   5513    ,0 };
            const unsigned long dim1252Kuo2Init[]       =   {   1   ,   3   ,   1   ,   9   ,   7   ,   43  ,   125 ,   63  ,   3   ,   365 ,   1345    ,   821 ,   731 ,   14007   ,0 };
            const unsigned long dim1253Kuo2Init[]       =   {   1   ,   3   ,   1   ,   7   ,   29  ,   49  ,   55  ,   37  ,   343 ,   481 ,   1319    ,   1167    ,   4835    ,   10549   ,0 };
            const unsigned long dim1254Kuo2Init[]       =   {   1   ,   1   ,   7   ,   13  ,   11  ,   33  ,   123 ,   91  ,   301 ,   765 ,   1037    ,   2935    ,   1213    ,   10141   ,0 };
            const unsigned long dim1255Kuo2Init[]       =   {   1   ,   1   ,   7   ,   1   ,   21  ,   55  ,   57  ,   79  ,   187 ,   977 ,   297 ,   63  ,   3431    ,   10597   ,0 };
            const unsigned long dim1256Kuo2Init[]       =   {   1   ,   3   ,   5   ,   9   ,   5   ,   25  ,   33  ,   25  ,   477 ,   105 ,   1757    ,   2051    ,   1263    ,   15421   ,0 };
            const unsigned long dim1257Kuo2Init[]       =   {   1   ,   3   ,   5   ,   11  ,   5   ,   39  ,   63  ,   209 ,   223 ,   497 ,   785 ,   2127    ,   4293    ,   7583    ,0 };
            const unsigned long dim1258Kuo2Init[]       =   {   1   ,   1   ,   5   ,   3   ,   9   ,   23  ,   107 ,   125 ,   319 ,   143 ,   947 ,   1135    ,   4391    ,   6297    ,0 };
            const unsigned long dim1259Kuo2Init[]       =   {   1   ,   3   ,   7   ,   15  ,   19  ,   21  ,   103 ,   161 ,   5   ,   739 ,   1537    ,   2697    ,   1827    ,   8287    ,0 };
            const unsigned long dim1260Kuo2Init[]       =   {   1   ,   1   ,   7   ,   3   ,   29  ,   37  ,   117 ,   199 ,   35  ,   683 ,   1539    ,   2915    ,   8129    ,   14785   ,0 };
            const unsigned long dim1261Kuo2Init[]       =   {   1   ,   3   ,   1   ,   3   ,   7   ,   35  ,   39  ,   181 ,   111 ,   165 ,   1919    ,   1583    ,   4219    ,   179 ,0 };
            const unsigned long dim1262Kuo2Init[]       =   {   1   ,   3   ,   3   ,   11  ,   15  ,   5   ,   13  ,   123 ,   93  ,   615 ,   1811    ,   3235    ,   4533    ,   2423    ,0 };
            const unsigned long dim1263Kuo2Init[]       =   {   1   ,   1   ,   1   ,   15  ,   25  ,   17  ,   21  ,   37  ,   5   ,   413 ,   1681    ,   2297    ,   5571    ,   6037    ,0 };
            const unsigned long dim1264Kuo2Init[]       =   {   1   ,   1   ,   1   ,   9   ,   27  ,   21  ,   7   ,   139 ,   235 ,   163 ,   1809    ,   783 ,   507 ,   8867    ,0 };
            const unsigned long dim1265Kuo2Init[]       =   {   1   ,   3   ,   7   ,   9   ,   23  ,   57  ,   1   ,   55  ,   311 ,   687 ,   1987    ,   463 ,   5639    ,   4795    ,0 };
            const unsigned long dim1266Kuo2Init[]       =   {   1   ,   1   ,   7   ,   9   ,   11  ,   55  ,   31  ,   77  ,   277 ,   447 ,   799 ,   993 ,   489 ,   3539    ,0 };
            const unsigned long dim1267Kuo2Init[]       =   {   1   ,   1   ,   1   ,   5   ,   27  ,   19  ,   23  ,   207 ,   11  ,   81  ,   1539    ,   1051    ,   1155    ,   11475   ,0 };
            const unsigned long dim1268Kuo2Init[]       =   {   1   ,   1   ,   5   ,   1   ,   23  ,   63  ,   23  ,   247 ,   261 ,   789 ,   1305    ,   3003    ,   1033    ,   1883    ,0 };
            const unsigned long dim1269Kuo2Init[]       =   {   1   ,   1   ,   5   ,   15  ,   1   ,   35  ,   1   ,   147 ,   51  ,   513 ,   35  ,   3285    ,   7461    ,   13733   ,0 };
            const unsigned long dim1270Kuo2Init[]       =   {   1   ,   1   ,   1   ,   5   ,   29  ,   25  ,   107 ,   91  ,   355 ,   213 ,   203 ,   2371    ,   1383    ,   14285   ,0 };
            const unsigned long dim1271Kuo2Init[]       =   {   1   ,   3   ,   7   ,   13  ,   11  ,   35  ,   109 ,   171 ,   229 ,   209 ,   375 ,   2911    ,   1943    ,   7037    ,0 };
            const unsigned long dim1272Kuo2Init[]       =   {   1   ,   1   ,   5   ,   11  ,   21  ,   39  ,   113 ,   5   ,   167 ,   769 ,   2047    ,   3083    ,   7421    ,   14085   ,0 };
            const unsigned long dim1273Kuo2Init[]       =   {   1   ,   3   ,   7   ,   13  ,   13  ,   27  ,   83  ,   67  ,   171 ,   671 ,   1675    ,   2473    ,   3241    ,   15283   ,0 };
            const unsigned long dim1274Kuo2Init[]       =   {   1   ,   1   ,   1   ,   7   ,   5   ,   27  ,   23  ,   171 ,   443 ,   243 ,   217 ,   3349    ,   1919    ,   2345    ,0 };
            const unsigned long dim1275Kuo2Init[]       =   {   1   ,   1   ,   1   ,   11  ,   7   ,   1   ,   41  ,   25  ,   47  ,   835 ,   1071    ,   3091    ,   6133    ,   7249    ,0 };
            const unsigned long dim1276Kuo2Init[]       =   {   1   ,   1   ,   7   ,   15  ,   27  ,   41  ,   3   ,   153 ,   479 ,   447 ,   1457    ,   2267    ,   3781    ,   9851    ,0 };
            const unsigned long dim1277Kuo2Init[]       =   {   1   ,   1   ,   7   ,   7   ,   25  ,   1   ,   125 ,   129 ,   231 ,   505 ,   1709    ,   3117    ,   6069    ,   3103    ,0 };
            const unsigned long dim1278Kuo2Init[]       =   {   1   ,   3   ,   3   ,   13  ,   9   ,   53  ,   73  ,   93  ,   185 ,   1007    ,   1883    ,   3087    ,   7453    ,   13335   ,0 };
            const unsigned long dim1279Kuo2Init[]       =   {   1   ,   1   ,   1   ,   5   ,   7   ,   51  ,   65  ,   189 ,   397 ,   73  ,   377 ,   2593    ,   6703    ,   16145   ,0 };
            const unsigned long dim1280Kuo2Init[]       =   {   1   ,   1   ,   7   ,   3   ,   17  ,   55  ,   93  ,   251 ,   29  ,   937 ,   223 ,   721 ,   3275    ,   11881   ,0 };
            const unsigned long dim1281Kuo2Init[]       =   {   1   ,   3   ,   7   ,   1   ,   1   ,   45  ,   99  ,   117 ,   279 ,   87  ,   2015    ,   3681    ,   5625    ,   5191    ,0 };
            const unsigned long dim1282Kuo2Init[]       =   {   1   ,   3   ,   7   ,   9   ,   7   ,   63  ,   23  ,   83  ,   277 ,   959 ,   693 ,   3945    ,   4551    ,   13711   ,0 };
            const unsigned long dim1283Kuo2Init[]       =   {   1   ,   3   ,   1   ,   1   ,   23  ,   57  ,   127 ,   7   ,   323 ,   545 ,   1909    ,   2399    ,   4483    ,   12165   ,0 };
            const unsigned long dim1284Kuo2Init[]       =   {   1   ,   3   ,   1   ,   7   ,   23  ,   37  ,   39  ,   213 ,   45  ,   949 ,   661 ,   397 ,   189 ,   2835    ,0 };
            const unsigned long dim1285Kuo2Init[]       =   {   1   ,   1   ,   1   ,   11  ,   17  ,   11  ,   97  ,   141 ,   389 ,   137 ,   907 ,   3939    ,   1205    ,   9751    ,0 };
            const unsigned long dim1286Kuo2Init[]       =   {   1   ,   1   ,   5   ,   9   ,   25  ,   45  ,   53  ,   101 ,   43  ,   503 ,   777 ,   1327    ,   8127    ,   1977    ,0 };
            const unsigned long dim1287Kuo2Init[]       =   {   1   ,   3   ,   1   ,   5   ,   19  ,   15  ,   71  ,   193 ,   443 ,   817 ,   1507    ,   291 ,   3781    ,   5841    ,0 };
            const unsigned long dim1288Kuo2Init[]       =   {   1   ,   3   ,   3   ,   1   ,   25  ,   41  ,   121 ,   17  ,   471 ,   357 ,   1193    ,   3079    ,   2025    ,   773 ,0 };
            const unsigned long dim1289Kuo2Init[]       =   {   1   ,   1   ,   3   ,   9   ,   5   ,   31  ,   123 ,   35  ,   103 ,   439 ,   1199    ,   1887    ,   6923    ,   9439    ,0 };
            const unsigned long dim1290Kuo2Init[]       =   {   1   ,   3   ,   1   ,   5   ,   7   ,   35  ,   1   ,   151 ,   467 ,   573 ,   517 ,   3681    ,   4563    ,   3143    ,0 };
            const unsigned long dim1291Kuo2Init[]       =   {   1   ,   3   ,   1   ,   7   ,   19  ,   61  ,   43  ,   57  ,   217 ,   113 ,   1045    ,   3773    ,   1197    ,   14627   ,0 };
            const unsigned long dim1292Kuo2Init[]       =   {   1   ,   3   ,   7   ,   11  ,   7   ,   51  ,   71  ,   31  ,   71  ,   143 ,   319 ,   2979    ,   849 ,   14403   ,0 };
            const unsigned long dim1293Kuo2Init[]       =   {   1   ,   1   ,   1   ,   15  ,   17  ,   25  ,   119 ,   253 ,   13  ,   959 ,   1217    ,   261 ,   2853    ,   10449   ,0 };
            const unsigned long dim1294Kuo2Init[]       =   {   1   ,   3   ,   7   ,   15  ,   23  ,   19  ,   71  ,   205 ,   63  ,   915 ,   317 ,   1637    ,   1963    ,   10047   ,0 };
            const unsigned long dim1295Kuo2Init[]       =   {   1   ,   1   ,   5   ,   9   ,   21  ,   57  ,   119 ,   223 ,   169 ,   87  ,   593 ,   2871    ,   1201    ,   4531    ,0 };
            const unsigned long dim1296Kuo2Init[]       =   {   1   ,   3   ,   5   ,   3   ,   7   ,   11  ,   25  ,   129 ,   469 ,   199 ,   323 ,   4021    ,   2983    ,   1459    ,0 };
            const unsigned long dim1297Kuo2Init[]       =   {   1   ,   3   ,   1   ,   1   ,   7   ,   39  ,   41  ,   75  ,   75  ,   113 ,   985 ,   3353    ,   5393    ,   7055    ,0 };
            const unsigned long dim1298Kuo2Init[]       =   {   1   ,   3   ,   3   ,   11  ,   19  ,   55  ,   37  ,   235 ,   221 ,   197 ,   1041    ,   3447    ,   8045    ,   801 ,0 };
            const unsigned long dim1299Kuo2Init[]       =   {   1   ,   1   ,   7   ,   7   ,   9   ,   19  ,   3   ,   221 ,   265 ,   963 ,   1815    ,   469 ,   2929    ,   10711   ,0 };
            const unsigned long dim1300Kuo2Init[]       =   {   1   ,   3   ,   1   ,   5   ,   27  ,   11  ,   59  ,   57  ,   91  ,   235 ,   1465    ,   825 ,   6429    ,   15537   ,0 };
            const unsigned long dim1301Kuo2Init[]       =   {   1   ,   1   ,   1   ,   1   ,   19  ,   21  ,   47  ,   235 ,   223 ,   761 ,   173 ,   3793    ,   2197    ,   10919   ,0 };
            const unsigned long dim1302Kuo2Init[]       =   {   1   ,   3   ,   3   ,   9   ,   11  ,   61  ,   19  ,   121 ,   51  ,   617 ,   441 ,   1889    ,   6225    ,   3317    ,0 };
            const unsigned long dim1303Kuo2Init[]       =   {   1   ,   3   ,   1   ,   7   ,   25  ,   27  ,   65  ,   163 ,   165 ,   7   ,   1303    ,   3275    ,   173 ,   6923    ,0 };
            const unsigned long dim1304Kuo2Init[]       =   {   1   ,   3   ,   1   ,   13  ,   23  ,   39  ,   35  ,   105 ,   267 ,   679 ,   1675    ,   2837    ,   1213    ,   14985   ,0 };
            const unsigned long dim1305Kuo2Init[]       =   {   1   ,   1   ,   5   ,   1   ,   21  ,   39  ,   1   ,   229 ,   101 ,   421 ,   513 ,   571 ,   7195    ,   9975    ,0 };
            const unsigned long dim1306Kuo2Init[]       =   {   1   ,   1   ,   1   ,   1   ,   13  ,   23  ,   97  ,   157 ,   1   ,   969 ,   2011    ,   295 ,   6775    ,   7397    ,0 };
            const unsigned long dim1307Kuo2Init[]       =   {   1   ,   1   ,   5   ,   5   ,   15  ,   9   ,   63  ,   21  ,   435 ,   51  ,   1279    ,   1309    ,   4303    ,   15771   ,0 };
            const unsigned long dim1308Kuo2Init[]       =   {   1   ,   3   ,   1   ,   3   ,   3   ,   43  ,   121 ,   59  ,   365 ,   437 ,   215 ,   2899    ,   761 ,   1369    ,0 };
            const unsigned long dim1309Kuo2Init[]       =   {   1   ,   1   ,   3   ,   7   ,   17  ,   25  ,   29  ,   149 ,   291 ,   117 ,   247 ,   2779    ,   7379    ,   11671   ,0 };
            const unsigned long dim1310Kuo2Init[]       =   {   1   ,   3   ,   5   ,   15  ,   7   ,   3   ,   87  ,   213 ,   467 ,   91  ,   145 ,   1241    ,   1557    ,   6427    ,0 };
            const unsigned long dim1311Kuo2Init[]       =   {   1   ,   1   ,   3   ,   9   ,   3   ,   51  ,   1   ,   61  ,   259 ,   535 ,   399 ,   3101    ,   5781    ,   14635   ,0 };
            const unsigned long dim1312Kuo2Init[]       =   {   1   ,   3   ,   3   ,   11  ,   21  ,   17  ,   127 ,   117 ,   203 ,   825 ,   1481    ,   3979    ,   3605    ,   10537   ,0 };
            const unsigned long dim1313Kuo2Init[]       =   {   1   ,   1   ,   5   ,   13  ,   15  ,   5   ,   119 ,   79  ,   439 ,   209 ,   1811    ,   3199    ,   7933    ,   1677    ,0 };
            const unsigned long dim1314Kuo2Init[]       =   {   1   ,   1   ,   3   ,   15  ,   9   ,   9   ,   111 ,   103 ,   145 ,   839 ,   187 ,   2937    ,   6871    ,   10573   ,0 };
            const unsigned long dim1315Kuo2Init[]       =   {   1   ,   3   ,   1   ,   9   ,   19  ,   7   ,   39  ,   37  ,   209 ,   83  ,   1143    ,   3767    ,   6013    ,   8829    ,0 };
            const unsigned long dim1316Kuo2Init[]       =   {   1   ,   1   ,   7   ,   5   ,   31  ,   39  ,   97  ,   167 ,   323 ,   343 ,   999 ,   2179    ,   6683    ,   4079    ,0 };
            const unsigned long dim1317Kuo2Init[]       =   {   1   ,   1   ,   1   ,   7   ,   1   ,   31  ,   5   ,   109 ,   391 ,   17  ,   1943    ,   2753    ,   3825    ,   327 ,0 };
            const unsigned long dim1318Kuo2Init[]       =   {   1   ,   1   ,   7   ,   9   ,   9   ,   19  ,   49  ,   137 ,   421 ,   669 ,   829 ,   3261    ,   3995    ,   4347    ,0 };
            const unsigned long dim1319Kuo2Init[]       =   {   1   ,   3   ,   5   ,   3   ,   25  ,   43  ,   51  ,   97  ,   315 ,   1021    ,   183 ,   87  ,   6465    ,   15415   ,0 };
            const unsigned long dim1320Kuo2Init[]       =   {   1   ,   3   ,   1   ,   1   ,   15  ,   61  ,   111 ,   249 ,   99  ,   1001    ,   791 ,   869 ,   7017    ,   5943    ,0 };
            const unsigned long dim1321Kuo2Init[]       =   {   1   ,   3   ,   7   ,   3   ,   3   ,   7   ,   51  ,   127 ,   275 ,   683 ,   1055    ,   3727    ,   1273    ,   3007    ,0 };
            const unsigned long dim1322Kuo2Init[]       =   {   1   ,   3   ,   1   ,   15  ,   19  ,   33  ,   113 ,   167 ,   473 ,   287 ,   233 ,   905 ,   2213    ,   12233   ,0 };
            const unsigned long dim1323Kuo2Init[]       =   {   1   ,   3   ,   3   ,   7   ,   3   ,   45  ,   83  ,   107 ,   7   ,   651 ,   603 ,   2663    ,   5261    ,   15521   ,0 };
            const unsigned long dim1324Kuo2Init[]       =   {   1   ,   3   ,   3   ,   9   ,   19  ,   49  ,   51  ,   81  ,   101 ,   23  ,   1345    ,   17  ,   4215    ,   7089    ,0 };
            const unsigned long dim1325Kuo2Init[]       =   {   1   ,   3   ,   5   ,   5   ,   7   ,   7   ,   65  ,   221 ,   7   ,   513 ,   1567    ,   4041    ,   6339    ,   8665    ,0 };
            const unsigned long dim1326Kuo2Init[]       =   {   1   ,   1   ,   7   ,   9   ,   15  ,   61  ,   91  ,   29  ,   85  ,   903 ,   1805    ,   1957    ,   63  ,   13941   ,0 };
            const unsigned long dim1327Kuo2Init[]       =   {   1   ,   1   ,   1   ,   1   ,   29  ,   19  ,   31  ,   21  ,   43  ,   353 ,   1419    ,   3999    ,   6651    ,   1897    ,0 };
            const unsigned long dim1328Kuo2Init[]       =   {   1   ,   1   ,   3   ,   13  ,   31  ,   7   ,   65  ,   57  ,   25  ,   1015    ,   1969    ,   1547    ,   4815    ,   4183    ,0 };
            const unsigned long dim1329Kuo2Init[]       =   {   1   ,   1   ,   7   ,   11  ,   25  ,   37  ,   91  ,   163 ,   287 ,   283 ,   1531    ,   149 ,   7799    ,   15  ,0 };
            const unsigned long dim1330Kuo2Init[]       =   {   1   ,   3   ,   5   ,   7   ,   23  ,   25  ,   71  ,   7   ,   439 ,   645 ,   843 ,   315 ,   4577    ,   6703    ,0 };
            const unsigned long dim1331Kuo2Init[]       =   {   1   ,   1   ,   5   ,   15  ,   13  ,   39  ,   57  ,   115 ,   91  ,   475 ,   2003    ,   3787    ,   4053    ,   13299   ,0 };
            const unsigned long dim1332Kuo2Init[]       =   {   1   ,   3   ,   3   ,   7   ,   1   ,   53  ,   53  ,   83  ,   7   ,   133 ,   1523    ,   347 ,   3569    ,   6495    ,0 };
            const unsigned long dim1333Kuo2Init[]       =   {   1   ,   1   ,   7   ,   13  ,   15  ,   53  ,   31  ,   153 ,   181 ,   465 ,   687 ,   3375    ,   3393    ,   565 ,0 };
            const unsigned long dim1334Kuo2Init[]       =   {   1   ,   3   ,   3   ,   3   ,   19  ,   21  ,   95  ,   45  ,   399 ,   1   ,   213 ,   2003    ,   221 ,   11961   ,0 };
            const unsigned long dim1335Kuo2Init[]       =   {   1   ,   1   ,   5   ,   13  ,   27  ,   29  ,   35  ,   55  ,   239 ,   999 ,   23  ,   2645    ,   5399    ,   865 ,0 };
            const unsigned long dim1336Kuo2Init[]       =   {   1   ,   1   ,   5   ,   11  ,   5   ,   3   ,   43  ,   217 ,   321 ,   159 ,   1611    ,   495 ,   435 ,   6989    ,0 };
            const unsigned long dim1337Kuo2Init[]       =   {   1   ,   1   ,   5   ,   1   ,   27  ,   15  ,   29  ,   109 ,   385 ,   449 ,   651 ,   1551    ,   5983    ,   12985   ,0 };
            const unsigned long dim1338Kuo2Init[]       =   {   1   ,   3   ,   5   ,   11  ,   27  ,   35  ,   11  ,   155 ,   279 ,   5   ,   389 ,   1157    ,   2401    ,   12485   ,0 };
            const unsigned long dim1339Kuo2Init[]       =   {   1   ,   1   ,   3   ,   3   ,   7   ,   7   ,   3   ,   185 ,   469 ,   995 ,   271 ,   2315    ,   3269    ,   909 ,0 };
            const unsigned long dim1340Kuo2Init[]       =   {   1   ,   3   ,   5   ,   3   ,   27  ,   31  ,   111 ,   47  ,   351 ,   17  ,   1321    ,   3973    ,   457 ,   15773   ,0 };
            const unsigned long dim1341Kuo2Init[]       =   {   1   ,   3   ,   1   ,   7   ,   19  ,   59  ,   71  ,   215 ,   93  ,   235 ,   1985    ,   1149    ,   1887    ,   2963    ,0 };
            const unsigned long dim1342Kuo2Init[]       =   {   1   ,   3   ,   5   ,   3   ,   3   ,   3   ,   87  ,   205 ,   187 ,   373 ,   1557    ,   311 ,   5807    ,   13343   ,0 };
            const unsigned long dim1343Kuo2Init[]       =   {   1   ,   1   ,   7   ,   13  ,   19  ,   17  ,   31  ,   113 ,   183 ,   31  ,   1117    ,   3241    ,   4781    ,   15071   ,0 };
            const unsigned long dim1344Kuo2Init[]       =   {   1   ,   3   ,   5   ,   5   ,   3   ,   25  ,   33  ,   93  ,   473 ,   353 ,   1005    ,   2385    ,   649 ,   5219    ,0 };
            const unsigned long dim1345Kuo2Init[]       =   {   1   ,   1   ,   3   ,   5   ,   25  ,   37  ,   39  ,   109 ,   511 ,   1005    ,   641 ,   3597    ,   7601    ,   613 ,0 };
            const unsigned long dim1346Kuo2Init[]       =   {   1   ,   1   ,   7   ,   7   ,   3   ,   17  ,   69  ,   145 ,   147 ,   321 ,   849 ,   3057    ,   7299    ,   12343   ,0 };
            const unsigned long dim1347Kuo2Init[]       =   {   1   ,   1   ,   3   ,   15  ,   11  ,   35  ,   19  ,   179 ,   337 ,   53  ,   679 ,   4009    ,   769 ,   2899    ,0 };
            const unsigned long dim1348Kuo2Init[]       =   {   1   ,   1   ,   7   ,   11  ,   19  ,   47  ,   61  ,   11  ,   349 ,   769 ,   445 ,   2051    ,   5653    ,   7143    ,0 };
            const unsigned long dim1349Kuo2Init[]       =   {   1   ,   1   ,   3   ,   9   ,   3   ,   39  ,   53  ,   193 ,   313 ,   299 ,   1579    ,   2237    ,   661 ,   6047    ,0 };
            const unsigned long dim1350Kuo2Init[]       =   {   1   ,   1   ,   7   ,   7   ,   29  ,   17  ,   9   ,   225 ,   143 ,   285 ,   1169    ,   3109    ,   927 ,   5257    ,0 };
            const unsigned long dim1351Kuo2Init[]       =   {   1   ,   3   ,   7   ,   5   ,   5   ,   27  ,   107 ,   253 ,   439 ,   1017    ,   253 ,   3437    ,   5435    ,   13479   ,0 };
            const unsigned long dim1352Kuo2Init[]       =   {   1   ,   3   ,   3   ,   1   ,   9   ,   61  ,   111 ,   75  ,   241 ,   451 ,   1361    ,   2853    ,   5457    ,   15853   ,0 };
            const unsigned long dim1353Kuo2Init[]       =   {   1   ,   1   ,   1   ,   5   ,   15  ,   23  ,   21  ,   115 ,   505 ,   265 ,   131 ,   1277    ,   2545    ,   8955    ,0 };
            const unsigned long dim1354Kuo2Init[]       =   {   1   ,   1   ,   1   ,   1   ,   17  ,   43  ,   3   ,   253 ,   315 ,   19  ,   787 ,   1477    ,   2529    ,   6527    ,0 };
            const unsigned long dim1355Kuo2Init[]       =   {   1   ,   1   ,   1   ,   7   ,   17  ,   31  ,   87  ,   121 ,   201 ,   367 ,   265 ,   3899    ,   1961    ,   12761   ,0 };
            const unsigned long dim1356Kuo2Init[]       =   {   1   ,   1   ,   7   ,   9   ,   9   ,   55  ,   101 ,   163 ,   39  ,   351 ,   1877    ,   2481    ,   3265    ,   15575   ,0 };
            const unsigned long dim1357Kuo2Init[]       =   {   1   ,   1   ,   3   ,   5   ,   29  ,   19  ,   21  ,   39  ,   187 ,   401 ,   547 ,   2477    ,   995 ,   3339    ,0 };
            const unsigned long dim1358Kuo2Init[]       =   {   1   ,   3   ,   3   ,   5   ,   19  ,   25  ,   59  ,   247 ,   389 ,   733 ,   1867    ,   3401    ,   5115    ,   12791   ,0 };
            const unsigned long dim1359Kuo2Init[]       =   {   1   ,   1   ,   3   ,   15  ,   11  ,   39  ,   29  ,   239 ,   157 ,   145 ,   1553    ,   3419    ,   5787    ,   3983    ,0 };
            const unsigned long dim1360Kuo2Init[]       =   {   1   ,   1   ,   1   ,   15  ,   23  ,   47  ,   99  ,   71  ,   335 ,   749 ,   875 ,   5   ,   8049    ,   4585    ,0 };
            const unsigned long dim1361Kuo2Init[]       =   {   1   ,   3   ,   7   ,   11  ,   19  ,   17  ,   25  ,   23  ,   243 ,   171 ,   929 ,   1981    ,   363 ,   13631   ,0 };
            const unsigned long dim1362Kuo2Init[]       =   {   1   ,   1   ,   1   ,   1   ,   19  ,   43  ,   101 ,   199 ,   157 ,   431 ,   1127    ,   3261    ,   4063    ,   10569   ,0 };
            const unsigned long dim1363Kuo2Init[]       =   {   1   ,   3   ,   5   ,   15  ,   15  ,   43  ,   1   ,   199 ,   281 ,   841 ,   1643    ,   2827    ,   6997    ,   11309   ,0 };
            const unsigned long dim1364Kuo2Init[]       =   {   1   ,   1   ,   1   ,   11  ,   5   ,   53  ,   127 ,   147 ,   141 ,   319 ,   1605    ,   3717    ,   6519    ,   15315   ,0 };
            const unsigned long dim1365Kuo2Init[]       =   {   1   ,   3   ,   3   ,   7   ,   31  ,   5   ,   39  ,   221 ,   91  ,   521 ,   527 ,   3253    ,   2877    ,   1439    ,0 };
            const unsigned long dim1366Kuo2Init[]       =   {   1   ,   3   ,   5   ,   1   ,   9   ,   61  ,   69  ,   27  ,   393 ,   557 ,   1949    ,   3035    ,   3717    ,   5099    ,0 };
            const unsigned long dim1367Kuo2Init[]       =   {   1   ,   3   ,   7   ,   3   ,   19  ,   43  ,   77  ,   161 ,   469 ,   721 ,   73  ,   685 ,   1681    ,   5581    ,0 };
            const unsigned long dim1368Kuo2Init[]       =   {   1   ,   1   ,   7   ,   5   ,   27  ,   3   ,   119 ,   223 ,   255 ,   905 ,   501 ,   1809    ,   5591    ,   29  ,0 };
            const unsigned long dim1369Kuo2Init[]       =   {   1   ,   1   ,   3   ,   11  ,   13  ,   1   ,   13  ,   103 ,   305 ,   679 ,   1485    ,   4077    ,   1419    ,   7893    ,0 };
            const unsigned long dim1370Kuo2Init[]       =   {   1   ,   1   ,   1   ,   15  ,   21  ,   7   ,   39  ,   49  ,   451 ,   291 ,   93  ,   3563    ,   5071    ,   15353   ,0 };
            const unsigned long dim1371Kuo2Init[]       =   {   1   ,   3   ,   7   ,   9   ,   19  ,   53  ,   57  ,   149 ,   171 ,   377 ,   1903    ,   235 ,   2181    ,   8207    ,0 };
            const unsigned long dim1372Kuo2Init[]       =   {   1   ,   3   ,   5   ,   15  ,   23  ,   25  ,   63  ,   53  ,   23  ,   263 ,   1681    ,   3305    ,   461 ,   9875    ,0 };
            const unsigned long dim1373Kuo2Init[]       =   {   1   ,   1   ,   3   ,   5   ,   5   ,   15  ,   71  ,   217 ,   493 ,   323 ,   1451    ,   1629    ,   7677    ,   15945   ,0 };
            const unsigned long dim1374Kuo2Init[]       =   {   1   ,   1   ,   5   ,   9   ,   31  ,   53  ,   95  ,   37  ,   357 ,   61  ,   1601    ,   3565    ,   3705    ,   6617    ,0 };
            const unsigned long dim1375Kuo2Init[]       =   {   1   ,   3   ,   1   ,   5   ,   17  ,   33  ,   81  ,   119 ,   233 ,   219 ,   1429    ,   2183    ,   771 ,   7199    ,0 };
            const unsigned long dim1376Kuo2Init[]       =   {   1   ,   1   ,   7   ,   9   ,   27  ,   17  ,   7   ,   17  ,   281 ,   959 ,   1065    ,   3489    ,   6665    ,   14469   ,0 };
            const unsigned long dim1377Kuo2Init[]       =   {   1   ,   3   ,   1   ,   15  ,   21  ,   15  ,   99  ,   77  ,   437 ,   571 ,   207 ,   609 ,   7287    ,   9395    ,0 };
            const unsigned long dim1378Kuo2Init[]       =   {   1   ,   1   ,   7   ,   1   ,   1   ,   49  ,   39  ,   189 ,   111 ,   65  ,   1345    ,   2559    ,   7835    ,   9869    ,0 };
            const unsigned long dim1379Kuo2Init[]       =   {   1   ,   1   ,   1   ,   1   ,   27  ,   17  ,   93  ,   83  ,   339 ,   465 ,   1445    ,   3557    ,   6489    ,   6985    ,0 };
            const unsigned long dim1380Kuo2Init[]       =   {   1   ,   1   ,   5   ,   3   ,   1   ,   7   ,   35  ,   39  ,   491 ,   563 ,   1815    ,   2513    ,   5523    ,   12691   ,0 };
            const unsigned long dim1381Kuo2Init[]       =   {   1   ,   1   ,   3   ,   5   ,   3   ,   31  ,   127 ,   91  ,   87  ,   447 ,   1237    ,   3419    ,   6979    ,   6311    ,0 };
            const unsigned long dim1382Kuo2Init[]       =   {   1   ,   1   ,   7   ,   3   ,   25  ,   61  ,   45  ,   219 ,   455 ,   859 ,   781 ,   2991    ,   6585    ,   11511   ,0 };
            const unsigned long dim1383Kuo2Init[]       =   {   1   ,   3   ,   3   ,   13  ,   15  ,   37  ,   77  ,   211 ,   403 ,   349 ,   35  ,   2785    ,   4439    ,   5715    ,0 };
            const unsigned long dim1384Kuo2Init[]       =   {   1   ,   1   ,   3   ,   1   ,   1   ,   19  ,   45  ,   37  ,   459 ,   839 ,   17  ,   3223    ,   6411    ,   9209    ,0 };
            const unsigned long dim1385Kuo2Init[]       =   {   1   ,   3   ,   1   ,   3   ,   27  ,   43  ,   63  ,   241 ,   455 ,   593 ,   1855    ,   3689    ,   7825    ,   10197   ,0 };
            const unsigned long dim1386Kuo2Init[]       =   {   1   ,   3   ,   3   ,   1   ,   5   ,   17  ,   71  ,   209 ,   57  ,   991 ,   617 ,   3097    ,   5431    ,   2057    ,0 };
            const unsigned long dim1387Kuo2Init[]       =   {   1   ,   1   ,   5   ,   15  ,   31  ,   21  ,   119 ,   53  ,   447 ,   969 ,   335 ,   3793    ,   3225    ,   9809    ,0 };
            const unsigned long dim1388Kuo2Init[]       =   {   1   ,   1   ,   1   ,   11  ,   17  ,   51  ,   17  ,   7   ,   257 ,   221 ,   119 ,   2631    ,   6395    ,   8495    ,0 };
            const unsigned long dim1389Kuo2Init[]       =   {   1   ,   3   ,   1   ,   15  ,   31  ,   13  ,   3   ,   9   ,   479 ,   233 ,   1051    ,   1953    ,   7863    ,   5969    ,0 };
            const unsigned long dim1390Kuo2Init[]       =   {   1   ,   3   ,   1   ,   9   ,   13  ,   49  ,   57  ,   161 ,   65  ,   871 ,   2021    ,   3337    ,   7071    ,   12969   ,0 };
            const unsigned long dim1391Kuo2Init[]       =   {   1   ,   3   ,   5   ,   15  ,   1   ,   53  ,   3   ,   77  ,   487 ,   207 ,   1761    ,   847 ,   3131    ,   1933    ,0 };
            const unsigned long dim1392Kuo2Init[]       =   {   1   ,   3   ,   1   ,   7   ,   7   ,   21  ,   55  ,   217 ,   167 ,   113 ,   1029    ,   2761    ,   2063    ,   5267    ,0 };
            const unsigned long dim1393Kuo2Init[]       =   {   1   ,   3   ,   3   ,   7   ,   15  ,   57  ,   119 ,   9   ,   211 ,   93  ,   209 ,   2905    ,   2589    ,   12341   ,0 };
            const unsigned long dim1394Kuo2Init[]       =   {   1   ,   3   ,   1   ,   15  ,   25  ,   7   ,   41  ,   87  ,   451 ,   613 ,   353 ,   2619    ,   6015    ,   14003   ,0 };
            const unsigned long dim1395Kuo2Init[]       =   {   1   ,   3   ,   3   ,   11  ,   23  ,   5   ,   3   ,   145 ,   471 ,   685 ,   3   ,   3431    ,   2843    ,   11725   ,0 };
            const unsigned long dim1396Kuo2Init[]       =   {   1   ,   3   ,   7   ,   11  ,   13  ,   45  ,   73  ,   45  ,   169 ,   395 ,   1279    ,   2589    ,   1419    ,   14965   ,0 };
            const unsigned long dim1397Kuo2Init[]       =   {   1   ,   3   ,   5   ,   9   ,   25  ,   23  ,   77  ,   149 ,   27  ,   441 ,   1391    ,   1853    ,   3993    ,   15507   ,0 };
            const unsigned long dim1398Kuo2Init[]       =   {   1   ,   1   ,   1   ,   7   ,   21  ,   33  ,   39  ,   59  ,   161 ,   399 ,   447 ,   1249    ,   5407    ,   11825   ,0 };
            const unsigned long dim1399Kuo2Init[]       =   {   1   ,   1   ,   5   ,   3   ,   25  ,   53  ,   37  ,   237 ,   91  ,   625 ,   1195    ,   1161    ,   7005    ,   7831    ,0 };
            const unsigned long dim1400Kuo2Init[]       =   {   1   ,   3   ,   7   ,   5   ,   7   ,   63  ,   37  ,   101 ,   399 ,   645 ,   545 ,   3419    ,   6619    ,   2179    ,0 };
            const unsigned long dim1401Kuo2Init[]       =   {   1   ,   1   ,   5   ,   11  ,   11  ,   51  ,   103 ,   207 ,   267 ,   909 ,   1975    ,   2279    ,   1991    ,   1891    ,0 };
            const unsigned long dim1402Kuo2Init[]       =   {   1   ,   1   ,   7   ,   1   ,   27  ,   45  ,   9   ,   1   ,   423 ,   197 ,   697 ,   2765    ,   2653    ,   2613    ,0 };
            const unsigned long dim1403Kuo2Init[]       =   {   1   ,   3   ,   7   ,   1   ,   27  ,   13  ,   45  ,   47  ,   493 ,   131 ,   565 ,   1363    ,   4921    ,   10023   ,0 };
            const unsigned long dim1404Kuo2Init[]       =   {   1   ,   3   ,   7   ,   1   ,   7   ,   47  ,   99  ,   107 ,   343 ,   933 ,   1017    ,   3931    ,   1329    ,   2833    ,0 };
            const unsigned long dim1405Kuo2Init[]       =   {   1   ,   1   ,   3   ,   1   ,   13  ,   9   ,   5   ,   131 ,   475 ,   537 ,   1857    ,   2579    ,   6377    ,   14363   ,0 };
            const unsigned long dim1406Kuo2Init[]       =   {   1   ,   1   ,   5   ,   5   ,   15  ,   7   ,   87  ,   39  ,   51  ,   563 ,   1145    ,   3639    ,   4261    ,   9585    ,0 };
            const unsigned long dim1407Kuo2Init[]       =   {   1   ,   3   ,   1   ,   3   ,   3   ,   61  ,   23  ,   223 ,   221 ,   165 ,   315 ,   2475    ,   7907    ,   10171   ,0 };
            const unsigned long dim1408Kuo2Init[]       =   {   1   ,   1   ,   3   ,   9   ,   23  ,   57  ,   121 ,   171 ,   439 ,   101 ,   261 ,   1115    ,   1291    ,   2951    ,0 };
            const unsigned long dim1409Kuo2Init[]       =   {   1   ,   1   ,   1   ,   9   ,   15  ,   23  ,   81  ,   189 ,   301 ,   515 ,   297 ,   915 ,   763 ,   11037   ,0 };
            const unsigned long dim1410Kuo2Init[]       =   {   1   ,   3   ,   5   ,   9   ,   17  ,   45  ,   33  ,   101 ,   65  ,   687 ,   2011    ,   1821    ,   6891    ,   7321    ,0 };
            const unsigned long dim1411Kuo2Init[]       =   {   1   ,   3   ,   5   ,   5   ,   9   ,   61  ,   11  ,   69  ,   345 ,   211 ,   393 ,   331 ,   3953    ,   13987   ,0 };
            const unsigned long dim1412Kuo2Init[]       =   {   1   ,   1   ,   5   ,   15  ,   9   ,   23  ,   11  ,   45  ,   325 ,   1019    ,   1499    ,   831 ,   127 ,   9237    ,0 };
            const unsigned long dim1413Kuo2Init[]       =   {   1   ,   1   ,   5   ,   5   ,   19  ,   9   ,   53  ,   213 ,   417 ,   761 ,   1275    ,   2117    ,   465 ,   15195   ,0 };
            const unsigned long dim1414Kuo2Init[]       =   {   1   ,   3   ,   1   ,   7   ,   25  ,   53  ,   57  ,   95  ,   31  ,   791 ,   345 ,   3769    ,   3487    ,   4269    ,0 };
            const unsigned long dim1415Kuo2Init[]       =   {   1   ,   3   ,   3   ,   1   ,   17  ,   9   ,   127 ,   203 ,   67  ,   435 ,   281 ,   3857    ,   819 ,   11235   ,0 };
            const unsigned long dim1416Kuo2Init[]       =   {   1   ,   3   ,   1   ,   7   ,   23  ,   45  ,   127 ,   197 ,   501 ,   11  ,   1125    ,   2781    ,   7979    ,   1699    ,0 };
            const unsigned long dim1417Kuo2Init[]       =   {   1   ,   3   ,   3   ,   13  ,   21  ,   21  ,   57  ,   125 ,   225 ,   593 ,   27  ,   539 ,   7989    ,   9597    ,0 };
            const unsigned long dim1418Kuo2Init[]       =   {   1   ,   3   ,   7   ,   7   ,   29  ,   27  ,   57  ,   19  ,   51  ,   981 ,   205 ,   169 ,   5713    ,   4777    ,0 };
            const unsigned long dim1419Kuo2Init[]       =   {   1   ,   3   ,   7   ,   7   ,   1   ,   59  ,   57  ,   61  ,   145 ,   683 ,   1043    ,   11  ,   5277    ,   5095    ,0 };
            const unsigned long dim1420Kuo2Init[]       =   {   1   ,   1   ,   1   ,   3   ,   7   ,   39  ,   123 ,   251 ,   259 ,   295 ,   1333    ,   395 ,   5463    ,   11593   ,0 };
            const unsigned long dim1421Kuo2Init[]       =   {   1   ,   1   ,   7   ,   15  ,   1   ,   9   ,   71  ,   79  ,   299 ,   1007    ,   1403    ,   237 ,   4845    ,   10521   ,0 };
            const unsigned long dim1422Kuo2Init[]       =   {   1   ,   1   ,   5   ,   5   ,   11  ,   51  ,   45  ,   51  ,   489 ,   865 ,   511 ,   1415    ,   7867    ,   4235    ,0 };
            const unsigned long dim1423Kuo2Init[]       =   {   1   ,   3   ,   3   ,   9   ,   23  ,   31  ,   83  ,   77  ,   461 ,   111 ,   1985    ,   3401    ,   271 ,   9763    ,0 };
            const unsigned long dim1424Kuo2Init[]       =   {   1   ,   3   ,   5   ,   11  ,   19  ,   53  ,   45  ,   135 ,   171 ,   593 ,   1761    ,   1019    ,   7715    ,   9835    ,0 };
            const unsigned long dim1425Kuo2Init[]       =   {   1   ,   1   ,   3   ,   5   ,   21  ,   47  ,   95  ,   1   ,   419 ,   547 ,   447 ,   3475    ,   7829    ,   12019   ,0 };
            const unsigned long dim1426Kuo2Init[]       =   {   1   ,   3   ,   7   ,   13  ,   15  ,   29  ,   21  ,   197 ,   491 ,   997 ,   453 ,   3415    ,   745 ,   11259   ,0 };
            const unsigned long dim1427Kuo2Init[]       =   {   1   ,   1   ,   3   ,   11  ,   19  ,   39  ,   99  ,   83  ,   35  ,   941 ,   1175    ,   2749    ,   4181    ,   8791    ,0 };
            const unsigned long dim1428Kuo2Init[]       =   {   1   ,   1   ,   5   ,   7   ,   23  ,   3   ,   101 ,   11  ,   81  ,   291 ,   1815    ,   3391    ,   7615    ,   1305    ,0 };
            const unsigned long dim1429Kuo2Init[]       =   {   1   ,   3   ,   5   ,   13  ,   27  ,   35  ,   71  ,   185 ,   87  ,   939 ,   1025    ,   1577    ,   5919    ,   5335    ,0 };
            const unsigned long dim1430Kuo2Init[]       =   {   1   ,   3   ,   3   ,   9   ,   3   ,   31  ,   23  ,   213 ,   371 ,   595 ,   2003    ,   647 ,   909 ,   7503    ,0 };
            const unsigned long dim1431Kuo2Init[]       =   {   1   ,   1   ,   5   ,   9   ,   31  ,   63  ,   125 ,   113 ,   273 ,   985 ,   1049    ,   1819    ,   2773    ,   6495    ,0 };
            const unsigned long dim1432Kuo2Init[]       =   {   1   ,   1   ,   1   ,   5   ,   21  ,   7   ,   87  ,   207 ,   455 ,   403 ,   677 ,   175 ,   531 ,   10467   ,0 };
            const unsigned long dim1433Kuo2Init[]       =   {   1   ,   1   ,   5   ,   1   ,   3   ,   45  ,   109 ,   75  ,   429 ,   983 ,   423 ,   3491    ,   7949    ,   15253   ,0 };
            const unsigned long dim1434Kuo2Init[]       =   {   1   ,   1   ,   3   ,   1   ,   23  ,   27  ,   91  ,   5   ,   235 ,   121 ,   1087    ,   325 ,   5191    ,   8271    ,0 };
            const unsigned long dim1435Kuo2Init[]       =   {   1   ,   3   ,   3   ,   15  ,   29  ,   63  ,   91  ,   129 ,   305 ,   827 ,   1311    ,   2765    ,   253 ,   1495    ,0 };
            const unsigned long dim1436Kuo2Init[]       =   {   1   ,   3   ,   1   ,   15  ,   21  ,   45  ,   7   ,   21  ,   169 ,   275 ,   629 ,   589 ,   4623    ,   5491    ,0 };
            const unsigned long dim1437Kuo2Init[]       =   {   1   ,   3   ,   1   ,   3   ,   13  ,   45  ,   127 ,   105 ,   377 ,   49  ,   1161    ,   489 ,   595 ,   7041    ,0 };
            const unsigned long dim1438Kuo2Init[]       =   {   1   ,   1   ,   5   ,   11  ,   31  ,   61  ,   91  ,   1   ,   35  ,   189 ,   1479    ,   2041    ,   1585    ,   13009   ,0 };
            const unsigned long dim1439Kuo2Init[]       =   {   1   ,   3   ,   3   ,   1   ,   23  ,   47  ,   69  ,   47  ,   361 ,   375 ,   77  ,   3777    ,   5769    ,   9925    ,0 };
            const unsigned long dim1440Kuo2Init[]       =   {   1   ,   3   ,   3   ,   9   ,   5   ,   13  ,   31  ,   119 ,   411 ,   735 ,   839 ,   2919    ,   543 ,   12929   ,0 };
            const unsigned long dim1441Kuo2Init[]       =   {   1   ,   3   ,   5   ,   11  ,   19  ,   5   ,   29  ,   185 ,   115 ,   841 ,   507 ,   1433    ,   5091    ,   3593    ,0 };
            const unsigned long dim1442Kuo2Init[]       =   {   1   ,   3   ,   1   ,   13  ,   15  ,   19  ,   23  ,   83  ,   319 ,   99  ,   785 ,   697 ,   3703    ,   8733    ,0 };
            const unsigned long dim1443Kuo2Init[]       =   {   1   ,   3   ,   5   ,   3   ,   7   ,   31  ,   53  ,   213 ,   287 ,   797 ,   1807    ,   879 ,   4227    ,   12013   ,0 };
            const unsigned long dim1444Kuo2Init[]       =   {   1   ,   3   ,   3   ,   1   ,   23  ,   53  ,   91  ,   173 ,   253 ,   971 ,   1017    ,   2185    ,   4911    ,   15359   ,0 };
            const unsigned long dim1445Kuo2Init[]       =   {   1   ,   3   ,   7   ,   3   ,   17  ,   27  ,   77  ,   225 ,   97  ,   199 ,   607 ,   1615    ,   963 ,   671 ,0 };
            const unsigned long dim1446Kuo2Init[]       =   {   1   ,   1   ,   5   ,   11  ,   9   ,   61  ,   3   ,   105 ,   455 ,   953 ,   455 ,   3925    ,   3947    ,   7873    ,0 };
            const unsigned long dim1447Kuo2Init[]       =   {   1   ,   3   ,   7   ,   7   ,   3   ,   49  ,   121 ,   109 ,   239 ,   855 ,   789 ,   3627    ,   3873    ,   6785    ,0 };
            const unsigned long dim1448Kuo2Init[]       =   {   1   ,   1   ,   5   ,   5   ,   29  ,   35  ,   99  ,   159 ,   17  ,   221 ,   855 ,   2081    ,   2791    ,   11413   ,0 };
            const unsigned long dim1449Kuo2Init[]       =   {   1   ,   3   ,   7   ,   5   ,   23  ,   53  ,   75  ,   93  ,   207 ,   189 ,   2035    ,   1829    ,   7581    ,   1425    ,0 };
            const unsigned long dim1450Kuo2Init[]       =   {   1   ,   1   ,   3   ,   9   ,   3   ,   61  ,   61  ,   237 ,   53  ,   875 ,   1617    ,   727 ,   2683    ,   15797   ,0 };
            const unsigned long dim1451Kuo2Init[]       =   {   1   ,   1   ,   1   ,   11  ,   23  ,   3   ,   105 ,   107 ,   353 ,   579 ,   119 ,   3041    ,   3005    ,   15581   ,0 };
            const unsigned long dim1452Kuo2Init[]       =   {   1   ,   1   ,   1   ,   11  ,   31  ,   15  ,   89  ,   215 ,   161 ,   39  ,   629 ,   2873    ,   1635    ,   1029    ,0 };
            const unsigned long dim1453Kuo2Init[]       =   {   1   ,   3   ,   3   ,   15  ,   15  ,   25  ,   53  ,   27  ,   167 ,   497 ,   1795    ,   1283    ,   7869    ,   15821   ,0 };
            const unsigned long dim1454Kuo2Init[]       =   {   1   ,   3   ,   7   ,   3   ,   27  ,   39  ,   111 ,   29  ,   445 ,   365 ,   273 ,   2183    ,   475 ,   12503   ,0 };
            const unsigned long dim1455Kuo2Init[]       =   {   1   ,   3   ,   7   ,   5   ,   9   ,   13  ,   25  ,   115 ,   419 ,   387 ,   1663    ,   3445    ,   157 ,   10549   ,0 };
            const unsigned long dim1456Kuo2Init[]       =   {   1   ,   1   ,   5   ,   13  ,   23  ,   39  ,   7   ,   247 ,   127 ,   285 ,   675 ,   1785    ,   1823    ,   10183   ,0 };
            const unsigned long dim1457Kuo2Init[]       =   {   1   ,   1   ,   1   ,   11  ,   31  ,   39  ,   93  ,   19  ,   125 ,   453 ,   1963    ,   2619    ,   6195    ,   2231    ,0 };
            const unsigned long dim1458Kuo2Init[]       =   {   1   ,   1   ,   7   ,   3   ,   31  ,   51  ,   59  ,   19  ,   15  ,   251 ,   749 ,   2469    ,   369 ,   15413   ,0 };
            const unsigned long dim1459Kuo2Init[]       =   {   1   ,   1   ,   7   ,   9   ,   19  ,   59  ,   111 ,   255 ,   91  ,   89  ,   1447    ,   353 ,   6701    ,   7955    ,0 };
            const unsigned long dim1460Kuo2Init[]       =   {   1   ,   1   ,   7   ,   3   ,   3   ,   21  ,   115 ,   159 ,   489 ,   609 ,   1993    ,   2961    ,   3117    ,   3977    ,0 };
            const unsigned long dim1461Kuo2Init[]       =   {   1   ,   3   ,   7   ,   13  ,   13  ,   13  ,   57  ,   41  ,   5   ,   337 ,   2005    ,   2587    ,   3873    ,   12829   ,0 };
            const unsigned long dim1462Kuo2Init[]       =   {   1   ,   1   ,   3   ,   9   ,   9   ,   3   ,   39  ,   165 ,   155 ,   353 ,   985 ,   2221    ,   2973    ,   6331    ,0 };
            const unsigned long dim1463Kuo2Init[]       =   {   1   ,   3   ,   3   ,   5   ,   3   ,   15  ,   121 ,   191 ,   367 ,   513 ,   1765    ,   3957    ,   3113    ,   10669   ,0 };
            const unsigned long dim1464Kuo2Init[]       =   {   1   ,   3   ,   7   ,   1   ,   13  ,   41  ,   63  ,   241 ,   405 ,   43  ,   1211    ,   1499    ,   1095    ,   11867   ,0 };
            const unsigned long dim1465Kuo2Init[]       =   {   1   ,   3   ,   7   ,   15  ,   17  ,   17  ,   63  ,   197 ,   377 ,   981 ,   1615    ,   2407    ,   6669    ,   411 ,0 };
            const unsigned long dim1466Kuo2Init[]       =   {   1   ,   1   ,   3   ,   1   ,   3   ,   39  ,   85  ,   15  ,   9   ,   747 ,   423 ,   3103    ,   7341    ,   10953   ,0 };
            const unsigned long dim1467Kuo2Init[]       =   {   1   ,   1   ,   7   ,   13  ,   25  ,   9   ,   17  ,   137 ,   267 ,   405 ,   1819    ,   641 ,   7305    ,   13971   ,0 };
            const unsigned long dim1468Kuo2Init[]       =   {   1   ,   3   ,   5   ,   5   ,   19  ,   45  ,   121 ,   65  ,   385 ,   715 ,   267 ,   1841    ,   6317    ,   5827    ,0 };
            const unsigned long dim1469Kuo2Init[]       =   {   1   ,   1   ,   7   ,   9   ,   11  ,   31  ,   57  ,   123 ,   107 ,   913 ,   669 ,   1053    ,   1213    ,   8791    ,0 };
            const unsigned long dim1470Kuo2Init[]       =   {   1   ,   1   ,   3   ,   7   ,   21  ,   3   ,   57  ,   227 ,   401 ,   443 ,   1847    ,   3523    ,   4717    ,   2505    ,0 };
            const unsigned long dim1471Kuo2Init[]       =   {   1   ,   3   ,   3   ,   15  ,   29  ,   41  ,   67  ,   217 ,   431 ,   37  ,   1775    ,   297 ,   1921    ,   8715    ,0 };
            const unsigned long dim1472Kuo2Init[]       =   {   1   ,   3   ,   7   ,   7   ,   27  ,   35  ,   113 ,   57  ,   23  ,   927 ,   141 ,   1363    ,   4209    ,   4399    ,0 };
            const unsigned long dim1473Kuo2Init[]       =   {   1   ,   1   ,   7   ,   9   ,   19  ,   51  ,   107 ,   189 ,   463 ,   695 ,   1931    ,   1541    ,   1285    ,   15703   ,0 };
            const unsigned long dim1474Kuo2Init[]       =   {   1   ,   3   ,   5   ,   7   ,   29  ,   43  ,   109 ,   117 ,   389 ,   7   ,   1129    ,   1945    ,   6987    ,   799 ,0 };
            const unsigned long dim1475Kuo2Init[]       =   {   1   ,   1   ,   1   ,   13  ,   9   ,   7   ,   119 ,   61  ,   385 ,   315 ,   1953    ,   1745    ,   8117    ,   15863   ,0 };
            const unsigned long dim1476Kuo2Init[]       =   {   1   ,   1   ,   7   ,   7   ,   21  ,   55  ,   95  ,   113 ,   293 ,   899 ,   1327    ,   161 ,   6613    ,   14949   ,0 };
            const unsigned long dim1477Kuo2Init[]       =   {   1   ,   1   ,   7   ,   9   ,   19  ,   21  ,   89  ,   5   ,   477 ,   441 ,   1477    ,   1995    ,   6633    ,   5523    ,0 };
            const unsigned long dim1478Kuo2Init[]       =   {   1   ,   3   ,   3   ,   11  ,   15  ,   27  ,   17  ,   239 ,   263 ,   267 ,   1729    ,   3583    ,   4117    ,   13637   ,0 };
            const unsigned long dim1479Kuo2Init[]       =   {   1   ,   1   ,   5   ,   5   ,   11  ,   53  ,   15  ,   201 ,   401 ,   845 ,   741 ,   2803    ,   3087    ,   9001    ,0 };
            const unsigned long dim1480Kuo2Init[]       =   {   1   ,   3   ,   3   ,   11  ,   17  ,   21  ,   65  ,   109 ,   163 ,   251 ,   1487    ,   1541    ,   481 ,   10545   ,0 };
            const unsigned long dim1481Kuo2Init[]       =   {   1   ,   3   ,   3   ,   13  ,   21  ,   61  ,   71  ,   121 ,   331 ,   551 ,   1237    ,   805 ,   3799    ,   1437    ,0 };
            const unsigned long dim1482Kuo2Init[]       =   {   1   ,   3   ,   1   ,   11  ,   21  ,   29  ,   61  ,   161 ,   193 ,   745 ,   935 ,   1013    ,   3507    ,   6091    ,0 };
            const unsigned long dim1483Kuo2Init[]       =   {   1   ,   3   ,   7   ,   15  ,   5   ,   59  ,   99  ,   147 ,   467 ,   625 ,   1873    ,   3459    ,   5749    ,   10571   ,0 };
            const unsigned long dim1484Kuo2Init[]       =   {   1   ,   3   ,   1   ,   3   ,   9   ,   21  ,   57  ,   63  ,   205 ,   891 ,   1963    ,   2271    ,   6303    ,   9565    ,0 };
            const unsigned long dim1485Kuo2Init[]       =   {   1   ,   1   ,   1   ,   1   ,   17  ,   55  ,   21  ,   105 ,   181 ,   763 ,   763 ,   337 ,   2283    ,   5203    ,0 };
            const unsigned long dim1486Kuo2Init[]       =   {   1   ,   3   ,   3   ,   11  ,   25  ,   33  ,   7   ,   107 ,   329 ,   49  ,   31  ,   995 ,   757 ,   9781    ,0 };
            const unsigned long dim1487Kuo2Init[]       =   {   1   ,   3   ,   7   ,   13  ,   17  ,   19  ,   107 ,   189 ,   359 ,   709 ,   819 ,   3943    ,   1891    ,   12173   ,0 };
            const unsigned long dim1488Kuo2Init[]       =   {   1   ,   1   ,   7   ,   5   ,   3   ,   55  ,   121 ,   37  ,   493 ,   579 ,   2021    ,   3129    ,   6519    ,   14569   ,0 };
            const unsigned long dim1489Kuo2Init[]       =   {   1   ,   1   ,   5   ,   15  ,   17  ,   3   ,   59  ,   75  ,   453 ,   323 ,   1559    ,   3695    ,   7711    ,   7665    ,0 };
            const unsigned long dim1490Kuo2Init[]       =   {   1   ,   3   ,   3   ,   7   ,   9   ,   35  ,   121 ,   179 ,   345 ,   973 ,   337 ,   3655    ,   3815    ,   6317    ,0 };
            const unsigned long dim1491Kuo2Init[]       =   {   1   ,   3   ,   7   ,   7   ,   9   ,   27  ,   59  ,   253 ,   9   ,   221 ,   587 ,   1087    ,   7295    ,   12193   ,0 };
            const unsigned long dim1492Kuo2Init[]       =   {   1   ,   1   ,   3   ,   7   ,   27  ,   1   ,   105 ,   99  ,   49  ,   555 ,   2019    ,   2863    ,   5211    ,   2485    ,0 };
            const unsigned long dim1493Kuo2Init[]       =   {   1   ,   1   ,   3   ,   15  ,   25  ,   55  ,   23  ,   221 ,   501 ,   671 ,   11  ,   1801    ,   6485    ,   14983   ,0 };
            const unsigned long dim1494Kuo2Init[]       =   {   1   ,   3   ,   1   ,   3   ,   27  ,   45  ,   69  ,   159 ,   77  ,   541 ,   1441    ,   2043    ,   5609    ,   6799    ,0 };
            const unsigned long dim1495Kuo2Init[]       =   {   1   ,   1   ,   5   ,   3   ,   21  ,   31  ,   7   ,   197 ,   511 ,   299 ,   1523    ,   3341    ,   301 ,   8639    ,0 };
            const unsigned long dim1496Kuo2Init[]       =   {   1   ,   1   ,   7   ,   1   ,   9   ,   59  ,   63  ,   253 ,   191 ,   653 ,   383 ,   1239    ,   7839    ,   3699    ,0 };
            const unsigned long dim1497Kuo2Init[]       =   {   1   ,   1   ,   3   ,   9   ,   13  ,   33  ,   73  ,   209 ,   125 ,   939 ,   1507    ,   3527    ,   6039    ,   14027   ,0 };
            const unsigned long dim1498Kuo2Init[]       =   {   1   ,   1   ,   3   ,   13  ,   13  ,   21  ,   5   ,   15  ,   301 ,   625 ,   713 ,   2867    ,   3129    ,   4939    ,0 };
            const unsigned long dim1499Kuo2Init[]       =   {   1   ,   1   ,   1   ,   13  ,   23  ,   5   ,   115 ,   69  ,   101 ,   327 ,   563 ,   2623    ,   6491    ,   7381    ,0 };
            const unsigned long dim1500Kuo2Init[]       =   {   1   ,   1   ,   5   ,   7   ,   9   ,   53  ,   1   ,   91  ,   53  ,   823 ,   281 ,   2281    ,   6509    ,   2515    ,0 };
            const unsigned long dim1501Kuo2Init[]       =   {   1   ,   1   ,   5   ,   5   ,   13  ,   61  ,   105 ,   183 ,   39  ,   493 ,   773 ,   1815    ,   1549    ,   7511    ,0 };
            const unsigned long dim1502Kuo2Init[]       =   {   1   ,   1   ,   3   ,   13  ,   29  ,   27  ,   97  ,   205 ,   425 ,   13  ,   1227    ,   625 ,   137 ,   6785    ,0 };
            const unsigned long dim1503Kuo2Init[]       =   {   1   ,   3   ,   7   ,   11  ,   13  ,   29  ,   105 ,   85  ,   251 ,   151 ,   1515    ,   3137    ,   8051    ,   14333   ,0 };
            const unsigned long dim1504Kuo2Init[]       =   {   1   ,   3   ,   7   ,   9   ,   25  ,   59  ,   99  ,   43  ,   435 ,   361 ,   297 ,   449 ,   5633    ,   5813    ,0 };
            const unsigned long dim1505Kuo2Init[]       =   {   1   ,   3   ,   1   ,   13  ,   17  ,   45  ,   113 ,   79  ,   315 ,   781 ,   785 ,   3463    ,   1299    ,   5959    ,0 };
            const unsigned long dim1506Kuo2Init[]       =   {   1   ,   1   ,   7   ,   15  ,   19  ,   47  ,   35  ,   13  ,   487 ,   227 ,   1965    ,   4035    ,   4285    ,   14395   ,0 };
            const unsigned long dim1507Kuo2Init[]       =   {   1   ,   3   ,   7   ,   5   ,   9   ,   41  ,   91  ,   127 ,   101 ,   519 ,   1147    ,   3851    ,   7929    ,   1807    ,0 };
            const unsigned long dim1508Kuo2Init[]       =   {   1   ,   1   ,   7   ,   1   ,   19  ,   63  ,   45  ,   75  ,   437 ,   309 ,   1327    ,   3913    ,   5961    ,   2825    ,0 };
            const unsigned long dim1509Kuo2Init[]       =   {   1   ,   1   ,   1   ,   15  ,   11  ,   1   ,   77  ,   11  ,   105 ,   499 ,   859 ,   2403    ,   4453    ,   12297   ,0 };
            const unsigned long dim1510Kuo2Init[]       =   {   1   ,   1   ,   5   ,   5   ,   9   ,   33  ,   87  ,   85  ,   167 ,   35  ,   501 ,   1027    ,   4489    ,   10513   ,0 };
            const unsigned long dim1511Kuo2Init[]       =   {   1   ,   1   ,   7   ,   13  ,   5   ,   3   ,   75  ,   75  ,   417 ,   319 ,   423 ,   1107    ,   371 ,   12493   ,0 };
            const unsigned long dim1512Kuo2Init[]       =   {   1   ,   3   ,   7   ,   7   ,   15  ,   57  ,   7   ,   105 ,   95  ,   287 ,   1865    ,   2795    ,   5187    ,   543 ,0 };
            const unsigned long dim1513Kuo2Init[]       =   {   1   ,   1   ,   1   ,   11  ,   1   ,   3   ,   121 ,   99  ,   299 ,   415 ,   47  ,   997 ,   2939    ,   8529    ,0 };
            const unsigned long dim1514Kuo2Init[]       =   {   1   ,   1   ,   7   ,   11  ,   23  ,   17  ,   101 ,   209 ,   267 ,   459 ,   1965    ,   2153    ,   6361    ,   9553    ,0 };
            const unsigned long dim1515Kuo2Init[]       =   {   1   ,   1   ,   3   ,   5   ,   1   ,   9   ,   7   ,   17  ,   109 ,   245 ,   701 ,   2161    ,   2109    ,   3301    ,0 };
            const unsigned long dim1516Kuo2Init[]       =   {   1   ,   3   ,   5   ,   3   ,   13  ,   21  ,   113 ,   159 ,   261 ,   467 ,   127 ,   437 ,   8071    ,   11647   ,0 };
            const unsigned long dim1517Kuo2Init[]       =   {   1   ,   3   ,   1   ,   15  ,   29  ,   63  ,   93  ,   31  ,   189 ,   881 ,   17  ,   2689    ,   1273    ,   4991    ,0 };
            const unsigned long dim1518Kuo2Init[]       =   {   1   ,   3   ,   3   ,   7   ,   31  ,   23  ,   39  ,   177 ,   505 ,   207 ,   67  ,   2897    ,   315 ,   5981    ,0 };
            const unsigned long dim1519Kuo2Init[]       =   {   1   ,   3   ,   5   ,   5   ,   31  ,   19  ,   87  ,   237 ,   119 ,   439 ,   1893    ,   3079    ,   3709    ,   15051   ,0 };
            const unsigned long dim1520Kuo2Init[]       =   {   1   ,   3   ,   5   ,   1   ,   31  ,   35  ,   125 ,   173 ,   197 ,   765 ,   1419    ,   465 ,   1919    ,   2641    ,0 };
            const unsigned long dim1521Kuo2Init[]       =   {   1   ,   1   ,   5   ,   11  ,   19  ,   15  ,   49  ,   255 ,   447 ,   939 ,   137 ,   2371    ,   6617    ,   79  ,0 };
            const unsigned long dim1522Kuo2Init[]       =   {   1   ,   3   ,   3   ,   1   ,   11  ,   29  ,   81  ,   81  ,   327 ,   593 ,   135 ,   415 ,   2417    ,   8671    ,0 };
            const unsigned long dim1523Kuo2Init[]       =   {   1   ,   1   ,   7   ,   11  ,   11  ,   35  ,   9   ,   143 ,   311 ,   57  ,   557 ,   793 ,   6597    ,   12979   ,0 };
            const unsigned long dim1524Kuo2Init[]       =   {   1   ,   1   ,   7   ,   1   ,   27  ,   1   ,   105 ,   149 ,   417 ,   605 ,   431 ,   2419    ,   4113    ,   5651    ,0 };
            const unsigned long dim1525Kuo2Init[]       =   {   1   ,   1   ,   3   ,   15  ,   1   ,   45  ,   63  ,   119 ,   167 ,   607 ,   961 ,   2561    ,   7839    ,   13343   ,0 };
            const unsigned long dim1526Kuo2Init[]       =   {   1   ,   1   ,   5   ,   15  ,   11  ,   47  ,   75  ,   209 ,   175 ,   219 ,   443 ,   1311    ,   3383    ,   14073   ,0 };
            const unsigned long dim1527Kuo2Init[]       =   {   1   ,   3   ,   5   ,   15  ,   17  ,   49  ,   97  ,   3   ,   343 ,   515 ,   129 ,   2243    ,   1567    ,   6119    ,0 };
            const unsigned long dim1528Kuo2Init[]       =   {   1   ,   1   ,   3   ,   13  ,   3   ,   31  ,   111 ,   173 ,   97  ,   211 ,   69  ,   1771    ,   3039    ,   10305   ,0 };
            const unsigned long dim1529Kuo2Init[]       =   {   1   ,   1   ,   1   ,   11  ,   7   ,   27  ,   5   ,   163 ,   277 ,   959 ,   1287    ,   3603    ,   2617    ,   12883   ,0 };
            const unsigned long dim1530Kuo2Init[]       =   {   1   ,   1   ,   7   ,   5   ,   7   ,   63  ,   25  ,   235 ,   301 ,   575 ,   1979    ,   1519    ,   5765    ,   6331    ,0 };
            const unsigned long dim1531Kuo2Init[]       =   {   1   ,   3   ,   5   ,   7   ,   23  ,   35  ,   43  ,   171 ,   481 ,   559 ,   1495    ,   665 ,   6325    ,   2733    ,0 };
            const unsigned long dim1532Kuo2Init[]       =   {   1   ,   1   ,   7   ,   5   ,   3   ,   55  ,   99  ,   141 ,   37  ,   757 ,   11  ,   911 ,   2523    ,   11031   ,0 };
            const unsigned long dim1533Kuo2Init[]       =   {   1   ,   1   ,   5   ,   7   ,   17  ,   9   ,   91  ,   9   ,   141 ,   693 ,   581 ,   2151    ,   3527    ,   11575   ,0 };
            const unsigned long dim1534Kuo2Init[]       =   {   1   ,   1   ,   7   ,   1   ,   19  ,   31  ,   23  ,   59  ,   215 ,   873 ,   161 ,   1719    ,   509 ,   4509    ,0 };
            const unsigned long dim1535Kuo2Init[]       =   {   1   ,   1   ,   7   ,   9   ,   17  ,   29  ,   57  ,   31  ,   233 ,   109 ,   865 ,   759 ,   3289    ,   11595   ,0 };
            const unsigned long dim1536Kuo2Init[]       =   {   1   ,   3   ,   3   ,   15  ,   5   ,   61  ,   17  ,   69  ,   413 ,   95  ,   41  ,   1375    ,   1959    ,   11945   ,0 };
            const unsigned long dim1537Kuo2Init[]       =   {   1   ,   3   ,   7   ,   1   ,   25  ,   3   ,   77  ,   9   ,   301 ,   117 ,   255 ,   3169    ,   3493    ,   5503    ,0 };
            const unsigned long dim1538Kuo2Init[]       =   {   1   ,   1   ,   5   ,   3   ,   31  ,   1   ,   87  ,   211 ,   45  ,   185 ,   329 ,   3883    ,   3621    ,   4007    ,0 };
            const unsigned long dim1539Kuo2Init[]       =   {   1   ,   1   ,   5   ,   15  ,   11  ,   5   ,   123 ,   131 ,   479 ,   981 ,   921 ,   3557    ,   5897    ,   9387    ,0 };
            const unsigned long dim1540Kuo2Init[]       =   {   1   ,   1   ,   3   ,   3   ,   9   ,   43  ,   25  ,   57  ,   151 ,   975 ,   403 ,   727 ,   1155    ,   13495   ,0 };
            const unsigned long dim1541Kuo2Init[]       =   {   1   ,   3   ,   3   ,   15  ,   29  ,   59  ,   31  ,   169 ,   201 ,   1007    ,   1017    ,   3121    ,   3989    ,   7835    ,0 };
            const unsigned long dim1542Kuo2Init[]       =   {   1   ,   1   ,   3   ,   9   ,   19  ,   7   ,   103 ,   137 ,   135 ,   855 ,   179 ,   2859    ,   6959    ,   9749    ,0 };
            const unsigned long dim1543Kuo2Init[]       =   {   1   ,   3   ,   3   ,   15  ,   3   ,   3   ,   121 ,   231 ,   461 ,   511 ,   1203    ,   2105    ,   6379    ,   6809    ,0 };
            const unsigned long dim1544Kuo2Init[]       =   {   1   ,   1   ,   7   ,   15  ,   29  ,   5   ,   19  ,   215 ,   13  ,   813 ,   1421    ,   1725    ,   7809    ,   2379    ,0 };
            const unsigned long dim1545Kuo2Init[]       =   {   1   ,   3   ,   5   ,   1   ,   9   ,   55  ,   127 ,   129 ,   95  ,   449 ,   1607    ,   4043    ,   5175    ,   6063    ,0 };
            const unsigned long dim1546Kuo2Init[]       =   {   1   ,   1   ,   1   ,   9   ,   17  ,   31  ,   103 ,   107 ,   89  ,   5   ,   1747    ,   3405    ,   5739    ,   1959    ,0 };
            const unsigned long dim1547Kuo2Init[]       =   {   1   ,   3   ,   1   ,   15  ,   5   ,   43  ,   41  ,   15  ,   453 ,   691 ,   739 ,   3755    ,   6597    ,   9737    ,0 };
            const unsigned long dim1548Kuo2Init[]       =   {   1   ,   1   ,   3   ,   5   ,   27  ,   55  ,   111 ,   19  ,   349 ,   753 ,   1667    ,   3683    ,   4617    ,   12663   ,0 };
            const unsigned long dim1549Kuo2Init[]       =   {   1   ,   1   ,   5   ,   11  ,   29  ,   57  ,   111 ,   187 ,   45  ,   957 ,   2001    ,   2743    ,   2813    ,   3831    ,0 };
            const unsigned long dim1550Kuo2Init[]       =   {   1   ,   3   ,   7   ,   5   ,   13  ,   63  ,   25  ,   59  ,   169 ,   577 ,   301 ,   1535    ,   5355    ,   15835   ,0 };
            const unsigned long dim1551Kuo2Init[]       =   {   1   ,   3   ,   3   ,   7   ,   7   ,   25  ,   101 ,   249 ,   59  ,   217 ,   557 ,   2565    ,   5103    ,   13787   ,0 };
            const unsigned long dim1552Kuo2Init[]       =   {   1   ,   3   ,   7   ,   9   ,   5   ,   31  ,   97  ,   19  ,   435 ,   205 ,   1217    ,   3659    ,   7687    ,   8381    ,0 };
            const unsigned long dim1553Kuo2Init[]       =   {   1   ,   1   ,   1   ,   5   ,   7   ,   33  ,   41  ,   89  ,   243 ,   769 ,   1915    ,   2235    ,   757 ,   367 ,0 };
            const unsigned long dim1554Kuo2Init[]       =   {   1   ,   1   ,   1   ,   15  ,   3   ,   11  ,   103 ,   41  ,   79  ,   585 ,   1661    ,   1739    ,   2075    ,   15113   ,0 };
            const unsigned long dim1555Kuo2Init[]       =   {   1   ,   1   ,   5   ,   5   ,   5   ,   53  ,   103 ,   167 ,   387 ,   591 ,   317 ,   2539    ,   3763    ,   3341    ,0 };
            const unsigned long dim1556Kuo2Init[]       =   {   1   ,   1   ,   7   ,   7   ,   31  ,   29  ,   69  ,   43  ,   257 ,   471 ,   435 ,   3895    ,   445 ,   10941   ,0 };
            const unsigned long dim1557Kuo2Init[]       =   {   1   ,   1   ,   1   ,   9   ,   13  ,   63  ,   3   ,   167 ,   447 ,   569 ,   949 ,   2289    ,   6383    ,   5887    ,0 };
            const unsigned long dim1558Kuo2Init[]       =   {   1   ,   1   ,   5   ,   5   ,   11  ,   31  ,   35  ,   9   ,   383 ,   933 ,   519 ,   3773    ,   4987    ,   2737    ,0 };
            const unsigned long dim1559Kuo2Init[]       =   {   1   ,   1   ,   7   ,   7   ,   15  ,   17  ,   121 ,   109 ,   299 ,   579 ,   683 ,   1365    ,   2471    ,   14275   ,0 };
            const unsigned long dim1560Kuo2Init[]       =   {   1   ,   1   ,   1   ,   9   ,   19  ,   9   ,   17  ,   229 ,   405 ,   545 ,   1923    ,   991 ,   473 ,   11233   ,0 };
            const unsigned long dim1561Kuo2Init[]       =   {   1   ,   1   ,   3   ,   13  ,   13  ,   1   ,   101 ,   241 ,   51  ,   885 ,   1835    ,   1095    ,   2917    ,   15167   ,0 };
            const unsigned long dim1562Kuo2Init[]       =   {   1   ,   3   ,   5   ,   3   ,   31  ,   35  ,   127 ,   17  ,   153 ,   415 ,   769 ,   299 ,   7791    ,   8697    ,0 };
            const unsigned long dim1563Kuo2Init[]       =   {   1   ,   3   ,   3   ,   3   ,   23  ,   35  ,   77  ,   237 ,   427 ,   611 ,   2025    ,   3201    ,   7855    ,   789 ,0 };
            const unsigned long dim1564Kuo2Init[]       =   {   1   ,   3   ,   1   ,   5   ,   9   ,   25  ,   123 ,   215 ,   387 ,   939 ,   471 ,   1123    ,   6747    ,   14803   ,0 };
            const unsigned long dim1565Kuo2Init[]       =   {   1   ,   1   ,   5   ,   15  ,   1   ,   55  ,   85  ,   249 ,   315 ,   253 ,   1705    ,   35  ,   7351    ,   11255   ,0 };
            const unsigned long dim1566Kuo2Init[]       =   {   1   ,   3   ,   5   ,   3   ,   7   ,   15  ,   5   ,   167 ,   351 ,   145 ,   1891    ,   2691    ,   3035    ,   873 ,0 };
            const unsigned long dim1567Kuo2Init[]       =   {   1   ,   3   ,   1   ,   15  ,   19  ,   33  ,   85  ,   241 ,   259 ,   625 ,   615 ,   97  ,   5977    ,   4777    ,0 };
            const unsigned long dim1568Kuo2Init[]       =   {   1   ,   3   ,   7   ,   9   ,   19  ,   17  ,   9   ,   197 ,   141 ,   5   ,   355 ,   3001    ,   6903    ,   13809   ,0 };
            const unsigned long dim1569Kuo2Init[]       =   {   1   ,   1   ,   7   ,   13  ,   9   ,   39  ,   79  ,   43  ,   465 ,   49  ,   849 ,   1509    ,   5251    ,   11239   ,0 };
            const unsigned long dim1570Kuo2Init[]       =   {   1   ,   1   ,   7   ,   3   ,   5   ,   21  ,   37  ,   159 ,   461 ,   1015    ,   329 ,   4095    ,   2513    ,   6213    ,0 };
            const unsigned long dim1571Kuo2Init[]       =   {   1   ,   1   ,   7   ,   7   ,   7   ,   19  ,   127 ,   15  ,   61  ,   439 ,   1381    ,   677 ,   2851    ,   1635    ,0 };
            const unsigned long dim1572Kuo2Init[]       =   {   1   ,   1   ,   3   ,   5   ,   17  ,   19  ,   125 ,   171 ,   175 ,   441 ,   1153    ,   2445    ,   2907    ,   13089   ,0 };
            const unsigned long dim1573Kuo2Init[]       =   {   1   ,   1   ,   7   ,   9   ,   17  ,   59  ,   53  ,   81  ,   5   ,   685 ,   1257    ,   3179    ,   5067    ,   10131   ,0 };
            const unsigned long dim1574Kuo2Init[]       =   {   1   ,   1   ,   5   ,   3   ,   9   ,   47  ,   25  ,   111 ,   235 ,   117 ,   65  ,   2565    ,   865 ,   15173   ,0 };
            const unsigned long dim1575Kuo2Init[]       =   {   1   ,   3   ,   1   ,   7   ,   27  ,   5   ,   55  ,   155 ,   221 ,   451 ,   1823    ,   1277    ,   1397    ,   15715   ,0 };
            const unsigned long dim1576Kuo2Init[]       =   {   1   ,   1   ,   1   ,   15  ,   17  ,   5   ,   115 ,   209 ,   393 ,   973 ,   177 ,   825 ,   3315    ,   8333    ,0 };
            const unsigned long dim1577Kuo2Init[]       =   {   1   ,   1   ,   3   ,   13  ,   23  ,   47  ,   121 ,   47  ,   379 ,   433 ,   807 ,   3723    ,   2595    ,   14727   ,0 };
            const unsigned long dim1578Kuo2Init[]       =   {   1   ,   1   ,   3   ,   5   ,   29  ,   59  ,   39  ,   205 ,   327 ,   999 ,   1897    ,   3441    ,   5355    ,   12881   ,0 };
            const unsigned long dim1579Kuo2Init[]       =   {   1   ,   1   ,   3   ,   3   ,   27  ,   63  ,   57  ,   175 ,   225 ,   353 ,   675 ,   2955    ,   6581    ,   2735    ,0 };
            const unsigned long dim1580Kuo2Init[]       =   {   1   ,   3   ,   3   ,   1   ,   11  ,   55  ,   55  ,   191 ,   239 ,   579 ,   1355    ,   2449    ,   563 ,   9757    ,0 };
            const unsigned long dim1581Kuo2Init[]       =   {   1   ,   1   ,   7   ,   5   ,   1   ,   49  ,   35  ,   19  ,   25  ,   783 ,   1453    ,   1063    ,   4811    ,   1449    ,0 };
            const unsigned long dim1582Kuo2Init[]       =   {   1   ,   3   ,   1   ,   13  ,   11  ,   63  ,   17  ,   203 ,   479 ,   133 ,   1951    ,   385 ,   6803    ,   13453   ,0 };
            const unsigned long dim1583Kuo2Init[]       =   {   1   ,   3   ,   7   ,   9   ,   21  ,   19  ,   15  ,   125 ,   397 ,   109 ,   1903    ,   4053    ,   6449    ,   8821    ,0 };
            const unsigned long dim1584Kuo2Init[]       =   {   1   ,   3   ,   7   ,   15  ,   1   ,   33  ,   45  ,   161 ,   43  ,   729 ,   397 ,   4093    ,   8123    ,   3123    ,0 };
            const unsigned long dim1585Kuo2Init[]       =   {   1   ,   3   ,   7   ,   7   ,   3   ,   35  ,   3   ,   183 ,   231 ,   351 ,   1381    ,   1251    ,   3423    ,   3361    ,0 };
            const unsigned long dim1586Kuo2Init[]       =   {   1   ,   1   ,   1   ,   5   ,   23  ,   11  ,   71  ,   137 ,   323 ,   809 ,   827 ,   1573    ,   4483    ,   8691    ,0 };
            const unsigned long dim1587Kuo2Init[]       =   {   1   ,   3   ,   7   ,   7   ,   17  ,   29  ,   121 ,   47  ,   139 ,   687 ,   95  ,   1507    ,   2009    ,   1679    ,0 };
            const unsigned long dim1588Kuo2Init[]       =   {   1   ,   3   ,   7   ,   9   ,   25  ,   53  ,   127 ,   137 ,   295 ,   863 ,   299 ,   1831    ,   5359    ,   4787    ,0 };
            const unsigned long dim1589Kuo2Init[]       =   {   1   ,   3   ,   1   ,   9   ,   29  ,   3   ,   83  ,   35  ,   367 ,   675 ,   1511    ,   381 ,   6675    ,   897 ,0 };
            const unsigned long dim1590Kuo2Init[]       =   {   1   ,   3   ,   5   ,   13  ,   1   ,   15  ,   57  ,   199 ,   349 ,   873 ,   1059    ,   3767    ,   841 ,   16259   ,0 };
            const unsigned long dim1591Kuo2Init[]       =   {   1   ,   3   ,   7   ,   3   ,   13  ,   5   ,   47  ,   203 ,   211 ,   899 ,   1107    ,   743 ,   6825    ,   8703    ,0 };
            const unsigned long dim1592Kuo2Init[]       =   {   1   ,   3   ,   7   ,   11  ,   31  ,   41  ,   11  ,   59  ,   59  ,   859 ,   787 ,   977 ,   2223    ,   15921   ,0 };
            const unsigned long dim1593Kuo2Init[]       =   {   1   ,   3   ,   5   ,   5   ,   31  ,   11  ,   39  ,   105 ,   109 ,   487 ,   1459    ,   2935    ,   3723    ,   4697    ,0 };
            const unsigned long dim1594Kuo2Init[]       =   {   1   ,   3   ,   3   ,   13  ,   9   ,   17  ,   39  ,   239 ,   183 ,   885 ,   123 ,   387 ,   2369    ,   16179   ,0 };
            const unsigned long dim1595Kuo2Init[]       =   {   1   ,   1   ,   1   ,   7   ,   25  ,   47  ,   23  ,   29  ,   95  ,   331 ,   2005    ,   633 ,   649 ,   1623    ,0 };
            const unsigned long dim1596Kuo2Init[]       =   {   1   ,   1   ,   3   ,   13  ,   9   ,   21  ,   41  ,   125 ,   287 ,   309 ,   1421    ,   1637    ,   2611    ,   2259    ,0 };
            const unsigned long dim1597Kuo2Init[]       =   {   1   ,   1   ,   3   ,   3   ,   11  ,   27  ,   117 ,   93  ,   419 ,   521 ,   1699    ,   233 ,   1967    ,   14147   ,0 };
            const unsigned long dim1598Kuo2Init[]       =   {   1   ,   1   ,   3   ,   11  ,   25  ,   21  ,   61  ,   15  ,   415 ,   693 ,   1685    ,   1727    ,   961 ,   3887    ,0 };
            const unsigned long dim1599Kuo2Init[]       =   {   1   ,   3   ,   7   ,   9   ,   15  ,   39  ,   41  ,   19  ,   237 ,   231 ,   1605    ,   2641    ,   6047    ,   10235   ,0 };
            const unsigned long dim1600Kuo2Init[]       =   {   1   ,   1   ,   3   ,   9   ,   5   ,   49  ,   113 ,   133 ,   499 ,   545 ,   467 ,   683 ,   579 ,   8123    ,0 };
            const unsigned long dim1601Kuo2Init[]       =   {   1   ,   3   ,   5   ,   13  ,   11  ,   31  ,   121 ,   229 ,   225 ,   135 ,   85  ,   1389    ,   5317    ,   13453   ,0 };
            const unsigned long dim1602Kuo2Init[]       =   {   1   ,   3   ,   7   ,   9   ,   17  ,   51  ,   77  ,   141 ,   413 ,   805 ,   1759    ,   1139    ,   5847    ,   16153   ,0 };
            const unsigned long dim1603Kuo2Init[]       =   {   1   ,   3   ,   5   ,   7   ,   23  ,   29  ,   79  ,   235 ,   113 ,   981 ,   1147    ,   219 ,   2609    ,   1089    ,0 };
            const unsigned long dim1604Kuo2Init[]       =   {   1   ,   1   ,   5   ,   7   ,   5   ,   45  ,   123 ,   123 ,   335 ,   409 ,   115 ,   971 ,   827 ,   427 ,0 };
            const unsigned long dim1605Kuo2Init[]       =   {   1   ,   3   ,   7   ,   13  ,   13  ,   45  ,   41  ,   173 ,   261 ,   151 ,   1199    ,   2803    ,   5801    ,   2449    ,0 };
            const unsigned long dim1606Kuo2Init[]       =   {   1   ,   3   ,   1   ,   7   ,   5   ,   17  ,   9   ,   91  ,   475 ,   505 ,   1983    ,   2539    ,   339 ,   16365   ,0 };
            const unsigned long dim1607Kuo2Init[]       =   {   1   ,   1   ,   7   ,   5   ,   25  ,   31  ,   77  ,   53  ,   293 ,   735 ,   1423    ,   165 ,   4363    ,   5923    ,0 };
            const unsigned long dim1608Kuo2Init[]       =   {   1   ,   1   ,   1   ,   15  ,   7   ,   51  ,   117 ,   91  ,   347 ,   677 ,   1059    ,   2861    ,   3583    ,   11995   ,0 };
            const unsigned long dim1609Kuo2Init[]       =   {   1   ,   3   ,   3   ,   13  ,   27  ,   27  ,   123 ,   251 ,   47  ,   885 ,   2015    ,   1089    ,   4077    ,   8591    ,0 };
            const unsigned long dim1610Kuo2Init[]       =   {   1   ,   1   ,   7   ,   11  ,   17  ,   41  ,   97  ,   47  ,   201 ,   789 ,   881 ,   1565    ,   1169    ,   10967   ,0 };
            const unsigned long dim1611Kuo2Init[]       =   {   1   ,   1   ,   3   ,   15  ,   25  ,   43  ,   45  ,   237 ,   73  ,   909 ,   1041    ,   1025    ,   3217    ,   9347    ,0 };
            const unsigned long dim1612Kuo2Init[]       =   {   1   ,   1   ,   1   ,   3   ,   7   ,   59  ,   51  ,   177 ,   223 ,   125 ,   63  ,   3311    ,   3171    ,   8725    ,0 };
            const unsigned long dim1613Kuo2Init[]       =   {   1   ,   1   ,   1   ,   1   ,   31  ,   25  ,   55  ,   117 ,   493 ,   419 ,   1749    ,   2799    ,   6943    ,   12427   ,0 };
            const unsigned long dim1614Kuo2Init[]       =   {   1   ,   3   ,   5   ,   13  ,   1   ,   59  ,   123 ,   239 ,   239 ,   589 ,   667 ,   1029    ,   1357    ,   6801    ,0 };
            const unsigned long dim1615Kuo2Init[]       =   {   1   ,   1   ,   7   ,   15  ,   21  ,   7   ,   43  ,   103 ,   159 ,   685 ,   1435    ,   529 ,   4225    ,   5051    ,0 };
            const unsigned long dim1616Kuo2Init[]       =   {   1   ,   3   ,   1   ,   13  ,   29  ,   21  ,   19  ,   229 ,   369 ,   317 ,   1681    ,   685 ,   5301    ,   11107   ,0 };
            const unsigned long dim1617Kuo2Init[]       =   {   1   ,   3   ,   7   ,   9   ,   17  ,   13  ,   117 ,   21  ,   203 ,   123 ,   1933    ,   3981    ,   5133    ,   11729   ,0 };
            const unsigned long dim1618Kuo2Init[]       =   {   1   ,   3   ,   3   ,   1   ,   29  ,   5   ,   119 ,   27  ,   135 ,   913 ,   1891    ,   2785    ,   6311    ,   8439    ,0 };
            const unsigned long dim1619Kuo2Init[]       =   {   1   ,   3   ,   7   ,   5   ,   3   ,   31  ,   95  ,   55  ,   89  ,   541 ,   609 ,   773 ,   6513    ,   10015   ,0 };
            const unsigned long dim1620Kuo2Init[]       =   {   1   ,   3   ,   7   ,   1   ,   25  ,   27  ,   71  ,   153 ,   215 ,   921 ,   29  ,   2103    ,   1605    ,   59  ,0 };
            const unsigned long dim1621Kuo2Init[]       =   {   1   ,   1   ,   3   ,   5   ,   15  ,   31  ,   65  ,   35  ,   221 ,   571 ,   873 ,   1271    ,   5399    ,   15659   ,0 };
            const unsigned long dim1622Kuo2Init[]       =   {   1   ,   3   ,   5   ,   3   ,   31  ,   31  ,   127 ,   251 ,   379 ,   801 ,   905 ,   1661    ,   225 ,   11939   ,0 };
            const unsigned long dim1623Kuo2Init[]       =   {   1   ,   3   ,   1   ,   15  ,   31  ,   59  ,   51  ,   117 ,   325 ,   491 ,   1801    ,   669 ,   2867    ,   5325    ,0 };
            const unsigned long dim1624Kuo2Init[]       =   {   1   ,   1   ,   5   ,   13  ,   1   ,   15  ,   3   ,   105 ,   81  ,   229 ,   573 ,   3047    ,   6373    ,   11947   ,0 };
            const unsigned long dim1625Kuo2Init[]       =   {   1   ,   1   ,   3   ,   9   ,   7   ,   51  ,   83  ,   103 ,   277 ,   5   ,   1189    ,   89  ,   7173    ,   1617    ,0 };
            const unsigned long dim1626Kuo2Init[]       =   {   1   ,   1   ,   5   ,   3   ,   11  ,   13  ,   5   ,   205 ,   179 ,   23  ,   1327    ,   3049    ,   23  ,   5123    ,0 };
            const unsigned long dim1627Kuo2Init[]       =   {   1   ,   3   ,   5   ,   7   ,   7   ,   23  ,   117 ,   133 ,   449 ,   265 ,   1087    ,   1565    ,   7185    ,   7099    ,0 };
            const unsigned long dim1628Kuo2Init[]       =   {   1   ,   1   ,   3   ,   5   ,   7   ,   13  ,   31  ,   197 ,   29  ,   659 ,   1485    ,   2395    ,   3239    ,   5837    ,0 };
            const unsigned long dim1629Kuo2Init[]       =   {   1   ,   3   ,   5   ,   13  ,   15  ,   53  ,   67  ,   107 ,   103 ,   675 ,   253 ,   471 ,   283 ,   1221    ,0 };
            const unsigned long dim1630Kuo2Init[]       =   {   1   ,   1   ,   5   ,   11  ,   17  ,   17  ,   65  ,   165 ,   141 ,   839 ,   217 ,   3779    ,   1151    ,   3551    ,0 };
            const unsigned long dim1631Kuo2Init[]       =   {   1   ,   1   ,   7   ,   13  ,   13  ,   23  ,   57  ,   241 ,   253 ,   565 ,   371 ,   3713    ,   4539    ,   1033    ,0 };
            const unsigned long dim1632Kuo2Init[]       =   {   1   ,   1   ,   7   ,   11  ,   17  ,   11  ,   121 ,   193 ,   9   ,   33  ,   695 ,   1361    ,   6433    ,   895 ,0 };
            const unsigned long dim1633Kuo2Init[]       =   {   1   ,   1   ,   1   ,   3   ,   15  ,   11  ,   87  ,   87  ,   419 ,   605 ,   163 ,   377 ,   3729    ,   4309    ,0 };
            const unsigned long dim1634Kuo2Init[]       =   {   1   ,   3   ,   5   ,   3   ,   17  ,   43  ,   13  ,   99  ,   409 ,   919 ,   1347    ,   1407    ,   1723    ,   11317   ,0 };
            const unsigned long dim1635Kuo2Init[]       =   {   1   ,   1   ,   1   ,   9   ,   1   ,   21  ,   111 ,   225 ,   191 ,   65  ,   1231    ,   27  ,   387 ,   12225   ,0 };
            const unsigned long dim1636Kuo2Init[]       =   {   1   ,   1   ,   3   ,   11  ,   19  ,   63  ,   61  ,   103 ,   31  ,   241 ,   1807    ,   3101    ,   7275    ,   2333    ,0 };
            const unsigned long dim1637Kuo2Init[]       =   {   1   ,   1   ,   7   ,   5   ,   31  ,   41  ,   115 ,   77  ,   347 ,   85  ,   69  ,   1911    ,   7345    ,   15269   ,0 };
            const unsigned long dim1638Kuo2Init[]       =   {   1   ,   1   ,   7   ,   1   ,   25  ,   41  ,   105 ,   5   ,   161 ,   721 ,   1667    ,   1285    ,   6197    ,   14823   ,0 };
            const unsigned long dim1639Kuo2Init[]       =   {   1   ,   3   ,   3   ,   9   ,   15  ,   33  ,   43  ,   155 ,   269 ,   215 ,   663 ,   3425    ,   7725    ,   15307   ,0 };
            const unsigned long dim1640Kuo2Init[]       =   {   1   ,   3   ,   5   ,   15  ,   23  ,   43  ,   71  ,   79  ,   163 ,   167 ,   1207    ,   741 ,   141 ,   8269    ,0 };
            const unsigned long dim1641Kuo2Init[]       =   {   1   ,   1   ,   7   ,   9   ,   21  ,   21  ,   39  ,   181 ,   379 ,   383 ,   343 ,   1685    ,   1713    ,   10377   ,0 };
            const unsigned long dim1642Kuo2Init[]       =   {   1   ,   3   ,   3   ,   7   ,   3   ,   17  ,   45  ,   11  ,   351 ,   111 ,   229 ,   3241    ,   6953    ,   7143    ,0 };
            const unsigned long dim1643Kuo2Init[]       =   {   1   ,   1   ,   7   ,   13  ,   15  ,   57  ,   87  ,   247 ,   361 ,   943 ,   801 ,   1861    ,   5705    ,   14499   ,0 };
            const unsigned long dim1644Kuo2Init[]       =   {   1   ,   1   ,   5   ,   7   ,   11  ,   9   ,   89  ,   235 ,   235 ,   225 ,   1053    ,   2909    ,   993 ,   15471   ,0 };
            const unsigned long dim1645Kuo2Init[]       =   {   1   ,   3   ,   3   ,   11  ,   15  ,   47  ,   73  ,   13  ,   477 ,   703 ,   1353    ,   3633    ,   4059    ,   16367   ,0 };
            const unsigned long dim1646Kuo2Init[]       =   {   1   ,   1   ,   5   ,   15  ,   9   ,   39  ,   49  ,   141 ,   181 ,   917 ,   1331    ,   427 ,   7455    ,   13753   ,0 };
            const unsigned long dim1647Kuo2Init[]       =   {   1   ,   3   ,   1   ,   13  ,   11  ,   7   ,   113 ,   101 ,   473 ,   851 ,   771 ,   1287    ,   8157    ,   2113    ,0 };
            const unsigned long dim1648Kuo2Init[]       =   {   1   ,   3   ,   1   ,   15  ,   23  ,   33  ,   87  ,   5   ,   245 ,   853 ,   1433    ,   85  ,   5655    ,   11579   ,0 };
            const unsigned long dim1649Kuo2Init[]       =   {   1   ,   1   ,   5   ,   15  ,   19  ,   59  ,   103 ,   249 ,   247 ,   589 ,   797 ,   2733    ,   6603    ,   2065    ,0 };
            const unsigned long dim1650Kuo2Init[]       =   {   1   ,   1   ,   1   ,   7   ,   25  ,   63  ,   105 ,   183 ,   497 ,   25  ,   739 ,   521 ,   7327    ,   12711   ,0 };
            const unsigned long dim1651Kuo2Init[]       =   {   1   ,   3   ,   3   ,   7   ,   27  ,   13  ,   63  ,   23  ,   417 ,   303 ,   155 ,   2227    ,   5621    ,   1643    ,0 };
            const unsigned long dim1652Kuo2Init[]       =   {   1   ,   1   ,   5   ,   9   ,   29  ,   13  ,   75  ,   199 ,   353 ,   577 ,   1715    ,   3939    ,   2659    ,   5969    ,0 };
            const unsigned long dim1653Kuo2Init[]       =   {   1   ,   1   ,   5   ,   11  ,   11  ,   37  ,   55  ,   185 ,   435 ,   763 ,   333 ,   3243    ,   7553    ,   9741    ,0 };
            const unsigned long dim1654Kuo2Init[]       =   {   1   ,   3   ,   7   ,   13  ,   7   ,   1   ,   65  ,   243 ,   443 ,   869 ,   769 ,   4027    ,   4831    ,   7333    ,0 };
            const unsigned long dim1655Kuo2Init[]       =   {   1   ,   1   ,   5   ,   11  ,   31  ,   23  ,   93  ,   117 ,   323 ,   157 ,   811 ,   1543    ,   449 ,   8371    ,0 };
            const unsigned long dim1656Kuo2Init[]       =   {   1   ,   3   ,   1   ,   9   ,   27  ,   25  ,   87  ,   223 ,   467 ,   761 ,   547 ,   2381    ,   411 ,   14035   ,0 };
            const unsigned long dim1657Kuo2Init[]       =   {   1   ,   1   ,   7   ,   9   ,   31  ,   27  ,   127 ,   11  ,   243 ,   263 ,   255 ,   831 ,   6801    ,   10923   ,0 };
            const unsigned long dim1658Kuo2Init[]       =   {   1   ,   3   ,   5   ,   15  ,   23  ,   11  ,   77  ,   5   ,   255 ,   699 ,   13  ,   3883    ,   4703    ,   2785    ,0 };
            const unsigned long dim1659Kuo2Init[]       =   {   1   ,   3   ,   1   ,   1   ,   25  ,   37  ,   47  ,   75  ,   79  ,   881 ,   1505    ,   2825    ,   2477    ,   1963    ,0 };
            const unsigned long dim1660Kuo2Init[]       =   {   1   ,   1   ,   3   ,   7   ,   29  ,   39  ,   33  ,   151 ,   153 ,   869 ,   65  ,   107 ,   7173    ,   5059    ,0 };
            const unsigned long dim1661Kuo2Init[]       =   {   1   ,   1   ,   5   ,   13  ,   29  ,   41  ,   61  ,   37  ,   353 ,   429 ,   587 ,   2431    ,   3711    ,   14149   ,0 };
            const unsigned long dim1662Kuo2Init[]       =   {   1   ,   1   ,   3   ,   7   ,   13  ,   23  ,   29  ,   157 ,   427 ,   353 ,   1485    ,   879 ,   3851    ,   15775   ,0 };
            const unsigned long dim1663Kuo2Init[]       =   {   1   ,   1   ,   5   ,   11  ,   7   ,   17  ,   91  ,   225 ,   275 ,   923 ,   1383    ,   2587    ,   7545    ,   7203    ,0 };
            const unsigned long dim1664Kuo2Init[]       =   {   1   ,   1   ,   5   ,   3   ,   3   ,   57  ,   75  ,   97  ,   233 ,   233 ,   1547    ,   1915    ,   1229    ,   8559    ,0 };
            const unsigned long dim1665Kuo2Init[]       =   {   1   ,   3   ,   3   ,   11  ,   11  ,   63  ,   15  ,   121 ,   209 ,   31  ,   171 ,   1117    ,   2373    ,   7545    ,0 };
            const unsigned long dim1666Kuo2Init[]       =   {   1   ,   1   ,   1   ,   3   ,   27  ,   7   ,   89  ,   155 ,   7   ,   829 ,   537 ,   3639    ,   1721    ,   8199    ,0 };
            const unsigned long dim1667Kuo2Init[]       =   {   1   ,   1   ,   1   ,   1   ,   5   ,   43  ,   59  ,   245 ,   321 ,   869 ,   785 ,   367 ,   2693    ,   8603    ,0 };
            const unsigned long dim1668Kuo2Init[]       =   {   1   ,   1   ,   3   ,   1   ,   23  ,   17  ,   19  ,   79  ,   303 ,   919 ,   1119    ,   303 ,   907 ,   9033    ,0 };
            const unsigned long dim1669Kuo2Init[]       =   {   1   ,   3   ,   1   ,   15  ,   17  ,   43  ,   1   ,   207 ,   395 ,   251 ,   727 ,   1341    ,   2073    ,   16319   ,0 };
            const unsigned long dim1670Kuo2Init[]       =   {   1   ,   3   ,   3   ,   15  ,   3   ,   31  ,   97  ,   111 ,   161 ,   947 ,   1509    ,   2143    ,   4773    ,   3569    ,0 };
            const unsigned long dim1671Kuo2Init[]       =   {   1   ,   3   ,   7   ,   3   ,   5   ,   33  ,   19  ,   149 ,   233 ,   221 ,   517 ,   21  ,   7651    ,   14987   ,0 };
            const unsigned long dim1672Kuo2Init[]       =   {   1   ,   3   ,   3   ,   7   ,   19  ,   1   ,   39  ,   253 ,   311 ,   681 ,   897 ,   2777    ,   5181    ,   6985    ,0 };
            const unsigned long dim1673Kuo2Init[]       =   {   1   ,   3   ,   5   ,   13  ,   3   ,   39  ,   75  ,   225 ,   491 ,   631 ,   513 ,   935 ,   2775    ,   6405    ,0 };
            const unsigned long dim1674Kuo2Init[]       =   {   1   ,   3   ,   5   ,   9   ,   29  ,   57  ,   67  ,   67  ,   53  ,   955 ,   1297    ,   3871    ,   131 ,   7317    ,0 };
            const unsigned long dim1675Kuo2Init[]       =   {   1   ,   3   ,   3   ,   9   ,   31  ,   55  ,   53  ,   179 ,   305 ,   625 ,   1097    ,   1203    ,   1973    ,   2807    ,0 };
            const unsigned long dim1676Kuo2Init[]       =   {   1   ,   1   ,   1   ,   5   ,   5   ,   47  ,   25  ,   7   ,   279 ,   751 ,   1149    ,   2871    ,   2985    ,   11941   ,0 };
            const unsigned long dim1677Kuo2Init[]       =   {   1   ,   3   ,   7   ,   5   ,   21  ,   47  ,   29  ,   23  ,   419 ,   181 ,   517 ,   1261    ,   2329    ,   11667   ,0 };
            const unsigned long dim1678Kuo2Init[]       =   {   1   ,   1   ,   1   ,   3   ,   5   ,   53  ,   61  ,   1   ,   273 ,   165 ,   1227    ,   2901    ,   2739    ,   2453    ,0 };
            const unsigned long dim1679Kuo2Init[]       =   {   1   ,   3   ,   1   ,   7   ,   1   ,   21  ,   89  ,   27  ,   487 ,   167 ,   717 ,   3653    ,   3971    ,   7825    ,0 };
            const unsigned long dim1680Kuo2Init[]       =   {   1   ,   1   ,   3   ,   5   ,   3   ,   43  ,   19  ,   221 ,   21  ,   329 ,   1415    ,   1137    ,   1073    ,   2869    ,0 };
            const unsigned long dim1681Kuo2Init[]       =   {   1   ,   1   ,   7   ,   13  ,   27  ,   61  ,   65  ,   93  ,   469 ,   975 ,   5   ,   3919    ,   7565    ,   13885   ,0 };
            const unsigned long dim1682Kuo2Init[]       =   {   1   ,   1   ,   5   ,   9   ,   9   ,   9   ,   111 ,   97  ,   35  ,   595 ,   997 ,   797 ,   5963    ,   12977   ,0 };
            const unsigned long dim1683Kuo2Init[]       =   {   1   ,   3   ,   7   ,   11  ,   5   ,   1   ,   125 ,   169 ,   393 ,   127 ,   1297    ,   3853    ,   1177    ,   10553   ,0 };
            const unsigned long dim1684Kuo2Init[]       =   {   1   ,   1   ,   3   ,   13  ,   11  ,   53  ,   55  ,   119 ,   41  ,   895 ,   1379    ,   1955    ,   7059    ,   143 ,0 };
            const unsigned long dim1685Kuo2Init[]       =   {   1   ,   3   ,   1   ,   7   ,   1   ,   9   ,   49  ,   165 ,   479 ,   247 ,   1555    ,   2199    ,   6695    ,   8983    ,0 };
            const unsigned long dim1686Kuo2Init[]       =   {   1   ,   3   ,   3   ,   3   ,   17  ,   53  ,   47  ,   97  ,   301 ,   649 ,   1121    ,   7   ,   1335    ,   11535   ,0 };
            const unsigned long dim1687Kuo2Init[]       =   {   1   ,   3   ,   1   ,   1   ,   25  ,   49  ,   5   ,   227 ,   335 ,   401 ,   651 ,   3089    ,   823 ,   3159    ,0 };
            const unsigned long dim1688Kuo2Init[]       =   {   1   ,   1   ,   5   ,   7   ,   7   ,   25  ,   79  ,   15  ,   79  ,   381 ,   991 ,   2319    ,   6087    ,   5919    ,0 };
            const unsigned long dim1689Kuo2Init[]       =   {   1   ,   3   ,   7   ,   3   ,   17  ,   53  ,   99  ,   237 ,   305 ,   411 ,   1751    ,   753 ,   4785    ,   57  ,0 };
            const unsigned long dim1690Kuo2Init[]       =   {   1   ,   1   ,   5   ,   7   ,   15  ,   27  ,   91  ,   159 ,   451 ,   877 ,   715 ,   1183    ,   3619    ,   4521    ,0 };
            const unsigned long dim1691Kuo2Init[]       =   {   1   ,   3   ,   7   ,   5   ,   25  ,   9   ,   9   ,   1   ,   421 ,   707 ,   1619    ,   1423    ,   7183    ,   687 ,0 };
            const unsigned long dim1692Kuo2Init[]       =   {   1   ,   3   ,   7   ,   3   ,   3   ,   49  ,   79  ,   249 ,   299 ,   559 ,   1619    ,   2071    ,   8135    ,   1975    ,0 };
            const unsigned long dim1693Kuo2Init[]       =   {   1   ,   3   ,   1   ,   9   ,   29  ,   63  ,   67  ,   93  ,   373 ,   505 ,   193 ,   791 ,   5761    ,   8329    ,0 };
            const unsigned long dim1694Kuo2Init[]       =   {   1   ,   1   ,   3   ,   3   ,   7   ,   33  ,   33  ,   233 ,   507 ,   471 ,   1369    ,   3355    ,   27  ,   6389    ,0 };
            const unsigned long dim1695Kuo2Init[]       =   {   1   ,   3   ,   3   ,   9   ,   21  ,   13  ,   25  ,   249 ,   39  ,   305 ,   827 ,   3559    ,   937 ,   6079    ,0 };
            const unsigned long dim1696Kuo2Init[]       =   {   1   ,   1   ,   1   ,   13  ,   25  ,   29  ,   115 ,   165 ,   65  ,   217 ,   1519    ,   3117    ,   3211    ,   7211    ,0 };
            const unsigned long dim1697Kuo2Init[]       =   {   1   ,   1   ,   1   ,   3   ,   21  ,   17  ,   55  ,   117 ,   61  ,   807 ,   1821    ,   703 ,   3403    ,   14181   ,0 };
            const unsigned long dim1698Kuo2Init[]       =   {   1   ,   3   ,   5   ,   1   ,   3   ,   5   ,   65  ,   199 ,   399 ,   695 ,   1527    ,   1187    ,   7887    ,   15437   ,0 };
            const unsigned long dim1699Kuo2Init[]       =   {   1   ,   3   ,   1   ,   3   ,   23  ,   47  ,   43  ,   49  ,   377 ,   893 ,   423 ,   3745    ,   4291    ,   2579    ,0 };
            const unsigned long dim1700Kuo2Init[]       =   {   1   ,   3   ,   3   ,   13  ,   15  ,   13  ,   19  ,   235 ,   209 ,   963 ,   1331    ,   1559    ,   5631    ,   2411    ,0 };
            const unsigned long dim1701Kuo2Init[]       =   {   1   ,   3   ,   5   ,   1   ,   3   ,   47  ,   5   ,   179 ,   483 ,   163 ,   123 ,   1241    ,   5767    ,   4361    ,0 };
            const unsigned long dim1702Kuo2Init[]       =   {   1   ,   3   ,   5   ,   13  ,   15  ,   25  ,   123 ,   13  ,   313 ,   959 ,   1845    ,   3949    ,   7395    ,   5727    ,0 };
            const unsigned long dim1703Kuo2Init[]       =   {   1   ,   3   ,   7   ,   7   ,   21  ,   35  ,   67  ,   209 ,   229 ,   213 ,   1403    ,   561 ,   4229    ,   13611   ,0 };
            const unsigned long dim1704Kuo2Init[]       =   {   1   ,   3   ,   5   ,   15  ,   17  ,   45  ,   93  ,   5   ,   493 ,   779 ,   1357    ,   243 ,   2067    ,   4621    ,0 };
            const unsigned long dim1705Kuo2Init[]       =   {   1   ,   3   ,   7   ,   1   ,   19  ,   29  ,   11  ,   219 ,   247 ,   133 ,   443 ,   1331    ,   5027    ,   6953    ,0 };
            const unsigned long dim1706Kuo2Init[]       =   {   1   ,   1   ,   7   ,   5   ,   29  ,   41  ,   21  ,   95  ,   175 ,   761 ,   1425    ,   2471    ,   4597    ,   7071    ,0 };
            const unsigned long dim1707Kuo2Init[]       =   {   1   ,   3   ,   7   ,   15  ,   25  ,   5   ,   49  ,   153 ,   415 ,   703 ,   321 ,   3727    ,   4541    ,   9711    ,0 };
            const unsigned long dim1708Kuo2Init[]       =   {   1   ,   3   ,   5   ,   5   ,   21  ,   49  ,   15  ,   41  ,   203 ,   909 ,   953 ,   1587    ,   6323    ,   12589   ,0 };
            const unsigned long dim1709Kuo2Init[]       =   {   1   ,   3   ,   3   ,   7   ,   29  ,   21  ,   59  ,   245 ,   287 ,   411 ,   1211    ,   2527    ,   5247    ,   11691   ,0 };
            const unsigned long dim1710Kuo2Init[]       =   {   1   ,   1   ,   3   ,   15  ,   5   ,   29  ,   25  ,   123 ,   29  ,   613 ,   627 ,   269 ,   6479    ,   817 ,0 };
            const unsigned long dim1711Kuo2Init[]       =   {   1   ,   3   ,   5   ,   3   ,   23  ,   11  ,   15  ,   177 ,   421 ,   803 ,   273 ,   3145    ,   7211    ,   7141    ,0 };
            const unsigned long dim1712Kuo2Init[]       =   {   1   ,   1   ,   7   ,   13  ,   13  ,   37  ,   59  ,   81  ,   57  ,   349 ,   833 ,   871 ,   1343    ,   8911    ,0 };
            const unsigned long dim1713Kuo2Init[]       =   {   1   ,   3   ,   5   ,   3   ,   23  ,   11  ,   33  ,   39  ,   123 ,   77  ,   259 ,   2667    ,   2121    ,   203 ,0 };
            const unsigned long dim1714Kuo2Init[]       =   {   1   ,   3   ,   7   ,   5   ,   5   ,   57  ,   31  ,   101 ,   463 ,   633 ,   1351    ,   2209    ,   5235    ,   8591    ,0 };
            const unsigned long dim1715Kuo2Init[]       =   {   1   ,   1   ,   7   ,   9   ,   1   ,   35  ,   39  ,   121 ,   155 ,   847 ,   1355    ,   2161    ,   3073    ,   5873    ,0 };
            const unsigned long dim1716Kuo2Init[]       =   {   1   ,   3   ,   7   ,   11  ,   19  ,   23  ,   29  ,   233 ,   501 ,   321 ,   1131    ,   641 ,   4023    ,   11955   ,0 };
            const unsigned long dim1717Kuo2Init[]       =   {   1   ,   3   ,   3   ,   7   ,   13  ,   39  ,   119 ,   157 ,   261 ,   167 ,   511 ,   2349    ,   1489    ,   11523   ,0 };
            const unsigned long dim1718Kuo2Init[]       =   {   1   ,   3   ,   5   ,   15  ,   27  ,   3   ,   19  ,   165 ,   229 ,   591 ,   1975    ,   3411    ,   4453    ,   3291    ,0 };
            const unsigned long dim1719Kuo2Init[]       =   {   1   ,   3   ,   5   ,   15  ,   11  ,   41  ,   61  ,   117 ,   505 ,   347 ,   957 ,   2545    ,   1123    ,   16229   ,0 };
            const unsigned long dim1720Kuo2Init[]       =   {   1   ,   1   ,   5   ,   15  ,   1   ,   21  ,   101 ,   123 ,   341 ,   613 ,   451 ,   3731    ,   8031    ,   1611    ,0 };
            const unsigned long dim1721Kuo2Init[]       =   {   1   ,   1   ,   5   ,   11  ,   27  ,   1   ,   53  ,   69  ,   509 ,   751 ,   407 ,   2777    ,   1319    ,   10897   ,0 };
            const unsigned long dim1722Kuo2Init[]       =   {   1   ,   3   ,   5   ,   7   ,   7   ,   5   ,   127 ,   91  ,   17  ,   699 ,   1761    ,   2083    ,   4587    ,   10747   ,0 };
            const unsigned long dim1723Kuo2Init[]       =   {   1   ,   3   ,   5   ,   11  ,   1   ,   57  ,   107 ,   245 ,   89  ,   899 ,   1765    ,   3521    ,   4933    ,   721 ,0 };
            const unsigned long dim1724Kuo2Init[]       =   {   1   ,   1   ,   1   ,   1   ,   27  ,   49  ,   113 ,   201 ,   491 ,   921 ,   417 ,   3541    ,   5889    ,   8417    ,0 };
            const unsigned long dim1725Kuo2Init[]       =   {   1   ,   3   ,   5   ,   11  ,   1   ,   37  ,   5   ,   43  ,   337 ,   883 ,   1351    ,   351 ,   269 ,   5501    ,0 };
            const unsigned long dim1726Kuo2Init[]       =   {   1   ,   3   ,   5   ,   5   ,   9   ,   11  ,   127 ,   87  ,   463 ,   999 ,   635 ,   2355    ,   763 ,   11959   ,0 };
            const unsigned long dim1727Kuo2Init[]       =   {   1   ,   3   ,   1   ,   15  ,   25  ,   21  ,   111 ,   241 ,   31  ,   35  ,   1869    ,   2913    ,   5963    ,   9989    ,0 };
            const unsigned long dim1728Kuo2Init[]       =   {   1   ,   1   ,   7   ,   13  ,   23  ,   57  ,   7   ,   225 ,   115 ,   649 ,   395 ,   515 ,   5335    ,   10829   ,0 };
            const unsigned long dim1729Kuo2Init[]       =   {   1   ,   3   ,   5   ,   13  ,   21  ,   21  ,   51  ,   1   ,   39  ,   927 ,   681 ,   1341    ,   7703    ,   631 ,0 };
            const unsigned long dim1730Kuo2Init[]       =   {   1   ,   1   ,   3   ,   13  ,   25  ,   35  ,   101 ,   215 ,   15  ,   207 ,   137 ,   409 ,   6239    ,   11609   ,0 };
            const unsigned long dim1731Kuo2Init[]       =   {   1   ,   3   ,   1   ,   15  ,   11  ,   47  ,   17  ,   131 ,   405 ,   633 ,   35  ,   3335    ,   821 ,   4681    ,0 };
            const unsigned long dim1732Kuo2Init[]       =   {   1   ,   1   ,   1   ,   9   ,   17  ,   17  ,   105 ,   81  ,   449 ,   421 ,   707 ,   1041    ,   3991    ,   2943    ,0 };
            const unsigned long dim1733Kuo2Init[]       =   {   1   ,   1   ,   1   ,   1   ,   29  ,   55  ,   43  ,   161 ,   449 ,   163 ,   1295    ,   2583    ,   1405    ,   14211   ,0 };
            const unsigned long dim1734Kuo2Init[]       =   {   1   ,   1   ,   1   ,   7   ,   15  ,   41  ,   125 ,   65  ,   347 ,   753 ,   1245    ,   2343    ,   733 ,   2269    ,0 };
            const unsigned long dim1735Kuo2Init[]       =   {   1   ,   3   ,   7   ,   5   ,   15  ,   25  ,   107 ,   35  ,   369 ,   395 ,   939 ,   427 ,   5201    ,   15761   ,0 };
            const unsigned long dim1736Kuo2Init[]       =   {   1   ,   1   ,   3   ,   13  ,   29  ,   11  ,   9   ,   53  ,   393 ,   855 ,   483 ,   3773    ,   3491    ,   14857   ,0 };
            const unsigned long dim1737Kuo2Init[]       =   {   1   ,   1   ,   5   ,   11  ,   25  ,   55  ,   25  ,   9   ,   161 ,   307 ,   1381    ,   2129    ,   1923    ,   8779    ,0 };
            const unsigned long dim1738Kuo2Init[]       =   {   1   ,   1   ,   5   ,   1   ,   5   ,   1   ,   99  ,   205 ,   207 ,   607 ,   639 ,   29  ,   8045    ,   7159    ,0 };
            const unsigned long dim1739Kuo2Init[]       =   {   1   ,   1   ,   3   ,   7   ,   29  ,   55  ,   15  ,   133 ,   217 ,   425 ,   1331    ,   2337    ,   2209    ,   1839    ,0 };
            const unsigned long dim1740Kuo2Init[]       =   {   1   ,   3   ,   5   ,   1   ,   25  ,   1   ,   47  ,   135 ,   433 ,   647 ,   1693    ,   3983    ,   975 ,   14725   ,0 };
            const unsigned long dim1741Kuo2Init[]       =   {   1   ,   3   ,   7   ,   13  ,   29  ,   41  ,   115 ,   227 ,   505 ,   17  ,   2041    ,   247 ,   7409    ,   12353   ,0 };
            const unsigned long dim1742Kuo2Init[]       =   {   1   ,   1   ,   1   ,   13  ,   19  ,   29  ,   25  ,   175 ,   189 ,   581 ,   877 ,   2775    ,   737 ,   223 ,0 };
            const unsigned long dim1743Kuo2Init[]       =   {   1   ,   3   ,   7   ,   11  ,   27  ,   61  ,   21  ,   209 ,   51  ,   653 ,   631 ,   3723    ,   7471    ,   11083   ,0 };
            const unsigned long dim1744Kuo2Init[]       =   {   1   ,   3   ,   1   ,   9   ,   9   ,   41  ,   33  ,   61  ,   97  ,   551 ,   1029    ,   663 ,   5373    ,   8443    ,0 };
            const unsigned long dim1745Kuo2Init[]       =   {   1   ,   3   ,   3   ,   9   ,   9   ,   39  ,   85  ,   103 ,   95  ,   865 ,   269 ,   1243    ,   5709    ,   3969    ,0 };
            const unsigned long dim1746Kuo2Init[]       =   {   1   ,   3   ,   5   ,   15  ,   9   ,   17  ,   75  ,   199 ,   3   ,   681 ,   1447    ,   1663    ,   7139    ,   10459   ,0 };
            const unsigned long dim1747Kuo2Init[]       =   {   1   ,   1   ,   1   ,   15  ,   3   ,   21  ,   99  ,   203 ,   161 ,   669 ,   611 ,   771 ,   7773    ,   6069    ,0 };
            const unsigned long dim1748Kuo2Init[]       =   {   1   ,   3   ,   3   ,   3   ,   17  ,   13  ,   111 ,   93  ,   35  ,   119 ,   113 ,   1169    ,   4891    ,   7753    ,0 };
            const unsigned long dim1749Kuo2Init[]       =   {   1   ,   1   ,   1   ,   11  ,   17  ,   7   ,   91  ,   39  ,   293 ,   443 ,   1819    ,   275 ,   7699    ,   8043    ,0 };
            const unsigned long dim1750Kuo2Init[]       =   {   1   ,   1   ,   3   ,   13  ,   21  ,   37  ,   93  ,   165 ,   281 ,   515 ,   1913    ,   839 ,   2177    ,   4907    ,0 };
            const unsigned long dim1751Kuo2Init[]       =   {   1   ,   3   ,   5   ,   11  ,   17  ,   23  ,   59  ,   161 ,   99  ,   371 ,   1223    ,   2815    ,   4019    ,   9225    ,0 };
            const unsigned long dim1752Kuo2Init[]       =   {   1   ,   3   ,   1   ,   5   ,   31  ,   31  ,   31  ,   111 ,   441 ,   585 ,   393 ,   1665    ,   1421    ,   7305    ,0 };
            const unsigned long dim1753Kuo2Init[]       =   {   1   ,   1   ,   5   ,   11  ,   19  ,   29  ,   81  ,   71  ,   223 ,   759 ,   2035    ,   1027    ,   2725    ,   4869    ,0 };
            const unsigned long dim1754Kuo2Init[]       =   {   1   ,   1   ,   1   ,   11  ,   19  ,   13  ,   33  ,   139 ,   457 ,   321 ,   793 ,   2043    ,   4823    ,   10819   ,0 };
            const unsigned long dim1755Kuo2Init[]       =   {   1   ,   3   ,   3   ,   5   ,   27  ,   9   ,   93  ,   141 ,   269 ,   177 ,   545 ,   1977    ,   1049    ,   8445    ,0 };
            const unsigned long dim1756Kuo2Init[]       =   {   1   ,   1   ,   7   ,   3   ,   3   ,   25  ,   11  ,   75  ,   241 ,   641 ,   523 ,   2255    ,   6649    ,   2191    ,0 };
            const unsigned long dim1757Kuo2Init[]       =   {   1   ,   1   ,   5   ,   11  ,   31  ,   31  ,   109 ,   235 ,   345 ,   291 ,   127 ,   2015    ,   7745    ,   4527    ,0 };
            const unsigned long dim1758Kuo2Init[]       =   {   1   ,   1   ,   3   ,   7   ,   3   ,   55  ,   107 ,   51  ,   145 ,   545 ,   1601    ,   1973    ,   4083    ,   3297    ,0 };
            const unsigned long dim1759Kuo2Init[]       =   {   1   ,   3   ,   5   ,   1   ,   21  ,   29  ,   31  ,   147 ,   245 ,   229 ,   1387    ,   2763    ,   769 ,   513 ,0 };
            const unsigned long dim1760Kuo2Init[]       =   {   1   ,   3   ,   5   ,   15  ,   3   ,   41  ,   119 ,   61  ,   401 ,   361 ,   443 ,   3663    ,   3691    ,   1851    ,0 };
            const unsigned long dim1761Kuo2Init[]       =   {   1   ,   3   ,   5   ,   5   ,   17  ,   7   ,   55  ,   159 ,   121 ,   841 ,   913 ,   3507    ,   5989    ,   15807   ,0 };
            const unsigned long dim1762Kuo2Init[]       =   {   1   ,   1   ,   3   ,   13  ,   5   ,   39  ,   65  ,   3   ,   261 ,   159 ,   439 ,   739 ,   6899    ,   61  ,0 };
            const unsigned long dim1763Kuo2Init[]       =   {   1   ,   1   ,   5   ,   13  ,   13  ,   43  ,   111 ,   17  ,   1   ,   503 ,   587 ,   1943    ,   7773    ,   14015   ,0 };
            const unsigned long dim1764Kuo2Init[]       =   {   1   ,   3   ,   1   ,   1   ,   5   ,   43  ,   55  ,   7   ,   171 ,   637 ,   39  ,   2341    ,   3027    ,   8521    ,0 };
            const unsigned long dim1765Kuo2Init[]       =   {   1   ,   3   ,   5   ,   7   ,   15  ,   49  ,   103 ,   157 ,   187 ,   773 ,   1391    ,   21  ,   5909    ,   14537   ,0 };
            const unsigned long dim1766Kuo2Init[]       =   {   1   ,   1   ,   5   ,   13  ,   1   ,   29  ,   63  ,   249 ,   473 ,   351 ,   705 ,   1999    ,   4135    ,   9683    ,0 };
            const unsigned long dim1767Kuo2Init[]       =   {   1   ,   3   ,   1   ,   7   ,   5   ,   7   ,   77  ,   179 ,   347 ,   101 ,   1055    ,   549 ,   297 ,   3195    ,0 };
            const unsigned long dim1768Kuo2Init[]       =   {   1   ,   1   ,   1   ,   5   ,   19  ,   51  ,   15  ,   247 ,   269 ,   545 ,   43  ,   2647    ,   1245    ,   6051    ,0 };
            const unsigned long dim1769Kuo2Init[]       =   {   1   ,   1   ,   5   ,   1   ,   9   ,   19  ,   45  ,   81  ,   145 ,   369 ,   1505    ,   1973    ,   3203    ,   10947   ,0 };
            const unsigned long dim1770Kuo2Init[]       =   {   1   ,   3   ,   5   ,   3   ,   11  ,   51  ,   113 ,   173 ,   121 ,   923 ,   969 ,   4005    ,   247 ,   1729    ,0 };
            const unsigned long dim1771Kuo2Init[]       =   {   1   ,   1   ,   7   ,   13  ,   1   ,   51  ,   111 ,   125 ,   163 ,   307 ,   1183    ,   1865    ,   2517    ,   9017    ,0 };
            const unsigned long dim1772Kuo2Init[]       =   {   1   ,   1   ,   5   ,   13  ,   7   ,   9   ,   115 ,   191 ,   53  ,   891 ,   1703    ,   2507    ,   3103    ,   12511   ,0 };
            const unsigned long dim1773Kuo2Init[]       =   {   1   ,   1   ,   3   ,   11  ,   9   ,   19  ,   3   ,   213 ,   421 ,   299 ,   1341    ,   3313    ,   637 ,   4977    ,0 };
            const unsigned long dim1774Kuo2Init[]       =   {   1   ,   3   ,   5   ,   13  ,   21  ,   7   ,   89  ,   185 ,   337 ,   105 ,   415 ,   375 ,   4357    ,   16305   ,0 };
            const unsigned long dim1775Kuo2Init[]       =   {   1   ,   1   ,   7   ,   7   ,   23  ,   25  ,   55  ,   19  ,   253 ,   879 ,   1779    ,   2451    ,   3485    ,   3383    ,0 };
            const unsigned long dim1776Kuo2Init[]       =   {   1   ,   3   ,   7   ,   7   ,   25  ,   33  ,   77  ,   67  ,   211 ,   571 ,   2029    ,   161 ,   3747    ,   11275   ,0 };
            const unsigned long dim1777Kuo2Init[]       =   {   1   ,   1   ,   3   ,   3   ,   15  ,   33  ,   55  ,   121 ,   97  ,   5   ,   907 ,   1869    ,   8171    ,   1553    ,0 };
            const unsigned long dim1778Kuo2Init[]       =   {   1   ,   1   ,   1   ,   9   ,   3   ,   33  ,   21  ,   165 ,   181 ,   211 ,   437 ,   1887    ,   895 ,   11549   ,0 };
            const unsigned long dim1779Kuo2Init[]       =   {   1   ,   3   ,   5   ,   3   ,   5   ,   7   ,   103 ,   95  ,   477 ,   585 ,   1307    ,   1017    ,   4115    ,   6011    ,0 };
            const unsigned long dim1780Kuo2Init[]       =   {   1   ,   1   ,   3   ,   11  ,   25  ,   45  ,   67  ,   239 ,   171 ,   869 ,   285 ,   2495    ,   4091    ,   15037   ,0 };
            const unsigned long dim1781Kuo2Init[]       =   {   1   ,   1   ,   3   ,   13  ,   5   ,   41  ,   103 ,   17  ,   283 ,   833 ,   179 ,   1183    ,   3323    ,   9777    ,0 };
            const unsigned long dim1782Kuo2Init[]       =   {   1   ,   3   ,   3   ,   5   ,   13  ,   29  ,   111 ,   37  ,   245 ,   329 ,   677 ,   4039    ,   2533    ,   3759    ,0 };
            const unsigned long dim1783Kuo2Init[]       =   {   1   ,   1   ,   7   ,   15  ,   23  ,   63  ,   115 ,   243 ,   347 ,   869 ,   587 ,   3073    ,   7727    ,   399 ,0 };
            const unsigned long dim1784Kuo2Init[]       =   {   1   ,   1   ,   7   ,   7   ,   5   ,   57  ,   49  ,   39  ,   463 ,   417 ,   1183    ,   1007    ,   3823    ,   15299   ,0 };
            const unsigned long dim1785Kuo2Init[]       =   {   1   ,   1   ,   1   ,   5   ,   23  ,   41  ,   65  ,   207 ,   151 ,   919 ,   1495    ,   2841    ,   1453    ,   6273    ,0 };
            const unsigned long dim1786Kuo2Init[]       =   {   1   ,   1   ,   1   ,   1   ,   15  ,   19  ,   51  ,   193 ,   187 ,   711 ,   11  ,   1291    ,   143 ,   15375   ,0 };
            const unsigned long dim1787Kuo2Init[]       =   {   1   ,   1   ,   3   ,   13  ,   25  ,   19  ,   3   ,   113 ,   157 ,   699 ,   1699    ,   1793    ,   29  ,   1869    ,0 };
            const unsigned long dim1788Kuo2Init[]       =   {   1   ,   3   ,   3   ,   7   ,   7   ,   37  ,   25  ,   73  ,   353 ,   443 ,   1517    ,   2621    ,   489 ,   15411   ,0 };
            const unsigned long dim1789Kuo2Init[]       =   {   1   ,   1   ,   1   ,   1   ,   7   ,   9   ,   91  ,   67  ,   471 ,   395 ,   1417    ,   431 ,   4793    ,   9517    ,0 };
            const unsigned long dim1790Kuo2Init[]       =   {   1   ,   3   ,   1   ,   5   ,   31  ,   15  ,   63  ,   67  ,   109 ,   537 ,   1543    ,   1021    ,   2549    ,   13587   ,0 };
            const unsigned long dim1791Kuo2Init[]       =   {   1   ,   1   ,   1   ,   3   ,   7   ,   63  ,   127 ,   141 ,   23  ,   281 ,   523 ,   283 ,   347 ,   2095    ,0 };
            const unsigned long dim1792Kuo2Init[]       =   {   1   ,   1   ,   7   ,   9   ,   21  ,   21  ,   71  ,   97  ,   357 ,   887 ,   1785    ,   3471    ,   3187    ,   12163   ,0 };
            const unsigned long dim1793Kuo2Init[]       =   {   1   ,   3   ,   3   ,   3   ,   25  ,   23  ,   127 ,   135 ,   199 ,   661 ,   7   ,   3279    ,   91  ,   11879   ,0 };
            const unsigned long dim1794Kuo2Init[]       =   {   1   ,   1   ,   7   ,   5   ,   29  ,   51  ,   41  ,   5   ,   387 ,   901 ,   233 ,   1559    ,   511 ,   7735    ,0 };
            const unsigned long dim1795Kuo2Init[]       =   {   1   ,   1   ,   5   ,   9   ,   3   ,   13  ,   45  ,   119 ,   177 ,   381 ,   1363    ,   3627    ,   259 ,   771 ,0 };
            const unsigned long dim1796Kuo2Init[]       =   {   1   ,   1   ,   5   ,   11  ,   5   ,   31  ,   73  ,   151 ,   487 ,   371 ,   1543    ,   1183    ,   697 ,   10225   ,0 };
            const unsigned long dim1797Kuo2Init[]       =   {   1   ,   3   ,   5   ,   7   ,   29  ,   39  ,   125 ,   245 ,   373 ,   43  ,   583 ,   233 ,   5799    ,   7865    ,0 };
            const unsigned long dim1798Kuo2Init[]       =   {   1   ,   1   ,   7   ,   13  ,   17  ,   41  ,   47  ,   7   ,   161 ,   717 ,   947 ,   2421    ,   1983    ,   12963   ,0 };
            const unsigned long dim1799Kuo2Init[]       =   {   1   ,   3   ,   3   ,   3   ,   31  ,   19  ,   11  ,   131 ,   217 ,   533 ,   1407    ,   421 ,   163 ,   5849    ,0 };
            const unsigned long dim1800Kuo2Init[]       =   {   1   ,   1   ,   3   ,   13  ,   31  ,   49  ,   77  ,   33  ,   335 ,   627 ,   1135    ,   4021    ,   6609    ,   819 ,0 };
            const unsigned long dim1801Kuo2Init[]       =   {   1   ,   3   ,   5   ,   5   ,   17  ,   51  ,   27  ,   83  ,   153 ,   469 ,   85  ,   2725    ,   4297    ,   8869    ,0 };
            const unsigned long dim1802Kuo2Init[]       =   {   1   ,   3   ,   5   ,   9   ,   29  ,   7   ,   93  ,   53  ,   251 ,   579 ,   1243    ,   2689    ,   4655    ,   9247    ,0 };
            const unsigned long dim1803Kuo2Init[]       =   {   1   ,   3   ,   1   ,   13  ,   9   ,   35  ,   59  ,   213 ,   181 ,   409 ,   1177    ,   3783    ,   6339    ,   339 ,0 };
            const unsigned long dim1804Kuo2Init[]       =   {   1   ,   1   ,   1   ,   7   ,   25  ,   55  ,   111 ,   133 ,   481 ,   757 ,   1683    ,   2193    ,   681 ,   12141   ,0 };
            const unsigned long dim1805Kuo2Init[]       =   {   1   ,   1   ,   3   ,   13  ,   11  ,   1   ,   57  ,   137 ,   359 ,   325 ,   67  ,   3069    ,   1125    ,   8305    ,0 };
            const unsigned long dim1806Kuo2Init[]       =   {   1   ,   1   ,   1   ,   5   ,   21  ,   11  ,   79  ,   195 ,   335 ,   85  ,   1749    ,   1859    ,   4253    ,   6391    ,0 };
            const unsigned long dim1807Kuo2Init[]       =   {   1   ,   1   ,   3   ,   13  ,   17  ,   55  ,   23  ,   119 ,   159 ,   843 ,   1677    ,   1725    ,   1139    ,   2051    ,0 };
            const unsigned long dim1808Kuo2Init[]       =   {   1   ,   3   ,   1   ,   11  ,   31  ,   35  ,   11  ,   49  ,   497 ,   807 ,   191 ,   3889    ,   7811    ,   13217   ,0 };
            const unsigned long dim1809Kuo2Init[]       =   {   1   ,   3   ,   1   ,   9   ,   29  ,   63  ,   117 ,   161 ,   191 ,   371 ,   245 ,   219 ,   1559    ,   13303   ,0 };
            const unsigned long dim1810Kuo2Init[]       =   {   1   ,   3   ,   3   ,   7   ,   5   ,   19  ,   37  ,   231 ,   381 ,   323 ,   233 ,   3973    ,   6531    ,   8807    ,0 };
            const unsigned long dim1811Kuo2Init[]       =   {   1   ,   1   ,   1   ,   11  ,   3   ,   43  ,   35  ,   103 ,   291 ,   363 ,   531 ,   1129    ,   3321    ,   3167    ,0 };
            const unsigned long dim1812Kuo2Init[]       =   {   1   ,   3   ,   3   ,   15  ,   25  ,   9   ,   47  ,   139 ,   143 ,   155 ,   2005    ,   1573    ,   2007    ,   6483    ,0 };
            const unsigned long dim1813Kuo2Init[]       =   {   1   ,   3   ,   7   ,   5   ,   25  ,   3   ,   125 ,   19  ,   255 ,   699 ,   1651    ,   1449    ,   3753    ,   12873   ,0 };
            const unsigned long dim1814Kuo2Init[]       =   {   1   ,   1   ,   7   ,   5   ,   7   ,   7   ,   63  ,   31  ,   255 ,   991 ,   1029    ,   1635    ,   4941    ,   7891    ,0 };
            const unsigned long dim1815Kuo2Init[]       =   {   1   ,   3   ,   3   ,   1   ,   3   ,   13  ,   59  ,   73  ,   429 ,   931 ,   53  ,   711 ,   3305    ,   7621    ,0 };
            const unsigned long dim1816Kuo2Init[]       =   {   1   ,   3   ,   3   ,   5   ,   15  ,   55  ,   103 ,   159 ,   185 ,   365 ,   1283    ,   1171    ,   6659    ,   2547    ,0 };
            const unsigned long dim1817Kuo2Init[]       =   {   1   ,   1   ,   5   ,   5   ,   25  ,   33  ,   91  ,   87  ,   323 ,   235 ,   1997    ,   1397    ,   6777    ,   10065   ,0 };
            const unsigned long dim1818Kuo2Init[]       =   {   1   ,   3   ,   1   ,   1   ,   7   ,   35  ,   19  ,   5   ,   327 ,   163 ,   651 ,   215 ,   7965    ,   1143    ,0 };
            const unsigned long dim1819Kuo2Init[]       =   {   1   ,   1   ,   1   ,   11  ,   7   ,   13  ,   69  ,   129 ,   277 ,   665 ,   1763    ,   33  ,   8145    ,   7131    ,0 };
            const unsigned long dim1820Kuo2Init[]       =   {   1   ,   3   ,   3   ,   13  ,   17  ,   25  ,   111 ,   247 ,   91  ,   537 ,   325 ,   2627    ,   1777    ,   2861    ,0 };
            const unsigned long dim1821Kuo2Init[]       =   {   1   ,   3   ,   3   ,   3   ,   29  ,   15  ,   87  ,   41  ,   103 ,   957 ,   105 ,   757 ,   2635    ,   4825    ,0 };
            const unsigned long dim1822Kuo2Init[]       =   {   1   ,   1   ,   3   ,   11  ,   21  ,   43  ,   73  ,   143 ,   31  ,   913 ,   1579    ,   1725    ,   6679    ,   2491    ,0 };
            const unsigned long dim1823Kuo2Init[]       =   {   1   ,   3   ,   1   ,   11  ,   1   ,   31  ,   93  ,   231 ,   327 ,   177 ,   921 ,   2715    ,   8015    ,   3295    ,0 };
            const unsigned long dim1824Kuo2Init[]       =   {   1   ,   3   ,   1   ,   3   ,   5   ,   47  ,   59  ,   1   ,   105 ,   991 ,   1581    ,   115 ,   1957    ,   10753   ,0 };
            const unsigned long dim1825Kuo2Init[]       =   {   1   ,   3   ,   7   ,   9   ,   13  ,   31  ,   65  ,   241 ,   353 ,   613 ,   1065    ,   1201    ,   7267    ,   2099    ,0 };
            const unsigned long dim1826Kuo2Init[]       =   {   1   ,   3   ,   7   ,   11  ,   5   ,   55  ,   121 ,   45  ,   279 ,   765 ,   1263    ,   993 ,   5047    ,   12059   ,0 };
            const unsigned long dim1827Kuo2Init[]       =   {   1   ,   1   ,   7   ,   3   ,   25  ,   27  ,   13  ,   123 ,   357 ,   339 ,   1299    ,   297 ,   361 ,   4951    ,0 };
            const unsigned long dim1828Kuo2Init[]       =   {   1   ,   3   ,   5   ,   3   ,   13  ,   43  ,   117 ,   237 ,   3   ,   463 ,   911 ,   3447    ,   7463    ,   5985    ,0 };
            const unsigned long dim1829Kuo2Init[]       =   {   1   ,   3   ,   1   ,   15  ,   13  ,   1   ,   21  ,   213 ,   115 ,   291 ,   1317    ,   2201    ,   7939    ,   16135   ,0 };
            const unsigned long dim1830Kuo2Init[]       =   {   1   ,   1   ,   3   ,   15  ,   13  ,   15  ,   55  ,   51  ,   203 ,   277 ,   71  ,   1025    ,   4237    ,   301 ,0 };
            const unsigned long dim1831Kuo2Init[]       =   {   1   ,   1   ,   5   ,   5   ,   17  ,   31  ,   95  ,   25  ,   7   ,   519 ,   1971    ,   3069    ,   1937    ,   5693    ,0 };
            const unsigned long dim1832Kuo2Init[]       =   {   1   ,   3   ,   7   ,   5   ,   11  ,   17  ,   31  ,   111 ,   219 ,   199 ,   303 ,   2527    ,   4185    ,   13185   ,0 };
            const unsigned long dim1833Kuo2Init[]       =   {   1   ,   3   ,   5   ,   13  ,   11  ,   17  ,   105 ,   147 ,   51  ,   885 ,   675 ,   3035    ,   4245    ,   3257    ,0 };
            const unsigned long dim1834Kuo2Init[]       =   {   1   ,   3   ,   3   ,   5   ,   9   ,   17  ,   37  ,   139 ,   251 ,   85  ,   361 ,   2569    ,   2727    ,   14111   ,0 };
            const unsigned long dim1835Kuo2Init[]       =   {   1   ,   3   ,   5   ,   9   ,   31  ,   5   ,   101 ,   1   ,   475 ,   645 ,   1515    ,   2713    ,   2771    ,   2981    ,0 };
            const unsigned long dim1836Kuo2Init[]       =   {   1   ,   3   ,   3   ,   13  ,   21  ,   21  ,   31  ,   117 ,   511 ,   525 ,   1105    ,   2881    ,   6361    ,   6991    ,0 };
            const unsigned long dim1837Kuo2Init[]       =   {   1   ,   1   ,   5   ,   5   ,   15  ,   5   ,   47  ,   11  ,   391 ,   239 ,   65  ,   1545    ,   691 ,   3973    ,0 };
            const unsigned long dim1838Kuo2Init[]       =   {   1   ,   3   ,   1   ,   1   ,   9   ,   3   ,   91  ,   49  ,   505 ,   353 ,   1593    ,   4009    ,   2535    ,   4947    ,0 };
            const unsigned long dim1839Kuo2Init[]       =   {   1   ,   3   ,   1   ,   13  ,   15  ,   27  ,   45  ,   255 ,   503 ,   487 ,   613 ,   3453    ,   655 ,   10083   ,0 };
            const unsigned long dim1840Kuo2Init[]       =   {   1   ,   1   ,   1   ,   9   ,   23  ,   17  ,   39  ,   219 ,   281 ,   669 ,   867 ,   773 ,   2555    ,   12565   ,0 };
            const unsigned long dim1841Kuo2Init[]       =   {   1   ,   3   ,   1   ,   7   ,   21  ,   29  ,   61  ,   85  ,   441 ,   273 ,   1889    ,   1441    ,   6291    ,   2599    ,0 };
            const unsigned long dim1842Kuo2Init[]       =   {   1   ,   1   ,   1   ,   7   ,   1   ,   7   ,   95  ,   9   ,   105 ,   77  ,   463 ,   701 ,   6001    ,   13491   ,0 };
            const unsigned long dim1843Kuo2Init[]       =   {   1   ,   3   ,   7   ,   3   ,   29  ,   57  ,   61  ,   123 ,   105 ,   909 ,   1925    ,   1255    ,   7069    ,   3049    ,0 };
            const unsigned long dim1844Kuo2Init[]       =   {   1   ,   1   ,   3   ,   11  ,   15  ,   37  ,   23  ,   25  ,   359 ,   743 ,   1599    ,   575 ,   4095    ,   11539   ,0 };
            const unsigned long dim1845Kuo2Init[]       =   {   1   ,   3   ,   3   ,   9   ,   17  ,   33  ,   59  ,   245 ,   119 ,   457 ,   147 ,   1027    ,   4051    ,   507 ,0 };
            const unsigned long dim1846Kuo2Init[]       =   {   1   ,   3   ,   1   ,   11  ,   25  ,   59  ,   97  ,   37  ,   199 ,   247 ,   985 ,   635 ,   3501    ,   11213   ,0 };
            const unsigned long dim1847Kuo2Init[]       =   {   1   ,   1   ,   5   ,   7   ,   21  ,   19  ,   17  ,   181 ,   183 ,   331 ,   817 ,   1069    ,   3863    ,   7547    ,0 };
            const unsigned long dim1848Kuo2Init[]       =   {   1   ,   3   ,   5   ,   1   ,   11  ,   25  ,   121 ,   127 ,   105 ,   57  ,   689 ,   701 ,   1527    ,   5301    ,0 };
            const unsigned long dim1849Kuo2Init[]       =   {   1   ,   1   ,   7   ,   3   ,   25  ,   59  ,   67  ,   103 ,   265 ,   649 ,   957 ,   3073    ,   3047    ,   12881   ,0 };
            const unsigned long dim1850Kuo2Init[]       =   {   1   ,   3   ,   1   ,   3   ,   17  ,   15  ,   103 ,   37  ,   359 ,   665 ,   1677    ,   2563    ,   7079    ,   2577    ,0 };
            const unsigned long dim1851Kuo2Init[]       =   {   1   ,   1   ,   1   ,   13  ,   21  ,   43  ,   117 ,   187 ,   447 ,   711 ,   749 ,   643 ,   183 ,   13053   ,0 };
            const unsigned long dim1852Kuo2Init[]       =   {   1   ,   3   ,   7   ,   3   ,   15  ,   45  ,   23  ,   169 ,   481 ,   841 ,   625 ,   3241    ,   7739    ,   13463   ,0 };
            const unsigned long dim1853Kuo2Init[]       =   {   1   ,   3   ,   7   ,   9   ,   7   ,   11  ,   9   ,   73  ,   57  ,   595 ,   1981    ,   3499    ,   53  ,   3445    ,0 };
            const unsigned long dim1854Kuo2Init[]       =   {   1   ,   3   ,   3   ,   11  ,   17  ,   13  ,   61  ,   139 ,   445 ,   287 ,   705 ,   3157    ,   7503    ,   15279   ,0 };
            const unsigned long dim1855Kuo2Init[]       =   {   1   ,   3   ,   5   ,   5   ,   15  ,   5   ,   15  ,   117 ,   177 ,   517 ,   1907    ,   399 ,   1613    ,   9495    ,0 };
            const unsigned long dim1856Kuo2Init[]       =   {   1   ,   3   ,   5   ,   13  ,   19  ,   37  ,   81  ,   179 ,   473 ,   477 ,   551 ,   1235    ,   6503    ,   14401   ,0 };
            const unsigned long dim1857Kuo2Init[]       =   {   1   ,   3   ,   5   ,   7   ,   1   ,   21  ,   29  ,   181 ,   25  ,   349 ,   1915    ,   1047    ,   6645    ,   5153    ,0 };
            const unsigned long dim1858Kuo2Init[]       =   {   1   ,   3   ,   1   ,   11  ,   31  ,   31  ,   13  ,   165 ,   147 ,   995 ,   1767    ,   4013    ,   4993    ,   51  ,0 };
            const unsigned long dim1859Kuo2Init[]       =   {   1   ,   1   ,   3   ,   11  ,   19  ,   33  ,   47  ,   179 ,   465 ,   811 ,   1535    ,   2555    ,   1843    ,   10877   ,0 };
            const unsigned long dim1860Kuo2Init[]       =   {   1   ,   3   ,   5   ,   7   ,   25  ,   13  ,   49  ,   105 ,   39  ,   893 ,   1729    ,   1297    ,   6071    ,   685 ,0 };
            const unsigned long dim1861Kuo2Init[]       =   {   1   ,   1   ,   3   ,   5   ,   3   ,   17  ,   69  ,   141 ,   83  ,   393 ,   1377    ,   3401    ,   1749    ,   10307   ,0 };
            const unsigned long dim1862Kuo2Init[]       =   {   1   ,   3   ,   7   ,   7   ,   5   ,   43  ,   27  ,   39  ,   61  ,   119 ,   829 ,   2105    ,   8155    ,   1647    ,0 };
            const unsigned long dim1863Kuo2Init[]       =   {   1   ,   1   ,   1   ,   3   ,   7   ,   61  ,   83  ,   253 ,   129 ,   459 ,   1241    ,   569 ,   7457    ,   3635    ,0 };
            const unsigned long dim1864Kuo2Init[]       =   {   1   ,   1   ,   1   ,   5   ,   21  ,   17  ,   97  ,   213 ,   463 ,   549 ,   1219    ,   3685    ,   3721    ,   8659    ,0 };
            const unsigned long dim1865Kuo2Init[]       =   {   1   ,   3   ,   7   ,   5   ,   19  ,   53  ,   45  ,   151 ,   15  ,   483 ,   1077    ,   1811    ,   4861    ,   9979    ,0 };
            const unsigned long dim1866Kuo2Init[]       =   {   1   ,   1   ,   5   ,   13  ,   7   ,   21  ,   11  ,   117 ,   267 ,   723 ,   573 ,   3335    ,   6331    ,   3255    ,0 };
            const unsigned long dim1867Kuo2Init[]       =   {   1   ,   3   ,   5   ,   3   ,   23  ,   17  ,   83  ,   171 ,   501 ,   249 ,   1937    ,   593 ,   3757    ,   6035    ,   30841   ,0 };
            const unsigned long dim1868Kuo2Init[]       =   {   1   ,   3   ,   7   ,   1   ,   13  ,   15  ,   19  ,   33  ,   191 ,   509 ,   1461    ,   2105    ,   2175    ,   10593   ,   27199   ,0 };
            const unsigned long dim1869Kuo2Init[]       =   {   1   ,   3   ,   3   ,   3   ,   9   ,   7   ,   7   ,   51  ,   199 ,   339 ,   393 ,   2543    ,   4269    ,   2447    ,   17671   ,0 };
            const unsigned long dim1870Kuo2Init[]       =   {   1   ,   3   ,   1   ,   1   ,   17  ,   3   ,   71  ,   119 ,   15  ,   333 ,   1313    ,   2609    ,   1077    ,   2493    ,   23109   ,0 };
            const unsigned long dim1871Kuo2Init[]       =   {   1   ,   1   ,   1   ,   11  ,   7   ,   45  ,   69  ,   69  ,   241 ,   995 ,   1111    ,   2141    ,   5467    ,   5003    ,   11135   ,0 };
            const unsigned long dim1872Kuo2Init[]       =   {   1   ,   1   ,   7   ,   11  ,   21  ,   1   ,   25  ,   35  ,   445 ,   379 ,   449 ,   3703    ,   3281    ,   3433    ,   5207    ,0 };
            const unsigned long dim1873Kuo2Init[]       =   {   1   ,   3   ,   3   ,   5   ,   15  ,   49  ,   119 ,   11  ,   331 ,   681 ,   1209    ,   161 ,   7969    ,   11115   ,   4775    ,0 };
            const unsigned long dim1874Kuo2Init[]       =   {   1   ,   3   ,   7   ,   15  ,   1   ,   1   ,   61  ,   131 ,   141 ,   849 ,   2035    ,   125 ,   2457    ,   9209    ,   18411   ,0 };
            const unsigned long dim1875Kuo2Init[]       =   {   1   ,   1   ,   5   ,   5   ,   25  ,   17  ,   91  ,   255 ,   137 ,   815 ,   1803    ,   3521    ,   181 ,   15683   ,   32099   ,0 };
            const unsigned long dim1876Kuo2Init[]       =   {   1   ,   3   ,   1   ,   11  ,   29  ,   13  ,   91  ,   129 ,   167 ,   323 ,   823 ,   3717    ,   473 ,   7617    ,   5431    ,0 };
            const unsigned long dim1877Kuo2Init[]       =   {   1   ,   1   ,   5   ,   9   ,   17  ,   41  ,   65  ,   5   ,   57  ,   441 ,   765 ,   2761    ,   5033    ,   14551   ,   28223   ,0 };
            const unsigned long dim1878Kuo2Init[]       =   {   1   ,   3   ,   1   ,   1   ,   11  ,   11  ,   35  ,   109 ,   197 ,   349 ,   1495    ,   4057    ,   5255    ,   8327    ,   29281   ,0 };
            const unsigned long dim1879Kuo2Init[]       =   {   1   ,   1   ,   7   ,   1   ,   27  ,   33  ,   7   ,   133 ,   439 ,   329 ,   655 ,   2459    ,   5369    ,   14681   ,   12271   ,0 };
            const unsigned long dim1880Kuo2Init[]       =   {   1   ,   1   ,   1   ,   13  ,   29  ,   11  ,   15  ,   71  ,   117 ,   849 ,   285 ,   3779    ,   1181    ,   6315    ,   23627   ,0 };
            const unsigned long dim1881Kuo2Init[]       =   {   1   ,   3   ,   3   ,   7   ,   15  ,   33  ,   93  ,   133 ,   239 ,   169 ,   197 ,   3005    ,   1035    ,   6771    ,   10881   ,0 };
            const unsigned long dim1882Kuo2Init[]       =   {   1   ,   3   ,   3   ,   13  ,   29  ,   13  ,   69  ,   157 ,   275 ,   335 ,   1977    ,   461 ,   8087    ,   10027   ,   6377    ,0 };
            const unsigned long dim1883Kuo2Init[]       =   {   1   ,   3   ,   1   ,   15  ,   25  ,   23  ,   107 ,   125 ,   27  ,   455 ,   207 ,   3425    ,   5319    ,   4861    ,   4853    ,0 };
            const unsigned long dim1884Kuo2Init[]       =   {   1   ,   1   ,   5   ,   15  ,   17  ,   49  ,   3   ,   129 ,   401 ,   451 ,   363 ,   553 ,   3999    ,   8115    ,   25621   ,0 };
            const unsigned long dim1885Kuo2Init[]       =   {   1   ,   3   ,   3   ,   3   ,   9   ,   63  ,   71  ,   219 ,   331 ,   43  ,   1369    ,   3821    ,   5189    ,   8799    ,   19413   ,0 };
            const unsigned long dim1886Kuo2Init[]       =   {   1   ,   1   ,   3   ,   13  ,   31  ,   19  ,   57  ,   127 ,   229 ,   429 ,   535 ,   591 ,   1135    ,   12559   ,   4219    ,0 };
            const unsigned long dim1887Kuo2Init[]       =   {   1   ,   1   ,   7   ,   3   ,   11  ,   41  ,   49  ,   45  ,   423 ,   299 ,   571 ,   1907    ,   1755    ,   10813   ,   14199   ,0 };
            const unsigned long dim1888Kuo2Init[]       =   {   1   ,   3   ,   1   ,   1   ,   1   ,   41  ,   53  ,   139 ,   345 ,   521 ,   1051    ,   691 ,   7367    ,   923 ,   2171    ,0 };
            const unsigned long dim1889Kuo2Init[]       =   {   1   ,   1   ,   7   ,   1   ,   23  ,   39  ,   21  ,   241 ,   433 ,   269 ,   1733    ,   3455    ,   8171    ,   585 ,   9035    ,0 };
            const unsigned long dim1890Kuo2Init[]       =   {   1   ,   1   ,   3   ,   3   ,   15  ,   31  ,   39  ,   75  ,   223 ,   827 ,   1799    ,   2699    ,   6927    ,   4251    ,   16799   ,0 };
            const unsigned long dim1891Kuo2Init[]       =   {   1   ,   1   ,   7   ,   11  ,   25  ,   17  ,   5   ,   199 ,   265 ,   953 ,   1609    ,   185 ,   5069    ,   10061   ,   31235   ,0 };
            const unsigned long dim1892Kuo2Init[]       =   {   1   ,   3   ,   3   ,   7   ,   11  ,   17  ,   125 ,   149 ,   29  ,   149 ,   1241    ,   3975    ,   2915    ,   13847   ,   8245    ,0 };
            const unsigned long dim1893Kuo2Init[]       =   {   1   ,   3   ,   5   ,   15  ,   19  ,   7   ,   115 ,   127 ,   325 ,   147 ,   505 ,   3199    ,   2261    ,   15309   ,   1925    ,0 };
            const unsigned long dim1894Kuo2Init[]       =   {   1   ,   1   ,   3   ,   15  ,   19  ,   27  ,   45  ,   131 ,   35  ,   69  ,   275 ,   2895    ,   8131    ,   6311    ,   15507   ,0 };
            const unsigned long dim1895Kuo2Init[]       =   {   1   ,   3   ,   7   ,   5   ,   5   ,   35  ,   115 ,   39  ,   343 ,   491 ,   199 ,   985 ,   7767    ,   13183   ,   23341   ,0 };
            const unsigned long dim1896Kuo2Init[]       =   {   1   ,   3   ,   3   ,   3   ,   17  ,   19  ,   125 ,   237 ,   393 ,   1   ,   1763    ,   2129    ,   2221    ,   4993    ,   23023   ,0 };
            const unsigned long dim1897Kuo2Init[]       =   {   1   ,   1   ,   5   ,   13  ,   1   ,   39  ,   53  ,   77  ,   43  ,   665 ,   173 ,   1923    ,   5071    ,   2713    ,   28507   ,0 };
            const unsigned long dim1898Kuo2Init[]       =   {   1   ,   1   ,   3   ,   5   ,   7   ,   57  ,   77  ,   19  ,   9   ,   441 ,   1941    ,   475 ,   4059    ,   3977    ,   25689   ,0 };
            const unsigned long dim1899Kuo2Init[]       =   {   1   ,   1   ,   3   ,   9   ,   23  ,   49  ,   35  ,   139 ,   189 ,   431 ,   1077    ,   1057    ,   1565    ,   5597    ,   23755   ,0 };
            const unsigned long dim1900Kuo2Init[]       =   {   1   ,   1   ,   7   ,   1   ,   7   ,   5   ,   121 ,   69  ,   277 ,   797 ,   1879    ,   3479    ,   3023    ,   1781    ,   31307   ,0 };
            const unsigned long dim1901Kuo2Init[]       =   {   1   ,   1   ,   5   ,   15  ,   5   ,   23  ,   75  ,   181 ,   97  ,   491 ,   399 ,   63  ,   3211    ,   7099    ,   14597   ,0 };
            const unsigned long dim1902Kuo2Init[]       =   {   1   ,   3   ,   1   ,   11  ,   9   ,   3   ,   85  ,   159 ,   143 ,   13  ,   599 ,   3685    ,   4291    ,   11457   ,   24571   ,0 };
            const unsigned long dim1903Kuo2Init[]       =   {   1   ,   3   ,   7   ,   9   ,   23  ,   45  ,   17  ,   197 ,   351 ,   997 ,   1789    ,   3965    ,   1163    ,   9027    ,   28843   ,0 };
            const unsigned long dim1904Kuo2Init[]       =   {   1   ,   1   ,   5   ,   15  ,   25  ,   13  ,   93  ,   141 ,   395 ,   905 ,   933 ,   3169    ,   7551    ,   1775    ,   22043   ,0 };
            const unsigned long dim1905Kuo2Init[]       =   {   1   ,   3   ,   3   ,   11  ,   17  ,   47  ,   115 ,   211 ,   157 ,   887 ,   1829    ,   767 ,   6821    ,   5243    ,   26423   ,0 };
            const unsigned long dim1906Kuo2Init[]       =   {   1   ,   1   ,   1   ,   1   ,   15  ,   33  ,   27  ,   239 ,   405 ,   5   ,   861 ,   1909    ,   1261    ,   1921    ,   32245   ,0 };
            const unsigned long dim1907Kuo2Init[]       =   {   1   ,   1   ,   1   ,   15  ,   5   ,   15  ,   21  ,   251 ,   489 ,   789 ,   507 ,   321 ,   6429    ,   1885    ,   23139   ,0 };
            const unsigned long dim1908Kuo2Init[]       =   {   1   ,   1   ,   1   ,   13  ,   19  ,   61  ,   59  ,   201 ,   69  ,   933 ,   83  ,   1657    ,   1331    ,   16183   ,   23085   ,0 };
            const unsigned long dim1909Kuo2Init[]       =   {   1   ,   3   ,   1   ,   5   ,   29  ,   59  ,   61  ,   149 ,   95  ,   897 ,   1107    ,   2853    ,   5271    ,   7069    ,   3759    ,0 };
            const unsigned long dim1910Kuo2Init[]       =   {   1   ,   1   ,   3   ,   9   ,   17  ,   53  ,   71  ,   183 ,   29  ,   353 ,   1621    ,   4015    ,   6395    ,   8737    ,   12507   ,0 };
            const unsigned long dim1911Kuo2Init[]       =   {   1   ,   3   ,   3   ,   15  ,   23  ,   31  ,   77  ,   187 ,   429 ,   59  ,   117 ,   641 ,   1039    ,   15113   ,   1177    ,0 };
            const unsigned long dim1912Kuo2Init[]       =   {   1   ,   1   ,   3   ,   7   ,   17  ,   9   ,   73  ,   37  ,   177 ,   901 ,   573 ,   2503    ,   5865    ,   13939   ,   10123   ,0 };
            const unsigned long dim1913Kuo2Init[]       =   {   1   ,   3   ,   1   ,   1   ,   7   ,   21  ,   55  ,   255 ,   85  ,   993 ,   583 ,   3345    ,   3211    ,   6027    ,   18677   ,0 };
            const unsigned long dim1914Kuo2Init[]       =   {   1   ,   3   ,   3   ,   1   ,   27  ,   37  ,   111 ,   235 ,   281 ,   363 ,   121 ,   3749    ,   7381    ,   14959   ,   22997   ,0 };
            const unsigned long dim1915Kuo2Init[]       =   {   1   ,   1   ,   1   ,   11  ,   13  ,   19  ,   47  ,   203 ,   157 ,   975 ,   1173    ,   1389    ,   1021    ,   6481    ,   547 ,0 };
            const unsigned long dim1916Kuo2Init[]       =   {   1   ,   3   ,   5   ,   11  ,   7   ,   51  ,   121 ,   99  ,   289 ,   303 ,   1759    ,   571 ,   3585    ,   2263    ,   21231   ,0 };
            const unsigned long dim1917Kuo2Init[]       =   {   1   ,   3   ,   5   ,   9   ,   9   ,   7   ,   25  ,   39  ,   443 ,   635 ,   1597    ,   2095    ,   6739    ,   16367   ,   11285   ,0 };
            const unsigned long dim1918Kuo2Init[]       =   {   1   ,   1   ,   1   ,   9   ,   5   ,   47  ,   107 ,   223 ,   389 ,   25  ,   1353    ,   501 ,   8025    ,   7269    ,   23617   ,0 };
            const unsigned long dim1919Kuo2Init[]       =   {   1   ,   1   ,   5   ,   5   ,   23  ,   25  ,   59  ,   133 ,   387 ,   761 ,   835 ,   105 ,   45  ,   5469    ,   13461   ,0 };
            const unsigned long dim1920Kuo2Init[]       =   {   1   ,   1   ,   3   ,   11  ,   5   ,   33  ,   121 ,   149 ,   121 ,   881 ,   603 ,   3455    ,   7667    ,   3039    ,   8471    ,0 };
            const unsigned long dim1921Kuo2Init[]       =   {   1   ,   3   ,   5   ,   9   ,   17  ,   41  ,   1   ,   109 ,   67  ,   511 ,   617 ,   1139    ,   7605    ,   12211   ,   8603    ,0 };
            const unsigned long dim1922Kuo2Init[]       =   {   1   ,   1   ,   3   ,   9   ,   13  ,   27  ,   5   ,   163 ,   17  ,   889 ,   199 ,   77  ,   5233    ,   6383    ,   15787   ,0 };
            const unsigned long dim1923Kuo2Init[]       =   {   1   ,   1   ,   1   ,   13  ,   5   ,   27  ,   45  ,   47  ,   213 ,   473 ,   1101    ,   623 ,   6169    ,   3139    ,   27561   ,0 };
            const unsigned long dim1924Kuo2Init[]       =   {   1   ,   1   ,   3   ,   13  ,   15  ,   63  ,   79  ,   51  ,   47  ,   227 ,   1771    ,   3167    ,   2975    ,   16075   ,   9117    ,0 };
            const unsigned long dim1925Kuo2Init[]       =   {   1   ,   1   ,   5   ,   5   ,   29  ,   9   ,   87  ,   213 ,   349 ,   763 ,   1341    ,   1299    ,   1183    ,   4763    ,   14869   ,0 };
            const unsigned long dim1926Kuo2Init[]       =   {   1   ,   3   ,   7   ,   1   ,   23  ,   3   ,   37  ,   99  ,   341 ,   67  ,   1273    ,   3383    ,   7805    ,   8373    ,   19401   ,0 };
            const unsigned long dim1927Kuo2Init[]       =   {   1   ,   3   ,   3   ,   3   ,   7   ,   25  ,   23  ,   231 ,   291 ,   591 ,   291 ,   1077    ,   7291    ,   9763    ,   515 ,0 };
            const unsigned long dim1928Kuo2Init[]       =   {   1   ,   1   ,   1   ,   3   ,   9   ,   49  ,   99  ,   127 ,   501 ,   375 ,   995 ,   1883    ,   6229    ,   1141    ,   16513   ,0 };
            const unsigned long dim1929Kuo2Init[]       =   {   1   ,   1   ,   1   ,   13  ,   21  ,   35  ,   43  ,   39  ,   405 ,   283 ,   1555    ,   3189    ,   4999    ,   139 ,   11031   ,0 };
            const unsigned long dim1930Kuo2Init[]       =   {   1   ,   3   ,   3   ,   1   ,   25  ,   3   ,   29  ,   89  ,   71  ,   975 ,   435 ,   2513    ,   4667    ,   14859   ,   22495   ,0 };
            const unsigned long dim1931Kuo2Init[]       =   {   1   ,   3   ,   5   ,   9   ,   27  ,   45  ,   29  ,   145 ,   97  ,   149 ,   341 ,   3253    ,   5385    ,   8099    ,   169 ,0 };
            const unsigned long dim1932Kuo2Init[]       =   {   1   ,   1   ,   1   ,   3   ,   17  ,   49  ,   73  ,   31  ,   291 ,   191 ,   959 ,   3139    ,   6221    ,   423 ,   5247    ,0 };
            const unsigned long dim1933Kuo2Init[]       =   {   1   ,   3   ,   7   ,   3   ,   11  ,   23  ,   25  ,   65  ,   291 ,   877 ,   15  ,   437 ,   3883    ,   6545    ,   16327   ,0 };
            const unsigned long dim1934Kuo2Init[]       =   {   1   ,   3   ,   1   ,   7   ,   21  ,   53  ,   35  ,   143 ,   433 ,   117 ,   1135    ,   467 ,   6825    ,   16041   ,   3351    ,0 };
            const unsigned long dim1935Kuo2Init[]       =   {   1   ,   3   ,   5   ,   1   ,   11  ,   57  ,   105 ,   147 ,   443 ,   1   ,   1879    ,   2963    ,   4041    ,   8929    ,   219 ,0 };
            const unsigned long dim1936Kuo2Init[]       =   {   1   ,   3   ,   3   ,   11  ,   13  ,   59  ,   57  ,   175 ,   151 ,   1015    ,   629 ,   1549    ,   4161    ,   16147   ,   25857   ,0 };
            const unsigned long dim1937Kuo2Init[]       =   {   1   ,   1   ,   7   ,   9   ,   15  ,   27  ,   19  ,   111 ,   495 ,   7   ,   1219    ,   2679    ,   3257    ,   7111    ,   23187   ,0 };
            const unsigned long dim1938Kuo2Init[]       =   {   1   ,   1   ,   1   ,   7   ,   5   ,   3   ,   37  ,   247 ,   43  ,   223 ,   1681    ,   225 ,   6573    ,   7523    ,   14433   ,0 };
            const unsigned long dim1939Kuo2Init[]       =   {   1   ,   1   ,   3   ,   3   ,   7   ,   31  ,   71  ,   141 ,   341 ,   319 ,   653 ,   3463    ,   963 ,   10133   ,   27313   ,0 };
            const unsigned long dim1940Kuo2Init[]       =   {   1   ,   1   ,   1   ,   11  ,   3   ,   61  ,   71  ,   9   ,   253 ,   69  ,   527 ,   2927    ,   1637    ,   15141   ,   32423   ,0 };
            const unsigned long dim1941Kuo2Init[]       =   {   1   ,   3   ,   3   ,   3   ,   21  ,   41  ,   119 ,   175 ,   363 ,   629 ,   909 ,   439 ,   8043    ,   12489   ,   18635   ,0 };
            const unsigned long dim1942Kuo2Init[]       =   {   1   ,   1   ,   5   ,   3   ,   11  ,   19  ,   73  ,   87  ,   133 ,   947 ,   537 ,   803 ,   965 ,   9139    ,   7429    ,0 };
            const unsigned long dim1943Kuo2Init[]       =   {   1   ,   1   ,   7   ,   3   ,   9   ,   45  ,   113 ,   213 ,   499 ,   493 ,   811 ,   389 ,   5323    ,   409 ,   17147   ,0 };
            const unsigned long dim1944Kuo2Init[]       =   {   1   ,   3   ,   1   ,   1   ,   9   ,   29  ,   61  ,   231 ,   55  ,   625 ,   1917    ,   3157    ,   4635    ,   11441   ,   24799   ,0 };
            const unsigned long dim1945Kuo2Init[]       =   {   1   ,   3   ,   3   ,   15  ,   15  ,   11  ,   93  ,   245 ,   247 ,   405 ,   1123    ,   2639    ,   1739    ,   11957   ,   19131   ,0 };
            const unsigned long dim1946Kuo2Init[]       =   {   1   ,   3   ,   5   ,   3   ,   19  ,   7   ,   47  ,   241 ,   97  ,   635 ,   1871    ,   2609    ,   5529    ,   11591   ,   13895   ,0 };
            const unsigned long dim1947Kuo2Init[]       =   {   1   ,   1   ,   1   ,   11  ,   11  ,   45  ,   11  ,   13  ,   109 ,   141 ,   83  ,   2109    ,   3259    ,   7441    ,   25881   ,0 };
            const unsigned long dim1948Kuo2Init[]       =   {   1   ,   1   ,   7   ,   13  ,   21  ,   61  ,   73  ,   129 ,   393 ,   175 ,   699 ,   2551    ,   4761    ,   13133   ,   9993    ,0 };
            const unsigned long dim1949Kuo2Init[]       =   {   1   ,   1   ,   7   ,   1   ,   31  ,   17  ,   85  ,   57  ,   457 ,   383 ,   1099    ,   1819    ,   503 ,   1389    ,   28341   ,0 };
            const unsigned long dim1950Kuo2Init[]       =   {   1   ,   3   ,   5   ,   7   ,   3   ,   17  ,   41  ,   33  ,   245 ,   815 ,   31  ,   2259    ,   1293    ,   9051    ,   13241   ,0 };
            const unsigned long dim1951Kuo2Init[]       =   {   1   ,   3   ,   7   ,   7   ,   23  ,   15  ,   81  ,   77  ,   193 ,   201 ,   785 ,   1271    ,   3061    ,   4429    ,   26665   ,0 };
            const unsigned long dim1952Kuo2Init[]       =   {   1   ,   1   ,   5   ,   15  ,   29  ,   23  ,   43  ,   227 ,   119 ,   843 ,   55  ,   1837    ,   51  ,   13115   ,   27725   ,0 };
            const unsigned long dim1953Kuo2Init[]       =   {   1   ,   3   ,   1   ,   11  ,   1   ,   33  ,   25  ,   233 ,   43  ,   367 ,   705 ,   2323    ,   2721    ,   12845   ,   29415   ,0 };
            const unsigned long dim1954Kuo2Init[]       =   {   1   ,   3   ,   5   ,   1   ,   25  ,   23  ,   5   ,   233 ,   507 ,   267 ,   607 ,   3009    ,   5035    ,   1069    ,   30735   ,0 };
            const unsigned long dim1955Kuo2Init[]       =   {   1   ,   1   ,   7   ,   5   ,   27  ,   5   ,   69  ,   5   ,   371 ,   951 ,   657 ,   421 ,   5707    ,   10455   ,   10961   ,0 };
            const unsigned long dim1956Kuo2Init[]       =   {   1   ,   1   ,   3   ,   11  ,   13  ,   13  ,   71  ,   113 ,   381 ,   619 ,   1713    ,   315 ,   7903    ,   4737    ,   24747   ,0 };
            const unsigned long dim1957Kuo2Init[]       =   {   1   ,   3   ,   5   ,   11  ,   15  ,   21  ,   39  ,   221 ,   9   ,   263 ,   1697    ,   3189    ,   7095    ,   7769    ,   10915   ,0 };
            const unsigned long dim1958Kuo2Init[]       =   {   1   ,   1   ,   1   ,   1   ,   17  ,   47  ,   77  ,   27  ,   377 ,   79  ,   1245    ,   1233    ,   3761    ,   3667    ,   12785   ,0 };
            const unsigned long dim1959Kuo2Init[]       =   {   1   ,   1   ,   3   ,   13  ,   11  ,   9   ,   89  ,   57  ,   339 ,   845 ,   393 ,   1717    ,   3415    ,   5725    ,   28655   ,0 };
            const unsigned long dim1960Kuo2Init[]       =   {   1   ,   3   ,   7   ,   3   ,   11  ,   61  ,   89  ,   37  ,   59  ,   117 ,   883 ,   1911    ,   4513    ,   561 ,   31273   ,0 };
            const unsigned long dim1961Kuo2Init[]       =   {   1   ,   1   ,   1   ,   13  ,   21  ,   29  ,   9   ,   147 ,   109 ,   783 ,   1647    ,   2147    ,   7285    ,   4035    ,   8181    ,0 };
            const unsigned long dim1962Kuo2Init[]       =   {   1   ,   3   ,   1   ,   13  ,   29  ,   15  ,   95  ,   71  ,   319 ,   489 ,   601 ,   1731    ,   7927    ,   9115    ,   32607   ,0 };
            const unsigned long dim1963Kuo2Init[]       =   {   1   ,   1   ,   3   ,   13  ,   1   ,   59  ,   53  ,   193 ,   105 ,   653 ,   249 ,   3285    ,   7817    ,   4157    ,   13299   ,0 };
            const unsigned long dim1964Kuo2Init[]       =   {   1   ,   1   ,   7   ,   15  ,   5   ,   63  ,   5   ,   111 ,   427 ,   781 ,   1077    ,   1267    ,   7253    ,   4883    ,   24539   ,0 };
            const unsigned long dim1965Kuo2Init[]       =   {   1   ,   1   ,   1   ,   5   ,   15  ,   13  ,   97  ,   187 ,   175 ,   617 ,   931 ,   3675    ,   8085    ,   14433   ,   22179   ,0 };
            const unsigned long dim1966Kuo2Init[]       =   {   1   ,   3   ,   3   ,   13  ,   29  ,   29  ,   7   ,   45  ,   295 ,   185 ,   15  ,   3171    ,   5701    ,   11099   ,   8663    ,0 };
            const unsigned long dim1967Kuo2Init[]       =   {   1   ,   3   ,   5   ,   15  ,   21  ,   47  ,   37  ,   189 ,   447 ,   123 ,   1979    ,   3633    ,   4875    ,   12667   ,   2349    ,0 };
            const unsigned long dim1968Kuo2Init[]       =   {   1   ,   3   ,   1   ,   3   ,   17  ,   37  ,   97  ,   167 ,   207 ,   347 ,   499 ,   129 ,   3131    ,   6649    ,   10051   ,0 };
            const unsigned long dim1969Kuo2Init[]       =   {   1   ,   3   ,   7   ,   13  ,   21  ,   21  ,   15  ,   125 ,   405 ,   175 ,   89  ,   3299    ,   1321    ,   15595   ,   26015   ,0 };
            const unsigned long dim1970Kuo2Init[]       =   {   1   ,   1   ,   5   ,   1   ,   25  ,   21  ,   11  ,   25  ,   351 ,   1011    ,   753 ,   1813    ,   2229    ,   9675    ,   10073   ,0 };
            const unsigned long dim1971Kuo2Init[]       =   {   1   ,   3   ,   5   ,   1   ,   21  ,   49  ,   61  ,   241 ,   303 ,   833 ,   653 ,   259 ,   4397    ,   545 ,   24857   ,0 };
            const unsigned long dim1972Kuo2Init[]       =   {   1   ,   1   ,   5   ,   5   ,   27  ,   61  ,   99  ,   181 ,   415 ,   945 ,   763 ,   1207    ,   5413    ,   11447   ,   11117   ,0 };
            const unsigned long dim1973Kuo2Init[]       =   {   1   ,   1   ,   3   ,   3   ,   27  ,   11  ,   5   ,   197 ,   211 ,   971 ,   1991    ,   4007    ,   2337    ,   2683    ,   27371   ,0 };
            const unsigned long dim1974Kuo2Init[]       =   {   1   ,   1   ,   5   ,   3   ,   5   ,   39  ,   25  ,   59  ,   223 ,   117 ,   1543    ,   2527    ,   2673    ,   15411   ,   8769    ,0 };
            const unsigned long dim1975Kuo2Init[]       =   {   1   ,   3   ,   7   ,   11  ,   31  ,   7   ,   33  ,   7   ,   319 ,   679 ,   685 ,   953 ,   491 ,   251 ,   16333   ,0 };
            const unsigned long dim1976Kuo2Init[]       =   {   1   ,   1   ,   5   ,   5   ,   1   ,   15  ,   69  ,   87  ,   447 ,   363 ,   1883    ,   937 ,   2691    ,   14777   ,   19757   ,0 };
            const unsigned long dim1977Kuo2Init[]       =   {   1   ,   1   ,   1   ,   1   ,   11  ,   51  ,   21  ,   183 ,   143 ,   577 ,   819 ,   3693    ,   5421    ,   9105    ,   10465   ,0 };
            const unsigned long dim1978Kuo2Init[]       =   {   1   ,   1   ,   7   ,   1   ,   7   ,   7   ,   25  ,   131 ,   63  ,   881 ,   931 ,   2039    ,   1371    ,   7479    ,   7009    ,0 };
            const unsigned long dim1979Kuo2Init[]       =   {   1   ,   1   ,   5   ,   5   ,   17  ,   37  ,   121 ,   101 ,   133 ,   829 ,   495 ,   1031    ,   6843    ,   13863   ,   6543    ,0 };
            const unsigned long dim1980Kuo2Init[]       =   {   1   ,   1   ,   1   ,   9   ,   5   ,   53  ,   15  ,   11  ,   215 ,   295 ,   1401    ,   1633    ,   395 ,   201 ,   18941   ,0 };
            const unsigned long dim1981Kuo2Init[]       =   {   1   ,   3   ,   7   ,   7   ,   21  ,   39  ,   107 ,   153 ,   139 ,   557 ,   683 ,   1993    ,   3495    ,   9929    ,   12067   ,0 };
            const unsigned long dim1982Kuo2Init[]       =   {   1   ,   3   ,   3   ,   13  ,   27  ,   57  ,   63  ,   151 ,   13  ,   827 ,   1079    ,   1209    ,   5631    ,   9627    ,   4765    ,0 };
            const unsigned long dim1983Kuo2Init[]       =   {   1   ,   1   ,   3   ,   13  ,   9   ,   9   ,   75  ,   209 ,   437 ,   475 ,   1905    ,   3021    ,   7097    ,   10637   ,   13301   ,0 };
            const unsigned long dim1984Kuo2Init[]       =   {   1   ,   1   ,   3   ,   1   ,   21  ,   43  ,   45  ,   7   ,   467 ,   825 ,   1189    ,   1525    ,   3295    ,   2575    ,   12071   ,0 };
            const unsigned long dim1985Kuo2Init[]       =   {   1   ,   3   ,   3   ,   1   ,   25  ,   57  ,   23  ,   243 ,   79  ,   39  ,   881 ,   467 ,   4183    ,   13881   ,   26719   ,0 };
            const unsigned long dim1986Kuo2Init[]       =   {   1   ,   1   ,   1   ,   11  ,   27  ,   41  ,   79  ,   223 ,   443 ,   489 ,   1531    ,   3671    ,   7905    ,   9593    ,   175 ,0 };
            const unsigned long dim1987Kuo2Init[]       =   {   1   ,   3   ,   3   ,   5   ,   1   ,   25  ,   89  ,   151 ,   165 ,   681 ,   1099    ,   665 ,   1209    ,   13617   ,   15259   ,0 };
            const unsigned long dim1988Kuo2Init[]       =   {   1   ,   3   ,   3   ,   3   ,   17  ,   25  ,   13  ,   85  ,   179 ,   335 ,   951 ,   4085    ,   3251    ,   15061   ,   14609   ,0 };
            const unsigned long dim1989Kuo2Init[]       =   {   1   ,   1   ,   5   ,   11  ,   15  ,   25  ,   89  ,   135 ,   27  ,   323 ,   1143    ,   7   ,   3569    ,   6841    ,   9243    ,0 };
            const unsigned long dim1990Kuo2Init[]       =   {   1   ,   1   ,   7   ,   5   ,   13  ,   43  ,   83  ,   23  ,   477 ,   229 ,   1623    ,   2641    ,   1235    ,   681 ,   17379   ,0 };
            const unsigned long dim1991Kuo2Init[]       =   {   1   ,   3   ,   1   ,   13  ,   1   ,   17  ,   77  ,   125 ,   121 ,   253 ,   1329    ,   1081    ,   1191    ,   7477    ,   13441   ,0 };
            const unsigned long dim1992Kuo2Init[]       =   {   1   ,   1   ,   7   ,   15  ,   5   ,   61  ,   23  ,   251 ,   407 ,   373 ,   1399    ,   1907    ,   4145    ,   15997   ,   12069   ,0 };
            const unsigned long dim1993Kuo2Init[]       =   {   1   ,   1   ,   3   ,   7   ,   9   ,   27  ,   85  ,   209 ,   21  ,   225 ,   745 ,   1405    ,   813 ,   1605    ,   17463   ,0 };
            const unsigned long dim1994Kuo2Init[]       =   {   1   ,   1   ,   7   ,   5   ,   19  ,   47  ,   111 ,   215 ,   335 ,   211 ,   1437    ,   425 ,   2325    ,   15761   ,   13937   ,0 };
            const unsigned long dim1995Kuo2Init[]       =   {   1   ,   3   ,   5   ,   9   ,   31  ,   21  ,   113 ,   89  ,   369 ,   165 ,   1673    ,   3143    ,   6765    ,   13897   ,   15295   ,0 };
            const unsigned long dim1996Kuo2Init[]       =   {   1   ,   3   ,   1   ,   1   ,   23  ,   57  ,   97  ,   9   ,   219 ,   513 ,   1001    ,   1903    ,   6033    ,   4561    ,   8157    ,0 };
            const unsigned long dim1997Kuo2Init[]       =   {   1   ,   3   ,   5   ,   13  ,   17  ,   59  ,   125 ,   93  ,   397 ,   277 ,   1273    ,   529 ,   3881    ,   3345    ,   32553   ,0 };
            const unsigned long dim1998Kuo2Init[]       =   {   1   ,   3   ,   1   ,   3   ,   29  ,   3   ,   73  ,   201 ,   129 ,   747 ,   1417    ,   77  ,   4399    ,   10283   ,   3677    ,0 };
            const unsigned long dim1999Kuo2Init[]       =   {   1   ,   3   ,   5   ,   11  ,   15  ,   13  ,   15  ,   193 ,   139 ,   683 ,   839 ,   1849    ,   1229    ,   7903    ,   15797   ,0 };
            const unsigned long dim2000Kuo2Init[]       =   {   1   ,   1   ,   7   ,   1   ,   19  ,   37  ,   103 ,   173 ,   63  ,   269 ,   1125    ,   425 ,   1235    ,   4193    ,   10095   ,0 };
            const unsigned long dim2001Kuo2Init[]       =   {   1   ,   3   ,   1   ,   3   ,   31  ,   33  ,   101 ,   71  ,   333 ,   565 ,   651 ,   1741    ,   3613    ,   2641    ,   27201   ,0 };
            const unsigned long dim2002Kuo2Init[]       =   {   1   ,   1   ,   7   ,   7   ,   9   ,   43  ,   111 ,   231 ,   383 ,   221 ,   1239    ,   1927    ,   715 ,   7663    ,   24709   ,0 };
            const unsigned long dim2003Kuo2Init[]       =   {   1   ,   1   ,   7   ,   1   ,   3   ,   41  ,   117 ,   129 ,   53  ,   527 ,   1973    ,   3377    ,   2803    ,   10855   ,   9405    ,0 };
            const unsigned long dim2004Kuo2Init[]       =   {   1   ,   3   ,   5   ,   15  ,   15  ,   61  ,   29  ,   75  ,   141 ,   517 ,   255 ,   1797    ,   1291    ,   13859   ,   10255   ,0 };
            const unsigned long dim2005Kuo2Init[]       =   {   1   ,   1   ,   5   ,   9   ,   1   ,   55  ,   99  ,   133 ,   9   ,   793 ,   1525    ,   2565    ,   1755    ,   7293    ,   27221   ,0 };
            const unsigned long dim2006Kuo2Init[]       =   {   1   ,   3   ,   7   ,   1   ,   19  ,   25  ,   113 ,   47  ,   117 ,   285 ,   1937    ,   513 ,   75  ,   2149    ,   25369   ,0 };
            const unsigned long dim2007Kuo2Init[]       =   {   1   ,   3   ,   3   ,   7   ,   3   ,   45  ,   1   ,   145 ,   501 ,   197 ,   423 ,   1131    ,   4649    ,   11027   ,   25237   ,0 };
            const unsigned long dim2008Kuo2Init[]       =   {   1   ,   1   ,   1   ,   9   ,   29  ,   49  ,   17  ,   85  ,   137 ,   765 ,   649 ,   809 ,   103 ,   7537    ,   5603    ,0 };
            const unsigned long dim2009Kuo2Init[]       =   {   1   ,   3   ,   1   ,   13  ,   23  ,   63  ,   85  ,   83  ,   135 ,   495 ,   2023    ,   2331    ,   1743    ,   16221   ,   2421    ,0 };
            const unsigned long dim2010Kuo2Init[]       =   {   1   ,   1   ,   1   ,   3   ,   25  ,   15  ,   75  ,   123 ,   309 ,   729 ,   481 ,   2335    ,   6963    ,   16309   ,   24593   ,0 };
            const unsigned long dim2011Kuo2Init[]       =   {   1   ,   3   ,   1   ,   7   ,   1   ,   3   ,   41  ,   57  ,   353 ,   821 ,   1439    ,   3445    ,   2295    ,   15269   ,   15821   ,0 };
            const unsigned long dim2012Kuo2Init[]       =   {   1   ,   3   ,   7   ,   1   ,   19  ,   41  ,   107 ,   159 ,   361 ,   1017    ,   1435    ,   2593    ,   3849    ,   1139    ,   5059    ,0 };
            const unsigned long dim2013Kuo2Init[]       =   {   1   ,   1   ,   7   ,   5   ,   23  ,   63  ,   23  ,   67  ,   263 ,   939 ,   1167    ,   3367    ,   2003    ,   7903    ,   1361    ,0 };
            const unsigned long dim2014Kuo2Init[]       =   {   1   ,   3   ,   5   ,   15  ,   11  ,   21  ,   109 ,   135 ,   219 ,   183 ,   1303    ,   2023    ,   6471    ,   8203    ,   21717   ,0 };
            const unsigned long dim2015Kuo2Init[]       =   {   1   ,   3   ,   3   ,   11  ,   11  ,   7   ,   109 ,   183 ,   313 ,   701 ,   1661    ,   2215    ,   7993    ,   6097    ,   15583   ,0 };
            const unsigned long dim2016Kuo2Init[]       =   {   1   ,   1   ,   7   ,   9   ,   5   ,   55  ,   85  ,   123 ,   507 ,   303 ,   1701    ,   767 ,   313 ,   12517   ,   10969   ,0 };
            const unsigned long dim2017Kuo2Init[]       =   {   1   ,   3   ,   1   ,   15  ,   1   ,   31  ,   97  ,   249 ,   407 ,   829 ,   693 ,   135 ,   5929    ,   13287   ,   23105   ,0 };
            const unsigned long dim2018Kuo2Init[]       =   {   1   ,   3   ,   3   ,   15  ,   23  ,   45  ,   13  ,   209 ,   365 ,   715 ,   827 ,   1451    ,   3817    ,   9039    ,   8601    ,0 };
            const unsigned long dim2019Kuo2Init[]       =   {   1   ,   1   ,   5   ,   3   ,   3   ,   5   ,   41  ,   175 ,   65  ,   467 ,   179 ,   1941    ,   2261    ,   14143   ,   10277   ,0 };
            const unsigned long dim2020Kuo2Init[]       =   {   1   ,   1   ,   7   ,   1   ,   15  ,   61  ,   37  ,   231 ,   275 ,   69  ,   7   ,   417 ,   8131    ,   12569   ,   16275   ,0 };
            const unsigned long dim2021Kuo2Init[]       =   {   1   ,   1   ,   7   ,   11  ,   3   ,   49  ,   1   ,   243 ,   237 ,   303 ,   2033    ,   1501    ,   8059    ,   1479    ,   1675    ,0 };
            const unsigned long dim2022Kuo2Init[]       =   {   1   ,   1   ,   3   ,   5   ,   21  ,   49  ,   87  ,   5   ,   235 ,   343 ,   1635    ,   3459    ,   3961    ,   14433   ,   10965   ,0 };
            const unsigned long dim2023Kuo2Init[]       =   {   1   ,   1   ,   7   ,   13  ,   29  ,   5   ,   105 ,   167 ,   57  ,   247 ,   1683    ,   1011    ,   2071    ,   10681   ,   679 ,0 };
            const unsigned long dim2024Kuo2Init[]       =   {   1   ,   1   ,   3   ,   7   ,   11  ,   31  ,   39  ,   19  ,   501 ,   945 ,   2043    ,   3591    ,   657 ,   8087    ,   21159   ,0 };
            const unsigned long dim2025Kuo2Init[]       =   {   1   ,   3   ,   5   ,   9   ,   7   ,   23  ,   3   ,   5   ,   35  ,   101 ,   409 ,   3769    ,   7163    ,   739 ,   28421   ,0 };
            const unsigned long dim2026Kuo2Init[]       =   {   1   ,   3   ,   5   ,   1   ,   23  ,   61  ,   65  ,   105 ,   303 ,   765 ,   1483    ,   883 ,   5743    ,   13479   ,   21489   ,0 };
            const unsigned long dim2027Kuo2Init[]       =   {   1   ,   1   ,   7   ,   13  ,   11  ,   5   ,   17  ,   237 ,   105 ,   539 ,   1237    ,   2727    ,   3781    ,   13195   ,   17029   ,0 };
            const unsigned long dim2028Kuo2Init[]       =   {   1   ,   3   ,   3   ,   15  ,   15  ,   23  ,   47  ,   129 ,   485 ,   797 ,   705 ,   931 ,   3865    ,   5029    ,   27847   ,0 };
            const unsigned long dim2029Kuo2Init[]       =   {   1   ,   3   ,   5   ,   13  ,   23  ,   61  ,   77  ,   221 ,   155 ,   429 ,   1933    ,   3007    ,   8001    ,   2137    ,   15737   ,0 };
            const unsigned long dim2030Kuo2Init[]       =   {   1   ,   3   ,   3   ,   15  ,   3   ,   57  ,   79  ,   95  ,   419 ,   953 ,   449 ,   1843    ,   6893    ,   1061    ,   28427   ,0 };
            const unsigned long dim2031Kuo2Init[]       =   {   1   ,   3   ,   3   ,   9   ,   19  ,   9   ,   39  ,   5   ,   189 ,   3   ,   181 ,   3091    ,   4215    ,   14881   ,   18273   ,0 };
            const unsigned long dim2032Kuo2Init[]       =   {   1   ,   1   ,   3   ,   13  ,   17  ,   63  ,   27  ,   49  ,   483 ,   947 ,   173 ,   3667    ,   3529    ,   15343   ,   23189   ,0 };
            const unsigned long dim2033Kuo2Init[]       =   {   1   ,   1   ,   3   ,   9   ,   17  ,   51  ,   37  ,   245 ,   177 ,   231 ,   689 ,   29  ,   1029    ,   12397   ,   26265   ,0 };
            const unsigned long dim2034Kuo2Init[]       =   {   1   ,   3   ,   1   ,   13  ,   15  ,   35  ,   55  ,   5   ,   303 ,   559 ,   1853    ,   3859    ,   1879    ,   3167    ,   6903    ,0 };
            const unsigned long dim2035Kuo2Init[]       =   {   1   ,   3   ,   3   ,   9   ,   9   ,   47  ,   59  ,   163 ,   373 ,   621 ,   1349    ,   2661    ,   1787    ,   13611   ,   28199   ,0 };
            const unsigned long dim2036Kuo2Init[]       =   {   1   ,   3   ,   3   ,   9   ,   15  ,   55  ,   93  ,   193 ,   267 ,   325 ,   1705    ,   155 ,   3001    ,   407 ,   20829   ,0 };
            const unsigned long dim2037Kuo2Init[]       =   {   1   ,   3   ,   5   ,   15  ,   13  ,   9   ,   65  ,   123 ,   325 ,   313 ,   25  ,   991 ,   2227    ,   6597    ,   30453   ,0 };
            const unsigned long dim2038Kuo2Init[]       =   {   1   ,   1   ,   5   ,   1   ,   23  ,   51  ,   61  ,   229 ,   27  ,   285 ,   1185    ,   1943    ,   7603    ,   10565   ,   22741   ,0 };
            const unsigned long dim2039Kuo2Init[]       =   {   1   ,   3   ,   7   ,   3   ,   17  ,   43  ,   21  ,   153 ,   359 ,   437 ,   1275    ,   3589    ,   1089    ,   3941    ,   16777   ,0 };
            const unsigned long dim2040Kuo2Init[]       =   {   1   ,   3   ,   5   ,   3   ,   19  ,   43  ,   127 ,   213 ,   181 ,   255 ,   929 ,   2553    ,   1101    ,   231 ,   2977    ,0 };
            const unsigned long dim2041Kuo2Init[]       =   {   1   ,   3   ,   7   ,   11  ,   7   ,   49  ,   39  ,   209 ,   141 ,   691 ,   1903    ,   1211    ,   687 ,   13107   ,   17055   ,0 };
            const unsigned long dim2042Kuo2Init[]       =   {   1   ,   1   ,   5   ,   1   ,   15  ,   21  ,   65  ,   209 ,   305 ,   65  ,   1383    ,   2247    ,   5813    ,   8961    ,   16281   ,0 };
            const unsigned long dim2043Kuo2Init[]       =   {   1   ,   3   ,   7   ,   11  ,   27  ,   1   ,   71  ,   97  ,   129 ,   317 ,   601 ,   981 ,   1947    ,   8897    ,   23489   ,0 };
            const unsigned long dim2044Kuo2Init[]       =   {   1   ,   1   ,   3   ,   9   ,   17  ,   35  ,   57  ,   37  ,   243 ,   217 ,   703 ,   3159    ,   4791    ,   1365    ,   14885   ,0 };
            const unsigned long dim2045Kuo2Init[]       =   {   1   ,   3   ,   5   ,   7   ,   7   ,   43  ,   59  ,   253 ,   141 ,   537 ,   547 ,   955 ,   5351    ,   13087   ,   5465    ,0 };
            const unsigned long dim2046Kuo2Init[]       =   {   1   ,   1   ,   5   ,   9   ,   15  ,   53  ,   47  ,   105 ,   55  ,   1017    ,   819 ,   3355    ,   3653    ,   11051   ,   29509   ,0 };
            const unsigned long dim2047Kuo2Init[]       =   {   1   ,   1   ,   3   ,   7   ,   17  ,   39  ,   13  ,   125 ,   161 ,   985 ,   1753    ,   1327    ,   7635    ,   11045   ,   3935    ,0 };
            const unsigned long dim2048Kuo2Init[]       =   {   1   ,   1   ,   7   ,   9   ,   1   ,   19  ,   55  ,   227 ,   111 ,   347 ,   85  ,   3871    ,   7153    ,   10527   ,   3137    ,0 };
            const unsigned long dim2049Kuo2Init[]       =   {   1   ,   3   ,   5   ,   9   ,   13  ,   7   ,   35  ,   25  ,   17  ,   269 ,   611 ,   1523    ,   3539    ,   15611   ,   11837   ,0 };
            const unsigned long dim2050Kuo2Init[]       =   {   1   ,   1   ,   5   ,   7   ,   29  ,   13  ,   109 ,   253 ,   355 ,   25  ,   243 ,   1433    ,   7463    ,   10149   ,   31469   ,0 };
            const unsigned long dim2051Kuo2Init[]       =   {   1   ,   3   ,   3   ,   15  ,   1   ,   51  ,   85  ,   27  ,   137 ,   299 ,   1943    ,   3805    ,   2661    ,   8091    ,   385 ,0 };
            const unsigned long dim2052Kuo2Init[]       =   {   1   ,   1   ,   5   ,   11  ,   19  ,   39  ,   103 ,   89  ,   285 ,   305 ,   1157    ,   149 ,   6661    ,   1423    ,   27163   ,0 };
            const unsigned long dim2053Kuo2Init[]       =   {   1   ,   1   ,   7   ,   9   ,   25  ,   35  ,   7   ,   227 ,   125 ,   811 ,   1867    ,   1851    ,   5271    ,   11405   ,   21457   ,0 };
            const unsigned long dim2054Kuo2Init[]       =   {   1   ,   1   ,   7   ,   1   ,   31  ,   19  ,   103 ,   239 ,   301 ,   395 ,   1737    ,   3467    ,   7153    ,   12437   ,   25795   ,0 };
            const unsigned long dim2055Kuo2Init[]       =   {   1   ,   3   ,   3   ,   15  ,   25  ,   43  ,   69  ,   155 ,   437 ,   9   ,   863 ,   2497    ,   6857    ,   3947    ,   12231   ,0 };
            const unsigned long dim2056Kuo2Init[]       =   {   1   ,   3   ,   7   ,   11  ,   3   ,   63  ,   27  ,   207 ,   87  ,   727 ,   791 ,   1663    ,   8121    ,   13149   ,   3927    ,0 };
            const unsigned long dim2057Kuo2Init[]       =   {   1   ,   3   ,   1   ,   7   ,   9   ,   19  ,   123 ,   131 ,   415 ,   565 ,   1571    ,   1085    ,   2989    ,   13693   ,   32487   ,0 };
            const unsigned long dim2058Kuo2Init[]       =   {   1   ,   3   ,   7   ,   9   ,   11  ,   27  ,   117 ,   63  ,   255 ,   297 ,   1147    ,   1501    ,   3435    ,   5861    ,   21351   ,0 };
            const unsigned long dim2059Kuo2Init[]       =   {   1   ,   3   ,   1   ,   5   ,   19  ,   5   ,   115 ,   89  ,   153 ,   705 ,   1233    ,   1803    ,   7237    ,   983 ,   9299    ,0 };
            const unsigned long dim2060Kuo2Init[]       =   {   1   ,   3   ,   3   ,   13  ,   21  ,   13  ,   89  ,   135 ,   5   ,   221 ,   191 ,   2611    ,   6007    ,   9177    ,   11101   ,0 };
            const unsigned long dim2061Kuo2Init[]       =   {   1   ,   3   ,   7   ,   15  ,   9   ,   43  ,   123 ,   123 ,   301 ,   233 ,   1727    ,   3973    ,   5241    ,   13437   ,   24507   ,0 };
            const unsigned long dim2062Kuo2Init[]       =   {   1   ,   3   ,   3   ,   5   ,   19  ,   29  ,   33  ,   127 ,   185 ,   319 ,   1767    ,   1133    ,   95  ,   7745    ,   32213   ,0 };
            const unsigned long dim2063Kuo2Init[]       =   {   1   ,   3   ,   7   ,   11  ,   13  ,   1   ,   91  ,   189 ,   75  ,   91  ,   583 ,   2317    ,   5257    ,   12133   ,   1951    ,0 };
            const unsigned long dim2064Kuo2Init[]       =   {   1   ,   1   ,   3   ,   9   ,   23  ,   23  ,   65  ,   7   ,   157 ,   143 ,   781 ,   997 ,   2357    ,   4907    ,   16797   ,0 };
            const unsigned long dim2065Kuo2Init[]       =   {   1   ,   1   ,   1   ,   11  ,   5   ,   43  ,   45  ,   131 ,   331 ,   253 ,   1975    ,   415 ,   2571    ,   14905   ,   9595    ,0 };
            const unsigned long dim2066Kuo2Init[]       =   {   1   ,   3   ,   7   ,   11  ,   1   ,   11  ,   95  ,   133 ,   393 ,   117 ,   833 ,   2661    ,   1683    ,   14333   ,   26279   ,0 };
            const unsigned long dim2067Kuo2Init[]       =   {   1   ,   3   ,   7   ,   5   ,   23  ,   51  ,   57  ,   9   ,   63  ,   377 ,   2043    ,   1967    ,   7039    ,   13783   ,   16075   ,0 };
            const unsigned long dim2068Kuo2Init[]       =   {   1   ,   3   ,   5   ,   5   ,   13  ,   49  ,   65  ,   59  ,   449 ,   665 ,   2017    ,   375 ,   3781    ,   15139   ,   26341   ,0 };
            const unsigned long dim2069Kuo2Init[]       =   {   1   ,   1   ,   7   ,   13  ,   15  ,   63  ,   83  ,   229 ,   465 ,   319 ,   469 ,   951 ,   2405    ,   15357   ,   6051    ,0 };
            const unsigned long dim2070Kuo2Init[]       =   {   1   ,   3   ,   3   ,   7   ,   15  ,   37  ,   97  ,   37  ,   89  ,   717 ,   449 ,   1821    ,   3319    ,   4957    ,   5455    ,0 };
            const unsigned long dim2071Kuo2Init[]       =   {   1   ,   1   ,   5   ,   3   ,   3   ,   7   ,   55  ,   39  ,   131 ,   497 ,   349 ,   4083    ,   6695    ,   14939   ,   13979   ,0 };
            const unsigned long dim2072Kuo2Init[]       =   {   1   ,   3   ,   3   ,   7   ,   15  ,   45  ,   109 ,   239 ,   415 ,   453 ,   729 ,   835 ,   4585    ,   12751   ,   8595    ,0 };
            const unsigned long dim2073Kuo2Init[]       =   {   1   ,   1   ,   7   ,   9   ,   29  ,   1   ,   13  ,   85  ,   185 ,   419 ,   1269    ,   981 ,   5079    ,   2723    ,   5275    ,0 };
            const unsigned long dim2074Kuo2Init[]       =   {   1   ,   1   ,   7   ,   11  ,   7   ,   41  ,   121 ,   75  ,   91  ,   943 ,   1487    ,   995 ,   6815    ,   15881   ,   27113   ,0 };
            const unsigned long dim2075Kuo2Init[]       =   {   1   ,   1   ,   7   ,   9   ,   1   ,   53  ,   83  ,   209 ,   327 ,   547 ,   1505    ,   1517    ,   6105    ,   5903    ,   13823   ,0 };
            const unsigned long dim2076Kuo2Init[]       =   {   1   ,   1   ,   3   ,   15  ,   25  ,   1   ,   33  ,   61  ,   249 ,   367 ,   1733    ,   3351    ,   273 ,   10053   ,   27115   ,0 };
            const unsigned long dim2077Kuo2Init[]       =   {   1   ,   3   ,   3   ,   11  ,   29  ,   27  ,   75  ,   143 ,   103 ,   509 ,   813 ,   1881    ,   2631    ,   933 ,   31141   ,0 };
            const unsigned long dim2078Kuo2Init[]       =   {   1   ,   3   ,   3   ,   7   ,   31  ,   9   ,   27  ,   69  ,   291 ,   927 ,   183 ,   1755    ,   7383    ,   7721    ,   4405    ,0 };
            const unsigned long dim2079Kuo2Init[]       =   {   1   ,   3   ,   1   ,   11  ,   25  ,   23  ,   89  ,   99  ,   101 ,   725 ,   1791    ,   2669    ,   7055    ,   6703    ,   27507   ,0 };
            const unsigned long dim2080Kuo2Init[]       =   {   1   ,   3   ,   5   ,   15  ,   1   ,   41  ,   23  ,   103 ,   123 ,   863 ,   1643    ,   947 ,   1349    ,   549 ,   29299   ,0 };
            const unsigned long dim2081Kuo2Init[]       =   {   1   ,   1   ,   3   ,   3   ,   9   ,   53  ,   67  ,   187 ,   203 ,   565 ,   1455    ,   645 ,   7983    ,   5311    ,   8827    ,0 };
            const unsigned long dim2082Kuo2Init[]       =   {   1   ,   1   ,   3   ,   7   ,   11  ,   29  ,   35  ,   207 ,   449 ,   537 ,   245 ,   945 ,   4067    ,   10961   ,   27301   ,0 };
            const unsigned long dim2083Kuo2Init[]       =   {   1   ,   3   ,   1   ,   13  ,   3   ,   23  ,   15  ,   171 ,   439 ,   217 ,   1245    ,   3301    ,   289 ,   6895    ,   30095   ,0 };
            const unsigned long dim2084Kuo2Init[]       =   {   1   ,   3   ,   5   ,   9   ,   21  ,   47  ,   71  ,   191 ,   425 ,   11  ,   1831    ,   3111    ,   7399    ,   13689   ,   28163   ,0 };
            const unsigned long dim2085Kuo2Init[]       =   {   1   ,   1   ,   5   ,   15  ,   5   ,   61  ,   97  ,   41  ,   451 ,   459 ,   1007    ,   2439    ,   5613    ,   7985    ,   9255    ,0 };
            const unsigned long dim2086Kuo2Init[]       =   {   1   ,   1   ,   1   ,   1   ,   9   ,   57  ,   61  ,   143 ,   3   ,   753 ,   1839    ,   2623    ,   5261    ,   6169    ,   13195   ,0 };
            const unsigned long dim2087Kuo2Init[]       =   {   1   ,   1   ,   1   ,   3   ,   19  ,   23  ,   15  ,   31  ,   481 ,   123 ,   653 ,   2763    ,   5663    ,   12771   ,   7469    ,0 };
            const unsigned long dim2088Kuo2Init[]       =   {   1   ,   1   ,   1   ,   11  ,   17  ,   55  ,   111 ,   181 ,   293 ,   9   ,   41  ,   3625    ,   2865    ,   9109    ,   14261   ,0 };
            const unsigned long dim2089Kuo2Init[]       =   {   1   ,   1   ,   5   ,   15  ,   27  ,   13  ,   45  ,   109 ,   321 ,   609 ,   551 ,   4071    ,   6467    ,   8245    ,   6889    ,0 };
            const unsigned long dim2090Kuo2Init[]       =   {   1   ,   1   ,   1   ,   11  ,   23  ,   15  ,   75  ,   9   ,   261 ,   189 ,   287 ,   2479    ,   2319    ,   2647    ,   3415    ,0 };
            const unsigned long dim2091Kuo2Init[]       =   {   1   ,   3   ,   1   ,   3   ,   31  ,   47  ,   45  ,   73  ,   439 ,   789 ,   687 ,   3205    ,   863 ,   2471    ,   27997   ,0 };
            const unsigned long dim2092Kuo2Init[]       =   {   1   ,   3   ,   3   ,   13  ,   17  ,   47  ,   111 ,   205 ,   307 ,   299 ,   439 ,   965 ,   6093    ,   3215    ,   29381   ,0 };
            const unsigned long dim2093Kuo2Init[]       =   {   1   ,   1   ,   7   ,   11  ,   11  ,   31  ,   119 ,   93  ,   385 ,   149 ,   843 ,   3081    ,   6793    ,   4427    ,   23967   ,0 };
            const unsigned long dim2094Kuo2Init[]       =   {   1   ,   3   ,   7   ,   11  ,   7   ,   9   ,   27  ,   135 ,   465 ,   169 ,   1229    ,   1491    ,   3527    ,   13801   ,   13187   ,0 };
            const unsigned long dim2095Kuo2Init[]       =   {   1   ,   3   ,   7   ,   1   ,   27  ,   43  ,   9   ,   213 ,   95  ,   273 ,   1067    ,   1819    ,   4087    ,   15955   ,   7865    ,0 };
            const unsigned long dim2096Kuo2Init[]       =   {   1   ,   1   ,   3   ,   5   ,   25  ,   53  ,   49  ,   15  ,   159 ,   713 ,   203 ,   1599    ,   4807    ,   7275    ,   20589   ,0 };
            const unsigned long dim2097Kuo2Init[]       =   {   1   ,   3   ,   1   ,   7   ,   9   ,   33  ,   5   ,   41  ,   289 ,   611 ,   153 ,   1019    ,   5571    ,   8459    ,   2903    ,0 };
            const unsigned long dim2098Kuo2Init[]       =   {   1   ,   1   ,   5   ,   11  ,   5   ,   9   ,   9   ,   245 ,   205 ,   821 ,   73  ,   2907    ,   2061    ,   10255   ,   24837   ,0 };
            const unsigned long dim2099Kuo2Init[]       =   {   1   ,   3   ,   3   ,   15  ,   3   ,   29  ,   79  ,   139 ,   57  ,   181 ,   741 ,   667 ,   3251    ,   15741   ,   2349    ,0 };
            const unsigned long dim2100Kuo2Init[]       =   {   1   ,   1   ,   7   ,   11  ,   15  ,   31  ,   63  ,   115 ,   475 ,   801 ,   195 ,   155 ,   2167    ,   9215    ,   11041   ,0 };
            const unsigned long dim2101Kuo2Init[]       =   {   1   ,   3   ,   7   ,   3   ,   7   ,   45  ,   17  ,   151 ,   353 ,   751 ,   373 ,   1523    ,   6271    ,   10307   ,   637 ,0 };
            const unsigned long dim2102Kuo2Init[]       =   {   1   ,   1   ,   1   ,   13  ,   15  ,   59  ,   99  ,   143 ,   261 ,   349 ,   327 ,   3931    ,   3365    ,   12179   ,   3017    ,0 };
            const unsigned long dim2103Kuo2Init[]       =   {   1   ,   3   ,   5   ,   1   ,   27  ,   13  ,   105 ,   127 ,   177 ,   857 ,   405 ,   185 ,   5793    ,   8043    ,   11671   ,0 };
            const unsigned long dim2104Kuo2Init[]       =   {   1   ,   1   ,   7   ,   5   ,   9   ,   51  ,   127 ,   47  ,   421 ,   687 ,   961 ,   1413    ,   6619    ,   3459    ,   9093    ,0 };
            const unsigned long dim2105Kuo2Init[]       =   {   1   ,   1   ,   7   ,   11  ,   15  ,   45  ,   7   ,   123 ,   479 ,   55  ,   491 ,   2901    ,   5183    ,   6081    ,   31719   ,0 };
            const unsigned long dim2106Kuo2Init[]       =   {   1   ,   3   ,   3   ,   13  ,   3   ,   49  ,   29  ,   147 ,   453 ,   659 ,   753 ,   233 ,   251 ,   12445   ,   6573    ,0 };
            const unsigned long dim2107Kuo2Init[]       =   {   1   ,   1   ,   3   ,   9   ,   7   ,   41  ,   3   ,   59  ,   187 ,   37  ,   1503    ,   4079    ,   5687    ,   16213   ,   9573    ,0 };
            const unsigned long dim2108Kuo2Init[]       =   {   1   ,   3   ,   7   ,   13  ,   19  ,   49  ,   75  ,   173 ,   219 ,   889 ,   859 ,   2021    ,   2249    ,   4031    ,   29485   ,0 };
            const unsigned long dim2109Kuo2Init[]       =   {   1   ,   3   ,   7   ,   9   ,   3   ,   21  ,   37  ,   79  ,   277 ,   307 ,   465 ,   1219    ,   4099    ,   5285    ,   27883   ,0 };
            const unsigned long dim2110Kuo2Init[]       =   {   1   ,   3   ,   5   ,   3   ,   19  ,   49  ,   97  ,   7   ,   483 ,   995 ,   775 ,   883 ,   2011    ,   16321   ,   31869   ,0 };
            const unsigned long dim2111Kuo2Init[]       =   {   1   ,   3   ,   7   ,   1   ,   25  ,   59  ,   113 ,   177 ,   507 ,   911 ,   1251    ,   571 ,   917 ,   2655    ,   31463   ,0 };
            const unsigned long dim2112Kuo2Init[]       =   {   1   ,   3   ,   1   ,   5   ,   15  ,   23  ,   47  ,   79  ,   401 ,   245 ,   317 ,   1033    ,   3377    ,   2605    ,   13021   ,0 };
            const unsigned long dim2113Kuo2Init[]       =   {   1   ,   1   ,   3   ,   7   ,   3   ,   1   ,   41  ,   253 ,   507 ,   283 ,   707 ,   1305    ,   2353    ,   5721    ,   15469   ,0 };
            const unsigned long dim2114Kuo2Init[]       =   {   1   ,   3   ,   1   ,   7   ,   25  ,   25  ,   15  ,   79  ,   193 ,   235 ,   1981    ,   2059    ,   2057    ,   1647    ,   5443    ,0 };
            const unsigned long dim2115Kuo2Init[]       =   {   1   ,   1   ,   1   ,   3   ,   7   ,   37  ,   51  ,   169 ,   163 ,   223 ,   457 ,   1841    ,   857 ,   15847   ,   27369   ,0 };
            const unsigned long dim2116Kuo2Init[]       =   {   1   ,   1   ,   3   ,   11  ,   23  ,   49  ,   93  ,   89  ,   347 ,   213 ,   615 ,   1305    ,   885 ,   6195    ,   6659    ,0 };
            const unsigned long dim2117Kuo2Init[]       =   {   1   ,   3   ,   7   ,   9   ,   9   ,   27  ,   3   ,   217 ,   337 ,   977 ,   1109    ,   1915    ,   677 ,   8927    ,   4449    ,0 };
            const unsigned long dim2118Kuo2Init[]       =   {   1   ,   1   ,   7   ,   5   ,   11  ,   39  ,   23  ,   189 ,   467 ,   735 ,   817 ,   2269    ,   5959    ,   2933    ,   15667   ,0 };
            const unsigned long dim2119Kuo2Init[]       =   {   1   ,   1   ,   5   ,   5   ,   21  ,   33  ,   21  ,   241 ,   75  ,   751 ,   671 ,   3053    ,   7865    ,   4261    ,   13455   ,0 };
            const unsigned long dim2120Kuo2Init[]       =   {   1   ,   3   ,   1   ,   7   ,   25  ,   11  ,   21  ,   149 ,   429 ,   275 ,   1815    ,   1153    ,   4471    ,   14993   ,   3543    ,0 };
            const unsigned long dim2121Kuo2Init[]       =   {   1   ,   3   ,   5   ,   1   ,   3   ,   47  ,   9   ,   113 ,   423 ,   205 ,   1905    ,   57  ,   6157    ,   6335    ,   7425    ,0 };
            const unsigned long dim2122Kuo2Init[]       =   {   1   ,   1   ,   7   ,   11  ,   17  ,   13  ,   97  ,   99  ,   145 ,   319 ,   945 ,   1559    ,   3379    ,   12559   ,   22479   ,0 };
            const unsigned long dim2123Kuo2Init[]       =   {   1   ,   1   ,   7   ,   9   ,   3   ,   47  ,   105 ,   251 ,   87  ,   293 ,   1191    ,   2211    ,   4631    ,   8443    ,   11789   ,0 };
            const unsigned long dim2124Kuo2Init[]       =   {   1   ,   1   ,   1   ,   3   ,   5   ,   1   ,   41  ,   115 ,   331 ,   401 ,   1649    ,   179 ,   1639    ,   16241   ,   18481   ,0 };
            const unsigned long dim2125Kuo2Init[]       =   {   1   ,   3   ,   1   ,   7   ,   15  ,   33  ,   117 ,   121 ,   81  ,   807 ,   1897    ,   1731    ,   2793    ,   12531   ,   8863    ,0 };
            const unsigned long dim2126Kuo2Init[]       =   {   1   ,   3   ,   1   ,   13  ,   1   ,   41  ,   83  ,   5   ,   331 ,   423 ,   855 ,   1327    ,   5799    ,   3679    ,   21461   ,0 };
            const unsigned long dim2127Kuo2Init[]       =   {   1   ,   3   ,   1   ,   3   ,   19  ,   37  ,   75  ,   39  ,   413 ,   123 ,   1073    ,   2299    ,   5379    ,   12805   ,   3527    ,0 };
            const unsigned long dim2128Kuo2Init[]       =   {   1   ,   3   ,   3   ,   1   ,   7   ,   17  ,   25  ,   227 ,   383 ,   501 ,   673 ,   2609    ,   1843    ,   3367    ,   9901    ,0 };
            const unsigned long dim2129Kuo2Init[]       =   {   1   ,   1   ,   7   ,   11  ,   25  ,   29  ,   95  ,   33  ,   431 ,   257 ,   1013    ,   3755    ,   2813    ,   15247   ,   1045    ,0 };
            const unsigned long dim2130Kuo2Init[]       =   {   1   ,   1   ,   7   ,   3   ,   15  ,   25  ,   79  ,   67  ,   175 ,   697 ,   1867    ,   1967    ,   8023    ,   2017    ,   15449   ,0 };
            const unsigned long dim2131Kuo2Init[]       =   {   1   ,   3   ,   3   ,   1   ,   21  ,   25  ,   67  ,   177 ,   479 ,   81  ,   495 ,   3531    ,   1997    ,   5169    ,   15861   ,0 };
            const unsigned long dim2132Kuo2Init[]       =   {   1   ,   3   ,   3   ,   9   ,   7   ,   23  ,   57  ,   97  ,   7   ,   917 ,   1969    ,   3421    ,   2149    ,   11379   ,   1035    ,0 };
            const unsigned long dim2133Kuo2Init[]       =   {   1   ,   1   ,   5   ,   9   ,   5   ,   45  ,   51  ,   29  ,   289 ,   49  ,   1311    ,   4005    ,   3619    ,   12337   ,   7693    ,0 };
            const unsigned long dim2134Kuo2Init[]       =   {   1   ,   3   ,   1   ,   7   ,   25  ,   9   ,   11  ,   23  ,   281 ,   517 ,   1309    ,   2719    ,   3133    ,   529 ,   21795   ,0 };
            const unsigned long dim2135Kuo2Init[]       =   {   1   ,   3   ,   3   ,   1   ,   5   ,   43  ,   79  ,   87  ,   389 ,   189 ,   547 ,   3161    ,   4765    ,   10751   ,   15557   ,0 };
            const unsigned long dim2136Kuo2Init[]       =   {   1   ,   3   ,   1   ,   13  ,   29  ,   11  ,   5   ,   75  ,   453 ,   323 ,   999 ,   341 ,   6991    ,   14803   ,   31057   ,0 };
            const unsigned long dim2137Kuo2Init[]       =   {   1   ,   1   ,   5   ,   11  ,   7   ,   3   ,   81  ,   201 ,   195 ,   221 ,   1949    ,   2205    ,   2273    ,   12029   ,   21089   ,0 };
            const unsigned long dim2138Kuo2Init[]       =   {   1   ,   1   ,   7   ,   1   ,   25  ,   53  ,   7   ,   135 ,   7   ,   179 ,   797 ,   567 ,   1459    ,   9773    ,   2839    ,0 };
            const unsigned long dim2139Kuo2Init[]       =   {   1   ,   1   ,   3   ,   7   ,   31  ,   27  ,   41  ,   253 ,   273 ,   749 ,   1605    ,   2259    ,   7171    ,   11861   ,   11769   ,0 };
            const unsigned long dim2140Kuo2Init[]       =   {   1   ,   1   ,   1   ,   5   ,   13  ,   49  ,   25  ,   131 ,   397 ,   783 ,   1849    ,   3585    ,   2727    ,   8213    ,   14849   ,0 };
            const unsigned long dim2141Kuo2Init[]       =   {   1   ,   3   ,   5   ,   1   ,   25  ,   35  ,   121 ,   159 ,   27  ,   305 ,   1841    ,   1893    ,   829 ,   11659   ,   9859    ,0 };
            const unsigned long dim2142Kuo2Init[]       =   {   1   ,   1   ,   5   ,   5   ,   3   ,   11  ,   79  ,   49  ,   95  ,   755 ,   1011    ,   2247    ,   5731    ,   12975   ,   20241   ,0 };
            const unsigned long dim2143Kuo2Init[]       =   {   1   ,   1   ,   5   ,   3   ,   25  ,   5   ,   103 ,   35  ,   337 ,   619 ,   35  ,   3655    ,   5137    ,   4867    ,   6055    ,0 };
            const unsigned long dim2144Kuo2Init[]       =   {   1   ,   1   ,   7   ,   1   ,   15  ,   35  ,   67  ,   113 ,   381 ,   1001    ,   177 ,   3981    ,   1059    ,   4433    ,   26845   ,0 };
            const unsigned long dim2145Kuo2Init[]       =   {   1   ,   3   ,   7   ,   9   ,   19  ,   3   ,   107 ,   161 ,   109 ,   129 ,   1447    ,   4053    ,   7435    ,   3483    ,   7595    ,0 };
            const unsigned long dim2146Kuo2Init[]       =   {   1   ,   1   ,   7   ,   9   ,   13  ,   25  ,   19  ,   137 ,   149 ,   191 ,   719 ,   1279    ,   6129    ,   6367    ,   9335    ,0 };
            const unsigned long dim2147Kuo2Init[]       =   {   1   ,   3   ,   5   ,   11  ,   19  ,   55  ,   19  ,   191 ,   31  ,   341 ,   485 ,   1165    ,   501 ,   11771   ,   17423   ,0 };
            const unsigned long dim2148Kuo2Init[]       =   {   1   ,   3   ,   7   ,   1   ,   21  ,   35  ,   115 ,   163 ,   99  ,   581 ,   1769    ,   1887    ,   6349    ,   13467   ,   10503   ,0 };
            const unsigned long dim2149Kuo2Init[]       =   {   1   ,   1   ,   7   ,   3   ,   23  ,   21  ,   105 ,   255 ,   501 ,   171 ,   1689    ,   1747    ,   2603    ,   11783   ,   7633    ,0 };
            const unsigned long dim2150Kuo2Init[]       =   {   1   ,   1   ,   7   ,   7   ,   9   ,   45  ,   65  ,   95  ,   47  ,   163 ,   1597    ,   2693    ,   7407    ,   8155    ,   13645   ,0 };
            const unsigned long dim2151Kuo2Init[]       =   {   1   ,   1   ,   1   ,   1   ,   9   ,   5   ,   105 ,   55  ,   217 ,   773 ,   43  ,   741 ,   1093    ,   8087    ,   8003    ,0 };
            const unsigned long dim2152Kuo2Init[]       =   {   1   ,   3   ,   5   ,   15  ,   19  ,   29  ,   127 ,   29  ,   23  ,   27  ,   2011    ,   623 ,   2649    ,   2979    ,   269 ,0 };
            const unsigned long dim2153Kuo2Init[]       =   {   1   ,   3   ,   3   ,   15  ,   3   ,   17  ,   121 ,   85  ,   313 ,   333 ,   89  ,   1   ,   1147    ,   15909   ,   24017   ,0 };
            const unsigned long dim2154Kuo2Init[]       =   {   1   ,   3   ,   5   ,   13  ,   17  ,   53  ,   89  ,   27  ,   219 ,   451 ,   629 ,   2539    ,   4151    ,   4811    ,   17057   ,0 };
            const unsigned long dim2155Kuo2Init[]       =   {   1   ,   1   ,   1   ,   3   ,   31  ,   47  ,   127 ,   63  ,   85  ,   387 ,   61  ,   685 ,   1103    ,   1257    ,   19305   ,0 };
            const unsigned long dim2156Kuo2Init[]       =   {   1   ,   3   ,   7   ,   11  ,   25  ,   7   ,   39  ,   209 ,   449 ,   667 ,   227 ,   605 ,   987 ,   1699    ,   11001   ,0 };
            const unsigned long dim2157Kuo2Init[]       =   {   1   ,   3   ,   5   ,   15  ,   23  ,   35  ,   15  ,   193 ,   25  ,   1021    ,   1371    ,   3015    ,   7837    ,   14087   ,   21779   ,0 };
            const unsigned long dim2158Kuo2Init[]       =   {   1   ,   1   ,   5   ,   1   ,   1   ,   25  ,   25  ,   241 ,   493 ,   901 ,   1261    ,   261 ,   6959    ,   16013   ,   2849    ,0 };
            const unsigned long dim2159Kuo2Init[]       =   {   1   ,   3   ,   1   ,   5   ,   13  ,   27  ,   85  ,   41  ,   139 ,   567 ,   753 ,   3255    ,   5633    ,   15687   ,   3019    ,0 };
            const unsigned long dim2160Kuo2Init[]       =   {   1   ,   1   ,   3   ,   11  ,   19  ,   61  ,   43  ,   153 ,   299 ,   931 ,   1481    ,   203 ,   593 ,   5019    ,   28309   ,0 };
            const unsigned long dim2161Kuo2Init[]       =   {   1   ,   3   ,   7   ,   7   ,   9   ,   11  ,   107 ,   243 ,   227 ,   69  ,   605 ,   3761    ,   1091    ,   4503    ,   21203   ,0 };
            const unsigned long dim2162Kuo2Init[]       =   {   1   ,   3   ,   5   ,   11  ,   11  ,   55  ,   47  ,   167 ,   291 ,   883 ,   1411    ,   3099    ,   1099    ,   11619   ,   9781    ,0 };
            const unsigned long dim2163Kuo2Init[]       =   {   1   ,   3   ,   7   ,   15  ,   23  ,   53  ,   109 ,   247 ,   459 ,   793 ,   839 ,   3069    ,   1829    ,   5151    ,   26255   ,0 };
            const unsigned long dim2164Kuo2Init[]       =   {   1   ,   3   ,   1   ,   15  ,   11  ,   31  ,   73  ,   193 ,   341 ,   791 ,   187 ,   3787    ,   4481    ,   11435   ,   715 ,0 };
            const unsigned long dim2165Kuo2Init[]       =   {   1   ,   3   ,   1   ,   13  ,   13  ,   61  ,   79  ,   193 ,   455 ,   571 ,   1277    ,   1565    ,   7385    ,   4393    ,   27447   ,0 };
            const unsigned long dim2166Kuo2Init[]       =   {   1   ,   3   ,   7   ,   3   ,   13  ,   35  ,   39  ,   89  ,   335 ,   701 ,   1423    ,   2441    ,   2895    ,   2795    ,   28511   ,0 };
            const unsigned long dim2167Kuo2Init[]       =   {   1   ,   3   ,   1   ,   1   ,   9   ,   27  ,   39  ,   163 ,   127 ,   163 ,   1881    ,   1995    ,   2629    ,   4541    ,   1669    ,0 };
            const unsigned long dim2168Kuo2Init[]       =   {   1   ,   1   ,   5   ,   3   ,   5   ,   53  ,   29  ,   223 ,   433 ,   643 ,   1035    ,   2339    ,   5947    ,   3047    ,   27409   ,0 };
            const unsigned long dim2169Kuo2Init[]       =   {   1   ,   3   ,   1   ,   15  ,   5   ,   35  ,   47  ,   49  ,   347 ,   579 ,   551 ,   2285    ,   3171    ,   8041    ,   8223    ,0 };
            const unsigned long dim2170Kuo2Init[]       =   {   1   ,   1   ,   1   ,   1   ,   1   ,   57  ,   51  ,   219 ,   339 ,   95  ,   1585    ,   2811    ,   4771    ,   1475    ,   7919    ,0 };
            const unsigned long dim2171Kuo2Init[]       =   {   1   ,   3   ,   7   ,   3   ,   21  ,   11  ,   51  ,   143 ,   111 ,   475 ,   1669    ,   3717    ,   6579    ,   7161    ,   9491    ,0 };
            const unsigned long dim2172Kuo2Init[]       =   {   1   ,   3   ,   1   ,   5   ,   1   ,   17  ,   15  ,   179 ,   485 ,   587 ,   571 ,   3259    ,   3299    ,   207 ,   2121    ,0 };
            const unsigned long dim2173Kuo2Init[]       =   {   1   ,   3   ,   5   ,   13  ,   27  ,   19  ,   23  ,   47  ,   289 ,   997 ,   523 ,   3183    ,   8087    ,   2621    ,   31471   ,0 };
            const unsigned long dim2174Kuo2Init[]       =   {   1   ,   1   ,   3   ,   7   ,   23  ,   21  ,   125 ,   181 ,   393 ,   177 ,   419 ,   2809    ,   6727    ,   1063    ,   10889   ,0 };
            const unsigned long dim2175Kuo2Init[]       =   {   1   ,   3   ,   5   ,   1   ,   19  ,   29  ,   115 ,   149 ,   439 ,   125 ,   7   ,   2173    ,   6691    ,   4063    ,   8193    ,0 };
            const unsigned long dim2176Kuo2Init[]       =   {   1   ,   1   ,   3   ,   11  ,   3   ,   17  ,   9   ,   251 ,   493 ,   977 ,   1713    ,   2711    ,   4377    ,   3171    ,   22195   ,0 };
            const unsigned long dim2177Kuo2Init[]       =   {   1   ,   1   ,   3   ,   1   ,   17  ,   63  ,   73  ,   141 ,   323 ,   875 ,   713 ,   2133    ,   3283    ,   10039   ,   11431   ,0 };
            const unsigned long dim2178Kuo2Init[]       =   {   1   ,   1   ,   7   ,   5   ,   29  ,   33  ,   63  ,   19  ,   191 ,   915 ,   1523    ,   3505    ,   2431    ,   9531    ,   5253    ,0 };
            const unsigned long dim2179Kuo2Init[]       =   {   1   ,   3   ,   5   ,   5   ,   15  ,   25  ,   81  ,   97  ,   455 ,   141 ,   1947    ,   2001    ,   6015    ,   11439   ,   29725   ,0 };
            const unsigned long dim2180Kuo2Init[]       =   {   1   ,   1   ,   1   ,   1   ,   7   ,   35  ,   71  ,   205 ,   485 ,   495 ,   345 ,   1991    ,   6905    ,   12535   ,   2209    ,0 };
            const unsigned long dim2181Kuo2Init[]       =   {   1   ,   1   ,   7   ,   15  ,   27  ,   23  ,   19  ,   119 ,   347 ,   519 ,   423 ,   1567    ,   7783    ,   15095   ,   23609   ,0 };
            const unsigned long dim2182Kuo2Init[]       =   {   1   ,   3   ,   7   ,   1   ,   23  ,   13  ,   23  ,   239 ,   199 ,   575 ,   1149    ,   771 ,   3779    ,   12437   ,   32301   ,0 };
            const unsigned long dim2183Kuo2Init[]       =   {   1   ,   3   ,   5   ,   3   ,   31  ,   55  ,   87  ,   111 ,   233 ,   633 ,   1037    ,   4053    ,   2007    ,   15197   ,   31531   ,0 };
            const unsigned long dim2184Kuo2Init[]       =   {   1   ,   1   ,   1   ,   7   ,   3   ,   47  ,   11  ,   169 ,   293 ,   959 ,   903 ,   2129    ,   1049    ,   1311    ,   4027    ,0 };
            const unsigned long dim2185Kuo2Init[]       =   {   1   ,   3   ,   3   ,   7   ,   1   ,   15  ,   37  ,   79  ,   299 ,   989 ,   969 ,   2699    ,   4531    ,   8455    ,   30613   ,0 };
            const unsigned long dim2186Kuo2Init[]       =   {   1   ,   3   ,   1   ,   15  ,   13  ,   25  ,   57  ,   237 ,   117 ,   659 ,   517 ,   1821    ,   4755    ,   6745    ,   13269   ,0 };
            const unsigned long dim2187Kuo2Init[]       =   {   1   ,   3   ,   7   ,   11  ,   3   ,   27  ,   53  ,   197 ,   285 ,   897 ,   5   ,   675 ,   255 ,   5999    ,   7629    ,0 };
            const unsigned long dim2188Kuo2Init[]       =   {   1   ,   3   ,   7   ,   7   ,   21  ,   21  ,   71  ,   117 ,   347 ,   251 ,   1025    ,   2455    ,   7177    ,   9135    ,   15569   ,0 };
            const unsigned long dim2189Kuo2Init[]       =   {   1   ,   3   ,   1   ,   3   ,   25  ,   7   ,   113 ,   229 ,   243 ,   173 ,   1903    ,   427 ,   2743    ,   2245    ,   12783   ,0 };
            const unsigned long dim2190Kuo2Init[]       =   {   1   ,   1   ,   1   ,   15  ,   11  ,   53  ,   45  ,   209 ,   181 ,   147 ,   183 ,   3133    ,   4793    ,   14731   ,   16491   ,0 };
            const unsigned long dim2191Kuo2Init[]       =   {   1   ,   3   ,   7   ,   5   ,   27  ,   45  ,   77  ,   211 ,   89  ,   555 ,   1829    ,   4073    ,   839 ,   12079   ,   27745   ,0 };
            const unsigned long dim2192Kuo2Init[]       =   {   1   ,   1   ,   3   ,   7   ,   1   ,   29  ,   17  ,   65  ,   47  ,   995 ,   85  ,   69  ,   641 ,   4139    ,   1221    ,0 };
            const unsigned long dim2193Kuo2Init[]       =   {   1   ,   3   ,   3   ,   3   ,   11  ,   57  ,   11  ,   69  ,   285 ,   837 ,   1129    ,   3465    ,   323 ,   4599    ,   6593    ,0 };
            const unsigned long dim2194Kuo2Init[]       =   {   1   ,   3   ,   5   ,   7   ,   19  ,   27  ,   1   ,   197 ,   259 ,   315 ,   567 ,   613 ,   5385    ,   4495    ,   13507   ,0 };
            const unsigned long dim2195Kuo2Init[]       =   {   1   ,   3   ,   3   ,   7   ,   13  ,   63  ,   3   ,   81  ,   363 ,   67  ,   1025    ,   1403    ,   4469    ,   9881    ,   3563    ,0 };
            const unsigned long dim2196Kuo2Init[]       =   {   1   ,   1   ,   1   ,   5   ,   23  ,   1   ,   103 ,   51  ,   445 ,   1017    ,   361 ,   811 ,   4555    ,   6013    ,   25117   ,0 };
            const unsigned long dim2197Kuo2Init[]       =   {   1   ,   1   ,   1   ,   11  ,   21  ,   33  ,   73  ,   13  ,   155 ,   419 ,   371 ,   499 ,   3003    ,   7577    ,   19191   ,0 };
            const unsigned long dim2198Kuo2Init[]       =   {   1   ,   1   ,   1   ,   13  ,   19  ,   57  ,   91  ,   251 ,   157 ,   551 ,   1529    ,   2649    ,   7785    ,   4357    ,   18533   ,0 };
            const unsigned long dim2199Kuo2Init[]       =   {   1   ,   1   ,   1   ,   15  ,   29  ,   23  ,   33  ,   139 ,   495 ,   87  ,   1291    ,   1441    ,   819 ,   321 ,   24597   ,0 };
            const unsigned long dim2200Kuo2Init[]       =   {   1   ,   1   ,   5   ,   7   ,   11  ,   21  ,   127 ,   135 ,   395 ,   511 ,   213 ,   4043    ,   5869    ,   5733    ,   23157   ,0 };
            const unsigned long dim2201Kuo2Init[]       =   {   1   ,   3   ,   5   ,   5   ,   1   ,   61  ,   99  ,   155 ,   415 ,   111 ,   1335    ,   241 ,   1935    ,   7457    ,   7127    ,0 };
            const unsigned long dim2202Kuo2Init[]       =   {   1   ,   1   ,   3   ,   7   ,   25  ,   43  ,   69  ,   97  ,   423 ,   989 ,   1433    ,   107 ,   7253    ,   6543    ,   1685    ,0 };
            const unsigned long dim2203Kuo2Init[]       =   {   1   ,   3   ,   1   ,   3   ,   9   ,   51  ,   77  ,   69  ,   1   ,   75  ,   1859    ,   3735    ,   1199    ,   13725   ,   6689    ,0 };
            const unsigned long dim2204Kuo2Init[]       =   {   1   ,   3   ,   5   ,   7   ,   13  ,   7   ,   103 ,   91  ,   465 ,   675 ,   1221    ,   2853    ,   6785    ,   7113    ,   12671   ,0 };
            const unsigned long dim2205Kuo2Init[]       =   {   1   ,   3   ,   1   ,   15  ,   9   ,   9   ,   75  ,   17  ,   357 ,   559 ,   1021    ,   1023    ,   779 ,   10805   ,   12707   ,0 };
            const unsigned long dim2206Kuo2Init[]       =   {   1   ,   3   ,   5   ,   1   ,   19  ,   41  ,   25  ,   209 ,   61  ,   225 ,   533 ,   2661    ,   6735    ,   4877    ,   20085   ,0 };
            const unsigned long dim2207Kuo2Init[]       =   {   1   ,   1   ,   5   ,   7   ,   29  ,   35  ,   61  ,   21  ,   203 ,   689 ,   1431    ,   2009    ,   3475    ,   2103    ,   19457   ,0 };
            const unsigned long dim2208Kuo2Init[]       =   {   1   ,   1   ,   3   ,   9   ,   7   ,   51  ,   119 ,   115 ,   357 ,   969 ,   779 ,   2407    ,   1669    ,   4671    ,   21945   ,0 };
            const unsigned long dim2209Kuo2Init[]       =   {   1   ,   1   ,   5   ,   13  ,   5   ,   33  ,   125 ,   61  ,   123 ,   569 ,   7   ,   2389    ,   3325    ,   15701   ,   16931   ,0 };
            const unsigned long dim2210Kuo2Init[]       =   {   1   ,   1   ,   3   ,   7   ,   13  ,   29  ,   7   ,   205 ,   397 ,   861 ,   1035    ,   675 ,   5963    ,   9229    ,   14537   ,0 };
            const unsigned long dim2211Kuo2Init[]       =   {   1   ,   3   ,   7   ,   5   ,   13  ,   41  ,   69  ,   25  ,   461 ,   745 ,   189 ,   3239    ,   4515    ,   2717    ,   22815   ,0 };
            const unsigned long dim2212Kuo2Init[]       =   {   1   ,   1   ,   1   ,   7   ,   29  ,   47  ,   49  ,   91  ,   403 ,   325 ,   1241    ,   2295    ,   6371    ,   6791    ,   911 ,0 };
            const unsigned long dim2213Kuo2Init[]       =   {   1   ,   1   ,   1   ,   5   ,   13  ,   47  ,   83  ,   161 ,   371 ,   779 ,   1935    ,   3445    ,   735 ,   1963    ,   17267   ,0 };
            const unsigned long dim2214Kuo2Init[]       =   {   1   ,   3   ,   7   ,   7   ,   19  ,   59  ,   41  ,   217 ,   71  ,   407 ,   1639    ,   4081    ,   8011    ,   4805    ,   1847    ,0 };
            const unsigned long dim2215Kuo2Init[]       =   {   1   ,   1   ,   7   ,   9   ,   21  ,   11  ,   113 ,   151 ,   269 ,   119 ,   615 ,   2657    ,   1835    ,   16071   ,   7475    ,0 };
            const unsigned long dim2216Kuo2Init[]       =   {   1   ,   1   ,   1   ,   13  ,   13  ,   53  ,   121 ,   231 ,   419 ,   439 ,   1287    ,   3043    ,   8187    ,   4355    ,   16377   ,0 };
            const unsigned long dim2217Kuo2Init[]       =   {   1   ,   1   ,   1   ,   3   ,   1   ,   51  ,   69  ,   49  ,   403 ,   549 ,   659 ,   501 ,   7497    ,   9461    ,   8813    ,0 };
            const unsigned long dim2218Kuo2Init[]       =   {   1   ,   3   ,   3   ,   9   ,   15  ,   13  ,   11  ,   65  ,   241 ,   77  ,   675 ,   2077    ,   927 ,   11841   ,   2527    ,0 };
            const unsigned long dim2219Kuo2Init[]       =   {   1   ,   1   ,   5   ,   9   ,   23  ,   35  ,   113 ,   33  ,   497 ,   1001    ,   1105    ,   2157    ,   1209    ,   7821    ,   20345   ,0 };
            const unsigned long dim2220Kuo2Init[]       =   {   1   ,   1   ,   1   ,   5   ,   23  ,   39  ,   81  ,   151 ,   259 ,   491 ,   1129    ,   1115    ,   4501    ,   5513    ,   20347   ,0 };
            const unsigned long dim2221Kuo2Init[]       =   {   1   ,   1   ,   5   ,   1   ,   11  ,   59  ,   55  ,   195 ,   157 ,   915 ,   35  ,   489 ,   2805    ,   9059    ,   14757   ,0 };
            const unsigned long dim2222Kuo2Init[]       =   {   1   ,   3   ,   5   ,   15  ,   9   ,   21  ,   91  ,   59  ,   425 ,   127 ,   1939    ,   2763    ,   1243    ,   4303    ,   4549    ,0 };
            const unsigned long dim2223Kuo2Init[]       =   {   1   ,   1   ,   5   ,   13  ,   19  ,   29  ,   127 ,   251 ,   91  ,   419 ,   333 ,   2943    ,   1649    ,   8653    ,   32401   ,0 };
            const unsigned long dim2224Kuo2Init[]       =   {   1   ,   3   ,   1   ,   11  ,   13  ,   13  ,   73  ,   153 ,   131 ,   675 ,   1235    ,   1565    ,   2153    ,   3999    ,   3075    ,0 };
            const unsigned long dim2225Kuo2Init[]       =   {   1   ,   1   ,   5   ,   13  ,   1   ,   29  ,   111 ,   229 ,   23  ,   649 ,   1017    ,   2257    ,   1967    ,   9839    ,   21179   ,0 };
            const unsigned long dim2226Kuo2Init[]       =   {   1   ,   1   ,   3   ,   9   ,   13  ,   59  ,   39  ,   243 ,   453 ,   51  ,   801 ,   727 ,   5391    ,   11987   ,   12431   ,0 };
            const unsigned long dim2227Kuo2Init[]       =   {   1   ,   1   ,   7   ,   5   ,   9   ,   7   ,   59  ,   203 ,   177 ,   255 ,   1525    ,   1869    ,   5187    ,   135 ,   11205   ,0 };
            const unsigned long dim2228Kuo2Init[]       =   {   1   ,   1   ,   5   ,   5   ,   17  ,   49  ,   115 ,   101 ,   45  ,   385 ,   1751    ,   883 ,   7437    ,   3839    ,   7997    ,0 };
            const unsigned long dim2229Kuo2Init[]       =   {   1   ,   1   ,   1   ,   13  ,   11  ,   17  ,   99  ,   127 ,   457 ,   333 ,   395 ,   1387    ,   4533    ,   2247    ,   5999    ,0 };
            const unsigned long dim2230Kuo2Init[]       =   {   1   ,   3   ,   7   ,   7   ,   15  ,   61  ,   103 ,   217 ,   175 ,   133 ,   1721    ,   1735    ,   8083    ,   14973   ,   20891   ,0 };
            const unsigned long dim2231Kuo2Init[]       =   {   1   ,   1   ,   1   ,   13  ,   31  ,   35  ,   69  ,   53  ,   191 ,   671 ,   1339    ,   1677    ,   4463    ,   1951    ,   28129   ,0 };
            const unsigned long dim2232Kuo2Init[]       =   {   1   ,   3   ,   3   ,   15  ,   19  ,   13  ,   73  ,   171 ,   481 ,   981 ,   343 ,   1831    ,   2967    ,   1801    ,   9923    ,0 };
            const unsigned long dim2233Kuo2Init[]       =   {   1   ,   1   ,   7   ,   7   ,   9   ,   29  ,   57  ,   113 ,   425 ,   705 ,   673 ,   3749    ,   1883    ,   12295   ,   9669    ,0 };
            const unsigned long dim2234Kuo2Init[]       =   {   1   ,   3   ,   1   ,   3   ,   13  ,   35  ,   23  ,   35  ,   303 ,   281 ,   773 ,   3053    ,   3841    ,   7943    ,   5023    ,0 };
            const unsigned long dim2235Kuo2Init[]       =   {   1   ,   1   ,   7   ,   15  ,   11  ,   27  ,   85  ,   163 ,   349 ,   289 ,   297 ,   3043    ,   6469    ,   12029   ,   8087    ,0 };
            const unsigned long dim2236Kuo2Init[]       =   {   1   ,   1   ,   5   ,   15  ,   27  ,   47  ,   3   ,   51  ,   87  ,   583 ,   893 ,   971 ,   6859    ,   7459    ,   6931    ,0 };
            const unsigned long dim2237Kuo2Init[]       =   {   1   ,   3   ,   3   ,   15  ,   21  ,   57  ,   65  ,   141 ,   129 ,   345 ,   745 ,   2281    ,   4753    ,   2899    ,   1891    ,0 };
            const unsigned long dim2238Kuo2Init[]       =   {   1   ,   1   ,   1   ,   11  ,   7   ,   3   ,   87  ,   127 ,   133 ,   243 ,   241 ,   557 ,   6537    ,   1645    ,   24843   ,0 };
            const unsigned long dim2239Kuo2Init[]       =   {   1   ,   1   ,   7   ,   1   ,   7   ,   37  ,   91  ,   9   ,   337 ,   787 ,   1833    ,   3259    ,   1335    ,   6365    ,   30259   ,0 };
            const unsigned long dim2240Kuo2Init[]       =   {   1   ,   1   ,   1   ,   9   ,   7   ,   3   ,   127 ,   41  ,   41  ,   377 ,   1633    ,   3031    ,   4483    ,   12555   ,   23921   ,0 };
            const unsigned long dim2241Kuo2Init[]       =   {   1   ,   1   ,   7   ,   1   ,   21  ,   9   ,   77  ,   243 ,   373 ,   289 ,   715 ,   2323    ,   2549    ,   8891    ,   22603   ,0 };
            const unsigned long dim2242Kuo2Init[]       =   {   1   ,   1   ,   1   ,   5   ,   19  ,   53  ,   81  ,   83  ,   111 ,   593 ,   1041    ,   499 ,   5849    ,   1261    ,   9561    ,0 };
            const unsigned long dim2243Kuo2Init[]       =   {   1   ,   3   ,   5   ,   15  ,   11  ,   45  ,   13  ,   217 ,   309 ,   249 ,   1597    ,   2829    ,   2185    ,   11919   ,   22975   ,0 };
            const unsigned long dim2244Kuo2Init[]       =   {   1   ,   1   ,   1   ,   3   ,   31  ,   3   ,   69  ,   105 ,   449 ,   339 ,   1705    ,   551 ,   1701    ,   1163    ,   22551   ,0 };
            const unsigned long dim2245Kuo2Init[]       =   {   1   ,   1   ,   3   ,   7   ,   29  ,   23  ,   115 ,   119 ,   25  ,   425 ,   225 ,   1379    ,   1443    ,   9549    ,   31083   ,0 };
            const unsigned long dim2246Kuo2Init[]       =   {   1   ,   1   ,   7   ,   11  ,   31  ,   37  ,   121 ,   29  ,   367 ,   171 ,   1611    ,   3147    ,   4887    ,   15191   ,   27943   ,0 };
            const unsigned long dim2247Kuo2Init[]       =   {   1   ,   1   ,   5   ,   1   ,   25  ,   11  ,   39  ,   137 ,   251 ,   557 ,   1833    ,   2803    ,   2081    ,   819 ,   12493   ,0 };
            const unsigned long dim2248Kuo2Init[]       =   {   1   ,   1   ,   1   ,   3   ,   17  ,   19  ,   87  ,   123 ,   333 ,   215 ,   685 ,   1235    ,   2327    ,   5927    ,   16909   ,0 };
            const unsigned long dim2249Kuo2Init[]       =   {   1   ,   1   ,   7   ,   7   ,   5   ,   31  ,   127 ,   205 ,   197 ,   671 ,   1101    ,   1767    ,   7871    ,   4157    ,   31301   ,0 };
            const unsigned long dim2250Kuo2Init[]       =   {   1   ,   1   ,   5   ,   5   ,   17  ,   5   ,   23  ,   175 ,   363 ,   831 ,   563 ,   2599    ,   7081    ,   11695   ,   10301   ,0 };
            const unsigned long dim2251Kuo2Init[]       =   {   1   ,   3   ,   3   ,   1   ,   21  ,   55  ,   97  ,   163 ,   443 ,   533 ,   1889    ,   1431    ,   6671    ,   14699   ,   26863   ,0 };
            const unsigned long dim2252Kuo2Init[]       =   {   1   ,   3   ,   5   ,   15  ,   27  ,   15  ,   7   ,   215 ,   7   ,   557 ,   819 ,   3855    ,   619 ,   10185   ,   23715   ,0 };
            const unsigned long dim2253Kuo2Init[]       =   {   1   ,   3   ,   7   ,   1   ,   17  ,   23  ,   25  ,   59  ,   369 ,   675 ,   463 ,   809 ,   8097    ,   9263    ,   8907    ,0 };
            const unsigned long dim2254Kuo2Init[]       =   {   1   ,   1   ,   1   ,   3   ,   1   ,   29  ,   63  ,   65  ,   103 ,   659 ,   1883    ,   1141    ,   641 ,   6083    ,   16237   ,0 };
            const unsigned long dim2255Kuo2Init[]       =   {   1   ,   1   ,   1   ,   3   ,   5   ,   41  ,   13  ,   143 ,   349 ,   811 ,   1645    ,   1871    ,   3385    ,   12795   ,   25105   ,0 };
            const unsigned long dim2256Kuo2Init[]       =   {   1   ,   1   ,   1   ,   15  ,   17  ,   7   ,   73  ,   195 ,   439 ,   85  ,   1701    ,   3989    ,   1619    ,   5319    ,   20593   ,0 };
            const unsigned long dim2257Kuo2Init[]       =   {   1   ,   1   ,   3   ,   15  ,   1   ,   9   ,   61  ,   223 ,   441 ,   543 ,   765 ,   3477    ,   2569    ,   6061    ,   25353   ,0 };
            const unsigned long dim2258Kuo2Init[]       =   {   1   ,   1   ,   3   ,   11  ,   29  ,   63  ,   9   ,   149 ,   307 ,   705 ,   871 ,   2975    ,   4145    ,   431 ,   14483   ,0 };
            const unsigned long dim2259Kuo2Init[]       =   {   1   ,   1   ,   1   ,   3   ,   11  ,   7   ,   103 ,   85  ,   389 ,   719 ,   1201    ,   3139    ,   2941    ,   13867   ,   1915    ,0 };
            const unsigned long dim2260Kuo2Init[]       =   {   1   ,   1   ,   3   ,   3   ,   21  ,   27  ,   65  ,   59  ,   223 ,   335 ,   1845    ,   1315    ,   3619    ,   1851    ,   4313    ,0 };
            const unsigned long dim2261Kuo2Init[]       =   {   1   ,   3   ,   7   ,   13  ,   21  ,   21  ,   19  ,   181 ,   39  ,   89  ,   1703    ,   791 ,   597 ,   3349    ,   14387   ,0 };
            const unsigned long dim2262Kuo2Init[]       =   {   1   ,   3   ,   5   ,   3   ,   9   ,   19  ,   49  ,   253 ,   61  ,   313 ,   1763    ,   629 ,   1951    ,   16157   ,   14435   ,0 };
            const unsigned long dim2263Kuo2Init[]       =   {   1   ,   1   ,   5   ,   13  ,   23  ,   41  ,   45  ,   107 ,   111 ,   287 ,   1829    ,   33  ,   5433    ,   7283    ,   20573   ,0 };
            const unsigned long dim2264Kuo2Init[]       =   {   1   ,   3   ,   3   ,   5   ,   17  ,   43  ,   117 ,   55  ,   275 ,   763 ,   1581    ,   2349    ,   1219    ,   16145   ,   545 ,0 };
            const unsigned long dim2265Kuo2Init[]       =   {   1   ,   1   ,   5   ,   3   ,   13  ,   23  ,   65  ,   97  ,   347 ,   169 ,   1371    ,   2373    ,   4577    ,   14667   ,   5713    ,0 };
            const unsigned long dim2266Kuo2Init[]       =   {   1   ,   1   ,   5   ,   11  ,   1   ,   43  ,   17  ,   87  ,   287 ,   943 ,   1257    ,   3383    ,   2389    ,   3349    ,   11641   ,0 };
            const unsigned long dim2267Kuo2Init[]       =   {   1   ,   3   ,   1   ,   7   ,   27  ,   37  ,   127 ,   27  ,   239 ,   5   ,   575 ,   1811    ,   7185    ,   10233   ,   20519   ,0 };
            const unsigned long dim2268Kuo2Init[]       =   {   1   ,   3   ,   1   ,   13  ,   21  ,   33  ,   91  ,   147 ,   351 ,   283 ,   1765    ,   3497    ,   3113    ,   7635    ,   21229   ,0 };
            const unsigned long dim2269Kuo2Init[]       =   {   1   ,   3   ,   7   ,   3   ,   19  ,   39  ,   47  ,   135 ,   495 ,   401 ,   843 ,   551 ,   5239    ,   6525    ,   32339   ,0 };
            const unsigned long dim2270Kuo2Init[]       =   {   1   ,   3   ,   3   ,   3   ,   1   ,   37  ,   61  ,   243 ,   491 ,   139 ,   613 ,   685 ,   3027    ,   8053    ,   4935    ,0 };
            const unsigned long dim2271Kuo2Init[]       =   {   1   ,   1   ,   7   ,   1   ,   13  ,   43  ,   25  ,   117 ,   43  ,   161 ,   747 ,   3005    ,   5445    ,   7559    ,   31029   ,0 };
            const unsigned long dim2272Kuo2Init[]       =   {   1   ,   1   ,   1   ,   1   ,   9   ,   31  ,   107 ,   57  ,   121 ,   733 ,   1813    ,   2057    ,   2473    ,   1769    ,   14649   ,0 };
            const unsigned long dim2273Kuo2Init[]       =   {   1   ,   1   ,   7   ,   7   ,   7   ,   53  ,   105 ,   163 ,   401 ,   391 ,   849 ,   2359    ,   6819    ,   13905   ,   17267   ,0 };
            const unsigned long dim2274Kuo2Init[]       =   {   1   ,   1   ,   5   ,   7   ,   23  ,   9   ,   31  ,   39  ,   7   ,   767 ,   153 ,   2969    ,   5539    ,   6581    ,   30515   ,0 };
            const unsigned long dim2275Kuo2Init[]       =   {   1   ,   3   ,   7   ,   5   ,   3   ,   63  ,   73  ,   167 ,   485 ,   837 ,   403 ,   193 ,   2707    ,   2017    ,   743 ,0 };
            const unsigned long dim2276Kuo2Init[]       =   {   1   ,   1   ,   5   ,   15  ,   17  ,   59  ,   115 ,   115 ,   463 ,   203 ,   607 ,   343 ,   577 ,   6985    ,   8797    ,0 };
            const unsigned long dim2277Kuo2Init[]       =   {   1   ,   1   ,   7   ,   3   ,   19  ,   61  ,   33  ,   245 ,   337 ,   105 ,   597 ,   807 ,   1669    ,   5447    ,   15173   ,0 };
            const unsigned long dim2278Kuo2Init[]       =   {   1   ,   3   ,   3   ,   9   ,   27  ,   51  ,   67  ,   135 ,   151 ,   215 ,   217 ,   221 ,   6467    ,   3693    ,   4563    ,0 };
            const unsigned long dim2279Kuo2Init[]       =   {   1   ,   1   ,   1   ,   3   ,   1   ,   5   ,   91  ,   223 ,   491 ,   783 ,   287 ,   4067    ,   6747    ,   2489    ,   11513   ,0 };
            const unsigned long dim2280Kuo2Init[]       =   {   1   ,   3   ,   1   ,   3   ,   25  ,   15  ,   9   ,   3   ,   445 ,   893 ,   399 ,   2073    ,   3715    ,   2493    ,   7761    ,0 };
            const unsigned long dim2281Kuo2Init[]       =   {   1   ,   3   ,   1   ,   11  ,   5   ,   41  ,   121 ,   107 ,   15  ,   547 ,   1417    ,   1275    ,   7255    ,   14979   ,   26993   ,0 };
            const unsigned long dim2282Kuo2Init[]       =   {   1   ,   1   ,   5   ,   5   ,   29  ,   41  ,   115 ,   195 ,   305 ,   887 ,   335 ,   3919    ,   6607    ,   15783   ,   21457   ,0 };
            const unsigned long dim2283Kuo2Init[]       =   {   1   ,   1   ,   7   ,   15  ,   11  ,   43  ,   53  ,   19  ,   37  ,   219 ,   1905    ,   1153    ,   4783    ,   11817   ,   11119   ,0 };
            const unsigned long dim2284Kuo2Init[]       =   {   1   ,   1   ,   5   ,   13  ,   31  ,   57  ,   73  ,   237 ,   443 ,   755 ,   1141    ,   793 ,   8177    ,   9053    ,   25949   ,0 };
            const unsigned long dim2285Kuo2Init[]       =   {   1   ,   1   ,   3   ,   9   ,   9   ,   11  ,   1   ,   151 ,   249 ,   701 ,   523 ,   2383    ,   647 ,   7979    ,   28965   ,0 };
            const unsigned long dim2286Kuo2Init[]       =   {   1   ,   3   ,   1   ,   9   ,   7   ,   47  ,   47  ,   117 ,   239 ,   213 ,   1747    ,   3093    ,   3111    ,   6689    ,   22451   ,0 };
            const unsigned long dim2287Kuo2Init[]       =   {   1   ,   1   ,   3   ,   1   ,   21  ,   43  ,   111 ,   97  ,   323 ,   733 ,   1051    ,   3567    ,   1217    ,   3731    ,   15475   ,0 };
            const unsigned long dim2288Kuo2Init[]       =   {   1   ,   3   ,   1   ,   9   ,   29  ,   13  ,   63  ,   151 ,   469 ,   211 ,   927 ,   1893    ,   7741    ,   9733    ,   5777    ,0 };
            const unsigned long dim2289Kuo2Init[]       =   {   1   ,   1   ,   1   ,   15  ,   17  ,   49  ,   65  ,   81  ,   325 ,   847 ,   193 ,   3777    ,   4979    ,   1621    ,   17401   ,0 };
            const unsigned long dim2290Kuo2Init[]       =   {   1   ,   3   ,   3   ,   15  ,   27  ,   31  ,   37  ,   185 ,   339 ,   731 ,   1317    ,   527 ,   1889    ,   10609   ,   27419   ,0 };
            const unsigned long dim2291Kuo2Init[]       =   {   1   ,   1   ,   3   ,   13  ,   27  ,   45  ,   85  ,   193 ,   293 ,   909 ,   205 ,   2245    ,   2845    ,   11575   ,   21997   ,0 };
            const unsigned long dim2292Kuo2Init[]       =   {   1   ,   3   ,   1   ,   3   ,   13  ,   63  ,   49  ,   97  ,   343 ,   331 ,   2015    ,   2117    ,   6367    ,   13501   ,   12029   ,0 };
            const unsigned long dim2293Kuo2Init[]       =   {   1   ,   3   ,   3   ,   1   ,   15  ,   9   ,   61  ,   41  ,   355 ,   327 ,   1285    ,   571 ,   3437    ,   10489   ,   24383   ,0 };
            const unsigned long dim2294Kuo2Init[]       =   {   1   ,   1   ,   7   ,   13  ,   23  ,   13  ,   7   ,   127 ,   329 ,   323 ,   2037    ,   31  ,   407 ,   11301   ,   28027   ,0 };
            const unsigned long dim2295Kuo2Init[]       =   {   1   ,   1   ,   3   ,   1   ,   23  ,   61  ,   99  ,   237 ,   15  ,   257 ,   1575    ,   2077    ,   6507    ,   3365    ,   21885   ,0 };
            const unsigned long dim2296Kuo2Init[]       =   {   1   ,   3   ,   5   ,   11  ,   15  ,   61  ,   111 ,   25  ,   309 ,   853 ,   1203    ,   2943    ,   4875    ,   9233    ,   197 ,0 };
            const unsigned long dim2297Kuo2Init[]       =   {   1   ,   3   ,   3   ,   7   ,   5   ,   15  ,   17  ,   97  ,   213 ,   991 ,   867 ,   2291    ,   8129    ,   9381    ,   787 ,0 };
            const unsigned long dim2298Kuo2Init[]       =   {   1   ,   3   ,   5   ,   1   ,   31  ,   45  ,   105 ,   35  ,   403 ,   595 ,   1007    ,   3633    ,   5167    ,   11245   ,   6643    ,0 };
            const unsigned long dim2299Kuo2Init[]       =   {   1   ,   3   ,   5   ,   5   ,   1   ,   51  ,   121 ,   161 ,   109 ,   369 ,   447 ,   213 ,   4879    ,   5639    ,   7233    ,0 };
            const unsigned long dim2300Kuo2Init[]       =   {   1   ,   1   ,   7   ,   5   ,   3   ,   27  ,   35  ,   143 ,   489 ,   435 ,   1483    ,   3261    ,   7427    ,   3105    ,   5297    ,0 };
            const unsigned long dim2301Kuo2Init[]       =   {   1   ,   3   ,   1   ,   5   ,   11  ,   1   ,   75  ,   63  ,   367 ,   523 ,   291 ,   1429    ,   5509    ,   8641    ,   6847    ,0 };
            const unsigned long dim2302Kuo2Init[]       =   {   1   ,   1   ,   7   ,   13  ,   5   ,   59  ,   69  ,   81  ,   359 ,   905 ,   723 ,   3449    ,   3961    ,   16231   ,   19789   ,0 };
            const unsigned long dim2303Kuo2Init[]       =   {   1   ,   1   ,   7   ,   7   ,   13  ,   53  ,   119 ,   129 ,   87  ,   325 ,   415 ,   3313    ,   6045    ,   5293    ,   21325   ,0 };
            const unsigned long dim2304Kuo2Init[]       =   {   1   ,   1   ,   3   ,   3   ,   3   ,   45  ,   73  ,   109 ,   337 ,   185 ,   781 ,   3277    ,   2575    ,   15999   ,   8309    ,0 };
            const unsigned long dim2305Kuo2Init[]       =   {   1   ,   1   ,   3   ,   15  ,   15  ,   59  ,   35  ,   77  ,   441 ,   635 ,   627 ,   183 ,   4599    ,   8749    ,   24149   ,0 };
            const unsigned long dim2306Kuo2Init[]       =   {   1   ,   3   ,   3   ,   7   ,   23  ,   51  ,   31  ,   23  ,   367 ,   473 ,   27  ,   1021    ,   2587    ,   3063    ,   32563   ,0 };
            const unsigned long dim2307Kuo2Init[]       =   {   1   ,   1   ,   7   ,   5   ,   9   ,   57  ,   45  ,   147 ,   95  ,   233 ,   1349    ,   1497    ,   7103    ,   4049    ,   905 ,0 };
            const unsigned long dim2308Kuo2Init[]       =   {   1   ,   1   ,   7   ,   9   ,   9   ,   21  ,   89  ,   59  ,   379 ,   467 ,   1619    ,   809 ,   3947    ,   1715    ,   9841    ,0 };
            const unsigned long dim2309Kuo2Init[]       =   {   1   ,   3   ,   5   ,   9   ,   23  ,   41  ,   43  ,   213 ,   429 ,   825 ,   279 ,   1117    ,   659 ,   13157   ,   3977    ,0 };
            const unsigned long dim2310Kuo2Init[]       =   {   1   ,   1   ,   3   ,   15  ,   1   ,   57  ,   79  ,   87  ,   285 ,   693 ,   1947    ,   3417    ,   4241    ,   8595    ,   18075   ,0 };
            const unsigned long dim2311Kuo2Init[]       =   {   1   ,   1   ,   1   ,   1   ,   11  ,   19  ,   9   ,   79  ,   407 ,   475 ,   655 ,   2925    ,   2919    ,   6343    ,   19637   ,0 };
            const unsigned long dim2312Kuo2Init[]       =   {   1   ,   3   ,   3   ,   13  ,   1   ,   41  ,   107 ,   185 ,   29  ,   169 ,   915 ,   3535    ,   3543    ,   8015    ,   7097    ,0 };
            const unsigned long dim2313Kuo2Init[]       =   {   1   ,   1   ,   3   ,   11  ,   29  ,   1   ,   25  ,   199 ,   129 ,   471 ,   1335    ,   3413    ,   7373    ,   9881    ,   32595   ,0 };
            const unsigned long dim2314Kuo2Init[]       =   {   1   ,   3   ,   3   ,   1   ,   25  ,   13  ,   95  ,   23  ,   145 ,   775 ,   1873    ,   2347    ,   323 ,   10021   ,   20551   ,0 };
            const unsigned long dim2315Kuo2Init[]       =   {   1   ,   3   ,   3   ,   15  ,   3   ,   35  ,   25  ,   79  ,   279 ,   251 ,   245 ,   2917    ,   1147    ,   9525    ,   9473    ,0 };
            const unsigned long dim2316Kuo2Init[]       =   {   1   ,   3   ,   3   ,   11  ,   25  ,   35  ,   115 ,   13  ,   191 ,   447 ,   1041    ,   3437    ,   5941    ,   903 ,   26727   ,0 };
            const unsigned long dim2317Kuo2Init[]       =   {   1   ,   3   ,   5   ,   15  ,   27  ,   49  ,   29  ,   169 ,   167 ,   685 ,   931 ,   1705    ,   4513    ,   16217   ,   8545    ,0 };
            const unsigned long dim2318Kuo2Init[]       =   {   1   ,   3   ,   5   ,   3   ,   5   ,   55  ,   55  ,   113 ,   129 ,   491 ,   1583    ,   2103    ,   2233    ,   12669   ,   2809    ,0 };
            const unsigned long dim2319Kuo2Init[]       =   {   1   ,   1   ,   1   ,   7   ,   7   ,   57  ,   47  ,   43  ,   91  ,   415 ,   1903    ,   1067    ,   4281    ,   16095   ,   19919   ,0 };
            const unsigned long dim2320Kuo2Init[]       =   {   1   ,   3   ,   7   ,   7   ,   29  ,   7   ,   65  ,   19  ,   119 ,   473 ,   493 ,   2861    ,   7765    ,   15057   ,   13809   ,0 };
            const unsigned long dim2321Kuo2Init[]       =   {   1   ,   1   ,   5   ,   9   ,   3   ,   49  ,   83  ,   73  ,   135 ,   971 ,   1233    ,   3623    ,   7351    ,   15255   ,   29921   ,0 };
            const unsigned long dim2322Kuo2Init[]       =   {   1   ,   3   ,   3   ,   11  ,   29  ,   53  ,   5   ,   57  ,   41  ,   215 ,   477 ,   1851    ,   2219    ,   16289   ,   23801   ,0 };
            const unsigned long dim2323Kuo2Init[]       =   {   1   ,   1   ,   3   ,   7   ,   9   ,   29  ,   55  ,   37  ,   415 ,   537 ,   1459    ,   431 ,   4559    ,   5569    ,   2031    ,0 };
            const unsigned long dim2324Kuo2Init[]       =   {   1   ,   1   ,   5   ,   1   ,   5   ,   15  ,   47  ,   159 ,   369 ,   677 ,   691 ,   2647    ,   6243    ,   15909   ,   17725   ,0 };
            const unsigned long dim2325Kuo2Init[]       =   {   1   ,   1   ,   5   ,   9   ,   13  ,   43  ,   45  ,   167 ,   503 ,   939 ,   1609    ,   1123    ,   2799    ,   3903    ,   27007   ,0 };
            const unsigned long dim2326Kuo2Init[]       =   {   1   ,   3   ,   1   ,   13  ,   19  ,   43  ,   95  ,   247 ,   475 ,   193 ,   1513    ,   885 ,   5391    ,   7041    ,   13473   ,0 };
            const unsigned long dim2327Kuo2Init[]       =   {   1   ,   1   ,   7   ,   13  ,   25  ,   61  ,   3   ,   237 ,   473 ,   671 ,   101 ,   1129    ,   787 ,   14697   ,   909 ,0 };
            const unsigned long dim2328Kuo2Init[]       =   {   1   ,   1   ,   7   ,   15  ,   25  ,   3   ,   53  ,   81  ,   147 ,   27  ,   1691    ,   2093    ,   2627    ,   9523    ,   16021   ,0 };
            const unsigned long dim2329Kuo2Init[]       =   {   1   ,   1   ,   7   ,   15  ,   11  ,   17  ,   35  ,   39  ,   187 ,   119 ,   541 ,   485 ,   3997    ,   16067   ,   12707   ,0 };
            const unsigned long dim2330Kuo2Init[]       =   {   1   ,   3   ,   7   ,   9   ,   17  ,   5   ,   57  ,   31  ,   75  ,   637 ,   1521    ,   3435    ,   387 ,   9325    ,   8165    ,0 };
            const unsigned long dim2331Kuo2Init[]       =   {   1   ,   1   ,   7   ,   3   ,   17  ,   33  ,   65  ,   55  ,   193 ,   55  ,   777 ,   151 ,   7605    ,   6927    ,   22731   ,0 };
            const unsigned long dim2332Kuo2Init[]       =   {   1   ,   1   ,   3   ,   15  ,   9   ,   21  ,   33  ,   29  ,   395 ,   179 ,   445 ,   895 ,   203 ,   14665   ,   28945   ,0 };
            const unsigned long dim2333Kuo2Init[]       =   {   1   ,   3   ,   3   ,   11  ,   25  ,   63  ,   35  ,   51  ,   261 ,   83  ,   1227    ,   2897    ,   8157    ,   11703   ,   21067   ,0 };
            const unsigned long dim2334Kuo2Init[]       =   {   1   ,   1   ,   5   ,   11  ,   15  ,   59  ,   1   ,   115 ,   225 ,   427 ,   397 ,   2113    ,   8117    ,   9387    ,   15817   ,0 };
            const unsigned long dim2335Kuo2Init[]       =   {   1   ,   1   ,   5   ,   3   ,   25  ,   23  ,   15  ,   3   ,   227 ,   341 ,   449 ,   217 ,   3583    ,   5617    ,   5215    ,0 };
            const unsigned long dim2336Kuo2Init[]       =   {   1   ,   3   ,   1   ,   9   ,   15  ,   61  ,   29  ,   241 ,   245 ,   857 ,   217 ,   1253    ,   2643    ,   435 ,   11991   ,0 };
            const unsigned long dim2337Kuo2Init[]       =   {   1   ,   1   ,   7   ,   7   ,   27  ,   17  ,   89  ,   123 ,   121 ,   97  ,   1749    ,   3341    ,   2279    ,   1067    ,   28845   ,0 };
            const unsigned long dim2338Kuo2Init[]       =   {   1   ,   1   ,   1   ,   13  ,   17  ,   21  ,   27  ,   153 ,   135 ,   831 ,   361 ,   3487    ,   7221    ,   9343    ,   26523   ,0 };
            const unsigned long dim2339Kuo2Init[]       =   {   1   ,   1   ,   7   ,   15  ,   11  ,   59  ,   115 ,   103 ,   425 ,   221 ,   1485    ,   3471    ,   773 ,   14989   ,   1193    ,0 };
            const unsigned long dim2340Kuo2Init[]       =   {   1   ,   1   ,   7   ,   11  ,   1   ,   19  ,   113 ,   115 ,   149 ,   429 ,   1695    ,   409 ,   2343    ,   11843   ,   16319   ,0 };
            const unsigned long dim2341Kuo2Init[]       =   {   1   ,   1   ,   1   ,   9   ,   29  ,   5   ,   37  ,   29  ,   197 ,   331 ,   2033    ,   1319    ,   5667    ,   1311    ,   3563    ,0 };
            const unsigned long dim2342Kuo2Init[]       =   {   1   ,   1   ,   1   ,   11  ,   13  ,   7   ,   73  ,   25  ,   391 ,   245 ,   1447    ,   4001    ,   7059    ,   3091    ,   27423   ,0 };
            const unsigned long dim2343Kuo2Init[]       =   {   1   ,   1   ,   5   ,   13  ,   21  ,   1   ,   125 ,   201 ,   103 ,   349 ,   1123    ,   2067    ,   7007    ,   12927   ,   27169   ,0 };
            const unsigned long dim2344Kuo2Init[]       =   {   1   ,   3   ,   3   ,   5   ,   25  ,   49  ,   73  ,   31  ,   55  ,   659 ,   1407    ,   3433    ,   6513    ,   151 ,   18409   ,0 };
            const unsigned long dim2345Kuo2Init[]       =   {   1   ,   1   ,   5   ,   13  ,   5   ,   19  ,   57  ,   73  ,   197 ,   69  ,   1477    ,   2763    ,   5691    ,   6639    ,   23613   ,0 };
            const unsigned long dim2346Kuo2Init[]       =   {   1   ,   3   ,   5   ,   5   ,   17  ,   47  ,   49  ,   77  ,   275 ,   341 ,   2017    ,   3257    ,   111 ,   925 ,   549 ,0 };
            const unsigned long dim2347Kuo2Init[]       =   {   1   ,   1   ,   7   ,   9   ,   29  ,   7   ,   27  ,   195 ,   411 ,   435 ,   1689    ,   1537    ,   3381    ,   11521   ,   4505    ,0 };
            const unsigned long dim2348Kuo2Init[]       =   {   1   ,   1   ,   1   ,   15  ,   13  ,   23  ,   97  ,   189 ,   285 ,   351 ,   1371    ,   3161    ,   2337    ,   5239    ,   31675   ,0 };
            const unsigned long dim2349Kuo2Init[]       =   {   1   ,   1   ,   5   ,   3   ,   27  ,   57  ,   105 ,   197 ,   423 ,   109 ,   1967    ,   3363    ,   1925    ,   14261   ,   13569   ,0 };
            const unsigned long dim2350Kuo2Init[]       =   {   1   ,   1   ,   3   ,   15  ,   21  ,   15  ,   7   ,   45  ,   411 ,   905 ,   1323    ,   3225    ,   5867    ,   2281    ,   13213   ,0 };
            const unsigned long dim2351Kuo2Init[]       =   {   1   ,   3   ,   7   ,   9   ,   29  ,   51  ,   41  ,   97  ,   393 ,   381 ,   1765    ,   1159    ,   965 ,   6013    ,   12541   ,0 };
            const unsigned long dim2352Kuo2Init[]       =   {   1   ,   1   ,   7   ,   5   ,   13  ,   35  ,   107 ,   231 ,   135 ,   745 ,   259 ,   2999    ,   1291    ,   14025   ,   2065    ,0 };
            const unsigned long dim2353Kuo2Init[]       =   {   1   ,   3   ,   5   ,   13  ,   13  ,   29  ,   65  ,   91  ,   101 ,   469 ,   1337    ,   2717    ,   1781    ,   10019   ,   23097   ,0 };
            const unsigned long dim2354Kuo2Init[]       =   {   1   ,   3   ,   3   ,   9   ,   3   ,   11  ,   23  ,   253 ,   175 ,   35  ,   1651    ,   3713    ,   3575    ,   4837    ,   10385   ,0 };
            const unsigned long dim2355Kuo2Init[]       =   {   1   ,   3   ,   1   ,   11  ,   13  ,   57  ,   33  ,   39  ,   55  ,   717 ,   679 ,   2417    ,   3665    ,   7753    ,   27659   ,0 };
            const unsigned long dim2356Kuo2Init[]       =   {   1   ,   3   ,   1   ,   15  ,   25  ,   9   ,   93  ,   205 ,   369 ,   567 ,   1525    ,   3797    ,   6047    ,   407 ,   2155    ,0 };
            const unsigned long dim2357Kuo2Init[]       =   {   1   ,   3   ,   7   ,   3   ,   5   ,   47  ,   15  ,   227 ,   325 ,   837 ,   1849    ,   3577    ,   1427    ,   5629    ,   17023   ,0 };
            const unsigned long dim2358Kuo2Init[]       =   {   1   ,   3   ,   5   ,   7   ,   17  ,   23  ,   43  ,   231 ,   23  ,   95  ,   423 ,   2225    ,   8051    ,   6775    ,   4087    ,0 };
            const unsigned long dim2359Kuo2Init[]       =   {   1   ,   1   ,   3   ,   15  ,   11  ,   45  ,   87  ,   213 ,   187 ,   159 ,   377 ,   3269    ,   2393    ,   7587    ,   8027    ,0 };
            const unsigned long dim2360Kuo2Init[]       =   {   1   ,   1   ,   5   ,   3   ,   5   ,   25  ,   115 ,   19  ,   229 ,   13  ,   935 ,   3269    ,   3877    ,   8081    ,   20583   ,0 };
            const unsigned long dim2361Kuo2Init[]       =   {   1   ,   3   ,   3   ,   15  ,   31  ,   47  ,   21  ,   179 ,   503 ,   43  ,   915 ,   2987    ,   323 ,   15843   ,   18869   ,0 };
            const unsigned long dim2362Kuo2Init[]       =   {   1   ,   1   ,   7   ,   7   ,   25  ,   9   ,   27  ,   169 ,   431 ,   495 ,   65  ,   3999    ,   4363    ,   8677    ,   12049   ,0 };
            const unsigned long dim2363Kuo2Init[]       =   {   1   ,   3   ,   5   ,   13  ,   17  ,   37  ,   123 ,   197 ,   185 ,   501 ,   1471    ,   149 ,   4649    ,   2337    ,   10279   ,0 };
            const unsigned long dim2364Kuo2Init[]       =   {   1   ,   1   ,   7   ,   13  ,   11  ,   11  ,   3   ,   103 ,   43  ,   97  ,   349 ,   3515    ,   2621    ,   11587   ,   26495   ,0 };
            const unsigned long dim2365Kuo2Init[]       =   {   1   ,   1   ,   1   ,   5   ,   3   ,   19  ,   99  ,   135 ,   355 ,   265 ,   1719    ,   2533    ,   4313    ,   11301   ,   18415   ,0 };
            const unsigned long dim2366Kuo2Init[]       =   {   1   ,   3   ,   7   ,   9   ,   1   ,   55  ,   115 ,   251 ,   293 ,   903 ,   931 ,   3687    ,   5019    ,   14277   ,   589 ,0 };
            const unsigned long dim2367Kuo2Init[]       =   {   1   ,   1   ,   3   ,   5   ,   31  ,   5   ,   123 ,   179 ,   51  ,   135 ,   9   ,   2779    ,   8143    ,   10033   ,   31473   ,0 };
            const unsigned long dim2368Kuo2Init[]       =   {   1   ,   3   ,   1   ,   5   ,   3   ,   39  ,   87  ,   235 ,   289 ,   907 ,   1211    ,   453 ,   611 ,   4009    ,   25787   ,0 };
            const unsigned long dim2369Kuo2Init[]       =   {   1   ,   3   ,   7   ,   9   ,   27  ,   7   ,   47  ,   123 ,   333 ,   713 ,   1349    ,   1345    ,   1641    ,   5935    ,   19127   ,0 };
            const unsigned long dim2370Kuo2Init[]       =   {   1   ,   1   ,   5   ,   11  ,   13  ,   17  ,   47  ,   235 ,   329 ,   955 ,   1761    ,   1957    ,   3611    ,   9705    ,   20307   ,0 };
            const unsigned long dim2371Kuo2Init[]       =   {   1   ,   1   ,   7   ,   15  ,   15  ,   59  ,   45  ,   17  ,   57  ,   871 ,   407 ,   943 ,   1881    ,   12929   ,   2181    ,0 };
            const unsigned long dim2372Kuo2Init[]       =   {   1   ,   1   ,   7   ,   1   ,   7   ,   43  ,   41  ,   193 ,   29  ,   675 ,   1449    ,   2121    ,   1681    ,   13951   ,   21659   ,0 };
            const unsigned long dim2373Kuo2Init[]       =   {   1   ,   1   ,   5   ,   3   ,   21  ,   53  ,   63  ,   75  ,   7   ,   793 ,   245 ,   349 ,   4801    ,   4057    ,   13269   ,0 };
            const unsigned long dim2374Kuo2Init[]       =   {   1   ,   1   ,   3   ,   13  ,   23  ,   9   ,   109 ,   167 ,   165 ,   573 ,   685 ,   129 ,   1781    ,   10331   ,   9429    ,0 };
            const unsigned long dim2375Kuo2Init[]       =   {   1   ,   3   ,   3   ,   15  ,   29  ,   43  ,   109 ,   25  ,   461 ,   991 ,   1509    ,   39  ,   6959    ,   13245   ,   18505   ,0 };
            const unsigned long dim2376Kuo2Init[]       =   {   1   ,   3   ,   1   ,   7   ,   25  ,   43  ,   57  ,   61  ,   399 ,   787 ,   1479    ,   1613    ,   4295    ,   4189    ,   1371    ,0 };
            const unsigned long dim2377Kuo2Init[]       =   {   1   ,   3   ,   3   ,   13  ,   27  ,   57  ,   49  ,   139 ,   189 ,   683 ,   115 ,   963 ,   5833    ,   3249    ,   28785   ,0 };
            const unsigned long dim2378Kuo2Init[]       =   {   1   ,   3   ,   5   ,   1   ,   3   ,   29  ,   41  ,   49  ,   341 ,   299 ,   1195    ,   57  ,   5791    ,   4409    ,   17733   ,0 };
            const unsigned long dim2379Kuo2Init[]       =   {   1   ,   1   ,   5   ,   7   ,   11  ,   21  ,   115 ,   145 ,   207 ,   41  ,   1531    ,   1897    ,   2627    ,   3653    ,   267 ,0 };
            const unsigned long dim2380Kuo2Init[]       =   {   1   ,   3   ,   3   ,   9   ,   1   ,   3   ,   5   ,   79  ,   449 ,   281 ,   1711    ,   2863    ,   6267    ,   14863   ,   29957   ,0 };
            const unsigned long dim2381Kuo2Init[]       =   {   1   ,   3   ,   1   ,   3   ,   15  ,   63  ,   127 ,   123 ,   457 ,   523 ,   1409    ,   2709    ,   4543    ,   13071   ,   17751   ,0 };
            const unsigned long dim2382Kuo2Init[]       =   {   1   ,   3   ,   1   ,   3   ,   7   ,   61  ,   21  ,   139 ,   481 ,   747 ,   107 ,   1041    ,   359 ,   7683    ,   30327   ,0 };
            const unsigned long dim2383Kuo2Init[]       =   {   1   ,   3   ,   3   ,   7   ,   3   ,   25  ,   7   ,   113 ,   39  ,   853 ,   1461    ,   2655    ,   2601    ,   7619    ,   2923    ,0 };
            const unsigned long dim2384Kuo2Init[]       =   {   1   ,   3   ,   3   ,   1   ,   19  ,   25  ,   5   ,   27  ,   439 ,   407 ,   1281    ,   2487    ,   2275    ,   8793    ,   119 ,0 };
            const unsigned long dim2385Kuo2Init[]       =   {   1   ,   3   ,   3   ,   1   ,   11  ,   37  ,   55  ,   131 ,   227 ,   79  ,   117 ,   2367    ,   1495    ,   15873   ,   24341   ,0 };
            const unsigned long dim2386Kuo2Init[]       =   {   1   ,   3   ,   1   ,   1   ,   11  ,   21  ,   35  ,   167 ,   413 ,   645 ,   1139    ,   2535    ,   8167    ,   5937    ,   29047   ,0 };
            const unsigned long dim2387Kuo2Init[]       =   {   1   ,   3   ,   7   ,   1   ,   5   ,   43  ,   31  ,   5   ,   419 ,   577 ,   1333    ,   1599    ,   3745    ,   10219   ,   27501   ,0 };
            const unsigned long dim2388Kuo2Init[]       =   {   1   ,   3   ,   1   ,   13  ,   21  ,   35  ,   115 ,   105 ,   187 ,   567 ,   1303    ,   1637    ,   3681    ,   14089   ,   12705   ,0 };
            const unsigned long dim2389Kuo2Init[]       =   {   1   ,   1   ,   3   ,   15  ,   19  ,   61  ,   65  ,   115 ,   395 ,   501 ,   1015    ,   2117    ,   7647    ,   3077    ,   15265   ,0 };
            const unsigned long dim2390Kuo2Init[]       =   {   1   ,   3   ,   1   ,   1   ,   29  ,   11  ,   21  ,   177 ,   15  ,   193 ,   1177    ,   1225    ,   6251    ,   9983    ,   12599   ,0 };
            const unsigned long dim2391Kuo2Init[]       =   {   1   ,   1   ,   1   ,   5   ,   5   ,   27  ,   27  ,   249 ,   197 ,   125 ,   571 ,   277 ,   4659    ,   14943   ,   19805   ,0 };
            const unsigned long dim2392Kuo2Init[]       =   {   1   ,   3   ,   3   ,   7   ,   27  ,   37  ,   75  ,   235 ,   249 ,   191 ,   1123    ,   1355    ,   4249    ,   14971   ,   11349   ,0 };
            const unsigned long dim2393Kuo2Init[]       =   {   1   ,   1   ,   5   ,   9   ,   31  ,   19  ,   7   ,   91  ,   87  ,   515 ,   293 ,   1289    ,   2495    ,   11311   ,   3939    ,0 };
            const unsigned long dim2394Kuo2Init[]       =   {   1   ,   1   ,   5   ,   15  ,   15  ,   51  ,   43  ,   231 ,   401 ,   557 ,   689 ,   1103    ,   457 ,   3493    ,   22193   ,0 };
            const unsigned long dim2395Kuo2Init[]       =   {   1   ,   3   ,   5   ,   9   ,   25  ,   39  ,   125 ,   91  ,   373 ,   331 ,   39  ,   35  ,   2397    ,   5043    ,   32061   ,0 };
            const unsigned long dim2396Kuo2Init[]       =   {   1   ,   1   ,   3   ,   11  ,   13  ,   37  ,   61  ,   27  ,   349 ,   377 ,   1589    ,   3711    ,   6551    ,   4795    ,   5507    ,0 };
            const unsigned long dim2397Kuo2Init[]       =   {   1   ,   3   ,   5   ,   7   ,   25  ,   49  ,   107 ,   189 ,   371 ,   363 ,   1011    ,   2133    ,   6923    ,   2441    ,   8295    ,0 };
            const unsigned long dim2398Kuo2Init[]       =   {   1   ,   1   ,   3   ,   11  ,   3   ,   11  ,   27  ,   9   ,   193 ,   979 ,   867 ,   3085    ,   3783    ,   14851   ,   11565   ,0 };
            const unsigned long dim2399Kuo2Init[]       =   {   1   ,   1   ,   3   ,   1   ,   1   ,   55  ,   17  ,   239 ,   351 ,   373 ,   939 ,   2115    ,   3697    ,   7881    ,   6603    ,0 };
            const unsigned long dim2400Kuo2Init[]       =   {   1   ,   3   ,   1   ,   7   ,   23  ,   17  ,   81  ,   143 ,   497 ,   875 ,   939 ,   3193    ,   7023    ,   3327    ,   29327   ,0 };
            const unsigned long dim2401Kuo2Init[]       =   {   1   ,   3   ,   7   ,   13  ,   27  ,   61  ,   1   ,   127 ,   369 ,   471 ,   1511    ,   2227    ,   565 ,   4121    ,   11155   ,0 };
            const unsigned long dim2402Kuo2Init[]       =   {   1   ,   3   ,   3   ,   5   ,   5   ,   53  ,   77  ,   185 ,   11  ,   801 ,   1289    ,   975 ,   5191    ,   3733    ,   18489   ,0 };
            const unsigned long dim2403Kuo2Init[]       =   {   1   ,   3   ,   1   ,   7   ,   19  ,   19  ,   67  ,   9   ,   121 ,   473 ,   377 ,   3771    ,   7561    ,   5751    ,   26295   ,0 };
            const unsigned long dim2404Kuo2Init[]       =   {   1   ,   3   ,   1   ,   5   ,   19  ,   41  ,   7   ,   227 ,   465 ,   803 ,   1979    ,   281 ,   1337    ,   7973    ,   20341   ,0 };
            const unsigned long dim2405Kuo2Init[]       =   {   1   ,   1   ,   1   ,   5   ,   9   ,   5   ,   123 ,   45  ,   65  ,   801 ,   373 ,   2093    ,   1473    ,   14503   ,   1401    ,0 };
            const unsigned long dim2406Kuo2Init[]       =   {   1   ,   1   ,   3   ,   9   ,   29  ,   39  ,   123 ,   17  ,   253 ,   649 ,   1755    ,   4081    ,   3415    ,   8805    ,   28635   ,0 };
            const unsigned long dim2407Kuo2Init[]       =   {   1   ,   3   ,   1   ,   1   ,   23  ,   7   ,   57  ,   163 ,   93  ,   829 ,   93  ,   1767    ,   1073    ,   14825   ,   21505   ,0 };
            const unsigned long dim2408Kuo2Init[]       =   {   1   ,   1   ,   7   ,   13  ,   21  ,   59  ,   59  ,   229 ,   175 ,   485 ,   621 ,   3785    ,   4949    ,   12281   ,   14569   ,0 };
            const unsigned long dim2409Kuo2Init[]       =   {   1   ,   3   ,   5   ,   13  ,   7   ,   51  ,   123 ,   221 ,   119 ,   61  ,   1271    ,   1857    ,   3505    ,   14413   ,   20975   ,0 };
            const unsigned long dim2410Kuo2Init[]       =   {   1   ,   3   ,   7   ,   5   ,   17  ,   51  ,   67  ,   123 ,   121 ,   379 ,   605 ,   3361    ,   5557    ,   9127    ,   15251   ,0 };
            const unsigned long dim2411Kuo2Init[]       =   {   1   ,   3   ,   7   ,   9   ,   27  ,   11  ,   33  ,   33  ,   439 ,   347 ,   1305    ,   11  ,   6839    ,   493 ,   771 ,0 };
            const unsigned long dim2412Kuo2Init[]       =   {   1   ,   3   ,   5   ,   15  ,   5   ,   13  ,   7   ,   131 ,   129 ,   865 ,   1115    ,   2651    ,   1567    ,   783 ,   15073   ,0 };
            const unsigned long dim2413Kuo2Init[]       =   {   1   ,   1   ,   3   ,   11  ,   11  ,   13  ,   95  ,   115 ,   361 ,   1009    ,   413 ,   2671    ,   6267    ,   5393    ,   18963   ,0 };
            const unsigned long dim2414Kuo2Init[]       =   {   1   ,   3   ,   1   ,   9   ,   27  ,   11  ,   33  ,   183 ,   343 ,   113 ,   1323    ,   3197    ,   3889    ,   8533    ,   31493   ,0 };
            const unsigned long dim2415Kuo2Init[]       =   {   1   ,   3   ,   3   ,   11  ,   11  ,   21  ,   93  ,   35  ,   177 ,   133 ,   539 ,   679 ,   6263    ,   8619    ,   2751    ,0 };
            const unsigned long dim2416Kuo2Init[]       =   {   1   ,   1   ,   3   ,   1   ,   27  ,   31  ,   21  ,   157 ,   365 ,   635 ,   1279    ,   575 ,   2245    ,   11597   ,   3899    ,0 };
            const unsigned long dim2417Kuo2Init[]       =   {   1   ,   3   ,   1   ,   5   ,   7   ,   45  ,   15  ,   237 ,   191 ,   541 ,   1631    ,   1459    ,   1973    ,   15333   ,   30419   ,0 };
            const unsigned long dim2418Kuo2Init[]       =   {   1   ,   3   ,   1   ,   13  ,   3   ,   47  ,   103 ,   101 ,   423 ,   739 ,   947 ,   3225    ,   783 ,   14771   ,   20183   ,0 };
            const unsigned long dim2419Kuo2Init[]       =   {   1   ,   3   ,   1   ,   5   ,   23  ,   19  ,   21  ,   245 ,   341 ,   907 ,   745 ,   1913    ,   1545    ,   7801    ,   9995    ,0 };
            const unsigned long dim2420Kuo2Init[]       =   {   1   ,   3   ,   1   ,   3   ,   7   ,   47  ,   55  ,   15  ,   193 ,   123 ,   73  ,   427 ,   6415    ,   6611    ,   6317    ,0 };
            const unsigned long dim2421Kuo2Init[]       =   {   1   ,   1   ,   1   ,   7   ,   3   ,   17  ,   27  ,   5   ,   295 ,   453 ,   1503    ,   2327    ,   4209    ,   3663    ,   16843   ,0 };
            const unsigned long dim2422Kuo2Init[]       =   {   1   ,   1   ,   1   ,   3   ,   25  ,   19  ,   41  ,   161 ,   333 ,   153 ,   999 ,   1129    ,   103 ,   9471    ,   12707   ,0 };
            const unsigned long dim2423Kuo2Init[]       =   {   1   ,   3   ,   3   ,   3   ,   21  ,   37  ,   73  ,   159 ,   179 ,   913 ,   691 ,   755 ,   3033    ,   1999    ,   15459   ,0 };
            const unsigned long dim2424Kuo2Init[]       =   {   1   ,   1   ,   5   ,   13  ,   1   ,   3   ,   67  ,   131 ,   119 ,   769 ,   747 ,   1047    ,   6751    ,   2753    ,   22611   ,0 };
            const unsigned long dim2425Kuo2Init[]       =   {   1   ,   3   ,   7   ,   5   ,   11  ,   1   ,   27  ,   81  ,   423 ,   691 ,   357 ,   3679    ,   2467    ,   15107   ,   13029   ,0 };
            const unsigned long dim2426Kuo2Init[]       =   {   1   ,   3   ,   5   ,   9   ,   21  ,   21  ,   109 ,   69  ,   319 ,   283 ,   1831    ,   1403    ,   5531    ,   14461   ,   14057   ,0 };
            const unsigned long dim2427Kuo2Init[]       =   {   1   ,   3   ,   7   ,   13  ,   21  ,   33  ,   77  ,   149 ,   495 ,   361 ,   401 ,   3313    ,   1849    ,   14009   ,   5933    ,0 };
            const unsigned long dim2428Kuo2Init[]       =   {   1   ,   3   ,   7   ,   15  ,   11  ,   3   ,   33  ,   181 ,   21  ,   437 ,   1843    ,   3591    ,   589 ,   1721    ,   27495   ,0 };
            const unsigned long dim2429Kuo2Init[]       =   {   1   ,   1   ,   1   ,   9   ,   25  ,   61  ,   103 ,   25  ,   71  ,   471 ,   219 ,   3623    ,   1517    ,   15471   ,   23179   ,0 };
            const unsigned long dim2430Kuo2Init[]       =   {   1   ,   1   ,   7   ,   1   ,   7   ,   59  ,   99  ,   143 ,   387 ,   231 ,   241 ,   1215    ,   3785    ,   13543   ,   27469   ,0 };
            const unsigned long dim2431Kuo2Init[]       =   {   1   ,   3   ,   1   ,   11  ,   25  ,   61  ,   93  ,   171 ,   421 ,   515 ,   1761    ,   1399    ,   7791    ,   291 ,   28455   ,0 };
            const unsigned long dim2432Kuo2Init[]       =   {   1   ,   1   ,   3   ,   11  ,   1   ,   23  ,   117 ,   251 ,   467 ,   445 ,   737 ,   2637    ,   4183    ,   8561    ,   24977   ,0 };
            const unsigned long dim2433Kuo2Init[]       =   {   1   ,   3   ,   5   ,   11  ,   11  ,   35  ,   111 ,   209 ,   415 ,   485 ,   1609    ,   1113    ,   3329    ,   15365   ,   1999    ,0 };
            const unsigned long dim2434Kuo2Init[]       =   {   1   ,   1   ,   3   ,   15  ,   19  ,   17  ,   43  ,   67  ,   191 ,   711 ,   395 ,   921 ,   2511    ,   8417    ,   4071    ,0 };
            const unsigned long dim2435Kuo2Init[]       =   {   1   ,   3   ,   3   ,   11  ,   9   ,   5   ,   89  ,   115 ,   451 ,   533 ,   1057    ,   1835    ,   2137    ,   9253    ,   19217   ,0 };
            const unsigned long dim2436Kuo2Init[]       =   {   1   ,   3   ,   7   ,   11  ,   9   ,   45  ,   119 ,   89  ,   485 ,   109 ,   1755    ,   709 ,   7729    ,   14659   ,   19203   ,0 };
            const unsigned long dim2437Kuo2Init[]       =   {   1   ,   1   ,   3   ,   7   ,   11  ,   7   ,   107 ,   113 ,   305 ,   431 ,   481 ,   2965    ,   6195    ,   2369    ,   14843   ,0 };
            const unsigned long dim2438Kuo2Init[]       =   {   1   ,   3   ,   1   ,   13  ,   5   ,   37  ,   11  ,   193 ,   135 ,   697 ,   665 ,   3085    ,   4691    ,   3799    ,   11915   ,0 };
            const unsigned long dim2439Kuo2Init[]       =   {   1   ,   1   ,   1   ,   9   ,   1   ,   47  ,   23  ,   143 ,   255 ,   93  ,   615 ,   2039    ,   6591    ,   14479   ,   24963   ,0 };
            const unsigned long dim2440Kuo2Init[]       =   {   1   ,   3   ,   3   ,   11  ,   15  ,   11  ,   75  ,   57  ,   151 ,   29  ,   1073    ,   2763    ,   5613    ,   15703   ,   2447    ,0 };
            const unsigned long dim2441Kuo2Init[]       =   {   1   ,   1   ,   5   ,   9   ,   19  ,   15  ,   79  ,   113 ,   359 ,   997 ,   1799    ,   695 ,   7075    ,   10091   ,   15271   ,0 };
            const unsigned long dim2442Kuo2Init[]       =   {   1   ,   3   ,   5   ,   9   ,   3   ,   63  ,   45  ,   253 ,   257 ,   739 ,   1759    ,   317 ,   1955    ,   1283    ,   11127   ,0 };
            const unsigned long dim2443Kuo2Init[]       =   {   1   ,   3   ,   5   ,   13  ,   29  ,   37  ,   67  ,   177 ,   203 ,   497 ,   947 ,   2339    ,   2401    ,   9199    ,   22331   ,0 };
            const unsigned long dim2444Kuo2Init[]       =   {   1   ,   1   ,   1   ,   5   ,   23  ,   15  ,   67  ,   131 ,   213 ,   21  ,   1167    ,   1729    ,   3297    ,   337 ,   12253   ,0 };
            const unsigned long dim2445Kuo2Init[]       =   {   1   ,   3   ,   3   ,   11  ,   21  ,   7   ,   69  ,   79  ,   297 ,   401 ,   1491    ,   1655    ,   2525    ,   4871    ,   17919   ,0 };
            const unsigned long dim2446Kuo2Init[]       =   {   1   ,   3   ,   5   ,   15  ,   5   ,   15  ,   113 ,   45  ,   205 ,   317 ,   109 ,   2793    ,   4975    ,   5563    ,   3649    ,0 };
            const unsigned long dim2447Kuo2Init[]       =   {   1   ,   3   ,   3   ,   7   ,   13  ,   59  ,   17  ,   63  ,   311 ,   161 ,   419 ,   1693    ,   2937    ,   8671    ,   24741   ,0 };
            const unsigned long dim2448Kuo2Init[]       =   {   1   ,   1   ,   5   ,   3   ,   17  ,   21  ,   103 ,   105 ,   117 ,   211 ,   941 ,   385 ,   6455    ,   9069    ,   10147   ,0 };
            const unsigned long dim2449Kuo2Init[]       =   {   1   ,   3   ,   5   ,   7   ,   5   ,   13  ,   43  ,   155 ,   89  ,   747 ,   359 ,   219 ,   6597    ,   5175    ,   24585   ,0 };
            const unsigned long dim2450Kuo2Init[]       =   {   1   ,   3   ,   5   ,   13  ,   7   ,   45  ,   17  ,   169 ,   421 ,   161 ,   207 ,   2033    ,   4805    ,   2737    ,   7799    ,0 };
            const unsigned long dim2451Kuo2Init[]       =   {   1   ,   1   ,   7   ,   9   ,   15  ,   55  ,   55  ,   219 ,   441 ,   409 ,   443 ,   3363    ,   6895    ,   16339   ,   16091   ,0 };
            const unsigned long dim2452Kuo2Init[]       =   {   1   ,   1   ,   3   ,   5   ,   3   ,   23  ,   17  ,   217 ,   93  ,   799 ,   1419    ,   3405    ,   7911    ,   15609   ,   25461   ,0 };
            const unsigned long dim2453Kuo2Init[]       =   {   1   ,   1   ,   1   ,   5   ,   29  ,   19  ,   17  ,   97  ,   319 ,   743 ,   1773    ,   2513    ,   2395    ,   1885    ,   6835    ,0 };
            const unsigned long dim2454Kuo2Init[]       =   {   1   ,   3   ,   3   ,   3   ,   25  ,   41  ,   79  ,   191 ,   175 ,   319 ,   1903    ,   2281    ,   5709    ,   2817    ,   14213   ,0 };
            const unsigned long dim2455Kuo2Init[]       =   {   1   ,   3   ,   5   ,   9   ,   21  ,   55  ,   9   ,   51  ,   223 ,   519 ,   225 ,   2543    ,   4183    ,   283 ,   599 ,0 };
            const unsigned long dim2456Kuo2Init[]       =   {   1   ,   1   ,   1   ,   15  ,   23  ,   39  ,   21  ,   171 ,   167 ,   739 ,   1779    ,   3469    ,   8121    ,   7013    ,   23445   ,0 };
            const unsigned long dim2457Kuo2Init[]       =   {   1   ,   3   ,   3   ,   9   ,   9   ,   5   ,   55  ,   253 ,   43  ,   397 ,   927 ,   237 ,   4459    ,   12817   ,   29563   ,0 };
            const unsigned long dim2458Kuo2Init[]       =   {   1   ,   3   ,   7   ,   3   ,   3   ,   57  ,   113 ,   221 ,   391 ,   365 ,   311 ,   1027    ,   3369    ,   7637    ,   25765   ,0 };
            const unsigned long dim2459Kuo2Init[]       =   {   1   ,   3   ,   3   ,   1   ,   19  ,   3   ,   91  ,   161 ,   421 ,   961 ,   1237    ,   4063    ,   2573    ,   483 ,   3375    ,0 };
            const unsigned long dim2460Kuo2Init[]       =   {   1   ,   1   ,   1   ,   5   ,   19  ,   35  ,   119 ,   211 ,   489 ,   649 ,   1233    ,   447 ,   3727    ,   15541   ,   4769    ,0 };
            const unsigned long dim2461Kuo2Init[]       =   {   1   ,   3   ,   7   ,   15  ,   7   ,   1   ,   15  ,   103 ,   81  ,   59  ,   1953    ,   3729    ,   6451    ,   15051   ,   17421   ,0 };
            const unsigned long dim2462Kuo2Init[]       =   {   1   ,   1   ,   3   ,   11  ,   11  ,   41  ,   73  ,   63  ,   485 ,   47  ,   585 ,   3503    ,   4787    ,   12445   ,   8513    ,0 };
            const unsigned long dim2463Kuo2Init[]       =   {   1   ,   1   ,   5   ,   1   ,   29  ,   35  ,   37  ,   99  ,   443 ,   171 ,   1455    ,   1705    ,   5587    ,   8791    ,   6269    ,0 };
            const unsigned long dim2464Kuo2Init[]       =   {   1   ,   3   ,   3   ,   9   ,   19  ,   23  ,   103 ,   19  ,   307 ,   615 ,   2037    ,   3289    ,   1453    ,   12955   ,   28069   ,0 };
            const unsigned long dim2465Kuo2Init[]       =   {   1   ,   3   ,   7   ,   5   ,   21  ,   57  ,   89  ,   29  ,   419 ,   681 ,   1413    ,   3375    ,   7341    ,   14825   ,   18251   ,0 };
            const unsigned long dim2466Kuo2Init[]       =   {   1   ,   3   ,   5   ,   11  ,   21  ,   49  ,   67  ,   69  ,   47  ,   33  ,   449 ,   1831    ,   1125    ,   1011    ,   16223   ,0 };
            const unsigned long dim2467Kuo2Init[]       =   {   1   ,   1   ,   3   ,   5   ,   17  ,   11  ,   67  ,   237 ,   299 ,   843 ,   1177    ,   1825    ,   1109    ,   12157   ,   9347    ,0 };
            const unsigned long dim2468Kuo2Init[]       =   {   1   ,   3   ,   3   ,   9   ,   21  ,   63  ,   47  ,   239 ,   219 ,   319 ,   1505    ,   393 ,   143 ,   5009    ,   9017    ,0 };
            const unsigned long dim2469Kuo2Init[]       =   {   1   ,   3   ,   5   ,   11  ,   1   ,   25  ,   21  ,   1   ,   15  ,   947 ,   1533    ,   1201    ,   3929    ,   7943    ,   9335    ,0 };
            const unsigned long dim2470Kuo2Init[]       =   {   1   ,   1   ,   1   ,   15  ,   29  ,   45  ,   29  ,   119 ,   419 ,   793 ,   345 ,   789 ,   6415    ,   5481    ,   31111   ,0 };
            const unsigned long dim2471Kuo2Init[]       =   {   1   ,   1   ,   1   ,   1   ,   19  ,   59  ,   3   ,   199 ,   247 ,   905 ,   1347    ,   4021    ,   1913    ,   15315   ,   15265   ,0 };
            const unsigned long dim2472Kuo2Init[]       =   {   1   ,   3   ,   3   ,   7   ,   23  ,   63  ,   1   ,   87  ,   455 ,   1013    ,   789 ,   603 ,   7027    ,   11527   ,   17571   ,0 };
            const unsigned long dim2473Kuo2Init[]       =   {   1   ,   3   ,   3   ,   15  ,   13  ,   27  ,   95  ,   61  ,   343 ,   201 ,   997 ,   2153    ,   3239    ,   2475    ,   9827    ,0 };
            const unsigned long dim2474Kuo2Init[]       =   {   1   ,   3   ,   3   ,   1   ,   29  ,   49  ,   77  ,   245 ,   167 ,   873 ,   1467    ,   2475    ,   3951    ,   10253   ,   8345    ,0 };
            const unsigned long dim2475Kuo2Init[]       =   {   1   ,   1   ,   1   ,   15  ,   5   ,   41  ,   71  ,   81  ,   491 ,   971 ,   1955    ,   1901    ,   4383    ,   13961   ,   3745    ,0 };
            const unsigned long dim2476Kuo2Init[]       =   {   1   ,   1   ,   1   ,   7   ,   21  ,   31  ,   11  ,   247 ,   47  ,   251 ,   509 ,   739 ,   2745    ,   14531   ,   19633   ,0 };
            const unsigned long dim2477Kuo2Init[]       =   {   1   ,   1   ,   7   ,   13  ,   31  ,   47  ,   105 ,   167 ,   443 ,   623 ,   503 ,   3283    ,   1935    ,   2627    ,   19807   ,0 };
            const unsigned long dim2478Kuo2Init[]       =   {   1   ,   1   ,   7   ,   3   ,   21  ,   9   ,   67  ,   39  ,   263 ,   395 ,   407 ,   3813    ,   5033    ,   14647   ,   11963   ,0 };
            const unsigned long dim2479Kuo2Init[]       =   {   1   ,   1   ,   1   ,   9   ,   3   ,   29  ,   9   ,   101 ,   313 ,   861 ,   529 ,   1171    ,   7943    ,   13701   ,   16497   ,0 };
            const unsigned long dim2480Kuo2Init[]       =   {   1   ,   1   ,   5   ,   15  ,   9   ,   3   ,   29  ,   53  ,   379 ,   721 ,   1777    ,   2767    ,   7591    ,   8923    ,   16909   ,0 };
            const unsigned long dim2481Kuo2Init[]       =   {   1   ,   1   ,   3   ,   9   ,   19  ,   7   ,   103 ,   189 ,   325 ,   995 ,   1419    ,   333 ,   4053    ,   10441   ,   1287    ,0 };
            const unsigned long dim2482Kuo2Init[]       =   {   1   ,   1   ,   7   ,   1   ,   1   ,   1   ,   53  ,   31  ,   425 ,   477 ,   695 ,   389 ,   6523    ,   7565    ,   95  ,0 };
            const unsigned long dim2483Kuo2Init[]       =   {   1   ,   1   ,   7   ,   15  ,   1   ,   5   ,   75  ,   211 ,   487 ,   641 ,   287 ,   3709    ,   2257    ,   12017   ,   31921   ,0 };
            const unsigned long dim2484Kuo2Init[]       =   {   1   ,   3   ,   5   ,   7   ,   31  ,   7   ,   33  ,   177 ,   349 ,   257 ,   1383    ,   303 ,   469 ,   10941   ,   16211   ,0 };
            const unsigned long dim2485Kuo2Init[]       =   {   1   ,   1   ,   5   ,   1   ,   7   ,   21  ,   89  ,   229 ,   237 ,   447 ,   1805    ,   3271    ,   7073    ,   3401    ,   7143    ,0 };
            const unsigned long dim2486Kuo2Init[]       =   {   1   ,   3   ,   1   ,   7   ,   11  ,   35  ,   127 ,   139 ,   497 ,   629 ,   543 ,   77  ,   2505    ,   8781    ,   30445   ,0 };
            const unsigned long dim2487Kuo2Init[]       =   {   1   ,   1   ,   7   ,   13  ,   1   ,   27  ,   59  ,   125 ,   399 ,   37  ,   597 ,   431 ,   4531    ,   7525    ,   2629    ,0 };
            const unsigned long dim2488Kuo2Init[]       =   {   1   ,   3   ,   3   ,   1   ,   25  ,   27  ,   75  ,   221 ,   269 ,   587 ,   669 ,   1807    ,   5437    ,   12715   ,   22419   ,0 };
            const unsigned long dim2489Kuo2Init[]       =   {   1   ,   1   ,   5   ,   7   ,   31  ,   29  ,   7   ,   135 ,   493 ,   1011    ,   1273    ,   3899    ,   6069    ,   9657    ,   21375   ,0 };
            const unsigned long dim2490Kuo2Init[]       =   {   1   ,   3   ,   5   ,   7   ,   3   ,   13  ,   123 ,   39  ,   37  ,   397 ,   1023    ,   2733    ,   2275    ,   8415    ,   12011   ,0 };
            const unsigned long dim2491Kuo2Init[]       =   {   1   ,   1   ,   5   ,   3   ,   1   ,   63  ,   115 ,   191 ,   159 ,   283 ,   7   ,   1073    ,   5609    ,   13113   ,   1755    ,0 };
            const unsigned long dim2492Kuo2Init[]       =   {   1   ,   1   ,   3   ,   5   ,   11  ,   35  ,   83  ,   91  ,   315 ,   823 ,   1961    ,   1023    ,   4191    ,   1013    ,   24451   ,0 };
            const unsigned long dim2493Kuo2Init[]       =   {   1   ,   1   ,   3   ,   15  ,   31  ,   13  ,   121 ,   213 ,   289 ,   763 ,   297 ,   3405    ,   4987    ,   13713   ,   16853   ,0 };
            const unsigned long dim2494Kuo2Init[]       =   {   1   ,   3   ,   1   ,   3   ,   9   ,   57  ,   53  ,   171 ,   203 ,   71  ,   1973    ,   3555    ,   5079    ,   16173   ,   6399    ,0 };
            const unsigned long dim2495Kuo2Init[]       =   {   1   ,   1   ,   3   ,   7   ,   21  ,   51  ,   105 ,   237 ,   491 ,   465 ,   597 ,   3055    ,   1657    ,   131 ,   19281   ,0 };
            const unsigned long dim2496Kuo2Init[]       =   {   1   ,   1   ,   7   ,   15  ,   15  ,   41  ,   1   ,   61  ,   3   ,   879 ,   1985    ,   651 ,   3945    ,   6979    ,   793 ,0 };
            const unsigned long dim2497Kuo2Init[]       =   {   1   ,   1   ,   1   ,   3   ,   15  ,   45  ,   51  ,   23  ,   1   ,   265 ,   1345    ,   3967    ,   1419    ,   247 ,   9225    ,0 };
            const unsigned long dim2498Kuo2Init[]       =   {   1   ,   1   ,   5   ,   5   ,   19  ,   31  ,   107 ,   73  ,   331 ,   135 ,   1645    ,   3465    ,   6809    ,   12793   ,   15589   ,0 };
            const unsigned long dim2499Kuo2Init[]       =   {   1   ,   1   ,   1   ,   1   ,   29  ,   47  ,   103 ,   133 ,   15  ,   113 ,   717 ,   891 ,   1967    ,   6063    ,   24975   ,0 };
            const unsigned long dim2500Kuo2Init[]       =   {   1   ,   1   ,   3   ,   1   ,   17  ,   39  ,   3   ,   103 ,   373 ,   379 ,   267 ,   131 ,   473 ,   10475   ,   18595   ,0 };
            const unsigned long dim2501Kuo2Init[]       =   {   1   ,   1   ,   3   ,   3   ,   27  ,   51  ,   17  ,   227 ,   425 ,   29  ,   969 ,   1515    ,   7045    ,   5869    ,   2031    ,0 };
            const unsigned long dim2502Kuo2Init[]       =   {   1   ,   3   ,   7   ,   5   ,   17  ,   59  ,   69  ,   21  ,   119 ,   359 ,   1805    ,   1225    ,   1969    ,   14601   ,   7175    ,0 };
            const unsigned long dim2503Kuo2Init[]       =   {   1   ,   3   ,   7   ,   9   ,   5   ,   23  ,   89  ,   89  ,   269 ,   159 ,   2017    ,   269 ,   4247    ,   14369   ,   4413    ,0 };
            const unsigned long dim2504Kuo2Init[]       =   {   1   ,   1   ,   7   ,   1   ,   19  ,   35  ,   67  ,   243 ,   71  ,   653 ,   1673    ,   647 ,   1171    ,   16167   ,   15223   ,0 };
            const unsigned long dim2505Kuo2Init[]       =   {   1   ,   1   ,   7   ,   7   ,   19  ,   37  ,   27  ,   69  ,   253 ,   233 ,   933 ,   649 ,   5449    ,   10651   ,   24709   ,0 };
            const unsigned long dim2506Kuo2Init[]       =   {   1   ,   1   ,   5   ,   15  ,   13  ,   39  ,   113 ,   253 ,   351 ,   281 ,   309 ,   2351    ,   2225    ,   14121   ,   24059   ,0 };
            const unsigned long dim2507Kuo2Init[]       =   {   1   ,   3   ,   1   ,   9   ,   13  ,   43  ,   115 ,   137 ,   107 ,   921 ,   37  ,   1743    ,   183 ,   5037    ,   4649    ,0 };
            const unsigned long dim2508Kuo2Init[]       =   {   1   ,   1   ,   5   ,   15  ,   23  ,   11  ,   121 ,   161 ,   337 ,   155 ,   1053    ,   2907    ,   1163    ,   12929   ,   9065    ,0 };
            const unsigned long dim2509Kuo2Init[]       =   {   1   ,   1   ,   3   ,   1   ,   1   ,   23  ,   85  ,   253 ,   245 ,   359 ,   1881    ,   581 ,   381 ,   15877   ,   15781   ,0 };
            const unsigned long dim2510Kuo2Init[]       =   {   1   ,   3   ,   5   ,   11  ,   17  ,   49  ,   33  ,   133 ,   247 ,   339 ,   969 ,   3347    ,   1791    ,   16115   ,   10821   ,0 };
            const unsigned long dim2511Kuo2Init[]       =   {   1   ,   3   ,   1   ,   9   ,   31  ,   41  ,   31  ,   103 ,   79  ,   339 ,   1311    ,   2609    ,   7455    ,   4543    ,   14765   ,0 };
            const unsigned long dim2512Kuo2Init[]       =   {   1   ,   3   ,   7   ,   13  ,   13  ,   25  ,   75  ,   219 ,   443 ,   543 ,   9   ,   1101    ,   6833    ,   2091    ,   20729   ,0 };
            const unsigned long dim2513Kuo2Init[]       =   {   1   ,   1   ,   3   ,   9   ,   3   ,   51  ,   125 ,   137 ,   107 ,   123 ,   1113    ,   3735    ,   655 ,   6897    ,   32047   ,0 };
            const unsigned long dim2514Kuo2Init[]       =   {   1   ,   1   ,   3   ,   1   ,   17  ,   49  ,   31  ,   87  ,   399 ,   405 ,   1497    ,   1199    ,   6135    ,   6889    ,   23913   ,0 };
            const unsigned long dim2515Kuo2Init[]       =   {   1   ,   1   ,   1   ,   13  ,   17  ,   59  ,   87  ,   153 ,   183 ,   539 ,   1231    ,   569 ,   1877    ,   11605   ,   9301    ,0 };
            const unsigned long dim2516Kuo2Init[]       =   {   1   ,   3   ,   1   ,   9   ,   15  ,   57  ,   41  ,   247 ,   107 ,   1019    ,   1423    ,   49  ,   3093    ,   1667    ,   28241   ,0 };
            const unsigned long dim2517Kuo2Init[]       =   {   1   ,   1   ,   1   ,   1   ,   19  ,   9   ,   41  ,   79  ,   383 ,   83  ,   1091    ,   457 ,   7245    ,   3639    ,   24411   ,0 };
            const unsigned long dim2518Kuo2Init[]       =   {   1   ,   3   ,   7   ,   11  ,   31  ,   23  ,   71  ,   157 ,   327 ,   973 ,   39  ,   1303    ,   3519    ,   6071    ,   19297   ,0 };
            const unsigned long dim2519Kuo2Init[]       =   {   1   ,   1   ,   7   ,   13  ,   27  ,   17  ,   105 ,   245 ,   83  ,   825 ,   33  ,   2331    ,   6469    ,   5547    ,   5409    ,0 };
            const unsigned long dim2520Kuo2Init[]       =   {   1   ,   3   ,   1   ,   9   ,   21  ,   1   ,   105 ,   161 ,   269 ,   767 ,   2039    ,   941 ,   5063    ,   2763    ,   31333   ,0 };
            const unsigned long dim2521Kuo2Init[]       =   {   1   ,   3   ,   1   ,   1   ,   17  ,   33  ,   37  ,   185 ,   505 ,   979 ,   239 ,   3001    ,   1441    ,   7803    ,   23377   ,0 };
            const unsigned long dim2522Kuo2Init[]       =   {   1   ,   1   ,   3   ,   3   ,   23  ,   5   ,   63  ,   249 ,   409 ,   501 ,   1177    ,   655 ,   5109    ,   309 ,   3791    ,0 };
            const unsigned long dim2523Kuo2Init[]       =   {   1   ,   1   ,   5   ,   1   ,   11  ,   37  ,   67  ,   5   ,   399 ,   479 ,   1119    ,   295 ,   2343    ,   11141   ,   28209   ,0 };
            const unsigned long dim2524Kuo2Init[]       =   {   1   ,   1   ,   5   ,   11  ,   3   ,   17  ,   75  ,   93  ,   69  ,   957 ,   1851    ,   2297    ,   6757    ,   9401    ,   15847   ,0 };
            const unsigned long dim2525Kuo2Init[]       =   {   1   ,   1   ,   3   ,   9   ,   5   ,   39  ,   75  ,   143 ,   235 ,   411 ,   1309    ,   1327    ,   863 ,   5579    ,   5153    ,0 };
            const unsigned long dim2526Kuo2Init[]       =   {   1   ,   1   ,   7   ,   13  ,   23  ,   3   ,   39  ,   7   ,   257 ,   941 ,   1441    ,   2931    ,   1651    ,   4669    ,   16193   ,0 };
            const unsigned long dim2527Kuo2Init[]       =   {   1   ,   1   ,   1   ,   15  ,   27  ,   43  ,   51  ,   197 ,   369 ,   531 ,   397 ,   387 ,   3021    ,   8885    ,   21943   ,0 };
            const unsigned long dim2528Kuo2Init[]       =   {   1   ,   1   ,   5   ,   11  ,   25  ,   49  ,   9   ,   227 ,   501 ,   687 ,   131 ,   2899    ,   7175    ,   11359   ,   19997   ,0 };
            const unsigned long dim2529Kuo2Init[]       =   {   1   ,   3   ,   3   ,   9   ,   23  ,   1   ,   51  ,   53  ,   437 ,   331 ,   1503    ,   3345    ,   5909    ,   439 ,   4771    ,0 };
            const unsigned long dim2530Kuo2Init[]       =   {   1   ,   3   ,   1   ,   7   ,   13  ,   13  ,   101 ,   71  ,   345 ,   675 ,   653 ,   2297    ,   1099    ,   11115   ,   28917   ,0 };
            const unsigned long dim2531Kuo2Init[]       =   {   1   ,   1   ,   3   ,   5   ,   5   ,   23  ,   69  ,   155 ,   133 ,   527 ,   1443    ,   1099    ,   2395    ,   2213    ,   16533   ,0 };
            const unsigned long dim2532Kuo2Init[]       =   {   1   ,   1   ,   7   ,   7   ,   9   ,   37  ,   75  ,   17  ,   397 ,   919 ,   1223    ,   491 ,   4719    ,   3915    ,   30101   ,0 };
            const unsigned long dim2533Kuo2Init[]       =   {   1   ,   3   ,   1   ,   7   ,   7   ,   51  ,   105 ,   31  ,   5   ,   149 ,   1275    ,   1381    ,   1963    ,   14737   ,   30975   ,0 };
            const unsigned long dim2534Kuo2Init[]       =   {   1   ,   1   ,   7   ,   5   ,   1   ,   51  ,   49  ,   21  ,   371 ,   293 ,   159 ,   3681    ,   7979    ,   15611   ,   255 ,0 };
            const unsigned long dim2535Kuo2Init[]       =   {   1   ,   1   ,   1   ,   7   ,   27  ,   53  ,   51  ,   135 ,   181 ,   419 ,   757 ,   2913    ,   941 ,   5863    ,   19619   ,0 };
            const unsigned long dim2536Kuo2Init[]       =   {   1   ,   1   ,   7   ,   5   ,   19  ,   1   ,   115 ,   113 ,   29  ,   355 ,   1203    ,   279 ,   7967    ,   12419   ,   1003    ,0 };
            const unsigned long dim2537Kuo2Init[]       =   {   1   ,   1   ,   1   ,   11  ,   1   ,   63  ,   109 ,   215 ,   285 ,   801 ,   1995    ,   3771    ,   4113    ,   11973   ,   16507   ,0 };
            const unsigned long dim2538Kuo2Init[]       =   {   1   ,   3   ,   3   ,   9   ,   19  ,   55  ,   107 ,   139 ,   95  ,   139 ,   1595    ,   929 ,   7023    ,   15245   ,   30163   ,0 };
            const unsigned long dim2539Kuo2Init[]       =   {   1   ,   1   ,   1   ,   13  ,   19  ,   1   ,   33  ,   23  ,   325 ,   245 ,   1671    ,   3065    ,   6767    ,   13583   ,   4683    ,0 };
            const unsigned long dim2540Kuo2Init[]       =   {   1   ,   1   ,   7   ,   5   ,   17  ,   55  ,   33  ,   235 ,   119 ,   715 ,   83  ,   1089    ,   5943    ,   11897   ,   13259   ,0 };
            const unsigned long dim2541Kuo2Init[]       =   {   1   ,   3   ,   7   ,   5   ,   25  ,   27  ,   111 ,   225 ,   99  ,   917 ,   731 ,   2617    ,   7915    ,   7283    ,   357 ,0 };
            const unsigned long dim2542Kuo2Init[]       =   {   1   ,   3   ,   1   ,   3   ,   13  ,   9   ,   127 ,   251 ,   189 ,   573 ,   1897    ,   2147    ,   2641    ,   13191   ,   3847    ,0 };
            const unsigned long dim2543Kuo2Init[]       =   {   1   ,   3   ,   7   ,   1   ,   19  ,   49  ,   49  ,   231 ,   321 ,   67  ,   1067    ,   2267    ,   7873    ,   2825    ,   445 ,0 };
            const unsigned long dim2544Kuo2Init[]       =   {   1   ,   3   ,   7   ,   5   ,   27  ,   55  ,   23  ,   155 ,   103 ,   969 ,   1929    ,   2171    ,   5601    ,   1267    ,   31897   ,0 };
            const unsigned long dim2545Kuo2Init[]       =   {   1   ,   3   ,   3   ,   7   ,   3   ,   27  ,   107 ,   7   ,   61  ,   601 ,   1957    ,   4075    ,   2955    ,   7717    ,   10473   ,0 };
            const unsigned long dim2546Kuo2Init[]       =   {   1   ,   3   ,   5   ,   13  ,   31  ,   25  ,   111 ,   11  ,   391 ,   727 ,   243 ,   1471    ,   143 ,   16161   ,   5591    ,0 };
            const unsigned long dim2547Kuo2Init[]       =   {   1   ,   1   ,   1   ,   11  ,   11  ,   43  ,   117 ,   117 ,   173 ,   933 ,   163 ,   1571    ,   1901    ,   3201    ,   1151    ,0 };
            const unsigned long dim2548Kuo2Init[]       =   {   1   ,   3   ,   5   ,   3   ,   17  ,   13  ,   47  ,   217 ,   351 ,   777 ,   607 ,   401 ,   3485    ,   8679    ,   20357   ,0 };
            const unsigned long dim2549Kuo2Init[]       =   {   1   ,   1   ,   7   ,   15  ,   5   ,   13  ,   37  ,   7   ,   469 ,   413 ,   1343    ,   947 ,   5883    ,   6125    ,   14803   ,0 };
            const unsigned long dim2550Kuo2Init[]       =   {   1   ,   3   ,   1   ,   3   ,   1   ,   61  ,   7   ,   67  ,   229 ,   931 ,   1997    ,   3067    ,   3507    ,   11675   ,   31837   ,0 };
            const unsigned long dim2551Kuo2Init[]       =   {   1   ,   1   ,   5   ,   1   ,   5   ,   35  ,   43  ,   65  ,   101 ,   75  ,   629 ,   1149    ,   625 ,   1865    ,   5215    ,0 };
            const unsigned long dim2552Kuo2Init[]       =   {   1   ,   1   ,   7   ,   7   ,   17  ,   61  ,   47  ,   29  ,   277 ,   861 ,   1901    ,   35  ,   1405    ,   4393    ,   26899   ,0 };
            const unsigned long dim2553Kuo2Init[]       =   {   1   ,   1   ,   5   ,   13  ,   13  ,   11  ,   21  ,   27  ,   473 ,   145 ,   867 ,   2781    ,   2865    ,   2793    ,   3265    ,0 };
            const unsigned long dim2554Kuo2Init[]       =   {   1   ,   1   ,   7   ,   11  ,   11  ,   31  ,   41  ,   119 ,   243 ,   989 ,   2047    ,   1455    ,   8037    ,   11633   ,   31199   ,0 };
            const unsigned long dim2555Kuo2Init[]       =   {   1   ,   1   ,   7   ,   5   ,   29  ,   19  ,   1   ,   89  ,   63  ,   41  ,   323 ,   2135    ,   2091    ,   15061   ,   29039   ,0 };
            const unsigned long dim2556Kuo2Init[]       =   {   1   ,   3   ,   5   ,   15  ,   9   ,   53  ,   15  ,   191 ,   229 ,   487 ,   1287    ,   1163    ,   6985    ,   881 ,   4789    ,0 };
            const unsigned long dim2557Kuo2Init[]       =   {   1   ,   1   ,   3   ,   9   ,   11  ,   57  ,   89  ,   205 ,   355 ,   551 ,   1643    ,   1565    ,   4181    ,   10805   ,   29359   ,0 };
            const unsigned long dim2558Kuo2Init[]       =   {   1   ,   1   ,   7   ,   7   ,   31  ,   3   ,   103 ,   165 ,   37  ,   195 ,   245 ,   1225    ,   2811    ,   14785   ,   9307    ,0 };
            const unsigned long dim2559Kuo2Init[]       =   {   1   ,   1   ,   7   ,   5   ,   19  ,   31  ,   47  ,   115 ,   361 ,   1019    ,   1517    ,   4079    ,   3717    ,   11301   ,   12051   ,0 };
            const unsigned long dim2560Kuo2Init[]       =   {   1   ,   3   ,   5   ,   9   ,   31  ,   21  ,   57  ,   27  ,   353 ,   623 ,   1217    ,   3079    ,   5429    ,   8645    ,   10773   ,0 };
            const unsigned long dim2561Kuo2Init[]       =   {   1   ,   1   ,   7   ,   7   ,   3   ,   11  ,   55  ,   209 ,   239 ,   509 ,   959 ,   2091    ,   123 ,   11365   ,   26593   ,0 };
            const unsigned long dim2562Kuo2Init[]       =   {   1   ,   1   ,   5   ,   9   ,   7   ,   19  ,   51  ,   179 ,   103 ,   901 ,   1717    ,   1837    ,   4327    ,   9753    ,   24147   ,0 };
            const unsigned long dim2563Kuo2Init[]       =   {   1   ,   1   ,   1   ,   5   ,   9   ,   63  ,   19  ,   103 ,   265 ,   893 ,   1519    ,   4063    ,   7951    ,   15601   ,   15019   ,0 };
            const unsigned long dim2564Kuo2Init[]       =   {   1   ,   1   ,   3   ,   1   ,   19  ,   29  ,   15  ,   11  ,   235 ,   785 ,   77  ,   3957    ,   4829    ,   9903    ,   8483    ,0 };
            const unsigned long dim2565Kuo2Init[]       =   {   1   ,   1   ,   3   ,   9   ,   27  ,   33  ,   25  ,   103 ,   387 ,   131 ,   1555    ,   301 ,   1471    ,   11331   ,   15091   ,0 };
            const unsigned long dim2566Kuo2Init[]       =   {   1   ,   3   ,   3   ,   13  ,   25  ,   63  ,   75  ,   245 ,   189 ,   527 ,   1829    ,   1943    ,   407 ,   15219   ,   24453   ,0 };
            const unsigned long dim2567Kuo2Init[]       =   {   1   ,   3   ,   1   ,   11  ,   29  ,   27  ,   35  ,   5   ,   215 ,   949 ,   97  ,   3843    ,   7911    ,   15947   ,   25037   ,0 };
            const unsigned long dim2568Kuo2Init[]       =   {   1   ,   3   ,   3   ,   9   ,   29  ,   49  ,   43  ,   245 ,   257 ,   741 ,   1067    ,   2405    ,   3793    ,   3787    ,   22921   ,0 };
            const unsigned long dim2569Kuo2Init[]       =   {   1   ,   1   ,   3   ,   15  ,   29  ,   29  ,   39  ,   25  ,   485 ,   257 ,   1003    ,   77  ,   1617    ,   5151    ,   29595   ,0 };
            const unsigned long dim2570Kuo2Init[]       =   {   1   ,   1   ,   3   ,   5   ,   9   ,   39  ,   61  ,   59  ,   197 ,   501 ,   527 ,   3371    ,   1739    ,   5697    ,   11857   ,0 };
            const unsigned long dim2571Kuo2Init[]       =   {   1   ,   1   ,   1   ,   15  ,   31  ,   27  ,   41  ,   91  ,   11  ,   373 ,   627 ,   361 ,   8001    ,   12231   ,   5715    ,0 };
            const unsigned long dim2572Kuo2Init[]       =   {   1   ,   3   ,   3   ,   9   ,   31  ,   27  ,   67  ,   203 ,   95  ,   221 ,   1911    ,   1045    ,   1311    ,   11633   ,   5375    ,0 };
            const unsigned long dim2573Kuo2Init[]       =   {   1   ,   1   ,   5   ,   15  ,   31  ,   49  ,   59  ,   185 ,   285 ,   173 ,   1351    ,   1665    ,   5941    ,   2075    ,   26087   ,0 };
            const unsigned long dim2574Kuo2Init[]       =   {   1   ,   1   ,   7   ,   9   ,   23  ,   33  ,   95  ,   171 ,   71  ,   743 ,   11  ,   33  ,   1263    ,   1595    ,   18281   ,0 };
            const unsigned long dim2575Kuo2Init[]       =   {   1   ,   1   ,   1   ,   7   ,   13  ,   59  ,   99  ,   245 ,   497 ,   951 ,   1195    ,   1503    ,   6069    ,   2293    ,   31661   ,0 };
            const unsigned long dim2576Kuo2Init[]       =   {   1   ,   3   ,   5   ,   3   ,   3   ,   3   ,   29  ,   3   ,   197 ,   697 ,   105 ,   435 ,   4583    ,   5407    ,   32727   ,0 };
            const unsigned long dim2577Kuo2Init[]       =   {   1   ,   3   ,   7   ,   1   ,   9   ,   41  ,   77  ,   199 ,   495 ,   999 ,   1791    ,   2183    ,   2819    ,   10771   ,   6957    ,0 };
            const unsigned long dim2578Kuo2Init[]       =   {   1   ,   1   ,   7   ,   15  ,   27  ,   13  ,   51  ,   33  ,   243 ,   947 ,   835 ,   2251    ,   1751    ,   10319   ,   3269    ,0 };
            const unsigned long dim2579Kuo2Init[]       =   {   1   ,   3   ,   5   ,   7   ,   17  ,   59  ,   125 ,   151 ,   113 ,   889 ,   21  ,   3839    ,   1475    ,   10643   ,   3291    ,0 };
            const unsigned long dim2580Kuo2Init[]       =   {   1   ,   3   ,   7   ,   3   ,   27  ,   55  ,   101 ,   13  ,   177 ,   967 ,   35  ,   3081    ,   3559    ,   6539    ,   13557   ,0 };
            const unsigned long dim2581Kuo2Init[]       =   {   1   ,   3   ,   5   ,   13  ,   27  ,   7   ,   107 ,   147 ,   455 ,   605 ,   1589    ,   2325    ,   6891    ,   10431   ,   18227   ,0 };
            const unsigned long dim2582Kuo2Init[]       =   {   1   ,   1   ,   3   ,   3   ,   27  ,   13  ,   121 ,   249 ,   205 ,   359 ,   1909    ,   4011    ,   6833    ,   11583   ,   8705    ,0 };
            const unsigned long dim2583Kuo2Init[]       =   {   1   ,   3   ,   5   ,   9   ,   21  ,   47  ,   77  ,   103 ,   173 ,   637 ,   1921    ,   991 ,   6471    ,   6053    ,   1963    ,0 };
            const unsigned long dim2584Kuo2Init[]       =   {   1   ,   1   ,   3   ,   15  ,   11  ,   9   ,   71  ,   161 ,   253 ,   1   ,   973 ,   3079    ,   4887    ,   10045   ,   20467   ,0 };
            const unsigned long dim2585Kuo2Init[]       =   {   1   ,   1   ,   5   ,   13  ,   27  ,   45  ,   11  ,   147 ,   79  ,   375 ,   533 ,   43  ,   7217    ,   9393    ,   20773   ,0 };
            const unsigned long dim2586Kuo2Init[]       =   {   1   ,   1   ,   5   ,   3   ,   21  ,   47  ,   107 ,   83  ,   505 ,   463 ,   967 ,   3437    ,   3051    ,   11239   ,   20271   ,0 };
            const unsigned long dim2587Kuo2Init[]       =   {   1   ,   1   ,   1   ,   15  ,   7   ,   9   ,   97  ,   143 ,   483 ,   393 ,   39  ,   447 ,   7779    ,   11037   ,   23283   ,0 };
            const unsigned long dim2588Kuo2Init[]       =   {   1   ,   3   ,   1   ,   1   ,   31  ,   59  ,   85  ,   231 ,   209 ,   297 ,   353 ,   2577    ,   6209    ,   6691    ,   2941    ,0 };
            const unsigned long dim2589Kuo2Init[]       =   {   1   ,   1   ,   7   ,   3   ,   19  ,   7   ,   81  ,   41  ,   419 ,   987 ,   1221    ,   2151    ,   5297    ,   1601    ,   1289    ,0 };
            const unsigned long dim2590Kuo2Init[]       =   {   1   ,   3   ,   3   ,   13  ,   25  ,   13  ,   57  ,   199 ,   101 ,   157 ,   1273    ,   159 ,   5417    ,   5153    ,   30529   ,0 };
            const unsigned long dim2591Kuo2Init[]       =   {   1   ,   1   ,   5   ,   1   ,   31  ,   41  ,   67  ,   71  ,   95  ,   929 ,   133 ,   3007    ,   6837    ,   7863    ,   23759   ,0 };
            const unsigned long dim2592Kuo2Init[]       =   {   1   ,   1   ,   5   ,   3   ,   13  ,   59  ,   35  ,   11  ,   187 ,   251 ,   1119    ,   1517    ,   1433    ,   15475   ,   17981   ,0 };
            const unsigned long dim2593Kuo2Init[]       =   {   1   ,   1   ,   3   ,   7   ,   5   ,   37  ,   11  ,   39  ,   111 ,   775 ,   761 ,   2853    ,   2283    ,   203 ,   22563   ,0 };
            const unsigned long dim2594Kuo2Init[]       =   {   1   ,   1   ,   1   ,   11  ,   5   ,   45  ,   119 ,   81  ,   125 ,   481 ,   547 ,   3659    ,   5583    ,   6577    ,   17239   ,0 };
            const unsigned long dim2595Kuo2Init[]       =   {   1   ,   3   ,   3   ,   13  ,   9   ,   25  ,   43  ,   103 ,   283 ,   637 ,   715 ,   3841    ,   5633    ,   503 ,   21933   ,0 };
            const unsigned long dim2596Kuo2Init[]       =   {   1   ,   3   ,   1   ,   13  ,   21  ,   39  ,   1   ,   215 ,   235 ,   171 ,   1377    ,   2131    ,   2907    ,   11311   ,   11337   ,0 };
            const unsigned long dim2597Kuo2Init[]       =   {   1   ,   3   ,   1   ,   3   ,   27  ,   19  ,   73  ,   83  ,   325 ,   451 ,   1779    ,   1665    ,   555 ,   7729    ,   26757   ,0 };
            const unsigned long dim2598Kuo2Init[]       =   {   1   ,   3   ,   5   ,   13  ,   9   ,   29  ,   123 ,   177 ,   289 ,   553 ,   1323    ,   2531    ,   4071    ,   7395    ,   25847   ,0 };
            const unsigned long dim2599Kuo2Init[]       =   {   1   ,   1   ,   7   ,   13  ,   1   ,   1   ,   29  ,   125 ,   253 ,   179 ,   1905    ,   821 ,   5329    ,   6181    ,   18445   ,0 };
            const unsigned long dim2600Kuo2Init[]       =   {   1   ,   1   ,   1   ,   9   ,   15  ,   19  ,   59  ,   227 ,   63  ,   637 ,   1315    ,   2611    ,   1163    ,   6451    ,   8919    ,0 };
            const unsigned long dim2601Kuo2Init[]       =   {   1   ,   1   ,   1   ,   5   ,   23  ,   39  ,   57  ,   143 ,   9   ,   811 ,   1101    ,   609 ,   1281    ,   9457    ,   14999   ,0 };
            const unsigned long dim2602Kuo2Init[]       =   {   1   ,   3   ,   3   ,   7   ,   5   ,   31  ,   21  ,   31  ,   33  ,   271 ,   381 ,   1179    ,   4039    ,   14033   ,   19741   ,0 };
            const unsigned long dim2603Kuo2Init[]       =   {   1   ,   3   ,   7   ,   15  ,   9   ,   41  ,   91  ,   247 ,   59  ,   201 ,   1837    ,   2609    ,   5481    ,   15479   ,   32253   ,0 };
            const unsigned long dim2604Kuo2Init[]       =   {   1   ,   1   ,   3   ,   5   ,   15  ,   53  ,   7   ,   193 ,   451 ,   935 ,   701 ,   177 ,   7943    ,   5687    ,   15725   ,0 };
            const unsigned long dim2605Kuo2Init[]       =   {   1   ,   3   ,   1   ,   13  ,   19  ,   9   ,   33  ,   173 ,   229 ,   993 ,   1637    ,   905 ,   5665    ,   9083    ,   10737   ,0 };
            const unsigned long dim2606Kuo2Init[]       =   {   1   ,   1   ,   7   ,   9   ,   17  ,   33  ,   3   ,   145 ,   213 ,   317 ,   1135    ,   3281    ,   4169    ,   4601    ,   30347   ,0 };
            const unsigned long dim2607Kuo2Init[]       =   {   1   ,   1   ,   1   ,   3   ,   23  ,   1   ,   49  ,   173 ,   351 ,   827 ,   937 ,   2423    ,   6443    ,   6413    ,   20383   ,0 };
            const unsigned long dim2608Kuo2Init[]       =   {   1   ,   3   ,   5   ,   3   ,   21  ,   9   ,   7   ,   29  ,   293 ,   489 ,   339 ,   2403    ,   6303    ,   12283   ,   18047   ,0 };
            const unsigned long dim2609Kuo2Init[]       =   {   1   ,   1   ,   7   ,   11  ,   23  ,   25  ,   49  ,   135 ,   389 ,   821 ,   17  ,   169 ,   4265    ,   5203    ,   3893    ,0 };
            const unsigned long dim2610Kuo2Init[]       =   {   1   ,   1   ,   3   ,   3   ,   27  ,   63  ,   109 ,   165 ,   141 ,   129 ,   1441    ,   1443    ,   3991    ,   1639    ,   31461   ,0 };
            const unsigned long dim2611Kuo2Init[]       =   {   1   ,   3   ,   7   ,   7   ,   25  ,   15  ,   85  ,   217 ,   223 ,   537 ,   1671    ,   2865    ,   4157    ,   8353    ,   18335   ,0 };
            const unsigned long dim2612Kuo2Init[]       =   {   1   ,   3   ,   5   ,   11  ,   19  ,   19  ,   95  ,   159 ,   161 ,   599 ,   1059    ,   2065    ,   7159    ,   11273   ,   17501   ,0 };
            const unsigned long dim2613Kuo2Init[]       =   {   1   ,   3   ,   7   ,   3   ,   15  ,   29  ,   5   ,   97  ,   95  ,   65  ,   1913    ,   1221    ,   329 ,   11601   ,   27033   ,0 };
            const unsigned long dim2614Kuo2Init[]       =   {   1   ,   3   ,   3   ,   1   ,   25  ,   39  ,   103 ,   139 ,   379 ,   547 ,   1415    ,   4027    ,   559 ,   2101    ,   6133    ,0 };
            const unsigned long dim2615Kuo2Init[]       =   {   1   ,   1   ,   7   ,   15  ,   21  ,   47  ,   115 ,   171 ,   17  ,   559 ,   1757    ,   745 ,   2573    ,   13969   ,   4193    ,0 };
            const unsigned long dim2616Kuo2Init[]       =   {   1   ,   3   ,   3   ,   9   ,   19  ,   15  ,   37  ,   131 ,   199 ,   377 ,   67  ,   2455    ,   6807    ,   3365    ,   19697   ,0 };
            const unsigned long dim2617Kuo2Init[]       =   {   1   ,   3   ,   5   ,   9   ,   25  ,   59  ,   9   ,   61  ,   369 ,   285 ,   231 ,   2911    ,   1817    ,   2125    ,   27637   ,0 };
            const unsigned long dim2618Kuo2Init[]       =   {   1   ,   3   ,   1   ,   13  ,   25  ,   55  ,   61  ,   153 ,   507 ,   105 ,   149 ,   1589    ,   7167    ,   12363   ,   10257   ,0 };
            const unsigned long dim2619Kuo2Init[]       =   {   1   ,   3   ,   1   ,   7   ,   9   ,   23  ,   109 ,   29  ,   167 ,   463 ,   101 ,   979 ,   3671    ,   10165   ,   12553   ,0 };
            const unsigned long dim2620Kuo2Init[]       =   {   1   ,   3   ,   3   ,   9   ,   21  ,   49  ,   103 ,   217 ,   269 ,   761 ,   915 ,   3529    ,   125 ,   429 ,   24131   ,0 };
            const unsigned long dim2621Kuo2Init[]       =   {   1   ,   3   ,   5   ,   15  ,   27  ,   61  ,   73  ,   13  ,   339 ,   1023    ,   967 ,   3883    ,   5515    ,   5893    ,   26183   ,0 };
            const unsigned long dim2622Kuo2Init[]       =   {   1   ,   1   ,   1   ,   5   ,   5   ,   39  ,   87  ,   183 ,   141 ,   147 ,   779 ,   1733    ,   6137    ,   5985    ,   1335    ,0 };
            const unsigned long dim2623Kuo2Init[]       =   {   1   ,   3   ,   5   ,   11  ,   23  ,   9   ,   61  ,   67  ,   235 ,   823 ,   1129    ,   401 ,   3049    ,   5593    ,   5193    ,0 };
            const unsigned long dim2624Kuo2Init[]       =   {   1   ,   1   ,   3   ,   3   ,   27  ,   39  ,   27  ,   111 ,   81  ,   27  ,   283 ,   4081    ,   6511    ,   6785    ,   20953   ,0 };
            const unsigned long dim2625Kuo2Init[]       =   {   1   ,   3   ,   7   ,   7   ,   17  ,   49  ,   93  ,   85  ,   327 ,   287 ,   1473    ,   3769    ,   2229    ,   12221   ,   27355   ,0 };
            const unsigned long dim2626Kuo2Init[]       =   {   1   ,   3   ,   1   ,   3   ,   15  ,   15  ,   71  ,   105 ,   139 ,   137 ,   1139    ,   859 ,   7923    ,   581 ,   7697    ,0 };
            const unsigned long dim2627Kuo2Init[]       =   {   1   ,   3   ,   7   ,   15  ,   29  ,   45  ,   17  ,   159 ,   233 ,   345 ,   1827    ,   2261    ,   6773    ,   5435    ,   17961   ,0 };
            const unsigned long dim2628Kuo2Init[]       =   {   1   ,   3   ,   5   ,   13  ,   21  ,   15  ,   33  ,   229 ,   71  ,   709 ,   1757    ,   1479    ,   7289    ,   15629   ,   24041   ,0 };
            const unsigned long dim2629Kuo2Init[]       =   {   1   ,   3   ,   3   ,   1   ,   27  ,   9   ,   103 ,   115 ,   137 ,   911 ,   603 ,   3615    ,   7671    ,   2439    ,   10041   ,0 };
            const unsigned long dim2630Kuo2Init[]       =   {   1   ,   1   ,   5   ,   5   ,   21  ,   59  ,   67  ,   193 ,   249 ,   781 ,   761 ,   403 ,   6299    ,   527 ,   8533    ,0 };
            const unsigned long dim2631Kuo2Init[]       =   {   1   ,   3   ,   3   ,   3   ,   5   ,   61  ,   71  ,   183 ,   151 ,   801 ,   1645    ,   2789    ,   5911    ,   11681   ,   5907    ,0 };
            const unsigned long dim2632Kuo2Init[]       =   {   1   ,   1   ,   5   ,   7   ,   9   ,   19  ,   53  ,   239 ,   259 ,   273 ,   679 ,   2101    ,   1525    ,   2139    ,   30353   ,0 };
            const unsigned long dim2633Kuo2Init[]       =   {   1   ,   1   ,   3   ,   1   ,   31  ,   25  ,   37  ,   225 ,   287 ,   31  ,   433 ,   2407    ,   1717    ,   6669    ,   10743   ,0 };
            const unsigned long dim2634Kuo2Init[]       =   {   1   ,   3   ,   1   ,   13  ,   31  ,   13  ,   89  ,   225 ,   173 ,   723 ,   651 ,   3315    ,   6457    ,   9387    ,   7759    ,0 };
            const unsigned long dim2635Kuo2Init[]       =   {   1   ,   1   ,   5   ,   7   ,   3   ,   25  ,   45  ,   79  ,   11  ,   843 ,   2031    ,   2827    ,   5029    ,   1359    ,   23489   ,0 };
            const unsigned long dim2636Kuo2Init[]       =   {   1   ,   3   ,   5   ,   9   ,   7   ,   1   ,   53  ,   135 ,   327 ,   13  ,   1051    ,   2911    ,   5553    ,   6557    ,   16415   ,0 };
            const unsigned long dim2637Kuo2Init[]       =   {   1   ,   1   ,   1   ,   5   ,   19  ,   19  ,   57  ,   147 ,   43  ,   503 ,   1899    ,   1709    ,   5589    ,   4251    ,   10871   ,0 };
            const unsigned long dim2638Kuo2Init[]       =   {   1   ,   1   ,   5   ,   7   ,   1   ,   41  ,   83  ,   51  ,   89  ,   1015    ,   1441    ,   1735    ,   7217    ,   11431   ,   29623   ,0 };
            const unsigned long dim2639Kuo2Init[]       =   {   1   ,   3   ,   7   ,   7   ,   29  ,   45  ,   73  ,   233 ,   399 ,   553 ,   961 ,   2053    ,   3537    ,   4015    ,   19963   ,0 };
            const unsigned long dim2640Kuo2Init[]       =   {   1   ,   1   ,   5   ,   3   ,   21  ,   5   ,   1   ,   105 ,   341 ,   355 ,   849 ,   2235    ,   2801    ,   13749   ,   17933   ,0 };
            const unsigned long dim2641Kuo2Init[]       =   {   1   ,   1   ,   7   ,   1   ,   25  ,   57  ,   29  ,   233 ,   485 ,   807 ,   1005    ,   1933    ,   3897    ,   1267    ,   30807   ,0 };
            const unsigned long dim2642Kuo2Init[]       =   {   1   ,   1   ,   7   ,   15  ,   31  ,   53  ,   49  ,   31  ,   325 ,   335 ,   671 ,   1183    ,   2955    ,   3   ,   31739   ,0 };
            const unsigned long dim2643Kuo2Init[]       =   {   1   ,   1   ,   5   ,   3   ,   17  ,   41  ,   19  ,   73  ,   215 ,   227 ,   1251    ,   907 ,   3303    ,   5343    ,   30385   ,0 };
            const unsigned long dim2644Kuo2Init[]       =   {   1   ,   1   ,   1   ,   5   ,   15  ,   61  ,   123 ,   61  ,   27  ,   759 ,   645 ,   2571    ,   411 ,   13925   ,   29905   ,0 };
            const unsigned long dim2645Kuo2Init[]       =   {   1   ,   3   ,   1   ,   5   ,   7   ,   55  ,   81  ,   241 ,   497 ,   1005    ,   1593    ,   2479    ,   25  ,   10627   ,   14043   ,0 };
            const unsigned long dim2646Kuo2Init[]       =   {   1   ,   3   ,   3   ,   9   ,   19  ,   33  ,   87  ,   63  ,   441 ,   773 ,   1367    ,   2945    ,   4897    ,   14209   ,   6667    ,0 };
            const unsigned long dim2647Kuo2Init[]       =   {   1   ,   1   ,   7   ,   15  ,   31  ,   59  ,   121 ,   153 ,   111 ,   329 ,   833 ,   2031    ,   4063    ,   13775   ,   9033    ,0 };
            const unsigned long dim2648Kuo2Init[]       =   {   1   ,   3   ,   1   ,   11  ,   3   ,   33  ,   47  ,   65  ,   329 ,   149 ,   1509    ,   3617    ,   1381    ,   9003    ,   4953    ,0 };
            const unsigned long dim2649Kuo2Init[]       =   {   1   ,   3   ,   1   ,   3   ,   7   ,   59  ,   39  ,   173 ,   157 ,   741 ,   343 ,   1039    ,   1695    ,   12077   ,   13203   ,0 };
            const unsigned long dim2650Kuo2Init[]       =   {   1   ,   3   ,   7   ,   1   ,   17  ,   21  ,   21  ,   5   ,   65  ,   163 ,   711 ,   2251    ,   4761    ,   11895   ,   15571   ,0 };
            const unsigned long dim2651Kuo2Init[]       =   {   1   ,   3   ,   7   ,   7   ,   7   ,   59  ,   31  ,   227 ,   283 ,   223 ,   163 ,   2639    ,   3037    ,   14423   ,   21507   ,0 };
            const unsigned long dim2652Kuo2Init[]       =   {   1   ,   1   ,   1   ,   5   ,   9   ,   37  ,   5   ,   149 ,   395 ,   491 ,   711 ,   447 ,   5657    ,   443 ,   7183    ,0 };
            const unsigned long dim2653Kuo2Init[]       =   {   1   ,   1   ,   5   ,   9   ,   17  ,   61  ,   31  ,   229 ,   207 ,   755 ,   199 ,   1585    ,   2771    ,   1499    ,   4911    ,0 };
            const unsigned long dim2654Kuo2Init[]       =   {   1   ,   3   ,   5   ,   7   ,   1   ,   1   ,   53  ,   185 ,   407 ,   57  ,   525 ,   543 ,   5009    ,   1825    ,   8583    ,0 };
            const unsigned long dim2655Kuo2Init[]       =   {   1   ,   1   ,   1   ,   3   ,   13  ,   49  ,   85  ,   175 ,   261 ,   653 ,   665 ,   3173    ,   4709    ,   11547   ,   14881   ,0 };
            const unsigned long dim2656Kuo2Init[]       =   {   1   ,   1   ,   1   ,   3   ,   21  ,   47  ,   21  ,   33  ,   265 ,   171 ,   765 ,   3871    ,   3577    ,   15135   ,   13483   ,0 };
            const unsigned long dim2657Kuo2Init[]       =   {   1   ,   3   ,   1   ,   3   ,   29  ,   5   ,   109 ,   3   ,   249 ,   379 ,   1259    ,   677 ,   1177    ,   13061   ,   31985   ,0 };
            const unsigned long dim2658Kuo2Init[]       =   {   1   ,   3   ,   7   ,   3   ,   15  ,   25  ,   123 ,   93  ,   193 ,   419 ,   1173    ,   2493    ,   2937    ,   14909   ,   3147    ,0 };
            const unsigned long dim2659Kuo2Init[]       =   {   1   ,   3   ,   3   ,   3   ,   9   ,   59  ,   55  ,   95  ,   93  ,   319 ,   1281    ,   587 ,   8065    ,   15697   ,   32101   ,0 };
            const unsigned long dim2660Kuo2Init[]       =   {   1   ,   3   ,   5   ,   7   ,   25  ,   61  ,   105 ,   199 ,   101 ,   255 ,   849 ,   377 ,   3245    ,   8139    ,   8383    ,0 };
            const unsigned long dim2661Kuo2Init[]       =   {   1   ,   1   ,   5   ,   9   ,   3   ,   47  ,   79  ,   155 ,   15  ,   661 ,   1417    ,   87  ,   4225    ,   13241   ,   15971   ,0 };
            const unsigned long dim2662Kuo2Init[]       =   {   1   ,   3   ,   3   ,   1   ,   13  ,   39  ,   101 ,   79  ,   271 ,   21  ,   1983    ,   2821    ,   4177    ,   14081   ,   31219   ,0 };
            const unsigned long dim2663Kuo2Init[]       =   {   1   ,   1   ,   1   ,   7   ,   15  ,   51  ,   123 ,   17  ,   29  ,   123 ,   1131    ,   147 ,   4191    ,   6729    ,   19359   ,0 };
            const unsigned long dim2664Kuo2Init[]       =   {   1   ,   1   ,   7   ,   7   ,   17  ,   39  ,   47  ,   215 ,   219 ,   643 ,   1877    ,   3965    ,   5799    ,   16131   ,   24675   ,0 };
            const unsigned long dim2665Kuo2Init[]       =   {   1   ,   3   ,   5   ,   7   ,   11  ,   53  ,   57  ,   133 ,   289 ,   33  ,   1631    ,   2037    ,   6699    ,   14153   ,   25309   ,0 };
            const unsigned long dim2666Kuo2Init[]       =   {   1   ,   3   ,   7   ,   7   ,   9   ,   17  ,   125 ,   123 ,   91  ,   15  ,   1279    ,   297 ,   2583    ,   11141   ,   7257    ,0 };
            const unsigned long dim2667Kuo2Init[]       =   {   1   ,   1   ,   7   ,   9   ,   13  ,   9   ,   83  ,   117 ,   365 ,   533 ,   343 ,   121 ,   7553    ,   14417   ,   23239   ,0 };
            const unsigned long dim2668Kuo2Init[]       =   {   1   ,   3   ,   3   ,   9   ,   5   ,   27  ,   97  ,   229 ,   321 ,   123 ,   1669    ,   621 ,   4937    ,   683 ,   31061   ,0 };
            const unsigned long dim2669Kuo2Init[]       =   {   1   ,   3   ,   1   ,   3   ,   1   ,   23  ,   63  ,   21  ,   271 ,   695 ,   601 ,   113 ,   6331    ,   8837    ,   16721   ,0 };
            const unsigned long dim2670Kuo2Init[]       =   {   1   ,   3   ,   7   ,   13  ,   15  ,   41  ,   53  ,   115 ,   225 ,   5   ,   1315    ,   2285    ,   3815    ,   5949    ,   4263    ,0 };
            const unsigned long dim2671Kuo2Init[]       =   {   1   ,   1   ,   1   ,   9   ,   21  ,   41  ,   25  ,   47  ,   9   ,   423 ,   1171    ,   2515    ,   3461    ,   13273   ,   13059   ,0 };
            const unsigned long dim2672Kuo2Init[]       =   {   1   ,   3   ,   5   ,   15  ,   5   ,   29  ,   111 ,   15  ,   505 ,   347 ,   347 ,   2521    ,   4583    ,   5729    ,   11645   ,0 };
            const unsigned long dim2673Kuo2Init[]       =   {   1   ,   3   ,   3   ,   1   ,   9   ,   15  ,   57  ,   211 ,   163 ,   749 ,   1549    ,   1015    ,   441 ,   16025   ,   28943   ,0 };
            const unsigned long dim2674Kuo2Init[]       =   {   1   ,   1   ,   3   ,   15  ,   15  ,   27  ,   111 ,   79  ,   131 ,   831 ,   2045    ,   1937    ,   6969    ,   11223   ,   22711   ,0 };
            const unsigned long dim2675Kuo2Init[]       =   {   1   ,   3   ,   5   ,   1   ,   3   ,   37  ,   37  ,   221 ,   143 ,   955 ,   1387    ,   3483    ,   5845    ,   13411   ,   16225   ,0 };
            const unsigned long dim2676Kuo2Init[]       =   {   1   ,   3   ,   1   ,   3   ,   7   ,   13  ,   75  ,   57  ,   357 ,   109 ,   241 ,   165 ,   1845    ,   213 ,   8337    ,0 };
            const unsigned long dim2677Kuo2Init[]       =   {   1   ,   3   ,   1   ,   15  ,   5   ,   41  ,   109 ,   129 ,   91  ,   459 ,   1237    ,   467 ,   8015    ,   14499   ,   19897   ,0 };
            const unsigned long dim2678Kuo2Init[]       =   {   1   ,   1   ,   5   ,   7   ,   23  ,   45  ,   73  ,   135 ,   161 ,   921 ,   127 ,   3723    ,   1795    ,   4047    ,   16607   ,0 };
            const unsigned long dim2679Kuo2Init[]       =   {   1   ,   3   ,   1   ,   9   ,   21  ,   17  ,   13  ,   133 ,   251 ,   927 ,   1551    ,   3087    ,   317 ,   2989    ,   30481   ,0 };
            const unsigned long dim2680Kuo2Init[]       =   {   1   ,   1   ,   3   ,   7   ,   15  ,   51  ,   17  ,   77  ,   345 ,   833 ,   983 ,   2855    ,   383 ,   10873   ,   20019   ,0 };
            const unsigned long dim2681Kuo2Init[]       =   {   1   ,   1   ,   7   ,   7   ,   3   ,   9   ,   65  ,   115 ,   449 ,   375 ,   27  ,   2053    ,   175 ,   6793    ,   9761    ,0 };
            const unsigned long dim2682Kuo2Init[]       =   {   1   ,   3   ,   5   ,   7   ,   17  ,   7   ,   51  ,   207 ,   389 ,   995 ,   1663    ,   3697    ,   5539    ,   13495   ,   4053    ,0 };
            const unsigned long dim2683Kuo2Init[]       =   {   1   ,   1   ,   1   ,   13  ,   5   ,   25  ,   125 ,   139 ,   123 ,   433 ,   1323    ,   273 ,   6061    ,   7145    ,   13929   ,0 };
            const unsigned long dim2684Kuo2Init[]       =   {   1   ,   3   ,   7   ,   3   ,   3   ,   1   ,   65  ,   217 ,   203 ,   887 ,   121 ,   961 ,   2843    ,   16171   ,   8433    ,0 };
            const unsigned long dim2685Kuo2Init[]       =   {   1   ,   1   ,   3   ,   1   ,   21  ,   25  ,   93  ,   237 ,   203 ,   721 ,   1135    ,   751 ,   7193    ,   7453    ,   29593   ,0 };
            const unsigned long dim2686Kuo2Init[]       =   {   1   ,   1   ,   5   ,   13  ,   27  ,   41  ,   125 ,   31  ,   145 ,   69  ,   1865    ,   471 ,   6031    ,   6589    ,   16185   ,0 };
            const unsigned long dim2687Kuo2Init[]       =   {   1   ,   1   ,   5   ,   1   ,   13  ,   39  ,   15  ,   251 ,   125 ,   781 ,   1635    ,   231 ,   7503    ,   12647   ,   9493    ,0 };
            const unsigned long dim2688Kuo2Init[]       =   {   1   ,   1   ,   3   ,   11  ,   31  ,   35  ,   39  ,   49  ,   279 ,   281 ,   285 ,   1707    ,   607 ,   3405    ,   5803    ,0 };
            const unsigned long dim2689Kuo2Init[]       =   {   1   ,   3   ,   5   ,   7   ,   29  ,   61  ,   43  ,   15  ,   413 ,   759 ,   159 ,   2027    ,   4065    ,   9501    ,   31597   ,0 };
            const unsigned long dim2690Kuo2Init[]       =   {   1   ,   1   ,   5   ,   3   ,   21  ,   19  ,   25  ,   31  ,   21  ,   139 ,   1273    ,   871 ,   423 ,   13737   ,   1633    ,0 };
            const unsigned long dim2691Kuo2Init[]       =   {   1   ,   1   ,   5   ,   11  ,   9   ,   15  ,   15  ,   247 ,   459 ,   49  ,   2047    ,   289 ,   7457    ,   13367   ,   3343    ,0 };
            const unsigned long dim2692Kuo2Init[]       =   {   1   ,   3   ,   7   ,   1   ,   5   ,   43  ,   49  ,   115 ,   201 ,   131 ,   297 ,   2069    ,   801 ,   7599    ,   675 ,0 };
            const unsigned long dim2693Kuo2Init[]       =   {   1   ,   1   ,   3   ,   15  ,   11  ,   19  ,   41  ,   135 ,   107 ,   325 ,   431 ,   2459    ,   7629    ,   3407    ,   19635   ,0 };
            const unsigned long dim2694Kuo2Init[]       =   {   1   ,   1   ,   7   ,   15  ,   1   ,   45  ,   89  ,   249 ,   463 ,   319 ,   1577    ,   549 ,   3623    ,   7053    ,   23037   ,0 };
            const unsigned long dim2695Kuo2Init[]       =   {   1   ,   3   ,   1   ,   11  ,   27  ,   23  ,   123 ,   63  ,   219 ,   77  ,   179 ,   3283    ,   2887    ,   14023   ,   28277   ,0 };
            const unsigned long dim2696Kuo2Init[]       =   {   1   ,   3   ,   1   ,   9   ,   29  ,   47  ,   39  ,   61  ,   413 ,   327 ,   1529    ,   593 ,   7779    ,   14857   ,   14165   ,0 };
            const unsigned long dim2697Kuo2Init[]       =   {   1   ,   3   ,   5   ,   3   ,   29  ,   15  ,   73  ,   231 ,   141 ,   435 ,   1835    ,   1023    ,   289 ,   1859    ,   1069    ,0 };
            const unsigned long dim2698Kuo2Init[]       =   {   1   ,   3   ,   3   ,   3   ,   9   ,   35  ,   1   ,   153 ,   111 ,   521 ,   461 ,   3001    ,   5401    ,   9529    ,   19845   ,0 };
            const unsigned long dim2699Kuo2Init[]       =   {   1   ,   1   ,   1   ,   15  ,   17  ,   43  ,   49  ,   227 ,   87  ,   199 ,   179 ,   1619    ,   6093    ,   15695   ,   30739   ,0 };
            const unsigned long dim2700Kuo2Init[]       =   {   1   ,   3   ,   3   ,   9   ,   17  ,   53  ,   67  ,   199 ,   263 ,   931 ,   413 ,   1019    ,   679 ,   15483   ,   22211   ,0 };
            const unsigned long dim2701Kuo2Init[]       =   {   1   ,   1   ,   7   ,   3   ,   19  ,   7   ,   55  ,   111 ,   499 ,   587 ,   1685    ,   1007    ,   5147    ,   6097    ,   31801   ,0 };
            const unsigned long dim2702Kuo2Init[]       =   {   1   ,   3   ,   7   ,   5   ,   17  ,   15  ,   41  ,   167 ,   495 ,   439 ,   1161    ,   3737    ,   5957    ,   4151    ,   26597   ,0 };
            const unsigned long dim2703Kuo2Init[]       =   {   1   ,   3   ,   5   ,   13  ,   17  ,   35  ,   103 ,   205 ,   83  ,   613 ,   1541    ,   2825    ,   2773    ,   14487   ,   177 ,0 };
            const unsigned long dim2704Kuo2Init[]       =   {   1   ,   3   ,   1   ,   5   ,   11  ,   31  ,   1   ,   93  ,   87  ,   917 ,   1209    ,   1489    ,   4063    ,   7811    ,   31291   ,0 };
            const unsigned long dim2705Kuo2Init[]       =   {   1   ,   1   ,   5   ,   9   ,   27  ,   25  ,   15  ,   121 ,   51  ,   207 ,   567 ,   2091    ,   6177    ,   121 ,   15305   ,0 };
            const unsigned long dim2706Kuo2Init[]       =   {   1   ,   3   ,   7   ,   13  ,   17  ,   27  ,   13  ,   31  ,   443 ,   487 ,   863 ,   3253    ,   4961    ,   2695    ,   6427    ,0 };
            const unsigned long dim2707Kuo2Init[]       =   {   1   ,   3   ,   7   ,   15  ,   23  ,   23  ,   85  ,   229 ,   123 ,   151 ,   519 ,   2581    ,   4073    ,   5915    ,   23715   ,0 };
            const unsigned long dim2708Kuo2Init[]       =   {   1   ,   3   ,   5   ,   9   ,   19  ,   41  ,   33  ,   71  ,   511 ,   683 ,   1387    ,   2975    ,   5069    ,   4327    ,   25881   ,0 };
            const unsigned long dim2709Kuo2Init[]       =   {   1   ,   1   ,   7   ,   3   ,   9   ,   25  ,   29  ,   73  ,   55  ,   323 ,   483 ,   1509    ,   4895    ,   5151    ,   13095   ,0 };
            const unsigned long dim2710Kuo2Init[]       =   {   1   ,   3   ,   7   ,   3   ,   7   ,   19  ,   95  ,   163 ,   239 ,   721 ,   747 ,   3965    ,   6943    ,   11771   ,   5991    ,0 };
            const unsigned long dim2711Kuo2Init[]       =   {   1   ,   3   ,   1   ,   9   ,   27  ,   11  ,   5   ,   9   ,   291 ,   987 ,   313 ,   129 ,   4147    ,   1005    ,   25719   ,0 };
            const unsigned long dim2712Kuo2Init[]       =   {   1   ,   1   ,   1   ,   5   ,   7   ,   25  ,   85  ,   197 ,   177 ,   759 ,   1083    ,   2615    ,   3647    ,   11873   ,   17035   ,0 };
            const unsigned long dim2713Kuo2Init[]       =   {   1   ,   3   ,   1   ,   13  ,   13  ,   11  ,   29  ,   45  ,   103 ,   319 ,   1325    ,   2987    ,   7253    ,   3967    ,   32547   ,0 };
            const unsigned long dim2714Kuo2Init[]       =   {   1   ,   3   ,   5   ,   9   ,   11  ,   27  ,   103 ,   75  ,   113 ,   741 ,   175 ,   2715    ,   1595    ,   7899    ,   261 ,0 };
            const unsigned long dim2715Kuo2Init[]       =   {   1   ,   1   ,   3   ,   5   ,   3   ,   39  ,   113 ,   19  ,   301 ,   19  ,   1369    ,   2759    ,   467 ,   10387   ,   15763   ,0 };
            const unsigned long dim2716Kuo2Init[]       =   {   1   ,   1   ,   5   ,   5   ,   5   ,   23  ,   47  ,   21  ,   69  ,   57  ,   1183    ,   2725    ,   2329    ,   4829    ,   22769   ,0 };
            const unsigned long dim2717Kuo2Init[]       =   {   1   ,   3   ,   3   ,   13  ,   1   ,   39  ,   111 ,   11  ,   133 ,   143 ,   465 ,   1551    ,   2969    ,   2353    ,   9381    ,0 };
            const unsigned long dim2718Kuo2Init[]       =   {   1   ,   3   ,   7   ,   3   ,   29  ,   19  ,   75  ,   97  ,   139 ,   325 ,   189 ,   477 ,   5687    ,   11105   ,   17325   ,0 };
            const unsigned long dim2719Kuo2Init[]       =   {   1   ,   1   ,   3   ,   15  ,   25  ,   51  ,   53  ,   177 ,   41  ,   677 ,   233 ,   467 ,   4977    ,   8055    ,   30257   ,0 };
            const unsigned long dim2720Kuo2Init[]       =   {   1   ,   3   ,   7   ,   5   ,   29  ,   49  ,   23  ,   247 ,   167 ,   229 ,   719 ,   1583    ,   3451    ,   11901   ,   10991   ,0 };
            const unsigned long dim2721Kuo2Init[]       =   {   1   ,   1   ,   1   ,   13  ,   5   ,   39  ,   35  ,   247 ,   185 ,   409 ,   1481    ,   3993    ,   2023    ,   9251    ,   1253    ,0 };
            const unsigned long dim2722Kuo2Init[]       =   {   1   ,   1   ,   3   ,   13  ,   27  ,   35  ,   49  ,   251 ,   359 ,   161 ,   1535    ,   3277    ,   1483    ,   8045    ,   22373   ,0 };
            const unsigned long dim2723Kuo2Init[]       =   {   1   ,   1   ,   1   ,   7   ,   3   ,   63  ,   103 ,   7   ,   141 ,   305 ,   1803    ,   3429    ,   6819    ,   3389    ,   12915   ,0 };
            const unsigned long dim2724Kuo2Init[]       =   {   1   ,   3   ,   3   ,   15  ,   15  ,   7   ,   19  ,   121 ,   421 ,   633 ,   1619    ,   1225    ,   4061    ,   13649   ,   25537   ,0 };
            const unsigned long dim2725Kuo2Init[]       =   {   1   ,   3   ,   7   ,   13  ,   5   ,   49  ,   117 ,   97  ,   85  ,   13  ,   55  ,   2991    ,   6215    ,   15087   ,   29311   ,0 };
            const unsigned long dim2726Kuo2Init[]       =   {   1   ,   1   ,   7   ,   13  ,   29  ,   37  ,   23  ,   75  ,   11  ,   899 ,   1519    ,   2151    ,   235 ,   7121    ,   20881   ,0 };
            const unsigned long dim2727Kuo2Init[]       =   {   1   ,   3   ,   5   ,   13  ,   7   ,   9   ,   97  ,   57  ,   307 ,   791 ,   745 ,   881 ,   6627    ,   1639    ,   22029   ,0 };
            const unsigned long dim2728Kuo2Init[]       =   {   1   ,   3   ,   7   ,   11  ,   31  ,   25  ,   67  ,   153 ,   397 ,   115 ,   1951    ,   553 ,   2511    ,   853 ,   22771   ,0 };
            const unsigned long dim2729Kuo2Init[]       =   {   1   ,   1   ,   7   ,   13  ,   13  ,   33  ,   3   ,   49  ,   281 ,   847 ,   395 ,   2671    ,   7173    ,   11035   ,   2809    ,0 };
            const unsigned long dim2730Kuo2Init[]       =   {   1   ,   3   ,   7   ,   11  ,   7   ,   13  ,   21  ,   93  ,   279 ,   453 ,   1793    ,   1183    ,   4071    ,   11357   ,   15871   ,0 };
            const unsigned long dim2731Kuo2Init[]       =   {   1   ,   3   ,   5   ,   1   ,   31  ,   27  ,   87  ,   115 ,   373 ,   519 ,   1691    ,   3339    ,   5319    ,   10765   ,   3497    ,0 };
            const unsigned long dim2732Kuo2Init[]       =   {   1   ,   3   ,   5   ,   11  ,   27  ,   13  ,   73  ,   111 ,   349 ,   889 ,   1447    ,   1733    ,   7701    ,   9263    ,   27267   ,0 };
            const unsigned long dim2733Kuo2Init[]       =   {   1   ,   1   ,   1   ,   13  ,   15  ,   25  ,   31  ,   187 ,   227 ,   119 ,   955 ,   1233    ,   2749    ,   8477    ,   12031   ,0 };
            const unsigned long dim2734Kuo2Init[]       =   {   1   ,   3   ,   1   ,   7   ,   19  ,   13  ,   89  ,   151 ,   61  ,   267 ,   1833    ,   3023    ,   8177    ,   12863   ,   13851   ,0 };
            const unsigned long dim2735Kuo2Init[]       =   {   1   ,   1   ,   7   ,   11  ,   11  ,   19  ,   65  ,   223 ,   175 ,   893 ,   1515    ,   2323    ,   6725    ,   747 ,   10299   ,0 };
            const unsigned long dim2736Kuo2Init[]       =   {   1   ,   1   ,   1   ,   15  ,   3   ,   17  ,   109 ,   119 ,   371 ,   193 ,   1289    ,   219 ,   463 ,   9043    ,   31471   ,0 };
            const unsigned long dim2737Kuo2Init[]       =   {   1   ,   1   ,   1   ,   3   ,   9   ,   9   ,   71  ,   185 ,   81  ,   849 ,   917 ,   2495    ,   4065    ,   14161   ,   14287   ,0 };
            const unsigned long dim2738Kuo2Init[]       =   {   1   ,   1   ,   5   ,   3   ,   5   ,   59  ,   111 ,   133 ,   351 ,   247 ,   1035    ,   3277    ,   3865    ,   5879    ,   16537   ,0 };
            const unsigned long dim2739Kuo2Init[]       =   {   1   ,   3   ,   3   ,   11  ,   23  ,   47  ,   3   ,   39  ,   343 ,   71  ,   109 ,   4005    ,   7453    ,   8327    ,   9797    ,0 };
            const unsigned long dim2740Kuo2Init[]       =   {   1   ,   1   ,   1   ,   5   ,   21  ,   31  ,   53  ,   33  ,   115 ,   287 ,   431 ,   2085    ,   5275    ,   5253    ,   18137   ,0 };
            const unsigned long dim2741Kuo2Init[]       =   {   1   ,   3   ,   1   ,   15  ,   21  ,   43  ,   95  ,   143 ,   103 ,   611 ,   583 ,   41  ,   483 ,   13973   ,   30231   ,0 };
            const unsigned long dim2742Kuo2Init[]       =   {   1   ,   1   ,   1   ,   9   ,   9   ,   33  ,   49  ,   35  ,   329 ,   887 ,   1593    ,   2301    ,   229 ,   10291   ,   21153   ,0 };
            const unsigned long dim2743Kuo2Init[]       =   {   1   ,   3   ,   5   ,   9   ,   25  ,   55  ,   55  ,   207 ,   381 ,   747 ,   997 ,   3105    ,   377 ,   15327   ,   24367   ,0 };
            const unsigned long dim2744Kuo2Init[]       =   {   1   ,   3   ,   7   ,   15  ,   23  ,   25  ,   101 ,   225 ,   345 ,   579 ,   975 ,   2383    ,   7833    ,   12727   ,   15991   ,0 };
            const unsigned long dim2745Kuo2Init[]       =   {   1   ,   3   ,   3   ,   1   ,   23  ,   13  ,   51  ,   45  ,   471 ,   271 ,   151 ,   1407    ,   2079    ,   3521    ,   6469    ,0 };
            const unsigned long dim2746Kuo2Init[]       =   {   1   ,   3   ,   1   ,   9   ,   3   ,   11  ,   117 ,   115 ,   89  ,   871 ,   1989    ,   1341    ,   2415    ,   12921   ,   3531    ,0 };
            const unsigned long dim2747Kuo2Init[]       =   {   1   ,   3   ,   3   ,   1   ,   3   ,   49  ,   41  ,   215 ,   277 ,   155 ,   1185    ,   2837    ,   4361    ,   14495   ,   19877   ,0 };
            const unsigned long dim2748Kuo2Init[]       =   {   1   ,   1   ,   1   ,   9   ,   9   ,   19  ,   31  ,   211 ,   179 ,   367 ,   1597    ,   561 ,   5397    ,   7525    ,   2963    ,0 };
            const unsigned long dim2749Kuo2Init[]       =   {   1   ,   3   ,   7   ,   11  ,   25  ,   45  ,   29  ,   151 ,   107 ,   13  ,   1461    ,   235 ,   6939    ,   6223    ,   13111   ,0 };
            const unsigned long dim2750Kuo2Init[]       =   {   1   ,   3   ,   7   ,   15  ,   17  ,   11  ,   43  ,   113 ,   65  ,   321 ,   1393    ,   1131    ,   2481    ,   13139   ,   14647   ,0 };
            const unsigned long dim2751Kuo2Init[]       =   {   1   ,   3   ,   7   ,   3   ,   15  ,   23  ,   45  ,   229 ,   175 ,   857 ,   1303    ,   783 ,   3929    ,   855 ,   13581   ,0 };
            const unsigned long dim2752Kuo2Init[]       =   {   1   ,   3   ,   7   ,   3   ,   7   ,   55  ,   99  ,   85  ,   415 ,   667 ,   603 ,   593 ,   427 ,   10725   ,   6973    ,0 };
            const unsigned long dim2753Kuo2Init[]       =   {   1   ,   3   ,   7   ,   3   ,   1   ,   23  ,   93  ,   177 ,   59  ,   601 ,   1947    ,   3253    ,   639 ,   10007   ,   1441    ,0 };
            const unsigned long dim2754Kuo2Init[]       =   {   1   ,   3   ,   5   ,   11  ,   13  ,   37  ,   79  ,   19  ,   395 ,   161 ,   945 ,   969 ,   1579    ,   13537   ,   6347    ,0 };
            const unsigned long dim2755Kuo2Init[]       =   {   1   ,   3   ,   1   ,   15  ,   3   ,   61  ,   101 ,   3   ,   289 ,   707 ,   61  ,   1405    ,   811 ,   6727    ,   22339   ,0 };
            const unsigned long dim2756Kuo2Init[]       =   {   1   ,   3   ,   5   ,   11  ,   3   ,   45  ,   19  ,   125 ,   503 ,   153 ,   1439    ,   2105    ,   7483    ,   1835    ,   22841   ,0 };
            const unsigned long dim2757Kuo2Init[]       =   {   1   ,   1   ,   7   ,   3   ,   23  ,   49  ,   49  ,   227 ,   413 ,   57  ,   1829    ,   37  ,   407 ,   4203    ,   20575   ,0 };
            const unsigned long dim2758Kuo2Init[]       =   {   1   ,   3   ,   5   ,   11  ,   3   ,   43  ,   93  ,   85  ,   37  ,   471 ,   1971    ,   1559    ,   5151    ,   11641   ,   13011   ,0 };
            const unsigned long dim2759Kuo2Init[]       =   {   1   ,   1   ,   5   ,   15  ,   5   ,   13  ,   17  ,   209 ,   125 ,   581 ,   57  ,   43  ,   7089    ,   3367    ,   16073   ,0 };
            const unsigned long dim2760Kuo2Init[]       =   {   1   ,   3   ,   3   ,   9   ,   29  ,   51  ,   93  ,   223 ,   73  ,   917 ,   1657    ,   2811    ,   1593    ,   8595    ,   23347   ,0 };
            const unsigned long dim2761Kuo2Init[]       =   {   1   ,   3   ,   1   ,   15  ,   25  ,   21  ,   111 ,   207 ,   61  ,   815 ,   805 ,   1439    ,   5745    ,   5065    ,   9877    ,0 };
            const unsigned long dim2762Kuo2Init[]       =   {   1   ,   3   ,   3   ,   15  ,   7   ,   1   ,   113 ,   225 ,   467 ,   341 ,   1145    ,   3839    ,   1127    ,   5085    ,   885 ,0 };
            const unsigned long dim2763Kuo2Init[]       =   {   1   ,   1   ,   7   ,   3   ,   13  ,   27  ,   103 ,   17  ,   175 ,   351 ,   1321    ,   2093    ,   3407    ,   3387    ,   1663    ,0 };
            const unsigned long dim2764Kuo2Init[]       =   {   1   ,   3   ,   5   ,   13  ,   5   ,   61  ,   31  ,   223 ,   249 ,   797 ,   1541    ,   3985    ,   2165    ,   8741    ,   561 ,0 };
            const unsigned long dim2765Kuo2Init[]       =   {   1   ,   1   ,   1   ,   7   ,   21  ,   51  ,   119 ,   9   ,   331 ,   305 ,   1279    ,   3331    ,   2769    ,   673 ,   26293   ,0 };
            const unsigned long dim2766Kuo2Init[]       =   {   1   ,   1   ,   3   ,   9   ,   7   ,   61  ,   65  ,   157 ,   7   ,   259 ,   871 ,   557 ,   7917    ,   12839   ,   18521   ,0 };
            const unsigned long dim2767Kuo2Init[]       =   {   1   ,   3   ,   3   ,   5   ,   9   ,   63  ,   21  ,   109 ,   409 ,   843 ,   1811    ,   1869    ,   6833    ,   6041    ,   31557   ,0 };
            const unsigned long dim2768Kuo2Init[]       =   {   1   ,   1   ,   5   ,   15  ,   17  ,   17  ,   25  ,   79  ,   157 ,   657 ,   1049    ,   2071    ,   4381    ,   14123   ,   4609    ,0 };
            const unsigned long dim2769Kuo2Init[]       =   {   1   ,   1   ,   5   ,   3   ,   3   ,   7   ,   45  ,   69  ,   311 ,   197 ,   1729    ,   2803    ,   2621    ,   1165    ,   26455   ,0 };
            const unsigned long dim2770Kuo2Init[]       =   {   1   ,   3   ,   7   ,   7   ,   23  ,   21  ,   21  ,   227 ,   481 ,   825 ,   1871    ,   325 ,   3199    ,   5591    ,   22061   ,0 };
            const unsigned long dim2771Kuo2Init[]       =   {   1   ,   3   ,   1   ,   7   ,   7   ,   53  ,   125 ,   163 ,   133 ,   1015    ,   727 ,   329 ,   403 ,   3347    ,   8957    ,0 };
            const unsigned long dim2772Kuo2Init[]       =   {   1   ,   1   ,   3   ,   1   ,   13  ,   19  ,   107 ,   31  ,   261 ,   771 ,   195 ,   2381    ,   2577    ,   5381    ,   7159    ,0 };
            const unsigned long dim2773Kuo2Init[]       =   {   1   ,   3   ,   7   ,   9   ,   7   ,   21  ,   61  ,   179 ,   405 ,   745 ,   689 ,   2527    ,   4189    ,   12869   ,   32101   ,0 };
            const unsigned long dim2774Kuo2Init[]       =   {   1   ,   3   ,   7   ,   1   ,   3   ,   25  ,   35  ,   103 ,   107 ,   433 ,   1039    ,   3557    ,   7207    ,   13945   ,   27751   ,0 };
            const unsigned long dim2775Kuo2Init[]       =   {   1   ,   3   ,   5   ,   5   ,   7   ,   9   ,   55  ,   93  ,   39  ,   965 ,   1391    ,   1069    ,   6003    ,   3785    ,   17337   ,0 };
            const unsigned long dim2776Kuo2Init[]       =   {   1   ,   1   ,   5   ,   5   ,   11  ,   33  ,   127 ,   247 ,   443 ,   697 ,   2007    ,   491 ,   3773    ,   4921    ,   21083   ,0 };
            const unsigned long dim2777Kuo2Init[]       =   {   1   ,   1   ,   5   ,   9   ,   1   ,   3   ,   33  ,   99  ,   31  ,   797 ,   295 ,   2855    ,   1911    ,   5741    ,   21255   ,0 };
            const unsigned long dim2778Kuo2Init[]       =   {   1   ,   1   ,   5   ,   9   ,   25  ,   27  ,   77  ,   111 ,   119 ,   1   ,   193 ,   2825    ,   3721    ,   2923    ,   1301    ,0 };
            const unsigned long dim2779Kuo2Init[]       =   {   1   ,   1   ,   1   ,   9   ,   11  ,   51  ,   51  ,   193 ,   331 ,   867 ,   1711    ,   969 ,   1245    ,   2815    ,   13165   ,0 };
            const unsigned long dim2780Kuo2Init[]       =   {   1   ,   3   ,   5   ,   1   ,   23  ,   29  ,   23  ,   193 ,   357 ,   567 ,   983 ,   2861    ,   3973    ,   3613    ,   8049    ,0 };
            const unsigned long dim2781Kuo2Init[]       =   {   1   ,   1   ,   7   ,   3   ,   27  ,   1   ,   117 ,   255 ,   149 ,   427 ,   1211    ,   2945    ,   4057    ,   725 ,   11061   ,0 };
            const unsigned long dim2782Kuo2Init[]       =   {   1   ,   1   ,   7   ,   5   ,   13  ,   5   ,   101 ,   21  ,   167 ,   535 ,   825 ,   639 ,   6659    ,   9779    ,   11939   ,0 };
            const unsigned long dim2783Kuo2Init[]       =   {   1   ,   3   ,   7   ,   1   ,   23  ,   25  ,   5   ,   15  ,   491 ,   531 ,   1035    ,   1989    ,   2125    ,   2779    ,   3207    ,0 };
            const unsigned long dim2784Kuo2Init[]       =   {   1   ,   1   ,   3   ,   13  ,   13  ,   25  ,   43  ,   155 ,   251 ,   761 ,   123 ,   1373    ,   649 ,   11709   ,   22331   ,0 };
            const unsigned long dim2785Kuo2Init[]       =   {   1   ,   3   ,   5   ,   13  ,   25  ,   37  ,   15  ,   247 ,   303 ,   95  ,   1837    ,   499 ,   2181    ,   11223   ,   1633    ,0 };
            const unsigned long dim2786Kuo2Init[]       =   {   1   ,   1   ,   1   ,   5   ,   15  ,   53  ,   29  ,   193 ,   11  ,   625 ,   21  ,   3893    ,   957 ,   493 ,   32393   ,0 };
            const unsigned long dim2787Kuo2Init[]       =   {   1   ,   1   ,   5   ,   9   ,   31  ,   13  ,   3   ,   235 ,   65  ,   563 ,   1787    ,   387 ,   2573    ,   3435    ,   8669    ,0 };
            const unsigned long dim2788Kuo2Init[]       =   {   1   ,   3   ,   5   ,   1   ,   29  ,   17  ,   73  ,   81  ,   207 ,   335 ,   1949    ,   3519    ,   6491    ,   9185    ,   2305    ,0 };
            const unsigned long dim2789Kuo2Init[]       =   {   1   ,   1   ,   3   ,   11  ,   5   ,   43  ,   15  ,   141 ,   347 ,   171 ,   1965    ,   2733    ,   7873    ,   7295    ,   10571   ,0 };
            const unsigned long dim2790Kuo2Init[]       =   {   1   ,   1   ,   5   ,   11  ,   5   ,   3   ,   73  ,   89  ,   189 ,   77  ,   877 ,   1127    ,   3609    ,   11649   ,   31191   ,0 };
            const unsigned long dim2791Kuo2Init[]       =   {   1   ,   1   ,   3   ,   15  ,   15  ,   25  ,   103 ,   73  ,   363 ,   261 ,   833 ,   2297    ,   4295    ,   13505   ,   8435    ,0 };
            const unsigned long dim2792Kuo2Init[]       =   {   1   ,   1   ,   3   ,   9   ,   23  ,   57  ,   15  ,   87  ,   215 ,   789 ,   975 ,   2659    ,   6407    ,   9883    ,   3863    ,0 };
            const unsigned long dim2793Kuo2Init[]       =   {   1   ,   1   ,   1   ,   9   ,   19  ,   51  ,   127 ,   77  ,   259 ,   161 ,   1527    ,   1911    ,   6707    ,   15573   ,   12479   ,0 };
            const unsigned long dim2794Kuo2Init[]       =   {   1   ,   3   ,   1   ,   15  ,   11  ,   5   ,   35  ,   89  ,   235 ,   281 ,   269 ,   511 ,   783 ,   307 ,   19887   ,0 };
            const unsigned long dim2795Kuo2Init[]       =   {   1   ,   3   ,   7   ,   15  ,   1   ,   43  ,   85  ,   25  ,   133 ,   577 ,   1719    ,   3999    ,   4597    ,   125 ,   12945   ,0 };
            const unsigned long dim2796Kuo2Init[]       =   {   1   ,   3   ,   5   ,   15  ,   3   ,   1   ,   3   ,   111 ,   145 ,   485 ,   1959    ,   1859    ,   3567    ,   4219    ,   7507    ,0 };
            const unsigned long dim2797Kuo2Init[]       =   {   1   ,   1   ,   3   ,   5   ,   27  ,   17  ,   45  ,   157 ,   87  ,   895 ,   497 ,   3   ,   1553    ,   2697    ,   24767   ,0 };
            const unsigned long dim2798Kuo2Init[]       =   {   1   ,   3   ,   5   ,   1   ,   31  ,   23  ,   123 ,   67  ,   349 ,   717 ,   1699    ,   823 ,   851 ,   1283    ,   25641   ,0 };
            const unsigned long dim2799Kuo2Init[]       =   {   1   ,   1   ,   5   ,   15  ,   9   ,   61  ,   43  ,   171 ,   33  ,   367 ,   1887    ,   917 ,   1901    ,   10881   ,   5699    ,0 };
            const unsigned long dim2800Kuo2Init[]       =   {   1   ,   3   ,   1   ,   7   ,   11  ,   33  ,   45  ,   19  ,   71  ,   913 ,   687 ,   265 ,   1579    ,   13857   ,   3833    ,0 };
            const unsigned long dim2801Kuo2Init[]       =   {   1   ,   3   ,   1   ,   5   ,   7   ,   15  ,   11  ,   87  ,   277 ,   435 ,   1479    ,   3025    ,   4663    ,   12063   ,   25185   ,0 };
            const unsigned long dim2802Kuo2Init[]       =   {   1   ,   1   ,   1   ,   1   ,   15  ,   11  ,   91  ,   99  ,   195 ,   77  ,   967 ,   2741    ,   4291    ,   4441    ,   2763    ,0 };
            const unsigned long dim2803Kuo2Init[]       =   {   1   ,   3   ,   7   ,   9   ,   9   ,   47  ,   25  ,   225 ,   193 ,   569 ,   561 ,   2567    ,   2751    ,   9035    ,   21613   ,0 };
            const unsigned long dim2804Kuo2Init[]       =   {   1   ,   1   ,   3   ,   9   ,   3   ,   25  ,   99  ,   95  ,   33  ,   811 ,   479 ,   99  ,   7893    ,   869 ,   6879    ,0 };
            const unsigned long dim2805Kuo2Init[]       =   {   1   ,   1   ,   5   ,   7   ,   11  ,   43  ,   111 ,   131 ,   107 ,   989 ,   147 ,   529 ,   6361    ,   769 ,   26651   ,0 };
            const unsigned long dim2806Kuo2Init[]       =   {   1   ,   1   ,   5   ,   9   ,   23  ,   47  ,   81  ,   97  ,   357 ,   613 ,   609 ,   357 ,   2001    ,   14795   ,   25093   ,0 };
            const unsigned long dim2807Kuo2Init[]       =   {   1   ,   1   ,   7   ,   3   ,   1   ,   13  ,   31  ,   171 ,   471 ,   981 ,   1923    ,   3053    ,   1577    ,   14433   ,   1795    ,0 };
            const unsigned long dim2808Kuo2Init[]       =   {   1   ,   1   ,   3   ,   11  ,   29  ,   15  ,   107 ,   5   ,   281 ,   773 ,   1323    ,   763 ,   6821    ,   15309   ,   21459   ,0 };
            const unsigned long dim2809Kuo2Init[]       =   {   1   ,   1   ,   5   ,   3   ,   17  ,   47  ,   49  ,   159 ,   17  ,   595 ,   1405    ,   1307    ,   3027    ,   3023    ,   17231   ,0 };
            const unsigned long dim2810Kuo2Init[]       =   {   1   ,   3   ,   3   ,   11  ,   5   ,   37  ,   39  ,   247 ,   391 ,   587 ,   1305    ,   969 ,   7183    ,   1079    ,   5723    ,0 };
            const unsigned long dim2811Kuo2Init[]       =   {   1   ,   3   ,   7   ,   5   ,   11  ,   57  ,   49  ,   55  ,   505 ,   961 ,   115 ,   2445    ,   583 ,   9175    ,   11897   ,0 };
            const unsigned long dim2812Kuo2Init[]       =   {   1   ,   1   ,   7   ,   13  ,   23  ,   35  ,   9   ,   59  ,   147 ,   139 ,   361 ,   1497    ,   1691    ,   1149    ,   19777   ,0 };
            const unsigned long dim2813Kuo2Init[]       =   {   1   ,   3   ,   7   ,   1   ,   19  ,   41  ,   71  ,   63  ,   401 ,   193 ,   107 ,   1891    ,   5353    ,   4619    ,   18311   ,0 };
            const unsigned long dim2814Kuo2Init[]       =   {   1   ,   1   ,   7   ,   1   ,   27  ,   7   ,   49  ,   69  ,   179 ,   461 ,   1625    ,   2431    ,   5149    ,   3221    ,   14221   ,0 };
            const unsigned long dim2815Kuo2Init[]       =   {   1   ,   3   ,   7   ,   13  ,   7   ,   11  ,   55  ,   93  ,   191 ,   1011    ,   139 ,   2323    ,   5103    ,   9655    ,   28605   ,0 };
            const unsigned long dim2816Kuo2Init[]       =   {   1   ,   1   ,   5   ,   7   ,   29  ,   53  ,   117 ,   253 ,   193 ,   29  ,   2003    ,   2051    ,   6113    ,   4941    ,   18407   ,0 };
            const unsigned long dim2817Kuo2Init[]       =   {   1   ,   1   ,   5   ,   5   ,   31  ,   7   ,   63  ,   179 ,   371 ,   1007    ,   107 ,   113 ,   6845    ,   8507    ,   9691    ,0 };
            const unsigned long dim2818Kuo2Init[]       =   {   1   ,   1   ,   7   ,   7   ,   9   ,   31  ,   91  ,   225 ,   115 ,   905 ,   1593    ,   2791    ,   1507    ,   13323   ,   787 ,0 };
            const unsigned long dim2819Kuo2Init[]       =   {   1   ,   3   ,   1   ,   9   ,   31  ,   13  ,   81  ,   111 ,   255 ,   467 ,   2007    ,   3591    ,   3849    ,   3445    ,   6859    ,0 };
            const unsigned long dim2820Kuo2Init[]       =   {   1   ,   3   ,   7   ,   15  ,   19  ,   3   ,   89  ,   171 ,   445 ,   803 ,   1583    ,   1875    ,   5631    ,   10583   ,   32527   ,0 };
            const unsigned long dim2821Kuo2Init[]       =   {   1   ,   3   ,   3   ,   15  ,   29  ,   29  ,   41  ,   199 ,   237 ,   507 ,   1715    ,   3887    ,   2951    ,   5803    ,   32749   ,0 };
            const unsigned long dim2822Kuo2Init[]       =   {   1   ,   1   ,   3   ,   5   ,   21  ,   51  ,   29  ,   131 ,   459 ,   611 ,   967 ,   99  ,   8057    ,   12685   ,   16021   ,0 };
            const unsigned long dim2823Kuo2Init[]       =   {   1   ,   3   ,   5   ,   5   ,   21  ,   19  ,   103 ,   131 ,   149 ,   477 ,   437 ,   4003    ,   7141    ,   2465    ,   32723   ,0 };
            const unsigned long dim2824Kuo2Init[]       =   {   1   ,   3   ,   7   ,   15  ,   9   ,   41  ,   115 ,   47  ,   261 ,   589 ,   1859    ,   631 ,   719 ,   4459    ,   26475   ,0 };
            const unsigned long dim2825Kuo2Init[]       =   {   1   ,   3   ,   1   ,   7   ,   9   ,   43  ,   79  ,   51  ,   383 ,   355 ,   1977    ,   739 ,   7333    ,   3573    ,   16061   ,0 };
            const unsigned long dim2826Kuo2Init[]       =   {   1   ,   1   ,   3   ,   1   ,   11  ,   9   ,   79  ,   141 ,   135 ,   195 ,   755 ,   585 ,   7125    ,   4643    ,   31791   ,0 };
            const unsigned long dim2827Kuo2Init[]       =   {   1   ,   1   ,   1   ,   7   ,   21  ,   23  ,   5   ,   225 ,   57  ,   307 ,   455 ,   3713    ,   8167    ,   4759    ,   10337   ,0 };
            const unsigned long dim2828Kuo2Init[]       =   {   1   ,   3   ,   1   ,   15  ,   3   ,   45  ,   9   ,   225 ,   167 ,   847 ,   1205    ,   2105    ,   2547    ,   10925   ,   31181   ,0 };
            const unsigned long dim2829Kuo2Init[]       =   {   1   ,   1   ,   3   ,   5   ,   3   ,   25  ,   57  ,   21  ,   45  ,   685 ,   53  ,   2447    ,   3139    ,   10949   ,   14939   ,0 };
            const unsigned long dim2830Kuo2Init[]       =   {   1   ,   3   ,   5   ,   7   ,   13  ,   55  ,   13  ,   185 ,   353 ,   613 ,   1655    ,   1171    ,   6261    ,   1703    ,   26859   ,0 };
            const unsigned long dim2831Kuo2Init[]       =   {   1   ,   1   ,   1   ,   15  ,   25  ,   29  ,   99  ,   81  ,   141 ,   387 ,   1369    ,   2177    ,   289 ,   14499   ,   23795   ,0 };
            const unsigned long dim2832Kuo2Init[]       =   {   1   ,   3   ,   7   ,   5   ,   1   ,   37  ,   123 ,   99  ,   1   ,   707 ,   1415    ,   1275    ,   3979    ,   3359    ,   17807   ,0 };
            const unsigned long dim2833Kuo2Init[]       =   {   1   ,   1   ,   5   ,   15  ,   19  ,   7   ,   113 ,   51  ,   329 ,   109 ,   1705    ,   517 ,   1887    ,   393 ,   8283    ,0 };
            const unsigned long dim2834Kuo2Init[]       =   {   1   ,   3   ,   1   ,   13  ,   23  ,   63  ,   89  ,   117 ,   89  ,   515 ,   1221    ,   3267    ,   2371    ,   5065    ,   16391   ,0 };
            const unsigned long dim2835Kuo2Init[]       =   {   1   ,   3   ,   5   ,   15  ,   1   ,   17  ,   39  ,   129 ,   453 ,   503 ,   841 ,   2997    ,   4279    ,   8639    ,   17623   ,0 };
            const unsigned long dim2836Kuo2Init[]       =   {   1   ,   3   ,   5   ,   15  ,   27  ,   23  ,   127 ,   85  ,   301 ,   491 ,   959 ,   3241    ,   7525    ,   13723   ,   12317   ,0 };
            const unsigned long dim2837Kuo2Init[]       =   {   1   ,   3   ,   5   ,   9   ,   31  ,   63  ,   45  ,   237 ,   173 ,   823 ,   223 ,   3129    ,   3285    ,   359 ,   3549    ,0 };
            const unsigned long dim2838Kuo2Init[]       =   {   1   ,   3   ,   5   ,   7   ,   19  ,   29  ,   85  ,   191 ,   395 ,   717 ,   1305    ,   2529    ,   7939    ,   10905   ,   8941    ,0 };
            const unsigned long dim2839Kuo2Init[]       =   {   1   ,   1   ,   5   ,   13  ,   31  ,   3   ,   41  ,   61  ,   461 ,   341 ,   683 ,   1973    ,   7127    ,   2183    ,   2137    ,0 };
            const unsigned long dim2840Kuo2Init[]       =   {   1   ,   3   ,   1   ,   7   ,   27  ,   33  ,   97  ,   97  ,   319 ,   329 ,   1075    ,   539 ,   3313    ,   6343    ,   11249   ,0 };
            const unsigned long dim2841Kuo2Init[]       =   {   1   ,   3   ,   5   ,   7   ,   25  ,   31  ,   99  ,   59  ,   337 ,   169 ,   861 ,   1989    ,   5999    ,   5397    ,   20563   ,0 };
            const unsigned long dim2842Kuo2Init[]       =   {   1   ,   1   ,   5   ,   13  ,   17  ,   37  ,   83  ,   33  ,   75  ,   887 ,   379 ,   3499    ,   733 ,   4343    ,   31503   ,0 };
            const unsigned long dim2843Kuo2Init[]       =   {   1   ,   1   ,   3   ,   1   ,   3   ,   45  ,   3   ,   245 ,   37  ,   797 ,   43  ,   943 ,   5301    ,   6857    ,   6653    ,0 };
            const unsigned long dim2844Kuo2Init[]       =   {   1   ,   1   ,   1   ,   3   ,   9   ,   7   ,   39  ,   231 ,   349 ,   481 ,   1831    ,   2115    ,   6381    ,   13021   ,   18775   ,0 };
            const unsigned long dim2845Kuo2Init[]       =   {   1   ,   3   ,   1   ,   11  ,   31  ,   27  ,   33  ,   103 ,   469 ,   987 ,   663 ,   3533    ,   1441    ,   4703    ,   23383   ,0 };
            const unsigned long dim2846Kuo2Init[]       =   {   1   ,   3   ,   3   ,   15  ,   27  ,   59  ,   31  ,   31  ,   463 ,   851 ,   1957    ,   2905    ,   4823    ,   12549   ,   849 ,0 };
            const unsigned long dim2847Kuo2Init[]       =   {   1   ,   3   ,   7   ,   5   ,   17  ,   31  ,   125 ,   183 ,   189 ,   297 ,   417 ,   1387    ,   5811    ,   10517   ,   19367   ,0 };
            const unsigned long dim2848Kuo2Init[]       =   {   1   ,   3   ,   7   ,   7   ,   7   ,   9   ,   91  ,   9   ,   343 ,   267 ,   493 ,   431 ,   2345    ,   7173    ,   1769    ,0 };
            const unsigned long dim2849Kuo2Init[]       =   {   1   ,   1   ,   1   ,   13  ,   13  ,   43  ,   75  ,   83  ,   19  ,   115 ,   1609    ,   2449    ,   5647    ,   12201   ,   27053   ,0 };
            const unsigned long dim2850Kuo2Init[]       =   {   1   ,   3   ,   7   ,   7   ,   1   ,   7   ,   81  ,   89  ,   371 ,   283 ,   993 ,   3875    ,   6853    ,   16039   ,   12725   ,0 };
            const unsigned long dim2851Kuo2Init[]       =   {   1   ,   1   ,   7   ,   1   ,   21  ,   61  ,   121 ,   135 ,   333 ,   339 ,   39  ,   2543    ,   3517    ,   9099    ,   27495   ,0 };
            const unsigned long dim2852Kuo2Init[]       =   {   1   ,   1   ,   7   ,   13  ,   19  ,   59  ,   65  ,   65  ,   407 ,   255 ,   915 ,   2679    ,   2841    ,   13981   ,   6351    ,0 };
            const unsigned long dim2853Kuo2Init[]       =   {   1   ,   3   ,   1   ,   5   ,   9   ,   45  ,   73  ,   217 ,   93  ,   3   ,   493 ,   531 ,   6709    ,   6543    ,   26243   ,0 };
            const unsigned long dim2854Kuo2Init[]       =   {   1   ,   3   ,   3   ,   13  ,   9   ,   25  ,   77  ,   223 ,   495 ,   789 ,   1885    ,   383 ,   7767    ,   4309    ,   29863   ,0 };
            const unsigned long dim2855Kuo2Init[]       =   {   1   ,   3   ,   5   ,   7   ,   21  ,   55  ,   11  ,   13  ,   19  ,   605 ,   1067    ,   667 ,   1163    ,   3847    ,   17997   ,0 };
            const unsigned long dim2856Kuo2Init[]       =   {   1   ,   1   ,   7   ,   1   ,   15  ,   51  ,   69  ,   115 ,   19  ,   1003    ,   1307    ,   3245    ,   7087    ,   6935    ,   25439   ,0 };
            const unsigned long dim2857Kuo2Init[]       =   {   1   ,   1   ,   7   ,   5   ,   11  ,   7   ,   85  ,   171 ,   301 ,   953 ,   507 ,   3873    ,   1121    ,   1663    ,   6321    ,0 };
            const unsigned long dim2858Kuo2Init[]       =   {   1   ,   1   ,   1   ,   13  ,   11  ,   1   ,   105 ,   175 ,   163 ,   847 ,   983 ,   3869    ,   3571    ,   3405    ,   25475   ,0 };
            const unsigned long dim2859Kuo2Init[]       =   {   1   ,   3   ,   5   ,   15  ,   7   ,   51  ,   89  ,   127 ,   347 ,   867 ,   429 ,   1483    ,   5203    ,   5215    ,   18521   ,0 };
            const unsigned long dim2860Kuo2Init[]       =   {   1   ,   1   ,   3   ,   7   ,   21  ,   35  ,   91  ,   107 ,   111 ,   895 ,   1827    ,   3273    ,   3189    ,   4253    ,   12643   ,0 };
            const unsigned long dim2861Kuo2Init[]       =   {   1   ,   3   ,   5   ,   15  ,   31  ,   33  ,   59  ,   27  ,   59  ,   297 ,   7   ,   1521    ,   1571    ,   10275   ,   32537   ,0 };
            const unsigned long dim2862Kuo2Init[]       =   {   1   ,   1   ,   5   ,   11  ,   1   ,   27  ,   17  ,   115 ,   349 ,   593 ,   1819    ,   707 ,   4915    ,   2053    ,   13393   ,0 };
            const unsigned long dim2863Kuo2Init[]       =   {   1   ,   1   ,   1   ,   1   ,   25  ,   47  ,   73  ,   163 ,   191 ,   637 ,   1723    ,   2431    ,   4821    ,   4283    ,   14647   ,0 };
            const unsigned long dim2864Kuo2Init[]       =   {   1   ,   3   ,   7   ,   13  ,   1   ,   55  ,   67  ,   89  ,   113 ,   757 ,   1675    ,   255 ,   99  ,   9427    ,   953 ,0 };
            const unsigned long dim2865Kuo2Init[]       =   {   1   ,   3   ,   7   ,   5   ,   1   ,   33  ,   7   ,   49  ,   299 ,   205 ,   1933    ,   2319    ,   6047    ,   6837    ,   6411    ,0 };
            const unsigned long dim2866Kuo2Init[]       =   {   1   ,   1   ,   7   ,   5   ,   11  ,   35  ,   61  ,   211 ,   487 ,   253 ,   1463    ,   2039    ,   3859    ,   4315    ,   29603   ,0 };
            const unsigned long dim2867Kuo2Init[]       =   {   1   ,   3   ,   1   ,   15  ,   23  ,   49  ,   99  ,   41  ,   263 ,   177 ,   193 ,   921 ,   1031    ,   13745   ,   14283   ,0 };
            const unsigned long dim2868Kuo2Init[]       =   {   1   ,   3   ,   7   ,   5   ,   19  ,   35  ,   91  ,   19  ,   481 ,   305 ,   477 ,   2989    ,   7319    ,   2477    ,   2481    ,0 };
            const unsigned long dim2869Kuo2Init[]       =   {   1   ,   3   ,   7   ,   13  ,   1   ,   15  ,   25  ,   133 ,   175 ,   571 ,   15  ,   505 ,   7459    ,   13283   ,   10253   ,0 };
            const unsigned long dim2870Kuo2Init[]       =   {   1   ,   3   ,   7   ,   11  ,   13  ,   27  ,   95  ,   179 ,   149 ,   267 ,   481 ,   543 ,   3643    ,   4261    ,   1241    ,0 };
            const unsigned long dim2871Kuo2Init[]       =   {   1   ,   1   ,   3   ,   1   ,   5   ,   25  ,   115 ,   27  ,   503 ,   595 ,   43  ,   975 ,   6223    ,   14173   ,   12977   ,0 };
            const unsigned long dim2872Kuo2Init[]       =   {   1   ,   1   ,   1   ,   13  ,   17  ,   55  ,   3   ,   129 ,   125 ,   565 ,   785 ,   2787    ,   5963    ,   12565   ,   11061   ,0 };
            const unsigned long dim2873Kuo2Init[]       =   {   1   ,   1   ,   1   ,   1   ,   23  ,   25  ,   51  ,   89  ,   317 ,   887 ,   1007    ,   1619    ,   3221    ,   4015    ,   18183   ,0 };
            const unsigned long dim2874Kuo2Init[]       =   {   1   ,   3   ,   3   ,   7   ,   27  ,   37  ,   33  ,   131 ,   379 ,   505 ,   1679    ,   2381    ,   629 ,   8593    ,   15553   ,0 };
            const unsigned long dim2875Kuo2Init[]       =   {   1   ,   3   ,   1   ,   9   ,   21  ,   53  ,   79  ,   85  ,   27  ,   541 ,   997 ,   1721    ,   2379    ,   9939    ,   32197   ,0 };
            const unsigned long dim2876Kuo2Init[]       =   {   1   ,   3   ,   7   ,   5   ,   29  ,   3   ,   29  ,   7   ,   499 ,   195 ,   157 ,   1431    ,   1117    ,   14781   ,   14811   ,0 };
            const unsigned long dim2877Kuo2Init[]       =   {   1   ,   3   ,   3   ,   9   ,   23  ,   19  ,   41  ,   59  ,   313 ,   391 ,   193 ,   3813    ,   2769    ,   7845    ,   4853    ,0 };
            const unsigned long dim2878Kuo2Init[]       =   {   1   ,   1   ,   7   ,   7   ,   7   ,   61  ,   103 ,   45  ,   225 ,   753 ,   339 ,   963 ,   3991    ,   8699    ,   19805   ,0 };
            const unsigned long dim2879Kuo2Init[]       =   {   1   ,   1   ,   3   ,   5   ,   15  ,   61  ,   125 ,   121 ,   53  ,   105 ,   121 ,   1939    ,   4745    ,   1193    ,   1733    ,0 };
            const unsigned long dim2880Kuo2Init[]       =   {   1   ,   1   ,   7   ,   7   ,   29  ,   29  ,   71  ,   137 ,   195 ,   757 ,   1251    ,   3483    ,   6977    ,   1189    ,   4929    ,0 };
            const unsigned long dim2881Kuo2Init[]       =   {   1   ,   1   ,   3   ,   9   ,   1   ,   11  ,   21  ,   39  ,   199 ,   873 ,   1063    ,   3327    ,   2171    ,   14611   ,   15283   ,0 };
            const unsigned long dim2882Kuo2Init[]       =   {   1   ,   3   ,   3   ,   15  ,   11  ,   19  ,   71  ,   185 ,   235 ,   211 ,   1435    ,   543 ,   3445    ,   2533    ,   25037   ,0 };
            const unsigned long dim2883Kuo2Init[]       =   {   1   ,   3   ,   1   ,   7   ,   15  ,   17  ,   33  ,   139 ,   301 ,   853 ,   2027    ,   1665    ,   4931    ,   2083    ,   9775    ,0 };
            const unsigned long dim2884Kuo2Init[]       =   {   1   ,   3   ,   5   ,   1   ,   31  ,   21  ,   91  ,   71  ,   273 ,   119 ,   801 ,   2071    ,   4537    ,   541 ,   743 ,0 };
            const unsigned long dim2885Kuo2Init[]       =   {   1   ,   1   ,   3   ,   5   ,   29  ,   23  ,   15  ,   225 ,   333 ,   979 ,   443 ,   1425    ,   3383    ,   5137    ,   27273   ,0 };
            const unsigned long dim2886Kuo2Init[]       =   {   1   ,   1   ,   1   ,   5   ,   21  ,   27  ,   121 ,   169 ,   495 ,   883 ,   355 ,   3923    ,   1505    ,   13033   ,   6421    ,0 };
            const unsigned long dim2887Kuo2Init[]       =   {   1   ,   3   ,   1   ,   5   ,   3   ,   17  ,   19  ,   169 ,   43  ,   659 ,   893 ,   1913    ,   885 ,   1239    ,   4055    ,0 };
            const unsigned long dim2888Kuo2Init[]       =   {   1   ,   1   ,   1   ,   13  ,   13  ,   39  ,   1   ,   123 ,   29  ,   573 ,   475 ,   535 ,   4685    ,   6377    ,   30009   ,0 };
            const unsigned long dim2889Kuo2Init[]       =   {   1   ,   3   ,   7   ,   7   ,   15  ,   53  ,   93  ,   183 ,   379 ,   125 ,   489 ,   4059    ,   5411    ,   7909    ,   25799   ,0 };
            const unsigned long dim2890Kuo2Init[]       =   {   1   ,   1   ,   1   ,   3   ,   23  ,   31  ,   75  ,   61  ,   75  ,   379 ,   1221    ,   3045    ,   7111    ,   10639   ,   3177    ,0 };
            const unsigned long dim2891Kuo2Init[]       =   {   1   ,   1   ,   1   ,   7   ,   27  ,   37  ,   77  ,   145 ,   35  ,   1005    ,   1105    ,   1301    ,   5217    ,   3011    ,   22631   ,0 };
            const unsigned long dim2892Kuo2Init[]       =   {   1   ,   1   ,   5   ,   5   ,   29  ,   7   ,   55  ,   139 ,   323 ,   413 ,   31  ,   1719    ,   2315    ,   5809    ,   9521    ,0 };
            const unsigned long dim2893Kuo2Init[]       =   {   1   ,   3   ,   1   ,   3   ,   21  ,   61  ,   45  ,   143 ,   427 ,   511 ,   1007    ,   1009    ,   6775    ,   4243    ,   29573   ,0 };
            const unsigned long dim2894Kuo2Init[]       =   {   1   ,   1   ,   3   ,   9   ,   11  ,   13  ,   29  ,   99  ,   179 ,   493 ,   1793    ,   375 ,   109 ,   8561    ,   14799   ,0 };
            const unsigned long dim2895Kuo2Init[]       =   {   1   ,   3   ,   1   ,   9   ,   5   ,   37  ,   83  ,   227 ,   241 ,   933 ,   869 ,   2125    ,   2733    ,   205 ,   14985   ,0 };
            const unsigned long dim2896Kuo2Init[]       =   {   1   ,   3   ,   7   ,   3   ,   11  ,   17  ,   39  ,   97  ,   229 ,   805 ,   491 ,   2881    ,   4147    ,   2159    ,   32461   ,0 };
            const unsigned long dim2897Kuo2Init[]       =   {   1   ,   3   ,   7   ,   15  ,   27  ,   17  ,   71  ,   227 ,   247 ,   731 ,   1287    ,   397 ,   7237    ,   10869   ,   7799    ,0 };
            const unsigned long dim2898Kuo2Init[]       =   {   1   ,   3   ,   7   ,   15  ,   7   ,   1   ,   63  ,   163 ,   249 ,   111 ,   1015    ,   2379    ,   7051    ,   11615   ,   20031   ,0 };
            const unsigned long dim2899Kuo2Init[]       =   {   1   ,   1   ,   7   ,   1   ,   1   ,   41  ,   61  ,   99  ,   385 ,   725 ,   161 ,   2595    ,   1789    ,   6619    ,   13397   ,0 };
            const unsigned long dim2900Kuo2Init[]       =   {   1   ,   1   ,   7   ,   15  ,   5   ,   53  ,   59  ,   153 ,   427 ,   251 ,   1977    ,   1757    ,   821 ,   6391    ,   7825    ,0 };
            const unsigned long dim2901Kuo2Init[]       =   {   1   ,   1   ,   3   ,   15  ,   17  ,   19  ,   83  ,   55  ,   199 ,   573 ,   337 ,   919 ,   649 ,   555 ,   27791   ,0 };
            const unsigned long dim2902Kuo2Init[]       =   {   1   ,   3   ,   1   ,   5   ,   25  ,   11  ,   23  ,   167 ,   443 ,   325 ,   2037    ,   2533    ,   643 ,   1815    ,   24747   ,0 };
            const unsigned long dim2903Kuo2Init[]       =   {   1   ,   1   ,   1   ,   7   ,   3   ,   15  ,   21  ,   73  ,   165 ,   241 ,   1089    ,   337 ,   4465    ,   11867   ,   31075   ,0 };
            const unsigned long dim2904Kuo2Init[]       =   {   1   ,   3   ,   1   ,   9   ,   31  ,   33  ,   41  ,   217 ,   185 ,   973 ,   1971    ,   1537    ,   1251    ,   4307    ,   22729   ,0 };
            const unsigned long dim2905Kuo2Init[]       =   {   1   ,   1   ,   1   ,   11  ,   7   ,   51  ,   17  ,   37  ,   113 ,   339 ,   211 ,   947 ,   4649    ,   8061    ,   1533    ,0 };
            const unsigned long dim2906Kuo2Init[]       =   {   1   ,   3   ,   3   ,   11  ,   7   ,   1   ,   29  ,   233 ,   145 ,   121 ,   83  ,   3311    ,   2007    ,   15325   ,   2507    ,0 };
            const unsigned long dim2907Kuo2Init[]       =   {   1   ,   3   ,   1   ,   5   ,   25  ,   9   ,   117 ,   227 ,   359 ,   401 ,   695 ,   841 ,   5825    ,   14661   ,   14937   ,0 };
            const unsigned long dim2908Kuo2Init[]       =   {   1   ,   3   ,   3   ,   13  ,   27  ,   43  ,   61  ,   117 ,   297 ,   773 ,   1589    ,   347 ,   3227    ,   10893   ,   3331    ,0 };
            const unsigned long dim2909Kuo2Init[]       =   {   1   ,   1   ,   5   ,   13  ,   5   ,   15  ,   65  ,   97  ,   155 ,   905 ,   655 ,   1665    ,   2759    ,   14709   ,   14327   ,0 };
            const unsigned long dim2910Kuo2Init[]       =   {   1   ,   3   ,   3   ,   11  ,   31  ,   9   ,   63  ,   207 ,   287 ,   895 ,   817 ,   1753    ,   1961    ,   15363   ,   5269    ,0 };
            const unsigned long dim2911Kuo2Init[]       =   {   1   ,   1   ,   1   ,   11  ,   3   ,   17  ,   21  ,   131 ,   295 ,   553 ,   1583    ,   1797    ,   6949    ,   7047    ,   30351   ,0 };
            const unsigned long dim2912Kuo2Init[]       =   {   1   ,   3   ,   7   ,   5   ,   23  ,   9   ,   123 ,   155 ,   1   ,   525 ,   469 ,   165 ,   591 ,   6113    ,   17593   ,0 };
            const unsigned long dim2913Kuo2Init[]       =   {   1   ,   3   ,   5   ,   13  ,   27  ,   21  ,   125 ,   175 ,   323 ,   113 ,   1805    ,   3775    ,   443 ,   15723   ,   187 ,0 };
            const unsigned long dim2914Kuo2Init[]       =   {   1   ,   3   ,   5   ,   13  ,   29  ,   7   ,   89  ,   123 ,   349 ,   469 ,   635 ,   629 ,   1227    ,   11601   ,   10409   ,0 };
            const unsigned long dim2915Kuo2Init[]       =   {   1   ,   3   ,   1   ,   5   ,   13  ,   7   ,   13  ,   209 ,   431 ,   59  ,   2001    ,   1347    ,   4619    ,   11581   ,   23281   ,0 };
            const unsigned long dim2916Kuo2Init[]       =   {   1   ,   3   ,   5   ,   3   ,   3   ,   9   ,   15  ,   169 ,   77  ,   293 ,   17  ,   3205    ,   5035    ,   14777   ,   2175    ,0 };
            const unsigned long dim2917Kuo2Init[]       =   {   1   ,   1   ,   1   ,   9   ,   27  ,   5   ,   39  ,   237 ,   275 ,   121 ,   217 ,   1283    ,   1619    ,   9115    ,   18019   ,0 };
            const unsigned long dim2918Kuo2Init[]       =   {   1   ,   1   ,   1   ,   11  ,   31  ,   43  ,   111 ,   235 ,   439 ,   217 ,   769 ,   513 ,   5147    ,   4033    ,   27329   ,0 };
            const unsigned long dim2919Kuo2Init[]       =   {   1   ,   1   ,   5   ,   5   ,   31  ,   25  ,   35  ,   71  ,   441 ,   1013    ,   963 ,   2945    ,   4507    ,   16329   ,   5965    ,0 };
            const unsigned long dim2920Kuo2Init[]       =   {   1   ,   3   ,   1   ,   7   ,   27  ,   25  ,   109 ,   43  ,   421 ,   963 ,   553 ,   3347    ,   3159    ,   11909   ,   29827   ,0 };
            const unsigned long dim2921Kuo2Init[]       =   {   1   ,   3   ,   3   ,   11  ,   5   ,   17  ,   1   ,   253 ,   497 ,   209 ,   423 ,   3837    ,   5233    ,   3919    ,   4593    ,0 };
            const unsigned long dim2922Kuo2Init[]       =   {   1   ,   3   ,   5   ,   11  ,   1   ,   61  ,   85  ,   61  ,   59  ,   637 ,   1017    ,   455 ,   5529    ,   7495    ,   16645   ,0 };
            const unsigned long dim2923Kuo2Init[]       =   {   1   ,   1   ,   1   ,   5   ,   11  ,   11  ,   111 ,   151 ,   297 ,   289 ,   259 ,   1737    ,   4033    ,   5545    ,   22049   ,0 };
            const unsigned long dim2924Kuo2Init[]       =   {   1   ,   1   ,   3   ,   9   ,   11  ,   33  ,   53  ,   43  ,   95  ,   859 ,   1113    ,   3201    ,   2705    ,   15009   ,   24945   ,0 };
            const unsigned long dim2925Kuo2Init[]       =   {   1   ,   1   ,   7   ,   3   ,   7   ,   1   ,   59  ,   201 ,   127 ,   1023    ,   879 ,   1359    ,   4323    ,   11829   ,   32591   ,0 };
            const unsigned long dim2926Kuo2Init[]       =   {   1   ,   1   ,   5   ,   13  ,   1   ,   15  ,   65  ,   3   ,   7   ,   373 ,   427 ,   3987    ,   3849    ,   16289   ,   3669    ,0 };
            const unsigned long dim2927Kuo2Init[]       =   {   1   ,   1   ,   1   ,   3   ,   23  ,   21  ,   49  ,   215 ,   159 ,   577 ,   963 ,   3257    ,   5919    ,   4397    ,   14811   ,0 };
            const unsigned long dim2928Kuo2Init[]       =   {   1   ,   3   ,   1   ,   9   ,   3   ,   43  ,   51  ,   49  ,   331 ,   717 ,   791 ,   279 ,   3305    ,   955 ,   14003   ,0 };
            const unsigned long dim2929Kuo2Init[]       =   {   1   ,   1   ,   7   ,   7   ,   19  ,   45  ,   11  ,   249 ,   9   ,   221 ,   1525    ,   2413    ,   1807    ,   191 ,   3063    ,0 };
            const unsigned long dim2930Kuo2Init[]       =   {   1   ,   3   ,   7   ,   1   ,   17  ,   33  ,   45  ,   201 ,   471 ,   683 ,   13  ,   3851    ,   7337    ,   2249    ,   8287    ,0 };
            const unsigned long dim2931Kuo2Init[]       =   {   1   ,   3   ,   3   ,   1   ,   11  ,   21  ,   85  ,   141 ,   445 ,   691 ,   577 ,   2781    ,   3677    ,   11799   ,   25555   ,0 };
            const unsigned long dim2932Kuo2Init[]       =   {   1   ,   1   ,   7   ,   1   ,   3   ,   27  ,   71  ,   245 ,   341 ,   69  ,   1355    ,   1289    ,   8165    ,   1159    ,   6001    ,0 };
            const unsigned long dim2933Kuo2Init[]       =   {   1   ,   3   ,   3   ,   15  ,   11  ,   31  ,   31  ,   117 ,   193 ,   847 ,   1755    ,   1629    ,   6707    ,   5447    ,   15961   ,0 };
            const unsigned long dim2934Kuo2Init[]       =   {   1   ,   1   ,   1   ,   1   ,   23  ,   13  ,   69  ,   107 ,   33  ,   731 ,   233 ,   2225    ,   201 ,   2667    ,   19935   ,0 };
            const unsigned long dim2935Kuo2Init[]       =   {   1   ,   1   ,   1   ,   9   ,   31  ,   9   ,   25  ,   225 ,   373 ,   99  ,   1725    ,   3029    ,   5019    ,   10927   ,   541 ,0 };
            const unsigned long dim2936Kuo2Init[]       =   {   1   ,   3   ,   3   ,   1   ,   29  ,   29  ,   37  ,   235 ,   135 ,   99  ,   599 ,   581 ,   7663    ,   1607    ,   7957    ,0 };
            const unsigned long dim2937Kuo2Init[]       =   {   1   ,   3   ,   1   ,   11  ,   1   ,   31  ,   79  ,   33  ,   319 ,   341 ,   1701    ,   973 ,   5783    ,   3003    ,   15311   ,0 };
            const unsigned long dim2938Kuo2Init[]       =   {   1   ,   3   ,   5   ,   15  ,   19  ,   11  ,   95  ,   51  ,   35  ,   887 ,   1505    ,   207 ,   7925    ,   1919    ,   30735   ,0 };
            const unsigned long dim2939Kuo2Init[]       =   {   1   ,   3   ,   5   ,   7   ,   3   ,   27  ,   55  ,   41  ,   293 ,   519 ,   1885    ,   617 ,   321 ,   5911    ,   26813   ,0 };
            const unsigned long dim2940Kuo2Init[]       =   {   1   ,   1   ,   7   ,   1   ,   7   ,   59  ,   9   ,   59  ,   291 ,   927 ,   1241    ,   1707    ,   3833    ,   13859   ,   3311    ,0 };
            const unsigned long dim2941Kuo2Init[]       =   {   1   ,   3   ,   3   ,   13  ,   17  ,   61  ,   1   ,   155 ,   355 ,   943 ,   159 ,   2519    ,   7835    ,   14097   ,   671 ,0 };
            const unsigned long dim2942Kuo2Init[]       =   {   1   ,   1   ,   3   ,   3   ,   29  ,   33  ,   59  ,   37  ,   363 ,   605 ,   365 ,   2431    ,   2529    ,   6449    ,   8847    ,0 };
            const unsigned long dim2943Kuo2Init[]       =   {   1   ,   3   ,   7   ,   5   ,   1   ,   33  ,   87  ,   215 ,   465 ,   937 ,   2031    ,   421 ,   2535    ,   10017   ,   18395   ,0 };
            const unsigned long dim2944Kuo2Init[]       =   {   1   ,   1   ,   7   ,   1   ,   27  ,   25  ,   7   ,   103 ,   501 ,   53  ,   1491    ,   99  ,   7433    ,   8251    ,   18849   ,0 };
            const unsigned long dim2945Kuo2Init[]       =   {   1   ,   3   ,   1   ,   1   ,   31  ,   15  ,   7   ,   89  ,   145 ,   457 ,   7   ,   2821    ,   6595    ,   5893    ,   29049   ,0 };
            const unsigned long dim2946Kuo2Init[]       =   {   1   ,   1   ,   3   ,   7   ,   15  ,   55  ,   127 ,   39  ,   493 ,   39  ,   1753    ,   643 ,   4005    ,   12041   ,   21811   ,0 };
            const unsigned long dim2947Kuo2Init[]       =   {   1   ,   3   ,   3   ,   3   ,   17  ,   23  ,   79  ,   129 ,   181 ,   1017    ,   767 ,   1869    ,   4589    ,   4059    ,   15391   ,0 };
            const unsigned long dim2948Kuo2Init[]       =   {   1   ,   1   ,   1   ,   3   ,   13  ,   17  ,   109 ,   253 ,   441 ,   361 ,   1641    ,   1647    ,   2139    ,   11629   ,   21061   ,0 };
            const unsigned long dim2949Kuo2Init[]       =   {   1   ,   1   ,   3   ,   15  ,   5   ,   15  ,   43  ,   113 ,   455 ,   17  ,   353 ,   3431    ,   4871    ,   14777   ,   3257    ,0 };
            const unsigned long dim2950Kuo2Init[]       =   {   1   ,   3   ,   1   ,   1   ,   7   ,   5   ,   49  ,   199 ,   295 ,   467 ,   1151    ,   2563    ,   1261    ,   1611    ,   8457    ,0 };
            const unsigned long dim2951Kuo2Init[]       =   {   1   ,   3   ,   3   ,   9   ,   21  ,   1   ,   107 ,   71  ,   67  ,   271 ,   427 ,   673 ,   2093    ,   13255   ,   15941   ,0 };
            const unsigned long dim2952Kuo2Init[]       =   {   1   ,   1   ,   7   ,   7   ,   19  ,   55  ,   127 ,   109 ,   101 ,   773 ,   1423    ,   2725    ,   2403    ,   1419    ,   2543    ,0 };
            const unsigned long dim2953Kuo2Init[]       =   {   1   ,   3   ,   5   ,   9   ,   25  ,   49  ,   15  ,   77  ,   423 ,   723 ,   2033    ,   275 ,   1593    ,   3055    ,   12043   ,0 };
            const unsigned long dim2954Kuo2Init[]       =   {   1   ,   3   ,   1   ,   9   ,   29  ,   27  ,   115 ,   5   ,   407 ,   797 ,   2041    ,   155 ,   2389    ,   9749    ,   31653   ,0 };
            const unsigned long dim2955Kuo2Init[]       =   {   1   ,   3   ,   3   ,   7   ,   5   ,   29  ,   123 ,   139 ,   467 ,   377 ,   1853    ,   2769    ,   6523    ,   14195   ,   31539   ,0 };
            const unsigned long dim2956Kuo2Init[]       =   {   1   ,   3   ,   3   ,   5   ,   25  ,   41  ,   103 ,   89  ,   469 ,   577 ,   1583    ,   1061    ,   7609    ,   4261    ,   17323   ,0 };
            const unsigned long dim2957Kuo2Init[]       =   {   1   ,   1   ,   5   ,   5   ,   29  ,   37  ,   79  ,   179 ,   189 ,   185 ,   535 ,   3433    ,   6091    ,   4935    ,   21853   ,0 };
            const unsigned long dim2958Kuo2Init[]       =   {   1   ,   1   ,   5   ,   13  ,   17  ,   63  ,   119 ,   191 ,   7   ,   353 ,   1591    ,   2881    ,   7857    ,   11355   ,   32427   ,0 };
            const unsigned long dim2959Kuo2Init[]       =   {   1   ,   3   ,   1   ,   7   ,   13  ,   53  ,   21  ,   151 ,   499 ,   721 ,   1495    ,   2101    ,   7061    ,   12529   ,   16733   ,0 };
            const unsigned long dim2960Kuo2Init[]       =   {   1   ,   1   ,   7   ,   1   ,   1   ,   1   ,   17  ,   141 ,   221 ,   875 ,   1569    ,   3165    ,   2583    ,   2803    ,   21001   ,0 };
            const unsigned long dim2961Kuo2Init[]       =   {   1   ,   3   ,   3   ,   13  ,   7   ,   7   ,   99  ,   129 ,   381 ,   237 ,   159 ,   2359    ,   2223    ,   7465    ,   25437   ,0 };
            const unsigned long dim2962Kuo2Init[]       =   {   1   ,   1   ,   3   ,   15  ,   31  ,   19  ,   103 ,   17  ,   475 ,   649 ,   279 ,   1321    ,   1751    ,   6957    ,   26055   ,0 };
            const unsigned long dim2963Kuo2Init[]       =   {   1   ,   1   ,   1   ,   7   ,   5   ,   41  ,   3   ,   29  ,   23  ,   513 ,   1619    ,   291 ,   3949    ,   12117   ,   15175   ,0 };
            const unsigned long dim2964Kuo2Init[]       =   {   1   ,   3   ,   1   ,   3   ,   13  ,   13  ,   73  ,   77  ,   373 ,   999 ,   1985    ,   161 ,   3223    ,   10873   ,   2967    ,0 };
            const unsigned long dim2965Kuo2Init[]       =   {   1   ,   1   ,   5   ,   7   ,   27  ,   1   ,   47  ,   79  ,   463 ,   499 ,   37  ,   113 ,   3207    ,   6753    ,   13207   ,0 };
            const unsigned long dim2966Kuo2Init[]       =   {   1   ,   1   ,   1   ,   9   ,   29  ,   33  ,   71  ,   225 ,   71  ,   943 ,   197 ,   3645    ,   7459    ,   5495    ,   22449   ,0 };
            const unsigned long dim2967Kuo2Init[]       =   {   1   ,   1   ,   3   ,   9   ,   1   ,   21  ,   101 ,   31  ,   357 ,   689 ,   1081    ,   165 ,   621 ,   11277   ,   21773   ,0 };
            const unsigned long dim2968Kuo2Init[]       =   {   1   ,   1   ,   7   ,   3   ,   13  ,   13  ,   21  ,   11  ,   87  ,   741 ,   765 ,   3299    ,   2741    ,   12825   ,   13383   ,0 };
            const unsigned long dim2969Kuo2Init[]       =   {   1   ,   3   ,   3   ,   5   ,   5   ,   51  ,   49  ,   215 ,   427 ,   513 ,   621 ,   4067    ,   661 ,   12261   ,   6839    ,0 };
            const unsigned long dim2970Kuo2Init[]       =   {   1   ,   1   ,   7   ,   9   ,   13  ,   9   ,   77  ,   17  ,   229 ,   139 ,   1053    ,   2061    ,   4915    ,   1047    ,   3621    ,0 };
            const unsigned long dim2971Kuo2Init[]       =   {   1   ,   1   ,   3   ,   1   ,   9   ,   7   ,   125 ,   165 ,   301 ,   721 ,   493 ,   543 ,   4885    ,   5011    ,   24545   ,0 };
            const unsigned long dim2972Kuo2Init[]       =   {   1   ,   3   ,   3   ,   11  ,   31  ,   11  ,   5   ,   61  ,   89  ,   35  ,   1551    ,   3031    ,   1175    ,   16223   ,   19433   ,0 };
            const unsigned long dim2973Kuo2Init[]       =   {   1   ,   1   ,   1   ,   11  ,   27  ,   31  ,   115 ,   181 ,   487 ,   53  ,   795 ,   1789    ,   7343    ,   7821    ,   20469   ,0 };
            const unsigned long dim2974Kuo2Init[]       =   {   1   ,   1   ,   5   ,   3   ,   7   ,   25  ,   89  ,   145 ,   107 ,   361 ,   171 ,   2407    ,   1159    ,   9341    ,   23877   ,0 };
            const unsigned long dim2975Kuo2Init[]       =   {   1   ,   3   ,   3   ,   15  ,   27  ,   47  ,   31  ,   53  ,   53  ,   479 ,   1479    ,   427 ,   461 ,   7287    ,   15947   ,0 };
            const unsigned long dim2976Kuo2Init[]       =   {   1   ,   3   ,   7   ,   5   ,   3   ,   11  ,   79  ,   151 ,   463 ,   315 ,   919 ,   1531    ,   2367    ,   5189    ,   6971    ,0 };
            const unsigned long dim2977Kuo2Init[]       =   {   1   ,   1   ,   5   ,   9   ,   25  ,   17  ,   3   ,   111 ,   167 ,   365 ,   871 ,   355 ,   6763    ,   11889   ,   5531    ,0 };
            const unsigned long dim2978Kuo2Init[]       =   {   1   ,   1   ,   1   ,   13  ,   25  ,   31  ,   25  ,   185 ,   63  ,   335 ,   177 ,   3017    ,   941 ,   11849   ,   6879    ,0 };
            const unsigned long dim2979Kuo2Init[]       =   {   1   ,   1   ,   3   ,   7   ,   29  ,   13  ,   81  ,   121 ,   119 ,   807 ,   1647    ,   569 ,   23  ,   9045    ,   15513   ,0 };
            const unsigned long dim2980Kuo2Init[]       =   {   1   ,   1   ,   1   ,   5   ,   5   ,   61  ,   115 ,   79  ,   115 ,   401 ,   383 ,   3679    ,   8159    ,   15199   ,   1531    ,0 };
            const unsigned long dim2981Kuo2Init[]       =   {   1   ,   3   ,   3   ,   11  ,   17  ,   45  ,   17  ,   37  ,   175 ,   735 ,   849 ,   3597    ,   7253    ,   11863   ,   5215    ,0 };
            const unsigned long dim2982Kuo2Init[]       =   {   1   ,   1   ,   1   ,   13  ,   7   ,   59  ,   99  ,   67  ,   223 ,   635 ,   1645    ,   3169    ,   2655    ,   15925   ,   14969   ,0 };
            const unsigned long dim2983Kuo2Init[]       =   {   1   ,   3   ,   7   ,   15  ,   25  ,   17  ,   17  ,   25  ,   1   ,   799 ,   829 ,   2085    ,   7371    ,   6905    ,   11573   ,0 };
            const unsigned long dim2984Kuo2Init[]       =   {   1   ,   1   ,   5   ,   1   ,   19  ,   33  ,   127 ,   31  ,   305 ,   523 ,   1141    ,   1171    ,   3707    ,   14687   ,   27897   ,0 };
            const unsigned long dim2985Kuo2Init[]       =   {   1   ,   1   ,   1   ,   11  ,   3   ,   7   ,   95  ,   243 ,   373 ,   481 ,   1663    ,   1449    ,   687 ,   5643    ,   3265    ,0 };
            const unsigned long dim2986Kuo2Init[]       =   {   1   ,   3   ,   7   ,   11  ,   17  ,   35  ,   45  ,   15  ,   201 ,   389 ,   1085    ,   1597    ,   5073    ,   5009    ,   6921    ,0 };
            const unsigned long dim2987Kuo2Init[]       =   {   1   ,   1   ,   5   ,   1   ,   25  ,   33  ,   99  ,   57  ,   429 ,   153 ,   1663    ,   1797    ,   785 ,   7323    ,   24763   ,0 };
            const unsigned long dim2988Kuo2Init[]       =   {   1   ,   1   ,   7   ,   15  ,   1   ,   63  ,   93  ,   225 ,   189 ,   297 ,   927 ,   2127    ,   7335    ,   5187    ,   12047   ,0 };
            const unsigned long dim2989Kuo2Init[]       =   {   1   ,   1   ,   5   ,   9   ,   23  ,   47  ,   27  ,   113 ,   469 ,   649 ,   1135    ,   1363    ,   6777    ,   2047    ,   5759    ,0 };
            const unsigned long dim2990Kuo2Init[]       =   {   1   ,   1   ,   3   ,   1   ,   9   ,   33  ,   21  ,   123 ,   309 ,   635 ,   1217    ,   1087    ,   4525    ,   10165   ,   18071   ,0 };
            const unsigned long dim2991Kuo2Init[]       =   {   1   ,   3   ,   3   ,   3   ,   19  ,   49  ,   13  ,   217 ,   155 ,   945 ,   1987    ,   2425    ,   3891    ,   13831   ,   21777   ,0 };
            const unsigned long dim2992Kuo2Init[]       =   {   1   ,   1   ,   3   ,   11  ,   25  ,   9   ,   75  ,   149 ,   241 ,   175 ,   799 ,   1285    ,   3265    ,   14103   ,   30147   ,0 };
            const unsigned long dim2993Kuo2Init[]       =   {   1   ,   1   ,   1   ,   3   ,   13  ,   41  ,   87  ,   127 ,   179 ,   723 ,   1345    ,   667 ,   4155    ,   8037    ,   29583   ,0 };
            const unsigned long dim2994Kuo2Init[]       =   {   1   ,   1   ,   1   ,   9   ,   15  ,   57  ,   67  ,   197 ,   131 ,   957 ,   1855    ,   2857    ,   2939    ,   6459    ,   5807    ,0 };
            const unsigned long dim2995Kuo2Init[]       =   {   1   ,   3   ,   5   ,   5   ,   23  ,   39  ,   19  ,   95  ,   103 ,   961 ,   239 ,   1151    ,   1069    ,   10261   ,   6065    ,0 };
            const unsigned long dim2996Kuo2Init[]       =   {   1   ,   3   ,   3   ,   15  ,   23  ,   29  ,   51  ,   121 ,   45  ,   561 ,   1317    ,   1777    ,   1325    ,   5833    ,   5765    ,0 };
            const unsigned long dim2997Kuo2Init[]       =   {   1   ,   1   ,   5   ,   1   ,   31  ,   3   ,   67  ,   9   ,   453 ,   281 ,   525 ,   1751    ,   5771    ,   14257   ,   11869   ,0 };
            const unsigned long dim2998Kuo2Init[]       =   {   1   ,   1   ,   7   ,   1   ,   13  ,   35  ,   33  ,   59  ,   117 ,   737 ,   929 ,   547 ,   4839    ,   7013    ,   28395   ,0 };
            const unsigned long dim2999Kuo2Init[]       =   {   1   ,   3   ,   7   ,   15  ,   21  ,   41  ,   19  ,   15  ,   73  ,   377 ,   665 ,   3867    ,   6445    ,   13137   ,   995 ,0 };
            const unsigned long dim3000Kuo2Init[]       =   {   1   ,   3   ,   7   ,   11  ,   13  ,   41  ,   53  ,   115 ,   435 ,   607 ,   871 ,   315 ,   6573    ,   6321    ,   10413   ,0 };
            const unsigned long dim3001Kuo2Init[]       =   {   1   ,   3   ,   3   ,   3   ,   27  ,   31  ,   23  ,   7   ,   151 ,   923 ,   1903    ,   789 ,   5031    ,   10631   ,   777 ,0 };
            const unsigned long dim3002Kuo2Init[]       =   {   1   ,   1   ,   7   ,   7   ,   9   ,   25  ,   17  ,   241 ,   19  ,   569 ,   1269    ,   1211    ,   3947    ,   11351   ,   31211   ,0 };
            const unsigned long dim3003Kuo2Init[]       =   {   1   ,   1   ,   7   ,   13  ,   25  ,   31  ,   79  ,   111 ,   157 ,   483 ,   1787    ,   1201    ,   3491    ,   7943    ,   4879    ,0 };
            const unsigned long dim3004Kuo2Init[]       =   {   1   ,   3   ,   7   ,   13  ,   19  ,   9   ,   31  ,   133 ,   125 ,   347 ,   15  ,   1287    ,   4355    ,   15357   ,   21391   ,0 };
            const unsigned long dim3005Kuo2Init[]       =   {   1   ,   3   ,   3   ,   13  ,   1   ,   29  ,   63  ,   175 ,   279 ,   799 ,   1867    ,   2221    ,   6043    ,   307 ,   12475   ,0 };
            const unsigned long dim3006Kuo2Init[]       =   {   1   ,   1   ,   7   ,   5   ,   23  ,   51  ,   93  ,   65  ,   209 ,   45  ,   409 ,   3405    ,   2053    ,   1399    ,   4423    ,0 };
            const unsigned long dim3007Kuo2Init[]       =   {   1   ,   3   ,   7   ,   13  ,   23  ,   55  ,   21  ,   121 ,   281 ,   115 ,   85  ,   1941    ,   2609    ,   7041    ,   16267   ,0 };
            const unsigned long dim3008Kuo2Init[]       =   {   1   ,   1   ,   3   ,   5   ,   25  ,   3   ,   113 ,   107 ,   37  ,   393 ,   737 ,   2671    ,   7405    ,   12191   ,   27679   ,0 };
            const unsigned long dim3009Kuo2Init[]       =   {   1   ,   1   ,   3   ,   13  ,   15  ,   13  ,   33  ,   233 ,   25  ,   163 ,   1097    ,   2977    ,   6933    ,   14625   ,   29587   ,0 };
            const unsigned long dim3010Kuo2Init[]       =   {   1   ,   1   ,   1   ,   1   ,   31  ,   13  ,   11  ,   25  ,   47  ,   589 ,   2035    ,   3595    ,   1299    ,   10171   ,   12695   ,0 };
            const unsigned long dim3011Kuo2Init[]       =   {   1   ,   3   ,   5   ,   11  ,   31  ,   63  ,   97  ,   67  ,   381 ,   1021    ,   1029    ,   2869    ,   5779    ,   15451   ,   4291    ,0 };
            const unsigned long dim3012Kuo2Init[]       =   {   1   ,   1   ,   1   ,   5   ,   9   ,   37  ,   9   ,   205 ,   169 ,   375 ,   1367    ,   1749    ,   8015    ,   1019    ,   12671   ,0 };
            const unsigned long dim3013Kuo2Init[]       =   {   1   ,   3   ,   5   ,   7   ,   13  ,   29  ,   75  ,   53  ,   173 ,   535 ,   1795    ,   1705    ,   3601    ,   2487    ,   29851   ,0 };
            const unsigned long dim3014Kuo2Init[]       =   {   1   ,   3   ,   3   ,   5   ,   5   ,   17  ,   81  ,   209 ,   229 ,   913 ,   959 ,   3597    ,   1515    ,   13401   ,   1531    ,0 };
            const unsigned long dim3015Kuo2Init[]       =   {   1   ,   3   ,   3   ,   5   ,   29  ,   47  ,   5   ,   43  ,   117 ,   215 ,   1469    ,   3743    ,   6801    ,   10511   ,   11097   ,0 };
            const unsigned long dim3016Kuo2Init[]       =   {   1   ,   1   ,   1   ,   9   ,   11  ,   5   ,   79  ,   247 ,   43  ,   853 ,   663 ,   1967    ,   355 ,   12263   ,   15877   ,0 };
            const unsigned long dim3017Kuo2Init[]       =   {   1   ,   3   ,   3   ,   5   ,   25  ,   23  ,   105 ,   203 ,   255 ,   677 ,   1855    ,   527 ,   825 ,   433 ,   27701   ,0 };
            const unsigned long dim3018Kuo2Init[]       =   {   1   ,   3   ,   3   ,   1   ,   13  ,   29  ,   119 ,   47  ,   331 ,   203 ,   1515    ,   1581    ,   1707    ,   10621   ,   19371   ,0 };
            const unsigned long dim3019Kuo2Init[]       =   {   1   ,   1   ,   1   ,   13  ,   27  ,   15  ,   5   ,   255 ,   201 ,   861 ,   1809    ,   703 ,   1707    ,   8209    ,   5175    ,0 };
            const unsigned long dim3020Kuo2Init[]       =   {   1   ,   3   ,   1   ,   13  ,   17  ,   61  ,   41  ,   143 ,   459 ,   291 ,   1747    ,   3605    ,   5545    ,   7441    ,   21493   ,0 };
            const unsigned long dim3021Kuo2Init[]       =   {   1   ,   1   ,   3   ,   9   ,   21  ,   61  ,   77  ,   61  ,   79  ,   855 ,   57  ,   1543    ,   2755    ,   6249    ,   3387    ,0 };
            const unsigned long dim3022Kuo2Init[]       =   {   1   ,   1   ,   3   ,   1   ,   17  ,   5   ,   117 ,   243 ,   153 ,   397 ,   693 ,   701 ,   2201    ,   11375   ,   23753   ,0 };
            const unsigned long dim3023Kuo2Init[]       =   {   1   ,   1   ,   7   ,   7   ,   17  ,   25  ,   123 ,   239 ,   411 ,   633 ,   395 ,   3809    ,   1483    ,   15125   ,   22221   ,0 };
            const unsigned long dim3024Kuo2Init[]       =   {   1   ,   3   ,   3   ,   3   ,   7   ,   45  ,   57  ,   51  ,   35  ,   463 ,   1287    ,   3101    ,   4213    ,   12931   ,   28639   ,0 };
            const unsigned long dim3025Kuo2Init[]       =   {   1   ,   3   ,   1   ,   5   ,   13  ,   7   ,   103 ,   177 ,   277 ,   645 ,   833 ,   883 ,   3607    ,   13545   ,   6405    ,0 };
            const unsigned long dim3026Kuo2Init[]       =   {   1   ,   3   ,   7   ,   11  ,   17  ,   3   ,   41  ,   161 ,   511 ,   483 ,   773 ,   147 ,   8113    ,   2203    ,   29711   ,0 };
            const unsigned long dim3027Kuo2Init[]       =   {   1   ,   3   ,   1   ,   5   ,   13  ,   29  ,   79  ,   189 ,   351 ,   65  ,   93  ,   1195    ,   5205    ,   7989    ,   14085   ,0 };
            const unsigned long dim3028Kuo2Init[]       =   {   1   ,   1   ,   3   ,   11  ,   13  ,   5   ,   13  ,   35  ,   231 ,   627 ,   1155    ,   1117    ,   3889    ,   3645    ,   25099   ,0 };
            const unsigned long dim3029Kuo2Init[]       =   {   1   ,   3   ,   5   ,   3   ,   13  ,   63  ,   73  ,   83  ,   249 ,   559 ,   303 ,   3303    ,   5387    ,   15773   ,   16249   ,0 };
            const unsigned long dim3030Kuo2Init[]       =   {   1   ,   1   ,   1   ,   9   ,   21  ,   23  ,   63  ,   153 ,   291 ,   951 ,   1409    ,   1333    ,   5385    ,   14621   ,   13763   ,0 };
            const unsigned long dim3031Kuo2Init[]       =   {   1   ,   3   ,   1   ,   9   ,   1   ,   45  ,   13  ,   139 ,   437 ,   495 ,   773 ,   1401    ,   1599    ,   2555    ,   21579   ,0 };
            const unsigned long dim3032Kuo2Init[]       =   {   1   ,   3   ,   1   ,   9   ,   11  ,   15  ,   51  ,   123 ,   393 ,   223 ,   1045    ,   1001    ,   6955    ,   12755   ,   5979    ,0 };
            const unsigned long dim3033Kuo2Init[]       =   {   1   ,   1   ,   1   ,   7   ,   31  ,   31  ,   57  ,   127 ,   339 ,   133 ,   1489    ,   3055    ,   3051    ,   14931   ,   7757    ,0 };
            const unsigned long dim3034Kuo2Init[]       =   {   1   ,   3   ,   5   ,   9   ,   1   ,   23  ,   97  ,   79  ,   491 ,   641 ,   295 ,   2597    ,   3857    ,   1889    ,   16295   ,0 };
            const unsigned long dim3035Kuo2Init[]       =   {   1   ,   3   ,   7   ,   7   ,   21  ,   31  ,   111 ,   77  ,   199 ,   401 ,   1829    ,   925 ,   6023    ,   16161   ,   1783    ,0 };
            const unsigned long dim3036Kuo2Init[]       =   {   1   ,   3   ,   3   ,   1   ,   5   ,   9   ,   53  ,   133 ,   347 ,   761 ,   841 ,   841 ,   3031    ,   10201   ,   27245   ,0 };
            const unsigned long dim3037Kuo2Init[]       =   {   1   ,   1   ,   3   ,   13  ,   31  ,   35  ,   21  ,   19  ,   95  ,   861 ,   1385    ,   3607    ,   335 ,   14643   ,   20435   ,0 };
            const unsigned long dim3038Kuo2Init[]       =   {   1   ,   3   ,   3   ,   13  ,   7   ,   61  ,   47  ,   225 ,   235 ,   333 ,   1923    ,   2961    ,   61  ,   14463   ,   12329   ,0 };
            const unsigned long dim3039Kuo2Init[]       =   {   1   ,   3   ,   5   ,   5   ,   23  ,   35  ,   43  ,   175 ,   25  ,   573 ,   1253    ,   233 ,   5377    ,   3153    ,   24543   ,0 };
            const unsigned long dim3040Kuo2Init[]       =   {   1   ,   1   ,   5   ,   11  ,   27  ,   55  ,   49  ,   239 ,   479 ,   475 ,   223 ,   2947    ,   89  ,   5517    ,   2925    ,0 };
            const unsigned long dim3041Kuo2Init[]       =   {   1   ,   3   ,   5   ,   13  ,   31  ,   45  ,   11  ,   255 ,   251 ,   647 ,   1161    ,   847 ,   1175    ,   1219    ,   19735   ,0 };
            const unsigned long dim3042Kuo2Init[]       =   {   1   ,   3   ,   5   ,   1   ,   13  ,   19  ,   13  ,   31  ,   309 ,   165 ,   49  ,   2281    ,   5607    ,   11987   ,   19255   ,0 };
            const unsigned long dim3043Kuo2Init[]       =   {   1   ,   3   ,   7   ,   7   ,   15  ,   35  ,   51  ,   243 ,   119 ,   179 ,   1055    ,   3751    ,   4133    ,   14873   ,   12317   ,0 };
            const unsigned long dim3044Kuo2Init[]       =   {   1   ,   3   ,   7   ,   13  ,   13  ,   47  ,   87  ,   169 ,   203 ,   27  ,   449 ,   2755    ,   5407    ,   11857   ,   291 ,0 };
            const unsigned long dim3045Kuo2Init[]       =   {   1   ,   3   ,   3   ,   5   ,   29  ,   27  ,   1   ,   225 ,   493 ,   649 ,   1075    ,   2197    ,   2201    ,   14425   ,   18853   ,0 };
            const unsigned long dim3046Kuo2Init[]       =   {   1   ,   3   ,   7   ,   1   ,   23  ,   21  ,   87  ,   237 ,   377 ,   575 ,   445 ,   1845    ,   6689    ,   10945   ,   17517   ,0 };
            const unsigned long dim3047Kuo2Init[]       =   {   1   ,   1   ,   5   ,   5   ,   19  ,   43  ,   5   ,   177 ,   507 ,   395 ,   945 ,   2933    ,   1237    ,   10069   ,   15707   ,0 };
            const unsigned long dim3048Kuo2Init[]       =   {   1   ,   1   ,   1   ,   9   ,   31  ,   61  ,   85  ,   45  ,   325 ,   915 ,   863 ,   839 ,   2335    ,   14631   ,   32479   ,0 };
            const unsigned long dim3049Kuo2Init[]       =   {   1   ,   1   ,   3   ,   5   ,   11  ,   61  ,   47  ,   119 ,   439 ,   407 ,   203 ,   2841    ,   2205    ,   5335    ,   29721   ,0 };
            const unsigned long dim3050Kuo2Init[]       =   {   1   ,   3   ,   5   ,   11  ,   9   ,   9   ,   63  ,   179 ,   315 ,   257 ,   901 ,   5   ,   4523    ,   11085   ,   23639   ,0 };
            const unsigned long dim3051Kuo2Init[]       =   {   1   ,   1   ,   5   ,   1   ,   11  ,   9   ,   83  ,   109 ,   265 ,   601 ,   1457    ,   795 ,   509 ,   5605    ,   13235   ,0 };
            const unsigned long dim3052Kuo2Init[]       =   {   1   ,   3   ,   3   ,   13  ,   31  ,   25  ,   59  ,   37  ,   25  ,   811 ,   657 ,   467 ,   1317    ,   3783    ,   27645   ,0 };
            const unsigned long dim3053Kuo2Init[]       =   {   1   ,   1   ,   1   ,   9   ,   31  ,   51  ,   81  ,   255 ,   379 ,   575 ,   1401    ,   2341    ,   2411    ,   11893   ,   29599   ,0 };
            const unsigned long dim3054Kuo2Init[]       =   {   1   ,   1   ,   5   ,   13  ,   3   ,   47  ,   79  ,   183 ,   111 ,   409 ,   1223    ,   1305    ,   1073    ,   9251    ,   28331   ,0 };
            const unsigned long dim3055Kuo2Init[]       =   {   1   ,   3   ,   3   ,   7   ,   31  ,   9   ,   39  ,   125 ,   269 ,   999 ,   1427    ,   1845    ,   4427    ,   3825    ,   29289   ,0 };
            const unsigned long dim3056Kuo2Init[]       =   {   1   ,   1   ,   3   ,   3   ,   17  ,   41  ,   3   ,   29  ,   9   ,   253 ,   925 ,   1279    ,   2263    ,   15521   ,   25455   ,0 };
            const unsigned long dim3057Kuo2Init[]       =   {   1   ,   1   ,   7   ,   7   ,   31  ,   31  ,   53  ,   95  ,   225 ,   967 ,   893 ,   1315    ,   5683    ,   3311    ,   15259   ,0 };
            const unsigned long dim3058Kuo2Init[]       =   {   1   ,   1   ,   7   ,   3   ,   31  ,   21  ,   123 ,   145 ,   117 ,   137 ,   499 ,   2381    ,   6467    ,   15677   ,   2079    ,0 };
            const unsigned long dim3059Kuo2Init[]       =   {   1   ,   1   ,   3   ,   15  ,   3   ,   15  ,   17  ,   235 ,   463 ,   493 ,   1465    ,   1717    ,   6511    ,   5331    ,   10265   ,0 };
            const unsigned long dim3060Kuo2Init[]       =   {   1   ,   3   ,   3   ,   1   ,   25  ,   41  ,   31  ,   225 ,   115 ,   199 ,   1831    ,   29  ,   5563    ,   2697    ,   26689   ,0 };
            const unsigned long dim3061Kuo2Init[]       =   {   1   ,   3   ,   7   ,   15  ,   7   ,   23  ,   29  ,   79  ,   35  ,   493 ,   997 ,   2139    ,   3813    ,   9915    ,   16693   ,0 };
            const unsigned long dim3062Kuo2Init[]       =   {   1   ,   1   ,   7   ,   13  ,   13  ,   59  ,   71  ,   127 ,   41  ,   271 ,   1513    ,   2239    ,   6739    ,   6351    ,   11331   ,0 };
            const unsigned long dim3063Kuo2Init[]       =   {   1   ,   3   ,   5   ,   1   ,   23  ,   63  ,   109 ,   225 ,   413 ,   277 ,   927 ,   1781    ,   843 ,   5959    ,   16655   ,0 };
            const unsigned long dim3064Kuo2Init[]       =   {   1   ,   3   ,   5   ,   13  ,   9   ,   55  ,   91  ,   143 ,   401 ,   695 ,   1857    ,   205 ,   3617    ,   11847   ,   27339   ,0 };
            const unsigned long dim3065Kuo2Init[]       =   {   1   ,   1   ,   3   ,   1   ,   21  ,   41  ,   5   ,   1   ,   207 ,   247 ,   1269    ,   3071    ,   8045    ,   8711    ,   3595    ,0 };
            const unsigned long dim3066Kuo2Init[]       =   {   1   ,   3   ,   7   ,   15  ,   5   ,   23  ,   29  ,   119 ,   435 ,   961 ,   333 ,   2337    ,   3469    ,   6363    ,   29145   ,0 };
            const unsigned long dim3067Kuo2Init[]       =   {   1   ,   3   ,   5   ,   5   ,   21  ,   19  ,   65  ,   177 ,   309 ,   115 ,   707 ,   1659    ,   941 ,   6825    ,   3595    ,0 };
            const unsigned long dim3068Kuo2Init[]       =   {   1   ,   1   ,   1   ,   5   ,   29  ,   5   ,   113 ,   15  ,   13  ,   897 ,   1535    ,   3453    ,   3539    ,   3937    ,   22647   ,0 };
            const unsigned long dim3069Kuo2Init[]       =   {   1   ,   1   ,   3   ,   9   ,   13  ,   41  ,   45  ,   11  ,   259 ,   127 ,   1421    ,   195 ,   4555    ,   11643   ,   30239   ,0 };
            const unsigned long dim3070Kuo2Init[]       =   {   1   ,   3   ,   3   ,   15  ,   5   ,   17  ,   115 ,   89  ,   183 ,   729 ,   1655    ,   3599    ,   1443    ,   11799   ,   29777   ,0 };
            const unsigned long dim3071Kuo2Init[]       =   {   1   ,   3   ,   3   ,   15  ,   19  ,   57  ,   61  ,   3   ,   53  ,   221 ,   799 ,   3243    ,   5255    ,   3271    ,   8677    ,0 };
            const unsigned long dim3072Kuo2Init[]       =   {   1   ,   3   ,   1   ,   5   ,   23  ,   45  ,   103 ,   173 ,   453 ,   111 ,   1145    ,   891 ,   4989    ,   3195    ,   20129   ,0 };
            const unsigned long dim3073Kuo2Init[]       =   {   1   ,   3   ,   1   ,   13  ,   13  ,   25  ,   71  ,   71  ,   125 ,   753 ,   397 ,   613 ,   1641    ,   2611    ,   1879    ,0 };
            const unsigned long dim3074Kuo2Init[]       =   {   1   ,   1   ,   1   ,   15  ,   23  ,   59  ,   25  ,   199 ,   285 ,   161 ,   225 ,   4039    ,   5427    ,   13711   ,   23643   ,0 };
            const unsigned long dim3075Kuo2Init[]       =   {   1   ,   1   ,   7   ,   13  ,   25  ,   9   ,   119 ,   121 ,   369 ,   811 ,   1857    ,   2227    ,   5711    ,   2761    ,   23033   ,0 };
            const unsigned long dim3076Kuo2Init[]       =   {   1   ,   1   ,   3   ,   15  ,   27  ,   43  ,   75  ,   231 ,   221 ,   869 ,   603 ,   537 ,   3679    ,   15963   ,   27183   ,0 };
            const unsigned long dim3077Kuo2Init[]       =   {   1   ,   1   ,   5   ,   3   ,   9   ,   35  ,   33  ,   39  ,   383 ,   415 ,   1769    ,   3837    ,   1479    ,   3411    ,   20917   ,0 };
            const unsigned long dim3078Kuo2Init[]       =   {   1   ,   1   ,   5   ,   15  ,   7   ,   47  ,   119 ,   105 ,   31  ,   79  ,   229 ,   135 ,   1509    ,   1571    ,   24355   ,0 };
            const unsigned long dim3079Kuo2Init[]       =   {   1   ,   1   ,   5   ,   7   ,   7   ,   15  ,   99  ,   79  ,   65  ,   869 ,   1781    ,   2111    ,   2405    ,   1901    ,   8419    ,0 };
            const unsigned long dim3080Kuo2Init[]       =   {   1   ,   1   ,   3   ,   15  ,   7   ,   59  ,   13  ,   211 ,   439 ,   543 ,   755 ,   2177    ,   971 ,   6045    ,   24943   ,0 };
            const unsigned long dim3081Kuo2Init[]       =   {   1   ,   1   ,   3   ,   7   ,   13  ,   25  ,   89  ,   71  ,   453 ,   71  ,   1205    ,   1419    ,   971 ,   11139   ,   25075   ,0 };
            const unsigned long dim3082Kuo2Init[]       =   {   1   ,   3   ,   5   ,   3   ,   7   ,   63  ,   21  ,   17  ,   179 ,   749 ,   915 ,   3977    ,   5399    ,   11385   ,   13825   ,0 };
            const unsigned long dim3083Kuo2Init[]       =   {   1   ,   1   ,   5   ,   9   ,   19  ,   49  ,   77  ,   47  ,   351 ,   815 ,   1925    ,   559 ,   4503    ,   8095    ,   3115    ,0 };
            const unsigned long dim3084Kuo2Init[]       =   {   1   ,   3   ,   1   ,   9   ,   13  ,   61  ,   79  ,   245 ,   509 ,   691 ,   387 ,   2067    ,   1493    ,   10783   ,   31911   ,0 };
            const unsigned long dim3085Kuo2Init[]       =   {   1   ,   3   ,   1   ,   11  ,   31  ,   27  ,   117 ,   55  ,   93  ,   389 ,   1299    ,   573 ,   6883    ,   10449   ,   19779   ,0 };
            const unsigned long dim3086Kuo2Init[]       =   {   1   ,   3   ,   3   ,   7   ,   15  ,   27  ,   5   ,   207 ,   215 ,   773 ,   1989    ,   2341    ,   2101    ,   10585   ,   6675    ,0 };
            const unsigned long dim3087Kuo2Init[]       =   {   1   ,   3   ,   3   ,   15  ,   19  ,   37  ,   93  ,   127 ,   233 ,   49  ,   557 ,   1683    ,   7299    ,   5399    ,   21571   ,0 };
            const unsigned long dim3088Kuo2Init[]       =   {   1   ,   3   ,   1   ,   15  ,   15  ,   21  ,   95  ,   81  ,   167 ,   113 ,   245 ,   3607    ,   2007    ,   1847    ,   2317    ,0 };
            const unsigned long dim3089Kuo2Init[]       =   {   1   ,   3   ,   3   ,   9   ,   21  ,   13  ,   1   ,   203 ,   361 ,   695 ,   539 ,   3927    ,   4149    ,   249 ,   8425    ,0 };
            const unsigned long dim3090Kuo2Init[]       =   {   1   ,   3   ,   3   ,   11  ,   1   ,   61  ,   25  ,   159 ,   215 ,   235 ,   1561    ,   99  ,   2965    ,   13035   ,   11673   ,0 };
            const unsigned long dim3091Kuo2Init[]       =   {   1   ,   3   ,   1   ,   1   ,   5   ,   27  ,   13  ,   83  ,   457 ,   57  ,   995 ,   2163    ,   257 ,   12671   ,   18391   ,0 };
            const unsigned long dim3092Kuo2Init[]       =   {   1   ,   1   ,   5   ,   13  ,   23  ,   25  ,   69  ,   15  ,   53  ,   253 ,   1789    ,   3415    ,   3667    ,   3255    ,   26871   ,0 };
            const unsigned long dim3093Kuo2Init[]       =   {   1   ,   1   ,   3   ,   3   ,   9   ,   13  ,   25  ,   81  ,   247 ,   483 ,   921 ,   649 ,   4759    ,   15037   ,   4923    ,0 };
            const unsigned long dim3094Kuo2Init[]       =   {   1   ,   3   ,   5   ,   1   ,   5   ,   1   ,   111 ,   179 ,   329 ,   395 ,   1661    ,   2647    ,   6487    ,   11109   ,   15465   ,0 };
            const unsigned long dim3095Kuo2Init[]       =   {   1   ,   1   ,   3   ,   7   ,   27  ,   27  ,   13  ,   123 ,   225 ,   181 ,   1205    ,   2703    ,   1753    ,   2513    ,   1671    ,0 };
            const unsigned long dim3096Kuo2Init[]       =   {   1   ,   3   ,   7   ,   9   ,   23  ,   51  ,   1   ,   9   ,   233 ,   709 ,   257 ,   999 ,   5761    ,   3281    ,   8083    ,0 };
            const unsigned long dim3097Kuo2Init[]       =   {   1   ,   1   ,   7   ,   11  ,   31  ,   33  ,   29  ,   83  ,   251 ,   1021    ,   1395    ,   447 ,   4071    ,   13257   ,   5821    ,0 };
            const unsigned long dim3098Kuo2Init[]       =   {   1   ,   1   ,   5   ,   1   ,   11  ,   19  ,   67  ,   35  ,   371 ,   425 ,   613 ,   1065    ,   269 ,   3895    ,   19507   ,0 };
            const unsigned long dim3099Kuo2Init[]       =   {   1   ,   3   ,   5   ,   1   ,   1   ,   53  ,   79  ,   229 ,   489 ,   373 ,   1195    ,   3221    ,   7425    ,   2731    ,   16607   ,0 };
            const unsigned long dim3100Kuo2Init[]       =   {   1   ,   1   ,   3   ,   3   ,   3   ,   11  ,   49  ,   241 ,   269 ,   727 ,   549 ,   2463    ,   6983    ,   13067   ,   19829   ,0 };
            const unsigned long dim3101Kuo2Init[]       =   {   1   ,   3   ,   5   ,   15  ,   13  ,   41  ,   53  ,   215 ,   131 ,   363 ,   595 ,   2777    ,   3929    ,   1305    ,   12725   ,0 };
            const unsigned long dim3102Kuo2Init[]       =   {   1   ,   3   ,   3   ,   1   ,   5   ,   41  ,   87  ,   243 ,   485 ,   401 ,   1599    ,   397 ,   4207    ,   3707    ,   28949   ,0 };
            const unsigned long dim3103Kuo2Init[]       =   {   1   ,   1   ,   1   ,   3   ,   7   ,   5   ,   31  ,   59  ,   25  ,   391 ,   1419    ,   2355    ,   953 ,   1179    ,   22605   ,0 };
            const unsigned long dim3104Kuo2Init[]       =   {   1   ,   3   ,   3   ,   5   ,   13  ,   47  ,   3   ,   167 ,   25  ,   381 ,   341 ,   1999    ,   3423    ,   4465    ,   19517   ,0 };
            const unsigned long dim3105Kuo2Init[]       =   {   1   ,   1   ,   3   ,   3   ,   21  ,   55  ,   93  ,   93  ,   243 ,   67  ,   1743    ,   139 ,   5783    ,   2261    ,   22411   ,0 };
            const unsigned long dim3106Kuo2Init[]       =   {   1   ,   1   ,   1   ,   15  ,   25  ,   19  ,   1   ,   247 ,   325 ,   641 ,   347 ,   1095    ,   7301    ,   7079    ,   25403   ,0 };
            const unsigned long dim3107Kuo2Init[]       =   {   1   ,   3   ,   5   ,   13  ,   29  ,   53  ,   23  ,   229 ,   425 ,   311 ,   1159    ,   2241    ,   105 ,   11165   ,   28737   ,0 };
            const unsigned long dim3108Kuo2Init[]       =   {   1   ,   1   ,   7   ,   9   ,   31  ,   23  ,   3   ,   191 ,   119 ,   113 ,   543 ,   1119    ,   3375    ,   13605   ,   27753   ,0 };
            const unsigned long dim3109Kuo2Init[]       =   {   1   ,   1   ,   7   ,   15  ,   1   ,   49  ,   23  ,   193 ,   165 ,   869 ,   509 ,   1017    ,   1045    ,   13709   ,   16293   ,0 };
            const unsigned long dim3110Kuo2Init[]       =   {   1   ,   1   ,   1   ,   13  ,   3   ,   15  ,   25  ,   239 ,   423 ,   359 ,   409 ,   1075    ,   2629    ,   2291    ,   20293   ,0 };
            const unsigned long dim3111Kuo2Init[]       =   {   1   ,   1   ,   3   ,   15  ,   27  ,   35  ,   67  ,   83  ,   11  ,   1   ,   1129    ,   1917    ,   1475    ,   14831   ,   13161   ,0 };
            const unsigned long dim3112Kuo2Init[]       =   {   1   ,   1   ,   7   ,   11  ,   27  ,   21  ,   117 ,   191 ,   321 ,   721 ,   929 ,   2293    ,   2257    ,   9259    ,   4849    ,0 };
            const unsigned long dim3113Kuo2Init[]       =   {   1   ,   1   ,   1   ,   1   ,   25  ,   61  ,   21  ,   211 ,   491 ,   127 ,   1423    ,   2291    ,   5121    ,   14065   ,   14835   ,0 };
            const unsigned long dim3114Kuo2Init[]       =   {   1   ,   1   ,   7   ,   15  ,   17  ,   15  ,   81  ,   221 ,   319 ,   1015    ,   1643    ,   665 ,   7429    ,   7829    ,   5369    ,0 };
            const unsigned long dim3115Kuo2Init[]       =   {   1   ,   3   ,   7   ,   13  ,   31  ,   45  ,   77  ,   103 ,   79  ,   891 ,   43  ,   1561    ,   295 ,   3977    ,   17699   ,0 };
            const unsigned long dim3116Kuo2Init[]       =   {   1   ,   3   ,   5   ,   1   ,   11  ,   19  ,   105 ,   1   ,   179 ,   699 ,   655 ,   3125    ,   5851    ,   9577    ,   29513   ,0 };
            const unsigned long dim3117Kuo2Init[]       =   {   1   ,   3   ,   7   ,   13  ,   1   ,   15  ,   105 ,   239 ,   113 ,   917 ,   219 ,   1857    ,   801 ,   11703   ,   20955   ,0 };
            const unsigned long dim3118Kuo2Init[]       =   {   1   ,   1   ,   1   ,   9   ,   3   ,   35  ,   69  ,   43  ,   133 ,   455 ,   471 ,   1785    ,   4659    ,   6325    ,   3183    ,0 };
            const unsigned long dim3119Kuo2Init[]       =   {   1   ,   1   ,   5   ,   3   ,   7   ,   57  ,   127 ,   13  ,   151 ,   359 ,   135 ,   2445    ,   1795    ,   3257    ,   32305   ,0 };
            const unsigned long dim3120Kuo2Init[]       =   {   1   ,   1   ,   3   ,   9   ,   11  ,   43  ,   51  ,   109 ,   323 ,   789 ,   1181    ,   1577    ,   4445    ,   1623    ,   28489   ,0 };
            const unsigned long dim3121Kuo2Init[]       =   {   1   ,   3   ,   7   ,   5   ,   19  ,   55  ,   71  ,   65  ,   35  ,   605 ,   165 ,   1573    ,   1429    ,   7123    ,   1967    ,0 };
            const unsigned long dim3122Kuo2Init[]       =   {   1   ,   1   ,   1   ,   15  ,   15  ,   27  ,   87  ,   67  ,   503 ,   833 ,   1841    ,   2639    ,   5629    ,   5511    ,   12393   ,0 };
            const unsigned long dim3123Kuo2Init[]       =   {   1   ,   3   ,   1   ,   11  ,   9   ,   59  ,   85  ,   59  ,   289 ,   679 ,   1957    ,   399 ,   7949    ,   4735    ,   709 ,0 };
            const unsigned long dim3124Kuo2Init[]       =   {   1   ,   3   ,   5   ,   7   ,   3   ,   37  ,   109 ,   3   ,   33  ,   971 ,   123 ,   1585    ,   3273    ,   8421    ,   20903   ,0 };
            const unsigned long dim3125Kuo2Init[]       =   {   1   ,   1   ,   1   ,   3   ,   17  ,   37  ,   99  ,   3   ,   405 ,   579 ,   755 ,   1337    ,   3649    ,   8217    ,   11011   ,0 };
            const unsigned long dim3126Kuo2Init[]       =   {   1   ,   1   ,   7   ,   9   ,   25  ,   59  ,   87  ,   115 ,   237 ,   749 ,   355 ,   1781    ,   5199    ,   7019    ,   10615   ,0 };
            const unsigned long dim3127Kuo2Init[]       =   {   1   ,   1   ,   3   ,   11  ,   23  ,   5   ,   105 ,   129 ,   453 ,   813 ,   1423    ,   2703    ,   6241    ,   14969   ,   22943   ,0 };
            const unsigned long dim3128Kuo2Init[]       =   {   1   ,   1   ,   5   ,   7   ,   21  ,   41  ,   89  ,   129 ,   473 ,   139 ,   595 ,   925 ,   3915    ,   12059   ,   19675   ,0 };
            const unsigned long dim3129Kuo2Init[]       =   {   1   ,   1   ,   5   ,   3   ,   17  ,   19  ,   83  ,   19  ,   191 ,   757 ,   1185    ,   1849    ,   3403    ,   779 ,   24773   ,0 };
            const unsigned long dim3130Kuo2Init[]       =   {   1   ,   1   ,   5   ,   11  ,   13  ,   31  ,   121 ,   211 ,   357 ,   487 ,   829 ,   1215    ,   1   ,   15103   ,   10835   ,0 };
            const unsigned long dim3131Kuo2Init[]       =   {   1   ,   1   ,   5   ,   15  ,   1   ,   15  ,   105 ,   137 ,   169 ,   295 ,   993 ,   261 ,   6787    ,   16339   ,   20307   ,0 };
            const unsigned long dim3132Kuo2Init[]       =   {   1   ,   1   ,   5   ,   15  ,   21  ,   11  ,   53  ,   149 ,   443 ,   285 ,   371 ,   2541    ,   7041    ,   5745    ,   12127   ,0 };
            const unsigned long dim3133Kuo2Init[]       =   {   1   ,   3   ,   5   ,   1   ,   9   ,   23  ,   89  ,   77  ,   337 ,   623 ,   329 ,   1021    ,   2013    ,   11023   ,   27045   ,0 };
            const unsigned long dim3134Kuo2Init[]       =   {   1   ,   1   ,   7   ,   7   ,   1   ,   11  ,   35  ,   67  ,   93  ,   121 ,   1155    ,   1063    ,   4997    ,   9567    ,   24023   ,0 };
            const unsigned long dim3135Kuo2Init[]       =   {   1   ,   1   ,   7   ,   7   ,   3   ,   33  ,   119 ,   243 ,   423 ,   615 ,   721 ,   3057    ,   7873    ,   9329    ,   20783   ,0 };
            const unsigned long dim3136Kuo2Init[]       =   {   1   ,   3   ,   3   ,   15  ,   3   ,   23  ,   41  ,   121 ,   111 ,   483 ,   1607    ,   181 ,   895 ,   12923   ,   21707   ,0 };
            const unsigned long dim3137Kuo2Init[]       =   {   1   ,   3   ,   7   ,   5   ,   9   ,   63  ,   65  ,   87  ,   235 ,   915 ,   1569    ,   3421    ,   1373    ,   2159    ,   353 ,0 };
            const unsigned long dim3138Kuo2Init[]       =   {   1   ,   1   ,   7   ,   7   ,   29  ,   3   ,   35  ,   193 ,   95  ,   977 ,   1405    ,   23  ,   1135    ,   5047    ,   16429   ,0 };
            const unsigned long dim3139Kuo2Init[]       =   {   1   ,   1   ,   7   ,   3   ,   7   ,   57  ,   123 ,   111 ,   33  ,   781 ,   613 ,   3423    ,   4701    ,   8727    ,   22665   ,0 };
            const unsigned long dim3140Kuo2Init[]       =   {   1   ,   1   ,   1   ,   11  ,   13  ,   43  ,   101 ,   83  ,   447 ,   503 ,   1955    ,   2125    ,   4469    ,   14937   ,   24067   ,0 };
            const unsigned long dim3141Kuo2Init[]       =   {   1   ,   1   ,   5   ,   15  ,   7   ,   41  ,   115 ,   17  ,   335 ,   497 ,   1753    ,   1131    ,   4623    ,   11483   ,   5181    ,0 };
            const unsigned long dim3142Kuo2Init[]       =   {   1   ,   1   ,   7   ,   9   ,   19  ,   37  ,   1   ,   89  ,   117 ,   211 ,   1203    ,   2469    ,   5989    ,   11429   ,   31597   ,0 };
            const unsigned long dim3143Kuo2Init[]       =   {   1   ,   3   ,   3   ,   11  ,   3   ,   7   ,   125 ,   97  ,   203 ,   935 ,   533 ,   2607    ,   6075    ,   12995   ,   9997    ,0 };
            const unsigned long dim3144Kuo2Init[]       =   {   1   ,   1   ,   7   ,   1   ,   27  ,   25  ,   53  ,   99  ,   243 ,   729 ,   1521    ,   875 ,   2363    ,   3309    ,   22621   ,0 };
            const unsigned long dim3145Kuo2Init[]       =   {   1   ,   3   ,   3   ,   1   ,   21  ,   7   ,   79  ,   105 ,   363 ,   227 ,   9   ,   2445    ,   3021    ,   15177   ,   19029   ,0 };
            const unsigned long dim3146Kuo2Init[]       =   {   1   ,   3   ,   7   ,   7   ,   13  ,   53  ,   55  ,   21  ,   137 ,   953 ,   1583    ,   3439    ,   2995    ,   4447    ,   19881   ,0 };
            const unsigned long dim3147Kuo2Init[]       =   {   1   ,   1   ,   7   ,   11  ,   29  ,   21  ,   83  ,   195 ,   387 ,   133 ,   1135    ,   1425    ,   4655    ,   9171    ,   19217   ,0 };
            const unsigned long dim3148Kuo2Init[]       =   {   1   ,   3   ,   5   ,   7   ,   3   ,   19  ,   45  ,   249 ,   451 ,   797 ,   693 ,   3511    ,   7213    ,   12851   ,   31915   ,0 };
            const unsigned long dim3149Kuo2Init[]       =   {   1   ,   3   ,   3   ,   15  ,   17  ,   35  ,   119 ,   153 ,   493 ,   815 ,   1621    ,   2305    ,   7963    ,   6405    ,   6253    ,0 };
            const unsigned long dim3150Kuo2Init[]       =   {   1   ,   1   ,   3   ,   15  ,   19  ,   23  ,   111 ,   223 ,   469 ,   1015    ,   799 ,   1453    ,   7639    ,   2307    ,   20595   ,0 };
            const unsigned long dim3151Kuo2Init[]       =   {   1   ,   1   ,   5   ,   9   ,   9   ,   49  ,   33  ,   233 ,   3   ,   531 ,   1723    ,   845 ,   3585    ,   13993   ,   11955   ,0 };
            const unsigned long dim3152Kuo2Init[]       =   {   1   ,   1   ,   7   ,   7   ,   1   ,   7   ,   3   ,   217 ,   191 ,   1   ,   2013    ,   3143    ,   5919    ,   3285    ,   10733   ,0 };
            const unsigned long dim3153Kuo2Init[]       =   {   1   ,   1   ,   3   ,   15  ,   25  ,   9   ,   93  ,   127 ,   471 ,   51  ,   1227    ,   341 ,   3455    ,   11015   ,   21427   ,0 };
            const unsigned long dim3154Kuo2Init[]       =   {   1   ,   1   ,   5   ,   9   ,   23  ,   45  ,   9   ,   1   ,   445 ,   927 ,   661 ,   829 ,   411 ,   13531   ,   30903   ,0 };
            const unsigned long dim3155Kuo2Init[]       =   {   1   ,   3   ,   3   ,   5   ,   21  ,   25  ,   17  ,   225 ,   43  ,   209 ,   659 ,   3681    ,   3281    ,   10135   ,   6121    ,0 };
            const unsigned long dim3156Kuo2Init[]       =   {   1   ,   3   ,   1   ,   7   ,   17  ,   33  ,   103 ,   209 ,   187 ,   759 ,   617 ,   2971    ,   3563    ,   12925   ,   23503   ,0 };
            const unsigned long dim3157Kuo2Init[]       =   {   1   ,   3   ,   5   ,   13  ,   21  ,   13  ,   35  ,   65  ,   507 ,   95  ,   695 ,   961 ,   7589    ,   9633    ,   29065   ,0 };
            const unsigned long dim3158Kuo2Init[]       =   {   1   ,   3   ,   5   ,   5   ,   13  ,   17  ,   11  ,   145 ,   11  ,   115 ,   1449    ,   2361    ,   2959    ,   6713    ,   15345   ,0 };
            const unsigned long dim3159Kuo2Init[]       =   {   1   ,   1   ,   3   ,   9   ,   21  ,   19  ,   15  ,   7   ,   303 ,   995 ,   1245    ,   349 ,   4117    ,   10993   ,   20627   ,0 };
            const unsigned long dim3160Kuo2Init[]       =   {   1   ,   3   ,   1   ,   3   ,   1   ,   9   ,   21  ,   43  ,   259 ,   725 ,   1735    ,   1515    ,   3465    ,   10867   ,   14895   ,0 };
            const unsigned long dim3161Kuo2Init[]       =   {   1   ,   1   ,   1   ,   3   ,   7   ,   25  ,   57  ,   29  ,   435 ,   663 ,   383 ,   1197    ,   981 ,   4943    ,   24527   ,0 };
            const unsigned long dim3162Kuo2Init[]       =   {   1   ,   1   ,   3   ,   13  ,   3   ,   29  ,   43  ,   149 ,   351 ,   293 ,   1549    ,   3177    ,   6771    ,   7291    ,   23227   ,0 };
            const unsigned long dim3163Kuo2Init[]       =   {   1   ,   1   ,   1   ,   9   ,   29  ,   53  ,   9   ,   43  ,   369 ,   763 ,   1269    ,   595 ,   6559    ,   13799   ,   27751   ,0 };
            const unsigned long dim3164Kuo2Init[]       =   {   1   ,   1   ,   3   ,   3   ,   7   ,   53  ,   79  ,   167 ,   121 ,   761 ,   1015    ,   1299    ,   1523    ,   55  ,   4431    ,0 };
            const unsigned long dim3165Kuo2Init[]       =   {   1   ,   3   ,   1   ,   15  ,   19  ,   9   ,   103 ,   17  ,   121 ,   621 ,   765 ,   2225    ,   4123    ,   815 ,   4305    ,0 };
            const unsigned long dim3166Kuo2Init[]       =   {   1   ,   3   ,   7   ,   7   ,   29  ,   59  ,   43  ,   73  ,   505 ,   755 ,   617 ,   2113    ,   5605    ,   11887   ,   10211   ,0 };
            const unsigned long dim3167Kuo2Init[]       =   {   1   ,   1   ,   5   ,   15  ,   23  ,   47  ,   27  ,   241 ,   287 ,   649 ,   1601    ,   2461    ,   7851    ,   7901    ,   24305   ,0 };
            const unsigned long dim3168Kuo2Init[]       =   {   1   ,   3   ,   1   ,   3   ,   9   ,   31  ,   43  ,   103 ,   227 ,   381 ,   813 ,   3871    ,   2293    ,   5317    ,   4907    ,0 };
            const unsigned long dim3169Kuo2Init[]       =   {   1   ,   3   ,   7   ,   1   ,   29  ,   41  ,   49  ,   75  ,   389 ,   799 ,   1283    ,   1823    ,   6907    ,   11715   ,   24101   ,0 };
            const unsigned long dim3170Kuo2Init[]       =   {   1   ,   1   ,   3   ,   9   ,   1   ,   17  ,   53  ,   61  ,   133 ,   485 ,   317 ,   1779    ,   2437    ,   6195    ,   4745    ,0 };
            const unsigned long dim3171Kuo2Init[]       =   {   1   ,   3   ,   7   ,   11  ,   1   ,   55  ,   79  ,   147 ,   511 ,   29  ,   495 ,   1145    ,   677 ,   8581    ,   9491    ,0 };
            const unsigned long dim3172Kuo2Init[]       =   {   1   ,   1   ,   1   ,   7   ,   15  ,   23  ,   51  ,   243 ,   139 ,   605 ,   2035    ,   1039    ,   1135    ,   10747   ,   28225   ,0 };
            const unsigned long dim3173Kuo2Init[]       =   {   1   ,   3   ,   3   ,   1   ,   15  ,   5   ,   121 ,   73  ,   403 ,   861 ,   1587    ,   2985    ,   2217    ,   1969    ,   10459   ,0 };
            const unsigned long dim3174Kuo2Init[]       =   {   1   ,   1   ,   3   ,   3   ,   15  ,   45  ,   123 ,   235 ,   85  ,   973 ,   915 ,   2217    ,   2405    ,   4473    ,   8119    ,0 };
            const unsigned long dim3175Kuo2Init[]       =   {   1   ,   1   ,   3   ,   3   ,   29  ,   29  ,   9   ,   47  ,   173 ,   121 ,   861 ,   1081    ,   917 ,   9757    ,   25923   ,0 };
            const unsigned long dim3176Kuo2Init[]       =   {   1   ,   1   ,   3   ,   7   ,   25  ,   59  ,   57  ,   35  ,   291 ,   529 ,   1069    ,   4037    ,   4865    ,   5047    ,   1133    ,0 };
            const unsigned long dim3177Kuo2Init[]       =   {   1   ,   1   ,   5   ,   7   ,   5   ,   25  ,   49  ,   171 ,   363 ,   427 ,   1807    ,   1903    ,   6455    ,   12201   ,   12043   ,0 };
            const unsigned long dim3178Kuo2Init[]       =   {   1   ,   1   ,   3   ,   3   ,   23  ,   13  ,   77  ,   61  ,   415 ,   307 ,   1237    ,   2785    ,   1133    ,   14337   ,   25755   ,0 };
            const unsigned long dim3179Kuo2Init[]       =   {   1   ,   1   ,   1   ,   3   ,   3   ,   49  ,   95  ,   117 ,   409 ,   641 ,   393 ,   3205    ,   5047    ,   4795    ,   22773   ,0 };
            const unsigned long dim3180Kuo2Init[]       =   {   1   ,   3   ,   7   ,   5   ,   15  ,   25  ,   39  ,   37  ,   15  ,   735 ,   1107    ,   2467    ,   8037    ,   6395    ,   26043   ,0 };
            const unsigned long dim3181Kuo2Init[]       =   {   1   ,   1   ,   3   ,   7   ,   15  ,   51  ,   53  ,   211 ,   223 ,   165 ,   217 ,   3519    ,   6973    ,   823 ,   15733   ,0 };
            const unsigned long dim3182Kuo2Init[]       =   {   1   ,   1   ,   3   ,   11  ,   25  ,   15  ,   57  ,   9   ,   373 ,   445 ,   369 ,   2199    ,   2793    ,   12279   ,   7413    ,0 };
            const unsigned long dim3183Kuo2Init[]       =   {   1   ,   1   ,   3   ,   1   ,   29  ,   13  ,   37  ,   91  ,   429 ,   5   ,   467 ,   813 ,   2375    ,   10467   ,   10373   ,0 };
            const unsigned long dim3184Kuo2Init[]       =   {   1   ,   1   ,   7   ,   15  ,   1   ,   21  ,   65  ,   145 ,   341 ,   839 ,   467 ,   1029    ,   3127    ,   11519   ,   6711    ,0 };
            const unsigned long dim3185Kuo2Init[]       =   {   1   ,   1   ,   3   ,   11  ,   29  ,   55  ,   115 ,   65  ,   501 ,   375 ,   1669    ,   2227    ,   125 ,   8059    ,   15489   ,0 };
            const unsigned long dim3186Kuo2Init[]       =   {   1   ,   3   ,   1   ,   3   ,   7   ,   9   ,   71  ,   179 ,   203 ,   1017    ,   107 ,   49  ,   1093    ,   4847    ,   14383   ,0 };
            const unsigned long dim3187Kuo2Init[]       =   {   1   ,   3   ,   7   ,   13  ,   9   ,   57  ,   75  ,   139 ,   67  ,   587 ,   1681    ,   1717    ,   459 ,   7835    ,   317 ,0 };
            const unsigned long dim3188Kuo2Init[]       =   {   1   ,   3   ,   1   ,   1   ,   13  ,   15  ,   37  ,   123 ,   233 ,   467 ,   1425    ,   2119    ,   2589    ,   10087   ,   31997   ,0 };
            const unsigned long dim3189Kuo2Init[]       =   {   1   ,   3   ,   7   ,   9   ,   29  ,   53  ,   119 ,   19  ,   387 ,   225 ,   1845    ,   2849    ,   181 ,   1991    ,   10323   ,0 };
            const unsigned long dim3190Kuo2Init[]       =   {   1   ,   3   ,   5   ,   11  ,   1   ,   61  ,   5   ,   221 ,   389 ,   595 ,   499 ,   2103    ,   5967    ,   4677    ,   17551   ,0 };
            const unsigned long dim3191Kuo2Init[]       =   {   1   ,   1   ,   7   ,   5   ,   31  ,   9   ,   63  ,   121 ,   505 ,   997 ,   939 ,   7   ,   7809    ,   11337   ,   11825   ,0 };
            const unsigned long dim3192Kuo2Init[]       =   {   1   ,   1   ,   1   ,   7   ,   27  ,   35  ,   99  ,   145 ,   401 ,   887 ,   657 ,   1023    ,   6215    ,   15775   ,   13477   ,0 };
            const unsigned long dim3193Kuo2Init[]       =   {   1   ,   3   ,   3   ,   15  ,   7   ,   39  ,   35  ,   63  ,   395 ,   621 ,   1551    ,   2061    ,   7817    ,   6159    ,   13765   ,0 };
            const unsigned long dim3194Kuo2Init[]       =   {   1   ,   1   ,   7   ,   3   ,   13  ,   5   ,   17  ,   19  ,   469 ,   541 ,   707 ,   1697    ,   699 ,   5031    ,   4121    ,0 };
            const unsigned long dim3195Kuo2Init[]       =   {   1   ,   3   ,   3   ,   1   ,   31  ,   7   ,   21  ,   15  ,   137 ,   151 ,   599 ,   857 ,   6267    ,   4725    ,   20317   ,0 };
            const unsigned long dim3196Kuo2Init[]       =   {   1   ,   1   ,   7   ,   9   ,   23  ,   13  ,   59  ,   169 ,   73  ,   517 ,   1209    ,   2063    ,   1881    ,   849 ,   3263    ,0 };
            const unsigned long dim3197Kuo2Init[]       =   {   1   ,   1   ,   1   ,   9   ,   3   ,   61  ,   115 ,   75  ,   161 ,   395 ,   1889    ,   1229    ,   6153    ,   597 ,   5667    ,0 };
            const unsigned long dim3198Kuo2Init[]       =   {   1   ,   1   ,   3   ,   11  ,   5   ,   31  ,   37  ,   163 ,   469 ,   237 ,   91  ,   3531    ,   7197    ,   12673   ,   13741   ,0 };
            const unsigned long dim3199Kuo2Init[]       =   {   1   ,   1   ,   3   ,   1   ,   17  ,   61  ,   31  ,   29  ,   3   ,   855 ,   1703    ,   2091    ,   2791    ,   1963    ,   7385    ,0 };
            const unsigned long dim3200Kuo2Init[]       =   {   1   ,   1   ,   1   ,   5   ,   13  ,   53  ,   105 ,   239 ,   143 ,   199 ,   2003    ,   4005    ,   1911    ,   4965    ,   31675   ,0 };
            const unsigned long dim3201Kuo2Init[]       =   {   1   ,   3   ,   7   ,   3   ,   7   ,   37  ,   73  ,   85  ,   219 ,   165 ,   887 ,   243 ,   117 ,   767 ,   5519    ,0 };
            const unsigned long dim3202Kuo2Init[]       =   {   1   ,   1   ,   7   ,   9   ,   21  ,   3   ,   89  ,   13  ,   261 ,   835 ,   529 ,   3851    ,   4905    ,   2881    ,   3889    ,0 };
            const unsigned long dim3203Kuo2Init[]       =   {   1   ,   1   ,   7   ,   3   ,   15  ,   41  ,   21  ,   89  ,   137 ,   959 ,   839 ,   4087    ,   1689    ,   11347   ,   17701   ,0 };
            const unsigned long dim3204Kuo2Init[]       =   {   1   ,   3   ,   7   ,   3   ,   15  ,   17  ,   63  ,   31  ,   493 ,   509 ,   1079    ,   3223    ,   7649    ,   4377    ,   20291   ,0 };
            const unsigned long dim3205Kuo2Init[]       =   {   1   ,   3   ,   5   ,   5   ,   5   ,   63  ,   101 ,   83  ,   431 ,   201 ,   1965    ,   1227    ,   2161    ,   14661   ,   2027    ,0 };
            const unsigned long dim3206Kuo2Init[]       =   {   1   ,   1   ,   7   ,   3   ,   29  ,   19  ,   81  ,   245 ,   125 ,   795 ,   181 ,   4017    ,   4061    ,   12909   ,   30591   ,0 };
            const unsigned long dim3207Kuo2Init[]       =   {   1   ,   1   ,   5   ,   11  ,   19  ,   63  ,   49  ,   205 ,   485 ,   789 ,   1123    ,   715 ,   6553    ,   5637    ,   7977    ,0 };
            const unsigned long dim3208Kuo2Init[]       =   {   1   ,   3   ,   3   ,   7   ,   13  ,   63  ,   113 ,   63  ,   187 ,   515 ,   725 ,   3583    ,   1699    ,   5165    ,   21823   ,0 };
            const unsigned long dim3209Kuo2Init[]       =   {   1   ,   3   ,   1   ,   3   ,   27  ,   17  ,   17  ,   241 ,   205 ,   593 ,   697 ,   4023    ,   7603    ,   15421   ,   9679    ,0 };
            const unsigned long dim3210Kuo2Init[]       =   {   1   ,   3   ,   1   ,   1   ,   3   ,   15  ,   39  ,   209 ,   53  ,   405 ,   601 ,   3315    ,   5487    ,   12163   ,   14741   ,0 };
            const unsigned long dim3211Kuo2Init[]       =   {   1   ,   1   ,   3   ,   13  ,   31  ,   9   ,   9   ,   159 ,   187 ,   307 ,   1115    ,   381 ,   7763    ,   4231    ,   2387    ,0 };
            const unsigned long dim3212Kuo2Init[]       =   {   1   ,   1   ,   7   ,   13  ,   1   ,   41  ,   17  ,   101 ,   321 ,   473 ,   515 ,   1959    ,   7589    ,   9323    ,   23947   ,0 };
            const unsigned long dim3213Kuo2Init[]       =   {   1   ,   3   ,   3   ,   9   ,   21  ,   55  ,   55  ,   137 ,   115 ,   791 ,   1411    ,   3543    ,   8173    ,   14403   ,   28019   ,0 };
            const unsigned long dim3214Kuo2Init[]       =   {   1   ,   3   ,   3   ,   7   ,   29  ,   35  ,   83  ,   119 ,   243 ,   561 ,   423 ,   2435    ,   7613    ,   1081    ,   29911   ,0 };
            const unsigned long dim3215Kuo2Init[]       =   {   1   ,   3   ,   1   ,   15  ,   13  ,   21  ,   9   ,   205 ,   3   ,   901 ,   1305    ,   4015    ,   4045    ,   11685   ,   489 ,0 };
            const unsigned long dim3216Kuo2Init[]       =   {   1   ,   1   ,   5   ,   13  ,   27  ,   57  ,   73  ,   173 ,   391 ,   259 ,   757 ,   3177    ,   3767    ,   5251    ,   12381   ,0 };
            const unsigned long dim3217Kuo2Init[]       =   {   1   ,   1   ,   3   ,   1   ,   3   ,   15  ,   121 ,   1   ,   85  ,   795 ,   1465    ,   1247    ,   5703    ,   3399    ,   28519   ,0 };
            const unsigned long dim3218Kuo2Init[]       =   {   1   ,   1   ,   1   ,   7   ,   27  ,   7   ,   41  ,   147 ,   277 ,   317 ,   1161    ,   2205    ,   1845    ,   12379   ,   9121    ,0 };
            const unsigned long dim3219Kuo2Init[]       =   {   1   ,   3   ,   3   ,   9   ,   5   ,   59  ,   107 ,   115 ,   13  ,   469 ,   595 ,   3979    ,   7563    ,   521 ,   6407    ,0 };
            const unsigned long dim3220Kuo2Init[]       =   {   1   ,   1   ,   7   ,   9   ,   31  ,   43  ,   107 ,   107 ,   49  ,   67  ,   475 ,   1371    ,   3205    ,   2035    ,   6349    ,0 };
            const unsigned long dim3221Kuo2Init[]       =   {   1   ,   3   ,   1   ,   3   ,   19  ,   33  ,   73  ,   171 ,   493 ,   473 ,   1963    ,   2453    ,   2833    ,   6847    ,   9899    ,0 };
            const unsigned long dim3222Kuo2Init[]       =   {   1   ,   3   ,   1   ,   1   ,   31  ,   57  ,   33  ,   201 ,   431 ,   623 ,   327 ,   481 ,   2377    ,   6533    ,   30339   ,0 };
            const unsigned long dim3223Kuo2Init[]       =   {   1   ,   1   ,   7   ,   1   ,   31  ,   11  ,   47  ,   183 ,   209 ,   661 ,   357 ,   3809    ,   4129    ,   5465    ,   29141   ,0 };
            const unsigned long dim3224Kuo2Init[]       =   {   1   ,   1   ,   3   ,   3   ,   19  ,   43  ,   99  ,   45  ,   471 ,   983 ,   627 ,   2397    ,   753 ,   6735    ,   24945   ,0 };
            const unsigned long dim3225Kuo2Init[]       =   {   1   ,   3   ,   5   ,   11  ,   3   ,   31  ,   127 ,   255 ,   175 ,   549 ,   1743    ,   3417    ,   3753    ,   11497   ,   5273    ,0 };
            const unsigned long dim3226Kuo2Init[]       =   {   1   ,   3   ,   5   ,   7   ,   5   ,   17  ,   77  ,   99  ,   109 ,   973 ,   93  ,   1287    ,   1781    ,   6055    ,   6987    ,0 };
            const unsigned long dim3227Kuo2Init[]       =   {   1   ,   1   ,   7   ,   5   ,   17  ,   31  ,   23  ,   231 ,   63  ,   315 ,   1119    ,   2175    ,   5807    ,   6497    ,   12991   ,0 };
            const unsigned long dim3228Kuo2Init[]       =   {   1   ,   3   ,   5   ,   1   ,   25  ,   17  ,   51  ,   81  ,   253 ,   725 ,   1351    ,   1941    ,   3893    ,   13059   ,   27373   ,0 };
            const unsigned long dim3229Kuo2Init[]       =   {   1   ,   3   ,   3   ,   7   ,   17  ,   43  ,   35  ,   151 ,   127 ,   587 ,   229 ,   2253    ,   4299    ,   10269   ,   169 ,0 };
            const unsigned long dim3230Kuo2Init[]       =   {   1   ,   1   ,   5   ,   11  ,   15  ,   11  ,   65  ,   31  ,   263 ,   897 ,   1345    ,   3011    ,   6767    ,   15851   ,   16801   ,0 };
            const unsigned long dim3231Kuo2Init[]       =   {   1   ,   3   ,   3   ,   15  ,   15  ,   1   ,   119 ,   235 ,   57  ,   339 ,   503 ,   2633    ,   1307    ,   14387   ,   28759   ,0 };
            const unsigned long dim3232Kuo2Init[]       =   {   1   ,   3   ,   1   ,   5   ,   5   ,   57  ,   111 ,   117 ,   273 ,   703 ,   1363    ,   1579    ,   97  ,   15323   ,   24897   ,0 };
            const unsigned long dim3233Kuo2Init[]       =   {   1   ,   3   ,   5   ,   13  ,   23  ,   29  ,   43  ,   171 ,   347 ,   651 ,   869 ,   3623    ,   1643    ,   14647   ,   16271   ,0 };
            const unsigned long dim3234Kuo2Init[]       =   {   1   ,   3   ,   5   ,   15  ,   11  ,   35  ,   41  ,   121 ,   117 ,   67  ,   1649    ,   1201    ,   213 ,   13013   ,   20403   ,0 };
            const unsigned long dim3235Kuo2Init[]       =   {   1   ,   3   ,   5   ,   7   ,   3   ,   33  ,   61  ,   97  ,   439 ,   371 ,   117 ,   1341    ,   6513    ,   11043   ,   10055   ,0 };
            const unsigned long dim3236Kuo2Init[]       =   {   1   ,   3   ,   3   ,   1   ,   25  ,   9   ,   25  ,   13  ,   121 ,   593 ,   1879    ,   2487    ,   385 ,   4341    ,   11515   ,0 };
            const unsigned long dim3237Kuo2Init[]       =   {   1   ,   3   ,   1   ,   7   ,   11  ,   23  ,   65  ,   181 ,   205 ,   381 ,   1749    ,   3837    ,   5343    ,   12389   ,   27243   ,0 };
            const unsigned long dim3238Kuo2Init[]       =   {   1   ,   3   ,   3   ,   15  ,   19  ,   9   ,   11  ,   185 ,   109 ,   49  ,   311 ,   1225    ,   7445    ,   14459   ,   12631   ,0 };
            const unsigned long dim3239Kuo2Init[]       =   {   1   ,   3   ,   7   ,   7   ,   27  ,   3   ,   87  ,   179 ,   467 ,   823 ,   1641    ,   189 ,   3489    ,   3203    ,   10447   ,0 };
            const unsigned long dim3240Kuo2Init[]       =   {   1   ,   3   ,   5   ,   3   ,   5   ,   15  ,   117 ,   139 ,   39  ,   497 ,   1137    ,   1193    ,   2281    ,   12945   ,   1329    ,0 };
            const unsigned long dim3241Kuo2Init[]       =   {   1   ,   3   ,   7   ,   15  ,   21  ,   9   ,   111 ,   115 ,   461 ,   987 ,   1091    ,   2363    ,   5453    ,   3929    ,   26887   ,0 };
            const unsigned long dim3242Kuo2Init[]       =   {   1   ,   1   ,   1   ,   11  ,   19  ,   25  ,   113 ,   65  ,   301 ,   491 ,   1679    ,   4007    ,   4709    ,   4547    ,   25155   ,0 };
            const unsigned long dim3243Kuo2Init[]       =   {   1   ,   3   ,   1   ,   15  ,   29  ,   23  ,   87  ,   209 ,   471 ,   247 ,   633 ,   2533    ,   1077    ,   5079    ,   10311   ,0 };
            const unsigned long dim3244Kuo2Init[]       =   {   1   ,   3   ,   3   ,   1   ,   27  ,   21  ,   27  ,   57  ,   69  ,   883 ,   145 ,   1559    ,   6751    ,   873 ,   28941   ,0 };
            const unsigned long dim3245Kuo2Init[]       =   {   1   ,   1   ,   7   ,   5   ,   23  ,   63  ,   113 ,   57  ,   243 ,   133 ,   1159    ,   3837    ,   4781    ,   6029    ,   391 ,0 };
            const unsigned long dim3246Kuo2Init[]       =   {   1   ,   1   ,   1   ,   7   ,   27  ,   47  ,   33  ,   211 ,   317 ,   891 ,   1955    ,   3201    ,   803 ,   1645    ,   26669   ,0 };
            const unsigned long dim3247Kuo2Init[]       =   {   1   ,   1   ,   7   ,   9   ,   17  ,   5   ,   69  ,   233 ,   491 ,   489 ,   49  ,   1395    ,   1907    ,   10189   ,   17853   ,0 };
            const unsigned long dim3248Kuo2Init[]       =   {   1   ,   1   ,   7   ,   13  ,   9   ,   29  ,   25  ,   169 ,   469 ,   1015    ,   1729    ,   3095    ,   281 ,   15537   ,   32565   ,0 };
            const unsigned long dim3249Kuo2Init[]       =   {   1   ,   1   ,   1   ,   7   ,   23  ,   63  ,   55  ,   65  ,   1   ,   913 ,   343 ,   3749    ,   3085    ,   13557   ,   32121   ,0 };
            const unsigned long dim3250Kuo2Init[]       =   {   1   ,   3   ,   7   ,   13  ,   15  ,   29  ,   15  ,   141 ,   113 ,   315 ,   645 ,   821 ,   5083    ,   12141   ,   1603    ,0 };
            const unsigned long dim3251Kuo2Init[]       =   {   1   ,   3   ,   5   ,   11  ,   19  ,   7   ,   73  ,   245 ,   139 ,   923 ,   731 ,   211 ,   7223    ,   8179    ,   20811   ,0 };
            const unsigned long dim3252Kuo2Init[]       =   {   1   ,   3   ,   7   ,   11  ,   31  ,   21  ,   35  ,   253 ,   373 ,   889 ,   1671    ,   1069    ,   3475    ,   1947    ,   16129   ,0 };
            const unsigned long dim3253Kuo2Init[]       =   {   1   ,   1   ,   1   ,   11  ,   17  ,   11  ,   3   ,   223 ,   169 ,   577 ,   2045    ,   2591    ,   2593    ,   10911   ,   20657   ,0 };
            const unsigned long dim3254Kuo2Init[]       =   {   1   ,   3   ,   3   ,   5   ,   21  ,   57  ,   121 ,   185 ,   329 ,   581 ,   471 ,   1799    ,   393 ,   1059    ,   19339   ,0 };
            const unsigned long dim3255Kuo2Init[]       =   {   1   ,   1   ,   3   ,   3   ,   17  ,   63  ,   39  ,   227 ,   447 ,   157 ,   1803    ,   239 ,   93  ,   5297    ,   801 ,0 };
            const unsigned long dim3256Kuo2Init[]       =   {   1   ,   3   ,   5   ,   5   ,   23  ,   47  ,   127 ,   211 ,   415 ,   943 ,   757 ,   213 ,   5071    ,   5811    ,   2135    ,0 };
            const unsigned long dim3257Kuo2Init[]       =   {   1   ,   3   ,   1   ,   5   ,   7   ,   63  ,   95  ,   83  ,   307 ,   241 ,   511 ,   1417    ,   2141    ,   2599    ,   1665    ,0 };
            const unsigned long dim3258Kuo2Init[]       =   {   1   ,   1   ,   7   ,   1   ,   17  ,   31  ,   113 ,   53  ,   27  ,   611 ,   45  ,   1909    ,   6519    ,   5687    ,   23335   ,0 };
            const unsigned long dim3259Kuo2Init[]       =   {   1   ,   1   ,   5   ,   9   ,   5   ,   43  ,   81  ,   99  ,   505 ,   739 ,   633 ,   1563    ,   6385    ,   14799   ,   24813   ,0 };
            const unsigned long dim3260Kuo2Init[]       =   {   1   ,   3   ,   7   ,   13  ,   19  ,   47  ,   11  ,   87  ,   403 ,   763 ,   483 ,   3793    ,   7445    ,   4865    ,   13187   ,0 };
            const unsigned long dim3261Kuo2Init[]       =   {   1   ,   1   ,   7   ,   9   ,   17  ,   49  ,   25  ,   57  ,   417 ,   253 ,   1901    ,   115 ,   3933    ,   11743   ,   1925    ,0 };
            const unsigned long dim3262Kuo2Init[]       =   {   1   ,   3   ,   5   ,   7   ,   15  ,   21  ,   13  ,   143 ,   399 ,   953 ,   1397    ,   3905    ,   7935    ,   11957   ,   4817    ,0 };
            const unsigned long dim3263Kuo2Init[]       =   {   1   ,   1   ,   5   ,   1   ,   7   ,   63  ,   49  ,   11  ,   321 ,   181 ,   545 ,   2035    ,   5339    ,   2727    ,   11623   ,0 };
            const unsigned long dim3264Kuo2Init[]       =   {   1   ,   3   ,   7   ,   1   ,   15  ,   13  ,   49  ,   101 ,   499 ,   831 ,   427 ,   531 ,   4165    ,   7087    ,   14077   ,0 };
            const unsigned long dim3265Kuo2Init[]       =   {   1   ,   3   ,   5   ,   9   ,   13  ,   19  ,   119 ,   115 ,   283 ,   1015    ,   1189    ,   2507    ,   2491    ,   13467   ,   10437   ,0 };
            const unsigned long dim3266Kuo2Init[]       =   {   1   ,   1   ,   5   ,   3   ,   11  ,   31  ,   103 ,   3   ,   421 ,   717 ,   611 ,   2711    ,   2221    ,   13187   ,   19973   ,0 };
            const unsigned long dim3267Kuo2Init[]       =   {   1   ,   1   ,   1   ,   11  ,   3   ,   53  ,   61  ,   3   ,   337 ,   55  ,   751 ,   3467    ,   7851    ,   3437    ,   31047   ,0 };
            const unsigned long dim3268Kuo2Init[]       =   {   1   ,   1   ,   1   ,   5   ,   1   ,   57  ,   81  ,   127 ,   425 ,   965 ,   1633    ,   2193    ,   7061    ,   2239    ,   18383   ,0 };
            const unsigned long dim3269Kuo2Init[]       =   {   1   ,   1   ,   5   ,   15  ,   11  ,   55  ,   49  ,   31  ,   497 ,   715 ,   1725    ,   1359    ,   177 ,   13011   ,   11291   ,0 };
            const unsigned long dim3270Kuo2Init[]       =   {   1   ,   1   ,   5   ,   9   ,   3   ,   11  ,   121 ,   217 ,   77  ,   627 ,   1857    ,   41  ,   7613    ,   3471    ,   10591   ,0 };
            const unsigned long dim3271Kuo2Init[]       =   {   1   ,   1   ,   7   ,   11  ,   13  ,   23  ,   81  ,   253 ,   263 ,   663 ,   1581    ,   1839    ,   1327    ,   7735    ,   22843   ,0 };
            const unsigned long dim3272Kuo2Init[]       =   {   1   ,   1   ,   7   ,   3   ,   31  ,   29  ,   97  ,   243 ,   215 ,   893 ,   1179    ,   953 ,   915 ,   16003   ,   32685   ,0 };
            const unsigned long dim3273Kuo2Init[]       =   {   1   ,   1   ,   7   ,   3   ,   5   ,   61  ,   7   ,   35  ,   439 ,   1023    ,   1079    ,   3639    ,   3179    ,   8793    ,   8413    ,0 };
            const unsigned long dim3274Kuo2Init[]       =   {   1   ,   1   ,   5   ,   13  ,   31  ,   1   ,   99  ,   199 ,   181 ,   75  ,   283 ,   1581    ,   1137    ,   2725    ,   27179   ,0 };
            const unsigned long dim3275Kuo2Init[]       =   {   1   ,   3   ,   7   ,   1   ,   27  ,   35  ,   45  ,   49  ,   179 ,   403 ,   593 ,   2849    ,   2349    ,   5835    ,   32533   ,0 };
            const unsigned long dim3276Kuo2Init[]       =   {   1   ,   1   ,   1   ,   1   ,   5   ,   31  ,   21  ,   153 ,   169 ,   19  ,   813 ,   3077    ,   2569    ,   12355   ,   3805    ,0 };
            const unsigned long dim3277Kuo2Init[]       =   {   1   ,   1   ,   5   ,   11  ,   1   ,   15  ,   77  ,   31  ,   17  ,   827 ,   2005    ,   799 ,   5663    ,   779 ,   28605   ,0 };
            const unsigned long dim3278Kuo2Init[]       =   {   1   ,   1   ,   5   ,   3   ,   29  ,   51  ,   3   ,   91  ,   505 ,   161 ,   2041    ,   2841    ,   269 ,   33  ,   21771   ,0 };
            const unsigned long dim3279Kuo2Init[]       =   {   1   ,   1   ,   3   ,   5   ,   17  ,   55  ,   105 ,   233 ,   9   ,   647 ,   987 ,   3801    ,   1421    ,   8589    ,   23389   ,0 };
            const unsigned long dim3280Kuo2Init[]       =   {   1   ,   1   ,   7   ,   15  ,   27  ,   63  ,   73  ,   87  ,   111 ,   233 ,   1231    ,   3631    ,   4355    ,   15843   ,   1227    ,0 };
            const unsigned long dim3281Kuo2Init[]       =   {   1   ,   1   ,   3   ,   15  ,   17  ,   5   ,   63  ,   23  ,   55  ,   655 ,   1909    ,   891 ,   7873    ,   11163   ,   26269   ,0 };
            const unsigned long dim3282Kuo2Init[]       =   {   1   ,   1   ,   1   ,   5   ,   19  ,   7   ,   47  ,   31  ,   41  ,   229 ,   1381    ,   773 ,   2183    ,   1687    ,   15761   ,0 };
            const unsigned long dim3283Kuo2Init[]       =   {   1   ,   1   ,   7   ,   3   ,   17  ,   57  ,   81  ,   115 ,   41  ,   907 ,   1353    ,   981 ,   2601    ,   10247   ,   13753   ,0 };
            const unsigned long dim3284Kuo2Init[]       =   {   1   ,   3   ,   3   ,   1   ,   5   ,   13  ,   83  ,   19  ,   107 ,   935 ,   1519    ,   795 ,   2035    ,   5649    ,   21125   ,0 };
            const unsigned long dim3285Kuo2Init[]       =   {   1   ,   1   ,   3   ,   9   ,   21  ,   7   ,   77  ,   229 ,   85  ,   837 ,   1085    ,   781 ,   6599    ,   5645    ,   15481   ,0 };
            const unsigned long dim3286Kuo2Init[]       =   {   1   ,   3   ,   3   ,   1   ,   13  ,   23  ,   109 ,   21  ,   77  ,   223 ,   131 ,   4013    ,   7507    ,   10463   ,   25621   ,0 };
            const unsigned long dim3287Kuo2Init[]       =   {   1   ,   3   ,   5   ,   11  ,   21  ,   55  ,   99  ,   195 ,   447 ,   57  ,   1395    ,   1283    ,   377 ,   15019   ,   26951   ,0 };
            const unsigned long dim3288Kuo2Init[]       =   {   1   ,   3   ,   5   ,   13  ,   13  ,   35  ,   119 ,   39  ,   161 ,   609 ,   247 ,   2541    ,   3133    ,   9625    ,   12129   ,0 };
            const unsigned long dim3289Kuo2Init[]       =   {   1   ,   3   ,   3   ,   15  ,   7   ,   55  ,   69  ,   169 ,   459 ,   461 ,   1359    ,   255 ,   4567    ,   7915    ,   16883   ,0 };
            const unsigned long dim3290Kuo2Init[]       =   {   1   ,   1   ,   1   ,   11  ,   25  ,   41  ,   93  ,   105 ,   147 ,   893 ,   515 ,   3161    ,   351 ,   12079   ,   3445    ,0 };
            const unsigned long dim3291Kuo2Init[]       =   {   1   ,   3   ,   3   ,   5   ,   15  ,   59  ,   93  ,   61  ,   235 ,   971 ,   1203    ,   2213    ,   3257    ,   16001   ,   20581   ,0 };
            const unsigned long dim3292Kuo2Init[]       =   {   1   ,   1   ,   1   ,   11  ,   19  ,   5   ,   53  ,   49  ,   101 ,   653 ,   897 ,   3761    ,   8087    ,   14891   ,   30595   ,0 };
            const unsigned long dim3293Kuo2Init[]       =   {   1   ,   1   ,   3   ,   7   ,   13  ,   27  ,   37  ,   29  ,   181 ,   875 ,   1429    ,   2483    ,   2987    ,   15951   ,   16217   ,0 };
            const unsigned long dim3294Kuo2Init[]       =   {   1   ,   3   ,   5   ,   7   ,   11  ,   53  ,   69  ,   33  ,   309 ,   467 ,   1149    ,   3103    ,   3713    ,   9701    ,   16637   ,0 };
            const unsigned long dim3295Kuo2Init[]       =   {   1   ,   1   ,   5   ,   1   ,   31  ,   63  ,   5   ,   187 ,   371 ,   895 ,   1365    ,   2671    ,   5899    ,   15493   ,   9685    ,0 };
            const unsigned long dim3296Kuo2Init[]       =   {   1   ,   3   ,   1   ,   5   ,   19  ,   3   ,   127 ,   251 ,   39  ,   335 ,   489 ,   3359    ,   4939    ,   413 ,   521 ,0 };
            const unsigned long dim3297Kuo2Init[]       =   {   1   ,   1   ,   7   ,   11  ,   11  ,   37  ,   51  ,   135 ,   191 ,   535 ,   1037    ,   2365    ,   381 ,   8253    ,   21177   ,0 };
            const unsigned long dim3298Kuo2Init[]       =   {   1   ,   1   ,   3   ,   13  ,   9   ,   23  ,   77  ,   147 ,   93  ,   793 ,   1223    ,   2655    ,   3873    ,   6049    ,   19201   ,0 };
            const unsigned long dim3299Kuo2Init[]       =   {   1   ,   3   ,   5   ,   5   ,   15  ,   41  ,   21  ,   217 ,   149 ,   965 ,   1173    ,   1779    ,   501 ,   4829    ,   17439   ,0 };
            const unsigned long dim3300Kuo2Init[]       =   {   1   ,   3   ,   5   ,   15  ,   11  ,   37  ,   63  ,   233 ,   103 ,   453 ,   799 ,   571 ,   1991    ,   439 ,   26713   ,0 };
            const unsigned long dim3301Kuo2Init[]       =   {   1   ,   1   ,   7   ,   9   ,   31  ,   1   ,   11  ,   155 ,   483 ,   875 ,   1943    ,   1355    ,   3125    ,   8771    ,   30531   ,0 };
            const unsigned long dim3302Kuo2Init[]       =   {   1   ,   3   ,   3   ,   13  ,   7   ,   31  ,   9   ,   139 ,   277 ,   203 ,   1937    ,   31  ,   4519    ,   2127    ,   22557   ,0 };
            const unsigned long dim3303Kuo2Init[]       =   {   1   ,   1   ,   1   ,   3   ,   19  ,   53  ,   117 ,   225 ,   391 ,   501 ,   607 ,   1493    ,   3119    ,   10817   ,   6307    ,0 };
            const unsigned long dim3304Kuo2Init[]       =   {   1   ,   1   ,   3   ,   11  ,   3   ,   19  ,   3   ,   51  ,   507 ,   245 ,   1563    ,   3987    ,   7461    ,   4851    ,   23697   ,0 };
            const unsigned long dim3305Kuo2Init[]       =   {   1   ,   3   ,   3   ,   9   ,   1   ,   39  ,   49  ,   21  ,   39  ,   631 ,   1777    ,   1689    ,   665 ,   5657    ,   26469   ,0 };
            const unsigned long dim3306Kuo2Init[]       =   {   1   ,   3   ,   7   ,   3   ,   17  ,   41  ,   73  ,   13  ,   97  ,   649 ,   1737    ,   653 ,   5093    ,   629 ,   4389    ,0 };
            const unsigned long dim3307Kuo2Init[]       =   {   1   ,   3   ,   5   ,   1   ,   13  ,   9   ,   99  ,   149 ,   389 ,   151 ,   313 ,   2517    ,   77  ,   2157    ,   4713    ,0 };
            const unsigned long dim3308Kuo2Init[]       =   {   1   ,   3   ,   5   ,   1   ,   15  ,   25  ,   3   ,   129 ,   111 ,   847 ,   585 ,   3103    ,   6237    ,   8491    ,   24519   ,0 };
            const unsigned long dim3309Kuo2Init[]       =   {   1   ,   1   ,   3   ,   1   ,   11  ,   25  ,   71  ,   229 ,   325 ,   51  ,   537 ,   3393    ,   6561    ,   6871    ,   30609   ,0 };
            const unsigned long dim3310Kuo2Init[]       =   {   1   ,   3   ,   7   ,   15  ,   5   ,   3   ,   109 ,   237 ,   469 ,   235 ,   1703    ,   525 ,   2579    ,   9107    ,   13261   ,0 };
            const unsigned long dim3311Kuo2Init[]       =   {   1   ,   3   ,   7   ,   13  ,   7   ,   31  ,   65  ,   67  ,   165 ,   1003    ,   957 ,   371 ,   5857    ,   16307   ,   4069    ,0 };
            const unsigned long dim3312Kuo2Init[]       =   {   1   ,   1   ,   7   ,   15  ,   7   ,   63  ,   51  ,   199 ,   63  ,   259 ,   1383    ,   2747    ,   5945    ,   8061    ,   30961   ,0 };
            const unsigned long dim3313Kuo2Init[]       =   {   1   ,   1   ,   1   ,   7   ,   13  ,   43  ,   67  ,   143 ,   407 ,   369 ,   1803    ,   2405    ,   4737    ,   5741    ,   21311   ,0 };
            const unsigned long dim3314Kuo2Init[]       =   {   1   ,   3   ,   7   ,   11  ,   19  ,   25  ,   127 ,   251 ,   325 ,   327 ,   169 ,   3745    ,   2115    ,   4591    ,   12383   ,0 };
            const unsigned long dim3315Kuo2Init[]       =   {   1   ,   1   ,   3   ,   13  ,   31  ,   61  ,   35  ,   209 ,   161 ,   171 ,   1155    ,   531 ,   5493    ,   13607   ,   23809   ,0 };
            const unsigned long dim3316Kuo2Init[]       =   {   1   ,   1   ,   3   ,   9   ,   21  ,   39  ,   93  ,   185 ,   421 ,   355 ,   2039    ,   259 ,   7309    ,   6747    ,   745 ,0 };
            const unsigned long dim3317Kuo2Init[]       =   {   1   ,   3   ,   5   ,   11  ,   1   ,   39  ,   43  ,   169 ,   137 ,   21  ,   1377    ,   3091    ,   4893    ,   1197    ,   1785    ,0 };
            const unsigned long dim3318Kuo2Init[]       =   {   1   ,   1   ,   5   ,   5   ,   29  ,   59  ,   89  ,   79  ,   457 ,   463 ,   193 ,   727 ,   2505    ,   10509   ,   5379    ,0 };
            const unsigned long dim3319Kuo2Init[]       =   {   1   ,   3   ,   3   ,   9   ,   23  ,   23  ,   59  ,   35  ,   425 ,   47  ,   1787    ,   1705    ,   3469    ,   14583   ,   3439    ,0 };
            const unsigned long dim3320Kuo2Init[]       =   {   1   ,   1   ,   7   ,   5   ,   27  ,   5   ,   85  ,   51  ,   31  ,   277 ,   851 ,   2731    ,   5613    ,   11305   ,   23727   ,0 };
            const unsigned long dim3321Kuo2Init[]       =   {   1   ,   3   ,   1   ,   15  ,   15  ,   47  ,   109 ,   211 ,   385 ,   505 ,   791 ,   253 ,   4123    ,   13393   ,   31515   ,0 };
            const unsigned long dim3322Kuo2Init[]       =   {   1   ,   3   ,   5   ,   3   ,   25  ,   9   ,   73  ,   35  ,   203 ,   721 ,   1111    ,   1791    ,   5241    ,   15307   ,   4385    ,0 };
            const unsigned long dim3323Kuo2Init[]       =   {   1   ,   3   ,   5   ,   1   ,   9   ,   43  ,   127 ,   127 ,   191 ,   21  ,   1967    ,   3523    ,   5959    ,   5775    ,   24635   ,0 };
            const unsigned long dim3324Kuo2Init[]       =   {   1   ,   3   ,   7   ,   5   ,   7   ,   23  ,   75  ,   37  ,   307 ,   375 ,   1731    ,   2945    ,   763 ,   11723   ,   32033   ,0 };
            const unsigned long dim3325Kuo2Init[]       =   {   1   ,   3   ,   3   ,   13  ,   31  ,   19  ,   83  ,   185 ,   457 ,   109 ,   349 ,   1085    ,   7381    ,   8851    ,   27633   ,0 };
            const unsigned long dim3326Kuo2Init[]       =   {   1   ,   3   ,   3   ,   1   ,   25  ,   17  ,   79  ,   73  ,   381 ,   655 ,   409 ,   2043    ,   6803    ,   9041    ,   25733   ,0 };
            const unsigned long dim3327Kuo2Init[]       =   {   1   ,   3   ,   5   ,   7   ,   17  ,   9   ,   61  ,   47  ,   429 ,   507 ,   1933    ,   1653    ,   2057    ,   15791   ,   12035   ,0 };
            const unsigned long dim3328Kuo2Init[]       =   {   1   ,   1   ,   7   ,   1   ,   5   ,   57  ,   31  ,   83  ,   223 ,   281 ,   1485    ,   2965    ,   4543    ,   9719    ,   22109   ,0 };
            const unsigned long dim3329Kuo2Init[]       =   {   1   ,   1   ,   3   ,   3   ,   5   ,   11  ,   69  ,   65  ,   229 ,   87  ,   1707    ,   2701    ,   985 ,   12261   ,   32209   ,0 };
            const unsigned long dim3330Kuo2Init[]       =   {   1   ,   3   ,   7   ,   15  ,   13  ,   29  ,   121 ,   1   ,   475 ,   905 ,   1843    ,   571 ,   5595    ,   1685    ,   8061    ,0 };
            const unsigned long dim3331Kuo2Init[]       =   {   1   ,   3   ,   5   ,   1   ,   9   ,   37  ,   115 ,   35  ,   309 ,   117 ,   17  ,   3537    ,   4491    ,   13863   ,   12611   ,0 };
            const unsigned long dim3332Kuo2Init[]       =   {   1   ,   1   ,   7   ,   15  ,   9   ,   1   ,   79  ,   135 ,   95  ,   417 ,   633 ,   2109    ,   7911    ,   15009   ,   32399   ,0 };
            const unsigned long dim3333Kuo2Init[]       =   {   1   ,   3   ,   7   ,   5   ,   9   ,   41  ,   23  ,   137 ,   135 ,   429 ,   267 ,   3787    ,   4643    ,   1221    ,   20257   ,0 };
            const unsigned long dim3334Kuo2Init[]       =   {   1   ,   3   ,   3   ,   1   ,   11  ,   37  ,   97  ,   107 ,   43  ,   273 ,   213 ,   1779    ,   2317    ,   7325    ,   2495    ,0 };
            const unsigned long dim3335Kuo2Init[]       =   {   1   ,   1   ,   3   ,   9   ,   13  ,   57  ,   29  ,   29  ,   5   ,   45  ,   633 ,   3437    ,   8147    ,   11819   ,   16397   ,0 };
            const unsigned long dim3336Kuo2Init[]       =   {   1   ,   3   ,   7   ,   5   ,   19  ,   47  ,   67  ,   39  ,   281 ,   1003    ,   1489    ,   199 ,   3099    ,   3925    ,   17517   ,0 };
            const unsigned long dim3337Kuo2Init[]       =   {   1   ,   3   ,   3   ,   3   ,   21  ,   49  ,   91  ,   47  ,   435 ,   193 ,   683 ,   1845    ,   6237    ,   5573    ,   10647   ,0 };
            const unsigned long dim3338Kuo2Init[]       =   {   1   ,   3   ,   1   ,   11  ,   3   ,   59  ,   5   ,   197 ,   393 ,   389 ,   1735    ,   1417    ,   2061    ,   11827   ,   7587    ,0 };
            const unsigned long dim3339Kuo2Init[]       =   {   1   ,   1   ,   5   ,   15  ,   1   ,   35  ,   113 ,   81  ,   233 ,   435 ,   1481    ,   2007    ,   6043    ,   9005    ,   19277   ,0 };
            const unsigned long dim3340Kuo2Init[]       =   {   1   ,   3   ,   1   ,   11  ,   9   ,   53  ,   109 ,   193 ,   329 ,   187 ,   35  ,   3693    ,   951 ,   9735    ,   5587    ,0 };
            const unsigned long dim3341Kuo2Init[]       =   {   1   ,   1   ,   1   ,   5   ,   19  ,   41  ,   25  ,   225 ,   269 ,   317 ,   201 ,   2817    ,   971 ,   2873    ,   22637   ,0 };
            const unsigned long dim3342Kuo2Init[]       =   {   1   ,   3   ,   7   ,   1   ,   9   ,   47  ,   105 ,   219 ,   69  ,   633 ,   1073    ,   1895    ,   1759    ,   12353   ,   9011    ,0 };
            const unsigned long dim3343Kuo2Init[]       =   {   1   ,   1   ,   7   ,   11  ,   11  ,   59  ,   3   ,   163 ,   361 ,   157 ,   1937    ,   3589    ,   1531    ,   9303    ,   3627    ,0 };
            const unsigned long dim3344Kuo2Init[]       =   {   1   ,   1   ,   3   ,   9   ,   31  ,   15  ,   23  ,   5   ,   447 ,   59  ,   1907    ,   197 ,   2775    ,   11303   ,   32553   ,0 };
            const unsigned long dim3345Kuo2Init[]       =   {   1   ,   3   ,   1   ,   13  ,   11  ,   35  ,   83  ,   97  ,   225 ,   761 ,   907 ,   3107    ,   1827    ,   14745   ,   235 ,0 };
            const unsigned long dim3346Kuo2Init[]       =   {   1   ,   3   ,   1   ,   13  ,   13  ,   37  ,   127 ,   55  ,   511 ,   861 ,   1769    ,   3999    ,   8135    ,   445 ,   12537   ,0 };
            const unsigned long dim3347Kuo2Init[]       =   {   1   ,   1   ,   7   ,   11  ,   21  ,   61  ,   111 ,   107 ,   129 ,   43  ,   837 ,   3811    ,   3261    ,   15273   ,   24319   ,0 };
            const unsigned long dim3348Kuo2Init[]       =   {   1   ,   3   ,   7   ,   1   ,   17  ,   15  ,   23  ,   27  ,   227 ,   573 ,   1739    ,   2611    ,   3977    ,   493 ,   24091   ,0 };
            const unsigned long dim3349Kuo2Init[]       =   {   1   ,   3   ,   3   ,   5   ,   17  ,   59  ,   17  ,   61  ,   443 ,   215 ,   1171    ,   1587    ,   55  ,   14417   ,   30225   ,0 };
            const unsigned long dim3350Kuo2Init[]       =   {   1   ,   3   ,   3   ,   5   ,   5   ,   51  ,   5   ,   101 ,   147 ,   951 ,   445 ,   2203    ,   6731    ,   3993    ,   12935   ,0 };
            const unsigned long dim3351Kuo2Init[]       =   {   1   ,   1   ,   3   ,   13  ,   31  ,   15  ,   117 ,   125 ,   413 ,   927 ,   1837    ,   2781    ,   2569    ,   12177   ,   20551   ,0 };
            const unsigned long dim3352Kuo2Init[]       =   {   1   ,   1   ,   1   ,   9   ,   27  ,   13  ,   115 ,   243 ,   211 ,   123 ,   1631    ,   1277    ,   1217    ,   15271   ,   26377   ,0 };
            const unsigned long dim3353Kuo2Init[]       =   {   1   ,   3   ,   7   ,   3   ,   27  ,   55  ,   5   ,   91  ,   255 ,   367 ,   503 ,   659 ,   1527    ,   15907   ,   4485    ,0 };
            const unsigned long dim3354Kuo2Init[]       =   {   1   ,   3   ,   3   ,   13  ,   9   ,   11  ,   29  ,   175 ,   305 ,   593 ,   1209    ,   551 ,   7579    ,   1273    ,   22545   ,0 };
            const unsigned long dim3355Kuo2Init[]       =   {   1   ,   3   ,   3   ,   13  ,   17  ,   33  ,   101 ,   83  ,   93  ,   261 ,   1395    ,   573 ,   6665    ,   10073   ,   26139   ,0 };
            const unsigned long dim3356Kuo2Init[]       =   {   1   ,   1   ,   3   ,   7   ,   25  ,   35  ,   93  ,   11  ,   29  ,   941 ,   525 ,   457 ,   1293    ,   15753   ,   19273   ,0 };
            const unsigned long dim3357Kuo2Init[]       =   {   1   ,   3   ,   5   ,   15  ,   5   ,   31  ,   101 ,   211 ,   383 ,   759 ,   1287    ,   429 ,   291 ,   12347   ,   12683   ,0 };
            const unsigned long dim3358Kuo2Init[]       =   {   1   ,   1   ,   7   ,   3   ,   9   ,   53  ,   119 ,   81  ,   91  ,   113 ,   1525    ,   1037    ,   8021    ,   2913    ,   11205   ,0 };
            const unsigned long dim3359Kuo2Init[]       =   {   1   ,   1   ,   3   ,   1   ,   9   ,   47  ,   59  ,   23  ,   497 ,   273 ,   453 ,   1063    ,   2523    ,   8929    ,   43  ,0 };
            const unsigned long dim3360Kuo2Init[]       =   {   1   ,   1   ,   3   ,   9   ,   1   ,   23  ,   65  ,   147 ,   327 ,   851 ,   177 ,   4047    ,   3881    ,   2885    ,   21021   ,0 };
            const unsigned long dim3361Kuo2Init[]       =   {   1   ,   1   ,   7   ,   1   ,   19  ,   53  ,   13  ,   117 ,   267 ,   803 ,   21  ,   3825    ,   2387    ,   7273    ,   19753   ,0 };
            const unsigned long dim3362Kuo2Init[]       =   {   1   ,   3   ,   5   ,   7   ,   9   ,   9   ,   107 ,   9   ,   261 ,   797 ,   1337    ,   1535    ,   663 ,   6635    ,   3383    ,0 };
            const unsigned long dim3363Kuo2Init[]       =   {   1   ,   1   ,   7   ,   13  ,   13  ,   27  ,   65  ,   75  ,   375 ,   805 ,   1079    ,   947 ,   993 ,   1793    ,   16235   ,0 };
            const unsigned long dim3364Kuo2Init[]       =   {   1   ,   3   ,   3   ,   9   ,   11  ,   17  ,   119 ,   57  ,   501 ,   105 ,   1109    ,   3607    ,   1805    ,   8277    ,   24869   ,0 };
            const unsigned long dim3365Kuo2Init[]       =   {   1   ,   3   ,   7   ,   1   ,   1   ,   29  ,   125 ,   205 ,   453 ,   607 ,   1371    ,   2169    ,   2723    ,   7381    ,   23411   ,0 };
            const unsigned long dim3366Kuo2Init[]       =   {   1   ,   3   ,   7   ,   13  ,   13  ,   57  ,   83  ,   149 ,   197 ,   739 ,   823 ,   1551    ,   5329    ,   10781   ,   6055    ,0 };
            const unsigned long dim3367Kuo2Init[]       =   {   1   ,   3   ,   7   ,   1   ,   5   ,   33  ,   9   ,   237 ,   37  ,   733 ,   879 ,   1303    ,   365 ,   5595    ,   32425   ,0 };
            const unsigned long dim3368Kuo2Init[]       =   {   1   ,   1   ,   1   ,   1   ,   17  ,   13  ,   117 ,   189 ,   323 ,   377 ,   1645    ,   2143    ,   4469    ,   9497    ,   11573   ,0 };
            const unsigned long dim3369Kuo2Init[]       =   {   1   ,   3   ,   1   ,   5   ,   13  ,   47  ,   19  ,   19  ,   155 ,   681 ,   1909    ,   913 ,   6985    ,   11449   ,   9981    ,0 };
            const unsigned long dim3370Kuo2Init[]       =   {   1   ,   3   ,   7   ,   11  ,   29  ,   55  ,   73  ,   113 ,   291 ,   865 ,   239 ,   209 ,   67  ,   10813   ,   19287   ,0 };
            const unsigned long dim3371Kuo2Init[]       =   {   1   ,   3   ,   5   ,   9   ,   3   ,   7   ,   27  ,   119 ,   267 ,   649 ,   1125    ,   2953    ,   3393    ,   3281    ,   15539   ,0 };
            const unsigned long dim3372Kuo2Init[]       =   {   1   ,   3   ,   3   ,   13  ,   9   ,   41  ,   25  ,   145 ,   233 ,   475 ,   873 ,   631 ,   2429    ,   12071   ,   18123   ,0 };
            const unsigned long dim3373Kuo2Init[]       =   {   1   ,   3   ,   1   ,   3   ,   17  ,   53  ,   85  ,   71  ,   193 ,   339 ,   493 ,   2495    ,   4369    ,   7239    ,   21395   ,0 };
            const unsigned long dim3374Kuo2Init[]       =   {   1   ,   3   ,   1   ,   3   ,   25  ,   51  ,   87  ,   11  ,   291 ,   659 ,   455 ,   1217    ,   5737    ,   3263    ,   15213   ,0 };
            const unsigned long dim3375Kuo2Init[]       =   {   1   ,   3   ,   1   ,   11  ,   5   ,   37  ,   93  ,   51  ,   359 ,   113 ,   1847    ,   4007    ,   2497    ,   12287   ,   16961   ,0 };
            const unsigned long dim3376Kuo2Init[]       =   {   1   ,   3   ,   5   ,   11  ,   1   ,   27  ,   107 ,   189 ,   353 ,   499 ,   395 ,   1717    ,   6313    ,   315 ,   17027   ,0 };
            const unsigned long dim3377Kuo2Init[]       =   {   1   ,   1   ,   3   ,   3   ,   15  ,   57  ,   105 ,   107 ,   17  ,   769 ,   1207    ,   3243    ,   179 ,   305 ,   16425   ,0 };
            const unsigned long dim3378Kuo2Init[]       =   {   1   ,   3   ,   5   ,   1   ,   19  ,   9   ,   87  ,   5   ,   303 ,   653 ,   481 ,   1673    ,   3101    ,   7751    ,   23301   ,0 };
            const unsigned long dim3379Kuo2Init[]       =   {   1   ,   1   ,   5   ,   9   ,   31  ,   39  ,   67  ,   225 ,   287 ,   285 ,   1937    ,   413 ,   2101    ,   10941   ,   32071   ,0 };
            const unsigned long dim3380Kuo2Init[]       =   {   1   ,   3   ,   5   ,   5   ,   1   ,   25  ,   49  ,   61  ,   309 ,   19  ,   641 ,   247 ,   7909    ,   13761   ,   14231   ,0 };
            const unsigned long dim3381Kuo2Init[]       =   {   1   ,   1   ,   5   ,   5   ,   13  ,   31  ,   55  ,   141 ,   223 ,   573 ,   895 ,   2853    ,   5663    ,   15519   ,   10871   ,0 };
            const unsigned long dim3382Kuo2Init[]       =   {   1   ,   3   ,   3   ,   11  ,   31  ,   27  ,   61  ,   239 ,   215 ,   897 ,   857 ,   2591    ,   6251    ,   14317   ,   10973   ,0 };
            const unsigned long dim3383Kuo2Init[]       =   {   1   ,   3   ,   1   ,   5   ,   29  ,   23  ,   115 ,   5   ,   449 ,   455 ,   1597    ,   3329    ,   4463    ,   10653   ,   17079   ,0 };
            const unsigned long dim3384Kuo2Init[]       =   {   1   ,   1   ,   1   ,   1   ,   31  ,   27  ,   25  ,   43  ,   99  ,   817 ,   1863    ,   1563    ,   7021    ,   4981    ,   5049    ,0 };
            const unsigned long dim3385Kuo2Init[]       =   {   1   ,   1   ,   7   ,   5   ,   1   ,   13  ,   95  ,   241 ,   269 ,   145 ,   737 ,   1479    ,   7469    ,   10069   ,   21421   ,0 };
            const unsigned long dim3386Kuo2Init[]       =   {   1   ,   1   ,   5   ,   5   ,   21  ,   47  ,   111 ,   153 ,   411 ,   763 ,   2003    ,   1809    ,   5887    ,   4543    ,   7505    ,0 };
            const unsigned long dim3387Kuo2Init[]       =   {   1   ,   3   ,   3   ,   7   ,   1   ,   31  ,   75  ,   131 ,   103 ,   873 ,   1893    ,   2025    ,   6525    ,   57  ,   19701   ,0 };
            const unsigned long dim3388Kuo2Init[]       =   {   1   ,   1   ,   1   ,   7   ,   13  ,   3   ,   41  ,   243 ,   19  ,   1007    ,   365 ,   601 ,   7015    ,   485 ,   6313    ,0 };
            const unsigned long dim3389Kuo2Init[]       =   {   1   ,   3   ,   5   ,   9   ,   9   ,   49  ,   27  ,   47  ,   139 ,   837 ,   43  ,   2843    ,   83  ,   9987    ,   29817   ,0 };
            const unsigned long dim3390Kuo2Init[]       =   {   1   ,   3   ,   7   ,   15  ,   27  ,   51  ,   95  ,   133 ,   275 ,   625 ,   1489    ,   4077    ,   3049    ,   15213   ,   23817   ,0 };
            const unsigned long dim3391Kuo2Init[]       =   {   1   ,   1   ,   1   ,   7   ,   1   ,   33  ,   101 ,   109 ,   131 ,   997 ,   101 ,   881 ,   3051    ,   1215    ,   4783    ,0 };
            const unsigned long dim3392Kuo2Init[]       =   {   1   ,   3   ,   7   ,   5   ,   9   ,   7   ,   3   ,   223 ,   343 ,   467 ,   697 ,   3441    ,   4723    ,   4265    ,   29517   ,0 };
            const unsigned long dim3393Kuo2Init[]       =   {   1   ,   1   ,   7   ,   3   ,   17  ,   25  ,   97  ,   191 ,   413 ,   211 ,   263 ,   2027    ,   2025    ,   6125    ,   9667    ,0 };
            const unsigned long dim3394Kuo2Init[]       =   {   1   ,   1   ,   7   ,   11  ,   15  ,   15  ,   57  ,   151 ,   351 ,   187 ,   103 ,   1231    ,   567 ,   7829    ,   17941   ,0 };
            const unsigned long dim3395Kuo2Init[]       =   {   1   ,   3   ,   3   ,   7   ,   13  ,   43  ,   49  ,   235 ,   195 ,   271 ,   1945    ,   1459    ,   2335    ,   141 ,   27615   ,0 };
            const unsigned long dim3396Kuo2Init[]       =   {   1   ,   1   ,   7   ,   9   ,   5   ,   49  ,   123 ,   231 ,   133 ,   781 ,   697 ,   157 ,   2889    ,   12929   ,   18099   ,0 };
            const unsigned long dim3397Kuo2Init[]       =   {   1   ,   1   ,   3   ,   7   ,   19  ,   25  ,   127 ,   181 ,   169 ,   981 ,   1955    ,   3289    ,   2077    ,   10627   ,   30383   ,0 };
            const unsigned long dim3398Kuo2Init[]       =   {   1   ,   3   ,   7   ,   9   ,   1   ,   53  ,   99  ,   163 ,   291 ,   701 ,   1367    ,   3941    ,   3211    ,   6661    ,   29143   ,0 };
            const unsigned long dim3399Kuo2Init[]       =   {   1   ,   1   ,   7   ,   9   ,   7   ,   49  ,   83  ,   77  ,   9   ,   549 ,   705 ,   527 ,   5609    ,   10933   ,   27715   ,0 };
            const unsigned long dim3400Kuo2Init[]       =   {   1   ,   3   ,   5   ,   7   ,   27  ,   15  ,   31  ,   75  ,   39  ,   823 ,   497 ,   875 ,   435 ,   14427   ,   30859   ,0 };
            const unsigned long dim3401Kuo2Init[]       =   {   1   ,   1   ,   7   ,   5   ,   21  ,   49  ,   63  ,   245 ,   357 ,   653 ,   45  ,   2053    ,   169 ,   7459    ,   2441    ,0 };
            const unsigned long dim3402Kuo2Init[]       =   {   1   ,   3   ,   1   ,   13  ,   7   ,   53  ,   109 ,   167 ,   243 ,   585 ,   1173    ,   1925    ,   601 ,   13899   ,   29  ,0 };
            const unsigned long dim3403Kuo2Init[]       =   {   1   ,   3   ,   3   ,   15  ,   21  ,   1   ,   119 ,   33  ,   129 ,   893 ,   1047    ,   1223    ,   3335    ,   5711    ,   28213   ,0 };
            const unsigned long dim3404Kuo2Init[]       =   {   1   ,   1   ,   5   ,   11  ,   21  ,   47  ,   97  ,   9   ,   201 ,   647 ,   1595    ,   3091    ,   3635    ,   15487   ,   21375   ,0 };
            const unsigned long dim3405Kuo2Init[]       =   {   1   ,   3   ,   1   ,   11  ,   11  ,   17  ,   11  ,   133 ,   447 ,   925 ,   97  ,   2841    ,   4147    ,   15803   ,   6311    ,0 };
            const unsigned long dim3406Kuo2Init[]       =   {   1   ,   3   ,   1   ,   7   ,   13  ,   27  ,   27  ,   201 ,   105 ,   37  ,   399 ,   1695    ,   4263    ,   6081    ,   21461   ,0 };
            const unsigned long dim3407Kuo2Init[]       =   {   1   ,   1   ,   7   ,   5   ,   7   ,   25  ,   45  ,   83  ,   327 ,   283 ,   1799    ,   691 ,   2795    ,   12961   ,   125 ,0 };
            const unsigned long dim3408Kuo2Init[]       =   {   1   ,   1   ,   5   ,   11  ,   23  ,   23  ,   11  ,   221 ,   189 ,   353 ,   887 ,   3087    ,   4937    ,   11335   ,   29193   ,0 };
            const unsigned long dim3409Kuo2Init[]       =   {   1   ,   1   ,   1   ,   9   ,   1   ,   47  ,   49  ,   141 ,   391 ,   661 ,   1289    ,   3937    ,   5833    ,   10731   ,   26501   ,0 };
            const unsigned long dim3410Kuo2Init[]       =   {   1   ,   3   ,   7   ,   5   ,   21  ,   57  ,   89  ,   123 ,   163 ,   137 ,   1575    ,   4073    ,   7663    ,   13611   ,   11511   ,0 };
            const unsigned long dim3411Kuo2Init[]       =   {   1   ,   1   ,   7   ,   15  ,   17  ,   27  ,   27  ,   149 ,   67  ,   147 ,   1069    ,   353 ,   3323    ,   7763    ,   11549   ,0 };
            const unsigned long dim3412Kuo2Init[]       =   {   1   ,   3   ,   1   ,   3   ,   19  ,   61  ,   11  ,   19  ,   403 ,   451 ,   737 ,   3107    ,   847 ,   13403   ,   25943   ,0 };
            const unsigned long dim3413Kuo2Init[]       =   {   1   ,   1   ,   1   ,   11  ,   11  ,   55  ,   93  ,   185 ,   241 ,   807 ,   1017    ,   1707    ,   2745    ,   1923    ,   20451   ,0 };
            const unsigned long dim3414Kuo2Init[]       =   {   1   ,   3   ,   1   ,   5   ,   25  ,   51  ,   125 ,   87  ,   275 ,   43  ,   65  ,   3327    ,   3313    ,   211 ,   24217   ,0 };
            const unsigned long dim3415Kuo2Init[]       =   {   1   ,   3   ,   3   ,   7   ,   23  ,   9   ,   3   ,   207 ,   367 ,   939 ,   423 ,   3741    ,   4051    ,   2673    ,   25363   ,0 };
            const unsigned long dim3416Kuo2Init[]       =   {   1   ,   1   ,   1   ,   9   ,   5   ,   41  ,   47  ,   255 ,   359 ,   179 ,   365 ,   3037    ,   6377    ,   14293   ,   21925   ,0 };
            const unsigned long dim3417Kuo2Init[]       =   {   1   ,   1   ,   7   ,   9   ,   31  ,   31  ,   97  ,   77  ,   479 ,   375 ,   491 ,   1101    ,   6717    ,   5499    ,   12693   ,0 };
            const unsigned long dim3418Kuo2Init[]       =   {   1   ,   1   ,   3   ,   1   ,   29  ,   55  ,   75  ,   209 ,   231 ,   15  ,   49  ,   3233    ,   3627    ,   11307   ,   11413   ,0 };
            const unsigned long dim3419Kuo2Init[]       =   {   1   ,   1   ,   7   ,   5   ,   13  ,   25  ,   105 ,   165 ,   15  ,   863 ,   827 ,   3691    ,   3531    ,   1487    ,   14615   ,0 };
            const unsigned long dim3420Kuo2Init[]       =   {   1   ,   3   ,   3   ,   1   ,   19  ,   17  ,   51  ,   37  ,   349 ,   71  ,   31  ,   2545    ,   1151    ,   6187    ,   26835   ,0 };
            const unsigned long dim3421Kuo2Init[]       =   {   1   ,   3   ,   3   ,   13  ,   21  ,   43  ,   19  ,   81  ,   307 ,   751 ,   61  ,   3319    ,   2129    ,   9275    ,   18825   ,0 };
            const unsigned long dim3422Kuo2Init[]       =   {   1   ,   3   ,   7   ,   11  ,   11  ,   57  ,   117 ,   29  ,   189 ,   267 ,   31  ,   579 ,   6103    ,   10553   ,   15037   ,0 };
            const unsigned long dim3423Kuo2Init[]       =   {   1   ,   3   ,   3   ,   5   ,   7   ,   41  ,   101 ,   149 ,   43  ,   693 ,   797 ,   2447    ,   2973    ,   11177   ,   5795    ,0 };
            const unsigned long dim3424Kuo2Init[]       =   {   1   ,   1   ,   1   ,   15  ,   25  ,   55  ,   117 ,   207 ,   187 ,   211 ,   109 ,   3719    ,   8161    ,   8091    ,   8301    ,0 };
            const unsigned long dim3425Kuo2Init[]       =   {   1   ,   1   ,   5   ,   11  ,   3   ,   27  ,   17  ,   61  ,   71  ,   555 ,   847 ,   1733    ,   6171    ,   6467    ,   2289    ,0 };
            const unsigned long dim3426Kuo2Init[]       =   {   1   ,   3   ,   7   ,   15  ,   31  ,   57  ,   103 ,   173 ,   127 ,   513 ,   2003    ,   1483    ,   3495    ,   1951    ,   30371   ,0 };
            const unsigned long dim3427Kuo2Init[]       =   {   1   ,   3   ,   7   ,   1   ,   13  ,   3   ,   107 ,   63  ,   77  ,   647 ,   1645    ,   2935    ,   4019    ,   595 ,   21603   ,0 };
            const unsigned long dim3428Kuo2Init[]       =   {   1   ,   3   ,   7   ,   15  ,   23  ,   23  ,   121 ,   51  ,   163 ,   321 ,   2005    ,   3873    ,   41  ,   15907   ,   26881   ,0 };
            const unsigned long dim3429Kuo2Init[]       =   {   1   ,   3   ,   1   ,   11  ,   27  ,   63  ,   1   ,   173 ,   423 ,   1023    ,   1123    ,   1683    ,   5395    ,   15017   ,   31877   ,0 };
            const unsigned long dim3430Kuo2Init[]       =   {   1   ,   1   ,   3   ,   1   ,   19  ,   19  ,   67  ,   121 ,   277 ,   409 ,   809 ,   4061    ,   6041    ,   2161    ,   13845   ,0 };
            const unsigned long dim3431Kuo2Init[]       =   {   1   ,   3   ,   7   ,   9   ,   13  ,   1   ,   111 ,   241 ,   193 ,   463 ,   1585    ,   2787    ,   2651    ,   4439    ,   27315   ,0 };
            const unsigned long dim3432Kuo2Init[]       =   {   1   ,   3   ,   1   ,   3   ,   5   ,   1   ,   83  ,   213 ,   425 ,   623 ,   1709    ,   3523    ,   4547    ,   5827    ,   14637   ,0 };
            const unsigned long dim3433Kuo2Init[]       =   {   1   ,   1   ,   5   ,   9   ,   11  ,   43  ,   87  ,   31  ,   455 ,   559 ,   1117    ,   2541    ,   6381    ,   6333    ,   16025   ,0 };
            const unsigned long dim3434Kuo2Init[]       =   {   1   ,   1   ,   3   ,   11  ,   5   ,   61  ,   69  ,   115 ,   253 ,   743 ,   719 ,   3745    ,   3739    ,   5511    ,   12207   ,0 };
            const unsigned long dim3435Kuo2Init[]       =   {   1   ,   3   ,   7   ,   1   ,   31  ,   63  ,   11  ,   239 ,   217 ,   569 ,   749 ,   2445    ,   7183    ,   11169   ,   15683   ,0 };
            const unsigned long dim3436Kuo2Init[]       =   {   1   ,   1   ,   1   ,   13  ,   21  ,   47  ,   11  ,   187 ,   65  ,   19  ,   217 ,   2205    ,   5435    ,   14333   ,   8095    ,0 };
            const unsigned long dim3437Kuo2Init[]       =   {   1   ,   1   ,   5   ,   1   ,   15  ,   63  ,   81  ,   3   ,   291 ,   289 ,   1575    ,   3499    ,   5583    ,   16189   ,   14587   ,0 };
            const unsigned long dim3438Kuo2Init[]       =   {   1   ,   1   ,   1   ,   11  ,   29  ,   1   ,   105 ,   195 ,   37  ,   875 ,   203 ,   3937    ,   3661    ,   1103    ,   569 ,0 };
            const unsigned long dim3439Kuo2Init[]       =   {   1   ,   3   ,   3   ,   7   ,   25  ,   55  ,   7   ,   75  ,   185 ,   995 ,   569 ,   1861    ,   1149    ,   14763   ,   7581    ,0 };
            const unsigned long dim3440Kuo2Init[]       =   {   1   ,   1   ,   3   ,   15  ,   13  ,   3   ,   107 ,   223 ,   273 ,   607 ,   1877    ,   3595    ,   3985    ,   1675    ,   9387    ,0 };
            const unsigned long dim3441Kuo2Init[]       =   {   1   ,   3   ,   7   ,   15  ,   15  ,   35  ,   57  ,   95  ,   99  ,   5   ,   351 ,   3767    ,   7617    ,   7071    ,   1465    ,0 };
            const unsigned long dim3442Kuo2Init[]       =   {   1   ,   1   ,   7   ,   1   ,   13  ,   11  ,   19  ,   169 ,   381 ,   245 ,   99  ,   1121    ,   7127    ,   6145    ,   625 ,0 };
            const unsigned long dim3443Kuo2Init[]       =   {   1   ,   3   ,   5   ,   15  ,   29  ,   13  ,   89  ,   81  ,   31  ,   827 ,   989 ,   445 ,   6361    ,   7953    ,   5035    ,0 };
            const unsigned long dim3444Kuo2Init[]       =   {   1   ,   1   ,   5   ,   11  ,   5   ,   9   ,   21  ,   27  ,   321 ,   739 ,   557 ,   3953    ,   3739    ,   15215   ,   5467    ,0 };
            const unsigned long dim3445Kuo2Init[]       =   {   1   ,   3   ,   5   ,   7   ,   9   ,   41  ,   111 ,   195 ,   385 ,   647 ,   829 ,   2109    ,   3313    ,   2457    ,   10925   ,0 };
            const unsigned long dim3446Kuo2Init[]       =   {   1   ,   3   ,   7   ,   1   ,   9   ,   47  ,   107 ,   15  ,   367 ,   385 ,   969 ,   1947    ,   1069    ,   6929    ,   16135   ,0 };
            const unsigned long dim3447Kuo2Init[]       =   {   1   ,   1   ,   5   ,   7   ,   31  ,   27  ,   19  ,   133 ,   413 ,   79  ,   1051    ,   173 ,   1033    ,   9609    ,   17485   ,0 };
            const unsigned long dim3448Kuo2Init[]       =   {   1   ,   1   ,   5   ,   9   ,   23  ,   19  ,   71  ,   135 ,   203 ,   659 ,   385 ,   771 ,   2053    ,   6803    ,   27443   ,0 };
            const unsigned long dim3449Kuo2Init[]       =   {   1   ,   1   ,   1   ,   11  ,   5   ,   11  ,   15  ,   117 ,   353 ,   949 ,   1873    ,   3925    ,   6831    ,   12609   ,   20761   ,0 };
            const unsigned long dim3450Kuo2Init[]       =   {   1   ,   1   ,   5   ,   9   ,   13  ,   51  ,   111 ,   105 ,   75  ,   147 ,   543 ,   2177    ,   2595    ,   4397    ,   20075   ,0 };
            const unsigned long dim3451Kuo2Init[]       =   {   1   ,   3   ,   7   ,   1   ,   31  ,   41  ,   17  ,   81  ,   477 ,   143 ,   1671    ,   131 ,   3861    ,   7937    ,   12421   ,0 };
            const unsigned long dim3452Kuo2Init[]       =   {   1   ,   1   ,   7   ,   9   ,   29  ,   55  ,   59  ,   141 ,   407 ,   287 ,   1205    ,   2349    ,   5033    ,   15907   ,   19857   ,0 };
            const unsigned long dim3453Kuo2Init[]       =   {   1   ,   3   ,   7   ,   11  ,   5   ,   7   ,   123 ,   183 ,   333 ,   547 ,   849 ,   3447    ,   2419    ,   6873    ,   16951   ,0 };
            const unsigned long dim3454Kuo2Init[]       =   {   1   ,   3   ,   7   ,   9   ,   9   ,   31  ,   63  ,   249 ,   211 ,   63  ,   1577    ,   235 ,   621 ,   4907    ,   25537   ,0 };
            const unsigned long dim3455Kuo2Init[]       =   {   1   ,   1   ,   1   ,   1   ,   13  ,   41  ,   1   ,   37  ,   455 ,   111 ,   1959    ,   1529    ,   4029    ,   10687   ,   17283   ,0 };
            const unsigned long dim3456Kuo2Init[]       =   {   1   ,   3   ,   1   ,   7   ,   9   ,   31  ,   105 ,   63  ,   39  ,   409 ,   1151    ,   2631    ,   1491    ,   2837    ,   2489    ,0 };
            const unsigned long dim3457Kuo2Init[]       =   {   1   ,   3   ,   1   ,   5   ,   23  ,   63  ,   61  ,   31  ,   229 ,   685 ,   1869    ,   3101    ,   4285    ,   4485    ,   15371   ,0 };
            const unsigned long dim3458Kuo2Init[]       =   {   1   ,   3   ,   1   ,   5   ,   17  ,   1   ,   9   ,   149 ,   239 ,   393 ,   1925    ,   2111    ,   1035    ,   16343   ,   10655   ,0 };
            const unsigned long dim3459Kuo2Init[]       =   {   1   ,   1   ,   5   ,   15  ,   5   ,   35  ,   107 ,   155 ,   381 ,   119 ,   1157    ,   3875    ,   2647    ,   3149    ,   141 ,0 };
            const unsigned long dim3460Kuo2Init[]       =   {   1   ,   1   ,   3   ,   1   ,   31  ,   1   ,   123 ,   163 ,   337 ,   653 ,   659 ,   2011    ,   2613    ,   15469   ,   16925   ,0 };
            const unsigned long dim3461Kuo2Init[]       =   {   1   ,   1   ,   7   ,   5   ,   19  ,   51  ,   3   ,   17  ,   151 ,   705 ,   349 ,   3015    ,   7791    ,   9899    ,   30235   ,0 };
            const unsigned long dim3462Kuo2Init[]       =   {   1   ,   1   ,   3   ,   11  ,   25  ,   55  ,   51  ,   201 ,   511 ,   67  ,   857 ,   1233    ,   2813    ,   10031   ,   17433   ,0 };
            const unsigned long dim3463Kuo2Init[]       =   {   1   ,   1   ,   5   ,   9   ,   17  ,   3   ,   27  ,   193 ,   81  ,   765 ,   711 ,   2551    ,   2569    ,   9081    ,   9997    ,0 };
            const unsigned long dim3464Kuo2Init[]       =   {   1   ,   3   ,   1   ,   9   ,   17  ,   19  ,   51  ,   41  ,   249 ,   157 ,   989 ,   597 ,   6059    ,   7699    ,   14761   ,0 };
            const unsigned long dim3465Kuo2Init[]       =   {   1   ,   1   ,   1   ,   11  ,   7   ,   3   ,   87  ,   101 ,   171 ,   287 ,   1259    ,   1161    ,   2765    ,   9293    ,   1343    ,0 };
            const unsigned long dim3466Kuo2Init[]       =   {   1   ,   3   ,   1   ,   11  ,   1   ,   49  ,   91  ,   179 ,   287 ,   661 ,   573 ,   255 ,   2873    ,   12153   ,   16505   ,0 };
            const unsigned long dim3467Kuo2Init[]       =   {   1   ,   3   ,   1   ,   5   ,   27  ,   9   ,   101 ,   15  ,   173 ,   327 ,   1523    ,   3727    ,   6089    ,   195 ,   29999   ,0 };
            const unsigned long dim3468Kuo2Init[]       =   {   1   ,   3   ,   3   ,   13  ,   5   ,   17  ,   31  ,   61  ,   395 ,   451 ,   743 ,   2337    ,   7295    ,   2411    ,   31039   ,0 };
            const unsigned long dim3469Kuo2Init[]       =   {   1   ,   1   ,   1   ,   11  ,   17  ,   25  ,   59  ,   209 ,   255 ,   401 ,   321 ,   1881    ,   4539    ,   5371    ,   25829   ,0 };
            const unsigned long dim3470Kuo2Init[]       =   {   1   ,   3   ,   5   ,   1   ,   21  ,   15  ,   57  ,   11  ,   233 ,   13  ,   1969    ,   1839    ,   7923    ,   11731   ,   32641   ,0 };
            const unsigned long dim3471Kuo2Init[]       =   {   1   ,   1   ,   1   ,   15  ,   15  ,   11  ,   69  ,   179 ,   167 ,   867 ,   1901    ,   1113    ,   5613    ,   1481    ,   18975   ,0 };
            const unsigned long dim3472Kuo2Init[]       =   {   1   ,   3   ,   3   ,   11  ,   15  ,   61  ,   1   ,   35  ,   111 ,   347 ,   141 ,   2749    ,   2987    ,   5587    ,   12385   ,0 };
            const unsigned long dim3473Kuo2Init[]       =   {   1   ,   1   ,   7   ,   9   ,   3   ,   41  ,   53  ,   101 ,   417 ,   729 ,   403 ,   2043    ,   453 ,   2479    ,   21141   ,0 };
            const unsigned long dim3474Kuo2Init[]       =   {   1   ,   1   ,   3   ,   7   ,   3   ,   55  ,   47  ,   227 ,   331 ,   425 ,   1361    ,   1657    ,   7231    ,   727 ,   4093    ,0 };
            const unsigned long dim3475Kuo2Init[]       =   {   1   ,   1   ,   1   ,   1   ,   9   ,   37  ,   91  ,   207 ,   493 ,   463 ,   161 ,   3887    ,   543 ,   8349    ,   17949   ,0 };
            const unsigned long dim3476Kuo2Init[]       =   {   1   ,   3   ,   3   ,   15  ,   25  ,   9   ,   89  ,   221 ,   293 ,   997 ,   969 ,   3239    ,   5147    ,   7173    ,   11277   ,0 };
            const unsigned long dim3477Kuo2Init[]       =   {   1   ,   1   ,   7   ,   3   ,   21  ,   29  ,   87  ,   7   ,   331 ,   939 ,   1919    ,   5   ,   113 ,   2189    ,   21471   ,0 };
            const unsigned long dim3478Kuo2Init[]       =   {   1   ,   3   ,   3   ,   5   ,   11  ,   47  ,   67  ,   119 ,   5   ,   357 ,   981 ,   4055    ,   6327    ,   13281   ,   185 ,0 };
            const unsigned long dim3479Kuo2Init[]       =   {   1   ,   3   ,   7   ,   13  ,   27  ,   55  ,   39  ,   23  ,   209 ,   721 ,   2043    ,   417 ,   655 ,   15333   ,   5015    ,0 };
            const unsigned long dim3480Kuo2Init[]       =   {   1   ,   3   ,   7   ,   5   ,   27  ,   7   ,   95  ,   125 ,   9   ,   1001    ,   557 ,   2947    ,   553 ,   4991    ,   11517   ,0 };
            const unsigned long dim3481Kuo2Init[]       =   {   1   ,   3   ,   5   ,   13  ,   3   ,   25  ,   5   ,   99  ,   355 ,   63  ,   1241    ,   2559    ,   4717    ,   9935    ,   19245   ,0 };
            const unsigned long dim3482Kuo2Init[]       =   {   1   ,   3   ,   1   ,   13  ,   3   ,   27  ,   45  ,   141 ,   123 ,   859 ,   603 ,   2495    ,   3815    ,   4229    ,   17501   ,0 };
            const unsigned long dim3483Kuo2Init[]       =   {   1   ,   3   ,   5   ,   3   ,   13  ,   53  ,   5   ,   157 ,   33  ,   493 ,   1611    ,   105 ,   7855    ,   12173   ,   6877    ,0 };
            const unsigned long dim3484Kuo2Init[]       =   {   1   ,   3   ,   7   ,   7   ,   13  ,   29  ,   89  ,   19  ,   267 ,   563 ,   645 ,   2637    ,   6889    ,   8283    ,   25807   ,0 };
            const unsigned long dim3485Kuo2Init[]       =   {   1   ,   3   ,   5   ,   15  ,   23  ,   55  ,   89  ,   69  ,   363 ,   333 ,   377 ,   3109    ,   5125    ,   1411    ,   22499   ,0 };
            const unsigned long dim3486Kuo2Init[]       =   {   1   ,   1   ,   5   ,   3   ,   29  ,   5   ,   11  ,   151 ,   455 ,   667 ,   1697    ,   1087    ,   1147    ,   12269   ,   14291   ,0 };
            const unsigned long dim3487Kuo2Init[]       =   {   1   ,   3   ,   7   ,   11  ,   1   ,   29  ,   21  ,   209 ,   173 ,   719 ,   1679    ,   1637    ,   6835    ,   8629    ,   27391   ,0 };
            const unsigned long dim3488Kuo2Init[]       =   {   1   ,   3   ,   7   ,   7   ,   23  ,   47  ,   125 ,   145 ,   93  ,   425 ,   2039    ,   2665    ,   3219    ,   12491   ,   12731   ,0 };
            const unsigned long dim3489Kuo2Init[]       =   {   1   ,   1   ,   3   ,   9   ,   19  ,   55  ,   99  ,   253 ,   39  ,   189 ,   559 ,   1893    ,   2155    ,   7165    ,   7705    ,0 };
            const unsigned long dim3490Kuo2Init[]       =   {   1   ,   1   ,   7   ,   15  ,   1   ,   63  ,   85  ,   9   ,   281 ,   551 ,   1697    ,   1341    ,   6039    ,   3315    ,   26821   ,0 };
            const unsigned long dim3491Kuo2Init[]       =   {   1   ,   3   ,   1   ,   15  ,   19  ,   29  ,   63  ,   221 ,   201 ,   671 ,   529 ,   1221    ,   4455    ,   2769    ,   941 ,0 };
            const unsigned long dim3492Kuo2Init[]       =   {   1   ,   3   ,   5   ,   11  ,   23  ,   17  ,   113 ,   167 ,   173 ,   17  ,   1911    ,   1995    ,   6459    ,   1517    ,   19169   ,0 };
            const unsigned long dim3493Kuo2Init[]       =   {   1   ,   3   ,   5   ,   7   ,   23  ,   7   ,   77  ,   231 ,   157 ,   237 ,   1855    ,   2771    ,   7871    ,   13523   ,   18143   ,0 };
            const unsigned long dim3494Kuo2Init[]       =   {   1   ,   3   ,   5   ,   5   ,   17  ,   49  ,   59  ,   175 ,   269 ,   765 ,   1863    ,   3661    ,   4979    ,   7029    ,   4071    ,0 };
            const unsigned long dim3495Kuo2Init[]       =   {   1   ,   3   ,   3   ,   7   ,   31  ,   53  ,   9   ,   89  ,   353 ,   557 ,   1805    ,   3153    ,   4761    ,   14547   ,   20487   ,0 };
            const unsigned long dim3496Kuo2Init[]       =   {   1   ,   1   ,   1   ,   15  ,   23  ,   45  ,   31  ,   151 ,   299 ,   745 ,   1899    ,   3113    ,   4177    ,   107 ,   28863   ,0 };
            const unsigned long dim3497Kuo2Init[]       =   {   1   ,   3   ,   7   ,   3   ,   5   ,   39  ,   21  ,   145 ,   487 ,   35  ,   1891    ,   3985    ,   2501    ,   16217   ,   11963   ,0 };
            const unsigned long dim3498Kuo2Init[]       =   {   1   ,   3   ,   1   ,   3   ,   29  ,   7   ,   15  ,   207 ,   235 ,   789 ,   1195    ,   253 ,   6443    ,   8921    ,   26009   ,0 };
            const unsigned long dim3499Kuo2Init[]       =   {   1   ,   3   ,   1   ,   15  ,   23  ,   53  ,   105 ,   233 ,   71  ,   595 ,   765 ,   2383    ,   7959    ,   3543    ,   5795    ,0 };
            const unsigned long dim3500Kuo2Init[]       =   {   1   ,   3   ,   1   ,   9   ,   1   ,   13  ,   95  ,   113 ,   311 ,   839 ,   1375    ,   4083    ,   7621    ,   4261    ,   6067    ,0 };
            const unsigned long dim3501Kuo2Init[]       =   {   1   ,   1   ,   3   ,   1   ,   25  ,   41  ,   111 ,   185 ,   15  ,   779 ,   375 ,   927 ,   3487    ,   8293    ,   11101   ,0 };
            const unsigned long dim3502Kuo2Init[]       =   {   1   ,   3   ,   3   ,   7   ,   3   ,   15  ,   79  ,   233 ,   255 ,   279 ,   171 ,   1659    ,   2131    ,   3659    ,   2381    ,0 };
            const unsigned long dim3503Kuo2Init[]       =   {   1   ,   3   ,   5   ,   15  ,   21  ,   47  ,   19  ,   163 ,   481 ,   135 ,   875 ,   1143    ,   5929    ,   1269    ,   30729   ,0 };
            const unsigned long dim3504Kuo2Init[]       =   {   1   ,   3   ,   5   ,   9   ,   15  ,   33  ,   45  ,   83  ,   345 ,   751 ,   2045    ,   3221    ,   3101    ,   8727    ,   3785    ,0 };
            const unsigned long dim3505Kuo2Init[]       =   {   1   ,   3   ,   3   ,   15  ,   29  ,   49  ,   95  ,   83  ,   31  ,   433 ,   693 ,   2459    ,   3563    ,   12973   ,   17627   ,0 };
            const unsigned long dim3506Kuo2Init[]       =   {   1   ,   3   ,   3   ,   15  ,   31  ,   19  ,   97  ,   165 ,   353 ,   109 ,   1957    ,   1493    ,   3025    ,   15721   ,   21733   ,0 };
            const unsigned long dim3507Kuo2Init[]       =   {   1   ,   1   ,   3   ,   15  ,   17  ,   55  ,   23  ,   47  ,   473 ,   803 ,   1921    ,   2523    ,   6493    ,   1881    ,   2277    ,0 };
            const unsigned long dim3508Kuo2Init[]       =   {   1   ,   3   ,   7   ,   1   ,   11  ,   11  ,   27  ,   137 ,   259 ,   825 ,   1093    ,   305 ,   1289    ,   9177    ,   22177   ,0 };
            const unsigned long dim3509Kuo2Init[]       =   {   1   ,   1   ,   3   ,   3   ,   5   ,   13  ,   85  ,   191 ,   319 ,   533 ,   1579    ,   351 ,   987 ,   13229   ,   8727    ,0 };
            const unsigned long dim3510Kuo2Init[]       =   {   1   ,   3   ,   5   ,   11  ,   21  ,   21  ,   21  ,   107 ,   201 ,   281 ,   809 ,   2331    ,   1513    ,   10721   ,   31379   ,0 };
            const unsigned long dim3511Kuo2Init[]       =   {   1   ,   3   ,   3   ,   3   ,   1   ,   21  ,   55  ,   157 ,   487 ,   303 ,   891 ,   2035    ,   7329    ,   16203   ,   11811   ,0 };
            const unsigned long dim3512Kuo2Init[]       =   {   1   ,   1   ,   3   ,   15  ,   3   ,   35  ,   91  ,   157 ,   501 ,   169 ,   1437    ,   1765    ,   395 ,   4991    ,   19543   ,0 };
            const unsigned long dim3513Kuo2Init[]       =   {   1   ,   3   ,   7   ,   7   ,   1   ,   43  ,   3   ,   33  ,   13  ,   393 ,   805 ,   1453    ,   6409    ,   3153    ,   3609    ,0 };
            const unsigned long dim3514Kuo2Init[]       =   {   1   ,   3   ,   5   ,   9   ,   25  ,   57  ,   13  ,   189 ,   333 ,   639 ,   1241    ,   277 ,   3665    ,   16025   ,   12687   ,0 };
            const unsigned long dim3515Kuo2Init[]       =   {   1   ,   1   ,   3   ,   9   ,   5   ,   31  ,   45  ,   93  ,   193 ,   469 ,   1325    ,   3109    ,   2115    ,   15903   ,   2211    ,0 };
            const unsigned long dim3516Kuo2Init[]       =   {   1   ,   3   ,   5   ,   15  ,   9   ,   37  ,   125 ,   163 ,   75  ,   505 ,   383 ,   705 ,   2155    ,   15999   ,   29139   ,0 };
            const unsigned long dim3517Kuo2Init[]       =   {   1   ,   1   ,   1   ,   5   ,   13  ,   33  ,   37  ,   101 ,   459 ,   943 ,   821 ,   975 ,   5511    ,   3923    ,   12037   ,0 };
            const unsigned long dim3518Kuo2Init[]       =   {   1   ,   3   ,   7   ,   3   ,   11  ,   45  ,   77  ,   79  ,   35  ,   289 ,   499 ,   2285    ,   1179    ,   953 ,   19641   ,0 };
            const unsigned long dim3519Kuo2Init[]       =   {   1   ,   3   ,   5   ,   1   ,   21  ,   25  ,   27  ,   13  ,   49  ,   365 ,   2047    ,   3015    ,   6483    ,   15287   ,   21813   ,0 };
            const unsigned long dim3520Kuo2Init[]       =   {   1   ,   1   ,   1   ,   11  ,   21  ,   5   ,   113 ,   215 ,   247 ,   377 ,   625 ,   2481    ,   7681    ,   343 ,   29051   ,0 };
            const unsigned long dim3521Kuo2Init[]       =   {   1   ,   1   ,   3   ,   15  ,   17  ,   33  ,   85  ,   143 ,   97  ,   443 ,   2029    ,   2343    ,   2539    ,   6033    ,   23371   ,0 };
            const unsigned long dim3522Kuo2Init[]       =   {   1   ,   1   ,   7   ,   13  ,   31  ,   31  ,   59  ,   135 ,   415 ,   1015    ,   1295    ,   2477    ,   1119    ,   15633   ,   23699   ,0 };
            const unsigned long dim3523Kuo2Init[]       =   {   1   ,   3   ,   5   ,   3   ,   3   ,   23  ,   61  ,   135 ,   273 ,   501 ,   1245    ,   427 ,   6981    ,   12135   ,   15081   ,0 };
            const unsigned long dim3524Kuo2Init[]       =   {   1   ,   3   ,   7   ,   1   ,   15  ,   35  ,   15  ,   127 ,   57  ,   813 ,   1769    ,   1919    ,   659 ,   7051    ,   30567   ,0 };
            const unsigned long dim3525Kuo2Init[]       =   {   1   ,   3   ,   1   ,   5   ,   31  ,   11  ,   85  ,   131 ,   483 ,   33  ,   1311    ,   1101    ,   3939    ,   8531    ,   21019   ,0 };
            const unsigned long dim3526Kuo2Init[]       =   {   1   ,   1   ,   5   ,   15  ,   17  ,   17  ,   127 ,   33  ,   111 ,   805 ,   1139    ,   2631    ,   7015    ,   11253   ,   13141   ,0 };
            const unsigned long dim3527Kuo2Init[]       =   {   1   ,   3   ,   3   ,   9   ,   5   ,   5   ,   55  ,   223 ,   237 ,   641 ,   1081    ,   1145    ,   3773    ,   15829   ,   2309    ,0 };
            const unsigned long dim3528Kuo2Init[]       =   {   1   ,   1   ,   7   ,   3   ,   29  ,   35  ,   97  ,   197 ,   259 ,   665 ,   1101    ,   2423    ,   1689    ,   12363   ,   16037   ,0 };
            const unsigned long dim3529Kuo2Init[]       =   {   1   ,   1   ,   7   ,   15  ,   7   ,   59  ,   59  ,   199 ,   187 ,   83  ,   803 ,   1391    ,   6701    ,   8105    ,   22967   ,0 };
            const unsigned long dim3530Kuo2Init[]       =   {   1   ,   1   ,   5   ,   9   ,   1   ,   59  ,   91  ,   253 ,   277 ,   1019    ,   87  ,   3727    ,   3367    ,   13351   ,   10217   ,0 };
            const unsigned long dim3531Kuo2Init[]       =   {   1   ,   3   ,   3   ,   11  ,   19  ,   39  ,   35  ,   29  ,   259 ,   525 ,   555 ,   1683    ,   1885    ,   8957    ,   17431   ,0 };
            const unsigned long dim3532Kuo2Init[]       =   {   1   ,   1   ,   1   ,   5   ,   3   ,   5   ,   99  ,   49  ,   19  ,   211 ,   135 ,   1677    ,   37  ,   14611   ,   25219   ,0 };
            const unsigned long dim3533Kuo2Init[]       =   {   1   ,   3   ,   7   ,   3   ,   15  ,   31  ,   3   ,   201 ,   329 ,   755 ,   1013    ,   1917    ,   785 ,   14243   ,   28877   ,0 };
            const unsigned long dim3534Kuo2Init[]       =   {   1   ,   3   ,   5   ,   15  ,   3   ,   13  ,   113 ,   41  ,   123 ,   157 ,   165 ,   125 ,   5147    ,   10085   ,   13903   ,0 };
            const unsigned long dim3535Kuo2Init[]       =   {   1   ,   3   ,   7   ,   11  ,   15  ,   29  ,   127 ,   249 ,   51  ,   543 ,   667 ,   2401    ,   2623    ,   2447    ,   9471    ,0 };
            const unsigned long dim3536Kuo2Init[]       =   {   1   ,   3   ,   3   ,   1   ,   7   ,   55  ,   127 ,   165 ,   235 ,   271 ,   1721    ,   2301    ,   1705    ,   8209    ,   10837   ,0 };
            const unsigned long dim3537Kuo2Init[]       =   {   1   ,   1   ,   3   ,   15  ,   31  ,   27  ,   49  ,   13  ,   33  ,   797 ,   1533    ,   3325    ,   7257    ,   21  ,   16787   ,0 };
            const unsigned long dim3538Kuo2Init[]       =   {   1   ,   1   ,   3   ,   13  ,   17  ,   29  ,   43  ,   79  ,   219 ,   133 ,   15  ,   2707    ,   5517    ,   15923   ,   18125   ,0 };
            const unsigned long dim3539Kuo2Init[]       =   {   1   ,   1   ,   7   ,   1   ,   21  ,   1   ,   61  ,   233 ,   123 ,   599 ,   201 ,   1431    ,   3637    ,   7921    ,   14069   ,0 };
            const unsigned long dim3540Kuo2Init[]       =   {   1   ,   3   ,   3   ,   9   ,   1   ,   1   ,   113 ,   187 ,   189 ,   1013    ,   1085    ,   3693    ,   337 ,   9807    ,   25367   ,0 };
            const unsigned long dim3541Kuo2Init[]       =   {   1   ,   3   ,   3   ,   13  ,   11  ,   13  ,   125 ,   243 ,   77  ,   671 ,   177 ,   647 ,   1985    ,   15007   ,   27919   ,0 };
            const unsigned long dim3542Kuo2Init[]       =   {   1   ,   3   ,   7   ,   13  ,   23  ,   3   ,   85  ,   13  ,   15  ,   443 ,   673 ,   1   ,   8119    ,   16227   ,   4497    ,0 };
            const unsigned long dim3543Kuo2Init[]       =   {   1   ,   1   ,   3   ,   3   ,   5   ,   25  ,   113 ,   125 ,   503 ,   141 ,   945 ,   631 ,   6413    ,   2997    ,   28309   ,0 };
            const unsigned long dim3544Kuo2Init[]       =   {   1   ,   3   ,   5   ,   5   ,   15  ,   37  ,   7   ,   169 ,   423 ,   917 ,   485 ,   3557    ,   1093    ,   11613   ,   11301   ,0 };
            const unsigned long dim3545Kuo2Init[]       =   {   1   ,   1   ,   7   ,   3   ,   5   ,   35  ,   45  ,   99  ,   215 ,   537 ,   1055    ,   3109    ,   7745    ,   15741   ,   30499   ,0 };
            const unsigned long dim3546Kuo2Init[]       =   {   1   ,   3   ,   5   ,   13  ,   19  ,   21  ,   37  ,   127 ,   83  ,   489 ,   1809    ,   2479    ,   5875    ,   2559    ,   13593   ,0 };
            const unsigned long dim3547Kuo2Init[]       =   {   1   ,   1   ,   3   ,   15  ,   11  ,   31  ,   19  ,   111 ,   245 ,   619 ,   793 ,   453 ,   2557    ,   16269   ,   2823    ,0 };
            const unsigned long dim3548Kuo2Init[]       =   {   1   ,   3   ,   3   ,   7   ,   31  ,   1   ,   99  ,   255 ,   187 ,   277 ,   71  ,   1985    ,   3919    ,   8445    ,   10791   ,0 };
            const unsigned long dim3549Kuo2Init[]       =   {   1   ,   3   ,   1   ,   11  ,   25  ,   1   ,   63  ,   5   ,   219 ,   601 ,   1155    ,   2411    ,   7673    ,   8979    ,   23729   ,0 };
            const unsigned long dim3550Kuo2Init[]       =   {   1   ,   1   ,   1   ,   11  ,   31  ,   59  ,   11  ,   37  ,   361 ,   145 ,   879 ,   2655    ,   3591    ,   13615   ,   25559   ,0 };
            const unsigned long dim3551Kuo2Init[]       =   {   1   ,   3   ,   5   ,   7   ,   15  ,   5   ,   117 ,   175 ,   417 ,   825 ,   585 ,   1429    ,   5163    ,   335 ,   32281   ,0 };
            const unsigned long dim3552Kuo2Init[]       =   {   1   ,   1   ,   1   ,   5   ,   27  ,   9   ,   107 ,   181 ,   493 ,   651 ,   1317    ,   2059    ,   4155    ,   15039   ,   27693   ,0 };
            const unsigned long dim3553Kuo2Init[]       =   {   1   ,   3   ,   7   ,   3   ,   1   ,   11  ,   13  ,   187 ,   173 ,   991 ,   1743    ,   2403    ,   5279    ,   11387   ,   11035   ,0 };
            const unsigned long dim3554Kuo2Init[]       =   {   1   ,   3   ,   1   ,   11  ,   31  ,   39  ,   87  ,   63  ,   177 ,   581 ,   609 ,   3315    ,   4859    ,   9597    ,   29901   ,0 };
            const unsigned long dim3555Kuo2Init[]       =   {   1   ,   3   ,   3   ,   11  ,   15  ,   47  ,   1   ,   53  ,   91  ,   39  ,   37  ,   1697    ,   6719    ,   383 ,   1793    ,0 };
            const unsigned long dim3556Kuo2Init[]       =   {   1   ,   3   ,   5   ,   5   ,   7   ,   19  ,   123 ,   167 ,   509 ,   585 ,   1153    ,   1803    ,   987 ,   2633    ,   14177   ,0 };
            const unsigned long dim3557Kuo2Init[]       =   {   1   ,   1   ,   5   ,   7   ,   9   ,   5   ,   59  ,   137 ,   209 ,   95  ,   519 ,   177 ,   5627    ,   7843    ,   23207   ,0 };
            const unsigned long dim3558Kuo2Init[]       =   {   1   ,   1   ,   1   ,   15  ,   7   ,   3   ,   89  ,   31  ,   125 ,   391 ,   545 ,   3781    ,   985 ,   7355    ,   27935   ,0 };
            const unsigned long dim3559Kuo2Init[]       =   {   1   ,   3   ,   7   ,   3   ,   31  ,   11  ,   7   ,   147 ,   155 ,   421 ,   195 ,   861 ,   3787    ,   10843   ,   2239    ,0 };
            const unsigned long dim3560Kuo2Init[]       =   {   1   ,   1   ,   3   ,   15  ,   23  ,   29  ,   71  ,   79  ,   279 ,   157 ,   853 ,   421 ,   6499    ,   7695    ,   5527    ,0 };
            const unsigned long dim3561Kuo2Init[]       =   {   1   ,   3   ,   3   ,   11  ,   3   ,   63  ,   49  ,   199 ,   5   ,   457 ,   479 ,   3419    ,   6649    ,   14259   ,   17789   ,0 };
            const unsigned long dim3562Kuo2Init[]       =   {   1   ,   1   ,   3   ,   3   ,   23  ,   7   ,   85  ,   105 ,   281 ,   929 ,   1487    ,   3797    ,   3487    ,   15959   ,   25859   ,0 };
            const unsigned long dim3563Kuo2Init[]       =   {   1   ,   3   ,   5   ,   5   ,   23  ,   27  ,   33  ,   5   ,   283 ,   853 ,   1939    ,   3745    ,   2995    ,   14715   ,   32739   ,0 };
            const unsigned long dim3564Kuo2Init[]       =   {   1   ,   3   ,   3   ,   5   ,   9   ,   7   ,   111 ,   53  ,   299 ,   237 ,   1263    ,   2053    ,   6359    ,   15617   ,   29347   ,0 };
            const unsigned long dim3565Kuo2Init[]       =   {   1   ,   3   ,   3   ,   13  ,   29  ,   61  ,   55  ,   181 ,   49  ,   377 ,   729 ,   117 ,   8125    ,   15317   ,   32363   ,0 };
            const unsigned long dim3566Kuo2Init[]       =   {   1   ,   1   ,   7   ,   1   ,   9   ,   31  ,   77  ,   107 ,   21  ,   375 ,   1719    ,   1701    ,   5439    ,   4187    ,   32397   ,0 };
            const unsigned long dim3567Kuo2Init[]       =   {   1   ,   1   ,   3   ,   11  ,   31  ,   41  ,   101 ,   1   ,   171 ,   401 ,   1711    ,   3027    ,   161 ,   441 ,   15589   ,0 };
            const unsigned long dim3568Kuo2Init[]       =   {   1   ,   1   ,   3   ,   15  ,   25  ,   35  ,   7   ,   85  ,   45  ,   409 ,   1769    ,   767 ,   2615    ,   1313    ,   7031    ,0 };
            const unsigned long dim3569Kuo2Init[]       =   {   1   ,   3   ,   5   ,   7   ,   29  ,   7   ,   73  ,   25  ,   397 ,   767 ,   315 ,   2105    ,   4677    ,   12081   ,   7763    ,0 };
            const unsigned long dim3570Kuo2Init[]       =   {   1   ,   3   ,   5   ,   13  ,   31  ,   63  ,   87  ,   29  ,   81  ,   969 ,   1931    ,   3713    ,   7217    ,   15899   ,   20505   ,0 };
            const unsigned long dim3571Kuo2Init[]       =   {   1   ,   1   ,   5   ,   3   ,   15  ,   31  ,   121 ,   141 ,   209 ,   697 ,   941 ,   1565    ,   1363    ,   11919   ,   24479   ,0 };
            const unsigned long dim3572Kuo2Init[]       =   {   1   ,   1   ,   7   ,   15  ,   23  ,   9   ,   5   ,   179 ,   437 ,   259 ,   1223    ,   2159    ,   4825    ,   2977    ,   20851   ,0 };
            const unsigned long dim3573Kuo2Init[]       =   {   1   ,   1   ,   1   ,   13  ,   13  ,   25  ,   33  ,   169 ,   91  ,   1001    ,   1931    ,   1853    ,   6173    ,   5411    ,   21081   ,0 };
            const unsigned long dim3574Kuo2Init[]       =   {   1   ,   1   ,   1   ,   13  ,   1   ,   53  ,   109 ,   53  ,   345 ,   519 ,   771 ,   1673    ,   6721    ,   3585    ,   6289    ,0 };
            const unsigned long dim3575Kuo2Init[]       =   {   1   ,   3   ,   3   ,   3   ,   27  ,   37  ,   77  ,   243 ,   207 ,   791 ,   1585    ,   1965    ,   3327    ,   4719    ,   7121    ,0 };
            const unsigned long dim3576Kuo2Init[]       =   {   1   ,   3   ,   1   ,   7   ,   3   ,   15  ,   115 ,   175 ,   303 ,   59  ,   835 ,   3347    ,   6055    ,   4959    ,   6963    ,0 };
            const unsigned long dim3577Kuo2Init[]       =   {   1   ,   3   ,   5   ,   1   ,   11  ,   5   ,   99  ,   139 ,   351 ,   709 ,   281 ,   2635    ,   379 ,   2543    ,   29879   ,0 };
            const unsigned long dim3578Kuo2Init[]       =   {   1   ,   3   ,   1   ,   11  ,   7   ,   9   ,   117 ,   209 ,   221 ,   667 ,   491 ,   2833    ,   1113    ,   12289   ,   7947    ,0 };
            const unsigned long dim3579Kuo2Init[]       =   {   1   ,   3   ,   5   ,   9   ,   17  ,   11  ,   11  ,   43  ,   475 ,   169 ,   1167    ,   2601    ,   2733    ,   243 ,   17279   ,0 };
            const unsigned long dim3580Kuo2Init[]       =   {   1   ,   1   ,   5   ,   15  ,   31  ,   43  ,   7   ,   107 ,   309 ,   993 ,   635 ,   3779    ,   1041    ,   9997    ,   23049   ,0 };
            const unsigned long dim3581Kuo2Init[]       =   {   1   ,   3   ,   5   ,   9   ,   21  ,   15  ,   21  ,   41  ,   451 ,   1021    ,   131 ,   1539    ,   7147    ,   1639    ,   31157   ,0 };
            const unsigned long dim3582Kuo2Init[]       =   {   1   ,   1   ,   3   ,   1   ,   25  ,   37  ,   67  ,   91  ,   463 ,   63  ,   653 ,   4045    ,   1255    ,   3315    ,   10747   ,0 };
            const unsigned long dim3583Kuo2Init[]       =   {   1   ,   1   ,   3   ,   3   ,   29  ,   3   ,   95  ,   125 ,   477 ,   845 ,   563 ,   3095    ,   2985    ,   8935    ,   23247   ,0 };
            const unsigned long dim3584Kuo2Init[]       =   {   1   ,   3   ,   5   ,   13  ,   11  ,   5   ,   21  ,   193 ,   121 ,   857 ,   997 ,   3487    ,   4613    ,   12231   ,   7819    ,0 };
            const unsigned long dim3585Kuo2Init[]       =   {   1   ,   1   ,   3   ,   15  ,   27  ,   1   ,   65  ,   219 ,   241 ,   99  ,   1193    ,   3157    ,   7641    ,   10327   ,   25241   ,0 };
            const unsigned long dim3586Kuo2Init[]       =   {   1   ,   3   ,   7   ,   1   ,   3   ,   25  ,   61  ,   191 ,   501 ,   219 ,   1863    ,   4039    ,   5879    ,   11203   ,   10581   ,0 };
            const unsigned long dim3587Kuo2Init[]       =   {   1   ,   1   ,   5   ,   3   ,   23  ,   39  ,   81  ,   37  ,   477 ,   969 ,   1   ,   1161    ,   7171    ,   3595    ,   375 ,0 };
            const unsigned long dim3588Kuo2Init[]       =   {   1   ,   3   ,   3   ,   3   ,   21  ,   17  ,   99  ,   101 ,   165 ,   849 ,   607 ,   1863    ,   6675    ,   13557   ,   3349    ,0 };
            const unsigned long dim3589Kuo2Init[]       =   {   1   ,   3   ,   5   ,   15  ,   9   ,   1   ,   125 ,   93  ,   211 ,   971 ,   1275    ,   545 ,   3625    ,   1503    ,   10735   ,0 };
            const unsigned long dim3590Kuo2Init[]       =   {   1   ,   3   ,   3   ,   7   ,   19  ,   53  ,   1   ,   169 ,   453 ,   227 ,   425 ,   2831    ,   6751    ,   11693   ,   20483   ,0 };
            const unsigned long dim3591Kuo2Init[]       =   {   1   ,   1   ,   1   ,   3   ,   9   ,   19  ,   121 ,   203 ,   55  ,   485 ,   37  ,   3977    ,   2193    ,   9819    ,   4635    ,0 };
            const unsigned long dim3592Kuo2Init[]       =   {   1   ,   3   ,   5   ,   15  ,   17  ,   35  ,   121 ,   207 ,   233 ,   149 ,   1607    ,   3599    ,   1013    ,   2295    ,   9113    ,0 };
            const unsigned long dim3593Kuo2Init[]       =   {   1   ,   3   ,   3   ,   13  ,   29  ,   17  ,   39  ,   85  ,   47  ,   477 ,   1967    ,   701 ,   1063    ,   3521    ,   23583   ,0 };
            const unsigned long dim3594Kuo2Init[]       =   {   1   ,   3   ,   3   ,   3   ,   23  ,   57  ,   69  ,   49  ,   55  ,   81  ,   1195    ,   347 ,   3869    ,   4843    ,   97  ,0 };
            const unsigned long dim3595Kuo2Init[]       =   {   1   ,   3   ,   3   ,   1   ,   9   ,   19  ,   31  ,   199 ,   173 ,   5   ,   1557    ,   2669    ,   5471    ,   7203    ,   27289   ,0 };
            const unsigned long dim3596Kuo2Init[]       =   {   1   ,   3   ,   3   ,   3   ,   3   ,   1   ,   121 ,   195 ,   193 ,   359 ,   1269    ,   973 ,   3443    ,   1669    ,   31895   ,0 };
            const unsigned long dim3597Kuo2Init[]       =   {   1   ,   1   ,   3   ,   7   ,   3   ,   43  ,   39  ,   77  ,   149 ,   111 ,   425 ,   1567    ,   1289    ,   13443   ,   30175   ,0 };
            const unsigned long dim3598Kuo2Init[]       =   {   1   ,   3   ,   3   ,   5   ,   15  ,   55  ,   67  ,   209 ,   167 ,   439 ,   825 ,   3165    ,   6537    ,   10291   ,   2561    ,0 };
            const unsigned long dim3599Kuo2Init[]       =   {   1   ,   1   ,   5   ,   7   ,   1   ,   27  ,   49  ,   67  ,   123 ,   905 ,   759 ,   1471    ,   4379    ,   8211    ,   1829    ,0 };
            const unsigned long dim3600Kuo2Init[]       =   {   1   ,   1   ,   7   ,   13  ,   23  ,   7   ,   99  ,   169 ,   349 ,   889 ,   1671    ,   3039    ,   6757    ,   7071    ,   20707   ,0 };
            const unsigned long dim3601Kuo2Init[]       =   {   1   ,   3   ,   3   ,   5   ,   21  ,   59  ,   103 ,   167 ,   69  ,   77  ,   155 ,   301 ,   6647    ,   13487   ,   19241   ,0 };
            const unsigned long dim3602Kuo2Init[]       =   {   1   ,   1   ,   7   ,   9   ,   5   ,   53  ,   19  ,   17  ,   239 ,   691 ,   1079    ,   3485    ,   4941    ,   1769    ,   6953    ,0 };
            const unsigned long dim3603Kuo2Init[]       =   {   1   ,   3   ,   5   ,   13  ,   15  ,   37  ,   87  ,   137 ,   205 ,   743 ,   965 ,   3911    ,   1421    ,   3003    ,   26639   ,0 };
            const unsigned long dim3604Kuo2Init[]       =   {   1   ,   1   ,   5   ,   7   ,   31  ,   17  ,   49  ,   151 ,   159 ,   323 ,   1393    ,   1677    ,   2335    ,   289 ,   30059   ,0 };
            const unsigned long dim3605Kuo2Init[]       =   {   1   ,   1   ,   7   ,   11  ,   25  ,   37  ,   93  ,   221 ,   441 ,   329 ,   421 ,   903 ,   2803    ,   11455   ,   20521   ,0 };
            const unsigned long dim3606Kuo2Init[]       =   {   1   ,   3   ,   5   ,   3   ,   11  ,   1   ,   127 ,   231 ,   451 ,   333 ,   1501    ,   1353    ,   1741    ,   7149    ,   17527   ,0 };
            const unsigned long dim3607Kuo2Init[]       =   {   1   ,   1   ,   7   ,   9   ,   25  ,   33  ,   101 ,   255 ,   69  ,   693 ,   1579    ,   2307    ,   6071    ,   6711    ,   20071   ,0 };
            const unsigned long dim3608Kuo2Init[]       =   {   1   ,   3   ,   5   ,   15  ,   29  ,   61  ,   39  ,   31  ,   193 ,   573 ,   1151    ,   2237    ,   125 ,   7851    ,   29569   ,0 };
            const unsigned long dim3609Kuo2Init[]       =   {   1   ,   3   ,   5   ,   15  ,   25  ,   15  ,   93  ,   93  ,   309 ,   765 ,   23  ,   2111    ,   7001    ,   3885    ,   18391   ,0 };
            const unsigned long dim3610Kuo2Init[]       =   {   1   ,   1   ,   1   ,   5   ,   29  ,   9   ,   119 ,   165 ,   127 ,   847 ,   753 ,   4089    ,   5739    ,   10385   ,   28691   ,0 };
            const unsigned long dim3611Kuo2Init[]       =   {   1   ,   1   ,   1   ,   7   ,   5   ,   23  ,   41  ,   39  ,   429 ,   851 ,   1221    ,   3817    ,   8049    ,   14547   ,   9277    ,0 };
            const unsigned long dim3612Kuo2Init[]       =   {   1   ,   3   ,   7   ,   1   ,   7   ,   23  ,   105 ,   231 ,   241 ,   49  ,   1083    ,   3903    ,   1469    ,   11055   ,   27949   ,0 };
            const unsigned long dim3613Kuo2Init[]       =   {   1   ,   1   ,   7   ,   15  ,   9   ,   11  ,   127 ,   83  ,   425 ,   671 ,   657 ,   2431    ,   1003    ,   2997    ,   5593    ,0 };
            const unsigned long dim3614Kuo2Init[]       =   {   1   ,   3   ,   1   ,   13  ,   15  ,   35  ,   73  ,   193 ,   155 ,   183 ,   2033    ,   2385    ,   7979    ,   8551    ,   2507    ,0 };
            const unsigned long dim3615Kuo2Init[]       =   {   1   ,   3   ,   5   ,   9   ,   29  ,   43  ,   59  ,   17  ,   227 ,   159 ,   1839    ,   4047    ,   1667    ,   2541    ,   30517   ,0 };
            const unsigned long dim3616Kuo2Init[]       =   {   1   ,   3   ,   7   ,   9   ,   31  ,   45  ,   109 ,   243 ,   163 ,   871 ,   951 ,   2463    ,   5573    ,   8415    ,   16857   ,0 };
            const unsigned long dim3617Kuo2Init[]       =   {   1   ,   3   ,   7   ,   7   ,   11  ,   57  ,   93  ,   243 ,   227 ,   863 ,   483 ,   2371    ,   6781    ,   6811    ,   23423   ,0 };
            const unsigned long dim3618Kuo2Init[]       =   {   1   ,   3   ,   1   ,   1   ,   13  ,   9   ,   113 ,   3   ,   207 ,   735 ,   1021    ,   1933    ,   6635    ,   13609   ,   19651   ,0 };
            const unsigned long dim3619Kuo2Init[]       =   {   1   ,   1   ,   7   ,   9   ,   5   ,   33  ,   33  ,   159 ,   273 ,   291 ,   159 ,   1693    ,   217 ,   12487   ,   21853   ,0 };
            const unsigned long dim3620Kuo2Init[]       =   {   1   ,   3   ,   1   ,   13  ,   27  ,   51  ,   65  ,   117 ,   485 ,   471 ,   29  ,   3929    ,   4509    ,   13577   ,   6745    ,0 };
            const unsigned long dim3621Kuo2Init[]       =   {   1   ,   3   ,   3   ,   13  ,   5   ,   29  ,   23  ,   81  ,   353 ,   727 ,   1437    ,   3239    ,   717 ,   13857   ,   18881   ,0 };
            const unsigned long dim3622Kuo2Init[]       =   {   1   ,   3   ,   1   ,   1   ,   11  ,   41  ,   97  ,   73  ,   271 ,   847 ,   471 ,   2857    ,   691 ,   10951   ,   31465   ,0 };
            const unsigned long dim3623Kuo2Init[]       =   {   1   ,   1   ,   1   ,   9   ,   11  ,   59  ,   3   ,   71  ,   235 ,   715 ,   1349    ,   199 ,   2853    ,   13641   ,   5025    ,0 };
            const unsigned long dim3624Kuo2Init[]       =   {   1   ,   3   ,   1   ,   9   ,   23  ,   7   ,   23  ,   49  ,   31  ,   145 ,   533 ,   2625    ,   207 ,   3661    ,   25215   ,0 };
            const unsigned long dim3625Kuo2Init[]       =   {   1   ,   3   ,   5   ,   15  ,   17  ,   17  ,   53  ,   129 ,   105 ,   285 ,   1423    ,   3985    ,   8015    ,   3565    ,   24495   ,0 };
            const unsigned long dim3626Kuo2Init[]       =   {   1   ,   1   ,   1   ,   13  ,   7   ,   29  ,   93  ,   123 ,   167 ,   845 ,   69  ,   2297    ,   543 ,   9165    ,   4029    ,0 };
            const unsigned long dim3627Kuo2Init[]       =   {   1   ,   3   ,   7   ,   15  ,   19  ,   57  ,   71  ,   99  ,   95  ,   165 ,   1875    ,   4011    ,   4779    ,   4057    ,   16331   ,0 };
            const unsigned long dim3628Kuo2Init[]       =   {   1   ,   1   ,   5   ,   11  ,   19  ,   43  ,   45  ,   83  ,   383 ,   111 ,   1859    ,   2767    ,   1469    ,   3687    ,   18343   ,0 };
            const unsigned long dim3629Kuo2Init[]       =   {   1   ,   1   ,   5   ,   1   ,   11  ,   7   ,   63  ,   51  ,   487 ,   609 ,   1619    ,   3399    ,   1   ,   7253    ,   24325   ,0 };
            const unsigned long dim3630Kuo2Init[]       =   {   1   ,   3   ,   3   ,   7   ,   29  ,   27  ,   83  ,   61  ,   353 ,   913 ,   1359    ,   689 ,   7741    ,   3357    ,   7731    ,0 };
            const unsigned long dim3631Kuo2Init[]       =   {   1   ,   1   ,   1   ,   7   ,   31  ,   9   ,   5   ,   221 ,   177 ,   379 ,   1953    ,   3389    ,   3861    ,   10265   ,   1347    ,0 };
            const unsigned long dim3632Kuo2Init[]       =   {   1   ,   3   ,   5   ,   13  ,   7   ,   13  ,   57  ,   9   ,   233 ,   601 ,   719 ,   1855    ,   6481    ,   9987    ,   16463   ,0 };
            const unsigned long dim3633Kuo2Init[]       =   {   1   ,   1   ,   3   ,   1   ,   15  ,   27  ,   67  ,   211 ,   289 ,   969 ,   1719    ,   2329    ,   3665    ,   7133    ,   24645   ,0 };
            const unsigned long dim3634Kuo2Init[]       =   {   1   ,   1   ,   3   ,   9   ,   21  ,   37  ,   63  ,   11  ,   19  ,   797 ,   1591    ,   2037    ,   7731    ,   12533   ,   2799    ,0 };
            const unsigned long dim3635Kuo2Init[]       =   {   1   ,   3   ,   5   ,   9   ,   15  ,   37  ,   1   ,   123 ,   209 ,   321 ,   737 ,   837 ,   561 ,   2631    ,   18547   ,0 };
            const unsigned long dim3636Kuo2Init[]       =   {   1   ,   3   ,   1   ,   1   ,   13  ,   33  ,   27  ,   119 ,   407 ,   675 ,   903 ,   2229    ,   1963    ,   3857    ,   5795    ,0 };
            const unsigned long dim3637Kuo2Init[]       =   {   1   ,   1   ,   5   ,   5   ,   9   ,   31  ,   25  ,   115 ,   471 ,   353 ,   113 ,   3297    ,   87  ,   9413    ,   25209   ,0 };
            const unsigned long dim3638Kuo2Init[]       =   {   1   ,   3   ,   5   ,   7   ,   23  ,   27  ,   29  ,   95  ,   387 ,   747 ,   155 ,   3991    ,   3505    ,   975 ,   6371    ,0 };
            const unsigned long dim3639Kuo2Init[]       =   {   1   ,   1   ,   5   ,   13  ,   15  ,   45  ,   3   ,   213 ,   299 ,   775 ,   525 ,   2639    ,   491 ,   12039   ,   20295   ,0 };
            const unsigned long dim3640Kuo2Init[]       =   {   1   ,   1   ,   7   ,   15  ,   5   ,   47  ,   59  ,   131 ,   165 ,   93  ,   583 ,   3823    ,   7781    ,   16339   ,   16523   ,0 };
            const unsigned long dim3641Kuo2Init[]       =   {   1   ,   3   ,   5   ,   3   ,   25  ,   61  ,   23  ,   75  ,   343 ,   15  ,   347 ,   259 ,   4621    ,   8275    ,   10323   ,0 };
            const unsigned long dim3642Kuo2Init[]       =   {   1   ,   3   ,   3   ,   13  ,   7   ,   25  ,   1   ,   153 ,   199 ,   225 ,   1217    ,   161 ,   1185    ,   10775   ,   6709    ,0 };
            const unsigned long dim3643Kuo2Init[]       =   {   1   ,   1   ,   1   ,   1   ,   1   ,   49  ,   27  ,   67  ,   411 ,   425 ,   1385    ,   3921    ,   6373    ,   14581   ,   25159   ,0 };
            const unsigned long dim3644Kuo2Init[]       =   {   1   ,   1   ,   1   ,   11  ,   5   ,   59  ,   125 ,   239 ,   391 ,   823 ,   969 ,   3287    ,   7257    ,   5405    ,   15  ,0 };
            const unsigned long dim3645Kuo2Init[]       =   {   1   ,   1   ,   5   ,   1   ,   1   ,   17  ,   109 ,   145 ,   289 ,   601 ,   1539    ,   807 ,   1123    ,   14541   ,   15885   ,0 };
            const unsigned long dim3646Kuo2Init[]       =   {   1   ,   1   ,   3   ,   9   ,   23  ,   29  ,   119 ,   79  ,   253 ,   921 ,   1935    ,   4047    ,   7639    ,   10419   ,   26597   ,0 };
            const unsigned long dim3647Kuo2Init[]       =   {   1   ,   1   ,   5   ,   11  ,   27  ,   53  ,   63  ,   123 ,   85  ,   121 ,   481 ,   3581    ,   4545    ,   7275    ,   981 ,0 };
            const unsigned long dim3648Kuo2Init[]       =   {   1   ,   3   ,   3   ,   1   ,   27  ,   5   ,   43  ,   247 ,   299 ,   411 ,   299 ,   2545    ,   4403    ,   9903    ,   16161   ,0 };
            const unsigned long dim3649Kuo2Init[]       =   {   1   ,   3   ,   1   ,   11  ,   29  ,   21  ,   83  ,   37  ,   425 ,   741 ,   899 ,   3643    ,   6641    ,   5241    ,   3627    ,0 };
            const unsigned long dim3650Kuo2Init[]       =   {   1   ,   1   ,   7   ,   9   ,   11  ,   23  ,   43  ,   191 ,   225 ,   101 ,   1423    ,   927 ,   1303    ,   15417   ,   25553   ,0 };
            const unsigned long dim3651Kuo2Init[]       =   {   1   ,   3   ,   1   ,   1   ,   23  ,   47  ,   103 ,   177 ,   437 ,   113 ,   295 ,   699 ,   1225    ,   8259    ,   19469   ,0 };
            const unsigned long dim3652Kuo2Init[]       =   {   1   ,   1   ,   7   ,   3   ,   13  ,   61  ,   119 ,   231 ,   33  ,   217 ,   1419    ,   989 ,   4413    ,   9171    ,   1199    ,0 };
            const unsigned long dim3653Kuo2Init[]       =   {   1   ,   1   ,   1   ,   1   ,   5   ,   13  ,   87  ,   185 ,   69  ,   471 ,   1665    ,   2201    ,   5889    ,   7931    ,   2913    ,0 };
            const unsigned long dim3654Kuo2Init[]       =   {   1   ,   1   ,   1   ,   15  ,   17  ,   7   ,   111 ,   27  ,   91  ,   127 ,   301 ,   3641    ,   561 ,   6697    ,   16767   ,0 };
            const unsigned long dim3655Kuo2Init[]       =   {   1   ,   3   ,   3   ,   9   ,   11  ,   25  ,   127 ,   139 ,   355 ,   475 ,   821 ,   843 ,   493 ,   2713    ,   16877   ,0 };
            const unsigned long dim3656Kuo2Init[]       =   {   1   ,   3   ,   5   ,   7   ,   27  ,   19  ,   51  ,   127 ,   211 ,   869 ,   339 ,   2469    ,   1515    ,   9007    ,   28807   ,0 };
            const unsigned long dim3657Kuo2Init[]       =   {   1   ,   1   ,   3   ,   7   ,   15  ,   23  ,   79  ,   49  ,   185 ,   465 ,   1165    ,   1103    ,   5975    ,   13867   ,   30039   ,0 };
            const unsigned long dim3658Kuo2Init[]       =   {   1   ,   3   ,   1   ,   5   ,   27  ,   41  ,   49  ,   195 ,   23  ,   441 ,   1225    ,   3861    ,   6661    ,   7261    ,   25415   ,0 };
            const unsigned long dim3659Kuo2Init[]       =   {   1   ,   1   ,   3   ,   3   ,   27  ,   41  ,   57  ,   139 ,   87  ,   403 ,   1351    ,   2063    ,   5445    ,   4559    ,   7181    ,0 };
            const unsigned long dim3660Kuo2Init[]       =   {   1   ,   1   ,   3   ,   15  ,   7   ,   35  ,   43  ,   177 ,   63  ,   1021    ,   1157    ,   3831    ,   1515    ,   8117    ,   16567   ,0 };
            const unsigned long dim3661Kuo2Init[]       =   {   1   ,   3   ,   7   ,   9   ,   5   ,   47  ,   59  ,   221 ,   415 ,   541 ,   857 ,   257 ,   6961    ,   16349   ,   28975   ,0 };
            const unsigned long dim3662Kuo2Init[]       =   {   1   ,   3   ,   5   ,   9   ,   13  ,   55  ,   55  ,   63  ,   7   ,   817 ,   693 ,   1183    ,   7417    ,   7087    ,   31173   ,0 };
            const unsigned long dim3663Kuo2Init[]       =   {   1   ,   3   ,   5   ,   9   ,   1   ,   29  ,   17  ,   15  ,   417 ,   559 ,   155 ,   495 ,   6269    ,   11163   ,   15073   ,0 };
            const unsigned long dim3664Kuo2Init[]       =   {   1   ,   1   ,   5   ,   1   ,   29  ,   59  ,   49  ,   253 ,   489 ,   657 ,   1723    ,   1969    ,   471 ,   10821   ,   3527    ,0 };
            const unsigned long dim3665Kuo2Init[]       =   {   1   ,   1   ,   1   ,   5   ,   29  ,   33  ,   121 ,   9   ,   73  ,   33  ,   89  ,   3419    ,   6491    ,   5361    ,   31587   ,0 };
            const unsigned long dim3666Kuo2Init[]       =   {   1   ,   3   ,   3   ,   5   ,   1   ,   49  ,   121 ,   25  ,   481 ,   813 ,   607 ,   757 ,   5493    ,   4353    ,   14355   ,0 };
            const unsigned long dim3667Kuo2Init[]       =   {   1   ,   1   ,   1   ,   1   ,   21  ,   45  ,   61  ,   101 ,   357 ,   875 ,   1381    ,   1731    ,   6505    ,   12891   ,   19595   ,   54305   ,0 };
            const unsigned long dim3668Kuo2Init[]       =   {   1   ,   3   ,   7   ,   7   ,   19  ,   55  ,   99  ,   131 ,   309 ,   303 ,   1575    ,   3241    ,   6799    ,   6543    ,   26973   ,   46843   ,0 };
            const unsigned long dim3669Kuo2Init[]       =   {   1   ,   1   ,   3   ,   3   ,   27  ,   11  ,   63  ,   237 ,   13  ,   949 ,   1359    ,   1563    ,   5501    ,   5293    ,   923 ,   3413    ,0 };
            const unsigned long dim3670Kuo2Init[]       =   {   1   ,   1   ,   3   ,   7   ,   13  ,   55  ,   117 ,   143 ,   225 ,   161 ,   805 ,   3259    ,   4113    ,   8457    ,   461 ,   55507   ,0 };
            const unsigned long dim3671Kuo2Init[]       =   {   1   ,   1   ,   1   ,   15  ,   15  ,   57  ,   127 ,   249 ,   409 ,   885 ,   587 ,   285 ,   7087    ,   11613   ,   725 ,   61955   ,0 };
            const unsigned long dim3672Kuo2Init[]       =   {   1   ,   1   ,   3   ,   9   ,   21  ,   5   ,   127 ,   179 ,   193 ,   683 ,   2045    ,   2555    ,   6707    ,   63  ,   815 ,   57953   ,0 };
            const unsigned long dim3673Kuo2Init[]       =   {   1   ,   3   ,   7   ,   7   ,   25  ,   45  ,   49  ,   37  ,   289 ,   431 ,   719 ,   3071    ,   7343    ,   1377    ,   18515   ,   33777   ,0 };
            const unsigned long dim3674Kuo2Init[]       =   {   1   ,   3   ,   3   ,   7   ,   19  ,   5   ,   127 ,   29  ,   385 ,   895 ,   1951    ,   3817    ,   3277    ,   7677    ,   19781   ,   31181   ,0 };
            const unsigned long dim3675Kuo2Init[]       =   {   1   ,   3   ,   3   ,   9   ,   19  ,   25  ,   105 ,   123 ,   445 ,   923 ,   801 ,   455 ,   955 ,   5529    ,   29133   ,   47531   ,0 };
            const unsigned long dim3676Kuo2Init[]       =   {   1   ,   1   ,   3   ,   11  ,   1   ,   15  ,   51  ,   5   ,   27  ,   335 ,   523 ,   2859    ,   5605    ,   15197   ,   14669   ,   29623   ,0 };
            const unsigned long dim3677Kuo2Init[]       =   {   1   ,   1   ,   3   ,   1   ,   11  ,   53  ,   97  ,   229 ,   477 ,   991 ,   749 ,   607 ,   6833    ,   6817    ,   2397    ,   31543   ,0 };
            const unsigned long dim3678Kuo2Init[]       =   {   1   ,   3   ,   3   ,   9   ,   9   ,   45  ,   123 ,   201 ,   129 ,   587 ,   1373    ,   3733    ,   2331    ,   9741    ,   29441   ,   24711   ,0 };
            const unsigned long dim3679Kuo2Init[]       =   {   1   ,   1   ,   7   ,   9   ,   1   ,   53  ,   29  ,   99  ,   475 ,   227 ,   1955    ,   351 ,   2583    ,   2335    ,   1821    ,   21105   ,0 };
            const unsigned long dim3680Kuo2Init[]       =   {   1   ,   1   ,   7   ,   7   ,   13  ,   7   ,   121 ,   121 ,   163 ,   325 ,   13  ,   779 ,   4173    ,   7285    ,   26205   ,   19693   ,0 };
            const unsigned long dim3681Kuo2Init[]       =   {   1   ,   1   ,   3   ,   13  ,   17  ,   23  ,   23  ,   167 ,   131 ,   821 ,   933 ,   1919    ,   853 ,   7525    ,   8195    ,   7357    ,0 };
            const unsigned long dim3682Kuo2Init[]       =   {   1   ,   1   ,   5   ,   9   ,   21  ,   29  ,   97  ,   105 ,   85  ,   865 ,   459 ,   1405    ,   5767    ,   8379    ,   23397   ,   35269   ,0 };
            const unsigned long dim3683Kuo2Init[]       =   {   1   ,   3   ,   3   ,   15  ,   7   ,   41  ,   125 ,   107 ,   507 ,   323 ,   149 ,   3475    ,   1149    ,   8811    ,   8295    ,   20543   ,0 };
            const unsigned long dim3684Kuo2Init[]       =   {   1   ,   1   ,   5   ,   11  ,   13  ,   35  ,   41  ,   249 ,   195 ,   603 ,   873 ,   3825    ,   1679    ,   11727   ,   13839   ,   22241   ,0 };
            const unsigned long dim3685Kuo2Init[]       =   {   1   ,   3   ,   5   ,   3   ,   13  ,   15  ,   19  ,   113 ,   141 ,   7   ,   627 ,   1285    ,   6701    ,   15921   ,   4235    ,   30879   ,0 };
            const unsigned long dim3686Kuo2Init[]       =   {   1   ,   1   ,   7   ,   9   ,   15  ,   57  ,   65  ,   239 ,   49  ,   657 ,   673 ,   4039    ,   5493    ,   2487    ,   1807    ,   14341   ,0 };
            const unsigned long dim3687Kuo2Init[]       =   {   1   ,   1   ,   1   ,   13  ,   25  ,   47  ,   19  ,   191 ,   177 ,   163 ,   1721    ,   3473    ,   6667    ,   10017   ,   19863   ,   43803   ,0 };
            const unsigned long dim3688Kuo2Init[]       =   {   1   ,   3   ,   5   ,   11  ,   15  ,   39  ,   33  ,   95  ,   353 ,   121 ,   503 ,   3025    ,   4773    ,   3853    ,   14263   ,   61149   ,0 };
            const unsigned long dim3689Kuo2Init[]       =   {   1   ,   1   ,   5   ,   3   ,   1   ,   57  ,   7   ,   105 ,   363 ,   797 ,   635 ,   23  ,   1795    ,   13501   ,   45  ,   22081   ,0 };
            const unsigned long dim3690Kuo2Init[]       =   {   1   ,   3   ,   7   ,   15  ,   19  ,   21  ,   9   ,   237 ,   249 ,   641 ,   1599    ,   3623    ,   2941    ,   12917   ,   5511    ,   18017   ,0 };
            const unsigned long dim3691Kuo2Init[]       =   {   1   ,   1   ,   5   ,   9   ,   17  ,   57  ,   81  ,   165 ,   315 ,   985 ,   1965    ,   3571    ,   2693    ,   11591   ,   13675   ,   36739   ,0 };
            const unsigned long dim3692Kuo2Init[]       =   {   1   ,   1   ,   7   ,   15  ,   5   ,   3   ,   11  ,   247 ,   79  ,   707 ,   1601    ,   2301    ,   7281    ,   4871    ,   26735   ,   14277   ,0 };
            const unsigned long dim3693Kuo2Init[]       =   {   1   ,   1   ,   5   ,   7   ,   17  ,   27  ,   101 ,   211 ,   507 ,   95  ,   1333    ,   2919    ,   1251    ,   3905    ,   1897    ,   12791   ,0 };
            const unsigned long dim3694Kuo2Init[]       =   {   1   ,   3   ,   5   ,   15  ,   29  ,   15  ,   119 ,   3   ,   181 ,   621 ,   349 ,   2509    ,   295 ,   15719   ,   1593    ,   27333   ,0 };
            const unsigned long dim3695Kuo2Init[]       =   {   1   ,   1   ,   7   ,   9   ,   9   ,   29  ,   75  ,   245 ,   353 ,   147 ,   1197    ,   2571    ,   7033    ,   15903   ,   28319   ,   7733    ,0 };
            const unsigned long dim3696Kuo2Init[]       =   {   1   ,   1   ,   7   ,   11  ,   19  ,   23  ,   53  ,   41  ,   377 ,   673 ,   1843    ,   1283    ,   1343    ,   2089    ,   12295   ,   33189   ,0 };
            const unsigned long dim3697Kuo2Init[]       =   {   1   ,   1   ,   1   ,   9   ,   17  ,   21  ,   17  ,   251 ,   399 ,   531 ,   2007    ,   3449    ,   265 ,   289 ,   15937   ,   19571   ,0 };
            const unsigned long dim3698Kuo2Init[]       =   {   1   ,   1   ,   5   ,   15  ,   15  ,   53  ,   21  ,   141 ,   387 ,   885 ,   1409    ,   2883    ,   2101    ,   4741    ,   597 ,   23435   ,0 };
            const unsigned long dim3699Kuo2Init[]       =   {   1   ,   3   ,   1   ,   7   ,   27  ,   63  ,   33  ,   157 ,   331 ,   921 ,   1097    ,   3963    ,   3275    ,   14909   ,   2359    ,   12327   ,0 };
            const unsigned long dim3700Kuo2Init[]       =   {   1   ,   1   ,   3   ,   11  ,   15  ,   7   ,   121 ,   235 ,   219 ,   181 ,   1305    ,   3877    ,   2233    ,   4921    ,   31409   ,   28211   ,0 };
            const unsigned long dim3701Kuo2Init[]       =   {   1   ,   3   ,   3   ,   7   ,   23  ,   25  ,   67  ,   147 ,   211 ,   757 ,   1903    ,   3077    ,   3931    ,   11885   ,   3053    ,   45709   ,0 };
            const unsigned long dim3702Kuo2Init[]       =   {   1   ,   1   ,   1   ,   5   ,   29  ,   21  ,   121 ,   115 ,   231 ,   957 ,   203 ,   689 ,   5521    ,   2467    ,   21489   ,   10057   ,0 };
            const unsigned long dim3703Kuo2Init[]       =   {   1   ,   1   ,   5   ,   9   ,   15  ,   17  ,   11  ,   177 ,   319 ,   547 ,   801 ,   3899    ,   5919    ,   8785    ,   12019   ,   2535    ,0 };
            const unsigned long dim3704Kuo2Init[]       =   {   1   ,   3   ,   1   ,   3   ,   21  ,   51  ,   83  ,   121 ,   317 ,   715 ,   467 ,   3587    ,   383 ,   3519    ,   26731   ,   64151   ,0 };
            const unsigned long dim3705Kuo2Init[]       =   {   1   ,   3   ,   3   ,   13  ,   31  ,   47  ,   41  ,   247 ,   211 ,   333 ,   691 ,   2679    ,   4599    ,   2479    ,   27687   ,   45793   ,0 };
            const unsigned long dim3706Kuo2Init[]       =   {   1   ,   1   ,   1   ,   11  ,   31  ,   9   ,   41  ,   241 ,   439 ,   391 ,   949 ,   2957    ,   5105    ,   5073    ,   7607    ,   4809    ,0 };
            const unsigned long dim3707Kuo2Init[]       =   {   1   ,   1   ,   1   ,   9   ,   5   ,   57  ,   11  ,   167 ,   193 ,   1019    ,   533 ,   2559    ,   3899    ,   15199   ,   13723   ,   1909    ,0 };
            const unsigned long dim3708Kuo2Init[]       =   {   1   ,   3   ,   3   ,   3   ,   5   ,   17  ,   5   ,   21  ,   445 ,   857 ,   2017    ,   2759    ,   5011    ,   8415    ,   881 ,   15747   ,0 };
            const unsigned long dim3709Kuo2Init[]       =   {   1   ,   1   ,   1   ,   1   ,   1   ,   25  ,   31  ,   61  ,   407 ,   71  ,   263 ,   681 ,   5325    ,   13747   ,   10833   ,   12191   ,0 };
            const unsigned long dim3710Kuo2Init[]       =   {   1   ,   3   ,   3   ,   9   ,   27  ,   3   ,   73  ,   145 ,   5   ,   851 ,   1519    ,   2809    ,   5857    ,   2327    ,   32659   ,   11145   ,0 };
            const unsigned long dim3711Kuo2Init[]       =   {   1   ,   3   ,   3   ,   5   ,   21  ,   3   ,   47  ,   11  ,   17  ,   1023    ,   621 ,   3527    ,   613 ,   5607    ,   19377   ,   49671   ,0 };
            const unsigned long dim3712Kuo2Init[]       =   {   1   ,   3   ,   1   ,   11  ,   5   ,   5   ,   57  ,   39  ,   221 ,   847 ,   735 ,   3951    ,   6661    ,   13959   ,   16735   ,   26471   ,0 };
            const unsigned long dim3713Kuo2Init[]       =   {   1   ,   3   ,   1   ,   11  ,   3   ,   31  ,   15  ,   121 ,   31  ,   689 ,   151 ,   849 ,   2859    ,   14903   ,   6065    ,   9203    ,0 };
            const unsigned long dim3714Kuo2Init[]       =   {   1   ,   3   ,   3   ,   9   ,   5   ,   63  ,   27  ,   51  ,   379 ,   657 ,   1381    ,   67  ,   2733    ,   7673    ,   13217   ,   14265   ,0 };
            const unsigned long dim3715Kuo2Init[]       =   {   1   ,   1   ,   7   ,   15  ,   27  ,   63  ,   21  ,   25  ,   315 ,   101 ,   851 ,   2585    ,   4423    ,   7857    ,   30561   ,   60301   ,0 };
            const unsigned long dim3716Kuo2Init[]       =   {   1   ,   3   ,   3   ,   11  ,   21  ,   51  ,   9   ,   39  ,   123 ,   763 ,   1777    ,   3427    ,   5071    ,   7905    ,   1017    ,   15373   ,0 };
            const unsigned long dim3717Kuo2Init[]       =   {   1   ,   3   ,   1   ,   7   ,   25  ,   59  ,   3   ,   223 ,   71  ,   33  ,   731 ,   2303    ,   7153    ,   12847   ,   9707    ,   50871   ,0 };
            const unsigned long dim3718Kuo2Init[]       =   {   1   ,   1   ,   5   ,   11  ,   31  ,   61  ,   13  ,   155 ,   467 ,   501 ,   371 ,   1447    ,   7925    ,   13135   ,   27629   ,   45097   ,0 };
            const unsigned long dim3719Kuo2Init[]       =   {   1   ,   1   ,   1   ,   1   ,   13  ,   25  ,   35  ,   153 ,   333 ,   9   ,   1609    ,   539 ,   5029    ,   2059    ,   3547    ,   55407   ,0 };
            const unsigned long dim3720Kuo2Init[]       =   {   1   ,   1   ,   7   ,   9   ,   21  ,   55  ,   101 ,   199 ,   279 ,   77  ,   1337    ,   3503    ,   2365    ,   4969    ,   2411    ,   16853   ,0 };
            const unsigned long dim3721Kuo2Init[]       =   {   1   ,   1   ,   7   ,   11  ,   29  ,   9   ,   111 ,   63  ,   435 ,   445 ,   1711    ,   3981    ,   6517    ,   14593   ,   29433   ,   31027   ,0 };
            const unsigned long dim3722Kuo2Init[]       =   {   1   ,   1   ,   7   ,   1   ,   15  ,   57  ,   43  ,   137 ,   189 ,   83  ,   187 ,   417 ,   3425    ,   14653   ,   17805   ,   57685   ,0 };
            const unsigned long dim3723Kuo2Init[]       =   {   1   ,   1   ,   5   ,   3   ,   27  ,   23  ,   71  ,   65  ,   399 ,   605 ,   421 ,   463 ,   2019    ,   10707   ,   26201   ,   23573   ,0 };
            const unsigned long dim3724Kuo2Init[]       =   {   1   ,   1   ,   3   ,   15  ,   17  ,   57  ,   119 ,   125 ,   365 ,   681 ,   1967    ,   241 ,   7647    ,   13317   ,   23661   ,   55683   ,0 };
            const unsigned long dim3725Kuo2Init[]       =   {   1   ,   1   ,   7   ,   1   ,   7   ,   63  ,   91  ,   161 ,   45  ,   533 ,   1833    ,   2465    ,   5081    ,   6341    ,   16713   ,   34685   ,0 };
            const unsigned long dim3726Kuo2Init[]       =   {   1   ,   3   ,   1   ,   9   ,   23  ,   9   ,   43  ,   157 ,   307 ,   521 ,   1611    ,   3851    ,   6943    ,   14939   ,   13833   ,   20099   ,0 };
            const unsigned long dim3727Kuo2Init[]       =   {   1   ,   1   ,   5   ,   3   ,   27  ,   13  ,   35  ,   131 ,   129 ,   257 ,   1015    ,   1849    ,   6645    ,   11567   ,   13385   ,   30943   ,0 };
            const unsigned long dim3728Kuo2Init[]       =   {   1   ,   1   ,   5   ,   1   ,   21  ,   49  ,   101 ,   77  ,   301 ,   541 ,   481 ,   2903    ,   8121    ,   2689    ,   21541   ,   25337   ,0 };
            const unsigned long dim3729Kuo2Init[]       =   {   1   ,   1   ,   5   ,   11  ,   1   ,   13  ,   125 ,   209 ,   327 ,   245 ,   1055    ,   1097    ,   4931    ,   15721   ,   8193    ,   19923   ,0 };
            const unsigned long dim3730Kuo2Init[]       =   {   1   ,   1   ,   1   ,   11  ,   17  ,   31  ,   85  ,   159 ,   405 ,   287 ,   1083    ,   1679    ,   6227    ,   305 ,   9549    ,   27303   ,0 };
            const unsigned long dim3731Kuo2Init[]       =   {   1   ,   1   ,   1   ,   13  ,   7   ,   45  ,   125 ,   181 ,   219 ,   811 ,   1471    ,   999 ,   93  ,   12845   ,   4597    ,   60641   ,0 };
            const unsigned long dim3732Kuo2Init[]       =   {   1   ,   3   ,   5   ,   7   ,   29  ,   23  ,   109 ,   37  ,   367 ,   185 ,   1467    ,   1871    ,   6349    ,   8563    ,   8743    ,   30277   ,0 };
            const unsigned long dim3733Kuo2Init[]       =   {   1   ,   1   ,   1   ,   15  ,   19  ,   23  ,   75  ,   93  ,   191 ,   993 ,   1111    ,   1531    ,   2105    ,   8869    ,   27217   ,   16609   ,0 };
            const unsigned long dim3734Kuo2Init[]       =   {   1   ,   3   ,   1   ,   15  ,   31  ,   29  ,   23  ,   145 ,   467 ,   465 ,   637 ,   3169    ,   889 ,   9225    ,   17917   ,   8379    ,0 };
            const unsigned long dim3735Kuo2Init[]       =   {   1   ,   3   ,   5   ,   1   ,   1   ,   33  ,   7   ,   209 ,   99  ,   753 ,   113 ,   2945    ,   345 ,   14091   ,   13169   ,   4799    ,0 };
            const unsigned long dim3736Kuo2Init[]       =   {   1   ,   3   ,   3   ,   11  ,   21  ,   5   ,   9   ,   97  ,   347 ,   537 ,   1287    ,   1401    ,   815 ,   9289    ,   17259   ,   11755   ,0 };
            const unsigned long dim3737Kuo2Init[]       =   {   1   ,   1   ,   5   ,   1   ,   7   ,   9   ,   21  ,   87  ,   495 ,   685 ,   1675    ,   1909    ,   4215    ,   1589    ,   18229   ,   2253    ,0 };
            const unsigned long dim3738Kuo2Init[]       =   {   1   ,   1   ,   1   ,   3   ,   7   ,   29  ,   121 ,   67  ,   383 ,   351 ,   1555    ,   3091    ,   3677    ,   3489    ,   30241   ,   55309   ,0 };
            const unsigned long dim3739Kuo2Init[]       =   {   1   ,   1   ,   7   ,   11  ,   21  ,   39  ,   33  ,   201 ,   195 ,   601 ,   1235    ,   3499    ,   1657    ,   4495    ,   26501   ,   49783   ,0 };
            const unsigned long dim3740Kuo2Init[]       =   {   1   ,   1   ,   5   ,   1   ,   15  ,   37  ,   91  ,   169 ,   65  ,   471 ,   1409    ,   3017    ,   1071    ,   15331   ,   27085   ,   22837   ,0 };
            const unsigned long dim3741Kuo2Init[]       =   {   1   ,   1   ,   5   ,   13  ,   23  ,   45  ,   115 ,   153 ,   27  ,   773 ,   637 ,   2673    ,   79  ,   4275    ,   21133   ,   38509   ,0 };
            const unsigned long dim3742Kuo2Init[]       =   {   1   ,   3   ,   1   ,   13  ,   17  ,   53  ,   49  ,   55  ,   439 ,   255 ,   383 ,   1057    ,   4995    ,   3853    ,   29839   ,   2529    ,0 };
            const unsigned long dim3743Kuo2Init[]       =   {   1   ,   3   ,   5   ,   9   ,   17  ,   29  ,   73  ,   205 ,   1   ,   685 ,   991 ,   3563    ,   4281    ,   2341    ,   27111   ,   36609   ,0 };
            const unsigned long dim3744Kuo2Init[]       =   {   1   ,   3   ,   5   ,   9   ,   19  ,   25  ,   65  ,   167 ,   425 ,   289 ,   963 ,   1653    ,   131 ,   4599    ,   4615    ,   57653   ,0 };
            const unsigned long dim3745Kuo2Init[]       =   {   1   ,   1   ,   1   ,   11  ,   19  ,   57  ,   93  ,   239 ,   375 ,   739 ,   1101    ,   881 ,   7391    ,   921 ,   11493   ,   50417   ,0 };
            const unsigned long dim3746Kuo2Init[]       =   {   1   ,   3   ,   5   ,   15  ,   7   ,   33  ,   5   ,   59  ,   381 ,   123 ,   33  ,   1903    ,   2639    ,   12073   ,   511 ,   659 ,0 };
            const unsigned long dim3747Kuo2Init[]       =   {   1   ,   1   ,   3   ,   1   ,   23  ,   1   ,   77  ,   55  ,   455 ,   687 ,   1825    ,   1355    ,   2997    ,   615 ,   123 ,   17611   ,0 };
            const unsigned long dim3748Kuo2Init[]       =   {   1   ,   1   ,   7   ,   13  ,   25  ,   61  ,   119 ,   101 ,   397 ,   645 ,   485 ,   3361    ,   899 ,   5103    ,   3307    ,   9765    ,0 };
            const unsigned long dim3749Kuo2Init[]       =   {   1   ,   3   ,   1   ,   3   ,   25  ,   59  ,   27  ,   245 ,   3   ,   915 ,   1001    ,   977 ,   7913    ,   4667    ,   24845   ,   37183   ,0 };
            const unsigned long dim3750Kuo2Init[]       =   {   1   ,   1   ,   7   ,   15  ,   19  ,   21  ,   1   ,   233 ,   175 ,   347 ,   1335    ,   2757    ,   1615    ,   8381    ,   10611   ,   32409   ,0 };
            const unsigned long dim3751Kuo2Init[]       =   {   1   ,   1   ,   7   ,   15  ,   23  ,   33  ,   7   ,   51  ,   235 ,   735 ,   615 ,   3305    ,   4555    ,   15079   ,   21309   ,   55397   ,0 };
            const unsigned long dim3752Kuo2Init[]       =   {   1   ,   3   ,   1   ,   3   ,   19  ,   49  ,   59  ,   213 ,   155 ,   23  ,   239 ,   753 ,   1473    ,   6869    ,   19993   ,   14383   ,0 };
            const unsigned long dim3753Kuo2Init[]       =   {   1   ,   1   ,   5   ,   9   ,   17  ,   37  ,   23  ,   163 ,   45  ,   409 ,   1103    ,   289 ,   4707    ,   12657   ,   26033   ,   27547   ,0 };
            const unsigned long dim3754Kuo2Init[]       =   {   1   ,   3   ,   7   ,   9   ,   23  ,   49  ,   83  ,   63  ,   79  ,   643 ,   107 ,   3041    ,   1921    ,   10343   ,   19179   ,   4961    ,0 };
            const unsigned long dim3755Kuo2Init[]       =   {   1   ,   1   ,   7   ,   7   ,   31  ,   7   ,   81  ,   145 ,   281 ,   751 ,   867 ,   535 ,   6593    ,   7255    ,   13785   ,   62137   ,0 };
            const unsigned long dim3756Kuo2Init[]       =   {   1   ,   1   ,   7   ,   1   ,   13  ,   55  ,   59  ,   203 ,   329 ,   421 ,   483 ,   2931    ,   521 ,   3013    ,   1077    ,   3405    ,0 };
            const unsigned long dim3757Kuo2Init[]       =   {   1   ,   3   ,   5   ,   1   ,   15  ,   13  ,   21  ,   205 ,   187 ,   491 ,   887 ,   893 ,   3917    ,   6989    ,   23569   ,   48729   ,0 };
            const unsigned long dim3758Kuo2Init[]       =   {   1   ,   1   ,   3   ,   11  ,   15  ,   35  ,   63  ,   109 ,   157 ,   643 ,   961 ,   3713    ,   1459    ,   8585    ,   6113    ,   7339    ,0 };
            const unsigned long dim3759Kuo2Init[]       =   {   1   ,   3   ,   7   ,   9   ,   7   ,   53  ,   57  ,   251 ,   203 ,   701 ,   2041    ,   73  ,   7637    ,   12745   ,   8003    ,   45849   ,0 };
            const unsigned long dim3760Kuo2Init[]       =   {   1   ,   3   ,   5   ,   1   ,   23  ,   27  ,   109 ,   211 ,   273 ,   1013    ,   19  ,   2499    ,   377 ,   5185    ,   25261   ,   8997    ,0 };
            const unsigned long dim3761Kuo2Init[]       =   {   1   ,   3   ,   7   ,   15  ,   3   ,   35  ,   41  ,   85  ,   391 ,   409 ,   1945    ,   25  ,   6069    ,   3499    ,   10879   ,   38239   ,0 };
            const unsigned long dim3762Kuo2Init[]       =   {   1   ,   3   ,   1   ,   3   ,   3   ,   37  ,   65  ,   229 ,   441 ,   675 ,   909 ,   3887    ,   7347    ,   3221    ,   1935    ,   65147   ,0 };
            const unsigned long dim3763Kuo2Init[]       =   {   1   ,   3   ,   5   ,   13  ,   13  ,   29  ,   119 ,   29  ,   277 ,   303 ,   685 ,   1921    ,   4863    ,   12483   ,   9303    ,   59355   ,0 };
            const unsigned long dim3764Kuo2Init[]       =   {   1   ,   1   ,   1   ,   9   ,   21  ,   43  ,   115 ,   253 ,   349 ,   655 ,   779 ,   1965    ,   4831    ,   6491    ,   16051   ,   6975    ,0 };
            const unsigned long dim3765Kuo2Init[]       =   {   1   ,   3   ,   3   ,   7   ,   1   ,   9   ,   55  ,   253 ,   21  ,   655 ,   133 ,   3777    ,   4239    ,   14447   ,   7221    ,   59819   ,0 };
            const unsigned long dim3766Kuo2Init[]       =   {   1   ,   1   ,   7   ,   3   ,   19  ,   49  ,   11  ,   93  ,   39  ,   149 ,   683 ,   3295    ,   6519    ,   12187   ,   2231    ,   39829   ,0 };
            const unsigned long dim3767Kuo2Init[]       =   {   1   ,   3   ,   7   ,   5   ,   23  ,   9   ,   101 ,   197 ,   3   ,   873 ,   1459    ,   747 ,   6939    ,   5629    ,   27305   ,   32235   ,0 };
            const unsigned long dim3768Kuo2Init[]       =   {   1   ,   1   ,   5   ,   7   ,   13  ,   9   ,   31  ,   251 ,   481 ,   877 ,   773 ,   993 ,   4431    ,   12081   ,   15987   ,   27081   ,0 };
            const unsigned long dim3769Kuo2Init[]       =   {   1   ,   1   ,   1   ,   1   ,   9   ,   49  ,   13  ,   89  ,   39  ,   797 ,   815 ,   319 ,   5917    ,   621 ,   25867   ,   46433   ,0 };
            const unsigned long dim3770Kuo2Init[]       =   {   1   ,   3   ,   1   ,   7   ,   23  ,   27  ,   83  ,   123 ,   83  ,   177 ,   1001    ,   995 ,   2765    ,   6401    ,   15593   ,   20205   ,0 };
            const unsigned long dim3771Kuo2Init[]       =   {   1   ,   1   ,   5   ,   13  ,   25  ,   9   ,   59  ,   123 ,   141 ,   737 ,   1111    ,   1883    ,   2647    ,   733 ,   3551    ,   44453   ,0 };
            const unsigned long dim3772Kuo2Init[]       =   {   1   ,   3   ,   1   ,   5   ,   7   ,   49  ,   25  ,   177 ,   297 ,   87  ,   1039    ,   1917    ,   3167    ,   11025   ,   22959   ,   63929   ,0 };
            const unsigned long dim3773Kuo2Init[]       =   {   1   ,   1   ,   1   ,   3   ,   17  ,   57  ,   103 ,   99  ,   127 ,   579 ,   903 ,   2835    ,   607 ,   9723    ,   22397   ,   36621   ,0 };
            const unsigned long dim3774Kuo2Init[]       =   {   1   ,   1   ,   3   ,   15  ,   13  ,   47  ,   77  ,   213 ,   383 ,   371 ,   887 ,   281 ,   6171    ,   10383   ,   12157   ,   57899   ,0 };
            const unsigned long dim3775Kuo2Init[]       =   {   1   ,   3   ,   1   ,   5   ,   5   ,   15  ,   65  ,   119 ,   303 ,   265 ,   783 ,   3075    ,   7893    ,   441 ,   22065   ,   52157   ,0 };
            const unsigned long dim3776Kuo2Init[]       =   {   1   ,   3   ,   5   ,   15  ,   7   ,   49  ,   105 ,   145 ,   301 ,   57  ,   1949    ,   3073    ,   435 ,   1171    ,   27691   ,   34821   ,0 };
            const unsigned long dim3777Kuo2Init[]       =   {   1   ,   3   ,   1   ,   11  ,   5   ,   31  ,   33  ,   159 ,   319 ,   437 ,   453 ,   1005    ,   3335    ,   5895    ,   15407   ,   159 ,0 };
            const unsigned long dim3778Kuo2Init[]       =   {   1   ,   3   ,   3   ,   9   ,   31  ,   15  ,   69  ,   129 ,   221 ,   991 ,   1879    ,   1251    ,   5303    ,   6677    ,   15823   ,   46319   ,0 };
            const unsigned long dim3779Kuo2Init[]       =   {   1   ,   3   ,   7   ,   11  ,   1   ,   47  ,   49  ,   13  ,   401 ,   755 ,   1703    ,   2021    ,   5195    ,   9489    ,   9191    ,   30427   ,0 };
            const unsigned long dim3780Kuo2Init[]       =   {   1   ,   1   ,   7   ,   9   ,   7   ,   11  ,   25  ,   147 ,   471 ,   31  ,   489 ,   3477    ,   3025    ,   4011    ,   7733    ,   22091   ,0 };
            const unsigned long dim3781Kuo2Init[]       =   {   1   ,   1   ,   5   ,   9   ,   7   ,   7   ,   99  ,   23  ,   397 ,   67  ,   2027    ,   3885    ,   3269    ,   15385   ,   24367   ,   1715    ,0 };
            const unsigned long dim3782Kuo2Init[]       =   {   1   ,   1   ,   5   ,   1   ,   7   ,   45  ,   83  ,   159 ,   501 ,   561 ,   993 ,   1495    ,   4415    ,   2559    ,   27  ,   60825   ,0 };
            const unsigned long dim3783Kuo2Init[]       =   {   1   ,   3   ,   1   ,   7   ,   7   ,   43  ,   99  ,   159 ,   117 ,   453 ,   1841    ,   1807    ,   1881    ,   12513   ,   16413   ,   37993   ,0 };
            const unsigned long dim3784Kuo2Init[]       =   {   1   ,   3   ,   1   ,   13  ,   15  ,   37  ,   53  ,   173 ,   279 ,   415 ,   535 ,   2553    ,   3995    ,   8245    ,   4507    ,   19575   ,0 };
            const unsigned long dim3785Kuo2Init[]       =   {   1   ,   1   ,   7   ,   9   ,   25  ,   7   ,   79  ,   25  ,   105 ,   525 ,   1179    ,   1793    ,   4899    ,   10253   ,   22655   ,   1419    ,0 };
            const unsigned long dim3786Kuo2Init[]       =   {   1   ,   1   ,   5   ,   11  ,   11  ,   35  ,   77  ,   139 ,   449 ,   879 ,   619 ,   1645    ,   3309    ,   37  ,   10607   ,   43757   ,0 };
            const unsigned long dim3787Kuo2Init[]       =   {   1   ,   1   ,   1   ,   13  ,   17  ,   17  ,   7   ,   69  ,   389 ,   733 ,   775 ,   791 ,   5451    ,   15737   ,   9043    ,   16573   ,0 };
            const unsigned long dim3788Kuo2Init[]       =   {   1   ,   1   ,   7   ,   5   ,   15  ,   57  ,   81  ,   19  ,   307 ,   207 ,   241 ,   449 ,   7745    ,   521 ,   16975   ,   11907   ,0 };
            const unsigned long dim3789Kuo2Init[]       =   {   1   ,   1   ,   1   ,   5   ,   25  ,   1   ,   55  ,   229 ,   123 ,   809 ,   1673    ,   3461    ,   7729    ,   8615    ,   8755    ,   35929   ,0 };
            const unsigned long dim3790Kuo2Init[]       =   {   1   ,   3   ,   3   ,   1   ,   23  ,   7   ,   19  ,   235 ,   375 ,   79  ,   899 ,   21  ,   7681    ,   7169    ,   13057   ,   1379    ,0 };
            const unsigned long dim3791Kuo2Init[]       =   {   1   ,   3   ,   3   ,   7   ,   29  ,   49  ,   93  ,   217 ,   409 ,   263 ,   225 ,   2333    ,   1621    ,   14741   ,   4933    ,   58263   ,0 };
            const unsigned long dim3792Kuo2Init[]       =   {   1   ,   3   ,   1   ,   13  ,   11  ,   35  ,   19  ,   55  ,   71  ,   937 ,   551 ,   1693    ,   3843    ,   7411    ,   29501   ,   31525   ,0 };
            const unsigned long dim3793Kuo2Init[]       =   {   1   ,   1   ,   5   ,   13  ,   17  ,   17  ,   105 ,   31  ,   181 ,   915 ,   321 ,   3161    ,   5247    ,   13317   ,   23595   ,   21783   ,0 };
            const unsigned long dim3794Kuo2Init[]       =   {   1   ,   3   ,   5   ,   15  ,   31  ,   19  ,   85  ,   71  ,   55  ,   933 ,   1425    ,   2419    ,   2891    ,   15813   ,   18209   ,   30347   ,0 };
            const unsigned long dim3795Kuo2Init[]       =   {   1   ,   1   ,   7   ,   9   ,   1   ,   45  ,   31  ,   1   ,   193 ,   445 ,   1293    ,   2563    ,   5837    ,   6113    ,   5741    ,   12245   ,0 };
            const unsigned long dim3796Kuo2Init[]       =   {   1   ,   3   ,   5   ,   9   ,   25  ,   25  ,   7   ,   155 ,   267 ,   65  ,   2015    ,   2021    ,   6331    ,   6973    ,   1303    ,   14833   ,0 };
            const unsigned long dim3797Kuo2Init[]       =   {   1   ,   1   ,   5   ,   11  ,   21  ,   31  ,   87  ,   21  ,   509 ,   555 ,   429 ,   1405    ,   4417    ,   1891    ,   15679   ,   3819    ,0 };
            const unsigned long dim3798Kuo2Init[]       =   {   1   ,   3   ,   7   ,   3   ,   19  ,   49  ,   53  ,   251 ,   423 ,   949 ,   971 ,   425 ,   1589    ,   5257    ,   19021   ,   16817   ,0 };
            const unsigned long dim3799Kuo2Init[]       =   {   1   ,   1   ,   5   ,   7   ,   1   ,   27  ,   71  ,   185 ,   357 ,   1009    ,   897 ,   491 ,   6433    ,   6717    ,   23445   ,   29097   ,0 };
            const unsigned long dim3800Kuo2Init[]       =   {   1   ,   3   ,   7   ,   15  ,   9   ,   39  ,   87  ,   25  ,   149 ,   291 ,   1249    ,   3987    ,   7441    ,   11225   ,   18051   ,   17747   ,0 };
            const unsigned long dim3801Kuo2Init[]       =   {   1   ,   1   ,   7   ,   1   ,   13  ,   61  ,   13  ,   141 ,   39  ,   317 ,   579 ,   2181    ,   6735    ,   14749   ,   25997   ,   16025   ,0 };
            const unsigned long dim3802Kuo2Init[]       =   {   1   ,   1   ,   7   ,   7   ,   27  ,   19  ,   127 ,   231 ,   13  ,   7   ,   1707    ,   87  ,   7379    ,   8171    ,   27435   ,   47911   ,0 };
            const unsigned long dim3803Kuo2Init[]       =   {   1   ,   1   ,   1   ,   15  ,   21  ,   33  ,   33  ,   85  ,   205 ,   321 ,   785 ,   4051    ,   2503    ,   15589   ,   10881   ,   63103   ,0 };
            const unsigned long dim3804Kuo2Init[]       =   {   1   ,   1   ,   1   ,   3   ,   15  ,   11  ,   31  ,   105 ,   185 ,   113 ,   293 ,   3279    ,   4867    ,   9249    ,   14403   ,   24951   ,0 };
            const unsigned long dim3805Kuo2Init[]       =   {   1   ,   3   ,   5   ,   13  ,   7   ,   47  ,   105 ,   253 ,   117 ,   577 ,   2019    ,   2089    ,   1357    ,   5299    ,   27961   ,   36675   ,0 };
            const unsigned long dim3806Kuo2Init[]       =   {   1   ,   1   ,   1   ,   7   ,   31  ,   33  ,   117 ,   51  ,   283 ,   453 ,   1657    ,   1343    ,   6491    ,   649 ,   1029    ,   44809   ,0 };
            const unsigned long dim3807Kuo2Init[]       =   {   1   ,   1   ,   1   ,   3   ,   3   ,   45  ,   121 ,   155 ,   191 ,   607 ,   1087    ,   3291    ,   8171    ,   3757    ,   14247   ,   49647   ,0 };
            const unsigned long dim3808Kuo2Init[]       =   {   1   ,   3   ,   1   ,   3   ,   7   ,   17  ,   109 ,   163 ,   87  ,   63  ,   573 ,   3127    ,   3361    ,   6733    ,   26729   ,   44427   ,0 };
            const unsigned long dim3809Kuo2Init[]       =   {   1   ,   1   ,   1   ,   5   ,   13  ,   23  ,   29  ,   173 ,   487 ,   345 ,   351 ,   3401    ,   4611    ,   5591    ,   6209    ,   58537   ,0 };
            const unsigned long dim3810Kuo2Init[]       =   {   1   ,   3   ,   1   ,   5   ,   23  ,   49  ,   31  ,   51  ,   157 ,   623 ,   1103    ,   681 ,   6645    ,   2147    ,   1715    ,   42253   ,0 };
            const unsigned long dim3811Kuo2Init[]       =   {   1   ,   1   ,   7   ,   9   ,   7   ,   61  ,   85  ,   129 ,   309 ,   701 ,   1139    ,   3211    ,   4943    ,   8365    ,   20727   ,   51025   ,0 };
            const unsigned long dim3812Kuo2Init[]       =   {   1   ,   1   ,   3   ,   11  ,   29  ,   47  ,   119 ,   181 ,   109 ,   403 ,   1459    ,   39  ,   609 ,   7405    ,   25651   ,   61395   ,0 };
            const unsigned long dim3813Kuo2Init[]       =   {   1   ,   3   ,   3   ,   7   ,   5   ,   55  ,   65  ,   155 ,   273 ,   177 ,   1879    ,   3901    ,   3219    ,   8699    ,   10445   ,   40141   ,0 };
            const unsigned long dim3814Kuo2Init[]       =   {   1   ,   3   ,   7   ,   15  ,   3   ,   11  ,   45  ,   85  ,   473 ,   1013    ,   1725    ,   155 ,   3369    ,   7345    ,   1475    ,   28781   ,0 };
            const unsigned long dim3815Kuo2Init[]       =   {   1   ,   1   ,   1   ,   11  ,   13  ,   61  ,   31  ,   217 ,   111 ,   719 ,   161 ,   1053    ,   1451    ,   10437   ,   28909   ,   1173    ,0 };
            const unsigned long dim3816Kuo2Init[]       =   {   1   ,   1   ,   3   ,   11  ,   9   ,   21  ,   119 ,   119 ,   355 ,   563 ,   855 ,   853 ,   4243    ,   5201    ,   2521    ,   41389   ,0 };
            const unsigned long dim3817Kuo2Init[]       =   {   1   ,   1   ,   5   ,   15  ,   9   ,   5   ,   15  ,   79  ,   489 ,   499 ,   1065    ,   2073    ,   7567    ,   5385    ,   9329    ,   17643   ,0 };
            const unsigned long dim3818Kuo2Init[]       =   {   1   ,   3   ,   3   ,   7   ,   15  ,   53  ,   61  ,   225 ,   381 ,   217 ,   163 ,   4005    ,   7269    ,   4481    ,   23697   ,   46055   ,0 };
            const unsigned long dim3819Kuo2Init[]       =   {   1   ,   3   ,   3   ,   13  ,   1   ,   55  ,   77  ,   95  ,   345 ,   113 ,   1359    ,   539 ,   6979    ,   5459    ,   9551    ,   45035   ,0 };
            const unsigned long dim3820Kuo2Init[]       =   {   1   ,   1   ,   5   ,   3   ,   5   ,   45  ,   93  ,   147 ,   387 ,   165 ,   1529    ,   1243    ,   7639    ,   5531    ,   22381   ,   63517   ,0 };
            const unsigned long dim3821Kuo2Init[]       =   {   1   ,   1   ,   3   ,   1   ,   9   ,   27  ,   109 ,   161 ,   71  ,   547 ,   259 ,   1399    ,   5815    ,   1569    ,   26627   ,   63133   ,0 };
            const unsigned long dim3822Kuo2Init[]       =   {   1   ,   1   ,   3   ,   3   ,   29  ,   37  ,   21  ,   205 ,   49  ,   863 ,   707 ,   307 ,   573 ,   3167    ,   17595   ,   63485   ,0 };
            const unsigned long dim3823Kuo2Init[]       =   {   1   ,   1   ,   5   ,   1   ,   17  ,   15  ,   95  ,   21  ,   111 ,   991 ,   657 ,   3865    ,   8091    ,   7589    ,   30487   ,   63229   ,0 };
            const unsigned long dim3824Kuo2Init[]       =   {   1   ,   1   ,   1   ,   7   ,   29  ,   11  ,   65  ,   201 ,   357 ,   269 ,   281 ,   1329    ,   6359    ,   7875    ,   23561   ,   44489   ,0 };
            const unsigned long dim3825Kuo2Init[]       =   {   1   ,   1   ,   7   ,   7   ,   27  ,   29  ,   1   ,   15  ,   475 ,   393 ,   1917    ,   3605    ,   4463    ,   7809    ,   16529   ,   64599   ,0 };
            const unsigned long dim3826Kuo2Init[]       =   {   1   ,   3   ,   5   ,   13  ,   29  ,   15  ,   11  ,   255 ,   315 ,   209 ,   895 ,   2813    ,   2263    ,   7095    ,   1447    ,   55307   ,0 };
            const unsigned long dim3827Kuo2Init[]       =   {   1   ,   3   ,   1   ,   15  ,   17  ,   11  ,   73  ,   57  ,   13  ,   529 ,   35  ,   1065    ,   81  ,   13801   ,   19531   ,   27249   ,0 };
            const unsigned long dim3828Kuo2Init[]       =   {   1   ,   3   ,   1   ,   11  ,   19  ,   15  ,   117 ,   235 ,   219 ,   109 ,   1885    ,   4005    ,   3565    ,   2591    ,   29233   ,   47111   ,0 };
            const unsigned long dim3829Kuo2Init[]       =   {   1   ,   3   ,   3   ,   11  ,   21  ,   35  ,   27  ,   227 ,   393 ,   57  ,   1637    ,   537 ,   2463    ,   4483    ,   32153   ,   22035   ,0 };
            const unsigned long dim3830Kuo2Init[]       =   {   1   ,   1   ,   7   ,   5   ,   13  ,   9   ,   79  ,   147 ,   341 ,   1007    ,   1093    ,   3837    ,   5581    ,   3363    ,   10637   ,   60109   ,0 };
            const unsigned long dim3831Kuo2Init[]       =   {   1   ,   3   ,   5   ,   13  ,   31  ,   49  ,   121 ,   231 ,   485 ,   947 ,   127 ,   1817    ,   3005    ,   11085   ,   14157   ,   21145   ,0 };
            const unsigned long dim3832Kuo2Init[]       =   {   1   ,   1   ,   3   ,   1   ,   1   ,   23  ,   17  ,   195 ,   351 ,   527 ,   1787    ,   3885    ,   2281    ,   729 ,   25083   ,   63523   ,0 };
            const unsigned long dim3833Kuo2Init[]       =   {   1   ,   3   ,   5   ,   15  ,   19  ,   31  ,   51  ,   5   ,   55  ,   551 ,   33  ,   1735    ,   427 ,   2859    ,   3941    ,   48639   ,0 };
            const unsigned long dim3834Kuo2Init[]       =   {   1   ,   3   ,   7   ,   7   ,   11  ,   59  ,   101 ,   125 ,   167 ,   311 ,   1487    ,   969 ,   5023    ,   3893    ,   22065   ,   48803   ,0 };
            const unsigned long dim3835Kuo2Init[]       =   {   1   ,   3   ,   3   ,   11  ,   27  ,   11  ,   77  ,   43  ,   295 ,   31  ,   1061    ,   2999    ,   4451    ,   10339   ,   9213    ,   63705   ,0 };
            const unsigned long dim3836Kuo2Init[]       =   {   1   ,   3   ,   1   ,   11  ,   1   ,   63  ,   83  ,   179 ,   413 ,   317 ,   1921    ,   2917    ,   6697    ,   6637    ,   5291    ,   48791   ,0 };
            const unsigned long dim3837Kuo2Init[]       =   {   1   ,   1   ,   3   ,   3   ,   19  ,   57  ,   55  ,   159 ,   137 ,   291 ,   1353    ,   81  ,   2495    ,   12315   ,   31031   ,   41081   ,0 };
            const unsigned long dim3838Kuo2Init[]       =   {   1   ,   1   ,   3   ,   3   ,   25  ,   49  ,   49  ,   239 ,   53  ,   999 ,   303 ,   3083    ,   2455    ,   3977    ,   13945   ,   22667   ,0 };
            const unsigned long dim3839Kuo2Init[]       =   {   1   ,   3   ,   5   ,   15  ,   7   ,   15  ,   85  ,   215 ,   463 ,   301 ,   1061    ,   3993    ,   5927    ,   703 ,   23083   ,   44633   ,0 };
            const unsigned long dim3840Kuo2Init[]       =   {   1   ,   3   ,   3   ,   1   ,   9   ,   53  ,   21  ,   165 ,   129 ,   697 ,   1867    ,   2427    ,   5383    ,   13223   ,   12127   ,   60477   ,0 };
            const unsigned long dim3841Kuo2Init[]       =   {   1   ,   3   ,   5   ,   5   ,   17  ,   13  ,   37  ,   151 ,   337 ,   425 ,   829 ,   2415    ,   1705    ,   10389   ,   26299   ,   45655   ,0 };
            const unsigned long dim3842Kuo2Init[]       =   {   1   ,   3   ,   1   ,   9   ,   13  ,   3   ,   99  ,   111 ,   371 ,   589 ,   605 ,   3137    ,   4515    ,   5303    ,   6151    ,   14225   ,0 };
            const unsigned long dim3843Kuo2Init[]       =   {   1   ,   1   ,   3   ,   11  ,   15  ,   59  ,   31  ,   95  ,   305 ,   133 ,   751 ,   13  ,   7691    ,   2603    ,   21761   ,   18577   ,0 };
            const unsigned long dim3844Kuo2Init[]       =   {   1   ,   1   ,   1   ,   13  ,   25  ,   47  ,   93  ,   251 ,   341 ,   465 ,   1043    ,   3391    ,   271 ,   3431    ,   16465   ,   53283   ,0 };
            const unsigned long dim3845Kuo2Init[]       =   {   1   ,   1   ,   5   ,   9   ,   31  ,   1   ,   31  ,   77  ,   383 ,   151 ,   1243    ,   3165    ,   2363    ,   13751   ,   10251   ,   55111   ,0 };
            const unsigned long dim3846Kuo2Init[]       =   {   1   ,   1   ,   5   ,   13  ,   23  ,   35  ,   123 ,   229 ,   193 ,   29  ,   201 ,   2661    ,   7539    ,   10407   ,   30831   ,   56091   ,0 };
            const unsigned long dim3847Kuo2Init[]       =   {   1   ,   1   ,   5   ,   15  ,   19  ,   17  ,   83  ,   157 ,   265 ,   649 ,   815 ,   531 ,   3905    ,   10691   ,   4137    ,   29875   ,0 };
            const unsigned long dim3848Kuo2Init[]       =   {   1   ,   1   ,   3   ,   7   ,   29  ,   41  ,   29  ,   153 ,   509 ,   617 ,   151 ,   2409    ,   4631    ,   6525    ,   8539    ,   20855   ,0 };
            const unsigned long dim3849Kuo2Init[]       =   {   1   ,   1   ,   1   ,   5   ,   27  ,   37  ,   55  ,   215 ,   135 ,   463 ,   163 ,   1415    ,   1457    ,   2675    ,   9461    ,   6585    ,0 };
            const unsigned long dim3850Kuo2Init[]       =   {   1   ,   3   ,   5   ,   7   ,   23  ,   17  ,   29  ,   183 ,   33  ,   593 ,   717 ,   1487    ,   4777    ,   6281    ,   3299    ,   37531   ,0 };
            const unsigned long dim3851Kuo2Init[]       =   {   1   ,   1   ,   5   ,   15  ,   15  ,   17  ,   91  ,   243 ,   259 ,   655 ,   1347    ,   3229    ,   1449    ,   10069   ,   2671    ,   11391   ,0 };
            const unsigned long dim3852Kuo2Init[]       =   {   1   ,   1   ,   7   ,   15  ,   17  ,   35  ,   35  ,   105 ,   285 ,   761 ,   1989    ,   3427    ,   3649    ,   223 ,   21887   ,   6153    ,0 };
            const unsigned long dim3853Kuo2Init[]       =   {   1   ,   3   ,   1   ,   5   ,   17  ,   55  ,   47  ,   235 ,   7   ,   829 ,   1437    ,   1669    ,   2837    ,   6573    ,   3847    ,   32347   ,0 };
            const unsigned long dim3854Kuo2Init[]       =   {   1   ,   1   ,   5   ,   11  ,   5   ,   19  ,   41  ,   151 ,   59  ,   427 ,   1917    ,   3339    ,   4139    ,   13781   ,   15171   ,   50597   ,0 };
            const unsigned long dim3855Kuo2Init[]       =   {   1   ,   3   ,   5   ,   1   ,   3   ,   47  ,   77  ,   175 ,   239 ,   729 ,   1957    ,   2941    ,   4749    ,   15157   ,   30351   ,   57657   ,0 };
            const unsigned long dim3856Kuo2Init[]       =   {   1   ,   3   ,   5   ,   1   ,   23  ,   39  ,   51  ,   239 ,   455 ,   347 ,   1673    ,   1355    ,   1707    ,   10125   ,   18923   ,   53953   ,0 };
            const unsigned long dim3857Kuo2Init[]       =   {   1   ,   3   ,   3   ,   11  ,   5   ,   21  ,   113 ,   161 ,   325 ,   663 ,   9   ,   641 ,   4089    ,   7043    ,   9771    ,   16719   ,0 };
            const unsigned long dim3858Kuo2Init[]       =   {   1   ,   3   ,   7   ,   7   ,   1   ,   63  ,   3   ,   77  ,   67  ,   575 ,   313 ,   1249    ,   1203    ,   8121    ,   13647   ,   563 ,0 };
            const unsigned long dim3859Kuo2Init[]       =   {   1   ,   3   ,   3   ,   15  ,   17  ,   61  ,   95  ,   75  ,   421 ,   671 ,   835 ,   4015    ,   6591    ,   7589    ,   8039    ,   10055   ,0 };
            const unsigned long dim3860Kuo2Init[]       =   {   1   ,   1   ,   3   ,   3   ,   27  ,   43  ,   75  ,   121 ,   143 ,   235 ,   1243    ,   2063    ,   7711    ,   7925    ,   2943    ,   23419   ,0 };
            const unsigned long dim3861Kuo2Init[]       =   {   1   ,   3   ,   5   ,   15  ,   1   ,   21  ,   41  ,   225 ,   427 ,   843 ,   1417    ,   3481    ,   5239    ,   16373   ,   11599   ,   37267   ,0 };
            const unsigned long dim3862Kuo2Init[]       =   {   1   ,   1   ,   7   ,   15  ,   17  ,   23  ,   95  ,   185 ,   193 ,   471 ,   1475    ,   861 ,   5115    ,   4573    ,   12909   ,   12337   ,0 };
            const unsigned long dim3863Kuo2Init[]       =   {   1   ,   3   ,   1   ,   7   ,   19  ,   25  ,   107 ,   171 ,   355 ,   57  ,   1301    ,   2343    ,   6425    ,   5537    ,   29391   ,   27187   ,0 };
            const unsigned long dim3864Kuo2Init[]       =   {   1   ,   3   ,   5   ,   9   ,   19  ,   35  ,   67  ,   157 ,   183 ,   477 ,   961 ,   1087    ,   7329    ,   15787   ,   9547    ,   59001   ,0 };
            const unsigned long dim3865Kuo2Init[]       =   {   1   ,   1   ,   3   ,   5   ,   29  ,   57  ,   29  ,   119 ,   315 ,   533 ,   717 ,   1099    ,   713 ,   1473    ,   19511   ,   59777   ,0 };
            const unsigned long dim3866Kuo2Init[]       =   {   1   ,   3   ,   7   ,   7   ,   17  ,   63  ,   71  ,   181 ,   103 ,   943 ,   305 ,   1657    ,   6117    ,   11921   ,   6119    ,   26437   ,0 };
            const unsigned long dim3867Kuo2Init[]       =   {   1   ,   3   ,   7   ,   9   ,   3   ,   61  ,   91  ,   65  ,   389 ,   331 ,   549 ,   2355    ,   4591    ,   3385    ,   19913   ,   62541   ,0 };
            const unsigned long dim3868Kuo2Init[]       =   {   1   ,   1   ,   5   ,   1   ,   17  ,   1   ,   127 ,   127 ,   471 ,   139 ,   1147    ,   1463    ,   5403    ,   10877   ,   11469   ,   65283   ,0 };
            const unsigned long dim3869Kuo2Init[]       =   {   1   ,   3   ,   1   ,   13  ,   27  ,   29  ,   31  ,   57  ,   433 ,   941 ,   1559    ,   2495    ,   83  ,   14701   ,   20787   ,   10761   ,0 };
            const unsigned long dim3870Kuo2Init[]       =   {   1   ,   3   ,   3   ,   9   ,   23  ,   59  ,   23  ,   175 ,   25  ,   865 ,   751 ,   3395    ,   4111    ,   10287   ,   28191   ,   38487   ,0 };
            const unsigned long dim3871Kuo2Init[]       =   {   1   ,   1   ,   3   ,   7   ,   21  ,   3   ,   107 ,   205 ,   349 ,   473 ,   1115    ,   1647    ,   2379    ,   4709    ,   21697   ,   47899   ,0 };
            const unsigned long dim3872Kuo2Init[]       =   {   1   ,   3   ,   5   ,   15  ,   27  ,   49  ,   113 ,   73  ,   487 ,   487 ,   1531    ,   357 ,   8073    ,   10275   ,   7713    ,   9169    ,0 };
            const unsigned long dim3873Kuo2Init[]       =   {   1   ,   3   ,   3   ,   13  ,   1   ,   55  ,   97  ,   19  ,   5   ,   997 ,   1   ,   437 ,   343 ,   4867    ,   4897    ,   20291   ,0 };
            const unsigned long dim3874Kuo2Init[]       =   {   1   ,   3   ,   7   ,   5   ,   19  ,   51  ,   25  ,   3   ,   143 ,   783 ,   1415    ,   2231    ,   1809    ,   2011    ,   32607   ,   33817   ,0 };
            const unsigned long dim3875Kuo2Init[]       =   {   1   ,   1   ,   5   ,   1   ,   3   ,   59  ,   49  ,   183 ,   447 ,   55  ,   17  ,   1455    ,   3753    ,   13531   ,   17193   ,   22375   ,0 };
            const unsigned long dim3876Kuo2Init[]       =   {   1   ,   3   ,   7   ,   5   ,   21  ,   55  ,   25  ,   79  ,   161 ,   763 ,   815 ,   2161    ,   1639    ,   7165    ,   15681   ,   53681   ,0 };
            const unsigned long dim3877Kuo2Init[]       =   {   1   ,   3   ,   7   ,   3   ,   15  ,   53  ,   75  ,   243 ,   49  ,   701 ,   519 ,   1781    ,   8073    ,   16349   ,   487 ,   35051   ,0 };
            const unsigned long dim3878Kuo2Init[]       =   {   1   ,   3   ,   3   ,   15  ,   11  ,   33  ,   65  ,   93  ,   157 ,   499 ,   1487    ,   627 ,   4483    ,   9001    ,   26869   ,   11713   ,0 };
            const unsigned long dim3879Kuo2Init[]       =   {   1   ,   1   ,   3   ,   9   ,   15  ,   43  ,   29  ,   159 ,   125 ,   329 ,   1619    ,   41  ,   2565    ,   6043    ,   23983   ,   33397   ,0 };
            const unsigned long dim3880Kuo2Init[]       =   {   1   ,   1   ,   1   ,   9   ,   31  ,   25  ,   71  ,   53  ,   109 ,   415 ,   1895    ,   309 ,   5121    ,   3659    ,   24591   ,   5207    ,0 };
            const unsigned long dim3881Kuo2Init[]       =   {   1   ,   3   ,   1   ,   5   ,   23  ,   33  ,   35  ,   95  ,   171 ,   395 ,   171 ,   1085    ,   6629    ,   757 ,   29347   ,   22703   ,0 };
            const unsigned long dim3882Kuo2Init[]       =   {   1   ,   1   ,   7   ,   5   ,   17  ,   59  ,   35  ,   9   ,   141 ,   311 ,   1075    ,   2499    ,   321 ,   14733   ,   4069    ,   46015   ,0 };
            const unsigned long dim3883Kuo2Init[]       =   {   1   ,   1   ,   1   ,   13  ,   5   ,   27  ,   93  ,   105 ,   205 ,   875 ,   1867    ,   2423    ,   3015    ,   2939    ,   27529   ,   24221   ,0 };
            const unsigned long dim3884Kuo2Init[]       =   {   1   ,   1   ,   1   ,   15  ,   21  ,   51  ,   55  ,   239 ,   95  ,   517 ,   1449    ,   537 ,   8083    ,   11519   ,   30293   ,   56663   ,0 };
            const unsigned long dim3885Kuo2Init[]       =   {   1   ,   3   ,   5   ,   7   ,   3   ,   1   ,   31  ,   77  ,   75  ,   343 ,   423 ,   3781    ,   1171    ,   5331    ,   8885    ,   1917    ,0 };
            const unsigned long dim3886Kuo2Init[]       =   {   1   ,   3   ,   1   ,   13  ,   1   ,   45  ,   127 ,   123 ,   153 ,   769 ,   1549    ,   3049    ,   4023    ,   10475   ,   3185    ,   7973    ,0 };
            const unsigned long dim3887Kuo2Init[]       =   {   1   ,   3   ,   5   ,   15  ,   17  ,   17  ,   63  ,   197 ,   483 ,   253 ,   1655    ,   4053    ,   3753    ,   4159    ,   23551   ,   34501   ,0 };
            const unsigned long dim3888Kuo2Init[]       =   {   1   ,   3   ,   7   ,   7   ,   27  ,   35  ,   91  ,   241 ,   377 ,   299 ,   1741    ,   2009    ,   4051    ,   3753    ,   18631   ,   56383   ,0 };
            const unsigned long dim3889Kuo2Init[]       =   {   1   ,   1   ,   3   ,   9   ,   11  ,   45  ,   101 ,   221 ,   167 ,   845 ,   249 ,   2297    ,   967 ,   5203    ,   29305   ,   8875    ,0 };
            const unsigned long dim3890Kuo2Init[]       =   {   1   ,   3   ,   5   ,   9   ,   31  ,   41  ,   21  ,   251 ,   421 ,   47  ,   421 ,   3037    ,   867 ,   12623   ,   23289   ,   39289   ,0 };
            const unsigned long dim3891Kuo2Init[]       =   {   1   ,   1   ,   7   ,   9   ,   31  ,   13  ,   99  ,   207 ,   181 ,   683 ,   1881    ,   2047    ,   2299    ,   15795   ,   24597   ,   46881   ,0 };
            const unsigned long dim3892Kuo2Init[]       =   {   1   ,   1   ,   1   ,   15  ,   1   ,   45  ,   57  ,   171 ,   149 ,   909 ,   1493    ,   181 ,   4539    ,   9625    ,   20357   ,   9291    ,0 };
            const unsigned long dim3893Kuo2Init[]       =   {   1   ,   3   ,   5   ,   15  ,   5   ,   19  ,   101 ,   167 ,   245 ,   213 ,   1453    ,   1447    ,   19  ,   2091    ,   27345   ,   45839   ,0 };
            const unsigned long dim3894Kuo2Init[]       =   {   1   ,   1   ,   5   ,   5   ,   5   ,   61  ,   59  ,   159 ,   343 ,   51  ,   2009    ,   3651    ,   5263    ,   1425    ,   4399    ,   34357   ,0 };
            const unsigned long dim3895Kuo2Init[]       =   {   1   ,   3   ,   1   ,   15  ,   15  ,   61  ,   49  ,   15  ,   415 ,   365 ,   551 ,   797 ,   7147    ,   12771   ,   17557   ,   30663   ,0 };
            const unsigned long dim3896Kuo2Init[]       =   {   1   ,   3   ,   1   ,   7   ,   17  ,   15  ,   125 ,   195 ,   455 ,   553 ,   1259    ,   1851    ,   5149    ,   6843    ,   29195   ,   15015   ,0 };
            const unsigned long dim3897Kuo2Init[]       =   {   1   ,   3   ,   7   ,   3   ,   23  ,   21  ,   5   ,   195 ,   77  ,   881 ,   941 ,   2839    ,   2837    ,   10405   ,   22289   ,   6211    ,0 };
            const unsigned long dim3898Kuo2Init[]       =   {   1   ,   1   ,   1   ,   7   ,   11  ,   61  ,   99  ,   247 ,   251 ,   183 ,   1263    ,   157 ,   511 ,   8627    ,   19503   ,   43933   ,0 };
            const unsigned long dim3899Kuo2Init[]       =   {   1   ,   3   ,   1   ,   9   ,   15  ,   15  ,   113 ,   255 ,   225 ,   173 ,   1427    ,   3837    ,   6607    ,   3777    ,   4913    ,   11173   ,0 };
            const unsigned long dim3900Kuo2Init[]       =   {   1   ,   1   ,   1   ,   5   ,   17  ,   51  ,   65  ,   249 ,   361 ,   377 ,   1597    ,   3065    ,   5949    ,   9363    ,   25597   ,   40657   ,0 };
            const unsigned long dim3901Kuo2Init[]       =   {   1   ,   1   ,   3   ,   13  ,   15  ,   41  ,   51  ,   75  ,   277 ,   611 ,   517 ,   3751    ,   721 ,   6673    ,   32715   ,   29053   ,0 };
            const unsigned long dim3902Kuo2Init[]       =   {   1   ,   3   ,   3   ,   11  ,   31  ,   61  ,   97  ,   163 ,   79  ,   267 ,   1601    ,   3155    ,   1623    ,   6301    ,   12897   ,   1555    ,0 };
            const unsigned long dim3903Kuo2Init[]       =   {   1   ,   1   ,   5   ,   1   ,   31  ,   33  ,   29  ,   239 ,   15  ,   235 ,   1675    ,   1453    ,   1005    ,   4811    ,   21601   ,   34343   ,0 };
            const unsigned long dim3904Kuo2Init[]       =   {   1   ,   3   ,   7   ,   9   ,   13  ,   23  ,   43  ,   119 ,   313 ,   249 ,   643 ,   3797    ,   5299    ,   5127    ,   5639    ,   8509    ,0 };
            const unsigned long dim3905Kuo2Init[]       =   {   1   ,   1   ,   7   ,   5   ,   11  ,   51  ,   85  ,   181 ,   231 ,   641 ,   1285    ,   1417    ,   4535    ,   10417   ,   21261   ,   4965    ,0 };
            const unsigned long dim3906Kuo2Init[]       =   {   1   ,   3   ,   5   ,   1   ,   15  ,   47  ,   109 ,   197 ,   93  ,   115 ,   47  ,   2315    ,   863 ,   2787    ,   25487   ,   4069    ,0 };
            const unsigned long dim3907Kuo2Init[]       =   {   1   ,   1   ,   7   ,   11  ,   13  ,   43  ,   117 ,   203 ,   189 ,   151 ,   549 ,   1045    ,   7953    ,   11003   ,   17637   ,   3627    ,0 };
            const unsigned long dim3908Kuo2Init[]       =   {   1   ,   3   ,   5   ,   7   ,   7   ,   29  ,   5   ,   55  ,   281 ,   767 ,   1809    ,   2489    ,   6417    ,   10927   ,   30935   ,   10769   ,0 };
            const unsigned long dim3909Kuo2Init[]       =   {   1   ,   3   ,   7   ,   1   ,   5   ,   21  ,   75  ,   247 ,   127 ,   223 ,   103 ,   1737    ,   6577    ,   389 ,   26319   ,   20005   ,0 };
            const unsigned long dim3910Kuo2Init[]       =   {   1   ,   1   ,   1   ,   5   ,   15  ,   55  ,   33  ,   171 ,   115 ,   535 ,   1523    ,   383 ,   2029    ,   16339   ,   12799   ,   46209   ,0 };
            const unsigned long dim3911Kuo2Init[]       =   {   1   ,   1   ,   5   ,   5   ,   21  ,   39  ,   9   ,   239 ,   221 ,   497 ,   913 ,   2465    ,   561 ,   15701   ,   18851   ,   11207   ,0 };
            const unsigned long dim3912Kuo2Init[]       =   {   1   ,   3   ,   7   ,   3   ,   5   ,   33  ,   97  ,   227 ,   91  ,   283 ,   1971    ,   3791    ,   2313    ,   9519    ,   11039   ,   46973   ,0 };
            const unsigned long dim3913Kuo2Init[]       =   {   1   ,   1   ,   7   ,   7   ,   1   ,   35  ,   1   ,   23  ,   391 ,   365 ,   1099    ,   429 ,   5159    ,   409 ,   17641   ,   55545   ,0 };
            const unsigned long dim3914Kuo2Init[]       =   {   1   ,   3   ,   7   ,   9   ,   29  ,   11  ,   5   ,   63  ,   1   ,   401 ,   265 ,   753 ,   6933    ,   899 ,   10407   ,   21391   ,0 };
            const unsigned long dim3915Kuo2Init[]       =   {   1   ,   1   ,   3   ,   5   ,   25  ,   7   ,   95  ,   251 ,   241 ,   625 ,   661 ,   829 ,   1077    ,   2157    ,   27767   ,   32355   ,0 };
            const unsigned long dim3916Kuo2Init[]       =   {   1   ,   3   ,   7   ,   3   ,   21  ,   17  ,   63  ,   85  ,   83  ,   507 ,   265 ,   1183    ,   6095    ,   343 ,   11011   ,   39071   ,0 };
            const unsigned long dim3917Kuo2Init[]       =   {   1   ,   3   ,   7   ,   11  ,   7   ,   23  ,   17  ,   67  ,   83  ,   875 ,   1487    ,   2565    ,   6619    ,   8651    ,   31249   ,   28487   ,0 };
            const unsigned long dim3918Kuo2Init[]       =   {   1   ,   3   ,   3   ,   7   ,   9   ,   27  ,   63  ,   23  ,   107 ,   103 ,   1059    ,   2029    ,   1417    ,   8883    ,   7003    ,   4729    ,0 };
            const unsigned long dim3919Kuo2Init[]       =   {   1   ,   1   ,   1   ,   11  ,   5   ,   53  ,   67  ,   41  ,   227 ,   643 ,   1767    ,   615 ,   2165    ,   14475   ,   21143   ,   22079   ,0 };
            const unsigned long dim3920Kuo2Init[]       =   {   1   ,   1   ,   7   ,   7   ,   1   ,   19  ,   77  ,   211 ,   379 ,   173 ,   1097    ,   3865    ,   3735    ,   10719   ,   9223    ,   209 ,0 };
            const unsigned long dim3921Kuo2Init[]       =   {   1   ,   1   ,   5   ,   7   ,   31  ,   53  ,   77  ,   73  ,   329 ,   587 ,   1121    ,   2139    ,   7433    ,   9569    ,   20289   ,   47867   ,0 };
            const unsigned long dim3922Kuo2Init[]       =   {   1   ,   1   ,   3   ,   3   ,   7   ,   21  ,   19  ,   181 ,   209 ,   603 ,   885 ,   2251    ,   1563    ,   6435    ,   9455    ,   1753    ,0 };
            const unsigned long dim3923Kuo2Init[]       =   {   1   ,   1   ,   5   ,   11  ,   5   ,   7   ,   67  ,   111 ,   355 ,   457 ,   193 ,   1311    ,   5987    ,   5223    ,   13489   ,   49679   ,0 };
            const unsigned long dim3924Kuo2Init[]       =   {   1   ,   1   ,   3   ,   9   ,   23  ,   11  ,   21  ,   217 ,   361 ,   803 ,   325 ,   2767    ,   3363    ,   9753    ,   26523   ,   20517   ,0 };
            const unsigned long dim3925Kuo2Init[]       =   {   1   ,   3   ,   5   ,   11  ,   27  ,   13  ,   71  ,   127 ,   249 ,   49  ,   153 ,   4017    ,   289 ,   10705   ,   19671   ,   21929   ,0 };
            const unsigned long dim3926Kuo2Init[]       =   {   1   ,   1   ,   5   ,   5   ,   13  ,   51  ,   113 ,   109 ,   135 ,   507 ,   719 ,   2011    ,   3803    ,   12049   ,   13629   ,   50781   ,0 };
            const unsigned long dim3927Kuo2Init[]       =   {   1   ,   3   ,   1   ,   9   ,   5   ,   33  ,   125 ,   223 ,   309 ,   909 ,   1209    ,   3099    ,   3431    ,   853 ,   2553    ,   1845    ,0 };
            const unsigned long dim3928Kuo2Init[]       =   {   1   ,   3   ,   1   ,   5   ,   21  ,   61  ,   63  ,   21  ,   485 ,   101 ,   1307    ,   3777    ,   6331    ,   3549    ,   20867   ,   41487   ,0 };
            const unsigned long dim3929Kuo2Init[]       =   {   1   ,   3   ,   7   ,   1   ,   1   ,   33  ,   115 ,   97  ,   445 ,   521 ,   1801    ,   2869    ,   7893    ,   10229   ,   27265   ,   7749    ,0 };
            const unsigned long dim3930Kuo2Init[]       =   {   1   ,   1   ,   1   ,   15  ,   7   ,   25  ,   23  ,   87  ,   277 ,   699 ,   783 ,   2891    ,   5033    ,   14031   ,   22947   ,   15921   ,0 };
            const unsigned long dim3931Kuo2Init[]       =   {   1   ,   3   ,   5   ,   11  ,   17  ,   1   ,   53  ,   195 ,   125 ,   937 ,   739 ,   999 ,   1729    ,   14809   ,   16855   ,   16407   ,0 };
            const unsigned long dim3932Kuo2Init[]       =   {   1   ,   1   ,   7   ,   15  ,   11  ,   21  ,   13  ,   205 ,   367 ,   167 ,   1967    ,   4003    ,   1145    ,   5679    ,   26437   ,   27597   ,0 };
            const unsigned long dim3933Kuo2Init[]       =   {   1   ,   1   ,   1   ,   15  ,   15  ,   45  ,   51  ,   23  ,   51  ,   279 ,   1463    ,   1585    ,   5863    ,   6625    ,   32427   ,   33877   ,0 };
            const unsigned long dim3934Kuo2Init[]       =   {   1   ,   1   ,   1   ,   9   ,   9   ,   47  ,   83  ,   61  ,   493 ,   399 ,   1055    ,   231 ,   671 ,   11643   ,   9151    ,   48169   ,0 };
            const unsigned long dim3935Kuo2Init[]       =   {   1   ,   3   ,   1   ,   13  ,   9   ,   43  ,   3   ,   119 ,   439 ,   335 ,   1167    ,   249 ,   193 ,   4305    ,   8821    ,   49611   ,0 };
            const unsigned long dim3936Kuo2Init[]       =   {   1   ,   3   ,   5   ,   1   ,   19  ,   55  ,   43  ,   251 ,   17  ,   865 ,   147 ,   507 ,   1315    ,   12029   ,   5077    ,   6003    ,0 };
            const unsigned long dim3937Kuo2Init[]       =   {   1   ,   3   ,   7   ,   7   ,   3   ,   23  ,   83  ,   7   ,   241 ,   197 ,   139 ,   1325    ,   7777    ,   8425    ,   27451   ,   47015   ,0 };
            const unsigned long dim3938Kuo2Init[]       =   {   1   ,   1   ,   3   ,   3   ,   27  ,   29  ,   37  ,   31  ,   325 ,   937 ,   59  ,   2615    ,   7331    ,   15635   ,   29159   ,   49697   ,0 };
            const unsigned long dim3939Kuo2Init[]       =   {   1   ,   3   ,   3   ,   5   ,   27  ,   47  ,   125 ,   165 ,   363 ,   573 ,   189 ,   3249    ,   1471    ,   7869    ,   6169    ,   43525   ,0 };
            const unsigned long dim3940Kuo2Init[]       =   {   1   ,   1   ,   3   ,   13  ,   19  ,   33  ,   15  ,   123 ,   245 ,   939 ,   209 ,   853 ,   6485    ,   5059    ,   21891   ,   7115    ,0 };
            const unsigned long dim3941Kuo2Init[]       =   {   1   ,   3   ,   7   ,   1   ,   23  ,   59  ,   91  ,   197 ,   79  ,   381 ,   151 ,   3533    ,   7553    ,   511 ,   3381    ,   29339   ,0 };
            const unsigned long dim3942Kuo2Init[]       =   {   1   ,   3   ,   5   ,   3   ,   21  ,   61  ,   73  ,   57  ,   317 ,   77  ,   2027    ,   1055    ,   3599    ,   15157   ,   17017   ,   23563   ,0 };
            const unsigned long dim3943Kuo2Init[]       =   {   1   ,   1   ,   7   ,   1   ,   15  ,   43  ,   97  ,   205 ,   421 ,   901 ,   1589    ,   2051    ,   5737    ,   13231   ,   15325   ,   54725   ,0 };
            const unsigned long dim3944Kuo2Init[]       =   {   1   ,   1   ,   5   ,   5   ,   31  ,   33  ,   93  ,   129 ,   385 ,   961 ,   733 ,   277 ,   3555    ,   5619    ,   16499   ,   57747   ,0 };
            const unsigned long dim3945Kuo2Init[]       =   {   1   ,   3   ,   1   ,   5   ,   17  ,   3   ,   65  ,   167 ,   237 ,   785 ,   771 ,   4033    ,   6053    ,   343 ,   9499    ,   55861   ,0 };
            const unsigned long dim3946Kuo2Init[]       =   {   1   ,   3   ,   3   ,   3   ,   19  ,   57  ,   23  ,   119 ,   75  ,   435 ,   1023    ,   2357    ,   4671    ,   14581   ,   15343   ,   2383    ,0 };

            const unsigned long * const Kuo2initializers[3946]
            =
            {
                dim1Kuo2Init,
                dim2Kuo2Init,
                dim3Kuo2Init,
                dim4Kuo2Init,
                dim5Kuo2Init,
                dim6Kuo2Init,
                dim7Kuo2Init,
                dim8Kuo2Init,
                dim9Kuo2Init,
                dim10Kuo2Init,
                dim11Kuo2Init,
                dim12Kuo2Init,
                dim13Kuo2Init,
                dim14Kuo2Init,
                dim15Kuo2Init,
                dim16Kuo2Init,
                dim17Kuo2Init,
                dim18Kuo2Init,
                dim19Kuo2Init,
                dim20Kuo2Init,
                dim21Kuo2Init,
                dim22Kuo2Init,
                dim23Kuo2Init,
                dim24Kuo2Init,
                dim25Kuo2Init,
                dim26Kuo2Init,
                dim27Kuo2Init,
                dim28Kuo2Init,
                dim29Kuo2Init,
                dim30Kuo2Init,
                dim31Kuo2Init,
                dim32Kuo2Init,
                dim33Kuo2Init,
                dim34Kuo2Init,
                dim35Kuo2Init,
                dim36Kuo2Init,
                dim37Kuo2Init,
                dim38Kuo2Init,
                dim39Kuo2Init,
                dim40Kuo2Init,
                dim41Kuo2Init,
                dim42Kuo2Init,
                dim43Kuo2Init,
                dim44Kuo2Init,
                dim45Kuo2Init,
                dim46Kuo2Init,
                dim47Kuo2Init,
                dim48Kuo2Init,
                dim49Kuo2Init,
                dim50Kuo2Init,
                dim51Kuo2Init,
                dim52Kuo2Init,
                dim53Kuo2Init,
                dim54Kuo2Init,
                dim55Kuo2Init,
                dim56Kuo2Init,
                dim57Kuo2Init,
                dim58Kuo2Init,
                dim59Kuo2Init,
                dim60Kuo2Init,
                dim61Kuo2Init,
                dim62Kuo2Init,
                dim63Kuo2Init,
                dim64Kuo2Init,
                dim65Kuo2Init,
                dim66Kuo2Init,
                dim67Kuo2Init,
                dim68Kuo2Init,
                dim69Kuo2Init,
                dim70Kuo2Init,
                dim71Kuo2Init,
                dim72Kuo2Init,
                dim73Kuo2Init,
                dim74Kuo2Init,
                dim75Kuo2Init,
                dim76Kuo2Init,
                dim77Kuo2Init,
                dim78Kuo2Init,
                dim79Kuo2Init,
                dim80Kuo2Init,
                dim81Kuo2Init,
                dim82Kuo2Init,
                dim83Kuo2Init,
                dim84Kuo2Init,
                dim85Kuo2Init,
                dim86Kuo2Init,
                dim87Kuo2Init,
                dim88Kuo2Init,
                dim89Kuo2Init,
                dim90Kuo2Init,
                dim91Kuo2Init,
                dim92Kuo2Init,
                dim93Kuo2Init,
                dim94Kuo2Init,
                dim95Kuo2Init,
                dim96Kuo2Init,
                dim97Kuo2Init,
                dim98Kuo2Init,
                dim99Kuo2Init,
                dim100Kuo2Init,
                dim101Kuo2Init,
                dim102Kuo2Init,
                dim103Kuo2Init,
                dim104Kuo2Init,
                dim105Kuo2Init,
                dim106Kuo2Init,
                dim107Kuo2Init,
                dim108Kuo2Init,
                dim109Kuo2Init,
                dim110Kuo2Init,
                dim111Kuo2Init,
                dim112Kuo2Init,
                dim113Kuo2Init,
                dim114Kuo2Init,
                dim115Kuo2Init,
                dim116Kuo2Init,
                dim117Kuo2Init,
                dim118Kuo2Init,
                dim119Kuo2Init,
                dim120Kuo2Init,
                dim121Kuo2Init,
                dim122Kuo2Init,
                dim123Kuo2Init,
                dim124Kuo2Init,
                dim125Kuo2Init,
                dim126Kuo2Init,
                dim127Kuo2Init,
                dim128Kuo2Init,
                dim129Kuo2Init,
                dim130Kuo2Init,
                dim131Kuo2Init,
                dim132Kuo2Init,
                dim133Kuo2Init,
                dim134Kuo2Init,
                dim135Kuo2Init,
                dim136Kuo2Init,
                dim137Kuo2Init,
                dim138Kuo2Init,
                dim139Kuo2Init,
                dim140Kuo2Init,
                dim141Kuo2Init,
                dim142Kuo2Init,
                dim143Kuo2Init,
                dim144Kuo2Init,
                dim145Kuo2Init,
                dim146Kuo2Init,
                dim147Kuo2Init,
                dim148Kuo2Init,
                dim149Kuo2Init,
                dim150Kuo2Init,
                dim151Kuo2Init,
                dim152Kuo2Init,
                dim153Kuo2Init,
                dim154Kuo2Init,
                dim155Kuo2Init,
                dim156Kuo2Init,
                dim157Kuo2Init,
                dim158Kuo2Init,
                dim159Kuo2Init,
                dim160Kuo2Init,
                dim161Kuo2Init,
                dim162Kuo2Init,
                dim163Kuo2Init,
                dim164Kuo2Init,
                dim165Kuo2Init,
                dim166Kuo2Init,
                dim167Kuo2Init,
                dim168Kuo2Init,
                dim169Kuo2Init,
                dim170Kuo2Init,
                dim171Kuo2Init,
                dim172Kuo2Init,
                dim173Kuo2Init,
                dim174Kuo2Init,
                dim175Kuo2Init,
                dim176Kuo2Init,
                dim177Kuo2Init,
                dim178Kuo2Init,
                dim179Kuo2Init,
                dim180Kuo2Init,
                dim181Kuo2Init,
                dim182Kuo2Init,
                dim183Kuo2Init,
                dim184Kuo2Init,
                dim185Kuo2Init,
                dim186Kuo2Init,
                dim187Kuo2Init,
                dim188Kuo2Init,
                dim189Kuo2Init,
                dim190Kuo2Init,
                dim191Kuo2Init,
                dim192Kuo2Init,
                dim193Kuo2Init,
                dim194Kuo2Init,
                dim195Kuo2Init,
                dim196Kuo2Init,
                dim197Kuo2Init,
                dim198Kuo2Init,
                dim199Kuo2Init,
                dim200Kuo2Init,
                dim201Kuo2Init,
                dim202Kuo2Init,
                dim203Kuo2Init,
                dim204Kuo2Init,
                dim205Kuo2Init,
                dim206Kuo2Init,
                dim207Kuo2Init,
                dim208Kuo2Init,
                dim209Kuo2Init,
                dim210Kuo2Init,
                dim211Kuo2Init,
                dim212Kuo2Init,
                dim213Kuo2Init,
                dim214Kuo2Init,
                dim215Kuo2Init,
                dim216Kuo2Init,
                dim217Kuo2Init,
                dim218Kuo2Init,
                dim219Kuo2Init,
                dim220Kuo2Init,
                dim221Kuo2Init,
                dim222Kuo2Init,
                dim223Kuo2Init,
                dim224Kuo2Init,
                dim225Kuo2Init,
                dim226Kuo2Init,
                dim227Kuo2Init,
                dim228Kuo2Init,
                dim229Kuo2Init,
                dim230Kuo2Init,
                dim231Kuo2Init,
                dim232Kuo2Init,
                dim233Kuo2Init,
                dim234Kuo2Init,
                dim235Kuo2Init,
                dim236Kuo2Init,
                dim237Kuo2Init,
                dim238Kuo2Init,
                dim239Kuo2Init,
                dim240Kuo2Init,
                dim241Kuo2Init,
                dim242Kuo2Init,
                dim243Kuo2Init,
                dim244Kuo2Init,
                dim245Kuo2Init,
                dim246Kuo2Init,
                dim247Kuo2Init,
                dim248Kuo2Init,
                dim249Kuo2Init,
                dim250Kuo2Init,
                dim251Kuo2Init,
                dim252Kuo2Init,
                dim253Kuo2Init,
                dim254Kuo2Init,
                dim255Kuo2Init,
                dim256Kuo2Init,
                dim257Kuo2Init,
                dim258Kuo2Init,
                dim259Kuo2Init,
                dim260Kuo2Init,
                dim261Kuo2Init,
                dim262Kuo2Init,
                dim263Kuo2Init,
                dim264Kuo2Init,
                dim265Kuo2Init,
                dim266Kuo2Init,
                dim267Kuo2Init,
                dim268Kuo2Init,
                dim269Kuo2Init,
                dim270Kuo2Init,
                dim271Kuo2Init,
                dim272Kuo2Init,
                dim273Kuo2Init,
                dim274Kuo2Init,
                dim275Kuo2Init,
                dim276Kuo2Init,
                dim277Kuo2Init,
                dim278Kuo2Init,
                dim279Kuo2Init,
                dim280Kuo2Init,
                dim281Kuo2Init,
                dim282Kuo2Init,
                dim283Kuo2Init,
                dim284Kuo2Init,
                dim285Kuo2Init,
                dim286Kuo2Init,
                dim287Kuo2Init,
                dim288Kuo2Init,
                dim289Kuo2Init,
                dim290Kuo2Init,
                dim291Kuo2Init,
                dim292Kuo2Init,
                dim293Kuo2Init,
                dim294Kuo2Init,
                dim295Kuo2Init,
                dim296Kuo2Init,
                dim297Kuo2Init,
                dim298Kuo2Init,
                dim299Kuo2Init,
                dim300Kuo2Init,
                dim301Kuo2Init,
                dim302Kuo2Init,
                dim303Kuo2Init,
                dim304Kuo2Init,
                dim305Kuo2Init,
                dim306Kuo2Init,
                dim307Kuo2Init,
                dim308Kuo2Init,
                dim309Kuo2Init,
                dim310Kuo2Init,
                dim311Kuo2Init,
                dim312Kuo2Init,
                dim313Kuo2Init,
                dim314Kuo2Init,
                dim315Kuo2Init,
                dim316Kuo2Init,
                dim317Kuo2Init,
                dim318Kuo2Init,
                dim319Kuo2Init,
                dim320Kuo2Init,
                dim321Kuo2Init,
                dim322Kuo2Init,
                dim323Kuo2Init,
                dim324Kuo2Init,
                dim325Kuo2Init,
                dim326Kuo2Init,
                dim327Kuo2Init,
                dim328Kuo2Init,
                dim329Kuo2Init,
                dim330Kuo2Init,
                dim331Kuo2Init,
                dim332Kuo2Init,
                dim333Kuo2Init,
                dim334Kuo2Init,
                dim335Kuo2Init,
                dim336Kuo2Init,
                dim337Kuo2Init,
                dim338Kuo2Init,
                dim339Kuo2Init,
                dim340Kuo2Init,
                dim341Kuo2Init,
                dim342Kuo2Init,
                dim343Kuo2Init,
                dim344Kuo2Init,
                dim345Kuo2Init,
                dim346Kuo2Init,
                dim347Kuo2Init,
                dim348Kuo2Init,
                dim349Kuo2Init,
                dim350Kuo2Init,
                dim351Kuo2Init,
                dim352Kuo2Init,
                dim353Kuo2Init,
                dim354Kuo2Init,
                dim355Kuo2Init,
                dim356Kuo2Init,
                dim357Kuo2Init,
                dim358Kuo2Init,
                dim359Kuo2Init,
                dim360Kuo2Init,
                dim361Kuo2Init,
                dim362Kuo2Init,
                dim363Kuo2Init,
                dim364Kuo2Init,
                dim365Kuo2Init,
                dim366Kuo2Init,
                dim367Kuo2Init,
                dim368Kuo2Init,
                dim369Kuo2Init,
                dim370Kuo2Init,
                dim371Kuo2Init,
                dim372Kuo2Init,
                dim373Kuo2Init,
                dim374Kuo2Init,
                dim375Kuo2Init,
                dim376Kuo2Init,
                dim377Kuo2Init,
                dim378Kuo2Init,
                dim379Kuo2Init,
                dim380Kuo2Init,
                dim381Kuo2Init,
                dim382Kuo2Init,
                dim383Kuo2Init,
                dim384Kuo2Init,
                dim385Kuo2Init,
                dim386Kuo2Init,
                dim387Kuo2Init,
                dim388Kuo2Init,
                dim389Kuo2Init,
                dim390Kuo2Init,
                dim391Kuo2Init,
                dim392Kuo2Init,
                dim393Kuo2Init,
                dim394Kuo2Init,
                dim395Kuo2Init,
                dim396Kuo2Init,
                dim397Kuo2Init,
                dim398Kuo2Init,
                dim399Kuo2Init,
                dim400Kuo2Init,
                dim401Kuo2Init,
                dim402Kuo2Init,
                dim403Kuo2Init,
                dim404Kuo2Init,
                dim405Kuo2Init,
                dim406Kuo2Init,
                dim407Kuo2Init,
                dim408Kuo2Init,
                dim409Kuo2Init,
                dim410Kuo2Init,
                dim411Kuo2Init,
                dim412Kuo2Init,
                dim413Kuo2Init,
                dim414Kuo2Init,
                dim415Kuo2Init,
                dim416Kuo2Init,
                dim417Kuo2Init,
                dim418Kuo2Init,
                dim419Kuo2Init,
                dim420Kuo2Init,
                dim421Kuo2Init,
                dim422Kuo2Init,
                dim423Kuo2Init,
                dim424Kuo2Init,
                dim425Kuo2Init,
                dim426Kuo2Init,
                dim427Kuo2Init,
                dim428Kuo2Init,
                dim429Kuo2Init,
                dim430Kuo2Init,
                dim431Kuo2Init,
                dim432Kuo2Init,
                dim433Kuo2Init,
                dim434Kuo2Init,
                dim435Kuo2Init,
                dim436Kuo2Init,
                dim437Kuo2Init,
                dim438Kuo2Init,
                dim439Kuo2Init,
                dim440Kuo2Init,
                dim441Kuo2Init,
                dim442Kuo2Init,
                dim443Kuo2Init,
                dim444Kuo2Init,
                dim445Kuo2Init,
                dim446Kuo2Init,
                dim447Kuo2Init,
                dim448Kuo2Init,
                dim449Kuo2Init,
                dim450Kuo2Init,
                dim451Kuo2Init,
                dim452Kuo2Init,
                dim453Kuo2Init,
                dim454Kuo2Init,
                dim455Kuo2Init,
                dim456Kuo2Init,
                dim457Kuo2Init,
                dim458Kuo2Init,
                dim459Kuo2Init,
                dim460Kuo2Init,
                dim461Kuo2Init,
                dim462Kuo2Init,
                dim463Kuo2Init,
                dim464Kuo2Init,
                dim465Kuo2Init,
                dim466Kuo2Init,
                dim467Kuo2Init,
                dim468Kuo2Init,
                dim469Kuo2Init,
                dim470Kuo2Init,
                dim471Kuo2Init,
                dim472Kuo2Init,
                dim473Kuo2Init,
                dim474Kuo2Init,
                dim475Kuo2Init,
                dim476Kuo2Init,
                dim477Kuo2Init,
                dim478Kuo2Init,
                dim479Kuo2Init,
                dim480Kuo2Init,
                dim481Kuo2Init,
                dim482Kuo2Init,
                dim483Kuo2Init,
                dim484Kuo2Init,
                dim485Kuo2Init,
                dim486Kuo2Init,
                dim487Kuo2Init,
                dim488Kuo2Init,
                dim489Kuo2Init,
                dim490Kuo2Init,
                dim491Kuo2Init,
                dim492Kuo2Init,
                dim493Kuo2Init,
                dim494Kuo2Init,
                dim495Kuo2Init,
                dim496Kuo2Init,
                dim497Kuo2Init,
                dim498Kuo2Init,
                dim499Kuo2Init,
                dim500Kuo2Init,
                dim501Kuo2Init,
                dim502Kuo2Init,
                dim503Kuo2Init,
                dim504Kuo2Init,
                dim505Kuo2Init,
                dim506Kuo2Init,
                dim507Kuo2Init,
                dim508Kuo2Init,
                dim509Kuo2Init,
                dim510Kuo2Init,
                dim511Kuo2Init,
                dim512Kuo2Init,
                dim513Kuo2Init,
                dim514Kuo2Init,
                dim515Kuo2Init,
                dim516Kuo2Init,
                dim517Kuo2Init,
                dim518Kuo2Init,
                dim519Kuo2Init,
                dim520Kuo2Init,
                dim521Kuo2Init,
                dim522Kuo2Init,
                dim523Kuo2Init,
                dim524Kuo2Init,
                dim525Kuo2Init,
                dim526Kuo2Init,
                dim527Kuo2Init,
                dim528Kuo2Init,
                dim529Kuo2Init,
                dim530Kuo2Init,
                dim531Kuo2Init,
                dim532Kuo2Init,
                dim533Kuo2Init,
                dim534Kuo2Init,
                dim535Kuo2Init,
                dim536Kuo2Init,
                dim537Kuo2Init,
                dim538Kuo2Init,
                dim539Kuo2Init,
                dim540Kuo2Init,
                dim541Kuo2Init,
                dim542Kuo2Init,
                dim543Kuo2Init,
                dim544Kuo2Init,
                dim545Kuo2Init,
                dim546Kuo2Init,
                dim547Kuo2Init,
                dim548Kuo2Init,
                dim549Kuo2Init,
                dim550Kuo2Init,
                dim551Kuo2Init,
                dim552Kuo2Init,
                dim553Kuo2Init,
                dim554Kuo2Init,
                dim555Kuo2Init,
                dim556Kuo2Init,
                dim557Kuo2Init,
                dim558Kuo2Init,
                dim559Kuo2Init,
                dim560Kuo2Init,
                dim561Kuo2Init,
                dim562Kuo2Init,
                dim563Kuo2Init,
                dim564Kuo2Init,
                dim565Kuo2Init,
                dim566Kuo2Init,
                dim567Kuo2Init,
                dim568Kuo2Init,
                dim569Kuo2Init,
                dim570Kuo2Init,
                dim571Kuo2Init,
                dim572Kuo2Init,
                dim573Kuo2Init,
                dim574Kuo2Init,
                dim575Kuo2Init,
                dim576Kuo2Init,
                dim577Kuo2Init,
                dim578Kuo2Init,
                dim579Kuo2Init,
                dim580Kuo2Init,
                dim581Kuo2Init,
                dim582Kuo2Init,
                dim583Kuo2Init,
                dim584Kuo2Init,
                dim585Kuo2Init,
                dim586Kuo2Init,
                dim587Kuo2Init,
                dim588Kuo2Init,
                dim589Kuo2Init,
                dim590Kuo2Init,
                dim591Kuo2Init,
                dim592Kuo2Init,
                dim593Kuo2Init,
                dim594Kuo2Init,
                dim595Kuo2Init,
                dim596Kuo2Init,
                dim597Kuo2Init,
                dim598Kuo2Init,
                dim599Kuo2Init,
                dim600Kuo2Init,
                dim601Kuo2Init,
                dim602Kuo2Init,
                dim603Kuo2Init,
                dim604Kuo2Init,
                dim605Kuo2Init,
                dim606Kuo2Init,
                dim607Kuo2Init,
                dim608Kuo2Init,
                dim609Kuo2Init,
                dim610Kuo2Init,
                dim611Kuo2Init,
                dim612Kuo2Init,
                dim613Kuo2Init,
                dim614Kuo2Init,
                dim615Kuo2Init,
                dim616Kuo2Init,
                dim617Kuo2Init,
                dim618Kuo2Init,
                dim619Kuo2Init,
                dim620Kuo2Init,
                dim621Kuo2Init,
                dim622Kuo2Init,
                dim623Kuo2Init,
                dim624Kuo2Init,
                dim625Kuo2Init,
                dim626Kuo2Init,
                dim627Kuo2Init,
                dim628Kuo2Init,
                dim629Kuo2Init,
                dim630Kuo2Init,
                dim631Kuo2Init,
                dim632Kuo2Init,
                dim633Kuo2Init,
                dim634Kuo2Init,
                dim635Kuo2Init,
                dim636Kuo2Init,
                dim637Kuo2Init,
                dim638Kuo2Init,
                dim639Kuo2Init,
                dim640Kuo2Init,
                dim641Kuo2Init,
                dim642Kuo2Init,
                dim643Kuo2Init,
                dim644Kuo2Init,
                dim645Kuo2Init,
                dim646Kuo2Init,
                dim647Kuo2Init,
                dim648Kuo2Init,
                dim649Kuo2Init,
                dim650Kuo2Init,
                dim651Kuo2Init,
                dim652Kuo2Init,
                dim653Kuo2Init,
                dim654Kuo2Init,
                dim655Kuo2Init,
                dim656Kuo2Init,
                dim657Kuo2Init,
                dim658Kuo2Init,
                dim659Kuo2Init,
                dim660Kuo2Init,
                dim661Kuo2Init,
                dim662Kuo2Init,
                dim663Kuo2Init,
                dim664Kuo2Init,
                dim665Kuo2Init,
                dim666Kuo2Init,
                dim667Kuo2Init,
                dim668Kuo2Init,
                dim669Kuo2Init,
                dim670Kuo2Init,
                dim671Kuo2Init,
                dim672Kuo2Init,
                dim673Kuo2Init,
                dim674Kuo2Init,
                dim675Kuo2Init,
                dim676Kuo2Init,
                dim677Kuo2Init,
                dim678Kuo2Init,
                dim679Kuo2Init,
                dim680Kuo2Init,
                dim681Kuo2Init,
                dim682Kuo2Init,
                dim683Kuo2Init,
                dim684Kuo2Init,
                dim685Kuo2Init,
                dim686Kuo2Init,
                dim687Kuo2Init,
                dim688Kuo2Init,
                dim689Kuo2Init,
                dim690Kuo2Init,
                dim691Kuo2Init,
                dim692Kuo2Init,
                dim693Kuo2Init,
                dim694Kuo2Init,
                dim695Kuo2Init,
                dim696Kuo2Init,
                dim697Kuo2Init,
                dim698Kuo2Init,
                dim699Kuo2Init,
                dim700Kuo2Init,
                dim701Kuo2Init,
                dim702Kuo2Init,
                dim703Kuo2Init,
                dim704Kuo2Init,
                dim705Kuo2Init,
                dim706Kuo2Init,
                dim707Kuo2Init,
                dim708Kuo2Init,
                dim709Kuo2Init,
                dim710Kuo2Init,
                dim711Kuo2Init,
                dim712Kuo2Init,
                dim713Kuo2Init,
                dim714Kuo2Init,
                dim715Kuo2Init,
                dim716Kuo2Init,
                dim717Kuo2Init,
                dim718Kuo2Init,
                dim719Kuo2Init,
                dim720Kuo2Init,
                dim721Kuo2Init,
                dim722Kuo2Init,
                dim723Kuo2Init,
                dim724Kuo2Init,
                dim725Kuo2Init,
                dim726Kuo2Init,
                dim727Kuo2Init,
                dim728Kuo2Init,
                dim729Kuo2Init,
                dim730Kuo2Init,
                dim731Kuo2Init,
                dim732Kuo2Init,
                dim733Kuo2Init,
                dim734Kuo2Init,
                dim735Kuo2Init,
                dim736Kuo2Init,
                dim737Kuo2Init,
                dim738Kuo2Init,
                dim739Kuo2Init,
                dim740Kuo2Init,
                dim741Kuo2Init,
                dim742Kuo2Init,
                dim743Kuo2Init,
                dim744Kuo2Init,
                dim745Kuo2Init,
                dim746Kuo2Init,
                dim747Kuo2Init,
                dim748Kuo2Init,
                dim749Kuo2Init,
                dim750Kuo2Init,
                dim751Kuo2Init,
                dim752Kuo2Init,
                dim753Kuo2Init,
                dim754Kuo2Init,
                dim755Kuo2Init,
                dim756Kuo2Init,
                dim757Kuo2Init,
                dim758Kuo2Init,
                dim759Kuo2Init,
                dim760Kuo2Init,
                dim761Kuo2Init,
                dim762Kuo2Init,
                dim763Kuo2Init,
                dim764Kuo2Init,
                dim765Kuo2Init,
                dim766Kuo2Init,
                dim767Kuo2Init,
                dim768Kuo2Init,
                dim769Kuo2Init,
                dim770Kuo2Init,
                dim771Kuo2Init,
                dim772Kuo2Init,
                dim773Kuo2Init,
                dim774Kuo2Init,
                dim775Kuo2Init,
                dim776Kuo2Init,
                dim777Kuo2Init,
                dim778Kuo2Init,
                dim779Kuo2Init,
                dim780Kuo2Init,
                dim781Kuo2Init,
                dim782Kuo2Init,
                dim783Kuo2Init,
                dim784Kuo2Init,
                dim785Kuo2Init,
                dim786Kuo2Init,
                dim787Kuo2Init,
                dim788Kuo2Init,
                dim789Kuo2Init,
                dim790Kuo2Init,
                dim791Kuo2Init,
                dim792Kuo2Init,
                dim793Kuo2Init,
                dim794Kuo2Init,
                dim795Kuo2Init,
                dim796Kuo2Init,
                dim797Kuo2Init,
                dim798Kuo2Init,
                dim799Kuo2Init,
                dim800Kuo2Init,
                dim801Kuo2Init,
                dim802Kuo2Init,
                dim803Kuo2Init,
                dim804Kuo2Init,
                dim805Kuo2Init,
                dim806Kuo2Init,
                dim807Kuo2Init,
                dim808Kuo2Init,
                dim809Kuo2Init,
                dim810Kuo2Init,
                dim811Kuo2Init,
                dim812Kuo2Init,
                dim813Kuo2Init,
                dim814Kuo2Init,
                dim815Kuo2Init,
                dim816Kuo2Init,
                dim817Kuo2Init,
                dim818Kuo2Init,
                dim819Kuo2Init,
                dim820Kuo2Init,
                dim821Kuo2Init,
                dim822Kuo2Init,
                dim823Kuo2Init,
                dim824Kuo2Init,
                dim825Kuo2Init,
                dim826Kuo2Init,
                dim827Kuo2Init,
                dim828Kuo2Init,
                dim829Kuo2Init,
                dim830Kuo2Init,
                dim831Kuo2Init,
                dim832Kuo2Init,
                dim833Kuo2Init,
                dim834Kuo2Init,
                dim835Kuo2Init,
                dim836Kuo2Init,
                dim837Kuo2Init,
                dim838Kuo2Init,
                dim839Kuo2Init,
                dim840Kuo2Init,
                dim841Kuo2Init,
                dim842Kuo2Init,
                dim843Kuo2Init,
                dim844Kuo2Init,
                dim845Kuo2Init,
                dim846Kuo2Init,
                dim847Kuo2Init,
                dim848Kuo2Init,
                dim849Kuo2Init,
                dim850Kuo2Init,
                dim851Kuo2Init,
                dim852Kuo2Init,
                dim853Kuo2Init,
                dim854Kuo2Init,
                dim855Kuo2Init,
                dim856Kuo2Init,
                dim857Kuo2Init,
                dim858Kuo2Init,
                dim859Kuo2Init,
                dim860Kuo2Init,
                dim861Kuo2Init,
                dim862Kuo2Init,
                dim863Kuo2Init,
                dim864Kuo2Init,
                dim865Kuo2Init,
                dim866Kuo2Init,
                dim867Kuo2Init,
                dim868Kuo2Init,
                dim869Kuo2Init,
                dim870Kuo2Init,
                dim871Kuo2Init,
                dim872Kuo2Init,
                dim873Kuo2Init,
                dim874Kuo2Init,
                dim875Kuo2Init,
                dim876Kuo2Init,
                dim877Kuo2Init,
                dim878Kuo2Init,
                dim879Kuo2Init,
                dim880Kuo2Init,
                dim881Kuo2Init,
                dim882Kuo2Init,
                dim883Kuo2Init,
                dim884Kuo2Init,
                dim885Kuo2Init,
                dim886Kuo2Init,
                dim887Kuo2Init,
                dim888Kuo2Init,
                dim889Kuo2Init,
                dim890Kuo2Init,
                dim891Kuo2Init,
                dim892Kuo2Init,
                dim893Kuo2Init,
                dim894Kuo2Init,
                dim895Kuo2Init,
                dim896Kuo2Init,
                dim897Kuo2Init,
                dim898Kuo2Init,
                dim899Kuo2Init,
                dim900Kuo2Init,
                dim901Kuo2Init,
                dim902Kuo2Init,
                dim903Kuo2Init,
                dim904Kuo2Init,
                dim905Kuo2Init,
                dim906Kuo2Init,
                dim907Kuo2Init,
                dim908Kuo2Init,
                dim909Kuo2Init,
                dim910Kuo2Init,
                dim911Kuo2Init,
                dim912Kuo2Init,
                dim913Kuo2Init,
                dim914Kuo2Init,
                dim915Kuo2Init,
                dim916Kuo2Init,
                dim917Kuo2Init,
                dim918Kuo2Init,
                dim919Kuo2Init,
                dim920Kuo2Init,
                dim921Kuo2Init,
                dim922Kuo2Init,
                dim923Kuo2Init,
                dim924Kuo2Init,
                dim925Kuo2Init,
                dim926Kuo2Init,
                dim927Kuo2Init,
                dim928Kuo2Init,
                dim929Kuo2Init,
                dim930Kuo2Init,
                dim931Kuo2Init,
                dim932Kuo2Init,
                dim933Kuo2Init,
                dim934Kuo2Init,
                dim935Kuo2Init,
                dim936Kuo2Init,
                dim937Kuo2Init,
                dim938Kuo2Init,
                dim939Kuo2Init,
                dim940Kuo2Init,
                dim941Kuo2Init,
                dim942Kuo2Init,
                dim943Kuo2Init,
                dim944Kuo2Init,
                dim945Kuo2Init,
                dim946Kuo2Init,
                dim947Kuo2Init,
                dim948Kuo2Init,
                dim949Kuo2Init,
                dim950Kuo2Init,
                dim951Kuo2Init,
                dim952Kuo2Init,
                dim953Kuo2Init,
                dim954Kuo2Init,
                dim955Kuo2Init,
                dim956Kuo2Init,
                dim957Kuo2Init,
                dim958Kuo2Init,
                dim959Kuo2Init,
                dim960Kuo2Init,
                dim961Kuo2Init,
                dim962Kuo2Init,
                dim963Kuo2Init,
                dim964Kuo2Init,
                dim965Kuo2Init,
                dim966Kuo2Init,
                dim967Kuo2Init,
                dim968Kuo2Init,
                dim969Kuo2Init,
                dim970Kuo2Init,
                dim971Kuo2Init,
                dim972Kuo2Init,
                dim973Kuo2Init,
                dim974Kuo2Init,
                dim975Kuo2Init,
                dim976Kuo2Init,
                dim977Kuo2Init,
                dim978Kuo2Init,
                dim979Kuo2Init,
                dim980Kuo2Init,
                dim981Kuo2Init,
                dim982Kuo2Init,
                dim983Kuo2Init,
                dim984Kuo2Init,
                dim985Kuo2Init,
                dim986Kuo2Init,
                dim987Kuo2Init,
                dim988Kuo2Init,
                dim989Kuo2Init,
                dim990Kuo2Init,
                dim991Kuo2Init,
                dim992Kuo2Init,
                dim993Kuo2Init,
                dim994Kuo2Init,
                dim995Kuo2Init,
                dim996Kuo2Init,
                dim997Kuo2Init,
                dim998Kuo2Init,
                dim999Kuo2Init,
                dim1000Kuo2Init,
                dim1001Kuo2Init,
                dim1002Kuo2Init,
                dim1003Kuo2Init,
                dim1004Kuo2Init,
                dim1005Kuo2Init,
                dim1006Kuo2Init,
                dim1007Kuo2Init,
                dim1008Kuo2Init,
                dim1009Kuo2Init,
                dim1010Kuo2Init,
                dim1011Kuo2Init,
                dim1012Kuo2Init,
                dim1013Kuo2Init,
                dim1014Kuo2Init,
                dim1015Kuo2Init,
                dim1016Kuo2Init,
                dim1017Kuo2Init,
                dim1018Kuo2Init,
                dim1019Kuo2Init,
                dim1020Kuo2Init,
                dim1021Kuo2Init,
                dim1022Kuo2Init,
                dim1023Kuo2Init,
                dim1024Kuo2Init,
                dim1025Kuo2Init,
                dim1026Kuo2Init,
                dim1027Kuo2Init,
                dim1028Kuo2Init,
                dim1029Kuo2Init,
                dim1030Kuo2Init,
                dim1031Kuo2Init,
                dim1032Kuo2Init,
                dim1033Kuo2Init,
                dim1034Kuo2Init,
                dim1035Kuo2Init,
                dim1036Kuo2Init,
                dim1037Kuo2Init,
                dim1038Kuo2Init,
                dim1039Kuo2Init,
                dim1040Kuo2Init,
                dim1041Kuo2Init,
                dim1042Kuo2Init,
                dim1043Kuo2Init,
                dim1044Kuo2Init,
                dim1045Kuo2Init,
                dim1046Kuo2Init,
                dim1047Kuo2Init,
                dim1048Kuo2Init,
                dim1049Kuo2Init,
                dim1050Kuo2Init,
                dim1051Kuo2Init,
                dim1052Kuo2Init,
                dim1053Kuo2Init,
                dim1054Kuo2Init,
                dim1055Kuo2Init,
                dim1056Kuo2Init,
                dim1057Kuo2Init,
                dim1058Kuo2Init,
                dim1059Kuo2Init,
                dim1060Kuo2Init,
                dim1061Kuo2Init,
                dim1062Kuo2Init,
                dim1063Kuo2Init,
                dim1064Kuo2Init,
                dim1065Kuo2Init,
                dim1066Kuo2Init,
                dim1067Kuo2Init,
                dim1068Kuo2Init,
                dim1069Kuo2Init,
                dim1070Kuo2Init,
                dim1071Kuo2Init,
                dim1072Kuo2Init,
                dim1073Kuo2Init,
                dim1074Kuo2Init,
                dim1075Kuo2Init,
                dim1076Kuo2Init,
                dim1077Kuo2Init,
                dim1078Kuo2Init,
                dim1079Kuo2Init,
                dim1080Kuo2Init,
                dim1081Kuo2Init,
                dim1082Kuo2Init,
                dim1083Kuo2Init,
                dim1084Kuo2Init,
                dim1085Kuo2Init,
                dim1086Kuo2Init,
                dim1087Kuo2Init,
                dim1088Kuo2Init,
                dim1089Kuo2Init,
                dim1090Kuo2Init,
                dim1091Kuo2Init,
                dim1092Kuo2Init,
                dim1093Kuo2Init,
                dim1094Kuo2Init,
                dim1095Kuo2Init,
                dim1096Kuo2Init,
                dim1097Kuo2Init,
                dim1098Kuo2Init,
                dim1099Kuo2Init,
                dim1100Kuo2Init,
                dim1101Kuo2Init,
                dim1102Kuo2Init,
                dim1103Kuo2Init,
                dim1104Kuo2Init,
                dim1105Kuo2Init,
                dim1106Kuo2Init,
                dim1107Kuo2Init,
                dim1108Kuo2Init,
                dim1109Kuo2Init,
                dim1110Kuo2Init,
                dim1111Kuo2Init,
                dim1112Kuo2Init,
                dim1113Kuo2Init,
                dim1114Kuo2Init,
                dim1115Kuo2Init,
                dim1116Kuo2Init,
                dim1117Kuo2Init,
                dim1118Kuo2Init,
                dim1119Kuo2Init,
                dim1120Kuo2Init,
                dim1121Kuo2Init,
                dim1122Kuo2Init,
                dim1123Kuo2Init,
                dim1124Kuo2Init,
                dim1125Kuo2Init,
                dim1126Kuo2Init,
                dim1127Kuo2Init,
                dim1128Kuo2Init,
                dim1129Kuo2Init,
                dim1130Kuo2Init,
                dim1131Kuo2Init,
                dim1132Kuo2Init,
                dim1133Kuo2Init,
                dim1134Kuo2Init,
                dim1135Kuo2Init,
                dim1136Kuo2Init,
                dim1137Kuo2Init,
                dim1138Kuo2Init,
                dim1139Kuo2Init,
                dim1140Kuo2Init,
                dim1141Kuo2Init,
                dim1142Kuo2Init,
                dim1143Kuo2Init,
                dim1144Kuo2Init,
                dim1145Kuo2Init,
                dim1146Kuo2Init,
                dim1147Kuo2Init,
                dim1148Kuo2Init,
                dim1149Kuo2Init,
                dim1150Kuo2Init,
                dim1151Kuo2Init,
                dim1152Kuo2Init,
                dim1153Kuo2Init,
                dim1154Kuo2Init,
                dim1155Kuo2Init,
                dim1156Kuo2Init,
                dim1157Kuo2Init,
                dim1158Kuo2Init,
                dim1159Kuo2Init,
                dim1160Kuo2Init,
                dim1161Kuo2Init,
                dim1162Kuo2Init,
                dim1163Kuo2Init,
                dim1164Kuo2Init,
                dim1165Kuo2Init,
                dim1166Kuo2Init,
                dim1167Kuo2Init,
                dim1168Kuo2Init,
                dim1169Kuo2Init,
                dim1170Kuo2Init,
                dim1171Kuo2Init,
                dim1172Kuo2Init,
                dim1173Kuo2Init,
                dim1174Kuo2Init,
                dim1175Kuo2Init,
                dim1176Kuo2Init,
                dim1177Kuo2Init,
                dim1178Kuo2Init,
                dim1179Kuo2Init,
                dim1180Kuo2Init,
                dim1181Kuo2Init,
                dim1182Kuo2Init,
                dim1183Kuo2Init,
                dim1184Kuo2Init,
                dim1185Kuo2Init,
                dim1186Kuo2Init,
                dim1187Kuo2Init,
                dim1188Kuo2Init,
                dim1189Kuo2Init,
                dim1190Kuo2Init,
                dim1191Kuo2Init,
                dim1192Kuo2Init,
                dim1193Kuo2Init,
                dim1194Kuo2Init,
                dim1195Kuo2Init,
                dim1196Kuo2Init,
                dim1197Kuo2Init,
                dim1198Kuo2Init,
                dim1199Kuo2Init,
                dim1200Kuo2Init,
                dim1201Kuo2Init,
                dim1202Kuo2Init,
                dim1203Kuo2Init,
                dim1204Kuo2Init,
                dim1205Kuo2Init,
                dim1206Kuo2Init,
                dim1207Kuo2Init,
                dim1208Kuo2Init,
                dim1209Kuo2Init,
                dim1210Kuo2Init,
                dim1211Kuo2Init,
                dim1212Kuo2Init,
                dim1213Kuo2Init,
                dim1214Kuo2Init,
                dim1215Kuo2Init,
                dim1216Kuo2Init,
                dim1217Kuo2Init,
                dim1218Kuo2Init,
                dim1219Kuo2Init,
                dim1220Kuo2Init,
                dim1221Kuo2Init,
                dim1222Kuo2Init,
                dim1223Kuo2Init,
                dim1224Kuo2Init,
                dim1225Kuo2Init,
                dim1226Kuo2Init,
                dim1227Kuo2Init,
                dim1228Kuo2Init,
                dim1229Kuo2Init,
                dim1230Kuo2Init,
                dim1231Kuo2Init,
                dim1232Kuo2Init,
                dim1233Kuo2Init,
                dim1234Kuo2Init,
                dim1235Kuo2Init,
                dim1236Kuo2Init,
                dim1237Kuo2Init,
                dim1238Kuo2Init,
                dim1239Kuo2Init,
                dim1240Kuo2Init,
                dim1241Kuo2Init,
                dim1242Kuo2Init,
                dim1243Kuo2Init,
                dim1244Kuo2Init,
                dim1245Kuo2Init,
                dim1246Kuo2Init,
                dim1247Kuo2Init,
                dim1248Kuo2Init,
                dim1249Kuo2Init,
                dim1250Kuo2Init,
                dim1251Kuo2Init,
                dim1252Kuo2Init,
                dim1253Kuo2Init,
                dim1254Kuo2Init,
                dim1255Kuo2Init,
                dim1256Kuo2Init,
                dim1257Kuo2Init,
                dim1258Kuo2Init,
                dim1259Kuo2Init,
                dim1260Kuo2Init,
                dim1261Kuo2Init,
                dim1262Kuo2Init,
                dim1263Kuo2Init,
                dim1264Kuo2Init,
                dim1265Kuo2Init,
                dim1266Kuo2Init,
                dim1267Kuo2Init,
                dim1268Kuo2Init,
                dim1269Kuo2Init,
                dim1270Kuo2Init,
                dim1271Kuo2Init,
                dim1272Kuo2Init,
                dim1273Kuo2Init,
                dim1274Kuo2Init,
                dim1275Kuo2Init,
                dim1276Kuo2Init,
                dim1277Kuo2Init,
                dim1278Kuo2Init,
                dim1279Kuo2Init,
                dim1280Kuo2Init,
                dim1281Kuo2Init,
                dim1282Kuo2Init,
                dim1283Kuo2Init,
                dim1284Kuo2Init,
                dim1285Kuo2Init,
                dim1286Kuo2Init,
                dim1287Kuo2Init,
                dim1288Kuo2Init,
                dim1289Kuo2Init,
                dim1290Kuo2Init,
                dim1291Kuo2Init,
                dim1292Kuo2Init,
                dim1293Kuo2Init,
                dim1294Kuo2Init,
                dim1295Kuo2Init,
                dim1296Kuo2Init,
                dim1297Kuo2Init,
                dim1298Kuo2Init,
                dim1299Kuo2Init,
                dim1300Kuo2Init,
                dim1301Kuo2Init,
                dim1302Kuo2Init,
                dim1303Kuo2Init,
                dim1304Kuo2Init,
                dim1305Kuo2Init,
                dim1306Kuo2Init,
                dim1307Kuo2Init,
                dim1308Kuo2Init,
                dim1309Kuo2Init,
                dim1310Kuo2Init,
                dim1311Kuo2Init,
                dim1312Kuo2Init,
                dim1313Kuo2Init,
                dim1314Kuo2Init,
                dim1315Kuo2Init,
                dim1316Kuo2Init,
                dim1317Kuo2Init,
                dim1318Kuo2Init,
                dim1319Kuo2Init,
                dim1320Kuo2Init,
                dim1321Kuo2Init,
                dim1322Kuo2Init,
                dim1323Kuo2Init,
                dim1324Kuo2Init,
                dim1325Kuo2Init,
                dim1326Kuo2Init,
                dim1327Kuo2Init,
                dim1328Kuo2Init,
                dim1329Kuo2Init,
                dim1330Kuo2Init,
                dim1331Kuo2Init,
                dim1332Kuo2Init,
                dim1333Kuo2Init,
                dim1334Kuo2Init,
                dim1335Kuo2Init,
                dim1336Kuo2Init,
                dim1337Kuo2Init,
                dim1338Kuo2Init,
                dim1339Kuo2Init,
                dim1340Kuo2Init,
                dim1341Kuo2Init,
                dim1342Kuo2Init,
                dim1343Kuo2Init,
                dim1344Kuo2Init,
                dim1345Kuo2Init,
                dim1346Kuo2Init,
                dim1347Kuo2Init,
                dim1348Kuo2Init,
                dim1349Kuo2Init,
                dim1350Kuo2Init,
                dim1351Kuo2Init,
                dim1352Kuo2Init,
                dim1353Kuo2Init,
                dim1354Kuo2Init,
                dim1355Kuo2Init,
                dim1356Kuo2Init,
                dim1357Kuo2Init,
                dim1358Kuo2Init,
                dim1359Kuo2Init,
                dim1360Kuo2Init,
                dim1361Kuo2Init,
                dim1362Kuo2Init,
                dim1363Kuo2Init,
                dim1364Kuo2Init,
                dim1365Kuo2Init,
                dim1366Kuo2Init,
                dim1367Kuo2Init,
                dim1368Kuo2Init,
                dim1369Kuo2Init,
                dim1370Kuo2Init,
                dim1371Kuo2Init,
                dim1372Kuo2Init,
                dim1373Kuo2Init,
                dim1374Kuo2Init,
                dim1375Kuo2Init,
                dim1376Kuo2Init,
                dim1377Kuo2Init,
                dim1378Kuo2Init,
                dim1379Kuo2Init,
                dim1380Kuo2Init,
                dim1381Kuo2Init,
                dim1382Kuo2Init,
                dim1383Kuo2Init,
                dim1384Kuo2Init,
                dim1385Kuo2Init,
                dim1386Kuo2Init,
                dim1387Kuo2Init,
                dim1388Kuo2Init,
                dim1389Kuo2Init,
                dim1390Kuo2Init,
                dim1391Kuo2Init,
                dim1392Kuo2Init,
                dim1393Kuo2Init,
                dim1394Kuo2Init,
                dim1395Kuo2Init,
                dim1396Kuo2Init,
                dim1397Kuo2Init,
                dim1398Kuo2Init,
                dim1399Kuo2Init,
                dim1400Kuo2Init,
                dim1401Kuo2Init,
                dim1402Kuo2Init,
                dim1403Kuo2Init,
                dim1404Kuo2Init,
                dim1405Kuo2Init,
                dim1406Kuo2Init,
                dim1407Kuo2Init,
                dim1408Kuo2Init,
                dim1409Kuo2Init,
                dim1410Kuo2Init,
                dim1411Kuo2Init,
                dim1412Kuo2Init,
                dim1413Kuo2Init,
                dim1414Kuo2Init,
                dim1415Kuo2Init,
                dim1416Kuo2Init,
                dim1417Kuo2Init,
                dim1418Kuo2Init,
                dim1419Kuo2Init,
                dim1420Kuo2Init,
                dim1421Kuo2Init,
                dim1422Kuo2Init,
                dim1423Kuo2Init,
                dim1424Kuo2Init,
                dim1425Kuo2Init,
                dim1426Kuo2Init,
                dim1427Kuo2Init,
                dim1428Kuo2Init,
                dim1429Kuo2Init,
                dim1430Kuo2Init,
                dim1431Kuo2Init,
                dim1432Kuo2Init,
                dim1433Kuo2Init,
                dim1434Kuo2Init,
                dim1435Kuo2Init,
                dim1436Kuo2Init,
                dim1437Kuo2Init,
                dim1438Kuo2Init,
                dim1439Kuo2Init,
                dim1440Kuo2Init,
                dim1441Kuo2Init,
                dim1442Kuo2Init,
                dim1443Kuo2Init,
                dim1444Kuo2Init,
                dim1445Kuo2Init,
                dim1446Kuo2Init,
                dim1447Kuo2Init,
                dim1448Kuo2Init,
                dim1449Kuo2Init,
                dim1450Kuo2Init,
                dim1451Kuo2Init,
                dim1452Kuo2Init,
                dim1453Kuo2Init,
                dim1454Kuo2Init,
                dim1455Kuo2Init,
                dim1456Kuo2Init,
                dim1457Kuo2Init,
                dim1458Kuo2Init,
                dim1459Kuo2Init,
                dim1460Kuo2Init,
                dim1461Kuo2Init,
                dim1462Kuo2Init,
                dim1463Kuo2Init,
                dim1464Kuo2Init,
                dim1465Kuo2Init,
                dim1466Kuo2Init,
                dim1467Kuo2Init,
                dim1468Kuo2Init,
                dim1469Kuo2Init,
                dim1470Kuo2Init,
                dim1471Kuo2Init,
                dim1472Kuo2Init,
                dim1473Kuo2Init,
                dim1474Kuo2Init,
                dim1475Kuo2Init,
                dim1476Kuo2Init,
                dim1477Kuo2Init,
                dim1478Kuo2Init,
                dim1479Kuo2Init,
                dim1480Kuo2Init,
                dim1481Kuo2Init,
                dim1482Kuo2Init,
                dim1483Kuo2Init,
                dim1484Kuo2Init,
                dim1485Kuo2Init,
                dim1486Kuo2Init,
                dim1487Kuo2Init,
                dim1488Kuo2Init,
                dim1489Kuo2Init,
                dim1490Kuo2Init,
                dim1491Kuo2Init,
                dim1492Kuo2Init,
                dim1493Kuo2Init,
                dim1494Kuo2Init,
                dim1495Kuo2Init,
                dim1496Kuo2Init,
                dim1497Kuo2Init,
                dim1498Kuo2Init,
                dim1499Kuo2Init,
                dim1500Kuo2Init,
                dim1501Kuo2Init,
                dim1502Kuo2Init,
                dim1503Kuo2Init,
                dim1504Kuo2Init,
                dim1505Kuo2Init,
                dim1506Kuo2Init,
                dim1507Kuo2Init,
                dim1508Kuo2Init,
                dim1509Kuo2Init,
                dim1510Kuo2Init,
                dim1511Kuo2Init,
                dim1512Kuo2Init,
                dim1513Kuo2Init,
                dim1514Kuo2Init,
                dim1515Kuo2Init,
                dim1516Kuo2Init,
                dim1517Kuo2Init,
                dim1518Kuo2Init,
                dim1519Kuo2Init,
                dim1520Kuo2Init,
                dim1521Kuo2Init,
                dim1522Kuo2Init,
                dim1523Kuo2Init,
                dim1524Kuo2Init,
                dim1525Kuo2Init,
                dim1526Kuo2Init,
                dim1527Kuo2Init,
                dim1528Kuo2Init,
                dim1529Kuo2Init,
                dim1530Kuo2Init,
                dim1531Kuo2Init,
                dim1532Kuo2Init,
                dim1533Kuo2Init,
                dim1534Kuo2Init,
                dim1535Kuo2Init,
                dim1536Kuo2Init,
                dim1537Kuo2Init,
                dim1538Kuo2Init,
                dim1539Kuo2Init,
                dim1540Kuo2Init,
                dim1541Kuo2Init,
                dim1542Kuo2Init,
                dim1543Kuo2Init,
                dim1544Kuo2Init,
                dim1545Kuo2Init,
                dim1546Kuo2Init,
                dim1547Kuo2Init,
                dim1548Kuo2Init,
                dim1549Kuo2Init,
                dim1550Kuo2Init,
                dim1551Kuo2Init,
                dim1552Kuo2Init,
                dim1553Kuo2Init,
                dim1554Kuo2Init,
                dim1555Kuo2Init,
                dim1556Kuo2Init,
                dim1557Kuo2Init,
                dim1558Kuo2Init,
                dim1559Kuo2Init,
                dim1560Kuo2Init,
                dim1561Kuo2Init,
                dim1562Kuo2Init,
                dim1563Kuo2Init,
                dim1564Kuo2Init,
                dim1565Kuo2Init,
                dim1566Kuo2Init,
                dim1567Kuo2Init,
                dim1568Kuo2Init,
                dim1569Kuo2Init,
                dim1570Kuo2Init,
                dim1571Kuo2Init,
                dim1572Kuo2Init,
                dim1573Kuo2Init,
                dim1574Kuo2Init,
                dim1575Kuo2Init,
                dim1576Kuo2Init,
                dim1577Kuo2Init,
                dim1578Kuo2Init,
                dim1579Kuo2Init,
                dim1580Kuo2Init,
                dim1581Kuo2Init,
                dim1582Kuo2Init,
                dim1583Kuo2Init,
                dim1584Kuo2Init,
                dim1585Kuo2Init,
                dim1586Kuo2Init,
                dim1587Kuo2Init,
                dim1588Kuo2Init,
                dim1589Kuo2Init,
                dim1590Kuo2Init,
                dim1591Kuo2Init,
                dim1592Kuo2Init,
                dim1593Kuo2Init,
                dim1594Kuo2Init,
                dim1595Kuo2Init,
                dim1596Kuo2Init,
                dim1597Kuo2Init,
                dim1598Kuo2Init,
                dim1599Kuo2Init,
                dim1600Kuo2Init,
                dim1601Kuo2Init,
                dim1602Kuo2Init,
                dim1603Kuo2Init,
                dim1604Kuo2Init,
                dim1605Kuo2Init,
                dim1606Kuo2Init,
                dim1607Kuo2Init,
                dim1608Kuo2Init,
                dim1609Kuo2Init,
                dim1610Kuo2Init,
                dim1611Kuo2Init,
                dim1612Kuo2Init,
                dim1613Kuo2Init,
                dim1614Kuo2Init,
                dim1615Kuo2Init,
                dim1616Kuo2Init,
                dim1617Kuo2Init,
                dim1618Kuo2Init,
                dim1619Kuo2Init,
                dim1620Kuo2Init,
                dim1621Kuo2Init,
                dim1622Kuo2Init,
                dim1623Kuo2Init,
                dim1624Kuo2Init,
                dim1625Kuo2Init,
                dim1626Kuo2Init,
                dim1627Kuo2Init,
                dim1628Kuo2Init,
                dim1629Kuo2Init,
                dim1630Kuo2Init,
                dim1631Kuo2Init,
                dim1632Kuo2Init,
                dim1633Kuo2Init,
                dim1634Kuo2Init,
                dim1635Kuo2Init,
                dim1636Kuo2Init,
                dim1637Kuo2Init,
                dim1638Kuo2Init,
                dim1639Kuo2Init,
                dim1640Kuo2Init,
                dim1641Kuo2Init,
                dim1642Kuo2Init,
                dim1643Kuo2Init,
                dim1644Kuo2Init,
                dim1645Kuo2Init,
                dim1646Kuo2Init,
                dim1647Kuo2Init,
                dim1648Kuo2Init,
                dim1649Kuo2Init,
                dim1650Kuo2Init,
                dim1651Kuo2Init,
                dim1652Kuo2Init,
                dim1653Kuo2Init,
                dim1654Kuo2Init,
                dim1655Kuo2Init,
                dim1656Kuo2Init,
                dim1657Kuo2Init,
                dim1658Kuo2Init,
                dim1659Kuo2Init,
                dim1660Kuo2Init,
                dim1661Kuo2Init,
                dim1662Kuo2Init,
                dim1663Kuo2Init,
                dim1664Kuo2Init,
                dim1665Kuo2Init,
                dim1666Kuo2Init,
                dim1667Kuo2Init,
                dim1668Kuo2Init,
                dim1669Kuo2Init,
                dim1670Kuo2Init,
                dim1671Kuo2Init,
                dim1672Kuo2Init,
                dim1673Kuo2Init,
                dim1674Kuo2Init,
                dim1675Kuo2Init,
                dim1676Kuo2Init,
                dim1677Kuo2Init,
                dim1678Kuo2Init,
                dim1679Kuo2Init,
                dim1680Kuo2Init,
                dim1681Kuo2Init,
                dim1682Kuo2Init,
                dim1683Kuo2Init,
                dim1684Kuo2Init,
                dim1685Kuo2Init,
                dim1686Kuo2Init,
                dim1687Kuo2Init,
                dim1688Kuo2Init,
                dim1689Kuo2Init,
                dim1690Kuo2Init,
                dim1691Kuo2Init,
                dim1692Kuo2Init,
                dim1693Kuo2Init,
                dim1694Kuo2Init,
                dim1695Kuo2Init,
                dim1696Kuo2Init,
                dim1697Kuo2Init,
                dim1698Kuo2Init,
                dim1699Kuo2Init,
                dim1700Kuo2Init,
                dim1701Kuo2Init,
                dim1702Kuo2Init,
                dim1703Kuo2Init,
                dim1704Kuo2Init,
                dim1705Kuo2Init,
                dim1706Kuo2Init,
                dim1707Kuo2Init,
                dim1708Kuo2Init,
                dim1709Kuo2Init,
                dim1710Kuo2Init,
                dim1711Kuo2Init,
                dim1712Kuo2Init,
                dim1713Kuo2Init,
                dim1714Kuo2Init,
                dim1715Kuo2Init,
                dim1716Kuo2Init,
                dim1717Kuo2Init,
                dim1718Kuo2Init,
                dim1719Kuo2Init,
                dim1720Kuo2Init,
                dim1721Kuo2Init,
                dim1722Kuo2Init,
                dim1723Kuo2Init,
                dim1724Kuo2Init,
                dim1725Kuo2Init,
                dim1726Kuo2Init,
                dim1727Kuo2Init,
                dim1728Kuo2Init,
                dim1729Kuo2Init,
                dim1730Kuo2Init,
                dim1731Kuo2Init,
                dim1732Kuo2Init,
                dim1733Kuo2Init,
                dim1734Kuo2Init,
                dim1735Kuo2Init,
                dim1736Kuo2Init,
                dim1737Kuo2Init,
                dim1738Kuo2Init,
                dim1739Kuo2Init,
                dim1740Kuo2Init,
                dim1741Kuo2Init,
                dim1742Kuo2Init,
                dim1743Kuo2Init,
                dim1744Kuo2Init,
                dim1745Kuo2Init,
                dim1746Kuo2Init,
                dim1747Kuo2Init,
                dim1748Kuo2Init,
                dim1749Kuo2Init,
                dim1750Kuo2Init,
                dim1751Kuo2Init,
                dim1752Kuo2Init,
                dim1753Kuo2Init,
                dim1754Kuo2Init,
                dim1755Kuo2Init,
                dim1756Kuo2Init,
                dim1757Kuo2Init,
                dim1758Kuo2Init,
                dim1759Kuo2Init,
                dim1760Kuo2Init,
                dim1761Kuo2Init,
                dim1762Kuo2Init,
                dim1763Kuo2Init,
                dim1764Kuo2Init,
                dim1765Kuo2Init,
                dim1766Kuo2Init,
                dim1767Kuo2Init,
                dim1768Kuo2Init,
                dim1769Kuo2Init,
                dim1770Kuo2Init,
                dim1771Kuo2Init,
                dim1772Kuo2Init,
                dim1773Kuo2Init,
                dim1774Kuo2Init,
                dim1775Kuo2Init,
                dim1776Kuo2Init,
                dim1777Kuo2Init,
                dim1778Kuo2Init,
                dim1779Kuo2Init,
                dim1780Kuo2Init,
                dim1781Kuo2Init,
                dim1782Kuo2Init,
                dim1783Kuo2Init,
                dim1784Kuo2Init,
                dim1785Kuo2Init,
                dim1786Kuo2Init,
                dim1787Kuo2Init,
                dim1788Kuo2Init,
                dim1789Kuo2Init,
                dim1790Kuo2Init,
                dim1791Kuo2Init,
                dim1792Kuo2Init,
                dim1793Kuo2Init,
                dim1794Kuo2Init,
                dim1795Kuo2Init,
                dim1796Kuo2Init,
                dim1797Kuo2Init,
                dim1798Kuo2Init,
                dim1799Kuo2Init,
                dim1800Kuo2Init,
                dim1801Kuo2Init,
                dim1802Kuo2Init,
                dim1803Kuo2Init,
                dim1804Kuo2Init,
                dim1805Kuo2Init,
                dim1806Kuo2Init,
                dim1807Kuo2Init,
                dim1808Kuo2Init,
                dim1809Kuo2Init,
                dim1810Kuo2Init,
                dim1811Kuo2Init,
                dim1812Kuo2Init,
                dim1813Kuo2Init,
                dim1814Kuo2Init,
                dim1815Kuo2Init,
                dim1816Kuo2Init,
                dim1817Kuo2Init,
                dim1818Kuo2Init,
                dim1819Kuo2Init,
                dim1820Kuo2Init,
                dim1821Kuo2Init,
                dim1822Kuo2Init,
                dim1823Kuo2Init,
                dim1824Kuo2Init,
                dim1825Kuo2Init,
                dim1826Kuo2Init,
                dim1827Kuo2Init,
                dim1828Kuo2Init,
                dim1829Kuo2Init,
                dim1830Kuo2Init,
                dim1831Kuo2Init,
                dim1832Kuo2Init,
                dim1833Kuo2Init,
                dim1834Kuo2Init,
                dim1835Kuo2Init,
                dim1836Kuo2Init,
                dim1837Kuo2Init,
                dim1838Kuo2Init,
                dim1839Kuo2Init,
                dim1840Kuo2Init,
                dim1841Kuo2Init,
                dim1842Kuo2Init,
                dim1843Kuo2Init,
                dim1844Kuo2Init,
                dim1845Kuo2Init,
                dim1846Kuo2Init,
                dim1847Kuo2Init,
                dim1848Kuo2Init,
                dim1849Kuo2Init,
                dim1850Kuo2Init,
                dim1851Kuo2Init,
                dim1852Kuo2Init,
                dim1853Kuo2Init,
                dim1854Kuo2Init,
                dim1855Kuo2Init,
                dim1856Kuo2Init,
                dim1857Kuo2Init,
                dim1858Kuo2Init,
                dim1859Kuo2Init,
                dim1860Kuo2Init,
                dim1861Kuo2Init,
                dim1862Kuo2Init,
                dim1863Kuo2Init,
                dim1864Kuo2Init,
                dim1865Kuo2Init,
                dim1866Kuo2Init,
                dim1867Kuo2Init,
                dim1868Kuo2Init,
                dim1869Kuo2Init,
                dim1870Kuo2Init,
                dim1871Kuo2Init,
                dim1872Kuo2Init,
                dim1873Kuo2Init,
                dim1874Kuo2Init,
                dim1875Kuo2Init,
                dim1876Kuo2Init,
                dim1877Kuo2Init,
                dim1878Kuo2Init,
                dim1879Kuo2Init,
                dim1880Kuo2Init,
                dim1881Kuo2Init,
                dim1882Kuo2Init,
                dim1883Kuo2Init,
                dim1884Kuo2Init,
                dim1885Kuo2Init,
                dim1886Kuo2Init,
                dim1887Kuo2Init,
                dim1888Kuo2Init,
                dim1889Kuo2Init,
                dim1890Kuo2Init,
                dim1891Kuo2Init,
                dim1892Kuo2Init,
                dim1893Kuo2Init,
                dim1894Kuo2Init,
                dim1895Kuo2Init,
                dim1896Kuo2Init,
                dim1897Kuo2Init,
                dim1898Kuo2Init,
                dim1899Kuo2Init,
                dim1900Kuo2Init,
                dim1901Kuo2Init,
                dim1902Kuo2Init,
                dim1903Kuo2Init,
                dim1904Kuo2Init,
                dim1905Kuo2Init,
                dim1906Kuo2Init,
                dim1907Kuo2Init,
                dim1908Kuo2Init,
                dim1909Kuo2Init,
                dim1910Kuo2Init,
                dim1911Kuo2Init,
                dim1912Kuo2Init,
                dim1913Kuo2Init,
                dim1914Kuo2Init,
                dim1915Kuo2Init,
                dim1916Kuo2Init,
                dim1917Kuo2Init,
                dim1918Kuo2Init,
                dim1919Kuo2Init,
                dim1920Kuo2Init,
                dim1921Kuo2Init,
                dim1922Kuo2Init,
                dim1923Kuo2Init,
                dim1924Kuo2Init,
                dim1925Kuo2Init,
                dim1926Kuo2Init,
                dim1927Kuo2Init,
                dim1928Kuo2Init,
                dim1929Kuo2Init,
                dim1930Kuo2Init,
                dim1931Kuo2Init,
                dim1932Kuo2Init,
                dim1933Kuo2Init,
                dim1934Kuo2Init,
                dim1935Kuo2Init,
                dim1936Kuo2Init,
                dim1937Kuo2Init,
                dim1938Kuo2Init,
                dim1939Kuo2Init,
                dim1940Kuo2Init,
                dim1941Kuo2Init,
                dim1942Kuo2Init,
                dim1943Kuo2Init,
                dim1944Kuo2Init,
                dim1945Kuo2Init,
                dim1946Kuo2Init,
                dim1947Kuo2Init,
                dim1948Kuo2Init,
                dim1949Kuo2Init,
                dim1950Kuo2Init,
                dim1951Kuo2Init,
                dim1952Kuo2Init,
                dim1953Kuo2Init,
                dim1954Kuo2Init,
                dim1955Kuo2Init,
                dim1956Kuo2Init,
                dim1957Kuo2Init,
                dim1958Kuo2Init,
                dim1959Kuo2Init,
                dim1960Kuo2Init,
                dim1961Kuo2Init,
                dim1962Kuo2Init,
                dim1963Kuo2Init,
                dim1964Kuo2Init,
                dim1965Kuo2Init,
                dim1966Kuo2Init,
                dim1967Kuo2Init,
                dim1968Kuo2Init,
                dim1969Kuo2Init,
                dim1970Kuo2Init,
                dim1971Kuo2Init,
                dim1972Kuo2Init,
                dim1973Kuo2Init,
                dim1974Kuo2Init,
                dim1975Kuo2Init,
                dim1976Kuo2Init,
                dim1977Kuo2Init,
                dim1978Kuo2Init,
                dim1979Kuo2Init,
                dim1980Kuo2Init,
                dim1981Kuo2Init,
                dim1982Kuo2Init,
                dim1983Kuo2Init,
                dim1984Kuo2Init,
                dim1985Kuo2Init,
                dim1986Kuo2Init,
                dim1987Kuo2Init,
                dim1988Kuo2Init,
                dim1989Kuo2Init,
                dim1990Kuo2Init,
                dim1991Kuo2Init,
                dim1992Kuo2Init,
                dim1993Kuo2Init,
                dim1994Kuo2Init,
                dim1995Kuo2Init,
                dim1996Kuo2Init,
                dim1997Kuo2Init,
                dim1998Kuo2Init,
                dim1999Kuo2Init,
                dim2000Kuo2Init,
                dim2001Kuo2Init,
                dim2002Kuo2Init,
                dim2003Kuo2Init,
                dim2004Kuo2Init,
                dim2005Kuo2Init,
                dim2006Kuo2Init,
                dim2007Kuo2Init,
                dim2008Kuo2Init,
                dim2009Kuo2Init,
                dim2010Kuo2Init,
                dim2011Kuo2Init,
                dim2012Kuo2Init,
                dim2013Kuo2Init,
                dim2014Kuo2Init,
                dim2015Kuo2Init,
                dim2016Kuo2Init,
                dim2017Kuo2Init,
                dim2018Kuo2Init,
                dim2019Kuo2Init,
                dim2020Kuo2Init,
                dim2021Kuo2Init,
                dim2022Kuo2Init,
                dim2023Kuo2Init,
                dim2024Kuo2Init,
                dim2025Kuo2Init,
                dim2026Kuo2Init,
                dim2027Kuo2Init,
                dim2028Kuo2Init,
                dim2029Kuo2Init,
                dim2030Kuo2Init,
                dim2031Kuo2Init,
                dim2032Kuo2Init,
                dim2033Kuo2Init,
                dim2034Kuo2Init,
                dim2035Kuo2Init,
                dim2036Kuo2Init,
                dim2037Kuo2Init,
                dim2038Kuo2Init,
                dim2039Kuo2Init,
                dim2040Kuo2Init,
                dim2041Kuo2Init,
                dim2042Kuo2Init,
                dim2043Kuo2Init,
                dim2044Kuo2Init,
                dim2045Kuo2Init,
                dim2046Kuo2Init,
                dim2047Kuo2Init,
                dim2048Kuo2Init,
                dim2049Kuo2Init,
                dim2050Kuo2Init,
                dim2051Kuo2Init,
                dim2052Kuo2Init,
                dim2053Kuo2Init,
                dim2054Kuo2Init,
                dim2055Kuo2Init,
                dim2056Kuo2Init,
                dim2057Kuo2Init,
                dim2058Kuo2Init,
                dim2059Kuo2Init,
                dim2060Kuo2Init,
                dim2061Kuo2Init,
                dim2062Kuo2Init,
                dim2063Kuo2Init,
                dim2064Kuo2Init,
                dim2065Kuo2Init,
                dim2066Kuo2Init,
                dim2067Kuo2Init,
                dim2068Kuo2Init,
                dim2069Kuo2Init,
                dim2070Kuo2Init,
                dim2071Kuo2Init,
                dim2072Kuo2Init,
                dim2073Kuo2Init,
                dim2074Kuo2Init,
                dim2075Kuo2Init,
                dim2076Kuo2Init,
                dim2077Kuo2Init,
                dim2078Kuo2Init,
                dim2079Kuo2Init,
                dim2080Kuo2Init,
                dim2081Kuo2Init,
                dim2082Kuo2Init,
                dim2083Kuo2Init,
                dim2084Kuo2Init,
                dim2085Kuo2Init,
                dim2086Kuo2Init,
                dim2087Kuo2Init,
                dim2088Kuo2Init,
                dim2089Kuo2Init,
                dim2090Kuo2Init,
                dim2091Kuo2Init,
                dim2092Kuo2Init,
                dim2093Kuo2Init,
                dim2094Kuo2Init,
                dim2095Kuo2Init,
                dim2096Kuo2Init,
                dim2097Kuo2Init,
                dim2098Kuo2Init,
                dim2099Kuo2Init,
                dim2100Kuo2Init,
                dim2101Kuo2Init,
                dim2102Kuo2Init,
                dim2103Kuo2Init,
                dim2104Kuo2Init,
                dim2105Kuo2Init,
                dim2106Kuo2Init,
                dim2107Kuo2Init,
                dim2108Kuo2Init,
                dim2109Kuo2Init,
                dim2110Kuo2Init,
                dim2111Kuo2Init,
                dim2112Kuo2Init,
                dim2113Kuo2Init,
                dim2114Kuo2Init,
                dim2115Kuo2Init,
                dim2116Kuo2Init,
                dim2117Kuo2Init,
                dim2118Kuo2Init,
                dim2119Kuo2Init,
                dim2120Kuo2Init,
                dim2121Kuo2Init,
                dim2122Kuo2Init,
                dim2123Kuo2Init,
                dim2124Kuo2Init,
                dim2125Kuo2Init,
                dim2126Kuo2Init,
                dim2127Kuo2Init,
                dim2128Kuo2Init,
                dim2129Kuo2Init,
                dim2130Kuo2Init,
                dim2131Kuo2Init,
                dim2132Kuo2Init,
                dim2133Kuo2Init,
                dim2134Kuo2Init,
                dim2135Kuo2Init,
                dim2136Kuo2Init,
                dim2137Kuo2Init,
                dim2138Kuo2Init,
                dim2139Kuo2Init,
                dim2140Kuo2Init,
                dim2141Kuo2Init,
                dim2142Kuo2Init,
                dim2143Kuo2Init,
                dim2144Kuo2Init,
                dim2145Kuo2Init,
                dim2146Kuo2Init,
                dim2147Kuo2Init,
                dim2148Kuo2Init,
                dim2149Kuo2Init,
                dim2150Kuo2Init,
                dim2151Kuo2Init,
                dim2152Kuo2Init,
                dim2153Kuo2Init,
                dim2154Kuo2Init,
                dim2155Kuo2Init,
                dim2156Kuo2Init,
                dim2157Kuo2Init,
                dim2158Kuo2Init,
                dim2159Kuo2Init,
                dim2160Kuo2Init,
                dim2161Kuo2Init,
                dim2162Kuo2Init,
                dim2163Kuo2Init,
                dim2164Kuo2Init,
                dim2165Kuo2Init,
                dim2166Kuo2Init,
                dim2167Kuo2Init,
                dim2168Kuo2Init,
                dim2169Kuo2Init,
                dim2170Kuo2Init,
                dim2171Kuo2Init,
                dim2172Kuo2Init,
                dim2173Kuo2Init,
                dim2174Kuo2Init,
                dim2175Kuo2Init,
                dim2176Kuo2Init,
                dim2177Kuo2Init,
                dim2178Kuo2Init,
                dim2179Kuo2Init,
                dim2180Kuo2Init,
                dim2181Kuo2Init,
                dim2182Kuo2Init,
                dim2183Kuo2Init,
                dim2184Kuo2Init,
                dim2185Kuo2Init,
                dim2186Kuo2Init,
                dim2187Kuo2Init,
                dim2188Kuo2Init,
                dim2189Kuo2Init,
                dim2190Kuo2Init,
                dim2191Kuo2Init,
                dim2192Kuo2Init,
                dim2193Kuo2Init,
                dim2194Kuo2Init,
                dim2195Kuo2Init,
                dim2196Kuo2Init,
                dim2197Kuo2Init,
                dim2198Kuo2Init,
                dim2199Kuo2Init,
                dim2200Kuo2Init,
                dim2201Kuo2Init,
                dim2202Kuo2Init,
                dim2203Kuo2Init,
                dim2204Kuo2Init,
                dim2205Kuo2Init,
                dim2206Kuo2Init,
                dim2207Kuo2Init,
                dim2208Kuo2Init,
                dim2209Kuo2Init,
                dim2210Kuo2Init,
                dim2211Kuo2Init,
                dim2212Kuo2Init,
                dim2213Kuo2Init,
                dim2214Kuo2Init,
                dim2215Kuo2Init,
                dim2216Kuo2Init,
                dim2217Kuo2Init,
                dim2218Kuo2Init,
                dim2219Kuo2Init,
                dim2220Kuo2Init,
                dim2221Kuo2Init,
                dim2222Kuo2Init,
                dim2223Kuo2Init,
                dim2224Kuo2Init,
                dim2225Kuo2Init,
                dim2226Kuo2Init,
                dim2227Kuo2Init,
                dim2228Kuo2Init,
                dim2229Kuo2Init,
                dim2230Kuo2Init,
                dim2231Kuo2Init,
                dim2232Kuo2Init,
                dim2233Kuo2Init,
                dim2234Kuo2Init,
                dim2235Kuo2Init,
                dim2236Kuo2Init,
                dim2237Kuo2Init,
                dim2238Kuo2Init,
                dim2239Kuo2Init,
                dim2240Kuo2Init,
                dim2241Kuo2Init,
                dim2242Kuo2Init,
                dim2243Kuo2Init,
                dim2244Kuo2Init,
                dim2245Kuo2Init,
                dim2246Kuo2Init,
                dim2247Kuo2Init,
                dim2248Kuo2Init,
                dim2249Kuo2Init,
                dim2250Kuo2Init,
                dim2251Kuo2Init,
                dim2252Kuo2Init,
                dim2253Kuo2Init,
                dim2254Kuo2Init,
                dim2255Kuo2Init,
                dim2256Kuo2Init,
                dim2257Kuo2Init,
                dim2258Kuo2Init,
                dim2259Kuo2Init,
                dim2260Kuo2Init,
                dim2261Kuo2Init,
                dim2262Kuo2Init,
                dim2263Kuo2Init,
                dim2264Kuo2Init,
                dim2265Kuo2Init,
                dim2266Kuo2Init,
                dim2267Kuo2Init,
                dim2268Kuo2Init,
                dim2269Kuo2Init,
                dim2270Kuo2Init,
                dim2271Kuo2Init,
                dim2272Kuo2Init,
                dim2273Kuo2Init,
                dim2274Kuo2Init,
                dim2275Kuo2Init,
                dim2276Kuo2Init,
                dim2277Kuo2Init,
                dim2278Kuo2Init,
                dim2279Kuo2Init,
                dim2280Kuo2Init,
                dim2281Kuo2Init,
                dim2282Kuo2Init,
                dim2283Kuo2Init,
                dim2284Kuo2Init,
                dim2285Kuo2Init,
                dim2286Kuo2Init,
                dim2287Kuo2Init,
                dim2288Kuo2Init,
                dim2289Kuo2Init,
                dim2290Kuo2Init,
                dim2291Kuo2Init,
                dim2292Kuo2Init,
                dim2293Kuo2Init,
                dim2294Kuo2Init,
                dim2295Kuo2Init,
                dim2296Kuo2Init,
                dim2297Kuo2Init,
                dim2298Kuo2Init,
                dim2299Kuo2Init,
                dim2300Kuo2Init,
                dim2301Kuo2Init,
                dim2302Kuo2Init,
                dim2303Kuo2Init,
                dim2304Kuo2Init,
                dim2305Kuo2Init,
                dim2306Kuo2Init,
                dim2307Kuo2Init,
                dim2308Kuo2Init,
                dim2309Kuo2Init,
                dim2310Kuo2Init,
                dim2311Kuo2Init,
                dim2312Kuo2Init,
                dim2313Kuo2Init,
                dim2314Kuo2Init,
                dim2315Kuo2Init,
                dim2316Kuo2Init,
                dim2317Kuo2Init,
                dim2318Kuo2Init,
                dim2319Kuo2Init,
                dim2320Kuo2Init,
                dim2321Kuo2Init,
                dim2322Kuo2Init,
                dim2323Kuo2Init,
                dim2324Kuo2Init,
                dim2325Kuo2Init,
                dim2326Kuo2Init,
                dim2327Kuo2Init,
                dim2328Kuo2Init,
                dim2329Kuo2Init,
                dim2330Kuo2Init,
                dim2331Kuo2Init,
                dim2332Kuo2Init,
                dim2333Kuo2Init,
                dim2334Kuo2Init,
                dim2335Kuo2Init,
                dim2336Kuo2Init,
                dim2337Kuo2Init,
                dim2338Kuo2Init,
                dim2339Kuo2Init,
                dim2340Kuo2Init,
                dim2341Kuo2Init,
                dim2342Kuo2Init,
                dim2343Kuo2Init,
                dim2344Kuo2Init,
                dim2345Kuo2Init,
                dim2346Kuo2Init,
                dim2347Kuo2Init,
                dim2348Kuo2Init,
                dim2349Kuo2Init,
                dim2350Kuo2Init,
                dim2351Kuo2Init,
                dim2352Kuo2Init,
                dim2353Kuo2Init,
                dim2354Kuo2Init,
                dim2355Kuo2Init,
                dim2356Kuo2Init,
                dim2357Kuo2Init,
                dim2358Kuo2Init,
                dim2359Kuo2Init,
                dim2360Kuo2Init,
                dim2361Kuo2Init,
                dim2362Kuo2Init,
                dim2363Kuo2Init,
                dim2364Kuo2Init,
                dim2365Kuo2Init,
                dim2366Kuo2Init,
                dim2367Kuo2Init,
                dim2368Kuo2Init,
                dim2369Kuo2Init,
                dim2370Kuo2Init,
                dim2371Kuo2Init,
                dim2372Kuo2Init,
                dim2373Kuo2Init,
                dim2374Kuo2Init,
                dim2375Kuo2Init,
                dim2376Kuo2Init,
                dim2377Kuo2Init,
                dim2378Kuo2Init,
                dim2379Kuo2Init,
                dim2380Kuo2Init,
                dim2381Kuo2Init,
                dim2382Kuo2Init,
                dim2383Kuo2Init,
                dim2384Kuo2Init,
                dim2385Kuo2Init,
                dim2386Kuo2Init,
                dim2387Kuo2Init,
                dim2388Kuo2Init,
                dim2389Kuo2Init,
                dim2390Kuo2Init,
                dim2391Kuo2Init,
                dim2392Kuo2Init,
                dim2393Kuo2Init,
                dim2394Kuo2Init,
                dim2395Kuo2Init,
                dim2396Kuo2Init,
                dim2397Kuo2Init,
                dim2398Kuo2Init,
                dim2399Kuo2Init,
                dim2400Kuo2Init,
                dim2401Kuo2Init,
                dim2402Kuo2Init,
                dim2403Kuo2Init,
                dim2404Kuo2Init,
                dim2405Kuo2Init,
                dim2406Kuo2Init,
                dim2407Kuo2Init,
                dim2408Kuo2Init,
                dim2409Kuo2Init,
                dim2410Kuo2Init,
                dim2411Kuo2Init,
                dim2412Kuo2Init,
                dim2413Kuo2Init,
                dim2414Kuo2Init,
                dim2415Kuo2Init,
                dim2416Kuo2Init,
                dim2417Kuo2Init,
                dim2418Kuo2Init,
                dim2419Kuo2Init,
                dim2420Kuo2Init,
                dim2421Kuo2Init,
                dim2422Kuo2Init,
                dim2423Kuo2Init,
                dim2424Kuo2Init,
                dim2425Kuo2Init,
                dim2426Kuo2Init,
                dim2427Kuo2Init,
                dim2428Kuo2Init,
                dim2429Kuo2Init,
                dim2430Kuo2Init,
                dim2431Kuo2Init,
                dim2432Kuo2Init,
                dim2433Kuo2Init,
                dim2434Kuo2Init,
                dim2435Kuo2Init,
                dim2436Kuo2Init,
                dim2437Kuo2Init,
                dim2438Kuo2Init,
                dim2439Kuo2Init,
                dim2440Kuo2Init,
                dim2441Kuo2Init,
                dim2442Kuo2Init,
                dim2443Kuo2Init,
                dim2444Kuo2Init,
                dim2445Kuo2Init,
                dim2446Kuo2Init,
                dim2447Kuo2Init,
                dim2448Kuo2Init,
                dim2449Kuo2Init,
                dim2450Kuo2Init,
                dim2451Kuo2Init,
                dim2452Kuo2Init,
                dim2453Kuo2Init,
                dim2454Kuo2Init,
                dim2455Kuo2Init,
                dim2456Kuo2Init,
                dim2457Kuo2Init,
                dim2458Kuo2Init,
                dim2459Kuo2Init,
                dim2460Kuo2Init,
                dim2461Kuo2Init,
                dim2462Kuo2Init,
                dim2463Kuo2Init,
                dim2464Kuo2Init,
                dim2465Kuo2Init,
                dim2466Kuo2Init,
                dim2467Kuo2Init,
                dim2468Kuo2Init,
                dim2469Kuo2Init,
                dim2470Kuo2Init,
                dim2471Kuo2Init,
                dim2472Kuo2Init,
                dim2473Kuo2Init,
                dim2474Kuo2Init,
                dim2475Kuo2Init,
                dim2476Kuo2Init,
                dim2477Kuo2Init,
                dim2478Kuo2Init,
                dim2479Kuo2Init,
                dim2480Kuo2Init,
                dim2481Kuo2Init,
                dim2482Kuo2Init,
                dim2483Kuo2Init,
                dim2484Kuo2Init,
                dim2485Kuo2Init,
                dim2486Kuo2Init,
                dim2487Kuo2Init,
                dim2488Kuo2Init,
                dim2489Kuo2Init,
                dim2490Kuo2Init,
                dim2491Kuo2Init,
                dim2492Kuo2Init,
                dim2493Kuo2Init,
                dim2494Kuo2Init,
                dim2495Kuo2Init,
                dim2496Kuo2Init,
                dim2497Kuo2Init,
                dim2498Kuo2Init,
                dim2499Kuo2Init,
                dim2500Kuo2Init,
                dim2501Kuo2Init,
                dim2502Kuo2Init,
                dim2503Kuo2Init,
                dim2504Kuo2Init,
                dim2505Kuo2Init,
                dim2506Kuo2Init,
                dim2507Kuo2Init,
                dim2508Kuo2Init,
                dim2509Kuo2Init,
                dim2510Kuo2Init,
                dim2511Kuo2Init,
                dim2512Kuo2Init,
                dim2513Kuo2Init,
                dim2514Kuo2Init,
                dim2515Kuo2Init,
                dim2516Kuo2Init,
                dim2517Kuo2Init,
                dim2518Kuo2Init,
                dim2519Kuo2Init,
                dim2520Kuo2Init,
                dim2521Kuo2Init,
                dim2522Kuo2Init,
                dim2523Kuo2Init,
                dim2524Kuo2Init,
                dim2525Kuo2Init,
                dim2526Kuo2Init,
                dim2527Kuo2Init,
                dim2528Kuo2Init,
                dim2529Kuo2Init,
                dim2530Kuo2Init,
                dim2531Kuo2Init,
                dim2532Kuo2Init,
                dim2533Kuo2Init,
                dim2534Kuo2Init,
                dim2535Kuo2Init,
                dim2536Kuo2Init,
                dim2537Kuo2Init,
                dim2538Kuo2Init,
                dim2539Kuo2Init,
                dim2540Kuo2Init,
                dim2541Kuo2Init,
                dim2542Kuo2Init,
                dim2543Kuo2Init,
                dim2544Kuo2Init,
                dim2545Kuo2Init,
                dim2546Kuo2Init,
                dim2547Kuo2Init,
                dim2548Kuo2Init,
                dim2549Kuo2Init,
                dim2550Kuo2Init,
                dim2551Kuo2Init,
                dim2552Kuo2Init,
                dim2553Kuo2Init,
                dim2554Kuo2Init,
                dim2555Kuo2Init,
                dim2556Kuo2Init,
                dim2557Kuo2Init,
                dim2558Kuo2Init,
                dim2559Kuo2Init,
                dim2560Kuo2Init,
                dim2561Kuo2Init,
                dim2562Kuo2Init,
                dim2563Kuo2Init,
                dim2564Kuo2Init,
                dim2565Kuo2Init,
                dim2566Kuo2Init,
                dim2567Kuo2Init,
                dim2568Kuo2Init,
                dim2569Kuo2Init,
                dim2570Kuo2Init,
                dim2571Kuo2Init,
                dim2572Kuo2Init,
                dim2573Kuo2Init,
                dim2574Kuo2Init,
                dim2575Kuo2Init,
                dim2576Kuo2Init,
                dim2577Kuo2Init,
                dim2578Kuo2Init,
                dim2579Kuo2Init,
                dim2580Kuo2Init,
                dim2581Kuo2Init,
                dim2582Kuo2Init,
                dim2583Kuo2Init,
                dim2584Kuo2Init,
                dim2585Kuo2Init,
                dim2586Kuo2Init,
                dim2587Kuo2Init,
                dim2588Kuo2Init,
                dim2589Kuo2Init,
                dim2590Kuo2Init,
                dim2591Kuo2Init,
                dim2592Kuo2Init,
                dim2593Kuo2Init,
                dim2594Kuo2Init,
                dim2595Kuo2Init,
                dim2596Kuo2Init,
                dim2597Kuo2Init,
                dim2598Kuo2Init,
                dim2599Kuo2Init,
                dim2600Kuo2Init,
                dim2601Kuo2Init,
                dim2602Kuo2Init,
                dim2603Kuo2Init,
                dim2604Kuo2Init,
                dim2605Kuo2Init,
                dim2606Kuo2Init,
                dim2607Kuo2Init,
                dim2608Kuo2Init,
                dim2609Kuo2Init,
                dim2610Kuo2Init,
                dim2611Kuo2Init,
                dim2612Kuo2Init,
                dim2613Kuo2Init,
                dim2614Kuo2Init,
                dim2615Kuo2Init,
                dim2616Kuo2Init,
                dim2617Kuo2Init,
                dim2618Kuo2Init,
                dim2619Kuo2Init,
                dim2620Kuo2Init,
                dim2621Kuo2Init,
                dim2622Kuo2Init,
                dim2623Kuo2Init,
                dim2624Kuo2Init,
                dim2625Kuo2Init,
                dim2626Kuo2Init,
                dim2627Kuo2Init,
                dim2628Kuo2Init,
                dim2629Kuo2Init,
                dim2630Kuo2Init,
                dim2631Kuo2Init,
                dim2632Kuo2Init,
                dim2633Kuo2Init,
                dim2634Kuo2Init,
                dim2635Kuo2Init,
                dim2636Kuo2Init,
                dim2637Kuo2Init,
                dim2638Kuo2Init,
                dim2639Kuo2Init,
                dim2640Kuo2Init,
                dim2641Kuo2Init,
                dim2642Kuo2Init,
                dim2643Kuo2Init,
                dim2644Kuo2Init,
                dim2645Kuo2Init,
                dim2646Kuo2Init,
                dim2647Kuo2Init,
                dim2648Kuo2Init,
                dim2649Kuo2Init,
                dim2650Kuo2Init,
                dim2651Kuo2Init,
                dim2652Kuo2Init,
                dim2653Kuo2Init,
                dim2654Kuo2Init,
                dim2655Kuo2Init,
                dim2656Kuo2Init,
                dim2657Kuo2Init,
                dim2658Kuo2Init,
                dim2659Kuo2Init,
                dim2660Kuo2Init,
                dim2661Kuo2Init,
                dim2662Kuo2Init,
                dim2663Kuo2Init,
                dim2664Kuo2Init,
                dim2665Kuo2Init,
                dim2666Kuo2Init,
                dim2667Kuo2Init,
                dim2668Kuo2Init,
                dim2669Kuo2Init,
                dim2670Kuo2Init,
                dim2671Kuo2Init,
                dim2672Kuo2Init,
                dim2673Kuo2Init,
                dim2674Kuo2Init,
                dim2675Kuo2Init,
                dim2676Kuo2Init,
                dim2677Kuo2Init,
                dim2678Kuo2Init,
                dim2679Kuo2Init,
                dim2680Kuo2Init,
                dim2681Kuo2Init,
                dim2682Kuo2Init,
                dim2683Kuo2Init,
                dim2684Kuo2Init,
                dim2685Kuo2Init,
                dim2686Kuo2Init,
                dim2687Kuo2Init,
                dim2688Kuo2Init,
                dim2689Kuo2Init,
                dim2690Kuo2Init,
                dim2691Kuo2Init,
                dim2692Kuo2Init,
                dim2693Kuo2Init,
                dim2694Kuo2Init,
                dim2695Kuo2Init,
                dim2696Kuo2Init,
                dim2697Kuo2Init,
                dim2698Kuo2Init,
                dim2699Kuo2Init,
                dim2700Kuo2Init,
                dim2701Kuo2Init,
                dim2702Kuo2Init,
                dim2703Kuo2Init,
                dim2704Kuo2Init,
                dim2705Kuo2Init,
                dim2706Kuo2Init,
                dim2707Kuo2Init,
                dim2708Kuo2Init,
                dim2709Kuo2Init,
                dim2710Kuo2Init,
                dim2711Kuo2Init,
                dim2712Kuo2Init,
                dim2713Kuo2Init,
                dim2714Kuo2Init,
                dim2715Kuo2Init,
                dim2716Kuo2Init,
                dim2717Kuo2Init,
                dim2718Kuo2Init,
                dim2719Kuo2Init,
                dim2720Kuo2Init,
                dim2721Kuo2Init,
                dim2722Kuo2Init,
                dim2723Kuo2Init,
                dim2724Kuo2Init,
                dim2725Kuo2Init,
                dim2726Kuo2Init,
                dim2727Kuo2Init,
                dim2728Kuo2Init,
                dim2729Kuo2Init,
                dim2730Kuo2Init,
                dim2731Kuo2Init,
                dim2732Kuo2Init,
                dim2733Kuo2Init,
                dim2734Kuo2Init,
                dim2735Kuo2Init,
                dim2736Kuo2Init,
                dim2737Kuo2Init,
                dim2738Kuo2Init,
                dim2739Kuo2Init,
                dim2740Kuo2Init,
                dim2741Kuo2Init,
                dim2742Kuo2Init,
                dim2743Kuo2Init,
                dim2744Kuo2Init,
                dim2745Kuo2Init,
                dim2746Kuo2Init,
                dim2747Kuo2Init,
                dim2748Kuo2Init,
                dim2749Kuo2Init,
                dim2750Kuo2Init,
                dim2751Kuo2Init,
                dim2752Kuo2Init,
                dim2753Kuo2Init,
                dim2754Kuo2Init,
                dim2755Kuo2Init,
                dim2756Kuo2Init,
                dim2757Kuo2Init,
                dim2758Kuo2Init,
                dim2759Kuo2Init,
                dim2760Kuo2Init,
                dim2761Kuo2Init,
                dim2762Kuo2Init,
                dim2763Kuo2Init,
                dim2764Kuo2Init,
                dim2765Kuo2Init,
                dim2766Kuo2Init,
                dim2767Kuo2Init,
                dim2768Kuo2Init,
                dim2769Kuo2Init,
                dim2770Kuo2Init,
                dim2771Kuo2Init,
                dim2772Kuo2Init,
                dim2773Kuo2Init,
                dim2774Kuo2Init,
                dim2775Kuo2Init,
                dim2776Kuo2Init,
                dim2777Kuo2Init,
                dim2778Kuo2Init,
                dim2779Kuo2Init,
                dim2780Kuo2Init,
                dim2781Kuo2Init,
                dim2782Kuo2Init,
                dim2783Kuo2Init,
                dim2784Kuo2Init,
                dim2785Kuo2Init,
                dim2786Kuo2Init,
                dim2787Kuo2Init,
                dim2788Kuo2Init,
                dim2789Kuo2Init,
                dim2790Kuo2Init,
                dim2791Kuo2Init,
                dim2792Kuo2Init,
                dim2793Kuo2Init,
                dim2794Kuo2Init,
                dim2795Kuo2Init,
                dim2796Kuo2Init,
                dim2797Kuo2Init,
                dim2798Kuo2Init,
                dim2799Kuo2Init,
                dim2800Kuo2Init,
                dim2801Kuo2Init,
                dim2802Kuo2Init,
                dim2803Kuo2Init,
                dim2804Kuo2Init,
                dim2805Kuo2Init,
                dim2806Kuo2Init,
                dim2807Kuo2Init,
                dim2808Kuo2Init,
                dim2809Kuo2Init,
                dim2810Kuo2Init,
                dim2811Kuo2Init,
                dim2812Kuo2Init,
                dim2813Kuo2Init,
                dim2814Kuo2Init,
                dim2815Kuo2Init,
                dim2816Kuo2Init,
                dim2817Kuo2Init,
                dim2818Kuo2Init,
                dim2819Kuo2Init,
                dim2820Kuo2Init,
                dim2821Kuo2Init,
                dim2822Kuo2Init,
                dim2823Kuo2Init,
                dim2824Kuo2Init,
                dim2825Kuo2Init,
                dim2826Kuo2Init,
                dim2827Kuo2Init,
                dim2828Kuo2Init,
                dim2829Kuo2Init,
                dim2830Kuo2Init,
                dim2831Kuo2Init,
                dim2832Kuo2Init,
                dim2833Kuo2Init,
                dim2834Kuo2Init,
                dim2835Kuo2Init,
                dim2836Kuo2Init,
                dim2837Kuo2Init,
                dim2838Kuo2Init,
                dim2839Kuo2Init,
                dim2840Kuo2Init,
                dim2841Kuo2Init,
                dim2842Kuo2Init,
                dim2843Kuo2Init,
                dim2844Kuo2Init,
                dim2845Kuo2Init,
                dim2846Kuo2Init,
                dim2847Kuo2Init,
                dim2848Kuo2Init,
                dim2849Kuo2Init,
                dim2850Kuo2Init,
                dim2851Kuo2Init,
                dim2852Kuo2Init,
                dim2853Kuo2Init,
                dim2854Kuo2Init,
                dim2855Kuo2Init,
                dim2856Kuo2Init,
                dim2857Kuo2Init,
                dim2858Kuo2Init,
                dim2859Kuo2Init,
                dim2860Kuo2Init,
                dim2861Kuo2Init,
                dim2862Kuo2Init,
                dim2863Kuo2Init,
                dim2864Kuo2Init,
                dim2865Kuo2Init,
                dim2866Kuo2Init,
                dim2867Kuo2Init,
                dim2868Kuo2Init,
                dim2869Kuo2Init,
                dim2870Kuo2Init,
                dim2871Kuo2Init,
                dim2872Kuo2Init,
                dim2873Kuo2Init,
                dim2874Kuo2Init,
                dim2875Kuo2Init,
                dim2876Kuo2Init,
                dim2877Kuo2Init,
                dim2878Kuo2Init,
                dim2879Kuo2Init,
                dim2880Kuo2Init,
                dim2881Kuo2Init,
                dim2882Kuo2Init,
                dim2883Kuo2Init,
                dim2884Kuo2Init,
                dim2885Kuo2Init,
                dim2886Kuo2Init,
                dim2887Kuo2Init,
                dim2888Kuo2Init,
                dim2889Kuo2Init,
                dim2890Kuo2Init,
                dim2891Kuo2Init,
                dim2892Kuo2Init,
                dim2893Kuo2Init,
                dim2894Kuo2Init,
                dim2895Kuo2Init,
                dim2896Kuo2Init,
                dim2897Kuo2Init,
                dim2898Kuo2Init,
                dim2899Kuo2Init,
                dim2900Kuo2Init,
                dim2901Kuo2Init,
                dim2902Kuo2Init,
                dim2903Kuo2Init,
                dim2904Kuo2Init,
                dim2905Kuo2Init,
                dim2906Kuo2Init,
                dim2907Kuo2Init,
                dim2908Kuo2Init,
                dim2909Kuo2Init,
                dim2910Kuo2Init,
                dim2911Kuo2Init,
                dim2912Kuo2Init,
                dim2913Kuo2Init,
                dim2914Kuo2Init,
                dim2915Kuo2Init,
                dim2916Kuo2Init,
                dim2917Kuo2Init,
                dim2918Kuo2Init,
                dim2919Kuo2Init,
                dim2920Kuo2Init,
                dim2921Kuo2Init,
                dim2922Kuo2Init,
                dim2923Kuo2Init,
                dim2924Kuo2Init,
                dim2925Kuo2Init,
                dim2926Kuo2Init,
                dim2927Kuo2Init,
                dim2928Kuo2Init,
                dim2929Kuo2Init,
                dim2930Kuo2Init,
                dim2931Kuo2Init,
                dim2932Kuo2Init,
                dim2933Kuo2Init,
                dim2934Kuo2Init,
                dim2935Kuo2Init,
                dim2936Kuo2Init,
                dim2937Kuo2Init,
                dim2938Kuo2Init,
                dim2939Kuo2Init,
                dim2940Kuo2Init,
                dim2941Kuo2Init,
                dim2942Kuo2Init,
                dim2943Kuo2Init,
                dim2944Kuo2Init,
                dim2945Kuo2Init,
                dim2946Kuo2Init,
                dim2947Kuo2Init,
                dim2948Kuo2Init,
                dim2949Kuo2Init,
                dim2950Kuo2Init,
                dim2951Kuo2Init,
                dim2952Kuo2Init,
                dim2953Kuo2Init,
                dim2954Kuo2Init,
                dim2955Kuo2Init,
                dim2956Kuo2Init,
                dim2957Kuo2Init,
                dim2958Kuo2Init,
                dim2959Kuo2Init,
                dim2960Kuo2Init,
                dim2961Kuo2Init,
                dim2962Kuo2Init,
                dim2963Kuo2Init,
                dim2964Kuo2Init,
                dim2965Kuo2Init,
                dim2966Kuo2Init,
                dim2967Kuo2Init,
                dim2968Kuo2Init,
                dim2969Kuo2Init,
                dim2970Kuo2Init,
                dim2971Kuo2Init,
                dim2972Kuo2Init,
                dim2973Kuo2Init,
                dim2974Kuo2Init,
                dim2975Kuo2Init,
                dim2976Kuo2Init,
                dim2977Kuo2Init,
                dim2978Kuo2Init,
                dim2979Kuo2Init,
                dim2980Kuo2Init,
                dim2981Kuo2Init,
                dim2982Kuo2Init,
                dim2983Kuo2Init,
                dim2984Kuo2Init,
                dim2985Kuo2Init,
                dim2986Kuo2Init,
                dim2987Kuo2Init,
                dim2988Kuo2Init,
                dim2989Kuo2Init,
                dim2990Kuo2Init,
                dim2991Kuo2Init,
                dim2992Kuo2Init,
                dim2993Kuo2Init,
                dim2994Kuo2Init,
                dim2995Kuo2Init,
                dim2996Kuo2Init,
                dim2997Kuo2Init,
                dim2998Kuo2Init,
                dim2999Kuo2Init,
                dim3000Kuo2Init,
                dim3001Kuo2Init,
                dim3002Kuo2Init,
                dim3003Kuo2Init,
                dim3004Kuo2Init,
                dim3005Kuo2Init,
                dim3006Kuo2Init,
                dim3007Kuo2Init,
                dim3008Kuo2Init,
                dim3009Kuo2Init,
                dim3010Kuo2Init,
                dim3011Kuo2Init,
                dim3012Kuo2Init,
                dim3013Kuo2Init,
                dim3014Kuo2Init,
                dim3015Kuo2Init,
                dim3016Kuo2Init,
                dim3017Kuo2Init,
                dim3018Kuo2Init,
                dim3019Kuo2Init,
                dim3020Kuo2Init,
                dim3021Kuo2Init,
                dim3022Kuo2Init,
                dim3023Kuo2Init,
                dim3024Kuo2Init,
                dim3025Kuo2Init,
                dim3026Kuo2Init,
                dim3027Kuo2Init,
                dim3028Kuo2Init,
                dim3029Kuo2Init,
                dim3030Kuo2Init,
                dim3031Kuo2Init,
                dim3032Kuo2Init,
                dim3033Kuo2Init,
                dim3034Kuo2Init,
                dim3035Kuo2Init,
                dim3036Kuo2Init,
                dim3037Kuo2Init,
                dim3038Kuo2Init,
                dim3039Kuo2Init,
                dim3040Kuo2Init,
                dim3041Kuo2Init,
                dim3042Kuo2Init,
                dim3043Kuo2Init,
                dim3044Kuo2Init,
                dim3045Kuo2Init,
                dim3046Kuo2Init,
                dim3047Kuo2Init,
                dim3048Kuo2Init,
                dim3049Kuo2Init,
                dim3050Kuo2Init,
                dim3051Kuo2Init,
                dim3052Kuo2Init,
                dim3053Kuo2Init,
                dim3054Kuo2Init,
                dim3055Kuo2Init,
                dim3056Kuo2Init,
                dim3057Kuo2Init,
                dim3058Kuo2Init,
                dim3059Kuo2Init,
                dim3060Kuo2Init,
                dim3061Kuo2Init,
                dim3062Kuo2Init,
                dim3063Kuo2Init,
                dim3064Kuo2Init,
                dim3065Kuo2Init,
                dim3066Kuo2Init,
                dim3067Kuo2Init,
                dim3068Kuo2Init,
                dim3069Kuo2Init,
                dim3070Kuo2Init,
                dim3071Kuo2Init,
                dim3072Kuo2Init,
                dim3073Kuo2Init,
                dim3074Kuo2Init,
                dim3075Kuo2Init,
                dim3076Kuo2Init,
                dim3077Kuo2Init,
                dim3078Kuo2Init,
                dim3079Kuo2Init,
                dim3080Kuo2Init,
                dim3081Kuo2Init,
                dim3082Kuo2Init,
                dim3083Kuo2Init,
                dim3084Kuo2Init,
                dim3085Kuo2Init,
                dim3086Kuo2Init,
                dim3087Kuo2Init,
                dim3088Kuo2Init,
                dim3089Kuo2Init,
                dim3090Kuo2Init,
                dim3091Kuo2Init,
                dim3092Kuo2Init,
                dim3093Kuo2Init,
                dim3094Kuo2Init,
                dim3095Kuo2Init,
                dim3096Kuo2Init,
                dim3097Kuo2Init,
                dim3098Kuo2Init,
                dim3099Kuo2Init,
                dim3100Kuo2Init,
                dim3101Kuo2Init,
                dim3102Kuo2Init,
                dim3103Kuo2Init,
                dim3104Kuo2Init,
                dim3105Kuo2Init,
                dim3106Kuo2Init,
                dim3107Kuo2Init,
                dim3108Kuo2Init,
                dim3109Kuo2Init,
                dim3110Kuo2Init,
                dim3111Kuo2Init,
                dim3112Kuo2Init,
                dim3113Kuo2Init,
                dim3114Kuo2Init,
                dim3115Kuo2Init,
                dim3116Kuo2Init,
                dim3117Kuo2Init,
                dim3118Kuo2Init,
                dim3119Kuo2Init,
                dim3120Kuo2Init,
                dim3121Kuo2Init,
                dim3122Kuo2Init,
                dim3123Kuo2Init,
                dim3124Kuo2Init,
                dim3125Kuo2Init,
                dim3126Kuo2Init,
                dim3127Kuo2Init,
                dim3128Kuo2Init,
                dim3129Kuo2Init,
                dim3130Kuo2Init,
                dim3131Kuo2Init,
                dim3132Kuo2Init,
                dim3133Kuo2Init,
                dim3134Kuo2Init,
                dim3135Kuo2Init,
                dim3136Kuo2Init,
                dim3137Kuo2Init,
                dim3138Kuo2Init,
                dim3139Kuo2Init,
                dim3140Kuo2Init,
                dim3141Kuo2Init,
                dim3142Kuo2Init,
                dim3143Kuo2Init,
                dim3144Kuo2Init,
                dim3145Kuo2Init,
                dim3146Kuo2Init,
                dim3147Kuo2Init,
                dim3148Kuo2Init,
                dim3149Kuo2Init,
                dim3150Kuo2Init,
                dim3151Kuo2Init,
                dim3152Kuo2Init,
                dim3153Kuo2Init,
                dim3154Kuo2Init,
                dim3155Kuo2Init,
                dim3156Kuo2Init,
                dim3157Kuo2Init,
                dim3158Kuo2Init,
                dim3159Kuo2Init,
                dim3160Kuo2Init,
                dim3161Kuo2Init,
                dim3162Kuo2Init,
                dim3163Kuo2Init,
                dim3164Kuo2Init,
                dim3165Kuo2Init,
                dim3166Kuo2Init,
                dim3167Kuo2Init,
                dim3168Kuo2Init,
                dim3169Kuo2Init,
                dim3170Kuo2Init,
                dim3171Kuo2Init,
                dim3172Kuo2Init,
                dim3173Kuo2Init,
                dim3174Kuo2Init,
                dim3175Kuo2Init,
                dim3176Kuo2Init,
                dim3177Kuo2Init,
                dim3178Kuo2Init,
                dim3179Kuo2Init,
                dim3180Kuo2Init,
                dim3181Kuo2Init,
                dim3182Kuo2Init,
                dim3183Kuo2Init,
                dim3184Kuo2Init,
                dim3185Kuo2Init,
                dim3186Kuo2Init,
                dim3187Kuo2Init,
                dim3188Kuo2Init,
                dim3189Kuo2Init,
                dim3190Kuo2Init,
                dim3191Kuo2Init,
                dim3192Kuo2Init,
                dim3193Kuo2Init,
                dim3194Kuo2Init,
                dim3195Kuo2Init,
                dim3196Kuo2Init,
                dim3197Kuo2Init,
                dim3198Kuo2Init,
                dim3199Kuo2Init,
                dim3200Kuo2Init,
                dim3201Kuo2Init,
                dim3202Kuo2Init,
                dim3203Kuo2Init,
                dim3204Kuo2Init,
                dim3205Kuo2Init,
                dim3206Kuo2Init,
                dim3207Kuo2Init,
                dim3208Kuo2Init,
                dim3209Kuo2Init,
                dim3210Kuo2Init,
                dim3211Kuo2Init,
                dim3212Kuo2Init,
                dim3213Kuo2Init,
                dim3214Kuo2Init,
                dim3215Kuo2Init,
                dim3216Kuo2Init,
                dim3217Kuo2Init,
                dim3218Kuo2Init,
                dim3219Kuo2Init,
                dim3220Kuo2Init,
                dim3221Kuo2Init,
                dim3222Kuo2Init,
                dim3223Kuo2Init,
                dim3224Kuo2Init,
                dim3225Kuo2Init,
                dim3226Kuo2Init,
                dim3227Kuo2Init,
                dim3228Kuo2Init,
                dim3229Kuo2Init,
                dim3230Kuo2Init,
                dim3231Kuo2Init,
                dim3232Kuo2Init,
                dim3233Kuo2Init,
                dim3234Kuo2Init,
                dim3235Kuo2Init,
                dim3236Kuo2Init,
                dim3237Kuo2Init,
                dim3238Kuo2Init,
                dim3239Kuo2Init,
                dim3240Kuo2Init,
                dim3241Kuo2Init,
                dim3242Kuo2Init,
                dim3243Kuo2Init,
                dim3244Kuo2Init,
                dim3245Kuo2Init,
                dim3246Kuo2Init,
                dim3247Kuo2Init,
                dim3248Kuo2Init,
                dim3249Kuo2Init,
                dim3250Kuo2Init,
                dim3251Kuo2Init,
                dim3252Kuo2Init,
                dim3253Kuo2Init,
                dim3254Kuo2Init,
                dim3255Kuo2Init,
                dim3256Kuo2Init,
                dim3257Kuo2Init,
                dim3258Kuo2Init,
                dim3259Kuo2Init,
                dim3260Kuo2Init,
                dim3261Kuo2Init,
                dim3262Kuo2Init,
                dim3263Kuo2Init,
                dim3264Kuo2Init,
                dim3265Kuo2Init,
                dim3266Kuo2Init,
                dim3267Kuo2Init,
                dim3268Kuo2Init,
                dim3269Kuo2Init,
                dim3270Kuo2Init,
                dim3271Kuo2Init,
                dim3272Kuo2Init,
                dim3273Kuo2Init,
                dim3274Kuo2Init,
                dim3275Kuo2Init,
                dim3276Kuo2Init,
                dim3277Kuo2Init,
                dim3278Kuo2Init,
                dim3279Kuo2Init,
                dim3280Kuo2Init,
                dim3281Kuo2Init,
                dim3282Kuo2Init,
                dim3283Kuo2Init,
                dim3284Kuo2Init,
                dim3285Kuo2Init,
                dim3286Kuo2Init,
                dim3287Kuo2Init,
                dim3288Kuo2Init,
                dim3289Kuo2Init,
                dim3290Kuo2Init,
                dim3291Kuo2Init,
                dim3292Kuo2Init,
                dim3293Kuo2Init,
                dim3294Kuo2Init,
                dim3295Kuo2Init,
                dim3296Kuo2Init,
                dim3297Kuo2Init,
                dim3298Kuo2Init,
                dim3299Kuo2Init,
                dim3300Kuo2Init,
                dim3301Kuo2Init,
                dim3302Kuo2Init,
                dim3303Kuo2Init,
                dim3304Kuo2Init,
                dim3305Kuo2Init,
                dim3306Kuo2Init,
                dim3307Kuo2Init,
                dim3308Kuo2Init,
                dim3309Kuo2Init,
                dim3310Kuo2Init,
                dim3311Kuo2Init,
                dim3312Kuo2Init,
                dim3313Kuo2Init,
                dim3314Kuo2Init,
                dim3315Kuo2Init,
                dim3316Kuo2Init,
                dim3317Kuo2Init,
                dim3318Kuo2Init,
                dim3319Kuo2Init,
                dim3320Kuo2Init,
                dim3321Kuo2Init,
                dim3322Kuo2Init,
                dim3323Kuo2Init,
                dim3324Kuo2Init,
                dim3325Kuo2Init,
                dim3326Kuo2Init,
                dim3327Kuo2Init,
                dim3328Kuo2Init,
                dim3329Kuo2Init,
                dim3330Kuo2Init,
                dim3331Kuo2Init,
                dim3332Kuo2Init,
                dim3333Kuo2Init,
                dim3334Kuo2Init,
                dim3335Kuo2Init,
                dim3336Kuo2Init,
                dim3337Kuo2Init,
                dim3338Kuo2Init,
                dim3339Kuo2Init,
                dim3340Kuo2Init,
                dim3341Kuo2Init,
                dim3342Kuo2Init,
                dim3343Kuo2Init,
                dim3344Kuo2Init,
                dim3345Kuo2Init,
                dim3346Kuo2Init,
                dim3347Kuo2Init,
                dim3348Kuo2Init,
                dim3349Kuo2Init,
                dim3350Kuo2Init,
                dim3351Kuo2Init,
                dim3352Kuo2Init,
                dim3353Kuo2Init,
                dim3354Kuo2Init,
                dim3355Kuo2Init,
                dim3356Kuo2Init,
                dim3357Kuo2Init,
                dim3358Kuo2Init,
                dim3359Kuo2Init,
                dim3360Kuo2Init,
                dim3361Kuo2Init,
                dim3362Kuo2Init,
                dim3363Kuo2Init,
                dim3364Kuo2Init,
                dim3365Kuo2Init,
                dim3366Kuo2Init,
                dim3367Kuo2Init,
                dim3368Kuo2Init,
                dim3369Kuo2Init,
                dim3370Kuo2Init,
                dim3371Kuo2Init,
                dim3372Kuo2Init,
                dim3373Kuo2Init,
                dim3374Kuo2Init,
                dim3375Kuo2Init,
                dim3376Kuo2Init,
                dim3377Kuo2Init,
                dim3378Kuo2Init,
                dim3379Kuo2Init,
                dim3380Kuo2Init,
                dim3381Kuo2Init,
                dim3382Kuo2Init,
                dim3383Kuo2Init,
                dim3384Kuo2Init,
                dim3385Kuo2Init,
                dim3386Kuo2Init,
                dim3387Kuo2Init,
                dim3388Kuo2Init,
                dim3389Kuo2Init,
                dim3390Kuo2Init,
                dim3391Kuo2Init,
                dim3392Kuo2Init,
                dim3393Kuo2Init,
                dim3394Kuo2Init,
                dim3395Kuo2Init,
                dim3396Kuo2Init,
                dim3397Kuo2Init,
                dim3398Kuo2Init,
                dim3399Kuo2Init,
                dim3400Kuo2Init,
                dim3401Kuo2Init,
                dim3402Kuo2Init,
                dim3403Kuo2Init,
                dim3404Kuo2Init,
                dim3405Kuo2Init,
                dim3406Kuo2Init,
                dim3407Kuo2Init,
                dim3408Kuo2Init,
                dim3409Kuo2Init,
                dim3410Kuo2Init,
                dim3411Kuo2Init,
                dim3412Kuo2Init,
                dim3413Kuo2Init,
                dim3414Kuo2Init,
                dim3415Kuo2Init,
                dim3416Kuo2Init,
                dim3417Kuo2Init,
                dim3418Kuo2Init,
                dim3419Kuo2Init,
                dim3420Kuo2Init,
                dim3421Kuo2Init,
                dim3422Kuo2Init,
                dim3423Kuo2Init,
                dim3424Kuo2Init,
                dim3425Kuo2Init,
                dim3426Kuo2Init,
                dim3427Kuo2Init,
                dim3428Kuo2Init,
                dim3429Kuo2Init,
                dim3430Kuo2Init,
                dim3431Kuo2Init,
                dim3432Kuo2Init,
                dim3433Kuo2Init,
                dim3434Kuo2Init,
                dim3435Kuo2Init,
                dim3436Kuo2Init,
                dim3437Kuo2Init,
                dim3438Kuo2Init,
                dim3439Kuo2Init,
                dim3440Kuo2Init,
                dim3441Kuo2Init,
                dim3442Kuo2Init,
                dim3443Kuo2Init,
                dim3444Kuo2Init,
                dim3445Kuo2Init,
                dim3446Kuo2Init,
                dim3447Kuo2Init,
                dim3448Kuo2Init,
                dim3449Kuo2Init,
                dim3450Kuo2Init,
                dim3451Kuo2Init,
                dim3452Kuo2Init,
                dim3453Kuo2Init,
                dim3454Kuo2Init,
                dim3455Kuo2Init,
                dim3456Kuo2Init,
                dim3457Kuo2Init,
                dim3458Kuo2Init,
                dim3459Kuo2Init,
                dim3460Kuo2Init,
                dim3461Kuo2Init,
                dim3462Kuo2Init,
                dim3463Kuo2Init,
                dim3464Kuo2Init,
                dim3465Kuo2Init,
                dim3466Kuo2Init,
                dim3467Kuo2Init,
                dim3468Kuo2Init,
                dim3469Kuo2Init,
                dim3470Kuo2Init,
                dim3471Kuo2Init,
                dim3472Kuo2Init,
                dim3473Kuo2Init,
                dim3474Kuo2Init,
                dim3475Kuo2Init,
                dim3476Kuo2Init,
                dim3477Kuo2Init,
                dim3478Kuo2Init,
                dim3479Kuo2Init,
                dim3480Kuo2Init,
                dim3481Kuo2Init,
                dim3482Kuo2Init,
                dim3483Kuo2Init,
                dim3484Kuo2Init,
                dim3485Kuo2Init,
                dim3486Kuo2Init,
                dim3487Kuo2Init,
                dim3488Kuo2Init,
                dim3489Kuo2Init,
                dim3490Kuo2Init,
                dim3491Kuo2Init,
                dim3492Kuo2Init,
                dim3493Kuo2Init,
                dim3494Kuo2Init,
                dim3495Kuo2Init,
                dim3496Kuo2Init,
                dim3497Kuo2Init,
                dim3498Kuo2Init,
                dim3499Kuo2Init,
                dim3500Kuo2Init,
                dim3501Kuo2Init,
                dim3502Kuo2Init,
                dim3503Kuo2Init,
                dim3504Kuo2Init,
                dim3505Kuo2Init,
                dim3506Kuo2Init,
                dim3507Kuo2Init,
                dim3508Kuo2Init,
                dim3509Kuo2Init,
                dim3510Kuo2Init,
                dim3511Kuo2Init,
                dim3512Kuo2Init,
                dim3513Kuo2Init,
                dim3514Kuo2Init,
                dim3515Kuo2Init,
                dim3516Kuo2Init,
                dim3517Kuo2Init,
                dim3518Kuo2Init,
                dim3519Kuo2Init,
                dim3520Kuo2Init,
                dim3521Kuo2Init,
                dim3522Kuo2Init,
                dim3523Kuo2Init,
                dim3524Kuo2Init,
                dim3525Kuo2Init,
                dim3526Kuo2Init,
                dim3527Kuo2Init,
                dim3528Kuo2Init,
                dim3529Kuo2Init,
                dim3530Kuo2Init,
                dim3531Kuo2Init,
                dim3532Kuo2Init,
                dim3533Kuo2Init,
                dim3534Kuo2Init,
                dim3535Kuo2Init,
                dim3536Kuo2Init,
                dim3537Kuo2Init,
                dim3538Kuo2Init,
                dim3539Kuo2Init,
                dim3540Kuo2Init,
                dim3541Kuo2Init,
                dim3542Kuo2Init,
                dim3543Kuo2Init,
                dim3544Kuo2Init,
                dim3545Kuo2Init,
                dim3546Kuo2Init,
                dim3547Kuo2Init,
                dim3548Kuo2Init,
                dim3549Kuo2Init,
                dim3550Kuo2Init,
                dim3551Kuo2Init,
                dim3552Kuo2Init,
                dim3553Kuo2Init,
                dim3554Kuo2Init,
                dim3555Kuo2Init,
                dim3556Kuo2Init,
                dim3557Kuo2Init,
                dim3558Kuo2Init,
                dim3559Kuo2Init,
                dim3560Kuo2Init,
                dim3561Kuo2Init,
                dim3562Kuo2Init,
                dim3563Kuo2Init,
                dim3564Kuo2Init,
                dim3565Kuo2Init,
                dim3566Kuo2Init,
                dim3567Kuo2Init,
                dim3568Kuo2Init,
                dim3569Kuo2Init,
                dim3570Kuo2Init,
                dim3571Kuo2Init,
                dim3572Kuo2Init,
                dim3573Kuo2Init,
                dim3574Kuo2Init,
                dim3575Kuo2Init,
                dim3576Kuo2Init,
                dim3577Kuo2Init,
                dim3578Kuo2Init,
                dim3579Kuo2Init,
                dim3580Kuo2Init,
                dim3581Kuo2Init,
                dim3582Kuo2Init,
                dim3583Kuo2Init,
                dim3584Kuo2Init,
                dim3585Kuo2Init,
                dim3586Kuo2Init,
                dim3587Kuo2Init,
                dim3588Kuo2Init,
                dim3589Kuo2Init,
                dim3590Kuo2Init,
                dim3591Kuo2Init,
                dim3592Kuo2Init,
                dim3593Kuo2Init,
                dim3594Kuo2Init,
                dim3595Kuo2Init,
                dim3596Kuo2Init,
                dim3597Kuo2Init,
                dim3598Kuo2Init,
                dim3599Kuo2Init,
                dim3600Kuo2Init,
                dim3601Kuo2Init,
                dim3602Kuo2Init,
                dim3603Kuo2Init,
                dim3604Kuo2Init,
                dim3605Kuo2Init,
                dim3606Kuo2Init,
                dim3607Kuo2Init,
                dim3608Kuo2Init,
                dim3609Kuo2Init,
                dim3610Kuo2Init,
                dim3611Kuo2Init,
                dim3612Kuo2Init,
                dim3613Kuo2Init,
                dim3614Kuo2Init,
                dim3615Kuo2Init,
                dim3616Kuo2Init,
                dim3617Kuo2Init,
                dim3618Kuo2Init,
                dim3619Kuo2Init,
                dim3620Kuo2Init,
                dim3621Kuo2Init,
                dim3622Kuo2Init,
                dim3623Kuo2Init,
                dim3624Kuo2Init,
                dim3625Kuo2Init,
                dim3626Kuo2Init,
                dim3627Kuo2Init,
                dim3628Kuo2Init,
                dim3629Kuo2Init,
                dim3630Kuo2Init,
                dim3631Kuo2Init,
                dim3632Kuo2Init,
                dim3633Kuo2Init,
                dim3634Kuo2Init,
                dim3635Kuo2Init,
                dim3636Kuo2Init,
                dim3637Kuo2Init,
                dim3638Kuo2Init,
                dim3639Kuo2Init,
                dim3640Kuo2Init,
                dim3641Kuo2Init,
                dim3642Kuo2Init,
                dim3643Kuo2Init,
                dim3644Kuo2Init,
                dim3645Kuo2Init,
                dim3646Kuo2Init,
                dim3647Kuo2Init,
                dim3648Kuo2Init,
                dim3649Kuo2Init,
                dim3650Kuo2Init,
                dim3651Kuo2Init,
                dim3652Kuo2Init,
                dim3653Kuo2Init,
                dim3654Kuo2Init,
                dim3655Kuo2Init,
                dim3656Kuo2Init,
                dim3657Kuo2Init,
                dim3658Kuo2Init,
                dim3659Kuo2Init,
                dim3660Kuo2Init,
                dim3661Kuo2Init,
                dim3662Kuo2Init,
                dim3663Kuo2Init,
                dim3664Kuo2Init,
                dim3665Kuo2Init,
                dim3666Kuo2Init,
                dim3667Kuo2Init,
                dim3668Kuo2Init,
                dim3669Kuo2Init,
                dim3670Kuo2Init,
                dim3671Kuo2Init,
                dim3672Kuo2Init,
                dim3673Kuo2Init,
                dim3674Kuo2Init,
                dim3675Kuo2Init,
                dim3676Kuo2Init,
                dim3677Kuo2Init,
                dim3678Kuo2Init,
                dim3679Kuo2Init,
                dim3680Kuo2Init,
                dim3681Kuo2Init,
                dim3682Kuo2Init,
                dim3683Kuo2Init,
                dim3684Kuo2Init,
                dim3685Kuo2Init,
                dim3686Kuo2Init,
                dim3687Kuo2Init,
                dim3688Kuo2Init,
                dim3689Kuo2Init,
                dim3690Kuo2Init,
                dim3691Kuo2Init,
                dim3692Kuo2Init,
                dim3693Kuo2Init,
                dim3694Kuo2Init,
                dim3695Kuo2Init,
                dim3696Kuo2Init,
                dim3697Kuo2Init,
                dim3698Kuo2Init,
                dim3699Kuo2Init,
                dim3700Kuo2Init,
                dim3701Kuo2Init,
                dim3702Kuo2Init,
                dim3703Kuo2Init,
                dim3704Kuo2Init,
                dim3705Kuo2Init,
                dim3706Kuo2Init,
                dim3707Kuo2Init,
                dim3708Kuo2Init,
                dim3709Kuo2Init,
                dim3710Kuo2Init,
                dim3711Kuo2Init,
                dim3712Kuo2Init,
                dim3713Kuo2Init,
                dim3714Kuo2Init,
                dim3715Kuo2Init,
                dim3716Kuo2Init,
                dim3717Kuo2Init,
                dim3718Kuo2Init,
                dim3719Kuo2Init,
                dim3720Kuo2Init,
                dim3721Kuo2Init,
                dim3722Kuo2Init,
                dim3723Kuo2Init,
                dim3724Kuo2Init,
                dim3725Kuo2Init,
                dim3726Kuo2Init,
                dim3727Kuo2Init,
                dim3728Kuo2Init,
                dim3729Kuo2Init,
                dim3730Kuo2Init,
                dim3731Kuo2Init,
                dim3732Kuo2Init,
                dim3733Kuo2Init,
                dim3734Kuo2Init,
                dim3735Kuo2Init,
                dim3736Kuo2Init,
                dim3737Kuo2Init,
                dim3738Kuo2Init,
                dim3739Kuo2Init,
                dim3740Kuo2Init,
                dim3741Kuo2Init,
                dim3742Kuo2Init,
                dim3743Kuo2Init,
                dim3744Kuo2Init,
                dim3745Kuo2Init,
                dim3746Kuo2Init,
                dim3747Kuo2Init,
                dim3748Kuo2Init,
                dim3749Kuo2Init,
                dim3750Kuo2Init,
                dim3751Kuo2Init,
                dim3752Kuo2Init,
                dim3753Kuo2Init,
                dim3754Kuo2Init,
                dim3755Kuo2Init,
                dim3756Kuo2Init,
                dim3757Kuo2Init,
                dim3758Kuo2Init,
                dim3759Kuo2Init,
                dim3760Kuo2Init,
                dim3761Kuo2Init,
                dim3762Kuo2Init,
                dim3763Kuo2Init,
                dim3764Kuo2Init,
                dim3765Kuo2Init,
                dim3766Kuo2Init,
                dim3767Kuo2Init,
                dim3768Kuo2Init,
                dim3769Kuo2Init,
                dim3770Kuo2Init,
                dim3771Kuo2Init,
                dim3772Kuo2Init,
                dim3773Kuo2Init,
                dim3774Kuo2Init,
                dim3775Kuo2Init,
                dim3776Kuo2Init,
                dim3777Kuo2Init,
                dim3778Kuo2Init,
                dim3779Kuo2Init,
                dim3780Kuo2Init,
                dim3781Kuo2Init,
                dim3782Kuo2Init,
                dim3783Kuo2Init,
                dim3784Kuo2Init,
                dim3785Kuo2Init,
                dim3786Kuo2Init,
                dim3787Kuo2Init,
                dim3788Kuo2Init,
                dim3789Kuo2Init,
                dim3790Kuo2Init,
                dim3791Kuo2Init,
                dim3792Kuo2Init,
                dim3793Kuo2Init,
                dim3794Kuo2Init,
                dim3795Kuo2Init,
                dim3796Kuo2Init,
                dim3797Kuo2Init,
                dim3798Kuo2Init,
                dim3799Kuo2Init,
                dim3800Kuo2Init,
                dim3801Kuo2Init,
                dim3802Kuo2Init,
                dim3803Kuo2Init,
                dim3804Kuo2Init,
                dim3805Kuo2Init,
                dim3806Kuo2Init,
                dim3807Kuo2Init,
                dim3808Kuo2Init,
                dim3809Kuo2Init,
                dim3810Kuo2Init,
                dim3811Kuo2Init,
                dim3812Kuo2Init,
                dim3813Kuo2Init,
                dim3814Kuo2Init,
                dim3815Kuo2Init,
                dim3816Kuo2Init,
                dim3817Kuo2Init,
                dim3818Kuo2Init,
                dim3819Kuo2Init,
                dim3820Kuo2Init,
                dim3821Kuo2Init,
                dim3822Kuo2Init,
                dim3823Kuo2Init,
                dim3824Kuo2Init,
                dim3825Kuo2Init,
                dim3826Kuo2Init,
                dim3827Kuo2Init,
                dim3828Kuo2Init,
                dim3829Kuo2Init,
                dim3830Kuo2Init,
                dim3831Kuo2Init,
                dim3832Kuo2Init,
                dim3833Kuo2Init,
                dim3834Kuo2Init,
                dim3835Kuo2Init,
                dim3836Kuo2Init,
                dim3837Kuo2Init,
                dim3838Kuo2Init,
                dim3839Kuo2Init,
                dim3840Kuo2Init,
                dim3841Kuo2Init,
                dim3842Kuo2Init,
                dim3843Kuo2Init,
                dim3844Kuo2Init,
                dim3845Kuo2Init,
                dim3846Kuo2Init,
                dim3847Kuo2Init,
                dim3848Kuo2Init,
                dim3849Kuo2Init,
                dim3850Kuo2Init,
                dim3851Kuo2Init,
                dim3852Kuo2Init,
                dim3853Kuo2Init,
                dim3854Kuo2Init,
                dim3855Kuo2Init,
                dim3856Kuo2Init,
                dim3857Kuo2Init,
                dim3858Kuo2Init,
                dim3859Kuo2Init,
                dim3860Kuo2Init,
                dim3861Kuo2Init,
                dim3862Kuo2Init,
                dim3863Kuo2Init,
                dim3864Kuo2Init,
                dim3865Kuo2Init,
                dim3866Kuo2Init,
                dim3867Kuo2Init,
                dim3868Kuo2Init,
                dim3869Kuo2Init,
                dim3870Kuo2Init,
                dim3871Kuo2Init,
                dim3872Kuo2Init,
                dim3873Kuo2Init,
                dim3874Kuo2Init,
                dim3875Kuo2Init,
                dim3876Kuo2Init,
                dim3877Kuo2Init,
                dim3878Kuo2Init,
                dim3879Kuo2Init,
                dim3880Kuo2Init,
                dim3881Kuo2Init,
                dim3882Kuo2Init,
                dim3883Kuo2Init,
                dim3884Kuo2Init,
                dim3885Kuo2Init,
                dim3886Kuo2Init,
                dim3887Kuo2Init,
                dim3888Kuo2Init,
                dim3889Kuo2Init,
                dim3890Kuo2Init,
                dim3891Kuo2Init,
                dim3892Kuo2Init,
                dim3893Kuo2Init,
                dim3894Kuo2Init,
                dim3895Kuo2Init,
                dim3896Kuo2Init,
                dim3897Kuo2Init,
                dim3898Kuo2Init,
                dim3899Kuo2Init,
                dim3900Kuo2Init,
                dim3901Kuo2Init,
                dim3902Kuo2Init,
                dim3903Kuo2Init,
                dim3904Kuo2Init,
                dim3905Kuo2Init,
                dim3906Kuo2Init,
                dim3907Kuo2Init,
                dim3908Kuo2Init,
                dim3909Kuo2Init,
                dim3910Kuo2Init,
                dim3911Kuo2Init,
                dim3912Kuo2Init,
                dim3913Kuo2Init,
                dim3914Kuo2Init,
                dim3915Kuo2Init,
                dim3916Kuo2Init,
                dim3917Kuo2Init,
                dim3918Kuo2Init,
                dim3919Kuo2Init,
                dim3920Kuo2Init,
                dim3921Kuo2Init,
                dim3922Kuo2Init,
                dim3923Kuo2Init,
                dim3924Kuo2Init,
                dim3925Kuo2Init,
                dim3926Kuo2Init,
                dim3927Kuo2Init,
                dim3928Kuo2Init,
                dim3929Kuo2Init,
                dim3930Kuo2Init,
                dim3931Kuo2Init,
                dim3932Kuo2Init,
                dim3933Kuo2Init,
                dim3934Kuo2Init,
                dim3935Kuo2Init,
                dim3936Kuo2Init,
                dim3937Kuo2Init,
                dim3938Kuo2Init,
                dim3939Kuo2Init,
                dim3940Kuo2Init,
                dim3941Kuo2Init,
                dim3942Kuo2Init,
                dim3943Kuo2Init,
                dim3944Kuo2Init,
                dim3945Kuo2Init,
                dim3946Kuo2Init,
            };

            const unsigned long dim1Kuo3Init[]  =   {   1   ,0 };
            const unsigned long dim2Kuo3Init[]  =   {   1   ,   1   ,0 };
            const unsigned long dim3Kuo3Init[]  =   {   1   ,   1   ,   1   ,0 };
            const unsigned long dim4Kuo3Init[]  =   {   1   ,   3   ,   1   ,0 };
            const unsigned long dim5Kuo3Init[]  =   {   1   ,   3   ,   7   ,   1   ,0 };
            const unsigned long dim6Kuo3Init[]  =   {   1   ,   1   ,   3   ,   7   ,0 };
            const unsigned long dim7Kuo3Init[]  =   {   1   ,   3   ,   1   ,   7   ,   23  ,0 };
            const unsigned long dim8Kuo3Init[]  =   {   1   ,   3   ,   1   ,   1   ,   13  ,0 };
            const unsigned long dim9Kuo3Init[]  =   {   1   ,   1   ,   7   ,   7   ,   9   ,0 };
            const unsigned long dim10Kuo3Init[] =   {   1   ,   1   ,   5   ,   15  ,   21  ,0 };
            const unsigned long dim11Kuo3Init[] =   {   1   ,   3   ,   5   ,   3   ,   23  ,0 };
            const unsigned long dim12Kuo3Init[] =   {   1   ,   1   ,   1   ,   15  ,   27  ,0 };
            const unsigned long dim13Kuo3Init[] =   {   1   ,   3   ,   7   ,   9   ,   1   ,   9   ,0 };
            const unsigned long dim14Kuo3Init[] =   {   1   ,   1   ,   7   ,   15  ,   31  ,   37  ,0 };
            const unsigned long dim15Kuo3Init[] =   {   1   ,   1   ,   5   ,   5   ,   31  ,   9   ,0 };
            const unsigned long dim16Kuo3Init[] =   {   1   ,   3   ,   1   ,   13  ,   3   ,   49  ,0 };
            const unsigned long dim17Kuo3Init[] =   {   1   ,   1   ,   7   ,   15  ,   1   ,   13  ,0 };
            const unsigned long dim18Kuo3Init[] =   {   1   ,   3   ,   3   ,   13  ,   29  ,   37  ,0 };
            const unsigned long dim19Kuo3Init[] =   {   1   ,   1   ,   1   ,   5   ,   13  ,   51  ,   45  ,0 };
            const unsigned long dim20Kuo3Init[] =   {   1   ,   3   ,   1   ,   7   ,   19  ,   23  ,   51  ,0 };
            const unsigned long dim21Kuo3Init[] =   {   1   ,   3   ,   7   ,   1   ,   27  ,   45  ,   61  ,0 };
            const unsigned long dim22Kuo3Init[] =   {   1   ,   3   ,   7   ,   3   ,   9   ,   51  ,   1   ,0 };
            const unsigned long dim23Kuo3Init[] =   {   1   ,   3   ,   1   ,   11  ,   27  ,   3   ,   23  ,0 };
            const unsigned long dim24Kuo3Init[] =   {   1   ,   3   ,   7   ,   9   ,   5   ,   35  ,   105 ,0 };
            const unsigned long dim25Kuo3Init[] =   {   1   ,   1   ,   1   ,   1   ,   19  ,   21  ,   69  ,0 };
            const unsigned long dim26Kuo3Init[] =   {   1   ,   3   ,   3   ,   5   ,   11  ,   61  ,   41  ,0 };
            const unsigned long dim27Kuo3Init[] =   {   1   ,   1   ,   3   ,   13  ,   15  ,   17  ,   25  ,0 };
            const unsigned long dim28Kuo3Init[] =   {   1   ,   1   ,   1   ,   13  ,   19  ,   47  ,   19  ,0 };
            const unsigned long dim29Kuo3Init[] =   {   1   ,   1   ,   7   ,   1   ,   31  ,   23  ,   127 ,0 };
            const unsigned long dim30Kuo3Init[] =   {   1   ,   1   ,   7   ,   1   ,   31  ,   59  ,   123 ,0 };
            const unsigned long dim31Kuo3Init[] =   {   1   ,   1   ,   5   ,   5   ,   13  ,   39  ,   109 ,0 };
            const unsigned long dim32Kuo3Init[] =   {   1   ,   3   ,   7   ,   15  ,   23  ,   19  ,   127 ,0 };
            const unsigned long dim33Kuo3Init[] =   {   1   ,   1   ,   5   ,   3   ,   1   ,   47  ,   15  ,0 };
            const unsigned long dim34Kuo3Init[] =   {   1   ,   1   ,   3   ,   13  ,   23  ,   1   ,   61  ,0 };
            const unsigned long dim35Kuo3Init[] =   {   1   ,   3   ,   3   ,   3   ,   3   ,   35  ,   83  ,0 };
            const unsigned long dim36Kuo3Init[] =   {   1   ,   1   ,   5   ,   5   ,   13  ,   5   ,   19  ,0 };
            const unsigned long dim37Kuo3Init[] =   {   1   ,   1   ,   5   ,   9   ,   13  ,   51  ,   25  ,   117 ,0 };
            const unsigned long dim38Kuo3Init[] =   {   1   ,   1   ,   5   ,   11  ,   19  ,   33  ,   111 ,   71  ,0 };
            const unsigned long dim39Kuo3Init[] =   {   1   ,   3   ,   1   ,   9   ,   5   ,   63  ,   33  ,   27  ,0 };
            const unsigned long dim40Kuo3Init[] =   {   1   ,   3   ,   1   ,   5   ,   19  ,   57  ,   15  ,   181 ,0 };
            const unsigned long dim41Kuo3Init[] =   {   1   ,   3   ,   7   ,   5   ,   5   ,   41  ,   115 ,   15  ,0 };
            const unsigned long dim42Kuo3Init[] =   {   1   ,   3   ,   1   ,   9   ,   9   ,   23  ,   45  ,   43  ,0 };
            const unsigned long dim43Kuo3Init[] =   {   1   ,   1   ,   1   ,   1   ,   13  ,   1   ,   109 ,   29  ,0 };
            const unsigned long dim44Kuo3Init[] =   {   1   ,   1   ,   5   ,   13  ,   31  ,   29  ,   95  ,   111 ,0 };
            const unsigned long dim45Kuo3Init[] =   {   1   ,   1   ,   1   ,   3   ,   9   ,   39  ,   109 ,   17  ,0 };
            const unsigned long dim46Kuo3Init[] =   {   1   ,   1   ,   1   ,   5   ,   17  ,   17  ,   47  ,   153 ,0 };
            const unsigned long dim47Kuo3Init[] =   {   1   ,   1   ,   1   ,   3   ,   31  ,   57  ,   43  ,   17  ,0 };
            const unsigned long dim48Kuo3Init[] =   {   1   ,   3   ,   7   ,   3   ,   29  ,   17  ,   91  ,   153 ,0 };
            const unsigned long dim49Kuo3Init[] =   {   1   ,   3   ,   3   ,   5   ,   21  ,   7   ,   73  ,   91  ,0 };
            const unsigned long dim50Kuo3Init[] =   {   1   ,   1   ,   7   ,   11  ,   13  ,   9   ,   83  ,   213 ,0 };
            const unsigned long dim51Kuo3Init[] =   {   1   ,   1   ,   7   ,   3   ,   17  ,   55  ,   37  ,   189 ,0 };
            const unsigned long dim52Kuo3Init[] =   {   1   ,   3   ,   7   ,   5   ,   7   ,   15  ,   39  ,   247 ,0 };
            const unsigned long dim53Kuo3Init[] =   {   1   ,   3   ,   1   ,   9   ,   27  ,   41  ,   83  ,   63  ,   89  ,0 };
            const unsigned long dim54Kuo3Init[] =   {   1   ,   3   ,   7   ,   11  ,   9   ,   51  ,   59  ,   89  ,   11  ,0 };
            const unsigned long dim55Kuo3Init[] =   {   1   ,   3   ,   5   ,   13  ,   29  ,   21  ,   55  ,   65  ,   45  ,0 };
            const unsigned long dim56Kuo3Init[] =   {   1   ,   3   ,   7   ,   5   ,   29  ,   33  ,   83  ,   229 ,   287 ,0 };
            const unsigned long dim57Kuo3Init[] =   {   1   ,   3   ,   5   ,   15  ,   13  ,   7   ,   89  ,   151 ,   391 ,0 };
            const unsigned long dim58Kuo3Init[] =   {   1   ,   3   ,   3   ,   9   ,   29  ,   3   ,   111 ,   177 ,   475 ,0 };
            const unsigned long dim59Kuo3Init[] =   {   1   ,   3   ,   1   ,   1   ,   5   ,   11  ,   23  ,   33  ,   337 ,0 };
            const unsigned long dim60Kuo3Init[] =   {   1   ,   1   ,   1   ,   15  ,   1   ,   23  ,   111 ,   175 ,   125 ,0 };
            const unsigned long dim61Kuo3Init[] =   {   1   ,   3   ,   7   ,   15  ,   27  ,   47  ,   31  ,   103 ,   245 ,0 };
            const unsigned long dim62Kuo3Init[] =   {   1   ,   3   ,   5   ,   7   ,   7   ,   3   ,   127 ,   187 ,   245 ,0 };
            const unsigned long dim63Kuo3Init[] =   {   1   ,   1   ,   1   ,   13  ,   13  ,   33  ,   33  ,   249 ,   381 ,0 };
            const unsigned long dim64Kuo3Init[] =   {   1   ,   1   ,   3   ,   11  ,   27  ,   45  ,   95  ,   31  ,   419 ,0 };
            const unsigned long dim65Kuo3Init[] =   {   1   ,   1   ,   1   ,   13  ,   17  ,   53  ,   35  ,   197 ,   455 ,0 };
            const unsigned long dim66Kuo3Init[] =   {   1   ,   1   ,   5   ,   7   ,   21  ,   43  ,   93  ,   237 ,   467 ,0 };
            const unsigned long dim67Kuo3Init[] =   {   1   ,   3   ,   5   ,   15  ,   1   ,   19  ,   43  ,   245 ,   193 ,0 };
            const unsigned long dim68Kuo3Init[] =   {   1   ,   1   ,   7   ,   11  ,   21  ,   21  ,   23  ,   147 ,   429 ,0 };
            const unsigned long dim69Kuo3Init[] =   {   1   ,   3   ,   3   ,   7   ,   27  ,   33  ,   99  ,   127 ,   97  ,0 };
            const unsigned long dim70Kuo3Init[] =   {   1   ,   1   ,   5   ,   11  ,   13  ,   21  ,   101 ,   73  ,   17  ,0 };
            const unsigned long dim71Kuo3Init[] =   {   1   ,   3   ,   1   ,   13  ,   3   ,   41  ,   7   ,   237 ,   311 ,0 };
            const unsigned long dim72Kuo3Init[] =   {   1   ,   3   ,   7   ,   1   ,   1   ,   43  ,   115 ,   215 ,   499 ,0 };
            const unsigned long dim73Kuo3Init[] =   {   1   ,   3   ,   1   ,   13  ,   1   ,   9   ,   113 ,   121 ,   473 ,0 };
            const unsigned long dim74Kuo3Init[] =   {   1   ,   1   ,   5   ,   5   ,   21  ,   7   ,   5   ,   171 ,   41  ,0 };
            const unsigned long dim75Kuo3Init[] =   {   1   ,   3   ,   7   ,   15  ,   31  ,   11  ,   87  ,   89  ,   495 ,0 };
            const unsigned long dim76Kuo3Init[] =   {   1   ,   1   ,   5   ,   1   ,   9   ,   9   ,   19  ,   161 ,   325 ,0 };
            const unsigned long dim77Kuo3Init[] =   {   1   ,   3   ,   3   ,   11  ,   15  ,   35  ,   75  ,   99  ,   331 ,0 };
            const unsigned long dim78Kuo3Init[] =   {   1   ,   3   ,   1   ,   1   ,   9   ,   27  ,   103 ,   93  ,   187 ,0 };
            const unsigned long dim79Kuo3Init[] =   {   1   ,   1   ,   3   ,   11  ,   19  ,   43  ,   123 ,   169 ,   325 ,0 };
            const unsigned long dim80Kuo3Init[] =   {   1   ,   1   ,   7   ,   9   ,   17  ,   59  ,   65  ,   139 ,   89  ,0 };
            const unsigned long dim81Kuo3Init[] =   {   1   ,   1   ,   5   ,   15  ,   19  ,   57  ,   29  ,   151 ,   131 ,0 };
            const unsigned long dim82Kuo3Init[] =   {   1   ,   1   ,   1   ,   1   ,   23  ,   49  ,   75  ,   85  ,   365 ,0 };
            const unsigned long dim83Kuo3Init[] =   {   1   ,   1   ,   5   ,   13  ,   11  ,   45  ,   31  ,   245 ,   311 ,0 };
            const unsigned long dim84Kuo3Init[] =   {   1   ,   3   ,   7   ,   5   ,   5   ,   13  ,   23  ,   1   ,   461 ,0 };
            const unsigned long dim85Kuo3Init[] =   {   1   ,   3   ,   1   ,   11  ,   17  ,   23  ,   19  ,   43  ,   393 ,0 };
            const unsigned long dim86Kuo3Init[] =   {   1   ,   1   ,   3   ,   5   ,   27  ,   9   ,   37  ,   105 ,   491 ,0 };
            const unsigned long dim87Kuo3Init[] =   {   1   ,   3   ,   3   ,   1   ,   15  ,   51  ,   107 ,   231 ,   39  ,0 };
            const unsigned long dim88Kuo3Init[] =   {   1   ,   1   ,   1   ,   5   ,   21  ,   57  ,   59  ,   143 ,   47  ,0 };
            const unsigned long dim89Kuo3Init[] =   {   1   ,   1   ,   7   ,   7   ,   23  ,   45  ,   29  ,   33  ,   403 ,0 };
            const unsigned long dim90Kuo3Init[] =   {   1   ,   3   ,   7   ,   5   ,   31  ,   7   ,   91  ,   5   ,   355 ,0 };
            const unsigned long dim91Kuo3Init[] =   {   1   ,   3   ,   5   ,   1   ,   25  ,   7   ,   15  ,   191 ,   45  ,0 };
            const unsigned long dim92Kuo3Init[] =   {   1   ,   1   ,   7   ,   5   ,   9   ,   33  ,   63  ,   75  ,   45  ,0 };
            const unsigned long dim93Kuo3Init[] =   {   1   ,   3   ,   5   ,   7   ,   27  ,   39  ,   83  ,   121 ,   495 ,0 };
            const unsigned long dim94Kuo3Init[] =   {   1   ,   1   ,   7   ,   5   ,   1   ,   25  ,   71  ,   31  ,   33  ,0 };
            const unsigned long dim95Kuo3Init[] =   {   1   ,   1   ,   7   ,   13  ,   13  ,   63  ,   85  ,   203 ,   491 ,0 };
            const unsigned long dim96Kuo3Init[] =   {   1   ,   3   ,   3   ,   11  ,   21  ,   39  ,   25  ,   213 ,   423 ,0 };
            const unsigned long dim97Kuo3Init[] =   {   1   ,   3   ,   5   ,   11  ,   3   ,   7   ,   115 ,   3   ,   191 ,0 };
            const unsigned long dim98Kuo3Init[] =   {   1   ,   1   ,   1   ,   9   ,   19  ,   1   ,   51  ,   247 ,   301 ,0 };
            const unsigned long dim99Kuo3Init[] =   {   1   ,   1   ,   1   ,   7   ,   13  ,   15  ,   45  ,   243 ,   453 ,0 };
            const unsigned long dim100Kuo3Init[]    =   {   1   ,   1   ,   1   ,   7   ,   25  ,   1   ,   113 ,   153 ,   105 ,0 };
            const unsigned long dim101Kuo3Init[]    =   {   1   ,   3   ,   5   ,   11  ,   19  ,   15  ,   3   ,   245 ,   443 ,   651 ,0 };
            const unsigned long dim102Kuo3Init[]    =   {   1   ,   3   ,   7   ,   15  ,   13  ,   35  ,   5   ,   167 ,   233 ,   9   ,0 };
            const unsigned long dim103Kuo3Init[]    =   {   1   ,   3   ,   5   ,   11  ,   21  ,   63  ,   95  ,   45  ,   411 ,   189 ,0 };
            const unsigned long dim104Kuo3Init[]    =   {   1   ,   1   ,   7   ,   7   ,   31  ,   11  ,   81  ,   87  ,   15  ,   677 ,0 };
            const unsigned long dim105Kuo3Init[]    =   {   1   ,   3   ,   3   ,   9   ,   13  ,   21  ,   27  ,   215 ,   459 ,   923 ,0 };
            const unsigned long dim106Kuo3Init[]    =   {   1   ,   1   ,   1   ,   9   ,   13  ,   49  ,   23  ,   7   ,   441 ,   31  ,0 };
            const unsigned long dim107Kuo3Init[]    =   {   1   ,   3   ,   1   ,   7   ,   27  ,   19  ,   5   ,   71  ,   391 ,   159 ,0 };
            const unsigned long dim108Kuo3Init[]    =   {   1   ,   3   ,   7   ,   9   ,   9   ,   55  ,   77  ,   95  ,   353 ,   519 ,0 };
            const unsigned long dim109Kuo3Init[]    =   {   1   ,   3   ,   1   ,   1   ,   23  ,   51  ,   51  ,   113 ,   33  ,   725 ,0 };
            const unsigned long dim110Kuo3Init[]    =   {   1   ,   3   ,   3   ,   1   ,   21  ,   3   ,   107 ,   223 ,   261 ,   913 ,0 };
            const unsigned long dim111Kuo3Init[]    =   {   1   ,   1   ,   1   ,   7   ,   9   ,   31  ,   83  ,   219 ,   285 ,   695 ,0 };
            const unsigned long dim112Kuo3Init[]    =   {   1   ,   1   ,   3   ,   15  ,   1   ,   1   ,   23  ,   49  ,   39  ,   485 ,0 };
            const unsigned long dim113Kuo3Init[]    =   {   1   ,   1   ,   1   ,   9   ,   5   ,   61  ,   71  ,   121 ,   427 ,   83  ,0 };
            const unsigned long dim114Kuo3Init[]    =   {   1   ,   1   ,   3   ,   3   ,   5   ,   45  ,   111 ,   227 ,   261 ,   957 ,0 };
            const unsigned long dim115Kuo3Init[]    =   {   1   ,   3   ,   7   ,   7   ,   17  ,   51  ,   61  ,   85  ,   183 ,   831 ,0 };
            const unsigned long dim116Kuo3Init[]    =   {   1   ,   3   ,   3   ,   5   ,   13  ,   35  ,   27  ,   135 ,   189 ,   743 ,0 };
            const unsigned long dim117Kuo3Init[]    =   {   1   ,   3   ,   1   ,   5   ,   7   ,   5   ,   7   ,   149 ,   333 ,   287 ,0 };
            const unsigned long dim118Kuo3Init[]    =   {   1   ,   3   ,   1   ,   13  ,   19  ,   57  ,   127 ,   201 ,   365 ,   359 ,0 };
            const unsigned long dim119Kuo3Init[]    =   {   1   ,   1   ,   5   ,   9   ,   27  ,   23  ,   79  ,   173 ,   227 ,   299 ,0 };
            const unsigned long dim120Kuo3Init[]    =   {   1   ,   1   ,   7   ,   7   ,   9   ,   35  ,   123 ,   1   ,   289 ,   43  ,0 };
            const unsigned long dim121Kuo3Init[]    =   {   1   ,   1   ,   5   ,   15  ,   15  ,   17  ,   57  ,   175 ,   97  ,   585 ,0 };
            const unsigned long dim122Kuo3Init[]    =   {   1   ,   1   ,   3   ,   13  ,   9   ,   23  ,   123 ,   45  ,   463 ,   137 ,0 };
            const unsigned long dim123Kuo3Init[]    =   {   1   ,   3   ,   1   ,   9   ,   19  ,   3   ,   81  ,   35  ,   221 ,   75  ,0 };
            const unsigned long dim124Kuo3Init[]    =   {   1   ,   3   ,   7   ,   13  ,   31  ,   41  ,   53  ,   207 ,   253 ,   957 ,0 };
            const unsigned long dim125Kuo3Init[]    =   {   1   ,   1   ,   1   ,   5   ,   1   ,   35  ,   51  ,   67  ,   289 ,   417 ,0 };
            const unsigned long dim126Kuo3Init[]    =   {   1   ,   1   ,   1   ,   11  ,   27  ,   21  ,   121 ,   235 ,   49  ,   3   ,0 };
            const unsigned long dim127Kuo3Init[]    =   {   1   ,   3   ,   7   ,   5   ,   15  ,   59  ,   25  ,   43  ,   141 ,   965 ,0 };
            const unsigned long dim128Kuo3Init[]    =   {   1   ,   1   ,   5   ,   3   ,   7   ,   23  ,   121 ,   161 ,   223 ,   245 ,0 };
            const unsigned long dim129Kuo3Init[]    =   {   1   ,   3   ,   7   ,   15  ,   5   ,   5   ,   49  ,   235 ,   329 ,   777 ,0 };
            const unsigned long dim130Kuo3Init[]    =   {   1   ,   3   ,   1   ,   11  ,   7   ,   49  ,   41  ,   187 ,   221 ,   639 ,0 };
            const unsigned long dim131Kuo3Init[]    =   {   1   ,   1   ,   5   ,   9   ,   1   ,   11  ,   41  ,   39  ,   319 ,   203 ,0 };
            const unsigned long dim132Kuo3Init[]    =   {   1   ,   3   ,   3   ,   3   ,   21  ,   61  ,   23  ,   233 ,   397 ,   251 ,0 };
            const unsigned long dim133Kuo3Init[]    =   {   1   ,   1   ,   1   ,   1   ,   27  ,   49  ,   87  ,   251 ,   371 ,   973 ,0 };
            const unsigned long dim134Kuo3Init[]    =   {   1   ,   1   ,   5   ,   11  ,   11  ,   43  ,   57  ,   59  ,   153 ,   695 ,0 };
            const unsigned long dim135Kuo3Init[]    =   {   1   ,   1   ,   3   ,   7   ,   29  ,   57  ,   87  ,   17  ,   129 ,   511 ,0 };
            const unsigned long dim136Kuo3Init[]    =   {   1   ,   3   ,   7   ,   7   ,   5   ,   39  ,   71  ,   27  ,   59  ,   49  ,0 };
            const unsigned long dim137Kuo3Init[]    =   {   1   ,   3   ,   3   ,   5   ,   1   ,   49  ,   65  ,   87  ,   453 ,   515 ,0 };
            const unsigned long dim138Kuo3Init[]    =   {   1   ,   3   ,   3   ,   9   ,   19  ,   3   ,   3   ,   167 ,   85  ,   653 ,0 };
            const unsigned long dim139Kuo3Init[]    =   {   1   ,   1   ,   7   ,   15  ,   7   ,   51  ,   67  ,   37  ,   245 ,   943 ,0 };
            const unsigned long dim140Kuo3Init[]    =   {   1   ,   1   ,   5   ,   7   ,   1   ,   7   ,   21  ,   5   ,   135 ,   999 ,0 };
            const unsigned long dim141Kuo3Init[]    =   {   1   ,   3   ,   3   ,   5   ,   19  ,   1   ,   7   ,   179 ,   405 ,   589 ,0 };
            const unsigned long dim142Kuo3Init[]    =   {   1   ,   1   ,   3   ,   15  ,   3   ,   3   ,   43  ,   79  ,   501 ,   287 ,0 };
            const unsigned long dim143Kuo3Init[]    =   {   1   ,   1   ,   3   ,   9   ,   21  ,   39  ,   59  ,   181 ,   99  ,   737 ,0 };
            const unsigned long dim144Kuo3Init[]    =   {   1   ,   3   ,   7   ,   1   ,   3   ,   11  ,   91  ,   33  ,   273 ,   651 ,0 };
            const unsigned long dim145Kuo3Init[]    =   {   1   ,   1   ,   5   ,   11  ,   23  ,   47  ,   97  ,   253 ,   465 ,   741 ,0 };
            const unsigned long dim146Kuo3Init[]    =   {   1   ,   1   ,   5   ,   1   ,   7   ,   45  ,   33  ,   203 ,   297 ,   193 ,0 };
            const unsigned long dim147Kuo3Init[]    =   {   1   ,   1   ,   5   ,   3   ,   31  ,   55  ,   63  ,   41  ,   243 ,   953 ,0 };
            const unsigned long dim148Kuo3Init[]    =   {   1   ,   3   ,   5   ,   9   ,   9   ,   53  ,   51  ,   71  ,   281 ,   447 ,0 };
            const unsigned long dim149Kuo3Init[]    =   {   1   ,   3   ,   3   ,   9   ,   5   ,   63  ,   77  ,   89  ,   213 ,   527 ,0 };
            const unsigned long dim150Kuo3Init[]    =   {   1   ,   1   ,   3   ,   11  ,   31  ,   15  ,   3   ,   91  ,   321 ,   997 ,0 };
            const unsigned long dim151Kuo3Init[]    =   {   1   ,   3   ,   5   ,   7   ,   31  ,   27  ,   117 ,   179 ,   389 ,   645 ,0 };
            const unsigned long dim152Kuo3Init[]    =   {   1   ,   3   ,   3   ,   9   ,   21  ,   37  ,   41  ,   49  ,   143 ,   401 ,0 };
            const unsigned long dim153Kuo3Init[]    =   {   1   ,   3   ,   1   ,   3   ,   31  ,   57  ,   77  ,   129 ,   49  ,   245 ,0 };
            const unsigned long dim154Kuo3Init[]    =   {   1   ,   3   ,   1   ,   15  ,   27  ,   7   ,   23  ,   69  ,   177 ,   1009    ,0 };
            const unsigned long dim155Kuo3Init[]    =   {   1   ,   3   ,   3   ,   3   ,   11  ,   9   ,   101 ,   141 ,   313 ,   739 ,0 };
            const unsigned long dim156Kuo3Init[]    =   {   1   ,   1   ,   3   ,   13  ,   27  ,   31  ,   27  ,   129 ,   139 ,   645 ,0 };
            const unsigned long dim157Kuo3Init[]    =   {   1   ,   3   ,   7   ,   5   ,   7   ,   11  ,   99  ,   145 ,   313 ,   641 ,0 };
            const unsigned long dim158Kuo3Init[]    =   {   1   ,   3   ,   1   ,   13  ,   15  ,   3   ,   55  ,   183 ,   153 ,   743 ,0 };
            const unsigned long dim159Kuo3Init[]    =   {   1   ,   3   ,   5   ,   1   ,   23  ,   63  ,   53  ,   237 ,   267 ,   333 ,0 };
            const unsigned long dim160Kuo3Init[]    =   {   1   ,   3   ,   1   ,   9   ,   5   ,   27  ,   45  ,   49  ,   9   ,   911 ,0 };
            const unsigned long dim161Kuo3Init[]    =   {   1   ,   1   ,   3   ,   9   ,   11  ,   51  ,   97  ,   231 ,   295 ,   877 ,   353 ,0 };
            const unsigned long dim162Kuo3Init[]    =   {   1   ,   3   ,   7   ,   3   ,   15  ,   55  ,   125 ,   241 ,   149 ,   417 ,   449 ,0 };
            const unsigned long dim163Kuo3Init[]    =   {   1   ,   1   ,   7   ,   1   ,   7   ,   31  ,   89  ,   91  ,   259 ,   819 ,   849 ,0 };
            const unsigned long dim164Kuo3Init[]    =   {   1   ,   3   ,   1   ,   11  ,   3   ,   9   ,   79  ,   77  ,   195 ,   39  ,   937 ,0 };
            const unsigned long dim165Kuo3Init[]    =   {   1   ,   1   ,   3   ,   15  ,   27  ,   37  ,   123 ,   233 ,   221 ,   661 ,   323 ,0 };
            const unsigned long dim166Kuo3Init[]    =   {   1   ,   3   ,   7   ,   7   ,   31  ,   59  ,   121 ,   71  ,   19  ,   723 ,   1713    ,0 };
            const unsigned long dim167Kuo3Init[]    =   {   1   ,   1   ,   5   ,   5   ,   23  ,   43  ,   75  ,   201 ,   7   ,   565 ,   293 ,0 };
            const unsigned long dim168Kuo3Init[]    =   {   1   ,   1   ,   1   ,   3   ,   13  ,   39  ,   115 ,   209 ,   239 ,   439 ,   1623    ,0 };
            const unsigned long dim169Kuo3Init[]    =   {   1   ,   3   ,   7   ,   5   ,   29  ,   39  ,   69  ,   241 ,   71  ,   465 ,   77  ,0 };
            const unsigned long dim170Kuo3Init[]    =   {   1   ,   1   ,   1   ,   13  ,   17  ,   1   ,   91  ,   67  ,   491 ,   969 ,   1709    ,0 };
            const unsigned long dim171Kuo3Init[]    =   {   1   ,   3   ,   1   ,   3   ,   7   ,   37  ,   17  ,   185 ,   365 ,   571 ,   757 ,0 };
            const unsigned long dim172Kuo3Init[]    =   {   1   ,   1   ,   5   ,   1   ,   29  ,   47  ,   87  ,   65  ,   171 ,   893 ,   1659    ,0 };
            const unsigned long dim173Kuo3Init[]    =   {   1   ,   3   ,   5   ,   5   ,   13  ,   5   ,   21  ,   151 ,   91  ,   801 ,   1415    ,0 };
            const unsigned long dim174Kuo3Init[]    =   {   1   ,   3   ,   3   ,   13  ,   27  ,   35  ,   81  ,   49  ,   417 ,   315 ,   1039    ,0 };
            const unsigned long dim175Kuo3Init[]    =   {   1   ,   3   ,   3   ,   7   ,   13  ,   11  ,   63  ,   19  ,   35  ,   659 ,   83  ,0 };
            const unsigned long dim176Kuo3Init[]    =   {   1   ,   1   ,   1   ,   15  ,   11  ,   55  ,   43  ,   181 ,   45  ,   657 ,   681 ,0 };
            const unsigned long dim177Kuo3Init[]    =   {   1   ,   3   ,   7   ,   1   ,   23  ,   9   ,   73  ,   55  ,   143 ,   731 ,   401 ,0 };
            const unsigned long dim178Kuo3Init[]    =   {   1   ,   1   ,   3   ,   1   ,   5   ,   49  ,   125 ,   1   ,   113 ,   41  ,   1733    ,0 };
            const unsigned long dim179Kuo3Init[]    =   {   1   ,   1   ,   3   ,   5   ,   21  ,   63  ,   45  ,   119 ,   67  ,   499 ,   141 ,0 };
            const unsigned long dim180Kuo3Init[]    =   {   1   ,   3   ,   5   ,   11  ,   29  ,   39  ,   45  ,   93  ,   415 ,   515 ,   413 ,0 };
            const unsigned long dim181Kuo3Init[]    =   {   1   ,   1   ,   7   ,   15  ,   5   ,   3   ,   79  ,   201 ,   5   ,   473 ,   55  ,0 };
            const unsigned long dim182Kuo3Init[]    =   {   1   ,   1   ,   3   ,   7   ,   25  ,   31  ,   55  ,   61  ,   175 ,   919 ,   2047    ,0 };
            const unsigned long dim183Kuo3Init[]    =   {   1   ,   1   ,   7   ,   3   ,   31  ,   61  ,   15  ,   147 ,   409 ,   249 ,   1677    ,0 };
            const unsigned long dim184Kuo3Init[]    =   {   1   ,   3   ,   3   ,   9   ,   11  ,   37  ,   9   ,   65  ,   383 ,   689 ,   111 ,0 };
            const unsigned long dim185Kuo3Init[]    =   {   1   ,   1   ,   5   ,   1   ,   31  ,   23  ,   93  ,   15  ,   327 ,   973 ,   765 ,0 };
            const unsigned long dim186Kuo3Init[]    =   {   1   ,   3   ,   7   ,   15  ,   25  ,   17  ,   115 ,   245 ,   205 ,   169 ,   13  ,0 };
            const unsigned long dim187Kuo3Init[]    =   {   1   ,   1   ,   5   ,   15  ,   15  ,   27  ,   109 ,   39  ,   289 ,   135 ,   693 ,0 };
            const unsigned long dim188Kuo3Init[]    =   {   1   ,   1   ,   7   ,   3   ,   23  ,   1   ,   119 ,   199 ,   1   ,   661 ,   1741    ,0 };
            const unsigned long dim189Kuo3Init[]    =   {   1   ,   3   ,   5   ,   13  ,   7   ,   29  ,   3   ,   55  ,   369 ,   903 ,   715 ,0 };
            const unsigned long dim190Kuo3Init[]    =   {   1   ,   1   ,   5   ,   1   ,   29  ,   19  ,   31  ,   7   ,   165 ,   1019    ,   467 ,0 };
            const unsigned long dim191Kuo3Init[]    =   {   1   ,   1   ,   7   ,   11  ,   27  ,   33  ,   105 ,   27  ,   271 ,   55  ,   475 ,0 };
            const unsigned long dim192Kuo3Init[]    =   {   1   ,   1   ,   5   ,   1   ,   27  ,   9   ,   63  ,   157 ,   507 ,   491 ,   1525    ,0 };
            const unsigned long dim193Kuo3Init[]    =   {   1   ,   3   ,   3   ,   9   ,   1   ,   41  ,   71  ,   239 ,   471 ,   147 ,   505 ,0 };
            const unsigned long dim194Kuo3Init[]    =   {   1   ,   3   ,   3   ,   1   ,   25  ,   11  ,   1   ,   35  ,   127 ,   583 ,   2043    ,0 };
            const unsigned long dim195Kuo3Init[]    =   {   1   ,   3   ,   7   ,   15  ,   1   ,   13  ,   119 ,   183 ,   89  ,   555 ,   1739    ,0 };
            const unsigned long dim196Kuo3Init[]    =   {   1   ,   1   ,   1   ,   3   ,   3   ,   45  ,   61  ,   163 ,   357 ,   679 ,   641 ,0 };
            const unsigned long dim197Kuo3Init[]    =   {   1   ,   3   ,   5   ,   13  ,   23  ,   27  ,   57  ,   125 ,   393 ,   917 ,   939 ,0 };
            const unsigned long dim198Kuo3Init[]    =   {   1   ,   1   ,   3   ,   9   ,   9   ,   21  ,   85  ,   105 ,   111 ,   537 ,   733 ,0 };
            const unsigned long dim199Kuo3Init[]    =   {   1   ,   1   ,   3   ,   15  ,   9   ,   61  ,   43  ,   103 ,   65  ,   227 ,   315 ,0 };
            const unsigned long dim200Kuo3Init[]    =   {   1   ,   1   ,   7   ,   9   ,   9   ,   7   ,   31  ,   219 ,   293 ,   441 ,   559 ,0 };
            const unsigned long dim201Kuo3Init[]    =   {   1   ,   1   ,   3   ,   1   ,   23  ,   21  ,   25  ,   229 ,   303 ,   537 ,   1655    ,0 };
            const unsigned long dim202Kuo3Init[]    =   {   1   ,   3   ,   7   ,   7   ,   31  ,   5   ,   21  ,   177 ,   435 ,   899 ,   1939    ,0 };
            const unsigned long dim203Kuo3Init[]    =   {   1   ,   3   ,   3   ,   3   ,   11  ,   43  ,   77  ,   131 ,   353 ,   11  ,   157 ,0 };
            const unsigned long dim204Kuo3Init[]    =   {   1   ,   3   ,   3   ,   3   ,   5   ,   45  ,   13  ,   39  ,   459 ,   261 ,   237 ,0 };
            const unsigned long dim205Kuo3Init[]    =   {   1   ,   3   ,   5   ,   7   ,   23  ,   1   ,   55  ,   63  ,   9   ,   755 ,   589 ,0 };
            const unsigned long dim206Kuo3Init[]    =   {   1   ,   1   ,   7   ,   5   ,   3   ,   11  ,   59  ,   167 ,   145 ,   459 ,   617 ,0 };
            const unsigned long dim207Kuo3Init[]    =   {   1   ,   3   ,   1   ,   13  ,   31  ,   49  ,   53  ,   235 ,   1   ,   37  ,   1373    ,0 };
            const unsigned long dim208Kuo3Init[]    =   {   1   ,   1   ,   7   ,   11  ,   27  ,   63  ,   21  ,   59  ,   183 ,   297 ,   15  ,0 };
            const unsigned long dim209Kuo3Init[]    =   {   1   ,   3   ,   7   ,   15  ,   3   ,   61  ,   57  ,   205 ,   359 ,   799 ,   1591    ,0 };
            const unsigned long dim210Kuo3Init[]    =   {   1   ,   1   ,   1   ,   3   ,   29  ,   29  ,   121 ,   193 ,   405 ,   161 ,   909 ,0 };
            const unsigned long dim211Kuo3Init[]    =   {   1   ,   1   ,   5   ,   5   ,   15  ,   43  ,   19  ,   87  ,   337 ,   595 ,   551 ,0 };
            const unsigned long dim212Kuo3Init[]    =   {   1   ,   3   ,   3   ,   7   ,   27  ,   23  ,   109 ,   127 ,   107 ,   793 ,   787 ,0 };
            const unsigned long dim213Kuo3Init[]    =   {   1   ,   1   ,   3   ,   5   ,   9   ,   11  ,   63  ,   21  ,   363 ,   403 ,   981 ,0 };
            const unsigned long dim214Kuo3Init[]    =   {   1   ,   1   ,   3   ,   5   ,   1   ,   39  ,   43  ,   71  ,   5   ,   151 ,   1535    ,0 };
            const unsigned long dim215Kuo3Init[]    =   {   1   ,   1   ,   7   ,   3   ,   1   ,   33  ,   65  ,   249 ,   81  ,   889 ,   1591    ,0 };
            const unsigned long dim216Kuo3Init[]    =   {   1   ,   3   ,   5   ,   11  ,   15  ,   53  ,   117 ,   47  ,   13  ,   215 ,   1953    ,0 };
            const unsigned long dim217Kuo3Init[]    =   {   1   ,   1   ,   3   ,   13  ,   15  ,   13  ,   49  ,   209 ,   137 ,   403 ,   1373    ,0 };
            const unsigned long dim218Kuo3Init[]    =   {   1   ,   1   ,   3   ,   15  ,   25  ,   13  ,   123 ,   193 ,   275 ,   235 ,   1975    ,0 };
            const unsigned long dim219Kuo3Init[]    =   {   1   ,   3   ,   1   ,   3   ,   21  ,   55  ,   109 ,   115 ,   411 ,   413 ,   2023    ,0 };
            const unsigned long dim220Kuo3Init[]    =   {   1   ,   3   ,   3   ,   1   ,   5   ,   7   ,   53  ,   161 ,   221 ,   529 ,   489 ,0 };
            const unsigned long dim221Kuo3Init[]    =   {   1   ,   1   ,   3   ,   5   ,   25  ,   27  ,   103 ,   31  ,   7   ,   103 ,   343 ,0 };
            const unsigned long dim222Kuo3Init[]    =   {   1   ,   1   ,   7   ,   11  ,   19  ,   47  ,   45  ,   201 ,   499 ,   255 ,   1501    ,0 };
            const unsigned long dim223Kuo3Init[]    =   {   1   ,   3   ,   1   ,   1   ,   5   ,   7   ,   79  ,   113 ,   445 ,   515 ,   287 ,0 };
            const unsigned long dim224Kuo3Init[]    =   {   1   ,   1   ,   7   ,   11  ,   7   ,   35  ,   87  ,   145 ,   115 ,   773 ,   1415    ,0 };
            const unsigned long dim225Kuo3Init[]    =   {   1   ,   1   ,   1   ,   13  ,   15  ,   25  ,   83  ,   143 ,   347 ,   473 ,   415 ,0 };
            const unsigned long dim226Kuo3Init[]    =   {   1   ,   3   ,   1   ,   11  ,   23  ,   45  ,   79  ,   245 ,   81  ,   835 ,   589 ,0 };
            const unsigned long dim227Kuo3Init[]    =   {   1   ,   1   ,   1   ,   13  ,   25  ,   49  ,   17  ,   15  ,   255 ,   421 ,   1195    ,0 };
            const unsigned long dim228Kuo3Init[]    =   {   1   ,   1   ,   3   ,   1   ,   13  ,   31  ,   103 ,   189 ,   43  ,   389 ,   1497    ,0 };
            const unsigned long dim229Kuo3Init[]    =   {   1   ,   3   ,   5   ,   7   ,   3   ,   63  ,   21  ,   151 ,   339 ,   525 ,   455 ,0 };
            const unsigned long dim230Kuo3Init[]    =   {   1   ,   3   ,   3   ,   13  ,   23  ,   29  ,   101 ,   25  ,   461 ,   631 ,   2017    ,0 };
            const unsigned long dim231Kuo3Init[]    =   {   1   ,   1   ,   5   ,   13  ,   7   ,   57  ,   11  ,   181 ,   215 ,   911 ,   463 ,0 };
            const unsigned long dim232Kuo3Init[]    =   {   1   ,   3   ,   7   ,   1   ,   1   ,   45  ,   41  ,   101 ,   263 ,   835 ,   1985    ,0 };
            const unsigned long dim233Kuo3Init[]    =   {   1   ,   3   ,   3   ,   3   ,   25  ,   13  ,   127 ,   245 ,   137 ,   709 ,   993 ,0 };
            const unsigned long dim234Kuo3Init[]    =   {   1   ,   3   ,   7   ,   3   ,   5   ,   19  ,   23  ,   55  ,   51  ,   367 ,   1513    ,0 };
            const unsigned long dim235Kuo3Init[]    =   {   1   ,   1   ,   5   ,   3   ,   13  ,   39  ,   101 ,   119 ,   119 ,   835 ,   875 ,0 };
            const unsigned long dim236Kuo3Init[]    =   {   1   ,   1   ,   7   ,   13  ,   17  ,   27  ,   103 ,   31  ,   223 ,   461 ,   1267    ,0 };
            const unsigned long dim237Kuo3Init[]    =   {   1   ,   3   ,   5   ,   13  ,   29  ,   29  ,   29  ,   189 ,   13  ,   647 ,   1931    ,0 };
            const unsigned long dim238Kuo3Init[]    =   {   1   ,   3   ,   5   ,   15  ,   9   ,   27  ,   127 ,   43  ,   341 ,   939 ,   1981    ,0 };
            const unsigned long dim239Kuo3Init[]    =   {   1   ,   1   ,   7   ,   11  ,   27  ,   27  ,   25  ,   163 ,   83  ,   709 ,   141 ,0 };
            const unsigned long dim240Kuo3Init[]    =   {   1   ,   3   ,   7   ,   15  ,   31  ,   45  ,   119 ,   141 ,   181 ,   231 ,   411 ,0 };
            const unsigned long dim241Kuo3Init[]    =   {   1   ,   1   ,   5   ,   9   ,   31  ,   25  ,   113 ,   243 ,   435 ,   55  ,   765 ,0 };
            const unsigned long dim242Kuo3Init[]    =   {   1   ,   1   ,   7   ,   9   ,   27  ,   5   ,   63  ,   23  ,   133 ,   767 ,   1963    ,0 };
            const unsigned long dim243Kuo3Init[]    =   {   1   ,   3   ,   7   ,   9   ,   9   ,   55  ,   97  ,   159 ,   371 ,   1013    ,   2013    ,0 };
            const unsigned long dim244Kuo3Init[]    =   {   1   ,   1   ,   3   ,   7   ,   3   ,   17  ,   13  ,   215 ,   15  ,   281 ,   291 ,0 };
            const unsigned long dim245Kuo3Init[]    =   {   1   ,   3   ,   5   ,   9   ,   29  ,   7   ,   71  ,   9   ,   181 ,   763 ,   1111    ,0 };
            const unsigned long dim246Kuo3Init[]    =   {   1   ,   3   ,   7   ,   13  ,   23  ,   55  ,   79  ,   99  ,   165 ,   447 ,   1947    ,0 };
            const unsigned long dim247Kuo3Init[]    =   {   1   ,   3   ,   7   ,   7   ,   29  ,   59  ,   25  ,   41  ,   313 ,   665 ,   1245    ,0 };
            const unsigned long dim248Kuo3Init[]    =   {   1   ,   1   ,   3   ,   3   ,   5   ,   29  ,   71  ,   71  ,   475 ,   781 ,   567 ,0 };
            const unsigned long dim249Kuo3Init[]    =   {   1   ,   1   ,   1   ,   3   ,   19  ,   53  ,   61  ,   215 ,   475 ,   885 ,   459 ,0 };
            const unsigned long dim250Kuo3Init[]    =   {   1   ,   3   ,   3   ,   13  ,   5   ,   31  ,   95  ,   137 ,   119 ,   249 ,   975 ,0 };
            const unsigned long dim251Kuo3Init[]    =   {   1   ,   1   ,   1   ,   15  ,   21  ,   51  ,   71  ,   21  ,   193 ,   709 ,   1167    ,0 };
            const unsigned long dim252Kuo3Init[]    =   {   1   ,   3   ,   7   ,   9   ,   29  ,   61  ,   45  ,   153 ,   351 ,   867 ,   995 ,0 };
            const unsigned long dim253Kuo3Init[]    =   {   1   ,   1   ,   1   ,   15  ,   7   ,   37  ,   65  ,   51  ,   45  ,   497 ,   1825    ,0 };
            const unsigned long dim254Kuo3Init[]    =   {   1   ,   3   ,   3   ,   1   ,   23  ,   59  ,   11  ,   127 ,   501 ,   351 ,   1525    ,0 };
            const unsigned long dim255Kuo3Init[]    =   {   1   ,   1   ,   3   ,   15  ,   31  ,   29  ,   47  ,   105 ,   179 ,   159 ,   991 ,0 };
            const unsigned long dim256Kuo3Init[]    =   {   1   ,   3   ,   3   ,   15  ,   27  ,   23  ,   93  ,   163 ,   169 ,   1003    ,   1369    ,0 };
            const unsigned long dim257Kuo3Init[]    =   {   1   ,   3   ,   1   ,   5   ,   31  ,   51  ,   85  ,   221 ,   87  ,   941 ,   317 ,0 };
            const unsigned long dim258Kuo3Init[]    =   {   1   ,   3   ,   7   ,   15  ,   21  ,   19  ,   71  ,   205 ,   385 ,   371 ,   1401    ,0 };
            const unsigned long dim259Kuo3Init[]    =   {   1   ,   1   ,   5   ,   11  ,   1   ,   51  ,   31  ,   185 ,   407 ,   233 ,   1171    ,0 };
            const unsigned long dim260Kuo3Init[]    =   {   1   ,   3   ,   3   ,   9   ,   19  ,   11  ,   93  ,   253 ,   373 ,   679 ,   69  ,0 };
            const unsigned long dim261Kuo3Init[]    =   {   1   ,   1   ,   7   ,   15  ,   23  ,   45  ,   87  ,   61  ,   113 ,   541 ,   1701    ,0 };
            const unsigned long dim262Kuo3Init[]    =   {   1   ,   1   ,   5   ,   7   ,   7   ,   61  ,   55  ,   167 ,   345 ,   633 ,   1699    ,0 };
            const unsigned long dim263Kuo3Init[]    =   {   1   ,   3   ,   7   ,   7   ,   1   ,   23  ,   121 ,   33  ,   507 ,   361 ,   1945    ,0 };
            const unsigned long dim264Kuo3Init[]    =   {   1   ,   3   ,   7   ,   1   ,   19  ,   47  ,   43  ,   101 ,   67  ,   57  ,   1685    ,0 };
            const unsigned long dim265Kuo3Init[]    =   {   1   ,   3   ,   3   ,   3   ,   25  ,   63  ,   53  ,   149 ,   483 ,   551 ,   1519    ,0 };
            const unsigned long dim266Kuo3Init[]    =   {   1   ,   3   ,   1   ,   3   ,   27  ,   1   ,   41  ,   213 ,   57  ,   897 ,   121 ,0 };
            const unsigned long dim267Kuo3Init[]    =   {   1   ,   3   ,   1   ,   5   ,   19  ,   13  ,   49  ,   81  ,   239 ,   455 ,   957 ,0 };
            const unsigned long dim268Kuo3Init[]    =   {   1   ,   1   ,   3   ,   7   ,   15  ,   23  ,   89  ,   171 ,   319 ,   117 ,   2041    ,0 };
            const unsigned long dim269Kuo3Init[]    =   {   1   ,   1   ,   3   ,   13  ,   3   ,   13  ,   41  ,   15  ,   145 ,   659 ,   973 ,0 };
            const unsigned long dim270Kuo3Init[]    =   {   1   ,   3   ,   7   ,   1   ,   15  ,   17  ,   69  ,   53  ,   245 ,   179 ,   159 ,0 };
            const unsigned long dim271Kuo3Init[]    =   {   1   ,   3   ,   3   ,   5   ,   5   ,   3   ,   113 ,   183 ,   219 ,   469 ,   1169    ,0 };
            const unsigned long dim272Kuo3Init[]    =   {   1   ,   1   ,   5   ,   15  ,   13  ,   45  ,   89  ,   87  ,   371 ,   267 ,   1643    ,0 };
            const unsigned long dim273Kuo3Init[]    =   {   1   ,   3   ,   5   ,   5   ,   1   ,   25  ,   3   ,   247 ,   309 ,   401 ,   2027    ,0 };
            const unsigned long dim274Kuo3Init[]    =   {   1   ,   3   ,   1   ,   11  ,   13  ,   7   ,   79  ,   17  ,   297 ,   679 ,   1699    ,0 };
            const unsigned long dim275Kuo3Init[]    =   {   1   ,   3   ,   1   ,   5   ,   13  ,   19  ,   123 ,   255 ,   343 ,   393 ,   1635    ,0 };
            const unsigned long dim276Kuo3Init[]    =   {   1   ,   3   ,   3   ,   15  ,   13  ,   13  ,   37  ,   229 ,   483 ,   283 ,   1423    ,0 };
            const unsigned long dim277Kuo3Init[]    =   {   1   ,   3   ,   1   ,   1   ,   7   ,   37  ,   65  ,   67  ,   77  ,   247 ,   1539    ,0 };
            const unsigned long dim278Kuo3Init[]    =   {   1   ,   1   ,   3   ,   5   ,   25  ,   39  ,   37  ,   19  ,   231 ,   835 ,   797 ,0 };
            const unsigned long dim279Kuo3Init[]    =   {   1   ,   1   ,   1   ,   11  ,   27  ,   53  ,   1   ,   217 ,   211 ,   7   ,   801 ,0 };
            const unsigned long dim280Kuo3Init[]    =   {   1   ,   3   ,   7   ,   1   ,   15  ,   55  ,   69  ,   77  ,   491 ,   61  ,   55  ,0 };
            const unsigned long dim281Kuo3Init[]    =   {   1   ,   1   ,   1   ,   1   ,   29  ,   21  ,   47  ,   167 ,   281 ,   105 ,   1363    ,0 };
            const unsigned long dim282Kuo3Init[]    =   {   1   ,   3   ,   1   ,   15  ,   15  ,   55  ,   57  ,   207 ,   479 ,   709 ,   867 ,0 };
            const unsigned long dim283Kuo3Init[]    =   {   1   ,   1   ,   1   ,   9   ,   7   ,   13  ,   115 ,   237 ,   47  ,   127 ,   1849    ,0 };
            const unsigned long dim284Kuo3Init[]    =   {   1   ,   1   ,   7   ,   9   ,   7   ,   35  ,   83  ,   63  ,   155 ,   441 ,   1105    ,0 };
            const unsigned long dim285Kuo3Init[]    =   {   1   ,   3   ,   3   ,   9   ,   17  ,   39  ,   19  ,   231 ,   491 ,   593 ,   1997    ,0 };
            const unsigned long dim286Kuo3Init[]    =   {   1   ,   3   ,   3   ,   13  ,   17  ,   19  ,   65  ,   21  ,   389 ,   391 ,   483 ,0 };
            const unsigned long dim287Kuo3Init[]    =   {   1   ,   1   ,   5   ,   7   ,   17  ,   7   ,   97  ,   173 ,   413 ,   917 ,   1691    ,0 };
            const unsigned long dim288Kuo3Init[]    =   {   1   ,   1   ,   3   ,   3   ,   29  ,   49  ,   3   ,   127 ,   337 ,   713 ,   1057    ,0 };
            const unsigned long dim289Kuo3Init[]    =   {   1   ,   3   ,   7   ,   5   ,   31  ,   23  ,   31  ,   119 ,   161 ,   51  ,   537 ,0 };
            const unsigned long dim290Kuo3Init[]    =   {   1   ,   1   ,   5   ,   1   ,   29  ,   25  ,   15  ,   187 ,   143 ,   927 ,   715 ,0 };
            const unsigned long dim291Kuo3Init[]    =   {   1   ,   1   ,   7   ,   11  ,   15  ,   47  ,   55  ,   103 ,   109 ,   465 ,   1439    ,0 };
            const unsigned long dim292Kuo3Init[]    =   {   1   ,   1   ,   3   ,   13  ,   31  ,   47  ,   115 ,   177 ,   355 ,   163 ,   833 ,0 };
            const unsigned long dim293Kuo3Init[]    =   {   1   ,   3   ,   5   ,   5   ,   31  ,   13  ,   15  ,   163 ,   57  ,   937 ,   1553    ,0 };
            const unsigned long dim294Kuo3Init[]    =   {   1   ,   3   ,   5   ,   15  ,   7   ,   29  ,   47  ,   139 ,   231 ,   507 ,   595 ,0 };
            const unsigned long dim295Kuo3Init[]    =   {   1   ,   3   ,   1   ,   15  ,   31  ,   13  ,   49  ,   253 ,   29  ,   109 ,   727 ,0 };
            const unsigned long dim296Kuo3Init[]    =   {   1   ,   1   ,   5   ,   15  ,   21  ,   1   ,   107 ,   89  ,   367 ,   825 ,   1297    ,0 };
            const unsigned long dim297Kuo3Init[]    =   {   1   ,   3   ,   3   ,   9   ,   31  ,   21  ,   31  ,   247 ,   375 ,   767 ,   1761    ,0 };
            const unsigned long dim298Kuo3Init[]    =   {   1   ,   1   ,   7   ,   3   ,   31  ,   33  ,   119 ,   119 ,   409 ,   955 ,   1411    ,0 };
            const unsigned long dim299Kuo3Init[]    =   {   1   ,   1   ,   7   ,   1   ,   17  ,   57  ,   101 ,   37  ,   147 ,   367 ,   265 ,0 };
            const unsigned long dim300Kuo3Init[]    =   {   1   ,   3   ,   3   ,   3   ,   25  ,   53  ,   91  ,   43  ,   351 ,   1021    ,   1243    ,0 };
            const unsigned long dim301Kuo3Init[]    =   {   1   ,   3   ,   3   ,   5   ,   11  ,   47  ,   107 ,   209 ,   251 ,   765 ,   1089    ,0 };
            const unsigned long dim302Kuo3Init[]    =   {   1   ,   1   ,   7   ,   5   ,   9   ,   13  ,   23  ,   153 ,   367 ,   585 ,   315 ,0 };
            const unsigned long dim303Kuo3Init[]    =   {   1   ,   3   ,   3   ,   15  ,   23  ,   57  ,   83  ,   173 ,   413 ,   555 ,   159 ,0 };
            const unsigned long dim304Kuo3Init[]    =   {   1   ,   3   ,   7   ,   15  ,   7   ,   59  ,   31  ,   233 ,   67  ,   767 ,   75  ,0 };
            const unsigned long dim305Kuo3Init[]    =   {   1   ,   3   ,   5   ,   3   ,   17  ,   1   ,   35  ,   37  ,   445 ,   847 ,   1853    ,0 };
            const unsigned long dim306Kuo3Init[]    =   {   1   ,   3   ,   5   ,   13  ,   3   ,   1   ,   75  ,   113 ,   83  ,   467 ,   1057    ,0 };
            const unsigned long dim307Kuo3Init[]    =   {   1   ,   1   ,   7   ,   1   ,   7   ,   15  ,   125 ,   31  ,   341 ,   133 ,   285 ,0 };
            const unsigned long dim308Kuo3Init[]    =   {   1   ,   1   ,   5   ,   5   ,   15  ,   11  ,   97  ,   3   ,   59  ,   777 ,   1125    ,0 };
            const unsigned long dim309Kuo3Init[]    =   {   1   ,   3   ,   7   ,   9   ,   17  ,   55  ,   117 ,   255 ,   429 ,   239 ,   1071    ,0 };
            const unsigned long dim310Kuo3Init[]    =   {   1   ,   1   ,   7   ,   3   ,   27  ,   37  ,   31  ,   221 ,   385 ,   877 ,   1767    ,0 };
            const unsigned long dim311Kuo3Init[]    =   {   1   ,   1   ,   5   ,   1   ,   5   ,   31  ,   43  ,   23  ,   77  ,   497 ,   1605    ,0 };
            const unsigned long dim312Kuo3Init[]    =   {   1   ,   3   ,   5   ,   15  ,   15  ,   37  ,   103 ,   227 ,   33  ,   185 ,   481 ,0 };
            const unsigned long dim313Kuo3Init[]    =   {   1   ,   3   ,   3   ,   11  ,   29  ,   61  ,   67  ,   53  ,   257 ,   609 ,   1581    ,0 };
            const unsigned long dim314Kuo3Init[]    =   {   1   ,   1   ,   3   ,   15  ,   31  ,   25  ,   123 ,   59  ,   395 ,   107 ,   587 ,0 };
            const unsigned long dim315Kuo3Init[]    =   {   1   ,   3   ,   1   ,   7   ,   29  ,   19  ,   51  ,   135 ,   503 ,   71  ,   33  ,0 };
            const unsigned long dim316Kuo3Init[]    =   {   1   ,   3   ,   1   ,   15  ,   21  ,   31  ,   23  ,   85  ,   17  ,   227 ,   657 ,0 };
            const unsigned long dim317Kuo3Init[]    =   {   1   ,   1   ,   1   ,   3   ,   9   ,   63  ,   61  ,   205 ,   61  ,   699 ,   599 ,0 };
            const unsigned long dim318Kuo3Init[]    =   {   1   ,   3   ,   1   ,   15  ,   1   ,   39  ,   45  ,   133 ,   1   ,   391 ,   255 ,0 };
            const unsigned long dim319Kuo3Init[]    =   {   1   ,   3   ,   5   ,   13  ,   17  ,   13  ,   107 ,   245 ,   55  ,   619 ,   1085    ,0 };
            const unsigned long dim320Kuo3Init[]    =   {   1   ,   3   ,   3   ,   9   ,   5   ,   25  ,   5   ,   49  ,   509 ,   661 ,   1225    ,0 };
            const unsigned long dim321Kuo3Init[]    =   {   1   ,   3   ,   7   ,   9   ,   29  ,   61  ,   13  ,   247 ,   341 ,   257 ,   1551    ,0 };
            const unsigned long dim322Kuo3Init[]    =   {   1   ,   3   ,   7   ,   9   ,   5   ,   39  ,   123 ,   221 ,   391 ,   501 ,   527 ,0 };
            const unsigned long dim323Kuo3Init[]    =   {   1   ,   3   ,   1   ,   7   ,   21  ,   23  ,   65  ,   233 ,   181 ,   405 ,   1297    ,0 };
            const unsigned long dim324Kuo3Init[]    =   {   1   ,   1   ,   1   ,   3   ,   23  ,   23  ,   15  ,   253 ,   37  ,   679 ,   1455    ,0 };
            const unsigned long dim325Kuo3Init[]    =   {   1   ,   3   ,   3   ,   3   ,   3   ,   31  ,   89  ,   105 ,   11  ,   1015    ,   1837    ,0 };
            const unsigned long dim326Kuo3Init[]    =   {   1   ,   3   ,   1   ,   7   ,   15  ,   59  ,   105 ,   179 ,   477 ,   923 ,   1189    ,0 };
            const unsigned long dim327Kuo3Init[]    =   {   1   ,   3   ,   5   ,   9   ,   1   ,   9   ,   21  ,   71  ,   145 ,   409 ,   1235    ,0 };
            const unsigned long dim328Kuo3Init[]    =   {   1   ,   1   ,   7   ,   13  ,   9   ,   43  ,   85  ,   241 ,   497 ,   683 ,   1169    ,0 };
            const unsigned long dim329Kuo3Init[]    =   {   1   ,   3   ,   3   ,   13  ,   29  ,   41  ,   23  ,   51  ,   1   ,   563 ,   905 ,0 };
            const unsigned long dim330Kuo3Init[]    =   {   1   ,   3   ,   5   ,   9   ,   25  ,   51  ,   1   ,   71  ,   25  ,   195 ,   363 ,0 };
            const unsigned long dim331Kuo3Init[]    =   {   1   ,   3   ,   7   ,   7   ,   31  ,   49  ,   13  ,   255 ,   471 ,   653 ,   1127    ,0 };
            const unsigned long dim332Kuo3Init[]    =   {   1   ,   3   ,   3   ,   7   ,   7   ,   61  ,   63  ,   21  ,   413 ,   817 ,   1369    ,0 };
            const unsigned long dim333Kuo3Init[]    =   {   1   ,   3   ,   5   ,   1   ,   7   ,   53  ,   43  ,   167 ,   173 ,   261 ,   633 ,0 };
            const unsigned long dim334Kuo3Init[]    =   {   1   ,   1   ,   1   ,   13  ,   21  ,   59  ,   49  ,   135 ,   351 ,   187 ,   447 ,0 };
            const unsigned long dim335Kuo3Init[]    =   {   1   ,   1   ,   3   ,   7   ,   5   ,   35  ,   103 ,   5   ,   247 ,   555 ,   1439    ,0 };
            const unsigned long dim336Kuo3Init[]    =   {   1   ,   1   ,   3   ,   1   ,   3   ,   11  ,   5   ,   153 ,   507 ,   21  ,   1331    ,0 };
            const unsigned long dim337Kuo3Init[]    =   {   1   ,   3   ,   7   ,   15  ,   25  ,   11  ,   29  ,   51  ,   353 ,   447 ,   337 ,   3621    ,0 };
            const unsigned long dim338Kuo3Init[]    =   {   1   ,   3   ,   5   ,   13  ,   31  ,   51  ,   103 ,   69  ,   311 ,   645 ,   1373    ,   1155    ,0 };
            const unsigned long dim339Kuo3Init[]    =   {   1   ,   1   ,   7   ,   7   ,   1   ,   55  ,   101 ,   203 ,   503 ,   711 ,   131 ,   937 ,0 };
            const unsigned long dim340Kuo3Init[]    =   {   1   ,   3   ,   3   ,   11  ,   7   ,   5   ,   75  ,   63  ,   217 ,   643 ,   871 ,   3511    ,0 };
            const unsigned long dim341Kuo3Init[]    =   {   1   ,   1   ,   3   ,   15  ,   21  ,   23  ,   75  ,   3   ,   113 ,   203 ,   13  ,   2727    ,0 };
            const unsigned long dim342Kuo3Init[]    =   {   1   ,   3   ,   5   ,   11  ,   7   ,   1   ,   27  ,   39  ,   411 ,   195 ,   607 ,   1705    ,0 };
            const unsigned long dim343Kuo3Init[]    =   {   1   ,   1   ,   7   ,   11  ,   23  ,   21  ,   11  ,   113 ,   437 ,   947 ,   1897    ,   1249    ,0 };
            const unsigned long dim344Kuo3Init[]    =   {   1   ,   1   ,   7   ,   5   ,   23  ,   45  ,   41  ,   169 ,   33  ,   981 ,   1331    ,   2013    ,0 };
            const unsigned long dim345Kuo3Init[]    =   {   1   ,   3   ,   1   ,   7   ,   11  ,   5   ,   11  ,   121 ,   239 ,   269 ,   1759    ,   829 ,0 };
            const unsigned long dim346Kuo3Init[]    =   {   1   ,   1   ,   1   ,   9   ,   31  ,   51  ,   77  ,   61  ,   19  ,   121 ,   969 ,   3627    ,0 };
            const unsigned long dim347Kuo3Init[]    =   {   1   ,   3   ,   1   ,   13  ,   23  ,   45  ,   39  ,   183 ,   21  ,   787 ,   967 ,   1911    ,0 };
            const unsigned long dim348Kuo3Init[]    =   {   1   ,   1   ,   1   ,   15  ,   21  ,   19  ,   91  ,   231 ,   325 ,   531 ,   473 ,   2543    ,0 };
            const unsigned long dim349Kuo3Init[]    =   {   1   ,   3   ,   5   ,   3   ,   7   ,   25  ,   91  ,   193 ,   387 ,   181 ,   493 ,   1225    ,0 };
            const unsigned long dim350Kuo3Init[]    =   {   1   ,   1   ,   7   ,   7   ,   23  ,   21  ,   55  ,   221 ,   283 ,   995 ,   793 ,   485 ,0 };
            const unsigned long dim351Kuo3Init[]    =   {   1   ,   1   ,   1   ,   15  ,   29  ,   61  ,   1   ,   33  ,   59  ,   725 ,   1053    ,   55  ,0 };
            const unsigned long dim352Kuo3Init[]    =   {   1   ,   3   ,   3   ,   5   ,   15  ,   43  ,   85  ,   87  ,   425 ,   39  ,   1559    ,   1403    ,0 };
            const unsigned long dim353Kuo3Init[]    =   {   1   ,   1   ,   5   ,   9   ,   3   ,   7   ,   65  ,   3   ,   177 ,   491 ,   121 ,   2755    ,0 };
            const unsigned long dim354Kuo3Init[]    =   {   1   ,   1   ,   3   ,   3   ,   15  ,   37  ,   67  ,   145 ,   295 ,   421 ,   1849    ,   2151    ,0 };
            const unsigned long dim355Kuo3Init[]    =   {   1   ,   1   ,   3   ,   15  ,   17  ,   41  ,   109 ,   125 ,   509 ,   407 ,   1927    ,   2401    ,0 };
            const unsigned long dim356Kuo3Init[]    =   {   1   ,   1   ,   7   ,   5   ,   9   ,   39  ,   69  ,   1   ,   259 ,   821 ,   1837    ,   947 ,0 };
            const unsigned long dim357Kuo3Init[]    =   {   1   ,   1   ,   7   ,   9   ,   25  ,   55  ,   127 ,   185 ,   143 ,   211 ,   1615    ,   1949    ,0 };
            const unsigned long dim358Kuo3Init[]    =   {   1   ,   3   ,   3   ,   13  ,   15  ,   25  ,   25  ,   231 ,   197 ,   329 ,   1319    ,   853 ,0 };
            const unsigned long dim359Kuo3Init[]    =   {   1   ,   3   ,   7   ,   11  ,   29  ,   1   ,   23  ,   131 ,   473 ,   497 ,   1147    ,   3779    ,0 };
            const unsigned long dim360Kuo3Init[]    =   {   1   ,   3   ,   1   ,   15  ,   21  ,   23  ,   67  ,   243 ,   191 ,   353 ,   1091    ,   2121    ,0 };
            const unsigned long dim361Kuo3Init[]    =   {   1   ,   1   ,   7   ,   11  ,   27  ,   15  ,   41  ,   183 ,   235 ,   219 ,   303 ,   795 ,0 };
            const unsigned long dim362Kuo3Init[]    =   {   1   ,   3   ,   7   ,   15  ,   3   ,   21  ,   19  ,   187 ,   365 ,   453 ,   1069    ,   1983    ,0 };
            const unsigned long dim363Kuo3Init[]    =   {   1   ,   1   ,   7   ,   1   ,   27  ,   11  ,   95  ,   117 ,   433 ,   453 ,   1839    ,   3863    ,0 };
            const unsigned long dim364Kuo3Init[]    =   {   1   ,   3   ,   3   ,   5   ,   29  ,   63  ,   17  ,   243 ,   247 ,   609 ,   1999    ,   785 ,0 };
            const unsigned long dim365Kuo3Init[]    =   {   1   ,   1   ,   1   ,   7   ,   21  ,   17  ,   57  ,   99  ,   227 ,   217 ,   585 ,   397 ,0 };
            const unsigned long dim366Kuo3Init[]    =   {   1   ,   1   ,   1   ,   13  ,   13  ,   41  ,   75  ,   153 ,   355 ,   175 ,   313 ,   2237    ,0 };
            const unsigned long dim367Kuo3Init[]    =   {   1   ,   3   ,   7   ,   9   ,   27  ,   63  ,   1   ,   33  ,   383 ,   507 ,   1315    ,   1691    ,0 };
            const unsigned long dim368Kuo3Init[]    =   {   1   ,   1   ,   3   ,   3   ,   17  ,   33  ,   65  ,   235 ,   395 ,   317 ,   1769    ,   2937    ,0 };
            const unsigned long dim369Kuo3Init[]    =   {   1   ,   3   ,   1   ,   5   ,   31  ,   19  ,   25  ,   15  ,   373 ,   823 ,   2003    ,   1309    ,0 };
            const unsigned long dim370Kuo3Init[]    =   {   1   ,   3   ,   3   ,   5   ,   3   ,   33  ,   43  ,   225 ,   445 ,   181 ,   325 ,   797 ,0 };
            const unsigned long dim371Kuo3Init[]    =   {   1   ,   1   ,   3   ,   1   ,   11  ,   19  ,   49  ,   129 ,   73  ,   983 ,   621 ,   2381    ,0 };
            const unsigned long dim372Kuo3Init[]    =   {   1   ,   3   ,   3   ,   15  ,   1   ,   33  ,   109 ,   33  ,   363 ,   925 ,   589 ,   3955    ,0 };
            const unsigned long dim373Kuo3Init[]    =   {   1   ,   1   ,   3   ,   3   ,   11  ,   21  ,   23  ,   99  ,   159 ,   949 ,   977 ,   3103    ,0 };
            const unsigned long dim374Kuo3Init[]    =   {   1   ,   1   ,   5   ,   3   ,   3   ,   15  ,   41  ,   45  ,   325 ,   703 ,   907 ,   2467    ,0 };
            const unsigned long dim375Kuo3Init[]    =   {   1   ,   1   ,   7   ,   15  ,   31  ,   11  ,   71  ,   225 ,   199 ,   521 ,   933 ,   2777    ,0 };
            const unsigned long dim376Kuo3Init[]    =   {   1   ,   3   ,   5   ,   9   ,   13  ,   55  ,   23  ,   199 ,   293 ,   297 ,   2031    ,   3815    ,0 };
            const unsigned long dim377Kuo3Init[]    =   {   1   ,   1   ,   3   ,   13  ,   25  ,   5   ,   25  ,   229 ,   301 ,   301 ,   1165    ,   2229    ,0 };
            const unsigned long dim378Kuo3Init[]    =   {   1   ,   1   ,   1   ,   11  ,   1   ,   39  ,   25  ,   211 ,   233 ,   233 ,   355 ,   3131    ,0 };
            const unsigned long dim379Kuo3Init[]    =   {   1   ,   3   ,   3   ,   9   ,   31  ,   37  ,   113 ,   17  ,   163 ,   547 ,   877 ,   333 ,0 };
            const unsigned long dim380Kuo3Init[]    =   {   1   ,   1   ,   7   ,   1   ,   31  ,   7   ,   117 ,   175 ,   191 ,   191 ,   1185    ,   2637    ,0 };
            const unsigned long dim381Kuo3Init[]    =   {   1   ,   1   ,   3   ,   11  ,   27  ,   59  ,   33  ,   79  ,   293 ,   657 ,   1899    ,   1547    ,0 };
            const unsigned long dim382Kuo3Init[]    =   {   1   ,   3   ,   1   ,   7   ,   29  ,   43  ,   89  ,   227 ,   315 ,   971 ,   1175    ,   693 ,0 };
            const unsigned long dim383Kuo3Init[]    =   {   1   ,   1   ,   1   ,   5   ,   9   ,   43  ,   17  ,   37  ,   321 ,   907 ,   441 ,   21  ,0 };
            const unsigned long dim384Kuo3Init[]    =   {   1   ,   3   ,   1   ,   15  ,   25  ,   47  ,   21  ,   11  ,   385 ,   353 ,   1103    ,   3445    ,0 };
            const unsigned long dim385Kuo3Init[]    =   {   1   ,   1   ,   7   ,   9   ,   21  ,   63  ,   9   ,   147 ,   79  ,   627 ,   471 ,   1683    ,0 };
            const unsigned long dim386Kuo3Init[]    =   {   1   ,   3   ,   7   ,   7   ,   1   ,   13  ,   29  ,   127 ,   3   ,   629 ,   1213    ,   2981    ,0 };
            const unsigned long dim387Kuo3Init[]    =   {   1   ,   1   ,   5   ,   9   ,   7   ,   41  ,   41  ,   235 ,   131 ,   379 ,   23  ,   1251    ,0 };
            const unsigned long dim388Kuo3Init[]    =   {   1   ,   3   ,   5   ,   3   ,   17  ,   31  ,   101 ,   193 ,   271 ,   189 ,   1923    ,   1849    ,0 };
            const unsigned long dim389Kuo3Init[]    =   {   1   ,   1   ,   1   ,   5   ,   21  ,   43  ,   35  ,   209 ,   359 ,   185 ,   1921    ,   2773    ,0 };
            const unsigned long dim390Kuo3Init[]    =   {   1   ,   1   ,   7   ,   5   ,   31  ,   5   ,   41  ,   127 ,   415 ,   151 ,   1027    ,   2171    ,0 };
            const unsigned long dim391Kuo3Init[]    =   {   1   ,   1   ,   1   ,   3   ,   31  ,   53  ,   127 ,   37  ,   273 ,   415 ,   1657    ,   467 ,0 };
            const unsigned long dim392Kuo3Init[]    =   {   1   ,   3   ,   7   ,   3   ,   29  ,   43  ,   77  ,   35  ,   347 ,   37  ,   1265    ,   1211    ,0 };
            const unsigned long dim393Kuo3Init[]    =   {   1   ,   1   ,   1   ,   1   ,   3   ,   31  ,   89  ,   57  ,   405 ,   491 ,   1685    ,   2727    ,0 };
            const unsigned long dim394Kuo3Init[]    =   {   1   ,   3   ,   5   ,   15  ,   29  ,   9   ,   23  ,   79  ,   253 ,   161 ,   273 ,   4001    ,0 };
            const unsigned long dim395Kuo3Init[]    =   {   1   ,   3   ,   3   ,   13  ,   29  ,   49  ,   73  ,   145 ,   203 ,   215 ,   423 ,   3293    ,0 };
            const unsigned long dim396Kuo3Init[]    =   {   1   ,   1   ,   5   ,   15  ,   29  ,   43  ,   49  ,   99  ,   335 ,   1017    ,   315 ,   2761    ,0 };
            const unsigned long dim397Kuo3Init[]    =   {   1   ,   3   ,   1   ,   9   ,   15  ,   39  ,   87  ,   243 ,   331 ,   1015    ,   1771    ,   1303    ,0 };
            const unsigned long dim398Kuo3Init[]    =   {   1   ,   1   ,   1   ,   5   ,   5   ,   1   ,   15  ,   19  ,   277 ,   699 ,   1957    ,   373 ,0 };
            const unsigned long dim399Kuo3Init[]    =   {   1   ,   1   ,   7   ,   1   ,   7   ,   13  ,   41  ,   211 ,   167 ,   705 ,   1011    ,   2097    ,0 };
            const unsigned long dim400Kuo3Init[]    =   {   1   ,   1   ,   7   ,   9   ,   3   ,   25  ,   127 ,   75  ,   271 ,   817 ,   629 ,   2027    ,0 };
            const unsigned long dim401Kuo3Init[]    =   {   1   ,   3   ,   1   ,   5   ,   31  ,   49  ,   17  ,   113 ,   71  ,   233 ,   443 ,   3005    ,0 };
            const unsigned long dim402Kuo3Init[]    =   {   1   ,   1   ,   1   ,   1   ,   1   ,   23  ,   87  ,   29  ,   121 ,   485 ,   227 ,   2981    ,0 };
            const unsigned long dim403Kuo3Init[]    =   {   1   ,   1   ,   1   ,   1   ,   19  ,   3   ,   127 ,   31  ,   57  ,   567 ,   375 ,   1121    ,0 };
            const unsigned long dim404Kuo3Init[]    =   {   1   ,   1   ,   1   ,   9   ,   15  ,   49  ,   67  ,   49  ,   473 ,   957 ,   1955    ,   29  ,0 };
            const unsigned long dim405Kuo3Init[]    =   {   1   ,   1   ,   1   ,   1   ,   15  ,   49  ,   71  ,   241 ,   43  ,   611 ,   953 ,   4013    ,0 };
            const unsigned long dim406Kuo3Init[]    =   {   1   ,   3   ,   3   ,   3   ,   13  ,   57  ,   45  ,   65  ,   5   ,   101 ,   511 ,   3317    ,0 };
            const unsigned long dim407Kuo3Init[]    =   {   1   ,   3   ,   3   ,   7   ,   25  ,   9   ,   97  ,   35  ,   495 ,   791 ,   223 ,   1555    ,0 };
            const unsigned long dim408Kuo3Init[]    =   {   1   ,   1   ,   3   ,   1   ,   25  ,   37  ,   73  ,   61  ,   31  ,   525 ,   855 ,   695 ,0 };
            const unsigned long dim409Kuo3Init[]    =   {   1   ,   3   ,   5   ,   11  ,   15  ,   45  ,   11  ,   53  ,   393 ,   557 ,   349 ,   2507    ,0 };
            const unsigned long dim410Kuo3Init[]    =   {   1   ,   1   ,   5   ,   11  ,   1   ,   43  ,   77  ,   215 ,   149 ,   81  ,   1663    ,   987 ,0 };
            const unsigned long dim411Kuo3Init[]    =   {   1   ,   3   ,   5   ,   5   ,   13  ,   55  ,   45  ,   79  ,   305 ,   65  ,   1721    ,   605 ,0 };
            const unsigned long dim412Kuo3Init[]    =   {   1   ,   1   ,   1   ,   3   ,   15  ,   19  ,   103 ,   249 ,   253 ,   935 ,   1933    ,   3809    ,0 };
            const unsigned long dim413Kuo3Init[]    =   {   1   ,   1   ,   1   ,   11  ,   1   ,   43  ,   119 ,   233 ,   79  ,   539 ,   1397    ,   1669    ,0 };
            const unsigned long dim414Kuo3Init[]    =   {   1   ,   1   ,   7   ,   15  ,   9   ,   27  ,   29  ,   125 ,   335 ,   573 ,   1721    ,   551 ,0 };
            const unsigned long dim415Kuo3Init[]    =   {   1   ,   1   ,   5   ,   11  ,   29  ,   33  ,   121 ,   47  ,   351 ,   857 ,   1155    ,   2939    ,0 };
            const unsigned long dim416Kuo3Init[]    =   {   1   ,   1   ,   1   ,   13  ,   21  ,   13  ,   107 ,   179 ,   135 ,   597 ,   39  ,   2129    ,0 };
            const unsigned long dim417Kuo3Init[]    =   {   1   ,   3   ,   5   ,   15  ,   11  ,   5   ,   101 ,   227 ,   457 ,   773 ,   1369    ,   2229    ,0 };
            const unsigned long dim418Kuo3Init[]    =   {   1   ,   1   ,   7   ,   3   ,   9   ,   47  ,   53  ,   47  ,   315 ,   221 ,   369 ,   2991    ,0 };
            const unsigned long dim419Kuo3Init[]    =   {   1   ,   3   ,   7   ,   9   ,   7   ,   1   ,   117 ,   229 ,   193 ,   105 ,   1591    ,   875 ,0 };
            const unsigned long dim420Kuo3Init[]    =   {   1   ,   3   ,   1   ,   13  ,   5   ,   49  ,   47  ,   95  ,   431 ,   37  ,   2037    ,   3793    ,0 };
            const unsigned long dim421Kuo3Init[]    =   {   1   ,   1   ,   3   ,   11  ,   17  ,   39  ,   91  ,   225 ,   109 ,   375 ,   1431    ,   3689    ,0 };
            const unsigned long dim422Kuo3Init[]    =   {   1   ,   3   ,   5   ,   7   ,   27  ,   39  ,   23  ,   215 ,   55  ,   573 ,   965 ,   321 ,0 };
            const unsigned long dim423Kuo3Init[]    =   {   1   ,   1   ,   3   ,   13  ,   1   ,   51  ,   23  ,   181 ,   399 ,   471 ,   429 ,   2409    ,0 };
            const unsigned long dim424Kuo3Init[]    =   {   1   ,   1   ,   7   ,   9   ,   17  ,   21  ,   43  ,   177 ,   327 ,   789 ,   1581    ,   3505    ,0 };
            const unsigned long dim425Kuo3Init[]    =   {   1   ,   1   ,   7   ,   7   ,   9   ,   39  ,   59  ,   187 ,   203 ,   655 ,   545 ,   2107    ,0 };
            const unsigned long dim426Kuo3Init[]    =   {   1   ,   1   ,   7   ,   3   ,   15  ,   13  ,   83  ,   189 ,   213 ,   575 ,   255 ,   1475    ,0 };
            const unsigned long dim427Kuo3Init[]    =   {   1   ,   1   ,   1   ,   3   ,   23  ,   27  ,   101 ,   67  ,   299 ,   1003    ,   1351    ,   1799    ,0 };
            const unsigned long dim428Kuo3Init[]    =   {   1   ,   3   ,   3   ,   9   ,   13  ,   29  ,   83  ,   251 ,   151 ,   607 ,   2047    ,   447 ,0 };
            const unsigned long dim429Kuo3Init[]    =   {   1   ,   3   ,   1   ,   5   ,   5   ,   7   ,   23  ,   145 ,   241 ,   437 ,   1215    ,   489 ,0 };
            const unsigned long dim430Kuo3Init[]    =   {   1   ,   3   ,   1   ,   11  ,   3   ,   63  ,   55  ,   53  ,   187 ,   179 ,   1137    ,   3923    ,0 };
            const unsigned long dim431Kuo3Init[]    =   {   1   ,   3   ,   7   ,   11  ,   31  ,   29  ,   127 ,   37  ,   509 ,   835 ,   1693    ,   965 ,0 };
            const unsigned long dim432Kuo3Init[]    =   {   1   ,   1   ,   1   ,   9   ,   29  ,   1   ,   11  ,   61  ,   459 ,   415 ,   317 ,   2981    ,0 };
            const unsigned long dim433Kuo3Init[]    =   {   1   ,   3   ,   7   ,   3   ,   1   ,   61  ,   31  ,   77  ,   55  ,   301 ,   1333    ,   2803    ,0 };
            const unsigned long dim434Kuo3Init[]    =   {   1   ,   1   ,   7   ,   13  ,   3   ,   1   ,   117 ,   185 ,   385 ,   771 ,   409 ,   3667    ,0 };
            const unsigned long dim435Kuo3Init[]    =   {   1   ,   1   ,   5   ,   13  ,   19  ,   3   ,   43  ,   53  ,   41  ,   111 ,   1617    ,   95  ,0 };
            const unsigned long dim436Kuo3Init[]    =   {   1   ,   3   ,   7   ,   13  ,   25  ,   25  ,   125 ,   91  ,   85  ,   99  ,   461 ,   3999    ,0 };
            const unsigned long dim437Kuo3Init[]    =   {   1   ,   3   ,   5   ,   7   ,   7   ,   29  ,   15  ,   155 ,   237 ,   443 ,   1187    ,   623 ,0 };
            const unsigned long dim438Kuo3Init[]    =   {   1   ,   3   ,   3   ,   1   ,   11  ,   3   ,   35  ,   169 ,   449 ,   917 ,   141 ,   1525    ,0 };
            const unsigned long dim439Kuo3Init[]    =   {   1   ,   1   ,   5   ,   7   ,   17  ,   11  ,   85  ,   31  ,   165 ,   35  ,   1069    ,   3395    ,0 };
            const unsigned long dim440Kuo3Init[]    =   {   1   ,   3   ,   7   ,   15  ,   11  ,   21  ,   89  ,   61  ,   459 ,   211 ,   959 ,   2411    ,0 };
            const unsigned long dim441Kuo3Init[]    =   {   1   ,   1   ,   5   ,   11  ,   25  ,   35  ,   1   ,   137 ,   491 ,   955 ,   1287    ,   3955    ,0 };
            const unsigned long dim442Kuo3Init[]    =   {   1   ,   3   ,   3   ,   11  ,   13  ,   47  ,   105 ,   69  ,   119 ,   455 ,   769 ,   1665    ,0 };
            const unsigned long dim443Kuo3Init[]    =   {   1   ,   3   ,   3   ,   13  ,   17  ,   9   ,   27  ,   25  ,   445 ,   611 ,   855 ,   1273    ,0 };
            const unsigned long dim444Kuo3Init[]    =   {   1   ,   1   ,   3   ,   9   ,   3   ,   5   ,   107 ,   89  ,   443 ,   991 ,   727 ,   871 ,0 };
            const unsigned long dim445Kuo3Init[]    =   {   1   ,   1   ,   5   ,   3   ,   19  ,   43  ,   93  ,   163 ,   193 ,   395 ,   99  ,   1717    ,0 };
            const unsigned long dim446Kuo3Init[]    =   {   1   ,   1   ,   1   ,   7   ,   23  ,   43  ,   121 ,   205 ,   249 ,   255 ,   1763    ,   1867    ,0 };
            const unsigned long dim447Kuo3Init[]    =   {   1   ,   1   ,   3   ,   3   ,   1   ,   45  ,   33  ,   191 ,   475 ,   503 ,   761 ,   1845    ,0 };
            const unsigned long dim448Kuo3Init[]    =   {   1   ,   1   ,   3   ,   3   ,   31  ,   39  ,   11  ,   203 ,   9   ,   183 ,   407 ,   3787    ,0 };
            const unsigned long dim449Kuo3Init[]    =   {   1   ,   1   ,   5   ,   11  ,   11  ,   3   ,   65  ,   175 ,   131 ,   757 ,   1717    ,   1003    ,0 };
            const unsigned long dim450Kuo3Init[]    =   {   1   ,   1   ,   1   ,   15  ,   13  ,   61  ,   29  ,   123 ,   269 ,   531 ,   257 ,   1399    ,0 };
            const unsigned long dim451Kuo3Init[]    =   {   1   ,   3   ,   5   ,   5   ,   13  ,   23  ,   23  ,   1   ,   293 ,   921 ,   79  ,   479 ,0 };
            const unsigned long dim452Kuo3Init[]    =   {   1   ,   1   ,   3   ,   13  ,   9   ,   55  ,   21  ,   15  ,   47  ,   685 ,   829 ,   2679    ,0 };
            const unsigned long dim453Kuo3Init[]    =   {   1   ,   3   ,   5   ,   11  ,   1   ,   39  ,   97  ,   59  ,   37  ,   787 ,   991 ,   263 ,0 };
            const unsigned long dim454Kuo3Init[]    =   {   1   ,   3   ,   5   ,   15  ,   31  ,   3   ,   45  ,   159 ,   219 ,   161 ,   1025    ,   665 ,0 };
            const unsigned long dim455Kuo3Init[]    =   {   1   ,   3   ,   7   ,   15  ,   11  ,   51  ,   51  ,   19  ,   181 ,   125 ,   1677    ,   2125    ,0 };
            const unsigned long dim456Kuo3Init[]    =   {   1   ,   1   ,   3   ,   5   ,   27  ,   23  ,   79  ,   161 ,   97  ,   581 ,   405 ,   2239    ,0 };
            const unsigned long dim457Kuo3Init[]    =   {   1   ,   3   ,   1   ,   11  ,   17  ,   1   ,   97  ,   33  ,   31  ,   771 ,   707 ,   1053    ,0 };
            const unsigned long dim458Kuo3Init[]    =   {   1   ,   1   ,   1   ,   11  ,   11  ,   51  ,   113 ,   47  ,   51  ,   827 ,   385 ,   3211    ,0 };
            const unsigned long dim459Kuo3Init[]    =   {   1   ,   3   ,   3   ,   7   ,   17  ,   3   ,   99  ,   57  ,   59  ,   199 ,   1359    ,   2459    ,0 };
            const unsigned long dim460Kuo3Init[]    =   {   1   ,   1   ,   5   ,   11  ,   1   ,   7   ,   19  ,   37  ,   53  ,   435 ,   1095    ,   495 ,0 };
            const unsigned long dim461Kuo3Init[]    =   {   1   ,   1   ,   5   ,   13  ,   19  ,   5   ,   95  ,   109 ,   245 ,   737 ,   1163    ,   33  ,0 };
            const unsigned long dim462Kuo3Init[]    =   {   1   ,   3   ,   5   ,   3   ,   1   ,   17  ,   79  ,   195 ,   153 ,   857 ,   603 ,   2905    ,0 };
            const unsigned long dim463Kuo3Init[]    =   {   1   ,   3   ,   5   ,   5   ,   31  ,   51  ,   71  ,   103 ,   121 ,   5   ,   931 ,   217 ,0 };
            const unsigned long dim464Kuo3Init[]    =   {   1   ,   1   ,   5   ,   7   ,   31  ,   39  ,   11  ,   9   ,   223 ,   365 ,   1317    ,   449 ,0 };
            const unsigned long dim465Kuo3Init[]    =   {   1   ,   3   ,   5   ,   3   ,   17  ,   19  ,   85  ,   121 ,   161 ,   75  ,   527 ,   3941    ,0 };
            const unsigned long dim466Kuo3Init[]    =   {   1   ,   3   ,   5   ,   9   ,   17  ,   53  ,   41  ,   255 ,   207 ,   251 ,   925 ,   2395    ,0 };
            const unsigned long dim467Kuo3Init[]    =   {   1   ,   3   ,   3   ,   3   ,   13  ,   5   ,   109 ,   125 ,   293 ,   165 ,   1715    ,   1017    ,0 };
            const unsigned long dim468Kuo3Init[]    =   {   1   ,   3   ,   7   ,   1   ,   23  ,   27  ,   45  ,   141 ,   267 ,   187 ,   611 ,   59  ,0 };
            const unsigned long dim469Kuo3Init[]    =   {   1   ,   3   ,   1   ,   9   ,   11  ,   9   ,   81  ,   239 ,   469 ,   905 ,   525 ,   1549    ,0 };
            const unsigned long dim470Kuo3Init[]    =   {   1   ,   1   ,   1   ,   15  ,   13  ,   21  ,   71  ,   87  ,   199 ,   389 ,   139 ,   3499    ,0 };
            const unsigned long dim471Kuo3Init[]    =   {   1   ,   3   ,   7   ,   5   ,   21  ,   41  ,   107 ,   195 ,   369 ,   391 ,   271 ,   3351    ,0 };
            const unsigned long dim472Kuo3Init[]    =   {   1   ,   3   ,   3   ,   1   ,   17  ,   19  ,   75  ,   219 ,   415 ,   241 ,   1441    ,   967 ,0 };
            const unsigned long dim473Kuo3Init[]    =   {   1   ,   1   ,   1   ,   15  ,   7   ,   7   ,   79  ,   213 ,   391 ,   831 ,   1649    ,   3741    ,0 };
            const unsigned long dim474Kuo3Init[]    =   {   1   ,   3   ,   5   ,   3   ,   29  ,   57  ,   99  ,   207 ,   243 ,   915 ,   1969    ,   2499    ,0 };
            const unsigned long dim475Kuo3Init[]    =   {   1   ,   1   ,   3   ,   13  ,   1   ,   3   ,   27  ,   211 ,   271 ,   657 ,   1251    ,   3333    ,0 };
            const unsigned long dim476Kuo3Init[]    =   {   1   ,   3   ,   3   ,   1   ,   3   ,   31  ,   49  ,   53  ,   273 ,   837 ,   1771    ,   1405    ,0 };
            const unsigned long dim477Kuo3Init[]    =   {   1   ,   3   ,   7   ,   7   ,   1   ,   33  ,   125 ,   221 ,   179 ,   601 ,   565 ,   2709    ,0 };
            const unsigned long dim478Kuo3Init[]    =   {   1   ,   1   ,   1   ,   7   ,   31  ,   19  ,   81  ,   173 ,   413 ,   849 ,   579 ,   2487    ,0 };
            const unsigned long dim479Kuo3Init[]    =   {   1   ,   1   ,   7   ,   11  ,   17  ,   23  ,   97  ,   145 ,   383 ,   1015    ,   547 ,   483 ,0 };
            const unsigned long dim480Kuo3Init[]    =   {   1   ,   3   ,   1   ,   3   ,   11  ,   63  ,   41  ,   175 ,   43  ,   283 ,   1441    ,   2695    ,0 };
            const unsigned long dim481Kuo3Init[]    =   {   1   ,   1   ,   7   ,   9   ,   5   ,   23  ,   61  ,   161 ,   149 ,   49  ,   1851    ,   1281    ,   4615    ,0 };
            const unsigned long dim482Kuo3Init[]    =   {   1   ,   1   ,   5   ,   7   ,   23  ,   47  ,   111 ,   213 ,   317 ,   107 ,   849 ,   1563    ,   6863    ,0 };
            const unsigned long dim483Kuo3Init[]    =   {   1   ,   1   ,   7   ,   9   ,   31  ,   53  ,   115 ,   233 ,   497 ,   23  ,   1395    ,   327 ,   551 ,0 };
            const unsigned long dim484Kuo3Init[]    =   {   1   ,   1   ,   7   ,   13  ,   21  ,   51  ,   17  ,   109 ,   29  ,   1003    ,   1221    ,   2175    ,   4645    ,0 };
            const unsigned long dim485Kuo3Init[]    =   {   1   ,   3   ,   3   ,   15  ,   7   ,   57  ,   103 ,   131 ,   63  ,   217 ,   607 ,   2311    ,   1237    ,0 };
            const unsigned long dim486Kuo3Init[]    =   {   1   ,   3   ,   5   ,   3   ,   19  ,   27  ,   107 ,   223 ,   15  ,   149 ,   1777    ,   2211    ,   2241    ,0 };
            const unsigned long dim487Kuo3Init[]    =   {   1   ,   1   ,   1   ,   1   ,   5   ,   37  ,   87  ,   59  ,   435 ,   37  ,   1771    ,   1675    ,   6241    ,0 };
            const unsigned long dim488Kuo3Init[]    =   {   1   ,   3   ,   1   ,   7   ,   29  ,   51  ,   83  ,   41  ,   11  ,   545 ,   995 ,   1625    ,   4883    ,0 };
            const unsigned long dim489Kuo3Init[]    =   {   1   ,   1   ,   1   ,   5   ,   15  ,   43  ,   57  ,   111 ,   55  ,   921 ,   567 ,   1767    ,   1849    ,0 };
            const unsigned long dim490Kuo3Init[]    =   {   1   ,   1   ,   5   ,   15  ,   5   ,   49  ,   3   ,   23  ,   307 ,   691 ,   1727    ,   2297    ,   6705    ,0 };
            const unsigned long dim491Kuo3Init[]    =   {   1   ,   3   ,   7   ,   5   ,   31  ,   1   ,   111 ,   53  ,   499 ,   1017    ,   1889    ,   1955    ,   1327    ,0 };
            const unsigned long dim492Kuo3Init[]    =   {   1   ,   3   ,   7   ,   11  ,   17  ,   35  ,   71  ,   223 ,   149 ,   743 ,   319 ,   3157    ,   5961    ,0 };
            const unsigned long dim493Kuo3Init[]    =   {   1   ,   1   ,   1   ,   3   ,   21  ,   31  ,   27  ,   79  ,   353 ,   753 ,   341 ,   2121    ,   5455    ,0 };
            const unsigned long dim494Kuo3Init[]    =   {   1   ,   1   ,   5   ,   3   ,   21  ,   29  ,   27  ,   91  ,   111 ,   359 ,   27  ,   2383    ,   7149    ,0 };
            const unsigned long dim495Kuo3Init[]    =   {   1   ,   3   ,   7   ,   9   ,   5   ,   51  ,   115 ,   65  ,   275 ,   211 ,   1519    ,   2247    ,   2065    ,0 };
            const unsigned long dim496Kuo3Init[]    =   {   1   ,   1   ,   1   ,   15  ,   27  ,   7   ,   21  ,   173 ,   79  ,   357 ,   693 ,   873 ,   879 ,0 };
            const unsigned long dim497Kuo3Init[]    =   {   1   ,   3   ,   1   ,   7   ,   25  ,   47  ,   53  ,   123 ,   491 ,   707 ,   1723    ,   1139    ,   3757    ,0 };
            const unsigned long dim498Kuo3Init[]    =   {   1   ,   1   ,   7   ,   5   ,   15  ,   55  ,   103 ,   17  ,   17  ,   345 ,   1477    ,   3423    ,   7527    ,0 };
            const unsigned long dim499Kuo3Init[]    =   {   1   ,   3   ,   1   ,   13  ,   1   ,   3   ,   55  ,   85  ,   251 ,   389 ,   631 ,   117 ,   4879    ,0 };
            const unsigned long dim500Kuo3Init[]    =   {   1   ,   1   ,   7   ,   1   ,   29  ,   33  ,   17  ,   3   ,   303 ,   299 ,   465 ,   1709    ,   1111    ,0 };
            const unsigned long dim501Kuo3Init[]    =   {   1   ,   1   ,   7   ,   5   ,   3   ,   29  ,   111 ,   249 ,   383 ,   611 ,   859 ,   1555    ,   7387    ,0 };
            const unsigned long dim502Kuo3Init[]    =   {   1   ,   3   ,   5   ,   5   ,   29  ,   41  ,   33  ,   63  ,   395 ,   53  ,   351 ,   247 ,   5317    ,0 };
            const unsigned long dim503Kuo3Init[]    =   {   1   ,   1   ,   5   ,   7   ,   11  ,   63  ,   121 ,   131 ,   249 ,   451 ,   619 ,   2395    ,   4575    ,0 };
            const unsigned long dim504Kuo3Init[]    =   {   1   ,   3   ,   7   ,   3   ,   3   ,   43  ,   37  ,   211 ,   385 ,   289 ,   791 ,   735 ,   4683    ,0 };
            const unsigned long dim505Kuo3Init[]    =   {   1   ,   1   ,   3   ,   11  ,   7   ,   13  ,   53  ,   7   ,   497 ,   771 ,   903 ,   119 ,   6601    ,0 };
            const unsigned long dim506Kuo3Init[]    =   {   1   ,   3   ,   7   ,   1   ,   3   ,   61  ,   117 ,   237 ,   427 ,   335 ,   1749    ,   3105    ,   6493    ,0 };
            const unsigned long dim507Kuo3Init[]    =   {   1   ,   3   ,   5   ,   7   ,   11  ,   55  ,   121 ,   93  ,   3   ,   953 ,   433 ,   2663    ,   1633    ,0 };
            const unsigned long dim508Kuo3Init[]    =   {   1   ,   1   ,   7   ,   1   ,   7   ,   51  ,   43  ,   31  ,   31  ,   819 ,   791 ,   1071    ,   569 ,0 };
            const unsigned long dim509Kuo3Init[]    =   {   1   ,   3   ,   5   ,   11  ,   17  ,   37  ,   71  ,   97  ,   69  ,   681 ,   1133    ,   2425    ,   3217    ,0 };
            const unsigned long dim510Kuo3Init[]    =   {   1   ,   1   ,   1   ,   15  ,   9   ,   15  ,   31  ,   131 ,   113 ,   567 ,   593 ,   1781    ,   6715    ,0 };
            const unsigned long dim511Kuo3Init[]    =   {   1   ,   1   ,   7   ,   9   ,   29  ,   25  ,   39  ,   159 ,   495 ,   507 ,   1253    ,   409 ,   6723    ,0 };
            const unsigned long dim512Kuo3Init[]    =   {   1   ,   1   ,   1   ,   5   ,   13  ,   29  ,   25  ,   197 ,   489 ,   853 ,   87  ,   3173    ,   6389    ,0 };
            const unsigned long dim513Kuo3Init[]    =   {   1   ,   1   ,   7   ,   11  ,   11  ,   7   ,   27  ,   227 ,   347 ,   777 ,   893 ,   1007    ,   1953    ,0 };
            const unsigned long dim514Kuo3Init[]    =   {   1   ,   1   ,   3   ,   7   ,   31  ,   9   ,   61  ,   39  ,   251 ,   1   ,   1297    ,   2861    ,   2593    ,0 };
            const unsigned long dim515Kuo3Init[]    =   {   1   ,   1   ,   5   ,   1   ,   15  ,   45  ,   3   ,   147 ,   143 ,   279 ,   2045    ,   649 ,   5579    ,0 };
            const unsigned long dim516Kuo3Init[]    =   {   1   ,   1   ,   5   ,   5   ,   9   ,   51  ,   3   ,   249 ,   253 ,   533 ,   1305    ,   1749    ,   1287    ,0 };
            const unsigned long dim517Kuo3Init[]    =   {   1   ,   3   ,   7   ,   9   ,   13  ,   15  ,   85  ,   165 ,   183 ,   943 ,   41  ,   4061    ,   5845    ,0 };
            const unsigned long dim518Kuo3Init[]    =   {   1   ,   3   ,   3   ,   1   ,   23  ,   31  ,   49  ,   247 ,   287 ,   83  ,   1833    ,   1249    ,   2789    ,0 };
            const unsigned long dim519Kuo3Init[]    =   {   1   ,   3   ,   3   ,   7   ,   7   ,   53  ,   125 ,   211 ,   145 ,   837 ,   515 ,   1971    ,   6201    ,0 };
            const unsigned long dim520Kuo3Init[]    =   {   1   ,   3   ,   3   ,   11  ,   13  ,   11  ,   87  ,   11  ,   33  ,   355 ,   1049    ,   3053    ,   4801    ,0 };
            const unsigned long dim521Kuo3Init[]    =   {   1   ,   3   ,   5   ,   13  ,   23  ,   25  ,   63  ,   221 ,   385 ,   217 ,   687 ,   455 ,   4083    ,0 };
            const unsigned long dim522Kuo3Init[]    =   {   1   ,   1   ,   5   ,   1   ,   3   ,   21  ,   119 ,   79  ,   207 ,   693 ,   925 ,   1519    ,   3981    ,0 };
            const unsigned long dim523Kuo3Init[]    =   {   1   ,   1   ,   3   ,   11  ,   21  ,   35  ,   21  ,   21  ,   297 ,   461 ,   793 ,   1679    ,   1285    ,0 };
            const unsigned long dim524Kuo3Init[]    =   {   1   ,   3   ,   5   ,   13  ,   31  ,   23  ,   11  ,   129 ,   189 ,   933 ,   283 ,   3435    ,   4593    ,0 };
            const unsigned long dim525Kuo3Init[]    =   {   1   ,   3   ,   1   ,   1   ,   13  ,   11  ,   113 ,   35  ,   199 ,   149 ,   181 ,   3183    ,   6205    ,0 };
            const unsigned long dim526Kuo3Init[]    =   {   1   ,   3   ,   1   ,   9   ,   7   ,   59  ,   105 ,   161 ,   63  ,   821 ,   13  ,   2561    ,   755 ,0 };
            const unsigned long dim527Kuo3Init[]    =   {   1   ,   1   ,   1   ,   7   ,   27  ,   3   ,   87  ,   243 ,   279 ,   13  ,   1341    ,   2383    ,   6005    ,0 };
            const unsigned long dim528Kuo3Init[]    =   {   1   ,   3   ,   3   ,   13  ,   27  ,   25  ,   91  ,   205 ,   275 ,   663 ,   1875    ,   2399    ,   4179    ,0 };
            const unsigned long dim529Kuo3Init[]    =   {   1   ,   1   ,   7   ,   15  ,   27  ,   25  ,   51  ,   235 ,   391 ,   951 ,   759 ,   285 ,   3443    ,0 };
            const unsigned long dim530Kuo3Init[]    =   {   1   ,   3   ,   3   ,   5   ,   1   ,   33  ,   57  ,   107 ,   137 ,   851 ,   1937    ,   1003    ,   2427    ,0 };
            const unsigned long dim531Kuo3Init[]    =   {   1   ,   1   ,   5   ,   11  ,   27  ,   3   ,   27  ,   81  ,   303 ,   151 ,   433 ,   1893    ,   587 ,0 };
            const unsigned long dim532Kuo3Init[]    =   {   1   ,   1   ,   1   ,   9   ,   25  ,   45  ,   113 ,   173 ,   471 ,   191 ,   1031    ,   2119    ,   4853    ,0 };
            const unsigned long dim533Kuo3Init[]    =   {   1   ,   1   ,   7   ,   11  ,   27  ,   23  ,   107 ,   229 ,   485 ,   197 ,   1925    ,   1809    ,   7775    ,0 };
            const unsigned long dim534Kuo3Init[]    =   {   1   ,   1   ,   1   ,   3   ,   31  ,   61  ,   91  ,   47  ,   3   ,   643 ,   1487    ,   2563    ,   3893    ,0 };
            const unsigned long dim535Kuo3Init[]    =   {   1   ,   1   ,   7   ,   15  ,   19  ,   27  ,   97  ,   109 ,   283 ,   67  ,   335 ,   3707    ,   1909    ,0 };
            const unsigned long dim536Kuo3Init[]    =   {   1   ,   1   ,   1   ,   11  ,   23  ,   59  ,   101 ,   249 ,   121 ,   919 ,   1385    ,   2119    ,   7389    ,0 };
            const unsigned long dim537Kuo3Init[]    =   {   1   ,   1   ,   7   ,   11  ,   7   ,   29  ,   33  ,   111 ,   243 ,   617 ,   125 ,   3887    ,   727 ,0 };
            const unsigned long dim538Kuo3Init[]    =   {   1   ,   1   ,   3   ,   1   ,   7   ,   41  ,   91  ,   177 ,   315 ,   555 ,   1653    ,   1725    ,   8069    ,0 };
            const unsigned long dim539Kuo3Init[]    =   {   1   ,   1   ,   5   ,   15  ,   3   ,   63  ,   53  ,   31  ,   177 ,   451 ,   141 ,   3727    ,   1695    ,0 };
            const unsigned long dim540Kuo3Init[]    =   {   1   ,   1   ,   3   ,   15  ,   25  ,   37  ,   17  ,   229 ,   75  ,   705 ,   327 ,   1367    ,   807 ,0 };
            const unsigned long dim541Kuo3Init[]    =   {   1   ,   3   ,   5   ,   5   ,   27  ,   1   ,   35  ,   97  ,   171 ,   167 ,   1077    ,   25  ,   5497    ,0 };
            const unsigned long dim542Kuo3Init[]    =   {   1   ,   1   ,   7   ,   7   ,   11  ,   17  ,   105 ,   41  ,   53  ,   389 ,   941 ,   3273    ,   967 ,0 };
            const unsigned long dim543Kuo3Init[]    =   {   1   ,   3   ,   3   ,   15  ,   11  ,   61  ,   125 ,   155 ,   361 ,   229 ,   1981    ,   1647    ,   5077    ,0 };
            const unsigned long dim544Kuo3Init[]    =   {   1   ,   3   ,   1   ,   15  ,   5   ,   11  ,   35  ,   203 ,   425 ,   509 ,   993 ,   1861    ,   1977    ,0 };
            const unsigned long dim545Kuo3Init[]    =   {   1   ,   3   ,   5   ,   15  ,   25  ,   7   ,   55  ,   89  ,   101 ,   939 ,   7   ,   3535    ,   7543    ,0 };
            const unsigned long dim546Kuo3Init[]    =   {   1   ,   1   ,   3   ,   1   ,   19  ,   45  ,   117 ,   249 ,   261 ,   529 ,   1299    ,   597 ,   7207    ,0 };
            const unsigned long dim547Kuo3Init[]    =   {   1   ,   3   ,   5   ,   13  ,   17  ,   47  ,   73  ,   231 ,   353 ,   941 ,   325 ,   115 ,   4005    ,0 };
            const unsigned long dim548Kuo3Init[]    =   {   1   ,   3   ,   1   ,   9   ,   5   ,   29  ,   11  ,   161 ,   461 ,   793 ,   813 ,   3755    ,   7017    ,0 };
            const unsigned long dim549Kuo3Init[]    =   {   1   ,   3   ,   3   ,   15  ,   23  ,   21  ,   83  ,   47  ,   143 ,   279 ,   1009    ,   3315    ,   3801    ,0 };
            const unsigned long dim550Kuo3Init[]    =   {   1   ,   1   ,   7   ,   9   ,   21  ,   55  ,   31  ,   129 ,   325 ,   59  ,   701 ,   2227    ,   1491    ,0 };
            const unsigned long dim551Kuo3Init[]    =   {   1   ,   1   ,   7   ,   1   ,   21  ,   41  ,   87  ,   151 ,   269 ,   885 ,   795 ,   4087    ,   2919    ,0 };
            const unsigned long dim552Kuo3Init[]    =   {   1   ,   1   ,   3   ,   11  ,   7   ,   19  ,   103 ,   45  ,   461 ,   109 ,   1921    ,   2497    ,   4969    ,0 };
            const unsigned long dim553Kuo3Init[]    =   {   1   ,   1   ,   3   ,   7   ,   7   ,   49  ,   9   ,   185 ,   95  ,   277 ,   1325    ,   2393    ,   7585    ,0 };
            const unsigned long dim554Kuo3Init[]    =   {   1   ,   3   ,   3   ,   13  ,   13  ,   1   ,   71  ,   77  ,   245 ,   865 ,   1693    ,   2187    ,   2589    ,0 };
            const unsigned long dim555Kuo3Init[]    =   {   1   ,   3   ,   1   ,   15  ,   19  ,   59  ,   17  ,   189 ,   101 ,   511 ,   23  ,   677 ,   4091    ,0 };
            const unsigned long dim556Kuo3Init[]    =   {   1   ,   3   ,   5   ,   7   ,   9   ,   45  ,   45  ,   233 ,   47  ,   973 ,   795 ,   2339    ,   595 ,0 };
            const unsigned long dim557Kuo3Init[]    =   {   1   ,   3   ,   1   ,   7   ,   15  ,   35  ,   79  ,   53  ,   429 ,   299 ,   389 ,   237 ,   2261    ,0 };
            const unsigned long dim558Kuo3Init[]    =   {   1   ,   3   ,   1   ,   1   ,   9   ,   35  ,   27  ,   169 ,   85  ,   755 ,   145 ,   91  ,   257 ,0 };
            const unsigned long dim559Kuo3Init[]    =   {   1   ,   3   ,   7   ,   11  ,   25  ,   33  ,   41  ,   151 ,   11  ,   401 ,   1335    ,   2111    ,   3309    ,0 };
            const unsigned long dim560Kuo3Init[]    =   {   1   ,   1   ,   7   ,   3   ,   17  ,   1   ,   61  ,   253 ,   289 ,   917 ,   1157    ,   1513    ,   4783    ,0 };
            const unsigned long dim561Kuo3Init[]    =   {   1   ,   1   ,   5   ,   7   ,   23  ,   51  ,   59  ,   171 ,   161 ,   815 ,   969 ,   1199    ,   4137    ,0 };
            const unsigned long dim562Kuo3Init[]    =   {   1   ,   3   ,   3   ,   7   ,   7   ,   27  ,   81  ,   175 ,   89  ,   145 ,   19  ,   1469    ,   3775    ,0 };
            const unsigned long dim563Kuo3Init[]    =   {   1   ,   3   ,   1   ,   7   ,   7   ,   33  ,   95  ,   179 ,   177 ,   375 ,   1077    ,   3711    ,   2903    ,0 };
            const unsigned long dim564Kuo3Init[]    =   {   1   ,   3   ,   1   ,   15  ,   15  ,   27  ,   127 ,   245 ,   307 ,   949 ,   1037    ,   3753    ,   4993    ,0 };
            const unsigned long dim565Kuo3Init[]    =   {   1   ,   3   ,   3   ,   7   ,   29  ,   15  ,   59  ,   39  ,   9   ,   331 ,   1877    ,   3565    ,   6483    ,0 };
            const unsigned long dim566Kuo3Init[]    =   {   1   ,   1   ,   5   ,   3   ,   1   ,   41  ,   29  ,   185 ,   351 ,   837 ,   747 ,   3263    ,   749 ,0 };
            const unsigned long dim567Kuo3Init[]    =   {   1   ,   1   ,   7   ,   1   ,   19  ,   47  ,   1   ,   239 ,   489 ,   79  ,   767 ,   1123    ,   8145    ,0 };
            const unsigned long dim568Kuo3Init[]    =   {   1   ,   3   ,   7   ,   15  ,   7   ,   39  ,   23  ,   239 ,   465 ,   981 ,   185 ,   837 ,   7837    ,0 };
            const unsigned long dim569Kuo3Init[]    =   {   1   ,   3   ,   5   ,   5   ,   23  ,   15  ,   21  ,   221 ,   259 ,   663 ,   185 ,   1203    ,   2857    ,0 };
            const unsigned long dim570Kuo3Init[]    =   {   1   ,   1   ,   3   ,   5   ,   31  ,   43  ,   107 ,   199 ,   117 ,   161 ,   403 ,   2227    ,   7969    ,0 };
            const unsigned long dim571Kuo3Init[]    =   {   1   ,   1   ,   3   ,   5   ,   5   ,   5   ,   105 ,   153 ,   349 ,   601 ,   815 ,   939 ,   3547    ,0 };
            const unsigned long dim572Kuo3Init[]    =   {   1   ,   1   ,   7   ,   15  ,   23  ,   33  ,   85  ,   51  ,   309 ,   63  ,   1799    ,   1599    ,   1711    ,0 };
            const unsigned long dim573Kuo3Init[]    =   {   1   ,   1   ,   7   ,   3   ,   15  ,   41  ,   71  ,   169 ,   87  ,   381 ,   1695    ,   3189    ,   3343    ,0 };
            const unsigned long dim574Kuo3Init[]    =   {   1   ,   1   ,   7   ,   11  ,   15  ,   57  ,   81  ,   107 ,   57  ,   773 ,   111 ,   2741    ,   6399    ,0 };
            const unsigned long dim575Kuo3Init[]    =   {   1   ,   1   ,   5   ,   15  ,   5   ,   55  ,   65  ,   35  ,   387 ,   797 ,   1093    ,   4039    ,   3041    ,0 };
            const unsigned long dim576Kuo3Init[]    =   {   1   ,   1   ,   5   ,   3   ,   29  ,   5   ,   87  ,   235 ,   503 ,   539 ,   483 ,   2785    ,   4211    ,0 };
            const unsigned long dim577Kuo3Init[]    =   {   1   ,   1   ,   3   ,   5   ,   11  ,   53  ,   101 ,   83  ,   61  ,   97  ,   411 ,   2063    ,   2951    ,0 };
            const unsigned long dim578Kuo3Init[]    =   {   1   ,   1   ,   1   ,   5   ,   25  ,   33  ,   45  ,   253 ,   183 ,   251 ,   1323    ,   1891    ,   5143    ,0 };
            const unsigned long dim579Kuo3Init[]    =   {   1   ,   1   ,   5   ,   11  ,   7   ,   61  ,   67  ,   201 ,   465 ,   851 ,   1   ,   1533    ,   2839    ,0 };
            const unsigned long dim580Kuo3Init[]    =   {   1   ,   3   ,   3   ,   7   ,   15  ,   3   ,   51  ,   9   ,   425 ,   671 ,   223 ,   2203    ,   481 ,0 };
            const unsigned long dim581Kuo3Init[]    =   {   1   ,   1   ,   1   ,   7   ,   25  ,   5   ,   45  ,   153 ,   479 ,   437 ,   221 ,   1085    ,   455 ,0 };
            const unsigned long dim582Kuo3Init[]    =   {   1   ,   3   ,   7   ,   1   ,   7   ,   53  ,   65  ,   35  ,   153 ,   423 ,   989 ,   2779    ,   4787    ,0 };
            const unsigned long dim583Kuo3Init[]    =   {   1   ,   3   ,   1   ,   3   ,   29  ,   43  ,   7   ,   219 ,   47  ,   431 ,   627 ,   3735    ,   383 ,0 };
            const unsigned long dim584Kuo3Init[]    =   {   1   ,   1   ,   5   ,   13  ,   31  ,   49  ,   121 ,   29  ,   235 ,   325 ,   1247    ,   1765    ,   7173    ,0 };
            const unsigned long dim585Kuo3Init[]    =   {   1   ,   1   ,   1   ,   9   ,   25  ,   15  ,   73  ,   167 ,   191 ,   215 ,   259 ,   111 ,   2289    ,0 };
            const unsigned long dim586Kuo3Init[]    =   {   1   ,   3   ,   7   ,   9   ,   23  ,   1   ,   119 ,   15  ,   155 ,   153 ,   1419    ,   667 ,   6309    ,0 };
            const unsigned long dim587Kuo3Init[]    =   {   1   ,   1   ,   5   ,   13  ,   7   ,   53  ,   91  ,   161 ,   445 ,   283 ,   571 ,   3623    ,   5737    ,0 };
            const unsigned long dim588Kuo3Init[]    =   {   1   ,   3   ,   3   ,   5   ,   3   ,   39  ,   29  ,   47  ,   393 ,   871 ,   345 ,   805 ,   8167    ,0 };
            const unsigned long dim589Kuo3Init[]    =   {   1   ,   3   ,   1   ,   5   ,   21  ,   9   ,   1   ,   169 ,   471 ,   693 ,   53  ,   967 ,   4145    ,0 };
            const unsigned long dim590Kuo3Init[]    =   {   1   ,   3   ,   7   ,   11  ,   17  ,   19  ,   123 ,   133 ,   511 ,   15  ,   675 ,   3935    ,   5465    ,0 };
            const unsigned long dim591Kuo3Init[]    =   {   1   ,   3   ,   5   ,   13  ,   25  ,   51  ,   63  ,   87  ,   503 ,   61  ,   907 ,   105 ,   7599    ,0 };
            const unsigned long dim592Kuo3Init[]    =   {   1   ,   3   ,   5   ,   13  ,   27  ,   3   ,   67  ,   189 ,   193 ,   131 ,   1227    ,   227 ,   1975    ,0 };
            const unsigned long dim593Kuo3Init[]    =   {   1   ,   3   ,   5   ,   5   ,   31  ,   17  ,   89  ,   187 ,   391 ,   575 ,   1715    ,   2551    ,   5371    ,0 };
            const unsigned long dim594Kuo3Init[]    =   {   1   ,   3   ,   1   ,   9   ,   11  ,   63  ,   9   ,   251 ,   399 ,   703 ,   1557    ,   107 ,   675 ,0 };
            const unsigned long dim595Kuo3Init[]    =   {   1   ,   3   ,   5   ,   15  ,   29  ,   55  ,   15  ,   65  ,   51  ,   811 ,   1387    ,   3815    ,   7973    ,0 };
            const unsigned long dim596Kuo3Init[]    =   {   1   ,   3   ,   1   ,   13  ,   19  ,   33  ,   87  ,   7   ,   183 ,   97  ,   2011    ,   3849    ,   3015    ,0 };
            const unsigned long dim597Kuo3Init[]    =   {   1   ,   1   ,   3   ,   11  ,   29  ,   57  ,   27  ,   185 ,   457 ,   181 ,   1341    ,   2033    ,   2181    ,0 };
            const unsigned long dim598Kuo3Init[]    =   {   1   ,   3   ,   7   ,   5   ,   25  ,   19  ,   19  ,   147 ,   101 ,   73  ,   499 ,   897 ,   3053    ,0 };
            const unsigned long dim599Kuo3Init[]    =   {   1   ,   3   ,   5   ,   5   ,   29  ,   29  ,   109 ,   5   ,   197 ,   581 ,   1829    ,   3679    ,   283 ,0 };
            const unsigned long dim600Kuo3Init[]    =   {   1   ,   3   ,   5   ,   1   ,   27  ,   63  ,   71  ,   131 ,   309 ,   193 ,   1663    ,   3009    ,   1041    ,0 };
            const unsigned long dim601Kuo3Init[]    =   {   1   ,   3   ,   3   ,   5   ,   19  ,   29  ,   95  ,   173 ,   195 ,   205 ,   207 ,   3597    ,   2223    ,0 };
            const unsigned long dim602Kuo3Init[]    =   {   1   ,   1   ,   5   ,   1   ,   3   ,   45  ,   127 ,   251 ,   175 ,   663 ,   415 ,   1789    ,   6577    ,0 };
            const unsigned long dim603Kuo3Init[]    =   {   1   ,   3   ,   1   ,   9   ,   1   ,   51  ,   73  ,   165 ,   377 ,   525 ,   381 ,   2829    ,   4619    ,0 };
            const unsigned long dim604Kuo3Init[]    =   {   1   ,   3   ,   1   ,   15  ,   23  ,   19  ,   79  ,   115 ,   297 ,   75  ,   1505    ,   1407    ,   5865    ,0 };
            const unsigned long dim605Kuo3Init[]    =   {   1   ,   1   ,   1   ,   5   ,   27  ,   33  ,   93  ,   123 ,   449 ,   747 ,   2027    ,   51  ,   4507    ,0 };
            const unsigned long dim606Kuo3Init[]    =   {   1   ,   3   ,   1   ,   7   ,   13  ,   39  ,   49  ,   111 ,   101 ,   247 ,   1375    ,   3577    ,   6769    ,0 };
            const unsigned long dim607Kuo3Init[]    =   {   1   ,   1   ,   7   ,   9   ,   19  ,   49  ,   35  ,   139 ,   57  ,   527 ,   1943    ,   1781    ,   1839    ,0 };
            const unsigned long dim608Kuo3Init[]    =   {   1   ,   3   ,   7   ,   7   ,   3   ,   7   ,   67  ,   163 ,   289 ,   139 ,   1415    ,   2149    ,   7501    ,0 };
            const unsigned long dim609Kuo3Init[]    =   {   1   ,   3   ,   5   ,   5   ,   13  ,   47  ,   21  ,   47  ,   3   ,   105 ,   473 ,   2013    ,   2443    ,0 };
            const unsigned long dim610Kuo3Init[]    =   {   1   ,   1   ,   3   ,   13  ,   31  ,   25  ,   21  ,   75  ,   505 ,   795 ,   1311    ,   2467    ,   5541    ,0 };
            const unsigned long dim611Kuo3Init[]    =   {   1   ,   1   ,   5   ,   1   ,   25  ,   15  ,   21  ,   113 ,   427 ,   249 ,   293 ,   1089    ,   1589    ,0 };
            const unsigned long dim612Kuo3Init[]    =   {   1   ,   1   ,   7   ,   5   ,   17  ,   11  ,   3   ,   211 ,   219 ,   37  ,   781 ,   1299    ,   3669    ,0 };
            const unsigned long dim613Kuo3Init[]    =   {   1   ,   3   ,   1   ,   15  ,   23  ,   3   ,   55  ,   59  ,   273 ,   947 ,   1647    ,   3111    ,   2615    ,0 };
            const unsigned long dim614Kuo3Init[]    =   {   1   ,   1   ,   3   ,   7   ,   21  ,   33  ,   81  ,   253 ,   267 ,   499 ,   299 ,   4015    ,   1669    ,0 };
            const unsigned long dim615Kuo3Init[]    =   {   1   ,   3   ,   7   ,   5   ,   5   ,   17  ,   75  ,   205 ,   53  ,   237 ,   1791    ,   3671    ,   4205    ,0 };
            const unsigned long dim616Kuo3Init[]    =   {   1   ,   3   ,   1   ,   5   ,   3   ,   59  ,   77  ,   39  ,   295 ,   649 ,   1219    ,   3971    ,   7675    ,0 };
            const unsigned long dim617Kuo3Init[]    =   {   1   ,   3   ,   5   ,   15  ,   1   ,   1   ,   13  ,   207 ,   445 ,   91  ,   571 ,   869 ,   359 ,0 };
            const unsigned long dim618Kuo3Init[]    =   {   1   ,   3   ,   5   ,   5   ,   15  ,   35  ,   79  ,   175 ,   313 ,   745 ,   1689    ,   1499    ,   5899    ,0 };
            const unsigned long dim619Kuo3Init[]    =   {   1   ,   3   ,   7   ,   7   ,   19  ,   51  ,   57  ,   131 ,   491 ,   301 ,   1405    ,   3789    ,   6081    ,0 };
            const unsigned long dim620Kuo3Init[]    =   {   1   ,   1   ,   3   ,   9   ,   27  ,   53  ,   55  ,   193 ,   413 ,   535 ,   715 ,   2823    ,   2063    ,0 };
            const unsigned long dim621Kuo3Init[]    =   {   1   ,   1   ,   7   ,   13  ,   31  ,   47  ,   47  ,   99  ,   441 ,   125 ,   999 ,   3575    ,   8007    ,0 };
            const unsigned long dim622Kuo3Init[]    =   {   1   ,   1   ,   1   ,   11  ,   17  ,   45  ,   67  ,   27  ,   329 ,   267 ,   773 ,   565 ,   5559    ,0 };
            const unsigned long dim623Kuo3Init[]    =   {   1   ,   1   ,   7   ,   11  ,   31  ,   23  ,   125 ,   117 ,   373 ,   523 ,   571 ,   1385    ,   3841    ,0 };
            const unsigned long dim624Kuo3Init[]    =   {   1   ,   3   ,   7   ,   15  ,   3   ,   15  ,   63  ,   91  ,   23  ,   97  ,   311 ,   3045    ,   2025    ,0 };
            const unsigned long dim625Kuo3Init[]    =   {   1   ,   3   ,   7   ,   5   ,   29  ,   37  ,   105 ,   63  ,   511 ,   35  ,   537 ,   4047    ,   7449    ,0 };
            const unsigned long dim626Kuo3Init[]    =   {   1   ,   3   ,   1   ,   13  ,   5   ,   9   ,   29  ,   51  ,   269 ,   395 ,   1737    ,   2097    ,   199 ,0 };
            const unsigned long dim627Kuo3Init[]    =   {   1   ,   3   ,   7   ,   15  ,   23  ,   33  ,   71  ,   111 ,   497 ,   369 ,   1265    ,   2937    ,   1883    ,0 };
            const unsigned long dim628Kuo3Init[]    =   {   1   ,   1   ,   3   ,   3   ,   3   ,   49  ,   117 ,   243 ,   57  ,   83  ,   3   ,   3319    ,   5711    ,0 };
            const unsigned long dim629Kuo3Init[]    =   {   1   ,   1   ,   5   ,   11  ,   25  ,   41  ,   29  ,   123 ,   405 ,   623 ,   1537    ,   3467    ,   699 ,0 };
            const unsigned long dim630Kuo3Init[]    =   {   1   ,   3   ,   7   ,   13  ,   17  ,   5   ,   87  ,   103 ,   501 ,   581 ,   713 ,   1535    ,   4637    ,0 };
            const unsigned long dim631Kuo3Init[]    =   {   1   ,   3   ,   3   ,   3   ,   25  ,   59  ,   91  ,   241 ,   163 ,   537 ,   1249    ,   3427    ,   6647    ,0 };
            const unsigned long dim632Kuo3Init[]    =   {   1   ,   3   ,   7   ,   1   ,   27  ,   7   ,   49  ,   219 ,   51  ,   989 ,   599 ,   2593    ,   6051    ,0 };
            const unsigned long dim633Kuo3Init[]    =   {   1   ,   1   ,   7   ,   9   ,   17  ,   19  ,   31  ,   23  ,   199 ,   729 ,   819 ,   287 ,   4689    ,0 };
            const unsigned long dim634Kuo3Init[]    =   {   1   ,   1   ,   1   ,   11  ,   13  ,   49  ,   97  ,   229 ,   411 ,   29  ,   1815    ,   3761    ,   3849    ,0 };
            const unsigned long dim635Kuo3Init[]    =   {   1   ,   3   ,   1   ,   1   ,   23  ,   35  ,   15  ,   33  ,   3   ,   865 ,   995 ,   1263    ,   3779    ,0 };
            const unsigned long dim636Kuo3Init[]    =   {   1   ,   1   ,   7   ,   7   ,   17  ,   1   ,   103 ,   117 ,   455 ,   93  ,   599 ,   1075    ,   4389    ,0 };
            const unsigned long dim637Kuo3Init[]    =   {   1   ,   3   ,   7   ,   5   ,   27  ,   3   ,   67  ,   117 ,   101 ,   443 ,   1773    ,   3731    ,   629 ,0 };
            const unsigned long dim638Kuo3Init[]    =   {   1   ,   3   ,   1   ,   11  ,   9   ,   51  ,   63  ,   217 ,   273 ,   425 ,   1889    ,   1113    ,   3799    ,0 };
            const unsigned long dim639Kuo3Init[]    =   {   1   ,   1   ,   3   ,   3   ,   25  ,   13  ,   37  ,   89  ,   429 ,   493 ,   637 ,   3287    ,   6733    ,0 };
            const unsigned long dim640Kuo3Init[]    =   {   1   ,   1   ,   1   ,   9   ,   17  ,   29  ,   45  ,   89  ,   141 ,   633 ,   243 ,   4035    ,   1835    ,0 };
            const unsigned long dim641Kuo3Init[]    =   {   1   ,   1   ,   1   ,   9   ,   15  ,   41  ,   61  ,   239 ,   349 ,   3   ,   1669    ,   1767    ,   6257    ,0 };
            const unsigned long dim642Kuo3Init[]    =   {   1   ,   3   ,   5   ,   11  ,   21  ,   41  ,   17  ,   65  ,   495 ,   879 ,   1891    ,   667 ,   3807    ,0 };
            const unsigned long dim643Kuo3Init[]    =   {   1   ,   1   ,   5   ,   11  ,   7   ,   25  ,   101 ,   231 ,   159 ,   287 ,   1125    ,   3859    ,   4331    ,0 };
            const unsigned long dim644Kuo3Init[]    =   {   1   ,   1   ,   1   ,   9   ,   13  ,   59  ,   23  ,   85  ,   393 ,   611 ,   1213    ,   7   ,   5707    ,0 };
            const unsigned long dim645Kuo3Init[]    =   {   1   ,   3   ,   5   ,   15  ,   13  ,   33  ,   61  ,   255 ,   163 ,   465 ,   641 ,   703 ,   6391    ,0 };
            const unsigned long dim646Kuo3Init[]    =   {   1   ,   1   ,   3   ,   3   ,   29  ,   41  ,   25  ,   19  ,   247 ,   973 ,   1815    ,   235 ,   7215    ,0 };
            const unsigned long dim647Kuo3Init[]    =   {   1   ,   1   ,   3   ,   11  ,   17  ,   43  ,   83  ,   105 ,   493 ,   663 ,   1395    ,   1079    ,   745 ,0 };
            const unsigned long dim648Kuo3Init[]    =   {   1   ,   1   ,   3   ,   13  ,   17  ,   53  ,   5   ,   55  ,   31  ,   459 ,   673 ,   1253    ,   5601    ,0 };
            const unsigned long dim649Kuo3Init[]    =   {   1   ,   3   ,   7   ,   7   ,   1   ,   49  ,   85  ,   73  ,   265 ,   555 ,   43  ,   2235    ,   6787    ,0 };
            const unsigned long dim650Kuo3Init[]    =   {   1   ,   1   ,   7   ,   5   ,   15  ,   3   ,   43  ,   201 ,   289 ,   755 ,   1605    ,   1927    ,   6179    ,0 };
            const unsigned long dim651Kuo3Init[]    =   {   1   ,   1   ,   7   ,   9   ,   15  ,   5   ,   101 ,   19  ,   17  ,   879 ,   1155    ,   3845    ,   3225    ,0 };
            const unsigned long dim652Kuo3Init[]    =   {   1   ,   3   ,   5   ,   15  ,   15  ,   63  ,   103 ,   133 ,   209 ,   121 ,   1865    ,   4039    ,   3335    ,0 };
            const unsigned long dim653Kuo3Init[]    =   {   1   ,   3   ,   7   ,   11  ,   15  ,   51  ,   87  ,   241 ,   455 ,   431 ,   1111    ,   1319    ,   1959    ,0 };
            const unsigned long dim654Kuo3Init[]    =   {   1   ,   3   ,   7   ,   5   ,   21  ,   47  ,   67  ,   251 ,   91  ,   575 ,   481 ,   3269    ,   1473    ,0 };
            const unsigned long dim655Kuo3Init[]    =   {   1   ,   3   ,   1   ,   1   ,   23  ,   5   ,   37  ,   169 ,   105 ,   803 ,   1765    ,   1823    ,   5081    ,0 };
            const unsigned long dim656Kuo3Init[]    =   {   1   ,   1   ,   7   ,   9   ,   1   ,   37  ,   55  ,   175 ,   431 ,   77  ,   171 ,   3759    ,   1227    ,0 };
            const unsigned long dim657Kuo3Init[]    =   {   1   ,   3   ,   7   ,   9   ,   3   ,   61  ,   111 ,   139 ,   87  ,   649 ,   243 ,   1109    ,   1303    ,0 };
            const unsigned long dim658Kuo3Init[]    =   {   1   ,   3   ,   1   ,   11  ,   17  ,   25  ,   9   ,   117 ,   227 ,   821 ,   1219    ,   2567    ,   5851    ,0 };
            const unsigned long dim659Kuo3Init[]    =   {   1   ,   1   ,   7   ,   1   ,   19  ,   53  ,   89  ,   57  ,   29  ,   885 ,   275 ,   2289    ,   4341    ,0 };
            const unsigned long dim660Kuo3Init[]    =   {   1   ,   3   ,   1   ,   5   ,   5   ,   1   ,   43  ,   237 ,   235 ,   667 ,   1433    ,   2497    ,   5535    ,0 };
            const unsigned long dim661Kuo3Init[]    =   {   1   ,   1   ,   1   ,   11  ,   21  ,   9   ,   29  ,   165 ,   153 ,   157 ,   1209    ,   817 ,   7819    ,0 };
            const unsigned long dim662Kuo3Init[]    =   {   1   ,   3   ,   1   ,   5   ,   1   ,   37  ,   33  ,   9   ,   493 ,   497 ,   1469    ,   357 ,   5929    ,0 };
            const unsigned long dim663Kuo3Init[]    =   {   1   ,   3   ,   3   ,   9   ,   25  ,   55  ,   5   ,   157 ,   397 ,   461 ,   537 ,   1085    ,   6737    ,0 };
            const unsigned long dim664Kuo3Init[]    =   {   1   ,   3   ,   3   ,   13  ,   3   ,   61  ,   83  ,   5   ,   5   ,   243 ,   1745    ,   845 ,   1965    ,0 };
            const unsigned long dim665Kuo3Init[]    =   {   1   ,   3   ,   1   ,   3   ,   31  ,   63  ,   3   ,   75  ,   355 ,   799 ,   417 ,   791 ,   2763    ,0 };
            const unsigned long dim666Kuo3Init[]    =   {   1   ,   3   ,   3   ,   11  ,   5   ,   31  ,   99  ,   163 ,   69  ,   699 ,   809 ,   1651    ,   7249    ,0 };
            const unsigned long dim667Kuo3Init[]    =   {   1   ,   1   ,   7   ,   3   ,   29  ,   53  ,   51  ,   185 ,   379 ,   743 ,   611 ,   963 ,   5631    ,0 };
            const unsigned long dim668Kuo3Init[]    =   {   1   ,   1   ,   5   ,   11  ,   21  ,   15  ,   59  ,   75  ,   121 ,   247 ,   307 ,   2483    ,   2861    ,0 };
            const unsigned long dim669Kuo3Init[]    =   {   1   ,   1   ,   1   ,   5   ,   21  ,   5   ,   33  ,   43  ,   225 ,   675 ,   213 ,   3081    ,   5047    ,0 };
            const unsigned long dim670Kuo3Init[]    =   {   1   ,   3   ,   3   ,   15  ,   29  ,   45  ,   19  ,   19  ,   15  ,   605 ,   1531    ,   1305    ,   7725    ,0 };
            const unsigned long dim671Kuo3Init[]    =   {   1   ,   3   ,   1   ,   5   ,   19  ,   21  ,   73  ,   131 ,   37  ,   667 ,   1953    ,   1211    ,   7499    ,0 };
            const unsigned long dim672Kuo3Init[]    =   {   1   ,   3   ,   1   ,   5   ,   13  ,   3   ,   105 ,   151 ,   131 ,   965 ,   1953    ,   335 ,   4991    ,0 };
            const unsigned long dim673Kuo3Init[]    =   {   1   ,   3   ,   1   ,   13  ,   23  ,   37  ,   81  ,   147 ,   353 ,   125 ,   141 ,   4017    ,   7543    ,0 };
            const unsigned long dim674Kuo3Init[]    =   {   1   ,   1   ,   1   ,   7   ,   15  ,   13  ,   25  ,   107 ,   431 ,   327 ,   1069    ,   155 ,   1729    ,0 };
            const unsigned long dim675Kuo3Init[]    =   {   1   ,   1   ,   5   ,   1   ,   19  ,   3   ,   51  ,   113 ,   349 ,   845 ,   1157    ,   1771    ,   4923    ,0 };
            const unsigned long dim676Kuo3Init[]    =   {   1   ,   1   ,   3   ,   7   ,   17  ,   19  ,   47  ,   249 ,   215 ,   959 ,   659 ,   3989    ,   3719    ,0 };
            const unsigned long dim677Kuo3Init[]    =   {   1   ,   3   ,   5   ,   5   ,   27  ,   57  ,   41  ,   29  ,   153 ,   1007    ,   1209    ,   1373    ,   2847    ,0 };
            const unsigned long dim678Kuo3Init[]    =   {   1   ,   3   ,   5   ,   7   ,   5   ,   21  ,   23  ,   67  ,   127 ,   755 ,   211 ,   2517    ,   8161    ,0 };
            const unsigned long dim679Kuo3Init[]    =   {   1   ,   1   ,   7   ,   15  ,   1   ,   5   ,   65  ,   197 ,   435 ,   275 ,   473 ,   1333    ,   3669    ,0 };
            const unsigned long dim680Kuo3Init[]    =   {   1   ,   3   ,   7   ,   1   ,   17  ,   19  ,   61  ,   11  ,   425 ,   717 ,   351 ,   2323    ,   6575    ,0 };
            const unsigned long dim681Kuo3Init[]    =   {   1   ,   1   ,   7   ,   3   ,   29  ,   13  ,   115 ,   19  ,   235 ,   371 ,   81  ,   3107    ,   3103    ,0 };
            const unsigned long dim682Kuo3Init[]    =   {   1   ,   1   ,   3   ,   1   ,   23  ,   19  ,   57  ,   179 ,   7   ,   405 ,   217 ,   3159    ,   6675    ,0 };
            const unsigned long dim683Kuo3Init[]    =   {   1   ,   3   ,   3   ,   3   ,   19  ,   39  ,   121 ,   195 ,   205 ,   489 ,   1959    ,   3617    ,   799 ,0 };
            const unsigned long dim684Kuo3Init[]    =   {   1   ,   3   ,   5   ,   9   ,   19  ,   35  ,   115 ,   3   ,   17  ,   201 ,   671 ,   1847    ,   597 ,0 };
            const unsigned long dim685Kuo3Init[]    =   {   1   ,   3   ,   1   ,   3   ,   7   ,   1   ,   121 ,   77  ,   163 ,   769 ,   977 ,   1223    ,   4215    ,0 };
            const unsigned long dim686Kuo3Init[]    =   {   1   ,   1   ,   1   ,   15  ,   9   ,   13  ,   33  ,   57  ,   271 ,   733 ,   1699    ,   3083    ,   813 ,0 };
            const unsigned long dim687Kuo3Init[]    =   {   1   ,   1   ,   7   ,   7   ,   15  ,   49  ,   101 ,   183 ,   339 ,   129 ,   1865    ,   2697    ,   3131    ,0 };
            const unsigned long dim688Kuo3Init[]    =   {   1   ,   3   ,   1   ,   3   ,   9   ,   39  ,   33  ,   141 ,   459 ,   971 ,   1621    ,   3893    ,   4393    ,0 };
            const unsigned long dim689Kuo3Init[]    =   {   1   ,   1   ,   3   ,   3   ,   27  ,   61  ,   53  ,   79  ,   375 ,   613 ,   663 ,   2325    ,   4707    ,0 };
            const unsigned long dim690Kuo3Init[]    =   {   1   ,   1   ,   7   ,   1   ,   27  ,   59  ,   59  ,   179 ,   173 ,   333 ,   649 ,   1053    ,   555 ,0 };
            const unsigned long dim691Kuo3Init[]    =   {   1   ,   3   ,   3   ,   13  ,   15  ,   55  ,   25  ,   91  ,   45  ,   739 ,   1769    ,   1341    ,   2007    ,0 };
            const unsigned long dim692Kuo3Init[]    =   {   1   ,   1   ,   3   ,   15  ,   9   ,   59  ,   27  ,   45  ,   133 ,   941 ,   1037    ,   1559    ,   7145    ,0 };
            const unsigned long dim693Kuo3Init[]    =   {   1   ,   3   ,   5   ,   1   ,   13  ,   5   ,   3   ,   167 ,   235 ,   609 ,   1023    ,   1967    ,   771 ,0 };
            const unsigned long dim694Kuo3Init[]    =   {   1   ,   1   ,   1   ,   15  ,   19  ,   15  ,   45  ,   143 ,   309 ,   697 ,   337 ,   3399    ,   3599    ,0 };
            const unsigned long dim695Kuo3Init[]    =   {   1   ,   1   ,   5   ,   3   ,   9   ,   1   ,   19  ,   41  ,   39  ,   1003    ,   997 ,   1357    ,   5849    ,0 };
            const unsigned long dim696Kuo3Init[]    =   {   1   ,   3   ,   3   ,   7   ,   29  ,   37  ,   9   ,   165 ,   9   ,   855 ,   337 ,   2363    ,   2633    ,0 };
            const unsigned long dim697Kuo3Init[]    =   {   1   ,   3   ,   5   ,   1   ,   23  ,   19  ,   97  ,   55  ,   411 ,   623 ,   287 ,   3365    ,   7675    ,0 };
            const unsigned long dim698Kuo3Init[]    =   {   1   ,   3   ,   3   ,   3   ,   11  ,   1   ,   47  ,   81  ,   377 ,   85  ,   859 ,   3717    ,   1979    ,0 };
            const unsigned long dim699Kuo3Init[]    =   {   1   ,   1   ,   3   ,   1   ,   11  ,   3   ,   125 ,   43  ,   471 ,   127 ,   341 ,   2943    ,   231 ,0 };
            const unsigned long dim700Kuo3Init[]    =   {   1   ,   1   ,   5   ,   15  ,   29  ,   21  ,   101 ,   193 ,   389 ,   727 ,   1397    ,   2799    ,   4387    ,0 };
            const unsigned long dim701Kuo3Init[]    =   {   1   ,   3   ,   5   ,   15  ,   27  ,   19  ,   13  ,   195 ,   105 ,   5   ,   985 ,   767 ,   4573    ,0 };
            const unsigned long dim702Kuo3Init[]    =   {   1   ,   1   ,   5   ,   3   ,   13  ,   29  ,   33  ,   123 ,   107 ,   353 ,   723 ,   3821    ,   5425    ,0 };
            const unsigned long dim703Kuo3Init[]    =   {   1   ,   3   ,   5   ,   11  ,   23  ,   11  ,   69  ,   219 ,   15  ,   735 ,   1015    ,   431 ,   6511    ,0 };
            const unsigned long dim704Kuo3Init[]    =   {   1   ,   3   ,   1   ,   1   ,   27  ,   49  ,   79  ,   91  ,   201 ,   365 ,   255 ,   1309    ,   7165    ,0 };
            const unsigned long dim705Kuo3Init[]    =   {   1   ,   3   ,   1   ,   13  ,   21  ,   61  ,   39  ,   111 ,   71  ,   137 ,   551 ,   3237    ,   6681    ,0 };
            const unsigned long dim706Kuo3Init[]    =   {   1   ,   3   ,   5   ,   13  ,   19  ,   21  ,   121 ,   227 ,   381 ,   677 ,   421 ,   449 ,   5375    ,0 };
            const unsigned long dim707Kuo3Init[]    =   {   1   ,   1   ,   7   ,   5   ,   1   ,   23  ,   95  ,   199 ,   509 ,   197 ,   829 ,   941 ,   2269    ,0 };
            const unsigned long dim708Kuo3Init[]    =   {   1   ,   3   ,   7   ,   5   ,   21  ,   61  ,   119 ,   59  ,   97  ,   841 ,   453 ,   2439    ,   7345    ,0 };
            const unsigned long dim709Kuo3Init[]    =   {   1   ,   1   ,   1   ,   3   ,   23  ,   21  ,   19  ,   245 ,   281 ,   841 ,   273 ,   1967    ,   2129    ,0 };
            const unsigned long dim710Kuo3Init[]    =   {   1   ,   3   ,   3   ,   5   ,   7   ,   43  ,   19  ,   253 ,   239 ,   421 ,   461 ,   4023    ,   6981    ,0 };
            const unsigned long dim711Kuo3Init[]    =   {   1   ,   3   ,   1   ,   11  ,   25  ,   3   ,   109 ,   107 ,   55  ,   781 ,   1109    ,   101 ,   889 ,0 };
            const unsigned long dim712Kuo3Init[]    =   {   1   ,   3   ,   7   ,   13  ,   1   ,   21  ,   27  ,   3   ,   303 ,   557 ,   1381    ,   407 ,   1849    ,0 };
            const unsigned long dim713Kuo3Init[]    =   {   1   ,   1   ,   3   ,   13  ,   3   ,   61  ,   121 ,   119 ,   453 ,   137 ,   2039    ,   3333    ,   333 ,0 };
            const unsigned long dim714Kuo3Init[]    =   {   1   ,   3   ,   5   ,   1   ,   21  ,   9   ,   113 ,   9   ,   413 ,   301 ,   1795    ,   2827    ,   1837    ,0 };
            const unsigned long dim715Kuo3Init[]    =   {   1   ,   3   ,   5   ,   15  ,   13  ,   31  ,   119 ,   107 ,   445 ,   277 ,   1555    ,   2121    ,   2173    ,0 };
            const unsigned long dim716Kuo3Init[]    =   {   1   ,   1   ,   1   ,   5   ,   9   ,   5   ,   5   ,   97  ,   191 ,   949 ,   1603    ,   2181    ,   6879    ,0 };
            const unsigned long dim717Kuo3Init[]    =   {   1   ,   3   ,   3   ,   9   ,   5   ,   47  ,   77  ,   43  ,   483 ,   137 ,   61  ,   1811    ,   5309    ,0 };
            const unsigned long dim718Kuo3Init[]    =   {   1   ,   3   ,   7   ,   11  ,   25  ,   63  ,   107 ,   207 ,   23  ,   231 ,   1685    ,   4009    ,   391 ,0 };
            const unsigned long dim719Kuo3Init[]    =   {   1   ,   3   ,   7   ,   9   ,   3   ,   35  ,   119 ,   83  ,   193 ,   269 ,   289 ,   3051    ,   8137    ,0 };
            const unsigned long dim720Kuo3Init[]    =   {   1   ,   3   ,   3   ,   11  ,   25  ,   25  ,   17  ,   243 ,   195 ,   503 ,   1797    ,   1913    ,   749 ,0 };
            const unsigned long dim721Kuo3Init[]    =   {   1   ,   3   ,   1   ,   13  ,   5   ,   35  ,   101 ,   57  ,   185 ,   881 ,   2005    ,   3493    ,   2715    ,0 };
            const unsigned long dim722Kuo3Init[]    =   {   1   ,   3   ,   1   ,   11  ,   25  ,   51  ,   113 ,   177 ,   175 ,   855 ,   1727    ,   1901    ,   8181    ,0 };
            const unsigned long dim723Kuo3Init[]    =   {   1   ,   1   ,   3   ,   9   ,   9   ,   63  ,   109 ,   233 ,   205 ,   357 ,   499 ,   3367    ,   1189    ,0 };
            const unsigned long dim724Kuo3Init[]    =   {   1   ,   3   ,   1   ,   15  ,   17  ,   11  ,   35  ,   135 ,   501 ,   319 ,   1465    ,   2095    ,   5269    ,0 };
            const unsigned long dim725Kuo3Init[]    =   {   1   ,   1   ,   3   ,   9   ,   1   ,   27  ,   99  ,   201 ,   67  ,   269 ,   135 ,   3891    ,   4067    ,0 };
            const unsigned long dim726Kuo3Init[]    =   {   1   ,   1   ,   1   ,   9   ,   15  ,   37  ,   83  ,   135 ,   3   ,   645 ,   499 ,   2597    ,   2705    ,0 };
            const unsigned long dim727Kuo3Init[]    =   {   1   ,   3   ,   1   ,   7   ,   17  ,   33  ,   89  ,   51  ,   295 ,   809 ,   743 ,   27  ,   2599    ,0 };
            const unsigned long dim728Kuo3Init[]    =   {   1   ,   1   ,   7   ,   1   ,   5   ,   55  ,   11  ,   97  ,   177 ,   515 ,   1651    ,   1183    ,   5803    ,0 };
            const unsigned long dim729Kuo3Init[]    =   {   1   ,   1   ,   3   ,   11  ,   21  ,   3   ,   39  ,   249 ,   55  ,   891 ,   1681    ,   709 ,   7833    ,0 };
            const unsigned long dim730Kuo3Init[]    =   {   1   ,   3   ,   5   ,   13  ,   11  ,   21  ,   81  ,   21  ,   99  ,   223 ,   355 ,   2015    ,   7195    ,0 };
            const unsigned long dim731Kuo3Init[]    =   {   1   ,   3   ,   7   ,   11  ,   25  ,   23  ,   9   ,   131 ,   417 ,   583 ,   1927    ,   277 ,   2923    ,0 };
            const unsigned long dim732Kuo3Init[]    =   {   1   ,   1   ,   5   ,   5   ,   1   ,   43  ,   35  ,   87  ,   471 ,   499 ,   1411    ,   2867    ,   2477    ,0 };
            const unsigned long dim733Kuo3Init[]    =   {   1   ,   1   ,   5   ,   9   ,   25  ,   43  ,   83  ,   103 ,   183 ,   527 ,   483 ,   935 ,   7097    ,0 };
            const unsigned long dim734Kuo3Init[]    =   {   1   ,   1   ,   5   ,   7   ,   31  ,   13  ,   49  ,   169 ,   135 ,   885 ,   1367    ,   2627    ,   5917    ,0 };
            const unsigned long dim735Kuo3Init[]    =   {   1   ,   3   ,   7   ,   5   ,   1   ,   39  ,   97  ,   167 ,   95  ,   745 ,   1067    ,   805 ,   2885    ,0 };
            const unsigned long dim736Kuo3Init[]    =   {   1   ,   3   ,   7   ,   15  ,   31  ,   41  ,   1   ,   147 ,   15  ,   795 ,   323 ,   309 ,   5075    ,0 };
            const unsigned long dim737Kuo3Init[]    =   {   1   ,   3   ,   7   ,   5   ,   23  ,   3   ,   65  ,   143 ,   183 ,   671 ,   993 ,   547 ,   5971    ,0 };
            const unsigned long dim738Kuo3Init[]    =   {   1   ,   3   ,   7   ,   1   ,   21  ,   59  ,   29  ,   45  ,   367 ,   363 ,   1707    ,   541 ,   7045    ,0 };
            const unsigned long dim739Kuo3Init[]    =   {   1   ,   1   ,   7   ,   15  ,   11  ,   25  ,   77  ,   33  ,   409 ,   495 ,   1713    ,   1479    ,   2857    ,0 };
            const unsigned long dim740Kuo3Init[]    =   {   1   ,   1   ,   3   ,   1   ,   11  ,   59  ,   113 ,   217 ,   313 ,   79  ,   379 ,   627 ,   3917    ,0 };
            const unsigned long dim741Kuo3Init[]    =   {   1   ,   1   ,   1   ,   15  ,   31  ,   61  ,   73  ,   211 ,   419 ,   965 ,   1817    ,   3965    ,   1425    ,0 };
            const unsigned long dim742Kuo3Init[]    =   {   1   ,   1   ,   7   ,   13  ,   9   ,   9   ,   61  ,   119 ,   223 ,   387 ,   595 ,   191 ,   1723    ,0 };
            const unsigned long dim743Kuo3Init[]    =   {   1   ,   3   ,   5   ,   3   ,   25  ,   49  ,   83  ,   175 ,   329 ,   1003    ,   729 ,   1399    ,   6553    ,0 };
            const unsigned long dim744Kuo3Init[]    =   {   1   ,   1   ,   5   ,   9   ,   5   ,   37  ,   31  ,   43  ,   365 ,   403 ,   1485    ,   1621    ,   3753    ,0 };
            const unsigned long dim745Kuo3Init[]    =   {   1   ,   3   ,   5   ,   7   ,   17  ,   5   ,   125 ,   219 ,   311 ,   1   ,   639 ,   1939    ,   4367    ,0 };
            const unsigned long dim746Kuo3Init[]    =   {   1   ,   3   ,   5   ,   3   ,   21  ,   63  ,   119 ,   31  ,   151 ,   325 ,   1063    ,   2477    ,   1223    ,0 };
            const unsigned long dim747Kuo3Init[]    =   {   1   ,   1   ,   5   ,   13  ,   15  ,   23  ,   1   ,   141 ,   105 ,   21  ,   747 ,   161 ,   3555    ,0 };
            const unsigned long dim748Kuo3Init[]    =   {   1   ,   3   ,   7   ,   1   ,   31  ,   21  ,   11  ,   27  ,   267 ,   205 ,   1569    ,   765 ,   1357    ,0 };
            const unsigned long dim749Kuo3Init[]    =   {   1   ,   1   ,   3   ,   9   ,   23  ,   59  ,   93  ,   51  ,   81  ,   367 ,   1545    ,   1161    ,   3481    ,0 };
            const unsigned long dim750Kuo3Init[]    =   {   1   ,   3   ,   7   ,   11  ,   15  ,   51  ,   43  ,   205 ,   237 ,   617 ,   1075    ,   1051    ,   5285    ,0 };
            const unsigned long dim751Kuo3Init[]    =   {   1   ,   3   ,   7   ,   3   ,   11  ,   5   ,   95  ,   203 ,   495 ,   661 ,   1119    ,   1095    ,   3391    ,0 };
            const unsigned long dim752Kuo3Init[]    =   {   1   ,   3   ,   3   ,   1   ,   25  ,   39  ,   59  ,   159 ,   273 ,   101 ,   1887    ,   417 ,   3423    ,0 };
            const unsigned long dim753Kuo3Init[]    =   {   1   ,   3   ,   5   ,   13  ,   23  ,   45  ,   13  ,   87  ,   139 ,   313 ,   1527    ,   3739    ,   7009    ,0 };
            const unsigned long dim754Kuo3Init[]    =   {   1   ,   1   ,   5   ,   1   ,   23  ,   43  ,   127 ,   149 ,   63  ,   823 ,   1075    ,   3123    ,   3079    ,0 };
            const unsigned long dim755Kuo3Init[]    =   {   1   ,   1   ,   7   ,   3   ,   11  ,   1   ,   29  ,   127 ,   313 ,   971 ,   1363    ,   1765    ,   6439    ,0 };
            const unsigned long dim756Kuo3Init[]    =   {   1   ,   1   ,   3   ,   13  ,   1   ,   19  ,   61  ,   125 ,   345 ,   51  ,   273 ,   2783    ,   5255    ,0 };
            const unsigned long dim757Kuo3Init[]    =   {   1   ,   1   ,   1   ,   7   ,   29  ,   33  ,   21  ,   239 ,   351 ,   481 ,   795 ,   2865    ,   2235    ,0 };
            const unsigned long dim758Kuo3Init[]    =   {   1   ,   3   ,   3   ,   3   ,   19  ,   31  ,   41  ,   125 ,   327 ,   169 ,   837 ,   753 ,   191 ,0 };
            const unsigned long dim759Kuo3Init[]    =   {   1   ,   3   ,   7   ,   15  ,   5   ,   15  ,   5   ,   109 ,   333 ,   127 ,   2021    ,   441 ,   6307    ,0 };
            const unsigned long dim760Kuo3Init[]    =   {   1   ,   3   ,   7   ,   9   ,   23  ,   39  ,   79  ,   235 ,   27  ,   573 ,   1989    ,   1401    ,   7597    ,0 };
            const unsigned long dim761Kuo3Init[]    =   {   1   ,   1   ,   7   ,   3   ,   23  ,   29  ,   119 ,   53  ,   105 ,   609 ,   1139    ,   321 ,   1111    ,0 };
            const unsigned long dim762Kuo3Init[]    =   {   1   ,   1   ,   7   ,   13  ,   13  ,   43  ,   3   ,   211 ,   389 ,   563 ,   1047    ,   1373    ,   4669    ,0 };
            const unsigned long dim763Kuo3Init[]    =   {   1   ,   3   ,   3   ,   1   ,   9   ,   7   ,   1   ,   135 ,   455 ,   539 ,   1681    ,   1851    ,   4585    ,0 };
            const unsigned long dim764Kuo3Init[]    =   {   1   ,   3   ,   1   ,   13  ,   19  ,   45  ,   65  ,   109 ,   391 ,   563 ,   753 ,   2405    ,   2623    ,0 };
            const unsigned long dim765Kuo3Init[]    =   {   1   ,   3   ,   1   ,   13  ,   31  ,   41  ,   29  ,   153 ,   507 ,   559 ,   877 ,   125 ,   4827    ,0 };
            const unsigned long dim766Kuo3Init[]    =   {   1   ,   1   ,   7   ,   3   ,   15  ,   53  ,   39  ,   1   ,   143 ,   351 ,   505 ,   1213    ,   4829    ,0 };
            const unsigned long dim767Kuo3Init[]    =   {   1   ,   1   ,   5   ,   5   ,   27  ,   49  ,   39  ,   101 ,   501 ,   239 ,   1233    ,   1695    ,   6929    ,0 };
            const unsigned long dim768Kuo3Init[]    =   {   1   ,   1   ,   5   ,   9   ,   13  ,   29  ,   103 ,   185 ,   343 ,   943 ,   1709    ,   801 ,   1227    ,0 };
            const unsigned long dim769Kuo3Init[]    =   {   1   ,   1   ,   1   ,   3   ,   3   ,   1   ,   63  ,   23  ,   199 ,   53  ,   1629    ,   325 ,   7609    ,0 };
            const unsigned long dim770Kuo3Init[]    =   {   1   ,   3   ,   7   ,   1   ,   11  ,   47  ,   61  ,   171 ,   71  ,   493 ,   2035    ,   869 ,   5925    ,0 };
            const unsigned long dim771Kuo3Init[]    =   {   1   ,   1   ,   7   ,   7   ,   5   ,   3   ,   95  ,   163 ,   89  ,   395 ,   1033    ,   2051    ,   5589    ,0 };
            const unsigned long dim772Kuo3Init[]    =   {   1   ,   3   ,   5   ,   9   ,   3   ,   21  ,   29  ,   143 ,   333 ,   805 ,   267 ,   4077    ,   383 ,0 };
            const unsigned long dim773Kuo3Init[]    =   {   1   ,   1   ,   3   ,   13  ,   9   ,   17  ,   87  ,   163 ,   107 ,   1011    ,   1099    ,   3395    ,   987 ,0 };
            const unsigned long dim774Kuo3Init[]    =   {   1   ,   3   ,   1   ,   1   ,   5   ,   51  ,   71  ,   249 ,   441 ,   513 ,   791 ,   1831    ,   2757    ,0 };
            const unsigned long dim775Kuo3Init[]    =   {   1   ,   1   ,   3   ,   13  ,   23  ,   7   ,   101 ,   211 ,   289 ,   1001    ,   1623    ,   3239    ,   1685    ,0 };
            const unsigned long dim776Kuo3Init[]    =   {   1   ,   1   ,   5   ,   15  ,   31  ,   53  ,   91  ,   245 ,   143 ,   237 ,   557 ,   3099    ,   3737    ,0 };
            const unsigned long dim777Kuo3Init[]    =   {   1   ,   3   ,   1   ,   1   ,   5   ,   27  ,   51  ,   173 ,   321 ,   85  ,   1385    ,   9   ,   1887    ,0 };
            const unsigned long dim778Kuo3Init[]    =   {   1   ,   3   ,   1   ,   15  ,   11  ,   17  ,   49  ,   97  ,   121 ,   189 ,   2031    ,   565 ,   7513    ,0 };
            const unsigned long dim779Kuo3Init[]    =   {   1   ,   1   ,   5   ,   5   ,   1   ,   3   ,   55  ,   157 ,   27  ,   783 ,   1543    ,   2645    ,   1079    ,0 };
            const unsigned long dim780Kuo3Init[]    =   {   1   ,   1   ,   5   ,   11  ,   27  ,   9   ,   83  ,   155 ,   17  ,   147 ,   1561    ,   3213    ,   4483    ,0 };
            const unsigned long dim781Kuo3Init[]    =   {   1   ,   3   ,   5   ,   15  ,   23  ,   63  ,   127 ,   97  ,   459 ,   869 ,   671 ,   3287    ,   995 ,0 };
            const unsigned long dim782Kuo3Init[]    =   {   1   ,   3   ,   3   ,   1   ,   5   ,   1   ,   89  ,   229 ,   269 ,   493 ,   709 ,   2719    ,   4803    ,0 };
            const unsigned long dim783Kuo3Init[]    =   {   1   ,   1   ,   3   ,   3   ,   27  ,   5   ,   85  ,   55  ,   491 ,   781 ,   711 ,   3193    ,   2051    ,0 };
            const unsigned long dim784Kuo3Init[]    =   {   1   ,   1   ,   5   ,   13  ,   21  ,   9   ,   31  ,   107 ,   381 ,   613 ,   991 ,   2289    ,   325 ,0 };
            const unsigned long dim785Kuo3Init[]    =   {   1   ,   3   ,   7   ,   11  ,   11  ,   47  ,   115 ,   103 ,   101 ,   945 ,   1993    ,   1345    ,   2031    ,0 };
            const unsigned long dim786Kuo3Init[]    =   {   1   ,   3   ,   7   ,   11  ,   27  ,   59  ,   123 ,   73  ,   85  ,   975 ,   397 ,   2619    ,   2095    ,0 };
            const unsigned long dim787Kuo3Init[]    =   {   1   ,   1   ,   1   ,   13  ,   19  ,   31  ,   87  ,   45  ,   39  ,   43  ,   1041    ,   2015    ,   5859    ,0 };
            const unsigned long dim788Kuo3Init[]    =   {   1   ,   1   ,   3   ,   15  ,   5   ,   19  ,   47  ,   193 ,   293 ,   961 ,   113 ,   3013    ,   889 ,0 };
            const unsigned long dim789Kuo3Init[]    =   {   1   ,   1   ,   1   ,   3   ,   31  ,   61  ,   61  ,   209 ,   493 ,   785 ,   1825    ,   3779    ,   8095    ,0 };
            const unsigned long dim790Kuo3Init[]    =   {   1   ,   1   ,   3   ,   9   ,   17  ,   51  ,   103 ,   255 ,   133 ,   165 ,   281 ,   3945    ,   2779    ,0 };
            const unsigned long dim791Kuo3Init[]    =   {   1   ,   1   ,   7   ,   3   ,   9   ,   33  ,   105 ,   29  ,   449 ,   231 ,   1393    ,   1071    ,   5043    ,0 };
            const unsigned long dim792Kuo3Init[]    =   {   1   ,   1   ,   7   ,   5   ,   29  ,   35  ,   115 ,   1   ,   467 ,   969 ,   343 ,   2291    ,   5099    ,0 };
            const unsigned long dim793Kuo3Init[]    =   {   1   ,   1   ,   3   ,   5   ,   13  ,   21  ,   99  ,   151 ,   13  ,   861 ,   189 ,   1709    ,   4943    ,0 };
            const unsigned long dim794Kuo3Init[]    =   {   1   ,   1   ,   7   ,   15  ,   5   ,   41  ,   53  ,   89  ,   235 ,   515 ,   1819    ,   3907    ,   4861    ,0 };
            const unsigned long dim795Kuo3Init[]    =   {   1   ,   3   ,   5   ,   1   ,   7   ,   15  ,   75  ,   65  ,   247 ,   581 ,   853 ,   445 ,   7623    ,0 };
            const unsigned long dim796Kuo3Init[]    =   {   1   ,   1   ,   3   ,   5   ,   25  ,   29  ,   27  ,   73  ,   359 ,   729 ,   1991    ,   1075    ,   8107    ,0 };
            const unsigned long dim797Kuo3Init[]    =   {   1   ,   1   ,   3   ,   1   ,   5   ,   35  ,   83  ,   119 ,   385 ,   657 ,   1085    ,   2707    ,   4403    ,0 };
            const unsigned long dim798Kuo3Init[]    =   {   1   ,   3   ,   5   ,   5   ,   25  ,   55  ,   51  ,   205 ,   257 ,   887 ,   827 ,   2795    ,   925 ,0 };
            const unsigned long dim799Kuo3Init[]    =   {   1   ,   1   ,   3   ,   1   ,   5   ,   29  ,   113 ,   195 ,   439 ,   241 ,   127 ,   2783    ,   2577    ,0 };
            const unsigned long dim800Kuo3Init[]    =   {   1   ,   3   ,   1   ,   11  ,   15  ,   47  ,   7   ,   149 ,   81  ,   897 ,   235 ,   3935    ,   6981    ,0 };
            const unsigned long dim801Kuo3Init[]    =   {   1   ,   3   ,   3   ,   1   ,   27  ,   5   ,   69  ,   125 ,   377 ,   829 ,   1549    ,   3025    ,   2319    ,0 };
            const unsigned long dim802Kuo3Init[]    =   {   1   ,   1   ,   7   ,   15  ,   15  ,   19  ,   3   ,   205 ,   291 ,   209 ,   1817    ,   2439    ,   1001    ,0 };
            const unsigned long dim803Kuo3Init[]    =   {   1   ,   1   ,   7   ,   9   ,   21  ,   7   ,   101 ,   17  ,   13  ,   615 ,   137 ,   1821    ,   1387    ,0 };
            const unsigned long dim804Kuo3Init[]    =   {   1   ,   3   ,   3   ,   11  ,   21  ,   27  ,   61  ,   133 ,   183 ,   699 ,   245 ,   527 ,   471 ,0 };
            const unsigned long dim805Kuo3Init[]    =   {   1   ,   3   ,   3   ,   13  ,   5   ,   25  ,   111 ,   69  ,   417 ,   271 ,   1545    ,   2729    ,   475 ,0 };
            const unsigned long dim806Kuo3Init[]    =   {   1   ,   3   ,   5   ,   15  ,   11  ,   47  ,   93  ,   93  ,   311 ,   775 ,   789 ,   1163    ,   1563    ,0 };
            const unsigned long dim807Kuo3Init[]    =   {   1   ,   3   ,   5   ,   1   ,   13  ,   63  ,   27  ,   113 ,   449 ,   971 ,   1021    ,   73  ,   1229    ,0 };
            const unsigned long dim808Kuo3Init[]    =   {   1   ,   3   ,   3   ,   15  ,   17  ,   37  ,   103 ,   13  ,   91  ,   823 ,   1767    ,   1295    ,   6941    ,0 };
            const unsigned long dim809Kuo3Init[]    =   {   1   ,   1   ,   3   ,   9   ,   15  ,   59  ,   125 ,   91  ,   195 ,   409 ,   65  ,   2873    ,   1921    ,0 };
            const unsigned long dim810Kuo3Init[]    =   {   1   ,   3   ,   1   ,   7   ,   17  ,   31  ,   21  ,   251 ,   145 ,   385 ,   1385    ,   1505    ,   5735    ,0 };
            const unsigned long dim811Kuo3Init[]    =   {   1   ,   3   ,   7   ,   11  ,   17  ,   19  ,   85  ,   85  ,   81  ,   583 ,   889 ,   669 ,   5187    ,0 };
            const unsigned long dim812Kuo3Init[]    =   {   1   ,   1   ,   5   ,   9   ,   23  ,   53  ,   97  ,   217 ,   129 ,   517 ,   271 ,   2967    ,   3937    ,0 };
            const unsigned long dim813Kuo3Init[]    =   {   1   ,   1   ,   1   ,   13  ,   3   ,   45  ,   49  ,   193 ,   35  ,   671 ,   1039    ,   2841    ,   4547    ,0 };
            const unsigned long dim814Kuo3Init[]    =   {   1   ,   3   ,   1   ,   15  ,   21  ,   29  ,   5   ,   91  ,   23  ,   247 ,   1701    ,   1585    ,   1493    ,0 };
            const unsigned long dim815Kuo3Init[]    =   {   1   ,   3   ,   3   ,   11  ,   15  ,   31  ,   21  ,   165 ,   153 ,   607 ,   221 ,   2781    ,   6159    ,0 };
            const unsigned long dim816Kuo3Init[]    =   {   1   ,   3   ,   7   ,   15  ,   15  ,   27  ,   107 ,   25  ,   457 ,   995 ,   849 ,   877 ,   407 ,0 };
            const unsigned long dim817Kuo3Init[]    =   {   1   ,   3   ,   1   ,   15  ,   7   ,   17  ,   37  ,   31  ,   203 ,   585 ,   369 ,   609 ,   2509    ,0 };
            const unsigned long dim818Kuo3Init[]    =   {   1   ,   3   ,   1   ,   11  ,   17  ,   5   ,   17  ,   71  ,   453 ,   171 ,   1317    ,   1985    ,   5817    ,0 };
            const unsigned long dim819Kuo3Init[]    =   {   1   ,   1   ,   5   ,   5   ,   23  ,   39  ,   123 ,   179 ,   49  ,   19  ,   1337    ,   1051    ,   5289    ,0 };
            const unsigned long dim820Kuo3Init[]    =   {   1   ,   3   ,   1   ,   13  ,   17  ,   43  ,   47  ,   139 ,   21  ,   851 ,   1253    ,   1939    ,   299 ,0 };
            const unsigned long dim821Kuo3Init[]    =   {   1   ,   1   ,   7   ,   1   ,   3   ,   35  ,   57  ,   115 ,   481 ,   653 ,   1975    ,   1319    ,   5225    ,0 };
            const unsigned long dim822Kuo3Init[]    =   {   1   ,   3   ,   7   ,   15  ,   5   ,   13  ,   3   ,   249 ,   21  ,   791 ,   1349    ,   1961    ,   6975    ,0 };
            const unsigned long dim823Kuo3Init[]    =   {   1   ,   1   ,   5   ,   11  ,   17  ,   3   ,   97  ,   103 ,   201 ,   313 ,   939 ,   693 ,   5397    ,0 };
            const unsigned long dim824Kuo3Init[]    =   {   1   ,   1   ,   7   ,   9   ,   29  ,   59  ,   35  ,   91  ,   135 ,   19  ,   1505    ,   1715    ,   3335    ,0 };
            const unsigned long dim825Kuo3Init[]    =   {   1   ,   3   ,   5   ,   1   ,   17  ,   53  ,   59  ,   217 ,   1   ,   305 ,   1257    ,   3773    ,   3509    ,0 };
            const unsigned long dim826Kuo3Init[]    =   {   1   ,   3   ,   7   ,   7   ,   5   ,   29  ,   47  ,   245 ,   487 ,   525 ,   1409    ,   245 ,   6427    ,0 };
            const unsigned long dim827Kuo3Init[]    =   {   1   ,   3   ,   5   ,   15  ,   25  ,   39  ,   49  ,   153 ,   461 ,   1003    ,   1527    ,   2883    ,   3499    ,0 };
            const unsigned long dim828Kuo3Init[]    =   {   1   ,   1   ,   7   ,   13  ,   23  ,   11  ,   15  ,   163 ,   9   ,   77  ,   1853    ,   1679    ,   1023    ,0 };
            const unsigned long dim829Kuo3Init[]    =   {   1   ,   3   ,   7   ,   13  ,   3   ,   55  ,   101 ,   229 ,   327 ,   537 ,   79  ,   3947    ,   6259    ,0 };
            const unsigned long dim830Kuo3Init[]    =   {   1   ,   3   ,   7   ,   5   ,   9   ,   39  ,   3   ,   151 ,   439 ,   181 ,   1473    ,   3613    ,   7425    ,0 };
            const unsigned long dim831Kuo3Init[]    =   {   1   ,   3   ,   7   ,   11  ,   21  ,   5   ,   53  ,   11  ,   39  ,   455 ,   1131    ,   393 ,   1525    ,0 };
            const unsigned long dim832Kuo3Init[]    =   {   1   ,   1   ,   1   ,   7   ,   5   ,   45  ,   49  ,   229 ,   173 ,   567 ,   1285    ,   2317    ,   2207    ,0 };
            const unsigned long dim833Kuo3Init[]    =   {   1   ,   1   ,   1   ,   7   ,   5   ,   33  ,   53  ,   241 ,   477 ,   161 ,   1607    ,   1253    ,   4497    ,0 };
            const unsigned long dim834Kuo3Init[]    =   {   1   ,   1   ,   7   ,   15  ,   11  ,   3   ,   107 ,   187 ,   249 ,   529 ,   905 ,   2883    ,   7777    ,0 };
            const unsigned long dim835Kuo3Init[]    =   {   1   ,   1   ,   5   ,   7   ,   17  ,   13  ,   123 ,   7   ,   449 ,   627 ,   1679    ,   569 ,   4145    ,0 };
            const unsigned long dim836Kuo3Init[]    =   {   1   ,   1   ,   1   ,   1   ,   3   ,   29  ,   57  ,   219 ,   209 ,   51  ,   1475    ,   2989    ,   7373    ,0 };
            const unsigned long dim837Kuo3Init[]    =   {   1   ,   3   ,   3   ,   5   ,   19  ,   33  ,   95  ,   241 ,   257 ,   465 ,   1305    ,   189 ,   269 ,0 };
            const unsigned long dim838Kuo3Init[]    =   {   1   ,   1   ,   3   ,   1   ,   13  ,   11  ,   67  ,   53  ,   157 ,   693 ,   1583    ,   739 ,   3827    ,0 };
            const unsigned long dim839Kuo3Init[]    =   {   1   ,   1   ,   3   ,   9   ,   17  ,   9   ,   5   ,   7   ,   421 ,   425 ,   1851    ,   3847    ,   5147    ,0 };
            const unsigned long dim840Kuo3Init[]    =   {   1   ,   1   ,   3   ,   3   ,   23  ,   47  ,   123 ,   215 ,   11  ,   643 ,   1543    ,   1859    ,   4073    ,0 };
            const unsigned long dim841Kuo3Init[]    =   {   1   ,   1   ,   1   ,   11  ,   5   ,   35  ,   79  ,   55  ,   225 ,   773 ,   1717    ,   2891    ,   563 ,0 };
            const unsigned long dim842Kuo3Init[]    =   {   1   ,   1   ,   1   ,   15  ,   5   ,   17  ,   115 ,   3   ,   107 ,   33  ,   1897    ,   3677    ,   4313    ,0 };
            const unsigned long dim843Kuo3Init[]    =   {   1   ,   1   ,   5   ,   7   ,   5   ,   57  ,   75  ,   65  ,   177 ,   269 ,   733 ,   3837    ,   1211    ,0 };
            const unsigned long dim844Kuo3Init[]    =   {   1   ,   1   ,   5   ,   15  ,   25  ,   45  ,   125 ,   159 ,   501 ,   17  ,   693 ,   1083    ,   1801    ,0 };
            const unsigned long dim845Kuo3Init[]    =   {   1   ,   3   ,   5   ,   3   ,   15  ,   17  ,   17  ,   159 ,   421 ,   449 ,   83  ,   1367    ,   7719    ,0 };
            const unsigned long dim846Kuo3Init[]    =   {   1   ,   3   ,   7   ,   3   ,   7   ,   31  ,   121 ,   43  ,   491 ,   501 ,   335 ,   2731    ,   1959    ,0 };
            const unsigned long dim847Kuo3Init[]    =   {   1   ,   1   ,   1   ,   11  ,   7   ,   1   ,   43  ,   187 ,   437 ,   113 ,   1205    ,   1255    ,   1319    ,0 };
            const unsigned long dim848Kuo3Init[]    =   {   1   ,   1   ,   5   ,   15  ,   15  ,   49  ,   103 ,   139 ,   187 ,   697 ,   907 ,   1229    ,   3205    ,0 };
            const unsigned long dim849Kuo3Init[]    =   {   1   ,   1   ,   5   ,   13  ,   9   ,   19  ,   25  ,   89  ,   345 ,   229 ,   1325    ,   859 ,   527 ,0 };
            const unsigned long dim850Kuo3Init[]    =   {   1   ,   1   ,   1   ,   3   ,   3   ,   63  ,   105 ,   11  ,   407 ,   131 ,   679 ,   3069    ,   4153    ,0 };
            const unsigned long dim851Kuo3Init[]    =   {   1   ,   1   ,   3   ,   11  ,   21  ,   27  ,   113 ,   59  ,   409 ,   845 ,   309 ,   2639    ,   129 ,0 };
            const unsigned long dim852Kuo3Init[]    =   {   1   ,   1   ,   5   ,   15  ,   21  ,   47  ,   55  ,   237 ,   201 ,   103 ,   331 ,   3531    ,   4965    ,0 };
            const unsigned long dim853Kuo3Init[]    =   {   1   ,   1   ,   5   ,   11  ,   9   ,   11  ,   85  ,   49  ,   209 ,   333 ,   1993    ,   511 ,   7347    ,0 };
            const unsigned long dim854Kuo3Init[]    =   {   1   ,   1   ,   5   ,   13  ,   25  ,   49  ,   109 ,   241 ,   297 ,   677 ,   313 ,   657 ,   4503    ,0 };
            const unsigned long dim855Kuo3Init[]    =   {   1   ,   1   ,   1   ,   15  ,   23  ,   57  ,   61  ,   117 ,   461 ,   559 ,   837 ,   2781    ,   709 ,0 };
            const unsigned long dim856Kuo3Init[]    =   {   1   ,   1   ,   3   ,   3   ,   31  ,   13  ,   109 ,   167 ,   389 ,   1021    ,   1437    ,   31  ,   811 ,0 };
            const unsigned long dim857Kuo3Init[]    =   {   1   ,   1   ,   1   ,   9   ,   21  ,   17  ,   19  ,   61  ,   241 ,   445 ,   1935    ,   3161    ,   4061    ,0 };
            const unsigned long dim858Kuo3Init[]    =   {   1   ,   3   ,   5   ,   5   ,   17  ,   5   ,   59  ,   117 ,   241 ,   913 ,   1315    ,   1597    ,   713 ,0 };
            const unsigned long dim859Kuo3Init[]    =   {   1   ,   3   ,   1   ,   15  ,   5   ,   41  ,   91  ,   191 ,   43  ,   49  ,   1491    ,   4051    ,   2317    ,0 };
            const unsigned long dim860Kuo3Init[]    =   {   1   ,   1   ,   1   ,   3   ,   13  ,   33  ,   53  ,   39  ,   341 ,   341 ,   617 ,   2627    ,   421 ,0 };
            const unsigned long dim861Kuo3Init[]    =   {   1   ,   3   ,   3   ,   15  ,   23  ,   25  ,   47  ,   251 ,   405 ,   659 ,   721 ,   229 ,   2269    ,0 };
            const unsigned long dim862Kuo3Init[]    =   {   1   ,   1   ,   3   ,   3   ,   25  ,   19  ,   123 ,   53  ,   59  ,   805 ,   2033    ,   1693    ,   6223    ,0 };
            const unsigned long dim863Kuo3Init[]    =   {   1   ,   1   ,   1   ,   3   ,   13  ,   45  ,   67  ,   67  ,   199 ,   925 ,   145 ,   775 ,   7923    ,0 };
            const unsigned long dim864Kuo3Init[]    =   {   1   ,   1   ,   5   ,   7   ,   17  ,   23  ,   49  ,   37  ,   185 ,   253 ,   1743    ,   2771    ,   6899    ,0 };
            const unsigned long dim865Kuo3Init[]    =   {   1   ,   3   ,   5   ,   5   ,   15  ,   53  ,   75  ,   185 ,   17  ,   379 ,   1091    ,   4029    ,   5399    ,0 };
            const unsigned long dim866Kuo3Init[]    =   {   1   ,   1   ,   1   ,   9   ,   13  ,   37  ,   101 ,   33  ,   499 ,   723 ,   57  ,   467 ,   7227    ,0 };
            const unsigned long dim867Kuo3Init[]    =   {   1   ,   1   ,   7   ,   1   ,   21  ,   49  ,   17  ,   49  ,   417 ,   895 ,   1099    ,   3199    ,   7059    ,0 };
            const unsigned long dim868Kuo3Init[]    =   {   1   ,   1   ,   5   ,   13  ,   27  ,   35  ,   49  ,   35  ,   213 ,   799 ,   197 ,   3599    ,   4135    ,0 };
            const unsigned long dim869Kuo3Init[]    =   {   1   ,   3   ,   5   ,   3   ,   31  ,   33  ,   61  ,   127 ,   87  ,   343 ,   845 ,   1699    ,   3897    ,0 };
            const unsigned long dim870Kuo3Init[]    =   {   1   ,   3   ,   1   ,   11  ,   7   ,   59  ,   19  ,   161 ,   119 ,   377 ,   1637    ,   2025    ,   849 ,0 };
            const unsigned long dim871Kuo3Init[]    =   {   1   ,   1   ,   3   ,   3   ,   3   ,   51  ,   3   ,   67  ,   77  ,   357 ,   295 ,   2383    ,   7423    ,0 };
            const unsigned long dim872Kuo3Init[]    =   {   1   ,   3   ,   1   ,   3   ,   23  ,   25  ,   67  ,   29  ,   433 ,   49  ,   901 ,   1641    ,   3021    ,0 };
            const unsigned long dim873Kuo3Init[]    =   {   1   ,   3   ,   7   ,   3   ,   25  ,   59  ,   15  ,   181 ,   193 ,   809 ,   1089    ,   423 ,   1629    ,0 };
            const unsigned long dim874Kuo3Init[]    =   {   1   ,   3   ,   5   ,   1   ,   3   ,   19  ,   53  ,   45  ,   277 ,   411 ,   455 ,   9   ,   3651    ,0 };
            const unsigned long dim875Kuo3Init[]    =   {   1   ,   3   ,   7   ,   15  ,   13  ,   33  ,   37  ,   15  ,   489 ,   275 ,   1113    ,   3371    ,   3915    ,0 };
            const unsigned long dim876Kuo3Init[]    =   {   1   ,   1   ,   5   ,   11  ,   3   ,   3   ,   97  ,   47  ,   471 ,   713 ,   1975    ,   1679    ,   7901    ,0 };
            const unsigned long dim877Kuo3Init[]    =   {   1   ,   1   ,   3   ,   15  ,   31  ,   47  ,   13  ,   33  ,   111 ,   607 ,   85  ,   4049    ,   2937    ,0 };
            const unsigned long dim878Kuo3Init[]    =   {   1   ,   1   ,   5   ,   1   ,   7   ,   61  ,   17  ,   179 ,   301 ,   245 ,   67  ,   2823    ,   4121    ,0 };
            const unsigned long dim879Kuo3Init[]    =   {   1   ,   3   ,   7   ,   15  ,   7   ,   25  ,   97  ,   73  ,   19  ,   899 ,   1357    ,   2041    ,   6493    ,0 };
            const unsigned long dim880Kuo3Init[]    =   {   1   ,   1   ,   1   ,   3   ,   3   ,   63  ,   117 ,   231 ,   125 ,   527 ,   1795    ,   1807    ,   7331    ,0 };
            const unsigned long dim881Kuo3Init[]    =   {   1   ,   1   ,   5   ,   11  ,   17  ,   35  ,   61  ,   161 ,   415 ,   687 ,   1015    ,   1239    ,   6925    ,0 };
            const unsigned long dim882Kuo3Init[]    =   {   1   ,   3   ,   3   ,   5   ,   15  ,   19  ,   1   ,   5   ,   209 ,   545 ,   1341    ,   4019    ,   6177    ,0 };
            const unsigned long dim883Kuo3Init[]    =   {   1   ,   3   ,   5   ,   15  ,   13  ,   3   ,   7   ,   39  ,   63  ,   727 ,   1033    ,   2551    ,   449 ,0 };
            const unsigned long dim884Kuo3Init[]    =   {   1   ,   1   ,   1   ,   5   ,   11  ,   35  ,   39  ,   143 ,   361 ,   659 ,   1407    ,   2399    ,   6421    ,0 };
            const unsigned long dim885Kuo3Init[]    =   {   1   ,   1   ,   7   ,   13  ,   3   ,   23  ,   21  ,   161 ,   153 ,   595 ,   1847    ,   1339    ,   1887    ,0 };
            const unsigned long dim886Kuo3Init[]    =   {   1   ,   3   ,   3   ,   9   ,   3   ,   43  ,   75  ,   45  ,   433 ,   811 ,   1045    ,   2773    ,   2415    ,0 };
            const unsigned long dim887Kuo3Init[]    =   {   1   ,   1   ,   3   ,   5   ,   23  ,   29  ,   73  ,   167 ,   211 ,   403 ,   539 ,   13  ,   4433    ,0 };
            const unsigned long dim888Kuo3Init[]    =   {   1   ,   3   ,   3   ,   13  ,   19  ,   25  ,   113 ,   181 ,   343 ,   349 ,   931 ,   115 ,   5687    ,0 };
            const unsigned long dim889Kuo3Init[]    =   {   1   ,   3   ,   1   ,   11  ,   25  ,   39  ,   23  ,   91  ,   359 ,   221 ,   37  ,   575 ,   3035    ,0 };
            const unsigned long dim890Kuo3Init[]    =   {   1   ,   1   ,   7   ,   1   ,   19  ,   61  ,   99  ,   81  ,   167 ,   53  ,   1493    ,   2623    ,   5797    ,0 };
            const unsigned long dim891Kuo3Init[]    =   {   1   ,   1   ,   3   ,   5   ,   13  ,   13  ,   105 ,   53  ,   449 ,   1003    ,   1405    ,   1403    ,   2557    ,0 };
            const unsigned long dim892Kuo3Init[]    =   {   1   ,   1   ,   3   ,   15  ,   1   ,   37  ,   51  ,   69  ,   13  ,   5   ,   1865    ,   3585    ,   3409    ,0 };
            const unsigned long dim893Kuo3Init[]    =   {   1   ,   1   ,   7   ,   13  ,   15  ,   31  ,   61  ,   53  ,   439 ,   665 ,   57  ,   445 ,   435 ,0 };
            const unsigned long dim894Kuo3Init[]    =   {   1   ,   3   ,   1   ,   3   ,   1   ,   27  ,   1   ,   155 ,   65  ,   191 ,   449 ,   2185    ,   1787    ,0 };
            const unsigned long dim895Kuo3Init[]    =   {   1   ,   1   ,   3   ,   11  ,   13  ,   59  ,   15  ,   43  ,   435 ,   477 ,   823 ,   717 ,   299 ,0 };
            const unsigned long dim896Kuo3Init[]    =   {   1   ,   1   ,   3   ,   1   ,   23  ,   9   ,   85  ,   155 ,   127 ,   543 ,   1081    ,   221 ,   7357    ,0 };
            const unsigned long dim897Kuo3Init[]    =   {   1   ,   3   ,   3   ,   1   ,   7   ,   35  ,   77  ,   9   ,   79  ,   283 ,   449 ,   2729    ,   4193    ,0 };
            const unsigned long dim898Kuo3Init[]    =   {   1   ,   1   ,   7   ,   1   ,   9   ,   15  ,   97  ,   99  ,   363 ,   513 ,   1063    ,   3051    ,   1529    ,0 };
            const unsigned long dim899Kuo3Init[]    =   {   1   ,   1   ,   5   ,   3   ,   1   ,   17  ,   101 ,   19  ,   425 ,   377 ,   1369    ,   3071    ,   6877    ,0 };
            const unsigned long dim900Kuo3Init[]    =   {   1   ,   1   ,   5   ,   1   ,   23  ,   63  ,   11  ,   7   ,   29  ,   717 ,   963 ,   2205    ,   7487    ,0 };
            const unsigned long dim901Kuo3Init[]    =   {   1   ,   3   ,   1   ,   13  ,   1   ,   15  ,   43  ,   151 ,   59  ,   315 ,   691 ,   2765    ,   7723    ,0 };
            const unsigned long dim902Kuo3Init[]    =   {   1   ,   3   ,   1   ,   5   ,   5   ,   31  ,   17  ,   81  ,   235 ,   419 ,   123 ,   2983    ,   1797    ,0 };
            const unsigned long dim903Kuo3Init[]    =   {   1   ,   1   ,   1   ,   11  ,   29  ,   17  ,   123 ,   115 ,   91  ,   531 ,   1393    ,   3537    ,   6387    ,0 };
            const unsigned long dim904Kuo3Init[]    =   {   1   ,   3   ,   5   ,   11  ,   31  ,   5   ,   107 ,   61  ,   463 ,   507 ,   1937    ,   4075    ,   4727    ,0 };
            const unsigned long dim905Kuo3Init[]    =   {   1   ,   3   ,   7   ,   1   ,   9   ,   27  ,   23  ,   97  ,   125 ,   511 ,   239 ,   1031    ,   153 ,0 };
            const unsigned long dim906Kuo3Init[]    =   {   1   ,   3   ,   5   ,   9   ,   13  ,   55  ,   53  ,   195 ,   265 ,   217 ,   357 ,   567 ,   1613    ,0 };
            const unsigned long dim907Kuo3Init[]    =   {   1   ,   1   ,   5   ,   15  ,   5   ,   51  ,   51  ,   235 ,   63  ,   393 ,   1915    ,   2495    ,   7067    ,0 };
            const unsigned long dim908Kuo3Init[]    =   {   1   ,   1   ,   7   ,   1   ,   15  ,   35  ,   97  ,   65  ,   115 ,   819 ,   1419    ,   2313    ,   6029    ,0 };
            const unsigned long dim909Kuo3Init[]    =   {   1   ,   1   ,   5   ,   13  ,   11  ,   35  ,   45  ,   149 ,   45  ,   617 ,   2001    ,   1629    ,   5611    ,0 };
            const unsigned long dim910Kuo3Init[]    =   {   1   ,   3   ,   5   ,   11  ,   15  ,   17  ,   39  ,   155 ,   143 ,   473 ,   1269    ,   2537    ,   2571    ,0 };
            const unsigned long dim911Kuo3Init[]    =   {   1   ,   1   ,   3   ,   3   ,   1   ,   7   ,   57  ,   227 ,   151 ,   731 ,   1871    ,   245 ,   7605    ,0 };
            const unsigned long dim912Kuo3Init[]    =   {   1   ,   1   ,   1   ,   13  ,   21  ,   1   ,   19  ,   189 ,   299 ,   483 ,   513 ,   2235    ,   3079    ,0 };
            const unsigned long dim913Kuo3Init[]    =   {   1   ,   1   ,   5   ,   15  ,   13  ,   53  ,   91  ,   55  ,   339 ,   161 ,   273 ,   2759    ,   7251    ,0 };
            const unsigned long dim914Kuo3Init[]    =   {   1   ,   3   ,   1   ,   1   ,   17  ,   47  ,   29  ,   107 ,   371 ,   437 ,   715 ,   739 ,   2195    ,0 };
            const unsigned long dim915Kuo3Init[]    =   {   1   ,   3   ,   5   ,   11  ,   27  ,   1   ,   35  ,   105 ,   443 ,   623 ,   1401    ,   2683    ,   4327    ,0 };
            const unsigned long dim916Kuo3Init[]    =   {   1   ,   1   ,   5   ,   7   ,   13  ,   23  ,   101 ,   235 ,   369 ,   951 ,   1983    ,   1021    ,   7835    ,0 };
            const unsigned long dim917Kuo3Init[]    =   {   1   ,   1   ,   1   ,   5   ,   21  ,   57  ,   45  ,   67  ,   89  ,   209 ,   1105    ,   2161    ,   2245    ,0 };
            const unsigned long dim918Kuo3Init[]    =   {   1   ,   1   ,   7   ,   15  ,   7   ,   33  ,   69  ,   129 ,   181 ,   231 ,   961 ,   1621    ,   7763    ,0 };
            const unsigned long dim919Kuo3Init[]    =   {   1   ,   3   ,   1   ,   5   ,   1   ,   31  ,   105 ,   123 ,   413 ,   989 ,   1311    ,   199 ,   5175    ,0 };
            const unsigned long dim920Kuo3Init[]    =   {   1   ,   3   ,   5   ,   13  ,   15  ,   9   ,   123 ,   215 ,   463 ,   197 ,   537 ,   531 ,   2331    ,0 };
            const unsigned long dim921Kuo3Init[]    =   {   1   ,   1   ,   3   ,   13  ,   13  ,   45  ,   61  ,   189 ,   91  ,   445 ,   1719    ,   3335    ,   5269    ,0 };
            const unsigned long dim922Kuo3Init[]    =   {   1   ,   3   ,   1   ,   11  ,   19  ,   41  ,   65  ,   143 ,   199 ,   177 ,   1695    ,   3647    ,   7169    ,0 };
            const unsigned long dim923Kuo3Init[]    =   {   1   ,   1   ,   3   ,   13  ,   13  ,   25  ,   37  ,   89  ,   175 ,   551 ,   1535    ,   1401    ,   1823    ,0 };
            const unsigned long dim924Kuo3Init[]    =   {   1   ,   3   ,   3   ,   1   ,   21  ,   15  ,   31  ,   101 ,   329 ,   907 ,   731 ,   1493    ,   451 ,0 };
            const unsigned long dim925Kuo3Init[]    =   {   1   ,   1   ,   3   ,   9   ,   5   ,   61  ,   85  ,   247 ,   289 ,   547 ,   1031    ,   3567    ,   1769    ,0 };
            const unsigned long dim926Kuo3Init[]    =   {   1   ,   3   ,   1   ,   7   ,   25  ,   35  ,   99  ,   43  ,   25  ,   855 ,   1451    ,   617 ,   2615    ,0 };
            const unsigned long dim927Kuo3Init[]    =   {   1   ,   1   ,   1   ,   7   ,   25  ,   9   ,   77  ,   39  ,   127 ,   327 ,   1501    ,   57  ,   1485    ,0 };
            const unsigned long dim928Kuo3Init[]    =   {   1   ,   3   ,   5   ,   5   ,   9   ,   13  ,   13  ,   239 ,   125 ,   473 ,   785 ,   2853    ,   5165    ,0 };
            const unsigned long dim929Kuo3Init[]    =   {   1   ,   1   ,   1   ,   3   ,   27  ,   43  ,   117 ,   63  ,   175 ,   487 ,   1505    ,   157 ,   6043    ,0 };
            const unsigned long dim930Kuo3Init[]    =   {   1   ,   1   ,   5   ,   5   ,   25  ,   43  ,   97  ,   231 ,   389 ,   191 ,   1795    ,   2659    ,   7409    ,0 };
            const unsigned long dim931Kuo3Init[]    =   {   1   ,   1   ,   7   ,   7   ,   11  ,   25  ,   47  ,   131 ,   363 ,   311 ,   523 ,   2773    ,   8015    ,0 };
            const unsigned long dim932Kuo3Init[]    =   {   1   ,   1   ,   5   ,   11  ,   25  ,   61  ,   123 ,   83  ,   265 ,   995 ,   613 ,   1893    ,   6341    ,0 };
            const unsigned long dim933Kuo3Init[]    =   {   1   ,   1   ,   3   ,   3   ,   27  ,   45  ,   41  ,   143 ,   173 ,   1009    ,   1963    ,   4009    ,   3675    ,0 };
            const unsigned long dim934Kuo3Init[]    =   {   1   ,   3   ,   3   ,   9   ,   5   ,   11  ,   95  ,   253 ,   415 ,   307 ,   501 ,   3117    ,   4621    ,0 };
            const unsigned long dim935Kuo3Init[]    =   {   1   ,   3   ,   5   ,   1   ,   31  ,   41  ,   95  ,   39  ,   473 ,   565 ,   1389    ,   4049    ,   5673    ,0 };
            const unsigned long dim936Kuo3Init[]    =   {   1   ,   1   ,   3   ,   11  ,   3   ,   13  ,   87  ,   101 ,   413 ,   769 ,   1223    ,   549 ,   4967    ,0 };
            const unsigned long dim937Kuo3Init[]    =   {   1   ,   3   ,   1   ,   3   ,   13  ,   41  ,   29  ,   187 ,   491 ,   197 ,   1685    ,   3859    ,   3799    ,0 };
            const unsigned long dim938Kuo3Init[]    =   {   1   ,   3   ,   1   ,   3   ,   17  ,   55  ,   91  ,   131 ,   429 ,   899 ,   29  ,   1193    ,   7323    ,0 };
            const unsigned long dim939Kuo3Init[]    =   {   1   ,   1   ,   3   ,   1   ,   11  ,   17  ,   107 ,   113 ,   71  ,   185 ,   1057    ,   897 ,   249 ,0 };
            const unsigned long dim940Kuo3Init[]    =   {   1   ,   1   ,   3   ,   5   ,   29  ,   27  ,   87  ,   33  ,   273 ,   405 ,   547 ,   763 ,   2195    ,0 };
            const unsigned long dim941Kuo3Init[]    =   {   1   ,   3   ,   1   ,   11  ,   7   ,   43  ,   39  ,   103 ,   475 ,   515 ,   1105    ,   767 ,   3881    ,0 };
            const unsigned long dim942Kuo3Init[]    =   {   1   ,   3   ,   3   ,   1   ,   13  ,   21  ,   93  ,   141 ,   407 ,   1001    ,   1939    ,   2747    ,   6479    ,0 };
            const unsigned long dim943Kuo3Init[]    =   {   1   ,   3   ,   1   ,   15  ,   1   ,   31  ,   15  ,   69  ,   313 ,   547 ,   987 ,   3569    ,   2795    ,0 };
            const unsigned long dim944Kuo3Init[]    =   {   1   ,   3   ,   5   ,   7   ,   1   ,   17  ,   105 ,   113 ,   253 ,   453 ,   1219    ,   2769    ,   197 ,0 };
            const unsigned long dim945Kuo3Init[]    =   {   1   ,   3   ,   7   ,   7   ,   29  ,   55  ,   77  ,   169 ,   227 ,   471 ,   1385    ,   2689    ,   7995    ,0 };
            const unsigned long dim946Kuo3Init[]    =   {   1   ,   3   ,   7   ,   15  ,   7   ,   15  ,   113 ,   133 ,   205 ,   271 ,   263 ,   3141    ,   5007    ,0 };
            const unsigned long dim947Kuo3Init[]    =   {   1   ,   1   ,   5   ,   7   ,   7   ,   57  ,   121 ,   209 ,   319 ,   733 ,   1689    ,   2817    ,   6621    ,0 };
            const unsigned long dim948Kuo3Init[]    =   {   1   ,   3   ,   7   ,   9   ,   23  ,   37  ,   101 ,   229 ,   269 ,   429 ,   157 ,   515 ,   6975    ,0 };
            const unsigned long dim949Kuo3Init[]    =   {   1   ,   1   ,   7   ,   7   ,   23  ,   31  ,   83  ,   237 ,   299 ,   333 ,   959 ,   1467    ,   419 ,0 };
            const unsigned long dim950Kuo3Init[]    =   {   1   ,   3   ,   7   ,   3   ,   23  ,   45  ,   17  ,   33  ,   293 ,   679 ,   625 ,   3651    ,   4205    ,0 };
            const unsigned long dim951Kuo3Init[]    =   {   1   ,   3   ,   1   ,   13  ,   17  ,   19  ,   87  ,   237 ,   219 ,   119 ,   841 ,   1923    ,   8141    ,0 };
            const unsigned long dim952Kuo3Init[]    =   {   1   ,   1   ,   1   ,   15  ,   15  ,   49  ,   21  ,   189 ,   287 ,   759 ,   695 ,   913 ,   5167    ,0 };
            const unsigned long dim953Kuo3Init[]    =   {   1   ,   1   ,   3   ,   15  ,   1   ,   9   ,   125 ,   5   ,   67  ,   603 ,   1691    ,   2863    ,   1299    ,0 };
            const unsigned long dim954Kuo3Init[]    =   {   1   ,   3   ,   7   ,   3   ,   5   ,   35  ,   105 ,   27  ,   51  ,   889 ,   1469    ,   3203    ,   6997    ,0 };
            const unsigned long dim955Kuo3Init[]    =   {   1   ,   3   ,   5   ,   7   ,   1   ,   53  ,   77  ,   1   ,   313 ,   475 ,   1515    ,   2233    ,   399 ,0 };
            const unsigned long dim956Kuo3Init[]    =   {   1   ,   3   ,   5   ,   7   ,   3   ,   31  ,   43  ,   17  ,   425 ,   673 ,   1817    ,   1939    ,   8027    ,0 };
            const unsigned long dim957Kuo3Init[]    =   {   1   ,   1   ,   5   ,   5   ,   31  ,   31  ,   119 ,   179 ,   157 ,   635 ,   1   ,   315 ,   1179    ,0 };
            const unsigned long dim958Kuo3Init[]    =   {   1   ,   1   ,   5   ,   5   ,   1   ,   7   ,   25  ,   15  ,   311 ,   489 ,   829 ,   1929    ,   623 ,0 };
            const unsigned long dim959Kuo3Init[]    =   {   1   ,   1   ,   5   ,   11  ,   1   ,   17  ,   43  ,   205 ,   347 ,   713 ,   2029    ,   977 ,   4345    ,0 };
            const unsigned long dim960Kuo3Init[]    =   {   1   ,   1   ,   3   ,   9   ,   17  ,   3   ,   109 ,   253 ,   427 ,   333 ,   203 ,   1609    ,   1851    ,0 };
            const unsigned long dim961Kuo3Init[]    =   {   1   ,   3   ,   1   ,   9   ,   21  ,   55  ,   107 ,   203 ,   483 ,   869 ,   1417    ,   587 ,   3965    ,0 };
            const unsigned long dim962Kuo3Init[]    =   {   1   ,   3   ,   3   ,   13  ,   27  ,   21  ,   103 ,   109 ,   313 ,   797 ,   723 ,   683 ,   349 ,0 };
            const unsigned long dim963Kuo3Init[]    =   {   1   ,   3   ,   5   ,   3   ,   1   ,   35  ,   43  ,   19  ,   367 ,   487 ,   703 ,   2211    ,   1199    ,0 };
            const unsigned long dim964Kuo3Init[]    =   {   1   ,   3   ,   1   ,   3   ,   19  ,   33  ,   13  ,   135 ,   329 ,   621 ,   843 ,   407 ,   6523    ,0 };
            const unsigned long dim965Kuo3Init[]    =   {   1   ,   1   ,   7   ,   5   ,   13  ,   1   ,   45  ,   133 ,   27  ,   267 ,   1975    ,   2761    ,   5061    ,0 };
            const unsigned long dim966Kuo3Init[]    =   {   1   ,   1   ,   3   ,   3   ,   19  ,   59  ,   13  ,   35  ,   489 ,   391 ,   1595    ,   2983    ,   3203    ,0 };
            const unsigned long dim967Kuo3Init[]    =   {   1   ,   3   ,   3   ,   7   ,   19  ,   27  ,   125 ,   55  ,   121 ,   733 ,   1437    ,   51  ,   4029    ,0 };
            const unsigned long dim968Kuo3Init[]    =   {   1   ,   3   ,   1   ,   15  ,   11  ,   7   ,   73  ,   187 ,   209 ,   127 ,   1259    ,   1733    ,   1897    ,0 };
            const unsigned long dim969Kuo3Init[]    =   {   1   ,   3   ,   7   ,   1   ,   27  ,   19  ,   87  ,   189 ,   91  ,   113 ,   237 ,   2073    ,   419 ,0 };
            const unsigned long dim970Kuo3Init[]    =   {   1   ,   3   ,   5   ,   3   ,   15  ,   27  ,   107 ,   77  ,   467 ,   953 ,   1317    ,   877 ,   6911    ,0 };
            const unsigned long dim971Kuo3Init[]    =   {   1   ,   3   ,   7   ,   13  ,   7   ,   17  ,   107 ,   157 ,   439 ,   871 ,   651 ,   797 ,   2095    ,0 };
            const unsigned long dim972Kuo3Init[]    =   {   1   ,   1   ,   3   ,   9   ,   13  ,   19  ,   19  ,   75  ,   13  ,   267 ,   435 ,   1067    ,   6603    ,0 };
            const unsigned long dim973Kuo3Init[]    =   {   1   ,   1   ,   7   ,   13  ,   25  ,   7   ,   81  ,   207 ,   149 ,   587 ,   1561    ,   2025    ,   2355    ,0 };
            const unsigned long dim974Kuo3Init[]    =   {   1   ,   3   ,   3   ,   5   ,   25  ,   25  ,   79  ,   249 ,   371 ,   305 ,   305 ,   53  ,   1051    ,0 };
            const unsigned long dim975Kuo3Init[]    =   {   1   ,   1   ,   3   ,   9   ,   19  ,   63  ,   57  ,   237 ,   39  ,   349 ,   1163    ,   2443    ,   1739    ,0 };
            const unsigned long dim976Kuo3Init[]    =   {   1   ,   1   ,   5   ,   5   ,   23  ,   57  ,   61  ,   21  ,   115 ,   107 ,   303 ,   1451    ,   2955    ,0 };
            const unsigned long dim977Kuo3Init[]    =   {   1   ,   3   ,   1   ,   3   ,   17  ,   19  ,   121 ,   123 ,   25  ,   367 ,   963 ,   545 ,   2885    ,0 };
            const unsigned long dim978Kuo3Init[]    =   {   1   ,   3   ,   3   ,   5   ,   15  ,   57  ,   123 ,   1   ,   193 ,   861 ,   1001    ,   1457    ,   4977    ,0 };
            const unsigned long dim979Kuo3Init[]    =   {   1   ,   3   ,   5   ,   9   ,   21  ,   31  ,   37  ,   83  ,   361 ,   491 ,   1933    ,   407 ,   1923    ,0 };
            const unsigned long dim980Kuo3Init[]    =   {   1   ,   3   ,   7   ,   5   ,   9   ,   47  ,   33  ,   43  ,   239 ,   837 ,   799 ,   3731    ,   5883    ,0 };
            const unsigned long dim981Kuo3Init[]    =   {   1   ,   3   ,   5   ,   5   ,   7   ,   39  ,   49  ,   201 ,   339 ,   783 ,   1937    ,   1473    ,   391 ,0 };
            const unsigned long dim982Kuo3Init[]    =   {   1   ,   3   ,   5   ,   13  ,   3   ,   63  ,   39  ,   91  ,   63  ,   645 ,   929 ,   1543    ,   577 ,0 };
            const unsigned long dim983Kuo3Init[]    =   {   1   ,   3   ,   3   ,   3   ,   9   ,   59  ,   107 ,   129 ,   455 ,   259 ,   1523    ,   107 ,   285 ,0 };
            const unsigned long dim984Kuo3Init[]    =   {   1   ,   1   ,   3   ,   9   ,   23  ,   21  ,   89  ,   13  ,   413 ,   87  ,   1703    ,   1611    ,   5667    ,0 };
            const unsigned long dim985Kuo3Init[]    =   {   1   ,   1   ,   3   ,   9   ,   21  ,   9   ,   83  ,   187 ,   159 ,   757 ,   1111    ,   3619    ,   3907    ,0 };
            const unsigned long dim986Kuo3Init[]    =   {   1   ,   3   ,   3   ,   9   ,   21  ,   35  ,   27  ,   57  ,   357 ,   659 ,   1243    ,   2725    ,   5769    ,0 };
            const unsigned long dim987Kuo3Init[]    =   {   1   ,   1   ,   7   ,   1   ,   3   ,   7   ,   75  ,   145 ,   119 ,   1019    ,   1169    ,   953 ,   2343    ,0 };
            const unsigned long dim988Kuo3Init[]    =   {   1   ,   3   ,   3   ,   11  ,   19  ,   15  ,   121 ,   135 ,   187 ,   569 ,   671 ,   131 ,   479 ,0 };
            const unsigned long dim989Kuo3Init[]    =   {   1   ,   3   ,   5   ,   1   ,   21  ,   27  ,   63  ,   187 ,   367 ,   181 ,   1885    ,   2513    ,   4499    ,0 };
            const unsigned long dim990Kuo3Init[]    =   {   1   ,   3   ,   3   ,   5   ,   27  ,   13  ,   107 ,   183 ,   345 ,   855 ,   123 ,   1491    ,   867 ,0 };
            const unsigned long dim991Kuo3Init[]    =   {   1   ,   3   ,   7   ,   9   ,   5   ,   41  ,   43  ,   47  ,   277 ,   657 ,   1227    ,   2265    ,   1777    ,0 };
            const unsigned long dim992Kuo3Init[]    =   {   1   ,   1   ,   5   ,   5   ,   31  ,   37  ,   119 ,   5   ,   395 ,   399 ,   787 ,   2869    ,   829 ,0 };
            const unsigned long dim993Kuo3Init[]    =   {   1   ,   1   ,   1   ,   5   ,   3   ,   55  ,   43  ,   117 ,   191 ,   577 ,   1185    ,   3407    ,   4673    ,0 };
            const unsigned long dim994Kuo3Init[]    =   {   1   ,   3   ,   5   ,   11  ,   15  ,   63  ,   103 ,   127 ,   505 ,   53  ,   297 ,   3411    ,   3919    ,0 };
            const unsigned long dim995Kuo3Init[]    =   {   1   ,   3   ,   7   ,   13  ,   15  ,   57  ,   45  ,   17  ,   253 ,   261 ,   767 ,   1951    ,   6599    ,0 };
            const unsigned long dim996Kuo3Init[]    =   {   1   ,   3   ,   1   ,   13  ,   3   ,   55  ,   11  ,   113 ,   223 ,   403 ,   249 ,   633 ,   3265    ,0 };
            const unsigned long dim997Kuo3Init[]    =   {   1   ,   1   ,   1   ,   5   ,   7   ,   49  ,   47  ,   47  ,   467 ,   633 ,   1991    ,   361 ,   8099    ,0 };
            const unsigned long dim998Kuo3Init[]    =   {   1   ,   1   ,   3   ,   5   ,   31  ,   45  ,   87  ,   117 ,   483 ,   331 ,   687 ,   3393    ,   7377    ,0 };
            const unsigned long dim999Kuo3Init[]    =   {   1   ,   1   ,   7   ,   9   ,   29  ,   3   ,   11  ,   147 ,   147 ,   631 ,   1131    ,   2651    ,   3141    ,0 };
            const unsigned long dim1000Kuo3Init[]   =   {   1   ,   3   ,   7   ,   9   ,   27  ,   49  ,   41  ,   207 ,   305 ,   403 ,   889 ,   2947    ,   5775    ,0 };
            const unsigned long dim1001Kuo3Init[]   =   {   1   ,   3   ,   1   ,   13  ,   1   ,   63  ,   25  ,   127 ,   401 ,   375 ,   727 ,   2375    ,   2945    ,0 };
            const unsigned long dim1002Kuo3Init[]   =   {   1   ,   3   ,   3   ,   15  ,   7   ,   19  ,   65  ,   29  ,   383 ,   809 ,   279 ,   2081    ,   1993    ,0 };
            const unsigned long dim1003Kuo3Init[]   =   {   1   ,   1   ,   7   ,   7   ,   29  ,   9   ,   39  ,   159 ,   457 ,   435 ,   947 ,   2309    ,   2219    ,0 };
            const unsigned long dim1004Kuo3Init[]   =   {   1   ,   1   ,   1   ,   11  ,   15  ,   31  ,   51  ,   147 ,   167 ,   113 ,   145 ,   125 ,   6997    ,0 };
            const unsigned long dim1005Kuo3Init[]   =   {   1   ,   1   ,   5   ,   1   ,   27  ,   21  ,   123 ,   249 ,   21  ,   487 ,   331 ,   35  ,   3101    ,0 };
            const unsigned long dim1006Kuo3Init[]   =   {   1   ,   3   ,   7   ,   3   ,   5   ,   7   ,   127 ,   173 ,   115 ,   607 ,   1061    ,   2923    ,   6869    ,0 };
            const unsigned long dim1007Kuo3Init[]   =   {   1   ,   3   ,   1   ,   5   ,   17  ,   55  ,   13  ,   15  ,   95  ,   365 ,   1419    ,   1111    ,   7561    ,0 };
            const unsigned long dim1008Kuo3Init[]   =   {   1   ,   3   ,   3   ,   15  ,   23  ,   45  ,   115 ,   123 ,   477 ,   335 ,   85  ,   431 ,   6893    ,0 };
            const unsigned long dim1009Kuo3Init[]   =   {   1   ,   1   ,   5   ,   1   ,   5   ,   49  ,   95  ,   19  ,   429 ,   549 ,   713 ,   661 ,   167 ,0 };
            const unsigned long dim1010Kuo3Init[]   =   {   1   ,   1   ,   3   ,   15  ,   3   ,   13  ,   127 ,   189 ,   385 ,   649 ,   67  ,   499 ,   5355    ,0 };
            const unsigned long dim1011Kuo3Init[]   =   {   1   ,   1   ,   1   ,   5   ,   9   ,   31  ,   17  ,   21  ,   419 ,   655 ,   287 ,   1173    ,   1023    ,0 };
            const unsigned long dim1012Kuo3Init[]   =   {   1   ,   1   ,   5   ,   1   ,   17  ,   11  ,   19  ,   17  ,   195 ,   1015    ,   1621    ,   1197    ,   3497    ,0 };
            const unsigned long dim1013Kuo3Init[]   =   {   1   ,   3   ,   1   ,   5   ,   29  ,   55  ,   85  ,   43  ,   381 ,   59  ,   149 ,   391 ,   2109    ,0 };
            const unsigned long dim1014Kuo3Init[]   =   {   1   ,   1   ,   7   ,   9   ,   11  ,   43  ,   115 ,   57  ,   211 ,   463 ,   787 ,   3665    ,   5619    ,0 };
            const unsigned long dim1015Kuo3Init[]   =   {   1   ,   3   ,   1   ,   7   ,   23  ,   5   ,   55  ,   93  ,   117 ,   925 ,   941 ,   91  ,   6221    ,0 };
            const unsigned long dim1016Kuo3Init[]   =   {   1   ,   1   ,   1   ,   13  ,   29  ,   35  ,   89  ,   7   ,   75  ,   759 ,   735 ,   3011    ,   1279    ,0 };
            const unsigned long dim1017Kuo3Init[]   =   {   1   ,   1   ,   1   ,   15  ,   13  ,   17  ,   13  ,   175 ,   167 ,   965 ,   1069    ,   2215    ,   6065    ,0 };
            const unsigned long dim1018Kuo3Init[]   =   {   1   ,   1   ,   3   ,   15  ,   1   ,   11  ,   109 ,   205 ,   501 ,   419 ,   1071    ,   543 ,   5377    ,0 };
            const unsigned long dim1019Kuo3Init[]   =   {   1   ,   1   ,   3   ,   5   ,   11  ,   1   ,   123 ,   147 ,   87  ,   289 ,   325 ,   3693    ,   4939    ,0 };
            const unsigned long dim1020Kuo3Init[]   =   {   1   ,   1   ,   3   ,   1   ,   9   ,   21  ,   29  ,   215 ,   365 ,   995 ,   1727    ,   2515    ,   5741    ,0 };
            const unsigned long dim1021Kuo3Init[]   =   {   1   ,   1   ,   3   ,   5   ,   3   ,   7   ,   5   ,   147 ,   413 ,   505 ,   661 ,   3071    ,   2521    ,0 };
            const unsigned long dim1022Kuo3Init[]   =   {   1   ,   3   ,   7   ,   5   ,   5   ,   41  ,   117 ,   145 ,   455 ,   379 ,   601 ,   455 ,   5377    ,0 };
            const unsigned long dim1023Kuo3Init[]   =   {   1   ,   1   ,   7   ,   15  ,   11  ,   45  ,   27  ,   37  ,   457 ,   935 ,   595 ,   3883    ,   1491    ,0 };
            const unsigned long dim1024Kuo3Init[]   =   {   1   ,   3   ,   5   ,   9   ,   15  ,   7   ,   49  ,   95  ,   225 ,   837 ,   805 ,   2553    ,   4891    ,0 };
            const unsigned long dim1025Kuo3Init[]   =   {   1   ,   3   ,   3   ,   5   ,   1   ,   29  ,   109 ,   99  ,   349 ,   221 ,   749 ,   3499    ,   4201    ,0 };
            const unsigned long dim1026Kuo3Init[]   =   {   1   ,   1   ,   1   ,   11  ,   27  ,   35  ,   47  ,   253 ,   151 ,   129 ,   927 ,   4039    ,   6133    ,0 };
            const unsigned long dim1027Kuo3Init[]   =   {   1   ,   3   ,   1   ,   3   ,   17  ,   59  ,   1   ,   63  ,   61  ,   927 ,   933 ,   3645    ,   3707    ,0 };
            const unsigned long dim1028Kuo3Init[]   =   {   1   ,   3   ,   1   ,   3   ,   29  ,   33  ,   45  ,   59  ,   183 ,   459 ,   903 ,   747 ,   1895    ,0 };
            const unsigned long dim1029Kuo3Init[]   =   {   1   ,   1   ,   5   ,   3   ,   19  ,   53  ,   71  ,   199 ,   91  ,   807 ,   507 ,   3821    ,   1815    ,0 };
            const unsigned long dim1030Kuo3Init[]   =   {   1   ,   1   ,   1   ,   7   ,   23  ,   43  ,   39  ,   17  ,   139 ,   751 ,   467 ,   617 ,   5237    ,0 };
            const unsigned long dim1031Kuo3Init[]   =   {   1   ,   1   ,   7   ,   13  ,   5   ,   53  ,   43  ,   181 ,   337 ,   731 ,   497 ,   3189    ,   4063    ,0 };
            const unsigned long dim1032Kuo3Init[]   =   {   1   ,   1   ,   7   ,   9   ,   1   ,   45  ,   103 ,   151 ,   265 ,   897 ,   1821    ,   2957    ,   5591    ,0 };
            const unsigned long dim1033Kuo3Init[]   =   {   1   ,   3   ,   7   ,   5   ,   13  ,   43  ,   65  ,   49  ,   233 ,   1   ,   1709    ,   2641    ,   5633    ,0 };
            const unsigned long dim1034Kuo3Init[]   =   {   1   ,   1   ,   1   ,   7   ,   15  ,   15  ,   95  ,   223 ,   447 ,   967 ,   1533    ,   1969    ,   6497    ,0 };
            const unsigned long dim1035Kuo3Init[]   =   {   1   ,   1   ,   7   ,   9   ,   29  ,   5   ,   55  ,   57  ,   377 ,   487 ,   443 ,   2589    ,   4735    ,0 };
            const unsigned long dim1036Kuo3Init[]   =   {   1   ,   1   ,   5   ,   1   ,   23  ,   49  ,   5   ,   47  ,   203 ,   727 ,   1955    ,   3465    ,   6691    ,0 };
            const unsigned long dim1037Kuo3Init[]   =   {   1   ,   1   ,   3   ,   7   ,   25  ,   13  ,   31  ,   153 ,   411 ,   79  ,   121 ,   3815    ,   2123    ,0 };
            const unsigned long dim1038Kuo3Init[]   =   {   1   ,   1   ,   3   ,   1   ,   3   ,   57  ,   11  ,   141 ,   465 ,   713 ,   659 ,   2849    ,   487 ,0 };
            const unsigned long dim1039Kuo3Init[]   =   {   1   ,   3   ,   7   ,   5   ,   29  ,   49  ,   91  ,   249 ,   353 ,   247 ,   1793    ,   1617    ,   3321    ,0 };
            const unsigned long dim1040Kuo3Init[]   =   {   1   ,   1   ,   7   ,   13  ,   31  ,   3   ,   59  ,   213 ,   265 ,   3   ,   543 ,   2237    ,   6617    ,0 };
            const unsigned long dim1041Kuo3Init[]   =   {   1   ,   1   ,   7   ,   13  ,   9   ,   23  ,   57  ,   89  ,   365 ,   11  ,   871 ,   1343    ,   7375    ,0 };
            const unsigned long dim1042Kuo3Init[]   =   {   1   ,   3   ,   7   ,   11  ,   7   ,   19  ,   15  ,   101 ,   239 ,   341 ,   1259    ,   3607    ,   3879    ,0 };
            const unsigned long dim1043Kuo3Init[]   =   {   1   ,   1   ,   5   ,   7   ,   29  ,   25  ,   117 ,   183 ,   131 ,   705 ,   1497    ,   3537    ,   1007    ,0 };
            const unsigned long dim1044Kuo3Init[]   =   {   1   ,   1   ,   1   ,   3   ,   11  ,   41  ,   113 ,   127 ,   95  ,   417 ,   1177    ,   1939    ,   1317    ,0 };
            const unsigned long dim1045Kuo3Init[]   =   {   1   ,   1   ,   5   ,   11  ,   1   ,   37  ,   1   ,   141 ,   417 ,   157 ,   993 ,   671 ,   4993    ,0 };
            const unsigned long dim1046Kuo3Init[]   =   {   1   ,   1   ,   7   ,   11  ,   23  ,   15  ,   41  ,   171 ,   389 ,   261 ,   1967    ,   3009    ,   2237    ,0 };
            const unsigned long dim1047Kuo3Init[]   =   {   1   ,   3   ,   5   ,   15  ,   31  ,   41  ,   49  ,   39  ,   73  ,   201 ,   495 ,   3121    ,   4247    ,0 };
            const unsigned long dim1048Kuo3Init[]   =   {   1   ,   1   ,   3   ,   1   ,   17  ,   21  ,   95  ,   219 ,   343 ,   103 ,   93  ,   2907    ,   1279    ,0 };
            const unsigned long dim1049Kuo3Init[]   =   {   1   ,   3   ,   3   ,   1   ,   7   ,   31  ,   79  ,   139 ,   347 ,   437 ,   1261    ,   3425    ,   6981    ,0 };
            const unsigned long dim1050Kuo3Init[]   =   {   1   ,   1   ,   5   ,   7   ,   9   ,   19  ,   55  ,   185 ,   425 ,   145 ,   1627    ,   3045    ,   8113    ,0 };
            const unsigned long dim1051Kuo3Init[]   =   {   1   ,   1   ,   5   ,   1   ,   15  ,   33  ,   99  ,   185 ,   485 ,   859 ,   481 ,   3063    ,   3209    ,0 };
            const unsigned long dim1052Kuo3Init[]   =   {   1   ,   1   ,   3   ,   13  ,   23  ,   7   ,   7   ,   81  ,   405 ,   705 ,   1933    ,   315 ,   5819    ,0 };
            const unsigned long dim1053Kuo3Init[]   =   {   1   ,   3   ,   1   ,   3   ,   9   ,   55  ,   95  ,   101 ,   391 ,   299 ,   185 ,   3687    ,   2833    ,0 };
            const unsigned long dim1054Kuo3Init[]   =   {   1   ,   3   ,   7   ,   13  ,   15  ,   5   ,   119 ,   29  ,   29  ,   893 ,   207 ,   1789    ,   3395    ,0 };
            const unsigned long dim1055Kuo3Init[]   =   {   1   ,   3   ,   3   ,   5   ,   1   ,   59  ,   49  ,   81  ,   391 ,   95  ,   289 ,   3167    ,   5417    ,0 };
            const unsigned long dim1056Kuo3Init[]   =   {   1   ,   3   ,   5   ,   15  ,   31  ,   21  ,   5   ,   189 ,   63  ,   567 ,   1515    ,   297 ,   5415    ,0 };
            const unsigned long dim1057Kuo3Init[]   =   {   1   ,   3   ,   3   ,   15  ,   3   ,   23  ,   105 ,   113 ,   403 ,   799 ,   591 ,   1977    ,   1117    ,0 };
            const unsigned long dim1058Kuo3Init[]   =   {   1   ,   3   ,   1   ,   13  ,   23  ,   55  ,   59  ,   229 ,   319 ,   357 ,   1703    ,   585 ,   3785    ,0 };
            const unsigned long dim1059Kuo3Init[]   =   {   1   ,   3   ,   7   ,   13  ,   7   ,   1   ,   97  ,   247 ,   451 ,   897 ,   137 ,   3289    ,   3621    ,0 };
            const unsigned long dim1060Kuo3Init[]   =   {   1   ,   3   ,   1   ,   5   ,   13  ,   43  ,   89  ,   75  ,   243 ,   489 ,   1999    ,   1179    ,   13  ,0 };
            const unsigned long dim1061Kuo3Init[]   =   {   1   ,   3   ,   7   ,   9   ,   11  ,   21  ,   51  ,   199 ,   15  ,   859 ,   83  ,   613 ,   5027    ,0 };
            const unsigned long dim1062Kuo3Init[]   =   {   1   ,   3   ,   1   ,   13  ,   15  ,   15  ,   81  ,   87  ,   311 ,   681 ,   1391    ,   2123    ,   6747    ,0 };
            const unsigned long dim1063Kuo3Init[]   =   {   1   ,   3   ,   1   ,   1   ,   5   ,   45  ,   5   ,   123 ,   19  ,   831 ,   1345    ,   3313    ,   7945    ,0 };
            const unsigned long dim1064Kuo3Init[]   =   {   1   ,   1   ,   7   ,   3   ,   5   ,   9   ,   93  ,   39  ,   221 ,   601 ,   315 ,   2179    ,   6573    ,0 };
            const unsigned long dim1065Kuo3Init[]   =   {   1   ,   3   ,   5   ,   13  ,   17  ,   17  ,   117 ,   145 ,   471 ,   893 ,   305 ,   3065    ,   7519    ,0 };
            const unsigned long dim1066Kuo3Init[]   =   {   1   ,   3   ,   1   ,   9   ,   1   ,   55  ,   119 ,   235 ,   417 ,   167 ,   467 ,   2947    ,   3235    ,0 };
            const unsigned long dim1067Kuo3Init[]   =   {   1   ,   3   ,   3   ,   13  ,   25  ,   51  ,   71  ,   101 ,   255 ,   671 ,   1929    ,   2911    ,   2837    ,0 };
            const unsigned long dim1068Kuo3Init[]   =   {   1   ,   1   ,   5   ,   15  ,   17  ,   1   ,   79  ,   241 ,   25  ,   635 ,   379 ,   1701    ,   7139    ,0 };
            const unsigned long dim1069Kuo3Init[]   =   {   1   ,   1   ,   7   ,   13  ,   1   ,   1   ,   119 ,   121 ,   469 ,   51  ,   939 ,   2161    ,   4247    ,0 };
            const unsigned long dim1070Kuo3Init[]   =   {   1   ,   3   ,   3   ,   15  ,   3   ,   35  ,   95  ,   107 ,   143 ,   701 ,   1843    ,   301 ,   6345    ,0 };
            const unsigned long dim1071Kuo3Init[]   =   {   1   ,   3   ,   5   ,   3   ,   27  ,   13  ,   117 ,   143 ,   49  ,   875 ,   1705    ,   3005    ,   3671    ,0 };
            const unsigned long dim1072Kuo3Init[]   =   {   1   ,   1   ,   7   ,   7   ,   5   ,   49  ,   111 ,   33  ,   169 ,   189 ,   1109    ,   805 ,   2753    ,0 };
            const unsigned long dim1073Kuo3Init[]   =   {   1   ,   3   ,   7   ,   9   ,   19  ,   49  ,   59  ,   3   ,   71  ,   859 ,   435 ,   3417    ,   1883    ,0 };
            const unsigned long dim1074Kuo3Init[]   =   {   1   ,   3   ,   5   ,   9   ,   9   ,   51  ,   35  ,   79  ,   385 ,   83  ,   1121    ,   1421    ,   7317    ,0 };
            const unsigned long dim1075Kuo3Init[]   =   {   1   ,   3   ,   7   ,   5   ,   13  ,   29  ,   49  ,   41  ,   219 ,   23  ,   1025    ,   849 ,   499 ,0 };
            const unsigned long dim1076Kuo3Init[]   =   {   1   ,   3   ,   3   ,   11  ,   3   ,   15  ,   13  ,   167 ,   441 ,   753 ,   1731    ,   527 ,   8099    ,0 };
            const unsigned long dim1077Kuo3Init[]   =   {   1   ,   3   ,   5   ,   3   ,   1   ,   57  ,   17  ,   193 ,   29  ,   849 ,   15  ,   539 ,   1713    ,0 };
            const unsigned long dim1078Kuo3Init[]   =   {   1   ,   1   ,   1   ,   11  ,   27  ,   47  ,   59  ,   249 ,   363 ,   175 ,   3   ,   2485    ,   3551    ,0 };
            const unsigned long dim1079Kuo3Init[]   =   {   1   ,   1   ,   3   ,   3   ,   25  ,   45  ,   79  ,   45  ,   141 ,   13  ,   1007    ,   1235    ,   5359    ,0 };
            const unsigned long dim1080Kuo3Init[]   =   {   1   ,   3   ,   1   ,   7   ,   29  ,   31  ,   111 ,   173 ,   281 ,   287 ,   173 ,   2627    ,   4125    ,0 };
            const unsigned long dim1081Kuo3Init[]   =   {   1   ,   3   ,   5   ,   5   ,   3   ,   13  ,   37  ,   157 ,   509 ,   323 ,   1695    ,   765 ,   6533    ,0 };
            const unsigned long dim1082Kuo3Init[]   =   {   1   ,   3   ,   5   ,   11  ,   7   ,   33  ,   77  ,   237 ,   377 ,   351 ,   205 ,   183 ,   5273    ,0 };
            const unsigned long dim1083Kuo3Init[]   =   {   1   ,   3   ,   7   ,   5   ,   11  ,   1   ,   123 ,   251 ,   3   ,   207 ,   1319    ,   553 ,   8143    ,0 };
            const unsigned long dim1084Kuo3Init[]   =   {   1   ,   1   ,   7   ,   15  ,   23  ,   25  ,   105 ,   43  ,   271 ,   633 ,   1331    ,   2197    ,   7229    ,0 };
            const unsigned long dim1085Kuo3Init[]   =   {   1   ,   3   ,   3   ,   1   ,   13  ,   37  ,   119 ,   195 ,   21  ,   751 ,   1809    ,   4025    ,   3037    ,0 };
            const unsigned long dim1086Kuo3Init[]   =   {   1   ,   1   ,   7   ,   13  ,   19  ,   1   ,   3   ,   33  ,   79  ,   477 ,   1203    ,   3657    ,   5573    ,0 };
            const unsigned long dim1087Kuo3Init[]   =   {   1   ,   1   ,   1   ,   1   ,   21  ,   21  ,   121 ,   183 ,   355 ,   259 ,   1033    ,   1621    ,   2843    ,0 };
            const unsigned long dim1088Kuo3Init[]   =   {   1   ,   3   ,   7   ,   15  ,   19  ,   37  ,   121 ,   233 ,   397 ,   27  ,   279 ,   2891    ,   4191    ,0 };
            const unsigned long dim1089Kuo3Init[]   =   {   1   ,   3   ,   1   ,   3   ,   13  ,   5   ,   31  ,   237 ,   453 ,   913 ,   1073    ,   3573    ,   7883    ,0 };
            const unsigned long dim1090Kuo3Init[]   =   {   1   ,   1   ,   1   ,   1   ,   1   ,   11  ,   57  ,   103 ,   453 ,   605 ,   1041    ,   2975    ,   4735    ,0 };
            const unsigned long dim1091Kuo3Init[]   =   {   1   ,   1   ,   5   ,   9   ,   29  ,   33  ,   119 ,   241 ,   27  ,   491 ,   341 ,   1945    ,   6599    ,0 };
            const unsigned long dim1092Kuo3Init[]   =   {   1   ,   1   ,   7   ,   1   ,   25  ,   41  ,   83  ,   191 ,   481 ,   995 ,   1819    ,   1577    ,   3837    ,0 };
            const unsigned long dim1093Kuo3Init[]   =   {   1   ,   3   ,   5   ,   5   ,   15  ,   21  ,   81  ,   111 ,   143 ,   281 ,   509 ,   2905    ,   177 ,0 };
            const unsigned long dim1094Kuo3Init[]   =   {   1   ,   1   ,   1   ,   3   ,   19  ,   19  ,   31  ,   107 ,   451 ,   1007    ,   2027    ,   3947    ,   3903    ,0 };
            const unsigned long dim1095Kuo3Init[]   =   {   1   ,   3   ,   3   ,   1   ,   27  ,   41  ,   23  ,   147 ,   299 ,   315 ,   895 ,   407 ,   4761    ,0 };
            const unsigned long dim1096Kuo3Init[]   =   {   1   ,   3   ,   7   ,   15  ,   27  ,   1   ,   105 ,   103 ,   507 ,   897 ,   1309    ,   3979    ,   5123    ,0 };
            const unsigned long dim1097Kuo3Init[]   =   {   1   ,   3   ,   1   ,   11  ,   7   ,   41  ,   99  ,   67  ,   183 ,   495 ,   1509    ,   2733    ,   6995    ,0 };
            const unsigned long dim1098Kuo3Init[]   =   {   1   ,   1   ,   1   ,   11  ,   13  ,   5   ,   69  ,   239 ,   95  ,   211 ,   691 ,   323 ,   7485    ,0 };
            const unsigned long dim1099Kuo3Init[]   =   {   1   ,   3   ,   5   ,   11  ,   23  ,   57  ,   5   ,   113 ,   417 ,   893 ,   1229    ,   2109    ,   7657    ,0 };
            const unsigned long dim1100Kuo3Init[]   =   {   1   ,   1   ,   5   ,   15  ,   19  ,   59  ,   111 ,   181 ,   61  ,   743 ,   227 ,   831 ,   3035    ,0 };
            const unsigned long dim1101Kuo3Init[]   =   {   1   ,   1   ,   5   ,   7   ,   19  ,   13  ,   107 ,   63  ,   127 ,   17  ,   1753    ,   337 ,   3403    ,0 };
            const unsigned long dim1102Kuo3Init[]   =   {   1   ,   3   ,   7   ,   7   ,   7   ,   21  ,   85  ,   219 ,   77  ,   311 ,   685 ,   145 ,   3451    ,0 };
            const unsigned long dim1103Kuo3Init[]   =   {   1   ,   3   ,   5   ,   5   ,   3   ,   55  ,   123 ,   187 ,   407 ,   539 ,   1167    ,   2741    ,   6885    ,0 };
            const unsigned long dim1104Kuo3Init[]   =   {   1   ,   3   ,   1   ,   13  ,   23  ,   57  ,   109 ,   93  ,   93  ,   483 ,   857 ,   1895    ,   7145    ,0 };
            const unsigned long dim1105Kuo3Init[]   =   {   1   ,   1   ,   1   ,   15  ,   19  ,   43  ,   37  ,   129 ,   487 ,   523 ,   1959    ,   2053    ,   6281    ,0 };
            const unsigned long dim1106Kuo3Init[]   =   {   1   ,   3   ,   1   ,   11  ,   9   ,   1   ,   105 ,   123 ,   483 ,   511 ,   1661    ,   1065    ,   7539    ,0 };
            const unsigned long dim1107Kuo3Init[]   =   {   1   ,   3   ,   3   ,   1   ,   27  ,   33  ,   87  ,   191 ,   89  ,   697 ,   1195    ,   1699    ,   2699    ,0 };
            const unsigned long dim1108Kuo3Init[]   =   {   1   ,   3   ,   7   ,   5   ,   21  ,   57  ,   5   ,   109 ,   343 ,   721 ,   1355    ,   1155    ,   4515    ,0 };
            const unsigned long dim1109Kuo3Init[]   =   {   1   ,   3   ,   3   ,   15  ,   27  ,   43  ,   47  ,   167 ,   457 ,   93  ,   221 ,   3693    ,   383 ,0 };
            const unsigned long dim1110Kuo3Init[]   =   {   1   ,   3   ,   1   ,   7   ,   13  ,   13  ,   55  ,   75  ,   113 ,   577 ,   397 ,   2491    ,   6871    ,0 };
            const unsigned long dim1111Kuo3Init[]   =   {   1   ,   1   ,   3   ,   13  ,   31  ,   31  ,   71  ,   189 ,   39  ,   397 ,   977 ,   2645    ,   1683    ,   5081    ,0 };
            const unsigned long dim1112Kuo3Init[]   =   {   1   ,   3   ,   5   ,   15  ,   27  ,   41  ,   125 ,   69  ,   269 ,   157 ,   1827    ,   41  ,   3105    ,   10615   ,0 };
            const unsigned long dim1113Kuo3Init[]   =   {   1   ,   1   ,   3   ,   9   ,   5   ,   45  ,   33  ,   169 ,   317 ,   371 ,   389 ,   595 ,   1071    ,   5323    ,0 };
            const unsigned long dim1114Kuo3Init[]   =   {   1   ,   1   ,   5   ,   5   ,   25  ,   25  ,   69  ,   45  ,   313 ,   105 ,   795 ,   3987    ,   5181    ,   155 ,0 };
            const unsigned long dim1115Kuo3Init[]   =   {   1   ,   3   ,   7   ,   5   ,   19  ,   17  ,   13  ,   119 ,   213 ,   289 ,   269 ,   1561    ,   6069    ,   15631   ,0 };
            const unsigned long dim1116Kuo3Init[]   =   {   1   ,   3   ,   7   ,   1   ,   15  ,   7   ,   17  ,   173 ,   491 ,   907 ,   315 ,   101 ,   6205    ,   4631    ,0 };
            const unsigned long dim1117Kuo3Init[]   =   {   1   ,   1   ,   5   ,   11  ,   17  ,   13  ,   9   ,   67  ,   267 ,   719 ,   1565    ,   2157    ,   5773    ,   5167    ,0 };
            const unsigned long dim1118Kuo3Init[]   =   {   1   ,   1   ,   5   ,   5   ,   11  ,   45  ,   65  ,   207 ,   403 ,   231 ,   1019    ,   3345    ,   2779    ,   3621    ,0 };
            const unsigned long dim1119Kuo3Init[]   =   {   1   ,   3   ,   5   ,   13  ,   13  ,   59  ,   65  ,   225 ,   307 ,   581 ,   1159    ,   2041    ,   3535    ,   16381   ,0 };
            const unsigned long dim1120Kuo3Init[]   =   {   1   ,   3   ,   1   ,   15  ,   9   ,   45  ,   17  ,   99  ,   285 ,   797 ,   477 ,   3041    ,   3123    ,   397 ,0 };
            const unsigned long dim1121Kuo3Init[]   =   {   1   ,   3   ,   1   ,   15  ,   3   ,   57  ,   99  ,   115 ,   105 ,   877 ,   9   ,   1185    ,   3147    ,   6285    ,0 };
            const unsigned long dim1122Kuo3Init[]   =   {   1   ,   1   ,   1   ,   1   ,   25  ,   31  ,   37  ,   13  ,   119 ,   935 ,   1339    ,   2147    ,   2371    ,   16107   ,0 };
            const unsigned long dim1123Kuo3Init[]   =   {   1   ,   1   ,   5   ,   15  ,   17  ,   31  ,   123 ,   131 ,   135 ,   977 ,   1943    ,   3643    ,   4457    ,   15405   ,0 };
            const unsigned long dim1124Kuo3Init[]   =   {   1   ,   3   ,   3   ,   1   ,   3   ,   57  ,   29  ,   33  ,   1   ,   519 ,   535 ,   681 ,   6255    ,   16263   ,0 };
            const unsigned long dim1125Kuo3Init[]   =   {   1   ,   1   ,   5   ,   3   ,   9   ,   45  ,   39  ,   215 ,   329 ,   457 ,   1341    ,   3037    ,   1631    ,   15335   ,0 };
            const unsigned long dim1126Kuo3Init[]   =   {   1   ,   1   ,   7   ,   13  ,   3   ,   27  ,   23  ,   73  ,   465 ,   771 ,   1453    ,   3821    ,   3099    ,   10075   ,0 };
            const unsigned long dim1127Kuo3Init[]   =   {   1   ,   1   ,   3   ,   3   ,   21  ,   45  ,   119 ,   233 ,   507 ,   959 ,   1743    ,   2711    ,   2465    ,   4257    ,0 };
            const unsigned long dim1128Kuo3Init[]   =   {   1   ,   1   ,   7   ,   5   ,   31  ,   37  ,   113 ,   207 ,   239 ,   305 ,   1085    ,   2447    ,   743 ,   15747   ,0 };
            const unsigned long dim1129Kuo3Init[]   =   {   1   ,   3   ,   5   ,   5   ,   27  ,   25  ,   83  ,   255 ,   93  ,   149 ,   627 ,   1887    ,   6913    ,   5793    ,0 };
            const unsigned long dim1130Kuo3Init[]   =   {   1   ,   1   ,   5   ,   11  ,   17  ,   61  ,   91  ,   75  ,   469 ,   503 ,   1405    ,   3609    ,   421 ,   10819   ,0 };
            const unsigned long dim1131Kuo3Init[]   =   {   1   ,   1   ,   5   ,   1   ,   13  ,   47  ,   53  ,   221 ,   351 ,   621 ,   1025    ,   3729    ,   367 ,   16091   ,0 };
            const unsigned long dim1132Kuo3Init[]   =   {   1   ,   1   ,   1   ,   5   ,   7   ,   41  ,   3   ,   177 ,   127 ,   71  ,   753 ,   1581    ,   2227    ,   257 ,0 };
            const unsigned long dim1133Kuo3Init[]   =   {   1   ,   1   ,   3   ,   3   ,   3   ,   17  ,   69  ,   31  ,   143 ,   825 ,   291 ,   257 ,   2147    ,   4199    ,0 };
            const unsigned long dim1134Kuo3Init[]   =   {   1   ,   1   ,   5   ,   9   ,   29  ,   21  ,   49  ,   1   ,   35  ,   321 ,   383 ,   1495    ,   4201    ,   7143    ,0 };
            const unsigned long dim1135Kuo3Init[]   =   {   1   ,   3   ,   7   ,   13  ,   5   ,   27  ,   31  ,   143 ,   117 ,   487 ,   585 ,   809 ,   2425    ,   15681   ,0 };
            const unsigned long dim1136Kuo3Init[]   =   {   1   ,   3   ,   3   ,   13  ,   23  ,   3   ,   33  ,   213 ,   209 ,   185 ,   603 ,   431 ,   5519    ,   5017    ,0 };
            const unsigned long dim1137Kuo3Init[]   =   {   1   ,   3   ,   7   ,   9   ,   21  ,   61  ,   3   ,   53  ,   303 ,   793 ,   759 ,   3869    ,   391 ,   8395    ,0 };
            const unsigned long dim1138Kuo3Init[]   =   {   1   ,   3   ,   5   ,   13  ,   25  ,   13  ,   119 ,   37  ,   419 ,   399 ,   825 ,   709 ,   5351    ,   1299    ,0 };
            const unsigned long dim1139Kuo3Init[]   =   {   1   ,   3   ,   7   ,   11  ,   17  ,   11  ,   93  ,   205 ,   33  ,   493 ,   1353    ,   3777    ,   2521    ,   13405   ,0 };
            const unsigned long dim1140Kuo3Init[]   =   {   1   ,   3   ,   3   ,   5   ,   5   ,   3   ,   25  ,   215 ,   209 ,   559 ,   465 ,   1809    ,   6621    ,   8733    ,0 };
            const unsigned long dim1141Kuo3Init[]   =   {   1   ,   3   ,   5   ,   11  ,   23  ,   17  ,   75  ,   11  ,   299 ,   613 ,   565 ,   3487    ,   5017    ,   6187    ,0 };
            const unsigned long dim1142Kuo3Init[]   =   {   1   ,   3   ,   5   ,   15  ,   15  ,   41  ,   121 ,   225 ,   259 ,   283 ,   1119    ,   3509    ,   6663    ,   241 ,0 };
            const unsigned long dim1143Kuo3Init[]   =   {   1   ,   3   ,   3   ,   5   ,   17  ,   49  ,   5   ,   227 ,   493 ,   677 ,   1365    ,   3681    ,   1651    ,   5079    ,0 };
            const unsigned long dim1144Kuo3Init[]   =   {   1   ,   3   ,   5   ,   11  ,   3   ,   13  ,   43  ,   219 ,   247 ,   223 ,   1451    ,   1449    ,   605 ,   5877    ,0 };
            const unsigned long dim1145Kuo3Init[]   =   {   1   ,   1   ,   7   ,   7   ,   21  ,   47  ,   7   ,   197 ,   387 ,   59  ,   7   ,   3209    ,   5623    ,   1221    ,0 };
            const unsigned long dim1146Kuo3Init[]   =   {   1   ,   3   ,   3   ,   5   ,   25  ,   5   ,   97  ,   1   ,   105 ,   801 ,   815 ,   1635    ,   3567    ,   5231    ,0 };
            const unsigned long dim1147Kuo3Init[]   =   {   1   ,   3   ,   7   ,   15  ,   17  ,   21  ,   101 ,   167 ,   265 ,   407 ,   1859    ,   3359    ,   1245    ,   4153    ,0 };
            const unsigned long dim1148Kuo3Init[]   =   {   1   ,   1   ,   1   ,   11  ,   31  ,   61  ,   21  ,   101 ,   77  ,   325 ,   645 ,   2059    ,   3201    ,   527 ,0 };
            const unsigned long dim1149Kuo3Init[]   =   {   1   ,   3   ,   7   ,   15  ,   29  ,   55  ,   107 ,   135 ,   55  ,   433 ,   1519    ,   1049    ,   6119    ,   15877   ,0 };
            const unsigned long dim1150Kuo3Init[]   =   {   1   ,   1   ,   7   ,   7   ,   19  ,   29  ,   1   ,   167 ,   13  ,   67  ,   1703    ,   1159    ,   7145    ,   4243    ,0 };
            const unsigned long dim1151Kuo3Init[]   =   {   1   ,   3   ,   5   ,   11  ,   1   ,   33  ,   7   ,   21  ,   407 ,   695 ,   345 ,   425 ,   7399    ,   3847    ,0 };
            const unsigned long dim1152Kuo3Init[]   =   {   1   ,   1   ,   7   ,   11  ,   15  ,   59  ,   121 ,   175 ,   21  ,   859 ,   1311    ,   1365    ,   3823    ,   3983    ,0 };
            const unsigned long dim1153Kuo3Init[]   =   {   1   ,   1   ,   7   ,   5   ,   29  ,   21  ,   111 ,   221 ,   429 ,   739 ,   985 ,   1989    ,   4947    ,   8311    ,0 };
            const unsigned long dim1154Kuo3Init[]   =   {   1   ,   3   ,   3   ,   15  ,   29  ,   17  ,   103 ,   39  ,   475 ,   193 ,   253 ,   1903    ,   4073    ,   15039   ,0 };
            const unsigned long dim1155Kuo3Init[]   =   {   1   ,   3   ,   7   ,   11  ,   1   ,   57  ,   35  ,   197 ,   89  ,   475 ,   283 ,   2291    ,   7129    ,   14357   ,0 };
            const unsigned long dim1156Kuo3Init[]   =   {   1   ,   1   ,   3   ,   11  ,   15  ,   63  ,   29  ,   157 ,   211 ,   517 ,   1643    ,   199 ,   285 ,   8621    ,0 };
            const unsigned long dim1157Kuo3Init[]   =   {   1   ,   3   ,   3   ,   15  ,   9   ,   47  ,   43  ,   215 ,   407 ,   979 ,   51  ,   635 ,   467 ,   6365    ,0 };
            const unsigned long dim1158Kuo3Init[]   =   {   1   ,   3   ,   3   ,   1   ,   23  ,   27  ,   97  ,   111 ,   51  ,   407 ,   1037    ,   1337    ,   1475    ,   1151    ,0 };
            const unsigned long dim1159Kuo3Init[]   =   {   1   ,   3   ,   3   ,   15  ,   13  ,   45  ,   67  ,   51  ,   331 ,   691 ,   1137    ,   1967    ,   5715    ,   3125    ,0 };
            const unsigned long dim1160Kuo3Init[]   =   {   1   ,   3   ,   3   ,   15  ,   31  ,   19  ,   85  ,   221 ,   449 ,   689 ,   1817    ,   1575    ,   7429    ,   12125   ,0 };
            const unsigned long dim1161Kuo3Init[]   =   {   1   ,   3   ,   1   ,   9   ,   21  ,   61  ,   77  ,   55  ,   65  ,   591 ,   9   ,   2477    ,   4081    ,   4631    ,0 };
            const unsigned long dim1162Kuo3Init[]   =   {   1   ,   3   ,   5   ,   9   ,   27  ,   45  ,   29  ,   159 ,   21  ,   931 ,   1939    ,   1297    ,   5589    ,   12635   ,0 };
            const unsigned long dim1163Kuo3Init[]   =   {   1   ,   1   ,   7   ,   9   ,   21  ,   41  ,   7   ,   171 ,   461 ,   373 ,   1655    ,   3475    ,   4073    ,   4859    ,0 };
            const unsigned long dim1164Kuo3Init[]   =   {   1   ,   3   ,   1   ,   13  ,   17  ,   35  ,   3   ,   41  ,   45  ,   37  ,   1409    ,   3939    ,   5497    ,   3501    ,0 };
            const unsigned long dim1165Kuo3Init[]   =   {   1   ,   1   ,   3   ,   15  ,   11  ,   57  ,   121 ,   177 ,   177 ,   515 ,   675 ,   1943    ,   2645    ,   2469    ,0 };
            const unsigned long dim1166Kuo3Init[]   =   {   1   ,   1   ,   7   ,   1   ,   23  ,   63  ,   29  ,   157 ,   157 ,   397 ,   1131    ,   1479    ,   4683    ,   10011   ,0 };
            const unsigned long dim1167Kuo3Init[]   =   {   1   ,   3   ,   7   ,   9   ,   15  ,   35  ,   75  ,   99  ,   195 ,   589 ,   19  ,   3427    ,   1623    ,   12843   ,0 };
            const unsigned long dim1168Kuo3Init[]   =   {   1   ,   3   ,   7   ,   3   ,   21  ,   19  ,   101 ,   135 ,   351 ,   549 ,   223 ,   517 ,   4023    ,   1863    ,0 };
            const unsigned long dim1169Kuo3Init[]   =   {   1   ,   1   ,   7   ,   7   ,   7   ,   23  ,   67  ,   15  ,   217 ,   915 ,   1617    ,   1527    ,   877 ,   15061   ,0 };
            const unsigned long dim1170Kuo3Init[]   =   {   1   ,   3   ,   5   ,   3   ,   11  ,   35  ,   95  ,   5   ,   399 ,   669 ,   1185    ,   167 ,   7405    ,   11169   ,0 };
            const unsigned long dim1171Kuo3Init[]   =   {   1   ,   3   ,   5   ,   5   ,   1   ,   49  ,   91  ,   125 ,   185 ,   73  ,   263 ,   1139    ,   3171    ,   5427    ,0 };
            const unsigned long dim1172Kuo3Init[]   =   {   1   ,   1   ,   5   ,   1   ,   27  ,   51  ,   89  ,   81  ,   179 ,   593 ,   67  ,   1495    ,   3669    ,   15981   ,0 };
            const unsigned long dim1173Kuo3Init[]   =   {   1   ,   3   ,   1   ,   11  ,   29  ,   63  ,   47  ,   105 ,   327 ,   747 ,   953 ,   2399    ,   6711    ,   16313   ,0 };
            const unsigned long dim1174Kuo3Init[]   =   {   1   ,   1   ,   5   ,   5   ,   21  ,   21  ,   117 ,   179 ,   427 ,   981 ,   2029    ,   1   ,   495 ,   4005    ,0 };
            const unsigned long dim1175Kuo3Init[]   =   {   1   ,   1   ,   3   ,   5   ,   23  ,   21  ,   59  ,   99  ,   459 ,   217 ,   1131    ,   1853    ,   3049    ,   2247    ,0 };
            const unsigned long dim1176Kuo3Init[]   =   {   1   ,   3   ,   1   ,   5   ,   29  ,   9   ,   71  ,   167 ,   55  ,   775 ,   627 ,   3967    ,   45  ,   11199   ,0 };
            const unsigned long dim1177Kuo3Init[]   =   {   1   ,   3   ,   5   ,   3   ,   29  ,   25  ,   1   ,   69  ,   317 ,   649 ,   1567    ,   2587    ,   2277    ,   15675   ,0 };
            const unsigned long dim1178Kuo3Init[]   =   {   1   ,   3   ,   3   ,   9   ,   21  ,   3   ,   39  ,   147 ,   279 ,   953 ,   1207    ,   333 ,   2747    ,   1717    ,0 };
            const unsigned long dim1179Kuo3Init[]   =   {   1   ,   3   ,   3   ,   3   ,   31  ,   63  ,   113 ,   149 ,   431 ,   699 ,   335 ,   2475    ,   6375    ,   13061   ,0 };
            const unsigned long dim1180Kuo3Init[]   =   {   1   ,   1   ,   7   ,   11  ,   11  ,   19  ,   51  ,   175 ,   377 ,   971 ,   35  ,   2749    ,   6309    ,   4755    ,0 };
            const unsigned long dim1181Kuo3Init[]   =   {   1   ,   1   ,   3   ,   15  ,   27  ,   53  ,   105 ,   221 ,   15  ,   495 ,   799 ,   1717    ,   6997    ,   15949   ,0 };
            const unsigned long dim1182Kuo3Init[]   =   {   1   ,   1   ,   1   ,   15  ,   31  ,   29  ,   69  ,   137 ,   421 ,   231 ,   2021    ,   1869    ,   4533    ,   1121    ,0 };
            const unsigned long dim1183Kuo3Init[]   =   {   1   ,   1   ,   7   ,   11  ,   15  ,   49  ,   17  ,   121 ,   285 ,   303 ,   1085    ,   2463    ,   357 ,   2519    ,0 };
            const unsigned long dim1184Kuo3Init[]   =   {   1   ,   1   ,   1   ,   5   ,   29  ,   47  ,   109 ,   95  ,   339 ,   15  ,   1195    ,   2171    ,   4649    ,   6517    ,0 };
            const unsigned long dim1185Kuo3Init[]   =   {   1   ,   3   ,   3   ,   3   ,   29  ,   19  ,   45  ,   93  ,   47  ,   827 ,   1113    ,   4055    ,   601 ,   10807   ,0 };
            const unsigned long dim1186Kuo3Init[]   =   {   1   ,   1   ,   7   ,   5   ,   15  ,   47  ,   11  ,   79  ,   137 ,   267 ,   1443    ,   1003    ,   2475    ,   147 ,0 };
            const unsigned long dim1187Kuo3Init[]   =   {   1   ,   3   ,   7   ,   15  ,   11  ,   39  ,   95  ,   191 ,   205 ,   967 ,   333 ,   3371    ,   4913    ,   15231   ,0 };
            const unsigned long dim1188Kuo3Init[]   =   {   1   ,   3   ,   7   ,   9   ,   25  ,   9   ,   47  ,   75  ,   77  ,   539 ,   487 ,   361 ,   2293    ,   7285    ,0 };
            const unsigned long dim1189Kuo3Init[]   =   {   1   ,   3   ,   7   ,   7   ,   27  ,   11  ,   93  ,   181 ,   455 ,   277 ,   1001    ,   3313    ,   7335    ,   4593    ,0 };
            const unsigned long dim1190Kuo3Init[]   =   {   1   ,   1   ,   5   ,   5   ,   15  ,   35  ,   107 ,   51  ,   223 ,   447 ,   371 ,   1739    ,   1937    ,   3589    ,0 };
            const unsigned long dim1191Kuo3Init[]   =   {   1   ,   1   ,   3   ,   11  ,   15  ,   51  ,   23  ,   129 ,   285 ,   895 ,   2015    ,   1389    ,   2043    ,   1989    ,0 };
            const unsigned long dim1192Kuo3Init[]   =   {   1   ,   1   ,   5   ,   7   ,   21  ,   57  ,   107 ,   149 ,   285 ,   825 ,   959 ,   827 ,   2931    ,   3207    ,0 };
            const unsigned long dim1193Kuo3Init[]   =   {   1   ,   1   ,   1   ,   13  ,   23  ,   19  ,   91  ,   15  ,   261 ,   785 ,   1031    ,   2245    ,   505 ,   14891   ,0 };
            const unsigned long dim1194Kuo3Init[]   =   {   1   ,   1   ,   3   ,   15  ,   23  ,   21  ,   33  ,   153 ,   71  ,   757 ,   1131    ,   3251    ,   877 ,   4533    ,0 };
            const unsigned long dim1195Kuo3Init[]   =   {   1   ,   1   ,   3   ,   11  ,   25  ,   39  ,   33  ,   231 ,   451 ,   587 ,   781 ,   1145    ,   4743    ,   467 ,0 };
            const unsigned long dim1196Kuo3Init[]   =   {   1   ,   1   ,   7   ,   11  ,   23  ,   7   ,   35  ,   45  ,   397 ,   41  ,   1443    ,   2087    ,   587 ,   14139   ,0 };
            const unsigned long dim1197Kuo3Init[]   =   {   1   ,   3   ,   3   ,   3   ,   19  ,   57  ,   61  ,   225 ,   197 ,   263 ,   1447    ,   1929    ,   2833    ,   5325    ,0 };
            const unsigned long dim1198Kuo3Init[]   =   {   1   ,   1   ,   5   ,   7   ,   25  ,   51  ,   45  ,   221 ,   403 ,   843 ,   727 ,   3099    ,   3883    ,   9151    ,0 };
            const unsigned long dim1199Kuo3Init[]   =   {   1   ,   1   ,   5   ,   13  ,   9   ,   23  ,   33  ,   119 ,   81  ,   525 ,   1527    ,   567 ,   2867    ,   12907   ,0 };
            const unsigned long dim1200Kuo3Init[]   =   {   1   ,   1   ,   5   ,   5   ,   11  ,   19  ,   125 ,   135 ,   427 ,   949 ,   643 ,   3403    ,   6531    ,   12607   ,0 };
            const unsigned long dim1201Kuo3Init[]   =   {   1   ,   3   ,   3   ,   7   ,   27  ,   23  ,   89  ,   83  ,   381 ,   333 ,   361 ,   2773    ,   4247    ,   10233   ,0 };
            const unsigned long dim1202Kuo3Init[]   =   {   1   ,   3   ,   7   ,   11  ,   25  ,   61  ,   105 ,   141 ,   473 ,   993 ,   379 ,   2949    ,   5757    ,   4003    ,0 };
            const unsigned long dim1203Kuo3Init[]   =   {   1   ,   1   ,   3   ,   11  ,   27  ,   29  ,   75  ,   175 ,   391 ,   567 ,   1017    ,   1575    ,   5287    ,   16289   ,0 };
            const unsigned long dim1204Kuo3Init[]   =   {   1   ,   3   ,   7   ,   15  ,   7   ,   33  ,   23  ,   137 ,   299 ,   19  ,   599 ,   2649    ,   6141    ,   3321    ,0 };
            const unsigned long dim1205Kuo3Init[]   =   {   1   ,   1   ,   3   ,   1   ,   3   ,   19  ,   101 ,   95  ,   499 ,   261 ,   51  ,   43  ,   5457    ,   313 ,0 };
            const unsigned long dim1206Kuo3Init[]   =   {   1   ,   3   ,   5   ,   11  ,   27  ,   5   ,   7   ,   85  ,   419 ,   983 ,   1965    ,   1639    ,   6397    ,   15823   ,0 };
            const unsigned long dim1207Kuo3Init[]   =   {   1   ,   1   ,   1   ,   11  ,   13  ,   55  ,   121 ,   13  ,   289 ,   791 ,   437 ,   3279    ,   7125    ,   10473   ,0 };
            const unsigned long dim1208Kuo3Init[]   =   {   1   ,   1   ,   7   ,   1   ,   13  ,   15  ,   25  ,   9   ,   21  ,   205 ,   61  ,   749 ,   8005    ,   9453    ,0 };
            const unsigned long dim1209Kuo3Init[]   =   {   1   ,   1   ,   3   ,   9   ,   19  ,   47  ,   55  ,   113 ,   325 ,   165 ,   771 ,   3655    ,   6265    ,   11157   ,0 };
            const unsigned long dim1210Kuo3Init[]   =   {   1   ,   1   ,   5   ,   13  ,   19  ,   45  ,   107 ,   37  ,   205 ,   677 ,   1273    ,   3817    ,   559 ,   11321   ,0 };
            const unsigned long dim1211Kuo3Init[]   =   {   1   ,   3   ,   7   ,   15  ,   31  ,   57  ,   87  ,   31  ,   431 ,   383 ,   327 ,   2703    ,   4197    ,   7861    ,0 };
            const unsigned long dim1212Kuo3Init[]   =   {   1   ,   1   ,   3   ,   9   ,   3   ,   33  ,   39  ,   253 ,   215 ,   903 ,   447 ,   2175    ,   5249    ,   3203    ,0 };
            const unsigned long dim1213Kuo3Init[]   =   {   1   ,   3   ,   7   ,   5   ,   15  ,   5   ,   27  ,   145 ,   391 ,   869 ,   1641    ,   2497    ,   4683    ,   3111    ,0 };
            const unsigned long dim1214Kuo3Init[]   =   {   1   ,   3   ,   1   ,   5   ,   3   ,   33  ,   39  ,   223 ,   379 ,   803 ,   1833    ,   1799    ,   2617    ,   2113    ,0 };
            const unsigned long dim1215Kuo3Init[]   =   {   1   ,   3   ,   1   ,   15  ,   29  ,   37  ,   57  ,   227 ,   425 ,   943 ,   311 ,   2883    ,   2973    ,   1831    ,0 };
            const unsigned long dim1216Kuo3Init[]   =   {   1   ,   3   ,   3   ,   5   ,   17  ,   63  ,   113 ,   139 ,   453 ,   871 ,   1029    ,   2705    ,   3379    ,   4503    ,0 };
            const unsigned long dim1217Kuo3Init[]   =   {   1   ,   3   ,   7   ,   5   ,   19  ,   35  ,   43  ,   151 ,   511 ,   657 ,   1885    ,   1425    ,   387 ,   2165    ,0 };
            const unsigned long dim1218Kuo3Init[]   =   {   1   ,   1   ,   3   ,   15  ,   9   ,   61  ,   123 ,   241 ,   485 ,   485 ,   955 ,   1215    ,   87  ,   3195    ,0 };
            const unsigned long dim1219Kuo3Init[]   =   {   1   ,   3   ,   5   ,   9   ,   23  ,   45  ,   51  ,   15  ,   23  ,   49  ,   2025    ,   853 ,   6609    ,   12099   ,0 };
            const unsigned long dim1220Kuo3Init[]   =   {   1   ,   3   ,   7   ,   7   ,   23  ,   35  ,   95  ,   67  ,   131 ,   281 ,   1821    ,   1951    ,   1363    ,   7477    ,0 };
            const unsigned long dim1221Kuo3Init[]   =   {   1   ,   3   ,   1   ,   7   ,   9   ,   53  ,   79  ,   255 ,   453 ,   77  ,   629 ,   1207    ,   4171    ,   4415    ,0 };
            const unsigned long dim1222Kuo3Init[]   =   {   1   ,   3   ,   5   ,   3   ,   23  ,   5   ,   11  ,   153 ,   57  ,   217 ,   1177    ,   2297    ,   3515    ,   15121   ,0 };
            const unsigned long dim1223Kuo3Init[]   =   {   1   ,   1   ,   1   ,   9   ,   31  ,   57  ,   93  ,   35  ,   177 ,   449 ,   1505    ,   2715    ,   1315    ,   8953    ,0 };
            const unsigned long dim1224Kuo3Init[]   =   {   1   ,   1   ,   3   ,   11  ,   7   ,   7   ,   109 ,   255 ,   317 ,   325 ,   267 ,   2573    ,   1511    ,   3371    ,0 };
            const unsigned long dim1225Kuo3Init[]   =   {   1   ,   3   ,   7   ,   11  ,   13  ,   51  ,   29  ,   203 ,   13  ,   889 ,   679 ,   1981    ,   5553    ,   671 ,0 };
            const unsigned long dim1226Kuo3Init[]   =   {   1   ,   1   ,   3   ,   5   ,   29  ,   11  ,   29  ,   249 ,   289 ,   817 ,   1983    ,   1865    ,   6705    ,   9667    ,0 };
            const unsigned long dim1227Kuo3Init[]   =   {   1   ,   1   ,   3   ,   13  ,   15  ,   19  ,   23  ,   59  ,   143 ,   815 ,   643 ,   763 ,   1437    ,   2909    ,0 };
            const unsigned long dim1228Kuo3Init[]   =   {   1   ,   3   ,   7   ,   5   ,   15  ,   61  ,   121 ,   73  ,   367 ,   13  ,   1455    ,   561 ,   7475    ,   8479    ,0 };
            const unsigned long dim1229Kuo3Init[]   =   {   1   ,   3   ,   1   ,   15  ,   17  ,   19  ,   105 ,   223 ,   481 ,   49  ,   217 ,   2935    ,   2731    ,   9857    ,0 };
            const unsigned long dim1230Kuo3Init[]   =   {   1   ,   3   ,   1   ,   7   ,   3   ,   41  ,   3   ,   195 ,   245 ,   969 ,   139 ,   2727    ,   6957    ,   2049    ,0 };
            const unsigned long dim1231Kuo3Init[]   =   {   1   ,   1   ,   5   ,   1   ,   13  ,   39  ,   55  ,   247 ,   279 ,   991 ,   875 ,   1393    ,   2649    ,   14477   ,0 };
            const unsigned long dim1232Kuo3Init[]   =   {   1   ,   1   ,   7   ,   3   ,   19  ,   1   ,   83  ,   87  ,   277 ,   939 ,   105 ,   681 ,   5377    ,   16067   ,0 };
            const unsigned long dim1233Kuo3Init[]   =   {   1   ,   3   ,   3   ,   3   ,   29  ,   9   ,   63  ,   67  ,   97  ,   457 ,   179 ,   1137    ,   5067    ,   14541   ,0 };
            const unsigned long dim1234Kuo3Init[]   =   {   1   ,   1   ,   7   ,   5   ,   25  ,   23  ,   17  ,   175 ,   33  ,   523 ,   2023    ,   649 ,   6889    ,   3481    ,0 };
            const unsigned long dim1235Kuo3Init[]   =   {   1   ,   1   ,   7   ,   7   ,   13  ,   7   ,   7   ,   145 ,   81  ,   69  ,   1029    ,   1351    ,   285 ,   13007   ,0 };
            const unsigned long dim1236Kuo3Init[]   =   {   1   ,   1   ,   5   ,   3   ,   3   ,   23  ,   107 ,   67  ,   327 ,   463 ,   1123    ,   1981    ,   1189    ,   6921    ,0 };
            const unsigned long dim1237Kuo3Init[]   =   {   1   ,   1   ,   7   ,   3   ,   15  ,   31  ,   15  ,   51  ,   421 ,   191 ,   1663    ,   2257    ,   2633    ,   4685    ,0 };
            const unsigned long dim1238Kuo3Init[]   =   {   1   ,   3   ,   7   ,   11  ,   9   ,   41  ,   101 ,   43  ,   23  ,   453 ,   1399    ,   2349    ,   8013    ,   15791   ,0 };
            const unsigned long dim1239Kuo3Init[]   =   {   1   ,   3   ,   3   ,   11  ,   27  ,   47  ,   15  ,   193 ,   99  ,   355 ,   1873    ,   1541    ,   5957    ,   13079   ,0 };
            const unsigned long dim1240Kuo3Init[]   =   {   1   ,   3   ,   1   ,   9   ,   11  ,   45  ,   43  ,   171 ,   355 ,   17  ,   955 ,   1153    ,   1437    ,   5477    ,0 };
            const unsigned long dim1241Kuo3Init[]   =   {   1   ,   3   ,   7   ,   7   ,   29  ,   25  ,   11  ,   231 ,   79  ,   951 ,   1619    ,   69  ,   2279    ,   7249    ,0 };
            const unsigned long dim1242Kuo3Init[]   =   {   1   ,   1   ,   5   ,   1   ,   5   ,   41  ,   47  ,   229 ,   353 ,   261 ,   443 ,   1181    ,   1813    ,   7929    ,0 };
            const unsigned long dim1243Kuo3Init[]   =   {   1   ,   3   ,   7   ,   13  ,   19  ,   45  ,   69  ,   11  ,   209 ,   477 ,   1623    ,   429 ,   1853    ,   13571   ,0 };
            const unsigned long dim1244Kuo3Init[]   =   {   1   ,   3   ,   3   ,   5   ,   19  ,   55  ,   49  ,   243 ,   357 ,   675 ,   129 ,   2801    ,   2511    ,   12965   ,0 };
            const unsigned long dim1245Kuo3Init[]   =   {   1   ,   1   ,   3   ,   15  ,   11  ,   47  ,   11  ,   11  ,   291 ,   943 ,   573 ,   2257    ,   1767    ,   2547    ,0 };
            const unsigned long dim1246Kuo3Init[]   =   {   1   ,   1   ,   7   ,   11  ,   1   ,   11  ,   103 ,   65  ,   395 ,   623 ,   697 ,   3375    ,   7977    ,   13449   ,0 };
            const unsigned long dim1247Kuo3Init[]   =   {   1   ,   1   ,   3   ,   3   ,   3   ,   35  ,   69  ,   59  ,   85  ,   379 ,   835 ,   557 ,   873 ,   16281   ,0 };
            const unsigned long dim1248Kuo3Init[]   =   {   1   ,   1   ,   7   ,   3   ,   31  ,   41  ,   59  ,   137 ,   297 ,   43  ,   1111    ,   2119    ,   3109    ,   1551    ,0 };
            const unsigned long dim1249Kuo3Init[]   =   {   1   ,   3   ,   7   ,   13  ,   11  ,   11  ,   87  ,   199 ,   283 ,   461 ,   951 ,   2429    ,   6739    ,   11499   ,0 };
            const unsigned long dim1250Kuo3Init[]   =   {   1   ,   3   ,   7   ,   7   ,   31  ,   13  ,   89  ,   75  ,   117 ,   755 ,   1761    ,   581 ,   7233    ,   8395    ,0 };
            const unsigned long dim1251Kuo3Init[]   =   {   1   ,   3   ,   7   ,   7   ,   1   ,   29  ,   117 ,   213 ,   81  ,   901 ,   357 ,   27  ,   127 ,   10541   ,0 };
            const unsigned long dim1252Kuo3Init[]   =   {   1   ,   1   ,   7   ,   5   ,   31  ,   1   ,   11  ,   243 ,   321 ,   973 ,   1753    ,   2363    ,   6717    ,   14861   ,0 };
            const unsigned long dim1253Kuo3Init[]   =   {   1   ,   1   ,   1   ,   1   ,   17  ,   25  ,   35  ,   211 ,   197 ,   697 ,   1389    ,   1371    ,   3687    ,   11941   ,0 };
            const unsigned long dim1254Kuo3Init[]   =   {   1   ,   3   ,   5   ,   11  ,   15  ,   35  ,   27  ,   179 ,   367 ,   891 ,   483 ,   3581    ,   4343    ,   9581    ,0 };
            const unsigned long dim1255Kuo3Init[]   =   {   1   ,   3   ,   5   ,   1   ,   13  ,   41  ,   67  ,   75  ,   309 ,   601 ,   247 ,   1609    ,   4365    ,   3003    ,0 };
            const unsigned long dim1256Kuo3Init[]   =   {   1   ,   1   ,   7   ,   11  ,   1   ,   59  ,   55  ,   213 ,   415 ,   445 ,   1763    ,   3615    ,   6847    ,   11931   ,0 };
            const unsigned long dim1257Kuo3Init[]   =   {   1   ,   1   ,   7   ,   13  ,   25  ,   53  ,   9   ,   63  ,   501 ,   95  ,   1745    ,   3273    ,   3551    ,   12733   ,0 };
            const unsigned long dim1258Kuo3Init[]   =   {   1   ,   3   ,   7   ,   7   ,   13  ,   23  ,   117 ,   75  ,   169 ,   11  ,   1401    ,   125 ,   2127    ,   15609   ,0 };
            const unsigned long dim1259Kuo3Init[]   =   {   1   ,   3   ,   3   ,   9   ,   21  ,   53  ,   113 ,   247 ,   495 ,   39  ,   1491    ,   985 ,   4943    ,   3683    ,0 };
            const unsigned long dim1260Kuo3Init[]   =   {   1   ,   1   ,   1   ,   7   ,   5   ,   49  ,   23  ,   207 ,   43  ,   863 ,   749 ,   3363    ,   1191    ,   1305    ,0 };
            const unsigned long dim1261Kuo3Init[]   =   {   1   ,   1   ,   5   ,   15  ,   3   ,   13  ,   19  ,   123 ,   377 ,   803 ,   319 ,   2617    ,   7781    ,   5113    ,0 };
            const unsigned long dim1262Kuo3Init[]   =   {   1   ,   3   ,   3   ,   13  ,   23  ,   53  ,   19  ,   223 ,   63  ,   729 ,   599 ,   2437    ,   1023    ,   13659   ,0 };
            const unsigned long dim1263Kuo3Init[]   =   {   1   ,   3   ,   7   ,   13  ,   7   ,   31  ,   101 ,   227 ,   53  ,   933 ,   405 ,   1077    ,   1341    ,   12049   ,0 };
            const unsigned long dim1264Kuo3Init[]   =   {   1   ,   1   ,   3   ,   11  ,   21  ,   31  ,   65  ,   191 ,   175 ,   735 ,   593 ,   805 ,   4405    ,   7893    ,0 };
            const unsigned long dim1265Kuo3Init[]   =   {   1   ,   3   ,   3   ,   11  ,   11  ,   23  ,   57  ,   71  ,   367 ,   275 ,   715 ,   81  ,   6455    ,   9509    ,0 };
            const unsigned long dim1266Kuo3Init[]   =   {   1   ,   1   ,   5   ,   9   ,   7   ,   5   ,   99  ,   19  ,   193 ,   501 ,   243 ,   1971    ,   4685    ,   455 ,0 };
            const unsigned long dim1267Kuo3Init[]   =   {   1   ,   3   ,   3   ,   1   ,   11  ,   59  ,   45  ,   241 ,   257 ,   107 ,   605 ,   3885    ,   813 ,   15667   ,0 };
            const unsigned long dim1268Kuo3Init[]   =   {   1   ,   3   ,   7   ,   9   ,   17  ,   63  ,   93  ,   129 ,   191 ,   839 ,   535 ,   271 ,   1087    ,   10393   ,0 };
            const unsigned long dim1269Kuo3Init[]   =   {   1   ,   1   ,   3   ,   9   ,   19  ,   15  ,   11  ,   85  ,   377 ,   653 ,   891 ,   1909    ,   5561    ,   1289    ,0 };
            const unsigned long dim1270Kuo3Init[]   =   {   1   ,   3   ,   1   ,   15  ,   7   ,   25  ,   25  ,   67  ,   55  ,   749 ,   1541    ,   1183    ,   1563    ,   5377    ,0 };
            const unsigned long dim1271Kuo3Init[]   =   {   1   ,   3   ,   1   ,   5   ,   11  ,   57  ,   121 ,   173 ,   317 ,   777 ,   749 ,   4037    ,   123 ,   10707   ,0 };
            const unsigned long dim1272Kuo3Init[]   =   {   1   ,   3   ,   3   ,   5   ,   29  ,   45  ,   99  ,   215 ,   103 ,   447 ,   1743    ,   991 ,   7649    ,   11773   ,0 };
            const unsigned long dim1273Kuo3Init[]   =   {   1   ,   1   ,   5   ,   7   ,   3   ,   33  ,   39  ,   169 ,   255 ,   383 ,   559 ,   3635    ,   5669    ,   4267    ,0 };
            const unsigned long dim1274Kuo3Init[]   =   {   1   ,   3   ,   1   ,   5   ,   15  ,   41  ,   51  ,   69  ,   373 ,   815 ,   293 ,   2125    ,   4903    ,   12429   ,0 };
            const unsigned long dim1275Kuo3Init[]   =   {   1   ,   3   ,   3   ,   1   ,   17  ,   53  ,   101 ,   153 ,   271 ,   491 ,   801 ,   2811    ,   619 ,   2021    ,0 };
            const unsigned long dim1276Kuo3Init[]   =   {   1   ,   1   ,   3   ,   5   ,   25  ,   37  ,   33  ,   139 ,   385 ,   323 ,   601 ,   1077    ,   753 ,   3925    ,0 };
            const unsigned long dim1277Kuo3Init[]   =   {   1   ,   1   ,   3   ,   1   ,   25  ,   21  ,   127 ,   67  ,   463 ,   121 ,   1765    ,   1561    ,   5   ,   5161    ,0 };
            const unsigned long dim1278Kuo3Init[]   =   {   1   ,   1   ,   7   ,   7   ,   25  ,   27  ,   37  ,   159 ,   453 ,   847 ,   353 ,   3051    ,   4269    ,   13981   ,0 };
            const unsigned long dim1279Kuo3Init[]   =   {   1   ,   3   ,   5   ,   11  ,   13  ,   1   ,   11  ,   123 ,   459 ,   989 ,   1459    ,   3793    ,   5915    ,   8411    ,0 };
            const unsigned long dim1280Kuo3Init[]   =   {   1   ,   3   ,   1   ,   15  ,   5   ,   21  ,   37  ,   41  ,   399 ,   471 ,   1147    ,   143 ,   2737    ,   9447    ,0 };
            const unsigned long dim1281Kuo3Init[]   =   {   1   ,   3   ,   1   ,   5   ,   27  ,   5   ,   121 ,   77  ,   183 ,   209 ,   49  ,   3737    ,   4317    ,   11225   ,0 };
            const unsigned long dim1282Kuo3Init[]   =   {   1   ,   1   ,   3   ,   7   ,   5   ,   9   ,   65  ,   25  ,   237 ,   21  ,   139 ,   1729    ,   965 ,   15419   ,0 };
            const unsigned long dim1283Kuo3Init[]   =   {   1   ,   1   ,   3   ,   11  ,   5   ,   47  ,   115 ,   211 ,   495 ,   97  ,   1439    ,   127 ,   4823    ,   11643   ,0 };
            const unsigned long dim1284Kuo3Init[]   =   {   1   ,   1   ,   5   ,   1   ,   11  ,   37  ,   19  ,   129 ,   35  ,   1017    ,   1057    ,   3501    ,   4293    ,   15271   ,0 };
            const unsigned long dim1285Kuo3Init[]   =   {   1   ,   3   ,   5   ,   5   ,   3   ,   31  ,   45  ,   25  ,   441 ,   729 ,   909 ,   3519    ,   5579    ,   10341   ,0 };
            const unsigned long dim1286Kuo3Init[]   =   {   1   ,   1   ,   5   ,   13  ,   23  ,   55  ,   13  ,   47  ,   25  ,   319 ,   697 ,   2839    ,   209 ,   5007    ,0 };
            const unsigned long dim1287Kuo3Init[]   =   {   1   ,   3   ,   1   ,   13  ,   31  ,   7   ,   81  ,   239 ,   305 ,   667 ,   233 ,   3555    ,   7385    ,   10113   ,0 };
            const unsigned long dim1288Kuo3Init[]   =   {   1   ,   3   ,   3   ,   13  ,   23  ,   7   ,   101 ,   193 ,   343 ,   435 ,   1065    ,   3399    ,   1855    ,   15231   ,0 };
            const unsigned long dim1289Kuo3Init[]   =   {   1   ,   3   ,   7   ,   7   ,   3   ,   31  ,   13  ,   51  ,   141 ,   523 ,   1545    ,   31  ,   3827    ,   13871   ,0 };
            const unsigned long dim1290Kuo3Init[]   =   {   1   ,   1   ,   1   ,   3   ,   29  ,   37  ,   117 ,   187 ,   337 ,   799 ,   923 ,   3919    ,   5861    ,   3167    ,0 };
            const unsigned long dim1291Kuo3Init[]   =   {   1   ,   3   ,   7   ,   1   ,   21  ,   17  ,   59  ,   73  ,   223 ,   455 ,   2025    ,   2701    ,   673 ,   8333    ,0 };
            const unsigned long dim1292Kuo3Init[]   =   {   1   ,   1   ,   1   ,   3   ,   9   ,   53  ,   99  ,   87  ,   35  ,   637 ,   1553    ,   1417    ,   1233    ,   855 ,0 };
            const unsigned long dim1293Kuo3Init[]   =   {   1   ,   1   ,   3   ,   11  ,   25  ,   21  ,   3   ,   227 ,   345 ,   711 ,   1347    ,   437 ,   7513    ,   7731    ,0 };
            const unsigned long dim1294Kuo3Init[]   =   {   1   ,   1   ,   1   ,   3   ,   5   ,   51  ,   125 ,   133 ,   241 ,   327 ,   589 ,   1895    ,   1627    ,   6945    ,0 };
            const unsigned long dim1295Kuo3Init[]   =   {   1   ,   1   ,   1   ,   3   ,   21  ,   19  ,   101 ,   95  ,   143 ,   353 ,   1783    ,   2815    ,   3691    ,   11209   ,0 };
            const unsigned long dim1296Kuo3Init[]   =   {   1   ,   1   ,   3   ,   1   ,   31  ,   29  ,   95  ,   121 ,   253 ,   925 ,   277 ,   2857    ,   963 ,   7701    ,0 };
            const unsigned long dim1297Kuo3Init[]   =   {   1   ,   3   ,   7   ,   5   ,   3   ,   23  ,   31  ,   63  ,   467 ,   91  ,   1235    ,   1927    ,   4347    ,   9281    ,0 };
            const unsigned long dim1298Kuo3Init[]   =   {   1   ,   1   ,   5   ,   15  ,   21  ,   23  ,   113 ,   219 ,   69  ,   13  ,   1165    ,   1043    ,   6335    ,   3499    ,0 };
            const unsigned long dim1299Kuo3Init[]   =   {   1   ,   1   ,   7   ,   15  ,   25  ,   35  ,   67  ,   67  ,   273 ,   421 ,   521 ,   1169    ,   1557    ,   1523    ,0 };
            const unsigned long dim1300Kuo3Init[]   =   {   1   ,   3   ,   5   ,   7   ,   25  ,   5   ,   19  ,   157 ,   385 ,   35  ,   1701    ,   1697    ,   1425    ,   12549   ,0 };
            const unsigned long dim1301Kuo3Init[]   =   {   1   ,   1   ,   3   ,   1   ,   15  ,   13  ,   31  ,   219 ,   159 ,   831 ,   909 ,   1907    ,   4251    ,   4231    ,0 };
            const unsigned long dim1302Kuo3Init[]   =   {   1   ,   3   ,   5   ,   7   ,   27  ,   51  ,   19  ,   205 ,   177 ,   769 ,   1881    ,   3961    ,   7715    ,   13351   ,0 };
            const unsigned long dim1303Kuo3Init[]   =   {   1   ,   1   ,   3   ,   5   ,   13  ,   19  ,   1   ,   91  ,   257 ,   201 ,   459 ,   525 ,   5093    ,   7259    ,0 };
            const unsigned long dim1304Kuo3Init[]   =   {   1   ,   3   ,   7   ,   9   ,   3   ,   37  ,   87  ,   117 ,   209 ,   9   ,   1019    ,   195 ,   1531    ,   14445   ,0 };
            const unsigned long dim1305Kuo3Init[]   =   {   1   ,   1   ,   1   ,   5   ,   23  ,   35  ,   21  ,   27  ,   455 ,   665 ,   455 ,   1009    ,   2315    ,   11465   ,0 };
            const unsigned long dim1306Kuo3Init[]   =   {   1   ,   3   ,   7   ,   1   ,   17  ,   55  ,   3   ,   71  ,   239 ,   183 ,   889 ,   2091    ,   2853    ,   12541   ,0 };
            const unsigned long dim1307Kuo3Init[]   =   {   1   ,   1   ,   3   ,   15  ,   5   ,   7   ,   67  ,   127 ,   411 ,   719 ,   835 ,   2299    ,   1339    ,   12149   ,0 };
            const unsigned long dim1308Kuo3Init[]   =   {   1   ,   1   ,   3   ,   3   ,   7   ,   21  ,   113 ,   199 ,   17  ,   161 ,   1945    ,   1141    ,   4879    ,   1579    ,0 };
            const unsigned long dim1309Kuo3Init[]   =   {   1   ,   3   ,   1   ,   7   ,   5   ,   9   ,   35  ,   169 ,   317 ,   355 ,   193 ,   599 ,   7653    ,   7689    ,0 };
            const unsigned long dim1310Kuo3Init[]   =   {   1   ,   1   ,   1   ,   13  ,   29  ,   35  ,   59  ,   85  ,   139 ,   199 ,   1075    ,   2799    ,   6263    ,   11133   ,0 };
            const unsigned long dim1311Kuo3Init[]   =   {   1   ,   1   ,   5   ,   3   ,   19  ,   63  ,   13  ,   41  ,   289 ,   845 ,   1455    ,   633 ,   4995    ,   2037    ,0 };
            const unsigned long dim1312Kuo3Init[]   =   {   1   ,   1   ,   3   ,   11  ,   19  ,   25  ,   65  ,   213 ,   27  ,   257 ,   1151    ,   4063    ,   5587    ,   2205    ,0 };
            const unsigned long dim1313Kuo3Init[]   =   {   1   ,   1   ,   7   ,   7   ,   9   ,   53  ,   103 ,   49  ,   187 ,   579 ,   801 ,   263 ,   2975    ,   12013   ,0 };
            const unsigned long dim1314Kuo3Init[]   =   {   1   ,   3   ,   1   ,   13  ,   7   ,   13  ,   85  ,   11  ,   365 ,   63  ,   595 ,   747 ,   1565    ,   1817    ,0 };
            const unsigned long dim1315Kuo3Init[]   =   {   1   ,   3   ,   7   ,   15  ,   11  ,   1   ,   45  ,   3   ,   181 ,   357 ,   447 ,   821 ,   6471    ,   10121   ,0 };
            const unsigned long dim1316Kuo3Init[]   =   {   1   ,   1   ,   7   ,   11  ,   7   ,   25  ,   75  ,   139 ,   463 ,   679 ,   1327    ,   2863    ,   2373    ,   10681   ,0 };
            const unsigned long dim1317Kuo3Init[]   =   {   1   ,   3   ,   5   ,   5   ,   17  ,   7   ,   103 ,   157 ,   241 ,   781 ,   399 ,   3951    ,   6587    ,   35  ,0 };
            const unsigned long dim1318Kuo3Init[]   =   {   1   ,   3   ,   5   ,   5   ,   17  ,   29  ,   71  ,   115 ,   415 ,   91  ,   729 ,   3055    ,   5219    ,   15271   ,0 };
            const unsigned long dim1319Kuo3Init[]   =   {   1   ,   3   ,   1   ,   9   ,   7   ,   5   ,   123 ,   235 ,   33  ,   103 ,   257 ,   1151    ,   3931    ,   16257   ,0 };
            const unsigned long dim1320Kuo3Init[]   =   {   1   ,   1   ,   3   ,   11  ,   27  ,   29  ,   29  ,   31  ,   463 ,   187 ,   103 ,   209 ,   747 ,   14311   ,0 };
            const unsigned long dim1321Kuo3Init[]   =   {   1   ,   3   ,   5   ,   1   ,   11  ,   61  ,   79  ,   103 ,   233 ,   733 ,   779 ,   705 ,   2377    ,   12689   ,0 };
            const unsigned long dim1322Kuo3Init[]   =   {   1   ,   3   ,   1   ,   9   ,   31  ,   11  ,   123 ,   239 ,   227 ,   81  ,   525 ,   2099    ,   8149    ,   5851    ,0 };
            const unsigned long dim1323Kuo3Init[]   =   {   1   ,   3   ,   7   ,   5   ,   1   ,   37  ,   9   ,   199 ,   313 ,   875 ,   975 ,   777 ,   5449    ,   3075    ,0 };
            const unsigned long dim1324Kuo3Init[]   =   {   1   ,   3   ,   5   ,   11  ,   7   ,   29  ,   5   ,   133 ,   419 ,   691 ,   859 ,   1089    ,   4331    ,   11079   ,0 };
            const unsigned long dim1325Kuo3Init[]   =   {   1   ,   1   ,   1   ,   1   ,   5   ,   57  ,   113 ,   17  ,   481 ,   989 ,   233 ,   1523    ,   3291    ,   1665    ,0 };
            const unsigned long dim1326Kuo3Init[]   =   {   1   ,   3   ,   7   ,   15  ,   27  ,   31  ,   43  ,   247 ,   403 ,   395 ,   69  ,   2561    ,   7281    ,   7291    ,0 };
            const unsigned long dim1327Kuo3Init[]   =   {   1   ,   3   ,   3   ,   11  ,   31  ,   43  ,   105 ,   237 ,   21  ,   719 ,   861 ,   199 ,   1045    ,   7689    ,0 };
            const unsigned long dim1328Kuo3Init[]   =   {   1   ,   3   ,   1   ,   13  ,   1   ,   15  ,   3   ,   35  ,   219 ,   993 ,   891 ,   3649    ,   3795    ,   11599   ,0 };
            const unsigned long dim1329Kuo3Init[]   =   {   1   ,   3   ,   7   ,   7   ,   7   ,   15  ,   125 ,   243 ,   449 ,   173 ,   1041    ,   1433    ,   6559    ,   14285   ,0 };
            const unsigned long dim1330Kuo3Init[]   =   {   1   ,   1   ,   7   ,   7   ,   1   ,   5   ,   75  ,   189 ,   439 ,   745 ,   1361    ,   3525    ,   2913    ,   10513   ,0 };
            const unsigned long dim1331Kuo3Init[]   =   {   1   ,   1   ,   5   ,   1   ,   31  ,   7   ,   49  ,   41  ,   413 ,   347 ,   881 ,   2953    ,   2173    ,   8667    ,0 };
            const unsigned long dim1332Kuo3Init[]   =   {   1   ,   3   ,   3   ,   1   ,   27  ,   33  ,   75  ,   99  ,   451 ,   207 ,   1847    ,   1285    ,   3217    ,   7899    ,0 };
            const unsigned long dim1333Kuo3Init[]   =   {   1   ,   3   ,   7   ,   11  ,   21  ,   63  ,   87  ,   11  ,   429 ,   591 ,   1471    ,   3221    ,   2441    ,   2889    ,0 };
            const unsigned long dim1334Kuo3Init[]   =   {   1   ,   3   ,   1   ,   15  ,   3   ,   47  ,   103 ,   219 ,   317 ,   605 ,   1709    ,   2473    ,   3511    ,   10035   ,0 };
            const unsigned long dim1335Kuo3Init[]   =   {   1   ,   1   ,   5   ,   3   ,   13  ,   21  ,   117 ,   113 ,   443 ,   863 ,   201 ,   1293    ,   1291    ,   5537    ,0 };
            const unsigned long dim1336Kuo3Init[]   =   {   1   ,   1   ,   7   ,   7   ,   3   ,   15  ,   59  ,   45  ,   135 ,   465 ,   1033    ,   605 ,   2783    ,   1001    ,0 };
            const unsigned long dim1337Kuo3Init[]   =   {   1   ,   3   ,   1   ,   13  ,   25  ,   49  ,   125 ,   171 ,   231 ,   691 ,   1759    ,   3493    ,   1703    ,   4405    ,0 };
            const unsigned long dim1338Kuo3Init[]   =   {   1   ,   1   ,   5   ,   11  ,   3   ,   31  ,   73  ,   31  ,   189 ,   533 ,   1191    ,   487 ,   6233    ,   5607    ,0 };
            const unsigned long dim1339Kuo3Init[]   =   {   1   ,   3   ,   7   ,   5   ,   11  ,   23  ,   91  ,   207 ,   337 ,   225 ,   339 ,   2085    ,   6223    ,   16299   ,0 };
            const unsigned long dim1340Kuo3Init[]   =   {   1   ,   1   ,   5   ,   13  ,   21  ,   35  ,   47  ,   81  ,   229 ,   815 ,   767 ,   1939    ,   1335    ,   1865    ,0 };
            const unsigned long dim1341Kuo3Init[]   =   {   1   ,   1   ,   5   ,   9   ,   3   ,   11  ,   19  ,   253 ,   499 ,   13  ,   1771    ,   1461    ,   2339    ,   1217    ,0 };
            const unsigned long dim1342Kuo3Init[]   =   {   1   ,   1   ,   1   ,   7   ,   15  ,   51  ,   91  ,   213 ,   435 ,   137 ,   551 ,   3483    ,   1485    ,   4061    ,0 };
            const unsigned long dim1343Kuo3Init[]   =   {   1   ,   3   ,   3   ,   9   ,   23  ,   25  ,   57  ,   1   ,   147 ,   705 ,   399 ,   3061    ,   5921    ,   4201    ,0 };
            const unsigned long dim1344Kuo3Init[]   =   {   1   ,   3   ,   3   ,   9   ,   9   ,   15  ,   53  ,   105 ,   175 ,   147 ,   1077    ,   2323    ,   6041    ,   9591    ,0 };
            const unsigned long dim1345Kuo3Init[]   =   {   1   ,   1   ,   5   ,   9   ,   5   ,   17  ,   125 ,   243 ,   5   ,   329 ,   161 ,   795 ,   8105    ,   8385    ,0 };
            const unsigned long dim1346Kuo3Init[]   =   {   1   ,   3   ,   1   ,   13  ,   9   ,   7   ,   83  ,   49  ,   83  ,   109 ,   759 ,   2197    ,   2587    ,   10715   ,0 };
            const unsigned long dim1347Kuo3Init[]   =   {   1   ,   3   ,   3   ,   9   ,   17  ,   11  ,   33  ,   173 ,   133 ,   853 ,   1005    ,   3027    ,   2137    ,   14471   ,0 };
            const unsigned long dim1348Kuo3Init[]   =   {   1   ,   3   ,   7   ,   9   ,   23  ,   19  ,   29  ,   169 ,   73  ,   405 ,   1175    ,   3179    ,   2449    ,   13447   ,0 };
            const unsigned long dim1349Kuo3Init[]   =   {   1   ,   1   ,   7   ,   1   ,   21  ,   43  ,   1   ,   85  ,   387 ,   979 ,   571 ,   1849    ,   3287    ,   1201    ,0 };
            const unsigned long dim1350Kuo3Init[]   =   {   1   ,   1   ,   5   ,   13  ,   15  ,   43  ,   125 ,   49  ,   365 ,   689 ,   1569    ,   719 ,   3925    ,   3781    ,0 };
            const unsigned long dim1351Kuo3Init[]   =   {   1   ,   1   ,   7   ,   1   ,   11  ,   5   ,   39  ,   181 ,   39  ,   729 ,   641 ,   1983    ,   7321    ,   14747   ,0 };
            const unsigned long dim1352Kuo3Init[]   =   {   1   ,   3   ,   7   ,   11  ,   11  ,   9   ,   115 ,   125 ,   305 ,   371 ,   1647    ,   463 ,   2473    ,   9029    ,0 };
            const unsigned long dim1353Kuo3Init[]   =   {   1   ,   1   ,   7   ,   15  ,   5   ,   11  ,   69  ,   177 ,   249 ,   235 ,   1843    ,   695 ,   7931    ,   10891   ,0 };
            const unsigned long dim1354Kuo3Init[]   =   {   1   ,   1   ,   5   ,   11  ,   17  ,   57  ,   127 ,   135 ,   357 ,   315 ,   55  ,   2867    ,   4187    ,   15803   ,0 };
            const unsigned long dim1355Kuo3Init[]   =   {   1   ,   3   ,   1   ,   13  ,   29  ,   23  ,   69  ,   179 ,   151 ,   813 ,   1317    ,   1943    ,   3995    ,   13023   ,0 };
            const unsigned long dim1356Kuo3Init[]   =   {   1   ,   1   ,   7   ,   11  ,   9   ,   55  ,   39  ,   65  ,   3   ,   127 ,   1781    ,   117 ,   2651    ,   16111   ,0 };
            const unsigned long dim1357Kuo3Init[]   =   {   1   ,   3   ,   5   ,   15  ,   31  ,   19  ,   89  ,   185 ,   351 ,   623 ,   1301    ,   3123    ,   5905    ,   7231    ,0 };
            const unsigned long dim1358Kuo3Init[]   =   {   1   ,   3   ,   3   ,   3   ,   31  ,   25  ,   31  ,   163 ,   81  ,   927 ,   1171    ,   2667    ,   5699    ,   7555    ,0 };
            const unsigned long dim1359Kuo3Init[]   =   {   1   ,   1   ,   3   ,   13  ,   17  ,   55  ,   83  ,   217 ,   97  ,   795 ,   1745    ,   875 ,   483 ,   3271    ,0 };
            const unsigned long dim1360Kuo3Init[]   =   {   1   ,   1   ,   1   ,   1   ,   9   ,   13  ,   19  ,   207 ,   325 ,   277 ,   1393    ,   2619    ,   3055    ,   8499    ,0 };
            const unsigned long dim1361Kuo3Init[]   =   {   1   ,   1   ,   7   ,   5   ,   1   ,   11  ,   113 ,   181 ,   411 ,   55  ,   505 ,   3125    ,   619 ,   1749    ,0 };
            const unsigned long dim1362Kuo3Init[]   =   {   1   ,   3   ,   3   ,   15  ,   11  ,   47  ,   25  ,   123 ,   143 ,   51  ,   1075    ,   715 ,   4647    ,   13995   ,0 };
            const unsigned long dim1363Kuo3Init[]   =   {   1   ,   1   ,   1   ,   1   ,   21  ,   23  ,   121 ,   217 ,   325 ,   589 ,   2017    ,   3027    ,   6175    ,   1293    ,0 };
            const unsigned long dim1364Kuo3Init[]   =   {   1   ,   3   ,   1   ,   9   ,   3   ,   47  ,   35  ,   233 ,   381 ,   905 ,   1031    ,   2229    ,   2575    ,   6891    ,0 };
            const unsigned long dim1365Kuo3Init[]   =   {   1   ,   1   ,   1   ,   5   ,   3   ,   33  ,   17  ,   39  ,   377 ,   635 ,   1281    ,   2673    ,   2101    ,   2921    ,0 };
            const unsigned long dim1366Kuo3Init[]   =   {   1   ,   1   ,   7   ,   7   ,   13  ,   21  ,   43  ,   55  ,   353 ,   643 ,   533 ,   57  ,   5583    ,   2189    ,0 };
            const unsigned long dim1367Kuo3Init[]   =   {   1   ,   3   ,   5   ,   3   ,   27  ,   15  ,   79  ,   47  ,   463 ,   327 ,   1231    ,   171 ,   4931    ,   7549    ,0 };
            const unsigned long dim1368Kuo3Init[]   =   {   1   ,   1   ,   5   ,   9   ,   11  ,   27  ,   109 ,   15  ,   57  ,   35  ,   1347    ,   3061    ,   8179    ,   5461    ,0 };
            const unsigned long dim1369Kuo3Init[]   =   {   1   ,   3   ,   3   ,   7   ,   29  ,   57  ,   103 ,   221 ,   409 ,   437 ,   1541    ,   2591    ,   4757    ,   13735   ,0 };
            const unsigned long dim1370Kuo3Init[]   =   {   1   ,   1   ,   7   ,   1   ,   3   ,   23  ,   115 ,   35  ,   165 ,   389 ,   1399    ,   2081    ,   3169    ,   1493    ,0 };
            const unsigned long dim1371Kuo3Init[]   =   {   1   ,   1   ,   3   ,   3   ,   25  ,   27  ,   93  ,   253 ,   83  ,   557 ,   957 ,   2017    ,   2005    ,   10005   ,0 };
            const unsigned long dim1372Kuo3Init[]   =   {   1   ,   3   ,   7   ,   7   ,   1   ,   33  ,   55  ,   227 ,   427 ,   595 ,   1235    ,   3815    ,   5397    ,   4827    ,0 };
            const unsigned long dim1373Kuo3Init[]   =   {   1   ,   3   ,   1   ,   7   ,   31  ,   23  ,   53  ,   155 ,   177 ,   851 ,   531 ,   2441    ,   2049    ,   12209   ,0 };
            const unsigned long dim1374Kuo3Init[]   =   {   1   ,   1   ,   7   ,   3   ,   21  ,   27  ,   13  ,   67  ,   467 ,   607 ,   611 ,   2565    ,   5907    ,   8441    ,0 };
            const unsigned long dim1375Kuo3Init[]   =   {   1   ,   1   ,   7   ,   5   ,   31  ,   59  ,   99  ,   65  ,   403 ,   439 ,   1407    ,   2587    ,   3111    ,   8715    ,0 };
            const unsigned long dim1376Kuo3Init[]   =   {   1   ,   1   ,   5   ,   13  ,   7   ,   3   ,   35  ,   69  ,   177 ,   591 ,   449 ,   369 ,   5047    ,   3435    ,0 };
            const unsigned long dim1377Kuo3Init[]   =   {   1   ,   3   ,   5   ,   5   ,   17  ,   27  ,   9   ,   49  ,   161 ,   925 ,   603 ,   2555    ,   3349    ,   4913    ,0 };
            const unsigned long dim1378Kuo3Init[]   =   {   1   ,   1   ,   1   ,   9   ,   11  ,   43  ,   89  ,   139 ,   163 ,   851 ,   1991    ,   3439    ,   227 ,   5875    ,0 };
            const unsigned long dim1379Kuo3Init[]   =   {   1   ,   3   ,   3   ,   9   ,   15  ,   7   ,   63  ,   133 ,   437 ,   447 ,   1227    ,   3945    ,   3467    ,   14919   ,0 };
            const unsigned long dim1380Kuo3Init[]   =   {   1   ,   1   ,   1   ,   7   ,   13  ,   49  ,   113 ,   39  ,   77  ,   773 ,   127 ,   3361    ,   2325    ,   987 ,0 };
            const unsigned long dim1381Kuo3Init[]   =   {   1   ,   1   ,   5   ,   9   ,   31  ,   23  ,   69  ,   203 ,   189 ,   35  ,   647 ,   3507    ,   6679    ,   2097    ,0 };
            const unsigned long dim1382Kuo3Init[]   =   {   1   ,   3   ,   7   ,   5   ,   23  ,   9   ,   99  ,   253 ,   203 ,   167 ,   1073    ,   435 ,   1745    ,   2155    ,0 };
            const unsigned long dim1383Kuo3Init[]   =   {   1   ,   1   ,   5   ,   11  ,   11  ,   27  ,   15  ,   125 ,   291 ,   261 ,   1953    ,   3695    ,   6345    ,   8519    ,0 };
            const unsigned long dim1384Kuo3Init[]   =   {   1   ,   1   ,   7   ,   11  ,   3   ,   39  ,   23  ,   143 ,   401 ,   587 ,   1729    ,   3083    ,   4255    ,   7427    ,0 };
            const unsigned long dim1385Kuo3Init[]   =   {   1   ,   1   ,   7   ,   9   ,   27  ,   7   ,   63  ,   153 ,   349 ,   1003    ,   1315    ,   841 ,   5991    ,   15647   ,0 };
            const unsigned long dim1386Kuo3Init[]   =   {   1   ,   3   ,   5   ,   1   ,   27  ,   45  ,   91  ,   245 ,   297 ,   485 ,   647 ,   91  ,   231 ,   5231    ,0 };
            const unsigned long dim1387Kuo3Init[]   =   {   1   ,   1   ,   5   ,   15  ,   29  ,   43  ,   53  ,   79  ,   387 ,   977 ,   247 ,   371 ,   5823    ,   8335    ,0 };
            const unsigned long dim1388Kuo3Init[]   =   {   1   ,   3   ,   3   ,   5   ,   5   ,   17  ,   29  ,   71  ,   417 ,   201 ,   1495    ,   277 ,   2087    ,   15421   ,0 };
            const unsigned long dim1389Kuo3Init[]   =   {   1   ,   3   ,   5   ,   11  ,   7   ,   47  ,   37  ,   171 ,   167 ,   355 ,   1845    ,   1747    ,   839 ,   7213    ,0 };
            const unsigned long dim1390Kuo3Init[]   =   {   1   ,   1   ,   3   ,   15  ,   15  ,   29  ,   107 ,   35  ,   183 ,   735 ,   183 ,   2993    ,   6495    ,   981 ,0 };
            const unsigned long dim1391Kuo3Init[]   =   {   1   ,   3   ,   1   ,   3   ,   3   ,   25  ,   25  ,   195 ,   461 ,   739 ,   861 ,   3027    ,   2007    ,   7771    ,0 };
            const unsigned long dim1392Kuo3Init[]   =   {   1   ,   3   ,   7   ,   5   ,   29  ,   55  ,   115 ,   57  ,   67  ,   913 ,   1603    ,   2397    ,   949 ,   14671   ,0 };
            const unsigned long dim1393Kuo3Init[]   =   {   1   ,   1   ,   1   ,   7   ,   27  ,   15  ,   51  ,   201 ,   127 ,   459 ,   1545    ,   3415    ,   1823    ,   1673    ,0 };
            const unsigned long dim1394Kuo3Init[]   =   {   1   ,   3   ,   1   ,   13  ,   21  ,   43  ,   75  ,   59  ,   379 ,   1007    ,   1697    ,   763 ,   741 ,   3801    ,0 };
            const unsigned long dim1395Kuo3Init[]   =   {   1   ,   1   ,   5   ,   13  ,   13  ,   41  ,   105 ,   235 ,   381 ,   465 ,   665 ,   877 ,   1025    ,   9041    ,0 };
            const unsigned long dim1396Kuo3Init[]   =   {   1   ,   3   ,   3   ,   3   ,   3   ,   63  ,   65  ,   151 ,   167 ,   683 ,   327 ,   3545    ,   2531    ,   13479   ,0 };
            const unsigned long dim1397Kuo3Init[]   =   {   1   ,   3   ,   5   ,   1   ,   7   ,   25  ,   61  ,   137 ,   155 ,   5   ,   623 ,   649 ,   7235    ,   13495   ,0 };
            const unsigned long dim1398Kuo3Init[]   =   {   1   ,   3   ,   5   ,   1   ,   23  ,   33  ,   3   ,   45  ,   235 ,   909 ,   1619    ,   2843    ,   5307    ,   14345   ,0 };
            const unsigned long dim1399Kuo3Init[]   =   {   1   ,   3   ,   3   ,   3   ,   29  ,   1   ,   67  ,   161 ,   237 ,   229 ,   1403    ,   3289    ,   399 ,   237 ,0 };
            const unsigned long dim1400Kuo3Init[]   =   {   1   ,   3   ,   1   ,   7   ,   5   ,   27  ,   83  ,   133 ,   189 ,   277 ,   1975    ,   1399    ,   65  ,   13645   ,0 };
            const unsigned long dim1401Kuo3Init[]   =   {   1   ,   3   ,   3   ,   9   ,   1   ,   41  ,   11  ,   121 ,   413 ,   413 ,   1817    ,   1533    ,   5299    ,   9089    ,0 };
            const unsigned long dim1402Kuo3Init[]   =   {   1   ,   1   ,   1   ,   7   ,   15  ,   59  ,   75  ,   255 ,   419 ,   441 ,   1993    ,   573 ,   6025    ,   16375   ,0 };
            const unsigned long dim1403Kuo3Init[]   =   {   1   ,   1   ,   1   ,   9   ,   3   ,   35  ,   117 ,   69  ,   473 ,   53  ,   1007    ,   2229    ,   1749    ,   14795   ,0 };
            const unsigned long dim1404Kuo3Init[]   =   {   1   ,   1   ,   3   ,   11  ,   23  ,   27  ,   117 ,   183 ,   69  ,   689 ,   1903    ,   1851    ,   1143    ,   11959   ,0 };
            const unsigned long dim1405Kuo3Init[]   =   {   1   ,   3   ,   7   ,   13  ,   3   ,   9   ,   125 ,   173 ,   365 ,   393 ,   1313    ,   3937    ,   1699    ,   15667   ,0 };
            const unsigned long dim1406Kuo3Init[]   =   {   1   ,   3   ,   5   ,   11  ,   19  ,   3   ,   45  ,   181 ,   439 ,   61  ,   1139    ,   149 ,   1929    ,   15617   ,0 };
            const unsigned long dim1407Kuo3Init[]   =   {   1   ,   3   ,   7   ,   1   ,   13  ,   41  ,   43  ,   131 ,   79  ,   947 ,   1579    ,   2055    ,   451 ,   10875   ,0 };
            const unsigned long dim1408Kuo3Init[]   =   {   1   ,   1   ,   1   ,   11  ,   27  ,   11  ,   105 ,   109 ,   503 ,   361 ,   27  ,   133 ,   801 ,   853 ,0 };
            const unsigned long dim1409Kuo3Init[]   =   {   1   ,   3   ,   5   ,   11  ,   11  ,   43  ,   3   ,   233 ,   449 ,   557 ,   2003    ,   467 ,   5617    ,   7431    ,0 };
            const unsigned long dim1410Kuo3Init[]   =   {   1   ,   1   ,   3   ,   9   ,   5   ,   31  ,   121 ,   13  ,   301 ,   433 ,   1719    ,   821 ,   8107    ,   13313   ,0 };
            const unsigned long dim1411Kuo3Init[]   =   {   1   ,   1   ,   5   ,   15  ,   25  ,   17  ,   117 ,   81  ,   239 ,   219 ,   195 ,   2653    ,   1287    ,   11595   ,0 };
            const unsigned long dim1412Kuo3Init[]   =   {   1   ,   1   ,   3   ,   9   ,   31  ,   37  ,   123 ,   35  ,   507 ,   91  ,   1345    ,   279 ,   2369    ,   1613    ,0 };
            const unsigned long dim1413Kuo3Init[]   =   {   1   ,   3   ,   3   ,   11  ,   23  ,   9   ,   83  ,   91  ,   257 ,   977 ,   697 ,   3813    ,   6839    ,   2755    ,0 };
            const unsigned long dim1414Kuo3Init[]   =   {   1   ,   3   ,   5   ,   5   ,   21  ,   23  ,   65  ,   63  ,   407 ,   515 ,   1817    ,   2893    ,   7539    ,   11803   ,0 };
            const unsigned long dim1415Kuo3Init[]   =   {   1   ,   3   ,   1   ,   15  ,   5   ,   1   ,   7   ,   195 ,   15  ,   319 ,   1177    ,   3191    ,   3489    ,   6213    ,0 };
            const unsigned long dim1416Kuo3Init[]   =   {   1   ,   1   ,   7   ,   5   ,   29  ,   11  ,   97  ,   125 ,   89  ,   1005    ,   455 ,   3397    ,   7345    ,   14481   ,0 };
            const unsigned long dim1417Kuo3Init[]   =   {   1   ,   1   ,   5   ,   11  ,   31  ,   37  ,   15  ,   35  ,   405 ,   181 ,   601 ,   543 ,   6639    ,   16103   ,0 };
            const unsigned long dim1418Kuo3Init[]   =   {   1   ,   3   ,   3   ,   1   ,   1   ,   33  ,   127 ,   177 ,   277 ,   681 ,   2039    ,   3109    ,   4657    ,   1303    ,0 };
            const unsigned long dim1419Kuo3Init[]   =   {   1   ,   1   ,   5   ,   7   ,   3   ,   61  ,   107 ,   215 ,   175 ,   973 ,   899 ,   3655    ,   823 ,   7125    ,0 };
            const unsigned long dim1420Kuo3Init[]   =   {   1   ,   1   ,   3   ,   1   ,   3   ,   47  ,   87  ,   179 ,   387 ,   987 ,   1171    ,   2767    ,   6329    ,   609 ,0 };
            const unsigned long dim1421Kuo3Init[]   =   {   1   ,   3   ,   3   ,   11  ,   7   ,   25  ,   33  ,   109 ,   485 ,   789 ,   979 ,   659 ,   2555    ,   10175   ,0 };
            const unsigned long dim1422Kuo3Init[]   =   {   1   ,   1   ,   5   ,   7   ,   15  ,   41  ,   77  ,   69  ,   487 ,   601 ,   753 ,   2785    ,   1075    ,   35  ,0 };
            const unsigned long dim1423Kuo3Init[]   =   {   1   ,   1   ,   1   ,   7   ,   23  ,   3   ,   125 ,   143 ,   463 ,   981 ,   499 ,   1249    ,   5461    ,   43  ,0 };
            const unsigned long dim1424Kuo3Init[]   =   {   1   ,   1   ,   5   ,   11  ,   9   ,   1   ,   55  ,   147 ,   85  ,   857 ,   1995    ,   3471    ,   6115    ,   5511    ,0 };
            const unsigned long dim1425Kuo3Init[]   =   {   1   ,   3   ,   3   ,   11  ,   11  ,   25  ,   27  ,   201 ,   497 ,   533 ,   643 ,   2015    ,   205 ,   13535   ,0 };
            const unsigned long dim1426Kuo3Init[]   =   {   1   ,   3   ,   3   ,   3   ,   23  ,   53  ,   77  ,   143 ,   231 ,   335 ,   471 ,   1821    ,   739 ,   4727    ,0 };
            const unsigned long dim1427Kuo3Init[]   =   {   1   ,   3   ,   1   ,   3   ,   13  ,   25  ,   53  ,   61  ,   155 ,   1013    ,   1439    ,   1745    ,   6393    ,   13329   ,0 };
            const unsigned long dim1428Kuo3Init[]   =   {   1   ,   3   ,   7   ,   1   ,   5   ,   11  ,   111 ,   21  ,   357 ,   599 ,   1213    ,   3581    ,   1221    ,   75  ,0 };
            const unsigned long dim1429Kuo3Init[]   =   {   1   ,   3   ,   7   ,   7   ,   23  ,   45  ,   79  ,   83  ,   97  ,   191 ,   53  ,   2157    ,   7737    ,   1533    ,0 };
            const unsigned long dim1430Kuo3Init[]   =   {   1   ,   3   ,   1   ,   9   ,   25  ,   17  ,   123 ,   179 ,   151 ,   77  ,   1903    ,   1457    ,   1403    ,   13557   ,0 };
            const unsigned long dim1431Kuo3Init[]   =   {   1   ,   3   ,   3   ,   1   ,   23  ,   59  ,   65  ,   47  ,   119 ,   113 ,   133 ,   3601    ,   7543    ,   16289   ,0 };
            const unsigned long dim1432Kuo3Init[]   =   {   1   ,   3   ,   5   ,   13  ,   17  ,   51  ,   33  ,   39  ,   257 ,   409 ,   659 ,   3907    ,   7301    ,   8893    ,0 };
            const unsigned long dim1433Kuo3Init[]   =   {   1   ,   1   ,   7   ,   11  ,   31  ,   25  ,   79  ,   245 ,   465 ,   809 ,   1317    ,   1109    ,   5983    ,   4359    ,0 };
            const unsigned long dim1434Kuo3Init[]   =   {   1   ,   1   ,   7   ,   7   ,   5   ,   25  ,   83  ,   239 ,   199 ,   497 ,   521 ,   1527    ,   4509    ,   12627   ,0 };
            const unsigned long dim1435Kuo3Init[]   =   {   1   ,   1   ,   1   ,   11  ,   25  ,   33  ,   97  ,   171 ,   443 ,   75  ,   1681    ,   407 ,   7669    ,   14893   ,0 };
            const unsigned long dim1436Kuo3Init[]   =   {   1   ,   1   ,   3   ,   13  ,   17  ,   31  ,   53  ,   113 ,   97  ,   653 ,   917 ,   11  ,   5697    ,   8083    ,0 };
            const unsigned long dim1437Kuo3Init[]   =   {   1   ,   1   ,   3   ,   9   ,   19  ,   27  ,   1   ,   155 ,   89  ,   579 ,   951 ,   1779    ,   343 ,   12053   ,0 };
            const unsigned long dim1438Kuo3Init[]   =   {   1   ,   3   ,   7   ,   7   ,   23  ,   29  ,   33  ,   143 ,   181 ,   25  ,   109 ,   3407    ,   3255    ,   15363   ,0 };
            const unsigned long dim1439Kuo3Init[]   =   {   1   ,   3   ,   3   ,   7   ,   5   ,   47  ,   119 ,   111 ,   315 ,   225 ,   1655    ,   3019    ,   4835    ,   11641   ,0 };
            const unsigned long dim1440Kuo3Init[]   =   {   1   ,   1   ,   7   ,   1   ,   19  ,   51  ,   103 ,   219 ,   233 ,   633 ,   1923    ,   1523    ,   467 ,   15363   ,0 };
            const unsigned long dim1441Kuo3Init[]   =   {   1   ,   1   ,   3   ,   3   ,   11  ,   5   ,   125 ,   143 ,   277 ,   5   ,   1597    ,   2445    ,   4261    ,   12449   ,0 };
            const unsigned long dim1442Kuo3Init[]   =   {   1   ,   3   ,   1   ,   1   ,   7   ,   1   ,   103 ,   63  ,   23  ,   65  ,   1449    ,   3097    ,   601 ,   12563   ,0 };
            const unsigned long dim1443Kuo3Init[]   =   {   1   ,   1   ,   7   ,   13  ,   31  ,   51  ,   23  ,   251 ,   193 ,   227 ,   1569    ,   439 ,   1037    ,   8013    ,0 };
            const unsigned long dim1444Kuo3Init[]   =   {   1   ,   3   ,   5   ,   5   ,   17  ,   9   ,   45  ,   241 ,   197 ,   787 ,   555 ,   1951    ,   2697    ,   725 ,0 };
            const unsigned long dim1445Kuo3Init[]   =   {   1   ,   3   ,   5   ,   15  ,   9   ,   43  ,   1   ,   211 ,   51  ,   645 ,   1977    ,   3085    ,   4003    ,   10331   ,0 };
            const unsigned long dim1446Kuo3Init[]   =   {   1   ,   1   ,   3   ,   11  ,   7   ,   31  ,   89  ,   209 ,   505 ,   753 ,   641 ,   3177    ,   2791    ,   10727   ,0 };
            const unsigned long dim1447Kuo3Init[]   =   {   1   ,   3   ,   1   ,   1   ,   31  ,   13  ,   53  ,   223 ,   435 ,   51  ,   571 ,   2033    ,   1689    ,   67  ,0 };
            const unsigned long dim1448Kuo3Init[]   =   {   1   ,   3   ,   1   ,   3   ,   21  ,   55  ,   19  ,   31  ,   273 ,   175 ,   1657    ,   3399    ,   3031    ,   10271   ,0 };
            const unsigned long dim1449Kuo3Init[]   =   {   1   ,   1   ,   7   ,   9   ,   27  ,   51  ,   87  ,   157 ,   395 ,   57  ,   1501    ,   2547    ,   5061    ,   11325   ,0 };
            const unsigned long dim1450Kuo3Init[]   =   {   1   ,   1   ,   1   ,   9   ,   19  ,   43  ,   55  ,   147 ,   193 ,   531 ,   511 ,   4019    ,   2317    ,   3727    ,0 };
            const unsigned long dim1451Kuo3Init[]   =   {   1   ,   1   ,   1   ,   15  ,   1   ,   15  ,   5   ,   157 ,   335 ,   183 ,   619 ,   2483    ,   5817    ,   3247    ,0 };
            const unsigned long dim1452Kuo3Init[]   =   {   1   ,   1   ,   1   ,   3   ,   29  ,   59  ,   27  ,   27  ,   355 ,   315 ,   193 ,   707 ,   5897    ,   4623    ,0 };
            const unsigned long dim1453Kuo3Init[]   =   {   1   ,   1   ,   7   ,   1   ,   23  ,   29  ,   33  ,   123 ,   413 ,   317 ,   1099    ,   1633    ,   1619    ,   9141    ,0 };
            const unsigned long dim1454Kuo3Init[]   =   {   1   ,   1   ,   5   ,   11  ,   11  ,   49  ,   85  ,   95  ,   73  ,   829 ,   929 ,   2013    ,   4653    ,   3045    ,0 };
            const unsigned long dim1455Kuo3Init[]   =   {   1   ,   1   ,   1   ,   13  ,   5   ,   63  ,   81  ,   177 ,   23  ,   455 ,   1527    ,   2271    ,   2523    ,   3711    ,0 };
            const unsigned long dim1456Kuo3Init[]   =   {   1   ,   3   ,   3   ,   13  ,   9   ,   13  ,   1   ,   241 ,   225 ,   191 ,   933 ,   657 ,   4557    ,   6903    ,0 };
            const unsigned long dim1457Kuo3Init[]   =   {   1   ,   3   ,   3   ,   15  ,   19  ,   29  ,   121 ,   3   ,   235 ,   729 ,   1705    ,   4035    ,   5429    ,   15339   ,0 };
            const unsigned long dim1458Kuo3Init[]   =   {   1   ,   1   ,   7   ,   5   ,   3   ,   63  ,   79  ,   89  ,   363 ,   407 ,   325 ,   3045    ,   7761    ,   8855    ,0 };
            const unsigned long dim1459Kuo3Init[]   =   {   1   ,   3   ,   3   ,   1   ,   13  ,   39  ,   69  ,   119 ,   45  ,   259 ,   577 ,   269 ,   2977    ,   2093    ,0 };
            const unsigned long dim1460Kuo3Init[]   =   {   1   ,   3   ,   3   ,   3   ,   5   ,   9   ,   115 ,   161 ,   319 ,   799 ,   777 ,   459 ,   1823    ,   10971   ,0 };
            const unsigned long dim1461Kuo3Init[]   =   {   1   ,   3   ,   7   ,   9   ,   1   ,   45  ,   31  ,   223 ,   421 ,   389 ,   1535    ,   1701    ,   3739    ,   9301    ,0 };
            const unsigned long dim1462Kuo3Init[]   =   {   1   ,   3   ,   5   ,   5   ,   5   ,   49  ,   53  ,   189 ,   101 ,   475 ,   453 ,   1921    ,   7161    ,   4529    ,0 };
            const unsigned long dim1463Kuo3Init[]   =   {   1   ,   3   ,   5   ,   11  ,   29  ,   37  ,   53  ,   213 ,   371 ,   985 ,   647 ,   765 ,   6303    ,   10947   ,0 };
            const unsigned long dim1464Kuo3Init[]   =   {   1   ,   3   ,   3   ,   15  ,   5   ,   23  ,   51  ,   199 ,   463 ,   11  ,   709 ,   2083    ,   6337    ,   1115    ,0 };
            const unsigned long dim1465Kuo3Init[]   =   {   1   ,   1   ,   1   ,   1   ,   17  ,   57  ,   1   ,   91  ,   159 ,   1013    ,   1229    ,   2147    ,   571 ,   335 ,0 };
            const unsigned long dim1466Kuo3Init[]   =   {   1   ,   3   ,   1   ,   11  ,   21  ,   53  ,   119 ,   221 ,   203 ,   467 ,   845 ,   645 ,   7939    ,   7257    ,0 };
            const unsigned long dim1467Kuo3Init[]   =   {   1   ,   1   ,   5   ,   1   ,   9   ,   45  ,   87  ,   213 ,   483 ,   247 ,   1423    ,   1855    ,   583 ,   13339   ,0 };
            const unsigned long dim1468Kuo3Init[]   =   {   1   ,   1   ,   3   ,   7   ,   13  ,   39  ,   89  ,   87  ,   467 ,   107 ,   185 ,   1507    ,   755 ,   12069   ,0 };
            const unsigned long dim1469Kuo3Init[]   =   {   1   ,   1   ,   5   ,   5   ,   5   ,   49  ,   17  ,   61  ,   483 ,   877 ,   1887    ,   2973    ,   6733    ,   3241    ,0 };
            const unsigned long dim1470Kuo3Init[]   =   {   1   ,   1   ,   7   ,   15  ,   31  ,   25  ,   105 ,   215 ,   449 ,   723 ,   973 ,   4011    ,   1079    ,   9713    ,0 };
            const unsigned long dim1471Kuo3Init[]   =   {   1   ,   3   ,   3   ,   5   ,   23  ,   61  ,   9   ,   85  ,   323 ,   675 ,   101 ,   2451    ,   5441    ,   14371   ,0 };
            const unsigned long dim1472Kuo3Init[]   =   {   1   ,   1   ,   5   ,   9   ,   1   ,   47  ,   109 ,   177 ,   279 ,   675 ,   867 ,   2053    ,   6975    ,   4829    ,0 };
            const unsigned long dim1473Kuo3Init[]   =   {   1   ,   3   ,   5   ,   15  ,   23  ,   15  ,   17  ,   49  ,   273 ,   839 ,   779 ,   3899    ,   3775    ,   4289    ,0 };
            const unsigned long dim1474Kuo3Init[]   =   {   1   ,   3   ,   3   ,   13  ,   3   ,   5   ,   9   ,   83  ,   291 ,   727 ,   2041    ,   2221    ,   979 ,   761 ,0 };
            const unsigned long dim1475Kuo3Init[]   =   {   1   ,   1   ,   5   ,   5   ,   7   ,   5   ,   23  ,   187 ,   293 ,   159 ,   847 ,   697 ,   7227    ,   9549    ,0 };
            const unsigned long dim1476Kuo3Init[]   =   {   1   ,   3   ,   5   ,   7   ,   19  ,   7   ,   77  ,   57  ,   225 ,   443 ,   539 ,   1531    ,   857 ,   12997   ,0 };
            const unsigned long dim1477Kuo3Init[]   =   {   1   ,   3   ,   3   ,   9   ,   9   ,   33  ,   79  ,   137 ,   15  ,   927 ,   1621    ,   1241    ,   5741    ,   13507   ,0 };
            const unsigned long dim1478Kuo3Init[]   =   {   1   ,   3   ,   7   ,   7   ,   11  ,   37  ,   87  ,   255 ,   293 ,   649 ,   393 ,   2549    ,   8083    ,   13417   ,0 };
            const unsigned long dim1479Kuo3Init[]   =   {   1   ,   1   ,   3   ,   5   ,   19  ,   31  ,   47  ,   243 ,   423 ,   663 ,   359 ,   3293    ,   7915    ,   16351   ,0 };
            const unsigned long dim1480Kuo3Init[]   =   {   1   ,   1   ,   1   ,   11  ,   19  ,   63  ,   87  ,   245 ,   445 ,   975 ,   1047    ,   935 ,   4221    ,   851 ,0 };
            const unsigned long dim1481Kuo3Init[]   =   {   1   ,   3   ,   5   ,   1   ,   19  ,   43  ,   11  ,   85  ,   257 ,   875 ,   369 ,   1075    ,   1343    ,   8487    ,0 };
            const unsigned long dim1482Kuo3Init[]   =   {   1   ,   1   ,   7   ,   13  ,   29  ,   35  ,   37  ,   201 ,   473 ,   853 ,   185 ,   3157    ,   6135    ,   7631    ,0 };
            const unsigned long dim1483Kuo3Init[]   =   {   1   ,   1   ,   1   ,   13  ,   13  ,   53  ,   43  ,   55  ,   65  ,   999 ,   729 ,   857 ,   3645    ,   9793    ,0 };
            const unsigned long dim1484Kuo3Init[]   =   {   1   ,   1   ,   3   ,   13  ,   21  ,   47  ,   85  ,   19  ,   1   ,   333 ,   989 ,   327 ,   4693    ,   3521    ,0 };
            const unsigned long dim1485Kuo3Init[]   =   {   1   ,   3   ,   1   ,   13  ,   11  ,   39  ,   101 ,   209 ,   51  ,   53  ,   631 ,   1439    ,   5909    ,   9599    ,0 };
            const unsigned long dim1486Kuo3Init[]   =   {   1   ,   1   ,   7   ,   15  ,   29  ,   57  ,   21  ,   181 ,   293 ,   807 ,   1299    ,   1001    ,   4265    ,   1811    ,0 };
            const unsigned long dim1487Kuo3Init[]   =   {   1   ,   1   ,   5   ,   13  ,   15  ,   1   ,   17  ,   63  ,   213 ,   85  ,   1025    ,   3955    ,   2371    ,   9065    ,0 };
            const unsigned long dim1488Kuo3Init[]   =   {   1   ,   1   ,   5   ,   15  ,   13  ,   57  ,   23  ,   169 ,   93  ,   501 ,   7   ,   2481    ,   123 ,   8985    ,0 };
            const unsigned long dim1489Kuo3Init[]   =   {   1   ,   3   ,   3   ,   3   ,   11  ,   61  ,   19  ,   133 ,   175 ,   833 ,   1753    ,   3871    ,   8169    ,   2771    ,0 };
            const unsigned long dim1490Kuo3Init[]   =   {   1   ,   3   ,   5   ,   1   ,   11  ,   13  ,   57  ,   55  ,   385 ,   463 ,   785 ,   2339    ,   7117    ,   12519   ,0 };
            const unsigned long dim1491Kuo3Init[]   =   {   1   ,   1   ,   1   ,   13  ,   15  ,   35  ,   27  ,   179 ,   375 ,   477 ,   1157    ,   2067    ,   6971    ,   8269    ,0 };
            const unsigned long dim1492Kuo3Init[]   =   {   1   ,   1   ,   3   ,   11  ,   27  ,   21  ,   57  ,   125 ,   219 ,   433 ,   67  ,   307 ,   7327    ,   2915    ,0 };
            const unsigned long dim1493Kuo3Init[]   =   {   1   ,   3   ,   5   ,   7   ,   21  ,   7   ,   111 ,   21  ,   85  ,   295 ,   1751    ,   821 ,   1175    ,   3109    ,0 };
            const unsigned long dim1494Kuo3Init[]   =   {   1   ,   1   ,   7   ,   1   ,   7   ,   51  ,   43  ,   119 ,   89  ,   751 ,   835 ,   3969    ,   861 ,   5775    ,0 };
            const unsigned long dim1495Kuo3Init[]   =   {   1   ,   3   ,   1   ,   13  ,   5   ,   37  ,   93  ,   185 ,   435 ,   283 ,   1221    ,   2965    ,   2845    ,   6025    ,0 };
            const unsigned long dim1496Kuo3Init[]   =   {   1   ,   1   ,   7   ,   7   ,   31  ,   27  ,   99  ,   143 ,   507 ,   461 ,   1999    ,   3743    ,   4277    ,   13729   ,0 };
            const unsigned long dim1497Kuo3Init[]   =   {   1   ,   3   ,   1   ,   13  ,   15  ,   43  ,   99  ,   121 ,   245 ,   73  ,   829 ,   3887    ,   3093    ,   16183   ,0 };
            const unsigned long dim1498Kuo3Init[]   =   {   1   ,   3   ,   3   ,   7   ,   15  ,   35  ,   29  ,   79  ,   457 ,   695 ,   1469    ,   1505    ,   3891    ,   14677   ,0 };
            const unsigned long dim1499Kuo3Init[]   =   {   1   ,   3   ,   7   ,   7   ,   5   ,   59  ,   71  ,   179 ,   305 ,   9   ,   2043    ,   245 ,   1291    ,   14409   ,0 };
            const unsigned long dim1500Kuo3Init[]   =   {   1   ,   1   ,   3   ,   1   ,   13  ,   7   ,   17  ,   43  ,   367 ,   163 ,   1649    ,   3671    ,   2663    ,   2277    ,0 };
            const unsigned long dim1501Kuo3Init[]   =   {   1   ,   1   ,   1   ,   5   ,   9   ,   57  ,   115 ,   177 ,   271 ,   735 ,   263 ,   1137    ,   2127    ,   15367   ,0 };
            const unsigned long dim1502Kuo3Init[]   =   {   1   ,   3   ,   5   ,   7   ,   17  ,   11  ,   37  ,   197 ,   459 ,   647 ,   307 ,   3617    ,   3201    ,   10293   ,0 };
            const unsigned long dim1503Kuo3Init[]   =   {   1   ,   1   ,   1   ,   7   ,   31  ,   37  ,   61  ,   59  ,   331 ,   483 ,   1849    ,   1737    ,   2617    ,   5053    ,0 };
            const unsigned long dim1504Kuo3Init[]   =   {   1   ,   3   ,   5   ,   13  ,   31  ,   43  ,   31  ,   185 ,   261 ,   249 ,   1137    ,   1285    ,   2885    ,   4289    ,0 };
            const unsigned long dim1505Kuo3Init[]   =   {   1   ,   1   ,   7   ,   3   ,   21  ,   53  ,   113 ,   163 ,   209 ,   25  ,   329 ,   2963    ,   7849    ,   10933   ,0 };
            const unsigned long dim1506Kuo3Init[]   =   {   1   ,   1   ,   5   ,   1   ,   31  ,   23  ,   101 ,   151 ,   173 ,   193 ,   1267    ,   487 ,   4701    ,   8543    ,0 };
            const unsigned long dim1507Kuo3Init[]   =   {   1   ,   1   ,   1   ,   1   ,   25  ,   33  ,   85  ,   31  ,   433 ,   537 ,   613 ,   1675    ,   6997    ,   11975   ,0 };
            const unsigned long dim1508Kuo3Init[]   =   {   1   ,   3   ,   7   ,   1   ,   7   ,   59  ,   57  ,   249 ,   511 ,   371 ,   121 ,   3435    ,   8029    ,   11785   ,0 };
            const unsigned long dim1509Kuo3Init[]   =   {   1   ,   1   ,   7   ,   3   ,   19  ,   61  ,   105 ,   27  ,   217 ,   47  ,   861 ,   2571    ,   6203    ,   805 ,0 };
            const unsigned long dim1510Kuo3Init[]   =   {   1   ,   1   ,   7   ,   13  ,   27  ,   33  ,   41  ,   217 ,   39  ,   917 ,   355 ,   3083    ,   2803    ,   747 ,0 };
            const unsigned long dim1511Kuo3Init[]   =   {   1   ,   3   ,   7   ,   3   ,   23  ,   33  ,   43  ,   27  ,   151 ,   399 ,   1307    ,   3269    ,   7941    ,   195 ,0 };
            const unsigned long dim1512Kuo3Init[]   =   {   1   ,   1   ,   3   ,   3   ,   3   ,   37  ,   71  ,   111 ,   389 ,   13  ,   1147    ,   3649    ,   1497    ,   10459   ,0 };
            const unsigned long dim1513Kuo3Init[]   =   {   1   ,   1   ,   5   ,   5   ,   25  ,   23  ,   19  ,   251 ,   87  ,   347 ,   711 ,   3379    ,   4677    ,   5651    ,0 };
            const unsigned long dim1514Kuo3Init[]   =   {   1   ,   1   ,   5   ,   3   ,   9   ,   21  ,   51  ,   85  ,   491 ,   67  ,   485 ,   2099    ,   3081    ,   11751   ,0 };
            const unsigned long dim1515Kuo3Init[]   =   {   1   ,   1   ,   3   ,   15  ,   27  ,   37  ,   75  ,   71  ,   41  ,   823 ,   1035    ,   133 ,   7537    ,   129 ,0 };
            const unsigned long dim1516Kuo3Init[]   =   {   1   ,   1   ,   1   ,   7   ,   31  ,   35  ,   75  ,   223 ,   379 ,   395 ,   2011    ,   1403    ,   4923    ,   391 ,0 };
            const unsigned long dim1517Kuo3Init[]   =   {   1   ,   1   ,   5   ,   1   ,   19  ,   37  ,   87  ,   109 ,   291 ,   73  ,   355 ,   1427    ,   741 ,   7393    ,0 };
            const unsigned long dim1518Kuo3Init[]   =   {   1   ,   3   ,   3   ,   11  ,   23  ,   1   ,   109 ,   81  ,   319 ,   1001    ,   1173    ,   2529    ,   5859    ,   3167    ,0 };
            const unsigned long dim1519Kuo3Init[]   =   {   1   ,   1   ,   1   ,   7   ,   17  ,   23  ,   121 ,   223 ,   49  ,   203 ,   1653    ,   1567    ,   7795    ,   1631    ,0 };
            const unsigned long dim1520Kuo3Init[]   =   {   1   ,   3   ,   1   ,   15  ,   15  ,   61  ,   123 ,   141 ,   127 ,   871 ,   897 ,   533 ,   231 ,   13767   ,0 };
            const unsigned long dim1521Kuo3Init[]   =   {   1   ,   3   ,   5   ,   15  ,   5   ,   17  ,   31  ,   5   ,   505 ,   65  ,   587 ,   3721    ,   5755    ,   3865    ,0 };
            const unsigned long dim1522Kuo3Init[]   =   {   1   ,   1   ,   7   ,   11  ,   27  ,   59  ,   17  ,   17  ,   33  ,   399 ,   1963    ,   3135    ,   5807    ,   2427    ,0 };
            const unsigned long dim1523Kuo3Init[]   =   {   1   ,   3   ,   1   ,   3   ,   29  ,   19  ,   77  ,   55  ,   511 ,   569 ,   729 ,   2893    ,   373 ,   643 ,0 };
            const unsigned long dim1524Kuo3Init[]   =   {   1   ,   1   ,   5   ,   13  ,   17  ,   17  ,   101 ,   107 ,   449 ,   581 ,   905 ,   3281    ,   4487    ,   9397    ,0 };
            const unsigned long dim1525Kuo3Init[]   =   {   1   ,   1   ,   7   ,   9   ,   23  ,   3   ,   103 ,   29  ,   291 ,   437 ,   1417    ,   2289    ,   3705    ,   14439   ,0 };
            const unsigned long dim1526Kuo3Init[]   =   {   1   ,   3   ,   1   ,   13  ,   11  ,   61  ,   7   ,   87  ,   471 ,   707 ,   1823    ,   3971    ,   7255    ,   14395   ,0 };
            const unsigned long dim1527Kuo3Init[]   =   {   1   ,   3   ,   7   ,   13  ,   21  ,   25  ,   125 ,   19  ,   221 ,   885 ,   1787    ,   3825    ,   7661    ,   8839    ,0 };
            const unsigned long dim1528Kuo3Init[]   =   {   1   ,   3   ,   5   ,   3   ,   29  ,   39  ,   21  ,   47  ,   475 ,   447 ,   405 ,   3365    ,   5545    ,   11759   ,0 };
            const unsigned long dim1529Kuo3Init[]   =   {   1   ,   1   ,   7   ,   1   ,   13  ,   11  ,   113 ,   123 ,   411 ,   385 ,   533 ,   4091    ,   6451    ,   6021    ,0 };
            const unsigned long dim1530Kuo3Init[]   =   {   1   ,   1   ,   3   ,   11  ,   7   ,   7   ,   25  ,   65  ,   105 ,   409 ,   1837    ,   2609    ,   6417    ,   16111   ,0 };
            const unsigned long dim1531Kuo3Init[]   =   {   1   ,   3   ,   7   ,   15  ,   25  ,   11  ,   101 ,   151 ,   91  ,   305 ,   865 ,   2653    ,   3031    ,   2295    ,0 };
            const unsigned long dim1532Kuo3Init[]   =   {   1   ,   1   ,   1   ,   15  ,   11  ,   53  ,   87  ,   203 ,   275 ,   915 ,   281 ,   687 ,   5857    ,   393 ,0 };
            const unsigned long dim1533Kuo3Init[]   =   {   1   ,   3   ,   3   ,   5   ,   15  ,   11  ,   97  ,   169 ,   175 ,   325 ,   285 ,   2287    ,   7095    ,   5245    ,0 };
            const unsigned long dim1534Kuo3Init[]   =   {   1   ,   1   ,   5   ,   13  ,   25  ,   3   ,   31  ,   79  ,   49  ,   637 ,   1927    ,   2361    ,   2233    ,   7935    ,0 };
            const unsigned long dim1535Kuo3Init[]   =   {   1   ,   1   ,   3   ,   11  ,   17  ,   41  ,   69  ,   9   ,   441 ,   719 ,   765 ,   689 ,   475 ,   8613    ,0 };
            const unsigned long dim1536Kuo3Init[]   =   {   1   ,   1   ,   7   ,   1   ,   9   ,   11  ,   51  ,   175 ,   231 ,   975 ,   1983    ,   1389    ,   205 ,   5051    ,0 };
            const unsigned long dim1537Kuo3Init[]   =   {   1   ,   1   ,   3   ,   7   ,   17  ,   37  ,   9   ,   193 ,   243 ,   331 ,   1545    ,   3905    ,   4601    ,   7731    ,0 };
            const unsigned long dim1538Kuo3Init[]   =   {   1   ,   1   ,   5   ,   9   ,   9   ,   41  ,   27  ,   29  ,   231 ,   775 ,   1753    ,   1915    ,   7247    ,   6821    ,0 };
            const unsigned long dim1539Kuo3Init[]   =   {   1   ,   3   ,   7   ,   13  ,   21  ,   9   ,   51  ,   165 ,   313 ,   951 ,   1897    ,   2123    ,   2999    ,   3171    ,0 };
            const unsigned long dim1540Kuo3Init[]   =   {   1   ,   1   ,   1   ,   11  ,   25  ,   43  ,   81  ,   249 ,   309 ,   523 ,   373 ,   3293    ,   6607    ,   9495    ,0 };
            const unsigned long dim1541Kuo3Init[]   =   {   1   ,   1   ,   3   ,   1   ,   15  ,   63  ,   11  ,   139 ,   319 ,   665 ,   1703    ,   1669    ,   4817    ,   3645    ,0 };
            const unsigned long dim1542Kuo3Init[]   =   {   1   ,   1   ,   7   ,   1   ,   11  ,   47  ,   117 ,   245 ,   251 ,   341 ,   415 ,   815 ,   7335    ,   14297   ,0 };
            const unsigned long dim1543Kuo3Init[]   =   {   1   ,   1   ,   5   ,   13  ,   19  ,   33  ,   109 ,   191 ,   253 ,   867 ,   1505    ,   2629    ,   6619    ,   7913    ,0 };
            const unsigned long dim1544Kuo3Init[]   =   {   1   ,   3   ,   5   ,   1   ,   7   ,   49  ,   53  ,   51  ,   285 ,   25  ,   1783    ,   2201    ,   7603    ,   11365   ,0 };
            const unsigned long dim1545Kuo3Init[]   =   {   1   ,   1   ,   3   ,   11  ,   1   ,   49  ,   9   ,   89  ,   365 ,   909 ,   1563    ,   2719    ,   6859    ,   14135   ,0 };
            const unsigned long dim1546Kuo3Init[]   =   {   1   ,   1   ,   1   ,   11  ,   17  ,   57  ,   21  ,   165 ,   419 ,   443 ,   1207    ,   3821    ,   1151    ,   14037   ,0 };
            const unsigned long dim1547Kuo3Init[]   =   {   1   ,   1   ,   1   ,   3   ,   1   ,   31  ,   97  ,   133 ,   423 ,   993 ,   745 ,   3559    ,   7725    ,   2069    ,0 };
            const unsigned long dim1548Kuo3Init[]   =   {   1   ,   1   ,   1   ,   7   ,   31  ,   1   ,   65  ,   3   ,   63  ,   255 ,   313 ,   515 ,   455 ,   4957    ,0 };
            const unsigned long dim1549Kuo3Init[]   =   {   1   ,   1   ,   1   ,   5   ,   5   ,   31  ,   37  ,   121 ,   201 ,   493 ,   55  ,   3735    ,   5019    ,   5665    ,0 };
            const unsigned long dim1550Kuo3Init[]   =   {   1   ,   3   ,   3   ,   9   ,   9   ,   57  ,   55  ,   155 ,   171 ,   687 ,   1415    ,   2017    ,   7021    ,   7409    ,0 };
            const unsigned long dim1551Kuo3Init[]   =   {   1   ,   3   ,   5   ,   3   ,   15  ,   23  ,   83  ,   79  ,   367 ,   485 ,   1933    ,   315 ,   2591    ,   7609    ,0 };
            const unsigned long dim1552Kuo3Init[]   =   {   1   ,   3   ,   7   ,   3   ,   25  ,   49  ,   79  ,   15  ,   435 ,   809 ,   265 ,   2215    ,   7207    ,   563 ,0 };
            const unsigned long dim1553Kuo3Init[]   =   {   1   ,   3   ,   7   ,   1   ,   19  ,   61  ,   103 ,   101 ,   51  ,   637 ,   639 ,   55  ,   4323    ,   179 ,0 };
            const unsigned long dim1554Kuo3Init[]   =   {   1   ,   3   ,   1   ,   3   ,   23  ,   27  ,   33  ,   79  ,   215 ,   849 ,   331 ,   2881    ,   2109    ,   10309   ,0 };
            const unsigned long dim1555Kuo3Init[]   =   {   1   ,   3   ,   5   ,   7   ,   21  ,   63  ,   113 ,   237 ,   63  ,   769 ,   63  ,   391 ,   1433    ,   6913    ,0 };
            const unsigned long dim1556Kuo3Init[]   =   {   1   ,   3   ,   3   ,   15  ,   3   ,   53  ,   25  ,   187 ,   313 ,   93  ,   1499    ,   3281    ,   557 ,   12441   ,0 };
            const unsigned long dim1557Kuo3Init[]   =   {   1   ,   1   ,   1   ,   5   ,   7   ,   27  ,   97  ,   211 ,   5   ,   31  ,   497 ,   1659    ,   7061    ,   807 ,0 };
            const unsigned long dim1558Kuo3Init[]   =   {   1   ,   3   ,   1   ,   9   ,   23  ,   3   ,   39  ,   55  ,   369 ,   197 ,   299 ,   2529    ,   2703    ,   247 ,0 };
            const unsigned long dim1559Kuo3Init[]   =   {   1   ,   3   ,   3   ,   3   ,   15  ,   49  ,   119 ,   241 ,   45  ,   303 ,   1741    ,   1077    ,   6735    ,   8079    ,0 };
            const unsigned long dim1560Kuo3Init[]   =   {   1   ,   3   ,   5   ,   5   ,   21  ,   55  ,   91  ,   17  ,   341 ,   123 ,   1389    ,   2557    ,   3667    ,   4891    ,0 };
            const unsigned long dim1561Kuo3Init[]   =   {   1   ,   1   ,   7   ,   5   ,   27  ,   21  ,   107 ,   111 ,   85  ,   39  ,   863 ,   1965    ,   5173    ,   49  ,0 };
            const unsigned long dim1562Kuo3Init[]   =   {   1   ,   1   ,   1   ,   5   ,   11  ,   5   ,   15  ,   185 ,   269 ,   711 ,   557 ,   2021    ,   7481    ,   4977    ,0 };
            const unsigned long dim1563Kuo3Init[]   =   {   1   ,   1   ,   5   ,   9   ,   5   ,   1   ,   93  ,   139 ,   323 ,   915 ,   549 ,   1345    ,   6031    ,   3979    ,0 };
            const unsigned long dim1564Kuo3Init[]   =   {   1   ,   3   ,   3   ,   7   ,   19  ,   9   ,   57  ,   177 ,   135 ,   65  ,   27  ,   795 ,   4637    ,   11977   ,0 };
            const unsigned long dim1565Kuo3Init[]   =   {   1   ,   3   ,   5   ,   13  ,   7   ,   17  ,   117 ,   35  ,   257 ,   671 ,   1741    ,   1993    ,   1887    ,   8149    ,0 };
            const unsigned long dim1566Kuo3Init[]   =   {   1   ,   1   ,   7   ,   13  ,   3   ,   45  ,   5   ,   199 ,   29  ,   227 ,   1613    ,   835 ,   2451    ,   13079   ,0 };
            const unsigned long dim1567Kuo3Init[]   =   {   1   ,   1   ,   5   ,   15  ,   23  ,   45  ,   87  ,   15  ,   349 ,   843 ,   159 ,   3283    ,   1017    ,   15813   ,0 };
            const unsigned long dim1568Kuo3Init[]   =   {   1   ,   3   ,   3   ,   3   ,   21  ,   21  ,   31  ,   51  ,   481 ,   49  ,   1561    ,   2171    ,   6025    ,   9955    ,0 };
            const unsigned long dim1569Kuo3Init[]   =   {   1   ,   1   ,   1   ,   3   ,   7   ,   51  ,   97  ,   195 ,   5   ,   251 ,   1509    ,   1479    ,   1165    ,   16375   ,0 };
            const unsigned long dim1570Kuo3Init[]   =   {   1   ,   3   ,   5   ,   7   ,   7   ,   17  ,   43  ,   101 ,   281 ,   807 ,   2003    ,   2821    ,   6877    ,   2457    ,0 };
            const unsigned long dim1571Kuo3Init[]   =   {   1   ,   1   ,   1   ,   1   ,   15  ,   23  ,   59  ,   91  ,   19  ,   933 ,   145 ,   673 ,   4729    ,   10369   ,0 };
            const unsigned long dim1572Kuo3Init[]   =   {   1   ,   3   ,   5   ,   13  ,   31  ,   37  ,   9   ,   219 ,   343 ,   907 ,   1401    ,   213 ,   5149    ,   14171   ,0 };
            const unsigned long dim1573Kuo3Init[]   =   {   1   ,   3   ,   1   ,   3   ,   17  ,   17  ,   61  ,   181 ,   281 ,   707 ,   113 ,   3595    ,   67  ,   15423   ,0 };
            const unsigned long dim1574Kuo3Init[]   =   {   1   ,   1   ,   5   ,   9   ,   3   ,   51  ,   69  ,   93  ,   375 ,   451 ,   1181    ,   1353    ,   939 ,   629 ,0 };
            const unsigned long dim1575Kuo3Init[]   =   {   1   ,   3   ,   5   ,   11  ,   13  ,   33  ,   113 ,   85  ,   397 ,   823 ,   1513    ,   283 ,   7167    ,   4627    ,0 };
            const unsigned long dim1576Kuo3Init[]   =   {   1   ,   3   ,   7   ,   9   ,   25  ,   31  ,   5   ,   17  ,   377 ,   429 ,   1841    ,   3915    ,   6023    ,   11791   ,0 };
            const unsigned long dim1577Kuo3Init[]   =   {   1   ,   1   ,   1   ,   7   ,   15  ,   35  ,   45  ,   31  ,   427 ,   509 ,   1827    ,   1171    ,   7001    ,   6419    ,0 };
            const unsigned long dim1578Kuo3Init[]   =   {   1   ,   1   ,   5   ,   5   ,   7   ,   37  ,   41  ,   35  ,   349 ,   73  ,   171 ,   35  ,   2333    ,   659 ,0 };
            const unsigned long dim1579Kuo3Init[]   =   {   1   ,   1   ,   5   ,   15  ,   13  ,   61  ,   53  ,   87  ,   463 ,   143 ,   257 ,   3151    ,   6149    ,   10155   ,0 };
            const unsigned long dim1580Kuo3Init[]   =   {   1   ,   1   ,   3   ,   9   ,   3   ,   17  ,   29  ,   151 ,   259 ,   881 ,   2003    ,   1057    ,   1867    ,   7773    ,0 };
            const unsigned long dim1581Kuo3Init[]   =   {   1   ,   3   ,   3   ,   1   ,   5   ,   55  ,   97  ,   49  ,   473 ,   795 ,   1545    ,   1169    ,   187 ,   10475   ,0 };
            const unsigned long dim1582Kuo3Init[]   =   {   1   ,   1   ,   3   ,   7   ,   9   ,   11  ,   61  ,   205 ,   71  ,   183 ,   1551    ,   621 ,   5715    ,   3401    ,0 };
            const unsigned long dim1583Kuo3Init[]   =   {   1   ,   3   ,   7   ,   13  ,   25  ,   41  ,   109 ,   211 ,   189 ,   859 ,   1681    ,   3281    ,   8153    ,   13303   ,0 };
            const unsigned long dim1584Kuo3Init[]   =   {   1   ,   1   ,   5   ,   3   ,   13  ,   17  ,   81  ,   151 ,   189 ,   419 ,   1707    ,   235 ,   355 ,   8295    ,0 };
            const unsigned long dim1585Kuo3Init[]   =   {   1   ,   3   ,   3   ,   13  ,   19  ,   33  ,   49  ,   249 ,   347 ,   617 ,   2037    ,   537 ,   755 ,   5131    ,0 };
            const unsigned long dim1586Kuo3Init[]   =   {   1   ,   1   ,   1   ,   13  ,   23  ,   17  ,   93  ,   35  ,   357 ,   523 ,   355 ,   1159    ,   1329    ,   7663    ,0 };
            const unsigned long dim1587Kuo3Init[]   =   {   1   ,   1   ,   5   ,   15  ,   15  ,   17  ,   81  ,   197 ,   479 ,   405 ,   55  ,   4095    ,   4931    ,   7663    ,0 };
            const unsigned long dim1588Kuo3Init[]   =   {   1   ,   3   ,   5   ,   13  ,   1   ,   1   ,   17  ,   173 ,   259 ,   191 ,   1929    ,   2063    ,   6623    ,   5065    ,0 };
            const unsigned long dim1589Kuo3Init[]   =   {   1   ,   3   ,   5   ,   15  ,   31  ,   21  ,   25  ,   33  ,   233 ,   663 ,   1687    ,   495 ,   5223    ,   439 ,0 };
            const unsigned long dim1590Kuo3Init[]   =   {   1   ,   3   ,   3   ,   5   ,   31  ,   47  ,   43  ,   119 ,   491 ,   893 ,   1513    ,   2453    ,   6131    ,   8675    ,0 };
            const unsigned long dim1591Kuo3Init[]   =   {   1   ,   3   ,   5   ,   1   ,   21  ,   51  ,   21  ,   27  ,   35  ,   613 ,   849 ,   3679    ,   4089    ,   9671    ,0 };
            const unsigned long dim1592Kuo3Init[]   =   {   1   ,   1   ,   3   ,   5   ,   29  ,   53  ,   125 ,   223 ,   415 ,   315 ,   811 ,   513 ,   4833    ,   10667   ,0 };
            const unsigned long dim1593Kuo3Init[]   =   {   1   ,   1   ,   7   ,   11  ,   5   ,   49  ,   87  ,   47  ,   85  ,   115 ,   757 ,   567 ,   3655    ,   4009    ,0 };
            const unsigned long dim1594Kuo3Init[]   =   {   1   ,   3   ,   1   ,   3   ,   9   ,   1   ,   65  ,   179 ,   55  ,   545 ,   467 ,   1615    ,   4349    ,   14745   ,0 };
            const unsigned long dim1595Kuo3Init[]   =   {   1   ,   3   ,   3   ,   13  ,   7   ,   49  ,   91  ,   133 ,   151 ,   113 ,   1373    ,   203 ,   2711    ,   8045    ,0 };
            const unsigned long dim1596Kuo3Init[]   =   {   1   ,   1   ,   3   ,   13  ,   13  ,   5   ,   53  ,   181 ,   397 ,   309 ,   1653    ,   2987    ,   175 ,   13419   ,0 };
            const unsigned long dim1597Kuo3Init[]   =   {   1   ,   1   ,   7   ,   13  ,   19  ,   35  ,   17  ,   177 ,   77  ,   1011    ,   1667    ,   3083    ,   7977    ,   14925   ,0 };
            const unsigned long dim1598Kuo3Init[]   =   {   1   ,   1   ,   3   ,   15  ,   25  ,   33  ,   59  ,   21  ,   345 ,   395 ,   1575    ,   2173    ,   159 ,   11799   ,0 };
            const unsigned long dim1599Kuo3Init[]   =   {   1   ,   3   ,   1   ,   13  ,   3   ,   19  ,   87  ,   135 ,   227 ,   995 ,   1965    ,   3469    ,   3801    ,   1979    ,0 };
            const unsigned long dim1600Kuo3Init[]   =   {   1   ,   1   ,   1   ,   15  ,   11  ,   1   ,   51  ,   235 ,   37  ,   695 ,   1761    ,   275 ,   5793    ,   4897    ,0 };
            const unsigned long dim1601Kuo3Init[]   =   {   1   ,   1   ,   7   ,   5   ,   13  ,   37  ,   123 ,   23  ,   363 ,   377 ,   2009    ,   2897    ,   207 ,   15103   ,0 };
            const unsigned long dim1602Kuo3Init[]   =   {   1   ,   1   ,   1   ,   15  ,   3   ,   43  ,   49  ,   23  ,   351 ,   429 ,   1181    ,   2003    ,   1397    ,   9489    ,0 };
            const unsigned long dim1603Kuo3Init[]   =   {   1   ,   3   ,   7   ,   7   ,   23  ,   61  ,   105 ,   149 ,   183 ,   83  ,   69  ,   121 ,   6719    ,   4729    ,0 };
            const unsigned long dim1604Kuo3Init[]   =   {   1   ,   1   ,   5   ,   11  ,   27  ,   9   ,   119 ,   59  ,   115 ,   207 ,   1865    ,   3953    ,   45  ,   9739    ,0 };
            const unsigned long dim1605Kuo3Init[]   =   {   1   ,   1   ,   5   ,   5   ,   7   ,   29  ,   81  ,   75  ,   361 ,   589 ,   1689    ,   49  ,   7125    ,   10563   ,0 };
            const unsigned long dim1606Kuo3Init[]   =   {   1   ,   1   ,   7   ,   7   ,   11  ,   21  ,   43  ,   31  ,   43  ,   991 ,   357 ,   3707    ,   7679    ,   4305    ,0 };
            const unsigned long dim1607Kuo3Init[]   =   {   1   ,   1   ,   1   ,   3   ,   23  ,   41  ,   33  ,   167 ,   207 ,   1015    ,   517 ,   1039    ,   4831    ,   753 ,0 };
            const unsigned long dim1608Kuo3Init[]   =   {   1   ,   3   ,   7   ,   9   ,   1   ,   29  ,   25  ,   227 ,   5   ,   935 ,   1201    ,   2781    ,   6947    ,   1597    ,0 };
            const unsigned long dim1609Kuo3Init[]   =   {   1   ,   1   ,   1   ,   11  ,   13  ,   27  ,   79  ,   175 ,   25  ,   681 ,   1805    ,   1435    ,   7325    ,   10439   ,0 };
            const unsigned long dim1610Kuo3Init[]   =   {   1   ,   3   ,   5   ,   15  ,   25  ,   21  ,   105 ,   175 ,   501 ,   43  ,   19  ,   965 ,   255 ,   4177    ,0 };
            const unsigned long dim1611Kuo3Init[]   =   {   1   ,   3   ,   7   ,   9   ,   1   ,   27  ,   115 ,   105 ,   43  ,   181 ,   181 ,   3747    ,   3827    ,   13451   ,0 };
            const unsigned long dim1612Kuo3Init[]   =   {   1   ,   1   ,   5   ,   1   ,   15  ,   27  ,   49  ,   57  ,   361 ,   911 ,   1735    ,   3815    ,   5935    ,   11793   ,0 };
            const unsigned long dim1613Kuo3Init[]   =   {   1   ,   3   ,   7   ,   11  ,   23  ,   7   ,   103 ,   83  ,   343 ,   59  ,   1383    ,   945 ,   4839    ,   1475    ,0 };
            const unsigned long dim1614Kuo3Init[]   =   {   1   ,   1   ,   3   ,   9   ,   13  ,   17  ,   15  ,   129 ,   449 ,   645 ,   719 ,   283 ,   7911    ,   14961   ,0 };
            const unsigned long dim1615Kuo3Init[]   =   {   1   ,   3   ,   1   ,   13  ,   1   ,   35  ,   27  ,   15  ,   177 ,   983 ,   1537    ,   2439    ,   7459    ,   7107    ,0 };
            const unsigned long dim1616Kuo3Init[]   =   {   1   ,   1   ,   7   ,   3   ,   25  ,   45  ,   3   ,   163 ,   149 ,   285 ,   917 ,   2233    ,   7463    ,   15869   ,0 };
            const unsigned long dim1617Kuo3Init[]   =   {   1   ,   1   ,   5   ,   9   ,   3   ,   3   ,   45  ,   175 ,   367 ,   261 ,   723 ,   3273    ,   5611    ,   3323    ,0 };
            const unsigned long dim1618Kuo3Init[]   =   {   1   ,   1   ,   5   ,   15  ,   25  ,   25  ,   9   ,   189 ,   437 ,   807 ,   801 ,   3655    ,   5759    ,   9099    ,0 };
            const unsigned long dim1619Kuo3Init[]   =   {   1   ,   3   ,   5   ,   3   ,   27  ,   37  ,   37  ,   177 ,   313 ,   179 ,   1897    ,   1299    ,   8041    ,   10523   ,0 };
            const unsigned long dim1620Kuo3Init[]   =   {   1   ,   1   ,   3   ,   13  ,   7   ,   61  ,   105 ,   125 ,   343 ,   201 ,   1089    ,   43  ,   5645    ,   7547    ,0 };
            const unsigned long dim1621Kuo3Init[]   =   {   1   ,   1   ,   3   ,   9   ,   17  ,   9   ,   69  ,   199 ,   333 ,   791 ,   1611    ,   1273    ,   7193    ,   6575    ,0 };
            const unsigned long dim1622Kuo3Init[]   =   {   1   ,   3   ,   1   ,   7   ,   7   ,   63  ,   79  ,   225 ,   89  ,   391 ,   503 ,   3199    ,   1409    ,   4669    ,0 };
            const unsigned long dim1623Kuo3Init[]   =   {   1   ,   1   ,   1   ,   3   ,   29  ,   7   ,   3   ,   119 ,   361 ,   269 ,   205 ,   1253    ,   113 ,   9665    ,0 };
            const unsigned long dim1624Kuo3Init[]   =   {   1   ,   3   ,   7   ,   13  ,   13  ,   59  ,   53  ,   75  ,   467 ,   317 ,   815 ,   3401    ,   7223    ,   8237    ,0 };
            const unsigned long dim1625Kuo3Init[]   =   {   1   ,   1   ,   7   ,   1   ,   7   ,   15  ,   49  ,   191 ,   333 ,   1017    ,   399 ,   55  ,   931 ,   14979   ,0 };
            const unsigned long dim1626Kuo3Init[]   =   {   1   ,   3   ,   1   ,   11  ,   23  ,   5   ,   113 ,   47  ,   105 ,   341 ,   615 ,   81  ,   7393    ,   1037    ,0 };
            const unsigned long dim1627Kuo3Init[]   =   {   1   ,   3   ,   5   ,   11  ,   3   ,   29  ,   95  ,   3   ,   275 ,   257 ,   613 ,   1627    ,   695 ,   5929    ,0 };
            const unsigned long dim1628Kuo3Init[]   =   {   1   ,   3   ,   5   ,   5   ,   11  ,   31  ,   57  ,   83  ,   77  ,   573 ,   969 ,   1237    ,   7449    ,   13959   ,0 };
            const unsigned long dim1629Kuo3Init[]   =   {   1   ,   3   ,   3   ,   5   ,   31  ,   47  ,   51  ,   37  ,   123 ,   817 ,   1293    ,   771 ,   4241    ,   3007    ,0 };
            const unsigned long dim1630Kuo3Init[]   =   {   1   ,   1   ,   1   ,   13  ,   9   ,   13  ,   23  ,   111 ,   311 ,   213 ,   1217    ,   2187    ,   5709    ,   6891    ,0 };
            const unsigned long dim1631Kuo3Init[]   =   {   1   ,   1   ,   3   ,   1   ,   7   ,   37  ,   29  ,   67  ,   109 ,   747 ,   723 ,   2631    ,   649 ,   13007   ,0 };
            const unsigned long dim1632Kuo3Init[]   =   {   1   ,   3   ,   5   ,   3   ,   5   ,   55  ,   121 ,   203 ,   69  ,   515 ,   767 ,   861 ,   7189    ,   3327    ,0 };
            const unsigned long dim1633Kuo3Init[]   =   {   1   ,   3   ,   7   ,   13  ,   27  ,   53  ,   91  ,   65  ,   247 ,   257 ,   799 ,   3115    ,   6301    ,   15527   ,0 };
            const unsigned long dim1634Kuo3Init[]   =   {   1   ,   3   ,   1   ,   7   ,   27  ,   29  ,   19  ,   95  ,   343 ,   265 ,   53  ,   3227    ,   8011    ,   10223   ,0 };
            const unsigned long dim1635Kuo3Init[]   =   {   1   ,   3   ,   1   ,   7   ,   23  ,   31  ,   89  ,   43  ,   455 ,   581 ,   2023    ,   845 ,   7765    ,   4811    ,0 };
            const unsigned long dim1636Kuo3Init[]   =   {   1   ,   3   ,   3   ,   15  ,   31  ,   37  ,   117 ,   173 ,   343 ,   971 ,   491 ,   1447    ,   1547    ,   6737    ,0 };
            const unsigned long dim1637Kuo3Init[]   =   {   1   ,   3   ,   1   ,   1   ,   7   ,   43  ,   75  ,   67  ,   407 ,   697 ,   51  ,   207 ,   2499    ,   9681    ,0 };
            const unsigned long dim1638Kuo3Init[]   =   {   1   ,   1   ,   3   ,   7   ,   17  ,   27  ,   55  ,   3   ,   237 ,   983 ,   1   ,   1227    ,   2605    ,   8441    ,0 };
            const unsigned long dim1639Kuo3Init[]   =   {   1   ,   1   ,   1   ,   7   ,   3   ,   31  ,   47  ,   65  ,   119 ,   823 ,   1055    ,   1047    ,   6067    ,   4267    ,0 };
            const unsigned long dim1640Kuo3Init[]   =   {   1   ,   3   ,   1   ,   15  ,   9   ,   11  ,   41  ,   211 ,   249 ,   739 ,   953 ,   1365    ,   4749    ,   3723    ,0 };
            const unsigned long dim1641Kuo3Init[]   =   {   1   ,   1   ,   5   ,   5   ,   25  ,   29  ,   31  ,   75  ,   207 ,   395 ,   1843    ,   1737    ,   387 ,   11871   ,0 };
            const unsigned long dim1642Kuo3Init[]   =   {   1   ,   1   ,   5   ,   15  ,   21  ,   5   ,   3   ,   221 ,   223 ,   975 ,   743 ,   2217    ,   7743    ,   4667    ,0 };
            const unsigned long dim1643Kuo3Init[]   =   {   1   ,   1   ,   5   ,   1   ,   23  ,   63  ,   113 ,   149 ,   427 ,   499 ,   1259    ,   735 ,   2839    ,   503 ,0 };
            const unsigned long dim1644Kuo3Init[]   =   {   1   ,   3   ,   5   ,   5   ,   19  ,   47  ,   53  ,   205 ,   385 ,   409 ,   475 ,   265 ,   8183    ,   11709   ,0 };
            const unsigned long dim1645Kuo3Init[]   =   {   1   ,   3   ,   7   ,   7   ,   21  ,   25  ,   91  ,   143 ,   373 ,   967 ,   1735    ,   1149    ,   3077    ,   8809    ,0 };
            const unsigned long dim1646Kuo3Init[]   =   {   1   ,   3   ,   1   ,   1   ,   23  ,   39  ,   39  ,   89  ,   241 ,   181 ,   1819    ,   3431    ,   1605    ,   2143    ,0 };
            const unsigned long dim1647Kuo3Init[]   =   {   1   ,   1   ,   7   ,   13  ,   27  ,   21  ,   69  ,   35  ,   327 ,   233 ,   495 ,   157 ,   5961    ,   15785   ,0 };
            const unsigned long dim1648Kuo3Init[]   =   {   1   ,   1   ,   5   ,   15  ,   1   ,   29  ,   77  ,   249 ,   471 ,   31  ,   1805    ,   1431    ,   1703    ,   3593    ,0 };
            const unsigned long dim1649Kuo3Init[]   =   {   1   ,   3   ,   1   ,   13  ,   13  ,   31  ,   15  ,   89  ,   35  ,   143 ,   487 ,   4047    ,   7681    ,   1557    ,0 };
            const unsigned long dim1650Kuo3Init[]   =   {   1   ,   1   ,   5   ,   3   ,   21  ,   3   ,   81  ,   1   ,   15  ,   531 ,   1443    ,   2421    ,   4465    ,   4431    ,0 };
            const unsigned long dim1651Kuo3Init[]   =   {   1   ,   1   ,   1   ,   13  ,   5   ,   25  ,   83  ,   1   ,   47  ,   723 ,   1427    ,   1013    ,   1525    ,   8877    ,0 };
            const unsigned long dim1652Kuo3Init[]   =   {   1   ,   1   ,   5   ,   15  ,   25  ,   23  ,   71  ,   87  ,   13  ,   543 ,   1715    ,   2449    ,   4639    ,   8339    ,0 };
            const unsigned long dim1653Kuo3Init[]   =   {   1   ,   1   ,   7   ,   9   ,   1   ,   27  ,   101 ,   129 ,   81  ,   747 ,   1745    ,   2545    ,   455 ,   14103   ,0 };
            const unsigned long dim1654Kuo3Init[]   =   {   1   ,   1   ,   1   ,   11  ,   23  ,   59  ,   125 ,   49  ,   323 ,   989 ,   1959    ,   2349    ,   2835    ,   115 ,0 };
            const unsigned long dim1655Kuo3Init[]   =   {   1   ,   3   ,   1   ,   11  ,   29  ,   23  ,   31  ,   185 ,   273 ,   43  ,   1093    ,   323 ,   751 ,   7473    ,0 };
            const unsigned long dim1656Kuo3Init[]   =   {   1   ,   1   ,   5   ,   7   ,   3   ,   51  ,   65  ,   231 ,   147 ,   69  ,   1635    ,   1593    ,   7385    ,   1251    ,0 };
            const unsigned long dim1657Kuo3Init[]   =   {   1   ,   1   ,   3   ,   7   ,   21  ,   7   ,   109 ,   61  ,   387 ,   819 ,   417 ,   1393    ,   5535    ,   12201   ,0 };
            const unsigned long dim1658Kuo3Init[]   =   {   1   ,   3   ,   1   ,   1   ,   11  ,   17  ,   41  ,   61  ,   361 ,   891 ,   69  ,   935 ,   2793    ,   1631    ,0 };
            const unsigned long dim1659Kuo3Init[]   =   {   1   ,   3   ,   3   ,   13  ,   25  ,   31  ,   43  ,   47  ,   247 ,   723 ,   1703    ,   1499    ,   139 ,   8809    ,0 };
            const unsigned long dim1660Kuo3Init[]   =   {   1   ,   3   ,   5   ,   1   ,   9   ,   47  ,   45  ,   117 ,   95  ,   345 ,   1431    ,   975 ,   6575    ,   1213    ,0 };
            const unsigned long dim1661Kuo3Init[]   =   {   1   ,   3   ,   1   ,   13  ,   29  ,   37  ,   45  ,   91  ,   443 ,   471 ,   1363    ,   1839    ,   6263    ,   3289    ,0 };
            const unsigned long dim1662Kuo3Init[]   =   {   1   ,   1   ,   3   ,   5   ,   17  ,   17  ,   57  ,   193 ,   477 ,   283 ,   1307    ,   967 ,   1793    ,   12551   ,0 };
            const unsigned long dim1663Kuo3Init[]   =   {   1   ,   1   ,   1   ,   7   ,   17  ,   1   ,   103 ,   67  ,   113 ,   805 ,   1993    ,   3019    ,   1911    ,   11961   ,0 };
            const unsigned long dim1664Kuo3Init[]   =   {   1   ,   1   ,   3   ,   1   ,   23  ,   47  ,   19  ,   137 ,   511 ,   685 ,   1525    ,   1703    ,   2513    ,   12541   ,0 };
            const unsigned long dim1665Kuo3Init[]   =   {   1   ,   3   ,   3   ,   3   ,   3   ,   7   ,   29  ,   45  ,   279 ,   855 ,   1931    ,   2261    ,   2097    ,   2373    ,0 };
            const unsigned long dim1666Kuo3Init[]   =   {   1   ,   3   ,   7   ,   13  ,   13  ,   23  ,   105 ,   119 ,   473 ,   809 ,   925 ,   3295    ,   7547    ,   4229    ,0 };
            const unsigned long dim1667Kuo3Init[]   =   {   1   ,   1   ,   1   ,   9   ,   27  ,   37  ,   3   ,   109 ,   257 ,   743 ,   1233    ,   2973    ,   3313    ,   2949    ,0 };
            const unsigned long dim1668Kuo3Init[]   =   {   1   ,   3   ,   7   ,   13  ,   25  ,   37  ,   103 ,   157 ,   333 ,   127 ,   1893    ,   649 ,   7773    ,   15639   ,0 };
            const unsigned long dim1669Kuo3Init[]   =   {   1   ,   1   ,   1   ,   11  ,   13  ,   53  ,   21  ,   67  ,   161 ,   859 ,   1715    ,   3303    ,   2527    ,   9487    ,0 };
            const unsigned long dim1670Kuo3Init[]   =   {   1   ,   1   ,   7   ,   3   ,   7   ,   27  ,   47  ,   173 ,   103 ,   413 ,   719 ,   2865    ,   2337    ,   2963    ,0 };
            const unsigned long dim1671Kuo3Init[]   =   {   1   ,   1   ,   5   ,   11  ,   25  ,   17  ,   3   ,   105 ,   85  ,   731 ,   933 ,   3967    ,   279 ,   14649   ,0 };
            const unsigned long dim1672Kuo3Init[]   =   {   1   ,   1   ,   7   ,   5   ,   11  ,   15  ,   37  ,   197 ,   443 ,   667 ,   1373    ,   2961    ,   615 ,   6011    ,0 };
            const unsigned long dim1673Kuo3Init[]   =   {   1   ,   1   ,   3   ,   9   ,   11  ,   45  ,   9   ,   43  ,   339 ,   203 ,   27  ,   3047    ,   6531    ,   1071    ,0 };
            const unsigned long dim1674Kuo3Init[]   =   {   1   ,   1   ,   3   ,   9   ,   1   ,   23  ,   111 ,   185 ,   485 ,   425 ,   2043    ,   3375    ,   4037    ,   8751    ,0 };
            const unsigned long dim1675Kuo3Init[]   =   {   1   ,   1   ,   7   ,   7   ,   21  ,   35  ,   37  ,   87  ,   393 ,   149 ,   1475    ,   4027    ,   8125    ,   16021   ,0 };
            const unsigned long dim1676Kuo3Init[]   =   {   1   ,   1   ,   3   ,   11  ,   21  ,   51  ,   125 ,   229 ,   49  ,   611 ,   919 ,   3103    ,   7239    ,   3553    ,0 };
            const unsigned long dim1677Kuo3Init[]   =   {   1   ,   3   ,   7   ,   5   ,   5   ,   27  ,   77  ,   99  ,   313 ,   667 ,   413 ,   1933    ,   7977    ,   8903    ,0 };
            const unsigned long dim1678Kuo3Init[]   =   {   1   ,   3   ,   3   ,   9   ,   5   ,   47  ,   47  ,   99  ,   181 ,   253 ,   433 ,   3385    ,   5789    ,   9043    ,0 };
            const unsigned long dim1679Kuo3Init[]   =   {   1   ,   3   ,   3   ,   13  ,   9   ,   9   ,   117 ,   51  ,   93  ,   615 ,   1159    ,   251 ,   3345    ,   925 ,0 };
            const unsigned long dim1680Kuo3Init[]   =   {   1   ,   3   ,   1   ,   7   ,   21  ,   27  ,   75  ,   201 ,   489 ,   901 ,   643 ,   2511    ,   3803    ,   11525   ,0 };
            const unsigned long dim1681Kuo3Init[]   =   {   1   ,   3   ,   3   ,   1   ,   15  ,   27  ,   117 ,   1   ,   69  ,   1019    ,   1853    ,   3745    ,   3549    ,   5721    ,0 };
            const unsigned long dim1682Kuo3Init[]   =   {   1   ,   3   ,   7   ,   7   ,   23  ,   63  ,   113 ,   235 ,   315 ,   613 ,   1691    ,   2725    ,   2941    ,   11339   ,0 };
            const unsigned long dim1683Kuo3Init[]   =   {   1   ,   1   ,   3   ,   15  ,   19  ,   1   ,   3   ,   157 ,   251 ,   443 ,   217 ,   2229    ,   6723    ,   15045   ,0 };
            const unsigned long dim1684Kuo3Init[]   =   {   1   ,   1   ,   1   ,   3   ,   1   ,   1   ,   93  ,   133 ,   69  ,   897 ,   197 ,   2241    ,   1727    ,   3645    ,0 };
            const unsigned long dim1685Kuo3Init[]   =   {   1   ,   3   ,   3   ,   15  ,   31  ,   33  ,   83  ,   21  ,   3   ,   243 ,   1857    ,   1343    ,   1359    ,   2631    ,0 };
            const unsigned long dim1686Kuo3Init[]   =   {   1   ,   1   ,   5   ,   13  ,   27  ,   17  ,   1   ,   147 ,   113 ,   849 ,   1899    ,   1037    ,   5029    ,   6829    ,0 };
            const unsigned long dim1687Kuo3Init[]   =   {   1   ,   3   ,   1   ,   1   ,   21  ,   3   ,   95  ,   69  ,   425 ,   185 ,   1983    ,   3887    ,   843 ,   14881   ,0 };
            const unsigned long dim1688Kuo3Init[]   =   {   1   ,   3   ,   5   ,   13  ,   11  ,   31  ,   49  ,   35  ,   349 ,   429 ,   1173    ,   3783    ,   7979    ,   13651   ,0 };
            const unsigned long dim1689Kuo3Init[]   =   {   1   ,   1   ,   5   ,   11  ,   25  ,   13  ,   49  ,   97  ,   293 ,   661 ,   387 ,   79  ,   5863    ,   9315    ,0 };
            const unsigned long dim1690Kuo3Init[]   =   {   1   ,   1   ,   5   ,   3   ,   27  ,   25  ,   9   ,   51  ,   185 ,   569 ,   1129    ,   2657    ,   6609    ,   8383    ,0 };
            const unsigned long dim1691Kuo3Init[]   =   {   1   ,   1   ,   1   ,   11  ,   29  ,   7   ,   67  ,   77  ,   77  ,   661 ,   259 ,   2365    ,   3649    ,   9989    ,0 };
            const unsigned long dim1692Kuo3Init[]   =   {   1   ,   1   ,   5   ,   9   ,   29  ,   59  ,   43  ,   241 ,   279 ,   177 ,   721 ,   1329    ,   3709    ,   8987    ,0 };
            const unsigned long dim1693Kuo3Init[]   =   {   1   ,   3   ,   5   ,   7   ,   25  ,   35  ,   51  ,   213 ,   367 ,   735 ,   11  ,   1061    ,   497 ,   13593   ,0 };
            const unsigned long dim1694Kuo3Init[]   =   {   1   ,   1   ,   1   ,   9   ,   1   ,   39  ,   79  ,   31  ,   497 ,   799 ,   1575    ,   2627    ,   2821    ,   4815    ,0 };
            const unsigned long dim1695Kuo3Init[]   =   {   1   ,   3   ,   7   ,   7   ,   15  ,   31  ,   99  ,   199 ,   369 ,   829 ,   721 ,   3509    ,   3973    ,   9491    ,0 };
            const unsigned long dim1696Kuo3Init[]   =   {   1   ,   1   ,   1   ,   15  ,   7   ,   49  ,   125 ,   249 ,   295 ,   907 ,   1243    ,   155 ,   2519    ,   2295    ,0 };
            const unsigned long dim1697Kuo3Init[]   =   {   1   ,   3   ,   7   ,   9   ,   7   ,   1   ,   63  ,   141 ,   505 ,   865 ,   1389    ,   1041    ,   191 ,   4109    ,0 };
            const unsigned long dim1698Kuo3Init[]   =   {   1   ,   3   ,   5   ,   13  ,   5   ,   45  ,   5   ,   53  ,   77  ,   525 ,   579 ,   3077    ,   6737    ,   6839    ,0 };
            const unsigned long dim1699Kuo3Init[]   =   {   1   ,   1   ,   3   ,   5   ,   23  ,   5   ,   127 ,   61  ,   421 ,   353 ,   589 ,   4095    ,   2841    ,   8095    ,0 };
            const unsigned long dim1700Kuo3Init[]   =   {   1   ,   1   ,   7   ,   9   ,   3   ,   11  ,   49  ,   103 ,   75  ,   403 ,   441 ,   3513    ,   6535    ,   7811    ,0 };
            const unsigned long dim1701Kuo3Init[]   =   {   1   ,   3   ,   7   ,   13  ,   21  ,   57  ,   115 ,   59  ,   341 ,   673 ,   759 ,   81  ,   6585    ,   9727    ,0 };
            const unsigned long dim1702Kuo3Init[]   =   {   1   ,   1   ,   3   ,   7   ,   29  ,   33  ,   35  ,   201 ,   149 ,   41  ,   175 ,   1355    ,   5311    ,   85  ,0 };
            const unsigned long dim1703Kuo3Init[]   =   {   1   ,   3   ,   1   ,   1   ,   31  ,   5   ,   51  ,   83  ,   225 ,   123 ,   1347    ,   2017    ,   6259    ,   1553    ,0 };
            const unsigned long dim1704Kuo3Init[]   =   {   1   ,   1   ,   5   ,   1   ,   17  ,   57  ,   77  ,   1   ,   393 ,   875 ,   887 ,   1   ,   4007    ,   6391    ,0 };
            const unsigned long dim1705Kuo3Init[]   =   {   1   ,   1   ,   3   ,   5   ,   3   ,   31  ,   83  ,   167 ,   511 ,   19  ,   525 ,   323 ,   6991    ,   10411   ,0 };
            const unsigned long dim1706Kuo3Init[]   =   {   1   ,   1   ,   7   ,   9   ,   15  ,   61  ,   77  ,   189 ,   43  ,   651 ,   305 ,   3107    ,   6433    ,   3035    ,0 };
            const unsigned long dim1707Kuo3Init[]   =   {   1   ,   3   ,   1   ,   11  ,   25  ,   15  ,   51  ,   233 ,   395 ,   867 ,   145 ,   2215    ,   917 ,   13815   ,0 };
            const unsigned long dim1708Kuo3Init[]   =   {   1   ,   3   ,   1   ,   13  ,   23  ,   15  ,   3   ,   103 ,   439 ,   691 ,   283 ,   3917    ,   5525    ,   11007   ,0 };
            const unsigned long dim1709Kuo3Init[]   =   {   1   ,   3   ,   7   ,   11  ,   11  ,   49  ,   39  ,   31  ,   127 ,   521 ,   1327    ,   43  ,   1509    ,   679 ,0 };
            const unsigned long dim1710Kuo3Init[]   =   {   1   ,   1   ,   7   ,   5   ,   17  ,   27  ,   17  ,   181 ,   337 ,   85  ,   1687    ,   3245    ,   5315    ,   5653    ,0 };
            const unsigned long dim1711Kuo3Init[]   =   {   1   ,   3   ,   5   ,   15  ,   31  ,   33  ,   3   ,   229 ,   103 ,   337 ,   1337    ,   1961    ,   6053    ,   8661    ,0 };
            const unsigned long dim1712Kuo3Init[]   =   {   1   ,   3   ,   1   ,   7   ,   31  ,   63  ,   15  ,   135 ,   77  ,   583 ,   697 ,   1353    ,   6543    ,   291 ,0 };
            const unsigned long dim1713Kuo3Init[]   =   {   1   ,   3   ,   7   ,   9   ,   17  ,   53  ,   55  ,   59  ,   149 ,   995 ,   1059    ,   1829    ,   7759    ,   10901   ,0 };
            const unsigned long dim1714Kuo3Init[]   =   {   1   ,   3   ,   5   ,   7   ,   5   ,   63  ,   51  ,   123 ,   369 ,   861 ,   1835    ,   3369    ,   4555    ,   1623    ,0 };
            const unsigned long dim1715Kuo3Init[]   =   {   1   ,   1   ,   3   ,   11  ,   23  ,   31  ,   21  ,   247 ,   417 ,   565 ,   689 ,   1401    ,   405 ,   6657    ,0 };
            const unsigned long dim1716Kuo3Init[]   =   {   1   ,   3   ,   7   ,   7   ,   21  ,   21  ,   37  ,   123 ,   359 ,   133 ,   397 ,   173 ,   2135    ,   12407   ,0 };
            const unsigned long dim1717Kuo3Init[]   =   {   1   ,   1   ,   1   ,   3   ,   1   ,   23  ,   91  ,   97  ,   421 ,   531 ,   975 ,   2647    ,   4337    ,   7775    ,0 };
            const unsigned long dim1718Kuo3Init[]   =   {   1   ,   1   ,   3   ,   13  ,   5   ,   9   ,   57  ,   149 ,   243 ,   225 ,   79  ,   2745    ,   523 ,   16133   ,0 };
            const unsigned long dim1719Kuo3Init[]   =   {   1   ,   1   ,   1   ,   11  ,   7   ,   5   ,   97  ,   91  ,   15  ,   217 ,   991 ,   1465    ,   3975    ,   8467    ,0 };
            const unsigned long dim1720Kuo3Init[]   =   {   1   ,   3   ,   7   ,   13  ,   9   ,   61  ,   97  ,   153 ,   411 ,   801 ,   59  ,   603 ,   5919    ,   9509    ,0 };
            const unsigned long dim1721Kuo3Init[]   =   {   1   ,   3   ,   5   ,   7   ,   29  ,   11  ,   71  ,   111 ,   473 ,   985 ,   931 ,   1211    ,   653 ,   13933   ,0 };
            const unsigned long dim1722Kuo3Init[]   =   {   1   ,   1   ,   3   ,   3   ,   29  ,   49  ,   71  ,   105 ,   29  ,   333 ,   1261    ,   3791    ,   5349    ,   15675   ,0 };
            const unsigned long dim1723Kuo3Init[]   =   {   1   ,   1   ,   7   ,   13  ,   3   ,   53  ,   73  ,   77  ,   451 ,   547 ,   51  ,   669 ,   7355    ,   8325    ,0 };
            const unsigned long dim1724Kuo3Init[]   =   {   1   ,   1   ,   1   ,   13  ,   7   ,   9   ,   37  ,   87  ,   273 ,   275 ,   191 ,   469 ,   7529    ,   1955    ,0 };
            const unsigned long dim1725Kuo3Init[]   =   {   1   ,   3   ,   5   ,   3   ,   21  ,   5   ,   121 ,   195 ,   327 ,   477 ,   1661    ,   3499    ,   7903    ,   9053    ,0 };
            const unsigned long dim1726Kuo3Init[]   =   {   1   ,   3   ,   3   ,   11  ,   29  ,   41  ,   25  ,   13  ,   429 ,   351 ,   723 ,   3901    ,   5841    ,   9783    ,0 };
            const unsigned long dim1727Kuo3Init[]   =   {   1   ,   3   ,   1   ,   5   ,   1   ,   25  ,   15  ,   203 ,   47  ,   289 ,   67  ,   847 ,   5547    ,   15845   ,0 };
            const unsigned long dim1728Kuo3Init[]   =   {   1   ,   3   ,   3   ,   11  ,   21  ,   7   ,   53  ,   39  ,   461 ,   153 ,   1739    ,   4021    ,   307 ,   1435    ,0 };
            const unsigned long dim1729Kuo3Init[]   =   {   1   ,   3   ,   5   ,   5   ,   7   ,   17  ,   87  ,   23  ,   47  ,   839 ,   1841    ,   3877    ,   2401    ,   853 ,0 };
            const unsigned long dim1730Kuo3Init[]   =   {   1   ,   3   ,   3   ,   9   ,   31  ,   37  ,   15  ,   49  ,   171 ,   757 ,   499 ,   1455    ,   2151    ,   6169    ,0 };
            const unsigned long dim1731Kuo3Init[]   =   {   1   ,   1   ,   5   ,   1   ,   31  ,   19  ,   33  ,   87  ,   507 ,   685 ,   1765    ,   3109    ,   7221    ,   8047    ,0 };
            const unsigned long dim1732Kuo3Init[]   =   {   1   ,   1   ,   7   ,   7   ,   23  ,   7   ,   25  ,   45  ,   167 ,   865 ,   465 ,   3377    ,   7203    ,   7825    ,0 };
            const unsigned long dim1733Kuo3Init[]   =   {   1   ,   3   ,   3   ,   13  ,   15  ,   53  ,   79  ,   111 ,   391 ,   777 ,   367 ,   1445    ,   5243    ,   1929    ,0 };
            const unsigned long dim1734Kuo3Init[]   =   {   1   ,   1   ,   3   ,   15  ,   11  ,   11  ,   57  ,   115 ,   239 ,   77  ,   797 ,   2073    ,   831 ,   11879   ,0 };
            const unsigned long dim1735Kuo3Init[]   =   {   1   ,   1   ,   7   ,   5   ,   23  ,   11  ,   91  ,   153 ,   29  ,   729 ,   241 ,   2627    ,   6447    ,   13239   ,0 };
            const unsigned long dim1736Kuo3Init[]   =   {   1   ,   1   ,   7   ,   9   ,   31  ,   11  ,   59  ,   243 ,   105 ,   83  ,   363 ,   3473    ,   315 ,   2371    ,0 };
            const unsigned long dim1737Kuo3Init[]   =   {   1   ,   3   ,   5   ,   1   ,   19  ,   31  ,   111 ,   91  ,   215 ,   697 ,   103 ,   3323    ,   7259    ,   13625   ,0 };
            const unsigned long dim1738Kuo3Init[]   =   {   1   ,   1   ,   1   ,   11  ,   5   ,   59  ,   83  ,   129 ,   45  ,   223 ,   835 ,   1965    ,   3975    ,   15233   ,0 };
            const unsigned long dim1739Kuo3Init[]   =   {   1   ,   3   ,   7   ,   11  ,   19  ,   11  ,   31  ,   225 ,   175 ,   209 ,   619 ,   3127    ,   4943    ,   9701    ,0 };
            const unsigned long dim1740Kuo3Init[]   =   {   1   ,   1   ,   1   ,   3   ,   3   ,   19  ,   21  ,   133 ,   7   ,   151 ,   1017    ,   3867    ,   3699    ,   13051   ,0 };
            const unsigned long dim1741Kuo3Init[]   =   {   1   ,   1   ,   5   ,   9   ,   27  ,   17  ,   81  ,   139 ,   439 ,   11  ,   265 ,   659 ,   1329    ,   4259    ,0 };
            const unsigned long dim1742Kuo3Init[]   =   {   1   ,   3   ,   5   ,   1   ,   1   ,   59  ,   105 ,   251 ,   131 ,   817 ,   1215    ,   3005    ,   3417    ,   5637    ,0 };
            const unsigned long dim1743Kuo3Init[]   =   {   1   ,   3   ,   3   ,   15  ,   31  ,   37  ,   27  ,   243 ,   225 ,   631 ,   2023    ,   2509    ,   249 ,   593 ,0 };
            const unsigned long dim1744Kuo3Init[]   =   {   1   ,   1   ,   3   ,   11  ,   27  ,   51  ,   89  ,   33  ,   423 ,   423 ,   789 ,   3627    ,   2995    ,   8477    ,0 };
            const unsigned long dim1745Kuo3Init[]   =   {   1   ,   3   ,   5   ,   5   ,   5   ,   31  ,   69  ,   65  ,   381 ,   375 ,   221 ,   477 ,   3311    ,   12103   ,0 };
            const unsigned long dim1746Kuo3Init[]   =   {   1   ,   3   ,   1   ,   15  ,   31  ,   1   ,   75  ,   97  ,   143 ,   749 ,   581 ,   4019    ,   4655    ,   15571   ,0 };
            const unsigned long dim1747Kuo3Init[]   =   {   1   ,   3   ,   7   ,   3   ,   1   ,   31  ,   51  ,   151 ,   127 ,   917 ,   975 ,   2839    ,   3815    ,   7405    ,0 };
            const unsigned long dim1748Kuo3Init[]   =   {   1   ,   3   ,   1   ,   3   ,   9   ,   61  ,   93  ,   49  ,   41  ,   881 ,   1205    ,   2425    ,   7727    ,   2687    ,0 };
            const unsigned long dim1749Kuo3Init[]   =   {   1   ,   3   ,   3   ,   3   ,   27  ,   11  ,   99  ,   51  ,   51  ,   937 ,   1511    ,   3891    ,   3955    ,   6055    ,0 };
            const unsigned long dim1750Kuo3Init[]   =   {   1   ,   3   ,   3   ,   3   ,   3   ,   27  ,   81  ,   107 ,   409 ,   895 ,   1869    ,   2721    ,   1505    ,   9877    ,0 };
            const unsigned long dim1751Kuo3Init[]   =   {   1   ,   1   ,   1   ,   5   ,   11  ,   43  ,   101 ,   7   ,   263 ,   881 ,   1029    ,   1491    ,   4147    ,   10339   ,0 };
            const unsigned long dim1752Kuo3Init[]   =   {   1   ,   1   ,   3   ,   11  ,   25  ,   21  ,   85  ,   145 ,   489 ,   335 ,   643 ,   4019    ,   951 ,   5885    ,0 };
            const unsigned long dim1753Kuo3Init[]   =   {   1   ,   1   ,   1   ,   15  ,   3   ,   55  ,   113 ,   87  ,   53  ,   839 ,   381 ,   1961    ,   3865    ,   2841    ,0 };
            const unsigned long dim1754Kuo3Init[]   =   {   1   ,   1   ,   3   ,   13  ,   11  ,   33  ,   25  ,   95  ,   47  ,   321 ,   649 ,   1641    ,   1715    ,   3815    ,0 };
            const unsigned long dim1755Kuo3Init[]   =   {   1   ,   1   ,   1   ,   15  ,   23  ,   59  ,   5   ,   3   ,   451 ,   781 ,   867 ,   747 ,   793 ,   6147    ,0 };
            const unsigned long dim1756Kuo3Init[]   =   {   1   ,   1   ,   1   ,   15  ,   3   ,   27  ,   31  ,   125 ,   355 ,   837 ,   689 ,   3599    ,   1451    ,   2167    ,0 };
            const unsigned long dim1757Kuo3Init[]   =   {   1   ,   1   ,   3   ,   9   ,   21  ,   5   ,   101 ,   37  ,   465 ,   115 ,   1607    ,   2999    ,   631 ,   1435    ,0 };
            const unsigned long dim1758Kuo3Init[]   =   {   1   ,   3   ,   5   ,   9   ,   19  ,   41  ,   91  ,   241 ,   375 ,   707 ,   1573    ,   105 ,   863 ,   11273   ,0 };
            const unsigned long dim1759Kuo3Init[]   =   {   1   ,   1   ,   5   ,   15  ,   17  ,   63  ,   95  ,   83  ,   273 ,   617 ,   1299    ,   1259    ,   7997    ,   5225    ,0 };
            const unsigned long dim1760Kuo3Init[]   =   {   1   ,   3   ,   5   ,   9   ,   11  ,   5   ,   127 ,   67  ,   189 ,   733 ,   1175    ,   3099    ,   6561    ,   7771    ,0 };
            const unsigned long dim1761Kuo3Init[]   =   {   1   ,   1   ,   7   ,   11  ,   13  ,   45  ,   75  ,   101 ,   279 ,   147 ,   177 ,   225 ,   557 ,   9413    ,0 };
            const unsigned long dim1762Kuo3Init[]   =   {   1   ,   3   ,   1   ,   13  ,   11  ,   53  ,   49  ,   71  ,   451 ,   327 ,   567 ,   531 ,   589 ,   15395   ,0 };
            const unsigned long dim1763Kuo3Init[]   =   {   1   ,   1   ,   7   ,   13  ,   29  ,   51  ,   95  ,   227 ,   87  ,   171 ,   1769    ,   2799    ,   2589    ,   7255    ,0 };
            const unsigned long dim1764Kuo3Init[]   =   {   1   ,   1   ,   5   ,   7   ,   1   ,   51  ,   109 ,   135 ,   99  ,   957 ,   1681    ,   2217    ,   2247    ,   6567    ,0 };
            const unsigned long dim1765Kuo3Init[]   =   {   1   ,   1   ,   3   ,   7   ,   27  ,   57  ,   101 ,   51  ,   237 ,   835 ,   1339    ,   955 ,   5969    ,   12511   ,0 };
            const unsigned long dim1766Kuo3Init[]   =   {   1   ,   1   ,   5   ,   1   ,   21  ,   9   ,   95  ,   247 ,   309 ,   413 ,   17  ,   3473    ,   3961    ,   8555    ,0 };
            const unsigned long dim1767Kuo3Init[]   =   {   1   ,   1   ,   1   ,   1   ,   21  ,   47  ,   29  ,   33  ,   135 ,   447 ,   1387    ,   107 ,   4781    ,   10987   ,0 };
            const unsigned long dim1768Kuo3Init[]   =   {   1   ,   1   ,   3   ,   11  ,   5   ,   17  ,   41  ,   103 ,   83  ,   829 ,   1449    ,   4047    ,   4807    ,   7277    ,0 };
            const unsigned long dim1769Kuo3Init[]   =   {   1   ,   3   ,   7   ,   15  ,   5   ,   1   ,   75  ,   241 ,   235 ,   701 ,   1529    ,   1621    ,   7995    ,   6243    ,0 };
            const unsigned long dim1770Kuo3Init[]   =   {   1   ,   1   ,   3   ,   5   ,   23  ,   49  ,   31  ,   173 ,   83  ,   961 ,   1293    ,   1279    ,   2495    ,   10805   ,0 };
            const unsigned long dim1771Kuo3Init[]   =   {   1   ,   1   ,   1   ,   1   ,   15  ,   39  ,   17  ,   81  ,   167 ,   209 ,   685 ,   1089    ,   7197    ,   15773   ,0 };
            const unsigned long dim1772Kuo3Init[]   =   {   1   ,   1   ,   1   ,   15  ,   29  ,   1   ,   73  ,   113 ,   21  ,   301 ,   1087    ,   2529    ,   6835    ,   8089    ,0 };
            const unsigned long dim1773Kuo3Init[]   =   {   1   ,   1   ,   5   ,   15  ,   1   ,   7   ,   99  ,   87  ,   177 ,   537 ,   283 ,   205 ,   6241    ,   12489   ,0 };
            const unsigned long dim1774Kuo3Init[]   =   {   1   ,   3   ,   7   ,   3   ,   11  ,   13  ,   89  ,   43  ,   37  ,   859 ,   263 ,   2623    ,   6941    ,   7433    ,0 };
            const unsigned long dim1775Kuo3Init[]   =   {   1   ,   3   ,   3   ,   13  ,   9   ,   47  ,   13  ,   155 ,   109 ,   507 ,   343 ,   1049    ,   4915    ,   14873   ,0 };
            const unsigned long dim1776Kuo3Init[]   =   {   1   ,   1   ,   5   ,   1   ,   5   ,   37  ,   5   ,   71  ,   1   ,   17  ,   417 ,   3137    ,   3045    ,   5835    ,0 };
            const unsigned long dim1777Kuo3Init[]   =   {   1   ,   3   ,   3   ,   5   ,   15  ,   43  ,   31  ,   235 ,   379 ,   807 ,   89  ,   3375    ,   123 ,   9783    ,0 };
            const unsigned long dim1778Kuo3Init[]   =   {   1   ,   3   ,   5   ,   3   ,   19  ,   23  ,   53  ,   129 ,   493 ,   393 ,   91  ,   2345    ,   1959    ,   745 ,0 };
            const unsigned long dim1779Kuo3Init[]   =   {   1   ,   3   ,   3   ,   13  ,   21  ,   33  ,   31  ,   125 ,   99  ,   167 ,   685 ,   733 ,   6479    ,   10391   ,0 };
            const unsigned long dim1780Kuo3Init[]   =   {   1   ,   1   ,   1   ,   11  ,   1   ,   17  ,   9   ,   245 ,   75  ,   281 ,   1427    ,   675 ,   5393    ,   8841    ,0 };
            const unsigned long dim1781Kuo3Init[]   =   {   1   ,   1   ,   3   ,   7   ,   15  ,   63  ,   31  ,   83  ,   359 ,   643 ,   77  ,   3941    ,   5477    ,   3435    ,0 };
            const unsigned long dim1782Kuo3Init[]   =   {   1   ,   1   ,   3   ,   9   ,   31  ,   21  ,   3   ,   111 ,   233 ,   745 ,   1819    ,   2617    ,   5735    ,   15237   ,0 };
            const unsigned long dim1783Kuo3Init[]   =   {   1   ,   3   ,   1   ,   9   ,   29  ,   23  ,   99  ,   25  ,   119 ,   69  ,   1063    ,   651 ,   4983    ,   14079   ,0 };
            const unsigned long dim1784Kuo3Init[]   =   {   1   ,   3   ,   3   ,   3   ,   9   ,   53  ,   47  ,   127 ,   443 ,   113 ,   1651    ,   2967    ,   6633    ,   10025   ,0 };
            const unsigned long dim1785Kuo3Init[]   =   {   1   ,   3   ,   7   ,   7   ,   19  ,   11  ,   5   ,   49  ,   23  ,   567 ,   1095    ,   3815    ,   4187    ,   5305    ,0 };
            const unsigned long dim1786Kuo3Init[]   =   {   1   ,   1   ,   1   ,   11  ,   29  ,   63  ,   85  ,   147 ,   105 ,   307 ,   1829    ,   27  ,   7499    ,   15661   ,0 };
            const unsigned long dim1787Kuo3Init[]   =   {   1   ,   3   ,   3   ,   9   ,   15  ,   19  ,   97  ,   65  ,   71  ,   597 ,   551 ,   947 ,   8047    ,   8513    ,0 };
            const unsigned long dim1788Kuo3Init[]   =   {   1   ,   1   ,   5   ,   1   ,   29  ,   31  ,   1   ,   135 ,   159 ,   769 ,   997 ,   3027    ,   1485    ,   12627   ,0 };
            const unsigned long dim1789Kuo3Init[]   =   {   1   ,   1   ,   3   ,   15  ,   29  ,   47  ,   37  ,   43  ,   45  ,   649 ,   719 ,   1757    ,   4619    ,   11571   ,0 };
            const unsigned long dim1790Kuo3Init[]   =   {   1   ,   1   ,   1   ,   5   ,   9   ,   9   ,   117 ,   207 ,   107 ,   651 ,   1155    ,   2047    ,   7489    ,   5959    ,0 };
            const unsigned long dim1791Kuo3Init[]   =   {   1   ,   3   ,   1   ,   13  ,   29  ,   37  ,   69  ,   243 ,   347 ,   849 ,   1503    ,   3985    ,   6081    ,   1471    ,0 };
            const unsigned long dim1792Kuo3Init[]   =   {   1   ,   1   ,   1   ,   11  ,   1   ,   21  ,   107 ,   9   ,   377 ,   147 ,   925 ,   3055    ,   397 ,   11337   ,0 };
            const unsigned long dim1793Kuo3Init[]   =   {   1   ,   3   ,   5   ,   9   ,   27  ,   45  ,   57  ,   177 ,   229 ,   925 ,   855 ,   2203    ,   8065    ,   299 ,0 };
            const unsigned long dim1794Kuo3Init[]   =   {   1   ,   3   ,   1   ,   11  ,   5   ,   1   ,   77  ,   149 ,   339 ,   1023    ,   89  ,   3509    ,   4315    ,   6571    ,0 };
            const unsigned long dim1795Kuo3Init[]   =   {   1   ,   1   ,   3   ,   11  ,   19  ,   51  ,   49  ,   191 ,   179 ,   941 ,   1409    ,   1789    ,   3363    ,   11507   ,0 };
            const unsigned long dim1796Kuo3Init[]   =   {   1   ,   3   ,   3   ,   13  ,   29  ,   25  ,   47  ,   61  ,   121 ,   243 ,   1155    ,   3915    ,   5397    ,   6289    ,0 };
            const unsigned long dim1797Kuo3Init[]   =   {   1   ,   1   ,   1   ,   3   ,   31  ,   55  ,   25  ,   125 ,   239 ,   277 ,   165 ,   3631    ,   3609    ,   13607   ,0 };
            const unsigned long dim1798Kuo3Init[]   =   {   1   ,   3   ,   1   ,   15  ,   11  ,   23  ,   41  ,   199 ,   239 ,   805 ,   2021    ,   3377    ,   3103    ,   10657   ,0 };
            const unsigned long dim1799Kuo3Init[]   =   {   1   ,   1   ,   7   ,   1   ,   27  ,   49  ,   9   ,   219 ,   383 ,   523 ,   1357    ,   501 ,   733 ,   631 ,0 };
            const unsigned long dim1800Kuo3Init[]   =   {   1   ,   3   ,   5   ,   13  ,   1   ,   51  ,   125 ,   185 ,   205 ,   231 ,   247 ,   2579    ,   1283    ,   3383    ,0 };
            const unsigned long dim1801Kuo3Init[]   =   {   1   ,   1   ,   7   ,   13  ,   13  ,   49  ,   81  ,   243 ,   111 ,   219 ,   409 ,   3755    ,   3783    ,   8159    ,0 };
            const unsigned long dim1802Kuo3Init[]   =   {   1   ,   3   ,   7   ,   5   ,   13  ,   33  ,   109 ,   245 ,   141 ,   391 ,   101 ,   2043    ,   4091    ,   379 ,0 };
            const unsigned long dim1803Kuo3Init[]   =   {   1   ,   1   ,   7   ,   3   ,   27  ,   39  ,   7   ,   223 ,   107 ,   969 ,   1117    ,   3599    ,   8063    ,   7737    ,0 };
            const unsigned long dim1804Kuo3Init[]   =   {   1   ,   1   ,   5   ,   5   ,   11  ,   19  ,   9   ,   47  ,   41  ,   571 ,   1831    ,   3   ,   5721    ,   14897   ,0 };
            const unsigned long dim1805Kuo3Init[]   =   {   1   ,   1   ,   7   ,   5   ,   11  ,   25  ,   75  ,   159 ,   77  ,   843 ,   61  ,   2283    ,   4033    ,   15099   ,0 };
            const unsigned long dim1806Kuo3Init[]   =   {   1   ,   3   ,   7   ,   1   ,   5   ,   23  ,   57  ,   117 ,   45  ,   373 ,   831 ,   3047    ,   5829    ,   3205    ,0 };
            const unsigned long dim1807Kuo3Init[]   =   {   1   ,   3   ,   5   ,   11  ,   31  ,   29  ,   45  ,   129 ,   25  ,   519 ,   661 ,   2033    ,   3003    ,   12469   ,0 };
            const unsigned long dim1808Kuo3Init[]   =   {   1   ,   3   ,   1   ,   15  ,   5   ,   37  ,   55  ,   247 ,   129 ,   299 ,   1317    ,   3543    ,   421 ,   12299   ,0 };
            const unsigned long dim1809Kuo3Init[]   =   {   1   ,   3   ,   1   ,   9   ,   23  ,   63  ,   5   ,   117 ,   425 ,   883 ,   1997    ,   1701    ,   415 ,   3279    ,0 };
            const unsigned long dim1810Kuo3Init[]   =   {   1   ,   3   ,   3   ,   5   ,   1   ,   21  ,   23  ,   109 ,   55  ,   377 ,   1565    ,   2821    ,   4995    ,   3151    ,0 };
            const unsigned long dim1811Kuo3Init[]   =   {   1   ,   3   ,   7   ,   7   ,   19  ,   39  ,   43  ,   169 ,   141 ,   353 ,   459 ,   2509    ,   1503    ,   3597    ,0 };
            const unsigned long dim1812Kuo3Init[]   =   {   1   ,   1   ,   1   ,   3   ,   17  ,   55  ,   121 ,   49  ,   229 ,   589 ,   579 ,   2935    ,   4647    ,   12887   ,0 };
            const unsigned long dim1813Kuo3Init[]   =   {   1   ,   3   ,   3   ,   11  ,   11  ,   47  ,   11  ,   119 ,   83  ,   701 ,   197 ,   583 ,   7997    ,   11635   ,0 };
            const unsigned long dim1814Kuo3Init[]   =   {   1   ,   1   ,   3   ,   5   ,   25  ,   47  ,   85  ,   21  ,   317 ,   887 ,   1551    ,   2309    ,   4227    ,   8115    ,0 };
            const unsigned long dim1815Kuo3Init[]   =   {   1   ,   1   ,   3   ,   5   ,   13  ,   23  ,   71  ,   7   ,   419 ,   873 ,   1423    ,   2213    ,   7485    ,   13237   ,0 };
            const unsigned long dim1816Kuo3Init[]   =   {   1   ,   3   ,   7   ,   7   ,   25  ,   55  ,   43  ,   141 ,   125 ,   583 ,   451 ,   3897    ,   4059    ,   7409    ,0 };
            const unsigned long dim1817Kuo3Init[]   =   {   1   ,   1   ,   7   ,   5   ,   19  ,   55  ,   13  ,   61  ,   461 ,   353 ,   981 ,   2271    ,   4881    ,   12761   ,0 };
            const unsigned long dim1818Kuo3Init[]   =   {   1   ,   3   ,   3   ,   9   ,   31  ,   61  ,   105 ,   105 ,   497 ,   173 ,   1111    ,   873 ,   1869    ,   9027    ,0 };
            const unsigned long dim1819Kuo3Init[]   =   {   1   ,   1   ,   1   ,   3   ,   31  ,   31  ,   93  ,   119 ,   399 ,   617 ,   1977    ,   3913    ,   8061    ,   13017   ,0 };
            const unsigned long dim1820Kuo3Init[]   =   {   1   ,   3   ,   7   ,   9   ,   17  ,   7   ,   89  ,   149 ,   119 ,   791 ,   839 ,   3643    ,   7545    ,   13329   ,0 };
            const unsigned long dim1821Kuo3Init[]   =   {   1   ,   1   ,   5   ,   1   ,   13  ,   59  ,   35  ,   231 ,   3   ,   703 ,   881 ,   45  ,   5979    ,   3249    ,0 };
            const unsigned long dim1822Kuo3Init[]   =   {   1   ,   1   ,   1   ,   7   ,   1   ,   21  ,   117 ,   105 ,   3   ,   791 ,   1563    ,   2303    ,   3533    ,   7853    ,0 };
            const unsigned long dim1823Kuo3Init[]   =   {   1   ,   3   ,   3   ,   5   ,   11  ,   11  ,   109 ,   115 ,   233 ,   471 ,   509 ,   821 ,   7209    ,   15463   ,0 };
            const unsigned long dim1824Kuo3Init[]   =   {   1   ,   1   ,   5   ,   5   ,   13  ,   1   ,   83  ,   99  ,   21  ,   369 ,   751 ,   277 ,   881 ,   8963    ,0 };
            const unsigned long dim1825Kuo3Init[]   =   {   1   ,   3   ,   1   ,   1   ,   27  ,   57  ,   3   ,   219 ,   509 ,   337 ,   1469    ,   2881    ,   7911    ,   1983    ,0 };
            const unsigned long dim1826Kuo3Init[]   =   {   1   ,   1   ,   7   ,   15  ,   1   ,   25  ,   35  ,   243 ,   243 ,   491 ,   1031    ,   2905    ,   4293    ,   8387    ,0 };
            const unsigned long dim1827Kuo3Init[]   =   {   1   ,   1   ,   1   ,   15  ,   5   ,   1   ,   121 ,   71  ,   129 ,   333 ,   1121    ,   917 ,   6627    ,   3291    ,0 };
            const unsigned long dim1828Kuo3Init[]   =   {   1   ,   3   ,   5   ,   3   ,   7   ,   37  ,   13  ,   7   ,   115 ,   863 ,   345 ,   603 ,   1471    ,   15843   ,0 };
            const unsigned long dim1829Kuo3Init[]   =   {   1   ,   3   ,   7   ,   3   ,   3   ,   49  ,   29  ,   79  ,   185 ,   685 ,   1691    ,   3469    ,   7111    ,   915 ,0 };
            const unsigned long dim1830Kuo3Init[]   =   {   1   ,   1   ,   3   ,   1   ,   19  ,   61  ,   67  ,   143 ,   133 ,   457 ,   1865    ,   3403    ,   2557    ,   2603    ,0 };
            const unsigned long dim1831Kuo3Init[]   =   {   1   ,   1   ,   1   ,   15  ,   27  ,   25  ,   9   ,   87  ,   489 ,   23  ,   1251    ,   1995    ,   4737    ,   3099    ,0 };
            const unsigned long dim1832Kuo3Init[]   =   {   1   ,   1   ,   7   ,   11  ,   5   ,   43  ,   55  ,   85  ,   43  ,   621 ,   145 ,   1507    ,   4895    ,   687 ,0 };
            const unsigned long dim1833Kuo3Init[]   =   {   1   ,   3   ,   7   ,   15  ,   7   ,   35  ,   101 ,   19  ,   35  ,   609 ,   1433    ,   3877    ,   3661    ,   2517    ,0 };
            const unsigned long dim1834Kuo3Init[]   =   {   1   ,   3   ,   1   ,   5   ,   31  ,   47  ,   89  ,   225 ,   95  ,   941 ,   287 ,   1163    ,   6537    ,   10907   ,0 };
            const unsigned long dim1835Kuo3Init[]   =   {   1   ,   3   ,   7   ,   1   ,   15  ,   61  ,   65  ,   205 ,   413 ,   147 ,   833 ,   2097    ,   5465    ,   6309    ,0 };
            const unsigned long dim1836Kuo3Init[]   =   {   1   ,   3   ,   5   ,   11  ,   27  ,   55  ,   37  ,   185 ,   265 ,   343 ,   805 ,   325 ,   283 ,   1935    ,0 };
            const unsigned long dim1837Kuo3Init[]   =   {   1   ,   1   ,   3   ,   13  ,   29  ,   7   ,   55  ,   243 ,   101 ,   87  ,   2015    ,   2897    ,   2087    ,   47  ,0 };
            const unsigned long dim1838Kuo3Init[]   =   {   1   ,   3   ,   5   ,   11  ,   27  ,   1   ,   125 ,   65  ,   205 ,   525 ,   655 ,   3881    ,   993 ,   3967    ,0 };
            const unsigned long dim1839Kuo3Init[]   =   {   1   ,   3   ,   3   ,   9   ,   3   ,   61  ,   85  ,   31  ,   257 ,   257 ,   1227    ,   3329    ,   731 ,   7223    ,0 };
            const unsigned long dim1840Kuo3Init[]   =   {   1   ,   1   ,   7   ,   15  ,   29  ,   49  ,   49  ,   171 ,   95  ,   475 ,   601 ,   1325    ,   3995    ,   14421   ,0 };
            const unsigned long dim1841Kuo3Init[]   =   {   1   ,   1   ,   1   ,   11  ,   21  ,   23  ,   97  ,   65  ,   53  ,   417 ,   181 ,   551 ,   6583    ,   12835   ,0 };
            const unsigned long dim1842Kuo3Init[]   =   {   1   ,   3   ,   3   ,   1   ,   27  ,   37  ,   37  ,   143 ,   181 ,   59  ,   1011    ,   97  ,   2757    ,   15537   ,0 };
            const unsigned long dim1843Kuo3Init[]   =   {   1   ,   3   ,   5   ,   9   ,   5   ,   47  ,   121 ,   91  ,   297 ,   897 ,   813 ,   2563    ,   1779    ,   12675   ,0 };
            const unsigned long dim1844Kuo3Init[]   =   {   1   ,   3   ,   7   ,   9   ,   17  ,   61  ,   31  ,   47  ,   509 ,   55  ,   1401    ,   2251    ,   4327    ,   10409   ,0 };
            const unsigned long dim1845Kuo3Init[]   =   {   1   ,   3   ,   1   ,   3   ,   15  ,   29  ,   97  ,   37  ,   295 ,   271 ,   893 ,   1135    ,   1651    ,   9271    ,0 };
            const unsigned long dim1846Kuo3Init[]   =   {   1   ,   3   ,   5   ,   15  ,   29  ,   37  ,   65  ,   247 ,   247 ,   341 ,   865 ,   2163    ,   6827    ,   1849    ,0 };
            const unsigned long dim1847Kuo3Init[]   =   {   1   ,   3   ,   3   ,   11  ,   27  ,   27  ,   105 ,   67  ,   283 ,   221 ,   1445    ,   3283    ,   865 ,   3747    ,0 };
            const unsigned long dim1848Kuo3Init[]   =   {   1   ,   3   ,   5   ,   7   ,   7   ,   33  ,   63  ,   193 ,   219 ,   357 ,   1755    ,   2521    ,   5375    ,   10049   ,0 };
            const unsigned long dim1849Kuo3Init[]   =   {   1   ,   3   ,   5   ,   7   ,   17  ,   43  ,   113 ,   171 ,   237 ,   75  ,   1257    ,   251 ,   4681    ,   14793   ,0 };
            const unsigned long dim1850Kuo3Init[]   =   {   1   ,   3   ,   5   ,   11  ,   15  ,   15  ,   103 ,   163 ,   61  ,   263 ,   1165    ,   4021    ,   6833    ,   2675    ,0 };
            const unsigned long dim1851Kuo3Init[]   =   {   1   ,   3   ,   1   ,   5   ,   21  ,   63  ,   69  ,   29  ,   237 ,   473 ,   1457    ,   3545    ,   1457    ,   7595    ,0 };
            const unsigned long dim1852Kuo3Init[]   =   {   1   ,   3   ,   3   ,   1   ,   29  ,   33  ,   85  ,   103 ,   469 ,   579 ,   387 ,   2113    ,   5365    ,   7955    ,0 };
            const unsigned long dim1853Kuo3Init[]   =   {   1   ,   1   ,   3   ,   13  ,   7   ,   37  ,   99  ,   87  ,   147 ,   401 ,   125 ,   1233    ,   1463    ,   15067   ,0 };
            const unsigned long dim1854Kuo3Init[]   =   {   1   ,   1   ,   1   ,   3   ,   29  ,   11  ,   17  ,   89  ,   55  ,   273 ,   1069    ,   203 ,   941 ,   13667   ,0 };
            const unsigned long dim1855Kuo3Init[]   =   {   1   ,   3   ,   7   ,   3   ,   3   ,   53  ,   75  ,   163 ,   79  ,   63  ,   1851    ,   1495    ,   6537    ,   6329    ,0 };
            const unsigned long dim1856Kuo3Init[]   =   {   1   ,   3   ,   3   ,   3   ,   7   ,   53  ,   39  ,   35  ,   211 ,   275 ,   553 ,   1247    ,   3137    ,   11513   ,0 };
            const unsigned long dim1857Kuo3Init[]   =   {   1   ,   1   ,   3   ,   15  ,   25  ,   37  ,   15  ,   249 ,   79  ,   227 ,   1297    ,   277 ,   7693    ,   2431    ,0 };
            const unsigned long dim1858Kuo3Init[]   =   {   1   ,   3   ,   7   ,   3   ,   15  ,   53  ,   61  ,   27  ,   313 ,   817 ,   699 ,   3743    ,   6683    ,   11909   ,0 };
            const unsigned long dim1859Kuo3Init[]   =   {   1   ,   1   ,   3   ,   7   ,   9   ,   3   ,   77  ,   25  ,   421 ,   403 ,   1729    ,   1349    ,   5995    ,   7579    ,0 };
            const unsigned long dim1860Kuo3Init[]   =   {   1   ,   1   ,   7   ,   5   ,   29  ,   25  ,   1   ,   3   ,   69  ,   479 ,   699 ,   1317    ,   6729    ,   7701    ,0 };
            const unsigned long dim1861Kuo3Init[]   =   {   1   ,   1   ,   1   ,   9   ,   15  ,   51  ,   41  ,   227 ,   195 ,   101 ,   1783    ,   485 ,   1157    ,   1849    ,0 };
            const unsigned long dim1862Kuo3Init[]   =   {   1   ,   3   ,   3   ,   11  ,   1   ,   31  ,   111 ,   149 ,   361 ,   291 ,   1353    ,   3531    ,   473 ,   8545    ,0 };
            const unsigned long dim1863Kuo3Init[]   =   {   1   ,   1   ,   1   ,   3   ,   9   ,   35  ,   31  ,   165 ,   175 ,   617 ,   1063    ,   179 ,   957 ,   1395    ,0 };
            const unsigned long dim1864Kuo3Init[]   =   {   1   ,   1   ,   3   ,   1   ,   1   ,   5   ,   21  ,   191 ,   379 ,   293 ,   439 ,   1575    ,   243 ,   5013    ,0 };
            const unsigned long dim1865Kuo3Init[]   =   {   1   ,   3   ,   3   ,   1   ,   1   ,   21  ,   107 ,   91  ,   473 ,   809 ,   993 ,   2599    ,   8151    ,   10927   ,0 };
            const unsigned long dim1866Kuo3Init[]   =   {   1   ,   1   ,   1   ,   13  ,   13  ,   35  ,   71  ,   37  ,   173 ,   545 ,   1279    ,   225 ,   3275    ,   8823    ,0 };
            const unsigned long dim1867Kuo3Init[]   =   {   1   ,   3   ,   1   ,   7   ,   3   ,   21  ,   97  ,   23  ,   197 ,   535 ,   741 ,   3481    ,   919 ,   14675   ,   14019   ,0 };
            const unsigned long dim1868Kuo3Init[]   =   {   1   ,   3   ,   1   ,   3   ,   23  ,   25  ,   73  ,   203 ,   405 ,   287 ,   1167    ,   2313    ,   3065    ,   6385    ,   16509   ,0 };
            const unsigned long dim1869Kuo3Init[]   =   {   1   ,   1   ,   7   ,   15  ,   9   ,   23  ,   59  ,   241 ,   245 ,   865 ,   97  ,   931 ,   3317    ,   747 ,   14415   ,0 };
            const unsigned long dim1870Kuo3Init[]   =   {   1   ,   3   ,   7   ,   13  ,   7   ,   13  ,   73  ,   213 ,   345 ,   707 ,   1223    ,   517 ,   4427    ,   9539    ,   2669    ,0 };
            const unsigned long dim1871Kuo3Init[]   =   {   1   ,   1   ,   5   ,   15  ,   29  ,   41  ,   75  ,   149 ,   445 ,   933 ,   1727    ,   1967    ,   1927    ,   1205    ,   30051   ,0 };
            const unsigned long dim1872Kuo3Init[]   =   {   1   ,   1   ,   7   ,   9   ,   17  ,   57  ,   119 ,   115 ,   39  ,   957 ,   309 ,   3197    ,   6849    ,   7151    ,   1771    ,0 };
            const unsigned long dim1873Kuo3Init[]   =   {   1   ,   1   ,   3   ,   11  ,   19  ,   57  ,   25  ,   89  ,   223 ,   921 ,   1741    ,   2695    ,   3571    ,   11933   ,   4973    ,0 };
            const unsigned long dim1874Kuo3Init[]   =   {   1   ,   1   ,   3   ,   3   ,   29  ,   37  ,   77  ,   239 ,   63  ,   231 ,   639 ,   1311    ,   725 ,   5699    ,   15709   ,0 };
            const unsigned long dim1875Kuo3Init[]   =   {   1   ,   3   ,   5   ,   5   ,   31  ,   23  ,   53  ,   27  ,   353 ,   851 ,   685 ,   169 ,   5325    ,   13675   ,   21493   ,0 };
            const unsigned long dim1876Kuo3Init[]   =   {   1   ,   3   ,   5   ,   9   ,   31  ,   13  ,   37  ,   7   ,   51  ,   211 ,   1807    ,   3107    ,   6485    ,   12595   ,   1097    ,0 };
            const unsigned long dim1877Kuo3Init[]   =   {   1   ,   1   ,   7   ,   15  ,   9   ,   3   ,   33  ,   197 ,   357 ,   841 ,   1233    ,   2065    ,   5009    ,   5863    ,   11287   ,0 };
            const unsigned long dim1878Kuo3Init[]   =   {   1   ,   3   ,   7   ,   9   ,   27  ,   15  ,   75  ,   195 ,   223 ,   689 ,   1181    ,   1597    ,   2553    ,   7985    ,   13593   ,0 };
            const unsigned long dim1879Kuo3Init[]   =   {   1   ,   1   ,   5   ,   5   ,   21  ,   41  ,   97  ,   121 ,   293 ,   639 ,   1799    ,   689 ,   6941    ,   14179   ,   11007   ,0 };
            const unsigned long dim1880Kuo3Init[]   =   {   1   ,   1   ,   3   ,   7   ,   21  ,   61  ,   91  ,   177 ,   213 ,   757 ,   123 ,   2021    ,   1493    ,   2365    ,   9713    ,0 };
            const unsigned long dim1881Kuo3Init[]   =   {   1   ,   1   ,   1   ,   7   ,   3   ,   57  ,   33  ,   229 ,   45  ,   351 ,   1671    ,   2047    ,   7685    ,   14375   ,   5773    ,0 };
            const unsigned long dim1882Kuo3Init[]   =   {   1   ,   1   ,   7   ,   7   ,   23  ,   55  ,   19  ,   83  ,   11  ,   463 ,   83  ,   3071    ,   5257    ,   2741    ,   31137   ,0 };
            const unsigned long dim1883Kuo3Init[]   =   {   1   ,   3   ,   3   ,   5   ,   5   ,   31  ,   53  ,   135 ,   255 ,   711 ,   1873    ,   609 ,   3395    ,   16137   ,   19947   ,0 };
            const unsigned long dim1884Kuo3Init[]   =   {   1   ,   3   ,   7   ,   7   ,   13  ,   29  ,   117 ,   203 ,   387 ,   529 ,   1301    ,   3199    ,   1843    ,   5381    ,   12491   ,0 };
            const unsigned long dim1885Kuo3Init[]   =   {   1   ,   3   ,   1   ,   15  ,   25  ,   61  ,   21  ,   75  ,   111 ,   757 ,   1611    ,   2911    ,   5805    ,   5787    ,   5099    ,0 };
            const unsigned long dim1886Kuo3Init[]   =   {   1   ,   3   ,   3   ,   5   ,   1   ,   33  ,   51  ,   45  ,   115 ,   45  ,   1155    ,   951 ,   6471    ,   10233   ,   8053    ,0 };
            const unsigned long dim1887Kuo3Init[]   =   {   1   ,   3   ,   7   ,   1   ,   11  ,   13  ,   17  ,   167 ,   391 ,   219 ,   277 ,   3311    ,   1761    ,   6741    ,   6253    ,0 };
            const unsigned long dim1888Kuo3Init[]   =   {   1   ,   1   ,   3   ,   7   ,   3   ,   9   ,   43  ,   187 ,   113 ,   747 ,   1207    ,   3513    ,   1055    ,   15145   ,   3793    ,0 };
            const unsigned long dim1889Kuo3Init[]   =   {   1   ,   1   ,   7   ,   13  ,   3   ,   33  ,   127 ,   95  ,   47  ,   663 ,   1581    ,   489 ,   2303    ,   15183   ,   21671   ,0 };
            const unsigned long dim1890Kuo3Init[]   =   {   1   ,   3   ,   7   ,   9   ,   15  ,   7   ,   53  ,   143 ,   313 ,   167 ,   425 ,   3153    ,   7241    ,   3019    ,   27323   ,0 };
            const unsigned long dim1891Kuo3Init[]   =   {   1   ,   1   ,   5   ,   1   ,   1   ,   47  ,   47  ,   211 ,   79  ,   495 ,   1383    ,   2373    ,   141 ,   2933    ,   11535   ,0 };
            const unsigned long dim1892Kuo3Init[]   =   {   1   ,   3   ,   1   ,   11  ,   13  ,   35  ,   79  ,   105 ,   329 ,   235 ,   667 ,   2431    ,   6427    ,   15335   ,   15013   ,0 };
            const unsigned long dim1893Kuo3Init[]   =   {   1   ,   3   ,   3   ,   9   ,   25  ,   33  ,   27  ,   241 ,   153 ,   197 ,   221 ,   3149    ,   2357    ,   4485    ,   17877   ,0 };
            const unsigned long dim1894Kuo3Init[]   =   {   1   ,   1   ,   1   ,   7   ,   31  ,   29  ,   43  ,   39  ,   439 ,   799 ,   293 ,   727 ,   6387    ,   12257   ,   15701   ,0 };
            const unsigned long dim1895Kuo3Init[]   =   {   1   ,   3   ,   1   ,   1   ,   17  ,   47  ,   105 ,   33  ,   275 ,   269 ,   259 ,   2293    ,   1829    ,   13141   ,   3015    ,0 };
            const unsigned long dim1896Kuo3Init[]   =   {   1   ,   1   ,   3   ,   3   ,   29  ,   27  ,   43  ,   161 ,   131 ,   95  ,   851 ,   1269    ,   491 ,   4073    ,   3481    ,0 };
            const unsigned long dim1897Kuo3Init[]   =   {   1   ,   3   ,   7   ,   3   ,   23  ,   13  ,   35  ,   137 ,   405 ,   899 ,   1521    ,   2193    ,   725 ,   11967   ,   16937   ,0 };
            const unsigned long dim1898Kuo3Init[]   =   {   1   ,   3   ,   1   ,   15  ,   5   ,   15  ,   113 ,   139 ,   307 ,   125 ,   635 ,   477 ,   493 ,   9879    ,   21125   ,0 };
            const unsigned long dim1899Kuo3Init[]   =   {   1   ,   1   ,   7   ,   11  ,   19  ,   9   ,   117 ,   173 ,   447 ,   485 ,   1073    ,   2893    ,   1767    ,   10299   ,   31449   ,0 };
            const unsigned long dim1900Kuo3Init[]   =   {   1   ,   3   ,   7   ,   3   ,   11  ,   53  ,   99  ,   109 ,   393 ,   281 ,   187 ,   1557    ,   3597    ,   5215    ,   4033    ,0 };
            const unsigned long dim1901Kuo3Init[]   =   {   1   ,   1   ,   1   ,   3   ,   5   ,   15  ,   33  ,   33  ,   37  ,   445 ,   1415    ,   2885    ,   2625    ,   465 ,   2745    ,0 };
            const unsigned long dim1902Kuo3Init[]   =   {   1   ,   3   ,   1   ,   13  ,   7   ,   13  ,   93  ,   141 ,   427 ,   663 ,   1863    ,   991 ,   2001    ,   701 ,   18287   ,0 };
            const unsigned long dim1903Kuo3Init[]   =   {   1   ,   3   ,   1   ,   7   ,   19  ,   47  ,   65  ,   49  ,   425 ,   691 ,   1487    ,   2725    ,   6145    ,   4983    ,   11871   ,0 };
            const unsigned long dim1904Kuo3Init[]   =   {   1   ,   1   ,   7   ,   11  ,   27  ,   35  ,   89  ,   147 ,   413 ,   1007    ,   755 ,   3883    ,   4737    ,   10399   ,   19073   ,0 };
            const unsigned long dim1905Kuo3Init[]   =   {   1   ,   3   ,   3   ,   1   ,   11  ,   41  ,   27  ,   225 ,   237 ,   407 ,   461 ,   1881    ,   3761    ,   6587    ,   28923   ,0 };
            const unsigned long dim1906Kuo3Init[]   =   {   1   ,   3   ,   5   ,   9   ,   1   ,   3   ,   93  ,   179 ,   157 ,   227 ,   571 ,   1901    ,   1407    ,   12059   ,   15281   ,0 };
            const unsigned long dim1907Kuo3Init[]   =   {   1   ,   3   ,   1   ,   15  ,   27  ,   29  ,   55  ,   169 ,   65  ,   517 ,   1801    ,   177 ,   5687    ,   9597    ,   5167    ,0 };
            const unsigned long dim1908Kuo3Init[]   =   {   1   ,   1   ,   3   ,   5   ,   25  ,   47  ,   57  ,   15  ,   491 ,   631 ,   1567    ,   25  ,   2087    ,   7513    ,   15107   ,0 };
            const unsigned long dim1909Kuo3Init[]   =   {   1   ,   1   ,   1   ,   11  ,   15  ,   55  ,   13  ,   39  ,   203 ,   487 ,   1235    ,   389 ,   281 ,   499 ,   20119   ,0 };
            const unsigned long dim1910Kuo3Init[]   =   {   1   ,   3   ,   3   ,   15  ,   29  ,   53  ,   127 ,   211 ,   301 ,   509 ,   611 ,   1293    ,   6495    ,   10211   ,   4905    ,0 };
            const unsigned long dim1911Kuo3Init[]   =   {   1   ,   3   ,   5   ,   3   ,   13  ,   15  ,   31  ,   205 ,   237 ,   429 ,   1721    ,   3953    ,   8153    ,   16275   ,   30421   ,0 };
            const unsigned long dim1912Kuo3Init[]   =   {   1   ,   3   ,   7   ,   11  ,   27  ,   17  ,   43  ,   123 ,   503 ,   507 ,   257 ,   447 ,   6315    ,   5023    ,   6995    ,0 };
            const unsigned long dim1913Kuo3Init[]   =   {   1   ,   3   ,   5   ,   15  ,   27  ,   33  ,   127 ,   33  ,   241 ,   1001    ,   1009    ,   1263    ,   4627    ,   9329    ,   29191   ,0 };
            const unsigned long dim1914Kuo3Init[]   =   {   1   ,   1   ,   1   ,   13  ,   13  ,   45  ,   125 ,   77  ,   247 ,   123 ,   1581    ,   537 ,   799 ,   15417   ,   30457   ,0 };
            const unsigned long dim1915Kuo3Init[]   =   {   1   ,   3   ,   5   ,   5   ,   9   ,   29  ,   39  ,   211 ,   243 ,   871 ,   1581    ,   4029    ,   4791    ,   13393   ,   25341   ,0 };
            const unsigned long dim1916Kuo3Init[]   =   {   1   ,   3   ,   1   ,   5   ,   15  ,   17  ,   123 ,   123 ,   303 ,   207 ,   1551    ,   3633    ,   5591    ,   2725    ,   5707    ,0 };
            const unsigned long dim1917Kuo3Init[]   =   {   1   ,   1   ,   7   ,   11  ,   29  ,   41  ,   111 ,   73  ,   163 ,   511 ,   1305    ,   3235    ,   4725    ,   13589   ,   9611    ,0 };
            const unsigned long dim1918Kuo3Init[]   =   {   1   ,   1   ,   1   ,   1   ,   19  ,   51  ,   17  ,   225 ,   499 ,   1013    ,   7   ,   1793    ,   7073    ,   10115   ,   18195   ,0 };
            const unsigned long dim1919Kuo3Init[]   =   {   1   ,   1   ,   7   ,   9   ,   17  ,   35  ,   107 ,   87  ,   251 ,   19  ,   959 ,   1389    ,   6725    ,   8555    ,   14571   ,0 };
            const unsigned long dim1920Kuo3Init[]   =   {   1   ,   1   ,   1   ,   9   ,   1   ,   13  ,   79  ,   163 ,   17  ,   87  ,   1121    ,   1813    ,   1213    ,   3325    ,   18083   ,0 };
            const unsigned long dim1921Kuo3Init[]   =   {   1   ,   3   ,   1   ,   3   ,   27  ,   5   ,   107 ,   143 ,   211 ,   459 ,   257 ,   957 ,   1183    ,   2367    ,   24687   ,0 };
            const unsigned long dim1922Kuo3Init[]   =   {   1   ,   1   ,   7   ,   15  ,   1   ,   33  ,   51  ,   25  ,   379 ,   829 ,   1657    ,   273 ,   4507    ,   9609    ,   879 ,0 };
            const unsigned long dim1923Kuo3Init[]   =   {   1   ,   1   ,   7   ,   15  ,   1   ,   31  ,   43  ,   181 ,   147 ,   373 ,   83  ,   2179    ,   6363    ,   1871    ,   10209   ,0 };
            const unsigned long dim1924Kuo3Init[]   =   {   1   ,   1   ,   1   ,   15  ,   9   ,   29  ,   105 ,   255 ,   483 ,   573 ,   1841    ,   1101    ,   2987    ,   1619    ,   18837   ,0 };
            const unsigned long dim1925Kuo3Init[]   =   {   1   ,   3   ,   5   ,   11  ,   17  ,   19  ,   67  ,   109 ,   13  ,   247 ,   1365    ,   2153    ,   5597    ,   5785    ,   25877   ,0 };
            const unsigned long dim1926Kuo3Init[]   =   {   1   ,   1   ,   3   ,   7   ,   17  ,   49  ,   121 ,   163 ,   49  ,   689 ,   1917    ,   2161    ,   7979    ,   8153    ,   12889   ,0 };
            const unsigned long dim1927Kuo3Init[]   =   {   1   ,   1   ,   3   ,   7   ,   5   ,   49  ,   39  ,   251 ,   311 ,   69  ,   553 ,   2789    ,   2415    ,   6521    ,   26847   ,0 };
            const unsigned long dim1928Kuo3Init[]   =   {   1   ,   1   ,   3   ,   5   ,   15  ,   51  ,   67  ,   227 ,   245 ,   893 ,   493 ,   1191    ,   3061    ,   249 ,   16603   ,0 };
            const unsigned long dim1929Kuo3Init[]   =   {   1   ,   3   ,   1   ,   3   ,   17  ,   63  ,   35  ,   41  ,   269 ,   75  ,   147 ,   3229    ,   7985    ,   5789    ,   6365    ,0 };
            const unsigned long dim1930Kuo3Init[]   =   {   1   ,   3   ,   1   ,   13  ,   21  ,   33  ,   125 ,   39  ,   343 ,   133 ,   389 ,   2467    ,   4079    ,   14315   ,   17047   ,0 };
            const unsigned long dim1931Kuo3Init[]   =   {   1   ,   3   ,   1   ,   7   ,   27  ,   61  ,   73  ,   107 ,   143 ,   665 ,   155 ,   2865    ,   3017    ,   7307    ,   8771    ,0 };
            const unsigned long dim1932Kuo3Init[]   =   {   1   ,   1   ,   5   ,   7   ,   25  ,   57  ,   57  ,   39  ,   195 ,   405 ,   1863    ,   2985    ,   5775    ,   9609    ,   24409   ,0 };
            const unsigned long dim1933Kuo3Init[]   =   {   1   ,   3   ,   3   ,   5   ,   11  ,   27  ,   91  ,   169 ,   343 ,   673 ,   1317    ,   1763    ,   1603    ,   527 ,   18945   ,0 };
            const unsigned long dim1934Kuo3Init[]   =   {   1   ,   3   ,   7   ,   7   ,   17  ,   39  ,   113 ,   47  ,   483 ,   793 ,   1295    ,   1305    ,   7025    ,   14849   ,   17703   ,0 };
            const unsigned long dim1935Kuo3Init[]   =   {   1   ,   1   ,   5   ,   9   ,   5   ,   43  ,   119 ,   149 ,   103 ,   63  ,   1369    ,   3791    ,   3081    ,   10587   ,   12711   ,0 };
            const unsigned long dim1936Kuo3Init[]   =   {   1   ,   3   ,   5   ,   15  ,   9   ,   55  ,   105 ,   171 ,   199 ,   883 ,   1113    ,   3559    ,   1979    ,   8823    ,   12407   ,0 };
            const unsigned long dim1937Kuo3Init[]   =   {   1   ,   3   ,   3   ,   9   ,   27  ,   51  ,   105 ,   163 ,   203 ,   25  ,   1579    ,   351 ,   8071    ,   5921    ,   28935   ,0 };
            const unsigned long dim1938Kuo3Init[]   =   {   1   ,   1   ,   5   ,   15  ,   15  ,   43  ,   115 ,   115 ,   5   ,   283 ,   1935    ,   2619    ,   2543    ,   7603    ,   2877    ,0 };
            const unsigned long dim1939Kuo3Init[]   =   {   1   ,   3   ,   7   ,   9   ,   31  ,   29  ,   17  ,   67  ,   121 ,   187 ,   563 ,   377 ,   3947    ,   7699    ,   8677    ,0 };
            const unsigned long dim1940Kuo3Init[]   =   {   1   ,   1   ,   5   ,   1   ,   5   ,   1   ,   121 ,   211 ,   109 ,   955 ,   1605    ,   3681    ,   5395    ,   797 ,   19243   ,0 };
            const unsigned long dim1941Kuo3Init[]   =   {   1   ,   1   ,   7   ,   13  ,   7   ,   63  ,   73  ,   215 ,   403 ,   465 ,   1317    ,   3141    ,   4511    ,   12579   ,   20839   ,0 };
            const unsigned long dim1942Kuo3Init[]   =   {   1   ,   3   ,   5   ,   7   ,   9   ,   63  ,   39  ,   55  ,   353 ,   395 ,   233 ,   1189    ,   1781    ,   9167    ,   887 ,0 };
            const unsigned long dim1943Kuo3Init[]   =   {   1   ,   1   ,   7   ,   1   ,   1   ,   55  ,   93  ,   63  ,   211 ,   635 ,   1767    ,   827 ,   7257    ,   119 ,   26333   ,0 };
            const unsigned long dim1944Kuo3Init[]   =   {   1   ,   1   ,   3   ,   11  ,   9   ,   33  ,   17  ,   205 ,   303 ,   713 ,   65  ,   3539    ,   5069    ,   8895    ,   3843    ,0 };
            const unsigned long dim1945Kuo3Init[]   =   {   1   ,   1   ,   3   ,   15  ,   3   ,   59  ,   45  ,   135 ,   103 ,   293 ,   731 ,   359 ,   3631    ,   7967    ,   30043   ,0 };
            const unsigned long dim1946Kuo3Init[]   =   {   1   ,   3   ,   3   ,   7   ,   25  ,   7   ,   27  ,   161 ,   9   ,   509 ,   1345    ,   3759    ,   7957    ,   5975    ,   17981   ,0 };
            const unsigned long dim1947Kuo3Init[]   =   {   1   ,   3   ,   1   ,   7   ,   13  ,   41  ,   105 ,   241 ,   331 ,   437 ,   2027    ,   2589    ,   5201    ,   11373   ,   22581   ,0 };
            const unsigned long dim1948Kuo3Init[]   =   {   1   ,   3   ,   7   ,   13  ,   5   ,   3   ,   37  ,   43  ,   125 ,   521 ,   1059    ,   291 ,   4421    ,   5215    ,   16835   ,0 };
            const unsigned long dim1949Kuo3Init[]   =   {   1   ,   1   ,   5   ,   15  ,   13  ,   15  ,   85  ,   175 ,   221 ,   527 ,   173 ,   1333    ,   7505    ,   883 ,   2215    ,0 };
            const unsigned long dim1950Kuo3Init[]   =   {   1   ,   3   ,   5   ,   11  ,   27  ,   63  ,   5   ,   215 ,   85  ,   229 ,   165 ,   525 ,   2693    ,   15089   ,   4907    ,0 };
            const unsigned long dim1951Kuo3Init[]   =   {   1   ,   3   ,   7   ,   15  ,   25  ,   13  ,   91  ,   81  ,   81  ,   189 ,   671 ,   539 ,   2393    ,   9627    ,   17667   ,0 };
            const unsigned long dim1952Kuo3Init[]   =   {   1   ,   1   ,   1   ,   5   ,   5   ,   9   ,   45  ,   43  ,   337 ,   841 ,   853 ,   3969    ,   5917    ,   5837    ,   17265   ,0 };
            const unsigned long dim1953Kuo3Init[]   =   {   1   ,   1   ,   7   ,   9   ,   1   ,   57  ,   33  ,   29  ,   235 ,   69  ,   815 ,   147 ,   7251    ,   2933    ,   32405   ,0 };
            const unsigned long dim1954Kuo3Init[]   =   {   1   ,   3   ,   1   ,   3   ,   29  ,   1   ,   3   ,   13  ,   483 ,   825 ,   31  ,   2289    ,   185 ,   9075    ,   15495   ,0 };
            const unsigned long dim1955Kuo3Init[]   =   {   1   ,   3   ,   3   ,   3   ,   1   ,   33  ,   127 ,   53  ,   229 ,   975 ,   91  ,   3259    ,   7969    ,   4193    ,   18433   ,0 };
            const unsigned long dim1956Kuo3Init[]   =   {   1   ,   1   ,   3   ,   7   ,   7   ,   43  ,   89  ,   193 ,   51  ,   563 ,   1961    ,   3959    ,   1901    ,   10393   ,   679 ,0 };
            const unsigned long dim1957Kuo3Init[]   =   {   1   ,   1   ,   5   ,   9   ,   29  ,   11  ,   57  ,   227 ,   163 ,   169 ,   1969    ,   3087    ,   3221    ,   15499   ,   12255   ,0 };
            const unsigned long dim1958Kuo3Init[]   =   {   1   ,   3   ,   7   ,   5   ,   31  ,   47  ,   45  ,   37  ,   257 ,   383 ,   1347    ,   3923    ,   7963    ,   13459   ,   8755    ,0 };
            const unsigned long dim1959Kuo3Init[]   =   {   1   ,   1   ,   1   ,   3   ,   5   ,   53  ,   19  ,   219 ,   261 ,   313 ,   1889    ,   3185    ,   5309    ,   14049   ,   5231    ,0 };
            const unsigned long dim1960Kuo3Init[]   =   {   1   ,   3   ,   3   ,   13  ,   29  ,   19  ,   49  ,   221 ,   443 ,   125 ,   929 ,   2445    ,   6713    ,   5639    ,   2657    ,0 };
            const unsigned long dim1961Kuo3Init[]   =   {   1   ,   1   ,   7   ,   3   ,   17  ,   59  ,   87  ,   227 ,   305 ,   67  ,   311 ,   3473    ,   4785    ,   10761   ,   2871    ,0 };
            const unsigned long dim1962Kuo3Init[]   =   {   1   ,   1   ,   5   ,   7   ,   31  ,   63  ,   103 ,   127 ,   49  ,   839 ,   1991    ,   41  ,   5019    ,   12779   ,   25489   ,0 };
            const unsigned long dim1963Kuo3Init[]   =   {   1   ,   3   ,   5   ,   9   ,   9   ,   45  ,   127 ,   29  ,   441 ,   481 ,   49  ,   1975    ,   1523    ,   759 ,   847 ,0 };
            const unsigned long dim1964Kuo3Init[]   =   {   1   ,   3   ,   5   ,   13  ,   1   ,   25  ,   89  ,   19  ,   53  ,   17  ,   1449    ,   603 ,   473 ,   9965    ,   14455   ,0 };
            const unsigned long dim1965Kuo3Init[]   =   {   1   ,   3   ,   1   ,   7   ,   19  ,   9   ,   27  ,   97  ,   321 ,   1007    ,   1283    ,   3805    ,   539 ,   2979    ,   4631    ,0 };
            const unsigned long dim1966Kuo3Init[]   =   {   1   ,   3   ,   1   ,   5   ,   9   ,   3   ,   9   ,   237 ,   349 ,   277 ,   1791    ,   3325    ,   865 ,   4305    ,   29293   ,0 };
            const unsigned long dim1967Kuo3Init[]   =   {   1   ,   1   ,   1   ,   9   ,   13  ,   23  ,   63  ,   33  ,   381 ,   711 ,   1921    ,   1385    ,   5949    ,   8837    ,   3589    ,0 };
            const unsigned long dim1968Kuo3Init[]   =   {   1   ,   3   ,   1   ,   7   ,   15  ,   5   ,   7   ,   175 ,   15  ,   343 ,   1587    ,   2917    ,   4189    ,   14973   ,   1691    ,0 };
            const unsigned long dim1969Kuo3Init[]   =   {   1   ,   1   ,   3   ,   15  ,   17  ,   59  ,   123 ,   195 ,   377 ,   35  ,   917 ,   1045    ,   721 ,   9759    ,   15287   ,0 };
            const unsigned long dim1970Kuo3Init[]   =   {   1   ,   1   ,   3   ,   15  ,   29  ,   3   ,   15  ,   147 ,   283 ,   449 ,   175 ,   1771    ,   2359    ,   7541    ,   11657   ,0 };
            const unsigned long dim1971Kuo3Init[]   =   {   1   ,   3   ,   1   ,   3   ,   1   ,   11  ,   45  ,   117 ,   235 ,   75  ,   247 ,   917 ,   6135    ,   13817   ,   27365   ,0 };
            const unsigned long dim1972Kuo3Init[]   =   {   1   ,   1   ,   7   ,   3   ,   1   ,   45  ,   31  ,   97  ,   281 ,   739 ,   829 ,   1837    ,   1709    ,   877 ,   25275   ,0 };
            const unsigned long dim1973Kuo3Init[]   =   {   1   ,   3   ,   7   ,   11  ,   3   ,   35  ,   11  ,   55  ,   233 ,   471 ,   1239    ,   1345    ,   2475    ,   13391   ,   25447   ,0 };
            const unsigned long dim1974Kuo3Init[]   =   {   1   ,   1   ,   5   ,   7   ,   27  ,   33  ,   111 ,   89  ,   173 ,   361 ,   1641    ,   381 ,   1629    ,   5663    ,   17715   ,0 };
            const unsigned long dim1975Kuo3Init[]   =   {   1   ,   3   ,   1   ,   1   ,   31  ,   59  ,   11  ,   165 ,   129 ,   513 ,   1729    ,   2707    ,   4665    ,   3381    ,   7915    ,0 };
            const unsigned long dim1976Kuo3Init[]   =   {   1   ,   3   ,   1   ,   11  ,   19  ,   15  ,   3   ,   99  ,   21  ,   255 ,   515 ,   1619    ,   1819    ,   4143    ,   24317   ,0 };
            const unsigned long dim1977Kuo3Init[]   =   {   1   ,   3   ,   1   ,   9   ,   11  ,   33  ,   13  ,   143 ,   39  ,   557 ,   901 ,   187 ,   2963    ,   13413   ,   5641    ,0 };
            const unsigned long dim1978Kuo3Init[]   =   {   1   ,   3   ,   3   ,   13  ,   5   ,   27  ,   87  ,   63  ,   41  ,   57  ,   2009    ,   493 ,   2565    ,   14931   ,   24071   ,0 };
            const unsigned long dim1979Kuo3Init[]   =   {   1   ,   3   ,   1   ,   1   ,   7   ,   11  ,   31  ,   107 ,   343 ,   395 ,   841 ,   467 ,   1019    ,   3861    ,   19595   ,0 };
            const unsigned long dim1980Kuo3Init[]   =   {   1   ,   1   ,   7   ,   13  ,   5   ,   15  ,   113 ,   23  ,   421 ,   693 ,   1421    ,   2213    ,   3589    ,   3713    ,   1353    ,0 };
            const unsigned long dim1981Kuo3Init[]   =   {   1   ,   3   ,   1   ,   7   ,   17  ,   1   ,   87  ,   217 ,   333 ,   797 ,   1971    ,   917 ,   3527    ,   10061   ,   19353   ,0 };
            const unsigned long dim1982Kuo3Init[]   =   {   1   ,   3   ,   1   ,   5   ,   5   ,   51  ,   17  ,   239 ,   275 ,   221 ,   1621    ,   215 ,   319 ,   12041   ,   15669   ,0 };
            const unsigned long dim1983Kuo3Init[]   =   {   1   ,   1   ,   7   ,   11  ,   27  ,   33  ,   31  ,   19  ,   365 ,   951 ,   1453    ,   141 ,   1249    ,   16313   ,   30607   ,0 };
            const unsigned long dim1984Kuo3Init[]   =   {   1   ,   3   ,   1   ,   9   ,   15  ,   19  ,   127 ,   247 ,   161 ,   921 ,   195 ,   2857    ,   1573    ,   11077   ,   15997   ,0 };
            const unsigned long dim1985Kuo3Init[]   =   {   1   ,   1   ,   1   ,   1   ,   9   ,   43  ,   57  ,   103 ,   215 ,   337 ,   1705    ,   961 ,   6895    ,   2671    ,   7157    ,0 };
            const unsigned long dim1986Kuo3Init[]   =   {   1   ,   1   ,   7   ,   3   ,   13  ,   55  ,   123 ,   19  ,   279 ,   571 ,   1039    ,   2545    ,   5961    ,   12373   ,   11479   ,0 };
            const unsigned long dim1987Kuo3Init[]   =   {   1   ,   3   ,   7   ,   11  ,   7   ,   35  ,   85  ,   89  ,   407 ,   359 ,   1897    ,   231 ,   7653    ,   9215    ,   9957    ,0 };
            const unsigned long dim1988Kuo3Init[]   =   {   1   ,   3   ,   7   ,   5   ,   23  ,   53  ,   13  ,   153 ,   221 ,   363 ,   1481    ,   1069    ,   1023    ,   8573    ,   29385   ,0 };
            const unsigned long dim1989Kuo3Init[]   =   {   1   ,   1   ,   1   ,   3   ,   15  ,   63  ,   1   ,   57  ,   373 ,   35  ,   725 ,   135 ,   5593    ,   5593    ,   8365    ,0 };
            const unsigned long dim1990Kuo3Init[]   =   {   1   ,   1   ,   5   ,   3   ,   7   ,   35  ,   25  ,   69  ,   441 ,   451 ,   1109    ,   1723    ,   1617    ,   3947    ,   22349   ,0 };
            const unsigned long dim1991Kuo3Init[]   =   {   1   ,   3   ,   5   ,   3   ,   27  ,   55  ,   75  ,   231 ,   245 ,   831 ,   1641    ,   1475    ,   51  ,   2683    ,   12489   ,0 };
            const unsigned long dim1992Kuo3Init[]   =   {   1   ,   1   ,   5   ,   5   ,   17  ,   5   ,   61  ,   167 ,   389 ,   149 ,   43  ,   825 ,   6079    ,   12955   ,   20121   ,0 };
            const unsigned long dim1993Kuo3Init[]   =   {   1   ,   1   ,   5   ,   7   ,   15  ,   43  ,   63  ,   83  ,   187 ,   1005    ,   1675    ,   2907    ,   2357    ,   5459    ,   18621   ,0 };
            const unsigned long dim1994Kuo3Init[]   =   {   1   ,   1   ,   5   ,   13  ,   27  ,   53  ,   49  ,   93  ,   9   ,   629 ,   1507    ,   357 ,   3207    ,   9209    ,   13697   ,0 };
            const unsigned long dim1995Kuo3Init[]   =   {   1   ,   3   ,   5   ,   3   ,   19  ,   49  ,   65  ,   31  ,   359 ,   13  ,   617 ,   815 ,   4673    ,   2475    ,   27491   ,0 };
            const unsigned long dim1996Kuo3Init[]   =   {   1   ,   3   ,   3   ,   13  ,   21  ,   39  ,   79  ,   107 ,   485 ,   905 ,   51  ,   987 ,   6579    ,   3899    ,   28785   ,0 };
            const unsigned long dim1997Kuo3Init[]   =   {   1   ,   1   ,   1   ,   11  ,   7   ,   11  ,   111 ,   223 ,   249 ,   767 ,   141 ,   691 ,   4711    ,   583 ,   7917    ,0 };
            const unsigned long dim1998Kuo3Init[]   =   {   1   ,   1   ,   5   ,   1   ,   9   ,   43  ,   115 ,   179 ,   313 ,   783 ,   2047    ,   1773    ,   2307    ,   5011    ,   7215    ,0 };
            const unsigned long dim1999Kuo3Init[]   =   {   1   ,   3   ,   3   ,   5   ,   23  ,   23  ,   85  ,   193 ,   83  ,   809 ,   1827    ,   3375    ,   5177    ,   2041    ,   19271   ,0 };
            const unsigned long dim2000Kuo3Init[]   =   {   1   ,   1   ,   1   ,   9   ,   23  ,   17  ,   31  ,   111 ,   75  ,   265 ,   1171    ,   219 ,   6309    ,   2263    ,   3807    ,0 };
            const unsigned long dim2001Kuo3Init[]   =   {   1   ,   1   ,   5   ,   1   ,   27  ,   7   ,   65  ,   117 ,   291 ,   811 ,   1797    ,   3395    ,   3495    ,   829 ,   23911   ,0 };
            const unsigned long dim2002Kuo3Init[]   =   {   1   ,   1   ,   3   ,   11  ,   19  ,   31  ,   43  ,   229 ,   361 ,   219 ,   1029    ,   1407    ,   5855    ,   12411   ,   16117   ,0 };
            const unsigned long dim2003Kuo3Init[]   =   {   1   ,   3   ,   3   ,   3   ,   5   ,   1   ,   99  ,   245 ,   21  ,   925 ,   1807    ,   2529    ,   627 ,   3363    ,   853 ,0 };
            const unsigned long dim2004Kuo3Init[]   =   {   1   ,   1   ,   3   ,   15  ,   17  ,   41  ,   83  ,   31  ,   207 ,   615 ,   851 ,   1631    ,   6789    ,   12061   ,   12327   ,0 };
            const unsigned long dim2005Kuo3Init[]   =   {   1   ,   1   ,   7   ,   15  ,   27  ,   19  ,   75  ,   3   ,   51  ,   655 ,   73  ,   2433    ,   7973    ,   1425    ,   31309   ,0 };
            const unsigned long dim2006Kuo3Init[]   =   {   1   ,   3   ,   1   ,   7   ,   25  ,   47  ,   73  ,   119 ,   91  ,   981 ,   1671    ,   2145    ,   3101    ,   3765    ,   8719    ,0 };
            const unsigned long dim2007Kuo3Init[]   =   {   1   ,   1   ,   1   ,   7   ,   31  ,   37  ,   51  ,   181 ,   289 ,   273 ,   1889    ,   1017    ,   2793    ,   9513    ,   32549   ,0 };
            const unsigned long dim2008Kuo3Init[]   =   {   1   ,   1   ,   3   ,   5   ,   5   ,   39  ,   57  ,   51  ,   135 ,   689 ,   1913    ,   709 ,   4759    ,   12223   ,   17377   ,0 };
            const unsigned long dim2009Kuo3Init[]   =   {   1   ,   3   ,   3   ,   15  ,   31  ,   9   ,   57  ,   245 ,   159 ,   97  ,   1485    ,   2229    ,   1393    ,   9699    ,   5807    ,0 };
            const unsigned long dim2010Kuo3Init[]   =   {   1   ,   1   ,   3   ,   9   ,   11  ,   3   ,   11  ,   169 ,   81  ,   667 ,   357 ,   2307    ,   7305    ,   8631    ,   2555    ,0 };
            const unsigned long dim2011Kuo3Init[]   =   {   1   ,   3   ,   7   ,   7   ,   15  ,   17  ,   119 ,   143 ,   15  ,   567 ,   1879    ,   1363    ,   1617    ,   8231    ,   23071   ,0 };
            const unsigned long dim2012Kuo3Init[]   =   {   1   ,   1   ,   5   ,   9   ,   5   ,   39  ,   49  ,   45  ,   371 ,   641 ,   943 ,   1231    ,   8139    ,   14423   ,   11057   ,0 };
            const unsigned long dim2013Kuo3Init[]   =   {   1   ,   1   ,   5   ,   1   ,   27  ,   51  ,   25  ,   111 ,   205 ,   655 ,   317 ,   1381    ,   2359    ,   8785    ,   31503   ,0 };
            const unsigned long dim2014Kuo3Init[]   =   {   1   ,   3   ,   3   ,   3   ,   9   ,   13  ,   127 ,   211 ,   423 ,   147 ,   735 ,   1323    ,   5801    ,   9291    ,   10059   ,0 };
            const unsigned long dim2015Kuo3Init[]   =   {   1   ,   3   ,   3   ,   13  ,   19  ,   1   ,   11  ,   85  ,   343 ,   781 ,   1661    ,   1803    ,   4541    ,   2271    ,   20541   ,0 };
            const unsigned long dim2016Kuo3Init[]   =   {   1   ,   1   ,   5   ,   9   ,   11  ,   15  ,   77  ,   63  ,   345 ,   71  ,   1311    ,   3809    ,   7535    ,   1269    ,   13949   ,0 };
            const unsigned long dim2017Kuo3Init[]   =   {   1   ,   3   ,   7   ,   7   ,   19  ,   27  ,   3   ,   107 ,   439 ,   195 ,   2007    ,   4009    ,   305 ,   3111    ,   17775   ,0 };
            const unsigned long dim2018Kuo3Init[]   =   {   1   ,   1   ,   3   ,   7   ,   27  ,   31  ,   25  ,   123 ,   177 ,   161 ,   287 ,   3389    ,   3013    ,   11427   ,   9165    ,0 };
            const unsigned long dim2019Kuo3Init[]   =   {   1   ,   1   ,   3   ,   7   ,   27  ,   15  ,   83  ,   131 ,   201 ,   277 ,   267 ,   3557    ,   1337    ,   4427    ,   2389    ,0 };
            const unsigned long dim2020Kuo3Init[]   =   {   1   ,   1   ,   5   ,   7   ,   25  ,   27  ,   75  ,   211 ,   147 ,   37  ,   491 ,   3467    ,   4525    ,   13073   ,   15289   ,0 };
            const unsigned long dim2021Kuo3Init[]   =   {   1   ,   3   ,   5   ,   9   ,   13  ,   37  ,   117 ,   159 ,   145 ,   453 ,   1451    ,   667 ,   1861    ,   6855    ,   25389   ,0 };
            const unsigned long dim2022Kuo3Init[]   =   {   1   ,   1   ,   5   ,   7   ,   19  ,   37  ,   3   ,   157 ,   251 ,   901 ,   433 ,   3289    ,   2723    ,   11941   ,   10103   ,0 };
            const unsigned long dim2023Kuo3Init[]   =   {   1   ,   3   ,   7   ,   3   ,   31  ,   51  ,   125 ,   211 ,   317 ,   775 ,   801 ,   3887    ,   2257    ,   13947   ,   2421    ,0 };
            const unsigned long dim2024Kuo3Init[]   =   {   1   ,   3   ,   3   ,   13  ,   11  ,   21  ,   67  ,   223 ,   203 ,   409 ,   1927    ,   3565    ,   679 ,   14197   ,   24987   ,0 };
            const unsigned long dim2025Kuo3Init[]   =   {   1   ,   3   ,   1   ,   1   ,   23  ,   29  ,   37  ,   203 ,   311 ,   825 ,   1957    ,   3069    ,   5295    ,   4073    ,   4407    ,0 };
            const unsigned long dim2026Kuo3Init[]   =   {   1   ,   3   ,   1   ,   15  ,   1   ,   49  ,   111 ,   205 ,   441 ,   907 ,   481 ,   225 ,   6589    ,   7523    ,   19807   ,0 };
            const unsigned long dim2027Kuo3Init[]   =   {   1   ,   3   ,   5   ,   15  ,   1   ,   55  ,   43  ,   121 ,   9   ,   239 ,   723 ,   3109    ,   5639    ,   9547    ,   20455   ,0 };
            const unsigned long dim2028Kuo3Init[]   =   {   1   ,   1   ,   1   ,   15  ,   29  ,   17  ,   121 ,   61  ,   233 ,   873 ,   1447    ,   1313    ,   5073    ,   12259   ,   32305   ,0 };
            const unsigned long dim2029Kuo3Init[]   =   {   1   ,   1   ,   5   ,   7   ,   25  ,   57  ,   15  ,   45  ,   243 ,   297 ,   219 ,   2307    ,   1459    ,   16005   ,   10041   ,0 };
            const unsigned long dim2030Kuo3Init[]   =   {   1   ,   3   ,   7   ,   3   ,   5   ,   53  ,   57  ,   187 ,   287 ,   243 ,   1223    ,   1625    ,   669 ,   5165    ,   8395    ,0 };
            const unsigned long dim2031Kuo3Init[]   =   {   1   ,   3   ,   7   ,   15  ,   23  ,   23  ,   25  ,   225 ,   471 ,   661 ,   905 ,   2293    ,   1131    ,   8469    ,   7929    ,0 };
            const unsigned long dim2032Kuo3Init[]   =   {   1   ,   1   ,   5   ,   11  ,   11  ,   33  ,   11  ,   129 ,   237 ,   855 ,   1417    ,   3661    ,   1561    ,   9635    ,   31905   ,0 };
            const unsigned long dim2033Kuo3Init[]   =   {   1   ,   3   ,   1   ,   3   ,   15  ,   45  ,   3   ,   253 ,   125 ,   449 ,   1675    ,   1707    ,   3917    ,   11829   ,   27439   ,0 };
            const unsigned long dim2034Kuo3Init[]   =   {   1   ,   3   ,   7   ,   1   ,   1   ,   55  ,   83  ,   69  ,   83  ,   745 ,   1225    ,   977 ,   7117    ,   15045   ,   28067   ,0 };
            const unsigned long dim2035Kuo3Init[]   =   {   1   ,   1   ,   7   ,   9   ,   3   ,   51  ,   23  ,   205 ,   5   ,   625 ,   531 ,   3557    ,   4101    ,   5709    ,   29487   ,0 };
            const unsigned long dim2036Kuo3Init[]   =   {   1   ,   1   ,   1   ,   5   ,   9   ,   23  ,   69  ,   99  ,   203 ,   355 ,   1935    ,   3165    ,   5701    ,   15491   ,   18151   ,0 };
            const unsigned long dim2037Kuo3Init[]   =   {   1   ,   1   ,   1   ,   7   ,   31  ,   41  ,   89  ,   85  ,   273 ,   227 ,   785 ,   1057    ,   4409    ,   9155    ,   11573   ,0 };
            const unsigned long dim2038Kuo3Init[]   =   {   1   ,   3   ,   7   ,   15  ,   17  ,   49  ,   117 ,   31  ,   181 ,   799 ,   417 ,   3931    ,   6069    ,   1253    ,   25793   ,0 };
            const unsigned long dim2039Kuo3Init[]   =   {   1   ,   1   ,   3   ,   7   ,   27  ,   23  ,   33  ,   189 ,   5   ,   841 ,   901 ,   2023    ,   331 ,   10015   ,   3781    ,0 };
            const unsigned long dim2040Kuo3Init[]   =   {   1   ,   3   ,   3   ,   5   ,   19  ,   47  ,   73  ,   103 ,   373 ,   1011    ,   1015    ,   1657    ,   6181    ,   69  ,   16243   ,0 };
            const unsigned long dim2041Kuo3Init[]   =   {   1   ,   1   ,   7   ,   9   ,   13  ,   9   ,   95  ,   161 ,   413 ,   155 ,   367 ,   3519    ,   1557    ,   14897   ,   31331   ,0 };
            const unsigned long dim2042Kuo3Init[]   =   {   1   ,   1   ,   7   ,   3   ,   15  ,   25  ,   105 ,   145 ,   157 ,   525 ,   1541    ,   3361    ,   1531    ,   7473    ,   28187   ,0 };
            const unsigned long dim2043Kuo3Init[]   =   {   1   ,   1   ,   7   ,   7   ,   15  ,   7   ,   119 ,   135 ,   395 ,   441 ,   1443    ,   537 ,   5185    ,   793 ,   13205   ,0 };
            const unsigned long dim2044Kuo3Init[]   =   {   1   ,   1   ,   1   ,   9   ,   15  ,   7   ,   101 ,   43  ,   437 ,   761 ,   1471    ,   3843    ,   6765    ,   4135    ,   7883    ,0 };
            const unsigned long dim2045Kuo3Init[]   =   {   1   ,   1   ,   5   ,   5   ,   23  ,   31  ,   125 ,   119 ,   361 ,   65  ,   1907    ,   1335    ,   1309    ,   10591   ,   11167   ,0 };
            const unsigned long dim2046Kuo3Init[]   =   {   1   ,   3   ,   1   ,   7   ,   21  ,   9   ,   29  ,   149 ,   371 ,   201 ,   1531    ,   2955    ,   1755    ,   14047   ,   5867    ,0 };
            const unsigned long dim2047Kuo3Init[]   =   {   1   ,   1   ,   5   ,   1   ,   19  ,   9   ,   13  ,   13  ,   313 ,   621 ,   595 ,   2765    ,   2963    ,   13081   ,   21795   ,0 };
            const unsigned long dim2048Kuo3Init[]   =   {   1   ,   1   ,   3   ,   11  ,   21  ,   61  ,   7   ,   113 ,   295 ,   995 ,   1649    ,   1715    ,   4737    ,   4769    ,   13945   ,0 };
            const unsigned long dim2049Kuo3Init[]   =   {   1   ,   3   ,   5   ,   15  ,   17  ,   53  ,   73  ,   15  ,   83  ,   813 ,   1757    ,   471 ,   6031    ,   7871    ,   32569   ,0 };
            const unsigned long dim2050Kuo3Init[]   =   {   1   ,   3   ,   1   ,   13  ,   29  ,   55  ,   45  ,   173 ,   231 ,   167 ,   415 ,   1487    ,   6403    ,   13823   ,   10121   ,0 };
            const unsigned long dim2051Kuo3Init[]   =   {   1   ,   1   ,   1   ,   5   ,   19  ,   25  ,   107 ,   177 ,   97  ,   703 ,   1349    ,   2217    ,   3863    ,   11211   ,   28875   ,0 };
            const unsigned long dim2052Kuo3Init[]   =   {   1   ,   3   ,   7   ,   9   ,   1   ,   9   ,   37  ,   127 ,   435 ,   917 ,   709 ,   2227    ,   7449    ,   13927   ,   1187    ,0 };
            const unsigned long dim2053Kuo3Init[]   =   {   1   ,   1   ,   3   ,   15  ,   19  ,   17  ,   19  ,   127 ,   343 ,   235 ,   931 ,   589 ,   6439    ,   8839    ,   18027   ,0 };
            const unsigned long dim2054Kuo3Init[]   =   {   1   ,   3   ,   5   ,   7   ,   25  ,   25  ,   71  ,   141 ,   343 ,   35  ,   15  ,   1439    ,   3163    ,   8063    ,   15337   ,0 };
            const unsigned long dim2055Kuo3Init[]   =   {   1   ,   3   ,   3   ,   13  ,   27  ,   31  ,   91  ,   215 ,   67  ,   835 ,   673 ,   3355    ,   7633    ,   11711   ,   31479   ,0 };
            const unsigned long dim2056Kuo3Init[]   =   {   1   ,   3   ,   5   ,   3   ,   5   ,   9   ,   35  ,   23  ,   415 ,   341 ,   1813    ,   3761    ,   5053    ,   9497    ,   10147   ,0 };
            const unsigned long dim2057Kuo3Init[]   =   {   1   ,   1   ,   3   ,   5   ,   25  ,   59  ,   107 ,   57  ,   43  ,   955 ,   1049    ,   2613    ,   7003    ,   5749    ,   28211   ,0 };
            const unsigned long dim2058Kuo3Init[]   =   {   1   ,   3   ,   5   ,   5   ,   5   ,   61  ,   25  ,   17  ,   41  ,   477 ,   131 ,   3331    ,   1193    ,   14291   ,   1467    ,0 };
            const unsigned long dim2059Kuo3Init[]   =   {   1   ,   3   ,   3   ,   3   ,   5   ,   7   ,   37  ,   41  ,   215 ,   723 ,   1735    ,   3735    ,   3981    ,   13405   ,   5837    ,0 };
            const unsigned long dim2060Kuo3Init[]   =   {   1   ,   3   ,   1   ,   7   ,   17  ,   9   ,   21  ,   51  ,   451 ,   691 ,   1853    ,   779 ,   635 ,   15461   ,   14789   ,0 };
            const unsigned long dim2061Kuo3Init[]   =   {   1   ,   1   ,   7   ,   13  ,   9   ,   51  ,   21  ,   183 ,   33  ,   565 ,   529 ,   393 ,   6847    ,   10149   ,   23807   ,0 };
            const unsigned long dim2062Kuo3Init[]   =   {   1   ,   3   ,   1   ,   15  ,   9   ,   27  ,   1   ,   245 ,   459 ,   935 ,   1219    ,   2277    ,   6309    ,   13511   ,   31255   ,0 };
            const unsigned long dim2063Kuo3Init[]   =   {   1   ,   1   ,   3   ,   3   ,   21  ,   63  ,   121 ,   95  ,   5   ,   545 ,   995 ,   1011    ,   2073    ,   6145    ,   4635    ,0 };
            const unsigned long dim2064Kuo3Init[]   =   {   1   ,   1   ,   3   ,   15  ,   11  ,   1   ,   61  ,   53  ,   445 ,   33  ,   153 ,   2179    ,   2899    ,   7633    ,   20417   ,0 };
            const unsigned long dim2065Kuo3Init[]   =   {   1   ,   1   ,   5   ,   3   ,   29  ,   47  ,   65  ,   55  ,   509 ,   175 ,   1875    ,   2669    ,   4583    ,   775 ,   27621   ,0 };
            const unsigned long dim2066Kuo3Init[]   =   {   1   ,   3   ,   1   ,   1   ,   5   ,   21  ,   43  ,   37  ,   349 ,   89  ,   1581    ,   2105    ,   6753    ,   14577   ,   14339   ,0 };
            const unsigned long dim2067Kuo3Init[]   =   {   1   ,   1   ,   1   ,   11  ,   21  ,   17  ,   55  ,   11  ,   87  ,   323 ,   647 ,   955 ,   1343    ,   10799   ,   15375   ,0 };
            const unsigned long dim2068Kuo3Init[]   =   {   1   ,   3   ,   7   ,   13  ,   11  ,   43  ,   23  ,   25  ,   327 ,   797 ,   1111    ,   231 ,   4423    ,   5907    ,   16191   ,0 };
            const unsigned long dim2069Kuo3Init[]   =   {   1   ,   3   ,   5   ,   15  ,   1   ,   15  ,   109 ,   43  ,   217 ,   905 ,   1295    ,   2373    ,   497 ,   11643   ,   14371   ,0 };
            const unsigned long dim2070Kuo3Init[]   =   {   1   ,   1   ,   3   ,   5   ,   3   ,   43  ,   125 ,   249 ,   343 ,   947 ,   1523    ,   2393    ,   3073    ,   11301   ,   18751   ,0 };
            const unsigned long dim2071Kuo3Init[]   =   {   1   ,   3   ,   1   ,   13  ,   31  ,   27  ,   23  ,   241 ,   103 ,   503 ,   1715    ,   877 ,   5269    ,   10665   ,   23927   ,0 };
            const unsigned long dim2072Kuo3Init[]   =   {   1   ,   3   ,   1   ,   9   ,   21  ,   1   ,   91  ,   55  ,   297 ,   941 ,   1419    ,   3391    ,   811 ,   8399    ,   8513    ,0 };
            const unsigned long dim2073Kuo3Init[]   =   {   1   ,   1   ,   5   ,   1   ,   7   ,   3   ,   17  ,   139 ,   461 ,   1007    ,   1835    ,   165 ,   1575    ,   15075   ,   18673   ,0 };
            const unsigned long dim2074Kuo3Init[]   =   {   1   ,   3   ,   7   ,   11  ,   21  ,   63  ,   11  ,   123 ,   229 ,   767 ,   201 ,   3415    ,   4997    ,   14791   ,   27193   ,0 };
            const unsigned long dim2075Kuo3Init[]   =   {   1   ,   3   ,   3   ,   9   ,   5   ,   13  ,   71  ,   163 ,   499 ,   933 ,   1829    ,   2321    ,   1773    ,   5255    ,   18219   ,0 };
            const unsigned long dim2076Kuo3Init[]   =   {   1   ,   1   ,   5   ,   15  ,   15  ,   5   ,   47  ,   81  ,   167 ,   27  ,   773 ,   3607    ,   25  ,   12437   ,   11525   ,0 };
            const unsigned long dim2077Kuo3Init[]   =   {   1   ,   3   ,   3   ,   11  ,   3   ,   21  ,   41  ,   1   ,   441 ,   1013    ,   1409    ,   3097    ,   5671    ,   3953    ,   51  ,0 };
            const unsigned long dim2078Kuo3Init[]   =   {   1   ,   1   ,   3   ,   15  ,   15  ,   13  ,   93  ,   111 ,   355 ,   207 ,   1653    ,   2711    ,   81  ,   7801    ,   14105   ,0 };
            const unsigned long dim2079Kuo3Init[]   =   {   1   ,   1   ,   5   ,   9   ,   15  ,   57  ,   47  ,   53  ,   79  ,   255 ,   517 ,   4015    ,   8123    ,   5557    ,   4245    ,0 };
            const unsigned long dim2080Kuo3Init[]   =   {   1   ,   3   ,   1   ,   9   ,   25  ,   45  ,   97  ,   107 ,   107 ,   365 ,   1647    ,   3831    ,   5567    ,   167 ,   26977   ,0 };
            const unsigned long dim2081Kuo3Init[]   =   {   1   ,   1   ,   1   ,   7   ,   31  ,   21  ,   79  ,   155 ,   391 ,   347 ,   1295    ,   255 ,   6439    ,   8201    ,   31975   ,0 };
            const unsigned long dim2082Kuo3Init[]   =   {   1   ,   1   ,   1   ,   13  ,   19  ,   57  ,   9   ,   167 ,   359 ,   833 ,   709 ,   3437    ,   5031    ,   16221   ,   18491   ,0 };
            const unsigned long dim2083Kuo3Init[]   =   {   1   ,   3   ,   3   ,   11  ,   17  ,   37  ,   39  ,   167 ,   271 ,   801 ,   601 ,   461 ,   3949    ,   5237    ,   3425    ,0 };
            const unsigned long dim2084Kuo3Init[]   =   {   1   ,   3   ,   3   ,   13  ,   29  ,   47  ,   51  ,   161 ,   427 ,   421 ,   671 ,   2545    ,   4793    ,   10083   ,   3499    ,0 };
            const unsigned long dim2085Kuo3Init[]   =   {   1   ,   3   ,   5   ,   7   ,   27  ,   25  ,   59  ,   75  ,   431 ,   891 ,   1557    ,   3615    ,   4759    ,   11267   ,   8473    ,0 };
            const unsigned long dim2086Kuo3Init[]   =   {   1   ,   3   ,   1   ,   7   ,   7   ,   31  ,   63  ,   89  ,   353 ,   761 ,   1333    ,   1367    ,   5963    ,   12311   ,   20717   ,0 };
            const unsigned long dim2087Kuo3Init[]   =   {   1   ,   3   ,   3   ,   1   ,   27  ,   19  ,   93  ,   209 ,   291 ,   119 ,   513 ,   3757    ,   3477    ,   2541    ,   4919    ,0 };
            const unsigned long dim2088Kuo3Init[]   =   {   1   ,   3   ,   7   ,   1   ,   15  ,   9   ,   109 ,   41  ,   401 ,   365 ,   907 ,   3603    ,   6401    ,   12115   ,   28691   ,0 };
            const unsigned long dim2089Kuo3Init[]   =   {   1   ,   1   ,   5   ,   9   ,   3   ,   21  ,   121 ,   27  ,   355 ,   571 ,   133 ,   1073    ,   4875    ,   13351   ,   13197   ,0 };
            const unsigned long dim2090Kuo3Init[]   =   {   1   ,   3   ,   1   ,   3   ,   1   ,   31  ,   35  ,   171 ,   201 ,   617 ,   1181    ,   3621    ,   823 ,   8599    ,   4407    ,0 };
            const unsigned long dim2091Kuo3Init[]   =   {   1   ,   3   ,   3   ,   3   ,   31  ,   61  ,   45  ,   19  ,   191 ,   843 ,   909 ,   391 ,   5899    ,   14059   ,   7879    ,0 };
            const unsigned long dim2092Kuo3Init[]   =   {   1   ,   1   ,   7   ,   9   ,   19  ,   51  ,   123 ,   201 ,   163 ,   737 ,   153 ,   161 ,   7949    ,   2865    ,   17357   ,0 };
            const unsigned long dim2093Kuo3Init[]   =   {   1   ,   3   ,   7   ,   11  ,   23  ,   51  ,   121 ,   165 ,   459 ,   959 ,   647 ,   535 ,   7903    ,   2935    ,   24895   ,0 };
            const unsigned long dim2094Kuo3Init[]   =   {   1   ,   3   ,   5   ,   3   ,   1   ,   63  ,   73  ,   179 ,   485 ,   661 ,   255 ,   1423    ,   7277    ,   2109    ,   25113   ,0 };
            const unsigned long dim2095Kuo3Init[]   =   {   1   ,   3   ,   5   ,   11  ,   27  ,   51  ,   79  ,   151 ,   389 ,   251 ,   277 ,   1785    ,   4073    ,   1729    ,   22705   ,0 };
            const unsigned long dim2096Kuo3Init[]   =   {   1   ,   3   ,   3   ,   11  ,   23  ,   15  ,   105 ,   205 ,   413 ,   253 ,   1181    ,   2605    ,   6181    ,   1771    ,   17869   ,0 };
            const unsigned long dim2097Kuo3Init[]   =   {   1   ,   3   ,   7   ,   13  ,   29  ,   45  ,   95  ,   37  ,   319 ,   609 ,   369 ,   3311    ,   5717    ,   7219    ,   25157   ,0 };
            const unsigned long dim2098Kuo3Init[]   =   {   1   ,   1   ,   7   ,   15  ,   23  ,   3   ,   61  ,   125 ,   429 ,   29  ,   479 ,   1151    ,   4479    ,   8909    ,   9763    ,0 };
            const unsigned long dim2099Kuo3Init[]   =   {   1   ,   3   ,   1   ,   13  ,   25  ,   21  ,   11  ,   147 ,   61  ,   903 ,   1719    ,   3065    ,   7089    ,   15041   ,   6387    ,0 };
            const unsigned long dim2100Kuo3Init[]   =   {   1   ,   1   ,   5   ,   11  ,   17  ,   33  ,   43  ,   1   ,   271 ,   223 ,   363 ,   2799    ,   1043    ,   3929    ,   25991   ,0 };
            const unsigned long dim2101Kuo3Init[]   =   {   1   ,   1   ,   7   ,   11  ,   3   ,   15  ,   115 ,   245 ,   383 ,   881 ,   221 ,   2045    ,   621 ,   14543   ,   1027    ,0 };
            const unsigned long dim2102Kuo3Init[]   =   {   1   ,   3   ,   1   ,   7   ,   1   ,   45  ,   19  ,   133 ,   339 ,   883 ,   1643    ,   853 ,   7141    ,   6259    ,   6373    ,0 };
            const unsigned long dim2103Kuo3Init[]   =   {   1   ,   1   ,   3   ,   11  ,   27  ,   23  ,   123 ,   159 ,   185 ,   693 ,   577 ,   473 ,   313 ,   13579   ,   19317   ,0 };
            const unsigned long dim2104Kuo3Init[]   =   {   1   ,   3   ,   3   ,   15  ,   27  ,   19  ,   43  ,   23  ,   183 ,   925 ,   1987    ,   1945    ,   6341    ,   10771   ,   9669    ,0 };
            const unsigned long dim2105Kuo3Init[]   =   {   1   ,   1   ,   1   ,   3   ,   15  ,   27  ,   49  ,   117 ,   347 ,   149 ,   23  ,   103 ,   3613    ,   3123    ,   32759   ,0 };
            const unsigned long dim2106Kuo3Init[]   =   {   1   ,   3   ,   7   ,   13  ,   1   ,   37  ,   13  ,   111 ,   407 ,   319 ,   563 ,   1499    ,   6551    ,   13333   ,   659 ,0 };
            const unsigned long dim2107Kuo3Init[]   =   {   1   ,   3   ,   3   ,   3   ,   23  ,   21  ,   9   ,   109 ,   393 ,   115 ,   415 ,   351 ,   6113    ,   2417    ,   14151   ,0 };
            const unsigned long dim2108Kuo3Init[]   =   {   1   ,   3   ,   5   ,   7   ,   25  ,   7   ,   35  ,   41  ,   43  ,   663 ,   1809    ,   535 ,   7437    ,   12487   ,   1059    ,0 };
            const unsigned long dim2109Kuo3Init[]   =   {   1   ,   3   ,   3   ,   11  ,   13  ,   13  ,   123 ,   111 ,   391 ,   199 ,   69  ,   2357    ,   4779    ,   7461    ,   5003    ,0 };
            const unsigned long dim2110Kuo3Init[]   =   {   1   ,   1   ,   1   ,   13  ,   17  ,   63  ,   5   ,   163 ,   107 ,   609 ,   1979    ,   3389    ,   499 ,   13111   ,   219 ,0 };
            const unsigned long dim2111Kuo3Init[]   =   {   1   ,   3   ,   7   ,   1   ,   19  ,   41  ,   43  ,   29  ,   35  ,   371 ,   1773    ,   3563    ,   2005    ,   8217    ,   5131    ,0 };
            const unsigned long dim2112Kuo3Init[]   =   {   1   ,   3   ,   7   ,   15  ,   13  ,   53  ,   43  ,   167 ,   201 ,   259 ,   1855    ,   3971    ,   1179    ,   8591    ,   21425   ,0 };
            const unsigned long dim2113Kuo3Init[]   =   {   1   ,   1   ,   1   ,   11  ,   27  ,   21  ,   45  ,   47  ,   185 ,   753 ,   531 ,   2553    ,   5515    ,   11501   ,   4623    ,0 };
            const unsigned long dim2114Kuo3Init[]   =   {   1   ,   1   ,   7   ,   9   ,   9   ,   49  ,   29  ,   207 ,   441 ,   265 ,   1059    ,   1703    ,   1625    ,   9371    ,   23045   ,0 };
            const unsigned long dim2115Kuo3Init[]   =   {   1   ,   3   ,   5   ,   9   ,   13  ,   29  ,   33  ,   185 ,   233 ,   11  ,   381 ,   3917    ,   4931    ,   3497    ,   24713   ,0 };
            const unsigned long dim2116Kuo3Init[]   =   {   1   ,   3   ,   1   ,   11  ,   13  ,   61  ,   43  ,   147 ,   505 ,   251 ,   1875    ,   3009    ,   4695    ,   6793    ,   3083    ,0 };
            const unsigned long dim2117Kuo3Init[]   =   {   1   ,   3   ,   3   ,   3   ,   31  ,   17  ,   119 ,   35  ,   111 ,   767 ,   211 ,   3227    ,   2539    ,   10755   ,   26177   ,0 };
            const unsigned long dim2118Kuo3Init[]   =   {   1   ,   3   ,   1   ,   7   ,   9   ,   63  ,   3   ,   45  ,   195 ,   515 ,   103 ,   2327    ,   5967    ,   8087    ,   9849    ,0 };
            const unsigned long dim2119Kuo3Init[]   =   {   1   ,   3   ,   7   ,   13  ,   31  ,   51  ,   55  ,   231 ,   451 ,   259 ,   1025    ,   2671    ,   5317    ,   7087    ,   7365    ,0 };
            const unsigned long dim2120Kuo3Init[]   =   {   1   ,   3   ,   5   ,   3   ,   5   ,   41  ,   33  ,   255 ,   305 ,   831 ,   1279    ,   4013    ,   3899    ,   171 ,   8543    ,0 };
            const unsigned long dim2121Kuo3Init[]   =   {   1   ,   1   ,   5   ,   9   ,   7   ,   9   ,   97  ,   125 ,   123 ,   3   ,   121 ,   4041    ,   3941    ,   1655    ,   25583   ,0 };
            const unsigned long dim2122Kuo3Init[]   =   {   1   ,   1   ,   3   ,   3   ,   9   ,   63  ,   99  ,   53  ,   85  ,   57  ,   1117    ,   969 ,   7067    ,   1393    ,   17497   ,0 };
            const unsigned long dim2123Kuo3Init[]   =   {   1   ,   1   ,   3   ,   7   ,   9   ,   51  ,   1   ,   207 ,   443 ,   967 ,   1387    ,   2515    ,   1789    ,   1493    ,   20941   ,0 };
            const unsigned long dim2124Kuo3Init[]   =   {   1   ,   1   ,   5   ,   9   ,   15  ,   37  ,   11  ,   191 ,   503 ,   823 ,   1789    ,   4071    ,   3355    ,   9177    ,   27281   ,0 };
            const unsigned long dim2125Kuo3Init[]   =   {   1   ,   3   ,   7   ,   5   ,   3   ,   61  ,   83  ,   19  ,   325 ,   543 ,   19  ,   239 ,   3513    ,   16147   ,   32395   ,0 };
            const unsigned long dim2126Kuo3Init[]   =   {   1   ,   3   ,   5   ,   7   ,   1   ,   17  ,   101 ,   23  ,   379 ,   649 ,   853 ,   3045    ,   1301    ,   7215    ,   18851   ,0 };
            const unsigned long dim2127Kuo3Init[]   =   {   1   ,   3   ,   5   ,   13  ,   25  ,   11  ,   41  ,   213 ,   137 ,   705 ,   1829    ,   1055    ,   3061    ,   9335    ,   5797    ,0 };
            const unsigned long dim2128Kuo3Init[]   =   {   1   ,   3   ,   1   ,   13  ,   11  ,   21  ,   17  ,   117 ,   473 ,   687 ,   125 ,   4007    ,   5283    ,   11961   ,   15365   ,0 };
            const unsigned long dim2129Kuo3Init[]   =   {   1   ,   3   ,   1   ,   13  ,   11  ,   19  ,   3   ,   189 ,   393 ,   219 ,   1229    ,   1521    ,   7137    ,   1811    ,   6745    ,0 };
            const unsigned long dim2130Kuo3Init[]   =   {   1   ,   1   ,   7   ,   9   ,   21  ,   59  ,   23  ,   193 ,   321 ,   691 ,   735 ,   3731    ,   593 ,   9351    ,   6571    ,0 };
            const unsigned long dim2131Kuo3Init[]   =   {   1   ,   3   ,   5   ,   5   ,   23  ,   55  ,   43  ,   107 ,   217 ,   111 ,   921 ,   1017    ,   7055    ,   16059   ,   8181    ,0 };
            const unsigned long dim2132Kuo3Init[]   =   {   1   ,   3   ,   3   ,   3   ,   29  ,   57  ,   93  ,   91  ,   9   ,   453 ,   1245    ,   691 ,   1571    ,   5755    ,   13387   ,0 };
            const unsigned long dim2133Kuo3Init[]   =   {   1   ,   1   ,   3   ,   1   ,   1   ,   23  ,   87  ,   239 ,   373 ,   433 ,   553 ,   2205    ,   4495    ,   3975    ,   25819   ,0 };
            const unsigned long dim2134Kuo3Init[]   =   {   1   ,   1   ,   3   ,   1   ,   23  ,   7   ,   9   ,   227 ,   503 ,   87  ,   819 ,   3167    ,   4135    ,   9281    ,   8443    ,0 };
            const unsigned long dim2135Kuo3Init[]   =   {   1   ,   1   ,   3   ,   5   ,   7   ,   27  ,   17  ,   81  ,   295 ,   131 ,   955 ,   4065    ,   797 ,   16165   ,   2385    ,0 };
            const unsigned long dim2136Kuo3Init[]   =   {   1   ,   3   ,   3   ,   1   ,   9   ,   47  ,   113 ,   105 ,   277 ,   187 ,   1983    ,   2833    ,   2749    ,   4473    ,   31275   ,0 };
            const unsigned long dim2137Kuo3Init[]   =   {   1   ,   3   ,   3   ,   3   ,   19  ,   61  ,   41  ,   51  ,   435 ,   933 ,   961 ,   459 ,   4187    ,   6619    ,   777 ,0 };
            const unsigned long dim2138Kuo3Init[]   =   {   1   ,   3   ,   3   ,   9   ,   11  ,   25  ,   51  ,   219 ,   203 ,   897 ,   777 ,   3047    ,   2601    ,   4387    ,   14185   ,0 };
            const unsigned long dim2139Kuo3Init[]   =   {   1   ,   3   ,   7   ,   15  ,   23  ,   17  ,   103 ,   55  ,   387 ,   127 ,   197 ,   1045    ,   3973    ,   10503   ,   23443   ,0 };
            const unsigned long dim2140Kuo3Init[]   =   {   1   ,   1   ,   5   ,   9   ,   13  ,   59  ,   87  ,   253 ,   7   ,   879 ,   1335    ,   1349    ,   6353    ,   5897    ,   10021   ,0 };
            const unsigned long dim2141Kuo3Init[]   =   {   1   ,   1   ,   3   ,   5   ,   19  ,   49  ,   107 ,   65  ,   229 ,   317 ,   169 ,   509 ,   1819    ,   2267    ,   13269   ,0 };
            const unsigned long dim2142Kuo3Init[]   =   {   1   ,   3   ,   5   ,   13  ,   21  ,   25  ,   11  ,   143 ,   511 ,   107 ,   637 ,   3237    ,   3985    ,   15197   ,   23343   ,0 };
            const unsigned long dim2143Kuo3Init[]   =   {   1   ,   1   ,   5   ,   15  ,   1   ,   15  ,   67  ,   91  ,   131 ,   501 ,   1803    ,   699 ,   2031    ,   11015   ,   17989   ,0 };
            const unsigned long dim2144Kuo3Init[]   =   {   1   ,   1   ,   7   ,   13  ,   3   ,   55  ,   103 ,   245 ,   229 ,   121 ,   245 ,   1299    ,   3357    ,   3283    ,   26181   ,0 };
            const unsigned long dim2145Kuo3Init[]   =   {   1   ,   1   ,   7   ,   1   ,   19  ,   53  ,   101 ,   125 ,   35  ,   963 ,   887 ,   3119    ,   1407    ,   723 ,   29169   ,0 };
            const unsigned long dim2146Kuo3Init[]   =   {   1   ,   3   ,   7   ,   3   ,   27  ,   39  ,   29  ,   71  ,   7   ,   529 ,   329 ,   2811    ,   4627    ,   2865    ,   26837   ,0 };
            const unsigned long dim2147Kuo3Init[]   =   {   1   ,   3   ,   5   ,   7   ,   29  ,   55  ,   71  ,   129 ,   455 ,   465 ,   1949    ,   3895    ,   6971    ,   15621   ,   26819   ,0 };
            const unsigned long dim2148Kuo3Init[]   =   {   1   ,   1   ,   3   ,   15  ,   5   ,   63  ,   95  ,   15  ,   425 ,   347 ,   835 ,   1973    ,   5171    ,   3963    ,   23691   ,0 };
            const unsigned long dim2149Kuo3Init[]   =   {   1   ,   3   ,   1   ,   9   ,   27  ,   23  ,   87  ,   237 ,   261 ,   349 ,   1777    ,   287 ,   3543    ,   35  ,   31891   ,0 };
            const unsigned long dim2150Kuo3Init[]   =   {   1   ,   1   ,   7   ,   7   ,   17  ,   59  ,   7   ,   65  ,   45  ,   733 ,   239 ,   221 ,   6419    ,   9161    ,   17499   ,0 };
            const unsigned long dim2151Kuo3Init[]   =   {   1   ,   3   ,   3   ,   15  ,   7   ,   53  ,   121 ,   97  ,   127 ,   583 ,   501 ,   1667    ,   6265    ,   3233    ,   19187   ,0 };
            const unsigned long dim2152Kuo3Init[]   =   {   1   ,   1   ,   3   ,   1   ,   29  ,   17  ,   121 ,   89  ,   107 ,   787 ,   1699    ,   3389    ,   6191    ,   11503   ,   2461    ,0 };
            const unsigned long dim2153Kuo3Init[]   =   {   1   ,   1   ,   1   ,   9   ,   13  ,   27  ,   97  ,   49  ,   447 ,   851 ,   147 ,   439 ,   277 ,   11129   ,   2201    ,0 };
            const unsigned long dim2154Kuo3Init[]   =   {   1   ,   3   ,   3   ,   11  ,   29  ,   47  ,   93  ,   243 ,   101 ,   627 ,   1607    ,   459 ,   27  ,   10635   ,   1345    ,0 };
            const unsigned long dim2155Kuo3Init[]   =   {   1   ,   1   ,   7   ,   3   ,   19  ,   11  ,   61  ,   253 ,   345 ,   523 ,   317 ,   1599    ,   5625    ,   14385   ,   3461    ,0 };
            const unsigned long dim2156Kuo3Init[]   =   {   1   ,   1   ,   5   ,   13  ,   27  ,   59  ,   33  ,   251 ,   29  ,   787 ,   1655    ,   3383    ,   2451    ,   2499    ,   10719   ,0 };
            const unsigned long dim2157Kuo3Init[]   =   {   1   ,   3   ,   7   ,   7   ,   1   ,   29  ,   97  ,   233 ,   193 ,   459 ,   1667    ,   47  ,   3059    ,   8713    ,   13259   ,0 };
            const unsigned long dim2158Kuo3Init[]   =   {   1   ,   1   ,   3   ,   9   ,   29  ,   63  ,   127 ,   35  ,   105 ,   95  ,   887 ,   2337    ,   5999    ,   13553   ,   30225   ,0 };
            const unsigned long dim2159Kuo3Init[]   =   {   1   ,   1   ,   7   ,   7   ,   17  ,   43  ,   47  ,   205 ,   497 ,   835 ,   1565    ,   3863    ,   785 ,   6149    ,   17595   ,0 };
            const unsigned long dim2160Kuo3Init[]   =   {   1   ,   1   ,   1   ,   13  ,   29  ,   17  ,   47  ,   37  ,   339 ,   131 ,   1219    ,   3545    ,   3499    ,   2223    ,   3107    ,0 };
            const unsigned long dim2161Kuo3Init[]   =   {   1   ,   1   ,   1   ,   15  ,   7   ,   17  ,   39  ,   25  ,   97  ,   691 ,   1645    ,   3333    ,   95  ,   15593   ,   11667   ,0 };
            const unsigned long dim2162Kuo3Init[]   =   {   1   ,   3   ,   5   ,   7   ,   13  ,   21  ,   61  ,   147 ,   115 ,   249 ,   981 ,   665 ,   1245    ,   14397   ,   9571    ,0 };
            const unsigned long dim2163Kuo3Init[]   =   {   1   ,   1   ,   3   ,   13  ,   15  ,   5   ,   7   ,   129 ,   151 ,   947 ,   375 ,   1631    ,   3247    ,   1321    ,   19495   ,0 };
            const unsigned long dim2164Kuo3Init[]   =   {   1   ,   1   ,   5   ,   11  ,   5   ,   5   ,   69  ,   247 ,   53  ,   37  ,   683 ,   1581    ,   2997    ,   5897    ,   4797    ,0 };
            const unsigned long dim2165Kuo3Init[]   =   {   1   ,   1   ,   5   ,   15  ,   5   ,   45  ,   115 ,   65  ,   509 ,   557 ,   1749    ,   2167    ,   4795    ,   629 ,   10749   ,0 };
            const unsigned long dim2166Kuo3Init[]   =   {   1   ,   1   ,   3   ,   1   ,   25  ,   23  ,   99  ,   93  ,   99  ,   721 ,   1235    ,   1259    ,   2739    ,   11797   ,   15985   ,0 };
            const unsigned long dim2167Kuo3Init[]   =   {   1   ,   3   ,   5   ,   11  ,   15  ,   29  ,   97  ,   103 ,   21  ,   981 ,   1717    ,   1483    ,   711 ,   12419   ,   14401   ,0 };
            const unsigned long dim2168Kuo3Init[]   =   {   1   ,   3   ,   1   ,   1   ,   5   ,   5   ,   125 ,   159 ,   301 ,   733 ,   1987    ,   1679    ,   7459    ,   3493    ,   29377   ,0 };
            const unsigned long dim2169Kuo3Init[]   =   {   1   ,   1   ,   7   ,   7   ,   29  ,   45  ,   53  ,   197 ,   303 ,   303 ,   931 ,   1171    ,   1771    ,   14895   ,   21525   ,0 };
            const unsigned long dim2170Kuo3Init[]   =   {   1   ,   3   ,   7   ,   13  ,   11  ,   59  ,   93  ,   113 ,   219 ,   557 ,   397 ,   2639    ,   5089    ,   6833    ,   20455   ,0 };
            const unsigned long dim2171Kuo3Init[]   =   {   1   ,   3   ,   1   ,   3   ,   11  ,   33  ,   117 ,   163 ,   359 ,   431 ,   915 ,   3285    ,   2443    ,   8225    ,   16341   ,0 };
            const unsigned long dim2172Kuo3Init[]   =   {   1   ,   3   ,   1   ,   9   ,   19  ,   1   ,   105 ,   65  ,   3   ,   71  ,   705 ,   2047    ,   2169    ,   13129   ,   8575    ,0 };
            const unsigned long dim2173Kuo3Init[]   =   {   1   ,   1   ,   1   ,   9   ,   23  ,   63  ,   67  ,   203 ,   429 ,   897 ,   1453    ,   101 ,   1445    ,   8553    ,   16355   ,0 };
            const unsigned long dim2174Kuo3Init[]   =   {   1   ,   1   ,   7   ,   5   ,   13  ,   61  ,   79  ,   217 ,   83  ,   339 ,   529 ,   3573    ,   4075    ,   5929    ,   2687    ,0 };
            const unsigned long dim2175Kuo3Init[]   =   {   1   ,   1   ,   7   ,   11  ,   27  ,   9   ,   93  ,   105 ,   179 ,   479 ,   1093    ,   3551    ,   5697    ,   8989    ,   1099    ,0 };
            const unsigned long dim2176Kuo3Init[]   =   {   1   ,   1   ,   1   ,   11  ,   15  ,   9   ,   21  ,   59  ,   193 ,   231 ,   1817    ,   463 ,   1945    ,   13795   ,   28655   ,0 };
            const unsigned long dim2177Kuo3Init[]   =   {   1   ,   1   ,   1   ,   5   ,   3   ,   27  ,   49  ,   179 ,   205 ,   49  ,   355 ,   667 ,   6573    ,   4317    ,   26599   ,0 };
            const unsigned long dim2178Kuo3Init[]   =   {   1   ,   3   ,   7   ,   5   ,   9   ,   45  ,   121 ,   219 ,   309 ,   371 ,   1119    ,   1991    ,   5889    ,   14257   ,   1105    ,0 };
            const unsigned long dim2179Kuo3Init[]   =   {   1   ,   1   ,   3   ,   3   ,   7   ,   19  ,   111 ,   191 ,   321 ,   77  ,   39  ,   1921    ,   563 ,   1049    ,   26623   ,0 };
            const unsigned long dim2180Kuo3Init[]   =   {   1   ,   3   ,   3   ,   13  ,   13  ,   43  ,   21  ,   121 ,   181 ,   185 ,   629 ,   211 ,   7897    ,   9445    ,   12679   ,0 };
            const unsigned long dim2181Kuo3Init[]   =   {   1   ,   1   ,   7   ,   15  ,   3   ,   19  ,   61  ,   159 ,   337 ,   711 ,   155 ,   2099    ,   2201    ,   14337   ,   21633   ,0 };
            const unsigned long dim2182Kuo3Init[]   =   {   1   ,   1   ,   3   ,   9   ,   27  ,   63  ,   43  ,   209 ,   17  ,   565 ,   225 ,   2381    ,   4577    ,   15793   ,   4861    ,0 };
            const unsigned long dim2183Kuo3Init[]   =   {   1   ,   3   ,   7   ,   3   ,   25  ,   51  ,   65  ,   187 ,   235 ,   479 ,   293 ,   789 ,   2453    ,   4367    ,   15835   ,0 };
            const unsigned long dim2184Kuo3Init[]   =   {   1   ,   1   ,   5   ,   11  ,   17  ,   7   ,   115 ,   165 ,   3   ,   457 ,   455 ,   969 ,   5859    ,   2663    ,   15337   ,0 };
            const unsigned long dim2185Kuo3Init[]   =   {   1   ,   3   ,   1   ,   13  ,   19  ,   1   ,   55  ,   177 ,   365 ,   395 ,   1043    ,   1095    ,   6531    ,   8599    ,   12867   ,0 };
            const unsigned long dim2186Kuo3Init[]   =   {   1   ,   3   ,   7   ,   1   ,   29  ,   57  ,   55  ,   139 ,   187 ,   649 ,   1899    ,   355 ,   6447    ,   4521    ,   31497   ,0 };
            const unsigned long dim2187Kuo3Init[]   =   {   1   ,   3   ,   5   ,   1   ,   19  ,   55  ,   73  ,   163 ,   127 ,   893 ,   1189    ,   2425    ,   5145    ,   15965   ,   5937    ,0 };
            const unsigned long dim2188Kuo3Init[]   =   {   1   ,   3   ,   1   ,   1   ,   21  ,   33  ,   91  ,   81  ,   75  ,   39  ,   1   ,   397 ,   2799    ,   11987   ,   27829   ,0 };
            const unsigned long dim2189Kuo3Init[]   =   {   1   ,   3   ,   7   ,   5   ,   25  ,   15  ,   95  ,   61  ,   351 ,   377 ,   1949    ,   3141    ,   6259    ,   135 ,   19307   ,0 };
            const unsigned long dim2190Kuo3Init[]   =   {   1   ,   3   ,   3   ,   7   ,   7   ,   45  ,   1   ,   197 ,   235 ,   747 ,   1333    ,   3293    ,   1439    ,   7447    ,   26119   ,0 };
            const unsigned long dim2191Kuo3Init[]   =   {   1   ,   3   ,   3   ,   3   ,   25  ,   31  ,   113 ,   159 ,   335 ,   587 ,   733 ,   1045    ,   5015    ,   9677    ,   21777   ,0 };
            const unsigned long dim2192Kuo3Init[]   =   {   1   ,   1   ,   5   ,   3   ,   15  ,   55  ,   91  ,   181 ,   465 ,   409 ,   995 ,   3243    ,   6165    ,   1547    ,   31785   ,0 };
            const unsigned long dim2193Kuo3Init[]   =   {   1   ,   3   ,   7   ,   3   ,   7   ,   47  ,   27  ,   87  ,   399 ,   265 ,   945 ,   1675    ,   5383    ,   4847    ,   27859   ,0 };
            const unsigned long dim2194Kuo3Init[]   =   {   1   ,   1   ,   7   ,   3   ,   21  ,   53  ,   69  ,   107 ,   121 ,   855 ,   629 ,   1947    ,   8149    ,   10707   ,   2981    ,0 };
            const unsigned long dim2195Kuo3Init[]   =   {   1   ,   3   ,   1   ,   11  ,   3   ,   25  ,   53  ,   87  ,   349 ,   237 ,   1745    ,   3705    ,   4273    ,   4109    ,   26059   ,0 };
            const unsigned long dim2196Kuo3Init[]   =   {   1   ,   1   ,   5   ,   11  ,   25  ,   19  ,   11  ,   179 ,   301 ,   27  ,   1343    ,   79  ,   2475    ,   5581    ,   27225   ,0 };
            const unsigned long dim2197Kuo3Init[]   =   {   1   ,   3   ,   7   ,   9   ,   29  ,   55  ,   119 ,   107 ,   431 ,   943 ,   1567    ,   3727    ,   7415    ,   2565    ,   24957   ,0 };
            const unsigned long dim2198Kuo3Init[]   =   {   1   ,   3   ,   3   ,   1   ,   15  ,   27  ,   91  ,   159 ,   285 ,   781 ,   357 ,   643 ,   7445    ,   9931    ,   11881   ,0 };
            const unsigned long dim2199Kuo3Init[]   =   {   1   ,   3   ,   7   ,   9   ,   5   ,   53  ,   65  ,   195 ,   273 ,   37  ,   1785    ,   467 ,   5211    ,   25  ,   1749    ,0 };
            const unsigned long dim2200Kuo3Init[]   =   {   1   ,   1   ,   1   ,   11  ,   7   ,   11  ,   43  ,   143 ,   305 ,   825 ,   1055    ,   2995    ,   3907    ,   7581    ,   9679    ,0 };
            const unsigned long dim2201Kuo3Init[]   =   {   1   ,   3   ,   1   ,   9   ,   31  ,   29  ,   93  ,   3   ,   299 ,   1   ,   1851    ,   395 ,   239 ,   8323    ,   18201   ,0 };
            const unsigned long dim2202Kuo3Init[]   =   {   1   ,   1   ,   1   ,   5   ,   29  ,   29  ,   75  ,   3   ,   273 ,   759 ,   801 ,   2841    ,   5789    ,   6651    ,   19351   ,0 };
            const unsigned long dim2203Kuo3Init[]   =   {   1   ,   3   ,   7   ,   1   ,   1   ,   11  ,   11  ,   215 ,   27  ,   955 ,   1397    ,   3483    ,   1317    ,   223 ,   20559   ,0 };
            const unsigned long dim2204Kuo3Init[]   =   {   1   ,   3   ,   5   ,   13  ,   21  ,   23  ,   39  ,   189 ,   295 ,   255 ,   1457    ,   2821    ,   3145    ,   15437   ,   26997   ,0 };
            const unsigned long dim2205Kuo3Init[]   =   {   1   ,   1   ,   3   ,   5   ,   9   ,   43  ,   13  ,   97  ,   259 ,   23  ,   933 ,   3867    ,   5921    ,   4199    ,   24721   ,0 };
            const unsigned long dim2206Kuo3Init[]   =   {   1   ,   1   ,   3   ,   11  ,   1   ,   19  ,   89  ,   87  ,   17  ,   247 ,   1535    ,   2629    ,   2893    ,   12209   ,   23739   ,0 };
            const unsigned long dim2207Kuo3Init[]   =   {   1   ,   1   ,   1   ,   1   ,   31  ,   15  ,   81  ,   37  ,   235 ,   249 ,   1437    ,   1485    ,   5889    ,   10017   ,   18593   ,0 };
            const unsigned long dim2208Kuo3Init[]   =   {   1   ,   3   ,   7   ,   5   ,   17  ,   33  ,   117 ,   5   ,   31  ,   63  ,   2015    ,   3063    ,   5139    ,   14771   ,   29107   ,0 };
            const unsigned long dim2209Kuo3Init[]   =   {   1   ,   1   ,   7   ,   15  ,   13  ,   3   ,   25  ,   127 ,   419 ,   945 ,   1739    ,   175 ,   1591    ,   4443    ,   27289   ,0 };
            const unsigned long dim2210Kuo3Init[]   =   {   1   ,   1   ,   1   ,   7   ,   11  ,   59  ,   99  ,   119 ,   307 ,   857 ,   1711    ,   1201    ,   1623    ,   11429   ,   5663    ,0 };
            const unsigned long dim2211Kuo3Init[]   =   {   1   ,   3   ,   1   ,   13  ,   17  ,   45  ,   65  ,   225 ,   131 ,   279 ,   587 ,   3487    ,   1123    ,   2753    ,   30789   ,0 };
            const unsigned long dim2212Kuo3Init[]   =   {   1   ,   3   ,   3   ,   15  ,   21  ,   55  ,   17  ,   141 ,   109 ,   1021    ,   1387    ,   1407    ,   8163    ,   1983    ,   25971   ,0 };
            const unsigned long dim2213Kuo3Init[]   =   {   1   ,   3   ,   3   ,   9   ,   21  ,   51  ,   103 ,   243 ,   435 ,   481 ,   857 ,   3353    ,   2717    ,   2295    ,   7999    ,0 };
            const unsigned long dim2214Kuo3Init[]   =   {   1   ,   1   ,   1   ,   7   ,   7   ,   27  ,   83  ,   45  ,   359 ,   191 ,   259 ,   3229    ,   7869    ,   10729   ,   28587   ,0 };
            const unsigned long dim2215Kuo3Init[]   =   {   1   ,   3   ,   7   ,   5   ,   19  ,   25  ,   51  ,   103 ,   37  ,   429 ,   995 ,   415 ,   2879    ,   2283    ,   29981   ,0 };
            const unsigned long dim2216Kuo3Init[]   =   {   1   ,   1   ,   3   ,   3   ,   21  ,   19  ,   29  ,   35  ,   401 ,   19  ,   1761    ,   3735    ,   7775    ,   3365    ,   31767   ,0 };
            const unsigned long dim2217Kuo3Init[]   =   {   1   ,   1   ,   7   ,   1   ,   29  ,   1   ,   9   ,   195 ,   279 ,   1003    ,   999 ,   4043    ,   3761    ,   11341   ,   7117    ,0 };
            const unsigned long dim2218Kuo3Init[]   =   {   1   ,   1   ,   5   ,   11  ,   19  ,   7   ,   97  ,   181 ,   37  ,   929 ,   667 ,   993 ,   2517    ,   12683   ,   24487   ,0 };
            const unsigned long dim2219Kuo3Init[]   =   {   1   ,   1   ,   3   ,   1   ,   31  ,   15  ,   47  ,   143 ,   511 ,   525 ,   1799    ,   1743    ,   3237    ,   12839   ,   2287    ,0 };
            const unsigned long dim2220Kuo3Init[]   =   {   1   ,   3   ,   3   ,   15  ,   31  ,   59  ,   59  ,   205 ,   99  ,   233 ,   1857    ,   2469    ,   2661    ,   3675    ,   391 ,0 };
            const unsigned long dim2221Kuo3Init[]   =   {   1   ,   3   ,   5   ,   15  ,   31  ,   23  ,   23  ,   1   ,   313 ,   401 ,   605 ,   3771    ,   2703    ,   6037    ,   7943    ,0 };
            const unsigned long dim2222Kuo3Init[]   =   {   1   ,   1   ,   3   ,   13  ,   21  ,   19  ,   55  ,   139 ,   273 ,   725 ,   1239    ,   595 ,   549 ,   15279   ,   20887   ,0 };
            const unsigned long dim2223Kuo3Init[]   =   {   1   ,   1   ,   5   ,   3   ,   13  ,   9   ,   39  ,   83  ,   257 ,   323 ,   1801    ,   167 ,   3747    ,   2867    ,   30151   ,0 };
            const unsigned long dim2224Kuo3Init[]   =   {   1   ,   3   ,   3   ,   3   ,   17  ,   3   ,   67  ,   163 ,   325 ,   569 ,   1099    ,   4037    ,   5003    ,   6985    ,   7445    ,0 };
            const unsigned long dim2225Kuo3Init[]   =   {   1   ,   3   ,   7   ,   13  ,   31  ,   37  ,   13  ,   91  ,   21  ,   763 ,   785 ,   1461    ,   5257    ,   15307   ,   16389   ,0 };
            const unsigned long dim2226Kuo3Init[]   =   {   1   ,   3   ,   5   ,   13  ,   1   ,   35  ,   27  ,   179 ,   429 ,   21  ,   865 ,   775 ,   4993    ,   335 ,   12479   ,0 };
            const unsigned long dim2227Kuo3Init[]   =   {   1   ,   3   ,   1   ,   1   ,   19  ,   1   ,   111 ,   201 ,   219 ,   709 ,   1539    ,   1905    ,   647 ,   8605    ,   14129   ,0 };
            const unsigned long dim2228Kuo3Init[]   =   {   1   ,   3   ,   5   ,   7   ,   21  ,   7   ,   117 ,   251 ,   311 ,   689 ,   1125    ,   1931    ,   6363    ,   7409    ,   4189    ,0 };
            const unsigned long dim2229Kuo3Init[]   =   {   1   ,   1   ,   3   ,   7   ,   1   ,   27  ,   93  ,   5   ,   395 ,   117 ,   113 ,   1297    ,   501 ,   13073   ,   30483   ,0 };
            const unsigned long dim2230Kuo3Init[]   =   {   1   ,   1   ,   3   ,   11  ,   19  ,   57  ,   65  ,   67  ,   13  ,   893 ,   551 ,   539 ,   2861    ,   10635   ,   16057   ,0 };
            const unsigned long dim2231Kuo3Init[]   =   {   1   ,   3   ,   3   ,   11  ,   19  ,   5   ,   41  ,   43  ,   161 ,   1015    ,   33  ,   471 ,   229 ,   13315   ,   10791   ,0 };
            const unsigned long dim2232Kuo3Init[]   =   {   1   ,   3   ,   1   ,   7   ,   1   ,   13  ,   97  ,   149 ,   321 ,   131 ,   1389    ,   715 ,   5421    ,   343 ,   4129    ,0 };
            const unsigned long dim2233Kuo3Init[]   =   {   1   ,   1   ,   3   ,   11  ,   15  ,   51  ,   91  ,   57  ,   229 ,   1007    ,   709 ,   2745    ,   147 ,   10427   ,   21151   ,0 };
            const unsigned long dim2234Kuo3Init[]   =   {   1   ,   3   ,   7   ,   15  ,   1   ,   31  ,   65  ,   251 ,   75  ,   553 ,   133 ,   1027    ,   4857    ,   7139    ,   10591   ,0 };
            const unsigned long dim2235Kuo3Init[]   =   {   1   ,   3   ,   5   ,   15  ,   13  ,   11  ,   49  ,   183 ,   263 ,   121 ,   981 ,   1421    ,   7781    ,   5313    ,   3245    ,0 };
            const unsigned long dim2236Kuo3Init[]   =   {   1   ,   3   ,   5   ,   15  ,   31  ,   27  ,   109 ,   85  ,   327 ,   999 ,   1063    ,   705 ,   6047    ,   13227   ,   5817    ,0 };
            const unsigned long dim2237Kuo3Init[]   =   {   1   ,   1   ,   5   ,   15  ,   9   ,   5   ,   99  ,   69  ,   91  ,   637 ,   1353    ,   2471    ,   7003    ,   9429    ,   25235   ,0 };
            const unsigned long dim2238Kuo3Init[]   =   {   1   ,   3   ,   3   ,   11  ,   1   ,   25  ,   77  ,   67  ,   173 ,   929 ,   1919    ,   4021    ,   5587    ,   3333    ,   12769   ,0 };
            const unsigned long dim2239Kuo3Init[]   =   {   1   ,   1   ,   5   ,   15  ,   15  ,   9   ,   11  ,   71  ,   465 ,   831 ,   1127    ,   3439    ,   6793    ,   1269    ,   1699    ,0 };
            const unsigned long dim2240Kuo3Init[]   =   {   1   ,   3   ,   1   ,   1   ,   17  ,   43  ,   45  ,   181 ,   147 ,   501 ,   283 ,   3875    ,   7221    ,   14487   ,   2549    ,0 };
            const unsigned long dim2241Kuo3Init[]   =   {   1   ,   3   ,   7   ,   11  ,   1   ,   21  ,   125 ,   179 ,   479 ,   367 ,   1961    ,   607 ,   255 ,   913 ,   16143   ,0 };
            const unsigned long dim2242Kuo3Init[]   =   {   1   ,   3   ,   3   ,   15  ,   7   ,   49  ,   117 ,   97  ,   45  ,   729 ,   1331    ,   1373    ,   2147    ,   11673   ,   9779    ,0 };
            const unsigned long dim2243Kuo3Init[]   =   {   1   ,   3   ,   5   ,   11  ,   1   ,   11  ,   105 ,   231 ,   237 ,   449 ,   1195    ,   33  ,   4531    ,   11591   ,   29539   ,0 };
            const unsigned long dim2244Kuo3Init[]   =   {   1   ,   3   ,   3   ,   3   ,   27  ,   47  ,   105 ,   85  ,   25  ,   663 ,   47  ,   3213    ,   6777    ,   11947   ,   28993   ,0 };
            const unsigned long dim2245Kuo3Init[]   =   {   1   ,   1   ,   5   ,   15  ,   11  ,   57  ,   121 ,   107 ,   327 ,   631 ,   285 ,   921 ,   3721    ,   7983    ,   9199    ,0 };
            const unsigned long dim2246Kuo3Init[]   =   {   1   ,   1   ,   1   ,   13  ,   3   ,   61  ,   49  ,   101 ,   415 ,   187 ,   1645    ,   2139    ,   143 ,   12033   ,   11409   ,0 };
            const unsigned long dim2247Kuo3Init[]   =   {   1   ,   1   ,   5   ,   13  ,   7   ,   15  ,   27  ,   143 ,   17  ,   977 ,   495 ,   2479    ,   1507    ,   10621   ,   2119    ,0 };
            const unsigned long dim2248Kuo3Init[]   =   {   1   ,   1   ,   5   ,   13  ,   11  ,   9   ,   53  ,   159 ,   479 ,   333 ,   1423    ,   2559    ,   5497    ,   12611   ,   22627   ,0 };
            const unsigned long dim2249Kuo3Init[]   =   {   1   ,   1   ,   5   ,   9   ,   7   ,   55  ,   3   ,   137 ,   421 ,   661 ,   113 ,   925 ,   1649    ,   12405   ,   30035   ,0 };
            const unsigned long dim2250Kuo3Init[]   =   {   1   ,   3   ,   1   ,   13  ,   9   ,   31  ,   25  ,   233 ,   241 ,   239 ,   505 ,   2685    ,   223 ,   4389    ,   28191   ,0 };
            const unsigned long dim2251Kuo3Init[]   =   {   1   ,   1   ,   3   ,   3   ,   23  ,   59  ,   5   ,   181 ,   491 ,   111 ,   1289    ,   3717    ,   4429    ,   15091   ,   25375   ,0 };
            const unsigned long dim2252Kuo3Init[]   =   {   1   ,   3   ,   1   ,   3   ,   25  ,   3   ,   3   ,   7   ,   163 ,   247 ,   1673    ,   3061    ,   2355    ,   12779   ,   20039   ,0 };
            const unsigned long dim2253Kuo3Init[]   =   {   1   ,   1   ,   1   ,   5   ,   3   ,   35  ,   5   ,   107 ,   387 ,   99  ,   1841    ,   3607    ,   2027    ,   6353    ,   15929   ,0 };
            const unsigned long dim2254Kuo3Init[]   =   {   1   ,   1   ,   7   ,   7   ,   29  ,   45  ,   43  ,   207 ,   383 ,   681 ,   769 ,   3939    ,   8161    ,   10235   ,   24209   ,0 };
            const unsigned long dim2255Kuo3Init[]   =   {   1   ,   3   ,   1   ,   5   ,   5   ,   13  ,   73  ,   99  ,   233 ,   327 ,   1429    ,   487 ,   3925    ,   5169    ,   5477    ,0 };
            const unsigned long dim2256Kuo3Init[]   =   {   1   ,   1   ,   1   ,   5   ,   21  ,   63  ,   39  ,   17  ,   227 ,   455 ,   1559    ,   113 ,   3993    ,   7935    ,   8995    ,0 };
            const unsigned long dim2257Kuo3Init[]   =   {   1   ,   1   ,   3   ,   11  ,   17  ,   51  ,   43  ,   231 ,   69  ,   475 ,   761 ,   4011    ,   4007    ,   14829   ,   29961   ,0 };
            const unsigned long dim2258Kuo3Init[]   =   {   1   ,   3   ,   1   ,   5   ,   19  ,   37  ,   49  ,   197 ,   277 ,   391 ,   583 ,   2711    ,   4041    ,   6029    ,   29559   ,0 };
            const unsigned long dim2259Kuo3Init[]   =   {   1   ,   3   ,   3   ,   1   ,   7   ,   53  ,   33  ,   33  ,   325 ,   171 ,   63  ,   2929    ,   641 ,   7143    ,   12585   ,0 };
            const unsigned long dim2260Kuo3Init[]   =   {   1   ,   1   ,   1   ,   9   ,   1   ,   31  ,   23  ,   157 ,   437 ,   355 ,   595 ,   239 ,   1275    ,   4917    ,   31055   ,0 };
            const unsigned long dim2261Kuo3Init[]   =   {   1   ,   3   ,   7   ,   15  ,   15  ,   43  ,   39  ,   51  ,   431 ,   131 ,   741 ,   3473    ,   1947    ,   4223    ,   27943   ,0 };
            const unsigned long dim2262Kuo3Init[]   =   {   1   ,   1   ,   7   ,   1   ,   27  ,   49  ,   5   ,   197 ,   33  ,   611 ,   1131    ,   429 ,   4931    ,   1243    ,   5787    ,0 };
            const unsigned long dim2263Kuo3Init[]   =   {   1   ,   1   ,   7   ,   9   ,   29  ,   19  ,   89  ,   139 ,   239 ,   421 ,   951 ,   3749    ,   1557    ,   3803    ,   7155    ,0 };
            const unsigned long dim2264Kuo3Init[]   =   {   1   ,   1   ,   7   ,   13  ,   15  ,   53  ,   121 ,   123 ,   99  ,   677 ,   1519    ,   3051    ,   5235    ,   8061    ,   21353   ,0 };
            const unsigned long dim2265Kuo3Init[]   =   {   1   ,   1   ,   3   ,   7   ,   21  ,   53  ,   65  ,   35  ,   389 ,   141 ,   253 ,   2867    ,   803 ,   11525   ,   9901    ,0 };
            const unsigned long dim2266Kuo3Init[]   =   {   1   ,   1   ,   7   ,   7   ,   21  ,   37  ,   81  ,   53  ,   121 ,   15  ,   1741    ,   2225    ,   183 ,   7155    ,   7597    ,0 };
            const unsigned long dim2267Kuo3Init[]   =   {   1   ,   1   ,   7   ,   15  ,   29  ,   29  ,   13  ,   197 ,   291 ,   765 ,   1479    ,   2211    ,   5949    ,   5045    ,   32519   ,0 };
            const unsigned long dim2268Kuo3Init[]   =   {   1   ,   3   ,   3   ,   5   ,   5   ,   19  ,   121 ,   219 ,   121 ,   17  ,   1223    ,   2685    ,   8179    ,   391 ,   31257   ,0 };
            const unsigned long dim2269Kuo3Init[]   =   {   1   ,   1   ,   5   ,   7   ,   13  ,   27  ,   29  ,   71  ,   7   ,   95  ,   255 ,   9   ,   2375    ,   3325    ,   29915   ,0 };
            const unsigned long dim2270Kuo3Init[]   =   {   1   ,   3   ,   3   ,   1   ,   1   ,   3   ,   11  ,   43  ,   233 ,   663 ,   715 ,   585 ,   3701    ,   13559   ,   3197    ,0 };
            const unsigned long dim2271Kuo3Init[]   =   {   1   ,   1   ,   7   ,   5   ,   17  ,   45  ,   103 ,   209 ,   61  ,   923 ,   517 ,   249 ,   5775    ,   14537   ,   15371   ,0 };
            const unsigned long dim2272Kuo3Init[]   =   {   1   ,   1   ,   5   ,   1   ,   7   ,   5   ,   123 ,   99  ,   137 ,   277 ,   1071    ,   3371    ,   7095    ,   2765    ,   28945   ,0 };
            const unsigned long dim2273Kuo3Init[]   =   {   1   ,   1   ,   3   ,   15  ,   1   ,   59  ,   85  ,   5   ,   475 ,   855 ,   1793    ,   2479    ,   4989    ,   7155    ,   6403    ,0 };
            const unsigned long dim2274Kuo3Init[]   =   {   1   ,   1   ,   5   ,   5   ,   3   ,   29  ,   41  ,   129 ,   373 ,   839 ,   1031    ,   2845    ,   6911    ,   4793    ,   25609   ,0 };
            const unsigned long dim2275Kuo3Init[]   =   {   1   ,   1   ,   3   ,   5   ,   23  ,   55  ,   79  ,   209 ,   131 ,   893 ,   1933    ,   1391    ,   6873    ,   407 ,   28823   ,0 };
            const unsigned long dim2276Kuo3Init[]   =   {   1   ,   1   ,   3   ,   15  ,   15  ,   57  ,   37  ,   153 ,   415 ,   947 ,   1959    ,   1513    ,   7961    ,   4931    ,   749 ,0 };
            const unsigned long dim2277Kuo3Init[]   =   {   1   ,   1   ,   3   ,   3   ,   21  ,   5   ,   27  ,   169 ,   349 ,   541 ,   533 ,   3411    ,   945 ,   5275    ,   665 ,0 };
            const unsigned long dim2278Kuo3Init[]   =   {   1   ,   1   ,   7   ,   15  ,   15  ,   17  ,   97  ,   225 ,   85  ,   887 ,   1251    ,   1485    ,   3987    ,   10753   ,   18013   ,0 };
            const unsigned long dim2279Kuo3Init[]   =   {   1   ,   3   ,   1   ,   15  ,   19  ,   53  ,   117 ,   153 ,   279 ,   397 ,   1929    ,   3621    ,   3887    ,   7753    ,   22075   ,0 };
            const unsigned long dim2280Kuo3Init[]   =   {   1   ,   3   ,   1   ,   5   ,   7   ,   21  ,   97  ,   63  ,   403 ,   715 ,   579 ,   1645    ,   2069    ,   9279    ,   14291   ,0 };
            const unsigned long dim2281Kuo3Init[]   =   {   1   ,   1   ,   1   ,   3   ,   21  ,   17  ,   63  ,   61  ,   207 ,   795 ,   737 ,   4047    ,   3559    ,   6971    ,   14903   ,0 };
            const unsigned long dim2282Kuo3Init[]   =   {   1   ,   3   ,   7   ,   5   ,   29  ,   1   ,   103 ,   109 ,   57  ,   485 ,   1373    ,   3399    ,   6351    ,   13037   ,   17193   ,0 };
            const unsigned long dim2283Kuo3Init[]   =   {   1   ,   1   ,   7   ,   7   ,   5   ,   15  ,   105 ,   225 ,   215 ,   53  ,   1125    ,   3255    ,   3277    ,   2125    ,   23359   ,0 };
            const unsigned long dim2284Kuo3Init[]   =   {   1   ,   3   ,   1   ,   7   ,   31  ,   39  ,   19  ,   163 ,   197 ,   831 ,   1959    ,   1227    ,   6515    ,   8839    ,   20105   ,0 };
            const unsigned long dim2285Kuo3Init[]   =   {   1   ,   3   ,   3   ,   1   ,   27  ,   31  ,   89  ,   133 ,   451 ,   371 ,   1513    ,   1595    ,   771 ,   7637    ,   32403   ,0 };
            const unsigned long dim2286Kuo3Init[]   =   {   1   ,   3   ,   3   ,   13  ,   29  ,   45  ,   29  ,   157 ,   117 ,   461 ,   913 ,   3907    ,   4225    ,   5617    ,   28917   ,0 };
            const unsigned long dim2287Kuo3Init[]   =   {   1   ,   3   ,   7   ,   3   ,   13  ,   63  ,   113 ,   49  ,   137 ,   641 ,   275 ,   1179    ,   8139    ,   519 ,   5851    ,0 };
            const unsigned long dim2288Kuo3Init[]   =   {   1   ,   1   ,   1   ,   1   ,   31  ,   43  ,   61  ,   193 ,   173 ,   651 ,   1491    ,   1005    ,   7559    ,   1963    ,   8323    ,0 };
            const unsigned long dim2289Kuo3Init[]   =   {   1   ,   3   ,   5   ,   5   ,   23  ,   23  ,   11  ,   111 ,   95  ,   487 ,   1091    ,   175 ,   25  ,   10557   ,   24493   ,0 };
            const unsigned long dim2290Kuo3Init[]   =   {   1   ,   1   ,   1   ,   9   ,   1   ,   31  ,   97  ,   175 ,   249 ,   309 ,   1005    ,   3777    ,   5551    ,   2605    ,   28071   ,0 };
            const unsigned long dim2291Kuo3Init[]   =   {   1   ,   3   ,   3   ,   7   ,   13  ,   1   ,   113 ,   103 ,   169 ,   853 ,   35  ,   3645    ,   6907    ,   13713   ,   12811   ,0 };
            const unsigned long dim2292Kuo3Init[]   =   {   1   ,   1   ,   7   ,   3   ,   25  ,   49  ,   53  ,   23  ,   33  ,   967 ,   2039    ,   2057    ,   799 ,   1527    ,   3427    ,0 };
            const unsigned long dim2293Kuo3Init[]   =   {   1   ,   1   ,   5   ,   7   ,   29  ,   43  ,   81  ,   155 ,   59  ,   201 ,   1571    ,   1719    ,   3343    ,   3231    ,   27929   ,0 };
            const unsigned long dim2294Kuo3Init[]   =   {   1   ,   1   ,   5   ,   1   ,   27  ,   17  ,   1   ,   159 ,   209 ,   749 ,   1883    ,   1885    ,   2511    ,   15957   ,   3471    ,0 };
            const unsigned long dim2295Kuo3Init[]   =   {   1   ,   3   ,   1   ,   5   ,   7   ,   51  ,   15  ,   245 ,   57  ,   627 ,   1315    ,   673 ,   5329    ,   1043    ,   555 ,0 };
            const unsigned long dim2296Kuo3Init[]   =   {   1   ,   1   ,   5   ,   11  ,   13  ,   55  ,   103 ,   157 ,   329 ,   171 ,   393 ,   3967    ,   373 ,   7225    ,   25133   ,0 };
            const unsigned long dim2297Kuo3Init[]   =   {   1   ,   1   ,   5   ,   7   ,   5   ,   51  ,   109 ,   11  ,   111 ,   859 ,   165 ,   3991    ,   7187    ,   697 ,   28035   ,0 };
            const unsigned long dim2298Kuo3Init[]   =   {   1   ,   3   ,   1   ,   9   ,   23  ,   55  ,   59  ,   173 ,   55  ,   105 ,   1103    ,   1569    ,   3099    ,   10483   ,   18821   ,0 };
            const unsigned long dim2299Kuo3Init[]   =   {   1   ,   3   ,   7   ,   5   ,   21  ,   3   ,   23  ,   247 ,   13  ,   59  ,   1327    ,   2393    ,   5023    ,   11163   ,   21443   ,0 };
            const unsigned long dim2300Kuo3Init[]   =   {   1   ,   3   ,   3   ,   15  ,   19  ,   47  ,   79  ,   45  ,   127 ,   313 ,   1609    ,   273 ,   4729    ,   7363    ,   13541   ,0 };
            const unsigned long dim2301Kuo3Init[]   =   {   1   ,   3   ,   7   ,   15  ,   21  ,   61  ,   91  ,   247 ,   477 ,   539 ,   1969    ,   3541    ,   3983    ,   8339    ,   32499   ,0 };
            const unsigned long dim2302Kuo3Init[]   =   {   1   ,   3   ,   7   ,   11  ,   13  ,   7   ,   47  ,   57  ,   299 ,   907 ,   701 ,   2039    ,   5509    ,   7489    ,   30175   ,0 };
            const unsigned long dim2303Kuo3Init[]   =   {   1   ,   3   ,   7   ,   7   ,   9   ,   21  ,   125 ,   241 ,   69  ,   405 ,   1989    ,   3777    ,   613 ,   9827    ,   19447   ,0 };
            const unsigned long dim2304Kuo3Init[]   =   {   1   ,   3   ,   7   ,   3   ,   19  ,   39  ,   19  ,   43  ,   1   ,   1007    ,   1829    ,   3099    ,   4001    ,   2181    ,   5849    ,0 };
            const unsigned long dim2305Kuo3Init[]   =   {   1   ,   1   ,   7   ,   1   ,   9   ,   49  ,   21  ,   103 ,   107 ,   473 ,   1913    ,   2983    ,   5669    ,   977 ,   21281   ,0 };
            const unsigned long dim2306Kuo3Init[]   =   {   1   ,   3   ,   5   ,   3   ,   3   ,   37  ,   79  ,   239 ,   425 ,   949 ,   1953    ,   3867    ,   2537    ,   16117   ,   945 ,0 };
            const unsigned long dim2307Kuo3Init[]   =   {   1   ,   3   ,   5   ,   13  ,   23  ,   7   ,   119 ,   221 ,   321 ,   881 ,   1505    ,   3993    ,   3677    ,   4095    ,   10501   ,0 };
            const unsigned long dim2308Kuo3Init[]   =   {   1   ,   1   ,   1   ,   3   ,   7   ,   17  ,   5   ,   223 ,   3   ,   891 ,   1699    ,   823 ,   3839    ,   4113    ,   16039   ,0 };
            const unsigned long dim2309Kuo3Init[]   =   {   1   ,   1   ,   7   ,   5   ,   27  ,   35  ,   67  ,   89  ,   291 ,   521 ,   411 ,   3025    ,   1427    ,   12131   ,   4449    ,0 };
            const unsigned long dim2310Kuo3Init[]   =   {   1   ,   1   ,   3   ,   7   ,   13  ,   61  ,   103 ,   113 ,   195 ,   269 ,   75  ,   3879    ,   7669    ,   9403    ,   13341   ,0 };
            const unsigned long dim2311Kuo3Init[]   =   {   1   ,   3   ,   7   ,   13  ,   21  ,   17  ,   101 ,   135 ,   153 ,   615 ,   755 ,   2589    ,   653 ,   11187   ,   2997    ,0 };
            const unsigned long dim2312Kuo3Init[]   =   {   1   ,   3   ,   1   ,   9   ,   11  ,   53  ,   61  ,   171 ,   423 ,   857 ,   465 ,   819 ,   4341    ,   6377    ,   21305   ,0 };
            const unsigned long dim2313Kuo3Init[]   =   {   1   ,   3   ,   5   ,   3   ,   17  ,   37  ,   63  ,   91  ,   157 ,   877 ,   1801    ,   587 ,   6183    ,   12923   ,   15991   ,0 };
            const unsigned long dim2314Kuo3Init[]   =   {   1   ,   3   ,   7   ,   5   ,   27  ,   63  ,   123 ,   27  ,   243 ,   375 ,   191 ,   3333    ,   965 ,   889 ,   15771   ,0 };
            const unsigned long dim2315Kuo3Init[]   =   {   1   ,   3   ,   7   ,   15  ,   23  ,   51  ,   49  ,   187 ,   201 ,   861 ,   1581    ,   1045    ,   493 ,   9939    ,   30893   ,0 };
            const unsigned long dim2316Kuo3Init[]   =   {   1   ,   3   ,   3   ,   15  ,   1   ,   31  ,   93  ,   125 ,   57  ,   147 ,   663 ,   2733    ,   4601    ,   16043   ,   6875    ,0 };
            const unsigned long dim2317Kuo3Init[]   =   {   1   ,   3   ,   7   ,   11  ,   15  ,   1   ,   45  ,   251 ,   173 ,   807 ,   1301    ,   507 ,   645 ,   10383   ,   14737   ,0 };
            const unsigned long dim2318Kuo3Init[]   =   {   1   ,   3   ,   7   ,   7   ,   27  ,   3   ,   3   ,   127 ,   153 ,   151 ,   1987    ,   2255    ,   5201    ,   11095   ,   20757   ,0 };
            const unsigned long dim2319Kuo3Init[]   =   {   1   ,   1   ,   5   ,   1   ,   13  ,   49  ,   69  ,   55  ,   205 ,   547 ,   701 ,   1449    ,   6689    ,   3047    ,   12025   ,0 };
            const unsigned long dim2320Kuo3Init[]   =   {   1   ,   1   ,   7   ,   9   ,   13  ,   27  ,   31  ,   87  ,   263 ,   499 ,   1573    ,   619 ,   6697    ,   9797    ,   7673    ,0 };
            const unsigned long dim2321Kuo3Init[]   =   {   1   ,   1   ,   3   ,   3   ,   1   ,   43  ,   45  ,   81  ,   241 ,   143 ,   2045    ,   2847    ,   7361    ,   12253   ,   27815   ,0 };
            const unsigned long dim2322Kuo3Init[]   =   {   1   ,   1   ,   3   ,   11  ,   27  ,   13  ,   71  ,   119 ,   365 ,   245 ,   113 ,   2635    ,   7157    ,   4951    ,   3381    ,0 };
            const unsigned long dim2323Kuo3Init[]   =   {   1   ,   1   ,   7   ,   7   ,   25  ,   23  ,   101 ,   21  ,   247 ,   149 ,   431 ,   3165    ,   4109    ,   14305   ,   22549   ,0 };
            const unsigned long dim2324Kuo3Init[]   =   {   1   ,   1   ,   5   ,   1   ,   3   ,   51  ,   79  ,   7   ,   473 ,   337 ,   709 ,   127 ,   555 ,   12147   ,   19843   ,0 };
            const unsigned long dim2325Kuo3Init[]   =   {   1   ,   1   ,   3   ,   5   ,   21  ,   35  ,   125 ,   51  ,   373 ,   859 ,   1429    ,   2623    ,   787 ,   9669    ,   14933   ,0 };
            const unsigned long dim2326Kuo3Init[]   =   {   1   ,   1   ,   1   ,   9   ,   25  ,   45  ,   47  ,   251 ,   475 ,   823 ,   1361    ,   1393    ,   6787    ,   14571   ,   11311   ,0 };
            const unsigned long dim2327Kuo3Init[]   =   {   1   ,   3   ,   3   ,   13  ,   1   ,   27  ,   75  ,   153 ,   19  ,   373 ,   209 ,   3083    ,   8071    ,   15401   ,   14619   ,0 };
            const unsigned long dim2328Kuo3Init[]   =   {   1   ,   1   ,   1   ,   13  ,   25  ,   39  ,   73  ,   235 ,   251 ,   245 ,   247 ,   1909    ,   2057    ,   14937   ,   16509   ,0 };
            const unsigned long dim2329Kuo3Init[]   =   {   1   ,   1   ,   5   ,   11  ,   31  ,   15  ,   117 ,   167 ,   79  ,   911 ,   403 ,   3595    ,   4633    ,   11017   ,   5021    ,0 };
            const unsigned long dim2330Kuo3Init[]   =   {   1   ,   1   ,   1   ,   9   ,   9   ,   37  ,   19  ,   179 ,   417 ,   285 ,   1237    ,   2667    ,   2131    ,   5943    ,   21265   ,0 };
            const unsigned long dim2331Kuo3Init[]   =   {   1   ,   3   ,   7   ,   13  ,   29  ,   53  ,   89  ,   223 ,   37  ,   101 ,   1255    ,   97  ,   3813    ,   8543    ,   17949   ,0 };
            const unsigned long dim2332Kuo3Init[]   =   {   1   ,   1   ,   1   ,   13  ,   25  ,   61  ,   33  ,   145 ,   271 ,   639 ,   1539    ,   2967    ,   1043    ,   10259   ,   27997   ,0 };
            const unsigned long dim2333Kuo3Init[]   =   {   1   ,   1   ,   3   ,   7   ,   11  ,   17  ,   21  ,   21  ,   159 ,   485 ,   1855    ,   3139    ,   6605    ,   9567    ,   21401   ,0 };
            const unsigned long dim2334Kuo3Init[]   =   {   1   ,   3   ,   7   ,   9   ,   27  ,   23  ,   23  ,   121 ,   17  ,   45  ,   537 ,   3155    ,   2409    ,   15359   ,   16105   ,0 };
            const unsigned long dim2335Kuo3Init[]   =   {   1   ,   1   ,   3   ,   13  ,   11  ,   51  ,   87  ,   251 ,   297 ,   227 ,   411 ,   3737    ,   3649    ,   14697   ,   665 ,0 };
            const unsigned long dim2336Kuo3Init[]   =   {   1   ,   3   ,   5   ,   11  ,   23  ,   23  ,   3   ,   75  ,   99  ,   149 ,   1505    ,   1137    ,   6495    ,   12629   ,   29339   ,0 };
            const unsigned long dim2337Kuo3Init[]   =   {   1   ,   3   ,   1   ,   9   ,   29  ,   1   ,   83  ,   233 ,   387 ,   801 ,   1387    ,   2721    ,   1399    ,   2275    ,   19865   ,0 };
            const unsigned long dim2338Kuo3Init[]   =   {   1   ,   3   ,   7   ,   7   ,   31  ,   33  ,   1   ,   253 ,   71  ,   435 ,   781 ,   2991    ,   6881    ,   13183   ,   10419   ,0 };
            const unsigned long dim2339Kuo3Init[]   =   {   1   ,   1   ,   5   ,   11  ,   19  ,   61  ,   25  ,   167 ,   79  ,   173 ,   447 ,   911 ,   6509    ,   5531    ,   9387    ,0 };
            const unsigned long dim2340Kuo3Init[]   =   {   1   ,   1   ,   1   ,   15  ,   31  ,   23  ,   117 ,   1   ,   119 ,   939 ,   1277    ,   2517    ,   2019    ,   4979    ,   4065    ,0 };
            const unsigned long dim2341Kuo3Init[]   =   {   1   ,   3   ,   3   ,   9   ,   5   ,   49  ,   87  ,   187 ,   159 ,   941 ,   303 ,   217 ,   5931    ,   9065    ,   20043   ,0 };
            const unsigned long dim2342Kuo3Init[]   =   {   1   ,   3   ,   1   ,   11  ,   31  ,   43  ,   25  ,   45  ,   61  ,   361 ,   2043    ,   817 ,   8023    ,   14309   ,   6781    ,0 };
            const unsigned long dim2343Kuo3Init[]   =   {   1   ,   3   ,   7   ,   9   ,   27  ,   45  ,   97  ,   189 ,   387 ,   379 ,   621 ,   2449    ,   39  ,   10313   ,   14603   ,0 };
            const unsigned long dim2344Kuo3Init[]   =   {   1   ,   3   ,   1   ,   5   ,   5   ,   45  ,   89  ,   251 ,   103 ,   33  ,   1141    ,   2957    ,   6511    ,   8497    ,   16609   ,0 };
            const unsigned long dim2345Kuo3Init[]   =   {   1   ,   3   ,   5   ,   15  ,   19  ,   11  ,   115 ,   255 ,   277 ,   419 ,   15  ,   999 ,   7509    ,   14191   ,   2749    ,0 };
            const unsigned long dim2346Kuo3Init[]   =   {   1   ,   1   ,   1   ,   7   ,   7   ,   51  ,   127 ,   225 ,   295 ,   389 ,   1039    ,   2775    ,   5531    ,   10645   ,   13391   ,0 };
            const unsigned long dim2347Kuo3Init[]   =   {   1   ,   3   ,   3   ,   11  ,   13  ,   63  ,   37  ,   5   ,   3   ,   737 ,   123 ,   777 ,   1469    ,   4009    ,   17037   ,0 };
            const unsigned long dim2348Kuo3Init[]   =   {   1   ,   1   ,   3   ,   5   ,   31  ,   43  ,   59  ,   119 ,   287 ,   973 ,   541 ,   751 ,   5697    ,   7909    ,   28673   ,0 };
            const unsigned long dim2349Kuo3Init[]   =   {   1   ,   1   ,   5   ,   7   ,   15  ,   61  ,   71  ,   213 ,   465 ,   961 ,   1399    ,   3227    ,   571 ,   10759   ,   13947   ,0 };
            const unsigned long dim2350Kuo3Init[]   =   {   1   ,   1   ,   5   ,   9   ,   17  ,   53  ,   85  ,   11  ,   185 ,   397 ,   567 ,   3763    ,   483 ,   6243    ,   31191   ,0 };
            const unsigned long dim2351Kuo3Init[]   =   {   1   ,   3   ,   7   ,   1   ,   31  ,   57  ,   13  ,   255 ,   495 ,   315 ,   763 ,   3153    ,   5543    ,   5719    ,   2235    ,0 };
            const unsigned long dim2352Kuo3Init[]   =   {   1   ,   1   ,   1   ,   1   ,   13  ,   25  ,   43  ,   125 ,   391 ,   483 ,   1029    ,   3739    ,   4991    ,   891 ,   24935   ,0 };
            const unsigned long dim2353Kuo3Init[]   =   {   1   ,   3   ,   7   ,   3   ,   21  ,   3   ,   79  ,   205 ,   87  ,   3   ,   1859    ,   1467    ,   6129    ,   14431   ,   18881   ,0 };
            const unsigned long dim2354Kuo3Init[]   =   {   1   ,   1   ,   1   ,   11  ,   15  ,   21  ,   79  ,   145 ,   253 ,   595 ,   1939    ,   2195    ,   5735    ,   5383    ,   5311    ,0 };
            const unsigned long dim2355Kuo3Init[]   =   {   1   ,   1   ,   7   ,   3   ,   13  ,   61  ,   89  ,   3   ,   231 ,   867 ,   1385    ,   183 ,   957 ,   791 ,   2879    ,0 };
            const unsigned long dim2356Kuo3Init[]   =   {   1   ,   3   ,   5   ,   7   ,   1   ,   39  ,   47  ,   107 ,   217 ,   725 ,   1707    ,   2241    ,   2813    ,   5373    ,   23191   ,0 };
            const unsigned long dim2357Kuo3Init[]   =   {   1   ,   1   ,   1   ,   5   ,   27  ,   29  ,   119 ,   159 ,   287 ,   343 ,   693 ,   519 ,   2127    ,   6083    ,   2991    ,0 };
            const unsigned long dim2358Kuo3Init[]   =   {   1   ,   3   ,   3   ,   11  ,   3   ,   51  ,   11  ,   119 ,   337 ,   567 ,   1175    ,   235 ,   6801    ,   1643    ,   26593   ,0 };
            const unsigned long dim2359Kuo3Init[]   =   {   1   ,   1   ,   5   ,   3   ,   27  ,   49  ,   91  ,   249 ,   489 ,   257 ,   1735    ,   3227    ,   1419    ,   4401    ,   12263   ,0 };
            const unsigned long dim2360Kuo3Init[]   =   {   1   ,   1   ,   3   ,   13  ,   27  ,   37  ,   101 ,   43  ,   289 ,   313 ,   1031    ,   2921    ,   3655    ,   6781    ,   29593   ,0 };
            const unsigned long dim2361Kuo3Init[]   =   {   1   ,   1   ,   5   ,   5   ,   17  ,   23  ,   117 ,   129 ,   261 ,   919 ,   265 ,   1561    ,   1297    ,   815 ,   4315    ,0 };
            const unsigned long dim2362Kuo3Init[]   =   {   1   ,   1   ,   3   ,   13  ,   23  ,   31  ,   95  ,   231 ,   99  ,   909 ,   1477    ,   3385    ,   5679    ,   15319   ,   28849   ,0 };
            const unsigned long dim2363Kuo3Init[]   =   {   1   ,   1   ,   5   ,   11  ,   25  ,   27  ,   69  ,   83  ,   15  ,   31  ,   1469    ,   751 ,   2339    ,   3251    ,   17343   ,0 };
            const unsigned long dim2364Kuo3Init[]   =   {   1   ,   1   ,   1   ,   9   ,   5   ,   19  ,   49  ,   25  ,   145 ,   763 ,   667 ,   2003    ,   6905    ,   15623   ,   3533    ,0 };
            const unsigned long dim2365Kuo3Init[]   =   {   1   ,   1   ,   1   ,   5   ,   5   ,   17  ,   65  ,   151 ,   333 ,   639 ,   1885    ,   3561    ,   1083    ,   7875    ,   22647   ,0 };
            const unsigned long dim2366Kuo3Init[]   =   {   1   ,   3   ,   5   ,   7   ,   25  ,   45  ,   79  ,   175 ,   87  ,   305 ,   1447    ,   1935    ,   3627    ,   6007    ,   27877   ,0 };
            const unsigned long dim2367Kuo3Init[]   =   {   1   ,   3   ,   7   ,   1   ,   11  ,   37  ,   63  ,   37  ,   323 ,   405 ,   1851    ,   435 ,   1501    ,   1345    ,   13729   ,0 };
            const unsigned long dim2368Kuo3Init[]   =   {   1   ,   3   ,   7   ,   13  ,   9   ,   37  ,   115 ,   151 ,   255 ,   795 ,   811 ,   681 ,   5631    ,   489 ,   21813   ,0 };
            const unsigned long dim2369Kuo3Init[]   =   {   1   ,   1   ,   1   ,   11  ,   13  ,   25  ,   61  ,   91  ,   367 ,   289 ,   1093    ,   41  ,   5073    ,   1105    ,   28059   ,0 };
            const unsigned long dim2370Kuo3Init[]   =   {   1   ,   1   ,   3   ,   13  ,   29  ,   53  ,   13  ,   173 ,   83  ,   313 ,   1223    ,   719 ,   517 ,   11689   ,   28719   ,0 };
            const unsigned long dim2371Kuo3Init[]   =   {   1   ,   1   ,   5   ,   3   ,   5   ,   21  ,   71  ,   229 ,   95  ,   187 ,   739 ,   2607    ,   601 ,   2359    ,   10603   ,0 };
            const unsigned long dim2372Kuo3Init[]   =   {   1   ,   1   ,   1   ,   5   ,   25  ,   35  ,   55  ,   131 ,   173 ,   677 ,   1887    ,   3293    ,   7641    ,   13535   ,   25561   ,0 };
            const unsigned long dim2373Kuo3Init[]   =   {   1   ,   3   ,   5   ,   13  ,   29  ,   47  ,   27  ,   1   ,   131 ,   345 ,   1751    ,   1807    ,   235 ,   1891    ,   5909    ,0 };
            const unsigned long dim2374Kuo3Init[]   =   {   1   ,   1   ,   3   ,   3   ,   15  ,   41  ,   55  ,   201 ,   263 ,   831 ,   401 ,   2685    ,   7403    ,   11827   ,   15353   ,0 };
            const unsigned long dim2375Kuo3Init[]   =   {   1   ,   3   ,   1   ,   7   ,   5   ,   3   ,   99  ,   61  ,   433 ,   641 ,   1817    ,   891 ,   1937    ,   7179    ,   14391   ,0 };
            const unsigned long dim2376Kuo3Init[]   =   {   1   ,   1   ,   5   ,   15  ,   19  ,   33  ,   37  ,   223 ,   329 ,   5   ,   1897    ,   2737    ,   4339    ,   9325    ,   12199   ,0 };
            const unsigned long dim2377Kuo3Init[]   =   {   1   ,   1   ,   5   ,   3   ,   3   ,   23  ,   47  ,   125 ,   85  ,   61  ,   1133    ,   2467    ,   1991    ,   8059    ,   153 ,0 };
            const unsigned long dim2378Kuo3Init[]   =   {   1   ,   1   ,   1   ,   15  ,   29  ,   45  ,   117 ,   133 ,   447 ,   627 ,   525 ,   173 ,   1165    ,   11939   ,   21807   ,0 };
            const unsigned long dim2379Kuo3Init[]   =   {   1   ,   1   ,   1   ,   3   ,   1   ,   15  ,   5   ,   49  ,   237 ,   675 ,   717 ,   2783    ,   4287    ,   11485   ,   3797    ,0 };
            const unsigned long dim2380Kuo3Init[]   =   {   1   ,   3   ,   5   ,   15  ,   19  ,   63  ,   23  ,   79  ,   313 ,   1007    ,   1321    ,   229 ,   7537    ,   825 ,   29127   ,0 };
            const unsigned long dim2381Kuo3Init[]   =   {   1   ,   1   ,   3   ,   15  ,   15  ,   55  ,   1   ,   229 ,   221 ,   359 ,   1021    ,   3307    ,   4993    ,   8781    ,   29275   ,0 };
            const unsigned long dim2382Kuo3Init[]   =   {   1   ,   3   ,   3   ,   11  ,   17  ,   57  ,   37  ,   207 ,   283 ,   599 ,   1867    ,   31  ,   6631    ,   10605   ,   29671   ,0 };
            const unsigned long dim2383Kuo3Init[]   =   {   1   ,   3   ,   3   ,   9   ,   25  ,   9   ,   111 ,   87  ,   325 ,   515 ,   1249    ,   2027    ,   7783    ,   6473    ,   23737   ,0 };
            const unsigned long dim2384Kuo3Init[]   =   {   1   ,   3   ,   3   ,   3   ,   31  ,   11  ,   39  ,   161 ,   65  ,   259 ,   1747    ,   597 ,   913 ,   12397   ,   25611   ,0 };
            const unsigned long dim2385Kuo3Init[]   =   {   1   ,   3   ,   5   ,   11  ,   31  ,   37  ,   27  ,   157 ,   411 ,   785 ,   1699    ,   467 ,   3993    ,   6357    ,   11039   ,0 };
            const unsigned long dim2386Kuo3Init[]   =   {   1   ,   3   ,   3   ,   7   ,   5   ,   41  ,   61  ,   103 ,   33  ,   789 ,   1861    ,   319 ,   7667    ,   10169   ,   10823   ,0 };
            const unsigned long dim2387Kuo3Init[]   =   {   1   ,   1   ,   7   ,   5   ,   31  ,   37  ,   45  ,   175 ,   453 ,   423 ,   291 ,   3989    ,   2215    ,   11379   ,   28713   ,0 };
            const unsigned long dim2388Kuo3Init[]   =   {   1   ,   3   ,   3   ,   1   ,   23  ,   39  ,   75  ,   103 ,   341 ,   961 ,   2039    ,   1539    ,   7831    ,   1665    ,   18579   ,0 };
            const unsigned long dim2389Kuo3Init[]   =   {   1   ,   1   ,   5   ,   13  ,   19  ,   25  ,   27  ,   183 ,   241 ,   539 ,   257 ,   3   ,   3709    ,   8133    ,   31081   ,0 };
            const unsigned long dim2390Kuo3Init[]   =   {   1   ,   1   ,   3   ,   15  ,   1   ,   53  ,   27  ,   229 ,   299 ,   913 ,   1277    ,   2687    ,   831 ,   1065    ,   4409    ,0 };
            const unsigned long dim2391Kuo3Init[]   =   {   1   ,   1   ,   7   ,   9   ,   13  ,   47  ,   127 ,   145 ,   481 ,   751 ,   1745    ,   3645    ,   2853    ,   5601    ,   14991   ,0 };
            const unsigned long dim2392Kuo3Init[]   =   {   1   ,   3   ,   7   ,   3   ,   3   ,   39  ,   31  ,   191 ,   69  ,   707 ,   1757    ,   223 ,   5055    ,   14777   ,   4433    ,0 };
            const unsigned long dim2393Kuo3Init[]   =   {   1   ,   1   ,   7   ,   15  ,   27  ,   33  ,   77  ,   75  ,   357 ,   1007    ,   1373    ,   3445    ,   2597    ,   6085    ,   19441   ,0 };
            const unsigned long dim2394Kuo3Init[]   =   {   1   ,   1   ,   1   ,   9   ,   13  ,   53  ,   77  ,   61  ,   425 ,   245 ,   1149    ,   1239    ,   1989    ,   1293    ,   12971   ,0 };
            const unsigned long dim2395Kuo3Init[]   =   {   1   ,   1   ,   1   ,   7   ,   23  ,   43  ,   29  ,   89  ,   349 ,   411 ,   2021    ,   2075    ,   4917    ,   11387   ,   27045   ,0 };
            const unsigned long dim2396Kuo3Init[]   =   {   1   ,   1   ,   5   ,   11  ,   27  ,   63  ,   71  ,   19  ,   363 ,   289 ,   1147    ,   2541    ,   4427    ,   591 ,   21287   ,0 };
            const unsigned long dim2397Kuo3Init[]   =   {   1   ,   1   ,   1   ,   3   ,   27  ,   3   ,   1   ,   159 ,   85  ,   651 ,   1109    ,   1885    ,   5639    ,   11725   ,   1527    ,0 };
            const unsigned long dim2398Kuo3Init[]   =   {   1   ,   1   ,   3   ,   1   ,   11  ,   63  ,   77  ,   185 ,   233 ,   129 ,   1721    ,   2419    ,   2043    ,   3121    ,   2159    ,0 };
            const unsigned long dim2399Kuo3Init[]   =   {   1   ,   1   ,   1   ,   3   ,   11  ,   27  ,   83  ,   193 ,   143 ,   89  ,   1927    ,   3753    ,   4041    ,   10397   ,   1973    ,0 };
            const unsigned long dim2400Kuo3Init[]   =   {   1   ,   1   ,   5   ,   1   ,   9   ,   61  ,   49  ,   29  ,   333 ,   621 ,   1555    ,   4051    ,   3011    ,   12075   ,   4699    ,0 };
            const unsigned long dim2401Kuo3Init[]   =   {   1   ,   3   ,   1   ,   5   ,   25  ,   37  ,   69  ,   81  ,   271 ,   827 ,   763 ,   2973    ,   3729    ,   2339    ,   4783    ,0 };
            const unsigned long dim2402Kuo3Init[]   =   {   1   ,   3   ,   1   ,   7   ,   23  ,   1   ,   95  ,   143 ,   417 ,   517 ,   1991    ,   1423    ,   7511    ,   125 ,   18675   ,0 };
            const unsigned long dim2403Kuo3Init[]   =   {   1   ,   3   ,   5   ,   9   ,   3   ,   43  ,   71  ,   73  ,   443 ,   335 ,   737 ,   1121    ,   8001    ,   13165   ,   11155   ,0 };
            const unsigned long dim2404Kuo3Init[]   =   {   1   ,   3   ,   1   ,   11  ,   1   ,   21  ,   39  ,   149 ,   261 ,   963 ,   117 ,   3421    ,   2603    ,   4035    ,   4801    ,0 };
            const unsigned long dim2405Kuo3Init[]   =   {   1   ,   3   ,   5   ,   7   ,   23  ,   63  ,   101 ,   71  ,   495 ,   649 ,   469 ,   2559    ,   541 ,   4499    ,   20357   ,0 };
            const unsigned long dim2406Kuo3Init[]   =   {   1   ,   1   ,   5   ,   7   ,   31  ,   37  ,   107 ,   239 ,   197 ,   695 ,   1403    ,   2193    ,   5583    ,   16019   ,   11599   ,0 };
            const unsigned long dim2407Kuo3Init[]   =   {   1   ,   1   ,   5   ,   3   ,   21  ,   25  ,   127 ,   255 ,   259 ,   209 ,   1491    ,   349 ,   737 ,   9781    ,   3697    ,0 };
            const unsigned long dim2408Kuo3Init[]   =   {   1   ,   1   ,   5   ,   9   ,   3   ,   47  ,   43  ,   123 ,   315 ,   983 ,   349 ,   2139    ,   6233    ,   731 ,   745 ,0 };
            const unsigned long dim2409Kuo3Init[]   =   {   1   ,   1   ,   1   ,   1   ,   25  ,   63  ,   95  ,   101 ,   483 ,   439 ,   1073    ,   1813    ,   3809    ,   11987   ,   4607    ,0 };
            const unsigned long dim2410Kuo3Init[]   =   {   1   ,   1   ,   3   ,   5   ,   23  ,   11  ,   11  ,   101 ,   9   ,   599 ,   931 ,   39  ,   6593    ,   14875   ,   13557   ,0 };
            const unsigned long dim2411Kuo3Init[]   =   {   1   ,   3   ,   3   ,   3   ,   21  ,   59  ,   107 ,   131 ,   221 ,   351 ,   1057    ,   3689    ,   7471    ,   4659    ,   27303   ,0 };
            const unsigned long dim2412Kuo3Init[]   =   {   1   ,   1   ,   5   ,   5   ,   19  ,   59  ,   79  ,   81  ,   169 ,   927 ,   1439    ,   2305    ,   2413    ,   2935    ,   28571   ,0 };
            const unsigned long dim2413Kuo3Init[]   =   {   1   ,   3   ,   7   ,   5   ,   17  ,   37  ,   81  ,   125 ,   347 ,   81  ,   1029    ,   705 ,   7321    ,   4877    ,   28743   ,0 };
            const unsigned long dim2414Kuo3Init[]   =   {   1   ,   3   ,   5   ,   5   ,   21  ,   27  ,   113 ,   161 ,   177 ,   49  ,   1655    ,   4035    ,   4987    ,   5545    ,   13137   ,0 };
            const unsigned long dim2415Kuo3Init[]   =   {   1   ,   1   ,   5   ,   13  ,   5   ,   19  ,   117 ,   13  ,   379 ,   449 ,   765 ,   1395    ,   1051    ,   14783   ,   22283   ,0 };
            const unsigned long dim2416Kuo3Init[]   =   {   1   ,   1   ,   7   ,   13  ,   25  ,   57  ,   59  ,   47  ,   15  ,   425 ,   1359    ,   1019    ,   4933    ,   6289    ,   2943    ,0 };
            const unsigned long dim2417Kuo3Init[]   =   {   1   ,   3   ,   7   ,   15  ,   7   ,   31  ,   95  ,   245 ,   243 ,   515 ,   893 ,   2451    ,   83  ,   6561    ,   20993   ,0 };
            const unsigned long dim2418Kuo3Init[]   =   {   1   ,   3   ,   1   ,   13  ,   7   ,   45  ,   39  ,   227 ,   245 ,   793 ,   1103    ,   3279    ,   6967    ,   14771   ,   23199   ,0 };
            const unsigned long dim2419Kuo3Init[]   =   {   1   ,   3   ,   5   ,   9   ,   21  ,   5   ,   103 ,   27  ,   509 ,   187 ,   1727    ,   1281    ,   923 ,   8305    ,   29315   ,0 };
            const unsigned long dim2420Kuo3Init[]   =   {   1   ,   3   ,   1   ,   13  ,   13  ,   3   ,   3   ,   57  ,   231 ,   609 ,   695 ,   3461    ,   6565    ,   16003   ,   12719   ,0 };
            const unsigned long dim2421Kuo3Init[]   =   {   1   ,   1   ,   5   ,   7   ,   13  ,   55  ,   49  ,   111 ,   111 ,   947 ,   1047    ,   3745    ,   8141    ,   13181   ,   25703   ,0 };
            const unsigned long dim2422Kuo3Init[]   =   {   1   ,   3   ,   7   ,   15  ,   17  ,   51  ,   37  ,   253 ,   97  ,   363 ,   1617    ,   1201    ,   3199    ,   11871   ,   29389   ,0 };
            const unsigned long dim2423Kuo3Init[]   =   {   1   ,   3   ,   3   ,   5   ,   23  ,   13  ,   73  ,   85  ,   53  ,   699 ,   1707    ,   2137    ,   4537    ,   15375   ,   28251   ,0 };
            const unsigned long dim2424Kuo3Init[]   =   {   1   ,   3   ,   3   ,   7   ,   17  ,   35  ,   29  ,   251 ,   455 ,   153 ,   1703    ,   1535    ,   4801    ,   12889   ,   5959    ,0 };
            const unsigned long dim2425Kuo3Init[]   =   {   1   ,   3   ,   1   ,   1   ,   15  ,   61  ,   75  ,   21  ,   211 ,   425 ,   1459    ,   1347    ,   7525    ,   15465   ,   28191   ,0 };
            const unsigned long dim2426Kuo3Init[]   =   {   1   ,   1   ,   5   ,   15  ,   3   ,   1   ,   7   ,   163 ,   317 ,   85  ,   385 ,   3183    ,   1901    ,   6265    ,   13773   ,0 };
            const unsigned long dim2427Kuo3Init[]   =   {   1   ,   1   ,   5   ,   1   ,   31  ,   43  ,   63  ,   15  ,   157 ,   711 ,   617 ,   2283    ,   1269    ,   15661   ,   28559   ,0 };
            const unsigned long dim2428Kuo3Init[]   =   {   1   ,   3   ,   5   ,   1   ,   19  ,   57  ,   1   ,   149 ,   183 ,   9   ,   1587    ,   2711    ,   6563    ,   9291    ,   287 ,0 };
            const unsigned long dim2429Kuo3Init[]   =   {   1   ,   3   ,   1   ,   7   ,   3   ,   47  ,   11  ,   85  ,   97  ,   245 ,   1673    ,   1147    ,   2187    ,   2361    ,   4417    ,0 };
            const unsigned long dim2430Kuo3Init[]   =   {   1   ,   3   ,   5   ,   7   ,   9   ,   17  ,   75  ,   3   ,   449 ,   145 ,   1695    ,   597 ,   6931    ,   2281    ,   27291   ,0 };
            const unsigned long dim2431Kuo3Init[]   =   {   1   ,   1   ,   1   ,   7   ,   7   ,   51  ,   53  ,   53  ,   463 ,   669 ,   677 ,   3579    ,   2609    ,   4147    ,   27567   ,0 };
            const unsigned long dim2432Kuo3Init[]   =   {   1   ,   1   ,   5   ,   13  ,   29  ,   63  ,   3   ,   195 ,   507 ,   43  ,   425 ,   2121    ,   8147    ,   1287    ,   15137   ,0 };
            const unsigned long dim2433Kuo3Init[]   =   {   1   ,   1   ,   7   ,   9   ,   17  ,   7   ,   3   ,   139 ,   301 ,   647 ,   817 ,   3657    ,   3217    ,   4651    ,   21497   ,0 };
            const unsigned long dim2434Kuo3Init[]   =   {   1   ,   1   ,   3   ,   13  ,   11  ,   43  ,   87  ,   139 ,   441 ,   13  ,   1555    ,   2855    ,   4891    ,   3835    ,   17465   ,0 };
            const unsigned long dim2435Kuo3Init[]   =   {   1   ,   3   ,   1   ,   3   ,   15  ,   47  ,   29  ,   99  ,   481 ,   377 ,   455 ,   929 ,   1977    ,   7221    ,   2517    ,0 };
            const unsigned long dim2436Kuo3Init[]   =   {   1   ,   3   ,   1   ,   9   ,   7   ,   41  ,   65  ,   19  ,   17  ,   573 ,   1007    ,   2097    ,   4251    ,   4067    ,   17217   ,0 };
            const unsigned long dim2437Kuo3Init[]   =   {   1   ,   3   ,   3   ,   13  ,   27  ,   53  ,   41  ,   223 ,   67  ,   747 ,   163 ,   1483    ,   1827    ,   11977   ,   17451   ,0 };
            const unsigned long dim2438Kuo3Init[]   =   {   1   ,   3   ,   5   ,   13  ,   17  ,   51  ,   109 ,   11  ,   105 ,   481 ,   239 ,   787 ,   5337    ,   14103   ,   10499   ,0 };
            const unsigned long dim2439Kuo3Init[]   =   {   1   ,   1   ,   3   ,   11  ,   31  ,   25  ,   13  ,   139 ,   37  ,   307 ,   331 ,   633 ,   7907    ,   4705    ,   26019   ,0 };
            const unsigned long dim2440Kuo3Init[]   =   {   1   ,   1   ,   7   ,   5   ,   25  ,   27  ,   11  ,   33  ,   463 ,   141 ,   1409    ,   2449    ,   2211    ,   10729   ,   28935   ,0 };
            const unsigned long dim2441Kuo3Init[]   =   {   1   ,   3   ,   3   ,   13  ,   1   ,   39  ,   53  ,   81  ,   117 ,   79  ,   563 ,   3735    ,   253 ,   10483   ,   28129   ,0 };
            const unsigned long dim2442Kuo3Init[]   =   {   1   ,   1   ,   3   ,   3   ,   7   ,   7   ,   103 ,   233 ,   341 ,   765 ,   11  ,   3735    ,   4011    ,   8075    ,   10239   ,0 };
            const unsigned long dim2443Kuo3Init[]   =   {   1   ,   3   ,   7   ,   7   ,   25  ,   23  ,   83  ,   139 ,   255 ,   441 ,   611 ,   1583    ,   733 ,   673 ,   21145   ,0 };
            const unsigned long dim2444Kuo3Init[]   =   {   1   ,   1   ,   3   ,   5   ,   29  ,   7   ,   49  ,   31  ,   417 ,   41  ,   133 ,   2119    ,   3871    ,   16029   ,   25649   ,0 };
            const unsigned long dim2445Kuo3Init[]   =   {   1   ,   3   ,   5   ,   5   ,   13  ,   17  ,   99  ,   213 ,   149 ,   1015    ,   987 ,   2497    ,   7419    ,   183 ,   27645   ,0 };
            const unsigned long dim2446Kuo3Init[]   =   {   1   ,   3   ,   3   ,   1   ,   25  ,   13  ,   107 ,   211 ,   113 ,   909 ,   1625    ,   2893    ,   3957    ,   10657   ,   16739   ,0 };
            const unsigned long dim2447Kuo3Init[]   =   {   1   ,   1   ,   5   ,   11  ,   5   ,   3   ,   97  ,   61  ,   189 ,   779 ,   1633    ,   71  ,   3751    ,   4589    ,   30795   ,0 };
            const unsigned long dim2448Kuo3Init[]   =   {   1   ,   3   ,   3   ,   15  ,   3   ,   15  ,   33  ,   175 ,   307 ,   937 ,   1713    ,   833 ,   2141    ,   1019    ,   26093   ,0 };
            const unsigned long dim2449Kuo3Init[]   =   {   1   ,   1   ,   7   ,   9   ,   15  ,   31  ,   9   ,   45  ,   197 ,   905 ,   1343    ,   27  ,   6835    ,   4353    ,   7229    ,0 };
            const unsigned long dim2450Kuo3Init[]   =   {   1   ,   1   ,   7   ,   15  ,   23  ,   15  ,   73  ,   217 ,   283 ,   853 ,   987 ,   389 ,   3875    ,   4353    ,   26667   ,0 };
            const unsigned long dim2451Kuo3Init[]   =   {   1   ,   3   ,   1   ,   7   ,   9   ,   1   ,   121 ,   39  ,   343 ,   225 ,   143 ,   343 ,   5573    ,   14997   ,   15723   ,0 };
            const unsigned long dim2452Kuo3Init[]   =   {   1   ,   1   ,   5   ,   5   ,   19  ,   35  ,   73  ,   177 ,   15  ,   697 ,   885 ,   3917    ,   6267    ,   14557   ,   30395   ,0 };
            const unsigned long dim2453Kuo3Init[]   =   {   1   ,   1   ,   3   ,   5   ,   3   ,   47  ,   59  ,   253 ,   107 ,   251 ,   1081    ,   3115    ,   1295    ,   9261    ,   22389   ,0 };
            const unsigned long dim2454Kuo3Init[]   =   {   1   ,   1   ,   5   ,   15  ,   29  ,   47  ,   113 ,   153 ,   205 ,   5   ,   507 ,   1313    ,   3329    ,   3043    ,   10135   ,0 };
            const unsigned long dim2455Kuo3Init[]   =   {   1   ,   1   ,   1   ,   3   ,   29  ,   39  ,   15  ,   147 ,   439 ,   975 ,   415 ,   3385    ,   5379    ,   8957    ,   3441    ,0 };
            const unsigned long dim2456Kuo3Init[]   =   {   1   ,   1   ,   1   ,   11  ,   23  ,   35  ,   97  ,   251 ,   399 ,   653 ,   1327    ,   3805    ,   6475    ,   1729    ,   29557   ,0 };
            const unsigned long dim2457Kuo3Init[]   =   {   1   ,   1   ,   3   ,   13  ,   1   ,   5   ,   63  ,   67  ,   209 ,   1005    ,   383 ,   2279    ,   509 ,   7039    ,   24973   ,0 };
            const unsigned long dim2458Kuo3Init[]   =   {   1   ,   1   ,   7   ,   7   ,   17  ,   59  ,   1   ,   5   ,   325 ,   703 ,   1937    ,   2699    ,   163 ,   10115   ,   8407    ,0 };
            const unsigned long dim2459Kuo3Init[]   =   {   1   ,   3   ,   1   ,   3   ,   25  ,   27  ,   73  ,   183 ,   41  ,   373 ,   909 ,   3957    ,   7135    ,   2401    ,   24963   ,0 };
            const unsigned long dim2460Kuo3Init[]   =   {   1   ,   1   ,   5   ,   7   ,   21  ,   3   ,   113 ,   19  ,   279 ,   729 ,   669 ,   1403    ,   881 ,   3581    ,   29297   ,0 };
            const unsigned long dim2461Kuo3Init[]   =   {   1   ,   3   ,   3   ,   1   ,   29  ,   9   ,   39  ,   97  ,   133 ,   369 ,   515 ,   1279    ,   2061    ,   5785    ,   29799   ,0 };
            const unsigned long dim2462Kuo3Init[]   =   {   1   ,   1   ,   1   ,   7   ,   15  ,   9   ,   33  ,   227 ,   13  ,   117 ,   1903    ,   2437    ,   451 ,   10029   ,   27469   ,0 };
            const unsigned long dim2463Kuo3Init[]   =   {   1   ,   1   ,   1   ,   5   ,   25  ,   3   ,   17  ,   11  ,   439 ,   701 ,   1319    ,   3357    ,   5487    ,   11165   ,   14077   ,0 };
            const unsigned long dim2464Kuo3Init[]   =   {   1   ,   3   ,   1   ,   9   ,   19  ,   61  ,   115 ,   83  ,   381 ,   261 ,   1025    ,   3215    ,   6889    ,   6123    ,   3761    ,0 };
            const unsigned long dim2465Kuo3Init[]   =   {   1   ,   3   ,   3   ,   11  ,   21  ,   17  ,   37  ,   151 ,   447 ,   683 ,   811 ,   2551    ,   6735    ,   16189   ,   13109   ,0 };
            const unsigned long dim2466Kuo3Init[]   =   {   1   ,   1   ,   1   ,   5   ,   5   ,   61  ,   11  ,   223 ,   221 ,   755 ,   65  ,   913 ,   6285    ,   13337   ,   27473   ,0 };
            const unsigned long dim2467Kuo3Init[]   =   {   1   ,   1   ,   1   ,   13  ,   15  ,   37  ,   93  ,   7   ,   249 ,   809 ,   1819    ,   3297    ,   4555    ,   9013    ,   16243   ,0 };
            const unsigned long dim2468Kuo3Init[]   =   {   1   ,   1   ,   1   ,   9   ,   3   ,   1   ,   13  ,   45  ,   407 ,   203 ,   963 ,   1055    ,   727 ,   2269    ,   18663   ,0 };
            const unsigned long dim2469Kuo3Init[]   =   {   1   ,   3   ,   7   ,   13  ,   3   ,   19  ,   95  ,   87  ,   341 ,   337 ,   557 ,   307 ,   7827    ,   15479   ,   24617   ,0 };
            const unsigned long dim2470Kuo3Init[]   =   {   1   ,   1   ,   1   ,   3   ,   31  ,   3   ,   125 ,   163 ,   439 ,   971 ,   815 ,   1785    ,   2637    ,   2973    ,   15405   ,0 };
            const unsigned long dim2471Kuo3Init[]   =   {   1   ,   3   ,   1   ,   15  ,   13  ,   61  ,   25  ,   151 ,   227 ,   173 ,   215 ,   3151    ,   4201    ,   6011    ,   13655   ,0 };
            const unsigned long dim2472Kuo3Init[]   =   {   1   ,   3   ,   1   ,   9   ,   27  ,   7   ,   13  ,   95  ,   511 ,   373 ,   2023    ,   1019    ,   3229    ,   2513    ,   20929   ,0 };
            const unsigned long dim2473Kuo3Init[]   =   {   1   ,   3   ,   7   ,   13  ,   11  ,   19  ,   93  ,   159 ,   461 ,   209 ,   549 ,   119 ,   7419    ,   6969    ,   29707   ,0 };
            const unsigned long dim2474Kuo3Init[]   =   {   1   ,   3   ,   1   ,   7   ,   17  ,   41  ,   53  ,   239 ,   121 ,   529 ,   27  ,   2129    ,   4675    ,   5159    ,   5767    ,0 };
            const unsigned long dim2475Kuo3Init[]   =   {   1   ,   1   ,   1   ,   15  ,   23  ,   45  ,   31  ,   189 ,   427 ,   583 ,   191 ,   1123    ,   3183    ,   5499    ,   25925   ,0 };
            const unsigned long dim2476Kuo3Init[]   =   {   1   ,   3   ,   7   ,   1   ,   3   ,   17  ,   39  ,   15  ,   257 ,   567 ,   459 ,   3731    ,   1315    ,   10975   ,   23725   ,0 };
            const unsigned long dim2477Kuo3Init[]   =   {   1   ,   3   ,   1   ,   11  ,   27  ,   43  ,   31  ,   101 ,   285 ,   249 ,   1467    ,   3937    ,   6413    ,   611 ,   10287   ,0 };
            const unsigned long dim2478Kuo3Init[]   =   {   1   ,   1   ,   5   ,   7   ,   19  ,   35  ,   5   ,   225 ,   211 ,   57  ,   1849    ,   3915    ,   5275    ,   13909   ,   18885   ,0 };
            const unsigned long dim2479Kuo3Init[]   =   {   1   ,   1   ,   5   ,   3   ,   17  ,   45  ,   45  ,   197 ,   311 ,   777 ,   2013    ,   927 ,   6209    ,   15495   ,   30027   ,0 };
            const unsigned long dim2480Kuo3Init[]   =   {   1   ,   1   ,   7   ,   11  ,   9   ,   13  ,   37  ,   81  ,   389 ,   459 ,   671 ,   2567    ,   7357    ,   5339    ,   13729   ,0 };
            const unsigned long dim2481Kuo3Init[]   =   {   1   ,   3   ,   1   ,   3   ,   29  ,   31  ,   63  ,   189 ,   271 ,   543 ,   229 ,   1385    ,   2031    ,   8993    ,   5665    ,0 };
            const unsigned long dim2482Kuo3Init[]   =   {   1   ,   3   ,   5   ,   11  ,   5   ,   25  ,   39  ,   225 ,   121 ,   207 ,   1161    ,   1739    ,   6627    ,   11129   ,   475 ,0 };
            const unsigned long dim2483Kuo3Init[]   =   {   1   ,   1   ,   7   ,   7   ,   19  ,   7   ,   29  ,   83  ,   25  ,   707 ,   197 ,   335 ,   2613    ,   10035   ,   20157   ,0 };
            const unsigned long dim2484Kuo3Init[]   =   {   1   ,   3   ,   3   ,   7   ,   7   ,   63  ,   99  ,   169 ,   221 ,   125 ,   563 ,   1351    ,   3867    ,   14903   ,   27469   ,0 };
            const unsigned long dim2485Kuo3Init[]   =   {   1   ,   3   ,   3   ,   7   ,   25  ,   5   ,   31  ,   151 ,   213 ,   555 ,   481 ,   3955    ,   4347    ,   8811    ,   30687   ,0 };
            const unsigned long dim2486Kuo3Init[]   =   {   1   ,   1   ,   5   ,   13  ,   23  ,   9   ,   59  ,   149 ,   351 ,   819 ,   311 ,   233 ,   4409    ,   3539    ,   6509    ,0 };
            const unsigned long dim2487Kuo3Init[]   =   {   1   ,   3   ,   3   ,   7   ,   1   ,   11  ,   3   ,   161 ,   473 ,   65  ,   1793    ,   367 ,   5115    ,   15743   ,   27175   ,0 };
            const unsigned long dim2488Kuo3Init[]   =   {   1   ,   3   ,   3   ,   9   ,   15  ,   61  ,   5   ,   125 ,   53  ,   523 ,   1365    ,   1909    ,   7689    ,   10799   ,   29199   ,0 };
            const unsigned long dim2489Kuo3Init[]   =   {   1   ,   1   ,   3   ,   7   ,   13  ,   43  ,   31  ,   179 ,   73  ,   611 ,   293 ,   2349    ,   4873    ,   11271   ,   29827   ,0 };
            const unsigned long dim2490Kuo3Init[]   =   {   1   ,   1   ,   5   ,   5   ,   27  ,   31  ,   107 ,   197 ,   409 ,   405 ,   789 ,   1529    ,   7793    ,   5679    ,   25127   ,0 };
            const unsigned long dim2491Kuo3Init[]   =   {   1   ,   3   ,   5   ,   1   ,   23  ,   35  ,   37  ,   119 ,   99  ,   891 ,   877 ,   1799    ,   5647    ,   13983   ,   22321   ,0 };
            const unsigned long dim2492Kuo3Init[]   =   {   1   ,   3   ,   3   ,   13  ,   9   ,   49  ,   29  ,   245 ,   447 ,   963 ,   211 ,   3233    ,   2883    ,   5467    ,   24057   ,0 };
            const unsigned long dim2493Kuo3Init[]   =   {   1   ,   1   ,   3   ,   5   ,   7   ,   29  ,   77  ,   7   ,   47  ,   141 ,   799 ,   2009    ,   7731    ,   4371    ,   24529   ,0 };
            const unsigned long dim2494Kuo3Init[]   =   {   1   ,   1   ,   7   ,   7   ,   19  ,   31  ,   65  ,   21  ,   417 ,   395 ,   653 ,   1751    ,   2989    ,   6721    ,   28155   ,0 };
            const unsigned long dim2495Kuo3Init[]   =   {   1   ,   1   ,   3   ,   9   ,   1   ,   5   ,   117 ,   119 ,   39  ,   949 ,   1081    ,   1919    ,   5689    ,   8367    ,   19043   ,0 };
            const unsigned long dim2496Kuo3Init[]   =   {   1   ,   3   ,   3   ,   13  ,   29  ,   45  ,   73  ,   49  ,   393 ,   545 ,   581 ,   689 ,   2597    ,   5757    ,   26953   ,0 };
            const unsigned long dim2497Kuo3Init[]   =   {   1   ,   1   ,   7   ,   13  ,   3   ,   13  ,   1   ,   97  ,   73  ,   517 ,   795 ,   2755    ,   5109    ,   15771   ,   19297   ,0 };
            const unsigned long dim2498Kuo3Init[]   =   {   1   ,   3   ,   1   ,   7   ,   11  ,   5   ,   117 ,   179 ,   507 ,   483 ,   143 ,   1441    ,   3247    ,   7557    ,   10939   ,0 };
            const unsigned long dim2499Kuo3Init[]   =   {   1   ,   1   ,   7   ,   1   ,   7   ,   9   ,   1   ,   1   ,   465 ,   1019    ,   37  ,   2011    ,   4065    ,   725 ,   10133   ,0 };
            const unsigned long dim2500Kuo3Init[]   =   {   1   ,   1   ,   7   ,   13  ,   15  ,   39  ,   69  ,   167 ,   399 ,   571 ,   1089    ,   1787    ,   8097    ,   3209    ,   25689   ,0 };
            const unsigned long dim2501Kuo3Init[]   =   {   1   ,   1   ,   7   ,   3   ,   7   ,   23  ,   17  ,   197 ,   43  ,   727 ,   543 ,   1307    ,   2701    ,   4121    ,   29165   ,0 };
            const unsigned long dim2502Kuo3Init[]   =   {   1   ,   1   ,   7   ,   3   ,   3   ,   29  ,   57  ,   89  ,   269 ,   667 ,   705 ,   2427    ,   47  ,   5535    ,   21861   ,0 };
            const unsigned long dim2503Kuo3Init[]   =   {   1   ,   1   ,   1   ,   13  ,   3   ,   43  ,   55  ,   55  ,   337 ,   567 ,   63  ,   3061    ,   5191    ,   16365   ,   24415   ,0 };
            const unsigned long dim2504Kuo3Init[]   =   {   1   ,   1   ,   5   ,   9   ,   3   ,   31  ,   123 ,   101 ,   149 ,   131 ,   1037    ,   3925    ,   861 ,   12811   ,   23787   ,0 };
            const unsigned long dim2505Kuo3Init[]   =   {   1   ,   3   ,   1   ,   5   ,   1   ,   5   ,   71  ,   7   ,   113 ,   797 ,   863 ,   1199    ,   1911    ,   2551    ,   12423   ,0 };
            const unsigned long dim2506Kuo3Init[]   =   {   1   ,   3   ,   1   ,   11  ,   19  ,   17  ,   25  ,   231 ,   7   ,   137 ,   127 ,   2297    ,   3345    ,   729 ,   909 ,0 };
            const unsigned long dim2507Kuo3Init[]   =   {   1   ,   1   ,   5   ,   9   ,   17  ,   13  ,   5   ,   227 ,   327 ,   335 ,   1463    ,   2685    ,   6301    ,   15547   ,   13151   ,0 };
            const unsigned long dim2508Kuo3Init[]   =   {   1   ,   3   ,   5   ,   9   ,   9   ,   25  ,   87  ,   53  ,   107 ,   219 ,   861 ,   1615    ,   371 ,   4983    ,   6445    ,0 };
            const unsigned long dim2509Kuo3Init[]   =   {   1   ,   3   ,   3   ,   13  ,   21  ,   59  ,   109 ,   53  ,   365 ,   293 ,   1373    ,   147 ,   4215    ,   1007    ,   20401   ,0 };
            const unsigned long dim2510Kuo3Init[]   =   {   1   ,   3   ,   3   ,   5   ,   5   ,   25  ,   113 ,   233 ,   317 ,   121 ,   427 ,   3561    ,   701 ,   10169   ,   23449   ,0 };
            const unsigned long dim2511Kuo3Init[]   =   {   1   ,   3   ,   5   ,   9   ,   9   ,   41  ,   123 ,   199 ,   305 ,   639 ,   1131    ,   3635    ,   8125    ,   4007    ,   6883    ,0 };
            const unsigned long dim2512Kuo3Init[]   =   {   1   ,   3   ,   1   ,   13  ,   23  ,   5   ,   105 ,   25  ,   281 ,   301 ,   1011    ,   3647    ,   263 ,   3389    ,   5527    ,0 };
            const unsigned long dim2513Kuo3Init[]   =   {   1   ,   1   ,   5   ,   15  ,   17  ,   59  ,   33  ,   175 ,   115 ,   625 ,   975 ,   3189    ,   6973    ,   8291    ,   14201   ,0 };
            const unsigned long dim2514Kuo3Init[]   =   {   1   ,   3   ,   3   ,   5   ,   1   ,   41  ,   39  ,   231 ,   57  ,   95  ,   623 ,   2199    ,   3821    ,   1627    ,   4035    ,0 };
            const unsigned long dim2515Kuo3Init[]   =   {   1   ,   3   ,   5   ,   11  ,   25  ,   5   ,   1   ,   21  ,   457 ,   875 ,   355 ,   389 ,   1623    ,   9575    ,   1283    ,0 };
            const unsigned long dim2516Kuo3Init[]   =   {   1   ,   3   ,   1   ,   1   ,   21  ,   29  ,   83  ,   45  ,   315 ,   735 ,   1253    ,   3743    ,   7163    ,   5163    ,   10833   ,0 };
            const unsigned long dim2517Kuo3Init[]   =   {   1   ,   3   ,   5   ,   13  ,   1   ,   5   ,   47  ,   21  ,   231 ,   1019    ,   1851    ,   21  ,   1699    ,   851 ,   11857   ,0 };
            const unsigned long dim2518Kuo3Init[]   =   {   1   ,   3   ,   5   ,   7   ,   19  ,   9   ,   61  ,   241 ,   175 ,   191 ,   415 ,   1197    ,   5643    ,   9765    ,   17375   ,0 };
            const unsigned long dim2519Kuo3Init[]   =   {   1   ,   3   ,   5   ,   1   ,   3   ,   5   ,   35  ,   231 ,   471 ,   175 ,   1715    ,   487 ,   2299    ,   12355   ,   5113    ,0 };
            const unsigned long dim2520Kuo3Init[]   =   {   1   ,   1   ,   3   ,   9   ,   15  ,   45  ,   33  ,   93  ,   289 ,   371 ,   275 ,   1703    ,   6925    ,   5511    ,   3203    ,0 };
            const unsigned long dim2521Kuo3Init[]   =   {   1   ,   3   ,   5   ,   7   ,   21  ,   55  ,   105 ,   199 ,   435 ,   955 ,   731 ,   719 ,   1571    ,   7341    ,   5255    ,0 };
            const unsigned long dim2522Kuo3Init[]   =   {   1   ,   3   ,   7   ,   7   ,   1   ,   3   ,   127 ,   85  ,   11  ,   165 ,   1423    ,   3889    ,   7509    ,   14687   ,   23383   ,0 };
            const unsigned long dim2523Kuo3Init[]   =   {   1   ,   1   ,   7   ,   11  ,   31  ,   61  ,   9   ,   141 ,   357 ,   801 ,   1785    ,   3813    ,   3437    ,   10853   ,   17191   ,0 };
            const unsigned long dim2524Kuo3Init[]   =   {   1   ,   1   ,   7   ,   5   ,   23  ,   59  ,   17  ,   205 ,   449 ,   107 ,   1123    ,   1089    ,   5499    ,   12251   ,   653 ,0 };
            const unsigned long dim2525Kuo3Init[]   =   {   1   ,   3   ,   1   ,   5   ,   3   ,   25  ,   71  ,   229 ,   263 ,   423 ,   1357    ,   3709    ,   1719    ,   2799    ,   10943   ,0 };
            const unsigned long dim2526Kuo3Init[]   =   {   1   ,   1   ,   5   ,   13  ,   1   ,   23  ,   97  ,   209 ,   401 ,   809 ,   783 ,   2455    ,   1137    ,   13941   ,   12751   ,0 };
            const unsigned long dim2527Kuo3Init[]   =   {   1   ,   3   ,   3   ,   11  ,   21  ,   47  ,   25  ,   29  ,   471 ,   677 ,   1129    ,   2689    ,   929 ,   5635    ,   26509   ,0 };
            const unsigned long dim2528Kuo3Init[]   =   {   1   ,   3   ,   1   ,   5   ,   7   ,   27  ,   5   ,   195 ,   435 ,   255 ,   395 ,   2775    ,   297 ,   6909    ,   2289    ,0 };
            const unsigned long dim2529Kuo3Init[]   =   {   1   ,   1   ,   1   ,   9   ,   7   ,   53  ,   109 ,   85  ,   141 ,   243 ,   1433    ,   1531    ,   5461    ,   1111    ,   7889    ,0 };
            const unsigned long dim2530Kuo3Init[]   =   {   1   ,   3   ,   7   ,   11  ,   13  ,   47  ,   91  ,   51  ,   489 ,   433 ,   1667    ,   2955    ,   5135    ,   11949   ,   14123   ,0 };
            const unsigned long dim2531Kuo3Init[]   =   {   1   ,   1   ,   7   ,   15  ,   5   ,   57  ,   57  ,   147 ,   97  ,   709 ,   1757    ,   1493    ,   2715    ,   12583   ,   18991   ,0 };
            const unsigned long dim2532Kuo3Init[]   =   {   1   ,   3   ,   1   ,   15  ,   7   ,   17  ,   79  ,   75  ,   397 ,   555 ,   787 ,   1405    ,   4749    ,   6723    ,   3749    ,0 };
            const unsigned long dim2533Kuo3Init[]   =   {   1   ,   1   ,   3   ,   11  ,   21  ,   59  ,   5   ,   211 ,   215 ,   393 ,   1445    ,   4057    ,   5115    ,   821 ,   7783    ,0 };
            const unsigned long dim2534Kuo3Init[]   =   {   1   ,   1   ,   3   ,   5   ,   29  ,   1   ,   97  ,   201 ,   21  ,   287 ,   1735    ,   2397    ,   3401    ,   2017    ,   32229   ,0 };
            const unsigned long dim2535Kuo3Init[]   =   {   1   ,   1   ,   3   ,   13  ,   17  ,   43  ,   71  ,   25  ,   243 ,   423 ,   733 ,   1917    ,   1235    ,   3527    ,   22695   ,0 };
            const unsigned long dim2536Kuo3Init[]   =   {   1   ,   1   ,   3   ,   15  ,   27  ,   7   ,   25  ,   239 ,   493 ,   945 ,   397 ,   3837    ,   5465    ,   12273   ,   26025   ,0 };
            const unsigned long dim2537Kuo3Init[]   =   {   1   ,   3   ,   1   ,   7   ,   17  ,   17  ,   61  ,   7   ,   487 ,   439 ,   1433    ,   1761    ,   6507    ,   487 ,   32633   ,0 };
            const unsigned long dim2538Kuo3Init[]   =   {   1   ,   3   ,   3   ,   9   ,   21  ,   29  ,   83  ,   27  ,   321 ,   73  ,   1247    ,   2593    ,   6935    ,   7009    ,   31617   ,0 };
            const unsigned long dim2539Kuo3Init[]   =   {   1   ,   3   ,   3   ,   1   ,   25  ,   37  ,   99  ,   177 ,   357 ,   527 ,   1259    ,   3625    ,   4277    ,   15243   ,   28615   ,0 };
            const unsigned long dim2540Kuo3Init[]   =   {   1   ,   1   ,   3   ,   13  ,   29  ,   31  ,   59  ,   177 ,   73  ,   1007    ,   1809    ,   2143    ,   7375    ,   4321    ,   3771    ,0 };
            const unsigned long dim2541Kuo3Init[]   =   {   1   ,   3   ,   3   ,   3   ,   21  ,   7   ,   123 ,   253 ,   87  ,   995 ,   103 ,   561 ,   7639    ,   10087   ,   4149    ,0 };
            const unsigned long dim2542Kuo3Init[]   =   {   1   ,   1   ,   5   ,   13  ,   11  ,   31  ,   53  ,   249 ,   435 ,   361 ,   367 ,   783 ,   7525    ,   5361    ,   21535   ,0 };
            const unsigned long dim2543Kuo3Init[]   =   {   1   ,   3   ,   7   ,   5   ,   15  ,   25  ,   31  ,   9   ,   29  ,   907 ,   77  ,   2585    ,   2929    ,   12153   ,   19401   ,0 };
            const unsigned long dim2544Kuo3Init[]   =   {   1   ,   3   ,   5   ,   5   ,   15  ,   1   ,   101 ,   99  ,   209 ,   583 ,   679 ,   1005    ,   2271    ,   269 ,   29527   ,0 };
            const unsigned long dim2545Kuo3Init[]   =   {   1   ,   1   ,   3   ,   1   ,   13  ,   35  ,   85  ,   83  ,   77  ,   999 ,   1485    ,   2817    ,   7001    ,   13659   ,   25061   ,0 };
            const unsigned long dim2546Kuo3Init[]   =   {   1   ,   3   ,   7   ,   13  ,   19  ,   15  ,   19  ,   65  ,   195 ,   693 ,   1399    ,   2875    ,   255 ,   6617    ,   7907    ,0 };
            const unsigned long dim2547Kuo3Init[]   =   {   1   ,   1   ,   5   ,   5   ,   13  ,   25  ,   67  ,   195 ,   131 ,   481 ,   1831    ,   873 ,   6281    ,   4935    ,   2311    ,0 };
            const unsigned long dim2548Kuo3Init[]   =   {   1   ,   1   ,   1   ,   15  ,   27  ,   47  ,   85  ,   209 ,   323 ,   809 ,   605 ,   2087    ,   1917    ,   8061    ,   13659   ,0 };
            const unsigned long dim2549Kuo3Init[]   =   {   1   ,   3   ,   1   ,   13  ,   5   ,   63  ,   69  ,   197 ,   241 ,   349 ,   813 ,   3943    ,   6267    ,   16197   ,   1417    ,0 };
            const unsigned long dim2550Kuo3Init[]   =   {   1   ,   1   ,   1   ,   5   ,   13  ,   21  ,   85  ,   55  ,   21  ,   791 ,   795 ,   4087    ,   2697    ,   6545    ,   30253   ,0 };
            const unsigned long dim2551Kuo3Init[]   =   {   1   ,   3   ,   5   ,   15  ,   31  ,   13  ,   117 ,   237 ,   99  ,   803 ,   1367    ,   251 ,   5143    ,   12153   ,   27565   ,0 };
            const unsigned long dim2552Kuo3Init[]   =   {   1   ,   1   ,   7   ,   7   ,   11  ,   7   ,   17  ,   47  ,   305 ,   9   ,   941 ,   3529    ,   6131    ,   11027   ,   20319   ,0 };
            const unsigned long dim2553Kuo3Init[]   =   {   1   ,   3   ,   1   ,   5   ,   19  ,   51  ,   1   ,   113 ,   277 ,   573 ,   1717    ,   1757    ,   4541    ,   943 ,   349 ,0 };
            const unsigned long dim2554Kuo3Init[]   =   {   1   ,   3   ,   1   ,   1   ,   21  ,   23  ,   11  ,   143 ,   413 ,   171 ,   355 ,   2801    ,   2431    ,   11597   ,   16833   ,0 };
            const unsigned long dim2555Kuo3Init[]   =   {   1   ,   3   ,   1   ,   1   ,   13  ,   31  ,   111 ,   109 ,   73  ,   125 ,   1011    ,   2249    ,   4039    ,   8757    ,   2765    ,0 };
            const unsigned long dim2556Kuo3Init[]   =   {   1   ,   1   ,   1   ,   7   ,   15  ,   53  ,   33  ,   133 ,   511 ,   595 ,   139 ,   815 ,   6093    ,   10215   ,   9507    ,0 };
            const unsigned long dim2557Kuo3Init[]   =   {   1   ,   3   ,   7   ,   7   ,   29  ,   1   ,   93  ,   13  ,   169 ,   435 ,   527 ,   3861    ,   3867    ,   9337    ,   26855   ,0 };
            const unsigned long dim2558Kuo3Init[]   =   {   1   ,   1   ,   5   ,   3   ,   13  ,   29  ,   91  ,   73  ,   257 ,   645 ,   679 ,   3245    ,   6467    ,   5917    ,   17891   ,0 };
            const unsigned long dim2559Kuo3Init[]   =   {   1   ,   1   ,   7   ,   15  ,   27  ,   1   ,   45  ,   17  ,   429 ,   571 ,   1127    ,   3843    ,   3801    ,   4773    ,   16577   ,0 };
            const unsigned long dim2560Kuo3Init[]   =   {   1   ,   1   ,   7   ,   11  ,   29  ,   29  ,   65  ,   111 ,   415 ,   941 ,   1411    ,   3003    ,   3945    ,   15281   ,   16851   ,0 };
            const unsigned long dim2561Kuo3Init[]   =   {   1   ,   3   ,   1   ,   3   ,   3   ,   29  ,   39  ,   121 ,   165 ,   285 ,   1895    ,   283 ,   4393    ,   13281   ,   3349    ,0 };
            const unsigned long dim2562Kuo3Init[]   =   {   1   ,   3   ,   1   ,   9   ,   27  ,   1   ,   35  ,   119 ,   91  ,   445 ,   1045    ,   1427    ,   6271    ,   955 ,   509 ,0 };
            const unsigned long dim2563Kuo3Init[]   =   {   1   ,   1   ,   7   ,   1   ,   11  ,   15  ,   3   ,   181 ,   17  ,   3   ,   1865    ,   2465    ,   357 ,   12149   ,   7359    ,0 };
            const unsigned long dim2564Kuo3Init[]   =   {   1   ,   3   ,   5   ,   5   ,   11  ,   47  ,   5   ,   65  ,   105 ,   53  ,   1341    ,   3729    ,   4883    ,   5179    ,   29353   ,0 };
            const unsigned long dim2565Kuo3Init[]   =   {   1   ,   1   ,   7   ,   5   ,   15  ,   53  ,   23  ,   181 ,   125 ,   441 ,   511 ,   3933    ,   2327    ,   1305    ,   14629   ,0 };
            const unsigned long dim2566Kuo3Init[]   =   {   1   ,   1   ,   3   ,   5   ,   3   ,   11  ,   91  ,   187 ,   177 ,   649 ,   1849    ,   2545    ,   3349    ,   6563    ,   8139    ,0 };
            const unsigned long dim2567Kuo3Init[]   =   {   1   ,   3   ,   1   ,   11  ,   31  ,   35  ,   71  ,   51  ,   235 ,   295 ,   521 ,   3869    ,   3083    ,   1327    ,   3277    ,0 };
            const unsigned long dim2568Kuo3Init[]   =   {   1   ,   1   ,   3   ,   15  ,   11  ,   11  ,   125 ,   163 ,   71  ,   721 ,   449 ,   3443    ,   1347    ,   6221    ,   19489   ,0 };
            const unsigned long dim2569Kuo3Init[]   =   {   1   ,   1   ,   7   ,   3   ,   25  ,   13  ,   13  ,   149 ,   115 ,   563 ,   311 ,   1823    ,   2687    ,   5997    ,   23063   ,0 };
            const unsigned long dim2570Kuo3Init[]   =   {   1   ,   1   ,   1   ,   13  ,   7   ,   47  ,   115 ,   19  ,   419 ,   395 ,   591 ,   2413    ,   7143    ,   11235   ,   25795   ,0 };
            const unsigned long dim2571Kuo3Init[]   =   {   1   ,   1   ,   1   ,   7   ,   17  ,   37  ,   63  ,   57  ,   175 ,   301 ,   493 ,   1849    ,   1903    ,   13149   ,   21429   ,0 };
            const unsigned long dim2572Kuo3Init[]   =   {   1   ,   1   ,   3   ,   3   ,   1   ,   53  ,   15  ,   91  ,   315 ,   687 ,   473 ,   1639    ,   7769    ,   11621   ,   24245   ,0 };
            const unsigned long dim2573Kuo3Init[]   =   {   1   ,   1   ,   5   ,   11  ,   13  ,   35  ,   13  ,   235 ,   365 ,   677 ,   611 ,   3175    ,   6349    ,   445 ,   13739   ,0 };
            const unsigned long dim2574Kuo3Init[]   =   {   1   ,   1   ,   1   ,   15  ,   17  ,   49  ,   49  ,   15  ,   205 ,   755 ,   665 ,   3865    ,   6499    ,   6521    ,   31925   ,0 };
            const unsigned long dim2575Kuo3Init[]   =   {   1   ,   1   ,   3   ,   5   ,   9   ,   49  ,   57  ,   119 ,   489 ,   633 ,   1407    ,   957 ,   4005    ,   14503   ,   2783    ,0 };
            const unsigned long dim2576Kuo3Init[]   =   {   1   ,   1   ,   3   ,   13  ,   17  ,   41  ,   73  ,   137 ,   195 ,   457 ,   1361    ,   3209    ,   3701    ,   9675    ,   26499   ,0 };
            const unsigned long dim2577Kuo3Init[]   =   {   1   ,   3   ,   7   ,   3   ,   31  ,   49  ,   71  ,   83  ,   141 ,   1009    ,   1503    ,   3393    ,   4653    ,   8051    ,   25861   ,0 };
            const unsigned long dim2578Kuo3Init[]   =   {   1   ,   3   ,   7   ,   9   ,   31  ,   17  ,   83  ,   211 ,   3   ,   81  ,   1579    ,   387 ,   4471    ,   5407    ,   28849   ,0 };
            const unsigned long dim2579Kuo3Init[]   =   {   1   ,   1   ,   5   ,   9   ,   3   ,   51  ,   45  ,   157 ,   197 ,   195 ,   1121    ,   2451    ,   3885    ,   9317    ,   29157   ,0 };
            const unsigned long dim2580Kuo3Init[]   =   {   1   ,   3   ,   7   ,   5   ,   23  ,   19  ,   49  ,   89  ,   107 ,   369 ,   129 ,   2381    ,   383 ,   8279    ,   29277   ,0 };
            const unsigned long dim2581Kuo3Init[]   =   {   1   ,   3   ,   3   ,   13  ,   21  ,   7   ,   123 ,   221 ,   85  ,   819 ,   1857    ,   1211    ,   7495    ,   2921    ,   9529    ,0 };
            const unsigned long dim2582Kuo3Init[]   =   {   1   ,   3   ,   3   ,   13  ,   19  ,   13  ,   63  ,   47  ,   465 ,   739 ,   1961    ,   2779    ,   7989    ,   8419    ,   19249   ,0 };
            const unsigned long dim2583Kuo3Init[]   =   {   1   ,   3   ,   5   ,   15  ,   25  ,   49  ,   65  ,   39  ,   333 ,   193 ,   443 ,   1221    ,   245 ,   523 ,   31885   ,0 };
            const unsigned long dim2584Kuo3Init[]   =   {   1   ,   3   ,   5   ,   7   ,   31  ,   3   ,   69  ,   173 ,   19  ,   593 ,   61  ,   1545    ,   415 ,   14489   ,   9723    ,0 };
            const unsigned long dim2585Kuo3Init[]   =   {   1   ,   3   ,   3   ,   1   ,   5   ,   9   ,   121 ,   87  ,   431 ,   893 ,   1031    ,   3589    ,   4263    ,   11355   ,   22725   ,0 };
            const unsigned long dim2586Kuo3Init[]   =   {   1   ,   1   ,   5   ,   3   ,   21  ,   9   ,   47  ,   151 ,   165 ,   941 ,   1369    ,   3991    ,   989 ,   1515    ,   12725   ,0 };
            const unsigned long dim2587Kuo3Init[]   =   {   1   ,   3   ,   7   ,   1   ,   7   ,   47  ,   57  ,   153 ,   239 ,   397 ,   1723    ,   1681    ,   521 ,   8173    ,   14567   ,0 };
            const unsigned long dim2588Kuo3Init[]   =   {   1   ,   3   ,   3   ,   5   ,   23  ,   9   ,   77  ,   201 ,   61  ,   969 ,   345 ,   2817    ,   4391    ,   2193    ,   18739   ,0 };
            const unsigned long dim2589Kuo3Init[]   =   {   1   ,   1   ,   5   ,   13  ,   15  ,   45  ,   83  ,   165 ,   421 ,   967 ,   449 ,   3419    ,   2475    ,   7283    ,   18199   ,0 };
            const unsigned long dim2590Kuo3Init[]   =   {   1   ,   3   ,   7   ,   11  ,   25  ,   37  ,   3   ,   175 ,   327 ,   35  ,   1179    ,   927 ,   5263    ,   1723    ,   32015   ,0 };
            const unsigned long dim2591Kuo3Init[]   =   {   1   ,   1   ,   5   ,   15  ,   29  ,   17  ,   63  ,   45  ,   495 ,   107 ,   459 ,   2223    ,   989 ,   11823   ,   6547    ,0 };
            const unsigned long dim2592Kuo3Init[]   =   {   1   ,   3   ,   1   ,   13  ,   17  ,   15  ,   127 ,   223 ,   369 ,   409 ,   303 ,   2229    ,   7349    ,   10165   ,   28327   ,0 };
            const unsigned long dim2593Kuo3Init[]   =   {   1   ,   3   ,   5   ,   15  ,   15  ,   17  ,   119 ,   185 ,   55  ,   299 ,   113 ,   3585    ,   2289    ,   7793    ,   26217   ,0 };
            const unsigned long dim2594Kuo3Init[]   =   {   1   ,   3   ,   7   ,   13  ,   21  ,   33  ,   3   ,   5   ,   169 ,   615 ,   1607    ,   1069    ,   1827    ,   6165    ,   16245   ,0 };
            const unsigned long dim2595Kuo3Init[]   =   {   1   ,   1   ,   5   ,   13  ,   21  ,   59  ,   119 ,   253 ,   181 ,   1001    ,   2001    ,   887 ,   4129    ,   3953    ,   29979   ,0 };
            const unsigned long dim2596Kuo3Init[]   =   {   1   ,   3   ,   3   ,   9   ,   29  ,   17  ,   33  ,   3   ,   375 ,   973 ,   989 ,   1725    ,   731 ,   12845   ,   21133   ,0 };
            const unsigned long dim2597Kuo3Init[]   =   {   1   ,   1   ,   3   ,   13  ,   21  ,   25  ,   65  ,   41  ,   379 ,   693 ,   1507    ,   1601    ,   2087    ,   8181    ,   12723   ,0 };
            const unsigned long dim2598Kuo3Init[]   =   {   1   ,   1   ,   7   ,   1   ,   21  ,   37  ,   117 ,   187 ,   483 ,   779 ,   1279    ,   685 ,   5479    ,   12039   ,   12559   ,0 };
            const unsigned long dim2599Kuo3Init[]   =   {   1   ,   1   ,   3   ,   5   ,   7   ,   39  ,   49  ,   79  ,   361 ,   191 ,   1749    ,   507 ,   19  ,   10671   ,   4745    ,0 };
            const unsigned long dim2600Kuo3Init[]   =   {   1   ,   1   ,   1   ,   15  ,   15  ,   45  ,   37  ,   49  ,   85  ,   311 ,   577 ,   2417    ,   4061    ,   1409    ,   19187   ,0 };
            const unsigned long dim2601Kuo3Init[]   =   {   1   ,   3   ,   5   ,   11  ,   5   ,   1   ,   41  ,   73  ,   313 ,   741 ,   1081    ,   3325    ,   977 ,   1693    ,   15249   ,0 };
            const unsigned long dim2602Kuo3Init[]   =   {   1   ,   1   ,   1   ,   9   ,   19  ,   49  ,   29  ,   169 ,   153 ,   665 ,   413 ,   907 ,   6129    ,   4961    ,   21295   ,0 };
            const unsigned long dim2603Kuo3Init[]   =   {   1   ,   3   ,   3   ,   3   ,   23  ,   17  ,   91  ,   27  ,   375 ,   231 ,   513 ,   1459    ,   7245    ,   8117    ,   19583   ,0 };
            const unsigned long dim2604Kuo3Init[]   =   {   1   ,   1   ,   7   ,   13  ,   19  ,   57  ,   111 ,   195 ,   491 ,   771 ,   1841    ,   1339    ,   5045    ,   12901   ,   17307   ,0 };
            const unsigned long dim2605Kuo3Init[]   =   {   1   ,   1   ,   1   ,   7   ,   1   ,   61  ,   61  ,   11  ,   151 ,   403 ,   173 ,   3629    ,   2157    ,   6125    ,   29233   ,0 };
            const unsigned long dim2606Kuo3Init[]   =   {   1   ,   3   ,   1   ,   13  ,   29  ,   63  ,   55  ,   9   ,   465 ,   439 ,   9   ,   1527    ,   4635    ,   4533    ,   13633   ,0 };
            const unsigned long dim2607Kuo3Init[]   =   {   1   ,   3   ,   5   ,   1   ,   15  ,   57  ,   113 ,   137 ,   483 ,   753 ,   1971    ,   1141    ,   2765    ,   7865    ,   32733   ,0 };
            const unsigned long dim2608Kuo3Init[]   =   {   1   ,   3   ,   5   ,   15  ,   9   ,   23  ,   125 ,   81  ,   471 ,   457 ,   1475    ,   3419    ,   6089    ,   4899    ,   19805   ,0 };
            const unsigned long dim2609Kuo3Init[]   =   {   1   ,   1   ,   5   ,   3   ,   25  ,   29  ,   1   ,   115 ,   273 ,   281 ,   899 ,   3129    ,   4877    ,   15459   ,   24615   ,0 };
            const unsigned long dim2610Kuo3Init[]   =   {   1   ,   1   ,   7   ,   11  ,   19  ,   41  ,   59  ,   165 ,   115 ,   369 ,   1173    ,   2545    ,   4285    ,   10899   ,   3823    ,0 };
            const unsigned long dim2611Kuo3Init[]   =   {   1   ,   3   ,   5   ,   13  ,   17  ,   47  ,   107 ,   191 ,   53  ,   1003    ,   747 ,   583 ,   6457    ,   12141   ,   2541    ,0 };
            const unsigned long dim2612Kuo3Init[]   =   {   1   ,   3   ,   7   ,   1   ,   9   ,   51  ,   33  ,   201 ,   399 ,   209 ,   271 ,   679 ,   2401    ,   10521   ,   9949    ,0 };
            const unsigned long dim2613Kuo3Init[]   =   {   1   ,   1   ,   5   ,   9   ,   5   ,   53  ,   117 ,   157 ,   137 ,   757 ,   257 ,   1407    ,   4831    ,   11205   ,   25075   ,0 };
            const unsigned long dim2614Kuo3Init[]   =   {   1   ,   1   ,   7   ,   13  ,   29  ,   63  ,   25  ,   11  ,   255 ,   251 ,   1393    ,   2089    ,   1701    ,   11773   ,   24011   ,0 };
            const unsigned long dim2615Kuo3Init[]   =   {   1   ,   3   ,   3   ,   11  ,   17  ,   5   ,   69  ,   119 ,   287 ,   607 ,   779 ,   2369    ,   55  ,   12381   ,   26051   ,0 };
            const unsigned long dim2616Kuo3Init[]   =   {   1   ,   3   ,   3   ,   13  ,   25  ,   1   ,   97  ,   109 ,   79  ,   453 ,   795 ,   2821    ,   4113    ,   3085    ,   30667   ,0 };
            const unsigned long dim2617Kuo3Init[]   =   {   1   ,   1   ,   1   ,   9   ,   25  ,   17  ,   87  ,   163 ,   483 ,   803 ,   315 ,   2755    ,   7147    ,   12537   ,   7707    ,0 };
            const unsigned long dim2618Kuo3Init[]   =   {   1   ,   3   ,   1   ,   5   ,   23  ,   29  ,   93  ,   129 ,   47  ,   181 ,   1577    ,   2645    ,   4103    ,   10753   ,   27119   ,0 };
            const unsigned long dim2619Kuo3Init[]   =   {   1   ,   1   ,   1   ,   1   ,   17  ,   61  ,   37  ,   175 ,   451 ,   5   ,   879 ,   1625    ,   7981    ,   11249   ,   12573   ,0 };
            const unsigned long dim2620Kuo3Init[]   =   {   1   ,   1   ,   3   ,   5   ,   9   ,   17  ,   85  ,   131 ,   111 ,   621 ,   903 ,   3143    ,   5951    ,   6567    ,   32227   ,0 };
            const unsigned long dim2621Kuo3Init[]   =   {   1   ,   3   ,   5   ,   3   ,   1   ,   49  ,   53  ,   255 ,   435 ,   845 ,   225 ,   2535    ,   7925    ,   8043    ,   17519   ,0 };
            const unsigned long dim2622Kuo3Init[]   =   {   1   ,   3   ,   3   ,   5   ,   23  ,   7   ,   63  ,   117 ,   79  ,   637 ,   1265    ,   2747    ,   6531    ,   7327    ,   17329   ,0 };
            const unsigned long dim2623Kuo3Init[]   =   {   1   ,   3   ,   7   ,   3   ,   27  ,   35  ,   69  ,   253 ,   489 ,   495 ,   561 ,   3615    ,   917 ,   6595    ,   10673   ,0 };
            const unsigned long dim2624Kuo3Init[]   =   {   1   ,   1   ,   1   ,   5   ,   25  ,   7   ,   99  ,   99  ,   499 ,   251 ,   1651    ,   3603    ,   689 ,   5837    ,   347 ,0 };
            const unsigned long dim2625Kuo3Init[]   =   {   1   ,   1   ,   3   ,   5   ,   1   ,   27  ,   55  ,   91  ,   137 ,   291 ,   517 ,   2149    ,   1097    ,   2187    ,   16191   ,0 };
            const unsigned long dim2626Kuo3Init[]   =   {   1   ,   1   ,   1   ,   11  ,   29  ,   3   ,   103 ,   81  ,   335 ,   85  ,   549 ,   1667    ,   5319    ,   5929    ,   15885   ,0 };
            const unsigned long dim2627Kuo3Init[]   =   {   1   ,   1   ,   3   ,   9   ,   7   ,   41  ,   103 ,   141 ,   103 ,   655 ,   179 ,   1513    ,   7105    ,   1919    ,   2245    ,0 };
            const unsigned long dim2628Kuo3Init[]   =   {   1   ,   3   ,   5   ,   11  ,   11  ,   13  ,   127 ,   65  ,   493 ,   233 ,   859 ,   3179    ,   5449    ,   13255   ,   14495   ,0 };
            const unsigned long dim2629Kuo3Init[]   =   {   1   ,   3   ,   1   ,   5   ,   27  ,   49  ,   119 ,   29  ,   55  ,   717 ,   1293    ,   1193    ,   7883    ,   4475    ,   13745   ,0 };
            const unsigned long dim2630Kuo3Init[]   =   {   1   ,   3   ,   7   ,   3   ,   21  ,   9   ,   1   ,   43  ,   425 ,   385 ,   447 ,   3135    ,   7671    ,   12317   ,   5649    ,0 };
            const unsigned long dim2631Kuo3Init[]   =   {   1   ,   1   ,   7   ,   5   ,   21  ,   51  ,   103 ,   199 ,   119 ,   875 ,   1457    ,   461 ,   6521    ,   7849    ,   7725    ,0 };
            const unsigned long dim2632Kuo3Init[]   =   {   1   ,   1   ,   1   ,   15  ,   9   ,   25  ,   89  ,   125 ,   395 ,   269 ,   1533    ,   1545    ,   2613    ,   11175   ,   17167   ,0 };
            const unsigned long dim2633Kuo3Init[]   =   {   1   ,   3   ,   5   ,   9   ,   17  ,   41  ,   113 ,   37  ,   253 ,   407 ,   795 ,   2117    ,   3041    ,   9269    ,   30097   ,0 };
            const unsigned long dim2634Kuo3Init[]   =   {   1   ,   1   ,   3   ,   11  ,   31  ,   13  ,   23  ,   29  ,   385 ,   603 ,   1721    ,   3817    ,   3453    ,   2021    ,   8829    ,0 };
            const unsigned long dim2635Kuo3Init[]   =   {   1   ,   1   ,   7   ,   9   ,   21  ,   3   ,   107 ,   121 ,   267 ,   385 ,   261 ,   3967    ,   5607    ,   13991   ,   8639    ,0 };
            const unsigned long dim2636Kuo3Init[]   =   {   1   ,   3   ,   5   ,   15  ,   17  ,   19  ,   99  ,   195 ,   415 ,   19  ,   865 ,   4049    ,   3819    ,   7581    ,   10779   ,0 };
            const unsigned long dim2637Kuo3Init[]   =   {   1   ,   1   ,   5   ,   15  ,   25  ,   29  ,   125 ,   105 ,   337 ,   153 ,   2023    ,   1475    ,   2741    ,   7779    ,   16929   ,0 };
            const unsigned long dim2638Kuo3Init[]   =   {   1   ,   3   ,   3   ,   5   ,   11  ,   1   ,   111 ,   157 ,   441 ,   751 ,   865 ,   3499    ,   6425    ,   9567    ,   6659    ,0 };
            const unsigned long dim2639Kuo3Init[]   =   {   1   ,   1   ,   5   ,   3   ,   23  ,   41  ,   57  ,   91  ,   413 ,   81  ,   1977    ,   2227    ,   2289    ,   1291    ,   19339   ,0 };
            const unsigned long dim2640Kuo3Init[]   =   {   1   ,   1   ,   3   ,   15  ,   15  ,   49  ,   47  ,   115 ,   153 ,   593 ,   1243    ,   2029    ,   1177    ,   8121    ,   14893   ,0 };
            const unsigned long dim2641Kuo3Init[]   =   {   1   ,   1   ,   5   ,   15  ,   29  ,   11  ,   55  ,   23  ,   173 ,   145 ,   143 ,   2275    ,   4537    ,   13221   ,   32575   ,0 };
            const unsigned long dim2642Kuo3Init[]   =   {   1   ,   1   ,   1   ,   9   ,   29  ,   53  ,   103 ,   37  ,   327 ,   543 ,   1327    ,   899 ,   6343    ,   6707    ,   1247    ,0 };
            const unsigned long dim2643Kuo3Init[]   =   {   1   ,   3   ,   1   ,   13  ,   21  ,   49  ,   113 ,   205 ,   349 ,   451 ,   1277    ,   989 ,   3387    ,   9275    ,   6629    ,0 };
            const unsigned long dim2644Kuo3Init[]   =   {   1   ,   3   ,   5   ,   5   ,   31  ,   59  ,   79  ,   65  ,   285 ,   841 ,   287 ,   1453    ,   1507    ,   14813   ,   14705   ,0 };
            const unsigned long dim2645Kuo3Init[]   =   {   1   ,   1   ,   3   ,   5   ,   9   ,   11  ,   79  ,   109 ,   313 ,   339 ,   1939    ,   2399    ,   5127    ,   8747    ,   11989   ,0 };
            const unsigned long dim2646Kuo3Init[]   =   {   1   ,   3   ,   3   ,   9   ,   9   ,   27  ,   83  ,   135 ,   249 ,   101 ,   1181    ,   451 ,   6063    ,   609 ,   15371   ,0 };
            const unsigned long dim2647Kuo3Init[]   =   {   1   ,   1   ,   5   ,   5   ,   15  ,   61  ,   77  ,   221 ,   19  ,   643 ,   1535    ,   1223    ,   7165    ,   10969   ,   10367   ,0 };
            const unsigned long dim2648Kuo3Init[]   =   {   1   ,   1   ,   7   ,   13  ,   5   ,   9   ,   61  ,   95  ,   493 ,   709 ,   71  ,   1571    ,   6805    ,   9687    ,   12217   ,0 };
            const unsigned long dim2649Kuo3Init[]   =   {   1   ,   3   ,   5   ,   7   ,   17  ,   55  ,   57  ,   179 ,   217 ,   699 ,   943 ,   127 ,   1603    ,   6851    ,   8579    ,0 };
            const unsigned long dim2650Kuo3Init[]   =   {   1   ,   3   ,   1   ,   11  ,   25  ,   51  ,   127 ,   13  ,   429 ,   15  ,   949 ,   2337    ,   5743    ,   14255   ,   15137   ,0 };
            const unsigned long dim2651Kuo3Init[]   =   {   1   ,   3   ,   7   ,   15  ,   5   ,   39  ,   39  ,   105 ,   119 ,   355 ,   1655    ,   3819    ,   1313    ,   5173    ,   7791    ,0 };
            const unsigned long dim2652Kuo3Init[]   =   {   1   ,   3   ,   7   ,   5   ,   3   ,   47  ,   93  ,   85  ,   19  ,   399 ,   303 ,   1333    ,   5435    ,   9425    ,   6503    ,0 };
            const unsigned long dim2653Kuo3Init[]   =   {   1   ,   3   ,   1   ,   7   ,   21  ,   33  ,   71  ,   253 ,   489 ,   149 ,   227 ,   797 ,   955 ,   6839    ,   23037   ,0 };
            const unsigned long dim2654Kuo3Init[]   =   {   1   ,   1   ,   7   ,   1   ,   27  ,   59  ,   15  ,   47  ,   249 ,   557 ,   1477    ,   1837    ,   4409    ,   14723   ,   14933   ,0 };
            const unsigned long dim2655Kuo3Init[]   =   {   1   ,   3   ,   5   ,   13  ,   1   ,   47  ,   11  ,   205 ,   131 ,   17  ,   373 ,   2039    ,   1873    ,   3041    ,   22869   ,0 };
            const unsigned long dim2656Kuo3Init[]   =   {   1   ,   1   ,   3   ,   1   ,   11  ,   27  ,   99  ,   135 ,   195 ,   419 ,   561 ,   1027    ,   1849    ,   13533   ,   7701    ,0 };
            const unsigned long dim2657Kuo3Init[]   =   {   1   ,   1   ,   7   ,   15  ,   27  ,   23  ,   77  ,   111 ,   157 ,   417 ,   591 ,   3435    ,   5999    ,   9481    ,   17073   ,0 };
            const unsigned long dim2658Kuo3Init[]   =   {   1   ,   3   ,   5   ,   3   ,   9   ,   61  ,   57  ,   43  ,   85  ,   585 ,   1861    ,   703 ,   2337    ,   8565    ,   16233   ,0 };
            const unsigned long dim2659Kuo3Init[]   =   {   1   ,   1   ,   3   ,   11  ,   7   ,   17  ,   41  ,   213 ,   145 ,   325 ,   1847    ,   3205    ,   3515    ,   1117    ,   285 ,0 };
            const unsigned long dim2660Kuo3Init[]   =   {   1   ,   3   ,   5   ,   11  ,   7   ,   35  ,   95  ,   127 ,   57  ,   123 ,   759 ,   129 ,   4335    ,   15183   ,   2107    ,0 };
            const unsigned long dim2661Kuo3Init[]   =   {   1   ,   3   ,   7   ,   7   ,   13  ,   1   ,   87  ,   193 ,   63  ,   201 ,   741 ,   2087    ,   1625    ,   1677    ,   31603   ,0 };
            const unsigned long dim2662Kuo3Init[]   =   {   1   ,   1   ,   7   ,   11  ,   13  ,   53  ,   127 ,   235 ,   489 ,   173 ,   1551    ,   1897    ,   5877    ,   15559   ,   24445   ,0 };
            const unsigned long dim2663Kuo3Init[]   =   {   1   ,   1   ,   7   ,   7   ,   9   ,   53  ,   59  ,   247 ,   83  ,   521 ,   2029    ,   3677    ,   2083    ,   11635   ,   30517   ,0 };
            const unsigned long dim2664Kuo3Init[]   =   {   1   ,   1   ,   1   ,   5   ,   9   ,   19  ,   71  ,   99  ,   213 ,   663 ,   947 ,   1929    ,   5275    ,   9715    ,   10033   ,0 };
            const unsigned long dim2665Kuo3Init[]   =   {   1   ,   1   ,   5   ,   11  ,   25  ,   31  ,   33  ,   239 ,   137 ,   445 ,   1227    ,   1051    ,   7405    ,   7983    ,   10079   ,0 };
            const unsigned long dim2666Kuo3Init[]   =   {   1   ,   3   ,   3   ,   15  ,   27  ,   35  ,   51  ,   207 ,   167 ,   519 ,   303 ,   3399    ,   4393    ,   1623    ,   21225   ,0 };
            const unsigned long dim2667Kuo3Init[]   =   {   1   ,   3   ,   1   ,   11  ,   17  ,   13  ,   41  ,   229 ,   441 ,   879 ,   1695    ,   1351    ,   5745    ,   1907    ,   15517   ,0 };
            const unsigned long dim2668Kuo3Init[]   =   {   1   ,   3   ,   7   ,   1   ,   3   ,   1   ,   95  ,   105 ,   333 ,   11  ,   1709    ,   2783    ,   3011    ,   5579    ,   14939   ,0 };
            const unsigned long dim2669Kuo3Init[]   =   {   1   ,   3   ,   1   ,   1   ,   5   ,   33  ,   67  ,   103 ,   71  ,   707 ,   325 ,   315 ,   567 ,   11969   ,   23071   ,0 };
            const unsigned long dim2670Kuo3Init[]   =   {   1   ,   1   ,   7   ,   11  ,   5   ,   61  ,   47  ,   131 ,   413 ,   507 ,   991 ,   27  ,   3873    ,   13151   ,   4115    ,0 };
            const unsigned long dim2671Kuo3Init[]   =   {   1   ,   1   ,   5   ,   11  ,   9   ,   31  ,   103 ,   91  ,   297 ,   419 ,   1251    ,   1155    ,   5539    ,   10347   ,   26811   ,0 };
            const unsigned long dim2672Kuo3Init[]   =   {   1   ,   3   ,   3   ,   3   ,   15  ,   37  ,   37  ,   167 ,   307 ,   397 ,   561 ,   2431    ,   6219    ,   3401    ,   23551   ,0 };
            const unsigned long dim2673Kuo3Init[]   =   {   1   ,   3   ,   7   ,   7   ,   29  ,   11  ,   33  ,   7   ,   177 ,   625 ,   853 ,   2317    ,   7371    ,   7435    ,   18135   ,0 };
            const unsigned long dim2674Kuo3Init[]   =   {   1   ,   1   ,   5   ,   1   ,   29  ,   7   ,   107 ,   97  ,   3   ,   97  ,   901 ,   717 ,   1911    ,   3877    ,   29583   ,0 };
            const unsigned long dim2675Kuo3Init[]   =   {   1   ,   1   ,   7   ,   3   ,   27  ,   33  ,   33  ,   105 ,   285 ,   3   ,   1427    ,   13  ,   3941    ,   299 ,   12025   ,0 };
            const unsigned long dim2676Kuo3Init[]   =   {   1   ,   1   ,   3   ,   7   ,   13  ,   57  ,   81  ,   131 ,   327 ,   587 ,   445 ,   1313    ,   4797    ,   8587    ,   5915    ,0 };
            const unsigned long dim2677Kuo3Init[]   =   {   1   ,   1   ,   3   ,   13  ,   21  ,   29  ,   43  ,   89  ,   499 ,   893 ,   501 ,   2409    ,   3145    ,   9435    ,   10831   ,0 };
            const unsigned long dim2678Kuo3Init[]   =   {   1   ,   1   ,   7   ,   11  ,   25  ,   21  ,   23  ,   211 ,   37  ,   993 ,   199 ,   3791    ,   291 ,   465 ,   32513   ,0 };
            const unsigned long dim2679Kuo3Init[]   =   {   1   ,   1   ,   5   ,   11  ,   15  ,   53  ,   123 ,   91  ,   407 ,   925 ,   1125    ,   927 ,   6793    ,   1905    ,   27477   ,0 };
            const unsigned long dim2680Kuo3Init[]   =   {   1   ,   1   ,   1   ,   11  ,   27  ,   51  ,   1   ,   77  ,   5   ,   273 ,   1363    ,   3835    ,   1627    ,   14925   ,   9431    ,0 };
            const unsigned long dim2681Kuo3Init[]   =   {   1   ,   1   ,   7   ,   3   ,   29  ,   47  ,   47  ,   79  ,   305 ,   655 ,   1171    ,   899 ,   95  ,   9413    ,   14371   ,0 };
            const unsigned long dim2682Kuo3Init[]   =   {   1   ,   1   ,   7   ,   13  ,   31  ,   45  ,   5   ,   211 ,   331 ,   677 ,   1045    ,   929 ,   975 ,   2377    ,   15407   ,0 };
            const unsigned long dim2683Kuo3Init[]   =   {   1   ,   3   ,   5   ,   13  ,   21  ,   59  ,   61  ,   111 ,   449 ,   641 ,   597 ,   4023    ,   6333    ,   4439    ,   605 ,0 };
            const unsigned long dim2684Kuo3Init[]   =   {   1   ,   1   ,   3   ,   5   ,   29  ,   7   ,   89  ,   21  ,   373 ,   311 ,   433 ,   1091    ,   4871    ,   8951    ,   22591   ,0 };
            const unsigned long dim2685Kuo3Init[]   =   {   1   ,   1   ,   1   ,   7   ,   27  ,   17  ,   81  ,   77  ,   119 ,   1005    ,   873 ,   2299    ,   6919    ,   657 ,   12479   ,0 };
            const unsigned long dim2686Kuo3Init[]   =   {   1   ,   3   ,   3   ,   5   ,   29  ,   41  ,   31  ,   133 ,   411 ,   655 ,   1753    ,   1365    ,   1807    ,   15975   ,   23783   ,0 };
            const unsigned long dim2687Kuo3Init[]   =   {   1   ,   1   ,   3   ,   9   ,   21  ,   11  ,   79  ,   65  ,   149 ,   121 ,   1393    ,   983 ,   885 ,   6991    ,   14943   ,0 };
            const unsigned long dim2688Kuo3Init[]   =   {   1   ,   1   ,   1   ,   3   ,   17  ,   39  ,   123 ,   35  ,   101 ,   363 ,   1991    ,   3309    ,   2117    ,   16229   ,   3017    ,0 };
            const unsigned long dim2689Kuo3Init[]   =   {   1   ,   3   ,   3   ,   15  ,   19  ,   61  ,   41  ,   201 ,   267 ,   985 ,   1635    ,   2681    ,   883 ,   13343   ,   22061   ,0 };
            const unsigned long dim2690Kuo3Init[]   =   {   1   ,   1   ,   7   ,   7   ,   9   ,   13  ,   15  ,   139 ,   25  ,   989 ,   1817    ,   867 ,   5913    ,   16197   ,   7521    ,0 };
            const unsigned long dim2691Kuo3Init[]   =   {   1   ,   1   ,   5   ,   5   ,   11  ,   29  ,   49  ,   75  ,   135 ,   705 ,   315 ,   2769    ,   3539    ,   13655   ,   11443   ,0 };
            const unsigned long dim2692Kuo3Init[]   =   {   1   ,   1   ,   7   ,   1   ,   25  ,   9   ,   53  ,   81  ,   217 ,   391 ,   1789    ,   3837    ,   1191    ,   12359   ,   21831   ,0 };
            const unsigned long dim2693Kuo3Init[]   =   {   1   ,   1   ,   5   ,   9   ,   3   ,   39  ,   1   ,   81  ,   367 ,   893 ,   1661    ,   3693    ,   9   ,   4061    ,   15321   ,0 };
            const unsigned long dim2694Kuo3Init[]   =   {   1   ,   1   ,   5   ,   5   ,   27  ,   21  ,   11  ,   23  ,   287 ,   901 ,   1695    ,   469 ,   1993    ,   11435   ,   12637   ,0 };
            const unsigned long dim2695Kuo3Init[]   =   {   1   ,   1   ,   3   ,   7   ,   21  ,   25  ,   113 ,   161 ,   361 ,   583 ,   2023    ,   831 ,   7161    ,   5125    ,   21097   ,0 };
            const unsigned long dim2696Kuo3Init[]   =   {   1   ,   1   ,   1   ,   13  ,   11  ,   29  ,   21  ,   221 ,   255 ,   255 ,   693 ,   701 ,   989 ,   10323   ,   3165    ,0 };
            const unsigned long dim2697Kuo3Init[]   =   {   1   ,   1   ,   5   ,   15  ,   5   ,   27  ,   107 ,   213 ,   85  ,   325 ,   1359    ,   735 ,   161 ,   11505   ,   24007   ,0 };
            const unsigned long dim2698Kuo3Init[]   =   {   1   ,   3   ,   1   ,   9   ,   1   ,   47  ,   17  ,   241 ,   53  ,   563 ,   635 ,   3159    ,   3231    ,   12167   ,   221 ,0 };
            const unsigned long dim2699Kuo3Init[]   =   {   1   ,   1   ,   1   ,   9   ,   21  ,   51  ,   15  ,   1   ,   491 ,   283 ,   1933    ,   1455    ,   6921    ,   10547   ,   29725   ,0 };
            const unsigned long dim2700Kuo3Init[]   =   {   1   ,   3   ,   3   ,   3   ,   31  ,   43  ,   123 ,   103 ,   291 ,   479 ,   1207    ,   1703    ,   5591    ,   6555    ,   37  ,0 };
            const unsigned long dim2701Kuo3Init[]   =   {   1   ,   3   ,   7   ,   7   ,   13  ,   53  ,   29  ,   155 ,   101 ,   651 ,   423 ,   2705    ,   6469    ,   13795   ,   9903    ,0 };
            const unsigned long dim2702Kuo3Init[]   =   {   1   ,   1   ,   7   ,   3   ,   23  ,   27  ,   39  ,   181 ,   47  ,   297 ,   353 ,   327 ,   6723    ,   5981    ,   8333    ,0 };
            const unsigned long dim2703Kuo3Init[]   =   {   1   ,   1   ,   1   ,   1   ,   29  ,   41  ,   55  ,   177 ,   459 ,   933 ,   675 ,   177 ,   1459    ,   6371    ,   8301    ,0 };
            const unsigned long dim2704Kuo3Init[]   =   {   1   ,   3   ,   5   ,   1   ,   11  ,   29  ,   45  ,   119 ,   145 ,   793 ,   1771    ,   2295    ,   2349    ,   13825   ,   10389   ,0 };
            const unsigned long dim2705Kuo3Init[]   =   {   1   ,   3   ,   7   ,   7   ,   13  ,   9   ,   85  ,   201 ,   509 ,   285 ,   1613    ,   2571    ,   3925    ,   4303    ,   14625   ,0 };
            const unsigned long dim2706Kuo3Init[]   =   {   1   ,   3   ,   5   ,   1   ,   17  ,   53  ,   89  ,   93  ,   401 ,   429 ,   865 ,   679 ,   1995    ,   13567   ,   32275   ,0 };
            const unsigned long dim2707Kuo3Init[]   =   {   1   ,   1   ,   1   ,   3   ,   17  ,   23  ,   97  ,   39  ,   309 ,   67  ,   1445    ,   101 ,   4305    ,   12195   ,   18559   ,0 };
            const unsigned long dim2708Kuo3Init[]   =   {   1   ,   1   ,   5   ,   11  ,   27  ,   35  ,   119 ,   67  ,   411 ,   461 ,   1149    ,   1651    ,   6149    ,   1543    ,   19747   ,0 };
            const unsigned long dim2709Kuo3Init[]   =   {   1   ,   1   ,   5   ,   9   ,   23  ,   51  ,   25  ,   109 ,   283 ,   331 ,   1651    ,   691 ,   7953    ,   1979    ,   7425    ,0 };
            const unsigned long dim2710Kuo3Init[]   =   {   1   ,   3   ,   3   ,   1   ,   31  ,   63  ,   113 ,   191 ,   481 ,   467 ,   589 ,   2375    ,   4619    ,   3745    ,   4343    ,0 };
            const unsigned long dim2711Kuo3Init[]   =   {   1   ,   1   ,   3   ,   1   ,   27  ,   35  ,   61  ,   29  ,   115 ,   909 ,   107 ,   815 ,   1579    ,   8373    ,   32657   ,0 };
            const unsigned long dim2712Kuo3Init[]   =   {   1   ,   3   ,   3   ,   11  ,   21  ,   45  ,   101 ,   27  ,   229 ,   1011    ,   497 ,   841 ,   7653    ,   5429    ,   17329   ,0 };
            const unsigned long dim2713Kuo3Init[]   =   {   1   ,   3   ,   7   ,   9   ,   27  ,   5   ,   105 ,   171 ,   75  ,   45  ,   865 ,   4061    ,   7947    ,   4979    ,   20577   ,0 };
            const unsigned long dim2714Kuo3Init[]   =   {   1   ,   3   ,   7   ,   15  ,   9   ,   45  ,   67  ,   221 ,   27  ,   31  ,   187 ,   2063    ,   4733    ,   10087   ,   23091   ,0 };
            const unsigned long dim2715Kuo3Init[]   =   {   1   ,   3   ,   5   ,   15  ,   3   ,   51  ,   79  ,   117 ,   379 ,   687 ,   721 ,   3645    ,   6639    ,   14161   ,   16187   ,0 };
            const unsigned long dim2716Kuo3Init[]   =   {   1   ,   3   ,   1   ,   1   ,   5   ,   15  ,   117 ,   177 ,   163 ,   577 ,   279 ,   1549    ,   2397    ,   8099    ,   2125    ,0 };
            const unsigned long dim2717Kuo3Init[]   =   {   1   ,   3   ,   3   ,   5   ,   17  ,   11  ,   121 ,   89  ,   313 ,   541 ,   47  ,   1993    ,   3313    ,   159 ,   3761    ,0 };
            const unsigned long dim2718Kuo3Init[]   =   {   1   ,   3   ,   5   ,   7   ,   11  ,   43  ,   7   ,   165 ,   283 ,   7   ,   1275    ,   3391    ,   417 ,   11697   ,   22357   ,0 };
            const unsigned long dim2719Kuo3Init[]   =   {   1   ,   1   ,   3   ,   13  ,   27  ,   47  ,   65  ,   51  ,   83  ,   223 ,   2015    ,   2939    ,   6121    ,   3873    ,   27573   ,0 };
            const unsigned long dim2720Kuo3Init[]   =   {   1   ,   3   ,   1   ,   3   ,   19  ,   55  ,   115 ,   105 ,   345 ,   393 ,   1609    ,   3131    ,   109 ,   11443   ,   23475   ,0 };
            const unsigned long dim2721Kuo3Init[]   =   {   1   ,   3   ,   3   ,   15  ,   21  ,   61  ,   39  ,   69  ,   381 ,   279 ,   1473    ,   1239    ,   661 ,   781 ,   21603   ,0 };
            const unsigned long dim2722Kuo3Init[]   =   {   1   ,   1   ,   7   ,   1   ,   3   ,   7   ,   7   ,   129 ,   229 ,   233 ,   77  ,   845 ,   7871    ,   3013    ,   14469   ,0 };
            const unsigned long dim2723Kuo3Init[]   =   {   1   ,   1   ,   7   ,   1   ,   17  ,   9   ,   3   ,   17  ,   375 ,   27  ,   473 ,   945 ,   4111    ,   7645    ,   7899    ,0 };
            const unsigned long dim2724Kuo3Init[]   =   {   1   ,   1   ,   7   ,   3   ,   25  ,   31  ,   71  ,   157 ,   269 ,   1003    ,   1875    ,   3573    ,   8121    ,   14073   ,   26019   ,0 };
            const unsigned long dim2725Kuo3Init[]   =   {   1   ,   3   ,   3   ,   13  ,   17  ,   25  ,   57  ,   1   ,   63  ,   425 ,   507 ,   2341    ,   4891    ,   12643   ,   32083   ,0 };
            const unsigned long dim2726Kuo3Init[]   =   {   1   ,   1   ,   3   ,   13  ,   23  ,   47  ,   55  ,   145 ,   269 ,   417 ,   559 ,   2065    ,   5801    ,   13751   ,   15617   ,0 };
            const unsigned long dim2727Kuo3Init[]   =   {   1   ,   3   ,   7   ,   9   ,   3   ,   39  ,   45  ,   143 ,   131 ,   713 ,   1549    ,   779 ,   1427    ,   10819   ,   10629   ,0 };
            const unsigned long dim2728Kuo3Init[]   =   {   1   ,   1   ,   5   ,   3   ,   13  ,   61  ,   99  ,   145 ,   293 ,   393 ,   1181    ,   4011    ,   6843    ,   12667   ,   21301   ,0 };
            const unsigned long dim2729Kuo3Init[]   =   {   1   ,   3   ,   7   ,   1   ,   19  ,   29  ,   105 ,   109 ,   391 ,   569 ,   1933    ,   1799    ,   8045    ,   509 ,   13251   ,0 };
            const unsigned long dim2730Kuo3Init[]   =   {   1   ,   1   ,   3   ,   3   ,   21  ,   55  ,   123 ,   239 ,   65  ,   891 ,   635 ,   3107    ,   2701    ,   9041    ,   28249   ,0 };
            const unsigned long dim2731Kuo3Init[]   =   {   1   ,   1   ,   3   ,   7   ,   1   ,   11  ,   9   ,   49  ,   51  ,   243 ,   51  ,   3   ,   6813    ,   14995   ,   27065   ,0 };
            const unsigned long dim2732Kuo3Init[]   =   {   1   ,   1   ,   3   ,   11  ,   29  ,   35  ,   107 ,   19  ,   415 ,   273 ,   881 ,   49  ,   3267    ,   4791    ,   9609    ,0 };
            const unsigned long dim2733Kuo3Init[]   =   {   1   ,   1   ,   5   ,   3   ,   5   ,   39  ,   59  ,   127 ,   267 ,   215 ,   375 ,   133 ,   8057    ,   8497    ,   23595   ,0 };
            const unsigned long dim2734Kuo3Init[]   =   {   1   ,   3   ,   7   ,   9   ,   1   ,   41  ,   81  ,   199 ,   287 ,   907 ,   743 ,   2875    ,   3873    ,   2851    ,   7385    ,0 };
            const unsigned long dim2735Kuo3Init[]   =   {   1   ,   3   ,   1   ,   13  ,   13  ,   49  ,   25  ,   43  ,   105 ,   511 ,   1483    ,   3277    ,   7561    ,   7211    ,   20671   ,0 };
            const unsigned long dim2736Kuo3Init[]   =   {   1   ,   3   ,   5   ,   13  ,   19  ,   21  ,   37  ,   87  ,   439 ,   69  ,   1145    ,   141 ,   6187    ,   6329    ,   26923   ,0 };
            const unsigned long dim2737Kuo3Init[]   =   {   1   ,   3   ,   1   ,   13  ,   13  ,   59  ,   123 ,   155 ,   189 ,   803 ,   1807    ,   363 ,   2239    ,   547 ,   30801   ,0 };
            const unsigned long dim2738Kuo3Init[]   =   {   1   ,   3   ,   7   ,   7   ,   1   ,   61  ,   93  ,   25  ,   205 ,   275 ,   1591    ,   1575    ,   3951    ,   12313   ,   9399    ,0 };
            const unsigned long dim2739Kuo3Init[]   =   {   1   ,   1   ,   1   ,   9   ,   11  ,   49  ,   43  ,   241 ,   435 ,   515 ,   283 ,   3667    ,   2089    ,   4755    ,   19427   ,0 };
            const unsigned long dim2740Kuo3Init[]   =   {   1   ,   1   ,   1   ,   5   ,   19  ,   57  ,   5   ,   23  ,   271 ,   113 ,   481 ,   2619    ,   4507    ,   15191   ,   2227    ,0 };
            const unsigned long dim2741Kuo3Init[]   =   {   1   ,   1   ,   1   ,   11  ,   11  ,   15  ,   49  ,   51  ,   121 ,   463 ,   1015    ,   791 ,   3993    ,   14311   ,   8723    ,0 };
            const unsigned long dim2742Kuo3Init[]   =   {   1   ,   1   ,   1   ,   7   ,   17  ,   19  ,   55  ,   137 ,   211 ,   307 ,   1647    ,   2719    ,   5231    ,   13005   ,   17527   ,0 };
            const unsigned long dim2743Kuo3Init[]   =   {   1   ,   1   ,   3   ,   3   ,   1   ,   25  ,   5   ,   119 ,   271 ,   611 ,   1799    ,   3149    ,   2159    ,   9147    ,   30765   ,0 };
            const unsigned long dim2744Kuo3Init[]   =   {   1   ,   1   ,   1   ,   15  ,   5   ,   19  ,   75  ,   137 ,   483 ,   457 ,   1913    ,   31  ,   9   ,   8333    ,   3315    ,0 };
            const unsigned long dim2745Kuo3Init[]   =   {   1   ,   3   ,   7   ,   7   ,   5   ,   11  ,   95  ,   83  ,   269 ,   105 ,   67  ,   2021    ,   6423    ,   10315   ,   28017   ,0 };
            const unsigned long dim2746Kuo3Init[]   =   {   1   ,   3   ,   1   ,   7   ,   27  ,   49  ,   7   ,   45  ,   297 ,   899 ,   665 ,   2473    ,   2863    ,   2171    ,   10693   ,0 };
            const unsigned long dim2747Kuo3Init[]   =   {   1   ,   1   ,   5   ,   5   ,   25  ,   17  ,   41  ,   155 ,   101 ,   355 ,   197 ,   3853    ,   2381    ,   9663    ,   25187   ,0 };
            const unsigned long dim2748Kuo3Init[]   =   {   1   ,   3   ,   7   ,   15  ,   15  ,   53  ,   119 ,   233 ,   79  ,   545 ,   1203    ,   1535    ,   3563    ,   2469    ,   29449   ,0 };
            const unsigned long dim2749Kuo3Init[]   =   {   1   ,   1   ,   5   ,   5   ,   27  ,   57  ,   113 ,   189 ,   3   ,   253 ,   1743    ,   3735    ,   5257    ,   10521   ,   18679   ,0 };
            const unsigned long dim2750Kuo3Init[]   =   {   1   ,   3   ,   7   ,   1   ,   5   ,   29  ,   117 ,   207 ,   101 ,   293 ,   615 ,   401 ,   7201    ,   11451   ,   13705   ,0 };
            const unsigned long dim2751Kuo3Init[]   =   {   1   ,   1   ,   5   ,   11  ,   13  ,   59  ,   55  ,   101 ,   253 ,   801 ,   1947    ,   3509    ,   655 ,   5555    ,   15517   ,0 };
            const unsigned long dim2752Kuo3Init[]   =   {   1   ,   3   ,   3   ,   5   ,   3   ,   19  ,   31  ,   93  ,   3   ,   1009    ,   1535    ,   261 ,   2259    ,   9835    ,   5797    ,0 };
            const unsigned long dim2753Kuo3Init[]   =   {   1   ,   3   ,   3   ,   9   ,   9   ,   53  ,   1   ,   213 ,   123 ,   271 ,   1137    ,   11  ,   6331    ,   9907    ,   13219   ,0 };
            const unsigned long dim2754Kuo3Init[]   =   {   1   ,   1   ,   7   ,   5   ,   17  ,   9   ,   83  ,   153 ,   231 ,   3   ,   1179    ,   735 ,   553 ,   10627   ,   4643    ,0 };
            const unsigned long dim2755Kuo3Init[]   =   {   1   ,   1   ,   1   ,   11  ,   25  ,   21  ,   127 ,   85  ,   17  ,   541 ,   535 ,   1041    ,   5071    ,   15451   ,   909 ,0 };
            const unsigned long dim2756Kuo3Init[]   =   {   1   ,   3   ,   1   ,   7   ,   29  ,   31  ,   75  ,   91  ,   67  ,   253 ,   971 ,   327 ,   7491    ,   7627    ,   31857   ,0 };
            const unsigned long dim2757Kuo3Init[]   =   {   1   ,   3   ,   1   ,   5   ,   23  ,   19  ,   79  ,   197 ,   25  ,   83  ,   1427    ,   2023    ,   1861    ,   1617    ,   10129   ,0 };
            const unsigned long dim2758Kuo3Init[]   =   {   1   ,   1   ,   5   ,   13  ,   27  ,   55  ,   113 ,   191 ,   119 ,   301 ,   1243    ,   333 ,   7609    ,   4875    ,   19035   ,0 };
            const unsigned long dim2759Kuo3Init[]   =   {   1   ,   1   ,   5   ,   5   ,   7   ,   51  ,   59  ,   157 ,   21  ,   697 ,   171 ,   2857    ,   3875    ,   15413   ,   28749   ,0 };
            const unsigned long dim2760Kuo3Init[]   =   {   1   ,   1   ,   1   ,   11  ,   13  ,   45  ,   49  ,   161 ,   389 ,   715 ,   991 ,   839 ,   2655    ,   11399   ,   5261    ,0 };
            const unsigned long dim2761Kuo3Init[]   =   {   1   ,   1   ,   7   ,   5   ,   17  ,   51  ,   87  ,   75  ,   65  ,   67  ,   1283    ,   1991    ,   7735    ,   9063    ,   313 ,0 };
            const unsigned long dim2762Kuo3Init[]   =   {   1   ,   3   ,   3   ,   13  ,   29  ,   39  ,   109 ,   159 ,   63  ,   103 ,   587 ,   1941    ,   3883    ,   1119    ,   4303    ,0 };
            const unsigned long dim2763Kuo3Init[]   =   {   1   ,   3   ,   3   ,   1   ,   15  ,   27  ,   69  ,   49  ,   417 ,   281 ,   1207    ,   289 ,   1873    ,   12287   ,   19321   ,0 };
            const unsigned long dim2764Kuo3Init[]   =   {   1   ,   1   ,   1   ,   3   ,   9   ,   57  ,   77  ,   81  ,   267 ,   45  ,   835 ,   671 ,   4569    ,   12713   ,   32657   ,0 };
            const unsigned long dim2765Kuo3Init[]   =   {   1   ,   1   ,   7   ,   5   ,   25  ,   25  ,   73  ,   165 ,   413 ,   845 ,   529 ,   2987    ,   2051    ,   6155    ,   29133   ,0 };
            const unsigned long dim2766Kuo3Init[]   =   {   1   ,   3   ,   1   ,   13  ,   3   ,   41  ,   107 ,   85  ,   429 ,   323 ,   1509    ,   649 ,   5897    ,   10309   ,   30927   ,0 };
            const unsigned long dim2767Kuo3Init[]   =   {   1   ,   1   ,   7   ,   9   ,   1   ,   1   ,   105 ,   213 ,   479 ,   755 ,   1863    ,   1429    ,   5927    ,   9923    ,   18531   ,0 };
            const unsigned long dim2768Kuo3Init[]   =   {   1   ,   3   ,   1   ,   5   ,   9   ,   13  ,   45  ,   105 ,   133 ,   117 ,   1837    ,   3019    ,   3809    ,   8713    ,   19865   ,0 };
            const unsigned long dim2769Kuo3Init[]   =   {   1   ,   1   ,   3   ,   7   ,   13  ,   7   ,   31  ,   175 ,   209 ,   313 ,   1095    ,   1093    ,   2707    ,   11955   ,   11137   ,0 };
            const unsigned long dim2770Kuo3Init[]   =   {   1   ,   3   ,   7   ,   13  ,   9   ,   31  ,   79  ,   25  ,   205 ,   309 ,   1977    ,   1207    ,   5629    ,   6869    ,   24269   ,0 };
            const unsigned long dim2771Kuo3Init[]   =   {   1   ,   1   ,   5   ,   3   ,   5   ,   43  ,   31  ,   81  ,   251 ,   569 ,   2033    ,   299 ,   345 ,   7437    ,   2183    ,0 };
            const unsigned long dim2772Kuo3Init[]   =   {   1   ,   1   ,   5   ,   1   ,   23  ,   17  ,   107 ,   143 ,   31  ,   349 ,   1501    ,   2487    ,   6911    ,   12591   ,   17599   ,0 };
            const unsigned long dim2773Kuo3Init[]   =   {   1   ,   1   ,   7   ,   9   ,   5   ,   13  ,   99  ,   145 ,   141 ,   947 ,   1587    ,   3691    ,   6029    ,   13153   ,   13971   ,0 };
            const unsigned long dim2774Kuo3Init[]   =   {   1   ,   3   ,   1   ,   1   ,   21  ,   37  ,   87  ,   19  ,   195 ,   263 ,   85  ,   3285    ,   1733    ,   11057   ,   6293    ,0 };
            const unsigned long dim2775Kuo3Init[]   =   {   1   ,   1   ,   3   ,   3   ,   31  ,   29  ,   97  ,   81  ,   479 ,   879 ,   2011    ,   1041    ,   8141    ,   14891   ,   175 ,0 };
            const unsigned long dim2776Kuo3Init[]   =   {   1   ,   3   ,   5   ,   7   ,   13  ,   35  ,   61  ,   219 ,   103 ,   161 ,   1187    ,   3925    ,   2041    ,   2609    ,   30195   ,0 };
            const unsigned long dim2777Kuo3Init[]   =   {   1   ,   3   ,   1   ,   13  ,   15  ,   43  ,   31  ,   163 ,   401 ,   451 ,   1881    ,   1307    ,   4201    ,   6821    ,   3717    ,0 };
            const unsigned long dim2778Kuo3Init[]   =   {   1   ,   3   ,   7   ,   3   ,   5   ,   3   ,   85  ,   149 ,   47  ,   1003    ,   405 ,   1801    ,   6695    ,   897 ,   28005   ,0 };
            const unsigned long dim2779Kuo3Init[]   =   {   1   ,   1   ,   3   ,   11  ,   19  ,   37  ,   81  ,   41  ,   315 ,   729 ,   303 ,   501 ,   7483    ,   10841   ,   12649   ,0 };
            const unsigned long dim2780Kuo3Init[]   =   {   1   ,   1   ,   1   ,   1   ,   1   ,   51  ,   121 ,   129 ,   49  ,   511 ,   719 ,   2235    ,   827 ,   9331    ,   22551   ,0 };
            const unsigned long dim2781Kuo3Init[]   =   {   1   ,   3   ,   3   ,   9   ,   21  ,   41  ,   47  ,   17  ,   157 ,   771 ,   867 ,   3685    ,   2387    ,   7409    ,   13889   ,0 };
            const unsigned long dim2782Kuo3Init[]   =   {   1   ,   3   ,   7   ,   7   ,   11  ,   53  ,   119 ,   19  ,   289 ,   337 ,   551 ,   2529    ,   321 ,   2599    ,   11153   ,0 };
            const unsigned long dim2783Kuo3Init[]   =   {   1   ,   3   ,   1   ,   9   ,   7   ,   33  ,   19  ,   7   ,   459 ,   729 ,   281 ,   947 ,   5469    ,   7497    ,   22893   ,0 };
            const unsigned long dim2784Kuo3Init[]   =   {   1   ,   1   ,   7   ,   13  ,   11  ,   41  ,   91  ,   217 ,   195 ,   483 ,   1659    ,   1619    ,   4201    ,   15825   ,   3357    ,0 };
            const unsigned long dim2785Kuo3Init[]   =   {   1   ,   1   ,   5   ,   9   ,   9   ,   29  ,   87  ,   153 ,   175 ,   967 ,   75  ,   385 ,   719 ,   12269   ,   28729   ,0 };
            const unsigned long dim2786Kuo3Init[]   =   {   1   ,   3   ,   3   ,   9   ,   13  ,   21  ,   115 ,   55  ,   303 ,   853 ,   1735    ,   3763    ,   31  ,   2547    ,   9735    ,0 };
            const unsigned long dim2787Kuo3Init[]   =   {   1   ,   3   ,   3   ,   9   ,   17  ,   25  ,   109 ,   207 ,   465 ,   627 ,   1171    ,   557 ,   2577    ,   13873   ,   30945   ,0 };
            const unsigned long dim2788Kuo3Init[]   =   {   1   ,   1   ,   1   ,   1   ,   27  ,   25  ,   65  ,   131 ,   373 ,   297 ,   1561    ,   2481    ,   3491    ,   3175    ,   12527   ,0 };
            const unsigned long dim2789Kuo3Init[]   =   {   1   ,   1   ,   7   ,   11  ,   13  ,   33  ,   81  ,   81  ,   337 ,   249 ,   739 ,   1451    ,   4479    ,   10253   ,   3585    ,0 };
            const unsigned long dim2790Kuo3Init[]   =   {   1   ,   1   ,   1   ,   13  ,   3   ,   9   ,   11  ,   109 ,   261 ,   663 ,   507 ,   897 ,   4081    ,   12393   ,   21621   ,0 };
            const unsigned long dim2791Kuo3Init[]   =   {   1   ,   1   ,   1   ,   9   ,   27  ,   35  ,   13  ,   191 ,   433 ,   111 ,   1567    ,   851 ,   6303    ,   2411    ,   25985   ,0 };
            const unsigned long dim2792Kuo3Init[]   =   {   1   ,   3   ,   3   ,   1   ,   7   ,   41  ,   91  ,   217 ,   203 ,   821 ,   883 ,   1549    ,   37  ,   1837    ,   19549   ,0 };
            const unsigned long dim2793Kuo3Init[]   =   {   1   ,   3   ,   3   ,   11  ,   23  ,   49  ,   27  ,   81  ,   475 ,   251 ,   985 ,   2527    ,   3857    ,   4579    ,   30231   ,0 };
            const unsigned long dim2794Kuo3Init[]   =   {   1   ,   1   ,   3   ,   11  ,   7   ,   33  ,   71  ,   207 ,   179 ,   77  ,   743 ,   375 ,   5649    ,   3165    ,   25063   ,0 };
            const unsigned long dim2795Kuo3Init[]   =   {   1   ,   1   ,   3   ,   9   ,   25  ,   47  ,   21  ,   109 ,   375 ,   525 ,   671 ,   1117    ,   5663    ,   9835    ,   19675   ,0 };
            const unsigned long dim2796Kuo3Init[]   =   {   1   ,   3   ,   7   ,   11  ,   19  ,   31  ,   9   ,   167 ,   501 ,   477 ,   1015    ,   109 ,   6233    ,   14131   ,   30015   ,0 };
            const unsigned long dim2797Kuo3Init[]   =   {   1   ,   3   ,   1   ,   3   ,   25  ,   9   ,   91  ,   121 ,   373 ,   511 ,   1925    ,   3609    ,   7681    ,   7321    ,   19931   ,0 };
            const unsigned long dim2798Kuo3Init[]   =   {   1   ,   1   ,   3   ,   7   ,   23  ,   23  ,   63  ,   225 ,   511 ,   407 ,   573 ,   587 ,   6877    ,   4747    ,   31503   ,0 };
            const unsigned long dim2799Kuo3Init[]   =   {   1   ,   3   ,   7   ,   3   ,   7   ,   11  ,   61  ,   119 ,   149 ,   119 ,   1497    ,   165 ,   5833    ,   14727   ,   21075   ,0 };
            const unsigned long dim2800Kuo3Init[]   =   {   1   ,   3   ,   7   ,   5   ,   23  ,   27  ,   113 ,   99  ,   97  ,   385 ,   563 ,   1745    ,   4321    ,   2201    ,   5707    ,0 };
            const unsigned long dim2801Kuo3Init[]   =   {   1   ,   1   ,   3   ,   15  ,   5   ,   51  ,   49  ,   213 ,   111 ,   587 ,   1433    ,   1473    ,   4673    ,   369 ,   31405   ,0 };
            const unsigned long dim2802Kuo3Init[]   =   {   1   ,   1   ,   5   ,   11  ,   25  ,   49  ,   67  ,   163 ,   109 ,   679 ,   975 ,   3711    ,   6627    ,   2273    ,   13451   ,0 };
            const unsigned long dim2803Kuo3Init[]   =   {   1   ,   3   ,   5   ,   3   ,   9   ,   3   ,   75  ,   123 ,   71  ,   181 ,   1007    ,   2459    ,   5671    ,   2805    ,   15839   ,0 };
            const unsigned long dim2804Kuo3Init[]   =   {   1   ,   3   ,   5   ,   3   ,   7   ,   51  ,   49  ,   19  ,   89  ,   209 ,   1097    ,   359 ,   2487    ,   12969   ,   24243   ,0 };
            const unsigned long dim2805Kuo3Init[]   =   {   1   ,   1   ,   7   ,   11  ,   1   ,   47  ,   61  ,   65  ,   161 ,   801 ,   511 ,   3711    ,   237 ,   14741   ,   14239   ,0 };
            const unsigned long dim2806Kuo3Init[]   =   {   1   ,   3   ,   7   ,   13  ,   31  ,   25  ,   11  ,   175 ,   151 ,   79  ,   1749    ,   751 ,   6129    ,   12139   ,   30837   ,0 };
            const unsigned long dim2807Kuo3Init[]   =   {   1   ,   3   ,   3   ,   11  ,   19  ,   13  ,   71  ,   75  ,   479 ,   211 ,   395 ,   1935    ,   7949    ,   12049   ,   31539   ,0 };
            const unsigned long dim2808Kuo3Init[]   =   {   1   ,   3   ,   5   ,   3   ,   19  ,   9   ,   93  ,   79  ,   69  ,   1001    ,   673 ,   1963    ,   8047    ,   8023    ,   8199    ,0 };
            const unsigned long dim2809Kuo3Init[]   =   {   1   ,   3   ,   7   ,   3   ,   11  ,   9   ,   107 ,   39  ,   319 ,   835 ,   1297    ,   1667    ,   393 ,   5619    ,   15019   ,0 };
            const unsigned long dim2810Kuo3Init[]   =   {   1   ,   3   ,   1   ,   3   ,   17  ,   25  ,   71  ,   227 ,   219 ,   455 ,   29  ,   1503    ,   4207    ,   11523   ,   13077   ,0 };
            const unsigned long dim2811Kuo3Init[]   =   {   1   ,   3   ,   7   ,   1   ,   11  ,   11  ,   91  ,   203 ,   483 ,   959 ,   911 ,   979 ,   4327    ,   15113   ,   23415   ,0 };
            const unsigned long dim2812Kuo3Init[]   =   {   1   ,   3   ,   7   ,   15  ,   5   ,   63  ,   57  ,   253 ,   369 ,   609 ,   1831    ,   2895    ,   1025    ,   2567    ,   14017   ,0 };
            const unsigned long dim2813Kuo3Init[]   =   {   1   ,   3   ,   5   ,   1   ,   19  ,   39  ,   107 ,   169 ,   107 ,   863 ,   591 ,   4077    ,   7441    ,   3201    ,   269 ,0 };
            const unsigned long dim2814Kuo3Init[]   =   {   1   ,   1   ,   3   ,   1   ,   7   ,   45  ,   19  ,   103 ,   119 ,   539 ,   9   ,   2909    ,   8135    ,   7913    ,   1403    ,0 };
            const unsigned long dim2815Kuo3Init[]   =   {   1   ,   3   ,   1   ,   13  ,   17  ,   39  ,   125 ,   57  ,   417 ,   543 ,   1067    ,   3807    ,   397 ,   11947   ,   7655    ,0 };
            const unsigned long dim2816Kuo3Init[]   =   {   1   ,   3   ,   5   ,   7   ,   31  ,   49  ,   65  ,   23  ,   197 ,   777 ,   1669    ,   1373    ,   773 ,   6947    ,   2835    ,0 };
            const unsigned long dim2817Kuo3Init[]   =   {   1   ,   3   ,   7   ,   7   ,   25  ,   45  ,   115 ,   31  ,   351 ,   203 ,   1791    ,   2931    ,   7127    ,   12117   ,   10781   ,0 };
            const unsigned long dim2818Kuo3Init[]   =   {   1   ,   1   ,   5   ,   9   ,   21  ,   61  ,   47  ,   59  ,   393 ,   621 ,   811 ,   2321    ,   7801    ,   2255    ,   13513   ,0 };
            const unsigned long dim2819Kuo3Init[]   =   {   1   ,   1   ,   5   ,   9   ,   27  ,   15  ,   85  ,   53  ,   379 ,   597 ,   19  ,   267 ,   5511    ,   8167    ,   14843   ,0 };
            const unsigned long dim2820Kuo3Init[]   =   {   1   ,   1   ,   7   ,   9   ,   19  ,   11  ,   43  ,   101 ,   309 ,   125 ,   2023    ,   689 ,   589 ,   3621    ,   10593   ,0 };
            const unsigned long dim2821Kuo3Init[]   =   {   1   ,   3   ,   1   ,   1   ,   25  ,   57  ,   55  ,   57  ,   111 ,   743 ,   1773    ,   2223    ,   6567    ,   7401    ,   14843   ,0 };
            const unsigned long dim2822Kuo3Init[]   =   {   1   ,   3   ,   1   ,   11  ,   17  ,   43  ,   1   ,   11  ,   155 ,   287 ,   519 ,   4051    ,   311 ,   10079   ,   29753   ,0 };
            const unsigned long dim2823Kuo3Init[]   =   {   1   ,   3   ,   7   ,   1   ,   15  ,   33  ,   47  ,   117 ,   343 ,   667 ,   1489    ,   1985    ,   5105    ,   11887   ,   29117   ,0 };
            const unsigned long dim2824Kuo3Init[]   =   {   1   ,   3   ,   3   ,   13  ,   9   ,   33  ,   115 ,   211 ,   7   ,   391 ,   475 ,   1843    ,   3277    ,   11067   ,   12857   ,0 };
            const unsigned long dim2825Kuo3Init[]   =   {   1   ,   1   ,   7   ,   7   ,   31  ,   11  ,   17  ,   147 ,   129 ,   875 ,   845 ,   2999    ,   6395    ,   361 ,   7395    ,0 };
            const unsigned long dim2826Kuo3Init[]   =   {   1   ,   3   ,   5   ,   1   ,   27  ,   17  ,   55  ,   115 ,   35  ,   989 ,   153 ,   77  ,   5713    ,   14653   ,   587 ,0 };
            const unsigned long dim2827Kuo3Init[]   =   {   1   ,   3   ,   3   ,   5   ,   3   ,   37  ,   107 ,   57  ,   201 ,   631 ,   1549    ,   3709    ,   3615    ,   12567   ,   21729   ,0 };
            const unsigned long dim2828Kuo3Init[]   =   {   1   ,   3   ,   3   ,   1   ,   15  ,   19  ,   21  ,   205 ,   253 ,   431 ,   915 ,   1057    ,   359 ,   431 ,   31417   ,0 };
            const unsigned long dim2829Kuo3Init[]   =   {   1   ,   3   ,   7   ,   3   ,   11  ,   47  ,   107 ,   101 ,   461 ,   805 ,   1569    ,   1215    ,   765 ,   1547    ,   32531   ,0 };
            const unsigned long dim2830Kuo3Init[]   =   {   1   ,   3   ,   3   ,   3   ,   21  ,   55  ,   61  ,   171 ,   75  ,   337 ,   231 ,   1925    ,   7415    ,   1073    ,   17519   ,0 };
            const unsigned long dim2831Kuo3Init[]   =   {   1   ,   3   ,   5   ,   7   ,   5   ,   23  ,   39  ,   119 ,   399 ,   265 ,   601 ,   3487    ,   1129    ,   10731   ,   14795   ,0 };
            const unsigned long dim2832Kuo3Init[]   =   {   1   ,   1   ,   5   ,   3   ,   11  ,   57  ,   33  ,   99  ,   283 ,   581 ,   1791    ,   1723    ,   411 ,   5779    ,   8989    ,0 };
            const unsigned long dim2833Kuo3Init[]   =   {   1   ,   3   ,   3   ,   13  ,   19  ,   33  ,   81  ,   245 ,   375 ,   763 ,   77  ,   175 ,   5417    ,   12525   ,   18197   ,0 };
            const unsigned long dim2834Kuo3Init[]   =   {   1   ,   3   ,   7   ,   5   ,   29  ,   31  ,   59  ,   207 ,   19  ,   369 ,   1449    ,   165 ,   6115    ,   15417   ,   14227   ,0 };
            const unsigned long dim2835Kuo3Init[]   =   {   1   ,   3   ,   5   ,   11  ,   7   ,   15  ,   109 ,   171 ,   179 ,   541 ,   1907    ,   2545    ,   5633    ,   8039    ,   23495   ,0 };
            const unsigned long dim2836Kuo3Init[]   =   {   1   ,   3   ,   5   ,   7   ,   5   ,   5   ,   51  ,   5   ,   347 ,   351 ,   177 ,   965 ,   6583    ,   3631    ,   26493   ,0 };
            const unsigned long dim2837Kuo3Init[]   =   {   1   ,   1   ,   1   ,   15  ,   21  ,   55  ,   101 ,   137 ,   165 ,   947 ,   1149    ,   2615    ,   3709    ,   8231    ,   2033    ,0 };
            const unsigned long dim2838Kuo3Init[]   =   {   1   ,   3   ,   5   ,   1   ,   1   ,   11  ,   29  ,   5   ,   353 ,   675 ,   2045    ,   2111    ,   2477    ,   7803    ,   18925   ,0 };
            const unsigned long dim2839Kuo3Init[]   =   {   1   ,   3   ,   3   ,   7   ,   11  ,   5   ,   43  ,   241 ,   103 ,   87  ,   1841    ,   841 ,   277 ,   7647    ,   6953    ,0 };
            const unsigned long dim2840Kuo3Init[]   =   {   1   ,   1   ,   3   ,   9   ,   23  ,   15  ,   77  ,   199 ,   457 ,   183 ,   1719    ,   4077    ,   2751    ,   6285    ,   2741    ,0 };
            const unsigned long dim2841Kuo3Init[]   =   {   1   ,   3   ,   5   ,   13  ,   11  ,   7   ,   119 ,   207 ,   89  ,   415 ,   1501    ,   2333    ,   2155    ,   13559   ,   9701    ,0 };
            const unsigned long dim2842Kuo3Init[]   =   {   1   ,   1   ,   7   ,   1   ,   9   ,   29  ,   119 ,   171 ,   259 ,   849 ,   493 ,   3103    ,   2559    ,   4323    ,   8993    ,0 };
            const unsigned long dim2843Kuo3Init[]   =   {   1   ,   3   ,   7   ,   15  ,   5   ,   3   ,   35  ,   29  ,   177 ,   19  ,   565 ,   2111    ,   4781    ,   13603   ,   10759   ,0 };
            const unsigned long dim2844Kuo3Init[]   =   {   1   ,   1   ,   3   ,   13  ,   7   ,   41  ,   45  ,   181 ,   55  ,   575 ,   685 ,   2407    ,   3499    ,   10073   ,   5717    ,0 };
            const unsigned long dim2845Kuo3Init[]   =   {   1   ,   3   ,   5   ,   11  ,   25  ,   31  ,   97  ,   247 ,   411 ,   663 ,   413 ,   771 ,   2325    ,   4849    ,   18605   ,0 };
            const unsigned long dim2846Kuo3Init[]   =   {   1   ,   1   ,   1   ,   1   ,   11  ,   55  ,   77  ,   183 ,   131 ,   873 ,   315 ,   3403    ,   7219    ,   5419    ,   9943    ,0 };
            const unsigned long dim2847Kuo3Init[]   =   {   1   ,   1   ,   5   ,   11  ,   17  ,   49  ,   125 ,   1   ,   379 ,   995 ,   1911    ,   9   ,   1201    ,   14599   ,   3237    ,0 };
            const unsigned long dim2848Kuo3Init[]   =   {   1   ,   1   ,   5   ,   7   ,   5   ,   47  ,   25  ,   47  ,   93  ,   1009    ,   361 ,   3111    ,   6221    ,   15799   ,   11049   ,0 };
            const unsigned long dim2849Kuo3Init[]   =   {   1   ,   1   ,   1   ,   13  ,   31  ,   41  ,   27  ,   119 ,   463 ,   357 ,   1399    ,   3023    ,   8145    ,   7233    ,   10461   ,0 };
            const unsigned long dim2850Kuo3Init[]   =   {   1   ,   3   ,   1   ,   13  ,   25  ,   1   ,   89  ,   167 ,   13  ,   477 ,   989 ,   1109    ,   3279    ,   1893    ,   22091   ,0 };
            const unsigned long dim2851Kuo3Init[]   =   {   1   ,   1   ,   3   ,   11  ,   27  ,   27  ,   127 ,   85  ,   135 ,   975 ,   557 ,   2247    ,   5675    ,   7397    ,   28829   ,0 };
            const unsigned long dim2852Kuo3Init[]   =   {   1   ,   1   ,   3   ,   3   ,   5   ,   57  ,   87  ,   115 ,   399 ,   189 ,   497 ,   2525    ,   323 ,   2849    ,   1539    ,0 };
            const unsigned long dim2853Kuo3Init[]   =   {   1   ,   3   ,   5   ,   15  ,   31  ,   57  ,   49  ,   13  ,   151 ,   483 ,   1117    ,   2177    ,   4011    ,   2467    ,   5465    ,0 };
            const unsigned long dim2854Kuo3Init[]   =   {   1   ,   3   ,   1   ,   13  ,   17  ,   49  ,   31  ,   167 ,   385 ,   183 ,   225 ,   465 ,   2099    ,   475 ,   28001   ,0 };
            const unsigned long dim2855Kuo3Init[]   =   {   1   ,   1   ,   3   ,   3   ,   23  ,   39  ,   37  ,   63  ,   419 ,   95  ,   1217    ,   1147    ,   7599    ,   985 ,   23455   ,0 };
            const unsigned long dim2856Kuo3Init[]   =   {   1   ,   3   ,   1   ,   5   ,   15  ,   53  ,   17  ,   221 ,   441 ,   303 ,   1595    ,   3731    ,   7833    ,   14649   ,   30843   ,0 };
            const unsigned long dim2857Kuo3Init[]   =   {   1   ,   3   ,   3   ,   15  ,   13  ,   63  ,   29  ,   137 ,   263 ,   79  ,   591 ,   679 ,   4681    ,   4121    ,   22733   ,0 };
            const unsigned long dim2858Kuo3Init[]   =   {   1   ,   1   ,   3   ,   3   ,   27  ,   31  ,   97  ,   191 ,   295 ,   601 ,   1353    ,   3817    ,   2305    ,   427 ,   5985    ,0 };
            const unsigned long dim2859Kuo3Init[]   =   {   1   ,   3   ,   5   ,   7   ,   29  ,   27  ,   27  ,   249 ,   441 ,   415 ,   489 ,   779 ,   5927    ,   10569   ,   7539    ,0 };
            const unsigned long dim2860Kuo3Init[]   =   {   1   ,   3   ,   7   ,   1   ,   1   ,   59  ,   21  ,   115 ,   25  ,   715 ,   697 ,   3585    ,   383 ,   8043    ,   21457   ,0 };
            const unsigned long dim2861Kuo3Init[]   =   {   1   ,   1   ,   7   ,   1   ,   3   ,   37  ,   25  ,   239 ,   25  ,   213 ,   1449    ,   1975    ,   1805    ,   11867   ,   24163   ,0 };
            const unsigned long dim2862Kuo3Init[]   =   {   1   ,   3   ,   5   ,   7   ,   9   ,   37  ,   95  ,   19  ,   105 ,   787 ,   249 ,   2501    ,   5073    ,   2677    ,   2871    ,0 };
            const unsigned long dim2863Kuo3Init[]   =   {   1   ,   1   ,   1   ,   7   ,   25  ,   17  ,   59  ,   79  ,   467 ,   489 ,   1981    ,   169 ,   7713    ,   3035    ,   6135    ,0 };
            const unsigned long dim2864Kuo3Init[]   =   {   1   ,   1   ,   1   ,   11  ,   21  ,   41  ,   75  ,   59  ,   495 ,   475 ,   803 ,   3117    ,   4999    ,   5153    ,   20541   ,0 };
            const unsigned long dim2865Kuo3Init[]   =   {   1   ,   3   ,   5   ,   5   ,   21  ,   5   ,   43  ,   69  ,   167 ,   743 ,   1395    ,   3701    ,   1329    ,   12039   ,   26595   ,0 };
            const unsigned long dim2866Kuo3Init[]   =   {   1   ,   3   ,   5   ,   5   ,   7   ,   51  ,   7   ,   191 ,   341 ,   561 ,   617 ,   803 ,   965 ,   7995    ,   7605    ,0 };
            const unsigned long dim2867Kuo3Init[]   =   {   1   ,   3   ,   5   ,   1   ,   25  ,   21  ,   21  ,   141 ,   461 ,   353 ,   1909    ,   1089    ,   4603    ,   7285    ,   30359   ,0 };
            const unsigned long dim2868Kuo3Init[]   =   {   1   ,   3   ,   1   ,   7   ,   13  ,   17  ,   47  ,   173 ,   151 ,   1003    ,   1823    ,   1463    ,   6441    ,   8647    ,   8671    ,0 };
            const unsigned long dim2869Kuo3Init[]   =   {   1   ,   1   ,   3   ,   13  ,   29  ,   29  ,   111 ,   237 ,   341 ,   693 ,   773 ,   637 ,   741 ,   7737    ,   12077   ,0 };
            const unsigned long dim2870Kuo3Init[]   =   {   1   ,   1   ,   5   ,   9   ,   1   ,   27  ,   71  ,   177 ,   379 ,   289 ,   1659    ,   697 ,   4765    ,   15945   ,   24463   ,0 };
            const unsigned long dim2871Kuo3Init[]   =   {   1   ,   1   ,   7   ,   7   ,   15  ,   13  ,   77  ,   59  ,   5   ,   483 ,   581 ,   2193    ,   7765    ,   14443   ,   21545   ,0 };
            const unsigned long dim2872Kuo3Init[]   =   {   1   ,   3   ,   5   ,   11  ,   9   ,   17  ,   111 ,   13  ,   427 ,   405 ,   1765    ,   2029    ,   4989    ,   1995    ,   18581   ,0 };
            const unsigned long dim2873Kuo3Init[]   =   {   1   ,   1   ,   1   ,   1   ,   17  ,   15  ,   9   ,   43  ,   177 ,   277 ,   217 ,   1869    ,   2847    ,   11839   ,   5483    ,0 };
            const unsigned long dim2874Kuo3Init[]   =   {   1   ,   3   ,   5   ,   13  ,   1   ,   61  ,   87  ,   211 ,   367 ,   317 ,   1681    ,   3913    ,   2635    ,   5095    ,   2993    ,0 };
            const unsigned long dim2875Kuo3Init[]   =   {   1   ,   3   ,   5   ,   11  ,   13  ,   47  ,   123 ,   111 ,   139 ,   193 ,   1575    ,   3549    ,   2113    ,   2997    ,   2319    ,0 };
            const unsigned long dim2876Kuo3Init[]   =   {   1   ,   3   ,   7   ,   9   ,   17  ,   29  ,   67  ,   195 ,   455 ,   137 ,   1159    ,   367 ,   469 ,   4663    ,   13811   ,0 };
            const unsigned long dim2877Kuo3Init[]   =   {   1   ,   3   ,   1   ,   9   ,   21  ,   55  ,   27  ,   165 ,   483 ,   487 ,   1199    ,   2989    ,   3195    ,   8847    ,   12465   ,0 };
            const unsigned long dim2878Kuo3Init[]   =   {   1   ,   3   ,   3   ,   5   ,   13  ,   31  ,   5   ,   77  ,   283 ,   837 ,   827 ,   3477    ,   5555    ,   15411   ,   803 ,0 };
            const unsigned long dim2879Kuo3Init[]   =   {   1   ,   3   ,   5   ,   1   ,   5   ,   47  ,   23  ,   21  ,   255 ,   193 ,   67  ,   1489    ,   5785    ,   3247    ,   18425   ,0 };
            const unsigned long dim2880Kuo3Init[]   =   {   1   ,   1   ,   5   ,   11  ,   3   ,   41  ,   1   ,   37  ,   281 ,   767 ,   419 ,   1067    ,   749 ,   4985    ,   18525   ,0 };
            const unsigned long dim2881Kuo3Init[]   =   {   1   ,   1   ,   3   ,   7   ,   29  ,   47  ,   37  ,   195 ,   359 ,   517 ,   1809    ,   2111    ,   419 ,   1895    ,   17813   ,0 };
            const unsigned long dim2882Kuo3Init[]   =   {   1   ,   3   ,   1   ,   13  ,   13  ,   61  ,   99  ,   183 ,   311 ,   515 ,   1359    ,   1077    ,   835 ,   12945   ,   27707   ,0 };
            const unsigned long dim2883Kuo3Init[]   =   {   1   ,   3   ,   7   ,   15  ,   5   ,   25  ,   77  ,   7   ,   249 ,   325 ,   929 ,   3121    ,   4183    ,   9347    ,   1209    ,0 };
            const unsigned long dim2884Kuo3Init[]   =   {   1   ,   1   ,   1   ,   7   ,   13  ,   59  ,   9   ,   251 ,   323 ,   479 ,   1791    ,   555 ,   2591    ,   417 ,   23821   ,0 };
            const unsigned long dim2885Kuo3Init[]   =   {   1   ,   3   ,   7   ,   1   ,   31  ,   25  ,   43  ,   167 ,   337 ,   321 ,   1223    ,   3693    ,   1945    ,   4937    ,   15613   ,0 };
            const unsigned long dim2886Kuo3Init[]   =   {   1   ,   3   ,   1   ,   13  ,   23  ,   27  ,   71  ,   3   ,   249 ,   427 ,   1259    ,   2493    ,   3463    ,   8339    ,   2361    ,0 };
            const unsigned long dim2887Kuo3Init[]   =   {   1   ,   3   ,   3   ,   3   ,   25  ,   49  ,   19  ,   167 ,   189 ,   211 ,   1247    ,   2883    ,   6983    ,   14573   ,   1021    ,0 };
            const unsigned long dim2888Kuo3Init[]   =   {   1   ,   1   ,   7   ,   11  ,   5   ,   53  ,   119 ,   115 ,   497 ,   57  ,   1255    ,   2771    ,   6055    ,   865 ,   2735    ,0 };
            const unsigned long dim2889Kuo3Init[]   =   {   1   ,   1   ,   7   ,   11  ,   25  ,   63  ,   83  ,   115 ,   185 ,   615 ,   1135    ,   1577    ,   3097    ,   2955    ,   7667    ,0 };
            const unsigned long dim2890Kuo3Init[]   =   {   1   ,   3   ,   7   ,   7   ,   19  ,   33  ,   61  ,   121 ,   339 ,   153 ,   89  ,   1841    ,   3219    ,   10177   ,   13655   ,0 };
            const unsigned long dim2891Kuo3Init[]   =   {   1   ,   3   ,   5   ,   3   ,   23  ,   51  ,   111 ,   149 ,   283 ,   863 ,   1687    ,   813 ,   951 ,   10593   ,   2491    ,0 };
            const unsigned long dim2892Kuo3Init[]   =   {   1   ,   1   ,   5   ,   7   ,   19  ,   63  ,   47  ,   73  ,   91  ,   501 ,   1837    ,   4079    ,   313 ,   14141   ,   7843    ,0 };
            const unsigned long dim2893Kuo3Init[]   =   {   1   ,   3   ,   1   ,   9   ,   9   ,   15  ,   49  ,   195 ,   433 ,   289 ,   1109    ,   3929    ,   5589    ,   8943    ,   6325    ,0 };
            const unsigned long dim2894Kuo3Init[]   =   {   1   ,   3   ,   5   ,   13  ,   13  ,   1   ,   97  ,   111 ,   379 ,   963 ,   837 ,   3681    ,   3281    ,   153 ,   9133    ,0 };
            const unsigned long dim2895Kuo3Init[]   =   {   1   ,   3   ,   5   ,   1   ,   19  ,   37  ,   7   ,   81  ,   443 ,   555 ,   1885    ,   3123    ,   7667    ,   2579    ,   3521    ,0 };
            const unsigned long dim2896Kuo3Init[]   =   {   1   ,   3   ,   1   ,   1   ,   21  ,   17  ,   77  ,   121 ,   157 ,   993 ,   627 ,   893 ,   7633    ,   6421    ,   7315    ,0 };
            const unsigned long dim2897Kuo3Init[]   =   {   1   ,   1   ,   3   ,   15  ,   25  ,   21  ,   105 ,   187 ,   17  ,   239 ,   1745    ,   13  ,   1021    ,   13319   ,   8511    ,0 };
            const unsigned long dim2898Kuo3Init[]   =   {   1   ,   3   ,   1   ,   5   ,   5   ,   19  ,   67  ,   79  ,   79  ,   361 ,   1189    ,   3967    ,   6643    ,   2367    ,   18247   ,0 };
            const unsigned long dim2899Kuo3Init[]   =   {   1   ,   3   ,   5   ,   15  ,   5   ,   15  ,   17  ,   139 ,   405 ,   229 ,   1819    ,   369 ,   4011    ,   11369   ,   5271    ,0 };
            const unsigned long dim2900Kuo3Init[]   =   {   1   ,   3   ,   1   ,   3   ,   19  ,   3   ,   97  ,   121 ,   345 ,   917 ,   1473    ,   1439    ,   1007    ,   2535    ,   17107   ,0 };
            const unsigned long dim2901Kuo3Init[]   =   {   1   ,   3   ,   7   ,   3   ,   3   ,   55  ,   117 ,   183 ,   65  ,   589 ,   107 ,   3933    ,   6911    ,   6425    ,   28363   ,0 };
            const unsigned long dim2902Kuo3Init[]   =   {   1   ,   1   ,   7   ,   11  ,   25  ,   55  ,   57  ,   45  ,   217 ,   213 ,   281 ,   2211    ,   6851    ,   11553   ,   30021   ,0 };
            const unsigned long dim2903Kuo3Init[]   =   {   1   ,   3   ,   7   ,   11  ,   29  ,   5   ,   1   ,   15  ,   477 ,   233 ,   677 ,   311 ,   3207    ,   10527   ,   5543    ,0 };
            const unsigned long dim2904Kuo3Init[]   =   {   1   ,   3   ,   3   ,   15  ,   21  ,   57  ,   43  ,   9   ,   145 ,   369 ,   1367    ,   3241    ,   2821    ,   8753    ,   18579   ,0 };
            const unsigned long dim2905Kuo3Init[]   =   {   1   ,   1   ,   3   ,   1   ,   13  ,   15  ,   99  ,   237 ,   101 ,   983 ,   657 ,   1169    ,   617 ,   12859   ,   32279   ,0 };
            const unsigned long dim2906Kuo3Init[]   =   {   1   ,   3   ,   3   ,   3   ,   11  ,   61  ,   53  ,   39  ,   487 ,   967 ,   1239    ,   1899    ,   7063    ,   7943    ,   6739    ,0 };
            const unsigned long dim2907Kuo3Init[]   =   {   1   ,   1   ,   7   ,   15  ,   1   ,   17  ,   27  ,   121 ,   13  ,   519 ,   1049    ,   3613    ,   2777    ,   11077   ,   28601   ,0 };
            const unsigned long dim2908Kuo3Init[]   =   {   1   ,   3   ,   7   ,   13  ,   27  ,   47  ,   39  ,   71  ,   179 ,   899 ,   1871    ,   3783    ,   2029    ,   2653    ,   5711    ,0 };
            const unsigned long dim2909Kuo3Init[]   =   {   1   ,   3   ,   3   ,   13  ,   15  ,   13  ,   67  ,   113 ,   321 ,   1017    ,   1615    ,   1575    ,   3879    ,   6113    ,   31893   ,0 };
            const unsigned long dim2910Kuo3Init[]   =   {   1   ,   1   ,   7   ,   3   ,   25  ,   45  ,   3   ,   121 ,   275 ,   313 ,   205 ,   1743    ,   5349    ,   4535    ,   22849   ,0 };
            const unsigned long dim2911Kuo3Init[]   =   {   1   ,   1   ,   3   ,   1   ,   3   ,   19  ,   21  ,   97  ,   455 ,   1021    ,   773 ,   1825    ,   6985    ,   9389    ,   21107   ,0 };
            const unsigned long dim2912Kuo3Init[]   =   {   1   ,   3   ,   5   ,   9   ,   5   ,   3   ,   29  ,   9   ,   261 ,   575 ,   1111    ,   1983    ,   7629    ,   15743   ,   11009   ,0 };
            const unsigned long dim2913Kuo3Init[]   =   {   1   ,   3   ,   1   ,   3   ,   25  ,   25  ,   79  ,   181 ,   417 ,   109 ,   29  ,   777 ,   4753    ,   6131    ,   4939    ,0 };
            const unsigned long dim2914Kuo3Init[]   =   {   1   ,   1   ,   3   ,   1   ,   1   ,   1   ,   73  ,   175 ,   377 ,   531 ,   1829    ,   37  ,   497 ,   12055   ,   24991   ,0 };
            const unsigned long dim2915Kuo3Init[]   =   {   1   ,   1   ,   5   ,   9   ,   27  ,   7   ,   63  ,   115 ,   381 ,   525 ,   1183    ,   1035    ,   5115    ,   2943    ,   2947    ,0 };
            const unsigned long dim2916Kuo3Init[]   =   {   1   ,   3   ,   7   ,   11  ,   13  ,   17  ,   25  ,   203 ,   479 ,   887 ,   1465    ,   249 ,   4347    ,   12201   ,   10365   ,0 };
            const unsigned long dim2917Kuo3Init[]   =   {   1   ,   3   ,   5   ,   13  ,   7   ,   19  ,   11  ,   173 ,   505 ,   509 ,   1311    ,   3535    ,   403 ,   13245   ,   25357   ,0 };
            const unsigned long dim2918Kuo3Init[]   =   {   1   ,   1   ,   5   ,   1   ,   11  ,   53  ,   111 ,   243 ,   509 ,   17  ,   1281    ,   1149    ,   4561    ,   4705    ,   5783    ,0 };
            const unsigned long dim2919Kuo3Init[]   =   {   1   ,   3   ,   7   ,   5   ,   29  ,   7   ,   39  ,   75  ,   385 ,   921 ,   147 ,   3419    ,   237 ,   12609   ,   10103   ,0 };
            const unsigned long dim2920Kuo3Init[]   =   {   1   ,   1   ,   5   ,   3   ,   5   ,   21  ,   75  ,   109 ,   451 ,   371 ,   743 ,   3   ,   5277    ,   10917   ,   2677    ,0 };
            const unsigned long dim2921Kuo3Init[]   =   {   1   ,   3   ,   5   ,   5   ,   21  ,   29  ,   85  ,   79  ,   359 ,   607 ,   191 ,   2481    ,   2089    ,   13441   ,   23575   ,0 };
            const unsigned long dim2922Kuo3Init[]   =   {   1   ,   3   ,   1   ,   11  ,   1   ,   43  ,   35  ,   119 ,   363 ,   477 ,   159 ,   3999    ,   7731    ,   15279   ,   20611   ,0 };
            const unsigned long dim2923Kuo3Init[]   =   {   1   ,   1   ,   5   ,   5   ,   1   ,   9   ,   111 ,   19  ,   191 ,   899 ,   2039    ,   1153    ,   7893    ,   1627    ,   32325   ,0 };
            const unsigned long dim2924Kuo3Init[]   =   {   1   ,   1   ,   1   ,   11  ,   1   ,   1   ,   43  ,   199 ,   329 ,   245 ,   1643    ,   2869    ,   5563    ,   3911    ,   21837   ,0 };
            const unsigned long dim2925Kuo3Init[]   =   {   1   ,   1   ,   5   ,   15  ,   11  ,   15  ,   35  ,   153 ,   489 ,   383 ,   1313    ,   1881    ,   5895    ,   10085   ,   2489    ,0 };
            const unsigned long dim2926Kuo3Init[]   =   {   1   ,   3   ,   1   ,   5   ,   27  ,   29  ,   15  ,   209 ,   27  ,   291 ,   77  ,   1759    ,   1565    ,   5477    ,   18701   ,0 };
            const unsigned long dim2927Kuo3Init[]   =   {   1   ,   3   ,   5   ,   5   ,   19  ,   39  ,   35  ,   207 ,   345 ,   625 ,   1797    ,   2563    ,   8043    ,   1559    ,   18415   ,0 };
            const unsigned long dim2928Kuo3Init[]   =   {   1   ,   3   ,   1   ,   1   ,   21  ,   43  ,   101 ,   89  ,   229 ,   435 ,   1445    ,   2789    ,   6903    ,   11629   ,   29933   ,0 };
            const unsigned long dim2929Kuo3Init[]   =   {   1   ,   1   ,   7   ,   7   ,   21  ,   23  ,   107 ,   21  ,   377 ,   67  ,   1007    ,   2323    ,   2957    ,   6553    ,   3225    ,0 };
            const unsigned long dim2930Kuo3Init[]   =   {   1   ,   1   ,   3   ,   13  ,   3   ,   13  ,   121 ,   119 ,   253 ,   391 ,   215 ,   851 ,   5047    ,   4577    ,   17691   ,0 };
            const unsigned long dim2931Kuo3Init[]   =   {   1   ,   3   ,   3   ,   3   ,   3   ,   37  ,   65  ,   103 ,   395 ,   745 ,   1897    ,   3031    ,   2321    ,   5731    ,   4657    ,0 };
            const unsigned long dim2932Kuo3Init[]   =   {   1   ,   3   ,   1   ,   7   ,   17  ,   19  ,   61  ,   159 ,   459 ,   207 ,   397 ,   1101    ,   5835    ,   1947    ,   28479   ,0 };
            const unsigned long dim2933Kuo3Init[]   =   {   1   ,   1   ,   7   ,   7   ,   23  ,   13  ,   63  ,   7   ,   231 ,   15  ,   855 ,   375 ,   673 ,   12523   ,   20981   ,0 };
            const unsigned long dim2934Kuo3Init[]   =   {   1   ,   1   ,   7   ,   3   ,   21  ,   13  ,   109 ,   93  ,   7   ,   767 ,   477 ,   2301    ,   3587    ,   6053    ,   427 ,0 };
            const unsigned long dim2935Kuo3Init[]   =   {   1   ,   1   ,   5   ,   5   ,   23  ,   9   ,   97  ,   175 ,   139 ,   983 ,   993 ,   2447    ,   2183    ,   16143   ,   10059   ,0 };
            const unsigned long dim2936Kuo3Init[]   =   {   1   ,   1   ,   1   ,   1   ,   17  ,   61  ,   9   ,   13  ,   241 ,   825 ,   707 ,   3385    ,   6551    ,   7321    ,   1825    ,0 };
            const unsigned long dim2937Kuo3Init[]   =   {   1   ,   1   ,   5   ,   1   ,   9   ,   63  ,   101 ,   33  ,   151 ,   629 ,   1903    ,   3159    ,   1083    ,   11109   ,   27679   ,0 };
            const unsigned long dim2938Kuo3Init[]   =   {   1   ,   3   ,   1   ,   13  ,   11  ,   57  ,   123 ,   191 ,   41  ,   765 ,   1039    ,   2639    ,   3793    ,   1403    ,   20745   ,0 };
            const unsigned long dim2939Kuo3Init[]   =   {   1   ,   3   ,   7   ,   13  ,   23  ,   45  ,   21  ,   39  ,   259 ,   531 ,   1205    ,   3537    ,   1927    ,   4357    ,   16185   ,0 };
            const unsigned long dim2940Kuo3Init[]   =   {   1   ,   1   ,   3   ,   1   ,   5   ,   59  ,   127 ,   187 ,   67  ,   377 ,   1577    ,   401 ,   4523    ,   8433    ,   20965   ,0 };
            const unsigned long dim2941Kuo3Init[]   =   {   1   ,   3   ,   3   ,   7   ,   9   ,   37  ,   9   ,   221 ,   459 ,   301 ,   1937    ,   283 ,   6973    ,   3545    ,   203 ,0 };
            const unsigned long dim2942Kuo3Init[]   =   {   1   ,   1   ,   3   ,   5   ,   27  ,   29  ,   59  ,   47  ,   11  ,   791 ,   1863    ,   1725    ,   2797    ,   2673    ,   9551    ,0 };
            const unsigned long dim2943Kuo3Init[]   =   {   1   ,   1   ,   5   ,   3   ,   9   ,   17  ,   75  ,   167 ,   31  ,   879 ,   1841    ,   801 ,   5659    ,   10199   ,   8639    ,0 };
            const unsigned long dim2944Kuo3Init[]   =   {   1   ,   3   ,   3   ,   13  ,   27  ,   35  ,   99  ,   97  ,   153 ,   987 ,   1847    ,   2209    ,   8139    ,   647 ,   30087   ,0 };
            const unsigned long dim2945Kuo3Init[]   =   {   1   ,   1   ,   5   ,   5   ,   3   ,   9   ,   85  ,   105 ,   503 ,   279 ,   1919    ,   1407    ,   8161    ,   3601    ,   26729   ,0 };
            const unsigned long dim2946Kuo3Init[]   =   {   1   ,   3   ,   7   ,   1   ,   23  ,   37  ,   71  ,   131 ,   315 ,   401 ,   389 ,   547 ,   7305    ,   581 ,   14619   ,0 };
            const unsigned long dim2947Kuo3Init[]   =   {   1   ,   1   ,   7   ,   5   ,   3   ,   47  ,   125 ,   87  ,   255 ,   397 ,   1561    ,   433 ,   7057    ,   4595    ,   17785   ,0 };
            const unsigned long dim2948Kuo3Init[]   =   {   1   ,   1   ,   5   ,   7   ,   17  ,   59  ,   107 ,   141 ,   437 ,   335 ,   397 ,   3913    ,   3777    ,   8625    ,   25539   ,0 };
            const unsigned long dim2949Kuo3Init[]   =   {   1   ,   1   ,   5   ,   1   ,   9   ,   21  ,   43  ,   13  ,   75  ,   667 ,   287 ,   347 ,   1503    ,   15905   ,   17281   ,0 };
            const unsigned long dim2950Kuo3Init[]   =   {   1   ,   1   ,   1   ,   5   ,   7   ,   1   ,   25  ,   91  ,   337 ,   909 ,   1555    ,   1719    ,   1169    ,   8537    ,   8859    ,0 };
            const unsigned long dim2951Kuo3Init[]   =   {   1   ,   1   ,   3   ,   15  ,   11  ,   17  ,   3   ,   67  ,   305 ,   565 ,   2023    ,   1175    ,   1135    ,   4453    ,   21431   ,0 };
            const unsigned long dim2952Kuo3Init[]   =   {   1   ,   3   ,   1   ,   9   ,   5   ,   17  ,   23  ,   41  ,   317 ,   63  ,   775 ,   3699    ,   3141    ,   12917   ,   1847    ,0 };
            const unsigned long dim2953Kuo3Init[]   =   {   1   ,   1   ,   5   ,   5   ,   13  ,   41  ,   27  ,   229 ,   113 ,   223 ,   1411    ,   2733    ,   7769    ,   831 ,   20447   ,0 };
            const unsigned long dim2954Kuo3Init[]   =   {   1   ,   1   ,   3   ,   13  ,   9   ,   15  ,   9   ,   235 ,   501 ,   709 ,   643 ,   79  ,   2163    ,   10511   ,   8763    ,0 };
            const unsigned long dim2955Kuo3Init[]   =   {   1   ,   3   ,   3   ,   7   ,   13  ,   3   ,   61  ,   173 ,   507 ,   619 ,   1169    ,   3133    ,   533 ,   7989    ,   14071   ,0 };
            const unsigned long dim2956Kuo3Init[]   =   {   1   ,   1   ,   5   ,   11  ,   9   ,   11  ,   63  ,   147 ,   73  ,   501 ,   371 ,   489 ,   5689    ,   3289    ,   23459   ,0 };
            const unsigned long dim2957Kuo3Init[]   =   {   1   ,   1   ,   1   ,   1   ,   23  ,   21  ,   83  ,   127 ,   507 ,   395 ,   445 ,   4021    ,   7401    ,   14939   ,   781 ,0 };
            const unsigned long dim2958Kuo3Init[]   =   {   1   ,   1   ,   5   ,   11  ,   21  ,   15  ,   57  ,   9   ,   223 ,   603 ,   271 ,   2277    ,   7509    ,   5183    ,   9645    ,0 };
            const unsigned long dim2959Kuo3Init[]   =   {   1   ,   3   ,   1   ,   5   ,   3   ,   37  ,   59  ,   75  ,   325 ,   481 ,   1503    ,   2891    ,   3333    ,   5251    ,   24987   ,0 };
            const unsigned long dim2960Kuo3Init[]   =   {   1   ,   1   ,   7   ,   5   ,   25  ,   3   ,   95  ,   55  ,   437 ,   809 ,   1459    ,   957 ,   4037    ,   4139    ,   22307   ,0 };
            const unsigned long dim2961Kuo3Init[]   =   {   1   ,   3   ,   5   ,   5   ,   7   ,   43  ,   111 ,   183 ,   321 ,   475 ,   633 ,   3183    ,   2885    ,   2083    ,   2515    ,0 };
            const unsigned long dim2962Kuo3Init[]   =   {   1   ,   1   ,   3   ,   3   ,   1   ,   1   ,   9   ,   159 ,   257 ,   257 ,   431 ,   3433    ,   6191    ,   15379   ,   29129   ,0 };
            const unsigned long dim2963Kuo3Init[]   =   {   1   ,   1   ,   3   ,   9   ,   19  ,   7   ,   69  ,   119 ,   391 ,   381 ,   1355    ,   877 ,   245 ,   5287    ,   451 ,0 };
            const unsigned long dim2964Kuo3Init[]   =   {   1   ,   3   ,   7   ,   3   ,   15  ,   27  ,   47  ,   19  ,   213 ,   213 ,   1435    ,   3835    ,   3639    ,   16037   ,   32613   ,0 };
            const unsigned long dim2965Kuo3Init[]   =   {   1   ,   3   ,   5   ,   15  ,   19  ,   9   ,   115 ,   1   ,   319 ,   893 ,   29  ,   3363    ,   2491    ,   8091    ,   29797   ,0 };
            const unsigned long dim2966Kuo3Init[]   =   {   1   ,   3   ,   3   ,   11  ,   19  ,   9   ,   111 ,   181 ,   37  ,   819 ,   843 ,   3591    ,   3617    ,   11209   ,   1731    ,0 };
            const unsigned long dim2967Kuo3Init[]   =   {   1   ,   3   ,   5   ,   7   ,   17  ,   31  ,   107 ,   147 ,   469 ,   409 ,   1391    ,   927 ,   4483    ,   15859   ,   11231   ,0 };
            const unsigned long dim2968Kuo3Init[]   =   {   1   ,   3   ,   7   ,   15  ,   17  ,   37  ,   37  ,   165 ,   289 ,   857 ,   243 ,   3803    ,   8131    ,   7409    ,   12197   ,0 };
            const unsigned long dim2969Kuo3Init[]   =   {   1   ,   1   ,   1   ,   9   ,   17  ,   61  ,   75  ,   177 ,   63  ,   835 ,   731 ,   2015    ,   7287    ,   11803   ,   11595   ,0 };
            const unsigned long dim2970Kuo3Init[]   =   {   1   ,   3   ,   1   ,   15  ,   17  ,   43  ,   33  ,   151 ,   259 ,   951 ,   73  ,   2229    ,   4851    ,   7875    ,   21497   ,0 };
            const unsigned long dim2971Kuo3Init[]   =   {   1   ,   3   ,   7   ,   11  ,   25  ,   5   ,   91  ,   15  ,   231 ,   729 ,   865 ,   545 ,   3043    ,   8925    ,   9111    ,0 };
            const unsigned long dim2972Kuo3Init[]   =   {   1   ,   3   ,   5   ,   3   ,   27  ,   13  ,   67  ,   175 ,   135 ,   999 ,   997 ,   2237    ,   617 ,   14805   ,   31347   ,0 };
            const unsigned long dim2973Kuo3Init[]   =   {   1   ,   3   ,   1   ,   7   ,   5   ,   37  ,   59  ,   155 ,   11  ,   401 ,   421 ,   305 ,   4727    ,   7087    ,   27231   ,0 };
            const unsigned long dim2974Kuo3Init[]   =   {   1   ,   3   ,   1   ,   7   ,   7   ,   39  ,   105 ,   215 ,   241 ,   959 ,   1065    ,   2045    ,   5807    ,   1219    ,   12545   ,0 };
            const unsigned long dim2975Kuo3Init[]   =   {   1   ,   1   ,   3   ,   5   ,   3   ,   61  ,   3   ,   141 ,   363 ,   677 ,   1149    ,   1143    ,   1697    ,   11339   ,   10503   ,0 };
            const unsigned long dim2976Kuo3Init[]   =   {   1   ,   1   ,   5   ,   5   ,   15  ,   1   ,   37  ,   183 ,   279 ,   885 ,   571 ,   1537    ,   5489    ,   12111   ,   4867    ,0 };
            const unsigned long dim2977Kuo3Init[]   =   {   1   ,   1   ,   3   ,   15  ,   7   ,   49  ,   49  ,   37  ,   323 ,   407 ,   1555    ,   1209    ,   8159    ,   12171   ,   4457    ,0 };
            const unsigned long dim2978Kuo3Init[]   =   {   1   ,   3   ,   3   ,   9   ,   9   ,   17  ,   111 ,   19  ,   353 ,   273 ,   1271    ,   975 ,   3935    ,   9689    ,   18355   ,0 };
            const unsigned long dim2979Kuo3Init[]   =   {   1   ,   3   ,   1   ,   7   ,   17  ,   37  ,   21  ,   127 ,   115 ,   19  ,   771 ,   667 ,   4219    ,   1829    ,   18899   ,0 };
            const unsigned long dim2980Kuo3Init[]   =   {   1   ,   3   ,   3   ,   7   ,   3   ,   43  ,   67  ,   43  ,   119 ,   319 ,   1165    ,   2759    ,   5389    ,   10891   ,   16529   ,0 };
            const unsigned long dim2981Kuo3Init[]   =   {   1   ,   3   ,   3   ,   11  ,   29  ,   17  ,   53  ,   137 ,   329 ,   71  ,   1499    ,   3381    ,   3533    ,   15551   ,   30561   ,0 };
            const unsigned long dim2982Kuo3Init[]   =   {   1   ,   3   ,   3   ,   3   ,   5   ,   61  ,   45  ,   109 ,   389 ,   505 ,   1803    ,   1845    ,   2165    ,   267 ,   14573   ,0 };
            const unsigned long dim2983Kuo3Init[]   =   {   1   ,   3   ,   7   ,   5   ,   25  ,   55  ,   47  ,   17  ,   421 ,   711 ,   613 ,   1537    ,   5255    ,   13437   ,   3425    ,0 };
            const unsigned long dim2984Kuo3Init[]   =   {   1   ,   1   ,   3   ,   7   ,   25  ,   51  ,   25  ,   67  ,   409 ,   309 ,   951 ,   131 ,   2781    ,   7445    ,   1487    ,0 };
            const unsigned long dim2985Kuo3Init[]   =   {   1   ,   3   ,   7   ,   9   ,   27  ,   21  ,   45  ,   73  ,   465 ,   977 ,   1327    ,   837 ,   1311    ,   11167   ,   21265   ,0 };
            const unsigned long dim2986Kuo3Init[]   =   {   1   ,   1   ,   3   ,   11  ,   17  ,   61  ,   27  ,   83  ,   271 ,   245 ,   187 ,   1947    ,   7989    ,   13633   ,   27023   ,0 };
            const unsigned long dim2987Kuo3Init[]   =   {   1   ,   3   ,   1   ,   9   ,   17  ,   7   ,   35  ,   195 ,   465 ,   745 ,   1833    ,   3231    ,   3061    ,   1631    ,   9081    ,0 };
            const unsigned long dim2988Kuo3Init[]   =   {   1   ,   3   ,   3   ,   1   ,   5   ,   39  ,   83  ,   13  ,   199 ,   821 ,   839 ,   3509    ,   3857    ,   12389   ,   28771   ,0 };
            const unsigned long dim2989Kuo3Init[]   =   {   1   ,   3   ,   3   ,   3   ,   5   ,   15  ,   91  ,   199 ,   239 ,   97  ,   1699    ,   549 ,   4735    ,   15127   ,   7407    ,0 };
            const unsigned long dim2990Kuo3Init[]   =   {   1   ,   3   ,   7   ,   9   ,   13  ,   1   ,   7   ,   109 ,   93  ,   733 ,   1221    ,   3361    ,   7765    ,   11317   ,   431 ,0 };
            const unsigned long dim2991Kuo3Init[]   =   {   1   ,   3   ,   7   ,   15  ,   31  ,   23  ,   17  ,   97  ,   477 ,   497 ,   1123    ,   2731    ,   3157    ,   5217    ,   23901   ,0 };
            const unsigned long dim2992Kuo3Init[]   =   {   1   ,   1   ,   5   ,   1   ,   1   ,   41  ,   63  ,   11  ,   507 ,   55  ,   1927    ,   2193    ,   6177    ,   3013    ,   24095   ,0 };
            const unsigned long dim2993Kuo3Init[]   =   {   1   ,   1   ,   7   ,   3   ,   25  ,   57  ,   113 ,   99  ,   383 ,   831 ,   1455    ,   2603    ,   4619    ,   9495    ,   5073    ,0 };
            const unsigned long dim2994Kuo3Init[]   =   {   1   ,   1   ,   5   ,   11  ,   25  ,   35  ,   67  ,   83  ,   125 ,   909 ,   227 ,   1337    ,   3729    ,   7623    ,   27899   ,0 };
            const unsigned long dim2995Kuo3Init[]   =   {   1   ,   1   ,   1   ,   13  ,   15  ,   23  ,   117 ,   73  ,   285 ,   341 ,   241 ,   3569    ,   7401    ,   16185   ,   18571   ,0 };
            const unsigned long dim2996Kuo3Init[]   =   {   1   ,   1   ,   7   ,   11  ,   23  ,   9   ,   123 ,   93  ,   287 ,   881 ,   1857    ,   195 ,   3999    ,   81  ,   26173   ,0 };
            const unsigned long dim2997Kuo3Init[]   =   {   1   ,   1   ,   3   ,   9   ,   21  ,   27  ,   57  ,   251 ,   321 ,   109 ,   811 ,   3665    ,   4575    ,   15971   ,   24111   ,0 };
            const unsigned long dim2998Kuo3Init[]   =   {   1   ,   1   ,   7   ,   11  ,   11  ,   53  ,   71  ,   11  ,   67  ,   433 ,   549 ,   3121    ,   193 ,   13099   ,   6635    ,0 };
            const unsigned long dim2999Kuo3Init[]   =   {   1   ,   3   ,   1   ,   9   ,   1   ,   39  ,   75  ,   163 ,   259 ,   523 ,   941 ,   3025    ,   3757    ,   2589    ,   15403   ,0 };
            const unsigned long dim3000Kuo3Init[]   =   {   1   ,   3   ,   5   ,   1   ,   1   ,   15  ,   47  ,   83  ,   33  ,   881 ,   215 ,   1717    ,   2255    ,   7349    ,   27345   ,0 };
            const unsigned long dim3001Kuo3Init[]   =   {   1   ,   1   ,   3   ,   11  ,   15  ,   55  ,   95  ,   125 ,   511 ,   829 ,   1021    ,   2095    ,   1107    ,   13071   ,   29279   ,0 };
            const unsigned long dim3002Kuo3Init[]   =   {   1   ,   1   ,   5   ,   11  ,   15  ,   63  ,   15  ,   123 ,   233 ,   351 ,   1687    ,   3383    ,   287 ,   2721    ,   20109   ,0 };
            const unsigned long dim3003Kuo3Init[]   =   {   1   ,   1   ,   1   ,   7   ,   7   ,   19  ,   103 ,   29  ,   349 ,   875 ,   1631    ,   119 ,   6095    ,   7623    ,   4063    ,0 };
            const unsigned long dim3004Kuo3Init[]   =   {   1   ,   3   ,   5   ,   5   ,   21  ,   41  ,   45  ,   91  ,   445 ,   193 ,   1077    ,   3849    ,   163 ,   15123   ,   9677    ,0 };
            const unsigned long dim3005Kuo3Init[]   =   {   1   ,   1   ,   3   ,   1   ,   7   ,   39  ,   127 ,   75  ,   53  ,   233 ,   531 ,   635 ,   8055    ,   8789    ,   6205    ,0 };
            const unsigned long dim3006Kuo3Init[]   =   {   1   ,   3   ,   7   ,   5   ,   21  ,   19  ,   53  ,   197 ,   207 ,   791 ,   1193    ,   365 ,   5747    ,   12155   ,   30361   ,0 };
            const unsigned long dim3007Kuo3Init[]   =   {   1   ,   1   ,   1   ,   15  ,   23  ,   39  ,   95  ,   75  ,   239 ,   639 ,   253 ,   3329    ,   6017    ,   6647    ,   16381   ,0 };
            const unsigned long dim3008Kuo3Init[]   =   {   1   ,   3   ,   3   ,   7   ,   5   ,   37  ,   85  ,   45  ,   473 ,   299 ,   767 ,   3795    ,   6895    ,   16255   ,   20411   ,0 };
            const unsigned long dim3009Kuo3Init[]   =   {   1   ,   1   ,   5   ,   15  ,   31  ,   33  ,   83  ,   79  ,   341 ,   11  ,   561 ,   2051    ,   485 ,   8653    ,   7751    ,0 };
            const unsigned long dim3010Kuo3Init[]   =   {   1   ,   1   ,   1   ,   1   ,   9   ,   39  ,   127 ,   115 ,   55  ,   281 ,   1439    ,   2527    ,   3975    ,   9037    ,   32221   ,0 };
            const unsigned long dim3011Kuo3Init[]   =   {   1   ,   1   ,   7   ,   5   ,   27  ,   45  ,   117 ,   67  ,   471 ,   101 ,   927 ,   2391    ,   4935    ,   14477   ,   17515   ,0 };
            const unsigned long dim3012Kuo3Init[]   =   {   1   ,   3   ,   1   ,   1   ,   5   ,   1   ,   45  ,   93  ,   169 ,   181 ,   609 ,   3489    ,   4779    ,   11641   ,   16535   ,0 };
            const unsigned long dim3013Kuo3Init[]   =   {   1   ,   3   ,   3   ,   5   ,   1   ,   53  ,   21  ,   169 ,   199 ,   769 ,   813 ,   1799    ,   5199    ,   13499   ,   25677   ,0 };
            const unsigned long dim3014Kuo3Init[]   =   {   1   ,   1   ,   3   ,   7   ,   13  ,   43  ,   127 ,   169 ,   503 ,   935 ,   1315    ,   1449    ,   1497    ,   6173    ,   20437   ,0 };
            const unsigned long dim3015Kuo3Init[]   =   {   1   ,   1   ,   5   ,   3   ,   19  ,   31  ,   83  ,   167 ,   23  ,   1019    ,   1223    ,   2705    ,   6471    ,   11619   ,   10589   ,0 };
            const unsigned long dim3016Kuo3Init[]   =   {   1   ,   1   ,   5   ,   3   ,   7   ,   39  ,   117 ,   67  ,   159 ,   301 ,   553 ,   2005    ,   5455    ,   10049   ,   3441    ,0 };
            const unsigned long dim3017Kuo3Init[]   =   {   1   ,   3   ,   5   ,   5   ,   7   ,   45  ,   59  ,   73  ,   105 ,   507 ,   387 ,   2697    ,   5679    ,   9825    ,   7689    ,0 };
            const unsigned long dim3018Kuo3Init[]   =   {   1   ,   1   ,   7   ,   13  ,   3   ,   35  ,   93  ,   129 ,   273 ,   165 ,   1615    ,   3631    ,   1391    ,   10631   ,   27157   ,0 };
            const unsigned long dim3019Kuo3Init[]   =   {   1   ,   1   ,   5   ,   5   ,   9   ,   25  ,   23  ,   101 ,   1   ,   741 ,   1539    ,   1893    ,   1573    ,   10037   ,   3079    ,0 };
            const unsigned long dim3020Kuo3Init[]   =   {   1   ,   1   ,   1   ,   13  ,   29  ,   63  ,   119 ,   83  ,   239 ,   993 ,   1663    ,   391 ,   1583    ,   11393   ,   8361    ,0 };
            const unsigned long dim3021Kuo3Init[]   =   {   1   ,   1   ,   5   ,   5   ,   11  ,   55  ,   41  ,   155 ,   439 ,   631 ,   1949    ,   1529    ,   5249    ,   883 ,   25125   ,0 };
            const unsigned long dim3022Kuo3Init[]   =   {   1   ,   3   ,   1   ,   1   ,   19  ,   15  ,   35  ,   175 ,   33  ,   199 ,   1979    ,   4025    ,   1543    ,   6451    ,   11485   ,0 };
            const unsigned long dim3023Kuo3Init[]   =   {   1   ,   3   ,   1   ,   1   ,   17  ,   63  ,   127 ,   191 ,   463 ,   973 ,   1509    ,   1925    ,   3469    ,   15967   ,   10661   ,0 };
            const unsigned long dim3024Kuo3Init[]   =   {   1   ,   1   ,   1   ,   11  ,   7   ,   57  ,   111 ,   127 ,   475 ,   625 ,   553 ,   1941    ,   2923    ,   5203    ,   16915   ,0 };
            const unsigned long dim3025Kuo3Init[]   =   {   1   ,   1   ,   1   ,   5   ,   15  ,   23  ,   61  ,   93  ,   251 ,   865 ,   2011    ,   2607    ,   1691    ,   10359   ,   4667    ,0 };
            const unsigned long dim3026Kuo3Init[]   =   {   1   ,   3   ,   1   ,   13  ,   29  ,   19  ,   5   ,   101 ,   205 ,   515 ,   1953    ,   3543    ,   929 ,   8745    ,   1773    ,0 };
            const unsigned long dim3027Kuo3Init[]   =   {   1   ,   1   ,   7   ,   15  ,   19  ,   29  ,   127 ,   153 ,   383 ,   799 ,   639 ,   3271    ,   6487    ,   2989    ,   8965    ,0 };
            const unsigned long dim3028Kuo3Init[]   =   {   1   ,   1   ,   5   ,   15  ,   7   ,   7   ,   73  ,   239 ,   117 ,   261 ,   1147    ,   3761    ,   1531    ,   15613   ,   13545   ,0 };
            const unsigned long dim3029Kuo3Init[]   =   {   1   ,   3   ,   1   ,   11  ,   17  ,   29  ,   53  ,   161 ,   47  ,   933 ,   1763    ,   1543    ,   3661    ,   9369    ,   14239   ,0 };
            const unsigned long dim3030Kuo3Init[]   =   {   1   ,   1   ,   5   ,   9   ,   13  ,   47  ,   19  ,   139 ,   115 ,   239 ,   799 ,   2211    ,   6205    ,   5273    ,   31205   ,0 };
            const unsigned long dim3031Kuo3Init[]   =   {   1   ,   1   ,   5   ,   9   ,   29  ,   53  ,   121 ,   3   ,   89  ,   253 ,   1201    ,   89  ,   297 ,   11561   ,   9655    ,0 };
            const unsigned long dim3032Kuo3Init[]   =   {   1   ,   1   ,   7   ,   5   ,   25  ,   43  ,   25  ,   165 ,   135 ,   789 ,   891 ,   1711    ,   711 ,   269 ,   17895   ,0 };
            const unsigned long dim3033Kuo3Init[]   =   {   1   ,   3   ,   1   ,   3   ,   27  ,   57  ,   63  ,   89  ,   33  ,   897 ,   691 ,   3129    ,   7253    ,   2041    ,   14341   ,0 };
            const unsigned long dim3034Kuo3Init[]   =   {   1   ,   1   ,   1   ,   1   ,   5   ,   25  ,   21  ,   65  ,   383 ,   687 ,   57  ,   2267    ,   1837    ,   7091    ,   5941    ,0 };
            const unsigned long dim3035Kuo3Init[]   =   {   1   ,   3   ,   7   ,   13  ,   9   ,   37  ,   75  ,   113 ,   243 ,   419 ,   387 ,   3479    ,   3295    ,   3659    ,   29215   ,0 };
            const unsigned long dim3036Kuo3Init[]   =   {   1   ,   1   ,   1   ,   13  ,   1   ,   39  ,   97  ,   237 ,   509 ,   51  ,   1239    ,   1889    ,   915 ,   15313   ,   3487    ,0 };
            const unsigned long dim3037Kuo3Init[]   =   {   1   ,   1   ,   3   ,   1   ,   1   ,   5   ,   81  ,   153 ,   83  ,   977 ,   1667    ,   571 ,   1161    ,   9347    ,   32353   ,0 };
            const unsigned long dim3038Kuo3Init[]   =   {   1   ,   1   ,   7   ,   5   ,   11  ,   13  ,   45  ,   91  ,   265 ,   815 ,   1141    ,   1187    ,   41  ,   15723   ,   5573    ,0 };
            const unsigned long dim3039Kuo3Init[]   =   {   1   ,   1   ,   7   ,   13  ,   31  ,   29  ,   27  ,   23  ,   163 ,   605 ,   1457    ,   37  ,   2075    ,   4581    ,   10571   ,0 };
            const unsigned long dim3040Kuo3Init[]   =   {   1   ,   1   ,   1   ,   11  ,   13  ,   31  ,   119 ,   241 ,   379 ,   427 ,   333 ,   1299    ,   3303    ,   13049   ,   15233   ,0 };
            const unsigned long dim3041Kuo3Init[]   =   {   1   ,   3   ,   7   ,   13  ,   23  ,   35  ,   69  ,   195 ,   149 ,   273 ,   41  ,   289 ,   3173    ,   13177   ,   10865   ,0 };
            const unsigned long dim3042Kuo3Init[]   =   {   1   ,   3   ,   3   ,   7   ,   1   ,   5   ,   85  ,   245 ,   417 ,   887 ,   1821    ,   3623    ,   1591    ,   15871   ,   10837   ,0 };
            const unsigned long dim3043Kuo3Init[]   =   {   1   ,   1   ,   3   ,   13  ,   19  ,   55  ,   11  ,   65  ,   227 ,   273 ,   1065    ,   1937    ,   5197    ,   7695    ,   24827   ,0 };
            const unsigned long dim3044Kuo3Init[]   =   {   1   ,   1   ,   3   ,   15  ,   11  ,   55  ,   11  ,   49  ,   141 ,   607 ,   509 ,   3785    ,   8009    ,   4127    ,   4757    ,0 };
            const unsigned long dim3045Kuo3Init[]   =   {   1   ,   1   ,   3   ,   11  ,   31  ,   23  ,   33  ,   241 ,   347 ,   261 ,   1609    ,   3177    ,   4065    ,   13767   ,   24607   ,0 };
            const unsigned long dim3046Kuo3Init[]   =   {   1   ,   3   ,   7   ,   13  ,   13  ,   7   ,   73  ,   63  ,   493 ,   607 ,   869 ,   1599    ,   7725    ,   2405    ,   12945   ,0 };
            const unsigned long dim3047Kuo3Init[]   =   {   1   ,   3   ,   1   ,   9   ,   9   ,   57  ,   85  ,   173 ,   259 ,   749 ,   1661    ,   2911    ,   7833    ,   7577    ,   21209   ,0 };
            const unsigned long dim3048Kuo3Init[]   =   {   1   ,   3   ,   5   ,   7   ,   7   ,   27  ,   61  ,   91  ,   417 ,   369 ,   1181    ,   1385    ,   2991    ,   12503   ,   13073   ,0 };
            const unsigned long dim3049Kuo3Init[]   =   {   1   ,   3   ,   5   ,   13  ,   17  ,   11  ,   51  ,   101 ,   45  ,   45  ,   1247    ,   4069    ,   3079    ,   9081    ,   12443   ,0 };
            const unsigned long dim3050Kuo3Init[]   =   {   1   ,   1   ,   3   ,   7   ,   13  ,   57  ,   63  ,   43  ,   451 ,   743 ,   1045    ,   2967    ,   6437    ,   93  ,   4589    ,0 };
            const unsigned long dim3051Kuo3Init[]   =   {   1   ,   3   ,   7   ,   3   ,   25  ,   17  ,   113 ,   89  ,   457 ,   261 ,   827 ,   197 ,   7177    ,   4135    ,   19733   ,0 };
            const unsigned long dim3052Kuo3Init[]   =   {   1   ,   3   ,   5   ,   9   ,   23  ,   59  ,   29  ,   207 ,   19  ,   623 ,   101 ,   1675    ,   579 ,   3255    ,   16259   ,0 };
            const unsigned long dim3053Kuo3Init[]   =   {   1   ,   1   ,   1   ,   1   ,   21  ,   23  ,   103 ,   123 ,   61  ,   233 ,   71  ,   897 ,   6147    ,   3335    ,   1027    ,0 };
            const unsigned long dim3054Kuo3Init[]   =   {   1   ,   3   ,   5   ,   13  ,   9   ,   11  ,   101 ,   55  ,   247 ,   285 ,   1387    ,   2063    ,   6853    ,   409 ,   10715   ,0 };
            const unsigned long dim3055Kuo3Init[]   =   {   1   ,   1   ,   7   ,   7   ,   11  ,   41  ,   5   ,   157 ,   325 ,   975 ,   1403    ,   3187    ,   4107    ,   4093    ,   18165   ,0 };
            const unsigned long dim3056Kuo3Init[]   =   {   1   ,   3   ,   1   ,   3   ,   25  ,   15  ,   69  ,   93  ,   87  ,   711 ,   1699    ,   733 ,   7703    ,   2191    ,   12293   ,0 };
            const unsigned long dim3057Kuo3Init[]   =   {   1   ,   3   ,   3   ,   1   ,   21  ,   41  ,   31  ,   229 ,   325 ,   507 ,   2019    ,   2225    ,   1733    ,   4447    ,   32477   ,0 };
            const unsigned long dim3058Kuo3Init[]   =   {   1   ,   3   ,   1   ,   5   ,   23  ,   41  ,   19  ,   1   ,   467 ,   315 ,   1513    ,   1671    ,   4159    ,   12477   ,   6199    ,0 };
            const unsigned long dim3059Kuo3Init[]   =   {   1   ,   1   ,   1   ,   11  ,   27  ,   61  ,   35  ,   115 ,   215 ,   979 ,   1887    ,   1913    ,   689 ,   3847    ,   11451   ,0 };
            const unsigned long dim3060Kuo3Init[]   =   {   1   ,   3   ,   3   ,   7   ,   25  ,   29  ,   65  ,   233 ,   433 ,   113 ,   1031    ,   2455    ,   6743    ,   5227    ,   7207    ,0 };
            const unsigned long dim3061Kuo3Init[]   =   {   1   ,   1   ,   3   ,   9   ,   5   ,   9   ,   85  ,   103 ,   185 ,   699 ,   663 ,   441 ,   8013    ,   1227    ,   12875   ,0 };
            const unsigned long dim3062Kuo3Init[]   =   {   1   ,   3   ,   5   ,   11  ,   9   ,   59  ,   35  ,   223 ,   131 ,   409 ,   125 ,   3437    ,   2015    ,   217 ,   16351   ,0 };
            const unsigned long dim3063Kuo3Init[]   =   {   1   ,   3   ,   7   ,   3   ,   9   ,   27  ,   43  ,   211 ,   365 ,   855 ,   1249    ,   3127    ,   3611    ,   9339    ,   16083   ,0 };
            const unsigned long dim3064Kuo3Init[]   =   {   1   ,   1   ,   1   ,   9   ,   29  ,   11  ,   61  ,   27  ,   195 ,   963 ,   1905    ,   1679    ,   1811    ,   7   ,   27243   ,0 };
            const unsigned long dim3065Kuo3Init[]   =   {   1   ,   3   ,   7   ,   15  ,   3   ,   41  ,   21  ,   31  ,   461 ,   3   ,   1033    ,   459 ,   1831    ,   12289   ,   19645   ,0 };
            const unsigned long dim3066Kuo3Init[]   =   {   1   ,   1   ,   7   ,   7   ,   15  ,   7   ,   25  ,   207 ,   37  ,   177 ,   973 ,   3207    ,   5329    ,   8541    ,   7929    ,0 };
            const unsigned long dim3067Kuo3Init[]   =   {   1   ,   1   ,   5   ,   1   ,   29  ,   35  ,   77  ,   251 ,   429 ,   681 ,   391 ,   3789    ,   5953    ,   13339   ,   16991   ,0 };
            const unsigned long dim3068Kuo3Init[]   =   {   1   ,   1   ,   5   ,   5   ,   5   ,   39  ,   21  ,   131 ,   511 ,   11  ,   921 ,   3073    ,   3239    ,   561 ,   28595   ,0 };
            const unsigned long dim3069Kuo3Init[]   =   {   1   ,   3   ,   1   ,   11  ,   3   ,   11  ,   71  ,   231 ,   439 ,   523 ,   209 ,   3213    ,   211 ,   13449   ,   23311   ,0 };
            const unsigned long dim3070Kuo3Init[]   =   {   1   ,   1   ,   1   ,   5   ,   31  ,   27  ,   59  ,   161 ,   71  ,   619 ,   1643    ,   3687    ,   753 ,   7443    ,   25743   ,0 };
            const unsigned long dim3071Kuo3Init[]   =   {   1   ,   3   ,   7   ,   11  ,   13  ,   49  ,   33  ,   223 ,   443 ,   463 ,   1833    ,   3757    ,   299 ,   9633    ,   16339   ,0 };
            const unsigned long dim3072Kuo3Init[]   =   {   1   ,   3   ,   7   ,   1   ,   21  ,   57  ,   63  ,   31  ,   491 ,   53  ,   75  ,   3059    ,   3013    ,   965 ,   6209    ,0 };
            const unsigned long dim3073Kuo3Init[]   =   {   1   ,   1   ,   5   ,   9   ,   3   ,   45  ,   63  ,   25  ,   511 ,   501 ,   1003    ,   1499    ,   887 ,   5583    ,   6329    ,0 };
            const unsigned long dim3074Kuo3Init[]   =   {   1   ,   3   ,   1   ,   1   ,   9   ,   11  ,   49  ,   225 ,   511 ,   563 ,   495 ,   1697    ,   7943    ,   13561   ,   16159   ,0 };
            const unsigned long dim3075Kuo3Init[]   =   {   1   ,   3   ,   7   ,   9   ,   11  ,   1   ,   95  ,   255 ,   437 ,   979 ,   1669    ,   179 ,   2189    ,   14431   ,   23919   ,0 };
            const unsigned long dim3076Kuo3Init[]   =   {   1   ,   3   ,   1   ,   13  ,   13  ,   5   ,   23  ,   233 ,   327 ,   903 ,   167 ,   797 ,   5205    ,   2203    ,   22137   ,0 };
            const unsigned long dim3077Kuo3Init[]   =   {   1   ,   1   ,   7   ,   11  ,   25  ,   63  ,   43  ,   113 ,   251 ,   39  ,   163 ,   1631    ,   2731    ,   12241   ,   32567   ,0 };
            const unsigned long dim3078Kuo3Init[]   =   {   1   ,   1   ,   5   ,   11  ,   3   ,   35  ,   9   ,   21  ,   477 ,   139 ,   1107    ,   1975    ,   6883    ,   6651    ,   20847   ,0 };
            const unsigned long dim3079Kuo3Init[]   =   {   1   ,   3   ,   5   ,   13  ,   25  ,   39  ,   89  ,   217 ,   95  ,   457 ,   1741    ,   2725    ,   5391    ,   13381   ,   2411    ,0 };
            const unsigned long dim3080Kuo3Init[]   =   {   1   ,   3   ,   5   ,   5   ,   1   ,   27  ,   9   ,   51  ,   3   ,   915 ,   1803    ,   931 ,   7651    ,   6497    ,   6635    ,0 };
            const unsigned long dim3081Kuo3Init[]   =   {   1   ,   3   ,   7   ,   9   ,   21  ,   53  ,   9   ,   151 ,   155 ,   827 ,   453 ,   4071    ,   3153    ,   2711    ,   25143   ,0 };
            const unsigned long dim3082Kuo3Init[]   =   {   1   ,   1   ,   3   ,   7   ,   21  ,   15  ,   5   ,   169 ,   27  ,   479 ,   911 ,   3949    ,   2819    ,   387 ,   2919    ,0 };
            const unsigned long dim3083Kuo3Init[]   =   {   1   ,   1   ,   3   ,   3   ,   27  ,   59  ,   27  ,   225 ,   507 ,   403 ,   1333    ,   747 ,   7731    ,   13007   ,   16283   ,0 };
            const unsigned long dim3084Kuo3Init[]   =   {   1   ,   1   ,   3   ,   13  ,   29  ,   11  ,   65  ,   169 ,   197 ,   209 ,   1703    ,   3503    ,   2661    ,   14703   ,   28529   ,0 };
            const unsigned long dim3085Kuo3Init[]   =   {   1   ,   1   ,   3   ,   13  ,   17  ,   63  ,   73  ,   11  ,   495 ,   603 ,   687 ,   2773    ,   3955    ,   12761   ,   13899   ,0 };
            const unsigned long dim3086Kuo3Init[]   =   {   1   ,   1   ,   1   ,   11  ,   7   ,   31  ,   55  ,   133 ,   197 ,   619 ,   1747    ,   3531    ,   1157    ,   7143    ,   19347   ,0 };
            const unsigned long dim3087Kuo3Init[]   =   {   1   ,   3   ,   1   ,   13  ,   15  ,   47  ,   5   ,   117 ,   329 ,   169 ,   2023    ,   901 ,   397 ,   15681   ,   8931    ,0 };
            const unsigned long dim3088Kuo3Init[]   =   {   1   ,   1   ,   5   ,   1   ,   29  ,   5   ,   79  ,   73  ,   463 ,   125 ,   1539    ,   2149    ,   4353    ,   15123   ,   27853   ,0 };
            const unsigned long dim3089Kuo3Init[]   =   {   1   ,   1   ,   5   ,   7   ,   13  ,   55  ,   59  ,   151 ,   105 ,   79  ,   1023    ,   1839    ,   3889    ,   3843    ,   6129    ,0 };
            const unsigned long dim3090Kuo3Init[]   =   {   1   ,   1   ,   3   ,   15  ,   9   ,   15  ,   109 ,   63  ,   95  ,   573 ,   385 ,   3773    ,   535 ,   11307   ,   4719    ,0 };
            const unsigned long dim3091Kuo3Init[]   =   {   1   ,   3   ,   7   ,   11  ,   9   ,   9   ,   91  ,   245 ,   403 ,   841 ,   1771    ,   2121    ,   6927    ,   1029    ,   26767   ,0 };
            const unsigned long dim3092Kuo3Init[]   =   {   1   ,   3   ,   1   ,   11  ,   25  ,   55  ,   21  ,   173 ,   327 ,   327 ,   1747    ,   1771    ,   6829    ,   12155   ,   29375   ,0 };
            const unsigned long dim3093Kuo3Init[]   =   {   1   ,   3   ,   7   ,   11  ,   17  ,   15  ,   31  ,   189 ,   179 ,   849 ,   1895    ,   3333    ,   7385    ,   4473    ,   18003   ,0 };
            const unsigned long dim3094Kuo3Init[]   =   {   1   ,   1   ,   7   ,   3   ,   29  ,   37  ,   11  ,   81  ,   181 ,   197 ,   1231    ,   167 ,   6611    ,   3437    ,   8857    ,0 };
            const unsigned long dim3095Kuo3Init[]   =   {   1   ,   3   ,   3   ,   15  ,   17  ,   53  ,   63  ,   51  ,   497 ,   725 ,   917 ,   2661    ,   2013    ,   10887   ,   29145   ,0 };
            const unsigned long dim3096Kuo3Init[]   =   {   1   ,   1   ,   3   ,   11  ,   15  ,   53  ,   75  ,   209 ,   429 ,   927 ,   1867    ,   677 ,   5305    ,   15901   ,   5079    ,0 };
            const unsigned long dim3097Kuo3Init[]   =   {   1   ,   1   ,   1   ,   5   ,   19  ,   57  ,   105 ,   169 ,   113 ,   701 ,   1123    ,   2203    ,   285 ,   12447   ,   18301   ,0 };
            const unsigned long dim3098Kuo3Init[]   =   {   1   ,   3   ,   1   ,   7   ,   25  ,   53  ,   51  ,   35  ,   245 ,   173 ,   1047    ,   1905    ,   1773    ,   10699   ,   32057   ,0 };
            const unsigned long dim3099Kuo3Init[]   =   {   1   ,   1   ,   1   ,   7   ,   31  ,   33  ,   15  ,   233 ,   105 ,   943 ,   201 ,   2681    ,   3121    ,   14145   ,   23741   ,0 };
            const unsigned long dim3100Kuo3Init[]   =   {   1   ,   1   ,   1   ,   1   ,   27  ,   61  ,   121 ,   199 ,   437 ,   717 ,   1481    ,   2421    ,   6681    ,   2457    ,   4411    ,0 };
            const unsigned long dim3101Kuo3Init[]   =   {   1   ,   1   ,   1   ,   13  ,   31  ,   9   ,   93  ,   149 ,   379 ,   639 ,   1889    ,   1385    ,   3513    ,   10105   ,   13933   ,0 };
            const unsigned long dim3102Kuo3Init[]   =   {   1   ,   1   ,   7   ,   7   ,   9   ,   37  ,   41  ,   221 ,   323 ,   989 ,   445 ,   3083    ,   6855    ,   11577   ,   29849   ,0 };
            const unsigned long dim3103Kuo3Init[]   =   {   1   ,   1   ,   7   ,   7   ,   27  ,   55  ,   53  ,   171 ,   87  ,   529 ,   1555    ,   2065    ,   1685    ,   1953    ,   20561   ,0 };
            const unsigned long dim3104Kuo3Init[]   =   {   1   ,   3   ,   7   ,   13  ,   1   ,   15  ,   11  ,   91  ,   81  ,   217 ,   7   ,   1687    ,   493 ,   13345   ,   14009   ,0 };
            const unsigned long dim3105Kuo3Init[]   =   {   1   ,   3   ,   1   ,   9   ,   5   ,   31  ,   89  ,   33  ,   39  ,   747 ,   753 ,   2193    ,   1937    ,   9685    ,   17689   ,0 };
            const unsigned long dim3106Kuo3Init[]   =   {   1   ,   1   ,   3   ,   7   ,   13  ,   61  ,   101 ,   89  ,   305 ,   123 ,   1773    ,   411 ,   595 ,   10113   ,   4725    ,0 };
            const unsigned long dim3107Kuo3Init[]   =   {   1   ,   1   ,   3   ,   15  ,   3   ,   13  ,   25  ,   205 ,   343 ,   869 ,   115 ,   2413    ,   4531    ,   13185   ,   15699   ,0 };
            const unsigned long dim3108Kuo3Init[]   =   {   1   ,   3   ,   5   ,   3   ,   25  ,   37  ,   71  ,   125 ,   85  ,   995 ,   769 ,   2737    ,   2495    ,   7571    ,   26825   ,0 };
            const unsigned long dim3109Kuo3Init[]   =   {   1   ,   3   ,   3   ,   11  ,   21  ,   3   ,   109 ,   175 ,   479 ,   605 ,   569 ,   2753    ,   2927    ,   10529   ,   28345   ,0 };
            const unsigned long dim3110Kuo3Init[]   =   {   1   ,   3   ,   7   ,   1   ,   5   ,   57  ,   75  ,   79  ,   383 ,   637 ,   1411    ,   3955    ,   5535    ,   9403    ,   10463   ,0 };
            const unsigned long dim3111Kuo3Init[]   =   {   1   ,   1   ,   7   ,   5   ,   27  ,   25  ,   67  ,   35  ,   95  ,   25  ,   1585    ,   1725    ,   7003    ,   2261    ,   3575    ,0 };
            const unsigned long dim3112Kuo3Init[]   =   {   1   ,   3   ,   7   ,   13  ,   11  ,   11  ,   85  ,   141 ,   389 ,   903 ,   995 ,   2625    ,   7163    ,   7229    ,   18205   ,0 };
            const unsigned long dim3113Kuo3Init[]   =   {   1   ,   1   ,   3   ,   5   ,   3   ,   53  ,   69  ,   141 ,   213 ,   675 ,   1799    ,   235 ,   3433    ,   2305    ,   29607   ,0 };
            const unsigned long dim3114Kuo3Init[]   =   {   1   ,   1   ,   7   ,   5   ,   21  ,   45  ,   119 ,   229 ,   493 ,   691 ,   1551    ,   3801    ,   663 ,   9693    ,   5255    ,0 };
            const unsigned long dim3115Kuo3Init[]   =   {   1   ,   3   ,   7   ,   11  ,   11  ,   37  ,   69  ,   65  ,   163 ,   61  ,   1567    ,   1879    ,   1853    ,   11099   ,   4523    ,0 };
            const unsigned long dim3116Kuo3Init[]   =   {   1   ,   1   ,   3   ,   9   ,   5   ,   43  ,   75  ,   93  ,   81  ,   555 ,   803 ,   129 ,   4391    ,   3025    ,   1013    ,0 };
            const unsigned long dim3117Kuo3Init[]   =   {   1   ,   1   ,   5   ,   5   ,   9   ,   49  ,   69  ,   185 ,   441 ,   567 ,   203 ,   2641    ,   4169    ,   15585   ,   10963   ,0 };
            const unsigned long dim3118Kuo3Init[]   =   {   1   ,   3   ,   7   ,   9   ,   9   ,   35  ,   121 ,   117 ,   509 ,   93  ,   1381    ,   1329    ,   751 ,   2351    ,   11079   ,0 };
            const unsigned long dim3119Kuo3Init[]   =   {   1   ,   3   ,   5   ,   15  ,   5   ,   61  ,   117 ,   83  ,   383 ,   961 ,   779 ,   1499    ,   4695    ,   3679    ,   20383   ,0 };
            const unsigned long dim3120Kuo3Init[]   =   {   1   ,   3   ,   5   ,   5   ,   21  ,   63  ,   105 ,   15  ,   297 ,   923 ,   1867    ,   1667    ,   3523    ,   12387   ,   14977   ,0 };
            const unsigned long dim3121Kuo3Init[]   =   {   1   ,   1   ,   5   ,   5   ,   7   ,   37  ,   123 ,   161 ,   343 ,   959 ,   1807    ,   1695    ,   6121    ,   8415    ,   5373    ,0 };
            const unsigned long dim3122Kuo3Init[]   =   {   1   ,   3   ,   5   ,   15  ,   29  ,   51  ,   107 ,   151 ,   5   ,   715 ,   3   ,   3005    ,   1917    ,   12243   ,   13903   ,0 };
            const unsigned long dim3123Kuo3Init[]   =   {   1   ,   3   ,   5   ,   11  ,   25  ,   29  ,   125 ,   165 ,   57  ,   729 ,   1187    ,   2273    ,   2869    ,   13149   ,   2831    ,0 };
            const unsigned long dim3124Kuo3Init[]   =   {   1   ,   1   ,   7   ,   15  ,   1   ,   15  ,   37  ,   15  ,   489 ,   879 ,   89  ,   2399    ,   547 ,   1959    ,   20073   ,0 };
            const unsigned long dim3125Kuo3Init[]   =   {   1   ,   3   ,   5   ,   13  ,   5   ,   43  ,   101 ,   27  ,   425 ,   531 ,   1681    ,   3301    ,   5741    ,   7005    ,   6607    ,0 };
            const unsigned long dim3126Kuo3Init[]   =   {   1   ,   1   ,   3   ,   13  ,   17  ,   53  ,   61  ,   165 ,   375 ,   7   ,   761 ,   1891    ,   1253    ,   2431    ,   15273   ,0 };
            const unsigned long dim3127Kuo3Init[]   =   {   1   ,   1   ,   1   ,   5   ,   13  ,   23  ,   27  ,   229 ,   215 ,   729 ,   9   ,   779 ,   3583    ,   9109    ,   19059   ,0 };
            const unsigned long dim3128Kuo3Init[]   =   {   1   ,   1   ,   3   ,   9   ,   3   ,   45  ,   99  ,   193 ,   203 ,   57  ,   147 ,   3953    ,   5263    ,   14641   ,   24729   ,0 };
            const unsigned long dim3129Kuo3Init[]   =   {   1   ,   3   ,   7   ,   3   ,   11  ,   49  ,   57  ,   121 ,   413 ,   293 ,   1251    ,   3791    ,   1803    ,   12549   ,   3743    ,0 };
            const unsigned long dim3130Kuo3Init[]   =   {   1   ,   3   ,   7   ,   13  ,   19  ,   55  ,   85  ,   1   ,   169 ,   539 ,   1565    ,   2157    ,   4493    ,   9169    ,   27583   ,0 };
            const unsigned long dim3131Kuo3Init[]   =   {   1   ,   1   ,   7   ,   7   ,   19  ,   59  ,   95  ,   171 ,   471 ,   881 ,   663 ,   1489    ,   8013    ,   8981    ,   11443   ,0 };
            const unsigned long dim3132Kuo3Init[]   =   {   1   ,   1   ,   3   ,   7   ,   17  ,   59  ,   25  ,   115 ,   439 ,   967 ,   1939    ,   905 ,   4141    ,   4177    ,   24827   ,0 };
            const unsigned long dim3133Kuo3Init[]   =   {   1   ,   3   ,   5   ,   3   ,   29  ,   17  ,   23  ,   241 ,   49  ,   815 ,   1409    ,   3997    ,   1073    ,   2331    ,   455 ,0 };
            const unsigned long dim3134Kuo3Init[]   =   {   1   ,   1   ,   5   ,   13  ,   27  ,   53  ,   51  ,   67  ,   405 ,   853 ,   1229    ,   2647    ,   4817    ,   521 ,   14343   ,0 };
            const unsigned long dim3135Kuo3Init[]   =   {   1   ,   3   ,   3   ,   13  ,   1   ,   55  ,   83  ,   15  ,   41  ,   327 ,   1501    ,   1967    ,   5787    ,   3557    ,   7461    ,0 };
            const unsigned long dim3136Kuo3Init[]   =   {   1   ,   3   ,   5   ,   1   ,   23  ,   43  ,   93  ,   179 ,   373 ,   173 ,   1625    ,   487 ,   6973    ,   12479   ,   24659   ,0 };
            const unsigned long dim3137Kuo3Init[]   =   {   1   ,   3   ,   7   ,   15  ,   3   ,   27  ,   9   ,   69  ,   501 ,   287 ,   945 ,   681 ,   3667    ,   7499    ,   24259   ,0 };
            const unsigned long dim3138Kuo3Init[]   =   {   1   ,   3   ,   7   ,   3   ,   21  ,   55  ,   55  ,   13  ,   191 ,   25  ,   1339    ,   185 ,   4345    ,   13649   ,   17793   ,0 };
            const unsigned long dim3139Kuo3Init[]   =   {   1   ,   3   ,   3   ,   1   ,   9   ,   13  ,   95  ,   127 ,   305 ,   961 ,   141 ,   3071    ,   683 ,   2141    ,   13757   ,0 };
            const unsigned long dim3140Kuo3Init[]   =   {   1   ,   1   ,   5   ,   5   ,   21  ,   59  ,   83  ,   39  ,   137 ,   259 ,   689 ,   339 ,   105 ,   147 ,   26143   ,0 };
            const unsigned long dim3141Kuo3Init[]   =   {   1   ,   1   ,   5   ,   7   ,   31  ,   21  ,   121 ,   45  ,   103 ,   507 ,   1243    ,   3325    ,   441 ,   6743    ,   31937   ,0 };
            const unsigned long dim3142Kuo3Init[]   =   {   1   ,   1   ,   5   ,   3   ,   7   ,   35  ,   5   ,   117 ,   433 ,   47  ,   261 ,   2405    ,   3361    ,   4361    ,   14397   ,0 };
            const unsigned long dim3143Kuo3Init[]   =   {   1   ,   3   ,   7   ,   9   ,   19  ,   7   ,   19  ,   129 ,   79  ,   247 ,   35  ,   1483    ,   285 ,   4923    ,   14255   ,0 };
            const unsigned long dim3144Kuo3Init[]   =   {   1   ,   3   ,   3   ,   11  ,   13  ,   3   ,   117 ,   169 ,   3   ,   527 ,   1295    ,   295 ,   2317    ,   5391    ,   27073   ,0 };
            const unsigned long dim3145Kuo3Init[]   =   {   1   ,   1   ,   7   ,   3   ,   3   ,   53  ,   9   ,   175 ,   157 ,   917 ,   1297    ,   1917    ,   5575    ,   16183   ,   20165   ,0 };
            const unsigned long dim3146Kuo3Init[]   =   {   1   ,   3   ,   3   ,   9   ,   25  ,   47  ,   61  ,   71  ,   129 ,   923 ,   367 ,   1623    ,   4281    ,   11241   ,   14759   ,0 };
            const unsigned long dim3147Kuo3Init[]   =   {   1   ,   3   ,   7   ,   11  ,   23  ,   29  ,   53  ,   135 ,   381 ,   943 ,   1611    ,   1609    ,   5897    ,   393 ,   24647   ,0 };
            const unsigned long dim3148Kuo3Init[]   =   {   1   ,   1   ,   7   ,   1   ,   3   ,   55  ,   5   ,   191 ,   305 ,   385 ,   1383    ,   913 ,   1017    ,   6463    ,   32323   ,0 };
            const unsigned long dim3149Kuo3Init[]   =   {   1   ,   1   ,   5   ,   9   ,   3   ,   57  ,   113 ,   189 ,   41  ,   413 ,   1685    ,   2145    ,   6071    ,   7229    ,   11413   ,0 };
            const unsigned long dim3150Kuo3Init[]   =   {   1   ,   3   ,   3   ,   9   ,   21  ,   5   ,   95  ,   87  ,   247 ,   175 ,   1265    ,   83  ,   511 ,   2821    ,   9881    ,0 };
            const unsigned long dim3151Kuo3Init[]   =   {   1   ,   3   ,   1   ,   3   ,   31  ,   41  ,   127 ,   115 ,   457 ,   809 ,   2017    ,   3979    ,   3425    ,   159 ,   15903   ,0 };
            const unsigned long dim3152Kuo3Init[]   =   {   1   ,   1   ,   7   ,   7   ,   29  ,   9   ,   85  ,   109 ,   299 ,   607 ,   1991    ,   3717    ,   4593    ,   8529    ,   10759   ,0 };
            const unsigned long dim3153Kuo3Init[]   =   {   1   ,   3   ,   1   ,   9   ,   15  ,   49  ,   123 ,   61  ,   49  ,   413 ,   1599    ,   3697    ,   2481    ,   739 ,   1641    ,0 };
            const unsigned long dim3154Kuo3Init[]   =   {   1   ,   3   ,   3   ,   1   ,   13  ,   55  ,   35  ,   217 ,   79  ,   167 ,   1215    ,   2687    ,   7303    ,   10331   ,   29985   ,0 };
            const unsigned long dim3155Kuo3Init[]   =   {   1   ,   1   ,   1   ,   9   ,   29  ,   29  ,   37  ,   177 ,   63  ,   515 ,   1551    ,   1053    ,   893 ,   6831    ,   18039   ,0 };
            const unsigned long dim3156Kuo3Init[]   =   {   1   ,   1   ,   7   ,   5   ,   25  ,   47  ,   57  ,   209 ,   345 ,   805 ,   999 ,   307 ,   3465    ,   13411   ,   17721   ,0 };
            const unsigned long dim3157Kuo3Init[]   =   {   1   ,   1   ,   5   ,   11  ,   3   ,   63  ,   95  ,   243 ,   163 ,   923 ,   1181    ,   3319    ,   5903    ,   8195    ,   21217   ,0 };
            const unsigned long dim3158Kuo3Init[]   =   {   1   ,   3   ,   7   ,   11  ,   15  ,   51  ,   75  ,   237 ,   25  ,   907 ,   395 ,   3721    ,   3371    ,   12425   ,   3795    ,0 };
            const unsigned long dim3159Kuo3Init[]   =   {   1   ,   1   ,   1   ,   11  ,   15  ,   61  ,   127 ,   39  ,   359 ,   381 ,   1879    ,   2605    ,   7511    ,   4329    ,   28251   ,0 };
            const unsigned long dim3160Kuo3Init[]   =   {   1   ,   3   ,   5   ,   9   ,   11  ,   61  ,   1   ,   215 ,   503 ,   939 ,   419 ,   3353    ,   3013    ,   5391    ,   12567   ,0 };
            const unsigned long dim3161Kuo3Init[]   =   {   1   ,   3   ,   1   ,   9   ,   7   ,   57  ,   123 ,   5   ,   473 ,   931 ,   759 ,   571 ,   2407    ,   13057   ,   7443    ,0 };
            const unsigned long dim3162Kuo3Init[]   =   {   1   ,   1   ,   3   ,   7   ,   31  ,   17  ,   55  ,   181 ,   359 ,   913 ,   271 ,   935 ,   5911    ,   9767    ,   23  ,0 };
            const unsigned long dim3163Kuo3Init[]   =   {   1   ,   1   ,   1   ,   15  ,   1   ,   35  ,   93  ,   233 ,   109 ,   135 ,   1793    ,   2453    ,   6547    ,   15435   ,   16227   ,0 };
            const unsigned long dim3164Kuo3Init[]   =   {   1   ,   3   ,   1   ,   13  ,   5   ,   37  ,   87  ,   51  ,   165 ,   455 ,   621 ,   881 ,   2377    ,   7977    ,   29763   ,0 };
            const unsigned long dim3165Kuo3Init[]   =   {   1   ,   1   ,   3   ,   13  ,   31  ,   3   ,   119 ,   113 ,   141 ,   179 ,   621 ,   677 ,   1047    ,   2405    ,   28777   ,0 };
            const unsigned long dim3166Kuo3Init[]   =   {   1   ,   3   ,   1   ,   1   ,   11  ,   25  ,   37  ,   15  ,   117 ,   509 ,   853 ,   747 ,   3065    ,   11569   ,   11731   ,0 };
            const unsigned long dim3167Kuo3Init[]   =   {   1   ,   1   ,   5   ,   11  ,   31  ,   9   ,   27  ,   219 ,   71  ,   919 ,   1553    ,   3067    ,   6315    ,   17  ,   18511   ,0 };
            const unsigned long dim3168Kuo3Init[]   =   {   1   ,   1   ,   7   ,   1   ,   27  ,   39  ,   111 ,   127 ,   315 ,   671 ,   845 ,   179 ,   8129    ,   15409   ,   21565   ,0 };
            const unsigned long dim3169Kuo3Init[]   =   {   1   ,   1   ,   5   ,   13  ,   31  ,   37  ,   105 ,   67  ,   37  ,   725 ,   1947    ,   2743    ,   6241    ,   5465    ,   15025   ,0 };
            const unsigned long dim3170Kuo3Init[]   =   {   1   ,   1   ,   7   ,   13  ,   7   ,   63  ,   63  ,   209 ,   431 ,   171 ,   845 ,   2067    ,   4829    ,   5627    ,   32333   ,0 };
            const unsigned long dim3171Kuo3Init[]   =   {   1   ,   3   ,   7   ,   15  ,   7   ,   61  ,   105 ,   35  ,   113 ,   551 ,   1227    ,   867 ,   5907    ,   4389    ,   13281   ,0 };
            const unsigned long dim3172Kuo3Init[]   =   {   1   ,   1   ,   1   ,   13  ,   31  ,   13  ,   123 ,   255 ,   497 ,   365 ,   921 ,   2009    ,   5365    ,   15443   ,   23455   ,0 };
            const unsigned long dim3173Kuo3Init[]   =   {   1   ,   1   ,   5   ,   15  ,   5   ,   45  ,   91  ,   199 ,   417 ,   557 ,   827 ,   1911    ,   4173    ,   8705    ,   19437   ,0 };
            const unsigned long dim3174Kuo3Init[]   =   {   1   ,   3   ,   3   ,   9   ,   31  ,   51  ,   69  ,   247 ,   407 ,   791 ,   1811    ,   1127    ,   6775    ,   8343    ,   19723   ,0 };
            const unsigned long dim3175Kuo3Init[]   =   {   1   ,   3   ,   3   ,   3   ,   21  ,   43  ,   71  ,   191 ,   339 ,   219 ,   1809    ,   3735    ,   151 ,   3743    ,   11803   ,0 };
            const unsigned long dim3176Kuo3Init[]   =   {   1   ,   1   ,   7   ,   3   ,   13  ,   21  ,   1   ,   233 ,   395 ,   335 ,   1929    ,   547 ,   2939    ,   9205    ,   22309   ,0 };
            const unsigned long dim3177Kuo3Init[]   =   {   1   ,   1   ,   1   ,   13  ,   7   ,   5   ,   97  ,   189 ,   195 ,   191 ,   833 ,   269 ,   3255    ,   15301   ,   11593   ,0 };
            const unsigned long dim3178Kuo3Init[]   =   {   1   ,   1   ,   3   ,   1   ,   31  ,   35  ,   53  ,   193 ,   143 ,   241 ,   763 ,   3221    ,   3217    ,   14915   ,   5603    ,0 };
            const unsigned long dim3179Kuo3Init[]   =   {   1   ,   3   ,   5   ,   15  ,   3   ,   15  ,   47  ,   171 ,   349 ,   995 ,   1803    ,   2123    ,   8009    ,   10339   ,   1827    ,0 };
            const unsigned long dim3180Kuo3Init[]   =   {   1   ,   1   ,   1   ,   3   ,   11  ,   37  ,   125 ,   149 ,   261 ,   507 ,   1603    ,   683 ,   137 ,   353 ,   26313   ,0 };
            const unsigned long dim3181Kuo3Init[]   =   {   1   ,   3   ,   3   ,   11  ,   5   ,   19  ,   7   ,   97  ,   257 ,   439 ,   825 ,   265 ,   2493    ,   5567    ,   8575    ,0 };
            const unsigned long dim3182Kuo3Init[]   =   {   1   ,   1   ,   5   ,   5   ,   9   ,   25  ,   101 ,   207 ,   345 ,   823 ,   1787    ,   2449    ,   4377    ,   10581   ,   19169   ,0 };
            const unsigned long dim3183Kuo3Init[]   =   {   1   ,   3   ,   1   ,   9   ,   19  ,   5   ,   15  ,   19  ,   405 ,   665 ,   715 ,   3743    ,   23  ,   5061    ,   21145   ,0 };
            const unsigned long dim3184Kuo3Init[]   =   {   1   ,   1   ,   3   ,   9   ,   11  ,   7   ,   73  ,   39  ,   77  ,   763 ,   1251    ,   3533    ,   6379    ,   16101   ,   8719    ,0 };
            const unsigned long dim3185Kuo3Init[]   =   {   1   ,   3   ,   3   ,   1   ,   15  ,   55  ,   89  ,   249 ,   475 ,   641 ,   511 ,   111 ,   5797    ,   5627    ,   2187    ,0 };
            const unsigned long dim3186Kuo3Init[]   =   {   1   ,   1   ,   5   ,   5   ,   7   ,   27  ,   117 ,   153 ,   387 ,   295 ,   309 ,   3293    ,   5197    ,   8359    ,   16643   ,0 };
            const unsigned long dim3187Kuo3Init[]   =   {   1   ,   3   ,   3   ,   3   ,   5   ,   35  ,   45  ,   159 ,   495 ,   311 ,   463 ,   1249    ,   5163    ,   7939    ,   21405   ,0 };
            const unsigned long dim3188Kuo3Init[]   =   {   1   ,   3   ,   3   ,   7   ,   31  ,   39  ,   125 ,   167 ,   43  ,   41  ,   589 ,   3329    ,   5511    ,   7983    ,   17329   ,0 };
            const unsigned long dim3189Kuo3Init[]   =   {   1   ,   1   ,   3   ,   13  ,   1   ,   17  ,   43  ,   95  ,   465 ,   587 ,   1189    ,   2629    ,   7613    ,   15115   ,   22299   ,0 };
            const unsigned long dim3190Kuo3Init[]   =   {   1   ,   3   ,   1   ,   9   ,   21  ,   57  ,   73  ,   241 ,   149 ,   531 ,   827 ,   731 ,   7431    ,   14375   ,   16307   ,0 };
            const unsigned long dim3191Kuo3Init[]   =   {   1   ,   1   ,   1   ,   7   ,   7   ,   25  ,   113 ,   115 ,   381 ,   951 ,   811 ,   305 ,   935 ,   2927    ,   18755   ,0 };
            const unsigned long dim3192Kuo3Init[]   =   {   1   ,   3   ,   5   ,   13  ,   11  ,   53  ,   91  ,   197 ,   497 ,   9   ,   1009    ,   25  ,   2473    ,   2467    ,   8827    ,0 };
            const unsigned long dim3193Kuo3Init[]   =   {   1   ,   3   ,   3   ,   7   ,   23  ,   23  ,   15  ,   3   ,   131 ,   591 ,   379 ,   1227    ,   3479    ,   8409    ,   20177   ,0 };
            const unsigned long dim3194Kuo3Init[]   =   {   1   ,   3   ,   5   ,   9   ,   17  ,   47  ,   91  ,   107 ,   17  ,   175 ,   1181    ,   2097    ,   457 ,   10253   ,   24291   ,0 };
            const unsigned long dim3195Kuo3Init[]   =   {   1   ,   1   ,   3   ,   5   ,   31  ,   57  ,   25  ,   117 ,   97  ,   707 ,   1875    ,   4059    ,   6997    ,   3533    ,   27781   ,0 };
            const unsigned long dim3196Kuo3Init[]   =   {   1   ,   1   ,   7   ,   11  ,   7   ,   7   ,   53  ,   137 ,   255 ,   1017    ,   501 ,   847 ,   4569    ,   3665    ,   21377   ,0 };
            const unsigned long dim3197Kuo3Init[]   =   {   1   ,   1   ,   3   ,   9   ,   19  ,   57  ,   39  ,   213 ,   67  ,   221 ,   149 ,   357 ,   1137    ,   12625   ,   13945   ,0 };
            const unsigned long dim3198Kuo3Init[]   =   {   1   ,   1   ,   7   ,   11  ,   19  ,   53  ,   75  ,   67  ,   71  ,   811 ,   837 ,   3399    ,   3993    ,   3119    ,   13459   ,0 };
            const unsigned long dim3199Kuo3Init[]   =   {   1   ,   1   ,   1   ,   3   ,   29  ,   7   ,   77  ,   249 ,   97  ,   159 ,   1097    ,   161 ,   1649    ,   4663    ,   8635    ,0 };
            const unsigned long dim3200Kuo3Init[]   =   {   1   ,   1   ,   5   ,   9   ,   9   ,   11  ,   69  ,   151 ,   419 ,   11  ,   1457    ,   2357    ,   7259    ,   4059    ,   6189    ,0 };
            const unsigned long dim3201Kuo3Init[]   =   {   1   ,   3   ,   5   ,   7   ,   29  ,   1   ,   109 ,   103 ,   129 ,   561 ,   1629    ,   237 ,   643 ,   14519   ,   22839   ,0 };
            const unsigned long dim3202Kuo3Init[]   =   {   1   ,   3   ,   3   ,   15  ,   17  ,   25  ,   109 ,   213 ,   253 ,   427 ,   849 ,   4053    ,   3845    ,   8935    ,   21233   ,0 };
            const unsigned long dim3203Kuo3Init[]   =   {   1   ,   3   ,   1   ,   3   ,   15  ,   53  ,   81  ,   127 ,   131 ,   523 ,   1303    ,   451 ,   3363    ,   14105   ,   28881   ,0 };
            const unsigned long dim3204Kuo3Init[]   =   {   1   ,   1   ,   1   ,   13  ,   1   ,   29  ,   105 ,   75  ,   235 ,   407 ,   1555    ,   1247    ,   2329    ,   1259    ,   8653    ,0 };
            const unsigned long dim3205Kuo3Init[]   =   {   1   ,   1   ,   1   ,   3   ,   31  ,   45  ,   9   ,   199 ,   315 ,   375 ,   205 ,   3157    ,   4247    ,   11961   ,   4039    ,0 };
            const unsigned long dim3206Kuo3Init[]   =   {   1   ,   3   ,   3   ,   5   ,   3   ,   39  ,   13  ,   125 ,   103 ,   359 ,   287 ,   3529    ,   2015    ,   3187    ,   26261   ,0 };
            const unsigned long dim3207Kuo3Init[]   =   {   1   ,   3   ,   3   ,   7   ,   3   ,   63  ,   81  ,   157 ,   169 ,   1005    ,   889 ,   2271    ,   7773    ,   8739    ,   11171   ,0 };
            const unsigned long dim3208Kuo3Init[]   =   {   1   ,   1   ,   1   ,   3   ,   11  ,   63  ,   61  ,   67  ,   1   ,   919 ,   349 ,   1645    ,   8155    ,   219 ,   357 ,0 };
            const unsigned long dim3209Kuo3Init[]   =   {   1   ,   3   ,   7   ,   9   ,   17  ,   39  ,   1   ,   253 ,   497 ,   409 ,   631 ,   1425    ,   7763    ,   3239    ,   5405    ,0 };
            const unsigned long dim3210Kuo3Init[]   =   {   1   ,   3   ,   3   ,   3   ,   21  ,   57  ,   97  ,   231 ,   445 ,   417 ,   1567    ,   739 ,   6413    ,   13499   ,   16947   ,0 };
            const unsigned long dim3211Kuo3Init[]   =   {   1   ,   1   ,   1   ,   3   ,   13  ,   59  ,   17  ,   15  ,   75  ,   197 ,   267 ,   1295    ,   3867    ,   6607    ,   31623   ,0 };
            const unsigned long dim3212Kuo3Init[]   =   {   1   ,   1   ,   3   ,   5   ,   27  ,   1   ,   83  ,   97  ,   381 ,   931 ,   1133    ,   4091    ,   341 ,   7105    ,   28761   ,0 };
            const unsigned long dim3213Kuo3Init[]   =   {   1   ,   3   ,   3   ,   11  ,   3   ,   39  ,   49  ,   237 ,   437 ,   353 ,   321 ,   3609    ,   4555    ,   8987    ,   28727   ,0 };
            const unsigned long dim3214Kuo3Init[]   =   {   1   ,   3   ,   1   ,   13  ,   15  ,   49  ,   101 ,   255 ,   409 ,   231 ,   355 ,   2333    ,   7277    ,   2549    ,   1613    ,0 };
            const unsigned long dim3215Kuo3Init[]   =   {   1   ,   3   ,   3   ,   13  ,   11  ,   49  ,   99  ,   201 ,   57  ,   705 ,   413 ,   2791    ,   5947    ,   8533    ,   22155   ,0 };
            const unsigned long dim3216Kuo3Init[]   =   {   1   ,   3   ,   1   ,   15  ,   13  ,   51  ,   91  ,   47  ,   165 ,   995 ,   1293    ,   3601    ,   5841    ,   4671    ,   27379   ,0 };
            const unsigned long dim3217Kuo3Init[]   =   {   1   ,   1   ,   3   ,   11  ,   15  ,   47  ,   43  ,   97  ,   151 ,   75  ,   1229    ,   2525    ,   5085    ,   10969   ,   5195    ,0 };
            const unsigned long dim3218Kuo3Init[]   =   {   1   ,   1   ,   5   ,   3   ,   7   ,   15  ,   65  ,   161 ,   423 ,   773 ,   291 ,   1081    ,   2127    ,   15861   ,   16051   ,0 };
            const unsigned long dim3219Kuo3Init[]   =   {   1   ,   3   ,   1   ,   1   ,   11  ,   47  ,   47  ,   157 ,   369 ,   889 ,   221 ,   83  ,   3165    ,   12911   ,   20687   ,0 };
            const unsigned long dim3220Kuo3Init[]   =   {   1   ,   1   ,   1   ,   9   ,   23  ,   63  ,   97  ,   191 ,   185 ,   863 ,   1611    ,   241 ,   4177    ,   14965   ,   22205   ,0 };
            const unsigned long dim3221Kuo3Init[]   =   {   1   ,   3   ,   3   ,   15  ,   13  ,   11  ,   85  ,   109 ,   13  ,   401 ,   1365    ,   2637    ,   2119    ,   14673   ,   28813   ,0 };
            const unsigned long dim3222Kuo3Init[]   =   {   1   ,   1   ,   5   ,   7   ,   23  ,   13  ,   11  ,   9   ,   105 ,   891 ,   1511    ,   3969    ,   2097    ,   2089    ,   18535   ,0 };
            const unsigned long dim3223Kuo3Init[]   =   {   1   ,   1   ,   3   ,   15  ,   3   ,   23  ,   67  ,   197 ,   309 ,   883 ,   1635    ,   359 ,   5023    ,   10693   ,   28537   ,0 };
            const unsigned long dim3224Kuo3Init[]   =   {   1   ,   1   ,   3   ,   1   ,   1   ,   25  ,   63  ,   233 ,   445 ,   665 ,   1971    ,   1713    ,   1827    ,   14269   ,   19539   ,0 };
            const unsigned long dim3225Kuo3Init[]   =   {   1   ,   3   ,   3   ,   13  ,   25  ,   27  ,   79  ,   131 ,   441 ,   465 ,   613 ,   2027    ,   2521    ,   11109   ,   14443   ,0 };
            const unsigned long dim3226Kuo3Init[]   =   {   1   ,   3   ,   5   ,   13  ,   27  ,   1   ,   27  ,   123 ,   405 ,   567 ,   159 ,   2319    ,   5625    ,   4643    ,   1185    ,0 };
            const unsigned long dim3227Kuo3Init[]   =   {   1   ,   1   ,   7   ,   15  ,   15  ,   55  ,   21  ,   105 ,   353 ,   431 ,   955 ,   3341    ,   2891    ,   13667   ,   17787   ,0 };
            const unsigned long dim3228Kuo3Init[]   =   {   1   ,   1   ,   1   ,   5   ,   25  ,   61  ,   11  ,   251 ,   459 ,   985 ,   1285    ,   811 ,   4279    ,   4579    ,   14313   ,0 };
            const unsigned long dim3229Kuo3Init[]   =   {   1   ,   1   ,   3   ,   15  ,   13  ,   1   ,   63  ,   103 ,   113 ,   91  ,   1241    ,   2513    ,   767 ,   9037    ,   29591   ,0 };
            const unsigned long dim3230Kuo3Init[]   =   {   1   ,   3   ,   7   ,   15  ,   31  ,   39  ,   87  ,   215 ,   421 ,   875 ,   913 ,   3593    ,   2215    ,   4889    ,   14595   ,0 };
            const unsigned long dim3231Kuo3Init[]   =   {   1   ,   3   ,   3   ,   3   ,   17  ,   61  ,   99  ,   13  ,   475 ,   331 ,   541 ,   2363    ,   5697    ,   14867   ,   11483   ,0 };
            const unsigned long dim3232Kuo3Init[]   =   {   1   ,   1   ,   5   ,   3   ,   17  ,   15  ,   7   ,   217 ,   219 ,   327 ,   1775    ,   939 ,   3953    ,   2393    ,   28241   ,0 };
            const unsigned long dim3233Kuo3Init[]   =   {   1   ,   1   ,   3   ,   3   ,   11  ,   57  ,   75  ,   139 ,   123 ,   607 ,   87  ,   539 ,   4955    ,   6533    ,   31827   ,0 };
            const unsigned long dim3234Kuo3Init[]   =   {   1   ,   3   ,   3   ,   15  ,   15  ,   15  ,   43  ,   127 ,   485 ,   1009    ,   1201    ,   919 ,   761 ,   385 ,   29889   ,0 };
            const unsigned long dim3235Kuo3Init[]   =   {   1   ,   3   ,   1   ,   3   ,   3   ,   53  ,   47  ,   101 ,   463 ,   899 ,   523 ,   1009    ,   3541    ,   11191   ,   2257    ,0 };
            const unsigned long dim3236Kuo3Init[]   =   {   1   ,   1   ,   1   ,   3   ,   31  ,   43  ,   9   ,   39  ,   201 ,   469 ,   839 ,   2509    ,   5645    ,   2861    ,   25533   ,0 };
            const unsigned long dim3237Kuo3Init[]   =   {   1   ,   3   ,   5   ,   7   ,   19  ,   27  ,   69  ,   45  ,   367 ,   267 ,   953 ,   3711    ,   603 ,   3511    ,   13419   ,0 };
            const unsigned long dim3238Kuo3Init[]   =   {   1   ,   1   ,   3   ,   1   ,   25  ,   39  ,   33  ,   89  ,   459 ,   573 ,   1643    ,   3291    ,   2371    ,   1749    ,   20693   ,0 };
            const unsigned long dim3239Kuo3Init[]   =   {   1   ,   1   ,   1   ,   13  ,   11  ,   29  ,   103 ,   213 ,   13  ,   219 ,   2023    ,   1327    ,   4315    ,   13933   ,   19559   ,0 };
            const unsigned long dim3240Kuo3Init[]   =   {   1   ,   1   ,   5   ,   3   ,   17  ,   49  ,   11  ,   241 ,   227 ,   157 ,   745 ,   1377    ,   5739    ,   14285   ,   26107   ,0 };
            const unsigned long dim3241Kuo3Init[]   =   {   1   ,   1   ,   7   ,   3   ,   23  ,   11  ,   5   ,   213 ,   405 ,   415 ,   1911    ,   3763    ,   3403    ,   16079   ,   10957   ,0 };
            const unsigned long dim3242Kuo3Init[]   =   {   1   ,   3   ,   5   ,   15  ,   31  ,   27  ,   79  ,   153 ,   403 ,   747 ,   269 ,   4039    ,   2791    ,   4711    ,   501 ,0 };
            const unsigned long dim3243Kuo3Init[]   =   {   1   ,   1   ,   7   ,   15  ,   1   ,   29  ,   105 ,   29  ,   131 ,   913 ,   265 ,   871 ,   2899    ,   11547   ,   20351   ,0 };
            const unsigned long dim3244Kuo3Init[]   =   {   1   ,   1   ,   1   ,   3   ,   9   ,   47  ,   55  ,   121 ,   487 ,   349 ,   233 ,   1657    ,   3015    ,   12345   ,   2213    ,0 };
            const unsigned long dim3245Kuo3Init[]   =   {   1   ,   1   ,   7   ,   9   ,   1   ,   45  ,   49  ,   241 ,   81  ,   673 ,   1933    ,   3203    ,   7433    ,   14745   ,   5811    ,0 };
            const unsigned long dim3246Kuo3Init[]   =   {   1   ,   1   ,   7   ,   5   ,   31  ,   31  ,   97  ,   97  ,   377 ,   795 ,   417 ,   3851    ,   4725    ,   9803    ,   16785   ,0 };
            const unsigned long dim3247Kuo3Init[]   =   {   1   ,   3   ,   1   ,   1   ,   9   ,   5   ,   45  ,   141 ,   361 ,   523 ,   297 ,   2535    ,   4433    ,   16353   ,   28191   ,0 };
            const unsigned long dim3248Kuo3Init[]   =   {   1   ,   3   ,   3   ,   7   ,   9   ,   29  ,   47  ,   19  ,   419 ,   49  ,   825 ,   1771    ,   5439    ,   8611    ,   12659   ,0 };
            const unsigned long dim3249Kuo3Init[]   =   {   1   ,   1   ,   1   ,   9   ,   19  ,   3   ,   89  ,   107 ,   407 ,   567 ,   807 ,   623 ,   7915    ,   8083    ,   6761    ,0 };
            const unsigned long dim3250Kuo3Init[]   =   {   1   ,   1   ,   5   ,   3   ,   23  ,   35  ,   109 ,   23  ,   461 ,   401 ,   713 ,   965 ,   7687    ,   9729    ,   17287   ,0 };
            const unsigned long dim3251Kuo3Init[]   =   {   1   ,   3   ,   5   ,   11  ,   11  ,   43  ,   79  ,   197 ,   249 ,   423 ,   953 ,   1621    ,   4045    ,   3913    ,   11143   ,0 };
            const unsigned long dim3252Kuo3Init[]   =   {   1   ,   1   ,   3   ,   7   ,   19  ,   59  ,   91  ,   217 ,   481 ,   861 ,   1915    ,   1689    ,   2061    ,   5377    ,   8535    ,0 };
            const unsigned long dim3253Kuo3Init[]   =   {   1   ,   3   ,   3   ,   9   ,   15  ,   57  ,   5   ,   81  ,   133 ,   833 ,   49  ,   3231    ,   6315    ,   9771    ,   5667    ,0 };
            const unsigned long dim3254Kuo3Init[]   =   {   1   ,   1   ,   1   ,   11  ,   27  ,   15  ,   83  ,   11  ,   483 ,   799 ,   1423    ,   2121    ,   5385    ,   4541    ,   597 ,0 };
            const unsigned long dim3255Kuo3Init[]   =   {   1   ,   3   ,   7   ,   3   ,   19  ,   25  ,   5   ,   111 ,   493 ,   531 ,   317 ,   5   ,   2601    ,   6909    ,   11515   ,0 };
            const unsigned long dim3256Kuo3Init[]   =   {   1   ,   3   ,   3   ,   7   ,   15  ,   33  ,   63  ,   89  ,   425 ,   913 ,   831 ,   3881    ,   2055    ,   4285    ,   13015   ,0 };
            const unsigned long dim3257Kuo3Init[]   =   {   1   ,   3   ,   1   ,   1   ,   11  ,   5   ,   87  ,   255 ,   17  ,   731 ,   1149    ,   115 ,   1223    ,   413 ,   579 ,0 };
            const unsigned long dim3258Kuo3Init[]   =   {   1   ,   1   ,   5   ,   15  ,   11  ,   27  ,   125 ,   97  ,   73  ,   221 ,   687 ,   623 ,   1617    ,   13347   ,   1369    ,0 };
            const unsigned long dim3259Kuo3Init[]   =   {   1   ,   1   ,   5   ,   9   ,   19  ,   19  ,   39  ,   9   ,   273 ,   687 ,   805 ,   1105    ,   267 ,   8321    ,   17899   ,0 };
            const unsigned long dim3260Kuo3Init[]   =   {   1   ,   1   ,   7   ,   7   ,   31  ,   25  ,   5   ,   209 ,   3   ,   503 ,   1757    ,   2071    ,   6251    ,   12821   ,   28241   ,0 };
            const unsigned long dim3261Kuo3Init[]   =   {   1   ,   1   ,   5   ,   3   ,   7   ,   51  ,   25  ,   37  ,   197 ,   143 ,   1261    ,   1069    ,   6559    ,   5625    ,   30631   ,0 };
            const unsigned long dim3262Kuo3Init[]   =   {   1   ,   3   ,   1   ,   5   ,   17  ,   15  ,   53  ,   71  ,   433 ,   133 ,   867 ,   3883    ,   3281    ,   3785    ,   11523   ,0 };
            const unsigned long dim3263Kuo3Init[]   =   {   1   ,   3   ,   1   ,   7   ,   1   ,   31  ,   89  ,   165 ,   239 ,   525 ,   507 ,   3637    ,   4423    ,   1295    ,   9441    ,0 };
            const unsigned long dim3264Kuo3Init[]   =   {   1   ,   1   ,   3   ,   13  ,   7   ,   11  ,   127 ,   151 ,   151 ,   821 ,   381 ,   23  ,   173 ,   14027   ,   8391    ,0 };
            const unsigned long dim3265Kuo3Init[]   =   {   1   ,   3   ,   5   ,   15  ,   1   ,   51  ,   85  ,   67  ,   321 ,   797 ,   1625    ,   475 ,   313 ,   8189    ,   13315   ,0 };
            const unsigned long dim3266Kuo3Init[]   =   {   1   ,   3   ,   5   ,   7   ,   25  ,   49  ,   1   ,   155 ,   497 ,   771 ,   457 ,   2097    ,   5997    ,   11363   ,   14715   ,0 };
            const unsigned long dim3267Kuo3Init[]   =   {   1   ,   1   ,   3   ,   5   ,   15  ,   5   ,   89  ,   219 ,   441 ,   997 ,   127 ,   699 ,   5245    ,   3941    ,   14083   ,0 };
            const unsigned long dim3268Kuo3Init[]   =   {   1   ,   3   ,   5   ,   15  ,   21  ,   13  ,   35  ,   239 ,   115 ,   57  ,   1479    ,   2813    ,   5249    ,   7733    ,   3135    ,0 };
            const unsigned long dim3269Kuo3Init[]   =   {   1   ,   3   ,   7   ,   3   ,   11  ,   59  ,   49  ,   55  ,   183 ,   631 ,   365 ,   323 ,   6891    ,   3727    ,   24231   ,0 };
            const unsigned long dim3270Kuo3Init[]   =   {   1   ,   1   ,   1   ,   13  ,   9   ,   29  ,   101 ,   59  ,   257 ,   995 ,   145 ,   3037    ,   7291    ,   16021   ,   31971   ,0 };
            const unsigned long dim3271Kuo3Init[]   =   {   1   ,   1   ,   1   ,   1   ,   19  ,   5   ,   51  ,   21  ,   207 ,   767 ,   1301    ,   3111    ,   7133    ,   9555    ,   24043   ,0 };
            const unsigned long dim3272Kuo3Init[]   =   {   1   ,   1   ,   3   ,   5   ,   23  ,   45  ,   25  ,   149 ,   225 ,   141 ,   827 ,   37  ,   3767    ,   16305   ,   30229   ,0 };
            const unsigned long dim3273Kuo3Init[]   =   {   1   ,   1   ,   3   ,   9   ,   21  ,   15  ,   51  ,   155 ,   185 ,   551 ,   243 ,   2493    ,   653 ,   10615   ,   28137   ,0 };
            const unsigned long dim3274Kuo3Init[]   =   {   1   ,   1   ,   1   ,   5   ,   13  ,   55  ,   77  ,   71  ,   383 ,   605 ,   957 ,   1605    ,   6905    ,   12193   ,   26729   ,0 };
            const unsigned long dim3275Kuo3Init[]   =   {   1   ,   3   ,   1   ,   15  ,   13  ,   19  ,   95  ,   219 ,   435 ,   165 ,   1531    ,   2733    ,   4213    ,   3739    ,   10943   ,0 };
            const unsigned long dim3276Kuo3Init[]   =   {   1   ,   3   ,   1   ,   7   ,   31  ,   35  ,   3   ,   85  ,   331 ,   161 ,   1167    ,   3627    ,   7075    ,   16189   ,   385 ,0 };
            const unsigned long dim3277Kuo3Init[]   =   {   1   ,   1   ,   3   ,   11  ,   3   ,   17  ,   103 ,   205 ,   299 ,   917 ,   105 ,   1675    ,   3461    ,   895 ,   3279    ,0 };
            const unsigned long dim3278Kuo3Init[]   =   {   1   ,   1   ,   1   ,   7   ,   25  ,   11  ,   47  ,   77  ,   91  ,   263 ,   1575    ,   1337    ,   5603    ,   6561    ,   4941    ,0 };
            const unsigned long dim3279Kuo3Init[]   =   {   1   ,   1   ,   1   ,   5   ,   15  ,   17  ,   77  ,   69  ,   413 ,   343 ,   1923    ,   2013    ,   269 ,   12245   ,   14611   ,0 };
            const unsigned long dim3280Kuo3Init[]   =   {   1   ,   3   ,   7   ,   13  ,   1   ,   29  ,   99  ,   83  ,   69  ,   871 ,   945 ,   1271    ,   5741    ,   2505    ,   4743    ,0 };
            const unsigned long dim3281Kuo3Init[]   =   {   1   ,   3   ,   5   ,   13  ,   23  ,   55  ,   95  ,   17  ,   321 ,   929 ,   1499    ,   205 ,   5529    ,   3819    ,   29763   ,0 };
            const unsigned long dim3282Kuo3Init[]   =   {   1   ,   3   ,   3   ,   1   ,   1   ,   47  ,   15  ,   151 ,   273 ,   991 ,   1993    ,   2325    ,   3633    ,   9033    ,   795 ,0 };
            const unsigned long dim3283Kuo3Init[]   =   {   1   ,   3   ,   1   ,   5   ,   19  ,   59  ,   115 ,   79  ,   51  ,   585 ,   253 ,   1545    ,   2871    ,   2303    ,   23189   ,0 };
            const unsigned long dim3284Kuo3Init[]   =   {   1   ,   1   ,   7   ,   11  ,   5   ,   49  ,   77  ,   79  ,   103 ,   719 ,   965 ,   2115    ,   2431    ,   7761    ,   7267    ,0 };
            const unsigned long dim3285Kuo3Init[]   =   {   1   ,   1   ,   7   ,   13  ,   1   ,   33  ,   13  ,   121 ,   297 ,   217 ,   1143    ,   2185    ,   6007    ,   13233   ,   4529    ,0 };
            const unsigned long dim3286Kuo3Init[]   =   {   1   ,   3   ,   7   ,   3   ,   1   ,   15  ,   17  ,   155 ,   321 ,   19  ,   1071    ,   3195    ,   3589    ,   12159   ,   30323   ,0 };
            const unsigned long dim3287Kuo3Init[]   =   {   1   ,   3   ,   5   ,   7   ,   3   ,   57  ,   35  ,   197 ,   247 ,   449 ,   557 ,   809 ,   909 ,   9141    ,   26613   ,0 };
            const unsigned long dim3288Kuo3Init[]   =   {   1   ,   3   ,   1   ,   5   ,   13  ,   43  ,   61  ,   177 ,   21  ,   161 ,   1823    ,   1077    ,   6523    ,   10765   ,   17427   ,0 };
            const unsigned long dim3289Kuo3Init[]   =   {   1   ,   1   ,   3   ,   3   ,   13  ,   27  ,   77  ,   77  ,   315 ,   763 ,   853 ,   3801    ,   1505    ,   13801   ,   28217   ,0 };
            const unsigned long dim3290Kuo3Init[]   =   {   1   ,   1   ,   1   ,   9   ,   15  ,   27  ,   45  ,   131 ,   271 ,   703 ,   1509    ,   443 ,   7739    ,   3011    ,   32463   ,0 };
            const unsigned long dim3291Kuo3Init[]   =   {   1   ,   1   ,   7   ,   11  ,   17  ,   3   ,   93  ,   55  ,   331 ,   785 ,   585 ,   3451    ,   283 ,   2623    ,   21459   ,0 };
            const unsigned long dim3292Kuo3Init[]   =   {   1   ,   3   ,   5   ,   15  ,   3   ,   3   ,   101 ,   25  ,   235 ,   721 ,   257 ,   3961    ,   7049    ,   9861    ,   21259   ,0 };
            const unsigned long dim3293Kuo3Init[]   =   {   1   ,   1   ,   7   ,   7   ,   9   ,   31  ,   17  ,   17  ,   473 ,   155 ,   677 ,   603 ,   7503    ,   2845    ,   21089   ,0 };
            const unsigned long dim3294Kuo3Init[]   =   {   1   ,   3   ,   3   ,   9   ,   11  ,   29  ,   5   ,   135 ,   141 ,   587 ,   1073    ,   461 ,   883 ,   16191   ,   16139   ,0 };
            const unsigned long dim3295Kuo3Init[]   =   {   1   ,   3   ,   1   ,   15  ,   25  ,   17  ,   87  ,   231 ,   63  ,   563 ,   1061    ,   3207    ,   2291    ,   913 ,   11871   ,0 };
            const unsigned long dim3296Kuo3Init[]   =   {   1   ,   1   ,   7   ,   11  ,   7   ,   45  ,   53  ,   79  ,   247 ,   21  ,   1203    ,   1857    ,   5095    ,   7533    ,   4427    ,0 };
            const unsigned long dim3297Kuo3Init[]   =   {   1   ,   3   ,   5   ,   9   ,   17  ,   57  ,   29  ,   181 ,   345 ,   709 ,   1889    ,   2625    ,   8157    ,   4681    ,   21681   ,0 };
            const unsigned long dim3298Kuo3Init[]   =   {   1   ,   3   ,   5   ,   13  ,   9   ,   63  ,   63  ,   183 ,   123 ,   957 ,   27  ,   3113    ,   1873    ,   8917    ,   8669    ,0 };
            const unsigned long dim3299Kuo3Init[]   =   {   1   ,   3   ,   5   ,   9   ,   29  ,   59  ,   31  ,   211 ,   459 ,   679 ,   259 ,   1885    ,   1895    ,   12815   ,   2437    ,0 };
            const unsigned long dim3300Kuo3Init[]   =   {   1   ,   1   ,   5   ,   11  ,   13  ,   51  ,   125 ,   5   ,   15  ,   865 ,   901 ,   1803    ,   3277    ,   2879    ,   13057   ,0 };
            const unsigned long dim3301Kuo3Init[]   =   {   1   ,   1   ,   7   ,   9   ,   27  ,   39  ,   115 ,   67  ,   389 ,   1   ,   649 ,   639 ,   7183    ,   15929   ,   12953   ,0 };
            const unsigned long dim3302Kuo3Init[]   =   {   1   ,   1   ,   5   ,   13  ,   3   ,   7   ,   21  ,   219 ,   241 ,   977 ,   1233    ,   1717    ,   7077    ,   8657    ,   1389    ,0 };
            const unsigned long dim3303Kuo3Init[]   =   {   1   ,   1   ,   5   ,   15  ,   9   ,   51  ,   101 ,   127 ,   383 ,   801 ,   657 ,   2793    ,   1037    ,   11525   ,   7195    ,0 };
            const unsigned long dim3304Kuo3Init[]   =   {   1   ,   3   ,   3   ,   7   ,   1   ,   29  ,   71  ,   227 ,   19  ,   371 ,   363 ,   809 ,   4241    ,   5739    ,   26497   ,0 };
            const unsigned long dim3305Kuo3Init[]   =   {   1   ,   3   ,   5   ,   11  ,   25  ,   23  ,   65  ,   161 ,   259 ,   135 ,   1047    ,   3943    ,   5749    ,   2931    ,   20647   ,0 };
            const unsigned long dim3306Kuo3Init[]   =   {   1   ,   1   ,   7   ,   3   ,   1   ,   55  ,   1   ,   85  ,   271 ,   67  ,   1813    ,   77  ,   5435    ,   11867   ,   28375   ,0 };
            const unsigned long dim3307Kuo3Init[]   =   {   1   ,   1   ,   7   ,   11  ,   31  ,   53  ,   43  ,   47  ,   285 ,   901 ,   901 ,   977 ,   733 ,   5291    ,   7043    ,0 };
            const unsigned long dim3308Kuo3Init[]   =   {   1   ,   3   ,   3   ,   7   ,   1   ,   59  ,   79  ,   123 ,   355 ,   721 ,   365 ,   2871    ,   6253    ,   14511   ,   4209    ,0 };
            const unsigned long dim3309Kuo3Init[]   =   {   1   ,   1   ,   5   ,   15  ,   5   ,   49  ,   77  ,   107 ,   237 ,   257 ,   703 ,   1013    ,   4563    ,   8603    ,   10469   ,0 };
            const unsigned long dim3310Kuo3Init[]   =   {   1   ,   3   ,   7   ,   11  ,   27  ,   59  ,   9   ,   129 ,   21  ,   991 ,   1283    ,   2871    ,   759 ,   5435    ,   30419   ,0 };
            const unsigned long dim3311Kuo3Init[]   =   {   1   ,   3   ,   5   ,   1   ,   19  ,   5   ,   5   ,   215 ,   91  ,   117 ,   939 ,   3595    ,   4331    ,   16083   ,   23279   ,0 };
            const unsigned long dim3312Kuo3Init[]   =   {   1   ,   1   ,   7   ,   7   ,   7   ,   45  ,   45  ,   71  ,   75  ,   261 ,   685 ,   63  ,   3553    ,   16349   ,   6803    ,0 };
            const unsigned long dim3313Kuo3Init[]   =   {   1   ,   3   ,   1   ,   9   ,   1   ,   23  ,   101 ,   11  ,   331 ,   407 ,   547 ,   1553    ,   7471    ,   6125    ,   3965    ,0 };
            const unsigned long dim3314Kuo3Init[]   =   {   1   ,   3   ,   7   ,   11  ,   19  ,   29  ,   1   ,   53  ,   125 ,   775 ,   1381    ,   1885    ,   2655    ,   7199    ,   24073   ,0 };
            const unsigned long dim3315Kuo3Init[]   =   {   1   ,   3   ,   5   ,   7   ,   27  ,   21  ,   79  ,   211 ,   433 ,   993 ,   143 ,   2241    ,   2607    ,   15447   ,   6121    ,0 };
            const unsigned long dim3316Kuo3Init[]   =   {   1   ,   3   ,   3   ,   7   ,   29  ,   39  ,   93  ,   65  ,   39  ,   577 ,   1471    ,   3491    ,   4413    ,   10553   ,   1339    ,0 };
            const unsigned long dim3317Kuo3Init[]   =   {   1   ,   1   ,   5   ,   11  ,   21  ,   23  ,   75  ,   237 ,   439 ,   171 ,   361 ,   3283    ,   4145    ,   15267   ,   19917   ,0 };
            const unsigned long dim3318Kuo3Init[]   =   {   1   ,   1   ,   1   ,   5   ,   9   ,   55  ,   39  ,   171 ,   423 ,   211 ,   43  ,   2597    ,   7135    ,   2437    ,   28965   ,0 };
            const unsigned long dim3319Kuo3Init[]   =   {   1   ,   3   ,   1   ,   9   ,   11  ,   15  ,   71  ,   109 ,   337 ,   907 ,   135 ,   3557    ,   7359    ,   3539    ,   26953   ,0 };
            const unsigned long dim3320Kuo3Init[]   =   {   1   ,   1   ,   1   ,   15  ,   15  ,   63  ,   9   ,   47  ,   153 ,   157 ,   457 ,   857 ,   6985    ,   3559    ,   23285   ,0 };
            const unsigned long dim3321Kuo3Init[]   =   {   1   ,   3   ,   3   ,   9   ,   21  ,   21  ,   3   ,   223 ,   67  ,   429 ,   497 ,   1303    ,   869 ,   3909    ,   17491   ,0 };
            const unsigned long dim3322Kuo3Init[]   =   {   1   ,   1   ,   1   ,   3   ,   23  ,   37  ,   7   ,   197 ,   223 ,   597 ,   421 ,   845 ,   1397    ,   927 ,   20621   ,0 };
            const unsigned long dim3323Kuo3Init[]   =   {   1   ,   3   ,   1   ,   13  ,   5   ,   37  ,   41  ,   83  ,   241 ,   301 ,   1157    ,   1087    ,   2365    ,   10905   ,   25195   ,0 };
            const unsigned long dim3324Kuo3Init[]   =   {   1   ,   3   ,   7   ,   11  ,   25  ,   11  ,   75  ,   9   ,   31  ,   5   ,   1909    ,   681 ,   2891    ,   867 ,   4661    ,0 };
            const unsigned long dim3325Kuo3Init[]   =   {   1   ,   1   ,   1   ,   13  ,   21  ,   33  ,   127 ,   233 ,   369 ,   381 ,   711 ,   133 ,   4463    ,   14119   ,   8877    ,0 };
            const unsigned long dim3326Kuo3Init[]   =   {   1   ,   1   ,   7   ,   7   ,   21  ,   13  ,   5   ,   61  ,   29  ,   247 ,   923 ,   3533    ,   2831    ,   9645    ,   1537    ,0 };
            const unsigned long dim3327Kuo3Init[]   =   {   1   ,   1   ,   1   ,   9   ,   29  ,   13  ,   29  ,   177 ,   195 ,   299 ,   445 ,   3753    ,   3699    ,   9383    ,   3249    ,0 };
            const unsigned long dim3328Kuo3Init[]   =   {   1   ,   3   ,   1   ,   3   ,   17  ,   55  ,   21  ,   245 ,   11  ,   93  ,   1097    ,   1601    ,   4431    ,   5847    ,   21521   ,0 };
            const unsigned long dim3329Kuo3Init[]   =   {   1   ,   1   ,   3   ,   9   ,   3   ,   37  ,   9   ,   179 ,   465 ,   123 ,   333 ,   3765    ,   1599    ,   8709    ,   11007   ,0 };
            const unsigned long dim3330Kuo3Init[]   =   {   1   ,   3   ,   3   ,   11  ,   15  ,   25  ,   23  ,   217 ,   193 ,   767 ,   1433    ,   3435    ,   4891    ,   4109    ,   32061   ,0 };
            const unsigned long dim3331Kuo3Init[]   =   {   1   ,   3   ,   7   ,   11  ,   21  ,   25  ,   123 ,   111 ,   377 ,   951 ,   1791    ,   2173    ,   569 ,   12831   ,   30045   ,0 };
            const unsigned long dim3332Kuo3Init[]   =   {   1   ,   1   ,   7   ,   11  ,   11  ,   19  ,   75  ,   63  ,   497 ,   1005    ,   1215    ,   3957    ,   4137    ,   4851    ,   23547   ,0 };
            const unsigned long dim3333Kuo3Init[]   =   {   1   ,   1   ,   1   ,   15  ,   15  ,   9   ,   59  ,   143 ,   27  ,   941 ,   117 ,   3035    ,   1201    ,   4373    ,   19291   ,0 };
            const unsigned long dim3334Kuo3Init[]   =   {   1   ,   3   ,   3   ,   5   ,   1   ,   3   ,   37  ,   187 ,   115 ,   901 ,   617 ,   1959    ,   2243    ,   6563    ,   3087    ,0 };
            const unsigned long dim3335Kuo3Init[]   =   {   1   ,   1   ,   1   ,   9   ,   17  ,   31  ,   67  ,   243 ,   35  ,   667 ,   1863    ,   1017    ,   6501    ,   9279    ,   18089   ,0 };
            const unsigned long dim3336Kuo3Init[]   =   {   1   ,   1   ,   5   ,   11  ,   25  ,   19  ,   25  ,   255 ,   371 ,   27  ,   763 ,   2575    ,   695 ,   6155    ,   27761   ,0 };
            const unsigned long dim3337Kuo3Init[]   =   {   1   ,   3   ,   1   ,   1   ,   5   ,   49  ,   79  ,   149 ,   331 ,   601 ,   199 ,   2927    ,   6293    ,   12449   ,   16397   ,0 };
            const unsigned long dim3338Kuo3Init[]   =   {   1   ,   1   ,   7   ,   11  ,   29  ,   41  ,   29  ,   161 ,   191 ,   519 ,   445 ,   1159    ,   3009    ,   3413    ,   21757   ,0 };
            const unsigned long dim3339Kuo3Init[]   =   {   1   ,   3   ,   7   ,   5   ,   23  ,   57  ,   93  ,   211 ,   57  ,   269 ,   1627    ,   3393    ,   873 ,   8189    ,   26059   ,0 };
            const unsigned long dim3340Kuo3Init[]   =   {   1   ,   1   ,   7   ,   11  ,   27  ,   59  ,   125 ,   251 ,   259 ,   251 ,   79  ,   2693    ,   1815    ,   8759    ,   15667   ,0 };
            const unsigned long dim3341Kuo3Init[]   =   {   1   ,   1   ,   3   ,   15  ,   31  ,   47  ,   15  ,   51  ,   17  ,   761 ,   589 ,   537 ,   3991    ,   2883    ,   6899    ,0 };
            const unsigned long dim3342Kuo3Init[]   =   {   1   ,   1   ,   3   ,   7   ,   19  ,   39  ,   95  ,   143 ,   37  ,   183 ,   1719    ,   1819    ,   1659    ,   1043    ,   31929   ,0 };
            const unsigned long dim3343Kuo3Init[]   =   {   1   ,   1   ,   3   ,   7   ,   15  ,   47  ,   105 ,   45  ,   67  ,   665 ,   1273    ,   221 ,   5423    ,   679 ,   8375    ,0 };
            const unsigned long dim3344Kuo3Init[]   =   {   1   ,   3   ,   1   ,   15  ,   23  ,   37  ,   123 ,   233 ,   117 ,   743 ,   785 ,   3871    ,   6031    ,   14379   ,   3339    ,0 };
            const unsigned long dim3345Kuo3Init[]   =   {   1   ,   3   ,   1   ,   13  ,   1   ,   47  ,   33  ,   63  ,   309 ,   961 ,   83  ,   2539    ,   5935    ,   3121    ,   27671   ,0 };
            const unsigned long dim3346Kuo3Init[]   =   {   1   ,   3   ,   7   ,   13  ,   21  ,   11  ,   103 ,   87  ,   433 ,   39  ,   293 ,   3217    ,   1533    ,   4511    ,   24493   ,0 };
            const unsigned long dim3347Kuo3Init[]   =   {   1   ,   1   ,   5   ,   7   ,   9   ,   51  ,   41  ,   133 ,   443 ,   605 ,   913 ,   377 ,   5969    ,   8401    ,   25427   ,0 };
            const unsigned long dim3348Kuo3Init[]   =   {   1   ,   1   ,   1   ,   3   ,   21  ,   27  ,   59  ,   41  ,   155 ,   35  ,   897 ,   2317    ,   6557    ,   679 ,   7863    ,0 };
            const unsigned long dim3349Kuo3Init[]   =   {   1   ,   1   ,   7   ,   11  ,   19  ,   47  ,   39  ,   53  ,   401 ,   199 ,   271 ,   1385    ,   3505    ,   719 ,   12963   ,0 };
            const unsigned long dim3350Kuo3Init[]   =   {   1   ,   1   ,   7   ,   7   ,   13  ,   7   ,   11  ,   131 ,   455 ,   605 ,   523 ,   3435    ,   2313    ,   9861    ,   22437   ,0 };
            const unsigned long dim3351Kuo3Init[]   =   {   1   ,   3   ,   7   ,   3   ,   19  ,   5   ,   125 ,   71  ,   425 ,   321 ,   1151    ,   231 ,   2585    ,   12899   ,   4789    ,0 };
            const unsigned long dim3352Kuo3Init[]   =   {   1   ,   3   ,   5   ,   11  ,   15  ,   27  ,   35  ,   13  ,   187 ,   863 ,   795 ,   1005    ,   6443    ,   3011    ,   5275    ,0 };
            const unsigned long dim3353Kuo3Init[]   =   {   1   ,   3   ,   7   ,   5   ,   25  ,   61  ,   33  ,   201 ,   343 ,   337 ,   1441    ,   2647    ,   3469    ,   5537    ,   30595   ,0 };
            const unsigned long dim3354Kuo3Init[]   =   {   1   ,   1   ,   7   ,   11  ,   17  ,   13  ,   73  ,   209 ,   121 ,   411 ,   1027    ,   1879    ,   845 ,   14019   ,   4269    ,0 };
            const unsigned long dim3355Kuo3Init[]   =   {   1   ,   3   ,   5   ,   9   ,   11  ,   51  ,   69  ,   103 ,   245 ,   677 ,   409 ,   135 ,   3775    ,   15781   ,   24159   ,0 };
            const unsigned long dim3356Kuo3Init[]   =   {   1   ,   1   ,   7   ,   3   ,   11  ,   9   ,   117 ,   101 ,   205 ,   745 ,   941 ,   125 ,   5915    ,   16305   ,   12547   ,0 };
            const unsigned long dim3357Kuo3Init[]   =   {   1   ,   1   ,   1   ,   9   ,   9   ,   51  ,   37  ,   135 ,   79  ,   327 ,   1695    ,   2517    ,   2685    ,   1103    ,   6035    ,0 };
            const unsigned long dim3358Kuo3Init[]   =   {   1   ,   1   ,   3   ,   13  ,   21  ,   33  ,   33  ,   229 ,   201 ,   939 ,   1625    ,   2043    ,   6661    ,   11821   ,   22881   ,0 };
            const unsigned long dim3359Kuo3Init[]   =   {   1   ,   1   ,   1   ,   11  ,   23  ,   3   ,   23  ,   5   ,   329 ,   467 ,   963 ,   3833    ,   7345    ,   3225    ,   3709    ,0 };
            const unsigned long dim3360Kuo3Init[]   =   {   1   ,   3   ,   1   ,   11  ,   27  ,   21  ,   35  ,   159 ,   255 ,   707 ,   353 ,   3095    ,   6701    ,   14311   ,   839 ,0 };
            const unsigned long dim3361Kuo3Init[]   =   {   1   ,   1   ,   5   ,   7   ,   7   ,   63  ,   107 ,   227 ,   233 ,   303 ,   1523    ,   2019    ,   227 ,   14695   ,   20253   ,0 };
            const unsigned long dim3362Kuo3Init[]   =   {   1   ,   3   ,   1   ,   1   ,   17  ,   29  ,   1   ,   167 ,   331 ,   795 ,   2027    ,   2333    ,   361 ,   1841    ,   4053    ,0 };
            const unsigned long dim3363Kuo3Init[]   =   {   1   ,   1   ,   7   ,   9   ,   1   ,   41  ,   71  ,   173 ,   103 ,   43  ,   635 ,   3311    ,   5883    ,   6517    ,   467 ,0 };
            const unsigned long dim3364Kuo3Init[]   =   {   1   ,   3   ,   7   ,   1   ,   3   ,   15  ,   101 ,   197 ,   449 ,   393 ,   889 ,   1189    ,   2665    ,   3647    ,   21977   ,0 };
            const unsigned long dim3365Kuo3Init[]   =   {   1   ,   3   ,   1   ,   11  ,   23  ,   5   ,   103 ,   199 ,   417 ,   729 ,   137 ,   1667    ,   5   ,   4769    ,   11981   ,0 };
            const unsigned long dim3366Kuo3Init[]   =   {   1   ,   3   ,   1   ,   1   ,   17  ,   37  ,   73  ,   231 ,   195 ,   497 ,   1835    ,   1597    ,   403 ,   12569   ,   24761   ,0 };
            const unsigned long dim3367Kuo3Init[]   =   {   1   ,   3   ,   3   ,   7   ,   31  ,   37  ,   33  ,   39  ,   359 ,   961 ,   1745    ,   2353    ,   827 ,   6439    ,   27419   ,0 };
            const unsigned long dim3368Kuo3Init[]   =   {   1   ,   1   ,   3   ,   1   ,   25  ,   25  ,   97  ,   111 ,   149 ,   799 ,   593 ,   1011    ,   5439    ,   2613    ,   3669    ,0 };
            const unsigned long dim3369Kuo3Init[]   =   {   1   ,   1   ,   3   ,   5   ,   5   ,   21  ,   55  ,   37  ,   483 ,   407 ,   1059    ,   905 ,   5225    ,   9441    ,   17461   ,0 };
            const unsigned long dim3370Kuo3Init[]   =   {   1   ,   1   ,   1   ,   11  ,   27  ,   15  ,   93  ,   87  ,   423 ,   625 ,   285 ,   4051    ,   7931    ,   9005    ,   4033    ,0 };
            const unsigned long dim3371Kuo3Init[]   =   {   1   ,   1   ,   1   ,   7   ,   13  ,   17  ,   13  ,   221 ,   29  ,   179 ,   645 ,   2721    ,   4781    ,   6151    ,   6435    ,0 };
            const unsigned long dim3372Kuo3Init[]   =   {   1   ,   1   ,   5   ,   9   ,   11  ,   59  ,   53  ,   249 ,   13  ,   295 ,   467 ,   1781    ,   2381    ,   2897    ,   11175   ,0 };
            const unsigned long dim3373Kuo3Init[]   =   {   1   ,   3   ,   5   ,   3   ,   5   ,   41  ,   111 ,   129 ,   159 ,   363 ,   1109    ,   1547    ,   1209    ,   15035   ,   17367   ,0 };
            const unsigned long dim3374Kuo3Init[]   =   {   1   ,   1   ,   7   ,   13  ,   23  ,   51  ,   55  ,   129 ,   435 ,   671 ,   947 ,   349 ,   1741    ,   15317   ,   28069   ,0 };
            const unsigned long dim3375Kuo3Init[]   =   {   1   ,   3   ,   5   ,   7   ,   13  ,   9   ,   17  ,   111 ,   309 ,   987 ,   1521    ,   1243    ,   1815    ,   10055   ,   4883    ,0 };
            const unsigned long dim3376Kuo3Init[]   =   {   1   ,   1   ,   5   ,   5   ,   21  ,   9   ,   19  ,   179 ,   315 ,   547 ,   1899    ,   1421    ,   589 ,   2563    ,   27523   ,0 };
            const unsigned long dim3377Kuo3Init[]   =   {   1   ,   1   ,   5   ,   11  ,   25  ,   23  ,   115 ,   229 ,   439 ,   225 ,   1957    ,   3251    ,   4377    ,   12751   ,   2983    ,0 };
            const unsigned long dim3378Kuo3Init[]   =   {   1   ,   3   ,   1   ,   3   ,   21  ,   23  ,   57  ,   187 ,   469 ,   575 ,   1587    ,   1983    ,   4353    ,   8605    ,   27269   ,0 };
            const unsigned long dim3379Kuo3Init[]   =   {   1   ,   3   ,   5   ,   15  ,   11  ,   17  ,   51  ,   131 ,   93  ,   599 ,   487 ,   3397    ,   761 ,   11503   ,   13405   ,0 };
            const unsigned long dim3380Kuo3Init[]   =   {   1   ,   3   ,   1   ,   15  ,   27  ,   7   ,   13  ,   105 ,   189 ,   141 ,   715 ,   1571    ,   8103    ,   3233    ,   10481   ,0 };
            const unsigned long dim3381Kuo3Init[]   =   {   1   ,   1   ,   7   ,   3   ,   21  ,   17  ,   103 ,   121 ,   229 ,   865 ,   1819    ,   171 ,   7281    ,   14085   ,   4483    ,0 };
            const unsigned long dim3382Kuo3Init[]   =   {   1   ,   3   ,   7   ,   13  ,   21  ,   47  ,   105 ,   233 ,   325 ,   49  ,   1305    ,   2269    ,   7717    ,   2771    ,   27583   ,0 };
            const unsigned long dim3383Kuo3Init[]   =   {   1   ,   3   ,   3   ,   5   ,   17  ,   11  ,   101 ,   55  ,   191 ,   115 ,   1749    ,   3015    ,   1493    ,   15301   ,   1629    ,0 };
            const unsigned long dim3384Kuo3Init[]   =   {   1   ,   1   ,   7   ,   1   ,   3   ,   39  ,   111 ,   51  ,   269 ,   451 ,   1327    ,   3389    ,   6461    ,   7361    ,   1311    ,0 };
            const unsigned long dim3385Kuo3Init[]   =   {   1   ,   1   ,   7   ,   7   ,   23  ,   21  ,   103 ,   241 ,   3   ,   295 ,   981 ,   3969    ,   1631    ,   6301    ,   9481    ,0 };
            const unsigned long dim3386Kuo3Init[]   =   {   1   ,   3   ,   7   ,   3   ,   17  ,   63  ,   49  ,   163 ,   89  ,   183 ,   555 ,   597 ,   3099    ,   8057    ,   28653   ,0 };
            const unsigned long dim3387Kuo3Init[]   =   {   1   ,   1   ,   7   ,   15  ,   9   ,   29  ,   35  ,   255 ,   173 ,   295 ,   2015    ,   283 ,   1179    ,   1677    ,   26417   ,0 };
            const unsigned long dim3388Kuo3Init[]   =   {   1   ,   3   ,   1   ,   5   ,   11  ,   29  ,   73  ,   67  ,   5   ,   995 ,   1563    ,   2611    ,   7605    ,   13831   ,   7405    ,0 };
            const unsigned long dim3389Kuo3Init[]   =   {   1   ,   3   ,   3   ,   1   ,   1   ,   27  ,   7   ,   251 ,   175 ,   683 ,   603 ,   2895    ,   6033    ,   4133    ,   2753    ,0 };
            const unsigned long dim3390Kuo3Init[]   =   {   1   ,   1   ,   7   ,   15  ,   23  ,   47  ,   45  ,   179 ,   289 ,   317 ,   1111    ,   1241    ,   7221    ,   8661    ,   31853   ,0 };
            const unsigned long dim3391Kuo3Init[]   =   {   1   ,   1   ,   3   ,   15  ,   19  ,   19  ,   35  ,   139 ,   145 ,   469 ,   1455    ,   2335    ,   8009    ,   6609    ,   14847   ,0 };
            const unsigned long dim3392Kuo3Init[]   =   {   1   ,   1   ,   7   ,   11  ,   17  ,   59  ,   115 ,   71  ,   197 ,   143 ,   1791    ,   3515    ,   3443    ,   5065    ,   22885   ,0 };
            const unsigned long dim3393Kuo3Init[]   =   {   1   ,   3   ,   7   ,   11  ,   25  ,   19  ,   47  ,   223 ,   469 ,   745 ,   107 ,   2801    ,   6279    ,   16311   ,   15699   ,0 };
            const unsigned long dim3394Kuo3Init[]   =   {   1   ,   1   ,   1   ,   5   ,   29  ,   47  ,   21  ,   227 ,   219 ,   331 ,   1179    ,   3409    ,   1787    ,   9113    ,   8359    ,0 };
            const unsigned long dim3395Kuo3Init[]   =   {   1   ,   3   ,   7   ,   3   ,   7   ,   29  ,   43  ,   135 ,   137 ,   203 ,   241 ,   2421    ,   3125    ,   3883    ,   23519   ,0 };
            const unsigned long dim3396Kuo3Init[]   =   {   1   ,   3   ,   5   ,   3   ,   3   ,   63  ,   5   ,   5   ,   425 ,   863 ,   1483    ,   4049    ,   2505    ,   3173    ,   3021    ,0 };
            const unsigned long dim3397Kuo3Init[]   =   {   1   ,   1   ,   1   ,   1   ,   31  ,   11  ,   37  ,   133 ,   349 ,   33  ,   1617    ,   7   ,   4945    ,   8143    ,   22027   ,0 };
            const unsigned long dim3398Kuo3Init[]   =   {   1   ,   3   ,   7   ,   9   ,   13  ,   61  ,   61  ,   111 ,   203 ,   255 ,   1107    ,   425 ,   4571    ,   1537    ,   2547    ,0 };
            const unsigned long dim3399Kuo3Init[]   =   {   1   ,   3   ,   1   ,   13  ,   19  ,   35  ,   5   ,   99  ,   33  ,   271 ,   203 ,   545 ,   1917    ,   8645    ,   30737   ,0 };
            const unsigned long dim3400Kuo3Init[]   =   {   1   ,   1   ,   5   ,   1   ,   31  ,   33  ,   125 ,   119 ,   475 ,   497 ,   1341    ,   593 ,   3497    ,   2191    ,   28715   ,0 };
            const unsigned long dim3401Kuo3Init[]   =   {   1   ,   1   ,   7   ,   15  ,   17  ,   57  ,   87  ,   101 ,   487 ,   659 ,   497 ,   1243    ,   3353    ,   4803    ,   11445   ,0 };
            const unsigned long dim3402Kuo3Init[]   =   {   1   ,   3   ,   1   ,   3   ,   31  ,   13  ,   55  ,   79  ,   233 ,   327 ,   1825    ,   245 ,   2693    ,   11761   ,   21155   ,0 };
            const unsigned long dim3403Kuo3Init[]   =   {   1   ,   3   ,   7   ,   7   ,   21  ,   1   ,   15  ,   183 ,   113 ,   871 ,   251 ,   3573    ,   517 ,   9973    ,   28567   ,0 };
            const unsigned long dim3404Kuo3Init[]   =   {   1   ,   1   ,   1   ,   11  ,   23  ,   53  ,   47  ,   143 ,   435 ,   107 ,   1091    ,   2139    ,   5985    ,   12541   ,   9857    ,0 };
            const unsigned long dim3405Kuo3Init[]   =   {   1   ,   3   ,   3   ,   13  ,   23  ,   47  ,   125 ,   95  ,   333 ,   13  ,   285 ,   1691    ,   1899    ,   5625    ,   29883   ,0 };
            const unsigned long dim3406Kuo3Init[]   =   {   1   ,   1   ,   7   ,   1   ,   1   ,   3   ,   21  ,   69  ,   425 ,   179 ,   1815    ,   2231    ,   6257    ,   1387    ,   26423   ,0 };
            const unsigned long dim3407Kuo3Init[]   =   {   1   ,   1   ,   1   ,   13  ,   23  ,   11  ,   37  ,   223 ,   259 ,   767 ,   1521    ,   2445    ,   7725    ,   2013    ,   8365    ,0 };
            const unsigned long dim3408Kuo3Init[]   =   {   1   ,   1   ,   7   ,   7   ,   13  ,   25  ,   125 ,   167 ,   423 ,   457 ,   1793    ,   4043    ,   913 ,   14917   ,   4577    ,0 };
            const unsigned long dim3409Kuo3Init[]   =   {   1   ,   3   ,   5   ,   11  ,   13  ,   49  ,   95  ,   145 ,   251 ,   305 ,   1883    ,   3717    ,   7927    ,   6635    ,   9647    ,0 };
            const unsigned long dim3410Kuo3Init[]   =   {   1   ,   3   ,   1   ,   1   ,   31  ,   33  ,   87  ,   189 ,   499 ,   933 ,   115 ,   3407    ,   1517    ,   157 ,   11815   ,0 };
            const unsigned long dim3411Kuo3Init[]   =   {   1   ,   1   ,   7   ,   7   ,   15  ,   1   ,   7   ,   155 ,   493 ,   413 ,   231 ,   69  ,   3433    ,   13471   ,   26705   ,0 };
            const unsigned long dim3412Kuo3Init[]   =   {   1   ,   3   ,   7   ,   15  ,   3   ,   59  ,   21  ,   81  ,   309 ,   445 ,   1467    ,   2655    ,   4871    ,   1667    ,   24557   ,0 };
            const unsigned long dim3413Kuo3Init[]   =   {   1   ,   3   ,   3   ,   15  ,   11  ,   25  ,   73  ,   181 ,   355 ,   887 ,   967 ,   1131    ,   5137    ,   12695   ,   8561    ,0 };
            const unsigned long dim3414Kuo3Init[]   =   {   1   ,   3   ,   3   ,   11  ,   27  ,   51  ,   19  ,   95  ,   165 ,   991 ,   2039    ,   3269    ,   2839    ,   12065   ,   26791   ,0 };
            const unsigned long dim3415Kuo3Init[]   =   {   1   ,   1   ,   5   ,   15  ,   29  ,   15  ,   47  ,   89  ,   251 ,   149 ,   7   ,   447 ,   1945    ,   157 ,   3865    ,0 };
            const unsigned long dim3416Kuo3Init[]   =   {   1   ,   1   ,   5   ,   15  ,   27  ,   33  ,   99  ,   211 ,   481 ,   95  ,   1361    ,   2933    ,   509 ,   16383   ,   11581   ,0 };
            const unsigned long dim3417Kuo3Init[]   =   {   1   ,   1   ,   3   ,   1   ,   29  ,   43  ,   103 ,   141 ,   117 ,   505 ,   733 ,   2719    ,   2281    ,   4753    ,   27511   ,0 };
            const unsigned long dim3418Kuo3Init[]   =   {   1   ,   3   ,   1   ,   15  ,   5   ,   15  ,   25  ,   87  ,   255 ,   819 ,   1589    ,   2623    ,   2043    ,   7735    ,   7813    ,0 };
            const unsigned long dim3419Kuo3Init[]   =   {   1   ,   1   ,   5   ,   15  ,   23  ,   57  ,   125 ,   145 ,   387 ,   729 ,   1845    ,   2809    ,   8067    ,   3487    ,   19919   ,0 };
            const unsigned long dim3420Kuo3Init[]   =   {   1   ,   3   ,   3   ,   5   ,   17  ,   59  ,   45  ,   11  ,   15  ,   505 ,   1859    ,   2193    ,   4581    ,   15553   ,   767 ,0 };
            const unsigned long dim3421Kuo3Init[]   =   {   1   ,   1   ,   1   ,   3   ,   27  ,   47  ,   95  ,   147 ,   493 ,   259 ,   759 ,   4013    ,   5611    ,   1379    ,   13017   ,0 };
            const unsigned long dim3422Kuo3Init[]   =   {   1   ,   3   ,   3   ,   9   ,   13  ,   3   ,   7   ,   193 ,   203 ,   379 ,   1445    ,   1599    ,   5915    ,   6025    ,   8037    ,0 };
            const unsigned long dim3423Kuo3Init[]   =   {   1   ,   3   ,   7   ,   5   ,   7   ,   33  ,   77  ,   5   ,   449 ,   509 ,   1409    ,   2711    ,   5603    ,   9343    ,   10613   ,0 };
            const unsigned long dim3424Kuo3Init[]   =   {   1   ,   3   ,   5   ,   1   ,   31  ,   19  ,   65  ,   169 ,   429 ,   951 ,   1659    ,   3045    ,   735 ,   5089    ,   27215   ,0 };
            const unsigned long dim3425Kuo3Init[]   =   {   1   ,   1   ,   7   ,   1   ,   13  ,   43  ,   73  ,   67  ,   449 ,   357 ,   285 ,   2291    ,   3493    ,   1575    ,   9281    ,0 };
            const unsigned long dim3426Kuo3Init[]   =   {   1   ,   3   ,   7   ,   13  ,   7   ,   59  ,   15  ,   119 ,   479 ,   523 ,   201 ,   1381    ,   3245    ,   15591   ,   18785   ,0 };
            const unsigned long dim3427Kuo3Init[]   =   {   1   ,   3   ,   1   ,   13  ,   13  ,   13  ,   29  ,   105 ,   65  ,   231 ,   161 ,   375 ,   3725    ,   5097    ,   23183   ,0 };
            const unsigned long dim3428Kuo3Init[]   =   {   1   ,   3   ,   5   ,   7   ,   7   ,   21  ,   103 ,   155 ,   357 ,   661 ,   1035    ,   2421    ,   5989    ,   14613   ,   11499   ,0 };
            const unsigned long dim3429Kuo3Init[]   =   {   1   ,   3   ,   3   ,   3   ,   13  ,   7   ,   45  ,   221 ,   189 ,   405 ,   1845    ,   3791    ,   2989    ,   495 ,   13899   ,0 };
            const unsigned long dim3430Kuo3Init[]   =   {   1   ,   3   ,   7   ,   15  ,   27  ,   23  ,   91  ,   39  ,   385 ,   59  ,   1367    ,   3177    ,   1945    ,   15335   ,   7543    ,0 };
            const unsigned long dim3431Kuo3Init[]   =   {   1   ,   3   ,   1   ,   5   ,   25  ,   49  ,   43  ,   215 ,   7   ,   1003    ,   831 ,   611 ,   183 ,   4627    ,   29423   ,0 };
            const unsigned long dim3432Kuo3Init[]   =   {   1   ,   1   ,   3   ,   9   ,   29  ,   15  ,   57  ,   7   ,   259 ,   543 ,   1313    ,   449 ,   2497    ,   4337    ,   20879   ,0 };
            const unsigned long dim3433Kuo3Init[]   =   {   1   ,   3   ,   1   ,   3   ,   17  ,   23  ,   111 ,   99  ,   185 ,   183 ,   1215    ,   1775    ,   8129    ,   4487    ,   7927    ,0 };
            const unsigned long dim3434Kuo3Init[]   =   {   1   ,   1   ,   1   ,   1   ,   21  ,   23  ,   55  ,   195 ,   219 ,   595 ,   703 ,   97  ,   7599    ,   15673   ,   25677   ,0 };
            const unsigned long dim3435Kuo3Init[]   =   {   1   ,   3   ,   1   ,   11  ,   19  ,   11  ,   121 ,   145 ,   237 ,   719 ,   201 ,   1335    ,   5147    ,   2195    ,   22909   ,0 };
            const unsigned long dim3436Kuo3Init[]   =   {   1   ,   1   ,   3   ,   11  ,   27  ,   27  ,   45  ,   185 ,   251 ,   329 ,   1253    ,   1071    ,   2097    ,   9669    ,   2999    ,0 };
            const unsigned long dim3437Kuo3Init[]   =   {   1   ,   3   ,   3   ,   11  ,   27  ,   35  ,   91  ,   183 ,   161 ,   393 ,   1269    ,   1945    ,   727 ,   9985    ,   9275    ,0 };
            const unsigned long dim3438Kuo3Init[]   =   {   1   ,   1   ,   1   ,   9   ,   15  ,   43  ,   109 ,   25  ,   431 ,   641 ,   1291    ,   1169    ,   649 ,   16145   ,   27663   ,0 };
            const unsigned long dim3439Kuo3Init[]   =   {   1   ,   3   ,   5   ,   5   ,   11  ,   59  ,   77  ,   193 ,   59  ,   325 ,   531 ,   365 ,   7285    ,   5111    ,   26647   ,0 };
            const unsigned long dim3440Kuo3Init[]   =   {   1   ,   1   ,   3   ,   9   ,   15  ,   1   ,   33  ,   51  ,   427 ,   513 ,   575 ,   3739    ,   4593    ,   6669    ,   25369   ,0 };
            const unsigned long dim3441Kuo3Init[]   =   {   1   ,   3   ,   1   ,   13  ,   5   ,   31  ,   109 ,   155 ,   83  ,   405 ,   657 ,   1095    ,   5087    ,   13835   ,   30395   ,0 };
            const unsigned long dim3442Kuo3Init[]   =   {   1   ,   3   ,   7   ,   9   ,   11  ,   21  ,   115 ,   129 ,   389 ,   25  ,   657 ,   3455    ,   1357    ,   1501    ,   10589   ,0 };
            const unsigned long dim3443Kuo3Init[]   =   {   1   ,   3   ,   5   ,   9   ,   29  ,   21  ,   67  ,   153 ,   197 ,   703 ,   123 ,   1645    ,   2175    ,   5581    ,   26465   ,0 };
            const unsigned long dim3444Kuo3Init[]   =   {   1   ,   1   ,   7   ,   11  ,   1   ,   21  ,   79  ,   71  ,   49  ,   719 ,   909 ,   3783    ,   5081    ,   13235   ,   19387   ,0 };
            const unsigned long dim3445Kuo3Init[]   =   {   1   ,   1   ,   1   ,   5   ,   11  ,   47  ,   93  ,   213 ,   305 ,   673 ,   191 ,   2787    ,   3293    ,   6721    ,   19105   ,0 };
            const unsigned long dim3446Kuo3Init[]   =   {   1   ,   1   ,   5   ,   15  ,   19  ,   1   ,   127 ,   189 ,   25  ,   201 ,   1867    ,   3541    ,   3215    ,   14025   ,   15823   ,0 };
            const unsigned long dim3447Kuo3Init[]   =   {   1   ,   1   ,   7   ,   5   ,   9   ,   43  ,   75  ,   129 ,   335 ,   73  ,   995 ,   15  ,   4423    ,   6817    ,   14311   ,0 };
            const unsigned long dim3448Kuo3Init[]   =   {   1   ,   3   ,   3   ,   7   ,   13  ,   53  ,   97  ,   29  ,   311 ,   749 ,   1025    ,   5   ,   6001    ,   6831    ,   24699   ,0 };
            const unsigned long dim3449Kuo3Init[]   =   {   1   ,   3   ,   7   ,   3   ,   29  ,   21  ,   117 ,   245 ,   389 ,   785 ,   405 ,   1455    ,   4249    ,   9663    ,   30027   ,0 };
            const unsigned long dim3450Kuo3Init[]   =   {   1   ,   1   ,   5   ,   13  ,   13  ,   27  ,   113 ,   221 ,   293 ,   583 ,   1329    ,   3387    ,   4233    ,   11511   ,   30005   ,0 };
            const unsigned long dim3451Kuo3Init[]   =   {   1   ,   3   ,   1   ,   11  ,   11  ,   59  ,   85  ,   23  ,   435 ,   769 ,   889 ,   2993    ,   8075    ,   10831   ,   23109   ,0 };
            const unsigned long dim3452Kuo3Init[]   =   {   1   ,   1   ,   7   ,   13  ,   31  ,   9   ,   59  ,   247 ,   387 ,   27  ,   791 ,   1381    ,   1881    ,   15223   ,   29481   ,0 };
            const unsigned long dim3453Kuo3Init[]   =   {   1   ,   1   ,   7   ,   7   ,   1   ,   39  ,   93  ,   67  ,   293 ,   777 ,   1463    ,   1363    ,   3159    ,   6569    ,   18361   ,0 };
            const unsigned long dim3454Kuo3Init[]   =   {   1   ,   3   ,   1   ,   11  ,   25  ,   31  ,   49  ,   133 ,   459 ,   399 ,   593 ,   3907    ,   701 ,   9321    ,   28589   ,0 };
            const unsigned long dim3455Kuo3Init[]   =   {   1   ,   1   ,   7   ,   5   ,   7   ,   35  ,   85  ,   247 ,   247 ,   577 ,   245 ,   1615    ,   337 ,   11289   ,   32149   ,0 };
            const unsigned long dim3456Kuo3Init[]   =   {   1   ,   3   ,   5   ,   7   ,   21  ,   35  ,   67  ,   37  ,   77  ,   395 ,   907 ,   595 ,   883 ,   14931   ,   21875   ,0 };
            const unsigned long dim3457Kuo3Init[]   =   {   1   ,   1   ,   1   ,   9   ,   5   ,   25  ,   1   ,   69  ,   307 ,   419 ,   1875    ,   3075    ,   15  ,   5853    ,   3177    ,0 };
            const unsigned long dim3458Kuo3Init[]   =   {   1   ,   1   ,   3   ,   1   ,   15  ,   7   ,   15  ,   17  ,   67  ,   577 ,   1525    ,   1847    ,   481 ,   15451   ,   21229   ,0 };
            const unsigned long dim3459Kuo3Init[]   =   {   1   ,   1   ,   1   ,   11  ,   23  ,   35  ,   79  ,   179 ,   397 ,   403 ,   1935    ,   3371    ,   3511    ,   13593   ,   30241   ,0 };
            const unsigned long dim3460Kuo3Init[]   =   {   1   ,   3   ,   5   ,   3   ,   13  ,   11  ,   79  ,   215 ,   409 ,   577 ,   1301    ,   451 ,   5171    ,   1555    ,   18445   ,0 };
            const unsigned long dim3461Kuo3Init[]   =   {   1   ,   3   ,   3   ,   13  ,   25  ,   35  ,   71  ,   65  ,   371 ,   667 ,   247 ,   3815    ,   4383    ,   6429    ,   4667    ,0 };
            const unsigned long dim3462Kuo3Init[]   =   {   1   ,   1   ,   3   ,   1   ,   17  ,   45  ,   115 ,   217 ,   353 ,   501 ,   1087    ,   1521    ,   2621    ,   4847    ,   12647   ,0 };
            const unsigned long dim3463Kuo3Init[]   =   {   1   ,   1   ,   7   ,   7   ,   5   ,   25  ,   75  ,   33  ,   421 ,   39  ,   1857    ,   2189    ,   2383    ,   4401    ,   11737   ,0 };
            const unsigned long dim3464Kuo3Init[]   =   {   1   ,   1   ,   7   ,   7   ,   7   ,   23  ,   123 ,   155 ,   343 ,   85  ,   879 ,   595 ,   6419    ,   3789    ,   31233   ,0 };
            const unsigned long dim3465Kuo3Init[]   =   {   1   ,   1   ,   7   ,   5   ,   29  ,   55  ,   75  ,   203 ,   319 ,   465 ,   1903    ,   595 ,   8127    ,   9773    ,   4083    ,0 };
            const unsigned long dim3466Kuo3Init[]   =   {   1   ,   3   ,   7   ,   11  ,   27  ,   31  ,   55  ,   147 ,   17  ,   893 ,   33  ,   3239    ,   1773    ,   2111    ,   19833   ,0 };
            const unsigned long dim3467Kuo3Init[]   =   {   1   ,   3   ,   7   ,   5   ,   27  ,   39  ,   3   ,   131 ,   401 ,   843 ,   1229    ,   227 ,   5221    ,   5845    ,   29029   ,0 };
            const unsigned long dim3468Kuo3Init[]   =   {   1   ,   1   ,   5   ,   13  ,   15  ,   47  ,   113 ,   11  ,   361 ,   95  ,   1303    ,   333 ,   5139    ,   8393    ,   2225    ,0 };
            const unsigned long dim3469Kuo3Init[]   =   {   1   ,   3   ,   5   ,   11  ,   27  ,   53  ,   45  ,   111 ,   377 ,   41  ,   985 ,   931 ,   3901    ,   13921   ,   25305   ,0 };
            const unsigned long dim3470Kuo3Init[]   =   {   1   ,   3   ,   7   ,   3   ,   23  ,   23  ,   95  ,   227 ,   207 ,   71  ,   525 ,   3843    ,   6025    ,   16335   ,   30519   ,0 };
            const unsigned long dim3471Kuo3Init[]   =   {   1   ,   3   ,   5   ,   1   ,   9   ,   55  ,   95  ,   69  ,   457 ,   823 ,   479 ,   4073    ,   3227    ,   1355    ,   12937   ,0 };
            const unsigned long dim3472Kuo3Init[]   =   {   1   ,   1   ,   1   ,   5   ,   7   ,   17  ,   67  ,   143 ,   189 ,   79  ,   653 ,   2791    ,   1193    ,   9409    ,   3561    ,0 };
            const unsigned long dim3473Kuo3Init[]   =   {   1   ,   1   ,   1   ,   5   ,   29  ,   53  ,   67  ,   253 ,   471 ,   649 ,   669 ,   1621    ,   1199    ,   12889   ,   27943   ,0 };
            const unsigned long dim3474Kuo3Init[]   =   {   1   ,   1   ,   3   ,   11  ,   15  ,   3   ,   93  ,   33  ,   335 ,   899 ,   1897    ,   1689    ,   2753    ,   473 ,   21935   ,0 };
            const unsigned long dim3475Kuo3Init[]   =   {   1   ,   1   ,   7   ,   7   ,   27  ,   29  ,   17  ,   167 ,   467 ,   261 ,   671 ,   2449    ,   7433    ,   5669    ,   77  ,0 };
            const unsigned long dim3476Kuo3Init[]   =   {   1   ,   1   ,   7   ,   9   ,   5   ,   37  ,   85  ,   103 ,   19  ,   971 ,   1851    ,   231 ,   7897    ,   1783    ,   12659   ,0 };
            const unsigned long dim3477Kuo3Init[]   =   {   1   ,   3   ,   7   ,   7   ,   17  ,   57  ,   1   ,   7   ,   491 ,   887 ,   341 ,   3859    ,   2783    ,   9547    ,   26867   ,0 };
            const unsigned long dim3478Kuo3Init[]   =   {   1   ,   1   ,   1   ,   3   ,   29  ,   45  ,   77  ,   219 ,   131 ,   573 ,   1015    ,   765 ,   6211    ,   3243    ,   32547   ,0 };
            const unsigned long dim3479Kuo3Init[]   =   {   1   ,   3   ,   1   ,   5   ,   19  ,   51  ,   61  ,   27  ,   415 ,   125 ,   69  ,   2151    ,   3385    ,   1405    ,   25509   ,0 };
            const unsigned long dim3480Kuo3Init[]   =   {   1   ,   3   ,   1   ,   3   ,   23  ,   17  ,   19  ,   35  ,   489 ,   869 ,   687 ,   1047    ,   6031    ,   589 ,   3611    ,0 };
            const unsigned long dim3481Kuo3Init[]   =   {   1   ,   3   ,   5   ,   15  ,   19  ,   27  ,   47  ,   175 ,   23  ,   427 ,   137 ,   2733    ,   1795    ,   483 ,   11083   ,0 };
            const unsigned long dim3482Kuo3Init[]   =   {   1   ,   1   ,   1   ,   1   ,   23  ,   41  ,   93  ,   229 ,   407 ,   751 ,   635 ,   3125    ,   831 ,   2121    ,   8463    ,0 };
            const unsigned long dim3483Kuo3Init[]   =   {   1   ,   3   ,   5   ,   11  ,   27  ,   61  ,   27  ,   21  ,   337 ,   33  ,   1373    ,   37  ,   89  ,   7527    ,   21165   ,0 };
            const unsigned long dim3484Kuo3Init[]   =   {   1   ,   1   ,   5   ,   9   ,   25  ,   55  ,   51  ,   111 ,   497 ,   737 ,   1097    ,   165 ,   7879    ,   1509    ,   5025    ,0 };
            const unsigned long dim3485Kuo3Init[]   =   {   1   ,   3   ,   5   ,   11  ,   3   ,   51  ,   61  ,   19  ,   75  ,   1001    ,   1535    ,   2689    ,   2903    ,   9675    ,   22209   ,0 };
            const unsigned long dim3486Kuo3Init[]   =   {   1   ,   1   ,   3   ,   13  ,   27  ,   7   ,   89  ,   219 ,   459 ,   403 ,   1963    ,   979 ,   3987    ,   4165    ,   5027    ,0 };
            const unsigned long dim3487Kuo3Init[]   =   {   1   ,   1   ,   1   ,   13  ,   15  ,   51  ,   113 ,   37  ,   105 ,   57  ,   589 ,   3583    ,   7205    ,   107 ,   29687   ,0 };
            const unsigned long dim3488Kuo3Init[]   =   {   1   ,   1   ,   7   ,   7   ,   1   ,   57  ,   125 ,   131 ,   265 ,   957 ,   461 ,   1413    ,   3961    ,   10543   ,   14951   ,0 };
            const unsigned long dim3489Kuo3Init[]   =   {   1   ,   1   ,   7   ,   13  ,   19  ,   53  ,   125 ,   41  ,   417 ,   629 ,   153 ,   2531    ,   5655    ,   1919    ,   14615   ,0 };
            const unsigned long dim3490Kuo3Init[]   =   {   1   ,   1   ,   7   ,   1   ,   1   ,   21  ,   19  ,   123 ,   3   ,   601 ,   1781    ,   2237    ,   705 ,   14229   ,   6739    ,0 };
            const unsigned long dim3491Kuo3Init[]   =   {   1   ,   3   ,   5   ,   5   ,   3   ,   3   ,   23  ,   71  ,   225 ,   507 ,   619 ,   493 ,   5917    ,   1145    ,   7303    ,0 };
            const unsigned long dim3492Kuo3Init[]   =   {   1   ,   1   ,   1   ,   13  ,   3   ,   19  ,   89  ,   237 ,   497 ,   3   ,   91  ,   2559    ,   6347    ,   8925    ,   9845    ,0 };
            const unsigned long dim3493Kuo3Init[]   =   {   1   ,   3   ,   5   ,   7   ,   11  ,   47  ,   85  ,   147 ,   339 ,   499 ,   511 ,   731 ,   3063    ,   5363    ,   19841   ,0 };
            const unsigned long dim3494Kuo3Init[]   =   {   1   ,   3   ,   7   ,   7   ,   19  ,   41  ,   105 ,   225 ,   37  ,   181 ,   127 ,   2167    ,   4247    ,   915 ,   14507   ,0 };
            const unsigned long dim3495Kuo3Init[]   =   {   1   ,   1   ,   5   ,   11  ,   31  ,   7   ,   23  ,   229 ,   165 ,   485 ,   693 ,   3279    ,   7407    ,   9883    ,   31803   ,0 };
            const unsigned long dim3496Kuo3Init[]   =   {   1   ,   3   ,   5   ,   9   ,   15  ,   33  ,   1   ,   99  ,   79  ,   589 ,   411 ,   2783    ,   3437    ,   2151    ,   11585   ,0 };
            const unsigned long dim3497Kuo3Init[]   =   {   1   ,   3   ,   1   ,   5   ,   25  ,   15  ,   95  ,   221 ,   399 ,   793 ,   617 ,   2431    ,   6859    ,   2893    ,   8919    ,0 };
            const unsigned long dim3498Kuo3Init[]   =   {   1   ,   3   ,   1   ,   11  ,   21  ,   57  ,   127 ,   1   ,   81  ,   657 ,   1935    ,   1937    ,   1635    ,   4079    ,   6677    ,0 };
            const unsigned long dim3499Kuo3Init[]   =   {   1   ,   1   ,   5   ,   1   ,   17  ,   7   ,   19  ,   167 ,   483 ,   53  ,   49  ,   1187    ,   789 ,   1787    ,   4103    ,0 };
            const unsigned long dim3500Kuo3Init[]   =   {   1   ,   3   ,   5   ,   15  ,   11  ,   1   ,   83  ,   223 ,   289 ,   119 ,   1263    ,   2995    ,   5697    ,   5835    ,   7127    ,0 };
            const unsigned long dim3501Kuo3Init[]   =   {   1   ,   3   ,   7   ,   13  ,   1   ,   25  ,   59  ,   59  ,   223 ,   309 ,   1133    ,   773 ,   8171    ,   12881   ,   21959   ,0 };
            const unsigned long dim3502Kuo3Init[]   =   {   1   ,   3   ,   7   ,   15  ,   7   ,   43  ,   63  ,   99  ,   503 ,   527 ,   1915    ,   1147    ,   2729    ,   8311    ,   13451   ,0 };
            const unsigned long dim3503Kuo3Init[]   =   {   1   ,   1   ,   1   ,   7   ,   25  ,   37  ,   61  ,   41  ,   25  ,   681 ,   1029    ,   1715    ,   63  ,   6913    ,   31299   ,0 };
            const unsigned long dim3504Kuo3Init[]   =   {   1   ,   3   ,   5   ,   13  ,   29  ,   41  ,   25  ,   135 ,   241 ,   331 ,   51  ,   2305    ,   5149    ,   12295   ,   8891    ,0 };
            const unsigned long dim3505Kuo3Init[]   =   {   1   ,   1   ,   3   ,   7   ,   7   ,   53  ,   25  ,   127 ,   227 ,   711 ,   1165    ,   2563    ,   2681    ,   1331    ,   8435    ,0 };
            const unsigned long dim3506Kuo3Init[]   =   {   1   ,   3   ,   3   ,   15  ,   29  ,   43  ,   27  ,   105 ,   371 ,   691 ,   77  ,   1847    ,   6097    ,   1253    ,   1163    ,0 };
            const unsigned long dim3507Kuo3Init[]   =   {   1   ,   3   ,   5   ,   1   ,   23  ,   43  ,   23  ,   185 ,   287 ,   407 ,   1523    ,   1923    ,   2121    ,   3297    ,   21549   ,0 };
            const unsigned long dim3508Kuo3Init[]   =   {   1   ,   3   ,   7   ,   15  ,   31  ,   45  ,   19  ,   85  ,   323 ,   837 ,   99  ,   581 ,   6607    ,   3685    ,   21853   ,0 };
            const unsigned long dim3509Kuo3Init[]   =   {   1   ,   1   ,   3   ,   3   ,   31  ,   27  ,   107 ,   55  ,   79  ,   339 ,   1739    ,   2373    ,   6393    ,   6557    ,   20651   ,0 };
            const unsigned long dim3510Kuo3Init[]   =   {   1   ,   3   ,   1   ,   11  ,   7   ,   23  ,   63  ,   43  ,   237 ,   753 ,   1347    ,   1775    ,   6663    ,   2907    ,   2905    ,0 };
            const unsigned long dim3511Kuo3Init[]   =   {   1   ,   1   ,   5   ,   5   ,   1   ,   3   ,   99  ,   135 ,   247 ,   249 ,   757 ,   619 ,   3497    ,   2421    ,   6601    ,0 };
            const unsigned long dim3512Kuo3Init[]   =   {   1   ,   3   ,   1   ,   7   ,   5   ,   61  ,   91  ,   151 ,   283 ,   723 ,   1931    ,   3799    ,   2737    ,   14833   ,   2363    ,0 };
            const unsigned long dim3513Kuo3Init[]   =   {   1   ,   1   ,   3   ,   1   ,   9   ,   15  ,   53  ,   233 ,   365 ,   323 ,   1141    ,   3523    ,   7771    ,   707 ,   1837    ,0 };
            const unsigned long dim3514Kuo3Init[]   =   {   1   ,   1   ,   1   ,   7   ,   11  ,   43  ,   87  ,   63  ,   425 ,   149 ,   1121    ,   4051    ,   5391    ,   591 ,   19849   ,0 };
            const unsigned long dim3515Kuo3Init[]   =   {   1   ,   1   ,   1   ,   5   ,   15  ,   39  ,   55  ,   233 ,   401 ,   337 ,   289 ,   693 ,   3659    ,   10381   ,   25723   ,0 };
            const unsigned long dim3516Kuo3Init[]   =   {   1   ,   1   ,   7   ,   11  ,   3   ,   11  ,   55  ,   89  ,   293 ,   235 ,   893 ,   853 ,   1281    ,   13845   ,   7123    ,0 };
            const unsigned long dim3517Kuo3Init[]   =   {   1   ,   1   ,   7   ,   15  ,   11  ,   1   ,   41  ,   187 ,   297 ,   557 ,   1951    ,   4007    ,   2429    ,   493 ,   31645   ,0 };
            const unsigned long dim3518Kuo3Init[]   =   {   1   ,   1   ,   1   ,   1   ,   31  ,   39  ,   113 ,   3   ,   11  ,   217 ,   1841    ,   545 ,   1953    ,   1667    ,   28317   ,0 };
            const unsigned long dim3519Kuo3Init[]   =   {   1   ,   1   ,   3   ,   7   ,   31  ,   31  ,   125 ,   217 ,   127 ,   561 ,   983 ,   3651    ,   6649    ,   11859   ,   3887    ,0 };
            const unsigned long dim3520Kuo3Init[]   =   {   1   ,   1   ,   1   ,   3   ,   17  ,   51  ,   107 ,   25  ,   511 ,   777 ,   771 ,   2359    ,   275 ,   16351   ,   15639   ,0 };
            const unsigned long dim3521Kuo3Init[]   =   {   1   ,   3   ,   5   ,   1   ,   11  ,   13  ,   9   ,   167 ,   505 ,   667 ,   155 ,   2857    ,   4043    ,   12145   ,   3369    ,0 };
            const unsigned long dim3522Kuo3Init[]   =   {   1   ,   3   ,   7   ,   1   ,   15  ,   33  ,   83  ,   217 ,   333 ,   847 ,   679 ,   3981    ,   925 ,   3465    ,   1217    ,0 };
            const unsigned long dim3523Kuo3Init[]   =   {   1   ,   1   ,   3   ,   7   ,   13  ,   55  ,   111 ,   101 ,   455 ,   929 ,   1755    ,   2177    ,   931 ,   8083    ,   1207    ,0 };
            const unsigned long dim3524Kuo3Init[]   =   {   1   ,   1   ,   1   ,   3   ,   21  ,   29  ,   25  ,   213 ,   113 ,   51  ,   1913    ,   1131    ,   1981    ,   16221   ,   32569   ,0 };
            const unsigned long dim3525Kuo3Init[]   =   {   1   ,   1   ,   3   ,   13  ,   13  ,   9   ,   105 ,   153 ,   317 ,   755 ,   363 ,   2833    ,   4463    ,   2607    ,   17587   ,0 };
            const unsigned long dim3526Kuo3Init[]   =   {   1   ,   1   ,   5   ,   1   ,   7   ,   15  ,   71  ,   133 ,   141 ,   639 ,   273 ,   1445    ,   4571    ,   13313   ,   319 ,0 };
            const unsigned long dim3527Kuo3Init[]   =   {   1   ,   1   ,   5   ,   9   ,   23  ,   17  ,   55  ,   49  ,   159 ,   713 ,   1127    ,   2537    ,   4927    ,   4389    ,   8361    ,0 };
            const unsigned long dim3528Kuo3Init[]   =   {   1   ,   3   ,   3   ,   1   ,   31  ,   23  ,   35  ,   237 ,   7   ,   801 ,   1939    ,   2243    ,   3891    ,   12379   ,   251 ,0 };
            const unsigned long dim3529Kuo3Init[]   =   {   1   ,   1   ,   7   ,   15  ,   17  ,   11  ,   33  ,   89  ,   39  ,   87  ,   1963    ,   1591    ,   7775    ,   5091    ,   23503   ,0 };
            const unsigned long dim3530Kuo3Init[]   =   {   1   ,   1   ,   7   ,   1   ,   31  ,   63  ,   93  ,   55  ,   197 ,   627 ,   1887    ,   899 ,   2491    ,   7713    ,   9921    ,0 };
            const unsigned long dim3531Kuo3Init[]   =   {   1   ,   1   ,   3   ,   11  ,   27  ,   17  ,   27  ,   33  ,   413 ,   131 ,   1713    ,   615 ,   1775    ,   237 ,   2775    ,0 };
            const unsigned long dim3532Kuo3Init[]   =   {   1   ,   3   ,   5   ,   9   ,   23  ,   3   ,   83  ,   69  ,   113 ,   311 ,   1199    ,   2593    ,   3745    ,   14473   ,   727 ,0 };
            const unsigned long dim3533Kuo3Init[]   =   {   1   ,   3   ,   3   ,   3   ,   3   ,   57  ,   111 ,   195 ,   303 ,   713 ,   1305    ,   955 ,   7415    ,   15625   ,   6491    ,0 };
            const unsigned long dim3534Kuo3Init[]   =   {   1   ,   3   ,   1   ,   15  ,   25  ,   37  ,   3   ,   149 ,   7   ,   923 ,   1367    ,   3743    ,   5887    ,   5683    ,   17437   ,0 };
            const unsigned long dim3535Kuo3Init[]   =   {   1   ,   3   ,   3   ,   7   ,   5   ,   33  ,   85  ,   149 ,   339 ,   223 ,   1359    ,   1881    ,   4383    ,   8949    ,   1331    ,0 };
            const unsigned long dim3536Kuo3Init[]   =   {   1   ,   1   ,   3   ,   15  ,   21  ,   27  ,   85  ,   209 ,   205 ,   597 ,   655 ,   2759    ,   109 ,   4963    ,   13055   ,0 };
            const unsigned long dim3537Kuo3Init[]   =   {   1   ,   3   ,   1   ,   1   ,   3   ,   29  ,   99  ,   15  ,   399 ,   233 ,   1697    ,   2613    ,   1819    ,   9059    ,   10329   ,0 };
            const unsigned long dim3538Kuo3Init[]   =   {   1   ,   1   ,   5   ,   1   ,   19  ,   13  ,   69  ,   87  ,   451 ,   389 ,   1065    ,   665 ,   7217    ,   12907   ,   32423   ,0 };
            const unsigned long dim3539Kuo3Init[]   =   {   1   ,   1   ,   3   ,   11  ,   23  ,   57  ,   127 ,   107 ,   181 ,   231 ,   1233    ,   1593    ,   6365    ,   13919   ,   29379   ,0 };
            const unsigned long dim3540Kuo3Init[]   =   {   1   ,   1   ,   7   ,   13  ,   29  ,   43  ,   11  ,   119 ,   165 ,   835 ,   323 ,   3051    ,   3641    ,   13841   ,   26399   ,0 };
            const unsigned long dim3541Kuo3Init[]   =   {   1   ,   1   ,   7   ,   5   ,   15  ,   27  ,   41  ,   97  ,   317 ,   565 ,   1589    ,   2091    ,   1213    ,   12143   ,   26777   ,0 };
            const unsigned long dim3542Kuo3Init[]   =   {   1   ,   3   ,   7   ,   13  ,   3   ,   17  ,   65  ,   221 ,   183 ,   265 ,   363 ,   1013    ,   7269    ,   13731   ,   25029   ,0 };
            const unsigned long dim3543Kuo3Init[]   =   {   1   ,   1   ,   3   ,   3   ,   13  ,   13  ,   123 ,   59  ,   193 ,   95  ,   1909    ,   1701    ,   6123    ,   5599    ,   17805   ,0 };
            const unsigned long dim3544Kuo3Init[]   =   {   1   ,   1   ,   1   ,   1   ,   3   ,   51  ,   1   ,   155 ,   59  ,   199 ,   1751    ,   813 ,   441 ,   5615    ,   28115   ,0 };
            const unsigned long dim3545Kuo3Init[]   =   {   1   ,   1   ,   5   ,   3   ,   13  ,   33  ,   37  ,   245 ,   343 ,   21  ,   1083    ,   2177    ,   4593    ,   13995   ,   12465   ,0 };
            const unsigned long dim3546Kuo3Init[]   =   {   1   ,   1   ,   7   ,   3   ,   21  ,   57  ,   31  ,   65  ,   15  ,   427 ,   981 ,   1047    ,   7731    ,   763 ,   26335   ,0 };
            const unsigned long dim3547Kuo3Init[]   =   {   1   ,   3   ,   3   ,   11  ,   1   ,   15  ,   19  ,   207 ,   127 ,   623 ,   949 ,   663 ,   2691    ,   505 ,   10667   ,0 };
            const unsigned long dim3548Kuo3Init[]   =   {   1   ,   1   ,   3   ,   9   ,   3   ,   7   ,   31  ,   175 ,   225 ,   707 ,   1393    ,   1661    ,   2761    ,   4675    ,   16195   ,0 };
            const unsigned long dim3549Kuo3Init[]   =   {   1   ,   3   ,   1   ,   5   ,   21  ,   35  ,   17  ,   61  ,   337 ,   307 ,   191 ,   1077    ,   7331    ,   8965    ,   18963   ,0 };
            const unsigned long dim3550Kuo3Init[]   =   {   1   ,   3   ,   3   ,   7   ,   23  ,   53  ,   37  ,   95  ,   229 ,   125 ,   813 ,   2279    ,   4915    ,   10637   ,   15533   ,0 };
            const unsigned long dim3551Kuo3Init[]   =   {   1   ,   3   ,   3   ,   1   ,   21  ,   9   ,   61  ,   183 ,   457 ,   997 ,   929 ,   2245    ,   6213    ,   11179   ,   1803    ,0 };
            const unsigned long dim3552Kuo3Init[]   =   {   1   ,   1   ,   5   ,   15  ,   11  ,   63  ,   21  ,   235 ,   259 ,   161 ,   1347    ,   2807    ,   3991    ,   14205   ,   26369   ,0 };
            const unsigned long dim3553Kuo3Init[]   =   {   1   ,   1   ,   7   ,   5   ,   13  ,   3   ,   33  ,   237 ,   473 ,   167 ,   1341    ,   3617    ,   3671    ,   15041   ,   29603   ,0 };
            const unsigned long dim3554Kuo3Init[]   =   {   1   ,   1   ,   5   ,   9   ,   17  ,   1   ,   105 ,   169 ,   25  ,   825 ,   1703    ,   751 ,   5461    ,   13121   ,   30917   ,0 };
            const unsigned long dim3555Kuo3Init[]   =   {   1   ,   3   ,   7   ,   5   ,   9   ,   31  ,   19  ,   231 ,   137 ,   601 ,   1177    ,   327 ,   3811    ,   7553    ,   13453   ,0 };
            const unsigned long dim3556Kuo3Init[]   =   {   1   ,   1   ,   7   ,   7   ,   7   ,   43  ,   7   ,   231 ,   171 ,   695 ,   331 ,   1493    ,   2437    ,   12487   ,   25451   ,0 };
            const unsigned long dim3557Kuo3Init[]   =   {   1   ,   3   ,   5   ,   13  ,   27  ,   19  ,   103 ,   157 ,   145 ,   267 ,   1417    ,   3867    ,   7477    ,   15399   ,   31311   ,0 };
            const unsigned long dim3558Kuo3Init[]   =   {   1   ,   1   ,   7   ,   1   ,   5   ,   11  ,   111 ,   89  ,   45  ,   685 ,   1283    ,   2581    ,   819 ,   3675    ,   25143   ,0 };
            const unsigned long dim3559Kuo3Init[]   =   {   1   ,   1   ,   1   ,   5   ,   31  ,   5   ,   85  ,   143 ,   113 ,   693 ,   529 ,   321 ,   5407    ,   14411   ,   27513   ,0 };
            const unsigned long dim3560Kuo3Init[]   =   {   1   ,   1   ,   1   ,   9   ,   19  ,   5   ,   103 ,   79  ,   439 ,   327 ,   1769    ,   3371    ,   7191    ,   5827    ,   10145   ,0 };
            const unsigned long dim3561Kuo3Init[]   =   {   1   ,   3   ,   1   ,   15  ,   1   ,   3   ,   63  ,   57  ,   331 ,   525 ,   239 ,   307 ,   4927    ,   13525   ,   26383   ,0 };
            const unsigned long dim3562Kuo3Init[]   =   {   1   ,   1   ,   1   ,   11  ,   11  ,   31  ,   49  ,   137 ,   407 ,   335 ,   905 ,   3441    ,   623 ,   11301   ,   8933    ,0 };
            const unsigned long dim3563Kuo3Init[]   =   {   1   ,   3   ,   7   ,   13  ,   29  ,   59  ,   41  ,   143 ,   103 ,   255 ,   1195    ,   3033    ,   6785    ,   12717   ,   5789    ,0 };
            const unsigned long dim3564Kuo3Init[]   =   {   1   ,   3   ,   7   ,   5   ,   11  ,   3   ,   107 ,   121 ,   173 ,   481 ,   3   ,   1485    ,   4557    ,   11095   ,   24611   ,0 };
            const unsigned long dim3565Kuo3Init[]   =   {   1   ,   3   ,   3   ,   1   ,   3   ,   47  ,   17  ,   107 ,   77  ,   323 ,   1349    ,   2285    ,   871 ,   7301    ,   27811   ,0 };
            const unsigned long dim3566Kuo3Init[]   =   {   1   ,   3   ,   3   ,   13  ,   9   ,   49  ,   95  ,   53  ,   337 ,   323 ,   1231    ,   3825    ,   7403    ,   12851   ,   31133   ,0 };
            const unsigned long dim3567Kuo3Init[]   =   {   1   ,   3   ,   1   ,   1   ,   29  ,   59  ,   123 ,   239 ,   251 ,   513 ,   259 ,   529 ,   1267    ,   13323   ,   15509   ,0 };
            const unsigned long dim3568Kuo3Init[]   =   {   1   ,   1   ,   1   ,   7   ,   25  ,   61  ,   91  ,   75  ,   83  ,   575 ,   671 ,   2575    ,   2871    ,   11909   ,   31371   ,0 };
            const unsigned long dim3569Kuo3Init[]   =   {   1   ,   3   ,   7   ,   3   ,   15  ,   63  ,   99  ,   161 ,   77  ,   539 ,   1069    ,   1263    ,   4173    ,   12691   ,   18091   ,0 };
            const unsigned long dim3570Kuo3Init[]   =   {   1   ,   3   ,   7   ,   1   ,   15  ,   55  ,   39  ,   33  ,   257 ,   925 ,   839 ,   4091    ,   2621    ,   315 ,   28153   ,0 };
            const unsigned long dim3571Kuo3Init[]   =   {   1   ,   1   ,   3   ,   3   ,   9   ,   9   ,   17  ,   207 ,   369 ,   773 ,   1661    ,   3175    ,   2367    ,   6363    ,   5363    ,0 };
            const unsigned long dim3572Kuo3Init[]   =   {   1   ,   1   ,   3   ,   15  ,   31  ,   11  ,   41  ,   145 ,   381 ,   969 ,   327 ,   487 ,   5545    ,   16163   ,   13891   ,0 };
            const unsigned long dim3573Kuo3Init[]   =   {   1   ,   1   ,   3   ,   5   ,   19  ,   15  ,   125 ,   183 ,   445 ,   771 ,   829 ,   3653    ,   7487    ,   14429   ,   14535   ,0 };
            const unsigned long dim3574Kuo3Init[]   =   {   1   ,   1   ,   1   ,   13  ,   5   ,   55  ,   41  ,   171 ,   163 ,   677 ,   1549    ,   2769    ,   3961    ,   4051    ,   25883   ,0 };
            const unsigned long dim3575Kuo3Init[]   =   {   1   ,   3   ,   5   ,   5   ,   15  ,   5   ,   121 ,   113 ,   423 ,   607 ,   1623    ,   2899    ,   7839    ,   8107    ,   13963   ,0 };
            const unsigned long dim3576Kuo3Init[]   =   {   1   ,   1   ,   7   ,   11  ,   17  ,   55  ,   85  ,   139 ,   93  ,   453 ,   75  ,   2405    ,   7601    ,   15647   ,   25349   ,0 };
            const unsigned long dim3577Kuo3Init[]   =   {   1   ,   3   ,   3   ,   3   ,   11  ,   27  ,   1   ,   11  ,   361 ,   811 ,   599 ,   405 ,   6981    ,   5705    ,   16427   ,0 };
            const unsigned long dim3578Kuo3Init[]   =   {   1   ,   3   ,   5   ,   9   ,   23  ,   61  ,   27  ,   7   ,   99  ,   495 ,   473 ,   477 ,   1733    ,   8043    ,   24843   ,0 };
            const unsigned long dim3579Kuo3Init[]   =   {   1   ,   3   ,   7   ,   3   ,   21  ,   27  ,   71  ,   157 ,   357 ,   175 ,   829 ,   2411    ,   2305    ,   4669    ,   24933   ,0 };
            const unsigned long dim3580Kuo3Init[]   =   {   1   ,   3   ,   5   ,   9   ,   21  ,   13  ,   121 ,   123 ,   119 ,   59  ,   429 ,   3757    ,   5833    ,   11263   ,   16013   ,0 };
            const unsigned long dim3581Kuo3Init[]   =   {   1   ,   3   ,   7   ,   15  ,   11  ,   49  ,   7   ,   31  ,   375 ,   647 ,   413 ,   2979    ,   7731    ,   15953   ,   13637   ,0 };
            const unsigned long dim3582Kuo3Init[]   =   {   1   ,   1   ,   1   ,   7   ,   9   ,   59  ,   121 ,   115 ,   239 ,   739 ,   121 ,   791 ,   769 ,   14223   ,   15127   ,0 };
            const unsigned long dim3583Kuo3Init[]   =   {   1   ,   3   ,   7   ,   15  ,   17  ,   31  ,   69  ,   9   ,   189 ,   503 ,   565 ,   1373    ,   6467    ,   15291   ,   20621   ,0 };
            const unsigned long dim3584Kuo3Init[]   =   {   1   ,   1   ,   5   ,   13  ,   31  ,   13  ,   71  ,   15  ,   197 ,   855 ,   1923    ,   2923    ,   2375    ,   7201    ,   29767   ,0 };
            const unsigned long dim3585Kuo3Init[]   =   {   1   ,   1   ,   7   ,   7   ,   17  ,   13  ,   53  ,   131 ,   447 ,   337 ,   1059    ,   1365    ,   1269    ,   6035    ,   32075   ,0 };
            const unsigned long dim3586Kuo3Init[]   =   {   1   ,   3   ,   5   ,   1   ,   15  ,   13  ,   43  ,   151 ,   207 ,   591 ,   717 ,   971 ,   5801    ,   10975   ,   3889    ,0 };
            const unsigned long dim3587Kuo3Init[]   =   {   1   ,   3   ,   5   ,   7   ,   15  ,   45  ,   125 ,   193 ,   17  ,   605 ,   651 ,   1987    ,   577 ,   2835    ,   6801    ,0 };
            const unsigned long dim3588Kuo3Init[]   =   {   1   ,   3   ,   5   ,   7   ,   31  ,   53  ,   83  ,   137 ,   399 ,   781 ,   135 ,   2863    ,   4901    ,   2861    ,   14025   ,0 };
            const unsigned long dim3589Kuo3Init[]   =   {   1   ,   3   ,   3   ,   13  ,   29  ,   45  ,   99  ,   43  ,   121 ,   141 ,   753 ,   11  ,   4033    ,   2011    ,   7493    ,0 };
            const unsigned long dim3590Kuo3Init[]   =   {   1   ,   3   ,   7   ,   11  ,   7   ,   39  ,   95  ,   199 ,   365 ,   9   ,   381 ,   3923    ,   3803    ,   6607    ,   30777   ,0 };
            const unsigned long dim3591Kuo3Init[]   =   {   1   ,   3   ,   5   ,   13  ,   13  ,   49  ,   17  ,   25  ,   195 ,   269 ,   83  ,   2115    ,   4013    ,   8817    ,   28017   ,0 };
            const unsigned long dim3592Kuo3Init[]   =   {   1   ,   1   ,   5   ,   15  ,   23  ,   49  ,   51  ,   49  ,   303 ,   439 ,   1353    ,   2385    ,   4547    ,   14405   ,   24115   ,0 };
            const unsigned long dim3593Kuo3Init[]   =   {   1   ,   3   ,   1   ,   9   ,   9   ,   3   ,   87  ,   159 ,   449 ,   139 ,   461 ,   465 ,   5585    ,   9313    ,   6337    ,0 };
            const unsigned long dim3594Kuo3Init[]   =   {   1   ,   1   ,   1   ,   9   ,   23  ,   17  ,   103 ,   167 ,   209 ,   179 ,   137 ,   57  ,   4265    ,   1743    ,   27575   ,0 };
            const unsigned long dim3595Kuo3Init[]   =   {   1   ,   3   ,   1   ,   15  ,   15  ,   17  ,   7   ,   93  ,   173 ,   277 ,   497 ,   4047    ,   2205    ,   10403   ,   6319    ,0 };
            const unsigned long dim3596Kuo3Init[]   =   {   1   ,   3   ,   1   ,   9   ,   13  ,   55  ,   37  ,   153 ,   69  ,   989 ,   1609    ,   1939    ,   1153    ,   5671    ,   16563   ,0 };
            const unsigned long dim3597Kuo3Init[]   =   {   1   ,   1   ,   3   ,   15  ,   3   ,   41  ,   67  ,   41  ,   253 ,   293 ,   1495    ,   3935    ,   271 ,   12067   ,   3891    ,0 };
            const unsigned long dim3598Kuo3Init[]   =   {   1   ,   3   ,   7   ,   15  ,   27  ,   21  ,   39  ,   41  ,   43  ,   713 ,   747 ,   995 ,   5389    ,   9811    ,   20489   ,0 };
            const unsigned long dim3599Kuo3Init[]   =   {   1   ,   3   ,   7   ,   15  ,   11  ,   47  ,   41  ,   155 ,   301 ,   867 ,   1467    ,   1503    ,   893 ,   6819    ,   6923    ,0 };
            const unsigned long dim3600Kuo3Init[]   =   {   1   ,   1   ,   1   ,   11  ,   29  ,   15  ,   57  ,   25  ,   219 ,   65  ,   997 ,   2109    ,   1921    ,   1771    ,   5997    ,0 };
            const unsigned long dim3601Kuo3Init[]   =   {   1   ,   1   ,   7   ,   5   ,   11  ,   59  ,   89  ,   59  ,   319 ,   841 ,   1   ,   1339    ,   5635    ,   3141    ,   8735    ,0 };
            const unsigned long dim3602Kuo3Init[]   =   {   1   ,   3   ,   3   ,   3   ,   17  ,   25  ,   13  ,   201 ,   37  ,   561 ,   1101    ,   1407    ,   3003    ,   99  ,   20189   ,0 };
            const unsigned long dim3603Kuo3Init[]   =   {   1   ,   1   ,   3   ,   3   ,   9   ,   7   ,   65  ,   27  ,   203 ,   399 ,   555 ,   2605    ,   7951    ,   833 ,   6103    ,0 };
            const unsigned long dim3604Kuo3Init[]   =   {   1   ,   3   ,   7   ,   15  ,   15  ,   25  ,   87  ,   215 ,   499 ,   935 ,   563 ,   1091    ,   6601    ,   12097   ,   22935   ,0 };
            const unsigned long dim3605Kuo3Init[]   =   {   1   ,   3   ,   5   ,   9   ,   3   ,   21  ,   63  ,   51  ,   353 ,   187 ,   1609    ,   2321    ,   1743    ,   12665   ,   9313    ,0 };
            const unsigned long dim3606Kuo3Init[]   =   {   1   ,   1   ,   3   ,   13  ,   21  ,   7   ,   101 ,   27  ,   15  ,   467 ,   833 ,   2439    ,   3559    ,   10509   ,   16327   ,0 };
            const unsigned long dim3607Kuo3Init[]   =   {   1   ,   3   ,   5   ,   1   ,   21  ,   37  ,   107 ,   77  ,   147 ,   107 ,   1493    ,   969 ,   7895    ,   14293   ,   14557   ,0 };
            const unsigned long dim3608Kuo3Init[]   =   {   1   ,   1   ,   3   ,   13  ,   1   ,   13  ,   57  ,   183 ,   303 ,   93  ,   1755    ,   845 ,   6857    ,   11061   ,   13661   ,0 };
            const unsigned long dim3609Kuo3Init[]   =   {   1   ,   1   ,   7   ,   13  ,   11  ,   23  ,   25  ,   77  ,   75  ,   279 ,   1375    ,   1235    ,   3225    ,   8143    ,   4055    ,0 };
            const unsigned long dim3610Kuo3Init[]   =   {   1   ,   3   ,   1   ,   11  ,   5   ,   1   ,   85  ,   1   ,   221 ,   831 ,   1673    ,   633 ,   4837    ,   14409   ,   28993   ,0 };
            const unsigned long dim3611Kuo3Init[]   =   {   1   ,   1   ,   7   ,   7   ,   23  ,   33  ,   51  ,   181 ,   201 ,   947 ,   697 ,   3417    ,   6169    ,   757 ,   9343    ,0 };
            const unsigned long dim3612Kuo3Init[]   =   {   1   ,   1   ,   3   ,   11  ,   21  ,   51  ,   45  ,   81  ,   447 ,   17  ,   789 ,   3773    ,   3647    ,   5913    ,   25419   ,0 };
            const unsigned long dim3613Kuo3Init[]   =   {   1   ,   1   ,   5   ,   11  ,   29  ,   39  ,   33  ,   67  ,   157 ,   583 ,   653 ,   43  ,   4441    ,   9571    ,   9895    ,0 };
            const unsigned long dim3614Kuo3Init[]   =   {   1   ,   3   ,   3   ,   13  ,   21  ,   13  ,   53  ,   3   ,   135 ,   61  ,   1975    ,   2107    ,   1395    ,   11395   ,   1741    ,0 };
            const unsigned long dim3615Kuo3Init[]   =   {   1   ,   1   ,   5   ,   3   ,   21  ,   29  ,   35  ,   25  ,   495 ,   783 ,   407 ,   1765    ,   1959    ,   3975    ,   23157   ,0 };
            const unsigned long dim3616Kuo3Init[]   =   {   1   ,   1   ,   7   ,   7   ,   1   ,   57  ,   107 ,   73  ,   483 ,   945 ,   73  ,   2757    ,   7211    ,   1053    ,   7625    ,0 };
            const unsigned long dim3617Kuo3Init[]   =   {   1   ,   1   ,   3   ,   13  ,   25  ,   53  ,   61  ,   117 ,   231 ,   261 ,   901 ,   1823    ,   3401    ,   8399    ,   17953   ,0 };
            const unsigned long dim3618Kuo3Init[]   =   {   1   ,   1   ,   7   ,   15  ,   29  ,   13  ,   55  ,   51  ,   469 ,   457 ,   1529    ,   2381    ,   6017    ,   10953   ,   18639   ,0 };
            const unsigned long dim3619Kuo3Init[]   =   {   1   ,   1   ,   1   ,   9   ,   21  ,   17  ,   45  ,   1   ,   457 ,   821 ,   367 ,   453 ,   1015    ,   1017    ,   21745   ,0 };
            const unsigned long dim3620Kuo3Init[]   =   {   1   ,   3   ,   1   ,   9   ,   15  ,   21  ,   121 ,   149 ,   417 ,   287 ,   897 ,   2595    ,   327 ,   16157   ,   11971   ,0 };
            const unsigned long dim3621Kuo3Init[]   =   {   1   ,   3   ,   7   ,   15  ,   17  ,   29  ,   41  ,   191 ,   475 ,   457 ,   1487    ,   1885    ,   6959    ,   10869   ,   7435    ,0 };
            const unsigned long dim3622Kuo3Init[]   =   {   1   ,   1   ,   5   ,   3   ,   9   ,   57  ,   29  ,   41  ,   81  ,   179 ,   1023    ,   3683    ,   4313    ,   4849    ,   14211   ,0 };
            const unsigned long dim3623Kuo3Init[]   =   {   1   ,   3   ,   1   ,   15  ,   15  ,   3   ,   123 ,   5   ,   201 ,   387 ,   1175    ,   1285    ,   4367    ,   459 ,   7771    ,0 };
            const unsigned long dim3624Kuo3Init[]   =   {   1   ,   1   ,   1   ,   3   ,   25  ,   47  ,   113 ,   41  ,   299 ,   217 ,   495 ,   565 ,   2337    ,   9877    ,   28731   ,0 };
            const unsigned long dim3625Kuo3Init[]   =   {   1   ,   1   ,   5   ,   5   ,   17  ,   11  ,   119 ,   187 ,   389 ,   19  ,   1947    ,   2607    ,   1119    ,   7677    ,   1561    ,0 };
            const unsigned long dim3626Kuo3Init[]   =   {   1   ,   1   ,   7   ,   9   ,   23  ,   59  ,   63  ,   175 ,   103 ,   211 ,   125 ,   3465    ,   2535    ,   14899   ,   1239    ,0 };
            const unsigned long dim3627Kuo3Init[]   =   {   1   ,   1   ,   3   ,   11  ,   17  ,   37  ,   111 ,   225 ,   51  ,   455 ,   1911    ,   1579    ,   841 ,   11451   ,   31521   ,0 };
            const unsigned long dim3628Kuo3Init[]   =   {   1   ,   1   ,   3   ,   15  ,   13  ,   1   ,   63  ,   33  ,   113 ,   83  ,   513 ,   3903    ,   5399    ,   5855    ,   11151   ,0 };
            const unsigned long dim3629Kuo3Init[]   =   {   1   ,   3   ,   7   ,   7   ,   1   ,   15  ,   55  ,   181 ,   453 ,   15  ,   333 ,   185 ,   2031    ,   14809   ,   6187    ,0 };
            const unsigned long dim3630Kuo3Init[]   =   {   1   ,   1   ,   5   ,   13  ,   27  ,   37  ,   41  ,   197 ,   191 ,   365 ,   1863    ,   559 ,   7731    ,   821 ,   8477    ,0 };
            const unsigned long dim3631Kuo3Init[]   =   {   1   ,   1   ,   1   ,   15  ,   19  ,   53  ,   91  ,   119 ,   185 ,   515 ,   1041    ,   3933    ,   601 ,   2971    ,   25445   ,0 };
            const unsigned long dim3632Kuo3Init[]   =   {   1   ,   3   ,   5   ,   15  ,   25  ,   57  ,   11  ,   79  ,   299 ,   5   ,   1433    ,   2811    ,   1807    ,   4057    ,   3283    ,0 };
            const unsigned long dim3633Kuo3Init[]   =   {   1   ,   1   ,   5   ,   13  ,   17  ,   17  ,   41  ,   157 ,   357 ,   981 ,   1323    ,   1189    ,   1249    ,   2857    ,   10163   ,0 };
            const unsigned long dim3634Kuo3Init[]   =   {   1   ,   3   ,   3   ,   3   ,   15  ,   47  ,   109 ,   249 ,   37  ,   403 ,   1611    ,   281 ,   2521    ,   14179   ,   3095    ,0 };
            const unsigned long dim3635Kuo3Init[]   =   {   1   ,   3   ,   5   ,   7   ,   23  ,   25  ,   5   ,   83  ,   503 ,   203 ,   1961    ,   1915    ,   5111    ,   4265    ,   27243   ,0 };
            const unsigned long dim3636Kuo3Init[]   =   {   1   ,   1   ,   3   ,   1   ,   7   ,   29  ,   43  ,   19  ,   139 ,   487 ,   1847    ,   1569    ,   643 ,   15999   ,   21539   ,0 };
            const unsigned long dim3637Kuo3Init[]   =   {   1   ,   1   ,   7   ,   3   ,   3   ,   7   ,   31  ,   217 ,   103 ,   127 ,   1389    ,   87  ,   7911    ,   12751   ,   11907   ,0 };
            const unsigned long dim3638Kuo3Init[]   =   {   1   ,   3   ,   1   ,   9   ,   31  ,   25  ,   27  ,   69  ,   477 ,   471 ,   1375    ,   2485    ,   4221    ,   175 ,   16841   ,0 };
            const unsigned long dim3639Kuo3Init[]   =   {   1   ,   3   ,   1   ,   15  ,   9   ,   31  ,   125 ,   233 ,   175 ,   649 ,   641 ,   3751    ,   7857    ,   3849    ,   8859    ,0 };
            const unsigned long dim3640Kuo3Init[]   =   {   1   ,   1   ,   1   ,   3   ,   25  ,   7   ,   31  ,   197 ,   435 ,   453 ,   1715    ,   3187    ,   6131    ,   12375   ,   29367   ,0 };
            const unsigned long dim3641Kuo3Init[]   =   {   1   ,   3   ,   5   ,   15  ,   9   ,   41  ,   75  ,   107 ,   205 ,   919 ,   1709    ,   2515    ,   7143    ,   10433   ,   6179    ,0 };
            const unsigned long dim3642Kuo3Init[]   =   {   1   ,   1   ,   5   ,   1   ,   27  ,   11  ,   107 ,   177 ,   17  ,   675 ,   677 ,   547 ,   7101    ,   7789    ,   29465   ,0 };
            const unsigned long dim3643Kuo3Init[]   =   {   1   ,   3   ,   5   ,   3   ,   9   ,   7   ,   123 ,   209 ,   443 ,   221 ,   35  ,   4045    ,   2871    ,   16167   ,   31745   ,0 };
            const unsigned long dim3644Kuo3Init[]   =   {   1   ,   1   ,   7   ,   1   ,   15  ,   59  ,   127 ,   173 ,   163 ,   151 ,   813 ,   967 ,   1165    ,   4147    ,   29355   ,0 };
            const unsigned long dim3645Kuo3Init[]   =   {   1   ,   3   ,   7   ,   13  ,   21  ,   17  ,   33  ,   35  ,   87  ,   897 ,   393 ,   2205    ,   2885    ,   7327    ,   28521   ,0 };
            const unsigned long dim3646Kuo3Init[]   =   {   1   ,   1   ,   5   ,   1   ,   1   ,   47  ,   67  ,   87  ,   371 ,   217 ,   1943    ,   1277    ,   27  ,   7205    ,   9751    ,0 };
            const unsigned long dim3647Kuo3Init[]   =   {   1   ,   1   ,   5   ,   13  ,   1   ,   25  ,   51  ,   103 ,   61  ,   279 ,   1151    ,   3745    ,   4159    ,   1895    ,   15869   ,0 };
            const unsigned long dim3648Kuo3Init[]   =   {   1   ,   3   ,   7   ,   9   ,   5   ,   59  ,   29  ,   83  ,   415 ,   851 ,   1415    ,   3719    ,   4345    ,   13897   ,   22409   ,0 };
            const unsigned long dim3649Kuo3Init[]   =   {   1   ,   1   ,   1   ,   7   ,   29  ,   19  ,   7   ,   23  ,   87  ,   1   ,   2035    ,   2287    ,   4985    ,   1371    ,   9323    ,0 };
            const unsigned long dim3650Kuo3Init[]   =   {   1   ,   1   ,   3   ,   1   ,   17  ,   33  ,   55  ,   49  ,   285 ,   953 ,   131 ,   3679    ,   3811    ,   6125    ,   18827   ,0 };
            const unsigned long dim3651Kuo3Init[]   =   {   1   ,   1   ,   3   ,   3   ,   9   ,   51  ,   61  ,   119 ,   163 ,   829 ,   1579    ,   2047    ,   909 ,   3499    ,   26535   ,0 };
            const unsigned long dim3652Kuo3Init[]   =   {   1   ,   3   ,   3   ,   3   ,   27  ,   49  ,   127 ,   239 ,   193 ,   359 ,   695 ,   1471    ,   6481    ,   3549    ,   22337   ,0 };
            const unsigned long dim3653Kuo3Init[]   =   {   1   ,   3   ,   1   ,   5   ,   29  ,   7   ,   83  ,   57  ,   509 ,   15  ,   391 ,   2287    ,   1023    ,   13993   ,   6073    ,0 };
            const unsigned long dim3654Kuo3Init[]   =   {   1   ,   1   ,   5   ,   15  ,   3   ,   21  ,   23  ,   217 ,   147 ,   511 ,   1817    ,   2425    ,   2347    ,   1433    ,   10625   ,0 };
            const unsigned long dim3655Kuo3Init[]   =   {   1   ,   1   ,   7   ,   13  ,   23  ,   21  ,   29  ,   197 ,   189 ,   367 ,   619 ,   1305    ,   5463    ,   7099    ,   3307    ,0 };
            const unsigned long dim3656Kuo3Init[]   =   {   1   ,   3   ,   1   ,   1   ,   27  ,   7   ,   33  ,   221 ,   107 ,   85  ,   633 ,   3325    ,   3109    ,   16377   ,   25633   ,0 };
            const unsigned long dim3657Kuo3Init[]   =   {   1   ,   1   ,   3   ,   11  ,   13  ,   59  ,   57  ,   251 ,   321 ,   45  ,   451 ,   237 ,   4499    ,   10549   ,   16473   ,0 };
            const unsigned long dim3658Kuo3Init[]   =   {   1   ,   1   ,   7   ,   15  ,   11  ,   39  ,   17  ,   85  ,   233 ,   865 ,   1285    ,   1771    ,   2913    ,   7729    ,   3815    ,0 };
            const unsigned long dim3659Kuo3Init[]   =   {   1   ,   1   ,   5   ,   15  ,   9   ,   1   ,   51  ,   47  ,   113 ,   917 ,   1007    ,   1883    ,   4087    ,   4061    ,   27965   ,0 };
            const unsigned long dim3660Kuo3Init[]   =   {   1   ,   1   ,   5   ,   15  ,   29  ,   33  ,   123 ,   183 ,   347 ,   99  ,   1215    ,   893 ,   7555    ,   6835    ,   1729    ,0 };
            const unsigned long dim3661Kuo3Init[]   =   {   1   ,   1   ,   5   ,   13  ,   5   ,   25  ,   7   ,   255 ,   315 ,   559 ,   51  ,   899 ,   7165    ,   2021    ,   17413   ,0 };
            const unsigned long dim3662Kuo3Init[]   =   {   1   ,   3   ,   3   ,   7   ,   3   ,   19  ,   7   ,   89  ,   77  ,   599 ,   1021    ,   3269    ,   3615    ,   10023   ,   31043   ,0 };
            const unsigned long dim3663Kuo3Init[]   =   {   1   ,   1   ,   7   ,   3   ,   9   ,   7   ,   119 ,   199 ,   315 ,   835 ,   1401    ,   1547    ,   609 ,   9015    ,   27589   ,0 };
            const unsigned long dim3664Kuo3Init[]   =   {   1   ,   3   ,   3   ,   15  ,   7   ,   33  ,   109 ,   239 ,   329 ,   213 ,   215 ,   1989    ,   1495    ,   7765    ,   3123    ,0 };
            const unsigned long dim3665Kuo3Init[]   =   {   1   ,   1   ,   1   ,   13  ,   31  ,   23  ,   79  ,   181 ,   211 ,   189 ,   1   ,   2103    ,   4247    ,   7209    ,   3733    ,0 };
            const unsigned long dim3666Kuo3Init[]   =   {   1   ,   3   ,   3   ,   7   ,   19  ,   31  ,   59  ,   103 ,   291 ,   877 ,   459 ,   3893    ,   4303    ,   10033   ,   31859   ,0 };
            const unsigned long dim3667Kuo3Init[]   =   {   1   ,   3   ,   5   ,   11  ,   11  ,   1   ,   11  ,   163 ,   477 ,   925 ,   659 ,   143 ,   3705    ,   11335   ,   23667   ,   53273   ,0 };
            const unsigned long dim3668Kuo3Init[]   =   {   1   ,   1   ,   7   ,   11  ,   21  ,   29  ,   119 ,   195 ,   303 ,   185 ,   1353    ,   1285    ,   2525    ,   11801   ,   10663   ,   49201   ,0 };
            const unsigned long dim3669Kuo3Init[]   =   {   1   ,   1   ,   7   ,   3   ,   29  ,   41  ,   115 ,   115 ,   389 ,   763 ,   1625    ,   691 ,   3425    ,   6451    ,   3355    ,   26967   ,0 };
            const unsigned long dim3670Kuo3Init[]   =   {   1   ,   1   ,   1   ,   9   ,   1   ,   3   ,   71  ,   129 ,   25  ,   231 ,   1849    ,   2087    ,   3253    ,   13151   ,   17039   ,   21893   ,0 };
            const unsigned long dim3671Kuo3Init[]   =   {   1   ,   3   ,   5   ,   9   ,   11  ,   57  ,   99  ,   187 ,   245 ,   965 ,   1437    ,   3431    ,   4657    ,   12581   ,   14883   ,   54985   ,0 };
            const unsigned long dim3672Kuo3Init[]   =   {   1   ,   3   ,   5   ,   3   ,   27  ,   55  ,   41  ,   203 ,   261 ,   271 ,   327 ,   71  ,   3659    ,   4517    ,   24071   ,   56297   ,0 };
            const unsigned long dim3673Kuo3Init[]   =   {   1   ,   1   ,   5   ,   1   ,   13  ,   63  ,   37  ,   109 ,   437 ,   611 ,   649 ,   1123    ,   4365    ,   10259   ,   26507   ,   60715   ,0 };
            const unsigned long dim3674Kuo3Init[]   =   {   1   ,   1   ,   5   ,   9   ,   21  ,   51  ,   113 ,   95  ,   221 ,   157 ,   843 ,   1429    ,   781 ,   13193   ,   16453   ,   26883   ,0 };
            const unsigned long dim3675Kuo3Init[]   =   {   1   ,   3   ,   3   ,   9   ,   29  ,   15  ,   13  ,   5   ,   441 ,   7   ,   139 ,   773 ,   1495    ,   2283    ,   31549   ,   4105    ,0 };
            const unsigned long dim3676Kuo3Init[]   =   {   1   ,   1   ,   5   ,   9   ,   27  ,   59  ,   51  ,   51  ,   211 ,   871 ,   1971    ,   1213    ,   7459    ,   14001   ,   32175   ,   53975   ,0 };
            const unsigned long dim3677Kuo3Init[]   =   {   1   ,   3   ,   5   ,   3   ,   3   ,   29  ,   47  ,   131 ,   263 ,   879 ,   731 ,   1063    ,   3647    ,   8729    ,   5407    ,   4001    ,0 };
            const unsigned long dim3678Kuo3Init[]   =   {   1   ,   1   ,   3   ,   5   ,   21  ,   45  ,   93  ,   143 ,   443 ,   295 ,   1683    ,   2335    ,   4213    ,   6297    ,   26611   ,   56661   ,0 };
            const unsigned long dim3679Kuo3Init[]   =   {   1   ,   1   ,   5   ,   3   ,   9   ,   63  ,   97  ,   143 ,   101 ,   579 ,   435 ,   3339    ,   7333    ,   6267    ,   1589    ,   30997   ,0 };
            const unsigned long dim3680Kuo3Init[]   =   {   1   ,   1   ,   3   ,   3   ,   31  ,   51  ,   81  ,   125 ,   285 ,   879 ,   1377    ,   107 ,   827 ,   7185    ,   27265   ,   54517   ,0 };
            const unsigned long dim3681Kuo3Init[]   =   {   1   ,   1   ,   3   ,   11  ,   9   ,   9   ,   121 ,   167 ,   195 ,   625 ,   1167    ,   1533    ,   3697    ,   7545    ,   27799   ,   11613   ,0 };
            const unsigned long dim3682Kuo3Init[]   =   {   1   ,   3   ,   7   ,   7   ,   21  ,   55  ,   65  ,   63  ,   355 ,   525 ,   1441    ,   3057    ,   1779    ,   1613    ,   6277    ,   37655   ,0 };
            const unsigned long dim3683Kuo3Init[]   =   {   1   ,   1   ,   7   ,   15  ,   9   ,   61  ,   7   ,   211 ,   11  ,   955 ,   499 ,   2795    ,   4017    ,   1459    ,   4323    ,   1453    ,0 };
            const unsigned long dim3684Kuo3Init[]   =   {   1   ,   1   ,   1   ,   9   ,   25  ,   17  ,   75  ,   219 ,   383 ,   721 ,   1661    ,   3325    ,   5917    ,   11423   ,   19265   ,   41063   ,0 };
            const unsigned long dim3685Kuo3Init[]   =   {   1   ,   3   ,   1   ,   7   ,   15  ,   39  ,   93  ,   9   ,   463 ,   611 ,   547 ,   173 ,   4179    ,   4641    ,   28201   ,   32139   ,0 };
            const unsigned long dim3686Kuo3Init[]   =   {   1   ,   1   ,   5   ,   7   ,   29  ,   51  ,   41  ,   157 ,   93  ,   289 ,   935 ,   3299    ,   2103    ,   4371    ,   15485   ,   62685   ,0 };
            const unsigned long dim3687Kuo3Init[]   =   {   1   ,   3   ,   3   ,   15  ,   31  ,   49  ,   93  ,   205 ,   13  ,   551 ,   1809    ,   4051    ,   5481    ,   2693    ,   6615    ,   18123   ,0 };
            const unsigned long dim3688Kuo3Init[]   =   {   1   ,   3   ,   7   ,   3   ,   1   ,   41  ,   27  ,   105 ,   469 ,   759 ,   625 ,   1579    ,   4253    ,   7435    ,   20545   ,   4365    ,0 };
            const unsigned long dim3689Kuo3Init[]   =   {   1   ,   3   ,   7   ,   11  ,   17  ,   45  ,   111 ,   91  ,   357 ,   651 ,   5   ,   937 ,   3199    ,   6485    ,   21145   ,   35085   ,0 };
            const unsigned long dim3690Kuo3Init[]   =   {   1   ,   3   ,   5   ,   1   ,   7   ,   21  ,   33  ,   213 ,   399 ,   735 ,   1923    ,   1047    ,   2765    ,   11057   ,   19007   ,   58345   ,0 };
            const unsigned long dim3691Kuo3Init[]   =   {   1   ,   3   ,   7   ,   13  ,   31  ,   57  ,   113 ,   81  ,   75  ,   985 ,   479 ,   1353    ,   1041    ,   13489   ,   12011   ,   52749   ,0 };
            const unsigned long dim3692Kuo3Init[]   =   {   1   ,   1   ,   1   ,   9   ,   23  ,   23  ,   83  ,   167 ,   473 ,   471 ,   171 ,   3303    ,   7167    ,   1109    ,   29263   ,   30047   ,0 };
            const unsigned long dim3693Kuo3Init[]   =   {   1   ,   3   ,   7   ,   15  ,   11  ,   63  ,   57  ,   73  ,   49  ,   921 ,   457 ,   1683    ,   4447    ,   14549   ,   5819    ,   18291   ,0 };
            const unsigned long dim3694Kuo3Init[]   =   {   1   ,   3   ,   5   ,   5   ,   27  ,   1   ,   127 ,   233 ,   509 ,   229 ,   1553    ,   3875    ,   2243    ,   16005   ,   17985   ,   28437   ,0 };
            const unsigned long dim3695Kuo3Init[]   =   {   1   ,   1   ,   1   ,   15  ,   7   ,   27  ,   75  ,   191 ,   433 ,   585 ,   1659    ,   3373    ,   4067    ,   55  ,   11095   ,   37179   ,0 };
            const unsigned long dim3696Kuo3Init[]   =   {   1   ,   1   ,   7   ,   15  ,   27  ,   31  ,   21  ,   105 ,   95  ,   357 ,   2019    ,   2193    ,   741 ,   8509    ,   349 ,   24797   ,0 };
            const unsigned long dim3697Kuo3Init[]   =   {   1   ,   1   ,   1   ,   15  ,   1   ,   51  ,   77  ,   139 ,   205 ,   81  ,   1849    ,   1687    ,   1597    ,   1531    ,   115 ,   50163   ,0 };
            const unsigned long dim3698Kuo3Init[]   =   {   1   ,   3   ,   5   ,   5   ,   7   ,   61  ,   113 ,   111 ,   69  ,   443 ,   179 ,   2955    ,   7   ,   13237   ,   26787   ,   33917   ,0 };
            const unsigned long dim3699Kuo3Init[]   =   {   1   ,   3   ,   5   ,   5   ,   21  ,   37  ,   11  ,   209 ,   177 ,   433 ,   1423    ,   1629    ,   5807    ,   4655    ,   19377   ,   58963   ,0 };
            const unsigned long dim3700Kuo3Init[]   =   {   1   ,   3   ,   5   ,   1   ,   11  ,   29  ,   107 ,   17  ,   135 ,   801 ,   1441    ,   3119    ,   4573    ,   8903    ,   11267   ,   2845    ,0 };
            const unsigned long dim3701Kuo3Init[]   =   {   1   ,   1   ,   3   ,   3   ,   21  ,   41  ,   115 ,   19  ,   473 ,   447 ,   1879    ,   4095    ,   2743    ,   10007   ,   5751    ,   36845   ,0 };
            const unsigned long dim3702Kuo3Init[]   =   {   1   ,   1   ,   3   ,   5   ,   19  ,   13  ,   85  ,   131 ,   85  ,   849 ,   1179    ,   1919    ,   3631    ,   13807   ,   10467   ,   24081   ,0 };
            const unsigned long dim3703Kuo3Init[]   =   {   1   ,   1   ,   5   ,   13  ,   5   ,   17  ,   21  ,   125 ,   169 ,   369 ,   761 ,   1419    ,   7795    ,   9041    ,   7613    ,   6125    ,0 };
            const unsigned long dim3704Kuo3Init[]   =   {   1   ,   3   ,   7   ,   9   ,   21  ,   53  ,   119 ,   55  ,   97  ,   563 ,   1531    ,   847 ,   3709    ,   10573   ,   18469   ,   29625   ,0 };
            const unsigned long dim3705Kuo3Init[]   =   {   1   ,   1   ,   5   ,   7   ,   23  ,   15  ,   99  ,   251 ,   415 ,   55  ,   563 ,   3535    ,   8017    ,   7771    ,   23601   ,   52955   ,0 };
            const unsigned long dim3706Kuo3Init[]   =   {   1   ,   3   ,   5   ,   15  ,   29  ,   11  ,   73  ,   251 ,   139 ,   215 ,   825 ,   1735    ,   7859    ,   10171   ,   17351   ,   24971   ,0 };
            const unsigned long dim3707Kuo3Init[]   =   {   1   ,   3   ,   1   ,   1   ,   21  ,   21  ,   21  ,   21  ,   107 ,   901 ,   1321    ,   647 ,   7521    ,   5505    ,   965 ,   36217   ,0 };
            const unsigned long dim3708Kuo3Init[]   =   {   1   ,   1   ,   5   ,   15  ,   9   ,   39  ,   87  ,   173 ,   201 ,   705 ,   563 ,   3479    ,   437 ,   1733    ,   18001   ,   39161   ,0 };
            const unsigned long dim3709Kuo3Init[]   =   {   1   ,   3   ,   7   ,   15  ,   31  ,   31  ,   89  ,   119 ,   471 ,   633 ,   249 ,   187 ,   3735    ,   4045    ,   1611    ,   10045   ,0 };
            const unsigned long dim3710Kuo3Init[]   =   {   1   ,   3   ,   5   ,   1   ,   7   ,   31  ,   117 ,   249 ,   175 ,   515 ,   1825    ,   3417    ,   307 ,   10909   ,   4453    ,   43833   ,0 };
            const unsigned long dim3711Kuo3Init[]   =   {   1   ,   3   ,   1   ,   9   ,   1   ,   21  ,   59  ,   103 ,   243 ,   105 ,   2037    ,   1759    ,   689 ,   15389   ,   16981   ,   35545   ,0 };
            const unsigned long dim3712Kuo3Init[]   =   {   1   ,   1   ,   7   ,   1   ,   13  ,   15  ,   105 ,   89  ,   463 ,   513 ,   191 ,   1489    ,   5239    ,   3723    ,   20511   ,   52139   ,0 };
            const unsigned long dim3713Kuo3Init[]   =   {   1   ,   3   ,   5   ,   1   ,   1   ,   25  ,   71  ,   89  ,   167 ,   1009    ,   829 ,   2267    ,   375 ,   10217   ,   31677   ,   58309   ,0 };
            const unsigned long dim3714Kuo3Init[]   =   {   1   ,   3   ,   5   ,   1   ,   31  ,   19  ,   21  ,   225 ,   331 ,   621 ,   1047    ,   3385    ,   5247    ,   1763    ,   19985   ,   9013    ,0 };
            const unsigned long dim3715Kuo3Init[]   =   {   1   ,   3   ,   1   ,   1   ,   5   ,   19  ,   69  ,   11  ,   15  ,   361 ,   617 ,   823 ,   2645    ,   207 ,   18775   ,   60509   ,0 };
            const unsigned long dim3716Kuo3Init[]   =   {   1   ,   3   ,   3   ,   13  ,   15  ,   1   ,   37  ,   3   ,   249 ,   277 ,   1729    ,   3599    ,   115 ,   9557    ,   12865   ,   6257    ,0 };
            const unsigned long dim3717Kuo3Init[]   =   {   1   ,   3   ,   7   ,   11  ,   15  ,   51  ,   21  ,   147 ,   23  ,   527 ,   167 ,   1673    ,   2299    ,   923 ,   22755   ,   48627   ,0 };
            const unsigned long dim3718Kuo3Init[]   =   {   1   ,   3   ,   7   ,   1   ,   17  ,   55  ,   23  ,   1   ,   213 ,   479 ,   287 ,   2741    ,   7255    ,   4571    ,   6581    ,   57509   ,0 };
            const unsigned long dim3719Kuo3Init[]   =   {   1   ,   3   ,   5   ,   7   ,   15  ,   49  ,   13  ,   149 ,   103 ,   513 ,   1977    ,   2275    ,   5535    ,   7867    ,   16305   ,   24713   ,0 };
            const unsigned long dim3720Kuo3Init[]   =   {   1   ,   1   ,   7   ,   1   ,   29  ,   41  ,   69  ,   211 ,   131 ,   87  ,   859 ,   3631    ,   611 ,   11463   ,   18029   ,   38771   ,0 };
            const unsigned long dim3721Kuo3Init[]   =   {   1   ,   3   ,   5   ,   11  ,   17  ,   23  ,   37  ,   149 ,   271 ,   547 ,   1129    ,   3615    ,   6647    ,   7023    ,   103 ,   29257   ,0 };
            const unsigned long dim3722Kuo3Init[]   =   {   1   ,   1   ,   7   ,   13  ,   1   ,   55  ,   71  ,   51  ,   249 ,   141 ,   1937    ,   3869    ,   5897    ,   13227   ,   22279   ,   11733   ,0 };
            const unsigned long dim3723Kuo3Init[]   =   {   1   ,   1   ,   5   ,   13  ,   1   ,   29  ,   85  ,   255 ,   93  ,   741 ,   687 ,   2195    ,   519 ,   8235    ,   21305   ,   19681   ,0 };
            const unsigned long dim3724Kuo3Init[]   =   {   1   ,   1   ,   7   ,   5   ,   13  ,   17  ,   103 ,   65  ,   211 ,   603 ,   543 ,   3629    ,   235 ,   7337    ,   27477   ,   2735    ,0 };
            const unsigned long dim3725Kuo3Init[]   =   {   1   ,   3   ,   1   ,   1   ,   17  ,   63  ,   75  ,   183 ,   107 ,   583 ,   1849    ,   545 ,   7987    ,   13679   ,   22797   ,   40095   ,0 };
            const unsigned long dim3726Kuo3Init[]   =   {   1   ,   1   ,   5   ,   1   ,   31  ,   13  ,   33  ,   21  ,   479 ,   879 ,   1223    ,   3925    ,   3003    ,   9975    ,   23465   ,   27775   ,0 };
            const unsigned long dim3727Kuo3Init[]   =   {   1   ,   1   ,   3   ,   1   ,   7   ,   23  ,   65  ,   107 ,   239 ,   867 ,   1245    ,   1549    ,   5981    ,   4677    ,   3761    ,   61449   ,0 };
            const unsigned long dim3728Kuo3Init[]   =   {   1   ,   3   ,   7   ,   9   ,   5   ,   31  ,   25  ,   53  ,   59  ,   79  ,   1545    ,   3721    ,   6979    ,   16351   ,   7305    ,   41653   ,0 };
            const unsigned long dim3729Kuo3Init[]   =   {   1   ,   1   ,   1   ,   13  ,   27  ,   53  ,   9   ,   7   ,   159 ,   353 ,   121 ,   2525    ,   8049    ,   12997   ,   3933    ,   19835   ,0 };
            const unsigned long dim3730Kuo3Init[]   =   {   1   ,   3   ,   3   ,   15  ,   1   ,   33  ,   127 ,   163 ,   399 ,   125 ,   1869    ,   569 ,   3853    ,   4819    ,   4403    ,   34367   ,0 };
            const unsigned long dim3731Kuo3Init[]   =   {   1   ,   3   ,   5   ,   15  ,   11  ,   23  ,   21  ,   95  ,   329 ,   937 ,   689 ,   3731    ,   1157    ,   9613    ,   12593   ,   58609   ,0 };
            const unsigned long dim3732Kuo3Init[]   =   {   1   ,   3   ,   5   ,   7   ,   13  ,   23  ,   7   ,   237 ,   413 ,   139 ,   523 ,   3517    ,   6829    ,   17  ,   17673   ,   57889   ,0 };
            const unsigned long dim3733Kuo3Init[]   =   {   1   ,   1   ,   3   ,   3   ,   3   ,   49  ,   63  ,   87  ,   147 ,   359 ,   2027    ,   1513    ,   3933    ,   13911   ,   25459   ,   11391   ,0 };
            const unsigned long dim3734Kuo3Init[]   =   {   1   ,   1   ,   1   ,   5   ,   23  ,   17  ,   81  ,   249 ,   487 ,   71  ,   395 ,   3635    ,   4773    ,   14909   ,   13523   ,   32617   ,0 };
            const unsigned long dim3735Kuo3Init[]   =   {   1   ,   1   ,   7   ,   5   ,   19  ,   39  ,   13  ,   91  ,   475 ,   899 ,   353 ,   3275    ,   4161    ,   2203    ,   20375   ,   14795   ,0 };
            const unsigned long dim3736Kuo3Init[]   =   {   1   ,   3   ,   3   ,   11  ,   17  ,   15  ,   117 ,   145 ,   223 ,   393 ,   1267    ,   1089    ,   6745    ,   10485   ,   20385   ,   25991   ,0 };
            const unsigned long dim3737Kuo3Init[]   =   {   1   ,   1   ,   7   ,   7   ,   27  ,   21  ,   57  ,   237 ,   405 ,   513 ,   1411    ,   2791    ,   1685    ,   11529   ,   2115    ,   36631   ,0 };
            const unsigned long dim3738Kuo3Init[]   =   {   1   ,   1   ,   1   ,   13  ,   5   ,   17  ,   111 ,   55  ,   277 ,   639 ,   1541    ,   2347    ,   1119    ,   10101   ,   8533    ,   15383   ,0 };
            const unsigned long dim3739Kuo3Init[]   =   {   1   ,   3   ,   5   ,   9   ,   1   ,   47  ,   123 ,   39  ,   473 ,   471 ,   299 ,   3785    ,   6589    ,   9415    ,   29193   ,   24367   ,0 };
            const unsigned long dim3740Kuo3Init[]   =   {   1   ,   3   ,   3   ,   9   ,   13  ,   17  ,   9   ,   131 ,   9   ,   309 ,   1529    ,   3721    ,   5927    ,   6699    ,   17439   ,   60313   ,0 };
            const unsigned long dim3741Kuo3Init[]   =   {   1   ,   3   ,   3   ,   7   ,   17  ,   3   ,   93  ,   33  ,   287 ,   47  ,   1295    ,   1021    ,   4267    ,   403 ,   31111   ,   5399    ,0 };
            const unsigned long dim3742Kuo3Init[]   =   {   1   ,   3   ,   5   ,   3   ,   29  ,   27  ,   73  ,   117 ,   123 ,   1007    ,   2003    ,   2921    ,   4875    ,   2043    ,   12221   ,   57865   ,0 };
            const unsigned long dim3743Kuo3Init[]   =   {   1   ,   3   ,   1   ,   15  ,   15  ,   51  ,   31  ,   217 ,   429 ,   1009    ,   1761    ,   389 ,   4859    ,   13667   ,   17955   ,   29493   ,0 };
            const unsigned long dim3744Kuo3Init[]   =   {   1   ,   3   ,   3   ,   7   ,   25  ,   11  ,   65  ,   109 ,   241 ,   421 ,   363 ,   2641    ,   781 ,   711 ,   5363    ,   9603    ,0 };
            const unsigned long dim3745Kuo3Init[]   =   {   1   ,   3   ,   1   ,   11  ,   1   ,   41  ,   99  ,   181 ,   103 ,   503 ,   1887    ,   1917    ,   4791    ,   6405    ,   27701   ,   62353   ,0 };
            const unsigned long dim3746Kuo3Init[]   =   {   1   ,   1   ,   5   ,   7   ,   25  ,   43  ,   23  ,   49  ,   333 ,   205 ,   903 ,   3915    ,   2899    ,   2719    ,   7887    ,   17313   ,0 };
            const unsigned long dim3747Kuo3Init[]   =   {   1   ,   3   ,   5   ,   3   ,   27  ,   19  ,   59  ,   125 ,   409 ,   979 ,   1203    ,   1737    ,   5269    ,   14755   ,   32173   ,   2127    ,0 };
            const unsigned long dim3748Kuo3Init[]   =   {   1   ,   1   ,   7   ,   11  ,   29  ,   19  ,   73  ,   63  ,   81  ,   717 ,   933 ,   2939    ,   2863    ,   5993    ,   25953   ,   31733   ,0 };
            const unsigned long dim3749Kuo3Init[]   =   {   1   ,   3   ,   7   ,   5   ,   11  ,   19  ,   55  ,   223 ,   51  ,   661 ,   547 ,   4055    ,   4849    ,   3009    ,   22303   ,   25  ,0 };
            const unsigned long dim3750Kuo3Init[]   =   {   1   ,   1   ,   7   ,   5   ,   7   ,   3   ,   55  ,   167 ,   207 ,   819 ,   791 ,   3813    ,   2455    ,   1305    ,   4661    ,   51189   ,0 };
            const unsigned long dim3751Kuo3Init[]   =   {   1   ,   3   ,   7   ,   5   ,   29  ,   1   ,   31  ,   191 ,   267 ,   705 ,   181 ,   3091    ,   4725    ,   12351   ,   31521   ,   56163   ,0 };
            const unsigned long dim3752Kuo3Init[]   =   {   1   ,   3   ,   3   ,   1   ,   19  ,   11  ,   27  ,   215 ,   249 ,   713 ,   723 ,   143 ,   2179    ,   11075   ,   22059   ,   62629   ,0 };
            const unsigned long dim3753Kuo3Init[]   =   {   1   ,   1   ,   3   ,   3   ,   23  ,   45  ,   101 ,   25  ,   115 ,   353 ,   565 ,   3491    ,   285 ,   2133    ,   27117   ,   42699   ,0 };
            const unsigned long dim3754Kuo3Init[]   =   {   1   ,   1   ,   5   ,   7   ,   1   ,   51  ,   113 ,   165 ,   157 ,   431 ,   1043    ,   3003    ,   5971    ,   15699   ,   20799   ,   13989   ,0 };
            const unsigned long dim3755Kuo3Init[]   =   {   1   ,   3   ,   7   ,   5   ,   23  ,   21  ,   121 ,   85  ,   69  ,   169 ,   1761    ,   1023    ,   5501    ,   1255    ,   25683   ,   48699   ,0 };
            const unsigned long dim3756Kuo3Init[]   =   {   1   ,   3   ,   5   ,   13  ,   21  ,   31  ,   95  ,   5   ,   237 ,   607 ,   1525    ,   763 ,   4725    ,   4221    ,   22769   ,   40899   ,0 };
            const unsigned long dim3757Kuo3Init[]   =   {   1   ,   1   ,   5   ,   1   ,   1   ,   9   ,   105 ,   225 ,   171 ,   999 ,   547 ,   2331    ,   4389    ,   12847   ,   18107   ,   11939   ,0 };
            const unsigned long dim3758Kuo3Init[]   =   {   1   ,   3   ,   5   ,   5   ,   15  ,   47  ,   73  ,   237 ,   463 ,   501 ,   1611    ,   2259    ,   1679    ,   1709    ,   24527   ,   53529   ,0 };
            const unsigned long dim3759Kuo3Init[]   =   {   1   ,   3   ,   1   ,   7   ,   21  ,   27  ,   27  ,   25  ,   463 ,   73  ,   2021    ,   3899    ,   487 ,   667 ,   5063    ,   5017    ,0 };
            const unsigned long dim3760Kuo3Init[]   =   {   1   ,   1   ,   3   ,   3   ,   15  ,   3   ,   87  ,   145 ,   45  ,   473 ,   1455    ,   1857    ,   2379    ,   4805    ,   10557   ,   27485   ,0 };
            const unsigned long dim3761Kuo3Init[]   =   {   1   ,   1   ,   3   ,   13  ,   17  ,   15  ,   13  ,   237 ,   177 ,   835 ,   1019    ,   975 ,   7085    ,   2871    ,   16473   ,   23693   ,0 };
            const unsigned long dim3762Kuo3Init[]   =   {   1   ,   1   ,   1   ,   1   ,   13  ,   61  ,   31  ,   197 ,   127 ,   213 ,   173 ,   3751    ,   3819    ,   2899    ,   22675   ,   23095   ,0 };
            const unsigned long dim3763Kuo3Init[]   =   {   1   ,   3   ,   1   ,   15  ,   7   ,   63  ,   89  ,   95  ,   497 ,   1021    ,   1751    ,   1379    ,   3809    ,   9493    ,   7941    ,   35217   ,0 };
            const unsigned long dim3764Kuo3Init[]   =   {   1   ,   3   ,   7   ,   3   ,   7   ,   63  ,   33  ,   155 ,   171 ,   651 ,   773 ,   1861    ,   4225    ,   1081    ,   29141   ,   53413   ,0 };
            const unsigned long dim3765Kuo3Init[]   =   {   1   ,   1   ,   3   ,   7   ,   15  ,   27  ,   31  ,   183 ,   377 ,   341 ,   205 ,   1817    ,   6185    ,   13047   ,   19585   ,   1533    ,0 };
            const unsigned long dim3766Kuo3Init[]   =   {   1   ,   1   ,   1   ,   7   ,   27  ,   23  ,   23  ,   117 ,   463 ,   73  ,   529 ,   1727    ,   6989    ,   15911   ,   11783   ,   1011    ,0 };
            const unsigned long dim3767Kuo3Init[]   =   {   1   ,   1   ,   3   ,   7   ,   9   ,   43  ,   55  ,   191 ,   203 ,   471 ,   515 ,   1315    ,   7911    ,   14281   ,   4767    ,   15725   ,0 };
            const unsigned long dim3768Kuo3Init[]   =   {   1   ,   1   ,   1   ,   15  ,   23  ,   13  ,   61  ,   231 ,   423 ,   79  ,   421 ,   1619    ,   7951    ,   15541   ,   20985   ,   42181   ,0 };
            const unsigned long dim3769Kuo3Init[]   =   {   1   ,   1   ,   7   ,   3   ,   31  ,   29  ,   81  ,   125 ,   217 ,   643 ,   449 ,   477 ,   4521    ,   2973    ,   13295   ,   943 ,0 };
            const unsigned long dim3770Kuo3Init[]   =   {   1   ,   3   ,   5   ,   3   ,   15  ,   37  ,   81  ,   167 ,   11  ,   53  ,   1455    ,   3719    ,   1777    ,   7253    ,   16365   ,   17021   ,0 };
            const unsigned long dim3771Kuo3Init[]   =   {   1   ,   3   ,   7   ,   5   ,   25  ,   39  ,   95  ,   189 ,   453 ,   697 ,   1761    ,   3837    ,   4697    ,   8429    ,   10455   ,   17239   ,0 };
            const unsigned long dim3772Kuo3Init[]   =   {   1   ,   1   ,   3   ,   5   ,   13  ,   43  ,   127 ,   251 ,   497 ,   475 ,   105 ,   3137    ,   5927    ,   7569    ,   5645    ,   25315   ,0 };
            const unsigned long dim3773Kuo3Init[]   =   {   1   ,   3   ,   7   ,   9   ,   1   ,   23  ,   15  ,   129 ,   165 ,   887 ,   1719    ,   2343    ,   1663    ,   1153    ,   25889   ,   10961   ,0 };
            const unsigned long dim3774Kuo3Init[]   =   {   1   ,   3   ,   1   ,   13  ,   29  ,   7   ,   37  ,   147 ,   369 ,   697 ,   481 ,   1845    ,   4499    ,   4885    ,   3201    ,   6773    ,0 };
            const unsigned long dim3775Kuo3Init[]   =   {   1   ,   1   ,   1   ,   3   ,   25  ,   23  ,   49  ,   149 ,   435 ,   561 ,   1233    ,   2821    ,   3121    ,   2825    ,   21237   ,   47491   ,0 };
            const unsigned long dim3776Kuo3Init[]   =   {   1   ,   1   ,   3   ,   3   ,   3   ,   59  ,   85  ,   205 ,   89  ,   423 ,   701 ,   1179    ,   4875    ,   3351    ,   17863   ,   35111   ,0 };
            const unsigned long dim3777Kuo3Init[]   =   {   1   ,   1   ,   3   ,   1   ,   5   ,   51  ,   105 ,   85  ,   411 ,   561 ,   155 ,   2809    ,   63  ,   8711    ,   12523   ,   28827   ,0 };
            const unsigned long dim3778Kuo3Init[]   =   {   1   ,   3   ,   7   ,   15  ,   29  ,   43  ,   103 ,   153 ,   105 ,   819 ,   1861    ,   1725    ,   2225    ,   1821    ,   25551   ,   65399   ,0 };
            const unsigned long dim3779Kuo3Init[]   =   {   1   ,   1   ,   1   ,   13  ,   7   ,   51  ,   71  ,   177 ,   49  ,   605 ,   1375    ,   737 ,   7325    ,   9597    ,   10239   ,   64761   ,0 };
            const unsigned long dim3780Kuo3Init[]   =   {   1   ,   3   ,   1   ,   3   ,   5   ,   15  ,   117 ,   187 ,   19  ,   149 ,   119 ,   3375    ,   3203    ,   6827    ,   22073   ,   35757   ,0 };
            const unsigned long dim3781Kuo3Init[]   =   {   1   ,   1   ,   1   ,   5   ,   13  ,   25  ,   7   ,   157 ,   23  ,   741 ,   465 ,   2225    ,   6697    ,   14371   ,   25115   ,   29699   ,0 };
            const unsigned long dim3782Kuo3Init[]   =   {   1   ,   3   ,   7   ,   13  ,   5   ,   29  ,   95  ,   209 ,   123 ,   947 ,   1631    ,   423 ,   1217    ,   12353   ,   7561    ,   1415    ,0 };
            const unsigned long dim3783Kuo3Init[]   =   {   1   ,   3   ,   7   ,   11  ,   3   ,   17  ,   63  ,   65  ,   371 ,   355 ,   1793    ,   1197    ,   2375    ,   3629    ,   13283   ,   41727   ,0 };
            const unsigned long dim3784Kuo3Init[]   =   {   1   ,   3   ,   1   ,   13  ,   5   ,   19  ,   103 ,   85  ,   271 ,   123 ,   1939    ,   2741    ,   5105    ,   8693    ,   4291    ,   49389   ,0 };
            const unsigned long dim3785Kuo3Init[]   =   {   1   ,   3   ,   3   ,   1   ,   31  ,   1   ,   57  ,   13  ,   265 ,   941 ,   1169    ,   3639    ,   5317    ,   10565   ,   30263   ,   64329   ,0 };
            const unsigned long dim3786Kuo3Init[]   =   {   1   ,   1   ,   7   ,   13  ,   7   ,   9   ,   115 ,   111 ,   301 ,   365 ,   1   ,   3093    ,   3209    ,   2607    ,   5749    ,   17047   ,0 };
            const unsigned long dim3787Kuo3Init[]   =   {   1   ,   3   ,   1   ,   1   ,   5   ,   41  ,   113 ,   73  ,   407 ,   427 ,   1271    ,   1361    ,   4869    ,   14845   ,   29515   ,   38769   ,0 };
            const unsigned long dim3788Kuo3Init[]   =   {   1   ,   3   ,   5   ,   13  ,   19  ,   47  ,   77  ,   27  ,   57  ,   43  ,   1633    ,   197 ,   5465    ,   14735   ,   25187   ,   38621   ,0 };
            const unsigned long dim3789Kuo3Init[]   =   {   1   ,   3   ,   1   ,   5   ,   11  ,   55  ,   23  ,   213 ,   495 ,   929 ,   697 ,   2829    ,   4185    ,   8321    ,   12673   ,   22759   ,0 };
            const unsigned long dim3790Kuo3Init[]   =   {   1   ,   1   ,   3   ,   9   ,   17  ,   29  ,   7   ,   41  ,   75  ,   139 ,   1165    ,   3387    ,   6015    ,   3679    ,   8353    ,   29981   ,0 };
            const unsigned long dim3791Kuo3Init[]   =   {   1   ,   3   ,   3   ,   1   ,   21  ,   47  ,   95  ,   163 ,   121 ,   877 ,   245 ,   3947    ,   3839    ,   15097   ,   4695    ,   26037   ,0 };
            const unsigned long dim3792Kuo3Init[]   =   {   1   ,   1   ,   7   ,   11  ,   19  ,   19  ,   7   ,   93  ,   77  ,   887 ,   153 ,   237 ,   543 ,   2631    ,   22023   ,   17207   ,0 };
            const unsigned long dim3793Kuo3Init[]   =   {   1   ,   1   ,   7   ,   13  ,   7   ,   23  ,   39  ,   181 ,   107 ,   313 ,   131 ,   3799    ,   981 ,   11145   ,   10969   ,   28733   ,0 };
            const unsigned long dim3794Kuo3Init[]   =   {   1   ,   3   ,   1   ,   11  ,   15  ,   45  ,   105 ,   187 ,   457 ,   41  ,   1739    ,   3229    ,   4289    ,   107 ,   2685    ,   15601   ,0 };
            const unsigned long dim3795Kuo3Init[]   =   {   1   ,   1   ,   1   ,   9   ,   15  ,   53  ,   11  ,   87  ,   421 ,   733 ,   1031    ,   121 ,   2413    ,   3383    ,   6525    ,   35781   ,0 };
            const unsigned long dim3796Kuo3Init[]   =   {   1   ,   3   ,   3   ,   11  ,   23  ,   15  ,   81  ,   13  ,   215 ,   573 ,   1953    ,   379 ,   6857    ,   2137    ,   2599    ,   29589   ,0 };
            const unsigned long dim3797Kuo3Init[]   =   {   1   ,   3   ,   3   ,   9   ,   29  ,   17  ,   25  ,   121 ,   291 ,   317 ,   81  ,   4073    ,   7111    ,   5045    ,   32685   ,   21527   ,0 };
            const unsigned long dim3798Kuo3Init[]   =   {   1   ,   1   ,   3   ,   11  ,   31  ,   53  ,   115 ,   179 ,   81  ,   919 ,   955 ,   2301    ,   647 ,   11433   ,   17761   ,   8187    ,0 };
            const unsigned long dim3799Kuo3Init[]   =   {   1   ,   1   ,   5   ,   9   ,   17  ,   15  ,   105 ,   55  ,   303 ,   505 ,   223 ,   1757    ,   3567    ,   9079    ,   29415   ,   37845   ,0 };
            const unsigned long dim3800Kuo3Init[]   =   {   1   ,   3   ,   7   ,   1   ,   23  ,   33  ,   33  ,   199 ,   119 ,   911 ,   349 ,   3641    ,   5339    ,   12803   ,   27979   ,   17951   ,0 };
            const unsigned long dim3801Kuo3Init[]   =   {   1   ,   1   ,   3   ,   13  ,   15  ,   53  ,   43  ,   25  ,   31  ,   609 ,   49  ,   1307    ,   6477    ,   2173    ,   32245   ,   34123   ,0 };
            const unsigned long dim3802Kuo3Init[]   =   {   1   ,   3   ,   1   ,   13  ,   11  ,   35  ,   55  ,   85  ,   381 ,   733 ,   1407    ,   2821    ,   1547    ,   9843    ,   13861   ,   13529   ,0 };
            const unsigned long dim3803Kuo3Init[]   =   {   1   ,   1   ,   1   ,   5   ,   13  ,   11  ,   35  ,   255 ,   421 ,   673 ,   387 ,   3429    ,   3137    ,   13619   ,   28689   ,   11145   ,0 };
            const unsigned long dim3804Kuo3Init[]   =   {   1   ,   3   ,   3   ,   9   ,   23  ,   61  ,   81  ,   25  ,   163 ,   595 ,   769 ,   3455    ,   5371    ,   9711    ,   14937   ,   34383   ,0 };
            const unsigned long dim3805Kuo3Init[]   =   {   1   ,   3   ,   7   ,   13  ,   11  ,   11  ,   115 ,   93  ,   307 ,   807 ,   1667    ,   483 ,   4815    ,   6305    ,   5197    ,   281 ,0 };
            const unsigned long dim3806Kuo3Init[]   =   {   1   ,   1   ,   7   ,   15  ,   29  ,   31  ,   113 ,   129 ,   483 ,   369 ,   855 ,   643 ,   7503    ,   1699    ,   20681   ,   51497   ,0 };
            const unsigned long dim3807Kuo3Init[]   =   {   1   ,   1   ,   3   ,   9   ,   3   ,   41  ,   7   ,   255 ,   171 ,   1003    ,   559 ,   775 ,   5333    ,   11563   ,   22473   ,   44299   ,0 };
            const unsigned long dim3808Kuo3Init[]   =   {   1   ,   1   ,   5   ,   15  ,   23  ,   51  ,   101 ,   161 ,   433 ,   757 ,   713 ,   403 ,   1789    ,   5353    ,   28901   ,   14249   ,0 };
            const unsigned long dim3809Kuo3Init[]   =   {   1   ,   3   ,   7   ,   9   ,   27  ,   5   ,   9   ,   71  ,   47  ,   287 ,   369 ,   653 ,   7595    ,   10127   ,   12547   ,   34119   ,0 };
            const unsigned long dim3810Kuo3Init[]   =   {   1   ,   1   ,   3   ,   7   ,   5   ,   33  ,   19  ,   183 ,   497 ,   209 ,   1999    ,   2897    ,   2553    ,   633 ,   7413    ,   7847    ,0 };
            const unsigned long dim3811Kuo3Init[]   =   {   1   ,   3   ,   3   ,   13  ,   27  ,   19  ,   73  ,   181 ,   15  ,   297 ,   1485    ,   2167    ,   5033    ,   11191   ,   24999   ,   30177   ,0 };
            const unsigned long dim3812Kuo3Init[]   =   {   1   ,   1   ,   1   ,   13  ,   5   ,   47  ,   111 ,   239 ,   259 ,   1019    ,   241 ,   2521    ,   1599    ,   4655    ,   5725    ,   56623   ,0 };
            const unsigned long dim3813Kuo3Init[]   =   {   1   ,   3   ,   1   ,   7   ,   29  ,   37  ,   29  ,   223 ,   413 ,   307 ,   939 ,   2009    ,   3961    ,   9481    ,   9121    ,   23927   ,0 };
            const unsigned long dim3814Kuo3Init[]   =   {   1   ,   1   ,   3   ,   13  ,   17  ,   37  ,   7   ,   121 ,   387 ,   573 ,   1751    ,   3431    ,   7199    ,   15041   ,   14453   ,   45757   ,0 };
            const unsigned long dim3815Kuo3Init[]   =   {   1   ,   3   ,   3   ,   7   ,   31  ,   53  ,   29  ,   247 ,   23  ,   579 ,   1919    ,   2925    ,   321 ,   653 ,   25925   ,   49439   ,0 };
            const unsigned long dim3816Kuo3Init[]   =   {   1   ,   3   ,   5   ,   5   ,   5   ,   55  ,   109 ,   193 ,   125 ,   823 ,   1707    ,   2389    ,   4135    ,   5197    ,   12563   ,   5977    ,0 };
            const unsigned long dim3817Kuo3Init[]   =   {   1   ,   3   ,   3   ,   7   ,   25  ,   47  ,   1   ,   145 ,   491 ,   661 ,   1125    ,   2567    ,   6783    ,   15009   ,   31871   ,   46645   ,0 };
            const unsigned long dim3818Kuo3Init[]   =   {   1   ,   3   ,   3   ,   7   ,   13  ,   43  ,   105 ,   69  ,   25  ,   157 ,   923 ,   1997    ,   4383    ,   14935   ,   8821    ,   6383    ,0 };
            const unsigned long dim3819Kuo3Init[]   =   {   1   ,   1   ,   7   ,   1   ,   23  ,   41  ,   51  ,   215 ,   509 ,   117 ,   1871    ,   3025    ,   4295    ,   8327    ,   23887   ,   53039   ,0 };
            const unsigned long dim3820Kuo3Init[]   =   {   1   ,   1   ,   7   ,   1   ,   17  ,   9   ,   41  ,   89  ,   227 ,   909 ,   523 ,   3755    ,   3527    ,   12117   ,   12455   ,   18087   ,0 };
            const unsigned long dim3821Kuo3Init[]   =   {   1   ,   3   ,   1   ,   11  ,   19  ,   1   ,   77  ,   73  ,   139 ,   707 ,   163 ,   2715    ,   819 ,   8401    ,   24751   ,   24619   ,0 };
            const unsigned long dim3822Kuo3Init[]   =   {   1   ,   3   ,   3   ,   15  ,   11  ,   55  ,   31  ,   89  ,   383 ,   429 ,   269 ,   1593    ,   4929    ,   12037   ,   17279   ,   6489    ,0 };
            const unsigned long dim3823Kuo3Init[]   =   {   1   ,   1   ,   7   ,   13  ,   5   ,   19  ,   65  ,   43  ,   257 ,   711 ,   349 ,   3231    ,   3169    ,   11759   ,   19069   ,   29043   ,0 };
            const unsigned long dim3824Kuo3Init[]   =   {   1   ,   3   ,   7   ,   15  ,   5   ,   51  ,   31  ,   245 ,   197 ,   685 ,   1527    ,   3409    ,   4925    ,   8923    ,   10927   ,   62873   ,0 };
            const unsigned long dim3825Kuo3Init[]   =   {   1   ,   1   ,   7   ,   15  ,   31  ,   31  ,   61  ,   111 ,   327 ,   785 ,   97  ,   2659    ,   4777    ,   14187   ,   19731   ,   5195    ,0 };
            const unsigned long dim3826Kuo3Init[]   =   {   1   ,   3   ,   5   ,   13  ,   7   ,   53  ,   3   ,   97  ,   109 ,   537 ,   1395    ,   1919    ,   4447    ,   4067    ,   30137   ,   60659   ,0 };
            const unsigned long dim3827Kuo3Init[]   =   {   1   ,   3   ,   7   ,   11  ,   19  ,   45  ,   51  ,   107 ,   161 ,   269 ,   1003    ,   2363    ,   311 ,   7051    ,   25943   ,   20881   ,0 };
            const unsigned long dim3828Kuo3Init[]   =   {   1   ,   3   ,   7   ,   11  ,   27  ,   49  ,   79  ,   187 ,   265 ,   903 ,   1547    ,   1147    ,   8049    ,   10451   ,   15371   ,   43705   ,0 };
            const unsigned long dim3829Kuo3Init[]   =   {   1   ,   3   ,   5   ,   1   ,   9   ,   25  ,   29  ,   229 ,   229 ,   337 ,   1821    ,   2377    ,   1325    ,   15079   ,   28661   ,   46107   ,0 };
            const unsigned long dim3830Kuo3Init[]   =   {   1   ,   3   ,   7   ,   9   ,   29  ,   13  ,   121 ,   121 ,   229 ,   523 ,   707 ,   2577    ,   7637    ,   9257    ,   31031   ,   56527   ,0 };
            const unsigned long dim3831Kuo3Init[]   =   {   1   ,   1   ,   5   ,   9   ,   11  ,   41  ,   111 ,   59  ,   347 ,   295 ,   895 ,   3477    ,   149 ,   6747    ,   14671   ,   3405    ,0 };
            const unsigned long dim3832Kuo3Init[]   =   {   1   ,   1   ,   3   ,   11  ,   17  ,   17  ,   77  ,   105 ,   65  ,   831 ,   1969    ,   41  ,   1001    ,   2753    ,   4427    ,   25655   ,0 };
            const unsigned long dim3833Kuo3Init[]   =   {   1   ,   1   ,   5   ,   9   ,   7   ,   37  ,   55  ,   121 ,   51  ,   155 ,   443 ,   321 ,   1845    ,   11587   ,   28855   ,   13905   ,0 };
            const unsigned long dim3834Kuo3Init[]   =   {   1   ,   3   ,   3   ,   15  ,   9   ,   51  ,   1   ,   1   ,   453 ,   193 ,   1615    ,   2085    ,   3685    ,   15933   ,   18913   ,   46073   ,0 };
            const unsigned long dim3835Kuo3Init[]   =   {   1   ,   1   ,   1   ,   9   ,   31  ,   29  ,   59  ,   55  ,   33  ,   413 ,   1261    ,   1367    ,   1895    ,   10291   ,   575 ,   12241   ,0 };
            const unsigned long dim3836Kuo3Init[]   =   {   1   ,   3   ,   3   ,   9   ,   7   ,   11  ,   1   ,   207 ,   351 ,   357 ,   77  ,   4059    ,   3951    ,   4589    ,   11861   ,   43855   ,0 };
            const unsigned long dim3837Kuo3Init[]   =   {   1   ,   1   ,   1   ,   9   ,   15  ,   1   ,   77  ,   227 ,   377 ,   25  ,   1665    ,   3693    ,   8129    ,   10059   ,   19517   ,   51103   ,0 };
            const unsigned long dim3838Kuo3Init[]   =   {   1   ,   3   ,   5   ,   15  ,   25  ,   43  ,   103 ,   3   ,   477 ,   87  ,   1943    ,   1163    ,   7619    ,   8345    ,   25321   ,   36191   ,0 };
            const unsigned long dim3839Kuo3Init[]   =   {   1   ,   1   ,   1   ,   5   ,   13  ,   5   ,   99  ,   11  ,   33  ,   113 ,   1905    ,   2559    ,   2679    ,   15005   ,   12859   ,   19233   ,0 };
            const unsigned long dim3840Kuo3Init[]   =   {   1   ,   3   ,   5   ,   3   ,   9   ,   61  ,   79  ,   95  ,   71  ,   963 ,   1367    ,   1843    ,   3179    ,   10633   ,   5451    ,   64379   ,0 };
            const unsigned long dim3841Kuo3Init[]   =   {   1   ,   1   ,   1   ,   5   ,   11  ,   11  ,   67  ,   11  ,   133 ,   759 ,   1823    ,   3281    ,   6741    ,   4995    ,   21753   ,   17093   ,0 };
            const unsigned long dim3842Kuo3Init[]   =   {   1   ,   1   ,   7   ,   5   ,   29  ,   41  ,   31  ,   17  ,   381 ,   741 ,   567 ,   1009    ,   4781    ,   8953    ,   12397   ,   51833   ,0 };
            const unsigned long dim3843Kuo3Init[]   =   {   1   ,   3   ,   3   ,   7   ,   9   ,   61  ,   35  ,   203 ,   69  ,   633 ,   477 ,   2839    ,   1917    ,   12097   ,   28693   ,   42999   ,0 };
            const unsigned long dim3844Kuo3Init[]   =   {   1   ,   1   ,   3   ,   11  ,   29  ,   25  ,   1   ,   219 ,   343 ,   249 ,   1077    ,   125 ,   3197    ,   5877    ,   28407   ,   59627   ,0 };
            const unsigned long dim3845Kuo3Init[]   =   {   1   ,   1   ,   3   ,   1   ,   25  ,   37  ,   113 ,   125 ,   253 ,   947 ,   1103    ,   1283    ,   3493    ,   7129    ,   17881   ,   21279   ,0 };
            const unsigned long dim3846Kuo3Init[]   =   {   1   ,   3   ,   7   ,   15  ,   29  ,   43  ,   125 ,   203 ,   495 ,   421 ,   1725    ,   1673    ,   3635    ,   8161    ,   3753    ,   46323   ,0 };
            const unsigned long dim3847Kuo3Init[]   =   {   1   ,   1   ,   5   ,   1   ,   9   ,   7   ,   73  ,   117 ,   397 ,   521 ,   1699    ,   2715    ,   6149    ,   2337    ,   30623   ,   3943    ,0 };
            const unsigned long dim3848Kuo3Init[]   =   {   1   ,   3   ,   5   ,   7   ,   13  ,   17  ,   121 ,   163 ,   491 ,   111 ,   1589    ,   351 ,   303 ,   7395    ,   5731    ,   65513   ,0 };
            const unsigned long dim3849Kuo3Init[]   =   {   1   ,   1   ,   1   ,   3   ,   11  ,   1   ,   19  ,   179 ,   295 ,   595 ,   897 ,   1929    ,   1379    ,   13651   ,   19067   ,   43507   ,0 };
            const unsigned long dim3850Kuo3Init[]   =   {   1   ,   3   ,   5   ,   15  ,   29  ,   51  ,   19  ,   221 ,   67  ,   507 ,   213 ,   1873    ,   6051    ,   11191   ,   31475   ,   23447   ,0 };
            const unsigned long dim3851Kuo3Init[]   =   {   1   ,   3   ,   5   ,   3   ,   13  ,   35  ,   97  ,   75  ,   195 ,   569 ,   1047    ,   2607    ,   1549    ,   9495    ,   5945    ,   7581    ,0 };
            const unsigned long dim3852Kuo3Init[]   =   {   1   ,   3   ,   3   ,   15  ,   23  ,   41  ,   125 ,   101 ,   415 ,   927 ,   1421    ,   2043    ,   8035    ,   4077    ,   11133   ,   17333   ,0 };
            const unsigned long dim3853Kuo3Init[]   =   {   1   ,   3   ,   7   ,   3   ,   5   ,   51  ,   111 ,   129 ,   29  ,   469 ,   991 ,   2595    ,   165 ,   13987   ,   30077   ,   63491   ,0 };
            const unsigned long dim3854Kuo3Init[]   =   {   1   ,   3   ,   1   ,   3   ,   1   ,   17  ,   93  ,   7   ,   511 ,   3   ,   861 ,   3979    ,   2173    ,   6571    ,   30099   ,   12907   ,0 };
            const unsigned long dim3855Kuo3Init[]   =   {   1   ,   1   ,   7   ,   5   ,   9   ,   29  ,   43  ,   153 ,   83  ,   961 ,   1585    ,   3615    ,   4409    ,   1937    ,   24107   ,   45235   ,0 };
            const unsigned long dim3856Kuo3Init[]   =   {   1   ,   1   ,   3   ,   1   ,   19  ,   29  ,   81  ,   183 ,   197 ,   565 ,   1455    ,   2993    ,   6731    ,   2749    ,   1013    ,   27777   ,0 };
            const unsigned long dim3857Kuo3Init[]   =   {   1   ,   3   ,   7   ,   1   ,   17  ,   11  ,   121 ,   245 ,   103 ,   851 ,   1367    ,   141 ,   1257    ,   719 ,   25581   ,   48999   ,0 };
            const unsigned long dim3858Kuo3Init[]   =   {   1   ,   1   ,   5   ,   15  ,   15  ,   17  ,   33  ,   255 ,   113 ,   715 ,   1497    ,   3137    ,   1415    ,   7791    ,   28563   ,   28415   ,0 };
            const unsigned long dim3859Kuo3Init[]   =   {   1   ,   1   ,   3   ,   9   ,   7   ,   45  ,   19  ,   143 ,   73  ,   95  ,   317 ,   1777    ,   7263    ,   11833   ,   25181   ,   27137   ,0 };
            const unsigned long dim3860Kuo3Init[]   =   {   1   ,   1   ,   5   ,   3   ,   15  ,   49  ,   71  ,   57  ,   101 ,   95  ,   667 ,   2745    ,   6387    ,   14167   ,   661 ,   31809   ,0 };
            const unsigned long dim3861Kuo3Init[]   =   {   1   ,   1   ,   7   ,   3   ,   7   ,   11  ,   119 ,   5   ,   3   ,   383 ,   933 ,   293 ,   5507    ,   3025    ,   18865   ,   26577   ,0 };
            const unsigned long dim3862Kuo3Init[]   =   {   1   ,   1   ,   7   ,   9   ,   21  ,   57  ,   113 ,   173 ,   103 ,   743 ,   1573    ,   2165    ,   2909    ,   14253   ,   2303    ,   51939   ,0 };
            const unsigned long dim3863Kuo3Init[]   =   {   1   ,   3   ,   3   ,   9   ,   11  ,   29  ,   85  ,   239 ,   501 ,   123 ,   1149    ,   165 ,   6341    ,   10205   ,   18829   ,   32327   ,0 };
            const unsigned long dim3864Kuo3Init[]   =   {   1   ,   3   ,   1   ,   15  ,   21  ,   15  ,   63  ,   117 ,   217 ,   701 ,   299 ,   2887    ,   1341    ,   1617    ,   15163   ,   45899   ,0 };
            const unsigned long dim3865Kuo3Init[]   =   {   1   ,   3   ,   3   ,   13  ,   25  ,   63  ,   91  ,   225 ,   399 ,   765 ,   1719    ,   3169    ,   1645    ,   10445   ,   1109    ,   13021   ,0 };
            const unsigned long dim3866Kuo3Init[]   =   {   1   ,   1   ,   3   ,   13  ,   3   ,   9   ,   115 ,   35  ,   59  ,   145 ,   551 ,   2601    ,   1737    ,   9753    ,   7665    ,   30995   ,0 };
            const unsigned long dim3867Kuo3Init[]   =   {   1   ,   1   ,   5   ,   9   ,   1   ,   33  ,   37  ,   173 ,   385 ,   635 ,   1419    ,   1377    ,   3161    ,   1915    ,   30933   ,   46117   ,0 };
            const unsigned long dim3868Kuo3Init[]   =   {   1   ,   1   ,   7   ,   9   ,   21  ,   35  ,   65  ,   227 ,   329 ,   797 ,   61  ,   3409    ,   3911    ,   751 ,   21951   ,   11969   ,0 };
            const unsigned long dim3869Kuo3Init[]   =   {   1   ,   3   ,   1   ,   13  ,   13  ,   33  ,   29  ,   197 ,   183 ,   979 ,   979 ,   1629    ,   6017    ,   8645    ,   4007    ,   52377   ,0 };
            const unsigned long dim3870Kuo3Init[]   =   {   1   ,   1   ,   5   ,   7   ,   5   ,   17  ,   29  ,   117 ,   197 ,   163 ,   685 ,   2561    ,   2849    ,   5441    ,   32031   ,   25761   ,0 };
            const unsigned long dim3871Kuo3Init[]   =   {   1   ,   3   ,   1   ,   5   ,   9   ,   17  ,   59  ,   237 ,   33  ,   11  ,   399 ,   3543    ,   7273    ,   8239    ,   15191   ,   57529   ,0 };
            const unsigned long dim3872Kuo3Init[]   =   {   1   ,   1   ,   3   ,   7   ,   1   ,   45  ,   75  ,   107 ,   123 ,   909 ,   389 ,   3549    ,   4961    ,   15671   ,   14139   ,   29273   ,0 };
            const unsigned long dim3873Kuo3Init[]   =   {   1   ,   3   ,   7   ,   3   ,   23  ,   3   ,   93  ,   73  ,   265 ,   889 ,   27  ,   805 ,   6021    ,   5851    ,   7181    ,   35309   ,0 };
            const unsigned long dim3874Kuo3Init[]   =   {   1   ,   3   ,   1   ,   9   ,   23  ,   61  ,   17  ,   89  ,   495 ,   813 ,   1189    ,   1137    ,   7363    ,   6075    ,   3053    ,   28313   ,0 };
            const unsigned long dim3875Kuo3Init[]   =   {   1   ,   1   ,   3   ,   13  ,   3   ,   1   ,   9   ,   133 ,   63  ,   417 ,   843 ,   2481    ,   5703    ,   11095   ,   867 ,   28097   ,0 };
            const unsigned long dim3876Kuo3Init[]   =   {   1   ,   1   ,   5   ,   9   ,   23  ,   15  ,   113 ,   23  ,   199 ,   863 ,   1751    ,   765 ,   1917    ,   2145    ,   5391    ,   4629    ,0 };
            const unsigned long dim3877Kuo3Init[]   =   {   1   ,   3   ,   3   ,   11  ,   13  ,   59  ,   127 ,   115 ,   341 ,   311 ,   677 ,   2305    ,   19  ,   13871   ,   9297    ,   4023    ,0 };
            const unsigned long dim3878Kuo3Init[]   =   {   1   ,   1   ,   5   ,   7   ,   13  ,   3   ,   25  ,   195 ,   403 ,   323 ,   515 ,   2809    ,   983 ,   8699    ,   10205   ,   14719   ,0 };
            const unsigned long dim3879Kuo3Init[]   =   {   1   ,   1   ,   1   ,   13  ,   5   ,   33  ,   123 ,   213 ,   99  ,   889 ,   1151    ,   3339    ,   4855    ,   3677    ,   30627   ,   23225   ,0 };
            const unsigned long dim3880Kuo3Init[]   =   {   1   ,   1   ,   7   ,   9   ,   21  ,   19  ,   77  ,   85  ,   155 ,   379 ,   119 ,   81  ,   1937    ,   3589    ,   22645   ,   51585   ,0 };
            const unsigned long dim3881Kuo3Init[]   =   {   1   ,   1   ,   3   ,   1   ,   5   ,   11  ,   121 ,   111 ,   423 ,   927 ,   1551    ,   2207    ,   3589    ,   691 ,   8519    ,   61261   ,0 };
            const unsigned long dim3882Kuo3Init[]   =   {   1   ,   3   ,   5   ,   3   ,   1   ,   59  ,   25  ,   111 ,   415 ,   41  ,   623 ,   2505    ,   6075    ,   973 ,   10419   ,   26877   ,0 };
            const unsigned long dim3883Kuo3Init[]   =   {   1   ,   3   ,   3   ,   11  ,   15  ,   57  ,   43  ,   27  ,   177 ,   733 ,   1521    ,   3629    ,   3003    ,   15351   ,   26327   ,   32921   ,0 };
            const unsigned long dim3884Kuo3Init[]   =   {   1   ,   3   ,   3   ,   13  ,   1   ,   37  ,   3   ,   183 ,   349 ,   731 ,   1583    ,   2787    ,   6233    ,   14787   ,   16299   ,   4663    ,0 };
            const unsigned long dim3885Kuo3Init[]   =   {   1   ,   3   ,   7   ,   15  ,   29  ,   47  ,   59  ,   79  ,   265 ,   193 ,   503 ,   2393    ,   643 ,   14341   ,   21055   ,   43803   ,0 };
            const unsigned long dim3886Kuo3Init[]   =   {   1   ,   3   ,   1   ,   9   ,   11  ,   43  ,   75  ,   77  ,   493 ,   411 ,   145 ,   227 ,   3045    ,   363 ,   19879   ,   41017   ,0 };
            const unsigned long dim3887Kuo3Init[]   =   {   1   ,   1   ,   5   ,   3   ,   1   ,   49  ,   87  ,   3   ,   377 ,   795 ,   1905    ,   2959    ,   3743    ,   1459    ,   32493   ,   6283    ,0 };
            const unsigned long dim3888Kuo3Init[]   =   {   1   ,   3   ,   1   ,   13  ,   13  ,   13  ,   39  ,   83  ,   131 ,   935 ,   599 ,   1871    ,   7293    ,   11749   ,   6627    ,   58295   ,0 };
            const unsigned long dim3889Kuo3Init[]   =   {   1   ,   1   ,   5   ,   3   ,   15  ,   9   ,   107 ,   239 ,   453 ,   53  ,   1825    ,   1493    ,   3611    ,   9993    ,   12535   ,   38879   ,0 };
            const unsigned long dim3890Kuo3Init[]   =   {   1   ,   3   ,   5   ,   7   ,   1   ,   21  ,   93  ,   189 ,   477 ,   923 ,   13  ,   3797    ,   8147    ,   8417    ,   21473   ,   62305   ,0 };
            const unsigned long dim3891Kuo3Init[]   =   {   1   ,   3   ,   3   ,   15  ,   21  ,   47  ,   75  ,   183 ,   425 ,   281 ,   813 ,   571 ,   1017    ,   565 ,   8541    ,   52977   ,0 };
            const unsigned long dim3892Kuo3Init[]   =   {   1   ,   1   ,   5   ,   13  ,   29  ,   53  ,   25  ,   217 ,   111 ,   605 ,   1771    ,   1827    ,   7245    ,   16063   ,   10769   ,   57483   ,0 };
            const unsigned long dim3893Kuo3Init[]   =   {   1   ,   3   ,   5   ,   7   ,   9   ,   27  ,   99  ,   75  ,   171 ,   299 ,   1751    ,   841 ,   4717    ,   7373    ,   14241   ,   7435    ,0 };
            const unsigned long dim3894Kuo3Init[]   =   {   1   ,   3   ,   7   ,   3   ,   27  ,   13  ,   121 ,   69  ,   453 ,   1009    ,   1353    ,   3289    ,   3221    ,   9127    ,   29967   ,   60919   ,0 };
            const unsigned long dim3895Kuo3Init[]   =   {   1   ,   1   ,   1   ,   11  ,   5   ,   41  ,   81  ,   151 ,   479 ,   599 ,   1751    ,   2841    ,   5721    ,   1725    ,   3465    ,   23385   ,0 };
            const unsigned long dim3896Kuo3Init[]   =   {   1   ,   1   ,   3   ,   11  ,   7   ,   7   ,   35  ,   169 ,   35  ,   255 ,   1657    ,   297 ,   3855    ,   2071    ,   10929   ,   15907   ,0 };
            const unsigned long dim3897Kuo3Init[]   =   {   1   ,   1   ,   1   ,   3   ,   25  ,   1   ,   43  ,   139 ,   165 ,   535 ,   95  ,   2769    ,   8105    ,   6865    ,   20297   ,   64539   ,0 };
            const unsigned long dim3898Kuo3Init[]   =   {   1   ,   3   ,   7   ,   7   ,   21  ,   27  ,   11  ,   209 ,   279 ,   769 ,   1961    ,   101 ,   2103    ,   4529    ,   4987    ,   29097   ,0 };
            const unsigned long dim3899Kuo3Init[]   =   {   1   ,   1   ,   1   ,   11  ,   29  ,   37  ,   45  ,   65  ,   75  ,   729 ,   1773    ,   4023    ,   7671    ,   6259    ,   22339   ,   45735   ,0 };
            const unsigned long dim3900Kuo3Init[]   =   {   1   ,   3   ,   7   ,   5   ,   25  ,   3   ,   41  ,   139 ,   365 ,   915 ,   111 ,   611 ,   7609    ,   6259    ,   12247   ,   58671   ,0 };
            const unsigned long dim3901Kuo3Init[]   =   {   1   ,   1   ,   7   ,   15  ,   11  ,   45  ,   95  ,   121 ,   3   ,   491 ,   1779    ,   1055    ,   7051    ,   15499   ,   17301   ,   13081   ,0 };
            const unsigned long dim3902Kuo3Init[]   =   {   1   ,   1   ,   1   ,   13  ,   5   ,   43  ,   105 ,   173 ,   267 ,   263 ,   873 ,   953 ,   5945    ,   11417   ,   23799   ,   54591   ,0 };
            const unsigned long dim3903Kuo3Init[]   =   {   1   ,   1   ,   1   ,   9   ,   1   ,   55  ,   15  ,   73  ,   87  ,   49  ,   1651    ,   775 ,   6485    ,   9543    ,   26199   ,   30419   ,0 };
            const unsigned long dim3904Kuo3Init[]   =   {   1   ,   3   ,   1   ,   3   ,   11  ,   5   ,   35  ,   37  ,   451 ,   85  ,   943 ,   4017    ,   6795    ,   8915    ,   28551   ,   45097   ,0 };
            const unsigned long dim3905Kuo3Init[]   =   {   1   ,   3   ,   7   ,   11  ,   11  ,   51  ,   93  ,   237 ,   487 ,   515 ,   1511    ,   383 ,   7571    ,   4673    ,   29395   ,   21659   ,0 };
            const unsigned long dim3906Kuo3Init[]   =   {   1   ,   1   ,   3   ,   1   ,   11  ,   63  ,   113 ,   253 ,   99  ,   733 ,   2001    ,   3767    ,   7431    ,   12131   ,   25203   ,   22251   ,0 };
            const unsigned long dim3907Kuo3Init[]   =   {   1   ,   3   ,   3   ,   1   ,   11  ,   29  ,   37  ,   181 ,   299 ,   9   ,   11  ,   317 ,   1079    ,   13441   ,   8925    ,   15847   ,0 };
            const unsigned long dim3908Kuo3Init[]   =   {   1   ,   1   ,   5   ,   5   ,   5   ,   9   ,   77  ,   67  ,   165 ,   273 ,   1161    ,   65  ,   7157    ,   13531   ,   2987    ,   1169    ,0 };
            const unsigned long dim3909Kuo3Init[]   =   {   1   ,   3   ,   5   ,   7   ,   9   ,   57  ,   17  ,   209 ,   357 ,   901 ,   1113    ,   61  ,   5751    ,   2255    ,   29677   ,   56397   ,0 };
            const unsigned long dim3910Kuo3Init[]   =   {   1   ,   1   ,   5   ,   11  ,   7   ,   5   ,   7   ,   53  ,   397 ,   279 ,   1387    ,   2447    ,   115 ,   2043    ,   19287   ,   45753   ,0 };
            const unsigned long dim3911Kuo3Init[]   =   {   1   ,   3   ,   3   ,   15  ,   25  ,   41  ,   45  ,   33  ,   167 ,   801 ,   1929    ,   1173    ,   1525    ,   3725    ,   22407   ,   37077   ,0 };
            const unsigned long dim3912Kuo3Init[]   =   {   1   ,   3   ,   1   ,   11  ,   13  ,   61  ,   107 ,   181 ,   179 ,   51  ,   1379    ,   3573    ,   7319    ,   4843    ,   10023   ,   28725   ,0 };
            const unsigned long dim3913Kuo3Init[]   =   {   1   ,   1   ,   5   ,   3   ,   11  ,   61  ,   61  ,   1   ,   211 ,   557 ,   417 ,   899 ,   4651    ,   3375    ,   19423   ,   11719   ,0 };
            const unsigned long dim3914Kuo3Init[]   =   {   1   ,   3   ,   1   ,   13  ,   9   ,   5   ,   117 ,   217 ,   193 ,   911 ,   991 ,   155 ,   3011    ,   4759    ,   30953   ,   63939   ,0 };
            const unsigned long dim3915Kuo3Init[]   =   {   1   ,   3   ,   1   ,   13  ,   21  ,   7   ,   77  ,   29  ,   457 ,   923 ,   871 ,   147 ,   5177    ,   13949   ,   1977    ,   27515   ,0 };
            const unsigned long dim3916Kuo3Init[]   =   {   1   ,   3   ,   1   ,   3   ,   5   ,   39  ,   83  ,   133 ,   471 ,   355 ,   615 ,   1627    ,   2471    ,   59  ,   21449   ,   52293   ,0 };
            const unsigned long dim3917Kuo3Init[]   =   {   1   ,   3   ,   7   ,   13  ,   23  ,   57  ,   97  ,   3   ,   385 ,   255 ,   1565    ,   3479    ,   5587    ,   15101   ,   27153   ,   9239    ,0 };
            const unsigned long dim3918Kuo3Init[]   =   {   1   ,   3   ,   5   ,   5   ,   13  ,   25  ,   99  ,   19  ,   49  ,   655 ,   1823    ,   641 ,   619 ,   5913    ,   11379   ,   23781   ,0 };
            const unsigned long dim3919Kuo3Init[]   =   {   1   ,   1   ,   1   ,   5   ,   21  ,   1   ,   91  ,   3   ,   31  ,   19  ,   19  ,   3627    ,   1575    ,   1439    ,   7473    ,   36215   ,0 };
            const unsigned long dim3920Kuo3Init[]   =   {   1   ,   3   ,   5   ,   13  ,   29  ,   61  ,   1   ,   151 ,   291 ,   315 ,   965 ,   3435    ,   1849    ,   13069   ,   9785    ,   41345   ,0 };
            const unsigned long dim3921Kuo3Init[]   =   {   1   ,   3   ,   3   ,   15  ,   17  ,   9   ,   97  ,   69  ,   233 ,   925 ,   1933    ,   3861    ,   89  ,   7889    ,   28853   ,   64959   ,0 };
            const unsigned long dim3922Kuo3Init[]   =   {   1   ,   3   ,   3   ,   3   ,   5   ,   51  ,   67  ,   175 ,   341 ,   191 ,   1643    ,   3327    ,   763 ,   327 ,   28069   ,   15317   ,0 };
            const unsigned long dim3923Kuo3Init[]   =   {   1   ,   3   ,   5   ,   3   ,   3   ,   37  ,   109 ,   43  ,   465 ,   57  ,   489 ,   915 ,   6943    ,   7947    ,   9229    ,   5647    ,0 };
            const unsigned long dim3924Kuo3Init[]   =   {   1   ,   1   ,   7   ,   7   ,   27  ,   1   ,   85  ,   171 ,   185 ,   615 ,   1451    ,   237 ,   1597    ,   4255    ,   24429   ,   34357   ,0 };
            const unsigned long dim3925Kuo3Init[]   =   {   1   ,   3   ,   1   ,   7   ,   29  ,   35  ,   69  ,   237 ,   117 ,   841 ,   1549    ,   2093    ,   105 ,   10197   ,   5229    ,   59299   ,0 };
            const unsigned long dim3926Kuo3Init[]   =   {   1   ,   3   ,   5   ,   3   ,   7   ,   33  ,   73  ,   81  ,   385 ,   177 ,   1849    ,   773 ,   1085    ,   5539    ,   29213   ,   30635   ,0 };
            const unsigned long dim3927Kuo3Init[]   =   {   1   ,   3   ,   7   ,   15  ,   15  ,   47  ,   53  ,   207 ,   337 ,   479 ,   775 ,   2851    ,   725 ,   14779   ,   29653   ,   35949   ,0 };
            const unsigned long dim3928Kuo3Init[]   =   {   1   ,   1   ,   3   ,   9   ,   15  ,   53  ,   53  ,   87  ,   315 ,   929 ,   919 ,   907 ,   6893    ,   4647    ,   2019    ,   57879   ,0 };
            const unsigned long dim3929Kuo3Init[]   =   {   1   ,   3   ,   5   ,   11  ,   5   ,   31  ,   21  ,   9   ,   215 ,   599 ,   617 ,   3737    ,   787 ,   10513   ,   17907   ,   22221   ,0 };
            const unsigned long dim3930Kuo3Init[]   =   {   1   ,   3   ,   3   ,   3   ,   27  ,   5   ,   27  ,   157 ,   423 ,   197 ,   1757    ,   2451    ,   6317    ,   1557    ,   6641    ,   12621   ,0 };
            const unsigned long dim3931Kuo3Init[]   =   {   1   ,   1   ,   5   ,   9   ,   19  ,   17  ,   109 ,   37  ,   329 ,   217 ,   1265    ,   3073    ,   6243    ,   6179    ,   25921   ,   9001    ,0 };
            const unsigned long dim3932Kuo3Init[]   =   {   1   ,   3   ,   5   ,   15  ,   13  ,   13  ,   103 ,   79  ,   399 ,   897 ,   1241    ,   2847    ,   5913    ,   2267    ,   5367    ,   26337   ,0 };
            const unsigned long dim3933Kuo3Init[]   =   {   1   ,   1   ,   3   ,   1   ,   17  ,   41  ,   19  ,   213 ,   123 ,   123 ,   1103    ,   1125    ,   2819    ,   7643    ,   26391   ,   20935   ,0 };
            const unsigned long dim3934Kuo3Init[]   =   {   1   ,   1   ,   1   ,   5   ,   29  ,   27  ,   79  ,   143 ,   93  ,   311 ,   433 ,   1945    ,   2783    ,   13787   ,   4473    ,   23279   ,0 };
            const unsigned long dim3935Kuo3Init[]   =   {   1   ,   3   ,   1   ,   1   ,   13  ,   49  ,   39  ,   21  ,   117 ,   933 ,   169 ,   551 ,   2871    ,   9879    ,   3715    ,   25143   ,0 };
            const unsigned long dim3936Kuo3Init[]   =   {   1   ,   3   ,   1   ,   3   ,   29  ,   15  ,   5   ,   89  ,   255 ,   779 ,   495 ,   2335    ,   6029    ,   8733    ,   5687    ,   40189   ,0 };
            const unsigned long dim3937Kuo3Init[]   =   {   1   ,   1   ,   5   ,   3   ,   5   ,   61  ,   37  ,   231 ,   395 ,   175 ,   711 ,   1149    ,   4819    ,   10119   ,   24929   ,   37155   ,0 };
            const unsigned long dim3938Kuo3Init[]   =   {   1   ,   1   ,   7   ,   3   ,   7   ,   49  ,   81  ,   221 ,   193 ,   685 ,   2029    ,   3979    ,   5071    ,   2965    ,   18951   ,   21497   ,0 };
            const unsigned long dim3939Kuo3Init[]   =   {   1   ,   1   ,   3   ,   5   ,   11  ,   17  ,   29  ,   153 ,   379 ,   387 ,   1271    ,   1261    ,   817 ,   5913    ,   22395   ,   885 ,0 };
            const unsigned long dim3940Kuo3Init[]   =   {   1   ,   1   ,   1   ,   1   ,   3   ,   37  ,   89  ,   255 ,   337 ,   135 ,   535 ,   1087    ,   7987    ,   8107    ,   22953   ,   18709   ,0 };
            const unsigned long dim3941Kuo3Init[]   =   {   1   ,   1   ,   5   ,   11  ,   25  ,   29  ,   15  ,   223 ,   181 ,   879 ,   1661    ,   1605    ,   3777    ,   5221    ,   7253    ,   22951   ,0 };
            const unsigned long dim3942Kuo3Init[]   =   {   1   ,   3   ,   1   ,   1   ,   3   ,   9   ,   57  ,   239 ,   439 ,   669 ,   1069    ,   2001    ,   2689    ,   9569    ,   24151   ,   54337   ,0 };
            const unsigned long dim3943Kuo3Init[]   =   {   1   ,   1   ,   1   ,   11  ,   7   ,   47  ,   21  ,   31  ,   35  ,   739 ,   1229    ,   2929    ,   5221    ,   6715    ,   32275   ,   42743   ,0 };
            const unsigned long dim3944Kuo3Init[]   =   {   1   ,   3   ,   1   ,   7   ,   19  ,   3   ,   85  ,   55  ,   397 ,   875 ,   949 ,   2835    ,   6077    ,   1069    ,   13307   ,   5897    ,0 };
            const unsigned long dim3945Kuo3Init[]   =   {   1   ,   3   ,   7   ,   15  ,   11  ,   57  ,   113 ,   193 ,   191 ,   53  ,   1319    ,   1585    ,   4797    ,   13903   ,   24875   ,   50229   ,0 };
            const unsigned long dim3946Kuo3Init[]   =   {   1   ,   3   ,   7   ,   7   ,   15  ,   17  ,   29  ,   125 ,   387 ,   775 ,   411 ,   1283    ,   6069    ,   10909   ,   14221   ,   52811   ,0 };
            const unsigned long dim3947Kuo3Init[]   =   {   1   ,   3   ,   7   ,   15  ,   17  ,   47  ,   73  ,   143 ,   133 ,   1009    ,   1671    ,   2307    ,   2991    ,   10319   ,   3161    ,   665 ,0 };
            const unsigned long dim3948Kuo3Init[]   =   {   1   ,   3   ,   7   ,   11  ,   23  ,   29  ,   115 ,   77  ,   167 ,   667 ,   1983    ,   3985    ,   1449    ,   10055   ,   27947   ,   63501   ,0 };
            const unsigned long dim3949Kuo3Init[]   =   {   1   ,   1   ,   3   ,   11  ,   27  ,   61  ,   83  ,   173 ,   427 ,   161 ,   1349    ,   297 ,   1579    ,   9929    ,   32663   ,   44385   ,0 };
            const unsigned long dim3950Kuo3Init[]   =   {   1   ,   1   ,   7   ,   9   ,   19  ,   31  ,   73  ,   233 ,   109 ,   991 ,   1381    ,   3223    ,   4033    ,   4447    ,   4773    ,   61631   ,0 };
            const unsigned long dim3951Kuo3Init[]   =   {   1   ,   1   ,   3   ,   5   ,   17  ,   53  ,   71  ,   201 ,   181 ,   395 ,   1305    ,   2049    ,   4317    ,   12689   ,   23133   ,   13401   ,0 };
            const unsigned long dim3952Kuo3Init[]   =   {   1   ,   3   ,   1   ,   1   ,   13  ,   63  ,   17  ,   107 ,   149 ,   739 ,   1829    ,   2441    ,   3491    ,   1649    ,   15839   ,   13021   ,0 };
            const unsigned long dim3953Kuo3Init[]   =   {   1   ,   1   ,   1   ,   3   ,   3   ,   15  ,   35  ,   247 ,   95  ,   973 ,   1625    ,   3097    ,   6075    ,   14331   ,   8345    ,   12225   ,0 };
            const unsigned long dim3954Kuo3Init[]   =   {   1   ,   3   ,   3   ,   7   ,   9   ,   21  ,   5   ,   85  ,   431 ,   317 ,   1313    ,   1981    ,   7761    ,   2857    ,   18659   ,   58677   ,0 };
            const unsigned long dim3955Kuo3Init[]   =   {   1   ,   3   ,   3   ,   1   ,   25  ,   23  ,   5   ,   213 ,   345 ,   119 ,   1979    ,   1023    ,   6849    ,   9587    ,   2519    ,   57219   ,0 };
            const unsigned long dim3956Kuo3Init[]   =   {   1   ,   1   ,   5   ,   11  ,   17  ,   1   ,   15  ,   213 ,   21  ,   435 ,   1719    ,   3065    ,   6197    ,   8993    ,   8479    ,   17383   ,0 };
            const unsigned long dim3957Kuo3Init[]   =   {   1   ,   1   ,   3   ,   5   ,   17  ,   35  ,   97  ,   167 ,   411 ,   57  ,   145 ,   663 ,   1319    ,   16029   ,   1649    ,   13097   ,0 };
            const unsigned long dim3958Kuo3Init[]   =   {   1   ,   3   ,   3   ,   9   ,   13  ,   1   ,   89  ,   179 ,   355 ,   655 ,   1217    ,   1421    ,   4303    ,   2705    ,   30107   ,   45801   ,0 };
            const unsigned long dim3959Kuo3Init[]   =   {   1   ,   1   ,   7   ,   1   ,   23  ,   49  ,   67  ,   119 ,   139 ,   397 ,   983 ,   3537    ,   1463    ,   10395   ,   4717    ,   50885   ,0 };
            const unsigned long dim3960Kuo3Init[]   =   {   1   ,   1   ,   1   ,   9   ,   9   ,   9   ,   95  ,   75  ,   489 ,   459 ,   951 ,   1829    ,   5017    ,   8703    ,   6229    ,   18921   ,0 };
            const unsigned long dim3961Kuo3Init[]   =   {   1   ,   3   ,   7   ,   5   ,   7   ,   55  ,   89  ,   75  ,   231 ,   141 ,   1611    ,   2275    ,   1953    ,   6389    ,   12899   ,   45117   ,0 };
            const unsigned long dim3962Kuo3Init[]   =   {   1   ,   3   ,   1   ,   1   ,   25  ,   53  ,   91  ,   25  ,   425 ,   275 ,   189 ,   863 ,   5777    ,   6179    ,   12273   ,   3473    ,0 };
            const unsigned long dim3963Kuo3Init[]   =   {   1   ,   1   ,   7   ,   15  ,   5   ,   17  ,   75  ,   83  ,   357 ,   523 ,   1075    ,   1981    ,   3993    ,   13133   ,   16441   ,   53935   ,0 };
            const unsigned long dim3964Kuo3Init[]   =   {   1   ,   3   ,   7   ,   5   ,   15  ,   13  ,   103 ,   89  ,   309 ,   961 ,   771 ,   1441    ,   1449    ,   15867   ,   20019   ,   22303   ,0 };
            const unsigned long dim3965Kuo3Init[]   =   {   1   ,   3   ,   5   ,   9   ,   7   ,   47  ,   33  ,   117 ,   151 ,   77  ,   893 ,   867 ,   2523    ,   13975   ,   20649   ,   15673   ,0 };
            const unsigned long dim3966Kuo3Init[]   =   {   1   ,   3   ,   5   ,   15  ,   5   ,   15  ,   57  ,   149 ,   405 ,   523 ,   763 ,   4071    ,   5915    ,   3255    ,   22243   ,   46779   ,0 };
            const unsigned long dim3967Kuo3Init[]   =   {   1   ,   3   ,   7   ,   5   ,   17  ,   39  ,   53  ,   181 ,   65  ,   541 ,   1995    ,   3213    ,   3837    ,   3527    ,   2433    ,   23527   ,0 };
            const unsigned long dim3968Kuo3Init[]   =   {   1   ,   3   ,   5   ,   3   ,   17  ,   9   ,   5   ,   79  ,   299 ,   441 ,   1891    ,   3607    ,   5087    ,   14629   ,   31767   ,   21443   ,0 };
            const unsigned long dim3969Kuo3Init[]   =   {   1   ,   1   ,   5   ,   11  ,   1   ,   47  ,   25  ,   13  ,   201 ,   691 ,   507 ,   2041    ,   3169    ,   2909    ,   5341    ,   7643    ,0 };
            const unsigned long dim3970Kuo3Init[]   =   {   1   ,   1   ,   1   ,   1   ,   31  ,   55  ,   109 ,   97  ,   335 ,   835 ,   933 ,   839 ,   3271    ,   14323   ,   25319   ,   18807   ,0 };
            const unsigned long dim3971Kuo3Init[]   =   {   1   ,   3   ,   7   ,   9   ,   19  ,   21  ,   19  ,   79  ,   449 ,   1007    ,   1559    ,   3143    ,   2633    ,   11691   ,   3273    ,   17381   ,0 };
            const unsigned long dim3972Kuo3Init[]   =   {   1   ,   1   ,   3   ,   9   ,   3   ,   61  ,   105 ,   255 ,   327 ,   649 ,   11  ,   715 ,   2629    ,   8425    ,   165 ,   6235    ,0 };
            const unsigned long dim3973Kuo3Init[]   =   {   1   ,   3   ,   5   ,   11  ,   9   ,   39  ,   45  ,   161 ,   421 ,   243 ,   1589    ,   157 ,   5309    ,   9769    ,   28229   ,   12947   ,0 };
            const unsigned long dim3974Kuo3Init[]   =   {   1   ,   3   ,   3   ,   13  ,   15  ,   17  ,   75  ,   23  ,   189 ,   601 ,   737 ,   687 ,   3387    ,   11913   ,   17245   ,   16073   ,0 };
            const unsigned long dim3975Kuo3Init[]   =   {   1   ,   1   ,   1   ,   3   ,   25  ,   45  ,   73  ,   243 ,   171 ,   179 ,   617 ,   2039    ,   4937    ,   963 ,   10791   ,   48989   ,0 };
            const unsigned long dim3976Kuo3Init[]   =   {   1   ,   1   ,   7   ,   3   ,   3   ,   61  ,   123 ,   27  ,   3   ,   385 ,   1357    ,   3773    ,   1639    ,   353 ,   21277   ,   52013   ,0 };
            const unsigned long dim3977Kuo3Init[]   =   {   1   ,   3   ,   5   ,   9   ,   27  ,   19  ,   49  ,   255 ,   213 ,   753 ,   297 ,   601 ,   4905    ,   15889   ,   18055   ,   31511   ,0 };
            const unsigned long dim3978Kuo3Init[]   =   {   1   ,   1   ,   7   ,   9   ,   1   ,   19  ,   93  ,   251 ,   279 ,   21  ,   891 ,   859 ,   4637    ,   2587    ,   17659   ,   19485   ,0 };
            const unsigned long dim3979Kuo3Init[]   =   {   1   ,   3   ,   3   ,   11  ,   29  ,   55  ,   69  ,   151 ,   445 ,   947 ,   719 ,   3935    ,   4625    ,   9321    ,   25645   ,   39153   ,0 };
            const unsigned long dim3980Kuo3Init[]   =   {   1   ,   3   ,   7   ,   5   ,   23  ,   57  ,   79  ,   21  ,   353 ,   7   ,   295 ,   3379    ,   3505    ,   15409   ,   9117    ,   19745   ,0 };
            const unsigned long dim3981Kuo3Init[]   =   {   1   ,   1   ,   5   ,   13  ,   7   ,   5   ,   41  ,   249 ,   429 ,   999 ,   1945    ,   1097    ,   795 ,   8647    ,   23297   ,   60707   ,0 };
            const unsigned long dim3982Kuo3Init[]   =   {   1   ,   1   ,   5   ,   13  ,   3   ,   57  ,   105 ,   175 ,   345 ,   53  ,   455 ,   1549    ,   5291    ,   6757    ,   12637   ,   45159   ,0 };
            const unsigned long dim3983Kuo3Init[]   =   {   1   ,   3   ,   3   ,   11  ,   5   ,   31  ,   5   ,   69  ,   373 ,   453 ,   1861    ,   1061    ,   6339    ,   3857    ,   13979   ,   24015   ,0 };
            const unsigned long dim3984Kuo3Init[]   =   {   1   ,   1   ,   5   ,   11  ,   19  ,   41  ,   113 ,   183 ,   299 ,   197 ,   397 ,   1597    ,   6471    ,   14855   ,   26067   ,   2705    ,0 };
            const unsigned long dim3985Kuo3Init[]   =   {   1   ,   1   ,   1   ,   13  ,   15  ,   59  ,   89  ,   73  ,   45  ,   891 ,   1285    ,   1565    ,   4675    ,   10669   ,   3067    ,   40585   ,0 };
            const unsigned long dim3986Kuo3Init[]   =   {   1   ,   1   ,   5   ,   7   ,   1   ,   51  ,   59  ,   27  ,   275 ,   961 ,   765 ,   3003    ,   1677    ,   2093    ,   15083   ,   51929   ,0 };
            const unsigned long dim3987Kuo3Init[]   =   {   1   ,   1   ,   5   ,   15  ,   19  ,   37  ,   101 ,   247 ,   223 ,   39  ,   2009    ,   3155    ,   6357    ,   13513   ,   30965   ,   36485   ,0 };
            const unsigned long dim3988Kuo3Init[]   =   {   1   ,   1   ,   5   ,   1   ,   27  ,   9   ,   71  ,   231 ,   115 ,   1   ,   719 ,   873 ,   4895    ,   4031    ,   23581   ,   17529   ,0 };
            const unsigned long dim3989Kuo3Init[]   =   {   1   ,   1   ,   3   ,   9   ,   1   ,   63  ,   13  ,   187 ,   325 ,   199 ,   1165    ,   491 ,   1361    ,   15377   ,   9463    ,   5077    ,0 };
            const unsigned long dim3990Kuo3Init[]   =   {   1   ,   3   ,   3   ,   7   ,   9   ,   13  ,   71  ,   17  ,   213 ,   915 ,   1779    ,   401 ,   5315    ,   2161    ,   14765   ,   11301   ,0 };
            const unsigned long dim3991Kuo3Init[]   =   {   1   ,   3   ,   7   ,   15  ,   5   ,   5   ,   3   ,   137 ,   17  ,   197 ,   1559    ,   3077    ,   4615    ,   923 ,   2143    ,   16743   ,0 };
            const unsigned long dim3992Kuo3Init[]   =   {   1   ,   1   ,   3   ,   5   ,   13  ,   19  ,   77  ,   155 ,   285 ,   209 ,   1089    ,   2213    ,   3131    ,   11685   ,   13287   ,   10439   ,0 };
            const unsigned long dim3993Kuo3Init[]   =   {   1   ,   1   ,   1   ,   9   ,   7   ,   11  ,   35  ,   113 ,   165 ,   795 ,   1657    ,   483 ,   4937    ,   5209    ,   19801   ,   13831   ,0 };
            const unsigned long dim3994Kuo3Init[]   =   {   1   ,   1   ,   7   ,   9   ,   29  ,   1   ,   67  ,   131 ,   277 ,   919 ,   1267    ,   3907    ,   2561    ,   2469    ,   25321   ,   17769   ,0 };
            const unsigned long dim3995Kuo3Init[]   =   {   1   ,   3   ,   5   ,   3   ,   25  ,   39  ,   93  ,   5   ,   151 ,   905 ,   201 ,   1499    ,   4069    ,   4691    ,   8293    ,   945 ,0 };
            const unsigned long dim3996Kuo3Init[]   =   {   1   ,   3   ,   3   ,   3   ,   3   ,   59  ,   47  ,   13  ,   79  ,   157 ,   231 ,   1105    ,   5449    ,   10889   ,   2625    ,   53303   ,0 };
            const unsigned long dim3997Kuo3Init[]   =   {   1   ,   3   ,   7   ,   15  ,   13  ,   23  ,   25  ,   171 ,   391 ,   497 ,   57  ,   4089    ,   3849    ,   6121    ,   1561    ,   29977   ,0 };
            const unsigned long dim3998Kuo3Init[]   =   {   1   ,   3   ,   7   ,   5   ,   17  ,   19  ,   33  ,   193 ,   351 ,   189 ,   619 ,   2091    ,   7881    ,   12689   ,   3799    ,   1199    ,0 };
            const unsigned long dim3999Kuo3Init[]   =   {   1   ,   3   ,   7   ,   7   ,   5   ,   35  ,   23  ,   219 ,   247 ,   377 ,   439 ,   3421    ,   2729    ,   7169    ,   24373   ,   48533   ,0 };
            const unsigned long dim4000Kuo3Init[]   =   {   1   ,   1   ,   1   ,   13  ,   23  ,   33  ,   77  ,   37  ,   347 ,   831 ,   1525    ,   2383    ,   6483    ,   15377   ,   22761   ,   51647   ,0 };
            const unsigned long dim4001Kuo3Init[]   =   {   1   ,   1   ,   3   ,   9   ,   5   ,   51  ,   89  ,   207 ,   131 ,   969 ,   463 ,   3917    ,   5985    ,   10797   ,   25183   ,   35637   ,0 };
            const unsigned long dim4002Kuo3Init[]   =   {   1   ,   1   ,   1   ,   15  ,   9   ,   45  ,   99  ,   25  ,   7   ,   279 ,   541 ,   879 ,   6681    ,   2591    ,   31667   ,   51957   ,0 };
            const unsigned long dim4003Kuo3Init[]   =   {   1   ,   1   ,   5   ,   1   ,   19  ,   13  ,   59  ,   119 ,   195 ,   385 ,   1715    ,   823 ,   7631    ,   9547    ,   27099   ,   30967   ,0 };
            const unsigned long dim4004Kuo3Init[]   =   {   1   ,   1   ,   3   ,   13  ,   21  ,   29  ,   9   ,   17  ,   51  ,   819 ,   365 ,   1855    ,   1505    ,   13923   ,   27397   ,   25347   ,0 };
            const unsigned long dim4005Kuo3Init[]   =   {   1   ,   3   ,   1   ,   9   ,   21  ,   1   ,   17  ,   109 ,   435 ,   887 ,   1951    ,   923 ,   181 ,   8365    ,   1531    ,   5533    ,0 };
            const unsigned long dim4006Kuo3Init[]   =   {   1   ,   3   ,   1   ,   15  ,   7   ,   29  ,   101 ,   221 ,   151 ,   49  ,   1655    ,   2889    ,   7179    ,   1019    ,   10765   ,   49489   ,0 };
            const unsigned long dim4007Kuo3Init[]   =   {   1   ,   3   ,   1   ,   13  ,   27  ,   15  ,   25  ,   249 ,   31  ,   187 ,   1875    ,   2937    ,   5831    ,   13499   ,   27997   ,   8003    ,0 };
            const unsigned long dim4008Kuo3Init[]   =   {   1   ,   3   ,   1   ,   5   ,   13  ,   55  ,   111 ,   231 ,   71  ,   245 ,   479 ,   2487    ,   7967    ,   10911   ,   12853   ,   56217   ,0 };
            const unsigned long dim4009Kuo3Init[]   =   {   1   ,   1   ,   7   ,   7   ,   15  ,   13  ,   77  ,   37  ,   497 ,   899 ,   1527    ,   275 ,   517 ,   10323   ,   2329    ,   19267   ,0 };
            const unsigned long dim4010Kuo3Init[]   =   {   1   ,   3   ,   5   ,   5   ,   13  ,   1   ,   17  ,   201 ,   211 ,   153 ,   1377    ,   347 ,   3481    ,   6785    ,   23745   ,   54849   ,0 };
            const unsigned long dim4011Kuo3Init[]   =   {   1   ,   1   ,   1   ,   13  ,   23  ,   29  ,   55  ,   91  ,   249 ,   815 ,   1375    ,   3081    ,   1445    ,   13763   ,   29983   ,   24581   ,0 };
            const unsigned long dim4012Kuo3Init[]   =   {   1   ,   1   ,   5   ,   15  ,   23  ,   19  ,   65  ,   31  ,   105 ,   797 ,   957 ,   1741    ,   4189    ,   3313    ,   17183   ,   34345   ,0 };
            const unsigned long dim4013Kuo3Init[]   =   {   1   ,   1   ,   7   ,   13  ,   23  ,   51  ,   13  ,   119 ,   19  ,   901 ,   1985    ,   3269    ,   3753    ,   5365    ,   3871    ,   30099   ,0 };
            const unsigned long dim4014Kuo3Init[]   =   {   1   ,   1   ,   1   ,   1   ,   13  ,   11  ,   3   ,   59  ,   45  ,   1007    ,   121 ,   1433    ,   1559    ,   15593   ,   187 ,   55939   ,0 };
            const unsigned long dim4015Kuo3Init[]   =   {   1   ,   3   ,   7   ,   13  ,   7   ,   15  ,   17  ,   227 ,   143 ,   889 ,   1011    ,   1195    ,   3849    ,   12643   ,   27303   ,   10473   ,0 };
            const unsigned long dim4016Kuo3Init[]   =   {   1   ,   3   ,   7   ,   11  ,   9   ,   63  ,   35  ,   181 ,   301 ,   945 ,   853 ,   877 ,   2983    ,   9945    ,   31237   ,   12011   ,0 };
            const unsigned long dim4017Kuo3Init[]   =   {   1   ,   1   ,   1   ,   5   ,   23  ,   17  ,   111 ,   15  ,   429 ,   925 ,   299 ,   173 ,   3379    ,   2407    ,   11999   ,   44485   ,0 };
            const unsigned long dim4018Kuo3Init[]   =   {   1   ,   1   ,   5   ,   5   ,   31  ,   55  ,   103 ,   227 ,   93  ,   15  ,   399 ,   1559    ,   3107    ,   5729    ,   24069   ,   55711   ,0 };
            const unsigned long dim4019Kuo3Init[]   =   {   1   ,   3   ,   7   ,   7   ,   25  ,   5   ,   107 ,   57  ,   23  ,   1005    ,   27  ,   2881    ,   2591    ,   10939   ,   8613    ,   50351   ,0 };
            const unsigned long dim4020Kuo3Init[]   =   {   1   ,   1   ,   5   ,   1   ,   19  ,   23  ,   107 ,   65  ,   11  ,   643 ,   1337    ,   3197    ,   1583    ,   8797    ,   6047    ,   55749   ,0 };
            const unsigned long dim4021Kuo3Init[]   =   {   1   ,   1   ,   1   ,   9   ,   3   ,   61  ,   31  ,   67  ,   369 ,   755 ,   1263    ,   2711    ,   1341    ,   15465   ,   25337   ,   20607   ,0 };
            const unsigned long dim4022Kuo3Init[]   =   {   1   ,   3   ,   7   ,   7   ,   23  ,   53  ,   95  ,   223 ,   225 ,   627 ,   471 ,   1203    ,   7689    ,   3211    ,   24697   ,   21153   ,0 };
            const unsigned long dim4023Kuo3Init[]   =   {   1   ,   3   ,   7   ,   1   ,   3   ,   29  ,   45  ,   15  ,   247 ,   45  ,   345 ,   4063    ,   3819    ,   263 ,   9705    ,   27991   ,0 };
            const unsigned long dim4024Kuo3Init[]   =   {   1   ,   3   ,   3   ,   1   ,   3   ,   55  ,   127 ,   247 ,   87  ,   969 ,   1409    ,   2105    ,   4133    ,   1213    ,   2719    ,   37043   ,0 };
            const unsigned long dim4025Kuo3Init[]   =   {   1   ,   1   ,   5   ,   15  ,   9   ,   15  ,   71  ,   9   ,   421 ,   1   ,   1367    ,   2557    ,   2121    ,   2919    ,   29167   ,   7035    ,0 };
            const unsigned long dim4026Kuo3Init[]   =   {   1   ,   3   ,   7   ,   5   ,   21  ,   35  ,   7   ,   1   ,   429 ,   491 ,   1077    ,   499 ,   2915    ,   13217   ,   25153   ,   327 ,0 };
            const unsigned long dim4027Kuo3Init[]   =   {   1   ,   1   ,   3   ,   9   ,   7   ,   57  ,   37  ,   235 ,   97  ,   357 ,   1171    ,   705 ,   7111    ,   5217    ,   13167   ,   38395   ,0 };
            const unsigned long dim4028Kuo3Init[]   =   {   1   ,   3   ,   7   ,   7   ,   9   ,   55  ,   43  ,   173 ,   429 ,   287 ,   355 ,   3009    ,   1667    ,   1845    ,   11477   ,   25437   ,0 };
            const unsigned long dim4029Kuo3Init[]   =   {   1   ,   3   ,   7   ,   3   ,   13  ,   41  ,   41  ,   239 ,   421 ,   347 ,   1423    ,   797 ,   7435    ,   16195   ,   3501    ,   56213   ,0 };
            const unsigned long dim4030Kuo3Init[]   =   {   1   ,   3   ,   7   ,   13  ,   25  ,   59  ,   67  ,   49  ,   363 ,   125 ,   365 ,   2517    ,   915 ,   2865    ,   25361   ,   24823   ,0 };
            const unsigned long dim4031Kuo3Init[]   =   {   1   ,   3   ,   7   ,   1   ,   31  ,   55  ,   109 ,   213 ,   19  ,   843 ,   1457    ,   851 ,   7245    ,   13587   ,   30111   ,   58737   ,0 };
            const unsigned long dim4032Kuo3Init[]   =   {   1   ,   1   ,   5   ,   5   ,   7   ,   29  ,   111 ,   121 ,   511 ,   401 ,   411 ,   3073    ,   3617    ,   4447    ,   12443   ,   49051   ,0 };
            const unsigned long dim4033Kuo3Init[]   =   {   1   ,   3   ,   3   ,   9   ,   5   ,   55  ,   89  ,   231 ,   343 ,   869 ,   465 ,   1421    ,   701 ,   13565   ,   12469   ,   62381   ,0 };
            const unsigned long dim4034Kuo3Init[]   =   {   1   ,   3   ,   3   ,   13  ,   7   ,   9   ,   91  ,   147 ,   37  ,   917 ,   1717    ,   2221    ,   5795    ,   7127    ,   21599   ,   28335   ,0 };
            const unsigned long dim4035Kuo3Init[]   =   {   1   ,   3   ,   1   ,   3   ,   9   ,   51  ,   109 ,   43  ,   411 ,   477 ,   549 ,   1901    ,   4295    ,   2915    ,   5701    ,   29005   ,0 };
            const unsigned long dim4036Kuo3Init[]   =   {   1   ,   1   ,   3   ,   9   ,   11  ,   51  ,   5   ,   81  ,   413 ,   461 ,   143 ,   2695    ,   2091    ,   8435    ,   14217   ,   60831   ,0 };
            const unsigned long dim4037Kuo3Init[]   =   {   1   ,   3   ,   5   ,   13  ,   1   ,   15  ,   17  ,   239 ,   219 ,   75  ,   339 ,   1565    ,   5029    ,   6351    ,   19429   ,   21483   ,0 };
            const unsigned long dim4038Kuo3Init[]   =   {   1   ,   1   ,   3   ,   5   ,   13  ,   7   ,   101 ,   215 ,   395 ,   387 ,   1215    ,   1967    ,   4141    ,   643 ,   14623   ,   58609   ,0 };
            const unsigned long dim4039Kuo3Init[]   =   {   1   ,   3   ,   5   ,   13  ,   17  ,   29  ,   85  ,   201 ,   327 ,   855 ,   1971    ,   335 ,   6201    ,   5933    ,   6131    ,   3179    ,0 };
            const unsigned long dim4040Kuo3Init[]   =   {   1   ,   3   ,   3   ,   13  ,   3   ,   49  ,   25  ,   223 ,   207 ,   477 ,   671 ,   1881    ,   3639    ,   10867   ,   22953   ,   58725   ,0 };
            const unsigned long dim4041Kuo3Init[]   =   {   1   ,   1   ,   1   ,   5   ,   21  ,   23  ,   31  ,   57  ,   387 ,   881 ,   343 ,   1781    ,   3241    ,   10939   ,   5569    ,   37555   ,0 };
            const unsigned long dim4042Kuo3Init[]   =   {   1   ,   1   ,   3   ,   1   ,   25  ,   3   ,   117 ,   153 ,   155 ,   61  ,   127 ,   3871    ,   7883    ,   11791   ,   18529   ,   55395   ,0 };
            const unsigned long dim4043Kuo3Init[]   =   {   1   ,   1   ,   7   ,   9   ,   15  ,   15  ,   89  ,   197 ,   109 ,   221 ,   317 ,   2827    ,   2159    ,   847 ,   2527    ,   9463    ,0 };
            const unsigned long dim4044Kuo3Init[]   =   {   1   ,   3   ,   5   ,   9   ,   25  ,   11  ,   39  ,   185 ,   457 ,   831 ,   637 ,   527 ,   2829    ,   4411    ,   4713    ,   53597   ,0 };
            const unsigned long dim4045Kuo3Init[]   =   {   1   ,   3   ,   7   ,   15  ,   5   ,   27  ,   45  ,   123 ,   385 ,   217 ,   907 ,   2877    ,   2703    ,   15987   ,   24963   ,   41659   ,0 };
            const unsigned long dim4046Kuo3Init[]   =   {   1   ,   3   ,   1   ,   13  ,   9   ,   31  ,   7   ,   225 ,   77  ,   227 ,   175 ,   1471    ,   3305    ,   4521    ,   6345    ,   55509   ,0 };
            const unsigned long dim4047Kuo3Init[]   =   {   1   ,   3   ,   1   ,   5   ,   21  ,   33  ,   37  ,   3   ,   121 ,   605 ,   1121    ,   537 ,   1771    ,   14415   ,   8323    ,   21795   ,0 };
            const unsigned long dim4048Kuo3Init[]   =   {   1   ,   3   ,   7   ,   5   ,   23  ,   23  ,   45  ,   231 ,   97  ,   535 ,   861 ,   1349    ,   6611    ,   7445    ,   25231   ,   50737   ,0 };
            const unsigned long dim4049Kuo3Init[]   =   {   1   ,   3   ,   1   ,   9   ,   11  ,   13  ,   61  ,   155 ,   483 ,   391 ,   615 ,   899 ,   5099    ,   4895    ,   2919    ,   60239   ,0 };
            const unsigned long dim4050Kuo3Init[]   =   {   1   ,   3   ,   3   ,   1   ,   31  ,   9   ,   47  ,   209 ,   405 ,   493 ,   103 ,   59  ,   3609    ,   12889   ,   6501    ,   32869   ,0 };
            const unsigned long dim4051Kuo3Init[]   =   {   1   ,   1   ,   7   ,   13  ,   15  ,   57  ,   23  ,   153 ,   57  ,   369 ,   1833    ,   1287    ,   4487    ,   3195    ,   19587   ,   24373   ,0 };
            const unsigned long dim4052Kuo3Init[]   =   {   1   ,   3   ,   7   ,   3   ,   25  ,   45  ,   51  ,   117 ,   293 ,   701 ,   1909    ,   659 ,   637 ,   3375    ,   23223   ,   29707   ,0 };
            const unsigned long dim4053Kuo3Init[]   =   {   1   ,   3   ,   7   ,   15  ,   1   ,   27  ,   13  ,   227 ,   369 ,   143 ,   1167    ,   1787    ,   4815    ,   1143    ,   7789    ,   16467   ,0 };
            const unsigned long dim4054Kuo3Init[]   =   {   1   ,   1   ,   5   ,   1   ,   13  ,   5   ,   63  ,   75  ,   13  ,   411 ,   1123    ,   459 ,   6323    ,   10445   ,   1303    ,   43103   ,0 };
            const unsigned long dim4055Kuo3Init[]   =   {   1   ,   3   ,   1   ,   9   ,   29  ,   21  ,   93  ,   49  ,   135 ,   525 ,   121 ,   1299    ,   5931    ,   13115   ,   32567   ,   60605   ,0 };
            const unsigned long dim4056Kuo3Init[]   =   {   1   ,   3   ,   3   ,   11  ,   25  ,   59  ,   77  ,   181 ,   39  ,   345 ,   1375    ,   2067    ,   3909    ,   8023    ,   6409    ,   12571   ,0 };
            const unsigned long dim4057Kuo3Init[]   =   {   1   ,   3   ,   5   ,   15  ,   15  ,   55  ,   95  ,   183 ,   269 ,   931 ,   285 ,   1945    ,   2105    ,   10589   ,   19787   ,   64461   ,0 };
            const unsigned long dim4058Kuo3Init[]   =   {   1   ,   1   ,   1   ,   5   ,   31  ,   19  ,   51  ,   109 ,   455 ,   877 ,   905 ,   2403    ,   2021    ,   573 ,   29837   ,   46871   ,0 };
            const unsigned long dim4059Kuo3Init[]   =   {   1   ,   3   ,   1   ,   9   ,   21  ,   23  ,   9   ,   201 ,   143 ,   661 ,   289 ,   621 ,   6729    ,   9573    ,   6359    ,   14981   ,0 };
            const unsigned long dim4060Kuo3Init[]   =   {   1   ,   3   ,   5   ,   5   ,   15  ,   53  ,   41  ,   57  ,   201 ,   37  ,   2047    ,   4045    ,   6501    ,   15641   ,   22031   ,   25851   ,0 };
            const unsigned long dim4061Kuo3Init[]   =   {   1   ,   1   ,   7   ,   7   ,   17  ,   3   ,   31  ,   31  ,   165 ,   765 ,   1101    ,   1161    ,   7247    ,   16383   ,   27715   ,   20835   ,0 };
            const unsigned long dim4062Kuo3Init[]   =   {   1   ,   3   ,   3   ,   13  ,   7   ,   41  ,   81  ,   55  ,   345 ,   913 ,   1457    ,   3801    ,   6937    ,   3781    ,   21125   ,   36899   ,0 };
            const unsigned long dim4063Kuo3Init[]   =   {   1   ,   1   ,   3   ,   11  ,   21  ,   61  ,   57  ,   89  ,   341 ,   877 ,   639 ,   3051    ,   6541    ,   2133    ,   11187   ,   9613    ,0 };
            const unsigned long dim4064Kuo3Init[]   =   {   1   ,   3   ,   3   ,   5   ,   3   ,   29  ,   3   ,   193 ,   331 ,   333 ,   1613    ,   3755    ,   6489    ,   4851    ,   10421   ,   11257   ,0 };
            const unsigned long dim4065Kuo3Init[]   =   {   1   ,   3   ,   3   ,   13  ,   15  ,   29  ,   115 ,   135 ,   283 ,   363 ,   905 ,   3199    ,   733 ,   7785    ,   5237    ,   20657   ,0 };
            const unsigned long dim4066Kuo3Init[]   =   {   1   ,   3   ,   5   ,   7   ,   3   ,   31  ,   89  ,   219 ,   61  ,   573 ,   1773    ,   677 ,   7447    ,   4625    ,   17011   ,   57851   ,0 };
            const unsigned long dim4067Kuo3Init[]   =   {   1   ,   3   ,   1   ,   7   ,   9   ,   31  ,   121 ,   219 ,   27  ,   583 ,   1769    ,   3667    ,   4415    ,   6789    ,   28069   ,   11833   ,0 };
            const unsigned long dim4068Kuo3Init[]   =   {   1   ,   1   ,   3   ,   1   ,   29  ,   11  ,   63  ,   201 ,   33  ,   153 ,   1075    ,   641 ,   5923    ,   10485   ,   24573   ,   29423   ,0 };
            const unsigned long dim4069Kuo3Init[]   =   {   1   ,   1   ,   3   ,   13  ,   17  ,   5   ,   87  ,   237 ,   243 ,   231 ,   915 ,   1755    ,   4021    ,   2127    ,   26147   ,   28397   ,0 };
            const unsigned long dim4070Kuo3Init[]   =   {   1   ,   3   ,   3   ,   3   ,   21  ,   49  ,   53  ,   163 ,   415 ,   553 ,   505 ,   3309    ,   4987    ,   16163   ,   14439   ,   30921   ,0 };
            const unsigned long dim4071Kuo3Init[]   =   {   1   ,   3   ,   1   ,   13  ,   9   ,   59  ,   89  ,   113 ,   203 ,   485 ,   177 ,   2243    ,   5267    ,   7105    ,   5297    ,   23333   ,0 };
            const unsigned long dim4072Kuo3Init[]   =   {   1   ,   3   ,   1   ,   1   ,   13  ,   21  ,   107 ,   183 ,   243 ,   985 ,   499 ,   385 ,   2409    ,   9317    ,   17393   ,   1913    ,0 };
            const unsigned long dim4073Kuo3Init[]   =   {   1   ,   3   ,   3   ,   13  ,   21  ,   47  ,   59  ,   21  ,   71  ,   523 ,   1843    ,   3131    ,   5907    ,   15401   ,   7957    ,   51365   ,0 };
            const unsigned long dim4074Kuo3Init[]   =   {   1   ,   3   ,   1   ,   3   ,   17  ,   19  ,   75  ,   195 ,   171 ,   11  ,   57  ,   303 ,   6801    ,   1825    ,   25027   ,   28587   ,0 };
            const unsigned long dim4075Kuo3Init[]   =   {   1   ,   3   ,   5   ,   15  ,   13  ,   47  ,   19  ,   161 ,   47  ,   385 ,   1227    ,   261 ,   5855    ,   1439    ,   18605   ,   55645   ,0 };
            const unsigned long dim4076Kuo3Init[]   =   {   1   ,   1   ,   5   ,   11  ,   23  ,   13  ,   87  ,   197 ,   507 ,   921 ,   1813    ,   579 ,   3585    ,   5117    ,   30913   ,   5739    ,0 };
            const unsigned long dim4077Kuo3Init[]   =   {   1   ,   1   ,   1   ,   5   ,   21  ,   7   ,   3   ,   103 ,   475 ,   929 ,   1189    ,   1005    ,   2933    ,   1377    ,   30183   ,   58513   ,0 };
            const unsigned long dim4078Kuo3Init[]   =   {   1   ,   1   ,   3   ,   11  ,   27  ,   51  ,   25  ,   139 ,   431 ,   275 ,   1541    ,   1459    ,   5787    ,   9865    ,   15077   ,   25965   ,0 };
            const unsigned long dim4079Kuo3Init[]   =   {   1   ,   3   ,   7   ,   15  ,   19  ,   33  ,   85  ,   5   ,   119 ,   523 ,   11  ,   1459    ,   3807    ,   3959    ,   31321   ,   1905    ,0 };
            const unsigned long dim4080Kuo3Init[]   =   {   1   ,   3   ,   1   ,   15  ,   9   ,   25  ,   107 ,   51  ,   149 ,   437 ,   995 ,   397 ,   753 ,   12421   ,   30637   ,   22865   ,0 };
            const unsigned long dim4081Kuo3Init[]   =   {   1   ,   3   ,   1   ,   9   ,   25  ,   13  ,   11  ,   129 ,   371 ,   497 ,   597 ,   1257    ,   7291    ,   4021    ,   16223   ,   21489   ,0 };
            const unsigned long dim4082Kuo3Init[]   =   {   1   ,   3   ,   5   ,   1   ,   5   ,   17  ,   119 ,   107 ,   267 ,   27  ,   87  ,   3641    ,   593 ,   4671    ,   16129   ,   25645   ,0 };
            const unsigned long dim4083Kuo3Init[]   =   {   1   ,   3   ,   1   ,   7   ,   19  ,   55  ,   39  ,   159 ,   283 ,   551 ,   321 ,   1979    ,   1259    ,   10049   ,   15195   ,   62799   ,0 };
            const unsigned long dim4084Kuo3Init[]   =   {   1   ,   3   ,   7   ,   3   ,   23  ,   43  ,   63  ,   187 ,   399 ,   725 ,   419 ,   1759    ,   5477    ,   4743    ,   13397   ,   21347   ,0 };
            const unsigned long dim4085Kuo3Init[]   =   {   1   ,   1   ,   1   ,   7   ,   3   ,   45  ,   55  ,   207 ,   33  ,   695 ,   1109    ,   1427    ,   5425    ,   4743    ,   28949   ,   54665   ,0 };
            const unsigned long dim4086Kuo3Init[]   =   {   1   ,   1   ,   5   ,   5   ,   19  ,   1   ,   97  ,   151 ,   465 ,   767 ,   1931    ,   1097    ,   7001    ,   5761    ,   30323   ,   17501   ,0 };
            const unsigned long dim4087Kuo3Init[]   =   {   1   ,   3   ,   1   ,   9   ,   9   ,   11  ,   9   ,   165 ,   181 ,   711 ,   45  ,   1163    ,   11  ,   9127    ,   31361   ,   11259   ,0 };
            const unsigned long dim4088Kuo3Init[]   =   {   1   ,   3   ,   3   ,   1   ,   23  ,   23  ,   7   ,   115 ,   243 ,   281 ,   847 ,   3091    ,   6383    ,   4593    ,   13507   ,   58001   ,0 };
            const unsigned long dim4089Kuo3Init[]   =   {   1   ,   1   ,   3   ,   13  ,   27  ,   3   ,   123 ,   91  ,   47  ,   483 ,   1913    ,   2139    ,   3877    ,   16373   ,   31969   ,   42977   ,0 };
            const unsigned long dim4090Kuo3Init[]   =   {   1   ,   1   ,   1   ,   3   ,   11  ,   9   ,   35  ,   3   ,   155 ,   585 ,   1193    ,   1887    ,   1657    ,   2479    ,   18803   ,   47309   ,0 };
            const unsigned long dim4091Kuo3Init[]   =   {   1   ,   3   ,   3   ,   5   ,   5   ,   57  ,   79  ,   77  ,   205 ,   361 ,   505 ,   2505    ,   2959    ,   1345    ,   8165    ,   46001   ,0 };
            const unsigned long dim4092Kuo3Init[]   =   {   1   ,   1   ,   1   ,   5   ,   29  ,   25  ,   35  ,   209 ,   193 ,   177 ,   291 ,   1095    ,   7137    ,   4323    ,   673 ,   37571   ,0 };
            const unsigned long dim4093Kuo3Init[]   =   {   1   ,   3   ,   5   ,   5   ,   27  ,   29  ,   107 ,   161 ,   93  ,   585 ,   1061    ,   3281    ,   5565    ,   5093    ,   4511    ,   45327   ,0 };
            const unsigned long dim4094Kuo3Init[]   =   {   1   ,   3   ,   7   ,   7   ,   15  ,   43  ,   41  ,   61  ,   425 ,   451 ,   1775    ,   345 ,   7851    ,   12907   ,   9467    ,   25705   ,0 };
            const unsigned long dim4095Kuo3Init[]   =   {   1   ,   1   ,   1   ,   15  ,   31  ,   37  ,   105 ,   227 ,   155 ,   201 ,   653 ,   2343    ,   4601    ,   1331    ,   9193    ,   28423   ,0 };
            const unsigned long dim4096Kuo3Init[]   =   {   1   ,   1   ,   5   ,   9   ,   31  ,   61  ,   19  ,   165 ,   491 ,   121 ,   159 ,   485 ,   7847    ,   4993    ,   30405   ,   22889   ,0 };
            const unsigned long dim4097Kuo3Init[]   =   {   1   ,   3   ,   1   ,   5   ,   27  ,   35  ,   81  ,   231 ,   345 ,   519 ,   849 ,   539 ,   627 ,   16141   ,   26509   ,   8965    ,0 };
            const unsigned long dim4098Kuo3Init[]   =   {   1   ,   1   ,   3   ,   3   ,   7   ,   33  ,   81  ,   51  ,   415 ,   785 ,   1617    ,   669 ,   4749    ,   737 ,   15557   ,   61779   ,0 };
            const unsigned long dim4099Kuo3Init[]   =   {   1   ,   3   ,   5   ,   15  ,   5   ,   39  ,   47  ,   199 ,   313 ,   193 ,   1797    ,   2425    ,   327 ,   5111    ,   275 ,   45933   ,0 };
            const unsigned long dim4100Kuo3Init[]   =   {   1   ,   3   ,   1   ,   5   ,   1   ,   27  ,   61  ,   1   ,   435 ,   1005    ,   665 ,   3099    ,   1153    ,   12721   ,   25537   ,   62559   ,0 };
            const unsigned long dim4101Kuo3Init[]   =   {   1   ,   3   ,   7   ,   9   ,   29  ,   43  ,   97  ,   213 ,   43  ,   333 ,   25  ,   3085    ,   2709    ,   5917    ,   5887    ,   3847    ,0 };
            const unsigned long dim4102Kuo3Init[]   =   {   1   ,   1   ,   7   ,   5   ,   17  ,   27  ,   93  ,   45  ,   205 ,   29  ,   2035    ,   4053    ,   6525    ,   12321   ,   26175   ,   43979   ,0 };
            const unsigned long dim4103Kuo3Init[]   =   {   1   ,   1   ,   3   ,   3   ,   9   ,   41  ,   23  ,   215 ,   163 ,   531 ,   1753    ,   2377    ,   5533    ,   8681    ,   12213   ,   36789   ,0 };
            const unsigned long dim4104Kuo3Init[]   =   {   1   ,   3   ,   1   ,   3   ,   7   ,   13  ,   25  ,   111 ,   439 ,   695 ,   1855    ,   2437    ,   4477    ,   10207   ,   4089    ,   43521   ,0 };
            const unsigned long dim4105Kuo3Init[]   =   {   1   ,   1   ,   3   ,   9   ,   23  ,   7   ,   85  ,   123 ,   361 ,   913 ,   1817    ,   1547    ,   7969    ,   9731    ,   10553   ,   53879   ,0 };
            const unsigned long dim4106Kuo3Init[]   =   {   1   ,   1   ,   7   ,   13  ,   29  ,   25  ,   5   ,   129 ,   211 ,   511 ,   45  ,   907 ,   1997    ,   15657   ,   9003    ,   47267   ,0 };
            const unsigned long dim4107Kuo3Init[]   =   {   1   ,   3   ,   1   ,   9   ,   15  ,   49  ,   13  ,   131 ,   43  ,   295 ,   1569    ,   2675    ,   7857    ,   15515   ,   9189    ,   33061   ,0 };
            const unsigned long dim4108Kuo3Init[]   =   {   1   ,   3   ,   7   ,   9   ,   29  ,   3   ,   85  ,   125 ,   403 ,   583 ,   1507    ,   73  ,   4701    ,   1117    ,   12271   ,   11387   ,0 };
            const unsigned long dim4109Kuo3Init[]   =   {   1   ,   1   ,   3   ,   5   ,   3   ,   15  ,   1   ,   63  ,   319 ,   43  ,   913 ,   3029    ,   557 ,   3087    ,   18605   ,   39477   ,0 };
            const unsigned long dim4110Kuo3Init[]   =   {   1   ,   1   ,   7   ,   13  ,   31  ,   51  ,   99  ,   103 ,   343 ,   835 ,   313 ,   2025    ,   7109    ,   5373    ,   20821   ,   55649   ,0 };
            const unsigned long dim4111Kuo3Init[]   =   {   1   ,   1   ,   3   ,   7   ,   7   ,   21  ,   53  ,   123 ,   189 ,   1013    ,   427 ,   2773    ,   6567    ,   5241    ,   29647   ,   41591   ,0 };
            const unsigned long dim4112Kuo3Init[]   =   {   1   ,   1   ,   1   ,   3   ,   29  ,   53  ,   127 ,   159 ,   327 ,   999 ,   1357    ,   989 ,   4401    ,   10949   ,   32555   ,   5097    ,0 };
            const unsigned long dim4113Kuo3Init[]   =   {   1   ,   3   ,   1   ,   9   ,   31  ,   11  ,   75  ,   47  ,   307 ,   251 ,   1845    ,   1543    ,   6901    ,   3959    ,   18371   ,   40149   ,0 };
            const unsigned long dim4114Kuo3Init[]   =   {   1   ,   1   ,   7   ,   11  ,   31  ,   47  ,   81  ,   17  ,   17  ,   21  ,   807 ,   3115    ,   5447    ,   11191   ,   2357    ,   2263    ,0 };
            const unsigned long dim4115Kuo3Init[]   =   {   1   ,   1   ,   5   ,   13  ,   25  ,   41  ,   93  ,   135 ,   251 ,   165 ,   805 ,   1805    ,   2895    ,   11725   ,   6389    ,   21875   ,0 };
            const unsigned long dim4116Kuo3Init[]   =   {   1   ,   1   ,   5   ,   5   ,   19  ,   37  ,   19  ,   247 ,   25  ,   241 ,   767 ,   2883    ,   5951    ,   531 ,   22893   ,   47493   ,0 };
            const unsigned long dim4117Kuo3Init[]   =   {   1   ,   3   ,   1   ,   11  ,   31  ,   49  ,   23  ,   199 ,   337 ,   311 ,   1929    ,   2703    ,   4339    ,   6477    ,   1821    ,   4729    ,0 };
            const unsigned long dim4118Kuo3Init[]   =   {   1   ,   1   ,   3   ,   11  ,   7   ,   59  ,   119 ,   3   ,   289 ,   951 ,   1387    ,   2999    ,   6617    ,   14551   ,   22081   ,   14187   ,0 };
            const unsigned long dim4119Kuo3Init[]   =   {   1   ,   1   ,   5   ,   5   ,   13  ,   33  ,   37  ,   97  ,   501 ,   247 ,   651 ,   209 ,   7517    ,   2847    ,   3549    ,   16479   ,0 };
            const unsigned long dim4120Kuo3Init[]   =   {   1   ,   1   ,   3   ,   1   ,   25  ,   45  ,   83  ,   207 ,   115 ,   1023    ,   543 ,   2851    ,   3383    ,   4177    ,   25567   ,   2797    ,0 };
            const unsigned long dim4121Kuo3Init[]   =   {   1   ,   3   ,   3   ,   5   ,   7   ,   9   ,   127 ,   243 ,   501 ,   653 ,   1071    ,   2547    ,   6995    ,   10811   ,   4791    ,   50157   ,0 };
            const unsigned long dim4122Kuo3Init[]   =   {   1   ,   1   ,   1   ,   11  ,   27  ,   61  ,   65  ,   117 ,   373 ,   191 ,   1643    ,   1057    ,   6013    ,   15167   ,   18815   ,   4225    ,0 };
            const unsigned long dim4123Kuo3Init[]   =   {   1   ,   1   ,   1   ,   11  ,   9   ,   25  ,   105 ,   225 ,   247 ,   609 ,   325 ,   2761    ,   515 ,   9985    ,   6977    ,   57207   ,0 };
            const unsigned long dim4124Kuo3Init[]   =   {   1   ,   1   ,   7   ,   15  ,   5   ,   43  ,   33  ,   45  ,   301 ,   1017    ,   1881    ,   83  ,   95  ,   10417   ,   2307    ,   6915    ,0 };
            const unsigned long dim4125Kuo3Init[]   =   {   1   ,   3   ,   1   ,   3   ,   25  ,   43  ,   27  ,   81  ,   249 ,   239 ,   1671    ,   1215    ,   3139    ,   7961    ,   28509   ,   27099   ,0 };
            const unsigned long dim4126Kuo3Init[]   =   {   1   ,   1   ,   3   ,   1   ,   17  ,   29  ,   49  ,   81  ,   503 ,   429 ,   2037    ,   2243    ,   7509    ,   12359   ,   14681   ,   21661   ,0 };
            const unsigned long dim4127Kuo3Init[]   =   {   1   ,   3   ,   5   ,   3   ,   3   ,   63  ,   71  ,   71  ,   375 ,   999 ,   11  ,   2475    ,   4491    ,   7439    ,   5915    ,   63169   ,0 };
            const unsigned long dim4128Kuo3Init[]   =   {   1   ,   1   ,   7   ,   13  ,   31  ,   35  ,   91  ,   27  ,   151 ,   893 ,   525 ,   1913    ,   2795    ,   10377   ,   1173    ,   44175   ,0 };
            const unsigned long dim4129Kuo3Init[]   =   {   1   ,   1   ,   7   ,   15  ,   23  ,   11  ,   69  ,   77  ,   133 ,   307 ,   1589    ,   2795    ,   1183    ,   1491    ,   4661    ,   51315   ,0 };
            const unsigned long dim4130Kuo3Init[]   =   {   1   ,   1   ,   7   ,   3   ,   27  ,   37  ,   9   ,   67  ,   481 ,   113 ,   7   ,   1679    ,   217 ,   6195    ,   18147   ,   50839   ,0 };
            const unsigned long dim4131Kuo3Init[]   =   {   1   ,   1   ,   5   ,   1   ,   3   ,   37  ,   53  ,   127 ,   19  ,   177 ,   1851    ,   921 ,   1009    ,   9577    ,   24637   ,   2491    ,0 };
            const unsigned long dim4132Kuo3Init[]   =   {   1   ,   3   ,   7   ,   5   ,   19  ,   33  ,   65  ,   25  ,   333 ,   749 ,   1577    ,   143 ,   6731    ,   2411    ,   25579   ,   57987   ,0 };
            const unsigned long dim4133Kuo3Init[]   =   {   1   ,   1   ,   5   ,   3   ,   29  ,   45  ,   89  ,   179 ,   147 ,   411 ,   1831    ,   2209    ,   499 ,   3141    ,   17357   ,   23229   ,0 };
            const unsigned long dim4134Kuo3Init[]   =   {   1   ,   1   ,   5   ,   1   ,   25  ,   41  ,   105 ,   165 ,   369 ,   241 ,   299 ,   875 ,   1231    ,   469 ,   4203    ,   27117   ,0 };
            const unsigned long dim4135Kuo3Init[]   =   {   1   ,   3   ,   1   ,   5   ,   29  ,   35  ,   95  ,   235 ,   229 ,   797 ,   411 ,   71  ,   4571    ,   2341    ,   4095    ,   28185   ,0 };
            const unsigned long dim4136Kuo3Init[]   =   {   1   ,   1   ,   1   ,   7   ,   21  ,   33  ,   23  ,   59  ,   195 ,   539 ,   1131    ,   3799    ,   8093    ,   8247    ,   6283    ,   44143   ,0 };
            const unsigned long dim4137Kuo3Init[]   =   {   1   ,   1   ,   5   ,   15  ,   11  ,   15  ,   45  ,   99  ,   73  ,   983 ,   1047    ,   361 ,   2925    ,   1985    ,   20379   ,   63235   ,0 };
            const unsigned long dim4138Kuo3Init[]   =   {   1   ,   3   ,   7   ,   5   ,   25  ,   7   ,   29  ,   129 ,   257 ,   1015    ,   1823    ,   23  ,   5031    ,   12153   ,   13511   ,   57897   ,0 };
            const unsigned long dim4139Kuo3Init[]   =   {   1   ,   1   ,   1   ,   7   ,   15  ,   39  ,   57  ,   173 ,   391 ,   979 ,   569 ,   2051    ,   3221    ,   9561    ,   24331   ,   39791   ,0 };
            const unsigned long dim4140Kuo3Init[]   =   {   1   ,   1   ,   3   ,   15  ,   1   ,   51  ,   25  ,   79  ,   373 ,   639 ,   5   ,   3237    ,   2983    ,   12491   ,   15991   ,   12467   ,0 };
            const unsigned long dim4141Kuo3Init[]   =   {   1   ,   1   ,   7   ,   15  ,   17  ,   21  ,   103 ,   55  ,   233 ,   605 ,   597 ,   1867    ,   2147    ,   11047   ,   1173    ,   27237   ,0 };
            const unsigned long dim4142Kuo3Init[]   =   {   1   ,   1   ,   7   ,   11  ,   11  ,   25  ,   43  ,   63  ,   143 ,   829 ,   1357    ,   2199    ,   5219    ,   15981   ,   15615   ,   22207   ,0 };
            const unsigned long dim4143Kuo3Init[]   =   {   1   ,   1   ,   3   ,   13  ,   17  ,   53  ,   121 ,   131 ,   501 ,   355 ,   533 ,   2343    ,   6515    ,   9115    ,   29065   ,   64569   ,0 };
            const unsigned long dim4144Kuo3Init[]   =   {   1   ,   3   ,   5   ,   15  ,   27  ,   57  ,   5   ,   105 ,   327 ,   861 ,   931 ,   3085    ,   5051    ,   12687   ,   22409   ,   57487   ,0 };
            const unsigned long dim4145Kuo3Init[]   =   {   1   ,   1   ,   5   ,   15  ,   1   ,   53  ,   65  ,   97  ,   315 ,   413 ,   37  ,   2563    ,   4805    ,   2327    ,   10261   ,   7281    ,0 };
            const unsigned long dim4146Kuo3Init[]   =   {   1   ,   3   ,   5   ,   9   ,   19  ,   57  ,   15  ,   75  ,   161 ,   245 ,   1215    ,   1735    ,   4567    ,   109 ,   8467    ,   43139   ,0 };
            const unsigned long dim4147Kuo3Init[]   =   {   1   ,   3   ,   1   ,   11  ,   21  ,   55  ,   95  ,   11  ,   193 ,   959 ,   793 ,   3321    ,   5865    ,   301 ,   7887    ,   7389    ,0 };
            const unsigned long dim4148Kuo3Init[]   =   {   1   ,   1   ,   1   ,   7   ,   31  ,   35  ,   27  ,   89  ,   247 ,   491 ,   1087    ,   3955    ,   6317    ,   1337    ,   21619   ,   11739   ,0 };
            const unsigned long dim4149Kuo3Init[]   =   {   1   ,   1   ,   7   ,   13  ,   21  ,   59  ,   53  ,   211 ,   403 ,   851 ,   887 ,   1457    ,   539 ,   4493    ,   16571   ,   55429   ,0 };
            const unsigned long dim4150Kuo3Init[]   =   {   1   ,   1   ,   1   ,   7   ,   23  ,   33  ,   69  ,   129 ,   33  ,   203 ,   267 ,   3261    ,   4179    ,   9653    ,   22523   ,   18779   ,0 };
            const unsigned long dim4151Kuo3Init[]   =   {   1   ,   1   ,   3   ,   9   ,   31  ,   35  ,   7   ,   189 ,   253 ,   583 ,   1475    ,   939 ,   2977    ,   5573    ,   20047   ,   38915   ,0 };
            const unsigned long dim4152Kuo3Init[]   =   {   1   ,   3   ,   3   ,   13  ,   19  ,   25  ,   61  ,   237 ,   479 ,   821 ,   1173    ,   599 ,   7035    ,   12695   ,   433 ,   26499   ,0 };
            const unsigned long dim4153Kuo3Init[]   =   {   1   ,   3   ,   7   ,   3   ,   27  ,   47  ,   29  ,   31  ,   327 ,   687 ,   495 ,   987 ,   1655    ,   951 ,   29265   ,   28641   ,0 };
            const unsigned long dim4154Kuo3Init[]   =   {   1   ,   1   ,   3   ,   1   ,   25  ,   29  ,   15  ,   139 ,   417 ,   837 ,   2011    ,   3703    ,   7809    ,   421 ,   207 ,   18961   ,0 };
            const unsigned long dim4155Kuo3Init[]   =   {   1   ,   3   ,   3   ,   13  ,   7   ,   29  ,   19  ,   63  ,   257 ,   695 ,   161 ,   3023    ,   149 ,   10707   ,   20873   ,   61881   ,0 };
            const unsigned long dim4156Kuo3Init[]   =   {   1   ,   3   ,   3   ,   7   ,   5   ,   9   ,   69  ,   31  ,   487 ,   525 ,   1441    ,   5   ,   3891    ,   14473   ,   22123   ,   40041   ,0 };
            const unsigned long dim4157Kuo3Init[]   =   {   1   ,   3   ,   7   ,   15  ,   13  ,   37  ,   75  ,   181 ,   431 ,   753 ,   1587    ,   595 ,   2567    ,   8155    ,   23199   ,   50647   ,0 };
            const unsigned long dim4158Kuo3Init[]   =   {   1   ,   1   ,   5   ,   1   ,   21  ,   63  ,   39  ,   245 ,   171 ,   217 ,   1643    ,   887 ,   2349    ,   2415    ,   31673   ,   51007   ,0 };
            const unsigned long dim4159Kuo3Init[]   =   {   1   ,   3   ,   5   ,   9   ,   19  ,   33  ,   107 ,   227 ,   187 ,   163 ,   1071    ,   2385    ,   2701    ,   15969   ,   18665   ,   62089   ,0 };
            const unsigned long dim4160Kuo3Init[]   =   {   1   ,   1   ,   1   ,   15  ,   27  ,   47  ,   25  ,   219 ,   139 ,   879 ,   459 ,   1521    ,   2153    ,   3429    ,   28293   ,   57703   ,0 };
            const unsigned long dim4161Kuo3Init[]   =   {   1   ,   3   ,   3   ,   15  ,   15  ,   63  ,   23  ,   237 ,   137 ,   271 ,   117 ,   3047    ,   3973    ,   3379    ,   11527   ,   52555   ,0 };
            const unsigned long dim4162Kuo3Init[]   =   {   1   ,   3   ,   7   ,   1   ,   19  ,   27  ,   19  ,   111 ,   441 ,   279 ,   321 ,   2015    ,   4451    ,   11841   ,   17223   ,   51691   ,0 };
            const unsigned long dim4163Kuo3Init[]   =   {   1   ,   3   ,   3   ,   1   ,   15  ,   35  ,   19  ,   31  ,   277 ,   977 ,   1125    ,   2435    ,   5171    ,   81  ,   31729   ,   18883   ,0 };
            const unsigned long dim4164Kuo3Init[]   =   {   1   ,   3   ,   1   ,   15  ,   19  ,   31  ,   105 ,   13  ,   117 ,   533 ,   191 ,   3651    ,   6915    ,   2423    ,   31819   ,   47709   ,0 };
            const unsigned long dim4165Kuo3Init[]   =   {   1   ,   1   ,   5   ,   11  ,   29  ,   45  ,   17  ,   51  ,   283 ,   297 ,   1605    ,   3009    ,   3137    ,   2665    ,   27799   ,   2293    ,0 };
            const unsigned long dim4166Kuo3Init[]   =   {   1   ,   3   ,   5   ,   13  ,   29  ,   61  ,   91  ,   7   ,   23  ,   835 ,   1809    ,   3283    ,   1091    ,   10443   ,   26547   ,   10685   ,0 };
            const unsigned long dim4167Kuo3Init[]   =   {   1   ,   3   ,   5   ,   1   ,   1   ,   29  ,   121 ,   181 ,   43  ,   827 ,   871 ,   3547    ,   4285    ,   12551   ,   2079    ,   19385   ,0 };
            const unsigned long dim4168Kuo3Init[]   =   {   1   ,   1   ,   1   ,   7   ,   3   ,   53  ,   73  ,   39  ,   1   ,   499 ,   1905    ,   3785    ,   1597    ,   1529    ,   9423    ,   11555   ,0 };
            const unsigned long dim4169Kuo3Init[]   =   {   1   ,   1   ,   5   ,   5   ,   27  ,   25  ,   33  ,   59  ,   327 ,   419 ,   1527    ,   255 ,   6877    ,   3927    ,   11877   ,   21747   ,0 };
            const unsigned long dim4170Kuo3Init[]   =   {   1   ,   1   ,   3   ,   11  ,   27  ,   37  ,   99  ,   161 ,   213 ,   321 ,   1127    ,   523 ,   5327    ,   7885    ,   17403   ,   2097    ,0 };
            const unsigned long dim4171Kuo3Init[]   =   {   1   ,   1   ,   1   ,   3   ,   5   ,   11  ,   119 ,   153 ,   503 ,   1007    ,   187 ,   1705    ,   7499    ,   12739   ,   16185   ,   11121   ,0 };
            const unsigned long dim4172Kuo3Init[]   =   {   1   ,   1   ,   3   ,   1   ,   27  ,   63  ,   29  ,   239 ,   5   ,   377 ,   831 ,   2883    ,   2167    ,   2551    ,   29891   ,   6165    ,0 };
            const unsigned long dim4173Kuo3Init[]   =   {   1   ,   3   ,   7   ,   11  ,   5   ,   57  ,   33  ,   15  ,   425 ,   93  ,   1435    ,   1449    ,   1629    ,   999 ,   27037   ,   51171   ,0 };
            const unsigned long dim4174Kuo3Init[]   =   {   1   ,   1   ,   5   ,   13  ,   1   ,   31  ,   83  ,   19  ,   373 ,   497 ,   1173    ,   775 ,   439 ,   5757    ,   32143   ,   52487   ,0 };
            const unsigned long dim4175Kuo3Init[]   =   {   1   ,   1   ,   5   ,   11  ,   15  ,   43  ,   73  ,   19  ,   259 ,   145 ,   1771    ,   2133    ,   6969    ,   8571    ,   29821   ,   19581   ,0 };
            const unsigned long dim4176Kuo3Init[]   =   {   1   ,   3   ,   3   ,   5   ,   25  ,   63  ,   43  ,   95  ,   215 ,   853 ,   127 ,   3737    ,   4803    ,   10089   ,   3537    ,   14671   ,0 };
            const unsigned long dim4177Kuo3Init[]   =   {   1   ,   1   ,   3   ,   1   ,   31  ,   17  ,   1   ,   139 ,   199 ,   841 ,   1859    ,   1767    ,   1545    ,   4869    ,   14501   ,   11729   ,0 };
            const unsigned long dim4178Kuo3Init[]   =   {   1   ,   3   ,   5   ,   13  ,   29  ,   57  ,   43  ,   143 ,   55  ,   893 ,   1219    ,   2415    ,   7605    ,   11091   ,   3441    ,   45749   ,0 };
            const unsigned long dim4179Kuo3Init[]   =   {   1   ,   3   ,   7   ,   13  ,   27  ,   11  ,   27  ,   15  ,   67  ,   49  ,   967 ,   1919    ,   1773    ,   9135    ,   4607    ,   65495   ,0 };
            const unsigned long dim4180Kuo3Init[]   =   {   1   ,   1   ,   1   ,   5   ,   7   ,   11  ,   63  ,   39  ,   209 ,   41  ,   951 ,   63  ,   1537    ,   9971    ,   9659    ,   17825   ,0 };
            const unsigned long dim4181Kuo3Init[]   =   {   1   ,   3   ,   7   ,   7   ,   27  ,   25  ,   57  ,   53  ,   91  ,   1013    ,   591 ,   261 ,   721 ,   2061    ,   629 ,   37817   ,0 };
            const unsigned long dim4182Kuo3Init[]   =   {   1   ,   1   ,   5   ,   15  ,   23  ,   35  ,   71  ,   43  ,   343 ,   329 ,   295 ,   3523    ,   275 ,   2119    ,   15165   ,   2809    ,0 };
            const unsigned long dim4183Kuo3Init[]   =   {   1   ,   1   ,   5   ,   13  ,   19  ,   3   ,   121 ,   67  ,   165 ,   909 ,   1731    ,   437 ,   8055    ,   11683   ,   12735   ,   29129   ,0 };
            const unsigned long dim4184Kuo3Init[]   =   {   1   ,   3   ,   5   ,   7   ,   21  ,   47  ,   113 ,   55  ,   169 ,   133 ,   2029    ,   825 ,   4079    ,   14651   ,   22373   ,   15009   ,0 };
            const unsigned long dim4185Kuo3Init[]   =   {   1   ,   1   ,   5   ,   7   ,   25  ,   23  ,   61  ,   147 ,   211 ,   665 ,   395 ,   505 ,   3757    ,   14545   ,   10683   ,   53211   ,0 };
            const unsigned long dim4186Kuo3Init[]   =   {   1   ,   1   ,   1   ,   3   ,   17  ,   25  ,   111 ,   171 ,   493 ,   815 ,   847 ,   2109    ,   8013    ,   751 ,   26793   ,   62903   ,0 };
            const unsigned long dim4187Kuo3Init[]   =   {   1   ,   3   ,   7   ,   11  ,   7   ,   27  ,   91  ,   61  ,   385 ,   3   ,   577 ,   3515    ,   4335    ,   4261    ,   6053    ,   62513   ,0 };
            const unsigned long dim4188Kuo3Init[]   =   {   1   ,   3   ,   7   ,   5   ,   17  ,   5   ,   21  ,   217 ,   227 ,   443 ,   1479    ,   2855    ,   2581    ,   6255    ,   23363   ,   37437   ,0 };
            const unsigned long dim4189Kuo3Init[]   =   {   1   ,   1   ,   5   ,   5   ,   5   ,   51  ,   5   ,   213 ,   233 ,   787 ,   719 ,   463 ,   2595    ,   4019    ,   16839   ,   14671   ,0 };
            const unsigned long dim4190Kuo3Init[]   =   {   1   ,   1   ,   1   ,   15  ,   11  ,   5   ,   49  ,   135 ,   85  ,   773 ,   1719    ,   2187    ,   449 ,   14913   ,   19721   ,   59583   ,0 };
            const unsigned long dim4191Kuo3Init[]   =   {   1   ,   1   ,   1   ,   13  ,   3   ,   49  ,   77  ,   217 ,   289 ,   889 ,   63  ,   2547    ,   5547    ,   14005   ,   18657   ,   65109   ,0 };
            const unsigned long dim4192Kuo3Init[]   =   {   1   ,   3   ,   5   ,   9   ,   11  ,   47  ,   19  ,   251 ,   77  ,   623 ,   1253    ,   635 ,   1047    ,   525 ,   8573    ,   25951   ,0 };
            const unsigned long dim4193Kuo3Init[]   =   {   1   ,   1   ,   1   ,   13  ,   5   ,   21  ,   35  ,   187 ,   395 ,   547 ,   2003    ,   3189    ,   4107    ,   567 ,   29239   ,   32647   ,0 };
            const unsigned long dim4194Kuo3Init[]   =   {   1   ,   1   ,   3   ,   13  ,   13  ,   61  ,   59  ,   57  ,   91  ,   881 ,   1665    ,   453 ,   6313    ,   2291    ,   19523   ,   60439   ,0 };
            const unsigned long dim4195Kuo3Init[]   =   {   1   ,   1   ,   3   ,   3   ,   7   ,   7   ,   7   ,   127 ,   411 ,   643 ,   1415    ,   2933    ,   6649    ,   4219    ,   31781   ,   8733    ,0 };
            const unsigned long dim4196Kuo3Init[]   =   {   1   ,   3   ,   1   ,   1   ,   25  ,   29  ,   127 ,   235 ,   271 ,   831 ,   1909    ,   2347    ,   7085    ,   8339    ,   22459   ,   48791   ,0 };
            const unsigned long dim4197Kuo3Init[]   =   {   1   ,   3   ,   7   ,   11  ,   21  ,   19  ,   31  ,   5   ,   509 ,   713 ,   767 ,   1801    ,   6937    ,   1763    ,   20703   ,   10153   ,0 };
            const unsigned long dim4198Kuo3Init[]   =   {   1   ,   3   ,   5   ,   3   ,   9   ,   35  ,   55  ,   215 ,   75  ,   733 ,   1773    ,   1373    ,   43  ,   5045    ,   17675   ,   29615   ,0 };
            const unsigned long dim4199Kuo3Init[]   =   {   1   ,   3   ,   5   ,   13  ,   25  ,   27  ,   33  ,   167 ,   277 ,   231 ,   449 ,   1055    ,   6227    ,   1441    ,   29535   ,   30421   ,0 };
            const unsigned long dim4200Kuo3Init[]   =   {   1   ,   1   ,   5   ,   15  ,   31  ,   63  ,   69  ,   137 ,   41  ,   67  ,   965 ,   3737    ,   3191    ,   2613    ,   8229    ,   47453   ,0 };
            const unsigned long dim4201Kuo3Init[]   =   {   1   ,   1   ,   3   ,   7   ,   9   ,   13  ,   97  ,   31  ,   279 ,   525 ,   185 ,   1673    ,   1269    ,   8305    ,   29745   ,   5799    ,0 };
            const unsigned long dim4202Kuo3Init[]   =   {   1   ,   3   ,   3   ,   13  ,   27  ,   47  ,   97  ,   185 ,   483 ,   787 ,   1049    ,   4073    ,   1985    ,   2129    ,   15099   ,   33307   ,0 };
            const unsigned long dim4203Kuo3Init[]   =   {   1   ,   1   ,   7   ,   13  ,   21  ,   25  ,   97  ,   49  ,   377 ,   75  ,   1875    ,   2389    ,   1381    ,   6691    ,   30447   ,   23969   ,0 };
            const unsigned long dim4204Kuo3Init[]   =   {   1   ,   1   ,   7   ,   1   ,   3   ,   55  ,   111 ,   185 ,   371 ,   331 ,   1471    ,   1211    ,   1523    ,   5393    ,   21763   ,   44333   ,0 };
            const unsigned long dim4205Kuo3Init[]   =   {   1   ,   3   ,   1   ,   3   ,   19  ,   39  ,   35  ,   49  ,   433 ,   937 ,   869 ,   239 ,   7775    ,   9591    ,   16969   ,   43455   ,0 };
            const unsigned long dim4206Kuo3Init[]   =   {   1   ,   3   ,   3   ,   11  ,   21  ,   11  ,   11  ,   201 ,   187 ,   409 ,   1603    ,   3471    ,   1473    ,   5283    ,   7109    ,   53219   ,0 };
            const unsigned long dim4207Kuo3Init[]   =   {   1   ,   3   ,   1   ,   1   ,   13  ,   55  ,   91  ,   37  ,   71  ,   837 ,   1779    ,   1097    ,   7433    ,   5969    ,   22781   ,   28547   ,0 };
            const unsigned long dim4208Kuo3Init[]   =   {   1   ,   3   ,   3   ,   11  ,   9   ,   41  ,   29  ,   125 ,   151 ,   51  ,   667 ,   1387    ,   6695    ,   7957    ,   20469   ,   20987   ,0 };
            const unsigned long dim4209Kuo3Init[]   =   {   1   ,   3   ,   1   ,   7   ,   23  ,   17  ,   91  ,   79  ,   249 ,   123 ,   1141    ,   2201    ,   7139    ,   5749    ,   20655   ,   18111   ,0 };
            const unsigned long dim4210Kuo3Init[]   =   {   1   ,   3   ,   5   ,   11  ,   1   ,   35  ,   5   ,   219 ,   255 ,   549 ,   1159    ,   1641    ,   5063    ,   457 ,   24023   ,   61667   ,0 };
            const unsigned long dim4211Kuo3Init[]   =   {   1   ,   1   ,   5   ,   9   ,   23  ,   45  ,   85  ,   81  ,   41  ,   327 ,   1923    ,   2043    ,   6737    ,   13609   ,   1833    ,   30563   ,0 };
            const unsigned long dim4212Kuo3Init[]   =   {   1   ,   3   ,   3   ,   1   ,   9   ,   15  ,   81  ,   137 ,   353 ,   509 ,   799 ,   787 ,   5371    ,   4863    ,   29781   ,   40405   ,0 };
            const unsigned long dim4213Kuo3Init[]   =   {   1   ,   3   ,   7   ,   5   ,   31  ,   17  ,   109 ,   123 ,   455 ,   155 ,   253 ,   1541    ,   2265    ,   16015   ,   32309   ,   40837   ,0 };
            const unsigned long dim4214Kuo3Init[]   =   {   1   ,   1   ,   5   ,   15  ,   21  ,   59  ,   109 ,   3   ,   367 ,   411 ,   1601    ,   3577    ,   3155    ,   1659    ,   15193   ,   32453   ,0 };
            const unsigned long dim4215Kuo3Init[]   =   {   1   ,   3   ,   5   ,   3   ,   3   ,   21  ,   123 ,   125 ,   163 ,   233 ,   125 ,   947 ,   1023    ,   7435    ,   30777   ,   14615   ,0 };
            const unsigned long dim4216Kuo3Init[]   =   {   1   ,   3   ,   5   ,   9   ,   15  ,   29  ,   31  ,   101 ,   427 ,   379 ,   379 ,   1123    ,   1411    ,   10523   ,   24017   ,   43629   ,0 };
            const unsigned long dim4217Kuo3Init[]   =   {   1   ,   1   ,   5   ,   15  ,   1   ,   27  ,   43  ,   239 ,   91  ,   967 ,   991 ,   2827    ,   6135    ,   1035    ,   25925   ,   42605   ,0 };
            const unsigned long dim4218Kuo3Init[]   =   {   1   ,   3   ,   7   ,   9   ,   11  ,   53  ,   83  ,   127 ,   51  ,   1021    ,   2017    ,   593 ,   1445    ,   14597   ,   6517    ,   19667   ,0 };
            const unsigned long dim4219Kuo3Init[]   =   {   1   ,   3   ,   7   ,   15  ,   9   ,   1   ,   17  ,   235 ,   321 ,   571 ,   1315    ,   3993    ,   5825    ,   8065    ,   13885   ,   48815   ,0 };
            const unsigned long dim4220Kuo3Init[]   =   {   1   ,   1   ,   1   ,   15  ,   3   ,   7   ,   5   ,   197 ,   89  ,   23  ,   711 ,   1601    ,   3577    ,   737 ,   27565   ,   43365   ,0 };
            const unsigned long dim4221Kuo3Init[]   =   {   1   ,   1   ,   5   ,   3   ,   19  ,   35  ,   15  ,   99  ,   177 ,   1013    ,   905 ,   755 ,   667 ,   21  ,   27039   ,   56601   ,0 };
            const unsigned long dim4222Kuo3Init[]   =   {   1   ,   1   ,   3   ,   9   ,   13  ,   37  ,   37  ,   73  ,   75  ,   827 ,   169 ,   2417    ,   5457    ,   13851   ,   5443    ,   15731   ,0 };
            const unsigned long dim4223Kuo3Init[]   =   {   1   ,   1   ,   3   ,   11  ,   11  ,   23  ,   21  ,   159 ,   59  ,   633 ,   1719    ,   203 ,   5655    ,   15887   ,   9067    ,   33597   ,0 };
            const unsigned long dim4224Kuo3Init[]   =   {   1   ,   1   ,   7   ,   7   ,   27  ,   11  ,   55  ,   139 ,   309 ,   721 ,   99  ,   3371    ,   1411    ,   4531    ,   12751   ,   37745   ,0 };
            const unsigned long dim4225Kuo3Init[]   =   {   1   ,   1   ,   1   ,   13  ,   31  ,   33  ,   37  ,   43  ,   125 ,   11  ,   1441    ,   3947    ,   7455    ,   6097    ,   15333   ,   32601   ,0 };
            const unsigned long dim4226Kuo3Init[]   =   {   1   ,   1   ,   3   ,   15  ,   9   ,   43  ,   79  ,   217 ,   247 ,   229 ,   171 ,   3607    ,   4155    ,   1675    ,   3945    ,   41499   ,0 };
            const unsigned long dim4227Kuo3Init[]   =   {   1   ,   1   ,   1   ,   9   ,   5   ,   59  ,   25  ,   175 ,   209 ,   291 ,   1419    ,   2339    ,   7039    ,   1327    ,   19653   ,   11563   ,0 };
            const unsigned long dim4228Kuo3Init[]   =   {   1   ,   1   ,   3   ,   11  ,   5   ,   1   ,   103 ,   197 ,   111 ,   891 ,   163 ,   361 ,   2885    ,   3443    ,   25481   ,   47399   ,0 };
            const unsigned long dim4229Kuo3Init[]   =   {   1   ,   3   ,   7   ,   1   ,   21  ,   49  ,   35  ,   179 ,   339 ,   505 ,   1121    ,   2393    ,   1181    ,   11389   ,   5851    ,   30553   ,0 };
            const unsigned long dim4230Kuo3Init[]   =   {   1   ,   3   ,   7   ,   15  ,   31  ,   3   ,   23  ,   205 ,   3   ,   861 ,   1767    ,   1971    ,   2279    ,   4901    ,   22077   ,   54183   ,0 };
            const unsigned long dim4231Kuo3Init[]   =   {   1   ,   3   ,   7   ,   11  ,   11  ,   1   ,   3   ,   165 ,   129 ,   767 ,   2007    ,   1913    ,   3701    ,   5761    ,   1985    ,   42383   ,0 };
            const unsigned long dim4232Kuo3Init[]   =   {   1   ,   3   ,   3   ,   5   ,   3   ,   39  ,   31  ,   73  ,   55  ,   805 ,   545 ,   1577    ,   1409    ,   5693    ,   31995   ,   58057   ,0 };
            const unsigned long dim4233Kuo3Init[]   =   {   1   ,   3   ,   1   ,   15  ,   21  ,   9   ,   69  ,   17  ,   497 ,   1003    ,   1123    ,   4023    ,   1921    ,   723 ,   19815   ,   11583   ,0 };
            const unsigned long dim4234Kuo3Init[]   =   {   1   ,   1   ,   7   ,   1   ,   13  ,   31  ,   61  ,   189 ,   1   ,   805 ,   1321    ,   1973    ,   4133    ,   7789    ,   9735    ,   50237   ,0 };
            const unsigned long dim4235Kuo3Init[]   =   {   1   ,   3   ,   7   ,   5   ,   1   ,   39  ,   67  ,   189 ,   139 ,   891 ,   171 ,   693 ,   5639    ,   12841   ,   12223   ,   64009   ,0 };
            const unsigned long dim4236Kuo3Init[]   =   {   1   ,   1   ,   3   ,   1   ,   7   ,   1   ,   115 ,   41  ,   283 ,   463 ,   53  ,   1349    ,   3233    ,   7247    ,   28515   ,   23931   ,0 };
            const unsigned long dim4237Kuo3Init[]   =   {   1   ,   3   ,   3   ,   1   ,   25  ,   31  ,   49  ,   177 ,   339 ,   103 ,   701 ,   3059    ,   6557    ,   8669    ,   13923   ,   46913   ,0 };
            const unsigned long dim4238Kuo3Init[]   =   {   1   ,   1   ,   5   ,   5   ,   7   ,   47  ,   7   ,   153 ,   481 ,   1005    ,   615 ,   2483    ,   3519    ,   5297    ,   21749   ,   55945   ,0 };
            const unsigned long dim4239Kuo3Init[]   =   {   1   ,   3   ,   5   ,   15  ,   15  ,   43  ,   105 ,   121 ,   419 ,   181 ,   1945    ,   3501    ,   3443    ,   5555    ,   27925   ,   18357   ,0 };
            const unsigned long dim4240Kuo3Init[]   =   {   1   ,   3   ,   3   ,   5   ,   5   ,   63  ,   107 ,   225 ,   17  ,   503 ,   2033    ,   2059    ,   2621    ,   12675   ,   25015   ,   15115   ,0 };
            const unsigned long dim4241Kuo3Init[]   =   {   1   ,   3   ,   5   ,   13  ,   5   ,   33  ,   39  ,   143 ,   293 ,   793 ,   159 ,   1349    ,   5483    ,   9303    ,   8269    ,   58221   ,0 };
            const unsigned long dim4242Kuo3Init[]   =   {   1   ,   1   ,   1   ,   3   ,   13  ,   21  ,   15  ,   11  ,   489 ,   921 ,   47  ,   7   ,   7677    ,   3281    ,   1339    ,   1447    ,0 };
            const unsigned long dim4243Kuo3Init[]   =   {   1   ,   1   ,   3   ,   11  ,   19  ,   61  ,   29  ,   45  ,   355 ,   671 ,   1029    ,   455 ,   5611    ,   14655   ,   22871   ,   17303   ,0 };
            const unsigned long dim4244Kuo3Init[]   =   {   1   ,   1   ,   1   ,   11  ,   3   ,   29  ,   93  ,   83  ,   425 ,   689 ,   437 ,   1943    ,   457 ,   2497    ,   2573    ,   56827   ,0 };
            const unsigned long dim4245Kuo3Init[]   =   {   1   ,   3   ,   7   ,   1   ,   3   ,   39  ,   17  ,   243 ,   475 ,   13  ,   199 ,   413 ,   5503    ,   1559    ,   8615    ,   5339    ,0 };
            const unsigned long dim4246Kuo3Init[]   =   {   1   ,   1   ,   5   ,   13  ,   7   ,   17  ,   55  ,   243 ,   465 ,   957 ,   1347    ,   3247    ,   4613    ,   10821   ,   17377   ,   21765   ,0 };
            const unsigned long dim4247Kuo3Init[]   =   {   1   ,   1   ,   3   ,   5   ,   15  ,   49  ,   109 ,   113 ,   19  ,   1007    ,   967 ,   685 ,   3195    ,   467 ,   1269    ,   27705   ,0 };
            const unsigned long dim4248Kuo3Init[]   =   {   1   ,   3   ,   3   ,   15  ,   31  ,   25  ,   73  ,   73  ,   483 ,   565 ,   1739    ,   641 ,   4085    ,   10481   ,   30053   ,   52773   ,0 };
            const unsigned long dim4249Kuo3Init[]   =   {   1   ,   1   ,   1   ,   1   ,   31  ,   17  ,   35  ,   85  ,   339 ,   93  ,   479 ,   2833    ,   681 ,   3123    ,   3753    ,   63989   ,0 };
            const unsigned long dim4250Kuo3Init[]   =   {   1   ,   3   ,   7   ,   5   ,   25  ,   17  ,   23  ,   47  ,   83  ,   59  ,   1525    ,   2445    ,   4831    ,   2605    ,   20711   ,   34129   ,0 };
            const unsigned long dim4251Kuo3Init[]   =   {   1   ,   3   ,   7   ,   9   ,   11  ,   29  ,   111 ,   239 ,   81  ,   855 ,   2039    ,   4051    ,   4619    ,   14095   ,   15765   ,   17195   ,0 };
            const unsigned long dim4252Kuo3Init[]   =   {   1   ,   1   ,   5   ,   9   ,   17  ,   29  ,   105 ,   175 ,   497 ,   741 ,   1151    ,   767 ,   7211    ,   14699   ,   9351    ,   22163   ,0 };
            const unsigned long dim4253Kuo3Init[]   =   {   1   ,   1   ,   3   ,   9   ,   23  ,   17  ,   23  ,   193 ,   25  ,   883 ,   347 ,   2429    ,   4023    ,   5477    ,   9365    ,   30625   ,0 };
            const unsigned long dim4254Kuo3Init[]   =   {   1   ,   1   ,   7   ,   13  ,   31  ,   57  ,   115 ,   119 ,   413 ,   733 ,   1095    ,   3877    ,   4021    ,   8415    ,   16129   ,   49425   ,0 };
            const unsigned long dim4255Kuo3Init[]   =   {   1   ,   3   ,   1   ,   3   ,   13  ,   39  ,   127 ,   97  ,   389 ,   993 ,   1941    ,   847 ,   5109    ,   8499    ,   18653   ,   20615   ,0 };
            const unsigned long dim4256Kuo3Init[]   =   {   1   ,   1   ,   3   ,   7   ,   9   ,   33  ,   21  ,   11  ,   357 ,   653 ,   1423    ,   3525    ,   277 ,   13085   ,   2525    ,   16547   ,0 };
            const unsigned long dim4257Kuo3Init[]   =   {   1   ,   3   ,   3   ,   1   ,   3   ,   63  ,   55  ,   137 ,   229 ,   359 ,   703 ,   2555    ,   3117    ,   9323    ,   3483    ,   52999   ,0 };
            const unsigned long dim4258Kuo3Init[]   =   {   1   ,   3   ,   7   ,   1   ,   23  ,   7   ,   67  ,   119 ,   239 ,   951 ,   495 ,   645 ,   3047    ,   12393   ,   14837   ,   33015   ,0 };
            const unsigned long dim4259Kuo3Init[]   =   {   1   ,   1   ,   5   ,   15  ,   23  ,   33  ,   19  ,   75  ,   365 ,   419 ,   905 ,   3485    ,   1097    ,   10639   ,   8627    ,   57441   ,0 };
            const unsigned long dim4260Kuo3Init[]   =   {   1   ,   1   ,   3   ,   3   ,   5   ,   37  ,   83  ,   163 ,   483 ,   403 ,   1467    ,   309 ,   5113    ,   3567    ,   28425   ,   63963   ,0 };
            const unsigned long dim4261Kuo3Init[]   =   {   1   ,   3   ,   3   ,   15  ,   21  ,   63  ,   59  ,   103 ,   261 ,   355 ,   1779    ,   1589    ,   7447    ,   10909   ,   23197   ,   31539   ,0 };
            const unsigned long dim4262Kuo3Init[]   =   {   1   ,   1   ,   3   ,   11  ,   9   ,   39  ,   121 ,   7   ,   383 ,   975 ,   179 ,   505 ,   1551    ,   15397   ,   28277   ,   53023   ,0 };
            const unsigned long dim4263Kuo3Init[]   =   {   1   ,   1   ,   5   ,   3   ,   31  ,   15  ,   29  ,   81  ,   291 ,   843 ,   507 ,   1291    ,   4533    ,   5333    ,   18841   ,   58999   ,0 };
            const unsigned long dim4264Kuo3Init[]   =   {   1   ,   3   ,   1   ,   3   ,   3   ,   15  ,   107 ,   93  ,   15  ,   349 ,   1359    ,   151 ,   4859    ,   10435   ,   3669    ,   31713   ,0 };
            const unsigned long dim4265Kuo3Init[]   =   {   1   ,   3   ,   5   ,   1   ,   21  ,   1   ,   45  ,   167 ,   25  ,   123 ,   905 ,   115 ,   2953    ,   15029   ,   31475   ,   37235   ,0 };
            const unsigned long dim4266Kuo3Init[]   =   {   1   ,   3   ,   5   ,   11  ,   9   ,   35  ,   115 ,   49  ,   61  ,   29  ,   1529    ,   1311    ,   6293    ,   13505   ,   15033   ,   31879   ,0 };
            const unsigned long dim4267Kuo3Init[]   =   {   1   ,   3   ,   1   ,   1   ,   1   ,   19  ,   21  ,   199 ,   489 ,   463 ,   1761    ,   2295    ,   2011    ,   14179   ,   22955   ,   22451   ,0 };
            const unsigned long dim4268Kuo3Init[]   =   {   1   ,   3   ,   1   ,   3   ,   21  ,   59  ,   71  ,   27  ,   425 ,   459 ,   177 ,   2301    ,   4703    ,   5589    ,   12799   ,   44729   ,0 };
            const unsigned long dim4269Kuo3Init[]   =   {   1   ,   3   ,   7   ,   15  ,   11  ,   43  ,   103 ,   207 ,   315 ,   805 ,   469 ,   4009    ,   4589    ,   9627    ,   13309   ,   929 ,0 };
            const unsigned long dim4270Kuo3Init[]   =   {   1   ,   3   ,   5   ,   15  ,   3   ,   45  ,   49  ,   59  ,   25  ,   371 ,   859 ,   2687    ,   3767    ,   7395    ,   24579   ,   2657    ,0 };
            const unsigned long dim4271Kuo3Init[]   =   {   1   ,   3   ,   3   ,   1   ,   13  ,   1   ,   39  ,   115 ,   497 ,   291 ,   1187    ,   3109    ,   6731    ,   7591    ,   2551    ,   49121   ,0 };
            const unsigned long dim4272Kuo3Init[]   =   {   1   ,   3   ,   1   ,   1   ,   11  ,   37  ,   75  ,   39  ,   377 ,   67  ,   291 ,   1275    ,   7893    ,   16261   ,   27093   ,   30007   ,0 };
            const unsigned long dim4273Kuo3Init[]   =   {   1   ,   1   ,   5   ,   15  ,   27  ,   31  ,   117 ,   145 ,   435 ,   739 ,   1855    ,   2703    ,   6125    ,   16333   ,   16209   ,   11649   ,0 };
            const unsigned long dim4274Kuo3Init[]   =   {   1   ,   3   ,   5   ,   7   ,   5   ,   37  ,   3   ,   173 ,   461 ,   535 ,   295 ,   3097    ,   6275    ,   15703   ,   20079   ,   45413   ,0 };
            const unsigned long dim4275Kuo3Init[]   =   {   1   ,   1   ,   5   ,   13  ,   9   ,   39  ,   39  ,   113 ,   113 ,   315 ,   697 ,   2433    ,   7637    ,   9895    ,   17979   ,   13837   ,0 };
            const unsigned long dim4276Kuo3Init[]   =   {   1   ,   1   ,   7   ,   11  ,   7   ,   35  ,   113 ,   255 ,   309 ,   389 ,   757 ,   1649    ,   1871    ,   311 ,   3531    ,   15617   ,0 };
            const unsigned long dim4277Kuo3Init[]   =   {   1   ,   1   ,   7   ,   5   ,   25  ,   7   ,   87  ,   99  ,   249 ,   687 ,   57  ,   2291    ,   3315    ,   14677   ,   15225   ,   16859   ,0 };
            const unsigned long dim4278Kuo3Init[]   =   {   1   ,   1   ,   7   ,   1   ,   11  ,   23  ,   91  ,   25  ,   209 ,   227 ,   567 ,   183 ,   7419    ,   9873    ,   10653   ,   41705   ,0 };
            const unsigned long dim4279Kuo3Init[]   =   {   1   ,   1   ,   1   ,   13  ,   19  ,   17  ,   47  ,   79  ,   187 ,   185 ,   187 ,   3959    ,   4221    ,   389 ,   1733    ,   62413   ,0 };
            const unsigned long dim4280Kuo3Init[]   =   {   1   ,   3   ,   7   ,   9   ,   1   ,   27  ,   109 ,   59  ,   223 ,   451 ,   865 ,   3623    ,   2545    ,   6531    ,   14837   ,   40749   ,0 };
            const unsigned long dim4281Kuo3Init[]   =   {   1   ,   1   ,   3   ,   7   ,   5   ,   9   ,   69  ,   201 ,   145 ,   445 ,   1787    ,   1995    ,   691 ,   11387   ,   21391   ,   19185   ,0 };
            const unsigned long dim4282Kuo3Init[]   =   {   1   ,   3   ,   7   ,   3   ,   15  ,   31  ,   61  ,   121 ,   295 ,   93  ,   1157    ,   2821    ,   2907    ,   6995    ,   17563   ,   4613    ,0 };
            const unsigned long dim4283Kuo3Init[]   =   {   1   ,   3   ,   5   ,   3   ,   3   ,   41  ,   43  ,   199 ,   319 ,   903 ,   1699    ,   3599    ,   6567    ,   12851   ,   21253   ,   49715   ,0 };
            const unsigned long dim4284Kuo3Init[]   =   {   1   ,   1   ,   5   ,   1   ,   25  ,   59  ,   7   ,   107 ,   405 ,   365 ,   1159    ,   2357    ,   1645    ,   16353   ,   31521   ,   13587   ,0 };
            const unsigned long dim4285Kuo3Init[]   =   {   1   ,   3   ,   3   ,   1   ,   29  ,   25  ,   109 ,   181 ,   475 ,   117 ,   809 ,   2741    ,   5113    ,   9973    ,   31763   ,   20201   ,0 };
            const unsigned long dim4286Kuo3Init[]   =   {   1   ,   3   ,   3   ,   9   ,   29  ,   43  ,   51  ,   239 ,   359 ,   137 ,   419 ,   43  ,   4091    ,   11107   ,   2121    ,   5057    ,0 };
            const unsigned long dim4287Kuo3Init[]   =   {   1   ,   3   ,   1   ,   7   ,   13  ,   17  ,   65  ,   151 ,   289 ,   37  ,   1321    ,   3853    ,   7513    ,   3161    ,   6517    ,   27777   ,0 };
            const unsigned long dim4288Kuo3Init[]   =   {   1   ,   1   ,   3   ,   13  ,   13  ,   31  ,   125 ,   13  ,   507 ,   491 ,   669 ,   2913    ,   8067    ,   4257    ,   11413   ,   8003    ,0 };
            const unsigned long dim4289Kuo3Init[]   =   {   1   ,   1   ,   7   ,   1   ,   23  ,   57  ,   125 ,   227 ,   341 ,   647 ,   907 ,   1503    ,   1541    ,   14311   ,   10591   ,   49967   ,0 };
            const unsigned long dim4290Kuo3Init[]   =   {   1   ,   3   ,   1   ,   15  ,   25  ,   29  ,   83  ,   39  ,   141 ,   545 ,   1213    ,   2001    ,   6627    ,   8701    ,   19205   ,   2859    ,0 };
            const unsigned long dim4291Kuo3Init[]   =   {   1   ,   1   ,   7   ,   13  ,   11  ,   25  ,   93  ,   91  ,   471 ,   907 ,   793 ,   3045    ,   8137    ,   5095    ,   12445   ,   41967   ,0 };
            const unsigned long dim4292Kuo3Init[]   =   {   1   ,   1   ,   5   ,   1   ,   25  ,   45  ,   13  ,   191 ,   373 ,   371 ,   1321    ,   3381    ,   3945    ,   9097    ,   7785    ,   44469   ,0 };
            const unsigned long dim4293Kuo3Init[]   =   {   1   ,   3   ,   3   ,   13  ,   13  ,   19  ,   53  ,   189 ,   183 ,   115 ,   1957    ,   91  ,   113 ,   5839    ,   23783   ,   63959   ,0 };
            const unsigned long dim4294Kuo3Init[]   =   {   1   ,   3   ,   1   ,   7   ,   3   ,   19  ,   71  ,   219 ,   393 ,   369 ,   107 ,   3729    ,   7457    ,   12379   ,   28647   ,   54159   ,0 };
            const unsigned long dim4295Kuo3Init[]   =   {   1   ,   1   ,   1   ,   9   ,   5   ,   43  ,   15  ,   125 ,   297 ,   145 ,   1837    ,   2143    ,   305 ,   793 ,   14663   ,   313 ,0 };
            const unsigned long dim4296Kuo3Init[]   =   {   1   ,   3   ,   1   ,   5   ,   17  ,   57  ,   67  ,   51  ,   211 ,   817 ,   2023    ,   1615    ,   4379    ,   12165   ,   30433   ,   44479   ,0 };
            const unsigned long dim4297Kuo3Init[]   =   {   1   ,   1   ,   7   ,   1   ,   9   ,   49  ,   71  ,   231 ,   159 ,   179 ,   2027    ,   4043    ,   1053    ,   3679    ,   9035    ,   4173    ,0 };
            const unsigned long dim4298Kuo3Init[]   =   {   1   ,   3   ,   7   ,   3   ,   17  ,   5   ,   33  ,   151 ,   223 ,   331 ,   1777    ,   2467    ,   7949    ,   2571    ,   9099    ,   58751   ,0 };
            const unsigned long dim4299Kuo3Init[]   =   {   1   ,   3   ,   1   ,   15  ,   5   ,   23  ,   85  ,   113 ,   369 ,   607 ,   505 ,   1791    ,   1597    ,   12933   ,   28853   ,   13207   ,0 };
            const unsigned long dim4300Kuo3Init[]   =   {   1   ,   1   ,   5   ,   15  ,   11  ,   39  ,   107 ,   81  ,   345 ,   815 ,   2025    ,   803 ,   1779    ,   15499   ,   26197   ,   6045    ,0 };
            const unsigned long dim4301Kuo3Init[]   =   {   1   ,   3   ,   1   ,   9   ,   1   ,   3   ,   41  ,   29  ,   213 ,   295 ,   1561    ,   2631    ,   6231    ,   3255    ,   11309   ,   22853   ,0 };
            const unsigned long dim4302Kuo3Init[]   =   {   1   ,   3   ,   7   ,   9   ,   19  ,   21  ,   127 ,   57  ,   425 ,   551 ,   171 ,   903 ,   7747    ,   5251    ,   11457   ,   57735   ,0 };
            const unsigned long dim4303Kuo3Init[]   =   {   1   ,   1   ,   1   ,   11  ,   29  ,   5   ,   33  ,   221 ,   91  ,   435 ,   1341    ,   1627    ,   4269    ,   8689    ,   19  ,   9609    ,0 };
            const unsigned long dim4304Kuo3Init[]   =   {   1   ,   3   ,   1   ,   9   ,   11  ,   1   ,   101 ,   219 ,   45  ,   317 ,   793 ,   823 ,   2449    ,   7911    ,   3739    ,   2441    ,0 };
            const unsigned long dim4305Kuo3Init[]   =   {   1   ,   3   ,   5   ,   7   ,   25  ,   51  ,   103 ,   179 ,   481 ,   999 ,   675 ,   1693    ,   193 ,   9987    ,   18519   ,   16777   ,0 };
            const unsigned long dim4306Kuo3Init[]   =   {   1   ,   3   ,   5   ,   5   ,   15  ,   59  ,   83  ,   61  ,   329 ,   537 ,   1293    ,   459 ,   5059    ,   7943    ,   3569    ,   34727   ,0 };
            const unsigned long dim4307Kuo3Init[]   =   {   1   ,   3   ,   3   ,   9   ,   29  ,   17  ,   41  ,   171 ,   303 ,   123 ,   1945    ,   1383    ,   5823    ,   1569    ,   10699   ,   14585   ,0 };
            const unsigned long dim4308Kuo3Init[]   =   {   1   ,   1   ,   1   ,   11  ,   7   ,   11  ,   35  ,   193 ,   69  ,   487 ,   879 ,   693 ,   6367    ,   7381    ,   11463   ,   50741   ,0 };
            const unsigned long dim4309Kuo3Init[]   =   {   1   ,   3   ,   1   ,   11  ,   9   ,   17  ,   5   ,   155 ,   443 ,   543 ,   643 ,   1595    ,   319 ,   2877    ,   27849   ,   8365    ,0 };
            const unsigned long dim4310Kuo3Init[]   =   {   1   ,   3   ,   7   ,   15  ,   3   ,   43  ,   111 ,   219 ,   35  ,   595 ,   541 ,   3299    ,   4117    ,   265 ,   2817    ,   60825   ,0 };
            const unsigned long dim4311Kuo3Init[]   =   {   1   ,   3   ,   5   ,   5   ,   13  ,   15  ,   45  ,   211 ,   21  ,   1019    ,   1015    ,   1051    ,   2879    ,   15077   ,   19851   ,   6615    ,0 };
            const unsigned long dim4312Kuo3Init[]   =   {   1   ,   1   ,   7   ,   13  ,   15  ,   59  ,   7   ,   201 ,   215 ,   217 ,   967 ,   2013    ,   907 ,   12073   ,   9105    ,   44873   ,0 };
            const unsigned long dim4313Kuo3Init[]   =   {   1   ,   1   ,   1   ,   13  ,   1   ,   3   ,   117 ,   253 ,   493 ,   727 ,   1997    ,   3325    ,   2757    ,   11313   ,   7439    ,   18863   ,0 };
            const unsigned long dim4314Kuo3Init[]   =   {   1   ,   1   ,   7   ,   11  ,   5   ,   23  ,   97  ,   169 ,   63  ,   243 ,   419 ,   3259    ,   2521    ,   13055   ,   3223    ,   36055   ,0 };
            const unsigned long dim4315Kuo3Init[]   =   {   1   ,   1   ,   5   ,   9   ,   9   ,   9   ,   63  ,   69  ,   305 ,   391 ,   1901    ,   2785    ,   7877    ,   7219    ,   10727   ,   13115   ,0 };
            const unsigned long dim4316Kuo3Init[]   =   {   1   ,   1   ,   3   ,   7   ,   1   ,   63  ,   83  ,   27  ,   417 ,   537 ,   1969    ,   3083    ,   4267    ,   11153   ,   11205   ,   61249   ,0 };
            const unsigned long dim4317Kuo3Init[]   =   {   1   ,   3   ,   3   ,   1   ,   11  ,   47  ,   79  ,   201 ,   449 ,   569 ,   809 ,   1575    ,   271 ,   651 ,   5459    ,   24805   ,0 };
            const unsigned long dim4318Kuo3Init[]   =   {   1   ,   3   ,   7   ,   15  ,   29  ,   25  ,   103 ,   95  ,   57  ,   209 ,   1175    ,   2579    ,   3735    ,   1969    ,   29903   ,   65397   ,0 };
            const unsigned long dim4319Kuo3Init[]   =   {   1   ,   3   ,   5   ,   1   ,   25  ,   41  ,   7   ,   239 ,   175 ,   395 ,   1261    ,   2567    ,   4875    ,   279 ,   25905   ,   43603   ,0 };
            const unsigned long dim4320Kuo3Init[]   =   {   1   ,   3   ,   7   ,   1   ,   27  ,   7   ,   89  ,   111 ,   467 ,   691 ,   39  ,   1213    ,   5377    ,   13247   ,   31513   ,   58015   ,0 };
            const unsigned long dim4321Kuo3Init[]   =   {   1   ,   1   ,   1   ,   15  ,   25  ,   1   ,   55  ,   131 ,   259 ,   923 ,   1623    ,   125 ,   2273    ,   4221    ,   21857   ,   49621   ,0 };
            const unsigned long dim4322Kuo3Init[]   =   {   1   ,   1   ,   5   ,   3   ,   27  ,   1   ,   99  ,   95  ,   199 ,   773 ,   87  ,   2981    ,   8127    ,   6493    ,   27507   ,   52789   ,0 };
            const unsigned long dim4323Kuo3Init[]   =   {   1   ,   1   ,   5   ,   9   ,   11  ,   39  ,   7   ,   7   ,   145 ,   933 ,   1301    ,   2975    ,   5237    ,   11619   ,   1423    ,   46351   ,0 };
            const unsigned long dim4324Kuo3Init[]   =   {   1   ,   1   ,   7   ,   5   ,   17  ,   3   ,   29  ,   233 ,   195 ,   727 ,   1893    ,   3709    ,   1285    ,   12877   ,   30821   ,   30383   ,0 };
            const unsigned long dim4325Kuo3Init[]   =   {   1   ,   3   ,   7   ,   15  ,   1   ,   7   ,   115 ,   55  ,   231 ,   769 ,   849 ,   1043    ,   3041    ,   5187    ,   14613   ,   58821   ,0 };
            const unsigned long dim4326Kuo3Init[]   =   {   1   ,   3   ,   7   ,   3   ,   13  ,   41  ,   5   ,   245 ,   109 ,   785 ,   187 ,   391 ,   3333    ,   8193    ,   823 ,   27985   ,0 };
            const unsigned long dim4327Kuo3Init[]   =   {   1   ,   1   ,   5   ,   11  ,   9   ,   33  ,   25  ,   231 ,   347 ,   313 ,   31  ,   59  ,   6189    ,   4249    ,   26963   ,   56975   ,0 };
            const unsigned long dim4328Kuo3Init[]   =   {   1   ,   1   ,   5   ,   5   ,   27  ,   57  ,   87  ,   77  ,   209 ,   231 ,   173 ,   3691    ,   8095    ,   10353   ,   14217   ,   3263    ,0 };
            const unsigned long dim4329Kuo3Init[]   =   {   1   ,   3   ,   1   ,   7   ,   1   ,   55  ,   115 ,   63  ,   13  ,   705 ,   865 ,   1901    ,   4209    ,   8195    ,   25245   ,   2313    ,0 };
            const unsigned long dim4330Kuo3Init[]   =   {   1   ,   3   ,   5   ,   1   ,   13  ,   3   ,   53  ,   215 ,   345 ,   635 ,   1601    ,   241 ,   6491    ,   911 ,   5825    ,   34247   ,0 };
            const unsigned long dim4331Kuo3Init[]   =   {   1   ,   1   ,   5   ,   5   ,   27  ,   21  ,   69  ,   83  ,   451 ,   33  ,   773 ,   2663    ,   5325    ,   16207   ,   15073   ,   32259   ,0 };
            const unsigned long dim4332Kuo3Init[]   =   {   1   ,   1   ,   3   ,   9   ,   25  ,   33  ,   1   ,   35  ,   335 ,   355 ,   2005    ,   3087    ,   4577    ,   10533   ,   23641   ,   45297   ,0 };
            const unsigned long dim4333Kuo3Init[]   =   {   1   ,   3   ,   3   ,   7   ,   21  ,   43  ,   103 ,   223 ,   269 ,   781 ,   1445    ,   461 ,   6777    ,   5251    ,   6483    ,   14425   ,0 };
            const unsigned long dim4334Kuo3Init[]   =   {   1   ,   3   ,   7   ,   7   ,   21  ,   55  ,   113 ,   227 ,   227 ,   173 ,   575 ,   3187    ,   3965    ,   13569   ,   4809    ,   16217   ,0 };
            const unsigned long dim4335Kuo3Init[]   =   {   1   ,   1   ,   5   ,   5   ,   23  ,   23  ,   67  ,   99  ,   257 ,   651 ,   495 ,   3427    ,   7653    ,   505 ,   2549    ,   9283    ,0 };
            const unsigned long dim4336Kuo3Init[]   =   {   1   ,   1   ,   5   ,   5   ,   15  ,   39  ,   123 ,   27  ,   1   ,   5   ,   107 ,   4055    ,   7347    ,   6411    ,   28857   ,   45591   ,0 };
            const unsigned long dim4337Kuo3Init[]   =   {   1   ,   3   ,   1   ,   5   ,   29  ,   57  ,   77  ,   67  ,   365 ,   25  ,   1475    ,   3583    ,   4315    ,   6725    ,   22635   ,   62003   ,0 };
            const unsigned long dim4338Kuo3Init[]   =   {   1   ,   3   ,   1   ,   1   ,   11  ,   9   ,   19  ,   25  ,   399 ,   467 ,   501 ,   2081    ,   977 ,   7695    ,   12997   ,   11841   ,0 };
            const unsigned long dim4339Kuo3Init[]   =   {   1   ,   1   ,   5   ,   15  ,   19  ,   37  ,   39  ,   69  ,   75  ,   343 ,   525 ,   433 ,   6771    ,   8697    ,   6041    ,   49701   ,0 };
            const unsigned long dim4340Kuo3Init[]   =   {   1   ,   1   ,   3   ,   1   ,   21  ,   31  ,   5   ,   193 ,   467 ,   161 ,   929 ,   3967    ,   2629    ,   3031    ,   30715   ,   45539   ,0 };
            const unsigned long dim4341Kuo3Init[]   =   {   1   ,   3   ,   3   ,   9   ,   25  ,   17  ,   87  ,   179 ,   385 ,   491 ,   965 ,   2809    ,   6937    ,   2669    ,   20963   ,   13333   ,0 };
            const unsigned long dim4342Kuo3Init[]   =   {   1   ,   3   ,   1   ,   7   ,   25  ,   1   ,   83  ,   47  ,   99  ,   425 ,   1007    ,   2903    ,   2959    ,   12403   ,   19707   ,   60863   ,0 };
            const unsigned long dim4343Kuo3Init[]   =   {   1   ,   1   ,   1   ,   15  ,   5   ,   19  ,   3   ,   181 ,   255 ,   135 ,   791 ,   3473    ,   3123    ,   5247    ,   28349   ,   5581    ,0 };
            const unsigned long dim4344Kuo3Init[]   =   {   1   ,   1   ,   5   ,   13  ,   27  ,   13  ,   93  ,   215 ,   333 ,   667 ,   251 ,   2027    ,   6345    ,   8985    ,   22669   ,   46509   ,0 };
            const unsigned long dim4345Kuo3Init[]   =   {   1   ,   1   ,   5   ,   9   ,   1   ,   27  ,   83  ,   107 ,   339 ,   699 ,   1197    ,   635 ,   2367    ,   6791    ,   9743    ,   9685    ,0 };
            const unsigned long dim4346Kuo3Init[]   =   {   1   ,   3   ,   1   ,   7   ,   11  ,   43  ,   9   ,   37  ,   511 ,   177 ,   1957    ,   2697    ,   1879    ,   8219    ,   27525   ,   17737   ,0 };
            const unsigned long dim4347Kuo3Init[]   =   {   1   ,   3   ,   1   ,   5   ,   3   ,   35  ,   81  ,   137 ,   193 ,   439 ,   35  ,   3743    ,   2973    ,   10245   ,   17111   ,   12025   ,0 };
            const unsigned long dim4348Kuo3Init[]   =   {   1   ,   3   ,   3   ,   9   ,   3   ,   51  ,   7   ,   103 ,   65  ,   787 ,   1859    ,   1217    ,   8189    ,   2917    ,   2161    ,   64559   ,0 };
            const unsigned long dim4349Kuo3Init[]   =   {   1   ,   3   ,   5   ,   5   ,   31  ,   9   ,   53  ,   45  ,   167 ,   877 ,   1169    ,   2375    ,   6631    ,   9699    ,   8345    ,   27223   ,0 };
            const unsigned long dim4350Kuo3Init[]   =   {   1   ,   3   ,   5   ,   9   ,   1   ,   33  ,   113 ,   205 ,   153 ,   255 ,   63  ,   1517    ,   5225    ,   14485   ,   22053   ,   53333   ,0 };
            const unsigned long dim4351Kuo3Init[]   =   {   1   ,   1   ,   1   ,   1   ,   7   ,   45  ,   95  ,   211 ,   13  ,   769 ,   1909    ,   1049    ,   2465    ,   11581   ,   2983    ,   55267   ,0 };
            const unsigned long dim4352Kuo3Init[]   =   {   1   ,   1   ,   5   ,   3   ,   7   ,   27  ,   3   ,   47  ,   503 ,   7   ,   1223    ,   1979    ,   815 ,   12593   ,   4229    ,   30539   ,0 };
            const unsigned long dim4353Kuo3Init[]   =   {   1   ,   3   ,   7   ,   15  ,   9   ,   3   ,   103 ,   217 ,   199 ,   567 ,   943 ,   1527    ,   7191    ,   15243   ,   26311   ,   44263   ,0 };
            const unsigned long dim4354Kuo3Init[]   =   {   1   ,   3   ,   5   ,   11  ,   29  ,   13  ,   107 ,   107 ,   35  ,   367 ,   849 ,   4047    ,   6027    ,   8155    ,   10963   ,   9941    ,0 };
            const unsigned long dim4355Kuo3Init[]   =   {   1   ,   1   ,   7   ,   1   ,   27  ,   49  ,   89  ,   59  ,   479 ,   477 ,   1351    ,   2165    ,   7611    ,   9745    ,   7023    ,   22153   ,0 };
            const unsigned long dim4356Kuo3Init[]   =   {   1   ,   1   ,   3   ,   1   ,   5   ,   7   ,   21  ,   91  ,   13  ,   547 ,   493 ,   813 ,   4325    ,   751 ,   14753   ,   52131   ,0 };
            const unsigned long dim4357Kuo3Init[]   =   {   1   ,   3   ,   1   ,   3   ,   23  ,   7   ,   55  ,   69  ,   335 ,   127 ,   1669    ,   2171    ,   6033    ,   3713    ,   16855   ,   36479   ,0 };
            const unsigned long dim4358Kuo3Init[]   =   {   1   ,   3   ,   3   ,   9   ,   17  ,   35  ,   117 ,   85  ,   415 ,   1019    ,   15  ,   3033    ,   2933    ,   2113    ,   10909   ,   107 ,0 };
            const unsigned long dim4359Kuo3Init[]   =   {   1   ,   1   ,   7   ,   3   ,   13  ,   45  ,   61  ,   31  ,   115 ,   569 ,   21  ,   375 ,   1955    ,   10699   ,   17307   ,   20119   ,0 };
            const unsigned long dim4360Kuo3Init[]   =   {   1   ,   1   ,   1   ,   13  ,   29  ,   29  ,   101 ,   137 ,   457 ,   845 ,   39  ,   461 ,   5113    ,   1429    ,   25443   ,   60127   ,0 };
            const unsigned long dim4361Kuo3Init[]   =   {   1   ,   1   ,   1   ,   3   ,   29  ,   39  ,   35  ,   45  ,   269 ,   327 ,   2023    ,   2799    ,   4893    ,   3751    ,   7579    ,   56975   ,0 };
            const unsigned long dim4362Kuo3Init[]   =   {   1   ,   1   ,   3   ,   3   ,   3   ,   49  ,   45  ,   219 ,   133 ,   279 ,   2003    ,   4007    ,   409 ,   3697    ,   11629   ,   19359   ,0 };
            const unsigned long dim4363Kuo3Init[]   =   {   1   ,   3   ,   5   ,   9   ,   25  ,   21  ,   85  ,   61  ,   131 ,   69  ,   1277    ,   249 ,   6265    ,   10245   ,   32575   ,   59735   ,0 };
            const unsigned long dim4364Kuo3Init[]   =   {   1   ,   3   ,   5   ,   7   ,   1   ,   11  ,   63  ,   125 ,   57  ,   17  ,   111 ,   3521    ,   7397    ,   10985   ,   10193   ,   45493   ,0 };
            const unsigned long dim4365Kuo3Init[]   =   {   1   ,   3   ,   5   ,   13  ,   31  ,   37  ,   115 ,   139 ,   203 ,   199 ,   559 ,   67  ,   7187    ,   5825    ,   6755    ,   43  ,0 };
            const unsigned long dim4366Kuo3Init[]   =   {   1   ,   3   ,   3   ,   15  ,   7   ,   33  ,   41  ,   59  ,   487 ,   711 ,   247 ,   1095    ,   4113    ,   11349   ,   26331   ,   33159   ,0 };
            const unsigned long dim4367Kuo3Init[]   =   {   1   ,   3   ,   7   ,   15  ,   25  ,   61  ,   57  ,   129 ,   409 ,   589 ,   1913    ,   1293    ,   7981    ,   8027    ,   789 ,   19655   ,0 };
            const unsigned long dim4368Kuo3Init[]   =   {   1   ,   1   ,   5   ,   11  ,   29  ,   33  ,   55  ,   165 ,   315 ,   757 ,   1213    ,   2071    ,   4827    ,   6905    ,   7087    ,   22317   ,0 };
            const unsigned long dim4369Kuo3Init[]   =   {   1   ,   3   ,   3   ,   1   ,   5   ,   7   ,   115 ,   161 ,   247 ,   91  ,   397 ,   1889    ,   2445    ,   13407   ,   20077   ,   50953   ,0 };
            const unsigned long dim4370Kuo3Init[]   =   {   1   ,   1   ,   5   ,   1   ,   31  ,   43  ,   53  ,   67  ,   183 ,   85  ,   285 ,   475 ,   3345    ,   7783    ,   25383   ,   26213   ,0 };
            const unsigned long dim4371Kuo3Init[]   =   {   1   ,   1   ,   1   ,   3   ,   21  ,   27  ,   127 ,   229 ,   135 ,   795 ,   287 ,   937 ,   5645    ,   2271    ,   26095   ,   18407   ,0 };
            const unsigned long dim4372Kuo3Init[]   =   {   1   ,   3   ,   5   ,   13  ,   27  ,   39  ,   109 ,   183 ,   1   ,   343 ,   695 ,   3825    ,   4347    ,   8399    ,   30219   ,   41839   ,0 };
            const unsigned long dim4373Kuo3Init[]   =   {   1   ,   1   ,   1   ,   3   ,   31  ,   3   ,   93  ,   139 ,   145 ,   107 ,   1311    ,   1919    ,   6631    ,   16363   ,   17473   ,   38661   ,0 };
            const unsigned long dim4374Kuo3Init[]   =   {   1   ,   3   ,   3   ,   1   ,   23  ,   7   ,   17  ,   131 ,   505 ,   335 ,   19  ,   1037    ,   6985    ,   15181   ,   24857   ,   37363   ,0 };
            const unsigned long dim4375Kuo3Init[]   =   {   1   ,   3   ,   7   ,   13  ,   5   ,   17  ,   51  ,   243 ,   221 ,   149 ,   801 ,   3179    ,   1929    ,   14145   ,   16541   ,   36911   ,0 };
            const unsigned long dim4376Kuo3Init[]   =   {   1   ,   3   ,   1   ,   7   ,   15  ,   5   ,   19  ,   93  ,   117 ,   21  ,   1517    ,   2041    ,   5281    ,   7579    ,   23087   ,   37619   ,0 };
            const unsigned long dim4377Kuo3Init[]   =   {   1   ,   3   ,   5   ,   1   ,   21  ,   33  ,   5   ,   187 ,   471 ,   483 ,   249 ,   1075    ,   4831    ,   5467    ,   2921    ,   60803   ,0 };
            const unsigned long dim4378Kuo3Init[]   =   {   1   ,   1   ,   1   ,   1   ,   27  ,   57  ,   23  ,   185 ,   397 ,   945 ,   495 ,   2097    ,   7561    ,   1111    ,   30773   ,   61855   ,0 };
            const unsigned long dim4379Kuo3Init[]   =   {   1   ,   1   ,   5   ,   5   ,   31  ,   53  ,   71  ,   125 ,   365 ,   947 ,   1685    ,   2871    ,   313 ,   5199    ,   11179   ,   18717   ,0 };
            const unsigned long dim4380Kuo3Init[]   =   {   1   ,   3   ,   7   ,   13  ,   23  ,   41  ,   83  ,   79  ,   427 ,   877 ,   281 ,   3205    ,   865 ,   8021    ,   3423    ,   28111   ,0 };
            const unsigned long dim4381Kuo3Init[]   =   {   1   ,   1   ,   3   ,   1   ,   29  ,   39  ,   11  ,   247 ,   45  ,   339 ,   1153    ,   1179    ,   6495    ,   6861    ,   15247   ,   44515   ,0 };
            const unsigned long dim4382Kuo3Init[]   =   {   1   ,   3   ,   1   ,   1   ,   1   ,   21  ,   41  ,   211 ,   379 ,   99  ,   887 ,   2035    ,   847 ,   4679    ,   9941    ,   49585   ,0 };
            const unsigned long dim4383Kuo3Init[]   =   {   1   ,   3   ,   1   ,   11  ,   17  ,   59  ,   29  ,   23  ,   355 ,   393 ,   323 ,   1843    ,   1373    ,   1995    ,   28417   ,   12167   ,0 };
            const unsigned long dim4384Kuo3Init[]   =   {   1   ,   3   ,   7   ,   1   ,   31  ,   55  ,   105 ,   239 ,   509 ,   709 ,   415 ,   4015    ,   4597    ,   13861   ,   11553   ,   87  ,0 };
            const unsigned long dim4385Kuo3Init[]   =   {   1   ,   3   ,   7   ,   15  ,   25  ,   55  ,   13  ,   19  ,   333 ,   371 ,   1829    ,   3879    ,   7199    ,   2849    ,   16885   ,   52575   ,0 };
            const unsigned long dim4386Kuo3Init[]   =   {   1   ,   3   ,   5   ,   9   ,   27  ,   23  ,   55  ,   109 ,   465 ,   923 ,   905 ,   197 ,   4017    ,   2079    ,   10775   ,   55543   ,0 };
            const unsigned long dim4387Kuo3Init[]   =   {   1   ,   1   ,   3   ,   7   ,   31  ,   49  ,   69  ,   181 ,   261 ,   71  ,   1727    ,   2717    ,   317 ,   13397   ,   26641   ,   37783   ,0 };
            const unsigned long dim4388Kuo3Init[]   =   {   1   ,   3   ,   3   ,   15  ,   17  ,   55  ,   107 ,   201 ,   21  ,   279 ,   437 ,   3267    ,   999 ,   14787   ,   23727   ,   2563    ,0 };
            const unsigned long dim4389Kuo3Init[]   =   {   1   ,   3   ,   1   ,   13  ,   27  ,   41  ,   19  ,   153 ,   435 ,   577 ,   327 ,   965 ,   7899    ,   11799   ,   4683    ,   49651   ,0 };
            const unsigned long dim4390Kuo3Init[]   =   {   1   ,   3   ,   5   ,   15  ,   25  ,   47  ,   13  ,   243 ,   281 ,   201 ,   1455    ,   2369    ,   7947    ,   10993   ,   14715   ,   30099   ,0 };
            const unsigned long dim4391Kuo3Init[]   =   {   1   ,   1   ,   3   ,   11  ,   17  ,   1   ,   49  ,   205 ,   303 ,   85  ,   823 ,   3721    ,   3155    ,   5473    ,   6019    ,   24837   ,0 };
            const unsigned long dim4392Kuo3Init[]   =   {   1   ,   3   ,   1   ,   7   ,   5   ,   27  ,   65  ,   177 ,   405 ,   841 ,   1523    ,   43  ,   2309    ,   14041   ,   6705    ,   18145   ,0 };
            const unsigned long dim4393Kuo3Init[]   =   {   1   ,   1   ,   7   ,   3   ,   25  ,   35  ,   53  ,   177 ,   233 ,   287 ,   1265    ,   3305    ,   4961    ,   2751    ,   8715    ,   56605   ,0 };
            const unsigned long dim4394Kuo3Init[]   =   {   1   ,   3   ,   5   ,   3   ,   7   ,   9   ,   45  ,   179 ,   65  ,   153 ,   1861    ,   111 ,   5501    ,   3671    ,   25663   ,   46493   ,0 };
            const unsigned long dim4395Kuo3Init[]   =   {   1   ,   3   ,   5   ,   11  ,   11  ,   37  ,   25  ,   199 ,   339 ,   313 ,   1823    ,   2933    ,   6459    ,   3359    ,   8577    ,   64109   ,0 };
            const unsigned long dim4396Kuo3Init[]   =   {   1   ,   3   ,   1   ,   11  ,   5   ,   3   ,   127 ,   53  ,   77  ,   431 ,   931 ,   3159    ,   2267    ,   8123    ,   7581    ,   52237   ,0 };
            const unsigned long dim4397Kuo3Init[]   =   {   1   ,   1   ,   5   ,   3   ,   1   ,   51  ,   41  ,   147 ,   481 ,   243 ,   2045    ,   3239    ,   8191    ,   7533    ,   25021   ,   46359   ,0 };
            const unsigned long dim4398Kuo3Init[]   =   {   1   ,   1   ,   1   ,   7   ,   17  ,   1   ,   57  ,   141 ,   239 ,   375 ,   23  ,   3695    ,   827 ,   10783   ,   18109   ,   54029   ,0 };
            const unsigned long dim4399Kuo3Init[]   =   {   1   ,   3   ,   1   ,   1   ,   13  ,   39  ,   35  ,   251 ,   271 ,   461 ,   313 ,   2807    ,   2927    ,   4639    ,   17175   ,   65325   ,0 };
            const unsigned long dim4400Kuo3Init[]   =   {   1   ,   1   ,   7   ,   15  ,   1   ,   9   ,   123 ,   243 ,   451 ,   433 ,   1339    ,   1223    ,   3399    ,   12669   ,   20663   ,   27505   ,0 };
            const unsigned long dim4401Kuo3Init[]   =   {   1   ,   1   ,   1   ,   5   ,   31  ,   25  ,   41  ,   225 ,   61  ,   525 ,   527 ,   1559    ,   5469    ,   2299    ,   3169    ,   12435   ,0 };
            const unsigned long dim4402Kuo3Init[]   =   {   1   ,   1   ,   7   ,   1   ,   3   ,   45  ,   21  ,   99  ,   89  ,   97  ,   379 ,   1815    ,   5883    ,   9437    ,   32597   ,   14795   ,0 };
            const unsigned long dim4403Kuo3Init[]   =   {   1   ,   3   ,   3   ,   3   ,   5   ,   17  ,   109 ,   185 ,   389 ,   291 ,   1041    ,   847 ,   957 ,   13413   ,   29763   ,   39789   ,0 };
            const unsigned long dim4404Kuo3Init[]   =   {   1   ,   3   ,   5   ,   7   ,   31  ,   39  ,   63  ,   181 ,   297 ,   71  ,   199 ,   1655    ,   1497    ,   8981    ,   23785   ,   40839   ,0 };
            const unsigned long dim4405Kuo3Init[]   =   {   1   ,   3   ,   5   ,   13  ,   23  ,   15  ,   25  ,   9   ,   361 ,   151 ,   1191    ,   943 ,   17  ,   715 ,   28593   ,   22609   ,0 };
            const unsigned long dim4406Kuo3Init[]   =   {   1   ,   3   ,   7   ,   3   ,   27  ,   13  ,   99  ,   109 ,   97  ,   415 ,   1515    ,   701 ,   7153    ,   14207   ,   24417   ,   52673   ,0 };
            const unsigned long dim4407Kuo3Init[]   =   {   1   ,   1   ,   1   ,   9   ,   15  ,   15  ,   15  ,   251 ,   23  ,   717 ,   219 ,   1783    ,   3657    ,   10141   ,   6525    ,   37683   ,0 };
            const unsigned long dim4408Kuo3Init[]   =   {   1   ,   1   ,   1   ,   1   ,   5   ,   55  ,   19  ,   213 ,   99  ,   73  ,   227 ,   2029    ,   1225    ,   14849   ,   17073   ,   30409   ,0 };
            const unsigned long dim4409Kuo3Init[]   =   {   1   ,   1   ,   3   ,   3   ,   15  ,   49  ,   49  ,   11  ,   289 ,   287 ,   999 ,   743 ,   1081    ,   3185    ,   30613   ,   25163   ,0 };
            const unsigned long dim4410Kuo3Init[]   =   {   1   ,   1   ,   3   ,   11  ,   15  ,   21  ,   33  ,   53  ,   465 ,   45  ,   113 ,   2865    ,   2641    ,   13797   ,   27785   ,   63159   ,0 };
            const unsigned long dim4411Kuo3Init[]   =   {   1   ,   1   ,   7   ,   13  ,   27  ,   5   ,   41  ,   21  ,   53  ,   675 ,   1217    ,   239 ,   7925    ,   6659    ,   15025   ,   56589   ,0 };
            const unsigned long dim4412Kuo3Init[]   =   {   1   ,   1   ,   3   ,   15  ,   15  ,   1   ,   7   ,   129 ,   377 ,   1017    ,   1015    ,   329 ,   3153    ,   5815    ,   7929    ,   62579   ,0 };
            const unsigned long dim4413Kuo3Init[]   =   {   1   ,   1   ,   7   ,   7   ,   19  ,   37  ,   23  ,   115 ,   29  ,   813 ,   1761    ,   751 ,   3461    ,   13027   ,   30853   ,   51399   ,0 };
            const unsigned long dim4414Kuo3Init[]   =   {   1   ,   3   ,   1   ,   15  ,   1   ,   37  ,   125 ,   143 ,   117 ,   33  ,   321 ,   1917    ,   741 ,   1281    ,   4811    ,   30907   ,0 };
            const unsigned long dim4415Kuo3Init[]   =   {   1   ,   1   ,   5   ,   9   ,   25  ,   55  ,   43  ,   69  ,   479 ,   37  ,   1891    ,   325 ,   7165    ,   11707   ,   11663   ,   9   ,0 };
            const unsigned long dim4416Kuo3Init[]   =   {   1   ,   1   ,   1   ,   7   ,   13  ,   1   ,   37  ,   5   ,   195 ,   959 ,   1707    ,   3099    ,   2295    ,   1363    ,   14463   ,   37013   ,0 };
            const unsigned long dim4417Kuo3Init[]   =   {   1   ,   1   ,   3   ,   1   ,   17  ,   45  ,   63  ,   189 ,   471 ,   273 ,   633 ,   977 ,   3377    ,   15191   ,   25127   ,   38949   ,0 };
            const unsigned long dim4418Kuo3Init[]   =   {   1   ,   3   ,   7   ,   11  ,   11  ,   49  ,   107 ,   237 ,   365 ,   813 ,   841 ,   215 ,   5087    ,   4661    ,   8045    ,   39713   ,0 };
            const unsigned long dim4419Kuo3Init[]   =   {   1   ,   1   ,   7   ,   7   ,   11  ,   39  ,   47  ,   23  ,   267 ,   929 ,   709 ,   2593    ,   3773    ,   8435    ,   15947   ,   43873   ,0 };
            const unsigned long dim4420Kuo3Init[]   =   {   1   ,   3   ,   3   ,   1   ,   7   ,   45  ,   109 ,   121 ,   389 ,   15  ,   599 ,   2887    ,   4009    ,   13557   ,   6845    ,   10149   ,0 };
            const unsigned long dim4421Kuo3Init[]   =   {   1   ,   1   ,   7   ,   11  ,   11  ,   45  ,   5   ,   39  ,   401 ,   467 ,   291 ,   1807    ,   3711    ,   13831   ,   11777   ,   28543   ,0 };
            const unsigned long dim4422Kuo3Init[]   =   {   1   ,   3   ,   1   ,   5   ,   29  ,   29  ,   121 ,   33  ,   185 ,   135 ,   479 ,   3681    ,   4763    ,   6915    ,   6689    ,   45981   ,0 };
            const unsigned long dim4423Kuo3Init[]   =   {   1   ,   1   ,   1   ,   13  ,   7   ,   57  ,   53  ,   39  ,   429 ,   985 ,   523 ,   2245    ,   8003    ,   3385    ,   31869   ,   30589   ,0 };
            const unsigned long dim4424Kuo3Init[]   =   {   1   ,   1   ,   5   ,   3   ,   23  ,   21  ,   71  ,   229 ,   83  ,   327 ,   399 ,   1413    ,   5321    ,   10593   ,   19219   ,   32837   ,0 };
            const unsigned long dim4425Kuo3Init[]   =   {   1   ,   3   ,   7   ,   7   ,   31  ,   37  ,   15  ,   119 ,   13  ,   949 ,   1589    ,   3247    ,   5671    ,   2141    ,   21189   ,   6353    ,0 };
            const unsigned long dim4426Kuo3Init[]   =   {   1   ,   1   ,   1   ,   11  ,   13  ,   25  ,   77  ,   19  ,   261 ,   237 ,   1651    ,   1165    ,   4317    ,   3265    ,   13471   ,   34163   ,0 };
            const unsigned long dim4427Kuo3Init[]   =   {   1   ,   3   ,   3   ,   3   ,   17  ,   53  ,   97  ,   245 ,   3   ,   177 ,   873 ,   2273    ,   3577    ,   10233   ,   3615    ,   38189   ,0 };
            const unsigned long dim4428Kuo3Init[]   =   {   1   ,   1   ,   7   ,   9   ,   23  ,   63  ,   5   ,   115 ,   55  ,   825 ,   877 ,   101 ,   1167    ,   10583   ,   14171   ,   42343   ,0 };
            const unsigned long dim4429Kuo3Init[]   =   {   1   ,   3   ,   7   ,   5   ,   3   ,   25  ,   119 ,   185 ,   295 ,   349 ,   587 ,   2497    ,   6103    ,   5755    ,   25191   ,   50015   ,0 };
            const unsigned long dim4430Kuo3Init[]   =   {   1   ,   1   ,   7   ,   13  ,   11  ,   55  ,   55  ,   37  ,   229 ,   803 ,   911 ,   3081    ,   1543    ,   297 ,   187 ,   51775   ,0 };
            const unsigned long dim4431Kuo3Init[]   =   {   1   ,   3   ,   1   ,   3   ,   31  ,   53  ,   105 ,   17  ,   219 ,   145 ,   1129    ,   431 ,   5651    ,   1963    ,   693 ,   18177   ,0 };
            const unsigned long dim4432Kuo3Init[]   =   {   1   ,   1   ,   7   ,   3   ,   7   ,   29  ,   61  ,   199 ,   101 ,   553 ,   1765    ,   1657    ,   1843    ,   12557   ,   27261   ,   4877    ,0 };
            const unsigned long dim4433Kuo3Init[]   =   {   1   ,   3   ,   1   ,   13  ,   19  ,   57  ,   33  ,   233 ,   179 ,   437 ,   1835    ,   3081    ,   7241    ,   4227    ,   18707   ,   27027   ,0 };
            const unsigned long dim4434Kuo3Init[]   =   {   1   ,   1   ,   3   ,   15  ,   7   ,   61  ,   45  ,   55  ,   153 ,   97  ,   1081    ,   2937    ,   2911    ,   3927    ,   17915   ,   275 ,0 };
            const unsigned long dim4435Kuo3Init[]   =   {   1   ,   3   ,   7   ,   15  ,   19  ,   25  ,   95  ,   85  ,   27  ,   589 ,   121 ,   2003    ,   4891    ,   6737    ,   4835    ,   10443   ,0 };
            const unsigned long dim4436Kuo3Init[]   =   {   1   ,   1   ,   5   ,   7   ,   5   ,   61  ,   21  ,   7   ,   439 ,   351 ,   1119    ,   937 ,   5517    ,   5001    ,   18859   ,   12427   ,0 };
            const unsigned long dim4437Kuo3Init[]   =   {   1   ,   1   ,   1   ,   1   ,   13  ,   21  ,   87  ,   73  ,   61  ,   71  ,   885 ,   4029    ,   2893    ,   833 ,   18813   ,   18617   ,0 };
            const unsigned long dim4438Kuo3Init[]   =   {   1   ,   1   ,   5   ,   7   ,   23  ,   63  ,   111 ,   159 ,   39  ,   937 ,   719 ,   1641    ,   7901    ,   7349    ,   31397   ,   38111   ,0 };
            const unsigned long dim4439Kuo3Init[]   =   {   1   ,   1   ,   5   ,   9   ,   9   ,   9   ,   105 ,   33  ,   205 ,   829 ,   1929    ,   1925    ,   1735    ,   10217   ,   1747    ,   38351   ,0 };
            const unsigned long dim4440Kuo3Init[]   =   {   1   ,   3   ,   1   ,   13  ,   9   ,   13  ,   111 ,   39  ,   465 ,   635 ,   989 ,   2079    ,   7781    ,   13229   ,   1039    ,   21061   ,0 };
            const unsigned long dim4441Kuo3Init[]   =   {   1   ,   3   ,   3   ,   1   ,   29  ,   15  ,   123 ,   137 ,   75  ,   561 ,   1699    ,   57  ,   8179    ,   5693    ,   6333    ,   25581   ,0 };
            const unsigned long dim4442Kuo3Init[]   =   {   1   ,   1   ,   7   ,   9   ,   3   ,   17  ,   27  ,   117 ,   95  ,   287 ,   1313    ,   1863    ,   7119    ,   16253   ,   27931   ,   42341   ,0 };
            const unsigned long dim4443Kuo3Init[]   =   {   1   ,   1   ,   1   ,   11  ,   21  ,   47  ,   83  ,   79  ,   451 ,   535 ,   1249    ,   1125    ,   5387    ,   11309   ,   12115   ,   45877   ,0 };
            const unsigned long dim4444Kuo3Init[]   =   {   1   ,   1   ,   3   ,   7   ,   21  ,   17  ,   93  ,   125 ,   467 ,   465 ,   619 ,   1899    ,   5501    ,   233 ,   28371   ,   7199    ,0 };
            const unsigned long dim4445Kuo3Init[]   =   {   1   ,   3   ,   7   ,   9   ,   3   ,   5   ,   5   ,   111 ,   107 ,   95  ,   2023    ,   3529    ,   5135    ,   3377    ,   22025   ,   41931   ,0 };
            const unsigned long dim4446Kuo3Init[]   =   {   1   ,   1   ,   3   ,   13  ,   31  ,   19  ,   115 ,   155 ,   349 ,   911 ,   1893    ,   3617    ,   689 ,   16323   ,   13523   ,   20373   ,0 };
            const unsigned long dim4447Kuo3Init[]   =   {   1   ,   1   ,   3   ,   5   ,   25  ,   7   ,   57  ,   117 ,   65  ,   469 ,   1669    ,   407 ,   3693    ,   1743    ,   16697   ,   30749   ,0 };
            const unsigned long dim4448Kuo3Init[]   =   {   1   ,   1   ,   1   ,   5   ,   29  ,   11  ,   117 ,   187 ,   13  ,   77  ,   45  ,   39  ,   6237    ,   9899    ,   28805   ,   9389    ,0 };
            const unsigned long dim4449Kuo3Init[]   =   {   1   ,   3   ,   1   ,   5   ,   5   ,   33  ,   43  ,   45  ,   331 ,   615 ,   961 ,   2143    ,   4403    ,   543 ,   22783   ,   47063   ,0 };
            const unsigned long dim4450Kuo3Init[]   =   {   1   ,   3   ,   1   ,   11  ,   7   ,   1   ,   93  ,   153 ,   417 ,   859 ,   451 ,   1211    ,   6439    ,   8043    ,   5759    ,   60265   ,0 };
            const unsigned long dim4451Kuo3Init[]   =   {   1   ,   3   ,   1   ,   9   ,   19  ,   53  ,   57  ,   93  ,   295 ,   773 ,   533 ,   2685    ,   7429    ,   2855    ,   7665    ,   46257   ,0 };
            const unsigned long dim4452Kuo3Init[]   =   {   1   ,   3   ,   5   ,   15  ,   29  ,   5   ,   93  ,   159 ,   7   ,   845 ,   337 ,   881 ,   2529    ,   15677   ,   9795    ,   31347   ,0 };
            const unsigned long dim4453Kuo3Init[]   =   {   1   ,   1   ,   3   ,   3   ,   11  ,   49  ,   97  ,   59  ,   157 ,   695 ,   65  ,   1865    ,   4107    ,   15283   ,   2309    ,   13841   ,0 };
            const unsigned long dim4454Kuo3Init[]   =   {   1   ,   3   ,   3   ,   15  ,   21  ,   31  ,   61  ,   85  ,   361 ,   893 ,   1223    ,   259 ,   2771    ,   4973    ,   18661   ,   16653   ,0 };
            const unsigned long dim4455Kuo3Init[]   =   {   1   ,   1   ,   1   ,   7   ,   19  ,   31  ,   3   ,   223 ,   151 ,   785 ,   447 ,   2123    ,   3377    ,   14159   ,   933 ,   23357   ,0 };
            const unsigned long dim4456Kuo3Init[]   =   {   1   ,   3   ,   5   ,   11  ,   1   ,   55  ,   77  ,   107 ,   245 ,   193 ,   583 ,   2729    ,   831 ,   11019   ,   25433   ,   6541    ,0 };
            const unsigned long dim4457Kuo3Init[]   =   {   1   ,   3   ,   3   ,   9   ,   3   ,   35  ,   19  ,   127 ,   419 ,   203 ,   1869    ,   1477    ,   5239    ,   6113    ,   4835    ,   21131   ,0 };
            const unsigned long dim4458Kuo3Init[]   =   {   1   ,   3   ,   3   ,   7   ,   17  ,   49  ,   17  ,   23  ,   111 ,   65  ,   1603    ,   157 ,   6433    ,   2965    ,   19845   ,   59041   ,0 };
            const unsigned long dim4459Kuo3Init[]   =   {   1   ,   3   ,   1   ,   3   ,   17  ,   1   ,   17  ,   207 ,   245 ,   129 ,   1309    ,   413 ,   6933    ,   7575    ,   17179   ,   24075   ,0 };
            const unsigned long dim4460Kuo3Init[]   =   {   1   ,   1   ,   5   ,   5   ,   29  ,   23  ,   65  ,   43  ,   157 ,   569 ,   645 ,   1821    ,   605 ,   3925    ,   2863    ,   55361   ,0 };
            const unsigned long dim4461Kuo3Init[]   =   {   1   ,   3   ,   3   ,   5   ,   15  ,   43  ,   1   ,   149 ,   253 ,   529 ,   37  ,   2589    ,   2851    ,   2529    ,   32689   ,   49303   ,0 };
            const unsigned long dim4462Kuo3Init[]   =   {   1   ,   3   ,   1   ,   15  ,   3   ,   29  ,   19  ,   241 ,   153 ,   189 ,   161 ,   797 ,   5419    ,   11269   ,   2159    ,   57893   ,0 };
            const unsigned long dim4463Kuo3Init[]   =   {   1   ,   1   ,   3   ,   13  ,   19  ,   23  ,   49  ,   179 ,   459 ,   701 ,   1003    ,   2673    ,   1729    ,   1901    ,   26003   ,   51441   ,0 };
            const unsigned long dim4464Kuo3Init[]   =   {   1   ,   1   ,   3   ,   9   ,   23  ,   19  ,   47  ,   85  ,   281 ,   371 ,   773 ,   2263    ,   3235    ,   647 ,   23085   ,   10053   ,0 };
            const unsigned long dim4465Kuo3Init[]   =   {   1   ,   3   ,   3   ,   13  ,   13  ,   23  ,   13  ,   95  ,   41  ,   963 ,   213 ,   2317    ,   3889    ,   3981    ,   4749    ,   55409   ,0 };
            const unsigned long dim4466Kuo3Init[]   =   {   1   ,   1   ,   7   ,   13  ,   9   ,   39  ,   33  ,   21  ,   17  ,   201 ,   1789    ,   2593    ,   1373    ,   14601   ,   17025   ,   31861   ,0 };
            const unsigned long dim4467Kuo3Init[]   =   {   1   ,   1   ,   5   ,   9   ,   5   ,   43  ,   107 ,   169 ,   247 ,   361 ,   1245    ,   2371    ,   119 ,   4987    ,   32661   ,   19363   ,0 };
            const unsigned long dim4468Kuo3Init[]   =   {   1   ,   3   ,   7   ,   5   ,   29  ,   35  ,   81  ,   13  ,   419 ,   1003    ,   701 ,   759 ,   4415    ,   11441   ,   28985   ,   46869   ,0 };
            const unsigned long dim4469Kuo3Init[]   =   {   1   ,   3   ,   7   ,   7   ,   9   ,   57  ,   3   ,   39  ,   457 ,   455 ,   29  ,   2983    ,   4879    ,   14359   ,   12913   ,   64711   ,0 };
            const unsigned long dim4470Kuo3Init[]   =   {   1   ,   1   ,   7   ,   3   ,   9   ,   31  ,   57  ,   47  ,   55  ,   525 ,   2015    ,   2053    ,   7101    ,   8487    ,   26865   ,   29149   ,0 };
            const unsigned long dim4471Kuo3Init[]   =   {   1   ,   1   ,   7   ,   11  ,   11  ,   37  ,   67  ,   241 ,   509 ,   711 ,   263 ,   1565    ,   6803    ,   2095    ,   3055    ,   28847   ,0 };
            const unsigned long dim4472Kuo3Init[]   =   {   1   ,   3   ,   1   ,   9   ,   25  ,   23  ,   53  ,   135 ,   391 ,   625 ,   673 ,   2807    ,   6511    ,   7465    ,   21959   ,   17561   ,0 };
            const unsigned long dim4473Kuo3Init[]   =   {   1   ,   1   ,   7   ,   3   ,   23  ,   23  ,   47  ,   247 ,   297 ,   191 ,   1137    ,   3921    ,   3669    ,   4439    ,   707 ,   33529   ,0 };
            const unsigned long dim4474Kuo3Init[]   =   {   1   ,   1   ,   7   ,   15  ,   19  ,   35  ,   91  ,   75  ,   237 ,   585 ,   1047    ,   1389    ,   7147    ,   9043    ,   22363   ,   14287   ,0 };
            const unsigned long dim4475Kuo3Init[]   =   {   1   ,   3   ,   1   ,   15  ,   17  ,   35  ,   45  ,   181 ,   439 ,   307 ,   709 ,   2925    ,   3397    ,   4577    ,   30645   ,   25167   ,0 };
            const unsigned long dim4476Kuo3Init[]   =   {   1   ,   1   ,   5   ,   9   ,   25  ,   1   ,   127 ,   173 ,   397 ,   477 ,   821 ,   2017    ,   1363    ,   8615    ,   8833    ,   12771   ,0 };
            const unsigned long dim4477Kuo3Init[]   =   {   1   ,   1   ,   5   ,   9   ,   9   ,   1   ,   33  ,   137 ,   383 ,   465 ,   1211    ,   3563    ,   5035    ,   6609    ,   31667   ,   10853   ,0 };
            const unsigned long dim4478Kuo3Init[]   =   {   1   ,   3   ,   7   ,   9   ,   11  ,   49  ,   61  ,   51  ,   31  ,   811 ,   623 ,   2213    ,   7131    ,   15679   ,   5335    ,   53001   ,0 };
            const unsigned long dim4479Kuo3Init[]   =   {   1   ,   1   ,   5   ,   7   ,   27  ,   55  ,   11  ,   229 ,   331 ,   109 ,   109 ,   877 ,   6087    ,   10023   ,   22627   ,   64253   ,0 };
            const unsigned long dim4480Kuo3Init[]   =   {   1   ,   1   ,   5   ,   3   ,   11  ,   13  ,   105 ,   31  ,   357 ,   589 ,   869 ,   1583    ,   3441    ,   15783   ,   25579   ,   25335   ,0 };
            const unsigned long dim4481Kuo3Init[]   =   {   1   ,   3   ,   7   ,   7   ,   7   ,   9   ,   35  ,   29  ,   153 ,   425 ,   1145    ,   2021    ,   115 ,   15865   ,   4305    ,   62417   ,0 };
            const unsigned long dim4482Kuo3Init[]   =   {   1   ,   3   ,   5   ,   3   ,   27  ,   17  ,   3   ,   57  ,   111 ,   393 ,   711 ,   2519    ,   7223    ,   10117   ,   5509    ,   63361   ,0 };
            const unsigned long dim4483Kuo3Init[]   =   {   1   ,   1   ,   5   ,   7   ,   9   ,   55  ,   101 ,   199 ,   441 ,   141 ,   1349    ,   3017    ,   7625    ,   1457    ,   27947   ,   13269   ,0 };
            const unsigned long dim4484Kuo3Init[]   =   {   1   ,   1   ,   5   ,   15  ,   5   ,   23  ,   3   ,   87  ,   61  ,   19  ,   1449    ,   2317    ,   4049    ,   3397    ,   6271    ,   53317   ,0 };
            const unsigned long dim4485Kuo3Init[]   =   {   1   ,   3   ,   5   ,   3   ,   11  ,   35  ,   95  ,   141 ,   355 ,   749 ,   1247    ,   3495    ,   5719    ,   16201   ,   9649    ,   12395   ,0 };
            const unsigned long dim4486Kuo3Init[]   =   {   1   ,   3   ,   7   ,   11  ,   23  ,   29  ,   79  ,   171 ,   259 ,   923 ,   2035    ,   783 ,   3571    ,   14987   ,   30025   ,   19185   ,0 };
            const unsigned long dim4487Kuo3Init[]   =   {   1   ,   1   ,   7   ,   13  ,   17  ,   31  ,   75  ,   141 ,   131 ,   391 ,   303 ,   2641    ,   5527    ,   12629   ,   1097    ,   47589   ,0 };
            const unsigned long dim4488Kuo3Init[]   =   {   1   ,   3   ,   3   ,   15  ,   9   ,   55  ,   23  ,   207 ,   211 ,   633 ,   805 ,   3079    ,   5495    ,   4651    ,   26483   ,   59323   ,0 };
            const unsigned long dim4489Kuo3Init[]   =   {   1   ,   1   ,   1   ,   9   ,   9   ,   17  ,   97  ,   173 ,   123 ,   139 ,   595 ,   3859    ,   4615    ,   15565   ,   9243    ,   19933   ,0 };
            const unsigned long dim4490Kuo3Init[]   =   {   1   ,   3   ,   5   ,   9   ,   31  ,   51  ,   1   ,   133 ,   213 ,   877 ,   23  ,   3661    ,   3451    ,   11705   ,   27943   ,   10995   ,0 };
            const unsigned long dim4491Kuo3Init[]   =   {   1   ,   3   ,   3   ,   11  ,   5   ,   1   ,   91  ,   113 ,   147 ,   825 ,   795 ,   2871    ,   5175    ,   6843    ,   17829   ,   25375   ,0 };
            const unsigned long dim4492Kuo3Init[]   =   {   1   ,   3   ,   3   ,   3   ,   7   ,   15  ,   105 ,   227 ,   97  ,   985 ,   347 ,   2833    ,   1237    ,   9419    ,   17853   ,   43467   ,0 };
            const unsigned long dim4493Kuo3Init[]   =   {   1   ,   3   ,   7   ,   7   ,   17  ,   47  ,   3   ,   237 ,   97  ,   915 ,   1463    ,   477 ,   2237    ,   7755    ,   11779   ,   35897   ,0 };
            const unsigned long dim4494Kuo3Init[]   =   {   1   ,   3   ,   1   ,   9   ,   17  ,   17  ,   97  ,   163 ,   181 ,   411 ,   1145    ,   623 ,   3017    ,   1335    ,   6999    ,   57149   ,0 };
            const unsigned long dim4495Kuo3Init[]   =   {   1   ,   1   ,   1   ,   7   ,   27  ,   49  ,   57  ,   49  ,   217 ,   635 ,   973 ,   1153    ,   8097    ,   4669    ,   30691   ,   1949    ,0 };
            const unsigned long dim4496Kuo3Init[]   =   {   1   ,   1   ,   1   ,   7   ,   13  ,   47  ,   107 ,   247 ,   269 ,   803 ,   1919    ,   3047    ,   2967    ,   63  ,   12287   ,   39925   ,0 };
            const unsigned long dim4497Kuo3Init[]   =   {   1   ,   1   ,   5   ,   3   ,   27  ,   15  ,   27  ,   255 ,   27  ,   463 ,   621 ,   2443    ,   5615    ,   14589   ,   10869   ,   5013    ,0 };
            const unsigned long dim4498Kuo3Init[]   =   {   1   ,   3   ,   5   ,   15  ,   25  ,   15  ,   75  ,   223 ,   17  ,   227 ,   911 ,   3069    ,   2583    ,   8811    ,   14191   ,   2819    ,0 };
            const unsigned long dim4499Kuo3Init[]   =   {   1   ,   3   ,   5   ,   15  ,   25  ,   13  ,   19  ,   157 ,   495 ,   997 ,   1997    ,   3175    ,   1515    ,   15693   ,   15585   ,   2753    ,0 };
            const unsigned long dim4500Kuo3Init[]   =   {   1   ,   1   ,   7   ,   7   ,   7   ,   5   ,   67  ,   95  ,   337 ,   539 ,   2019    ,   2009    ,   7171    ,   1089    ,   7955    ,   17557   ,0 };
            const unsigned long dim4501Kuo3Init[]   =   {   1   ,   3   ,   3   ,   3   ,   7   ,   63  ,   61  ,   31  ,   61  ,   483 ,   365 ,   1153    ,   495 ,   7357    ,   2831    ,   23191   ,0 };
            const unsigned long dim4502Kuo3Init[]   =   {   1   ,   1   ,   7   ,   1   ,   17  ,   35  ,   121 ,   143 ,   411 ,   895 ,   1859    ,   2205    ,   6565    ,   11611   ,   20941   ,   47151   ,0 };
            const unsigned long dim4503Kuo3Init[]   =   {   1   ,   3   ,   7   ,   9   ,   27  ,   21  ,   53  ,   51  ,   77  ,   251 ,   1355    ,   175 ,   3949    ,   2903    ,   26991   ,   41571   ,0 };
            const unsigned long dim4504Kuo3Init[]   =   {   1   ,   1   ,   1   ,   15  ,   17  ,   1   ,   15  ,   253 ,   99  ,   611 ,   1949    ,   1917    ,   4363    ,   2887    ,   30421   ,   42197   ,0 };
            const unsigned long dim4505Kuo3Init[]   =   {   1   ,   3   ,   3   ,   11  ,   29  ,   45  ,   111 ,   49  ,   389 ,   171 ,   1069    ,   3061    ,   6657    ,   12097   ,   533 ,   9459    ,0 };
            const unsigned long dim4506Kuo3Init[]   =   {   1   ,   1   ,   3   ,   1   ,   5   ,   27  ,   39  ,   15  ,   381 ,   119 ,   137 ,   3641    ,   8047    ,   1391    ,   13093   ,   50659   ,0 };
            const unsigned long dim4507Kuo3Init[]   =   {   1   ,   1   ,   7   ,   9   ,   27  ,   59  ,   123 ,   235 ,   301 ,   933 ,   335 ,   3927    ,   2811    ,   16133   ,   5999    ,   19933   ,0 };
            const unsigned long dim4508Kuo3Init[]   =   {   1   ,   3   ,   3   ,   13  ,   15  ,   55  ,   69  ,   199 ,   353 ,   131 ,   753 ,   2765    ,   4005    ,   13575   ,   16607   ,   51253   ,0 };
            const unsigned long dim4509Kuo3Init[]   =   {   1   ,   1   ,   7   ,   1   ,   19  ,   53  ,   31  ,   233 ,   289 ,   939 ,   31  ,   1741    ,   5467    ,   1987    ,   13923   ,   49857   ,0 };
            const unsigned long dim4510Kuo3Init[]   =   {   1   ,   1   ,   7   ,   7   ,   31  ,   45  ,   45  ,   237 ,   427 ,   305 ,   1753    ,   1237    ,   3573    ,   10813   ,   10609   ,   24187   ,0 };
            const unsigned long dim4511Kuo3Init[]   =   {   1   ,   1   ,   1   ,   13  ,   21  ,   41  ,   81  ,   219 ,   185 ,   107 ,   71  ,   2017    ,   6587    ,   3351    ,   10207   ,   10545   ,0 };
            const unsigned long dim4512Kuo3Init[]   =   {   1   ,   1   ,   1   ,   5   ,   5   ,   21  ,   127 ,   129 ,   209 ,   313 ,   1273    ,   3229    ,   6733    ,   4431    ,   7297    ,   41569   ,0 };
            const unsigned long dim4513Kuo3Init[]   =   {   1   ,   1   ,   3   ,   15  ,   17  ,   61  ,   55  ,   127 ,   85  ,   913 ,   1173    ,   3837    ,   5579    ,   10309   ,   24659   ,   22697   ,0 };
            const unsigned long dim4514Kuo3Init[]   =   {   1   ,   1   ,   3   ,   3   ,   11  ,   11  ,   91  ,   165 ,   367 ,   79  ,   43  ,   679 ,   3285    ,   15469   ,   21511   ,   57755   ,0 };
            const unsigned long dim4515Kuo3Init[]   =   {   1   ,   1   ,   1   ,   13  ,   15  ,   1   ,   125 ,   47  ,   101 ,   999 ,   521 ,   543 ,   5535    ,   10191   ,   7497    ,   15677   ,0 };
            const unsigned long dim4516Kuo3Init[]   =   {   1   ,   3   ,   3   ,   9   ,   9   ,   15  ,   63  ,   239 ,   471 ,   903 ,   1473    ,   717 ,   6547    ,   8445    ,   20737   ,   41457   ,0 };
            const unsigned long dim4517Kuo3Init[]   =   {   1   ,   1   ,   5   ,   15  ,   13  ,   31  ,   43  ,   93  ,   325 ,   393 ,   373 ,   971 ,   6045    ,   14981   ,   28011   ,   35997   ,0 };
            const unsigned long dim4518Kuo3Init[]   =   {   1   ,   3   ,   1   ,   15  ,   11  ,   45  ,   41  ,   35  ,   381 ,   779 ,   443 ,   3395    ,   2495    ,   14795   ,   4695    ,   22745   ,0 };
            const unsigned long dim4519Kuo3Init[]   =   {   1   ,   3   ,   3   ,   7   ,   21  ,   9   ,   89  ,   173 ,   233 ,   935 ,   683 ,   1533    ,   953 ,   2627    ,   30975   ,   43449   ,0 };
            const unsigned long dim4520Kuo3Init[]   =   {   1   ,   1   ,   5   ,   3   ,   25  ,   55  ,   107 ,   105 ,   99  ,   959 ,   1935    ,   3489    ,   3671    ,   14999   ,   11707   ,   10285   ,0 };
            const unsigned long dim4521Kuo3Init[]   =   {   1   ,   3   ,   5   ,   3   ,   31  ,   17  ,   123 ,   119 ,   63  ,   169 ,   1283    ,   3691    ,   1411    ,   3437    ,   27941   ,   57089   ,0 };
            const unsigned long dim4522Kuo3Init[]   =   {   1   ,   3   ,   3   ,   1   ,   17  ,   9   ,   115 ,   69  ,   481 ,   689 ,   2007    ,   3815    ,   4419    ,   11249   ,   31469   ,   55783   ,0 };
            const unsigned long dim4523Kuo3Init[]   =   {   1   ,   3   ,   7   ,   15  ,   9   ,   39  ,   65  ,   125 ,   253 ,   413 ,   1283    ,   2073    ,   6563    ,   2111    ,   22865   ,   59393   ,0 };
            const unsigned long dim4524Kuo3Init[]   =   {   1   ,   3   ,   3   ,   15  ,   19  ,   1   ,   67  ,   253 ,   489 ,   185 ,   297 ,   261 ,   5203    ,   8685    ,   10515   ,   58595   ,0 };
            const unsigned long dim4525Kuo3Init[]   =   {   1   ,   1   ,   1   ,   13  ,   23  ,   39  ,   95  ,   233 ,   171 ,   653 ,   2025    ,   2285    ,   4477    ,   9275    ,   9717    ,   43321   ,0 };
            const unsigned long dim4526Kuo3Init[]   =   {   1   ,   1   ,   7   ,   3   ,   23  ,   35  ,   77  ,   119 ,   193 ,   41  ,   1057    ,   4017    ,   235 ,   13819   ,   7911    ,   3981    ,0 };
            const unsigned long dim4527Kuo3Init[]   =   {   1   ,   3   ,   5   ,   9   ,   15  ,   33  ,   41  ,   221 ,   1   ,   307 ,   1325    ,   2899    ,   399 ,   801 ,   24611   ,   31941   ,0 };
            const unsigned long dim4528Kuo3Init[]   =   {   1   ,   1   ,   1   ,   15  ,   1   ,   7   ,   91  ,   27  ,   247 ,   667 ,   483 ,   573 ,   481 ,   11771   ,   31139   ,   31979   ,0 };
            const unsigned long dim4529Kuo3Init[]   =   {   1   ,   3   ,   5   ,   1   ,   3   ,   23  ,   39  ,   73  ,   101 ,   367 ,   1259    ,   4053    ,   3055    ,   1231    ,   27981   ,   54287   ,0 };
            const unsigned long dim4530Kuo3Init[]   =   {   1   ,   3   ,   3   ,   7   ,   3   ,   31  ,   9   ,   71  ,   339 ,   33  ,   1255    ,   1353    ,   5575    ,   12481   ,   27881   ,   17293   ,0 };
            const unsigned long dim4531Kuo3Init[]   =   {   1   ,   3   ,   5   ,   13  ,   5   ,   57  ,   123 ,   193 ,   443 ,   185 ,   851 ,   539 ,   4217    ,   915 ,   1247    ,   20163   ,0 };
            const unsigned long dim4532Kuo3Init[]   =   {   1   ,   3   ,   7   ,   5   ,   11  ,   13  ,   99  ,   141 ,   439 ,   259 ,   1177    ,   453 ,   6037    ,   5813    ,   25147   ,   20381   ,0 };
            const unsigned long dim4533Kuo3Init[]   =   {   1   ,   1   ,   1   ,   5   ,   25  ,   17  ,   105 ,   91  ,   33  ,   15  ,   2035    ,   3693    ,   2249    ,   5825    ,   29037   ,   12529   ,0 };
            const unsigned long dim4534Kuo3Init[]   =   {   1   ,   3   ,   1   ,   13  ,   1   ,   33  ,   113 ,   237 ,   13  ,   845 ,   1127    ,   4067    ,   825 ,   8391    ,   1269    ,   51473   ,0 };
            const unsigned long dim4535Kuo3Init[]   =   {   1   ,   3   ,   5   ,   3   ,   27  ,   5   ,   77  ,   97  ,   33  ,   569 ,   1327    ,   751 ,   2817    ,   775 ,   32275   ,   55757   ,0 };
            const unsigned long dim4536Kuo3Init[]   =   {   1   ,   1   ,   1   ,   7   ,   25  ,   1   ,   73  ,   105 ,   19  ,   1007    ,   371 ,   269 ,   4833    ,   13881   ,   21503   ,   35303   ,0 };
            const unsigned long dim4537Kuo3Init[]   =   {   1   ,   3   ,   1   ,   3   ,   23  ,   39  ,   27  ,   203 ,   351 ,   433 ,   399 ,   405 ,   39  ,   7773    ,   27345   ,   57109   ,0 };
            const unsigned long dim4538Kuo3Init[]   =   {   1   ,   1   ,   5   ,   13  ,   7   ,   7   ,   21  ,   33  ,   481 ,   201 ,   7   ,   2377    ,   1009    ,   7593    ,   19381   ,   10325   ,0 };
            const unsigned long dim4539Kuo3Init[]   =   {   1   ,   1   ,   1   ,   13  ,   15  ,   61  ,   29  ,   59  ,   255 ,   213 ,   1207    ,   3783    ,   4375    ,   8137    ,   6733    ,   10809   ,0 };
            const unsigned long dim4540Kuo3Init[]   =   {   1   ,   1   ,   1   ,   9   ,   25  ,   37  ,   83  ,   169 ,   267 ,   21  ,   1211    ,   3099    ,   239 ,   2489    ,   12843   ,   4395    ,0 };
            const unsigned long dim4541Kuo3Init[]   =   {   1   ,   1   ,   1   ,   13  ,   17  ,   47  ,   77  ,   63  ,   361 ,   567 ,   245 ,   3569    ,   7707    ,   15463   ,   29579   ,   55017   ,0 };
            const unsigned long dim4542Kuo3Init[]   =   {   1   ,   1   ,   7   ,   13  ,   23  ,   3   ,   111 ,   21  ,   417 ,   609 ,   547 ,   3297    ,   7487    ,   3285    ,   16913   ,   34005   ,0 };
            const unsigned long dim4543Kuo3Init[]   =   {   1   ,   3   ,   7   ,   5   ,   9   ,   37  ,   39  ,   173 ,   129 ,   215 ,   1939    ,   2745    ,   319 ,   12707   ,   4665    ,   32065   ,0 };
            const unsigned long dim4544Kuo3Init[]   =   {   1   ,   1   ,   3   ,   3   ,   21  ,   39  ,   67  ,   85  ,   223 ,   591 ,   681 ,   1655    ,   6651    ,   3547    ,   22459   ,   291 ,0 };
            const unsigned long dim4545Kuo3Init[]   =   {   1   ,   1   ,   7   ,   11  ,   17  ,   9   ,   61  ,   121 ,   489 ,   129 ,   489 ,   1261    ,   2377    ,   9243    ,   15123   ,   59249   ,0 };
            const unsigned long dim4546Kuo3Init[]   =   {   1   ,   1   ,   5   ,   9   ,   1   ,   31  ,   1   ,   171 ,   331 ,   233 ,   1033    ,   1975    ,   1515    ,   13295   ,   9383    ,   52107   ,0 };
            const unsigned long dim4547Kuo3Init[]   =   {   1   ,   1   ,   1   ,   13  ,   17  ,   45  ,   75  ,   209 ,   187 ,   441 ,   1227    ,   4063    ,   5699    ,   3397    ,   25793   ,   58187   ,0 };
            const unsigned long dim4548Kuo3Init[]   =   {   1   ,   3   ,   1   ,   3   ,   17  ,   13  ,   67  ,   73  ,   141 ,   451 ,   1307    ,   3115    ,   1273    ,   5275    ,   32383   ,   4663    ,0 };
            const unsigned long dim4549Kuo3Init[]   =   {   1   ,   1   ,   5   ,   15  ,   5   ,   37  ,   77  ,   159 ,   339 ,   443 ,   1663    ,   345 ,   1513    ,   3195    ,   15131   ,   32473   ,0 };
            const unsigned long dim4550Kuo3Init[]   =   {   1   ,   1   ,   3   ,   11  ,   29  ,   19  ,   13  ,   47  ,   409 ,   141 ,   1907    ,   3473    ,   4799    ,   213 ,   341 ,   45091   ,0 };
            const unsigned long dim4551Kuo3Init[]   =   {   1   ,   3   ,   1   ,   1   ,   15  ,   29  ,   27  ,   179 ,   367 ,   439 ,   1073    ,   1783    ,   735 ,   11731   ,   20101   ,   7889    ,0 };
            const unsigned long dim4552Kuo3Init[]   =   {   1   ,   1   ,   3   ,   15  ,   23  ,   35  ,   107 ,   53  ,   325 ,   51  ,   831 ,   1693    ,   5161    ,   13655   ,   31957   ,   60089   ,0 };
            const unsigned long dim4553Kuo3Init[]   =   {   1   ,   1   ,   5   ,   11  ,   31  ,   13  ,   103 ,   109 ,   71  ,   237 ,   1271    ,   629 ,   6813    ,   10253   ,   17161   ,   58597   ,0 };
            const unsigned long dim4554Kuo3Init[]   =   {   1   ,   3   ,   3   ,   1   ,   29  ,   37  ,   47  ,   173 ,   131 ,   787 ,   39  ,   879 ,   215 ,   13625   ,   4031    ,   56677   ,0 };
            const unsigned long dim4555Kuo3Init[]   =   {   1   ,   1   ,   7   ,   5   ,   29  ,   3   ,   61  ,   227 ,   23  ,   699 ,   1291    ,   3909    ,   6995    ,   4085    ,   8961    ,   40275   ,0 };
            const unsigned long dim4556Kuo3Init[]   =   {   1   ,   1   ,   5   ,   7   ,   27  ,   21  ,   25  ,   213 ,   359 ,   701 ,   1453    ,   3753    ,   5493    ,   5971    ,   30375   ,   54707   ,0 };
            const unsigned long dim4557Kuo3Init[]   =   {   1   ,   3   ,   7   ,   7   ,   7   ,   63  ,   67  ,   47  ,   341 ,   451 ,   1205    ,   3125    ,   1575    ,   6229    ,   13951   ,   4539    ,0 };
            const unsigned long dim4558Kuo3Init[]   =   {   1   ,   1   ,   3   ,   13  ,   25  ,   3   ,   93  ,   91  ,   507 ,   843 ,   623 ,   865 ,   5873    ,   10397   ,   4135    ,   11313   ,0 };
            const unsigned long dim4559Kuo3Init[]   =   {   1   ,   3   ,   7   ,   15  ,   11  ,   39  ,   111 ,   149 ,   439 ,   519 ,   1015    ,   2777    ,   1643    ,   9703    ,   13767   ,   7351    ,0 };
            const unsigned long dim4560Kuo3Init[]   =   {   1   ,   1   ,   5   ,   5   ,   27  ,   55  ,   79  ,   39  ,   23  ,   451 ,   1507    ,   2117    ,   6881    ,   4623    ,   26289   ,   23435   ,0 };
            const unsigned long dim4561Kuo3Init[]   =   {   1   ,   3   ,   3   ,   11  ,   27  ,   29  ,   21  ,   63  ,   229 ,   173 ,   1777    ,   2163    ,   5185    ,   15487   ,   19661   ,   49185   ,0 };
            const unsigned long dim4562Kuo3Init[]   =   {   1   ,   3   ,   1   ,   9   ,   3   ,   29  ,   59  ,   63  ,   477 ,   51  ,   1257    ,   2301    ,   3495    ,   9475    ,   32201   ,   47511   ,0 };
            const unsigned long dim4563Kuo3Init[]   =   {   1   ,   1   ,   1   ,   7   ,   29  ,   45  ,   115 ,   235 ,   201 ,   847 ,   57  ,   1783    ,   2203    ,   13077   ,   4917    ,   6773    ,0 };
            const unsigned long dim4564Kuo3Init[]   =   {   1   ,   1   ,   5   ,   9   ,   13  ,   49  ,   77  ,   1   ,   105 ,   747 ,   1389    ,   2461    ,   3823    ,   3315    ,   25295   ,   24761   ,0 };
            const unsigned long dim4565Kuo3Init[]   =   {   1   ,   3   ,   7   ,   7   ,   19  ,   5   ,   83  ,   157 ,   37  ,   499 ,   1893    ,   2787    ,   6427    ,   8575    ,   26547   ,   28871   ,0 };
            const unsigned long dim4566Kuo3Init[]   =   {   1   ,   1   ,   3   ,   9   ,   29  ,   47  ,   109 ,   151 ,   125 ,   1005    ,   751 ,   2059    ,   7045    ,   15483   ,   5227    ,   58193   ,0 };
            const unsigned long dim4567Kuo3Init[]   =   {   1   ,   1   ,   3   ,   1   ,   11  ,   41  ,   63  ,   155 ,   175 ,   943 ,   161 ,   889 ,   6001    ,   3537    ,   23609   ,   23595   ,0 };
            const unsigned long dim4568Kuo3Init[]   =   {   1   ,   3   ,   1   ,   11  ,   5   ,   51  ,   121 ,   11  ,   81  ,   175 ,   1953    ,   451 ,   4723    ,   11083   ,   26527   ,   41887   ,0 };
            const unsigned long dim4569Kuo3Init[]   =   {   1   ,   1   ,   3   ,   7   ,   13  ,   59  ,   47  ,   205 ,   503 ,   249 ,   1263    ,   1767    ,   2553    ,   9755    ,   909 ,   27129   ,0 };
            const unsigned long dim4570Kuo3Init[]   =   {   1   ,   1   ,   5   ,   13  ,   11  ,   19  ,   123 ,   1   ,   429 ,   455 ,   85  ,   1021    ,   7119    ,   8563    ,   11115   ,   24741   ,0 };
            const unsigned long dim4571Kuo3Init[]   =   {   1   ,   1   ,   1   ,   1   ,   5   ,   55  ,   1   ,   1   ,   17  ,   221 ,   1633    ,   4065    ,   9   ,   4091    ,   29557   ,   33855   ,0 };
            const unsigned long dim4572Kuo3Init[]   =   {   1   ,   3   ,   3   ,   11  ,   9   ,   3   ,   15  ,   159 ,   243 ,   245 ,   1845    ,   3841    ,   3321    ,   8631    ,   1131    ,   7327    ,0 };
            const unsigned long dim4573Kuo3Init[]   =   {   1   ,   1   ,   5   ,   5   ,   3   ,   9   ,   119 ,   99  ,   337 ,   969 ,   1313    ,   4083    ,   2053    ,   10277   ,   15941   ,   16969   ,0 };
            const unsigned long dim4574Kuo3Init[]   =   {   1   ,   1   ,   7   ,   13  ,   15  ,   35  ,   25  ,   239 ,   345 ,   785 ,   747 ,   233 ,   7175    ,   10319   ,   23783   ,   55803   ,0 };
            const unsigned long dim4575Kuo3Init[]   =   {   1   ,   1   ,   7   ,   1   ,   7   ,   5   ,   91  ,   119 ,   317 ,   831 ,   351 ,   2149    ,   7531    ,   16373   ,   23461   ,   19863   ,0 };
            const unsigned long dim4576Kuo3Init[]   =   {   1   ,   1   ,   7   ,   1   ,   29  ,   7   ,   5   ,   43  ,   7   ,   717 ,   1459    ,   713 ,   5453    ,   7471    ,   12909   ,   36941   ,0 };
            const unsigned long dim4577Kuo3Init[]   =   {   1   ,   3   ,   5   ,   15  ,   9   ,   53  ,   59  ,   201 ,   333 ,   603 ,   629 ,   1819    ,   3397    ,   2823    ,   6459    ,   16213   ,0 };
            const unsigned long dim4578Kuo3Init[]   =   {   1   ,   3   ,   1   ,   13  ,   27  ,   47  ,   67  ,   119 ,   145 ,   81  ,   507 ,   53  ,   7407    ,   6327    ,   11387   ,   45391   ,0 };
            const unsigned long dim4579Kuo3Init[]   =   {   1   ,   1   ,   7   ,   13  ,   7   ,   63  ,   53  ,   171 ,   381 ,   523 ,   825 ,   2785    ,   7077    ,   13601   ,   329 ,   26677   ,0 };
            const unsigned long dim4580Kuo3Init[]   =   {   1   ,   1   ,   5   ,   13  ,   7   ,   25  ,   91  ,   67  ,   267 ,   463 ,   1869    ,   2041    ,   5949    ,   1635    ,   21551   ,   37331   ,0 };
            const unsigned long dim4581Kuo3Init[]   =   {   1   ,   1   ,   1   ,   5   ,   25  ,   1   ,   17  ,   113 ,   445 ,   685 ,   1201    ,   3299    ,   8139    ,   1463    ,   26067   ,   14613   ,0 };
            const unsigned long dim4582Kuo3Init[]   =   {   1   ,   3   ,   1   ,   13  ,   3   ,   19  ,   103 ,   159 ,   337 ,   999 ,   755 ,   3935    ,   5161    ,   8681    ,   17375   ,   20163   ,0 };
            const unsigned long dim4583Kuo3Init[]   =   {   1   ,   3   ,   5   ,   9   ,   1   ,   13  ,   117 ,   209 ,   93  ,   613 ,   271 ,   1321    ,   2265    ,   2739    ,   5055    ,   27581   ,0 };
            const unsigned long dim4584Kuo3Init[]   =   {   1   ,   1   ,   1   ,   1   ,   7   ,   21  ,   111 ,   45  ,   361 ,   665 ,   1243    ,   1485    ,   7833    ,   15515   ,   26437   ,   55983   ,0 };
            const unsigned long dim4585Kuo3Init[]   =   {   1   ,   1   ,   5   ,   7   ,   1   ,   31  ,   123 ,   199 ,   347 ,   235 ,   1367    ,   2451    ,   7211    ,   6095    ,   19245   ,   55993   ,0 };
            const unsigned long dim4586Kuo3Init[]   =   {   1   ,   3   ,   7   ,   3   ,   25  ,   9   ,   25  ,   117 ,   315 ,   497 ,   1235    ,   1969    ,   7981    ,   9743    ,   20951   ,   30635   ,0 };

            const unsigned long * const Kuo3initializers[4586]
            =
            {
                dim1Kuo3Init,
                dim2Kuo3Init,
                dim3Kuo3Init,
                dim4Kuo3Init,
                dim5Kuo3Init,
                dim6Kuo3Init,
                dim7Kuo3Init,
                dim8Kuo3Init,
                dim9Kuo3Init,
                dim10Kuo3Init,
                dim11Kuo3Init,
                dim12Kuo3Init,
                dim13Kuo3Init,
                dim14Kuo3Init,
                dim15Kuo3Init,
                dim16Kuo3Init,
                dim17Kuo3Init,
                dim18Kuo3Init,
                dim19Kuo3Init,
                dim20Kuo3Init,
                dim21Kuo3Init,
                dim22Kuo3Init,
                dim23Kuo3Init,
                dim24Kuo3Init,
                dim25Kuo3Init,
                dim26Kuo3Init,
                dim27Kuo3Init,
                dim28Kuo3Init,
                dim29Kuo3Init,
                dim30Kuo3Init,
                dim31Kuo3Init,
                dim32Kuo3Init,
                dim33Kuo3Init,
                dim34Kuo3Init,
                dim35Kuo3Init,
                dim36Kuo3Init,
                dim37Kuo3Init,
                dim38Kuo3Init,
                dim39Kuo3Init,
                dim40Kuo3Init,
                dim41Kuo3Init,
                dim42Kuo3Init,
                dim43Kuo3Init,
                dim44Kuo3Init,
                dim45Kuo3Init,
                dim46Kuo3Init,
                dim47Kuo3Init,
                dim48Kuo3Init,
                dim49Kuo3Init,
                dim50Kuo3Init,
                dim51Kuo3Init,
                dim52Kuo3Init,
                dim53Kuo3Init,
                dim54Kuo3Init,
                dim55Kuo3Init,
                dim56Kuo3Init,
                dim57Kuo3Init,
                dim58Kuo3Init,
                dim59Kuo3Init,
                dim60Kuo3Init,
                dim61Kuo3Init,
                dim62Kuo3Init,
                dim63Kuo3Init,
                dim64Kuo3Init,
                dim65Kuo3Init,
                dim66Kuo3Init,
                dim67Kuo3Init,
                dim68Kuo3Init,
                dim69Kuo3Init,
                dim70Kuo3Init,
                dim71Kuo3Init,
                dim72Kuo3Init,
                dim73Kuo3Init,
                dim74Kuo3Init,
                dim75Kuo3Init,
                dim76Kuo3Init,
                dim77Kuo3Init,
                dim78Kuo3Init,
                dim79Kuo3Init,
                dim80Kuo3Init,
                dim81Kuo3Init,
                dim82Kuo3Init,
                dim83Kuo3Init,
                dim84Kuo3Init,
                dim85Kuo3Init,
                dim86Kuo3Init,
                dim87Kuo3Init,
                dim88Kuo3Init,
                dim89Kuo3Init,
                dim90Kuo3Init,
                dim91Kuo3Init,
                dim92Kuo3Init,
                dim93Kuo3Init,
                dim94Kuo3Init,
                dim95Kuo3Init,
                dim96Kuo3Init,
                dim97Kuo3Init,
                dim98Kuo3Init,
                dim99Kuo3Init,
                dim100Kuo3Init,
                dim101Kuo3Init,
                dim102Kuo3Init,
                dim103Kuo3Init,
                dim104Kuo3Init,
                dim105Kuo3Init,
                dim106Kuo3Init,
                dim107Kuo3Init,
                dim108Kuo3Init,
                dim109Kuo3Init,
                dim110Kuo3Init,
                dim111Kuo3Init,
                dim112Kuo3Init,
                dim113Kuo3Init,
                dim114Kuo3Init,
                dim115Kuo3Init,
                dim116Kuo3Init,
                dim117Kuo3Init,
                dim118Kuo3Init,
                dim119Kuo3Init,
                dim120Kuo3Init,
                dim121Kuo3Init,
                dim122Kuo3Init,
                dim123Kuo3Init,
                dim124Kuo3Init,
                dim125Kuo3Init,
                dim126Kuo3Init,
                dim127Kuo3Init,
                dim128Kuo3Init,
                dim129Kuo3Init,
                dim130Kuo3Init,
                dim131Kuo3Init,
                dim132Kuo3Init,
                dim133Kuo3Init,
                dim134Kuo3Init,
                dim135Kuo3Init,
                dim136Kuo3Init,
                dim137Kuo3Init,
                dim138Kuo3Init,
                dim139Kuo3Init,
                dim140Kuo3Init,
                dim141Kuo3Init,
                dim142Kuo3Init,
                dim143Kuo3Init,
                dim144Kuo3Init,
                dim145Kuo3Init,
                dim146Kuo3Init,
                dim147Kuo3Init,
                dim148Kuo3Init,
                dim149Kuo3Init,
                dim150Kuo3Init,
                dim151Kuo3Init,
                dim152Kuo3Init,
                dim153Kuo3Init,
                dim154Kuo3Init,
                dim155Kuo3Init,
                dim156Kuo3Init,
                dim157Kuo3Init,
                dim158Kuo3Init,
                dim159Kuo3Init,
                dim160Kuo3Init,
                dim161Kuo3Init,
                dim162Kuo3Init,
                dim163Kuo3Init,
                dim164Kuo3Init,
                dim165Kuo3Init,
                dim166Kuo3Init,
                dim167Kuo3Init,
                dim168Kuo3Init,
                dim169Kuo3Init,
                dim170Kuo3Init,
                dim171Kuo3Init,
                dim172Kuo3Init,
                dim173Kuo3Init,
                dim174Kuo3Init,
                dim175Kuo3Init,
                dim176Kuo3Init,
                dim177Kuo3Init,
                dim178Kuo3Init,
                dim179Kuo3Init,
                dim180Kuo3Init,
                dim181Kuo3Init,
                dim182Kuo3Init,
                dim183Kuo3Init,
                dim184Kuo3Init,
                dim185Kuo3Init,
                dim186Kuo3Init,
                dim187Kuo3Init,
                dim188Kuo3Init,
                dim189Kuo3Init,
                dim190Kuo3Init,
                dim191Kuo3Init,
                dim192Kuo3Init,
                dim193Kuo3Init,
                dim194Kuo3Init,
                dim195Kuo3Init,
                dim196Kuo3Init,
                dim197Kuo3Init,
                dim198Kuo3Init,
                dim199Kuo3Init,
                dim200Kuo3Init,
                dim201Kuo3Init,
                dim202Kuo3Init,
                dim203Kuo3Init,
                dim204Kuo3Init,
                dim205Kuo3Init,
                dim206Kuo3Init,
                dim207Kuo3Init,
                dim208Kuo3Init,
                dim209Kuo3Init,
                dim210Kuo3Init,
                dim211Kuo3Init,
                dim212Kuo3Init,
                dim213Kuo3Init,
                dim214Kuo3Init,
                dim215Kuo3Init,
                dim216Kuo3Init,
                dim217Kuo3Init,
                dim218Kuo3Init,
                dim219Kuo3Init,
                dim220Kuo3Init,
                dim221Kuo3Init,
                dim222Kuo3Init,
                dim223Kuo3Init,
                dim224Kuo3Init,
                dim225Kuo3Init,
                dim226Kuo3Init,
                dim227Kuo3Init,
                dim228Kuo3Init,
                dim229Kuo3Init,
                dim230Kuo3Init,
                dim231Kuo3Init,
                dim232Kuo3Init,
                dim233Kuo3Init,
                dim234Kuo3Init,
                dim235Kuo3Init,
                dim236Kuo3Init,
                dim237Kuo3Init,
                dim238Kuo3Init,
                dim239Kuo3Init,
                dim240Kuo3Init,
                dim241Kuo3Init,
                dim242Kuo3Init,
                dim243Kuo3Init,
                dim244Kuo3Init,
                dim245Kuo3Init,
                dim246Kuo3Init,
                dim247Kuo3Init,
                dim248Kuo3Init,
                dim249Kuo3Init,
                dim250Kuo3Init,
                dim251Kuo3Init,
                dim252Kuo3Init,
                dim253Kuo3Init,
                dim254Kuo3Init,
                dim255Kuo3Init,
                dim256Kuo3Init,
                dim257Kuo3Init,
                dim258Kuo3Init,
                dim259Kuo3Init,
                dim260Kuo3Init,
                dim261Kuo3Init,
                dim262Kuo3Init,
                dim263Kuo3Init,
                dim264Kuo3Init,
                dim265Kuo3Init,
                dim266Kuo3Init,
                dim267Kuo3Init,
                dim268Kuo3Init,
                dim269Kuo3Init,
                dim270Kuo3Init,
                dim271Kuo3Init,
                dim272Kuo3Init,
                dim273Kuo3Init,
                dim274Kuo3Init,
                dim275Kuo3Init,
                dim276Kuo3Init,
                dim277Kuo3Init,
                dim278Kuo3Init,
                dim279Kuo3Init,
                dim280Kuo3Init,
                dim281Kuo3Init,
                dim282Kuo3Init,
                dim283Kuo3Init,
                dim284Kuo3Init,
                dim285Kuo3Init,
                dim286Kuo3Init,
                dim287Kuo3Init,
                dim288Kuo3Init,
                dim289Kuo3Init,
                dim290Kuo3Init,
                dim291Kuo3Init,
                dim292Kuo3Init,
                dim293Kuo3Init,
                dim294Kuo3Init,
                dim295Kuo3Init,
                dim296Kuo3Init,
                dim297Kuo3Init,
                dim298Kuo3Init,
                dim299Kuo3Init,
                dim300Kuo3Init,
                dim301Kuo3Init,
                dim302Kuo3Init,
                dim303Kuo3Init,
                dim304Kuo3Init,
                dim305Kuo3Init,
                dim306Kuo3Init,
                dim307Kuo3Init,
                dim308Kuo3Init,
                dim309Kuo3Init,
                dim310Kuo3Init,
                dim311Kuo3Init,
                dim312Kuo3Init,
                dim313Kuo3Init,
                dim314Kuo3Init,
                dim315Kuo3Init,
                dim316Kuo3Init,
                dim317Kuo3Init,
                dim318Kuo3Init,
                dim319Kuo3Init,
                dim320Kuo3Init,
                dim321Kuo3Init,
                dim322Kuo3Init,
                dim323Kuo3Init,
                dim324Kuo3Init,
                dim325Kuo3Init,
                dim326Kuo3Init,
                dim327Kuo3Init,
                dim328Kuo3Init,
                dim329Kuo3Init,
                dim330Kuo3Init,
                dim331Kuo3Init,
                dim332Kuo3Init,
                dim333Kuo3Init,
                dim334Kuo3Init,
                dim335Kuo3Init,
                dim336Kuo3Init,
                dim337Kuo3Init,
                dim338Kuo3Init,
                dim339Kuo3Init,
                dim340Kuo3Init,
                dim341Kuo3Init,
                dim342Kuo3Init,
                dim343Kuo3Init,
                dim344Kuo3Init,
                dim345Kuo3Init,
                dim346Kuo3Init,
                dim347Kuo3Init,
                dim348Kuo3Init,
                dim349Kuo3Init,
                dim350Kuo3Init,
                dim351Kuo3Init,
                dim352Kuo3Init,
                dim353Kuo3Init,
                dim354Kuo3Init,
                dim355Kuo3Init,
                dim356Kuo3Init,
                dim357Kuo3Init,
                dim358Kuo3Init,
                dim359Kuo3Init,
                dim360Kuo3Init,
                dim361Kuo3Init,
                dim362Kuo3Init,
                dim363Kuo3Init,
                dim364Kuo3Init,
                dim365Kuo3Init,
                dim366Kuo3Init,
                dim367Kuo3Init,
                dim368Kuo3Init,
                dim369Kuo3Init,
                dim370Kuo3Init,
                dim371Kuo3Init,
                dim372Kuo3Init,
                dim373Kuo3Init,
                dim374Kuo3Init,
                dim375Kuo3Init,
                dim376Kuo3Init,
                dim377Kuo3Init,
                dim378Kuo3Init,
                dim379Kuo3Init,
                dim380Kuo3Init,
                dim381Kuo3Init,
                dim382Kuo3Init,
                dim383Kuo3Init,
                dim384Kuo3Init,
                dim385Kuo3Init,
                dim386Kuo3Init,
                dim387Kuo3Init,
                dim388Kuo3Init,
                dim389Kuo3Init,
                dim390Kuo3Init,
                dim391Kuo3Init,
                dim392Kuo3Init,
                dim393Kuo3Init,
                dim394Kuo3Init,
                dim395Kuo3Init,
                dim396Kuo3Init,
                dim397Kuo3Init,
                dim398Kuo3Init,
                dim399Kuo3Init,
                dim400Kuo3Init,
                dim401Kuo3Init,
                dim402Kuo3Init,
                dim403Kuo3Init,
                dim404Kuo3Init,
                dim405Kuo3Init,
                dim406Kuo3Init,
                dim407Kuo3Init,
                dim408Kuo3Init,
                dim409Kuo3Init,
                dim410Kuo3Init,
                dim411Kuo3Init,
                dim412Kuo3Init,
                dim413Kuo3Init,
                dim414Kuo3Init,
                dim415Kuo3Init,
                dim416Kuo3Init,
                dim417Kuo3Init,
                dim418Kuo3Init,
                dim419Kuo3Init,
                dim420Kuo3Init,
                dim421Kuo3Init,
                dim422Kuo3Init,
                dim423Kuo3Init,
                dim424Kuo3Init,
                dim425Kuo3Init,
                dim426Kuo3Init,
                dim427Kuo3Init,
                dim428Kuo3Init,
                dim429Kuo3Init,
                dim430Kuo3Init,
                dim431Kuo3Init,
                dim432Kuo3Init,
                dim433Kuo3Init,
                dim434Kuo3Init,
                dim435Kuo3Init,
                dim436Kuo3Init,
                dim437Kuo3Init,
                dim438Kuo3Init,
                dim439Kuo3Init,
                dim440Kuo3Init,
                dim441Kuo3Init,
                dim442Kuo3Init,
                dim443Kuo3Init,
                dim444Kuo3Init,
                dim445Kuo3Init,
                dim446Kuo3Init,
                dim447Kuo3Init,
                dim448Kuo3Init,
                dim449Kuo3Init,
                dim450Kuo3Init,
                dim451Kuo3Init,
                dim452Kuo3Init,
                dim453Kuo3Init,
                dim454Kuo3Init,
                dim455Kuo3Init,
                dim456Kuo3Init,
                dim457Kuo3Init,
                dim458Kuo3Init,
                dim459Kuo3Init,
                dim460Kuo3Init,
                dim461Kuo3Init,
                dim462Kuo3Init,
                dim463Kuo3Init,
                dim464Kuo3Init,
                dim465Kuo3Init,
                dim466Kuo3Init,
                dim467Kuo3Init,
                dim468Kuo3Init,
                dim469Kuo3Init,
                dim470Kuo3Init,
                dim471Kuo3Init,
                dim472Kuo3Init,
                dim473Kuo3Init,
                dim474Kuo3Init,
                dim475Kuo3Init,
                dim476Kuo3Init,
                dim477Kuo3Init,
                dim478Kuo3Init,
                dim479Kuo3Init,
                dim480Kuo3Init,
                dim481Kuo3Init,
                dim482Kuo3Init,
                dim483Kuo3Init,
                dim484Kuo3Init,
                dim485Kuo3Init,
                dim486Kuo3Init,
                dim487Kuo3Init,
                dim488Kuo3Init,
                dim489Kuo3Init,
                dim490Kuo3Init,
                dim491Kuo3Init,
                dim492Kuo3Init,
                dim493Kuo3Init,
                dim494Kuo3Init,
                dim495Kuo3Init,
                dim496Kuo3Init,
                dim497Kuo3Init,
                dim498Kuo3Init,
                dim499Kuo3Init,
                dim500Kuo3Init,
                dim501Kuo3Init,
                dim502Kuo3Init,
                dim503Kuo3Init,
                dim504Kuo3Init,
                dim505Kuo3Init,
                dim506Kuo3Init,
                dim507Kuo3Init,
                dim508Kuo3Init,
                dim509Kuo3Init,
                dim510Kuo3Init,
                dim511Kuo3Init,
                dim512Kuo3Init,
                dim513Kuo3Init,
                dim514Kuo3Init,
                dim515Kuo3Init,
                dim516Kuo3Init,
                dim517Kuo3Init,
                dim518Kuo3Init,
                dim519Kuo3Init,
                dim520Kuo3Init,
                dim521Kuo3Init,
                dim522Kuo3Init,
                dim523Kuo3Init,
                dim524Kuo3Init,
                dim525Kuo3Init,
                dim526Kuo3Init,
                dim527Kuo3Init,
                dim528Kuo3Init,
                dim529Kuo3Init,
                dim530Kuo3Init,
                dim531Kuo3Init,
                dim532Kuo3Init,
                dim533Kuo3Init,
                dim534Kuo3Init,
                dim535Kuo3Init,
                dim536Kuo3Init,
                dim537Kuo3Init,
                dim538Kuo3Init,
                dim539Kuo3Init,
                dim540Kuo3Init,
                dim541Kuo3Init,
                dim542Kuo3Init,
                dim543Kuo3Init,
                dim544Kuo3Init,
                dim545Kuo3Init,
                dim546Kuo3Init,
                dim547Kuo3Init,
                dim548Kuo3Init,
                dim549Kuo3Init,
                dim550Kuo3Init,
                dim551Kuo3Init,
                dim552Kuo3Init,
                dim553Kuo3Init,
                dim554Kuo3Init,
                dim555Kuo3Init,
                dim556Kuo3Init,
                dim557Kuo3Init,
                dim558Kuo3Init,
                dim559Kuo3Init,
                dim560Kuo3Init,
                dim561Kuo3Init,
                dim562Kuo3Init,
                dim563Kuo3Init,
                dim564Kuo3Init,
                dim565Kuo3Init,
                dim566Kuo3Init,
                dim567Kuo3Init,
                dim568Kuo3Init,
                dim569Kuo3Init,
                dim570Kuo3Init,
                dim571Kuo3Init,
                dim572Kuo3Init,
                dim573Kuo3Init,
                dim574Kuo3Init,
                dim575Kuo3Init,
                dim576Kuo3Init,
                dim577Kuo3Init,
                dim578Kuo3Init,
                dim579Kuo3Init,
                dim580Kuo3Init,
                dim581Kuo3Init,
                dim582Kuo3Init,
                dim583Kuo3Init,
                dim584Kuo3Init,
                dim585Kuo3Init,
                dim586Kuo3Init,
                dim587Kuo3Init,
                dim588Kuo3Init,
                dim589Kuo3Init,
                dim590Kuo3Init,
                dim591Kuo3Init,
                dim592Kuo3Init,
                dim593Kuo3Init,
                dim594Kuo3Init,
                dim595Kuo3Init,
                dim596Kuo3Init,
                dim597Kuo3Init,
                dim598Kuo3Init,
                dim599Kuo3Init,
                dim600Kuo3Init,
                dim601Kuo3Init,
                dim602Kuo3Init,
                dim603Kuo3Init,
                dim604Kuo3Init,
                dim605Kuo3Init,
                dim606Kuo3Init,
                dim607Kuo3Init,
                dim608Kuo3Init,
                dim609Kuo3Init,
                dim610Kuo3Init,
                dim611Kuo3Init,
                dim612Kuo3Init,
                dim613Kuo3Init,
                dim614Kuo3Init,
                dim615Kuo3Init,
                dim616Kuo3Init,
                dim617Kuo3Init,
                dim618Kuo3Init,
                dim619Kuo3Init,
                dim620Kuo3Init,
                dim621Kuo3Init,
                dim622Kuo3Init,
                dim623Kuo3Init,
                dim624Kuo3Init,
                dim625Kuo3Init,
                dim626Kuo3Init,
                dim627Kuo3Init,
                dim628Kuo3Init,
                dim629Kuo3Init,
                dim630Kuo3Init,
                dim631Kuo3Init,
                dim632Kuo3Init,
                dim633Kuo3Init,
                dim634Kuo3Init,
                dim635Kuo3Init,
                dim636Kuo3Init,
                dim637Kuo3Init,
                dim638Kuo3Init,
                dim639Kuo3Init,
                dim640Kuo3Init,
                dim641Kuo3Init,
                dim642Kuo3Init,
                dim643Kuo3Init,
                dim644Kuo3Init,
                dim645Kuo3Init,
                dim646Kuo3Init,
                dim647Kuo3Init,
                dim648Kuo3Init,
                dim649Kuo3Init,
                dim650Kuo3Init,
                dim651Kuo3Init,
                dim652Kuo3Init,
                dim653Kuo3Init,
                dim654Kuo3Init,
                dim655Kuo3Init,
                dim656Kuo3Init,
                dim657Kuo3Init,
                dim658Kuo3Init,
                dim659Kuo3Init,
                dim660Kuo3Init,
                dim661Kuo3Init,
                dim662Kuo3Init,
                dim663Kuo3Init,
                dim664Kuo3Init,
                dim665Kuo3Init,
                dim666Kuo3Init,
                dim667Kuo3Init,
                dim668Kuo3Init,
                dim669Kuo3Init,
                dim670Kuo3Init,
                dim671Kuo3Init,
                dim672Kuo3Init,
                dim673Kuo3Init,
                dim674Kuo3Init,
                dim675Kuo3Init,
                dim676Kuo3Init,
                dim677Kuo3Init,
                dim678Kuo3Init,
                dim679Kuo3Init,
                dim680Kuo3Init,
                dim681Kuo3Init,
                dim682Kuo3Init,
                dim683Kuo3Init,
                dim684Kuo3Init,
                dim685Kuo3Init,
                dim686Kuo3Init,
                dim687Kuo3Init,
                dim688Kuo3Init,
                dim689Kuo3Init,
                dim690Kuo3Init,
                dim691Kuo3Init,
                dim692Kuo3Init,
                dim693Kuo3Init,
                dim694Kuo3Init,
                dim695Kuo3Init,
                dim696Kuo3Init,
                dim697Kuo3Init,
                dim698Kuo3Init,
                dim699Kuo3Init,
                dim700Kuo3Init,
                dim701Kuo3Init,
                dim702Kuo3Init,
                dim703Kuo3Init,
                dim704Kuo3Init,
                dim705Kuo3Init,
                dim706Kuo3Init,
                dim707Kuo3Init,
                dim708Kuo3Init,
                dim709Kuo3Init,
                dim710Kuo3Init,
                dim711Kuo3Init,
                dim712Kuo3Init,
                dim713Kuo3Init,
                dim714Kuo3Init,
                dim715Kuo3Init,
                dim716Kuo3Init,
                dim717Kuo3Init,
                dim718Kuo3Init,
                dim719Kuo3Init,
                dim720Kuo3Init,
                dim721Kuo3Init,
                dim722Kuo3Init,
                dim723Kuo3Init,
                dim724Kuo3Init,
                dim725Kuo3Init,
                dim726Kuo3Init,
                dim727Kuo3Init,
                dim728Kuo3Init,
                dim729Kuo3Init,
                dim730Kuo3Init,
                dim731Kuo3Init,
                dim732Kuo3Init,
                dim733Kuo3Init,
                dim734Kuo3Init,
                dim735Kuo3Init,
                dim736Kuo3Init,
                dim737Kuo3Init,
                dim738Kuo3Init,
                dim739Kuo3Init,
                dim740Kuo3Init,
                dim741Kuo3Init,
                dim742Kuo3Init,
                dim743Kuo3Init,
                dim744Kuo3Init,
                dim745Kuo3Init,
                dim746Kuo3Init,
                dim747Kuo3Init,
                dim748Kuo3Init,
                dim749Kuo3Init,
                dim750Kuo3Init,
                dim751Kuo3Init,
                dim752Kuo3Init,
                dim753Kuo3Init,
                dim754Kuo3Init,
                dim755Kuo3Init,
                dim756Kuo3Init,
                dim757Kuo3Init,
                dim758Kuo3Init,
                dim759Kuo3Init,
                dim760Kuo3Init,
                dim761Kuo3Init,
                dim762Kuo3Init,
                dim763Kuo3Init,
                dim764Kuo3Init,
                dim765Kuo3Init,
                dim766Kuo3Init,
                dim767Kuo3Init,
                dim768Kuo3Init,
                dim769Kuo3Init,
                dim770Kuo3Init,
                dim771Kuo3Init,
                dim772Kuo3Init,
                dim773Kuo3Init,
                dim774Kuo3Init,
                dim775Kuo3Init,
                dim776Kuo3Init,
                dim777Kuo3Init,
                dim778Kuo3Init,
                dim779Kuo3Init,
                dim780Kuo3Init,
                dim781Kuo3Init,
                dim782Kuo3Init,
                dim783Kuo3Init,
                dim784Kuo3Init,
                dim785Kuo3Init,
                dim786Kuo3Init,
                dim787Kuo3Init,
                dim788Kuo3Init,
                dim789Kuo3Init,
                dim790Kuo3Init,
                dim791Kuo3Init,
                dim792Kuo3Init,
                dim793Kuo3Init,
                dim794Kuo3Init,
                dim795Kuo3Init,
                dim796Kuo3Init,
                dim797Kuo3Init,
                dim798Kuo3Init,
                dim799Kuo3Init,
                dim800Kuo3Init,
                dim801Kuo3Init,
                dim802Kuo3Init,
                dim803Kuo3Init,
                dim804Kuo3Init,
                dim805Kuo3Init,
                dim806Kuo3Init,
                dim807Kuo3Init,
                dim808Kuo3Init,
                dim809Kuo3Init,
                dim810Kuo3Init,
                dim811Kuo3Init,
                dim812Kuo3Init,
                dim813Kuo3Init,
                dim814Kuo3Init,
                dim815Kuo3Init,
                dim816Kuo3Init,
                dim817Kuo3Init,
                dim818Kuo3Init,
                dim819Kuo3Init,
                dim820Kuo3Init,
                dim821Kuo3Init,
                dim822Kuo3Init,
                dim823Kuo3Init,
                dim824Kuo3Init,
                dim825Kuo3Init,
                dim826Kuo3Init,
                dim827Kuo3Init,
                dim828Kuo3Init,
                dim829Kuo3Init,
                dim830Kuo3Init,
                dim831Kuo3Init,
                dim832Kuo3Init,
                dim833Kuo3Init,
                dim834Kuo3Init,
                dim835Kuo3Init,
                dim836Kuo3Init,
                dim837Kuo3Init,
                dim838Kuo3Init,
                dim839Kuo3Init,
                dim840Kuo3Init,
                dim841Kuo3Init,
                dim842Kuo3Init,
                dim843Kuo3Init,
                dim844Kuo3Init,
                dim845Kuo3Init,
                dim846Kuo3Init,
                dim847Kuo3Init,
                dim848Kuo3Init,
                dim849Kuo3Init,
                dim850Kuo3Init,
                dim851Kuo3Init,
                dim852Kuo3Init,
                dim853Kuo3Init,
                dim854Kuo3Init,
                dim855Kuo3Init,
                dim856Kuo3Init,
                dim857Kuo3Init,
                dim858Kuo3Init,
                dim859Kuo3Init,
                dim860Kuo3Init,
                dim861Kuo3Init,
                dim862Kuo3Init,
                dim863Kuo3Init,
                dim864Kuo3Init,
                dim865Kuo3Init,
                dim866Kuo3Init,
                dim867Kuo3Init,
                dim868Kuo3Init,
                dim869Kuo3Init,
                dim870Kuo3Init,
                dim871Kuo3Init,
                dim872Kuo3Init,
                dim873Kuo3Init,
                dim874Kuo3Init,
                dim875Kuo3Init,
                dim876Kuo3Init,
                dim877Kuo3Init,
                dim878Kuo3Init,
                dim879Kuo3Init,
                dim880Kuo3Init,
                dim881Kuo3Init,
                dim882Kuo3Init,
                dim883Kuo3Init,
                dim884Kuo3Init,
                dim885Kuo3Init,
                dim886Kuo3Init,
                dim887Kuo3Init,
                dim888Kuo3Init,
                dim889Kuo3Init,
                dim890Kuo3Init,
                dim891Kuo3Init,
                dim892Kuo3Init,
                dim893Kuo3Init,
                dim894Kuo3Init,
                dim895Kuo3Init,
                dim896Kuo3Init,
                dim897Kuo3Init,
                dim898Kuo3Init,
                dim899Kuo3Init,
                dim900Kuo3Init,
                dim901Kuo3Init,
                dim902Kuo3Init,
                dim903Kuo3Init,
                dim904Kuo3Init,
                dim905Kuo3Init,
                dim906Kuo3Init,
                dim907Kuo3Init,
                dim908Kuo3Init,
                dim909Kuo3Init,
                dim910Kuo3Init,
                dim911Kuo3Init,
                dim912Kuo3Init,
                dim913Kuo3Init,
                dim914Kuo3Init,
                dim915Kuo3Init,
                dim916Kuo3Init,
                dim917Kuo3Init,
                dim918Kuo3Init,
                dim919Kuo3Init,
                dim920Kuo3Init,
                dim921Kuo3Init,
                dim922Kuo3Init,
                dim923Kuo3Init,
                dim924Kuo3Init,
                dim925Kuo3Init,
                dim926Kuo3Init,
                dim927Kuo3Init,
                dim928Kuo3Init,
                dim929Kuo3Init,
                dim930Kuo3Init,
                dim931Kuo3Init,
                dim932Kuo3Init,
                dim933Kuo3Init,
                dim934Kuo3Init,
                dim935Kuo3Init,
                dim936Kuo3Init,
                dim937Kuo3Init,
                dim938Kuo3Init,
                dim939Kuo3Init,
                dim940Kuo3Init,
                dim941Kuo3Init,
                dim942Kuo3Init,
                dim943Kuo3Init,
                dim944Kuo3Init,
                dim945Kuo3Init,
                dim946Kuo3Init,
                dim947Kuo3Init,
                dim948Kuo3Init,
                dim949Kuo3Init,
                dim950Kuo3Init,
                dim951Kuo3Init,
                dim952Kuo3Init,
                dim953Kuo3Init,
                dim954Kuo3Init,
                dim955Kuo3Init,
                dim956Kuo3Init,
                dim957Kuo3Init,
                dim958Kuo3Init,
                dim959Kuo3Init,
                dim960Kuo3Init,
                dim961Kuo3Init,
                dim962Kuo3Init,
                dim963Kuo3Init,
                dim964Kuo3Init,
                dim965Kuo3Init,
                dim966Kuo3Init,
                dim967Kuo3Init,
                dim968Kuo3Init,
                dim969Kuo3Init,
                dim970Kuo3Init,
                dim971Kuo3Init,
                dim972Kuo3Init,
                dim973Kuo3Init,
                dim974Kuo3Init,
                dim975Kuo3Init,
                dim976Kuo3Init,
                dim977Kuo3Init,
                dim978Kuo3Init,
                dim979Kuo3Init,
                dim980Kuo3Init,
                dim981Kuo3Init,
                dim982Kuo3Init,
                dim983Kuo3Init,
                dim984Kuo3Init,
                dim985Kuo3Init,
                dim986Kuo3Init,
                dim987Kuo3Init,
                dim988Kuo3Init,
                dim989Kuo3Init,
                dim990Kuo3Init,
                dim991Kuo3Init,
                dim992Kuo3Init,
                dim993Kuo3Init,
                dim994Kuo3Init,
                dim995Kuo3Init,
                dim996Kuo3Init,
                dim997Kuo3Init,
                dim998Kuo3Init,
                dim999Kuo3Init,
                dim1000Kuo3Init,
                dim1001Kuo3Init,
                dim1002Kuo3Init,
                dim1003Kuo3Init,
                dim1004Kuo3Init,
                dim1005Kuo3Init,
                dim1006Kuo3Init,
                dim1007Kuo3Init,
                dim1008Kuo3Init,
                dim1009Kuo3Init,
                dim1010Kuo3Init,
                dim1011Kuo3Init,
                dim1012Kuo3Init,
                dim1013Kuo3Init,
                dim1014Kuo3Init,
                dim1015Kuo3Init,
                dim1016Kuo3Init,
                dim1017Kuo3Init,
                dim1018Kuo3Init,
                dim1019Kuo3Init,
                dim1020Kuo3Init,
                dim1021Kuo3Init,
                dim1022Kuo3Init,
                dim1023Kuo3Init,
                dim1024Kuo3Init,
                dim1025Kuo3Init,
                dim1026Kuo3Init,
                dim1027Kuo3Init,
                dim1028Kuo3Init,
                dim1029Kuo3Init,
                dim1030Kuo3Init,
                dim1031Kuo3Init,
                dim1032Kuo3Init,
                dim1033Kuo3Init,
                dim1034Kuo3Init,
                dim1035Kuo3Init,
                dim1036Kuo3Init,
                dim1037Kuo3Init,
                dim1038Kuo3Init,
                dim1039Kuo3Init,
                dim1040Kuo3Init,
                dim1041Kuo3Init,
                dim1042Kuo3Init,
                dim1043Kuo3Init,
                dim1044Kuo3Init,
                dim1045Kuo3Init,
                dim1046Kuo3Init,
                dim1047Kuo3Init,
                dim1048Kuo3Init,
                dim1049Kuo3Init,
                dim1050Kuo3Init,
                dim1051Kuo3Init,
                dim1052Kuo3Init,
                dim1053Kuo3Init,
                dim1054Kuo3Init,
                dim1055Kuo3Init,
                dim1056Kuo3Init,
                dim1057Kuo3Init,
                dim1058Kuo3Init,
                dim1059Kuo3Init,
                dim1060Kuo3Init,
                dim1061Kuo3Init,
                dim1062Kuo3Init,
                dim1063Kuo3Init,
                dim1064Kuo3Init,
                dim1065Kuo3Init,
                dim1066Kuo3Init,
                dim1067Kuo3Init,
                dim1068Kuo3Init,
                dim1069Kuo3Init,
                dim1070Kuo3Init,
                dim1071Kuo3Init,
                dim1072Kuo3Init,
                dim1073Kuo3Init,
                dim1074Kuo3Init,
                dim1075Kuo3Init,
                dim1076Kuo3Init,
                dim1077Kuo3Init,
                dim1078Kuo3Init,
                dim1079Kuo3Init,
                dim1080Kuo3Init,
                dim1081Kuo3Init,
                dim1082Kuo3Init,
                dim1083Kuo3Init,
                dim1084Kuo3Init,
                dim1085Kuo3Init,
                dim1086Kuo3Init,
                dim1087Kuo3Init,
                dim1088Kuo3Init,
                dim1089Kuo3Init,
                dim1090Kuo3Init,
                dim1091Kuo3Init,
                dim1092Kuo3Init,
                dim1093Kuo3Init,
                dim1094Kuo3Init,
                dim1095Kuo3Init,
                dim1096Kuo3Init,
                dim1097Kuo3Init,
                dim1098Kuo3Init,
                dim1099Kuo3Init,
                dim1100Kuo3Init,
                dim1101Kuo3Init,
                dim1102Kuo3Init,
                dim1103Kuo3Init,
                dim1104Kuo3Init,
                dim1105Kuo3Init,
                dim1106Kuo3Init,
                dim1107Kuo3Init,
                dim1108Kuo3Init,
                dim1109Kuo3Init,
                dim1110Kuo3Init,
                dim1111Kuo3Init,
                dim1112Kuo3Init,
                dim1113Kuo3Init,
                dim1114Kuo3Init,
                dim1115Kuo3Init,
                dim1116Kuo3Init,
                dim1117Kuo3Init,
                dim1118Kuo3Init,
                dim1119Kuo3Init,
                dim1120Kuo3Init,
                dim1121Kuo3Init,
                dim1122Kuo3Init,
                dim1123Kuo3Init,
                dim1124Kuo3Init,
                dim1125Kuo3Init,
                dim1126Kuo3Init,
                dim1127Kuo3Init,
                dim1128Kuo3Init,
                dim1129Kuo3Init,
                dim1130Kuo3Init,
                dim1131Kuo3Init,
                dim1132Kuo3Init,
                dim1133Kuo3Init,
                dim1134Kuo3Init,
                dim1135Kuo3Init,
                dim1136Kuo3Init,
                dim1137Kuo3Init,
                dim1138Kuo3Init,
                dim1139Kuo3Init,
                dim1140Kuo3Init,
                dim1141Kuo3Init,
                dim1142Kuo3Init,
                dim1143Kuo3Init,
                dim1144Kuo3Init,
                dim1145Kuo3Init,
                dim1146Kuo3Init,
                dim1147Kuo3Init,
                dim1148Kuo3Init,
                dim1149Kuo3Init,
                dim1150Kuo3Init,
                dim1151Kuo3Init,
                dim1152Kuo3Init,
                dim1153Kuo3Init,
                dim1154Kuo3Init,
                dim1155Kuo3Init,
                dim1156Kuo3Init,
                dim1157Kuo3Init,
                dim1158Kuo3Init,
                dim1159Kuo3Init,
                dim1160Kuo3Init,
                dim1161Kuo3Init,
                dim1162Kuo3Init,
                dim1163Kuo3Init,
                dim1164Kuo3Init,
                dim1165Kuo3Init,
                dim1166Kuo3Init,
                dim1167Kuo3Init,
                dim1168Kuo3Init,
                dim1169Kuo3Init,
                dim1170Kuo3Init,
                dim1171Kuo3Init,
                dim1172Kuo3Init,
                dim1173Kuo3Init,
                dim1174Kuo3Init,
                dim1175Kuo3Init,
                dim1176Kuo3Init,
                dim1177Kuo3Init,
                dim1178Kuo3Init,
                dim1179Kuo3Init,
                dim1180Kuo3Init,
                dim1181Kuo3Init,
                dim1182Kuo3Init,
                dim1183Kuo3Init,
                dim1184Kuo3Init,
                dim1185Kuo3Init,
                dim1186Kuo3Init,
                dim1187Kuo3Init,
                dim1188Kuo3Init,
                dim1189Kuo3Init,
                dim1190Kuo3Init,
                dim1191Kuo3Init,
                dim1192Kuo3Init,
                dim1193Kuo3Init,
                dim1194Kuo3Init,
                dim1195Kuo3Init,
                dim1196Kuo3Init,
                dim1197Kuo3Init,
                dim1198Kuo3Init,
                dim1199Kuo3Init,
                dim1200Kuo3Init,
                dim1201Kuo3Init,
                dim1202Kuo3Init,
                dim1203Kuo3Init,
                dim1204Kuo3Init,
                dim1205Kuo3Init,
                dim1206Kuo3Init,
                dim1207Kuo3Init,
                dim1208Kuo3Init,
                dim1209Kuo3Init,
                dim1210Kuo3Init,
                dim1211Kuo3Init,
                dim1212Kuo3Init,
                dim1213Kuo3Init,
                dim1214Kuo3Init,
                dim1215Kuo3Init,
                dim1216Kuo3Init,
                dim1217Kuo3Init,
                dim1218Kuo3Init,
                dim1219Kuo3Init,
                dim1220Kuo3Init,
                dim1221Kuo3Init,
                dim1222Kuo3Init,
                dim1223Kuo3Init,
                dim1224Kuo3Init,
                dim1225Kuo3Init,
                dim1226Kuo3Init,
                dim1227Kuo3Init,
                dim1228Kuo3Init,
                dim1229Kuo3Init,
                dim1230Kuo3Init,
                dim1231Kuo3Init,
                dim1232Kuo3Init,
                dim1233Kuo3Init,
                dim1234Kuo3Init,
                dim1235Kuo3Init,
                dim1236Kuo3Init,
                dim1237Kuo3Init,
                dim1238Kuo3Init,
                dim1239Kuo3Init,
                dim1240Kuo3Init,
                dim1241Kuo3Init,
                dim1242Kuo3Init,
                dim1243Kuo3Init,
                dim1244Kuo3Init,
                dim1245Kuo3Init,
                dim1246Kuo3Init,
                dim1247Kuo3Init,
                dim1248Kuo3Init,
                dim1249Kuo3Init,
                dim1250Kuo3Init,
                dim1251Kuo3Init,
                dim1252Kuo3Init,
                dim1253Kuo3Init,
                dim1254Kuo3Init,
                dim1255Kuo3Init,
                dim1256Kuo3Init,
                dim1257Kuo3Init,
                dim1258Kuo3Init,
                dim1259Kuo3Init,
                dim1260Kuo3Init,
                dim1261Kuo3Init,
                dim1262Kuo3Init,
                dim1263Kuo3Init,
                dim1264Kuo3Init,
                dim1265Kuo3Init,
                dim1266Kuo3Init,
                dim1267Kuo3Init,
                dim1268Kuo3Init,
                dim1269Kuo3Init,
                dim1270Kuo3Init,
                dim1271Kuo3Init,
                dim1272Kuo3Init,
                dim1273Kuo3Init,
                dim1274Kuo3Init,
                dim1275Kuo3Init,
                dim1276Kuo3Init,
                dim1277Kuo3Init,
                dim1278Kuo3Init,
                dim1279Kuo3Init,
                dim1280Kuo3Init,
                dim1281Kuo3Init,
                dim1282Kuo3Init,
                dim1283Kuo3Init,
                dim1284Kuo3Init,
                dim1285Kuo3Init,
                dim1286Kuo3Init,
                dim1287Kuo3Init,
                dim1288Kuo3Init,
                dim1289Kuo3Init,
                dim1290Kuo3Init,
                dim1291Kuo3Init,
                dim1292Kuo3Init,
                dim1293Kuo3Init,
                dim1294Kuo3Init,
                dim1295Kuo3Init,
                dim1296Kuo3Init,
                dim1297Kuo3Init,
                dim1298Kuo3Init,
                dim1299Kuo3Init,
                dim1300Kuo3Init,
                dim1301Kuo3Init,
                dim1302Kuo3Init,
                dim1303Kuo3Init,
                dim1304Kuo3Init,
                dim1305Kuo3Init,
                dim1306Kuo3Init,
                dim1307Kuo3Init,
                dim1308Kuo3Init,
                dim1309Kuo3Init,
                dim1310Kuo3Init,
                dim1311Kuo3Init,
                dim1312Kuo3Init,
                dim1313Kuo3Init,
                dim1314Kuo3Init,
                dim1315Kuo3Init,
                dim1316Kuo3Init,
                dim1317Kuo3Init,
                dim1318Kuo3Init,
                dim1319Kuo3Init,
                dim1320Kuo3Init,
                dim1321Kuo3Init,
                dim1322Kuo3Init,
                dim1323Kuo3Init,
                dim1324Kuo3Init,
                dim1325Kuo3Init,
                dim1326Kuo3Init,
                dim1327Kuo3Init,
                dim1328Kuo3Init,
                dim1329Kuo3Init,
                dim1330Kuo3Init,
                dim1331Kuo3Init,
                dim1332Kuo3Init,
                dim1333Kuo3Init,
                dim1334Kuo3Init,
                dim1335Kuo3Init,
                dim1336Kuo3Init,
                dim1337Kuo3Init,
                dim1338Kuo3Init,
                dim1339Kuo3Init,
                dim1340Kuo3Init,
                dim1341Kuo3Init,
                dim1342Kuo3Init,
                dim1343Kuo3Init,
                dim1344Kuo3Init,
                dim1345Kuo3Init,
                dim1346Kuo3Init,
                dim1347Kuo3Init,
                dim1348Kuo3Init,
                dim1349Kuo3Init,
                dim1350Kuo3Init,
                dim1351Kuo3Init,
                dim1352Kuo3Init,
                dim1353Kuo3Init,
                dim1354Kuo3Init,
                dim1355Kuo3Init,
                dim1356Kuo3Init,
                dim1357Kuo3Init,
                dim1358Kuo3Init,
                dim1359Kuo3Init,
                dim1360Kuo3Init,
                dim1361Kuo3Init,
                dim1362Kuo3Init,
                dim1363Kuo3Init,
                dim1364Kuo3Init,
                dim1365Kuo3Init,
                dim1366Kuo3Init,
                dim1367Kuo3Init,
                dim1368Kuo3Init,
                dim1369Kuo3Init,
                dim1370Kuo3Init,
                dim1371Kuo3Init,
                dim1372Kuo3Init,
                dim1373Kuo3Init,
                dim1374Kuo3Init,
                dim1375Kuo3Init,
                dim1376Kuo3Init,
                dim1377Kuo3Init,
                dim1378Kuo3Init,
                dim1379Kuo3Init,
                dim1380Kuo3Init,
                dim1381Kuo3Init,
                dim1382Kuo3Init,
                dim1383Kuo3Init,
                dim1384Kuo3Init,
                dim1385Kuo3Init,
                dim1386Kuo3Init,
                dim1387Kuo3Init,
                dim1388Kuo3Init,
                dim1389Kuo3Init,
                dim1390Kuo3Init,
                dim1391Kuo3Init,
                dim1392Kuo3Init,
                dim1393Kuo3Init,
                dim1394Kuo3Init,
                dim1395Kuo3Init,
                dim1396Kuo3Init,
                dim1397Kuo3Init,
                dim1398Kuo3Init,
                dim1399Kuo3Init,
                dim1400Kuo3Init,
                dim1401Kuo3Init,
                dim1402Kuo3Init,
                dim1403Kuo3Init,
                dim1404Kuo3Init,
                dim1405Kuo3Init,
                dim1406Kuo3Init,
                dim1407Kuo3Init,
                dim1408Kuo3Init,
                dim1409Kuo3Init,
                dim1410Kuo3Init,
                dim1411Kuo3Init,
                dim1412Kuo3Init,
                dim1413Kuo3Init,
                dim1414Kuo3Init,
                dim1415Kuo3Init,
                dim1416Kuo3Init,
                dim1417Kuo3Init,
                dim1418Kuo3Init,
                dim1419Kuo3Init,
                dim1420Kuo3Init,
                dim1421Kuo3Init,
                dim1422Kuo3Init,
                dim1423Kuo3Init,
                dim1424Kuo3Init,
                dim1425Kuo3Init,
                dim1426Kuo3Init,
                dim1427Kuo3Init,
                dim1428Kuo3Init,
                dim1429Kuo3Init,
                dim1430Kuo3Init,
                dim1431Kuo3Init,
                dim1432Kuo3Init,
                dim1433Kuo3Init,
                dim1434Kuo3Init,
                dim1435Kuo3Init,
                dim1436Kuo3Init,
                dim1437Kuo3Init,
                dim1438Kuo3Init,
                dim1439Kuo3Init,
                dim1440Kuo3Init,
                dim1441Kuo3Init,
                dim1442Kuo3Init,
                dim1443Kuo3Init,
                dim1444Kuo3Init,
                dim1445Kuo3Init,
                dim1446Kuo3Init,
                dim1447Kuo3Init,
                dim1448Kuo3Init,
                dim1449Kuo3Init,
                dim1450Kuo3Init,
                dim1451Kuo3Init,
                dim1452Kuo3Init,
                dim1453Kuo3Init,
                dim1454Kuo3Init,
                dim1455Kuo3Init,
                dim1456Kuo3Init,
                dim1457Kuo3Init,
                dim1458Kuo3Init,
                dim1459Kuo3Init,
                dim1460Kuo3Init,
                dim1461Kuo3Init,
                dim1462Kuo3Init,
                dim1463Kuo3Init,
                dim1464Kuo3Init,
                dim1465Kuo3Init,
                dim1466Kuo3Init,
                dim1467Kuo3Init,
                dim1468Kuo3Init,
                dim1469Kuo3Init,
                dim1470Kuo3Init,
                dim1471Kuo3Init,
                dim1472Kuo3Init,
                dim1473Kuo3Init,
                dim1474Kuo3Init,
                dim1475Kuo3Init,
                dim1476Kuo3Init,
                dim1477Kuo3Init,
                dim1478Kuo3Init,
                dim1479Kuo3Init,
                dim1480Kuo3Init,
                dim1481Kuo3Init,
                dim1482Kuo3Init,
                dim1483Kuo3Init,
                dim1484Kuo3Init,
                dim1485Kuo3Init,
                dim1486Kuo3Init,
                dim1487Kuo3Init,
                dim1488Kuo3Init,
                dim1489Kuo3Init,
                dim1490Kuo3Init,
                dim1491Kuo3Init,
                dim1492Kuo3Init,
                dim1493Kuo3Init,
                dim1494Kuo3Init,
                dim1495Kuo3Init,
                dim1496Kuo3Init,
                dim1497Kuo3Init,
                dim1498Kuo3Init,
                dim1499Kuo3Init,
                dim1500Kuo3Init,
                dim1501Kuo3Init,
                dim1502Kuo3Init,
                dim1503Kuo3Init,
                dim1504Kuo3Init,
                dim1505Kuo3Init,
                dim1506Kuo3Init,
                dim1507Kuo3Init,
                dim1508Kuo3Init,
                dim1509Kuo3Init,
                dim1510Kuo3Init,
                dim1511Kuo3Init,
                dim1512Kuo3Init,
                dim1513Kuo3Init,
                dim1514Kuo3Init,
                dim1515Kuo3Init,
                dim1516Kuo3Init,
                dim1517Kuo3Init,
                dim1518Kuo3Init,
                dim1519Kuo3Init,
                dim1520Kuo3Init,
                dim1521Kuo3Init,
                dim1522Kuo3Init,
                dim1523Kuo3Init,
                dim1524Kuo3Init,
                dim1525Kuo3Init,
                dim1526Kuo3Init,
                dim1527Kuo3Init,
                dim1528Kuo3Init,
                dim1529Kuo3Init,
                dim1530Kuo3Init,
                dim1531Kuo3Init,
                dim1532Kuo3Init,
                dim1533Kuo3Init,
                dim1534Kuo3Init,
                dim1535Kuo3Init,
                dim1536Kuo3Init,
                dim1537Kuo3Init,
                dim1538Kuo3Init,
                dim1539Kuo3Init,
                dim1540Kuo3Init,
                dim1541Kuo3Init,
                dim1542Kuo3Init,
                dim1543Kuo3Init,
                dim1544Kuo3Init,
                dim1545Kuo3Init,
                dim1546Kuo3Init,
                dim1547Kuo3Init,
                dim1548Kuo3Init,
                dim1549Kuo3Init,
                dim1550Kuo3Init,
                dim1551Kuo3Init,
                dim1552Kuo3Init,
                dim1553Kuo3Init,
                dim1554Kuo3Init,
                dim1555Kuo3Init,
                dim1556Kuo3Init,
                dim1557Kuo3Init,
                dim1558Kuo3Init,
                dim1559Kuo3Init,
                dim1560Kuo3Init,
                dim1561Kuo3Init,
                dim1562Kuo3Init,
                dim1563Kuo3Init,
                dim1564Kuo3Init,
                dim1565Kuo3Init,
                dim1566Kuo3Init,
                dim1567Kuo3Init,
                dim1568Kuo3Init,
                dim1569Kuo3Init,
                dim1570Kuo3Init,
                dim1571Kuo3Init,
                dim1572Kuo3Init,
                dim1573Kuo3Init,
                dim1574Kuo3Init,
                dim1575Kuo3Init,
                dim1576Kuo3Init,
                dim1577Kuo3Init,
                dim1578Kuo3Init,
                dim1579Kuo3Init,
                dim1580Kuo3Init,
                dim1581Kuo3Init,
                dim1582Kuo3Init,
                dim1583Kuo3Init,
                dim1584Kuo3Init,
                dim1585Kuo3Init,
                dim1586Kuo3Init,
                dim1587Kuo3Init,
                dim1588Kuo3Init,
                dim1589Kuo3Init,
                dim1590Kuo3Init,
                dim1591Kuo3Init,
                dim1592Kuo3Init,
                dim1593Kuo3Init,
                dim1594Kuo3Init,
                dim1595Kuo3Init,
                dim1596Kuo3Init,
                dim1597Kuo3Init,
                dim1598Kuo3Init,
                dim1599Kuo3Init,
                dim1600Kuo3Init,
                dim1601Kuo3Init,
                dim1602Kuo3Init,
                dim1603Kuo3Init,
                dim1604Kuo3Init,
                dim1605Kuo3Init,
                dim1606Kuo3Init,
                dim1607Kuo3Init,
                dim1608Kuo3Init,
                dim1609Kuo3Init,
                dim1610Kuo3Init,
                dim1611Kuo3Init,
                dim1612Kuo3Init,
                dim1613Kuo3Init,
                dim1614Kuo3Init,
                dim1615Kuo3Init,
                dim1616Kuo3Init,
                dim1617Kuo3Init,
                dim1618Kuo3Init,
                dim1619Kuo3Init,
                dim1620Kuo3Init,
                dim1621Kuo3Init,
                dim1622Kuo3Init,
                dim1623Kuo3Init,
                dim1624Kuo3Init,
                dim1625Kuo3Init,
                dim1626Kuo3Init,
                dim1627Kuo3Init,
                dim1628Kuo3Init,
                dim1629Kuo3Init,
                dim1630Kuo3Init,
                dim1631Kuo3Init,
                dim1632Kuo3Init,
                dim1633Kuo3Init,
                dim1634Kuo3Init,
                dim1635Kuo3Init,
                dim1636Kuo3Init,
                dim1637Kuo3Init,
                dim1638Kuo3Init,
                dim1639Kuo3Init,
                dim1640Kuo3Init,
                dim1641Kuo3Init,
                dim1642Kuo3Init,
                dim1643Kuo3Init,
                dim1644Kuo3Init,
                dim1645Kuo3Init,
                dim1646Kuo3Init,
                dim1647Kuo3Init,
                dim1648Kuo3Init,
                dim1649Kuo3Init,
                dim1650Kuo3Init,
                dim1651Kuo3Init,
                dim1652Kuo3Init,
                dim1653Kuo3Init,
                dim1654Kuo3Init,
                dim1655Kuo3Init,
                dim1656Kuo3Init,
                dim1657Kuo3Init,
                dim1658Kuo3Init,
                dim1659Kuo3Init,
                dim1660Kuo3Init,
                dim1661Kuo3Init,
                dim1662Kuo3Init,
                dim1663Kuo3Init,
                dim1664Kuo3Init,
                dim1665Kuo3Init,
                dim1666Kuo3Init,
                dim1667Kuo3Init,
                dim1668Kuo3Init,
                dim1669Kuo3Init,
                dim1670Kuo3Init,
                dim1671Kuo3Init,
                dim1672Kuo3Init,
                dim1673Kuo3Init,
                dim1674Kuo3Init,
                dim1675Kuo3Init,
                dim1676Kuo3Init,
                dim1677Kuo3Init,
                dim1678Kuo3Init,
                dim1679Kuo3Init,
                dim1680Kuo3Init,
                dim1681Kuo3Init,
                dim1682Kuo3Init,
                dim1683Kuo3Init,
                dim1684Kuo3Init,
                dim1685Kuo3Init,
                dim1686Kuo3Init,
                dim1687Kuo3Init,
                dim1688Kuo3Init,
                dim1689Kuo3Init,
                dim1690Kuo3Init,
                dim1691Kuo3Init,
                dim1692Kuo3Init,
                dim1693Kuo3Init,
                dim1694Kuo3Init,
                dim1695Kuo3Init,
                dim1696Kuo3Init,
                dim1697Kuo3Init,
                dim1698Kuo3Init,
                dim1699Kuo3Init,
                dim1700Kuo3Init,
                dim1701Kuo3Init,
                dim1702Kuo3Init,
                dim1703Kuo3Init,
                dim1704Kuo3Init,
                dim1705Kuo3Init,
                dim1706Kuo3Init,
                dim1707Kuo3Init,
                dim1708Kuo3Init,
                dim1709Kuo3Init,
                dim1710Kuo3Init,
                dim1711Kuo3Init,
                dim1712Kuo3Init,
                dim1713Kuo3Init,
                dim1714Kuo3Init,
                dim1715Kuo3Init,
                dim1716Kuo3Init,
                dim1717Kuo3Init,
                dim1718Kuo3Init,
                dim1719Kuo3Init,
                dim1720Kuo3Init,
                dim1721Kuo3Init,
                dim1722Kuo3Init,
                dim1723Kuo3Init,
                dim1724Kuo3Init,
                dim1725Kuo3Init,
                dim1726Kuo3Init,
                dim1727Kuo3Init,
                dim1728Kuo3Init,
                dim1729Kuo3Init,
                dim1730Kuo3Init,
                dim1731Kuo3Init,
                dim1732Kuo3Init,
                dim1733Kuo3Init,
                dim1734Kuo3Init,
                dim1735Kuo3Init,
                dim1736Kuo3Init,
                dim1737Kuo3Init,
                dim1738Kuo3Init,
                dim1739Kuo3Init,
                dim1740Kuo3Init,
                dim1741Kuo3Init,
                dim1742Kuo3Init,
                dim1743Kuo3Init,
                dim1744Kuo3Init,
                dim1745Kuo3Init,
                dim1746Kuo3Init,
                dim1747Kuo3Init,
                dim1748Kuo3Init,
                dim1749Kuo3Init,
                dim1750Kuo3Init,
                dim1751Kuo3Init,
                dim1752Kuo3Init,
                dim1753Kuo3Init,
                dim1754Kuo3Init,
                dim1755Kuo3Init,
                dim1756Kuo3Init,
                dim1757Kuo3Init,
                dim1758Kuo3Init,
                dim1759Kuo3Init,
                dim1760Kuo3Init,
                dim1761Kuo3Init,
                dim1762Kuo3Init,
                dim1763Kuo3Init,
                dim1764Kuo3Init,
                dim1765Kuo3Init,
                dim1766Kuo3Init,
                dim1767Kuo3Init,
                dim1768Kuo3Init,
                dim1769Kuo3Init,
                dim1770Kuo3Init,
                dim1771Kuo3Init,
                dim1772Kuo3Init,
                dim1773Kuo3Init,
                dim1774Kuo3Init,
                dim1775Kuo3Init,
                dim1776Kuo3Init,
                dim1777Kuo3Init,
                dim1778Kuo3Init,
                dim1779Kuo3Init,
                dim1780Kuo3Init,
                dim1781Kuo3Init,
                dim1782Kuo3Init,
                dim1783Kuo3Init,
                dim1784Kuo3Init,
                dim1785Kuo3Init,
                dim1786Kuo3Init,
                dim1787Kuo3Init,
                dim1788Kuo3Init,
                dim1789Kuo3Init,
                dim1790Kuo3Init,
                dim1791Kuo3Init,
                dim1792Kuo3Init,
                dim1793Kuo3Init,
                dim1794Kuo3Init,
                dim1795Kuo3Init,
                dim1796Kuo3Init,
                dim1797Kuo3Init,
                dim1798Kuo3Init,
                dim1799Kuo3Init,
                dim1800Kuo3Init,
                dim1801Kuo3Init,
                dim1802Kuo3Init,
                dim1803Kuo3Init,
                dim1804Kuo3Init,
                dim1805Kuo3Init,
                dim1806Kuo3Init,
                dim1807Kuo3Init,
                dim1808Kuo3Init,
                dim1809Kuo3Init,
                dim1810Kuo3Init,
                dim1811Kuo3Init,
                dim1812Kuo3Init,
                dim1813Kuo3Init,
                dim1814Kuo3Init,
                dim1815Kuo3Init,
                dim1816Kuo3Init,
                dim1817Kuo3Init,
                dim1818Kuo3Init,
                dim1819Kuo3Init,
                dim1820Kuo3Init,
                dim1821Kuo3Init,
                dim1822Kuo3Init,
                dim1823Kuo3Init,
                dim1824Kuo3Init,
                dim1825Kuo3Init,
                dim1826Kuo3Init,
                dim1827Kuo3Init,
                dim1828Kuo3Init,
                dim1829Kuo3Init,
                dim1830Kuo3Init,
                dim1831Kuo3Init,
                dim1832Kuo3Init,
                dim1833Kuo3Init,
                dim1834Kuo3Init,
                dim1835Kuo3Init,
                dim1836Kuo3Init,
                dim1837Kuo3Init,
                dim1838Kuo3Init,
                dim1839Kuo3Init,
                dim1840Kuo3Init,
                dim1841Kuo3Init,
                dim1842Kuo3Init,
                dim1843Kuo3Init,
                dim1844Kuo3Init,
                dim1845Kuo3Init,
                dim1846Kuo3Init,
                dim1847Kuo3Init,
                dim1848Kuo3Init,
                dim1849Kuo3Init,
                dim1850Kuo3Init,
                dim1851Kuo3Init,
                dim1852Kuo3Init,
                dim1853Kuo3Init,
                dim1854Kuo3Init,
                dim1855Kuo3Init,
                dim1856Kuo3Init,
                dim1857Kuo3Init,
                dim1858Kuo3Init,
                dim1859Kuo3Init,
                dim1860Kuo3Init,
                dim1861Kuo3Init,
                dim1862Kuo3Init,
                dim1863Kuo3Init,
                dim1864Kuo3Init,
                dim1865Kuo3Init,
                dim1866Kuo3Init,
                dim1867Kuo3Init,
                dim1868Kuo3Init,
                dim1869Kuo3Init,
                dim1870Kuo3Init,
                dim1871Kuo3Init,
                dim1872Kuo3Init,
                dim1873Kuo3Init,
                dim1874Kuo3Init,
                dim1875Kuo3Init,
                dim1876Kuo3Init,
                dim1877Kuo3Init,
                dim1878Kuo3Init,
                dim1879Kuo3Init,
                dim1880Kuo3Init,
                dim1881Kuo3Init,
                dim1882Kuo3Init,
                dim1883Kuo3Init,
                dim1884Kuo3Init,
                dim1885Kuo3Init,
                dim1886Kuo3Init,
                dim1887Kuo3Init,
                dim1888Kuo3Init,
                dim1889Kuo3Init,
                dim1890Kuo3Init,
                dim1891Kuo3Init,
                dim1892Kuo3Init,
                dim1893Kuo3Init,
                dim1894Kuo3Init,
                dim1895Kuo3Init,
                dim1896Kuo3Init,
                dim1897Kuo3Init,
                dim1898Kuo3Init,
                dim1899Kuo3Init,
                dim1900Kuo3Init,
                dim1901Kuo3Init,
                dim1902Kuo3Init,
                dim1903Kuo3Init,
                dim1904Kuo3Init,
                dim1905Kuo3Init,
                dim1906Kuo3Init,
                dim1907Kuo3Init,
                dim1908Kuo3Init,
                dim1909Kuo3Init,
                dim1910Kuo3Init,
                dim1911Kuo3Init,
                dim1912Kuo3Init,
                dim1913Kuo3Init,
                dim1914Kuo3Init,
                dim1915Kuo3Init,
                dim1916Kuo3Init,
                dim1917Kuo3Init,
                dim1918Kuo3Init,
                dim1919Kuo3Init,
                dim1920Kuo3Init,
                dim1921Kuo3Init,
                dim1922Kuo3Init,
                dim1923Kuo3Init,
                dim1924Kuo3Init,
                dim1925Kuo3Init,
                dim1926Kuo3Init,
                dim1927Kuo3Init,
                dim1928Kuo3Init,
                dim1929Kuo3Init,
                dim1930Kuo3Init,
                dim1931Kuo3Init,
                dim1932Kuo3Init,
                dim1933Kuo3Init,
                dim1934Kuo3Init,
                dim1935Kuo3Init,
                dim1936Kuo3Init,
                dim1937Kuo3Init,
                dim1938Kuo3Init,
                dim1939Kuo3Init,
                dim1940Kuo3Init,
                dim1941Kuo3Init,
                dim1942Kuo3Init,
                dim1943Kuo3Init,
                dim1944Kuo3Init,
                dim1945Kuo3Init,
                dim1946Kuo3Init,
                dim1947Kuo3Init,
                dim1948Kuo3Init,
                dim1949Kuo3Init,
                dim1950Kuo3Init,
                dim1951Kuo3Init,
                dim1952Kuo3Init,
                dim1953Kuo3Init,
                dim1954Kuo3Init,
                dim1955Kuo3Init,
                dim1956Kuo3Init,
                dim1957Kuo3Init,
                dim1958Kuo3Init,
                dim1959Kuo3Init,
                dim1960Kuo3Init,
                dim1961Kuo3Init,
                dim1962Kuo3Init,
                dim1963Kuo3Init,
                dim1964Kuo3Init,
                dim1965Kuo3Init,
                dim1966Kuo3Init,
                dim1967Kuo3Init,
                dim1968Kuo3Init,
                dim1969Kuo3Init,
                dim1970Kuo3Init,
                dim1971Kuo3Init,
                dim1972Kuo3Init,
                dim1973Kuo3Init,
                dim1974Kuo3Init,
                dim1975Kuo3Init,
                dim1976Kuo3Init,
                dim1977Kuo3Init,
                dim1978Kuo3Init,
                dim1979Kuo3Init,
                dim1980Kuo3Init,
                dim1981Kuo3Init,
                dim1982Kuo3Init,
                dim1983Kuo3Init,
                dim1984Kuo3Init,
                dim1985Kuo3Init,
                dim1986Kuo3Init,
                dim1987Kuo3Init,
                dim1988Kuo3Init,
                dim1989Kuo3Init,
                dim1990Kuo3Init,
                dim1991Kuo3Init,
                dim1992Kuo3Init,
                dim1993Kuo3Init,
                dim1994Kuo3Init,
                dim1995Kuo3Init,
                dim1996Kuo3Init,
                dim1997Kuo3Init,
                dim1998Kuo3Init,
                dim1999Kuo3Init,
                dim2000Kuo3Init,
                dim2001Kuo3Init,
                dim2002Kuo3Init,
                dim2003Kuo3Init,
                dim2004Kuo3Init,
                dim2005Kuo3Init,
                dim2006Kuo3Init,
                dim2007Kuo3Init,
                dim2008Kuo3Init,
                dim2009Kuo3Init,
                dim2010Kuo3Init,
                dim2011Kuo3Init,
                dim2012Kuo3Init,
                dim2013Kuo3Init,
                dim2014Kuo3Init,
                dim2015Kuo3Init,
                dim2016Kuo3Init,
                dim2017Kuo3Init,
                dim2018Kuo3Init,
                dim2019Kuo3Init,
                dim2020Kuo3Init,
                dim2021Kuo3Init,
                dim2022Kuo3Init,
                dim2023Kuo3Init,
                dim2024Kuo3Init,
                dim2025Kuo3Init,
                dim2026Kuo3Init,
                dim2027Kuo3Init,
                dim2028Kuo3Init,
                dim2029Kuo3Init,
                dim2030Kuo3Init,
                dim2031Kuo3Init,
                dim2032Kuo3Init,
                dim2033Kuo3Init,
                dim2034Kuo3Init,
                dim2035Kuo3Init,
                dim2036Kuo3Init,
                dim2037Kuo3Init,
                dim2038Kuo3Init,
                dim2039Kuo3Init,
                dim2040Kuo3Init,
                dim2041Kuo3Init,
                dim2042Kuo3Init,
                dim2043Kuo3Init,
                dim2044Kuo3Init,
                dim2045Kuo3Init,
                dim2046Kuo3Init,
                dim2047Kuo3Init,
                dim2048Kuo3Init,
                dim2049Kuo3Init,
                dim2050Kuo3Init,
                dim2051Kuo3Init,
                dim2052Kuo3Init,
                dim2053Kuo3Init,
                dim2054Kuo3Init,
                dim2055Kuo3Init,
                dim2056Kuo3Init,
                dim2057Kuo3Init,
                dim2058Kuo3Init,
                dim2059Kuo3Init,
                dim2060Kuo3Init,
                dim2061Kuo3Init,
                dim2062Kuo3Init,
                dim2063Kuo3Init,
                dim2064Kuo3Init,
                dim2065Kuo3Init,
                dim2066Kuo3Init,
                dim2067Kuo3Init,
                dim2068Kuo3Init,
                dim2069Kuo3Init,
                dim2070Kuo3Init,
                dim2071Kuo3Init,
                dim2072Kuo3Init,
                dim2073Kuo3Init,
                dim2074Kuo3Init,
                dim2075Kuo3Init,
                dim2076Kuo3Init,
                dim2077Kuo3Init,
                dim2078Kuo3Init,
                dim2079Kuo3Init,
                dim2080Kuo3Init,
                dim2081Kuo3Init,
                dim2082Kuo3Init,
                dim2083Kuo3Init,
                dim2084Kuo3Init,
                dim2085Kuo3Init,
                dim2086Kuo3Init,
                dim2087Kuo3Init,
                dim2088Kuo3Init,
                dim2089Kuo3Init,
                dim2090Kuo3Init,
                dim2091Kuo3Init,
                dim2092Kuo3Init,
                dim2093Kuo3Init,
                dim2094Kuo3Init,
                dim2095Kuo3Init,
                dim2096Kuo3Init,
                dim2097Kuo3Init,
                dim2098Kuo3Init,
                dim2099Kuo3Init,
                dim2100Kuo3Init,
                dim2101Kuo3Init,
                dim2102Kuo3Init,
                dim2103Kuo3Init,
                dim2104Kuo3Init,
                dim2105Kuo3Init,
                dim2106Kuo3Init,
                dim2107Kuo3Init,
                dim2108Kuo3Init,
                dim2109Kuo3Init,
                dim2110Kuo3Init,
                dim2111Kuo3Init,
                dim2112Kuo3Init,
                dim2113Kuo3Init,
                dim2114Kuo3Init,
                dim2115Kuo3Init,
                dim2116Kuo3Init,
                dim2117Kuo3Init,
                dim2118Kuo3Init,
                dim2119Kuo3Init,
                dim2120Kuo3Init,
                dim2121Kuo3Init,
                dim2122Kuo3Init,
                dim2123Kuo3Init,
                dim2124Kuo3Init,
                dim2125Kuo3Init,
                dim2126Kuo3Init,
                dim2127Kuo3Init,
                dim2128Kuo3Init,
                dim2129Kuo3Init,
                dim2130Kuo3Init,
                dim2131Kuo3Init,
                dim2132Kuo3Init,
                dim2133Kuo3Init,
                dim2134Kuo3Init,
                dim2135Kuo3Init,
                dim2136Kuo3Init,
                dim2137Kuo3Init,
                dim2138Kuo3Init,
                dim2139Kuo3Init,
                dim2140Kuo3Init,
                dim2141Kuo3Init,
                dim2142Kuo3Init,
                dim2143Kuo3Init,
                dim2144Kuo3Init,
                dim2145Kuo3Init,
                dim2146Kuo3Init,
                dim2147Kuo3Init,
                dim2148Kuo3Init,
                dim2149Kuo3Init,
                dim2150Kuo3Init,
                dim2151Kuo3Init,
                dim2152Kuo3Init,
                dim2153Kuo3Init,
                dim2154Kuo3Init,
                dim2155Kuo3Init,
                dim2156Kuo3Init,
                dim2157Kuo3Init,
                dim2158Kuo3Init,
                dim2159Kuo3Init,
                dim2160Kuo3Init,
                dim2161Kuo3Init,
                dim2162Kuo3Init,
                dim2163Kuo3Init,
                dim2164Kuo3Init,
                dim2165Kuo3Init,
                dim2166Kuo3Init,
                dim2167Kuo3Init,
                dim2168Kuo3Init,
                dim2169Kuo3Init,
                dim2170Kuo3Init,
                dim2171Kuo3Init,
                dim2172Kuo3Init,
                dim2173Kuo3Init,
                dim2174Kuo3Init,
                dim2175Kuo3Init,
                dim2176Kuo3Init,
                dim2177Kuo3Init,
                dim2178Kuo3Init,
                dim2179Kuo3Init,
                dim2180Kuo3Init,
                dim2181Kuo3Init,
                dim2182Kuo3Init,
                dim2183Kuo3Init,
                dim2184Kuo3Init,
                dim2185Kuo3Init,
                dim2186Kuo3Init,
                dim2187Kuo3Init,
                dim2188Kuo3Init,
                dim2189Kuo3Init,
                dim2190Kuo3Init,
                dim2191Kuo3Init,
                dim2192Kuo3Init,
                dim2193Kuo3Init,
                dim2194Kuo3Init,
                dim2195Kuo3Init,
                dim2196Kuo3Init,
                dim2197Kuo3Init,
                dim2198Kuo3Init,
                dim2199Kuo3Init,
                dim2200Kuo3Init,
                dim2201Kuo3Init,
                dim2202Kuo3Init,
                dim2203Kuo3Init,
                dim2204Kuo3Init,
                dim2205Kuo3Init,
                dim2206Kuo3Init,
                dim2207Kuo3Init,
                dim2208Kuo3Init,
                dim2209Kuo3Init,
                dim2210Kuo3Init,
                dim2211Kuo3Init,
                dim2212Kuo3Init,
                dim2213Kuo3Init,
                dim2214Kuo3Init,
                dim2215Kuo3Init,
                dim2216Kuo3Init,
                dim2217Kuo3Init,
                dim2218Kuo3Init,
                dim2219Kuo3Init,
                dim2220Kuo3Init,
                dim2221Kuo3Init,
                dim2222Kuo3Init,
                dim2223Kuo3Init,
                dim2224Kuo3Init,
                dim2225Kuo3Init,
                dim2226Kuo3Init,
                dim2227Kuo3Init,
                dim2228Kuo3Init,
                dim2229Kuo3Init,
                dim2230Kuo3Init,
                dim2231Kuo3Init,
                dim2232Kuo3Init,
                dim2233Kuo3Init,
                dim2234Kuo3Init,
                dim2235Kuo3Init,
                dim2236Kuo3Init,
                dim2237Kuo3Init,
                dim2238Kuo3Init,
                dim2239Kuo3Init,
                dim2240Kuo3Init,
                dim2241Kuo3Init,
                dim2242Kuo3Init,
                dim2243Kuo3Init,
                dim2244Kuo3Init,
                dim2245Kuo3Init,
                dim2246Kuo3Init,
                dim2247Kuo3Init,
                dim2248Kuo3Init,
                dim2249Kuo3Init,
                dim2250Kuo3Init,
                dim2251Kuo3Init,
                dim2252Kuo3Init,
                dim2253Kuo3Init,
                dim2254Kuo3Init,
                dim2255Kuo3Init,
                dim2256Kuo3Init,
                dim2257Kuo3Init,
                dim2258Kuo3Init,
                dim2259Kuo3Init,
                dim2260Kuo3Init,
                dim2261Kuo3Init,
                dim2262Kuo3Init,
                dim2263Kuo3Init,
                dim2264Kuo3Init,
                dim2265Kuo3Init,
                dim2266Kuo3Init,
                dim2267Kuo3Init,
                dim2268Kuo3Init,
                dim2269Kuo3Init,
                dim2270Kuo3Init,
                dim2271Kuo3Init,
                dim2272Kuo3Init,
                dim2273Kuo3Init,
                dim2274Kuo3Init,
                dim2275Kuo3Init,
                dim2276Kuo3Init,
                dim2277Kuo3Init,
                dim2278Kuo3Init,
                dim2279Kuo3Init,
                dim2280Kuo3Init,
                dim2281Kuo3Init,
                dim2282Kuo3Init,
                dim2283Kuo3Init,
                dim2284Kuo3Init,
                dim2285Kuo3Init,
                dim2286Kuo3Init,
                dim2287Kuo3Init,
                dim2288Kuo3Init,
                dim2289Kuo3Init,
                dim2290Kuo3Init,
                dim2291Kuo3Init,
                dim2292Kuo3Init,
                dim2293Kuo3Init,
                dim2294Kuo3Init,
                dim2295Kuo3Init,
                dim2296Kuo3Init,
                dim2297Kuo3Init,
                dim2298Kuo3Init,
                dim2299Kuo3Init,
                dim2300Kuo3Init,
                dim2301Kuo3Init,
                dim2302Kuo3Init,
                dim2303Kuo3Init,
                dim2304Kuo3Init,
                dim2305Kuo3Init,
                dim2306Kuo3Init,
                dim2307Kuo3Init,
                dim2308Kuo3Init,
                dim2309Kuo3Init,
                dim2310Kuo3Init,
                dim2311Kuo3Init,
                dim2312Kuo3Init,
                dim2313Kuo3Init,
                dim2314Kuo3Init,
                dim2315Kuo3Init,
                dim2316Kuo3Init,
                dim2317Kuo3Init,
                dim2318Kuo3Init,
                dim2319Kuo3Init,
                dim2320Kuo3Init,
                dim2321Kuo3Init,
                dim2322Kuo3Init,
                dim2323Kuo3Init,
                dim2324Kuo3Init,
                dim2325Kuo3Init,
                dim2326Kuo3Init,
                dim2327Kuo3Init,
                dim2328Kuo3Init,
                dim2329Kuo3Init,
                dim2330Kuo3Init,
                dim2331Kuo3Init,
                dim2332Kuo3Init,
                dim2333Kuo3Init,
                dim2334Kuo3Init,
                dim2335Kuo3Init,
                dim2336Kuo3Init,
                dim2337Kuo3Init,
                dim2338Kuo3Init,
                dim2339Kuo3Init,
                dim2340Kuo3Init,
                dim2341Kuo3Init,
                dim2342Kuo3Init,
                dim2343Kuo3Init,
                dim2344Kuo3Init,
                dim2345Kuo3Init,
                dim2346Kuo3Init,
                dim2347Kuo3Init,
                dim2348Kuo3Init,
                dim2349Kuo3Init,
                dim2350Kuo3Init,
                dim2351Kuo3Init,
                dim2352Kuo3Init,
                dim2353Kuo3Init,
                dim2354Kuo3Init,
                dim2355Kuo3Init,
                dim2356Kuo3Init,
                dim2357Kuo3Init,
                dim2358Kuo3Init,
                dim2359Kuo3Init,
                dim2360Kuo3Init,
                dim2361Kuo3Init,
                dim2362Kuo3Init,
                dim2363Kuo3Init,
                dim2364Kuo3Init,
                dim2365Kuo3Init,
                dim2366Kuo3Init,
                dim2367Kuo3Init,
                dim2368Kuo3Init,
                dim2369Kuo3Init,
                dim2370Kuo3Init,
                dim2371Kuo3Init,
                dim2372Kuo3Init,
                dim2373Kuo3Init,
                dim2374Kuo3Init,
                dim2375Kuo3Init,
                dim2376Kuo3Init,
                dim2377Kuo3Init,
                dim2378Kuo3Init,
                dim2379Kuo3Init,
                dim2380Kuo3Init,
                dim2381Kuo3Init,
                dim2382Kuo3Init,
                dim2383Kuo3Init,
                dim2384Kuo3Init,
                dim2385Kuo3Init,
                dim2386Kuo3Init,
                dim2387Kuo3Init,
                dim2388Kuo3Init,
                dim2389Kuo3Init,
                dim2390Kuo3Init,
                dim2391Kuo3Init,
                dim2392Kuo3Init,
                dim2393Kuo3Init,
                dim2394Kuo3Init,
                dim2395Kuo3Init,
                dim2396Kuo3Init,
                dim2397Kuo3Init,
                dim2398Kuo3Init,
                dim2399Kuo3Init,
                dim2400Kuo3Init,
                dim2401Kuo3Init,
                dim2402Kuo3Init,
                dim2403Kuo3Init,
                dim2404Kuo3Init,
                dim2405Kuo3Init,
                dim2406Kuo3Init,
                dim2407Kuo3Init,
                dim2408Kuo3Init,
                dim2409Kuo3Init,
                dim2410Kuo3Init,
                dim2411Kuo3Init,
                dim2412Kuo3Init,
                dim2413Kuo3Init,
                dim2414Kuo3Init,
                dim2415Kuo3Init,
                dim2416Kuo3Init,
                dim2417Kuo3Init,
                dim2418Kuo3Init,
                dim2419Kuo3Init,
                dim2420Kuo3Init,
                dim2421Kuo3Init,
                dim2422Kuo3Init,
                dim2423Kuo3Init,
                dim2424Kuo3Init,
                dim2425Kuo3Init,
                dim2426Kuo3Init,
                dim2427Kuo3Init,
                dim2428Kuo3Init,
                dim2429Kuo3Init,
                dim2430Kuo3Init,
                dim2431Kuo3Init,
                dim2432Kuo3Init,
                dim2433Kuo3Init,
                dim2434Kuo3Init,
                dim2435Kuo3Init,
                dim2436Kuo3Init,
                dim2437Kuo3Init,
                dim2438Kuo3Init,
                dim2439Kuo3Init,
                dim2440Kuo3Init,
                dim2441Kuo3Init,
                dim2442Kuo3Init,
                dim2443Kuo3Init,
                dim2444Kuo3Init,
                dim2445Kuo3Init,
                dim2446Kuo3Init,
                dim2447Kuo3Init,
                dim2448Kuo3Init,
                dim2449Kuo3Init,
                dim2450Kuo3Init,
                dim2451Kuo3Init,
                dim2452Kuo3Init,
                dim2453Kuo3Init,
                dim2454Kuo3Init,
                dim2455Kuo3Init,
                dim2456Kuo3Init,
                dim2457Kuo3Init,
                dim2458Kuo3Init,
                dim2459Kuo3Init,
                dim2460Kuo3Init,
                dim2461Kuo3Init,
                dim2462Kuo3Init,
                dim2463Kuo3Init,
                dim2464Kuo3Init,
                dim2465Kuo3Init,
                dim2466Kuo3Init,
                dim2467Kuo3Init,
                dim2468Kuo3Init,
                dim2469Kuo3Init,
                dim2470Kuo3Init,
                dim2471Kuo3Init,
                dim2472Kuo3Init,
                dim2473Kuo3Init,
                dim2474Kuo3Init,
                dim2475Kuo3Init,
                dim2476Kuo3Init,
                dim2477Kuo3Init,
                dim2478Kuo3Init,
                dim2479Kuo3Init,
                dim2480Kuo3Init,
                dim2481Kuo3Init,
                dim2482Kuo3Init,
                dim2483Kuo3Init,
                dim2484Kuo3Init,
                dim2485Kuo3Init,
                dim2486Kuo3Init,
                dim2487Kuo3Init,
                dim2488Kuo3Init,
                dim2489Kuo3Init,
                dim2490Kuo3Init,
                dim2491Kuo3Init,
                dim2492Kuo3Init,
                dim2493Kuo3Init,
                dim2494Kuo3Init,
                dim2495Kuo3Init,
                dim2496Kuo3Init,
                dim2497Kuo3Init,
                dim2498Kuo3Init,
                dim2499Kuo3Init,
                dim2500Kuo3Init,
                dim2501Kuo3Init,
                dim2502Kuo3Init,
                dim2503Kuo3Init,
                dim2504Kuo3Init,
                dim2505Kuo3Init,
                dim2506Kuo3Init,
                dim2507Kuo3Init,
                dim2508Kuo3Init,
                dim2509Kuo3Init,
                dim2510Kuo3Init,
                dim2511Kuo3Init,
                dim2512Kuo3Init,
                dim2513Kuo3Init,
                dim2514Kuo3Init,
                dim2515Kuo3Init,
                dim2516Kuo3Init,
                dim2517Kuo3Init,
                dim2518Kuo3Init,
                dim2519Kuo3Init,
                dim2520Kuo3Init,
                dim2521Kuo3Init,
                dim2522Kuo3Init,
                dim2523Kuo3Init,
                dim2524Kuo3Init,
                dim2525Kuo3Init,
                dim2526Kuo3Init,
                dim2527Kuo3Init,
                dim2528Kuo3Init,
                dim2529Kuo3Init,
                dim2530Kuo3Init,
                dim2531Kuo3Init,
                dim2532Kuo3Init,
                dim2533Kuo3Init,
                dim2534Kuo3Init,
                dim2535Kuo3Init,
                dim2536Kuo3Init,
                dim2537Kuo3Init,
                dim2538Kuo3Init,
                dim2539Kuo3Init,
                dim2540Kuo3Init,
                dim2541Kuo3Init,
                dim2542Kuo3Init,
                dim2543Kuo3Init,
                dim2544Kuo3Init,
                dim2545Kuo3Init,
                dim2546Kuo3Init,
                dim2547Kuo3Init,
                dim2548Kuo3Init,
                dim2549Kuo3Init,
                dim2550Kuo3Init,
                dim2551Kuo3Init,
                dim2552Kuo3Init,
                dim2553Kuo3Init,
                dim2554Kuo3Init,
                dim2555Kuo3Init,
                dim2556Kuo3Init,
                dim2557Kuo3Init,
                dim2558Kuo3Init,
                dim2559Kuo3Init,
                dim2560Kuo3Init,
                dim2561Kuo3Init,
                dim2562Kuo3Init,
                dim2563Kuo3Init,
                dim2564Kuo3Init,
                dim2565Kuo3Init,
                dim2566Kuo3Init,
                dim2567Kuo3Init,
                dim2568Kuo3Init,
                dim2569Kuo3Init,
                dim2570Kuo3Init,
                dim2571Kuo3Init,
                dim2572Kuo3Init,
                dim2573Kuo3Init,
                dim2574Kuo3Init,
                dim2575Kuo3Init,
                dim2576Kuo3Init,
                dim2577Kuo3Init,
                dim2578Kuo3Init,
                dim2579Kuo3Init,
                dim2580Kuo3Init,
                dim2581Kuo3Init,
                dim2582Kuo3Init,
                dim2583Kuo3Init,
                dim2584Kuo3Init,
                dim2585Kuo3Init,
                dim2586Kuo3Init,
                dim2587Kuo3Init,
                dim2588Kuo3Init,
                dim2589Kuo3Init,
                dim2590Kuo3Init,
                dim2591Kuo3Init,
                dim2592Kuo3Init,
                dim2593Kuo3Init,
                dim2594Kuo3Init,
                dim2595Kuo3Init,
                dim2596Kuo3Init,
                dim2597Kuo3Init,
                dim2598Kuo3Init,
                dim2599Kuo3Init,
                dim2600Kuo3Init,
                dim2601Kuo3Init,
                dim2602Kuo3Init,
                dim2603Kuo3Init,
                dim2604Kuo3Init,
                dim2605Kuo3Init,
                dim2606Kuo3Init,
                dim2607Kuo3Init,
                dim2608Kuo3Init,
                dim2609Kuo3Init,
                dim2610Kuo3Init,
                dim2611Kuo3Init,
                dim2612Kuo3Init,
                dim2613Kuo3Init,
                dim2614Kuo3Init,
                dim2615Kuo3Init,
                dim2616Kuo3Init,
                dim2617Kuo3Init,
                dim2618Kuo3Init,
                dim2619Kuo3Init,
                dim2620Kuo3Init,
                dim2621Kuo3Init,
                dim2622Kuo3Init,
                dim2623Kuo3Init,
                dim2624Kuo3Init,
                dim2625Kuo3Init,
                dim2626Kuo3Init,
                dim2627Kuo3Init,
                dim2628Kuo3Init,
                dim2629Kuo3Init,
                dim2630Kuo3Init,
                dim2631Kuo3Init,
                dim2632Kuo3Init,
                dim2633Kuo3Init,
                dim2634Kuo3Init,
                dim2635Kuo3Init,
                dim2636Kuo3Init,
                dim2637Kuo3Init,
                dim2638Kuo3Init,
                dim2639Kuo3Init,
                dim2640Kuo3Init,
                dim2641Kuo3Init,
                dim2642Kuo3Init,
                dim2643Kuo3Init,
                dim2644Kuo3Init,
                dim2645Kuo3Init,
                dim2646Kuo3Init,
                dim2647Kuo3Init,
                dim2648Kuo3Init,
                dim2649Kuo3Init,
                dim2650Kuo3Init,
                dim2651Kuo3Init,
                dim2652Kuo3Init,
                dim2653Kuo3Init,
                dim2654Kuo3Init,
                dim2655Kuo3Init,
                dim2656Kuo3Init,
                dim2657Kuo3Init,
                dim2658Kuo3Init,
                dim2659Kuo3Init,
                dim2660Kuo3Init,
                dim2661Kuo3Init,
                dim2662Kuo3Init,
                dim2663Kuo3Init,
                dim2664Kuo3Init,
                dim2665Kuo3Init,
                dim2666Kuo3Init,
                dim2667Kuo3Init,
                dim2668Kuo3Init,
                dim2669Kuo3Init,
                dim2670Kuo3Init,
                dim2671Kuo3Init,
                dim2672Kuo3Init,
                dim2673Kuo3Init,
                dim2674Kuo3Init,
                dim2675Kuo3Init,
                dim2676Kuo3Init,
                dim2677Kuo3Init,
                dim2678Kuo3Init,
                dim2679Kuo3Init,
                dim2680Kuo3Init,
                dim2681Kuo3Init,
                dim2682Kuo3Init,
                dim2683Kuo3Init,
                dim2684Kuo3Init,
                dim2685Kuo3Init,
                dim2686Kuo3Init,
                dim2687Kuo3Init,
                dim2688Kuo3Init,
                dim2689Kuo3Init,
                dim2690Kuo3Init,
                dim2691Kuo3Init,
                dim2692Kuo3Init,
                dim2693Kuo3Init,
                dim2694Kuo3Init,
                dim2695Kuo3Init,
                dim2696Kuo3Init,
                dim2697Kuo3Init,
                dim2698Kuo3Init,
                dim2699Kuo3Init,
                dim2700Kuo3Init,
                dim2701Kuo3Init,
                dim2702Kuo3Init,
                dim2703Kuo3Init,
                dim2704Kuo3Init,
                dim2705Kuo3Init,
                dim2706Kuo3Init,
                dim2707Kuo3Init,
                dim2708Kuo3Init,
                dim2709Kuo3Init,
                dim2710Kuo3Init,
                dim2711Kuo3Init,
                dim2712Kuo3Init,
                dim2713Kuo3Init,
                dim2714Kuo3Init,
                dim2715Kuo3Init,
                dim2716Kuo3Init,
                dim2717Kuo3Init,
                dim2718Kuo3Init,
                dim2719Kuo3Init,
                dim2720Kuo3Init,
                dim2721Kuo3Init,
                dim2722Kuo3Init,
                dim2723Kuo3Init,
                dim2724Kuo3Init,
                dim2725Kuo3Init,
                dim2726Kuo3Init,
                dim2727Kuo3Init,
                dim2728Kuo3Init,
                dim2729Kuo3Init,
                dim2730Kuo3Init,
                dim2731Kuo3Init,
                dim2732Kuo3Init,
                dim2733Kuo3Init,
                dim2734Kuo3Init,
                dim2735Kuo3Init,
                dim2736Kuo3Init,
                dim2737Kuo3Init,
                dim2738Kuo3Init,
                dim2739Kuo3Init,
                dim2740Kuo3Init,
                dim2741Kuo3Init,
                dim2742Kuo3Init,
                dim2743Kuo3Init,
                dim2744Kuo3Init,
                dim2745Kuo3Init,
                dim2746Kuo3Init,
                dim2747Kuo3Init,
                dim2748Kuo3Init,
                dim2749Kuo3Init,
                dim2750Kuo3Init,
                dim2751Kuo3Init,
                dim2752Kuo3Init,
                dim2753Kuo3Init,
                dim2754Kuo3Init,
                dim2755Kuo3Init,
                dim2756Kuo3Init,
                dim2757Kuo3Init,
                dim2758Kuo3Init,
                dim2759Kuo3Init,
                dim2760Kuo3Init,
                dim2761Kuo3Init,
                dim2762Kuo3Init,
                dim2763Kuo3Init,
                dim2764Kuo3Init,
                dim2765Kuo3Init,
                dim2766Kuo3Init,
                dim2767Kuo3Init,
                dim2768Kuo3Init,
                dim2769Kuo3Init,
                dim2770Kuo3Init,
                dim2771Kuo3Init,
                dim2772Kuo3Init,
                dim2773Kuo3Init,
                dim2774Kuo3Init,
                dim2775Kuo3Init,
                dim2776Kuo3Init,
                dim2777Kuo3Init,
                dim2778Kuo3Init,
                dim2779Kuo3Init,
                dim2780Kuo3Init,
                dim2781Kuo3Init,
                dim2782Kuo3Init,
                dim2783Kuo3Init,
                dim2784Kuo3Init,
                dim2785Kuo3Init,
                dim2786Kuo3Init,
                dim2787Kuo3Init,
                dim2788Kuo3Init,
                dim2789Kuo3Init,
                dim2790Kuo3Init,
                dim2791Kuo3Init,
                dim2792Kuo3Init,
                dim2793Kuo3Init,
                dim2794Kuo3Init,
                dim2795Kuo3Init,
                dim2796Kuo3Init,
                dim2797Kuo3Init,
                dim2798Kuo3Init,
                dim2799Kuo3Init,
                dim2800Kuo3Init,
                dim2801Kuo3Init,
                dim2802Kuo3Init,
                dim2803Kuo3Init,
                dim2804Kuo3Init,
                dim2805Kuo3Init,
                dim2806Kuo3Init,
                dim2807Kuo3Init,
                dim2808Kuo3Init,
                dim2809Kuo3Init,
                dim2810Kuo3Init,
                dim2811Kuo3Init,
                dim2812Kuo3Init,
                dim2813Kuo3Init,
                dim2814Kuo3Init,
                dim2815Kuo3Init,
                dim2816Kuo3Init,
                dim2817Kuo3Init,
                dim2818Kuo3Init,
                dim2819Kuo3Init,
                dim2820Kuo3Init,
                dim2821Kuo3Init,
                dim2822Kuo3Init,
                dim2823Kuo3Init,
                dim2824Kuo3Init,
                dim2825Kuo3Init,
                dim2826Kuo3Init,
                dim2827Kuo3Init,
                dim2828Kuo3Init,
                dim2829Kuo3Init,
                dim2830Kuo3Init,
                dim2831Kuo3Init,
                dim2832Kuo3Init,
                dim2833Kuo3Init,
                dim2834Kuo3Init,
                dim2835Kuo3Init,
                dim2836Kuo3Init,
                dim2837Kuo3Init,
                dim2838Kuo3Init,
                dim2839Kuo3Init,
                dim2840Kuo3Init,
                dim2841Kuo3Init,
                dim2842Kuo3Init,
                dim2843Kuo3Init,
                dim2844Kuo3Init,
                dim2845Kuo3Init,
                dim2846Kuo3Init,
                dim2847Kuo3Init,
                dim2848Kuo3Init,
                dim2849Kuo3Init,
                dim2850Kuo3Init,
                dim2851Kuo3Init,
                dim2852Kuo3Init,
                dim2853Kuo3Init,
                dim2854Kuo3Init,
                dim2855Kuo3Init,
                dim2856Kuo3Init,
                dim2857Kuo3Init,
                dim2858Kuo3Init,
                dim2859Kuo3Init,
                dim2860Kuo3Init,
                dim2861Kuo3Init,
                dim2862Kuo3Init,
                dim2863Kuo3Init,
                dim2864Kuo3Init,
                dim2865Kuo3Init,
                dim2866Kuo3Init,
                dim2867Kuo3Init,
                dim2868Kuo3Init,
                dim2869Kuo3Init,
                dim2870Kuo3Init,
                dim2871Kuo3Init,
                dim2872Kuo3Init,
                dim2873Kuo3Init,
                dim2874Kuo3Init,
                dim2875Kuo3Init,
                dim2876Kuo3Init,
                dim2877Kuo3Init,
                dim2878Kuo3Init,
                dim2879Kuo3Init,
                dim2880Kuo3Init,
                dim2881Kuo3Init,
                dim2882Kuo3Init,
                dim2883Kuo3Init,
                dim2884Kuo3Init,
                dim2885Kuo3Init,
                dim2886Kuo3Init,
                dim2887Kuo3Init,
                dim2888Kuo3Init,
                dim2889Kuo3Init,
                dim2890Kuo3Init,
                dim2891Kuo3Init,
                dim2892Kuo3Init,
                dim2893Kuo3Init,
                dim2894Kuo3Init,
                dim2895Kuo3Init,
                dim2896Kuo3Init,
                dim2897Kuo3Init,
                dim2898Kuo3Init,
                dim2899Kuo3Init,
                dim2900Kuo3Init,
                dim2901Kuo3Init,
                dim2902Kuo3Init,
                dim2903Kuo3Init,
                dim2904Kuo3Init,
                dim2905Kuo3Init,
                dim2906Kuo3Init,
                dim2907Kuo3Init,
                dim2908Kuo3Init,
                dim2909Kuo3Init,
                dim2910Kuo3Init,
                dim2911Kuo3Init,
                dim2912Kuo3Init,
                dim2913Kuo3Init,
                dim2914Kuo3Init,
                dim2915Kuo3Init,
                dim2916Kuo3Init,
                dim2917Kuo3Init,
                dim2918Kuo3Init,
                dim2919Kuo3Init,
                dim2920Kuo3Init,
                dim2921Kuo3Init,
                dim2922Kuo3Init,
                dim2923Kuo3Init,
                dim2924Kuo3Init,
                dim2925Kuo3Init,
                dim2926Kuo3Init,
                dim2927Kuo3Init,
                dim2928Kuo3Init,
                dim2929Kuo3Init,
                dim2930Kuo3Init,
                dim2931Kuo3Init,
                dim2932Kuo3Init,
                dim2933Kuo3Init,
                dim2934Kuo3Init,
                dim2935Kuo3Init,
                dim2936Kuo3Init,
                dim2937Kuo3Init,
                dim2938Kuo3Init,
                dim2939Kuo3Init,
                dim2940Kuo3Init,
                dim2941Kuo3Init,
                dim2942Kuo3Init,
                dim2943Kuo3Init,
                dim2944Kuo3Init,
                dim2945Kuo3Init,
                dim2946Kuo3Init,
                dim2947Kuo3Init,
                dim2948Kuo3Init,
                dim2949Kuo3Init,
                dim2950Kuo3Init,
                dim2951Kuo3Init,
                dim2952Kuo3Init,
                dim2953Kuo3Init,
                dim2954Kuo3Init,
                dim2955Kuo3Init,
                dim2956Kuo3Init,
                dim2957Kuo3Init,
                dim2958Kuo3Init,
                dim2959Kuo3Init,
                dim2960Kuo3Init,
                dim2961Kuo3Init,
                dim2962Kuo3Init,
                dim2963Kuo3Init,
                dim2964Kuo3Init,
                dim2965Kuo3Init,
                dim2966Kuo3Init,
                dim2967Kuo3Init,
                dim2968Kuo3Init,
                dim2969Kuo3Init,
                dim2970Kuo3Init,
                dim2971Kuo3Init,
                dim2972Kuo3Init,
                dim2973Kuo3Init,
                dim2974Kuo3Init,
                dim2975Kuo3Init,
                dim2976Kuo3Init,
                dim2977Kuo3Init,
                dim2978Kuo3Init,
                dim2979Kuo3Init,
                dim2980Kuo3Init,
                dim2981Kuo3Init,
                dim2982Kuo3Init,
                dim2983Kuo3Init,
                dim2984Kuo3Init,
                dim2985Kuo3Init,
                dim2986Kuo3Init,
                dim2987Kuo3Init,
                dim2988Kuo3Init,
                dim2989Kuo3Init,
                dim2990Kuo3Init,
                dim2991Kuo3Init,
                dim2992Kuo3Init,
                dim2993Kuo3Init,
                dim2994Kuo3Init,
                dim2995Kuo3Init,
                dim2996Kuo3Init,
                dim2997Kuo3Init,
                dim2998Kuo3Init,
                dim2999Kuo3Init,
                dim3000Kuo3Init,
                dim3001Kuo3Init,
                dim3002Kuo3Init,
                dim3003Kuo3Init,
                dim3004Kuo3Init,
                dim3005Kuo3Init,
                dim3006Kuo3Init,
                dim3007Kuo3Init,
                dim3008Kuo3Init,
                dim3009Kuo3Init,
                dim3010Kuo3Init,
                dim3011Kuo3Init,
                dim3012Kuo3Init,
                dim3013Kuo3Init,
                dim3014Kuo3Init,
                dim3015Kuo3Init,
                dim3016Kuo3Init,
                dim3017Kuo3Init,
                dim3018Kuo3Init,
                dim3019Kuo3Init,
                dim3020Kuo3Init,
                dim3021Kuo3Init,
                dim3022Kuo3Init,
                dim3023Kuo3Init,
                dim3024Kuo3Init,
                dim3025Kuo3Init,
                dim3026Kuo3Init,
                dim3027Kuo3Init,
                dim3028Kuo3Init,
                dim3029Kuo3Init,
                dim3030Kuo3Init,
                dim3031Kuo3Init,
                dim3032Kuo3Init,
                dim3033Kuo3Init,
                dim3034Kuo3Init,
                dim3035Kuo3Init,
                dim3036Kuo3Init,
                dim3037Kuo3Init,
                dim3038Kuo3Init,
                dim3039Kuo3Init,
                dim3040Kuo3Init,
                dim3041Kuo3Init,
                dim3042Kuo3Init,
                dim3043Kuo3Init,
                dim3044Kuo3Init,
                dim3045Kuo3Init,
                dim3046Kuo3Init,
                dim3047Kuo3Init,
                dim3048Kuo3Init,
                dim3049Kuo3Init,
                dim3050Kuo3Init,
                dim3051Kuo3Init,
                dim3052Kuo3Init,
                dim3053Kuo3Init,
                dim3054Kuo3Init,
                dim3055Kuo3Init,
                dim3056Kuo3Init,
                dim3057Kuo3Init,
                dim3058Kuo3Init,
                dim3059Kuo3Init,
                dim3060Kuo3Init,
                dim3061Kuo3Init,
                dim3062Kuo3Init,
                dim3063Kuo3Init,
                dim3064Kuo3Init,
                dim3065Kuo3Init,
                dim3066Kuo3Init,
                dim3067Kuo3Init,
                dim3068Kuo3Init,
                dim3069Kuo3Init,
                dim3070Kuo3Init,
                dim3071Kuo3Init,
                dim3072Kuo3Init,
                dim3073Kuo3Init,
                dim3074Kuo3Init,
                dim3075Kuo3Init,
                dim3076Kuo3Init,
                dim3077Kuo3Init,
                dim3078Kuo3Init,
                dim3079Kuo3Init,
                dim3080Kuo3Init,
                dim3081Kuo3Init,
                dim3082Kuo3Init,
                dim3083Kuo3Init,
                dim3084Kuo3Init,
                dim3085Kuo3Init,
                dim3086Kuo3Init,
                dim3087Kuo3Init,
                dim3088Kuo3Init,
                dim3089Kuo3Init,
                dim3090Kuo3Init,
                dim3091Kuo3Init,
                dim3092Kuo3Init,
                dim3093Kuo3Init,
                dim3094Kuo3Init,
                dim3095Kuo3Init,
                dim3096Kuo3Init,
                dim3097Kuo3Init,
                dim3098Kuo3Init,
                dim3099Kuo3Init,
                dim3100Kuo3Init,
                dim3101Kuo3Init,
                dim3102Kuo3Init,
                dim3103Kuo3Init,
                dim3104Kuo3Init,
                dim3105Kuo3Init,
                dim3106Kuo3Init,
                dim3107Kuo3Init,
                dim3108Kuo3Init,
                dim3109Kuo3Init,
                dim3110Kuo3Init,
                dim3111Kuo3Init,
                dim3112Kuo3Init,
                dim3113Kuo3Init,
                dim3114Kuo3Init,
                dim3115Kuo3Init,
                dim3116Kuo3Init,
                dim3117Kuo3Init,
                dim3118Kuo3Init,
                dim3119Kuo3Init,
                dim3120Kuo3Init,
                dim3121Kuo3Init,
                dim3122Kuo3Init,
                dim3123Kuo3Init,
                dim3124Kuo3Init,
                dim3125Kuo3Init,
                dim3126Kuo3Init,
                dim3127Kuo3Init,
                dim3128Kuo3Init,
                dim3129Kuo3Init,
                dim3130Kuo3Init,
                dim3131Kuo3Init,
                dim3132Kuo3Init,
                dim3133Kuo3Init,
                dim3134Kuo3Init,
                dim3135Kuo3Init,
                dim3136Kuo3Init,
                dim3137Kuo3Init,
                dim3138Kuo3Init,
                dim3139Kuo3Init,
                dim3140Kuo3Init,
                dim3141Kuo3Init,
                dim3142Kuo3Init,
                dim3143Kuo3Init,
                dim3144Kuo3Init,
                dim3145Kuo3Init,
                dim3146Kuo3Init,
                dim3147Kuo3Init,
                dim3148Kuo3Init,
                dim3149Kuo3Init,
                dim3150Kuo3Init,
                dim3151Kuo3Init,
                dim3152Kuo3Init,
                dim3153Kuo3Init,
                dim3154Kuo3Init,
                dim3155Kuo3Init,
                dim3156Kuo3Init,
                dim3157Kuo3Init,
                dim3158Kuo3Init,
                dim3159Kuo3Init,
                dim3160Kuo3Init,
                dim3161Kuo3Init,
                dim3162Kuo3Init,
                dim3163Kuo3Init,
                dim3164Kuo3Init,
                dim3165Kuo3Init,
                dim3166Kuo3Init,
                dim3167Kuo3Init,
                dim3168Kuo3Init,
                dim3169Kuo3Init,
                dim3170Kuo3Init,
                dim3171Kuo3Init,
                dim3172Kuo3Init,
                dim3173Kuo3Init,
                dim3174Kuo3Init,
                dim3175Kuo3Init,
                dim3176Kuo3Init,
                dim3177Kuo3Init,
                dim3178Kuo3Init,
                dim3179Kuo3Init,
                dim3180Kuo3Init,
                dim3181Kuo3Init,
                dim3182Kuo3Init,
                dim3183Kuo3Init,
                dim3184Kuo3Init,
                dim3185Kuo3Init,
                dim3186Kuo3Init,
                dim3187Kuo3Init,
                dim3188Kuo3Init,
                dim3189Kuo3Init,
                dim3190Kuo3Init,
                dim3191Kuo3Init,
                dim3192Kuo3Init,
                dim3193Kuo3Init,
                dim3194Kuo3Init,
                dim3195Kuo3Init,
                dim3196Kuo3Init,
                dim3197Kuo3Init,
                dim3198Kuo3Init,
                dim3199Kuo3Init,
                dim3200Kuo3Init,
                dim3201Kuo3Init,
                dim3202Kuo3Init,
                dim3203Kuo3Init,
                dim3204Kuo3Init,
                dim3205Kuo3Init,
                dim3206Kuo3Init,
                dim3207Kuo3Init,
                dim3208Kuo3Init,
                dim3209Kuo3Init,
                dim3210Kuo3Init,
                dim3211Kuo3Init,
                dim3212Kuo3Init,
                dim3213Kuo3Init,
                dim3214Kuo3Init,
                dim3215Kuo3Init,
                dim3216Kuo3Init,
                dim3217Kuo3Init,
                dim3218Kuo3Init,
                dim3219Kuo3Init,
                dim3220Kuo3Init,
                dim3221Kuo3Init,
                dim3222Kuo3Init,
                dim3223Kuo3Init,
                dim3224Kuo3Init,
                dim3225Kuo3Init,
                dim3226Kuo3Init,
                dim3227Kuo3Init,
                dim3228Kuo3Init,
                dim3229Kuo3Init,
                dim3230Kuo3Init,
                dim3231Kuo3Init,
                dim3232Kuo3Init,
                dim3233Kuo3Init,
                dim3234Kuo3Init,
                dim3235Kuo3Init,
                dim3236Kuo3Init,
                dim3237Kuo3Init,
                dim3238Kuo3Init,
                dim3239Kuo3Init,
                dim3240Kuo3Init,
                dim3241Kuo3Init,
                dim3242Kuo3Init,
                dim3243Kuo3Init,
                dim3244Kuo3Init,
                dim3245Kuo3Init,
                dim3246Kuo3Init,
                dim3247Kuo3Init,
                dim3248Kuo3Init,
                dim3249Kuo3Init,
                dim3250Kuo3Init,
                dim3251Kuo3Init,
                dim3252Kuo3Init,
                dim3253Kuo3Init,
                dim3254Kuo3Init,
                dim3255Kuo3Init,
                dim3256Kuo3Init,
                dim3257Kuo3Init,
                dim3258Kuo3Init,
                dim3259Kuo3Init,
                dim3260Kuo3Init,
                dim3261Kuo3Init,
                dim3262Kuo3Init,
                dim3263Kuo3Init,
                dim3264Kuo3Init,
                dim3265Kuo3Init,
                dim3266Kuo3Init,
                dim3267Kuo3Init,
                dim3268Kuo3Init,
                dim3269Kuo3Init,
                dim3270Kuo3Init,
                dim3271Kuo3Init,
                dim3272Kuo3Init,
                dim3273Kuo3Init,
                dim3274Kuo3Init,
                dim3275Kuo3Init,
                dim3276Kuo3Init,
                dim3277Kuo3Init,
                dim3278Kuo3Init,
                dim3279Kuo3Init,
                dim3280Kuo3Init,
                dim3281Kuo3Init,
                dim3282Kuo3Init,
                dim3283Kuo3Init,
                dim3284Kuo3Init,
                dim3285Kuo3Init,
                dim3286Kuo3Init,
                dim3287Kuo3Init,
                dim3288Kuo3Init,
                dim3289Kuo3Init,
                dim3290Kuo3Init,
                dim3291Kuo3Init,
                dim3292Kuo3Init,
                dim3293Kuo3Init,
                dim3294Kuo3Init,
                dim3295Kuo3Init,
                dim3296Kuo3Init,
                dim3297Kuo3Init,
                dim3298Kuo3Init,
                dim3299Kuo3Init,
                dim3300Kuo3Init,
                dim3301Kuo3Init,
                dim3302Kuo3Init,
                dim3303Kuo3Init,
                dim3304Kuo3Init,
                dim3305Kuo3Init,
                dim3306Kuo3Init,
                dim3307Kuo3Init,
                dim3308Kuo3Init,
                dim3309Kuo3Init,
                dim3310Kuo3Init,
                dim3311Kuo3Init,
                dim3312Kuo3Init,
                dim3313Kuo3Init,
                dim3314Kuo3Init,
                dim3315Kuo3Init,
                dim3316Kuo3Init,
                dim3317Kuo3Init,
                dim3318Kuo3Init,
                dim3319Kuo3Init,
                dim3320Kuo3Init,
                dim3321Kuo3Init,
                dim3322Kuo3Init,
                dim3323Kuo3Init,
                dim3324Kuo3Init,
                dim3325Kuo3Init,
                dim3326Kuo3Init,
                dim3327Kuo3Init,
                dim3328Kuo3Init,
                dim3329Kuo3Init,
                dim3330Kuo3Init,
                dim3331Kuo3Init,
                dim3332Kuo3Init,
                dim3333Kuo3Init,
                dim3334Kuo3Init,
                dim3335Kuo3Init,
                dim3336Kuo3Init,
                dim3337Kuo3Init,
                dim3338Kuo3Init,
                dim3339Kuo3Init,
                dim3340Kuo3Init,
                dim3341Kuo3Init,
                dim3342Kuo3Init,
                dim3343Kuo3Init,
                dim3344Kuo3Init,
                dim3345Kuo3Init,
                dim3346Kuo3Init,
                dim3347Kuo3Init,
                dim3348Kuo3Init,
                dim3349Kuo3Init,
                dim3350Kuo3Init,
                dim3351Kuo3Init,
                dim3352Kuo3Init,
                dim3353Kuo3Init,
                dim3354Kuo3Init,
                dim3355Kuo3Init,
                dim3356Kuo3Init,
                dim3357Kuo3Init,
                dim3358Kuo3Init,
                dim3359Kuo3Init,
                dim3360Kuo3Init,
                dim3361Kuo3Init,
                dim3362Kuo3Init,
                dim3363Kuo3Init,
                dim3364Kuo3Init,
                dim3365Kuo3Init,
                dim3366Kuo3Init,
                dim3367Kuo3Init,
                dim3368Kuo3Init,
                dim3369Kuo3Init,
                dim3370Kuo3Init,
                dim3371Kuo3Init,
                dim3372Kuo3Init,
                dim3373Kuo3Init,
                dim3374Kuo3Init,
                dim3375Kuo3Init,
                dim3376Kuo3Init,
                dim3377Kuo3Init,
                dim3378Kuo3Init,
                dim3379Kuo3Init,
                dim3380Kuo3Init,
                dim3381Kuo3Init,
                dim3382Kuo3Init,
                dim3383Kuo3Init,
                dim3384Kuo3Init,
                dim3385Kuo3Init,
                dim3386Kuo3Init,
                dim3387Kuo3Init,
                dim3388Kuo3Init,
                dim3389Kuo3Init,
                dim3390Kuo3Init,
                dim3391Kuo3Init,
                dim3392Kuo3Init,
                dim3393Kuo3Init,
                dim3394Kuo3Init,
                dim3395Kuo3Init,
                dim3396Kuo3Init,
                dim3397Kuo3Init,
                dim3398Kuo3Init,
                dim3399Kuo3Init,
                dim3400Kuo3Init,
                dim3401Kuo3Init,
                dim3402Kuo3Init,
                dim3403Kuo3Init,
                dim3404Kuo3Init,
                dim3405Kuo3Init,
                dim3406Kuo3Init,
                dim3407Kuo3Init,
                dim3408Kuo3Init,
                dim3409Kuo3Init,
                dim3410Kuo3Init,
                dim3411Kuo3Init,
                dim3412Kuo3Init,
                dim3413Kuo3Init,
                dim3414Kuo3Init,
                dim3415Kuo3Init,
                dim3416Kuo3Init,
                dim3417Kuo3Init,
                dim3418Kuo3Init,
                dim3419Kuo3Init,
                dim3420Kuo3Init,
                dim3421Kuo3Init,
                dim3422Kuo3Init,
                dim3423Kuo3Init,
                dim3424Kuo3Init,
                dim3425Kuo3Init,
                dim3426Kuo3Init,
                dim3427Kuo3Init,
                dim3428Kuo3Init,
                dim3429Kuo3Init,
                dim3430Kuo3Init,
                dim3431Kuo3Init,
                dim3432Kuo3Init,
                dim3433Kuo3Init,
                dim3434Kuo3Init,
                dim3435Kuo3Init,
                dim3436Kuo3Init,
                dim3437Kuo3Init,
                dim3438Kuo3Init,
                dim3439Kuo3Init,
                dim3440Kuo3Init,
                dim3441Kuo3Init,
                dim3442Kuo3Init,
                dim3443Kuo3Init,
                dim3444Kuo3Init,
                dim3445Kuo3Init,
                dim3446Kuo3Init,
                dim3447Kuo3Init,
                dim3448Kuo3Init,
                dim3449Kuo3Init,
                dim3450Kuo3Init,
                dim3451Kuo3Init,
                dim3452Kuo3Init,
                dim3453Kuo3Init,
                dim3454Kuo3Init,
                dim3455Kuo3Init,
                dim3456Kuo3Init,
                dim3457Kuo3Init,
                dim3458Kuo3Init,
                dim3459Kuo3Init,
                dim3460Kuo3Init,
                dim3461Kuo3Init,
                dim3462Kuo3Init,
                dim3463Kuo3Init,
                dim3464Kuo3Init,
                dim3465Kuo3Init,
                dim3466Kuo3Init,
                dim3467Kuo3Init,
                dim3468Kuo3Init,
                dim3469Kuo3Init,
                dim3470Kuo3Init,
                dim3471Kuo3Init,
                dim3472Kuo3Init,
                dim3473Kuo3Init,
                dim3474Kuo3Init,
                dim3475Kuo3Init,
                dim3476Kuo3Init,
                dim3477Kuo3Init,
                dim3478Kuo3Init,
                dim3479Kuo3Init,
                dim3480Kuo3Init,
                dim3481Kuo3Init,
                dim3482Kuo3Init,
                dim3483Kuo3Init,
                dim3484Kuo3Init,
                dim3485Kuo3Init,
                dim3486Kuo3Init,
                dim3487Kuo3Init,
                dim3488Kuo3Init,
                dim3489Kuo3Init,
                dim3490Kuo3Init,
                dim3491Kuo3Init,
                dim3492Kuo3Init,
                dim3493Kuo3Init,
                dim3494Kuo3Init,
                dim3495Kuo3Init,
                dim3496Kuo3Init,
                dim3497Kuo3Init,
                dim3498Kuo3Init,
                dim3499Kuo3Init,
                dim3500Kuo3Init,
                dim3501Kuo3Init,
                dim3502Kuo3Init,
                dim3503Kuo3Init,
                dim3504Kuo3Init,
                dim3505Kuo3Init,
                dim3506Kuo3Init,
                dim3507Kuo3Init,
                dim3508Kuo3Init,
                dim3509Kuo3Init,
                dim3510Kuo3Init,
                dim3511Kuo3Init,
                dim3512Kuo3Init,
                dim3513Kuo3Init,
                dim3514Kuo3Init,
                dim3515Kuo3Init,
                dim3516Kuo3Init,
                dim3517Kuo3Init,
                dim3518Kuo3Init,
                dim3519Kuo3Init,
                dim3520Kuo3Init,
                dim3521Kuo3Init,
                dim3522Kuo3Init,
                dim3523Kuo3Init,
                dim3524Kuo3Init,
                dim3525Kuo3Init,
                dim3526Kuo3Init,
                dim3527Kuo3Init,
                dim3528Kuo3Init,
                dim3529Kuo3Init,
                dim3530Kuo3Init,
                dim3531Kuo3Init,
                dim3532Kuo3Init,
                dim3533Kuo3Init,
                dim3534Kuo3Init,
                dim3535Kuo3Init,
                dim3536Kuo3Init,
                dim3537Kuo3Init,
                dim3538Kuo3Init,
                dim3539Kuo3Init,
                dim3540Kuo3Init,
                dim3541Kuo3Init,
                dim3542Kuo3Init,
                dim3543Kuo3Init,
                dim3544Kuo3Init,
                dim3545Kuo3Init,
                dim3546Kuo3Init,
                dim3547Kuo3Init,
                dim3548Kuo3Init,
                dim3549Kuo3Init,
                dim3550Kuo3Init,
                dim3551Kuo3Init,
                dim3552Kuo3Init,
                dim3553Kuo3Init,
                dim3554Kuo3Init,
                dim3555Kuo3Init,
                dim3556Kuo3Init,
                dim3557Kuo3Init,
                dim3558Kuo3Init,
                dim3559Kuo3Init,
                dim3560Kuo3Init,
                dim3561Kuo3Init,
                dim3562Kuo3Init,
                dim3563Kuo3Init,
                dim3564Kuo3Init,
                dim3565Kuo3Init,
                dim3566Kuo3Init,
                dim3567Kuo3Init,
                dim3568Kuo3Init,
                dim3569Kuo3Init,
                dim3570Kuo3Init,
                dim3571Kuo3Init,
                dim3572Kuo3Init,
                dim3573Kuo3Init,
                dim3574Kuo3Init,
                dim3575Kuo3Init,
                dim3576Kuo3Init,
                dim3577Kuo3Init,
                dim3578Kuo3Init,
                dim3579Kuo3Init,
                dim3580Kuo3Init,
                dim3581Kuo3Init,
                dim3582Kuo3Init,
                dim3583Kuo3Init,
                dim3584Kuo3Init,
                dim3585Kuo3Init,
                dim3586Kuo3Init,
                dim3587Kuo3Init,
                dim3588Kuo3Init,
                dim3589Kuo3Init,
                dim3590Kuo3Init,
                dim3591Kuo3Init,
                dim3592Kuo3Init,
                dim3593Kuo3Init,
                dim3594Kuo3Init,
                dim3595Kuo3Init,
                dim3596Kuo3Init,
                dim3597Kuo3Init,
                dim3598Kuo3Init,
                dim3599Kuo3Init,
                dim3600Kuo3Init,
                dim3601Kuo3Init,
                dim3602Kuo3Init,
                dim3603Kuo3Init,
                dim3604Kuo3Init,
                dim3605Kuo3Init,
                dim3606Kuo3Init,
                dim3607Kuo3Init,
                dim3608Kuo3Init,
                dim3609Kuo3Init,
                dim3610Kuo3Init,
                dim3611Kuo3Init,
                dim3612Kuo3Init,
                dim3613Kuo3Init,
                dim3614Kuo3Init,
                dim3615Kuo3Init,
                dim3616Kuo3Init,
                dim3617Kuo3Init,
                dim3618Kuo3Init,
                dim3619Kuo3Init,
                dim3620Kuo3Init,
                dim3621Kuo3Init,
                dim3622Kuo3Init,
                dim3623Kuo3Init,
                dim3624Kuo3Init,
                dim3625Kuo3Init,
                dim3626Kuo3Init,
                dim3627Kuo3Init,
                dim3628Kuo3Init,
                dim3629Kuo3Init,
                dim3630Kuo3Init,
                dim3631Kuo3Init,
                dim3632Kuo3Init,
                dim3633Kuo3Init,
                dim3634Kuo3Init,
                dim3635Kuo3Init,
                dim3636Kuo3Init,
                dim3637Kuo3Init,
                dim3638Kuo3Init,
                dim3639Kuo3Init,
                dim3640Kuo3Init,
                dim3641Kuo3Init,
                dim3642Kuo3Init,
                dim3643Kuo3Init,
                dim3644Kuo3Init,
                dim3645Kuo3Init,
                dim3646Kuo3Init,
                dim3647Kuo3Init,
                dim3648Kuo3Init,
                dim3649Kuo3Init,
                dim3650Kuo3Init,
                dim3651Kuo3Init,
                dim3652Kuo3Init,
                dim3653Kuo3Init,
                dim3654Kuo3Init,
                dim3655Kuo3Init,
                dim3656Kuo3Init,
                dim3657Kuo3Init,
                dim3658Kuo3Init,
                dim3659Kuo3Init,
                dim3660Kuo3Init,
                dim3661Kuo3Init,
                dim3662Kuo3Init,
                dim3663Kuo3Init,
                dim3664Kuo3Init,
                dim3665Kuo3Init,
                dim3666Kuo3Init,
                dim3667Kuo3Init,
                dim3668Kuo3Init,
                dim3669Kuo3Init,
                dim3670Kuo3Init,
                dim3671Kuo3Init,
                dim3672Kuo3Init,
                dim3673Kuo3Init,
                dim3674Kuo3Init,
                dim3675Kuo3Init,
                dim3676Kuo3Init,
                dim3677Kuo3Init,
                dim3678Kuo3Init,
                dim3679Kuo3Init,
                dim3680Kuo3Init,
                dim3681Kuo3Init,
                dim3682Kuo3Init,
                dim3683Kuo3Init,
                dim3684Kuo3Init,
                dim3685Kuo3Init,
                dim3686Kuo3Init,
                dim3687Kuo3Init,
                dim3688Kuo3Init,
                dim3689Kuo3Init,
                dim3690Kuo3Init,
                dim3691Kuo3Init,
                dim3692Kuo3Init,
                dim3693Kuo3Init,
                dim3694Kuo3Init,
                dim3695Kuo3Init,
                dim3696Kuo3Init,
                dim3697Kuo3Init,
                dim3698Kuo3Init,
                dim3699Kuo3Init,
                dim3700Kuo3Init,
                dim3701Kuo3Init,
                dim3702Kuo3Init,
                dim3703Kuo3Init,
                dim3704Kuo3Init,
                dim3705Kuo3Init,
                dim3706Kuo3Init,
                dim3707Kuo3Init,
                dim3708Kuo3Init,
                dim3709Kuo3Init,
                dim3710Kuo3Init,
                dim3711Kuo3Init,
                dim3712Kuo3Init,
                dim3713Kuo3Init,
                dim3714Kuo3Init,
                dim3715Kuo3Init,
                dim3716Kuo3Init,
                dim3717Kuo3Init,
                dim3718Kuo3Init,
                dim3719Kuo3Init,
                dim3720Kuo3Init,
                dim3721Kuo3Init,
                dim3722Kuo3Init,
                dim3723Kuo3Init,
                dim3724Kuo3Init,
                dim3725Kuo3Init,
                dim3726Kuo3Init,
                dim3727Kuo3Init,
                dim3728Kuo3Init,
                dim3729Kuo3Init,
                dim3730Kuo3Init,
                dim3731Kuo3Init,
                dim3732Kuo3Init,
                dim3733Kuo3Init,
                dim3734Kuo3Init,
                dim3735Kuo3Init,
                dim3736Kuo3Init,
                dim3737Kuo3Init,
                dim3738Kuo3Init,
                dim3739Kuo3Init,
                dim3740Kuo3Init,
                dim3741Kuo3Init,
                dim3742Kuo3Init,
                dim3743Kuo3Init,
                dim3744Kuo3Init,
                dim3745Kuo3Init,
                dim3746Kuo3Init,
                dim3747Kuo3Init,
                dim3748Kuo3Init,
                dim3749Kuo3Init,
                dim3750Kuo3Init,
                dim3751Kuo3Init,
                dim3752Kuo3Init,
                dim3753Kuo3Init,
                dim3754Kuo3Init,
                dim3755Kuo3Init,
                dim3756Kuo3Init,
                dim3757Kuo3Init,
                dim3758Kuo3Init,
                dim3759Kuo3Init,
                dim3760Kuo3Init,
                dim3761Kuo3Init,
                dim3762Kuo3Init,
                dim3763Kuo3Init,
                dim3764Kuo3Init,
                dim3765Kuo3Init,
                dim3766Kuo3Init,
                dim3767Kuo3Init,
                dim3768Kuo3Init,
                dim3769Kuo3Init,
                dim3770Kuo3Init,
                dim3771Kuo3Init,
                dim3772Kuo3Init,
                dim3773Kuo3Init,
                dim3774Kuo3Init,
                dim3775Kuo3Init,
                dim3776Kuo3Init,
                dim3777Kuo3Init,
                dim3778Kuo3Init,
                dim3779Kuo3Init,
                dim3780Kuo3Init,
                dim3781Kuo3Init,
                dim3782Kuo3Init,
                dim3783Kuo3Init,
                dim3784Kuo3Init,
                dim3785Kuo3Init,
                dim3786Kuo3Init,
                dim3787Kuo3Init,
                dim3788Kuo3Init,
                dim3789Kuo3Init,
                dim3790Kuo3Init,
                dim3791Kuo3Init,
                dim3792Kuo3Init,
                dim3793Kuo3Init,
                dim3794Kuo3Init,
                dim3795Kuo3Init,
                dim3796Kuo3Init,
                dim3797Kuo3Init,
                dim3798Kuo3Init,
                dim3799Kuo3Init,
                dim3800Kuo3Init,
                dim3801Kuo3Init,
                dim3802Kuo3Init,
                dim3803Kuo3Init,
                dim3804Kuo3Init,
                dim3805Kuo3Init,
                dim3806Kuo3Init,
                dim3807Kuo3Init,
                dim3808Kuo3Init,
                dim3809Kuo3Init,
                dim3810Kuo3Init,
                dim3811Kuo3Init,
                dim3812Kuo3Init,
                dim3813Kuo3Init,
                dim3814Kuo3Init,
                dim3815Kuo3Init,
                dim3816Kuo3Init,
                dim3817Kuo3Init,
                dim3818Kuo3Init,
                dim3819Kuo3Init,
                dim3820Kuo3Init,
                dim3821Kuo3Init,
                dim3822Kuo3Init,
                dim3823Kuo3Init,
                dim3824Kuo3Init,
                dim3825Kuo3Init,
                dim3826Kuo3Init,
                dim3827Kuo3Init,
                dim3828Kuo3Init,
                dim3829Kuo3Init,
                dim3830Kuo3Init,
                dim3831Kuo3Init,
                dim3832Kuo3Init,
                dim3833Kuo3Init,
                dim3834Kuo3Init,
                dim3835Kuo3Init,
                dim3836Kuo3Init,
                dim3837Kuo3Init,
                dim3838Kuo3Init,
                dim3839Kuo3Init,
                dim3840Kuo3Init,
                dim3841Kuo3Init,
                dim3842Kuo3Init,
                dim3843Kuo3Init,
                dim3844Kuo3Init,
                dim3845Kuo3Init,
                dim3846Kuo3Init,
                dim3847Kuo3Init,
                dim3848Kuo3Init,
                dim3849Kuo3Init,
                dim3850Kuo3Init,
                dim3851Kuo3Init,
                dim3852Kuo3Init,
                dim3853Kuo3Init,
                dim3854Kuo3Init,
                dim3855Kuo3Init,
                dim3856Kuo3Init,
                dim3857Kuo3Init,
                dim3858Kuo3Init,
                dim3859Kuo3Init,
                dim3860Kuo3Init,
                dim3861Kuo3Init,
                dim3862Kuo3Init,
                dim3863Kuo3Init,
                dim3864Kuo3Init,
                dim3865Kuo3Init,
                dim3866Kuo3Init,
                dim3867Kuo3Init,
                dim3868Kuo3Init,
                dim3869Kuo3Init,
                dim3870Kuo3Init,
                dim3871Kuo3Init,
                dim3872Kuo3Init,
                dim3873Kuo3Init,
                dim3874Kuo3Init,
                dim3875Kuo3Init,
                dim3876Kuo3Init,
                dim3877Kuo3Init,
                dim3878Kuo3Init,
                dim3879Kuo3Init,
                dim3880Kuo3Init,
                dim3881Kuo3Init,
                dim3882Kuo3Init,
                dim3883Kuo3Init,
                dim3884Kuo3Init,
                dim3885Kuo3Init,
                dim3886Kuo3Init,
                dim3887Kuo3Init,
                dim3888Kuo3Init,
                dim3889Kuo3Init,
                dim3890Kuo3Init,
                dim3891Kuo3Init,
                dim3892Kuo3Init,
                dim3893Kuo3Init,
                dim3894Kuo3Init,
                dim3895Kuo3Init,
                dim3896Kuo3Init,
                dim3897Kuo3Init,
                dim3898Kuo3Init,
                dim3899Kuo3Init,
                dim3900Kuo3Init,
                dim3901Kuo3Init,
                dim3902Kuo3Init,
                dim3903Kuo3Init,
                dim3904Kuo3Init,
                dim3905Kuo3Init,
                dim3906Kuo3Init,
                dim3907Kuo3Init,
                dim3908Kuo3Init,
                dim3909Kuo3Init,
                dim3910Kuo3Init,
                dim3911Kuo3Init,
                dim3912Kuo3Init,
                dim3913Kuo3Init,
                dim3914Kuo3Init,
                dim3915Kuo3Init,
                dim3916Kuo3Init,
                dim3917Kuo3Init,
                dim3918Kuo3Init,
                dim3919Kuo3Init,
                dim3920Kuo3Init,
                dim3921Kuo3Init,
                dim3922Kuo3Init,
                dim3923Kuo3Init,
                dim3924Kuo3Init,
                dim3925Kuo3Init,
                dim3926Kuo3Init,
                dim3927Kuo3Init,
                dim3928Kuo3Init,
                dim3929Kuo3Init,
                dim3930Kuo3Init,
                dim3931Kuo3Init,
                dim3932Kuo3Init,
                dim3933Kuo3Init,
                dim3934Kuo3Init,
                dim3935Kuo3Init,
                dim3936Kuo3Init,
                dim3937Kuo3Init,
                dim3938Kuo3Init,
                dim3939Kuo3Init,
                dim3940Kuo3Init,
                dim3941Kuo3Init,
                dim3942Kuo3Init,
                dim3943Kuo3Init,
                dim3944Kuo3Init,
                dim3945Kuo3Init,
                dim3946Kuo3Init,
                dim3947Kuo3Init,
                dim3948Kuo3Init,
                dim3949Kuo3Init,
                dim3950Kuo3Init,
                dim3951Kuo3Init,
                dim3952Kuo3Init,
                dim3953Kuo3Init,
                dim3954Kuo3Init,
                dim3955Kuo3Init,
                dim3956Kuo3Init,
                dim3957Kuo3Init,
                dim3958Kuo3Init,
                dim3959Kuo3Init,
                dim3960Kuo3Init,
                dim3961Kuo3Init,
                dim3962Kuo3Init,
                dim3963Kuo3Init,
                dim3964Kuo3Init,
                dim3965Kuo3Init,
                dim3966Kuo3Init,
                dim3967Kuo3Init,
                dim3968Kuo3Init,
                dim3969Kuo3Init,
                dim3970Kuo3Init,
                dim3971Kuo3Init,
                dim3972Kuo3Init,
                dim3973Kuo3Init,
                dim3974Kuo3Init,
                dim3975Kuo3Init,
                dim3976Kuo3Init,
                dim3977Kuo3Init,
                dim3978Kuo3Init,
                dim3979Kuo3Init,
                dim3980Kuo3Init,
                dim3981Kuo3Init,
                dim3982Kuo3Init,
                dim3983Kuo3Init,
                dim3984Kuo3Init,
                dim3985Kuo3Init,
                dim3986Kuo3Init,
                dim3987Kuo3Init,
                dim3988Kuo3Init,
                dim3989Kuo3Init,
                dim3990Kuo3Init,
                dim3991Kuo3Init,
                dim3992Kuo3Init,
                dim3993Kuo3Init,
                dim3994Kuo3Init,
                dim3995Kuo3Init,
                dim3996Kuo3Init,
                dim3997Kuo3Init,
                dim3998Kuo3Init,
                dim3999Kuo3Init,
                dim4000Kuo3Init,
                dim4001Kuo3Init,
                dim4002Kuo3Init,
                dim4003Kuo3Init,
                dim4004Kuo3Init,
                dim4005Kuo3Init,
                dim4006Kuo3Init,
                dim4007Kuo3Init,
                dim4008Kuo3Init,
                dim4009Kuo3Init,
                dim4010Kuo3Init,
                dim4011Kuo3Init,
                dim4012Kuo3Init,
                dim4013Kuo3Init,
                dim4014Kuo3Init,
                dim4015Kuo3Init,
                dim4016Kuo3Init,
                dim4017Kuo3Init,
                dim4018Kuo3Init,
                dim4019Kuo3Init,
                dim4020Kuo3Init,
                dim4021Kuo3Init,
                dim4022Kuo3Init,
                dim4023Kuo3Init,
                dim4024Kuo3Init,
                dim4025Kuo3Init,
                dim4026Kuo3Init,
                dim4027Kuo3Init,
                dim4028Kuo3Init,
                dim4029Kuo3Init,
                dim4030Kuo3Init,
                dim4031Kuo3Init,
                dim4032Kuo3Init,
                dim4033Kuo3Init,
                dim4034Kuo3Init,
                dim4035Kuo3Init,
                dim4036Kuo3Init,
                dim4037Kuo3Init,
                dim4038Kuo3Init,
                dim4039Kuo3Init,
                dim4040Kuo3Init,
                dim4041Kuo3Init,
                dim4042Kuo3Init,
                dim4043Kuo3Init,
                dim4044Kuo3Init,
                dim4045Kuo3Init,
                dim4046Kuo3Init,
                dim4047Kuo3Init,
                dim4048Kuo3Init,
                dim4049Kuo3Init,
                dim4050Kuo3Init,
                dim4051Kuo3Init,
                dim4052Kuo3Init,
                dim4053Kuo3Init,
                dim4054Kuo3Init,
                dim4055Kuo3Init,
                dim4056Kuo3Init,
                dim4057Kuo3Init,
                dim4058Kuo3Init,
                dim4059Kuo3Init,
                dim4060Kuo3Init,
                dim4061Kuo3Init,
                dim4062Kuo3Init,
                dim4063Kuo3Init,
                dim4064Kuo3Init,
                dim4065Kuo3Init,
                dim4066Kuo3Init,
                dim4067Kuo3Init,
                dim4068Kuo3Init,
                dim4069Kuo3Init,
                dim4070Kuo3Init,
                dim4071Kuo3Init,
                dim4072Kuo3Init,
                dim4073Kuo3Init,
                dim4074Kuo3Init,
                dim4075Kuo3Init,
                dim4076Kuo3Init,
                dim4077Kuo3Init,
                dim4078Kuo3Init,
                dim4079Kuo3Init,
                dim4080Kuo3Init,
                dim4081Kuo3Init,
                dim4082Kuo3Init,
                dim4083Kuo3Init,
                dim4084Kuo3Init,
                dim4085Kuo3Init,
                dim4086Kuo3Init,
                dim4087Kuo3Init,
                dim4088Kuo3Init,
                dim4089Kuo3Init,
                dim4090Kuo3Init,
                dim4091Kuo3Init,
                dim4092Kuo3Init,
                dim4093Kuo3Init,
                dim4094Kuo3Init,
                dim4095Kuo3Init,
                dim4096Kuo3Init,
                dim4097Kuo3Init,
                dim4098Kuo3Init,
                dim4099Kuo3Init,
                dim4100Kuo3Init,
                dim4101Kuo3Init,
                dim4102Kuo3Init,
                dim4103Kuo3Init,
                dim4104Kuo3Init,
                dim4105Kuo3Init,
                dim4106Kuo3Init,
                dim4107Kuo3Init,
                dim4108Kuo3Init,
                dim4109Kuo3Init,
                dim4110Kuo3Init,
                dim4111Kuo3Init,
                dim4112Kuo3Init,
                dim4113Kuo3Init,
                dim4114Kuo3Init,
                dim4115Kuo3Init,
                dim4116Kuo3Init,
                dim4117Kuo3Init,
                dim4118Kuo3Init,
                dim4119Kuo3Init,
                dim4120Kuo3Init,
                dim4121Kuo3Init,
                dim4122Kuo3Init,
                dim4123Kuo3Init,
                dim4124Kuo3Init,
                dim4125Kuo3Init,
                dim4126Kuo3Init,
                dim4127Kuo3Init,
                dim4128Kuo3Init,
                dim4129Kuo3Init,
                dim4130Kuo3Init,
                dim4131Kuo3Init,
                dim4132Kuo3Init,
                dim4133Kuo3Init,
                dim4134Kuo3Init,
                dim4135Kuo3Init,
                dim4136Kuo3Init,
                dim4137Kuo3Init,
                dim4138Kuo3Init,
                dim4139Kuo3Init,
                dim4140Kuo3Init,
                dim4141Kuo3Init,
                dim4142Kuo3Init,
                dim4143Kuo3Init,
                dim4144Kuo3Init,
                dim4145Kuo3Init,
                dim4146Kuo3Init,
                dim4147Kuo3Init,
                dim4148Kuo3Init,
                dim4149Kuo3Init,
                dim4150Kuo3Init,
                dim4151Kuo3Init,
                dim4152Kuo3Init,
                dim4153Kuo3Init,
                dim4154Kuo3Init,
                dim4155Kuo3Init,
                dim4156Kuo3Init,
                dim4157Kuo3Init,
                dim4158Kuo3Init,
                dim4159Kuo3Init,
                dim4160Kuo3Init,
                dim4161Kuo3Init,
                dim4162Kuo3Init,
                dim4163Kuo3Init,
                dim4164Kuo3Init,
                dim4165Kuo3Init,
                dim4166Kuo3Init,
                dim4167Kuo3Init,
                dim4168Kuo3Init,
                dim4169Kuo3Init,
                dim4170Kuo3Init,
                dim4171Kuo3Init,
                dim4172Kuo3Init,
                dim4173Kuo3Init,
                dim4174Kuo3Init,
                dim4175Kuo3Init,
                dim4176Kuo3Init,
                dim4177Kuo3Init,
                dim4178Kuo3Init,
                dim4179Kuo3Init,
                dim4180Kuo3Init,
                dim4181Kuo3Init,
                dim4182Kuo3Init,
                dim4183Kuo3Init,
                dim4184Kuo3Init,
                dim4185Kuo3Init,
                dim4186Kuo3Init,
                dim4187Kuo3Init,
                dim4188Kuo3Init,
                dim4189Kuo3Init,
                dim4190Kuo3Init,
                dim4191Kuo3Init,
                dim4192Kuo3Init,
                dim4193Kuo3Init,
                dim4194Kuo3Init,
                dim4195Kuo3Init,
                dim4196Kuo3Init,
                dim4197Kuo3Init,
                dim4198Kuo3Init,
                dim4199Kuo3Init,
                dim4200Kuo3Init,
                dim4201Kuo3Init,
                dim4202Kuo3Init,
                dim4203Kuo3Init,
                dim4204Kuo3Init,
                dim4205Kuo3Init,
                dim4206Kuo3Init,
                dim4207Kuo3Init,
                dim4208Kuo3Init,
                dim4209Kuo3Init,
                dim4210Kuo3Init,
                dim4211Kuo3Init,
                dim4212Kuo3Init,
                dim4213Kuo3Init,
                dim4214Kuo3Init,
                dim4215Kuo3Init,
                dim4216Kuo3Init,
                dim4217Kuo3Init,
                dim4218Kuo3Init,
                dim4219Kuo3Init,
                dim4220Kuo3Init,
                dim4221Kuo3Init,
                dim4222Kuo3Init,
                dim4223Kuo3Init,
                dim4224Kuo3Init,
                dim4225Kuo3Init,
                dim4226Kuo3Init,
                dim4227Kuo3Init,
                dim4228Kuo3Init,
                dim4229Kuo3Init,
                dim4230Kuo3Init,
                dim4231Kuo3Init,
                dim4232Kuo3Init,
                dim4233Kuo3Init,
                dim4234Kuo3Init,
                dim4235Kuo3Init,
                dim4236Kuo3Init,
                dim4237Kuo3Init,
                dim4238Kuo3Init,
                dim4239Kuo3Init,
                dim4240Kuo3Init,
                dim4241Kuo3Init,
                dim4242Kuo3Init,
                dim4243Kuo3Init,
                dim4244Kuo3Init,
                dim4245Kuo3Init,
                dim4246Kuo3Init,
                dim4247Kuo3Init,
                dim4248Kuo3Init,
                dim4249Kuo3Init,
                dim4250Kuo3Init,
                dim4251Kuo3Init,
                dim4252Kuo3Init,
                dim4253Kuo3Init,
                dim4254Kuo3Init,
                dim4255Kuo3Init,
                dim4256Kuo3Init,
                dim4257Kuo3Init,
                dim4258Kuo3Init,
                dim4259Kuo3Init,
                dim4260Kuo3Init,
                dim4261Kuo3Init,
                dim4262Kuo3Init,
                dim4263Kuo3Init,
                dim4264Kuo3Init,
                dim4265Kuo3Init,
                dim4266Kuo3Init,
                dim4267Kuo3Init,
                dim4268Kuo3Init,
                dim4269Kuo3Init,
                dim4270Kuo3Init,
                dim4271Kuo3Init,
                dim4272Kuo3Init,
                dim4273Kuo3Init,
                dim4274Kuo3Init,
                dim4275Kuo3Init,
                dim4276Kuo3Init,
                dim4277Kuo3Init,
                dim4278Kuo3Init,
                dim4279Kuo3Init,
                dim4280Kuo3Init,
                dim4281Kuo3Init,
                dim4282Kuo3Init,
                dim4283Kuo3Init,
                dim4284Kuo3Init,
                dim4285Kuo3Init,
                dim4286Kuo3Init,
                dim4287Kuo3Init,
                dim4288Kuo3Init,
                dim4289Kuo3Init,
                dim4290Kuo3Init,
                dim4291Kuo3Init,
                dim4292Kuo3Init,
                dim4293Kuo3Init,
                dim4294Kuo3Init,
                dim4295Kuo3Init,
                dim4296Kuo3Init,
                dim4297Kuo3Init,
                dim4298Kuo3Init,
                dim4299Kuo3Init,
                dim4300Kuo3Init,
                dim4301Kuo3Init,
                dim4302Kuo3Init,
                dim4303Kuo3Init,
                dim4304Kuo3Init,
                dim4305Kuo3Init,
                dim4306Kuo3Init,
                dim4307Kuo3Init,
                dim4308Kuo3Init,
                dim4309Kuo3Init,
                dim4310Kuo3Init,
                dim4311Kuo3Init,
                dim4312Kuo3Init,
                dim4313Kuo3Init,
                dim4314Kuo3Init,
                dim4315Kuo3Init,
                dim4316Kuo3Init,
                dim4317Kuo3Init,
                dim4318Kuo3Init,
                dim4319Kuo3Init,
                dim4320Kuo3Init,
                dim4321Kuo3Init,
                dim4322Kuo3Init,
                dim4323Kuo3Init,
                dim4324Kuo3Init,
                dim4325Kuo3Init,
                dim4326Kuo3Init,
                dim4327Kuo3Init,
                dim4328Kuo3Init,
                dim4329Kuo3Init,
                dim4330Kuo3Init,
                dim4331Kuo3Init,
                dim4332Kuo3Init,
                dim4333Kuo3Init,
                dim4334Kuo3Init,
                dim4335Kuo3Init,
                dim4336Kuo3Init,
                dim4337Kuo3Init,
                dim4338Kuo3Init,
                dim4339Kuo3Init,
                dim4340Kuo3Init,
                dim4341Kuo3Init,
                dim4342Kuo3Init,
                dim4343Kuo3Init,
                dim4344Kuo3Init,
                dim4345Kuo3Init,
                dim4346Kuo3Init,
                dim4347Kuo3Init,
                dim4348Kuo3Init,
                dim4349Kuo3Init,
                dim4350Kuo3Init,
                dim4351Kuo3Init,
                dim4352Kuo3Init,
                dim4353Kuo3Init,
                dim4354Kuo3Init,
                dim4355Kuo3Init,
                dim4356Kuo3Init,
                dim4357Kuo3Init,
                dim4358Kuo3Init,
                dim4359Kuo3Init,
                dim4360Kuo3Init,
                dim4361Kuo3Init,
                dim4362Kuo3Init,
                dim4363Kuo3Init,
                dim4364Kuo3Init,
                dim4365Kuo3Init,
                dim4366Kuo3Init,
                dim4367Kuo3Init,
                dim4368Kuo3Init,
                dim4369Kuo3Init,
                dim4370Kuo3Init,
                dim4371Kuo3Init,
                dim4372Kuo3Init,
                dim4373Kuo3Init,
                dim4374Kuo3Init,
                dim4375Kuo3Init,
                dim4376Kuo3Init,
                dim4377Kuo3Init,
                dim4378Kuo3Init,
                dim4379Kuo3Init,
                dim4380Kuo3Init,
                dim4381Kuo3Init,
                dim4382Kuo3Init,
                dim4383Kuo3Init,
                dim4384Kuo3Init,
                dim4385Kuo3Init,
                dim4386Kuo3Init,
                dim4387Kuo3Init,
                dim4388Kuo3Init,
                dim4389Kuo3Init,
                dim4390Kuo3Init,
                dim4391Kuo3Init,
                dim4392Kuo3Init,
                dim4393Kuo3Init,
                dim4394Kuo3Init,
                dim4395Kuo3Init,
                dim4396Kuo3Init,
                dim4397Kuo3Init,
                dim4398Kuo3Init,
                dim4399Kuo3Init,
                dim4400Kuo3Init,
                dim4401Kuo3Init,
                dim4402Kuo3Init,
                dim4403Kuo3Init,
                dim4404Kuo3Init,
                dim4405Kuo3Init,
                dim4406Kuo3Init,
                dim4407Kuo3Init,
                dim4408Kuo3Init,
                dim4409Kuo3Init,
                dim4410Kuo3Init,
                dim4411Kuo3Init,
                dim4412Kuo3Init,
                dim4413Kuo3Init,
                dim4414Kuo3Init,
                dim4415Kuo3Init,
                dim4416Kuo3Init,
                dim4417Kuo3Init,
                dim4418Kuo3Init,
                dim4419Kuo3Init,
                dim4420Kuo3Init,
                dim4421Kuo3Init,
                dim4422Kuo3Init,
                dim4423Kuo3Init,
                dim4424Kuo3Init,
                dim4425Kuo3Init,
                dim4426Kuo3Init,
                dim4427Kuo3Init,
                dim4428Kuo3Init,
                dim4429Kuo3Init,
                dim4430Kuo3Init,
                dim4431Kuo3Init,
                dim4432Kuo3Init,
                dim4433Kuo3Init,
                dim4434Kuo3Init,
                dim4435Kuo3Init,
                dim4436Kuo3Init,
                dim4437Kuo3Init,
                dim4438Kuo3Init,
                dim4439Kuo3Init,
                dim4440Kuo3Init,
                dim4441Kuo3Init,
                dim4442Kuo3Init,
                dim4443Kuo3Init,
                dim4444Kuo3Init,
                dim4445Kuo3Init,
                dim4446Kuo3Init,
                dim4447Kuo3Init,
                dim4448Kuo3Init,
                dim4449Kuo3Init,
                dim4450Kuo3Init,
                dim4451Kuo3Init,
                dim4452Kuo3Init,
                dim4453Kuo3Init,
                dim4454Kuo3Init,
                dim4455Kuo3Init,
                dim4456Kuo3Init,
                dim4457Kuo3Init,
                dim4458Kuo3Init,
                dim4459Kuo3Init,
                dim4460Kuo3Init,
                dim4461Kuo3Init,
                dim4462Kuo3Init,
                dim4463Kuo3Init,
                dim4464Kuo3Init,
                dim4465Kuo3Init,
                dim4466Kuo3Init,
                dim4467Kuo3Init,
                dim4468Kuo3Init,
                dim4469Kuo3Init,
                dim4470Kuo3Init,
                dim4471Kuo3Init,
                dim4472Kuo3Init,
                dim4473Kuo3Init,
                dim4474Kuo3Init,
                dim4475Kuo3Init,
                dim4476Kuo3Init,
                dim4477Kuo3Init,
                dim4478Kuo3Init,
                dim4479Kuo3Init,
                dim4480Kuo3Init,
                dim4481Kuo3Init,
                dim4482Kuo3Init,
                dim4483Kuo3Init,
                dim4484Kuo3Init,
                dim4485Kuo3Init,
                dim4486Kuo3Init,
                dim4487Kuo3Init,
                dim4488Kuo3Init,
                dim4489Kuo3Init,
                dim4490Kuo3Init,
                dim4491Kuo3Init,
                dim4492Kuo3Init,
                dim4493Kuo3Init,
                dim4494Kuo3Init,
                dim4495Kuo3Init,
                dim4496Kuo3Init,
                dim4497Kuo3Init,
                dim4498Kuo3Init,
                dim4499Kuo3Init,
                dim4500Kuo3Init,
                dim4501Kuo3Init,
                dim4502Kuo3Init,
                dim4503Kuo3Init,
                dim4504Kuo3Init,
                dim4505Kuo3Init,
                dim4506Kuo3Init,
                dim4507Kuo3Init,
                dim4508Kuo3Init,
                dim4509Kuo3Init,
                dim4510Kuo3Init,
                dim4511Kuo3Init,
                dim4512Kuo3Init,
                dim4513Kuo3Init,
                dim4514Kuo3Init,
                dim4515Kuo3Init,
                dim4516Kuo3Init,
                dim4517Kuo3Init,
                dim4518Kuo3Init,
                dim4519Kuo3Init,
                dim4520Kuo3Init,
                dim4521Kuo3Init,
                dim4522Kuo3Init,
                dim4523Kuo3Init,
                dim4524Kuo3Init,
                dim4525Kuo3Init,
                dim4526Kuo3Init,
                dim4527Kuo3Init,
                dim4528Kuo3Init,
                dim4529Kuo3Init,
                dim4530Kuo3Init,
                dim4531Kuo3Init,
                dim4532Kuo3Init,
                dim4533Kuo3Init,
                dim4534Kuo3Init,
                dim4535Kuo3Init,
                dim4536Kuo3Init,
                dim4537Kuo3Init,
                dim4538Kuo3Init,
                dim4539Kuo3Init,
                dim4540Kuo3Init,
                dim4541Kuo3Init,
                dim4542Kuo3Init,
                dim4543Kuo3Init,
                dim4544Kuo3Init,
                dim4545Kuo3Init,
                dim4546Kuo3Init,
                dim4547Kuo3Init,
                dim4548Kuo3Init,
                dim4549Kuo3Init,
                dim4550Kuo3Init,
                dim4551Kuo3Init,
                dim4552Kuo3Init,
                dim4553Kuo3Init,
                dim4554Kuo3Init,
                dim4555Kuo3Init,
                dim4556Kuo3Init,
                dim4557Kuo3Init,
                dim4558Kuo3Init,
                dim4559Kuo3Init,
                dim4560Kuo3Init,
                dim4561Kuo3Init,
                dim4562Kuo3Init,
                dim4563Kuo3Init,
                dim4564Kuo3Init,
                dim4565Kuo3Init,
                dim4566Kuo3Init,
                dim4567Kuo3Init,
                dim4568Kuo3Init,
                dim4569Kuo3Init,
                dim4570Kuo3Init,
                dim4571Kuo3Init,
                dim4572Kuo3Init,
                dim4573Kuo3Init,
                dim4574Kuo3Init,
                dim4575Kuo3Init,
                dim4576Kuo3Init,
                dim4577Kuo3Init,
                dim4578Kuo3Init,
                dim4579Kuo3Init,
                dim4580Kuo3Init,
                dim4581Kuo3Init,
                dim4582Kuo3Init,
                dim4583Kuo3Init,
                dim4584Kuo3Init,
                dim4585Kuo3Init,
                dim4586Kuo3Init
            };

            const unsigned long dim1JoeKuoD6Init[]  =   {   1   ,0 };
            const unsigned long dim2JoeKuoD6Init[]  =   {   1   ,   3   ,0 };
            const unsigned long dim3JoeKuoD6Init[]  =   {   1   ,   3   ,   1   ,0 };
            const unsigned long dim4JoeKuoD6Init[]  =   {   1   ,   1   ,   1   ,0 };
            const unsigned long dim5JoeKuoD6Init[]  =   {   1   ,   1   ,   3   ,   3   ,0 };
            const unsigned long dim6JoeKuoD6Init[]  =   {   1   ,   3   ,   5   ,   13  ,0 };
            const unsigned long dim7JoeKuoD6Init[]  =   {   1   ,   1   ,   5   ,   5   ,   17  ,0 };
            const unsigned long dim8JoeKuoD6Init[]  =   {   1   ,   1   ,   5   ,   5   ,   5   ,0 };
            const unsigned long dim9JoeKuoD6Init[]  =   {   1   ,   1   ,   7   ,   11  ,   19  ,0 };
            const unsigned long dim10JoeKuoD6Init[] =   {   1   ,   1   ,   5   ,   1   ,   1   ,0 };
            const unsigned long dim11JoeKuoD6Init[] =   {   1   ,   1   ,   1   ,   3   ,   11  ,0 };
            const unsigned long dim12JoeKuoD6Init[] =   {   1   ,   3   ,   5   ,   5   ,   31  ,0 };
            const unsigned long dim13JoeKuoD6Init[] =   {   1   ,   3   ,   3   ,   9   ,   7   ,   49  ,0 };
            const unsigned long dim14JoeKuoD6Init[] =   {   1   ,   1   ,   1   ,   15  ,   21  ,   21  ,0 };
            const unsigned long dim15JoeKuoD6Init[] =   {   1   ,   3   ,   1   ,   13  ,   27  ,   49  ,0 };
            const unsigned long dim16JoeKuoD6Init[] =   {   1   ,   1   ,   1   ,   15  ,   7   ,   5   ,0 };
            const unsigned long dim17JoeKuoD6Init[] =   {   1   ,   3   ,   1   ,   15  ,   13  ,   25  ,0 };
            const unsigned long dim18JoeKuoD6Init[] =   {   1   ,   1   ,   5   ,   5   ,   19  ,   61  ,0 };
            const unsigned long dim19JoeKuoD6Init[] =   {   1   ,   3   ,   7   ,   11  ,   23  ,   15  ,   103 ,0 };
            const unsigned long dim20JoeKuoD6Init[] =   {   1   ,   3   ,   7   ,   13  ,   13  ,   15  ,   69  ,0 };
            const unsigned long dim21JoeKuoD6Init[] =   {   1   ,   1   ,   3   ,   13  ,   7   ,   35  ,   63  ,0 };
            const unsigned long dim22JoeKuoD6Init[] =   {   1   ,   3   ,   5   ,   9   ,   1   ,   25  ,   53  ,0 };
            const unsigned long dim23JoeKuoD6Init[] =   {   1   ,   3   ,   1   ,   13  ,   9   ,   35  ,   107 ,0 };
            const unsigned long dim24JoeKuoD6Init[] =   {   1   ,   3   ,   1   ,   5   ,   27  ,   61  ,   31  ,0 };
            const unsigned long dim25JoeKuoD6Init[] =   {   1   ,   1   ,   5   ,   11  ,   19  ,   41  ,   61  ,0 };
            const unsigned long dim26JoeKuoD6Init[] =   {   1   ,   3   ,   5   ,   3   ,   3   ,   13  ,   69  ,0 };
            const unsigned long dim27JoeKuoD6Init[] =   {   1   ,   1   ,   7   ,   13  ,   1   ,   19  ,   1   ,0 };
            const unsigned long dim28JoeKuoD6Init[] =   {   1   ,   3   ,   7   ,   5   ,   13  ,   19  ,   59  ,0 };
            const unsigned long dim29JoeKuoD6Init[] =   {   1   ,   1   ,   3   ,   9   ,   25  ,   29  ,   41  ,0 };
            const unsigned long dim30JoeKuoD6Init[] =   {   1   ,   3   ,   5   ,   13  ,   23  ,   1   ,   55  ,0 };
            const unsigned long dim31JoeKuoD6Init[] =   {   1   ,   3   ,   7   ,   3   ,   13  ,   59  ,   17  ,0 };
            const unsigned long dim32JoeKuoD6Init[] =   {   1   ,   3   ,   1   ,   3   ,   5   ,   53  ,   69  ,0 };
            const unsigned long dim33JoeKuoD6Init[] =   {   1   ,   1   ,   5   ,   5   ,   23  ,   33  ,   13  ,0 };
            const unsigned long dim34JoeKuoD6Init[] =   {   1   ,   1   ,   7   ,   7   ,   1   ,   61  ,   123 ,0 };
            const unsigned long dim35JoeKuoD6Init[] =   {   1   ,   1   ,   7   ,   9   ,   13  ,   61  ,   49  ,0 };
            const unsigned long dim36JoeKuoD6Init[] =   {   1   ,   3   ,   3   ,   5   ,   3   ,   55  ,   33  ,0 };
            const unsigned long dim37JoeKuoD6Init[] =   {   1   ,   3   ,   1   ,   15  ,   31  ,   13  ,   49  ,   245 ,0 };
            const unsigned long dim38JoeKuoD6Init[] =   {   1   ,   3   ,   5   ,   15  ,   31  ,   59  ,   63  ,   97  ,0 };
            const unsigned long dim39JoeKuoD6Init[] =   {   1   ,   3   ,   1   ,   11  ,   11  ,   11  ,   77  ,   249 ,0 };
            const unsigned long dim40JoeKuoD6Init[] =   {   1   ,   3   ,   1   ,   11  ,   27  ,   43  ,   71  ,   9   ,0 };
            const unsigned long dim41JoeKuoD6Init[] =   {   1   ,   1   ,   7   ,   15  ,   21  ,   11  ,   81  ,   45  ,0 };
            const unsigned long dim42JoeKuoD6Init[] =   {   1   ,   3   ,   7   ,   3   ,   25  ,   31  ,   65  ,   79  ,0 };
            const unsigned long dim43JoeKuoD6Init[] =   {   1   ,   3   ,   1   ,   1   ,   19  ,   11  ,   3   ,   205 ,0 };
            const unsigned long dim44JoeKuoD6Init[] =   {   1   ,   1   ,   5   ,   9   ,   19  ,   21  ,   29  ,   157 ,0 };
            const unsigned long dim45JoeKuoD6Init[] =   {   1   ,   3   ,   7   ,   11  ,   1   ,   33  ,   89  ,   185 ,0 };
            const unsigned long dim46JoeKuoD6Init[] =   {   1   ,   3   ,   3   ,   3   ,   15  ,   9   ,   79  ,   71  ,0 };
            const unsigned long dim47JoeKuoD6Init[] =   {   1   ,   3   ,   7   ,   11  ,   15  ,   39  ,   119 ,   27  ,0 };
            const unsigned long dim48JoeKuoD6Init[] =   {   1   ,   1   ,   3   ,   1   ,   11  ,   31  ,   97  ,   225 ,0 };
            const unsigned long dim49JoeKuoD6Init[] =   {   1   ,   1   ,   1   ,   3   ,   23  ,   43  ,   57  ,   177 ,0 };
            const unsigned long dim50JoeKuoD6Init[] =   {   1   ,   3   ,   7   ,   7   ,   17  ,   17  ,   37  ,   71  ,0 };
            const unsigned long dim51JoeKuoD6Init[] =   {   1   ,   3   ,   1   ,   5   ,   27  ,   63  ,   123 ,   213 ,0 };
            const unsigned long dim52JoeKuoD6Init[] =   {   1   ,   1   ,   3   ,   5   ,   11  ,   43  ,   53  ,   133 ,0 };
            const unsigned long dim53JoeKuoD6Init[] =   {   1   ,   3   ,   5   ,   5   ,   29  ,   17  ,   47  ,   173 ,   479 ,0 };
            const unsigned long dim54JoeKuoD6Init[] =   {   1   ,   3   ,   3   ,   11  ,   3   ,   1   ,   109 ,   9   ,   69  ,0 };
            const unsigned long dim55JoeKuoD6Init[] =   {   1   ,   1   ,   1   ,   5   ,   17  ,   39  ,   23  ,   5   ,   343 ,0 };
            const unsigned long dim56JoeKuoD6Init[] =   {   1   ,   3   ,   1   ,   5   ,   25  ,   15  ,   31  ,   103 ,   499 ,0 };
            const unsigned long dim57JoeKuoD6Init[] =   {   1   ,   1   ,   1   ,   11  ,   11  ,   17  ,   63  ,   105 ,   183 ,0 };
            const unsigned long dim58JoeKuoD6Init[] =   {   1   ,   1   ,   5   ,   11  ,   9   ,   29  ,   97  ,   231 ,   363 ,0 };
            const unsigned long dim59JoeKuoD6Init[] =   {   1   ,   1   ,   5   ,   15  ,   19  ,   45  ,   41  ,   7   ,   383 ,0 };
            const unsigned long dim60JoeKuoD6Init[] =   {   1   ,   3   ,   7   ,   7   ,   31  ,   19  ,   83  ,   137 ,   221 ,0 };
            const unsigned long dim61JoeKuoD6Init[] =   {   1   ,   1   ,   1   ,   3   ,   23  ,   15  ,   111 ,   223 ,   83  ,0 };
            const unsigned long dim62JoeKuoD6Init[] =   {   1   ,   1   ,   5   ,   13  ,   31  ,   15  ,   55  ,   25  ,   161 ,0 };
            const unsigned long dim63JoeKuoD6Init[] =   {   1   ,   1   ,   3   ,   13  ,   25  ,   47  ,   39  ,   87  ,   257 ,0 };
            const unsigned long dim64JoeKuoD6Init[] =   {   1   ,   1   ,   1   ,   11  ,   21  ,   53  ,   125 ,   249 ,   293 ,0 };
            const unsigned long dim65JoeKuoD6Init[] =   {   1   ,   1   ,   7   ,   11  ,   11  ,   7   ,   57  ,   79  ,   323 ,0 };
            const unsigned long dim66JoeKuoD6Init[] =   {   1   ,   1   ,   5   ,   5   ,   17  ,   13  ,   81  ,   3   ,   131 ,0 };
            const unsigned long dim67JoeKuoD6Init[] =   {   1   ,   1   ,   7   ,   13  ,   23  ,   7   ,   65  ,   251 ,   475 ,0 };
            const unsigned long dim68JoeKuoD6Init[] =   {   1   ,   3   ,   5   ,   1   ,   9   ,   43  ,   3   ,   149 ,   11  ,0 };
            const unsigned long dim69JoeKuoD6Init[] =   {   1   ,   1   ,   3   ,   13  ,   31  ,   13  ,   13  ,   255 ,   487 ,0 };
            const unsigned long dim70JoeKuoD6Init[] =   {   1   ,   3   ,   3   ,   1   ,   5   ,   63  ,   89  ,   91  ,   127 ,0 };
            const unsigned long dim71JoeKuoD6Init[] =   {   1   ,   1   ,   3   ,   3   ,   1   ,   19  ,   123 ,   127 ,   237 ,0 };
            const unsigned long dim72JoeKuoD6Init[] =   {   1   ,   1   ,   5   ,   7   ,   23  ,   31  ,   37  ,   243 ,   289 ,0 };
            const unsigned long dim73JoeKuoD6Init[] =   {   1   ,   1   ,   5   ,   11  ,   17  ,   53  ,   117 ,   183 ,   491 ,0 };
            const unsigned long dim74JoeKuoD6Init[] =   {   1   ,   1   ,   1   ,   5   ,   1   ,   13  ,   13  ,   209 ,   345 ,0 };
            const unsigned long dim75JoeKuoD6Init[] =   {   1   ,   1   ,   3   ,   15  ,   1   ,   57  ,   115 ,   7   ,   33  ,0 };
            const unsigned long dim76JoeKuoD6Init[] =   {   1   ,   3   ,   1   ,   11  ,   7   ,   43  ,   81  ,   207 ,   175 ,0 };
            const unsigned long dim77JoeKuoD6Init[] =   {   1   ,   3   ,   1   ,   1   ,   15  ,   27  ,   63  ,   255 ,   49  ,0 };
            const unsigned long dim78JoeKuoD6Init[] =   {   1   ,   3   ,   5   ,   3   ,   27  ,   61  ,   105 ,   171 ,   305 ,0 };
            const unsigned long dim79JoeKuoD6Init[] =   {   1   ,   1   ,   5   ,   3   ,   1   ,   3   ,   57  ,   249 ,   149 ,0 };
            const unsigned long dim80JoeKuoD6Init[] =   {   1   ,   1   ,   3   ,   5   ,   5   ,   57  ,   15  ,   13  ,   159 ,0 };
            const unsigned long dim81JoeKuoD6Init[] =   {   1   ,   1   ,   1   ,   11  ,   7   ,   11  ,   105 ,   141 ,   225 ,0 };
            const unsigned long dim82JoeKuoD6Init[] =   {   1   ,   3   ,   3   ,   5   ,   27  ,   59  ,   121 ,   101 ,   271 ,0 };
            const unsigned long dim83JoeKuoD6Init[] =   {   1   ,   3   ,   5   ,   9   ,   11  ,   49  ,   51  ,   59  ,   115 ,0 };
            const unsigned long dim84JoeKuoD6Init[] =   {   1   ,   1   ,   7   ,   1   ,   23  ,   45  ,   125 ,   71  ,   419 ,0 };
            const unsigned long dim85JoeKuoD6Init[] =   {   1   ,   1   ,   3   ,   5   ,   23  ,   5   ,   105 ,   109 ,   75  ,0 };
            const unsigned long dim86JoeKuoD6Init[] =   {   1   ,   1   ,   7   ,   15  ,   7   ,   11  ,   67  ,   121 ,   453 ,0 };
            const unsigned long dim87JoeKuoD6Init[] =   {   1   ,   3   ,   7   ,   3   ,   9   ,   13  ,   31  ,   27  ,   449 ,0 };
            const unsigned long dim88JoeKuoD6Init[] =   {   1   ,   3   ,   1   ,   15  ,   19  ,   39  ,   39  ,   89  ,   15  ,0 };
            const unsigned long dim89JoeKuoD6Init[] =   {   1   ,   1   ,   1   ,   1   ,   1   ,   33  ,   73  ,   145 ,   379 ,0 };
            const unsigned long dim90JoeKuoD6Init[] =   {   1   ,   3   ,   1   ,   15  ,   15  ,   43  ,   29  ,   13  ,   483 ,0 };
            const unsigned long dim91JoeKuoD6Init[] =   {   1   ,   1   ,   7   ,   3   ,   19  ,   27  ,   85  ,   131 ,   431 ,0 };
            const unsigned long dim92JoeKuoD6Init[] =   {   1   ,   3   ,   3   ,   3   ,   5   ,   35  ,   23  ,   195 ,   349 ,0 };
            const unsigned long dim93JoeKuoD6Init[] =   {   1   ,   3   ,   3   ,   7   ,   9   ,   27  ,   39  ,   59  ,   297 ,0 };
            const unsigned long dim94JoeKuoD6Init[] =   {   1   ,   1   ,   3   ,   9   ,   11  ,   17  ,   13  ,   241 ,   157 ,0 };
            const unsigned long dim95JoeKuoD6Init[] =   {   1   ,   3   ,   7   ,   15  ,   25  ,   57  ,   33  ,   189 ,   213 ,0 };
            const unsigned long dim96JoeKuoD6Init[] =   {   1   ,   1   ,   7   ,   1   ,   9   ,   55  ,   73  ,   83  ,   217 ,0 };
            const unsigned long dim97JoeKuoD6Init[] =   {   1   ,   3   ,   3   ,   13  ,   19  ,   27  ,   23  ,   113 ,   249 ,0 };
            const unsigned long dim98JoeKuoD6Init[] =   {   1   ,   3   ,   5   ,   3   ,   23  ,   43  ,   3   ,   253 ,   479 ,0 };
            const unsigned long dim99JoeKuoD6Init[] =   {   1   ,   1   ,   5   ,   5   ,   11  ,   5   ,   45  ,   117 ,   217 ,0 };
            const unsigned long dim100JoeKuoD6Init[]    =   {   1   ,   3   ,   3   ,   7   ,   29  ,   37  ,   33  ,   123 ,   147 ,0 };
            const unsigned long dim101JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   15  ,   5   ,   5   ,   37  ,   227 ,   223 ,   459 ,0 };
            const unsigned long dim102JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   5   ,   5   ,   39  ,   63  ,   255 ,   135 ,   487 ,0 };
            const unsigned long dim103JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   7   ,   9   ,   7   ,   87  ,   249 ,   217 ,   599 ,0 };
            const unsigned long dim104JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   13  ,   9   ,   47  ,   7   ,   225 ,   363 ,   247 ,0 };
            const unsigned long dim105JoeKuoD6Init[]    =   {   1   ,   3   ,   7   ,   13  ,   19  ,   13  ,   9   ,   67  ,   9   ,   737 ,0 };
            const unsigned long dim106JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   5   ,   19  ,   59  ,   7   ,   41  ,   319 ,   677 ,0 };
            const unsigned long dim107JoeKuoD6Init[]    =   {   1   ,   1   ,   5   ,   3   ,   31  ,   63  ,   15  ,   43  ,   207 ,   789 ,0 };
            const unsigned long dim108JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   9   ,   13  ,   39  ,   3   ,   47  ,   497 ,   169 ,0 };
            const unsigned long dim109JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   7   ,   21  ,   17  ,   97  ,   19  ,   415 ,   905 ,0 };
            const unsigned long dim110JoeKuoD6Init[]    =   {   1   ,   3   ,   7   ,   1   ,   3   ,   31  ,   71  ,   111 ,   165 ,   127 ,0 };
            const unsigned long dim111JoeKuoD6Init[]    =   {   1   ,   1   ,   5   ,   11  ,   1   ,   61  ,   83  ,   119 ,   203 ,   847 ,0 };
            const unsigned long dim112JoeKuoD6Init[]    =   {   1   ,   3   ,   3   ,   13  ,   9   ,   61  ,   19  ,   97  ,   47  ,   35  ,0 };
            const unsigned long dim113JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   7   ,   15  ,   29  ,   63  ,   95  ,   417 ,   469 ,0 };
            const unsigned long dim114JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   9   ,   25  ,   9   ,   71  ,   57  ,   213 ,   385 ,0 };
            const unsigned long dim115JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   13  ,   31  ,   47  ,   101 ,   57  ,   39  ,   341 ,0 };
            const unsigned long dim116JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   3   ,   31  ,   57  ,   125 ,   173 ,   365 ,   551 ,0 };
            const unsigned long dim117JoeKuoD6Init[]    =   {   1   ,   3   ,   7   ,   1   ,   13  ,   57  ,   67  ,   157 ,   451 ,   707 ,0 };
            const unsigned long dim118JoeKuoD6Init[]    =   {   1   ,   1   ,   1   ,   7   ,   21  ,   13  ,   105 ,   89  ,   429 ,   965 ,0 };
            const unsigned long dim119JoeKuoD6Init[]    =   {   1   ,   1   ,   5   ,   9   ,   17  ,   51  ,   45  ,   119 ,   157 ,   141 ,0 };
            const unsigned long dim120JoeKuoD6Init[]    =   {   1   ,   3   ,   7   ,   7   ,   13  ,   45  ,   91  ,   9   ,   129 ,   741 ,0 };
            const unsigned long dim121JoeKuoD6Init[]    =   {   1   ,   3   ,   7   ,   1   ,   23  ,   57  ,   67  ,   141 ,   151 ,   571 ,0 };
            const unsigned long dim122JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   11  ,   17  ,   47  ,   93  ,   107 ,   375 ,   157 ,0 };
            const unsigned long dim123JoeKuoD6Init[]    =   {   1   ,   3   ,   3   ,   5   ,   11  ,   21  ,   43  ,   51  ,   169 ,   915 ,0 };
            const unsigned long dim124JoeKuoD6Init[]    =   {   1   ,   1   ,   5   ,   3   ,   15  ,   55  ,   101 ,   67  ,   455 ,   625 ,0 };
            const unsigned long dim125JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   9   ,   1   ,   23  ,   29  ,   47  ,   345 ,   595 ,0 };
            const unsigned long dim126JoeKuoD6Init[]    =   {   1   ,   3   ,   7   ,   7   ,   5   ,   49  ,   29  ,   155 ,   323 ,   589 ,0 };
            const unsigned long dim127JoeKuoD6Init[]    =   {   1   ,   3   ,   3   ,   7   ,   5   ,   41  ,   127 ,   61  ,   261 ,   717 ,0 };
            const unsigned long dim128JoeKuoD6Init[]    =   {   1   ,   3   ,   7   ,   7   ,   17  ,   23  ,   117 ,   67  ,   129 ,   1009    ,0 };
            const unsigned long dim129JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   13  ,   11  ,   39  ,   21  ,   207 ,   123 ,   305 ,0 };
            const unsigned long dim130JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   9   ,   29  ,   3   ,   95  ,   47  ,   231 ,   73  ,0 };
            const unsigned long dim131JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   9   ,   1   ,   29  ,   117 ,   21  ,   441 ,   259 ,0 };
            const unsigned long dim132JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   13  ,   21  ,   39  ,   125 ,   211 ,   439 ,   723 ,0 };
            const unsigned long dim133JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   3   ,   17  ,   63  ,   115 ,   89  ,   49  ,   773 ,0 };
            const unsigned long dim134JoeKuoD6Init[]    =   {   1   ,   3   ,   7   ,   13  ,   11  ,   33  ,   101 ,   107 ,   63  ,   73  ,0 };
            const unsigned long dim135JoeKuoD6Init[]    =   {   1   ,   1   ,   5   ,   5   ,   13  ,   57  ,   63  ,   135 ,   437 ,   177 ,0 };
            const unsigned long dim136JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   7   ,   27  ,   63  ,   93  ,   47  ,   417 ,   483 ,0 };
            const unsigned long dim137JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   1   ,   23  ,   29  ,   1   ,   191 ,   49  ,   23  ,0 };
            const unsigned long dim138JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   15  ,   25  ,   55  ,   9   ,   101 ,   219 ,   607 ,0 };
            const unsigned long dim139JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   7   ,   7   ,   19  ,   51  ,   251 ,   393 ,   307 ,0 };
            const unsigned long dim140JoeKuoD6Init[]    =   {   1   ,   3   ,   3   ,   3   ,   25  ,   55  ,   17  ,   75  ,   337 ,   3   ,0 };
            const unsigned long dim141JoeKuoD6Init[]    =   {   1   ,   1   ,   1   ,   13  ,   25  ,   17  ,   65  ,   45  ,   479 ,   413 ,0 };
            const unsigned long dim142JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   7   ,   27  ,   49  ,   99  ,   161 ,   213 ,   727 ,0 };
            const unsigned long dim143JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   1   ,   23  ,   5   ,   43  ,   41  ,   251 ,   857 ,0 };
            const unsigned long dim144JoeKuoD6Init[]    =   {   1   ,   3   ,   3   ,   7   ,   11  ,   61  ,   39  ,   87  ,   383 ,   835 ,0 };
            const unsigned long dim145JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   15  ,   13  ,   7   ,   29  ,   7   ,   505 ,   923 ,0 };
            const unsigned long dim146JoeKuoD6Init[]    =   {   1   ,   3   ,   7   ,   1   ,   5   ,   31  ,   47  ,   157 ,   445 ,   501 ,0 };
            const unsigned long dim147JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   7   ,   1   ,   43  ,   9   ,   147 ,   115 ,   605 ,0 };
            const unsigned long dim148JoeKuoD6Init[]    =   {   1   ,   3   ,   3   ,   13  ,   5   ,   1   ,   119 ,   211 ,   455 ,   1001    ,0 };
            const unsigned long dim149JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   5   ,   13  ,   19  ,   3   ,   243 ,   75  ,   843 ,0 };
            const unsigned long dim150JoeKuoD6Init[]    =   {   1   ,   3   ,   7   ,   7   ,   1   ,   19  ,   91  ,   249 ,   357 ,   589 ,0 };
            const unsigned long dim151JoeKuoD6Init[]    =   {   1   ,   1   ,   1   ,   9   ,   1   ,   25  ,   109 ,   197 ,   279 ,   411 ,0 };
            const unsigned long dim152JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   15  ,   23  ,   57  ,   59  ,   135 ,   191 ,   75  ,0 };
            const unsigned long dim153JoeKuoD6Init[]    =   {   1   ,   1   ,   5   ,   15  ,   29  ,   21  ,   39  ,   253 ,   383 ,   349 ,0 };
            const unsigned long dim154JoeKuoD6Init[]    =   {   1   ,   3   ,   3   ,   5   ,   19  ,   45  ,   61  ,   151 ,   199 ,   981 ,0 };
            const unsigned long dim155JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   13  ,   9   ,   61  ,   107 ,   141 ,   141 ,   1   ,0 };
            const unsigned long dim156JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   11  ,   27  ,   25  ,   85  ,   105 ,   309 ,   979 ,0 };
            const unsigned long dim157JoeKuoD6Init[]    =   {   1   ,   3   ,   3   ,   11  ,   19  ,   7   ,   115 ,   223 ,   349 ,   43  ,0 };
            const unsigned long dim158JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   9   ,   21  ,   39  ,   123 ,   21  ,   275 ,   927 ,0 };
            const unsigned long dim159JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   13  ,   15  ,   41  ,   47  ,   243 ,   303 ,   437 ,0 };
            const unsigned long dim160JoeKuoD6Init[]    =   {   1   ,   1   ,   1   ,   7   ,   7   ,   3   ,   15  ,   99  ,   409 ,   719 ,0 };
            const unsigned long dim161JoeKuoD6Init[]    =   {   1   ,   3   ,   3   ,   15  ,   27  ,   49  ,   113 ,   123 ,   113 ,   67  ,   469 ,0 };
            const unsigned long dim162JoeKuoD6Init[]    =   {   1   ,   3   ,   7   ,   11  ,   3   ,   23  ,   87  ,   169 ,   119 ,   483 ,   199 ,0 };
            const unsigned long dim163JoeKuoD6Init[]    =   {   1   ,   1   ,   5   ,   15  ,   7   ,   17  ,   109 ,   229 ,   179 ,   213 ,   741 ,0 };
            const unsigned long dim164JoeKuoD6Init[]    =   {   1   ,   1   ,   5   ,   13  ,   11  ,   17  ,   25  ,   135 ,   403 ,   557 ,   1433    ,0 };
            const unsigned long dim165JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   1   ,   1   ,   61  ,   67  ,   215 ,   189 ,   945 ,   1243    ,0 };
            const unsigned long dim166JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   13  ,   17  ,   33  ,   9   ,   221 ,   429 ,   217 ,   1679    ,0 };
            const unsigned long dim167JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   11  ,   27  ,   3   ,   15  ,   93  ,   93  ,   865 ,   1049    ,0 };
            const unsigned long dim168JoeKuoD6Init[]    =   {   1   ,   3   ,   7   ,   7   ,   25  ,   41  ,   121 ,   35  ,   373 ,   379 ,   1547    ,0 };
            const unsigned long dim169JoeKuoD6Init[]    =   {   1   ,   3   ,   3   ,   9   ,   11  ,   35  ,   45  ,   205 ,   241 ,   9   ,   59  ,0 };
            const unsigned long dim170JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   7   ,   3   ,   51  ,   7   ,   177 ,   53  ,   975 ,   89  ,0 };
            const unsigned long dim171JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   5   ,   27  ,   1   ,   113 ,   231 ,   299 ,   759 ,   861 ,0 };
            const unsigned long dim172JoeKuoD6Init[]    =   {   1   ,   3   ,   3   ,   15  ,   25  ,   29  ,   5   ,   255 ,   139 ,   891 ,   2031    ,0 };
            const unsigned long dim173JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   1   ,   13  ,   9   ,   109 ,   193 ,   419 ,   95  ,   17  ,0 };
            const unsigned long dim174JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   9   ,   3   ,   7   ,   29  ,   41  ,   135 ,   839 ,   867 ,0 };
            const unsigned long dim175JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   9   ,   25  ,   49  ,   123 ,   217 ,   113 ,   909 ,   215 ,0 };
            const unsigned long dim176JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   3   ,   23  ,   15  ,   43  ,   133 ,   217 ,   327 ,   901 ,0 };
            const unsigned long dim177JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   3   ,   13  ,   53  ,   63  ,   123 ,   477 ,   711 ,   1387    ,0 };
            const unsigned long dim178JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   15  ,   7   ,   29  ,   75  ,   119 ,   181 ,   957 ,   247 ,0 };
            const unsigned long dim179JoeKuoD6Init[]    =   {   1   ,   1   ,   1   ,   11  ,   27  ,   25  ,   109 ,   151 ,   267 ,   99  ,   1461    ,0 };
            const unsigned long dim180JoeKuoD6Init[]    =   {   1   ,   3   ,   7   ,   15  ,   5   ,   5   ,   53  ,   145 ,   11  ,   725 ,   1501    ,0 };
            const unsigned long dim181JoeKuoD6Init[]    =   {   1   ,   3   ,   7   ,   1   ,   9   ,   43  ,   71  ,   229 ,   157 ,   607 ,   1835    ,0 };
            const unsigned long dim182JoeKuoD6Init[]    =   {   1   ,   3   ,   3   ,   13  ,   25  ,   1   ,   5   ,   27  ,   471 ,   349 ,   127 ,0 };
            const unsigned long dim183JoeKuoD6Init[]    =   {   1   ,   1   ,   1   ,   1   ,   23  ,   37  ,   9   ,   221 ,   269 ,   897 ,   1685    ,0 };
            const unsigned long dim184JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   3   ,   31  ,   29  ,   51  ,   19  ,   311 ,   553 ,   1969    ,0 };
            const unsigned long dim185JoeKuoD6Init[]    =   {   1   ,   3   ,   7   ,   5   ,   5   ,   55  ,   17  ,   39  ,   475 ,   671 ,   1529    ,0 };
            const unsigned long dim186JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   1   ,   1   ,   35  ,   47  ,   27  ,   437 ,   395 ,   1635    ,0 };
            const unsigned long dim187JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   3   ,   13  ,   23  ,   43  ,   135 ,   327 ,   139 ,   389 ,0 };
            const unsigned long dim188JoeKuoD6Init[]    =   {   1   ,   3   ,   7   ,   3   ,   9   ,   25  ,   91  ,   25  ,   429 ,   219 ,   513 ,0 };
            const unsigned long dim189JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   5   ,   13  ,   29  ,   119 ,   201 ,   277 ,   157 ,   2043    ,0 };
            const unsigned long dim190JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   3   ,   29  ,   57  ,   13  ,   17  ,   167 ,   739 ,   1031    ,0 };
            const unsigned long dim191JoeKuoD6Init[]    =   {   1   ,   3   ,   3   ,   5   ,   29  ,   21  ,   95  ,   27  ,   255 ,   679 ,   1531    ,0 };
            const unsigned long dim192JoeKuoD6Init[]    =   {   1   ,   3   ,   7   ,   15  ,   9   ,   5   ,   21  ,   71  ,   61  ,   961 ,   1201    ,0 };
            const unsigned long dim193JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   13  ,   15  ,   57  ,   33  ,   93  ,   459 ,   867 ,   223 ,0 };
            const unsigned long dim194JoeKuoD6Init[]    =   {   1   ,   1   ,   1   ,   15  ,   17  ,   43  ,   127 ,   191 ,   67  ,   177 ,   1073    ,0 };
            const unsigned long dim195JoeKuoD6Init[]    =   {   1   ,   1   ,   1   ,   15  ,   23  ,   7   ,   21  ,   199 ,   75  ,   293 ,   1611    ,0 };
            const unsigned long dim196JoeKuoD6Init[]    =   {   1   ,   3   ,   7   ,   13  ,   15  ,   39  ,   21  ,   149 ,   65  ,   741 ,   319 ,0 };
            const unsigned long dim197JoeKuoD6Init[]    =   {   1   ,   3   ,   7   ,   11  ,   23  ,   13  ,   101 ,   89  ,   277 ,   519 ,   711 ,0 };
            const unsigned long dim198JoeKuoD6Init[]    =   {   1   ,   3   ,   7   ,   15  ,   19  ,   27  ,   85  ,   203 ,   441 ,   97  ,   1895    ,0 };
            const unsigned long dim199JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   3   ,   29  ,   25  ,   21  ,   155 ,   11  ,   191 ,   197 ,0 };
            const unsigned long dim200JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   5   ,   27  ,   11  ,   81  ,   101 ,   457 ,   675 ,   1687    ,0 };
            const unsigned long dim201JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   5   ,   25  ,   5   ,   65  ,   193 ,   41  ,   567 ,   781 ,0 };
            const unsigned long dim202JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   5   ,   11  ,   15  ,   113 ,   77  ,   411 ,   695 ,   1111    ,0 };
            const unsigned long dim203JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   9   ,   11  ,   53  ,   119 ,   171 ,   55  ,   297 ,   509 ,0 };
            const unsigned long dim204JoeKuoD6Init[]    =   {   1   ,   1   ,   1   ,   1   ,   11  ,   39  ,   113 ,   139 ,   165 ,   347 ,   595 ,0 };
            const unsigned long dim205JoeKuoD6Init[]    =   {   1   ,   3   ,   7   ,   11  ,   9   ,   17  ,   101 ,   13  ,   81  ,   325 ,   1733    ,0 };
            const unsigned long dim206JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   1   ,   21  ,   43  ,   115 ,   9   ,   113 ,   907 ,   645 ,0 };
            const unsigned long dim207JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   3   ,   9   ,   25  ,   117 ,   197 ,   159 ,   471 ,   475 ,0 };
            const unsigned long dim208JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   9   ,   11  ,   21  ,   57  ,   207 ,   485 ,   613 ,   1661    ,0 };
            const unsigned long dim209JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   7   ,   27  ,   55  ,   49  ,   223 ,   89  ,   85  ,   1523    ,0 };
            const unsigned long dim210JoeKuoD6Init[]    =   {   1   ,   1   ,   5   ,   3   ,   19  ,   41  ,   45  ,   51  ,   447 ,   299 ,   1355    ,0 };
            const unsigned long dim211JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   15  ,   1   ,   35  ,   119 ,   135 ,   427 ,   93  ,   1311    ,0 };
            const unsigned long dim212JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   15  ,   31  ,   39  ,   121 ,   145 ,   129 ,   903 ,   333 ,0 };
            const unsigned long dim213JoeKuoD6Init[]    =   {   1   ,   1   ,   5   ,   3   ,   29  ,   7   ,   9   ,   11  ,   395 ,   119 ,   873 ,0 };
            const unsigned long dim214JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   11  ,   27  ,   29  ,   5   ,   117 ,   469 ,   31  ,   1811    ,0 };
            const unsigned long dim215JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   15  ,   21  ,   25  ,   71  ,   15  ,   89  ,   1021    ,   679 ,0 };
            const unsigned long dim216JoeKuoD6Init[]    =   {   1   ,   3   ,   7   ,   9   ,   21  ,   39  ,   51  ,   111 ,   327 ,   809 ,   845 ,0 };
            const unsigned long dim217JoeKuoD6Init[]    =   {   1   ,   3   ,   3   ,   13  ,   3   ,   33  ,   79  ,   17  ,   183 ,   599 ,   1859    ,0 };
            const unsigned long dim218JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   9   ,   31  ,   9   ,   1   ,   5   ,   353 ,   395 ,   833 ,0 };
            const unsigned long dim219JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   7   ,   3   ,   53  ,   81  ,   133 ,   155 ,   355 ,   1559    ,0 };
            const unsigned long dim220JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   5   ,   31  ,   13  ,   89  ,   195 ,   357 ,   769 ,   889 ,0 };
            const unsigned long dim221JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   1   ,   9   ,   21  ,   77  ,   31  ,   269 ,   857 ,   923 ,0 };
            const unsigned long dim222JoeKuoD6Init[]    =   {   1   ,   1   ,   1   ,   11  ,   11  ,   27  ,   61  ,   185 ,   329 ,   685 ,   325 ,0 };
            const unsigned long dim223JoeKuoD6Init[]    =   {   1   ,   1   ,   1   ,   5   ,   11  ,   31  ,   71  ,   79  ,   293 ,   29  ,   17  ,0 };
            const unsigned long dim224JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   7   ,   7   ,   41  ,   65  ,   167 ,   67  ,   153 ,   1305    ,0 };
            const unsigned long dim225JoeKuoD6Init[]    =   {   1   ,   3   ,   3   ,   5   ,   23  ,   45  ,   109 ,   205 ,   191 ,   649 ,   749 ,0 };
            const unsigned long dim226JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   7   ,   17  ,   1   ,   73  ,   211 ,   19  ,   797 ,   1009    ,0 };
            const unsigned long dim227JoeKuoD6Init[]    =   {   1   ,   1   ,   5   ,   9   ,   23  ,   15  ,   117 ,   213 ,   183 ,   477 ,   1301    ,0 };
            const unsigned long dim228JoeKuoD6Init[]    =   {   1   ,   1   ,   5   ,   13  ,   23  ,   35  ,   55  ,   151 ,   267 ,   883 ,   1479    ,0 };
            const unsigned long dim229JoeKuoD6Init[]    =   {   1   ,   1   ,   1   ,   7   ,   5   ,   43  ,   77  ,   11  ,   149 ,   297 ,   329 ,0 };
            const unsigned long dim230JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   15  ,   3   ,   41  ,   55  ,   123 ,   219 ,   499 ,   377 ,0 };
            const unsigned long dim231JoeKuoD6Init[]    =   {   1   ,   3   ,   3   ,   1   ,   27  ,   7   ,   113 ,   141 ,   173 ,   379 ,   1087    ,0 };
            const unsigned long dim232JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   9   ,   7   ,   25  ,   45  ,   49  ,   415 ,   329 ,   483 ,0 };
            const unsigned long dim233JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   7   ,   7   ,   15  ,   75  ,   191 ,   305 ,   757 ,   813 ,0 };
            const unsigned long dim234JoeKuoD6Init[]    =   {   1   ,   3   ,   7   ,   5   ,   31  ,   15  ,   21  ,   247 ,   29  ,   297 ,   1015    ,0 };
            const unsigned long dim235JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   13  ,   3   ,   41  ,   69  ,   235 ,   431 ,   515 ,   1889    ,0 };
            const unsigned long dim236JoeKuoD6Init[]    =   {   1   ,   1   ,   5   ,   7   ,   31  ,   5   ,   69  ,   193 ,   229 ,   439 ,   1043    ,0 };
            const unsigned long dim237JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   13  ,   31  ,   33  ,   45  ,   159 ,   477 ,   707 ,   1605    ,0 };
            const unsigned long dim238JoeKuoD6Init[]    =   {   1   ,   1   ,   1   ,   7   ,   23  ,   49  ,   107 ,   223 ,   7   ,   429 ,   2047    ,0 };
            const unsigned long dim239JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   1   ,   31  ,   17  ,   117 ,   133 ,   331 ,   97  ,   833 ,0 };
            const unsigned long dim240JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   11  ,   7   ,   31  ,   103 ,   5   ,   299 ,   601 ,   981 ,0 };
            const unsigned long dim241JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   5   ,   15  ,   57  ,   87  ,   39  ,   217 ,   155 ,   361 ,0 };
            const unsigned long dim242JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   15  ,   29  ,   19  ,   125 ,   131 ,   37  ,   635 ,   1429    ,0 };
            const unsigned long dim243JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   9   ,   29  ,   3   ,   59  ,   107 ,   341 ,   809 ,   101 ,0 };
            const unsigned long dim244JoeKuoD6Init[]    =   {   1   ,   1   ,   5   ,   15  ,   9   ,   37  ,   113 ,   225 ,   139 ,   165 ,   393 ,0 };
            const unsigned long dim245JoeKuoD6Init[]    =   {   1   ,   3   ,   3   ,   13  ,   27  ,   21  ,   35  ,   255 ,   189 ,   45  ,   103 ,0 };
            const unsigned long dim246JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   5   ,   7   ,   59  ,   39  ,   87  ,   319 ,   723 ,   955 ,0 };
            const unsigned long dim247JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   1   ,   9   ,   23  ,   83  ,   47  ,   413 ,   445 ,   787 ,0 };
            const unsigned long dim248JoeKuoD6Init[]    =   {   1   ,   3   ,   3   ,   3   ,   9   ,   27  ,   113 ,   135 ,   279 ,   915 ,   1221    ,0 };
            const unsigned long dim249JoeKuoD6Init[]    =   {   1   ,   3   ,   7   ,   15  ,   3   ,   49  ,   99  ,   67  ,   11  ,   1017    ,   891 ,0 };
            const unsigned long dim250JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   9   ,   31  ,   41  ,   9   ,   205 ,   39  ,   993 ,   105 ,0 };
            const unsigned long dim251JoeKuoD6Init[]    =   {   1   ,   3   ,   3   ,   7   ,   11  ,   45  ,   67  ,   123 ,   303 ,   87  ,   1943    ,0 };
            const unsigned long dim252JoeKuoD6Init[]    =   {   1   ,   1   ,   1   ,   9   ,   1   ,   51  ,   101 ,   253 ,   499 ,   879 ,   833 ,0 };
            const unsigned long dim253JoeKuoD6Init[]    =   {   1   ,   3   ,   7   ,   15  ,   19  ,   51  ,   97  ,   13  ,   499 ,   231 ,   829 ,0 };
            const unsigned long dim254JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   3   ,   25  ,   49  ,   29  ,   165 ,   43  ,   95  ,   779 ,0 };
            const unsigned long dim255JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   11  ,   13  ,   63  ,   1   ,   133 ,   207 ,   349 ,   221 ,0 };
            const unsigned long dim256JoeKuoD6Init[]    =   {   1   ,   3   ,   7   ,   11  ,   23  ,   37  ,   51  ,   123 ,   399 ,   295 ,   1057    ,0 };
            const unsigned long dim257JoeKuoD6Init[]    =   {   1   ,   1   ,   5   ,   11  ,   11  ,   43  ,   97  ,   163 ,   477 ,   777 ,   1217    ,0 };
            const unsigned long dim258JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   13  ,   3   ,   33  ,   65  ,   191 ,   25  ,   993 ,   291 ,0 };
            const unsigned long dim259JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   9   ,   25  ,   55  ,   101 ,   231 ,   219 ,   215 ,   517 ,0 };
            const unsigned long dim260JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   7   ,   17  ,   31  ,   87  ,   1   ,   475 ,   371 ,   771 ,0 };
            const unsigned long dim261JoeKuoD6Init[]    =   {   1   ,   1   ,   1   ,   7   ,   3   ,   25  ,   35  ,   255 ,   211 ,   67  ,   447 ,0 };
            const unsigned long dim262JoeKuoD6Init[]    =   {   1   ,   1   ,   1   ,   13  ,   9   ,   33  ,   103 ,   17  ,   369 ,   455 ,   265 ,0 };
            const unsigned long dim263JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   5   ,   29  ,   61  ,   61  ,   207 ,   345 ,   87  ,   1035    ,0 };
            const unsigned long dim264JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   7   ,   19  ,   21  ,   117 ,   49  ,   191 ,   133 ,   1889    ,0 };
            const unsigned long dim265JoeKuoD6Init[]    =   {   1   ,   3   ,   3   ,   13  ,   5   ,   37  ,   63  ,   229 ,   53  ,   891 ,   1939    ,0 };
            const unsigned long dim266JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   5   ,   1   ,   43  ,   7   ,   1   ,   427 ,   155 ,   1587    ,0 };
            const unsigned long dim267JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   15  ,   15  ,   33  ,   33  ,   233 ,   481 ,   709 ,   905 ,0 };
            const unsigned long dim268JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   13  ,   7   ,   11  ,   125 ,   77  ,   39  ,   55  ,   1745    ,0 };
            const unsigned long dim269JoeKuoD6Init[]    =   {   1   ,   3   ,   3   ,   3   ,   7   ,   49  ,   35  ,   123 ,   267 ,   229 ,   1873    ,0 };
            const unsigned long dim270JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   3   ,   17  ,   11  ,   35  ,   217 ,   215 ,   577 ,   1467    ,0 };
            const unsigned long dim271JoeKuoD6Init[]    =   {   1   ,   3   ,   3   ,   5   ,   7   ,   11  ,   51  ,   27  ,   277 ,   113 ,   767 ,0 };
            const unsigned long dim272JoeKuoD6Init[]    =   {   1   ,   3   ,   3   ,   3   ,   17  ,   1   ,   61  ,   151 ,   227 ,   841 ,   1421    ,0 };
            const unsigned long dim273JoeKuoD6Init[]    =   {   1   ,   3   ,   7   ,   7   ,   7   ,   29  ,   95  ,   219 ,   407 ,   257 ,   417 ,0 };
            const unsigned long dim274JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   13  ,   5   ,   39  ,   63  ,   67  ,   467 ,   99  ,   457 ,0 };
            const unsigned long dim275JoeKuoD6Init[]    =   {   1   ,   3   ,   3   ,   9   ,   15  ,   19  ,   107 ,   217 ,   371 ,   411 ,   867 ,0 };
            const unsigned long dim276JoeKuoD6Init[]    =   {   1   ,   3   ,   3   ,   13  ,   13  ,   1   ,   101 ,   81  ,   69  ,   371 ,   415 ,0 };
            const unsigned long dim277JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   1   ,   5   ,   31  ,   103 ,   125 ,   147 ,   85  ,   1357    ,0 };
            const unsigned long dim278JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   5   ,   25  ,   61  ,   9   ,   83  ,   371 ,   691 ,   1189    ,0 };
            const unsigned long dim279JoeKuoD6Init[]    =   {   1   ,   1   ,   5   ,   3   ,   3   ,   9   ,   1   ,   123 ,   333 ,   741 ,   961 ,0 };
            const unsigned long dim280JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   1   ,   7   ,   37  ,   73  ,   39  ,   101 ,   1005    ,   1361    ,0 };
            const unsigned long dim281JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   15  ,   11  ,   51  ,   29  ,   159 ,   361 ,   685 ,   1643    ,0 };
            const unsigned long dim282JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   3   ,   17  ,   51  ,   79  ,   143 ,   373 ,   425 ,   703 ,0 };
            const unsigned long dim283JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   15  ,   17  ,   21  ,   63  ,   97  ,   335 ,   591 ,   1697    ,0 };
            const unsigned long dim284JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   7   ,   1   ,   57  ,   21  ,   253 ,   199 ,   661 ,   77  ,0 };
            const unsigned long dim285JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   13  ,   27  ,   13  ,   101 ,   247 ,   497 ,   535 ,   273 ,0 };
            const unsigned long dim286JoeKuoD6Init[]    =   {   1   ,   1   ,   1   ,   7   ,   25  ,   5   ,   83  ,   27  ,   107 ,   549 ,   1885    ,0 };
            const unsigned long dim287JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   5   ,   29  ,   45  ,   107 ,   159 ,   59  ,   303 ,   1941    ,0 };
            const unsigned long dim288JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   9   ,   5   ,   43  ,   85  ,   89  ,   151 ,   39  ,   221 ,0 };
            const unsigned long dim289JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   5   ,   5   ,   47  ,   65  ,   205 ,   141 ,   777 ,   1873    ,0 };
            const unsigned long dim290JoeKuoD6Init[]    =   {   1   ,   1   ,   5   ,   15  ,   25  ,   41  ,   121 ,   133 ,   41  ,   171 ,   521 ,0 };
            const unsigned long dim291JoeKuoD6Init[]    =   {   1   ,   3   ,   3   ,   9   ,   19  ,   25  ,   15  ,   161 ,   339 ,   155 ,   413 ,0 };
            const unsigned long dim292JoeKuoD6Init[]    =   {   1   ,   3   ,   7   ,   1   ,   7   ,   1   ,   67  ,   23  ,   453 ,   625 ,   1375    ,0 };
            const unsigned long dim293JoeKuoD6Init[]    =   {   1   ,   1   ,   1   ,   15  ,   5   ,   9   ,   87  ,   89  ,   209 ,   361 ,   1711    ,0 };
            const unsigned long dim294JoeKuoD6Init[]    =   {   1   ,   3   ,   7   ,   11  ,   21  ,   39  ,   87  ,   21  ,   23  ,   727 ,   15  ,0 };
            const unsigned long dim295JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   13  ,   15  ,   15  ,   9   ,   247 ,   73  ,   9   ,   319 ,0 };
            const unsigned long dim296JoeKuoD6Init[]    =   {   1   ,   1   ,   5   ,   7   ,   29  ,   21  ,   13  ,   121 ,   359 ,   679 ,   1659    ,0 };
            const unsigned long dim297JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   13  ,   19  ,   63  ,   79  ,   99  ,   107 ,   13  ,   137 ,0 };
            const unsigned long dim298JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   15  ,   1   ,   43  ,   117 ,   79  ,   305 ,   285 ,   1151    ,0 };
            const unsigned long dim299JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   9   ,   23  ,   9   ,   53  ,   105 ,   205 ,   121 ,   1451    ,0 };
            const unsigned long dim300JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   3   ,   3   ,   63  ,   99  ,   71  ,   495 ,   233 ,   1633    ,0 };
            const unsigned long dim301JoeKuoD6Init[]    =   {   1   ,   3   ,   7   ,   11  ,   17  ,   5   ,   99  ,   211 ,   409 ,   911 ,   1743    ,0 };
            const unsigned long dim302JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   3   ,   1   ,   61  ,   81  ,   107 ,   493 ,   625 ,   847 ,0 };
            const unsigned long dim303JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   13  ,   15  ,   29  ,   13  ,   57  ,   305 ,   409 ,   1267    ,0 };
            const unsigned long dim304JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   13  ,   11  ,   3   ,   121 ,   253 ,   161 ,   891 ,   1463    ,0 };
            const unsigned long dim305JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   13  ,   23  ,   1   ,   125 ,   7   ,   273 ,   221 ,   2001    ,0 };
            const unsigned long dim306JoeKuoD6Init[]    =   {   1   ,   3   ,   3   ,   5   ,   3   ,   49  ,   115 ,   205 ,   11  ,   45  ,   1563    ,0 };
            const unsigned long dim307JoeKuoD6Init[]    =   {   1   ,   3   ,   7   ,   5   ,   21  ,   59  ,   81  ,   191 ,   31  ,   485 ,   743 ,0 };
            const unsigned long dim308JoeKuoD6Init[]    =   {   1   ,   1   ,   5   ,   13  ,   9   ,   9   ,   21  ,   77  ,   487 ,   615 ,   423 ,0 };
            const unsigned long dim309JoeKuoD6Init[]    =   {   1   ,   1   ,   1   ,   3   ,   27  ,   43  ,   53  ,   53  ,   345 ,   605 ,   949 ,0 };
            const unsigned long dim310JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   7   ,   31  ,   49  ,   97  ,   39  ,   279 ,   597 ,   189 ,0 };
            const unsigned long dim311JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   3   ,   17  ,   23  ,   87  ,   89  ,   457 ,   103 ,   1969    ,0 };
            const unsigned long dim312JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   3   ,   21  ,   45  ,   19  ,   127 ,   49  ,   231 ,   273 ,0 };
            const unsigned long dim313JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   11  ,   13  ,   15  ,   127 ,   83  ,   185 ,   553 ,   537 ,0 };
            const unsigned long dim314JoeKuoD6Init[]    =   {   1   ,   3   ,   7   ,   11  ,   19  ,   45  ,   77  ,   123 ,   471 ,   731 ,   605 ,0 };
            const unsigned long dim315JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   5   ,   11  ,   25  ,   123 ,   25  ,   51  ,   501 ,   2025    ,0 };
            const unsigned long dim316JoeKuoD6Init[]    =   {   1   ,   3   ,   3   ,   9   ,   15  ,   17  ,   67  ,   173 ,   93  ,   217 ,   1507    ,0 };
            const unsigned long dim317JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   1   ,   29  ,   7   ,   29  ,   211 ,   369 ,   983 ,   1331    ,0 };
            const unsigned long dim318JoeKuoD6Init[]    =   {   1   ,   1   ,   5   ,   7   ,   17  ,   63  ,   1   ,   9   ,   493 ,   1009    ,   1275    ,0 };
            const unsigned long dim319JoeKuoD6Init[]    =   {   1   ,   1   ,   1   ,   7   ,   31  ,   37  ,   51  ,   255 ,   253 ,   387 ,   591 ,0 };
            const unsigned long dim320JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   1   ,   11  ,   25  ,   41  ,   245 ,   215 ,   329 ,   1877    ,0 };
            const unsigned long dim321JoeKuoD6Init[]    =   {   1   ,   1   ,   5   ,   7   ,   17  ,   21  ,   115 ,   63  ,   31  ,   435 ,   39  ,0 };
            const unsigned long dim322JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   9   ,   1   ,   17  ,   95  ,   139 ,   505 ,   365 ,   825 ,0 };
            const unsigned long dim323JoeKuoD6Init[]    =   {   1   ,   3   ,   7   ,   7   ,   13  ,   31  ,   105 ,   167 ,   203 ,   337 ,   813 ,0 };
            const unsigned long dim324JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   11  ,   3   ,   7   ,   111 ,   91  ,   491 ,   93  ,   1473    ,0 };
            const unsigned long dim325JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   7   ,   29  ,   55  ,   103 ,   193 ,   219 ,   693 ,   1241    ,0 };
            const unsigned long dim326JoeKuoD6Init[]    =   {   1   ,   3   ,   7   ,   9   ,   27  ,   7   ,   37  ,   135 ,   455 ,   87  ,   257 ,0 };
            const unsigned long dim327JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   5   ,   23  ,   19  ,   23  ,   33  ,   417 ,   263 ,   1693    ,0 };
            const unsigned long dim328JoeKuoD6Init[]    =   {   1   ,   3   ,   3   ,   7   ,   29  ,   39  ,   127 ,   221 ,   305 ,   717 ,   989 ,0 };
            const unsigned long dim329JoeKuoD6Init[]    =   {   1   ,   3   ,   3   ,   1   ,   27  ,   21  ,   49  ,   51  ,   355 ,   241 ,   1853    ,0 };
            const unsigned long dim330JoeKuoD6Init[]    =   {   1   ,   1   ,   1   ,   15  ,   21  ,   3   ,   51  ,   137 ,   257 ,   431 ,   1641    ,0 };
            const unsigned long dim331JoeKuoD6Init[]    =   {   1   ,   3   ,   7   ,   13  ,   25  ,   3   ,   69  ,   111 ,   163 ,   955 ,   459 ,0 };
            const unsigned long dim332JoeKuoD6Init[]    =   {   1   ,   3   ,   7   ,   5   ,   5   ,   43  ,   111 ,   53  ,   463 ,   177 ,   605 ,0 };
            const unsigned long dim333JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   11  ,   23  ,   11  ,   79  ,   105 ,   419 ,   645 ,   423 ,0 };
            const unsigned long dim334JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   3   ,   15  ,   5   ,   105 ,   31  ,   117 ,   961 ,   325 ,0 };
            const unsigned long dim335JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   15  ,   21  ,   55  ,   81  ,   199 ,   231 ,   973 ,   373 ,0 };
            const unsigned long dim336JoeKuoD6Init[]    =   {   1   ,   3   ,   3   ,   15  ,   23  ,   17  ,   71  ,   63  ,   59  ,   719 ,   405 ,0 };
            const unsigned long dim337JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   3   ,   9   ,   37  ,   45  ,   205 ,   267 ,   261 ,   1173    ,   3681    ,0 };
            const unsigned long dim338JoeKuoD6Init[]    =   {   1   ,   3   ,   7   ,   11  ,   1   ,   53  ,   91  ,   207 ,   487 ,   593 ,   1397    ,   1255    ,0 };
            const unsigned long dim339JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   7   ,   17  ,   3   ,   111 ,   145 ,   237 ,   667 ,   1199    ,   3793    ,0 };
            const unsigned long dim340JoeKuoD6Init[]    =   {   1   ,   3   ,   7   ,   3   ,   15  ,   59  ,   69  ,   223 ,   509 ,   211 ,   1431    ,   2455    ,0 };
            const unsigned long dim341JoeKuoD6Init[]    =   {   1   ,   3   ,   7   ,   1   ,   27  ,   47  ,   75  ,   123 ,   455 ,   553 ,   1153    ,   1675    ,0 };
            const unsigned long dim342JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   13  ,   17  ,   3   ,   85  ,   103 ,   205 ,   443 ,   1093    ,   4083    ,0 };
            const unsigned long dim343JoeKuoD6Init[]    =   {   1   ,   3   ,   7   ,   13  ,   13  ,   63  ,   13  ,   103 ,   223 ,   473 ,   1125    ,   1565    ,0 };
            const unsigned long dim344JoeKuoD6Init[]    =   {   1   ,   1   ,   1   ,   13  ,   21  ,   55  ,   21  ,   39  ,   385 ,   809 ,   457 ,   1991    ,0 };
            const unsigned long dim345JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   5   ,   29  ,   57  ,   81  ,   1   ,   165 ,   181 ,   57  ,   1463    ,0 };
            const unsigned long dim346JoeKuoD6Init[]    =   {   1   ,   3   ,   7   ,   7   ,   21  ,   29  ,   59  ,   127 ,   393 ,   981 ,   577 ,   2707    ,0 };
            const unsigned long dim347JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   11  ,   17  ,   47  ,   123 ,   149 ,   273 ,   825 ,   1843    ,   479 ,0 };
            const unsigned long dim348JoeKuoD6Init[]    =   {   1   ,   1   ,   5   ,   7   ,   1   ,   33  ,   91  ,   251 ,   371 ,   655 ,   383 ,   3123    ,0 };
            const unsigned long dim349JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   7   ,   3   ,   55  ,   15  ,   195 ,   345 ,   1017    ,   385 ,   1869    ,0 };
            const unsigned long dim350JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   7   ,   15  ,   25  ,   25  ,   67  ,   223 ,   935 ,   905 ,   923 ,0 };
            const unsigned long dim351JoeKuoD6Init[]    =   {   1   ,   1   ,   1   ,   13  ,   29  ,   5   ,   23  ,   81  ,   227 ,   175 ,   1025    ,   355 ,0 };
            const unsigned long dim352JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   9   ,   25  ,   27  ,   43  ,   85  ,   401 ,   955 ,   989 ,   3117    ,0 };
            const unsigned long dim353JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   13  ,   27  ,   31  ,   27  ,   235 ,   99  ,   539 ,   1093    ,   2821    ,0 };
            const unsigned long dim354JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   7   ,   13  ,   19  ,   45  ,   231 ,   203 ,   489 ,   1929    ,   3615    ,0 };
            const unsigned long dim355JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   3   ,   3   ,   15  ,   3   ,   191 ,   181 ,   463 ,   1247    ,   2229    ,0 };
            const unsigned long dim356JoeKuoD6Init[]    =   {   1   ,   3   ,   7   ,   3   ,   1   ,   37  ,   1   ,   59  ,   29  ,   283 ,   679 ,   429 ,0 };
            const unsigned long dim357JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   15  ,   17  ,   25  ,   5   ,   113 ,   191 ,   505 ,   365 ,   3943    ,0 };
            const unsigned long dim358JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   3   ,   13  ,   31  ,   95  ,   141 ,   29  ,   353 ,   661 ,   3637    ,0 };
            const unsigned long dim359JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   1   ,   5   ,   53  ,   19  ,   9   ,   81  ,   497 ,   1915    ,   2997    ,0 };
            const unsigned long dim360JoeKuoD6Init[]    =   {   1   ,   3   ,   7   ,   15  ,   5   ,   13  ,   35  ,   163 ,   293 ,   309 ,   503 ,   2387    ,0 };
            const unsigned long dim361JoeKuoD6Init[]    =   {   1   ,   1   ,   5   ,   3   ,   11  ,   25  ,   127 ,   27  ,   435 ,   247 ,   1485    ,   3917    ,0 };
            const unsigned long dim362JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   5   ,   21  ,   7   ,   15  ,   213 ,   355 ,   209 ,   1107    ,   1317    ,0 };
            const unsigned long dim363JoeKuoD6Init[]    =   {   1   ,   3   ,   3   ,   11  ,   25  ,   29  ,   73  ,   203 ,   445 ,   849 ,   291 ,   1567    ,0 };
            const unsigned long dim364JoeKuoD6Init[]    =   {   1   ,   3   ,   3   ,   15  ,   15  ,   31  ,   5   ,   15  ,   19  ,   345 ,   635 ,   1815    ,0 };
            const unsigned long dim365JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   3   ,   19  ,   21  ,   15  ,   95  ,   111 ,   167 ,   1051    ,   479 ,0 };
            const unsigned long dim366JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   13  ,   15  ,   61  ,   125 ,   237 ,   11  ,   789 ,   1681    ,   437 ,0 };
            const unsigned long dim367JoeKuoD6Init[]    =   {   1   ,   3   ,   3   ,   15  ,   21  ,   23  ,   105 ,   145 ,   145 ,   453 ,   917 ,   1727    ,0 };
            const unsigned long dim368JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   11  ,   9   ,   37  ,   27  ,   91  ,   199 ,   561 ,   1705    ,   3849    ,0 };
            const unsigned long dim369JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   5   ,   11  ,   53  ,   1   ,   21  ,   63  ,   879 ,   629 ,   3219    ,0 };
            const unsigned long dim370JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   13  ,   3   ,   5   ,   121 ,   17  ,   511 ,   809 ,   627 ,   2627    ,0 };
            const unsigned long dim371JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   3   ,   27  ,   21  ,   115 ,   117 ,   193 ,   185 ,   1855    ,   2473    ,0 };
            const unsigned long dim372JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   9   ,   11  ,   37  ,   79  ,   123 ,   391 ,   715 ,   573 ,   699 ,0 };
            const unsigned long dim373JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   9   ,   5   ,   35  ,   105 ,   5   ,   55  ,   347 ,   1217    ,   3483    ,0 };
            const unsigned long dim374JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   11  ,   19  ,   59  ,   21  ,   221 ,   329 ,   873 ,   943 ,   3679    ,0 };
            const unsigned long dim375JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   13  ,   21  ,   57  ,   119 ,   33  ,   111 ,   581 ,   1133    ,   131 ,0 };
            const unsigned long dim376JoeKuoD6Init[]    =   {   1   ,   3   ,   3   ,   7   ,   13  ,   31  ,   67  ,   205 ,   211 ,   143 ,   1117    ,   4067    ,0 };
            const unsigned long dim377JoeKuoD6Init[]    =   {   1   ,   3   ,   7   ,   9   ,   9   ,   53  ,   71  ,   251 ,   41  ,   875 ,   691 ,   3681    ,0 };
            const unsigned long dim378JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   9   ,   23  ,   55  ,   57  ,   29  ,   123 ,   453 ,   523 ,   3237    ,0 };
            const unsigned long dim379JoeKuoD6Init[]    =   {   1   ,   1   ,   5   ,   11  ,   15  ,   19  ,   91  ,   105 ,   221 ,   1005    ,   169 ,   461 ,0 };
            const unsigned long dim380JoeKuoD6Init[]    =   {   1   ,   1   ,   1   ,   11  ,   7   ,   61  ,   49  ,   41  ,   337 ,   739 ,   699 ,   2379    ,0 };
            const unsigned long dim381JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   13  ,   21  ,   31  ,   103 ,   247 ,   87  ,   537 ,   757 ,   785 ,0 };
            const unsigned long dim382JoeKuoD6Init[]    =   {   1   ,   1   ,   5   ,   9   ,   7   ,   49  ,   127 ,   149 ,   277 ,   87  ,   475 ,   127 ,0 };
            const unsigned long dim383JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   15  ,   5   ,   25  ,   95  ,   229 ,   389 ,   121 ,   95  ,   2787    ,0 };
            const unsigned long dim384JoeKuoD6Init[]    =   {   1   ,   1   ,   1   ,   5   ,   9   ,   27  ,   29  ,   23  ,   419 ,   371 ,   127 ,   961 ,0 };
            const unsigned long dim385JoeKuoD6Init[]    =   {   1   ,   1   ,   1   ,   15  ,   25  ,   19  ,   123 ,   35  ,   41  ,   785 ,   605 ,   711 ,0 };
            const unsigned long dim386JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   3   ,   1   ,   13  ,   69  ,   161 ,   449 ,   225 ,   1795    ,   2083    ,0 };
            const unsigned long dim387JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   5   ,   21  ,   1   ,   19  ,   211 ,   191 ,   697 ,   89  ,   177 ,0 };
            const unsigned long dim388JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   3   ,   19  ,   29  ,   49  ,   151 ,   365 ,   541 ,   1681    ,   1943    ,0 };
            const unsigned long dim389JoeKuoD6Init[]    =   {   1   ,   3   ,   7   ,   7   ,   9   ,   5   ,   75  ,   109 ,   243 ,   769 ,   755 ,   2733    ,0 };
            const unsigned long dim390JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   11  ,   11  ,   41  ,   7   ,   87  ,   363 ,   893 ,   1799    ,   1303    ,0 };
            const unsigned long dim391JoeKuoD6Init[]    =   {   1   ,   1   ,   1   ,   9   ,   11  ,   5   ,   19  ,   101 ,   41  ,   401 ,   1863    ,   965 ,0 };
            const unsigned long dim392JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   11  ,   13  ,   15  ,   71  ,   249 ,   511 ,   625 ,   1101    ,   2615    ,0 };
            const unsigned long dim393JoeKuoD6Init[]    =   {   1   ,   3   ,   3   ,   9   ,   31  ,   35  ,   85  ,   107 ,   383 ,   201 ,   1163    ,   981 ,0 };
            const unsigned long dim394JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   9   ,   21  ,   13  ,   33  ,   9   ,   243 ,   93  ,   461 ,   741 ,0 };
            const unsigned long dim395JoeKuoD6Init[]    =   {   1   ,   1   ,   1   ,   9   ,   1   ,   55  ,   85  ,   179 ,   321 ,   163 ,   1807    ,   339 ,0 };
            const unsigned long dim396JoeKuoD6Init[]    =   {   1   ,   3   ,   7   ,   11  ,   29  ,   7   ,   57  ,   133 ,   245 ,   87  ,   917 ,   109 ,0 };
            const unsigned long dim397JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   9   ,   11  ,   13  ,   93  ,   81  ,   361 ,   343 ,   647 ,   1113    ,0 };
            const unsigned long dim398JoeKuoD6Init[]    =   {   1   ,   3   ,   7   ,   5   ,   23  ,   57  ,   117 ,   105 ,   221 ,   697 ,   1159    ,   1307    ,0 };
            const unsigned long dim399JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   7   ,   27  ,   55  ,   15  ,   235 ,   279 ,   941 ,   655 ,   3975    ,0 };
            const unsigned long dim400JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   9   ,   27  ,   47  ,   45  ,   41  ,   361 ,   135 ,   379 ,   2121    ,0 };
            const unsigned long dim401JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   15  ,   19  ,   51  ,   31  ,   39  ,   187 ,   573 ,   273 ,   383 ,0 };
            const unsigned long dim402JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   11  ,   9   ,   47  ,   107 ,   29  ,   435 ,   989 ,   1315    ,   2815    ,0 };
            const unsigned long dim403JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   3   ,   21  ,   31  ,   7   ,   41  ,   45  ,   115 ,   1639    ,   1861    ,0 };
            const unsigned long dim404JoeKuoD6Init[]    =   {   1   ,   1   ,   1   ,   15  ,   1   ,   21  ,   57  ,   233 ,   51  ,   149 ,   1251    ,   1809    ,0 };
            const unsigned long dim405JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   11  ,   1   ,   53  ,   111 ,   35  ,   447 ,   657 ,   1557    ,   349 ,0 };
            const unsigned long dim406JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   11  ,   25  ,   33  ,   39  ,   209 ,   127 ,   913 ,   659 ,   177 ,0 };
            const unsigned long dim407JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   9   ,   13  ,   53  ,   25  ,   191 ,   373 ,   997 ,   1413    ,   2859    ,0 };
            const unsigned long dim408JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   3   ,   31  ,   51  ,   17  ,   253 ,   149 ,   719 ,   1681    ,   2713    ,0 };
            const unsigned long dim409JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   1   ,   23  ,   49  ,   9   ,   211 ,   381 ,   321 ,   2033    ,   361 ,0 };
            const unsigned long dim410JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   15  ,   31  ,   3   ,   121 ,   21  ,   465 ,   417 ,   1023    ,   3921    ,0 };
            const unsigned long dim411JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   1   ,   17  ,   49  ,   101 ,   37  ,   429 ,   877 ,   505 ,   821 ,0 };
            const unsigned long dim412JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   7   ,   23  ,   59  ,   33  ,   127 ,   43  ,   987 ,   1065    ,   2675    ,0 };
            const unsigned long dim413JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   9   ,   7   ,   19  ,   45  ,   157 ,   475 ,   53  ,   1223    ,   523 ,0 };
            const unsigned long dim414JoeKuoD6Init[]    =   {   1   ,   3   ,   7   ,   9   ,   23  ,   61  ,   31  ,   133 ,   85  ,   623 ,   871 ,   965 ,0 };
            const unsigned long dim415JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   7   ,   3   ,   17  ,   21  ,   197 ,   377 ,   819 ,   661 ,   1803    ,0 };
            const unsigned long dim416JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   3   ,   11  ,   37  ,   1   ,   135 ,   509 ,   927 ,   23  ,   1657    ,0 };
            const unsigned long dim417JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   5   ,   15  ,   57  ,   55  ,   179 ,   317 ,   31  ,   611 ,   4057    ,0 };
            const unsigned long dim418JoeKuoD6Init[]    =   {   1   ,   1   ,   5   ,   5   ,   29  ,   57  ,   93  ,   105 ,   103 ,   157 ,   701 ,   249 ,0 };
            const unsigned long dim419JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   13  ,   27  ,   41  ,   23  ,   213 ,   363 ,   911 ,   85  ,   2733    ,0 };
            const unsigned long dim420JoeKuoD6Init[]    =   {   1   ,   1   ,   5   ,   15  ,   15  ,   15  ,   93  ,   203 ,   61  ,   759 ,   1735    ,   459 ,0 };
            const unsigned long dim421JoeKuoD6Init[]    =   {   1   ,   1   ,   5   ,   13  ,   1   ,   59  ,   27  ,   203 ,   165 ,   87  ,   835 ,   3647    ,0 };
            const unsigned long dim422JoeKuoD6Init[]    =   {   1   ,   3   ,   3   ,   5   ,   25  ,   57  ,   43  ,   209 ,   71  ,   35  ,   1985    ,   1521    ,0 };
            const unsigned long dim423JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   5   ,   15  ,   19  ,   49  ,   103 ,   195 ,   17  ,   2041    ,   1267    ,0 };
            const unsigned long dim424JoeKuoD6Init[]    =   {   1   ,   1   ,   5   ,   5   ,   5   ,   19  ,   91  ,   131 ,   157 ,   625 ,   1851    ,   3691    ,0 };
            const unsigned long dim425JoeKuoD6Init[]    =   {   1   ,   1   ,   1   ,   9   ,   9   ,   47  ,   3   ,   195 ,   25  ,   1003    ,   533 ,   833 ,0 };
            const unsigned long dim426JoeKuoD6Init[]    =   {   1   ,   1   ,   1   ,   15  ,   21  ,   33  ,   27  ,   193 ,   65  ,   545 ,   261 ,   3095    ,0 };
            const unsigned long dim427JoeKuoD6Init[]    =   {   1   ,   3   ,   7   ,   5   ,   7   ,   39  ,   35  ,   9   ,   387 ,   241 ,   1015    ,   1557    ,0 };
            const unsigned long dim428JoeKuoD6Init[]    =   {   1   ,   3   ,   7   ,   7   ,   13  ,   29  ,   43  ,   101 ,   457 ,   363 ,   115 ,   915 ,0 };
            const unsigned long dim429JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   13  ,   17  ,   57  ,   79  ,   11  ,   155 ,   39  ,   215 ,   157 ,0 };
            const unsigned long dim430JoeKuoD6Init[]    =   {   1   ,   3   ,   7   ,   9   ,   23  ,   33  ,   69  ,   189 ,   59  ,   823 ,   1941    ,   423 ,0 };
            const unsigned long dim431JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   11  ,   9   ,   39  ,   79  ,   43  ,   431 ,   227 ,   163 ,   2653    ,0 };
            const unsigned long dim432JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   3   ,   5   ,   63  ,   55  ,   141 ,   207 ,   349 ,   1873    ,   773 ,0 };
            const unsigned long dim433JoeKuoD6Init[]    =   {   1   ,   1   ,   5   ,   7   ,   21  ,   59  ,   89  ,   111 ,   53  ,   217 ,   1897    ,   1619    ,0 };
            const unsigned long dim434JoeKuoD6Init[]    =   {   1   ,   1   ,   5   ,   9   ,   7   ,   27  ,   127 ,   247 ,   183 ,   925 ,   989 ,   1059    ,0 };
            const unsigned long dim435JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   3   ,   21  ,   47  ,   39  ,   71  ,   343 ,   285 ,   1053    ,   3007    ,0 };
            const unsigned long dim436JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   11  ,   1   ,   7   ,   27  ,   109 ,   19  ,   325 ,   157 ,   2779    ,0 };
            const unsigned long dim437JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   9   ,   23  ,   27  ,   63  ,   115 ,   415 ,   1005    ,   1159    ,   3699    ,0 };
            const unsigned long dim438JoeKuoD6Init[]    =   {   1   ,   1   ,   5   ,   13  ,   27  ,   27  ,   5   ,   245 ,   95  ,   1021    ,   641 ,   3875    ,0 };
            const unsigned long dim439JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   9   ,   5   ,   9   ,   53  ,   183 ,   379 ,   251 ,   1747    ,   969 ,0 };
            const unsigned long dim440JoeKuoD6Init[]    =   {   1   ,   1   ,   5   ,   13  ,   7   ,   23  ,   53  ,   195 ,   353 ,   225 ,   1871    ,   2131    ,0 };
            const unsigned long dim441JoeKuoD6Init[]    =   {   1   ,   3   ,   7   ,   3   ,   21  ,   5   ,   25  ,   253 ,   59  ,   821 ,   1699    ,   1911    ,0 };
            const unsigned long dim442JoeKuoD6Init[]    =   {   1   ,   3   ,   3   ,   7   ,   9   ,   53  ,   121 ,   155 ,   281 ,   969 ,   1943    ,   2291    ,0 };
            const unsigned long dim443JoeKuoD6Init[]    =   {   1   ,   1   ,   1   ,   1   ,   27  ,   25  ,   65  ,   111 ,   295 ,   213 ,   81  ,   415 ,0 };
            const unsigned long dim444JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   7   ,   29  ,   51  ,   39  ,   201 ,   133 ,   987 ,   243 ,   3369    ,0 };
            const unsigned long dim445JoeKuoD6Init[]    =   {   1   ,   1   ,   1   ,   7   ,   27  ,   5   ,   57  ,   205 ,   395 ,   813 ,   125 ,   1293    ,0 };
            const unsigned long dim446JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   13  ,   15  ,   19  ,   57  ,   11  ,   151 ,   405 ,   2013    ,   2607    ,0 };
            const unsigned long dim447JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   1   ,   9   ,   53  ,   17  ,   25  ,   7   ,   483 ,   525 ,   967 ,0 };
            const unsigned long dim448JoeKuoD6Init[]    =   {   1   ,   3   ,   7   ,   13  ,   11  ,   31  ,   111 ,   253 ,   439 ,   689 ,   919 ,   3563    ,0 };
            const unsigned long dim449JoeKuoD6Init[]    =   {   1   ,   1   ,   5   ,   9   ,   13  ,   59  ,   123 ,   69  ,   339 ,   17  ,   297 ,   2511    ,0 };
            const unsigned long dim450JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   5   ,   13  ,   1   ,   85  ,   113 ,   45  ,   107 ,   565 ,   795 ,0 };
            const unsigned long dim451JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   11  ,   15  ,   5   ,   109 ,   163 ,   73  ,   103 ,   195 ,   639 ,0 };
            const unsigned long dim452JoeKuoD6Init[]    =   {   1   ,   1   ,   5   ,   9   ,   21  ,   11  ,   93  ,   23  ,   217 ,   275 ,   139 ,   1159    ,0 };
            const unsigned long dim453JoeKuoD6Init[]    =   {   1   ,   1   ,   1   ,   3   ,   31  ,   63  ,   73  ,   61  ,   63  ,   923 ,   1237    ,   3887    ,0 };
            const unsigned long dim454JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   9   ,   27  ,   45  ,   5   ,   93  ,   3   ,   303 ,   357 ,   2973    ,0 };
            const unsigned long dim455JoeKuoD6Init[]    =   {   1   ,   1   ,   1   ,   11  ,   7   ,   51  ,   41  ,   27  ,   83  ,   357 ,   1679    ,   2015    ,0 };
            const unsigned long dim456JoeKuoD6Init[]    =   {   1   ,   3   ,   3   ,   13  ,   9   ,   53  ,   1   ,   193 ,   343 ,   59  ,   451 ,   1579    ,0 };
            const unsigned long dim457JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   3   ,   9   ,   37  ,   37  ,   131 ,   207 ,   29  ,   2027    ,   733 ,0 };
            const unsigned long dim458JoeKuoD6Init[]    =   {   1   ,   1   ,   5   ,   7   ,   3   ,   61  ,   51  ,   221 ,   71  ,   659 ,   1955    ,   1289    ,0 };
            const unsigned long dim459JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   15  ,   5   ,   19  ,   3   ,   205 ,   319 ,   155 ,   325 ,   2643    ,0 };
            const unsigned long dim460JoeKuoD6Init[]    =   {   1   ,   1   ,   5   ,   9   ,   3   ,   15  ,   1   ,   247 ,   323 ,   143 ,   37  ,   2069    ,0 };
            const unsigned long dim461JoeKuoD6Init[]    =   {   1   ,   1   ,   5   ,   15  ,   23  ,   43  ,   35  ,   59  ,   285 ,   779 ,   605 ,   1739    ,0 };
            const unsigned long dim462JoeKuoD6Init[]    =   {   1   ,   3   ,   3   ,   5   ,   29  ,   31  ,   57  ,   225 ,   417 ,   521 ,   1337    ,   3051    ,0 };
            const unsigned long dim463JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   7   ,   23  ,   57  ,   41  ,   165 ,   437 ,   669 ,   1087    ,   1275    ,0 };
            const unsigned long dim464JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   5   ,   1   ,   51  ,   25  ,   233 ,   449 ,   503 ,   1237    ,   1303    ,0 };
            const unsigned long dim465JoeKuoD6Init[]    =   {   1   ,   1   ,   1   ,   11  ,   9   ,   45  ,   79  ,   149 ,   229 ,   991 ,   1725    ,   395 ,0 };
            const unsigned long dim466JoeKuoD6Init[]    =   {   1   ,   1   ,   5   ,   11  ,   31  ,   37  ,   13  ,   117 ,   19  ,   939 ,   1347    ,   2251    ,0 };
            const unsigned long dim467JoeKuoD6Init[]    =   {   1   ,   1   ,   5   ,   5   ,   1   ,   29  ,   25  ,   251 ,   493 ,   793 ,   649 ,   1197    ,0 };
            const unsigned long dim468JoeKuoD6Init[]    =   {   1   ,   3   ,   3   ,   11  ,   5   ,   53  ,   49  ,   33  ,   37  ,   937 ,   497 ,   361 ,0 };
            const unsigned long dim469JoeKuoD6Init[]    =   {   1   ,   1   ,   1   ,   1   ,   5   ,   57  ,   19  ,   141 ,   115 ,   371 ,   1341    ,   3335    ,0 };
            const unsigned long dim470JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   5   ,   23  ,   47  ,   77  ,   31  ,   225 ,   589 ,   551 ,   2625    ,0 };
            const unsigned long dim471JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   1   ,   25  ,   37  ,   29  ,   213 ,   283 ,   207 ,   385 ,   2019    ,0 };
            const unsigned long dim472JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   11  ,   5   ,   17  ,   21  ,   37  ,   167 ,   49  ,   1595    ,   1647    ,0 };
            const unsigned long dim473JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   15  ,   29  ,   15  ,   79  ,   203 ,   329 ,   627 ,   1799    ,   895 ,0 };
            const unsigned long dim474JoeKuoD6Init[]    =   {   1   ,   1   ,   5   ,   7   ,   23  ,   53  ,   69  ,   45  ,   271 ,   661 ,   1437    ,   2429    ,0 };
            const unsigned long dim475JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   1   ,   21  ,   7   ,   115 ,   233 ,   267 ,   873 ,   2023    ,   3431    ,0 };
            const unsigned long dim476JoeKuoD6Init[]    =   {   1   ,   1   ,   5   ,   5   ,   7   ,   63  ,   7   ,   125 ,   129 ,   815 ,   771 ,   1233    ,0 };
            const unsigned long dim477JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   9   ,   5   ,   29  ,   121 ,   203 ,   373 ,   137 ,   1181    ,   3773    ,0 };
            const unsigned long dim478JoeKuoD6Init[]    =   {   1   ,   1   ,   5   ,   1   ,   15  ,   47  ,   25  ,   137 ,   329 ,   531 ,   1957    ,   1933    ,0 };
            const unsigned long dim479JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   11  ,   23  ,   37  ,   67  ,   123 ,   181 ,   567 ,   893 ,   2307    ,0 };
            const unsigned long dim480JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   3   ,   5   ,   23  ,   41  ,   227 ,   9   ,   213 ,   821 ,   1689    ,0 };
            const unsigned long dim481JoeKuoD6Init[]    =   {   1   ,   1   ,   1   ,   13  ,   29  ,   33  ,   23  ,   241 ,   199 ,   753 ,   1759    ,   833 ,   5261    ,0 };
            const unsigned long dim482JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   15  ,   31  ,   7   ,   61  ,   225 ,   209 ,   117 ,   1983    ,   395 ,   3583    ,0 };
            const unsigned long dim483JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   5   ,   27  ,   49  ,   7   ,   125 ,   365 ,   513 ,   1001    ,   283 ,   627 ,0 };
            const unsigned long dim484JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   3   ,   7   ,   53  ,   99  ,   123 ,   493 ,   481 ,   173 ,   1483    ,   805 ,0 };
            const unsigned long dim485JoeKuoD6Init[]    =   {   1   ,   1   ,   5   ,   7   ,   15  ,   1   ,   57  ,   19  ,   391 ,   847 ,   1641    ,   3497    ,   2807    ,0 };
            const unsigned long dim486JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   11  ,   19  ,   49  ,   77  ,   139 ,   275 ,   115 ,   1401    ,   2541    ,   7239    ,0 };
            const unsigned long dim487JoeKuoD6Init[]    =   {   1   ,   1   ,   1   ,   15  ,   11  ,   13  ,   13  ,   211 ,   119 ,   317 ,   271 ,   1867    ,   4299    ,0 };
            const unsigned long dim488JoeKuoD6Init[]    =   {   1   ,   1   ,   1   ,   13  ,   15  ,   37  ,   121 ,   197 ,   485 ,   605 ,   1025    ,   3235    ,   6205    ,0 };
            const unsigned long dim489JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   7   ,   25  ,   39  ,   31  ,   3   ,   465 ,   33  ,   625 ,   1069    ,   6043    ,0 };
            const unsigned long dim490JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   9   ,   7   ,   37  ,   115 ,   169 ,   27  ,   967 ,   505 ,   1493    ,   4333    ,0 };
            const unsigned long dim491JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   13  ,   1   ,   51  ,   19  ,   117 ,   365 ,   587 ,   1597    ,   279 ,   99  ,0 };
            const unsigned long dim492JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   9   ,   15  ,   57  ,   89  ,   153 ,   495 ,   535 ,   775 ,   4063    ,   6049    ,0 };
            const unsigned long dim493JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   1   ,   9   ,   7   ,   111 ,   43  ,   225 ,   677 ,   1379    ,   481 ,   5003    ,0 };
            const unsigned long dim494JoeKuoD6Init[]    =   {   1   ,   1   ,   1   ,   15  ,   11  ,   47  ,   63  ,   37  ,   41  ,   833 ,   467 ,   3795    ,   5643    ,0 };
            const unsigned long dim495JoeKuoD6Init[]    =   {   1   ,   3   ,   3   ,   7   ,   25  ,   49  ,   21  ,   91  ,   191 ,   579 ,   507 ,   3001    ,   6549    ,0 };
            const unsigned long dim496JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   9   ,   25  ,   17  ,   17  ,   153 ,   467 ,   359 ,   1523    ,   3175    ,   7307    ,0 };
            const unsigned long dim497JoeKuoD6Init[]    =   {   1   ,   3   ,   7   ,   7   ,   29  ,   57  ,   51  ,   145 ,   71  ,   515 ,   495 ,   1091    ,   7911    ,0 };
            const unsigned long dim498JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   7   ,   31  ,   31  ,   53  ,   149 ,   303 ,   969 ,   383 ,   939 ,   2659    ,0 };
            const unsigned long dim499JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   9   ,   1   ,   21  ,   15  ,   145 ,   9   ,   199 ,   695 ,   2967    ,   1179    ,0 };
            const unsigned long dim500JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   3   ,   23  ,   33  ,   89  ,   23  ,   387 ,   591 ,   1543    ,   3941    ,   7021    ,0 };
            const unsigned long dim501JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   7   ,   27  ,   17  ,   25  ,   71  ,   43  ,   761 ,   803 ,   833 ,   1023    ,0 };
            const unsigned long dim502JoeKuoD6Init[]    =   {   1   ,   1   ,   1   ,   3   ,   11  ,   25  ,   3   ,   171 ,   377 ,   843 ,   1147    ,   1469    ,   1773    ,0 };
            const unsigned long dim503JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   15  ,   1   ,   13  ,   65  ,   197 ,   381 ,   683 ,   895 ,   2181    ,   1973    ,0 };
            const unsigned long dim504JoeKuoD6Init[]    =   {   1   ,   1   ,   5   ,   7   ,   19  ,   25  ,   109 ,   49  ,   451 ,   277 ,   1417    ,   2965    ,   7677    ,0 };
            const unsigned long dim505JoeKuoD6Init[]    =   {   1   ,   3   ,   3   ,   7   ,   25  ,   51  ,   115 ,   43  ,   175 ,   317 ,   953 ,   81  ,   2533    ,0 };
            const unsigned long dim506JoeKuoD6Init[]    =   {   1   ,   3   ,   7   ,   15  ,   27  ,   57  ,   5   ,   181 ,   489 ,   369 ,   1339    ,   3643    ,   7911    ,0 };
            const unsigned long dim507JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   5   ,   21  ,   11  ,   87  ,   81  ,   85  ,   5   ,   867 ,   3541    ,   3071    ,0 };
            const unsigned long dim508JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   3   ,   19  ,   37  ,   111 ,   43  ,   27  ,   549 ,   367 ,   3021    ,   6563    ,0 };
            const unsigned long dim509JoeKuoD6Init[]    =   {   1   ,   1   ,   1   ,   13  ,   3   ,   63  ,   83  ,   107 ,   327 ,   599 ,   425 ,   3941    ,   4319    ,0 };
            const unsigned long dim510JoeKuoD6Init[]    =   {   1   ,   3   ,   7   ,   15  ,   5   ,   37  ,   127 ,   169 ,   117 ,   987 ,   1821    ,   997 ,   63  ,0 };
            const unsigned long dim511JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   15  ,   17  ,   41  ,   103 ,   129 ,   449 ,   373 ,   733 ,   55  ,   5987    ,0 };
            const unsigned long dim512JoeKuoD6Init[]    =   {   1   ,   1   ,   1   ,   15  ,   13  ,   61  ,   15  ,   117 ,   249 ,   149 ,   885 ,   3723    ,   5525    ,0 };
            const unsigned long dim513JoeKuoD6Init[]    =   {   1   ,   1   ,   1   ,   1   ,   7   ,   3   ,   39  ,   47  ,   367 ,   355 ,   1943    ,   1035    ,   2367    ,0 };
            const unsigned long dim514JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   11  ,   19  ,   49  ,   21  ,   239 ,   403 ,   921 ,   693 ,   2123    ,   7387    ,0 };
            const unsigned long dim515JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   5   ,   23  ,   45  ,   13  ,   81  ,   105 ,   91  ,   1173    ,   3059    ,   5969    ,0 };
            const unsigned long dim516JoeKuoD6Init[]    =   {   1   ,   1   ,   1   ,   5   ,   1   ,   31  ,   73  ,   253 ,   395 ,   941 ,   1949    ,   455 ,   453 ,0 };
            const unsigned long dim517JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   15  ,   27  ,   57  ,   115 ,   17  ,   269 ,   819 ,   145 ,   1219    ,   6471    ,0 };
            const unsigned long dim518JoeKuoD6Init[]    =   {   1   ,   3   ,   3   ,   7   ,   29  ,   19  ,   9   ,   67  ,   501 ,   87  ,   883 ,   1985    ,   4859    ,0 };
            const unsigned long dim519JoeKuoD6Init[]    =   {   1   ,   1   ,   1   ,   3   ,   29  ,   3   ,   55  ,   107 ,   79  ,   941 ,   1437    ,   3607    ,   6221    ,0 };
            const unsigned long dim520JoeKuoD6Init[]    =   {   1   ,   3   ,   3   ,   15  ,   9   ,   63  ,   59  ,   113 ,   319 ,   825 ,   641 ,   491 ,   4015    ,0 };
            const unsigned long dim521JoeKuoD6Init[]    =   {   1   ,   1   ,   1   ,   3   ,   21  ,   59  ,   55  ,   107 ,   215 ,   167 ,   315 ,   81  ,   3491    ,0 };
            const unsigned long dim522JoeKuoD6Init[]    =   {   1   ,   3   ,   7   ,   11  ,   15  ,   7   ,   71  ,   145 ,   481 ,   843 ,   1437    ,   3809    ,   7729    ,0 };
            const unsigned long dim523JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   11  ,   23  ,   59  ,   65  ,   45  ,   397 ,   117 ,   325 ,   2933    ,   541 ,0 };
            const unsigned long dim524JoeKuoD6Init[]    =   {   1   ,   3   ,   3   ,   15  ,   15  ,   47  ,   77  ,   35  ,   143 ,   327 ,   965 ,   3409    ,   1817    ,0 };
            const unsigned long dim525JoeKuoD6Init[]    =   {   1   ,   1   ,   1   ,   9   ,   31  ,   7   ,   31  ,   57  ,   331 ,   615 ,   645 ,   605 ,   2185    ,0 };
            const unsigned long dim526JoeKuoD6Init[]    =   {   1   ,   1   ,   5   ,   1   ,   15  ,   53  ,   109 ,   135 ,   259 ,   591 ,   1419    ,   3349    ,   5541    ,0 };
            const unsigned long dim527JoeKuoD6Init[]    =   {   1   ,   1   ,   5   ,   13  ,   21  ,   23  ,   87  ,   115 ,   211 ,   737 ,   1433    ,   3247    ,   6899    ,0 };
            const unsigned long dim528JoeKuoD6Init[]    =   {   1   ,   3   ,   3   ,   7   ,   23  ,   5   ,   43  ,   43  ,   253 ,   753 ,   1305    ,   2179    ,   1375    ,0 };
            const unsigned long dim529JoeKuoD6Init[]    =   {   1   ,   3   ,   3   ,   7   ,   21  ,   49  ,   127 ,   129 ,   229 ,   415 ,   1593    ,   2315    ,   3273    ,0 };
            const unsigned long dim530JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   9   ,   15  ,   61  ,   97  ,   101 ,   199 ,   651 ,   1421    ,   2471    ,   6109    ,0 };
            const unsigned long dim531JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   11  ,   13  ,   27  ,   19  ,   243 ,   185 ,   707 ,   559 ,   1687    ,   981 ,0 };
            const unsigned long dim532JoeKuoD6Init[]    =   {   1   ,   3   ,   7   ,   3   ,   5   ,   9   ,   101 ,   55  ,   39  ,   811 ,   1491    ,   3725    ,   6691    ,0 };
            const unsigned long dim533JoeKuoD6Init[]    =   {   1   ,   1   ,   5   ,   9   ,   27  ,   33  ,   41  ,   61  ,   271 ,   695 ,   1563    ,   547 ,   1569    ,0 };
            const unsigned long dim534JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   1   ,   29  ,   21  ,   9   ,   83  ,   131 ,   287 ,   1505    ,   2711    ,   1617    ,0 };
            const unsigned long dim535JoeKuoD6Init[]    =   {   1   ,   1   ,   1   ,   7   ,   29  ,   57  ,   19  ,   99  ,   303 ,   609 ,   1397    ,   3241    ,   3463    ,0 };
            const unsigned long dim536JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   9   ,   13  ,   33  ,   45  ,   101 ,   195 ,   615 ,   2017    ,   2473    ,   1251    ,0 };
            const unsigned long dim537JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   7   ,   31  ,   7   ,   121 ,   67  ,   275 ,   153 ,   1189    ,   2343    ,   5593    ,0 };
            const unsigned long dim538JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   15  ,   19  ,   37  ,   79  ,   207 ,   473 ,   927 ,   915 ,   1971    ,   8033    ,0 };
            const unsigned long dim539JoeKuoD6Init[]    =   {   1   ,   3   ,   3   ,   13  ,   7   ,   55  ,   123 ,   163 ,   47  ,   599 ,   209 ,   2235    ,   7643    ,0 };
            const unsigned long dim540JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   11  ,   31  ,   3   ,   31  ,   69  ,   61  ,   257 ,   11  ,   845 ,   2659    ,0 };
            const unsigned long dim541JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   9   ,   27  ,   15  ,   73  ,   237 ,   79  ,   517 ,   1905    ,   1595    ,   6337    ,0 };
            const unsigned long dim542JoeKuoD6Init[]    =   {   1   ,   1   ,   5   ,   15  ,   11  ,   19  ,   95  ,   213 ,   321 ,   303 ,   1957    ,   2291    ,   5991    ,0 };
            const unsigned long dim543JoeKuoD6Init[]    =   {   1   ,   1   ,   1   ,   11  ,   15  ,   17  ,   85  ,   45  ,   143 ,   425 ,   55  ,   3595    ,   685 ,0 };
            const unsigned long dim544JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   3   ,   25  ,   5   ,   25  ,   203 ,   353 ,   537 ,   1921    ,   1885    ,   1157    ,0 };
            const unsigned long dim545JoeKuoD6Init[]    =   {   1   ,   1   ,   1   ,   7   ,   29  ,   23  ,   117 ,   141 ,   257 ,   81  ,   25  ,   3237    ,   7223    ,0 };
            const unsigned long dim546JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   11  ,   27  ,   23  ,   47  ,   29  ,   143 ,   341 ,   1855    ,   3973    ,   5977    ,0 };
            const unsigned long dim547JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   7   ,   7   ,   45  ,   11  ,   95  ,   135 ,   261 ,   1475    ,   3723    ,   6999    ,0 };
            const unsigned long dim548JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   13  ,   19  ,   39  ,   89  ,   93  ,   187 ,   1011    ,   1319    ,   67  ,   399 ,0 };
            const unsigned long dim549JoeKuoD6Init[]    =   {   1   ,   1   ,   5   ,   5   ,   17  ,   1   ,   91  ,   69  ,   405 ,   815 ,   225 ,   2469    ,   6287    ,0 };
            const unsigned long dim550JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   7   ,   3   ,   19  ,   41  ,   37  ,   125 ,   221 ,   1011    ,   3603    ,   4881    ,0 };
            const unsigned long dim551JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   11  ,   15  ,   13  ,   45  ,   235 ,   141 ,   17  ,   1991    ,   2581    ,   4965    ,0 };
            const unsigned long dim552JoeKuoD6Init[]    =   {   1   ,   3   ,   3   ,   13  ,   5   ,   41  ,   117 ,   143 ,   261 ,   717 ,   111 ,   395 ,   4381    ,0 };
            const unsigned long dim553JoeKuoD6Init[]    =   {   1   ,   3   ,   7   ,   5   ,   15  ,   47  ,   117 ,   197 ,   147 ,   117 ,   295 ,   2005    ,   4081    ,0 };
            const unsigned long dim554JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   3   ,   29  ,   61  ,   93  ,   245 ,   59  ,   119 ,   1121    ,   2549    ,   6471    ,0 };
            const unsigned long dim555JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   5   ,   13  ,   43  ,   93  ,   133 ,   313 ,   923 ,   295 ,   1239    ,   6355    ,0 };
            const unsigned long dim556JoeKuoD6Init[]    =   {   1   ,   3   ,   3   ,   11  ,   5   ,   47  ,   99  ,   115 ,   107 ,   585 ,   671 ,   2077    ,   7005    ,0 };
            const unsigned long dim557JoeKuoD6Init[]    =   {   1   ,   1   ,   1   ,   9   ,   13  ,   31  ,   79  ,   253 ,   245 ,   597 ,   1071    ,   2273    ,   2481    ,0 };
            const unsigned long dim558JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   1   ,   25  ,   3   ,   39  ,   47  ,   201 ,   187 ,   419 ,   3881    ,   4031    ,0 };
            const unsigned long dim559JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   11  ,   17  ,   47  ,   119 ,   243 ,   217 ,   321 ,   727 ,   3227    ,   6041    ,0 };
            const unsigned long dim560JoeKuoD6Init[]    =   {   1   ,   1   ,   5   ,   11  ,   9   ,   3   ,   85  ,   121 ,   161 ,   663 ,   2035    ,   4073    ,   7401    ,0 };
            const unsigned long dim561JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   5   ,   11  ,   45  ,   51  ,   99  ,   431 ,   307 ,   157 ,   2941    ,   1645    ,0 };
            const unsigned long dim562JoeKuoD6Init[]    =   {   1   ,   3   ,   3   ,   7   ,   7   ,   43  ,   77  ,   19  ,   169 ,   535 ,   1797    ,   2907    ,   531 ,0 };
            const unsigned long dim563JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   9   ,   19  ,   11  ,   45  ,   51  ,   169 ,   175 ,   995 ,   2231    ,   147 ,0 };
            const unsigned long dim564JoeKuoD6Init[]    =   {   1   ,   1   ,   5   ,   1   ,   3   ,   25  ,   123 ,   11  ,   387 ,   757 ,   1067    ,   1359    ,   2227    ,0 };
            const unsigned long dim565JoeKuoD6Init[]    =   {   1   ,   1   ,   5   ,   15  ,   13  ,   41  ,   45  ,   135 ,   49  ,   391 ,   1551    ,   1729    ,   4799    ,0 };
            const unsigned long dim566JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   9   ,   3   ,   11  ,   109 ,   59  ,   217 ,   507 ,   47  ,   1609    ,   7145    ,0 };
            const unsigned long dim567JoeKuoD6Init[]    =   {   1   ,   1   ,   5   ,   15  ,   13  ,   35  ,   7   ,   213 ,   495 ,   811 ,   1403    ,   3423    ,   8139    ,0 };
            const unsigned long dim568JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   15  ,   3   ,   55  ,   107 ,   47  ,   337 ,   305 ,   1869    ,   943 ,   7533    ,0 };
            const unsigned long dim569JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   15  ,   15  ,   51  ,   9   ,   209 ,   477 ,   173 ,   1921    ,   3295    ,   4283    ,0 };
            const unsigned long dim570JoeKuoD6Init[]    =   {   1   ,   3   ,   7   ,   13  ,   17  ,   31  ,   111 ,   73  ,   143 ,   773 ,   1499    ,   651 ,   6659    ,0 };
            const unsigned long dim571JoeKuoD6Init[]    =   {   1   ,   1   ,   1   ,   1   ,   17  ,   5   ,   1   ,   95  ,   347 ,   613 ,   49  ,   7   ,   2373    ,0 };
            const unsigned long dim572JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   13  ,   5   ,   21  ,   33  ,   227 ,   375 ,   635 ,   149 ,   1585    ,   6159    ,0 };
            const unsigned long dim573JoeKuoD6Init[]    =   {   1   ,   3   ,   7   ,   5   ,   25  ,   7   ,   121 ,   185 ,   129 ,   1021    ,   167 ,   873 ,   949 ,0 };
            const unsigned long dim574JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   11  ,   29  ,   23  ,   25  ,   219 ,   203 ,   651 ,   629 ,   155 ,   5385    ,0 };
            const unsigned long dim575JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   9   ,   25  ,   1   ,   91  ,   213 ,   87  ,   807 ,   599 ,   2527    ,   6971    ,0 };
            const unsigned long dim576JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   11  ,   19  ,   13  ,   91  ,   11  ,   73  ,   617 ,   1721    ,   1803    ,   3005    ,0 };
            const unsigned long dim577JoeKuoD6Init[]    =   {   1   ,   1   ,   5   ,   9   ,   25  ,   33  ,   19  ,   167 ,   195 ,   861 ,   1209    ,   1887    ,   3165    ,0 };
            const unsigned long dim578JoeKuoD6Init[]    =   {   1   ,   1   ,   5   ,   1   ,   5   ,   21  ,   57  ,   253 ,   325 ,   907 ,   1237    ,   1037    ,   3639    ,0 };
            const unsigned long dim579JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   1   ,   5   ,   17  ,   35  ,   65  ,   193 ,   57  ,   1045    ,   2117    ,   7433    ,0 };
            const unsigned long dim580JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   1   ,   3   ,   23  ,   89  ,   157 ,   379 ,   831 ,   479 ,   3285    ,   5107    ,0 };
            const unsigned long dim581JoeKuoD6Init[]    =   {   1   ,   3   ,   3   ,   1   ,   5   ,   35  ,   57  ,   107 ,   465 ,   81  ,   387 ,   1035    ,   7231    ,0 };
            const unsigned long dim582JoeKuoD6Init[]    =   {   1   ,   3   ,   3   ,   9   ,   13  ,   25  ,   23  ,   39  ,   439 ,   89  ,   1651    ,   781 ,   6167    ,0 };
            const unsigned long dim583JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   1   ,   19  ,   33  ,   103 ,   139 ,   259 ,   201 ,   1569    ,   3453    ,   5143    ,0 };
            const unsigned long dim584JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   1   ,   5   ,   51  ,   33  ,   199 ,   509 ,   593 ,   1671    ,   243 ,   3455    ,0 };
            const unsigned long dim585JoeKuoD6Init[]    =   {   1   ,   1   ,   5   ,   15  ,   31  ,   15  ,   81  ,   173 ,   357 ,   447 ,   595 ,   1177    ,   2435    ,0 };
            const unsigned long dim586JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   1   ,   9   ,   35  ,   53  ,   111 ,   277 ,   143 ,   1837    ,   385 ,   7057    ,0 };
            const unsigned long dim587JoeKuoD6Init[]    =   {   1   ,   3   ,   3   ,   1   ,   7   ,   25  ,   115 ,   117 ,   1   ,   97  ,   51  ,   3459    ,   3057    ,0 };
            const unsigned long dim588JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   5   ,   11  ,   63  ,   85  ,   69  ,   1   ,   943 ,   279 ,   4091    ,   7753    ,0 };
            const unsigned long dim589JoeKuoD6Init[]    =   {   1   ,   3   ,   7   ,   9   ,   7   ,   59  ,   19  ,   17  ,   435 ,   779 ,   1049    ,   2125    ,   3295    ,0 };
            const unsigned long dim590JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   13  ,   29  ,   45  ,   95  ,   37  ,   179 ,   463 ,   1503    ,   417 ,   1959    ,0 };
            const unsigned long dim591JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   9   ,   21  ,   27  ,   27  ,   53  ,   359 ,   781 ,   1387    ,   997 ,   2123    ,0 };
            const unsigned long dim592JoeKuoD6Init[]    =   {   1   ,   3   ,   3   ,   13  ,   17  ,   39  ,   103 ,   175 ,   131 ,   973 ,   1055    ,   2935    ,   5671    ,0 };
            const unsigned long dim593JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   9   ,   7   ,   41  ,   55  ,   133 ,   223 ,   1019    ,   1341    ,   2559    ,   4225    ,0 };
            const unsigned long dim594JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   9   ,   13  ,   7   ,   85  ,   101 ,   161 ,   461 ,   227 ,   2073    ,   6483    ,0 };
            const unsigned long dim595JoeKuoD6Init[]    =   {   1   ,   3   ,   3   ,   1   ,   11  ,   27  ,   67  ,   125 ,   287 ,   229 ,   1945    ,   2461    ,   7111    ,0 };
            const unsigned long dim596JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   3   ,   23  ,   59  ,   39  ,   137 ,   421 ,   245 ,   955 ,   1941    ,   3743    ,0 };
            const unsigned long dim597JoeKuoD6Init[]    =   {   1   ,   1   ,   1   ,   9   ,   19  ,   55  ,   45  ,   93  ,   101 ,   777 ,   729 ,   867 ,   7123    ,0 };
            const unsigned long dim598JoeKuoD6Init[]    =   {   1   ,   1   ,   5   ,   5   ,   21  ,   27  ,   49  ,   125 ,   245 ,   271 ,   1643    ,   3995    ,   1845    ,0 };
            const unsigned long dim599JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   15  ,   21  ,   63  ,   113 ,   157 ,   123 ,   347 ,   333 ,   2445    ,   3401    ,0 };
            const unsigned long dim600JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   5   ,   25  ,   7   ,   113 ,   13  ,   167 ,   935 ,   329 ,   3483    ,   4237    ,0 };
            const unsigned long dim601JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   9   ,   25  ,   43  ,   101 ,   115 ,   439 ,   633 ,   2001    ,   283 ,   4415    ,0 };
            const unsigned long dim602JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   7   ,   7   ,   39  ,   41  ,   201 ,   423 ,   351 ,   563 ,   653 ,   6769    ,0 };
            const unsigned long dim603JoeKuoD6Init[]    =   {   1   ,   3   ,   7   ,   11  ,   3   ,   55  ,   29  ,   75  ,   17  ,   283 ,   837 ,   1303    ,   7917    ,0 };
            const unsigned long dim604JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   11  ,   9   ,   17  ,   47  ,   39  ,   399 ,   645 ,   575 ,   805 ,   5751    ,0 };
            const unsigned long dim605JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   11  ,   7   ,   9   ,   99  ,   239 ,   503 ,   995 ,   1265    ,   1651    ,   2095    ,0 };
            const unsigned long dim606JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   15  ,   17  ,   13  ,   107 ,   165 ,   341 ,   337 ,   51  ,   1619    ,   7801    ,0 };
            const unsigned long dim607JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   3   ,   19  ,   55  ,   73  ,   105 ,   97  ,   557 ,   1701    ,   799 ,   5591    ,0 };
            const unsigned long dim608JoeKuoD6Init[]    =   {   1   ,   1   ,   5   ,   13  ,   15  ,   21  ,   35  ,   39  ,   215 ,   357 ,   171 ,   161 ,   4527    ,0 };
            const unsigned long dim609JoeKuoD6Init[]    =   {   1   ,   3   ,   3   ,   15  ,   19  ,   7   ,   27  ,   153 ,   399 ,   415 ,   2031    ,   627 ,   8169    ,0 };
            const unsigned long dim610JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   11  ,   27  ,   37  ,   21  ,   217 ,   413 ,   787 ,   1637    ,   1199    ,   1735    ,0 };
            const unsigned long dim611JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   7   ,   25  ,   21  ,   95  ,   13  ,   159 ,   195 ,   1493    ,   1395    ,   7453    ,0 };
            const unsigned long dim612JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   1   ,   3   ,   17  ,   47  ,   95  ,   327 ,   493 ,   1501    ,   2857    ,   7215    ,0 };
            const unsigned long dim613JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   3   ,   19  ,   25  ,   67  ,   89  ,   243 ,   913 ,   815 ,   2471    ,   6537    ,0 };
            const unsigned long dim614JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   1   ,   29  ,   53  ,   13  ,   21  ,   131 ,   893 ,   1113    ,   525 ,   6723    ,0 };
            const unsigned long dim615JoeKuoD6Init[]    =   {   1   ,   1   ,   5   ,   1   ,   25  ,   45  ,   97  ,   167 ,   465 ,   65  ,   1923    ,   3971    ,   6365    ,0 };
            const unsigned long dim616JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   7   ,   3   ,   7   ,   113 ,   85  ,   371 ,   143 ,   201 ,   1023    ,   2827    ,0 };
            const unsigned long dim617JoeKuoD6Init[]    =   {   1   ,   3   ,   3   ,   1   ,   5   ,   31  ,   5   ,   37  ,   57  ,   203 ,   241 ,   2473    ,   593 ,0 };
            const unsigned long dim618JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   5   ,   25  ,   59  ,   87  ,   207 ,   227 ,   103 ,   491 ,   1299    ,   4173    ,0 };
            const unsigned long dim619JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   9   ,   15  ,   61  ,   75  ,   171 ,   421 ,   651 ,   375 ,   511 ,   2723    ,0 };
            const unsigned long dim620JoeKuoD6Init[]    =   {   1   ,   3   ,   7   ,   11  ,   9   ,   19  ,   55  ,   5   ,   365 ,   7   ,   1401    ,   3049    ,   6669    ,0 };
            const unsigned long dim621JoeKuoD6Init[]    =   {   1   ,   1   ,   1   ,   5   ,   31  ,   37  ,   97  ,   105 ,   343 ,   999 ,   1641    ,   1951    ,   1395    ,0 };
            const unsigned long dim622JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   1   ,   5   ,   13  ,   83  ,   189 ,   395 ,   335 ,   1767    ,   887 ,   3431    ,0 };
            const unsigned long dim623JoeKuoD6Init[]    =   {   1   ,   1   ,   1   ,   13  ,   17  ,   45  ,   71  ,   227 ,   171 ,   831 ,   617 ,   1203    ,   6477    ,0 };
            const unsigned long dim624JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   13  ,   15  ,   49  ,   91  ,   61  ,   155 ,   951 ,   1031    ,   1153    ,   5201    ,0 };
            const unsigned long dim625JoeKuoD6Init[]    =   {   1   ,   1   ,   1   ,   13  ,   23  ,   51  ,   75  ,   89  ,   359 ,   257 ,   1695    ,   1739    ,   5439    ,0 };
            const unsigned long dim626JoeKuoD6Init[]    =   {   1   ,   3   ,   3   ,   13  ,   9   ,   47  ,   31  ,   225 ,   191 ,   787 ,   355 ,   2013    ,   5897    ,0 };
            const unsigned long dim627JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   3   ,   7   ,   41  ,   91  ,   83  ,   9   ,   643 ,   1589    ,   3491    ,   135 ,0 };
            const unsigned long dim628JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   9   ,   5   ,   29  ,   29  ,   249 ,   181 ,   641 ,   17  ,   3187    ,   6767    ,0 };
            const unsigned long dim629JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   13  ,   25  ,   17  ,   91  ,   187 ,   61  ,   305 ,   199 ,   2397    ,   2287    ,0 };
            const unsigned long dim630JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   1   ,   25  ,   45  ,   85  ,   35  ,   361 ,   353 ,   631 ,   1113    ,   135 ,0 };
            const unsigned long dim631JoeKuoD6Init[]    =   {   1   ,   3   ,   7   ,   9   ,   13  ,   11  ,   107 ,   67  ,   127 ,   523 ,   1129    ,   81  ,   7627    ,0 };
            const unsigned long dim632JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   7   ,   17  ,   43  ,   89  ,   225 ,   45  ,   597 ,   1683    ,   943 ,   3801    ,0 };
            const unsigned long dim633JoeKuoD6Init[]    =   {   1   ,   3   ,   7   ,   1   ,   31  ,   47  ,   37  ,   31  ,   269 ,   731 ,   1817    ,   3977    ,   7107    ,0 };
            const unsigned long dim634JoeKuoD6Init[]    =   {   1   ,   1   ,   5   ,   5   ,   21  ,   37  ,   93  ,   219 ,   117 ,   369 ,   561 ,   4031    ,   365 ,0 };
            const unsigned long dim635JoeKuoD6Init[]    =   {   1   ,   1   ,   5   ,   13  ,   31  ,   21  ,   65  ,   237 ,   157 ,   775 ,   843 ,   1725    ,   583 ,0 };
            const unsigned long dim636JoeKuoD6Init[]    =   {   1   ,   1   ,   1   ,   11  ,   11  ,   31  ,   15  ,   217 ,   377 ,   395 ,   1633    ,   3841    ,   4417    ,0 };
            const unsigned long dim637JoeKuoD6Init[]    =   {   1   ,   3   ,   7   ,   9   ,   23  ,   23  ,   93  ,   109 ,   265 ,   379 ,   1341    ,   1857    ,   5567    ,0 };
            const unsigned long dim638JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   1   ,   13  ,   37  ,   109 ,   121 ,   7   ,   157 ,   1639    ,   483 ,   1433    ,0 };
            const unsigned long dim639JoeKuoD6Init[]    =   {   1   ,   1   ,   1   ,   3   ,   5   ,   37  ,   29  ,   107 ,   77  ,   771 ,   1631    ,   3051    ,   8097    ,0 };
            const unsigned long dim640JoeKuoD6Init[]    =   {   1   ,   3   ,   7   ,   13  ,   27  ,   5   ,   59  ,   89  ,   401 ,   223 ,   1013    ,   1089    ,   3407    ,0 };
            const unsigned long dim641JoeKuoD6Init[]    =   {   1   ,   1   ,   5   ,   5   ,   11  ,   47  ,   113 ,   237 ,   69  ,   833 ,   845 ,   179 ,   3687    ,0 };
            const unsigned long dim642JoeKuoD6Init[]    =   {   1   ,   3   ,   7   ,   3   ,   19  ,   43  ,   117 ,   223 ,   373 ,   145 ,   463 ,   1719    ,   3621    ,0 };
            const unsigned long dim643JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   7   ,   5   ,   53  ,   5   ,   53  ,   385 ,   723 ,   809 ,   1785    ,   6495    ,0 };
            const unsigned long dim644JoeKuoD6Init[]    =   {   1   ,   3   ,   7   ,   5   ,   3   ,   63  ,   125 ,   63  ,   275 ,   575 ,   1461    ,   821 ,   5607    ,0 };
            const unsigned long dim645JoeKuoD6Init[]    =   {   1   ,   1   ,   5   ,   9   ,   17  ,   33  ,   71  ,   145 ,   215 ,   741 ,   1843    ,   1987    ,   4263    ,0 };
            const unsigned long dim646JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   3   ,   31  ,   63  ,   29  ,   233 ,   441 ,   447 ,   853 ,   3937    ,   4747    ,0 };
            const unsigned long dim647JoeKuoD6Init[]    =   {   1   ,   3   ,   3   ,   15  ,   21  ,   35  ,   15  ,   221 ,   63  ,   481 ,   403 ,   2795    ,   5533    ,0 };
            const unsigned long dim648JoeKuoD6Init[]    =   {   1   ,   3   ,   7   ,   9   ,   1   ,   47  ,   123 ,   129 ,   365 ,   23  ,   1975    ,   1179    ,   7957    ,0 };
            const unsigned long dim649JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   3   ,   11  ,   23  ,   119 ,   229 ,   141 ,   687 ,   1639    ,   2105    ,   319 ,0 };
            const unsigned long dim650JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   11  ,   17  ,   49  ,   3   ,   113 ,   297 ,   727 ,   1135    ,   2657    ,   187 ,0 };
            const unsigned long dim651JoeKuoD6Init[]    =   {   1   ,   3   ,   7   ,   5   ,   5   ,   31  ,   83  ,   173 ,   159 ,   389 ,   1373    ,   3049    ,   827 ,0 };
            const unsigned long dim652JoeKuoD6Init[]    =   {   1   ,   1   ,   1   ,   15  ,   27  ,   51  ,   107 ,   97  ,   119 ,   831 ,   247 ,   2613    ,   7125    ,0 };
            const unsigned long dim653JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   7   ,   23  ,   53  ,   35  ,   181 ,   81  ,   213 ,   1135    ,   2901    ,   1555    ,0 };
            const unsigned long dim654JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   7   ,   9   ,   45  ,   25  ,   137 ,   399 ,   345 ,   339 ,   1869    ,   3469    ,0 };
            const unsigned long dim655JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   5   ,   25  ,   49  ,   71  ,   139 ,   473 ,   883 ,   677 ,   1229    ,   6661    ,0 };
            const unsigned long dim656JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   13  ,   1   ,   55  ,   31  ,   67  ,   321 ,   221 ,   1425    ,   1335    ,   3805    ,0 };
            const unsigned long dim657JoeKuoD6Init[]    =   {   1   ,   1   ,   5   ,   9   ,   7   ,   31  ,   53  ,   169 ,   67  ,   509 ,   1155    ,   1739    ,   615 ,0 };
            const unsigned long dim658JoeKuoD6Init[]    =   {   1   ,   3   ,   7   ,   9   ,   1   ,   13  ,   49  ,   39  ,   263 ,   215 ,   767 ,   2881    ,   1517    ,0 };
            const unsigned long dim659JoeKuoD6Init[]    =   {   1   ,   1   ,   5   ,   5   ,   27  ,   41  ,   15  ,   5   ,   203 ,   539 ,   463 ,   1983    ,   1949    ,0 };
            const unsigned long dim660JoeKuoD6Init[]    =   {   1   ,   3   ,   3   ,   1   ,   21  ,   3   ,   101 ,   225 ,   255 ,   805 ,   1035    ,   813 ,   8125    ,0 };
            const unsigned long dim661JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   1   ,   17  ,   59  ,   55  ,   107 ,   117 ,   489 ,   541 ,   689 ,   581 ,0 };
            const unsigned long dim662JoeKuoD6Init[]    =   {   1   ,   3   ,   7   ,   15  ,   17  ,   57  ,   75  ,   135 ,   295 ,   11  ,   1415    ,   1843    ,   7079    ,0 };
            const unsigned long dim663JoeKuoD6Init[]    =   {   1   ,   1   ,   5   ,   5   ,   17  ,   49  ,   49  ,   91  ,   227 ,   295 ,   1501    ,   3723    ,   5879    ,0 };
            const unsigned long dim664JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   9   ,   27  ,   17  ,   63  ,   127 ,   503 ,   869 ,   53  ,   627 ,   5085    ,0 };
            const unsigned long dim665JoeKuoD6Init[]    =   {   1   ,   3   ,   7   ,   3   ,   25  ,   25  ,   1   ,   83  ,   461 ,   1019    ,   1737    ,   1901    ,   2611    ,0 };
            const unsigned long dim666JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   11  ,   23  ,   59  ,   59  ,   235 ,   33  ,   747 ,   1141    ,   2429    ,   3833    ,0 };
            const unsigned long dim667JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   9   ,   29  ,   55  ,   85  ,   179 ,   385 ,   645 ,   1009    ,   1193    ,   5643    ,0 };
            const unsigned long dim668JoeKuoD6Init[]    =   {   1   ,   1   ,   5   ,   15  ,   29  ,   11  ,   77  ,   147 ,   453 ,   655 ,   469 ,   2727    ,   6205    ,0 };
            const unsigned long dim669JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   3   ,   5   ,   13  ,   103 ,   115 ,   445 ,   87  ,   1803    ,   601 ,   5159    ,0 };
            const unsigned long dim670JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   5   ,   1   ,   37  ,   55  ,   129 ,   5   ,   387 ,   51  ,   3585    ,   6771    ,0 };
            const unsigned long dim671JoeKuoD6Init[]    =   {   1   ,   3   ,   3   ,   15  ,   15  ,   15  ,   21  ,   13  ,   67  ,   391 ,   1745    ,   1137    ,   7039    ,0 };
            const unsigned long dim672JoeKuoD6Init[]    =   {   1   ,   3   ,   7   ,   9   ,   25  ,   1   ,   83  ,   135 ,   275 ,   11  ,   339 ,   269 ,   6121    ,0 };
            const unsigned long dim673JoeKuoD6Init[]    =   {   1   ,   1   ,   5   ,   3   ,   27  ,   11  ,   109 ,   15  ,   499 ,   489 ,   615 ,   1713    ,   7713    ,0 };
            const unsigned long dim674JoeKuoD6Init[]    =   {   1   ,   3   ,   7   ,   15  ,   25  ,   53  ,   1   ,   169 ,   275 ,   623 ,   411 ,   195 ,   6331    ,0 };
            const unsigned long dim675JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   7   ,   3   ,   5   ,   73  ,   97  ,   407 ,   687 ,   311 ,   2141    ,   6687    ,0 };
            const unsigned long dim676JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   3   ,   27  ,   43  ,   123 ,   143 ,   99  ,   963 ,   2029    ,   3759    ,   1483    ,0 };
            const unsigned long dim677JoeKuoD6Init[]    =   {   1   ,   1   ,   5   ,   3   ,   5   ,   43  ,   11  ,   51  ,   423 ,   111 ,   875 ,   3605    ,   4023    ,0 };
            const unsigned long dim678JoeKuoD6Init[]    =   {   1   ,   1   ,   1   ,   15  ,   25  ,   11  ,   89  ,   21  ,   425 ,   585 ,   1983    ,   3481    ,   7329    ,0 };
            const unsigned long dim679JoeKuoD6Init[]    =   {   1   ,   3   ,   7   ,   3   ,   1   ,   5   ,   107 ,   29  ,   285 ,   929 ,   663 ,   319 ,   3139    ,0 };
            const unsigned long dim680JoeKuoD6Init[]    =   {   1   ,   3   ,   3   ,   7   ,   7   ,   37  ,   21  ,   51  ,   385 ,   539 ,   59  ,   1595    ,   9   ,0 };
            const unsigned long dim681JoeKuoD6Init[]    =   {   1   ,   3   ,   3   ,   1   ,   9   ,   21  ,   87  ,   149 ,   21  ,   211 ,   787 ,   351 ,   5597    ,0 };
            const unsigned long dim682JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   11  ,   1   ,   41  ,   21  ,   65  ,   7   ,   83  ,   693 ,   1905    ,   5935    ,0 };
            const unsigned long dim683JoeKuoD6Init[]    =   {   1   ,   3   ,   3   ,   11  ,   17  ,   3   ,   55  ,   205 ,   461 ,   897 ,   273 ,   2849    ,   1685    ,0 };
            const unsigned long dim684JoeKuoD6Init[]    =   {   1   ,   3   ,   7   ,   5   ,   23  ,   15  ,   121 ,   129 ,   387 ,   967 ,   1131    ,   3451    ,   3545    ,0 };
            const unsigned long dim685JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   11  ,   5   ,   29  ,   59  ,   69  ,   27  ,   699 ,   1639    ,   2137    ,   6305    ,0 };
            const unsigned long dim686JoeKuoD6Init[]    =   {   1   ,   3   ,   7   ,   7   ,   23  ,   23  ,   69  ,   205 ,   399 ,   661 ,   1187    ,   1639    ,   5099    ,0 };
            const unsigned long dim687JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   5   ,   31  ,   17  ,   37  ,   173 ,   371 ,   649 ,   449 ,   243 ,   3665    ,0 };
            const unsigned long dim688JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   1   ,   21  ,   5   ,   51  ,   105 ,   173 ,   699 ,   475 ,   3489    ,   7411    ,0 };
            const unsigned long dim689JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   13  ,   17  ,   41  ,   55  ,   131 ,   359 ,   461 ,   1885    ,   4045    ,   3615    ,0 };
            const unsigned long dim690JoeKuoD6Init[]    =   {   1   ,   3   ,   7   ,   11  ,   27  ,   33  ,   79  ,   127 ,   227 ,   293 ,   607 ,   1713    ,   167 ,0 };
            const unsigned long dim691JoeKuoD6Init[]    =   {   1   ,   3   ,   3   ,   1   ,   5   ,   9   ,   13  ,   59  ,   425 ,   319 ,   1813    ,   2905    ,   4647    ,0 };
            const unsigned long dim692JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   5   ,   13  ,   59  ,   69  ,   95  ,   379 ,   119 ,   1175    ,   1853    ,   7573    ,0 };
            const unsigned long dim693JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   3   ,   3   ,   47  ,   51  ,   203 ,   395 ,   623 ,   13  ,   3369    ,   663 ,0 };
            const unsigned long dim694JoeKuoD6Init[]    =   {   1   ,   1   ,   5   ,   3   ,   3   ,   39  ,   107 ,   203 ,   125 ,   627 ,   1817    ,   1101    ,   935 ,0 };
            const unsigned long dim695JoeKuoD6Init[]    =   {   1   ,   1   ,   1   ,   5   ,   7   ,   31  ,   71  ,   141 ,   507 ,   101 ,   913 ,   567 ,   753 ,0 };
            const unsigned long dim696JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   11  ,   1   ,   25  ,   81  ,   211 ,   79  ,   277 ,   1445    ,   783 ,   5   ,0 };
            const unsigned long dim697JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   3   ,   7   ,   9   ,   63  ,   87  ,   335 ,   413 ,   1595    ,   2313    ,   5883    ,0 };
            const unsigned long dim698JoeKuoD6Init[]    =   {   1   ,   1   ,   1   ,   11  ,   21  ,   61  ,   87  ,   119 ,   155 ,   259 ,   1629    ,   2565    ,   5975    ,0 };
            const unsigned long dim699JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   3   ,   29  ,   59  ,   57  ,   205 ,   65  ,   779 ,   1433    ,   2845    ,   1311    ,0 };
            const unsigned long dim700JoeKuoD6Init[]    =   {   1   ,   3   ,   3   ,   3   ,   29  ,   59  ,   1   ,   125 ,   493 ,   25  ,   315 ,   1171    ,   4837    ,0 };
            const unsigned long dim701JoeKuoD6Init[]    =   {   1   ,   1   ,   1   ,   11  ,   1   ,   47  ,   119 ,   141 ,   95  ,   87  ,   1497    ,   3613    ,   5009    ,0 };
            const unsigned long dim702JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   5   ,   7   ,   19  ,   61  ,   233 ,   321 ,   585 ,   547 ,   179 ,   6787    ,0 };
            const unsigned long dim703JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   7   ,   15  ,   29  ,   51  ,   177 ,   189 ,   435 ,   89  ,   2507    ,   4975    ,0 };
            const unsigned long dim704JoeKuoD6Init[]    =   {   1   ,   3   ,   7   ,   5   ,   19  ,   49  ,   125 ,   145 ,   329 ,   975 ,   717 ,   161 ,   1099    ,0 };
            const unsigned long dim705JoeKuoD6Init[]    =   {   1   ,   3   ,   3   ,   13  ,   1   ,   21  ,   81  ,   63  ,   263 ,   597 ,   1713    ,   1759    ,   1745    ,0 };
            const unsigned long dim706JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   9   ,   17  ,   61  ,   19  ,   119 ,   229 ,   25  ,   1759    ,   343 ,   7925    ,0 };
            const unsigned long dim707JoeKuoD6Init[]    =   {   1   ,   1   ,   1   ,   9   ,   27  ,   49  ,   51  ,   149 ,   453 ,   717 ,   1791    ,   2529    ,   331 ,0 };
            const unsigned long dim708JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   7   ,   11  ,   15  ,   125 ,   237 ,   291 ,   111 ,   1051    ,   3739    ,   3805    ,0 };
            const unsigned long dim709JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   9   ,   7   ,   63  ,   49  ,   73  ,   53  ,   429 ,   1167    ,   2019    ,   911 ,0 };
            const unsigned long dim710JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   5   ,   29  ,   17  ,   3   ,   193 ,   243 ,   29  ,   2007    ,   1881    ,   7143    ,0 };
            const unsigned long dim711JoeKuoD6Init[]    =   {   1   ,   3   ,   3   ,   3   ,   5   ,   19  ,   127 ,   149 ,   57  ,   601 ,   1263    ,   1909    ,   3835    ,0 };
            const unsigned long dim712JoeKuoD6Init[]    =   {   1   ,   1   ,   1   ,   5   ,   25  ,   43  ,   21  ,   115 ,   193 ,   735 ,   571 ,   401 ,   1573    ,0 };
            const unsigned long dim713JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   9   ,   25  ,   41  ,   53  ,   165 ,   5   ,   575 ,   1381    ,   995 ,   5851    ,0 };
            const unsigned long dim714JoeKuoD6Init[]    =   {   1   ,   1   ,   5   ,   15  ,   21  ,   43  ,   25  ,   125 ,   49  ,   207 ,   345 ,   2981    ,   709 ,0 };
            const unsigned long dim715JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   5   ,   29  ,   37  ,   89  ,   145 ,   123 ,   771 ,   631 ,   3259    ,   1431    ,0 };
            const unsigned long dim716JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   3   ,   1   ,   37  ,   67  ,   243 ,   335 ,   913 ,   1927    ,   3027    ,   6059    ,0 };
            const unsigned long dim717JoeKuoD6Init[]    =   {   1   ,   1   ,   5   ,   15  ,   11  ,   31  ,   61  ,   59  ,   511 ,   543 ,   1747    ,   3329    ,   599 ,0 };
            const unsigned long dim718JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   13  ,   17  ,   31  ,   9   ,   143 ,   381 ,   713 ,   135 ,   1037    ,   5981    ,0 };
            const unsigned long dim719JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   11  ,   25  ,   7   ,   79  ,   3   ,   403 ,   979 ,   1021    ,   457 ,   999 ,0 };
            const unsigned long dim720JoeKuoD6Init[]    =   {   1   ,   3   ,   3   ,   11  ,   3   ,   19  ,   127 ,   135 ,   255 ,   645 ,   437 ,   3793    ,   5809    ,0 };
            const unsigned long dim721JoeKuoD6Init[]    =   {   1   ,   3   ,   7   ,   1   ,   5   ,   59  ,   69  ,   135 ,   307 ,   769 ,   307 ,   3225    ,   2869    ,0 };
            const unsigned long dim722JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   3   ,   9   ,   33  ,   121 ,   41  ,   245 ,   1009    ,   2039    ,   925 ,   3341    ,0 };
            const unsigned long dim723JoeKuoD6Init[]    =   {   1   ,   3   ,   3   ,   15  ,   19  ,   23  ,   55  ,   155 ,   389 ,   817 ,   475 ,   1005    ,   1971    ,0 };
            const unsigned long dim724JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   7   ,   11  ,   17  ,   7   ,   113 ,   249 ,   241 ,   1149    ,   2359    ,   8037    ,0 };
            const unsigned long dim725JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   5   ,   23  ,   59  ,   115 ,   5   ,   503 ,   221 ,   485 ,   3857    ,   4399    ,0 };
            const unsigned long dim726JoeKuoD6Init[]    =   {   1   ,   3   ,   7   ,   3   ,   5   ,   53  ,   119 ,   107 ,   103 ,   99  ,   1313    ,   623 ,   3073    ,0 };
            const unsigned long dim727JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   13  ,   7   ,   21  ,   85  ,   47  ,   345 ,   355 ,   607 ,   2213    ,   2489    ,0 };
            const unsigned long dim728JoeKuoD6Init[]    =   {   1   ,   1   ,   5   ,   11  ,   7   ,   51  ,   43  ,   53  ,   281 ,   311 ,   1221    ,   3919    ,   4821    ,0 };
            const unsigned long dim729JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   9   ,   21  ,   15  ,   105 ,   251 ,   367 ,   311 ,   1247    ,   2357    ,   5411    ,0 };
            const unsigned long dim730JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   5   ,   17  ,   21  ,   1   ,   213 ,   505 ,   97  ,   603 ,   4057    ,   6853    ,0 };
            const unsigned long dim731JoeKuoD6Init[]    =   {   1   ,   3   ,   7   ,   5   ,   21  ,   3   ,   105 ,   61  ,   5   ,   39  ,   461 ,   351 ,   4379    ,0 };
            const unsigned long dim732JoeKuoD6Init[]    =   {   1   ,   3   ,   3   ,   9   ,   15  ,   39  ,   87  ,   21  ,   393 ,   525 ,   1699    ,   2427    ,   211 ,0 };
            const unsigned long dim733JoeKuoD6Init[]    =   {   1   ,   1   ,   5   ,   9   ,   21  ,   25  ,   53  ,   175 ,   387 ,   83  ,   1155    ,   1225    ,   5681    ,0 };
            const unsigned long dim734JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   15  ,   7   ,   49  ,   43  ,   127 ,   257 ,   499 ,   79  ,   33  ,   2843    ,0 };
            const unsigned long dim735JoeKuoD6Init[]    =   {   1   ,   3   ,   7   ,   15  ,   21  ,   17  ,   89  ,   175 ,   161 ,   371 ,   389 ,   2289    ,   7751    ,0 };
            const unsigned long dim736JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   3   ,   19  ,   23  ,   65  ,   5   ,   149 ,   563 ,   1237    ,   3445    ,   6817    ,0 };
            const unsigned long dim737JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   11  ,   23  ,   21  ,   65  ,   11  ,   329 ,   323 ,   551 ,   3859    ,   749 ,0 };
            const unsigned long dim738JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   9   ,   13  ,   23  ,   11  ,   169 ,   203 ,   349 ,   1733    ,   2537    ,   3777    ,0 };
            const unsigned long dim739JoeKuoD6Init[]    =   {   1   ,   3   ,   7   ,   7   ,   5   ,   19  ,   47  ,   81  ,   215 ,   677 ,   1861    ,   2703    ,   6373    ,0 };
            const unsigned long dim740JoeKuoD6Init[]    =   {   1   ,   1   ,   5   ,   5   ,   11  ,   5   ,   19  ,   151 ,   215 ,   695 ,   1347    ,   1359    ,   1979    ,0 };
            const unsigned long dim741JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   3   ,   1   ,   59  ,   11  ,   185 ,   379 ,   247 ,   1885    ,   1819    ,   637 ,0 };
            const unsigned long dim742JoeKuoD6Init[]    =   {   1   ,   3   ,   7   ,   3   ,   21  ,   3   ,   53  ,   161 ,   49  ,   697 ,   113 ,   3561    ,   95  ,0 };
            const unsigned long dim743JoeKuoD6Init[]    =   {   1   ,   3   ,   7   ,   7   ,   29  ,   15  ,   85  ,   123 ,   415 ,   41  ,   703 ,   1501    ,   6939    ,0 };
            const unsigned long dim744JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   5   ,   7   ,   37  ,   75  ,   185 ,   115 ,   423 ,   625 ,   3557    ,   7851    ,0 };
            const unsigned long dim745JoeKuoD6Init[]    =   {   1   ,   1   ,   1   ,   7   ,   11  ,   35  ,   119 ,   131 ,   331 ,   197 ,   1939    ,   1951    ,   7957    ,0 };
            const unsigned long dim746JoeKuoD6Init[]    =   {   1   ,   1   ,   5   ,   3   ,   9   ,   43  ,   93  ,   11  ,   97  ,   483 ,   131 ,   1095    ,   601 ,0 };
            const unsigned long dim747JoeKuoD6Init[]    =   {   1   ,   1   ,   1   ,   15  ,   11  ,   43  ,   49  ,   205 ,   393 ,   921 ,   625 ,   359 ,   3997    ,0 };
            const unsigned long dim748JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   7   ,   31  ,   21  ,   27  ,   231 ,   159 ,   67  ,   1127    ,   2153    ,   6291    ,0 };
            const unsigned long dim749JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   1   ,   13  ,   15  ,   45  ,   85  ,   155 ,   763 ,   1341    ,   9   ,   6229    ,0 };
            const unsigned long dim750JoeKuoD6Init[]    =   {   1   ,   3   ,   7   ,   15  ,   9   ,   21  ,   47  ,   67  ,   207 ,   385 ,   401 ,   1665    ,   7849    ,0 };
            const unsigned long dim751JoeKuoD6Init[]    =   {   1   ,   1   ,   5   ,   9   ,   13  ,   59  ,   89  ,   163 ,   19  ,   997 ,   2021    ,   3301    ,   2945    ,0 };
            const unsigned long dim752JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   15  ,   17  ,   9   ,   33  ,   117 ,   33  ,   701 ,   73  ,   2955    ,   1781    ,0 };
            const unsigned long dim753JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   7   ,   3   ,   11  ,   81  ,   209 ,   131 ,   247 ,   1927    ,   3997    ,   1709    ,0 };
            const unsigned long dim754JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   5   ,   21  ,   47  ,   85  ,   163 ,   247 ,   273 ,   545 ,   3477    ,   4565    ,0 };
            const unsigned long dim755JoeKuoD6Init[]    =   {   1   ,   1   ,   1   ,   15  ,   3   ,   5   ,   55  ,   171 ,   501 ,   149 ,   121 ,   2317    ,   2023    ,0 };
            const unsigned long dim756JoeKuoD6Init[]    =   {   1   ,   1   ,   5   ,   1   ,   25  ,   31  ,   25  ,   171 ,   471 ,   149 ,   49  ,   3059    ,   7871    ,0 };
            const unsigned long dim757JoeKuoD6Init[]    =   {   1   ,   3   ,   3   ,   1   ,   3   ,   39  ,   91  ,   75  ,   183 ,   115 ,   869 ,   4005    ,   1443    ,0 };
            const unsigned long dim758JoeKuoD6Init[]    =   {   1   ,   1   ,   1   ,   13  ,   31  ,   37  ,   3   ,   23  ,   443 ,   177 ,   219 ,   1683    ,   7063    ,0 };
            const unsigned long dim759JoeKuoD6Init[]    =   {   1   ,   1   ,   1   ,   11  ,   17  ,   51  ,   111 ,   157 ,   455 ,   415 ,   1171    ,   531 ,   7289    ,0 };
            const unsigned long dim760JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   13  ,   3   ,   45  ,   103 ,   185 ,   63  ,   299 ,   1471    ,   3579    ,   3751    ,0 };
            const unsigned long dim761JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   7   ,   25  ,   45  ,   91  ,   145 ,   285 ,   525 ,   1675    ,   665 ,   883 ,0 };
            const unsigned long dim762JoeKuoD6Init[]    =   {   1   ,   3   ,   7   ,   5   ,   23  ,   23  ,   33  ,   47  ,   271 ,   65  ,   1009    ,   2363    ,   7231    ,0 };
            const unsigned long dim763JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   11  ,   25  ,   3   ,   13  ,   105 ,   105 ,   649 ,   531 ,   535 ,   6019    ,0 };
            const unsigned long dim764JoeKuoD6Init[]    =   {   1   ,   3   ,   3   ,   5   ,   9   ,   1   ,   9   ,   15  ,   131 ,   793 ,   1841    ,   2249    ,   5775    ,0 };
            const unsigned long dim765JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   7   ,   27  ,   35  ,   125 ,   249 ,   125 ,   973 ,   233 ,   229 ,   8165    ,0 };
            const unsigned long dim766JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   7   ,   19  ,   1   ,   39  ,   239 ,   159 ,   627 ,   601 ,   1537    ,   7473    ,0 };
            const unsigned long dim767JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   3   ,   1   ,   45  ,   99  ,   149 ,   205 ,   277 ,   1701    ,   121 ,   1577    ,0 };
            const unsigned long dim768JoeKuoD6Init[]    =   {   1   ,   3   ,   7   ,   15  ,   9   ,   43  ,   7   ,   139 ,   331 ,   995 ,   647 ,   2707    ,   4493    ,0 };
            const unsigned long dim769JoeKuoD6Init[]    =   {   1   ,   3   ,   3   ,   13  ,   19  ,   31  ,   67  ,   213 ,   355 ,   85  ,   1829    ,   2897    ,   6799    ,0 };
            const unsigned long dim770JoeKuoD6Init[]    =   {   1   ,   3   ,   7   ,   1   ,   29  ,   51  ,   49  ,   209 ,   257 ,   913 ,   811 ,   895 ,   8051    ,0 };
            const unsigned long dim771JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   3   ,   31  ,   33  ,   19  ,   37  ,   297 ,   473 ,   1987    ,   3433    ,   1981    ,0 };
            const unsigned long dim772JoeKuoD6Init[]    =   {   1   ,   1   ,   1   ,   13  ,   1   ,   3   ,   127 ,   183 ,   241 ,   941 ,   973 ,   577 ,   6891    ,0 };
            const unsigned long dim773JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   15  ,   31  ,   21  ,   125 ,   193 ,   23  ,   731 ,   493 ,   3173    ,   4381    ,0 };
            const unsigned long dim774JoeKuoD6Init[]    =   {   1   ,   3   ,   3   ,   13  ,   23  ,   31  ,   65  ,   59  ,   63  ,   285 ,   1191    ,   3005    ,   2211    ,0 };
            const unsigned long dim775JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   3   ,   13  ,   11  ,   121 ,   153 ,   199 ,   359 ,   1159    ,   363 ,   3477    ,0 };
            const unsigned long dim776JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   3   ,   29  ,   27  ,   105 ,   215 ,   143 ,   365 ,   369 ,   3341    ,   5017    ,0 };
            const unsigned long dim777JoeKuoD6Init[]    =   {   1   ,   1   ,   5   ,   9   ,   23  ,   43  ,   119 ,   89  ,   307 ,   607 ,   1087    ,   287 ,   7305    ,0 };
            const unsigned long dim778JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   3   ,   19  ,   9   ,   115 ,   133 ,   387 ,   717 ,   687 ,   3447    ,   201 ,0 };
            const unsigned long dim779JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   1   ,   23  ,   7   ,   81  ,   77  ,   483 ,   725 ,   1499    ,   1565    ,   8073    ,0 };
            const unsigned long dim780JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   9   ,   23  ,   25  ,   119 ,   247 ,   195 ,   205 ,   1851    ,   2463    ,   6827    ,0 };
            const unsigned long dim781JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   9   ,   17  ,   51  ,   11  ,   233 ,   115 ,   719 ,   1333    ,   3829    ,   4941    ,0 };
            const unsigned long dim782JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   13  ,   25  ,   17  ,   103 ,   241 ,   133 ,   237 ,   1375    ,   4005    ,   7853    ,0 };
            const unsigned long dim783JoeKuoD6Init[]    =   {   1   ,   3   ,   7   ,   1   ,   19  ,   31  ,   115 ,   193 ,   305 ,   1   ,   1167    ,   325 ,   985 ,0 };
            const unsigned long dim784JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   15  ,   13  ,   11  ,   33  ,   27  ,   453 ,   941 ,   293 ,   2819    ,   3247    ,0 };
            const unsigned long dim785JoeKuoD6Init[]    =   {   1   ,   3   ,   3   ,   3   ,   11  ,   27  ,   37  ,   95  ,   371 ,   441 ,   1347    ,   13  ,   4219    ,0 };
            const unsigned long dim786JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   11  ,   31  ,   39  ,   17  ,   63  ,   323 ,   1013    ,   479 ,   4077    ,   313 ,0 };
            const unsigned long dim787JoeKuoD6Init[]    =   {   1   ,   3   ,   7   ,   15  ,   3   ,   59  ,   121 ,   19  ,   43  ,   351 ,   285 ,   2239    ,   8117    ,0 };
            const unsigned long dim788JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   3   ,   7   ,   7   ,   13  ,   31  ,   403 ,   925 ,   789 ,   275 ,   6959    ,0 };
            const unsigned long dim789JoeKuoD6Init[]    =   {   1   ,   1   ,   1   ,   9   ,   3   ,   11  ,   89  ,   195 ,   393 ,   301 ,   517 ,   1675    ,   963 ,0 };
            const unsigned long dim790JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   13  ,   17  ,   63  ,   79  ,   17  ,   127 ,   119 ,   1127    ,   2121    ,   7545    ,0 };
            const unsigned long dim791JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   5   ,   29  ,   15  ,   97  ,   183 ,   157 ,   319 ,   1047    ,   895 ,   7297    ,0 };
            const unsigned long dim792JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   3   ,   25  ,   37  ,   101 ,   233 ,   275 ,   705 ,   1147    ,   1511    ,   231 ,0 };
            const unsigned long dim793JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   11  ,   3   ,   63  ,   45  ,   219 ,   315 ,   905 ,   9   ,   3105    ,   785 ,0 };
            const unsigned long dim794JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   1   ,   17  ,   27  ,   89  ,   15  ,   337 ,   981 ,   403 ,   1609    ,   349 ,0 };
            const unsigned long dim795JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   5   ,   27  ,   21  ,   39  ,   209 ,   461 ,   211 ,   71  ,   1413    ,   1185    ,0 };
            const unsigned long dim796JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   1   ,   11  ,   9   ,   19  ,   229 ,   403 ,   323 ,   1651    ,   475 ,   1207    ,0 };
            const unsigned long dim797JoeKuoD6Init[]    =   {   1   ,   3   ,   7   ,   5   ,   31  ,   13  ,   39  ,   9   ,   365 ,   857 ,   1715    ,   3831    ,   2467    ,0 };
            const unsigned long dim798JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   1   ,   31  ,   41  ,   25  ,   101 ,   297 ,   143 ,   345 ,   2097    ,   1679    ,0 };
            const unsigned long dim799JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   7   ,   27  ,   61  ,   77  ,   107 ,   291 ,   265 ,   759 ,   933 ,   2529    ,0 };
            const unsigned long dim800JoeKuoD6Init[]    =   {   1   ,   3   ,   7   ,   11  ,   15  ,   51  ,   31  ,   191 ,   39  ,   469 ,   1711    ,   1215    ,   7603    ,0 };
            const unsigned long dim801JoeKuoD6Init[]    =   {   1   ,   1   ,   5   ,   13  ,   19  ,   27  ,   121 ,   235 ,   475 ,   11  ,   1193    ,   543 ,   5689    ,0 };
            const unsigned long dim802JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   11  ,   19  ,   1   ,   97  ,   85  ,   393 ,   971 ,   1915    ,   3487    ,   7569    ,0 };
            const unsigned long dim803JoeKuoD6Init[]    =   {   1   ,   3   ,   7   ,   9   ,   21  ,   25  ,   81  ,   197 ,   9   ,   403 ,   1305    ,   2977    ,   6261    ,0 };
            const unsigned long dim804JoeKuoD6Init[]    =   {   1   ,   3   ,   3   ,   15  ,   5   ,   3   ,   5   ,   91  ,   133 ,   957 ,   437 ,   3821    ,   4395    ,0 };
            const unsigned long dim805JoeKuoD6Init[]    =   {   1   ,   3   ,   7   ,   7   ,   23  ,   47  ,   85  ,   139 ,   511 ,   389 ,   461 ,   2967    ,   7305    ,0 };
            const unsigned long dim806JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   5   ,   13  ,   3   ,   33  ,   171 ,   391 ,   109 ,   197 ,   2097    ,   5357    ,0 };
            const unsigned long dim807JoeKuoD6Init[]    =   {   1   ,   3   ,   7   ,   15  ,   31  ,   61  ,   7   ,   61  ,   175 ,   455 ,   1867    ,   2829    ,   197 ,0 };
            const unsigned long dim808JoeKuoD6Init[]    =   {   1   ,   3   ,   7   ,   13  ,   21  ,   33  ,   119 ,   97  ,   297 ,   69  ,   1503    ,   2887    ,   3785    ,0 };
            const unsigned long dim809JoeKuoD6Init[]    =   {   1   ,   3   ,   3   ,   7   ,   1   ,   63  ,   105 ,   53  ,   443 ,   775 ,   1855    ,   2635    ,   4453    ,0 };
            const unsigned long dim810JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   11  ,   1   ,   13  ,   99  ,   227 ,   421 ,   739 ,   583 ,   4011    ,   2095    ,0 };
            const unsigned long dim811JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   11  ,   7   ,   21  ,   125 ,   45  ,   411 ,   621 ,   137 ,   259 ,   3401    ,0 };
            const unsigned long dim812JoeKuoD6Init[]    =   {   1   ,   3   ,   3   ,   7   ,   9   ,   15  ,   47  ,   213 ,   71  ,   725 ,   2015    ,   3119    ,   3295    ,0 };
            const unsigned long dim813JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   7   ,   21  ,   63  ,   83  ,   63  ,   29  ,   969 ,   119 ,   3535    ,   989 ,0 };
            const unsigned long dim814JoeKuoD6Init[]    =   {   1   ,   3   ,   3   ,   9   ,   13  ,   15  ,   71  ,   83  ,   209 ,   191 ,   1049    ,   3423    ,   5829    ,0 };
            const unsigned long dim815JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   11  ,   27  ,   7   ,   65  ,   107 ,   387 ,   853 ,   919 ,   1257    ,   719 ,0 };
            const unsigned long dim816JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   9   ,   29  ,   63  ,   115 ,   91  ,   181 ,   217 ,   1097    ,   1023    ,   3641    ,0 };
            const unsigned long dim817JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   7   ,   7   ,   57  ,   37  ,   209 ,   99  ,   959 ,   693 ,   1573    ,   3407    ,0 };
            const unsigned long dim818JoeKuoD6Init[]    =   {   1   ,   3   ,   7   ,   7   ,   9   ,   51  ,   51  ,   127 ,   153 ,   977 ,   1565    ,   1329    ,   4603    ,0 };
            const unsigned long dim819JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   3   ,   9   ,   57  ,   7   ,   225 ,   139 ,   515 ,   1355    ,   191 ,   659 ,0 };
            const unsigned long dim820JoeKuoD6Init[]    =   {   1   ,   3   ,   3   ,   3   ,   7   ,   45  ,   85  ,   231 ,   153 ,   93  ,   371 ,   2753    ,   5331    ,0 };
            const unsigned long dim821JoeKuoD6Init[]    =   {   1   ,   1   ,   5   ,   5   ,   1   ,   9   ,   45  ,   29  ,   131 ,   203 ,   31  ,   1219    ,   1635    ,0 };
            const unsigned long dim822JoeKuoD6Init[]    =   {   1   ,   1   ,   1   ,   11  ,   5   ,   13  ,   115 ,   51  ,   471 ,   987 ,   1937    ,   3921    ,   7885    ,0 };
            const unsigned long dim823JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   3   ,   1   ,   23  ,   49  ,   183 ,   495 ,   427 ,   1397    ,   3383    ,   3459    ,0 };
            const unsigned long dim824JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   1   ,   23  ,   39  ,   3   ,   205 ,   447 ,   603 ,   1213    ,   667 ,   1039    ,0 };
            const unsigned long dim825JoeKuoD6Init[]    =   {   1   ,   1   ,   1   ,   5   ,   7   ,   63  ,   27  ,   169 ,   403 ,   775 ,   241 ,   3951    ,   3411    ,0 };
            const unsigned long dim826JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   1   ,   7   ,   1   ,   63  ,   185 ,   197 ,   813 ,   1757    ,   2687    ,   7289    ,0 };
            const unsigned long dim827JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   3   ,   21  ,   21  ,   97  ,   73  ,   73  ,   147 ,   1905    ,   2391    ,   3743    ,0 };
            const unsigned long dim828JoeKuoD6Init[]    =   {   1   ,   3   ,   7   ,   1   ,   29  ,   45  ,   27  ,   93  ,   71  ,   749 ,   337 ,   7   ,   3959    ,0 };
            const unsigned long dim829JoeKuoD6Init[]    =   {   1   ,   3   ,   3   ,   9   ,   25  ,   13  ,   61  ,   131 ,   183 ,   409 ,   1547    ,   1261    ,   5541    ,0 };
            const unsigned long dim830JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   9   ,   9   ,   13  ,   13  ,   115 ,   87  ,   587 ,   1561    ,   1845    ,   2383    ,0 };
            const unsigned long dim831JoeKuoD6Init[]    =   {   1   ,   3   ,   3   ,   3   ,   13  ,   19  ,   99  ,   75  ,   435 ,   815 ,   251 ,   1639    ,   2157    ,0 };
            const unsigned long dim832JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   9   ,   19  ,   61  ,   49  ,   237 ,   325 ,   429 ,   1019    ,   2649    ,   6671    ,0 };
            const unsigned long dim833JoeKuoD6Init[]    =   {   1   ,   3   ,   3   ,   11  ,   1   ,   59  ,   71  ,   143 ,   37  ,   445 ,   1289    ,   2089    ,   5693    ,0 };
            const unsigned long dim834JoeKuoD6Init[]    =   {   1   ,   3   ,   3   ,   7   ,   13  ,   13  ,   119 ,   83  ,   111 ,   497 ,   739 ,   863 ,   2009    ,0 };
            const unsigned long dim835JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   1   ,   3   ,   21  ,   89  ,   27  ,   171 ,   121 ,   1453    ,   3273    ,   2891    ,0 };
            const unsigned long dim836JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   5   ,   29  ,   19  ,   79  ,   77  ,   343 ,   969 ,   1759    ,   1921    ,   4189    ,0 };
            const unsigned long dim837JoeKuoD6Init[]    =   {   1   ,   1   ,   1   ,   3   ,   21  ,   55  ,   99  ,   49  ,   435 ,   697 ,   559 ,   3475    ,   6905    ,0 };
            const unsigned long dim838JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   5   ,   9   ,   37  ,   83  ,   79  ,   171 ,   935 ,   751 ,   3293    ,   299 ,0 };
            const unsigned long dim839JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   13  ,   13  ,   45  ,   35  ,   159 ,   283 ,   451 ,   495 ,   3239    ,   4429    ,0 };
            const unsigned long dim840JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   1   ,   13  ,   43  ,   107 ,   189 ,   475 ,   679 ,   1843    ,   33  ,   6895    ,0 };
            const unsigned long dim841JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   9   ,   17  ,   13  ,   107 ,   245 ,   415 ,   531 ,   265 ,   3165    ,   3189    ,0 };
            const unsigned long dim842JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   3   ,   31  ,   29  ,   103 ,   233 ,   31  ,   445 ,   649 ,   3377    ,   3207    ,0 };
            const unsigned long dim843JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   9   ,   19  ,   3   ,   1   ,   75  ,   423 ,   445 ,   1351    ,   2889    ,   5785    ,0 };
            const unsigned long dim844JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   3   ,   21  ,   19  ,   113 ,   35  ,   173 ,   871 ,   275 ,   1901    ,   6725    ,0 };
            const unsigned long dim845JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   7   ,   27  ,   45  ,   41  ,   93  ,   463 ,   209 ,   1721    ,   1863    ,   7893    ,0 };
            const unsigned long dim846JoeKuoD6Init[]    =   {   1   ,   1   ,   1   ,   13  ,   25  ,   63  ,   17  ,   171 ,   403 ,   645 ,   1669    ,   1663    ,   4335    ,0 };
            const unsigned long dim847JoeKuoD6Init[]    =   {   1   ,   3   ,   3   ,   7   ,   17  ,   15  ,   103 ,   5   ,   375 ,   7   ,   1117    ,   3851    ,   651 ,0 };
            const unsigned long dim848JoeKuoD6Init[]    =   {   1   ,   1   ,   1   ,   15  ,   29  ,   43  ,   13  ,   241 ,   403 ,   349 ,   685 ,   987 ,   4177    ,0 };
            const unsigned long dim849JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   15  ,   31  ,   37  ,   111 ,   91  ,   391 ,   513 ,   1701    ,   2861    ,   4713    ,0 };
            const unsigned long dim850JoeKuoD6Init[]    =   {   1   ,   1   ,   5   ,   9   ,   23  ,   59  ,   125 ,   117 ,   375 ,   461 ,   1697    ,   1283    ,   7395    ,0 };
            const unsigned long dim851JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   7   ,   21  ,   35  ,   101 ,   131 ,   129 ,   757 ,   199 ,   1881    ,   737 ,0 };
            const unsigned long dim852JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   3   ,   31  ,   15  ,   121 ,   143 ,   67  ,   593 ,   1143    ,   3761    ,   3441    ,0 };
            const unsigned long dim853JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   9   ,   11  ,   43  ,   7   ,   159 ,   25  ,   725 ,   1169    ,   1757    ,   3499    ,0 };
            const unsigned long dim854JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   9   ,   5   ,   55  ,   77  ,   85  ,   309 ,   113 ,   1419    ,   1255    ,   4777    ,0 };
            const unsigned long dim855JoeKuoD6Init[]    =   {   1   ,   3   ,   3   ,   15  ,   19  ,   41  ,   75  ,   163 ,   51  ,   511 ,   1785    ,   4033    ,   7791    ,0 };
            const unsigned long dim856JoeKuoD6Init[]    =   {   1   ,   3   ,   7   ,   15  ,   29  ,   45  ,   81  ,   189 ,   401 ,   71  ,   1597    ,   3671    ,   2133    ,0 };
            const unsigned long dim857JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   1   ,   3   ,   17  ,   119 ,   55  ,   217 ,   485 ,   1217    ,   1475    ,   5791    ,0 };
            const unsigned long dim858JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   9   ,   21  ,   61  ,   71  ,   175 ,   205 ,   413 ,   1645    ,   2375    ,   89  ,0 };
            const unsigned long dim859JoeKuoD6Init[]    =   {   1   ,   3   ,   3   ,   7   ,   25  ,   35  ,   23  ,   49  ,   395 ,   675 ,   341 ,   959 ,   4665    ,0 };
            const unsigned long dim860JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   15  ,   21  ,   37  ,   99  ,   151 ,   197 ,   145 ,   2043    ,   755 ,   2523    ,0 };
            const unsigned long dim861JoeKuoD6Init[]    =   {   1   ,   1   ,   1   ,   7   ,   1   ,   61  ,   75  ,   131 ,   65  ,   891 ,   349 ,   2767    ,   3855    ,0 };
            const unsigned long dim862JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   9   ,   29  ,   57  ,   61  ,   79  ,   303 ,   997 ,   1053    ,   1973    ,   1547    ,0 };
            const unsigned long dim863JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   7   ,   19  ,   9   ,   57  ,   197 ,   253 ,   597 ,   1971    ,   1831    ,   8097    ,0 };
            const unsigned long dim864JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   9   ,   3   ,   13  ,   29  ,   169 ,   25  ,   957 ,   1575    ,   2955    ,   3549    ,0 };
            const unsigned long dim865JoeKuoD6Init[]    =   {   1   ,   1   ,   5   ,   11  ,   17  ,   37  ,   79  ,   91  ,   11  ,   687 ,   1511    ,   1065    ,   1539    ,0 };
            const unsigned long dim866JoeKuoD6Init[]    =   {   1   ,   1   ,   1   ,   7   ,   1   ,   59  ,   1   ,   49  ,   23  ,   73  ,   881 ,   1661    ,   3841    ,0 };
            const unsigned long dim867JoeKuoD6Init[]    =   {   1   ,   1   ,   1   ,   1   ,   13  ,   57  ,   111 ,   67  ,   511 ,   833 ,   1095    ,   15  ,   569 ,0 };
            const unsigned long dim868JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   5   ,   17  ,   33  ,   1   ,   161 ,   311 ,   29  ,   685 ,   729 ,   5981    ,0 };
            const unsigned long dim869JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   1   ,   7   ,   13  ,   49  ,   55  ,   251 ,   825 ,   709 ,   445 ,   7439    ,0 };
            const unsigned long dim870JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   5   ,   13  ,   55  ,   101 ,   89  ,   239 ,   907 ,   1559    ,   1421    ,   87  ,0 };
            const unsigned long dim871JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   9   ,   23  ,   27  ,   41  ,   189 ,   35  ,   155 ,   683 ,   2491    ,   7495    ,0 };
            const unsigned long dim872JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   3   ,   5   ,   47  ,   39  ,   161 ,   279 ,   945 ,   365 ,   15  ,   4643    ,0 };
            const unsigned long dim873JoeKuoD6Init[]    =   {   1   ,   1   ,   1   ,   11  ,   21  ,   7   ,   125 ,   195 ,   23  ,   635 ,   1949    ,   2237    ,   1885    ,0 };
            const unsigned long dim874JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   7   ,   13  ,   25  ,   105 ,   145 ,   203 ,   701 ,   1525    ,   1391    ,   1637    ,0 };
            const unsigned long dim875JoeKuoD6Init[]    =   {   1   ,   3   ,   3   ,   13  ,   7   ,   43  ,   13  ,   39  ,   301 ,   507 ,   405 ,   2333    ,   2405    ,0 };
            const unsigned long dim876JoeKuoD6Init[]    =   {   1   ,   1   ,   5   ,   15  ,   25  ,   19  ,   31  ,   67  ,   81  ,   781 ,   1541    ,   3743    ,   2787    ,0 };
            const unsigned long dim877JoeKuoD6Init[]    =   {   1   ,   1   ,   1   ,   11  ,   15  ,   7   ,   87  ,   129 ,   151 ,   225 ,   655 ,   1569    ,   5299    ,0 };
            const unsigned long dim878JoeKuoD6Init[]    =   {   1   ,   1   ,   5   ,   3   ,   3   ,   33  ,   97  ,   255 ,   9   ,   667 ,   787 ,   3557    ,   5273    ,0 };
            const unsigned long dim879JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   3   ,   1   ,   7   ,   13  ,   31  ,   31  ,   1021    ,   1863    ,   2897    ,   6389    ,0 };
            const unsigned long dim880JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   15  ,   13  ,   29  ,   93  ,   27  ,   229 ,   599 ,   521 ,   1631    ,   3683    ,0 };
            const unsigned long dim881JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   9   ,   11  ,   27  ,   43  ,   95  ,   149 ,   159 ,   495 ,   2711    ,   2189    ,0 };
            const unsigned long dim882JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   9   ,   11  ,   51  ,   43  ,   121 ,   299 ,   73  ,   1991    ,   561 ,   6781    ,0 };
            const unsigned long dim883JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   3   ,   7   ,   61  ,   125 ,   13  ,   415 ,   11  ,   15  ,   2463    ,   385 ,0 };
            const unsigned long dim884JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   1   ,   27  ,   23  ,   1   ,   47  ,   149 ,   371 ,   1599    ,   341 ,   1595    ,0 };
            const unsigned long dim885JoeKuoD6Init[]    =   {   1   ,   1   ,   1   ,   15  ,   3   ,   39  ,   65  ,   5   ,   33  ,   713 ,   399 ,   2935    ,   3825    ,0 };
            const unsigned long dim886JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   7   ,   17  ,   39  ,   3   ,   235 ,   227 ,   231 ,   323 ,   2869    ,   4747    ,0 };
            const unsigned long dim887JoeKuoD6Init[]    =   {   1   ,   1   ,   5   ,   9   ,   13  ,   11  ,   41  ,   155 ,   393 ,   701 ,   909 ,   2999    ,   5053    ,0 };
            const unsigned long dim888JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   7   ,   5   ,   41  ,   105 ,   13  ,   123 ,   429 ,   1455    ,   2501    ,   1065    ,0 };
            const unsigned long dim889JoeKuoD6Init[]    =   {   1   ,   1   ,   1   ,   7   ,   17  ,   21  ,   95  ,   141 ,   305 ,   951 ,   1305    ,   3639    ,   3797    ,0 };
            const unsigned long dim890JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   9   ,   7   ,   59  ,   29  ,   129 ,   265 ,   889 ,   993 ,   1411    ,   2069    ,0 };
            const unsigned long dim891JoeKuoD6Init[]    =   {   1   ,   1   ,   1   ,   11  ,   3   ,   51  ,   9   ,   35  ,   355 ,   131 ,   1307    ,   1547    ,   701 ,0 };
            const unsigned long dim892JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   3   ,   17  ,   57  ,   113 ,   197 ,   483 ,   831 ,   1741    ,   3897    ,   3649    ,0 };
            const unsigned long dim893JoeKuoD6Init[]    =   {   1   ,   1   ,   5   ,   11  ,   1   ,   33  ,   7   ,   243 ,   95  ,   905 ,   1933    ,   2123    ,   7381    ,0 };
            const unsigned long dim894JoeKuoD6Init[]    =   {   1   ,   3   ,   3   ,   15  ,   5   ,   37  ,   1   ,   235 ,   77  ,   375 ,   1231    ,   1029    ,   1839    ,0 };
            const unsigned long dim895JoeKuoD6Init[]    =   {   1   ,   1   ,   5   ,   3   ,   31  ,   11  ,   103 ,   245 ,   373 ,   103 ,   25  ,   1283    ,   6155    ,0 };
            const unsigned long dim896JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   1   ,   25  ,   25  ,   51  ,   159 ,   381 ,   97  ,   799 ,   565 ,   5165    ,0 };
            const unsigned long dim897JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   13  ,   25  ,   21  ,   109 ,   3   ,   253 ,   459 ,   623 ,   2817    ,   7815    ,0 };
            const unsigned long dim898JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   1   ,   23  ,   5   ,   127 ,   157 ,   187 ,   187 ,   1383    ,   2777    ,   2413    ,0 };
            const unsigned long dim899JoeKuoD6Init[]    =   {   1   ,   1   ,   5   ,   3   ,   29  ,   47  ,   97  ,   197 ,   215 ,   725 ,   445 ,   2111    ,   1701    ,0 };
            const unsigned long dim900JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   9   ,   27  ,   25  ,   105 ,   59  ,   333 ,   79  ,   1099    ,   3695    ,   5725    ,0 };
            const unsigned long dim901JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   3   ,   11  ,   15  ,   127 ,   151 ,   163 ,   967 ,   665 ,   733 ,   509 ,0 };
            const unsigned long dim902JoeKuoD6Init[]    =   {   1   ,   3   ,   3   ,   3   ,   15  ,   5   ,   51  ,   17  ,   409 ,   679 ,   229 ,   1037    ,   353 ,0 };
            const unsigned long dim903JoeKuoD6Init[]    =   {   1   ,   3   ,   7   ,   15  ,   7   ,   9   ,   17  ,   115 ,   205 ,   105 ,   597 ,   2913    ,   1941    ,0 };
            const unsigned long dim904JoeKuoD6Init[]    =   {   1   ,   1   ,   5   ,   11  ,   31  ,   1   ,   99  ,   213 ,   33  ,   319 ,   617 ,   1493    ,   7287    ,0 };
            const unsigned long dim905JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   5   ,   21  ,   31  ,   115 ,   43  ,   309 ,   217 ,   1   ,   3363    ,   6673    ,0 };
            const unsigned long dim906JoeKuoD6Init[]    =   {   1   ,   3   ,   7   ,   7   ,   11  ,   29  ,   125 ,   75  ,   209 ,   217 ,   1749    ,   1565    ,   4431    ,0 };
            const unsigned long dim907JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   1   ,   5   ,   37  ,   103 ,   237 ,   19  ,   763 ,   915 ,   753 ,   7479    ,0 };
            const unsigned long dim908JoeKuoD6Init[]    =   {   1   ,   3   ,   3   ,   9   ,   11  ,   59  ,   9   ,   133 ,   455 ,   683 ,   251 ,   963 ,   3743    ,0 };
            const unsigned long dim909JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   9   ,   7   ,   41  ,   97  ,   185 ,   117 ,   367 ,   2045    ,   3013    ,   6249    ,0 };
            const unsigned long dim910JoeKuoD6Init[]    =   {   1   ,   1   ,   1   ,   5   ,   29  ,   51  ,   53  ,   147 ,   379 ,   769 ,   865 ,   2273    ,   1755    ,0 };
            const unsigned long dim911JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   7   ,   13  ,   61  ,   125 ,   153 ,   297 ,   385 ,   1981    ,   2221    ,   4209    ,0 };
            const unsigned long dim912JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   15  ,   7   ,   33  ,   41  ,   43  ,   201 ,   819 ,   129 ,   757 ,   4087    ,0 };
            const unsigned long dim913JoeKuoD6Init[]    =   {   1   ,   1   ,   5   ,   7   ,   31  ,   15  ,   19  ,   15  ,   197 ,   921 ,   1967    ,   491 ,   4911    ,0 };
            const unsigned long dim914JoeKuoD6Init[]    =   {   1   ,   1   ,   1   ,   13  ,   21  ,   25  ,   119 ,   123 ,   255 ,   155 ,   1765    ,   2895    ,   4201    ,0 };
            const unsigned long dim915JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   9   ,   1   ,   63  ,   57  ,   193 ,   327 ,   455 ,   1729    ,   1259    ,   7483    ,0 };
            const unsigned long dim916JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   1   ,   5   ,   19  ,   71  ,   255 ,   257 ,   477 ,   1799    ,   3855    ,   955 ,0 };
            const unsigned long dim917JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   1   ,   29  ,   27  ,   7   ,   151 ,   327 ,   311 ,   235 ,   3599    ,   6165    ,0 };
            const unsigned long dim918JoeKuoD6Init[]    =   {   1   ,   3   ,   3   ,   3   ,   7   ,   29  ,   113 ,   87  ,   339 ,   271 ,   1141    ,   1111    ,   7971    ,0 };
            const unsigned long dim919JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   15  ,   23  ,   51  ,   57  ,   233 ,   179 ,   397 ,   1059    ,   123 ,   697 ,0 };
            const unsigned long dim920JoeKuoD6Init[]    =   {   1   ,   3   ,   3   ,   13  ,   17  ,   11  ,   15  ,   189 ,   437 ,   449 ,   1533    ,   2209    ,   1849    ,0 };
            const unsigned long dim921JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   9   ,   9   ,   19  ,   53  ,   165 ,   279 ,   875 ,   1223    ,   1585    ,   6647    ,0 };
            const unsigned long dim922JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   5   ,   5   ,   49  ,   119 ,   31  ,   273 ,   919 ,   413 ,   1531    ,   6233    ,0 };
            const unsigned long dim923JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   5   ,   1   ,   19  ,   27  ,   213 ,   413 ,   271 ,   223 ,   299 ,   5635    ,0 };
            const unsigned long dim924JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   3   ,   11  ,   55  ,   25  ,   43  ,   147 ,   53  ,   95  ,   1849    ,   7701    ,0 };
            const unsigned long dim925JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   11  ,   17  ,   35  ,   9   ,   3   ,   117 ,   729 ,   21  ,   3379    ,   5963    ,0 };
            const unsigned long dim926JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   7   ,   23  ,   25  ,   45  ,   103 ,   181 ,   677 ,   1155    ,   205 ,   1967    ,0 };
            const unsigned long dim927JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   9   ,   19  ,   53  ,   41  ,   243 ,   97  ,   199 ,   407 ,   3299    ,   5309    ,0 };
            const unsigned long dim928JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   11  ,   27  ,   27  ,   59  ,   153 ,   501 ,   855 ,   1235    ,   2295    ,   5535    ,0 };
            const unsigned long dim929JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   7   ,   11  ,   37  ,   1   ,   179 ,   289 ,   895 ,   511 ,   697 ,   5067    ,0 };
            const unsigned long dim930JoeKuoD6Init[]    =   {   1   ,   3   ,   7   ,   15  ,   25  ,   27  ,   33  ,   37  ,   351 ,   351 ,   813 ,   2239    ,   1561    ,0 };
            const unsigned long dim931JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   7   ,   25  ,   51  ,   79  ,   65  ,   427 ,   293 ,   1031    ,   685 ,   533 ,0 };
            const unsigned long dim932JoeKuoD6Init[]    =   {   1   ,   3   ,   3   ,   15  ,   21  ,   21  ,   77  ,   31  ,   299 ,   643 ,   161 ,   2991    ,   7911    ,0 };
            const unsigned long dim933JoeKuoD6Init[]    =   {   1   ,   3   ,   3   ,   7   ,   7   ,   61  ,   99  ,   19  ,   43  ,   35  ,   687 ,   1449    ,   5211    ,0 };
            const unsigned long dim934JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   5   ,   5   ,   59  ,   101 ,   175 ,   149 ,   799 ,   439 ,   19  ,   7565    ,0 };
            const unsigned long dim935JoeKuoD6Init[]    =   {   1   ,   3   ,   7   ,   9   ,   31  ,   13  ,   75  ,   227 ,   337 ,   207 ,   609 ,   2119    ,   3539    ,0 };
            const unsigned long dim936JoeKuoD6Init[]    =   {   1   ,   1   ,   1   ,   11  ,   3   ,   59  ,   1   ,   139 ,   425 ,   583 ,   263 ,   2687    ,   4445    ,0 };
            const unsigned long dim937JoeKuoD6Init[]    =   {   1   ,   1   ,   5   ,   7   ,   21  ,   31  ,   55  ,   117 ,   377 ,   469 ,   1907    ,   3091    ,   3221    ,0 };
            const unsigned long dim938JoeKuoD6Init[]    =   {   1   ,   1   ,   5   ,   1   ,   7   ,   63  ,   89  ,   197 ,   271 ,   425 ,   1493    ,   2337    ,   6397    ,0 };
            const unsigned long dim939JoeKuoD6Init[]    =   {   1   ,   1   ,   1   ,   7   ,   9   ,   35  ,   17  ,   237 ,   215 ,   211 ,   523 ,   3095    ,   1771    ,0 };
            const unsigned long dim940JoeKuoD6Init[]    =   {   1   ,   1   ,   5   ,   5   ,   19  ,   23  ,   119 ,   31  ,   331 ,   533 ,   71  ,   319 ,   7567    ,0 };
            const unsigned long dim941JoeKuoD6Init[]    =   {   1   ,   1   ,   5   ,   7   ,   1   ,   33  ,   41  ,   9   ,   421 ,   789 ,   1305    ,   3775    ,   6545    ,0 };
            const unsigned long dim942JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   11  ,   17  ,   9   ,   47  ,   237 ,   227 ,   859 ,   113 ,   429 ,   5399    ,0 };
            const unsigned long dim943JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   11  ,   25  ,   29  ,   15  ,   1   ,   73  ,   1015    ,   2023    ,   367 ,   1523    ,0 };
            const unsigned long dim944JoeKuoD6Init[]    =   {   1   ,   1   ,   5   ,   11  ,   19  ,   31  ,   121 ,   243 ,   411 ,   717 ,   671 ,   1329    ,   2669    ,0 };
            const unsigned long dim945JoeKuoD6Init[]    =   {   1   ,   1   ,   1   ,   11  ,   27  ,   41  ,   23  ,   13  ,   353 ,   533 ,   1627    ,   1625    ,   3705    ,0 };
            const unsigned long dim946JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   11  ,   31  ,   41  ,   17  ,   3   ,   163 ,   207 ,   407 ,   1475    ,   5911    ,0 };
            const unsigned long dim947JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   3   ,   3   ,   19  ,   87  ,   199 ,   77  ,   207 ,   1957    ,   1691    ,   5045    ,0 };
            const unsigned long dim948JoeKuoD6Init[]    =   {   1   ,   1   ,   1   ,   3   ,   19  ,   57  ,   107 ,   37  ,   419 ,   943 ,   133 ,   519 ,   4391    ,0 };
            const unsigned long dim949JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   11  ,   7   ,   5   ,   85  ,   1   ,   265 ,   801 ,   2025    ,   2655    ,   4635    ,0 };
            const unsigned long dim950JoeKuoD6Init[]    =   {   1   ,   3   ,   3   ,   3   ,   7   ,   13  ,   15  ,   155 ,   283 ,   97  ,   1311    ,   607 ,   2879    ,0 };
            const unsigned long dim951JoeKuoD6Init[]    =   {   1   ,   1   ,   5   ,   9   ,   21  ,   31  ,   3   ,   237 ,   323 ,   749 ,   1879    ,   611 ,   6763    ,0 };
            const unsigned long dim952JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   9   ,   19  ,   51  ,   25  ,   39  ,   7   ,   973 ,   837 ,   105 ,   1987    ,0 };
            const unsigned long dim953JoeKuoD6Init[]    =   {   1   ,   1   ,   5   ,   13  ,   9   ,   63  ,   39  ,   233 ,   391 ,   93  ,   1741    ,   425 ,   901 ,0 };
            const unsigned long dim954JoeKuoD6Init[]    =   {   1   ,   3   ,   7   ,   13  ,   13  ,   23  ,   95  ,   195 ,   263 ,   71  ,   255 ,   1839    ,   4301    ,0 };
            const unsigned long dim955JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   3   ,   29  ,   35  ,   123 ,   145 ,   377 ,   1021    ,   1361    ,   2319    ,   1487    ,0 };
            const unsigned long dim956JoeKuoD6Init[]    =   {   1   ,   3   ,   7   ,   15  ,   9   ,   27  ,   61  ,   85  ,   81  ,   371 ,   1803    ,   3683    ,   4771    ,0 };
            const unsigned long dim957JoeKuoD6Init[]    =   {   1   ,   3   ,   7   ,   15  ,   9   ,   35  ,   51  ,   137 ,   3   ,   721 ,   1269    ,   2667    ,   7293    ,0 };
            const unsigned long dim958JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   15  ,   3   ,   53  ,   121 ,   227 ,   55  ,   937 ,   1623    ,   2691    ,   2353    ,0 };
            const unsigned long dim959JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   5   ,   7   ,   53  ,   53  ,   121 ,   299 ,   633 ,   263 ,   3651    ,   3187    ,0 };
            const unsigned long dim960JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   9   ,   31  ,   59  ,   73  ,   5   ,   475 ,   763 ,   115 ,   2263    ,   2475    ,0 };
            const unsigned long dim961JoeKuoD6Init[]    =   {   1   ,   3   ,   3   ,   7   ,   13  ,   15  ,   49  ,   227 ,   285 ,   41  ,   1817    ,   2573    ,   4991    ,0 };
            const unsigned long dim962JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   3   ,   17  ,   7   ,   29  ,   93  ,   253 ,   527 ,   657 ,   1859    ,   763 ,0 };
            const unsigned long dim963JoeKuoD6Init[]    =   {   1   ,   1   ,   5   ,   11  ,   9   ,   11  ,   75  ,   47  ,   469 ,   119 ,   1505    ,   3807    ,   2769    ,0 };
            const unsigned long dim964JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   11  ,   1   ,   37  ,   65  ,   135 ,   345 ,   307 ,   305 ,   1167    ,   1767    ,0 };
            const unsigned long dim965JoeKuoD6Init[]    =   {   1   ,   3   ,   3   ,   13  ,   5   ,   33  ,   9   ,   245 ,   271 ,   123 ,   1865    ,   2215    ,   2791    ,0 };
            const unsigned long dim966JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   1   ,   9   ,   23  ,   3   ,   21  ,   149 ,   923 ,   1335    ,   1015    ,   45  ,0 };
            const unsigned long dim967JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   15  ,   11  ,   3   ,   65  ,   219 ,   233 ,   353 ,   771 ,   3961    ,   1515    ,0 };
            const unsigned long dim968JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   5   ,   15  ,   41  ,   87  ,   141 ,   19  ,   905 ,   449 ,   3727    ,   1875    ,0 };
            const unsigned long dim969JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   9   ,   11  ,   15  ,   53  ,   151 ,   243 ,   279 ,   1695    ,   2317    ,   6661    ,0 };
            const unsigned long dim970JoeKuoD6Init[]    =   {   1   ,   3   ,   3   ,   7   ,   19  ,   51  ,   31  ,   157 ,   79  ,   399 ,   465 ,   661 ,   1029    ,0 };
            const unsigned long dim971JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   1   ,   5   ,   41  ,   3   ,   33  ,   157 ,   385 ,   1943    ,   1851    ,   3121    ,0 };
            const unsigned long dim972JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   11  ,   25  ,   59  ,   105 ,   191 ,   151 ,   285 ,   801 ,   201 ,   6871    ,0 };
            const unsigned long dim973JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   15  ,   27  ,   35  ,   17  ,   195 ,   387 ,   907 ,   1905    ,   2177    ,   2597    ,0 };
            const unsigned long dim974JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   9   ,   31  ,   43  ,   71  ,   185 ,   469 ,   809 ,   541 ,   571 ,   107 ,0 };
            const unsigned long dim975JoeKuoD6Init[]    =   {   1   ,   3   ,   7   ,   13  ,   5   ,   15  ,   55  ,   211 ,   305 ,   645 ,   1317    ,   2751    ,   603 ,0 };
            const unsigned long dim976JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   15  ,   19  ,   59  ,   45  ,   147 ,   165 ,   223 ,   259 ,   1205    ,   431 ,0 };
            const unsigned long dim977JoeKuoD6Init[]    =   {   1   ,   1   ,   5   ,   13  ,   5   ,   53  ,   115 ,   57  ,   203 ,   151 ,   1039    ,   871 ,   3881    ,0 };
            const unsigned long dim978JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   15  ,   1   ,   29  ,   115 ,   249 ,   79  ,   957 ,   143 ,   3425    ,   5769    ,0 };
            const unsigned long dim979JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   15  ,   15  ,   15  ,   117 ,   181 ,   363 ,   17  ,   1871    ,   1049    ,   2997    ,0 };
            const unsigned long dim980JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   7   ,   23  ,   11  ,   77  ,   123 ,   149 ,   1019    ,   93  ,   3637    ,   6975    ,0 };
            const unsigned long dim981JoeKuoD6Init[]    =   {   1   ,   1   ,   1   ,   7   ,   5   ,   51  ,   77  ,   89  ,   121 ,   155 ,   1839    ,   1301    ,   2771    ,0 };
            const unsigned long dim982JoeKuoD6Init[]    =   {   1   ,   3   ,   7   ,   7   ,   7   ,   39  ,   33  ,   69  ,   169 ,   535 ,   1699    ,   2117    ,   3233    ,0 };
            const unsigned long dim983JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   7   ,   3   ,   59  ,   81  ,   113 ,   261 ,   627 ,   79  ,   227 ,   7623    ,0 };
            const unsigned long dim984JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   5   ,   13  ,   57  ,   121 ,   41  ,   11  ,   111 ,   1055    ,   3313    ,   3371    ,0 };
            const unsigned long dim985JoeKuoD6Init[]    =   {   1   ,   1   ,   3   ,   13  ,   13  ,   9   ,   47  ,   197 ,   303 ,   987 ,   1125    ,   2461    ,   5007    ,0 };
            const unsigned long dim986JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   5   ,   1   ,   59  ,   109 ,   49  ,   453 ,   673 ,   2017    ,   205 ,   4673    ,0 };
            const unsigned long dim987JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   13  ,   19  ,   43  ,   95  ,   247 ,   475 ,   193 ,   1513    ,   885 ,   5391    ,0 };
            const unsigned long dim988JoeKuoD6Init[]    =   {   1   ,   1   ,   1   ,   11  ,   29  ,   57  ,   77  ,   101 ,   433 ,   537 ,   1681    ,   2463    ,   6623    ,0 };
            const unsigned long dim989JoeKuoD6Init[]    =   {   1   ,   3   ,   7   ,   13  ,   3   ,   61  ,   29  ,   53  ,   29  ,   673 ,   1037    ,   3877    ,   7421    ,0 };
            const unsigned long dim990JoeKuoD6Init[]    =   {   1   ,   3   ,   3   ,   11  ,   5   ,   59  ,   61  ,   91  ,   13  ,   949 ,   1093    ,   1291    ,   7395    ,0 };
            const unsigned long dim991JoeKuoD6Init[]    =   {   1   ,   3   ,   3   ,   13  ,   21  ,   27  ,   49  ,   37  ,   429 ,   877 ,   1247    ,   2375    ,   737 ,0 };
            const unsigned long dim992JoeKuoD6Init[]    =   {   1   ,   3   ,   5   ,   3   ,   25  ,   19  ,   75  ,   187 ,   239 ,   971 ,   859 ,   2519    ,   7323    ,0 };
            const unsigned long dim993JoeKuoD6Init[]    =   {   1   ,   3   ,   3   ,   7   ,   25  ,   19  ,   11  ,   223 ,   75  ,   801 ,   313 ,   2875    ,   6857    ,0 };
            const unsigned long dim994JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   3   ,   13  ,   63  ,   49  ,   125 ,   65  ,   1005    ,   1757    ,   3685    ,   4385    ,0 };
            const unsigned long dim995JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   15  ,   5   ,   41  ,   95  ,   231 ,   505 ,   907 ,   751 ,   3207    ,   4115    ,0 };
            const unsigned long dim996JoeKuoD6Init[]    =   {   1   ,   1   ,   7   ,   11  ,   31  ,   7   ,   29  ,   51  ,   449 ,   151 ,   1821    ,   3317    ,   931 ,0 };
            const unsigned long dim997JoeKuoD6Init[]    =   {   1   ,   3   ,   1   ,   11  ,   19  ,   11  ,   19  ,   47  ,   111 ,   637 ,   37  ,   3707    ,   2599    ,0 };
            const unsigned long dim998JoeKuoD6Init[]    =   {   1   ,   3   ,   7   ,   5   ,   23  ,   27  ,   65  ,   217 ,   491 ,   129 ,   1715    ,   2479    ,   2015    ,0 };
            const unsigned long dim999JoeKuoD6Init[]    =   {   1   ,   1   ,   5   ,   3   ,   5   ,   51  ,   43  ,   253 ,   495 ,   105 ,   1183    ,   3409    ,   1469    ,0 };
            const unsigned long dim1000JoeKuoD6Init[]   =   {   1   ,   3   ,   7   ,   9   ,   21  ,   5   ,   87  ,   67  ,   481 ,   405 ,   1967    ,   2837    ,   5161    ,0 };
            const unsigned long dim1001JoeKuoD6Init[]   =   {   1   ,   1   ,   7   ,   3   ,   29  ,   25  ,   15  ,   211 ,   233 ,   301 ,   1037    ,   3853    ,   7115    ,0 };
            const unsigned long dim1002JoeKuoD6Init[]   =   {   1   ,   3   ,   5   ,   9   ,   3   ,   15  ,   85  ,   81  ,   141 ,   395 ,   845 ,   17  ,   7485    ,0 };
            const unsigned long dim1003JoeKuoD6Init[]   =   {   1   ,   3   ,   5   ,   3   ,   21  ,   37  ,   39  ,   93  ,   461 ,   615 ,   1725    ,   1745    ,   5137    ,0 };
            const unsigned long dim1004JoeKuoD6Init[]   =   {   1   ,   1   ,   3   ,   1   ,   1   ,   21  ,   21  ,   207 ,   271 ,   447 ,   1415    ,   2507    ,   3291    ,0 };
            const unsigned long dim1005JoeKuoD6Init[]   =   {   1   ,   3   ,   7   ,   11  ,   23  ,   3   ,   115 ,   191 ,   257 ,   1015    ,   463 ,   1501    ,   7439    ,0 };
            const unsigned long dim1006JoeKuoD6Init[]   =   {   1   ,   1   ,   7   ,   11  ,   3   ,   53  ,   101 ,   221 ,   9   ,   225 ,   1799    ,   1701    ,   5669    ,0 };
            const unsigned long dim1007JoeKuoD6Init[]   =   {   1   ,   1   ,   7   ,   13  ,   17  ,   27  ,   7   ,   115 ,   67  ,   645 ,   1283    ,   75  ,   1881    ,0 };
            const unsigned long dim1008JoeKuoD6Init[]   =   {   1   ,   1   ,   3   ,   15  ,   7   ,   39  ,   75  ,   85  ,   127 ,   379 ,   1509    ,   2215    ,   5697    ,0 };
            const unsigned long dim1009JoeKuoD6Init[]   =   {   1   ,   1   ,   7   ,   1   ,   19  ,   31  ,   11  ,   67  ,   263 ,   877 ,   337 ,   1655    ,   1855    ,0 };
            const unsigned long dim1010JoeKuoD6Init[]   =   {   1   ,   3   ,   5   ,   9   ,   11  ,   47  ,   125 ,   89  ,   235 ,   331 ,   159 ,   685 ,   999 ,0 };
            const unsigned long dim1011JoeKuoD6Init[]   =   {   1   ,   3   ,   7   ,   3   ,   23  ,   39  ,   59  ,   69  ,   221 ,   291 ,   725 ,   2831    ,   4743    ,0 };
            const unsigned long dim1012JoeKuoD6Init[]   =   {   1   ,   3   ,   7   ,   11  ,   1   ,   43  ,   127 ,   235 ,   143 ,   329 ,   1823    ,   21  ,   73  ,0 };
            const unsigned long dim1013JoeKuoD6Init[]   =   {   1   ,   3   ,   1   ,   3   ,   27  ,   23  ,   33  ,   29  ,   493 ,   775 ,   1873    ,   1005    ,   2869    ,0 };
            const unsigned long dim1014JoeKuoD6Init[]   =   {   1   ,   3   ,   3   ,   9   ,   3   ,   31  ,   63  ,   249 ,   353 ,   849 ,   5   ,   2787    ,   4759    ,0 };
            const unsigned long dim1015JoeKuoD6Init[]   =   {   1   ,   3   ,   3   ,   11  ,   15  ,   15  ,   73  ,   159 ,   433 ,   35  ,   1203    ,   2541    ,   6713    ,0 };
            const unsigned long dim1016JoeKuoD6Init[]   =   {   1   ,   1   ,   5   ,   1   ,   13  ,   17  ,   107 ,   101 ,   411 ,   861 ,   1075    ,   2161    ,   3923    ,0 };
            const unsigned long dim1017JoeKuoD6Init[]   =   {   1   ,   1   ,   3   ,   5   ,   29  ,   29  ,   55  ,   205 ,   235 ,   685 ,   1445    ,   2413    ,   6837    ,0 };
            const unsigned long dim1018JoeKuoD6Init[]   =   {   1   ,   1   ,   1   ,   13  ,   13  ,   1   ,   125 ,   37  ,   331 ,   33  ,   47  ,   2761    ,   3449    ,0 };
            const unsigned long dim1019JoeKuoD6Init[]   =   {   1   ,   1   ,   7   ,   3   ,   17  ,   21  ,   21  ,   93  ,   145 ,   423 ,   1057    ,   997 ,   6117    ,0 };
            const unsigned long dim1020JoeKuoD6Init[]   =   {   1   ,   3   ,   3   ,   13  ,   25  ,   47  ,   13  ,   159 ,   41  ,   907 ,   1543    ,   1919    ,   5547    ,0 };
            const unsigned long dim1021JoeKuoD6Init[]   =   {   1   ,   1   ,   3   ,   9   ,   29  ,   53  ,   83  ,   45  ,   133 ,   453 ,   183 ,   2939    ,   4945    ,0 };
            const unsigned long dim1022JoeKuoD6Init[]   =   {   1   ,   1   ,   1   ,   7   ,   9   ,   7   ,   89  ,   71  ,   315 ,   983 ,   1471    ,   2475    ,   7077    ,0 };
            const unsigned long dim1023JoeKuoD6Init[]   =   {   1   ,   1   ,   3   ,   3   ,   31  ,   21  ,   55  ,   37  ,   337 ,   63  ,   299 ,   2415    ,   153 ,0 };
            const unsigned long dim1024JoeKuoD6Init[]   =   {   1   ,   1   ,   7   ,   3   ,   27  ,   59  ,   85  ,   109 ,   405 ,   337 ,   393 ,   2291    ,   5423    ,0 };
            const unsigned long dim1025JoeKuoD6Init[]   =   {   1   ,   1   ,   1   ,   1   ,   5   ,   31  ,   33  ,   43  ,   141 ,   9   ,   1271    ,   1749    ,   6281    ,0 };
            const unsigned long dim1026JoeKuoD6Init[]   =   {   1   ,   1   ,   3   ,   9   ,   23  ,   35  ,   25  ,   17  ,   49  ,   119 ,   883 ,   2907    ,   3573    ,0 };
            const unsigned long dim1027JoeKuoD6Init[]   =   {   1   ,   3   ,   1   ,   15  ,   25  ,   61  ,   89  ,   15  ,   315 ,   1015    ,   563 ,   1895    ,   4123    ,0 };
            const unsigned long dim1028JoeKuoD6Init[]   =   {   1   ,   3   ,   3   ,   7   ,   15  ,   53  ,   61  ,   225 ,   381 ,   217 ,   163 ,   4005    ,   7269    ,0 };
            const unsigned long dim1029JoeKuoD6Init[]   =   {   1   ,   3   ,   7   ,   5   ,   1   ,   59  ,   13  ,   201 ,   53  ,   629 ,   1917    ,   3105    ,   3047    ,0 };
            const unsigned long dim1030JoeKuoD6Init[]   =   {   1   ,   3   ,   1   ,   1   ,   19  ,   45  ,   71  ,   11  ,   15  ,   365 ,   1223    ,   3861    ,   5005    ,0 };
            const unsigned long dim1031JoeKuoD6Init[]   =   {   1   ,   1   ,   5   ,   11  ,   15  ,   15  ,   31  ,   245 ,   105 ,   609 ,   1051    ,   1911    ,   7531    ,0 };
            const unsigned long dim1032JoeKuoD6Init[]   =   {   1   ,   3   ,   7   ,   11  ,   27  ,   5   ,   79  ,   31  ,   201 ,   347 ,   1295    ,   2589    ,   5179    ,0 };
            const unsigned long dim1033JoeKuoD6Init[]   =   {   1   ,   1   ,   3   ,   1   ,   31  ,   23  ,   65  ,   203 ,   359 ,   591 ,   1459    ,   1941    ,   2307    ,0 };
            const unsigned long dim1034JoeKuoD6Init[]   =   {   1   ,   1   ,   5   ,   5   ,   3   ,   55  ,   79  ,   179 ,   147 ,   223 ,   1011    ,   1437    ,   3715    ,0 };
            const unsigned long dim1035JoeKuoD6Init[]   =   {   1   ,   1   ,   5   ,   11  ,   23  ,   19  ,   57  ,   241 ,   363 ,   729 ,   1647    ,   2395    ,   3237    ,0 };
            const unsigned long dim1036JoeKuoD6Init[]   =   {   1   ,   3   ,   7   ,   7   ,   7   ,   63  ,   115 ,   53  ,   485 ,   965 ,   551 ,   3945    ,   1535    ,0 };
            const unsigned long dim1037JoeKuoD6Init[]   =   {   1   ,   1   ,   1   ,   15  ,   5   ,   61  ,   33  ,   213 ,   451 ,   363 ,   95  ,   1185    ,   3633    ,0 };
            const unsigned long dim1038JoeKuoD6Init[]   =   {   1   ,   3   ,   5   ,   1   ,   17  ,   49  ,   39  ,   159 ,   327 ,   1021    ,   289 ,   3375    ,   8055    ,0 };
            const unsigned long dim1039JoeKuoD6Init[]   =   {   1   ,   3   ,   3   ,   5   ,   11  ,   1   ,   73  ,   3   ,   121 ,   561 ,   209 ,   1913    ,   4549    ,0 };
            const unsigned long dim1040JoeKuoD6Init[]   =   {   1   ,   3   ,   1   ,   3   ,   23  ,   39  ,   101 ,   231 ,   317 ,   397 ,   693 ,   1579    ,   3585    ,0 };
            const unsigned long dim1041JoeKuoD6Init[]   =   {   1   ,   1   ,   1   ,   1   ,   1   ,   5   ,   81  ,   209 ,   373 ,   689 ,   259 ,   3167    ,   1665    ,0 };
            const unsigned long dim1042JoeKuoD6Init[]   =   {   1   ,   1   ,   1   ,   11  ,   21  ,   19  ,   83  ,   201 ,   7   ,   573 ,   1563    ,   1649    ,   5827    ,0 };
            const unsigned long dim1043JoeKuoD6Init[]   =   {   1   ,   3   ,   5   ,   1   ,   31  ,   35  ,   53  ,   149 ,   185 ,   925 ,   701 ,   3727    ,   4425    ,0 };
            const unsigned long dim1044JoeKuoD6Init[]   =   {   1   ,   1   ,   5   ,   7   ,   19  ,   11  ,   67  ,   47  ,   19  ,   789 ,   1855    ,   293 ,   7569    ,0 };
            const unsigned long dim1045JoeKuoD6Init[]   =   {   1   ,   1   ,   5   ,   13  ,   25  ,   55  ,   25  ,   187 ,   405 ,   1013    ,   493 ,   1645    ,   6753    ,0 };
            const unsigned long dim1046JoeKuoD6Init[]   =   {   1   ,   1   ,   3   ,   11  ,   13  ,   23  ,   35  ,   219 ,   247 ,   723 ,   1079    ,   1523    ,   7383    ,0 };
            const unsigned long dim1047JoeKuoD6Init[]   =   {   1   ,   1   ,   3   ,   13  ,   5   ,   1   ,   27  ,   151 ,   61  ,   889 ,   1911    ,   3859    ,   4661    ,0 };
            const unsigned long dim1048JoeKuoD6Init[]   =   {   1   ,   3   ,   1   ,   7   ,   15  ,   59  ,   57  ,   157 ,   333 ,   367 ,   245 ,   3043    ,   1683    ,0 };
            const unsigned long dim1049JoeKuoD6Init[]   =   {   1   ,   3   ,   7   ,   11  ,   15  ,   53  ,   107 ,   81  ,   287 ,   637 ,   557 ,   315 ,   4361    ,0 };
            const unsigned long dim1050JoeKuoD6Init[]   =   {   1   ,   1   ,   5   ,   11  ,   11  ,   21  ,   35  ,   59  ,   429 ,   459 ,   1439    ,   639 ,   6907    ,0 };
            const unsigned long dim1051JoeKuoD6Init[]   =   {   1   ,   3   ,   3   ,   11  ,   13  ,   51  ,   101 ,   121 ,   437 ,   9   ,   889 ,   1907    ,   5217    ,0 };
            const unsigned long dim1052JoeKuoD6Init[]   =   {   1   ,   1   ,   7   ,   13  ,   7   ,   5   ,   99  ,   161 ,   207 ,   237 ,   1547    ,   403 ,   2269    ,0 };
            const unsigned long dim1053JoeKuoD6Init[]   =   {   1   ,   1   ,   5   ,   9   ,   23  ,   17  ,   43  ,   113 ,   119 ,   213 ,   977 ,   1197    ,   1039    ,0 };
            const unsigned long dim1054JoeKuoD6Init[]   =   {   1   ,   3   ,   3   ,   9   ,   25  ,   51  ,   79  ,   5   ,   355 ,   279 ,   875 ,   1947    ,   1261    ,0 };
            const unsigned long dim1055JoeKuoD6Init[]   =   {   1   ,   1   ,   7   ,   3   ,   5   ,   15  ,   85  ,   137 ,   361 ,   509 ,   959 ,   3227    ,   3755    ,0 };
            const unsigned long dim1056JoeKuoD6Init[]   =   {   1   ,   1   ,   7   ,   3   ,   29  ,   31  ,   45  ,   139 ,   143 ,   891 ,   877 ,   2195    ,   4405    ,0 };
            const unsigned long dim1057JoeKuoD6Init[]   =   {   1   ,   1   ,   1   ,   7   ,   17  ,   7   ,   81  ,   19  ,   469 ,   713 ,   57  ,   3559    ,   4925    ,0 };
            const unsigned long dim1058JoeKuoD6Init[]   =   {   1   ,   1   ,   1   ,   9   ,   13  ,   43  ,   69  ,   77  ,   225 ,   497 ,   1263    ,   1153    ,   7375    ,0 };
            const unsigned long dim1059JoeKuoD6Init[]   =   {   1   ,   1   ,   5   ,   7   ,   15  ,   19  ,   67  ,   75  ,   207 ,   851 ,   1827    ,   3523    ,   2159    ,0 };
            const unsigned long dim1060JoeKuoD6Init[]   =   {   1   ,   3   ,   7   ,   3   ,   25  ,   63  ,   71  ,   157 ,   509 ,   219 ,   1529    ,   2513    ,   1697    ,0 };
            const unsigned long dim1061JoeKuoD6Init[]   =   {   1   ,   3   ,   1   ,   5   ,   9   ,   9   ,   13  ,   147 ,   363 ,   561 ,   1707    ,   2991    ,   1671    ,0 };
            const unsigned long dim1062JoeKuoD6Init[]   =   {   1   ,   3   ,   1   ,   15  ,   13  ,   55  ,   77  ,   27  ,   427 ,   277 ,   725 ,   2861    ,   4841    ,0 };
            const unsigned long dim1063JoeKuoD6Init[]   =   {   1   ,   1   ,   5   ,   15  ,   25  ,   55  ,   21  ,   135 ,   221 ,   965 ,   1723    ,   2097    ,   1277    ,0 };
            const unsigned long dim1064JoeKuoD6Init[]   =   {   1   ,   1   ,   5   ,   3   ,   1   ,   11  ,   45  ,   179 ,   503 ,   283 ,   1075    ,   1283    ,   7603    ,0 };
            const unsigned long dim1065JoeKuoD6Init[]   =   {   1   ,   3   ,   1   ,   1   ,   1   ,   39  ,   61  ,   107 ,   183 ,   817 ,   1577    ,   1143    ,   783 ,0 };
            const unsigned long dim1066JoeKuoD6Init[]   =   {   1   ,   1   ,   5   ,   13  ,   9   ,   59  ,   51  ,   89  ,   397 ,   705 ,   1081    ,   3053    ,   4001    ,0 };
            const unsigned long dim1067JoeKuoD6Init[]   =   {   1   ,   3   ,   3   ,   11  ,   25  ,   53  ,   3   ,   125 ,   419 ,   827 ,   1907    ,   903 ,   3169    ,0 };
            const unsigned long dim1068JoeKuoD6Init[]   =   {   1   ,   3   ,   5   ,   9   ,   1   ,   3   ,   119 ,   153 ,   353 ,   315 ,   1555    ,   1273    ,   6605    ,0 };
            const unsigned long dim1069JoeKuoD6Init[]   =   {   1   ,   1   ,   3   ,   13  ,   13  ,   39  ,   117 ,   59  ,   225 ,   735 ,   1297    ,   1281    ,   8049    ,0 };
            const unsigned long dim1070JoeKuoD6Init[]   =   {   1   ,   3   ,   1   ,   3   ,   27  ,   53  ,   57  ,   95  ,   303 ,   803 ,   343 ,   4021    ,   5047    ,0 };
            const unsigned long dim1071JoeKuoD6Init[]   =   {   1   ,   3   ,   1   ,   11  ,   25  ,   1   ,   91  ,   107 ,   181 ,   839 ,   1691    ,   3859    ,   2363    ,0 };
            const unsigned long dim1072JoeKuoD6Init[]   =   {   1   ,   1   ,   7   ,   1   ,   19  ,   45  ,   23  ,   37  ,   345 ,   675 ,   5   ,   3731    ,   2961    ,0 };
            const unsigned long dim1073JoeKuoD6Init[]   =   {   1   ,   3   ,   1   ,   11  ,   13  ,   47  ,   19  ,   43  ,   221 ,   1007    ,   1021    ,   2967    ,   6931    ,0 };
            const unsigned long dim1074JoeKuoD6Init[]   =   {   1   ,   3   ,   7   ,   5   ,   27  ,   21  ,   19  ,   57  ,   59  ,   485 ,   967 ,   983 ,   4989    ,0 };
            const unsigned long dim1075JoeKuoD6Init[]   =   {   1   ,   3   ,   7   ,   3   ,   11  ,   23  ,   1   ,   221 ,   223 ,   535 ,   155 ,   2805    ,   661 ,0 };
            const unsigned long dim1076JoeKuoD6Init[]   =   {   1   ,   1   ,   7   ,   5   ,   3   ,   49  ,   127 ,   75  ,   215 ,   517 ,   1391    ,   589 ,   493 ,0 };
            const unsigned long dim1077JoeKuoD6Init[]   =   {   1   ,   1   ,   7   ,   5   ,   23  ,   29  ,   79  ,   171 ,   319 ,   811 ,   191 ,   1131    ,   7689    ,0 };
            const unsigned long dim1078JoeKuoD6Init[]   =   {   1   ,   1   ,   1   ,   7   ,   21  ,   9   ,   23  ,   21  ,   171 ,   631 ,   299 ,   2295    ,   5691    ,0 };
            const unsigned long dim1079JoeKuoD6Init[]   =   {   1   ,   1   ,   3   ,   7   ,   17  ,   59  ,   47  ,   205 ,   39  ,   407 ,   1107    ,   1323    ,   6769    ,0 };
            const unsigned long dim1080JoeKuoD6Init[]   =   {   1   ,   3   ,   7   ,   1   ,   9   ,   33  ,   117 ,   249 ,   167 ,   631 ,   717 ,   3901    ,   6125    ,0 };
            const unsigned long dim1081JoeKuoD6Init[]   =   {   1   ,   3   ,   7   ,   9   ,   31  ,   11  ,   75  ,   125 ,   417 ,   803 ,   1117    ,   1609    ,   3263    ,0 };
            const unsigned long dim1082JoeKuoD6Init[]   =   {   1   ,   3   ,   1   ,   3   ,   15  ,   33  ,   35  ,   205 ,   477 ,   939 ,   1891    ,   2579    ,   2035    ,0 };
            const unsigned long dim1083JoeKuoD6Init[]   =   {   1   ,   3   ,   5   ,   1   ,   5   ,   21  ,   5   ,   221 ,   105 ,   103 ,   587 ,   2111    ,   5891    ,0 };
            const unsigned long dim1084JoeKuoD6Init[]   =   {   1   ,   3   ,   1   ,   11  ,   3   ,   45  ,   91  ,   113 ,   311 ,   617 ,   1629    ,   1677    ,   2349    ,0 };
            const unsigned long dim1085JoeKuoD6Init[]   =   {   1   ,   1   ,   7   ,   13  ,   27  ,   17  ,   69  ,   11  ,   75  ,   545 ,   1245    ,   785 ,   645 ,0 };
            const unsigned long dim1086JoeKuoD6Init[]   =   {   1   ,   3   ,   3   ,   5   ,   23  ,   13  ,   87  ,   193 ,   407 ,   445 ,   1299    ,   81  ,   2679    ,0 };
            const unsigned long dim1087JoeKuoD6Init[]   =   {   1   ,   1   ,   7   ,   15  ,   17  ,   43  ,   13  ,   233 ,   255 ,   51  ,   1277    ,   965 ,   1615    ,0 };
            const unsigned long dim1088JoeKuoD6Init[]   =   {   1   ,   3   ,   7   ,   1   ,   31  ,   49  ,   21  ,   61  ,   269 ,   265 ,   25  ,   3141    ,   7687    ,0 };
            const unsigned long dim1089JoeKuoD6Init[]   =   {   1   ,   3   ,   3   ,   13  ,   27  ,   39  ,   65  ,   69  ,   453 ,   203 ,   461 ,   1585    ,   5681    ,0 };
            const unsigned long dim1090JoeKuoD6Init[]   =   {   1   ,   1   ,   3   ,   3   ,   23  ,   19  ,   63  ,   97  ,   339 ,   83  ,   671 ,   433 ,   5285    ,0 };
            const unsigned long dim1091JoeKuoD6Init[]   =   {   1   ,   3   ,   7   ,   11  ,   9   ,   3   ,   119 ,   157 ,   385 ,   847 ,   1159    ,   1389    ,   5833    ,0 };
            const unsigned long dim1092JoeKuoD6Init[]   =   {   1   ,   1   ,   5   ,   5   ,   31  ,   51  ,   39  ,   247 ,   71  ,   815 ,   1807    ,   1973    ,   1841    ,0 };
            const unsigned long dim1093JoeKuoD6Init[]   =   {   1   ,   1   ,   3   ,   3   ,   31  ,   39  ,   121 ,   75  ,   343 ,   25  ,   59  ,   575 ,   3503    ,0 };
            const unsigned long dim1094JoeKuoD6Init[]   =   {   1   ,   1   ,   5   ,   11  ,   23  ,   17  ,   35  ,   33  ,   437 ,   1001    ,   1941    ,   351 ,   2949    ,0 };
            const unsigned long dim1095JoeKuoD6Init[]   =   {   1   ,   1   ,   3   ,   7   ,   9   ,   51  ,   125 ,   3   ,   279 ,   373 ,   157 ,   1537    ,   5547    ,0 };
            const unsigned long dim1096JoeKuoD6Init[]   =   {   1   ,   3   ,   1   ,   11  ,   11  ,   23  ,   63  ,   91  ,   139 ,   911 ,   463 ,   1831    ,   7325    ,0 };
            const unsigned long dim1097JoeKuoD6Init[]   =   {   1   ,   3   ,   7   ,   1   ,   9   ,   11  ,   33  ,   125 ,   471 ,   583 ,   881 ,   2533    ,   7173    ,0 };
            const unsigned long dim1098JoeKuoD6Init[]   =   {   1   ,   1   ,   5   ,   1   ,   15  ,   41  ,   53  ,   77  ,   43  ,   237 ,   787 ,   3299    ,   5131    ,0 };
            const unsigned long dim1099JoeKuoD6Init[]   =   {   1   ,   1   ,   7   ,   7   ,   17  ,   19  ,   115 ,   145 ,   231 ,   555 ,   1279    ,   605 ,   2977    ,0 };
            const unsigned long dim1100JoeKuoD6Init[]   =   {   1   ,   1   ,   3   ,   7   ,   21  ,   47  ,   61  ,   171 ,   337 ,   441 ,   1961    ,   2815    ,   4565    ,0 };
            const unsigned long dim1101JoeKuoD6Init[]   =   {   1   ,   1   ,   7   ,   9   ,   25  ,   25  ,   3   ,   199 ,   85  ,   129 ,   1853    ,   3293    ,   71  ,0 };
            const unsigned long dim1102JoeKuoD6Init[]   =   {   1   ,   1   ,   3   ,   1   ,   25  ,   25  ,   113 ,   9   ,   453 ,   353 ,   1955    ,   1443    ,   7027    ,0 };
            const unsigned long dim1103JoeKuoD6Init[]   =   {   1   ,   3   ,   3   ,   11  ,   9   ,   35  ,   89  ,   149 ,   335 ,   485 ,   37  ,   1519    ,   7341    ,0 };
            const unsigned long dim1104JoeKuoD6Init[]   =   {   1   ,   3   ,   7   ,   15  ,   27  ,   1   ,   119 ,   175 ,   181 ,   353 ,   1667    ,   2023    ,   3239    ,0 };
            const unsigned long dim1105JoeKuoD6Init[]   =   {   1   ,   1   ,   1   ,   5   ,   27  ,   11  ,   39  ,   255 ,   283 ,   101 ,   603 ,   541 ,   2429    ,0 };
            const unsigned long dim1106JoeKuoD6Init[]   =   {   1   ,   1   ,   7   ,   11  ,   27  ,   51  ,   87  ,   187 ,   183 ,   1013    ,   1207    ,   3431    ,   3321    ,0 };
            const unsigned long dim1107JoeKuoD6Init[]   =   {   1   ,   1   ,   5   ,   3   ,   5   ,   15  ,   107 ,   171 ,   139 ,   919 ,   235 ,   3047    ,   3401    ,0 };
            const unsigned long dim1108JoeKuoD6Init[]   =   {   1   ,   3   ,   3   ,   1   ,   17  ,   49  ,   35  ,   37  ,   59  ,   61  ,   227 ,   3175    ,   4093    ,0 };
            const unsigned long dim1109JoeKuoD6Init[]   =   {   1   ,   1   ,   3   ,   1   ,   23  ,   51  ,   43  ,   143 ,   389 ,   559 ,   349 ,   1933    ,   6149    ,0 };
            const unsigned long dim1110JoeKuoD6Init[]   =   {   1   ,   3   ,   5   ,   13  ,   9   ,   19  ,   111 ,   233 ,   373 ,   883 ,   1653    ,   1591    ,   2971    ,0 };
            const unsigned long dim1111JoeKuoD6Init[]   =   {   1   ,   3   ,   3   ,   13  ,   7   ,   63  ,   95  ,   29  ,   173 ,   175 ,   355 ,   877 ,   5819    ,   15873   ,0 };
            const unsigned long dim1112JoeKuoD6Init[]   =   {   1   ,   1   ,   1   ,   7   ,   1   ,   63  ,   23  ,   73  ,   469 ,   893 ,   773 ,   3647    ,   1191    ,   8731    ,0 };
            const unsigned long dim1113JoeKuoD6Init[]   =   {   1   ,   3   ,   3   ,   9   ,   17  ,   9   ,   85  ,   137 ,   53  ,   793 ,   439 ,   2097    ,   4257    ,   8439    ,0 };
            const unsigned long dim1114JoeKuoD6Init[]   =   {   1   ,   1   ,   7   ,   15  ,   3   ,   57  ,   83  ,   83  ,   123 ,   399 ,   1659    ,   2621    ,   3355    ,   255 ,0 };
            const unsigned long dim1115JoeKuoD6Init[]   =   {   1   ,   1   ,   5   ,   13  ,   9   ,   55  ,   21  ,   55  ,   391 ,   765 ,   2023    ,   1211    ,   1265    ,   5391    ,0 };
            const unsigned long dim1116JoeKuoD6Init[]   =   {   1   ,   1   ,   1   ,   11  ,   31  ,   63  ,   43  ,   153 ,   423 ,   969 ,   931 ,   531 ,   1635    ,   4695    ,0 };
            const unsigned long dim1117JoeKuoD6Init[]   =   {   1   ,   1   ,   5   ,   11  ,   19  ,   11  ,   115 ,   213 ,   473 ,   13  ,   853 ,   3771    ,   7841    ,   1249    ,0 };
            const unsigned long dim1118JoeKuoD6Init[]   =   {   1   ,   1   ,   1   ,   5   ,   31  ,   51  ,   121 ,   97  ,   483 ,   511 ,   299 ,   3307    ,   3363    ,   12639   ,0 };
            const unsigned long dim1119JoeKuoD6Init[]   =   {   1   ,   3   ,   3   ,   1   ,   31  ,   31  ,   89  ,   221 ,   451 ,   253 ,   1371    ,   2457    ,   573 ,   4359    ,0 };
            const unsigned long dim1120JoeKuoD6Init[]   =   {   1   ,   1   ,   7   ,   5   ,   5   ,   21  ,   59  ,   151 ,   275 ,   937 ,   1603    ,   3337    ,   3157    ,   15289   ,0 };
            const unsigned long dim1121JoeKuoD6Init[]   =   {   1   ,   3   ,   1   ,   3   ,   19  ,   11  ,   27  ,   5   ,   405 ,   899 ,   1315    ,   3299    ,   945 ,   1421    ,0 };
            const unsigned long dim1122JoeKuoD6Init[]   =   {   1   ,   3   ,   7   ,   3   ,   31  ,   55  ,   85  ,   49  ,   303 ,   859 ,   469 ,   723 ,   2221    ,   4099    ,0 };
            const unsigned long dim1123JoeKuoD6Init[]   =   {   1   ,   1   ,   5   ,   13  ,   31  ,   27  ,   57  ,   37  ,   493 ,   883 ,   1781    ,   2915    ,   6435    ,   6905    ,0 };
            const unsigned long dim1124JoeKuoD6Init[]   =   {   1   ,   1   ,   1   ,   7   ,   21  ,   49  ,   111 ,   69  ,   221 ,   845 ,   219 ,   1257    ,   7909    ,   15225   ,0 };
            const unsigned long dim1125JoeKuoD6Init[]   =   {   1   ,   3   ,   1   ,   3   ,   3   ,   1   ,   89  ,   143 ,   359 ,   717 ,   327 ,   1063    ,   3845    ,   8963    ,0 };
            const unsigned long dim1126JoeKuoD6Init[]   =   {   1   ,   1   ,   3   ,   7   ,   31  ,   3   ,   5   ,   79  ,   241 ,   877 ,   309 ,   3735    ,   1409    ,   11947   ,0 };
            const unsigned long dim1127JoeKuoD6Init[]   =   {   1   ,   1   ,   1   ,   7   ,   13  ,   63  ,   89  ,   69  ,   417 ,   809 ,   1251    ,   3213    ,   1107    ,   9143    ,0 };
            const unsigned long dim1128JoeKuoD6Init[]   =   {   1   ,   3   ,   7   ,   9   ,   21  ,   3   ,   17  ,   231 ,   273 ,   545 ,   1385    ,   1367    ,   5491    ,   7497    ,0 };
            const unsigned long dim1129JoeKuoD6Init[]   =   {   1   ,   1   ,   7   ,   1   ,   31  ,   35  ,   117 ,   31  ,   173 ,   711 ,   555 ,   1551    ,   443 ,   10469   ,0 };
            const unsigned long dim1130JoeKuoD6Init[]   =   {   1   ,   1   ,   5   ,   9   ,   3   ,   61  ,   55  ,   31  ,   321 ,   517 ,   47  ,   1379    ,   4329    ,   2305    ,0 };
            const unsigned long dim1131JoeKuoD6Init[]   =   {   1   ,   1   ,   5   ,   15  ,   15  ,   13  ,   23  ,   153 ,   209 ,   115 ,   913 ,   2013    ,   841 ,   2015    ,0 };
            const unsigned long dim1132JoeKuoD6Init[]   =   {   1   ,   3   ,   1   ,   5   ,   15  ,   43  ,   31  ,   103 ,   55  ,   489 ,   1003    ,   2723    ,   2613    ,   15743   ,0 };
            const unsigned long dim1133JoeKuoD6Init[]   =   {   1   ,   3   ,   5   ,   3   ,   31  ,   45  ,   21  ,   37  ,   453 ,   99  ,   1047    ,   235 ,   1133    ,   11199   ,0 };
            const unsigned long dim1134JoeKuoD6Init[]   =   {   1   ,   1   ,   1   ,   9   ,   13  ,   25  ,   109 ,   129 ,   25  ,   291 ,   445 ,   2775    ,   1011    ,   4645    ,0 };
            const unsigned long dim1135JoeKuoD6Init[]   =   {   1   ,   1   ,   3   ,   9   ,   9   ,   45  ,   7   ,   161 ,   217 ,   645 ,   421 ,   3429    ,   2941    ,   5319    ,0 };
            const unsigned long dim1136JoeKuoD6Init[]   =   {   1   ,   1   ,   3   ,   15  ,   13  ,   57  ,   13  ,   63  ,   89  ,   739 ,   1989    ,   1679    ,   2031    ,   265 ,0 };
            const unsigned long dim1137JoeKuoD6Init[]   =   {   1   ,   3   ,   7   ,   9   ,   21  ,   63  ,   117 ,   21  ,   345 ,   19  ,   1357    ,   611 ,   6591    ,   8075    ,0 };
            const unsigned long dim1138JoeKuoD6Init[]   =   {   1   ,   1   ,   7   ,   13  ,   29  ,   37  ,   37  ,   219 ,   81  ,   143 ,   1505    ,   2921    ,   1497    ,   4359    ,0 };
            const unsigned long dim1139JoeKuoD6Init[]   =   {   1   ,   3   ,   5   ,   9   ,   25  ,   23  ,   39  ,   145 ,   121 ,   877 ,   1491    ,   2687    ,   865 ,   3417    ,0 };
            const unsigned long dim1140JoeKuoD6Init[]   =   {   1   ,   1   ,   1   ,   13  ,   25  ,   7   ,   45  ,   143 ,   61  ,   591 ,   455 ,   3039    ,   4147    ,   1695    ,0 };
            const unsigned long dim1141JoeKuoD6Init[]   =   {   1   ,   1   ,   1   ,   9   ,   27  ,   7   ,   61  ,   139 ,   509 ,   843 ,   1175    ,   3787    ,   4695    ,   13851   ,0 };
            const unsigned long dim1142JoeKuoD6Init[]   =   {   1   ,   3   ,   1   ,   5   ,   15  ,   19  ,   73  ,   109 ,   263 ,   791 ,   455 ,   3503    ,   1203    ,   4993    ,0 };
            const unsigned long dim1143JoeKuoD6Init[]   =   {   1   ,   1   ,   1   ,   11  ,   21  ,   59  ,   51  ,   101 ,   407 ,   25  ,   355 ,   2817    ,   3323    ,   14641   ,0 };
            const unsigned long dim1144JoeKuoD6Init[]   =   {   1   ,   3   ,   1   ,   13  ,   3   ,   45  ,   73  ,   241 ,   511 ,   763 ,   189 ,   2913    ,   489 ,   10441   ,0 };
            const unsigned long dim1145JoeKuoD6Init[]   =   {   1   ,   1   ,   7   ,   9   ,   15  ,   9   ,   113 ,   103 ,   375 ,   773 ,   621 ,   3035    ,   4359    ,   13259   ,0 };
            const unsigned long dim1146JoeKuoD6Init[]   =   {   1   ,   3   ,   1   ,   1   ,   17  ,   13  ,   81  ,   207 ,   493 ,   737 ,   427 ,   2895    ,   539 ,   4125    ,0 };
            const unsigned long dim1147JoeKuoD6Init[]   =   {   1   ,   1   ,   1   ,   11  ,   9   ,   5   ,   89  ,   209 ,   175 ,   105 ,   293 ,   3337    ,   7235    ,   6481    ,0 };
            const unsigned long dim1148JoeKuoD6Init[]   =   {   1   ,   3   ,   1   ,   13  ,   5   ,   19  ,   3   ,   245 ,   25  ,   733 ,   885 ,   3747    ,   3997    ,   7945    ,0 };
            const unsigned long dim1149JoeKuoD6Init[]   =   {   1   ,   3   ,   3   ,   11  ,   13  ,   7   ,   29  ,   169 ,   239 ,   205 ,   1399    ,   1447    ,   5211    ,   11789   ,0 };
            const unsigned long dim1150JoeKuoD6Init[]   =   {   1   ,   3   ,   1   ,   15  ,   7   ,   19  ,   65  ,   23  ,   183 ,   547 ,   671 ,   2873    ,   5101    ,   4187    ,0 };
            const unsigned long dim1151JoeKuoD6Init[]   =   {   1   ,   3   ,   5   ,   13  ,   15  ,   35  ,   119 ,   75  ,   53  ,   773 ,   1769    ,   1945    ,   7781    ,   11349   ,0 };
            const unsigned long dim1152JoeKuoD6Init[]   =   {   1   ,   3   ,   7   ,   9   ,   31  ,   37  ,   39  ,   55  ,   333 ,   567 ,   1095    ,   3707    ,   4265    ,   117 ,0 };
            const unsigned long dim1153JoeKuoD6Init[]   =   {   1   ,   3   ,   1   ,   3   ,   1   ,   61  ,   9   ,   7   ,   253 ,   737 ,   1441    ,   2591    ,   2807    ,   13685   ,0 };
            const unsigned long dim1154JoeKuoD6Init[]   =   {   1   ,   1   ,   3   ,   11  ,   21  ,   41  ,   81  ,   61  ,   497 ,   827 ,   783 ,   3799    ,   4803    ,   141 ,0 };
            const unsigned long dim1155JoeKuoD6Init[]   =   {   1   ,   3   ,   7   ,   13  ,   11  ,   15  ,   3   ,   165 ,   401 ,   747 ,   1961    ,   3265    ,   8123    ,   2151    ,0 };
            const unsigned long dim1156JoeKuoD6Init[]   =   {   1   ,   3   ,   3   ,   15  ,   21  ,   39  ,   47  ,   135 ,   61  ,   803 ,   1447    ,   1763    ,   2033    ,   12907   ,0 };
            const unsigned long dim1157JoeKuoD6Init[]   =   {   1   ,   1   ,   1   ,   9   ,   25  ,   21  ,   15  ,   233 ,   493 ,   115 ,   921 ,   2047    ,   3715    ,   9089    ,0 };
            const unsigned long dim1158JoeKuoD6Init[]   =   {   1   ,   3   ,   5   ,   5   ,   11  ,   17  ,   43  ,   49  ,   249 ,   873 ,   771 ,   47  ,   5357    ,   12459   ,0 };
            const unsigned long dim1159JoeKuoD6Init[]   =   {   1   ,   3   ,   3   ,   13  ,   31  ,   39  ,   85  ,   97  ,   241 ,   529 ,   65  ,   1995    ,   2411    ,   761 ,0 };
            const unsigned long dim1160JoeKuoD6Init[]   =   {   1   ,   3   ,   5   ,   15  ,   3   ,   1   ,   63  ,   229 ,   11  ,   733 ,   901 ,   63  ,   3585    ,   10273   ,0 };
            const unsigned long dim1161JoeKuoD6Init[]   =   {   1   ,   1   ,   1   ,   13  ,   23  ,   39  ,   123 ,   187 ,   27  ,   893 ,   1127    ,   2781    ,   5977    ,   13797   ,0 };
            const unsigned long dim1162JoeKuoD6Init[]   =   {   1   ,   1   ,   5   ,   13  ,   15  ,   35  ,   97  ,   29  ,   345 ,   447 ,   897 ,   793 ,   5155    ,   15939   ,0 };
            const unsigned long dim1163JoeKuoD6Init[]   =   {   1   ,   1   ,   3   ,   3   ,   1   ,   15  ,   19  ,   189 ,   183 ,   137 ,   1059    ,   2135    ,   6891    ,   10319   ,0 };
            const unsigned long dim1164JoeKuoD6Init[]   =   {   1   ,   1   ,   1   ,   11  ,   29  ,   27  ,   83  ,   111 ,   113 ,   225 ,   1527    ,   1071    ,   635 ,   3471    ,0 };
            const unsigned long dim1165JoeKuoD6Init[]   =   {   1   ,   3   ,   1   ,   3   ,   27  ,   9   ,   43  ,   19  ,   409 ,   579 ,   1983    ,   811 ,   5477    ,   559 ,0 };
            const unsigned long dim1166JoeKuoD6Init[]   =   {   1   ,   3   ,   7   ,   9   ,   1   ,   39  ,   121 ,   249 ,   355 ,   677 ,   899 ,   729 ,   2611    ,   9463    ,0 };
            const unsigned long dim1167JoeKuoD6Init[]   =   {   1   ,   1   ,   1   ,   11  ,   31  ,   43  ,   47  ,   157 ,   309 ,   529 ,   787 ,   1935    ,   5679    ,   1553    ,0 };
            const unsigned long dim1168JoeKuoD6Init[]   =   {   1   ,   1   ,   5   ,   15  ,   19  ,   53  ,   63  ,   71  ,   371 ,   455 ,   1943    ,   477 ,   6227    ,   1033    ,0 };
            const unsigned long dim1169JoeKuoD6Init[]   =   {   1   ,   1   ,   3   ,   15  ,   9   ,   55  ,   45  ,   247 ,   189 ,   57  ,   1573    ,   591 ,   2831    ,   14273   ,0 };
            const unsigned long dim1170JoeKuoD6Init[]   =   {   1   ,   3   ,   5   ,   7   ,   3   ,   37  ,   33  ,   219 ,   13  ,   163 ,   763 ,   3921    ,   7617    ,   15439   ,0 };
            const unsigned long dim1171JoeKuoD6Init[]   =   {   1   ,   3   ,   1   ,   13  ,   21  ,   41  ,   85  ,   151 ,   117 ,   587 ,   1513    ,   17  ,   7695    ,   4135    ,0 };
            const unsigned long dim1172JoeKuoD6Init[]   =   {   1   ,   3   ,   1   ,   5   ,   21  ,   51  ,   107 ,   209 ,   107 ,   459 ,   243 ,   3043    ,   473 ,   1357    ,0 };
            const unsigned long dim1173JoeKuoD6Init[]   =   {   1   ,   3   ,   1   ,   1   ,   5   ,   49  ,   87  ,   203 ,   261 ,   935 ,   765 ,   2879    ,   7677    ,   607 ,0 };
            const unsigned long dim1174JoeKuoD6Init[]   =   {   1   ,   3   ,   7   ,   15  ,   25  ,   31  ,   7   ,   53  ,   391 ,   569 ,   187 ,   2761    ,   1369    ,   8871    ,0 };
            const unsigned long dim1175JoeKuoD6Init[]   =   {   1   ,   1   ,   1   ,   15  ,   9   ,   25  ,   31  ,   59  ,   485 ,   253 ,   381 ,   1239    ,   4797    ,   4691    ,0 };
            const unsigned long dim1176JoeKuoD6Init[]   =   {   1   ,   3   ,   5   ,   9   ,   13  ,   17  ,   35  ,   97  ,   339 ,   47  ,   1053    ,   293 ,   8103    ,   8489    ,0 };
            const unsigned long dim1177JoeKuoD6Init[]   =   {   1   ,   1   ,   1   ,   11  ,   13  ,   7   ,   43  ,   33  ,   187 ,   135 ,   277 ,   1011    ,   297 ,   8443    ,0 };
            const unsigned long dim1178JoeKuoD6Init[]   =   {   1   ,   3   ,   5   ,   1   ,   13  ,   45  ,   125 ,   185 ,   369 ,   643 ,   791 ,   119 ,   7205    ,   6785    ,0 };
            const unsigned long dim1179JoeKuoD6Init[]   =   {   1   ,   1   ,   3   ,   15  ,   15  ,   39  ,   89  ,   33  ,   47  ,   801 ,   1451    ,   2583    ,   445 ,   14029   ,0 };
            const unsigned long dim1180JoeKuoD6Init[]   =   {   1   ,   3   ,   1   ,   13  ,   27  ,   53  ,   113 ,   159 ,   395 ,   835 ,   655 ,   3503    ,   3733    ,   1547    ,0 };
            const unsigned long dim1181JoeKuoD6Init[]   =   {   1   ,   1   ,   7   ,   5   ,   29  ,   61  ,   43  ,   181 ,   127 ,   839 ,   755 ,   579 ,   6805    ,   4591    ,0 };
            const unsigned long dim1182JoeKuoD6Init[]   =   {   1   ,   3   ,   1   ,   1   ,   29  ,   63  ,   71  ,   145 ,   149 ,   571 ,   1129    ,   2503    ,   5761    ,   6251    ,0 };
            const unsigned long dim1183JoeKuoD6Init[]   =   {   1   ,   3   ,   5   ,   7   ,   7   ,   21  ,   73  ,   101 ,   29  ,   517 ,   235 ,   195 ,   3561    ,   15347   ,0 };
            const unsigned long dim1184JoeKuoD6Init[]   =   {   1   ,   3   ,   5   ,   9   ,   11  ,   11  ,   109 ,   81  ,   483 ,   183 ,   753 ,   1635    ,   3763    ,   8141    ,0 };
            const unsigned long dim1185JoeKuoD6Init[]   =   {   1   ,   3   ,   5   ,   1   ,   21  ,   35  ,   111 ,   251 ,   255 ,   605 ,   1613    ,   283 ,   3035    ,   11681   ,0 };
            const unsigned long dim1186JoeKuoD6Init[]   =   {   1   ,   3   ,   7   ,   3   ,   29  ,   49  ,   43  ,   255 ,   217 ,   791 ,   1281    ,   2327    ,   5115    ,   10261   ,0 };
            const unsigned long dim1187JoeKuoD6Init[]   =   {   1   ,   3   ,   7   ,   9   ,   19  ,   23  ,   93  ,   105 ,   69  ,   429 ,   1729    ,   3547    ,   217 ,   15075   ,0 };
            const unsigned long dim1188JoeKuoD6Init[]   =   {   1   ,   3   ,   1   ,   9   ,   7   ,   21  ,   13  ,   81  ,   395 ,   997 ,   399 ,   3925    ,   7277    ,   1733    ,0 };
            const unsigned long dim1189JoeKuoD6Init[]   =   {   1   ,   3   ,   7   ,   13  ,   11  ,   51  ,   65  ,   29  ,   489 ,   489 ,   1639    ,   3949    ,   6271    ,   2015    ,0 };
            const unsigned long dim1190JoeKuoD6Init[]   =   {   1   ,   1   ,   7   ,   1   ,   15  ,   39  ,   23  ,   67  ,   121 ,   617 ,   1689    ,   1887    ,   3485    ,   7913    ,0 };
            const unsigned long dim1191JoeKuoD6Init[]   =   {   1   ,   3   ,   7   ,   3   ,   25  ,   23  ,   81  ,   59  ,   87  ,   457 ,   937 ,   3229    ,   2731    ,   5003    ,0 };
            const unsigned long dim1192JoeKuoD6Init[]   =   {   1   ,   1   ,   5   ,   11  ,   29  ,   41  ,   21  ,   67  ,   475 ,   127 ,   1113    ,   1693    ,   3705    ,   1051    ,0 };
            const unsigned long dim1193JoeKuoD6Init[]   =   {   1   ,   3   ,   7   ,   3   ,   13  ,   7   ,   33  ,   95  ,   123 ,   65  ,   2039    ,   1011    ,   5175    ,   10459   ,0 };
            const unsigned long dim1194JoeKuoD6Init[]   =   {   1   ,   1   ,   7   ,   9   ,   29  ,   53  ,   113 ,   57  ,   323 ,   309 ,   501 ,   3971    ,   5999    ,   16335   ,0 };
            const unsigned long dim1195JoeKuoD6Init[]   =   {   1   ,   1   ,   1   ,   13  ,   29  ,   9   ,   17  ,   49  ,   173 ,   473 ,   555 ,   3753    ,   7471    ,   10099   ,0 };
            const unsigned long dim1196JoeKuoD6Init[]   =   {   1   ,   3   ,   3   ,   3   ,   23  ,   3   ,   101 ,   199 ,   351 ,   435 ,   1919    ,   1175    ,   8037    ,   2259    ,0 };
            const unsigned long dim1197JoeKuoD6Init[]   =   {   1   ,   3   ,   1   ,   7   ,   29  ,   27  ,   23  ,   61  ,   197 ,   883 ,   153 ,   683 ,   2109    ,   14991   ,0 };
            const unsigned long dim1198JoeKuoD6Init[]   =   {   1   ,   1   ,   5   ,   11  ,   15  ,   21  ,   23  ,   9   ,   341 ,   391 ,   1537    ,   3483    ,   4641    ,   12489   ,0 };
            const unsigned long dim1199JoeKuoD6Init[]   =   {   1   ,   3   ,   7   ,   1   ,   1   ,   9   ,   3   ,   7   ,   485 ,   377 ,   597 ,   2277    ,   7811    ,   1303    ,0 };
            const unsigned long dim1200JoeKuoD6Init[]   =   {   1   ,   3   ,   3   ,   9   ,   13  ,   23  ,   13  ,   1   ,   353 ,   517 ,   1787    ,   2095    ,   3095    ,   4999    ,0 };
            const unsigned long dim1201JoeKuoD6Init[]   =   {   1   ,   3   ,   5   ,   13  ,   17  ,   13  ,   55  ,   165 ,   487 ,   387 ,   839 ,   2867    ,   3939    ,   7945    ,0 };
            const unsigned long dim1202JoeKuoD6Init[]   =   {   1   ,   3   ,   5   ,   3   ,   15  ,   25  ,   103 ,   225 ,   275 ,   571 ,   441 ,   567 ,   4759    ,   12965   ,0 };
            const unsigned long dim1203JoeKuoD6Init[]   =   {   1   ,   1   ,   3   ,   5   ,   19  ,   31  ,   121 ,   109 ,   89  ,   315 ,   1517    ,   187 ,   1645    ,   2331    ,0 };
            const unsigned long dim1204JoeKuoD6Init[]   =   {   1   ,   1   ,   7   ,   3   ,   17  ,   49  ,   87  ,   241 ,   491 ,   99  ,   1559    ,   843 ,   5627    ,   2383    ,0 };
            const unsigned long dim1205JoeKuoD6Init[]   =   {   1   ,   3   ,   7   ,   7   ,   21  ,   55  ,   69  ,   27  ,   461 ,   663 ,   1609    ,   4059    ,   4097    ,   2299    ,0 };
            const unsigned long dim1206JoeKuoD6Init[]   =   {   1   ,   1   ,   1   ,   13  ,   1   ,   27  ,   35  ,   63  ,   125 ,   293 ,   721 ,   1433    ,   4813    ,   15439   ,0 };
            const unsigned long dim1207JoeKuoD6Init[]   =   {   1   ,   1   ,   1   ,   15  ,   19  ,   13  ,   65  ,   55  ,   101 ,   981 ,   989 ,   3501    ,   8125    ,   8849    ,0 };
            const unsigned long dim1208JoeKuoD6Init[]   =   {   1   ,   1   ,   5   ,   13  ,   25  ,   45  ,   23  ,   157 ,   165 ,   581 ,   1613    ,   3721    ,   7569    ,   9347    ,0 };
            const unsigned long dim1209JoeKuoD6Init[]   =   {   1   ,   1   ,   1   ,   5   ,   29  ,   47  ,   51  ,   43  ,   221 ,   809 ,   469 ,   2089    ,   2721    ,   3799    ,0 };
            const unsigned long dim1210JoeKuoD6Init[]   =   {   1   ,   3   ,   7   ,   9   ,   11  ,   63  ,   107 ,   33  ,   433 ,   381 ,   819 ,   739 ,   7861    ,   12219   ,0 };
            const unsigned long dim1211JoeKuoD6Init[]   =   {   1   ,   3   ,   5   ,   7   ,   29  ,   27  ,   105 ,   219 ,   227 ,   851 ,   1287    ,   1825    ,   1861    ,   8171    ,0 };
            const unsigned long dim1212JoeKuoD6Init[]   =   {   1   ,   3   ,   1   ,   11  ,   1   ,   31  ,   77  ,   183 ,   81  ,   781 ,   541 ,   1663    ,   3   ,   12533   ,0 };
            const unsigned long dim1213JoeKuoD6Init[]   =   {   1   ,   3   ,   3   ,   1   ,   7   ,   59  ,   107 ,   49  ,   147 ,   625 ,   593 ,   729 ,   4335    ,   12651   ,0 };
            const unsigned long dim1214JoeKuoD6Init[]   =   {   1   ,   3   ,   1   ,   11  ,   25  ,   45  ,   55  ,   117 ,   247 ,   499 ,   1403    ,   3321    ,   6769    ,   15183   ,0 };
            const unsigned long dim1215JoeKuoD6Init[]   =   {   1   ,   1   ,   3   ,   13  ,   15  ,   51  ,   41  ,   111 ,   297 ,   283 ,   1943    ,   2741    ,   5087    ,   12203   ,0 };
            const unsigned long dim1216JoeKuoD6Init[]   =   {   1   ,   3   ,   1   ,   13  ,   19  ,   37  ,   59  ,   149 ,   481 ,   211 ,   69  ,   1253    ,   8019    ,   10607   ,0 };
            const unsigned long dim1217JoeKuoD6Init[]   =   {   1   ,   3   ,   1   ,   1   ,   21  ,   61  ,   103 ,   93  ,   75  ,   91  ,   1695    ,   3683    ,   679 ,   14789   ,0 };
            const unsigned long dim1218JoeKuoD6Init[]   =   {   1   ,   3   ,   1   ,   3   ,   27  ,   19  ,   101 ,   215 ,   165 ,   887 ,   1647    ,   4033    ,   7709    ,   15581   ,0 };
            const unsigned long dim1219JoeKuoD6Init[]   =   {   1   ,   1   ,   1   ,   9   ,   17  ,   9   ,   15  ,   231 ,   335 ,   227 ,   661 ,   2505    ,   6525    ,   15019   ,0 };
            const unsigned long dim1220JoeKuoD6Init[]   =   {   1   ,   3   ,   3   ,   13  ,   15  ,   63  ,   81  ,   221 ,   117 ,   893 ,   875 ,   485 ,   7323    ,   11105   ,0 };
            const unsigned long dim1221JoeKuoD6Init[]   =   {   1   ,   3   ,   1   ,   3   ,   7   ,   57  ,   115 ,   159 ,   457 ,   235 ,   1245    ,   973 ,   7801    ,   4753    ,0 };
            const unsigned long dim1222JoeKuoD6Init[]   =   {   1   ,   3   ,   5   ,   7   ,   7   ,   21  ,   3   ,   49  ,   455 ,   733 ,   323 ,   261 ,   7809    ,   4699    ,0 };
            const unsigned long dim1223JoeKuoD6Init[]   =   {   1   ,   1   ,   7   ,   3   ,   17  ,   31  ,   15  ,   105 ,   81  ,   901 ,   1107    ,   2161    ,   6631    ,   14523   ,0 };
            const unsigned long dim1224JoeKuoD6Init[]   =   {   1   ,   1   ,   7   ,   15  ,   21  ,   51  ,   125 ,   227 ,   321 ,   665 ,   13  ,   3203    ,   6117    ,   15401   ,0 };
            const unsigned long dim1225JoeKuoD6Init[]   =   {   1   ,   1   ,   5   ,   11  ,   23  ,   23  ,   27  ,   31  ,   489 ,   625 ,   1973    ,   2005    ,   7097    ,   11557   ,0 };
            const unsigned long dim1226JoeKuoD6Init[]   =   {   1   ,   1   ,   7   ,   11  ,   27  ,   37  ,   41  ,   79  ,   373 ,   207 ,   1399    ,   2629    ,   953 ,   8649    ,0 };
            const unsigned long dim1227JoeKuoD6Init[]   =   {   1   ,   1   ,   5   ,   5   ,   27  ,   15  ,   73  ,   37  ,   277 ,   301 ,   207 ,   2283    ,   2787    ,   8475    ,0 };
            const unsigned long dim1228JoeKuoD6Init[]   =   {   1   ,   3   ,   5   ,   11  ,   7   ,   47  ,   3   ,   169 ,   201 ,   573 ,   1953    ,   915 ,   4509    ,   7767    ,0 };
            const unsigned long dim1229JoeKuoD6Init[]   =   {   1   ,   3   ,   3   ,   7   ,   21  ,   21  ,   97  ,   69  ,   501 ,   491 ,   1471    ,   657 ,   5661    ,   2315    ,0 };
            const unsigned long dim1230JoeKuoD6Init[]   =   {   1   ,   1   ,   1   ,   1   ,   21  ,   17  ,   127 ,   207 ,   255 ,   357 ,   59  ,   1035    ,   4497    ,   7397    ,0 };
            const unsigned long dim1231JoeKuoD6Init[]   =   {   1   ,   3   ,   7   ,   1   ,   17  ,   55  ,   71  ,   35  ,   85  ,   5   ,   507 ,   1445    ,   2201    ,   14137   ,0 };
            const unsigned long dim1232JoeKuoD6Init[]   =   {   1   ,   1   ,   1   ,   9   ,   23  ,   11  ,   21  ,   67  ,   347 ,   845 ,   443 ,   2461    ,   2767    ,   1603    ,0 };
            const unsigned long dim1233JoeKuoD6Init[]   =   {   1   ,   3   ,   3   ,   11  ,   25  ,   11  ,   97  ,   225 ,   301 ,   111 ,   1163    ,   1783    ,   7879    ,   15191   ,0 };
            const unsigned long dim1234JoeKuoD6Init[]   =   {   1   ,   3   ,   5   ,   11  ,   29  ,   55  ,   9   ,   137 ,   233 ,   51  ,   1957    ,   3807    ,   2631    ,   2255    ,0 };
            const unsigned long dim1235JoeKuoD6Init[]   =   {   1   ,   1   ,   7   ,   1   ,   19  ,   41  ,   83  ,   121 ,   283 ,   839 ,   179 ,   3307    ,   4679    ,   1163    ,0 };
            const unsigned long dim1236JoeKuoD6Init[]   =   {   1   ,   1   ,   3   ,   9   ,   17  ,   1   ,   11  ,   47  ,   295 ,   189 ,   1863    ,   1145    ,   367 ,   41  ,0 };
            const unsigned long dim1237JoeKuoD6Init[]   =   {   1   ,   3   ,   1   ,   5   ,   7   ,   29  ,   85  ,   169 ,   479 ,   27  ,   1681    ,   917 ,   3983    ,   5041    ,0 };
            const unsigned long dim1238JoeKuoD6Init[]   =   {   1   ,   1   ,   7   ,   15  ,   13  ,   29  ,   39  ,   131 ,   491 ,   361 ,   565 ,   271 ,   6255    ,   10815   ,0 };
            const unsigned long dim1239JoeKuoD6Init[]   =   {   1   ,   1   ,   3   ,   3   ,   5   ,   63  ,   23  ,   201 ,   377 ,   593 ,   825 ,   339 ,   1065    ,   8989    ,0 };
            const unsigned long dim1240JoeKuoD6Init[]   =   {   1   ,   3   ,   1   ,   1   ,   19  ,   33  ,   103 ,   187 ,   327 ,   85  ,   55  ,   2561    ,   6833    ,   3889    ,0 };
            const unsigned long dim1241JoeKuoD6Init[]   =   {   1   ,   1   ,   5   ,   5   ,   5   ,   7   ,   85  ,   103 ,   463 ,   279 ,   47  ,   2479    ,   2149    ,   1421    ,0 };
            const unsigned long dim1242JoeKuoD6Init[]   =   {   1   ,   1   ,   3   ,   13  ,   17  ,   31  ,   61  ,   247 ,   59  ,   869 ,   1513    ,   2405    ,   2129    ,   15647   ,0 };
            const unsigned long dim1243JoeKuoD6Init[]   =   {   1   ,   1   ,   3   ,   13  ,   7   ,   31  ,   99  ,   63  ,   289 ,   123 ,   617 ,   3263    ,   6445    ,   14101   ,0 };
            const unsigned long dim1244JoeKuoD6Init[]   =   {   1   ,   3   ,   7   ,   5   ,   13  ,   23  ,   107 ,   73  ,   19  ,   595 ,   1623    ,   1583    ,   15  ,   10907   ,0 };
            const unsigned long dim1245JoeKuoD6Init[]   =   {   1   ,   1   ,   5   ,   15  ,   19  ,   33  ,   113 ,   45  ,   435 ,   99  ,   1909    ,   3317    ,   1733    ,   3629    ,0 };
            const unsigned long dim1246JoeKuoD6Init[]   =   {   1   ,   1   ,   7   ,   13  ,   3   ,   9   ,   107 ,   59  ,   69  ,   781 ,   1861    ,   393 ,   7143    ,   4465    ,0 };
            const unsigned long dim1247JoeKuoD6Init[]   =   {   1   ,   3   ,   5   ,   3   ,   23  ,   49  ,   3   ,   173 ,   483 ,   353 ,   783 ,   717 ,   2903    ,   11101   ,0 };
            const unsigned long dim1248JoeKuoD6Init[]   =   {   1   ,   3   ,   7   ,   15  ,   27  ,   25  ,   105 ,   47  ,   387 ,   103 ,   323 ,   455 ,   1607    ,   3673    ,0 };
            const unsigned long dim1249JoeKuoD6Init[]   =   {   1   ,   1   ,   7   ,   5   ,   19  ,   49  ,   59  ,   159 ,   5   ,   595 ,   75  ,   1963    ,   313 ,   5593    ,0 };
            const unsigned long dim1250JoeKuoD6Init[]   =   {   1   ,   3   ,   3   ,   15  ,   27  ,   35  ,   51  ,   135 ,   347 ,   375 ,   493 ,   589 ,   507 ,   4661    ,0 };
            const unsigned long dim1251JoeKuoD6Init[]   =   {   1   ,   1   ,   1   ,   5   ,   5   ,   59  ,   103 ,   249 ,   473 ,   503 ,   1401    ,   3219    ,   4337    ,   3607    ,0 };
            const unsigned long dim1252JoeKuoD6Init[]   =   {   1   ,   3   ,   1   ,   15  ,   29  ,   15  ,   123 ,   165 ,   55  ,   515 ,   1009    ,   3231    ,   3567    ,   7871    ,0 };
            const unsigned long dim1253JoeKuoD6Init[]   =   {   1   ,   3   ,   5   ,   7   ,   5   ,   11  ,   45  ,   87  ,   67  ,   359 ,   43  ,   457 ,   6483    ,   5165    ,0 };
            const unsigned long dim1254JoeKuoD6Init[]   =   {   1   ,   1   ,   1   ,   15  ,   13  ,   55  ,   85  ,   51  ,   127 ,   937 ,   1505    ,   1445    ,   889 ,   789 ,0 };
            const unsigned long dim1255JoeKuoD6Init[]   =   {   1   ,   1   ,   5   ,   1   ,   31  ,   27  ,   5   ,   225 ,   209 ,   291 ,   1047    ,   3417    ,   2899    ,   9691    ,0 };
            const unsigned long dim1256JoeKuoD6Init[]   =   {   1   ,   1   ,   1   ,   5   ,   23  ,   9   ,   83  ,   201 ,   511 ,   805 ,   1341    ,   2791    ,   4803    ,   13523   ,0 };
            const unsigned long dim1257JoeKuoD6Init[]   =   {   1   ,   3   ,   5   ,   7   ,   29  ,   29  ,   31  ,   239 ,   175 ,   1005    ,   1443    ,   2543    ,   6235    ,   15409   ,0 };
            const unsigned long dim1258JoeKuoD6Init[]   =   {   1   ,   3   ,   5   ,   5   ,   31  ,   7   ,   17  ,   195 ,   15  ,   65  ,   1297    ,   2707    ,   3097    ,   2281    ,0 };
            const unsigned long dim1259JoeKuoD6Init[]   =   {   1   ,   1   ,   3   ,   1   ,   25  ,   43  ,   57  ,   219 ,   419 ,   9   ,   1121    ,   593 ,   4403    ,   147 ,0 };
            const unsigned long dim1260JoeKuoD6Init[]   =   {   1   ,   1   ,   3   ,   7   ,   5   ,   31  ,   27  ,   43  ,   343 ,   95  ,   1917    ,   2425    ,   4939    ,   2937    ,0 };
            const unsigned long dim1261JoeKuoD6Init[]   =   {   1   ,   3   ,   1   ,   1   ,   31  ,   35  ,   19  ,   151 ,   339 ,   411 ,   77  ,   3881    ,   7401    ,   927 ,0 };
            const unsigned long dim1262JoeKuoD6Init[]   =   {   1   ,   3   ,   1   ,   7   ,   3   ,   61  ,   37  ,   135 ,   159 ,   81  ,   157 ,   2945    ,   6371    ,   3583    ,0 };
            const unsigned long dim1263JoeKuoD6Init[]   =   {   1   ,   1   ,   7   ,   1   ,   15  ,   27  ,   113 ,   213 ,   127 ,   141 ,   1431    ,   2163    ,   2527    ,   4027    ,0 };
            const unsigned long dim1264JoeKuoD6Init[]   =   {   1   ,   3   ,   1   ,   13  ,   1   ,   39  ,   103 ,   187 ,   395 ,   283 ,   907 ,   3977    ,   5901    ,   6043    ,0 };
            const unsigned long dim1265JoeKuoD6Init[]   =   {   1   ,   3   ,   7   ,   1   ,   31  ,   47  ,   45  ,   95  ,   53  ,   165 ,   357 ,   3893    ,   1555    ,   2263    ,0 };
            const unsigned long dim1266JoeKuoD6Init[]   =   {   1   ,   1   ,   1   ,   7   ,   1   ,   57  ,   39  ,   39  ,   365 ,   193 ,   887 ,   815 ,   7309    ,   15089   ,0 };
            const unsigned long dim1267JoeKuoD6Init[]   =   {   1   ,   1   ,   3   ,   11  ,   25  ,   61  ,   21  ,   139 ,   241 ,   167 ,   843 ,   299 ,   7737    ,   4625    ,0 };
            const unsigned long dim1268JoeKuoD6Init[]   =   {   1   ,   3   ,   5   ,   3   ,   29  ,   49  ,   115 ,   55  ,   185 ,   157 ,   515 ,   1985    ,   5891    ,   11189   ,0 };
            const unsigned long dim1269JoeKuoD6Init[]   =   {   1   ,   1   ,   1   ,   1   ,   1   ,   9   ,   41  ,   227 ,   267 ,   581 ,   5   ,   2677    ,   1231    ,   6989    ,0 };
            const unsigned long dim1270JoeKuoD6Init[]   =   {   1   ,   1   ,   3   ,   5   ,   13  ,   57  ,   33  ,   211 ,   333 ,   277 ,   1387    ,   2673    ,   5923    ,   15623   ,0 };
            const unsigned long dim1271JoeKuoD6Init[]   =   {   1   ,   1   ,   1   ,   9   ,   5   ,   57  ,   123 ,   93  ,   387 ,   975 ,   251 ,   677 ,   2843    ,   2237    ,0 };
            const unsigned long dim1272JoeKuoD6Init[]   =   {   1   ,   1   ,   7   ,   1   ,   5   ,   15  ,   19  ,   241 ,   171 ,   713 ,   1761    ,   3363    ,   2707    ,   7629    ,0 };
            const unsigned long dim1273JoeKuoD6Init[]   =   {   1   ,   1   ,   1   ,   13  ,   13  ,   25  ,   125 ,   233 ,   377 ,   1005    ,   1591    ,   1327    ,   7661    ,   15701   ,0 };
            const unsigned long dim1274JoeKuoD6Init[]   =   {   1   ,   1   ,   5   ,   15  ,   23  ,   39  ,   77  ,   121 ,   255 ,   105 ,   1739    ,   3511    ,   2369    ,   2209    ,0 };
            const unsigned long dim1275JoeKuoD6Init[]   =   {   1   ,   3   ,   5   ,   11  ,   5   ,   23  ,   11  ,   29  ,   501 ,   753 ,   1317    ,   1927    ,   2127    ,   7753    ,0 };
            const unsigned long dim1276JoeKuoD6Init[]   =   {   1   ,   3   ,   1   ,   1   ,   31  ,   25  ,   45  ,   197 ,   483 ,   787 ,   721 ,   2227    ,   4237    ,   1985    ,0 };
            const unsigned long dim1277JoeKuoD6Init[]   =   {   1   ,   3   ,   7   ,   7   ,   1   ,   53  ,   51  ,   235 ,   255 ,   517 ,   1389    ,   2381    ,   7845    ,   8707    ,0 };
            const unsigned long dim1278JoeKuoD6Init[]   =   {   1   ,   1   ,   1   ,   5   ,   27  ,   9   ,   19  ,   251 ,   29  ,   211 ,   1665    ,   1181    ,   983 ,   16139   ,0 };
            const unsigned long dim1279JoeKuoD6Init[]   =   {   1   ,   1   ,   5   ,   11  ,   21  ,   35  ,   39  ,   47  ,   445 ,   685 ,   773 ,   3563    ,   3339    ,   14855   ,0 };
            const unsigned long dim1280JoeKuoD6Init[]   =   {   1   ,   3   ,   7   ,   7   ,   29  ,   61  ,   89  ,   203 ,   239 ,   11  ,   1407    ,   1529    ,   3581    ,   11443   ,0 };
            const unsigned long dim1281JoeKuoD6Init[]   =   {   1   ,   3   ,   3   ,   5   ,   19  ,   21  ,   3   ,   253 ,   437 ,   755 ,   1409    ,   1449    ,   623 ,   10533   ,0 };
            const unsigned long dim1282JoeKuoD6Init[]   =   {   1   ,   3   ,   3   ,   15  ,   13  ,   25  ,   65  ,   123 ,   273 ,   611 ,   1305    ,   2751    ,   5123    ,   1515    ,0 };
            const unsigned long dim1283JoeKuoD6Init[]   =   {   1   ,   1   ,   3   ,   5   ,   13  ,   63  ,   79  ,   139 ,   253 ,   917 ,   19  ,   1129    ,   7801    ,   14215   ,0 };
            const unsigned long dim1284JoeKuoD6Init[]   =   {   1   ,   1   ,   5   ,   9   ,   5   ,   21  ,   29  ,   179 ,   79  ,   803 ,   1163    ,   2481    ,   3537    ,   11015   ,0 };
            const unsigned long dim1285JoeKuoD6Init[]   =   {   1   ,   3   ,   5   ,   1   ,   9   ,   51  ,   35  ,   241 ,   473 ,   291 ,   147 ,   3021    ,   2147    ,   4437    ,0 };
            const unsigned long dim1286JoeKuoD6Init[]   =   {   1   ,   1   ,   7   ,   15  ,   7   ,   19  ,   81  ,   215 ,   205 ,   19  ,   1713    ,   1347    ,   7519    ,   9851    ,0 };
            const unsigned long dim1287JoeKuoD6Init[]   =   {   1   ,   1   ,   7   ,   1   ,   9   ,   49  ,   35  ,   219 ,   73  ,   479 ,   559 ,   1619    ,   7891    ,   1221    ,0 };
            const unsigned long dim1288JoeKuoD6Init[]   =   {   1   ,   1   ,   3   ,   1   ,   9   ,   39  ,   17  ,   67  ,   139 ,   287 ,   703 ,   3797    ,   4263    ,   10847   ,0 };
            const unsigned long dim1289JoeKuoD6Init[]   =   {   1   ,   3   ,   7   ,   9   ,   23  ,   23  ,   17  ,   221 ,   185 ,   209 ,   1827    ,   3117    ,   8183    ,   15405   ,0 };
            const unsigned long dim1290JoeKuoD6Init[]   =   {   1   ,   3   ,   5   ,   9   ,   9   ,   9   ,   95  ,   1   ,   225 ,   727 ,   1577    ,   3947    ,   7995    ,   2587    ,0 };
            const unsigned long dim1291JoeKuoD6Init[]   =   {   1   ,   1   ,   5   ,   7   ,   5   ,   29  ,   89  ,   193 ,   427 ,   1009    ,   1123    ,   717 ,   3191    ,   14381   ,0 };
            const unsigned long dim1292JoeKuoD6Init[]   =   {   1   ,   3   ,   1   ,   9   ,   7   ,   53  ,   57  ,   147 ,   511 ,   887 ,   501 ,   431 ,   7537    ,   363 ,0 };
            const unsigned long dim1293JoeKuoD6Init[]   =   {   1   ,   1   ,   3   ,   7   ,   3   ,   21  ,   127 ,   157 ,   45  ,   773 ,   1973    ,   2577    ,   6057    ,   8153    ,0 };
            const unsigned long dim1294JoeKuoD6Init[]   =   {   1   ,   3   ,   3   ,   9   ,   31  ,   63  ,   23  ,   47  ,   365 ,   369 ,   97  ,   1627    ,   831 ,   12769   ,0 };
            const unsigned long dim1295JoeKuoD6Init[]   =   {   1   ,   3   ,   5   ,   9   ,   5   ,   61  ,   53  ,   131 ,   447 ,   281 ,   1115    ,   2523    ,   3789    ,   11581   ,0 };
            const unsigned long dim1296JoeKuoD6Init[]   =   {   1   ,   1   ,   3   ,   5   ,   19  ,   39  ,   95  ,   243 ,   95  ,   601 ,   1385    ,   2523    ,   2973    ,   2759    ,0 };
            const unsigned long dim1297JoeKuoD6Init[]   =   {   1   ,   3   ,   1   ,   15  ,   25  ,   15  ,   19  ,   133 ,   1   ,   197 ,   1259    ,   2501    ,   2513    ,   9417    ,0 };
            const unsigned long dim1298JoeKuoD6Init[]   =   {   1   ,   1   ,   7   ,   1   ,   13  ,   49  ,   47  ,   61  ,   167 ,   971 ,   835 ,   2259    ,   3747    ,   13077   ,0 };
            const unsigned long dim1299JoeKuoD6Init[]   =   {   1   ,   1   ,   5   ,   9   ,   17  ,   35  ,   33  ,   223 ,   173 ,   691 ,   855 ,   2743    ,   5383    ,   2017    ,0 };
            const unsigned long dim1300JoeKuoD6Init[]   =   {   1   ,   3   ,   1   ,   5   ,   11  ,   35  ,   19  ,   255 ,   201 ,   801 ,   1921    ,   2511    ,   2497    ,   13317   ,0 };
            const unsigned long dim1301JoeKuoD6Init[]   =   {   1   ,   1   ,   7   ,   9   ,   17  ,   1   ,   47  ,   187 ,   63  ,   641 ,   1395    ,   1591    ,   6145    ,   4549    ,0 };
            const unsigned long dim1302JoeKuoD6Init[]   =   {   1   ,   1   ,   1   ,   1   ,   11  ,   35  ,   55  ,   167 ,   75  ,   801 ,   257 ,   2531    ,   1821    ,   11325   ,0 };
            const unsigned long dim1303JoeKuoD6Init[]   =   {   1   ,   1   ,   1   ,   11  ,   21  ,   27  ,   117 ,   131 ,   167 ,   461 ,   1459    ,   1   ,   4847    ,   4347    ,0 };
            const unsigned long dim1304JoeKuoD6Init[]   =   {   1   ,   1   ,   5   ,   15  ,   7   ,   19  ,   103 ,   233 ,   509 ,   397 ,   1147    ,   1325    ,   6173    ,   9643    ,0 };
            const unsigned long dim1305JoeKuoD6Init[]   =   {   1   ,   3   ,   1   ,   11  ,   31  ,   37  ,   51  ,   43  ,   173 ,   575 ,   7   ,   3595    ,   5865    ,   7539    ,0 };
            const unsigned long dim1306JoeKuoD6Init[]   =   {   1   ,   1   ,   5   ,   15  ,   17  ,   63  ,   109 ,   29  ,   325 ,   57  ,   889 ,   1197    ,   7839    ,   13931   ,0 };
            const unsigned long dim1307JoeKuoD6Init[]   =   {   1   ,   1   ,   5   ,   15  ,   3   ,   19  ,   105 ,   251 ,   447 ,   451 ,   925 ,   1857    ,   4343    ,   14273   ,0 };
            const unsigned long dim1308JoeKuoD6Init[]   =   {   1   ,   3   ,   1   ,   9   ,   27  ,   29  ,   97  ,   113 ,   155 ,   591 ,   1435    ,   2175    ,   4197    ,   10661   ,0 };
            const unsigned long dim1309JoeKuoD6Init[]   =   {   1   ,   3   ,   1   ,   3   ,   9   ,   11  ,   33  ,   107 ,   187 ,   463 ,   1479    ,   1127    ,   6699    ,   2183    ,0 };
            const unsigned long dim1310JoeKuoD6Init[]   =   {   1   ,   3   ,   3   ,   1   ,   5   ,   15  ,   73  ,   231 ,   73  ,   161 ,   525 ,   2265    ,   619 ,   8085    ,0 };
            const unsigned long dim1311JoeKuoD6Init[]   =   {   1   ,   1   ,   1   ,   1   ,   1   ,   43  ,   7   ,   137 ,   163 ,   953 ,   775 ,   3839    ,   7377    ,   12321   ,0 };
            const unsigned long dim1312JoeKuoD6Init[]   =   {   1   ,   1   ,   3   ,   3   ,   27  ,   57  ,   11  ,   131 ,   407 ,   41  ,   173 ,   2595    ,   5179    ,   1455    ,0 };
            const unsigned long dim1313JoeKuoD6Init[]   =   {   1   ,   3   ,   5   ,   3   ,   31  ,   9   ,   93  ,   233 ,   61  ,   149 ,   1973    ,   3493    ,   5535    ,   9595    ,0 };
            const unsigned long dim1314JoeKuoD6Init[]   =   {   1   ,   1   ,   5   ,   9   ,   1   ,   37  ,   13  ,   71  ,   451 ,   645 ,   51  ,   3663    ,   1917    ,   10641   ,0 };
            const unsigned long dim1315JoeKuoD6Init[]   =   {   1   ,   3   ,   1   ,   11  ,   13  ,   53  ,   111 ,   71  ,   61  ,   539 ,   319 ,   1951    ,   6763    ,   751 ,0 };
            const unsigned long dim1316JoeKuoD6Init[]   =   {   1   ,   1   ,   7   ,   1   ,   27  ,   27  ,   53  ,   189 ,   39  ,   29  ,   1645    ,   235 ,   7643    ,   15621   ,0 };
            const unsigned long dim1317JoeKuoD6Init[]   =   {   1   ,   1   ,   7   ,   1   ,   13  ,   13  ,   115 ,   125 ,   109 ,   377 ,   463 ,   3819    ,   7397    ,   5205    ,0 };
            const unsigned long dim1318JoeKuoD6Init[]   =   {   1   ,   1   ,   3   ,   9   ,   29  ,   59  ,   61  ,   145 ,   187 ,   875 ,   1571    ,   4035    ,   2861    ,   10897   ,0 };
            const unsigned long dim1319JoeKuoD6Init[]   =   {   1   ,   1   ,   7   ,   1   ,   11  ,   5   ,   9   ,   19  ,   477 ,   201 ,   1061    ,   1483    ,   7597    ,   7161    ,0 };
            const unsigned long dim1320JoeKuoD6Init[]   =   {   1   ,   3   ,   1   ,   13  ,   17  ,   35  ,   55  ,   211 ,   285 ,   161 ,   385 ,   4093    ,   3685    ,   3635    ,0 };
            const unsigned long dim1321JoeKuoD6Init[]   =   {   1   ,   3   ,   1   ,   9   ,   19  ,   43  ,   95  ,   253 ,   453 ,   829 ,   857 ,   3913    ,   4949    ,   9057    ,0 };
            const unsigned long dim1322JoeKuoD6Init[]   =   {   1   ,   1   ,   1   ,   13  ,   3   ,   45  ,   57  ,   185 ,   7   ,   457 ,   1805    ,   3947    ,   1515    ,   4863    ,0 };
            const unsigned long dim1323JoeKuoD6Init[]   =   {   1   ,   1   ,   5   ,   1   ,   9   ,   27  ,   75  ,   243 ,   401 ,   593 ,   1669    ,   3315    ,   3023    ,   14677   ,0 };
            const unsigned long dim1324JoeKuoD6Init[]   =   {   1   ,   1   ,   3   ,   13  ,   29  ,   57  ,   37  ,   11  ,   153 ,   255 ,   1039    ,   2067    ,   2033    ,   7927    ,0 };
            const unsigned long dim1325JoeKuoD6Init[]   =   {   1   ,   3   ,   3   ,   13  ,   27  ,   63  ,   23  ,   73  ,   355 ,   913 ,   327 ,   43  ,   2095    ,   433 ,0 };
            const unsigned long dim1326JoeKuoD6Init[]   =   {   1   ,   3   ,   3   ,   7   ,   13  ,   1   ,   61  ,   193 ,   73  ,   465 ,   1483    ,   1359    ,   7459    ,   6855    ,0 };
            const unsigned long dim1327JoeKuoD6Init[]   =   {   1   ,   1   ,   7   ,   3   ,   15  ,   39  ,   85  ,   165 ,   285 ,   713 ,   1455    ,   403 ,   7723    ,   6595    ,0 };
            const unsigned long dim1328JoeKuoD6Init[]   =   {   1   ,   1   ,   3   ,   9   ,   11  ,   39  ,   67  ,   43  ,   47  ,   85  ,   1831    ,   1387    ,   1423    ,   12473   ,0 };
            const unsigned long dim1329JoeKuoD6Init[]   =   {   1   ,   3   ,   5   ,   3   ,   23  ,   11  ,   49  ,   195 ,   207 ,   403 ,   1761    ,   3633    ,   8069    ,   12455   ,0 };
            const unsigned long dim1330JoeKuoD6Init[]   =   {   1   ,   1   ,   3   ,   7   ,   19  ,   11  ,   27  ,   131 ,   123 ,   685 ,   931 ,   2433    ,   7133    ,   1997    ,0 };
            const unsigned long dim1331JoeKuoD6Init[]   =   {   1   ,   3   ,   3   ,   7   ,   27  ,   3   ,   69  ,   77  ,   227 ,   227 ,   213 ,   1277    ,   6597    ,   8343    ,0 };
            const unsigned long dim1332JoeKuoD6Init[]   =   {   1   ,   3   ,   7   ,   5   ,   13  ,   61  ,   69  ,   63  ,   59  ,   949 ,   2019    ,   2723    ,   451 ,   9641    ,0 };
            const unsigned long dim1333JoeKuoD6Init[]   =   {   1   ,   1   ,   7   ,   11  ,   19  ,   29  ,   75  ,   149 ,   481 ,   583 ,   861 ,   1353    ,   7525    ,   115 ,0 };
            const unsigned long dim1334JoeKuoD6Init[]   =   {   1   ,   3   ,   5   ,   13  ,   19  ,   3   ,   27  ,   155 ,   81  ,   907 ,   683 ,   1997    ,   189 ,   12139   ,0 };
            const unsigned long dim1335JoeKuoD6Init[]   =   {   1   ,   1   ,   5   ,   9   ,   19  ,   1   ,   61  ,   97  ,   433 ,   289 ,   1191    ,   2347    ,   5533    ,   14327   ,0 };
            const unsigned long dim1336JoeKuoD6Init[]   =   {   1   ,   3   ,   1   ,   5   ,   13  ,   9   ,   7   ,   91  ,   49  ,   75  ,   1839    ,   415 ,   795 ,   9127    ,0 };
            const unsigned long dim1337JoeKuoD6Init[]   =   {   1   ,   1   ,   1   ,   11  ,   13  ,   11  ,   123 ,   31  ,   261 ,   159 ,   681 ,   231 ,   117 ,   14923   ,0 };
            const unsigned long dim1338JoeKuoD6Init[]   =   {   1   ,   1   ,   7   ,   1   ,   21  ,   35  ,   93  ,   17  ,   251 ,   229 ,   553 ,   3413    ,   7717    ,   891 ,0 };
            const unsigned long dim1339JoeKuoD6Init[]   =   {   1   ,   3   ,   5   ,   3   ,   7   ,   47  ,   85  ,   155 ,   63  ,   785 ,   939 ,   3135    ,   3319    ,   1997    ,0 };
            const unsigned long dim1340JoeKuoD6Init[]   =   {   1   ,   3   ,   7   ,   1   ,   21  ,   19  ,   7   ,   69  ,   505 ,   327 ,   1015    ,   1821    ,   2583    ,   15031   ,0 };
            const unsigned long dim1341JoeKuoD6Init[]   =   {   1   ,   3   ,   5   ,   15  ,   3   ,   55  ,   87  ,   111 ,   499 ,   845 ,   375 ,   2735    ,   6895    ,   291 ,0 };
            const unsigned long dim1342JoeKuoD6Init[]   =   {   1   ,   1   ,   3   ,   3   ,   15  ,   5   ,   51  ,   105 ,   369 ,   91  ,   945 ,   3633    ,   7847    ,   13811   ,0 };
            const unsigned long dim1343JoeKuoD6Init[]   =   {   1   ,   1   ,   1   ,   5   ,   21  ,   57  ,   79  ,   213 ,   225 ,   71  ,   881 ,   391 ,   4329    ,   2481    ,0 };
            const unsigned long dim1344JoeKuoD6Init[]   =   {   1   ,   1   ,   7   ,   11  ,   19  ,   49  ,   39  ,   193 ,   271 ,   463 ,   1009    ,   137 ,   3751    ,   8439    ,0 };
            const unsigned long dim1345JoeKuoD6Init[]   =   {   1   ,   3   ,   1   ,   3   ,   17  ,   37  ,   51  ,   211 ,   103 ,   929 ,   1323    ,   1175    ,   5465    ,   12229   ,0 };
            const unsigned long dim1346JoeKuoD6Init[]   =   {   1   ,   1   ,   3   ,   9   ,   21  ,   9   ,   123 ,   221 ,   201 ,   769 ,   723 ,   2627    ,   2367    ,   13169   ,0 };
            const unsigned long dim1347JoeKuoD6Init[]   =   {   1   ,   1   ,   1   ,   11  ,   31  ,   23  ,   93  ,   103 ,   257 ,   263 ,   1771    ,   801 ,   4735    ,   13557   ,0 };
            const unsigned long dim1348JoeKuoD6Init[]   =   {   1   ,   3   ,   1   ,   9   ,   21  ,   29  ,   67  ,   225 ,   367 ,   909 ,   1333    ,   655 ,   7691    ,   1251    ,0 };
            const unsigned long dim1349JoeKuoD6Init[]   =   {   1   ,   1   ,   3   ,   7   ,   19  ,   11  ,   75  ,   175 ,   383 ,   121 ,   1917    ,   2371    ,   4191    ,   16111   ,0 };
            const unsigned long dim1350JoeKuoD6Init[]   =   {   1   ,   1   ,   3   ,   13  ,   17  ,   5   ,   103 ,   53  ,   173 ,   981 ,   1185    ,   3071    ,   179 ,   9197    ,0 };
            const unsigned long dim1351JoeKuoD6Init[]   =   {   1   ,   1   ,   3   ,   9   ,   5   ,   41  ,   25  ,   219 ,   211 ,   291 ,   1735    ,   397 ,   6317    ,   11209   ,0 };
            const unsigned long dim1352JoeKuoD6Init[]   =   {   1   ,   3   ,   7   ,   5   ,   5   ,   1   ,   11  ,   125 ,   405 ,   583 ,   773 ,   1711    ,   757 ,   5187    ,0 };
            const unsigned long dim1353JoeKuoD6Init[]   =   {   1   ,   1   ,   1   ,   7   ,   1   ,   45  ,   127 ,   11  ,   157 ,   255 ,   1167    ,   3315    ,   5059    ,   8219    ,0 };
            const unsigned long dim1354JoeKuoD6Init[]   =   {   1   ,   1   ,   1   ,   5   ,   15  ,   55  ,   49  ,   81  ,   451 ,   887 ,   309 ,   2321    ,   5759    ,   4673    ,0 };
            const unsigned long dim1355JoeKuoD6Init[]   =   {   1   ,   1   ,   3   ,   3   ,   29  ,   51  ,   15  ,   113 ,   245 ,   77  ,   109 ,   2813    ,   225 ,   5479    ,0 };
            const unsigned long dim1356JoeKuoD6Init[]   =   {   1   ,   1   ,   3   ,   13  ,   13  ,   43  ,   5   ,   99  ,   89  ,   559 ,   903 ,   1003    ,   5401    ,   1839    ,0 };
            const unsigned long dim1357JoeKuoD6Init[]   =   {   1   ,   1   ,   1   ,   5   ,   25  ,   11  ,   3   ,   225 ,   341 ,   439 ,   521 ,   545 ,   5529    ,   5365    ,0 };
            const unsigned long dim1358JoeKuoD6Init[]   =   {   1   ,   1   ,   7   ,   7   ,   13  ,   41  ,   41  ,   183 ,   319 ,   593 ,   1961    ,   2211    ,   7155    ,   1683    ,0 };
            const unsigned long dim1359JoeKuoD6Init[]   =   {   1   ,   1   ,   3   ,   1   ,   9   ,   27  ,   115 ,   119 ,   289 ,   935 ,   1173    ,   1783    ,   7151    ,   7157    ,0 };
            const unsigned long dim1360JoeKuoD6Init[]   =   {   1   ,   3   ,   3   ,   7   ,   17  ,   33  ,   31  ,   109 ,   245 ,   785 ,   1563    ,   999 ,   7989    ,   3493    ,0 };
            const unsigned long dim1361JoeKuoD6Init[]   =   {   1   ,   1   ,   7   ,   7   ,   9   ,   3   ,   85  ,   91  ,   21  ,   831 ,   1669    ,   1857    ,   1737    ,   12513   ,0 };
            const unsigned long dim1362JoeKuoD6Init[]   =   {   1   ,   3   ,   3   ,   5   ,   17  ,   51  ,   37  ,   175 ,   325 ,   581 ,   1809    ,   2669    ,   6179    ,   2187    ,0 };
            const unsigned long dim1363JoeKuoD6Init[]   =   {   1   ,   3   ,   5   ,   9   ,   9   ,   39  ,   53  ,   87  ,   47  ,   677 ,   1081    ,   171 ,   4885    ,   2311    ,0 };
            const unsigned long dim1364JoeKuoD6Init[]   =   {   1   ,   1   ,   5   ,   9   ,   29  ,   3   ,   79  ,   13  ,   433 ,   891 ,   131 ,   3559    ,   797 ,   3405    ,0 };
            const unsigned long dim1365JoeKuoD6Init[]   =   {   1   ,   3   ,   7   ,   11  ,   25  ,   5   ,   73  ,   189 ,   251 ,   927 ,   2001    ,   3179    ,   859 ,   15931   ,0 };
            const unsigned long dim1366JoeKuoD6Init[]   =   {   1   ,   1   ,   1   ,   7   ,   15  ,   61  ,   67  ,   149 ,   99  ,   795 ,   53  ,   2417    ,   4943    ,   7367    ,0 };
            const unsigned long dim1367JoeKuoD6Init[]   =   {   1   ,   3   ,   1   ,   3   ,   25  ,   61  ,   79  ,   145 ,   283 ,   161 ,   887 ,   137 ,   6027    ,   12585   ,0 };
            const unsigned long dim1368JoeKuoD6Init[]   =   {   1   ,   3   ,   3   ,   3   ,   31  ,   5   ,   117 ,   33  ,   21  ,   883 ,   117 ,   1609    ,   7219    ,   2511    ,0 };
            const unsigned long dim1369JoeKuoD6Init[]   =   {   1   ,   3   ,   3   ,   13  ,   1   ,   59  ,   29  ,   177 ,   419 ,   287 ,   973 ,   3649    ,   3333    ,   9779    ,0 };
            const unsigned long dim1370JoeKuoD6Init[]   =   {   1   ,   3   ,   5   ,   5   ,   11  ,   51  ,   11  ,   125 ,   237 ,   305 ,   37  ,   3137    ,   7645    ,   6317    ,0 };
            const unsigned long dim1371JoeKuoD6Init[]   =   {   1   ,   1   ,   3   ,   3   ,   29  ,   39  ,   127 ,   43  ,   113 ,   437 ,   1501    ,   13  ,   1915    ,   7561    ,0 };
            const unsigned long dim1372JoeKuoD6Init[]   =   {   1   ,   3   ,   5   ,   13  ,   9   ,   55  ,   7   ,   209 ,   381 ,   725 ,   1867    ,   637 ,   6335    ,   6057    ,0 };
            const unsigned long dim1373JoeKuoD6Init[]   =   {   1   ,   3   ,   3   ,   5   ,   1   ,   57  ,   57  ,   45  ,   247 ,   23  ,   1663    ,   489 ,   2503    ,   9033    ,0 };
            const unsigned long dim1374JoeKuoD6Init[]   =   {   1   ,   3   ,   1   ,   3   ,   23  ,   7   ,   57  ,   157 ,   291 ,   523 ,   313 ,   2501    ,   1197    ,   3353    ,0 };
            const unsigned long dim1375JoeKuoD6Init[]   =   {   1   ,   3   ,   5   ,   11  ,   31  ,   13  ,   35  ,   253 ,   349 ,   393 ,   1531    ,   159 ,   5311    ,   9225    ,0 };
            const unsigned long dim1376JoeKuoD6Init[]   =   {   1   ,   1   ,   7   ,   1   ,   1   ,   47  ,   75  ,   187 ,   41  ,   713 ,   1933    ,   2139    ,   2211    ,   7425    ,0 };
            const unsigned long dim1377JoeKuoD6Init[]   =   {   1   ,   1   ,   3   ,   5   ,   7   ,   21  ,   67  ,   27  ,   303 ,   617 ,   1943    ,   3173    ,   3081    ,   10977   ,0 };
            const unsigned long dim1378JoeKuoD6Init[]   =   {   1   ,   1   ,   7   ,   7   ,   27  ,   57  ,   103 ,   65  ,   195 ,   267 ,   1347    ,   3831    ,   5335    ,   435 ,0 };
            const unsigned long dim1379JoeKuoD6Init[]   =   {   1   ,   1   ,   7   ,   13  ,   11  ,   35  ,   83  ,   21  ,   273 ,   227 ,   1523    ,   3565    ,   7221    ,   3105    ,0 };
            const unsigned long dim1380JoeKuoD6Init[]   =   {   1   ,   1   ,   5   ,   15  ,   7   ,   29  ,   123 ,   207 ,   63  ,   535 ,   1355    ,   2751    ,   4791    ,   10285   ,0 };
            const unsigned long dim1381JoeKuoD6Init[]   =   {   1   ,   1   ,   3   ,   5   ,   3   ,   19  ,   79  ,   91  ,   481 ,   205 ,   1741    ,   319 ,   4651    ,   12631   ,0 };
            const unsigned long dim1382JoeKuoD6Init[]   =   {   1   ,   3   ,   3   ,   7   ,   25  ,   61  ,   39  ,   181 ,   377 ,   243 ,   593 ,   1383    ,   1427    ,   11865   ,0 };
            const unsigned long dim1383JoeKuoD6Init[]   =   {   1   ,   1   ,   3   ,   9   ,   13  ,   13  ,   97  ,   27  ,   335 ,   973 ,   1857    ,   425 ,   2729    ,   12013   ,0 };
            const unsigned long dim1384JoeKuoD6Init[]   =   {   1   ,   1   ,   5   ,   7   ,   23  ,   57  ,   99  ,   37  ,   117 ,   337 ,   751 ,   2471    ,   1105    ,   1415    ,0 };
            const unsigned long dim1385JoeKuoD6Init[]   =   {   1   ,   3   ,   5   ,   1   ,   17  ,   33  ,   117 ,   87  ,   211 ,   899 ,   825 ,   1043    ,   4273    ,   9093    ,0 };
            const unsigned long dim1386JoeKuoD6Init[]   =   {   1   ,   1   ,   3   ,   13  ,   11  ,   7   ,   81  ,   251 ,   421 ,   643 ,   1931    ,   1205    ,   1033    ,   8741    ,0 };
            const unsigned long dim1387JoeKuoD6Init[]   =   {   1   ,   3   ,   7   ,   5   ,   17  ,   27  ,   113 ,   43  ,   63  ,   819 ,   503 ,   1427    ,   5787    ,   1133    ,0 };
            const unsigned long dim1388JoeKuoD6Init[]   =   {   1   ,   1   ,   1   ,   13  ,   7   ,   31  ,   85  ,   123 ,   17  ,   243 ,   749 ,   1447    ,   6045    ,   729 ,0 };
            const unsigned long dim1389JoeKuoD6Init[]   =   {   1   ,   1   ,   1   ,   3   ,   19  ,   15  ,   97  ,   49  ,   101 ,   37  ,   1123    ,   1565    ,   6957    ,   1209    ,0 };
            const unsigned long dim1390JoeKuoD6Init[]   =   {   1   ,   3   ,   1   ,   5   ,   1   ,   39  ,   23  ,   47  ,   387 ,   727 ,   81  ,   4023    ,   7635    ,   8691    ,0 };
            const unsigned long dim1391JoeKuoD6Init[]   =   {   1   ,   3   ,   7   ,   11  ,   19  ,   27  ,   71  ,   37  ,   281 ,   981 ,   1777    ,   1751    ,   7387    ,   8391    ,0 };
            const unsigned long dim1392JoeKuoD6Init[]   =   {   1   ,   3   ,   7   ,   11  ,   21  ,   55  ,   55  ,   35  ,   205 ,   277 ,   1573    ,   327 ,   7643    ,   4901    ,0 };
            const unsigned long dim1393JoeKuoD6Init[]   =   {   1   ,   1   ,   5   ,   11  ,   15  ,   45  ,   91  ,   157 ,   339 ,   489 ,   1871    ,   2841    ,   6839    ,   4067    ,0 };
            const unsigned long dim1394JoeKuoD6Init[]   =   {   1   ,   1   ,   7   ,   3   ,   11  ,   13  ,   57  ,   193 ,   499 ,   729 ,   603 ,   2235    ,   6323    ,   16307   ,0 };
            const unsigned long dim1395JoeKuoD6Init[]   =   {   1   ,   1   ,   5   ,   9   ,   13  ,   41  ,   127 ,   117 ,   115 ,   159 ,   1335    ,   1473    ,   6509    ,   8165    ,0 };
            const unsigned long dim1396JoeKuoD6Init[]   =   {   1   ,   3   ,   7   ,   15  ,   29  ,   39  ,   57  ,   161 ,   71  ,   97  ,   1201    ,   649 ,   5863    ,   4765    ,0 };
            const unsigned long dim1397JoeKuoD6Init[]   =   {   1   ,   3   ,   1   ,   1   ,   21  ,   3   ,   3   ,   129 ,   421 ,   575 ,   1693    ,   337 ,   7313    ,   4801    ,0 };
            const unsigned long dim1398JoeKuoD6Init[]   =   {   1   ,   3   ,   3   ,   15  ,   27  ,   51  ,   105 ,   91  ,   51  ,   683 ,   795 ,   603 ,   7759    ,   251 ,0 };
            const unsigned long dim1399JoeKuoD6Init[]   =   {   1   ,   1   ,   5   ,   7   ,   9   ,   45  ,   55  ,   199 ,   83  ,   575 ,   521 ,   2169    ,   5741    ,   1579    ,0 };
            const unsigned long dim1400JoeKuoD6Init[]   =   {   1   ,   3   ,   1   ,   7   ,   13  ,   47  ,   125 ,   227 ,   45  ,   355 ,   127 ,   1841    ,   727 ,   8555    ,0 };
            const unsigned long dim1401JoeKuoD6Init[]   =   {   1   ,   3   ,   7   ,   11  ,   27  ,   9   ,   57  ,   133 ,   393 ,   315 ,   1411    ,   453 ,   3787    ,   7651    ,0 };
            const unsigned long dim1402JoeKuoD6Init[]   =   {   1   ,   1   ,   1   ,   15  ,   11  ,   35  ,   77  ,   15  ,   223 ,   1009    ,   291 ,   2645    ,   4585    ,   5839    ,0 };
            const unsigned long dim1403JoeKuoD6Init[]   =   {   1   ,   3   ,   1   ,   7   ,   7   ,   61  ,   89  ,   55  ,   399 ,   651 ,   1847    ,   1901    ,   1887    ,   5841    ,0 };
            const unsigned long dim1404JoeKuoD6Init[]   =   {   1   ,   3   ,   3   ,   3   ,   17  ,   29  ,   41  ,   19  ,   353 ,   475 ,   1193    ,   949 ,   3351    ,   5073    ,0 };
            const unsigned long dim1405JoeKuoD6Init[]   =   {   1   ,   1   ,   1   ,   5   ,   29  ,   45  ,   37  ,   39  ,   343 ,   661 ,   1405    ,   229 ,   493 ,   471 ,0 };
            const unsigned long dim1406JoeKuoD6Init[]   =   {   1   ,   3   ,   5   ,   5   ,   25  ,   27  ,   37  ,   33  ,   33  ,   139 ,   1537    ,   1125    ,   4331    ,   253 ,0 };
            const unsigned long dim1407JoeKuoD6Init[]   =   {   1   ,   3   ,   3   ,   7   ,   19  ,   31  ,   111 ,   243 ,   201 ,   563 ,   1913    ,   2153    ,   749 ,   3521    ,0 };
            const unsigned long dim1408JoeKuoD6Init[]   =   {   1   ,   3   ,   5   ,   1   ,   9   ,   41  ,   19  ,   163 ,   185 ,   435 ,   1997    ,   3383    ,   1721    ,   11137   ,0 };
            const unsigned long dim1409JoeKuoD6Init[]   =   {   1   ,   3   ,   3   ,   15  ,   25  ,   1   ,   67  ,   231 ,   11  ,   197 ,   1431    ,   1697    ,   3337    ,   8791    ,0 };
            const unsigned long dim1410JoeKuoD6Init[]   =   {   1   ,   1   ,   1   ,   1   ,   13  ,   59  ,   11  ,   199 ,   243 ,   105 ,   751 ,   811 ,   345 ,   8183    ,0 };
            const unsigned long dim1411JoeKuoD6Init[]   =   {   1   ,   1   ,   3   ,   1   ,   29  ,   35  ,   123 ,   187 ,   411 ,   523 ,   989 ,   2545    ,   4969    ,   14489   ,0 };
            const unsigned long dim1412JoeKuoD6Init[]   =   {   1   ,   3   ,   7   ,   5   ,   29  ,   23  ,   51  ,   83  ,   335 ,   769 ,   171 ,   1631    ,   1651    ,   10549   ,0 };
            const unsigned long dim1413JoeKuoD6Init[]   =   {   1   ,   1   ,   3   ,   9   ,   13  ,   29  ,   57  ,   167 ,   113 ,   703 ,   1027    ,   163 ,   1327    ,   5649    ,0 };
            const unsigned long dim1414JoeKuoD6Init[]   =   {   1   ,   3   ,   5   ,   11  ,   1   ,   47  ,   101 ,   251 ,   377 ,   421 ,   1825    ,   3153    ,   3425    ,   10447   ,0 };
            const unsigned long dim1415JoeKuoD6Init[]   =   {   1   ,   3   ,   1   ,   1   ,   11  ,   49  ,   17  ,   215 ,   261 ,   899 ,   811 ,   3451    ,   3955    ,   791 ,0 };
            const unsigned long dim1416JoeKuoD6Init[]   =   {   1   ,   1   ,   1   ,   11  ,   19  ,   23  ,   41  ,   239 ,   203 ,   321 ,   993 ,   1201    ,   5073    ,   8329    ,0 };
            const unsigned long dim1417JoeKuoD6Init[]   =   {   1   ,   3   ,   3   ,   7   ,   23  ,   55  ,   81  ,   227 ,   379 ,   571 ,   1869    ,   2159    ,   8181    ,   12901   ,0 };
            const unsigned long dim1418JoeKuoD6Init[]   =   {   1   ,   1   ,   7   ,   3   ,   31  ,   55  ,   13  ,   57  ,   413 ,   927 ,   873 ,   3649    ,   7167    ,   3985    ,0 };
            const unsigned long dim1419JoeKuoD6Init[]   =   {   1   ,   1   ,   7   ,   1   ,   9   ,   49  ,   7   ,   41  ,   461 ,   959 ,   1477    ,   4037    ,   1119    ,   3525    ,0 };
            const unsigned long dim1420JoeKuoD6Init[]   =   {   1   ,   3   ,   5   ,   9   ,   9   ,   31  ,   67  ,   245 ,   57  ,   395 ,   1035    ,   3005    ,   4211    ,   7155    ,0 };
            const unsigned long dim1421JoeKuoD6Init[]   =   {   1   ,   1   ,   7   ,   9   ,   5   ,   59  ,   55  ,   7   ,   97  ,   707 ,   1755    ,   1921    ,   6239    ,   233 ,0 };
            const unsigned long dim1422JoeKuoD6Init[]   =   {   1   ,   1   ,   7   ,   5   ,   11  ,   11  ,   47  ,   195 ,   197 ,   37  ,   501 ,   2013    ,   6579    ,   15115   ,0 };
            const unsigned long dim1423JoeKuoD6Init[]   =   {   1   ,   1   ,   3   ,   15  ,   23  ,   61  ,   27  ,   199 ,   41  ,   479 ,   871 ,   2365    ,   1589    ,   13  ,0 };
            const unsigned long dim1424JoeKuoD6Init[]   =   {   1   ,   1   ,   3   ,   3   ,   11  ,   9   ,   1   ,   225 ,   295 ,   403 ,   1065    ,   3039    ,   6025    ,   963 ,0 };
            const unsigned long dim1425JoeKuoD6Init[]   =   {   1   ,   3   ,   7   ,   13  ,   9   ,   61  ,   29  ,   253 ,   371 ,   527 ,   1697    ,   571 ,   2865    ,   15967   ,0 };
            const unsigned long dim1426JoeKuoD6Init[]   =   {   1   ,   1   ,   3   ,   1   ,   13  ,   33  ,   55  ,   67  ,   441 ,   451 ,   379 ,   1365    ,   2111    ,   6169    ,0 };
            const unsigned long dim1427JoeKuoD6Init[]   =   {   1   ,   1   ,   7   ,   1   ,   23  ,   39  ,   127 ,   43  ,   205 ,   391 ,   1463    ,   2227    ,   6877    ,   11483   ,0 };
            const unsigned long dim1428JoeKuoD6Init[]   =   {   1   ,   3   ,   3   ,   9   ,   31  ,   41  ,   5   ,   65  ,   245 ,   405 ,   633 ,   1379    ,   4111    ,   13129   ,0 };
            const unsigned long dim1429JoeKuoD6Init[]   =   {   1   ,   3   ,   5   ,   1   ,   1   ,   3   ,   57  ,   179 ,   23  ,   615 ,   761 ,   2339    ,   8191    ,   1995    ,0 };
            const unsigned long dim1430JoeKuoD6Init[]   =   {   1   ,   3   ,   3   ,   5   ,   11  ,   53  ,   67  ,   119 ,   169 ,   313 ,   873 ,   531 ,   63  ,   10827   ,0 };
            const unsigned long dim1431JoeKuoD6Init[]   =   {   1   ,   1   ,   5   ,   7   ,   31  ,   3   ,   53  ,   139 ,   33  ,   639 ,   855 ,   43  ,   7895    ,   495 ,0 };
            const unsigned long dim1432JoeKuoD6Init[]   =   {   1   ,   3   ,   3   ,   15  ,   15  ,   15  ,   13  ,   237 ,   427 ,   257 ,   1973    ,   2287    ,   7597    ,   13397   ,0 };
            const unsigned long dim1433JoeKuoD6Init[]   =   {   1   ,   1   ,   3   ,   11  ,   31  ,   5   ,   29  ,   97  ,   99  ,   1003    ,   333 ,   773 ,   4457    ,   8629    ,0 };
            const unsigned long dim1434JoeKuoD6Init[]   =   {   1   ,   1   ,   5   ,   9   ,   11  ,   55  ,   59  ,   201 ,   25  ,   559 ,   1647    ,   485 ,   7041    ,   4965    ,0 };
            const unsigned long dim1435JoeKuoD6Init[]   =   {   1   ,   1   ,   3   ,   5   ,   7   ,   37  ,   117 ,   159 ,   319 ,   431 ,   341 ,   2497    ,   755 ,   5363    ,0 };
            const unsigned long dim1436JoeKuoD6Init[]   =   {   1   ,   3   ,   5   ,   15  ,   25  ,   29  ,   99  ,   61  ,   5   ,   643 ,   1605    ,   2273    ,   3189    ,   5505    ,0 };
            const unsigned long dim1437JoeKuoD6Init[]   =   {   1   ,   1   ,   1   ,   1   ,   29  ,   47  ,   79  ,   177 ,   473 ,   361 ,   1337    ,   1199    ,   2395    ,   1079    ,0 };
            const unsigned long dim1438JoeKuoD6Init[]   =   {   1   ,   3   ,   5   ,   3   ,   7   ,   31  ,   101 ,   189 ,   223 ,   719 ,   1815    ,   1383    ,   7233    ,   11693   ,0 };
            const unsigned long dim1439JoeKuoD6Init[]   =   {   1   ,   1   ,   1   ,   5   ,   29  ,   39  ,   109 ,   211 ,   485 ,   659 ,   683 ,   3331    ,   4277    ,   809 ,0 };
            const unsigned long dim1440JoeKuoD6Init[]   =   {   1   ,   1   ,   3   ,   7   ,   23  ,   49  ,   37  ,   163 ,   127 ,   389 ,   367 ,   2223    ,   1175    ,   12717   ,0 };
            const unsigned long dim1441JoeKuoD6Init[]   =   {   1   ,   3   ,   3   ,   7   ,   27  ,   27  ,   51  ,   57  ,   115 ,   941 ,   243 ,   2741    ,   245 ,   341 ,0 };
            const unsigned long dim1442JoeKuoD6Init[]   =   {   1   ,   1   ,   5   ,   5   ,   15  ,   33  ,   113 ,   3   ,   267 ,   343 ,   563 ,   1769    ,   4095    ,   11557   ,0 };
            const unsigned long dim1443JoeKuoD6Init[]   =   {   1   ,   3   ,   3   ,   13  ,   13  ,   11  ,   105 ,   161 ,   145 ,   125 ,   1509    ,   1459    ,   843 ,   4553    ,0 };
            const unsigned long dim1444JoeKuoD6Init[]   =   {   1   ,   1   ,   1   ,   3   ,   11  ,   7   ,   67  ,   205 ,   335 ,   595 ,   1553    ,   2319    ,   6565    ,   14473   ,0 };
            const unsigned long dim1445JoeKuoD6Init[]   =   {   1   ,   3   ,   5   ,   15  ,   23  ,   13  ,   119 ,   117 ,   263 ,   463 ,   1919    ,   2089    ,   2809    ,   9957    ,0 };
            const unsigned long dim1446JoeKuoD6Init[]   =   {   1   ,   3   ,   7   ,   1   ,   27  ,   41  ,   5   ,   155 ,   31  ,   261 ,   1383    ,   1271    ,   2557    ,   6041    ,0 };
            const unsigned long dim1447JoeKuoD6Init[]   =   {   1   ,   3   ,   3   ,   13  ,   13  ,   43  ,   89  ,   45  ,   443 ,   13  ,   465 ,   703 ,   3405    ,   6471    ,0 };
            const unsigned long dim1448JoeKuoD6Init[]   =   {   1   ,   3   ,   1   ,   7   ,   31  ,   49  ,   49  ,   57  ,   97  ,   987 ,   187 ,   141 ,   7537    ,   11021   ,0 };
            const unsigned long dim1449JoeKuoD6Init[]   =   {   1   ,   1   ,   5   ,   15  ,   29  ,   43  ,   103 ,   89  ,   113 ,   729 ,   1523    ,   2893    ,   1143    ,   9049    ,0 };
            const unsigned long dim1450JoeKuoD6Init[]   =   {   1   ,   1   ,   3   ,   3   ,   29  ,   49  ,   43  ,   255 ,   371 ,   903 ,   661 ,   3355    ,   5981    ,   861 ,0 };
            const unsigned long dim1451JoeKuoD6Init[]   =   {   1   ,   3   ,   7   ,   11  ,   29  ,   13  ,   75  ,   5   ,   239 ,   983 ,   733 ,   2949    ,   4635    ,   6789    ,0 };
            const unsigned long dim1452JoeKuoD6Init[]   =   {   1   ,   1   ,   5   ,   3   ,   17  ,   51  ,   17  ,   49  ,   349 ,   73  ,   111 ,   897 ,   2757    ,   9817    ,0 };
            const unsigned long dim1453JoeKuoD6Init[]   =   {   1   ,   1   ,   1   ,   9   ,   17  ,   31  ,   73  ,   125 ,   207 ,   951 ,   681 ,   3685    ,   893 ,   13725   ,0 };
            const unsigned long dim1454JoeKuoD6Init[]   =   {   1   ,   1   ,   3   ,   1   ,   29  ,   13  ,   51  ,   181 ,   27  ,   807 ,   59  ,   3781    ,   263 ,   13865   ,0 };
            const unsigned long dim1455JoeKuoD6Init[]   =   {   1   ,   1   ,   7   ,   5   ,   15  ,   29  ,   55  ,   159 ,   341 ,   365 ,   1101    ,   3139    ,   4795    ,   4305    ,0 };
            const unsigned long dim1456JoeKuoD6Init[]   =   {   1   ,   3   ,   7   ,   15  ,   19  ,   35  ,   99  ,   251 ,   241 ,   115 ,   1197    ,   1167    ,   4663    ,   5523    ,0 };
            const unsigned long dim1457JoeKuoD6Init[]   =   {   1   ,   1   ,   7   ,   5   ,   13  ,   43  ,   59  ,   251 ,   455 ,   601 ,   891 ,   2279    ,   3447    ,   7787    ,0 };
            const unsigned long dim1458JoeKuoD6Init[]   =   {   1   ,   3   ,   3   ,   9   ,   31  ,   17  ,   97  ,   27  ,   373 ,   183 ,   7   ,   1729    ,   4513    ,   9763    ,0 };
            const unsigned long dim1459JoeKuoD6Init[]   =   {   1   ,   1   ,   3   ,   15  ,   29  ,   15  ,   57  ,   75  ,   211 ,   551 ,   141 ,   811 ,   27  ,   15107   ,0 };
            const unsigned long dim1460JoeKuoD6Init[]   =   {   1   ,   3   ,   1   ,   15  ,   31  ,   9   ,   57  ,   73  ,   431 ,   375 ,   1269    ,   2589    ,   1567    ,   9293    ,0 };
            const unsigned long dim1461JoeKuoD6Init[]   =   {   1   ,   1   ,   3   ,   13  ,   21  ,   13  ,   41  ,   211 ,   301 ,   177 ,   1667    ,   3957    ,   6261    ,   3275    ,0 };
            const unsigned long dim1462JoeKuoD6Init[]   =   {   1   ,   1   ,   5   ,   5   ,   15  ,   3   ,   33  ,   101 ,   119 ,   345 ,   1853    ,   2807    ,   337 ,   10163   ,0 };
            const unsigned long dim1463JoeKuoD6Init[]   =   {   1   ,   3   ,   3   ,   5   ,   27  ,   61  ,   11  ,   161 ,   25  ,   159 ,   1775    ,   3039    ,   3807    ,   9373    ,0 };
            const unsigned long dim1464JoeKuoD6Init[]   =   {   1   ,   1   ,   1   ,   15  ,   19  ,   35  ,   15  ,   191 ,   87  ,   735 ,   1079    ,   2373    ,   7087    ,   7505    ,0 };
            const unsigned long dim1465JoeKuoD6Init[]   =   {   1   ,   1   ,   5   ,   15  ,   29  ,   37  ,   43  ,   25  ,   161 ,   925 ,   259 ,   3281    ,   2643    ,   2359    ,0 };
            const unsigned long dim1466JoeKuoD6Init[]   =   {   1   ,   3   ,   1   ,   11  ,   3   ,   21  ,   13  ,   169 ,   17  ,   637 ,   1711    ,   1877    ,   6081    ,   5115    ,0 };
            const unsigned long dim1467JoeKuoD6Init[]   =   {   1   ,   3   ,   5   ,   1   ,   25  ,   5   ,   125 ,   103 ,   277 ,   533 ,   1485    ,   3263    ,   4899    ,   11851   ,0 };
            const unsigned long dim1468JoeKuoD6Init[]   =   {   1   ,   3   ,   7   ,   9   ,   23  ,   13  ,   61  ,   59  ,   405 ,   73  ,   817 ,   3831    ,   5059    ,   7671    ,0 };
            const unsigned long dim1469JoeKuoD6Init[]   =   {   1   ,   1   ,   7   ,   3   ,   1   ,   53  ,   113 ,   135 ,   481 ,   861 ,   1041    ,   931 ,   2973    ,   7957    ,0 };
            const unsigned long dim1470JoeKuoD6Init[]   =   {   1   ,   1   ,   7   ,   11  ,   9   ,   9   ,   123 ,   175 ,   65  ,   447 ,   1709    ,   4083    ,   6089    ,   7869    ,0 };
            const unsigned long dim1471JoeKuoD6Init[]   =   {   1   ,   3   ,   7   ,   7   ,   5   ,   11  ,   43  ,   59  ,   439 ,   539 ,   283 ,   3221    ,   2631    ,   13025   ,0 };
            const unsigned long dim1472JoeKuoD6Init[]   =   {   1   ,   3   ,   7   ,   7   ,   27  ,   1   ,   93  ,   51  ,   361 ,   441 ,   2005    ,   127 ,   6763    ,   12699   ,0 };
            const unsigned long dim1473JoeKuoD6Init[]   =   {   1   ,   3   ,   5   ,   15  ,   27  ,   29  ,   49  ,   49  ,   115 ,   167 ,   445 ,   2811    ,   1809    ,   11141   ,0 };
            const unsigned long dim1474JoeKuoD6Init[]   =   {   1   ,   1   ,   7   ,   5   ,   1   ,   57  ,   69  ,   49  ,   197 ,   459 ,   1613    ,   2891    ,   4437    ,   1621    ,0 };
            const unsigned long dim1475JoeKuoD6Init[]   =   {   1   ,   1   ,   1   ,   5   ,   25  ,   51  ,   33  ,   47  ,   115 ,   349 ,   1373    ,   1985    ,   5339    ,   4331    ,0 };
            const unsigned long dim1476JoeKuoD6Init[]   =   {   1   ,   3   ,   7   ,   11  ,   21  ,   55  ,   113 ,   183 ,   91  ,   955 ,   835 ,   2277    ,   5989    ,   1383    ,0 };
            const unsigned long dim1477JoeKuoD6Init[]   =   {   1   ,   3   ,   5   ,   1   ,   1   ,   11  ,   113 ,   121 ,   207 ,   897 ,   125 ,   393 ,   6281    ,   15003   ,0 };
            const unsigned long dim1478JoeKuoD6Init[]   =   {   1   ,   3   ,   7   ,   15  ,   13  ,   37  ,   89  ,   169 ,   437 ,   787 ,   1253    ,   2835    ,   4923    ,   15271   ,0 };
            const unsigned long dim1479JoeKuoD6Init[]   =   {   1   ,   3   ,   5   ,   9   ,   23  ,   37  ,   65  ,   135 ,   271 ,   381 ,   551 ,   3819    ,   5683    ,   3689    ,0 };
            const unsigned long dim1480JoeKuoD6Init[]   =   {   1   ,   1   ,   1   ,   9   ,   27  ,   29  ,   61  ,   147 ,   347 ,   745 ,   1793    ,   611 ,   7415    ,   4927    ,0 };
            const unsigned long dim1481JoeKuoD6Init[]   =   {   1   ,   1   ,   3   ,   5   ,   23  ,   59  ,   35  ,   5   ,   179 ,   247 ,   1265    ,   2503    ,   1705    ,   10011   ,0 };
            const unsigned long dim1482JoeKuoD6Init[]   =   {   1   ,   3   ,   5   ,   7   ,   11  ,   19  ,   57  ,   175 ,   93  ,   777 ,   353 ,   371 ,   6655    ,   10271   ,0 };
            const unsigned long dim1483JoeKuoD6Init[]   =   {   1   ,   3   ,   7   ,   13  ,   15  ,   51  ,   29  ,   213 ,   113 ,   143 ,   2011    ,   2223    ,   2099    ,   4103    ,0 };
            const unsigned long dim1484JoeKuoD6Init[]   =   {   1   ,   1   ,   7   ,   13  ,   19  ,   31  ,   75  ,   99  ,   89  ,   219 ,   2019    ,   1739    ,   3813    ,   5619    ,0 };
            const unsigned long dim1485JoeKuoD6Init[]   =   {   1   ,   1   ,   3   ,   7   ,   27  ,   53  ,   69  ,   215 ,   299 ,   401 ,   719 ,   3123    ,   3463    ,   13967   ,0 };
            const unsigned long dim1486JoeKuoD6Init[]   =   {   1   ,   3   ,   5   ,   3   ,   29  ,   63  ,   5   ,   247 ,   21  ,   651 ,   879 ,   3033    ,   117 ,   11511   ,0 };
            const unsigned long dim1487JoeKuoD6Init[]   =   {   1   ,   1   ,   5   ,   1   ,   5   ,   47  ,   63  ,   45  ,   145 ,   905 ,   885 ,   2367    ,   6721    ,   3869    ,0 };
            const unsigned long dim1488JoeKuoD6Init[]   =   {   1   ,   1   ,   3   ,   3   ,   25  ,   13  ,   117 ,   17  ,   367 ,   425 ,   367 ,   2863    ,   3087    ,   9115    ,0 };
            const unsigned long dim1489JoeKuoD6Init[]   =   {   1   ,   3   ,   5   ,   1   ,   1   ,   31  ,   13  ,   65  ,   255 ,   873 ,   831 ,   3439    ,   7459    ,   7713    ,0 };
            const unsigned long dim1490JoeKuoD6Init[]   =   {   1   ,   1   ,   7   ,   11  ,   19  ,   53  ,   93  ,   33  ,   505 ,   1007    ,   1331    ,   2317    ,   2817    ,   2549    ,0 };
            const unsigned long dim1491JoeKuoD6Init[]   =   {   1   ,   1   ,   1   ,   11  ,   9   ,   1   ,   67  ,   109 ,   397 ,   831 ,   1371    ,   3277    ,   151 ,   6523    ,0 };
            const unsigned long dim1492JoeKuoD6Init[]   =   {   1   ,   3   ,   1   ,   3   ,   25  ,   11  ,   35  ,   233 ,   231 ,   89  ,   1029    ,   1053    ,   6579    ,   455 ,0 };
            const unsigned long dim1493JoeKuoD6Init[]   =   {   1   ,   1   ,   5   ,   13  ,   11  ,   7   ,   117 ,   203 ,   125 ,   863 ,   1927    ,   2081    ,   1229    ,   6373    ,0 };
            const unsigned long dim1494JoeKuoD6Init[]   =   {   1   ,   1   ,   1   ,   7   ,   23  ,   53  ,   27  ,   207 ,   189 ,   287 ,   635 ,   3483    ,   4551    ,   15085   ,0 };
            const unsigned long dim1495JoeKuoD6Init[]   =   {   1   ,   1   ,   3   ,   3   ,   23  ,   45  ,   99  ,   177 ,   59  ,   835 ,   1123    ,   3397    ,   4237    ,   11895   ,0 };
            const unsigned long dim1496JoeKuoD6Init[]   =   {   1   ,   1   ,   7   ,   3   ,   31  ,   59  ,   37  ,   7   ,   387 ,   849 ,   1857    ,   3469    ,   3663    ,   1465    ,0 };
            const unsigned long dim1497JoeKuoD6Init[]   =   {   1   ,   1   ,   5   ,   9   ,   13  ,   7   ,   73  ,   211 ,   83  ,   831 ,   1069    ,   3103    ,   2355    ,   8817    ,0 };
            const unsigned long dim1498JoeKuoD6Init[]   =   {   1   ,   3   ,   7   ,   1   ,   1   ,   7   ,   63  ,   195 ,   69  ,   945 ,   39  ,   1417    ,   879 ,   9223    ,0 };
            const unsigned long dim1499JoeKuoD6Init[]   =   {   1   ,   1   ,   7   ,   9   ,   3   ,   53  ,   111 ,   41  ,   29  ,   597 ,   1667    ,   3903    ,   7401    ,   6939    ,0 };
            const unsigned long dim1500JoeKuoD6Init[]   =   {   1   ,   3   ,   5   ,   7   ,   9   ,   35  ,   55  ,   61  ,   509 ,   729 ,   25  ,   2357    ,   3607    ,   2215    ,0 };
            const unsigned long dim1501JoeKuoD6Init[]   =   {   1   ,   3   ,   7   ,   7   ,   19  ,   49  ,   47  ,   241 ,   79  ,   363 ,   1987    ,   3269    ,   5161    ,   1379    ,0 };
            const unsigned long dim1502JoeKuoD6Init[]   =   {   1   ,   3   ,   5   ,   5   ,   1   ,   33  ,   21  ,   249 ,   321 ,   281 ,   2025    ,   425 ,   4955    ,   5799    ,0 };
            const unsigned long dim1503JoeKuoD6Init[]   =   {   1   ,   1   ,   7   ,   5   ,   19  ,   21  ,   93  ,   69  ,   487 ,   619 ,   839 ,   1359    ,   3219    ,   8303    ,0 };
            const unsigned long dim1504JoeKuoD6Init[]   =   {   1   ,   3   ,   5   ,   15  ,   27  ,   33  ,   3   ,   55  ,   149 ,   361 ,   887 ,   1341    ,   1219    ,   4769    ,0 };
            const unsigned long dim1505JoeKuoD6Init[]   =   {   1   ,   1   ,   1   ,   15  ,   5   ,   13  ,   65  ,   53  ,   181 ,   923 ,   1643    ,   3853    ,   3067    ,   4281    ,0 };
            const unsigned long dim1506JoeKuoD6Init[]   =   {   1   ,   1   ,   1   ,   1   ,   13  ,   5   ,   23  ,   27  ,   211 ,   377 ,   1421    ,   3085    ,   509 ,   509 ,0 };
            const unsigned long dim1507JoeKuoD6Init[]   =   {   1   ,   1   ,   7   ,   9   ,   21  ,   13  ,   3   ,   193 ,   19  ,   497 ,   1989    ,   885 ,   1157    ,   7213    ,0 };
            const unsigned long dim1508JoeKuoD6Init[]   =   {   1   ,   3   ,   7   ,   11  ,   1   ,   5   ,   105 ,   57  ,   245 ,   25  ,   153 ,   795 ,   3781    ,   11735   ,0 };
            const unsigned long dim1509JoeKuoD6Init[]   =   {   1   ,   1   ,   1   ,   13  ,   5   ,   7   ,   87  ,   63  ,   81  ,   491 ,   971 ,   805 ,   2349    ,   15627   ,0 };
            const unsigned long dim1510JoeKuoD6Init[]   =   {   1   ,   1   ,   5   ,   1   ,   27  ,   41  ,   31  ,   165 ,   383 ,   85  ,   1713    ,   2427    ,   5329    ,   15169   ,0 };
            const unsigned long dim1511JoeKuoD6Init[]   =   {   1   ,   1   ,   5   ,   7   ,   9   ,   47  ,   75  ,   51  ,   329 ,   651 ,   481 ,   2469    ,   1761    ,   1367    ,0 };
            const unsigned long dim1512JoeKuoD6Init[]   =   {   1   ,   1   ,   1   ,   11  ,   13  ,   51  ,   89  ,   163 ,   431 ,   117 ,   1909    ,   2685    ,   7197    ,   1537    ,0 };
            const unsigned long dim1513JoeKuoD6Init[]   =   {   1   ,   1   ,   1   ,   15  ,   29  ,   47  ,   19  ,   231 ,   35  ,   597 ,   997 ,   105 ,   7701    ,   899 ,0 };
            const unsigned long dim1514JoeKuoD6Init[]   =   {   1   ,   3   ,   7   ,   15  ,   5   ,   19  ,   63  ,   87  ,   49  ,   685 ,   525 ,   3709    ,   4461    ,   15697   ,0 };
            const unsigned long dim1515JoeKuoD6Init[]   =   {   1   ,   3   ,   1   ,   13  ,   31  ,   63  ,   65  ,   91  ,   169 ,   857 ,   1143    ,   2831    ,   7307    ,   11755   ,0 };
            const unsigned long dim1516JoeKuoD6Init[]   =   {   1   ,   3   ,   1   ,   13  ,   27  ,   61  ,   105 ,   191 ,   325 ,   71  ,   1155    ,   2291    ,   8023    ,   1281    ,0 };
            const unsigned long dim1517JoeKuoD6Init[]   =   {   1   ,   1   ,   3   ,   11  ,   21  ,   45  ,   45  ,   19  ,   453 ,   527 ,   75  ,   3489    ,   3011    ,   12687   ,0 };
            const unsigned long dim1518JoeKuoD6Init[]   =   {   1   ,   1   ,   5   ,   7   ,   23  ,   15  ,   27  ,   105 ,   335 ,   549 ,   737 ,   3671    ,   5071    ,   10957   ,0 };
            const unsigned long dim1519JoeKuoD6Init[]   =   {   1   ,   1   ,   7   ,   9   ,   9   ,   13  ,   121 ,   169 ,   87  ,   871 ,   917 ,   3071    ,   7817    ,   14493   ,0 };
            const unsigned long dim1520JoeKuoD6Init[]   =   {   1   ,   1   ,   3   ,   9   ,   3   ,   1   ,   63  ,   195 ,   67  ,   123 ,   175 ,   3779    ,   2681    ,   15967   ,0 };
            const unsigned long dim1521JoeKuoD6Init[]   =   {   1   ,   3   ,   7   ,   9   ,   11  ,   31  ,   111 ,   147 ,   203 ,   205 ,   1591    ,   3003    ,   631 ,   15567   ,0 };
            const unsigned long dim1522JoeKuoD6Init[]   =   {   1   ,   1   ,   3   ,   5   ,   23  ,   11  ,   43  ,   117 ,   37  ,   413 ,   1657    ,   4061    ,   5491    ,   10329   ,0 };
            const unsigned long dim1523JoeKuoD6Init[]   =   {   1   ,   1   ,   1   ,   11  ,   31  ,   35  ,   107 ,   105 ,   107 ,   117 ,   2041    ,   1415    ,   2153    ,   9063    ,0 };
            const unsigned long dim1524JoeKuoD6Init[]   =   {   1   ,   3   ,   5   ,   13  ,   9   ,   35  ,   77  ,   1   ,   343 ,   341 ,   1841    ,   2963    ,   1105    ,   4103    ,0 };
            const unsigned long dim1525JoeKuoD6Init[]   =   {   1   ,   1   ,   7   ,   3   ,   25  ,   21  ,   45  ,   63  ,   429 ,   843 ,   585 ,   741 ,   6641    ,   6327    ,0 };
            const unsigned long dim1526JoeKuoD6Init[]   =   {   1   ,   1   ,   1   ,   5   ,   19  ,   23  ,   3   ,   63  ,   273 ,   1009    ,   131 ,   3541    ,   3769    ,   13045   ,0 };
            const unsigned long dim1527JoeKuoD6Init[]   =   {   1   ,   3   ,   5   ,   15  ,   11  ,   41  ,   127 ,   139 ,   105 ,   15  ,   1311    ,   2995    ,   3151    ,   12913   ,0 };
            const unsigned long dim1528JoeKuoD6Init[]   =   {   1   ,   3   ,   1   ,   3   ,   1   ,   37  ,   71  ,   55  ,   17  ,   155 ,   397 ,   121 ,   2917    ,   3943    ,0 };
            const unsigned long dim1529JoeKuoD6Init[]   =   {   1   ,   3   ,   7   ,   11  ,   5   ,   1   ,   23  ,   29  ,   233 ,   947 ,   893 ,   269 ,   6487    ,   13831   ,0 };
            const unsigned long dim1530JoeKuoD6Init[]   =   {   1   ,   3   ,   1   ,   5   ,   7   ,   17  ,   63  ,   177 ,   153 ,   595 ,   1217    ,   2225    ,   5973    ,   15853   ,0 };
            const unsigned long dim1531JoeKuoD6Init[]   =   {   1   ,   3   ,   7   ,   9   ,   19  ,   9   ,   45  ,   197 ,   285 ,   85  ,   1533    ,   673 ,   5969    ,   501 ,0 };
            const unsigned long dim1532JoeKuoD6Init[]   =   {   1   ,   3   ,   5   ,   15  ,   15  ,   23  ,   123 ,   151 ,   49  ,   583 ,   385 ,   3671    ,   3617    ,   16285   ,0 };
            const unsigned long dim1533JoeKuoD6Init[]   =   {   1   ,   3   ,   7   ,   13  ,   11  ,   55  ,   111 ,   57  ,   51  ,   249 ,   13  ,   1929    ,   4711    ,   9847    ,0 };
            const unsigned long dim1534JoeKuoD6Init[]   =   {   1   ,   1   ,   1   ,   3   ,   17  ,   13  ,   101 ,   117 ,   433 ,   327 ,   151 ,   2137    ,   4885    ,   12775   ,0 };
            const unsigned long dim1535JoeKuoD6Init[]   =   {   1   ,   3   ,   3   ,   11  ,   13  ,   39  ,   113 ,   35  ,   99  ,   985 ,   765 ,   1239    ,   1311    ,   4591    ,0 };
            const unsigned long dim1536JoeKuoD6Init[]   =   {   1   ,   1   ,   1   ,   11  ,   11  ,   31  ,   41  ,   205 ,   339 ,   773 ,   1291    ,   1665    ,   4345    ,   4139    ,0 };
            const unsigned long dim1537JoeKuoD6Init[]   =   {   1   ,   1   ,   1   ,   1   ,   27  ,   41  ,   27  ,   139 ,   109 ,   555 ,   691 ,   2263    ,   5321    ,   12865   ,0 };
            const unsigned long dim1538JoeKuoD6Init[]   =   {   1   ,   1   ,   3   ,   11  ,   23  ,   13  ,   41  ,   73  ,   253 ,   481 ,   1345    ,   3339    ,   1121    ,   665 ,0 };
            const unsigned long dim1539JoeKuoD6Init[]   =   {   1   ,   1   ,   7   ,   7   ,   23  ,   57  ,   109 ,   161 ,   35  ,   911 ,   411 ,   3117    ,   5997    ,   6843    ,0 };
            const unsigned long dim1540JoeKuoD6Init[]   =   {   1   ,   3   ,   7   ,   13  ,   3   ,   39  ,   97  ,   85  ,   399 ,   733 ,   43  ,   2437    ,   2755    ,   10693   ,0 };
            const unsigned long dim1541JoeKuoD6Init[]   =   {   1   ,   3   ,   5   ,   15  ,   15  ,   47  ,   121 ,   157 ,   379 ,   805 ,   1483    ,   4013    ,   79  ,   10943   ,0 };
            const unsigned long dim1542JoeKuoD6Init[]   =   {   1   ,   1   ,   3   ,   3   ,   23  ,   51  ,   39  ,   61  ,   157 ,   139 ,   273 ,   2385    ,   2451    ,   12437   ,0 };
            const unsigned long dim1543JoeKuoD6Init[]   =   {   1   ,   1   ,   5   ,   1   ,   21  ,   7   ,   77  ,   197 ,   469 ,   391 ,   1665    ,   1241    ,   1847    ,   14129   ,0 };
            const unsigned long dim1544JoeKuoD6Init[]   =   {   1   ,   1   ,   5   ,   5   ,   27  ,   23  ,   83  ,   245 ,   319 ,   247 ,   1345    ,   1229    ,   1119    ,   5749    ,0 };
            const unsigned long dim1545JoeKuoD6Init[]   =   {   1   ,   1   ,   5   ,   3   ,   31  ,   63  ,   121 ,   245 ,   371 ,   359 ,   1929    ,   817 ,   2603    ,   5135    ,0 };
            const unsigned long dim1546JoeKuoD6Init[]   =   {   1   ,   3   ,   1   ,   15  ,   21  ,   17  ,   15  ,   11  ,   425 ,   115 ,   1773    ,   1491    ,   2295    ,   11883   ,0 };
            const unsigned long dim1547JoeKuoD6Init[]   =   {   1   ,   1   ,   5   ,   1   ,   7   ,   21  ,   91  ,   203 ,   435 ,   119 ,   819 ,   707 ,   3567    ,   517 ,0 };
            const unsigned long dim1548JoeKuoD6Init[]   =   {   1   ,   1   ,   5   ,   3   ,   7   ,   7   ,   101 ,   41  ,   475 ,   539 ,   1059    ,   2245    ,   6731    ,   6635    ,0 };
            const unsigned long dim1549JoeKuoD6Init[]   =   {   1   ,   3   ,   5   ,   1   ,   11  ,   25  ,   91  ,   89  ,   193 ,   463 ,   1667    ,   661 ,   801 ,   7087    ,0 };
            const unsigned long dim1550JoeKuoD6Init[]   =   {   1   ,   3   ,   7   ,   15  ,   1   ,   43  ,   17  ,   97  ,   129 ,   843 ,   1319    ,   723 ,   2725    ,   10743   ,0 };
            const unsigned long dim1551JoeKuoD6Init[]   =   {   1   ,   1   ,   3   ,   11  ,   19  ,   27  ,   119 ,   113 ,   85  ,   385 ,   1841    ,   2821    ,   2867    ,   6199    ,0 };
            const unsigned long dim1552JoeKuoD6Init[]   =   {   1   ,   1   ,   7   ,   1   ,   23  ,   27  ,   95  ,   175 ,   99  ,   839 ,   1039    ,   735 ,   6355    ,   1651    ,0 };
            const unsigned long dim1553JoeKuoD6Init[]   =   {   1   ,   3   ,   3   ,   1   ,   11  ,   13  ,   71  ,   91  ,   7   ,   959 ,   979 ,   2571    ,   117 ,   11479   ,0 };
            const unsigned long dim1554JoeKuoD6Init[]   =   {   1   ,   1   ,   7   ,   11  ,   21  ,   45  ,   113 ,   205 ,   107 ,   881 ,   2039    ,   3181    ,   4125    ,   13801   ,0 };
            const unsigned long dim1555JoeKuoD6Init[]   =   {   1   ,   1   ,   1   ,   13  ,   3   ,   63  ,   123 ,   53  ,   427 ,   469 ,   685 ,   3685    ,   1019    ,   7069    ,0 };
            const unsigned long dim1556JoeKuoD6Init[]   =   {   1   ,   1   ,   1   ,   13  ,   31  ,   33  ,   39  ,   155 ,   19  ,   975 ,   659 ,   11  ,   6873    ,   10105   ,0 };
            const unsigned long dim1557JoeKuoD6Init[]   =   {   1   ,   3   ,   5   ,   3   ,   31  ,   33  ,   91  ,   3   ,   207 ,   63  ,   559 ,   3745    ,   637 ,   6869    ,0 };
            const unsigned long dim1558JoeKuoD6Init[]   =   {   1   ,   3   ,   1   ,   11  ,   1   ,   11  ,   41  ,   51  ,   439 ,   273 ,   1977    ,   2987    ,   5265    ,   11779   ,0 };
            const unsigned long dim1559JoeKuoD6Init[]   =   {   1   ,   1   ,   5   ,   7   ,   13  ,   63  ,   111 ,   125 ,   65  ,   1003    ,   2045    ,   1823    ,   6799    ,   2847    ,0 };
            const unsigned long dim1560JoeKuoD6Init[]   =   {   1   ,   1   ,   1   ,   1   ,   11  ,   9   ,   59  ,   187 ,   431 ,   773 ,   1817    ,   2891    ,   2927    ,   1145    ,0 };
            const unsigned long dim1561JoeKuoD6Init[]   =   {   1   ,   1   ,   1   ,   15  ,   25  ,   19  ,   101 ,   231 ,   371 ,   857 ,   261 ,   1901    ,   2165    ,   7275    ,0 };
            const unsigned long dim1562JoeKuoD6Init[]   =   {   1   ,   3   ,   1   ,   15  ,   31  ,   57  ,   125 ,   77  ,   447 ,   417 ,   1503    ,   3003    ,   7229    ,   14275   ,0 };
            const unsigned long dim1563JoeKuoD6Init[]   =   {   1   ,   1   ,   1   ,   11  ,   23  ,   63  ,   17  ,   251 ,   237 ,   755 ,   641 ,   1875    ,   1537    ,   2465    ,0 };
            const unsigned long dim1564JoeKuoD6Init[]   =   {   1   ,   3   ,   7   ,   3   ,   25  ,   59  ,   45  ,   97  ,   173 ,   179 ,   421 ,   1269    ,   2933    ,   337 ,0 };
            const unsigned long dim1565JoeKuoD6Init[]   =   {   1   ,   3   ,   5   ,   5   ,   15  ,   25  ,   81  ,   191 ,   475 ,   723 ,   107 ,   277 ,   681 ,   15147   ,0 };
            const unsigned long dim1566JoeKuoD6Init[]   =   {   1   ,   1   ,   5   ,   15  ,   3   ,   17  ,   65  ,   39  ,   281 ,   831 ,   85  ,   1445    ,   4779    ,   5559    ,0 };
            const unsigned long dim1567JoeKuoD6Init[]   =   {   1   ,   3   ,   1   ,   13  ,   21  ,   1   ,   7   ,   5   ,   141 ,   65  ,   745 ,   1389    ,   1935    ,   9317    ,0 };
            const unsigned long dim1568JoeKuoD6Init[]   =   {   1   ,   1   ,   5   ,   11  ,   27  ,   31  ,   33  ,   27  ,   185 ,   625 ,   1723    ,   3499    ,   6225    ,   7993    ,0 };
            const unsigned long dim1569JoeKuoD6Init[]   =   {   1   ,   1   ,   5   ,   7   ,   5   ,   51  ,   19  ,   27  ,   301 ,   845 ,   949 ,   1963    ,   7201    ,   5399    ,0 };
            const unsigned long dim1570JoeKuoD6Init[]   =   {   1   ,   3   ,   3   ,   9   ,   1   ,   59  ,   113 ,   87  ,   237 ,   83  ,   815 ,   161 ,   1685    ,   1155    ,0 };
            const unsigned long dim1571JoeKuoD6Init[]   =   {   1   ,   3   ,   3   ,   15  ,   3   ,   47  ,   105 ,   77  ,   343 ,   997 ,   647 ,   3051    ,   4877    ,   11519   ,0 };
            const unsigned long dim1572JoeKuoD6Init[]   =   {   1   ,   1   ,   5   ,   7   ,   15  ,   13  ,   123 ,   253 ,   343 ,   701 ,   1231    ,   3997    ,   3203    ,   7411    ,0 };
            const unsigned long dim1573JoeKuoD6Init[]   =   {   1   ,   3   ,   3   ,   7   ,   21  ,   53  ,   103 ,   49  ,   511 ,   535 ,   1159    ,   747 ,   4859    ,   2733    ,0 };
            const unsigned long dim1574JoeKuoD6Init[]   =   {   1   ,   3   ,   3   ,   13  ,   19  ,   13  ,   113 ,   93  ,   61  ,   1007    ,   219 ,   3817    ,   4819    ,   9959    ,0 };
            const unsigned long dim1575JoeKuoD6Init[]   =   {   1   ,   3   ,   1   ,   7   ,   1   ,   57  ,   45  ,   167 ,   209 ,   757 ,   1861    ,   1863    ,   3495    ,   4295    ,0 };
            const unsigned long dim1576JoeKuoD6Init[]   =   {   1   ,   1   ,   3   ,   13  ,   25  ,   53  ,   63  ,   187 ,   27  ,   663 ,   1697    ,   2477    ,   1685    ,   6609    ,0 };
            const unsigned long dim1577JoeKuoD6Init[]   =   {   1   ,   3   ,   7   ,   1   ,   19  ,   51  ,   91  ,   125 ,   275 ,   737 ,   61  ,   685 ,   5739    ,   12795   ,0 };
            const unsigned long dim1578JoeKuoD6Init[]   =   {   1   ,   1   ,   7   ,   15  ,   1   ,   63  ,   21  ,   253 ,   247 ,   399 ,   1207    ,   3357    ,   1029    ,   4719    ,0 };
            const unsigned long dim1579JoeKuoD6Init[]   =   {   1   ,   1   ,   5   ,   13  ,   21  ,   1   ,   45  ,   223 ,   511 ,   963 ,   1099    ,   2913    ,   7385    ,   13903   ,0 };
            const unsigned long dim1580JoeKuoD6Init[]   =   {   1   ,   3   ,   7   ,   7   ,   1   ,   33  ,   123 ,   81  ,   431 ,   947 ,   1821    ,   3763    ,   1679    ,   2391    ,0 };
            const unsigned long dim1581JoeKuoD6Init[]   =   {   1   ,   1   ,   5   ,   1   ,   9   ,   9   ,   9   ,   55  ,   425 ,   539 ,   221 ,   409 ,   4967    ,   8239    ,0 };
            const unsigned long dim1582JoeKuoD6Init[]   =   {   1   ,   3   ,   7   ,   11  ,   19  ,   5   ,   97  ,   77  ,   95  ,   955 ,   705 ,   3627    ,   3743    ,   8687    ,0 };
            const unsigned long dim1583JoeKuoD6Init[]   =   {   1   ,   3   ,   1   ,   11  ,   27  ,   51  ,   37  ,   177 ,   109 ,   343 ,   35  ,   1361    ,   1557    ,   13487   ,0 };
            const unsigned long dim1584JoeKuoD6Init[]   =   {   1   ,   1   ,   1   ,   11  ,   19  ,   25  ,   93  ,   155 ,   421 ,   765 ,   1529    ,   2685    ,   2729    ,   4763    ,0 };
            const unsigned long dim1585JoeKuoD6Init[]   =   {   1   ,   3   ,   3   ,   11  ,   7   ,   21  ,   99  ,   179 ,   205 ,   473 ,   1435    ,   711 ,   6975    ,   10497   ,0 };
            const unsigned long dim1586JoeKuoD6Init[]   =   {   1   ,   1   ,   5   ,   3   ,   1   ,   55  ,   29  ,   113 ,   201 ,   123 ,   201 ,   2749    ,   75  ,   5839    ,0 };
            const unsigned long dim1587JoeKuoD6Init[]   =   {   1   ,   3   ,   1   ,   9   ,   13  ,   9   ,   19  ,   55  ,   57  ,   99  ,   811 ,   3143    ,   5143    ,   5149    ,0 };
            const unsigned long dim1588JoeKuoD6Init[]   =   {   1   ,   3   ,   5   ,   1   ,   5   ,   49  ,   89  ,   229 ,   351 ,   93  ,   951 ,   3587    ,   6819    ,   119 ,0 };
            const unsigned long dim1589JoeKuoD6Init[]   =   {   1   ,   1   ,   5   ,   5   ,   17  ,   1   ,   95  ,   129 ,   79  ,   165 ,   209 ,   2859    ,   4269    ,   7613    ,0 };
            const unsigned long dim1590JoeKuoD6Init[]   =   {   1   ,   3   ,   5   ,   11  ,   3   ,   1   ,   115 ,   215 ,   215 ,   449 ,   1365    ,   3429    ,   5711    ,   10053   ,0 };
            const unsigned long dim1591JoeKuoD6Init[]   =   {   1   ,   1   ,   1   ,   9   ,   17  ,   59  ,   11  ,   121 ,   229 ,   37  ,   781 ,   421 ,   145 ,   1359    ,0 };
            const unsigned long dim1592JoeKuoD6Init[]   =   {   1   ,   1   ,   3   ,   9   ,   31  ,   31  ,   73  ,   255 ,   451 ,   337 ,   617 ,   3169    ,   7275    ,   4783    ,0 };
            const unsigned long dim1593JoeKuoD6Init[]   =   {   1   ,   1   ,   7   ,   15  ,   3   ,   31  ,   17  ,   163 ,   243 ,   695 ,   1073    ,   2625    ,   6237    ,   4011    ,0 };
            const unsigned long dim1594JoeKuoD6Init[]   =   {   1   ,   1   ,   5   ,   5   ,   27  ,   17  ,   45  ,   167 ,   363 ,   741 ,   1873    ,   3699    ,   7577    ,   7389    ,0 };
            const unsigned long dim1595JoeKuoD6Init[]   =   {   1   ,   1   ,   7   ,   9   ,   5   ,   61  ,   43  ,   99  ,   175 ,   645 ,   1351    ,   1405    ,   2443    ,   15481   ,0 };
            const unsigned long dim1596JoeKuoD6Init[]   =   {   1   ,   3   ,   7   ,   11  ,   7   ,   19  ,   11  ,   27  ,   481 ,   271 ,   881 ,   143 ,   1087    ,   7443    ,0 };
            const unsigned long dim1597JoeKuoD6Init[]   =   {   1   ,   1   ,   3   ,   15  ,   11  ,   59  ,   47  ,   165 ,   421 ,   211 ,   451 ,   3957    ,   1715    ,   5045    ,0 };
            const unsigned long dim1598JoeKuoD6Init[]   =   {   1   ,   1   ,   1   ,   3   ,   1   ,   55  ,   121 ,   71  ,   483 ,   79  ,   61  ,   3107    ,   6367    ,   8787    ,0 };
            const unsigned long dim1599JoeKuoD6Init[]   =   {   1   ,   1   ,   5   ,   11  ,   5   ,   35  ,   7   ,   197 ,   269 ,   437 ,   1529    ,   1757    ,   8165    ,   10897   ,0 };
            const unsigned long dim1600JoeKuoD6Init[]   =   {   1   ,   1   ,   3   ,   13  ,   25  ,   41  ,   121 ,   115 ,   177 ,   435 ,   839 ,   2701    ,   2813    ,   5055    ,0 };
            const unsigned long dim1601JoeKuoD6Init[]   =   {   1   ,   1   ,   3   ,   15  ,   5   ,   57  ,   17  ,   99  ,   31  ,   415 ,   1391    ,   3077    ,   4317    ,   12963   ,0 };
            const unsigned long dim1602JoeKuoD6Init[]   =   {   1   ,   3   ,   1   ,   7   ,   31  ,   45  ,   95  ,   97  ,   43  ,   601 ,   2033    ,   2967    ,   3985    ,   10845   ,0 };
            const unsigned long dim1603JoeKuoD6Init[]   =   {   1   ,   3   ,   1   ,   15  ,   19  ,   55  ,   99  ,   125 ,   391 ,   339 ,   1277    ,   2669    ,   359 ,   3983    ,0 };
            const unsigned long dim1604JoeKuoD6Init[]   =   {   1   ,   3   ,   7   ,   11  ,   11  ,   47  ,   79  ,   197 ,   157 ,   959 ,   777 ,   133 ,   337 ,   14891   ,0 };
            const unsigned long dim1605JoeKuoD6Init[]   =   {   1   ,   1   ,   5   ,   7   ,   9   ,   3   ,   95  ,   77  ,   217 ,   399 ,   1269    ,   2505    ,   1055    ,   5483    ,0 };
            const unsigned long dim1606JoeKuoD6Init[]   =   {   1   ,   1   ,   5   ,   3   ,   27  ,   5   ,   3   ,   109 ,   481 ,   169 ,   1787    ,   725 ,   2827    ,   13599   ,0 };
            const unsigned long dim1607JoeKuoD6Init[]   =   {   1   ,   3   ,   1   ,   5   ,   7   ,   11  ,   127 ,   119 ,   141 ,   999 ,   295 ,   2727    ,   7329    ,   11371   ,0 };
            const unsigned long dim1608JoeKuoD6Init[]   =   {   1   ,   3   ,   3   ,   7   ,   31  ,   17  ,   105 ,   139 ,   287 ,   333 ,   1149    ,   3373    ,   5165    ,   9555    ,0 };
            const unsigned long dim1609JoeKuoD6Init[]   =   {   1   ,   3   ,   5   ,   5   ,   5   ,   13  ,   53  ,   25  ,   125 ,   953 ,   511 ,   3695    ,   7935    ,   10901   ,0 };
            const unsigned long dim1610JoeKuoD6Init[]   =   {   1   ,   1   ,   5   ,   15  ,   5   ,   37  ,   73  ,   29  ,   471 ,   423 ,   2025    ,   3759    ,   4271    ,   9589    ,0 };
            const unsigned long dim1611JoeKuoD6Init[]   =   {   1   ,   1   ,   1   ,   5   ,   9   ,   35  ,   105 ,   207 ,   245 ,   935 ,   801 ,   2115    ,   7935    ,   11439   ,0 };
            const unsigned long dim1612JoeKuoD6Init[]   =   {   1   ,   3   ,   7   ,   15  ,   23  ,   39  ,   67  ,   187 ,   333 ,   823 ,   1015    ,   1463    ,   6445    ,   5079    ,0 };
            const unsigned long dim1613JoeKuoD6Init[]   =   {   1   ,   1   ,   7   ,   7   ,   15  ,   43  ,   41  ,   243 ,   493 ,   703 ,   3   ,   3585    ,   6775    ,   2169    ,0 };
            const unsigned long dim1614JoeKuoD6Init[]   =   {   1   ,   3   ,   7   ,   9   ,   7   ,   53  ,   29  ,   225 ,   399 ,   885 ,   1051    ,   665 ,   1447    ,   5913    ,0 };
            const unsigned long dim1615JoeKuoD6Init[]   =   {   1   ,   1   ,   1   ,   7   ,   3   ,   19  ,   47  ,   219 ,   203 ,   903 ,   1399    ,   1793    ,   5949    ,   4311    ,0 };
            const unsigned long dim1616JoeKuoD6Init[]   =   {   1   ,   1   ,   1   ,   5   ,   9   ,   9   ,   117 ,   207 ,   107 ,   651 ,   1155    ,   2047    ,   7489    ,   5959    ,0 };
            const unsigned long dim1617JoeKuoD6Init[]   =   {   1   ,   3   ,   7   ,   9   ,   31  ,   59  ,   105 ,   73  ,   251 ,   209 ,   515 ,   3467    ,   721 ,   13423   ,0 };
            const unsigned long dim1618JoeKuoD6Init[]   =   {   1   ,   1   ,   7   ,   13  ,   13  ,   45  ,   75  ,   171 ,   61  ,   249 ,   1641    ,   289 ,   201 ,   6185    ,0 };
            const unsigned long dim1619JoeKuoD6Init[]   =   {   1   ,   1   ,   1   ,   11  ,   13  ,   37  ,   91  ,   225 ,   241 ,   899 ,   1471    ,   3553    ,   6741    ,   11623   ,0 };
            const unsigned long dim1620JoeKuoD6Init[]   =   {   1   ,   3   ,   7   ,   15  ,   19  ,   61  ,   79  ,   81  ,   75  ,   343 ,   145 ,   3975    ,   4807    ,   4887    ,0 };
            const unsigned long dim1621JoeKuoD6Init[]   =   {   1   ,   1   ,   3   ,   9   ,   17  ,   15  ,   29  ,   1   ,   331 ,   649 ,   1455    ,   2365    ,   6309    ,   6701    ,0 };
            const unsigned long dim1622JoeKuoD6Init[]   =   {   1   ,   1   ,   1   ,   1   ,   15  ,   19  ,   37  ,   5   ,   417 ,   933 ,   269 ,   2391    ,   8063    ,   14231   ,0 };
            const unsigned long dim1623JoeKuoD6Init[]   =   {   1   ,   1   ,   7   ,   3   ,   17  ,   49  ,   21  ,   41  ,   127 ,   507 ,   1927    ,   1971    ,   7057    ,   12047   ,0 };
            const unsigned long dim1624JoeKuoD6Init[]   =   {   1   ,   1   ,   7   ,   15  ,   31  ,   53  ,   33  ,   225 ,   105 ,   907 ,   775 ,   3699    ,   439 ,   10087   ,0 };
            const unsigned long dim1625JoeKuoD6Init[]   =   {   1   ,   1   ,   5   ,   3   ,   23  ,   5   ,   3   ,   19  ,   387 ,   811 ,   865 ,   2817    ,   3757    ,   12849   ,0 };
            const unsigned long dim1626JoeKuoD6Init[]   =   {   1   ,   3   ,   5   ,   3   ,   3   ,   23  ,   51  ,   239 ,   489 ,   667 ,   301 ,   3879    ,   4141    ,   1131    ,0 };
            const unsigned long dim1627JoeKuoD6Init[]   =   {   1   ,   3   ,   3   ,   1   ,   21  ,   59  ,   115 ,   89  ,   235 ,   1009    ,   1371    ,   2141    ,   2489    ,   763 ,0 };
            const unsigned long dim1628JoeKuoD6Init[]   =   {   1   ,   1   ,   7   ,   1   ,   5   ,   7   ,   75  ,   83  ,   491 ,   233 ,   1027    ,   1421    ,   5111    ,   445 ,0 };
            const unsigned long dim1629JoeKuoD6Init[]   =   {   1   ,   3   ,   3   ,   3   ,   19  ,   27  ,   1   ,   181 ,   193 ,   949 ,   69  ,   2689    ,   3421    ,   13921   ,0 };
            const unsigned long dim1630JoeKuoD6Init[]   =   {   1   ,   3   ,   7   ,   3   ,   9   ,   35  ,   109 ,   203 ,   123 ,   217 ,   225 ,   3355    ,   2331    ,   11853   ,0 };
            const unsigned long dim1631JoeKuoD6Init[]   =   {   1   ,   1   ,   5   ,   13  ,   1   ,   55  ,   41  ,   225 ,   425 ,   801 ,   353 ,   2851    ,   4403    ,   7885    ,0 };
            const unsigned long dim1632JoeKuoD6Init[]   =   {   1   ,   3   ,   5   ,   11  ,   19  ,   47  ,   83  ,   17  ,   91  ,   875 ,   1739    ,   971 ,   8157    ,   11047   ,0 };
            const unsigned long dim1633JoeKuoD6Init[]   =   {   1   ,   3   ,   1   ,   15  ,   29  ,   47  ,   123 ,   209 ,   183 ,   473 ,   1019    ,   537 ,   1035    ,   4473    ,0 };
            const unsigned long dim1634JoeKuoD6Init[]   =   {   1   ,   3   ,   7   ,   7   ,   25  ,   9   ,   5   ,   115 ,   297 ,   411 ,   563 ,   883 ,   3509    ,   1637    ,0 };
            const unsigned long dim1635JoeKuoD6Init[]   =   {   1   ,   1   ,   1   ,   15  ,   3   ,   17  ,   55  ,   31  ,   123 ,   931 ,   1239    ,   2137    ,   3673    ,   10325   ,0 };
            const unsigned long dim1636JoeKuoD6Init[]   =   {   1   ,   1   ,   5   ,   1   ,   31  ,   49  ,   35  ,   51  ,   223 ,   65  ,   925 ,   2133    ,   3849    ,   2243    ,0 };
            const unsigned long dim1637JoeKuoD6Init[]   =   {   1   ,   1   ,   5   ,   15  ,   1   ,   27  ,   73  ,   111 ,   237 ,   877 ,   13  ,   3147    ,   7115    ,   13467   ,0 };
            const unsigned long dim1638JoeKuoD6Init[]   =   {   1   ,   3   ,   5   ,   7   ,   9   ,   53  ,   123 ,   189 ,   459 ,   383 ,   1259    ,   297 ,   3307    ,   3305    ,0 };
            const unsigned long dim1639JoeKuoD6Init[]   =   {   1   ,   3   ,   5   ,   15  ,   27  ,   45  ,   65  ,   175 ,   417 ,   377 ,   195 ,   1419    ,   1013    ,   8419    ,0 };
            const unsigned long dim1640JoeKuoD6Init[]   =   {   1   ,   3   ,   7   ,   3   ,   29  ,   13  ,   19  ,   37  ,   339 ,   515 ,   981 ,   335 ,   3965    ,   7577    ,0 };
            const unsigned long dim1641JoeKuoD6Init[]   =   {   1   ,   1   ,   1   ,   5   ,   27  ,   37  ,   85  ,   217 ,   463 ,   173 ,   1723    ,   493 ,   6847    ,   12253   ,0 };
            const unsigned long dim1642JoeKuoD6Init[]   =   {   1   ,   1   ,   3   ,   9   ,   19  ,   35  ,   53  ,   123 ,   365 ,   771 ,   961 ,   2545    ,   3191    ,   7891    ,0 };
            const unsigned long dim1643JoeKuoD6Init[]   =   {   1   ,   3   ,   7   ,   3   ,   27  ,   47  ,   81  ,   243 ,   89  ,   79  ,   13  ,   1733    ,   965 ,   39  ,0 };
            const unsigned long dim1644JoeKuoD6Init[]   =   {   1   ,   3   ,   1   ,   7   ,   21  ,   1   ,   57  ,   137 ,   211 ,   491 ,   1071    ,   2425    ,   5225    ,   16141   ,0 };
            const unsigned long dim1645JoeKuoD6Init[]   =   {   1   ,   3   ,   7   ,   1   ,   5   ,   23  ,   57  ,   117 ,   45  ,   373 ,   831 ,   3047    ,   5829    ,   3205    ,0 };
            const unsigned long dim1646JoeKuoD6Init[]   =   {   1   ,   3   ,   1   ,   7   ,   3   ,   59  ,   17  ,   171 ,   263 ,   723 ,   535 ,   3791    ,   4969    ,   4325    ,0 };
            const unsigned long dim1647JoeKuoD6Init[]   =   {   1   ,   1   ,   5   ,   7   ,   19  ,   5   ,   75  ,   169 ,   231 ,   25  ,   1925    ,   161 ,   8051    ,   13051   ,0 };
            const unsigned long dim1648JoeKuoD6Init[]   =   {   1   ,   1   ,   1   ,   9   ,   29  ,   5   ,   5   ,   197 ,   135 ,   713 ,   1895    ,   533 ,   6703    ,   9307    ,0 };
            const unsigned long dim1649JoeKuoD6Init[]   =   {   1   ,   3   ,   7   ,   3   ,   1   ,   47  ,   69  ,   247 ,   49  ,   15  ,   1533    ,   3051    ,   8097    ,   4881    ,0 };
            const unsigned long dim1650JoeKuoD6Init[]   =   {   1   ,   3   ,   1   ,   3   ,   21  ,   29  ,   81  ,   147 ,   123 ,   983 ,   1925    ,   3959    ,   3803    ,   13461   ,0 };
            const unsigned long dim1651JoeKuoD6Init[]   =   {   1   ,   1   ,   5   ,   11  ,   25  ,   59  ,   99  ,   253 ,   85  ,   37  ,   1821    ,   93  ,   1813    ,   6001    ,0 };
            const unsigned long dim1652JoeKuoD6Init[]   =   {   1   ,   1   ,   1   ,   15  ,   7   ,   53  ,   69  ,   129 ,   273 ,   5   ,   837 ,   441 ,   7573    ,   1803    ,0 };
            const unsigned long dim1653JoeKuoD6Init[]   =   {   1   ,   1   ,   5   ,   15  ,   25  ,   13  ,   125 ,   113 ,   405 ,   41  ,   1151    ,   3659    ,   3947    ,   12983   ,0 };
            const unsigned long dim1654JoeKuoD6Init[]   =   {   1   ,   1   ,   3   ,   15  ,   19  ,   1   ,   63  ,   223 ,   343 ,   523 ,   1857    ,   1273    ,   5251    ,   13665   ,0 };
            const unsigned long dim1655JoeKuoD6Init[]   =   {   1   ,   3   ,   1   ,   9   ,   3   ,   39  ,   45  ,   161 ,   483 ,   421 ,   159 ,   3257    ,   7749    ,   4481    ,0 };
            const unsigned long dim1656JoeKuoD6Init[]   =   {   1   ,   1   ,   5   ,   15  ,   13  ,   53  ,   25  ,   213 ,   179 ,   753 ,   1195    ,   343 ,   5803    ,   15289   ,0 };
            const unsigned long dim1657JoeKuoD6Init[]   =   {   1   ,   1   ,   3   ,   3   ,   21  ,   41  ,   67  ,   149 ,   225 ,   103 ,   1567    ,   2681    ,   3731    ,   9779    ,0 };
            const unsigned long dim1658JoeKuoD6Init[]   =   {   1   ,   3   ,   7   ,   3   ,   1   ,   25  ,   9   ,   185 ,   373 ,   851 ,   1245    ,   1635    ,   2255    ,   8411    ,0 };
            const unsigned long dim1659JoeKuoD6Init[]   =   {   1   ,   1   ,   7   ,   9   ,   1   ,   39  ,   99  ,   193 ,   115 ,   763 ,   1079    ,   1923    ,   6111    ,   2327    ,0 };
            const unsigned long dim1660JoeKuoD6Init[]   =   {   1   ,   1   ,   5   ,   13  ,   29  ,   51  ,   73  ,   91  ,   427 ,   959 ,   1321    ,   3105    ,   5093    ,   8935    ,0 };
            const unsigned long dim1661JoeKuoD6Init[]   =   {   1   ,   3   ,   3   ,   7   ,   3   ,   37  ,   19  ,   187 ,   341 ,   907 ,   119 ,   2177    ,   7459    ,   14047   ,0 };
            const unsigned long dim1662JoeKuoD6Init[]   =   {   1   ,   1   ,   7   ,   11  ,   29  ,   39  ,   95  ,   29  ,   17  ,   833 ,   1657    ,   2867    ,   4945    ,   6435    ,0 };
            const unsigned long dim1663JoeKuoD6Init[]   =   {   1   ,   3   ,   7   ,   9   ,   15  ,   51  ,   123 ,   201 ,   161 ,   239 ,   1937    ,   627 ,   7529    ,   15029   ,0 };
            const unsigned long dim1664JoeKuoD6Init[]   =   {   1   ,   1   ,   1   ,   1   ,   11  ,   57  ,   95  ,   155 ,   459 ,   157 ,   477 ,   1751    ,   4569    ,   13395   ,0 };
            const unsigned long dim1665JoeKuoD6Init[]   =   {   1   ,   1   ,   1   ,   13  ,   13  ,   51  ,   1   ,   33  ,   47  ,   781 ,   1527    ,   2471    ,   7245    ,   8565    ,0 };
            const unsigned long dim1666JoeKuoD6Init[]   =   {   1   ,   1   ,   3   ,   15  ,   1   ,   51  ,   81  ,   113 ,   317 ,   323 ,   1255    ,   2387    ,   6913    ,   2587    ,0 };
            const unsigned long dim1667JoeKuoD6Init[]   =   {   1   ,   1   ,   5   ,   3   ,   25  ,   9   ,   73  ,   163 ,   31  ,   367 ,   477 ,   3427    ,   7957    ,   12235   ,0 };
            const unsigned long dim1668JoeKuoD6Init[]   =   {   1   ,   1   ,   5   ,   15  ,   19  ,   31  ,   31  ,   93  ,   307 ,   371 ,   961 ,   3967    ,   4205    ,   12459   ,0 };
            const unsigned long dim1669JoeKuoD6Init[]   =   {   1   ,   1   ,   7   ,   3   ,   7   ,   13  ,   29  ,   75  ,   31  ,   583 ,   923 ,   2993    ,   5707    ,   715 ,0 };
            const unsigned long dim1670JoeKuoD6Init[]   =   {   1   ,   1   ,   1   ,   13  ,   21  ,   1   ,   103 ,   171 ,   469 ,   373 ,   415 ,   3031    ,   1411    ,   14589   ,0 };
            const unsigned long dim1671JoeKuoD6Init[]   =   {   1   ,   1   ,   7   ,   3   ,   17  ,   11  ,   57  ,   89  ,   441 ,   709 ,   1339    ,   47  ,   759 ,   8273    ,0 };
            const unsigned long dim1672JoeKuoD6Init[]   =   {   1   ,   3   ,   1   ,   11  ,   13  ,   61  ,   73  ,   31  ,   503 ,   619 ,   371 ,   3533    ,   4173    ,   5011    ,0 };
            const unsigned long dim1673JoeKuoD6Init[]   =   {   1   ,   1   ,   3   ,   11  ,   23  ,   1   ,   79  ,   7   ,   363 ,   919 ,   495 ,   15  ,   2595    ,   2691    ,0 };
            const unsigned long dim1674JoeKuoD6Init[]   =   {   1   ,   3   ,   1   ,   11  ,   25  ,   5   ,   57  ,   241 ,   23  ,   907 ,   43  ,   2895    ,   3697    ,   11003   ,0 };
            const unsigned long dim1675JoeKuoD6Init[]   =   {   1   ,   3   ,   3   ,   1   ,   3   ,   7   ,   57  ,   133 ,   491 ,   365 ,   1793    ,   3765    ,   3623    ,   11377   ,0 };
            const unsigned long dim1676JoeKuoD6Init[]   =   {   1   ,   3   ,   1   ,   13  ,   7   ,   37  ,   101 ,   133 ,   29  ,   689 ,   1997    ,   3011    ,   2321    ,   3139    ,0 };
            const unsigned long dim1677JoeKuoD6Init[]   =   {   1   ,   1   ,   5   ,   13  ,   17  ,   21  ,   119 ,   253 ,   323 ,   469 ,   1959    ,   537 ,   675 ,   4933    ,0 };
            const unsigned long dim1678JoeKuoD6Init[]   =   {   1   ,   1   ,   7   ,   11  ,   19  ,   19  ,   29  ,   7   ,   451 ,   111 ,   1743    ,   379 ,   5223    ,   3477    ,0 };
            const unsigned long dim1679JoeKuoD6Init[]   =   {   1   ,   1   ,   5   ,   15  ,   31  ,   35  ,   71  ,   217 ,   185 ,   599 ,   1233    ,   2079    ,   7127    ,   4785    ,0 };
            const unsigned long dim1680JoeKuoD6Init[]   =   {   1   ,   1   ,   7   ,   11  ,   1   ,   33  ,   41  ,   239 ,   429 ,   987 ,   1453    ,   81  ,   705 ,   6347    ,0 };
            const unsigned long dim1681JoeKuoD6Init[]   =   {   1   ,   3   ,   7   ,   13  ,   29  ,   49  ,   107 ,   217 ,   295 ,   797 ,   563 ,   1917    ,   537 ,   4355    ,0 };
            const unsigned long dim1682JoeKuoD6Init[]   =   {   1   ,   3   ,   1   ,   13  ,   17  ,   49  ,   27  ,   67  ,   5   ,   239 ,   53  ,   1759    ,   3235    ,   1321    ,0 };
            const unsigned long dim1683JoeKuoD6Init[]   =   {   1   ,   1   ,   7   ,   13  ,   11  ,   33  ,   85  ,   47  ,   285 ,   669 ,   341 ,   3011    ,   287 ,   9693    ,0 };
            const unsigned long dim1684JoeKuoD6Init[]   =   {   1   ,   3   ,   7   ,   7   ,   19  ,   57  ,   99  ,   97  ,   249 ,   567 ,   1709    ,   2053    ,   7603    ,   4011    ,0 };
            const unsigned long dim1685JoeKuoD6Init[]   =   {   1   ,   1   ,   3   ,   1   ,   3   ,   55  ,   105 ,   245 ,   167 ,   785 ,   1409    ,   3395    ,   6503    ,   14809   ,0 };
            const unsigned long dim1686JoeKuoD6Init[]   =   {   1   ,   3   ,   5   ,   13  ,   9   ,   63  ,   91  ,   49  ,   125 ,   123 ,   861 ,   1327    ,   5517    ,   3319    ,0 };
            const unsigned long dim1687JoeKuoD6Init[]   =   {   1   ,   3   ,   5   ,   7   ,   3   ,   63  ,   101 ,   167 ,   403 ,   43  ,   1767    ,   1365    ,   4577    ,   15423   ,0 };
            const unsigned long dim1688JoeKuoD6Init[]   =   {   1   ,   3   ,   3   ,   1   ,   3   ,   31  ,   81  ,   189 ,   273 ,   935 ,   1195    ,   3075    ,   4703    ,   7855    ,0 };
            const unsigned long dim1689JoeKuoD6Init[]   =   {   1   ,   1   ,   7   ,   9   ,   11  ,   21  ,   57  ,   69  ,   405 ,   865 ,   281 ,   3813    ,   2543    ,   4281    ,0 };
            const unsigned long dim1690JoeKuoD6Init[]   =   {   1   ,   3   ,   7   ,   11  ,   19  ,   49  ,   117 ,   21  ,   507 ,   481 ,   1961    ,   2893    ,   6627    ,   13545   ,0 };
            const unsigned long dim1691JoeKuoD6Init[]   =   {   1   ,   1   ,   1   ,   5   ,   7   ,   41  ,   57  ,   39  ,   233 ,   85  ,   35  ,   539 ,   8009    ,   7223    ,0 };
            const unsigned long dim1692JoeKuoD6Init[]   =   {   1   ,   1   ,   3   ,   11  ,   17  ,   51  ,   47  ,   219 ,   319 ,   141 ,   1993    ,   1473    ,   2403    ,   13697   ,0 };
            const unsigned long dim1693JoeKuoD6Init[]   =   {   1   ,   1   ,   5   ,   1   ,   5   ,   61  ,   69  ,   251 ,   499 ,   813 ,   123 ,   925 ,   2093    ,   3253    ,0 };
            const unsigned long dim1694JoeKuoD6Init[]   =   {   1   ,   1   ,   3   ,   11  ,   23  ,   19  ,   97  ,   19  ,   345 ,   127 ,   1571    ,   27  ,   2915    ,   1497    ,0 };
            const unsigned long dim1695JoeKuoD6Init[]   =   {   1   ,   3   ,   5   ,   7   ,   25  ,   23  ,   41  ,   41  ,   75  ,   571 ,   527 ,   517 ,   7057    ,   2737    ,0 };
            const unsigned long dim1696JoeKuoD6Init[]   =   {   1   ,   3   ,   7   ,   3   ,   1   ,   59  ,   27  ,   177 ,   253 ,   21  ,   2023    ,   3393    ,   3195    ,   879 ,0 };
            const unsigned long dim1697JoeKuoD6Init[]   =   {   1   ,   3   ,   3   ,   9   ,   7   ,   37  ,   1   ,   185 ,   381 ,   211 ,   2033    ,   1415    ,   3275    ,   15281   ,0 };
            const unsigned long dim1698JoeKuoD6Init[]   =   {   1   ,   1   ,   1   ,   15  ,   3   ,   49  ,   101 ,   89  ,   495 ,   683 ,   1945    ,   3017    ,   5209    ,   9879    ,0 };
            const unsigned long dim1699JoeKuoD6Init[]   =   {   1   ,   3   ,   3   ,   7   ,   17  ,   63  ,   55  ,   243 ,   215 ,   289 ,   1119    ,   4087    ,   3513    ,   2961    ,0 };
            const unsigned long dim1700JoeKuoD6Init[]   =   {   1   ,   1   ,   7   ,   13  ,   11  ,   49  ,   17  ,   103 ,   151 ,   725 ,   1433    ,   967 ,   4127    ,   14955   ,0 };
            const unsigned long dim1701JoeKuoD6Init[]   =   {   1   ,   3   ,   5   ,   9   ,   11  ,   5   ,   69  ,   179 ,   163 ,   565 ,   245 ,   2013    ,   699 ,   7109    ,0 };
            const unsigned long dim1702JoeKuoD6Init[]   =   {   1   ,   3   ,   7   ,   15  ,   29  ,   23  ,   39  ,   93  ,   211 ,   1017    ,   1677    ,   3883    ,   6157    ,   11339   ,0 };
            const unsigned long dim1703JoeKuoD6Init[]   =   {   1   ,   1   ,   7   ,   9   ,   7   ,   15  ,   27  ,   109 ,   449 ,   165 ,   1785    ,   1945    ,   2095    ,   14831   ,0 };
            const unsigned long dim1704JoeKuoD6Init[]   =   {   1   ,   1   ,   5   ,   13  ,   17  ,   43  ,   71  ,   49  ,   9   ,   305 ,   561 ,   2309    ,   6409    ,   2097    ,0 };
            const unsigned long dim1705JoeKuoD6Init[]   =   {   1   ,   1   ,   7   ,   5   ,   9   ,   63  ,   77  ,   149 ,   119 ,   727 ,   1781    ,   3429    ,   141 ,   13329   ,0 };
            const unsigned long dim1706JoeKuoD6Init[]   =   {   1   ,   1   ,   1   ,   7   ,   23  ,   27  ,   111 ,   193 ,   97  ,   727 ,   425 ,   2665    ,   2217    ,   7463    ,0 };
            const unsigned long dim1707JoeKuoD6Init[]   =   {   1   ,   3   ,   3   ,   5   ,   15  ,   45  ,   85  ,   237 ,   207 ,   759 ,   1007    ,   2545    ,   7897    ,   10953   ,0 };
            const unsigned long dim1708JoeKuoD6Init[]   =   {   1   ,   1   ,   1   ,   3   ,   7   ,   47  ,   81  ,   57  ,   369 ,   777 ,   155 ,   1341    ,   6679    ,   5877    ,0 };
            const unsigned long dim1709JoeKuoD6Init[]   =   {   1   ,   1   ,   3   ,   9   ,   9   ,   31  ,   47  ,   135 ,   31  ,   275 ,   1779    ,   2537    ,   2157    ,   7347    ,0 };
            const unsigned long dim1710JoeKuoD6Init[]   =   {   1   ,   1   ,   5   ,   9   ,   23  ,   51  ,   59  ,   107 ,   43  ,   639 ,   1893    ,   1713    ,   2409    ,   6301    ,0 };
            const unsigned long dim1711JoeKuoD6Init[]   =   {   1   ,   3   ,   5   ,   13  ,   9   ,   13  ,   13  ,   231 ,   419 ,   373 ,   1071    ,   2903    ,   3565    ,   15075   ,0 };
            const unsigned long dim1712JoeKuoD6Init[]   =   {   1   ,   3   ,   5   ,   5   ,   1   ,   3   ,   33  ,   123 ,   53  ,   273 ,   731 ,   2049    ,   4617    ,   12341   ,0 };
            const unsigned long dim1713JoeKuoD6Init[]   =   {   1   ,   3   ,   1   ,   1   ,   17  ,   49  ,   75  ,   141 ,   499 ,   863 ,   635 ,   2967    ,   3483    ,   5087    ,0 };
            const unsigned long dim1714JoeKuoD6Init[]   =   {   1   ,   3   ,   1   ,   9   ,   17  ,   7   ,   33  ,   247 ,   143 ,   73  ,   551 ,   65  ,   1261    ,   10177   ,0 };
            const unsigned long dim1715JoeKuoD6Init[]   =   {   1   ,   3   ,   7   ,   3   ,   11  ,   13  ,   123 ,   137 ,   29  ,   175 ,   1353    ,   627 ,   1223    ,   191 ,0 };
            const unsigned long dim1716JoeKuoD6Init[]   =   {   1   ,   3   ,   5   ,   15  ,   29  ,   23  ,   89  ,   183 ,   427 ,   629 ,   1747    ,   141 ,   95  ,   8319    ,0 };
            const unsigned long dim1717JoeKuoD6Init[]   =   {   1   ,   1   ,   1   ,   5   ,   17  ,   45  ,   53  ,   95  ,   433 ,   523 ,   1175    ,   1227    ,   6893    ,   2081    ,0 };
            const unsigned long dim1718JoeKuoD6Init[]   =   {   1   ,   3   ,   3   ,   15  ,   29  ,   21  ,   107 ,   117 ,   407 ,   937 ,   301 ,   1443    ,   6841    ,   5673    ,0 };
            const unsigned long dim1719JoeKuoD6Init[]   =   {   1   ,   3   ,   5   ,   11  ,   15  ,   13  ,   81  ,   213 ,   259 ,   619 ,   375 ,   1777    ,   455 ,   8185    ,0 };
            const unsigned long dim1720JoeKuoD6Init[]   =   {   1   ,   1   ,   5   ,   7   ,   13  ,   25  ,   113 ,   175 ,   31  ,   265 ,   1901    ,   1779    ,   1787    ,   14551   ,0 };
            const unsigned long dim1721JoeKuoD6Init[]   =   {   1   ,   3   ,   1   ,   9   ,   17  ,   59  ,   41  ,   183 ,   405 ,   477 ,   133 ,   3993    ,   7217    ,   9373    ,0 };
            const unsigned long dim1722JoeKuoD6Init[]   =   {   1   ,   1   ,   1   ,   13  ,   7   ,   27  ,   85  ,   133 ,   165 ,   1023    ,   443 ,   2549    ,   1923    ,   6561    ,0 };
            const unsigned long dim1723JoeKuoD6Init[]   =   {   1   ,   3   ,   7   ,   3   ,   15  ,   51  ,   11  ,   167 ,   321 ,   37  ,   1299    ,   1953    ,   5365    ,   8423    ,0 };
            const unsigned long dim1724JoeKuoD6Init[]   =   {   1   ,   1   ,   3   ,   1   ,   7   ,   45  ,   41  ,   219 ,   149 ,   355 ,   1975    ,   1711    ,   1643    ,   13669   ,0 };
            const unsigned long dim1725JoeKuoD6Init[]   =   {   1   ,   1   ,   7   ,   7   ,   11  ,   9   ,   105 ,   221 ,   275 ,   845 ,   1651    ,   483 ,   7463    ,   7729    ,0 };
            const unsigned long dim1726JoeKuoD6Init[]   =   {   1   ,   3   ,   3   ,   9   ,   7   ,   29  ,   81  ,   141 ,   503 ,   321 ,   2039    ,   405 ,   7503    ,   3751    ,0 };
            const unsigned long dim1727JoeKuoD6Init[]   =   {   1   ,   1   ,   1   ,   7   ,   21  ,   31  ,   19  ,   71  ,   299 ,   301 ,   1101    ,   2821    ,   6005    ,   5361    ,0 };
            const unsigned long dim1728JoeKuoD6Init[]   =   {   1   ,   1   ,   7   ,   15  ,   11  ,   61  ,   37  ,   67  ,   151 ,   175 ,   1815    ,   2779    ,   5507    ,   153 ,0 };
            const unsigned long dim1729JoeKuoD6Init[]   =   {   1   ,   1   ,   5   ,   11  ,   9   ,   41  ,   83  ,   225 ,   109 ,   23  ,   563 ,   1021    ,   3523    ,   14027   ,0 };
            const unsigned long dim1730JoeKuoD6Init[]   =   {   1   ,   3   ,   3   ,   7   ,   3   ,   19  ,   33  ,   251 ,   337 ,   925 ,   1309    ,   1895    ,   4113    ,   14017   ,0 };
            const unsigned long dim1731JoeKuoD6Init[]   =   {   1   ,   3   ,   7   ,   9   ,   15  ,   57  ,   15  ,   121 ,   311 ,   929 ,   1697    ,   1877    ,   803 ,   3361    ,0 };
            const unsigned long dim1732JoeKuoD6Init[]   =   {   1   ,   3   ,   5   ,   1   ,   11  ,   29  ,   19  ,   173 ,   497 ,   879 ,   249 ,   3111    ,   4227    ,   14681   ,0 };
            const unsigned long dim1733JoeKuoD6Init[]   =   {   1   ,   3   ,   7   ,   13  ,   13  ,   41  ,   51  ,   99  ,   137 ,   867 ,   1207    ,   3343    ,   613 ,   15047   ,0 };
            const unsigned long dim1734JoeKuoD6Init[]   =   {   1   ,   3   ,   3   ,   15  ,   11  ,   41  ,   57  ,   7   ,   203 ,   569 ,   669 ,   3185    ,   1233    ,   527 ,0 };
            const unsigned long dim1735JoeKuoD6Init[]   =   {   1   ,   3   ,   5   ,   5   ,   27  ,   45  ,   83  ,   49  ,   335 ,   747 ,   1075    ,   2639    ,   3057    ,   4137    ,0 };
            const unsigned long dim1736JoeKuoD6Init[]   =   {   1   ,   3   ,   7   ,   15  ,   3   ,   11  ,   9   ,   177 ,   153 ,   401 ,   61  ,   3689    ,   3853    ,   15605   ,0 };
            const unsigned long dim1737JoeKuoD6Init[]   =   {   1   ,   3   ,   3   ,   7   ,   27  ,   13  ,   79  ,   51  ,   213 ,   465 ,   1917    ,   3353    ,   3553    ,   14853   ,0 };
            const unsigned long dim1738JoeKuoD6Init[]   =   {   1   ,   3   ,   1   ,   9   ,   25  ,   27  ,   67  ,   253 ,   29  ,   635 ,   239 ,   679 ,   3653    ,   15435   ,0 };
            const unsigned long dim1739JoeKuoD6Init[]   =   {   1   ,   1   ,   1   ,   7   ,   23  ,   9   ,   77  ,   167 ,   373 ,   295 ,   1423    ,   1895    ,   1541    ,   14563   ,0 };
            const unsigned long dim1740JoeKuoD6Init[]   =   {   1   ,   3   ,   3   ,   1   ,   9   ,   37  ,   117 ,   223 ,   109 ,   969 ,   149 ,   1611    ,   3511    ,   11357   ,0 };
            const unsigned long dim1741JoeKuoD6Init[]   =   {   1   ,   3   ,   5   ,   11  ,   19  ,   61  ,   105 ,   137 ,   113 ,   443 ,   359 ,   1705    ,   6893    ,   11371   ,0 };
            const unsigned long dim1742JoeKuoD6Init[]   =   {   1   ,   3   ,   1   ,   13  ,   21  ,   11  ,   65  ,   3   ,   275 ,   311 ,   1157    ,   87  ,   4841    ,   2995    ,0 };
            const unsigned long dim1743JoeKuoD6Init[]   =   {   1   ,   1   ,   7   ,   3   ,   17  ,   41  ,   15  ,   203 ,   249 ,   353 ,   861 ,   2519    ,   537 ,   15203   ,0 };
            const unsigned long dim1744JoeKuoD6Init[]   =   {   1   ,   3   ,   3   ,   13  ,   29  ,   7   ,   87  ,   121 ,   69  ,   733 ,   1035    ,   2731    ,   8079    ,   7243    ,0 };
            const unsigned long dim1745JoeKuoD6Init[]   =   {   1   ,   1   ,   1   ,   3   ,   13  ,   3   ,   109 ,   47  ,   161 ,   313 ,   1641    ,   2621    ,   7629    ,   13713   ,0 };
            const unsigned long dim1746JoeKuoD6Init[]   =   {   1   ,   1   ,   3   ,   1   ,   5   ,   5   ,   85  ,   99  ,   463 ,   1005    ,   89  ,   1209    ,   8013    ,   14393   ,0 };
            const unsigned long dim1747JoeKuoD6Init[]   =   {   1   ,   1   ,   3   ,   1   ,   27  ,   1   ,   127 ,   185 ,   193 ,   275 ,   891 ,   763 ,   3775    ,   2569    ,0 };
            const unsigned long dim1748JoeKuoD6Init[]   =   {   1   ,   3   ,   7   ,   1   ,   7   ,   43  ,   123 ,   3   ,   131 ,   441 ,   771 ,   3413    ,   6641    ,   2363    ,0 };
            const unsigned long dim1749JoeKuoD6Init[]   =   {   1   ,   3   ,   3   ,   9   ,   21  ,   39  ,   81  ,   243 ,   179 ,   765 ,   181 ,   417 ,   8027    ,   1025    ,0 };
            const unsigned long dim1750JoeKuoD6Init[]   =   {   1   ,   1   ,   3   ,   11  ,   11  ,   37  ,   23  ,   167 ,   257 ,   103 ,   305 ,   1097    ,   6099    ,   11321   ,0 };
            const unsigned long dim1751JoeKuoD6Init[]   =   {   1   ,   3   ,   1   ,   7   ,   27  ,   29  ,   71  ,   207 ,   83  ,   367 ,   1575    ,   3171    ,   759 ,   10981   ,0 };
            const unsigned long dim1752JoeKuoD6Init[]   =   {   1   ,   1   ,   7   ,   1   ,   23  ,   29  ,   93  ,   89  ,   41  ,   577 ,   1657    ,   415 ,   655 ,   2501    ,0 };
            const unsigned long dim1753JoeKuoD6Init[]   =   {   1   ,   1   ,   1   ,   11  ,   27  ,   39  ,   111 ,   139 ,   425 ,   513 ,   1341    ,   2059    ,   2187    ,   6047    ,0 };
            const unsigned long dim1754JoeKuoD6Init[]   =   {   1   ,   1   ,   1   ,   13  ,   25  ,   55  ,   63  ,   193 ,   183 ,   429 ,   611 ,   2119    ,   2495    ,   15361   ,0 };
            const unsigned long dim1755JoeKuoD6Init[]   =   {   1   ,   1   ,   1   ,   3   ,   9   ,   31  ,   15  ,   185 ,   139 ,   603 ,   1159    ,   2065    ,   1159    ,   14869   ,0 };
            const unsigned long dim1756JoeKuoD6Init[]   =   {   1   ,   1   ,   3   ,   5   ,   21  ,   5   ,   117 ,   191 ,   357 ,   61  ,   1517    ,   2075    ,   2041    ,   14485   ,0 };
            const unsigned long dim1757JoeKuoD6Init[]   =   {   1   ,   1   ,   3   ,   15  ,   21  ,   59  ,   13  ,   67  ,   393 ,   809 ,   1539    ,   3631    ,   5437    ,   7787    ,0 };
            const unsigned long dim1758JoeKuoD6Init[]   =   {   1   ,   3   ,   1   ,   5   ,   23  ,   19  ,   53  ,   237 ,   123 ,   123 ,   351 ,   1797    ,   3415    ,   4721    ,0 };
            const unsigned long dim1759JoeKuoD6Init[]   =   {   1   ,   1   ,   3   ,   1   ,   13  ,   39  ,   75  ,   41  ,   425 ,   557 ,   985 ,   1485    ,   7613    ,   11109   ,0 };
            const unsigned long dim1760JoeKuoD6Init[]   =   {   1   ,   3   ,   1   ,   11  ,   25  ,   29  ,   23  ,   117 ,   411 ,   721 ,   21  ,   1707    ,   7685    ,   11979   ,0 };
            const unsigned long dim1761JoeKuoD6Init[]   =   {   1   ,   3   ,   1   ,   5   ,   25  ,   55  ,   1   ,   221 ,   339 ,   705 ,   641 ,   515 ,   5883    ,   10709   ,0 };
            const unsigned long dim1762JoeKuoD6Init[]   =   {   1   ,   1   ,   3   ,   9   ,   13  ,   63  ,   69  ,   121 ,   371 ,   643 ,   597 ,   1977    ,   1295    ,   10941   ,0 };
            const unsigned long dim1763JoeKuoD6Init[]   =   {   1   ,   3   ,   5   ,   5   ,   21  ,   49  ,   91  ,   37  ,   53  ,   549 ,   905 ,   1707    ,   7371    ,   7539    ,0 };
            const unsigned long dim1764JoeKuoD6Init[]   =   {   1   ,   1   ,   5   ,   9   ,   9   ,   43  ,   45  ,   205 ,   49  ,   1001    ,   1617    ,   2307    ,   3359    ,   5871    ,0 };
            const unsigned long dim1765JoeKuoD6Init[]   =   {   1   ,   3   ,   5   ,   7   ,   7   ,   43  ,   73  ,   129 ,   381 ,   97  ,   529 ,   607 ,   2505    ,   5171    ,0 };
            const unsigned long dim1766JoeKuoD6Init[]   =   {   1   ,   1   ,   5   ,   9   ,   3   ,   31  ,   81  ,   63  ,   185 ,   443 ,   1215    ,   3123    ,   7973    ,   8457    ,0 };
            const unsigned long dim1767JoeKuoD6Init[]   =   {   1   ,   1   ,   5   ,   5   ,   13  ,   31  ,   59  ,   141 ,   93  ,   509 ,   149 ,   3039    ,   6011    ,   11489   ,0 };
            const unsigned long dim1768JoeKuoD6Init[]   =   {   1   ,   1   ,   5   ,   5   ,   17  ,   13  ,   93  ,   221 ,   23  ,   325 ,   415 ,   829 ,   7611    ,   2183    ,0 };
            const unsigned long dim1769JoeKuoD6Init[]   =   {   1   ,   3   ,   3   ,   3   ,   13  ,   55  ,   47  ,   77  ,   375 ,   41  ,   1821    ,   2599    ,   843 ,   13411   ,0 };
            const unsigned long dim1770JoeKuoD6Init[]   =   {   1   ,   3   ,   1   ,   15  ,   17  ,   15  ,   101 ,   225 ,   47  ,   217 ,   137 ,   2637    ,   895 ,   10679   ,0 };
            const unsigned long dim1771JoeKuoD6Init[]   =   {   1   ,   1   ,   7   ,   3   ,   27  ,   37  ,   123 ,   173 ,   83  ,   119 ,   599 ,   17  ,   89  ,   15621   ,0 };
            const unsigned long dim1772JoeKuoD6Init[]   =   {   1   ,   3   ,   7   ,   11  ,   25  ,   41  ,   105 ,   43  ,   247 ,   841 ,   1331    ,   2323    ,   3499    ,   5443    ,0 };
            const unsigned long dim1773JoeKuoD6Init[]   =   {   1   ,   1   ,   1   ,   7   ,   23  ,   49  ,   101 ,   179 ,   233 ,   463 ,   443 ,   2825    ,   3431    ,   12585   ,0 };
            const unsigned long dim1774JoeKuoD6Init[]   =   {   1   ,   3   ,   7   ,   13  ,   31  ,   49  ,   115 ,   173 ,   193 ,   615 ,   663 ,   2167    ,   5655    ,   15331   ,0 };
            const unsigned long dim1775JoeKuoD6Init[]   =   {   1   ,   3   ,   7   ,   1   ,   1   ,   55  ,   73  ,   145 ,   345 ,   475 ,   871 ,   3809    ,   5329    ,   15785   ,0 };
            const unsigned long dim1776JoeKuoD6Init[]   =   {   1   ,   1   ,   1   ,   3   ,   7   ,   7   ,   45  ,   3   ,   9   ,   753 ,   703 ,   3381    ,   4831    ,   5145    ,0 };
            const unsigned long dim1777JoeKuoD6Init[]   =   {   1   ,   3   ,   1   ,   15  ,   21  ,   51  ,   93  ,   181 ,   275 ,   127 ,   1743    ,   509 ,   2231    ,   13423   ,0 };
            const unsigned long dim1778JoeKuoD6Init[]   =   {   1   ,   1   ,   3   ,   11  ,   31  ,   29  ,   33  ,   49  ,   495 ,   645 ,   49  ,   3139    ,   5281    ,   12193   ,0 };
            const unsigned long dim1779JoeKuoD6Init[]   =   {   1   ,   3   ,   3   ,   3   ,   3   ,   49  ,   41  ,   31  ,   101 ,   681 ,   1331    ,   1001    ,   6679    ,   4525    ,0 };
            const unsigned long dim1780JoeKuoD6Init[]   =   {   1   ,   3   ,   5   ,   11  ,   7   ,   51  ,   49  ,   197 ,   43  ,   419 ,   965 ,   3105    ,   1325    ,   3137    ,0 };
            const unsigned long dim1781JoeKuoD6Init[]   =   {   1   ,   1   ,   5   ,   3   ,   15  ,   31  ,   73  ,   147 ,   1   ,   471 ,   1569    ,   3647    ,   7251    ,   6261    ,0 };
            const unsigned long dim1782JoeKuoD6Init[]   =   {   1   ,   3   ,   3   ,   1   ,   5   ,   43  ,   41  ,   215 ,   161 ,   521 ,   1349    ,   1627    ,   1177    ,   11719   ,0 };
            const unsigned long dim1783JoeKuoD6Init[]   =   {   1   ,   1   ,   3   ,   11  ,   31  ,   5   ,   125 ,   79  ,   199 ,   581 ,   1153    ,   2937    ,   7875    ,   16361   ,0 };
            const unsigned long dim1784JoeKuoD6Init[]   =   {   1   ,   1   ,   5   ,   1   ,   1   ,   31  ,   7   ,   7   ,   475 ,   807 ,   1249    ,   811 ,   6865    ,   13897   ,0 };
            const unsigned long dim1785JoeKuoD6Init[]   =   {   1   ,   3   ,   1   ,   9   ,   5   ,   7   ,   11  ,   203 ,   215 ,   669 ,   1963    ,   1255    ,   4785    ,   11215   ,0 };
            const unsigned long dim1786JoeKuoD6Init[]   =   {   1   ,   1   ,   5   ,   13  ,   19  ,   57  ,   89  ,   133 ,   287 ,   469 ,   191 ,   3177    ,   1425    ,   14149   ,0 };
            const unsigned long dim1787JoeKuoD6Init[]   =   {   1   ,   3   ,   5   ,   15  ,   9   ,   9   ,   13  ,   229 ,   477 ,   261 ,   1015    ,   1669    ,   7685    ,   3337    ,0 };
            const unsigned long dim1788JoeKuoD6Init[]   =   {   1   ,   1   ,   1   ,   13  ,   21  ,   7   ,   117 ,   35  ,   233 ,   347 ,   1347    ,   2171    ,   2419    ,   11555   ,0 };
            const unsigned long dim1789JoeKuoD6Init[]   =   {   1   ,   3   ,   7   ,   5   ,   17  ,   55  ,   37  ,   243 ,   159 ,   605 ,   1079    ,   111 ,   3149    ,   3973    ,0 };
            const unsigned long dim1790JoeKuoD6Init[]   =   {   1   ,   3   ,   5   ,   7   ,   29  ,   29  ,   111 ,   49  ,   445 ,   241 ,   875 ,   3895    ,   3915    ,   6337    ,0 };
            const unsigned long dim1791JoeKuoD6Init[]   =   {   1   ,   3   ,   1   ,   3   ,   13  ,   7   ,   21  ,   185 ,   121 ,   319 ,   1391    ,   1303    ,   3123    ,   7699    ,0 };
            const unsigned long dim1792JoeKuoD6Init[]   =   {   1   ,   1   ,   5   ,   5   ,   3   ,   9   ,   69  ,   205 ,   423 ,   189 ,   687 ,   1453    ,   2427    ,   11231   ,0 };
            const unsigned long dim1793JoeKuoD6Init[]   =   {   1   ,   3   ,   3   ,   7   ,   5   ,   31  ,   91  ,   235 ,   287 ,   185 ,   1439    ,   2011    ,   7429    ,   4313    ,0 };
            const unsigned long dim1794JoeKuoD6Init[]   =   {   1   ,   1   ,   7   ,   1   ,   3   ,   57  ,   11  ,   11  ,   377 ,   699 ,   883 ,   221 ,   4849    ,   11373   ,0 };
            const unsigned long dim1795JoeKuoD6Init[]   =   {   1   ,   1   ,   1   ,   11  ,   3   ,   15  ,   51  ,   79  ,   83  ,   915 ,   1429    ,   811 ,   2651    ,   829 ,0 };
            const unsigned long dim1796JoeKuoD6Init[]   =   {   1   ,   1   ,   1   ,   5   ,   21  ,   17  ,   5   ,   227 ,   379 ,   383 ,   745 ,   2761    ,   323 ,   753 ,0 };
            const unsigned long dim1797JoeKuoD6Init[]   =   {   1   ,   1   ,   3   ,   9   ,   11  ,   33  ,   123 ,   59  ,   241 ,   829 ,   879 ,   3339    ,   5315    ,   10777   ,0 };
            const unsigned long dim1798JoeKuoD6Init[]   =   {   1   ,   3   ,   7   ,   13  ,   25  ,   27  ,   117 ,   5   ,   11  ,   793 ,   243 ,   2571    ,   369 ,   1371    ,0 };
            const unsigned long dim1799JoeKuoD6Init[]   =   {   1   ,   1   ,   1   ,   3   ,   21  ,   31  ,   95  ,   217 ,   175 ,   837 ,   583 ,   1149    ,   8073    ,   12751   ,0 };


            const unsigned long * const JoeKuoD6initializers[1799]
            =
            {
                dim1JoeKuoD6Init,
                dim2JoeKuoD6Init,
                dim3JoeKuoD6Init,
                dim4JoeKuoD6Init,
                dim5JoeKuoD6Init,
                dim6JoeKuoD6Init,
                dim7JoeKuoD6Init,
                dim8JoeKuoD6Init,
                dim9JoeKuoD6Init,
                dim10JoeKuoD6Init,
                dim11JoeKuoD6Init,
                dim12JoeKuoD6Init,
                dim13JoeKuoD6Init,
                dim14JoeKuoD6Init,
                dim15JoeKuoD6Init,
                dim16JoeKuoD6Init,
                dim17JoeKuoD6Init,
                dim18JoeKuoD6Init,
                dim19JoeKuoD6Init,
                dim20JoeKuoD6Init,
                dim21JoeKuoD6Init,
                dim22JoeKuoD6Init,
                dim23JoeKuoD6Init,
                dim24JoeKuoD6Init,
                dim25JoeKuoD6Init,
                dim26JoeKuoD6Init,
                dim27JoeKuoD6Init,
                dim28JoeKuoD6Init,
                dim29JoeKuoD6Init,
                dim30JoeKuoD6Init,
                dim31JoeKuoD6Init,
                dim32JoeKuoD6Init,
                dim33JoeKuoD6Init,
                dim34JoeKuoD6Init,
                dim35JoeKuoD6Init,
                dim36JoeKuoD6Init,
                dim37JoeKuoD6Init,
                dim38JoeKuoD6Init,
                dim39JoeKuoD6Init,
                dim40JoeKuoD6Init,
                dim41JoeKuoD6Init,
                dim42JoeKuoD6Init,
                dim43JoeKuoD6Init,
                dim44JoeKuoD6Init,
                dim45JoeKuoD6Init,
                dim46JoeKuoD6Init,
                dim47JoeKuoD6Init,
                dim48JoeKuoD6Init,
                dim49JoeKuoD6Init,
                dim50JoeKuoD6Init,
                dim51JoeKuoD6Init,
                dim52JoeKuoD6Init,
                dim53JoeKuoD6Init,
                dim54JoeKuoD6Init,
                dim55JoeKuoD6Init,
                dim56JoeKuoD6Init,
                dim57JoeKuoD6Init,
                dim58JoeKuoD6Init,
                dim59JoeKuoD6Init,
                dim60JoeKuoD6Init,
                dim61JoeKuoD6Init,
                dim62JoeKuoD6Init,
                dim63JoeKuoD6Init,
                dim64JoeKuoD6Init,
                dim65JoeKuoD6Init,
                dim66JoeKuoD6Init,
                dim67JoeKuoD6Init,
                dim68JoeKuoD6Init,
                dim69JoeKuoD6Init,
                dim70JoeKuoD6Init,
                dim71JoeKuoD6Init,
                dim72JoeKuoD6Init,
                dim73JoeKuoD6Init,
                dim74JoeKuoD6Init,
                dim75JoeKuoD6Init,
                dim76JoeKuoD6Init,
                dim77JoeKuoD6Init,
                dim78JoeKuoD6Init,
                dim79JoeKuoD6Init,
                dim80JoeKuoD6Init,
                dim81JoeKuoD6Init,
                dim82JoeKuoD6Init,
                dim83JoeKuoD6Init,
                dim84JoeKuoD6Init,
                dim85JoeKuoD6Init,
                dim86JoeKuoD6Init,
                dim87JoeKuoD6Init,
                dim88JoeKuoD6Init,
                dim89JoeKuoD6Init,
                dim90JoeKuoD6Init,
                dim91JoeKuoD6Init,
                dim92JoeKuoD6Init,
                dim93JoeKuoD6Init,
                dim94JoeKuoD6Init,
                dim95JoeKuoD6Init,
                dim96JoeKuoD6Init,
                dim97JoeKuoD6Init,
                dim98JoeKuoD6Init,
                dim99JoeKuoD6Init,
                dim100JoeKuoD6Init,
                dim101JoeKuoD6Init,
                dim102JoeKuoD6Init,
                dim103JoeKuoD6Init,
                dim104JoeKuoD6Init,
                dim105JoeKuoD6Init,
                dim106JoeKuoD6Init,
                dim107JoeKuoD6Init,
                dim108JoeKuoD6Init,
                dim109JoeKuoD6Init,
                dim110JoeKuoD6Init,
                dim111JoeKuoD6Init,
                dim112JoeKuoD6Init,
                dim113JoeKuoD6Init,
                dim114JoeKuoD6Init,
                dim115JoeKuoD6Init,
                dim116JoeKuoD6Init,
                dim117JoeKuoD6Init,
                dim118JoeKuoD6Init,
                dim119JoeKuoD6Init,
                dim120JoeKuoD6Init,
                dim121JoeKuoD6Init,
                dim122JoeKuoD6Init,
                dim123JoeKuoD6Init,
                dim124JoeKuoD6Init,
                dim125JoeKuoD6Init,
                dim126JoeKuoD6Init,
                dim127JoeKuoD6Init,
                dim128JoeKuoD6Init,
                dim129JoeKuoD6Init,
                dim130JoeKuoD6Init,
                dim131JoeKuoD6Init,
                dim132JoeKuoD6Init,
                dim133JoeKuoD6Init,
                dim134JoeKuoD6Init,
                dim135JoeKuoD6Init,
                dim136JoeKuoD6Init,
                dim137JoeKuoD6Init,
                dim138JoeKuoD6Init,
                dim139JoeKuoD6Init,
                dim140JoeKuoD6Init,
                dim141JoeKuoD6Init,
                dim142JoeKuoD6Init,
                dim143JoeKuoD6Init,
                dim144JoeKuoD6Init,
                dim145JoeKuoD6Init,
                dim146JoeKuoD6Init,
                dim147JoeKuoD6Init,
                dim148JoeKuoD6Init,
                dim149JoeKuoD6Init,
                dim150JoeKuoD6Init,
                dim151JoeKuoD6Init,
                dim152JoeKuoD6Init,
                dim153JoeKuoD6Init,
                dim154JoeKuoD6Init,
                dim155JoeKuoD6Init,
                dim156JoeKuoD6Init,
                dim157JoeKuoD6Init,
                dim158JoeKuoD6Init,
                dim159JoeKuoD6Init,
                dim160JoeKuoD6Init,
                dim161JoeKuoD6Init,
                dim162JoeKuoD6Init,
                dim163JoeKuoD6Init,
                dim164JoeKuoD6Init,
                dim165JoeKuoD6Init,
                dim166JoeKuoD6Init,
                dim167JoeKuoD6Init,
                dim168JoeKuoD6Init,
                dim169JoeKuoD6Init,
                dim170JoeKuoD6Init,
                dim171JoeKuoD6Init,
                dim172JoeKuoD6Init,
                dim173JoeKuoD6Init,
                dim174JoeKuoD6Init,
                dim175JoeKuoD6Init,
                dim176JoeKuoD6Init,
                dim177JoeKuoD6Init,
                dim178JoeKuoD6Init,
                dim179JoeKuoD6Init,
                dim180JoeKuoD6Init,
                dim181JoeKuoD6Init,
                dim182JoeKuoD6Init,
                dim183JoeKuoD6Init,
                dim184JoeKuoD6Init,
                dim185JoeKuoD6Init,
                dim186JoeKuoD6Init,
                dim187JoeKuoD6Init,
                dim188JoeKuoD6Init,
                dim189JoeKuoD6Init,
                dim190JoeKuoD6Init,
                dim191JoeKuoD6Init,
                dim192JoeKuoD6Init,
                dim193JoeKuoD6Init,
                dim194JoeKuoD6Init,
                dim195JoeKuoD6Init,
                dim196JoeKuoD6Init,
                dim197JoeKuoD6Init,
                dim198JoeKuoD6Init,
                dim199JoeKuoD6Init,
                dim200JoeKuoD6Init,
                dim201JoeKuoD6Init,
                dim202JoeKuoD6Init,
                dim203JoeKuoD6Init,
                dim204JoeKuoD6Init,
                dim205JoeKuoD6Init,
                dim206JoeKuoD6Init,
                dim207JoeKuoD6Init,
                dim208JoeKuoD6Init,
                dim209JoeKuoD6Init,
                dim210JoeKuoD6Init,
                dim211JoeKuoD6Init,
                dim212JoeKuoD6Init,
                dim213JoeKuoD6Init,
                dim214JoeKuoD6Init,
                dim215JoeKuoD6Init,
                dim216JoeKuoD6Init,
                dim217JoeKuoD6Init,
                dim218JoeKuoD6Init,
                dim219JoeKuoD6Init,
                dim220JoeKuoD6Init,
                dim221JoeKuoD6Init,
                dim222JoeKuoD6Init,
                dim223JoeKuoD6Init,
                dim224JoeKuoD6Init,
                dim225JoeKuoD6Init,
                dim226JoeKuoD6Init,
                dim227JoeKuoD6Init,
                dim228JoeKuoD6Init,
                dim229JoeKuoD6Init,
                dim230JoeKuoD6Init,
                dim231JoeKuoD6Init,
                dim232JoeKuoD6Init,
                dim233JoeKuoD6Init,
                dim234JoeKuoD6Init,
                dim235JoeKuoD6Init,
                dim236JoeKuoD6Init,
                dim237JoeKuoD6Init,
                dim238JoeKuoD6Init,
                dim239JoeKuoD6Init,
                dim240JoeKuoD6Init,
                dim241JoeKuoD6Init,
                dim242JoeKuoD6Init,
                dim243JoeKuoD6Init,
                dim244JoeKuoD6Init,
                dim245JoeKuoD6Init,
                dim246JoeKuoD6Init,
                dim247JoeKuoD6Init,
                dim248JoeKuoD6Init,
                dim249JoeKuoD6Init,
                dim250JoeKuoD6Init,
                dim251JoeKuoD6Init,
                dim252JoeKuoD6Init,
                dim253JoeKuoD6Init,
                dim254JoeKuoD6Init,
                dim255JoeKuoD6Init,
                dim256JoeKuoD6Init,
                dim257JoeKuoD6Init,
                dim258JoeKuoD6Init,
                dim259JoeKuoD6Init,
                dim260JoeKuoD6Init,
                dim261JoeKuoD6Init,
                dim262JoeKuoD6Init,
                dim263JoeKuoD6Init,
                dim264JoeKuoD6Init,
                dim265JoeKuoD6Init,
                dim266JoeKuoD6Init,
                dim267JoeKuoD6Init,
                dim268JoeKuoD6Init,
                dim269JoeKuoD6Init,
                dim270JoeKuoD6Init,
                dim271JoeKuoD6Init,
                dim272JoeKuoD6Init,
                dim273JoeKuoD6Init,
                dim274JoeKuoD6Init,
                dim275JoeKuoD6Init,
                dim276JoeKuoD6Init,
                dim277JoeKuoD6Init,
                dim278JoeKuoD6Init,
                dim279JoeKuoD6Init,
                dim280JoeKuoD6Init,
                dim281JoeKuoD6Init,
                dim282JoeKuoD6Init,
                dim283JoeKuoD6Init,
                dim284JoeKuoD6Init,
                dim285JoeKuoD6Init,
                dim286JoeKuoD6Init,
                dim287JoeKuoD6Init,
                dim288JoeKuoD6Init,
                dim289JoeKuoD6Init,
                dim290JoeKuoD6Init,
                dim291JoeKuoD6Init,
                dim292JoeKuoD6Init,
                dim293JoeKuoD6Init,
                dim294JoeKuoD6Init,
                dim295JoeKuoD6Init,
                dim296JoeKuoD6Init,
                dim297JoeKuoD6Init,
                dim298JoeKuoD6Init,
                dim299JoeKuoD6Init,
                dim300JoeKuoD6Init,
                dim301JoeKuoD6Init,
                dim302JoeKuoD6Init,
                dim303JoeKuoD6Init,
                dim304JoeKuoD6Init,
                dim305JoeKuoD6Init,
                dim306JoeKuoD6Init,
                dim307JoeKuoD6Init,
                dim308JoeKuoD6Init,
                dim309JoeKuoD6Init,
                dim310JoeKuoD6Init,
                dim311JoeKuoD6Init,
                dim312JoeKuoD6Init,
                dim313JoeKuoD6Init,
                dim314JoeKuoD6Init,
                dim315JoeKuoD6Init,
                dim316JoeKuoD6Init,
                dim317JoeKuoD6Init,
                dim318JoeKuoD6Init,
                dim319JoeKuoD6Init,
                dim320JoeKuoD6Init,
                dim321JoeKuoD6Init,
                dim322JoeKuoD6Init,
                dim323JoeKuoD6Init,
                dim324JoeKuoD6Init,
                dim325JoeKuoD6Init,
                dim326JoeKuoD6Init,
                dim327JoeKuoD6Init,
                dim328JoeKuoD6Init,
                dim329JoeKuoD6Init,
                dim330JoeKuoD6Init,
                dim331JoeKuoD6Init,
                dim332JoeKuoD6Init,
                dim333JoeKuoD6Init,
                dim334JoeKuoD6Init,
                dim335JoeKuoD6Init,
                dim336JoeKuoD6Init,
                dim337JoeKuoD6Init,
                dim338JoeKuoD6Init,
                dim339JoeKuoD6Init,
                dim340JoeKuoD6Init,
                dim341JoeKuoD6Init,
                dim342JoeKuoD6Init,
                dim343JoeKuoD6Init,
                dim344JoeKuoD6Init,
                dim345JoeKuoD6Init,
                dim346JoeKuoD6Init,
                dim347JoeKuoD6Init,
                dim348JoeKuoD6Init,
                dim349JoeKuoD6Init,
                dim350JoeKuoD6Init,
                dim351JoeKuoD6Init,
                dim352JoeKuoD6Init,
                dim353JoeKuoD6Init,
                dim354JoeKuoD6Init,
                dim355JoeKuoD6Init,
                dim356JoeKuoD6Init,
                dim357JoeKuoD6Init,
                dim358JoeKuoD6Init,
                dim359JoeKuoD6Init,
                dim360JoeKuoD6Init,
                dim361JoeKuoD6Init,
                dim362JoeKuoD6Init,
                dim363JoeKuoD6Init,
                dim364JoeKuoD6Init,
                dim365JoeKuoD6Init,
                dim366JoeKuoD6Init,
                dim367JoeKuoD6Init,
                dim368JoeKuoD6Init,
                dim369JoeKuoD6Init,
                dim370JoeKuoD6Init,
                dim371JoeKuoD6Init,
                dim372JoeKuoD6Init,
                dim373JoeKuoD6Init,
                dim374JoeKuoD6Init,
                dim375JoeKuoD6Init,
                dim376JoeKuoD6Init,
                dim377JoeKuoD6Init,
                dim378JoeKuoD6Init,
                dim379JoeKuoD6Init,
                dim380JoeKuoD6Init,
                dim381JoeKuoD6Init,
                dim382JoeKuoD6Init,
                dim383JoeKuoD6Init,
                dim384JoeKuoD6Init,
                dim385JoeKuoD6Init,
                dim386JoeKuoD6Init,
                dim387JoeKuoD6Init,
                dim388JoeKuoD6Init,
                dim389JoeKuoD6Init,
                dim390JoeKuoD6Init,
                dim391JoeKuoD6Init,
                dim392JoeKuoD6Init,
                dim393JoeKuoD6Init,
                dim394JoeKuoD6Init,
                dim395JoeKuoD6Init,
                dim396JoeKuoD6Init,
                dim397JoeKuoD6Init,
                dim398JoeKuoD6Init,
                dim399JoeKuoD6Init,
                dim400JoeKuoD6Init,
                dim401JoeKuoD6Init,
                dim402JoeKuoD6Init,
                dim403JoeKuoD6Init,
                dim404JoeKuoD6Init,
                dim405JoeKuoD6Init,
                dim406JoeKuoD6Init,
                dim407JoeKuoD6Init,
                dim408JoeKuoD6Init,
                dim409JoeKuoD6Init,
                dim410JoeKuoD6Init,
                dim411JoeKuoD6Init,
                dim412JoeKuoD6Init,
                dim413JoeKuoD6Init,
                dim414JoeKuoD6Init,
                dim415JoeKuoD6Init,
                dim416JoeKuoD6Init,
                dim417JoeKuoD6Init,
                dim418JoeKuoD6Init,
                dim419JoeKuoD6Init,
                dim420JoeKuoD6Init,
                dim421JoeKuoD6Init,
                dim422JoeKuoD6Init,
                dim423JoeKuoD6Init,
                dim424JoeKuoD6Init,
                dim425JoeKuoD6Init,
                dim426JoeKuoD6Init,
                dim427JoeKuoD6Init,
                dim428JoeKuoD6Init,
                dim429JoeKuoD6Init,
                dim430JoeKuoD6Init,
                dim431JoeKuoD6Init,
                dim432JoeKuoD6Init,
                dim433JoeKuoD6Init,
                dim434JoeKuoD6Init,
                dim435JoeKuoD6Init,
                dim436JoeKuoD6Init,
                dim437JoeKuoD6Init,
                dim438JoeKuoD6Init,
                dim439JoeKuoD6Init,
                dim440JoeKuoD6Init,
                dim441JoeKuoD6Init,
                dim442JoeKuoD6Init,
                dim443JoeKuoD6Init,
                dim444JoeKuoD6Init,
                dim445JoeKuoD6Init,
                dim446JoeKuoD6Init,
                dim447JoeKuoD6Init,
                dim448JoeKuoD6Init,
                dim449JoeKuoD6Init,
                dim450JoeKuoD6Init,
                dim451JoeKuoD6Init,
                dim452JoeKuoD6Init,
                dim453JoeKuoD6Init,
                dim454JoeKuoD6Init,
                dim455JoeKuoD6Init,
                dim456JoeKuoD6Init,
                dim457JoeKuoD6Init,
                dim458JoeKuoD6Init,
                dim459JoeKuoD6Init,
                dim460JoeKuoD6Init,
                dim461JoeKuoD6Init,
                dim462JoeKuoD6Init,
                dim463JoeKuoD6Init,
                dim464JoeKuoD6Init,
                dim465JoeKuoD6Init,
                dim466JoeKuoD6Init,
                dim467JoeKuoD6Init,
                dim468JoeKuoD6Init,
                dim469JoeKuoD6Init,
                dim470JoeKuoD6Init,
                dim471JoeKuoD6Init,
                dim472JoeKuoD6Init,
                dim473JoeKuoD6Init,
                dim474JoeKuoD6Init,
                dim475JoeKuoD6Init,
                dim476JoeKuoD6Init,
                dim477JoeKuoD6Init,
                dim478JoeKuoD6Init,
                dim479JoeKuoD6Init,
                dim480JoeKuoD6Init,
                dim481JoeKuoD6Init,
                dim482JoeKuoD6Init,
                dim483JoeKuoD6Init,
                dim484JoeKuoD6Init,
                dim485JoeKuoD6Init,
                dim486JoeKuoD6Init,
                dim487JoeKuoD6Init,
                dim488JoeKuoD6Init,
                dim489JoeKuoD6Init,
                dim490JoeKuoD6Init,
                dim491JoeKuoD6Init,
                dim492JoeKuoD6Init,
                dim493JoeKuoD6Init,
                dim494JoeKuoD6Init,
                dim495JoeKuoD6Init,
                dim496JoeKuoD6Init,
                dim497JoeKuoD6Init,
                dim498JoeKuoD6Init,
                dim499JoeKuoD6Init,
                dim500JoeKuoD6Init,
                dim501JoeKuoD6Init,
                dim502JoeKuoD6Init,
                dim503JoeKuoD6Init,
                dim504JoeKuoD6Init,
                dim505JoeKuoD6Init,
                dim506JoeKuoD6Init,
                dim507JoeKuoD6Init,
                dim508JoeKuoD6Init,
                dim509JoeKuoD6Init,
                dim510JoeKuoD6Init,
                dim511JoeKuoD6Init,
                dim512JoeKuoD6Init,
                dim513JoeKuoD6Init,
                dim514JoeKuoD6Init,
                dim515JoeKuoD6Init,
                dim516JoeKuoD6Init,
                dim517JoeKuoD6Init,
                dim518JoeKuoD6Init,
                dim519JoeKuoD6Init,
                dim520JoeKuoD6Init,
                dim521JoeKuoD6Init,
                dim522JoeKuoD6Init,
                dim523JoeKuoD6Init,
                dim524JoeKuoD6Init,
                dim525JoeKuoD6Init,
                dim526JoeKuoD6Init,
                dim527JoeKuoD6Init,
                dim528JoeKuoD6Init,
                dim529JoeKuoD6Init,
                dim530JoeKuoD6Init,
                dim531JoeKuoD6Init,
                dim532JoeKuoD6Init,
                dim533JoeKuoD6Init,
                dim534JoeKuoD6Init,
                dim535JoeKuoD6Init,
                dim536JoeKuoD6Init,
                dim537JoeKuoD6Init,
                dim538JoeKuoD6Init,
                dim539JoeKuoD6Init,
                dim540JoeKuoD6Init,
                dim541JoeKuoD6Init,
                dim542JoeKuoD6Init,
                dim543JoeKuoD6Init,
                dim544JoeKuoD6Init,
                dim545JoeKuoD6Init,
                dim546JoeKuoD6Init,
                dim547JoeKuoD6Init,
                dim548JoeKuoD6Init,
                dim549JoeKuoD6Init,
                dim550JoeKuoD6Init,
                dim551JoeKuoD6Init,
                dim552JoeKuoD6Init,
                dim553JoeKuoD6Init,
                dim554JoeKuoD6Init,
                dim555JoeKuoD6Init,
                dim556JoeKuoD6Init,
                dim557JoeKuoD6Init,
                dim558JoeKuoD6Init,
                dim559JoeKuoD6Init,
                dim560JoeKuoD6Init,
                dim561JoeKuoD6Init,
                dim562JoeKuoD6Init,
                dim563JoeKuoD6Init,
                dim564JoeKuoD6Init,
                dim565JoeKuoD6Init,
                dim566JoeKuoD6Init,
                dim567JoeKuoD6Init,
                dim568JoeKuoD6Init,
                dim569JoeKuoD6Init,
                dim570JoeKuoD6Init,
                dim571JoeKuoD6Init,
                dim572JoeKuoD6Init,
                dim573JoeKuoD6Init,
                dim574JoeKuoD6Init,
                dim575JoeKuoD6Init,
                dim576JoeKuoD6Init,
                dim577JoeKuoD6Init,
                dim578JoeKuoD6Init,
                dim579JoeKuoD6Init,
                dim580JoeKuoD6Init,
                dim581JoeKuoD6Init,
                dim582JoeKuoD6Init,
                dim583JoeKuoD6Init,
                dim584JoeKuoD6Init,
                dim585JoeKuoD6Init,
                dim586JoeKuoD6Init,
                dim587JoeKuoD6Init,
                dim588JoeKuoD6Init,
                dim589JoeKuoD6Init,
                dim590JoeKuoD6Init,
                dim591JoeKuoD6Init,
                dim592JoeKuoD6Init,
                dim593JoeKuoD6Init,
                dim594JoeKuoD6Init,
                dim595JoeKuoD6Init,
                dim596JoeKuoD6Init,
                dim597JoeKuoD6Init,
                dim598JoeKuoD6Init,
                dim599JoeKuoD6Init,
                dim600JoeKuoD6Init,
                dim601JoeKuoD6Init,
                dim602JoeKuoD6Init,
                dim603JoeKuoD6Init,
                dim604JoeKuoD6Init,
                dim605JoeKuoD6Init,
                dim606JoeKuoD6Init,
                dim607JoeKuoD6Init,
                dim608JoeKuoD6Init,
                dim609JoeKuoD6Init,
                dim610JoeKuoD6Init,
                dim611JoeKuoD6Init,
                dim612JoeKuoD6Init,
                dim613JoeKuoD6Init,
                dim614JoeKuoD6Init,
                dim615JoeKuoD6Init,
                dim616JoeKuoD6Init,
                dim617JoeKuoD6Init,
                dim618JoeKuoD6Init,
                dim619JoeKuoD6Init,
                dim620JoeKuoD6Init,
                dim621JoeKuoD6Init,
                dim622JoeKuoD6Init,
                dim623JoeKuoD6Init,
                dim624JoeKuoD6Init,
                dim625JoeKuoD6Init,
                dim626JoeKuoD6Init,
                dim627JoeKuoD6Init,
                dim628JoeKuoD6Init,
                dim629JoeKuoD6Init,
                dim630JoeKuoD6Init,
                dim631JoeKuoD6Init,
                dim632JoeKuoD6Init,
                dim633JoeKuoD6Init,
                dim634JoeKuoD6Init,
                dim635JoeKuoD6Init,
                dim636JoeKuoD6Init,
                dim637JoeKuoD6Init,
                dim638JoeKuoD6Init,
                dim639JoeKuoD6Init,
                dim640JoeKuoD6Init,
                dim641JoeKuoD6Init,
                dim642JoeKuoD6Init,
                dim643JoeKuoD6Init,
                dim644JoeKuoD6Init,
                dim645JoeKuoD6Init,
                dim646JoeKuoD6Init,
                dim647JoeKuoD6Init,
                dim648JoeKuoD6Init,
                dim649JoeKuoD6Init,
                dim650JoeKuoD6Init,
                dim651JoeKuoD6Init,
                dim652JoeKuoD6Init,
                dim653JoeKuoD6Init,
                dim654JoeKuoD6Init,
                dim655JoeKuoD6Init,
                dim656JoeKuoD6Init,
                dim657JoeKuoD6Init,
                dim658JoeKuoD6Init,
                dim659JoeKuoD6Init,
                dim660JoeKuoD6Init,
                dim661JoeKuoD6Init,
                dim662JoeKuoD6Init,
                dim663JoeKuoD6Init,
                dim664JoeKuoD6Init,
                dim665JoeKuoD6Init,
                dim666JoeKuoD6Init,
                dim667JoeKuoD6Init,
                dim668JoeKuoD6Init,
                dim669JoeKuoD6Init,
                dim670JoeKuoD6Init,
                dim671JoeKuoD6Init,
                dim672JoeKuoD6Init,
                dim673JoeKuoD6Init,
                dim674JoeKuoD6Init,
                dim675JoeKuoD6Init,
                dim676JoeKuoD6Init,
                dim677JoeKuoD6Init,
                dim678JoeKuoD6Init,
                dim679JoeKuoD6Init,
                dim680JoeKuoD6Init,
                dim681JoeKuoD6Init,
                dim682JoeKuoD6Init,
                dim683JoeKuoD6Init,
                dim684JoeKuoD6Init,
                dim685JoeKuoD6Init,
                dim686JoeKuoD6Init,
                dim687JoeKuoD6Init,
                dim688JoeKuoD6Init,
                dim689JoeKuoD6Init,
                dim690JoeKuoD6Init,
                dim691JoeKuoD6Init,
                dim692JoeKuoD6Init,
                dim693JoeKuoD6Init,
                dim694JoeKuoD6Init,
                dim695JoeKuoD6Init,
                dim696JoeKuoD6Init,
                dim697JoeKuoD6Init,
                dim698JoeKuoD6Init,
                dim699JoeKuoD6Init,
                dim700JoeKuoD6Init,
                dim701JoeKuoD6Init,
                dim702JoeKuoD6Init,
                dim703JoeKuoD6Init,
                dim704JoeKuoD6Init,
                dim705JoeKuoD6Init,
                dim706JoeKuoD6Init,
                dim707JoeKuoD6Init,
                dim708JoeKuoD6Init,
                dim709JoeKuoD6Init,
                dim710JoeKuoD6Init,
                dim711JoeKuoD6Init,
                dim712JoeKuoD6Init,
                dim713JoeKuoD6Init,
                dim714JoeKuoD6Init,
                dim715JoeKuoD6Init,
                dim716JoeKuoD6Init,
                dim717JoeKuoD6Init,
                dim718JoeKuoD6Init,
                dim719JoeKuoD6Init,
                dim720JoeKuoD6Init,
                dim721JoeKuoD6Init,
                dim722JoeKuoD6Init,
                dim723JoeKuoD6Init,
                dim724JoeKuoD6Init,
                dim725JoeKuoD6Init,
                dim726JoeKuoD6Init,
                dim727JoeKuoD6Init,
                dim728JoeKuoD6Init,
                dim729JoeKuoD6Init,
                dim730JoeKuoD6Init,
                dim731JoeKuoD6Init,
                dim732JoeKuoD6Init,
                dim733JoeKuoD6Init,
                dim734JoeKuoD6Init,
                dim735JoeKuoD6Init,
                dim736JoeKuoD6Init,
                dim737JoeKuoD6Init,
                dim738JoeKuoD6Init,
                dim739JoeKuoD6Init,
                dim740JoeKuoD6Init,
                dim741JoeKuoD6Init,
                dim742JoeKuoD6Init,
                dim743JoeKuoD6Init,
                dim744JoeKuoD6Init,
                dim745JoeKuoD6Init,
                dim746JoeKuoD6Init,
                dim747JoeKuoD6Init,
                dim748JoeKuoD6Init,
                dim749JoeKuoD6Init,
                dim750JoeKuoD6Init,
                dim751JoeKuoD6Init,
                dim752JoeKuoD6Init,
                dim753JoeKuoD6Init,
                dim754JoeKuoD6Init,
                dim755JoeKuoD6Init,
                dim756JoeKuoD6Init,
                dim757JoeKuoD6Init,
                dim758JoeKuoD6Init,
                dim759JoeKuoD6Init,
                dim760JoeKuoD6Init,
                dim761JoeKuoD6Init,
                dim762JoeKuoD6Init,
                dim763JoeKuoD6Init,
                dim764JoeKuoD6Init,
                dim765JoeKuoD6Init,
                dim766JoeKuoD6Init,
                dim767JoeKuoD6Init,
                dim768JoeKuoD6Init,
                dim769JoeKuoD6Init,
                dim770JoeKuoD6Init,
                dim771JoeKuoD6Init,
                dim772JoeKuoD6Init,
                dim773JoeKuoD6Init,
                dim774JoeKuoD6Init,
                dim775JoeKuoD6Init,
                dim776JoeKuoD6Init,
                dim777JoeKuoD6Init,
                dim778JoeKuoD6Init,
                dim779JoeKuoD6Init,
                dim780JoeKuoD6Init,
                dim781JoeKuoD6Init,
                dim782JoeKuoD6Init,
                dim783JoeKuoD6Init,
                dim784JoeKuoD6Init,
                dim785JoeKuoD6Init,
                dim786JoeKuoD6Init,
                dim787JoeKuoD6Init,
                dim788JoeKuoD6Init,
                dim789JoeKuoD6Init,
                dim790JoeKuoD6Init,
                dim791JoeKuoD6Init,
                dim792JoeKuoD6Init,
                dim793JoeKuoD6Init,
                dim794JoeKuoD6Init,
                dim795JoeKuoD6Init,
                dim796JoeKuoD6Init,
                dim797JoeKuoD6Init,
                dim798JoeKuoD6Init,
                dim799JoeKuoD6Init,
                dim800JoeKuoD6Init,
                dim801JoeKuoD6Init,
                dim802JoeKuoD6Init,
                dim803JoeKuoD6Init,
                dim804JoeKuoD6Init,
                dim805JoeKuoD6Init,
                dim806JoeKuoD6Init,
                dim807JoeKuoD6Init,
                dim808JoeKuoD6Init,
                dim809JoeKuoD6Init,
                dim810JoeKuoD6Init,
                dim811JoeKuoD6Init,
                dim812JoeKuoD6Init,
                dim813JoeKuoD6Init,
                dim814JoeKuoD6Init,
                dim815JoeKuoD6Init,
                dim816JoeKuoD6Init,
                dim817JoeKuoD6Init,
                dim818JoeKuoD6Init,
                dim819JoeKuoD6Init,
                dim820JoeKuoD6Init,
                dim821JoeKuoD6Init,
                dim822JoeKuoD6Init,
                dim823JoeKuoD6Init,
                dim824JoeKuoD6Init,
                dim825JoeKuoD6Init,
                dim826JoeKuoD6Init,
                dim827JoeKuoD6Init,
                dim828JoeKuoD6Init,
                dim829JoeKuoD6Init,
                dim830JoeKuoD6Init,
                dim831JoeKuoD6Init,
                dim832JoeKuoD6Init,
                dim833JoeKuoD6Init,
                dim834JoeKuoD6Init,
                dim835JoeKuoD6Init,
                dim836JoeKuoD6Init,
                dim837JoeKuoD6Init,
                dim838JoeKuoD6Init,
                dim839JoeKuoD6Init,
                dim840JoeKuoD6Init,
                dim841JoeKuoD6Init,
                dim842JoeKuoD6Init,
                dim843JoeKuoD6Init,
                dim844JoeKuoD6Init,
                dim845JoeKuoD6Init,
                dim846JoeKuoD6Init,
                dim847JoeKuoD6Init,
                dim848JoeKuoD6Init,
                dim849JoeKuoD6Init,
                dim850JoeKuoD6Init,
                dim851JoeKuoD6Init,
                dim852JoeKuoD6Init,
                dim853JoeKuoD6Init,
                dim854JoeKuoD6Init,
                dim855JoeKuoD6Init,
                dim856JoeKuoD6Init,
                dim857JoeKuoD6Init,
                dim858JoeKuoD6Init,
                dim859JoeKuoD6Init,
                dim860JoeKuoD6Init,
                dim861JoeKuoD6Init,
                dim862JoeKuoD6Init,
                dim863JoeKuoD6Init,
                dim864JoeKuoD6Init,
                dim865JoeKuoD6Init,
                dim866JoeKuoD6Init,
                dim867JoeKuoD6Init,
                dim868JoeKuoD6Init,
                dim869JoeKuoD6Init,
                dim870JoeKuoD6Init,
                dim871JoeKuoD6Init,
                dim872JoeKuoD6Init,
                dim873JoeKuoD6Init,
                dim874JoeKuoD6Init,
                dim875JoeKuoD6Init,
                dim876JoeKuoD6Init,
                dim877JoeKuoD6Init,
                dim878JoeKuoD6Init,
                dim879JoeKuoD6Init,
                dim880JoeKuoD6Init,
                dim881JoeKuoD6Init,
                dim882JoeKuoD6Init,
                dim883JoeKuoD6Init,
                dim884JoeKuoD6Init,
                dim885JoeKuoD6Init,
                dim886JoeKuoD6Init,
                dim887JoeKuoD6Init,
                dim888JoeKuoD6Init,
                dim889JoeKuoD6Init,
                dim890JoeKuoD6Init,
                dim891JoeKuoD6Init,
                dim892JoeKuoD6Init,
                dim893JoeKuoD6Init,
                dim894JoeKuoD6Init,
                dim895JoeKuoD6Init,
                dim896JoeKuoD6Init,
                dim897JoeKuoD6Init,
                dim898JoeKuoD6Init,
                dim899JoeKuoD6Init,
                dim900JoeKuoD6Init,
                dim901JoeKuoD6Init,
                dim902JoeKuoD6Init,
                dim903JoeKuoD6Init,
                dim904JoeKuoD6Init,
                dim905JoeKuoD6Init,
                dim906JoeKuoD6Init,
                dim907JoeKuoD6Init,
                dim908JoeKuoD6Init,
                dim909JoeKuoD6Init,
                dim910JoeKuoD6Init,
                dim911JoeKuoD6Init,
                dim912JoeKuoD6Init,
                dim913JoeKuoD6Init,
                dim914JoeKuoD6Init,
                dim915JoeKuoD6Init,
                dim916JoeKuoD6Init,
                dim917JoeKuoD6Init,
                dim918JoeKuoD6Init,
                dim919JoeKuoD6Init,
                dim920JoeKuoD6Init,
                dim921JoeKuoD6Init,
                dim922JoeKuoD6Init,
                dim923JoeKuoD6Init,
                dim924JoeKuoD6Init,
                dim925JoeKuoD6Init,
                dim926JoeKuoD6Init,
                dim927JoeKuoD6Init,
                dim928JoeKuoD6Init,
                dim929JoeKuoD6Init,
                dim930JoeKuoD6Init,
                dim931JoeKuoD6Init,
                dim932JoeKuoD6Init,
                dim933JoeKuoD6Init,
                dim934JoeKuoD6Init,
                dim935JoeKuoD6Init,
                dim936JoeKuoD6Init,
                dim937JoeKuoD6Init,
                dim938JoeKuoD6Init,
                dim939JoeKuoD6Init,
                dim940JoeKuoD6Init,
                dim941JoeKuoD6Init,
                dim942JoeKuoD6Init,
                dim943JoeKuoD6Init,
                dim944JoeKuoD6Init,
                dim945JoeKuoD6Init,
                dim946JoeKuoD6Init,
                dim947JoeKuoD6Init,
                dim948JoeKuoD6Init,
                dim949JoeKuoD6Init,
                dim950JoeKuoD6Init,
                dim951JoeKuoD6Init,
                dim952JoeKuoD6Init,
                dim953JoeKuoD6Init,
                dim954JoeKuoD6Init,
                dim955JoeKuoD6Init,
                dim956JoeKuoD6Init,
                dim957JoeKuoD6Init,
                dim958JoeKuoD6Init,
                dim959JoeKuoD6Init,
                dim960JoeKuoD6Init,
                dim961JoeKuoD6Init,
                dim962JoeKuoD6Init,
                dim963JoeKuoD6Init,
                dim964JoeKuoD6Init,
                dim965JoeKuoD6Init,
                dim966JoeKuoD6Init,
                dim967JoeKuoD6Init,
                dim968JoeKuoD6Init,
                dim969JoeKuoD6Init,
                dim970JoeKuoD6Init,
                dim971JoeKuoD6Init,
                dim972JoeKuoD6Init,
                dim973JoeKuoD6Init,
                dim974JoeKuoD6Init,
                dim975JoeKuoD6Init,
                dim976JoeKuoD6Init,
                dim977JoeKuoD6Init,
                dim978JoeKuoD6Init,
                dim979JoeKuoD6Init,
                dim980JoeKuoD6Init,
                dim981JoeKuoD6Init,
                dim982JoeKuoD6Init,
                dim983JoeKuoD6Init,
                dim984JoeKuoD6Init,
                dim985JoeKuoD6Init,
                dim986JoeKuoD6Init,
                dim987JoeKuoD6Init,
                dim988JoeKuoD6Init,
                dim989JoeKuoD6Init,
                dim990JoeKuoD6Init,
                dim991JoeKuoD6Init,
                dim992JoeKuoD6Init,
                dim993JoeKuoD6Init,
                dim994JoeKuoD6Init,
                dim995JoeKuoD6Init,
                dim996JoeKuoD6Init,
                dim997JoeKuoD6Init,
                dim998JoeKuoD6Init,
                dim999JoeKuoD6Init,
                dim1000JoeKuoD6Init,
                dim1001JoeKuoD6Init,
                dim1002JoeKuoD6Init,
                dim1003JoeKuoD6Init,
                dim1004JoeKuoD6Init,
                dim1005JoeKuoD6Init,
                dim1006JoeKuoD6Init,
                dim1007JoeKuoD6Init,
                dim1008JoeKuoD6Init,
                dim1009JoeKuoD6Init,
                dim1010JoeKuoD6Init,
                dim1011JoeKuoD6Init,
                dim1012JoeKuoD6Init,
                dim1013JoeKuoD6Init,
                dim1014JoeKuoD6Init,
                dim1015JoeKuoD6Init,
                dim1016JoeKuoD6Init,
                dim1017JoeKuoD6Init,
                dim1018JoeKuoD6Init,
                dim1019JoeKuoD6Init,
                dim1020JoeKuoD6Init,
                dim1021JoeKuoD6Init,
                dim1022JoeKuoD6Init,
                dim1023JoeKuoD6Init,
                dim1024JoeKuoD6Init,
                dim1025JoeKuoD6Init,
                dim1026JoeKuoD6Init,
                dim1027JoeKuoD6Init,
                dim1028JoeKuoD6Init,
                dim1029JoeKuoD6Init,
                dim1030JoeKuoD6Init,
                dim1031JoeKuoD6Init,
                dim1032JoeKuoD6Init,
                dim1033JoeKuoD6Init,
                dim1034JoeKuoD6Init,
                dim1035JoeKuoD6Init,
                dim1036JoeKuoD6Init,
                dim1037JoeKuoD6Init,
                dim1038JoeKuoD6Init,
                dim1039JoeKuoD6Init,
                dim1040JoeKuoD6Init,
                dim1041JoeKuoD6Init,
                dim1042JoeKuoD6Init,
                dim1043JoeKuoD6Init,
                dim1044JoeKuoD6Init,
                dim1045JoeKuoD6Init,
                dim1046JoeKuoD6Init,
                dim1047JoeKuoD6Init,
                dim1048JoeKuoD6Init,
                dim1049JoeKuoD6Init,
                dim1050JoeKuoD6Init,
                dim1051JoeKuoD6Init,
                dim1052JoeKuoD6Init,
                dim1053JoeKuoD6Init,
                dim1054JoeKuoD6Init,
                dim1055JoeKuoD6Init,
                dim1056JoeKuoD6Init,
                dim1057JoeKuoD6Init,
                dim1058JoeKuoD6Init,
                dim1059JoeKuoD6Init,
                dim1060JoeKuoD6Init,
                dim1061JoeKuoD6Init,
                dim1062JoeKuoD6Init,
                dim1063JoeKuoD6Init,
                dim1064JoeKuoD6Init,
                dim1065JoeKuoD6Init,
                dim1066JoeKuoD6Init,
                dim1067JoeKuoD6Init,
                dim1068JoeKuoD6Init,
                dim1069JoeKuoD6Init,
                dim1070JoeKuoD6Init,
                dim1071JoeKuoD6Init,
                dim1072JoeKuoD6Init,
                dim1073JoeKuoD6Init,
                dim1074JoeKuoD6Init,
                dim1075JoeKuoD6Init,
                dim1076JoeKuoD6Init,
                dim1077JoeKuoD6Init,
                dim1078JoeKuoD6Init,
                dim1079JoeKuoD6Init,
                dim1080JoeKuoD6Init,
                dim1081JoeKuoD6Init,
                dim1082JoeKuoD6Init,
                dim1083JoeKuoD6Init,
                dim1084JoeKuoD6Init,
                dim1085JoeKuoD6Init,
                dim1086JoeKuoD6Init,
                dim1087JoeKuoD6Init,
                dim1088JoeKuoD6Init,
                dim1089JoeKuoD6Init,
                dim1090JoeKuoD6Init,
                dim1091JoeKuoD6Init,
                dim1092JoeKuoD6Init,
                dim1093JoeKuoD6Init,
                dim1094JoeKuoD6Init,
                dim1095JoeKuoD6Init,
                dim1096JoeKuoD6Init,
                dim1097JoeKuoD6Init,
                dim1098JoeKuoD6Init,
                dim1099JoeKuoD6Init,
                dim1100JoeKuoD6Init,
                dim1101JoeKuoD6Init,
                dim1102JoeKuoD6Init,
                dim1103JoeKuoD6Init,
                dim1104JoeKuoD6Init,
                dim1105JoeKuoD6Init,
                dim1106JoeKuoD6Init,
                dim1107JoeKuoD6Init,
                dim1108JoeKuoD6Init,
                dim1109JoeKuoD6Init,
                dim1110JoeKuoD6Init,
                dim1111JoeKuoD6Init,
                dim1112JoeKuoD6Init,
                dim1113JoeKuoD6Init,
                dim1114JoeKuoD6Init,
                dim1115JoeKuoD6Init,
                dim1116JoeKuoD6Init,
                dim1117JoeKuoD6Init,
                dim1118JoeKuoD6Init,
                dim1119JoeKuoD6Init,
                dim1120JoeKuoD6Init,
                dim1121JoeKuoD6Init,
                dim1122JoeKuoD6Init,
                dim1123JoeKuoD6Init,
                dim1124JoeKuoD6Init,
                dim1125JoeKuoD6Init,
                dim1126JoeKuoD6Init,
                dim1127JoeKuoD6Init,
                dim1128JoeKuoD6Init,
                dim1129JoeKuoD6Init,
                dim1130JoeKuoD6Init,
                dim1131JoeKuoD6Init,
                dim1132JoeKuoD6Init,
                dim1133JoeKuoD6Init,
                dim1134JoeKuoD6Init,
                dim1135JoeKuoD6Init,
                dim1136JoeKuoD6Init,
                dim1137JoeKuoD6Init,
                dim1138JoeKuoD6Init,
                dim1139JoeKuoD6Init,
                dim1140JoeKuoD6Init,
                dim1141JoeKuoD6Init,
                dim1142JoeKuoD6Init,
                dim1143JoeKuoD6Init,
                dim1144JoeKuoD6Init,
                dim1145JoeKuoD6Init,
                dim1146JoeKuoD6Init,
                dim1147JoeKuoD6Init,
                dim1148JoeKuoD6Init,
                dim1149JoeKuoD6Init,
                dim1150JoeKuoD6Init,
                dim1151JoeKuoD6Init,
                dim1152JoeKuoD6Init,
                dim1153JoeKuoD6Init,
                dim1154JoeKuoD6Init,
                dim1155JoeKuoD6Init,
                dim1156JoeKuoD6Init,
                dim1157JoeKuoD6Init,
                dim1158JoeKuoD6Init,
                dim1159JoeKuoD6Init,
                dim1160JoeKuoD6Init,
                dim1161JoeKuoD6Init,
                dim1162JoeKuoD6Init,
                dim1163JoeKuoD6Init,
                dim1164JoeKuoD6Init,
                dim1165JoeKuoD6Init,
                dim1166JoeKuoD6Init,
                dim1167JoeKuoD6Init,
                dim1168JoeKuoD6Init,
                dim1169JoeKuoD6Init,
                dim1170JoeKuoD6Init,
                dim1171JoeKuoD6Init,
                dim1172JoeKuoD6Init,
                dim1173JoeKuoD6Init,
                dim1174JoeKuoD6Init,
                dim1175JoeKuoD6Init,
                dim1176JoeKuoD6Init,
                dim1177JoeKuoD6Init,
                dim1178JoeKuoD6Init,
                dim1179JoeKuoD6Init,
                dim1180JoeKuoD6Init,
                dim1181JoeKuoD6Init,
                dim1182JoeKuoD6Init,
                dim1183JoeKuoD6Init,
                dim1184JoeKuoD6Init,
                dim1185JoeKuoD6Init,
                dim1186JoeKuoD6Init,
                dim1187JoeKuoD6Init,
                dim1188JoeKuoD6Init,
                dim1189JoeKuoD6Init,
                dim1190JoeKuoD6Init,
                dim1191JoeKuoD6Init,
                dim1192JoeKuoD6Init,
                dim1193JoeKuoD6Init,
                dim1194JoeKuoD6Init,
                dim1195JoeKuoD6Init,
                dim1196JoeKuoD6Init,
                dim1197JoeKuoD6Init,
                dim1198JoeKuoD6Init,
                dim1199JoeKuoD6Init,
                dim1200JoeKuoD6Init,
                dim1201JoeKuoD6Init,
                dim1202JoeKuoD6Init,
                dim1203JoeKuoD6Init,
                dim1204JoeKuoD6Init,
                dim1205JoeKuoD6Init,
                dim1206JoeKuoD6Init,
                dim1207JoeKuoD6Init,
                dim1208JoeKuoD6Init,
                dim1209JoeKuoD6Init,
                dim1210JoeKuoD6Init,
                dim1211JoeKuoD6Init,
                dim1212JoeKuoD6Init,
                dim1213JoeKuoD6Init,
                dim1214JoeKuoD6Init,
                dim1215JoeKuoD6Init,
                dim1216JoeKuoD6Init,
                dim1217JoeKuoD6Init,
                dim1218JoeKuoD6Init,
                dim1219JoeKuoD6Init,
                dim1220JoeKuoD6Init,
                dim1221JoeKuoD6Init,
                dim1222JoeKuoD6Init,
                dim1223JoeKuoD6Init,
                dim1224JoeKuoD6Init,
                dim1225JoeKuoD6Init,
                dim1226JoeKuoD6Init,
                dim1227JoeKuoD6Init,
                dim1228JoeKuoD6Init,
                dim1229JoeKuoD6Init,
                dim1230JoeKuoD6Init,
                dim1231JoeKuoD6Init,
                dim1232JoeKuoD6Init,
                dim1233JoeKuoD6Init,
                dim1234JoeKuoD6Init,
                dim1235JoeKuoD6Init,
                dim1236JoeKuoD6Init,
                dim1237JoeKuoD6Init,
                dim1238JoeKuoD6Init,
                dim1239JoeKuoD6Init,
                dim1240JoeKuoD6Init,
                dim1241JoeKuoD6Init,
                dim1242JoeKuoD6Init,
                dim1243JoeKuoD6Init,
                dim1244JoeKuoD6Init,
                dim1245JoeKuoD6Init,
                dim1246JoeKuoD6Init,
                dim1247JoeKuoD6Init,
                dim1248JoeKuoD6Init,
                dim1249JoeKuoD6Init,
                dim1250JoeKuoD6Init,
                dim1251JoeKuoD6Init,
                dim1252JoeKuoD6Init,
                dim1253JoeKuoD6Init,
                dim1254JoeKuoD6Init,
                dim1255JoeKuoD6Init,
                dim1256JoeKuoD6Init,
                dim1257JoeKuoD6Init,
                dim1258JoeKuoD6Init,
                dim1259JoeKuoD6Init,
                dim1260JoeKuoD6Init,
                dim1261JoeKuoD6Init,
                dim1262JoeKuoD6Init,
                dim1263JoeKuoD6Init,
                dim1264JoeKuoD6Init,
                dim1265JoeKuoD6Init,
                dim1266JoeKuoD6Init,
                dim1267JoeKuoD6Init,
                dim1268JoeKuoD6Init,
                dim1269JoeKuoD6Init,
                dim1270JoeKuoD6Init,
                dim1271JoeKuoD6Init,
                dim1272JoeKuoD6Init,
                dim1273JoeKuoD6Init,
                dim1274JoeKuoD6Init,
                dim1275JoeKuoD6Init,
                dim1276JoeKuoD6Init,
                dim1277JoeKuoD6Init,
                dim1278JoeKuoD6Init,
                dim1279JoeKuoD6Init,
                dim1280JoeKuoD6Init,
                dim1281JoeKuoD6Init,
                dim1282JoeKuoD6Init,
                dim1283JoeKuoD6Init,
                dim1284JoeKuoD6Init,
                dim1285JoeKuoD6Init,
                dim1286JoeKuoD6Init,
                dim1287JoeKuoD6Init,
                dim1288JoeKuoD6Init,
                dim1289JoeKuoD6Init,
                dim1290JoeKuoD6Init,
                dim1291JoeKuoD6Init,
                dim1292JoeKuoD6Init,
                dim1293JoeKuoD6Init,
                dim1294JoeKuoD6Init,
                dim1295JoeKuoD6Init,
                dim1296JoeKuoD6Init,
                dim1297JoeKuoD6Init,
                dim1298JoeKuoD6Init,
                dim1299JoeKuoD6Init,
                dim1300JoeKuoD6Init,
                dim1301JoeKuoD6Init,
                dim1302JoeKuoD6Init,
                dim1303JoeKuoD6Init,
                dim1304JoeKuoD6Init,
                dim1305JoeKuoD6Init,
                dim1306JoeKuoD6Init,
                dim1307JoeKuoD6Init,
                dim1308JoeKuoD6Init,
                dim1309JoeKuoD6Init,
                dim1310JoeKuoD6Init,
                dim1311JoeKuoD6Init,
                dim1312JoeKuoD6Init,
                dim1313JoeKuoD6Init,
                dim1314JoeKuoD6Init,
                dim1315JoeKuoD6Init,
                dim1316JoeKuoD6Init,
                dim1317JoeKuoD6Init,
                dim1318JoeKuoD6Init,
                dim1319JoeKuoD6Init,
                dim1320JoeKuoD6Init,
                dim1321JoeKuoD6Init,
                dim1322JoeKuoD6Init,
                dim1323JoeKuoD6Init,
                dim1324JoeKuoD6Init,
                dim1325JoeKuoD6Init,
                dim1326JoeKuoD6Init,
                dim1327JoeKuoD6Init,
                dim1328JoeKuoD6Init,
                dim1329JoeKuoD6Init,
                dim1330JoeKuoD6Init,
                dim1331JoeKuoD6Init,
                dim1332JoeKuoD6Init,
                dim1333JoeKuoD6Init,
                dim1334JoeKuoD6Init,
                dim1335JoeKuoD6Init,
                dim1336JoeKuoD6Init,
                dim1337JoeKuoD6Init,
                dim1338JoeKuoD6Init,
                dim1339JoeKuoD6Init,
                dim1340JoeKuoD6Init,
                dim1341JoeKuoD6Init,
                dim1342JoeKuoD6Init,
                dim1343JoeKuoD6Init,
                dim1344JoeKuoD6Init,
                dim1345JoeKuoD6Init,
                dim1346JoeKuoD6Init,
                dim1347JoeKuoD6Init,
                dim1348JoeKuoD6Init,
                dim1349JoeKuoD6Init,
                dim1350JoeKuoD6Init,
                dim1351JoeKuoD6Init,
                dim1352JoeKuoD6Init,
                dim1353JoeKuoD6Init,
                dim1354JoeKuoD6Init,
                dim1355JoeKuoD6Init,
                dim1356JoeKuoD6Init,
                dim1357JoeKuoD6Init,
                dim1358JoeKuoD6Init,
                dim1359JoeKuoD6Init,
                dim1360JoeKuoD6Init,
                dim1361JoeKuoD6Init,
                dim1362JoeKuoD6Init,
                dim1363JoeKuoD6Init,
                dim1364JoeKuoD6Init,
                dim1365JoeKuoD6Init,
                dim1366JoeKuoD6Init,
                dim1367JoeKuoD6Init,
                dim1368JoeKuoD6Init,
                dim1369JoeKuoD6Init,
                dim1370JoeKuoD6Init,
                dim1371JoeKuoD6Init,
                dim1372JoeKuoD6Init,
                dim1373JoeKuoD6Init,
                dim1374JoeKuoD6Init,
                dim1375JoeKuoD6Init,
                dim1376JoeKuoD6Init,
                dim1377JoeKuoD6Init,
                dim1378JoeKuoD6Init,
                dim1379JoeKuoD6Init,
                dim1380JoeKuoD6Init,
                dim1381JoeKuoD6Init,
                dim1382JoeKuoD6Init,
                dim1383JoeKuoD6Init,
                dim1384JoeKuoD6Init,
                dim1385JoeKuoD6Init,
                dim1386JoeKuoD6Init,
                dim1387JoeKuoD6Init,
                dim1388JoeKuoD6Init,
                dim1389JoeKuoD6Init,
                dim1390JoeKuoD6Init,
                dim1391JoeKuoD6Init,
                dim1392JoeKuoD6Init,
                dim1393JoeKuoD6Init,
                dim1394JoeKuoD6Init,
                dim1395JoeKuoD6Init,
                dim1396JoeKuoD6Init,
                dim1397JoeKuoD6Init,
                dim1398JoeKuoD6Init,
                dim1399JoeKuoD6Init,
                dim1400JoeKuoD6Init,
                dim1401JoeKuoD6Init,
                dim1402JoeKuoD6Init,
                dim1403JoeKuoD6Init,
                dim1404JoeKuoD6Init,
                dim1405JoeKuoD6Init,
                dim1406JoeKuoD6Init,
                dim1407JoeKuoD6Init,
                dim1408JoeKuoD6Init,
                dim1409JoeKuoD6Init,
                dim1410JoeKuoD6Init,
                dim1411JoeKuoD6Init,
                dim1412JoeKuoD6Init,
                dim1413JoeKuoD6Init,
                dim1414JoeKuoD6Init,
                dim1415JoeKuoD6Init,
                dim1416JoeKuoD6Init,
                dim1417JoeKuoD6Init,
                dim1418JoeKuoD6Init,
                dim1419JoeKuoD6Init,
                dim1420JoeKuoD6Init,
                dim1421JoeKuoD6Init,
                dim1422JoeKuoD6Init,
                dim1423JoeKuoD6Init,
                dim1424JoeKuoD6Init,
                dim1425JoeKuoD6Init,
                dim1426JoeKuoD6Init,
                dim1427JoeKuoD6Init,
                dim1428JoeKuoD6Init,
                dim1429JoeKuoD6Init,
                dim1430JoeKuoD6Init,
                dim1431JoeKuoD6Init,
                dim1432JoeKuoD6Init,
                dim1433JoeKuoD6Init,
                dim1434JoeKuoD6Init,
                dim1435JoeKuoD6Init,
                dim1436JoeKuoD6Init,
                dim1437JoeKuoD6Init,
                dim1438JoeKuoD6Init,
                dim1439JoeKuoD6Init,
                dim1440JoeKuoD6Init,
                dim1441JoeKuoD6Init,
                dim1442JoeKuoD6Init,
                dim1443JoeKuoD6Init,
                dim1444JoeKuoD6Init,
                dim1445JoeKuoD6Init,
                dim1446JoeKuoD6Init,
                dim1447JoeKuoD6Init,
                dim1448JoeKuoD6Init,
                dim1449JoeKuoD6Init,
                dim1450JoeKuoD6Init,
                dim1451JoeKuoD6Init,
                dim1452JoeKuoD6Init,
                dim1453JoeKuoD6Init,
                dim1454JoeKuoD6Init,
                dim1455JoeKuoD6Init,
                dim1456JoeKuoD6Init,
                dim1457JoeKuoD6Init,
                dim1458JoeKuoD6Init,
                dim1459JoeKuoD6Init,
                dim1460JoeKuoD6Init,
                dim1461JoeKuoD6Init,
                dim1462JoeKuoD6Init,
                dim1463JoeKuoD6Init,
                dim1464JoeKuoD6Init,
                dim1465JoeKuoD6Init,
                dim1466JoeKuoD6Init,
                dim1467JoeKuoD6Init,
                dim1468JoeKuoD6Init,
                dim1469JoeKuoD6Init,
                dim1470JoeKuoD6Init,
                dim1471JoeKuoD6Init,
                dim1472JoeKuoD6Init,
                dim1473JoeKuoD6Init,
                dim1474JoeKuoD6Init,
                dim1475JoeKuoD6Init,
                dim1476JoeKuoD6Init,
                dim1477JoeKuoD6Init,
                dim1478JoeKuoD6Init,
                dim1479JoeKuoD6Init,
                dim1480JoeKuoD6Init,
                dim1481JoeKuoD6Init,
                dim1482JoeKuoD6Init,
                dim1483JoeKuoD6Init,
                dim1484JoeKuoD6Init,
                dim1485JoeKuoD6Init,
                dim1486JoeKuoD6Init,
                dim1487JoeKuoD6Init,
                dim1488JoeKuoD6Init,
                dim1489JoeKuoD6Init,
                dim1490JoeKuoD6Init,
                dim1491JoeKuoD6Init,
                dim1492JoeKuoD6Init,
                dim1493JoeKuoD6Init,
                dim1494JoeKuoD6Init,
                dim1495JoeKuoD6Init,
                dim1496JoeKuoD6Init,
                dim1497JoeKuoD6Init,
                dim1498JoeKuoD6Init,
                dim1499JoeKuoD6Init,
                dim1500JoeKuoD6Init,
                dim1501JoeKuoD6Init,
                dim1502JoeKuoD6Init,
                dim1503JoeKuoD6Init,
                dim1504JoeKuoD6Init,
                dim1505JoeKuoD6Init,
                dim1506JoeKuoD6Init,
                dim1507JoeKuoD6Init,
                dim1508JoeKuoD6Init,
                dim1509JoeKuoD6Init,
                dim1510JoeKuoD6Init,
                dim1511JoeKuoD6Init,
                dim1512JoeKuoD6Init,
                dim1513JoeKuoD6Init,
                dim1514JoeKuoD6Init,
                dim1515JoeKuoD6Init,
                dim1516JoeKuoD6Init,
                dim1517JoeKuoD6Init,
                dim1518JoeKuoD6Init,
                dim1519JoeKuoD6Init,
                dim1520JoeKuoD6Init,
                dim1521JoeKuoD6Init,
                dim1522JoeKuoD6Init,
                dim1523JoeKuoD6Init,
                dim1524JoeKuoD6Init,
                dim1525JoeKuoD6Init,
                dim1526JoeKuoD6Init,
                dim1527JoeKuoD6Init,
                dim1528JoeKuoD6Init,
                dim1529JoeKuoD6Init,
                dim1530JoeKuoD6Init,
                dim1531JoeKuoD6Init,
                dim1532JoeKuoD6Init,
                dim1533JoeKuoD6Init,
                dim1534JoeKuoD6Init,
                dim1535JoeKuoD6Init,
                dim1536JoeKuoD6Init,
                dim1537JoeKuoD6Init,
                dim1538JoeKuoD6Init,
                dim1539JoeKuoD6Init,
                dim1540JoeKuoD6Init,
                dim1541JoeKuoD6Init,
                dim1542JoeKuoD6Init,
                dim1543JoeKuoD6Init,
                dim1544JoeKuoD6Init,
                dim1545JoeKuoD6Init,
                dim1546JoeKuoD6Init,
                dim1547JoeKuoD6Init,
                dim1548JoeKuoD6Init,
                dim1549JoeKuoD6Init,
                dim1550JoeKuoD6Init,
                dim1551JoeKuoD6Init,
                dim1552JoeKuoD6Init,
                dim1553JoeKuoD6Init,
                dim1554JoeKuoD6Init,
                dim1555JoeKuoD6Init,
                dim1556JoeKuoD6Init,
                dim1557JoeKuoD6Init,
                dim1558JoeKuoD6Init,
                dim1559JoeKuoD6Init,
                dim1560JoeKuoD6Init,
                dim1561JoeKuoD6Init,
                dim1562JoeKuoD6Init,
                dim1563JoeKuoD6Init,
                dim1564JoeKuoD6Init,
                dim1565JoeKuoD6Init,
                dim1566JoeKuoD6Init,
                dim1567JoeKuoD6Init,
                dim1568JoeKuoD6Init,
                dim1569JoeKuoD6Init,
                dim1570JoeKuoD6Init,
                dim1571JoeKuoD6Init,
                dim1572JoeKuoD6Init,
                dim1573JoeKuoD6Init,
                dim1574JoeKuoD6Init,
                dim1575JoeKuoD6Init,
                dim1576JoeKuoD6Init,
                dim1577JoeKuoD6Init,
                dim1578JoeKuoD6Init,
                dim1579JoeKuoD6Init,
                dim1580JoeKuoD6Init,
                dim1581JoeKuoD6Init,
                dim1582JoeKuoD6Init,
                dim1583JoeKuoD6Init,
                dim1584JoeKuoD6Init,
                dim1585JoeKuoD6Init,
                dim1586JoeKuoD6Init,
                dim1587JoeKuoD6Init,
                dim1588JoeKuoD6Init,
                dim1589JoeKuoD6Init,
                dim1590JoeKuoD6Init,
                dim1591JoeKuoD6Init,
                dim1592JoeKuoD6Init,
                dim1593JoeKuoD6Init,
                dim1594JoeKuoD6Init,
                dim1595JoeKuoD6Init,
                dim1596JoeKuoD6Init,
                dim1597JoeKuoD6Init,
                dim1598JoeKuoD6Init,
                dim1599JoeKuoD6Init,
                dim1600JoeKuoD6Init,
                dim1601JoeKuoD6Init,
                dim1602JoeKuoD6Init,
                dim1603JoeKuoD6Init,
                dim1604JoeKuoD6Init,
                dim1605JoeKuoD6Init,
                dim1606JoeKuoD6Init,
                dim1607JoeKuoD6Init,
                dim1608JoeKuoD6Init,
                dim1609JoeKuoD6Init,
                dim1610JoeKuoD6Init,
                dim1611JoeKuoD6Init,
                dim1612JoeKuoD6Init,
                dim1613JoeKuoD6Init,
                dim1614JoeKuoD6Init,
                dim1615JoeKuoD6Init,
                dim1616JoeKuoD6Init,
                dim1617JoeKuoD6Init,
                dim1618JoeKuoD6Init,
                dim1619JoeKuoD6Init,
                dim1620JoeKuoD6Init,
                dim1621JoeKuoD6Init,
                dim1622JoeKuoD6Init,
                dim1623JoeKuoD6Init,
                dim1624JoeKuoD6Init,
                dim1625JoeKuoD6Init,
                dim1626JoeKuoD6Init,
                dim1627JoeKuoD6Init,
                dim1628JoeKuoD6Init,
                dim1629JoeKuoD6Init,
                dim1630JoeKuoD6Init,
                dim1631JoeKuoD6Init,
                dim1632JoeKuoD6Init,
                dim1633JoeKuoD6Init,
                dim1634JoeKuoD6Init,
                dim1635JoeKuoD6Init,
                dim1636JoeKuoD6Init,
                dim1637JoeKuoD6Init,
                dim1638JoeKuoD6Init,
                dim1639JoeKuoD6Init,
                dim1640JoeKuoD6Init,
                dim1641JoeKuoD6Init,
                dim1642JoeKuoD6Init,
                dim1643JoeKuoD6Init,
                dim1644JoeKuoD6Init,
                dim1645JoeKuoD6Init,
                dim1646JoeKuoD6Init,
                dim1647JoeKuoD6Init,
                dim1648JoeKuoD6Init,
                dim1649JoeKuoD6Init,
                dim1650JoeKuoD6Init,
                dim1651JoeKuoD6Init,
                dim1652JoeKuoD6Init,
                dim1653JoeKuoD6Init,
                dim1654JoeKuoD6Init,
                dim1655JoeKuoD6Init,
                dim1656JoeKuoD6Init,
                dim1657JoeKuoD6Init,
                dim1658JoeKuoD6Init,
                dim1659JoeKuoD6Init,
                dim1660JoeKuoD6Init,
                dim1661JoeKuoD6Init,
                dim1662JoeKuoD6Init,
                dim1663JoeKuoD6Init,
                dim1664JoeKuoD6Init,
                dim1665JoeKuoD6Init,
                dim1666JoeKuoD6Init,
                dim1667JoeKuoD6Init,
                dim1668JoeKuoD6Init,
                dim1669JoeKuoD6Init,
                dim1670JoeKuoD6Init,
                dim1671JoeKuoD6Init,
                dim1672JoeKuoD6Init,
                dim1673JoeKuoD6Init,
                dim1674JoeKuoD6Init,
                dim1675JoeKuoD6Init,
                dim1676JoeKuoD6Init,
                dim1677JoeKuoD6Init,
                dim1678JoeKuoD6Init,
                dim1679JoeKuoD6Init,
                dim1680JoeKuoD6Init,
                dim1681JoeKuoD6Init,
                dim1682JoeKuoD6Init,
                dim1683JoeKuoD6Init,
                dim1684JoeKuoD6Init,
                dim1685JoeKuoD6Init,
                dim1686JoeKuoD6Init,
                dim1687JoeKuoD6Init,
                dim1688JoeKuoD6Init,
                dim1689JoeKuoD6Init,
                dim1690JoeKuoD6Init,
                dim1691JoeKuoD6Init,
                dim1692JoeKuoD6Init,
                dim1693JoeKuoD6Init,
                dim1694JoeKuoD6Init,
                dim1695JoeKuoD6Init,
                dim1696JoeKuoD6Init,
                dim1697JoeKuoD6Init,
                dim1698JoeKuoD6Init,
                dim1699JoeKuoD6Init,
                dim1700JoeKuoD6Init,
                dim1701JoeKuoD6Init,
                dim1702JoeKuoD6Init,
                dim1703JoeKuoD6Init,
                dim1704JoeKuoD6Init,
                dim1705JoeKuoD6Init,
                dim1706JoeKuoD6Init,
                dim1707JoeKuoD6Init,
                dim1708JoeKuoD6Init,
                dim1709JoeKuoD6Init,
                dim1710JoeKuoD6Init,
                dim1711JoeKuoD6Init,
                dim1712JoeKuoD6Init,
                dim1713JoeKuoD6Init,
                dim1714JoeKuoD6Init,
                dim1715JoeKuoD6Init,
                dim1716JoeKuoD6Init,
                dim1717JoeKuoD6Init,
                dim1718JoeKuoD6Init,
                dim1719JoeKuoD6Init,
                dim1720JoeKuoD6Init,
                dim1721JoeKuoD6Init,
                dim1722JoeKuoD6Init,
                dim1723JoeKuoD6Init,
                dim1724JoeKuoD6Init,
                dim1725JoeKuoD6Init,
                dim1726JoeKuoD6Init,
                dim1727JoeKuoD6Init,
                dim1728JoeKuoD6Init,
                dim1729JoeKuoD6Init,
                dim1730JoeKuoD6Init,
                dim1731JoeKuoD6Init,
                dim1732JoeKuoD6Init,
                dim1733JoeKuoD6Init,
                dim1734JoeKuoD6Init,
                dim1735JoeKuoD6Init,
                dim1736JoeKuoD6Init,
                dim1737JoeKuoD6Init,
                dim1738JoeKuoD6Init,
                dim1739JoeKuoD6Init,
                dim1740JoeKuoD6Init,
                dim1741JoeKuoD6Init,
                dim1742JoeKuoD6Init,
                dim1743JoeKuoD6Init,
                dim1744JoeKuoD6Init,
                dim1745JoeKuoD6Init,
                dim1746JoeKuoD6Init,
                dim1747JoeKuoD6Init,
                dim1748JoeKuoD6Init,
                dim1749JoeKuoD6Init,
                dim1750JoeKuoD6Init,
                dim1751JoeKuoD6Init,
                dim1752JoeKuoD6Init,
                dim1753JoeKuoD6Init,
                dim1754JoeKuoD6Init,
                dim1755JoeKuoD6Init,
                dim1756JoeKuoD6Init,
                dim1757JoeKuoD6Init,
                dim1758JoeKuoD6Init,
                dim1759JoeKuoD6Init,
                dim1760JoeKuoD6Init,
                dim1761JoeKuoD6Init,
                dim1762JoeKuoD6Init,
                dim1763JoeKuoD6Init,
                dim1764JoeKuoD6Init,
                dim1765JoeKuoD6Init,
                dim1766JoeKuoD6Init,
                dim1767JoeKuoD6Init,
                dim1768JoeKuoD6Init,
                dim1769JoeKuoD6Init,
                dim1770JoeKuoD6Init,
                dim1771JoeKuoD6Init,
                dim1772JoeKuoD6Init,
                dim1773JoeKuoD6Init,
                dim1774JoeKuoD6Init,
                dim1775JoeKuoD6Init,
                dim1776JoeKuoD6Init,
                dim1777JoeKuoD6Init,
                dim1778JoeKuoD6Init,
                dim1779JoeKuoD6Init,
                dim1780JoeKuoD6Init,
                dim1781JoeKuoD6Init,
                dim1782JoeKuoD6Init,
                dim1783JoeKuoD6Init,
                dim1784JoeKuoD6Init,
                dim1785JoeKuoD6Init,
                dim1786JoeKuoD6Init,
                dim1787JoeKuoD6Init,
                dim1788JoeKuoD6Init,
                dim1789JoeKuoD6Init,
                dim1790JoeKuoD6Init,
                dim1791JoeKuoD6Init,
                dim1792JoeKuoD6Init,
                dim1793JoeKuoD6Init,
                dim1794JoeKuoD6Init,
                dim1795JoeKuoD6Init,
                dim1796JoeKuoD6Init,
                dim1797JoeKuoD6Init,
                dim1798JoeKuoD6Init,
                dim1799JoeKuoD6Init
            };

            const unsigned long dim1JoeKuoD7Init[]  =   {   1   ,0 };
            const unsigned long dim2JoeKuoD7Init[]  =   {   1   ,   3   ,0 };
            const unsigned long dim3JoeKuoD7Init[]  =   {   1   ,   3   ,   1   ,0 };
            const unsigned long dim4JoeKuoD7Init[]  =   {   1   ,   1   ,   1   ,0 };
            const unsigned long dim5JoeKuoD7Init[]  =   {   1   ,   1   ,   3   ,   3   ,0 };
            const unsigned long dim6JoeKuoD7Init[]  =   {   1   ,   3   ,   5   ,   13  ,0 };
            const unsigned long dim7JoeKuoD7Init[]  =   {   1   ,   1   ,   5   ,   5   ,   17  ,0 };
            const unsigned long dim8JoeKuoD7Init[]  =   {   1   ,   1   ,   5   ,   5   ,   5   ,0 };
            const unsigned long dim9JoeKuoD7Init[]  =   {   1   ,   1   ,   7   ,   11  ,   19  ,0 };
            const unsigned long dim10JoeKuoD7Init[] =   {   1   ,   1   ,   5   ,   1   ,   1   ,0 };
            const unsigned long dim11JoeKuoD7Init[] =   {   1   ,   1   ,   1   ,   3   ,   11  ,0 };
            const unsigned long dim12JoeKuoD7Init[] =   {   1   ,   3   ,   5   ,   5   ,   31  ,0 };
            const unsigned long dim13JoeKuoD7Init[] =   {   1   ,   3   ,   3   ,   9   ,   7   ,   49  ,0 };
            const unsigned long dim14JoeKuoD7Init[] =   {   1   ,   1   ,   1   ,   15  ,   23  ,   17  ,0 };
            const unsigned long dim15JoeKuoD7Init[] =   {   1   ,   3   ,   1   ,   13  ,   27  ,   49  ,0 };
            const unsigned long dim16JoeKuoD7Init[] =   {   1   ,   1   ,   1   ,   15  ,   7   ,   5   ,0 };
            const unsigned long dim17JoeKuoD7Init[] =   {   1   ,   3   ,   7   ,   7   ,   19  ,   13  ,0 };
            const unsigned long dim18JoeKuoD7Init[] =   {   1   ,   1   ,   5   ,   5   ,   19  ,   59  ,0 };
            const unsigned long dim19JoeKuoD7Init[] =   {   1   ,   3   ,   7   ,   11  ,   23  ,   7   ,   121 ,0 };
            const unsigned long dim20JoeKuoD7Init[] =   {   1   ,   3   ,   5   ,   13  ,   9   ,   19  ,   113 ,0 };
            const unsigned long dim21JoeKuoD7Init[] =   {   1   ,   1   ,   7   ,   15  ,   3   ,   53  ,   87  ,0 };
            const unsigned long dim22JoeKuoD7Init[] =   {   1   ,   3   ,   7   ,   7   ,   25  ,   33  ,   41  ,0 };
            const unsigned long dim23JoeKuoD7Init[] =   {   1   ,   3   ,   1   ,   9   ,   15  ,   47  ,   85  ,0 };
            const unsigned long dim24JoeKuoD7Init[] =   {   1   ,   1   ,   7   ,   3   ,   5   ,   25  ,   37  ,0 };
            const unsigned long dim25JoeKuoD7Init[] =   {   1   ,   3   ,   7   ,   9   ,   31  ,   11  ,   27  ,0 };
            const unsigned long dim26JoeKuoD7Init[] =   {   1   ,   1   ,   7   ,   15  ,   1   ,   17  ,   121 ,0 };
            const unsigned long dim27JoeKuoD7Init[] =   {   1   ,   3   ,   7   ,   13  ,   17  ,   49  ,   5   ,0 };
            const unsigned long dim28JoeKuoD7Init[] =   {   1   ,   3   ,   5   ,   9   ,   7   ,   9   ,   31  ,0 };
            const unsigned long dim29JoeKuoD7Init[] =   {   1   ,   3   ,   1   ,   9   ,   7   ,   29  ,   15  ,0 };
            const unsigned long dim30JoeKuoD7Init[] =   {   1   ,   3   ,   1   ,   1   ,   23  ,   43  ,   37  ,0 };
            const unsigned long dim31JoeKuoD7Init[] =   {   1   ,   3   ,   7   ,   5   ,   3   ,   43  ,   95  ,0 };
            const unsigned long dim32JoeKuoD7Init[] =   {   1   ,   1   ,   7   ,   9   ,   31  ,   25  ,   91  ,0 };
            const unsigned long dim33JoeKuoD7Init[] =   {   1   ,   3   ,   1   ,   11  ,   15  ,   17  ,   91  ,0 };
            const unsigned long dim34JoeKuoD7Init[] =   {   1   ,   1   ,   1   ,   1   ,   21  ,   29  ,   71  ,0 };
            const unsigned long dim35JoeKuoD7Init[] =   {   1   ,   1   ,   5   ,   13  ,   27  ,   3   ,   27  ,0 };
            const unsigned long dim36JoeKuoD7Init[] =   {   1   ,   3   ,   1   ,   9   ,   31  ,   25  ,   3   ,0 };
            const unsigned long dim37JoeKuoD7Init[] =   {   1   ,   3   ,   3   ,   1   ,   25  ,   23  ,   51  ,   45  ,0 };
            const unsigned long dim38JoeKuoD7Init[] =   {   1   ,   1   ,   5   ,   13  ,   9   ,   25  ,   17  ,   227 ,0 };
            const unsigned long dim39JoeKuoD7Init[] =   {   1   ,   1   ,   1   ,   1   ,   15  ,   9   ,   91  ,   45  ,0 };
            const unsigned long dim40JoeKuoD7Init[] =   {   1   ,   1   ,   7   ,   7   ,   9   ,   35  ,   87  ,   31  ,0 };
            const unsigned long dim41JoeKuoD7Init[] =   {   1   ,   1   ,   5   ,   13  ,   29  ,   33  ,   13  ,   119 ,0 };
            const unsigned long dim42JoeKuoD7Init[] =   {   1   ,   3   ,   5   ,   3   ,   17  ,   9   ,   73  ,   85  ,0 };
            const unsigned long dim43JoeKuoD7Init[] =   {   1   ,   3   ,   7   ,   5   ,   13  ,   33  ,   11  ,   17  ,0 };
            const unsigned long dim44JoeKuoD7Init[] =   {   1   ,   1   ,   3   ,   13  ,   29  ,   25  ,   15  ,   75  ,0 };
            const unsigned long dim45JoeKuoD7Init[] =   {   1   ,   3   ,   7   ,   9   ,   3   ,   11  ,   11  ,   59  ,0 };
            const unsigned long dim46JoeKuoD7Init[] =   {   1   ,   3   ,   1   ,   13  ,   5   ,   27  ,   19  ,   71  ,0 };
            const unsigned long dim47JoeKuoD7Init[] =   {   1   ,   1   ,   7   ,   3   ,   13  ,   41  ,   79  ,   137 ,0 };
            const unsigned long dim48JoeKuoD7Init[] =   {   1   ,   3   ,   3   ,   7   ,   9   ,   41  ,   89  ,   63  ,0 };
            const unsigned long dim49JoeKuoD7Init[] =   {   1   ,   1   ,   1   ,   5   ,   29  ,   57  ,   53  ,   103 ,0 };
            const unsigned long dim50JoeKuoD7Init[] =   {   1   ,   1   ,   3   ,   9   ,   3   ,   41  ,   71  ,   119 ,0 };
            const unsigned long dim51JoeKuoD7Init[] =   {   1   ,   3   ,   5   ,   5   ,   3   ,   1   ,   39  ,   185 ,0 };
            const unsigned long dim52JoeKuoD7Init[] =   {   1   ,   3   ,   3   ,   15  ,   23  ,   3   ,   43  ,   95  ,0 };
            const unsigned long dim53JoeKuoD7Init[] =   {   1   ,   1   ,   3   ,   7   ,   1   ,   35  ,   19  ,   19  ,   127 ,0 };
            const unsigned long dim54JoeKuoD7Init[] =   {   1   ,   1   ,   3   ,   9   ,   9   ,   31  ,   9   ,   157 ,   105 ,0 };
            const unsigned long dim55JoeKuoD7Init[] =   {   1   ,   3   ,   1   ,   1   ,   7   ,   49  ,   13  ,   125 ,   471 ,0 };
            const unsigned long dim56JoeKuoD7Init[] =   {   1   ,   3   ,   3   ,   1   ,   17  ,   25  ,   111 ,   73  ,   41  ,0 };
            const unsigned long dim57JoeKuoD7Init[] =   {   1   ,   3   ,   5   ,   11  ,   3   ,   61  ,   117 ,   93  ,   325 ,0 };
            const unsigned long dim58JoeKuoD7Init[] =   {   1   ,   1   ,   7   ,   15  ,   21  ,   63  ,   103 ,   145 ,   61  ,0 };
            const unsigned long dim59JoeKuoD7Init[] =   {   1   ,   3   ,   7   ,   7   ,   27  ,   49  ,   101 ,   227 ,   55  ,0 };
            const unsigned long dim60JoeKuoD7Init[] =   {   1   ,   1   ,   7   ,   5   ,   1   ,   1   ,   33  ,   225 ,   501 ,0 };
            const unsigned long dim61JoeKuoD7Init[] =   {   1   ,   3   ,   5   ,   5   ,   1   ,   17  ,   113 ,   149 ,   71  ,0 };
            const unsigned long dim62JoeKuoD7Init[] =   {   1   ,   1   ,   5   ,   9   ,   27  ,   9   ,   97  ,   7   ,   35  ,0 };
            const unsigned long dim63JoeKuoD7Init[] =   {   1   ,   3   ,   1   ,   7   ,   13  ,   39  ,   49  ,   213 ,   483 ,0 };
            const unsigned long dim64JoeKuoD7Init[] =   {   1   ,   3   ,   3   ,   11  ,   13  ,   47  ,   43  ,   89  ,   329 ,0 };
            const unsigned long dim65JoeKuoD7Init[] =   {   1   ,   3   ,   1   ,   1   ,   19  ,   61  ,   97  ,   205 ,   285 ,0 };
            const unsigned long dim66JoeKuoD7Init[] =   {   1   ,   1   ,   3   ,   11  ,   11  ,   3   ,   105 ,   151 ,   469 ,0 };
            const unsigned long dim67JoeKuoD7Init[] =   {   1   ,   1   ,   1   ,   1   ,   5   ,   39  ,   89  ,   137 ,   409 ,0 };
            const unsigned long dim68JoeKuoD7Init[] =   {   1   ,   1   ,   3   ,   11  ,   23  ,   39  ,   85  ,   251 ,   141 ,0 };
            const unsigned long dim69JoeKuoD7Init[] =   {   1   ,   1   ,   3   ,   5   ,   23  ,   23  ,   59  ,   111 ,   19  ,0 };
            const unsigned long dim70JoeKuoD7Init[] =   {   1   ,   3   ,   5   ,   9   ,   19  ,   63  ,   1   ,   251 ,   305 ,0 };
            const unsigned long dim71JoeKuoD7Init[] =   {   1   ,   3   ,   1   ,   1   ,   3   ,   35  ,   65  ,   153 ,   387 ,0 };
            const unsigned long dim72JoeKuoD7Init[] =   {   1   ,   1   ,   5   ,   13  ,   21  ,   45  ,   37  ,   255 ,   295 ,0 };
            const unsigned long dim73JoeKuoD7Init[] =   {   1   ,   1   ,   5   ,   1   ,   21  ,   5   ,   39  ,   143 ,   293 ,0 };
            const unsigned long dim74JoeKuoD7Init[] =   {   1   ,   1   ,   1   ,   5   ,   21  ,   45  ,   91  ,   211 ,   347 ,0 };
            const unsigned long dim75JoeKuoD7Init[] =   {   1   ,   3   ,   5   ,   11  ,   17  ,   21  ,   73  ,   185 ,   259 ,0 };
            const unsigned long dim76JoeKuoD7Init[] =   {   1   ,   3   ,   1   ,   11  ,   29  ,   25  ,   85  ,   103 ,   97  ,0 };
            const unsigned long dim77JoeKuoD7Init[] =   {   1   ,   3   ,   5   ,   5   ,   27  ,   37  ,   43  ,   243 ,   257 ,0 };
            const unsigned long dim78JoeKuoD7Init[] =   {   1   ,   3   ,   5   ,   11  ,   31  ,   63  ,   105 ,   41  ,   229 ,0 };
            const unsigned long dim79JoeKuoD7Init[] =   {   1   ,   3   ,   3   ,   15  ,   21  ,   55  ,   29  ,   39  ,   279 ,0 };
            const unsigned long dim80JoeKuoD7Init[] =   {   1   ,   3   ,   5   ,   7   ,   1   ,   47  ,   97  ,   15  ,   173 ,0 };
            const unsigned long dim81JoeKuoD7Init[] =   {   1   ,   1   ,   5   ,   7   ,   21  ,   55  ,   29  ,   93  ,   97  ,0 };
            const unsigned long dim82JoeKuoD7Init[] =   {   1   ,   3   ,   3   ,   3   ,   7   ,   29  ,   27  ,   179 ,   105 ,0 };
            const unsigned long dim83JoeKuoD7Init[] =   {   1   ,   1   ,   3   ,   3   ,   1   ,   43  ,   37  ,   221 ,   259 ,0 };
            const unsigned long dim84JoeKuoD7Init[] =   {   1   ,   1   ,   7   ,   9   ,   31  ,   25  ,   1   ,   47  ,   71  ,0 };
            const unsigned long dim85JoeKuoD7Init[] =   {   1   ,   3   ,   7   ,   7   ,   7   ,   25  ,   37  ,   1   ,   487 ,0 };
            const unsigned long dim86JoeKuoD7Init[] =   {   1   ,   3   ,   1   ,   1   ,   31  ,   43  ,   25  ,   117 ,   147 ,0 };
            const unsigned long dim87JoeKuoD7Init[] =   {   1   ,   3   ,   7   ,   3   ,   27  ,   7   ,   103 ,   241 ,   173 ,0 };
            const unsigned long dim88JoeKuoD7Init[] =   {   1   ,   1   ,   5   ,   1   ,   27  ,   25  ,   91  ,   189 ,   53  ,0 };
            const unsigned long dim89JoeKuoD7Init[] =   {   1   ,   1   ,   3   ,   3   ,   29  ,   39  ,   79  ,   125 ,   281 ,0 };
            const unsigned long dim90JoeKuoD7Init[] =   {   1   ,   1   ,   5   ,   11  ,   13  ,   57  ,   15  ,   219 ,   407 ,0 };
            const unsigned long dim91JoeKuoD7Init[] =   {   1   ,   3   ,   3   ,   5   ,   15  ,   9   ,   65  ,   111 ,   1   ,0 };
            const unsigned long dim92JoeKuoD7Init[] =   {   1   ,   1   ,   5   ,   3   ,   31  ,   45  ,   17  ,   227 ,   67  ,0 };
            const unsigned long dim93JoeKuoD7Init[] =   {   1   ,   1   ,   3   ,   1   ,   31  ,   59  ,   63  ,   205 ,   27  ,0 };
            const unsigned long dim94JoeKuoD7Init[] =   {   1   ,   1   ,   1   ,   9   ,   27  ,   9   ,   49  ,   187 ,   451 ,0 };
            const unsigned long dim95JoeKuoD7Init[] =   {   1   ,   1   ,   5   ,   15  ,   11  ,   43  ,   93  ,   49  ,   129 ,0 };
            const unsigned long dim96JoeKuoD7Init[] =   {   1   ,   3   ,   7   ,   7   ,   5   ,   63  ,   31  ,   239 ,   11  ,0 };
            const unsigned long dim97JoeKuoD7Init[] =   {   1   ,   1   ,   1   ,   11  ,   7   ,   59  ,   37  ,   157 ,   505 ,0 };
            const unsigned long dim98JoeKuoD7Init[] =   {   1   ,   3   ,   5   ,   9   ,   3   ,   41  ,   87  ,   229 ,   269 ,0 };
            const unsigned long dim99JoeKuoD7Init[] =   {   1   ,   1   ,   5   ,   3   ,   25  ,   11  ,   69  ,   125 ,   35  ,0 };
            const unsigned long dim100JoeKuoD7Init[]    =   {   1   ,   3   ,   5   ,   13  ,   11  ,   53  ,   47  ,   21  ,   79  ,0 };
            const unsigned long dim101JoeKuoD7Init[]    =   {   1   ,   3   ,   3   ,   7   ,   11  ,   23  ,   1   ,   99  ,   417 ,   35  ,0 };
            const unsigned long dim102JoeKuoD7Init[]    =   {   1   ,   1   ,   7   ,   7   ,   7   ,   53  ,   57  ,   229 ,   473 ,   533 ,0 };
            const unsigned long dim103JoeKuoD7Init[]    =   {   1   ,   1   ,   3   ,   9   ,   23  ,   31  ,   87  ,   27  ,   313 ,   853 ,0 };
            const unsigned long dim104JoeKuoD7Init[]    =   {   1   ,   3   ,   3   ,   1   ,   13  ,   43  ,   27  ,   227 ,   231 ,   91  ,0 };
            const unsigned long dim105JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   5   ,   11  ,   17  ,   109 ,   29  ,   417 ,   597 ,0 };
            const unsigned long dim106JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   7   ,   31  ,   5   ,   23  ,   137 ,   35  ,   583 ,0 };
            const unsigned long dim107JoeKuoD7Init[]    =   {   1   ,   3   ,   1   ,   5   ,   15  ,   23  ,   87  ,   171 ,   171 ,   889 ,0 };
            const unsigned long dim108JoeKuoD7Init[]    =   {   1   ,   3   ,   3   ,   7   ,   17  ,   21  ,   19  ,   233 ,   275 ,   651 ,0 };
            const unsigned long dim109JoeKuoD7Init[]    =   {   1   ,   1   ,   3   ,   5   ,   31  ,   19  ,   127 ,   115 ,   211 ,   715 ,0 };
            const unsigned long dim110JoeKuoD7Init[]    =   {   1   ,   1   ,   5   ,   3   ,   13  ,   5   ,   37  ,   83  ,   123 ,   567 ,0 };
            const unsigned long dim111JoeKuoD7Init[]    =   {   1   ,   3   ,   1   ,   7   ,   3   ,   41  ,   123 ,   61  ,   311 ,   45  ,0 };
            const unsigned long dim112JoeKuoD7Init[]    =   {   1   ,   1   ,   5   ,   5   ,   19  ,   27  ,   1   ,   159 ,   87  ,   155 ,0 };
            const unsigned long dim113JoeKuoD7Init[]    =   {   1   ,   3   ,   3   ,   15  ,   31  ,   17  ,   29  ,   105 ,   293 ,   205 ,0 };
            const unsigned long dim114JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   13  ,   15  ,   11  ,   125 ,   241 ,   1   ,   755 ,0 };
            const unsigned long dim115JoeKuoD7Init[]    =   {   1   ,   1   ,   7   ,   7   ,   27  ,   47  ,   59  ,   143 ,   451 ,   1005    ,0 };
            const unsigned long dim116JoeKuoD7Init[]    =   {   1   ,   3   ,   1   ,   11  ,   19  ,   37  ,   49  ,   105 ,   127 ,   683 ,0 };
            const unsigned long dim117JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   1   ,   11  ,   31  ,   9   ,   15  ,   355 ,   951 ,0 };
            const unsigned long dim118JoeKuoD7Init[]    =   {   1   ,   1   ,   7   ,   15  ,   5   ,   19  ,   69  ,   51  ,   121 ,   695 ,0 };
            const unsigned long dim119JoeKuoD7Init[]    =   {   1   ,   3   ,   1   ,   9   ,   15  ,   33  ,   125 ,   63  ,   191 ,   773 ,0 };
            const unsigned long dim120JoeKuoD7Init[]    =   {   1   ,   1   ,   3   ,   1   ,   11  ,   13  ,   67  ,   123 ,   143 ,   427 ,0 };
            const unsigned long dim121JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   7   ,   17  ,   45  ,   73  ,   209 ,   501 ,   781 ,0 };
            const unsigned long dim122JoeKuoD7Init[]    =   {   1   ,   1   ,   3   ,   5   ,   21  ,   25  ,   31  ,   229 ,   107 ,   447 ,0 };
            const unsigned long dim123JoeKuoD7Init[]    =   {   1   ,   3   ,   3   ,   5   ,   15  ,   39  ,   19  ,   111 ,   221 ,   35  ,0 };
            const unsigned long dim124JoeKuoD7Init[]    =   {   1   ,   3   ,   5   ,   7   ,   7   ,   23  ,   111 ,   201 ,   469 ,   859 ,0 };
            const unsigned long dim125JoeKuoD7Init[]    =   {   1   ,   3   ,   1   ,   9   ,   7   ,   55  ,   1   ,   57  ,   245 ,   865 ,0 };
            const unsigned long dim126JoeKuoD7Init[]    =   {   1   ,   1   ,   5   ,   11  ,   19  ,   17  ,   63  ,   107 ,   423 ,   977 ,0 };
            const unsigned long dim127JoeKuoD7Init[]    =   {   1   ,   1   ,   3   ,   5   ,   23  ,   35  ,   59  ,   143 ,   403 ,   469 ,0 };
            const unsigned long dim128JoeKuoD7Init[]    =   {   1   ,   3   ,   3   ,   9   ,   29  ,   11  ,   11  ,   137 ,   341 ,   669 ,0 };
            const unsigned long dim129JoeKuoD7Init[]    =   {   1   ,   3   ,   5   ,   5   ,   31  ,   29  ,   73  ,   25  ,   21  ,   543 ,0 };
            const unsigned long dim130JoeKuoD7Init[]    =   {   1   ,   1   ,   5   ,   3   ,   17  ,   47  ,   69  ,   103 ,   49  ,   869 ,0 };
            const unsigned long dim131JoeKuoD7Init[]    =   {   1   ,   1   ,   3   ,   9   ,   1   ,   5   ,   69  ,   45  ,   73  ,   533 ,0 };
            const unsigned long dim132JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   11  ,   13  ,   11  ,   61  ,   71  ,   397 ,   919 ,0 };
            const unsigned long dim133JoeKuoD7Init[]    =   {   1   ,   1   ,   7   ,   11  ,   1   ,   3   ,   21  ,   127 ,   263 ,   597 ,0 };
            const unsigned long dim134JoeKuoD7Init[]    =   {   1   ,   3   ,   1   ,   9   ,   3   ,   31  ,   41  ,   179 ,   499 ,   313 ,0 };
            const unsigned long dim135JoeKuoD7Init[]    =   {   1   ,   1   ,   7   ,   11  ,   5   ,   43  ,   33  ,   41  ,   53  ,   917 ,0 };
            const unsigned long dim136JoeKuoD7Init[]    =   {   1   ,   3   ,   3   ,   9   ,   1   ,   63  ,   35  ,   139 ,   169 ,   945 ,0 };
            const unsigned long dim137JoeKuoD7Init[]    =   {   1   ,   1   ,   7   ,   3   ,   31  ,   13  ,   51  ,   105 ,   333 ,   703 ,0 };
            const unsigned long dim138JoeKuoD7Init[]    =   {   1   ,   1   ,   5   ,   3   ,   5   ,   5   ,   127 ,   103 ,   465 ,   989 ,0 };
            const unsigned long dim139JoeKuoD7Init[]    =   {   1   ,   1   ,   3   ,   1   ,   31  ,   3   ,   87  ,   41  ,   511 ,   713 ,0 };
            const unsigned long dim140JoeKuoD7Init[]    =   {   1   ,   1   ,   7   ,   1   ,   7   ,   29  ,   19  ,   237 ,   235 ,   531 ,0 };
            const unsigned long dim141JoeKuoD7Init[]    =   {   1   ,   1   ,   7   ,   13  ,   21  ,   11  ,   111 ,   51  ,   461 ,   525 ,0 };
            const unsigned long dim142JoeKuoD7Init[]    =   {   1   ,   3   ,   3   ,   13  ,   27  ,   49  ,   123 ,   241 ,   365 ,   137 ,0 };
            const unsigned long dim143JoeKuoD7Init[]    =   {   1   ,   1   ,   7   ,   13  ,   17  ,   43  ,   43  ,   115 ,   483 ,   291 ,0 };
            const unsigned long dim144JoeKuoD7Init[]    =   {   1   ,   1   ,   3   ,   7   ,   5   ,   17  ,   127 ,   167 ,   371 ,   417 ,0 };
            const unsigned long dim145JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   9   ,   11  ,   21  ,   123 ,   29  ,   133 ,   9   ,0 };
            const unsigned long dim146JoeKuoD7Init[]    =   {   1   ,   3   ,   1   ,   7   ,   13  ,   35  ,   99  ,   161 ,   45  ,   347 ,0 };
            const unsigned long dim147JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   7   ,   15  ,   21  ,   31  ,   131 ,   441 ,   635 ,0 };
            const unsigned long dim148JoeKuoD7Init[]    =   {   1   ,   1   ,   3   ,   7   ,   21  ,   63  ,   119 ,   175 ,   183 ,   189 ,0 };
            const unsigned long dim149JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   3   ,   17  ,   39  ,   21  ,   201 ,   31  ,   747 ,0 };
            const unsigned long dim150JoeKuoD7Init[]    =   {   1   ,   1   ,   5   ,   15  ,   7   ,   23  ,   127 ,   169 ,   341 ,   975 ,0 };
            const unsigned long dim151JoeKuoD7Init[]    =   {   1   ,   3   ,   3   ,   15  ,   27  ,   39  ,   67  ,   111 ,   501 ,   343 ,0 };
            const unsigned long dim152JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   15  ,   13  ,   3   ,   99  ,   145 ,   259 ,   579 ,0 };
            const unsigned long dim153JoeKuoD7Init[]    =   {   1   ,   1   ,   5   ,   13  ,   27  ,   45  ,   9   ,   229 ,   139 ,   1015    ,0 };
            const unsigned long dim154JoeKuoD7Init[]    =   {   1   ,   1   ,   3   ,   11  ,   1   ,   61  ,   35  ,   77  ,   423 ,   987 ,0 };
            const unsigned long dim155JoeKuoD7Init[]    =   {   1   ,   1   ,   7   ,   11  ,   13  ,   53  ,   99  ,   165 ,   329 ,   191 ,0 };
            const unsigned long dim156JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   11  ,   13  ,   31  ,   15  ,   237 ,   359 ,   337 ,0 };
            const unsigned long dim157JoeKuoD7Init[]    =   {   1   ,   3   ,   1   ,   1   ,   1   ,   15  ,   35  ,   175 ,   347 ,   485 ,0 };
            const unsigned long dim158JoeKuoD7Init[]    =   {   1   ,   1   ,   5   ,   9   ,   3   ,   15  ,   63  ,   19  ,   123 ,   633 ,0 };
            const unsigned long dim159JoeKuoD7Init[]    =   {   1   ,   3   ,   5   ,   3   ,   9   ,   5   ,   51  ,   123 ,   239 ,   829 ,0 };
            const unsigned long dim160JoeKuoD7Init[]    =   {   1   ,   1   ,   3   ,   15  ,   27  ,   49  ,   1   ,   177 ,   433 ,   237 ,0 };
            const unsigned long dim161JoeKuoD7Init[]    =   {   1   ,   3   ,   3   ,   1   ,   25  ,   1   ,   71  ,   187 ,   3   ,   455 ,   1559    ,0 };
            const unsigned long dim162JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   5   ,   17  ,   53  ,   71  ,   115 ,   87  ,   811 ,   1559    ,0 };
            const unsigned long dim163JoeKuoD7Init[]    =   {   1   ,   3   ,   3   ,   9   ,   3   ,   5   ,   93  ,   233 ,   413 ,   105 ,   335 ,0 };
            const unsigned long dim164JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   5   ,   13  ,   33  ,   45  ,   45  ,   255 ,   691 ,   1827    ,0 };
            const unsigned long dim165JoeKuoD7Init[]    =   {   1   ,   1   ,   5   ,   7   ,   3   ,   51  ,   57  ,   119 ,   355 ,   151 ,   297 ,0 };
            const unsigned long dim166JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   13  ,   17  ,   11  ,   19  ,   5   ,   369 ,   661 ,   877 ,0 };
            const unsigned long dim167JoeKuoD7Init[]    =   {   1   ,   1   ,   7   ,   3   ,   3   ,   61  ,   15  ,   207 ,   161 ,   811 ,   1585    ,0 };
            const unsigned long dim168JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   9   ,   11  ,   63  ,   105 ,   45  ,   209 ,   605 ,   835 ,0 };
            const unsigned long dim169JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   7   ,   11  ,   47  ,   35  ,   73  ,   303 ,   809 ,   1813    ,0 };
            const unsigned long dim170JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   1   ,   9   ,   35  ,   69  ,   219 ,   5   ,   281 ,   1263    ,0 };
            const unsigned long dim171JoeKuoD7Init[]    =   {   1   ,   1   ,   5   ,   1   ,   9   ,   25  ,   85  ,   159 ,   415 ,   471 ,   345 ,0 };
            const unsigned long dim172JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   11  ,   11  ,   61  ,   43  ,   109 ,   327 ,   865 ,   1635    ,0 };
            const unsigned long dim173JoeKuoD7Init[]    =   {   1   ,   1   ,   3   ,   13  ,   21  ,   39  ,   123 ,   211 ,   355 ,   91  ,   1481    ,0 };
            const unsigned long dim174JoeKuoD7Init[]    =   {   1   ,   3   ,   1   ,   15  ,   17  ,   39  ,   69  ,   149 ,   133 ,   779 ,   575 ,0 };
            const unsigned long dim175JoeKuoD7Init[]    =   {   1   ,   3   ,   5   ,   9   ,   15  ,   49  ,   7   ,   183 ,   25  ,   823 ,   971 ,0 };
            const unsigned long dim176JoeKuoD7Init[]    =   {   1   ,   3   ,   1   ,   7   ,   31  ,   63  ,   9   ,   187 ,   211 ,   473 ,   485 ,0 };
            const unsigned long dim177JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   11  ,   9   ,   1   ,   35  ,   177 ,   435 ,   307 ,   1493    ,0 };
            const unsigned long dim178JoeKuoD7Init[]    =   {   1   ,   1   ,   7   ,   3   ,   27  ,   51  ,   91  ,   63  ,   511 ,   179 ,   251 ,0 };
            const unsigned long dim179JoeKuoD7Init[]    =   {   1   ,   1   ,   5   ,   11  ,   27  ,   49  ,   125 ,   111 ,   283 ,   577 ,   397 ,0 };
            const unsigned long dim180JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   15  ,   11  ,   7   ,   19  ,   85  ,   111 ,   629 ,   435 ,0 };
            const unsigned long dim181JoeKuoD7Init[]    =   {   1   ,   3   ,   3   ,   9   ,   7   ,   11  ,   127 ,   25  ,   139 ,   271 ,   263 ,0 };
            const unsigned long dim182JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   15  ,   21  ,   25  ,   55  ,   129 ,   211 ,   755 ,   1695    ,0 };
            const unsigned long dim183JoeKuoD7Init[]    =   {   1   ,   1   ,   7   ,   15  ,   13  ,   23  ,   107 ,   241 ,   231 ,   319 ,   385 ,0 };
            const unsigned long dim184JoeKuoD7Init[]    =   {   1   ,   1   ,   5   ,   11  ,   31  ,   31  ,   97  ,   141 ,   93  ,   759 ,   1199    ,0 };
            const unsigned long dim185JoeKuoD7Init[]    =   {   1   ,   3   ,   5   ,   9   ,   25  ,   15  ,   63  ,   79  ,   321 ,   53  ,   1877    ,0 };
            const unsigned long dim186JoeKuoD7Init[]    =   {   1   ,   1   ,   7   ,   5   ,   11  ,   59  ,   113 ,   77  ,   75  ,   453 ,   1617    ,0 };
            const unsigned long dim187JoeKuoD7Init[]    =   {   1   ,   3   ,   1   ,   15  ,   19  ,   33  ,   65  ,   141 ,   109 ,   211 ,   213 ,0 };
            const unsigned long dim188JoeKuoD7Init[]    =   {   1   ,   3   ,   3   ,   13  ,   17  ,   1   ,   123 ,   1   ,   305 ,   271 ,   1835    ,0 };
            const unsigned long dim189JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   7   ,   9   ,   61  ,   83  ,   75  ,   355 ,   595 ,   773 ,0 };
            const unsigned long dim190JoeKuoD7Init[]    =   {   1   ,   3   ,   5   ,   13  ,   31  ,   53  ,   97  ,   193 ,   95  ,   933 ,   535 ,0 };
            const unsigned long dim191JoeKuoD7Init[]    =   {   1   ,   1   ,   3   ,   11  ,   21  ,   47  ,   95  ,   31  ,   399 ,   115 ,   265 ,0 };
            const unsigned long dim192JoeKuoD7Init[]    =   {   1   ,   3   ,   3   ,   3   ,   19  ,   43  ,   97  ,   29  ,   259 ,   671 ,   1789    ,0 };
            const unsigned long dim193JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   7   ,   11  ,   3   ,   1   ,   167 ,   73  ,   21  ,   587 ,0 };
            const unsigned long dim194JoeKuoD7Init[]    =   {   1   ,   1   ,   7   ,   1   ,   7   ,   29  ,   113 ,   243 ,   389 ,   605 ,   1267    ,0 };
            const unsigned long dim195JoeKuoD7Init[]    =   {   1   ,   3   ,   3   ,   1   ,   19  ,   17  ,   43  ,   89  ,   467 ,   501 ,   137 ,0 };
            const unsigned long dim196JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   7   ,   13  ,   41  ,   97  ,   123 ,   153 ,   603 ,   1469    ,0 };
            const unsigned long dim197JoeKuoD7Init[]    =   {   1   ,   3   ,   5   ,   3   ,   25  ,   29  ,   45  ,   231 ,   489 ,   305 ,   1071    ,0 };
            const unsigned long dim198JoeKuoD7Init[]    =   {   1   ,   1   ,   7   ,   3   ,   31  ,   51  ,   59  ,   113 ,   309 ,   923 ,   1171    ,0 };
            const unsigned long dim199JoeKuoD7Init[]    =   {   1   ,   1   ,   5   ,   11  ,   7   ,   41  ,   55  ,   207 ,   181 ,   565 ,   913 ,0 };
            const unsigned long dim200JoeKuoD7Init[]    =   {   1   ,   1   ,   3   ,   9   ,   9   ,   45  ,   25  ,   49  ,   135 ,   777 ,   593 ,0 };
            const unsigned long dim201JoeKuoD7Init[]    =   {   1   ,   3   ,   5   ,   3   ,   31  ,   3   ,   51  ,   253 ,   369 ,   853 ,   1985    ,0 };
            const unsigned long dim202JoeKuoD7Init[]    =   {   1   ,   3   ,   1   ,   3   ,   25  ,   7   ,   89  ,   193 ,   153 ,   835 ,   341 ,0 };
            const unsigned long dim203JoeKuoD7Init[]    =   {   1   ,   3   ,   1   ,   15  ,   29  ,   51  ,   1   ,   159 ,   77  ,   423 ,   1319    ,0 };
            const unsigned long dim204JoeKuoD7Init[]    =   {   1   ,   1   ,   5   ,   11  ,   31  ,   51  ,   93  ,   249 ,   321 ,   841 ,   619 ,0 };
            const unsigned long dim205JoeKuoD7Init[]    =   {   1   ,   3   ,   1   ,   7   ,   17  ,   11  ,   99  ,   45  ,   407 ,   123 ,   1363    ,0 };
            const unsigned long dim206JoeKuoD7Init[]    =   {   1   ,   3   ,   5   ,   5   ,   13  ,   39  ,   65  ,   195 ,   55  ,   991 ,   1105    ,0 };
            const unsigned long dim207JoeKuoD7Init[]    =   {   1   ,   1   ,   5   ,   11  ,   13  ,   11  ,   79  ,   239 ,   93  ,   683 ,   1459    ,0 };
            const unsigned long dim208JoeKuoD7Init[]    =   {   1   ,   1   ,   5   ,   11  ,   23  ,   5   ,   29  ,   73  ,   197 ,   17  ,   953 ,0 };
            const unsigned long dim209JoeKuoD7Init[]    =   {   1   ,   1   ,   5   ,   3   ,   27  ,   13  ,   109 ,   193 ,   447 ,   271 ,   429 ,0 };
            const unsigned long dim210JoeKuoD7Init[]    =   {   1   ,   3   ,   5   ,   15  ,   17  ,   41  ,   45  ,   205 ,   235 ,   1015    ,   1129    ,0 };
            const unsigned long dim211JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   7   ,   5   ,   33  ,   121 ,   23  ,   265 ,   199 ,   525 ,0 };
            const unsigned long dim212JoeKuoD7Init[]    =   {   1   ,   1   ,   3   ,   5   ,   19  ,   35  ,   17  ,   177 ,   351 ,   935 ,   1769    ,0 };
            const unsigned long dim213JoeKuoD7Init[]    =   {   1   ,   1   ,   5   ,   5   ,   1   ,   39  ,   71  ,   89  ,   27  ,   695 ,   289 ,0 };
            const unsigned long dim214JoeKuoD7Init[]    =   {   1   ,   3   ,   5   ,   5   ,   27  ,   9   ,   27  ,   239 ,   83  ,   349 ,   1989    ,0 };
            const unsigned long dim215JoeKuoD7Init[]    =   {   1   ,   3   ,   1   ,   11  ,   13  ,   63  ,   99  ,   177 ,   91  ,   899 ,   987 ,0 };
            const unsigned long dim216JoeKuoD7Init[]    =   {   1   ,   3   ,   1   ,   1   ,   5   ,   23  ,   85  ,   61  ,   307 ,   491 ,   663 ,0 };
            const unsigned long dim217JoeKuoD7Init[]    =   {   1   ,   1   ,   7   ,   7   ,   19  ,   27  ,   127 ,   205 ,   489 ,   857 ,   1077    ,0 };
            const unsigned long dim218JoeKuoD7Init[]    =   {   1   ,   1   ,   3   ,   1   ,   1   ,   45  ,   43  ,   93  ,   445 ,   391 ,   1703    ,0 };
            const unsigned long dim219JoeKuoD7Init[]    =   {   1   ,   1   ,   7   ,   9   ,   31  ,   55  ,   85  ,   103 ,   439 ,   123 ,   3   ,0 };
            const unsigned long dim220JoeKuoD7Init[]    =   {   1   ,   1   ,   5   ,   15  ,   9   ,   51  ,   51  ,   213 ,   119 ,   245 ,   417 ,0 };
            const unsigned long dim221JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   1   ,   31  ,   49  ,   49  ,   227 ,   279 ,   891 ,   69  ,0 };
            const unsigned long dim222JoeKuoD7Init[]    =   {   1   ,   3   ,   1   ,   13  ,   7   ,   49  ,   19  ,   147 ,   69  ,   451 ,   669 ,0 };
            const unsigned long dim223JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   9   ,   27  ,   27  ,   67  ,   131 ,   191 ,   497 ,   1189    ,0 };
            const unsigned long dim224JoeKuoD7Init[]    =   {   1   ,   3   ,   5   ,   1   ,   13  ,   49  ,   93  ,   1   ,   161 ,   725 ,   709 ,0 };
            const unsigned long dim225JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   1   ,   13  ,   49  ,   67  ,   37  ,   153 ,   31  ,   409 ,0 };
            const unsigned long dim226JoeKuoD7Init[]    =   {   1   ,   3   ,   1   ,   7   ,   19  ,   55  ,   103 ,   111 ,   381 ,   209 ,   271 ,0 };
            const unsigned long dim227JoeKuoD7Init[]    =   {   1   ,   1   ,   5   ,   13  ,   17  ,   29  ,   15  ,   155 ,   81  ,   737 ,   565 ,0 };
            const unsigned long dim228JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   11  ,   29  ,   17  ,   119 ,   247 ,   189 ,   897 ,   1385    ,0 };
            const unsigned long dim229JoeKuoD7Init[]    =   {   1   ,   1   ,   7   ,   13  ,   15  ,   51  ,   71  ,   85  ,   211 ,   839 ,   1085    ,0 };
            const unsigned long dim230JoeKuoD7Init[]    =   {   1   ,   1   ,   3   ,   9   ,   15  ,   37  ,   93  ,   155 ,   357 ,   475 ,   175 ,0 };
            const unsigned long dim231JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   7   ,   1   ,   31  ,   125 ,   181 ,   339 ,   275 ,   301 ,0 };
            const unsigned long dim232JoeKuoD7Init[]    =   {   1   ,   1   ,   3   ,   7   ,   27  ,   33  ,   3   ,   125 ,   153 ,   343 ,   577 ,0 };
            const unsigned long dim233JoeKuoD7Init[]    =   {   1   ,   1   ,   3   ,   1   ,   17  ,   17  ,   51  ,   87  ,   409 ,   845 ,   1857    ,0 };
            const unsigned long dim234JoeKuoD7Init[]    =   {   1   ,   1   ,   3   ,   11  ,   5   ,   47  ,   101 ,   45  ,   425 ,   721 ,   1303    ,0 };
            const unsigned long dim235JoeKuoD7Init[]    =   {   1   ,   3   ,   1   ,   15  ,   9   ,   57  ,   15  ,   7   ,   479 ,   611 ,   1837    ,0 };
            const unsigned long dim236JoeKuoD7Init[]    =   {   1   ,   3   ,   3   ,   13  ,   21  ,   57  ,   89  ,   145 ,   401 ,   1007    ,   753 ,0 };
            const unsigned long dim237JoeKuoD7Init[]    =   {   1   ,   3   ,   3   ,   15  ,   19  ,   63  ,   5   ,   243 ,   111 ,   999 ,   435 ,0 };
            const unsigned long dim238JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   13  ,   31  ,   27  ,   27  ,   217 ,   71  ,   145 ,   695 ,0 };
            const unsigned long dim239JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   7   ,   11  ,   61  ,   113 ,   145 ,   481 ,   487 ,   1609    ,0 };
            const unsigned long dim240JoeKuoD7Init[]    =   {   1   ,   1   ,   5   ,   13  ,   7   ,   51  ,   55  ,   1   ,   307 ,   219 ,   475 ,0 };
            const unsigned long dim241JoeKuoD7Init[]    =   {   1   ,   3   ,   5   ,   5   ,   21  ,   27  ,   77  ,   193 ,   263 ,   37  ,   1421    ,0 };
            const unsigned long dim242JoeKuoD7Init[]    =   {   1   ,   1   ,   5   ,   11  ,   5   ,   23  ,   101 ,   85  ,   41  ,   179 ,   1311    ,0 };
            const unsigned long dim243JoeKuoD7Init[]    =   {   1   ,   1   ,   3   ,   9   ,   27  ,   5   ,   49  ,   227 ,   317 ,   695 ,   581 ,0 };
            const unsigned long dim244JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   15  ,   9   ,   23  ,   119 ,   145 ,   231 ,   15  ,   1349    ,0 };
            const unsigned long dim245JoeKuoD7Init[]    =   {   1   ,   3   ,   1   ,   5   ,   1   ,   17  ,   37  ,   247 ,   261 ,   993 ,   173 ,0 };
            const unsigned long dim246JoeKuoD7Init[]    =   {   1   ,   3   ,   1   ,   15  ,   31  ,   19  ,   23  ,   189 ,   353 ,   445 ,   1145    ,0 };
            const unsigned long dim247JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   11  ,   9   ,   1   ,   65  ,   211 ,   119 ,   531 ,   1389    ,0 };
            const unsigned long dim248JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   1   ,   31  ,   49  ,   31  ,   239 ,   131 ,   159 ,   55  ,0 };
            const unsigned long dim249JoeKuoD7Init[]    =   {   1   ,   1   ,   7   ,   7   ,   3   ,   13  ,   85  ,   89  ,   459 ,   293 ,   1621    ,0 };
            const unsigned long dim250JoeKuoD7Init[]    =   {   1   ,   1   ,   7   ,   7   ,   23  ,   55  ,   79  ,   61  ,   499 ,   915 ,   1641    ,0 };
            const unsigned long dim251JoeKuoD7Init[]    =   {   1   ,   3   ,   1   ,   7   ,   17  ,   47  ,   69  ,   91  ,   157 ,   171 ,   1827    ,0 };
            const unsigned long dim252JoeKuoD7Init[]    =   {   1   ,   3   ,   5   ,   5   ,   29  ,   55  ,   105 ,   31  ,   467 ,   3   ,   241 ,0 };
            const unsigned long dim253JoeKuoD7Init[]    =   {   1   ,   1   ,   5   ,   11  ,   19  ,   27  ,   87  ,   199 ,   295 ,   949 ,   323 ,0 };
            const unsigned long dim254JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   1   ,   25  ,   59  ,   125 ,   33  ,   35  ,   75  ,   603 ,0 };
            const unsigned long dim255JoeKuoD7Init[]    =   {   1   ,   3   ,   5   ,   5   ,   25  ,   63  ,   19  ,   87  ,   475 ,   711 ,   1549    ,0 };
            const unsigned long dim256JoeKuoD7Init[]    =   {   1   ,   3   ,   3   ,   7   ,   11  ,   35  ,   29  ,   193 ,   27  ,   757 ,   133 ,0 };
            const unsigned long dim257JoeKuoD7Init[]    =   {   1   ,   1   ,   5   ,   13  ,   9   ,   29  ,   21  ,   229 ,   137 ,   111 ,   1827    ,0 };
            const unsigned long dim258JoeKuoD7Init[]    =   {   1   ,   3   ,   5   ,   7   ,   7   ,   5   ,   35  ,   217 ,   511 ,   309 ,   2013    ,0 };
            const unsigned long dim259JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   1   ,   3   ,   25  ,   99  ,   187 ,   141 ,   599 ,   223 ,0 };
            const unsigned long dim260JoeKuoD7Init[]    =   {   1   ,   1   ,   7   ,   9   ,   23  ,   9   ,   3   ,   223 ,   357 ,   261 ,   393 ,0 };
            const unsigned long dim261JoeKuoD7Init[]    =   {   1   ,   3   ,   5   ,   9   ,   3   ,   47  ,   73  ,   75  ,   153 ,   249 ,   149 ,0 };
            const unsigned long dim262JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   13  ,   5   ,   31  ,   25  ,   77  ,   501 ,   221 ,   941 ,0 };
            const unsigned long dim263JoeKuoD7Init[]    =   {   1   ,   1   ,   7   ,   1   ,   21  ,   41  ,   107 ,   95  ,   245 ,   35  ,   1843    ,0 };
            const unsigned long dim264JoeKuoD7Init[]    =   {   1   ,   3   ,   3   ,   3   ,   19  ,   59  ,   95  ,   51  ,   211 ,   647 ,   61  ,0 };
            const unsigned long dim265JoeKuoD7Init[]    =   {   1   ,   3   ,   3   ,   15  ,   31  ,   9   ,   91  ,   191 ,   395 ,   161 ,   601 ,0 };
            const unsigned long dim266JoeKuoD7Init[]    =   {   1   ,   3   ,   1   ,   7   ,   25  ,   11  ,   45  ,   147 ,   457 ,   457 ,   1331    ,0 };
            const unsigned long dim267JoeKuoD7Init[]    =   {   1   ,   1   ,   5   ,   15  ,   31  ,   13  ,   49  ,   87  ,   481 ,   155 ,   745 ,0 };
            const unsigned long dim268JoeKuoD7Init[]    =   {   1   ,   1   ,   5   ,   11  ,   17  ,   31  ,   93  ,   39  ,   195 ,   957 ,   877 ,0 };
            const unsigned long dim269JoeKuoD7Init[]    =   {   1   ,   3   ,   3   ,   1   ,   1   ,   9   ,   1   ,   175 ,   63  ,   715 ,   887 ,0 };
            const unsigned long dim270JoeKuoD7Init[]    =   {   1   ,   1   ,   7   ,   3   ,   11  ,   55  ,   15  ,   221 ,   103 ,   169 ,   1961    ,0 };
            const unsigned long dim271JoeKuoD7Init[]    =   {   1   ,   1   ,   5   ,   15  ,   21  ,   41  ,   87  ,   141 ,   233 ,   879 ,   1869    ,0 };
            const unsigned long dim272JoeKuoD7Init[]    =   {   1   ,   3   ,   5   ,   13  ,   13  ,   15  ,   127 ,   29  ,   103 ,   533 ,   925 ,0 };
            const unsigned long dim273JoeKuoD7Init[]    =   {   1   ,   3   ,   1   ,   5   ,   23  ,   59  ,   1   ,   93  ,   97  ,   979 ,   1123    ,0 };
            const unsigned long dim274JoeKuoD7Init[]    =   {   1   ,   3   ,   1   ,   7   ,   31  ,   29  ,   55  ,   185 ,   27  ,   759 ,   1037    ,0 };
            const unsigned long dim275JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   9   ,   9   ,   21  ,   95  ,   111 ,   291 ,   331 ,   889 ,0 };
            const unsigned long dim276JoeKuoD7Init[]    =   {   1   ,   1   ,   3   ,   3   ,   15  ,   17  ,   113 ,   21  ,   97  ,   653 ,   769 ,0 };
            const unsigned long dim277JoeKuoD7Init[]    =   {   1   ,   3   ,   1   ,   9   ,   19  ,   9   ,   67  ,   43  ,   9   ,   7   ,   1751    ,0 };
            const unsigned long dim278JoeKuoD7Init[]    =   {   1   ,   1   ,   3   ,   9   ,   29  ,   15  ,   103 ,   69  ,   249 ,   465 ,   1747    ,0 };
            const unsigned long dim279JoeKuoD7Init[]    =   {   1   ,   1   ,   3   ,   15  ,   17  ,   13  ,   81  ,   149 ,   253 ,   993 ,   25  ,0 };
            const unsigned long dim280JoeKuoD7Init[]    =   {   1   ,   1   ,   3   ,   7   ,   27  ,   19  ,   63  ,   95  ,   165 ,   641 ,   1141    ,0 };
            const unsigned long dim281JoeKuoD7Init[]    =   {   1   ,   3   ,   3   ,   1   ,   13  ,   39  ,   101 ,   131 ,   359 ,   825 ,   41  ,0 };
            const unsigned long dim282JoeKuoD7Init[]    =   {   1   ,   1   ,   5   ,   7   ,   31  ,   31  ,   43  ,   83  ,   147 ,   427 ,   2031    ,0 };
            const unsigned long dim283JoeKuoD7Init[]    =   {   1   ,   3   ,   5   ,   9   ,   27  ,   33  ,   109 ,   45  ,   123 ,   121 ,   485 ,0 };
            const unsigned long dim284JoeKuoD7Init[]    =   {   1   ,   1   ,   3   ,   9   ,   27  ,   35  ,   13  ,   161 ,   363 ,   651 ,   1865    ,0 };
            const unsigned long dim285JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   11  ,   11  ,   45  ,   11  ,   127 ,   19  ,   459 ,   605 ,0 };
            const unsigned long dim286JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   15  ,   13  ,   37  ,   105 ,   159 ,   61  ,   157 ,   157 ,0 };
            const unsigned long dim287JoeKuoD7Init[]    =   {   1   ,   3   ,   1   ,   15  ,   5   ,   39  ,   121 ,   85  ,   383 ,   579 ,   2001    ,0 };
            const unsigned long dim288JoeKuoD7Init[]    =   {   1   ,   3   ,   1   ,   13  ,   13  ,   7   ,   101 ,   213 ,   455 ,   961 ,   1327    ,0 };
            const unsigned long dim289JoeKuoD7Init[]    =   {   1   ,   3   ,   5   ,   13  ,   31  ,   39  ,   49  ,   33  ,   267 ,   783 ,   67  ,0 };
            const unsigned long dim290JoeKuoD7Init[]    =   {   1   ,   1   ,   3   ,   9   ,   31  ,   43  ,   45  ,   85  ,   1   ,   187 ,   899 ,0 };
            const unsigned long dim291JoeKuoD7Init[]    =   {   1   ,   1   ,   7   ,   1   ,   23  ,   11  ,   93  ,   227 ,   41  ,   1013    ,   1689    ,0 };
            const unsigned long dim292JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   15  ,   17  ,   31  ,   33  ,   67  ,   79  ,   113 ,   803 ,0 };
            const unsigned long dim293JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   9   ,   13  ,   53  ,   27  ,   203 ,   143 ,   469 ,   1435    ,0 };
            const unsigned long dim294JoeKuoD7Init[]    =   {   1   ,   1   ,   5   ,   1   ,   19  ,   1   ,   119 ,   37  ,   135 ,   797 ,   1935    ,0 };
            const unsigned long dim295JoeKuoD7Init[]    =   {   1   ,   1   ,   3   ,   11  ,   29  ,   29  ,   111 ,   79  ,   81  ,   973 ,   899 ,0 };
            const unsigned long dim296JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   1   ,   21  ,   3   ,   83  ,   43  ,   95  ,   445 ,   1529    ,0 };
            const unsigned long dim297JoeKuoD7Init[]    =   {   1   ,   1   ,   3   ,   9   ,   11  ,   5   ,   27  ,   67  ,   101 ,   395 ,   195 ,0 };
            const unsigned long dim298JoeKuoD7Init[]    =   {   1   ,   3   ,   1   ,   3   ,   29  ,   13  ,   41  ,   219 ,   157 ,   325 ,   955 ,0 };
            const unsigned long dim299JoeKuoD7Init[]    =   {   1   ,   3   ,   3   ,   13  ,   27  ,   29  ,   113 ,   209 ,   323 ,   747 ,   759 ,0 };
            const unsigned long dim300JoeKuoD7Init[]    =   {   1   ,   3   ,   1   ,   13  ,   1   ,   57  ,   113 ,   133 ,   363 ,   57  ,   715 ,0 };
            const unsigned long dim301JoeKuoD7Init[]    =   {   1   ,   1   ,   3   ,   3   ,   25  ,   43  ,   123 ,   27  ,   483 ,   1023    ,   739 ,0 };
            const unsigned long dim302JoeKuoD7Init[]    =   {   1   ,   1   ,   3   ,   13  ,   27  ,   43  ,   117 ,   217 ,   55  ,   977 ,   487 ,0 };
            const unsigned long dim303JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   3   ,   25  ,   31  ,   95  ,   67  ,   7   ,   19  ,   321 ,0 };
            const unsigned long dim304JoeKuoD7Init[]    =   {   1   ,   3   ,   1   ,   3   ,   25  ,   7   ,   17  ,   123 ,   487 ,   709 ,   971 ,0 };
            const unsigned long dim305JoeKuoD7Init[]    =   {   1   ,   3   ,   5   ,   11  ,   15  ,   9   ,   127 ,   129 ,   35  ,   123 ,   1217    ,0 };
            const unsigned long dim306JoeKuoD7Init[]    =   {   1   ,   1   ,   3   ,   5   ,   17  ,   63  ,   67  ,   137 ,   253 ,   691 ,   447 ,0 };
            const unsigned long dim307JoeKuoD7Init[]    =   {   1   ,   3   ,   3   ,   1   ,   27  ,   53  ,   25  ,   205 ,   75  ,   827 ,   1609    ,0 };
            const unsigned long dim308JoeKuoD7Init[]    =   {   1   ,   1   ,   7   ,   13  ,   3   ,   49  ,   59  ,   207 ,   357 ,   269 ,   309 ,0 };
            const unsigned long dim309JoeKuoD7Init[]    =   {   1   ,   3   ,   1   ,   5   ,   25  ,   57  ,   9   ,   117 ,   359 ,   9   ,   405 ,0 };
            const unsigned long dim310JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   15  ,   15  ,   21  ,   53  ,   163 ,   91  ,   35  ,   277 ,0 };
            const unsigned long dim311JoeKuoD7Init[]    =   {   1   ,   3   ,   1   ,   1   ,   23  ,   5   ,   69  ,   83  ,   11  ,   161 ,   141 ,0 };
            const unsigned long dim312JoeKuoD7Init[]    =   {   1   ,   1   ,   5   ,   13  ,   13  ,   25  ,   15  ,   209 ,   385 ,   389 ,   817 ,0 };
            const unsigned long dim313JoeKuoD7Init[]    =   {   1   ,   3   ,   1   ,   13  ,   11  ,   11  ,   17  ,   191 ,   327 ,   731 ,   973 ,0 };
            const unsigned long dim314JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   15  ,   1   ,   63  ,   55  ,   51  ,   379 ,   359 ,   305 ,0 };
            const unsigned long dim315JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   11  ,   17  ,   19  ,   81  ,   123 ,   335 ,   391 ,   647 ,0 };
            const unsigned long dim316JoeKuoD7Init[]    =   {   1   ,   3   ,   3   ,   13  ,   27  ,   15  ,   53  ,   243 ,   115 ,   733 ,   1095    ,0 };
            const unsigned long dim317JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   3   ,   5   ,   47  ,   111 ,   153 ,   479 ,   133 ,   723 ,0 };
            const unsigned long dim318JoeKuoD7Init[]    =   {   1   ,   1   ,   3   ,   1   ,   15  ,   47  ,   85  ,   27  ,   449 ,   673 ,   931 ,0 };
            const unsigned long dim319JoeKuoD7Init[]    =   {   1   ,   3   ,   1   ,   5   ,   23  ,   23  ,   15  ,   21  ,   113 ,   1003    ,   1077    ,0 };
            const unsigned long dim320JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   9   ,   27  ,   55  ,   29  ,   95  ,   21  ,   29  ,   1409    ,0 };
            const unsigned long dim321JoeKuoD7Init[]    =   {   1   ,   1   ,   5   ,   5   ,   19  ,   53  ,   5   ,   135 ,   219 ,   607 ,   151 ,0 };
            const unsigned long dim322JoeKuoD7Init[]    =   {   1   ,   1   ,   5   ,   1   ,   29  ,   31  ,   83  ,   103 ,   493 ,   227 ,   423 ,0 };
            const unsigned long dim323JoeKuoD7Init[]    =   {   1   ,   3   ,   5   ,   13  ,   23  ,   3   ,   111 ,   147 ,   333 ,   159 ,   983 ,0 };
            const unsigned long dim324JoeKuoD7Init[]    =   {   1   ,   3   ,   5   ,   7   ,   13  ,   1   ,   45  ,   57  ,   111 ,   459 ,   1503    ,0 };
            const unsigned long dim325JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   9   ,   9   ,   21  ,   27  ,   155 ,   377 ,   617 ,   1429    ,0 };
            const unsigned long dim326JoeKuoD7Init[]    =   {   1   ,   3   ,   1   ,   15  ,   3   ,   7   ,   113 ,   11  ,   349 ,   961 ,   749 ,0 };
            const unsigned long dim327JoeKuoD7Init[]    =   {   1   ,   3   ,   5   ,   15  ,   31  ,   13  ,   51  ,   63  ,   315 ,   691 ,   1351    ,0 };
            const unsigned long dim328JoeKuoD7Init[]    =   {   1   ,   1   ,   5   ,   1   ,   21  ,   51  ,   5   ,   53  ,   215 ,   199 ,   891 ,0 };
            const unsigned long dim329JoeKuoD7Init[]    =   {   1   ,   1   ,   3   ,   1   ,   23  ,   19  ,   37  ,   29  ,   263 ,   341 ,   453 ,0 };
            const unsigned long dim330JoeKuoD7Init[]    =   {   1   ,   3   ,   3   ,   1   ,   13  ,   37  ,   113 ,   121 ,   483 ,   879 ,   1749    ,0 };
            const unsigned long dim331JoeKuoD7Init[]    =   {   1   ,   3   ,   3   ,   7   ,   13  ,   41  ,   19  ,   69  ,   323 ,   547 ,   297 ,0 };
            const unsigned long dim332JoeKuoD7Init[]    =   {   1   ,   3   ,   5   ,   9   ,   23  ,   13  ,   21  ,   179 ,   377 ,   217 ,   559 ,0 };
            const unsigned long dim333JoeKuoD7Init[]    =   {   1   ,   3   ,   5   ,   15  ,   15  ,   5   ,   17  ,   41  ,   3   ,   391 ,   1241    ,0 };
            const unsigned long dim334JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   15  ,   31  ,   57  ,   97  ,   57  ,   183 ,   415 ,   917 ,0 };
            const unsigned long dim335JoeKuoD7Init[]    =   {   1   ,   1   ,   7   ,   9   ,   17  ,   31  ,   121 ,   15  ,   425 ,   9   ,   655 ,0 };
            const unsigned long dim336JoeKuoD7Init[]    =   {   1   ,   1   ,   7   ,   15  ,   3   ,   55  ,   53  ,   73  ,   463 ,   887 ,   1999    ,0 };
            const unsigned long dim337JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   7   ,   3   ,   25  ,   17  ,   5   ,   55  ,   833 ,   1899    ,   4077    ,0 };
            const unsigned long dim338JoeKuoD7Init[]    =   {   1   ,   1   ,   5   ,   11  ,   23  ,   47  ,   67  ,   45  ,   9   ,   41  ,   835 ,   373 ,0 };
            const unsigned long dim339JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   11  ,   5   ,   55  ,   25  ,   137 ,   11  ,   119 ,   967 ,   3969    ,0 };
            const unsigned long dim340JoeKuoD7Init[]    =   {   1   ,   3   ,   3   ,   13  ,   3   ,   5   ,   17  ,   11  ,   259 ,   7   ,   1587    ,   3919    ,0 };
            const unsigned long dim341JoeKuoD7Init[]    =   {   1   ,   1   ,   7   ,   3   ,   29  ,   57  ,   113 ,   207 ,   187 ,   173 ,   575 ,   935 ,0 };
            const unsigned long dim342JoeKuoD7Init[]    =   {   1   ,   3   ,   1   ,   9   ,   25  ,   31  ,   89  ,   29  ,   345 ,   753 ,   1675    ,   85  ,0 };
            const unsigned long dim343JoeKuoD7Init[]    =   {   1   ,   1   ,   3   ,   1   ,   9   ,   41  ,   91  ,   245 ,   63  ,   995 ,   1751    ,   2881    ,0 };
            const unsigned long dim344JoeKuoD7Init[]    =   {   1   ,   1   ,   7   ,   3   ,   21  ,   63  ,   121 ,   237 ,   243 ,   815 ,   1507    ,   265 ,0 };
            const unsigned long dim345JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   1   ,   23  ,   27  ,   121 ,   215 ,   381 ,   317 ,   669 ,   357 ,0 };
            const unsigned long dim346JoeKuoD7Init[]    =   {   1   ,   3   ,   1   ,   3   ,   1   ,   45  ,   87  ,   19  ,   43  ,   915 ,   1083    ,   3259    ,0 };
            const unsigned long dim347JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   5   ,   15  ,   31  ,   15  ,   229 ,   489 ,   897 ,   791 ,   3019    ,0 };
            const unsigned long dim348JoeKuoD7Init[]    =   {   1   ,   3   ,   3   ,   5   ,   25  ,   19  ,   89  ,   211 ,   413 ,   839 ,   2029    ,   1785    ,0 };
            const unsigned long dim349JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   1   ,   19  ,   29  ,   67  ,   11  ,   483 ,   303 ,   469 ,   795 ,0 };
            const unsigned long dim350JoeKuoD7Init[]    =   {   1   ,   3   ,   5   ,   5   ,   19  ,   15  ,   81  ,   147 ,   337 ,   641 ,   1863    ,   2497    ,0 };
            const unsigned long dim351JoeKuoD7Init[]    =   {   1   ,   1   ,   7   ,   5   ,   17  ,   43  ,   83  ,   173 ,   307 ,   397 ,   1231    ,   2275    ,0 };
            const unsigned long dim352JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   9   ,   27  ,   49  ,   61  ,   29  ,   169 ,   811 ,   171 ,   1373    ,0 };
            const unsigned long dim353JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   1   ,   13  ,   63  ,   11  ,   197 ,   145 ,   987 ,   1709    ,   895 ,0 };
            const unsigned long dim354JoeKuoD7Init[]    =   {   1   ,   1   ,   7   ,   3   ,   15  ,   3   ,   5   ,   213 ,   155 ,   271 ,   867 ,   1665    ,0 };
            const unsigned long dim355JoeKuoD7Init[]    =   {   1   ,   1   ,   3   ,   1   ,   23  ,   51  ,   99  ,   49  ,   83  ,   159 ,   1589    ,   3961    ,0 };
            const unsigned long dim356JoeKuoD7Init[]    =   {   1   ,   3   ,   3   ,   9   ,   21  ,   13  ,   1   ,   91  ,   187 ,   615 ,   293 ,   3791    ,0 };
            const unsigned long dim357JoeKuoD7Init[]    =   {   1   ,   3   ,   1   ,   5   ,   19  ,   43  ,   107 ,   235 ,   339 ,   65  ,   579 ,   2529    ,0 };
            const unsigned long dim358JoeKuoD7Init[]    =   {   1   ,   3   ,   1   ,   1   ,   17  ,   39  ,   63  ,   67  ,   289 ,   593 ,   959 ,   1503    ,0 };
            const unsigned long dim359JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   11  ,   3   ,   15  ,   87  ,   153 ,   487 ,   537 ,   849 ,   385 ,0 };
            const unsigned long dim360JoeKuoD7Init[]    =   {   1   ,   1   ,   3   ,   3   ,   5   ,   11  ,   99  ,   239 ,   241 ,   251 ,   793 ,   3677    ,0 };
            const unsigned long dim361JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   7   ,   13  ,   21  ,   111 ,   113 ,   69  ,   861 ,   1455    ,   3655    ,0 };
            const unsigned long dim362JoeKuoD7Init[]    =   {   1   ,   1   ,   7   ,   11  ,   9   ,   21  ,   33  ,   29  ,   505 ,   81  ,   713 ,   2545    ,0 };
            const unsigned long dim363JoeKuoD7Init[]    =   {   1   ,   3   ,   1   ,   11  ,   5   ,   29  ,   83  ,   73  ,   377 ,   611 ,   1099    ,   3225    ,0 };
            const unsigned long dim364JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   3   ,   25  ,   53  ,   75  ,   75  ,   235 ,   877 ,   609 ,   1535    ,0 };
            const unsigned long dim365JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   1   ,   29  ,   45  ,   27  ,   13  ,   437 ,   317 ,   581 ,   1179    ,0 };
            const unsigned long dim366JoeKuoD7Init[]    =   {   1   ,   3   ,   3   ,   1   ,   3   ,   25  ,   127 ,   157 ,   443 ,   229 ,   277 ,   4059    ,0 };
            const unsigned long dim367JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   13  ,   17  ,   61  ,   31  ,   1   ,   483 ,   915 ,   1085    ,   1497    ,0 };
            const unsigned long dim368JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   11  ,   13  ,   57  ,   5   ,   49  ,   181 ,   881 ,   765 ,   1767    ,0 };
            const unsigned long dim369JoeKuoD7Init[]    =   {   1   ,   1   ,   3   ,   1   ,   29  ,   17  ,   91  ,   95  ,   319 ,   339 ,   187 ,   2427    ,0 };
            const unsigned long dim370JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   5   ,   1   ,   17  ,   33  ,   181 ,   165 ,   237 ,   815 ,   3387    ,0 };
            const unsigned long dim371JoeKuoD7Init[]    =   {   1   ,   1   ,   7   ,   1   ,   5   ,   17  ,   81  ,   193 ,   61  ,   411 ,   457 ,   2367    ,0 };
            const unsigned long dim372JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   7   ,   23  ,   17  ,   73  ,   185 ,   269 ,   1017    ,   299 ,   1879    ,0 };
            const unsigned long dim373JoeKuoD7Init[]    =   {   1   ,   3   ,   5   ,   11  ,   7   ,   33  ,   75  ,   57  ,   275 ,   355 ,   1157    ,   2297    ,0 };
            const unsigned long dim374JoeKuoD7Init[]    =   {   1   ,   3   ,   5   ,   7   ,   13  ,   7   ,   69  ,   25  ,   411 ,   319 ,   1385    ,   2081    ,0 };
            const unsigned long dim375JoeKuoD7Init[]    =   {   1   ,   1   ,   7   ,   15  ,   27  ,   55  ,   35  ,   51  ,   71  ,   759 ,   1071    ,   51  ,0 };
            const unsigned long dim376JoeKuoD7Init[]    =   {   1   ,   3   ,   5   ,   7   ,   29  ,   3   ,   25  ,   19  ,   123 ,   717 ,   1071    ,   257 ,0 };
            const unsigned long dim377JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   1   ,   7   ,   37  ,   51  ,   175 ,   19  ,   511 ,   469 ,   3757    ,0 };
            const unsigned long dim378JoeKuoD7Init[]    =   {   1   ,   1   ,   5   ,   5   ,   19  ,   45  ,   107 ,   207 ,   369 ,   479 ,   853 ,   195 ,0 };
            const unsigned long dim379JoeKuoD7Init[]    =   {   1   ,   3   ,   1   ,   9   ,   11  ,   11  ,   89  ,   115 ,   149 ,   789 ,   523 ,   2949    ,0 };
            const unsigned long dim380JoeKuoD7Init[]    =   {   1   ,   3   ,   1   ,   11  ,   15  ,   53  ,   123 ,   167 ,   471 ,   355 ,   1515    ,   2291    ,0 };
            const unsigned long dim381JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   1   ,   23  ,   1   ,   5   ,   21  ,   315 ,   45  ,   1901    ,   3351    ,0 };
            const unsigned long dim382JoeKuoD7Init[]    =   {   1   ,   1   ,   7   ,   5   ,   3   ,   49  ,   31  ,   89  ,   11  ,   777 ,   1003    ,   389 ,0 };
            const unsigned long dim383JoeKuoD7Init[]    =   {   1   ,   1   ,   5   ,   3   ,   23  ,   3   ,   55  ,   213 ,   333 ,   257 ,   309 ,   4079    ,0 };
            const unsigned long dim384JoeKuoD7Init[]    =   {   1   ,   3   ,   3   ,   7   ,   9   ,   29  ,   103 ,   41  ,   161 ,   253 ,   681 ,   1971    ,0 };
            const unsigned long dim385JoeKuoD7Init[]    =   {   1   ,   1   ,   7   ,   13  ,   13  ,   33  ,   71  ,   121 ,   325 ,   423 ,   1485    ,   3205    ,0 };
            const unsigned long dim386JoeKuoD7Init[]    =   {   1   ,   3   ,   1   ,   7   ,   9   ,   55  ,   69  ,   201 ,   111 ,   561 ,   1733    ,   1881    ,0 };
            const unsigned long dim387JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   7   ,   9   ,   41  ,   31  ,   113 ,   79  ,   193 ,   757 ,   329 ,0 };
            const unsigned long dim388JoeKuoD7Init[]    =   {   1   ,   1   ,   7   ,   5   ,   3   ,   23  ,   27  ,   99  ,   179 ,   7   ,   1087    ,   27  ,0 };
            const unsigned long dim389JoeKuoD7Init[]    =   {   1   ,   3   ,   3   ,   7   ,   13  ,   23  ,   9   ,   61  ,   157 ,   641 ,   299 ,   2077    ,0 };
            const unsigned long dim390JoeKuoD7Init[]    =   {   1   ,   1   ,   5   ,   5   ,   29  ,   59  ,   63  ,   67  ,   493 ,   119 ,   959 ,   2683    ,0 };
            const unsigned long dim391JoeKuoD7Init[]    =   {   1   ,   1   ,   3   ,   1   ,   19  ,   25  ,   117 ,   161 ,   249 ,   545 ,   997 ,   17  ,0 };
            const unsigned long dim392JoeKuoD7Init[]    =   {   1   ,   3   ,   5   ,   5   ,   5   ,   31  ,   107 ,   181 ,   191 ,   913 ,   1911    ,   2329    ,0 };
            const unsigned long dim393JoeKuoD7Init[]    =   {   1   ,   3   ,   1   ,   13  ,   31  ,   11  ,   109 ,   229 ,   159 ,   235 ,   975 ,   717 ,0 };
            const unsigned long dim394JoeKuoD7Init[]    =   {   1   ,   1   ,   5   ,   3   ,   5   ,   7   ,   117 ,   233 ,   83  ,   31  ,   1641    ,   251 ,0 };
            const unsigned long dim395JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   1   ,   3   ,   45  ,   53  ,   179 ,   113 ,   353 ,   1805    ,   1097    ,0 };
            const unsigned long dim396JoeKuoD7Init[]    =   {   1   ,   3   ,   3   ,   5   ,   17  ,   15  ,   83  ,   177 ,   399 ,   831 ,   231 ,   2307    ,0 };
            const unsigned long dim397JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   15  ,   19  ,   37  ,   19  ,   119 ,   243 ,   703 ,   1813    ,   2415    ,0 };
            const unsigned long dim398JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   9   ,   19  ,   25  ,   119 ,   123 ,   235 ,   817 ,   871 ,   763 ,0 };
            const unsigned long dim399JoeKuoD7Init[]    =   {   1   ,   1   ,   5   ,   3   ,   7   ,   49  ,   97  ,   207 ,   297 ,   283 ,   467 ,   1843    ,0 };
            const unsigned long dim400JoeKuoD7Init[]    =   {   1   ,   3   ,   5   ,   1   ,   21  ,   25  ,   103 ,   227 ,   239 ,   683 ,   797 ,   1763    ,0 };
            const unsigned long dim401JoeKuoD7Init[]    =   {   1   ,   1   ,   5   ,   11  ,   27  ,   27  ,   127 ,   185 ,   137 ,   449 ,   1549    ,   3561    ,0 };
            const unsigned long dim402JoeKuoD7Init[]    =   {   1   ,   1   ,   5   ,   5   ,   3   ,   45  ,   113 ,   35  ,   427 ,   847 ,   1761    ,   1721    ,0 };
            const unsigned long dim403JoeKuoD7Init[]    =   {   1   ,   1   ,   3   ,   7   ,   25  ,   63  ,   33  ,   197 ,   183 ,   803 ,   963 ,   2179    ,0 };
            const unsigned long dim404JoeKuoD7Init[]    =   {   1   ,   1   ,   3   ,   9   ,   19  ,   1   ,   3   ,   229 ,   225 ,   479 ,   593 ,   597 ,0 };
            const unsigned long dim405JoeKuoD7Init[]    =   {   1   ,   3   ,   1   ,   5   ,   19  ,   35  ,   105 ,   245 ,   1   ,   249 ,   217 ,   2917    ,0 };
            const unsigned long dim406JoeKuoD7Init[]    =   {   1   ,   3   ,   1   ,   1   ,   11  ,   3   ,   59  ,   51  ,   249 ,   135 ,   419 ,   3411    ,0 };
            const unsigned long dim407JoeKuoD7Init[]    =   {   1   ,   3   ,   3   ,   15  ,   27  ,   45  ,   51  ,   15  ,   17  ,   473 ,   557 ,   2679    ,0 };
            const unsigned long dim408JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   11  ,   31  ,   43  ,   45  ,   163 ,   289 ,   723 ,   337 ,   2817    ,0 };
            const unsigned long dim409JoeKuoD7Init[]    =   {   1   ,   1   ,   3   ,   11  ,   11  ,   51  ,   73  ,   101 ,   217 ,   255 ,   1547    ,   925 ,0 };
            const unsigned long dim410JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   5   ,   7   ,   53  ,   5   ,   163 ,   323 ,   757 ,   1443    ,   2703    ,0 };
            const unsigned long dim411JoeKuoD7Init[]    =   {   1   ,   1   ,   5   ,   15  ,   19  ,   35  ,   23  ,   129 ,   199 ,   303 ,   1801    ,   1999    ,0 };
            const unsigned long dim412JoeKuoD7Init[]    =   {   1   ,   1   ,   5   ,   15  ,   23  ,   37  ,   25  ,   61  ,   307 ,   901 ,   1117    ,   1301    ,0 };
            const unsigned long dim413JoeKuoD7Init[]    =   {   1   ,   1   ,   5   ,   11  ,   15  ,   59  ,   71  ,   163 ,   93  ,   471 ,   341 ,   1825    ,0 };
            const unsigned long dim414JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   11  ,   27  ,   63  ,   73  ,   209 ,   453 ,   49  ,   1161    ,   1399    ,0 };
            const unsigned long dim415JoeKuoD7Init[]    =   {   1   ,   3   ,   5   ,   3   ,   9   ,   5   ,   101 ,   47  ,   225 ,   451 ,   2005    ,   223 ,0 };
            const unsigned long dim416JoeKuoD7Init[]    =   {   1   ,   3   ,   3   ,   3   ,   25  ,   43  ,   33  ,   171 ,   413 ,   371 ,   1133    ,   1057    ,0 };
            const unsigned long dim417JoeKuoD7Init[]    =   {   1   ,   3   ,   3   ,   9   ,   5   ,   35  ,   39  ,   219 ,   319 ,   301 ,   1475    ,   3155    ,0 };
            const unsigned long dim418JoeKuoD7Init[]    =   {   1   ,   3   ,   5   ,   15  ,   17  ,   39  ,   123 ,   165 ,   337 ,   17  ,   679 ,   3155    ,0 };
            const unsigned long dim419JoeKuoD7Init[]    =   {   1   ,   1   ,   5   ,   3   ,   29  ,   33  ,   13  ,   165 ,   57  ,   723 ,   1201    ,   3523    ,0 };
            const unsigned long dim420JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   9   ,   25  ,   23  ,   67  ,   139 ,   321 ,   555 ,   723 ,   1017    ,0 };
            const unsigned long dim421JoeKuoD7Init[]    =   {   1   ,   1   ,   3   ,   9   ,   17  ,   59  ,   25  ,   29  ,   473 ,   473 ,   1425    ,   965 ,0 };
            const unsigned long dim422JoeKuoD7Init[]    =   {   1   ,   1   ,   5   ,   3   ,   13  ,   53  ,   29  ,   239 ,   321 ,   455 ,   1539    ,   1279    ,0 };
            const unsigned long dim423JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   3   ,   3   ,   37  ,   127 ,   149 ,   205 ,   313 ,   827 ,   1713    ,0 };
            const unsigned long dim424JoeKuoD7Init[]    =   {   1   ,   1   ,   7   ,   5   ,   11  ,   41  ,   25  ,   73  ,   279 ,   539 ,   1761    ,   2565    ,0 };
            const unsigned long dim425JoeKuoD7Init[]    =   {   1   ,   1   ,   5   ,   9   ,   21  ,   7   ,   75  ,   209 ,   83  ,   885 ,   821 ,   3887    ,0 };
            const unsigned long dim426JoeKuoD7Init[]    =   {   1   ,   1   ,   3   ,   1   ,   21  ,   1   ,   13  ,   245 ,   291 ,   615 ,   1747    ,   1477    ,0 };
            const unsigned long dim427JoeKuoD7Init[]    =   {   1   ,   1   ,   5   ,   1   ,   1   ,   11  ,   107 ,   105 ,   19  ,   263 ,   1497    ,   2311    ,0 };
            const unsigned long dim428JoeKuoD7Init[]    =   {   1   ,   1   ,   5   ,   5   ,   15  ,   15  ,   105 ,   171 ,   143 ,   879 ,   1611    ,   277 ,0 };
            const unsigned long dim429JoeKuoD7Init[]    =   {   1   ,   3   ,   1   ,   15  ,   15  ,   51  ,   15  ,   15  ,   25  ,   309 ,   1531    ,   977 ,0 };
            const unsigned long dim430JoeKuoD7Init[]    =   {   1   ,   1   ,   7   ,   11  ,   29  ,   63  ,   65  ,   209 ,   155 ,   975 ,   107 ,   611 ,0 };
            const unsigned long dim431JoeKuoD7Init[]    =   {   1   ,   3   ,   3   ,   15  ,   11  ,   55  ,   21  ,   149 ,   291 ,   221 ,   449 ,   133 ,0 };
            const unsigned long dim432JoeKuoD7Init[]    =   {   1   ,   1   ,   3   ,   13  ,   3   ,   59  ,   41  ,   167 ,   37  ,   951 ,   1561    ,   2087    ,0 };
            const unsigned long dim433JoeKuoD7Init[]    =   {   1   ,   3   ,   3   ,   15  ,   31  ,   59  ,   97  ,   49  ,   319 ,   935 ,   1419    ,   3215    ,0 };
            const unsigned long dim434JoeKuoD7Init[]    =   {   1   ,   1   ,   3   ,   1   ,   27  ,   17  ,   91  ,   91  ,   327 ,   395 ,   1553    ,   201 ,0 };
            const unsigned long dim435JoeKuoD7Init[]    =   {   1   ,   3   ,   1   ,   7   ,   5   ,   59  ,   51  ,   77  ,   207 ,   697 ,   521 ,   635 ,0 };
            const unsigned long dim436JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   5   ,   5   ,   37  ,   95  ,   91  ,   399 ,   151 ,   1795    ,   905 ,0 };
            const unsigned long dim437JoeKuoD7Init[]    =   {   1   ,   3   ,   5   ,   5   ,   19  ,   29  ,   49  ,   55  ,   377 ,   793 ,   1429    ,   3393    ,0 };
            const unsigned long dim438JoeKuoD7Init[]    =   {   1   ,   1   ,   5   ,   9   ,   9   ,   43  ,   127 ,   203 ,   355 ,   951 ,   711 ,   563 ,0 };
            const unsigned long dim439JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   13  ,   23  ,   21  ,   123 ,   41  ,   465 ,   33  ,   855 ,   1685    ,0 };
            const unsigned long dim440JoeKuoD7Init[]    =   {   1   ,   1   ,   5   ,   7   ,   7   ,   11  ,   65  ,   195 ,   235 ,   823 ,   23  ,   2723    ,0 };
            const unsigned long dim441JoeKuoD7Init[]    =   {   1   ,   1   ,   7   ,   7   ,   9   ,   41  ,   93  ,   109 ,   475 ,   587 ,   27  ,   1061    ,0 };
            const unsigned long dim442JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   13  ,   5   ,   45  ,   29  ,   87  ,   75  ,   853 ,   2023    ,   795 ,0 };
            const unsigned long dim443JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   3   ,   3   ,   7   ,   59  ,   235 ,   139 ,   457 ,   1855    ,   3739    ,0 };
            const unsigned long dim444JoeKuoD7Init[]    =   {   1   ,   1   ,   7   ,   11  ,   23  ,   57  ,   25  ,   11  ,   473 ,   699 ,   887 ,   805 ,0 };
            const unsigned long dim445JoeKuoD7Init[]    =   {   1   ,   1   ,   5   ,   1   ,   27  ,   3   ,   23  ,   115 ,   499 ,   287 ,   531 ,   1115    ,0 };
            const unsigned long dim446JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   15  ,   7   ,   7   ,   33  ,   83  ,   29  ,   477 ,   727 ,   3037    ,0 };
            const unsigned long dim447JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   9   ,   27  ,   43  ,   113 ,   203 ,   413 ,   649 ,   869 ,   3351    ,0 };
            const unsigned long dim448JoeKuoD7Init[]    =   {   1   ,   1   ,   3   ,   9   ,   27  ,   15  ,   105 ,   167 ,   431 ,   87  ,   931 ,   4017    ,0 };
            const unsigned long dim449JoeKuoD7Init[]    =   {   1   ,   3   ,   3   ,   1   ,   15  ,   23  ,   85  ,   145 ,   349 ,   291 ,   1137    ,   605 ,0 };
            const unsigned long dim450JoeKuoD7Init[]    =   {   1   ,   3   ,   3   ,   15  ,   31  ,   53  ,   127 ,   25  ,   15  ,   277 ,   575 ,   1327    ,0 };
            const unsigned long dim451JoeKuoD7Init[]    =   {   1   ,   3   ,   1   ,   15  ,   17  ,   29  ,   125 ,   251 ,   469 ,   15  ,   1803    ,   237 ,0 };
            const unsigned long dim452JoeKuoD7Init[]    =   {   1   ,   3   ,   5   ,   13  ,   9   ,   21  ,   23  ,   111 ,   103 ,   861 ,   1137    ,   3843    ,0 };
            const unsigned long dim453JoeKuoD7Init[]    =   {   1   ,   1   ,   5   ,   5   ,   29  ,   49  ,   101 ,   237 ,   187 ,   825 ,   137 ,   957 ,0 };
            const unsigned long dim454JoeKuoD7Init[]    =   {   1   ,   1   ,   3   ,   13  ,   13  ,   9   ,   101 ,   171 ,   15  ,   245 ,   107 ,   2179    ,0 };
            const unsigned long dim455JoeKuoD7Init[]    =   {   1   ,   1   ,   3   ,   11  ,   21  ,   33  ,   51  ,   211 ,   113 ,   569 ,   295 ,   3593    ,0 };
            const unsigned long dim456JoeKuoD7Init[]    =   {   1   ,   3   ,   1   ,   5   ,   9   ,   5   ,   119 ,   159 ,   211 ,   121 ,   1547    ,   2267    ,0 };
            const unsigned long dim457JoeKuoD7Init[]    =   {   1   ,   1   ,   3   ,   9   ,   7   ,   19  ,   53  ,   39  ,   271 ,   437 ,   765 ,   1511    ,0 };
            const unsigned long dim458JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   9   ,   3   ,   5   ,   119 ,   83  ,   161 ,   275 ,   1443    ,   465 ,0 };
            const unsigned long dim459JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   1   ,   27  ,   61  ,   103 ,   207 ,   249 ,   479 ,   651 ,   1793    ,0 };
            const unsigned long dim460JoeKuoD7Init[]    =   {   1   ,   3   ,   5   ,   9   ,   11  ,   53  ,   43  ,   41  ,   271 ,   859 ,   25  ,   1401    ,0 };
            const unsigned long dim461JoeKuoD7Init[]    =   {   1   ,   3   ,   1   ,   5   ,   9   ,   39  ,   87  ,   21  ,   425 ,   561 ,   861 ,   2801    ,0 };
            const unsigned long dim462JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   1   ,   1   ,   25  ,   21  ,   199 ,   315 ,   355 ,   429 ,   3093    ,0 };
            const unsigned long dim463JoeKuoD7Init[]    =   {   1   ,   1   ,   7   ,   9   ,   11  ,   63  ,   123 ,   253 ,   483 ,   781 ,   1685    ,   977 ,0 };
            const unsigned long dim464JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   15  ,   17  ,   57  ,   19  ,   251 ,   173 ,   519 ,   1523    ,   749 ,0 };
            const unsigned long dim465JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   11  ,   23  ,   5   ,   119 ,   7   ,   327 ,   507 ,   1777    ,   951 ,0 };
            const unsigned long dim466JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   5   ,   1   ,   37  ,   41  ,   99  ,   273 ,   241 ,   293 ,   2289    ,0 };
            const unsigned long dim467JoeKuoD7Init[]    =   {   1   ,   3   ,   3   ,   9   ,   3   ,   51  ,   65  ,   115 ,   159 ,   341 ,   1603    ,   3769    ,0 };
            const unsigned long dim468JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   3   ,   15  ,   61  ,   105 ,   103 ,   113 ,   435 ,   415 ,   161 ,0 };
            const unsigned long dim469JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   7   ,   9   ,   9   ,   17  ,   13  ,   135 ,   923 ,   951 ,   3603    ,0 };
            const unsigned long dim470JoeKuoD7Init[]    =   {   1   ,   1   ,   3   ,   13  ,   9   ,   31  ,   25  ,   175 ,   449 ,   541 ,   745 ,   2771    ,0 };
            const unsigned long dim471JoeKuoD7Init[]    =   {   1   ,   3   ,   3   ,   9   ,   5   ,   5   ,   61  ,   17  ,   357 ,   623 ,   873 ,   975 ,0 };
            const unsigned long dim472JoeKuoD7Init[]    =   {   1   ,   1   ,   3   ,   9   ,   9   ,   31  ,   11  ,   225 ,   81  ,   965 ,   1837    ,   1975    ,0 };
            const unsigned long dim473JoeKuoD7Init[]    =   {   1   ,   1   ,   7   ,   1   ,   3   ,   57  ,   69  ,   169 ,   11  ,   633 ,   885 ,   535 ,0 };
            const unsigned long dim474JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   9   ,   21  ,   63  ,   73  ,   233 ,   151 ,   519 ,   1319    ,   1485    ,0 };
            const unsigned long dim475JoeKuoD7Init[]    =   {   1   ,   1   ,   3   ,   11  ,   23  ,   5   ,   51  ,   115 ,   501 ,   627 ,   1875    ,   3003    ,0 };
            const unsigned long dim476JoeKuoD7Init[]    =   {   1   ,   1   ,   5   ,   1   ,   9   ,   3   ,   91  ,   177 ,   305 ,   897 ,   317 ,   3231    ,0 };
            const unsigned long dim477JoeKuoD7Init[]    =   {   1   ,   3   ,   3   ,   3   ,   9   ,   13  ,   121 ,   93  ,   303 ,   551 ,   1087    ,   905 ,0 };
            const unsigned long dim478JoeKuoD7Init[]    =   {   1   ,   1   ,   5   ,   7   ,   1   ,   63  ,   123 ,   177 ,   87  ,   489 ,   1247    ,   1831    ,0 };
            const unsigned long dim479JoeKuoD7Init[]    =   {   1   ,   3   ,   3   ,   5   ,   25  ,   61  ,   1   ,   225 ,   247 ,   991 ,   231 ,   657 ,0 };
            const unsigned long dim480JoeKuoD7Init[]    =   {   1   ,   3   ,   5   ,   5   ,   19  ,   23  ,   113 ,   101 ,   427 ,   327 ,   1859    ,   3115    ,0 };
            const unsigned long dim481JoeKuoD7Init[]    =   {   1   ,   1   ,   5   ,   13  ,   11  ,   5   ,   125 ,   133 ,   131 ,   441 ,   1277    ,   2583    ,   7133    ,0 };
            const unsigned long dim482JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   3   ,   29  ,   45  ,   21  ,   45  ,   477 ,   135 ,   13  ,   2531    ,   7657    ,0 };
            const unsigned long dim483JoeKuoD7Init[]    =   {   1   ,   1   ,   7   ,   9   ,   19  ,   3   ,   47  ,   37  ,   491 ,   237 ,   1493    ,   4053    ,   6025    ,0 };
            const unsigned long dim484JoeKuoD7Init[]    =   {   1   ,   1   ,   5   ,   7   ,   15  ,   51  ,   23  ,   181 ,   63  ,   755 ,   761 ,   3215    ,   3547    ,0 };
            const unsigned long dim485JoeKuoD7Init[]    =   {   1   ,   1   ,   7   ,   5   ,   11  ,   3   ,   27  ,   187 ,   221 ,   305 ,   1941    ,   1443    ,   5161    ,0 };
            const unsigned long dim486JoeKuoD7Init[]    =   {   1   ,   1   ,   7   ,   13  ,   13  ,   53  ,   63  ,   19  ,   5   ,   273 ,   1305    ,   919 ,   3335    ,0 };
            const unsigned long dim487JoeKuoD7Init[]    =   {   1   ,   3   ,   1   ,   5   ,   15  ,   19  ,   85  ,   35  ,   41  ,   117 ,   1269    ,   289 ,   7433    ,0 };
            const unsigned long dim488JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   9   ,   27  ,   49  ,   119 ,   101 ,   57  ,   267 ,   1225    ,   3779    ,   3357    ,0 };
            const unsigned long dim489JoeKuoD7Init[]    =   {   1   ,   3   ,   5   ,   11  ,   5   ,   1   ,   9   ,   251 ,   411 ,   19  ,   763 ,   1297    ,   5309    ,0 };
            const unsigned long dim490JoeKuoD7Init[]    =   {   1   ,   3   ,   3   ,   13  ,   31  ,   55  ,   67  ,   63  ,   23  ,   1021    ,   199 ,   2885    ,   1777    ,0 };
            const unsigned long dim491JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   9   ,   29  ,   27  ,   123 ,   191 ,   125 ,   295 ,   891 ,   3507    ,   2895    ,0 };
            const unsigned long dim492JoeKuoD7Init[]    =   {   1   ,   1   ,   5   ,   7   ,   15  ,   41  ,   35  ,   119 ,   343 ,   603 ,   819 ,   41  ,   8131    ,0 };
            const unsigned long dim493JoeKuoD7Init[]    =   {   1   ,   3   ,   3   ,   9   ,   21  ,   61  ,   43  ,   199 ,   359 ,   607 ,   623 ,   1579    ,   7899    ,0 };
            const unsigned long dim494JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   15  ,   25  ,   59  ,   47  ,   47  ,   365 ,   125 ,   921 ,   3081    ,   2081    ,0 };
            const unsigned long dim495JoeKuoD7Init[]    =   {   1   ,   3   ,   3   ,   1   ,   23  ,   23  ,   111 ,   53  ,   395 ,   945 ,   59  ,   3345    ,   1671    ,0 };
            const unsigned long dim496JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   15  ,   15  ,   57  ,   31  ,   15  ,   235 ,   247 ,   1645    ,   3085    ,   4381    ,0 };
            const unsigned long dim497JoeKuoD7Init[]    =   {   1   ,   1   ,   3   ,   11  ,   23  ,   21  ,   9   ,   89  ,   61  ,   347 ,   1423    ,   3613    ,   697 ,0 };
            const unsigned long dim498JoeKuoD7Init[]    =   {   1   ,   1   ,   5   ,   13  ,   21  ,   37  ,   5   ,   159 ,   409 ,   131 ,   559 ,   3505    ,   6631    ,0 };
            const unsigned long dim499JoeKuoD7Init[]    =   {   1   ,   3   ,   3   ,   11  ,   23  ,   51  ,   97  ,   247 ,   137 ,   183 ,   1775    ,   2169    ,   2869    ,0 };
            const unsigned long dim500JoeKuoD7Init[]    =   {   1   ,   1   ,   7   ,   7   ,   29  ,   53  ,   109 ,   75  ,   327 ,   755 ,   825 ,   329 ,   5795    ,0 };
            const unsigned long dim501JoeKuoD7Init[]    =   {   1   ,   3   ,   1   ,   5   ,   17  ,   63  ,   109 ,   199 ,   215 ,   869 ,   2009    ,   1153    ,   1233    ,0 };
            const unsigned long dim502JoeKuoD7Init[]    =   {   1   ,   1   ,   7   ,   15  ,   15  ,   37  ,   63  ,   93  ,   151 ,   621 ,   1749    ,   2211    ,   5437    ,0 };
            const unsigned long dim503JoeKuoD7Init[]    =   {   1   ,   3   ,   5   ,   15  ,   13  ,   63  ,   57  ,   17  ,   323 ,   271 ,   1451    ,   459 ,   1187    ,0 };
            const unsigned long dim504JoeKuoD7Init[]    =   {   1   ,   1   ,   5   ,   5   ,   31  ,   47  ,   15  ,   83  ,   389 ,   857 ,   505 ,   635 ,   7341    ,0 };
            const unsigned long dim505JoeKuoD7Init[]    =   {   1   ,   3   ,   5   ,   3   ,   19  ,   7   ,   83  ,   17  ,   51  ,   947 ,   1193    ,   163 ,   3895    ,0 };
            const unsigned long dim506JoeKuoD7Init[]    =   {   1   ,   1   ,   7   ,   1   ,   23  ,   5   ,   115 ,   113 ,   445 ,   703 ,   747 ,   963 ,   7483    ,0 };
            const unsigned long dim507JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   5   ,   27  ,   31  ,   41  ,   221 ,   313 ,   317 ,   517 ,   3901    ,   3937    ,0 };
            const unsigned long dim508JoeKuoD7Init[]    =   {   1   ,   3   ,   5   ,   11  ,   29  ,   41  ,   23  ,   249 ,   169 ,   51  ,   1783    ,   2493    ,   6109    ,0 };
            const unsigned long dim509JoeKuoD7Init[]    =   {   1   ,   1   ,   5   ,   13  ,   23  ,   23  ,   85  ,   71  ,   321 ,   623 ,   207 ,   3027    ,   319 ,0 };
            const unsigned long dim510JoeKuoD7Init[]    =   {   1   ,   1   ,   7   ,   3   ,   3   ,   19  ,   63  ,   233 ,   377 ,   509 ,   1969    ,   67  ,   5059    ,0 };
            const unsigned long dim511JoeKuoD7Init[]    =   {   1   ,   1   ,   3   ,   7   ,   13  ,   27  ,   39  ,   67  ,   483 ,   587 ,   1581    ,   177 ,   5213    ,0 };
            const unsigned long dim512JoeKuoD7Init[]    =   {   1   ,   1   ,   3   ,   1   ,   23  ,   55  ,   3   ,   177 ,   97  ,   551 ,   1733    ,   1943    ,   6607    ,0 };
            const unsigned long dim513JoeKuoD7Init[]    =   {   1   ,   3   ,   3   ,   5   ,   5   ,   9   ,   39  ,   189 ,   473 ,   5   ,   19  ,   2793    ,   6937    ,0 };
            const unsigned long dim514JoeKuoD7Init[]    =   {   1   ,   1   ,   3   ,   13  ,   31  ,   35  ,   105 ,   19  ,   279 ,   749 ,   523 ,   1503    ,   6231    ,0 };
            const unsigned long dim515JoeKuoD7Init[]    =   {   1   ,   3   ,   1   ,   9   ,   11  ,   59  ,   89  ,   153 ,   155 ,   219 ,   549 ,   2263    ,   8101    ,0 };
            const unsigned long dim516JoeKuoD7Init[]    =   {   1   ,   1   ,   7   ,   13  ,   29  ,   9   ,   9   ,   61  ,   43  ,   165 ,   1395    ,   91  ,   2871    ,0 };
            const unsigned long dim517JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   1   ,   27  ,   63  ,   63  ,   97  ,   47  ,   969 ,   991 ,   111 ,   2267    ,0 };
            const unsigned long dim518JoeKuoD7Init[]    =   {   1   ,   1   ,   3   ,   7   ,   5   ,   13  ,   51  ,   195 ,   369 ,   551 ,   961 ,   1123    ,   7001    ,0 };
            const unsigned long dim519JoeKuoD7Init[]    =   {   1   ,   1   ,   5   ,   7   ,   27  ,   57  ,   49  ,   51  ,   321 ,   187 ,   1869    ,   1181    ,   6333    ,0 };
            const unsigned long dim520JoeKuoD7Init[]    =   {   1   ,   3   ,   1   ,   3   ,   25  ,   63  ,   15  ,   165 ,   175 ,   401 ,   1659    ,   2345    ,   683 ,0 };
            const unsigned long dim521JoeKuoD7Init[]    =   {   1   ,   1   ,   7   ,   1   ,   9   ,   41  ,   121 ,   123 ,   49  ,   303 ,   229 ,   2799    ,   3247    ,0 };
            const unsigned long dim522JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   15  ,   29  ,   11  ,   113 ,   137 ,   221 ,   859 ,   1601    ,   215 ,   6257    ,0 };
            const unsigned long dim523JoeKuoD7Init[]    =   {   1   ,   1   ,   5   ,   9   ,   9   ,   21  ,   91  ,   179 ,   369 ,   389 ,   1737    ,   1235    ,   3033    ,0 };
            const unsigned long dim524JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   7   ,   21  ,   53  ,   85  ,   217 ,   505 ,   299 ,   1505    ,   3289    ,   3811    ,0 };
            const unsigned long dim525JoeKuoD7Init[]    =   {   1   ,   1   ,   3   ,   15  ,   25  ,   25  ,   109 ,   229 ,   511 ,   733 ,   1047    ,   3295    ,   307 ,0 };
            const unsigned long dim526JoeKuoD7Init[]    =   {   1   ,   1   ,   3   ,   15  ,   29  ,   25  ,   21  ,   201 ,   353 ,   97  ,   231 ,   3791    ,   3019    ,0 };
            const unsigned long dim527JoeKuoD7Init[]    =   {   1   ,   3   ,   1   ,   5   ,   1   ,   7   ,   75  ,   57  ,   501 ,   217 ,   1647    ,   1539    ,   2961    ,0 };
            const unsigned long dim528JoeKuoD7Init[]    =   {   1   ,   3   ,   3   ,   5   ,   21  ,   13  ,   123 ,   153 ,   83  ,   253 ,   19  ,   2469    ,   5239    ,0 };
            const unsigned long dim529JoeKuoD7Init[]    =   {   1   ,   3   ,   1   ,   3   ,   31  ,   13  ,   103 ,   107 ,   91  ,   3   ,   1795    ,   3587    ,   3345    ,0 };
            const unsigned long dim530JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   13  ,   1   ,   47  ,   71  ,   61  ,   335 ,   347 ,   753 ,   3685    ,   3895    ,0 };
            const unsigned long dim531JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   11  ,   25  ,   53  ,   73  ,   65  ,   203 ,   923 ,   661 ,   1349    ,   8089    ,0 };
            const unsigned long dim532JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   5   ,   11  ,   23  ,   93  ,   149 ,   367 ,   577 ,   1095    ,   993 ,   5279    ,0 };
            const unsigned long dim533JoeKuoD7Init[]    =   {   1   ,   3   ,   3   ,   7   ,   11  ,   33  ,   23  ,   117 ,   457 ,   1023    ,   1789    ,   3103    ,   2979    ,0 };
            const unsigned long dim534JoeKuoD7Init[]    =   {   1   ,   3   ,   3   ,   11  ,   3   ,   3   ,   3   ,   177 ,   9   ,   11  ,   1753    ,   3291    ,   7617    ,0 };
            const unsigned long dim535JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   7   ,   1   ,   27  ,   105 ,   199 ,   465 ,   505 ,   427 ,   2605    ,   7215    ,0 };
            const unsigned long dim536JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   1   ,   19  ,   23  ,   21  ,   167 ,   205 ,   561 ,   175 ,   677 ,   117 ,0 };
            const unsigned long dim537JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   7   ,   31  ,   23  ,   83  ,   219 ,   391 ,   101 ,   1433    ,   3973    ,   4217    ,0 };
            const unsigned long dim538JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   3   ,   9   ,   41  ,   21  ,   189 ,   437 ,   187 ,   1973    ,   445 ,   7439    ,0 };
            const unsigned long dim539JoeKuoD7Init[]    =   {   1   ,   1   ,   5   ,   7   ,   21  ,   3   ,   53  ,   9   ,   295 ,   331 ,   2029    ,   3063    ,   6095    ,0 };
            const unsigned long dim540JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   7   ,   19  ,   9   ,   119 ,   247 ,   371 ,   635 ,   1411    ,   419 ,   3297    ,0 };
            const unsigned long dim541JoeKuoD7Init[]    =   {   1   ,   1   ,   3   ,   1   ,   3   ,   31  ,   113 ,   235 ,   57  ,   923 ,   2033    ,   1523    ,   6381    ,0 };
            const unsigned long dim542JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   3   ,   17  ,   51  ,   115 ,   215 ,   271 ,   411 ,   573 ,   1761    ,   1793    ,0 };
            const unsigned long dim543JoeKuoD7Init[]    =   {   1   ,   3   ,   1   ,   11  ,   5   ,   15  ,   69  ,   193 ,   107 ,   577 ,   909 ,   1997    ,   4503    ,0 };
            const unsigned long dim544JoeKuoD7Init[]    =   {   1   ,   3   ,   5   ,   3   ,   1   ,   3   ,   17  ,   227 ,   381 ,   931 ,   1933    ,   807 ,   3295    ,0 };
            const unsigned long dim545JoeKuoD7Init[]    =   {   1   ,   1   ,   7   ,   15  ,   15  ,   23  ,   15  ,   183 ,   247 ,   167 ,   193 ,   1643    ,   4139    ,0 };
            const unsigned long dim546JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   13  ,   31  ,   39  ,   13  ,   213 ,   45  ,   643 ,   515 ,   1167    ,   6991    ,0 };
            const unsigned long dim547JoeKuoD7Init[]    =   {   1   ,   3   ,   1   ,   9   ,   15  ,   7   ,   5   ,   223 ,   53  ,   413 ,   403 ,   3003    ,   4135    ,0 };
            const unsigned long dim548JoeKuoD7Init[]    =   {   1   ,   3   ,   3   ,   5   ,   25  ,   25  ,   31  ,   209 ,   113 ,   971 ,   629 ,   189 ,   2309    ,0 };
            const unsigned long dim549JoeKuoD7Init[]    =   {   1   ,   3   ,   3   ,   3   ,   23  ,   29  ,   37  ,   39  ,   73  ,   797 ,   589 ,   1351    ,   4495    ,0 };
            const unsigned long dim550JoeKuoD7Init[]    =   {   1   ,   3   ,   1   ,   13  ,   31  ,   25  ,   15  ,   147 ,   351 ,   541 ,   493 ,   1361    ,   3151    ,0 };
            const unsigned long dim551JoeKuoD7Init[]    =   {   1   ,   3   ,   5   ,   9   ,   31  ,   45  ,   43  ,   125 ,   503 ,   987 ,   1485    ,   521 ,   957 ,0 };
            const unsigned long dim552JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   5   ,   1   ,   51  ,   7   ,   189 ,   111 ,   735 ,   993 ,   2313    ,   7027    ,0 };
            const unsigned long dim553JoeKuoD7Init[]    =   {   1   ,   1   ,   7   ,   1   ,   5   ,   11  ,   123 ,   241 ,   325 ,   337 ,   1369    ,   3515    ,   8077    ,0 };
            const unsigned long dim554JoeKuoD7Init[]    =   {   1   ,   3   ,   3   ,   13  ,   9   ,   57  ,   89  ,   89  ,   439 ,   803 ,   545 ,   1097    ,   2443    ,0 };
            const unsigned long dim555JoeKuoD7Init[]    =   {   1   ,   3   ,   5   ,   7   ,   1   ,   31  ,   23  ,   121 ,   19  ,   469 ,   887 ,   3925    ,   6591    ,0 };
            const unsigned long dim556JoeKuoD7Init[]    =   {   1   ,   1   ,   3   ,   7   ,   11  ,   39  ,   95  ,   65  ,   39  ,   407 ,   1187    ,   2803    ,   643 ,0 };
            const unsigned long dim557JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   3   ,   23  ,   43  ,   97  ,   39  ,   463 ,   929 ,   469 ,   255 ,   7219    ,0 };
            const unsigned long dim558JoeKuoD7Init[]    =   {   1   ,   1   ,   5   ,   5   ,   21  ,   47  ,   99  ,   117 ,   233 ,   921 ,   801 ,   1105    ,   2993    ,0 };
            const unsigned long dim559JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   5   ,   7   ,   1   ,   33  ,   61  ,   389 ,   969 ,   389 ,   2081    ,   4215    ,0 };
            const unsigned long dim560JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   7   ,   15  ,   5   ,   103 ,   115 ,   59  ,   699 ,   1287    ,   3023    ,   7599    ,0 };
            const unsigned long dim561JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   15  ,   23  ,   49  ,   19  ,   85  ,   219 ,   837 ,   15  ,   1213    ,   1011    ,0 };
            const unsigned long dim562JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   3   ,   9   ,   19  ,   55  ,   247 ,   97  ,   125 ,   1377    ,   3107    ,   1787    ,0 };
            const unsigned long dim563JoeKuoD7Init[]    =   {   1   ,   3   ,   5   ,   7   ,   31  ,   13  ,   35  ,   53  ,   269 ,   603 ,   729 ,   1827    ,   1083    ,0 };
            const unsigned long dim564JoeKuoD7Init[]    =   {   1   ,   1   ,   5   ,   11  ,   7   ,   55  ,   81  ,   61  ,   3   ,   119 ,   995 ,   1721    ,   3205    ,0 };
            const unsigned long dim565JoeKuoD7Init[]    =   {   1   ,   1   ,   7   ,   13  ,   11  ,   57  ,   81  ,   63  ,   271 ,   1001    ,   1901    ,   471 ,   2217    ,0 };
            const unsigned long dim566JoeKuoD7Init[]    =   {   1   ,   1   ,   5   ,   1   ,   1   ,   47  ,   79  ,   175 ,   145 ,   615 ,   1393    ,   137 ,   589 ,0 };
            const unsigned long dim567JoeKuoD7Init[]    =   {   1   ,   1   ,   7   ,   11  ,   25  ,   47  ,   93  ,   81  ,   157 ,   913 ,   1117    ,   819 ,   8079    ,0 };
            const unsigned long dim568JoeKuoD7Init[]    =   {   1   ,   1   ,   7   ,   1   ,   23  ,   51  ,   85  ,   101 ,   41  ,   115 ,   303 ,   1721    ,   6515    ,0 };
            const unsigned long dim569JoeKuoD7Init[]    =   {   1   ,   3   ,   5   ,   13  ,   23  ,   31  ,   1   ,   31  ,   63  ,   595 ,   611 ,   3153    ,   7865    ,0 };
            const unsigned long dim570JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   9   ,   17  ,   35  ,   53  ,   53  ,   421 ,   537 ,   557 ,   2735    ,   7813    ,0 };
            const unsigned long dim571JoeKuoD7Init[]    =   {   1   ,   3   ,   3   ,   11  ,   7   ,   17  ,   105 ,   1   ,   201 ,   45  ,   1517    ,   3303    ,   3287    ,0 };
            const unsigned long dim572JoeKuoD7Init[]    =   {   1   ,   1   ,   7   ,   15  ,   15  ,   63  ,   45  ,   129 ,   81  ,   901 ,   1735    ,   3551    ,   2885    ,0 };
            const unsigned long dim573JoeKuoD7Init[]    =   {   1   ,   1   ,   5   ,   15  ,   21  ,   9   ,   63  ,   133 ,   245 ,   807 ,   1425    ,   961 ,   7425    ,0 };
            const unsigned long dim574JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   15  ,   27  ,   25  ,   41  ,   93  ,   253 ,   345 ,   875 ,   1663    ,   5713    ,0 };
            const unsigned long dim575JoeKuoD7Init[]    =   {   1   ,   3   ,   1   ,   1   ,   9   ,   61  ,   117 ,   183 ,   381 ,   739 ,   1739    ,   3285    ,   747 ,0 };
            const unsigned long dim576JoeKuoD7Init[]    =   {   1   ,   3   ,   5   ,   15  ,   7   ,   49  ,   39  ,   233 ,   465 ,   511 ,   1853    ,   637 ,   6823    ,0 };
            const unsigned long dim577JoeKuoD7Init[]    =   {   1   ,   3   ,   3   ,   15  ,   29  ,   19  ,   1   ,   71  ,   109 ,   959 ,   1575    ,   3225    ,   3951    ,0 };
            const unsigned long dim578JoeKuoD7Init[]    =   {   1   ,   3   ,   5   ,   7   ,   3   ,   21  ,   23  ,   25  ,   423 ,   507 ,   617 ,   2175    ,   7331    ,0 };
            const unsigned long dim579JoeKuoD7Init[]    =   {   1   ,   1   ,   5   ,   5   ,   31  ,   19  ,   111 ,   139 ,   217 ,   301 ,   1845    ,   1789    ,   5935    ,0 };
            const unsigned long dim580JoeKuoD7Init[]    =   {   1   ,   3   ,   1   ,   11  ,   19  ,   63  ,   53  ,   47  ,   493 ,   181 ,   709 ,   3601    ,   6163    ,0 };
            const unsigned long dim581JoeKuoD7Init[]    =   {   1   ,   3   ,   3   ,   9   ,   7   ,   57  ,   57  ,   89  ,   25  ,   757 ,   1959    ,   285 ,   2121    ,0 };
            const unsigned long dim582JoeKuoD7Init[]    =   {   1   ,   3   ,   5   ,   13  ,   13  ,   23  ,   25  ,   61  ,   481 ,   473 ,   1933    ,   1861    ,   881 ,0 };
            const unsigned long dim583JoeKuoD7Init[]    =   {   1   ,   3   ,   1   ,   15  ,   11  ,   27  ,   53  ,   243 ,   439 ,   221 ,   1281    ,   1129    ,   6685    ,0 };
            const unsigned long dim584JoeKuoD7Init[]    =   {   1   ,   1   ,   3   ,   7   ,   11  ,   55  ,   11  ,   99  ,   153 ,   517 ,   985 ,   3533    ,   1363    ,0 };
            const unsigned long dim585JoeKuoD7Init[]    =   {   1   ,   1   ,   3   ,   3   ,   11  ,   7   ,   95  ,   193 ,   161 ,   183 ,   1417    ,   187 ,   4787    ,0 };
            const unsigned long dim586JoeKuoD7Init[]    =   {   1   ,   1   ,   3   ,   13  ,   13  ,   21  ,   1   ,   183 ,   257 ,   393 ,   831 ,   2149    ,   2219    ,0 };
            const unsigned long dim587JoeKuoD7Init[]    =   {   1   ,   1   ,   3   ,   9   ,   1   ,   19  ,   9   ,   95  ,   65  ,   57  ,   1999    ,   125 ,   2181    ,0 };
            const unsigned long dim588JoeKuoD7Init[]    =   {   1   ,   1   ,   3   ,   11  ,   3   ,   29  ,   57  ,   249 ,   93  ,   143 ,   1163    ,   2635    ,   837 ,0 };
            const unsigned long dim589JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   3   ,   19  ,   3   ,   61  ,   151 ,   115 ,   601 ,   1085    ,   2505    ,   4171    ,0 };
            const unsigned long dim590JoeKuoD7Init[]    =   {   1   ,   1   ,   5   ,   11  ,   19  ,   35  ,   89  ,   117 ,   5   ,   197 ,   1763    ,   415 ,   5455    ,0 };
            const unsigned long dim591JoeKuoD7Init[]    =   {   1   ,   3   ,   3   ,   13  ,   1   ,   41  ,   77  ,   127 ,   93  ,   785 ,   269 ,   915 ,   5723    ,0 };
            const unsigned long dim592JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   5   ,   25  ,   39  ,   77  ,   77  ,   499 ,   931 ,   1707    ,   3401    ,   5707    ,0 };
            const unsigned long dim593JoeKuoD7Init[]    =   {   1   ,   1   ,   7   ,   13  ,   7   ,   41  ,   99  ,   3   ,   463 ,   433 ,   651 ,   4079    ,   409 ,0 };
            const unsigned long dim594JoeKuoD7Init[]    =   {   1   ,   1   ,   5   ,   3   ,   25  ,   25  ,   51  ,   211 ,   269 ,   867 ,   1403    ,   517 ,   4371    ,0 };
            const unsigned long dim595JoeKuoD7Init[]    =   {   1   ,   3   ,   1   ,   15  ,   23  ,   53  ,   53  ,   41  ,   269 ,   571 ,   525 ,   2309    ,   431 ,0 };
            const unsigned long dim596JoeKuoD7Init[]    =   {   1   ,   1   ,   5   ,   5   ,   19  ,   25  ,   125 ,   133 ,   249 ,   555 ,   781 ,   2853    ,   2185    ,0 };
            const unsigned long dim597JoeKuoD7Init[]    =   {   1   ,   3   ,   1   ,   13  ,   27  ,   59  ,   47  ,   67  ,   197 ,   865 ,   101 ,   2291    ,   6219    ,0 };
            const unsigned long dim598JoeKuoD7Init[]    =   {   1   ,   1   ,   7   ,   7   ,   5   ,   43  ,   95  ,   209 ,   407 ,   821 ,   215 ,   2371    ,   5059    ,0 };
            const unsigned long dim599JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   1   ,   7   ,   25  ,   75  ,   239 ,   327 ,   969 ,   575 ,   2519    ,   3515    ,0 };
            const unsigned long dim600JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   3   ,   3   ,   13  ,   105 ,   1   ,   319 ,   549 ,   1003    ,   3441    ,   185 ,0 };
            const unsigned long dim601JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   9   ,   29  ,   17  ,   75  ,   95  ,   407 ,   199 ,   1609    ,   3537    ,   7293    ,0 };
            const unsigned long dim602JoeKuoD7Init[]    =   {   1   ,   1   ,   5   ,   7   ,   15  ,   27  ,   39  ,   175 ,   431 ,   905 ,   925 ,   3277    ,   8029    ,0 };
            const unsigned long dim603JoeKuoD7Init[]    =   {   1   ,   3   ,   1   ,   3   ,   3   ,   19  ,   73  ,   87  ,   107 ,   605 ,   45  ,   3267    ,   3033    ,0 };
            const unsigned long dim604JoeKuoD7Init[]    =   {   1   ,   3   ,   1   ,   1   ,   31  ,   9   ,   73  ,   151 ,   71  ,   929 ,   1995    ,   739 ,   1483    ,0 };
            const unsigned long dim605JoeKuoD7Init[]    =   {   1   ,   1   ,   7   ,   3   ,   7   ,   55  ,   27  ,   177 ,   225 ,   333 ,   663 ,   4065    ,   2559    ,0 };
            const unsigned long dim606JoeKuoD7Init[]    =   {   1   ,   3   ,   3   ,   5   ,   21  ,   61  ,   121 ,   159 ,   269 ,   267 ,   1307    ,   1213    ,   5057    ,0 };
            const unsigned long dim607JoeKuoD7Init[]    =   {   1   ,   3   ,   3   ,   15  ,   3   ,   35  ,   29  ,   243 ,   199 ,   695 ,   1225    ,   3437    ,   7451    ,0 };
            const unsigned long dim608JoeKuoD7Init[]    =   {   1   ,   3   ,   3   ,   9   ,   15  ,   23  ,   15  ,   43  ,   383 ,   95  ,   239 ,   945 ,   5339    ,0 };
            const unsigned long dim609JoeKuoD7Init[]    =   {   1   ,   1   ,   5   ,   9   ,   23  ,   11  ,   81  ,   93  ,   493 ,   435 ,   1291    ,   873 ,   5101    ,0 };
            const unsigned long dim610JoeKuoD7Init[]    =   {   1   ,   1   ,   5   ,   7   ,   23  ,   11  ,   89  ,   95  ,   203 ,   533 ,   583 ,   981 ,   2029    ,0 };
            const unsigned long dim611JoeKuoD7Init[]    =   {   1   ,   1   ,   7   ,   3   ,   27  ,   57  ,   59  ,   161 ,   361 ,   511 ,   1723    ,   1893    ,   4113    ,0 };
            const unsigned long dim612JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   15  ,   27  ,   5   ,   7   ,   213 ,   509 ,   483 ,   1645    ,   3487    ,   2927    ,0 };
            const unsigned long dim613JoeKuoD7Init[]    =   {   1   ,   3   ,   5   ,   13  ,   17  ,   47  ,   87  ,   25  ,   331 ,   793 ,   1467    ,   875 ,   7147    ,0 };
            const unsigned long dim614JoeKuoD7Init[]    =   {   1   ,   3   ,   5   ,   1   ,   29  ,   51  ,   21  ,   223 ,   289 ,   461 ,   1153    ,   579 ,   2621    ,0 };
            const unsigned long dim615JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   5   ,   9   ,   9   ,   69  ,   21  ,   257 ,   679 ,   1031    ,   809 ,   1633    ,0 };
            const unsigned long dim616JoeKuoD7Init[]    =   {   1   ,   1   ,   3   ,   13  ,   3   ,   59  ,   27  ,   193 ,   507 ,   407 ,   831 ,   1921    ,   4411    ,0 };
            const unsigned long dim617JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   9   ,   19  ,   61  ,   77  ,   163 ,   103 ,   35  ,   803 ,   2327    ,   7349    ,0 };
            const unsigned long dim618JoeKuoD7Init[]    =   {   1   ,   1   ,   7   ,   1   ,   5   ,   19  ,   49  ,   79  ,   469 ,   595 ,   625 ,   333 ,   4861    ,0 };
            const unsigned long dim619JoeKuoD7Init[]    =   {   1   ,   1   ,   3   ,   13  ,   9   ,   17  ,   97  ,   229 ,   263 ,   63  ,   1907    ,   2493    ,   1001    ,0 };
            const unsigned long dim620JoeKuoD7Init[]    =   {   1   ,   3   ,   3   ,   11  ,   25  ,   23  ,   111 ,   51  ,   7   ,   551 ,   1669    ,   2495    ,   3409    ,0 };
            const unsigned long dim621JoeKuoD7Init[]    =   {   1   ,   1   ,   3   ,   13  ,   5   ,   39  ,   115 ,   179 ,   473 ,   179 ,   2047    ,   467 ,   5797    ,0 };
            const unsigned long dim622JoeKuoD7Init[]    =   {   1   ,   1   ,   7   ,   9   ,   21  ,   3   ,   9   ,   245 ,   29  ,   457 ,   101 ,   253 ,   5757    ,0 };
            const unsigned long dim623JoeKuoD7Init[]    =   {   1   ,   1   ,   3   ,   5   ,   1   ,   11  ,   123 ,   29  ,   43  ,   259 ,   1387    ,   2973    ,   2509    ,0 };
            const unsigned long dim624JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   9   ,   15  ,   35  ,   85  ,   165 ,   91  ,   793 ,   1963    ,   595 ,   1721    ,0 };
            const unsigned long dim625JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   11  ,   29  ,   29  ,   79  ,   33  ,   439 ,   635 ,   1639    ,   803 ,   847 ,0 };
            const unsigned long dim626JoeKuoD7Init[]    =   {   1   ,   1   ,   3   ,   5   ,   29  ,   9   ,   121 ,   239 ,   33  ,   535 ,   1957    ,   1459    ,   7997    ,0 };
            const unsigned long dim627JoeKuoD7Init[]    =   {   1   ,   3   ,   1   ,   9   ,   7   ,   15  ,   11  ,   91  ,   495 ,   939 ,   1597    ,   4011    ,   97  ,0 };
            const unsigned long dim628JoeKuoD7Init[]    =   {   1   ,   3   ,   5   ,   7   ,   27  ,   1   ,   115 ,   29  ,   345 ,   665 ,   1433    ,   3291    ,   1431    ,0 };
            const unsigned long dim629JoeKuoD7Init[]    =   {   1   ,   3   ,   5   ,   11  ,   21  ,   27  ,   127 ,   55  ,   401 ,   787 ,   405 ,   715 ,   5385    ,0 };
            const unsigned long dim630JoeKuoD7Init[]    =   {   1   ,   3   ,   5   ,   7   ,   7   ,   59  ,   109 ,   221 ,   179 ,   217 ,   1045    ,   1979    ,   4003    ,0 };
            const unsigned long dim631JoeKuoD7Init[]    =   {   1   ,   1   ,   5   ,   9   ,   17  ,   57  ,   73  ,   219 ,   369 ,   235 ,   43  ,   65  ,   919 ,0 };
            const unsigned long dim632JoeKuoD7Init[]    =   {   1   ,   3   ,   3   ,   13  ,   19  ,   29  ,   73  ,   7   ,   405 ,   691 ,   1007    ,   341 ,   7153    ,0 };
            const unsigned long dim633JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   9   ,   1   ,   17  ,   67  ,   101 ,   33  ,   851 ,   1227    ,   41  ,   8057    ,0 };
            const unsigned long dim634JoeKuoD7Init[]    =   {   1   ,   1   ,   3   ,   15  ,   21  ,   37  ,   43  ,   161 ,   63  ,   27  ,   855 ,   3983    ,   7661    ,0 };
            const unsigned long dim635JoeKuoD7Init[]    =   {   1   ,   3   ,   3   ,   3   ,   25  ,   11  ,   85  ,   127 ,   309 ,   979 ,   1071    ,   3485    ,   3513    ,0 };
            const unsigned long dim636JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   1   ,   21  ,   51  ,   37  ,   87  ,   71  ,   923 ,   1983    ,   1013    ,   4677    ,0 };
            const unsigned long dim637JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   3   ,   29  ,   13  ,   93  ,   193 ,   207 ,   229 ,   185 ,   2109    ,   8085    ,0 };
            const unsigned long dim638JoeKuoD7Init[]    =   {   1   ,   3   ,   5   ,   5   ,   15  ,   43  ,   101 ,   27  ,   185 ,   541 ,   1371    ,   1487    ,   11  ,0 };
            const unsigned long dim639JoeKuoD7Init[]    =   {   1   ,   1   ,   7   ,   7   ,   11  ,   35  ,   83  ,   163 ,   3   ,   391 ,   343 ,   2931    ,   1183    ,0 };
            const unsigned long dim640JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   7   ,   5   ,   1   ,   109 ,   195 ,   389 ,   679 ,   571 ,   121 ,   2669    ,0 };
            const unsigned long dim641JoeKuoD7Init[]    =   {   1   ,   1   ,   5   ,   13  ,   11  ,   11  ,   47  ,   105 ,   389 ,   661 ,   113 ,   2849    ,   2861    ,0 };
            const unsigned long dim642JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   7   ,   27  ,   3   ,   33  ,   105 ,   81  ,   817 ,   691 ,   1017    ,   4477    ,0 };
            const unsigned long dim643JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   15  ,   19  ,   63  ,   11  ,   179 ,   473 ,   695 ,   1033    ,   363 ,   1479    ,0 };
            const unsigned long dim644JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   5   ,   23  ,   33  ,   121 ,   101 ,   349 ,   599 ,   1821    ,   3975    ,   263 ,0 };
            const unsigned long dim645JoeKuoD7Init[]    =   {   1   ,   3   ,   1   ,   11  ,   21  ,   1   ,   115 ,   123 ,   367 ,   497 ,   609 ,   1473    ,   515 ,0 };
            const unsigned long dim646JoeKuoD7Init[]    =   {   1   ,   3   ,   5   ,   5   ,   13  ,   49  ,   63  ,   63  ,   137 ,   39  ,   1281    ,   1289    ,   6687    ,0 };
            const unsigned long dim647JoeKuoD7Init[]    =   {   1   ,   3   ,   3   ,   3   ,   9   ,   49  ,   11  ,   205 ,   451 ,   163 ,   125 ,   2639    ,   1921    ,0 };
            const unsigned long dim648JoeKuoD7Init[]    =   {   1   ,   1   ,   7   ,   5   ,   5   ,   29  ,   67  ,   33  ,   169 ,   133 ,   599 ,   2559    ,   6495    ,0 };
            const unsigned long dim649JoeKuoD7Init[]    =   {   1   ,   3   ,   5   ,   13  ,   21  ,   47  ,   119 ,   39  ,   389 ,   695 ,   717 ,   1003    ,   4107    ,0 };
            const unsigned long dim650JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   7   ,   23  ,   49  ,   113 ,   245 ,   485 ,   291 ,   1431    ,   3577    ,   1787    ,0 };
            const unsigned long dim651JoeKuoD7Init[]    =   {   1   ,   1   ,   5   ,   11  ,   27  ,   59  ,   127 ,   205 ,   177 ,   73  ,   935 ,   2041    ,   6899    ,0 };
            const unsigned long dim652JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   13  ,   27  ,   41  ,   115 ,   121 ,   239 ,   939 ,   1963    ,   2045    ,   2457    ,0 };
            const unsigned long dim653JoeKuoD7Init[]    =   {   1   ,   3   ,   1   ,   11  ,   23  ,   51  ,   13  ,   177 ,   51  ,   145 ,   651 ,   3817    ,   2399    ,0 };
            const unsigned long dim654JoeKuoD7Init[]    =   {   1   ,   3   ,   1   ,   3   ,   11  ,   35  ,   119 ,   251 ,   303 ,   673 ,   1475    ,   1415    ,   3809    ,0 };
            const unsigned long dim655JoeKuoD7Init[]    =   {   1   ,   1   ,   3   ,   3   ,   31  ,   25  ,   117 ,   87  ,   19  ,   491 ,   1681    ,   2291    ,   3559    ,0 };
            const unsigned long dim656JoeKuoD7Init[]    =   {   1   ,   1   ,   3   ,   15  ,   25  ,   13  ,   83  ,   243 ,   499 ,   77  ,   1809    ,   81  ,   4211    ,0 };
            const unsigned long dim657JoeKuoD7Init[]    =   {   1   ,   1   ,   7   ,   3   ,   27  ,   7   ,   119 ,   141 ,   305 ,   323 ,   1679    ,   1143    ,   2277    ,0 };
            const unsigned long dim658JoeKuoD7Init[]    =   {   1   ,   3   ,   1   ,   7   ,   27  ,   31  ,   31  ,   219 ,   55  ,   437 ,   1113    ,   2801    ,   1409    ,0 };
            const unsigned long dim659JoeKuoD7Init[]    =   {   1   ,   3   ,   5   ,   7   ,   25  ,   51  ,   77  ,   255 ,   211 ,   265 ,   1597    ,   3883    ,   5913    ,0 };
            const unsigned long dim660JoeKuoD7Init[]    =   {   1   ,   3   ,   3   ,   1   ,   23  ,   55  ,   21  ,   243 ,   329 ,   221 ,   901 ,   1363    ,   1449    ,0 };
            const unsigned long dim661JoeKuoD7Init[]    =   {   1   ,   1   ,   7   ,   3   ,   7   ,   53  ,   63  ,   143 ,   503 ,   295 ,   135 ,   4037    ,   6363    ,0 };
            const unsigned long dim662JoeKuoD7Init[]    =   {   1   ,   3   ,   1   ,   7   ,   7   ,   35  ,   7   ,   19  ,   437 ,   361 ,   807 ,   277 ,   6335    ,0 };
            const unsigned long dim663JoeKuoD7Init[]    =   {   1   ,   3   ,   5   ,   5   ,   5   ,   53  ,   121 ,   219 ,   447 ,   453 ,   1001    ,   339 ,   3959    ,0 };
            const unsigned long dim664JoeKuoD7Init[]    =   {   1   ,   1   ,   7   ,   13  ,   27  ,   29  ,   107 ,   255 ,   385 ,   539 ,   1787    ,   3091    ,   6319    ,0 };
            const unsigned long dim665JoeKuoD7Init[]    =   {   1   ,   1   ,   3   ,   5   ,   1   ,   57  ,   15  ,   93  ,   331 ,   155 ,   107 ,   3295    ,   7081    ,0 };
            const unsigned long dim666JoeKuoD7Init[]    =   {   1   ,   1   ,   5   ,   15  ,   1   ,   35  ,   57  ,   249 ,   171 ,   309 ,   1413    ,   2677    ,   1593    ,0 };
            const unsigned long dim667JoeKuoD7Init[]    =   {   1   ,   1   ,   7   ,   7   ,   27  ,   51  ,   107 ,   233 ,   163 ,   317 ,   1815    ,   2443    ,   3431    ,0 };
            const unsigned long dim668JoeKuoD7Init[]    =   {   1   ,   1   ,   3   ,   5   ,   13  ,   31  ,   119 ,   25  ,   339 ,   671 ,   1143    ,   2317    ,   1963    ,0 };
            const unsigned long dim669JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   11  ,   15  ,   39  ,   15  ,   135 ,   183 ,   445 ,   1607    ,   2397    ,   7609    ,0 };
            const unsigned long dim670JoeKuoD7Init[]    =   {   1   ,   3   ,   5   ,   7   ,   13  ,   61  ,   29  ,   5   ,   309 ,   1009    ,   1241    ,   3471    ,   4505    ,0 };
            const unsigned long dim671JoeKuoD7Init[]    =   {   1   ,   1   ,   7   ,   9   ,   7   ,   39  ,   13  ,   67  ,   421 ,   169 ,   1751    ,   2939    ,   3315    ,0 };
            const unsigned long dim672JoeKuoD7Init[]    =   {   1   ,   1   ,   7   ,   9   ,   5   ,   35  ,   91  ,   59  ,   195 ,   621 ,   449 ,   2245    ,   3379    ,0 };
            const unsigned long dim673JoeKuoD7Init[]    =   {   1   ,   1   ,   3   ,   9   ,   27  ,   23  ,   87  ,   181 ,   171 ,   37  ,   1137    ,   597 ,   7443    ,0 };
            const unsigned long dim674JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   13  ,   15  ,   41  ,   109 ,   81  ,   149 ,   709 ,   1537    ,   2349    ,   3999    ,0 };
            const unsigned long dim675JoeKuoD7Init[]    =   {   1   ,   3   ,   1   ,   3   ,   9   ,   27  ,   93  ,   211 ,   239 ,   391 ,   411 ,   1329    ,   4213    ,0 };
            const unsigned long dim676JoeKuoD7Init[]    =   {   1   ,   3   ,   5   ,   15  ,   9   ,   33  ,   19  ,   105 ,   89  ,   181 ,   157 ,   461 ,   6505    ,0 };
            const unsigned long dim677JoeKuoD7Init[]    =   {   1   ,   3   ,   5   ,   15  ,   15  ,   3   ,   113 ,   79  ,   49  ,   463 ,   299 ,   1621    ,   2719    ,0 };
            const unsigned long dim678JoeKuoD7Init[]    =   {   1   ,   1   ,   5   ,   13  ,   23  ,   13  ,   57  ,   161 ,   347 ,   483 ,   1521    ,   3611    ,   4111    ,0 };
            const unsigned long dim679JoeKuoD7Init[]    =   {   1   ,   1   ,   7   ,   13  ,   15  ,   21  ,   79  ,   145 ,   329 ,   649 ,   1809    ,   1925    ,   7157    ,0 };
            const unsigned long dim680JoeKuoD7Init[]    =   {   1   ,   1   ,   3   ,   1   ,   29  ,   5   ,   35  ,   229 ,   375 ,   789 ,   481 ,   1863    ,   6467    ,0 };
            const unsigned long dim681JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   15  ,   1   ,   19  ,   39  ,   3   ,   327 ,   721 ,   1763    ,   963 ,   4397    ,0 };
            const unsigned long dim682JoeKuoD7Init[]    =   {   1   ,   3   ,   1   ,   3   ,   3   ,   51  ,   97  ,   1   ,   109 ,   337 ,   535 ,   2795    ,   1155    ,0 };
            const unsigned long dim683JoeKuoD7Init[]    =   {   1   ,   3   ,   5   ,   13  ,   5   ,   55  ,   91  ,   147 ,   395 ,   141 ,   947 ,   2647    ,   1193    ,0 };
            const unsigned long dim684JoeKuoD7Init[]    =   {   1   ,   1   ,   7   ,   11  ,   31  ,   49  ,   71  ,   175 ,   231 ,   81  ,   155 ,   3257    ,   173 ,0 };
            const unsigned long dim685JoeKuoD7Init[]    =   {   1   ,   3   ,   5   ,   13  ,   15  ,   27  ,   71  ,   43  ,   267 ,   611 ,   2041    ,   1127    ,   3905    ,0 };
            const unsigned long dim686JoeKuoD7Init[]    =   {   1   ,   1   ,   7   ,   15  ,   3   ,   63  ,   25  ,   129 ,   409 ,   837 ,   1601    ,   2387    ,   5063    ,0 };
            const unsigned long dim687JoeKuoD7Init[]    =   {   1   ,   3   ,   3   ,   7   ,   19  ,   63  ,   27  ,   141 ,   423 ,   751 ,   669 ,   1945    ,   3015    ,0 };
            const unsigned long dim688JoeKuoD7Init[]    =   {   1   ,   1   ,   3   ,   11  ,   15  ,   59  ,   103 ,   123 ,   49  ,   71  ,   929 ,   3495    ,   3591    ,0 };
            const unsigned long dim689JoeKuoD7Init[]    =   {   1   ,   1   ,   3   ,   5   ,   9   ,   17  ,   49  ,   3   ,   67  ,   63  ,   1645    ,   585 ,   83  ,0 };
            const unsigned long dim690JoeKuoD7Init[]    =   {   1   ,   3   ,   3   ,   3   ,   25  ,   7   ,   103 ,   109 ,   241 ,   501 ,   2043    ,   139 ,   3865    ,0 };
            const unsigned long dim691JoeKuoD7Init[]    =   {   1   ,   3   ,   1   ,   5   ,   19  ,   41  ,   35  ,   223 ,   351 ,   853 ,   961 ,   2963    ,   3315    ,0 };
            const unsigned long dim692JoeKuoD7Init[]    =   {   1   ,   1   ,   3   ,   11  ,   9   ,   61  ,   103 ,   95  ,   393 ,   131 ,   611 ,   177 ,   635 ,0 };
            const unsigned long dim693JoeKuoD7Init[]    =   {   1   ,   1   ,   7   ,   9   ,   15  ,   27  ,   107 ,   35  ,   211 ,   299 ,   923 ,   2629    ,   2585    ,0 };
            const unsigned long dim694JoeKuoD7Init[]    =   {   1   ,   1   ,   3   ,   3   ,   9   ,   57  ,   89  ,   161 ,   323 ,   123 ,   1969    ,   3545    ,   5149    ,0 };
            const unsigned long dim695JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   9   ,   25  ,   15  ,   39  ,   7   ,   81  ,   503 ,   1549    ,   227 ,   7267    ,0 };
            const unsigned long dim696JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   1   ,   29  ,   41  ,   77  ,   131 ,   181 ,   555 ,   1165    ,   563 ,   5197    ,0 };
            const unsigned long dim697JoeKuoD7Init[]    =   {   1   ,   3   ,   3   ,   11  ,   15  ,   11  ,   69  ,   183 ,   113 ,   651 ,   1265    ,   1741    ,   8025    ,0 };
            const unsigned long dim698JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   13  ,   15  ,   61  ,   113 ,   181 ,   305 ,   669 ,   1903    ,   575 ,   6561    ,0 };
            const unsigned long dim699JoeKuoD7Init[]    =   {   1   ,   1   ,   3   ,   1   ,   27  ,   63  ,   103 ,   21  ,   387 ,   803 ,   1611    ,   1793    ,   1011    ,0 };
            const unsigned long dim700JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   9   ,   19  ,   41  ,   61  ,   153 ,   387 ,   509 ,   1771    ,   1715    ,   4643    ,0 };
            const unsigned long dim701JoeKuoD7Init[]    =   {   1   ,   3   ,   1   ,   9   ,   9   ,   23  ,   99  ,   157 ,   185 ,   457 ,   1485    ,   3153    ,   6361    ,0 };
            const unsigned long dim702JoeKuoD7Init[]    =   {   1   ,   1   ,   3   ,   9   ,   15  ,   3   ,   71  ,   95  ,   353 ,   549 ,   1269    ,   3451    ,   403 ,0 };
            const unsigned long dim703JoeKuoD7Init[]    =   {   1   ,   3   ,   5   ,   5   ,   17  ,   45  ,   49  ,   249 ,   185 ,   367 ,   1499    ,   2921    ,   7485    ,0 };
            const unsigned long dim704JoeKuoD7Init[]    =   {   1   ,   3   ,   1   ,   13  ,   1   ,   61  ,   57  ,   95  ,   213 ,   333 ,   603 ,   697 ,   6737    ,0 };
            const unsigned long dim705JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   7   ,   31  ,   37  ,   45  ,   71  ,   339 ,   629 ,   1181    ,   979 ,   1473    ,0 };
            const unsigned long dim706JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   5   ,   7   ,   5   ,   9   ,   139 ,   89  ,   683 ,   569 ,   1887    ,   2173    ,0 };
            const unsigned long dim707JoeKuoD7Init[]    =   {   1   ,   3   ,   1   ,   11  ,   19  ,   7   ,   33  ,   65  ,   299 ,   791 ,   1281    ,   3793    ,   7019    ,0 };
            const unsigned long dim708JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   7   ,   5   ,   45  ,   11  ,   247 ,   335 ,   199 ,   17  ,   2829    ,   2811    ,0 };
            const unsigned long dim709JoeKuoD7Init[]    =   {   1   ,   1   ,   7   ,   7   ,   9   ,   37  ,   17  ,   37  ,   385 ,   81  ,   1455    ,   1921    ,   6957    ,0 };
            const unsigned long dim710JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   1   ,   23  ,   53  ,   13  ,   7   ,   475 ,   167 ,   1573    ,   2235    ,   4097    ,0 };
            const unsigned long dim711JoeKuoD7Init[]    =   {   1   ,   3   ,   3   ,   9   ,   23  ,   53  ,   107 ,   63  ,   457 ,   723 ,   53  ,   3025    ,   3153    ,0 };
            const unsigned long dim712JoeKuoD7Init[]    =   {   1   ,   3   ,   1   ,   5   ,   21  ,   39  ,   1   ,   123 ,   293 ,   275 ,   1441    ,   2897    ,   4645    ,0 };
            const unsigned long dim713JoeKuoD7Init[]    =   {   1   ,   3   ,   3   ,   7   ,   21  ,   5   ,   93  ,   63  ,   427 ,   47  ,   559 ,   2383    ,   4137    ,0 };
            const unsigned long dim714JoeKuoD7Init[]    =   {   1   ,   3   ,   3   ,   1   ,   13  ,   39  ,   81  ,   63  ,   353 ,   937 ,   679 ,   79  ,   3319    ,0 };
            const unsigned long dim715JoeKuoD7Init[]    =   {   1   ,   1   ,   7   ,   3   ,   19  ,   3   ,   57  ,   207 ,   115 ,   439 ,   835 ,   389 ,   4413    ,0 };
            const unsigned long dim716JoeKuoD7Init[]    =   {   1   ,   3   ,   5   ,   7   ,   21  ,   57  ,   81  ,   137 ,   265 ,   845 ,   843 ,   3083    ,   4453    ,0 };
            const unsigned long dim717JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   13  ,   13  ,   29  ,   77  ,   181 ,   483 ,   871 ,   1739    ,   3859    ,   5005    ,0 };
            const unsigned long dim718JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   13  ,   9   ,   45  ,   73  ,   73  ,   427 ,   783 ,   617 ,   1771    ,   5953    ,0 };
            const unsigned long dim719JoeKuoD7Init[]    =   {   1   ,   1   ,   5   ,   13  ,   31  ,   39  ,   83  ,   17  ,   275 ,   419 ,   1405    ,   3189    ,   7751    ,0 };
            const unsigned long dim720JoeKuoD7Init[]    =   {   1   ,   1   ,   5   ,   13  ,   15  ,   17  ,   15  ,   67  ,   455 ,   477 ,   1911    ,   2225    ,   4095    ,0 };
            const unsigned long dim721JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   1   ,   5   ,   39  ,   73  ,   205 ,   331 ,   819 ,   1499    ,   1243    ,   7369    ,0 };
            const unsigned long dim722JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   1   ,   1   ,   49  ,   11  ,   167 ,   489 ,   169 ,   1683    ,   2999    ,   4371    ,0 };
            const unsigned long dim723JoeKuoD7Init[]    =   {   1   ,   1   ,   7   ,   9   ,   13  ,   7   ,   29  ,   105 ,   155 ,   85  ,   1311    ,   617 ,   1793    ,0 };
            const unsigned long dim724JoeKuoD7Init[]    =   {   1   ,   1   ,   5   ,   5   ,   13  ,   3   ,   71  ,   27  ,   83  ,   145 ,   89  ,   691 ,   1081    ,0 };
            const unsigned long dim725JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   1   ,   9   ,   43  ,   23  ,   63  ,   43  ,   263 ,   1849    ,   1963    ,   1043    ,0 };
            const unsigned long dim726JoeKuoD7Init[]    =   {   1   ,   3   ,   3   ,   1   ,   25  ,   15  ,   17  ,   211 ,   453 ,   385 ,   349 ,   789 ,   787 ,0 };
            const unsigned long dim727JoeKuoD7Init[]    =   {   1   ,   3   ,   5   ,   11  ,   1   ,   63  ,   83  ,   63  ,   317 ,   841 ,   1779    ,   669 ,   253 ,0 };
            const unsigned long dim728JoeKuoD7Init[]    =   {   1   ,   3   ,   3   ,   5   ,   7   ,   49  ,   67  ,   69  ,   167 ,   29  ,   217 ,   1759    ,   6167    ,0 };
            const unsigned long dim729JoeKuoD7Init[]    =   {   1   ,   1   ,   7   ,   1   ,   15  ,   63  ,   43  ,   3   ,   183 ,   577 ,   1301    ,   1479    ,   4189    ,0 };
            const unsigned long dim730JoeKuoD7Init[]    =   {   1   ,   3   ,   1   ,   13  ,   29  ,   57  ,   103 ,   253 ,   45  ,   661 ,   1593    ,   1993    ,   523 ,0 };
            const unsigned long dim731JoeKuoD7Init[]    =   {   1   ,   3   ,   3   ,   15  ,   25  ,   17  ,   11  ,   71  ,   213 ,   287 ,   627 ,   2699    ,   4909    ,0 };
            const unsigned long dim732JoeKuoD7Init[]    =   {   1   ,   1   ,   3   ,   9   ,   11  ,   53  ,   117 ,   59  ,   225 ,   469 ,   109 ,   583 ,   4677    ,0 };
            const unsigned long dim733JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   15  ,   27  ,   41  ,   109 ,   37  ,   153 ,   885 ,   1339    ,   2783    ,   7833    ,0 };
            const unsigned long dim734JoeKuoD7Init[]    =   {   1   ,   3   ,   1   ,   7   ,   5   ,   47  ,   5   ,   233 ,   507 ,   953 ,   1359    ,   2505    ,   6943    ,0 };
            const unsigned long dim735JoeKuoD7Init[]    =   {   1   ,   1   ,   7   ,   5   ,   11  ,   21  ,   33  ,   207 ,   17  ,   621 ,   1943    ,   597 ,   3993    ,0 };
            const unsigned long dim736JoeKuoD7Init[]    =   {   1   ,   3   ,   5   ,   3   ,   13  ,   55  ,   11  ,   247 ,   321 ,   533 ,   715 ,   2941    ,   6435    ,0 };
            const unsigned long dim737JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   15  ,   19  ,   3   ,   59  ,   1   ,   473 ,   17  ,   691 ,   747 ,   3917    ,0 };
            const unsigned long dim738JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   1   ,   25  ,   49  ,   11  ,   185 ,   185 ,   821 ,   1857    ,   3927    ,   2641    ,0 };
            const unsigned long dim739JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   11  ,   27  ,   31  ,   47  ,   33  ,   63  ,   935 ,   1709    ,   1495    ,   3869    ,0 };
            const unsigned long dim740JoeKuoD7Init[]    =   {   1   ,   3   ,   3   ,   3   ,   23  ,   35  ,   73  ,   171 ,   207 ,   535 ,   751 ,   3693    ,   3015    ,0 };
            const unsigned long dim741JoeKuoD7Init[]    =   {   1   ,   3   ,   5   ,   9   ,   11  ,   9   ,   101 ,   183 ,   83  ,   507 ,   15  ,   3993    ,   3631    ,0 };
            const unsigned long dim742JoeKuoD7Init[]    =   {   1   ,   1   ,   3   ,   1   ,   5   ,   13  ,   41  ,   239 ,   355 ,   463 ,   997 ,   1163    ,   651 ,0 };
            const unsigned long dim743JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   11  ,   19  ,   25  ,   21  ,   233 ,   315 ,   1009    ,   1433    ,   2281    ,   4685    ,0 };
            const unsigned long dim744JoeKuoD7Init[]    =   {   1   ,   1   ,   5   ,   5   ,   17  ,   11  ,   31  ,   47  ,   47  ,   873 ,   1957    ,   2307    ,   1757    ,0 };
            const unsigned long dim745JoeKuoD7Init[]    =   {   1   ,   1   ,   3   ,   15  ,   1   ,   43  ,   93  ,   27  ,   315 ,   357 ,   787 ,   3843    ,   4615    ,0 };
            const unsigned long dim746JoeKuoD7Init[]    =   {   1   ,   3   ,   1   ,   13  ,   13  ,   41  ,   67  ,   137 ,   241 ,   707 ,   1207    ,   1923    ,   2753    ,0 };
            const unsigned long dim747JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   3   ,   1   ,   15  ,   21  ,   175 ,   21  ,   917 ,   625 ,   543 ,   7235    ,0 };
            const unsigned long dim748JoeKuoD7Init[]    =   {   1   ,   1   ,   7   ,   15  ,   3   ,   49  ,   97  ,   195 ,   443 ,   689 ,   1393    ,   2385    ,   4111    ,0 };
            const unsigned long dim749JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   3   ,   19  ,   53  ,   55  ,   243 ,   491 ,   213 ,   1371    ,   3867    ,   5323    ,0 };
            const unsigned long dim750JoeKuoD7Init[]    =   {   1   ,   3   ,   1   ,   1   ,   9   ,   47  ,   7   ,   169 ,   151 ,   53  ,   1481    ,   3689    ,   2539    ,0 };
            const unsigned long dim751JoeKuoD7Init[]    =   {   1   ,   1   ,   7   ,   9   ,   31  ,   7   ,   49  ,   27  ,   57  ,   769 ,   687 ,   1007    ,   3619    ,0 };
            const unsigned long dim752JoeKuoD7Init[]    =   {   1   ,   3   ,   3   ,   9   ,   1   ,   49  ,   29  ,   153 ,   273 ,   965 ,   719 ,   4023    ,   1755    ,0 };
            const unsigned long dim753JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   1   ,   23  ,   29  ,   85  ,   179 ,   137 ,   237 ,   1639    ,   3759    ,   5533    ,0 };
            const unsigned long dim754JoeKuoD7Init[]    =   {   1   ,   3   ,   1   ,   3   ,   7   ,   3   ,   1   ,   107 ,   259 ,   827 ,   799 ,   2677    ,   1677    ,0 };
            const unsigned long dim755JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   5   ,   9   ,   51  ,   89  ,   255 ,   97  ,   419 ,   703 ,   3903    ,   833 ,0 };
            const unsigned long dim756JoeKuoD7Init[]    =   {   1   ,   3   ,   5   ,   3   ,   21  ,   27  ,   115 ,   75  ,   329 ,   747 ,   1949    ,   1309    ,   3611    ,0 };
            const unsigned long dim757JoeKuoD7Init[]    =   {   1   ,   3   ,   1   ,   3   ,   13  ,   3   ,   21  ,   5   ,   337 ,   757 ,   1765    ,   667 ,   5607    ,0 };
            const unsigned long dim758JoeKuoD7Init[]    =   {   1   ,   1   ,   5   ,   1   ,   19  ,   29  ,   113 ,   7   ,   51  ,   283 ,   1987    ,   3939    ,   583 ,0 };
            const unsigned long dim759JoeKuoD7Init[]    =   {   1   ,   1   ,   3   ,   3   ,   23  ,   13  ,   9   ,   187 ,   251 ,   899 ,   85  ,   701 ,   2597    ,0 };
            const unsigned long dim760JoeKuoD7Init[]    =   {   1   ,   3   ,   3   ,   7   ,   1   ,   15  ,   83  ,   183 ,   471 ,   533 ,   1119    ,   3157    ,   827 ,0 };
            const unsigned long dim761JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   11  ,   27  ,   49  ,   117 ,   57  ,   337 ,   603 ,   1247    ,   3957    ,   1739    ,0 };
            const unsigned long dim762JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   9   ,   29  ,   23  ,   123 ,   215 ,   479 ,   459 ,   361 ,   305 ,   237 ,0 };
            const unsigned long dim763JoeKuoD7Init[]    =   {   1   ,   3   ,   5   ,   11  ,   1   ,   27  ,   103 ,   241 ,   493 ,   441 ,   1171    ,   3625    ,   3369    ,0 };
            const unsigned long dim764JoeKuoD7Init[]    =   {   1   ,   3   ,   3   ,   9   ,   15  ,   59  ,   13  ,   225 ,   485 ,   493 ,   97  ,   2705    ,   7077    ,0 };
            const unsigned long dim765JoeKuoD7Init[]    =   {   1   ,   3   ,   3   ,   1   ,   5   ,   27  ,   85  ,   135 ,   451 ,   745 ,   613 ,   1443    ,   6047    ,0 };
            const unsigned long dim766JoeKuoD7Init[]    =   {   1   ,   1   ,   5   ,   7   ,   5   ,   41  ,   57  ,   167 ,   105 ,   721 ,   365 ,   833 ,   4485    ,0 };
            const unsigned long dim767JoeKuoD7Init[]    =   {   1   ,   1   ,   3   ,   1   ,   19  ,   13  ,   119 ,   161 ,   475 ,   935 ,   1707    ,   2971    ,   6797    ,0 };
            const unsigned long dim768JoeKuoD7Init[]    =   {   1   ,   3   ,   1   ,   15  ,   19  ,   59  ,   23  ,   29  ,   339 ,   671 ,   2027    ,   545 ,   6197    ,0 };
            const unsigned long dim769JoeKuoD7Init[]    =   {   1   ,   1   ,   7   ,   3   ,   7   ,   41  ,   3   ,   169 ,   479 ,   567 ,   865 ,   2481    ,   3979    ,0 };
            const unsigned long dim770JoeKuoD7Init[]    =   {   1   ,   3   ,   1   ,   13  ,   25  ,   23  ,   59  ,   175 ,   17  ,   291 ,   1283    ,   1405    ,   4939    ,0 };
            const unsigned long dim771JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   13  ,   13  ,   21  ,   127 ,   75  ,   275 ,   549 ,   971 ,   2019    ,   199 ,0 };
            const unsigned long dim772JoeKuoD7Init[]    =   {   1   ,   1   ,   3   ,   3   ,   9   ,   61  ,   47  ,   23  ,   51  ,   589 ,   73  ,   265 ,   7315    ,0 };
            const unsigned long dim773JoeKuoD7Init[]    =   {   1   ,   1   ,   5   ,   1   ,   3   ,   21  ,   15  ,   33  ,   443 ,   831 ,   375 ,   3045    ,   4215    ,0 };
            const unsigned long dim774JoeKuoD7Init[]    =   {   1   ,   1   ,   7   ,   1   ,   27  ,   47  ,   123 ,   9   ,   131 ,   487 ,   393 ,   2959    ,   1509    ,0 };
            const unsigned long dim775JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   9   ,   15  ,   49  ,   61  ,   77  ,   65  ,   411 ,   293 ,   3827    ,   5723    ,0 };
            const unsigned long dim776JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   9   ,   17  ,   33  ,   97  ,   207 ,   227 ,   679 ,   1157    ,   1237    ,   2203    ,0 };
            const unsigned long dim777JoeKuoD7Init[]    =   {   1   ,   1   ,   5   ,   1   ,   31  ,   31  ,   23  ,   59  ,   363 ,   225 ,   827 ,   2757    ,   1979    ,0 };
            const unsigned long dim778JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   11  ,   13  ,   31  ,   19  ,   27  ,   225 ,   901 ,   727 ,   2491    ,   3075    ,0 };
            const unsigned long dim779JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   9   ,   17  ,   9   ,   63  ,   117 ,   331 ,   887 ,   407 ,   111 ,   125 ,0 };
            const unsigned long dim780JoeKuoD7Init[]    =   {   1   ,   3   ,   3   ,   13  ,   5   ,   25  ,   37  ,   61  ,   241 ,   15  ,   559 ,   1981    ,   3585    ,0 };
            const unsigned long dim781JoeKuoD7Init[]    =   {   1   ,   1   ,   3   ,   9   ,   13  ,   57  ,   9   ,   189 ,   135 ,   485 ,   105 ,   327 ,   7233    ,0 };
            const unsigned long dim782JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   11  ,   11  ,   37  ,   13  ,   139 ,   93  ,   233 ,   1171    ,   165 ,   4151    ,0 };
            const unsigned long dim783JoeKuoD7Init[]    =   {   1   ,   1   ,   3   ,   9   ,   7   ,   61  ,   81  ,   243 ,   111 ,   681 ,   859 ,   2163    ,   7019    ,0 };
            const unsigned long dim784JoeKuoD7Init[]    =   {   1   ,   3   ,   1   ,   5   ,   31  ,   57  ,   85  ,   105 ,   323 ,   729 ,   131 ,   2055    ,   1893    ,0 };
            const unsigned long dim785JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   15  ,   17  ,   55  ,   35  ,   211 ,   341 ,   281 ,   613 ,   697 ,   6479    ,0 };
            const unsigned long dim786JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   1   ,   9   ,   35  ,   107 ,   47  ,   243 ,   917 ,   683 ,   1333    ,   7797    ,0 };
            const unsigned long dim787JoeKuoD7Init[]    =   {   1   ,   3   ,   5   ,   7   ,   31  ,   15  ,   9   ,   11  ,   109 ,   509 ,   973 ,   3817    ,   2877    ,0 };
            const unsigned long dim788JoeKuoD7Init[]    =   {   1   ,   1   ,   3   ,   15  ,   31  ,   21  ,   43  ,   79  ,   367 ,   201 ,   549 ,   1335    ,   7231    ,0 };
            const unsigned long dim789JoeKuoD7Init[]    =   {   1   ,   1   ,   3   ,   5   ,   23  ,   21  ,   63  ,   207 ,   333 ,   395 ,   925 ,   705 ,   8143    ,0 };
            const unsigned long dim790JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   5   ,   13  ,   23  ,   93  ,   123 ,   317 ,   511 ,   867 ,   1791    ,   5621    ,0 };
            const unsigned long dim791JoeKuoD7Init[]    =   {   1   ,   3   ,   3   ,   3   ,   1   ,   1   ,   61  ,   133 ,   23  ,   885 ,   1281    ,   3115    ,   2717    ,0 };
            const unsigned long dim792JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   13  ,   5   ,   3   ,   59  ,   59  ,   453 ,   787 ,   1933    ,   2813    ,   6405    ,0 };
            const unsigned long dim793JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   9   ,   7   ,   39  ,   85  ,   65  ,   33  ,   85  ,   1597    ,   17  ,   3779    ,0 };
            const unsigned long dim794JoeKuoD7Init[]    =   {   1   ,   1   ,   7   ,   1   ,   3   ,   33  ,   85  ,   129 ,   249 ,   845 ,   1481    ,   3249    ,   7207    ,0 };
            const unsigned long dim795JoeKuoD7Init[]    =   {   1   ,   1   ,   7   ,   3   ,   9   ,   5   ,   51  ,   79  ,   241 ,   173 ,   567 ,   3047    ,   6779    ,0 };
            const unsigned long dim796JoeKuoD7Init[]    =   {   1   ,   3   ,   1   ,   1   ,   1   ,   49  ,   29  ,   243 ,   95  ,   609 ,   1887    ,   407 ,   7875    ,0 };
            const unsigned long dim797JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   15  ,   29  ,   55  ,   69  ,   171 ,   337 ,   701 ,   1255    ,   1889    ,   3365    ,0 };
            const unsigned long dim798JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   5   ,   11  ,   27  ,   3   ,   167 ,   325 ,   397 ,   545 ,   2097    ,   1323    ,0 };
            const unsigned long dim799JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   11  ,   7   ,   3   ,   17  ,   47  ,   227 ,   573 ,   713 ,   975 ,   5991    ,0 };
            const unsigned long dim800JoeKuoD7Init[]    =   {   1   ,   3   ,   1   ,   15  ,   7   ,   55  ,   71  ,   41  ,   195 ,   227 ,   821 ,   3209    ,   1363    ,0 };
            const unsigned long dim801JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   9   ,   7   ,   49  ,   105 ,   5   ,   441 ,   629 ,   201 ,   2951    ,   5757    ,0 };
            const unsigned long dim802JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   7   ,   13  ,   5   ,   5   ,   31  ,   355 ,   107 ,   1937    ,   3553    ,   697 ,0 };
            const unsigned long dim803JoeKuoD7Init[]    =   {   1   ,   3   ,   3   ,   7   ,   23  ,   59  ,   109 ,   185 ,   289 ,   257 ,   1055    ,   3221    ,   6445    ,0 };
            const unsigned long dim804JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   7   ,   3   ,   27  ,   61  ,   85  ,   383 ,   93  ,   639 ,   1113    ,   1511    ,0 };
            const unsigned long dim805JoeKuoD7Init[]    =   {   1   ,   3   ,   3   ,   7   ,   17  ,   29  ,   31  ,   255 ,   57  ,   557 ,   721 ,   4015    ,   7587    ,0 };
            const unsigned long dim806JoeKuoD7Init[]    =   {   1   ,   1   ,   7   ,   9   ,   23  ,   15  ,   89  ,   85  ,   201 ,   177 ,   571 ,   2691    ,   4521    ,0 };
            const unsigned long dim807JoeKuoD7Init[]    =   {   1   ,   1   ,   5   ,   9   ,   9   ,   25  ,   91  ,   75  ,   405 ,   13  ,   587 ,   2299    ,   7879    ,0 };
            const unsigned long dim808JoeKuoD7Init[]    =   {   1   ,   1   ,   5   ,   15  ,   3   ,   59  ,   109 ,   175 ,   411 ,   309 ,   787 ,   2979    ,   5385    ,0 };
            const unsigned long dim809JoeKuoD7Init[]    =   {   1   ,   3   ,   5   ,   1   ,   9   ,   21  ,   109 ,   109 ,   179 ,   955 ,   423 ,   3637    ,   3283    ,0 };
            const unsigned long dim810JoeKuoD7Init[]    =   {   1   ,   3   ,   1   ,   1   ,   7   ,   51  ,   31  ,   83  ,   273 ,   545 ,   1063    ,   1463    ,   1857    ,0 };
            const unsigned long dim811JoeKuoD7Init[]    =   {   1   ,   3   ,   5   ,   9   ,   1   ,   51  ,   49  ,   241 ,   365 ,   423 ,   1299    ,   3533    ,   1771    ,0 };
            const unsigned long dim812JoeKuoD7Init[]    =   {   1   ,   1   ,   7   ,   15  ,   25  ,   21  ,   95  ,   221 ,   65  ,   335 ,   697 ,   759 ,   4331    ,0 };
            const unsigned long dim813JoeKuoD7Init[]    =   {   1   ,   1   ,   5   ,   11  ,   15  ,   53  ,   17  ,   7   ,   461 ,   543 ,   233 ,   3993    ,   725 ,0 };
            const unsigned long dim814JoeKuoD7Init[]    =   {   1   ,   3   ,   5   ,   1   ,   13  ,   3   ,   53  ,   213 ,   5   ,   709 ,   883 ,   3261    ,   855 ,0 };
            const unsigned long dim815JoeKuoD7Init[]    =   {   1   ,   3   ,   1   ,   13  ,   1   ,   33  ,   61  ,   255 ,   385 ,   793 ,   525 ,   929 ,   4263    ,0 };
            const unsigned long dim816JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   11  ,   1   ,   33  ,   97  ,   235 ,   427 ,   1021    ,   1683    ,   2821    ,   1347    ,0 };
            const unsigned long dim817JoeKuoD7Init[]    =   {   1   ,   3   ,   1   ,   9   ,   15  ,   7   ,   39  ,   163 ,   13  ,   535 ,   1259    ,   653 ,   7359    ,0 };
            const unsigned long dim818JoeKuoD7Init[]    =   {   1   ,   3   ,   5   ,   15  ,   31  ,   27  ,   19  ,   255 ,   341 ,   509 ,   1775    ,   1493    ,   6977    ,0 };
            const unsigned long dim819JoeKuoD7Init[]    =   {   1   ,   1   ,   5   ,   15  ,   3   ,   23  ,   53  ,   221 ,   271 ,   729 ,   1013    ,   101 ,   2837    ,0 };
            const unsigned long dim820JoeKuoD7Init[]    =   {   1   ,   1   ,   7   ,   3   ,   19  ,   63  ,   83  ,   149 ,   385 ,   813 ,   1783    ,   4021    ,   3645    ,0 };
            const unsigned long dim821JoeKuoD7Init[]    =   {   1   ,   1   ,   5   ,   7   ,   11  ,   57  ,   121 ,   97  ,   479 ,   33  ,   579 ,   1009    ,   1821    ,0 };
            const unsigned long dim822JoeKuoD7Init[]    =   {   1   ,   3   ,   5   ,   15  ,   5   ,   25  ,   69  ,   159 ,   441 ,   661 ,   151 ,   103 ,   2659    ,0 };
            const unsigned long dim823JoeKuoD7Init[]    =   {   1   ,   1   ,   5   ,   7   ,   11  ,   7   ,   15  ,   159 ,   19  ,   457 ,   1873    ,   3017    ,   3927    ,0 };
            const unsigned long dim824JoeKuoD7Init[]    =   {   1   ,   1   ,   5   ,   11  ,   9   ,   43  ,   73  ,   187 ,   107 ,   603 ,   387 ,   3713    ,   853 ,0 };
            const unsigned long dim825JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   15  ,   29  ,   55  ,   87  ,   247 ,   283 ,   11  ,   927 ,   3313    ,   5147    ,0 };
            const unsigned long dim826JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   5   ,   27  ,   25  ,   1   ,   123 ,   19  ,   237 ,   1171    ,   1731    ,   7079    ,0 };
            const unsigned long dim827JoeKuoD7Init[]    =   {   1   ,   1   ,   5   ,   7   ,   29  ,   43  ,   15  ,   91  ,   61  ,   329 ,   1343    ,   1349    ,   471 ,0 };
            const unsigned long dim828JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   11  ,   11  ,   7   ,   25  ,   143 ,   277 ,   791 ,   775 ,   3711    ,   7115    ,0 };
            const unsigned long dim829JoeKuoD7Init[]    =   {   1   ,   1   ,   5   ,   5   ,   29  ,   27  ,   53  ,   241 ,   399 ,   165 ,   1347    ,   1661    ,   4495    ,0 };
            const unsigned long dim830JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   3   ,   21  ,   7   ,   73  ,   61  ,   155 ,   635 ,   1095    ,   3295    ,   7109    ,0 };
            const unsigned long dim831JoeKuoD7Init[]    =   {   1   ,   1   ,   3   ,   11  ,   5   ,   39  ,   11  ,   167 ,   317 ,   317 ,   161 ,   2141    ,   1327    ,0 };
            const unsigned long dim832JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   7   ,   23  ,   45  ,   59  ,   13  ,   107 ,   47  ,   2001    ,   1491    ,   4463    ,0 };
            const unsigned long dim833JoeKuoD7Init[]    =   {   1   ,   3   ,   3   ,   15  ,   27  ,   63  ,   41  ,   143 ,   503 ,   711 ,   277 ,   2523    ,   6331    ,0 };
            const unsigned long dim834JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   15  ,   13  ,   49  ,   121 ,   107 ,   145 ,   769 ,   699 ,   3793    ,   2333    ,0 };
            const unsigned long dim835JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   11  ,   17  ,   7   ,   57  ,   139 ,   17  ,   433 ,   1053    ,   1279    ,   833 ,0 };
            const unsigned long dim836JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   3   ,   21  ,   19  ,   61  ,   229 ,   271 ,   831 ,   495 ,   1663    ,   2109    ,0 };
            const unsigned long dim837JoeKuoD7Init[]    =   {   1   ,   3   ,   3   ,   11  ,   17  ,   27  ,   23  ,   153 ,   151 ,   309 ,   1935    ,   3993    ,   4149    ,0 };
            const unsigned long dim838JoeKuoD7Init[]    =   {   1   ,   3   ,   1   ,   1   ,   11  ,   29  ,   33  ,   77  ,   357 ,   543 ,   217 ,   893 ,   2989    ,0 };
            const unsigned long dim839JoeKuoD7Init[]    =   {   1   ,   3   ,   3   ,   13  ,   21  ,   25  ,   121 ,   211 ,   491 ,   983 ,   1395    ,   2879    ,   1173    ,0 };
            const unsigned long dim840JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   11  ,   31  ,   61  ,   45  ,   125 ,   21  ,   175 ,   1869    ,   403 ,   579 ,0 };
            const unsigned long dim841JoeKuoD7Init[]    =   {   1   ,   3   ,   1   ,   11  ,   27  ,   5   ,   47  ,   211 ,   149 ,   753 ,   301 ,   933 ,   3339    ,0 };
            const unsigned long dim842JoeKuoD7Init[]    =   {   1   ,   1   ,   5   ,   15  ,   7   ,   57  ,   105 ,   243 ,   27  ,   629 ,   337 ,   1917    ,   2653    ,0 };
            const unsigned long dim843JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   1   ,   27  ,   27  ,   45  ,   39  ,   249 ,   167 ,   2031    ,   3427    ,   6051    ,0 };
            const unsigned long dim844JoeKuoD7Init[]    =   {   1   ,   1   ,   5   ,   5   ,   9   ,   13  ,   55  ,   167 ,   473 ,   371 ,   677 ,   2171    ,   1859    ,0 };
            const unsigned long dim845JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   13  ,   21  ,   49  ,   25  ,   193 ,   507 ,   531 ,   1639    ,   3181    ,   2763    ,0 };
            const unsigned long dim846JoeKuoD7Init[]    =   {   1   ,   3   ,   3   ,   11  ,   15  ,   3   ,   115 ,   163 ,   63  ,   383 ,   835 ,   2507    ,   5525    ,0 };
            const unsigned long dim847JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   1   ,   1   ,   63  ,   59  ,   75  ,   293 ,   609 ,   59  ,   1729    ,   4909    ,0 };
            const unsigned long dim848JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   15  ,   11  ,   57  ,   59  ,   109 ,   465 ,   285 ,   1973    ,   505 ,   4447    ,0 };
            const unsigned long dim849JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   13  ,   11  ,   63  ,   87  ,   149 ,   185 ,   493 ,   831 ,   2927    ,   11  ,0 };
            const unsigned long dim850JoeKuoD7Init[]    =   {   1   ,   1   ,   3   ,   3   ,   3   ,   19  ,   111 ,   199 ,   371 ,   411 ,   659 ,   1191    ,   6279    ,0 };
            const unsigned long dim851JoeKuoD7Init[]    =   {   1   ,   1   ,   5   ,   9   ,   21  ,   57  ,   73  ,   189 ,   307 ,   515 ,   1797    ,   2073    ,   1085    ,0 };
            const unsigned long dim852JoeKuoD7Init[]    =   {   1   ,   1   ,   5   ,   9   ,   23  ,   21  ,   21  ,   23  ,   487 ,   715 ,   449 ,   2157    ,   5063    ,0 };
            const unsigned long dim853JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   11  ,   27  ,   27  ,   121 ,   61  ,   281 ,   805 ,   55  ,   2111    ,   2193    ,0 };
            const unsigned long dim854JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   3   ,   5   ,   45  ,   23  ,   71  ,   369 ,   875 ,   1185    ,   1611    ,   2009    ,0 };
            const unsigned long dim855JoeKuoD7Init[]    =   {   1   ,   3   ,   1   ,   15  ,   17  ,   59  ,   17  ,   99  ,   403 ,   103 ,   847 ,   849 ,   569 ,0 };
            const unsigned long dim856JoeKuoD7Init[]    =   {   1   ,   3   ,   5   ,   7   ,   7   ,   13  ,   47  ,   125 ,   35  ,   615 ,   281 ,   1443    ,   1199    ,0 };
            const unsigned long dim857JoeKuoD7Init[]    =   {   1   ,   3   ,   3   ,   15  ,   13  ,   63  ,   73  ,   191 ,   137 ,   257 ,   1609    ,   409 ,   5135    ,0 };
            const unsigned long dim858JoeKuoD7Init[]    =   {   1   ,   3   ,   3   ,   11  ,   13  ,   57  ,   63  ,   189 ,   367 ,   7   ,   1261    ,   211 ,   2223    ,0 };
            const unsigned long dim859JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   3   ,   31  ,   21  ,   117 ,   135 ,   231 ,   981 ,   857 ,   3737    ,   1631    ,0 };
            const unsigned long dim860JoeKuoD7Init[]    =   {   1   ,   1   ,   3   ,   5   ,   11  ,   13  ,   55  ,   193 ,   99  ,   969 ,   1093    ,   1575    ,   2367    ,0 };
            const unsigned long dim861JoeKuoD7Init[]    =   {   1   ,   3   ,   5   ,   9   ,   9   ,   45  ,   81  ,   139 ,   479 ,   561 ,   1983    ,   3117    ,   2539    ,0 };
            const unsigned long dim862JoeKuoD7Init[]    =   {   1   ,   1   ,   3   ,   11  ,   19  ,   37  ,   107 ,   125 ,   415 ,   653 ,   517 ,   2191    ,   415 ,0 };
            const unsigned long dim863JoeKuoD7Init[]    =   {   1   ,   3   ,   3   ,   9   ,   13  ,   41  ,   59  ,   193 ,   205 ,   485 ,   1439    ,   1333    ,   2519    ,0 };
            const unsigned long dim864JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   13  ,   9   ,   41  ,   95  ,   141 ,   21  ,   953 ,   1453    ,   3585    ,   2521    ,0 };
            const unsigned long dim865JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   11  ,   25  ,   21  ,   43  ,   221 ,   201 ,   761 ,   1457    ,   1597    ,   2147    ,0 };
            const unsigned long dim866JoeKuoD7Init[]    =   {   1   ,   1   ,   3   ,   9   ,   27  ,   63  ,   35  ,   229 ,   147 ,   293 ,   699 ,   673 ,   4779    ,0 };
            const unsigned long dim867JoeKuoD7Init[]    =   {   1   ,   3   ,   5   ,   13  ,   15  ,   57  ,   57  ,   197 ,   23  ,   925 ,   1433    ,   681 ,   1721    ,0 };
            const unsigned long dim868JoeKuoD7Init[]    =   {   1   ,   1   ,   5   ,   11  ,   17  ,   25  ,   127 ,   195 ,   487 ,   679 ,   1529    ,   453 ,   3955    ,0 };
            const unsigned long dim869JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   1   ,   29  ,   17  ,   49  ,   13  ,   299 ,   305 ,   11  ,   3913    ,   1633    ,0 };
            const unsigned long dim870JoeKuoD7Init[]    =   {   1   ,   3   ,   5   ,   3   ,   15  ,   15  ,   79  ,   233 ,   183 ,   515 ,   773 ,   1761    ,   5439    ,0 };
            const unsigned long dim871JoeKuoD7Init[]    =   {   1   ,   1   ,   5   ,   7   ,   23  ,   43  ,   3   ,   59  ,   37  ,   985 ,   1957    ,   2373    ,   7823    ,0 };
            const unsigned long dim872JoeKuoD7Init[]    =   {   1   ,   1   ,   5   ,   5   ,   21  ,   59  ,   47  ,   205 ,   69  ,   435 ,   1567    ,   2661    ,   3589    ,0 };
            const unsigned long dim873JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   1   ,   7   ,   29  ,   15  ,   61  ,   197 ,   339 ,   1955    ,   2815    ,   4343    ,0 };
            const unsigned long dim874JoeKuoD7Init[]    =   {   1   ,   3   ,   5   ,   5   ,   13  ,   33  ,   109 ,   33  ,   475 ,   219 ,   1707    ,   3747    ,   7359    ,0 };
            const unsigned long dim875JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   3   ,   21  ,   15  ,   53  ,   29  ,   493 ,   729 ,   1847    ,   1387    ,   1221    ,0 };
            const unsigned long dim876JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   11  ,   19  ,   27  ,   107 ,   29  ,   217 ,   19  ,   337 ,   2621    ,   1199    ,0 };
            const unsigned long dim877JoeKuoD7Init[]    =   {   1   ,   1   ,   5   ,   11  ,   11  ,   39  ,   111 ,   195 ,   373 ,   239 ,   355 ,   359 ,   1187    ,0 };
            const unsigned long dim878JoeKuoD7Init[]    =   {   1   ,   1   ,   3   ,   15  ,   29  ,   63  ,   13  ,   235 ,   413 ,   67  ,   1437    ,   3227    ,   5067    ,0 };
            const unsigned long dim879JoeKuoD7Init[]    =   {   1   ,   3   ,   3   ,   3   ,   29  ,   35  ,   11  ,   189 ,   15  ,   181 ,   1807    ,   1711    ,   3317    ,0 };
            const unsigned long dim880JoeKuoD7Init[]    =   {   1   ,   3   ,   5   ,   5   ,   25  ,   21  ,   93  ,   101 ,   65  ,   495 ,   375 ,   2873    ,   5253    ,0 };
            const unsigned long dim881JoeKuoD7Init[]    =   {   1   ,   1   ,   7   ,   3   ,   19  ,   47  ,   13  ,   229 ,   3   ,   377 ,   1769    ,   673 ,   5713    ,0 };
            const unsigned long dim882JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   15  ,   11  ,   19  ,   25  ,   249 ,   95  ,   777 ,   1763    ,   2607    ,   6439    ,0 };
            const unsigned long dim883JoeKuoD7Init[]    =   {   1   ,   1   ,   5   ,   1   ,   31  ,   47  ,   101 ,   155 ,   139 ,   201 ,   1687    ,   1503    ,   4783    ,0 };
            const unsigned long dim884JoeKuoD7Init[]    =   {   1   ,   1   ,   5   ,   13  ,   27  ,   11  ,   53  ,   199 ,   305 ,   481 ,   1071    ,   143 ,   3667    ,0 };
            const unsigned long dim885JoeKuoD7Init[]    =   {   1   ,   3   ,   5   ,   11  ,   15  ,   29  ,   95  ,   201 ,   271 ,   607 ,   511 ,   4029    ,   4269    ,0 };
            const unsigned long dim886JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   3   ,   29  ,   13  ,   99  ,   161 ,   55  ,   997 ,   1123    ,   3107    ,   4255    ,0 };
            const unsigned long dim887JoeKuoD7Init[]    =   {   1   ,   3   ,   5   ,   11  ,   13  ,   5   ,   121 ,   253 ,   315 ,   465 ,   851 ,   3147    ,   5415    ,0 };
            const unsigned long dim888JoeKuoD7Init[]    =   {   1   ,   3   ,   3   ,   13  ,   29  ,   15  ,   39  ,   7   ,   439 ,   787 ,   895 ,   1859    ,   5139    ,0 };
            const unsigned long dim889JoeKuoD7Init[]    =   {   1   ,   3   ,   5   ,   9   ,   29  ,   37  ,   61  ,   61  ,   321 ,   307 ,   1741    ,   3143    ,   1977    ,0 };
            const unsigned long dim890JoeKuoD7Init[]    =   {   1   ,   3   ,   3   ,   5   ,   19  ,   5   ,   105 ,   89  ,   189 ,   723 ,   407 ,   2557    ,   3817    ,0 };
            const unsigned long dim891JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   7   ,   15  ,   9   ,   93  ,   15  ,   461 ,   337 ,   1157    ,   531 ,   797 ,0 };
            const unsigned long dim892JoeKuoD7Init[]    =   {   1   ,   3   ,   1   ,   1   ,   15  ,   33  ,   49  ,   87  ,   185 ,   581 ,   1449    ,   735 ,   4891    ,0 };
            const unsigned long dim893JoeKuoD7Init[]    =   {   1   ,   1   ,   3   ,   7   ,   31  ,   49  ,   125 ,   99  ,   237 ,   655 ,   825 ,   2769    ,   4589    ,0 };
            const unsigned long dim894JoeKuoD7Init[]    =   {   1   ,   3   ,   3   ,   15  ,   1   ,   51  ,   97  ,   205 ,   15  ,   47  ,   1453    ,   3399    ,   5539    ,0 };
            const unsigned long dim895JoeKuoD7Init[]    =   {   1   ,   3   ,   5   ,   13  ,   29  ,   43  ,   105 ,   39  ,   329 ,   719 ,   563 ,   3817    ,   2901    ,0 };
            const unsigned long dim896JoeKuoD7Init[]    =   {   1   ,   1   ,   3   ,   5   ,   25  ,   19  ,   7   ,   209 ,   89  ,   455 ,   2019    ,   1481    ,   7409    ,0 };
            const unsigned long dim897JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   1   ,   25  ,   47  ,   53  ,   115 ,   263 ,   547 ,   565 ,   3495    ,   3759    ,0 };
            const unsigned long dim898JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   15  ,   25  ,   47  ,   121 ,   193 ,   99  ,   45  ,   1251    ,   1221    ,   6425    ,0 };
            const unsigned long dim899JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   9   ,   3   ,   47  ,   91  ,   231 ,   189 ,   39  ,   1811    ,   3215    ,   6593    ,0 };
            const unsigned long dim900JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   5   ,   3   ,   57  ,   67  ,   63  ,   23  ,   627 ,   2045    ,   2377    ,   3805    ,0 };
            const unsigned long dim901JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   13  ,   7   ,   47  ,   29  ,   99  ,   481 ,   923 ,   119 ,   1911    ,   5187    ,0 };
            const unsigned long dim902JoeKuoD7Init[]    =   {   1   ,   1   ,   5   ,   7   ,   1   ,   57  ,   7   ,   197 ,   451 ,   531 ,   29  ,   1851    ,   7069    ,0 };
            const unsigned long dim903JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   5   ,   31  ,   43  ,   3   ,   207 ,   67  ,   279 ,   37  ,   2381    ,   5997    ,0 };
            const unsigned long dim904JoeKuoD7Init[]    =   {   1   ,   3   ,   5   ,   7   ,   23  ,   57  ,   91  ,   109 ,   165 ,   549 ,   837 ,   3135    ,   2251    ,0 };
            const unsigned long dim905JoeKuoD7Init[]    =   {   1   ,   1   ,   7   ,   1   ,   3   ,   41  ,   67  ,   127 ,   273 ,   385 ,   995 ,   3263    ,   5211    ,0 };
            const unsigned long dim906JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   9   ,   21  ,   9   ,   101 ,   105 ,   43  ,   291 ,   1909    ,   1641    ,   7965    ,0 };
            const unsigned long dim907JoeKuoD7Init[]    =   {   1   ,   3   ,   5   ,   11  ,   19  ,   21  ,   117 ,   141 ,   197 ,   83  ,   203 ,   393 ,   7995    ,0 };
            const unsigned long dim908JoeKuoD7Init[]    =   {   1   ,   1   ,   5   ,   13  ,   11  ,   33  ,   21  ,   207 ,   421 ,   91  ,   1233    ,   1239    ,   5683    ,0 };
            const unsigned long dim909JoeKuoD7Init[]    =   {   1   ,   1   ,   3   ,   15  ,   5   ,   1   ,   127 ,   119 ,   349 ,   953 ,   1297    ,   1805    ,   2385    ,0 };
            const unsigned long dim910JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   1   ,   19  ,   17  ,   17  ,   37  ,   217 ,   857 ,   859 ,   53  ,   2091    ,0 };
            const unsigned long dim911JoeKuoD7Init[]    =   {   1   ,   3   ,   1   ,   15  ,   27  ,   37  ,   69  ,   121 ,   101 ,   1011    ,   1859    ,   3297    ,   5755    ,0 };
            const unsigned long dim912JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   13  ,   21  ,   13  ,   107 ,   105 ,   17  ,   743 ,   1849    ,   3169    ,   5901    ,0 };
            const unsigned long dim913JoeKuoD7Init[]    =   {   1   ,   3   ,   3   ,   13  ,   31  ,   37  ,   125 ,   83  ,   127 ,   687 ,   1593    ,   1257    ,   4729    ,0 };
            const unsigned long dim914JoeKuoD7Init[]    =   {   1   ,   3   ,   5   ,   3   ,   23  ,   21  ,   51  ,   157 ,   469 ,   349 ,   213 ,   1303    ,   123 ,0 };
            const unsigned long dim915JoeKuoD7Init[]    =   {   1   ,   3   ,   1   ,   15  ,   17  ,   23  ,   45  ,   157 ,   313 ,   241 ,   361 ,   3967    ,   3635    ,0 };
            const unsigned long dim916JoeKuoD7Init[]    =   {   1   ,   1   ,   7   ,   7   ,   25  ,   41  ,   65  ,   25  ,   97  ,   787 ,   951 ,   2471    ,   4621    ,0 };
            const unsigned long dim917JoeKuoD7Init[]    =   {   1   ,   1   ,   7   ,   3   ,   31  ,   29  ,   75  ,   173 ,   55  ,   613 ,   283 ,   195 ,   5009    ,0 };
            const unsigned long dim918JoeKuoD7Init[]    =   {   1   ,   1   ,   5   ,   3   ,   9   ,   47  ,   115 ,   149 ,   501 ,   199 ,   1957    ,   785 ,   3391    ,0 };
            const unsigned long dim919JoeKuoD7Init[]    =   {   1   ,   1   ,   7   ,   11  ,   31  ,   57  ,   71  ,   191 ,   333 ,   157 ,   809 ,   3039    ,   7471    ,0 };
            const unsigned long dim920JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   9   ,   19  ,   17  ,   103 ,   243 ,   411 ,   449 ,   1495    ,   3843    ,   5361    ,0 };
            const unsigned long dim921JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   1   ,   11  ,   13  ,   99  ,   229 ,   365 ,   909 ,   87  ,   3669    ,   6609    ,0 };
            const unsigned long dim922JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   15  ,   7   ,   61  ,   95  ,   95  ,   147 ,   959 ,   971 ,   649 ,   5047    ,0 };
            const unsigned long dim923JoeKuoD7Init[]    =   {   1   ,   3   ,   1   ,   15  ,   15  ,   37  ,   113 ,   29  ,   373 ,   379 ,   151 ,   2741    ,   1259    ,0 };
            const unsigned long dim924JoeKuoD7Init[]    =   {   1   ,   3   ,   3   ,   7   ,   1   ,   7   ,   91  ,   61  ,   213 ,   131 ,   1607    ,   3957    ,   4197    ,0 };
            const unsigned long dim925JoeKuoD7Init[]    =   {   1   ,   1   ,   7   ,   15  ,   11  ,   41  ,   79  ,   25  ,   267 ,   511 ,   111 ,   1779    ,   351 ,0 };
            const unsigned long dim926JoeKuoD7Init[]    =   {   1   ,   3   ,   5   ,   5   ,   17  ,   9   ,   29  ,   117 ,   207 ,   655 ,   221 ,   3567    ,   6069    ,0 };
            const unsigned long dim927JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   13  ,   31  ,   47  ,   3   ,   51  ,   451 ,   675 ,   91  ,   1529    ,   5729    ,0 };
            const unsigned long dim928JoeKuoD7Init[]    =   {   1   ,   1   ,   7   ,   15  ,   27  ,   37  ,   27  ,   189 ,   183 ,   619 ,   509 ,   2421    ,   5903    ,0 };
            const unsigned long dim929JoeKuoD7Init[]    =   {   1   ,   1   ,   3   ,   7   ,   1   ,   31  ,   79  ,   233 ,   41  ,   949 ,   1313    ,   3807    ,   171 ,0 };
            const unsigned long dim930JoeKuoD7Init[]    =   {   1   ,   1   ,   3   ,   5   ,   31  ,   39  ,   29  ,   43  ,   309 ,   11  ,   1041    ,   1893    ,   4443    ,0 };
            const unsigned long dim931JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   11  ,   13  ,   51  ,   5   ,   199 ,   51  ,   563 ,   819 ,   3467    ,   7003    ,0 };
            const unsigned long dim932JoeKuoD7Init[]    =   {   1   ,   1   ,   3   ,   15  ,   19  ,   27  ,   1   ,   3   ,   393 ,   901 ,   1603    ,   1949    ,   7119    ,0 };
            const unsigned long dim933JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   3   ,   13  ,   39  ,   35  ,   53  ,   437 ,   263 ,   783 ,   1669    ,   7271    ,0 };
            const unsigned long dim934JoeKuoD7Init[]    =   {   1   ,   1   ,   5   ,   11  ,   23  ,   11  ,   17  ,   187 ,   303 ,   833 ,   805 ,   3395    ,   181 ,0 };
            const unsigned long dim935JoeKuoD7Init[]    =   {   1   ,   3   ,   5   ,   9   ,   15  ,   23  ,   113 ,   37  ,   131 ,   89  ,   923 ,   2265    ,   1621    ,0 };
            const unsigned long dim936JoeKuoD7Init[]    =   {   1   ,   3   ,   1   ,   5   ,   3   ,   35  ,   53  ,   139 ,   21  ,   547 ,   1765    ,   3993    ,   2617    ,0 };
            const unsigned long dim937JoeKuoD7Init[]    =   {   1   ,   3   ,   5   ,   1   ,   3   ,   35  ,   17  ,   185 ,   291 ,   973 ,   1267    ,   301 ,   7443    ,0 };
            const unsigned long dim938JoeKuoD7Init[]    =   {   1   ,   3   ,   5   ,   1   ,   5   ,   45  ,   107 ,   209 ,   345 ,   821 ,   173 ,   1777    ,   4215    ,0 };
            const unsigned long dim939JoeKuoD7Init[]    =   {   1   ,   1   ,   5   ,   3   ,   13  ,   11  ,   17  ,   5   ,   445 ,   411 ,   825 ,   3031    ,   4295    ,0 };
            const unsigned long dim940JoeKuoD7Init[]    =   {   1   ,   3   ,   5   ,   1   ,   15  ,   45  ,   57  ,   243 ,   263 ,   167 ,   1889    ,   2901    ,   2315    ,0 };
            const unsigned long dim941JoeKuoD7Init[]    =   {   1   ,   1   ,   5   ,   13  ,   27  ,   37  ,   27  ,   39  ,   89  ,   571 ,   775 ,   269 ,   2633    ,0 };
            const unsigned long dim942JoeKuoD7Init[]    =   {   1   ,   3   ,   1   ,   15  ,   5   ,   41  ,   59  ,   89  ,   475 ,   745 ,   431 ,   489 ,   2121    ,0 };
            const unsigned long dim943JoeKuoD7Init[]    =   {   1   ,   3   ,   1   ,   11  ,   3   ,   53  ,   65  ,   107 ,   475 ,   529 ,   169 ,   2555    ,   5189    ,0 };
            const unsigned long dim944JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   7   ,   27  ,   7   ,   55  ,   11  ,   493 ,   687 ,   365 ,   3105    ,   8067    ,0 };
            const unsigned long dim945JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   7   ,   13  ,   7   ,   101 ,   139 ,   387 ,   289 ,   1157    ,   785 ,   1457    ,0 };
            const unsigned long dim946JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   7   ,   1   ,   23  ,   81  ,   147 ,   509 ,   241 ,   1507    ,   3947    ,   795 ,0 };
            const unsigned long dim947JoeKuoD7Init[]    =   {   1   ,   3   ,   1   ,   3   ,   3   ,   3   ,   39  ,   205 ,   67  ,   409 ,   1463    ,   2503    ,   7003    ,0 };
            const unsigned long dim948JoeKuoD7Init[]    =   {   1   ,   3   ,   5   ,   7   ,   13  ,   3   ,   49  ,   157 ,   271 ,   813 ,   1793    ,   2059    ,   7233    ,0 };
            const unsigned long dim949JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   11  ,   17  ,   7   ,   35  ,   251 ,   91  ,   725 ,   157 ,   2503    ,   7007    ,0 };
            const unsigned long dim950JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   7   ,   31  ,   59  ,   115 ,   1   ,   421 ,   181 ,   809 ,   1739    ,   5641    ,0 };
            const unsigned long dim951JoeKuoD7Init[]    =   {   1   ,   1   ,   7   ,   1   ,   3   ,   53  ,   29  ,   37  ,   181 ,   187 ,   11  ,   1829    ,   5031    ,0 };
            const unsigned long dim952JoeKuoD7Init[]    =   {   1   ,   3   ,   3   ,   3   ,   5   ,   49  ,   11  ,   69  ,   109 ,   11  ,   1825    ,   2119    ,   4697    ,0 };
            const unsigned long dim953JoeKuoD7Init[]    =   {   1   ,   3   ,   3   ,   7   ,   7   ,   31  ,   115 ,   197 ,   39  ,   177 ,   1263    ,   1173    ,   3845    ,0 };
            const unsigned long dim954JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   1   ,   15  ,   61  ,   49  ,   3   ,   141 ,   599 ,   809 ,   2827    ,   6349    ,0 };
            const unsigned long dim955JoeKuoD7Init[]    =   {   1   ,   3   ,   1   ,   9   ,   23  ,   29  ,   121 ,   175 ,   507 ,   273 ,   1945    ,   2991    ,   885 ,0 };
            const unsigned long dim956JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   1   ,   3   ,   55  ,   125 ,   37  ,   57  ,   901 ,   85  ,   3319    ,   3477    ,0 };
            const unsigned long dim957JoeKuoD7Init[]    =   {   1   ,   1   ,   7   ,   15  ,   3   ,   27  ,   69  ,   143 ,   431 ,   47  ,   603 ,   1321    ,   5053    ,0 };
            const unsigned long dim958JoeKuoD7Init[]    =   {   1   ,   3   ,   1   ,   3   ,   3   ,   37  ,   65  ,   163 ,   391 ,   509 ,   87  ,   747 ,   4301    ,0 };
            const unsigned long dim959JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   9   ,   5   ,   53  ,   75  ,   179 ,   389 ,   97  ,   373 ,   963 ,   3647    ,0 };
            const unsigned long dim960JoeKuoD7Init[]    =   {   1   ,   3   ,   1   ,   5   ,   13  ,   21  ,   49  ,   211 ,   5   ,   469 ,   1807    ,   891 ,   1543    ,0 };
            const unsigned long dim961JoeKuoD7Init[]    =   {   1   ,   3   ,   1   ,   9   ,   13  ,   29  ,   39  ,   55  ,   39  ,   667 ,   1967    ,   1887    ,   2619    ,0 };
            const unsigned long dim962JoeKuoD7Init[]    =   {   1   ,   1   ,   7   ,   9   ,   25  ,   47  ,   1   ,   221 ,   421 ,   175 ,   461 ,   1931    ,   3525    ,0 };
            const unsigned long dim963JoeKuoD7Init[]    =   {   1   ,   3   ,   1   ,   9   ,   11  ,   35  ,   23  ,   121 ,   493 ,   675 ,   857 ,   2989    ,   4863    ,0 };
            const unsigned long dim964JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   9   ,   5   ,   39  ,   121 ,   9   ,   249 ,   507 ,   1637    ,   2991    ,   4545    ,0 };
            const unsigned long dim965JoeKuoD7Init[]    =   {   1   ,   1   ,   5   ,   9   ,   5   ,   9   ,   27  ,   135 ,   299 ,   713 ,   1385    ,   729 ,   29  ,0 };
            const unsigned long dim966JoeKuoD7Init[]    =   {   1   ,   3   ,   1   ,   11  ,   3   ,   3   ,   67  ,   17  ,   163 ,   535 ,   1731    ,   1401    ,   4495    ,0 };
            const unsigned long dim967JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   5   ,   31  ,   29  ,   81  ,   251 ,   471 ,   461 ,   1151    ,   1037    ,   4069    ,0 };
            const unsigned long dim968JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   5   ,   11  ,   29  ,   13  ,   157 ,   329 ,   237 ,   691 ,   1895    ,   653 ,0 };
            const unsigned long dim969JoeKuoD7Init[]    =   {   1   ,   3   ,   5   ,   13  ,   23  ,   23  ,   57  ,   149 ,   111 ,   607 ,   1835    ,   3129    ,   1827    ,0 };
            const unsigned long dim970JoeKuoD7Init[]    =   {   1   ,   3   ,   1   ,   5   ,   11  ,   11  ,   95  ,   143 ,   495 ,   805 ,   1163    ,   1853    ,   1017    ,0 };
            const unsigned long dim971JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   9   ,   15  ,   19  ,   89  ,   143 ,   305 ,   121 ,   1217    ,   807 ,   5003    ,0 };
            const unsigned long dim972JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   13  ,   27  ,   39  ,   117 ,   65  ,   473 ,   191 ,   1863    ,   3533    ,   243 ,0 };
            const unsigned long dim973JoeKuoD7Init[]    =   {   1   ,   3   ,   3   ,   9   ,   31  ,   5   ,   91  ,   245 ,   175 ,   635 ,   257 ,   1197    ,   5965    ,0 };
            const unsigned long dim974JoeKuoD7Init[]    =   {   1   ,   3   ,   5   ,   1   ,   9   ,   13  ,   37  ,   19  ,   177 ,   949 ,   49  ,   3855    ,   5105    ,0 };
            const unsigned long dim975JoeKuoD7Init[]    =   {   1   ,   1   ,   3   ,   3   ,   1   ,   41  ,   33  ,   65  ,   295 ,   917 ,   503 ,   2665    ,   1485    ,0 };
            const unsigned long dim976JoeKuoD7Init[]    =   {   1   ,   1   ,   3   ,   5   ,   3   ,   25  ,   11  ,   9   ,   439 ,   293 ,   1919    ,   693 ,   5797    ,0 };
            const unsigned long dim977JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   15  ,   17  ,   51  ,   7   ,   17  ,   153 ,   105 ,   807 ,   1029    ,   4649    ,0 };
            const unsigned long dim978JoeKuoD7Init[]    =   {   1   ,   1   ,   5   ,   11  ,   27  ,   25  ,   37  ,   49  ,   481 ,   599 ,   655 ,   1497    ,   6455    ,0 };
            const unsigned long dim979JoeKuoD7Init[]    =   {   1   ,   1   ,   7   ,   5   ,   25  ,   23  ,   127 ,   203 ,   381 ,   61  ,   1651    ,   3535    ,   6345    ,0 };
            const unsigned long dim980JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   5   ,   27  ,   13  ,   93  ,   151 ,   421 ,   713 ,   305 ,   3541    ,   6741    ,0 };
            const unsigned long dim981JoeKuoD7Init[]    =   {   1   ,   1   ,   5   ,   5   ,   3   ,   1   ,   93  ,   55  ,   65  ,   437 ,   1283    ,   55  ,   861 ,0 };
            const unsigned long dim982JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   9   ,   27  ,   59  ,   15  ,   195 ,   303 ,   705 ,   143 ,   1423    ,   6793    ,0 };
            const unsigned long dim983JoeKuoD7Init[]    =   {   1   ,   1   ,   7   ,   13  ,   3   ,   21  ,   81  ,   89  ,   325 ,   103 ,   573 ,   1891    ,   3023    ,0 };
            const unsigned long dim984JoeKuoD7Init[]    =   {   1   ,   3   ,   3   ,   7   ,   15  ,   47  ,   23  ,   139 ,   437 ,   701 ,   1453    ,   1419    ,   1487    ,0 };
            const unsigned long dim985JoeKuoD7Init[]    =   {   1   ,   3   ,   3   ,   1   ,   13  ,   35  ,   49  ,   127 ,   335 ,   185 ,   393 ,   2239    ,   2039    ,0 };
            const unsigned long dim986JoeKuoD7Init[]    =   {   1   ,   1   ,   3   ,   5   ,   29  ,   3   ,   83  ,   179 ,   487 ,   381 ,   1477    ,   1849    ,   7593    ,0 };
            const unsigned long dim987JoeKuoD7Init[]    =   {   1   ,   1   ,   3   ,   3   ,   15  ,   61  ,   75  ,   115 ,   487 ,   237 ,   615 ,   1011    ,   4725    ,0 };
            const unsigned long dim988JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   5   ,   25  ,   39  ,   105 ,   81  ,   401 ,   939 ,   343 ,   3281    ,   3907    ,0 };
            const unsigned long dim989JoeKuoD7Init[]    =   {   1   ,   3   ,   5   ,   9   ,   31  ,   31  ,   43  ,   131 ,   275 ,   575 ,   727 ,   3035    ,   363 ,0 };
            const unsigned long dim990JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   11  ,   31  ,   9   ,   27  ,   145 ,   99  ,   261 ,   1927    ,   2619    ,   5853    ,0 };
            const unsigned long dim991JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   5   ,   25  ,   33  ,   85  ,   77  ,   435 ,   453 ,   1115    ,   1025    ,   2477    ,0 };
            const unsigned long dim992JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   13  ,   19  ,   55  ,   85  ,   155 ,   455 ,   385 ,   1033    ,   3789    ,   709 ,0 };
            const unsigned long dim993JoeKuoD7Init[]    =   {   1   ,   3   ,   1   ,   11  ,   31  ,   47  ,   61  ,   45  ,   247 ,   557 ,   929 ,   1579    ,   3927    ,0 };
            const unsigned long dim994JoeKuoD7Init[]    =   {   1   ,   1   ,   3   ,   7   ,   13  ,   41  ,   121 ,   249 ,   393 ,   527 ,   1807    ,   2493    ,   67  ,0 };
            const unsigned long dim995JoeKuoD7Init[]    =   {   1   ,   3   ,   7   ,   13  ,   3   ,   17  ,   9   ,   215 ,   181 ,   841 ,   1381    ,   1573    ,   1531    ,0 };
            const unsigned long dim996JoeKuoD7Init[]    =   {   1   ,   1   ,   7   ,   3   ,   21  ,   39  ,   119 ,   89  ,   267 ,   395 ,   1263    ,   3927    ,   5419    ,0 };
            const unsigned long dim997JoeKuoD7Init[]    =   {   1   ,   1   ,   1   ,   5   ,   17  ,   21  ,   35  ,   217 ,   165 ,   837 ,   1093    ,   1347    ,   6361    ,0 };
            const unsigned long dim998JoeKuoD7Init[]    =   {   1   ,   3   ,   3   ,   9   ,   9   ,   1   ,   59  ,   31  ,   41  ,   229 ,   925 ,   2703    ,   6597    ,0 };
            const unsigned long dim999JoeKuoD7Init[]    =   {   1   ,   3   ,   1   ,   9   ,   13  ,   21  ,   35  ,   159 ,   329 ,   981 ,   955 ,   1871    ,   3571    ,0 };
            const unsigned long dim1000JoeKuoD7Init[]   =   {   1   ,   3   ,   1   ,   11  ,   31  ,   47  ,   121 ,   201 ,   339 ,   1021    ,   1413    ,   1405    ,   1261    ,0 };
            const unsigned long dim1001JoeKuoD7Init[]   =   {   1   ,   1   ,   3   ,   11  ,   15  ,   23  ,   115 ,   247 ,   333 ,   757 ,   369 ,   3839    ,   5217    ,0 };
            const unsigned long dim1002JoeKuoD7Init[]   =   {   1   ,   1   ,   1   ,   15  ,   1   ,   37  ,   47  ,   67  ,   461 ,   333 ,   837 ,   2831    ,   2293    ,0 };
            const unsigned long dim1003JoeKuoD7Init[]   =   {   1   ,   1   ,   5   ,   11  ,   31  ,   29  ,   75  ,   9   ,   287 ,   17  ,   1771    ,   3523    ,   7321    ,0 };
            const unsigned long dim1004JoeKuoD7Init[]   =   {   1   ,   3   ,   7   ,   5   ,   29  ,   47  ,   37  ,   25  ,   497 ,   455 ,   1929    ,   2711    ,   777 ,0 };
            const unsigned long dim1005JoeKuoD7Init[]   =   {   1   ,   3   ,   1   ,   15  ,   17  ,   21  ,   121 ,   205 ,   133 ,   139 ,   1343    ,   2181    ,   6281    ,0 };
            const unsigned long dim1006JoeKuoD7Init[]   =   {   1   ,   3   ,   5   ,   3   ,   17  ,   59  ,   93  ,   105 ,   359 ,   27  ,   1515    ,   115 ,   7585    ,0 };
            const unsigned long dim1007JoeKuoD7Init[]   =   {   1   ,   3   ,   1   ,   3   ,   25  ,   43  ,   71  ,   71  ,   315 ,   137 ,   1245    ,   1751    ,   1589    ,0 };
            const unsigned long dim1008JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   1   ,   23  ,   51  ,   61  ,   247 ,   317 ,   125 ,   859 ,   3807    ,   1175    ,0 };
            const unsigned long dim1009JoeKuoD7Init[]   =   {   1   ,   1   ,   5   ,   9   ,   31  ,   33  ,   25  ,   41  ,   21  ,   1001    ,   63  ,   2105    ,   1929    ,0 };
            const unsigned long dim1010JoeKuoD7Init[]   =   {   1   ,   1   ,   3   ,   15  ,   29  ,   61  ,   59  ,   173 ,   265 ,   661 ,   115 ,   3127    ,   4431    ,0 };
            const unsigned long dim1011JoeKuoD7Init[]   =   {   1   ,   1   ,   1   ,   15  ,   7   ,   63  ,   55  ,   253 ,   131 ,   425 ,   735 ,   2917    ,   5311    ,0 };
            const unsigned long dim1012JoeKuoD7Init[]   =   {   1   ,   3   ,   1   ,   3   ,   9   ,   7   ,   81  ,   39  ,   261 ,   633 ,   485 ,   2091    ,   3763    ,0 };
            const unsigned long dim1013JoeKuoD7Init[]   =   {   1   ,   3   ,   1   ,   11  ,   27  ,   19  ,   99  ,   99  ,   27  ,   225 ,   1007    ,   1089    ,   921 ,0 };
            const unsigned long dim1014JoeKuoD7Init[]   =   {   1   ,   1   ,   3   ,   13  ,   21  ,   7   ,   85  ,   105 ,   379 ,   223 ,   955 ,   3533    ,   3879    ,0 };
            const unsigned long dim1015JoeKuoD7Init[]   =   {   1   ,   3   ,   3   ,   9   ,   13  ,   45  ,   55  ,   57  ,   315 ,   255 ,   141 ,   3255    ,   5185    ,0 };
            const unsigned long dim1016JoeKuoD7Init[]   =   {   1   ,   3   ,   7   ,   15  ,   9   ,   27  ,   53  ,   15  ,   25  ,   547 ,   1573    ,   1387    ,   7243    ,0 };
            const unsigned long dim1017JoeKuoD7Init[]   =   {   1   ,   1   ,   3   ,   15  ,   19  ,   17  ,   77  ,   253 ,   105 ,   1009    ,   1331    ,   2773    ,   7563    ,0 };
            const unsigned long dim1018JoeKuoD7Init[]   =   {   1   ,   1   ,   5   ,   5   ,   3   ,   19  ,   103 ,   63  ,   491 ,   709 ,   1299    ,   1245    ,   3879    ,0 };
            const unsigned long dim1019JoeKuoD7Init[]   =   {   1   ,   3   ,   1   ,   1   ,   13  ,   43  ,   37  ,   41  ,   167 ,   349 ,   1821    ,   2741    ,   3025    ,0 };
            const unsigned long dim1020JoeKuoD7Init[]   =   {   1   ,   1   ,   1   ,   5   ,   29  ,   39  ,   25  ,   221 ,   259 ,   939 ,   471 ,   2547    ,   7119    ,0 };
            const unsigned long dim1021JoeKuoD7Init[]   =   {   1   ,   3   ,   3   ,   3   ,   25  ,   59  ,   53  ,   105 ,   39  ,   737 ,   987 ,   2961    ,   7703    ,0 };
            const unsigned long dim1022JoeKuoD7Init[]   =   {   1   ,   3   ,   5   ,   3   ,   29  ,   11  ,   109 ,   233 ,   221 ,   515 ,   1463    ,   1233    ,   7591    ,0 };
            const unsigned long dim1023JoeKuoD7Init[]   =   {   1   ,   1   ,   1   ,   3   ,   17  ,   17  ,   113 ,   91  ,   253 ,   547 ,   1203    ,   2327    ,   6933    ,0 };
            const unsigned long dim1024JoeKuoD7Init[]   =   {   1   ,   1   ,   3   ,   11  ,   25  ,   3   ,   17  ,   27  ,   239 ,   723 ,   1751    ,   3907    ,   1199    ,0 };
            const unsigned long dim1025JoeKuoD7Init[]   =   {   1   ,   1   ,   5   ,   15  ,   13  ,   35  ,   93  ,   9   ,   501 ,   949 ,   659 ,   1863    ,   5319    ,0 };
            const unsigned long dim1026JoeKuoD7Init[]   =   {   1   ,   3   ,   7   ,   1   ,   1   ,   55  ,   99  ,   73  ,   185 ,   299 ,   181 ,   4005    ,   7239    ,0 };
            const unsigned long dim1027JoeKuoD7Init[]   =   {   1   ,   1   ,   1   ,   5   ,   13  ,   57  ,   21  ,   41  ,   461 ,   445 ,   199 ,   689 ,   2005    ,0 };
            const unsigned long dim1028JoeKuoD7Init[]   =   {   1   ,   1   ,   1   ,   9   ,   11  ,   39  ,   45  ,   157 ,   295 ,   907 ,   1839    ,   3313    ,   6397    ,0 };
            const unsigned long dim1029JoeKuoD7Init[]   =   {   1   ,   1   ,   5   ,   15  ,   21  ,   49  ,   39  ,   147 ,   325 ,   437 ,   1773    ,   2533    ,   8183    ,0 };
            const unsigned long dim1030JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   11  ,   11  ,   45  ,   9   ,   149 ,   439 ,   403 ,   1717    ,   3141    ,   7259    ,0 };
            const unsigned long dim1031JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   3   ,   13  ,   41  ,   79  ,   93  ,   123 ,   135 ,   541 ,   1133    ,   4389    ,0 };
            const unsigned long dim1032JoeKuoD7Init[]   =   {   1   ,   3   ,   1   ,   11  ,   17  ,   45  ,   3   ,   49  ,   117 ,   193 ,   1737    ,   2995    ,   6821    ,0 };
            const unsigned long dim1033JoeKuoD7Init[]   =   {   1   ,   3   ,   7   ,   9   ,   19  ,   53  ,   77  ,   119 ,   83  ,   161 ,   1029    ,   95  ,   7065    ,0 };
            const unsigned long dim1034JoeKuoD7Init[]   =   {   1   ,   3   ,   7   ,   5   ,   19  ,   33  ,   13  ,   85  ,   63  ,   1011    ,   1857    ,   2657    ,   7513    ,0 };
            const unsigned long dim1035JoeKuoD7Init[]   =   {   1   ,   1   ,   5   ,   13  ,   15  ,   31  ,   121 ,   75  ,   343 ,   915 ,   1571    ,   1291    ,   965 ,0 };
            const unsigned long dim1036JoeKuoD7Init[]   =   {   1   ,   1   ,   1   ,   15  ,   17  ,   27  ,   31  ,   183 ,   79  ,   219 ,   341 ,   2491    ,   7329    ,0 };
            const unsigned long dim1037JoeKuoD7Init[]   =   {   1   ,   3   ,   3   ,   15  ,   5   ,   5   ,   53  ,   151 ,   149 ,   471 ,   413 ,   967 ,   4207    ,0 };
            const unsigned long dim1038JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   15  ,   21  ,   17  ,   97  ,   107 ,   51  ,   533 ,   1195    ,   1495    ,   3843    ,0 };
            const unsigned long dim1039JoeKuoD7Init[]   =   {   1   ,   1   ,   3   ,   7   ,   3   ,   5   ,   109 ,   19  ,   401 ,   179 ,   1595    ,   3147    ,   3847    ,0 };
            const unsigned long dim1040JoeKuoD7Init[]   =   {   1   ,   1   ,   3   ,   9   ,   11  ,   49  ,   9   ,   245 ,   417 ,   933 ,   627 ,   551 ,   3643    ,0 };
            const unsigned long dim1041JoeKuoD7Init[]   =   {   1   ,   1   ,   1   ,   11  ,   5   ,   33  ,   101 ,   247 ,   57  ,   597 ,   377 ,   7   ,   7117    ,0 };
            const unsigned long dim1042JoeKuoD7Init[]   =   {   1   ,   3   ,   7   ,   11  ,   7   ,   47  ,   113 ,   29  ,   503 ,   385 ,   743 ,   563 ,   4775    ,0 };
            const unsigned long dim1043JoeKuoD7Init[]   =   {   1   ,   3   ,   3   ,   15  ,   7   ,   63  ,   57  ,   171 ,   77  ,   663 ,   1113    ,   1901    ,   5361    ,0 };
            const unsigned long dim1044JoeKuoD7Init[]   =   {   1   ,   3   ,   1   ,   11  ,   1   ,   1   ,   39  ,   83  ,   497 ,   725 ,   1105    ,   863 ,   113 ,0 };
            const unsigned long dim1045JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   1   ,   1   ,   23  ,   95  ,   137 ,   447 ,   593 ,   561 ,   1665    ,   7565    ,0 };
            const unsigned long dim1046JoeKuoD7Init[]   =   {   1   ,   1   ,   1   ,   5   ,   19  ,   15  ,   25  ,   185 ,   33  ,   747 ,   641 ,   3205    ,   1373    ,0 };
            const unsigned long dim1047JoeKuoD7Init[]   =   {   1   ,   1   ,   3   ,   1   ,   15  ,   17  ,   65  ,   77  ,   1   ,   689 ,   1751    ,   1885    ,   3873    ,0 };
            const unsigned long dim1048JoeKuoD7Init[]   =   {   1   ,   3   ,   7   ,   15  ,   23  ,   23  ,   9   ,   97  ,   129 ,   185 ,   1527    ,   3721    ,   5161    ,0 };
            const unsigned long dim1049JoeKuoD7Init[]   =   {   1   ,   1   ,   1   ,   5   ,   23  ,   33  ,   33  ,   131 ,   99  ,   929 ,   1077    ,   3129    ,   8153    ,0 };
            const unsigned long dim1050JoeKuoD7Init[]   =   {   1   ,   1   ,   3   ,   1   ,   1   ,   37  ,   69  ,   155 ,   69  ,   121 ,   395 ,   717 ,   841 ,0 };
            const unsigned long dim1051JoeKuoD7Init[]   =   {   1   ,   1   ,   5   ,   7   ,   17  ,   63  ,   101 ,   109 ,   323 ,   559 ,   427 ,   2279    ,   5345    ,0 };
            const unsigned long dim1052JoeKuoD7Init[]   =   {   1   ,   1   ,   1   ,   1   ,   9   ,   27  ,   77  ,   19  ,   375 ,   313 ,   855 ,   65  ,   1359    ,0 };
            const unsigned long dim1053JoeKuoD7Init[]   =   {   1   ,   3   ,   7   ,   5   ,   23  ,   11  ,   5   ,   57  ,   69  ,   267 ,   895 ,   1279    ,   5083    ,0 };
            const unsigned long dim1054JoeKuoD7Init[]   =   {   1   ,   3   ,   7   ,   9   ,   17  ,   5   ,   13  ,   27  ,   299 ,   279 ,   389 ,   3161    ,   4633    ,0 };
            const unsigned long dim1055JoeKuoD7Init[]   =   {   1   ,   3   ,   7   ,   1   ,   19  ,   47  ,   31  ,   93  ,   503 ,   473 ,   1501    ,   617 ,   1621    ,0 };
            const unsigned long dim1056JoeKuoD7Init[]   =   {   1   ,   3   ,   1   ,   15  ,   17  ,   31  ,   35  ,   75  ,   41  ,   113 ,   149 ,   861 ,   5845    ,0 };
            const unsigned long dim1057JoeKuoD7Init[]   =   {   1   ,   3   ,   7   ,   1   ,   7   ,   21  ,   31  ,   77  ,   307 ,   709 ,   531 ,   1133    ,   5887    ,0 };
            const unsigned long dim1058JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   7   ,   25  ,   37  ,   23  ,   233 ,   339 ,   705 ,   123 ,   79  ,   3827    ,0 };
            const unsigned long dim1059JoeKuoD7Init[]   =   {   1   ,   1   ,   5   ,   9   ,   15  ,   61  ,   13  ,   215 ,   175 ,   999 ,   1147    ,   523 ,   7831    ,0 };
            const unsigned long dim1060JoeKuoD7Init[]   =   {   1   ,   1   ,   1   ,   7   ,   15  ,   31  ,   117 ,   41  ,   43  ,   851 ,   1657    ,   669 ,   7243    ,0 };
            const unsigned long dim1061JoeKuoD7Init[]   =   {   1   ,   1   ,   1   ,   1   ,   29  ,   47  ,   29  ,   179 ,   151 ,   833 ,   1911    ,   3031    ,   3835    ,0 };
            const unsigned long dim1062JoeKuoD7Init[]   =   {   1   ,   1   ,   1   ,   13  ,   27  ,   41  ,   127 ,   203 ,   381 ,   395 ,   201 ,   1713    ,   4781    ,0 };
            const unsigned long dim1063JoeKuoD7Init[]   =   {   1   ,   1   ,   3   ,   7   ,   11  ,   11  ,   127 ,   13  ,   257 ,   1   ,   1285    ,   2829    ,   8077    ,0 };
            const unsigned long dim1064JoeKuoD7Init[]   =   {   1   ,   3   ,   3   ,   5   ,   15  ,   1   ,   35  ,   105 ,   183 ,   543 ,   1865    ,   675 ,   7747    ,0 };
            const unsigned long dim1065JoeKuoD7Init[]   =   {   1   ,   1   ,   3   ,   1   ,   15  ,   47  ,   111 ,   147 ,   373 ,   445 ,   1273    ,   1989    ,   2057    ,0 };
            const unsigned long dim1066JoeKuoD7Init[]   =   {   1   ,   1   ,   5   ,   9   ,   5   ,   31  ,   63  ,   127 ,   33  ,   323 ,   477 ,   213 ,   8137    ,0 };
            const unsigned long dim1067JoeKuoD7Init[]   =   {   1   ,   1   ,   5   ,   9   ,   13  ,   1   ,   9   ,   135 ,   113 ,   395 ,   127 ,   3561    ,   5903    ,0 };
            const unsigned long dim1068JoeKuoD7Init[]   =   {   1   ,   1   ,   3   ,   15  ,   1   ,   31  ,   85  ,   181 ,   49  ,   585 ,   1515    ,   2105    ,   541 ,0 };
            const unsigned long dim1069JoeKuoD7Init[]   =   {   1   ,   1   ,   1   ,   15  ,   9   ,   1   ,   121 ,   31  ,   221 ,   445 ,   735 ,   1107    ,   4073    ,0 };
            const unsigned long dim1070JoeKuoD7Init[]   =   {   1   ,   3   ,   1   ,   15  ,   3   ,   13  ,   13  ,   247 ,   509 ,   199 ,   1575    ,   2721    ,   2303    ,0 };
            const unsigned long dim1071JoeKuoD7Init[]   =   {   1   ,   1   ,   3   ,   7   ,   25  ,   17  ,   63  ,   59  ,   409 ,   69  ,   1947    ,   427 ,   5357    ,0 };
            const unsigned long dim1072JoeKuoD7Init[]   =   {   1   ,   1   ,   3   ,   3   ,   13  ,   5   ,   99  ,   161 ,   163 ,   437 ,   1467    ,   2215    ,   3321    ,0 };
            const unsigned long dim1073JoeKuoD7Init[]   =   {   1   ,   3   ,   1   ,   3   ,   15  ,   55  ,   101 ,   207 ,   75  ,   601 ,   1327    ,   3901    ,   5773    ,0 };
            const unsigned long dim1074JoeKuoD7Init[]   =   {   1   ,   1   ,   3   ,   3   ,   13  ,   45  ,   115 ,   119 ,   431 ,   5   ,   563 ,   3255    ,   5077    ,0 };
            const unsigned long dim1075JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   3   ,   21  ,   51  ,   41  ,   3   ,   375 ,   203 ,   1117    ,   3921    ,   387 ,0 };
            const unsigned long dim1076JoeKuoD7Init[]   =   {   1   ,   3   ,   7   ,   7   ,   9   ,   5   ,   15  ,   133 ,   179 ,   37  ,   1593    ,   193 ,   4975    ,0 };
            const unsigned long dim1077JoeKuoD7Init[]   =   {   1   ,   3   ,   3   ,   7   ,   31  ,   37  ,   89  ,   1   ,   241 ,   445 ,   381 ,   3283    ,   3851    ,0 };
            const unsigned long dim1078JoeKuoD7Init[]   =   {   1   ,   3   ,   5   ,   1   ,   5   ,   45  ,   21  ,   123 ,   377 ,   83  ,   2029    ,   1913    ,   659 ,0 };
            const unsigned long dim1079JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   3   ,   5   ,   5   ,   35  ,   85  ,   173 ,   397 ,   1695    ,   3517    ,   175 ,0 };
            const unsigned long dim1080JoeKuoD7Init[]   =   {   1   ,   3   ,   7   ,   13  ,   25  ,   51  ,   93  ,   177 ,   371 ,   507 ,   1777    ,   3335    ,   5953    ,0 };
            const unsigned long dim1081JoeKuoD7Init[]   =   {   1   ,   3   ,   5   ,   15  ,   13  ,   41  ,   103 ,   105 ,   303 ,   401 ,   1339    ,   21  ,   5723    ,0 };
            const unsigned long dim1082JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   15  ,   7   ,   57  ,   63  ,   79  ,   35  ,   363 ,   831 ,   3911    ,   5777    ,0 };
            const unsigned long dim1083JoeKuoD7Init[]   =   {   1   ,   1   ,   5   ,   11  ,   13  ,   15  ,   127 ,   141 ,   231 ,   619 ,   835 ,   691 ,   525 ,0 };
            const unsigned long dim1084JoeKuoD7Init[]   =   {   1   ,   1   ,   3   ,   5   ,   7   ,   33  ,   81  ,   83  ,   497 ,   459 ,   493 ,   725 ,   2137    ,0 };
            const unsigned long dim1085JoeKuoD7Init[]   =   {   1   ,   3   ,   3   ,   5   ,   25  ,   7   ,   77  ,   163 ,   47  ,   445 ,   629 ,   463 ,   3597    ,0 };
            const unsigned long dim1086JoeKuoD7Init[]   =   {   1   ,   1   ,   3   ,   11  ,   19  ,   13  ,   67  ,   117 ,   351 ,   169 ,   1905    ,   3105    ,   93  ,0 };
            const unsigned long dim1087JoeKuoD7Init[]   =   {   1   ,   3   ,   1   ,   13  ,   17  ,   17  ,   67  ,   135 ,   111 ,   649 ,   495 ,   425 ,   1621    ,0 };
            const unsigned long dim1088JoeKuoD7Init[]   =   {   1   ,   3   ,   5   ,   1   ,   19  ,   59  ,   11  ,   227 ,   137 ,   525 ,   1751    ,   915 ,   4397    ,0 };
            const unsigned long dim1089JoeKuoD7Init[]   =   {   1   ,   1   ,   3   ,   3   ,   15  ,   53  ,   87  ,   153 ,   69  ,   867 ,   833 ,   2363    ,   4363    ,0 };
            const unsigned long dim1090JoeKuoD7Init[]   =   {   1   ,   3   ,   5   ,   13  ,   13  ,   19  ,   35  ,   103 ,   441 ,   23  ,   1049    ,   2469    ,   6595    ,0 };
            const unsigned long dim1091JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   7   ,   23  ,   55  ,   57  ,   25  ,   3   ,   157 ,   523 ,   1029    ,   4177    ,0 };
            const unsigned long dim1092JoeKuoD7Init[]   =   {   1   ,   1   ,   3   ,   5   ,   9   ,   13  ,   1   ,   109 ,   219 ,   703 ,   237 ,   435 ,   2607    ,0 };
            const unsigned long dim1093JoeKuoD7Init[]   =   {   1   ,   1   ,   1   ,   13  ,   21  ,   23  ,   5   ,   109 ,   467 ,   197 ,   1467    ,   3625    ,   4997    ,0 };
            const unsigned long dim1094JoeKuoD7Init[]   =   {   1   ,   3   ,   7   ,   3   ,   25  ,   53  ,   119 ,   163 ,   117 ,   721 ,   1017    ,   2499    ,   4139    ,0 };
            const unsigned long dim1095JoeKuoD7Init[]   =   {   1   ,   1   ,   1   ,   9   ,   9   ,   13  ,   33  ,   5   ,   263 ,   21  ,   1393    ,   3181    ,   4785    ,0 };
            const unsigned long dim1096JoeKuoD7Init[]   =   {   1   ,   1   ,   5   ,   13  ,   15  ,   5   ,   57  ,   233 ,   99  ,   305 ,   745 ,   2327    ,   2259    ,0 };
            const unsigned long dim1097JoeKuoD7Init[]   =   {   1   ,   1   ,   3   ,   3   ,   15  ,   39  ,   99  ,   115 ,   395 ,   99  ,   947 ,   3171    ,   711 ,0 };
            const unsigned long dim1098JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   13  ,   17  ,   33  ,   29  ,   37  ,   399 ,   479 ,   1083    ,   285 ,   6897    ,0 };
            const unsigned long dim1099JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   5   ,   7   ,   1   ,   9   ,   239 ,   479 ,   265 ,   1129    ,   3615    ,   4865    ,0 };
            const unsigned long dim1100JoeKuoD7Init[]   =   {   1   ,   3   ,   3   ,   15  ,   31  ,   45  ,   7   ,   179 ,   467 ,   675 ,   1257    ,   1729    ,   7879    ,0 };
            const unsigned long dim1101JoeKuoD7Init[]   =   {   1   ,   3   ,   5   ,   13  ,   5   ,   23  ,   11  ,   219 ,   503 ,   613 ,   1383    ,   1733    ,   3269    ,0 };
            const unsigned long dim1102JoeKuoD7Init[]   =   {   1   ,   1   ,   3   ,   13  ,   25  ,   5   ,   81  ,   205 ,   389 ,   925 ,   735 ,   2981    ,   4271    ,0 };
            const unsigned long dim1103JoeKuoD7Init[]   =   {   1   ,   1   ,   3   ,   7   ,   19  ,   47  ,   107 ,   15  ,   87  ,   889 ,   725 ,   1123    ,   2883    ,0 };
            const unsigned long dim1104JoeKuoD7Init[]   =   {   1   ,   1   ,   3   ,   13  ,   31  ,   9   ,   45  ,   129 ,   353 ,   919 ,   1529    ,   827 ,   3645    ,0 };
            const unsigned long dim1105JoeKuoD7Init[]   =   {   1   ,   3   ,   5   ,   5   ,   13  ,   45  ,   17  ,   219 ,   23  ,   873 ,   1127    ,   1757    ,   1103    ,0 };
            const unsigned long dim1106JoeKuoD7Init[]   =   {   1   ,   3   ,   3   ,   11  ,   31  ,   45  ,   33  ,   215 ,   107 ,   941 ,   1785    ,   4039    ,   3393    ,0 };
            const unsigned long dim1107JoeKuoD7Init[]   =   {   1   ,   3   ,   3   ,   7   ,   15  ,   61  ,   107 ,   189 ,   31  ,   457 ,   881 ,   1449    ,   7351    ,0 };
            const unsigned long dim1108JoeKuoD7Init[]   =   {   1   ,   3   ,   1   ,   15  ,   13  ,   63  ,   95  ,   27  ,   3   ,   771 ,   363 ,   2569    ,   2195    ,0 };
            const unsigned long dim1109JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   9   ,   15  ,   11  ,   5   ,   193 ,   439 ,   745 ,   1705    ,   1249    ,   7319    ,0 };
            const unsigned long dim1110JoeKuoD7Init[]   =   {   1   ,   1   ,   3   ,   11  ,   29  ,   19  ,   111 ,   73  ,   149 ,   707 ,   1337    ,   1417    ,   761 ,0 };
            const unsigned long dim1111JoeKuoD7Init[]   =   {   1   ,   3   ,   7   ,   9   ,   17  ,   7   ,   127 ,   113 ,   353 ,   575 ,   1979    ,   3589    ,   3221    ,   1801    ,0 };
            const unsigned long dim1112JoeKuoD7Init[]   =   {   1   ,   1   ,   5   ,   13  ,   29  ,   21  ,   31  ,   135 ,   45  ,   571 ,   229 ,   791 ,   2017    ,   1023    ,0 };
            const unsigned long dim1113JoeKuoD7Init[]   =   {   1   ,   1   ,   1   ,   1   ,   3   ,   37  ,   29  ,   233 ,   379 ,   579 ,   1895    ,   2359    ,   5511    ,   15901   ,0 };
            const unsigned long dim1114JoeKuoD7Init[]   =   {   1   ,   3   ,   1   ,   3   ,   11  ,   63  ,   83  ,   219 ,   275 ,   441 ,   581 ,   3735    ,   6845    ,   12669   ,0 };
            const unsigned long dim1115JoeKuoD7Init[]   =   {   1   ,   3   ,   3   ,   9   ,   3   ,   5   ,   87  ,   143 ,   207 ,   211 ,   1123    ,   443 ,   3139    ,   15123   ,0 };
            const unsigned long dim1116JoeKuoD7Init[]   =   {   1   ,   3   ,   3   ,   11  ,   11  ,   21  ,   69  ,   69  ,   307 ,   935 ,   1353    ,   163 ,   1535    ,   4821    ,0 };
            const unsigned long dim1117JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   13  ,   11  ,   47  ,   41  ,   107 ,   463 ,   1003    ,   457 ,   1019    ,   1699    ,   4425    ,0 };
            const unsigned long dim1118JoeKuoD7Init[]   =   {   1   ,   3   ,   7   ,   7   ,   9   ,   5   ,   99  ,   95  ,   207 ,   697 ,   9   ,   771 ,   6895    ,   9157    ,0 };
            const unsigned long dim1119JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   11  ,   13  ,   5   ,   15  ,   107 ,   3   ,   319 ,   1187    ,   3357    ,   295 ,   2863    ,0 };
            const unsigned long dim1120JoeKuoD7Init[]   =   {   1   ,   3   ,   3   ,   1   ,   13  ,   23  ,   89  ,   149 ,   337 ,   435 ,   633 ,   885 ,   1361    ,   3795    ,0 };
            const unsigned long dim1121JoeKuoD7Init[]   =   {   1   ,   3   ,   1   ,   13  ,   1   ,   9   ,   63  ,   203 ,   377 ,   305 ,   1471    ,   3193    ,   5507    ,   2219    ,0 };
            const unsigned long dim1122JoeKuoD7Init[]   =   {   1   ,   1   ,   3   ,   9   ,   9   ,   9   ,   63  ,   31  ,   35  ,   551 ,   1379    ,   135 ,   4949    ,   10345   ,0 };
            const unsigned long dim1123JoeKuoD7Init[]   =   {   1   ,   3   ,   1   ,   3   ,   7   ,   61  ,   121 ,   27  ,   127 ,   901 ,   1935    ,   1347    ,   3179    ,   13785   ,0 };
            const unsigned long dim1124JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   11  ,   13  ,   17  ,   57  ,   73  ,   219 ,   73  ,   335 ,   3705    ,   591 ,   3547    ,0 };
            const unsigned long dim1125JoeKuoD7Init[]   =   {   1   ,   1   ,   1   ,   7   ,   13  ,   7   ,   127 ,   153 ,   503 ,   99  ,   187 ,   1697    ,   3497    ,   11913   ,0 };
            const unsigned long dim1126JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   5   ,   3   ,   21  ,   99  ,   125 ,   87  ,   889 ,   1471    ,   625 ,   2671    ,   11729   ,0 };
            const unsigned long dim1127JoeKuoD7Init[]   =   {   1   ,   3   ,   5   ,   7   ,   9   ,   55  ,   127 ,   201 ,   383 ,   769 ,   769 ,   3907    ,   7087    ,   1933    ,0 };
            const unsigned long dim1128JoeKuoD7Init[]   =   {   1   ,   3   ,   7   ,   5   ,   9   ,   47  ,   5   ,   29  ,   155 ,   327 ,   1413    ,   2761    ,   7699    ,   9129    ,0 };
            const unsigned long dim1129JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   15  ,   9   ,   11  ,   7   ,   91  ,   71  ,   563 ,   535 ,   3387    ,   4769    ,   4875    ,0 };
            const unsigned long dim1130JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   11  ,   17  ,   47  ,   29  ,   201 ,   299 ,   947 ,   39  ,   583 ,   1723    ,   877 ,0 };
            const unsigned long dim1131JoeKuoD7Init[]   =   {   1   ,   3   ,   1   ,   5   ,   27  ,   49  ,   105 ,   31  ,   91  ,   39  ,   1355    ,   2729    ,   6965    ,   14081   ,0 };
            const unsigned long dim1132JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   7   ,   17  ,   63  ,   65  ,   229 ,   497 ,   431 ,   1167    ,   1299    ,   4641    ,   10975   ,0 };
            const unsigned long dim1133JoeKuoD7Init[]   =   {   1   ,   1   ,   5   ,   9   ,   17  ,   3   ,   101 ,   191 ,   187 ,   87  ,   1229    ,   369 ,   2633    ,   13765   ,0 };
            const unsigned long dim1134JoeKuoD7Init[]   =   {   1   ,   1   ,   1   ,   7   ,   29  ,   7   ,   57  ,   17  ,   319 ,   839 ,   799 ,   151 ,   431 ,   9513    ,0 };
            const unsigned long dim1135JoeKuoD7Init[]   =   {   1   ,   1   ,   1   ,   11  ,   25  ,   47  ,   15  ,   29  ,   71  ,   617 ,   697 ,   2633    ,   1787    ,   10343   ,0 };
            const unsigned long dim1136JoeKuoD7Init[]   =   {   1   ,   1   ,   3   ,   3   ,   13  ,   1   ,   41  ,   211 ,   265 ,   501 ,   1619    ,   3781    ,   4091    ,   7785    ,0 };
            const unsigned long dim1137JoeKuoD7Init[]   =   {   1   ,   3   ,   3   ,   9   ,   11  ,   37  ,   69  ,   45  ,   107 ,   69  ,   1227    ,   311 ,   187 ,   10379   ,0 };
            const unsigned long dim1138JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   1   ,   31  ,   37  ,   85  ,   135 ,   283 ,   869 ,   1159    ,   3971    ,   6339    ,   5511    ,0 };
            const unsigned long dim1139JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   3   ,   19  ,   43  ,   75  ,   59  ,   365 ,   763 ,   323 ,   3513    ,   437 ,   2371    ,0 };
            const unsigned long dim1140JoeKuoD7Init[]   =   {   1   ,   3   ,   3   ,   13  ,   31  ,   39  ,   81  ,   199 ,   43  ,   863 ,   1563    ,   2811    ,   4551    ,   7469    ,0 };
            const unsigned long dim1141JoeKuoD7Init[]   =   {   1   ,   1   ,   3   ,   5   ,   17  ,   59  ,   17  ,   9   ,   53  ,   527 ,   1689    ,   2365    ,   7673    ,   14581   ,0 };
            const unsigned long dim1142JoeKuoD7Init[]   =   {   1   ,   3   ,   7   ,   5   ,   5   ,   5   ,   59  ,   21  ,   303 ,   43  ,   1379    ,   3297    ,   3053    ,   11115   ,0 };
            const unsigned long dim1143JoeKuoD7Init[]   =   {   1   ,   1   ,   5   ,   15  ,   3   ,   33  ,   5   ,   253 ,   409 ,   321 ,   1487    ,   2083    ,   405 ,   15075   ,0 };
            const unsigned long dim1144JoeKuoD7Init[]   =   {   1   ,   1   ,   3   ,   15  ,   11  ,   53  ,   99  ,   151 ,   33  ,   675 ,   1781    ,   1475    ,   2193    ,   6073    ,0 };
            const unsigned long dim1145JoeKuoD7Init[]   =   {   1   ,   1   ,   3   ,   13  ,   27  ,   49  ,   83  ,   15  ,   335 ,   35  ,   1999    ,   237 ,   7359    ,   12173   ,0 };
            const unsigned long dim1146JoeKuoD7Init[]   =   {   1   ,   3   ,   1   ,   15  ,   15  ,   61  ,   85  ,   243 ,   347 ,   745 ,   637 ,   1057    ,   5475    ,   15039   ,0 };
            const unsigned long dim1147JoeKuoD7Init[]   =   {   1   ,   1   ,   3   ,   7   ,   1   ,   1   ,   31  ,   197 ,   499 ,   741 ,   191 ,   3799    ,   5485    ,   10129   ,0 };
            const unsigned long dim1148JoeKuoD7Init[]   =   {   1   ,   1   ,   5   ,   5   ,   29  ,   15  ,   15  ,   185 ,   55  ,   685 ,   329 ,   605 ,   3365    ,   10203   ,0 };
            const unsigned long dim1149JoeKuoD7Init[]   =   {   1   ,   3   ,   1   ,   13  ,   5   ,   1   ,   1   ,   169 ,   349 ,   591 ,   1961    ,   2989    ,   131 ,   2647    ,0 };
            const unsigned long dim1150JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   9   ,   17  ,   49  ,   95  ,   7   ,   57  ,   881 ,   1065    ,   3759    ,   5851    ,   7439    ,0 };
            const unsigned long dim1151JoeKuoD7Init[]   =   {   1   ,   1   ,   3   ,   5   ,   27  ,   49  ,   65  ,   1   ,   177 ,   477 ,   453 ,   1823    ,   7615    ,   1285    ,0 };
            const unsigned long dim1152JoeKuoD7Init[]   =   {   1   ,   1   ,   5   ,   9   ,   23  ,   25  ,   17  ,   149 ,   85  ,   373 ,   1585    ,   1227    ,   6127    ,   6003    ,0 };
            const unsigned long dim1153JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   13  ,   15  ,   61  ,   41  ,   141 ,   507 ,   391 ,   587 ,   2819    ,   3311    ,   853 ,0 };
            const unsigned long dim1154JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   13  ,   13  ,   23  ,   67  ,   171 ,   381 ,   285 ,   619 ,   457 ,   2751    ,   435 ,0 };
            const unsigned long dim1155JoeKuoD7Init[]   =   {   1   ,   1   ,   5   ,   15  ,   13  ,   9   ,   21  ,   197 ,   385 ,   511 ,   1683    ,   2965    ,   1367    ,   7523    ,0 };
            const unsigned long dim1156JoeKuoD7Init[]   =   {   1   ,   3   ,   7   ,   1   ,   23  ,   13  ,   21  ,   55  ,   415 ,   503 ,   1765    ,   1745    ,   5329    ,   4665    ,0 };
            const unsigned long dim1157JoeKuoD7Init[]   =   {   1   ,   1   ,   1   ,   1   ,   21  ,   51  ,   49  ,   31  ,   195 ,   243 ,   1825    ,   2831    ,   467 ,   6527    ,0 };
            const unsigned long dim1158JoeKuoD7Init[]   =   {   1   ,   1   ,   3   ,   3   ,   31  ,   37  ,   41  ,   139 ,   63  ,   517 ,   717 ,   415 ,   1137    ,   1855    ,0 };
            const unsigned long dim1159JoeKuoD7Init[]   =   {   1   ,   1   ,   3   ,   15  ,   1   ,   45  ,   39  ,   9   ,   441 ,   809 ,   801 ,   3133    ,   2283    ,   8947    ,0 };
            const unsigned long dim1160JoeKuoD7Init[]   =   {   1   ,   1   ,   3   ,   13  ,   13  ,   47  ,   55  ,   115 ,   155 ,   347 ,   1597    ,   415 ,   271 ,   4543    ,0 };
            const unsigned long dim1161JoeKuoD7Init[]   =   {   1   ,   1   ,   1   ,   1   ,   1   ,   45  ,   5   ,   239 ,   187 ,   293 ,   59  ,   1319    ,   3839    ,   9945    ,0 };
            const unsigned long dim1162JoeKuoD7Init[]   =   {   1   ,   1   ,   1   ,   15  ,   1   ,   41  ,   13  ,   19  ,   455 ,   881 ,   451 ,   1795    ,   239 ,   2695    ,0 };
            const unsigned long dim1163JoeKuoD7Init[]   =   {   1   ,   3   ,   3   ,   9   ,   5   ,   25  ,   3   ,   19  ,   189 ,   377 ,   1017    ,   2501    ,   2707    ,   15333   ,0 };
            const unsigned long dim1164JoeKuoD7Init[]   =   {   1   ,   1   ,   5   ,   5   ,   29  ,   57  ,   45  ,   35  ,   297 ,   663 ,   359 ,   2597    ,   7223    ,   6265    ,0 };
            const unsigned long dim1165JoeKuoD7Init[]   =   {   1   ,   3   ,   1   ,   7   ,   1   ,   47  ,   125 ,   209 ,   65  ,   129 ,   211 ,   2535    ,   3617    ,   16229   ,0 };
            const unsigned long dim1166JoeKuoD7Init[]   =   {   1   ,   1   ,   3   ,   11  ,   19  ,   29  ,   39  ,   177 ,   39  ,   807 ,   975 ,   835 ,   5509    ,   791 ,0 };
            const unsigned long dim1167JoeKuoD7Init[]   =   {   1   ,   3   ,   3   ,   13  ,   31  ,   49  ,   69  ,   169 ,   433 ,   833 ,   1743    ,   3187    ,   8111    ,   15475   ,0 };
            const unsigned long dim1168JoeKuoD7Init[]   =   {   1   ,   1   ,   5   ,   7   ,   31  ,   9   ,   27  ,   73  ,   335 ,   363 ,   1707    ,   1227    ,   6459    ,   12513   ,0 };
            const unsigned long dim1169JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   1   ,   1   ,   47  ,   109 ,   115 ,   429 ,   551 ,   1805    ,   3235    ,   579 ,   10039   ,0 };
            const unsigned long dim1170JoeKuoD7Init[]   =   {   1   ,   1   ,   1   ,   7   ,   17  ,   61  ,   61  ,   1   ,   35  ,   917 ,   1513    ,   2141    ,   3939    ,   15755   ,0 };
            const unsigned long dim1171JoeKuoD7Init[]   =   {   1   ,   1   ,   5   ,   1   ,   3   ,   27  ,   7   ,   219 ,   473 ,   875 ,   711 ,   703 ,   129 ,   16359   ,0 };
            const unsigned long dim1172JoeKuoD7Init[]   =   {   1   ,   3   ,   5   ,   3   ,   27  ,   33  ,   97  ,   137 ,   19  ,   893 ,   233 ,   2605    ,   13  ,   875 ,0 };
            const unsigned long dim1173JoeKuoD7Init[]   =   {   1   ,   3   ,   7   ,   11  ,   31  ,   43  ,   77  ,   225 ,   403 ,   861 ,   141 ,   2121    ,   707 ,   15753   ,0 };
            const unsigned long dim1174JoeKuoD7Init[]   =   {   1   ,   1   ,   1   ,   3   ,   17  ,   33  ,   29  ,   103 ,   381 ,   179 ,   735 ,   2939    ,   7213    ,   12679   ,0 };
            const unsigned long dim1175JoeKuoD7Init[]   =   {   1   ,   1   ,   1   ,   7   ,   5   ,   41  ,   43  ,   215 ,   239 ,   19  ,   625 ,   579 ,   5157    ,   12457   ,0 };
            const unsigned long dim1176JoeKuoD7Init[]   =   {   1   ,   3   ,   1   ,   7   ,   15  ,   29  ,   93  ,   37  ,   241 ,   141 ,   1035    ,   3283    ,   193 ,   9873    ,0 };
            const unsigned long dim1177JoeKuoD7Init[]   =   {   1   ,   1   ,   3   ,   13  ,   23  ,   11  ,   113 ,   153 ,   271 ,   515 ,   197 ,   3723    ,   1827    ,   11941   ,0 };
            const unsigned long dim1178JoeKuoD7Init[]   =   {   1   ,   3   ,   7   ,   1   ,   29  ,   13  ,   45  ,   23  ,   421 ,   957 ,   1819    ,   1287    ,   5615    ,   4253    ,0 };
            const unsigned long dim1179JoeKuoD7Init[]   =   {   1   ,   1   ,   3   ,   3   ,   11  ,   55  ,   47  ,   211 ,   85  ,   481 ,   1199    ,   3225    ,   5233    ,   4295    ,0 };
            const unsigned long dim1180JoeKuoD7Init[]   =   {   1   ,   3   ,   3   ,   13  ,   11  ,   51  ,   93  ,   139 ,   271 ,   929 ,   863 ,   3227    ,   3075    ,   3587    ,0 };
            const unsigned long dim1181JoeKuoD7Init[]   =   {   1   ,   1   ,   5   ,   11  ,   19  ,   59  ,   9   ,   221 ,   275 ,   171 ,   1703    ,   1495    ,   523 ,   5993    ,0 };
            const unsigned long dim1182JoeKuoD7Init[]   =   {   1   ,   3   ,   1   ,   13  ,   15  ,   11  ,   21  ,   147 ,   291 ,   639 ,   1411    ,   33  ,   6845    ,   1431    ,0 };
            const unsigned long dim1183JoeKuoD7Init[]   =   {   1   ,   3   ,   5   ,   1   ,   11  ,   11  ,   39  ,   195 ,   261 ,   19  ,   1803    ,   2869    ,   4507    ,   1131    ,0 };
            const unsigned long dim1184JoeKuoD7Init[]   =   {   1   ,   3   ,   3   ,   11  ,   15  ,   27  ,   125 ,   5   ,   309 ,   569 ,   361 ,   4061    ,   2559    ,   1631    ,0 };
            const unsigned long dim1185JoeKuoD7Init[]   =   {   1   ,   1   ,   5   ,   7   ,   23  ,   57  ,   37  ,   175 ,   35  ,   753 ,   755 ,   519 ,   5617    ,   1419    ,0 };
            const unsigned long dim1186JoeKuoD7Init[]   =   {   1   ,   1   ,   5   ,   11  ,   21  ,   9   ,   107 ,   37  ,   121 ,   767 ,   1005    ,   2609    ,   2531    ,   14811   ,0 };
            const unsigned long dim1187JoeKuoD7Init[]   =   {   1   ,   3   ,   7   ,   7   ,   3   ,   9   ,   91  ,   203 ,   219 ,   523 ,   319 ,   3021    ,   3465    ,   15579   ,0 };
            const unsigned long dim1188JoeKuoD7Init[]   =   {   1   ,   3   ,   5   ,   1   ,   3   ,   29  ,   97  ,   249 ,   429 ,   665 ,   283 ,   595 ,   1527    ,   4559    ,0 };
            const unsigned long dim1189JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   15  ,   19  ,   23  ,   9   ,   169 ,   427 ,   171 ,   897 ,   3099    ,   7339    ,   13237   ,0 };
            const unsigned long dim1190JoeKuoD7Init[]   =   {   1   ,   3   ,   5   ,   13  ,   11  ,   25  ,   11  ,   123 ,   229 ,   197 ,   1395    ,   2999    ,   3591    ,   5017    ,0 };
            const unsigned long dim1191JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   15  ,   3   ,   11  ,   49  ,   5   ,   217 ,   575 ,   445 ,   1381    ,   4407    ,   9777    ,0 };
            const unsigned long dim1192JoeKuoD7Init[]   =   {   1   ,   1   ,   1   ,   13  ,   23  ,   43  ,   55  ,   19  ,   129 ,   925 ,   1623    ,   4037    ,   2535    ,   871 ,0 };
            const unsigned long dim1193JoeKuoD7Init[]   =   {   1   ,   3   ,   3   ,   15  ,   3   ,   7   ,   29  ,   109 ,   251 ,   627 ,   1751    ,   3613    ,   5741    ,   7859    ,0 };
            const unsigned long dim1194JoeKuoD7Init[]   =   {   1   ,   3   ,   5   ,   3   ,   23  ,   61  ,   115 ,   127 ,   157 ,   725 ,   1321    ,   1143    ,   3737    ,   12559   ,0 };
            const unsigned long dim1195JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   11  ,   25  ,   41  ,   117 ,   217 ,   389 ,   839 ,   1029    ,   547 ,   6149    ,   8335    ,0 };
            const unsigned long dim1196JoeKuoD7Init[]   =   {   1   ,   3   ,   5   ,   1   ,   29  ,   61  ,   101 ,   81  ,   137 ,   547 ,   1989    ,   3907    ,   4855    ,   15949   ,0 };
            const unsigned long dim1197JoeKuoD7Init[]   =   {   1   ,   3   ,   1   ,   3   ,   5   ,   9   ,   15  ,   33  ,   415 ,   725 ,   1183    ,   4091    ,   411 ,   4541    ,0 };
            const unsigned long dim1198JoeKuoD7Init[]   =   {   1   ,   3   ,   3   ,   7   ,   31  ,   61  ,   53  ,   205 ,   157 ,   861 ,   989 ,   3265    ,   2943    ,   12941   ,0 };
            const unsigned long dim1199JoeKuoD7Init[]   =   {   1   ,   3   ,   5   ,   11  ,   11  ,   13  ,   89  ,   39  ,   241 ,   153 ,   439 ,   2087    ,   5161    ,   5705    ,0 };
            const unsigned long dim1200JoeKuoD7Init[]   =   {   1   ,   3   ,   3   ,   11  ,   15  ,   49  ,   39  ,   101 ,   1   ,   793 ,   1429    ,   2109    ,   1233    ,   3717    ,0 };
            const unsigned long dim1201JoeKuoD7Init[]   =   {   1   ,   3   ,   7   ,   9   ,   29  ,   59  ,   45  ,   111 ,   157 ,   421 ,   853 ,   273 ,   1595    ,   1799    ,0 };
            const unsigned long dim1202JoeKuoD7Init[]   =   {   1   ,   3   ,   7   ,   11  ,   3   ,   45  ,   31  ,   89  ,   475 ,   887 ,   563 ,   3985    ,   3011    ,   6903    ,0 };
            const unsigned long dim1203JoeKuoD7Init[]   =   {   1   ,   3   ,   5   ,   15  ,   25  ,   13  ,   69  ,   3   ,   249 ,   435 ,   1333    ,   4079    ,   2259    ,   1971    ,0 };
            const unsigned long dim1204JoeKuoD7Init[]   =   {   1   ,   3   ,   3   ,   11  ,   9   ,   51  ,   7   ,   203 ,   231 ,   155 ,   375 ,   2079    ,   6821    ,   12973   ,0 };
            const unsigned long dim1205JoeKuoD7Init[]   =   {   1   ,   1   ,   1   ,   9   ,   5   ,   23  ,   115 ,   1   ,   307 ,   553 ,   1005    ,   2113    ,   6443    ,   8545    ,0 };
            const unsigned long dim1206JoeKuoD7Init[]   =   {   1   ,   1   ,   3   ,   11  ,   7   ,   49  ,   49  ,   255 ,   73  ,   373 ,   413 ,   2997    ,   1403    ,   9501    ,0 };
            const unsigned long dim1207JoeKuoD7Init[]   =   {   1   ,   3   ,   1   ,   15  ,   11  ,   31  ,   29  ,   43  ,   341 ,   119 ,   1209    ,   1471    ,   5785    ,   9809    ,0 };
            const unsigned long dim1208JoeKuoD7Init[]   =   {   1   ,   1   ,   3   ,   15  ,   3   ,   35  ,   75  ,   187 ,   421 ,   911 ,   169 ,   1427    ,   7243    ,   10511   ,0 };
            const unsigned long dim1209JoeKuoD7Init[]   =   {   1   ,   1   ,   1   ,   9   ,   3   ,   25  ,   21  ,   137 ,   437 ,   737 ,   1373    ,   3417    ,   2951    ,   7623    ,0 };
            const unsigned long dim1210JoeKuoD7Init[]   =   {   1   ,   3   ,   1   ,   15  ,   27  ,   19  ,   97  ,   237 ,   341 ,   679 ,   473 ,   2775    ,   7195    ,   8093    ,0 };
            const unsigned long dim1211JoeKuoD7Init[]   =   {   1   ,   3   ,   7   ,   3   ,   7   ,   45  ,   19  ,   109 ,   499 ,   883 ,   1473    ,   1907    ,   7229    ,   8063    ,0 };
            const unsigned long dim1212JoeKuoD7Init[]   =   {   1   ,   3   ,   1   ,   1   ,   5   ,   49  ,   65  ,   213 ,   49  ,   941 ,   833 ,   3859    ,   3423    ,   5723    ,0 };
            const unsigned long dim1213JoeKuoD7Init[]   =   {   1   ,   3   ,   3   ,   1   ,   27  ,   43  ,   53  ,   21  ,   379 ,   139 ,   1811    ,   133 ,   3641    ,   3021    ,0 };
            const unsigned long dim1214JoeKuoD7Init[]   =   {   1   ,   3   ,   7   ,   9   ,   31  ,   49  ,   13  ,   17  ,   99  ,   773 ,   1127    ,   1237    ,   3593    ,   13233   ,0 };
            const unsigned long dim1215JoeKuoD7Init[]   =   {   1   ,   1   ,   1   ,   9   ,   13  ,   21  ,   117 ,   29  ,   105 ,   671 ,   1423    ,   2449    ,   1979    ,   3717    ,0 };
            const unsigned long dim1216JoeKuoD7Init[]   =   {   1   ,   3   ,   1   ,   13  ,   17  ,   55  ,   37  ,   45  ,   163 ,   475 ,   665 ,   115 ,   2175    ,   10065   ,0 };
            const unsigned long dim1217JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   9   ,   7   ,   31  ,   67  ,   215 ,   153 ,   981 ,   569 ,   2911    ,   6111    ,   12189   ,0 };
            const unsigned long dim1218JoeKuoD7Init[]   =   {   1   ,   1   ,   3   ,   11  ,   19  ,   19  ,   101 ,   27  ,   131 ,   153 ,   505 ,   2897    ,   4573    ,   13713   ,0 };
            const unsigned long dim1219JoeKuoD7Init[]   =   {   1   ,   1   ,   1   ,   9   ,   15  ,   21  ,   99  ,   187 ,   155 ,   135 ,   1421    ,   1421    ,   843 ,   3271    ,0 };
            const unsigned long dim1220JoeKuoD7Init[]   =   {   1   ,   3   ,   1   ,   5   ,   1   ,   9   ,   17  ,   243 ,   111 ,   223 ,   1343    ,   3765    ,   2439    ,   717 ,0 };
            const unsigned long dim1221JoeKuoD7Init[]   =   {   1   ,   1   ,   5   ,   3   ,   21  ,   9   ,   83  ,   191 ,   45  ,   813 ,   1047    ,   1483    ,   209 ,   2895    ,0 };
            const unsigned long dim1222JoeKuoD7Init[]   =   {   1   ,   3   ,   5   ,   11  ,   29  ,   41  ,   119 ,   11  ,   245 ,   145 ,   969 ,   3017    ,   6403    ,   11387   ,0 };
            const unsigned long dim1223JoeKuoD7Init[]   =   {   1   ,   1   ,   1   ,   13  ,   1   ,   19  ,   107 ,   35  ,   49  ,   381 ,   225 ,   595 ,   7145    ,   697 ,0 };
            const unsigned long dim1224JoeKuoD7Init[]   =   {   1   ,   1   ,   1   ,   11  ,   5   ,   15  ,   37  ,   31  ,   117 ,   649 ,   1263    ,   1045    ,   3749    ,   16177   ,0 };
            const unsigned long dim1225JoeKuoD7Init[]   =   {   1   ,   3   ,   7   ,   3   ,   7   ,   7   ,   91  ,   45  ,   395 ,   403 ,   1953    ,   743 ,   2207    ,   12887   ,0 };
            const unsigned long dim1226JoeKuoD7Init[]   =   {   1   ,   1   ,   1   ,   11  ,   13  ,   31  ,   67  ,   7   ,   279 ,   389 ,   775 ,   301 ,   87  ,   6705    ,0 };
            const unsigned long dim1227JoeKuoD7Init[]   =   {   1   ,   3   ,   1   ,   5   ,   27  ,   7   ,   85  ,   183 ,   325 ,   103 ,   1943    ,   2421    ,   4399    ,   7339    ,0 };
            const unsigned long dim1228JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   7   ,   21  ,   13  ,   69  ,   1   ,   289 ,   727 ,   1587    ,   3033    ,   2921    ,   12035   ,0 };
            const unsigned long dim1229JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   11  ,   31  ,   35  ,   103 ,   239 ,   89  ,   983 ,   433 ,   3027    ,   3089    ,   14439   ,0 };
            const unsigned long dim1230JoeKuoD7Init[]   =   {   1   ,   3   ,   5   ,   9   ,   11  ,   9   ,   107 ,   207 ,   87  ,   563 ,   129 ,   2647    ,   3163    ,   11327   ,0 };
            const unsigned long dim1231JoeKuoD7Init[]   =   {   1   ,   3   ,   7   ,   11  ,   9   ,   35  ,   115 ,   189 ,   293 ,   637 ,   213 ,   1111    ,   331 ,   10115   ,0 };
            const unsigned long dim1232JoeKuoD7Init[]   =   {   1   ,   3   ,   5   ,   7   ,   21  ,   59  ,   127 ,   145 ,   157 ,   431 ,   295 ,   4055    ,   6207    ,   5353    ,0 };
            const unsigned long dim1233JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   5   ,   9   ,   53  ,   13  ,   249 ,   275 ,   841 ,   987 ,   3243    ,   7229    ,   6215    ,0 };
            const unsigned long dim1234JoeKuoD7Init[]   =   {   1   ,   3   ,   5   ,   9   ,   11  ,   5   ,   47  ,   247 ,   57  ,   799 ,   1787    ,   2667    ,   3615    ,   4343    ,0 };
            const unsigned long dim1235JoeKuoD7Init[]   =   {   1   ,   1   ,   1   ,   11  ,   31  ,   37  ,   73  ,   77  ,   181 ,   591 ,   445 ,   3081    ,   5337    ,   14259   ,0 };
            const unsigned long dim1236JoeKuoD7Init[]   =   {   1   ,   3   ,   3   ,   11  ,   11  ,   49  ,   59  ,   229 ,   159 ,   139 ,   1435    ,   313 ,   7995    ,   5623    ,0 };
            const unsigned long dim1237JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   13  ,   17  ,   39  ,   53  ,   39  ,   413 ,   451 ,   2029    ,   1549    ,   485 ,   1311    ,0 };
            const unsigned long dim1238JoeKuoD7Init[]   =   {   1   ,   1   ,   5   ,   11  ,   15  ,   57  ,   9   ,   245 ,   397 ,   587 ,   1733    ,   459 ,   2127    ,   16085   ,0 };
            const unsigned long dim1239JoeKuoD7Init[]   =   {   1   ,   3   ,   3   ,   3   ,   17  ,   5   ,   17  ,   231 ,   439 ,   355 ,   1899    ,   1613    ,   1529    ,   12557   ,0 };
            const unsigned long dim1240JoeKuoD7Init[]   =   {   1   ,   1   ,   1   ,   1   ,   5   ,   55  ,   113 ,   209 ,   441 ,   813 ,   1025    ,   191 ,   853 ,   2039    ,0 };
            const unsigned long dim1241JoeKuoD7Init[]   =   {   1   ,   3   ,   3   ,   11  ,   9   ,   47  ,   89  ,   137 ,   187 ,   457 ,   1169    ,   3607    ,   3799    ,   15121   ,0 };
            const unsigned long dim1242JoeKuoD7Init[]   =   {   1   ,   1   ,   3   ,   15  ,   15  ,   27  ,   93  ,   211 ,   137 ,   809 ,   1723    ,   1865    ,   6291    ,   13209   ,0 };
            const unsigned long dim1243JoeKuoD7Init[]   =   {   1   ,   1   ,   5   ,   15  ,   5   ,   15  ,   67  ,   55  ,   445 ,   781 ,   829 ,   3237    ,   4031    ,   5147    ,0 };
            const unsigned long dim1244JoeKuoD7Init[]   =   {   1   ,   3   ,   1   ,   3   ,   17  ,   59  ,   65  ,   35  ,   103 ,   515 ,   1409    ,   395 ,   711 ,   11773   ,0 };
            const unsigned long dim1245JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   11  ,   29  ,   5   ,   21  ,   117 ,   269 ,   963 ,   337 ,   3149    ,   2569    ,   9881    ,0 };
            const unsigned long dim1246JoeKuoD7Init[]   =   {   1   ,   1   ,   5   ,   7   ,   1   ,   57  ,   61  ,   185 ,   431 ,   977 ,   661 ,   1285    ,   2607    ,   13465   ,0 };
            const unsigned long dim1247JoeKuoD7Init[]   =   {   1   ,   3   ,   7   ,   5   ,   25  ,   35  ,   61  ,   227 ,   437 ,   75  ,   1677    ,   3337    ,   4601    ,   10447   ,0 };
            const unsigned long dim1248JoeKuoD7Init[]   =   {   1   ,   3   ,   7   ,   1   ,   29  ,   29  ,   5   ,   163 ,   237 ,   217 ,   97  ,   1607    ,   911 ,   6195    ,0 };
            const unsigned long dim1249JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   3   ,   17  ,   7   ,   101 ,   31  ,   323 ,   191 ,   871 ,   1691    ,   3535    ,   8935    ,0 };
            const unsigned long dim1250JoeKuoD7Init[]   =   {   1   ,   3   ,   5   ,   3   ,   11  ,   47  ,   117 ,   179 ,   25  ,   735 ,   989 ,   1623    ,   5321    ,   12603   ,0 };
            const unsigned long dim1251JoeKuoD7Init[]   =   {   1   ,   1   ,   5   ,   11  ,   1   ,   55  ,   121 ,   221 ,   383 ,   969 ,   777 ,   3419    ,   2621    ,   1651    ,0 };
            const unsigned long dim1252JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   7   ,   21  ,   9   ,   95  ,   239 ,   221 ,   577 ,   1479    ,   1649    ,   6765    ,   5601    ,0 };
            const unsigned long dim1253JoeKuoD7Init[]   =   {   1   ,   3   ,   3   ,   15  ,   29  ,   43  ,   49  ,   71  ,   255 ,   1011    ,   115 ,   1859    ,   7369    ,   16233   ,0 };
            const unsigned long dim1254JoeKuoD7Init[]   =   {   1   ,   1   ,   3   ,   1   ,   13  ,   9   ,   21  ,   193 ,   255 ,   315 ,   1295    ,   1413    ,   6851    ,   4395    ,0 };
            const unsigned long dim1255JoeKuoD7Init[]   =   {   1   ,   3   ,   1   ,   13  ,   21  ,   11  ,   75  ,   87  ,   453 ,   39  ,   503 ,   2937    ,   5089    ,   14211   ,0 };
            const unsigned long dim1256JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   13  ,   1   ,   17  ,   121 ,   153 ,   89  ,   113 ,   1969    ,   65  ,   4907    ,   10165   ,0 };
            const unsigned long dim1257JoeKuoD7Init[]   =   {   1   ,   3   ,   5   ,   1   ,   17  ,   49  ,   75  ,   25  ,   147 ,   357 ,   617 ,   3245    ,   7313    ,   753 ,0 };
            const unsigned long dim1258JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   15  ,   25  ,   3   ,   71  ,   97  ,   463 ,   515 ,   989 ,   69  ,   687 ,   15151   ,0 };
            const unsigned long dim1259JoeKuoD7Init[]   =   {   1   ,   3   ,   3   ,   3   ,   25  ,   61  ,   95  ,   219 ,   429 ,   143 ,   565 ,   463 ,   4165    ,   2619    ,0 };
            const unsigned long dim1260JoeKuoD7Init[]   =   {   1   ,   1   ,   5   ,   13  ,   3   ,   45  ,   11  ,   59  ,   237 ,   387 ,   109 ,   253 ,   651 ,   13061   ,0 };
            const unsigned long dim1261JoeKuoD7Init[]   =   {   1   ,   3   ,   7   ,   1   ,   25  ,   41  ,   111 ,   179 ,   321 ,   139 ,   453 ,   1061    ,   6395    ,   6391    ,0 };
            const unsigned long dim1262JoeKuoD7Init[]   =   {   1   ,   3   ,   3   ,   1   ,   5   ,   11  ,   73  ,   131 ,   469 ,   839 ,   459 ,   2237    ,   2481    ,   10087   ,0 };
            const unsigned long dim1263JoeKuoD7Init[]   =   {   1   ,   3   ,   7   ,   3   ,   13  ,   55  ,   75  ,   111 ,   411 ,   1007    ,   1303    ,   1613    ,   2449    ,   11337   ,0 };
            const unsigned long dim1264JoeKuoD7Init[]   =   {   1   ,   1   ,   3   ,   1   ,   11  ,   39  ,   3   ,   193 ,   113 ,   927 ,   1345    ,   2747    ,   4951    ,   10543   ,0 };
            const unsigned long dim1265JoeKuoD7Init[]   =   {   1   ,   3   ,   1   ,   15  ,   23  ,   39  ,   47  ,   115 ,   129 ,   51  ,   921 ,   1021    ,   6907    ,   8139    ,0 };
            const unsigned long dim1266JoeKuoD7Init[]   =   {   1   ,   3   ,   1   ,   7   ,   31  ,   13  ,   75  ,   79  ,   331 ,   773 ,   1147    ,   335 ,   1729    ,   4015    ,0 };
            const unsigned long dim1267JoeKuoD7Init[]   =   {   1   ,   3   ,   1   ,   7   ,   1   ,   47  ,   15  ,   183 ,   243 ,   525 ,   1515    ,   3209    ,   6597    ,   1895    ,0 };
            const unsigned long dim1268JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   3   ,   7   ,   59  ,   115 ,   223 ,   161 ,   181 ,   533 ,   1085    ,   3405    ,   2565    ,0 };
            const unsigned long dim1269JoeKuoD7Init[]   =   {   1   ,   1   ,   5   ,   1   ,   9   ,   23  ,   57  ,   135 ,   113 ,   57  ,   143 ,   2949    ,   2703    ,   1813    ,0 };
            const unsigned long dim1270JoeKuoD7Init[]   =   {   1   ,   3   ,   1   ,   9   ,   25  ,   43  ,   103 ,   7   ,   167 ,   393 ,   1515    ,   2871    ,   5167    ,   9943    ,0 };
            const unsigned long dim1271JoeKuoD7Init[]   =   {   1   ,   1   ,   5   ,   1   ,   25  ,   1   ,   61  ,   49  ,   137 ,   157 ,   693 ,   3615    ,   1655    ,   14975   ,0 };
            const unsigned long dim1272JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   11  ,   21  ,   41  ,   89  ,   7   ,   171 ,   49  ,   1021    ,   3227    ,   3989    ,   11739   ,0 };
            const unsigned long dim1273JoeKuoD7Init[]   =   {   1   ,   1   ,   3   ,   5   ,   31  ,   49  ,   13  ,   109 ,   435 ,   99  ,   1435    ,   709 ,   4085    ,   1233    ,0 };
            const unsigned long dim1274JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   1   ,   1   ,   55  ,   5   ,   107 ,   435 ,   663 ,   441 ,   3503    ,   5313    ,   159 ,0 };
            const unsigned long dim1275JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   13  ,   19  ,   57  ,   17  ,   13  ,   25  ,   1021    ,   2005    ,   3445    ,   4263    ,   3185    ,0 };
            const unsigned long dim1276JoeKuoD7Init[]   =   {   1   ,   3   ,   1   ,   7   ,   5   ,   5   ,   37  ,   99  ,   237 ,   563 ,   483 ,   1991    ,   7217    ,   5433    ,0 };
            const unsigned long dim1277JoeKuoD7Init[]   =   {   1   ,   3   ,   5   ,   9   ,   13  ,   13  ,   123 ,   51  ,   31  ,   957 ,   1125    ,   2125    ,   2157    ,   11437   ,0 };
            const unsigned long dim1278JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   11  ,   5   ,   59  ,   53  ,   193 ,   337 ,   5   ,   1037    ,   1573    ,   1147    ,   7055    ,0 };
            const unsigned long dim1279JoeKuoD7Init[]   =   {   1   ,   3   ,   1   ,   9   ,   27  ,   47  ,   53  ,   121 ,   421 ,   551 ,   23  ,   1603    ,   1815    ,   2741    ,0 };
            const unsigned long dim1280JoeKuoD7Init[]   =   {   1   ,   3   ,   3   ,   13  ,   27  ,   7   ,   79  ,   255 ,   209 ,   441 ,   1761    ,   1755    ,   7077    ,   3581    ,0 };
            const unsigned long dim1281JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   13  ,   3   ,   43  ,   113 ,   227 ,   43  ,   387 ,   1255    ,   1117    ,   6849    ,   11715   ,0 };
            const unsigned long dim1282JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   13  ,   15  ,   19  ,   59  ,   159 ,   265 ,   283 ,   615 ,   3745    ,   539 ,   15195   ,0 };
            const unsigned long dim1283JoeKuoD7Init[]   =   {   1   ,   3   ,   1   ,   3   ,   7   ,   21  ,   29  ,   135 ,   203 ,   887 ,   549 ,   85  ,   8093    ,   12527   ,0 };
            const unsigned long dim1284JoeKuoD7Init[]   =   {   1   ,   3   ,   7   ,   11  ,   5   ,   35  ,   41  ,   61  ,   273 ,   187 ,   1655    ,   2449    ,   2173    ,   14737   ,0 };
            const unsigned long dim1285JoeKuoD7Init[]   =   {   1   ,   3   ,   1   ,   9   ,   29  ,   49  ,   11  ,   33  ,   477 ,   671 ,   1455    ,   2577    ,   5239    ,   5025    ,0 };
            const unsigned long dim1286JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   5   ,   23  ,   15  ,   67  ,   73  ,   73  ,   45  ,   1727    ,   645 ,   2623    ,   10859   ,0 };
            const unsigned long dim1287JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   15  ,   29  ,   37  ,   45  ,   183 ,   97  ,   297 ,   593 ,   833 ,   5833    ,   4777    ,0 };
            const unsigned long dim1288JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   5   ,   27  ,   21  ,   5   ,   253 ,   243 ,   993 ,   1701    ,   1695    ,   4037    ,   1891    ,0 };
            const unsigned long dim1289JoeKuoD7Init[]   =   {   1   ,   3   ,   7   ,   1   ,   17  ,   37  ,   123 ,   21  ,   95  ,   533 ,   111 ,   2605    ,   7011    ,   7957    ,0 };
            const unsigned long dim1290JoeKuoD7Init[]   =   {   1   ,   3   ,   1   ,   1   ,   5   ,   43  ,   11  ,   143 ,   457 ,   607 ,   469 ,   1479    ,   6181    ,   7579    ,0 };
            const unsigned long dim1291JoeKuoD7Init[]   =   {   1   ,   3   ,   3   ,   15  ,   5   ,   59  ,   105 ,   111 ,   47  ,   53  ,   215 ,   1601    ,   8057    ,   4233    ,0 };
            const unsigned long dim1292JoeKuoD7Init[]   =   {   1   ,   3   ,   1   ,   15  ,   7   ,   5   ,   17  ,   157 ,   47  ,   541 ,   1487    ,   2623    ,   1705    ,   5565    ,0 };
            const unsigned long dim1293JoeKuoD7Init[]   =   {   1   ,   1   ,   1   ,   7   ,   21  ,   27  ,   63  ,   117 ,   323 ,   379 ,   1407    ,   4043    ,   3763    ,   10367   ,0 };
            const unsigned long dim1294JoeKuoD7Init[]   =   {   1   ,   1   ,   3   ,   7   ,   27  ,   11  ,   35  ,   81  ,   447 ,   279 ,   811 ,   1355    ,   4185    ,   8281    ,0 };
            const unsigned long dim1295JoeKuoD7Init[]   =   {   1   ,   3   ,   5   ,   15  ,   19  ,   41  ,   111 ,   225 ,   133 ,   79  ,   347 ,   3595    ,   7863    ,   5239    ,0 };
            const unsigned long dim1296JoeKuoD7Init[]   =   {   1   ,   3   ,   1   ,   5   ,   17  ,   41  ,   77  ,   125 ,   221 ,   93  ,   551 ,   1263    ,   1973    ,   12081   ,0 };
            const unsigned long dim1297JoeKuoD7Init[]   =   {   1   ,   3   ,   5   ,   3   ,   17  ,   55  ,   9   ,   149 ,   203 ,   217 ,   1235    ,   1905    ,   6139    ,   7223    ,0 };
            const unsigned long dim1298JoeKuoD7Init[]   =   {   1   ,   3   ,   1   ,   9   ,   1   ,   3   ,   17  ,   5   ,   157 ,   739 ,   691 ,   1471    ,   709 ,   1355    ,0 };
            const unsigned long dim1299JoeKuoD7Init[]   =   {   1   ,   1   ,   1   ,   9   ,   3   ,   13  ,   87  ,   153 ,   49  ,   391 ,   47  ,   1703    ,   1913    ,   12885   ,0 };
            const unsigned long dim1300JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   3   ,   19  ,   49  ,   91  ,   47  ,   263 ,   287 ,   1885    ,   3137    ,   3709    ,   8895    ,0 };
            const unsigned long dim1301JoeKuoD7Init[]   =   {   1   ,   1   ,   5   ,   5   ,   19  ,   55  ,   59  ,   63  ,   229 ,   705 ,   1791    ,   1789    ,   4083    ,   11827   ,0 };
            const unsigned long dim1302JoeKuoD7Init[]   =   {   1   ,   1   ,   3   ,   13  ,   21  ,   31  ,   127 ,   143 ,   57  ,   933 ,   125 ,   813 ,   6121    ,   5883    ,0 };
            const unsigned long dim1303JoeKuoD7Init[]   =   {   1   ,   3   ,   5   ,   7   ,   7   ,   45  ,   39  ,   137 ,   219 ,   249 ,   791 ,   751 ,   4877    ,   6935    ,0 };
            const unsigned long dim1304JoeKuoD7Init[]   =   {   1   ,   3   ,   1   ,   5   ,   31  ,   21  ,   11  ,   191 ,   399 ,   339 ,   815 ,   2459    ,   7621    ,   1279    ,0 };
            const unsigned long dim1305JoeKuoD7Init[]   =   {   1   ,   1   ,   5   ,   13  ,   11  ,   49  ,   111 ,   101 ,   367 ,   645 ,   99  ,   1759    ,   2275    ,   8289    ,0 };
            const unsigned long dim1306JoeKuoD7Init[]   =   {   1   ,   1   ,   1   ,   9   ,   15  ,   23  ,   17  ,   93  ,   339 ,   705 ,   1487    ,   3325    ,   6585    ,   13281   ,0 };
            const unsigned long dim1307JoeKuoD7Init[]   =   {   1   ,   1   ,   3   ,   1   ,   17  ,   25  ,   69  ,   115 ,   339 ,   241 ,   777 ,   1585    ,   2565    ,   14843   ,0 };
            const unsigned long dim1308JoeKuoD7Init[]   =   {   1   ,   3   ,   5   ,   7   ,   5   ,   45  ,   99  ,   161 ,   505 ,   511 ,   591 ,   2109    ,   5885    ,   12547   ,0 };
            const unsigned long dim1309JoeKuoD7Init[]   =   {   1   ,   3   ,   5   ,   11  ,   13  ,   45  ,   87  ,   105 ,   387 ,   339 ,   1079    ,   705 ,   2385    ,   9375    ,0 };
            const unsigned long dim1310JoeKuoD7Init[]   =   {   1   ,   3   ,   3   ,   1   ,   5   ,   19  ,   111 ,   155 ,   147 ,   897 ,   1879    ,   243 ,   3745    ,   2309    ,0 };
            const unsigned long dim1311JoeKuoD7Init[]   =   {   1   ,   3   ,   3   ,   1   ,   1   ,   1   ,   123 ,   153 ,   409 ,   101 ,   105 ,   1713    ,   5161    ,   8915    ,0 };
            const unsigned long dim1312JoeKuoD7Init[]   =   {   1   ,   3   ,   3   ,   11  ,   17  ,   5   ,   53  ,   145 ,   243 ,   811 ,   1413    ,   3735    ,   6857    ,   1151    ,0 };
            const unsigned long dim1313JoeKuoD7Init[]   =   {   1   ,   1   ,   3   ,   13  ,   9   ,   35  ,   127 ,   245 ,   479 ,   691 ,   281 ,   845 ,   4017    ,   14011   ,0 };
            const unsigned long dim1314JoeKuoD7Init[]   =   {   1   ,   1   ,   1   ,   15  ,   3   ,   27  ,   11  ,   231 ,   47  ,   1023    ,   1785    ,   4039    ,   137 ,   3429    ,0 };
            const unsigned long dim1315JoeKuoD7Init[]   =   {   1   ,   1   ,   5   ,   5   ,   19  ,   45  ,   7   ,   251 ,   191 ,   387 ,   1853    ,   3851    ,   6237    ,   6475    ,0 };
            const unsigned long dim1316JoeKuoD7Init[]   =   {   1   ,   1   ,   3   ,   15  ,   27  ,   19  ,   71  ,   13  ,   65  ,   679 ,   281 ,   2141    ,   1211    ,   4063    ,0 };
            const unsigned long dim1317JoeKuoD7Init[]   =   {   1   ,   3   ,   5   ,   15  ,   7   ,   31  ,   73  ,   103 ,   393 ,   177 ,   1929    ,   809 ,   3389    ,   5673    ,0 };
            const unsigned long dim1318JoeKuoD7Init[]   =   {   1   ,   1   ,   5   ,   7   ,   1   ,   45  ,   35  ,   247 ,   485 ,   25  ,   1079    ,   4019    ,   6101    ,   4715    ,0 };
            const unsigned long dim1319JoeKuoD7Init[]   =   {   1   ,   3   ,   7   ,   7   ,   27  ,   13  ,   121 ,   177 ,   377 ,   585 ,   1503    ,   1249    ,   715 ,   309 ,0 };
            const unsigned long dim1320JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   7   ,   23  ,   19  ,   37  ,   19  ,   243 ,   261 ,   139 ,   1261    ,   3775    ,   10299   ,0 };
            const unsigned long dim1321JoeKuoD7Init[]   =   {   1   ,   1   ,   1   ,   11  ,   23  ,   1   ,   123 ,   225 ,   497 ,   131 ,   1863    ,   4061    ,   1247    ,   12511   ,0 };
            const unsigned long dim1322JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   15  ,   5   ,   25  ,   59  ,   213 ,   157 ,   697 ,   671 ,   1431    ,   2891    ,   6881    ,0 };
            const unsigned long dim1323JoeKuoD7Init[]   =   {   1   ,   1   ,   5   ,   9   ,   1   ,   33  ,   67  ,   121 ,   165 ,   867 ,   393 ,   1439    ,   1255    ,   11839   ,0 };
            const unsigned long dim1324JoeKuoD7Init[]   =   {   1   ,   3   ,   5   ,   11  ,   23  ,   43  ,   11  ,   23  ,   507 ,   585 ,   277 ,   1077    ,   2697    ,   4901    ,0 };
            const unsigned long dim1325JoeKuoD7Init[]   =   {   1   ,   3   ,   5   ,   15  ,   17  ,   35  ,   75  ,   177 ,   373 ,   113 ,   401 ,   1999    ,   1595    ,   10635   ,0 };
            const unsigned long dim1326JoeKuoD7Init[]   =   {   1   ,   3   ,   7   ,   11  ,   31  ,   49  ,   65  ,   225 ,   217 ,   783 ,   787 ,   1421    ,   5701    ,   2185    ,0 };
            const unsigned long dim1327JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   15  ,   13  ,   45  ,   111 ,   23  ,   89  ,   263 ,   699 ,   4037    ,   6937    ,   6265    ,0 };
            const unsigned long dim1328JoeKuoD7Init[]   =   {   1   ,   1   ,   1   ,   1   ,   15  ,   51  ,   49  ,   9   ,   221 ,   85  ,   1983    ,   2865    ,   2841    ,   3105    ,0 };
            const unsigned long dim1329JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   9   ,   5   ,   35  ,   75  ,   213 ,   177 ,   131 ,   1227    ,   4069    ,   5035    ,   4371    ,0 };
            const unsigned long dim1330JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   5   ,   1   ,   33  ,   75  ,   125 ,   5   ,   283 ,   667 ,   1521    ,   3267    ,   9641    ,0 };
            const unsigned long dim1331JoeKuoD7Init[]   =   {   1   ,   3   ,   3   ,   11  ,   11  ,   49  ,   75  ,   125 ,   349 ,   887 ,   269 ,   2295    ,   4049    ,   1217    ,0 };
            const unsigned long dim1332JoeKuoD7Init[]   =   {   1   ,   1   ,   3   ,   13  ,   17  ,   23  ,   19  ,   141 ,   213 ,   33  ,   475 ,   2543    ,   3617    ,   5791    ,0 };
            const unsigned long dim1333JoeKuoD7Init[]   =   {   1   ,   3   ,   3   ,   11  ,   3   ,   39  ,   95  ,   139 ,   277 ,   109 ,   601 ,   2011    ,   2459    ,   5915    ,0 };
            const unsigned long dim1334JoeKuoD7Init[]   =   {   1   ,   1   ,   5   ,   5   ,   27  ,   61  ,   35  ,   221 ,   329 ,   411 ,   1811    ,   3397    ,   1545    ,   10615   ,0 };
            const unsigned long dim1335JoeKuoD7Init[]   =   {   1   ,   3   ,   5   ,   3   ,   3   ,   55  ,   71  ,   187 ,   371 ,   223 ,   1711    ,   3279    ,   5541    ,   4203    ,0 };
            const unsigned long dim1336JoeKuoD7Init[]   =   {   1   ,   3   ,   7   ,   3   ,   27  ,   61  ,   23  ,   143 ,   475 ,   693 ,   1717    ,   353 ,   4381    ,   11355   ,0 };
            const unsigned long dim1337JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   13  ,   9   ,   23  ,   109 ,   11  ,   483 ,   1005    ,   1709    ,   2883    ,   5021    ,   10079   ,0 };
            const unsigned long dim1338JoeKuoD7Init[]   =   {   1   ,   1   ,   5   ,   9   ,   25  ,   19  ,   79  ,   169 ,   433 ,   177 ,   1071    ,   3413    ,   2071    ,   3565    ,0 };
            const unsigned long dim1339JoeKuoD7Init[]   =   {   1   ,   1   ,   5   ,   7   ,   7   ,   47  ,   105 ,   225 ,   153 ,   949 ,   9   ,   2383    ,   3947    ,   2711    ,0 };
            const unsigned long dim1340JoeKuoD7Init[]   =   {   1   ,   3   ,   7   ,   9   ,   1   ,   17  ,   87  ,   159 ,   107 ,   41  ,   1459    ,   2133    ,   6261    ,   15637   ,0 };
            const unsigned long dim1341JoeKuoD7Init[]   =   {   1   ,   3   ,   3   ,   3   ,   1   ,   7   ,   45  ,   255 ,   377 ,   811 ,   1911    ,   2727    ,   7223    ,   3825    ,0 };
            const unsigned long dim1342JoeKuoD7Init[]   =   {   1   ,   1   ,   1   ,   13  ,   13  ,   13  ,   5   ,   59  ,   421 ,   585 ,   1651    ,   55  ,   2877    ,   11643   ,0 };
            const unsigned long dim1343JoeKuoD7Init[]   =   {   1   ,   1   ,   3   ,   11  ,   15  ,   57  ,   1   ,   173 ,   189 ,   485 ,   1881    ,   3449    ,   1831    ,   2501    ,0 };
            const unsigned long dim1344JoeKuoD7Init[]   =   {   1   ,   3   ,   3   ,   7   ,   21  ,   29  ,   49  ,   255 ,   313 ,   137 ,   51  ,   2281    ,   973 ,   16053   ,0 };
            const unsigned long dim1345JoeKuoD7Init[]   =   {   1   ,   1   ,   1   ,   11  ,   23  ,   61  ,   99  ,   77  ,   399 ,   649 ,   1567    ,   1701    ,   4747    ,   1301    ,0 };
            const unsigned long dim1346JoeKuoD7Init[]   =   {   1   ,   3   ,   7   ,   3   ,   11  ,   57  ,   45  ,   169 ,   247 ,   73  ,   1685    ,   341 ,   6791    ,   2833    ,0 };
            const unsigned long dim1347JoeKuoD7Init[]   =   {   1   ,   1   ,   5   ,   13  ,   9   ,   5   ,   17  ,   209 ,   329 ,   379 ,   1787    ,   737 ,   7271    ,   7357    ,0 };
            const unsigned long dim1348JoeKuoD7Init[]   =   {   1   ,   3   ,   1   ,   9   ,   13  ,   7   ,   97  ,   165 ,   99  ,   599 ,   471 ,   2945    ,   3125    ,   10413   ,0 };
            const unsigned long dim1349JoeKuoD7Init[]   =   {   1   ,   3   ,   1   ,   11  ,   31  ,   29  ,   45  ,   3   ,   457 ,   673 ,   589 ,   2521    ,   5655    ,   13973   ,0 };
            const unsigned long dim1350JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   1   ,   15  ,   15  ,   61  ,   77  ,   321 ,   317 ,   993 ,   1101    ,   5885    ,   5105    ,0 };
            const unsigned long dim1351JoeKuoD7Init[]   =   {   1   ,   1   ,   1   ,   5   ,   7   ,   45  ,   111 ,   129 ,   407 ,   117 ,   93  ,   2399    ,   1049    ,   12435   ,0 };
            const unsigned long dim1352JoeKuoD7Init[]   =   {   1   ,   3   ,   7   ,   3   ,   11  ,   29  ,   103 ,   189 ,   29  ,   899 ,   1141    ,   1799    ,   4439    ,   2907    ,0 };
            const unsigned long dim1353JoeKuoD7Init[]   =   {   1   ,   1   ,   1   ,   5   ,   3   ,   39  ,   25  ,   255 ,   97  ,   1015    ,   727 ,   377 ,   8015    ,   12617   ,0 };
            const unsigned long dim1354JoeKuoD7Init[]   =   {   1   ,   3   ,   3   ,   11  ,   29  ,   23  ,   103 ,   51  ,   95  ,   333 ,   171 ,   3019    ,   1373    ,   13379   ,0 };
            const unsigned long dim1355JoeKuoD7Init[]   =   {   1   ,   3   ,   1   ,   9   ,   21  ,   49  ,   57  ,   157 ,   337 ,   141 ,   307 ,   2333    ,   1857    ,   8739    ,0 };
            const unsigned long dim1356JoeKuoD7Init[]   =   {   1   ,   1   ,   1   ,   11  ,   9   ,   35  ,   33  ,   93  ,   211 ,   969 ,   1093    ,   4053    ,   1313    ,   12771   ,0 };
            const unsigned long dim1357JoeKuoD7Init[]   =   {   1   ,   3   ,   1   ,   9   ,   27  ,   7   ,   77  ,   65  ,   425 ,   603 ,   261 ,   2321    ,   579 ,   1249    ,0 };
            const unsigned long dim1358JoeKuoD7Init[]   =   {   1   ,   3   ,   5   ,   13  ,   29  ,   39  ,   21  ,   243 ,   115 ,   201 ,   749 ,   3049    ,   3113    ,   1485    ,0 };
            const unsigned long dim1359JoeKuoD7Init[]   =   {   1   ,   1   ,   5   ,   15  ,   23  ,   53  ,   27  ,   189 ,   193 ,   281 ,   149 ,   2205    ,   6149    ,   2669    ,0 };
            const unsigned long dim1360JoeKuoD7Init[]   =   {   1   ,   1   ,   5   ,   13  ,   9   ,   13  ,   65  ,   37  ,   161 ,   59  ,   1691    ,   435 ,   6077    ,   14827   ,0 };
            const unsigned long dim1361JoeKuoD7Init[]   =   {   1   ,   1   ,   3   ,   15  ,   5   ,   35  ,   41  ,   239 ,   177 ,   265 ,   389 ,   1829    ,   1229    ,   14351   ,0 };
            const unsigned long dim1362JoeKuoD7Init[]   =   {   1   ,   1   ,   5   ,   3   ,   7   ,   27  ,   61  ,   107 ,   171 ,   411 ,   685 ,   455 ,   5175    ,   4981    ,0 };
            const unsigned long dim1363JoeKuoD7Init[]   =   {   1   ,   1   ,   3   ,   7   ,   13  ,   23  ,   107 ,   63  ,   71  ,   755 ,   535 ,   1819    ,   1545    ,   2075    ,0 };
            const unsigned long dim1364JoeKuoD7Init[]   =   {   1   ,   1   ,   5   ,   15  ,   1   ,   27  ,   61  ,   247 ,   333 ,   549 ,   981 ,   3201    ,   919 ,   1505    ,0 };
            const unsigned long dim1365JoeKuoD7Init[]   =   {   1   ,   1   ,   3   ,   5   ,   29  ,   27  ,   93  ,   147 ,   105 ,   83  ,   209 ,   61  ,   3045    ,   4639    ,0 };
            const unsigned long dim1366JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   7   ,   25  ,   35  ,   77  ,   59  ,   157 ,   993 ,   325 ,   3085    ,   671 ,   3573    ,0 };
            const unsigned long dim1367JoeKuoD7Init[]   =   {   1   ,   3   ,   1   ,   15  ,   15  ,   45  ,   31  ,   93  ,   367 ,   747 ,   969 ,   3443    ,   3825    ,   2997    ,0 };
            const unsigned long dim1368JoeKuoD7Init[]   =   {   1   ,   1   ,   1   ,   5   ,   27  ,   21  ,   23  ,   63  ,   271 ,   515 ,   261 ,   1947    ,   6257    ,   12861   ,0 };
            const unsigned long dim1369JoeKuoD7Init[]   =   {   1   ,   3   ,   7   ,   7   ,   27  ,   51  ,   31  ,   229 ,   225 ,   283 ,   947 ,   759 ,   5431    ,   2161    ,0 };
            const unsigned long dim1370JoeKuoD7Init[]   =   {   1   ,   1   ,   3   ,   13  ,   7   ,   39  ,   5   ,   223 ,   55  ,   999 ,   697 ,   3383    ,   1605    ,   15315   ,0 };
            const unsigned long dim1371JoeKuoD7Init[]   =   {   1   ,   1   ,   3   ,   3   ,   29  ,   13  ,   103 ,   215 ,   73  ,   715 ,   1117    ,   3399    ,   1867    ,   8049    ,0 };
            const unsigned long dim1372JoeKuoD7Init[]   =   {   1   ,   3   ,   3   ,   3   ,   1   ,   33  ,   121 ,   135 ,   105 ,   335 ,   105 ,   815 ,   1601    ,   15785   ,0 };
            const unsigned long dim1373JoeKuoD7Init[]   =   {   1   ,   1   ,   3   ,   13  ,   13  ,   27  ,   39  ,   119 ,   271 ,   75  ,   1531    ,   649 ,   8035    ,   5533    ,0 };
            const unsigned long dim1374JoeKuoD7Init[]   =   {   1   ,   3   ,   7   ,   3   ,   9   ,   47  ,   109 ,   85  ,   135 ,   553 ,   1091    ,   3077    ,   5527    ,   15225   ,0 };
            const unsigned long dim1375JoeKuoD7Init[]   =   {   1   ,   3   ,   3   ,   5   ,   31  ,   11  ,   127 ,   197 ,   257 ,   281 ,   1939    ,   3711    ,   2871    ,   11953   ,0 };
            const unsigned long dim1376JoeKuoD7Init[]   =   {   1   ,   1   ,   1   ,   15  ,   21  ,   51  ,   57  ,   159 ,   445 ,   521 ,   27  ,   1313    ,   6031    ,   2891    ,0 };
            const unsigned long dim1377JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   15  ,   21  ,   57  ,   53  ,   29  ,   39  ,   547 ,   1995    ,   3605    ,   3899    ,   1199    ,0 };
            const unsigned long dim1378JoeKuoD7Init[]   =   {   1   ,   1   ,   5   ,   3   ,   1   ,   59  ,   69  ,   21  ,   383 ,   743 ,   87  ,   3369    ,   2309    ,   9651    ,0 };
            const unsigned long dim1379JoeKuoD7Init[]   =   {   1   ,   3   ,   7   ,   13  ,   3   ,   9   ,   41  ,   77  ,   201 ,   331 ,   1007    ,   1503    ,   5771    ,   567 ,0 };
            const unsigned long dim1380JoeKuoD7Init[]   =   {   1   ,   1   ,   3   ,   15  ,   19  ,   19  ,   61  ,   221 ,   257 ,   903 ,   83  ,   1923    ,   6453    ,   11275   ,0 };
            const unsigned long dim1381JoeKuoD7Init[]   =   {   1   ,   1   ,   5   ,   13  ,   21  ,   21  ,   101 ,   109 ,   419 ,   409 ,   125 ,   3337    ,   5763    ,   4945    ,0 };
            const unsigned long dim1382JoeKuoD7Init[]   =   {   1   ,   1   ,   5   ,   11  ,   27  ,   21  ,   39  ,   63  ,   475 ,   285 ,   917 ,   4017    ,   1491    ,   9479    ,0 };
            const unsigned long dim1383JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   1   ,   23  ,   59  ,   119 ,   89  ,   309 ,   567 ,   1781    ,   2369    ,   4311    ,   16033   ,0 };
            const unsigned long dim1384JoeKuoD7Init[]   =   {   1   ,   1   ,   1   ,   3   ,   13  ,   51  ,   69  ,   71  ,   89  ,   1013    ,   329 ,   69  ,   1107    ,   14509   ,0 };
            const unsigned long dim1385JoeKuoD7Init[]   =   {   1   ,   1   ,   3   ,   7   ,   23  ,   39  ,   9   ,   137 ,   65  ,   83  ,   1001    ,   271 ,   2127    ,   9715    ,0 };
            const unsigned long dim1386JoeKuoD7Init[]   =   {   1   ,   3   ,   3   ,   9   ,   13  ,   1   ,   67  ,   233 ,   333 ,   365 ,   1717    ,   77  ,   7173    ,   1271    ,0 };
            const unsigned long dim1387JoeKuoD7Init[]   =   {   1   ,   3   ,   7   ,   7   ,   15  ,   3   ,   115 ,   237 ,   467 ,   827 ,   389 ,   1901    ,   6767    ,   7809    ,0 };
            const unsigned long dim1388JoeKuoD7Init[]   =   {   1   ,   3   ,   1   ,   15  ,   23  ,   27  ,   117 ,   99  ,   49  ,   863 ,   1775    ,   1729    ,   8069    ,   9917    ,0 };
            const unsigned long dim1389JoeKuoD7Init[]   =   {   1   ,   3   ,   1   ,   11  ,   23  ,   47  ,   75  ,   91  ,   47  ,   99  ,   1187    ,   1911    ,   2453    ,   5055    ,0 };
            const unsigned long dim1390JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   11  ,   15  ,   51  ,   43  ,   133 ,   175 ,   647 ,   1053    ,   439 ,   1933    ,   7339    ,0 };
            const unsigned long dim1391JoeKuoD7Init[]   =   {   1   ,   1   ,   5   ,   13  ,   5   ,   37  ,   105 ,   243 ,   467 ,   255 ,   715 ,   1023    ,   4331    ,   8943    ,0 };
            const unsigned long dim1392JoeKuoD7Init[]   =   {   1   ,   1   ,   1   ,   15  ,   15  ,   55  ,   59  ,   111 ,   351 ,   345 ,   1517    ,   3009    ,   4149    ,   8511    ,0 };
            const unsigned long dim1393JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   5   ,   3   ,   55  ,   123 ,   61  ,   405 ,   719 ,   245 ,   2745    ,   7471    ,   2205    ,0 };
            const unsigned long dim1394JoeKuoD7Init[]   =   {   1   ,   1   ,   1   ,   9   ,   19  ,   25  ,   43  ,   105 ,   215 ,   117 ,   563 ,   513 ,   4609    ,   13307   ,0 };
            const unsigned long dim1395JoeKuoD7Init[]   =   {   1   ,   3   ,   5   ,   5   ,   21  ,   27  ,   107 ,   131 ,   329 ,   519 ,   369 ,   2523    ,   4099    ,   13943   ,0 };
            const unsigned long dim1396JoeKuoD7Init[]   =   {   1   ,   1   ,   3   ,   13  ,   27  ,   15  ,   25  ,   63  ,   5   ,   513 ,   1641    ,   3955    ,   4233    ,   12263   ,0 };
            const unsigned long dim1397JoeKuoD7Init[]   =   {   1   ,   3   ,   5   ,   5   ,   3   ,   51  ,   49  ,   49  ,   197 ,   919 ,   281 ,   3713    ,   2179    ,   4573    ,0 };
            const unsigned long dim1398JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   5   ,   25  ,   15  ,   49  ,   159 ,   53  ,   609 ,   573 ,   1533    ,   4589    ,   10593   ,0 };
            const unsigned long dim1399JoeKuoD7Init[]   =   {   1   ,   3   ,   5   ,   15  ,   7   ,   21  ,   61  ,   129 ,   53  ,   95  ,   1107    ,   57  ,   2559    ,   11977   ,0 };
            const unsigned long dim1400JoeKuoD7Init[]   =   {   1   ,   1   ,   5   ,   5   ,   5   ,   5   ,   75  ,   153 ,   359 ,   345 ,   1739    ,   2047    ,   6355    ,   5801    ,0 };
            const unsigned long dim1401JoeKuoD7Init[]   =   {   1   ,   3   ,   1   ,   13  ,   17  ,   5   ,   39  ,   209 ,   399 ,   127 ,   1215    ,   2377    ,   7707    ,   5191    ,0 };
            const unsigned long dim1402JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   3   ,   25  ,   31  ,   33  ,   93  ,   1   ,   229 ,   1379    ,   1587    ,   2245    ,   16041   ,0 };
            const unsigned long dim1403JoeKuoD7Init[]   =   {   1   ,   3   ,   5   ,   13  ,   25  ,   51  ,   5   ,   1   ,   237 ,   471 ,   637 ,   3205    ,   4629    ,   15929   ,0 };
            const unsigned long dim1404JoeKuoD7Init[]   =   {   1   ,   3   ,   5   ,   11  ,   27  ,   57  ,   45  ,   141 ,   167 ,   949 ,   1839    ,   289 ,   2637    ,   2041    ,0 };
            const unsigned long dim1405JoeKuoD7Init[]   =   {   1   ,   1   ,   3   ,   7   ,   9   ,   57  ,   69  ,   93  ,   301 ,   805 ,   1265    ,   3415    ,   2395    ,   2949    ,0 };
            const unsigned long dim1406JoeKuoD7Init[]   =   {   1   ,   1   ,   3   ,   15  ,   9   ,   61  ,   17  ,   185 ,   201 ,   743 ,   1903    ,   1161    ,   1693    ,   15125   ,0 };
            const unsigned long dim1407JoeKuoD7Init[]   =   {   1   ,   3   ,   3   ,   3   ,   21  ,   47  ,   125 ,   117 ,   213 ,   761 ,   1357    ,   2455    ,   7065    ,   7335    ,0 };
            const unsigned long dim1408JoeKuoD7Init[]   =   {   1   ,   3   ,   5   ,   7   ,   23  ,   15  ,   39  ,   17  ,   373 ,   777 ,   761 ,   3853    ,   3607    ,   2119    ,0 };
            const unsigned long dim1409JoeKuoD7Init[]   =   {   1   ,   3   ,   1   ,   1   ,   31  ,   9   ,   45  ,   109 ,   209 ,   797 ,   777 ,   2381    ,   5047    ,   12453   ,0 };
            const unsigned long dim1410JoeKuoD7Init[]   =   {   1   ,   3   ,   1   ,   1   ,   1   ,   9   ,   13  ,   171 ,   223 ,   987 ,   1815    ,   1953    ,   4439    ,   3973    ,0 };
            const unsigned long dim1411JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   9   ,   19  ,   11  ,   37  ,   87  ,   117 ,   399 ,   705 ,   2839    ,   3699    ,   13249   ,0 };
            const unsigned long dim1412JoeKuoD7Init[]   =   {   1   ,   3   ,   3   ,   1   ,   21  ,   45  ,   25  ,   41  ,   133 ,   721 ,   289 ,   727 ,   7835    ,   9811    ,0 };
            const unsigned long dim1413JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   9   ,   23  ,   7   ,   75  ,   233 ,   113 ,   945 ,   1447    ,   4031    ,   6791    ,   6285    ,0 };
            const unsigned long dim1414JoeKuoD7Init[]   =   {   1   ,   1   ,   3   ,   9   ,   3   ,   1   ,   127 ,   139 ,   391 ,   745 ,   807 ,   3229    ,   1809    ,   7587    ,0 };
            const unsigned long dim1415JoeKuoD7Init[]   =   {   1   ,   1   ,   5   ,   3   ,   29  ,   47  ,   99  ,   115 ,   37  ,   487 ,   1897    ,   2433    ,   4925    ,   9747    ,0 };
            const unsigned long dim1416JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   15  ,   23  ,   25  ,   95  ,   165 ,   257 ,   67  ,   1445    ,   4055    ,   3761    ,   10411   ,0 };
            const unsigned long dim1417JoeKuoD7Init[]   =   {   1   ,   3   ,   3   ,   11  ,   5   ,   27  ,   77  ,   197 ,   9   ,   679 ,   1259    ,   3289    ,   621 ,   2733    ,0 };
            const unsigned long dim1418JoeKuoD7Init[]   =   {   1   ,   3   ,   7   ,   9   ,   27  ,   51  ,   57  ,   189 ,   511 ,   633 ,   431 ,   43  ,   2613    ,   5863    ,0 };
            const unsigned long dim1419JoeKuoD7Init[]   =   {   1   ,   1   ,   1   ,   11  ,   29  ,   53  ,   79  ,   237 ,   15  ,   377 ,   45  ,   347 ,   5733    ,   3531    ,0 };
            const unsigned long dim1420JoeKuoD7Init[]   =   {   1   ,   1   ,   1   ,   1   ,   25  ,   7   ,   87  ,   211 ,   429 ,   745 ,   673 ,   3967    ,   2059    ,   785 ,0 };
            const unsigned long dim1421JoeKuoD7Init[]   =   {   1   ,   3   ,   1   ,   5   ,   21  ,   31  ,   7   ,   43  ,   381 ,   635 ,   35  ,   1247    ,   8061    ,   3907    ,0 };
            const unsigned long dim1422JoeKuoD7Init[]   =   {   1   ,   3   ,   1   ,   9   ,   11  ,   39  ,   79  ,   9   ,   395 ,   611 ,   1087    ,   2837    ,   6979    ,   7651    ,0 };
            const unsigned long dim1423JoeKuoD7Init[]   =   {   1   ,   3   ,   3   ,   13  ,   11  ,   37  ,   35  ,   61  ,   373 ,   1001    ,   1849    ,   2169    ,   903 ,   10993   ,0 };
            const unsigned long dim1424JoeKuoD7Init[]   =   {   1   ,   1   ,   3   ,   3   ,   29  ,   3   ,   43  ,   155 ,   403 ,   597 ,   1437    ,   4031    ,   7169    ,   15333   ,0 };
            const unsigned long dim1425JoeKuoD7Init[]   =   {   1   ,   3   ,   5   ,   1   ,   17  ,   29  ,   57  ,   245 ,   505 ,   915 ,   933 ,   1313    ,   7295    ,   12105   ,0 };
            const unsigned long dim1426JoeKuoD7Init[]   =   {   1   ,   1   ,   3   ,   9   ,   29  ,   13  ,   49  ,   89  ,   377 ,   131 ,   31  ,   1869    ,   1717    ,   7679    ,0 };
            const unsigned long dim1427JoeKuoD7Init[]   =   {   1   ,   1   ,   1   ,   7   ,   27  ,   43  ,   1   ,   107 ,   167 ,   851 ,   743 ,   2837    ,   6705    ,   2913    ,0 };
            const unsigned long dim1428JoeKuoD7Init[]   =   {   1   ,   1   ,   3   ,   13  ,   5   ,   27  ,   77  ,   91  ,   305 ,   941 ,   1293    ,   2343    ,   5397    ,   6487    ,0 };
            const unsigned long dim1429JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   5   ,   11  ,   29  ,   25  ,   233 ,   391 ,   221 ,   1285    ,   3995    ,   2713    ,   9125    ,0 };
            const unsigned long dim1430JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   3   ,   11  ,   47  ,   77  ,   31  ,   233 ,   775 ,   1399    ,   1519    ,   3427    ,   10027   ,0 };
            const unsigned long dim1431JoeKuoD7Init[]   =   {   1   ,   3   ,   3   ,   3   ,   5   ,   7   ,   37  ,   111 ,   265 ,   541 ,   77  ,   3327    ,   8115    ,   5541    ,0 };
            const unsigned long dim1432JoeKuoD7Init[]   =   {   1   ,   1   ,   1   ,   5   ,   25  ,   27  ,   91  ,   185 ,   447 ,   225 ,   615 ,   2141    ,   2813    ,   9509    ,0 };
            const unsigned long dim1433JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   5   ,   15  ,   21  ,   107 ,   61  ,   259 ,   651 ,   1401    ,   3733    ,   6255    ,   5995    ,0 };
            const unsigned long dim1434JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   11  ,   21  ,   5   ,   121 ,   229 ,   497 ,   267 ,   1077    ,   3835    ,   4619    ,   8791    ,0 };
            const unsigned long dim1435JoeKuoD7Init[]   =   {   1   ,   1   ,   3   ,   3   ,   19  ,   59  ,   85  ,   115 ,   19  ,   69  ,   1937    ,   711 ,   7911    ,   2615    ,0 };
            const unsigned long dim1436JoeKuoD7Init[]   =   {   1   ,   3   ,   5   ,   11  ,   1   ,   9   ,   117 ,   229 ,   307 ,   355 ,   1635    ,   863 ,   4931    ,   5997    ,0 };
            const unsigned long dim1437JoeKuoD7Init[]   =   {   1   ,   3   ,   5   ,   11  ,   21  ,   17  ,   119 ,   163 ,   297 ,   641 ,   291 ,   2287    ,   4611    ,   4419    ,0 };
            const unsigned long dim1438JoeKuoD7Init[]   =   {   1   ,   1   ,   5   ,   7   ,   13  ,   59  ,   53  ,   17  ,   321 ,   263 ,   1929    ,   3277    ,   6209    ,   2649    ,0 };
            const unsigned long dim1439JoeKuoD7Init[]   =   {   1   ,   3   ,   5   ,   13  ,   3   ,   35  ,   35  ,   249 ,   359 ,   479 ,   1067    ,   2507    ,   4809    ,   7427    ,0 };
            const unsigned long dim1440JoeKuoD7Init[]   =   {   1   ,   3   ,   3   ,   1   ,   27  ,   57  ,   101 ,   47  ,   383 ,   937 ,   969 ,   307 ,   7617    ,   9579    ,0 };
            const unsigned long dim1441JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   13  ,   27  ,   61  ,   39  ,   177 ,   413 ,   435 ,   1529    ,   2773    ,   2267    ,   12597   ,0 };
            const unsigned long dim1442JoeKuoD7Init[]   =   {   1   ,   3   ,   3   ,   5   ,   25  ,   61  ,   57  ,   175 ,   407 ,   341 ,   1979    ,   2237    ,   3557    ,   6841    ,0 };
            const unsigned long dim1443JoeKuoD7Init[]   =   {   1   ,   3   ,   7   ,   5   ,   11  ,   23  ,   97  ,   247 ,   187 ,   555 ,   1603    ,   1871    ,   7427    ,   10605   ,0 };
            const unsigned long dim1444JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   15  ,   7   ,   21  ,   23  ,   165 ,   235 ,   937 ,   739 ,   2579    ,   6621    ,   12661   ,0 };
            const unsigned long dim1445JoeKuoD7Init[]   =   {   1   ,   1   ,   5   ,   1   ,   13  ,   47  ,   103 ,   107 ,   87  ,   941 ,   563 ,   1007    ,   7483    ,   11721   ,0 };
            const unsigned long dim1446JoeKuoD7Init[]   =   {   1   ,   1   ,   1   ,   15  ,   29  ,   63  ,   55  ,   215 ,   95  ,   991 ,   1017    ,   921 ,   8113    ,   11747   ,0 };
            const unsigned long dim1447JoeKuoD7Init[]   =   {   1   ,   1   ,   3   ,   15  ,   3   ,   25  ,   79  ,   5   ,   401 ,   473 ,   125 ,   2135    ,   6443    ,   10773   ,0 };
            const unsigned long dim1448JoeKuoD7Init[]   =   {   1   ,   3   ,   1   ,   9   ,   31  ,   21  ,   71  ,   185 ,   75  ,   43  ,   1131    ,   3885    ,   3879    ,   2695    ,0 };
            const unsigned long dim1449JoeKuoD7Init[]   =   {   1   ,   1   ,   5   ,   13  ,   13  ,   41  ,   117 ,   109 ,   429 ,   473 ,   35  ,   123 ,   5545    ,   7015    ,0 };
            const unsigned long dim1450JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   13  ,   21  ,   57  ,   55  ,   241 ,   107 ,   689 ,   707 ,   45  ,   8175    ,   15185   ,0 };
            const unsigned long dim1451JoeKuoD7Init[]   =   {   1   ,   3   ,   3   ,   7   ,   25  ,   5   ,   123 ,   173 ,   57  ,   517 ,   1605    ,   2355    ,   411 ,   14485   ,0 };
            const unsigned long dim1452JoeKuoD7Init[]   =   {   1   ,   1   ,   3   ,   13  ,   11  ,   17  ,   95  ,   129 ,   9   ,   1013    ,   1683    ,   461 ,   697 ,   8089    ,0 };
            const unsigned long dim1453JoeKuoD7Init[]   =   {   1   ,   1   ,   1   ,   3   ,   19  ,   49  ,   49  ,   247 ,   371 ,   175 ,   1803    ,   119 ,   365 ,   10281   ,0 };
            const unsigned long dim1454JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   15  ,   13  ,   57  ,   93  ,   245 ,   469 ,   933 ,   225 ,   2645    ,   6325    ,   9545    ,0 };
            const unsigned long dim1455JoeKuoD7Init[]   =   {   1   ,   3   ,   1   ,   9   ,   11  ,   37  ,   1   ,   203 ,   227 ,   809 ,   2035    ,   99  ,   2693    ,   13591   ,0 };
            const unsigned long dim1456JoeKuoD7Init[]   =   {   1   ,   3   ,   1   ,   3   ,   29  ,   7   ,   71  ,   7   ,   295 ,   203 ,   401 ,   461 ,   5351    ,   8223    ,0 };
            const unsigned long dim1457JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   1   ,   25  ,   31  ,   57  ,   35  ,   131 ,   411 ,   945 ,   1441    ,   4487    ,   2659    ,0 };
            const unsigned long dim1458JoeKuoD7Init[]   =   {   1   ,   3   ,   7   ,   15  ,   15  ,   19  ,   89  ,   253 ,   141 ,   549 ,   751 ,   1267    ,   717 ,   7185    ,0 };
            const unsigned long dim1459JoeKuoD7Init[]   =   {   1   ,   3   ,   5   ,   13  ,   1   ,   15  ,   61  ,   19  ,   331 ,   347 ,   1019    ,   1913    ,   6233    ,   8913    ,0 };
            const unsigned long dim1460JoeKuoD7Init[]   =   {   1   ,   3   ,   7   ,   15  ,   7   ,   25  ,   59  ,   75  ,   455 ,   411 ,   325 ,   553 ,   5861    ,   915 ,0 };
            const unsigned long dim1461JoeKuoD7Init[]   =   {   1   ,   1   ,   3   ,   5   ,   31  ,   49  ,   91  ,   107 ,   279 ,   405 ,   517 ,   55  ,   2121    ,   6971    ,0 };
            const unsigned long dim1462JoeKuoD7Init[]   =   {   1   ,   3   ,   5   ,   7   ,   5   ,   39  ,   29  ,   205 ,   33  ,   827 ,   1593    ,   3503    ,   5461    ,   16147   ,0 };
            const unsigned long dim1463JoeKuoD7Init[]   =   {   1   ,   1   ,   1   ,   13  ,   11  ,   13  ,   83  ,   99  ,   369 ,   283 ,   873 ,   1247    ,   6159    ,   15495   ,0 };
            const unsigned long dim1464JoeKuoD7Init[]   =   {   1   ,   3   ,   1   ,   15  ,   25  ,   61  ,   71  ,   63  ,   333 ,   183 ,   1317    ,   3467    ,   5483    ,   11591   ,0 };
            const unsigned long dim1465JoeKuoD7Init[]   =   {   1   ,   1   ,   1   ,   15  ,   27  ,   23  ,   1   ,   173 ,   407 ,   525 ,   1841    ,   347 ,   6103    ,   6449    ,0 };
            const unsigned long dim1466JoeKuoD7Init[]   =   {   1   ,   1   ,   5   ,   3   ,   9   ,   45  ,   39  ,   215 ,   329 ,   457 ,   1341    ,   3037    ,   1631    ,   15335   ,0 };
            const unsigned long dim1467JoeKuoD7Init[]   =   {   1   ,   3   ,   5   ,   3   ,   13  ,   53  ,   65  ,   177 ,   105 ,   21  ,   685 ,   3111    ,   7487    ,   1463    ,0 };
            const unsigned long dim1468JoeKuoD7Init[]   =   {   1   ,   3   ,   5   ,   5   ,   5   ,   31  ,   69  ,   65  ,   381 ,   375 ,   221 ,   477 ,   3311    ,   12103   ,0 };
            const unsigned long dim1469JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   9   ,   25  ,   59  ,   57  ,   221 ,   293 ,   237 ,   429 ,   503 ,   6539    ,   14773   ,0 };
            const unsigned long dim1470JoeKuoD7Init[]   =   {   1   ,   3   ,   7   ,   13  ,   9   ,   3   ,   47  ,   21  ,   115 ,   321 ,   389 ,   1631    ,   2715    ,   6607    ,0 };
            const unsigned long dim1471JoeKuoD7Init[]   =   {   1   ,   3   ,   7   ,   15  ,   7   ,   31  ,   105 ,   5   ,   449 ,   195 ,   65  ,   3895    ,   7093    ,   3229    ,0 };
            const unsigned long dim1472JoeKuoD7Init[]   =   {   1   ,   3   ,   7   ,   1   ,   1   ,   55  ,   119 ,   5   ,   39  ,   143 ,   905 ,   2813    ,   1723    ,   3705    ,0 };
            const unsigned long dim1473JoeKuoD7Init[]   =   {   1   ,   3   ,   7   ,   3   ,   23  ,   47  ,   61  ,   239 ,   369 ,   485 ,   1857    ,   2105    ,   2681    ,   13723   ,0 };
            const unsigned long dim1474JoeKuoD7Init[]   =   {   1   ,   3   ,   3   ,   5   ,   19  ,   5   ,   43  ,   155 ,   445 ,   597 ,   1201    ,   1625    ,   5273    ,   13569   ,0 };
            const unsigned long dim1475JoeKuoD7Init[]   =   {   1   ,   1   ,   3   ,   9   ,   3   ,   63  ,   35  ,   103 ,   67  ,   259 ,   493 ,   2847    ,   1149    ,   8367    ,0 };
            const unsigned long dim1476JoeKuoD7Init[]   =   {   1   ,   3   ,   7   ,   11  ,   19  ,   35  ,   15  ,   245 ,   59  ,   65  ,   75  ,   1297    ,   1637    ,   9081    ,0 };
            const unsigned long dim1477JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   11  ,   3   ,   23  ,   73  ,   163 ,   323 ,   857 ,   441 ,   399 ,   5447    ,   16365   ,0 };
            const unsigned long dim1478JoeKuoD7Init[]   =   {   1   ,   3   ,   1   ,   7   ,   31  ,   55  ,   125 ,   255 ,   173 ,   203 ,   1537    ,   475 ,   2955    ,   12461   ,0 };
            const unsigned long dim1479JoeKuoD7Init[]   =   {   1   ,   3   ,   5   ,   9   ,   27  ,   37  ,   69  ,   183 ,   319 ,   435 ,   807 ,   557 ,   7093    ,   2935    ,0 };
            const unsigned long dim1480JoeKuoD7Init[]   =   {   1   ,   1   ,   3   ,   5   ,   23  ,   53  ,   105 ,   211 ,   49  ,   815 ,   97  ,   915 ,   7607    ,   8847    ,0 };
            const unsigned long dim1481JoeKuoD7Init[]   =   {   1   ,   1   ,   1   ,   13  ,   13  ,   49  ,   43  ,   203 ,   293 ,   771 ,   1021    ,   535 ,   4655    ,   14727   ,0 };
            const unsigned long dim1482JoeKuoD7Init[]   =   {   1   ,   3   ,   3   ,   5   ,   13  ,   49  ,   89  ,   75  ,   227 ,   821 ,   7   ,   2539    ,   7627    ,   6301    ,0 };
            const unsigned long dim1483JoeKuoD7Init[]   =   {   1   ,   1   ,   3   ,   9   ,   5   ,   11  ,   35  ,   71  ,   187 ,   739 ,   1191    ,   2255    ,   2131    ,   13649   ,0 };
            const unsigned long dim1484JoeKuoD7Init[]   =   {   1   ,   3   ,   1   ,   5   ,   5   ,   37  ,   105 ,   29  ,   453 ,   251 ,   1111    ,   2345    ,   6297    ,   14701   ,0 };
            const unsigned long dim1485JoeKuoD7Init[]   =   {   1   ,   3   ,   1   ,   7   ,   31  ,   21  ,   5   ,   33  ,   245 ,   1007    ,   1981    ,   3775    ,   8169    ,   955 ,0 };
            const unsigned long dim1486JoeKuoD7Init[]   =   {   1   ,   1   ,   3   ,   11  ,   11  ,   51  ,   45  ,   35  ,   351 ,   479 ,   1307    ,   3009    ,   7285    ,   5261    ,0 };
            const unsigned long dim1487JoeKuoD7Init[]   =   {   1   ,   3   ,   3   ,   13  ,   5   ,   13  ,   119 ,   41  ,   175 ,   995 ,   1461    ,   2739    ,   7281    ,   7533    ,0 };
            const unsigned long dim1488JoeKuoD7Init[]   =   {   1   ,   3   ,   5   ,   9   ,   7   ,   5   ,   7   ,   153 ,   233 ,   209 ,   1235    ,   3457    ,   3113    ,   13215   ,0 };
            const unsigned long dim1489JoeKuoD7Init[]   =   {   1   ,   1   ,   3   ,   15  ,   9   ,   29  ,   41  ,   103 ,   389 ,   541 ,   1697    ,   893 ,   6507    ,   10805   ,0 };
            const unsigned long dim1490JoeKuoD7Init[]   =   {   1   ,   1   ,   3   ,   3   ,   19  ,   11  ,   73  ,   73  ,   439 ,   985 ,   1215    ,   1481    ,   617 ,   5343    ,0 };
            const unsigned long dim1491JoeKuoD7Init[]   =   {   1   ,   3   ,   7   ,   11  ,   7   ,   15  ,   81  ,   203 ,   41  ,   789 ,   283 ,   691 ,   3495    ,   12143   ,0 };
            const unsigned long dim1492JoeKuoD7Init[]   =   {   1   ,   3   ,   7   ,   9   ,   21  ,   61  ,   101 ,   185 ,   193 ,   159 ,   13  ,   2571    ,   545 ,   3043    ,0 };
            const unsigned long dim1493JoeKuoD7Init[]   =   {   1   ,   3   ,   7   ,   1   ,   11  ,   29  ,   107 ,   179 ,   69  ,   899 ,   1275    ,   1647    ,   1061    ,   12283   ,0 };
            const unsigned long dim1494JoeKuoD7Init[]   =   {   1   ,   3   ,   1   ,   7   ,   15  ,   43  ,   3   ,   207 ,   75  ,   669 ,   1239    ,   1339    ,   571 ,   5483    ,0 };
            const unsigned long dim1495JoeKuoD7Init[]   =   {   1   ,   3   ,   1   ,   1   ,   7   ,   23  ,   49  ,   183 ,   305 ,   309 ,   657 ,   1951    ,   6377    ,   4165    ,0 };
            const unsigned long dim1496JoeKuoD7Init[]   =   {   1   ,   1   ,   1   ,   3   ,   5   ,   21  ,   51  ,   177 ,   101 ,   95  ,   1331    ,   357 ,   6777    ,   13199   ,0 };
            const unsigned long dim1497JoeKuoD7Init[]   =   {   1   ,   3   ,   7   ,   3   ,   31  ,   3   ,   11  ,   179 ,   95  ,   219 ,   395 ,   603 ,   7397    ,   12045   ,0 };
            const unsigned long dim1498JoeKuoD7Init[]   =   {   1   ,   3   ,   7   ,   1   ,   21  ,   35  ,   79  ,   255 ,   443 ,   123 ,   551 ,   1113    ,   8133    ,   11621   ,0 };
            const unsigned long dim1499JoeKuoD7Init[]   =   {   1   ,   3   ,   7   ,   1   ,   11  ,   7   ,   125 ,   133 ,   429 ,   81  ,   91  ,   1135    ,   5165    ,   14953   ,0 };
            const unsigned long dim1500JoeKuoD7Init[]   =   {   1   ,   3   ,   5   ,   9   ,   25  ,   17  ,   103 ,   39  ,   167 ,   459 ,   703 ,   759 ,   1771    ,   723 ,0 };
            const unsigned long dim1501JoeKuoD7Init[]   =   {   1   ,   3   ,   5   ,   5   ,   7   ,   21  ,   113 ,   121 ,   97  ,   231 ,   59  ,   1679    ,   1409    ,   5045    ,0 };
            const unsigned long dim1502JoeKuoD7Init[]   =   {   1   ,   3   ,   1   ,   9   ,   19  ,   17  ,   117 ,   225 ,   163 ,   183 ,   89  ,   2251    ,   4777    ,   14553   ,0 };
            const unsigned long dim1503JoeKuoD7Init[]   =   {   1   ,   3   ,   1   ,   7   ,   7   ,   61  ,   25  ,   211 ,   65  ,   273 ,   1427    ,   585 ,   5907    ,   1031    ,0 };
            const unsigned long dim1504JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   3   ,   25  ,   27  ,   69  ,   113 ,   29  ,   67  ,   2025    ,   37  ,   7357    ,   14879   ,0 };
            const unsigned long dim1505JoeKuoD7Init[]   =   {   1   ,   1   ,   5   ,   11  ,   23  ,   59  ,   107 ,   179 ,   187 ,   989 ,   1043    ,   409 ,   489 ,   3289    ,0 };
            const unsigned long dim1506JoeKuoD7Init[]   =   {   1   ,   1   ,   1   ,   7   ,   5   ,   39  ,   79  ,   23  ,   443 ,   701 ,   789 ,   1723    ,   7595    ,   8003    ,0 };
            const unsigned long dim1507JoeKuoD7Init[]   =   {   1   ,   3   ,   1   ,   11  ,   13  ,   25  ,   29  ,   167 ,   143 ,   735 ,   1623    ,   2005    ,   3731    ,   1413    ,0 };
            const unsigned long dim1508JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   13  ,   23  ,   19  ,   3   ,   143 ,   69  ,   331 ,   71  ,   3233    ,   4187    ,   13927   ,0 };
            const unsigned long dim1509JoeKuoD7Init[]   =   {   1   ,   3   ,   3   ,   5   ,   23  ,   21  ,   121 ,   47  ,   291 ,   883 ,   43  ,   419 ,   297 ,   8385    ,0 };
            const unsigned long dim1510JoeKuoD7Init[]   =   {   1   ,   1   ,   3   ,   5   ,   19  ,   63  ,   57  ,   127 ,   59  ,   539 ,   207 ,   3805    ,   4741    ,   12519   ,0 };
            const unsigned long dim1511JoeKuoD7Init[]   =   {   1   ,   3   ,   3   ,   5   ,   5   ,   23  ,   117 ,   45  ,   501 ,   903 ,   1241    ,   3853    ,   989 ,   7997    ,0 };
            const unsigned long dim1512JoeKuoD7Init[]   =   {   1   ,   1   ,   3   ,   5   ,   11  ,   27  ,   17  ,   21  ,   373 ,   755 ,   1149    ,   3443    ,   3489    ,   11701   ,0 };
            const unsigned long dim1513JoeKuoD7Init[]   =   {   1   ,   1   ,   5   ,   7   ,   7   ,   55  ,   75  ,   29  ,   241 ,   707 ,   555 ,   1683    ,   1311    ,   14769   ,0 };
            const unsigned long dim1514JoeKuoD7Init[]   =   {   1   ,   3   ,   5   ,   13  ,   19  ,   55  ,   87  ,   145 ,   27  ,   935 ,   519 ,   1083    ,   1115    ,   10963   ,0 };
            const unsigned long dim1515JoeKuoD7Init[]   =   {   1   ,   1   ,   3   ,   13  ,   5   ,   11  ,   25  ,   215 ,   483 ,   817 ,   101 ,   981 ,   5125    ,   15257   ,0 };
            const unsigned long dim1516JoeKuoD7Init[]   =   {   1   ,   3   ,   1   ,   11  ,   21  ,   59  ,   93  ,   55  ,   447 ,   5   ,   1225    ,   827 ,   1539    ,   4035    ,0 };
            const unsigned long dim1517JoeKuoD7Init[]   =   {   1   ,   1   ,   1   ,   13  ,   3   ,   53  ,   5   ,   39  ,   475 ,   475 ,   905 ,   355 ,   6319    ,   415 ,0 };
            const unsigned long dim1518JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   9   ,   21  ,   49  ,   69  ,   109 ,   3   ,   441 ,   1533    ,   2359    ,   4025    ,   15949   ,0 };
            const unsigned long dim1519JoeKuoD7Init[]   =   {   1   ,   1   ,   3   ,   13  ,   3   ,   59  ,   67  ,   147 ,   151 ,   27  ,   1039    ,   2331    ,   2839    ,   15939   ,0 };
            const unsigned long dim1520JoeKuoD7Init[]   =   {   1   ,   3   ,   7   ,   9   ,   7   ,   17  ,   13  ,   153 ,   415 ,   147 ,   181 ,   3441    ,   7997    ,   3175    ,0 };
            const unsigned long dim1521JoeKuoD7Init[]   =   {   1   ,   3   ,   5   ,   13  ,   9   ,   61  ,   123 ,   177 ,   497 ,   989 ,   525 ,   1819    ,   5963    ,   2051    ,0 };
            const unsigned long dim1522JoeKuoD7Init[]   =   {   1   ,   1   ,   1   ,   1   ,   11  ,   13  ,   17  ,   147 ,   57  ,   367 ,   477 ,   1009    ,   4613    ,   16357   ,0 };
            const unsigned long dim1523JoeKuoD7Init[]   =   {   1   ,   3   ,   7   ,   9   ,   27  ,   27  ,   29  ,   61  ,   499 ,   191 ,   43  ,   1109    ,   5805    ,   12799   ,0 };
            const unsigned long dim1524JoeKuoD7Init[]   =   {   1   ,   3   ,   5   ,   9   ,   11  ,   61  ,   57  ,   139 ,   173 ,   725 ,   1879    ,   1861    ,   7123    ,   9505    ,0 };
            const unsigned long dim1525JoeKuoD7Init[]   =   {   1   ,   3   ,   1   ,   5   ,   11  ,   1   ,   93  ,   179 ,   3   ,   893 ,   1917    ,   2293    ,   5105    ,   12093   ,0 };
            const unsigned long dim1526JoeKuoD7Init[]   =   {   1   ,   1   ,   5   ,   9   ,   31  ,   45  ,   107 ,   19  ,   285 ,   81  ,   339 ,   3051    ,   1369    ,   10841   ,0 };
            const unsigned long dim1527JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   9   ,   27  ,   49  ,   19  ,   29  ,   155 ,   389 ,   1287    ,   3   ,   2311    ,   7561    ,0 };
            const unsigned long dim1528JoeKuoD7Init[]   =   {   1   ,   3   ,   3   ,   9   ,   9   ,   1   ,   45  ,   171 ,   245 ,   897 ,   1881    ,   2683    ,   4185    ,   14049   ,0 };
            const unsigned long dim1529JoeKuoD7Init[]   =   {   1   ,   3   ,   1   ,   15  ,   19  ,   37  ,   19  ,   227 ,   133 ,   573 ,   1897    ,   1231    ,   1649    ,   8481    ,0 };
            const unsigned long dim1530JoeKuoD7Init[]   =   {   1   ,   3   ,   1   ,   15  ,   5   ,   19  ,   27  ,   85  ,   281 ,   471 ,   3   ,   2149    ,   4785    ,   12391   ,0 };
            const unsigned long dim1531JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   3   ,   17  ,   5   ,   127 ,   37  ,   299 ,   467 ,   1523    ,   3421    ,   4119    ,   11151   ,0 };
            const unsigned long dim1532JoeKuoD7Init[]   =   {   1   ,   3   ,   3   ,   13  ,   11  ,   5   ,   15  ,   137 ,   459 ,   553 ,   785 ,   1149    ,   6107    ,   321 ,0 };
            const unsigned long dim1533JoeKuoD7Init[]   =   {   1   ,   1   ,   3   ,   3   ,   7   ,   19  ,   85  ,   133 ,   7   ,   753 ,   65  ,   3341    ,   5859    ,   3829    ,0 };
            const unsigned long dim1534JoeKuoD7Init[]   =   {   1   ,   3   ,   3   ,   11  ,   19  ,   13  ,   17  ,   191 ,   121 ,   895 ,   761 ,   2431    ,   5887    ,   10961   ,0 };
            const unsigned long dim1535JoeKuoD7Init[]   =   {   1   ,   3   ,   5   ,   7   ,   1   ,   63  ,   47  ,   43  ,   119 ,   311 ,   935 ,   2299    ,   1509    ,   4821    ,0 };
            const unsigned long dim1536JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   7   ,   27  ,   51  ,   93  ,   195 ,   397 ,   155 ,   471 ,   2547    ,   8007    ,   15575   ,0 };
            const unsigned long dim1537JoeKuoD7Init[]   =   {   1   ,   3   ,   3   ,   5   ,   25  ,   13  ,   125 ,   239 ,   93  ,   983 ,   293 ,   1317    ,   115 ,   6417    ,0 };
            const unsigned long dim1538JoeKuoD7Init[]   =   {   1   ,   3   ,   3   ,   3   ,   9   ,   61  ,   93  ,   15  ,   495 ,   295 ,   1391    ,   2833    ,   45  ,   9039    ,0 };
            const unsigned long dim1539JoeKuoD7Init[]   =   {   1   ,   1   ,   1   ,   9   ,   11  ,   31  ,   125 ,   5   ,   65  ,   173 ,   1189    ,   2363    ,   6603    ,   15891   ,0 };
            const unsigned long dim1540JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   5   ,   5   ,   63  ,   53  ,   35  ,   311 ,   935 ,   1591    ,   309 ,   7829    ,   1495    ,0 };
            const unsigned long dim1541JoeKuoD7Init[]   =   {   1   ,   3   ,   5   ,   5   ,   29  ,   21  ,   121 ,   199 ,   153 ,   933 ,   1577    ,   1681    ,   1883    ,   7697    ,0 };
            const unsigned long dim1542JoeKuoD7Init[]   =   {   1   ,   1   ,   1   ,   11  ,   19  ,   23  ,   111 ,   113 ,   73  ,   99  ,   1385    ,   591 ,   6977    ,   11995   ,0 };
            const unsigned long dim1543JoeKuoD7Init[]   =   {   1   ,   1   ,   1   ,   13  ,   27  ,   11  ,   23  ,   79  ,   429 ,   505 ,   1693    ,   919 ,   3699    ,   3815    ,0 };
            const unsigned long dim1544JoeKuoD7Init[]   =   {   1   ,   1   ,   3   ,   9   ,   7   ,   23  ,   29  ,   197 ,   465 ,   267 ,   915 ,   1455    ,   2155    ,   15105   ,0 };
            const unsigned long dim1545JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   9   ,   17  ,   3   ,   7   ,   73  ,   417 ,   841 ,   1123    ,   3425    ,   4691    ,   14703   ,0 };
            const unsigned long dim1546JoeKuoD7Init[]   =   {   1   ,   3   ,   5   ,   15  ,   15  ,   7   ,   75  ,   85  ,   441 ,   5   ,   1593    ,   43  ,   1207    ,   12117   ,0 };
            const unsigned long dim1547JoeKuoD7Init[]   =   {   1   ,   1   ,   1   ,   7   ,   21  ,   45  ,   93  ,   133 ,   31  ,   249 ,   701 ,   97  ,   1033    ,   15359   ,0 };
            const unsigned long dim1548JoeKuoD7Init[]   =   {   1   ,   1   ,   3   ,   7   ,   3   ,   5   ,   9   ,   21  ,   339 ,   661 ,   1841    ,   2209    ,   5869    ,   10291   ,0 };
            const unsigned long dim1549JoeKuoD7Init[]   =   {   1   ,   1   ,   5   ,   11  ,   13  ,   3   ,   67  ,   135 ,   69  ,   117 ,   43  ,   2885    ,   5719    ,   427 ,0 };
            const unsigned long dim1550JoeKuoD7Init[]   =   {   1   ,   3   ,   1   ,   9   ,   15  ,   55  ,   45  ,   85  ,   423 ,   351 ,   169 ,   2693    ,   3063    ,   13471   ,0 };
            const unsigned long dim1551JoeKuoD7Init[]   =   {   1   ,   1   ,   1   ,   7   ,   21  ,   17  ,   51  ,   59  ,   251 ,   697 ,   787 ,   3331    ,   7567    ,   15413   ,0 };
            const unsigned long dim1552JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   7   ,   29  ,   29  ,   113 ,   37  ,   347 ,   21  ,   223 ,   427 ,   4163    ,   6773    ,0 };
            const unsigned long dim1553JoeKuoD7Init[]   =   {   1   ,   3   ,   5   ,   9   ,   25  ,   55  ,   89  ,   101 ,   327 ,   483 ,   1197    ,   1377    ,   3367    ,   15363   ,0 };
            const unsigned long dim1554JoeKuoD7Init[]   =   {   1   ,   3   ,   3   ,   5   ,   15  ,   37  ,   87  ,   131 ,   279 ,   981 ,   623 ,   105 ,   3537    ,   6021    ,0 };
            const unsigned long dim1555JoeKuoD7Init[]   =   {   1   ,   3   ,   5   ,   13  ,   19  ,   27  ,   49  ,   187 ,   249 ,   17  ,   1499    ,   3559    ,   1691    ,   525 ,0 };
            const unsigned long dim1556JoeKuoD7Init[]   =   {   1   ,   1   ,   3   ,   11  ,   25  ,   55  ,   7   ,   145 ,   335 ,   975 ,   1007    ,   1853    ,   2509    ,   12895   ,0 };
            const unsigned long dim1557JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   15  ,   31  ,   29  ,   43  ,   91  ,   51  ,   543 ,   251 ,   3137    ,   1449    ,   2465    ,0 };
            const unsigned long dim1558JoeKuoD7Init[]   =   {   1   ,   3   ,   7   ,   1   ,   25  ,   33  ,   107 ,   117 ,   439 ,   851 ,   1853    ,   1309    ,   7329    ,   6953    ,0 };
            const unsigned long dim1559JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   9   ,   13  ,   13  ,   105 ,   113 ,   161 ,   307 ,   939 ,   579 ,   6945    ,   3275    ,0 };
            const unsigned long dim1560JoeKuoD7Init[]   =   {   1   ,   3   ,   3   ,   3   ,   19  ,   61  ,   85  ,   247 ,   31  ,   781 ,   599 ,   1731    ,   7907    ,   10855   ,0 };
            const unsigned long dim1561JoeKuoD7Init[]   =   {   1   ,   3   ,   7   ,   15  ,   23  ,   13  ,   111 ,   197 ,   399 ,   273 ,   1339    ,   717 ,   6365    ,   11211   ,0 };
            const unsigned long dim1562JoeKuoD7Init[]   =   {   1   ,   1   ,   1   ,   1   ,   7   ,   37  ,   83  ,   217 ,   351 ,   325 ,   787 ,   2735    ,   6797    ,   649 ,0 };
            const unsigned long dim1563JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   5   ,   25  ,   9   ,   29  ,   3   ,   305 ,   773 ,   933 ,   4091    ,   7515    ,   13513   ,0 };
            const unsigned long dim1564JoeKuoD7Init[]   =   {   1   ,   1   ,   1   ,   7   ,   5   ,   9   ,   53  ,   79  ,   509 ,   101 ,   187 ,   4087    ,   5473    ,   8581    ,0 };
            const unsigned long dim1565JoeKuoD7Init[]   =   {   1   ,   1   ,   3   ,   3   ,   9   ,   7   ,   59  ,   235 ,   219 ,   85  ,   1275    ,   329 ,   3661    ,   2135    ,0 };
            const unsigned long dim1566JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   15  ,   7   ,   43  ,   39  ,   153 ,   59  ,   91  ,   755 ,   827 ,   2399    ,   5395    ,0 };
            const unsigned long dim1567JoeKuoD7Init[]   =   {   1   ,   3   ,   7   ,   3   ,   11  ,   29  ,   75  ,   195 ,   249 ,   789 ,   489 ,   1297    ,   1465    ,   449 ,0 };
            const unsigned long dim1568JoeKuoD7Init[]   =   {   1   ,   3   ,   5   ,   13  ,   29  ,   13  ,   61  ,   111 ,   85  ,   335 ,   1835    ,   4025    ,   3433    ,   12839   ,0 };
            const unsigned long dim1569JoeKuoD7Init[]   =   {   1   ,   1   ,   3   ,   1   ,   9   ,   45  ,   13  ,   179 ,   145 ,   161 ,   459 ,   3011    ,   7809    ,   11047   ,0 };
            const unsigned long dim1570JoeKuoD7Init[]   =   {   1   ,   1   ,   5   ,   15  ,   27  ,   1   ,   81  ,   241 ,   89  ,   159 ,   681 ,   3525    ,   4487    ,   9129    ,0 };
            const unsigned long dim1571JoeKuoD7Init[]   =   {   1   ,   3   ,   7   ,   15  ,   7   ,   1   ,   15  ,   103 ,   81  ,   59  ,   1953    ,   3729    ,   6451    ,   15051   ,0 };
            const unsigned long dim1572JoeKuoD7Init[]   =   {   1   ,   3   ,   1   ,   1   ,   29  ,   7   ,   77  ,   77  ,   279 ,   1007    ,   1315    ,   1149    ,   301 ,   14183   ,0 };
            const unsigned long dim1573JoeKuoD7Init[]   =   {   1   ,   1   ,   5   ,   11  ,   1   ,   1   ,   85  ,   23  ,   27  ,   211 ,   2023    ,   591 ,   2949    ,   10741   ,0 };
            const unsigned long dim1574JoeKuoD7Init[]   =   {   1   ,   3   ,   1   ,   11  ,   31  ,   63  ,   49  ,   87  ,   291 ,   701 ,   531 ,   287 ,   6117    ,   16303   ,0 };
            const unsigned long dim1575JoeKuoD7Init[]   =   {   1   ,   3   ,   3   ,   9   ,   27  ,   19  ,   19  ,   143 ,   95  ,   65  ,   1027    ,   2765    ,   2483    ,   14347   ,0 };
            const unsigned long dim1576JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   9   ,   3   ,   31  ,   31  ,   243 ,   45  ,   195 ,   1997    ,   3361    ,   7151    ,   12021   ,0 };
            const unsigned long dim1577JoeKuoD7Init[]   =   {   1   ,   3   ,   3   ,   9   ,   11  ,   43  ,   83  ,   1   ,   389 ,   843 ,   1635    ,   3405    ,   2497    ,   4473    ,0 };
            const unsigned long dim1578JoeKuoD7Init[]   =   {   1   ,   3   ,   5   ,   5   ,   15  ,   9   ,   65  ,   187 ,   107 ,   271 ,   1049    ,   1657    ,   1831    ,   12291   ,0 };
            const unsigned long dim1579JoeKuoD7Init[]   =   {   1   ,   3   ,   7   ,   3   ,   29  ,   17  ,   107 ,   17  ,   253 ,   399 ,   481 ,   83  ,   3905    ,   5817    ,0 };
            const unsigned long dim1580JoeKuoD7Init[]   =   {   1   ,   3   ,   5   ,   5   ,   21  ,   1   ,   93  ,   189 ,   451 ,   183 ,   1489    ,   379 ,   4673    ,   9323    ,0 };
            const unsigned long dim1581JoeKuoD7Init[]   =   {   1   ,   3   ,   1   ,   9   ,   19  ,   25  ,   105 ,   97  ,   505 ,   569 ,   773 ,   3155    ,   3275    ,   11389   ,0 };
            const unsigned long dim1582JoeKuoD7Init[]   =   {   1   ,   3   ,   3   ,   11  ,   29  ,   25  ,   87  ,   205 ,   321 ,   187 ,   239 ,   2217    ,   1063    ,   13323   ,0 };
            const unsigned long dim1583JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   13  ,   19  ,   47  ,   17  ,   243 ,   103 ,   515 ,   1049    ,   2119    ,   2185    ,   3137    ,0 };
            const unsigned long dim1584JoeKuoD7Init[]   =   {   1   ,   3   ,   1   ,   15  ,   17  ,   19  ,   45  ,   191 ,   257 ,   453 ,   779 ,   3759    ,   6825    ,   5711    ,0 };
            const unsigned long dim1585JoeKuoD7Init[]   =   {   1   ,   3   ,   3   ,   15  ,   15  ,   41  ,   41  ,   181 ,   185 ,   91  ,   375 ,   2825    ,   3841    ,   14179   ,0 };
            const unsigned long dim1586JoeKuoD7Init[]   =   {   1   ,   1   ,   5   ,   13  ,   23  ,   11  ,   93  ,   117 ,   135 ,   333 ,   1985    ,   3355    ,   2119    ,   10511   ,0 };
            const unsigned long dim1587JoeKuoD7Init[]   =   {   1   ,   3   ,   7   ,   15  ,   5   ,   59  ,   17  ,   141 ,   291 ,   897 ,   657 ,   2805    ,   4423    ,   559 ,0 };
            const unsigned long dim1588JoeKuoD7Init[]   =   {   1   ,   3   ,   7   ,   1   ,   19  ,   55  ,   33  ,   77  ,   321 ,   989 ,   1713    ,   1553    ,   5541    ,   12995   ,0 };
            const unsigned long dim1589JoeKuoD7Init[]   =   {   1   ,   1   ,   3   ,   5   ,   11  ,   9   ,   69  ,   123 ,   223 ,   201 ,   1505    ,   1119    ,   381 ,   12361   ,0 };
            const unsigned long dim1590JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   7   ,   13  ,   13  ,   15  ,   107 ,   143 ,   925 ,   1741    ,   3697    ,   5379    ,   10661   ,0 };
            const unsigned long dim1591JoeKuoD7Init[]   =   {   1   ,   1   ,   5   ,   1   ,   25  ,   43  ,   47  ,   209 ,   57  ,   1015    ,   1721    ,   1707    ,   3675    ,   2651    ,0 };
            const unsigned long dim1592JoeKuoD7Init[]   =   {   1   ,   3   ,   3   ,   3   ,   21  ,   27  ,   101 ,   41  ,   309 ,   585 ,   159 ,   2029    ,   5307    ,   5919    ,0 };
            const unsigned long dim1593JoeKuoD7Init[]   =   {   1   ,   1   ,   5   ,   7   ,   13  ,   53  ,   43  ,   71  ,   259 ,   331 ,   1317    ,   3367    ,   2593    ,   4667    ,0 };
            const unsigned long dim1594JoeKuoD7Init[]   =   {   1   ,   3   ,   7   ,   13  ,   1   ,   31  ,   115 ,   159 ,   203 ,   39  ,   19  ,   1939    ,   2601    ,   14819   ,0 };
            const unsigned long dim1595JoeKuoD7Init[]   =   {   1   ,   3   ,   7   ,   9   ,   1   ,   33  ,   101 ,   145 ,   401 ,   655 ,   543 ,   2063    ,   425 ,   6397    ,0 };
            const unsigned long dim1596JoeKuoD7Init[]   =   {   1   ,   3   ,   3   ,   1   ,   1   ,   13  ,   15  ,   19  ,   347 ,   679 ,   107 ,   933 ,   8097    ,   16213   ,0 };
            const unsigned long dim1597JoeKuoD7Init[]   =   {   1   ,   3   ,   1   ,   5   ,   9   ,   47  ,   15  ,   95  ,   191 ,   887 ,   69  ,   3775    ,   7839    ,   767 ,0 };
            const unsigned long dim1598JoeKuoD7Init[]   =   {   1   ,   1   ,   5   ,   13  ,   17  ,   37  ,   43  ,   241 ,   401 ,   273 ,   1819    ,   2437    ,   2027    ,   5763    ,0 };
            const unsigned long dim1599JoeKuoD7Init[]   =   {   1   ,   1   ,   3   ,   11  ,   25  ,   49  ,   31  ,   17  ,   153 ,   771 ,   1151    ,   1575    ,   1719    ,   12567   ,0 };
            const unsigned long dim1600JoeKuoD7Init[]   =   {   1   ,   3   ,   7   ,   9   ,   19  ,   31  ,   33  ,   227 ,   15  ,   645 ,   1499    ,   967 ,   6459    ,   975 ,0 };
            const unsigned long dim1601JoeKuoD7Init[]   =   {   1   ,   3   ,   1   ,   1   ,   5   ,   23  ,   31  ,   105 ,   47  ,   761 ,   1911    ,   3889    ,   67  ,   5543    ,0 };
            const unsigned long dim1602JoeKuoD7Init[]   =   {   1   ,   1   ,   5   ,   3   ,   15  ,   1   ,   7   ,   119 ,   247 ,   759 ,   1277    ,   1233    ,   1055    ,   15651   ,0 };
            const unsigned long dim1603JoeKuoD7Init[]   =   {   1   ,   1   ,   1   ,   1   ,   15  ,   33  ,   117 ,   189 ,   93  ,   511 ,   1709    ,   3329    ,   619 ,   3561    ,0 };
            const unsigned long dim1604JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   9   ,   21  ,   39  ,   35  ,   59  ,   121 ,   149 ,   2015    ,   1127    ,   8035    ,   8197    ,0 };
            const unsigned long dim1605JoeKuoD7Init[]   =   {   1   ,   3   ,   7   ,   9   ,   25  ,   3   ,   43  ,   189 ,   257 ,   343 ,   607 ,   285 ,   6003    ,   1119    ,0 };
            const unsigned long dim1606JoeKuoD7Init[]   =   {   1   ,   1   ,   5   ,   11  ,   7   ,   41  ,   53  ,   193 ,   343 ,   353 ,   1743    ,   1877    ,   2207    ,   10765   ,0 };
            const unsigned long dim1607JoeKuoD7Init[]   =   {   1   ,   1   ,   3   ,   1   ,   3   ,   59  ,   117 ,   15  ,   199 ,   639 ,   1597    ,   3833    ,   7167    ,   9437    ,0 };
            const unsigned long dim1608JoeKuoD7Init[]   =   {   1   ,   3   ,   7   ,   13  ,   15  ,   49  ,   79  ,   19  ,   131 ,   1021    ,   1881    ,   1113    ,   6191    ,   7549    ,0 };
            const unsigned long dim1609JoeKuoD7Init[]   =   {   1   ,   1   ,   5   ,   5   ,   29  ,   57  ,   69  ,   181 ,   167 ,   179 ,   317 ,   4069    ,   3885    ,   2641    ,0 };
            const unsigned long dim1610JoeKuoD7Init[]   =   {   1   ,   3   ,   3   ,   7   ,   1   ,   15  ,   13  ,   69  ,   425 ,   491 ,   327 ,   1387    ,   5333    ,   14555   ,0 };
            const unsigned long dim1611JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   3   ,   3   ,   33  ,   99  ,   169 ,   245 ,   155 ,   1843    ,   2435    ,   2853    ,   8871    ,0 };
            const unsigned long dim1612JoeKuoD7Init[]   =   {   1   ,   1   ,   1   ,   11  ,   27  ,   37  ,   107 ,   163 ,   399 ,   933 ,   1867    ,   3321    ,   2737    ,   4941    ,0 };
            const unsigned long dim1613JoeKuoD7Init[]   =   {   1   ,   3   ,   5   ,   11  ,   27  ,   25  ,   31  ,   193 ,   219 ,   21  ,   1811    ,   2515    ,   1027    ,   10337   ,0 };
            const unsigned long dim1614JoeKuoD7Init[]   =   {   1   ,   3   ,   7   ,   1   ,   13  ,   23  ,   75  ,   175 ,   265 ,   71  ,   597 ,   521 ,   5343    ,   5615    ,0 };
            const unsigned long dim1615JoeKuoD7Init[]   =   {   1   ,   3   ,   1   ,   11  ,   27  ,   13  ,   23  ,   187 ,   281 ,   593 ,   1559    ,   1355    ,   637 ,   15835   ,0 };
            const unsigned long dim1616JoeKuoD7Init[]   =   {   1   ,   3   ,   5   ,   15  ,   25  ,   7   ,   29  ,   1   ,   403 ,   175 ,   401 ,   3521    ,   1077    ,   8699    ,0 };
            const unsigned long dim1617JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   5   ,   31  ,   27  ,   79  ,   135 ,   439 ,   935 ,   783 ,   2981    ,   3747    ,   10697   ,0 };
            const unsigned long dim1618JoeKuoD7Init[]   =   {   1   ,   3   ,   1   ,   3   ,   25  ,   1   ,   119 ,   53  ,   107 ,   841 ,   909 ,   265 ,   3845    ,   8613    ,0 };
            const unsigned long dim1619JoeKuoD7Init[]   =   {   1   ,   3   ,   3   ,   7   ,   13  ,   53  ,   3   ,   159 ,   453 ,   711 ,   399 ,   3267    ,   1245    ,   8747    ,0 };
            const unsigned long dim1620JoeKuoD7Init[]   =   {   1   ,   1   ,   1   ,   9   ,   15  ,   59  ,   99  ,   255 ,   331 ,   241 ,   1389    ,   3885    ,   7497    ,   13169   ,0 };
            const unsigned long dim1621JoeKuoD7Init[]   =   {   1   ,   3   ,   5   ,   9   ,   11  ,   23  ,   49  ,   77  ,   239 ,   341 ,   453 ,   763 ,   3241    ,   13097   ,0 };
            const unsigned long dim1622JoeKuoD7Init[]   =   {   1   ,   3   ,   7   ,   15  ,   21  ,   15  ,   65  ,   221 ,   167 ,   17  ,   1621    ,   4043    ,   1643    ,   13481   ,0 };
            const unsigned long dim1623JoeKuoD7Init[]   =   {   1   ,   3   ,   1   ,   9   ,   1   ,   35  ,   125 ,   157 ,   437 ,   49  ,   1145    ,   861 ,   4323    ,   11839   ,0 };
            const unsigned long dim1624JoeKuoD7Init[]   =   {   1   ,   1   ,   3   ,   5   ,   27  ,   27  ,   125 ,   61  ,   401 ,   251 ,   1171    ,   2409    ,   63  ,   11279   ,0 };
            const unsigned long dim1625JoeKuoD7Init[]   =   {   1   ,   3   ,   3   ,   5   ,   27  ,   57  ,   39  ,   37  ,   337 ,   999 ,   549 ,   2071    ,   5547    ,   315 ,0 };
            const unsigned long dim1626JoeKuoD7Init[]   =   {   1   ,   1   ,   1   ,   15  ,   29  ,   29  ,   119 ,   89  ,   325 ,   665 ,   783 ,   1683    ,   6185    ,   4301    ,0 };
            const unsigned long dim1627JoeKuoD7Init[]   =   {   1   ,   3   ,   7   ,   7   ,   7   ,   25  ,   111 ,   99  ,   411 ,   549 ,   949 ,   735 ,   6881    ,   4681    ,0 };
            const unsigned long dim1628JoeKuoD7Init[]   =   {   1   ,   3   ,   5   ,   15  ,   21  ,   3   ,   117 ,   83  ,   473 ,   595 ,   1061    ,   3905    ,   4985    ,   6149    ,0 };
            const unsigned long dim1629JoeKuoD7Init[]   =   {   1   ,   3   ,   7   ,   1   ,   17  ,   13  ,   27  ,   101 ,   95  ,   773 ,   1095    ,   2813    ,   6253    ,   2817    ,0 };
            const unsigned long dim1630JoeKuoD7Init[]   =   {   1   ,   1   ,   5   ,   9   ,   1   ,   9   ,   91  ,   43  ,   153 ,   627 ,   49  ,   1109    ,   1395    ,   2971    ,0 };
            const unsigned long dim1631JoeKuoD7Init[]   =   {   1   ,   3   ,   7   ,   11  ,   25  ,   57  ,   1   ,   141 ,   227 ,   161 ,   3   ,   2509    ,   2675    ,   14093   ,0 };
            const unsigned long dim1632JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   7   ,   27  ,   35  ,   85  ,   103 ,   321 ,   331 ,   351 ,   2831    ,   911 ,   13259   ,0 };
            const unsigned long dim1633JoeKuoD7Init[]   =   {   1   ,   1   ,   1   ,   9   ,   29  ,   59  ,   57  ,   199 ,   27  ,   303 ,   1001    ,   1393    ,   3345    ,   11169   ,0 };
            const unsigned long dim1634JoeKuoD7Init[]   =   {   1   ,   3   ,   5   ,   15  ,   11  ,   37  ,   83  ,   127 ,   79  ,   683 ,   669 ,   1793    ,   2271    ,   1643    ,0 };
            const unsigned long dim1635JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   5   ,   9   ,   45  ,   65  ,   143 ,   489 ,   91  ,   1819    ,   2679    ,   4003    ,   13961   ,0 };
            const unsigned long dim1636JoeKuoD7Init[]   =   {   1   ,   3   ,   1   ,   1   ,   5   ,   9   ,   127 ,   61  ,   209 ,   597 ,   463 ,   961 ,   1739    ,   13955   ,0 };
            const unsigned long dim1637JoeKuoD7Init[]   =   {   1   ,   3   ,   5   ,   11  ,   29  ,   13  ,   73  ,   61  ,   233 ,   891 ,   1477    ,   501 ,   7781    ,   15911   ,0 };
            const unsigned long dim1638JoeKuoD7Init[]   =   {   1   ,   3   ,   5   ,   3   ,   23  ,   23  ,   99  ,   227 ,   385 ,   503 ,   405 ,   3825    ,   535 ,   5997    ,0 };
            const unsigned long dim1639JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   7   ,   11  ,   31  ,   105 ,   103 ,   383 ,   897 ,   1049    ,   2757    ,   7687    ,   13175   ,0 };
            const unsigned long dim1640JoeKuoD7Init[]   =   {   1   ,   1   ,   3   ,   15  ,   7   ,   55  ,   43  ,   205 ,   451 ,   195 ,   1973    ,   1963    ,   5791    ,   14413   ,0 };
            const unsigned long dim1641JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   13  ,   7   ,   5   ,   35  ,   215 ,   415 ,   221 ,   1245    ,   399 ,   1471    ,   9559    ,0 };
            const unsigned long dim1642JoeKuoD7Init[]   =   {   1   ,   3   ,   1   ,   13  ,   3   ,   27  ,   119 ,   245 ,   291 ,   633 ,   641 ,   779 ,   533 ,   12031   ,0 };
            const unsigned long dim1643JoeKuoD7Init[]   =   {   1   ,   1   ,   1   ,   11  ,   21  ,   41  ,   83  ,   93  ,   91  ,   671 ,   1579    ,   3375    ,   7109    ,   12321   ,0 };
            const unsigned long dim1644JoeKuoD7Init[]   =   {   1   ,   1   ,   5   ,   1   ,   27  ,   47  ,   19  ,   161 ,   103 ,   455 ,   2027    ,   1461    ,   633 ,   5815    ,0 };
            const unsigned long dim1645JoeKuoD7Init[]   =   {   1   ,   3   ,   5   ,   11  ,   25  ,   11  ,   115 ,   95  ,   15  ,   979 ,   89  ,   2489    ,   255 ,   13683   ,0 };
            const unsigned long dim1646JoeKuoD7Init[]   =   {   1   ,   1   ,   5   ,   5   ,   31  ,   55  ,   85  ,   223 ,   231 ,   9   ,   493 ,   3117    ,   3431    ,   12139   ,0 };
            const unsigned long dim1647JoeKuoD7Init[]   =   {   1   ,   1   ,   1   ,   13  ,   1   ,   59  ,   5   ,   33  ,   111 ,   711 ,   1111    ,   3337    ,   6595    ,   15149   ,0 };
            const unsigned long dim1648JoeKuoD7Init[]   =   {   1   ,   3   ,   3   ,   5   ,   27  ,   9   ,   85  ,   243 ,   141 ,   155 ,   2007    ,   673 ,   3613    ,   9837    ,0 };
            const unsigned long dim1649JoeKuoD7Init[]   =   {   1   ,   3   ,   1   ,   13  ,   17  ,   53  ,   105 ,   125 ,   295 ,   467 ,   29  ,   1501    ,   2841    ,   1531    ,0 };
            const unsigned long dim1650JoeKuoD7Init[]   =   {   1   ,   3   ,   1   ,   1   ,   15  ,   35  ,   95  ,   73  ,   273 ,   5   ,   515 ,   185 ,   5845    ,   15793   ,0 };
            const unsigned long dim1651JoeKuoD7Init[]   =   {   1   ,   3   ,   7   ,   15  ,   1   ,   9   ,   117 ,   229 ,   219 ,   723 ,   1641    ,   2639    ,   5711    ,   10337   ,0 };
            const unsigned long dim1652JoeKuoD7Init[]   =   {   1   ,   1   ,   5   ,   1   ,   9   ,   47  ,   13  ,   179 ,   353 ,   167 ,   1255    ,   4053    ,   1279    ,   6181    ,0 };
            const unsigned long dim1653JoeKuoD7Init[]   =   {   1   ,   1   ,   5   ,   11  ,   3   ,   29  ,   47  ,   63  ,   127 ,   355 ,   1459    ,   441 ,   6849    ,   12995   ,0 };
            const unsigned long dim1654JoeKuoD7Init[]   =   {   1   ,   1   ,   5   ,   5   ,   15  ,   27  ,   99  ,   11  ,   63  ,   405 ,   1025    ,   2047    ,   7239    ,   13019   ,0 };
            const unsigned long dim1655JoeKuoD7Init[]   =   {   1   ,   3   ,   3   ,   15  ,   31  ,   1   ,   17  ,   109 ,   7   ,   251 ,   517 ,   2923    ,   2967    ,   15609   ,0 };
            const unsigned long dim1656JoeKuoD7Init[]   =   {   1   ,   3   ,   5   ,   5   ,   5   ,   9   ,   89  ,   77  ,   77  ,   619 ,   793 ,   195 ,   3507    ,   4203    ,0 };
            const unsigned long dim1657JoeKuoD7Init[]   =   {   1   ,   3   ,   1   ,   1   ,   29  ,   47  ,   83  ,   117 ,   387 ,   601 ,   35  ,   1327    ,   6381    ,   7673    ,0 };
            const unsigned long dim1658JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   3   ,   21  ,   7   ,   33  ,   41  ,   161 ,   53  ,   1635    ,   787 ,   6197    ,   4841    ,0 };
            const unsigned long dim1659JoeKuoD7Init[]   =   {   1   ,   3   ,   7   ,   11  ,   5   ,   33  ,   29  ,   43  ,   289 ,   805 ,   109 ,   2099    ,   7851    ,   809 ,0 };
            const unsigned long dim1660JoeKuoD7Init[]   =   {   1   ,   3   ,   7   ,   7   ,   7   ,   39  ,   79  ,   147 ,   129 ,   307 ,   93  ,   3349    ,   7329    ,   14213   ,0 };
            const unsigned long dim1661JoeKuoD7Init[]   =   {   1   ,   3   ,   5   ,   13  ,   27  ,   59  ,   35  ,   235 ,   413 ,   631 ,   1201    ,   1317    ,   5489    ,   13277   ,0 };
            const unsigned long dim1662JoeKuoD7Init[]   =   {   1   ,   3   ,   5   ,   7   ,   27  ,   57  ,   57  ,   83  ,   273 ,   973 ,   767 ,   3993    ,   5337    ,   15679   ,0 };
            const unsigned long dim1663JoeKuoD7Init[]   =   {   1   ,   1   ,   1   ,   11  ,   29  ,   49  ,   63  ,   211 ,   253 ,   947 ,   21  ,   1993    ,   6775    ,   1551    ,0 };
            const unsigned long dim1664JoeKuoD7Init[]   =   {   1   ,   3   ,   5   ,   15  ,   27  ,   17  ,   37  ,   81  ,   83  ,   991 ,   1509    ,   1931    ,   7389    ,   6053    ,0 };
            const unsigned long dim1665JoeKuoD7Init[]   =   {   1   ,   3   ,   3   ,   15  ,   19  ,   23  ,   53  ,   225 ,   131 ,   935 ,   333 ,   577 ,   7893    ,   13339   ,0 };
            const unsigned long dim1666JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   1   ,   27  ,   41  ,   97  ,   131 ,   135 ,   857 ,   89  ,   2285    ,   223 ,   15595   ,0 };
            const unsigned long dim1667JoeKuoD7Init[]   =   {   1   ,   1   ,   5   ,   13  ,   13  ,   29  ,   47  ,   179 ,   63  ,   261 ,   1409    ,   3921    ,   653 ,   12425   ,0 };
            const unsigned long dim1668JoeKuoD7Init[]   =   {   1   ,   3   ,   1   ,   7   ,   15  ,   53  ,   121 ,   7   ,   287 ,   251 ,   579 ,   1415    ,   453 ,   15615   ,0 };
            const unsigned long dim1669JoeKuoD7Init[]   =   {   1   ,   3   ,   1   ,   5   ,   3   ,   9   ,   41  ,   105 ,   505 ,   37  ,   665 ,   3323    ,   3613    ,   4533    ,0 };
            const unsigned long dim1670JoeKuoD7Init[]   =   {   1   ,   1   ,   3   ,   1   ,   5   ,   57  ,   53  ,   107 ,   279 ,   169 ,   1797    ,   3395    ,   7325    ,   5825    ,0 };
            const unsigned long dim1671JoeKuoD7Init[]   =   {   1   ,   3   ,   3   ,   13  ,   19  ,   23  ,   33  ,   145 ,   15  ,   459 ,   711 ,   2883    ,   2879    ,   7437    ,0 };
            const unsigned long dim1672JoeKuoD7Init[]   =   {   1   ,   1   ,   1   ,   1   ,   23  ,   17  ,   81  ,   233 ,   85  ,   435 ,   169 ,   287 ,   5913    ,   11063   ,0 };
            const unsigned long dim1673JoeKuoD7Init[]   =   {   1   ,   3   ,   1   ,   9   ,   23  ,   5   ,   7   ,   9   ,   393 ,   243 ,   1301    ,   365 ,   7529    ,   13875   ,0 };
            const unsigned long dim1674JoeKuoD7Init[]   =   {   1   ,   1   ,   1   ,   1   ,   9   ,   43  ,   127 ,   81  ,   49  ,   783 ,   1079    ,   4059    ,   5829    ,   7873    ,0 };
            const unsigned long dim1675JoeKuoD7Init[]   =   {   1   ,   3   ,   7   ,   7   ,   29  ,   49  ,   123 ,   81  ,   73  ,   137 ,   895 ,   791 ,   6801    ,   1209    ,0 };
            const unsigned long dim1676JoeKuoD7Init[]   =   {   1   ,   1   ,   5   ,   13  ,   7   ,   63  ,   93  ,   177 ,   263 ,   373 ,   39  ,   1713    ,   6793    ,   215 ,0 };
            const unsigned long dim1677JoeKuoD7Init[]   =   {   1   ,   3   ,   5   ,   9   ,   31  ,   19  ,   105 ,   187 ,   55  ,   715 ,   505 ,   2597    ,   2923    ,   16351   ,0 };
            const unsigned long dim1678JoeKuoD7Init[]   =   {   1   ,   3   ,   1   ,   15  ,   9   ,   29  ,   113 ,   41  ,   83  ,   667 ,   605 ,   1033    ,   1525    ,   2781    ,0 };
            const unsigned long dim1679JoeKuoD7Init[]   =   {   1   ,   3   ,   3   ,   5   ,   29  ,   27  ,   123 ,   159 ,   3   ,   659 ,   1757    ,   3603    ,   6419    ,   10125   ,0 };
            const unsigned long dim1680JoeKuoD7Init[]   =   {   1   ,   3   ,   3   ,   3   ,   17  ,   19  ,   1   ,   19  ,   459 ,   833 ,   1149    ,   413 ,   1835    ,   12449   ,0 };
            const unsigned long dim1681JoeKuoD7Init[]   =   {   1   ,   3   ,   3   ,   15  ,   27  ,   13  ,   109 ,   233 ,   341 ,   943 ,   173 ,   2491    ,   445 ,   5669    ,0 };
            const unsigned long dim1682JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   13  ,   17  ,   37  ,   53  ,   79  ,   107 ,   263 ,   353 ,   1193    ,   4445    ,   13969   ,0 };
            const unsigned long dim1683JoeKuoD7Init[]   =   {   1   ,   1   ,   5   ,   7   ,   17  ,   55  ,   49  ,   53  ,   9   ,   831 ,   1551    ,   1105    ,   7797    ,   10667   ,0 };
            const unsigned long dim1684JoeKuoD7Init[]   =   {   1   ,   3   ,   3   ,   13  ,   23  ,   9   ,   77  ,   85  ,   297 ,   875 ,   1339    ,   3019    ,   7211    ,   363 ,0 };
            const unsigned long dim1685JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   3   ,   13  ,   11  ,   13  ,   239 ,   235 ,   299 ,   1433    ,   3913    ,   6793    ,   8471    ,0 };
            const unsigned long dim1686JoeKuoD7Init[]   =   {   1   ,   3   ,   7   ,   15  ,   15  ,   43  ,   109 ,   33  ,   221 ,   543 ,   1067    ,   559 ,   2541    ,   5601    ,0 };
            const unsigned long dim1687JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   11  ,   15  ,   51  ,   77  ,   125 ,   195 ,   435 ,   1769    ,   1047    ,   2509    ,   11143   ,0 };
            const unsigned long dim1688JoeKuoD7Init[]   =   {   1   ,   1   ,   5   ,   13  ,   19  ,   43  ,   31  ,   115 ,   357 ,   197 ,   119 ,   1709    ,   1851    ,   5625    ,0 };
            const unsigned long dim1689JoeKuoD7Init[]   =   {   1   ,   1   ,   1   ,   15  ,   25  ,   19  ,   127 ,   165 ,   241 ,   255 ,   1737    ,   4053    ,   5987    ,   2573    ,0 };
            const unsigned long dim1690JoeKuoD7Init[]   =   {   1   ,   3   ,   3   ,   1   ,   5   ,   5   ,   21  ,   203 ,   379 ,   147 ,   1793    ,   461 ,   4863    ,   14659   ,0 };
            const unsigned long dim1691JoeKuoD7Init[]   =   {   1   ,   1   ,   3   ,   13  ,   15  ,   41  ,   125 ,   191 ,   471 ,   863 ,   485 ,   1029    ,   6375    ,   7825    ,0 };
            const unsigned long dim1692JoeKuoD7Init[]   =   {   1   ,   1   ,   5   ,   1   ,   13  ,   47  ,   53  ,   221 ,   351 ,   621 ,   1025    ,   3729    ,   367 ,   16091   ,0 };
            const unsigned long dim1693JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   11  ,   1   ,   47  ,   11  ,   91  ,   485 ,   601 ,   235 ,   241 ,   6631    ,   3159    ,0 };
            const unsigned long dim1694JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   11  ,   9   ,   59  ,   41  ,   121 ,   87  ,   969 ,   667 ,   3845    ,   3317    ,   14095   ,0 };
            const unsigned long dim1695JoeKuoD7Init[]   =   {   1   ,   3   ,   3   ,   9   ,   17  ,   43  ,   95  ,   111 ,   253 ,   497 ,   1263    ,   1323    ,   2669    ,   11567   ,0 };
            const unsigned long dim1696JoeKuoD7Init[]   =   {   1   ,   3   ,   1   ,   5   ,   17  ,   45  ,   15  ,   231 ,   7   ,   125 ,   67  ,   27  ,   5397    ,   15347   ,0 };
            const unsigned long dim1697JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   13  ,   15  ,   59  ,   119 ,   203 ,   89  ,   987 ,   337 ,   3313    ,   5533    ,   13157   ,0 };
            const unsigned long dim1698JoeKuoD7Init[]   =   {   1   ,   3   ,   1   ,   5   ,   31  ,   21  ,   11  ,   79  ,   485 ,   431 ,   1333    ,   3111    ,   2081    ,   4327    ,0 };
            const unsigned long dim1699JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   7   ,   1   ,   19  ,   111 ,   127 ,   23  ,   597 ,   1049    ,   2337    ,   6891    ,   7421    ,0 };
            const unsigned long dim1700JoeKuoD7Init[]   =   {   1   ,   3   ,   7   ,   7   ,   23  ,   3   ,   121 ,   171 ,   41  ,   3   ,   147 ,   397 ,   5413    ,   15457   ,0 };
            const unsigned long dim1701JoeKuoD7Init[]   =   {   1   ,   1   ,   1   ,   9   ,   11  ,   9   ,   107 ,   55  ,   303 ,   285 ,   599 ,   2507    ,   699 ,   8593    ,0 };
            const unsigned long dim1702JoeKuoD7Init[]   =   {   1   ,   3   ,   5   ,   3   ,   19  ,   61  ,   9   ,   33  ,   385 ,   761 ,   1185    ,   1163    ,   2251    ,   8497    ,0 };
            const unsigned long dim1703JoeKuoD7Init[]   =   {   1   ,   3   ,   5   ,   3   ,   5   ,   35  ,   31  ,   19  ,   173 ,   425 ,   833 ,   119 ,   6691    ,   11103   ,0 };
            const unsigned long dim1704JoeKuoD7Init[]   =   {   1   ,   3   ,   7   ,   3   ,   11  ,   15  ,   105 ,   239 ,   21  ,   477 ,   1303    ,   3437    ,   5511    ,   2469    ,0 };
            const unsigned long dim1705JoeKuoD7Init[]   =   {   1   ,   3   ,   7   ,   5   ,   13  ,   11  ,   49  ,   81  ,   505 ,   527 ,   1799    ,   403 ,   6849    ,   4551    ,0 };
            const unsigned long dim1706JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   1   ,   5   ,   1   ,   21  ,   83  ,   273 ,   477 ,   1281    ,   3173    ,   2065    ,   1559    ,0 };
            const unsigned long dim1707JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   5   ,   19  ,   43  ,   33  ,   191 ,   411 ,   879 ,   83  ,   2653    ,   1271    ,   10557   ,0 };
            const unsigned long dim1708JoeKuoD7Init[]   =   {   1   ,   3   ,   5   ,   1   ,   25  ,   27  ,   81  ,   81  ,   23  ,   575 ,   787 ,   1715    ,   7983    ,   14397   ,0 };
            const unsigned long dim1709JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   11  ,   27  ,   57  ,   31  ,   139 ,   65  ,   445 ,   1993    ,   3693    ,   2627    ,   941 ,0 };
            const unsigned long dim1710JoeKuoD7Init[]   =   {   1   ,   1   ,   1   ,   9   ,   5   ,   55  ,   23  ,   235 ,   103 ,   451 ,   1299    ,   3901    ,   2705    ,   12485   ,0 };
            const unsigned long dim1711JoeKuoD7Init[]   =   {   1   ,   1   ,   1   ,   11  ,   9   ,   57  ,   57  ,   129 ,   269 ,   841 ,   1859    ,   1149    ,   531 ,   11977   ,0 };
            const unsigned long dim1712JoeKuoD7Init[]   =   {   1   ,   3   ,   3   ,   7   ,   17  ,   11  ,   9   ,   89  ,   205 ,   3   ,   437 ,   475 ,   313 ,   7593    ,0 };
            const unsigned long dim1713JoeKuoD7Init[]   =   {   1   ,   3   ,   5   ,   13  ,   9   ,   37  ,   9   ,   109 ,   453 ,   965 ,   1259    ,   2713    ,   7301    ,   15453   ,0 };
            const unsigned long dim1714JoeKuoD7Init[]   =   {   1   ,   1   ,   1   ,   7   ,   25  ,   57  ,   53  ,   149 ,   291 ,   45  ,   783 ,   2327    ,   7439    ,   6567    ,0 };
            const unsigned long dim1715JoeKuoD7Init[]   =   {   1   ,   1   ,   1   ,   13  ,   5   ,   17  ,   65  ,   15  ,   279 ,   435 ,   1015    ,   3157    ,   1411    ,   3259    ,0 };
            const unsigned long dim1716JoeKuoD7Init[]   =   {   1   ,   3   ,   5   ,   1   ,   27  ,   9   ,   119 ,   11  ,   159 ,   829 ,   1499    ,   3797    ,   3059    ,   675 ,0 };
            const unsigned long dim1717JoeKuoD7Init[]   =   {   1   ,   1   ,   3   ,   7   ,   27  ,   5   ,   65  ,   141 ,   403 ,   431 ,   645 ,   3547    ,   7779    ,   12565   ,0 };
            const unsigned long dim1718JoeKuoD7Init[]   =   {   1   ,   1   ,   3   ,   15  ,   7   ,   11  ,   3   ,   53  ,   479 ,   841 ,   541 ,   2041    ,   631 ,   13969   ,0 };
            const unsigned long dim1719JoeKuoD7Init[]   =   {   1   ,   3   ,   7   ,   9   ,   19  ,   29  ,   15  ,   137 ,   317 ,   929 ,   509 ,   2231    ,   6273    ,   305 ,0 };
            const unsigned long dim1720JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   1   ,   19  ,   43  ,   99  ,   221 ,   429 ,   293 ,   133 ,   3233    ,   5097    ,   7521    ,0 };
            const unsigned long dim1721JoeKuoD7Init[]   =   {   1   ,   3   ,   3   ,   7   ,   5   ,   19  ,   11  ,   253 ,   375 ,   33  ,   941 ,   573 ,   1855    ,   13119   ,0 };
            const unsigned long dim1722JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   5   ,   27  ,   59  ,   61  ,   195 ,   197 ,   635 ,   889 ,   777 ,   2559    ,   3013    ,0 };
            const unsigned long dim1723JoeKuoD7Init[]   =   {   1   ,   1   ,   1   ,   7   ,   21  ,   45  ,   95  ,   61  ,   253 ,   239 ,   585 ,   75  ,   3737    ,   2199    ,0 };
            const unsigned long dim1724JoeKuoD7Init[]   =   {   1   ,   1   ,   1   ,   5   ,   1   ,   17  ,   61  ,   21  ,   485 ,   199 ,   7   ,   431 ,   1117    ,   2039    ,0 };
            const unsigned long dim1725JoeKuoD7Init[]   =   {   1   ,   1   ,   5   ,   7   ,   13  ,   23  ,   53  ,   211 ,   319 ,   47  ,   299 ,   2681    ,   6913    ,   5265    ,0 };
            const unsigned long dim1726JoeKuoD7Init[]   =   {   1   ,   1   ,   3   ,   1   ,   21  ,   49  ,   1   ,   197 ,   413 ,   475 ,   79  ,   3761    ,   6379    ,   10689   ,0 };
            const unsigned long dim1727JoeKuoD7Init[]   =   {   1   ,   3   ,   1   ,   11  ,   15  ,   37  ,   51  ,   81  ,   319 ,   985 ,   129 ,   3105    ,   1321    ,   9749    ,0 };
            const unsigned long dim1728JoeKuoD7Init[]   =   {   1   ,   1   ,   5   ,   13  ,   23  ,   31  ,   61  ,   49  ,   497 ,   987 ,   1401    ,   2609    ,   307 ,   4839    ,0 };
            const unsigned long dim1729JoeKuoD7Init[]   =   {   1   ,   3   ,   5   ,   7   ,   11  ,   25  ,   9   ,   97  ,   337 ,   535 ,   543 ,   145 ,   241 ,   1481    ,0 };
            const unsigned long dim1730JoeKuoD7Init[]   =   {   1   ,   3   ,   1   ,   11  ,   17  ,   55  ,   81  ,   161 ,   367 ,   825 ,   879 ,   301 ,   13  ,   7089    ,0 };
            const unsigned long dim1731JoeKuoD7Init[]   =   {   1   ,   3   ,   3   ,   3   ,   13  ,   11  ,   43  ,   237 ,   485 ,   3   ,   1191    ,   223 ,   767 ,   10297   ,0 };
            const unsigned long dim1732JoeKuoD7Init[]   =   {   1   ,   3   ,   3   ,   13  ,   21  ,   21  ,   93  ,   241 ,   119 ,   647 ,   921 ,   2847    ,   6889    ,   7055    ,0 };
            const unsigned long dim1733JoeKuoD7Init[]   =   {   1   ,   3   ,   5   ,   5   ,   5   ,   29  ,   3   ,   225 ,   307 ,   257 ,   1961    ,   2455    ,   2239    ,   3405    ,0 };
            const unsigned long dim1734JoeKuoD7Init[]   =   {   1   ,   3   ,   7   ,   11  ,   15  ,   33  ,   17  ,   179 ,   469 ,   5   ,   1577    ,   3539    ,   7123    ,   16093   ,0 };
            const unsigned long dim1735JoeKuoD7Init[]   =   {   1   ,   1   ,   3   ,   7   ,   27  ,   43  ,   45  ,   115 ,   507 ,   657 ,   1531    ,   2221    ,   3175    ,   1877    ,0 };
            const unsigned long dim1736JoeKuoD7Init[]   =   {   1   ,   3   ,   5   ,   1   ,   11  ,   25  ,   49  ,   155 ,   441 ,   703 ,   813 ,   1143    ,   219 ,   12675   ,0 };
            const unsigned long dim1737JoeKuoD7Init[]   =   {   1   ,   1   ,   1   ,   3   ,   15  ,   55  ,   15  ,   83  ,   333 ,   573 ,   1527    ,   3421    ,   6225    ,   3115    ,0 };
            const unsigned long dim1738JoeKuoD7Init[]   =   {   1   ,   3   ,   5   ,   1   ,   27  ,   3   ,   117 ,   149 ,   35  ,   587 ,   1123    ,   3037    ,   8083    ,   113 ,0 };
            const unsigned long dim1739JoeKuoD7Init[]   =   {   1   ,   3   ,   1   ,   5   ,   1   ,   23  ,   79  ,   99  ,   287 ,   505 ,   807 ,   4003    ,   4551    ,   2093    ,0 };
            const unsigned long dim1740JoeKuoD7Init[]   =   {   1   ,   3   ,   7   ,   9   ,   29  ,   5   ,   59  ,   111 ,   401 ,   179 ,   1263    ,   749 ,   689 ,   8905    ,0 };
            const unsigned long dim1741JoeKuoD7Init[]   =   {   1   ,   3   ,   7   ,   3   ,   17  ,   41  ,   81  ,   191 ,   313 ,   343 ,   809 ,   83  ,   405 ,   10665   ,0 };
            const unsigned long dim1742JoeKuoD7Init[]   =   {   1   ,   3   ,   5   ,   9   ,   7   ,   19  ,   59  ,   157 ,   169 ,   273 ,   935 ,   1435    ,   4725    ,   14049   ,0 };
            const unsigned long dim1743JoeKuoD7Init[]   =   {   1   ,   1   ,   5   ,   7   ,   11  ,   41  ,   61  ,   47  ,   43  ,   13  ,   783 ,   3685    ,   6417    ,   7859    ,0 };
            const unsigned long dim1744JoeKuoD7Init[]   =   {   1   ,   1   ,   3   ,   3   ,   13  ,   37  ,   127 ,   107 ,   439 ,   855 ,   1969    ,   1257    ,   1459    ,   3617    ,0 };
            const unsigned long dim1745JoeKuoD7Init[]   =   {   1   ,   3   ,   5   ,   1   ,   5   ,   7   ,   41  ,   77  ,   175 ,   679 ,   353 ,   2899    ,   1513    ,   9241    ,0 };
            const unsigned long dim1746JoeKuoD7Init[]   =   {   1   ,   1   ,   5   ,   1   ,   31  ,   41  ,   75  ,   217 ,   131 ,   261 ,   1659    ,   509 ,   3327    ,   14651   ,0 };
            const unsigned long dim1747JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   15  ,   15  ,   1   ,   111 ,   131 ,   19  ,   123 ,   701 ,   2633    ,   5125    ,   15109   ,0 };
            const unsigned long dim1748JoeKuoD7Init[]   =   {   1   ,   3   ,   3   ,   3   ,   15  ,   51  ,   19  ,   133 ,   41  ,   561 ,   869 ,   447 ,   1421    ,   8659    ,0 };
            const unsigned long dim1749JoeKuoD7Init[]   =   {   1   ,   1   ,   3   ,   9   ,   15  ,   47  ,   103 ,   5   ,   23  ,   97  ,   1413    ,   3307    ,   407 ,   6943    ,0 };
            const unsigned long dim1750JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   15  ,   19  ,   37  ,   113 ,   217 ,   451 ,   733 ,   1407    ,   2805    ,   6121    ,   12471   ,0 };
            const unsigned long dim1751JoeKuoD7Init[]   =   {   1   ,   1   ,   3   ,   13  ,   7   ,   53  ,   1   ,   159 ,   503 ,   603 ,   1523    ,   2327    ,   909 ,   3685    ,0 };
            const unsigned long dim1752JoeKuoD7Init[]   =   {   1   ,   3   ,   1   ,   13  ,   11  ,   47  ,   79  ,   69  ,   459 ,   761 ,   1181    ,   3369    ,   7983    ,   8561    ,0 };
            const unsigned long dim1753JoeKuoD7Init[]   =   {   1   ,   1   ,   5   ,   9   ,   3   ,   51  ,   69  ,   93  ,   375 ,   451 ,   1181    ,   1353    ,   939 ,   629 ,0 };
            const unsigned long dim1754JoeKuoD7Init[]   =   {   1   ,   1   ,   5   ,   9   ,   29  ,   57  ,   89  ,   145 ,   477 ,   17  ,   55  ,   2073    ,   1805    ,   5389    ,0 };
            const unsigned long dim1755JoeKuoD7Init[]   =   {   1   ,   3   ,   3   ,   7   ,   1   ,   59  ,   85  ,   69  ,   369 ,   191 ,   1419    ,   1331    ,   7451    ,   8769    ,0 };
            const unsigned long dim1756JoeKuoD7Init[]   =   {   1   ,   3   ,   3   ,   11  ,   11  ,   49  ,   31  ,   249 ,   285 ,   29  ,   633 ,   3387    ,   1197    ,   6363    ,0 };
            const unsigned long dim1757JoeKuoD7Init[]   =   {   1   ,   3   ,   7   ,   7   ,   15  ,   33  ,   3   ,   253 ,   247 ,   573 ,   2029    ,   3419    ,   8037    ,   2889    ,0 };
            const unsigned long dim1758JoeKuoD7Init[]   =   {   1   ,   3   ,   5   ,   1   ,   23  ,   31  ,   35  ,   113 ,   221 ,   659 ,   755 ,   2777    ,   3709    ,   7511    ,0 };
            const unsigned long dim1759JoeKuoD7Init[]   =   {   1   ,   1   ,   5   ,   7   ,   7   ,   29  ,   113 ,   81  ,   203 ,   711 ,   1427    ,   2233    ,   6049    ,   11417   ,0 };
            const unsigned long dim1760JoeKuoD7Init[]   =   {   1   ,   3   ,   3   ,   1   ,   19  ,   55  ,   107 ,   87  ,   383 ,   25  ,   935 ,   1287    ,   1867    ,   2133    ,0 };
            const unsigned long dim1761JoeKuoD7Init[]   =   {   1   ,   3   ,   7   ,   11  ,   9   ,   45  ,   15  ,   171 ,   293 ,   533 ,   1393    ,   759 ,   3009    ,   7879    ,0 };
            const unsigned long dim1762JoeKuoD7Init[]   =   {   1   ,   1   ,   1   ,   1   ,   1   ,   59  ,   51  ,   15  ,   471 ,   377 ,   1031    ,   3427    ,   6447    ,   10465   ,0 };
            const unsigned long dim1763JoeKuoD7Init[]   =   {   1   ,   3   ,   5   ,   15  ,   7   ,   15  ,   3   ,   1   ,   359 ,   331 ,   1295    ,   2777    ,   373 ,   4043    ,0 };
            const unsigned long dim1764JoeKuoD7Init[]   =   {   1   ,   1   ,   3   ,   9   ,   21  ,   41  ,   123 ,   61  ,   225 ,   305 ,   149 ,   2041    ,   7569    ,   14055   ,0 };
            const unsigned long dim1765JoeKuoD7Init[]   =   {   1   ,   3   ,   5   ,   3   ,   27  ,   15  ,   79  ,   47  ,   463 ,   327 ,   1231    ,   171 ,   4931    ,   7549    ,0 };
            const unsigned long dim1766JoeKuoD7Init[]   =   {   1   ,   3   ,   3   ,   15  ,   27  ,   11  ,   35  ,   179 ,   205 ,   579 ,   1297    ,   3149    ,   1253    ,   3067    ,0 };
            const unsigned long dim1767JoeKuoD7Init[]   =   {   1   ,   3   ,   1   ,   5   ,   15  ,   41  ,   17  ,   69  ,   369 ,   57  ,   779 ,   3055    ,   7799    ,   10755   ,0 };
            const unsigned long dim1768JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   11  ,   1   ,   11  ,   121 ,   177 ,   297 ,   933 ,   1319    ,   3123    ,   1233    ,   15521   ,0 };
            const unsigned long dim1769JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   7   ,   27  ,   59  ,   17  ,   155 ,   17  ,   109 ,   1623    ,   3527    ,   6227    ,   14295   ,0 };
            const unsigned long dim1770JoeKuoD7Init[]   =   {   1   ,   3   ,   1   ,   13  ,   27  ,   19  ,   47  ,   35  ,   403 ,   647 ,   1257    ,   1661    ,   7011    ,   11981   ,0 };
            const unsigned long dim1771JoeKuoD7Init[]   =   {   1   ,   1   ,   1   ,   13  ,   27  ,   47  ,   73  ,   53  ,   355 ,   857 ,   1091    ,   3165    ,   2729    ,   7685    ,0 };
            const unsigned long dim1772JoeKuoD7Init[]   =   {   1   ,   3   ,   1   ,   9   ,   7   ,   55  ,   53  ,   183 ,   109 ,   655 ,   1143    ,   975 ,   2809    ,   2709    ,0 };
            const unsigned long dim1773JoeKuoD7Init[]   =   {   1   ,   3   ,   3   ,   9   ,   11  ,   5   ,   121 ,   63  ,   221 ,   247 ,   1649    ,   709 ,   5215    ,   12159   ,0 };
            const unsigned long dim1774JoeKuoD7Init[]   =   {   1   ,   1   ,   3   ,   5   ,   5   ,   37  ,   83  ,   89  ,   475 ,   513 ,   563 ,   1965    ,   2335    ,   7921    ,0 };
            const unsigned long dim1775JoeKuoD7Init[]   =   {   1   ,   1   ,   3   ,   3   ,   15  ,   23  ,   79  ,   79  ,   499 ,   343 ,   917 ,   3495    ,   7685    ,   5939    ,0 };
            const unsigned long dim1776JoeKuoD7Init[]   =   {   1   ,   1   ,   1   ,   13  ,   5   ,   59  ,   87  ,   7   ,   469 ,   979 ,   683 ,   3397    ,   3787    ,   5063    ,0 };
            const unsigned long dim1777JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   5   ,   1   ,   27  ,   73  ,   85  ,   427 ,   619 ,   989 ,   2323    ,   4069    ,   13721   ,0 };
            const unsigned long dim1778JoeKuoD7Init[]   =   {   1   ,   3   ,   5   ,   3   ,   3   ,   39  ,   55  ,   153 ,   293 ,   483 ,   235 ,   2041    ,   549 ,   15519   ,0 };
            const unsigned long dim1779JoeKuoD7Init[]   =   {   1   ,   3   ,   1   ,   9   ,   17  ,   13  ,   61  ,   227 ,   435 ,   383 ,   513 ,   1169    ,   3335    ,   11257   ,0 };
            const unsigned long dim1780JoeKuoD7Init[]   =   {   1   ,   3   ,   3   ,   7   ,   1   ,   47  ,   61  ,   117 ,   483 ,   979 ,   163 ,   2621    ,   1701    ,   13331   ,0 };
            const unsigned long dim1781JoeKuoD7Init[]   =   {   1   ,   3   ,   3   ,   1   ,   17  ,   49  ,   115 ,   235 ,   511 ,   681 ,   1493    ,   137 ,   539 ,   5429    ,0 };
            const unsigned long dim1782JoeKuoD7Init[]   =   {   1   ,   3   ,   3   ,   13  ,   13  ,   61  ,   3   ,   1   ,   291 ,   101 ,   363 ,   1677    ,   2133    ,   3629    ,0 };
            const unsigned long dim1783JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   3   ,   1   ,   11  ,   83  ,   127 ,   291 ,   599 ,   819 ,   2637    ,   1197    ,   7435    ,0 };
            const unsigned long dim1784JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   1   ,   1   ,   23  ,   115 ,   47  ,   403 ,   377 ,   833 ,   3241    ,   6843    ,   10279   ,0 };
            const unsigned long dim1785JoeKuoD7Init[]   =   {   1   ,   1   ,   1   ,   15  ,   7   ,   51  ,   81  ,   215 ,   43  ,   445 ,   203 ,   3375    ,   6479    ,   3745    ,0 };
            const unsigned long dim1786JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   5   ,   7   ,   13  ,   25  ,   43  ,   233 ,   725 ,   267 ,   1875    ,   6481    ,   9685    ,0 };
            const unsigned long dim1787JoeKuoD7Init[]   =   {   1   ,   1   ,   1   ,   7   ,   23  ,   9   ,   37  ,   1   ,   51  ,   861 ,   1509    ,   11  ,   8119    ,   9171    ,0 };
            const unsigned long dim1788JoeKuoD7Init[]   =   {   1   ,   1   ,   5   ,   9   ,   17  ,   41  ,   69  ,   51  ,   445 ,   727 ,   159 ,   3317    ,   4951    ,   7519    ,0 };
            const unsigned long dim1789JoeKuoD7Init[]   =   {   1   ,   1   ,   3   ,   13  ,   11  ,   3   ,   9   ,   211 ,   453 ,   957 ,   1499    ,   953 ,   403 ,   10715   ,0 };
            const unsigned long dim1790JoeKuoD7Init[]   =   {   1   ,   3   ,   7   ,   7   ,   19  ,   1   ,   13  ,   255 ,   351 ,   325 ,   1055    ,   3719    ,   2343    ,   12693   ,0 };
            const unsigned long dim1791JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   3   ,   25  ,   29  ,   77  ,   185 ,   421 ,   33  ,   911 ,   2791    ,   4865    ,   4591    ,0 };
            const unsigned long dim1792JoeKuoD7Init[]   =   {   1   ,   3   ,   7   ,   1   ,   5   ,   53  ,   77  ,   25  ,   359 ,   69  ,   169 ,   225 ,   237 ,   2059    ,0 };
            const unsigned long dim1793JoeKuoD7Init[]   =   {   1   ,   3   ,   3   ,   3   ,   25  ,   63  ,   21  ,   111 ,   83  ,   21  ,   1513    ,   565 ,   1097    ,   7331    ,0 };
            const unsigned long dim1794JoeKuoD7Init[]   =   {   1   ,   3   ,   5   ,   15  ,   23  ,   33  ,   127 ,   221 ,   121 ,   95  ,   909 ,   1207    ,   2957    ,   3367    ,0 };
            const unsigned long dim1795JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   15  ,   11  ,   49  ,   35  ,   147 ,   265 ,   517 ,   2011    ,   3585    ,   3877    ,   2907    ,0 };
            const unsigned long dim1796JoeKuoD7Init[]   =   {   1   ,   3   ,   3   ,   1   ,   11  ,   37  ,   53  ,   47  ,   35  ,   323 ,   1283    ,   399 ,   7785    ,   8341    ,0 };
            const unsigned long dim1797JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   15  ,   5   ,   51  ,   83  ,   255 ,   231 ,   141 ,   1005    ,   3681    ,   3155    ,   11815   ,0 };
            const unsigned long dim1798JoeKuoD7Init[]   =   {   1   ,   1   ,   3   ,   7   ,   3   ,   15  ,   75  ,   81  ,   23  ,   137 ,   897 ,   2439    ,   5277    ,   6377    ,0 };
            const unsigned long dim1799JoeKuoD7Init[]   =   {   1   ,   1   ,   5   ,   1   ,   3   ,   11  ,   69  ,   231 ,   483 ,   797 ,   133 ,   497 ,   3973    ,   14993   ,0 };
            const unsigned long dim1800JoeKuoD7Init[]   =   {   1   ,   1   ,   3   ,   15  ,   23  ,   17  ,   89  ,   207 ,   111 ,   157 ,   1689    ,   3165    ,   7147    ,   15265   ,0 };
            const unsigned long dim1801JoeKuoD7Init[]   =   {   1   ,   3   ,   1   ,   5   ,   9   ,   31  ,   45  ,   131 ,   423 ,   53  ,   1723    ,   2009    ,   485 ,   15329   ,0 };
            const unsigned long dim1802JoeKuoD7Init[]   =   {   1   ,   1   ,   3   ,   3   ,   25  ,   5   ,   97  ,   81  ,   431 ,   701 ,   1799    ,   3595    ,   3147    ,   8919    ,0 };
            const unsigned long dim1803JoeKuoD7Init[]   =   {   1   ,   3   ,   7   ,   3   ,   3   ,   37  ,   3   ,   219 ,   291 ,   553 ,   1879    ,   15  ,   3357    ,   8479    ,0 };
            const unsigned long dim1804JoeKuoD7Init[]   =   {   1   ,   3   ,   7   ,   3   ,   3   ,   21  ,   55  ,   253 ,   49  ,   191 ,   951 ,   13  ,   5163    ,   12243   ,0 };
            const unsigned long dim1805JoeKuoD7Init[]   =   {   1   ,   1   ,   1   ,   1   ,   29  ,   39  ,   93  ,   123 ,   133 ,   465 ,   1665    ,   2023    ,   6773    ,   10521   ,0 };
            const unsigned long dim1806JoeKuoD7Init[]   =   {   1   ,   1   ,   1   ,   15  ,   15  ,   15  ,   107 ,   209 ,   103 ,   59  ,   33  ,   1927    ,   6517    ,   8479    ,0 };
            const unsigned long dim1807JoeKuoD7Init[]   =   {   1   ,   3   ,   3   ,   15  ,   3   ,   37  ,   19  ,   3   ,   43  ,   193 ,   1907    ,   2579    ,   7415    ,   9165    ,0 };
            const unsigned long dim1808JoeKuoD7Init[]   =   {   1   ,   3   ,   5   ,   11  ,   29  ,   63  ,   47  ,   79  ,   227 ,   569 ,   67  ,   4015    ,   1275    ,   11963   ,0 };
            const unsigned long dim1809JoeKuoD7Init[]   =   {   1   ,   3   ,   7   ,   1   ,   1   ,   33  ,   103 ,   111 ,   129 ,   189 ,   837 ,   3741    ,   699 ,   14433   ,0 };
            const unsigned long dim1810JoeKuoD7Init[]   =   {   1   ,   3   ,   7   ,   11  ,   19  ,   15  ,   43  ,   35  ,   367 ,   535 ,   913 ,   2109    ,   1397    ,   12199   ,0 };
            const unsigned long dim1811JoeKuoD7Init[]   =   {   1   ,   3   ,   7   ,   15  ,   27  ,   15  ,   41  ,   59  ,   429 ,   423 ,   377 ,   905 ,   7403    ,   6175    ,0 };
            const unsigned long dim1812JoeKuoD7Init[]   =   {   1   ,   1   ,   5   ,   5   ,   27  ,   55  ,   49  ,   247 ,   353 ,   773 ,   707 ,   2501    ,   7567    ,   1821    ,0 };
            const unsigned long dim1813JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   15  ,   5   ,   11  ,   31  ,   49  ,   331 ,   1017    ,   909 ,   967 ,   5557    ,   7279    ,0 };
            const unsigned long dim1814JoeKuoD7Init[]   =   {   1   ,   1   ,   5   ,   5   ,   13  ,   55  ,   127 ,   111 ,   193 ,   77  ,   1353    ,   349 ,   5785    ,   16325   ,0 };
            const unsigned long dim1815JoeKuoD7Init[]   =   {   1   ,   1   ,   5   ,   11  ,   23  ,   61  ,   83  ,   79  ,   409 ,   877 ,   1625    ,   705 ,   6159    ,   1701    ,0 };
            const unsigned long dim1816JoeKuoD7Init[]   =   {   1   ,   1   ,   5   ,   1   ,   29  ,   39  ,   69  ,   221 ,   29  ,   83  ,   21  ,   3525    ,   7387    ,   9249    ,0 };
            const unsigned long dim1817JoeKuoD7Init[]   =   {   1   ,   3   ,   7   ,   3   ,   13  ,   57  ,   63  ,   159 ,   25  ,   151 ,   1749    ,   3191    ,   4805    ,   6003    ,0 };
            const unsigned long dim1818JoeKuoD7Init[]   =   {   1   ,   1   ,   3   ,   15  ,   11  ,   49  ,   71  ,   241 ,   421 ,   859 ,   1923    ,   539 ,   1721    ,   2903    ,0 };
            const unsigned long dim1819JoeKuoD7Init[]   =   {   1   ,   3   ,   1   ,   1   ,   15  ,   33  ,   123 ,   133 ,   323 ,   173 ,   271 ,   2641    ,   2649    ,   5485    ,0 };
            const unsigned long dim1820JoeKuoD7Init[]   =   {   1   ,   1   ,   3   ,   13  ,   13  ,   31  ,   39  ,   147 ,   313 ,   497 ,   2019    ,   3269    ,   2671    ,   6599    ,0 };
            const unsigned long dim1821JoeKuoD7Init[]   =   {   1   ,   3   ,   7   ,   9   ,   13  ,   37  ,   111 ,   87  ,   3   ,   417 ,   923 ,   1833    ,   5383    ,   12045   ,0 };
            const unsigned long dim1822JoeKuoD7Init[]   =   {   1   ,   3   ,   3   ,   13  ,   25  ,   27  ,   91  ,   121 ,   151 ,   407 ,   899 ,   469 ,   6647    ,   2799    ,0 };
            const unsigned long dim1823JoeKuoD7Init[]   =   {   1   ,   3   ,   1   ,   7   ,   19  ,   39  ,   29  ,   43  ,   195 ,   439 ,   1065    ,   2035    ,   749 ,   14013   ,0 };
            const unsigned long dim1824JoeKuoD7Init[]   =   {   1   ,   1   ,   5   ,   13  ,   21  ,   1   ,   69  ,   223 ,   83  ,   457 ,   1589    ,   1199    ,   7635    ,   1987    ,0 };
            const unsigned long dim1825JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   13  ,   5   ,   21  ,   35  ,   169 ,   71  ,   855 ,   497 ,   1119    ,   6681    ,   3905    ,0 };
            const unsigned long dim1826JoeKuoD7Init[]   =   {   1   ,   3   ,   1   ,   13  ,   25  ,   55  ,   109 ,   113 ,   41  ,   461 ,   185 ,   3677    ,   1807    ,   13773   ,0 };
            const unsigned long dim1827JoeKuoD7Init[]   =   {   1   ,   3   ,   1   ,   5   ,   15  ,   61  ,   91  ,   219 ,   365 ,   399 ,   505 ,   1207    ,   2237    ,   9267    ,0 };
            const unsigned long dim1828JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   3   ,   7   ,   1   ,   29  ,   151 ,   215 ,   587 ,   1499    ,   355 ,   5121    ,   14963   ,0 };
            const unsigned long dim1829JoeKuoD7Init[]   =   {   1   ,   3   ,   1   ,   1   ,   3   ,   23  ,   127 ,   189 ,   139 ,   987 ,   1291    ,   3583    ,   2881    ,   1927    ,0 };
            const unsigned long dim1830JoeKuoD7Init[]   =   {   1   ,   1   ,   1   ,   11  ,   19  ,   51  ,   25  ,   179 ,   35  ,   517 ,   1531    ,   3657    ,   7025    ,   7079    ,0 };
            const unsigned long dim1831JoeKuoD7Init[]   =   {   1   ,   1   ,   5   ,   7   ,   23  ,   51  ,   35  ,   157 ,   113 ,   993 ,   265 ,   1119    ,   2639    ,   14475   ,0 };
            const unsigned long dim1832JoeKuoD7Init[]   =   {   1   ,   3   ,   7   ,   1   ,   23  ,   19  ,   111 ,   83  ,   413 ,   205 ,   87  ,   2389    ,   7593    ,   9189    ,0 };
            const unsigned long dim1833JoeKuoD7Init[]   =   {   1   ,   3   ,   5   ,   11  ,   25  ,   25  ,   39  ,   59  ,   419 ,   477 ,   1189    ,   807 ,   7679    ,   9867    ,0 };
            const unsigned long dim1834JoeKuoD7Init[]   =   {   1   ,   1   ,   5   ,   1   ,   9   ,   59  ,   99  ,   183 ,   493 ,   811 ,   441 ,   1039    ,   1099    ,   10229   ,0 };
            const unsigned long dim1835JoeKuoD7Init[]   =   {   1   ,   1   ,   5   ,   9   ,   31  ,   27  ,   51  ,   179 ,   277 ,   549 ,   153 ,   1785    ,   5227    ,   13695   ,0 };
            const unsigned long dim1836JoeKuoD7Init[]   =   {   1   ,   3   ,   7   ,   11  ,   29  ,   43  ,   9   ,   205 ,   499 ,   55  ,   745 ,   2217    ,   793 ,   14517   ,0 };
            const unsigned long dim1837JoeKuoD7Init[]   =   {   1   ,   3   ,   5   ,   7   ,   29  ,   47  ,   15  ,   153 ,   73  ,   971 ,   149 ,   2579    ,   1699    ,   14849   ,0 };
            const unsigned long dim1838JoeKuoD7Init[]   =   {   1   ,   3   ,   3   ,   7   ,   31  ,   33  ,   61  ,   103 ,   503 ,   531 ,   2033    ,   2415    ,   4151    ,   4937    ,0 };
            const unsigned long dim1839JoeKuoD7Init[]   =   {   1   ,   3   ,   1   ,   15  ,   15  ,   53  ,   27  ,   67  ,   351 ,   273 ,   637 ,   4085    ,   3445    ,   2085    ,0 };
            const unsigned long dim1840JoeKuoD7Init[]   =   {   1   ,   3   ,   5   ,   15  ,   7   ,   57  ,   121 ,   23  ,   73  ,   765 ,   1949    ,   67  ,   8061    ,   11167   ,0 };
            const unsigned long dim1841JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   1   ,   25  ,   7   ,   107 ,   13  ,   409 ,   115 ,   1451    ,   1215    ,   39  ,   11577   ,0 };
            const unsigned long dim1842JoeKuoD7Init[]   =   {   1   ,   1   ,   5   ,   15  ,   3   ,   15  ,   71  ,   1   ,   337 ,   929 ,   1435    ,   2847    ,   3257    ,   12233   ,0 };
            const unsigned long dim1843JoeKuoD7Init[]   =   {   1   ,   1   ,   5   ,   11  ,   31  ,   37  ,   61  ,   243 ,   255 ,   79  ,   1735    ,   3769    ,   4893    ,   15103   ,0 };
            const unsigned long dim1844JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   11  ,   5   ,   63  ,   49  ,   197 ,   301 ,   883 ,   951 ,   2357    ,   7501    ,   15499   ,0 };
            const unsigned long dim1845JoeKuoD7Init[]   =   {   1   ,   3   ,   7   ,   3   ,   13  ,   53  ,   71  ,   235 ,   183 ,   951 ,   845 ,   3361    ,   4757    ,   7745    ,0 };
            const unsigned long dim1846JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   3   ,   25  ,   61  ,   119 ,   3   ,   487 ,   503 ,   179 ,   2085    ,   4093    ,   1339    ,0 };
            const unsigned long dim1847JoeKuoD7Init[]   =   {   1   ,   1   ,   5   ,   3   ,   21  ,   15  ,   113 ,   175 ,   61  ,   789 ,   1369    ,   109 ,   5963    ,   9401    ,0 };
            const unsigned long dim1848JoeKuoD7Init[]   =   {   1   ,   1   ,   5   ,   5   ,   21  ,   35  ,   103 ,   33  ,   147 ,   561 ,   1033    ,   1497    ,   309 ,   12505   ,0 };
            const unsigned long dim1849JoeKuoD7Init[]   =   {   1   ,   3   ,   3   ,   13  ,   25  ,   17  ,   29  ,   101 ,   103 ,   719 ,   329 ,   1317    ,   4487    ,   9115    ,0 };
            const unsigned long dim1850JoeKuoD7Init[]   =   {   1   ,   1   ,   3   ,   13  ,   5   ,   61  ,   25  ,   79  ,   161 ,   673 ,   47  ,   2485    ,   7711    ,   4959    ,0 };
            const unsigned long dim1851JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   13  ,   19  ,   13  ,   1   ,   27  ,   443 ,   369 ,   669 ,   2555    ,   3003    ,   5117    ,0 };
            const unsigned long dim1852JoeKuoD7Init[]   =   {   1   ,   3   ,   3   ,   5   ,   21  ,   29  ,   65  ,   211 ,   343 ,   717 ,   949 ,   1987    ,   6571    ,   11299   ,0 };
            const unsigned long dim1853JoeKuoD7Init[]   =   {   1   ,   3   ,   3   ,   9   ,   19  ,   7   ,   115 ,   119 ,   369 ,   247 ,   1165    ,   4033    ,   861 ,   5667    ,0 };
            const unsigned long dim1854JoeKuoD7Init[]   =   {   1   ,   3   ,   5   ,   3   ,   3   ,   53  ,   15  ,   219 ,   223 ,   873 ,   1145    ,   2781    ,   3297    ,   10551   ,0 };
            const unsigned long dim1855JoeKuoD7Init[]   =   {   1   ,   3   ,   1   ,   15  ,   5   ,   61  ,   15  ,   121 ,   201 ,   89  ,   647 ,   3079    ,   1505    ,   10003   ,0 };
            const unsigned long dim1856JoeKuoD7Init[]   =   {   1   ,   3   ,   1   ,   13  ,   17  ,   55  ,   27  ,   215 ,   67  ,   461 ,   1955    ,   765 ,   483 ,   11695   ,0 };
            const unsigned long dim1857JoeKuoD7Init[]   =   {   1   ,   3   ,   1   ,   1   ,   1   ,   21  ,   59  ,   19  ,   339 ,   157 ,   2019    ,   971 ,   1797    ,   5237    ,0 };
            const unsigned long dim1858JoeKuoD7Init[]   =   {   1   ,   3   ,   7   ,   9   ,   15  ,   1   ,   67  ,   119 ,   63  ,   59  ,   1107    ,   2201    ,   5499    ,   3757    ,0 };
            const unsigned long dim1859JoeKuoD7Init[]   =   {   1   ,   3   ,   3   ,   13  ,   29  ,   59  ,   73  ,   89  ,   469 ,   335 ,   53  ,   315 ,   4371    ,   6417    ,0 };
            const unsigned long dim1860JoeKuoD7Init[]   =   {   1   ,   1   ,   5   ,   3   ,   21  ,   7   ,   11  ,   205 ,   485 ,   695 ,   1959    ,   2407    ,   5423    ,   15359   ,0 };
            const unsigned long dim1861JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   1   ,   1   ,   43  ,   61  ,   103 ,   65  ,   747 ,   695 ,   3855    ,   7765    ,   13299   ,0 };
            const unsigned long dim1862JoeKuoD7Init[]   =   {   1   ,   1   ,   3   ,   11  ,   9   ,   11  ,   47  ,   59  ,   287 ,   363 ,   1677    ,   613 ,   4181    ,   16221   ,0 };
            const unsigned long dim1863JoeKuoD7Init[]   =   {   1   ,   1   ,   1   ,   3   ,   9   ,   43  ,   1   ,   107 ,   291 ,   493 ,   571 ,   2329    ,   2707    ,   13683   ,0 };
            const unsigned long dim1864JoeKuoD7Init[]   =   {   1   ,   1   ,   5   ,   1   ,   13  ,   27  ,   29  ,   159 ,   357 ,   187 ,   613 ,   1803    ,   5563    ,   14913   ,0 };
            const unsigned long dim1865JoeKuoD7Init[]   =   {   1   ,   3   ,   5   ,   13  ,   1   ,   5   ,   69  ,   69  ,   17  ,   527 ,   1019    ,   2565    ,   1383    ,   3695    ,0 };
            const unsigned long dim1866JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   3   ,   9   ,   21  ,   9   ,   151 ,   145 ,   891 ,   1961    ,   257 ,   5885    ,   5201    ,0 };
            const unsigned long dim1867JoeKuoD7Init[]   =   {   1   ,   1   ,   3   ,   15  ,   9   ,   7   ,   31  ,   3   ,   115 ,   697 ,   369 ,   1531    ,   3581    ,   5681    ,   27421   ,0 };
            const unsigned long dim1868JoeKuoD7Init[]   =   {   1   ,   1   ,   5   ,   9   ,   13  ,   33  ,   19  ,   191 ,   433 ,   651 ,   2023    ,   373 ,   4665    ,   15147   ,   28277   ,0 };
            const unsigned long dim1869JoeKuoD7Init[]   =   {   1   ,   3   ,   5   ,   7   ,   27  ,   3   ,   15  ,   57  ,   137 ,   903 ,   355 ,   503 ,   3   ,   13015   ,   32415   ,0 };
            const unsigned long dim1870JoeKuoD7Init[]   =   {   1   ,   3   ,   5   ,   1   ,   13  ,   57  ,   69  ,   55  ,   75  ,   633 ,   1271    ,   189 ,   4279    ,   1293    ,   5719    ,0 };
            const unsigned long dim1871JoeKuoD7Init[]   =   {   1   ,   3   ,   5   ,   15  ,   27  ,   9   ,   13  ,   185 ,   211 ,   593 ,   723 ,   3725    ,   5165    ,   15375   ,   1055    ,0 };
            const unsigned long dim1872JoeKuoD7Init[]   =   {   1   ,   3   ,   7   ,   13  ,   21  ,   11  ,   67  ,   11  ,   279 ,   503 ,   1839    ,   3509    ,   1285    ,   5251    ,   27835   ,0 };
            const unsigned long dim1873JoeKuoD7Init[]   =   {   1   ,   1   ,   3   ,   15  ,   5   ,   45  ,   35  ,   145 ,   383 ,   227 ,   327 ,   2419    ,   5207    ,   15897   ,   20869   ,0 };
            const unsigned long dim1874JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   11  ,   31  ,   25  ,   69  ,   169 ,   363 ,   819 ,   1855    ,   61  ,   5613    ,   11835   ,   24761   ,0 };
            const unsigned long dim1875JoeKuoD7Init[]   =   {   1   ,   3   ,   7   ,   9   ,   15  ,   41  ,   123 ,   61  ,   359 ,   29  ,   369 ,   1167    ,   489 ,   12947   ,   14765   ,0 };
            const unsigned long dim1876JoeKuoD7Init[]   =   {   1   ,   3   ,   3   ,   5   ,   7   ,   33  ,   115 ,   193 ,   427 ,   123 ,   1795    ,   3153    ,   2771    ,   6761    ,   719 ,0 };
            const unsigned long dim1877JoeKuoD7Init[]   =   {   1   ,   1   ,   1   ,   13  ,   31  ,   39  ,   5   ,   179 ,   181 ,   787 ,   1891    ,   2287    ,   2085    ,   15081   ,   2035    ,0 };
            const unsigned long dim1878JoeKuoD7Init[]   =   {   1   ,   1   ,   5   ,   15  ,   1   ,   47  ,   15  ,   235 ,   431 ,   687 ,   1787    ,   225 ,   7749    ,   1179    ,   29243   ,0 };
            const unsigned long dim1879JoeKuoD7Init[]   =   {   1   ,   1   ,   1   ,   13  ,   19  ,   53  ,   119 ,   131 ,   417 ,   505 ,   1179    ,   2291    ,   4795    ,   10475   ,   6307    ,0 };
            const unsigned long dim1880JoeKuoD7Init[]   =   {   1   ,   1   ,   5   ,   13  ,   15  ,   43  ,   73  ,   3   ,   193 ,   5   ,   1843    ,   1617    ,   7557    ,   4165    ,   4089    ,0 };
            const unsigned long dim1881JoeKuoD7Init[]   =   {   1   ,   1   ,   1   ,   5   ,   3   ,   35  ,   121 ,   183 ,   197 ,   605 ,   1775    ,   1647    ,   4929    ,   365 ,   43  ,0 };
            const unsigned long dim1882JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   7   ,   13  ,   31  ,   69  ,   157 ,   309 ,   99  ,   705 ,   1705    ,   2093    ,   13055   ,   6727    ,0 };
            const unsigned long dim1883JoeKuoD7Init[]   =   {   1   ,   1   ,   1   ,   3   ,   29  ,   41  ,   109 ,   129 ,   223 ,   755 ,   595 ,   1931    ,   3967    ,   8249    ,   9439    ,0 };
            const unsigned long dim1884JoeKuoD7Init[]   =   {   1   ,   3   ,   5   ,   1   ,   21  ,   45  ,   101 ,   229 ,   343 ,   455 ,   125 ,   1335    ,   749 ,   3967    ,   22777   ,0 };
            const unsigned long dim1885JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   13  ,   31  ,   49  ,   125 ,   199 ,   37  ,   685 ,   1171    ,   1793    ,   1057    ,   10041   ,   12693   ,0 };
            const unsigned long dim1886JoeKuoD7Init[]   =   {   1   ,   3   ,   7   ,   5   ,   19  ,   33  ,   35  ,   115 ,   269 ,   151 ,   731 ,   1803    ,   5241    ,   3667    ,   25371   ,0 };
            const unsigned long dim1887JoeKuoD7Init[]   =   {   1   ,   3   ,   5   ,   3   ,   7   ,   27  ,   25  ,   249 ,   275 ,   331 ,   1579    ,   2911    ,   5859    ,   9893    ,   27807   ,0 };
            const unsigned long dim1888JoeKuoD7Init[]   =   {   1   ,   3   ,   1   ,   7   ,   21  ,   41  ,   81  ,   239 ,   141 ,   861 ,   827 ,   1261    ,   6873    ,   6413    ,   19313   ,0 };
            const unsigned long dim1889JoeKuoD7Init[]   =   {   1   ,   1   ,   7   ,   1   ,   9   ,   51  ,   39  ,   123 ,   145 ,   929 ,   1101    ,   2505    ,   6493    ,   95  ,   17247   ,0 };
            const unsigned long dim1890JoeKuoD7Init[]   =   {   1   ,   1   ,   5   ,   5   ,   11  ,   3   ,   47  ,   167 ,   5   ,   229 ,   1603    ,   3541    ,   5505    ,   1221    ,   899 ,0 };
            const unsigned long dim1891JoeKuoD7Init[]   =   {   1   ,   1   ,   1   ,   9   ,   5   ,   17  ,   7   ,   21  ,   499 ,   597 ,   1775    ,   3187    ,   831 ,   10995   ,   7721    ,0 };
            const unsigned long dim1892JoeKuoD7Init[]   =   {   1   ,   3   ,   3   ,   15  ,   27  ,   25  ,   7   ,   205 ,   341 ,   965 ,   519 ,   2907    ,   373 ,   5827    ,   14243   ,0 };
            const unsigned long dim1893JoeKuoD7Init[]   =   {   1   ,   3   ,   1   ,   3   ,   19  ,   27  ,   13  ,   207 ,   287 ,   983 ,   575 ,   891 ,   3869    ,   11213   ,   11377   ,0 };
            const unsigned long dim1894JoeKuoD7Init[]   =   {   1   ,   1   ,   3   ,   9   ,   1   ,   5   ,   53  ,   7   ,   507 ,   505 ,   133 ,   3567    ,   6601    ,   7985    ,   10461   ,0 };
            const unsigned long dim1895JoeKuoD7Init[]   =   {   1   ,   1   ,   3   ,   9   ,   15  ,   9   ,   87  ,   231 ,   39  ,   701 ,   625 ,   2783    ,   6163    ,   6945    ,   4521    ,0 };
            const unsigned long dim1896JoeKuoD7Init[]   =   {   1   ,   3   ,   5   ,   7   ,   25  ,   39  ,   63  ,   215 ,   89  ,   775 ,   283 ,   2919    ,   151 ,   7233    ,   32163   ,0 };
            const unsigned long dim1897JoeKuoD7Init[]   =   {   1   ,   3   ,   3   ,   13  ,   31  ,   31  ,   55  ,   125 ,   433 ,   77  ,   1123    ,   3151    ,   6421    ,   6343    ,   14893   ,0 };
            const unsigned long dim1898JoeKuoD7Init[]   =   {   1   ,   1   ,   5   ,   9   ,   17  ,   55  ,   63  ,   31  ,   331 ,   945 ,   1953    ,   4093    ,   6875    ,   3519    ,   23329   ,0 };
            const unsigned long dim1899JoeKuoD7Init[]   =   {   1   ,   1   ,   1   ,   3   ,   23  ,   17  ,   17  ,   39  ,   173 ,   1013    ,   527 ,   2563    ,   3623    ,   10049   ,   10919   ,0 };

            const unsigned long * const JoeKuoD7initializers[1899]
            =
            {
                dim1JoeKuoD7Init,
                dim2JoeKuoD7Init,
                dim3JoeKuoD7Init,
                dim4JoeKuoD7Init,
                dim5JoeKuoD7Init,
                dim6JoeKuoD7Init,
                dim7JoeKuoD7Init,
                dim8JoeKuoD7Init,
                dim9JoeKuoD7Init,
                dim10JoeKuoD7Init,
                dim11JoeKuoD7Init,
                dim12JoeKuoD7Init,
                dim13JoeKuoD7Init,
                dim14JoeKuoD7Init,
                dim15JoeKuoD7Init,
                dim16JoeKuoD7Init,
                dim17JoeKuoD7Init,
                dim18JoeKuoD7Init,
                dim19JoeKuoD7Init,
                dim20JoeKuoD7Init,
                dim21JoeKuoD7Init,
                dim22JoeKuoD7Init,
                dim23JoeKuoD7Init,
                dim24JoeKuoD7Init,
                dim25JoeKuoD7Init,
                dim26JoeKuoD7Init,
                dim27JoeKuoD7Init,
                dim28JoeKuoD7Init,
                dim29JoeKuoD7Init,
                dim30JoeKuoD7Init,
                dim31JoeKuoD7Init,
                dim32JoeKuoD7Init,
                dim33JoeKuoD7Init,
                dim34JoeKuoD7Init,
                dim35JoeKuoD7Init,
                dim36JoeKuoD7Init,
                dim37JoeKuoD7Init,
                dim38JoeKuoD7Init,
                dim39JoeKuoD7Init,
                dim40JoeKuoD7Init,
                dim41JoeKuoD7Init,
                dim42JoeKuoD7Init,
                dim43JoeKuoD7Init,
                dim44JoeKuoD7Init,
                dim45JoeKuoD7Init,
                dim46JoeKuoD7Init,
                dim47JoeKuoD7Init,
                dim48JoeKuoD7Init,
                dim49JoeKuoD7Init,
                dim50JoeKuoD7Init,
                dim51JoeKuoD7Init,
                dim52JoeKuoD7Init,
                dim53JoeKuoD7Init,
                dim54JoeKuoD7Init,
                dim55JoeKuoD7Init,
                dim56JoeKuoD7Init,
                dim57JoeKuoD7Init,
                dim58JoeKuoD7Init,
                dim59JoeKuoD7Init,
                dim60JoeKuoD7Init,
                dim61JoeKuoD7Init,
                dim62JoeKuoD7Init,
                dim63JoeKuoD7Init,
                dim64JoeKuoD7Init,
                dim65JoeKuoD7Init,
                dim66JoeKuoD7Init,
                dim67JoeKuoD7Init,
                dim68JoeKuoD7Init,
                dim69JoeKuoD7Init,
                dim70JoeKuoD7Init,
                dim71JoeKuoD7Init,
                dim72JoeKuoD7Init,
                dim73JoeKuoD7Init,
                dim74JoeKuoD7Init,
                dim75JoeKuoD7Init,
                dim76JoeKuoD7Init,
                dim77JoeKuoD7Init,
                dim78JoeKuoD7Init,
                dim79JoeKuoD7Init,
                dim80JoeKuoD7Init,
                dim81JoeKuoD7Init,
                dim82JoeKuoD7Init,
                dim83JoeKuoD7Init,
                dim84JoeKuoD7Init,
                dim85JoeKuoD7Init,
                dim86JoeKuoD7Init,
                dim87JoeKuoD7Init,
                dim88JoeKuoD7Init,
                dim89JoeKuoD7Init,
                dim90JoeKuoD7Init,
                dim91JoeKuoD7Init,
                dim92JoeKuoD7Init,
                dim93JoeKuoD7Init,
                dim94JoeKuoD7Init,
                dim95JoeKuoD7Init,
                dim96JoeKuoD7Init,
                dim97JoeKuoD7Init,
                dim98JoeKuoD7Init,
                dim99JoeKuoD7Init,
                dim100JoeKuoD7Init,
                dim101JoeKuoD7Init,
                dim102JoeKuoD7Init,
                dim103JoeKuoD7Init,
                dim104JoeKuoD7Init,
                dim105JoeKuoD7Init,
                dim106JoeKuoD7Init,
                dim107JoeKuoD7Init,
                dim108JoeKuoD7Init,
                dim109JoeKuoD7Init,
                dim110JoeKuoD7Init,
                dim111JoeKuoD7Init,
                dim112JoeKuoD7Init,
                dim113JoeKuoD7Init,
                dim114JoeKuoD7Init,
                dim115JoeKuoD7Init,
                dim116JoeKuoD7Init,
                dim117JoeKuoD7Init,
                dim118JoeKuoD7Init,
                dim119JoeKuoD7Init,
                dim120JoeKuoD7Init,
                dim121JoeKuoD7Init,
                dim122JoeKuoD7Init,
                dim123JoeKuoD7Init,
                dim124JoeKuoD7Init,
                dim125JoeKuoD7Init,
                dim126JoeKuoD7Init,
                dim127JoeKuoD7Init,
                dim128JoeKuoD7Init,
                dim129JoeKuoD7Init,
                dim130JoeKuoD7Init,
                dim131JoeKuoD7Init,
                dim132JoeKuoD7Init,
                dim133JoeKuoD7Init,
                dim134JoeKuoD7Init,
                dim135JoeKuoD7Init,
                dim136JoeKuoD7Init,
                dim137JoeKuoD7Init,
                dim138JoeKuoD7Init,
                dim139JoeKuoD7Init,
                dim140JoeKuoD7Init,
                dim141JoeKuoD7Init,
                dim142JoeKuoD7Init,
                dim143JoeKuoD7Init,
                dim144JoeKuoD7Init,
                dim145JoeKuoD7Init,
                dim146JoeKuoD7Init,
                dim147JoeKuoD7Init,
                dim148JoeKuoD7Init,
                dim149JoeKuoD7Init,
                dim150JoeKuoD7Init,
                dim151JoeKuoD7Init,
                dim152JoeKuoD7Init,
                dim153JoeKuoD7Init,
                dim154JoeKuoD7Init,
                dim155JoeKuoD7Init,
                dim156JoeKuoD7Init,
                dim157JoeKuoD7Init,
                dim158JoeKuoD7Init,
                dim159JoeKuoD7Init,
                dim160JoeKuoD7Init,
                dim161JoeKuoD7Init,
                dim162JoeKuoD7Init,
                dim163JoeKuoD7Init,
                dim164JoeKuoD7Init,
                dim165JoeKuoD7Init,
                dim166JoeKuoD7Init,
                dim167JoeKuoD7Init,
                dim168JoeKuoD7Init,
                dim169JoeKuoD7Init,
                dim170JoeKuoD7Init,
                dim171JoeKuoD7Init,
                dim172JoeKuoD7Init,
                dim173JoeKuoD7Init,
                dim174JoeKuoD7Init,
                dim175JoeKuoD7Init,
                dim176JoeKuoD7Init,
                dim177JoeKuoD7Init,
                dim178JoeKuoD7Init,
                dim179JoeKuoD7Init,
                dim180JoeKuoD7Init,
                dim181JoeKuoD7Init,
                dim182JoeKuoD7Init,
                dim183JoeKuoD7Init,
                dim184JoeKuoD7Init,
                dim185JoeKuoD7Init,
                dim186JoeKuoD7Init,
                dim187JoeKuoD7Init,
                dim188JoeKuoD7Init,
                dim189JoeKuoD7Init,
                dim190JoeKuoD7Init,
                dim191JoeKuoD7Init,
                dim192JoeKuoD7Init,
                dim193JoeKuoD7Init,
                dim194JoeKuoD7Init,
                dim195JoeKuoD7Init,
                dim196JoeKuoD7Init,
                dim197JoeKuoD7Init,
                dim198JoeKuoD7Init,
                dim199JoeKuoD7Init,
                dim200JoeKuoD7Init,
                dim201JoeKuoD7Init,
                dim202JoeKuoD7Init,
                dim203JoeKuoD7Init,
                dim204JoeKuoD7Init,
                dim205JoeKuoD7Init,
                dim206JoeKuoD7Init,
                dim207JoeKuoD7Init,
                dim208JoeKuoD7Init,
                dim209JoeKuoD7Init,
                dim210JoeKuoD7Init,
                dim211JoeKuoD7Init,
                dim212JoeKuoD7Init,
                dim213JoeKuoD7Init,
                dim214JoeKuoD7Init,
                dim215JoeKuoD7Init,
                dim216JoeKuoD7Init,
                dim217JoeKuoD7Init,
                dim218JoeKuoD7Init,
                dim219JoeKuoD7Init,
                dim220JoeKuoD7Init,
                dim221JoeKuoD7Init,
                dim222JoeKuoD7Init,
                dim223JoeKuoD7Init,
                dim224JoeKuoD7Init,
                dim225JoeKuoD7Init,
                dim226JoeKuoD7Init,
                dim227JoeKuoD7Init,
                dim228JoeKuoD7Init,
                dim229JoeKuoD7Init,
                dim230JoeKuoD7Init,
                dim231JoeKuoD7Init,
                dim232JoeKuoD7Init,
                dim233JoeKuoD7Init,
                dim234JoeKuoD7Init,
                dim235JoeKuoD7Init,
                dim236JoeKuoD7Init,
                dim237JoeKuoD7Init,
                dim238JoeKuoD7Init,
                dim239JoeKuoD7Init,
                dim240JoeKuoD7Init,
                dim241JoeKuoD7Init,
                dim242JoeKuoD7Init,
                dim243JoeKuoD7Init,
                dim244JoeKuoD7Init,
                dim245JoeKuoD7Init,
                dim246JoeKuoD7Init,
                dim247JoeKuoD7Init,
                dim248JoeKuoD7Init,
                dim249JoeKuoD7Init,
                dim250JoeKuoD7Init,
                dim251JoeKuoD7Init,
                dim252JoeKuoD7Init,
                dim253JoeKuoD7Init,
                dim254JoeKuoD7Init,
                dim255JoeKuoD7Init,
                dim256JoeKuoD7Init,
                dim257JoeKuoD7Init,
                dim258JoeKuoD7Init,
                dim259JoeKuoD7Init,
                dim260JoeKuoD7Init,
                dim261JoeKuoD7Init,
                dim262JoeKuoD7Init,
                dim263JoeKuoD7Init,
                dim264JoeKuoD7Init,
                dim265JoeKuoD7Init,
                dim266JoeKuoD7Init,
                dim267JoeKuoD7Init,
                dim268JoeKuoD7Init,
                dim269JoeKuoD7Init,
                dim270JoeKuoD7Init,
                dim271JoeKuoD7Init,
                dim272JoeKuoD7Init,
                dim273JoeKuoD7Init,
                dim274JoeKuoD7Init,
                dim275JoeKuoD7Init,
                dim276JoeKuoD7Init,
                dim277JoeKuoD7Init,
                dim278JoeKuoD7Init,
                dim279JoeKuoD7Init,
                dim280JoeKuoD7Init,
                dim281JoeKuoD7Init,
                dim282JoeKuoD7Init,
                dim283JoeKuoD7Init,
                dim284JoeKuoD7Init,
                dim285JoeKuoD7Init,
                dim286JoeKuoD7Init,
                dim287JoeKuoD7Init,
                dim288JoeKuoD7Init,
                dim289JoeKuoD7Init,
                dim290JoeKuoD7Init,
                dim291JoeKuoD7Init,
                dim292JoeKuoD7Init,
                dim293JoeKuoD7Init,
                dim294JoeKuoD7Init,
                dim295JoeKuoD7Init,
                dim296JoeKuoD7Init,
                dim297JoeKuoD7Init,
                dim298JoeKuoD7Init,
                dim299JoeKuoD7Init,
                dim300JoeKuoD7Init,
                dim301JoeKuoD7Init,
                dim302JoeKuoD7Init,
                dim303JoeKuoD7Init,
                dim304JoeKuoD7Init,
                dim305JoeKuoD7Init,
                dim306JoeKuoD7Init,
                dim307JoeKuoD7Init,
                dim308JoeKuoD7Init,
                dim309JoeKuoD7Init,
                dim310JoeKuoD7Init,
                dim311JoeKuoD7Init,
                dim312JoeKuoD7Init,
                dim313JoeKuoD7Init,
                dim314JoeKuoD7Init,
                dim315JoeKuoD7Init,
                dim316JoeKuoD7Init,
                dim317JoeKuoD7Init,
                dim318JoeKuoD7Init,
                dim319JoeKuoD7Init,
                dim320JoeKuoD7Init,
                dim321JoeKuoD7Init,
                dim322JoeKuoD7Init,
                dim323JoeKuoD7Init,
                dim324JoeKuoD7Init,
                dim325JoeKuoD7Init,
                dim326JoeKuoD7Init,
                dim327JoeKuoD7Init,
                dim328JoeKuoD7Init,
                dim329JoeKuoD7Init,
                dim330JoeKuoD7Init,
                dim331JoeKuoD7Init,
                dim332JoeKuoD7Init,
                dim333JoeKuoD7Init,
                dim334JoeKuoD7Init,
                dim335JoeKuoD7Init,
                dim336JoeKuoD7Init,
                dim337JoeKuoD7Init,
                dim338JoeKuoD7Init,
                dim339JoeKuoD7Init,
                dim340JoeKuoD7Init,
                dim341JoeKuoD7Init,
                dim342JoeKuoD7Init,
                dim343JoeKuoD7Init,
                dim344JoeKuoD7Init,
                dim345JoeKuoD7Init,
                dim346JoeKuoD7Init,
                dim347JoeKuoD7Init,
                dim348JoeKuoD7Init,
                dim349JoeKuoD7Init,
                dim350JoeKuoD7Init,
                dim351JoeKuoD7Init,
                dim352JoeKuoD7Init,
                dim353JoeKuoD7Init,
                dim354JoeKuoD7Init,
                dim355JoeKuoD7Init,
                dim356JoeKuoD7Init,
                dim357JoeKuoD7Init,
                dim358JoeKuoD7Init,
                dim359JoeKuoD7Init,
                dim360JoeKuoD7Init,
                dim361JoeKuoD7Init,
                dim362JoeKuoD7Init,
                dim363JoeKuoD7Init,
                dim364JoeKuoD7Init,
                dim365JoeKuoD7Init,
                dim366JoeKuoD7Init,
                dim367JoeKuoD7Init,
                dim368JoeKuoD7Init,
                dim369JoeKuoD7Init,
                dim370JoeKuoD7Init,
                dim371JoeKuoD7Init,
                dim372JoeKuoD7Init,
                dim373JoeKuoD7Init,
                dim374JoeKuoD7Init,
                dim375JoeKuoD7Init,
                dim376JoeKuoD7Init,
                dim377JoeKuoD7Init,
                dim378JoeKuoD7Init,
                dim379JoeKuoD7Init,
                dim380JoeKuoD7Init,
                dim381JoeKuoD7Init,
                dim382JoeKuoD7Init,
                dim383JoeKuoD7Init,
                dim384JoeKuoD7Init,
                dim385JoeKuoD7Init,
                dim386JoeKuoD7Init,
                dim387JoeKuoD7Init,
                dim388JoeKuoD7Init,
                dim389JoeKuoD7Init,
                dim390JoeKuoD7Init,
                dim391JoeKuoD7Init,
                dim392JoeKuoD7Init,
                dim393JoeKuoD7Init,
                dim394JoeKuoD7Init,
                dim395JoeKuoD7Init,
                dim396JoeKuoD7Init,
                dim397JoeKuoD7Init,
                dim398JoeKuoD7Init,
                dim399JoeKuoD7Init,
                dim400JoeKuoD7Init,
                dim401JoeKuoD7Init,
                dim402JoeKuoD7Init,
                dim403JoeKuoD7Init,
                dim404JoeKuoD7Init,
                dim405JoeKuoD7Init,
                dim406JoeKuoD7Init,
                dim407JoeKuoD7Init,
                dim408JoeKuoD7Init,
                dim409JoeKuoD7Init,
                dim410JoeKuoD7Init,
                dim411JoeKuoD7Init,
                dim412JoeKuoD7Init,
                dim413JoeKuoD7Init,
                dim414JoeKuoD7Init,
                dim415JoeKuoD7Init,
                dim416JoeKuoD7Init,
                dim417JoeKuoD7Init,
                dim418JoeKuoD7Init,
                dim419JoeKuoD7Init,
                dim420JoeKuoD7Init,
                dim421JoeKuoD7Init,
                dim422JoeKuoD7Init,
                dim423JoeKuoD7Init,
                dim424JoeKuoD7Init,
                dim425JoeKuoD7Init,
                dim426JoeKuoD7Init,
                dim427JoeKuoD7Init,
                dim428JoeKuoD7Init,
                dim429JoeKuoD7Init,
                dim430JoeKuoD7Init,
                dim431JoeKuoD7Init,
                dim432JoeKuoD7Init,
                dim433JoeKuoD7Init,
                dim434JoeKuoD7Init,
                dim435JoeKuoD7Init,
                dim436JoeKuoD7Init,
                dim437JoeKuoD7Init,
                dim438JoeKuoD7Init,
                dim439JoeKuoD7Init,
                dim440JoeKuoD7Init,
                dim441JoeKuoD7Init,
                dim442JoeKuoD7Init,
                dim443JoeKuoD7Init,
                dim444JoeKuoD7Init,
                dim445JoeKuoD7Init,
                dim446JoeKuoD7Init,
                dim447JoeKuoD7Init,
                dim448JoeKuoD7Init,
                dim449JoeKuoD7Init,
                dim450JoeKuoD7Init,
                dim451JoeKuoD7Init,
                dim452JoeKuoD7Init,
                dim453JoeKuoD7Init,
                dim454JoeKuoD7Init,
                dim455JoeKuoD7Init,
                dim456JoeKuoD7Init,
                dim457JoeKuoD7Init,
                dim458JoeKuoD7Init,
                dim459JoeKuoD7Init,
                dim460JoeKuoD7Init,
                dim461JoeKuoD7Init,
                dim462JoeKuoD7Init,
                dim463JoeKuoD7Init,
                dim464JoeKuoD7Init,
                dim465JoeKuoD7Init,
                dim466JoeKuoD7Init,
                dim467JoeKuoD7Init,
                dim468JoeKuoD7Init,
                dim469JoeKuoD7Init,
                dim470JoeKuoD7Init,
                dim471JoeKuoD7Init,
                dim472JoeKuoD7Init,
                dim473JoeKuoD7Init,
                dim474JoeKuoD7Init,
                dim475JoeKuoD7Init,
                dim476JoeKuoD7Init,
                dim477JoeKuoD7Init,
                dim478JoeKuoD7Init,
                dim479JoeKuoD7Init,
                dim480JoeKuoD7Init,
                dim481JoeKuoD7Init,
                dim482JoeKuoD7Init,
                dim483JoeKuoD7Init,
                dim484JoeKuoD7Init,
                dim485JoeKuoD7Init,
                dim486JoeKuoD7Init,
                dim487JoeKuoD7Init,
                dim488JoeKuoD7Init,
                dim489JoeKuoD7Init,
                dim490JoeKuoD7Init,
                dim491JoeKuoD7Init,
                dim492JoeKuoD7Init,
                dim493JoeKuoD7Init,
                dim494JoeKuoD7Init,
                dim495JoeKuoD7Init,
                dim496JoeKuoD7Init,
                dim497JoeKuoD7Init,
                dim498JoeKuoD7Init,
                dim499JoeKuoD7Init,
                dim500JoeKuoD7Init,
                dim501JoeKuoD7Init,
                dim502JoeKuoD7Init,
                dim503JoeKuoD7Init,
                dim504JoeKuoD7Init,
                dim505JoeKuoD7Init,
                dim506JoeKuoD7Init,
                dim507JoeKuoD7Init,
                dim508JoeKuoD7Init,
                dim509JoeKuoD7Init,
                dim510JoeKuoD7Init,
                dim511JoeKuoD7Init,
                dim512JoeKuoD7Init,
                dim513JoeKuoD7Init,
                dim514JoeKuoD7Init,
                dim515JoeKuoD7Init,
                dim516JoeKuoD7Init,
                dim517JoeKuoD7Init,
                dim518JoeKuoD7Init,
                dim519JoeKuoD7Init,
                dim520JoeKuoD7Init,
                dim521JoeKuoD7Init,
                dim522JoeKuoD7Init,
                dim523JoeKuoD7Init,
                dim524JoeKuoD7Init,
                dim525JoeKuoD7Init,
                dim526JoeKuoD7Init,
                dim527JoeKuoD7Init,
                dim528JoeKuoD7Init,
                dim529JoeKuoD7Init,
                dim530JoeKuoD7Init,
                dim531JoeKuoD7Init,
                dim532JoeKuoD7Init,
                dim533JoeKuoD7Init,
                dim534JoeKuoD7Init,
                dim535JoeKuoD7Init,
                dim536JoeKuoD7Init,
                dim537JoeKuoD7Init,
                dim538JoeKuoD7Init,
                dim539JoeKuoD7Init,
                dim540JoeKuoD7Init,
                dim541JoeKuoD7Init,
                dim542JoeKuoD7Init,
                dim543JoeKuoD7Init,
                dim544JoeKuoD7Init,
                dim545JoeKuoD7Init,
                dim546JoeKuoD7Init,
                dim547JoeKuoD7Init,
                dim548JoeKuoD7Init,
                dim549JoeKuoD7Init,
                dim550JoeKuoD7Init,
                dim551JoeKuoD7Init,
                dim552JoeKuoD7Init,
                dim553JoeKuoD7Init,
                dim554JoeKuoD7Init,
                dim555JoeKuoD7Init,
                dim556JoeKuoD7Init,
                dim557JoeKuoD7Init,
                dim558JoeKuoD7Init,
                dim559JoeKuoD7Init,
                dim560JoeKuoD7Init,
                dim561JoeKuoD7Init,
                dim562JoeKuoD7Init,
                dim563JoeKuoD7Init,
                dim564JoeKuoD7Init,
                dim565JoeKuoD7Init,
                dim566JoeKuoD7Init,
                dim567JoeKuoD7Init,
                dim568JoeKuoD7Init,
                dim569JoeKuoD7Init,
                dim570JoeKuoD7Init,
                dim571JoeKuoD7Init,
                dim572JoeKuoD7Init,
                dim573JoeKuoD7Init,
                dim574JoeKuoD7Init,
                dim575JoeKuoD7Init,
                dim576JoeKuoD7Init,
                dim577JoeKuoD7Init,
                dim578JoeKuoD7Init,
                dim579JoeKuoD7Init,
                dim580JoeKuoD7Init,
                dim581JoeKuoD7Init,
                dim582JoeKuoD7Init,
                dim583JoeKuoD7Init,
                dim584JoeKuoD7Init,
                dim585JoeKuoD7Init,
                dim586JoeKuoD7Init,
                dim587JoeKuoD7Init,
                dim588JoeKuoD7Init,
                dim589JoeKuoD7Init,
                dim590JoeKuoD7Init,
                dim591JoeKuoD7Init,
                dim592JoeKuoD7Init,
                dim593JoeKuoD7Init,
                dim594JoeKuoD7Init,
                dim595JoeKuoD7Init,
                dim596JoeKuoD7Init,
                dim597JoeKuoD7Init,
                dim598JoeKuoD7Init,
                dim599JoeKuoD7Init,
                dim600JoeKuoD7Init,
                dim601JoeKuoD7Init,
                dim602JoeKuoD7Init,
                dim603JoeKuoD7Init,
                dim604JoeKuoD7Init,
                dim605JoeKuoD7Init,
                dim606JoeKuoD7Init,
                dim607JoeKuoD7Init,
                dim608JoeKuoD7Init,
                dim609JoeKuoD7Init,
                dim610JoeKuoD7Init,
                dim611JoeKuoD7Init,
                dim612JoeKuoD7Init,
                dim613JoeKuoD7Init,
                dim614JoeKuoD7Init,
                dim615JoeKuoD7Init,
                dim616JoeKuoD7Init,
                dim617JoeKuoD7Init,
                dim618JoeKuoD7Init,
                dim619JoeKuoD7Init,
                dim620JoeKuoD7Init,
                dim621JoeKuoD7Init,
                dim622JoeKuoD7Init,
                dim623JoeKuoD7Init,
                dim624JoeKuoD7Init,
                dim625JoeKuoD7Init,
                dim626JoeKuoD7Init,
                dim627JoeKuoD7Init,
                dim628JoeKuoD7Init,
                dim629JoeKuoD7Init,
                dim630JoeKuoD7Init,
                dim631JoeKuoD7Init,
                dim632JoeKuoD7Init,
                dim633JoeKuoD7Init,
                dim634JoeKuoD7Init,
                dim635JoeKuoD7Init,
                dim636JoeKuoD7Init,
                dim637JoeKuoD7Init,
                dim638JoeKuoD7Init,
                dim639JoeKuoD7Init,
                dim640JoeKuoD7Init,
                dim641JoeKuoD7Init,
                dim642JoeKuoD7Init,
                dim643JoeKuoD7Init,
                dim644JoeKuoD7Init,
                dim645JoeKuoD7Init,
                dim646JoeKuoD7Init,
                dim647JoeKuoD7Init,
                dim648JoeKuoD7Init,
                dim649JoeKuoD7Init,
                dim650JoeKuoD7Init,
                dim651JoeKuoD7Init,
                dim652JoeKuoD7Init,
                dim653JoeKuoD7Init,
                dim654JoeKuoD7Init,
                dim655JoeKuoD7Init,
                dim656JoeKuoD7Init,
                dim657JoeKuoD7Init,
                dim658JoeKuoD7Init,
                dim659JoeKuoD7Init,
                dim660JoeKuoD7Init,
                dim661JoeKuoD7Init,
                dim662JoeKuoD7Init,
                dim663JoeKuoD7Init,
                dim664JoeKuoD7Init,
                dim665JoeKuoD7Init,
                dim666JoeKuoD7Init,
                dim667JoeKuoD7Init,
                dim668JoeKuoD7Init,
                dim669JoeKuoD7Init,
                dim670JoeKuoD7Init,
                dim671JoeKuoD7Init,
                dim672JoeKuoD7Init,
                dim673JoeKuoD7Init,
                dim674JoeKuoD7Init,
                dim675JoeKuoD7Init,
                dim676JoeKuoD7Init,
                dim677JoeKuoD7Init,
                dim678JoeKuoD7Init,
                dim679JoeKuoD7Init,
                dim680JoeKuoD7Init,
                dim681JoeKuoD7Init,
                dim682JoeKuoD7Init,
                dim683JoeKuoD7Init,
                dim684JoeKuoD7Init,
                dim685JoeKuoD7Init,
                dim686JoeKuoD7Init,
                dim687JoeKuoD7Init,
                dim688JoeKuoD7Init,
                dim689JoeKuoD7Init,
                dim690JoeKuoD7Init,
                dim691JoeKuoD7Init,
                dim692JoeKuoD7Init,
                dim693JoeKuoD7Init,
                dim694JoeKuoD7Init,
                dim695JoeKuoD7Init,
                dim696JoeKuoD7Init,
                dim697JoeKuoD7Init,
                dim698JoeKuoD7Init,
                dim699JoeKuoD7Init,
                dim700JoeKuoD7Init,
                dim701JoeKuoD7Init,
                dim702JoeKuoD7Init,
                dim703JoeKuoD7Init,
                dim704JoeKuoD7Init,
                dim705JoeKuoD7Init,
                dim706JoeKuoD7Init,
                dim707JoeKuoD7Init,
                dim708JoeKuoD7Init,
                dim709JoeKuoD7Init,
                dim710JoeKuoD7Init,
                dim711JoeKuoD7Init,
                dim712JoeKuoD7Init,
                dim713JoeKuoD7Init,
                dim714JoeKuoD7Init,
                dim715JoeKuoD7Init,
                dim716JoeKuoD7Init,
                dim717JoeKuoD7Init,
                dim718JoeKuoD7Init,
                dim719JoeKuoD7Init,
                dim720JoeKuoD7Init,
                dim721JoeKuoD7Init,
                dim722JoeKuoD7Init,
                dim723JoeKuoD7Init,
                dim724JoeKuoD7Init,
                dim725JoeKuoD7Init,
                dim726JoeKuoD7Init,
                dim727JoeKuoD7Init,
                dim728JoeKuoD7Init,
                dim729JoeKuoD7Init,
                dim730JoeKuoD7Init,
                dim731JoeKuoD7Init,
                dim732JoeKuoD7Init,
                dim733JoeKuoD7Init,
                dim734JoeKuoD7Init,
                dim735JoeKuoD7Init,
                dim736JoeKuoD7Init,
                dim737JoeKuoD7Init,
                dim738JoeKuoD7Init,
                dim739JoeKuoD7Init,
                dim740JoeKuoD7Init,
                dim741JoeKuoD7Init,
                dim742JoeKuoD7Init,
                dim743JoeKuoD7Init,
                dim744JoeKuoD7Init,
                dim745JoeKuoD7Init,
                dim746JoeKuoD7Init,
                dim747JoeKuoD7Init,
                dim748JoeKuoD7Init,
                dim749JoeKuoD7Init,
                dim750JoeKuoD7Init,
                dim751JoeKuoD7Init,
                dim752JoeKuoD7Init,
                dim753JoeKuoD7Init,
                dim754JoeKuoD7Init,
                dim755JoeKuoD7Init,
                dim756JoeKuoD7Init,
                dim757JoeKuoD7Init,
                dim758JoeKuoD7Init,
                dim759JoeKuoD7Init,
                dim760JoeKuoD7Init,
                dim761JoeKuoD7Init,
                dim762JoeKuoD7Init,
                dim763JoeKuoD7Init,
                dim764JoeKuoD7Init,
                dim765JoeKuoD7Init,
                dim766JoeKuoD7Init,
                dim767JoeKuoD7Init,
                dim768JoeKuoD7Init,
                dim769JoeKuoD7Init,
                dim770JoeKuoD7Init,
                dim771JoeKuoD7Init,
                dim772JoeKuoD7Init,
                dim773JoeKuoD7Init,
                dim774JoeKuoD7Init,
                dim775JoeKuoD7Init,
                dim776JoeKuoD7Init,
                dim777JoeKuoD7Init,
                dim778JoeKuoD7Init,
                dim779JoeKuoD7Init,
                dim780JoeKuoD7Init,
                dim781JoeKuoD7Init,
                dim782JoeKuoD7Init,
                dim783JoeKuoD7Init,
                dim784JoeKuoD7Init,
                dim785JoeKuoD7Init,
                dim786JoeKuoD7Init,
                dim787JoeKuoD7Init,
                dim788JoeKuoD7Init,
                dim789JoeKuoD7Init,
                dim790JoeKuoD7Init,
                dim791JoeKuoD7Init,
                dim792JoeKuoD7Init,
                dim793JoeKuoD7Init,
                dim794JoeKuoD7Init,
                dim795JoeKuoD7Init,
                dim796JoeKuoD7Init,
                dim797JoeKuoD7Init,
                dim798JoeKuoD7Init,
                dim799JoeKuoD7Init,
                dim800JoeKuoD7Init,
                dim801JoeKuoD7Init,
                dim802JoeKuoD7Init,
                dim803JoeKuoD7Init,
                dim804JoeKuoD7Init,
                dim805JoeKuoD7Init,
                dim806JoeKuoD7Init,
                dim807JoeKuoD7Init,
                dim808JoeKuoD7Init,
                dim809JoeKuoD7Init,
                dim810JoeKuoD7Init,
                dim811JoeKuoD7Init,
                dim812JoeKuoD7Init,
                dim813JoeKuoD7Init,
                dim814JoeKuoD7Init,
                dim815JoeKuoD7Init,
                dim816JoeKuoD7Init,
                dim817JoeKuoD7Init,
                dim818JoeKuoD7Init,
                dim819JoeKuoD7Init,
                dim820JoeKuoD7Init,
                dim821JoeKuoD7Init,
                dim822JoeKuoD7Init,
                dim823JoeKuoD7Init,
                dim824JoeKuoD7Init,
                dim825JoeKuoD7Init,
                dim826JoeKuoD7Init,
                dim827JoeKuoD7Init,
                dim828JoeKuoD7Init,
                dim829JoeKuoD7Init,
                dim830JoeKuoD7Init,
                dim831JoeKuoD7Init,
                dim832JoeKuoD7Init,
                dim833JoeKuoD7Init,
                dim834JoeKuoD7Init,
                dim835JoeKuoD7Init,
                dim836JoeKuoD7Init,
                dim837JoeKuoD7Init,
                dim838JoeKuoD7Init,
                dim839JoeKuoD7Init,
                dim840JoeKuoD7Init,
                dim841JoeKuoD7Init,
                dim842JoeKuoD7Init,
                dim843JoeKuoD7Init,
                dim844JoeKuoD7Init,
                dim845JoeKuoD7Init,
                dim846JoeKuoD7Init,
                dim847JoeKuoD7Init,
                dim848JoeKuoD7Init,
                dim849JoeKuoD7Init,
                dim850JoeKuoD7Init,
                dim851JoeKuoD7Init,
                dim852JoeKuoD7Init,
                dim853JoeKuoD7Init,
                dim854JoeKuoD7Init,
                dim855JoeKuoD7Init,
                dim856JoeKuoD7Init,
                dim857JoeKuoD7Init,
                dim858JoeKuoD7Init,
                dim859JoeKuoD7Init,
                dim860JoeKuoD7Init,
                dim861JoeKuoD7Init,
                dim862JoeKuoD7Init,
                dim863JoeKuoD7Init,
                dim864JoeKuoD7Init,
                dim865JoeKuoD7Init,
                dim866JoeKuoD7Init,
                dim867JoeKuoD7Init,
                dim868JoeKuoD7Init,
                dim869JoeKuoD7Init,
                dim870JoeKuoD7Init,
                dim871JoeKuoD7Init,
                dim872JoeKuoD7Init,
                dim873JoeKuoD7Init,
                dim874JoeKuoD7Init,
                dim875JoeKuoD7Init,
                dim876JoeKuoD7Init,
                dim877JoeKuoD7Init,
                dim878JoeKuoD7Init,
                dim879JoeKuoD7Init,
                dim880JoeKuoD7Init,
                dim881JoeKuoD7Init,
                dim882JoeKuoD7Init,
                dim883JoeKuoD7Init,
                dim884JoeKuoD7Init,
                dim885JoeKuoD7Init,
                dim886JoeKuoD7Init,
                dim887JoeKuoD7Init,
                dim888JoeKuoD7Init,
                dim889JoeKuoD7Init,
                dim890JoeKuoD7Init,
                dim891JoeKuoD7Init,
                dim892JoeKuoD7Init,
                dim893JoeKuoD7Init,
                dim894JoeKuoD7Init,
                dim895JoeKuoD7Init,
                dim896JoeKuoD7Init,
                dim897JoeKuoD7Init,
                dim898JoeKuoD7Init,
                dim899JoeKuoD7Init,
                dim900JoeKuoD7Init,
                dim901JoeKuoD7Init,
                dim902JoeKuoD7Init,
                dim903JoeKuoD7Init,
                dim904JoeKuoD7Init,
                dim905JoeKuoD7Init,
                dim906JoeKuoD7Init,
                dim907JoeKuoD7Init,
                dim908JoeKuoD7Init,
                dim909JoeKuoD7Init,
                dim910JoeKuoD7Init,
                dim911JoeKuoD7Init,
                dim912JoeKuoD7Init,
                dim913JoeKuoD7Init,
                dim914JoeKuoD7Init,
                dim915JoeKuoD7Init,
                dim916JoeKuoD7Init,
                dim917JoeKuoD7Init,
                dim918JoeKuoD7Init,
                dim919JoeKuoD7Init,
                dim920JoeKuoD7Init,
                dim921JoeKuoD7Init,
                dim922JoeKuoD7Init,
                dim923JoeKuoD7Init,
                dim924JoeKuoD7Init,
                dim925JoeKuoD7Init,
                dim926JoeKuoD7Init,
                dim927JoeKuoD7Init,
                dim928JoeKuoD7Init,
                dim929JoeKuoD7Init,
                dim930JoeKuoD7Init,
                dim931JoeKuoD7Init,
                dim932JoeKuoD7Init,
                dim933JoeKuoD7Init,
                dim934JoeKuoD7Init,
                dim935JoeKuoD7Init,
                dim936JoeKuoD7Init,
                dim937JoeKuoD7Init,
                dim938JoeKuoD7Init,
                dim939JoeKuoD7Init,
                dim940JoeKuoD7Init,
                dim941JoeKuoD7Init,
                dim942JoeKuoD7Init,
                dim943JoeKuoD7Init,
                dim944JoeKuoD7Init,
                dim945JoeKuoD7Init,
                dim946JoeKuoD7Init,
                dim947JoeKuoD7Init,
                dim948JoeKuoD7Init,
                dim949JoeKuoD7Init,
                dim950JoeKuoD7Init,
                dim951JoeKuoD7Init,
                dim952JoeKuoD7Init,
                dim953JoeKuoD7Init,
                dim954JoeKuoD7Init,
                dim955JoeKuoD7Init,
                dim956JoeKuoD7Init,
                dim957JoeKuoD7Init,
                dim958JoeKuoD7Init,
                dim959JoeKuoD7Init,
                dim960JoeKuoD7Init,
                dim961JoeKuoD7Init,
                dim962JoeKuoD7Init,
                dim963JoeKuoD7Init,
                dim964JoeKuoD7Init,
                dim965JoeKuoD7Init,
                dim966JoeKuoD7Init,
                dim967JoeKuoD7Init,
                dim968JoeKuoD7Init,
                dim969JoeKuoD7Init,
                dim970JoeKuoD7Init,
                dim971JoeKuoD7Init,
                dim972JoeKuoD7Init,
                dim973JoeKuoD7Init,
                dim974JoeKuoD7Init,
                dim975JoeKuoD7Init,
                dim976JoeKuoD7Init,
                dim977JoeKuoD7Init,
                dim978JoeKuoD7Init,
                dim979JoeKuoD7Init,
                dim980JoeKuoD7Init,
                dim981JoeKuoD7Init,
                dim982JoeKuoD7Init,
                dim983JoeKuoD7Init,
                dim984JoeKuoD7Init,
                dim985JoeKuoD7Init,
                dim986JoeKuoD7Init,
                dim987JoeKuoD7Init,
                dim988JoeKuoD7Init,
                dim989JoeKuoD7Init,
                dim990JoeKuoD7Init,
                dim991JoeKuoD7Init,
                dim992JoeKuoD7Init,
                dim993JoeKuoD7Init,
                dim994JoeKuoD7Init,
                dim995JoeKuoD7Init,
                dim996JoeKuoD7Init,
                dim997JoeKuoD7Init,
                dim998JoeKuoD7Init,
                dim999JoeKuoD7Init,
                dim1000JoeKuoD7Init,
                dim1001JoeKuoD7Init,
                dim1002JoeKuoD7Init,
                dim1003JoeKuoD7Init,
                dim1004JoeKuoD7Init,
                dim1005JoeKuoD7Init,
                dim1006JoeKuoD7Init,
                dim1007JoeKuoD7Init,
                dim1008JoeKuoD7Init,
                dim1009JoeKuoD7Init,
                dim1010JoeKuoD7Init,
                dim1011JoeKuoD7Init,
                dim1012JoeKuoD7Init,
                dim1013JoeKuoD7Init,
                dim1014JoeKuoD7Init,
                dim1015JoeKuoD7Init,
                dim1016JoeKuoD7Init,
                dim1017JoeKuoD7Init,
                dim1018JoeKuoD7Init,
                dim1019JoeKuoD7Init,
                dim1020JoeKuoD7Init,
                dim1021JoeKuoD7Init,
                dim1022JoeKuoD7Init,
                dim1023JoeKuoD7Init,
                dim1024JoeKuoD7Init,
                dim1025JoeKuoD7Init,
                dim1026JoeKuoD7Init,
                dim1027JoeKuoD7Init,
                dim1028JoeKuoD7Init,
                dim1029JoeKuoD7Init,
                dim1030JoeKuoD7Init,
                dim1031JoeKuoD7Init,
                dim1032JoeKuoD7Init,
                dim1033JoeKuoD7Init,
                dim1034JoeKuoD7Init,
                dim1035JoeKuoD7Init,
                dim1036JoeKuoD7Init,
                dim1037JoeKuoD7Init,
                dim1038JoeKuoD7Init,
                dim1039JoeKuoD7Init,
                dim1040JoeKuoD7Init,
                dim1041JoeKuoD7Init,
                dim1042JoeKuoD7Init,
                dim1043JoeKuoD7Init,
                dim1044JoeKuoD7Init,
                dim1045JoeKuoD7Init,
                dim1046JoeKuoD7Init,
                dim1047JoeKuoD7Init,
                dim1048JoeKuoD7Init,
                dim1049JoeKuoD7Init,
                dim1050JoeKuoD7Init,
                dim1051JoeKuoD7Init,
                dim1052JoeKuoD7Init,
                dim1053JoeKuoD7Init,
                dim1054JoeKuoD7Init,
                dim1055JoeKuoD7Init,
                dim1056JoeKuoD7Init,
                dim1057JoeKuoD7Init,
                dim1058JoeKuoD7Init,
                dim1059JoeKuoD7Init,
                dim1060JoeKuoD7Init,
                dim1061JoeKuoD7Init,
                dim1062JoeKuoD7Init,
                dim1063JoeKuoD7Init,
                dim1064JoeKuoD7Init,
                dim1065JoeKuoD7Init,
                dim1066JoeKuoD7Init,
                dim1067JoeKuoD7Init,
                dim1068JoeKuoD7Init,
                dim1069JoeKuoD7Init,
                dim1070JoeKuoD7Init,
                dim1071JoeKuoD7Init,
                dim1072JoeKuoD7Init,
                dim1073JoeKuoD7Init,
                dim1074JoeKuoD7Init,
                dim1075JoeKuoD7Init,
                dim1076JoeKuoD7Init,
                dim1077JoeKuoD7Init,
                dim1078JoeKuoD7Init,
                dim1079JoeKuoD7Init,
                dim1080JoeKuoD7Init,
                dim1081JoeKuoD7Init,
                dim1082JoeKuoD7Init,
                dim1083JoeKuoD7Init,
                dim1084JoeKuoD7Init,
                dim1085JoeKuoD7Init,
                dim1086JoeKuoD7Init,
                dim1087JoeKuoD7Init,
                dim1088JoeKuoD7Init,
                dim1089JoeKuoD7Init,
                dim1090JoeKuoD7Init,
                dim1091JoeKuoD7Init,
                dim1092JoeKuoD7Init,
                dim1093JoeKuoD7Init,
                dim1094JoeKuoD7Init,
                dim1095JoeKuoD7Init,
                dim1096JoeKuoD7Init,
                dim1097JoeKuoD7Init,
                dim1098JoeKuoD7Init,
                dim1099JoeKuoD7Init,
                dim1100JoeKuoD7Init,
                dim1101JoeKuoD7Init,
                dim1102JoeKuoD7Init,
                dim1103JoeKuoD7Init,
                dim1104JoeKuoD7Init,
                dim1105JoeKuoD7Init,
                dim1106JoeKuoD7Init,
                dim1107JoeKuoD7Init,
                dim1108JoeKuoD7Init,
                dim1109JoeKuoD7Init,
                dim1110JoeKuoD7Init,
                dim1111JoeKuoD7Init,
                dim1112JoeKuoD7Init,
                dim1113JoeKuoD7Init,
                dim1114JoeKuoD7Init,
                dim1115JoeKuoD7Init,
                dim1116JoeKuoD7Init,
                dim1117JoeKuoD7Init,
                dim1118JoeKuoD7Init,
                dim1119JoeKuoD7Init,
                dim1120JoeKuoD7Init,
                dim1121JoeKuoD7Init,
                dim1122JoeKuoD7Init,
                dim1123JoeKuoD7Init,
                dim1124JoeKuoD7Init,
                dim1125JoeKuoD7Init,
                dim1126JoeKuoD7Init,
                dim1127JoeKuoD7Init,
                dim1128JoeKuoD7Init,
                dim1129JoeKuoD7Init,
                dim1130JoeKuoD7Init,
                dim1131JoeKuoD7Init,
                dim1132JoeKuoD7Init,
                dim1133JoeKuoD7Init,
                dim1134JoeKuoD7Init,
                dim1135JoeKuoD7Init,
                dim1136JoeKuoD7Init,
                dim1137JoeKuoD7Init,
                dim1138JoeKuoD7Init,
                dim1139JoeKuoD7Init,
                dim1140JoeKuoD7Init,
                dim1141JoeKuoD7Init,
                dim1142JoeKuoD7Init,
                dim1143JoeKuoD7Init,
                dim1144JoeKuoD7Init,
                dim1145JoeKuoD7Init,
                dim1146JoeKuoD7Init,
                dim1147JoeKuoD7Init,
                dim1148JoeKuoD7Init,
                dim1149JoeKuoD7Init,
                dim1150JoeKuoD7Init,
                dim1151JoeKuoD7Init,
                dim1152JoeKuoD7Init,
                dim1153JoeKuoD7Init,
                dim1154JoeKuoD7Init,
                dim1155JoeKuoD7Init,
                dim1156JoeKuoD7Init,
                dim1157JoeKuoD7Init,
                dim1158JoeKuoD7Init,
                dim1159JoeKuoD7Init,
                dim1160JoeKuoD7Init,
                dim1161JoeKuoD7Init,
                dim1162JoeKuoD7Init,
                dim1163JoeKuoD7Init,
                dim1164JoeKuoD7Init,
                dim1165JoeKuoD7Init,
                dim1166JoeKuoD7Init,
                dim1167JoeKuoD7Init,
                dim1168JoeKuoD7Init,
                dim1169JoeKuoD7Init,
                dim1170JoeKuoD7Init,
                dim1171JoeKuoD7Init,
                dim1172JoeKuoD7Init,
                dim1173JoeKuoD7Init,
                dim1174JoeKuoD7Init,
                dim1175JoeKuoD7Init,
                dim1176JoeKuoD7Init,
                dim1177JoeKuoD7Init,
                dim1178JoeKuoD7Init,
                dim1179JoeKuoD7Init,
                dim1180JoeKuoD7Init,
                dim1181JoeKuoD7Init,
                dim1182JoeKuoD7Init,
                dim1183JoeKuoD7Init,
                dim1184JoeKuoD7Init,
                dim1185JoeKuoD7Init,
                dim1186JoeKuoD7Init,
                dim1187JoeKuoD7Init,
                dim1188JoeKuoD7Init,
                dim1189JoeKuoD7Init,
                dim1190JoeKuoD7Init,
                dim1191JoeKuoD7Init,
                dim1192JoeKuoD7Init,
                dim1193JoeKuoD7Init,
                dim1194JoeKuoD7Init,
                dim1195JoeKuoD7Init,
                dim1196JoeKuoD7Init,
                dim1197JoeKuoD7Init,
                dim1198JoeKuoD7Init,
                dim1199JoeKuoD7Init,
                dim1200JoeKuoD7Init,
                dim1201JoeKuoD7Init,
                dim1202JoeKuoD7Init,
                dim1203JoeKuoD7Init,
                dim1204JoeKuoD7Init,
                dim1205JoeKuoD7Init,
                dim1206JoeKuoD7Init,
                dim1207JoeKuoD7Init,
                dim1208JoeKuoD7Init,
                dim1209JoeKuoD7Init,
                dim1210JoeKuoD7Init,
                dim1211JoeKuoD7Init,
                dim1212JoeKuoD7Init,
                dim1213JoeKuoD7Init,
                dim1214JoeKuoD7Init,
                dim1215JoeKuoD7Init,
                dim1216JoeKuoD7Init,
                dim1217JoeKuoD7Init,
                dim1218JoeKuoD7Init,
                dim1219JoeKuoD7Init,
                dim1220JoeKuoD7Init,
                dim1221JoeKuoD7Init,
                dim1222JoeKuoD7Init,
                dim1223JoeKuoD7Init,
                dim1224JoeKuoD7Init,
                dim1225JoeKuoD7Init,
                dim1226JoeKuoD7Init,
                dim1227JoeKuoD7Init,
                dim1228JoeKuoD7Init,
                dim1229JoeKuoD7Init,
                dim1230JoeKuoD7Init,
                dim1231JoeKuoD7Init,
                dim1232JoeKuoD7Init,
                dim1233JoeKuoD7Init,
                dim1234JoeKuoD7Init,
                dim1235JoeKuoD7Init,
                dim1236JoeKuoD7Init,
                dim1237JoeKuoD7Init,
                dim1238JoeKuoD7Init,
                dim1239JoeKuoD7Init,
                dim1240JoeKuoD7Init,
                dim1241JoeKuoD7Init,
                dim1242JoeKuoD7Init,
                dim1243JoeKuoD7Init,
                dim1244JoeKuoD7Init,
                dim1245JoeKuoD7Init,
                dim1246JoeKuoD7Init,
                dim1247JoeKuoD7Init,
                dim1248JoeKuoD7Init,
                dim1249JoeKuoD7Init,
                dim1250JoeKuoD7Init,
                dim1251JoeKuoD7Init,
                dim1252JoeKuoD7Init,
                dim1253JoeKuoD7Init,
                dim1254JoeKuoD7Init,
                dim1255JoeKuoD7Init,
                dim1256JoeKuoD7Init,
                dim1257JoeKuoD7Init,
                dim1258JoeKuoD7Init,
                dim1259JoeKuoD7Init,
                dim1260JoeKuoD7Init,
                dim1261JoeKuoD7Init,
                dim1262JoeKuoD7Init,
                dim1263JoeKuoD7Init,
                dim1264JoeKuoD7Init,
                dim1265JoeKuoD7Init,
                dim1266JoeKuoD7Init,
                dim1267JoeKuoD7Init,
                dim1268JoeKuoD7Init,
                dim1269JoeKuoD7Init,
                dim1270JoeKuoD7Init,
                dim1271JoeKuoD7Init,
                dim1272JoeKuoD7Init,
                dim1273JoeKuoD7Init,
                dim1274JoeKuoD7Init,
                dim1275JoeKuoD7Init,
                dim1276JoeKuoD7Init,
                dim1277JoeKuoD7Init,
                dim1278JoeKuoD7Init,
                dim1279JoeKuoD7Init,
                dim1280JoeKuoD7Init,
                dim1281JoeKuoD7Init,
                dim1282JoeKuoD7Init,
                dim1283JoeKuoD7Init,
                dim1284JoeKuoD7Init,
                dim1285JoeKuoD7Init,
                dim1286JoeKuoD7Init,
                dim1287JoeKuoD7Init,
                dim1288JoeKuoD7Init,
                dim1289JoeKuoD7Init,
                dim1290JoeKuoD7Init,
                dim1291JoeKuoD7Init,
                dim1292JoeKuoD7Init,
                dim1293JoeKuoD7Init,
                dim1294JoeKuoD7Init,
                dim1295JoeKuoD7Init,
                dim1296JoeKuoD7Init,
                dim1297JoeKuoD7Init,
                dim1298JoeKuoD7Init,
                dim1299JoeKuoD7Init,
                dim1300JoeKuoD7Init,
                dim1301JoeKuoD7Init,
                dim1302JoeKuoD7Init,
                dim1303JoeKuoD7Init,
                dim1304JoeKuoD7Init,
                dim1305JoeKuoD7Init,
                dim1306JoeKuoD7Init,
                dim1307JoeKuoD7Init,
                dim1308JoeKuoD7Init,
                dim1309JoeKuoD7Init,
                dim1310JoeKuoD7Init,
                dim1311JoeKuoD7Init,
                dim1312JoeKuoD7Init,
                dim1313JoeKuoD7Init,
                dim1314JoeKuoD7Init,
                dim1315JoeKuoD7Init,
                dim1316JoeKuoD7Init,
                dim1317JoeKuoD7Init,
                dim1318JoeKuoD7Init,
                dim1319JoeKuoD7Init,
                dim1320JoeKuoD7Init,
                dim1321JoeKuoD7Init,
                dim1322JoeKuoD7Init,
                dim1323JoeKuoD7Init,
                dim1324JoeKuoD7Init,
                dim1325JoeKuoD7Init,
                dim1326JoeKuoD7Init,
                dim1327JoeKuoD7Init,
                dim1328JoeKuoD7Init,
                dim1329JoeKuoD7Init,
                dim1330JoeKuoD7Init,
                dim1331JoeKuoD7Init,
                dim1332JoeKuoD7Init,
                dim1333JoeKuoD7Init,
                dim1334JoeKuoD7Init,
                dim1335JoeKuoD7Init,
                dim1336JoeKuoD7Init,
                dim1337JoeKuoD7Init,
                dim1338JoeKuoD7Init,
                dim1339JoeKuoD7Init,
                dim1340JoeKuoD7Init,
                dim1341JoeKuoD7Init,
                dim1342JoeKuoD7Init,
                dim1343JoeKuoD7Init,
                dim1344JoeKuoD7Init,
                dim1345JoeKuoD7Init,
                dim1346JoeKuoD7Init,
                dim1347JoeKuoD7Init,
                dim1348JoeKuoD7Init,
                dim1349JoeKuoD7Init,
                dim1350JoeKuoD7Init,
                dim1351JoeKuoD7Init,
                dim1352JoeKuoD7Init,
                dim1353JoeKuoD7Init,
                dim1354JoeKuoD7Init,
                dim1355JoeKuoD7Init,
                dim1356JoeKuoD7Init,
                dim1357JoeKuoD7Init,
                dim1358JoeKuoD7Init,
                dim1359JoeKuoD7Init,
                dim1360JoeKuoD7Init,
                dim1361JoeKuoD7Init,
                dim1362JoeKuoD7Init,
                dim1363JoeKuoD7Init,
                dim1364JoeKuoD7Init,
                dim1365JoeKuoD7Init,
                dim1366JoeKuoD7Init,
                dim1367JoeKuoD7Init,
                dim1368JoeKuoD7Init,
                dim1369JoeKuoD7Init,
                dim1370JoeKuoD7Init,
                dim1371JoeKuoD7Init,
                dim1372JoeKuoD7Init,
                dim1373JoeKuoD7Init,
                dim1374JoeKuoD7Init,
                dim1375JoeKuoD7Init,
                dim1376JoeKuoD7Init,
                dim1377JoeKuoD7Init,
                dim1378JoeKuoD7Init,
                dim1379JoeKuoD7Init,
                dim1380JoeKuoD7Init,
                dim1381JoeKuoD7Init,
                dim1382JoeKuoD7Init,
                dim1383JoeKuoD7Init,
                dim1384JoeKuoD7Init,
                dim1385JoeKuoD7Init,
                dim1386JoeKuoD7Init,
                dim1387JoeKuoD7Init,
                dim1388JoeKuoD7Init,
                dim1389JoeKuoD7Init,
                dim1390JoeKuoD7Init,
                dim1391JoeKuoD7Init,
                dim1392JoeKuoD7Init,
                dim1393JoeKuoD7Init,
                dim1394JoeKuoD7Init,
                dim1395JoeKuoD7Init,
                dim1396JoeKuoD7Init,
                dim1397JoeKuoD7Init,
                dim1398JoeKuoD7Init,
                dim1399JoeKuoD7Init,
                dim1400JoeKuoD7Init,
                dim1401JoeKuoD7Init,
                dim1402JoeKuoD7Init,
                dim1403JoeKuoD7Init,
                dim1404JoeKuoD7Init,
                dim1405JoeKuoD7Init,
                dim1406JoeKuoD7Init,
                dim1407JoeKuoD7Init,
                dim1408JoeKuoD7Init,
                dim1409JoeKuoD7Init,
                dim1410JoeKuoD7Init,
                dim1411JoeKuoD7Init,
                dim1412JoeKuoD7Init,
                dim1413JoeKuoD7Init,
                dim1414JoeKuoD7Init,
                dim1415JoeKuoD7Init,
                dim1416JoeKuoD7Init,
                dim1417JoeKuoD7Init,
                dim1418JoeKuoD7Init,
                dim1419JoeKuoD7Init,
                dim1420JoeKuoD7Init,
                dim1421JoeKuoD7Init,
                dim1422JoeKuoD7Init,
                dim1423JoeKuoD7Init,
                dim1424JoeKuoD7Init,
                dim1425JoeKuoD7Init,
                dim1426JoeKuoD7Init,
                dim1427JoeKuoD7Init,
                dim1428JoeKuoD7Init,
                dim1429JoeKuoD7Init,
                dim1430JoeKuoD7Init,
                dim1431JoeKuoD7Init,
                dim1432JoeKuoD7Init,
                dim1433JoeKuoD7Init,
                dim1434JoeKuoD7Init,
                dim1435JoeKuoD7Init,
                dim1436JoeKuoD7Init,
                dim1437JoeKuoD7Init,
                dim1438JoeKuoD7Init,
                dim1439JoeKuoD7Init,
                dim1440JoeKuoD7Init,
                dim1441JoeKuoD7Init,
                dim1442JoeKuoD7Init,
                dim1443JoeKuoD7Init,
                dim1444JoeKuoD7Init,
                dim1445JoeKuoD7Init,
                dim1446JoeKuoD7Init,
                dim1447JoeKuoD7Init,
                dim1448JoeKuoD7Init,
                dim1449JoeKuoD7Init,
                dim1450JoeKuoD7Init,
                dim1451JoeKuoD7Init,
                dim1452JoeKuoD7Init,
                dim1453JoeKuoD7Init,
                dim1454JoeKuoD7Init,
                dim1455JoeKuoD7Init,
                dim1456JoeKuoD7Init,
                dim1457JoeKuoD7Init,
                dim1458JoeKuoD7Init,
                dim1459JoeKuoD7Init,
                dim1460JoeKuoD7Init,
                dim1461JoeKuoD7Init,
                dim1462JoeKuoD7Init,
                dim1463JoeKuoD7Init,
                dim1464JoeKuoD7Init,
                dim1465JoeKuoD7Init,
                dim1466JoeKuoD7Init,
                dim1467JoeKuoD7Init,
                dim1468JoeKuoD7Init,
                dim1469JoeKuoD7Init,
                dim1470JoeKuoD7Init,
                dim1471JoeKuoD7Init,
                dim1472JoeKuoD7Init,
                dim1473JoeKuoD7Init,
                dim1474JoeKuoD7Init,
                dim1475JoeKuoD7Init,
                dim1476JoeKuoD7Init,
                dim1477JoeKuoD7Init,
                dim1478JoeKuoD7Init,
                dim1479JoeKuoD7Init,
                dim1480JoeKuoD7Init,
                dim1481JoeKuoD7Init,
                dim1482JoeKuoD7Init,
                dim1483JoeKuoD7Init,
                dim1484JoeKuoD7Init,
                dim1485JoeKuoD7Init,
                dim1486JoeKuoD7Init,
                dim1487JoeKuoD7Init,
                dim1488JoeKuoD7Init,
                dim1489JoeKuoD7Init,
                dim1490JoeKuoD7Init,
                dim1491JoeKuoD7Init,
                dim1492JoeKuoD7Init,
                dim1493JoeKuoD7Init,
                dim1494JoeKuoD7Init,
                dim1495JoeKuoD7Init,
                dim1496JoeKuoD7Init,
                dim1497JoeKuoD7Init,
                dim1498JoeKuoD7Init,
                dim1499JoeKuoD7Init,
                dim1500JoeKuoD7Init,
                dim1501JoeKuoD7Init,
                dim1502JoeKuoD7Init,
                dim1503JoeKuoD7Init,
                dim1504JoeKuoD7Init,
                dim1505JoeKuoD7Init,
                dim1506JoeKuoD7Init,
                dim1507JoeKuoD7Init,
                dim1508JoeKuoD7Init,
                dim1509JoeKuoD7Init,
                dim1510JoeKuoD7Init,
                dim1511JoeKuoD7Init,
                dim1512JoeKuoD7Init,
                dim1513JoeKuoD7Init,
                dim1514JoeKuoD7Init,
                dim1515JoeKuoD7Init,
                dim1516JoeKuoD7Init,
                dim1517JoeKuoD7Init,
                dim1518JoeKuoD7Init,
                dim1519JoeKuoD7Init,
                dim1520JoeKuoD7Init,
                dim1521JoeKuoD7Init,
                dim1522JoeKuoD7Init,
                dim1523JoeKuoD7Init,
                dim1524JoeKuoD7Init,
                dim1525JoeKuoD7Init,
                dim1526JoeKuoD7Init,
                dim1527JoeKuoD7Init,
                dim1528JoeKuoD7Init,
                dim1529JoeKuoD7Init,
                dim1530JoeKuoD7Init,
                dim1531JoeKuoD7Init,
                dim1532JoeKuoD7Init,
                dim1533JoeKuoD7Init,
                dim1534JoeKuoD7Init,
                dim1535JoeKuoD7Init,
                dim1536JoeKuoD7Init,
                dim1537JoeKuoD7Init,
                dim1538JoeKuoD7Init,
                dim1539JoeKuoD7Init,
                dim1540JoeKuoD7Init,
                dim1541JoeKuoD7Init,
                dim1542JoeKuoD7Init,
                dim1543JoeKuoD7Init,
                dim1544JoeKuoD7Init,
                dim1545JoeKuoD7Init,
                dim1546JoeKuoD7Init,
                dim1547JoeKuoD7Init,
                dim1548JoeKuoD7Init,
                dim1549JoeKuoD7Init,
                dim1550JoeKuoD7Init,
                dim1551JoeKuoD7Init,
                dim1552JoeKuoD7Init,
                dim1553JoeKuoD7Init,
                dim1554JoeKuoD7Init,
                dim1555JoeKuoD7Init,
                dim1556JoeKuoD7Init,
                dim1557JoeKuoD7Init,
                dim1558JoeKuoD7Init,
                dim1559JoeKuoD7Init,
                dim1560JoeKuoD7Init,
                dim1561JoeKuoD7Init,
                dim1562JoeKuoD7Init,
                dim1563JoeKuoD7Init,
                dim1564JoeKuoD7Init,
                dim1565JoeKuoD7Init,
                dim1566JoeKuoD7Init,
                dim1567JoeKuoD7Init,
                dim1568JoeKuoD7Init,
                dim1569JoeKuoD7Init,
                dim1570JoeKuoD7Init,
                dim1571JoeKuoD7Init,
                dim1572JoeKuoD7Init,
                dim1573JoeKuoD7Init,
                dim1574JoeKuoD7Init,
                dim1575JoeKuoD7Init,
                dim1576JoeKuoD7Init,
                dim1577JoeKuoD7Init,
                dim1578JoeKuoD7Init,
                dim1579JoeKuoD7Init,
                dim1580JoeKuoD7Init,
                dim1581JoeKuoD7Init,
                dim1582JoeKuoD7Init,
                dim1583JoeKuoD7Init,
                dim1584JoeKuoD7Init,
                dim1585JoeKuoD7Init,
                dim1586JoeKuoD7Init,
                dim1587JoeKuoD7Init,
                dim1588JoeKuoD7Init,
                dim1589JoeKuoD7Init,
                dim1590JoeKuoD7Init,
                dim1591JoeKuoD7Init,
                dim1592JoeKuoD7Init,
                dim1593JoeKuoD7Init,
                dim1594JoeKuoD7Init,
                dim1595JoeKuoD7Init,
                dim1596JoeKuoD7Init,
                dim1597JoeKuoD7Init,
                dim1598JoeKuoD7Init,
                dim1599JoeKuoD7Init,
                dim1600JoeKuoD7Init,
                dim1601JoeKuoD7Init,
                dim1602JoeKuoD7Init,
                dim1603JoeKuoD7Init,
                dim1604JoeKuoD7Init,
                dim1605JoeKuoD7Init,
                dim1606JoeKuoD7Init,
                dim1607JoeKuoD7Init,
                dim1608JoeKuoD7Init,
                dim1609JoeKuoD7Init,
                dim1610JoeKuoD7Init,
                dim1611JoeKuoD7Init,
                dim1612JoeKuoD7Init,
                dim1613JoeKuoD7Init,
                dim1614JoeKuoD7Init,
                dim1615JoeKuoD7Init,
                dim1616JoeKuoD7Init,
                dim1617JoeKuoD7Init,
                dim1618JoeKuoD7Init,
                dim1619JoeKuoD7Init,
                dim1620JoeKuoD7Init,
                dim1621JoeKuoD7Init,
                dim1622JoeKuoD7Init,
                dim1623JoeKuoD7Init,
                dim1624JoeKuoD7Init,
                dim1625JoeKuoD7Init,
                dim1626JoeKuoD7Init,
                dim1627JoeKuoD7Init,
                dim1628JoeKuoD7Init,
                dim1629JoeKuoD7Init,
                dim1630JoeKuoD7Init,
                dim1631JoeKuoD7Init,
                dim1632JoeKuoD7Init,
                dim1633JoeKuoD7Init,
                dim1634JoeKuoD7Init,
                dim1635JoeKuoD7Init,
                dim1636JoeKuoD7Init,
                dim1637JoeKuoD7Init,
                dim1638JoeKuoD7Init,
                dim1639JoeKuoD7Init,
                dim1640JoeKuoD7Init,
                dim1641JoeKuoD7Init,
                dim1642JoeKuoD7Init,
                dim1643JoeKuoD7Init,
                dim1644JoeKuoD7Init,
                dim1645JoeKuoD7Init,
                dim1646JoeKuoD7Init,
                dim1647JoeKuoD7Init,
                dim1648JoeKuoD7Init,
                dim1649JoeKuoD7Init,
                dim1650JoeKuoD7Init,
                dim1651JoeKuoD7Init,
                dim1652JoeKuoD7Init,
                dim1653JoeKuoD7Init,
                dim1654JoeKuoD7Init,
                dim1655JoeKuoD7Init,
                dim1656JoeKuoD7Init,
                dim1657JoeKuoD7Init,
                dim1658JoeKuoD7Init,
                dim1659JoeKuoD7Init,
                dim1660JoeKuoD7Init,
                dim1661JoeKuoD7Init,
                dim1662JoeKuoD7Init,
                dim1663JoeKuoD7Init,
                dim1664JoeKuoD7Init,
                dim1665JoeKuoD7Init,
                dim1666JoeKuoD7Init,
                dim1667JoeKuoD7Init,
                dim1668JoeKuoD7Init,
                dim1669JoeKuoD7Init,
                dim1670JoeKuoD7Init,
                dim1671JoeKuoD7Init,
                dim1672JoeKuoD7Init,
                dim1673JoeKuoD7Init,
                dim1674JoeKuoD7Init,
                dim1675JoeKuoD7Init,
                dim1676JoeKuoD7Init,
                dim1677JoeKuoD7Init,
                dim1678JoeKuoD7Init,
                dim1679JoeKuoD7Init,
                dim1680JoeKuoD7Init,
                dim1681JoeKuoD7Init,
                dim1682JoeKuoD7Init,
                dim1683JoeKuoD7Init,
                dim1684JoeKuoD7Init,
                dim1685JoeKuoD7Init,
                dim1686JoeKuoD7Init,
                dim1687JoeKuoD7Init,
                dim1688JoeKuoD7Init,
                dim1689JoeKuoD7Init,
                dim1690JoeKuoD7Init,
                dim1691JoeKuoD7Init,
                dim1692JoeKuoD7Init,
                dim1693JoeKuoD7Init,
                dim1694JoeKuoD7Init,
                dim1695JoeKuoD7Init,
                dim1696JoeKuoD7Init,
                dim1697JoeKuoD7Init,
                dim1698JoeKuoD7Init,
                dim1699JoeKuoD7Init,
                dim1700JoeKuoD7Init,
                dim1701JoeKuoD7Init,
                dim1702JoeKuoD7Init,
                dim1703JoeKuoD7Init,
                dim1704JoeKuoD7Init,
                dim1705JoeKuoD7Init,
                dim1706JoeKuoD7Init,
                dim1707JoeKuoD7Init,
                dim1708JoeKuoD7Init,
                dim1709JoeKuoD7Init,
                dim1710JoeKuoD7Init,
                dim1711JoeKuoD7Init,
                dim1712JoeKuoD7Init,
                dim1713JoeKuoD7Init,
                dim1714JoeKuoD7Init,
                dim1715JoeKuoD7Init,
                dim1716JoeKuoD7Init,
                dim1717JoeKuoD7Init,
                dim1718JoeKuoD7Init,
                dim1719JoeKuoD7Init,
                dim1720JoeKuoD7Init,
                dim1721JoeKuoD7Init,
                dim1722JoeKuoD7Init,
                dim1723JoeKuoD7Init,
                dim1724JoeKuoD7Init,
                dim1725JoeKuoD7Init,
                dim1726JoeKuoD7Init,
                dim1727JoeKuoD7Init,
                dim1728JoeKuoD7Init,
                dim1729JoeKuoD7Init,
                dim1730JoeKuoD7Init,
                dim1731JoeKuoD7Init,
                dim1732JoeKuoD7Init,
                dim1733JoeKuoD7Init,
                dim1734JoeKuoD7Init,
                dim1735JoeKuoD7Init,
                dim1736JoeKuoD7Init,
                dim1737JoeKuoD7Init,
                dim1738JoeKuoD7Init,
                dim1739JoeKuoD7Init,
                dim1740JoeKuoD7Init,
                dim1741JoeKuoD7Init,
                dim1742JoeKuoD7Init,
                dim1743JoeKuoD7Init,
                dim1744JoeKuoD7Init,
                dim1745JoeKuoD7Init,
                dim1746JoeKuoD7Init,
                dim1747JoeKuoD7Init,
                dim1748JoeKuoD7Init,
                dim1749JoeKuoD7Init,
                dim1750JoeKuoD7Init,
                dim1751JoeKuoD7Init,
                dim1752JoeKuoD7Init,
                dim1753JoeKuoD7Init,
                dim1754JoeKuoD7Init,
                dim1755JoeKuoD7Init,
                dim1756JoeKuoD7Init,
                dim1757JoeKuoD7Init,
                dim1758JoeKuoD7Init,
                dim1759JoeKuoD7Init,
                dim1760JoeKuoD7Init,
                dim1761JoeKuoD7Init,
                dim1762JoeKuoD7Init,
                dim1763JoeKuoD7Init,
                dim1764JoeKuoD7Init,
                dim1765JoeKuoD7Init,
                dim1766JoeKuoD7Init,
                dim1767JoeKuoD7Init,
                dim1768JoeKuoD7Init,
                dim1769JoeKuoD7Init,
                dim1770JoeKuoD7Init,
                dim1771JoeKuoD7Init,
                dim1772JoeKuoD7Init,
                dim1773JoeKuoD7Init,
                dim1774JoeKuoD7Init,
                dim1775JoeKuoD7Init,
                dim1776JoeKuoD7Init,
                dim1777JoeKuoD7Init,
                dim1778JoeKuoD7Init,
                dim1779JoeKuoD7Init,
                dim1780JoeKuoD7Init,
                dim1781JoeKuoD7Init,
                dim1782JoeKuoD7Init,
                dim1783JoeKuoD7Init,
                dim1784JoeKuoD7Init,
                dim1785JoeKuoD7Init,
                dim1786JoeKuoD7Init,
                dim1787JoeKuoD7Init,
                dim1788JoeKuoD7Init,
                dim1789JoeKuoD7Init,
                dim1790JoeKuoD7Init,
                dim1791JoeKuoD7Init,
                dim1792JoeKuoD7Init,
                dim1793JoeKuoD7Init,
                dim1794JoeKuoD7Init,
                dim1795JoeKuoD7Init,
                dim1796JoeKuoD7Init,
                dim1797JoeKuoD7Init,
                dim1798JoeKuoD7Init,
                dim1799JoeKuoD7Init,
                dim1800JoeKuoD7Init,
                dim1801JoeKuoD7Init,
                dim1802JoeKuoD7Init,
                dim1803JoeKuoD7Init,
                dim1804JoeKuoD7Init,
                dim1805JoeKuoD7Init,
                dim1806JoeKuoD7Init,
                dim1807JoeKuoD7Init,
                dim1808JoeKuoD7Init,
                dim1809JoeKuoD7Init,
                dim1810JoeKuoD7Init,
                dim1811JoeKuoD7Init,
                dim1812JoeKuoD7Init,
                dim1813JoeKuoD7Init,
                dim1814JoeKuoD7Init,
                dim1815JoeKuoD7Init,
                dim1816JoeKuoD7Init,
                dim1817JoeKuoD7Init,
                dim1818JoeKuoD7Init,
                dim1819JoeKuoD7Init,
                dim1820JoeKuoD7Init,
                dim1821JoeKuoD7Init,
                dim1822JoeKuoD7Init,
                dim1823JoeKuoD7Init,
                dim1824JoeKuoD7Init,
                dim1825JoeKuoD7Init,
                dim1826JoeKuoD7Init,
                dim1827JoeKuoD7Init,
                dim1828JoeKuoD7Init,
                dim1829JoeKuoD7Init,
                dim1830JoeKuoD7Init,
                dim1831JoeKuoD7Init,
                dim1832JoeKuoD7Init,
                dim1833JoeKuoD7Init,
                dim1834JoeKuoD7Init,
                dim1835JoeKuoD7Init,
                dim1836JoeKuoD7Init,
                dim1837JoeKuoD7Init,
                dim1838JoeKuoD7Init,
                dim1839JoeKuoD7Init,
                dim1840JoeKuoD7Init,
                dim1841JoeKuoD7Init,
                dim1842JoeKuoD7Init,
                dim1843JoeKuoD7Init,
                dim1844JoeKuoD7Init,
                dim1845JoeKuoD7Init,
                dim1846JoeKuoD7Init,
                dim1847JoeKuoD7Init,
                dim1848JoeKuoD7Init,
                dim1849JoeKuoD7Init,
                dim1850JoeKuoD7Init,
                dim1851JoeKuoD7Init,
                dim1852JoeKuoD7Init,
                dim1853JoeKuoD7Init,
                dim1854JoeKuoD7Init,
                dim1855JoeKuoD7Init,
                dim1856JoeKuoD7Init,
                dim1857JoeKuoD7Init,
                dim1858JoeKuoD7Init,
                dim1859JoeKuoD7Init,
                dim1860JoeKuoD7Init,
                dim1861JoeKuoD7Init,
                dim1862JoeKuoD7Init,
                dim1863JoeKuoD7Init,
                dim1864JoeKuoD7Init,
                dim1865JoeKuoD7Init,
                dim1866JoeKuoD7Init,
                dim1867JoeKuoD7Init,
                dim1868JoeKuoD7Init,
                dim1869JoeKuoD7Init,
                dim1870JoeKuoD7Init,
                dim1871JoeKuoD7Init,
                dim1872JoeKuoD7Init,
                dim1873JoeKuoD7Init,
                dim1874JoeKuoD7Init,
                dim1875JoeKuoD7Init,
                dim1876JoeKuoD7Init,
                dim1877JoeKuoD7Init,
                dim1878JoeKuoD7Init,
                dim1879JoeKuoD7Init,
                dim1880JoeKuoD7Init,
                dim1881JoeKuoD7Init,
                dim1882JoeKuoD7Init,
                dim1883JoeKuoD7Init,
                dim1884JoeKuoD7Init,
                dim1885JoeKuoD7Init,
                dim1886JoeKuoD7Init,
                dim1887JoeKuoD7Init,
                dim1888JoeKuoD7Init,
                dim1889JoeKuoD7Init,
                dim1890JoeKuoD7Init,
                dim1891JoeKuoD7Init,
                dim1892JoeKuoD7Init,
                dim1893JoeKuoD7Init,
                dim1894JoeKuoD7Init,
                dim1895JoeKuoD7Init,
                dim1896JoeKuoD7Init,
                dim1897JoeKuoD7Init,
                dim1898JoeKuoD7Init,
                dim1899JoeKuoD7Init
            };

            const unsigned long dim1JoeKuoD5Init[]  =   {   1   ,0 };
            const unsigned long dim2JoeKuoD5Init[]  =   {   1   ,   3   ,0 };
            const unsigned long dim3JoeKuoD5Init[]  =   {   1   ,   3   ,   1   ,0 };
            const unsigned long dim4JoeKuoD5Init[]  =   {   1   ,   1   ,   1   ,0 };
            const unsigned long dim5JoeKuoD5Init[]  =   {   1   ,   1   ,   3   ,   3   ,0 };
            const unsigned long dim6JoeKuoD5Init[]  =   {   1   ,   3   ,   5   ,   13  ,0 };
            const unsigned long dim7JoeKuoD5Init[]  =   {   1   ,   1   ,   5   ,   5   ,   17  ,0 };
            const unsigned long dim8JoeKuoD5Init[]  =   {   1   ,   1   ,   5   ,   5   ,   5   ,0 };
            const unsigned long dim9JoeKuoD5Init[]  =   {   1   ,   1   ,   7   ,   11  ,   19  ,0 };
            const unsigned long dim10JoeKuoD5Init[] =   {   1   ,   1   ,   5   ,   1   ,   1   ,0 };
            const unsigned long dim11JoeKuoD5Init[] =   {   1   ,   3   ,   7   ,   1   ,   19  ,0 };
            const unsigned long dim12JoeKuoD5Init[] =   {   1   ,   3   ,   3   ,   5   ,   7   ,0 };
            const unsigned long dim13JoeKuoD5Init[] =   {   1   ,   3   ,   3   ,   13  ,   9   ,   53  ,0 };
            const unsigned long dim14JoeKuoD5Init[] =   {   1   ,   1   ,   5   ,   11  ,   1   ,   1   ,0 };
            const unsigned long dim15JoeKuoD5Init[] =   {   1   ,   1   ,   3   ,   7   ,   21  ,   51  ,0 };
            const unsigned long dim16JoeKuoD5Init[] =   {   1   ,   1   ,   1   ,   15  ,   1   ,   5   ,0 };
            const unsigned long dim17JoeKuoD5Init[] =   {   1   ,   3   ,   1   ,   9   ,   9   ,   1   ,0 };
            const unsigned long dim18JoeKuoD5Init[] =   {   1   ,   1   ,   5   ,   5   ,   17  ,   61  ,0 };
            const unsigned long dim19JoeKuoD5Init[] =   {   1   ,   3   ,   1   ,   15  ,   29  ,   57  ,   87  ,0 };
            const unsigned long dim20JoeKuoD5Init[] =   {   1   ,   3   ,   5   ,   15  ,   3   ,   11  ,   17  ,0 };
            const unsigned long dim21JoeKuoD5Init[] =   {   1   ,   3   ,   3   ,   7   ,   5   ,   17  ,   65  ,0 };
            const unsigned long dim22JoeKuoD5Init[] =   {   1   ,   3   ,   5   ,   1   ,   25  ,   29  ,   49  ,0 };
            const unsigned long dim23JoeKuoD5Init[] =   {   1   ,   1   ,   3   ,   7   ,   15  ,   39  ,   119 ,0 };
            const unsigned long dim24JoeKuoD5Init[] =   {   1   ,   3   ,   3   ,   5   ,   19  ,   51  ,   61  ,0 };
            const unsigned long dim25JoeKuoD5Init[] =   {   1   ,   1   ,   5   ,   15  ,   11  ,   47  ,   15  ,0 };
            const unsigned long dim26JoeKuoD5Init[] =   {   1   ,   1   ,   7   ,   3   ,   29  ,   51  ,   51  ,0 };
            const unsigned long dim27JoeKuoD5Init[] =   {   1   ,   1   ,   3   ,   15  ,   19  ,   17  ,   13  ,0 };
            const unsigned long dim28JoeKuoD5Init[] =   {   1   ,   3   ,   7   ,   3   ,   17  ,   9   ,   93  ,0 };
            const unsigned long dim29JoeKuoD5Init[] =   {   1   ,   3   ,   7   ,   5   ,   7   ,   29  ,   111 ,0 };
            const unsigned long dim30JoeKuoD5Init[] =   {   1   ,   1   ,   7   ,   9   ,   25  ,   19  ,   105 ,0 };
            const unsigned long dim31JoeKuoD5Init[] =   {   1   ,   1   ,   1   ,   11  ,   21  ,   35  ,   107 ,0 };
            const unsigned long dim32JoeKuoD5Init[] =   {   1   ,   1   ,   5   ,   11  ,   19  ,   53  ,   25  ,0 };
            const unsigned long dim33JoeKuoD5Init[] =   {   1   ,   3   ,   1   ,   3   ,   27  ,   29  ,   31  ,0 };
            const unsigned long dim34JoeKuoD5Init[] =   {   1   ,   1   ,   5   ,   13  ,   27  ,   19  ,   61  ,0 };
            const unsigned long dim35JoeKuoD5Init[] =   {   1   ,   3   ,   1   ,   3   ,   25  ,   33  ,   105 ,0 };
            const unsigned long dim36JoeKuoD5Init[] =   {   1   ,   3   ,   7   ,   11  ,   27  ,   55  ,   1   ,0 };
            const unsigned long dim37JoeKuoD5Init[] =   {   1   ,   1   ,   7   ,   1   ,   9   ,   45  ,   97  ,   63  ,0 };
            const unsigned long dim38JoeKuoD5Init[] =   {   1   ,   1   ,   7   ,   9   ,   3   ,   17  ,   85  ,   213 ,0 };
            const unsigned long dim39JoeKuoD5Init[] =   {   1   ,   1   ,   1   ,   3   ,   31  ,   35  ,   93  ,   35  ,0 };
            const unsigned long dim40JoeKuoD5Init[] =   {   1   ,   3   ,   5   ,   9   ,   1   ,   63  ,   117 ,   35  ,0 };
            const unsigned long dim41JoeKuoD5Init[] =   {   1   ,   3   ,   1   ,   9   ,   21  ,   3   ,   53  ,   29  ,0 };
            const unsigned long dim42JoeKuoD5Init[] =   {   1   ,   3   ,   1   ,   9   ,   29  ,   33  ,   43  ,   181 ,0 };
            const unsigned long dim43JoeKuoD5Init[] =   {   1   ,   3   ,   7   ,   3   ,   21  ,   45  ,   121 ,   141 ,0 };
            const unsigned long dim44JoeKuoD5Init[] =   {   1   ,   1   ,   1   ,   13  ,   5   ,   49  ,   45  ,   77  ,0 };
            const unsigned long dim45JoeKuoD5Init[] =   {   1   ,   1   ,   3   ,   3   ,   1   ,   47  ,   37  ,   151 ,0 };
            const unsigned long dim46JoeKuoD5Init[] =   {   1   ,   3   ,   7   ,   5   ,   9   ,   51  ,   61  ,   95  ,0 };
            const unsigned long dim47JoeKuoD5Init[] =   {   1   ,   1   ,   1   ,   7   ,   31  ,   23  ,   81  ,   105 ,0 };
            const unsigned long dim48JoeKuoD5Init[] =   {   1   ,   3   ,   5   ,   15  ,   15  ,   9   ,   115 ,   55  ,0 };
            const unsigned long dim49JoeKuoD5Init[] =   {   1   ,   3   ,   3   ,   13  ,   15  ,   1   ,   87  ,   11  ,0 };
            const unsigned long dim50JoeKuoD5Init[] =   {   1   ,   3   ,   5   ,   1   ,   5   ,   9   ,   29  ,   241 ,0 };
            const unsigned long dim51JoeKuoD5Init[] =   {   1   ,   1   ,   1   ,   9   ,   19  ,   5   ,   115 ,   191 ,0 };
            const unsigned long dim52JoeKuoD5Init[] =   {   1   ,   1   ,   1   ,   15  ,   1   ,   57  ,   107 ,   49  ,0 };
            const unsigned long dim53JoeKuoD5Init[] =   {   1   ,   1   ,   7   ,   7   ,   23  ,   21  ,   71  ,   187 ,   207 ,0 };
            const unsigned long dim54JoeKuoD5Init[] =   {   1   ,   3   ,   3   ,   5   ,   11  ,   35  ,   101 ,   7   ,   501 ,0 };
            const unsigned long dim55JoeKuoD5Init[] =   {   1   ,   3   ,   5   ,   15  ,   29  ,   5   ,   61  ,   205 ,   301 ,0 };
            const unsigned long dim56JoeKuoD5Init[] =   {   1   ,   1   ,   7   ,   13  ,   7   ,   39  ,   127 ,   243 ,   307 ,0 };
            const unsigned long dim57JoeKuoD5Init[] =   {   1   ,   3   ,   7   ,   13  ,   29  ,   9   ,   93  ,   187 ,   429 ,0 };
            const unsigned long dim58JoeKuoD5Init[] =   {   1   ,   3   ,   3   ,   11  ,   15  ,   35  ,   85  ,   159 ,   223 ,0 };
            const unsigned long dim59JoeKuoD5Init[] =   {   1   ,   1   ,   3   ,   1   ,   13  ,   3   ,   111 ,   17  ,   411 ,0 };
            const unsigned long dim60JoeKuoD5Init[] =   {   1   ,   1   ,   1   ,   7   ,   31  ,   21  ,   103 ,   175 ,   97  ,0 };
            const unsigned long dim61JoeKuoD5Init[] =   {   1   ,   1   ,   1   ,   15  ,   11  ,   21  ,   63  ,   45  ,   29  ,0 };
            const unsigned long dim62JoeKuoD5Init[] =   {   1   ,   3   ,   5   ,   3   ,   13  ,   45  ,   53  ,   191 ,   455 ,0 };
            const unsigned long dim63JoeKuoD5Init[] =   {   1   ,   3   ,   3   ,   13  ,   11  ,   37  ,   65  ,   45  ,   371 ,0 };
            const unsigned long dim64JoeKuoD5Init[] =   {   1   ,   1   ,   1   ,   15  ,   23  ,   9   ,   123 ,   97  ,   497 ,0 };
            const unsigned long dim65JoeKuoD5Init[] =   {   1   ,   1   ,   7   ,   7   ,   5   ,   13  ,   33  ,   169 ,   411 ,0 };
            const unsigned long dim66JoeKuoD5Init[] =   {   1   ,   1   ,   3   ,   13  ,   29  ,   61  ,   67  ,   1   ,   167 ,0 };
            const unsigned long dim67JoeKuoD5Init[] =   {   1   ,   1   ,   7   ,   3   ,   11  ,   21  ,   25  ,   87  ,   507 ,0 };
            const unsigned long dim68JoeKuoD5Init[] =   {   1   ,   3   ,   1   ,   9   ,   31  ,   37  ,   3   ,   89  ,   113 ,0 };
            const unsigned long dim69JoeKuoD5Init[] =   {   1   ,   1   ,   5   ,   11  ,   5   ,   11  ,   83  ,   85  ,   421 ,0 };
            const unsigned long dim70JoeKuoD5Init[] =   {   1   ,   3   ,   5   ,   7   ,   23  ,   9   ,   111 ,   135 ,   337 ,0 };
            const unsigned long dim71JoeKuoD5Init[] =   {   1   ,   3   ,   5   ,   15  ,   3   ,   39  ,   81  ,   249 ,   363 ,0 };
            const unsigned long dim72JoeKuoD5Init[] =   {   1   ,   1   ,   7   ,   15  ,   9   ,   49  ,   79  ,   103 ,   19  ,0 };
            const unsigned long dim73JoeKuoD5Init[] =   {   1   ,   3   ,   1   ,   1   ,   17  ,   31  ,   45  ,   205 ,   381 ,0 };
            const unsigned long dim74JoeKuoD5Init[] =   {   1   ,   1   ,   1   ,   1   ,   11  ,   45  ,   89  ,   1   ,   365 ,0 };
            const unsigned long dim75JoeKuoD5Init[] =   {   1   ,   1   ,   1   ,   15  ,   7   ,   63  ,   55  ,   185 ,   373 ,0 };
            const unsigned long dim76JoeKuoD5Init[] =   {   1   ,   1   ,   1   ,   3   ,   17  ,   19  ,   23  ,   7   ,   265 ,0 };
            const unsigned long dim77JoeKuoD5Init[] =   {   1   ,   1   ,   3   ,   15  ,   13  ,   31  ,   53  ,   235 ,   309 ,0 };
            const unsigned long dim78JoeKuoD5Init[] =   {   1   ,   3   ,   1   ,   5   ,   1   ,   63  ,   73  ,   155 ,   33  ,0 };
            const unsigned long dim79JoeKuoD5Init[] =   {   1   ,   1   ,   1   ,   15  ,   19  ,   57  ,   21  ,   45  ,   203 ,0 };
            const unsigned long dim80JoeKuoD5Init[] =   {   1   ,   3   ,   5   ,   9   ,   13  ,   27  ,   55  ,   215 ,   181 ,0 };
            const unsigned long dim81JoeKuoD5Init[] =   {   1   ,   1   ,   7   ,   3   ,   11  ,   39  ,   55  ,   219 ,   401 ,0 };
            const unsigned long dim82JoeKuoD5Init[] =   {   1   ,   1   ,   1   ,   7   ,   9   ,   13  ,   3   ,   181 ,   395 ,0 };
            const unsigned long dim83JoeKuoD5Init[] =   {   1   ,   3   ,   1   ,   15  ,   13  ,   19  ,   23  ,   145 ,   97  ,0 };
            const unsigned long dim84JoeKuoD5Init[] =   {   1   ,   1   ,   1   ,   15  ,   23  ,   15  ,   55  ,   3   ,   243 ,0 };
            const unsigned long dim85JoeKuoD5Init[] =   {   1   ,   1   ,   3   ,   15  ,   15  ,   5   ,   115 ,   169 ,   415 ,0 };
            const unsigned long dim86JoeKuoD5Init[] =   {   1   ,   1   ,   7   ,   15  ,   31  ,   15  ,   119 ,   89  ,   37  ,0 };
            const unsigned long dim87JoeKuoD5Init[] =   {   1   ,   1   ,   5   ,   3   ,   25  ,   23  ,   23  ,   133 ,   467 ,0 };
            const unsigned long dim88JoeKuoD5Init[] =   {   1   ,   3   ,   5   ,   15  ,   11  ,   7   ,   73  ,   209 ,   331 ,0 };
            const unsigned long dim89JoeKuoD5Init[] =   {   1   ,   3   ,   7   ,   11  ,   29  ,   51  ,   119 ,   105 ,   63  ,0 };
            const unsigned long dim90JoeKuoD5Init[] =   {   1   ,   1   ,   5   ,   9   ,   19  ,   21  ,   53  ,   211 ,   231 ,0 };
            const unsigned long dim91JoeKuoD5Init[] =   {   1   ,   3   ,   5   ,   1   ,   7   ,   23  ,   27  ,   67  ,   445 ,0 };
            const unsigned long dim92JoeKuoD5Init[] =   {   1   ,   1   ,   5   ,   15  ,   21  ,   53  ,   117 ,   229 ,   13  ,0 };
            const unsigned long dim93JoeKuoD5Init[] =   {   1   ,   3   ,   5   ,   11  ,   19  ,   55  ,   53  ,   13  ,   349 ,0 };
            const unsigned long dim94JoeKuoD5Init[] =   {   1   ,   3   ,   1   ,   9   ,   19  ,   25  ,   79  ,   55  ,   355 ,0 };
            const unsigned long dim95JoeKuoD5Init[] =   {   1   ,   3   ,   3   ,   5   ,   31  ,   27  ,   105 ,   63  ,   21  ,0 };
            const unsigned long dim96JoeKuoD5Init[] =   {   1   ,   3   ,   3   ,   15  ,   13  ,   37  ,   39  ,   13  ,   459 ,0 };
            const unsigned long dim97JoeKuoD5Init[] =   {   1   ,   1   ,   1   ,   7   ,   25  ,   61  ,   93  ,   195 ,   441 ,0 };
            const unsigned long dim98JoeKuoD5Init[] =   {   1   ,   3   ,   3   ,   3   ,   7   ,   35  ,   55  ,   103 ,   159 ,0 };
            const unsigned long dim99JoeKuoD5Init[] =   {   1   ,   1   ,   7   ,   11  ,   7   ,   33  ,   49  ,   113 ,   331 ,0 };
            const unsigned long dim100JoeKuoD5Init[]    =   {   1   ,   3   ,   1   ,   1   ,   1   ,   31  ,   35  ,   63  ,   465 ,0 };
            const unsigned long dim101JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   1   ,   13  ,   43  ,   83  ,   177 ,   461 ,   747 ,0 };
            const unsigned long dim102JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   3   ,   21  ,   5   ,   19  ,   135 ,   483 ,   181 ,0 };
            const unsigned long dim103JoeKuoD5Init[]    =   {   1   ,   3   ,   1   ,   9   ,   25  ,   3   ,   37  ,   147 ,   483 ,   743 ,0 };
            const unsigned long dim104JoeKuoD5Init[]    =   {   1   ,   1   ,   5   ,   3   ,   1   ,   1   ,   101 ,   163 ,   165 ,   957 ,0 };
            const unsigned long dim105JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   1   ,   15  ,   41  ,   117 ,   7   ,   71  ,   357 ,0 };
            const unsigned long dim106JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   9   ,   27  ,   11  ,   55  ,   5   ,   11  ,   863 ,0 };
            const unsigned long dim107JoeKuoD5Init[]    =   {   1   ,   1   ,   1   ,   1   ,   27  ,   43  ,   51  ,   211 ,   265 ,   403 ,0 };
            const unsigned long dim108JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   3   ,   31  ,   35  ,   61  ,   43  ,   223 ,   441 ,0 };
            const unsigned long dim109JoeKuoD5Init[]    =   {   1   ,   3   ,   1   ,   7   ,   19  ,   61  ,   59  ,   63  ,   401 ,   767 ,0 };
            const unsigned long dim110JoeKuoD5Init[]    =   {   1   ,   3   ,   1   ,   9   ,   23  ,   39  ,   83  ,   249 ,   129 ,   843 ,0 };
            const unsigned long dim111JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   13  ,   1   ,   49  ,   61  ,   115 ,   289 ,   85  ,0 };
            const unsigned long dim112JoeKuoD5Init[]    =   {   1   ,   1   ,   5   ,   13  ,   7   ,   17  ,   35  ,   95  ,   235 ,   49  ,0 };
            const unsigned long dim113JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   11  ,   9   ,   9   ,   91  ,   141 ,   305 ,   955 ,0 };
            const unsigned long dim114JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   11  ,   17  ,   3   ,   77  ,   95  ,   507 ,   627 ,0 };
            const unsigned long dim115JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   1   ,   31  ,   43  ,   31  ,   217 ,   67  ,   853 ,0 };
            const unsigned long dim116JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   15  ,   17  ,   33  ,   31  ,   91  ,   465 ,   209 ,0 };
            const unsigned long dim117JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   7   ,   31  ,   41  ,   101 ,   95  ,   431 ,   203 ,0 };
            const unsigned long dim118JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   1   ,   23  ,   45  ,   123 ,   111 ,   457 ,   655 ,0 };
            const unsigned long dim119JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   5   ,   9   ,   27  ,   37  ,   195 ,   11  ,   99  ,0 };
            const unsigned long dim120JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   9   ,   7   ,   39  ,   81  ,   171 ,   97  ,   775 ,0 };
            const unsigned long dim121JoeKuoD5Init[]    =   {   1   ,   1   ,   1   ,   3   ,   3   ,   27  ,   93  ,   149 ,   321 ,   537 ,0 };
            const unsigned long dim122JoeKuoD5Init[]    =   {   1   ,   3   ,   1   ,   3   ,   13  ,   3   ,   59  ,   137 ,   505 ,   395 ,0 };
            const unsigned long dim123JoeKuoD5Init[]    =   {   1   ,   1   ,   1   ,   11  ,   15  ,   53  ,   77  ,   235 ,   439 ,   829 ,0 };
            const unsigned long dim124JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   1   ,   25  ,   61  ,   65  ,   53  ,   207 ,   891 ,0 };
            const unsigned long dim125JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   15  ,   29  ,   39  ,   67  ,   203 ,   495 ,   795 ,0 };
            const unsigned long dim126JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   1   ,   15  ,   17  ,   119 ,   83  ,   411 ,   1015    ,0 };
            const unsigned long dim127JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   1   ,   13  ,   37  ,   69  ,   147 ,   141 ,   229 ,0 };
            const unsigned long dim128JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   1   ,   11  ,   31  ,   73  ,   99  ,   133 ,   45  ,0 };
            const unsigned long dim129JoeKuoD5Init[]    =   {   1   ,   1   ,   5   ,   11  ,   31  ,   49  ,   51  ,   45  ,   27  ,   415 ,0 };
            const unsigned long dim130JoeKuoD5Init[]    =   {   1   ,   3   ,   1   ,   11  ,   23  ,   19  ,   85  ,   17  ,   41  ,   625 ,0 };
            const unsigned long dim131JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   9   ,   3   ,   37  ,   9   ,   57  ,   59  ,   383 ,0 };
            const unsigned long dim132JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   3   ,   19  ,   53  ,   11  ,   215 ,   391 ,   51  ,0 };
            const unsigned long dim133JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   15  ,   17  ,   1   ,   53  ,   89  ,   291 ,   91  ,0 };
            const unsigned long dim134JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   1   ,   5   ,   33  ,   69  ,   29  ,   281 ,   13  ,0 };
            const unsigned long dim135JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   15  ,   21  ,   61  ,   61  ,   69  ,   395 ,   785 ,0 };
            const unsigned long dim136JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   7   ,   25  ,   51  ,   27  ,   11  ,   375 ,   865 ,0 };
            const unsigned long dim137JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   1   ,   11  ,   47  ,   45  ,   85  ,   507 ,   11  ,0 };
            const unsigned long dim138JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   5   ,   17  ,   53  ,   101 ,   57  ,   213 ,   795 ,0 };
            const unsigned long dim139JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   7   ,   15  ,   19  ,   117 ,   213 ,   397 ,   343 ,0 };
            const unsigned long dim140JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   7   ,   9   ,   19  ,   27  ,   23  ,   205 ,   707 ,0 };
            const unsigned long dim141JoeKuoD5Init[]    =   {   1   ,   3   ,   1   ,   13  ,   11  ,   51  ,   1   ,   3   ,   63  ,   483 ,0 };
            const unsigned long dim142JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   13  ,   27  ,   49  ,   97  ,   247 ,   273 ,   785 ,0 };
            const unsigned long dim143JoeKuoD5Init[]    =   {   1   ,   3   ,   1   ,   15  ,   31  ,   3   ,   43  ,   199 ,   81  ,   317 ,0 };
            const unsigned long dim144JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   15  ,   17  ,   3   ,   101 ,   13  ,   131 ,   631 ,0 };
            const unsigned long dim145JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   1   ,   25  ,   23  ,   17  ,   145 ,   247 ,   889 ,0 };
            const unsigned long dim146JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   7   ,   17  ,   5   ,   11  ,   133 ,   19  ,   507 ,0 };
            const unsigned long dim147JoeKuoD5Init[]    =   {   1   ,   3   ,   1   ,   7   ,   31  ,   53  ,   39  ,   107 ,   183 ,   335 ,0 };
            const unsigned long dim148JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   15  ,   3   ,   19  ,   39  ,   155 ,   477 ,   833 ,0 };
            const unsigned long dim149JoeKuoD5Init[]    =   {   1   ,   1   ,   1   ,   9   ,   31  ,   7   ,   5   ,   5   ,   399 ,   831 ,0 };
            const unsigned long dim150JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   1   ,   19  ,   37  ,   89  ,   243 ,   131 ,   901 ,0 };
            const unsigned long dim151JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   13  ,   23  ,   3   ,   127 ,   213 ,   97  ,   325 ,0 };
            const unsigned long dim152JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   9   ,   23  ,   27  ,   7   ,   161 ,   307 ,   451 ,0 };
            const unsigned long dim153JoeKuoD5Init[]    =   {   1   ,   1   ,   5   ,   1   ,   5   ,   25  ,   23  ,   103 ,   59  ,   431 ,0 };
            const unsigned long dim154JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   7   ,   3   ,   43  ,   121 ,   117 ,   33  ,   231 ,0 };
            const unsigned long dim155JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   7   ,   11  ,   61  ,   73  ,   231 ,   225 ,   97  ,0 };
            const unsigned long dim156JoeKuoD5Init[]    =   {   1   ,   1   ,   5   ,   11  ,   1   ,   9   ,   61  ,   3   ,   407 ,   425 ,0 };
            const unsigned long dim157JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   7   ,   1   ,   25  ,   95  ,   161 ,   387 ,   379 ,0 };
            const unsigned long dim158JoeKuoD5Init[]    =   {   1   ,   3   ,   1   ,   1   ,   15  ,   53  ,   55  ,   107 ,   425 ,   629 ,0 };
            const unsigned long dim159JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   11  ,   31  ,   59  ,   97  ,   99  ,   339 ,   417 ,0 };
            const unsigned long dim160JoeKuoD5Init[]    =   {   1   ,   3   ,   1   ,   9   ,   9   ,   5   ,   7   ,   157 ,   401 ,   155 ,0 };
            const unsigned long dim161JoeKuoD5Init[]    =   {   1   ,   3   ,   1   ,   7   ,   31  ,   15  ,   37  ,   253 ,   93  ,   149 ,   1255    ,0 };
            const unsigned long dim162JoeKuoD5Init[]    =   {   1   ,   1   ,   1   ,   9   ,   11  ,   61  ,   65  ,   71  ,   53  ,   929 ,   1153    ,0 };
            const unsigned long dim163JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   15  ,   13  ,   25  ,   93  ,   55  ,   55  ,   923 ,   601 ,0 };
            const unsigned long dim164JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   7   ,   21  ,   43  ,   71  ,   115 ,   253 ,   953 ,   1455    ,0 };
            const unsigned long dim165JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   7   ,   3   ,   31  ,   115 ,   213 ,   39  ,   589 ,   1029    ,0 };
            const unsigned long dim166JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   1   ,   7   ,   59  ,   31  ,   87  ,   233 ,   75  ,   365 ,0 };
            const unsigned long dim167JoeKuoD5Init[]    =   {   1   ,   1   ,   1   ,   15  ,   13  ,   27  ,   55  ,   117 ,   327 ,   465 ,   1735    ,0 };
            const unsigned long dim168JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   9   ,   25  ,   9   ,   125 ,   247 ,   81  ,   437 ,   483 ,0 };
            const unsigned long dim169JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   15  ,   27  ,   39  ,   27  ,   57  ,   45  ,   263 ,   281 ,0 };
            const unsigned long dim170JoeKuoD5Init[]    =   {   1   ,   3   ,   1   ,   5   ,   29  ,   5   ,   73  ,   231 ,   97  ,   773 ,   1349    ,0 };
            const unsigned long dim171JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   5   ,   25  ,   27  ,   81  ,   87  ,   91  ,   169 ,   235 ,0 };
            const unsigned long dim172JoeKuoD5Init[]    =   {   1   ,   3   ,   1   ,   11  ,   19  ,   61  ,   17  ,   175 ,   253 ,   673 ,   175 ,0 };
            const unsigned long dim173JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   7   ,   17  ,   43  ,   65  ,   245 ,   125 ,   137 ,   1475    ,0 };
            const unsigned long dim174JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   5   ,   15  ,   57  ,   39  ,   151 ,   197 ,   529 ,   393 ,0 };
            const unsigned long dim175JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   3   ,   29  ,   41  ,   93  ,   47  ,   375 ,   729 ,   709 ,0 };
            const unsigned long dim176JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   9   ,   13  ,   35  ,   49  ,   197 ,   107 ,   381 ,   1531    ,0 };
            const unsigned long dim177JoeKuoD5Init[]    =   {   1   ,   1   ,   5   ,   9   ,   15  ,   35  ,   105 ,   105 ,   259 ,   201 ,   317 ,0 };
            const unsigned long dim178JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   15  ,   13  ,   23  ,   89  ,   203 ,   335 ,   1003    ,   107 ,0 };
            const unsigned long dim179JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   15  ,   25  ,   25  ,   7   ,   145 ,   213 ,   845 ,   949 ,0 };
            const unsigned long dim180JoeKuoD5Init[]    =   {   1   ,   1   ,   5   ,   1   ,   23  ,   11  ,   101 ,   59  ,   57  ,   261 ,   1627    ,0 };
            const unsigned long dim181JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   11  ,   23  ,   63  ,   43  ,   137 ,   49  ,   249 ,   1369    ,0 };
            const unsigned long dim182JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   13  ,   25  ,   3   ,   81  ,   157 ,   511 ,   725 ,   2027    ,0 };
            const unsigned long dim183JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   7   ,   5   ,   37  ,   33  ,   7   ,   287 ,   307 ,   147 ,0 };
            const unsigned long dim184JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   5   ,   27  ,   19  ,   93  ,   173 ,   145 ,   821 ,   139 ,0 };
            const unsigned long dim185JoeKuoD5Init[]    =   {   1   ,   1   ,   1   ,   15  ,   23  ,   61  ,   123 ,   85  ,   375 ,   699 ,   229 ,0 };
            const unsigned long dim186JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   3   ,   25  ,   21  ,   127 ,   53  ,   247 ,   1005    ,   831 ,0 };
            const unsigned long dim187JoeKuoD5Init[]    =   {   1   ,   1   ,   5   ,   7   ,   25  ,   39  ,   27  ,   247 ,   319 ,   659 ,   1453    ,0 };
            const unsigned long dim188JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   11  ,   31  ,   45  ,   67  ,   195 ,   11  ,   481 ,   83  ,0 };
            const unsigned long dim189JoeKuoD5Init[]    =   {   1   ,   1   ,   5   ,   13  ,   27  ,   61  ,   5   ,   173 ,   353 ,   733 ,   1189    ,0 };
            const unsigned long dim190JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   7   ,   23  ,   49  ,   119 ,   145 ,   285 ,   873 ,   641 ,0 };
            const unsigned long dim191JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   9   ,   1   ,   17  ,   119 ,   121 ,   203 ,   483 ,   1601    ,0 };
            const unsigned long dim192JoeKuoD5Init[]    =   {   1   ,   3   ,   1   ,   11  ,   21  ,   35  ,   121 ,   11  ,   213 ,   93  ,   77  ,0 };
            const unsigned long dim193JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   7   ,   7   ,   5   ,   27  ,   153 ,   223 ,   831 ,   679 ,0 };
            const unsigned long dim194JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   5   ,   1   ,   39  ,   39  ,   233 ,   483 ,   667 ,   1367    ,0 };
            const unsigned long dim195JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   13  ,   19  ,   51  ,   21  ,   209 ,   381 ,   787 ,   1451    ,0 };
            const unsigned long dim196JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   3   ,   15  ,   9   ,   29  ,   225 ,   225 ,   389 ,   1075    ,0 };
            const unsigned long dim197JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   5   ,   13  ,   57  ,   55  ,   229 ,   279 ,   1019    ,   1491    ,0 };
            const unsigned long dim198JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   9   ,   19  ,   19  ,   35  ,   15  ,   335 ,   685 ,   1987    ,0 };
            const unsigned long dim199JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   11  ,   1   ,   5   ,   3   ,   29  ,   19  ,   363 ,   247 ,0 };
            const unsigned long dim200JoeKuoD5Init[]    =   {   1   ,   1   ,   5   ,   11  ,   13  ,   17  ,   17  ,   253 ,   365 ,   397 ,   1643    ,0 };
            const unsigned long dim201JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   15  ,   25  ,   23  ,   81  ,   243 ,   343 ,   441 ,   1675    ,0 };
            const unsigned long dim202JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   11  ,   15  ,   51  ,   23  ,   213 ,   235 ,   997 ,   1205    ,0 };
            const unsigned long dim203JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   7   ,   5   ,   39  ,   99  ,   165 ,   331 ,   795 ,   803 ,0 };
            const unsigned long dim204JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   13  ,   31  ,   63  ,   109 ,   171 ,   151 ,   269 ,   581 ,0 };
            const unsigned long dim205JoeKuoD5Init[]    =   {   1   ,   3   ,   1   ,   7   ,   25  ,   13  ,   41  ,   75  ,   411 ,   425 ,   267 ,0 };
            const unsigned long dim206JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   7   ,   1   ,   31  ,   49  ,   113 ,   473 ,   677 ,   395 ,0 };
            const unsigned long dim207JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   5   ,   23  ,   43  ,   111 ,   171 ,   489 ,   949 ,   1681    ,0 };
            const unsigned long dim208JoeKuoD5Init[]    =   {   1   ,   3   ,   1   ,   5   ,   5   ,   37  ,   67  ,   23  ,   115 ,   909 ,   853 ,0 };
            const unsigned long dim209JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   3   ,   27  ,   35  ,   63  ,   45  ,   481 ,   571 ,   793 ,0 };
            const unsigned long dim210JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   15  ,   11  ,   45  ,   53  ,   225 ,   147 ,   935 ,   1189    ,0 };
            const unsigned long dim211JoeKuoD5Init[]    =   {   1   ,   1   ,   1   ,   5   ,   15  ,   17  ,   77  ,   59  ,   169 ,   831 ,   163 ,0 };
            const unsigned long dim212JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   15  ,   7   ,   29  ,   85  ,   19  ,   313 ,   35  ,   401 ,0 };
            const unsigned long dim213JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   13  ,   3   ,   1   ,   45  ,   139 ,   17  ,   847 ,   1309    ,0 };
            const unsigned long dim214JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   3   ,   25  ,   1   ,   117 ,   185 ,   291 ,   251 ,   1049    ,0 };
            const unsigned long dim215JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   1   ,   27  ,   63  ,   33  ,   161 ,   181 ,   79  ,   667 ,0 };
            const unsigned long dim216JoeKuoD5Init[]    =   {   1   ,   3   ,   1   ,   11  ,   9   ,   63  ,   91  ,   207 ,   329 ,   3   ,   1155    ,0 };
            const unsigned long dim217JoeKuoD5Init[]    =   {   1   ,   1   ,   1   ,   5   ,   3   ,   59  ,   39  ,   37  ,   231 ,   993 ,   1147    ,0 };
            const unsigned long dim218JoeKuoD5Init[]    =   {   1   ,   3   ,   1   ,   3   ,   7   ,   41  ,   85  ,   223 ,   407 ,   403 ,   573 ,0 };
            const unsigned long dim219JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   3   ,   15  ,   51  ,   1   ,   191 ,   123 ,   103 ,   1201    ,0 };
            const unsigned long dim220JoeKuoD5Init[]    =   {   1   ,   1   ,   5   ,   3   ,   31  ,   9   ,   83  ,   231 ,   419 ,   109 ,   1455    ,0 };
            const unsigned long dim221JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   11  ,   13  ,   39  ,   89  ,   9   ,   237 ,   185 ,   1113    ,0 };
            const unsigned long dim222JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   9   ,   1   ,   43  ,   121 ,   241 ,   165 ,   263 ,   1205    ,0 };
            const unsigned long dim223JoeKuoD5Init[]    =   {   1   ,   1   ,   1   ,   3   ,   21  ,   3   ,   61  ,   219 ,   49  ,   733 ,   25  ,0 };
            const unsigned long dim224JoeKuoD5Init[]    =   {   1   ,   3   ,   1   ,   7   ,   31  ,   27  ,   121 ,   61  ,   447 ,   401 ,   529 ,0 };
            const unsigned long dim225JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   11  ,   27  ,   47  ,   85  ,   5   ,   305 ,   763 ,   1255    ,0 };
            const unsigned long dim226JoeKuoD5Init[]    =   {   1   ,   1   ,   1   ,   9   ,   13  ,   41  ,   9   ,   51  ,   195 ,   103 ,   983 ,0 };
            const unsigned long dim227JoeKuoD5Init[]    =   {   1   ,   1   ,   1   ,   15  ,   1   ,   27  ,   65  ,   91  ,   61  ,   591 ,   2039    ,0 };
            const unsigned long dim228JoeKuoD5Init[]    =   {   1   ,   1   ,   1   ,   1   ,   15  ,   19  ,   107 ,   197 ,   121 ,   879 ,   771 ,0 };
            const unsigned long dim229JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   7   ,   19  ,   53  ,   9   ,   3   ,   67  ,   893 ,   1817    ,0 };
            const unsigned long dim230JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   9   ,   11  ,   63  ,   53  ,   247 ,   65  ,   681 ,   1721    ,0 };
            const unsigned long dim231JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   3   ,   25  ,   47  ,   91  ,   55  ,   471 ,   731 ,   939 ,0 };
            const unsigned long dim232JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   9   ,   7   ,   45  ,   121 ,   69  ,   423 ,   599 ,   2027    ,0 };
            const unsigned long dim233JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   11  ,   23  ,   7   ,   43  ,   179 ,   511 ,   571 ,   1707    ,0 };
            const unsigned long dim234JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   9   ,   17  ,   11  ,   69  ,   13  ,   303 ,   299 ,   653 ,0 };
            const unsigned long dim235JoeKuoD5Init[]    =   {   1   ,   1   ,   1   ,   7   ,   7   ,   29  ,   69  ,   237 ,   237 ,   425 ,   1413    ,0 };
            const unsigned long dim236JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   3   ,   19  ,   31  ,   55  ,   55  ,   225 ,   943 ,   1027    ,0 };
            const unsigned long dim237JoeKuoD5Init[]    =   {   1   ,   1   ,   5   ,   7   ,   25  ,   9   ,   9   ,   29  ,   485 ,   885 ,   1229    ,0 };
            const unsigned long dim238JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   9   ,   7   ,   11  ,   73  ,   151 ,   17  ,   669 ,   773 ,0 };
            const unsigned long dim239JoeKuoD5Init[]    =   {   1   ,   1   ,   5   ,   13  ,   25  ,   13  ,   51  ,   53  ,   411 ,   555 ,   795 ,0 };
            const unsigned long dim240JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   9   ,   21  ,   33  ,   7   ,   113 ,   325 ,   593 ,   1647    ,0 };
            const unsigned long dim241JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   15  ,   3   ,   13  ,   83  ,   205 ,   153 ,   7   ,   1181    ,0 };
            const unsigned long dim242JoeKuoD5Init[]    =   {   1   ,   1   ,   5   ,   5   ,   1   ,   55  ,   91  ,   17  ,   383 ,   453 ,   1749    ,0 };
            const unsigned long dim243JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   1   ,   5   ,   13  ,   91  ,   19  ,   241 ,   569 ,   291 ,0 };
            const unsigned long dim244JoeKuoD5Init[]    =   {   1   ,   1   ,   1   ,   9   ,   13  ,   19  ,   109 ,   195 ,   17  ,   203 ,   473 ,0 };
            const unsigned long dim245JoeKuoD5Init[]    =   {   1   ,   1   ,   1   ,   11  ,   17  ,   17  ,   43  ,   201 ,   297 ,   159 ,   685 ,0 };
            const unsigned long dim246JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   15  ,   15  ,   33  ,   91  ,   53  ,   337 ,   237 ,   1063    ,0 };
            const unsigned long dim247JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   9   ,   13  ,   15  ,   59  ,   115 ,   457 ,   169 ,   29  ,0 };
            const unsigned long dim248JoeKuoD5Init[]    =   {   1   ,   1   ,   1   ,   7   ,   21  ,   51  ,   41  ,   49  ,   467 ,   171 ,   301 ,0 };
            const unsigned long dim249JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   5   ,   19  ,   57  ,   27  ,   57  ,   119 ,   183 ,   1519    ,0 };
            const unsigned long dim250JoeKuoD5Init[]    =   {   1   ,   1   ,   1   ,   9   ,   19  ,   21  ,   117 ,   35  ,   43  ,   829 ,   1817    ,0 };
            const unsigned long dim251JoeKuoD5Init[]    =   {   1   ,   3   ,   1   ,   7   ,   9   ,   27  ,   127 ,   233 ,   229 ,   467 ,   2033    ,0 };
            const unsigned long dim252JoeKuoD5Init[]    =   {   1   ,   3   ,   1   ,   1   ,   7   ,   21  ,   113 ,   23  ,   15  ,   43  ,   375 ,0 };
            const unsigned long dim253JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   3   ,   11  ,   21  ,   13  ,   87  ,   57  ,   805 ,   1529    ,0 };
            const unsigned long dim254JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   15  ,   3   ,   39  ,   115 ,   179 ,   199 ,   907 ,   1487    ,0 };
            const unsigned long dim255JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   1   ,   17  ,   17  ,   83  ,   23  ,   421 ,   813 ,   29  ,0 };
            const unsigned long dim256JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   3   ,   29  ,   39  ,   81  ,   69  ,   339 ,   495 ,   281 ,0 };
            const unsigned long dim257JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   5   ,   27  ,   39  ,   9   ,   233 ,   19  ,   663 ,   57  ,0 };
            const unsigned long dim258JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   15  ,   17  ,   7   ,   17  ,   39  ,   299 ,   97  ,   1329    ,0 };
            const unsigned long dim259JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   7   ,   9   ,   27  ,   5   ,   245 ,   477 ,   591 ,   1021    ,0 };
            const unsigned long dim260JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   13  ,   9   ,   37  ,   19  ,   141 ,   201 ,   57  ,   1117    ,0 };
            const unsigned long dim261JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   1   ,   7   ,   27  ,   33  ,   235 ,   247 ,   701 ,   1293    ,0 };
            const unsigned long dim262JoeKuoD5Init[]    =   {   1   ,   1   ,   5   ,   3   ,   17  ,   25  ,   81  ,   213 ,   457 ,   877 ,   741 ,0 };
            const unsigned long dim263JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   13  ,   15  ,   47  ,   113 ,   29  ,   111 ,   913 ,   1695    ,0 };
            const unsigned long dim264JoeKuoD5Init[]    =   {   1   ,   1   ,   1   ,   3   ,   5   ,   53  ,   119 ,   215 ,   163 ,   933 ,   1447    ,0 };
            const unsigned long dim265JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   7   ,   15  ,   27  ,   111 ,   19  ,   231 ,   287 ,   1439    ,0 };
            const unsigned long dim266JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   5   ,   27  ,   29  ,   35  ,   109 ,   249 ,   989 ,   1837    ,0 };
            const unsigned long dim267JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   9   ,   29  ,   35  ,   51  ,   241 ,   509 ,   163 ,   831 ,0 };
            const unsigned long dim268JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   9   ,   7   ,   59  ,   43  ,   111 ,   119 ,   639 ,   899 ,0 };
            const unsigned long dim269JoeKuoD5Init[]    =   {   1   ,   3   ,   1   ,   3   ,   15  ,   47  ,   95  ,   219 ,   377 ,   899 ,   535 ,0 };
            const unsigned long dim270JoeKuoD5Init[]    =   {   1   ,   1   ,   1   ,   11  ,   3   ,   19  ,   115 ,   59  ,   143 ,   13  ,   701 ,0 };
            const unsigned long dim271JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   11  ,   5   ,   17  ,   17  ,   91  ,   223 ,   923 ,   1299    ,0 };
            const unsigned long dim272JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   15  ,   21  ,   21  ,   85  ,   127 ,   253 ,   271 ,   725 ,0 };
            const unsigned long dim273JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   15  ,   9   ,   11  ,   113 ,   67  ,   509 ,   697 ,   1163    ,0 };
            const unsigned long dim274JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   11  ,   13  ,   53  ,   15  ,   221 ,   253 ,   219 ,   1839    ,0 };
            const unsigned long dim275JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   3   ,   31  ,   63  ,   85  ,   171 ,   345 ,   243 ,   711 ,0 };
            const unsigned long dim276JoeKuoD5Init[]    =   {   1   ,   1   ,   5   ,   1   ,   25  ,   13  ,   65  ,   91  ,   441 ,   609 ,   1751    ,0 };
            const unsigned long dim277JoeKuoD5Init[]    =   {   1   ,   1   ,   1   ,   5   ,   23  ,   1   ,   15  ,   83  ,   115 ,   367 ,   735 ,0 };
            const unsigned long dim278JoeKuoD5Init[]    =   {   1   ,   1   ,   1   ,   3   ,   11  ,   15  ,   41  ,   1   ,   437 ,   231 ,   1529    ,0 };
            const unsigned long dim279JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   3   ,   29  ,   11  ,   89  ,   133 ,   473 ,   811 ,   87  ,0 };
            const unsigned long dim280JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   1   ,   15  ,   39  ,   97  ,   197 ,   475 ,   105 ,   527 ,0 };
            const unsigned long dim281JoeKuoD5Init[]    =   {   1   ,   1   ,   1   ,   1   ,   9   ,   17  ,   21  ,   167 ,   255 ,   341 ,   765 ,0 };
            const unsigned long dim282JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   7   ,   23  ,   47  ,   121 ,   219 ,   343 ,   169 ,   1147    ,0 };
            const unsigned long dim283JoeKuoD5Init[]    =   {   1   ,   3   ,   1   ,   7   ,   3   ,   57  ,   27  ,   147 ,   383 ,   157 ,   1851    ,0 };
            const unsigned long dim284JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   1   ,   25  ,   17  ,   35  ,   123 ,   371 ,   281 ,   881 ,0 };
            const unsigned long dim285JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   7   ,   11  ,   11  ,   21  ,   5   ,   53  ,   155 ,   1811    ,0 };
            const unsigned long dim286JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   15  ,   31  ,   27  ,   117 ,   169 ,   389 ,   651 ,   1513    ,0 };
            const unsigned long dim287JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   5   ,   31  ,   15  ,   103 ,   59  ,   73  ,   575 ,   1597    ,0 };
            const unsigned long dim288JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   9   ,   15  ,   5   ,   17  ,   183 ,   471 ,   561 ,   607 ,0 };
            const unsigned long dim289JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   5   ,   9   ,   39  ,   25  ,   87  ,   171 ,   559 ,   481 ,0 };
            const unsigned long dim290JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   5   ,   17  ,   61  ,   101 ,   255 ,   147 ,   481 ,   661 ,0 };
            const unsigned long dim291JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   1   ,   17  ,   9   ,   119 ,   31  ,   177 ,   475 ,   1243    ,0 };
            const unsigned long dim292JoeKuoD5Init[]    =   {   1   ,   1   ,   1   ,   13  ,   27  ,   45  ,   111 ,   229 ,   201 ,   927 ,   339 ,0 };
            const unsigned long dim293JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   15  ,   9   ,   25  ,   23  ,   91  ,   453 ,   861 ,   1919    ,0 };
            const unsigned long dim294JoeKuoD5Init[]    =   {   1   ,   3   ,   1   ,   3   ,   7   ,   57  ,   93  ,   75  ,   223 ,   63  ,   7   ,0 };
            const unsigned long dim295JoeKuoD5Init[]    =   {   1   ,   1   ,   5   ,   13  ,   7   ,   29  ,   71  ,   197 ,   405 ,   401 ,   585 ,0 };
            const unsigned long dim296JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   13  ,   11  ,   11  ,   7   ,   157 ,   1   ,   105 ,   473 ,0 };
            const unsigned long dim297JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   1   ,   29  ,   3   ,   127 ,   243 ,   93  ,   123 ,   1041    ,0 };
            const unsigned long dim298JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   9   ,   25  ,   55  ,   13  ,   243 ,   37  ,   565 ,   1167    ,0 };
            const unsigned long dim299JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   15  ,   31  ,   29  ,   75  ,   61  ,   43  ,   159 ,   443 ,0 };
            const unsigned long dim300JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   9   ,   15  ,   63  ,   43  ,   251 ,   97  ,   141 ,   791 ,0 };
            const unsigned long dim301JoeKuoD5Init[]    =   {   1   ,   3   ,   1   ,   3   ,   27  ,   43  ,   17  ,   49  ,   109 ,   777 ,   1999    ,0 };
            const unsigned long dim302JoeKuoD5Init[]    =   {   1   ,   3   ,   1   ,   1   ,   25  ,   5   ,   15  ,   145 ,   75  ,   855 ,   771 ,0 };
            const unsigned long dim303JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   13  ,   7   ,   27  ,   3   ,   221 ,   451 ,   533 ,   1059    ,0 };
            const unsigned long dim304JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   13  ,   11  ,   61  ,   53  ,   33  ,   217 ,   967 ,   177 ,0 };
            const unsigned long dim305JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   7   ,   1   ,   7   ,   85  ,   105 ,   417 ,   87  ,   417 ,0 };
            const unsigned long dim306JoeKuoD5Init[]    =   {   1   ,   1   ,   1   ,   3   ,   13  ,   59  ,   11  ,   219 ,   363 ,   481 ,   893 ,0 };
            const unsigned long dim307JoeKuoD5Init[]    =   {   1   ,   1   ,   5   ,   11  ,   27  ,   23  ,   85  ,   239 ,   343 ,   43  ,   1597    ,0 };
            const unsigned long dim308JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   3   ,   29  ,   5   ,   127 ,   49  ,   223 ,   797 ,   2003    ,0 };
            const unsigned long dim309JoeKuoD5Init[]    =   {   1   ,   3   ,   1   ,   9   ,   29  ,   61  ,   21  ,   191 ,   157 ,   355 ,   2033    ,0 };
            const unsigned long dim310JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   5   ,   5   ,   49  ,   53  ,   207 ,   121 ,   451 ,   319 ,0 };
            const unsigned long dim311JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   11  ,   9   ,   7   ,   111 ,   153 ,   151 ,   395 ,   1389    ,0 };
            const unsigned long dim312JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   11  ,   25  ,   45  ,   113 ,   99  ,   263 ,   561 ,   1181    ,0 };
            const unsigned long dim313JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   1   ,   13  ,   27  ,   77  ,   1   ,   109 ,   741 ,   59  ,0 };
            const unsigned long dim314JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   1   ,   15  ,   57  ,   43  ,   7   ,   507 ,   885 ,   747 ,0 };
            const unsigned long dim315JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   5   ,   25  ,   7   ,   45  ,   147 ,   375 ,   975 ,   619 ,0 };
            const unsigned long dim316JoeKuoD5Init[]    =   {   1   ,   1   ,   5   ,   11  ,   13  ,   11  ,   107 ,   81  ,   199 ,   11  ,   1267    ,0 };
            const unsigned long dim317JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   3   ,   15  ,   11  ,   113 ,   61  ,   425 ,   43  ,   1889    ,0 };
            const unsigned long dim318JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   7   ,   29  ,   11  ,   123 ,   237 ,   173 ,   249 ,   1091    ,0 };
            const unsigned long dim319JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   15  ,   29  ,   13  ,   21  ,   159 ,   149 ,   379 ,   1665    ,0 };
            const unsigned long dim320JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   15  ,   13  ,   53  ,   23  ,   47  ,   115 ,   183 ,   577 ,0 };
            const unsigned long dim321JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   9   ,   11  ,   49  ,   77  ,   81  ,   193 ,   133 ,   489 ,0 };
            const unsigned long dim322JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   15  ,   7   ,   27  ,   53  ,   187 ,   347 ,   211 ,   233 ,0 };
            const unsigned long dim323JoeKuoD5Init[]    =   {   1   ,   1   ,   5   ,   9   ,   15  ,   17  ,   45  ,   75  ,   449 ,   971 ,   1123    ,0 };
            const unsigned long dim324JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   11  ,   25  ,   21  ,   109 ,   71  ,   439 ,   439 ,   1397    ,0 };
            const unsigned long dim325JoeKuoD5Init[]    =   {   1   ,   1   ,   1   ,   7   ,   9   ,   47  ,   117 ,   117 ,   165 ,   531 ,   271 ,0 };
            const unsigned long dim326JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   9   ,   11  ,   5   ,   31  ,   199 ,   159 ,   87  ,   1729    ,0 };
            const unsigned long dim327JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   5   ,   27  ,   29  ,   17  ,   237 ,   175 ,   881 ,   989 ,0 };
            const unsigned long dim328JoeKuoD5Init[]    =   {   1   ,   1   ,   5   ,   13  ,   25  ,   63  ,   19  ,   159 ,   409 ,   247 ,   683 ,0 };
            const unsigned long dim329JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   15  ,   15  ,   39  ,   101 ,   129 ,   253 ,   487 ,   719 ,0 };
            const unsigned long dim330JoeKuoD5Init[]    =   {   1   ,   3   ,   1   ,   1   ,   7   ,   13  ,   107 ,   249 ,   331 ,   553 ,   1199    ,0 };
            const unsigned long dim331JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   1   ,   5   ,   51  ,   81  ,   37  ,   349 ,   561 ,   295 ,0 };
            const unsigned long dim332JoeKuoD5Init[]    =   {   1   ,   3   ,   1   ,   5   ,   27  ,   61  ,   49  ,   89  ,   379 ,   67  ,   1063    ,0 };
            const unsigned long dim333JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   9   ,   15  ,   37  ,   119 ,   141 ,   81  ,   341 ,   2003    ,0 };
            const unsigned long dim334JoeKuoD5Init[]    =   {   1   ,   3   ,   1   ,   13  ,   29  ,   31  ,   29  ,   143 ,   463 ,   399 ,   1345    ,0 };
            const unsigned long dim335JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   3   ,   29  ,   35  ,   29  ,   233 ,   499 ,   503 ,   903 ,0 };
            const unsigned long dim336JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   13  ,   29  ,   23  ,   127 ,   185 ,   77  ,   555 ,   1311    ,0 };
            const unsigned long dim337JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   11  ,   23  ,   3   ,   111 ,   159 ,   503 ,   889 ,   1043    ,   1153    ,0 };
            const unsigned long dim338JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   1   ,   13  ,   41  ,   109 ,   133 ,   81  ,   525 ,   2027    ,   3059    ,0 };
            const unsigned long dim339JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   11  ,   29  ,   53  ,   111 ,   129 ,   399 ,   479 ,   467 ,   363 ,0 };
            const unsigned long dim340JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   3   ,   5   ,   7   ,   9   ,   21  ,   391 ,   851 ,   575 ,   1317    ,0 };
            const unsigned long dim341JoeKuoD5Init[]    =   {   1   ,   1   ,   1   ,   5   ,   19  ,   59  ,   91  ,   133 ,   403 ,   71  ,   1895    ,   3029    ,0 };
            const unsigned long dim342JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   11  ,   21  ,   29  ,   113 ,   109 ,   463 ,   251 ,   393 ,   3169    ,0 };
            const unsigned long dim343JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   11  ,   25  ,   3   ,   47  ,   195 ,   223 ,   1003    ,   947 ,   121 ,0 };
            const unsigned long dim344JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   9   ,   31  ,   61  ,   63  ,   31  ,   49  ,   907 ,   389 ,   3713    ,0 };
            const unsigned long dim345JoeKuoD5Init[]    =   {   1   ,   1   ,   1   ,   13  ,   13  ,   19  ,   55  ,   87  ,   489 ,   665 ,   945 ,   2081    ,0 };
            const unsigned long dim346JoeKuoD5Init[]    =   {   1   ,   1   ,   1   ,   7   ,   23  ,   5   ,   39  ,   19  ,   355 ,   399 ,   929 ,   3077    ,0 };
            const unsigned long dim347JoeKuoD5Init[]    =   {   1   ,   3   ,   1   ,   11  ,   7   ,   59  ,   43  ,   69  ,   285 ,   753 ,   75  ,   2261    ,0 };
            const unsigned long dim348JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   3   ,   27  ,   45  ,   29  ,   181 ,   347 ,   863 ,   1421    ,   2077    ,0 };
            const unsigned long dim349JoeKuoD5Init[]    =   {   1   ,   1   ,   1   ,   3   ,   7   ,   27  ,   77  ,   67  ,   399 ,   919 ,   917 ,   1465    ,0 };
            const unsigned long dim350JoeKuoD5Init[]    =   {   1   ,   1   ,   5   ,   5   ,   11  ,   41  ,   17  ,   65  ,   495 ,   643 ,   1641    ,   323 ,0 };
            const unsigned long dim351JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   1   ,   9   ,   37  ,   107 ,   171 ,   189 ,   405 ,   2005    ,   2811    ,0 };
            const unsigned long dim352JoeKuoD5Init[]    =   {   1   ,   1   ,   5   ,   11  ,   3   ,   63  ,   51  ,   27  ,   479 ,   571 ,   575 ,   2859    ,0 };
            const unsigned long dim353JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   9   ,   29  ,   23  ,   89  ,   7   ,   265 ,   41  ,   481 ,   2177    ,0 };
            const unsigned long dim354JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   7   ,   29  ,   15  ,   79  ,   217 ,   411 ,   867 ,   49  ,   469 ,0 };
            const unsigned long dim355JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   3   ,   3   ,   27  ,   69  ,   177 ,   291 ,   965 ,   637 ,   3629    ,0 };
            const unsigned long dim356JoeKuoD5Init[]    =   {   1   ,   1   ,   1   ,   13  ,   11  ,   45  ,   83  ,   63  ,   275 ,   851 ,   779 ,   2615    ,0 };
            const unsigned long dim357JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   13  ,   1   ,   27  ,   89  ,   153 ,   355 ,   811 ,   515 ,   1541    ,0 };
            const unsigned long dim358JoeKuoD5Init[]    =   {   1   ,   3   ,   1   ,   13  ,   1   ,   21  ,   75  ,   5   ,   255 ,   813 ,   1347    ,   2301    ,0 };
            const unsigned long dim359JoeKuoD5Init[]    =   {   1   ,   1   ,   1   ,   15  ,   9   ,   49  ,   3   ,   203 ,   505 ,   591 ,   713 ,   2893    ,0 };
            const unsigned long dim360JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   3   ,   1   ,   51  ,   11  ,   161 ,   41  ,   17  ,   435 ,   3045    ,0 };
            const unsigned long dim361JoeKuoD5Init[]    =   {   1   ,   1   ,   5   ,   9   ,   15  ,   23  ,   115 ,   73  ,   343 ,   985 ,   1559    ,   1615    ,0 };
            const unsigned long dim362JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   5   ,   9   ,   43  ,   17  ,   187 ,   311 ,   749 ,   1841    ,   609 ,0 };
            const unsigned long dim363JoeKuoD5Init[]    =   {   1   ,   3   ,   1   ,   15  ,   13  ,   3   ,   113 ,   83  ,   287 ,   931 ,   399 ,   2143    ,0 };
            const unsigned long dim364JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   13  ,   17  ,   11  ,   99  ,   235 ,   313 ,   293 ,   2005    ,   2557    ,0 };
            const unsigned long dim365JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   13  ,   7   ,   57  ,   79  ,   225 ,   415 ,   749 ,   1243    ,   1303    ,0 };
            const unsigned long dim366JoeKuoD5Init[]    =   {   1   ,   1   ,   5   ,   5   ,   21  ,   37  ,   55  ,   53  ,   389 ,   141 ,   1231    ,   1639    ,0 };
            const unsigned long dim367JoeKuoD5Init[]    =   {   1   ,   3   ,   1   ,   13  ,   15  ,   37  ,   83  ,   219 ,   471 ,   751 ,   1241    ,   269 ,0 };
            const unsigned long dim368JoeKuoD5Init[]    =   {   1   ,   3   ,   1   ,   11  ,   7   ,   51  ,   37  ,   81  ,   97  ,   857 ,   1431    ,   883 ,0 };
            const unsigned long dim369JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   7   ,   27  ,   31  ,   29  ,   223 ,   439 ,   25  ,   379 ,   3721    ,0 };
            const unsigned long dim370JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   7   ,   13  ,   11  ,   55  ,   127 ,   493 ,   493 ,   143 ,   1595    ,0 };
            const unsigned long dim371JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   15  ,   27  ,   55  ,   93  ,   91  ,   49  ,   931 ,   99  ,   1887    ,0 };
            const unsigned long dim372JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   1   ,   13  ,   11  ,   81  ,   175 ,   171 ,   203 ,   679 ,   239 ,0 };
            const unsigned long dim373JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   9   ,   19  ,   35  ,   79  ,   51  ,   163 ,   571 ,   363 ,   3903    ,0 };
            const unsigned long dim374JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   5   ,   3   ,   11  ,   99  ,   57  ,   479 ,   571 ,   487 ,   2141    ,0 };
            const unsigned long dim375JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   13  ,   1   ,   3   ,   123 ,   191 ,   349 ,   523 ,   53  ,   2991    ,0 };
            const unsigned long dim376JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   5   ,   17  ,   1   ,   5   ,   131 ,   279 ,   717 ,   1725    ,   35  ,0 };
            const unsigned long dim377JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   1   ,   9   ,   43  ,   29  ,   9   ,   487 ,   349 ,   457 ,   2551    ,0 };
            const unsigned long dim378JoeKuoD5Init[]    =   {   1   ,   1   ,   5   ,   5   ,   7   ,   55  ,   75  ,   245 ,   249 ,   623 ,   1681    ,   2345    ,0 };
            const unsigned long dim379JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   11  ,   15  ,   35  ,   111 ,   185 ,   269 ,   913 ,   1899    ,   4059    ,0 };
            const unsigned long dim380JoeKuoD5Init[]    =   {   1   ,   1   ,   5   ,   3   ,   1   ,   51  ,   43  ,   159 ,   273 ,   329 ,   863 ,   831 ,0 };
            const unsigned long dim381JoeKuoD5Init[]    =   {   1   ,   1   ,   5   ,   15  ,   31  ,   35  ,   23  ,   135 ,   223 ,   333 ,   1265    ,   1183    ,0 };
            const unsigned long dim382JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   9   ,   9   ,   21  ,   93  ,   33  ,   341 ,   649 ,   1707    ,   1995    ,0 };
            const unsigned long dim383JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   5   ,   9   ,   9   ,   9   ,   175 ,   331 ,   709 ,   927 ,   423 ,0 };
            const unsigned long dim384JoeKuoD5Init[]    =   {   1   ,   3   ,   1   ,   1   ,   5   ,   41  ,   31  ,   105 ,   223 ,   17  ,   1485    ,   2133    ,0 };
            const unsigned long dim385JoeKuoD5Init[]    =   {   1   ,   1   ,   5   ,   11  ,   23  ,   7   ,   95  ,   87  ,   303 ,   817 ,   1019    ,   3335    ,0 };
            const unsigned long dim386JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   1   ,   21  ,   3   ,   7   ,   133 ,   23  ,   235 ,   1311    ,   531 ,0 };
            const unsigned long dim387JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   15  ,   25  ,   35  ,   69  ,   251 ,   37  ,   5   ,   1147    ,   2593    ,0 };
            const unsigned long dim388JoeKuoD5Init[]    =   {   1   ,   1   ,   1   ,   9   ,   21  ,   55  ,   27  ,   129 ,   239 ,   887 ,   1759    ,   3211    ,0 };
            const unsigned long dim389JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   15  ,   3   ,   41  ,   13  ,   141 ,   339 ,   921 ,   1081    ,   4047    ,0 };
            const unsigned long dim390JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   5   ,   1   ,   49  ,   51  ,   91  ,   357 ,   259 ,   547 ,   189 ,0 };
            const unsigned long dim391JoeKuoD5Init[]    =   {   1   ,   3   ,   1   ,   13  ,   15  ,   29  ,   43  ,   165 ,   213 ,   59  ,   429 ,   1831    ,0 };
            const unsigned long dim392JoeKuoD5Init[]    =   {   1   ,   3   ,   1   ,   9   ,   25  ,   11  ,   37  ,   67  ,   5   ,   77  ,   915 ,   3865    ,0 };
            const unsigned long dim393JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   9   ,   31  ,   63  ,   61  ,   209 ,   283 ,   223 ,   1253    ,   2137    ,0 };
            const unsigned long dim394JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   11  ,   13  ,   45  ,   65  ,   105 ,   419 ,   909 ,   1943    ,   2201    ,0 };
            const unsigned long dim395JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   7   ,   9   ,   27  ,   21  ,   233 ,   37  ,   23  ,   921 ,   969 ,0 };
            const unsigned long dim396JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   3   ,   31  ,   55  ,   39  ,   127 ,   455 ,   397 ,   65  ,   2381    ,0 };
            const unsigned long dim397JoeKuoD5Init[]    =   {   1   ,   3   ,   1   ,   3   ,   27  ,   57  ,   39  ,   137 ,   263 ,   519 ,   427 ,   3289    ,0 };
            const unsigned long dim398JoeKuoD5Init[]    =   {   1   ,   1   ,   1   ,   5   ,   9   ,   29  ,   21  ,   99  ,   21  ,   807 ,   1871    ,   2875    ,0 };
            const unsigned long dim399JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   15  ,   27  ,   59  ,   3   ,   15  ,   189 ,   681 ,   305 ,   2969    ,0 };
            const unsigned long dim400JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   13  ,   25  ,   43  ,   111 ,   179 ,   281 ,   377 ,   1885    ,   815 ,0 };
            const unsigned long dim401JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   1   ,   15  ,   13  ,   17  ,   99  ,   53  ,   269 ,   1199    ,   1771    ,0 };
            const unsigned long dim402JoeKuoD5Init[]    =   {   1   ,   1   ,   1   ,   7   ,   7   ,   59  ,   115 ,   209 ,   327 ,   913 ,   715 ,   279 ,0 };
            const unsigned long dim403JoeKuoD5Init[]    =   {   1   ,   1   ,   5   ,   5   ,   11  ,   29  ,   81  ,   69  ,   191 ,   453 ,   379 ,   1379    ,0 };
            const unsigned long dim404JoeKuoD5Init[]    =   {   1   ,   3   ,   1   ,   13  ,   1   ,   29  ,   85  ,   181 ,   281 ,   463 ,   137 ,   2779    ,0 };
            const unsigned long dim405JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   7   ,   25  ,   39  ,   45  ,   241 ,   87  ,   11  ,   511 ,   1919    ,0 };
            const unsigned long dim406JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   13  ,   21  ,   17  ,   57  ,   249 ,   91  ,   165 ,   1867    ,   615 ,0 };
            const unsigned long dim407JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   3   ,   29  ,   47  ,   79  ,   83  ,   3   ,   765 ,   1803    ,   2741    ,0 };
            const unsigned long dim408JoeKuoD5Init[]    =   {   1   ,   1   ,   1   ,   15  ,   29  ,   41  ,   23  ,   9   ,   205 ,   657 ,   721 ,   2877    ,0 };
            const unsigned long dim409JoeKuoD5Init[]    =   {   1   ,   1   ,   5   ,   5   ,   31  ,   21  ,   71  ,   217 ,   19  ,   589 ,   281 ,   719 ,0 };
            const unsigned long dim410JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   9   ,   11  ,   37  ,   3   ,   159 ,   41  ,   823 ,   1519    ,   3395    ,0 };
            const unsigned long dim411JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   3   ,   7   ,   51  ,   49  ,   193 ,   37  ,   981 ,   687 ,   3219    ,0 };
            const unsigned long dim412JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   15  ,   3   ,   27  ,   79  ,   195 ,   155 ,   613 ,   1933    ,   2083    ,0 };
            const unsigned long dim413JoeKuoD5Init[]    =   {   1   ,   3   ,   1   ,   11  ,   13  ,   39  ,   75  ,   109 ,   395 ,   809 ,   545 ,   499 ,0 };
            const unsigned long dim414JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   11  ,   1   ,   63  ,   47  ,   77  ,   455 ,   617 ,   739 ,   2885    ,0 };
            const unsigned long dim415JoeKuoD5Init[]    =   {   1   ,   1   ,   5   ,   9   ,   23  ,   59  ,   117 ,   47  ,   379 ,   349 ,   1967    ,   1895    ,0 };
            const unsigned long dim416JoeKuoD5Init[]    =   {   1   ,   1   ,   1   ,   15  ,   1   ,   3   ,   7   ,   7   ,   105 ,   703 ,   1777    ,   113 ,0 };
            const unsigned long dim417JoeKuoD5Init[]    =   {   1   ,   1   ,   5   ,   7   ,   7   ,   25  ,   69  ,   123 ,   257 ,   513 ,   41  ,   2689    ,0 };
            const unsigned long dim418JoeKuoD5Init[]    =   {   1   ,   3   ,   1   ,   15  ,   7   ,   9   ,   11  ,   67  ,   27  ,   283 ,   1139    ,   1961    ,0 };
            const unsigned long dim419JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   13  ,   3   ,   5   ,   53  ,   251 ,   139 ,   913 ,   267 ,   1931    ,0 };
            const unsigned long dim420JoeKuoD5Init[]    =   {   1   ,   1   ,   1   ,   13  ,   3   ,   11  ,   79  ,   211 ,   27  ,   551 ,   339 ,   3383    ,0 };
            const unsigned long dim421JoeKuoD5Init[]    =   {   1   ,   1   ,   1   ,   1   ,   9   ,   47  ,   111 ,   47  ,   399 ,   353 ,   1707    ,   603 ,0 };
            const unsigned long dim422JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   15  ,   1   ,   43  ,   17  ,   19  ,   335 ,   713 ,   645 ,   3227    ,0 };
            const unsigned long dim423JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   9   ,   5   ,   27  ,   17  ,   209 ,   363 ,   821 ,   1365    ,   143 ,0 };
            const unsigned long dim424JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   7   ,   29  ,   11  ,   47  ,   253 ,   421 ,   599 ,   465 ,   2413    ,0 };
            const unsigned long dim425JoeKuoD5Init[]    =   {   1   ,   1   ,   5   ,   9   ,   17  ,   39  ,   5   ,   47  ,   315 ,   645 ,   713 ,   4023    ,0 };
            const unsigned long dim426JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   9   ,   9   ,   3   ,   11  ,   45  ,   9   ,   831 ,   1513    ,   2655    ,0 };
            const unsigned long dim427JoeKuoD5Init[]    =   {   1   ,   1   ,   5   ,   3   ,   29  ,   55  ,   113 ,   181 ,   281 ,   329 ,   193 ,   2969    ,0 };
            const unsigned long dim428JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   7   ,   9   ,   15  ,   29  ,   77  ,   11  ,   627 ,   1191    ,   3589    ,0 };
            const unsigned long dim429JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   3   ,   27  ,   53  ,   83  ,   13  ,   409 ,   931 ,   1581    ,   371 ,0 };
            const unsigned long dim430JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   3   ,   11  ,   7   ,   89  ,   143 ,   369 ,   519 ,   947 ,   2047    ,0 };
            const unsigned long dim431JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   11  ,   27  ,   59  ,   27  ,   75  ,   199 ,   965 ,   1669    ,   1713    ,0 };
            const unsigned long dim432JoeKuoD5Init[]    =   {   1   ,   3   ,   1   ,   1   ,   23  ,   63  ,   91  ,   159 ,   193 ,   355 ,   653 ,   2659    ,0 };
            const unsigned long dim433JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   11  ,   23  ,   37  ,   15  ,   73  ,   457 ,   789 ,   1207    ,   2573    ,0 };
            const unsigned long dim434JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   7   ,   13  ,   23  ,   71  ,   171 ,   479 ,   183 ,   1285    ,   1649    ,0 };
            const unsigned long dim435JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   1   ,   19  ,   13  ,   89  ,   5   ,   319 ,   15  ,   857 ,   3175    ,0 };
            const unsigned long dim436JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   7   ,   29  ,   7   ,   1   ,   249 ,   191 ,   237 ,   683 ,   1261    ,0 };
            const unsigned long dim437JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   7   ,   17  ,   53  ,   99  ,   119 ,   35  ,   63  ,   1845    ,   2681    ,0 };
            const unsigned long dim438JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   7   ,   23  ,   53  ,   39  ,   157 ,   323 ,   537 ,   1989    ,   1233    ,0 };
            const unsigned long dim439JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   7   ,   25  ,   19  ,   9   ,   67  ,   315 ,   499 ,   919 ,   2299    ,0 };
            const unsigned long dim440JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   9   ,   21  ,   49  ,   109 ,   185 ,   403 ,   179 ,   1967    ,   1185    ,0 };
            const unsigned long dim441JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   1   ,   27  ,   53  ,   33  ,   203 ,   179 ,   515 ,   1867    ,   1775    ,0 };
            const unsigned long dim442JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   15  ,   19  ,   25  ,   23  ,   77  ,   51  ,   467 ,   143 ,   1585    ,0 };
            const unsigned long dim443JoeKuoD5Init[]    =   {   1   ,   1   ,   1   ,   15  ,   19  ,   35  ,   41  ,   97  ,   407 ,   319 ,   1175    ,   241 ,0 };
            const unsigned long dim444JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   3   ,   29  ,   51  ,   41  ,   91  ,   223 ,   671 ,   729 ,   2009    ,0 };
            const unsigned long dim445JoeKuoD5Init[]    =   {   1   ,   1   ,   5   ,   7   ,   7   ,   55  ,   125 ,   75  ,   425 ,   699 ,   1837    ,   1515    ,0 };
            const unsigned long dim446JoeKuoD5Init[]    =   {   1   ,   1   ,   5   ,   1   ,   27  ,   59  ,   59  ,   235 ,   43  ,   77  ,   1433    ,   3689    ,0 };
            const unsigned long dim447JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   9   ,   19  ,   59  ,   69  ,   85  ,   199 ,   173 ,   1947    ,   1383    ,0 };
            const unsigned long dim448JoeKuoD5Init[]    =   {   1   ,   3   ,   1   ,   13  ,   29  ,   47  ,   121 ,   131 ,   45  ,   341 ,   85  ,   257 ,0 };
            const unsigned long dim449JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   1   ,   17  ,   17  ,   113 ,   103 ,   407 ,   815 ,   225 ,   2267    ,0 };
            const unsigned long dim450JoeKuoD5Init[]    =   {   1   ,   3   ,   1   ,   3   ,   23  ,   19  ,   65  ,   173 ,   475 ,   527 ,   271 ,   261 ,0 };
            const unsigned long dim451JoeKuoD5Init[]    =   {   1   ,   3   ,   1   ,   9   ,   23  ,   51  ,   5   ,   75  ,   403 ,   277 ,   1897    ,   353 ,0 };
            const unsigned long dim452JoeKuoD5Init[]    =   {   1   ,   1   ,   5   ,   3   ,   31  ,   49  ,   73  ,   93  ,   55  ,   99  ,   403 ,   659 ,0 };
            const unsigned long dim453JoeKuoD5Init[]    =   {   1   ,   3   ,   1   ,   11  ,   29  ,   17  ,   57  ,   141 ,   209 ,   907 ,   431 ,   2265    ,0 };
            const unsigned long dim454JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   15  ,   25  ,   55  ,   105 ,   61  ,   273 ,   201 ,   23  ,   1211    ,0 };
            const unsigned long dim455JoeKuoD5Init[]    =   {   1   ,   1   ,   5   ,   5   ,   3   ,   63  ,   41  ,   121 ,   161 ,   713 ,   1885    ,   225 ,0 };
            const unsigned long dim456JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   11  ,   11  ,   53  ,   63  ,   175 ,   439 ,   1   ,   953 ,   481 ,0 };
            const unsigned long dim457JoeKuoD5Init[]    =   {   1   ,   3   ,   1   ,   7   ,   1   ,   27  ,   65  ,   189 ,   223 ,   659 ,   413 ,   3677    ,0 };
            const unsigned long dim458JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   5   ,   17  ,   9   ,   51  ,   39  ,   307 ,   811 ,   941 ,   2297    ,0 };
            const unsigned long dim459JoeKuoD5Init[]    =   {   1   ,   3   ,   1   ,   3   ,   1   ,   41  ,   77  ,   237 ,   47  ,   533 ,   1783    ,   1385    ,0 };
            const unsigned long dim460JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   13  ,   19  ,   39  ,   31  ,   249 ,   449 ,   639 ,   1789    ,   3479    ,0 };
            const unsigned long dim461JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   11  ,   9   ,   9   ,   9   ,   19  ,   481 ,   411 ,   1669    ,   863 ,0 };
            const unsigned long dim462JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   7   ,   15  ,   15  ,   89  ,   161 ,   171 ,   377 ,   2031    ,   389 ,0 };
            const unsigned long dim463JoeKuoD5Init[]    =   {   1   ,   1   ,   5   ,   5   ,   7   ,   49  ,   83  ,   81  ,   181 ,   395 ,   1197    ,   2455    ,0 };
            const unsigned long dim464JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   9   ,   15  ,   25  ,   21  ,   35  ,   287 ,   369 ,   693 ,   1753    ,0 };
            const unsigned long dim465JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   7   ,   5   ,   19  ,   41  ,   241 ,   459 ,   427 ,   631 ,   1109    ,0 };
            const unsigned long dim466JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   3   ,   23  ,   13  ,   115 ,   55  ,   145 ,   933 ,   1985    ,   753 ,0 };
            const unsigned long dim467JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   5   ,   19  ,   15  ,   73  ,   67  ,   335 ,   583 ,   315 ,   1559    ,0 };
            const unsigned long dim468JoeKuoD5Init[]    =   {   1   ,   1   ,   5   ,   5   ,   23  ,   41  ,   65  ,   155 ,   171 ,   1017    ,   1283    ,   1989    ,0 };
            const unsigned long dim469JoeKuoD5Init[]    =   {   1   ,   1   ,   5   ,   13  ,   23  ,   53  ,   75  ,   91  ,   253 ,   41  ,   101 ,   1943    ,0 };
            const unsigned long dim470JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   13  ,   25  ,   13  ,   5   ,   71  ,   437 ,   553 ,   701 ,   805 ,0 };
            const unsigned long dim471JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   5   ,   3   ,   3   ,   61  ,   255 ,   409 ,   753 ,   1265    ,   2739    ,0 };
            const unsigned long dim472JoeKuoD5Init[]    =   {   1   ,   3   ,   1   ,   11  ,   25  ,   33  ,   69  ,   41  ,   367 ,   133 ,   809 ,   1421    ,0 };
            const unsigned long dim473JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   9   ,   3   ,   59  ,   15  ,   229 ,   313 ,   667 ,   45  ,   1485    ,0 };
            const unsigned long dim474JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   11  ,   3   ,   45  ,   73  ,   245 ,   47  ,   489 ,   1715    ,   3167    ,0 };
            const unsigned long dim475JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   15  ,   23  ,   23  ,   107 ,   1   ,   171 ,   993 ,   1617    ,   3665    ,0 };
            const unsigned long dim476JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   5   ,   5   ,   41  ,   9   ,   221 ,   91  ,   611 ,   639 ,   3709    ,0 };
            const unsigned long dim477JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   1   ,   17  ,   31  ,   55  ,   23  ,   75  ,   501 ,   1605    ,   2361    ,0 };
            const unsigned long dim478JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   11  ,   19  ,   27  ,   37  ,   111 ,   393 ,   75  ,   523 ,   2079    ,0 };
            const unsigned long dim479JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   1   ,   13  ,   17  ,   103 ,   177 ,   331 ,   431 ,   419 ,   2781    ,0 };
            const unsigned long dim480JoeKuoD5Init[]    =   {   1   ,   1   ,   5   ,   1   ,   31  ,   57  ,   9   ,   87  ,   111 ,   761 ,   1259    ,   3645    ,0 };
            const unsigned long dim481JoeKuoD5Init[]    =   {   1   ,   3   ,   1   ,   1   ,   29  ,   39  ,   117 ,   207 ,   25  ,   263 ,   1875    ,   3325    ,   1773    ,0 };
            const unsigned long dim482JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   11  ,   21  ,   55  ,   105 ,   43  ,   155 ,   933 ,   585 ,   1617    ,   1705    ,0 };
            const unsigned long dim483JoeKuoD5Init[]    =   {   1   ,   1   ,   5   ,   5   ,   29  ,   37  ,   67  ,   165 ,   229 ,   517 ,   71  ,   3927    ,   1131    ,0 };
            const unsigned long dim484JoeKuoD5Init[]    =   {   1   ,   1   ,   1   ,   3   ,   1   ,   15  ,   7   ,   139 ,   431 ,   599 ,   101 ,   1167    ,   55  ,0 };
            const unsigned long dim485JoeKuoD5Init[]    =   {   1   ,   3   ,   1   ,   3   ,   5   ,   31  ,   15  ,   99  ,   375 ,   491 ,   293 ,   2521    ,   1599    ,0 };
            const unsigned long dim486JoeKuoD5Init[]    =   {   1   ,   1   ,   1   ,   13  ,   29  ,   59  ,   99  ,   155 ,   295 ,   21  ,   1459    ,   2263    ,   1997    ,0 };
            const unsigned long dim487JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   5   ,   29  ,   61  ,   103 ,   151 ,   37  ,   431 ,   1893    ,   2835    ,   6509    ,0 };
            const unsigned long dim488JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   7   ,   17  ,   15  ,   1   ,   163 ,   481 ,   547 ,   701 ,   2957    ,   7071    ,0 };
            const unsigned long dim489JoeKuoD5Init[]    =   {   1   ,   1   ,   5   ,   5   ,   29  ,   41  ,   51  ,   223 ,   133 ,   49  ,   753 ,   3769    ,   8139    ,0 };
            const unsigned long dim490JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   1   ,   23  ,   23  ,   21  ,   107 ,   9   ,   445 ,   215 ,   857 ,   7913    ,0 };
            const unsigned long dim491JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   5   ,   17  ,   13  ,   11  ,   111 ,   419 ,   433 ,   1289    ,   2855    ,   2157    ,0 };
            const unsigned long dim492JoeKuoD5Init[]    =   {   1   ,   1   ,   5   ,   11  ,   31  ,   3   ,   97  ,   223 ,   143 ,   117 ,   563 ,   2179    ,   1053    ,0 };
            const unsigned long dim493JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   13  ,   7   ,   25  ,   115 ,   151 ,   181 ,   999 ,   1027    ,   795 ,   679 ,0 };
            const unsigned long dim494JoeKuoD5Init[]    =   {   1   ,   1   ,   5   ,   15  ,   31  ,   23  ,   85  ,   125 ,   135 ,   1001    ,   909 ,   339 ,   5693    ,0 };
            const unsigned long dim495JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   3   ,   29  ,   17  ,   105 ,   239 ,   467 ,   875 ,   1135    ,   1859    ,   6399    ,0 };
            const unsigned long dim496JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   13  ,   19  ,   59  ,   99  ,   29  ,   177 ,   879 ,   1817    ,   3747    ,   1855    ,0 };
            const unsigned long dim497JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   1   ,   23  ,   5   ,   29  ,   53  ,   111 ,   341 ,   1713    ,   2285    ,   7033    ,0 };
            const unsigned long dim498JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   5   ,   7   ,   55  ,   67  ,   173 ,   273 ,   881 ,   1405    ,   1663    ,   2135    ,0 };
            const unsigned long dim499JoeKuoD5Init[]    =   {   1   ,   3   ,   1   ,   9   ,   1   ,   39  ,   11  ,   63  ,   107 ,   905 ,   629 ,   1773    ,   1059    ,0 };
            const unsigned long dim500JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   15  ,   29  ,   51  ,   111 ,   189 ,   337 ,   505 ,   453 ,   1549    ,   3697    ,0 };
            const unsigned long dim501JoeKuoD5Init[]    =   {   1   ,   3   ,   1   ,   9   ,   23  ,   57  ,   21  ,   61  ,   161 ,   695 ,   1097    ,   809 ,   5737    ,0 };
            const unsigned long dim502JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   15  ,   15  ,   55  ,   93  ,   65  ,   101 ,   521 ,   1273    ,   1949    ,   7325    ,0 };
            const unsigned long dim503JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   3   ,   23  ,   31  ,   37  ,   51  ,   205 ,   261 ,   647 ,   1905    ,   4407    ,0 };
            const unsigned long dim504JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   3   ,   3   ,   9   ,   91  ,   51  ,   271 ,   623 ,   1611    ,   563 ,   4687    ,0 };
            const unsigned long dim505JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   3   ,   11  ,   61  ,   95  ,   215 ,   347 ,   171 ,   519 ,   2331    ,   2189    ,0 };
            const unsigned long dim506JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   1   ,   25  ,   13  ,   87  ,   159 ,   87  ,   915 ,   463 ,   1345    ,   5901    ,0 };
            const unsigned long dim507JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   11  ,   31  ,   21  ,   81  ,   75  ,   153 ,   337 ,   2025    ,   233 ,   4999    ,0 };
            const unsigned long dim508JoeKuoD5Init[]    =   {   1   ,   1   ,   1   ,   11  ,   23  ,   25  ,   81  ,   149 ,   225 ,   799 ,   159 ,   799 ,   687 ,0 };
            const unsigned long dim509JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   5   ,   23  ,   55  ,   67  ,   47  ,   375 ,   657 ,   877 ,   1505    ,   1757    ,0 };
            const unsigned long dim510JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   9   ,   17  ,   7   ,   123 ,   71  ,   203 ,   457 ,   201 ,   9   ,   6671    ,0 };
            const unsigned long dim511JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   3   ,   1   ,   55  ,   7   ,   133 ,   65  ,   891 ,   1705    ,   389 ,   4601    ,0 };
            const unsigned long dim512JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   3   ,   21  ,   59  ,   101 ,   105 ,   241 ,   231 ,   363 ,   4029    ,   1279    ,0 };
            const unsigned long dim513JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   11  ,   31  ,   61  ,   115 ,   219 ,   249 ,   575 ,   201 ,   547 ,   5315    ,0 };
            const unsigned long dim514JoeKuoD5Init[]    =   {   1   ,   1   ,   5   ,   11  ,   11  ,   3   ,   75  ,   219 ,   183 ,   771 ,   725 ,   2175    ,   4077    ,0 };
            const unsigned long dim515JoeKuoD5Init[]    =   {   1   ,   3   ,   1   ,   13  ,   1   ,   37  ,   1   ,   165 ,   431 ,   423 ,   2021    ,   475 ,   5151    ,0 };
            const unsigned long dim516JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   7   ,   15  ,   59  ,   25  ,   133 ,   377 ,   747 ,   23  ,   1195    ,   3303    ,0 };
            const unsigned long dim517JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   15  ,   23  ,   63  ,   121 ,   159 ,   403 ,   143 ,   187 ,   1481    ,   4755    ,0 };
            const unsigned long dim518JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   13  ,   21  ,   29  ,   55  ,   165 ,   483 ,   495 ,   579 ,   1197    ,   4841    ,0 };
            const unsigned long dim519JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   3   ,   27  ,   7   ,   111 ,   57  ,   353 ,   1023    ,   1593    ,   1447    ,   5819    ,0 };
            const unsigned long dim520JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   3   ,   13  ,   37  ,   115 ,   65  ,   23  ,   707 ,   603 ,   1805    ,   6011    ,0 };
            const unsigned long dim521JoeKuoD5Init[]    =   {   1   ,   1   ,   5   ,   15  ,   9   ,   47  ,   87  ,   195 ,   125 ,   515 ,   1885    ,   89  ,   1377    ,0 };
            const unsigned long dim522JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   11  ,   1   ,   25  ,   11  ,   73  ,   183 ,   897 ,   981 ,   275 ,   331 ,0 };
            const unsigned long dim523JoeKuoD5Init[]    =   {   1   ,   1   ,   5   ,   15  ,   13  ,   53  ,   125 ,   37  ,   145 ,   763 ,   1991    ,   1971    ,   4385    ,0 };
            const unsigned long dim524JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   1   ,   29  ,   53  ,   69  ,   97  ,   415 ,   151 ,   1389    ,   2867    ,   3085    ,0 };
            const unsigned long dim525JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   7   ,   21  ,   51  ,   77  ,   115 ,   81  ,   197 ,   91  ,   3417    ,   2357    ,0 };
            const unsigned long dim526JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   11  ,   31  ,   31  ,   21  ,   21  ,   93  ,   221 ,   1401    ,   3253    ,   6875    ,0 };
            const unsigned long dim527JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   9   ,   11  ,   17  ,   87  ,   75  ,   333 ,   871 ,   1679    ,   2943    ,   4803    ,0 };
            const unsigned long dim528JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   11  ,   11  ,   61  ,   67  ,   141 ,   79  ,   757 ,   965 ,   1999    ,   6363    ,0 };
            const unsigned long dim529JoeKuoD5Init[]    =   {   1   ,   1   ,   5   ,   5   ,   27  ,   13  ,   109 ,   137 ,   235 ,   1007    ,   1307    ,   341 ,   3957    ,0 };
            const unsigned long dim530JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   9   ,   15  ,   37  ,   47  ,   247 ,   295 ,   867 ,   1433    ,   553 ,   5365    ,0 };
            const unsigned long dim531JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   3   ,   13  ,   29  ,   77  ,   155 ,   423 ,   823 ,   1117    ,   3939    ,   1423    ,0 };
            const unsigned long dim532JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   9   ,   17  ,   27  ,   47  ,   73  ,   79  ,   329 ,   1473    ,   3241    ,   697 ,0 };
            const unsigned long dim533JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   9   ,   23  ,   5   ,   47  ,   89  ,   427 ,   893 ,   2031    ,   3415    ,   6367    ,0 };
            const unsigned long dim534JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   1   ,   17  ,   47  ,   31  ,   113 ,   461 ,   417 ,   2017    ,   41  ,   2417    ,0 };
            const unsigned long dim535JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   5   ,   11  ,   35  ,   119 ,   95  ,   389 ,   31  ,   871 ,   563 ,   7547    ,0 };
            const unsigned long dim536JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   9   ,   3   ,   49  ,   63  ,   237 ,   511 ,   619 ,   589 ,   3571    ,   1883    ,0 };
            const unsigned long dim537JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   1   ,   1   ,   29  ,   17  ,   117 ,   173 ,   399 ,   443 ,   2625    ,   2009    ,0 };
            const unsigned long dim538JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   9   ,   23  ,   47  ,   5   ,   167 ,   413 ,   513 ,   509 ,   853 ,   3509    ,0 };
            const unsigned long dim539JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   5   ,   13  ,   15  ,   33  ,   165 ,   21  ,   163 ,   1613    ,   3387    ,   645 ,0 };
            const unsigned long dim540JoeKuoD5Init[]    =   {   1   ,   1   ,   1   ,   3   ,   7   ,   33  ,   59  ,   25  ,   65  ,   243 ,   1253    ,   1893    ,   1637    ,0 };
            const unsigned long dim541JoeKuoD5Init[]    =   {   1   ,   1   ,   5   ,   15  ,   21  ,   63  ,   51  ,   167 ,   131 ,   171 ,   651 ,   295 ,   5775    ,0 };
            const unsigned long dim542JoeKuoD5Init[]    =   {   1   ,   1   ,   5   ,   9   ,   7   ,   23  ,   31  ,   171 ,   85  ,   859 ,   1691    ,   2757    ,   1351    ,0 };
            const unsigned long dim543JoeKuoD5Init[]    =   {   1   ,   1   ,   5   ,   3   ,   31  ,   7   ,   25  ,   69  ,   183 ,   417 ,   39  ,   2671    ,   5197    ,0 };
            const unsigned long dim544JoeKuoD5Init[]    =   {   1   ,   1   ,   5   ,   9   ,   17  ,   21  ,   57  ,   145 ,   23  ,   933 ,   2031    ,   65  ,   4583    ,0 };
            const unsigned long dim545JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   5   ,   1   ,   59  ,   117 ,   191 ,   197 ,   627 ,   659 ,   2873    ,   3865    ,0 };
            const unsigned long dim546JoeKuoD5Init[]    =   {   1   ,   1   ,   5   ,   15  ,   23  ,   51  ,   45  ,   47  ,   147 ,   779 ,   1619    ,   1017    ,   3769    ,0 };
            const unsigned long dim547JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   3   ,   9   ,   1   ,   75  ,   151 ,   117 ,   483 ,   1499    ,   2143    ,   5873    ,0 };
            const unsigned long dim548JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   1   ,   13  ,   31  ,   105 ,   115 ,   199 ,   111 ,   1403    ,   1833    ,   7923    ,0 };
            const unsigned long dim549JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   7   ,   29  ,   53  ,   121 ,   149 ,   419 ,   107 ,   1299    ,   1925    ,   4409    ,0 };
            const unsigned long dim550JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   11  ,   21  ,   25  ,   63  ,   97  ,   145 ,   71  ,   1693    ,   465 ,   5607    ,0 };
            const unsigned long dim551JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   15  ,   25  ,   43  ,   77  ,   177 ,   53  ,   495 ,   1983    ,   4083    ,   2107    ,0 };
            const unsigned long dim552JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   11  ,   7   ,   51  ,   109 ,   29  ,   171 ,   847 ,   673 ,   2929    ,   3887    ,0 };
            const unsigned long dim553JoeKuoD5Init[]    =   {   1   ,   1   ,   5   ,   9   ,   31  ,   47  ,   63  ,   225 ,   371 ,   453 ,   1075    ,   2293    ,   3323    ,0 };
            const unsigned long dim554JoeKuoD5Init[]    =   {   1   ,   3   ,   1   ,   9   ,   31  ,   29  ,   67  ,   227 ,   135 ,   369 ,   481 ,   187 ,   3237    ,0 };
            const unsigned long dim555JoeKuoD5Init[]    =   {   1   ,   3   ,   1   ,   7   ,   1   ,   29  ,   49  ,   157 ,   99  ,   741 ,   279 ,   1963    ,   7881    ,0 };
            const unsigned long dim556JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   13  ,   17  ,   49  ,   33  ,   73  ,   103 ,   941 ,   209 ,   1329    ,   3   ,0 };
            const unsigned long dim557JoeKuoD5Init[]    =   {   1   ,   1   ,   1   ,   1   ,   17  ,   41  ,   53  ,   57  ,   163 ,   761 ,   1855    ,   3423    ,   5317    ,0 };
            const unsigned long dim558JoeKuoD5Init[]    =   {   1   ,   1   ,   5   ,   1   ,   11  ,   13  ,   59  ,   37  ,   351 ,   561 ,   1213    ,   2355    ,   8095    ,0 };
            const unsigned long dim559JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   3   ,   3   ,   31  ,   47  ,   237 ,   101 ,   167 ,   1623    ,   645 ,   4787    ,0 };
            const unsigned long dim560JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   9   ,   21  ,   33  ,   55  ,   15  ,   433 ,   129 ,   279 ,   2131    ,   2943    ,0 };
            const unsigned long dim561JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   13  ,   9   ,   55  ,   71  ,   151 ,   273 ,   901 ,   427 ,   3749    ,   8163    ,0 };
            const unsigned long dim562JoeKuoD5Init[]    =   {   1   ,   1   ,   1   ,   13  ,   13  ,   63  ,   11  ,   63  ,   477 ,   743 ,   1391    ,   2045    ,   6985    ,0 };
            const unsigned long dim563JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   11  ,   31  ,   25  ,   93  ,   217 ,   39  ,   263 ,   1411    ,   3   ,   7313    ,0 };
            const unsigned long dim564JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   1   ,   21  ,   13  ,   3   ,   255 ,   107 ,   851 ,   1281    ,   959 ,   3955    ,0 };
            const unsigned long dim565JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   9   ,   19  ,   55  ,   53  ,   201 ,   199 ,   361 ,   805 ,   579 ,   1459    ,0 };
            const unsigned long dim566JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   15  ,   9   ,   59  ,   109 ,   245 ,   29  ,   21  ,   137 ,   717 ,   607 ,0 };
            const unsigned long dim567JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   11  ,   15  ,   23  ,   49  ,   3   ,   195 ,   185 ,   85  ,   3885    ,   5859    ,0 };
            const unsigned long dim568JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   3   ,   13  ,   21  ,   7   ,   65  ,   185 ,   541 ,   305 ,   79  ,   3125    ,0 };
            const unsigned long dim569JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   3   ,   7   ,   59  ,   11  ,   125 ,   127 ,   283 ,   943 ,   3545    ,   1617    ,0 };
            const unsigned long dim570JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   3   ,   19  ,   39  ,   73  ,   167 ,   431 ,   147 ,   3   ,   1099    ,   6311    ,0 };
            const unsigned long dim571JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   13  ,   29  ,   57  ,   109 ,   169 ,   49  ,   457 ,   469 ,   3093    ,   7505    ,0 };
            const unsigned long dim572JoeKuoD5Init[]    =   {   1   ,   3   ,   1   ,   1   ,   15  ,   29  ,   69  ,   133 ,   423 ,   737 ,   673 ,   2529    ,   2065    ,0 };
            const unsigned long dim573JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   7   ,   21  ,   47  ,   15  ,   175 ,   17  ,   419 ,   1917    ,   1183    ,   429 ,0 };
            const unsigned long dim574JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   9   ,   31  ,   43  ,   19  ,   63  ,   395 ,   331 ,   385 ,   3879    ,   3233    ,0 };
            const unsigned long dim575JoeKuoD5Init[]    =   {   1   ,   1   ,   1   ,   9   ,   5   ,   11  ,   101 ,   65  ,   315 ,   805 ,   719 ,   641 ,   343 ,0 };
            const unsigned long dim576JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   1   ,   15  ,   5   ,   17  ,   115 ,   503 ,   395 ,   531 ,   1201    ,   7225    ,0 };
            const unsigned long dim577JoeKuoD5Init[]    =   {   1   ,   3   ,   1   ,   7   ,   3   ,   35  ,   29  ,   1   ,   297 ,   421 ,   1365    ,   1491    ,   7973    ,0 };
            const unsigned long dim578JoeKuoD5Init[]    =   {   1   ,   1   ,   1   ,   13  ,   1   ,   59  ,   79  ,   13  ,   337 ,   717 ,   1229    ,   2587    ,   5659    ,0 };
            const unsigned long dim579JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   9   ,   21  ,   15  ,   27  ,   19  ,   195 ,   27  ,   267 ,   381 ,   1969    ,0 };
            const unsigned long dim580JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   7   ,   25  ,   63  ,   119 ,   155 ,   243 ,   229 ,   897 ,   629 ,   7515    ,0 };
            const unsigned long dim581JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   1   ,   13  ,   37  ,   35  ,   31  ,   485 ,   729 ,   123 ,   1645    ,   457 ,0 };
            const unsigned long dim582JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   11  ,   17  ,   15  ,   47  ,   149 ,   311 ,   189 ,   1925    ,   9   ,   7639    ,0 };
            const unsigned long dim583JoeKuoD5Init[]    =   {   1   ,   1   ,   1   ,   3   ,   1   ,   1   ,   127 ,   197 ,   109 ,   49  ,   265 ,   3643    ,   3629    ,0 };
            const unsigned long dim584JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   15  ,   21  ,   49  ,   71  ,   187 ,   189 ,   631 ,   1449    ,   775 ,   5973    ,0 };
            const unsigned long dim585JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   5   ,   27  ,   15  ,   17  ,   137 ,   393 ,   807 ,   1189    ,   2731    ,   6337    ,0 };
            const unsigned long dim586JoeKuoD5Init[]    =   {   1   ,   3   ,   1   ,   11  ,   3   ,   43  ,   3   ,   77  ,   487 ,   539 ,   1781    ,   3261    ,   2775    ,0 };
            const unsigned long dim587JoeKuoD5Init[]    =   {   1   ,   1   ,   1   ,   13  ,   29  ,   31  ,   83  ,   225 ,   159 ,   971 ,   1899    ,   1035    ,   5383    ,0 };
            const unsigned long dim588JoeKuoD5Init[]    =   {   1   ,   3   ,   1   ,   7   ,   27  ,   1   ,   15  ,   141 ,   485 ,   639 ,   1895    ,   3129    ,   4489    ,0 };
            const unsigned long dim589JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   5   ,   9   ,   17  ,   21  ,   231 ,   363 ,   637 ,   1851    ,   3675    ,   5371    ,0 };
            const unsigned long dim590JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   11  ,   9   ,   17  ,   91  ,   243 ,   51  ,   565 ,   491 ,   3333    ,   3329    ,0 };
            const unsigned long dim591JoeKuoD5Init[]    =   {   1   ,   1   ,   5   ,   3   ,   13  ,   9   ,   19  ,   227 ,   353 ,   111 ,   1805    ,   3917    ,   6849    ,0 };
            const unsigned long dim592JoeKuoD5Init[]    =   {   1   ,   1   ,   1   ,   1   ,   29  ,   31  ,   27  ,   57  ,   421 ,   155 ,   1385    ,   999 ,   1581    ,0 };
            const unsigned long dim593JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   7   ,   1   ,   55  ,   35  ,   35  ,   311 ,   357 ,   1569    ,   2693    ,   2251    ,0 };
            const unsigned long dim594JoeKuoD5Init[]    =   {   1   ,   3   ,   1   ,   9   ,   27  ,   41  ,   111 ,   119 ,   265 ,   165 ,   1999    ,   2067    ,   7801    ,0 };
            const unsigned long dim595JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   15  ,   31  ,   55  ,   31  ,   39  ,   305 ,   581 ,   373 ,   2523    ,   2153    ,0 };
            const unsigned long dim596JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   7   ,   9   ,   19  ,   115 ,   41  ,   261 ,   209 ,   897 ,   409 ,   5201    ,0 };
            const unsigned long dim597JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   13  ,   3   ,   15  ,   95  ,   143 ,   407 ,   719 ,   1763    ,   1763    ,   1173    ,0 };
            const unsigned long dim598JoeKuoD5Init[]    =   {   1   ,   1   ,   5   ,   11  ,   17  ,   13  ,   69  ,   17  ,   293 ,   815 ,   1361    ,   259 ,   6751    ,0 };
            const unsigned long dim599JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   15  ,   11  ,   53  ,   13  ,   195 ,   153 ,   445 ,   1873    ,   1159    ,   4739    ,0 };
            const unsigned long dim600JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   7   ,   3   ,   25  ,   57  ,   229 ,   269 ,   299 ,   1687    ,   2707    ,   7049    ,0 };
            const unsigned long dim601JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   1   ,   5   ,   23  ,   89  ,   171 ,   207 ,   523 ,   2031    ,   2513    ,   2475    ,0 };
            const unsigned long dim602JoeKuoD5Init[]    =   {   1   ,   1   ,   1   ,   7   ,   25  ,   57  ,   125 ,   109 ,   203 ,   671 ,   781 ,   295 ,   4001    ,0 };
            const unsigned long dim603JoeKuoD5Init[]    =   {   1   ,   1   ,   1   ,   3   ,   17  ,   37  ,   51  ,   169 ,   441 ,   797 ,   871 ,   3267    ,   5695    ,0 };
            const unsigned long dim604JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   5   ,   17  ,   25  ,   97  ,   41  ,   377 ,   643 ,   1463    ,   141 ,   3961    ,0 };
            const unsigned long dim605JoeKuoD5Init[]    =   {   1   ,   1   ,   5   ,   1   ,   17  ,   35  ,   111 ,   91  ,   253 ,   237 ,   1491    ,   2839    ,   2265    ,0 };
            const unsigned long dim606JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   13  ,   23  ,   7   ,   47  ,   61  ,   263 ,   591 ,   1365    ,   2371    ,   4209    ,0 };
            const unsigned long dim607JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   5   ,   7   ,   51  ,   117 ,   161 ,   383 ,   303 ,   1765    ,   3105    ,   3961    ,0 };
            const unsigned long dim608JoeKuoD5Init[]    =   {   1   ,   1   ,   5   ,   11  ,   11  ,   55  ,   111 ,   55  ,   417 ,   713 ,   305 ,   1781    ,   5283    ,0 };
            const unsigned long dim609JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   11  ,   25  ,   51  ,   17  ,   215 ,   335 ,   47  ,   1789    ,   2049    ,   5349    ,0 };
            const unsigned long dim610JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   3   ,   1   ,   9   ,   71  ,   105 ,   397 ,   517 ,   1093    ,   765 ,   5301    ,0 };
            const unsigned long dim611JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   9   ,   31  ,   41  ,   95  ,   153 ,   383 ,   91  ,   1649    ,   3059    ,   6135    ,0 };
            const unsigned long dim612JoeKuoD5Init[]    =   {   1   ,   1   ,   5   ,   11  ,   19  ,   51  ,   67  ,   119 ,   507 ,   179 ,   571 ,   2767    ,   5517    ,0 };
            const unsigned long dim613JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   9   ,   13  ,   19  ,   35  ,   249 ,   39  ,   425 ,   233 ,   1635    ,   5915    ,0 };
            const unsigned long dim614JoeKuoD5Init[]    =   {   1   ,   1   ,   1   ,   11  ,   17  ,   15  ,   43  ,   29  ,   351 ,   25  ,   1879    ,   3941    ,   189 ,0 };
            const unsigned long dim615JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   5   ,   13  ,   55  ,   9   ,   7   ,   91  ,   951 ,   1681    ,   2723    ,   4349    ,0 };
            const unsigned long dim616JoeKuoD5Init[]    =   {   1   ,   3   ,   1   ,   9   ,   27  ,   31  ,   49  ,   33  ,   287 ,   629 ,   851 ,   1353    ,   6391    ,0 };
            const unsigned long dim617JoeKuoD5Init[]    =   {   1   ,   1   ,   5   ,   7   ,   19  ,   27  ,   45  ,   209 ,   257 ,   141 ,   1771    ,   931 ,   7839    ,0 };
            const unsigned long dim618JoeKuoD5Init[]    =   {   1   ,   1   ,   5   ,   1   ,   9   ,   47  ,   87  ,   71  ,   183 ,   249 ,   311 ,   1989    ,   1753    ,0 };
            const unsigned long dim619JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   5   ,   7   ,   35  ,   45  ,   199 ,   207 ,   203 ,   831 ,   2643    ,   1155    ,0 };
            const unsigned long dim620JoeKuoD5Init[]    =   {   1   ,   1   ,   1   ,   11  ,   5   ,   11  ,   27  ,   187 ,   405 ,   747 ,   261 ,   1279    ,   6153    ,0 };
            const unsigned long dim621JoeKuoD5Init[]    =   {   1   ,   1   ,   5   ,   9   ,   7   ,   43  ,   23  ,   117 ,   421 ,   775 ,   1657    ,   1071    ,   4551    ,0 };
            const unsigned long dim622JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   9   ,   1   ,   51  ,   5   ,   27  ,   121 ,   459 ,   1251    ,   901 ,   2301    ,0 };
            const unsigned long dim623JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   11  ,   15  ,   47  ,   107 ,   93  ,   79  ,   719 ,   571 ,   65  ,   7589    ,0 };
            const unsigned long dim624JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   3   ,   19  ,   33  ,   103 ,   253 ,   469 ,   109 ,   913 ,   2251    ,   4737    ,0 };
            const unsigned long dim625JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   5   ,   23  ,   29  ,   89  ,   79  ,   253 ,   513 ,   723 ,   3823    ,   5769    ,0 };
            const unsigned long dim626JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   1   ,   27  ,   3   ,   103 ,   171 ,   353 ,   673 ,   1147    ,   529 ,   4737    ,0 };
            const unsigned long dim627JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   7   ,   27  ,   5   ,   103 ,   61  ,   101 ,   759 ,   443 ,   2003    ,   5537    ,0 };
            const unsigned long dim628JoeKuoD5Init[]    =   {   1   ,   1   ,   1   ,   5   ,   11  ,   15  ,   109 ,   119 ,   473 ,   585 ,   1759    ,   319 ,   5461    ,0 };
            const unsigned long dim629JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   9   ,   19  ,   61  ,   5   ,   255 ,   171 ,   843 ,   823 ,   2713    ,   5313    ,0 };
            const unsigned long dim630JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   3   ,   9   ,   11  ,   57  ,   3   ,   365 ,   471 ,   1179    ,   1999    ,   3333    ,0 };
            const unsigned long dim631JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   9   ,   1   ,   3   ,   3   ,   195 ,   441 ,   193 ,   1905    ,   1753    ,   1839    ,0 };
            const unsigned long dim632JoeKuoD5Init[]    =   {   1   ,   3   ,   1   ,   3   ,   1   ,   49  ,   99  ,   85  ,   175 ,   603 ,   1569    ,   2201    ,   1979    ,0 };
            const unsigned long dim633JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   5   ,   27  ,   37  ,   61  ,   137 ,   219 ,   469 ,   973 ,   1979    ,   1135    ,0 };
            const unsigned long dim634JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   13  ,   3   ,   59  ,   11  ,   203 ,   415 ,   513 ,   1469    ,   1655    ,   5913    ,0 };
            const unsigned long dim635JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   7   ,   21  ,   23  ,   87  ,   83  ,   21  ,   351 ,   899 ,   1633    ,   6589    ,0 };
            const unsigned long dim636JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   5   ,   5   ,   63  ,   45  ,   201 ,   193 ,   27  ,   1365    ,   1197    ,   1729    ,0 };
            const unsigned long dim637JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   5   ,   1   ,   35  ,   59  ,   157 ,   295 ,   359 ,   383 ,   3191    ,   7019    ,0 };
            const unsigned long dim638JoeKuoD5Init[]    =   {   1   ,   1   ,   1   ,   13  ,   3   ,   23  ,   17  ,   149 ,   59  ,   115 ,   1101    ,   1879    ,   4243    ,0 };
            const unsigned long dim639JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   5   ,   3   ,   21  ,   71  ,   31  ,   85  ,   93  ,   1691    ,   379 ,   7901    ,0 };
            const unsigned long dim640JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   13  ,   15  ,   7   ,   29  ,   59  ,   191 ,   817 ,   439 ,   453 ,   5073    ,0 };
            const unsigned long dim641JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   9   ,   29  ,   7   ,   119 ,   35  ,   393 ,   9   ,   509 ,   3907    ,   7031    ,0 };
            const unsigned long dim642JoeKuoD5Init[]    =   {   1   ,   3   ,   1   ,   3   ,   29  ,   51  ,   19  ,   127 ,   399 ,   309 ,   117 ,   3491    ,   5417    ,0 };
            const unsigned long dim643JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   5   ,   31  ,   33  ,   17  ,   119 ,   365 ,   301 ,   527 ,   3341    ,   779 ,0 };
            const unsigned long dim644JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   1   ,   29  ,   41  ,   43  ,   85  ,   133 ,   191 ,   229 ,   3407    ,   3147    ,0 };
            const unsigned long dim645JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   3   ,   7   ,   9   ,   49  ,   121 ,   193 ,   569 ,   467 ,   999 ,   6813    ,0 };
            const unsigned long dim646JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   13  ,   7   ,   23  ,   121 ,   43  ,   173 ,   761 ,   525 ,   3221    ,   5435    ,0 };
            const unsigned long dim647JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   3   ,   15  ,   43  ,   47  ,   149 ,   227 ,   357 ,   1219    ,   4087    ,   1215    ,0 };
            const unsigned long dim648JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   15  ,   1   ,   33  ,   81  ,   195 ,   201 ,   307 ,   1081    ,   3201    ,   4293    ,0 };
            const unsigned long dim649JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   11  ,   11  ,   1   ,   125 ,   119 ,   105 ,   783 ,   117 ,   3465    ,   5713    ,0 };
            const unsigned long dim650JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   1   ,   1   ,   51  ,   17  ,   141 ,   107 ,   875 ,   1135    ,   1213    ,   7113    ,0 };
            const unsigned long dim651JoeKuoD5Init[]    =   {   1   ,   1   ,   1   ,   1   ,   7   ,   41  ,   63  ,   91  ,   465 ,   893 ,   1663    ,   2717    ,   6313    ,0 };
            const unsigned long dim652JoeKuoD5Init[]    =   {   1   ,   3   ,   1   ,   15  ,   23  ,   43  ,   77  ,   49  ,   131 ,   953 ,   1591    ,   869 ,   3779    ,0 };
            const unsigned long dim653JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   15  ,   5   ,   9   ,   41  ,   183 ,   403 ,   775 ,   1163    ,   2963    ,   861 ,0 };
            const unsigned long dim654JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   11  ,   27  ,   19  ,   51  ,   139 ,   87  ,   315 ,   831 ,   2587    ,   4847    ,0 };
            const unsigned long dim655JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   13  ,   23  ,   23  ,   117 ,   189 ,   405 ,   735 ,   681 ,   457 ,   337 ,0 };
            const unsigned long dim656JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   5   ,   13  ,   21  ,   25  ,   207 ,   179 ,   715 ,   629 ,   593 ,   6351    ,0 };
            const unsigned long dim657JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   3   ,   25  ,   13  ,   31  ,   245 ,   147 ,   953 ,   1061    ,   3749    ,   6927    ,0 };
            const unsigned long dim658JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   9   ,   29  ,   27  ,   57  ,   5   ,   345 ,   471 ,   599 ,   3677    ,   1801    ,0 };
            const unsigned long dim659JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   5   ,   21  ,   47  ,   27  ,   21  ,   473 ,   881 ,   1973    ,   995 ,   6513    ,0 };
            const unsigned long dim660JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   7   ,   23  ,   1   ,   19  ,   197 ,   43  ,   955 ,   1503    ,   2825    ,   7241    ,0 };
            const unsigned long dim661JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   13  ,   5   ,   51  ,   51  ,   33  ,   349 ,   835 ,   1367    ,   1913    ,   1963    ,0 };
            const unsigned long dim662JoeKuoD5Init[]    =   {   1   ,   1   ,   5   ,   5   ,   19  ,   63  ,   9   ,   145 ,   335 ,   843 ,   655 ,   1049    ,   3421    ,0 };
            const unsigned long dim663JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   1   ,   1   ,   33  ,   1   ,   9   ,   35  ,   833 ,   629 ,   3453    ,   6341    ,0 };
            const unsigned long dim664JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   11  ,   3   ,   55  ,   119 ,   87  ,   441 ,   43  ,   169 ,   761 ,   753 ,0 };
            const unsigned long dim665JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   5   ,   17  ,   11  ,   87  ,   165 ,   421 ,   1005    ,   1227    ,   3381    ,   1005    ,0 };
            const unsigned long dim666JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   13  ,   17  ,   19  ,   63  ,   71  ,   251 ,   355 ,   1127    ,   2575    ,   1193    ,0 };
            const unsigned long dim667JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   13  ,   9   ,   23  ,   71  ,   173 ,   421 ,   179 ,   1899    ,   2507    ,   8083    ,0 };
            const unsigned long dim668JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   15  ,   1   ,   13  ,   81  ,   169 ,   85  ,   957 ,   1109    ,   2767    ,   447 ,0 };
            const unsigned long dim669JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   15  ,   19  ,   41  ,   61  ,   247 ,   325 ,   273 ,   469 ,   1859    ,   6869    ,0 };
            const unsigned long dim670JoeKuoD5Init[]    =   {   1   ,   1   ,   1   ,   9   ,   15  ,   3   ,   3   ,   119 ,   377 ,   579 ,   1155    ,   325 ,   143 ,0 };
            const unsigned long dim671JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   11  ,   31  ,   1   ,   43  ,   135 ,   375 ,   383 ,   1497    ,   2759    ,   43  ,0 };
            const unsigned long dim672JoeKuoD5Init[]    =   {   1   ,   1   ,   5   ,   5   ,   1   ,   63  ,   75  ,   179 ,   447 ,   113 ,   1037    ,   631 ,   4969    ,0 };
            const unsigned long dim673JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   13  ,   13  ,   61  ,   63  ,   229 ,   85  ,   223 ,   153 ,   3987    ,   4685    ,0 };
            const unsigned long dim674JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   13  ,   17  ,   33  ,   17  ,   247 ,   399 ,   559 ,   369 ,   1525    ,   4923    ,0 };
            const unsigned long dim675JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   13  ,   1   ,   63  ,   69  ,   167 ,   407 ,   191 ,   499 ,   1697    ,   3267    ,0 };
            const unsigned long dim676JoeKuoD5Init[]    =   {   1   ,   1   ,   5   ,   9   ,   31  ,   15  ,   97  ,   183 ,   165 ,   271 ,   1465    ,   931 ,   4061    ,0 };
            const unsigned long dim677JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   1   ,   31  ,   7   ,   121 ,   37  ,   429 ,   375 ,   1539    ,   1383    ,   7317    ,0 };
            const unsigned long dim678JoeKuoD5Init[]    =   {   1   ,   1   ,   5   ,   9   ,   13  ,   41  ,   49  ,   205 ,   233 ,   847 ,   187 ,   359 ,   2341    ,0 };
            const unsigned long dim679JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   5   ,   25  ,   5   ,   57  ,   119 ,   79  ,   131 ,   1707    ,   1601    ,   1657    ,0 };
            const unsigned long dim680JoeKuoD5Init[]    =   {   1   ,   1   ,   5   ,   9   ,   21  ,   11  ,   89  ,   149 ,   369 ,   401 ,   623 ,   3001    ,   85  ,0 };
            const unsigned long dim681JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   15  ,   31  ,   9   ,   109 ,   33  ,   31  ,   93  ,   2035    ,   3785    ,   5893    ,0 };
            const unsigned long dim682JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   3   ,   31  ,   57  ,   15  ,   161 ,   333 ,   559 ,   1487    ,   1037    ,   1055    ,0 };
            const unsigned long dim683JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   13  ,   7   ,   9   ,   7   ,   109 ,   151 ,   629 ,   295 ,   105 ,   1121    ,0 };
            const unsigned long dim684JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   3   ,   7   ,   25  ,   49  ,   5   ,   95  ,   321 ,   303 ,   2571    ,   6967    ,0 };
            const unsigned long dim685JoeKuoD5Init[]    =   {   1   ,   1   ,   5   ,   3   ,   27  ,   13  ,   65  ,   27  ,   507 ,   847 ,   155 ,   3183    ,   5985    ,0 };
            const unsigned long dim686JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   9   ,   17  ,   29  ,   69  ,   23  ,   181 ,   855 ,   1659    ,   889 ,   1273    ,0 };
            const unsigned long dim687JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   5   ,   19  ,   41  ,   109 ,   97  ,   373 ,   591 ,   1715    ,   3927    ,   3101    ,0 };
            const unsigned long dim688JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   3   ,   9   ,   9   ,   75  ,   63  ,   1   ,   39  ,   941 ,   3677    ,   7253    ,0 };
            const unsigned long dim689JoeKuoD5Init[]    =   {   1   ,   1   ,   1   ,   11  ,   27  ,   37  ,   25  ,   33  ,   213 ,   917 ,   2037    ,   145 ,   4395    ,0 };
            const unsigned long dim690JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   11  ,   23  ,   21  ,   15  ,   195 ,   87  ,   699 ,   1475    ,   429 ,   1641    ,0 };
            const unsigned long dim691JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   1   ,   1   ,   51  ,   117 ,   13  ,   369 ,   711 ,   625 ,   3171    ,   1867    ,0 };
            const unsigned long dim692JoeKuoD5Init[]    =   {   1   ,   1   ,   1   ,   15  ,   23  ,   11  ,   47  ,   177 ,   511 ,   709 ,   255 ,   2355    ,   71  ,0 };
            const unsigned long dim693JoeKuoD5Init[]    =   {   1   ,   1   ,   1   ,   9   ,   7   ,   21  ,   21  ,   35  ,   303 ,   105 ,   525 ,   589 ,   5449    ,0 };
            const unsigned long dim694JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   7   ,   21  ,   27  ,   71  ,   177 ,   205 ,   143 ,   385 ,   2949    ,   8169    ,0 };
            const unsigned long dim695JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   9   ,   19  ,   39  ,   79  ,   241 ,   103 ,   489 ,   1399    ,   2781    ,   5533    ,0 };
            const unsigned long dim696JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   5   ,   23  ,   9   ,   121 ,   51  ,   145 ,   927 ,   1457    ,   1891    ,   7475    ,0 };
            const unsigned long dim697JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   1   ,   7   ,   23  ,   67  ,   145 ,   161 ,   63  ,   569 ,   395 ,   6719    ,0 };
            const unsigned long dim698JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   11  ,   5   ,   31  ,   13  ,   9   ,   135 ,   367 ,   1591    ,   1963    ,   7249    ,0 };
            const unsigned long dim699JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   13  ,   7   ,   47  ,   119 ,   195 ,   439 ,   619 ,   551 ,   1955    ,   6913    ,0 };
            const unsigned long dim700JoeKuoD5Init[]    =   {   1   ,   3   ,   1   ,   1   ,   29  ,   57  ,   43  ,   213 ,   113 ,   339 ,   1943    ,   183 ,   3063    ,0 };
            const unsigned long dim701JoeKuoD5Init[]    =   {   1   ,   3   ,   1   ,   11  ,   1   ,   3   ,   57  ,   149 ,   495 ,   547 ,   603 ,   907 ,   1349    ,0 };
            const unsigned long dim702JoeKuoD5Init[]    =   {   1   ,   3   ,   1   ,   7   ,   7   ,   31  ,   87  ,   49  ,   239 ,   827 ,   1451    ,   3171    ,   287 ,0 };
            const unsigned long dim703JoeKuoD5Init[]    =   {   1   ,   3   ,   1   ,   15  ,   11  ,   41  ,   29  ,   233 ,   101 ,   573 ,   737 ,   3813    ,   7739    ,0 };
            const unsigned long dim704JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   9   ,   19  ,   13  ,   57  ,   177 ,   259 ,   595 ,   263 ,   2851    ,   5771    ,0 };
            const unsigned long dim705JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   13  ,   13  ,   5   ,   27  ,   133 ,   17  ,   725 ,   417 ,   2277    ,   5649    ,0 };
            const unsigned long dim706JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   13  ,   25  ,   57  ,   35  ,   229 ,   253 ,   761 ,   275 ,   2519    ,   7061    ,0 };
            const unsigned long dim707JoeKuoD5Init[]    =   {   1   ,   3   ,   1   ,   13  ,   9   ,   31  ,   61  ,   181 ,   129 ,   259 ,   1569    ,   817 ,   4665    ,0 };
            const unsigned long dim708JoeKuoD5Init[]    =   {   1   ,   1   ,   1   ,   5   ,   23  ,   3   ,   69  ,   161 ,   429 ,   143 ,   455 ,   905 ,   3337    ,0 };
            const unsigned long dim709JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   3   ,   21  ,   33  ,   81  ,   187 ,   481 ,   481 ,   1947    ,   3563    ,   4797    ,0 };
            const unsigned long dim710JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   7   ,   15  ,   35  ,   23  ,   187 ,   455 ,   535 ,   85  ,   1067    ,   1793    ,0 };
            const unsigned long dim711JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   3   ,   11  ,   17  ,   89  ,   255 ,   443 ,   791 ,   1617    ,   2979    ,   1357    ,0 };
            const unsigned long dim712JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   15  ,   17  ,   49  ,   1   ,   233 ,   3   ,   587 ,   1203    ,   3185    ,   1173    ,0 };
            const unsigned long dim713JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   9   ,   5   ,   53  ,   37  ,   13  ,   47  ,   1011    ,   1589    ,   1073    ,   5445    ,0 };
            const unsigned long dim714JoeKuoD5Init[]    =   {   1   ,   1   ,   5   ,   1   ,   15  ,   43  ,   85  ,   167 ,   347 ,   935 ,   1681    ,   261 ,   3623    ,0 };
            const unsigned long dim715JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   3   ,   3   ,   31  ,   121 ,   241 ,   129 ,   555 ,   1737    ,   3557    ,   6515    ,0 };
            const unsigned long dim716JoeKuoD5Init[]    =   {   1   ,   3   ,   1   ,   15  ,   25  ,   7   ,   125 ,   105 ,   13  ,   927 ,   1929    ,   3869    ,   7429    ,0 };
            const unsigned long dim717JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   1   ,   31  ,   47  ,   3   ,   29  ,   149 ,   827 ,   771 ,   2113    ,   1607    ,0 };
            const unsigned long dim718JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   1   ,   17  ,   17  ,   65  ,   247 ,   109 ,   613 ,   1975    ,   2393    ,   6057    ,0 };
            const unsigned long dim719JoeKuoD5Init[]    =   {   1   ,   3   ,   1   ,   7   ,   9   ,   43  ,   41  ,   119 ,   335 ,   533 ,   1053    ,   1343    ,   6529    ,0 };
            const unsigned long dim720JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   1   ,   31  ,   59  ,   85  ,   247 ,   485 ,   811 ,   749 ,   89  ,   6677    ,0 };
            const unsigned long dim721JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   13  ,   7   ,   47  ,   77  ,   193 ,   25  ,   27  ,   639 ,   109 ,   6455    ,0 };
            const unsigned long dim722JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   1   ,   15  ,   63  ,   17  ,   81  ,   425 ,   403 ,   1537    ,   3205    ,   6237    ,0 };
            const unsigned long dim723JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   9   ,   21  ,   29  ,   105 ,   135 ,   407 ,   119 ,   793 ,   559 ,   973 ,0 };
            const unsigned long dim724JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   15  ,   1   ,   31  ,   45  ,   97  ,   511 ,   463 ,   1923    ,   2487    ,   1311    ,0 };
            const unsigned long dim725JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   11  ,   13  ,   53  ,   91  ,   149 ,   441 ,   841 ,   1121    ,   3305    ,   975 ,0 };
            const unsigned long dim726JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   15  ,   5   ,   55  ,   81  ,   181 ,   333 ,   63  ,   1157    ,   2831    ,   6231    ,0 };
            const unsigned long dim727JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   15  ,   29  ,   21  ,   69  ,   87  ,   41  ,   623 ,   475 ,   723 ,   5693    ,0 };
            const unsigned long dim728JoeKuoD5Init[]    =   {   1   ,   3   ,   1   ,   5   ,   23  ,   19  ,   15  ,   25  ,   269 ,   79  ,   1699    ,   691 ,   525 ,0 };
            const unsigned long dim729JoeKuoD5Init[]    =   {   1   ,   3   ,   1   ,   1   ,   29  ,   23  ,   85  ,   5   ,   431 ,   649 ,   1187    ,   3503    ,   105 ,0 };
            const unsigned long dim730JoeKuoD5Init[]    =   {   1   ,   1   ,   1   ,   1   ,   27  ,   59  ,   1   ,   81  ,   351 ,   383 ,   1167    ,   3747    ,   5935    ,0 };
            const unsigned long dim731JoeKuoD5Init[]    =   {   1   ,   1   ,   1   ,   13  ,   21  ,   49  ,   93  ,   5   ,   471 ,   231 ,   1469    ,   1089    ,   5059    ,0 };
            const unsigned long dim732JoeKuoD5Init[]    =   {   1   ,   3   ,   1   ,   11  ,   31  ,   61  ,   117 ,   161 ,   307 ,   621 ,   1713    ,   1325    ,   283 ,0 };
            const unsigned long dim733JoeKuoD5Init[]    =   {   1   ,   3   ,   1   ,   7   ,   21  ,   35  ,   7   ,   79  ,   157 ,   723 ,   57  ,   25  ,   2789    ,0 };
            const unsigned long dim734JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   13  ,   9   ,   47  ,   97  ,   1   ,   257 ,   941 ,   553 ,   3811    ,   3775    ,0 };
            const unsigned long dim735JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   13  ,   19  ,   5   ,   73  ,   233 ,   289 ,   241 ,   175 ,   2831    ,   4613    ,0 };
            const unsigned long dim736JoeKuoD5Init[]    =   {   1   ,   1   ,   1   ,   9   ,   11  ,   61  ,   1   ,   245 ,   223 ,   641 ,   77  ,   1811    ,   3459    ,0 };
            const unsigned long dim737JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   1   ,   21  ,   41  ,   9   ,   175 ,   377 ,   331 ,   1615    ,   325 ,   5413    ,0 };
            const unsigned long dim738JoeKuoD5Init[]    =   {   1   ,   1   ,   1   ,   3   ,   23  ,   57  ,   99  ,   161 ,   127 ,   47  ,   1923    ,   165 ,   2123    ,0 };
            const unsigned long dim739JoeKuoD5Init[]    =   {   1   ,   1   ,   5   ,   15  ,   11  ,   15  ,   77  ,   15  ,   323 ,   919 ,   1001    ,   377 ,   6095    ,0 };
            const unsigned long dim740JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   7   ,   25  ,   47  ,   5   ,   25  ,   197 ,   811 ,   179 ,   215 ,   5393    ,0 };
            const unsigned long dim741JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   11  ,   29  ,   37  ,   7   ,   125 ,   3   ,   7   ,   1881    ,   3823    ,   2117    ,0 };
            const unsigned long dim742JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   3   ,   9   ,   61  ,   97  ,   233 ,   233 ,   603 ,   511 ,   17  ,   2081    ,0 };
            const unsigned long dim743JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   9   ,   15  ,   35  ,   73  ,   199 ,   113 ,   575 ,   537 ,   883 ,   6897    ,0 };
            const unsigned long dim744JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   9   ,   17  ,   63  ,   7   ,   181 ,   261 ,   719 ,   591 ,   2575    ,   1065    ,0 };
            const unsigned long dim745JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   1   ,   13  ,   15  ,   1   ,   207 ,   385 ,   277 ,   547 ,   1069    ,   6421    ,0 };
            const unsigned long dim746JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   3   ,   5   ,   3   ,   43  ,   213 ,   509 ,   969 ,   1799    ,   3519    ,   7759    ,0 };
            const unsigned long dim747JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   15  ,   29  ,   51  ,   87  ,   63  ,   257 ,   1001    ,   741 ,   1747    ,   975 ,0 };
            const unsigned long dim748JoeKuoD5Init[]    =   {   1   ,   3   ,   1   ,   15  ,   5   ,   27  ,   7   ,   27  ,   493 ,   817 ,   319 ,   1435    ,   3243    ,0 };
            const unsigned long dim749JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   15  ,   5   ,   9   ,   113 ,   153 ,   397 ,   63  ,   2037    ,   3319    ,   6355    ,0 };
            const unsigned long dim750JoeKuoD5Init[]    =   {   1   ,   1   ,   1   ,   1   ,   29  ,   9   ,   97  ,   171 ,   113 ,   729 ,   1939    ,   2741    ,   4699    ,0 };
            const unsigned long dim751JoeKuoD5Init[]    =   {   1   ,   1   ,   1   ,   11  ,   29  ,   3   ,   49  ,   183 ,   69  ,   313 ,   153 ,   2757    ,   3353    ,0 };
            const unsigned long dim752JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   7   ,   5   ,   39  ,   65  ,   65  ,   405 ,   301 ,   849 ,   1211    ,   3627    ,0 };
            const unsigned long dim753JoeKuoD5Init[]    =   {   1   ,   1   ,   5   ,   13  ,   19  ,   11  ,   81  ,   181 ,   471 ,   669 ,   139 ,   4019    ,   4057    ,0 };
            const unsigned long dim754JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   13  ,   3   ,   31  ,   125 ,   229 ,   181 ,   913 ,   2035    ,   2081    ,   6573    ,0 };
            const unsigned long dim755JoeKuoD5Init[]    =   {   1   ,   1   ,   1   ,   5   ,   13  ,   25  ,   37  ,   209 ,   255 ,   761 ,   1485    ,   2833    ,   6617    ,0 };
            const unsigned long dim756JoeKuoD5Init[]    =   {   1   ,   1   ,   5   ,   7   ,   29  ,   17  ,   55  ,   59  ,   337 ,   953 ,   775 ,   3865    ,   5671    ,0 };
            const unsigned long dim757JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   13  ,   15  ,   37  ,   109 ,   93  ,   303 ,   57  ,   1727    ,   615 ,   3337    ,0 };
            const unsigned long dim758JoeKuoD5Init[]    =   {   1   ,   1   ,   1   ,   13  ,   31  ,   57  ,   119 ,   57  ,   295 ,   707 ,   1409    ,   3769    ,   6359    ,0 };
            const unsigned long dim759JoeKuoD5Init[]    =   {   1   ,   3   ,   1   ,   5   ,   31  ,   25  ,   35  ,   191 ,   27  ,   463 ,   875 ,   129 ,   3829    ,0 };
            const unsigned long dim760JoeKuoD5Init[]    =   {   1   ,   1   ,   5   ,   13  ,   19  ,   29  ,   95  ,   227 ,   487 ,   519 ,   289 ,   965 ,   2121    ,0 };
            const unsigned long dim761JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   5   ,   21  ,   37  ,   7   ,   43  ,   213 ,   673 ,   25  ,   1911    ,   7229    ,0 };
            const unsigned long dim762JoeKuoD5Init[]    =   {   1   ,   1   ,   1   ,   5   ,   9   ,   55  ,   115 ,   155 ,   119 ,   49  ,   653 ,   3425    ,   299 ,0 };
            const unsigned long dim763JoeKuoD5Init[]    =   {   1   ,   1   ,   1   ,   11  ,   7   ,   55  ,   65  ,   21  ,   93  ,   295 ,   1097    ,   145 ,   3401    ,0 };
            const unsigned long dim764JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   3   ,   1   ,   19  ,   109 ,   247 ,   499 ,   391 ,   309 ,   59  ,   695 ,0 };
            const unsigned long dim765JoeKuoD5Init[]    =   {   1   ,   1   ,   5   ,   5   ,   9   ,   37  ,   113 ,   233 ,   57  ,   955 ,   437 ,   775 ,   2673    ,0 };
            const unsigned long dim766JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   3   ,   5   ,   27  ,   57  ,   53  ,   149 ,   929 ,   41  ,   1473    ,   4751    ,0 };
            const unsigned long dim767JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   11  ,   5   ,   3   ,   107 ,   57  ,   3   ,   727 ,   2001    ,   3463    ,   4753    ,0 };
            const unsigned long dim768JoeKuoD5Init[]    =   {   1   ,   1   ,   1   ,   11  ,   3   ,   37  ,   111 ,   97  ,   225 ,   809 ,   135 ,   2049    ,   2373    ,0 };
            const unsigned long dim769JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   15  ,   3   ,   37  ,   57  ,   153 ,   175 ,   851 ,   137 ,   495 ,   4423    ,0 };
            const unsigned long dim770JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   7   ,   25  ,   31  ,   53  ,   51  ,   421 ,   33  ,   1241    ,   2157    ,   867 ,0 };
            const unsigned long dim771JoeKuoD5Init[]    =   {   1   ,   1   ,   1   ,   13  ,   3   ,   33  ,   15  ,   77  ,   73  ,   51  ,   1131    ,   3387    ,   7935    ,0 };
            const unsigned long dim772JoeKuoD5Init[]    =   {   1   ,   1   ,   5   ,   9   ,   11  ,   25  ,   27  ,   235 ,   341 ,   133 ,   261 ,   2087    ,   4079    ,0 };
            const unsigned long dim773JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   13  ,   11  ,   63  ,   33  ,   99  ,   65  ,   485 ,   561 ,   2269    ,   8143    ,0 };
            const unsigned long dim774JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   5   ,   13  ,   25  ,   53  ,   65  ,   59  ,   861 ,   1705    ,   627 ,   1097    ,0 };
            const unsigned long dim775JoeKuoD5Init[]    =   {   1   ,   3   ,   1   ,   3   ,   25  ,   15  ,   67  ,   187 ,   161 ,   747 ,   947 ,   4029    ,   7547    ,0 };
            const unsigned long dim776JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   5   ,   7   ,   27  ,   119 ,   239 ,   197 ,   165 ,   457 ,   1927    ,   6195    ,0 };
            const unsigned long dim777JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   7   ,   27  ,   21  ,   23  ,   93  ,   319 ,   551 ,   1317    ,   69  ,   6735    ,0 };
            const unsigned long dim778JoeKuoD5Init[]    =   {   1   ,   1   ,   5   ,   11  ,   25  ,   45  ,   23  ,   127 ,   327 ,   949 ,   1451    ,   2245    ,   6705    ,0 };
            const unsigned long dim779JoeKuoD5Init[]    =   {   1   ,   1   ,   1   ,   15  ,   19  ,   23  ,   69  ,   113 ,   53  ,   215 ,   1505    ,   1255    ,   5063    ,0 };
            const unsigned long dim780JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   9   ,   7   ,   19  ,   39  ,   135 ,   437 ,   251 ,   1947    ,   2219    ,   4015    ,0 };
            const unsigned long dim781JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   3   ,   23  ,   39  ,   91  ,   111 ,   395 ,   545 ,   179 ,   937 ,   2531    ,0 };
            const unsigned long dim782JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   11  ,   1   ,   53  ,   113 ,   137 ,   131 ,   461 ,   151 ,   1617    ,   6399    ,0 };
            const unsigned long dim783JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   3   ,   13  ,   3   ,   75  ,   31  ,   175 ,   773 ,   1293    ,   625 ,   6563    ,0 };
            const unsigned long dim784JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   15  ,   9   ,   51  ,   87  ,   223 ,   405 ,   751 ,   1053    ,   1431    ,   4701    ,0 };
            const unsigned long dim785JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   13  ,   21  ,   21  ,   111 ,   145 ,   311 ,   733 ,   635 ,   1369    ,   297 ,0 };
            const unsigned long dim786JoeKuoD5Init[]    =   {   1   ,   1   ,   1   ,   7   ,   25  ,   5   ,   23  ,   201 ,   179 ,   593 ,   1531    ,   1197    ,   3525    ,0 };
            const unsigned long dim787JoeKuoD5Init[]    =   {   1   ,   1   ,   5   ,   13  ,   7   ,   17  ,   19  ,   247 ,   401 ,   15  ,   93  ,   245 ,   5987    ,0 };
            const unsigned long dim788JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   11  ,   7   ,   45  ,   119 ,   99  ,   197 ,   163 ,   637 ,   3143    ,   1775    ,0 };
            const unsigned long dim789JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   3   ,   17  ,   7   ,   35  ,   121 ,   155 ,   925 ,   1001    ,   2941    ,   107 ,0 };
            const unsigned long dim790JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   15  ,   15  ,   3   ,   89  ,   205 ,   171 ,   279 ,   763 ,   2343    ,   7825    ,0 };
            const unsigned long dim791JoeKuoD5Init[]    =   {   1   ,   3   ,   1   ,   15  ,   15  ,   3   ,   113 ,   135 ,   159 ,   217 ,   139 ,   167 ,   589 ,0 };
            const unsigned long dim792JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   1   ,   31  ,   59  ,   89  ,   3   ,   261 ,   953 ,   1527    ,   447 ,   1211    ,0 };
            const unsigned long dim793JoeKuoD5Init[]    =   {   1   ,   1   ,   5   ,   9   ,   31  ,   39  ,   93  ,   171 ,   497 ,   657 ,   809 ,   2905    ,   2399    ,0 };
            const unsigned long dim794JoeKuoD5Init[]    =   {   1   ,   1   ,   5   ,   15  ,   27  ,   19  ,   25  ,   179 ,   293 ,   829 ,   1197    ,   3077    ,   6631    ,0 };
            const unsigned long dim795JoeKuoD5Init[]    =   {   1   ,   1   ,   1   ,   7   ,   31  ,   55  ,   53  ,   177 ,   157 ,   619 ,   197 ,   3675    ,   4691    ,0 };
            const unsigned long dim796JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   5   ,   7   ,   33  ,   85  ,   59  ,   435 ,   653 ,   1363    ,   731 ,   3923    ,0 };
            const unsigned long dim797JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   3   ,   15  ,   35  ,   15  ,   217 ,   333 ,   717 ,   143 ,   4071    ,   4769    ,0 };
            const unsigned long dim798JoeKuoD5Init[]    =   {   1   ,   1   ,   5   ,   7   ,   23  ,   31  ,   65  ,   77  ,   261 ,   389 ,   1733    ,   4077    ,   3387    ,0 };
            const unsigned long dim799JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   9   ,   21  ,   45  ,   21  ,   103 ,   163 ,   237 ,   1447    ,   141 ,   3995    ,0 };
            const unsigned long dim800JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   3   ,   27  ,   35  ,   61  ,   149 ,   359 ,   131 ,   1309    ,   3283    ,   5905    ,0 };
            const unsigned long dim801JoeKuoD5Init[]    =   {   1   ,   3   ,   1   ,   15  ,   9   ,   15  ,   99  ,   43  ,   115 ,   871 ,   419 ,   2787    ,   7741    ,0 };
            const unsigned long dim802JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   5   ,   1   ,   45  ,   31  ,   87  ,   165 ,   103 ,   2041    ,   1529    ,   5721    ,0 };
            const unsigned long dim803JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   15  ,   3   ,   11  ,   107 ,   9   ,   367 ,   401 ,   1571    ,   2657    ,   4745    ,0 };
            const unsigned long dim804JoeKuoD5Init[]    =   {   1   ,   1   ,   5   ,   13  ,   31  ,   51  ,   35  ,   53  ,   131 ,   441 ,   1421    ,   2367    ,   7713    ,0 };
            const unsigned long dim805JoeKuoD5Init[]    =   {   1   ,   3   ,   1   ,   5   ,   21  ,   49  ,   109 ,   117 ,   263 ,   827 ,   1975    ,   2639    ,   1249    ,0 };
            const unsigned long dim806JoeKuoD5Init[]    =   {   1   ,   1   ,   5   ,   13  ,   7   ,   3   ,   87  ,   185 ,   493 ,   721 ,   1363    ,   2201    ,   1067    ,0 };
            const unsigned long dim807JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   5   ,   13  ,   41  ,   113 ,   105 ,   111 ,   65  ,   705 ,   4079    ,   2461    ,0 };
            const unsigned long dim808JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   13  ,   9   ,   1   ,   75  ,   103 ,   181 ,   587 ,   1531    ,   461 ,   6551    ,0 };
            const unsigned long dim809JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   11  ,   13  ,   43  ,   19  ,   131 ,   233 ,   209 ,   175 ,   625 ,   985 ,0 };
            const unsigned long dim810JoeKuoD5Init[]    =   {   1   ,   1   ,   1   ,   11  ,   23  ,   1   ,   87  ,   233 ,   355 ,   765 ,   869 ,   2569    ,   5919    ,0 };
            const unsigned long dim811JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   5   ,   13  ,   27  ,   41  ,   127 ,   105 ,   299 ,   189 ,   3801    ,   4677    ,0 };
            const unsigned long dim812JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   13  ,   21  ,   49  ,   111 ,   95  ,   27  ,   433 ,   1715    ,   1167    ,   4943    ,0 };
            const unsigned long dim813JoeKuoD5Init[]    =   {   1   ,   3   ,   1   ,   11  ,   5   ,   17  ,   93  ,   91  ,   79  ,   355 ,   111 ,   1159    ,   4629    ,0 };
            const unsigned long dim814JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   1   ,   29  ,   43  ,   107 ,   91  ,   111 ,   813 ,   537 ,   97  ,   5337    ,0 };
            const unsigned long dim815JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   5   ,   27  ,   1   ,   69  ,   181 ,   247 ,   51  ,   409 ,   1965    ,   4709    ,0 };
            const unsigned long dim816JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   1   ,   7   ,   53  ,   119 ,   109 ,   331 ,   189 ,   761 ,   385 ,   5227    ,0 };
            const unsigned long dim817JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   15  ,   5   ,   61  ,   57  ,   3   ,   157 ,   737 ,   1605    ,   3701    ,   1069    ,0 };
            const unsigned long dim818JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   1   ,   17  ,   27  ,   89  ,   235 ,   53  ,   521 ,   1975    ,   1383    ,   467 ,0 };
            const unsigned long dim819JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   7   ,   25  ,   25  ,   39  ,   149 ,   157 ,   385 ,   1651    ,   105 ,   1487    ,0 };
            const unsigned long dim820JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   7   ,   15  ,   43  ,   37  ,   167 ,   303 ,   655 ,   331 ,   1595    ,   5405    ,0 };
            const unsigned long dim821JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   9   ,   13  ,   29  ,   13  ,   163 ,   405 ,   903 ,   1277    ,   985 ,   1479    ,0 };
            const unsigned long dim822JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   13  ,   1   ,   7   ,   61  ,   9   ,   163 ,   67  ,   209 ,   1923    ,   6587    ,0 };
            const unsigned long dim823JoeKuoD5Init[]    =   {   1   ,   1   ,   5   ,   15  ,   13  ,   3   ,   99  ,   177 ,   507 ,   1017    ,   1565    ,   757 ,   5829    ,0 };
            const unsigned long dim824JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   11  ,   7   ,   49  ,   51  ,   71  ,   177 ,   1015    ,   1321    ,   187 ,   875 ,0 };
            const unsigned long dim825JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   1   ,   29  ,   49  ,   111 ,   143 ,   99  ,   297 ,   659 ,   3147    ,   7531    ,0 };
            const unsigned long dim826JoeKuoD5Init[]    =   {   1   ,   1   ,   5   ,   3   ,   15  ,   21  ,   11  ,   229 ,   249 ,   185 ,   147 ,   173 ,   7895    ,0 };
            const unsigned long dim827JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   1   ,   17  ,   23  ,   123 ,   157 ,   373 ,   501 ,   411 ,   2487    ,   4873    ,0 };
            const unsigned long dim828JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   11  ,   21  ,   61  ,   83  ,   65  ,   345 ,   105 ,   1533    ,   981 ,   1635    ,0 };
            const unsigned long dim829JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   1   ,   15  ,   15  ,   19  ,   119 ,   59  ,   1003    ,   1595    ,   2871    ,   627 ,0 };
            const unsigned long dim830JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   3   ,   7   ,   47  ,   43  ,   3   ,   359 ,   175 ,   1149    ,   213 ,   79  ,0 };
            const unsigned long dim831JoeKuoD5Init[]    =   {   1   ,   1   ,   1   ,   1   ,   5   ,   19  ,   7   ,   93  ,   335 ,   809 ,   1867    ,   1359    ,   1017    ,0 };
            const unsigned long dim832JoeKuoD5Init[]    =   {   1   ,   1   ,   1   ,   11  ,   31  ,   25  ,   75  ,   139 ,   125 ,   479 ,   101 ,   3969    ,   1173    ,0 };
            const unsigned long dim833JoeKuoD5Init[]    =   {   1   ,   3   ,   1   ,   3   ,   5   ,   17  ,   33  ,   145 ,   57  ,   31  ,   1033    ,   3975    ,   5561    ,0 };
            const unsigned long dim834JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   3   ,   31  ,   37  ,   47  ,   149 ,   341 ,   663 ,   303 ,   3395    ,   4327    ,0 };
            const unsigned long dim835JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   1   ,   31  ,   31  ,   37  ,   207 ,   189 ,   435 ,   347 ,   2791    ,   2203    ,0 };
            const unsigned long dim836JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   5   ,   3   ,   61  ,   71  ,   199 ,   207 ,   261 ,   27  ,   2281    ,   6215    ,0 };
            const unsigned long dim837JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   13  ,   19  ,   61  ,   51  ,   99  ,   279 ,   535 ,   473 ,   2233    ,   5637    ,0 };
            const unsigned long dim838JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   15  ,   7   ,   51  ,   25  ,   159 ,   63  ,   611 ,   1015    ,   3561    ,   5763    ,0 };
            const unsigned long dim839JoeKuoD5Init[]    =   {   1   ,   3   ,   1   ,   7   ,   9   ,   13  ,   101 ,   13  ,   203 ,   733 ,   1657    ,   919 ,   7857    ,0 };
            const unsigned long dim840JoeKuoD5Init[]    =   {   1   ,   3   ,   1   ,   9   ,   5   ,   7   ,   105 ,   137 ,   321 ,   817 ,   467 ,   4043    ,   603 ,0 };
            const unsigned long dim841JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   9   ,   5   ,   7   ,   105 ,   93  ,   295 ,   617 ,   713 ,   3075    ,   1839    ,0 };
            const unsigned long dim842JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   5   ,   11  ,   17  ,   89  ,   229 ,   417 ,   53  ,   693 ,   2101    ,   1293    ,0 };
            const unsigned long dim843JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   5   ,   3   ,   57  ,   95  ,   175 ,   161 ,   291 ,   219 ,   3101    ,   3387    ,0 };
            const unsigned long dim844JoeKuoD5Init[]    =   {   1   ,   3   ,   1   ,   3   ,   25  ,   53  ,   89  ,   7   ,   9   ,   863 ,   675 ,   1011    ,   1753    ,0 };
            const unsigned long dim845JoeKuoD5Init[]    =   {   1   ,   1   ,   5   ,   15  ,   13  ,   41  ,   57  ,   141 ,   109 ,   541 ,   387 ,   3805    ,   2219    ,0 };
            const unsigned long dim846JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   11  ,   19  ,   45  ,   45  ,   63  ,   283 ,   277 ,   1503    ,   3909    ,   4825    ,0 };
            const unsigned long dim847JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   13  ,   11  ,   27  ,   33  ,   245 ,   335 ,   785 ,   1651    ,   2503    ,   7247    ,0 };
            const unsigned long dim848JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   5   ,   19  ,   3   ,   29  ,   9   ,   259 ,   1023    ,   71  ,   3659    ,   3615    ,0 };
            const unsigned long dim849JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   9   ,   29  ,   41  ,   59  ,   137 ,   139 ,   193 ,   267 ,   3293    ,   4769    ,0 };
            const unsigned long dim850JoeKuoD5Init[]    =   {   1   ,   3   ,   1   ,   5   ,   21  ,   29  ,   107 ,   227 ,   107 ,   929 ,   1009    ,   2013    ,   2791    ,0 };
            const unsigned long dim851JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   7   ,   25  ,   61  ,   105 ,   159 ,   125 ,   873 ,   293 ,   1475    ,   1745    ,0 };
            const unsigned long dim852JoeKuoD5Init[]    =   {   1   ,   3   ,   1   ,   5   ,   13  ,   11  ,   63  ,   109 ,   329 ,   847 ,   521 ,   3045    ,   2673    ,0 };
            const unsigned long dim853JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   1   ,   21  ,   1   ,   15  ,   161 ,   369 ,   339 ,   417 ,   3165    ,   5047    ,0 };
            const unsigned long dim854JoeKuoD5Init[]    =   {   1   ,   1   ,   1   ,   9   ,   31  ,   61  ,   109 ,   25  ,   491 ,   969 ,   1369    ,   403 ,   835 ,0 };
            const unsigned long dim855JoeKuoD5Init[]    =   {   1   ,   1   ,   1   ,   9   ,   29  ,   29  ,   27  ,   55  ,   395 ,   979 ,   27  ,   3091    ,   2383    ,0 };
            const unsigned long dim856JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   1   ,   25  ,   17  ,   73  ,   117 ,   55  ,   381 ,   641 ,   2549    ,   7049    ,0 };
            const unsigned long dim857JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   11  ,   17  ,   5   ,   97  ,   89  ,   187 ,   973 ,   1343    ,   3777    ,   3549    ,0 };
            const unsigned long dim858JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   1   ,   9   ,   51  ,   49  ,   169 ,   135 ,   827 ,   1941    ,   3421    ,   2351    ,0 };
            const unsigned long dim859JoeKuoD5Init[]    =   {   1   ,   1   ,   1   ,   11  ,   25  ,   59  ,   9   ,   63  ,   259 ,   551 ,   983 ,   2743    ,   3439    ,0 };
            const unsigned long dim860JoeKuoD5Init[]    =   {   1   ,   3   ,   1   ,   5   ,   1   ,   55  ,   19  ,   35  ,   39  ,   629 ,   1601    ,   773 ,   3697    ,0 };
            const unsigned long dim861JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   7   ,   17  ,   59  ,   55  ,   201 ,   467 ,   945 ,   707 ,   2197    ,   6907    ,0 };
            const unsigned long dim862JoeKuoD5Init[]    =   {   1   ,   1   ,   1   ,   15  ,   11  ,   9   ,   1   ,   87  ,   299 ,   509 ,   117 ,   3249    ,   3811    ,0 };
            const unsigned long dim863JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   7   ,   23  ,   35  ,   43  ,   129 ,   357 ,   501 ,   837 ,   305 ,   7967    ,0 };
            const unsigned long dim864JoeKuoD5Init[]    =   {   1   ,   1   ,   5   ,   1   ,   23  ,   45  ,   85  ,   245 ,   157 ,   193 ,   215 ,   1021    ,   5115    ,0 };
            const unsigned long dim865JoeKuoD5Init[]    =   {   1   ,   3   ,   1   ,   9   ,   5   ,   47  ,   125 ,   27  ,   295 ,   407 ,   1601    ,   859 ,   1203    ,0 };
            const unsigned long dim866JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   7   ,   21  ,   3   ,   23  ,   141 ,   75  ,   841 ,   199 ,   2719    ,   2131    ,0 };
            const unsigned long dim867JoeKuoD5Init[]    =   {   1   ,   3   ,   1   ,   1   ,   3   ,   7   ,   35  ,   227 ,   275 ,   37  ,   523 ,   2849    ,   4363    ,0 };
            const unsigned long dim868JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   9   ,   5   ,   39  ,   95  ,   77  ,   201 ,   891 ,   339 ,   375 ,   7115    ,0 };
            const unsigned long dim869JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   3   ,   3   ,   57  ,   57  ,   49  ,   305 ,   141 ,   1149    ,   3909    ,   6981    ,0 };
            const unsigned long dim870JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   7   ,   19  ,   17  ,   53  ,   31  ,   233 ,   317 ,   1541    ,   235 ,   1831    ,0 };
            const unsigned long dim871JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   5   ,   7   ,   63  ,   125 ,   169 ,   141 ,   399 ,   277 ,   1417    ,   7989    ,0 };
            const unsigned long dim872JoeKuoD5Init[]    =   {   1   ,   1   ,   1   ,   1   ,   17  ,   61  ,   89  ,   11  ,   411 ,   505 ,   1191    ,   2651    ,   1175    ,0 };
            const unsigned long dim873JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   1   ,   11  ,   41  ,   39  ,   69  ,   279 ,   229 ,   1247    ,   1001    ,   7163    ,0 };
            const unsigned long dim874JoeKuoD5Init[]    =   {   1   ,   1   ,   1   ,   5   ,   13  ,   29  ,   115 ,   51  ,   79  ,   57  ,   315 ,   173 ,   5875    ,0 };
            const unsigned long dim875JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   5   ,   1   ,   43  ,   121 ,   5   ,   99  ,   451 ,   1121    ,   425 ,   4581    ,0 };
            const unsigned long dim876JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   5   ,   19  ,   5   ,   101 ,   33  ,   19  ,   5   ,   1325    ,   3527    ,   1733    ,0 };
            const unsigned long dim877JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   7   ,   15  ,   35  ,   91  ,   141 ,   127 ,   1005    ,   459 ,   3707    ,   6551    ,0 };
            const unsigned long dim878JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   11  ,   15  ,   27  ,   31  ,   49  ,   153 ,   337 ,   1235    ,   2063    ,   211 ,0 };
            const unsigned long dim879JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   3   ,   25  ,   45  ,   17  ,   233 ,   161 ,   559 ,   1687    ,   3833    ,   5451    ,0 };
            const unsigned long dim880JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   13  ,   23  ,   39  ,   127 ,   183 ,   379 ,   655 ,   129 ,   37  ,   2283    ,0 };
            const unsigned long dim881JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   13  ,   23  ,   33  ,   35  ,   111 ,   491 ,   343 ,   1771    ,   509 ,   937 ,0 };
            const unsigned long dim882JoeKuoD5Init[]    =   {   1   ,   1   ,   1   ,   3   ,   19  ,   59  ,   59  ,   79  ,   327 ,   911 ,   1103    ,   2695    ,   3673    ,0 };
            const unsigned long dim883JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   5   ,   11  ,   23  ,   43  ,   31  ,   455 ,   843 ,   1515    ,   3059    ,   505 ,0 };
            const unsigned long dim884JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   7   ,   15  ,   1   ,   63  ,   35  ,   327 ,   801 ,   237 ,   1137    ,   3447    ,0 };
            const unsigned long dim885JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   1   ,   7   ,   35  ,   105 ,   113 ,   281 ,   153 ,   461 ,   3165    ,   659 ,0 };
            const unsigned long dim886JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   13  ,   29  ,   5   ,   119 ,   45  ,   313 ,   713 ,   1989    ,   99  ,   883 ,0 };
            const unsigned long dim887JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   15  ,   17  ,   43  ,   115 ,   201 ,   51  ,   947 ,   1271    ,   2465    ,   5319    ,0 };
            const unsigned long dim888JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   11  ,   25  ,   39  ,   49  ,   103 ,   99  ,   469 ,   1777    ,   33  ,   7115    ,0 };
            const unsigned long dim889JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   1   ,   17  ,   43  ,   53  ,   97  ,   473 ,   231 ,   1035    ,   19  ,   995 ,0 };
            const unsigned long dim890JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   1   ,   5   ,   1   ,   89  ,   255 ,   201 ,   945 ,   2017    ,   3181    ,   3961    ,0 };
            const unsigned long dim891JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   7   ,   9   ,   9   ,   45  ,   91  ,   451 ,   671 ,   613 ,   4051    ,   1233    ,0 };
            const unsigned long dim892JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   11  ,   27  ,   49  ,   37  ,   151 ,   59  ,   489 ,   341 ,   507 ,   5839    ,0 };
            const unsigned long dim893JoeKuoD5Init[]    =   {   1   ,   1   ,   1   ,   3   ,   1   ,   41  ,   11  ,   165 ,   509 ,   615 ,   793 ,   2741    ,   7269    ,0 };
            const unsigned long dim894JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   13  ,   11  ,   33  ,   115 ,   87  ,   131 ,   653 ,   995 ,   1903    ,   4449    ,0 };
            const unsigned long dim895JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   11  ,   27  ,   41  ,   49  ,   211 ,   229 ,   797 ,   469 ,   839 ,   5047    ,0 };
            const unsigned long dim896JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   3   ,   9   ,   43  ,   57  ,   13  ,   77  ,   623 ,   245 ,   2349    ,   3611    ,0 };
            const unsigned long dim897JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   11  ,   23  ,   31  ,   101 ,   3   ,   355 ,   739 ,   1287    ,   3973    ,   6923    ,0 };
            const unsigned long dim898JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   5   ,   17  ,   61  ,   83  ,   75  ,   329 ,   849 ,   645 ,   3125    ,   8159    ,0 };
            const unsigned long dim899JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   11  ,   15  ,   31  ,   91  ,   149 ,   195 ,   585 ,   1415    ,   119 ,   6737    ,0 };
            const unsigned long dim900JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   13  ,   7   ,   3   ,   41  ,   25  ,   481 ,   175 ,   1147    ,   153 ,   6483    ,0 };
            const unsigned long dim901JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   1   ,   29  ,   59  ,   11  ,   225 ,   275 ,   299 ,   1083    ,   401 ,   6809    ,0 };
            const unsigned long dim902JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   11  ,   7   ,   19  ,   119 ,   145 ,   299 ,   273 ,   1571    ,   627 ,   6597    ,0 };
            const unsigned long dim903JoeKuoD5Init[]    =   {   1   ,   1   ,   1   ,   13  ,   17  ,   7   ,   63  ,   19  ,   141 ,   359 ,   879 ,   2741    ,   3139    ,0 };
            const unsigned long dim904JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   15  ,   9   ,   19  ,   45  ,   127 ,   511 ,   767 ,   47  ,   2389    ,   7691    ,0 };
            const unsigned long dim905JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   15  ,   11  ,   53  ,   15  ,   207 ,   31  ,   215 ,   183 ,   2745    ,   4703    ,0 };
            const unsigned long dim906JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   7   ,   13  ,   43  ,   13  ,   65  ,   165 ,   157 ,   1139    ,   2417    ,   547 ,0 };
            const unsigned long dim907JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   7   ,   29  ,   29  ,   45  ,   165 ,   401 ,   327 ,   119 ,   1449    ,   1281    ,0 };
            const unsigned long dim908JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   11  ,   27  ,   25  ,   35  ,   5   ,   447 ,   205 ,   1487    ,   4089    ,   4929    ,0 };
            const unsigned long dim909JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   13  ,   3   ,   45  ,   73  ,   19  ,   349 ,   657 ,   771 ,   1029    ,   5047    ,0 };
            const unsigned long dim910JoeKuoD5Init[]    =   {   1   ,   1   ,   5   ,   7   ,   29  ,   63  ,   23  ,   69  ,   425 ,   997 ,   9   ,   2365    ,   3279    ,0 };
            const unsigned long dim911JoeKuoD5Init[]    =   {   1   ,   1   ,   1   ,   1   ,   7   ,   31  ,   47  ,   205 ,   29  ,   851 ,   1735    ,   1641    ,   2623    ,0 };
            const unsigned long dim912JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   11  ,   25  ,   33  ,   73  ,   109 ,   183 ,   357 ,   1585    ,   3323    ,   4993    ,0 };
            const unsigned long dim913JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   15  ,   13  ,   53  ,   43  ,   61  ,   137 ,   127 ,   1769    ,   2107    ,   4025    ,0 };
            const unsigned long dim914JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   7   ,   25  ,   59  ,   93  ,   57  ,   391 ,   155 ,   1367    ,   4093    ,   625 ,0 };
            const unsigned long dim915JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   1   ,   25  ,   27  ,   53  ,   217 ,   75  ,   961 ,   371 ,   1281    ,   6553    ,0 };
            const unsigned long dim916JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   5   ,   19  ,   5   ,   89  ,   87  ,   377 ,   199 ,   1533    ,   2219    ,   2705    ,0 };
            const unsigned long dim917JoeKuoD5Init[]    =   {   1   ,   3   ,   1   ,   7   ,   15  ,   1   ,   61  ,   143 ,   137 ,   379 ,   1443    ,   355 ,   971 ,0 };
            const unsigned long dim918JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   11  ,   17  ,   57  ,   85  ,   225 ,   37  ,   345 ,   1509    ,   597 ,   5731    ,0 };
            const unsigned long dim919JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   11  ,   29  ,   55  ,   69  ,   27  ,   155 ,   481 ,   925 ,   3391    ,   5277    ,0 };
            const unsigned long dim920JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   9   ,   15  ,   19  ,   25  ,   5   ,   109 ,   83  ,   635 ,   3073    ,   3531    ,0 };
            const unsigned long dim921JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   15  ,   23  ,   5   ,   13  ,   7   ,   19  ,   625 ,   1481    ,   1827    ,   3991    ,0 };
            const unsigned long dim922JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   9   ,   7   ,   35  ,   123 ,   161 ,   303 ,   423 ,   25  ,   2467    ,   3411    ,0 };
            const unsigned long dim923JoeKuoD5Init[]    =   {   1   ,   1   ,   1   ,   1   ,   9   ,   47  ,   27  ,   159 ,   249 ,   573 ,   1987    ,   3449    ,   7639    ,0 };
            const unsigned long dim924JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   5   ,   19  ,   59  ,   7   ,   209 ,   27  ,   495 ,   1491    ,   3217    ,   1321    ,0 };
            const unsigned long dim925JoeKuoD5Init[]    =   {   1   ,   3   ,   1   ,   11  ,   17  ,   11  ,   109 ,   87  ,   315 ,   917 ,   1105    ,   215 ,   1295    ,0 };
            const unsigned long dim926JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   9   ,   25  ,   11  ,   77  ,   81  ,   453 ,   871 ,   1541    ,   141 ,   1625    ,0 };
            const unsigned long dim927JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   3   ,   1   ,   37  ,   11  ,   149 ,   45  ,   213 ,   975 ,   2557    ,   4263    ,0 };
            const unsigned long dim928JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   13  ,   17  ,   55  ,   59  ,   33  ,   39  ,   285 ,   1767    ,   3687    ,   7087    ,0 };
            const unsigned long dim929JoeKuoD5Init[]    =   {   1   ,   1   ,   5   ,   7   ,   27  ,   9   ,   103 ,   191 ,   405 ,   25  ,   595 ,   1765    ,   5695    ,0 };
            const unsigned long dim930JoeKuoD5Init[]    =   {   1   ,   3   ,   1   ,   7   ,   31  ,   27  ,   29  ,   195 ,   355 ,   199 ,   1297    ,   3195    ,   683 ,0 };
            const unsigned long dim931JoeKuoD5Init[]    =   {   1   ,   3   ,   1   ,   1   ,   21  ,   3   ,   19  ,   227 ,   325 ,   279 ,   1593    ,   613 ,   7527    ,0 };
            const unsigned long dim932JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   15  ,   27  ,   17  ,   91  ,   179 ,   385 ,   133 ,   823 ,   3731    ,   2957    ,0 };
            const unsigned long dim933JoeKuoD5Init[]    =   {   1   ,   1   ,   1   ,   7   ,   13  ,   9   ,   1   ,   137 ,   347 ,   67  ,   287 ,   1403    ,   5233    ,0 };
            const unsigned long dim934JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   7   ,   17  ,   15  ,   25  ,   95  ,   225 ,   469 ,   1585    ,   2513    ,   1489    ,0 };
            const unsigned long dim935JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   13  ,   31  ,   33  ,   81  ,   121 ,   435 ,   971 ,   877 ,   1603    ,   373 ,0 };
            const unsigned long dim936JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   9   ,   7   ,   31  ,   29  ,   27  ,   135 ,   181 ,   549 ,   1901    ,   813 ,0 };
            const unsigned long dim937JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   3   ,   3   ,   5   ,   11  ,   229 ,   255 ,   981 ,   1843    ,   2785    ,   2573    ,0 };
            const unsigned long dim938JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   13  ,   25  ,   29  ,   13  ,   87  ,   361 ,   431 ,   969 ,   1893    ,   5257    ,0 };
            const unsigned long dim939JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   3   ,   21  ,   53  ,   99  ,   243 ,   313 ,   681 ,   655 ,   2733    ,   4329    ,0 };
            const unsigned long dim940JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   7   ,   13  ,   3   ,   107 ,   61  ,   285 ,   687 ,   213 ,   551 ,   5039    ,0 };
            const unsigned long dim941JoeKuoD5Init[]    =   {   1   ,   3   ,   1   ,   3   ,   9   ,   47  ,   99  ,   239 ,   313 ,   975 ,   1403    ,   3641    ,   1951    ,0 };
            const unsigned long dim942JoeKuoD5Init[]    =   {   1   ,   1   ,   1   ,   3   ,   19  ,   43  ,   77  ,   139 ,   357 ,   973 ,   977 ,   369 ,   2775    ,0 };
            const unsigned long dim943JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   15  ,   23  ,   43  ,   109 ,   91  ,   321 ,   1   ,   1917    ,   3341    ,   1441    ,0 };
            const unsigned long dim944JoeKuoD5Init[]    =   {   1   ,   1   ,   1   ,   9   ,   7   ,   19  ,   61  ,   255 ,   25  ,   729 ,   479 ,   837 ,   13  ,0 };
            const unsigned long dim945JoeKuoD5Init[]    =   {   1   ,   1   ,   5   ,   9   ,   27  ,   29  ,   123 ,   101 ,   349 ,   443 ,   1759    ,   3   ,   4685    ,0 };
            const unsigned long dim946JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   1   ,   29  ,   47  ,   33  ,   183 ,   223 ,   927 ,   1341    ,   797 ,   8007    ,0 };
            const unsigned long dim947JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   15  ,   27  ,   33  ,   71  ,   195 ,   57  ,   897 ,   1337    ,   3455    ,   5201    ,0 };
            const unsigned long dim948JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   7   ,   31  ,   55  ,   121 ,   49  ,   343 ,   501 ,   1511    ,   113 ,   1549    ,0 };
            const unsigned long dim949JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   3   ,   9   ,   61  ,   19  ,   215 ,   323 ,   427 ,   1777    ,   685 ,   63  ,0 };
            const unsigned long dim950JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   15  ,   17  ,   23  ,   33  ,   129 ,   257 ,   527 ,   825 ,   3611    ,   2123    ,0 };
            const unsigned long dim951JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   1   ,   5   ,   45  ,   79  ,   9   ,   211 ,   493 ,   1095    ,   3031    ,   4093    ,0 };
            const unsigned long dim952JoeKuoD5Init[]    =   {   1   ,   1   ,   5   ,   9   ,   21  ,   29  ,   29  ,   247 ,   489 ,   735 ,   11  ,   1723    ,   4459    ,0 };
            const unsigned long dim953JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   5   ,   21  ,   41  ,   59  ,   65  ,   151 ,   113 ,   851 ,   1213    ,   6367    ,0 };
            const unsigned long dim954JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   5   ,   9   ,   17  ,   85  ,   207 ,   219 ,   45  ,   85  ,   2433    ,   2219    ,0 };
            const unsigned long dim955JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   5   ,   7   ,   9   ,   39  ,   201 ,   369 ,   369 ,   113 ,   2667    ,   5137    ,0 };
            const unsigned long dim956JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   3   ,   27  ,   59  ,   127 ,   189 ,   289 ,   683 ,   1285    ,   2713    ,   7037    ,0 };
            const unsigned long dim957JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   3   ,   5   ,   19  ,   51  ,   23  ,   139 ,   379 ,   651 ,   19  ,   7705    ,0 };
            const unsigned long dim958JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   11  ,   31  ,   23  ,   11  ,   37  ,   161 ,   679 ,   1581    ,   217 ,   7973    ,0 };
            const unsigned long dim959JoeKuoD5Init[]    =   {   1   ,   3   ,   1   ,   5   ,   9   ,   49  ,   23  ,   177 ,   475 ,   261 ,   403 ,   1415    ,   2299    ,0 };
            const unsigned long dim960JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   15  ,   19  ,   35  ,   117 ,   223 ,   159 ,   805 ,   1039    ,   1359    ,   6635    ,0 };
            const unsigned long dim961JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   15  ,   19  ,   1   ,   121 ,   223 ,   123 ,   161 ,   1631    ,   1161    ,   6997    ,0 };
            const unsigned long dim962JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   15  ,   5   ,   31  ,   41  ,   133 ,   121 ,   523 ,   1941    ,   2583    ,   6231    ,0 };
            const unsigned long dim963JoeKuoD5Init[]    =   {   1   ,   1   ,   1   ,   15  ,   7   ,   31  ,   95  ,   49  ,   501 ,   737 ,   363 ,   2879    ,   6561    ,0 };
            const unsigned long dim964JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   11  ,   31  ,   47  ,   61  ,   17  ,   399 ,   3   ,   605 ,   907 ,   4605    ,0 };
            const unsigned long dim965JoeKuoD5Init[]    =   {   1   ,   1   ,   5   ,   1   ,   23  ,   61  ,   47  ,   93  ,   25  ,   643 ,   881 ,   3559    ,   5251    ,0 };
            const unsigned long dim966JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   5   ,   27  ,   63  ,   1   ,   147 ,   425 ,   639 ,   1229    ,   3131    ,   5833    ,0 };
            const unsigned long dim967JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   15  ,   31  ,   31  ,   57  ,   121 ,   183 ,   39  ,   1067    ,   1915    ,   7321    ,0 };
            const unsigned long dim968JoeKuoD5Init[]    =   {   1   ,   1   ,   5   ,   9   ,   21  ,   25  ,   125 ,   29  ,   53  ,   433 ,   189 ,   3465    ,   3847    ,0 };
            const unsigned long dim969JoeKuoD5Init[]    =   {   1   ,   1   ,   1   ,   1   ,   11  ,   13  ,   99  ,   229 ,   365 ,   909 ,   87  ,   3669    ,   6609    ,0 };
            const unsigned long dim970JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   1   ,   29  ,   3   ,   43  ,   13  ,   19  ,   897 ,   1269    ,   1091    ,   3207    ,0 };
            const unsigned long dim971JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   13  ,   11  ,   33  ,   69  ,   251 ,   337 ,   235 ,   523 ,   2053    ,   3655    ,0 };
            const unsigned long dim972JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   15  ,   1   ,   11  ,   75  ,   169 ,   507 ,   391 ,   1009    ,   3165    ,   3691    ,0 };
            const unsigned long dim973JoeKuoD5Init[]    =   {   1   ,   3   ,   1   ,   13  ,   3   ,   39  ,   119 ,   193 ,   169 ,   661 ,   813 ,   143 ,   7825    ,0 };
            const unsigned long dim974JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   11  ,   13  ,   33  ,   91  ,   209 ,   469 ,   141 ,   391 ,   1037    ,   6591    ,0 };
            const unsigned long dim975JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   7   ,   17  ,   45  ,   39  ,   19  ,   449 ,   691 ,   187 ,   2739    ,   7671    ,0 };
            const unsigned long dim976JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   15  ,   13  ,   57  ,   7   ,   75  ,   435 ,   287 ,   1479    ,   2143    ,   6501    ,0 };
            const unsigned long dim977JoeKuoD5Init[]    =   {   1   ,   1   ,   1   ,   11  ,   11  ,   29  ,   111 ,   223 ,   505 ,   139 ,   1587    ,   3769    ,   5839    ,0 };
            const unsigned long dim978JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   7   ,   5   ,   61  ,   7   ,   209 ,   461 ,   37  ,   1771    ,   3683    ,   4283    ,0 };
            const unsigned long dim979JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   9   ,   1   ,   5   ,   123 ,   69  ,   75  ,   451 ,   963 ,   3273    ,   2785    ,0 };
            const unsigned long dim980JoeKuoD5Init[]    =   {   1   ,   1   ,   1   ,   3   ,   7   ,   17  ,   87  ,   63  ,   505 ,   863 ,   1955    ,   3253    ,   463 ,0 };
            const unsigned long dim981JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   15  ,   15  ,   55  ,   127 ,   213 ,   277 ,   829 ,   165 ,   2885    ,   6693    ,0 };
            const unsigned long dim982JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   1   ,   9   ,   35  ,   5   ,   233 ,   329 ,   827 ,   531 ,   1435    ,   899 ,0 };
            const unsigned long dim983JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   11  ,   15  ,   25  ,   19  ,   233 ,   375 ,   327 ,   241 ,   3519    ,   4511    ,0 };
            const unsigned long dim984JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   11  ,   1   ,   1   ,   117 ,   29  ,   185 ,   529 ,   873 ,   1769    ,   6857    ,0 };
            const unsigned long dim985JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   11  ,   1   ,   31  ,   125 ,   27  ,   77  ,   295 ,   43  ,   205 ,   3349    ,0 };
            const unsigned long dim986JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   13  ,   1   ,   59  ,   11  ,   195 ,   483 ,   391 ,   381 ,   1251    ,   205 ,0 };
            const unsigned long dim987JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   1   ,   5   ,   51  ,   33  ,   159 ,   143 ,   213 ,   573 ,   1329    ,   2327    ,0 };
            const unsigned long dim988JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   15  ,   25  ,   5   ,   11  ,   203 ,   217 ,   397 ,   819 ,   949 ,   3987    ,0 };
            const unsigned long dim989JoeKuoD5Init[]    =   {   1   ,   1   ,   5   ,   13  ,   17  ,   1   ,   29  ,   219 ,   161 ,   437 ,   685 ,   2743    ,   7509    ,0 };
            const unsigned long dim990JoeKuoD5Init[]    =   {   1   ,   1   ,   3   ,   13  ,   3   ,   31  ,   29  ,   51  ,   41  ,   217 ,   997 ,   2581    ,   4273    ,0 };
            const unsigned long dim991JoeKuoD5Init[]    =   {   1   ,   1   ,   7   ,   5   ,   31  ,   33  ,   45  ,   113 ,   463 ,   537 ,   237 ,   1501    ,   315 ,0 };
            const unsigned long dim992JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   5   ,   13  ,   3   ,   49  ,   155 ,   175 ,   655 ,   1995    ,   2131    ,   6105    ,0 };
            const unsigned long dim993JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   15  ,   23  ,   3   ,   17  ,   165 ,   67  ,   137 ,   337 ,   3805    ,   257 ,0 };
            const unsigned long dim994JoeKuoD5Init[]    =   {   1   ,   1   ,   5   ,   13  ,   31  ,   11  ,   39  ,   111 ,   79  ,   585 ,   1911    ,   2395    ,   6239    ,0 };
            const unsigned long dim995JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   13  ,   5   ,   13  ,   61  ,   87  ,   309 ,   571 ,   321 ,   2485    ,   807 ,0 };
            const unsigned long dim996JoeKuoD5Init[]    =   {   1   ,   3   ,   5   ,   13  ,   31  ,   21  ,   9   ,   177 ,   9   ,   395 ,   351 ,   529 ,   4977    ,0 };
            const unsigned long dim997JoeKuoD5Init[]    =   {   1   ,   3   ,   1   ,   5   ,   5   ,   41  ,   95  ,   145 ,   319 ,   339 ,   1559    ,   203 ,   2883    ,0 };
            const unsigned long dim998JoeKuoD5Init[]    =   {   1   ,   3   ,   3   ,   7   ,   9   ,   13  ,   121 ,   111 ,   107 ,   421 ,   1763    ,   2671    ,   259 ,0 };
            const unsigned long dim999JoeKuoD5Init[]    =   {   1   ,   3   ,   7   ,   13  ,   25  ,   47  ,   71  ,   249 ,   119 ,   83  ,   1651    ,   2715    ,   4819    ,0 };
            const unsigned long dim1000JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   3   ,   5   ,   59  ,   99  ,   139 ,   435 ,   653 ,   153 ,   3605    ,   753 ,0 };
            const unsigned long dim1001JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   13  ,   19  ,   13  ,   3   ,   17  ,   215 ,   1017    ,   1685    ,   3795    ,   2363    ,0 };
            const unsigned long dim1002JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   11  ,   15  ,   13  ,   97  ,   145 ,   383 ,   39  ,   667 ,   1217    ,   1473    ,0 };
            const unsigned long dim1003JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   3   ,   11  ,   25  ,   107 ,   149 ,   11  ,   835 ,   1013    ,   1587    ,   1485    ,0 };
            const unsigned long dim1004JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   7   ,   15  ,   33  ,   15  ,   251 ,   473 ,   723 ,   959 ,   3991    ,   7145    ,0 };
            const unsigned long dim1005JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   13  ,   1   ,   49  ,   73  ,   195 ,   139 ,   893 ,   1677    ,   707 ,   667 ,0 };
            const unsigned long dim1006JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   11  ,   23  ,   3   ,   79  ,   255 ,   371 ,   885 ,   469 ,   3673    ,   5477    ,0 };
            const unsigned long dim1007JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   15  ,   21  ,   1   ,   45  ,   65  ,   403 ,   129 ,   123 ,   1171    ,   8177    ,0 };
            const unsigned long dim1008JoeKuoD5Init[]   =   {   1   ,   1   ,   1   ,   3   ,   29  ,   25  ,   89  ,   231 ,   81  ,   503 ,   629 ,   1925    ,   2853    ,0 };
            const unsigned long dim1009JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   13  ,   9   ,   15  ,   107 ,   81  ,   479 ,   235 ,   1483    ,   3593    ,   2289    ,0 };
            const unsigned long dim1010JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   5   ,   9   ,   49  ,   119 ,   161 ,   233 ,   321 ,   1505    ,   3969    ,   3131    ,0 };
            const unsigned long dim1011JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   15  ,   9   ,   5   ,   91  ,   57  ,   13  ,   271 ,   999 ,   747 ,   3399    ,0 };
            const unsigned long dim1012JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   1   ,   1   ,   21  ,   1   ,   179 ,   449 ,   963 ,   33  ,   2259    ,   259 ,0 };
            const unsigned long dim1013JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   13  ,   7   ,   11  ,   81  ,   53  ,   157 ,   373 ,   767 ,   2489    ,   2275    ,0 };
            const unsigned long dim1014JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   3   ,   9   ,   55  ,   123 ,   135 ,   9   ,   499 ,   3   ,   2039    ,   2387    ,0 };
            const unsigned long dim1015JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   13  ,   7   ,   47  ,   119 ,   81  ,   351 ,   949 ,   1159    ,   859 ,   99  ,0 };
            const unsigned long dim1016JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   9   ,   1   ,   5   ,   83  ,   3   ,   387 ,   455 ,   1997    ,   1253    ,   77  ,0 };
            const unsigned long dim1017JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   13  ,   13  ,   49  ,   111 ,   133 ,   193 ,   893 ,   1549    ,   4003    ,   3461    ,0 };
            const unsigned long dim1018JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   13  ,   21  ,   3   ,   63  ,   209 ,   491 ,   447 ,   1635    ,   2297    ,   7667    ,0 };
            const unsigned long dim1019JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   1   ,   3   ,   61  ,   93  ,   115 ,   417 ,   465 ,   1075    ,   2157    ,   861 ,0 };
            const unsigned long dim1020JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   1   ,   7   ,   33  ,   7   ,   61  ,   509 ,   539 ,   1579    ,   2089    ,   5633    ,0 };
            const unsigned long dim1021JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   1   ,   9   ,   5   ,   75  ,   125 ,   345 ,   133 ,   1699    ,   3183    ,   5403    ,0 };
            const unsigned long dim1022JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   15  ,   17  ,   9   ,   115 ,   213 ,   417 ,   713 ,   989 ,   3987    ,   3043    ,0 };
            const unsigned long dim1023JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   3   ,   25  ,   9   ,   115 ,   83  ,   255 ,   695 ,   471 ,   1819    ,   2661    ,0 };
            const unsigned long dim1024JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   9   ,   31  ,   35  ,   33  ,   197 ,   335 ,   543 ,   323 ,   3241    ,   7039    ,0 };
            const unsigned long dim1025JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   7   ,   23  ,   5   ,   23  ,   193 ,   327 ,   3   ,   1425    ,   2787    ,   5659    ,0 };
            const unsigned long dim1026JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   9   ,   27  ,   25  ,   37  ,   241 ,   373 ,   411 ,   783 ,   621 ,   2129    ,0 };
            const unsigned long dim1027JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   5   ,   13  ,   19  ,   119 ,   39  ,   303 ,   383 ,   1965    ,   725 ,   1909    ,0 };
            const unsigned long dim1028JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   15  ,   25  ,   27  ,   121 ,   245 ,   165 ,   985 ,   595 ,   3325    ,   7319    ,0 };
            const unsigned long dim1029JoeKuoD5Init[]   =   {   1   ,   1   ,   1   ,   5   ,   25  ,   7   ,   109 ,   75  ,   277 ,   25  ,   715 ,   495 ,   3911    ,0 };
            const unsigned long dim1030JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   9   ,   21  ,   37  ,   13  ,   23  ,   161 ,   907 ,   1551    ,   2453    ,   5323    ,0 };
            const unsigned long dim1031JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   9   ,   17  ,   11  ,   101 ,   237 ,   219 ,   735 ,   1865    ,   209 ,   5605    ,0 };
            const unsigned long dim1032JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   15  ,   5   ,   63  ,   87  ,   173 ,   299 ,   739 ,   617 ,   1883    ,   2525    ,0 };
            const unsigned long dim1033JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   1   ,   11  ,   5   ,   25  ,   207 ,   271 ,   471 ,   921 ,   3819    ,   5627    ,0 };
            const unsigned long dim1034JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   13  ,   1   ,   7   ,   27  ,   185 ,   245 ,   629 ,   1329    ,   611 ,   7183    ,0 };
            const unsigned long dim1035JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   15  ,   9   ,   21  ,   55  ,   17  ,   157 ,   987 ,   553 ,   3823    ,   6923    ,0 };
            const unsigned long dim1036JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   11  ,   19  ,   47  ,   113 ,   149 ,   495 ,   891 ,   1885    ,   2699    ,   6019    ,0 };
            const unsigned long dim1037JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   9   ,   23  ,   53  ,   27  ,   241 ,   453 ,   103 ,   1879    ,   289 ,   5195    ,0 };
            const unsigned long dim1038JoeKuoD5Init[]   =   {   1   ,   1   ,   1   ,   3   ,   13  ,   37  ,   125 ,   83  ,   341 ,   793 ,   193 ,   297 ,   3337    ,0 };
            const unsigned long dim1039JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   7   ,   13  ,   43  ,   101 ,   121 ,   319 ,   845 ,   601 ,   3357    ,   3037    ,0 };
            const unsigned long dim1040JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   1   ,   27  ,   3   ,   53  ,   111 ,   287 ,   791 ,   2017    ,   3869    ,   5105    ,0 };
            const unsigned long dim1041JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   1   ,   19  ,   33  ,   107 ,   203 ,   135 ,   783 ,   497 ,   1007    ,   4587    ,0 };
            const unsigned long dim1042JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   13  ,   7   ,   41  ,   101 ,   59  ,   407 ,   525 ,   1941    ,   961 ,   7059    ,0 };
            const unsigned long dim1043JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   5   ,   17  ,   33  ,   9   ,   217 ,   31  ,   695 ,   1111    ,   391 ,   5617    ,0 };
            const unsigned long dim1044JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   5   ,   15  ,   13  ,   107 ,   223 ,   477 ,   91  ,   449 ,   901 ,   3075    ,0 };
            const unsigned long dim1045JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   13  ,   31  ,   47  ,   97  ,   49  ,   47  ,   301 ,   305 ,   1159    ,   6977    ,0 };
            const unsigned long dim1046JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   9   ,   29  ,   9   ,   17  ,   237 ,   461 ,   593 ,   495 ,   1099    ,   5135    ,0 };
            const unsigned long dim1047JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   7   ,   3   ,   51  ,   5   ,   113 ,   409 ,   777 ,   1323    ,   2719    ,   3647    ,0 };
            const unsigned long dim1048JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   5   ,   15  ,   45  ,   33  ,   49  ,   167 ,   933 ,   1831    ,   3195    ,   3121    ,0 };
            const unsigned long dim1049JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   15  ,   3   ,   33  ,   123 ,   19  ,   173 ,   69  ,   593 ,   3709    ,   7193    ,0 };
            const unsigned long dim1050JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   15  ,   9   ,   9   ,   63  ,   81  ,   325 ,   473 ,   1517    ,   3483    ,   7585    ,0 };
            const unsigned long dim1051JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   9   ,   31  ,   47  ,   77  ,   67  ,   55  ,   673 ,   1963    ,   111 ,   839 ,0 };
            const unsigned long dim1052JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   15  ,   23  ,   45  ,   5   ,   159 ,   225 ,   595 ,   1573    ,   1891    ,   301 ,0 };
            const unsigned long dim1053JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   1   ,   3   ,   21  ,   123 ,   29  ,   331 ,   793 ,   1885    ,   3299    ,   3433    ,0 };
            const unsigned long dim1054JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   1   ,   23  ,   51  ,   21  ,   23  ,   265 ,   919 ,   853 ,   3969    ,   2043    ,0 };
            const unsigned long dim1055JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   3   ,   25  ,   59  ,   111 ,   13  ,   217 ,   893 ,   1005    ,   3795    ,   3233    ,0 };
            const unsigned long dim1056JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   7   ,   7   ,   11  ,   69  ,   183 ,   509 ,   51  ,   727 ,   2093    ,   2615    ,0 };
            const unsigned long dim1057JoeKuoD5Init[]   =   {   1   ,   1   ,   1   ,   11  ,   3   ,   13  ,   119 ,   209 ,   365 ,   895 ,   1563    ,   427 ,   5519    ,0 };
            const unsigned long dim1058JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   5   ,   23  ,   9   ,   87  ,   29  ,   19  ,   519 ,   763 ,   3553    ,   575 ,0 };
            const unsigned long dim1059JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   1   ,   3   ,   21  ,   15  ,   237 ,   501 ,   627 ,   1557    ,   545 ,   2415    ,0 };
            const unsigned long dim1060JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   3   ,   7   ,   53  ,   83  ,   19  ,   385 ,   425 ,   1145    ,   1039    ,   6667    ,0 };
            const unsigned long dim1061JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   15  ,   31  ,   31  ,   39  ,   51  ,   233 ,   755 ,   1105    ,   925 ,   6113    ,0 };
            const unsigned long dim1062JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   13  ,   17  ,   25  ,   45  ,   135 ,   347 ,   707 ,   1035    ,   1405    ,   7105    ,0 };
            const unsigned long dim1063JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   9   ,   17  ,   25  ,   119 ,   77  ,   279 ,   467 ,   195 ,   1919    ,   4959    ,0 };
            const unsigned long dim1064JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   7   ,   31  ,   41  ,   5   ,   21  ,   349 ,   607 ,   737 ,   2033    ,   2323    ,0 };
            const unsigned long dim1065JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   9   ,   29  ,   19  ,   45  ,   223 ,   391 ,   495 ,   1905    ,   735 ,   6309    ,0 };
            const unsigned long dim1066JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   15  ,   19  ,   9   ,   93  ,   89  ,   43  ,   297 ,   653 ,   1343    ,   5897    ,0 };
            const unsigned long dim1067JoeKuoD5Init[]   =   {   1   ,   1   ,   1   ,   1   ,   21  ,   7   ,   29  ,   187 ,   115 ,   279 ,   1029    ,   2817    ,   1349    ,0 };
            const unsigned long dim1068JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   9   ,   5   ,   61  ,   35  ,   33  ,   151 ,   119 ,   1713    ,   1713    ,   1645    ,0 };
            const unsigned long dim1069JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   1   ,   23  ,   21  ,   101 ,   131 ,   355 ,   75  ,   1233    ,   1677    ,   2463    ,0 };
            const unsigned long dim1070JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   15  ,   5   ,   57  ,   79  ,   51  ,   299 ,   307 ,   1977    ,   3473    ,   6153    ,0 };
            const unsigned long dim1071JoeKuoD5Init[]   =   {   1   ,   1   ,   1   ,   15  ,   19  ,   1   ,   59  ,   69  ,   175 ,   189 ,   303 ,   43  ,   7561    ,0 };
            const unsigned long dim1072JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   1   ,   31  ,   37  ,   117 ,   9   ,   373 ,   279 ,   1187    ,   3501    ,   715 ,0 };
            const unsigned long dim1073JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   13  ,   1   ,   29  ,   79  ,   161 ,   223 ,   437 ,   577 ,   921 ,   5535    ,0 };
            const unsigned long dim1074JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   1   ,   31  ,   27  ,   93  ,   63  ,   281 ,   187 ,   1739    ,   4085    ,   5669    ,0 };
            const unsigned long dim1075JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   7   ,   11  ,   25  ,   87  ,   245 ,   339 ,   741 ,   927 ,   1279    ,   3889    ,0 };
            const unsigned long dim1076JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   7   ,   31  ,   5   ,   45  ,   205 ,   289 ,   999 ,   361 ,   3595    ,   569 ,0 };
            const unsigned long dim1077JoeKuoD5Init[]   =   {   1   ,   1   ,   1   ,   13  ,   19  ,   23  ,   103 ,   73  ,   403 ,   85  ,   1623    ,   325 ,   5369    ,0 };
            const unsigned long dim1078JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   15  ,   13  ,   23  ,   27  ,   155 ,   359 ,   777 ,   1751    ,   915 ,   949 ,0 };
            const unsigned long dim1079JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   11  ,   23  ,   59  ,   57  ,   215 ,   77  ,   581 ,   369 ,   953 ,   6987    ,0 };
            const unsigned long dim1080JoeKuoD5Init[]   =   {   1   ,   1   ,   1   ,   15  ,   27  ,   55  ,   103 ,   173 ,   485 ,   771 ,   1693    ,   1227    ,   3257    ,0 };
            const unsigned long dim1081JoeKuoD5Init[]   =   {   1   ,   1   ,   1   ,   9   ,   5   ,   23  ,   95  ,   121 ,   81  ,   107 ,   1897    ,   1647    ,   3047    ,0 };
            const unsigned long dim1082JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   11  ,   17  ,   47  ,   119 ,   83  ,   137 ,   897 ,   1893    ,   653 ,   5031    ,0 };
            const unsigned long dim1083JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   13  ,   31  ,   3   ,   73  ,   129 ,   159 ,   529 ,   1433    ,   2313    ,   6143    ,0 };
            const unsigned long dim1084JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   15  ,   29  ,   19  ,   123 ,   141 ,   51  ,   427 ,   935 ,   2831    ,   5799    ,0 };
            const unsigned long dim1085JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   1   ,   31  ,   3   ,   119 ,   227 ,   37  ,   435 ,   921 ,   3313    ,   2129    ,0 };
            const unsigned long dim1086JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   13  ,   1   ,   19  ,   75  ,   35  ,   307 ,   419 ,   813 ,   2217    ,   6603    ,0 };
            const unsigned long dim1087JoeKuoD5Init[]   =   {   1   ,   1   ,   1   ,   3   ,   17  ,   47  ,   79  ,   75  ,   47  ,   835 ,   287 ,   3361    ,   5875    ,0 };
            const unsigned long dim1088JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   9   ,   5   ,   5   ,   3   ,   19  ,   341 ,   717 ,   45  ,   1169    ,   1305    ,0 };
            const unsigned long dim1089JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   1   ,   3   ,   11  ,   81  ,   233 ,   195 ,   987 ,   593 ,   2495    ,   5213    ,0 };
            const unsigned long dim1090JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   1   ,   27  ,   1   ,   29  ,   251 ,   221 ,   267 ,   593 ,   361 ,   5629    ,0 };
            const unsigned long dim1091JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   15  ,   21  ,   33  ,   15  ,   37  ,   341 ,   301 ,   293 ,   2787    ,   3531    ,0 };
            const unsigned long dim1092JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   7   ,   3   ,   3   ,   9   ,   7   ,   257 ,   509 ,   1545    ,   4095    ,   3309    ,0 };
            const unsigned long dim1093JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   5   ,   11  ,   7   ,   27  ,   71  ,   317 ,   221 ,   391 ,   1257    ,   5885    ,0 };
            const unsigned long dim1094JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   7   ,   15  ,   51  ,   29  ,   107 ,   461 ,   597 ,   961 ,   3589    ,   2325    ,0 };
            const unsigned long dim1095JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   3   ,   29  ,   1   ,   91  ,   181 ,   477 ,   125 ,   1869    ,   3209    ,   3513    ,0 };
            const unsigned long dim1096JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   7   ,   13  ,   17  ,   49  ,   145 ,   215 ,   1003    ,   1053    ,   1413    ,   8011    ,0 };
            const unsigned long dim1097JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   5   ,   23  ,   63  ,   9   ,   175 ,   159 ,   627 ,   705 ,   2769    ,   2469    ,0 };
            const unsigned long dim1098JoeKuoD5Init[]   =   {   1   ,   1   ,   1   ,   11  ,   27  ,   21  ,   5   ,   61  ,   249 ,   581 ,   829 ,   2195    ,   4241    ,0 };
            const unsigned long dim1099JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   11  ,   27  ,   39  ,   67  ,   3   ,   23  ,   819 ,   1879    ,   3775    ,   6949    ,0 };
            const unsigned long dim1100JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   5   ,   19  ,   35  ,   93  ,   113 ,   371 ,   511 ,   811 ,   577 ,   1121    ,0 };
            const unsigned long dim1101JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   9   ,   9   ,   25  ,   103 ,   139 ,   151 ,   177 ,   557 ,   2123    ,   6677    ,0 };
            const unsigned long dim1102JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   5   ,   17  ,   63  ,   61  ,   241 ,   351 ,   371 ,   1745    ,   3133    ,   7663    ,0 };
            const unsigned long dim1103JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   11  ,   19  ,   39  ,   105 ,   93  ,   77  ,   445 ,   1433    ,   1793    ,   2957    ,0 };
            const unsigned long dim1104JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   15  ,   5   ,   15  ,   29  ,   211 ,   229 ,   887 ,   413 ,   701 ,   737 ,0 };
            const unsigned long dim1105JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   13  ,   17  ,   7   ,   69  ,   213 ,   49  ,   91  ,   1143    ,   3743    ,   4385    ,0 };
            const unsigned long dim1106JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   7   ,   21  ,   47  ,   41  ,   157 ,   299 ,   29  ,   751 ,   2427    ,   1521    ,0 };
            const unsigned long dim1107JoeKuoD5Init[]   =   {   1   ,   1   ,   1   ,   1   ,   29  ,   45  ,   119 ,   79  ,   141 ,   477 ,   1289    ,   515 ,   8143    ,0 };
            const unsigned long dim1108JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   7   ,   3   ,   7   ,   123 ,   197 ,   441 ,   233 ,   1841    ,   267 ,   6553    ,0 };
            const unsigned long dim1109JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   15  ,   3   ,   41  ,   33  ,   95  ,   271 ,   461 ,   1505    ,   2989    ,   5503    ,0 };
            const unsigned long dim1110JoeKuoD5Init[]   =   {   1   ,   1   ,   1   ,   11  ,   19  ,   15  ,   1   ,   23  ,   13  ,   737 ,   51  ,   289 ,   6731    ,0 };
            const unsigned long dim1111JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   1   ,   15  ,   11  ,   53  ,   241 ,   17  ,   107 ,   1931    ,   3759    ,   5421    ,   1889    ,0 };
            const unsigned long dim1112JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   13  ,   15  ,   29  ,   107 ,   163 ,   395 ,   645 ,   299 ,   799 ,   4331    ,   335 ,0 };
            const unsigned long dim1113JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   13  ,   5   ,   47  ,   91  ,   41  ,   439 ,   319 ,   1213    ,   763 ,   6101    ,   1543    ,0 };
            const unsigned long dim1114JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   15  ,   19  ,   51  ,   117 ,   159 ,   315 ,   767 ,   1957    ,   3655    ,   6573    ,   5419    ,0 };
            const unsigned long dim1115JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   11  ,   23  ,   51  ,   115 ,   223 ,   125 ,   633 ,   637 ,   3443    ,   1993    ,   1887    ,0 };
            const unsigned long dim1116JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   15  ,   27  ,   59  ,   49  ,   123 ,   49  ,   187 ,   963 ,   3893    ,   3921    ,   14411   ,0 };
            const unsigned long dim1117JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   7   ,   29  ,   3   ,   77  ,   3   ,   79  ,   409 ,   1151    ,   3547    ,   3693    ,   8367    ,0 };
            const unsigned long dim1118JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   9   ,   23  ,   31  ,   123 ,   133 ,   215 ,   921 ,   329 ,   1449    ,   5535    ,   9725    ,0 };
            const unsigned long dim1119JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   5   ,   11  ,   45  ,   109 ,   117 ,   493 ,   743 ,   1473    ,   2073    ,   4771    ,   16321   ,0 };
            const unsigned long dim1120JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   9   ,   27  ,   29  ,   25  ,   223 ,   371 ,   113 ,   1183    ,   1723    ,   6127    ,   9949    ,0 };
            const unsigned long dim1121JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   15  ,   27  ,   55  ,   119 ,   31  ,   21  ,   849 ,   2001    ,   2541    ,   2611    ,   15429   ,0 };
            const unsigned long dim1122JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   7   ,   17  ,   1   ,   93  ,   243 ,   311 ,   175 ,   559 ,   2177    ,   5641    ,   15293   ,0 };
            const unsigned long dim1123JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   15  ,   25  ,   31  ,   121 ,   179 ,   169 ,   61  ,   1837    ,   2233    ,   1735    ,   6597    ,0 };
            const unsigned long dim1124JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   13  ,   21  ,   59  ,   61  ,   239 ,   501 ,   523 ,   257 ,   573 ,   893 ,   7275    ,0 };
            const unsigned long dim1125JoeKuoD5Init[]   =   {   1   ,   1   ,   1   ,   13  ,   29  ,   33  ,   77  ,   225 ,   81  ,   879 ,   1403    ,   3279    ,   2225    ,   11571   ,0 };
            const unsigned long dim1126JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   5   ,   15  ,   5   ,   29  ,   7   ,   157 ,   717 ,   397 ,   2079    ,   5839    ,   13297   ,0 };
            const unsigned long dim1127JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   7   ,   17  ,   3   ,   93  ,   241 ,   301 ,   433 ,   2003    ,   2089    ,   5781    ,   15223   ,0 };
            const unsigned long dim1128JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   13  ,   5   ,   19  ,   53  ,   189 ,   41  ,   17  ,   897 ,   2327    ,   3481    ,   7185    ,0 };
            const unsigned long dim1129JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   3   ,   25  ,   23  ,   23  ,   155 ,   367 ,   391 ,   1001    ,   1179    ,   3781    ,   14225   ,0 };
            const unsigned long dim1130JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   7   ,   9   ,   23  ,   63  ,   73  ,   439 ,   361 ,   233 ,   3387    ,   887 ,   5425    ,0 };
            const unsigned long dim1131JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   13  ,   5   ,   57  ,   55  ,   35  ,   369 ,   85  ,   1585    ,   2267    ,   2927    ,   13997   ,0 };
            const unsigned long dim1132JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   1   ,   7   ,   7   ,   55  ,   109 ,   401 ,   443 ,   1777    ,   3831    ,   6933    ,   3661    ,0 };
            const unsigned long dim1133JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   15  ,   17  ,   27  ,   5   ,   17  ,   419 ,   949 ,   1483    ,   791 ,   7353    ,   1425    ,0 };
            const unsigned long dim1134JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   9   ,   27  ,   41  ,   67  ,   135 ,   129 ,   863 ,   1679    ,   4001    ,   6841    ,   13561   ,0 };
            const unsigned long dim1135JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   3   ,   21  ,   43  ,   45  ,   65  ,   103 ,   141 ,   1261    ,   2865    ,   5621    ,   5131    ,0 };
            const unsigned long dim1136JoeKuoD5Init[]   =   {   1   ,   1   ,   1   ,   15  ,   19  ,   3   ,   97  ,   159 ,   465 ,   31  ,   1757    ,   2765    ,   667 ,   6943    ,0 };
            const unsigned long dim1137JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   3   ,   3   ,   5   ,   111 ,   203 ,   313 ,   495 ,   123 ,   1899    ,   7765    ,   2737    ,0 };
            const unsigned long dim1138JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   15  ,   19  ,   63  ,   19  ,   233 ,   283 ,   25  ,   1009    ,   2117    ,   6233    ,   5059    ,0 };
            const unsigned long dim1139JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   9   ,   29  ,   11  ,   35  ,   111 ,   111 ,   49  ,   1681    ,   3483    ,   2449    ,   13877   ,0 };
            const unsigned long dim1140JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   13  ,   7   ,   61  ,   27  ,   217 ,   275 ,   137 ,   2025    ,   2745    ,   5565    ,   7999    ,0 };
            const unsigned long dim1141JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   1   ,   13  ,   19  ,   113 ,   169 ,   425 ,   691 ,   1425    ,   1645    ,   1045    ,   9237    ,0 };
            const unsigned long dim1142JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   11  ,   23  ,   19  ,   67  ,   5   ,   225 ,   523 ,   1809    ,   341 ,   7919    ,   3675    ,0 };
            const unsigned long dim1143JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   7   ,   3   ,   33  ,   25  ,   229 ,   393 ,   141 ,   1953    ,   1433    ,   1593    ,   11569   ,0 };
            const unsigned long dim1144JoeKuoD5Init[]   =   {   1   ,   1   ,   1   ,   1   ,   5   ,   23  ,   53  ,   59  ,   141 ,   385 ,   1765    ,   4079    ,   2901    ,   593 ,0 };
            const unsigned long dim1145JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   9   ,   15  ,   43  ,   115 ,   93  ,   121 ,   209 ,   1797    ,   633 ,   2595    ,   5539    ,0 };
            const unsigned long dim1146JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   15  ,   5   ,   25  ,   9   ,   141 ,   37  ,   313 ,   1937    ,   2259    ,   1051    ,   8251    ,0 };
            const unsigned long dim1147JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   5   ,   17  ,   1   ,   89  ,   173 ,   169 ,   463 ,   2003    ,   4005    ,   6009    ,   4373    ,0 };
            const unsigned long dim1148JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   7   ,   17  ,   21  ,   59  ,   207 ,   333 ,   741 ,   1847    ,   683 ,   2847    ,   11007   ,0 };
            const unsigned long dim1149JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   5   ,   5   ,   39  ,   111 ,   91  ,   49  ,   559 ,   1937    ,   1311    ,   6157    ,   517 ,0 };
            const unsigned long dim1150JoeKuoD5Init[]   =   {   1   ,   1   ,   1   ,   15  ,   29  ,   1   ,   113 ,   125 ,   343 ,   939 ,   1989    ,   2569    ,   7215    ,   5099    ,0 };
            const unsigned long dim1151JoeKuoD5Init[]   =   {   1   ,   1   ,   1   ,   5   ,   5   ,   55  ,   103 ,   237 ,   313 ,   43  ,   909 ,   201 ,   175 ,   16025   ,0 };
            const unsigned long dim1152JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   3   ,   25  ,   29  ,   33  ,   59  ,   127 ,   865 ,   1753    ,   3649    ,   5517    ,   5001    ,0 };
            const unsigned long dim1153JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   15  ,   27  ,   31  ,   97  ,   153 ,   451 ,   241 ,   59  ,   515 ,   2869    ,   12909   ,0 };
            const unsigned long dim1154JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   3   ,   13  ,   51  ,   3   ,   55  ,   105 ,   497 ,   701 ,   483 ,   5165    ,   4721    ,0 };
            const unsigned long dim1155JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   3   ,   19  ,   57  ,   39  ,   55  ,   197 ,   409 ,   199 ,   1635    ,   1965    ,   2489    ,0 };
            const unsigned long dim1156JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   9   ,   1   ,   13  ,   123 ,   125 ,   341 ,   981 ,   1957    ,   1619    ,   1973    ,   8641    ,0 };
            const unsigned long dim1157JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   1   ,   29  ,   13  ,   31  ,   71  ,   443 ,   867 ,   1755    ,   843 ,   7349    ,   6015    ,0 };
            const unsigned long dim1158JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   13  ,   15  ,   39  ,   81  ,   123 ,   9   ,   991 ,   803 ,   3281    ,   7859    ,   15455   ,0 };
            const unsigned long dim1159JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   1   ,   3   ,   33  ,   87  ,   81  ,   111 ,   595 ,   483 ,   3273    ,   847 ,   2061    ,0 };
            const unsigned long dim1160JoeKuoD5Init[]   =   {   1   ,   1   ,   1   ,   11  ,   9   ,   21  ,   79  ,   49  ,   453 ,   125 ,   603 ,   1733    ,   7213    ,   7309    ,0 };
            const unsigned long dim1161JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   3   ,   29  ,   47  ,   21  ,   99  ,   35  ,   275 ,   69  ,   3773    ,   389 ,   10615   ,0 };
            const unsigned long dim1162JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   5   ,   15  ,   21  ,   9   ,   55  ,   293 ,   639 ,   135 ,   903 ,   973 ,   9467    ,0 };
            const unsigned long dim1163JoeKuoD5Init[]   =   {   1   ,   1   ,   1   ,   15  ,   19  ,   63  ,   73  ,   71  ,   397 ,   387 ,   1859    ,   2741    ,   7323    ,   369 ,0 };
            const unsigned long dim1164JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   9   ,   21  ,   23  ,   55  ,   129 ,   183 ,   721 ,   1293    ,   3579    ,   3629    ,   13303   ,0 };
            const unsigned long dim1165JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   1   ,   21  ,   35  ,   79  ,   255 ,   443 ,   123 ,   551 ,   1113    ,   8133    ,   11621   ,0 };
            const unsigned long dim1166JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   7   ,   19  ,   61  ,   35  ,   161 ,   145 ,   291 ,   1503    ,   3085    ,   4589    ,   7971    ,0 };
            const unsigned long dim1167JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   15  ,   9   ,   59  ,   119 ,   133 ,   69  ,   413 ,   335 ,   2089    ,   8085    ,   12727   ,0 };
            const unsigned long dim1168JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   11  ,   17  ,   5   ,   83  ,   125 ,   161 ,   745 ,   1889    ,   345 ,   8107    ,   10693   ,0 };
            const unsigned long dim1169JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   9   ,   5   ,   29  ,   59  ,   69  ,   113 ,   529 ,   199 ,   1565    ,   1611    ,   12297   ,0 };
            const unsigned long dim1170JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   15  ,   27  ,   61  ,   49  ,   59  ,   249 ,   121 ,   1569    ,   407 ,   1443    ,   5705    ,0 };
            const unsigned long dim1171JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   13  ,   5   ,   3   ,   93  ,   99  ,   417 ,   499 ,   1867    ,   1269    ,   4293    ,   14633   ,0 };
            const unsigned long dim1172JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   5   ,   15  ,   37  ,   29  ,   75  ,   191 ,   41  ,   11  ,   339 ,   485 ,   13635   ,0 };
            const unsigned long dim1173JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   7   ,   27  ,   53  ,   13  ,   121 ,   209 ,   411 ,   51  ,   1003    ,   6587    ,   8247    ,0 };
            const unsigned long dim1174JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   5   ,   13  ,   7   ,   43  ,   155 ,   467 ,   491 ,   1181    ,   1105    ,   2165    ,   16347   ,0 };
            const unsigned long dim1175JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   3   ,   21  ,   19  ,   89  ,   25  ,   353 ,   223 ,   1063    ,   111 ,   611 ,   2225    ,0 };
            const unsigned long dim1176JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   9   ,   3   ,   41  ,   95  ,   183 ,   135 ,   919 ,   861 ,   2929    ,   7189    ,   5505    ,0 };
            const unsigned long dim1177JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   13  ,   13  ,   39  ,   85  ,   135 ,   403 ,   39  ,   1893    ,   3667    ,   2609    ,   9251    ,0 };
            const unsigned long dim1178JoeKuoD5Init[]   =   {   1   ,   1   ,   1   ,   11  ,   25  ,   5   ,   1   ,   179 ,   491 ,   953 ,   393 ,   2005    ,   1401    ,   12589   ,0 };
            const unsigned long dim1179JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   7   ,   25  ,   19  ,   25  ,   73  ,   63  ,   897 ,   701 ,   345 ,   4995    ,   2411    ,0 };
            const unsigned long dim1180JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   3   ,   31  ,   15  ,   121 ,   161 ,   127 ,   847 ,   1547    ,   3379    ,   4763    ,   8349    ,0 };
            const unsigned long dim1181JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   7   ,   25  ,   61  ,   87  ,   235 ,   117 ,   861 ,   977 ,   2979    ,   3333    ,   10911   ,0 };
            const unsigned long dim1182JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   9   ,   5   ,   57  ,   33  ,   129 ,   295 ,   225 ,   1497    ,   2367    ,   5379    ,   12721   ,0 };
            const unsigned long dim1183JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   11  ,   19  ,   43  ,   43  ,   217 ,   139 ,   637 ,   567 ,   3661    ,   2807    ,   11061   ,0 };
            const unsigned long dim1184JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   7   ,   27  ,   41  ,   97  ,   89  ,   53  ,   375 ,   1871    ,   2575    ,   1545    ,   455 ,0 };
            const unsigned long dim1185JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   3   ,   13  ,   15  ,   77  ,   229 ,   329 ,   663 ,   1943    ,   1411    ,   5755    ,   16279   ,0 };
            const unsigned long dim1186JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   9   ,   1   ,   3   ,   65  ,   21  ,   239 ,   619 ,   1801    ,   233 ,   3441    ,   12777   ,0 };
            const unsigned long dim1187JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   5   ,   9   ,   63  ,   9   ,   19  ,   65  ,   759 ,   1871    ,   3537    ,   4453    ,   15443   ,0 };
            const unsigned long dim1188JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   5   ,   17  ,   35  ,   43  ,   199 ,   167 ,   507 ,   669 ,   3593    ,   1645    ,   11791   ,0 };
            const unsigned long dim1189JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   7   ,   25  ,   21  ,   111 ,   171 ,   349 ,   423 ,   1793    ,   659 ,   5211    ,   3547    ,0 };
            const unsigned long dim1190JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   13  ,   23  ,   53  ,   95  ,   81  ,   383 ,   639 ,   1113    ,   2021    ,   7897    ,   4803    ,0 };
            const unsigned long dim1191JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   9   ,   19  ,   63  ,   61  ,   165 ,   3   ,   565 ,   829 ,   3071    ,   6605    ,   3625    ,0 };
            const unsigned long dim1192JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   15  ,   13  ,   53  ,   89  ,   97  ,   307 ,   555 ,   2039    ,   2753    ,   169 ,   941 ,0 };
            const unsigned long dim1193JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   3   ,   9   ,   41  ,   45  ,   89  ,   155 ,   269 ,   51  ,   3791    ,   5563    ,   4757    ,0 };
            const unsigned long dim1194JoeKuoD5Init[]   =   {   1   ,   1   ,   1   ,   7   ,   23  ,   35  ,   19  ,   93  ,   205 ,   51  ,   375 ,   2107    ,   6357    ,   16257   ,0 };
            const unsigned long dim1195JoeKuoD5Init[]   =   {   1   ,   1   ,   1   ,   5   ,   3   ,   45  ,   107 ,   141 ,   315 ,   107 ,   219 ,   51  ,   7629    ,   6865    ,0 };
            const unsigned long dim1196JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   1   ,   19  ,   23  ,   17  ,   77  ,   409 ,   59  ,   1649    ,   4029    ,   6541    ,   1075    ,0 };
            const unsigned long dim1197JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   11  ,   1   ,   37  ,   13  ,   221 ,   277 ,   61  ,   1509    ,   1713    ,   4597    ,   5907    ,0 };
            const unsigned long dim1198JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   1   ,   19  ,   53  ,   109 ,   123 ,   243 ,   1001    ,   291 ,   2265    ,   45  ,   437 ,0 };
            const unsigned long dim1199JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   1   ,   17  ,   63  ,   65  ,   67  ,   359 ,   139 ,   699 ,   3037    ,   6123    ,   11885   ,0 };
            const unsigned long dim1200JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   1   ,   3   ,   45  ,   125 ,   107 ,   75  ,   631 ,   769 ,   1431    ,   2089    ,   4919    ,0 };
            const unsigned long dim1201JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   5   ,   27  ,   31  ,   63  ,   75  ,   265 ,   727 ,   1197    ,   3549    ,   7677    ,   10831   ,0 };
            const unsigned long dim1202JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   15  ,   17  ,   61  ,   67  ,   169 ,   325 ,   627 ,   489 ,   729 ,   3585    ,   14253   ,0 };
            const unsigned long dim1203JoeKuoD5Init[]   =   {   1   ,   1   ,   1   ,   5   ,   1   ,   39  ,   23  ,   11  ,   23  ,   3   ,   973 ,   2161    ,   5613    ,   5711    ,0 };
            const unsigned long dim1204JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   11  ,   3   ,   33  ,   31  ,   139 ,   275 ,   391 ,   1625    ,   2037    ,   389 ,   3853    ,0 };
            const unsigned long dim1205JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   1   ,   17  ,   47  ,   19  ,   171 ,   219 ,   959 ,   1623    ,   1069    ,   1833    ,   4467    ,0 };
            const unsigned long dim1206JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   11  ,   3   ,   29  ,   59  ,   159 ,   385 ,   183 ,   1091    ,   491 ,   8087    ,   4047    ,0 };
            const unsigned long dim1207JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   5   ,   19  ,   3   ,   121 ,   29  ,   97  ,   345 ,   1573    ,   2993    ,   7987    ,   5397    ,0 };
            const unsigned long dim1208JoeKuoD5Init[]   =   {   1   ,   1   ,   1   ,   1   ,   31  ,   9   ,   57  ,   33  ,   489 ,   389 ,   1065    ,   2715    ,   5955    ,   11267   ,0 };
            const unsigned long dim1209JoeKuoD5Init[]   =   {   1   ,   1   ,   1   ,   1   ,   25  ,   7   ,   67  ,   245 ,   159 ,   987 ,   1297    ,   277 ,   2223    ,   9865    ,0 };
            const unsigned long dim1210JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   1   ,   5   ,   11  ,   109 ,   171 ,   181 ,   303 ,   1875    ,   1915    ,   5007    ,   15563   ,0 };
            const unsigned long dim1211JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   1   ,   21  ,   7   ,   67  ,   155 ,   503 ,   751 ,   1721    ,   3405    ,   4717    ,   12567   ,0 };
            const unsigned long dim1212JoeKuoD5Init[]   =   {   1   ,   1   ,   1   ,   1   ,   1   ,   51  ,   67  ,   57  ,   373 ,   835 ,   911 ,   3899    ,   7235    ,   5943    ,0 };
            const unsigned long dim1213JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   1   ,   21  ,   63  ,   21  ,   209 ,   321 ,   75  ,   551 ,   2741    ,   535 ,   14693   ,0 };
            const unsigned long dim1214JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   11  ,   21  ,   53  ,   23  ,   189 ,   463 ,   707 ,   637 ,   4055    ,   851 ,   12377   ,0 };
            const unsigned long dim1215JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   3   ,   13  ,   39  ,   17  ,   7   ,   487 ,   731 ,   1965    ,   2395    ,   5129    ,   6677    ,0 };
            const unsigned long dim1216JoeKuoD5Init[]   =   {   1   ,   1   ,   1   ,   9   ,   5   ,   57  ,   15  ,   247 ,   325 ,   789 ,   1771    ,   2811    ,   6453    ,   7031    ,0 };
            const unsigned long dim1217JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   13  ,   21  ,   21  ,   27  ,   113 ,   345 ,   999 ,   7   ,   3241    ,   2569    ,   9175    ,0 };
            const unsigned long dim1218JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   5   ,   19  ,   41  ,   51  ,   37  ,   409 ,   553 ,   693 ,   1877    ,   4015    ,   1749    ,0 };
            const unsigned long dim1219JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   11  ,   3   ,   7   ,   11  ,   197 ,   5   ,   601 ,   451 ,   2117    ,   4519    ,   5913    ,0 };
            const unsigned long dim1220JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   7   ,   11  ,   53  ,   81  ,   187 ,   169 ,   419 ,   175 ,   2041    ,   2537    ,   16333   ,0 };
            const unsigned long dim1221JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   1   ,   1   ,   9   ,   117 ,   33  ,   233 ,   581 ,   1575    ,   2131    ,   2065    ,   6597    ,0 };
            const unsigned long dim1222JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   3   ,   27  ,   7   ,   111 ,   51  ,   311 ,   297 ,   1773    ,   193 ,   5599    ,   271 ,0 };
            const unsigned long dim1223JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   7   ,   17  ,   49  ,   85  ,   247 ,   7   ,   1005    ,   1191    ,   399 ,   7775    ,   5635    ,0 };
            const unsigned long dim1224JoeKuoD5Init[]   =   {   1   ,   1   ,   1   ,   9   ,   31  ,   23  ,   9   ,   251 ,   97  ,   835 ,   813 ,   1335    ,   7955    ,   3977    ,0 };
            const unsigned long dim1225JoeKuoD5Init[]   =   {   1   ,   1   ,   1   ,   15  ,   3   ,   39  ,   5   ,   19  ,   111 ,   523 ,   1475    ,   2169    ,   4121    ,   1171    ,0 };
            const unsigned long dim1226JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   9   ,   3   ,   19  ,   73  ,   43  ,   45  ,   57  ,   165 ,   3659    ,   7585    ,   8549    ,0 };
            const unsigned long dim1227JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   11  ,   25  ,   57  ,   117 ,   15  ,   289 ,   953 ,   1123    ,   2327    ,   7957    ,   6043    ,0 };
            const unsigned long dim1228JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   13  ,   7   ,   45  ,   63  ,   39  ,   349 ,   871 ,   1215    ,   2915    ,   1061    ,   7633    ,0 };
            const unsigned long dim1229JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   15  ,   1   ,   33  ,   55  ,   213 ,   245 ,   221 ,   259 ,   1679    ,   1507    ,   14275   ,0 };
            const unsigned long dim1230JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   3   ,   5   ,   15  ,   59  ,   55  ,   349 ,   121 ,   1471    ,   2119    ,   2559    ,   6379    ,0 };
            const unsigned long dim1231JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   3   ,   7   ,   23  ,   125 ,   137 ,   171 ,   775 ,   1069    ,   605 ,   2945    ,   16089   ,0 };
            const unsigned long dim1232JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   7   ,   9   ,   9   ,   43  ,   131 ,   385 ,   527 ,   757 ,   1263    ,   5285    ,   16309   ,0 };
            const unsigned long dim1233JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   3   ,   3   ,   39  ,   5   ,   163 ,   459 ,   697 ,   715 ,   3827    ,   3295    ,   9163    ,0 };
            const unsigned long dim1234JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   15  ,   7   ,   29  ,   7   ,   177 ,   207 ,   465 ,   59  ,   1485    ,   7731    ,   9843    ,0 };
            const unsigned long dim1235JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   13  ,   21  ,   57  ,   87  ,   191 ,   399 ,   689 ,   935 ,   2771    ,   1025    ,   715 ,0 };
            const unsigned long dim1236JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   1   ,   29  ,   19  ,   107 ,   253 ,   155 ,   163 ,   659 ,   3711    ,   2127    ,   10465   ,0 };
            const unsigned long dim1237JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   13  ,   7   ,   43  ,   27  ,   211 ,   407 ,   153 ,   1939    ,   3243    ,   6655    ,   15983   ,0 };
            const unsigned long dim1238JoeKuoD5Init[]   =   {   1   ,   1   ,   1   ,   5   ,   27  ,   21  ,   23  ,   63  ,   271 ,   515 ,   261 ,   1947    ,   6257    ,   12861   ,0 };
            const unsigned long dim1239JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   1   ,   19  ,   57  ,   67  ,   221 ,   179 ,   725 ,   1179    ,   3983    ,   1585    ,   5899    ,0 };
            const unsigned long dim1240JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   15  ,   7   ,   61  ,   97  ,   131 ,   145 ,   637 ,   733 ,   2533    ,   1993    ,   14399   ,0 };
            const unsigned long dim1241JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   15  ,   13  ,   39  ,   83  ,   247 ,   235 ,   621 ,   1557    ,   3075    ,   3165    ,   16027   ,0 };
            const unsigned long dim1242JoeKuoD5Init[]   =   {   1   ,   1   ,   1   ,   15  ,   9   ,   11  ,   111 ,   153 ,   437 ,   687 ,   1845    ,   3547    ,   2097    ,   2219    ,0 };
            const unsigned long dim1243JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   13  ,   21  ,   13  ,   43  ,   33  ,   51  ,   1021    ,   1883    ,   1261    ,   2823    ,   12771   ,0 };
            const unsigned long dim1244JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   9   ,   3   ,   35  ,   19  ,   127 ,   419 ,   203 ,   1869    ,   1477    ,   5239    ,   6113    ,0 };
            const unsigned long dim1245JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   13  ,   31  ,   63  ,   75  ,   185 ,   507 ,   401 ,   1079    ,   67  ,   1621    ,   2849    ,0 };
            const unsigned long dim1246JoeKuoD5Init[]   =   {   1   ,   1   ,   1   ,   1   ,   31  ,   61  ,   125 ,   77  ,   169 ,   51  ,   1115    ,   1625    ,   3533    ,   5953    ,0 };
            const unsigned long dim1247JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   15  ,   29  ,   51  ,   23  ,   43  ,   505 ,   313 ,   887 ,   551 ,   4401    ,   2133    ,0 };
            const unsigned long dim1248JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   15  ,   11  ,   37  ,   55  ,   161 ,   353 ,   347 ,   1991    ,   4009    ,   2073    ,   12169   ,0 };
            const unsigned long dim1249JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   11  ,   29  ,   51  ,   15  ,   13  ,   391 ,   677 ,   225 ,   1467    ,   6615    ,   1895    ,0 };
            const unsigned long dim1250JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   3   ,   29  ,   15  ,   93  ,   103 ,   147 ,   323 ,   837 ,   1983    ,   1613    ,   10667   ,0 };
            const unsigned long dim1251JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   9   ,   21  ,   11  ,   39  ,   95  ,   473 ,   591 ,   1645    ,   67  ,   2793    ,   11217   ,0 };
            const unsigned long dim1252JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   5   ,   9   ,   51  ,   1   ,   243 ,   445 ,   583 ,   375 ,   2289    ,   4911    ,   10511   ,0 };
            const unsigned long dim1253JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   7   ,   23  ,   15  ,   111 ,   239 ,   247 ,   511 ,   1723    ,   2911    ,   2627    ,   10549   ,0 };
            const unsigned long dim1254JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   9   ,   15  ,   37  ,   31  ,   135 ,   347 ,   599 ,   765 ,   2561    ,   4635    ,   10659   ,0 };
            const unsigned long dim1255JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   11  ,   9   ,   49  ,   93  ,   187 ,   415 ,   247 ,   1871    ,   617 ,   6711    ,   3283    ,0 };
            const unsigned long dim1256JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   11  ,   25  ,   25  ,   37  ,   97  ,   365 ,   477 ,   1383    ,   1357    ,   693 ,   14743   ,0 };
            const unsigned long dim1257JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   11  ,   25  ,   19  ,   109 ,   99  ,   471 ,   661 ,   1633    ,   3773    ,   921 ,   4113    ,0 };
            const unsigned long dim1258JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   11  ,   13  ,   61  ,   21  ,   255 ,   161 ,   233 ,   1181    ,   3003    ,   1465    ,   9299    ,0 };
            const unsigned long dim1259JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   5   ,   7   ,   59  ,   53  ,   129 ,   127 ,   961 ,   893 ,   499 ,   5265    ,   2299    ,0 };
            const unsigned long dim1260JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   9   ,   29  ,   45  ,   85  ,   173 ,   131 ,   775 ,   1527    ,   1899    ,   4833    ,   12763   ,0 };
            const unsigned long dim1261JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   3   ,   1   ,   1   ,   115 ,   245 ,   205 ,   609 ,   1729    ,   915 ,   5965    ,   11001   ,0 };
            const unsigned long dim1262JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   15  ,   3   ,   61  ,   91  ,   93  ,   363 ,   457 ,   1497    ,   2539    ,   553 ,   7581    ,0 };
            const unsigned long dim1263JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   1   ,   29  ,   5   ,   7   ,   171 ,   393 ,   381 ,   935 ,   3467    ,   6199    ,   6625    ,0 };
            const unsigned long dim1264JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   13  ,   15  ,   25  ,   65  ,   137 ,   349 ,   61  ,   1035    ,   591 ,   4317    ,   13949   ,0 };
            const unsigned long dim1265JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   9   ,   27  ,   17  ,   87  ,   233 ,   227 ,   341 ,   639 ,   1813    ,   871 ,   12871   ,0 };
            const unsigned long dim1266JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   13  ,   29  ,   15  ,   31  ,   81  ,   61  ,   857 ,   1305    ,   3631    ,   2919    ,   2093    ,0 };
            const unsigned long dim1267JoeKuoD5Init[]   =   {   1   ,   1   ,   1   ,   3   ,   27  ,   5   ,   49  ,   1   ,   49  ,   745 ,   543 ,   847 ,   2469    ,   3513    ,0 };
            const unsigned long dim1268JoeKuoD5Init[]   =   {   1   ,   1   ,   1   ,   3   ,   15  ,   39  ,   47  ,   73  ,   33  ,   469 ,   1809    ,   2105    ,   7995    ,   11285   ,0 };
            const unsigned long dim1269JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   15  ,   9   ,   9   ,   81  ,   211 ,   295 ,   791 ,   1267    ,   2945    ,   5639    ,   6967    ,0 };
            const unsigned long dim1270JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   9   ,   13  ,   11  ,   43  ,   221 ,   161 ,   263 ,   905 ,   2767    ,   4491    ,   7605    ,0 };
            const unsigned long dim1271JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   15  ,   13  ,   9   ,   91  ,   123 ,   449 ,   21  ,   941 ,   1391    ,   3469    ,   16027   ,0 };
            const unsigned long dim1272JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   11  ,   5   ,   13  ,   77  ,   69  ,   245 ,   905 ,   231 ,   547 ,   2933    ,   4307    ,0 };
            const unsigned long dim1273JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   1   ,   25  ,   59  ,   49  ,   49  ,   183 ,   213 ,   29  ,   1801    ,   6271    ,   16283   ,0 };
            const unsigned long dim1274JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   11  ,   3   ,   15  ,   35  ,   157 ,   87  ,   453 ,   1939    ,   2697    ,   3325    ,   8679    ,0 };
            const unsigned long dim1275JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   13  ,   7   ,   45  ,   77  ,   73  ,   203 ,   321 ,   425 ,   581 ,   481 ,   15367   ,0 };
            const unsigned long dim1276JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   15  ,   11  ,   51  ,   11  ,   59  ,   355 ,   677 ,   1565    ,   123 ,   2403    ,   12835   ,0 };
            const unsigned long dim1277JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   5   ,   7   ,   27  ,   17  ,   81  ,   295 ,   131 ,   955 ,   4065    ,   797 ,   16165   ,0 };
            const unsigned long dim1278JoeKuoD5Init[]   =   {   1   ,   1   ,   1   ,   5   ,   29  ,   63  ,   51  ,   215 ,   269 ,   1013    ,   517 ,   1857    ,   141 ,   4495    ,0 };
            const unsigned long dim1279JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   5   ,   17  ,   47  ,   121 ,   199 ,   177 ,   1023    ,   1009    ,   3535    ,   4825    ,   16349   ,0 };
            const unsigned long dim1280JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   1   ,   21  ,   25  ,   107 ,   165 ,   43  ,   213 ,   1847    ,   1945    ,   3463    ,   2259    ,0 };
            const unsigned long dim1281JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   7   ,   5   ,   5   ,   5   ,   65  ,   493 ,   725 ,   755 ,   111 ,   6673    ,   213 ,0 };
            const unsigned long dim1282JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   1   ,   3   ,   29  ,   17  ,   87  ,   207 ,   793 ,   873 ,   2341    ,   3505    ,   5751    ,0 };
            const unsigned long dim1283JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   7   ,   27  ,   47  ,   123 ,   157 ,   427 ,   273 ,   139 ,   4043    ,   4083    ,   14121   ,0 };
            const unsigned long dim1284JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   9   ,   25  ,   17  ,   41  ,   245 ,   441 ,   47  ,   1   ,   2393    ,   405 ,   4021    ,0 };
            const unsigned long dim1285JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   13  ,   31  ,   9   ,   115 ,   165 ,   93  ,   701 ,   255 ,   895 ,   995 ,   12371   ,0 };
            const unsigned long dim1286JoeKuoD5Init[]   =   {   1   ,   1   ,   1   ,   11  ,   23  ,   13  ,   9   ,   141 ,   31  ,   973 ,   441 ,   3335    ,   2567    ,   6993    ,0 };
            const unsigned long dim1287JoeKuoD5Init[]   =   {   1   ,   1   ,   1   ,   13  ,   19  ,   25  ,   61  ,   139 ,   39  ,   987 ,   385 ,   2199    ,   7675    ,   13301   ,0 };
            const unsigned long dim1288JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   3   ,   1   ,   35  ,   121 ,   187 ,   273 ,   905 ,   245 ,   1031    ,   6203    ,   8165    ,0 };
            const unsigned long dim1289JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   13  ,   31  ,   59  ,   97  ,   45  ,   25  ,   803 ,   1245    ,   2659    ,   7471    ,   8367    ,0 };
            const unsigned long dim1290JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   11  ,   21  ,   41  ,   47  ,   183 ,   419 ,   279 ,   1901    ,   1081    ,   3575    ,   8591    ,0 };
            const unsigned long dim1291JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   5   ,   11  ,   39  ,   119 ,   245 ,   403 ,   703 ,   1547    ,   743 ,   7957    ,   15123   ,0 };
            const unsigned long dim1292JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   11  ,   11  ,   31  ,   71  ,   55  ,   3   ,   961 ,   991 ,   1665    ,   5539    ,   8187    ,0 };
            const unsigned long dim1293JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   3   ,   19  ,   7   ,   1   ,   35  ,   463 ,   1003    ,   555 ,   669 ,   2119    ,   10939   ,0 };
            const unsigned long dim1294JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   13  ,   17  ,   29  ,   99  ,   133 ,   107 ,   907 ,   7   ,   3047    ,   1263    ,   7497    ,0 };
            const unsigned long dim1295JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   15  ,   1   ,   53  ,   39  ,   191 ,   201 ,   189 ,   465 ,   453 ,   1967    ,   1033    ,0 };
            const unsigned long dim1296JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   13  ,   25  ,   19  ,   5   ,   155 ,   105 ,   923 ,   2045    ,   2889    ,   7685    ,   13847   ,0 };
            const unsigned long dim1297JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   13  ,   7   ,   61  ,   93  ,   41  ,   183 ,   783 ,   2011    ,   2967    ,   2949    ,   10247   ,0 };
            const unsigned long dim1298JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   9   ,   3   ,   53  ,   105 ,   63  ,   9   ,   131 ,   897 ,   347 ,   7683    ,   16027   ,0 };
            const unsigned long dim1299JoeKuoD5Init[]   =   {   1   ,   1   ,   1   ,   9   ,   19  ,   37  ,   37  ,   143 ,   37  ,   493 ,   533 ,   733 ,   2295    ,   4203    ,0 };
            const unsigned long dim1300JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   5   ,   15  ,   23  ,   83  ,   185 ,   495 ,   1023    ,   1473    ,   1501    ,   373 ,   201 ,0 };
            const unsigned long dim1301JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   13  ,   31  ,   51  ,   115 ,   233 ,   289 ,   201 ,   869 ,   3177    ,   4649    ,   16111   ,0 };
            const unsigned long dim1302JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   5   ,   23  ,   7   ,   59  ,   233 ,   447 ,   781 ,   1249    ,   71  ,   5857    ,   15481   ,0 };
            const unsigned long dim1303JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   11  ,   31  ,   1   ,   67  ,   243 ,   129 ,   737 ,   1443    ,   3647    ,   2391    ,   3635    ,0 };
            const unsigned long dim1304JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   1   ,   17  ,   39  ,   117 ,   81  ,   395 ,   119 ,   413 ,   1295    ,   7889    ,   13569   ,0 };
            const unsigned long dim1305JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   5   ,   29  ,   5   ,   123 ,   89  ,   143 ,   779 ,   1173    ,   3211    ,   3027    ,   10145   ,0 };
            const unsigned long dim1306JoeKuoD5Init[]   =   {   1   ,   1   ,   1   ,   7   ,   19  ,   53  ,   39  ,   199 ,   487 ,   797 ,   123 ,   871 ,   6335    ,   7957    ,0 };
            const unsigned long dim1307JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   7   ,   13  ,   11  ,   105 ,   121 ,   403 ,   283 ,   413 ,   1957    ,   6557    ,   8429    ,0 };
            const unsigned long dim1308JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   3   ,   23  ,   21  ,   109 ,   37  ,   463 ,   613 ,   927 ,   1857    ,   7003    ,   3477    ,0 };
            const unsigned long dim1309JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   1   ,   15  ,   37  ,   81  ,   27  ,   259 ,   661 ,   287 ,   615 ,   6151    ,   13759   ,0 };
            const unsigned long dim1310JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   7   ,   27  ,   59  ,   85  ,   223 ,   499 ,   571 ,   1853    ,   1419    ,   7761    ,   8385    ,0 };
            const unsigned long dim1311JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   9   ,   23  ,   57  ,   53  ,   163 ,   437 ,   657 ,   851 ,   3177    ,   6477    ,   13003   ,0 };
            const unsigned long dim1312JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   13  ,   3   ,   7   ,   37  ,   167 ,   49  ,   595 ,   1493    ,   369 ,   687 ,   13463   ,0 };
            const unsigned long dim1313JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   11  ,   27  ,   63  ,   49  ,   157 ,   379 ,   779 ,   99  ,   3457    ,   4477    ,   6531    ,0 };
            const unsigned long dim1314JoeKuoD5Init[]   =   {   1   ,   1   ,   1   ,   1   ,   5   ,   55  ,   77  ,   159 ,   371 ,   369 ,   743 ,   3571    ,   1877    ,   14767   ,0 };
            const unsigned long dim1315JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   15  ,   25  ,   11  ,   65  ,   187 ,   253 ,   437 ,   1301    ,   2871    ,   6219    ,   817 ,0 };
            const unsigned long dim1316JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   13  ,   23  ,   33  ,   25  ,   49  ,   491 ,   315 ,   11  ,   2163    ,   6155    ,   23  ,0 };
            const unsigned long dim1317JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   7   ,   27  ,   31  ,   81  ,   7   ,   355 ,   289 ,   1481    ,   2969    ,   1067    ,   7399    ,0 };
            const unsigned long dim1318JoeKuoD5Init[]   =   {   1   ,   1   ,   1   ,   7   ,   1   ,   57  ,   39  ,   201 ,   271 ,   223 ,   1117    ,   727 ,   7491    ,   4043    ,0 };
            const unsigned long dim1319JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   7   ,   9   ,   57  ,   17  ,   161 ,   221 ,   385 ,   2027    ,   1195    ,   2489    ,   12377   ,0 };
            const unsigned long dim1320JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   3   ,   27  ,   59  ,   101 ,   27  ,   177 ,   1005    ,   63  ,   3029    ,   7345    ,   14429   ,0 };
            const unsigned long dim1321JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   7   ,   17  ,   5   ,   111 ,   239 ,   85  ,   57  ,   1625    ,   657 ,   5931    ,   4929    ,0 };
            const unsigned long dim1322JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   3   ,   7   ,   41  ,   29  ,   9   ,   73  ,   875 ,   1665    ,   325 ,   27  ,   997 ,0 };
            const unsigned long dim1323JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   7   ,   21  ,   61  ,   119 ,   247 ,   307 ,   1011    ,   1489    ,   2361    ,   5781    ,   2465    ,0 };
            const unsigned long dim1324JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   3   ,   1   ,   21  ,   105 ,   77  ,   57  ,   983 ,   1519    ,   3543    ,   5025    ,   14051   ,0 };
            const unsigned long dim1325JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   5   ,   13  ,   39  ,   113 ,   69  ,   81  ,   155 ,   101 ,   427 ,   733 ,   10085   ,0 };
            const unsigned long dim1326JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   3   ,   1   ,   25  ,   109 ,   109 ,   303 ,   323 ,   565 ,   2267    ,   2755    ,   9165    ,0 };
            const unsigned long dim1327JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   3   ,   15  ,   43  ,   103 ,   163 ,   265 ,   849 ,   1969    ,   2247    ,   4495    ,   7301    ,0 };
            const unsigned long dim1328JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   3   ,   23  ,   9   ,   81  ,   103 ,   193 ,   845 ,   1603    ,   2493    ,   4919    ,   10649   ,0 };
            const unsigned long dim1329JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   1   ,   9   ,   23  ,   91  ,   33  ,   115 ,   599 ,   1755    ,   53  ,   1757    ,   145 ,0 };
            const unsigned long dim1330JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   11  ,   1   ,   39  ,   5   ,   233 ,   399 ,   187 ,   943 ,   2325    ,   437 ,   4421    ,0 };
            const unsigned long dim1331JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   13  ,   11  ,   39  ,   29  ,   131 ,   363 ,   885 ,   1921    ,   3703    ,   4197    ,   9703    ,0 };
            const unsigned long dim1332JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   5   ,   21  ,   43  ,   27  ,   151 ,   193 ,   211 ,   1229    ,   4031    ,   681 ,   5103    ,0 };
            const unsigned long dim1333JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   3   ,   19  ,   23  ,   41  ,   47  ,   315 ,   169 ,   271 ,   1877    ,   6357    ,   7709    ,0 };
            const unsigned long dim1334JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   1   ,   11  ,   27  ,   123 ,   33  ,   287 ,   293 ,   335 ,   2331    ,   141 ,   10095   ,0 };
            const unsigned long dim1335JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   5   ,   15  ,   33  ,   93  ,   123 ,   277 ,   833 ,   115 ,   3799    ,   1519    ,   153 ,0 };
            const unsigned long dim1336JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   15  ,   11  ,   3   ,   97  ,   225 ,   179 ,   601 ,   687 ,   253 ,   2839    ,   10985   ,0 };
            const unsigned long dim1337JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   3   ,   11  ,   51  ,   5   ,   141 ,   487 ,   325 ,   1691    ,   1291    ,   4677    ,   9087    ,0 };
            const unsigned long dim1338JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   9   ,   5   ,   11  ,   99  ,   195 ,   459 ,   171 ,   361 ,   1621    ,   5377    ,   4651    ,0 };
            const unsigned long dim1339JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   3   ,   29  ,   47  ,   5   ,   29  ,   107 ,   751 ,   739 ,   2815    ,   3709    ,   15493   ,0 };
            const unsigned long dim1340JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   1   ,   3   ,   63  ,   47  ,   15  ,   433 ,   501 ,   1687    ,   2035    ,   6263    ,   12681   ,0 };
            const unsigned long dim1341JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   13  ,   1   ,   57  ,   91  ,   33  ,   217 ,   141 ,   2005    ,   2405    ,   1987    ,   14957   ,0 };
            const unsigned long dim1342JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   13  ,   25  ,   23  ,   93  ,   37  ,   129 ,   747 ,   1607    ,   849 ,   2119    ,   11855   ,0 };
            const unsigned long dim1343JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   9   ,   15  ,   55  ,   55  ,   125 ,   235 ,   455 ,   2027    ,   1709    ,   7217    ,   10341   ,0 };
            const unsigned long dim1344JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   7   ,   1   ,   3   ,   69  ,   167 ,   373 ,   87  ,   901 ,   2333    ,   6751    ,   5809    ,0 };
            const unsigned long dim1345JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   9   ,   19  ,   25  ,   67  ,   197 ,   395 ,   735 ,   941 ,   1753    ,   3923    ,   8805    ,0 };
            const unsigned long dim1346JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   9   ,   17  ,   29  ,   1   ,   205 ,   511 ,   179 ,   1191    ,   17  ,   3179    ,   6891    ,0 };
            const unsigned long dim1347JoeKuoD5Init[]   =   {   1   ,   1   ,   1   ,   9   ,   9   ,   31  ,   99  ,   133 ,   253 ,   239 ,   1729    ,   4093    ,   5759    ,   15357   ,0 };
            const unsigned long dim1348JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   3   ,   31  ,   35  ,   125 ,   167 ,   417 ,   431 ,   709 ,   415 ,   1093    ,   11361   ,0 };
            const unsigned long dim1349JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   11  ,   7   ,   59  ,   57  ,   89  ,   119 ,   537 ,   1157    ,   2539    ,   5783    ,   15093   ,0 };
            const unsigned long dim1350JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   7   ,   17  ,   41  ,   105 ,   33  ,   301 ,   601 ,   537 ,   3877    ,   797 ,   1319    ,0 };
            const unsigned long dim1351JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   7   ,   3   ,   59  ,   79  ,   47  ,   191 ,   985 ,   83  ,   3535    ,   4135    ,   16165   ,0 };
            const unsigned long dim1352JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   3   ,   13  ,   61  ,   35  ,   193 ,   141 ,   961 ,   1733    ,   4051    ,   2657    ,   9183    ,0 };
            const unsigned long dim1353JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   9   ,   25  ,   49  ,   127 ,   233 ,   291 ,   445 ,   1639    ,   4023    ,   4791    ,   279 ,0 };
            const unsigned long dim1354JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   11  ,   15  ,   39  ,   91  ,   81  ,   433 ,   7   ,   1897    ,   2659    ,   7877    ,   15733   ,0 };
            const unsigned long dim1355JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   5   ,   19  ,   5   ,   125 ,   181 ,   305 ,   377 ,   1699    ,   2157    ,   4617    ,   7165    ,0 };
            const unsigned long dim1356JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   13  ,   11  ,   27  ,   19  ,   217 ,   171 ,   343 ,   61  ,   3799    ,   4923    ,   14279   ,0 };
            const unsigned long dim1357JoeKuoD5Init[]   =   {   1   ,   1   ,   1   ,   15  ,   13  ,   41  ,   19  ,   19  ,   359 ,   101 ,   1795    ,   127 ,   7067    ,   4327    ,0 };
            const unsigned long dim1358JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   15  ,   15  ,   5   ,   43  ,   47  ,   103 ,   549 ,   767 ,   2695    ,   1689    ,   5569    ,0 };
            const unsigned long dim1359JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   9   ,   7   ,   33  ,   35  ,   105 ,   283 ,   957 ,   1255    ,   2085    ,   6263    ,   4537    ,0 };
            const unsigned long dim1360JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   9   ,   27  ,   23  ,   49  ,   169 ,   455 ,   163 ,   301 ,   3107    ,   6859    ,   14477   ,0 };
            const unsigned long dim1361JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   15  ,   13  ,   11  ,   25  ,   231 ,   171 ,   173 ,   661 ,   1921    ,   3535    ,   10157   ,0 };
            const unsigned long dim1362JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   11  ,   27  ,   19  ,   23  ,   97  ,   409 ,   347 ,   1413    ,   2273    ,   7305    ,   12597   ,0 };
            const unsigned long dim1363JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   15  ,   11  ,   63  ,   49  ,   247 ,   459 ,   195 ,   1579    ,   539 ,   6283    ,   14829   ,0 };
            const unsigned long dim1364JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   5   ,   19  ,   37  ,   105 ,   31  ,   87  ,   413 ,   511 ,   271 ,   6265    ,   9499    ,0 };
            const unsigned long dim1365JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   13  ,   23  ,   39  ,   97  ,   239 ,   397 ,   975 ,   1369    ,   2397    ,   409 ,   3495    ,0 };
            const unsigned long dim1366JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   1   ,   25  ,   45  ,   11  ,   81  ,   233 ,   299 ,   1269    ,   1129    ,   1679    ,   10195   ,0 };
            const unsigned long dim1367JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   5   ,   13  ,   29  ,   3   ,   189 ,   231 ,   535 ,   1201    ,   1215    ,   1889    ,   6169    ,0 };
            const unsigned long dim1368JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   15  ,   31  ,   53  ,   17  ,   197 ,   453 ,   13  ,   181 ,   2663    ,   3869    ,   7269    ,0 };
            const unsigned long dim1369JoeKuoD5Init[]   =   {   1   ,   1   ,   1   ,   5   ,   27  ,   37  ,   71  ,   89  ,   505 ,   769 ,   1359    ,   295 ,   6061    ,   8363    ,0 };
            const unsigned long dim1370JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   1   ,   9   ,   1   ,   111 ,   157 ,   89  ,   777 ,   1713    ,   117 ,   285 ,   6353    ,0 };
            const unsigned long dim1371JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   11  ,   29  ,   53  ,   57  ,   3   ,   225 ,   885 ,   1445    ,   3673    ,   7857    ,   3843    ,0 };
            const unsigned long dim1372JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   11  ,   29  ,   23  ,   43  ,   19  ,   175 ,   573 ,   1709    ,   2303    ,   5607    ,   4347    ,0 };
            const unsigned long dim1373JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   7   ,   19  ,   41  ,   93  ,   13  ,   3   ,   483 ,   1365    ,   411 ,   5147    ,   10505   ,0 };
            const unsigned long dim1374JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   7   ,   3   ,   17  ,   51  ,   23  ,   19  ,   411 ,   741 ,   877 ,   7121    ,   7639    ,0 };
            const unsigned long dim1375JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   3   ,   7   ,   45  ,   103 ,   69  ,   387 ,   803 ,   29  ,   1469    ,   2139    ,   6397    ,0 };
            const unsigned long dim1376JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   3   ,   11  ,   43  ,   5   ,   19  ,   441 ,   285 ,   1657    ,   2133    ,   6343    ,   11817   ,0 };
            const unsigned long dim1377JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   1   ,   31  ,   31  ,   19  ,   15  ,   475 ,   131 ,   1687    ,   1647    ,   4685    ,   1135    ,0 };
            const unsigned long dim1378JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   7   ,   17  ,   59  ,   125 ,   127 ,   63  ,   451 ,   949 ,   4041    ,   6649    ,   12187   ,0 };
            const unsigned long dim1379JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   1   ,   25  ,   13  ,   59  ,   159 ,   107 ,   509 ,   787 ,   2517    ,   6679    ,   2809    ,0 };
            const unsigned long dim1380JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   5   ,   17  ,   5   ,   87  ,   235 ,   379 ,   599 ,   1971    ,   969 ,   853 ,   10481   ,0 };
            const unsigned long dim1381JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   15  ,   7   ,   35  ,   47  ,   113 ,   349 ,   451 ,   1827    ,   2647    ,   367 ,   2581    ,0 };
            const unsigned long dim1382JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   9   ,   1   ,   39  ,   29  ,   45  ,   365 ,   317 ,   129 ,   137 ,   5975    ,   3353    ,0 };
            const unsigned long dim1383JoeKuoD5Init[]   =   {   1   ,   1   ,   1   ,   15  ,   17  ,   3   ,   5   ,   61  ,   11  ,   587 ,   769 ,   2127    ,   2625    ,   12545   ,0 };
            const unsigned long dim1384JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   7   ,   21  ,   1   ,   127 ,   235 ,   343 ,   807 ,   147 ,   3517    ,   3471    ,   4625    ,0 };
            const unsigned long dim1385JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   15  ,   31  ,   23  ,   81  ,   23  ,   171 ,   403 ,   1083    ,   4049    ,   1959    ,   16307   ,0 };
            const unsigned long dim1386JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   9   ,   21  ,   37  ,   99  ,   83  ,   33  ,   705 ,   369 ,   2445    ,   3253    ,   16171   ,0 };
            const unsigned long dim1387JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   9   ,   17  ,   61  ,   121 ,   75  ,   89  ,   823 ,   1519    ,   1997    ,   6433    ,   5013    ,0 };
            const unsigned long dim1388JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   11  ,   7   ,   47  ,   7   ,   71  ,   475 ,   517 ,   1271    ,   3815    ,   5969    ,   607 ,0 };
            const unsigned long dim1389JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   3   ,   13  ,   15  ,   37  ,   89  ,   43  ,   489 ,   1853    ,   1195    ,   3097    ,   10297   ,0 };
            const unsigned long dim1390JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   11  ,   25  ,   39  ,   95  ,   43  ,   145 ,   231 ,   1859    ,   3201    ,   1377    ,   7091    ,0 };
            const unsigned long dim1391JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   3   ,   13  ,   43  ,   117 ,   111 ,   231 ,   101 ,   1801    ,   739 ,   945 ,   15585   ,0 };
            const unsigned long dim1392JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   7   ,   23  ,   3   ,   35  ,   167 ,   485 ,   951 ,   1729    ,   1831    ,   2639    ,   6561    ,0 };
            const unsigned long dim1393JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   15  ,   5   ,   31  ,   103 ,   85  ,   111 ,   545 ,   789 ,   945 ,   2691    ,   327 ,0 };
            const unsigned long dim1394JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   11  ,   29  ,   61  ,   127 ,   59  ,   137 ,   485 ,   1673    ,   3295    ,   4185    ,   6489    ,0 };
            const unsigned long dim1395JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   9   ,   31  ,   9   ,   115 ,   11  ,   73  ,   267 ,   195 ,   1445    ,   873 ,   7285    ,0 };
            const unsigned long dim1396JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   7   ,   3   ,   19  ,   9   ,   71  ,   287 ,   89  ,   329 ,   953 ,   2237    ,   16341   ,0 };
            const unsigned long dim1397JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   13  ,   19  ,   31  ,   5   ,   49  ,   293 ,   65  ,   291 ,   93  ,   2553    ,   8407    ,0 };
            const unsigned long dim1398JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   15  ,   17  ,   13  ,   15  ,   111 ,   211 ,   935 ,   1165    ,   2975    ,   339 ,   16333   ,0 };
            const unsigned long dim1399JoeKuoD5Init[]   =   {   1   ,   1   ,   1   ,   3   ,   17  ,   57  ,   9   ,   63  ,   243 ,   431 ,   289 ,   3493    ,   2879    ,   4801    ,0 };
            const unsigned long dim1400JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   3   ,   17  ,   15  ,   53  ,   191 ,   439 ,   981 ,   751 ,   4025    ,   7177    ,   4887    ,0 };
            const unsigned long dim1401JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   3   ,   11  ,   61  ,   43  ,   103 ,   151 ,   33  ,   421 ,   1949    ,   5915    ,   5515    ,0 };
            const unsigned long dim1402JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   3   ,   1   ,   55  ,   45  ,   227 ,   27  ,   491 ,   1479    ,   3323    ,   5485    ,   5493    ,0 };
            const unsigned long dim1403JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   1   ,   9   ,   11  ,   1   ,   249 ,   113 ,   415 ,   295 ,   3437    ,   3877    ,   6675    ,0 };
            const unsigned long dim1404JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   11  ,   9   ,   43  ,   127 ,   57  ,   115 ,   685 ,   165 ,   973 ,   2707    ,   2503    ,0 };
            const unsigned long dim1405JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   7   ,   1   ,   29  ,   125 ,   45  ,   307 ,   625 ,   1477    ,   2565    ,   2949    ,   1729    ,0 };
            const unsigned long dim1406JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   1   ,   23  ,   7   ,   45  ,   245 ,   301 ,   531 ,   1419    ,   1795    ,   5757    ,   15219   ,0 };
            const unsigned long dim1407JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   7   ,   15  ,   49  ,   71  ,   109 ,   145 ,   79  ,   1333    ,   1589    ,   3851    ,   2879    ,0 };
            const unsigned long dim1408JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   5   ,   11  ,   15  ,   37  ,   239 ,   217 ,   193 ,   1687    ,   1721    ,   8059    ,   9027    ,0 };
            const unsigned long dim1409JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   13  ,   13  ,   49  ,   57  ,   189 ,   433 ,   569 ,   1285    ,   1891    ,   6079    ,   13469   ,0 };
            const unsigned long dim1410JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   1   ,   19  ,   61  ,   27  ,   181 ,   365 ,   121 ,   883 ,   1611    ,   1521    ,   11437   ,0 };
            const unsigned long dim1411JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   3   ,   17  ,   37  ,   71  ,   71  ,   495 ,   519 ,   879 ,   2993    ,   6275    ,   14345   ,0 };
            const unsigned long dim1412JoeKuoD5Init[]   =   {   1   ,   1   ,   1   ,   5   ,   29  ,   5   ,   121 ,   29  ,   293 ,   745 ,   1839    ,   2061    ,   2721    ,   11741   ,0 };
            const unsigned long dim1413JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   13  ,   23  ,   61  ,   55  ,   99  ,   409 ,   211 ,   783 ,   1841    ,   193 ,   1941    ,0 };
            const unsigned long dim1414JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   7   ,   11  ,   49  ,   87  ,   81  ,   225 ,   211 ,   1263    ,   1403    ,   6169    ,   6235    ,0 };
            const unsigned long dim1415JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   9   ,   11  ,   35  ,   71  ,   145 ,   417 ,   485 ,   1565    ,   2101    ,   4153    ,   5239    ,0 };
            const unsigned long dim1416JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   1   ,   13  ,   41  ,   45  ,   79  ,   157 ,   477 ,   677 ,   3961    ,   1127    ,   1139    ,0 };
            const unsigned long dim1417JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   13  ,   27  ,   39  ,   23  ,   197 ,   123 ,   431 ,   273 ,   2723    ,   1303    ,   13271   ,0 };
            const unsigned long dim1418JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   9   ,   13  ,   21  ,   59  ,   109 ,   267 ,   173 ,   997 ,   2701    ,   3719    ,   3703    ,0 };
            const unsigned long dim1419JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   1   ,   1   ,   5   ,   57  ,   181 ,   241 ,   1   ,   1063    ,   199 ,   8181    ,   12721   ,0 };
            const unsigned long dim1420JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   9   ,   15  ,   61  ,   107 ,   231 ,   65  ,   933 ,   677 ,   3883    ,   2621    ,   8821    ,0 };
            const unsigned long dim1421JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   1   ,   9   ,   33  ,   41  ,   55  ,   503 ,   683 ,   747 ,   3619    ,   7885    ,   851 ,0 };
            const unsigned long dim1422JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   5   ,   9   ,   25  ,   27  ,   123 ,   97  ,   731 ,   583 ,   2535    ,   1267    ,   5921    ,0 };
            const unsigned long dim1423JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   15  ,   27  ,   63  ,   69  ,   99  ,   475 ,   709 ,   1239    ,   861 ,   1229    ,   11369   ,0 };
            const unsigned long dim1424JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   13  ,   17  ,   47  ,   123 ,   159 ,   57  ,   871 ,   465 ,   783 ,   4093    ,   15277   ,0 };
            const unsigned long dim1425JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   9   ,   17  ,   47  ,   89  ,   139 ,   301 ,   45  ,   627 ,   4073    ,   3187    ,   9633    ,0 };
            const unsigned long dim1426JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   11  ,   21  ,   43  ,   19  ,   25  ,   457 ,   879 ,   113 ,   847 ,   201 ,   15683   ,0 };
            const unsigned long dim1427JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   15  ,   9   ,   47  ,   43  ,   215 ,   407 ,   979 ,   51  ,   635 ,   467 ,   6365    ,0 };
            const unsigned long dim1428JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   5   ,   7   ,   11  ,   53  ,   5   ,   471 ,   317 ,   1719    ,   755 ,   5211    ,   7599    ,0 };
            const unsigned long dim1429JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   3   ,   11  ,   43  ,   67  ,   203 ,   15  ,   75  ,   1063    ,   1763    ,   3537    ,   13511   ,0 };
            const unsigned long dim1430JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   1   ,   17  ,   37  ,   91  ,   55  ,   95  ,   843 ,   751 ,   3501    ,   6203    ,   2999    ,0 };
            const unsigned long dim1431JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   13  ,   19  ,   61  ,   37  ,   157 ,   309 ,   361 ,   1499    ,   1845    ,   3675    ,   6221    ,0 };
            const unsigned long dim1432JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   11  ,   15  ,   9   ,   25  ,   123 ,   25  ,   525 ,   227 ,   3429    ,   1573    ,   12321   ,0 };
            const unsigned long dim1433JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   11  ,   9   ,   17  ,   91  ,   61  ,   263 ,   129 ,   1853    ,   1911    ,   5065    ,   775 ,0 };
            const unsigned long dim1434JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   3   ,   31  ,   15  ,   3   ,   207 ,   505 ,   123 ,   477 ,   1285    ,   7007    ,   2873    ,0 };
            const unsigned long dim1435JoeKuoD5Init[]   =   {   1   ,   1   ,   1   ,   7   ,   7   ,   15  ,   123 ,   249 ,   467 ,   229 ,   845 ,   1913    ,   461 ,   6235    ,0 };
            const unsigned long dim1436JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   9   ,   15  ,   9   ,   121 ,   211 ,   231 ,   491 ,   521 ,   3621    ,   7285    ,   3165    ,0 };
            const unsigned long dim1437JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   15  ,   21  ,   43  ,   1   ,   35  ,   339 ,   671 ,   719 ,   1739    ,   501 ,   9573    ,0 };
            const unsigned long dim1438JoeKuoD5Init[]   =   {   1   ,   1   ,   1   ,   5   ,   13  ,   37  ,   81  ,   1   ,   281 ,   785 ,   831 ,   991 ,   7485    ,   15619   ,0 };
            const unsigned long dim1439JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   1   ,   29  ,   47  ,   29  ,   99  ,   311 ,   519 ,   545 ,   1115    ,   6651    ,   793 ,0 };
            const unsigned long dim1440JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   11  ,   17  ,   15  ,   17  ,   65  ,   101 ,   411 ,   231 ,   2959    ,   8077    ,   673 ,0 };
            const unsigned long dim1441JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   9   ,   27  ,   47  ,   127 ,   237 ,   319 ,   293 ,   1109    ,   3863    ,   213 ,   2149    ,0 };
            const unsigned long dim1442JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   13  ,   11  ,   61  ,   27  ,   129 ,   57  ,   743 ,   889 ,   3707    ,   469 ,   11949   ,0 };
            const unsigned long dim1443JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   13  ,   23  ,   13  ,   119 ,   125 ,   75  ,   753 ,   1951    ,   1181    ,   291 ,   11737   ,0 };
            const unsigned long dim1444JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   5   ,   31  ,   19  ,   43  ,   113 ,   309 ,   721 ,   175 ,   1041    ,   7123    ,   103 ,0 };
            const unsigned long dim1445JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   9   ,   21  ,   19  ,   71  ,   175 ,   309 ,   9   ,   807 ,   961 ,   6741    ,   5075    ,0 };
            const unsigned long dim1446JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   9   ,   15  ,   27  ,   83  ,   207 ,   67  ,   71  ,   289 ,   2901    ,   7637    ,   9525    ,0 };
            const unsigned long dim1447JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   15  ,   21  ,   9   ,   19  ,   67  ,   401 ,   165 ,   1297    ,   3159    ,   2881    ,   15979   ,0 };
            const unsigned long dim1448JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   1   ,   11  ,   21  ,   27  ,   33  ,   95  ,   853 ,   1699    ,   875 ,   6519    ,   9109    ,0 };
            const unsigned long dim1449JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   3   ,   31  ,   5   ,   123 ,   215 ,   165 ,   405 ,   623 ,   845 ,   4149    ,   5015    ,0 };
            const unsigned long dim1450JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   7   ,   31  ,   13  ,   23  ,   55  ,   129 ,   789 ,   803 ,   2077    ,   1885    ,   1669    ,0 };
            const unsigned long dim1451JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   11  ,   31  ,   39  ,   93  ,   7   ,   485 ,   571 ,   417 ,   3839    ,   1289    ,   4127    ,0 };
            const unsigned long dim1452JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   11  ,   23  ,   63  ,   123 ,   249 ,   75  ,   515 ,   2009    ,   949 ,   3291    ,   727 ,0 };
            const unsigned long dim1453JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   13  ,   21  ,   15  ,   57  ,   235 ,   25  ,   507 ,   1055    ,   3161    ,   4351    ,   8855    ,0 };
            const unsigned long dim1454JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   7   ,   23  ,   39  ,   71  ,   151 ,   377 ,   469 ,   665 ,   1197    ,   3503    ,   655 ,0 };
            const unsigned long dim1455JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   3   ,   5   ,   35  ,   67  ,   203 ,   33  ,   319 ,   925 ,   1021    ,   6869    ,   5145    ,0 };
            const unsigned long dim1456JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   3   ,   21  ,   61  ,   119 ,   195 ,   383 ,   573 ,   135 ,   2371    ,   1665    ,   4957    ,0 };
            const unsigned long dim1457JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   1   ,   7   ,   47  ,   93  ,   201 ,   269 ,   329 ,   209 ,   1659    ,   1547    ,   4605    ,0 };
            const unsigned long dim1458JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   1   ,   9   ,   47  ,   71  ,   13  ,   451 ,   51  ,   1073    ,   3691    ,   6881    ,   14801   ,0 };
            const unsigned long dim1459JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   3   ,   13  ,   33  ,   75  ,   155 ,   283 ,   599 ,   517 ,   2251    ,   6217    ,   10487   ,0 };
            const unsigned long dim1460JoeKuoD5Init[]   =   {   1   ,   1   ,   1   ,   3   ,   7   ,   49  ,   65  ,   235 ,   159 ,   733 ,   939 ,   283 ,   6935    ,   14367   ,0 };
            const unsigned long dim1461JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   13  ,   7   ,   19  ,   71  ,   227 ,   307 ,   515 ,   1701    ,   747 ,   3475    ,   8165    ,0 };
            const unsigned long dim1462JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   11  ,   13  ,   19  ,   1   ,   97  ,   425 ,   739 ,   451 ,   3789    ,   5337    ,   2023    ,0 };
            const unsigned long dim1463JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   9   ,   7   ,   27  ,   23  ,   169 ,   137 ,   537 ,   61  ,   2207    ,   917 ,   1209    ,0 };
            const unsigned long dim1464JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   3   ,   29  ,   5   ,   121 ,   197 ,   159 ,   467 ,   581 ,   1679    ,   6605    ,   1989    ,0 };
            const unsigned long dim1465JoeKuoD5Init[]   =   {   1   ,   1   ,   1   ,   13  ,   5   ,   33  ,   61  ,   77  ,   383 ,   977 ,   781 ,   175 ,   8151    ,   7979    ,0 };
            const unsigned long dim1466JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   13  ,   1   ,   1   ,   3   ,   95  ,   193 ,   453 ,   649 ,   1137    ,   485 ,   14345   ,0 };
            const unsigned long dim1467JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   7   ,   1   ,   13  ,   41  ,   159 ,   327 ,   105 ,   1569    ,   475 ,   1295    ,   3767    ,0 };
            const unsigned long dim1468JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   3   ,   25  ,   7   ,   73  ,   235 ,   271 ,   491 ,   1385    ,   2567    ,   1463    ,   12731   ,0 };
            const unsigned long dim1469JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   1   ,   9   ,   9   ,   37  ,   199 ,   249 ,   9   ,   299 ,   3891    ,   2373    ,   11553   ,0 };
            const unsigned long dim1470JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   13  ,   19  ,   47  ,   81  ,   25  ,   125 ,   933 ,   1637    ,   469 ,   6351    ,   4219    ,0 };
            const unsigned long dim1471JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   7   ,   31  ,   5   ,   67  ,   13  ,   63  ,   615 ,   1089    ,   2291    ,   3105    ,   7009    ,0 };
            const unsigned long dim1472JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   1   ,   1   ,   43  ,   49  ,   141 ,   189 ,   1015    ,   1527    ,   1511    ,   3093    ,   5497    ,0 };
            const unsigned long dim1473JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   1   ,   3   ,   51  ,   121 ,   89  ,   161 ,   517 ,   467 ,   2837    ,   2275    ,   4987    ,0 };
            const unsigned long dim1474JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   13  ,   9   ,   63  ,   31  ,   109 ,   331 ,   607 ,   1271    ,   3639    ,   617 ,   13177   ,0 };
            const unsigned long dim1475JoeKuoD5Init[]   =   {   1   ,   1   ,   1   ,   15  ,   9   ,   25  ,   73  ,   237 ,   145 ,   885 ,   1945    ,   1871    ,   5401    ,   15403   ,0 };
            const unsigned long dim1476JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   5   ,   1   ,   63  ,   29  ,   145 ,   377 ,   429 ,   1663    ,   1643    ,   2713    ,   6621    ,0 };
            const unsigned long dim1477JoeKuoD5Init[]   =   {   1   ,   1   ,   1   ,   5   ,   29  ,   7   ,   87  ,   53  ,   129 ,   503 ,   961 ,   3535    ,   3255    ,   7621    ,0 };
            const unsigned long dim1478JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   1   ,   5   ,   47  ,   13  ,   243 ,   155 ,   863 ,   103 ,   1699    ,   5063    ,   8221    ,0 };
            const unsigned long dim1479JoeKuoD5Init[]   =   {   1   ,   1   ,   1   ,   9   ,   31  ,   61  ,   1   ,   253 ,   67  ,   369 ,   521 ,   3429    ,   6935    ,   3383    ,0 };
            const unsigned long dim1480JoeKuoD5Init[]   =   {   1   ,   1   ,   1   ,   3   ,   11  ,   61  ,   11  ,   187 ,   145 ,   7   ,   535 ,   831 ,   933 ,   7779    ,0 };
            const unsigned long dim1481JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   1   ,   25  ,   33  ,   117 ,   161 ,   169 ,   33  ,   1415    ,   1493    ,   1599    ,   1109    ,0 };
            const unsigned long dim1482JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   3   ,   29  ,   19  ,   75  ,   143 ,   467 ,   785 ,   455 ,   2593    ,   7539    ,   6283    ,0 };
            const unsigned long dim1483JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   15  ,   11  ,   5   ,   19  ,   187 ,   357 ,   955 ,   631 ,   3697    ,   4641    ,   14353   ,0 };
            const unsigned long dim1484JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   15  ,   3   ,   21  ,   25  ,   79  ,   375 ,   611 ,   915 ,   2491    ,   5691    ,   773 ,0 };
            const unsigned long dim1485JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   7   ,   29  ,   23  ,   7   ,   27  ,   207 ,   157 ,   1021    ,   2411    ,   5061    ,   7493    ,0 };
            const unsigned long dim1486JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   5   ,   13  ,   7   ,   7   ,   159 ,   393 ,   51  ,   1573    ,   1353    ,   2373    ,   12721   ,0 };
            const unsigned long dim1487JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   13  ,   5   ,   55  ,   43  ,   7   ,   411 ,   353 ,   379 ,   2213    ,   6257    ,   5825    ,0 };
            const unsigned long dim1488JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   1   ,   5   ,   37  ,   65  ,   69  ,   251 ,   893 ,   1747    ,   4065    ,   3937    ,   1855    ,0 };
            const unsigned long dim1489JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   1   ,   3   ,   11  ,   97  ,   111 ,   509 ,   569 ,   839 ,   2431    ,   3475    ,   12283   ,0 };
            const unsigned long dim1490JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   11  ,   13  ,   63  ,   85  ,   193 ,   351 ,   491 ,   205 ,   1051    ,   403 ,   1749    ,0 };
            const unsigned long dim1491JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   1   ,   27  ,   61  ,   61  ,   169 ,   189 ,   549 ,   1589    ,   3567    ,   7301    ,   12723   ,0 };
            const unsigned long dim1492JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   7   ,   13  ,   39  ,   121 ,   75  ,   261 ,   919 ,   1557    ,   635 ,   2123    ,   3771    ,0 };
            const unsigned long dim1493JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   13  ,   7   ,   39  ,   107 ,   109 ,   165 ,   91  ,   1049    ,   3897    ,   1395    ,   9573    ,0 };
            const unsigned long dim1494JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   5   ,   19  ,   9   ,   61  ,   209 ,   205 ,   363 ,   823 ,   2445    ,   7301    ,   7141    ,0 };
            const unsigned long dim1495JoeKuoD5Init[]   =   {   1   ,   1   ,   1   ,   15  ,   5   ,   31  ,   13  ,   61  ,   153 ,   847 ,   67  ,   2227    ,   4119    ,   9231    ,0 };
            const unsigned long dim1496JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   9   ,   13  ,   63  ,   93  ,   199 ,   437 ,   319 ,   735 ,   2015    ,   1719    ,   3253    ,0 };
            const unsigned long dim1497JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   5   ,   31  ,   19  ,   27  ,   141 ,   357 ,   647 ,   895 ,   345 ,   5937    ,   15711   ,0 };
            const unsigned long dim1498JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   1   ,   15  ,   35  ,   73  ,   69  ,   205 ,   545 ,   387 ,   3487    ,   7391    ,   3337    ,0 };
            const unsigned long dim1499JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   5   ,   27  ,   55  ,   51  ,   179 ,   125 ,   1013    ,   35  ,   2741    ,   7793    ,   4347    ,0 };
            const unsigned long dim1500JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   9   ,   31  ,   57  ,   55  ,   3   ,   385 ,   421 ,   1543    ,   2809    ,   1887    ,   13709   ,0 };
            const unsigned long dim1501JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   7   ,   29  ,   23  ,   115 ,   25  ,   295 ,   267 ,   101 ,   3005    ,   2601    ,   4959    ,0 };
            const unsigned long dim1502JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   13  ,   27  ,   11  ,   11  ,   35  ,   257 ,   23  ,   2013    ,   1369    ,   6503    ,   2589    ,0 };
            const unsigned long dim1503JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   15  ,   5   ,   43  ,   101 ,   87  ,   231 ,   761 ,   1991    ,   3167    ,   5689    ,   9565    ,0 };
            const unsigned long dim1504JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   11  ,   17  ,   37  ,   65  ,   249 ,   119 ,   727 ,   793 ,   929 ,   6275    ,   12173   ,0 };
            const unsigned long dim1505JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   9   ,   1   ,   11  ,   93  ,   27  ,   291 ,   411 ,   1069    ,   1283    ,   7593    ,   4335    ,0 };
            const unsigned long dim1506JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   1   ,   23  ,   45  ,   51  ,   93  ,   401 ,   515 ,   749 ,   1293    ,   8155    ,   15123   ,0 };
            const unsigned long dim1507JoeKuoD5Init[]   =   {   1   ,   1   ,   1   ,   3   ,   21  ,   17  ,   67  ,   69  ,   371 ,   507 ,   143 ,   2393    ,   6267    ,   7143    ,0 };
            const unsigned long dim1508JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   11  ,   1   ,   5   ,   79  ,   75  ,   361 ,   689 ,   339 ,   1855    ,   6863    ,   15841   ,0 };
            const unsigned long dim1509JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   13  ,   23  ,   3   ,   43  ,   57  ,   233 ,   927 ,   1095    ,   1827    ,   401 ,   825 ,0 };
            const unsigned long dim1510JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   9   ,   31  ,   7   ,   67  ,   201 ,   27  ,   699 ,   535 ,   3073    ,   6895    ,   3021    ,0 };
            const unsigned long dim1511JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   13  ,   1   ,   19  ,   13  ,   83  ,   221 ,   727 ,   745 ,   2131    ,   3757    ,   1493    ,0 };
            const unsigned long dim1512JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   9   ,   13  ,   29  ,   101 ,   255 ,   243 ,   763 ,   535 ,   169 ,   1987    ,   4071    ,0 };
            const unsigned long dim1513JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   3   ,   21  ,   33  ,   117 ,   213 ,   143 ,   351 ,   1735    ,   1651    ,   5781    ,   8803    ,0 };
            const unsigned long dim1514JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   7   ,   17  ,   39  ,   57  ,   209 ,   141 ,   865 ,   1731    ,   3349    ,   7107    ,   15983   ,0 };
            const unsigned long dim1515JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   3   ,   21  ,   33  ,   117 ,   201 ,   481 ,   711 ,   1207    ,   1971    ,   3353    ,   5827    ,0 };
            const unsigned long dim1516JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   13  ,   13  ,   1   ,   121 ,   115 ,   449 ,   801 ,   1507    ,   2323    ,   6709    ,   5533    ,0 };
            const unsigned long dim1517JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   7   ,   5   ,   5   ,   35  ,   81  ,   469 ,   691 ,   443 ,   501 ,   6745    ,   421 ,0 };
            const unsigned long dim1518JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   13  ,   27  ,   41  ,   83  ,   39  ,   349 ,   585 ,   551 ,   643 ,   6659    ,   61  ,0 };
            const unsigned long dim1519JoeKuoD5Init[]   =   {   1   ,   1   ,   1   ,   3   ,   25  ,   41  ,   45  ,   203 ,   89  ,   305 ,   1433    ,   1879    ,   6703    ,   14323   ,0 };
            const unsigned long dim1520JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   15  ,   27  ,   37  ,   97  ,   117 ,   431 ,   197 ,   807 ,   1841    ,   8075    ,   4613    ,0 };
            const unsigned long dim1521JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   3   ,   23  ,   47  ,   69  ,   129 ,   451 ,   113 ,   1397    ,   3005    ,   7599    ,   101 ,0 };
            const unsigned long dim1522JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   1   ,   11  ,   47  ,   117 ,   181 ,   85  ,   545 ,   571 ,   3227    ,   4097    ,   7937    ,0 };
            const unsigned long dim1523JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   9   ,   27  ,   53  ,   11  ,   181 ,   177 ,   121 ,   363 ,   2751    ,   4799    ,   14215   ,0 };
            const unsigned long dim1524JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   15  ,   27  ,   15  ,   101 ,   229 ,   411 ,   947 ,   189 ,   119 ,   6707    ,   5177    ,0 };
            const unsigned long dim1525JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   7   ,   29  ,   39  ,   65  ,   149 ,   185 ,   915 ,   889 ,   1651    ,   5977    ,   273 ,0 };
            const unsigned long dim1526JoeKuoD5Init[]   =   {   1   ,   1   ,   1   ,   7   ,   7   ,   45  ,   113 ,   203 ,   459 ,   747 ,   1577    ,   2247    ,   5005    ,   2375    ,0 };
            const unsigned long dim1527JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   9   ,   9   ,   5   ,   63  ,   83  ,   193 ,   363 ,   257 ,   4075    ,   7497    ,   4579    ,0 };
            const unsigned long dim1528JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   15  ,   1   ,   55  ,   107 ,   161 ,   471 ,   245 ,   1303    ,   1821    ,   3395    ,   8957    ,0 };
            const unsigned long dim1529JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   15  ,   5   ,   17  ,   83  ,   215 ,   467 ,   489 ,   827 ,   1951    ,   3753    ,   13333   ,0 };
            const unsigned long dim1530JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   11  ,   27  ,   31  ,   89  ,   55  ,   445 ,   489 ,   1171    ,   107 ,   1479    ,   1389    ,0 };
            const unsigned long dim1531JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   15  ,   21  ,   23  ,   51  ,   21  ,   129 ,   349 ,   195 ,   2177    ,   529 ,   5479    ,0 };
            const unsigned long dim1532JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   5   ,   15  ,   31  ,   99  ,   169 ,   299 ,   21  ,   1379    ,   3845    ,   4991    ,   8755    ,0 };
            const unsigned long dim1533JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   3   ,   23  ,   29  ,   47  ,   219 ,   289 ,   559 ,   393 ,   793 ,   3217    ,   12103   ,0 };
            const unsigned long dim1534JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   15  ,   5   ,   13  ,   107 ,   119 ,   159 ,   421 ,   243 ,   3231    ,   5331    ,   14511   ,0 };
            const unsigned long dim1535JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   13  ,   13  ,   9   ,   1   ,   99  ,   7   ,   743 ,   1125    ,   2969    ,   1205    ,   2963    ,0 };
            const unsigned long dim1536JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   15  ,   5   ,   1   ,   25  ,   101 ,   397 ,   75  ,   141 ,   3503    ,   3003    ,   11363   ,0 };
            const unsigned long dim1537JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   1   ,   21  ,   25  ,   55  ,   7   ,   189 ,   599 ,   1071    ,   343 ,   2877    ,   5131    ,0 };
            const unsigned long dim1538JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   7   ,   13  ,   25  ,   113 ,   175 ,   31  ,   265 ,   1901    ,   1779    ,   1787    ,   14551   ,0 };
            const unsigned long dim1539JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   13  ,   31  ,   53  ,   69  ,   49  ,   281 ,   57  ,   1865    ,   3211    ,   5545    ,   12597   ,0 };
            const unsigned long dim1540JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   11  ,   3   ,   51  ,   99  ,   23  ,   303 ,   455 ,   2021    ,   2903    ,   2521    ,   10211   ,0 };
            const unsigned long dim1541JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   3   ,   17  ,   49  ,   55  ,   147 ,   177 ,   515 ,   1333    ,   3357    ,   6483    ,   4599    ,0 };
            const unsigned long dim1542JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   1   ,   31  ,   3   ,   1   ,   15  ,   263 ,   1007    ,   1377    ,   1245    ,   313 ,   10227   ,0 };
            const unsigned long dim1543JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   3   ,   21  ,   7   ,   61  ,   201 ,   95  ,   407 ,   661 ,   2159    ,   3255    ,   5749    ,0 };
            const unsigned long dim1544JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   9   ,   13  ,   31  ,   65  ,   135 ,   117 ,   737 ,   1165    ,   2305    ,   5347    ,   2739    ,0 };
            const unsigned long dim1545JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   3   ,   25  ,   49  ,   63  ,   115 ,   65  ,   693 ,   1211    ,   763 ,   6949    ,   11655   ,0 };
            const unsigned long dim1546JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   5   ,   13  ,   37  ,   15  ,   61  ,   175 ,   377 ,   765 ,   269 ,   1363    ,   8199    ,0 };
            const unsigned long dim1547JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   7   ,   27  ,   27  ,   113 ,   131 ,   207 ,   921 ,   1051    ,   2205    ,   1197    ,   16233   ,0 };
            const unsigned long dim1548JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   7   ,   5   ,   19  ,   115 ,   95  ,   93  ,   87  ,   1379    ,   3295    ,   5211    ,   9113    ,0 };
            const unsigned long dim1549JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   11  ,   21  ,   19  ,   9   ,   39  ,   101 ,   5   ,   1903    ,   1067    ,   2845    ,   2025    ,0 };
            const unsigned long dim1550JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   9   ,   31  ,   5   ,   39  ,   23  ,   505 ,   563 ,   1445    ,   1177    ,   1485    ,   13551   ,0 };
            const unsigned long dim1551JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   3   ,   31  ,   57  ,   107 ,   241 ,   361 ,   353 ,   699 ,   171 ,   5211    ,   2235    ,0 };
            const unsigned long dim1552JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   5   ,   15  ,   7   ,   115 ,   119 ,   157 ,   293 ,   827 ,   3249    ,   3463    ,   11873   ,0 };
            const unsigned long dim1553JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   3   ,   11  ,   15  ,   17  ,   195 ,   257 ,   31  ,   1207    ,   549 ,   7807    ,   14135   ,0 };
            const unsigned long dim1554JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   7   ,   25  ,   59  ,   37  ,   205 ,   507 ,   855 ,   303 ,   683 ,   4277    ,   9387    ,0 };
            const unsigned long dim1555JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   1   ,   25  ,   1   ,   121 ,   29  ,   485 ,   707 ,   319 ,   3717    ,   2741    ,   5241    ,0 };
            const unsigned long dim1556JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   3   ,   15  ,   55  ,   107 ,   227 ,   129 ,   1011    ,   319 ,   713 ,   5263    ,   7865    ,0 };
            const unsigned long dim1557JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   5   ,   17  ,   11  ,   7   ,   107 ,   21  ,   349 ,   1101    ,   3279    ,   5541    ,   12485   ,0 };
            const unsigned long dim1558JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   13  ,   25  ,   9   ,   45  ,   149 ,   187 ,   229 ,   671 ,   1219    ,   5171    ,   2073    ,0 };
            const unsigned long dim1559JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   11  ,   17  ,   49  ,   23  ,   7   ,   219 ,   639 ,   1497    ,   3103    ,   3047    ,   7723    ,0 };
            const unsigned long dim1560JoeKuoD5Init[]   =   {   1   ,   1   ,   1   ,   5   ,   13  ,   5   ,   3   ,   249 ,   429 ,   289 ,   625 ,   325 ,   1257    ,   16251   ,0 };
            const unsigned long dim1561JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   1   ,   17  ,   21  ,   31  ,   11  ,   189 ,   73  ,   583 ,   2843    ,   1873    ,   1215    ,0 };
            const unsigned long dim1562JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   9   ,   25  ,   61  ,   123 ,   251 ,   485 ,   543 ,   1851    ,   2827    ,   397 ,   7313    ,0 };
            const unsigned long dim1563JoeKuoD5Init[]   =   {   1   ,   1   ,   1   ,   9   ,   29  ,   59  ,   103 ,   159 ,   295 ,   227 ,   1127    ,   1905    ,   4121    ,   3233    ,0 };
            const unsigned long dim1564JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   15  ,   17  ,   55  ,   7   ,   1   ,   381 ,   33  ,   39  ,   485 ,   3967    ,   4401    ,0 };
            const unsigned long dim1565JoeKuoD5Init[]   =   {   1   ,   1   ,   1   ,   3   ,   5   ,   29  ,   7   ,   95  ,   191 ,   23  ,   1205    ,   2427    ,   5439    ,   12585   ,0 };
            const unsigned long dim1566JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   9   ,   27  ,   9   ,   73  ,   195 ,   225 ,   597 ,   683 ,   3335    ,   6341    ,   10527   ,0 };
            const unsigned long dim1567JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   5   ,   13  ,   15  ,   83  ,   41  ,   241 ,   833 ,   1253    ,   3389    ,   2927    ,   2629    ,0 };
            const unsigned long dim1568JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   11  ,   25  ,   55  ,   23  ,   249 ,   151 ,   123 ,   667 ,   3835    ,   2215    ,   6189    ,0 };
            const unsigned long dim1569JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   13  ,   31  ,   43  ,   91  ,   181 ,   509 ,   453 ,   171 ,   2883    ,   1247    ,   5105    ,0 };
            const unsigned long dim1570JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   15  ,   3   ,   17  ,   59  ,   127 ,   173 ,   343 ,   901 ,   3419    ,   4755    ,   12367   ,0 };
            const unsigned long dim1571JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   9   ,   15  ,   49  ,   119 ,   1   ,   183 ,   1021    ,   561 ,   2405    ,   4093    ,   963 ,0 };
            const unsigned long dim1572JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   15  ,   21  ,   61  ,   63  ,   119 ,   11  ,   695 ,   1591    ,   2175    ,   5143    ,   10491   ,0 };
            const unsigned long dim1573JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   15  ,   27  ,   23  ,   91  ,   131 ,   23  ,   891 ,   1033    ,   1341    ,   1747    ,   12085   ,0 };
            const unsigned long dim1574JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   9   ,   11  ,   1   ,   35  ,   189 ,   131 ,   617 ,   417 ,   3347    ,   3995    ,   11723   ,0 };
            const unsigned long dim1575JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   3   ,   31  ,   35  ,   113 ,   83  ,   27  ,   399 ,   1615    ,   47  ,   1455    ,   4163    ,0 };
            const unsigned long dim1576JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   5   ,   29  ,   61  ,   103 ,   105 ,   495 ,   495 ,   1751    ,   2035    ,   7827    ,   11193   ,0 };
            const unsigned long dim1577JoeKuoD5Init[]   =   {   1   ,   1   ,   1   ,   11  ,   29  ,   35  ,   125 ,   193 ,   59  ,   15  ,   1319    ,   1169    ,   3789    ,   2003    ,0 };
            const unsigned long dim1578JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   9   ,   27  ,   9   ,   33  ,   197 ,   65  ,   191 ,   783 ,   3685    ,   7505    ,   13407   ,0 };
            const unsigned long dim1579JoeKuoD5Init[]   =   {   1   ,   1   ,   1   ,   1   ,   1   ,   35  ,   97  ,   55  ,   259 ,   477 ,   1835    ,   3083    ,   7879    ,   4701    ,0 };
            const unsigned long dim1580JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   11  ,   5   ,   31  ,   5   ,   125 ,   393 ,   317 ,   1577    ,   3741    ,   3823    ,   12447   ,0 };
            const unsigned long dim1581JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   13  ,   9   ,   59  ,   81  ,   85  ,   233 ,   465 ,   239 ,   1525    ,   3095    ,   5793    ,0 };
            const unsigned long dim1582JoeKuoD5Init[]   =   {   1   ,   1   ,   1   ,   1   ,   9   ,   49  ,   49  ,   239 ,   81  ,   475 ,   799 ,   2999    ,   2985    ,   11587   ,0 };
            const unsigned long dim1583JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   5   ,   27  ,   55  ,   43  ,   149 ,   191 ,   325 ,   2035    ,   1645    ,   2153    ,   13237   ,0 };
            const unsigned long dim1584JoeKuoD5Init[]   =   {   1   ,   1   ,   1   ,   15  ,   27  ,   59  ,   5   ,   49  ,   277 ,   75  ,   1759    ,   2753    ,   95  ,   2959    ,0 };
            const unsigned long dim1585JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   1   ,   21  ,   57  ,   119 ,   35  ,   457 ,   137 ,   1877    ,   2613    ,   4209    ,   9669    ,0 };
            const unsigned long dim1586JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   1   ,   25  ,   21  ,   51  ,   65  ,   155 ,   37  ,   783 ,   3427    ,   2763    ,   14361   ,0 };
            const unsigned long dim1587JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   7   ,   23  ,   11  ,   53  ,   181 ,   267 ,   285 ,   1927    ,   3591    ,   3735    ,   11471   ,0 };
            const unsigned long dim1588JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   1   ,   5   ,   25  ,   51  ,   31  ,   201 ,   851 ,   871 ,   3665    ,   193 ,   10929   ,0 };
            const unsigned long dim1589JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   5   ,   1   ,   25  ,   35  ,   133 ,   459 ,   45  ,   525 ,   3171    ,   1123    ,   5679    ,0 };
            const unsigned long dim1590JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   5   ,   29  ,   47  ,   123 ,   75  ,   101 ,   89  ,   1949    ,   1801    ,   3859    ,   6557    ,0 };
            const unsigned long dim1591JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   3   ,   25  ,   1   ,   93  ,   205 ,   447 ,   865 ,   1309    ,   3009    ,   945 ,   6961    ,0 };
            const unsigned long dim1592JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   11  ,   21  ,   47  ,   91  ,   9   ,   23  ,   607 ,   1905    ,   2291    ,   5315    ,   6673    ,0 };
            const unsigned long dim1593JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   13  ,   3   ,   33  ,   83  ,   27  ,   491 ,   467 ,   1819    ,   3295    ,   1589    ,   4771    ,0 };
            const unsigned long dim1594JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   15  ,   17  ,   57  ,   73  ,   49  ,   69  ,   821 ,   1773    ,   459 ,   7945    ,   14471   ,0 };
            const unsigned long dim1595JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   1   ,   19  ,   29  ,   9   ,   181 ,   311 ,   353 ,   2045    ,   2873    ,   7417    ,   15243   ,0 };
            const unsigned long dim1596JoeKuoD5Init[]   =   {   1   ,   1   ,   1   ,   7   ,   29  ,   45  ,   43  ,   17  ,   237 ,   915 ,   1069    ,   1429    ,   5629    ,   4501    ,0 };
            const unsigned long dim1597JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   9   ,   9   ,   53  ,   57  ,   55  ,   275 ,   233 ,   1697    ,   2857    ,   919 ,   4507    ,0 };
            const unsigned long dim1598JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   13  ,   17  ,   45  ,   81  ,   101 ,   209 ,   769 ,   783 ,   1949    ,   5933    ,   137 ,0 };
            const unsigned long dim1599JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   7   ,   1   ,   33  ,   17  ,   87  ,   35  ,   927 ,   1595    ,   2443    ,   285 ,   12821   ,0 };
            const unsigned long dim1600JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   9   ,   3   ,   33  ,   113 ,   243 ,   51  ,   203 ,   1683    ,   389 ,   2789    ,   9255    ,0 };
            const unsigned long dim1601JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   3   ,   9   ,   1   ,   7   ,   227 ,   27  ,   773 ,   621 ,   3743    ,   6591    ,   407 ,0 };
            const unsigned long dim1602JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   5   ,   29  ,   51  ,   113 ,   99  ,   305 ,   267 ,   907 ,   3861    ,   5335    ,   10851   ,0 };
            const unsigned long dim1603JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   9   ,   27  ,   63  ,   65  ,   249 ,   87  ,   105 ,   2023    ,   1383    ,   4267    ,   8995    ,0 };
            const unsigned long dim1604JoeKuoD5Init[]   =   {   1   ,   1   ,   1   ,   13  ,   15  ,   11  ,   77  ,   9   ,   365 ,   711 ,   1367    ,   2101    ,   5833    ,   9799    ,0 };
            const unsigned long dim1605JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   3   ,   7   ,   29  ,   13  ,   157 ,   259 ,   455 ,   389 ,   3177    ,   4243    ,   7615    ,0 };
            const unsigned long dim1606JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   3   ,   25  ,   27  ,   31  ,   91  ,   195 ,   109 ,   1767    ,   987 ,   2715    ,   7613    ,0 };
            const unsigned long dim1607JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   7   ,   23  ,   51  ,   89  ,   127 ,   485 ,   361 ,   1555    ,   441 ,   4963    ,   3371    ,0 };
            const unsigned long dim1608JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   7   ,   27  ,   31  ,   25  ,   249 ,   35  ,   115 ,   1021    ,   1051    ,   3449    ,   3395    ,0 };
            const unsigned long dim1609JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   3   ,   17  ,   41  ,   117 ,   173 ,   491 ,   281 ,   885 ,   471 ,   6665    ,   10041   ,0 };
            const unsigned long dim1610JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   3   ,   7   ,   53  ,   117 ,   117 ,   479 ,   449 ,   569 ,   4049    ,   2747    ,   12963   ,0 };
            const unsigned long dim1611JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   7   ,   11  ,   13  ,   107 ,   171 ,   7   ,   547 ,   1635    ,   1697    ,   1005    ,   11137   ,0 };
            const unsigned long dim1612JoeKuoD5Init[]   =   {   1   ,   1   ,   1   ,   13  ,   17  ,   51  ,   109 ,   81  ,   111 ,   395 ,   349 ,   1467    ,   1399    ,   15545   ,0 };
            const unsigned long dim1613JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   9   ,   13  ,   47  ,   79  ,   119 ,   31  ,   239 ,   41  ,   2043    ,   2849    ,   16079   ,0 };
            const unsigned long dim1614JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   1   ,   17  ,   41  ,   95  ,   193 ,   511 ,   879 ,   1223    ,   3133    ,   1675    ,   3929    ,0 };
            const unsigned long dim1615JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   3   ,   15  ,   55  ,   21  ,   143 ,   279 ,   11  ,   717 ,   3021    ,   6207    ,   4499    ,0 };
            const unsigned long dim1616JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   3   ,   3   ,   25  ,   37  ,   91  ,   227 ,   619 ,   1873    ,   1991    ,   793 ,   10021   ,0 };
            const unsigned long dim1617JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   11  ,   17  ,   39  ,   75  ,   191 ,   117 ,   997 ,   735 ,   3771    ,   4243    ,   2491    ,0 };
            const unsigned long dim1618JoeKuoD5Init[]   =   {   1   ,   1   ,   1   ,   7   ,   19  ,   9   ,   71  ,   89  ,   427 ,   791 ,   623 ,   903 ,   6685    ,   9029    ,0 };
            const unsigned long dim1619JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   3   ,   13  ,   49  ,   13  ,   151 ,   389 ,   677 ,   727 ,   3135    ,   1029    ,   12669   ,0 };
            const unsigned long dim1620JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   11  ,   27  ,   31  ,   119 ,   219 ,   491 ,   819 ,   755 ,   3529    ,   3071    ,   16095   ,0 };
            const unsigned long dim1621JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   9   ,   1   ,   57  ,   127 ,   171 ,   481 ,   467 ,   1131    ,   1481    ,   2491    ,   2717    ,0 };
            const unsigned long dim1622JoeKuoD5Init[]   =   {   1   ,   1   ,   1   ,   15  ,   7   ,   25  ,   55  ,   87  ,   225 ,   363 ,   843 ,   3581    ,   2511    ,   6685    ,0 };
            const unsigned long dim1623JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   3   ,   29  ,   21  ,   91  ,   251 ,   403 ,   1007    ,   307 ,   2869    ,   6033    ,   14169   ,0 };
            const unsigned long dim1624JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   13  ,   31  ,   43  ,   77  ,   71  ,   101 ,   995 ,   625 ,   2763    ,   7537    ,   1213    ,0 };
            const unsigned long dim1625JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   5   ,   9   ,   45  ,   11  ,   235 ,   59  ,   561 ,   547 ,   815 ,   6123    ,   8173    ,0 };
            const unsigned long dim1626JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   3   ,   25  ,   19  ,   107 ,   47  ,   103 ,   467 ,   1889    ,   2021    ,   2861    ,   7617    ,0 };
            const unsigned long dim1627JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   9   ,   9   ,   3   ,   11  ,   231 ,   217 ,   213 ,   1497    ,   3125    ,   7421    ,   6221    ,0 };
            const unsigned long dim1628JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   7   ,   27  ,   57  ,   27  ,   9   ,   507 ,   191 ,   1297    ,   3307    ,   4687    ,   13299   ,0 };
            const unsigned long dim1629JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   1   ,   25  ,   17  ,   109 ,   57  ,   231 ,   123 ,   1297    ,   77  ,   785 ,   13731   ,0 };
            const unsigned long dim1630JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   3   ,   25  ,   5   ,   85  ,   177 ,   365 ,   301 ,   655 ,   1743    ,   2009    ,   7759    ,0 };
            const unsigned long dim1631JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   7   ,   19  ,   29  ,   105 ,   57  ,   469 ,   735 ,   875 ,   3749    ,   619 ,   15569   ,0 };
            const unsigned long dim1632JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   1   ,   19  ,   21  ,   105 ,   19  ,   83  ,   117 ,   1599    ,   655 ,   63  ,   12143   ,0 };
            const unsigned long dim1633JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   7   ,   9   ,   35  ,   123 ,   195 ,   109 ,   321 ,   1713    ,   793 ,   8067    ,   6903    ,0 };
            const unsigned long dim1634JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   9   ,   1   ,   41  ,   43  ,   41  ,   51  ,   711 ,   1713    ,   3063    ,   6427    ,   3577    ,0 };
            const unsigned long dim1635JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   11  ,   17  ,   3   ,   95  ,   159 ,   23  ,   605 ,   1431    ,   1289    ,   2225    ,   1689    ,0 };
            const unsigned long dim1636JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   15  ,   19  ,   3   ,   83  ,   61  ,   133 ,   707 ,   453 ,   2487    ,   551 ,   13605   ,0 };
            const unsigned long dim1637JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   5   ,   31  ,   37  ,   51  ,   157 ,   51  ,   423 ,   2021    ,   1837    ,   1873    ,   2919    ,0 };
            const unsigned long dim1638JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   13  ,   3   ,   17  ,   19  ,   19  ,   181 ,   961 ,   669 ,   47  ,   7513    ,   7551    ,0 };
            const unsigned long dim1639JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   1   ,   9   ,   7   ,   61  ,   219 ,   347 ,   21  ,   467 ,   955 ,   3255    ,   275 ,0 };
            const unsigned long dim1640JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   1   ,   31  ,   57  ,   51  ,   9   ,   49  ,   491 ,   119 ,   1155    ,   3641    ,   16095   ,0 };
            const unsigned long dim1641JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   1   ,   15  ,   37  ,   93  ,   27  ,   205 ,   889 ,   1463    ,   1567    ,   453 ,   13757   ,0 };
            const unsigned long dim1642JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   15  ,   27  ,   51  ,   43  ,   189 ,   357 ,   591 ,   1783    ,   549 ,   761 ,   8683    ,0 };
            const unsigned long dim1643JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   15  ,   13  ,   27  ,   25  ,   67  ,   145 ,   471 ,   1589    ,   2395    ,   6625    ,   3837    ,0 };
            const unsigned long dim1644JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   15  ,   25  ,   61  ,   111 ,   125 ,   475 ,   489 ,   15  ,   3835    ,   5077    ,   14487   ,0 };
            const unsigned long dim1645JoeKuoD5Init[]   =   {   1   ,   1   ,   1   ,   9   ,   15  ,   55  ,   35  ,   169 ,   267 ,   135 ,   383 ,   733 ,   6913    ,   14153   ,0 };
            const unsigned long dim1646JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   1   ,   19  ,   55  ,   89  ,   61  ,   175 ,   467 ,   1243    ,   1431    ,   1743    ,   8641    ,0 };
            const unsigned long dim1647JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   7   ,   19  ,   41  ,   13  ,   45  ,   503 ,   285 ,   1727    ,   587 ,   5073    ,   13053   ,0 };
            const unsigned long dim1648JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   3   ,   11  ,   39  ,   93  ,   167 ,   385 ,   165 ,   881 ,   1037    ,   1471    ,   14527   ,0 };
            const unsigned long dim1649JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   15  ,   9   ,   7   ,   49  ,   163 ,   219 ,   955 ,   1083    ,   2723    ,   3749    ,   15415   ,0 };
            const unsigned long dim1650JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   3   ,   7   ,   43  ,   67  ,   87  ,   375 ,   405 ,   1663    ,   1263    ,   1895    ,   2229    ,0 };
            const unsigned long dim1651JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   1   ,   27  ,   47  ,   77  ,   147 ,   417 ,   843 ,   1829    ,   3451    ,   2973    ,   7313    ,0 };
            const unsigned long dim1652JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   1   ,   3   ,   25  ,   69  ,   255 ,   413 ,   43  ,   837 ,   957 ,   261 ,   8663    ,0 };
            const unsigned long dim1653JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   7   ,   29  ,   35  ,   9   ,   57  ,   421 ,   751 ,   593 ,   3745    ,   3203    ,   14179   ,0 };
            const unsigned long dim1654JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   3   ,   13  ,   25  ,   109 ,   169 ,   331 ,   537 ,   69  ,   2089    ,   7263    ,   15133   ,0 };
            const unsigned long dim1655JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   5   ,   21  ,   21  ,   21  ,   145 ,   223 ,   167 ,   1297    ,   3257    ,   7465    ,   1557    ,0 };
            const unsigned long dim1656JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   3   ,   11  ,   7   ,   31  ,   205 ,   97  ,   737 ,   79  ,   4083    ,   5601    ,   8411    ,0 };
            const unsigned long dim1657JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   3   ,   7   ,   39  ,   117 ,   203 ,   225 ,   801 ,   741 ,   1861    ,   8091    ,   3169    ,0 };
            const unsigned long dim1658JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   3   ,   15  ,   49  ,   109 ,   171 ,   109 ,   845 ,   1757    ,   2687    ,   5643    ,   3967    ,0 };
            const unsigned long dim1659JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   13  ,   7   ,   41  ,   93  ,   165 ,   127 ,   629 ,   959 ,   1483    ,   3837    ,   2093    ,0 };
            const unsigned long dim1660JoeKuoD5Init[]   =   {   1   ,   1   ,   1   ,   5   ,   9   ,   39  ,   115 ,   153 ,   395 ,   777 ,   1601    ,   3443    ,   3581    ,   1419    ,0 };
            const unsigned long dim1661JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   3   ,   23  ,   61  ,   87  ,   179 ,   463 ,   55  ,   1727    ,   99  ,   7527    ,   15281   ,0 };
            const unsigned long dim1662JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   9   ,   9   ,   49  ,   27  ,   103 ,   221 ,   277 ,   1093    ,   3547    ,   8009    ,   8711    ,0 };
            const unsigned long dim1663JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   7   ,   15  ,   59  ,   79  ,   59  ,   451 ,   763 ,   1687    ,   389 ,   1665    ,   12149   ,0 };
            const unsigned long dim1664JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   9   ,   25  ,   25  ,   37  ,   53  ,   173 ,   1003    ,   1175    ,   3881    ,   4355    ,   6247    ,0 };
            const unsigned long dim1665JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   3   ,   15  ,   31  ,   123 ,   133 ,   79  ,   581 ,   405 ,   2869    ,   2759    ,   2295    ,0 };
            const unsigned long dim1666JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   13  ,   19  ,   3   ,   35  ,   135 ,   287 ,   433 ,   205 ,   2119    ,   6293    ,   2931    ,0 };
            const unsigned long dim1667JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   13  ,   21  ,   29  ,   69  ,   109 ,   401 ,   753 ,   1371    ,   3777    ,   5473    ,   8357    ,0 };
            const unsigned long dim1668JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   7   ,   11  ,   27  ,   101 ,   153 ,   483 ,   639 ,   687 ,   2325    ,   329 ,   12427   ,0 };
            const unsigned long dim1669JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   13  ,   31  ,   41  ,   93  ,   83  ,   121 ,   645 ,   479 ,   1417    ,   1967    ,   2807    ,0 };
            const unsigned long dim1670JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   5   ,   3   ,   29  ,   41  ,   115 ,   41  ,   81  ,   5   ,   1063    ,   943 ,   10151   ,0 };
            const unsigned long dim1671JoeKuoD5Init[]   =   {   1   ,   1   ,   1   ,   13  ,   5   ,   33  ,   79  ,   245 ,   271 ,   373 ,   1339    ,   1471    ,   4695    ,   12791   ,0 };
            const unsigned long dim1672JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   3   ,   5   ,   5   ,   19  ,   101 ,   109 ,   429 ,   1587    ,   3453    ,   4549    ,   8173    ,0 };
            const unsigned long dim1673JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   7   ,   5   ,   33  ,   5   ,   15  ,   183 ,   241 ,   871 ,   1263    ,   85  ,   13525   ,0 };
            const unsigned long dim1674JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   1   ,   15  ,   5   ,   115 ,   165 ,   451 ,   789 ,   515 ,   3359    ,   1231    ,   3129    ,0 };
            const unsigned long dim1675JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   3   ,   5   ,   45  ,   109 ,   233 ,   369 ,   819 ,   987 ,   3157    ,   5961    ,   5705    ,0 };
            const unsigned long dim1676JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   11  ,   7   ,   63  ,   43  ,   249 ,   95  ,   853 ,   917 ,   1749    ,   4275    ,   10109   ,0 };
            const unsigned long dim1677JoeKuoD5Init[]   =   {   1   ,   1   ,   1   ,   9   ,   21  ,   5   ,   53  ,   249 ,   3   ,   11  ,   1171    ,   3289    ,   6659    ,   6555    ,0 };
            const unsigned long dim1678JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   11  ,   9   ,   19  ,   35  ,   253 ,   13  ,   229 ,   149 ,   3153    ,   3833    ,   11635   ,0 };
            const unsigned long dim1679JoeKuoD5Init[]   =   {   1   ,   1   ,   1   ,   11  ,   1   ,   19  ,   127 ,   119 ,   405 ,   19  ,   589 ,   2613    ,   399 ,   3869    ,0 };
            const unsigned long dim1680JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   1   ,   3   ,   11  ,   35  ,   147 ,   11  ,   765 ,   237 ,   2451    ,   5041    ,   4919    ,0 };
            const unsigned long dim1681JoeKuoD5Init[]   =   {   1   ,   1   ,   1   ,   11  ,   15  ,   7   ,   91  ,   7   ,   5   ,   223 ,   1505    ,   3513    ,   185 ,   2767    ,0 };
            const unsigned long dim1682JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   7   ,   27  ,   17  ,   81  ,   181 ,   11  ,   803 ,   319 ,   3891    ,   4505    ,   6035    ,0 };
            const unsigned long dim1683JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   15  ,   29  ,   13  ,   47  ,   137 ,   301 ,   975 ,   77  ,   3351    ,   6307    ,   2613    ,0 };
            const unsigned long dim1684JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   5   ,   15  ,   41  ,   99  ,   171 ,   465 ,   145 ,   1859    ,   2949    ,   7915    ,   7755    ,0 };
            const unsigned long dim1685JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   5   ,   3   ,   25  ,   123 ,   251 ,   423 ,   921 ,   987 ,   793 ,   2199    ,   1255    ,0 };
            const unsigned long dim1686JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   3   ,   1   ,   63  ,   75  ,   101 ,   219 ,   1013    ,   1761    ,   2171    ,   2763    ,   11185   ,0 };
            const unsigned long dim1687JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   13  ,   13  ,   15  ,   33  ,   147 ,   5   ,   915 ,   1903    ,   2607    ,   3847    ,   167 ,0 };
            const unsigned long dim1688JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   1   ,   23  ,   27  ,   71  ,   17  ,   381 ,   269 ,   603 ,   2303    ,   1399    ,   13795   ,0 };
            const unsigned long dim1689JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   5   ,   29  ,   5   ,   51  ,   151 ,   271 ,   369 ,   595 ,   531 ,   7155    ,   15871   ,0 };
            const unsigned long dim1690JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   13  ,   9   ,   49  ,   49  ,   119 ,   191 ,   105 ,   1203    ,   3431    ,   7063    ,   10831   ,0 };
            const unsigned long dim1691JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   15  ,   31  ,   49  ,   37  ,   147 ,   155 ,   865 ,   339 ,   257 ,   4065    ,   7249    ,0 };
            const unsigned long dim1692JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   11  ,   3   ,   23  ,   127 ,   19  ,   363 ,   733 ,   1059    ,   3693    ,   4623    ,   2853    ,0 };
            const unsigned long dim1693JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   3   ,   3   ,   55  ,   13  ,   177 ,   453 ,   447 ,   183 ,   3247    ,   5923    ,   15485   ,0 };
            const unsigned long dim1694JoeKuoD5Init[]   =   {   1   ,   1   ,   1   ,   7   ,   3   ,   57  ,   93  ,   13  ,   347 ,   225 ,   535 ,   3187    ,   6047    ,   11315   ,0 };
            const unsigned long dim1695JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   9   ,   17  ,   5   ,   121 ,   35  ,   35  ,   393 ,   1941    ,   1901    ,   7099    ,   2639    ,0 };
            const unsigned long dim1696JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   13  ,   13  ,   43  ,   61  ,   143 ,   291 ,   667 ,   103 ,   3679    ,   7899    ,   3603    ,0 };
            const unsigned long dim1697JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   15  ,   31  ,   57  ,   41  ,   101 ,   455 ,   541 ,   1001    ,   1879    ,   1879    ,   7411    ,0 };
            const unsigned long dim1698JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   7   ,   13  ,   29  ,   113 ,   151 ,   305 ,   593 ,   147 ,   737 ,   6643    ,   14463   ,0 };
            const unsigned long dim1699JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   3   ,   17  ,   55  ,   3   ,   179 ,   277 ,   873 ,   1915    ,   2541    ,   4245    ,   4357    ,0 };
            const unsigned long dim1700JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   11  ,   21  ,   43  ,   97  ,   251 ,   15  ,   443 ,   957 ,   923 ,   7497    ,   9377    ,0 };
            const unsigned long dim1701JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   7   ,   23  ,   19  ,   27  ,   3   ,   35  ,   407 ,   451 ,   3653    ,   813 ,   8833    ,0 };
            const unsigned long dim1702JoeKuoD5Init[]   =   {   1   ,   1   ,   1   ,   5   ,   15  ,   25  ,   5   ,   165 ,   231 ,   221 ,   1325    ,   641 ,   4545    ,   7667    ,0 };
            const unsigned long dim1703JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   1   ,   13  ,   17  ,   25  ,   225 ,   237 ,   289 ,   187 ,   2881    ,   4723    ,   15579   ,0 };
            const unsigned long dim1704JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   3   ,   15  ,   43  ,   53  ,   241 ,   163 ,   853 ,   491 ,   2497    ,   761 ,   1707    ,0 };
            const unsigned long dim1705JoeKuoD5Init[]   =   {   1   ,   1   ,   1   ,   3   ,   13  ,   19  ,   83  ,   97  ,   7   ,   639 ,   1985    ,   3553    ,   3971    ,   10661   ,0 };
            const unsigned long dim1706JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   1   ,   15  ,   37  ,   21  ,   139 ,   33  ,   643 ,   63  ,   2213    ,   1807    ,   14483   ,0 };
            const unsigned long dim1707JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   11  ,   3   ,   55  ,   67  ,   45  ,   313 ,   539 ,   1057    ,   455 ,   6473    ,   1499    ,0 };
            const unsigned long dim1708JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   9   ,   1   ,   29  ,   91  ,   133 ,   461 ,   301 ,   539 ,   2001    ,   8189    ,   2009    ,0 };
            const unsigned long dim1709JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   9   ,   17  ,   5   ,   77  ,   221 ,   423 ,   505 ,   419 ,   1987    ,   8171    ,   12985   ,0 };
            const unsigned long dim1710JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   13  ,   21  ,   11  ,   69  ,   53  ,   43  ,   819 ,   1513    ,   1767    ,   2981    ,   4333    ,0 };
            const unsigned long dim1711JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   9   ,   9   ,   29  ,   83  ,   183 ,   9   ,   815 ,   1051    ,   819 ,   2089    ,   13917   ,0 };
            const unsigned long dim1712JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   1   ,   21  ,   41  ,   87  ,   133 ,   149 ,   847 ,   205 ,   1511    ,   2441    ,   13537   ,0 };
            const unsigned long dim1713JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   3   ,   23  ,   1   ,   121 ,   89  ,   349 ,   355 ,   1919    ,   2045    ,   1723    ,   11859   ,0 };
            const unsigned long dim1714JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   1   ,   21  ,   47  ,   111 ,   65  ,   367 ,   505 ,   805 ,   2823    ,   5807    ,   6221    ,0 };
            const unsigned long dim1715JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   7   ,   7   ,   39  ,   35  ,   35  ,   497 ,   411 ,   387 ,   203 ,   1513    ,   15385   ,0 };
            const unsigned long dim1716JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   7   ,   3   ,   7   ,   23  ,   195 ,   353 ,   241 ,   327 ,   1041    ,   4667    ,   4333    ,0 };
            const unsigned long dim1717JoeKuoD5Init[]   =   {   1   ,   1   ,   1   ,   9   ,   31  ,   51  ,   45  ,   243 ,   485 ,   785 ,   419 ,   1107    ,   7691    ,   2303    ,0 };
            const unsigned long dim1718JoeKuoD5Init[]   =   {   1   ,   1   ,   1   ,   7   ,   31  ,   1   ,   121 ,   79  ,   263 ,   153 ,   695 ,   3617    ,   7435    ,   11587   ,0 };
            const unsigned long dim1719JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   7   ,   27  ,   33  ,   43  ,   223 ,   153 ,   755 ,   1151    ,   3343    ,   2795    ,   7781    ,0 };
            const unsigned long dim1720JoeKuoD5Init[]   =   {   1   ,   1   ,   1   ,   15  ,   5   ,   41  ,   39  ,   181 ,   485 ,   129 ,   529 ,   2335    ,   6843    ,   7733    ,0 };
            const unsigned long dim1721JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   11  ,   31  ,   55  ,   117 ,   157 ,   227 ,   125 ,   557 ,   315 ,   3031    ,   8671    ,0 };
            const unsigned long dim1722JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   7   ,   15  ,   5   ,   117 ,   149 ,   239 ,   85  ,   341 ,   995 ,   1709    ,   9303    ,0 };
            const unsigned long dim1723JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   9   ,   3   ,   23  ,   17  ,   87  ,   327 ,   929 ,   519 ,   3441    ,   7599    ,   15021   ,0 };
            const unsigned long dim1724JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   9   ,   23  ,   31  ,   81  ,   37  ,   493 ,   451 ,   603 ,   1943    ,   1055    ,   3959    ,0 };
            const unsigned long dim1725JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   13  ,   7   ,   61  ,   61  ,   127 ,   303 ,   675 ,   955 ,   249 ,   7653    ,   8441    ,0 };
            const unsigned long dim1726JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   5   ,   23  ,   1   ,   95  ,   109 ,   155 ,   853 ,   1567    ,   4007    ,   4205    ,   7839    ,0 };
            const unsigned long dim1727JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   1   ,   11  ,   27  ,   57  ,   167 ,   285 ,   421 ,   143 ,   3937    ,   4865    ,   10581   ,0 };
            const unsigned long dim1728JoeKuoD5Init[]   =   {   1   ,   1   ,   1   ,   11  ,   15  ,   19  ,   49  ,   39  ,   383 ,   549 ,   1563    ,   2499    ,   7889    ,   239 ,0 };
            const unsigned long dim1729JoeKuoD5Init[]   =   {   1   ,   1   ,   1   ,   7   ,   13  ,   47  ,   61  ,   77  ,   443 ,   961 ,   1979    ,   931 ,   433 ,   6457    ,0 };
            const unsigned long dim1730JoeKuoD5Init[]   =   {   1   ,   1   ,   1   ,   5   ,   5   ,   1   ,   49  ,   137 ,   417 ,   579 ,   1079    ,   1511    ,   1611    ,   16083   ,0 };
            const unsigned long dim1731JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   11  ,   7   ,   51  ,   43  ,   29  ,   199 ,   525 ,   801 ,   3887    ,   619 ,   3389    ,0 };
            const unsigned long dim1732JoeKuoD5Init[]   =   {   1   ,   1   ,   1   ,   11  ,   25  ,   23  ,   65  ,   65  ,   201 ,   875 ,   787 ,   3747    ,   7275    ,   6191    ,0 };
            const unsigned long dim1733JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   1   ,   19  ,   21  ,   117 ,   169 ,   39  ,   901 ,   3   ,   3579    ,   6119    ,   2057    ,0 };
            const unsigned long dim1734JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   9   ,   13  ,   59  ,   33  ,   43  ,   237 ,   767 ,   819 ,   3555    ,   1337    ,   13469   ,0 };
            const unsigned long dim1735JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   1   ,   25  ,   51  ,   59  ,   11  ,   427 ,   353 ,   1601    ,   3905    ,   6975    ,   1065    ,0 };
            const unsigned long dim1736JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   5   ,   19  ,   35  ,   43  ,   137 ,   227 ,   17  ,   803 ,   919 ,   6651    ,   3339    ,0 };
            const unsigned long dim1737JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   15  ,   21  ,   21  ,   41  ,   197 ,   373 ,   849 ,   1753    ,   515 ,   4093    ,   15407   ,0 };
            const unsigned long dim1738JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   3   ,   11  ,   5   ,   49  ,   223 ,   351 ,   349 ,   987 ,   1785    ,   269 ,   3037    ,0 };
            const unsigned long dim1739JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   9   ,   3   ,   47  ,   103 ,   225 ,   21  ,   285 ,   1529    ,   399 ,   4951    ,   10767   ,0 };
            const unsigned long dim1740JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   11  ,   25  ,   27  ,   97  ,   87  ,   487 ,   613 ,   607 ,   1905    ,   6019    ,   423 ,0 };
            const unsigned long dim1741JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   3   ,   23  ,   53  ,   23  ,   147 ,   69  ,   781 ,   373 ,   1261    ,   8011    ,   9611    ,0 };
            const unsigned long dim1742JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   11  ,   21  ,   43  ,   67  ,   87  ,   173 ,   57  ,   1147    ,   1841    ,   6031    ,   11261   ,0 };
            const unsigned long dim1743JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   7   ,   29  ,   23  ,   15  ,   113 ,   485 ,   699 ,   259 ,   1175    ,   7489    ,   5119    ,0 };
            const unsigned long dim1744JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   13  ,   15  ,   13  ,   85  ,   219 ,   107 ,   811 ,   1599    ,   2267    ,   8047    ,   12427   ,0 };
            const unsigned long dim1745JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   11  ,   9   ,   7   ,   127 ,   201 ,   275 ,   293 ,   1313    ,   3251    ,   3745    ,   15237   ,0 };
            const unsigned long dim1746JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   15  ,   1   ,   3   ,   57  ,   235 ,   509 ,   353 ,   513 ,   467 ,   1409    ,   4733    ,0 };
            const unsigned long dim1747JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   7   ,   13  ,   43  ,   9   ,   41  ,   39  ,   919 ,   1545    ,   43  ,   8029    ,   1413    ,0 };
            const unsigned long dim1748JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   3   ,   21  ,   7   ,   33  ,   41  ,   161 ,   53  ,   1635    ,   787 ,   6197    ,   4841    ,0 };
            const unsigned long dim1749JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   9   ,   29  ,   41  ,   19  ,   161 ,   205 ,   1003    ,   1899    ,   7   ,   4329    ,   11151   ,0 };
            const unsigned long dim1750JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   7   ,   23  ,   23  ,   27  ,   145 ,   375 ,   433 ,   1729    ,   3787    ,   4985    ,   2167    ,0 };
            const unsigned long dim1751JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   5   ,   15  ,   21  ,   77  ,   13  ,   33  ,   227 ,   837 ,   1373    ,   2745    ,   2339    ,0 };
            const unsigned long dim1752JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   13  ,   17  ,   57  ,   31  ,   83  ,   135 ,   535 ,   693 ,   803 ,   1459    ,   319 ,0 };
            const unsigned long dim1753JoeKuoD5Init[]   =   {   1   ,   1   ,   1   ,   15  ,   11  ,   39  ,   105 ,   111 ,   247 ,   901 ,   395 ,   573 ,   3359    ,   8955    ,0 };
            const unsigned long dim1754JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   5   ,   23  ,   27  ,   99  ,   83  ,   277 ,   229 ,   429 ,   1451    ,   4755    ,   13951   ,0 };
            const unsigned long dim1755JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   15  ,   27  ,   17  ,   37  ,   81  ,   83  ,   991 ,   1509    ,   1931    ,   7389    ,   6053    ,0 };
            const unsigned long dim1756JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   13  ,   27  ,   55  ,   117 ,   93  ,   55  ,   65  ,   155 ,   3933    ,   3159    ,   9507    ,0 };
            const unsigned long dim1757JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   7   ,   15  ,   47  ,   109 ,   79  ,   379 ,   175 ,   419 ,   3285    ,   3207    ,   1675    ,0 };
            const unsigned long dim1758JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   15  ,   31  ,   19  ,   123 ,   229 ,   151 ,   549 ,   1329    ,   3835    ,   511 ,   14679   ,0 };
            const unsigned long dim1759JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   5   ,   25  ,   43  ,   89  ,   145 ,   419 ,   5   ,   1131    ,   2143    ,   7473    ,   14101   ,0 };
            const unsigned long dim1760JoeKuoD5Init[]   =   {   1   ,   1   ,   1   ,   11  ,   21  ,   43  ,   7   ,   123 ,   15  ,   743 ,   241 ,   1737    ,   2239    ,   11007   ,0 };
            const unsigned long dim1761JoeKuoD5Init[]   =   {   1   ,   1   ,   1   ,   15  ,   27  ,   3   ,   119 ,   157 ,   361 ,   471 ,   1673    ,   1873    ,   4555    ,   16337   ,0 };
            const unsigned long dim1762JoeKuoD5Init[]   =   {   1   ,   1   ,   1   ,   7   ,   15  ,   39  ,   95  ,   189 ,   353 ,   605 ,   349 ,   2763    ,   5125    ,   7943    ,0 };
            const unsigned long dim1763JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   13  ,   13  ,   1   ,   45  ,   69  ,   191 ,   183 ,   1967    ,   2455    ,   3879    ,   2397    ,0 };
            const unsigned long dim1764JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   1   ,   3   ,   23  ,   109 ,   123 ,   127 ,   813 ,   1499    ,   39  ,   1991    ,   9767    ,0 };
            const unsigned long dim1765JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   1   ,   7   ,   3   ,   125 ,   93  ,   279 ,   803 ,   1203    ,   3623    ,   4359    ,   4251    ,0 };
            const unsigned long dim1766JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   15  ,   25  ,   57  ,   61  ,   103 ,   219 ,   983 ,   957 ,   895 ,   4077    ,   11799   ,0 };
            const unsigned long dim1767JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   9   ,   21  ,   11  ,   15  ,   137 ,   491 ,   643 ,   1737    ,   3459    ,   4367    ,   5727    ,0 };
            const unsigned long dim1768JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   9   ,   25  ,   57  ,   111 ,   129 ,   101 ,   807 ,   1481    ,   3703    ,   2713    ,   6375    ,0 };
            const unsigned long dim1769JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   1   ,   31  ,   59  ,   75  ,   173 ,   209 ,   273 ,   121 ,   747 ,   257 ,   9713    ,0 };
            const unsigned long dim1770JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   11  ,   5   ,   31  ,   107 ,   243 ,   189 ,   407 ,   773 ,   503 ,   6197    ,   9455    ,0 };
            const unsigned long dim1771JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   5   ,   29  ,   25  ,   31  ,   169 ,   393 ,   857 ,   1739    ,   883 ,   2147    ,   15569   ,0 };
            const unsigned long dim1772JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   11  ,   31  ,   47  ,   17  ,   119 ,   499 ,   595 ,   207 ,   1709    ,   4585    ,   12855   ,0 };
            const unsigned long dim1773JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   3   ,   21  ,   33  ,   55  ,   197 ,   349 ,   41  ,   453 ,   1913    ,   2301    ,   6461    ,0 };
            const unsigned long dim1774JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   13  ,   27  ,   37  ,   105 ,   217 ,   145 ,   1007    ,   259 ,   2681    ,   477 ,   15931   ,0 };
            const unsigned long dim1775JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   13  ,   3   ,   39  ,   9   ,   231 ,   395 ,   735 ,   501 ,   1631    ,   2931    ,   11947   ,0 };
            const unsigned long dim1776JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   11  ,   13  ,   61  ,   113 ,   113 ,   473 ,   591 ,   499 ,   2169    ,   6419    ,   10619   ,0 };
            const unsigned long dim1777JoeKuoD5Init[]   =   {   1   ,   1   ,   1   ,   5   ,   29  ,   63  ,   43  ,   127 ,   243 ,   647 ,   1633    ,   1361    ,   3755    ,   11315   ,0 };
            const unsigned long dim1778JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   1   ,   13  ,   31  ,   111 ,   63  ,   69  ,   1005    ,   1955    ,   339 ,   2415    ,   4855    ,0 };
            const unsigned long dim1779JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   13  ,   25  ,   11  ,   21  ,   1   ,   343 ,   259 ,   1359    ,   597 ,   7029    ,   16229   ,0 };
            const unsigned long dim1780JoeKuoD5Init[]   =   {   1   ,   1   ,   1   ,   7   ,   27  ,   13  ,   17  ,   43  ,   509 ,   105 ,   347 ,   443 ,   5939    ,   12173   ,0 };
            const unsigned long dim1781JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   13  ,   15  ,   25  ,   93  ,   199 ,   305 ,   725 ,   597 ,   1497    ,   313 ,   10677   ,0 };
            const unsigned long dim1782JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   15  ,   17  ,   7   ,   103 ,   95  ,   83  ,   433 ,   441 ,   2587    ,   3365    ,   14771   ,0 };
            const unsigned long dim1783JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   1   ,   17  ,   7   ,   55  ,   189 ,   451 ,   729 ,   817 ,   243 ,   4089    ,   6569    ,0 };
            const unsigned long dim1784JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   1   ,   5   ,   7   ,   19  ,   113 ,   191 ,   211 ,   1205    ,   4005    ,   3221    ,   7521    ,0 };
            const unsigned long dim1785JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   15  ,   25  ,   49  ,   77  ,   187 ,   105 ,   87  ,   955 ,   2381    ,   1243    ,   11335   ,0 };
            const unsigned long dim1786JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   1   ,   19  ,   63  ,   113 ,   3   ,   7   ,   105 ,   185 ,   1499    ,   6885    ,   9063    ,0 };
            const unsigned long dim1787JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   11  ,   31  ,   47  ,   37  ,   107 ,   31  ,   691 ,   1049    ,   2273    ,   1595    ,   4431    ,0 };
            const unsigned long dim1788JoeKuoD5Init[]   =   {   1   ,   1   ,   1   ,   3   ,   23  ,   57  ,   51  ,   17  ,   225 ,   551 ,   965 ,   3497    ,   2549    ,   1153    ,0 };
            const unsigned long dim1789JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   5   ,   25  ,   41  ,   123 ,   143 ,   13  ,   437 ,   1081    ,   1625    ,   147 ,   6239    ,0 };
            const unsigned long dim1790JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   3   ,   31  ,   49  ,   17  ,   119 ,   315 ,   155 ,   37  ,   1125    ,   1223    ,   341 ,0 };
            const unsigned long dim1791JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   7   ,   9   ,   49  ,   35  ,   223 ,   449 ,   147 ,   317 ,   29  ,   4651    ,   15995   ,0 };
            const unsigned long dim1792JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   3   ,   9   ,   23  ,   97  ,   209 ,   343 ,   279 ,   1947    ,   3465    ,   7775    ,   5779    ,0 };
            const unsigned long dim1793JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   1   ,   7   ,   61  ,   111 ,   135 ,   473 ,   105 ,   1529    ,   2855    ,   3457    ,   7053    ,0 };
            const unsigned long dim1794JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   7   ,   29  ,   3   ,   63  ,   25  ,   115 ,   423 ,   1425    ,   1363    ,   77  ,   11485   ,0 };
            const unsigned long dim1795JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   15  ,   23  ,   59  ,   101 ,   87  ,   319 ,   785 ,   1639    ,   3427    ,   3165    ,   11273   ,0 };
            const unsigned long dim1796JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   13  ,   23  ,   7   ,   47  ,   221 ,   137 ,   775 ,   329 ,   1777    ,   3091    ,   4693    ,0 };
            const unsigned long dim1797JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   13  ,   1   ,   33  ,   31  ,   177 ,   481 ,   865 ,   739 ,   97  ,   5113    ,   16371   ,0 };
            const unsigned long dim1798JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   9   ,   31  ,   57  ,   99  ,   241 ,   465 ,   915 ,   1181    ,   225 ,   6795    ,   5743    ,0 };
            const unsigned long dim1799JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   3   ,   13  ,   25  ,   71  ,   119 ,   31  ,   169 ,   1745    ,   4085    ,   2945    ,   13357   ,0 };
            const unsigned long dim1800JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   5   ,   31  ,   61  ,   123 ,   71  ,   375 ,   1003    ,   1303    ,   2149    ,   5867    ,   10523   ,0 };
            const unsigned long dim1801JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   11  ,   21  ,   13  ,   25  ,   147 ,   147 ,   591 ,   259 ,   1707    ,   1777    ,   5869    ,0 };
            const unsigned long dim1802JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   13  ,   7   ,   13  ,   87  ,   227 ,   305 ,   235 ,   1263    ,   953 ,   4509    ,   11375   ,0 };
            const unsigned long dim1803JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   1   ,   11  ,   23  ,   103 ,   177 ,   9   ,   329 ,   1519    ,   2393    ,   6627    ,   14631   ,0 };
            const unsigned long dim1804JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   9   ,   25  ,   19  ,   75  ,   87  ,   361 ,   741 ,   1745    ,   3281    ,   6771    ,   3111    ,0 };
            const unsigned long dim1805JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   5   ,   23  ,   5   ,   1   ,   247 ,   43  ,   61  ,   1489    ,   3537    ,   5079    ,   11545   ,0 };
            const unsigned long dim1806JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   15  ,   9   ,   43  ,   49  ,   213 ,   191 ,   567 ,   1237    ,   681 ,   6715    ,   6471    ,0 };
            const unsigned long dim1807JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   1   ,   31  ,   53  ,   83  ,   53  ,   117 ,   1001    ,   525 ,   841 ,   2891    ,   4281    ,0 };
            const unsigned long dim1808JoeKuoD5Init[]   =   {   1   ,   1   ,   1   ,   9   ,   21  ,   49  ,   21  ,   215 ,   209 ,   611 ,   1849    ,   969 ,   3081    ,   9485    ,0 };
            const unsigned long dim1809JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   15  ,   13  ,   25  ,   37  ,   31  ,   357 ,   611 ,   83  ,   1615    ,   8137    ,   14505   ,0 };
            const unsigned long dim1810JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   11  ,   29  ,   31  ,   9   ,   27  ,   427 ,   883 ,   555 ,   2559    ,   7039    ,   11345   ,0 };
            const unsigned long dim1811JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   11  ,   3   ,   17  ,   9   ,   251 ,   493 ,   977 ,   1713    ,   2711    ,   4377    ,   3171    ,0 };
            const unsigned long dim1812JoeKuoD5Init[]   =   {   1   ,   1   ,   1   ,   7   ,   9   ,   1   ,   27  ,   179 ,   345 ,   425 ,   413 ,   2101    ,   3417    ,   1497    ,0 };
            const unsigned long dim1813JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   3   ,   13  ,   33  ,   121 ,   81  ,   247 ,   1003    ,   1405    ,   2769    ,   1919    ,   10807   ,0 };
            const unsigned long dim1814JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   15  ,   29  ,   51  ,   13  ,   101 ,   237 ,   165 ,   1483    ,   3961    ,   2389    ,   8379    ,0 };
            const unsigned long dim1815JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   11  ,   17  ,   3   ,   51  ,   141 ,   295 ,   165 ,   1089    ,   3889    ,   6415    ,   4969    ,0 };
            const unsigned long dim1816JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   3   ,   25  ,   47  ,   1   ,   17  ,   499 ,   409 ,   1417    ,   2975    ,   3935    ,   13363   ,0 };
            const unsigned long dim1817JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   13  ,   7   ,   55  ,   91  ,   121 ,   509 ,   675 ,   1203    ,   795 ,   1225    ,   15329   ,0 };
            const unsigned long dim1818JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   5   ,   3   ,   29  ,   19  ,   103 ,   11  ,   511 ,   799 ,   773 ,   515 ,   9931    ,0 };
            const unsigned long dim1819JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   3   ,   27  ,   63  ,   19  ,   83  ,   213 ,   977 ,   1923    ,   999 ,   7935    ,   2081    ,0 };
            const unsigned long dim1820JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   3   ,   19  ,   17  ,   127 ,   251 ,   417 ,   711 ,   85  ,   2757    ,   6461    ,   83  ,0 };
            const unsigned long dim1821JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   9   ,   11  ,   7   ,   113 ,   127 ,   287 ,   647 ,   1775    ,   3201    ,   2551    ,   13389   ,0 };
            const unsigned long dim1822JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   15  ,   31  ,   33  ,   81  ,   53  ,   207 ,   361 ,   665 ,   2073    ,   4249    ,   6699    ,0 };
            const unsigned long dim1823JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   11  ,   25  ,   43  ,   83  ,   159 ,   31  ,   693 ,   1315    ,   2043    ,   3463    ,   15771   ,0 };
            const unsigned long dim1824JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   15  ,   9   ,   1   ,   93  ,   207 ,   97  ,   293 ,   67  ,   2411    ,   1241    ,   10819   ,0 };
            const unsigned long dim1825JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   15  ,   17  ,   41  ,   113 ,   153 ,   7   ,   499 ,   131 ,   737 ,   7881    ,   2691    ,0 };
            const unsigned long dim1826JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   11  ,   5   ,   3   ,   125 ,   49  ,   63  ,   375 ,   811 ,   3295    ,   7997    ,   1063    ,0 };
            const unsigned long dim1827JoeKuoD5Init[]   =   {   1   ,   1   ,   1   ,   5   ,   1   ,   27  ,   37  ,   101 ,   179 ,   975 ,   421 ,   2785    ,   8093    ,   11803   ,0 };
            const unsigned long dim1828JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   3   ,   27  ,   35  ,   89  ,   111 ,   481 ,   821 ,   223 ,   2983    ,   2369    ,   4861    ,0 };
            const unsigned long dim1829JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   3   ,   21  ,   59  ,   25  ,   237 ,   73  ,   451 ,   309 ,   241 ,   4955    ,   9889    ,0 };
            const unsigned long dim1830JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   15  ,   25  ,   37  ,   85  ,   29  ,   103 ,   893 ,   337 ,   2831    ,   6679    ,   15171   ,0 };
            const unsigned long dim1831JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   3   ,   13  ,   1   ,   101 ,   109 ,   223 ,   243 ,   1749    ,   3611    ,   4097    ,   14439   ,0 };
            const unsigned long dim1832JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   11  ,   31  ,   49  ,   97  ,   181 ,   269 ,   567 ,   1801    ,   3129    ,   2697    ,   15167   ,0 };
            const unsigned long dim1833JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   5   ,   11  ,   11  ,   105 ,   161 ,   163 ,   363 ,   187 ,   3671    ,   6793    ,   4197    ,0 };
            const unsigned long dim1834JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   1   ,   23  ,   61  ,   127 ,   9   ,   101 ,   493 ,   1875    ,   3943    ,   3501    ,   1685    ,0 };
            const unsigned long dim1835JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   3   ,   29  ,   13  ,   11  ,   185 ,   437 ,   151 ,   927 ,   199 ,   7739    ,   14771   ,0 };
            const unsigned long dim1836JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   5   ,   17  ,   17  ,   35  ,   147 ,   387 ,   401 ,   1385    ,   1993    ,   1551    ,   2421    ,0 };
            const unsigned long dim1837JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   7   ,   25  ,   27  ,   1   ,   43  ,   407 ,   503 ,   1917    ,   23  ,   3459    ,   7653    ,0 };
            const unsigned long dim1838JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   1   ,   29  ,   37  ,   69  ,   193 ,   359 ,   901 ,   1661    ,   4049    ,   2923    ,   15553   ,0 };
            const unsigned long dim1839JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   11  ,   31  ,   53  ,   45  ,   5   ,   143 ,   823 ,   483 ,   669 ,   8037    ,   14021   ,0 };
            const unsigned long dim1840JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   15  ,   23  ,   47  ,   105 ,   185 ,   267 ,   347 ,   85  ,   3121    ,   5811    ,   6229    ,0 };
            const unsigned long dim1841JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   5   ,   27  ,   43  ,   25  ,   143 ,   197 ,   425 ,   1537    ,   1701    ,   6061    ,   12125   ,0 };
            const unsigned long dim1842JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   1   ,   31  ,   57  ,   21  ,   21  ,   301 ,   573 ,   947 ,   2087    ,   3135    ,   1767    ,0 };
            const unsigned long dim1843JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   15  ,   23  ,   35  ,   35  ,   191 ,   195 ,   131 ,   1669    ,   3311    ,   4107    ,   9311    ,0 };
            const unsigned long dim1844JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   5   ,   23  ,   13  ,   111 ,   165 ,   283 ,   949 ,   1575    ,   2359    ,   5891    ,   9911    ,0 };
            const unsigned long dim1845JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   13  ,   17  ,   31  ,   31  ,   179 ,   45  ,   475 ,   975 ,   251 ,   353 ,   6331    ,0 };
            const unsigned long dim1846JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   13  ,   15  ,   9   ,   127 ,   35  ,   463 ,   711 ,   1783    ,   837 ,   7859    ,   14299   ,0 };
            const unsigned long dim1847JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   5   ,   31  ,   57  ,   79  ,   7   ,   319 ,   517 ,   677 ,   3923    ,   1363    ,   11355   ,0 };
            const unsigned long dim1848JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   3   ,   17  ,   17  ,   49  ,   171 ,   95  ,   519 ,   775 ,   2265    ,   7953    ,   617 ,0 };
            const unsigned long dim1849JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   9   ,   3   ,   35  ,   87  ,   185 ,   369 ,   795 ,   1171    ,   3777    ,   4983    ,   1153    ,0 };
            const unsigned long dim1850JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   3   ,   9   ,   3   ,   1   ,   131 ,   275 ,   921 ,   507 ,   1877    ,   1573    ,   6231    ,0 };
            const unsigned long dim1851JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   15  ,   7   ,   53  ,   77  ,   255 ,   329 ,   449 ,   1859    ,   3625    ,   7027    ,   9921    ,0 };
            const unsigned long dim1852JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   7   ,   1   ,   47  ,   57  ,   111 ,   303 ,   749 ,   689 ,   2963    ,   8085    ,   9097    ,0 };
            const unsigned long dim1853JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   11  ,   19  ,   47  ,   61  ,   3   ,   481 ,   371 ,   1737    ,   287 ,   5485    ,   15433   ,0 };
            const unsigned long dim1854JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   13  ,   23  ,   45  ,   33  ,   83  ,   483 ,   411 ,   257 ,   749 ,   3959    ,   10269   ,0 };
            const unsigned long dim1855JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   5   ,   25  ,   15  ,   11  ,   181 ,   131 ,   943 ,   125 ,   3367    ,   2429    ,   2151    ,0 };
            const unsigned long dim1856JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   15  ,   25  ,   35  ,   37  ,   163 ,   399 ,   575 ,   843 ,   1567    ,   4401    ,   14757   ,0 };
            const unsigned long dim1857JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   1   ,   5   ,   3   ,   31  ,   199 ,   479 ,   525 ,   973 ,   2843    ,   7831    ,   6693    ,0 };
            const unsigned long dim1858JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   3   ,   21  ,   7   ,   123 ,   31  ,   207 ,   741 ,   425 ,   2139    ,   2965    ,   1017    ,0 };
            const unsigned long dim1859JoeKuoD5Init[]   =   {   1   ,   1   ,   1   ,   11  ,   3   ,   31  ,   61  ,   85  ,   303 ,   627 ,   1995    ,   3211    ,   5747    ,   14897   ,0 };
            const unsigned long dim1860JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   7   ,   9   ,   25  ,   45  ,   73  ,   427 ,   945 ,   1505    ,   1121    ,   7585    ,   13115   ,0 };
            const unsigned long dim1861JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   1   ,   9   ,   43  ,   75  ,   63  ,   445 ,   709 ,   645 ,   1105    ,   1375    ,   12303   ,0 };
            const unsigned long dim1862JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   5   ,   25  ,   37  ,   101 ,   209 ,   239 ,   429 ,   1455    ,   3587    ,   4231    ,   6501    ,0 };
            const unsigned long dim1863JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   1   ,   21  ,   23  ,   113 ,   227 ,   151 ,   97  ,   367 ,   889 ,   7675    ,   14093   ,0 };
            const unsigned long dim1864JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   11  ,   13  ,   31  ,   63  ,   7   ,   31  ,   973 ,   1935    ,   53  ,   4941    ,   9057    ,0 };
            const unsigned long dim1865JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   13  ,   21  ,   23  ,   79  ,   43  ,   255 ,   1003    ,   1741    ,   1003    ,   355 ,   3839    ,0 };
            const unsigned long dim1866JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   9   ,   31  ,   17  ,   45  ,   235 ,   271 ,   711 ,   61  ,   235 ,   513 ,   11321   ,0 };
            const unsigned long dim1867JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   9   ,   29  ,   35  ,   43  ,   115 ,   9   ,   787 ,   325 ,   625 ,   7905    ,   8191    ,   29233   ,0 };
            const unsigned long dim1868JoeKuoD5Init[]   =   {   1   ,   1   ,   1   ,   13  ,   5   ,   17  ,   67  ,   161 ,   357 ,   163 ,   1129    ,   3819    ,   3601    ,   5961    ,   9259    ,0 };
            const unsigned long dim1869JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   13  ,   25  ,   17  ,   83  ,   41  ,   463 ,   879 ,   1881    ,   1107    ,   4337    ,   409 ,   8633    ,0 };
            const unsigned long dim1870JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   1   ,   17  ,   45  ,   127 ,   213 ,   285 ,   791 ,   479 ,   3153    ,   3001    ,   16207   ,   32669   ,0 };
            const unsigned long dim1871JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   11  ,   23  ,   41  ,   103 ,   209 ,   81  ,   903 ,   1787    ,   907 ,   161 ,   15739   ,   17367   ,0 };
            const unsigned long dim1872JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   11  ,   17  ,   61  ,   69  ,   251 ,   321 ,   361 ,   905 ,   1081    ,   5749    ,   2639    ,   21195   ,0 };
            const unsigned long dim1873JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   13  ,   27  ,   39  ,   61  ,   161 ,   139 ,   837 ,   579 ,   1979    ,   4413    ,   8701    ,   16301   ,0 };
            const unsigned long dim1874JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   7   ,   3   ,   49  ,   111 ,   77  ,   423 ,   491 ,   1489    ,   2615    ,   4997    ,   3695    ,   15109   ,0 };
            const unsigned long dim1875JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   7   ,   31  ,   3   ,   53  ,   113 ,   45  ,   767 ,   1561    ,   2925    ,   605 ,   3211    ,   8413    ,0 };
            const unsigned long dim1876JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   15  ,   31  ,   29  ,   65  ,   13  ,   19  ,   159 ,   639 ,   2155    ,   2699    ,   7001    ,   15575   ,0 };
            const unsigned long dim1877JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   1   ,   3   ,   33  ,   37  ,   89  ,   189 ,   661 ,   1685    ,   503 ,   6037    ,   12993   ,   9571    ,0 };
            const unsigned long dim1878JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   13  ,   31  ,   13  ,   41  ,   35  ,   217 ,   193 ,   1273    ,   443 ,   1385    ,   13661   ,   14873   ,0 };
            const unsigned long dim1879JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   15  ,   3   ,   9   ,   123 ,   251 ,   155 ,   545 ,   813 ,   317 ,   5127    ,   15305   ,   5571    ,0 };
            const unsigned long dim1880JoeKuoD5Init[]   =   {   1   ,   1   ,   1   ,   15  ,   3   ,   39  ,   41  ,   39  ,   273 ,   881 ,   809 ,   1643    ,   687 ,   5525    ,   3473    ,0 };
            const unsigned long dim1881JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   11  ,   7   ,   25  ,   25  ,   27  ,   263 ,   773 ,   155 ,   1541    ,   7545    ,   873 ,   32713   ,0 };
            const unsigned long dim1882JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   9   ,   1   ,   49  ,   57  ,   17  ,   433 ,   73  ,   635 ,   1073    ,   7459    ,   15653   ,   47  ,0 };
            const unsigned long dim1883JoeKuoD5Init[]   =   {   1   ,   1   ,   1   ,   13  ,   1   ,   5   ,   65  ,   105 ,   353 ,   361 ,   339 ,   3983    ,   2707    ,   10165   ,   20883   ,0 };
            const unsigned long dim1884JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   3   ,   23  ,   29  ,   125 ,   223 ,   463 ,   493 ,   1989    ,   1583    ,   8153    ,   13475   ,   5789    ,0 };
            const unsigned long dim1885JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   7   ,   27  ,   39  ,   61  ,   127 ,   387 ,   215 ,   1769    ,   2279    ,   7883    ,   1491    ,   29625   ,0 };
            const unsigned long dim1886JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   5   ,   5   ,   13  ,   11  ,   47  ,   409 ,   673 ,   1927    ,   1277    ,   6971    ,   8195    ,   22663   ,0 };
            const unsigned long dim1887JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   13  ,   1   ,   53  ,   123 ,   187 ,   25  ,   29  ,   1107    ,   1929    ,   6995    ,   5081    ,   8861    ,0 };
            const unsigned long dim1888JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   5   ,   25  ,   27  ,   53  ,   65  ,   239 ,   299 ,   21  ,   3389    ,   1587    ,   1437    ,   26209   ,0 };
            const unsigned long dim1889JoeKuoD5Init[]   =   {   1   ,   1   ,   1   ,   7   ,   15  ,   35  ,   37  ,   105 ,   169 ,   843 ,   623 ,   3763    ,   7101    ,   5245    ,   17203   ,0 };
            const unsigned long dim1890JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   15  ,   21  ,   7   ,   29  ,   29  ,   465 ,   149 ,   189 ,   1115    ,   6471    ,   11275   ,   28769   ,0 };
            const unsigned long dim1891JoeKuoD5Init[]   =   {   1   ,   1   ,   1   ,   9   ,   31  ,   1   ,   99  ,   101 ,   321 ,   675 ,   1631    ,   1371    ,   7733    ,   1697    ,   4437    ,0 };
            const unsigned long dim1892JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   9   ,   7   ,   27  ,   47  ,   207 ,   295 ,   787 ,   1539    ,   2233    ,   7849    ,   15843   ,   11149   ,0 };
            const unsigned long dim1893JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   3   ,   17  ,   29  ,   61  ,   143 ,   289 ,   939 ,   729 ,   297 ,   1513    ,   4387    ,   3347    ,0 };
            const unsigned long dim1894JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   13  ,   15  ,   25  ,   125 ,   13  ,   157 ,   377 ,   317 ,   827 ,   5137    ,   829 ,   21215   ,0 };
            const unsigned long dim1895JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   13  ,   17  ,   37  ,   79  ,   87  ,   163 ,   847 ,   1009    ,   2259    ,   6543    ,   8697    ,   4837    ,0 };
            const unsigned long dim1896JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   11  ,   29  ,   7   ,   13  ,   91  ,   279 ,   439 ,   691 ,   4047    ,   159 ,   9403    ,   27735   ,0 };
            const unsigned long dim1897JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   5   ,   23  ,   61  ,   53  ,   93  ,   343 ,   349 ,   355 ,   1275    ,   3573    ,   12847   ,   8969    ,0 };
            const unsigned long dim1898JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   11  ,   1   ,   49  ,   47  ,   255 ,   371 ,   511 ,   579 ,   2385    ,   2237    ,   2015    ,   23973   ,0 };
            const unsigned long dim1899JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   11  ,   15  ,   31  ,   13  ,   121 ,   399 ,   957 ,   731 ,   3743    ,   1573    ,   2293    ,   27755   ,0 };
            const unsigned long dim1900JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   5   ,   23  ,   19  ,   57  ,   211 ,   133 ,   197 ,   379 ,   1037    ,   625 ,   9405    ,   11547   ,0 };
            const unsigned long dim1901JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   15  ,   7   ,   13  ,   75  ,   93  ,   415 ,   551 ,   1885    ,   2259    ,   23  ,   14321   ,   21509   ,0 };
            const unsigned long dim1902JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   5   ,   29  ,   37  ,   23  ,   59  ,   199 ,   559 ,   1761    ,   821 ,   5077    ,   12017   ,   16505   ,0 };
            const unsigned long dim1903JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   7   ,   19  ,   29  ,   31  ,   101 ,   121 ,   605 ,   1679    ,   2317    ,   3359    ,   13557   ,   17567   ,0 };
            const unsigned long dim1904JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   15  ,   3   ,   5   ,   45  ,   201 ,   401 ,   689 ,   1775    ,   3615    ,   2641    ,   14149   ,   29241   ,0 };
            const unsigned long dim1905JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   5   ,   9   ,   13  ,   51  ,   101 ,   41  ,   347 ,   57  ,   613 ,   5813    ,   3753    ,   22007   ,0 };
            const unsigned long dim1906JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   15  ,   19  ,   59  ,   57  ,   57  ,   405 ,   647 ,   1943    ,   353 ,   1691    ,   9287    ,   29567   ,0 };
            const unsigned long dim1907JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   13  ,   29  ,   11  ,   51  ,   137 ,   163 ,   365 ,   153 ,   3791    ,   5367    ,   4649    ,   11255   ,0 };
            const unsigned long dim1908JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   1   ,   27  ,   37  ,   99  ,   211 ,   377 ,   543 ,   649 ,   2107    ,   3511    ,   10385   ,   1093    ,0 };
            const unsigned long dim1909JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   9   ,   29  ,   25  ,   85  ,   75  ,   55  ,   75  ,   825 ,   2129    ,   6867    ,   9053    ,   22687   ,0 };
            const unsigned long dim1910JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   1   ,   19  ,   57  ,   29  ,   99  ,   369 ,   107 ,   1187    ,   2919    ,   6163    ,   15209   ,   7911    ,0 };
            const unsigned long dim1911JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   11  ,   25  ,   55  ,   111 ,   197 ,   329 ,   307 ,   1531    ,   1753    ,   1275    ,   5543    ,   23411   ,0 };
            const unsigned long dim1912JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   9   ,   21  ,   59  ,   97  ,   211 ,   117 ,   331 ,   1919    ,   1763    ,   1755    ,   12517   ,   16579   ,0 };
            const unsigned long dim1913JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   11  ,   5   ,   37  ,   21  ,   229 ,   109 ,   341 ,   421 ,   2413    ,   785 ,   3251    ,   10245   ,0 };
            const unsigned long dim1914JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   11  ,   27  ,   3   ,   89  ,   15  ,   455 ,   53  ,   1447    ,   137 ,   681 ,   7811    ,   16817   ,0 };
            const unsigned long dim1915JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   13  ,   9   ,   57  ,   45  ,   101 ,   459 ,   801 ,   409 ,   911 ,   749 ,   11875   ,   6039    ,0 };
            const unsigned long dim1916JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   13  ,   23  ,   31  ,   49  ,   105 ,   255 ,   485 ,   1587    ,   1137    ,   2113    ,   16313   ,   19073   ,0 };
            const unsigned long dim1917JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   9   ,   31  ,   59  ,   117 ,   213 ,   293 ,   195 ,   991 ,   3531    ,   157 ,   1747    ,   20883   ,0 };
            const unsigned long dim1918JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   11  ,   17  ,   31  ,   63  ,   33  ,   415 ,   341 ,   779 ,   423 ,   5661    ,   2533    ,   23031   ,0 };
            const unsigned long dim1919JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   3   ,   11  ,   41  ,   69  ,   117 ,   441 ,   3   ,   885 ,   3387    ,   4291    ,   497 ,   9991    ,0 };
            const unsigned long dim1920JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   15  ,   27  ,   29  ,   17  ,   109 ,   315 ,   433 ,   291 ,   577 ,   3209    ,   3305    ,   6759    ,0 };
            const unsigned long dim1921JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   1   ,   29  ,   37  ,   75  ,   139 ,   243 ,   217 ,   319 ,   1025    ,   2415    ,   5957    ,   8303    ,0 };
            const unsigned long dim1922JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   9   ,   15  ,   41  ,   19  ,   21  ,   279 ,   563 ,   893 ,   1391    ,   4907    ,   14381   ,   7165    ,0 };
            const unsigned long dim1923JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   13  ,   21  ,   35  ,   23  ,   139 ,   441 ,   463 ,   563 ,   155 ,   2009    ,   14887   ,   30943   ,0 };
            const unsigned long dim1924JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   3   ,   23  ,   23  ,   13  ,   157 ,   511 ,   401 ,   1573    ,   3019    ,   1791    ,   587 ,   10927   ,0 };
            const unsigned long dim1925JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   5   ,   1   ,   45  ,   127 ,   85  ,   27  ,   729 ,   993 ,   1487    ,   5577    ,   14113   ,   23163   ,0 };
            const unsigned long dim1926JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   5   ,   31  ,   15  ,   19  ,   7   ,   503 ,   359 ,   595 ,   1471    ,   2587    ,   7827    ,   31497   ,0 };
            const unsigned long dim1927JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   1   ,   13  ,   29  ,   35  ,   49  ,   381 ,   677 ,   407 ,   2681    ,   7923    ,   2917    ,   4001    ,0 };
            const unsigned long dim1928JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   5   ,   25  ,   51  ,   3   ,   33  ,   229 ,   105 ,   1449    ,   417 ,   2577    ,   5185    ,   18737   ,0 };
            const unsigned long dim1929JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   9   ,   29  ,   17  ,   9   ,   185 ,   253 ,   929 ,   1267    ,   2559    ,   2271    ,   10501   ,   21687   ,0 };
            const unsigned long dim1930JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   13  ,   9   ,   29  ,   89  ,   193 ,   309 ,   385 ,   73  ,   1835    ,   5193    ,   10235   ,   29827   ,0 };
            const unsigned long dim1931JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   7   ,   29  ,   23  ,   9   ,   119 ,   325 ,   277 ,   1883    ,   613 ,   6299    ,   10799   ,   16271   ,0 };
            const unsigned long dim1932JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   11  ,   29  ,   55  ,   71  ,   5   ,   99  ,   303 ,   1853    ,   2225    ,   3047    ,   16019   ,   23257   ,0 };
            const unsigned long dim1933JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   1   ,   19  ,   19  ,   35  ,   39  ,   389 ,   737 ,   433 ,   1489    ,   5903    ,   10181   ,   11249   ,0 };
            const unsigned long dim1934JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   13  ,   15  ,   13  ,   21  ,   191 ,   369 ,   939 ,   1353    ,   2645    ,   221 ,   49  ,   14783   ,0 };
            const unsigned long dim1935JoeKuoD5Init[]   =   {   1   ,   1   ,   1   ,   1   ,   17  ,   5   ,   21  ,   121 ,   359 ,   863 ,   893 ,   2887    ,   4431    ,   5245    ,   20865   ,0 };
            const unsigned long dim1936JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   11  ,   15  ,   25  ,   59  ,   37  ,   263 ,   425 ,   1143    ,   3843    ,   4955    ,   6639    ,   8411    ,0 };
            const unsigned long dim1937JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   15  ,   3   ,   5   ,   123 ,   131 ,   229 ,   93  ,   113 ,   3609    ,   2007    ,   7709    ,   32141   ,0 };
            const unsigned long dim1938JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   11  ,   25  ,   43  ,   103 ,   199 ,   243 ,   533 ,   1947    ,   683 ,   1013    ,   13155   ,   29877   ,0 };
            const unsigned long dim1939JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   9   ,   17  ,   41  ,   57  ,   99  ,   137 ,   839 ,   1437    ,   3997    ,   2473    ,   10169   ,   20253   ,0 };
            const unsigned long dim1940JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   5   ,   19  ,   49  ,   55  ,   185 ,   175 ,   319 ,   1173    ,   1049    ,   5289    ,   3297    ,   23755   ,0 };
            const unsigned long dim1941JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   5   ,   25  ,   7   ,   81  ,   119 ,   319 ,   743 ,   1655    ,   3719    ,   5731    ,   11015   ,   11309   ,0 };
            const unsigned long dim1942JoeKuoD5Init[]   =   {   1   ,   1   ,   1   ,   9   ,   31  ,   53  ,   61  ,   119 ,   207 ,   689 ,   81  ,   367 ,   2495    ,   1965    ,   15559   ,0 };
            const unsigned long dim1943JoeKuoD5Init[]   =   {   1   ,   1   ,   1   ,   3   ,   21  ,   47  ,   9   ,   199 ,   273 ,   397 ,   1293    ,   2709    ,   3687    ,   5755    ,   6015    ,0 };
            const unsigned long dim1944JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   15  ,   1   ,   7   ,   55  ,   55  ,   343 ,   363 ,   1207    ,   3731    ,   7489    ,   2365    ,   25301   ,0 };
            const unsigned long dim1945JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   7   ,   3   ,   31  ,   87  ,   15  ,   327 ,   479 ,   177 ,   1261    ,   913 ,   4189    ,   4565    ,0 };
            const unsigned long dim1946JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   9   ,   19  ,   37  ,   29  ,   17  ,   185 ,   827 ,   239 ,   281 ,   287 ,   16105   ,   23549   ,0 };
            const unsigned long dim1947JoeKuoD5Init[]   =   {   1   ,   1   ,   1   ,   3   ,   23  ,   17  ,   73  ,   123 ,   287 ,   665 ,   297 ,   215 ,   359 ,   11741   ,   29159   ,0 };
            const unsigned long dim1948JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   11  ,   13  ,   45  ,   35  ,   81  ,   211 ,   257 ,   655 ,   911 ,   5949    ,   9597    ,   11677   ,0 };
            const unsigned long dim1949JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   15  ,   7   ,   39  ,   31  ,   103 ,   341 ,   871 ,   299 ,   635 ,   951 ,   13669   ,   30083   ,0 };
            const unsigned long dim1950JoeKuoD5Init[]   =   {   1   ,   1   ,   1   ,   7   ,   7   ,   15  ,   55  ,   45  ,   375 ,   711 ,   1339    ,   3045    ,   4691    ,   13247   ,   14611   ,0 };
            const unsigned long dim1951JoeKuoD5Init[]   =   {   1   ,   1   ,   1   ,   9   ,   19  ,   33  ,   67  ,   91  ,   83  ,   541 ,   1229    ,   3209    ,   2783    ,   11519   ,   10729   ,0 };
            const unsigned long dim1952JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   7   ,   15  ,   15  ,   89  ,   7   ,   173 ,   809 ,   1229    ,   117 ,   2323    ,   15145   ,   2863    ,0 };
            const unsigned long dim1953JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   13  ,   1   ,   55  ,   39  ,   187 ,   491 ,   675 ,   1357    ,   1109    ,   2701    ,   14891   ,   17061   ,0 };
            const unsigned long dim1954JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   5   ,   7   ,   47  ,   127 ,   111 ,   481 ,   177 ,   1553    ,   2949    ,   6403    ,   1993    ,   26227   ,0 };
            const unsigned long dim1955JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   11  ,   31  ,   3   ,   111 ,   193 ,   445 ,   359 ,   1039    ,   667 ,   7463    ,   7983    ,   10901   ,0 };
            const unsigned long dim1956JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   15  ,   29  ,   17  ,   15  ,   69  ,   111 ,   27  ,   647 ,   3377    ,   185 ,   12083   ,   21197   ,0 };
            const unsigned long dim1957JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   5   ,   17  ,   57  ,   5   ,   189 ,   383 ,   407 ,   145 ,   1185    ,   3307    ,   12373   ,   28531   ,0 };
            const unsigned long dim1958JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   9   ,   23  ,   43  ,   41  ,   19  ,   127 ,   687 ,   1841    ,   3707    ,   4983    ,   14585   ,   15253   ,0 };
            const unsigned long dim1959JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   9   ,   13  ,   61  ,   13  ,   113 ,   43  ,   423 ,   1073    ,   1843    ,   1071    ,   7869    ,   21137   ,0 };
            const unsigned long dim1960JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   3   ,   7   ,   43  ,   87  ,   185 ,   375 ,   699 ,   283 ,   3265    ,   4905    ,   9317    ,   3055    ,0 };
            const unsigned long dim1961JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   9   ,   17  ,   15  ,   47  ,   73  ,   211 ,   37  ,   427 ,   3841    ,   2801    ,   12131   ,   19619   ,0 };
            const unsigned long dim1962JoeKuoD5Init[]   =   {   1   ,   1   ,   1   ,   7   ,   17  ,   35  ,   53  ,   67  ,   221 ,   97  ,   199 ,   2825    ,   1997    ,   9903    ,   3003    ,0 };
            const unsigned long dim1963JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   9   ,   23  ,   53  ,   69  ,   225 ,   413 ,   307 ,   553 ,   2631    ,   6499    ,   6277    ,   30337   ,0 };
            const unsigned long dim1964JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   13  ,   5   ,   23  ,   85  ,   193 ,   437 ,   607 ,   397 ,   2003    ,   4345    ,   3575    ,   27939   ,0 };
            const unsigned long dim1965JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   13  ,   13  ,   45  ,   117 ,   191 ,   143 ,   477 ,   473 ,   3051    ,   5903    ,   6253    ,   22679   ,0 };
            const unsigned long dim1966JoeKuoD5Init[]   =   {   1   ,   1   ,   1   ,   13  ,   31  ,   49  ,   31  ,   217 ,   119 ,   893 ,   1415    ,   1633    ,   7573    ,   8353    ,   32231   ,0 };
            const unsigned long dim1967JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   5   ,   29  ,   47  ,   25  ,   3   ,   425 ,   883 ,   1461    ,   3679    ,   6635    ,   9103    ,   26793   ,0 };
            const unsigned long dim1968JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   15  ,   11  ,   27  ,   87  ,   31  ,   153 ,   115 ,   1573    ,   745 ,   6515    ,   14321   ,   21301   ,0 };
            const unsigned long dim1969JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   11  ,   15  ,   57  ,   41  ,   157 ,   147 ,   867 ,   1959    ,   943 ,   6431    ,   14581   ,   9579    ,0 };
            const unsigned long dim1970JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   9   ,   5   ,   13  ,   1   ,   203 ,   317 ,   185 ,   1569    ,   3621    ,   4873    ,   13249   ,   19153   ,0 };
            const unsigned long dim1971JoeKuoD5Init[]   =   {   1   ,   1   ,   1   ,   5   ,   5   ,   29  ,   41  ,   3   ,   37  ,   919 ,   1037    ,   237 ,   4699    ,   12011   ,   18669   ,0 };
            const unsigned long dim1972JoeKuoD5Init[]   =   {   1   ,   1   ,   1   ,   1   ,   21  ,   1   ,   55  ,   195 ,   289 ,   635 ,   981 ,   1395    ,   1827    ,   7481    ,   19163   ,0 };
            const unsigned long dim1973JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   13  ,   15  ,   15  ,   89  ,   109 ,   17  ,   111 ,   815 ,   2637    ,   6917    ,   7973    ,   9471    ,0 };
            const unsigned long dim1974JoeKuoD5Init[]   =   {   1   ,   1   ,   1   ,   15  ,   9   ,   25  ,   73  ,   101 ,   417 ,   519 ,   1697    ,   2861    ,   2281    ,   10959   ,   30433   ,0 };
            const unsigned long dim1975JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   3   ,   21  ,   51  ,   89  ,   157 ,   181 ,   307 ,   355 ,   661 ,   4885    ,   12411   ,   23473   ,0 };
            const unsigned long dim1976JoeKuoD5Init[]   =   {   1   ,   1   ,   1   ,   15  ,   17  ,   37  ,   19  ,   139 ,   205 ,   73  ,   97  ,   2463    ,   2785    ,   9355    ,   23989   ,0 };
            const unsigned long dim1977JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   1   ,   7   ,   39  ,   117 ,   191 ,   395 ,   539 ,   1823    ,   2333    ,   1205    ,   14131   ,   2301    ,0 };
            const unsigned long dim1978JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   13  ,   21  ,   11  ,   49  ,   43  ,   381 ,   15  ,   1743    ,   641 ,   95  ,   10581   ,   15437   ,0 };
            const unsigned long dim1979JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   7   ,   5   ,   53  ,   55  ,   23  ,   265 ,   163 ,   173 ,   1399    ,   7257    ,   15097   ,   13491   ,0 };
            const unsigned long dim1980JoeKuoD5Init[]   =   {   1   ,   3   ,   3   ,   13  ,   27  ,   55  ,   97  ,   51  ,   469 ,   7   ,   871 ,   3213    ,   5719    ,   871 ,   11669   ,0 };
            const unsigned long dim1981JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   3   ,   29  ,   35  ,   47  ,   31  ,   77  ,   335 ,   537 ,   1695    ,   461 ,   14417   ,   23945   ,0 };
            const unsigned long dim1982JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   13  ,   1   ,   47  ,   97  ,   113 ,   235 ,   593 ,   1437    ,   3893    ,   5299    ,   857 ,   451 ,0 };
            const unsigned long dim1983JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   3   ,   3   ,   53  ,   51  ,   203 ,   177 ,   205 ,   773 ,   281 ,   7689    ,   8039    ,   7275    ,0 };
            const unsigned long dim1984JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   3   ,   3   ,   37  ,   33  ,   113 ,   429 ,   791 ,   1593    ,   3259    ,   1275    ,   11113   ,   16001   ,0 };
            const unsigned long dim1985JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   7   ,   27  ,   9   ,   109 ,   229 ,   453 ,   633 ,   2047    ,   1803    ,   4127    ,   3453    ,   15625   ,0 };
            const unsigned long dim1986JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   15  ,   17  ,   59  ,   15  ,   73  ,   105 ,   627 ,   1181    ,   2925    ,   2077    ,   16067   ,   15829   ,0 };
            const unsigned long dim1987JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   1   ,   15  ,   41  ,   1   ,   143 ,   493 ,   261 ,   845 ,   737 ,   6249    ,   7663    ,   32439   ,0 };
            const unsigned long dim1988JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   9   ,   11  ,   45  ,   101 ,   227 ,   227 ,   211 ,   243 ,   2817    ,   179 ,   3361    ,   20535   ,0 };
            const unsigned long dim1989JoeKuoD5Init[]   =   {   1   ,   3   ,   1   ,   3   ,   11  ,   25  ,   105 ,   173 ,   249 ,   465 ,   853 ,   2365    ,   5035    ,   11541   ,   5481    ,0 };
            const unsigned long dim1990JoeKuoD5Init[]   =   {   1   ,   3   ,   5   ,   9   ,   21  ,   29  ,   77  ,   245 ,   311 ,   879 ,   1007    ,   2545    ,   1561    ,   2949    ,   24855   ,0 };
            const unsigned long dim1991JoeKuoD5Init[]   =   {   1   ,   1   ,   5   ,   13  ,   1   ,   31  ,   105 ,   151 ,   127 ,   413 ,   553 ,   645 ,   863 ,   15943   ,   32731   ,0 };
            const unsigned long dim1992JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   11  ,   21  ,   13  ,   71  ,   229 ,   283 ,   493 ,   1445    ,   15  ,   4681    ,   3137    ,   18199   ,0 };
            const unsigned long dim1993JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   15  ,   27  ,   17  ,   37  ,   213 ,   253 ,   117 ,   205 ,   1489    ,   2997    ,   6483    ,   17201   ,0 };
            const unsigned long dim1994JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   13  ,   3   ,   29  ,   83  ,   109 ,   503 ,   139 ,   1119    ,   59  ,   5259    ,   8289    ,   7717    ,0 };
            const unsigned long dim1995JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   13  ,   13  ,   5   ,   65  ,   231 ,   117 ,   1009    ,   163 ,   21  ,   7639    ,   16275   ,   10661   ,0 };
            const unsigned long dim1996JoeKuoD5Init[]   =   {   1   ,   3   ,   7   ,   7   ,   19  ,   63  ,   31  ,   211 ,   487 ,   277 ,   27  ,   3685    ,   5371    ,   8157    ,   29735   ,0 };
            const unsigned long dim1997JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   1   ,   11  ,   49  ,   113 ,   191 ,   3   ,   923 ,   797 ,   2055    ,   3999    ,   8511    ,   23931   ,0 };
            const unsigned long dim1998JoeKuoD5Init[]   =   {   1   ,   1   ,   7   ,   7   ,   13  ,   27  ,   87  ,   43  ,   433 ,   401 ,   1441    ,   1301    ,   2639    ,   5773    ,   12431   ,0 };
            const unsigned long dim1999JoeKuoD5Init[]   =   {   1   ,   1   ,   3   ,   15  ,   21  ,   55  ,   17  ,   57  ,   449 ,   811 ,   519 ,   2329    ,   7607    ,   4255    ,   2845    ,0 };

            const unsigned long * const JoeKuoD5initializers[1999]
            =
            {
                dim1JoeKuoD5Init,
                dim2JoeKuoD5Init,
                dim3JoeKuoD5Init,
                dim4JoeKuoD5Init,
                dim5JoeKuoD5Init,
                dim6JoeKuoD5Init,
                dim7JoeKuoD5Init,
                dim8JoeKuoD5Init,
                dim9JoeKuoD5Init,
                dim10JoeKuoD5Init,
                dim11JoeKuoD5Init,
                dim12JoeKuoD5Init,
                dim13JoeKuoD5Init,
                dim14JoeKuoD5Init,
                dim15JoeKuoD5Init,
                dim16JoeKuoD5Init,
                dim17JoeKuoD5Init,
                dim18JoeKuoD5Init,
                dim19JoeKuoD5Init,
                dim20JoeKuoD5Init,
                dim21JoeKuoD5Init,
                dim22JoeKuoD5Init,
                dim23JoeKuoD5Init,
                dim24JoeKuoD5Init,
                dim25JoeKuoD5Init,
                dim26JoeKuoD5Init,
                dim27JoeKuoD5Init,
                dim28JoeKuoD5Init,
                dim29JoeKuoD5Init,
                dim30JoeKuoD5Init,
                dim31JoeKuoD5Init,
                dim32JoeKuoD5Init,
                dim33JoeKuoD5Init,
                dim34JoeKuoD5Init,
                dim35JoeKuoD5Init,
                dim36JoeKuoD5Init,
                dim37JoeKuoD5Init,
                dim38JoeKuoD5Init,
                dim39JoeKuoD5Init,
                dim40JoeKuoD5Init,
                dim41JoeKuoD5Init,
                dim42JoeKuoD5Init,
                dim43JoeKuoD5Init,
                dim44JoeKuoD5Init,
                dim45JoeKuoD5Init,
                dim46JoeKuoD5Init,
                dim47JoeKuoD5Init,
                dim48JoeKuoD5Init,
                dim49JoeKuoD5Init,
                dim50JoeKuoD5Init,
                dim51JoeKuoD5Init,
                dim52JoeKuoD5Init,
                dim53JoeKuoD5Init,
                dim54JoeKuoD5Init,
                dim55JoeKuoD5Init,
                dim56JoeKuoD5Init,
                dim57JoeKuoD5Init,
                dim58JoeKuoD5Init,
                dim59JoeKuoD5Init,
                dim60JoeKuoD5Init,
                dim61JoeKuoD5Init,
                dim62JoeKuoD5Init,
                dim63JoeKuoD5Init,
                dim64JoeKuoD5Init,
                dim65JoeKuoD5Init,
                dim66JoeKuoD5Init,
                dim67JoeKuoD5Init,
                dim68JoeKuoD5Init,
                dim69JoeKuoD5Init,
                dim70JoeKuoD5Init,
                dim71JoeKuoD5Init,
                dim72JoeKuoD5Init,
                dim73JoeKuoD5Init,
                dim74JoeKuoD5Init,
                dim75JoeKuoD5Init,
                dim76JoeKuoD5Init,
                dim77JoeKuoD5Init,
                dim78JoeKuoD5Init,
                dim79JoeKuoD5Init,
                dim80JoeKuoD5Init,
                dim81JoeKuoD5Init,
                dim82JoeKuoD5Init,
                dim83JoeKuoD5Init,
                dim84JoeKuoD5Init,
                dim85JoeKuoD5Init,
                dim86JoeKuoD5Init,
                dim87JoeKuoD5Init,
                dim88JoeKuoD5Init,
                dim89JoeKuoD5Init,
                dim90JoeKuoD5Init,
                dim91JoeKuoD5Init,
                dim92JoeKuoD5Init,
                dim93JoeKuoD5Init,
                dim94JoeKuoD5Init,
                dim95JoeKuoD5Init,
                dim96JoeKuoD5Init,
                dim97JoeKuoD5Init,
                dim98JoeKuoD5Init,
                dim99JoeKuoD5Init,
                dim100JoeKuoD5Init,
                dim101JoeKuoD5Init,
                dim102JoeKuoD5Init,
                dim103JoeKuoD5Init,
                dim104JoeKuoD5Init,
                dim105JoeKuoD5Init,
                dim106JoeKuoD5Init,
                dim107JoeKuoD5Init,
                dim108JoeKuoD5Init,
                dim109JoeKuoD5Init,
                dim110JoeKuoD5Init,
                dim111JoeKuoD5Init,
                dim112JoeKuoD5Init,
                dim113JoeKuoD5Init,
                dim114JoeKuoD5Init,
                dim115JoeKuoD5Init,
                dim116JoeKuoD5Init,
                dim117JoeKuoD5Init,
                dim118JoeKuoD5Init,
                dim119JoeKuoD5Init,
                dim120JoeKuoD5Init,
                dim121JoeKuoD5Init,
                dim122JoeKuoD5Init,
                dim123JoeKuoD5Init,
                dim124JoeKuoD5Init,
                dim125JoeKuoD5Init,
                dim126JoeKuoD5Init,
                dim127JoeKuoD5Init,
                dim128JoeKuoD5Init,
                dim129JoeKuoD5Init,
                dim130JoeKuoD5Init,
                dim131JoeKuoD5Init,
                dim132JoeKuoD5Init,
                dim133JoeKuoD5Init,
                dim134JoeKuoD5Init,
                dim135JoeKuoD5Init,
                dim136JoeKuoD5Init,
                dim137JoeKuoD5Init,
                dim138JoeKuoD5Init,
                dim139JoeKuoD5Init,
                dim140JoeKuoD5Init,
                dim141JoeKuoD5Init,
                dim142JoeKuoD5Init,
                dim143JoeKuoD5Init,
                dim144JoeKuoD5Init,
                dim145JoeKuoD5Init,
                dim146JoeKuoD5Init,
                dim147JoeKuoD5Init,
                dim148JoeKuoD5Init,
                dim149JoeKuoD5Init,
                dim150JoeKuoD5Init,
                dim151JoeKuoD5Init,
                dim152JoeKuoD5Init,
                dim153JoeKuoD5Init,
                dim154JoeKuoD5Init,
                dim155JoeKuoD5Init,
                dim156JoeKuoD5Init,
                dim157JoeKuoD5Init,
                dim158JoeKuoD5Init,
                dim159JoeKuoD5Init,
                dim160JoeKuoD5Init,
                dim161JoeKuoD5Init,
                dim162JoeKuoD5Init,
                dim163JoeKuoD5Init,
                dim164JoeKuoD5Init,
                dim165JoeKuoD5Init,
                dim166JoeKuoD5Init,
                dim167JoeKuoD5Init,
                dim168JoeKuoD5Init,
                dim169JoeKuoD5Init,
                dim170JoeKuoD5Init,
                dim171JoeKuoD5Init,
                dim172JoeKuoD5Init,
                dim173JoeKuoD5Init,
                dim174JoeKuoD5Init,
                dim175JoeKuoD5Init,
                dim176JoeKuoD5Init,
                dim177JoeKuoD5Init,
                dim178JoeKuoD5Init,
                dim179JoeKuoD5Init,
                dim180JoeKuoD5Init,
                dim181JoeKuoD5Init,
                dim182JoeKuoD5Init,
                dim183JoeKuoD5Init,
                dim184JoeKuoD5Init,
                dim185JoeKuoD5Init,
                dim186JoeKuoD5Init,
                dim187JoeKuoD5Init,
                dim188JoeKuoD5Init,
                dim189JoeKuoD5Init,
                dim190JoeKuoD5Init,
                dim191JoeKuoD5Init,
                dim192JoeKuoD5Init,
                dim193JoeKuoD5Init,
                dim194JoeKuoD5Init,
                dim195JoeKuoD5Init,
                dim196JoeKuoD5Init,
                dim197JoeKuoD5Init,
                dim198JoeKuoD5Init,
                dim199JoeKuoD5Init,
                dim200JoeKuoD5Init,
                dim201JoeKuoD5Init,
                dim202JoeKuoD5Init,
                dim203JoeKuoD5Init,
                dim204JoeKuoD5Init,
                dim205JoeKuoD5Init,
                dim206JoeKuoD5Init,
                dim207JoeKuoD5Init,
                dim208JoeKuoD5Init,
                dim209JoeKuoD5Init,
                dim210JoeKuoD5Init,
                dim211JoeKuoD5Init,
                dim212JoeKuoD5Init,
                dim213JoeKuoD5Init,
                dim214JoeKuoD5Init,
                dim215JoeKuoD5Init,
                dim216JoeKuoD5Init,
                dim217JoeKuoD5Init,
                dim218JoeKuoD5Init,
                dim219JoeKuoD5Init,
                dim220JoeKuoD5Init,
                dim221JoeKuoD5Init,
                dim222JoeKuoD5Init,
                dim223JoeKuoD5Init,
                dim224JoeKuoD5Init,
                dim225JoeKuoD5Init,
                dim226JoeKuoD5Init,
                dim227JoeKuoD5Init,
                dim228JoeKuoD5Init,
                dim229JoeKuoD5Init,
                dim230JoeKuoD5Init,
                dim231JoeKuoD5Init,
                dim232JoeKuoD5Init,
                dim233JoeKuoD5Init,
                dim234JoeKuoD5Init,
                dim235JoeKuoD5Init,
                dim236JoeKuoD5Init,
                dim237JoeKuoD5Init,
                dim238JoeKuoD5Init,
                dim239JoeKuoD5Init,
                dim240JoeKuoD5Init,
                dim241JoeKuoD5Init,
                dim242JoeKuoD5Init,
                dim243JoeKuoD5Init,
                dim244JoeKuoD5Init,
                dim245JoeKuoD5Init,
                dim246JoeKuoD5Init,
                dim247JoeKuoD5Init,
                dim248JoeKuoD5Init,
                dim249JoeKuoD5Init,
                dim250JoeKuoD5Init,
                dim251JoeKuoD5Init,
                dim252JoeKuoD5Init,
                dim253JoeKuoD5Init,
                dim254JoeKuoD5Init,
                dim255JoeKuoD5Init,
                dim256JoeKuoD5Init,
                dim257JoeKuoD5Init,
                dim258JoeKuoD5Init,
                dim259JoeKuoD5Init,
                dim260JoeKuoD5Init,
                dim261JoeKuoD5Init,
                dim262JoeKuoD5Init,
                dim263JoeKuoD5Init,
                dim264JoeKuoD5Init,
                dim265JoeKuoD5Init,
                dim266JoeKuoD5Init,
                dim267JoeKuoD5Init,
                dim268JoeKuoD5Init,
                dim269JoeKuoD5Init,
                dim270JoeKuoD5Init,
                dim271JoeKuoD5Init,
                dim272JoeKuoD5Init,
                dim273JoeKuoD5Init,
                dim274JoeKuoD5Init,
                dim275JoeKuoD5Init,
                dim276JoeKuoD5Init,
                dim277JoeKuoD5Init,
                dim278JoeKuoD5Init,
                dim279JoeKuoD5Init,
                dim280JoeKuoD5Init,
                dim281JoeKuoD5Init,
                dim282JoeKuoD5Init,
                dim283JoeKuoD5Init,
                dim284JoeKuoD5Init,
                dim285JoeKuoD5Init,
                dim286JoeKuoD5Init,
                dim287JoeKuoD5Init,
                dim288JoeKuoD5Init,
                dim289JoeKuoD5Init,
                dim290JoeKuoD5Init,
                dim291JoeKuoD5Init,
                dim292JoeKuoD5Init,
                dim293JoeKuoD5Init,
                dim294JoeKuoD5Init,
                dim295JoeKuoD5Init,
                dim296JoeKuoD5Init,
                dim297JoeKuoD5Init,
                dim298JoeKuoD5Init,
                dim299JoeKuoD5Init,
                dim300JoeKuoD5Init,
                dim301JoeKuoD5Init,
                dim302JoeKuoD5Init,
                dim303JoeKuoD5Init,
                dim304JoeKuoD5Init,
                dim305JoeKuoD5Init,
                dim306JoeKuoD5Init,
                dim307JoeKuoD5Init,
                dim308JoeKuoD5Init,
                dim309JoeKuoD5Init,
                dim310JoeKuoD5Init,
                dim311JoeKuoD5Init,
                dim312JoeKuoD5Init,
                dim313JoeKuoD5Init,
                dim314JoeKuoD5Init,
                dim315JoeKuoD5Init,
                dim316JoeKuoD5Init,
                dim317JoeKuoD5Init,
                dim318JoeKuoD5Init,
                dim319JoeKuoD5Init,
                dim320JoeKuoD5Init,
                dim321JoeKuoD5Init,
                dim322JoeKuoD5Init,
                dim323JoeKuoD5Init,
                dim324JoeKuoD5Init,
                dim325JoeKuoD5Init,
                dim326JoeKuoD5Init,
                dim327JoeKuoD5Init,
                dim328JoeKuoD5Init,
                dim329JoeKuoD5Init,
                dim330JoeKuoD5Init,
                dim331JoeKuoD5Init,
                dim332JoeKuoD5Init,
                dim333JoeKuoD5Init,
                dim334JoeKuoD5Init,
                dim335JoeKuoD5Init,
                dim336JoeKuoD5Init,
                dim337JoeKuoD5Init,
                dim338JoeKuoD5Init,
                dim339JoeKuoD5Init,
                dim340JoeKuoD5Init,
                dim341JoeKuoD5Init,
                dim342JoeKuoD5Init,
                dim343JoeKuoD5Init,
                dim344JoeKuoD5Init,
                dim345JoeKuoD5Init,
                dim346JoeKuoD5Init,
                dim347JoeKuoD5Init,
                dim348JoeKuoD5Init,
                dim349JoeKuoD5Init,
                dim350JoeKuoD5Init,
                dim351JoeKuoD5Init,
                dim352JoeKuoD5Init,
                dim353JoeKuoD5Init,
                dim354JoeKuoD5Init,
                dim355JoeKuoD5Init,
                dim356JoeKuoD5Init,
                dim357JoeKuoD5Init,
                dim358JoeKuoD5Init,
                dim359JoeKuoD5Init,
                dim360JoeKuoD5Init,
                dim361JoeKuoD5Init,
                dim362JoeKuoD5Init,
                dim363JoeKuoD5Init,
                dim364JoeKuoD5Init,
                dim365JoeKuoD5Init,
                dim366JoeKuoD5Init,
                dim367JoeKuoD5Init,
                dim368JoeKuoD5Init,
                dim369JoeKuoD5Init,
                dim370JoeKuoD5Init,
                dim371JoeKuoD5Init,
                dim372JoeKuoD5Init,
                dim373JoeKuoD5Init,
                dim374JoeKuoD5Init,
                dim375JoeKuoD5Init,
                dim376JoeKuoD5Init,
                dim377JoeKuoD5Init,
                dim378JoeKuoD5Init,
                dim379JoeKuoD5Init,
                dim380JoeKuoD5Init,
                dim381JoeKuoD5Init,
                dim382JoeKuoD5Init,
                dim383JoeKuoD5Init,
                dim384JoeKuoD5Init,
                dim385JoeKuoD5Init,
                dim386JoeKuoD5Init,
                dim387JoeKuoD5Init,
                dim388JoeKuoD5Init,
                dim389JoeKuoD5Init,
                dim390JoeKuoD5Init,
                dim391JoeKuoD5Init,
                dim392JoeKuoD5Init,
                dim393JoeKuoD5Init,
                dim394JoeKuoD5Init,
                dim395JoeKuoD5Init,
                dim396JoeKuoD5Init,
                dim397JoeKuoD5Init,
                dim398JoeKuoD5Init,
                dim399JoeKuoD5Init,
                dim400JoeKuoD5Init,
                dim401JoeKuoD5Init,
                dim402JoeKuoD5Init,
                dim403JoeKuoD5Init,
                dim404JoeKuoD5Init,
                dim405JoeKuoD5Init,
                dim406JoeKuoD5Init,
                dim407JoeKuoD5Init,
                dim408JoeKuoD5Init,
                dim409JoeKuoD5Init,
                dim410JoeKuoD5Init,
                dim411JoeKuoD5Init,
                dim412JoeKuoD5Init,
                dim413JoeKuoD5Init,
                dim414JoeKuoD5Init,
                dim415JoeKuoD5Init,
                dim416JoeKuoD5Init,
                dim417JoeKuoD5Init,
                dim418JoeKuoD5Init,
                dim419JoeKuoD5Init,
                dim420JoeKuoD5Init,
                dim421JoeKuoD5Init,
                dim422JoeKuoD5Init,
                dim423JoeKuoD5Init,
                dim424JoeKuoD5Init,
                dim425JoeKuoD5Init,
                dim426JoeKuoD5Init,
                dim427JoeKuoD5Init,
                dim428JoeKuoD5Init,
                dim429JoeKuoD5Init,
                dim430JoeKuoD5Init,
                dim431JoeKuoD5Init,
                dim432JoeKuoD5Init,
                dim433JoeKuoD5Init,
                dim434JoeKuoD5Init,
                dim435JoeKuoD5Init,
                dim436JoeKuoD5Init,
                dim437JoeKuoD5Init,
                dim438JoeKuoD5Init,
                dim439JoeKuoD5Init,
                dim440JoeKuoD5Init,
                dim441JoeKuoD5Init,
                dim442JoeKuoD5Init,
                dim443JoeKuoD5Init,
                dim444JoeKuoD5Init,
                dim445JoeKuoD5Init,
                dim446JoeKuoD5Init,
                dim447JoeKuoD5Init,
                dim448JoeKuoD5Init,
                dim449JoeKuoD5Init,
                dim450JoeKuoD5Init,
                dim451JoeKuoD5Init,
                dim452JoeKuoD5Init,
                dim453JoeKuoD5Init,
                dim454JoeKuoD5Init,
                dim455JoeKuoD5Init,
                dim456JoeKuoD5Init,
                dim457JoeKuoD5Init,
                dim458JoeKuoD5Init,
                dim459JoeKuoD5Init,
                dim460JoeKuoD5Init,
                dim461JoeKuoD5Init,
                dim462JoeKuoD5Init,
                dim463JoeKuoD5Init,
                dim464JoeKuoD5Init,
                dim465JoeKuoD5Init,
                dim466JoeKuoD5Init,
                dim467JoeKuoD5Init,
                dim468JoeKuoD5Init,
                dim469JoeKuoD5Init,
                dim470JoeKuoD5Init,
                dim471JoeKuoD5Init,
                dim472JoeKuoD5Init,
                dim473JoeKuoD5Init,
                dim474JoeKuoD5Init,
                dim475JoeKuoD5Init,
                dim476JoeKuoD5Init,
                dim477JoeKuoD5Init,
                dim478JoeKuoD5Init,
                dim479JoeKuoD5Init,
                dim480JoeKuoD5Init,
                dim481JoeKuoD5Init,
                dim482JoeKuoD5Init,
                dim483JoeKuoD5Init,
                dim484JoeKuoD5Init,
                dim485JoeKuoD5Init,
                dim486JoeKuoD5Init,
                dim487JoeKuoD5Init,
                dim488JoeKuoD5Init,
                dim489JoeKuoD5Init,
                dim490JoeKuoD5Init,
                dim491JoeKuoD5Init,
                dim492JoeKuoD5Init,
                dim493JoeKuoD5Init,
                dim494JoeKuoD5Init,
                dim495JoeKuoD5Init,
                dim496JoeKuoD5Init,
                dim497JoeKuoD5Init,
                dim498JoeKuoD5Init,
                dim499JoeKuoD5Init,
                dim500JoeKuoD5Init,
                dim501JoeKuoD5Init,
                dim502JoeKuoD5Init,
                dim503JoeKuoD5Init,
                dim504JoeKuoD5Init,
                dim505JoeKuoD5Init,
                dim506JoeKuoD5Init,
                dim507JoeKuoD5Init,
                dim508JoeKuoD5Init,
                dim509JoeKuoD5Init,
                dim510JoeKuoD5Init,
                dim511JoeKuoD5Init,
                dim512JoeKuoD5Init,
                dim513JoeKuoD5Init,
                dim514JoeKuoD5Init,
                dim515JoeKuoD5Init,
                dim516JoeKuoD5Init,
                dim517JoeKuoD5Init,
                dim518JoeKuoD5Init,
                dim519JoeKuoD5Init,
                dim520JoeKuoD5Init,
                dim521JoeKuoD5Init,
                dim522JoeKuoD5Init,
                dim523JoeKuoD5Init,
                dim524JoeKuoD5Init,
                dim525JoeKuoD5Init,
                dim526JoeKuoD5Init,
                dim527JoeKuoD5Init,
                dim528JoeKuoD5Init,
                dim529JoeKuoD5Init,
                dim530JoeKuoD5Init,
                dim531JoeKuoD5Init,
                dim532JoeKuoD5Init,
                dim533JoeKuoD5Init,
                dim534JoeKuoD5Init,
                dim535JoeKuoD5Init,
                dim536JoeKuoD5Init,
                dim537JoeKuoD5Init,
                dim538JoeKuoD5Init,
                dim539JoeKuoD5Init,
                dim540JoeKuoD5Init,
                dim541JoeKuoD5Init,
                dim542JoeKuoD5Init,
                dim543JoeKuoD5Init,
                dim544JoeKuoD5Init,
                dim545JoeKuoD5Init,
                dim546JoeKuoD5Init,
                dim547JoeKuoD5Init,
                dim548JoeKuoD5Init,
                dim549JoeKuoD5Init,
                dim550JoeKuoD5Init,
                dim551JoeKuoD5Init,
                dim552JoeKuoD5Init,
                dim553JoeKuoD5Init,
                dim554JoeKuoD5Init,
                dim555JoeKuoD5Init,
                dim556JoeKuoD5Init,
                dim557JoeKuoD5Init,
                dim558JoeKuoD5Init,
                dim559JoeKuoD5Init,
                dim560JoeKuoD5Init,
                dim561JoeKuoD5Init,
                dim562JoeKuoD5Init,
                dim563JoeKuoD5Init,
                dim564JoeKuoD5Init,
                dim565JoeKuoD5Init,
                dim566JoeKuoD5Init,
                dim567JoeKuoD5Init,
                dim568JoeKuoD5Init,
                dim569JoeKuoD5Init,
                dim570JoeKuoD5Init,
                dim571JoeKuoD5Init,
                dim572JoeKuoD5Init,
                dim573JoeKuoD5Init,
                dim574JoeKuoD5Init,
                dim575JoeKuoD5Init,
                dim576JoeKuoD5Init,
                dim577JoeKuoD5Init,
                dim578JoeKuoD5Init,
                dim579JoeKuoD5Init,
                dim580JoeKuoD5Init,
                dim581JoeKuoD5Init,
                dim582JoeKuoD5Init,
                dim583JoeKuoD5Init,
                dim584JoeKuoD5Init,
                dim585JoeKuoD5Init,
                dim586JoeKuoD5Init,
                dim587JoeKuoD5Init,
                dim588JoeKuoD5Init,
                dim589JoeKuoD5Init,
                dim590JoeKuoD5Init,
                dim591JoeKuoD5Init,
                dim592JoeKuoD5Init,
                dim593JoeKuoD5Init,
                dim594JoeKuoD5Init,
                dim595JoeKuoD5Init,
                dim596JoeKuoD5Init,
                dim597JoeKuoD5Init,
                dim598JoeKuoD5Init,
                dim599JoeKuoD5Init,
                dim600JoeKuoD5Init,
                dim601JoeKuoD5Init,
                dim602JoeKuoD5Init,
                dim603JoeKuoD5Init,
                dim604JoeKuoD5Init,
                dim605JoeKuoD5Init,
                dim606JoeKuoD5Init,
                dim607JoeKuoD5Init,
                dim608JoeKuoD5Init,
                dim609JoeKuoD5Init,
                dim610JoeKuoD5Init,
                dim611JoeKuoD5Init,
                dim612JoeKuoD5Init,
                dim613JoeKuoD5Init,
                dim614JoeKuoD5Init,
                dim615JoeKuoD5Init,
                dim616JoeKuoD5Init,
                dim617JoeKuoD5Init,
                dim618JoeKuoD5Init,
                dim619JoeKuoD5Init,
                dim620JoeKuoD5Init,
                dim621JoeKuoD5Init,
                dim622JoeKuoD5Init,
                dim623JoeKuoD5Init,
                dim624JoeKuoD5Init,
                dim625JoeKuoD5Init,
                dim626JoeKuoD5Init,
                dim627JoeKuoD5Init,
                dim628JoeKuoD5Init,
                dim629JoeKuoD5Init,
                dim630JoeKuoD5Init,
                dim631JoeKuoD5Init,
                dim632JoeKuoD5Init,
                dim633JoeKuoD5Init,
                dim634JoeKuoD5Init,
                dim635JoeKuoD5Init,
                dim636JoeKuoD5Init,
                dim637JoeKuoD5Init,
                dim638JoeKuoD5Init,
                dim639JoeKuoD5Init,
                dim640JoeKuoD5Init,
                dim641JoeKuoD5Init,
                dim642JoeKuoD5Init,
                dim643JoeKuoD5Init,
                dim644JoeKuoD5Init,
                dim645JoeKuoD5Init,
                dim646JoeKuoD5Init,
                dim647JoeKuoD5Init,
                dim648JoeKuoD5Init,
                dim649JoeKuoD5Init,
                dim650JoeKuoD5Init,
                dim651JoeKuoD5Init,
                dim652JoeKuoD5Init,
                dim653JoeKuoD5Init,
                dim654JoeKuoD5Init,
                dim655JoeKuoD5Init,
                dim656JoeKuoD5Init,
                dim657JoeKuoD5Init,
                dim658JoeKuoD5Init,
                dim659JoeKuoD5Init,
                dim660JoeKuoD5Init,
                dim661JoeKuoD5Init,
                dim662JoeKuoD5Init,
                dim663JoeKuoD5Init,
                dim664JoeKuoD5Init,
                dim665JoeKuoD5Init,
                dim666JoeKuoD5Init,
                dim667JoeKuoD5Init,
                dim668JoeKuoD5Init,
                dim669JoeKuoD5Init,
                dim670JoeKuoD5Init,
                dim671JoeKuoD5Init,
                dim672JoeKuoD5Init,
                dim673JoeKuoD5Init,
                dim674JoeKuoD5Init,
                dim675JoeKuoD5Init,
                dim676JoeKuoD5Init,
                dim677JoeKuoD5Init,
                dim678JoeKuoD5Init,
                dim679JoeKuoD5Init,
                dim680JoeKuoD5Init,
                dim681JoeKuoD5Init,
                dim682JoeKuoD5Init,
                dim683JoeKuoD5Init,
                dim684JoeKuoD5Init,
                dim685JoeKuoD5Init,
                dim686JoeKuoD5Init,
                dim687JoeKuoD5Init,
                dim688JoeKuoD5Init,
                dim689JoeKuoD5Init,
                dim690JoeKuoD5Init,
                dim691JoeKuoD5Init,
                dim692JoeKuoD5Init,
                dim693JoeKuoD5Init,
                dim694JoeKuoD5Init,
                dim695JoeKuoD5Init,
                dim696JoeKuoD5Init,
                dim697JoeKuoD5Init,
                dim698JoeKuoD5Init,
                dim699JoeKuoD5Init,
                dim700JoeKuoD5Init,
                dim701JoeKuoD5Init,
                dim702JoeKuoD5Init,
                dim703JoeKuoD5Init,
                dim704JoeKuoD5Init,
                dim705JoeKuoD5Init,
                dim706JoeKuoD5Init,
                dim707JoeKuoD5Init,
                dim708JoeKuoD5Init,
                dim709JoeKuoD5Init,
                dim710JoeKuoD5Init,
                dim711JoeKuoD5Init,
                dim712JoeKuoD5Init,
                dim713JoeKuoD5Init,
                dim714JoeKuoD5Init,
                dim715JoeKuoD5Init,
                dim716JoeKuoD5Init,
                dim717JoeKuoD5Init,
                dim718JoeKuoD5Init,
                dim719JoeKuoD5Init,
                dim720JoeKuoD5Init,
                dim721JoeKuoD5Init,
                dim722JoeKuoD5Init,
                dim723JoeKuoD5Init,
                dim724JoeKuoD5Init,
                dim725JoeKuoD5Init,
                dim726JoeKuoD5Init,
                dim727JoeKuoD5Init,
                dim728JoeKuoD5Init,
                dim729JoeKuoD5Init,
                dim730JoeKuoD5Init,
                dim731JoeKuoD5Init,
                dim732JoeKuoD5Init,
                dim733JoeKuoD5Init,
                dim734JoeKuoD5Init,
                dim735JoeKuoD5Init,
                dim736JoeKuoD5Init,
                dim737JoeKuoD5Init,
                dim738JoeKuoD5Init,
                dim739JoeKuoD5Init,
                dim740JoeKuoD5Init,
                dim741JoeKuoD5Init,
                dim742JoeKuoD5Init,
                dim743JoeKuoD5Init,
                dim744JoeKuoD5Init,
                dim745JoeKuoD5Init,
                dim746JoeKuoD5Init,
                dim747JoeKuoD5Init,
                dim748JoeKuoD5Init,
                dim749JoeKuoD5Init,
                dim750JoeKuoD5Init,
                dim751JoeKuoD5Init,
                dim752JoeKuoD5Init,
                dim753JoeKuoD5Init,
                dim754JoeKuoD5Init,
                dim755JoeKuoD5Init,
                dim756JoeKuoD5Init,
                dim757JoeKuoD5Init,
                dim758JoeKuoD5Init,
                dim759JoeKuoD5Init,
                dim760JoeKuoD5Init,
                dim761JoeKuoD5Init,
                dim762JoeKuoD5Init,
                dim763JoeKuoD5Init,
                dim764JoeKuoD5Init,
                dim765JoeKuoD5Init,
                dim766JoeKuoD5Init,
                dim767JoeKuoD5Init,
                dim768JoeKuoD5Init,
                dim769JoeKuoD5Init,
                dim770JoeKuoD5Init,
                dim771JoeKuoD5Init,
                dim772JoeKuoD5Init,
                dim773JoeKuoD5Init,
                dim774JoeKuoD5Init,
                dim775JoeKuoD5Init,
                dim776JoeKuoD5Init,
                dim777JoeKuoD5Init,
                dim778JoeKuoD5Init,
                dim779JoeKuoD5Init,
                dim780JoeKuoD5Init,
                dim781JoeKuoD5Init,
                dim782JoeKuoD5Init,
                dim783JoeKuoD5Init,
                dim784JoeKuoD5Init,
                dim785JoeKuoD5Init,
                dim786JoeKuoD5Init,
                dim787JoeKuoD5Init,
                dim788JoeKuoD5Init,
                dim789JoeKuoD5Init,
                dim790JoeKuoD5Init,
                dim791JoeKuoD5Init,
                dim792JoeKuoD5Init,
                dim793JoeKuoD5Init,
                dim794JoeKuoD5Init,
                dim795JoeKuoD5Init,
                dim796JoeKuoD5Init,
                dim797JoeKuoD5Init,
                dim798JoeKuoD5Init,
                dim799JoeKuoD5Init,
                dim800JoeKuoD5Init,
                dim801JoeKuoD5Init,
                dim802JoeKuoD5Init,
                dim803JoeKuoD5Init,
                dim804JoeKuoD5Init,
                dim805JoeKuoD5Init,
                dim806JoeKuoD5Init,
                dim807JoeKuoD5Init,
                dim808JoeKuoD5Init,
                dim809JoeKuoD5Init,
                dim810JoeKuoD5Init,
                dim811JoeKuoD5Init,
                dim812JoeKuoD5Init,
                dim813JoeKuoD5Init,
                dim814JoeKuoD5Init,
                dim815JoeKuoD5Init,
                dim816JoeKuoD5Init,
                dim817JoeKuoD5Init,
                dim818JoeKuoD5Init,
                dim819JoeKuoD5Init,
                dim820JoeKuoD5Init,
                dim821JoeKuoD5Init,
                dim822JoeKuoD5Init,
                dim823JoeKuoD5Init,
                dim824JoeKuoD5Init,
                dim825JoeKuoD5Init,
                dim826JoeKuoD5Init,
                dim827JoeKuoD5Init,
                dim828JoeKuoD5Init,
                dim829JoeKuoD5Init,
                dim830JoeKuoD5Init,
                dim831JoeKuoD5Init,
                dim832JoeKuoD5Init,
                dim833JoeKuoD5Init,
                dim834JoeKuoD5Init,
                dim835JoeKuoD5Init,
                dim836JoeKuoD5Init,
                dim837JoeKuoD5Init,
                dim838JoeKuoD5Init,
                dim839JoeKuoD5Init,
                dim840JoeKuoD5Init,
                dim841JoeKuoD5Init,
                dim842JoeKuoD5Init,
                dim843JoeKuoD5Init,
                dim844JoeKuoD5Init,
                dim845JoeKuoD5Init,
                dim846JoeKuoD5Init,
                dim847JoeKuoD5Init,
                dim848JoeKuoD5Init,
                dim849JoeKuoD5Init,
                dim850JoeKuoD5Init,
                dim851JoeKuoD5Init,
                dim852JoeKuoD5Init,
                dim853JoeKuoD5Init,
                dim854JoeKuoD5Init,
                dim855JoeKuoD5Init,
                dim856JoeKuoD5Init,
                dim857JoeKuoD5Init,
                dim858JoeKuoD5Init,
                dim859JoeKuoD5Init,
                dim860JoeKuoD5Init,
                dim861JoeKuoD5Init,
                dim862JoeKuoD5Init,
                dim863JoeKuoD5Init,
                dim864JoeKuoD5Init,
                dim865JoeKuoD5Init,
                dim866JoeKuoD5Init,
                dim867JoeKuoD5Init,
                dim868JoeKuoD5Init,
                dim869JoeKuoD5Init,
                dim870JoeKuoD5Init,
                dim871JoeKuoD5Init,
                dim872JoeKuoD5Init,
                dim873JoeKuoD5Init,
                dim874JoeKuoD5Init,
                dim875JoeKuoD5Init,
                dim876JoeKuoD5Init,
                dim877JoeKuoD5Init,
                dim878JoeKuoD5Init,
                dim879JoeKuoD5Init,
                dim880JoeKuoD5Init,
                dim881JoeKuoD5Init,
                dim882JoeKuoD5Init,
                dim883JoeKuoD5Init,
                dim884JoeKuoD5Init,
                dim885JoeKuoD5Init,
                dim886JoeKuoD5Init,
                dim887JoeKuoD5Init,
                dim888JoeKuoD5Init,
                dim889JoeKuoD5Init,
                dim890JoeKuoD5Init,
                dim891JoeKuoD5Init,
                dim892JoeKuoD5Init,
                dim893JoeKuoD5Init,
                dim894JoeKuoD5Init,
                dim895JoeKuoD5Init,
                dim896JoeKuoD5Init,
                dim897JoeKuoD5Init,
                dim898JoeKuoD5Init,
                dim899JoeKuoD5Init,
                dim900JoeKuoD5Init,
                dim901JoeKuoD5Init,
                dim902JoeKuoD5Init,
                dim903JoeKuoD5Init,
                dim904JoeKuoD5Init,
                dim905JoeKuoD5Init,
                dim906JoeKuoD5Init,
                dim907JoeKuoD5Init,
                dim908JoeKuoD5Init,
                dim909JoeKuoD5Init,
                dim910JoeKuoD5Init,
                dim911JoeKuoD5Init,
                dim912JoeKuoD5Init,
                dim913JoeKuoD5Init,
                dim914JoeKuoD5Init,
                dim915JoeKuoD5Init,
                dim916JoeKuoD5Init,
                dim917JoeKuoD5Init,
                dim918JoeKuoD5Init,
                dim919JoeKuoD5Init,
                dim920JoeKuoD5Init,
                dim921JoeKuoD5Init,
                dim922JoeKuoD5Init,
                dim923JoeKuoD5Init,
                dim924JoeKuoD5Init,
                dim925JoeKuoD5Init,
                dim926JoeKuoD5Init,
                dim927JoeKuoD5Init,
                dim928JoeKuoD5Init,
                dim929JoeKuoD5Init,
                dim930JoeKuoD5Init,
                dim931JoeKuoD5Init,
                dim932JoeKuoD5Init,
                dim933JoeKuoD5Init,
                dim934JoeKuoD5Init,
                dim935JoeKuoD5Init,
                dim936JoeKuoD5Init,
                dim937JoeKuoD5Init,
                dim938JoeKuoD5Init,
                dim939JoeKuoD5Init,
                dim940JoeKuoD5Init,
                dim941JoeKuoD5Init,
                dim942JoeKuoD5Init,
                dim943JoeKuoD5Init,
                dim944JoeKuoD5Init,
                dim945JoeKuoD5Init,
                dim946JoeKuoD5Init,
                dim947JoeKuoD5Init,
                dim948JoeKuoD5Init,
                dim949JoeKuoD5Init,
                dim950JoeKuoD5Init,
                dim951JoeKuoD5Init,
                dim952JoeKuoD5Init,
                dim953JoeKuoD5Init,
                dim954JoeKuoD5Init,
                dim955JoeKuoD5Init,
                dim956JoeKuoD5Init,
                dim957JoeKuoD5Init,
                dim958JoeKuoD5Init,
                dim959JoeKuoD5Init,
                dim960JoeKuoD5Init,
                dim961JoeKuoD5Init,
                dim962JoeKuoD5Init,
                dim963JoeKuoD5Init,
                dim964JoeKuoD5Init,
                dim965JoeKuoD5Init,
                dim966JoeKuoD5Init,
                dim967JoeKuoD5Init,
                dim968JoeKuoD5Init,
                dim969JoeKuoD5Init,
                dim970JoeKuoD5Init,
                dim971JoeKuoD5Init,
                dim972JoeKuoD5Init,
                dim973JoeKuoD5Init,
                dim974JoeKuoD5Init,
                dim975JoeKuoD5Init,
                dim976JoeKuoD5Init,
                dim977JoeKuoD5Init,
                dim978JoeKuoD5Init,
                dim979JoeKuoD5Init,
                dim980JoeKuoD5Init,
                dim981JoeKuoD5Init,
                dim982JoeKuoD5Init,
                dim983JoeKuoD5Init,
                dim984JoeKuoD5Init,
                dim985JoeKuoD5Init,
                dim986JoeKuoD5Init,
                dim987JoeKuoD5Init,
                dim988JoeKuoD5Init,
                dim989JoeKuoD5Init,
                dim990JoeKuoD5Init,
                dim991JoeKuoD5Init,
                dim992JoeKuoD5Init,
                dim993JoeKuoD5Init,
                dim994JoeKuoD5Init,
                dim995JoeKuoD5Init,
                dim996JoeKuoD5Init,
                dim997JoeKuoD5Init,
                dim998JoeKuoD5Init,
                dim999JoeKuoD5Init,
                dim1000JoeKuoD5Init,
                dim1001JoeKuoD5Init,
                dim1002JoeKuoD5Init,
                dim1003JoeKuoD5Init,
                dim1004JoeKuoD5Init,
                dim1005JoeKuoD5Init,
                dim1006JoeKuoD5Init,
                dim1007JoeKuoD5Init,
                dim1008JoeKuoD5Init,
                dim1009JoeKuoD5Init,
                dim1010JoeKuoD5Init,
                dim1011JoeKuoD5Init,
                dim1012JoeKuoD5Init,
                dim1013JoeKuoD5Init,
                dim1014JoeKuoD5Init,
                dim1015JoeKuoD5Init,
                dim1016JoeKuoD5Init,
                dim1017JoeKuoD5Init,
                dim1018JoeKuoD5Init,
                dim1019JoeKuoD5Init,
                dim1020JoeKuoD5Init,
                dim1021JoeKuoD5Init,
                dim1022JoeKuoD5Init,
                dim1023JoeKuoD5Init,
                dim1024JoeKuoD5Init,
                dim1025JoeKuoD5Init,
                dim1026JoeKuoD5Init,
                dim1027JoeKuoD5Init,
                dim1028JoeKuoD5Init,
                dim1029JoeKuoD5Init,
                dim1030JoeKuoD5Init,
                dim1031JoeKuoD5Init,
                dim1032JoeKuoD5Init,
                dim1033JoeKuoD5Init,
                dim1034JoeKuoD5Init,
                dim1035JoeKuoD5Init,
                dim1036JoeKuoD5Init,
                dim1037JoeKuoD5Init,
                dim1038JoeKuoD5Init,
                dim1039JoeKuoD5Init,
                dim1040JoeKuoD5Init,
                dim1041JoeKuoD5Init,
                dim1042JoeKuoD5Init,
                dim1043JoeKuoD5Init,
                dim1044JoeKuoD5Init,
                dim1045JoeKuoD5Init,
                dim1046JoeKuoD5Init,
                dim1047JoeKuoD5Init,
                dim1048JoeKuoD5Init,
                dim1049JoeKuoD5Init,
                dim1050JoeKuoD5Init,
                dim1051JoeKuoD5Init,
                dim1052JoeKuoD5Init,
                dim1053JoeKuoD5Init,
                dim1054JoeKuoD5Init,
                dim1055JoeKuoD5Init,
                dim1056JoeKuoD5Init,
                dim1057JoeKuoD5Init,
                dim1058JoeKuoD5Init,
                dim1059JoeKuoD5Init,
                dim1060JoeKuoD5Init,
                dim1061JoeKuoD5Init,
                dim1062JoeKuoD5Init,
                dim1063JoeKuoD5Init,
                dim1064JoeKuoD5Init,
                dim1065JoeKuoD5Init,
                dim1066JoeKuoD5Init,
                dim1067JoeKuoD5Init,
                dim1068JoeKuoD5Init,
                dim1069JoeKuoD5Init,
                dim1070JoeKuoD5Init,
                dim1071JoeKuoD5Init,
                dim1072JoeKuoD5Init,
                dim1073JoeKuoD5Init,
                dim1074JoeKuoD5Init,
                dim1075JoeKuoD5Init,
                dim1076JoeKuoD5Init,
                dim1077JoeKuoD5Init,
                dim1078JoeKuoD5Init,
                dim1079JoeKuoD5Init,
                dim1080JoeKuoD5Init,
                dim1081JoeKuoD5Init,
                dim1082JoeKuoD5Init,
                dim1083JoeKuoD5Init,
                dim1084JoeKuoD5Init,
                dim1085JoeKuoD5Init,
                dim1086JoeKuoD5Init,
                dim1087JoeKuoD5Init,
                dim1088JoeKuoD5Init,
                dim1089JoeKuoD5Init,
                dim1090JoeKuoD5Init,
                dim1091JoeKuoD5Init,
                dim1092JoeKuoD5Init,
                dim1093JoeKuoD5Init,
                dim1094JoeKuoD5Init,
                dim1095JoeKuoD5Init,
                dim1096JoeKuoD5Init,
                dim1097JoeKuoD5Init,
                dim1098JoeKuoD5Init,
                dim1099JoeKuoD5Init,
                dim1100JoeKuoD5Init,
                dim1101JoeKuoD5Init,
                dim1102JoeKuoD5Init,
                dim1103JoeKuoD5Init,
                dim1104JoeKuoD5Init,
                dim1105JoeKuoD5Init,
                dim1106JoeKuoD5Init,
                dim1107JoeKuoD5Init,
                dim1108JoeKuoD5Init,
                dim1109JoeKuoD5Init,
                dim1110JoeKuoD5Init,
                dim1111JoeKuoD5Init,
                dim1112JoeKuoD5Init,
                dim1113JoeKuoD5Init,
                dim1114JoeKuoD5Init,
                dim1115JoeKuoD5Init,
                dim1116JoeKuoD5Init,
                dim1117JoeKuoD5Init,
                dim1118JoeKuoD5Init,
                dim1119JoeKuoD5Init,
                dim1120JoeKuoD5Init,
                dim1121JoeKuoD5Init,
                dim1122JoeKuoD5Init,
                dim1123JoeKuoD5Init,
                dim1124JoeKuoD5Init,
                dim1125JoeKuoD5Init,
                dim1126JoeKuoD5Init,
                dim1127JoeKuoD5Init,
                dim1128JoeKuoD5Init,
                dim1129JoeKuoD5Init,
                dim1130JoeKuoD5Init,
                dim1131JoeKuoD5Init,
                dim1132JoeKuoD5Init,
                dim1133JoeKuoD5Init,
                dim1134JoeKuoD5Init,
                dim1135JoeKuoD5Init,
                dim1136JoeKuoD5Init,
                dim1137JoeKuoD5Init,
                dim1138JoeKuoD5Init,
                dim1139JoeKuoD5Init,
                dim1140JoeKuoD5Init,
                dim1141JoeKuoD5Init,
                dim1142JoeKuoD5Init,
                dim1143JoeKuoD5Init,
                dim1144JoeKuoD5Init,
                dim1145JoeKuoD5Init,
                dim1146JoeKuoD5Init,
                dim1147JoeKuoD5Init,
                dim1148JoeKuoD5Init,
                dim1149JoeKuoD5Init,
                dim1150JoeKuoD5Init,
                dim1151JoeKuoD5Init,
                dim1152JoeKuoD5Init,
                dim1153JoeKuoD5Init,
                dim1154JoeKuoD5Init,
                dim1155JoeKuoD5Init,
                dim1156JoeKuoD5Init,
                dim1157JoeKuoD5Init,
                dim1158JoeKuoD5Init,
                dim1159JoeKuoD5Init,
                dim1160JoeKuoD5Init,
                dim1161JoeKuoD5Init,
                dim1162JoeKuoD5Init,
                dim1163JoeKuoD5Init,
                dim1164JoeKuoD5Init,
                dim1165JoeKuoD5Init,
                dim1166JoeKuoD5Init,
                dim1167JoeKuoD5Init,
                dim1168JoeKuoD5Init,
                dim1169JoeKuoD5Init,
                dim1170JoeKuoD5Init,
                dim1171JoeKuoD5Init,
                dim1172JoeKuoD5Init,
                dim1173JoeKuoD5Init,
                dim1174JoeKuoD5Init,
                dim1175JoeKuoD5Init,
                dim1176JoeKuoD5Init,
                dim1177JoeKuoD5Init,
                dim1178JoeKuoD5Init,
                dim1179JoeKuoD5Init,
                dim1180JoeKuoD5Init,
                dim1181JoeKuoD5Init,
                dim1182JoeKuoD5Init,
                dim1183JoeKuoD5Init,
                dim1184JoeKuoD5Init,
                dim1185JoeKuoD5Init,
                dim1186JoeKuoD5Init,
                dim1187JoeKuoD5Init,
                dim1188JoeKuoD5Init,
                dim1189JoeKuoD5Init,
                dim1190JoeKuoD5Init,
                dim1191JoeKuoD5Init,
                dim1192JoeKuoD5Init,
                dim1193JoeKuoD5Init,
                dim1194JoeKuoD5Init,
                dim1195JoeKuoD5Init,
                dim1196JoeKuoD5Init,
                dim1197JoeKuoD5Init,
                dim1198JoeKuoD5Init,
                dim1199JoeKuoD5Init,
                dim1200JoeKuoD5Init,
                dim1201JoeKuoD5Init,
                dim1202JoeKuoD5Init,
                dim1203JoeKuoD5Init,
                dim1204JoeKuoD5Init,
                dim1205JoeKuoD5Init,
                dim1206JoeKuoD5Init,
                dim1207JoeKuoD5Init,
                dim1208JoeKuoD5Init,
                dim1209JoeKuoD5Init,
                dim1210JoeKuoD5Init,
                dim1211JoeKuoD5Init,
                dim1212JoeKuoD5Init,
                dim1213JoeKuoD5Init,
                dim1214JoeKuoD5Init,
                dim1215JoeKuoD5Init,
                dim1216JoeKuoD5Init,
                dim1217JoeKuoD5Init,
                dim1218JoeKuoD5Init,
                dim1219JoeKuoD5Init,
                dim1220JoeKuoD5Init,
                dim1221JoeKuoD5Init,
                dim1222JoeKuoD5Init,
                dim1223JoeKuoD5Init,
                dim1224JoeKuoD5Init,
                dim1225JoeKuoD5Init,
                dim1226JoeKuoD5Init,
                dim1227JoeKuoD5Init,
                dim1228JoeKuoD5Init,
                dim1229JoeKuoD5Init,
                dim1230JoeKuoD5Init,
                dim1231JoeKuoD5Init,
                dim1232JoeKuoD5Init,
                dim1233JoeKuoD5Init,
                dim1234JoeKuoD5Init,
                dim1235JoeKuoD5Init,
                dim1236JoeKuoD5Init,
                dim1237JoeKuoD5Init,
                dim1238JoeKuoD5Init,
                dim1239JoeKuoD5Init,
                dim1240JoeKuoD5Init,
                dim1241JoeKuoD5Init,
                dim1242JoeKuoD5Init,
                dim1243JoeKuoD5Init,
                dim1244JoeKuoD5Init,
                dim1245JoeKuoD5Init,
                dim1246JoeKuoD5Init,
                dim1247JoeKuoD5Init,
                dim1248JoeKuoD5Init,
                dim1249JoeKuoD5Init,
                dim1250JoeKuoD5Init,
                dim1251JoeKuoD5Init,
                dim1252JoeKuoD5Init,
                dim1253JoeKuoD5Init,
                dim1254JoeKuoD5Init,
                dim1255JoeKuoD5Init,
                dim1256JoeKuoD5Init,
                dim1257JoeKuoD5Init,
                dim1258JoeKuoD5Init,
                dim1259JoeKuoD5Init,
                dim1260JoeKuoD5Init,
                dim1261JoeKuoD5Init,
                dim1262JoeKuoD5Init,
                dim1263JoeKuoD5Init,
                dim1264JoeKuoD5Init,
                dim1265JoeKuoD5Init,
                dim1266JoeKuoD5Init,
                dim1267JoeKuoD5Init,
                dim1268JoeKuoD5Init,
                dim1269JoeKuoD5Init,
                dim1270JoeKuoD5Init,
                dim1271JoeKuoD5Init,
                dim1272JoeKuoD5Init,
                dim1273JoeKuoD5Init,
                dim1274JoeKuoD5Init,
                dim1275JoeKuoD5Init,
                dim1276JoeKuoD5Init,
                dim1277JoeKuoD5Init,
                dim1278JoeKuoD5Init,
                dim1279JoeKuoD5Init,
                dim1280JoeKuoD5Init,
                dim1281JoeKuoD5Init,
                dim1282JoeKuoD5Init,
                dim1283JoeKuoD5Init,
                dim1284JoeKuoD5Init,
                dim1285JoeKuoD5Init,
                dim1286JoeKuoD5Init,
                dim1287JoeKuoD5Init,
                dim1288JoeKuoD5Init,
                dim1289JoeKuoD5Init,
                dim1290JoeKuoD5Init,
                dim1291JoeKuoD5Init,
                dim1292JoeKuoD5Init,
                dim1293JoeKuoD5Init,
                dim1294JoeKuoD5Init,
                dim1295JoeKuoD5Init,
                dim1296JoeKuoD5Init,
                dim1297JoeKuoD5Init,
                dim1298JoeKuoD5Init,
                dim1299JoeKuoD5Init,
                dim1300JoeKuoD5Init,
                dim1301JoeKuoD5Init,
                dim1302JoeKuoD5Init,
                dim1303JoeKuoD5Init,
                dim1304JoeKuoD5Init,
                dim1305JoeKuoD5Init,
                dim1306JoeKuoD5Init,
                dim1307JoeKuoD5Init,
                dim1308JoeKuoD5Init,
                dim1309JoeKuoD5Init,
                dim1310JoeKuoD5Init,
                dim1311JoeKuoD5Init,
                dim1312JoeKuoD5Init,
                dim1313JoeKuoD5Init,
                dim1314JoeKuoD5Init,
                dim1315JoeKuoD5Init,
                dim1316JoeKuoD5Init,
                dim1317JoeKuoD5Init,
                dim1318JoeKuoD5Init,
                dim1319JoeKuoD5Init,
                dim1320JoeKuoD5Init,
                dim1321JoeKuoD5Init,
                dim1322JoeKuoD5Init,
                dim1323JoeKuoD5Init,
                dim1324JoeKuoD5Init,
                dim1325JoeKuoD5Init,
                dim1326JoeKuoD5Init,
                dim1327JoeKuoD5Init,
                dim1328JoeKuoD5Init,
                dim1329JoeKuoD5Init,
                dim1330JoeKuoD5Init,
                dim1331JoeKuoD5Init,
                dim1332JoeKuoD5Init,
                dim1333JoeKuoD5Init,
                dim1334JoeKuoD5Init,
                dim1335JoeKuoD5Init,
                dim1336JoeKuoD5Init,
                dim1337JoeKuoD5Init,
                dim1338JoeKuoD5Init,
                dim1339JoeKuoD5Init,
                dim1340JoeKuoD5Init,
                dim1341JoeKuoD5Init,
                dim1342JoeKuoD5Init,
                dim1343JoeKuoD5Init,
                dim1344JoeKuoD5Init,
                dim1345JoeKuoD5Init,
                dim1346JoeKuoD5Init,
                dim1347JoeKuoD5Init,
                dim1348JoeKuoD5Init,
                dim1349JoeKuoD5Init,
                dim1350JoeKuoD5Init,
                dim1351JoeKuoD5Init,
                dim1352JoeKuoD5Init,
                dim1353JoeKuoD5Init,
                dim1354JoeKuoD5Init,
                dim1355JoeKuoD5Init,
                dim1356JoeKuoD5Init,
                dim1357JoeKuoD5Init,
                dim1358JoeKuoD5Init,
                dim1359JoeKuoD5Init,
                dim1360JoeKuoD5Init,
                dim1361JoeKuoD5Init,
                dim1362JoeKuoD5Init,
                dim1363JoeKuoD5Init,
                dim1364JoeKuoD5Init,
                dim1365JoeKuoD5Init,
                dim1366JoeKuoD5Init,
                dim1367JoeKuoD5Init,
                dim1368JoeKuoD5Init,
                dim1369JoeKuoD5Init,
                dim1370JoeKuoD5Init,
                dim1371JoeKuoD5Init,
                dim1372JoeKuoD5Init,
                dim1373JoeKuoD5Init,
                dim1374JoeKuoD5Init,
                dim1375JoeKuoD5Init,
                dim1376JoeKuoD5Init,
                dim1377JoeKuoD5Init,
                dim1378JoeKuoD5Init,
                dim1379JoeKuoD5Init,
                dim1380JoeKuoD5Init,
                dim1381JoeKuoD5Init,
                dim1382JoeKuoD5Init,
                dim1383JoeKuoD5Init,
                dim1384JoeKuoD5Init,
                dim1385JoeKuoD5Init,
                dim1386JoeKuoD5Init,
                dim1387JoeKuoD5Init,
                dim1388JoeKuoD5Init,
                dim1389JoeKuoD5Init,
                dim1390JoeKuoD5Init,
                dim1391JoeKuoD5Init,
                dim1392JoeKuoD5Init,
                dim1393JoeKuoD5Init,
                dim1394JoeKuoD5Init,
                dim1395JoeKuoD5Init,
                dim1396JoeKuoD5Init,
                dim1397JoeKuoD5Init,
                dim1398JoeKuoD5Init,
                dim1399JoeKuoD5Init,
                dim1400JoeKuoD5Init,
                dim1401JoeKuoD5Init,
                dim1402JoeKuoD5Init,
                dim1403JoeKuoD5Init,
                dim1404JoeKuoD5Init,
                dim1405JoeKuoD5Init,
                dim1406JoeKuoD5Init,
                dim1407JoeKuoD5Init,
                dim1408JoeKuoD5Init,
                dim1409JoeKuoD5Init,
                dim1410JoeKuoD5Init,
                dim1411JoeKuoD5Init,
                dim1412JoeKuoD5Init,
                dim1413JoeKuoD5Init,
                dim1414JoeKuoD5Init,
                dim1415JoeKuoD5Init,
                dim1416JoeKuoD5Init,
                dim1417JoeKuoD5Init,
                dim1418JoeKuoD5Init,
                dim1419JoeKuoD5Init,
                dim1420JoeKuoD5Init,
                dim1421JoeKuoD5Init,
                dim1422JoeKuoD5Init,
                dim1423JoeKuoD5Init,
                dim1424JoeKuoD5Init,
                dim1425JoeKuoD5Init,
                dim1426JoeKuoD5Init,
                dim1427JoeKuoD5Init,
                dim1428JoeKuoD5Init,
                dim1429JoeKuoD5Init,
                dim1430JoeKuoD5Init,
                dim1431JoeKuoD5Init,
                dim1432JoeKuoD5Init,
                dim1433JoeKuoD5Init,
                dim1434JoeKuoD5Init,
                dim1435JoeKuoD5Init,
                dim1436JoeKuoD5Init,
                dim1437JoeKuoD5Init,
                dim1438JoeKuoD5Init,
                dim1439JoeKuoD5Init,
                dim1440JoeKuoD5Init,
                dim1441JoeKuoD5Init,
                dim1442JoeKuoD5Init,
                dim1443JoeKuoD5Init,
                dim1444JoeKuoD5Init,
                dim1445JoeKuoD5Init,
                dim1446JoeKuoD5Init,
                dim1447JoeKuoD5Init,
                dim1448JoeKuoD5Init,
                dim1449JoeKuoD5Init,
                dim1450JoeKuoD5Init,
                dim1451JoeKuoD5Init,
                dim1452JoeKuoD5Init,
                dim1453JoeKuoD5Init,
                dim1454JoeKuoD5Init,
                dim1455JoeKuoD5Init,
                dim1456JoeKuoD5Init,
                dim1457JoeKuoD5Init,
                dim1458JoeKuoD5Init,
                dim1459JoeKuoD5Init,
                dim1460JoeKuoD5Init,
                dim1461JoeKuoD5Init,
                dim1462JoeKuoD5Init,
                dim1463JoeKuoD5Init,
                dim1464JoeKuoD5Init,
                dim1465JoeKuoD5Init,
                dim1466JoeKuoD5Init,
                dim1467JoeKuoD5Init,
                dim1468JoeKuoD5Init,
                dim1469JoeKuoD5Init,
                dim1470JoeKuoD5Init,
                dim1471JoeKuoD5Init,
                dim1472JoeKuoD5Init,
                dim1473JoeKuoD5Init,
                dim1474JoeKuoD5Init,
                dim1475JoeKuoD5Init,
                dim1476JoeKuoD5Init,
                dim1477JoeKuoD5Init,
                dim1478JoeKuoD5Init,
                dim1479JoeKuoD5Init,
                dim1480JoeKuoD5Init,
                dim1481JoeKuoD5Init,
                dim1482JoeKuoD5Init,
                dim1483JoeKuoD5Init,
                dim1484JoeKuoD5Init,
                dim1485JoeKuoD5Init,
                dim1486JoeKuoD5Init,
                dim1487JoeKuoD5Init,
                dim1488JoeKuoD5Init,
                dim1489JoeKuoD5Init,
                dim1490JoeKuoD5Init,
                dim1491JoeKuoD5Init,
                dim1492JoeKuoD5Init,
                dim1493JoeKuoD5Init,
                dim1494JoeKuoD5Init,
                dim1495JoeKuoD5Init,
                dim1496JoeKuoD5Init,
                dim1497JoeKuoD5Init,
                dim1498JoeKuoD5Init,
                dim1499JoeKuoD5Init,
                dim1500JoeKuoD5Init,
                dim1501JoeKuoD5Init,
                dim1502JoeKuoD5Init,
                dim1503JoeKuoD5Init,
                dim1504JoeKuoD5Init,
                dim1505JoeKuoD5Init,
                dim1506JoeKuoD5Init,
                dim1507JoeKuoD5Init,
                dim1508JoeKuoD5Init,
                dim1509JoeKuoD5Init,
                dim1510JoeKuoD5Init,
                dim1511JoeKuoD5Init,
                dim1512JoeKuoD5Init,
                dim1513JoeKuoD5Init,
                dim1514JoeKuoD5Init,
                dim1515JoeKuoD5Init,
                dim1516JoeKuoD5Init,
                dim1517JoeKuoD5Init,
                dim1518JoeKuoD5Init,
                dim1519JoeKuoD5Init,
                dim1520JoeKuoD5Init,
                dim1521JoeKuoD5Init,
                dim1522JoeKuoD5Init,
                dim1523JoeKuoD5Init,
                dim1524JoeKuoD5Init,
                dim1525JoeKuoD5Init,
                dim1526JoeKuoD5Init,
                dim1527JoeKuoD5Init,
                dim1528JoeKuoD5Init,
                dim1529JoeKuoD5Init,
                dim1530JoeKuoD5Init,
                dim1531JoeKuoD5Init,
                dim1532JoeKuoD5Init,
                dim1533JoeKuoD5Init,
                dim1534JoeKuoD5Init,
                dim1535JoeKuoD5Init,
                dim1536JoeKuoD5Init,
                dim1537JoeKuoD5Init,
                dim1538JoeKuoD5Init,
                dim1539JoeKuoD5Init,
                dim1540JoeKuoD5Init,
                dim1541JoeKuoD5Init,
                dim1542JoeKuoD5Init,
                dim1543JoeKuoD5Init,
                dim1544JoeKuoD5Init,
                dim1545JoeKuoD5Init,
                dim1546JoeKuoD5Init,
                dim1547JoeKuoD5Init,
                dim1548JoeKuoD5Init,
                dim1549JoeKuoD5Init,
                dim1550JoeKuoD5Init,
                dim1551JoeKuoD5Init,
                dim1552JoeKuoD5Init,
                dim1553JoeKuoD5Init,
                dim1554JoeKuoD5Init,
                dim1555JoeKuoD5Init,
                dim1556JoeKuoD5Init,
                dim1557JoeKuoD5Init,
                dim1558JoeKuoD5Init,
                dim1559JoeKuoD5Init,
                dim1560JoeKuoD5Init,
                dim1561JoeKuoD5Init,
                dim1562JoeKuoD5Init,
                dim1563JoeKuoD5Init,
                dim1564JoeKuoD5Init,
                dim1565JoeKuoD5Init,
                dim1566JoeKuoD5Init,
                dim1567JoeKuoD5Init,
                dim1568JoeKuoD5Init,
                dim1569JoeKuoD5Init,
                dim1570JoeKuoD5Init,
                dim1571JoeKuoD5Init,
                dim1572JoeKuoD5Init,
                dim1573JoeKuoD5Init,
                dim1574JoeKuoD5Init,
                dim1575JoeKuoD5Init,
                dim1576JoeKuoD5Init,
                dim1577JoeKuoD5Init,
                dim1578JoeKuoD5Init,
                dim1579JoeKuoD5Init,
                dim1580JoeKuoD5Init,
                dim1581JoeKuoD5Init,
                dim1582JoeKuoD5Init,
                dim1583JoeKuoD5Init,
                dim1584JoeKuoD5Init,
                dim1585JoeKuoD5Init,
                dim1586JoeKuoD5Init,
                dim1587JoeKuoD5Init,
                dim1588JoeKuoD5Init,
                dim1589JoeKuoD5Init,
                dim1590JoeKuoD5Init,
                dim1591JoeKuoD5Init,
                dim1592JoeKuoD5Init,
                dim1593JoeKuoD5Init,
                dim1594JoeKuoD5Init,
                dim1595JoeKuoD5Init,
                dim1596JoeKuoD5Init,
                dim1597JoeKuoD5Init,
                dim1598JoeKuoD5Init,
                dim1599JoeKuoD5Init,
                dim1600JoeKuoD5Init,
                dim1601JoeKuoD5Init,
                dim1602JoeKuoD5Init,
                dim1603JoeKuoD5Init,
                dim1604JoeKuoD5Init,
                dim1605JoeKuoD5Init,
                dim1606JoeKuoD5Init,
                dim1607JoeKuoD5Init,
                dim1608JoeKuoD5Init,
                dim1609JoeKuoD5Init,
                dim1610JoeKuoD5Init,
                dim1611JoeKuoD5Init,
                dim1612JoeKuoD5Init,
                dim1613JoeKuoD5Init,
                dim1614JoeKuoD5Init,
                dim1615JoeKuoD5Init,
                dim1616JoeKuoD5Init,
                dim1617JoeKuoD5Init,
                dim1618JoeKuoD5Init,
                dim1619JoeKuoD5Init,
                dim1620JoeKuoD5Init,
                dim1621JoeKuoD5Init,
                dim1622JoeKuoD5Init,
                dim1623JoeKuoD5Init,
                dim1624JoeKuoD5Init,
                dim1625JoeKuoD5Init,
                dim1626JoeKuoD5Init,
                dim1627JoeKuoD5Init,
                dim1628JoeKuoD5Init,
                dim1629JoeKuoD5Init,
                dim1630JoeKuoD5Init,
                dim1631JoeKuoD5Init,
                dim1632JoeKuoD5Init,
                dim1633JoeKuoD5Init,
                dim1634JoeKuoD5Init,
                dim1635JoeKuoD5Init,
                dim1636JoeKuoD5Init,
                dim1637JoeKuoD5Init,
                dim1638JoeKuoD5Init,
                dim1639JoeKuoD5Init,
                dim1640JoeKuoD5Init,
                dim1641JoeKuoD5Init,
                dim1642JoeKuoD5Init,
                dim1643JoeKuoD5Init,
                dim1644JoeKuoD5Init,
                dim1645JoeKuoD5Init,
                dim1646JoeKuoD5Init,
                dim1647JoeKuoD5Init,
                dim1648JoeKuoD5Init,
                dim1649JoeKuoD5Init,
                dim1650JoeKuoD5Init,
                dim1651JoeKuoD5Init,
                dim1652JoeKuoD5Init,
                dim1653JoeKuoD5Init,
                dim1654JoeKuoD5Init,
                dim1655JoeKuoD5Init,
                dim1656JoeKuoD5Init,
                dim1657JoeKuoD5Init,
                dim1658JoeKuoD5Init,
                dim1659JoeKuoD5Init,
                dim1660JoeKuoD5Init,
                dim1661JoeKuoD5Init,
                dim1662JoeKuoD5Init,
                dim1663JoeKuoD5Init,
                dim1664JoeKuoD5Init,
                dim1665JoeKuoD5Init,
                dim1666JoeKuoD5Init,
                dim1667JoeKuoD5Init,
                dim1668JoeKuoD5Init,
                dim1669JoeKuoD5Init,
                dim1670JoeKuoD5Init,
                dim1671JoeKuoD5Init,
                dim1672JoeKuoD5Init,
                dim1673JoeKuoD5Init,
                dim1674JoeKuoD5Init,
                dim1675JoeKuoD5Init,
                dim1676JoeKuoD5Init,
                dim1677JoeKuoD5Init,
                dim1678JoeKuoD5Init,
                dim1679JoeKuoD5Init,
                dim1680JoeKuoD5Init,
                dim1681JoeKuoD5Init,
                dim1682JoeKuoD5Init,
                dim1683JoeKuoD5Init,
                dim1684JoeKuoD5Init,
                dim1685JoeKuoD5Init,
                dim1686JoeKuoD5Init,
                dim1687JoeKuoD5Init,
                dim1688JoeKuoD5Init,
                dim1689JoeKuoD5Init,
                dim1690JoeKuoD5Init,
                dim1691JoeKuoD5Init,
                dim1692JoeKuoD5Init,
                dim1693JoeKuoD5Init,
                dim1694JoeKuoD5Init,
                dim1695JoeKuoD5Init,
                dim1696JoeKuoD5Init,
                dim1697JoeKuoD5Init,
                dim1698JoeKuoD5Init,
                dim1699JoeKuoD5Init,
                dim1700JoeKuoD5Init,
                dim1701JoeKuoD5Init,
                dim1702JoeKuoD5Init,
                dim1703JoeKuoD5Init,
                dim1704JoeKuoD5Init,
                dim1705JoeKuoD5Init,
                dim1706JoeKuoD5Init,
                dim1707JoeKuoD5Init,
                dim1708JoeKuoD5Init,
                dim1709JoeKuoD5Init,
                dim1710JoeKuoD5Init,
                dim1711JoeKuoD5Init,
                dim1712JoeKuoD5Init,
                dim1713JoeKuoD5Init,
                dim1714JoeKuoD5Init,
                dim1715JoeKuoD5Init,
                dim1716JoeKuoD5Init,
                dim1717JoeKuoD5Init,
                dim1718JoeKuoD5Init,
                dim1719JoeKuoD5Init,
                dim1720JoeKuoD5Init,
                dim1721JoeKuoD5Init,
                dim1722JoeKuoD5Init,
                dim1723JoeKuoD5Init,
                dim1724JoeKuoD5Init,
                dim1725JoeKuoD5Init,
                dim1726JoeKuoD5Init,
                dim1727JoeKuoD5Init,
                dim1728JoeKuoD5Init,
                dim1729JoeKuoD5Init,
                dim1730JoeKuoD5Init,
                dim1731JoeKuoD5Init,
                dim1732JoeKuoD5Init,
                dim1733JoeKuoD5Init,
                dim1734JoeKuoD5Init,
                dim1735JoeKuoD5Init,
                dim1736JoeKuoD5Init,
                dim1737JoeKuoD5Init,
                dim1738JoeKuoD5Init,
                dim1739JoeKuoD5Init,
                dim1740JoeKuoD5Init,
                dim1741JoeKuoD5Init,
                dim1742JoeKuoD5Init,
                dim1743JoeKuoD5Init,
                dim1744JoeKuoD5Init,
                dim1745JoeKuoD5Init,
                dim1746JoeKuoD5Init,
                dim1747JoeKuoD5Init,
                dim1748JoeKuoD5Init,
                dim1749JoeKuoD5Init,
                dim1750JoeKuoD5Init,
                dim1751JoeKuoD5Init,
                dim1752JoeKuoD5Init,
                dim1753JoeKuoD5Init,
                dim1754JoeKuoD5Init,
                dim1755JoeKuoD5Init,
                dim1756JoeKuoD5Init,
                dim1757JoeKuoD5Init,
                dim1758JoeKuoD5Init,
                dim1759JoeKuoD5Init,
                dim1760JoeKuoD5Init,
                dim1761JoeKuoD5Init,
                dim1762JoeKuoD5Init,
                dim1763JoeKuoD5Init,
                dim1764JoeKuoD5Init,
                dim1765JoeKuoD5Init,
                dim1766JoeKuoD5Init,
                dim1767JoeKuoD5Init,
                dim1768JoeKuoD5Init,
                dim1769JoeKuoD5Init,
                dim1770JoeKuoD5Init,
                dim1771JoeKuoD5Init,
                dim1772JoeKuoD5Init,
                dim1773JoeKuoD5Init,
                dim1774JoeKuoD5Init,
                dim1775JoeKuoD5Init,
                dim1776JoeKuoD5Init,
                dim1777JoeKuoD5Init,
                dim1778JoeKuoD5Init,
                dim1779JoeKuoD5Init,
                dim1780JoeKuoD5Init,
                dim1781JoeKuoD5Init,
                dim1782JoeKuoD5Init,
                dim1783JoeKuoD5Init,
                dim1784JoeKuoD5Init,
                dim1785JoeKuoD5Init,
                dim1786JoeKuoD5Init,
                dim1787JoeKuoD5Init,
                dim1788JoeKuoD5Init,
                dim1789JoeKuoD5Init,
                dim1790JoeKuoD5Init,
                dim1791JoeKuoD5Init,
                dim1792JoeKuoD5Init,
                dim1793JoeKuoD5Init,
                dim1794JoeKuoD5Init,
                dim1795JoeKuoD5Init,
                dim1796JoeKuoD5Init,
                dim1797JoeKuoD5Init,
                dim1798JoeKuoD5Init,
                dim1799JoeKuoD5Init,
                dim1800JoeKuoD5Init,
                dim1801JoeKuoD5Init,
                dim1802JoeKuoD5Init,
                dim1803JoeKuoD5Init,
                dim1804JoeKuoD5Init,
                dim1805JoeKuoD5Init,
                dim1806JoeKuoD5Init,
                dim1807JoeKuoD5Init,
                dim1808JoeKuoD5Init,
                dim1809JoeKuoD5Init,
                dim1810JoeKuoD5Init,
                dim1811JoeKuoD5Init,
                dim1812JoeKuoD5Init,
                dim1813JoeKuoD5Init,
                dim1814JoeKuoD5Init,
                dim1815JoeKuoD5Init,
                dim1816JoeKuoD5Init,
                dim1817JoeKuoD5Init,
                dim1818JoeKuoD5Init,
                dim1819JoeKuoD5Init,
                dim1820JoeKuoD5Init,
                dim1821JoeKuoD5Init,
                dim1822JoeKuoD5Init,
                dim1823JoeKuoD5Init,
                dim1824JoeKuoD5Init,
                dim1825JoeKuoD5Init,
                dim1826JoeKuoD5Init,
                dim1827JoeKuoD5Init,
                dim1828JoeKuoD5Init,
                dim1829JoeKuoD5Init,
                dim1830JoeKuoD5Init,
                dim1831JoeKuoD5Init,
                dim1832JoeKuoD5Init,
                dim1833JoeKuoD5Init,
                dim1834JoeKuoD5Init,
                dim1835JoeKuoD5Init,
                dim1836JoeKuoD5Init,
                dim1837JoeKuoD5Init,
                dim1838JoeKuoD5Init,
                dim1839JoeKuoD5Init,
                dim1840JoeKuoD5Init,
                dim1841JoeKuoD5Init,
                dim1842JoeKuoD5Init,
                dim1843JoeKuoD5Init,
                dim1844JoeKuoD5Init,
                dim1845JoeKuoD5Init,
                dim1846JoeKuoD5Init,
                dim1847JoeKuoD5Init,
                dim1848JoeKuoD5Init,
                dim1849JoeKuoD5Init,
                dim1850JoeKuoD5Init,
                dim1851JoeKuoD5Init,
                dim1852JoeKuoD5Init,
                dim1853JoeKuoD5Init,
                dim1854JoeKuoD5Init,
                dim1855JoeKuoD5Init,
                dim1856JoeKuoD5Init,
                dim1857JoeKuoD5Init,
                dim1858JoeKuoD5Init,
                dim1859JoeKuoD5Init,
                dim1860JoeKuoD5Init,
                dim1861JoeKuoD5Init,
                dim1862JoeKuoD5Init,
                dim1863JoeKuoD5Init,
                dim1864JoeKuoD5Init,
                dim1865JoeKuoD5Init,
                dim1866JoeKuoD5Init,
                dim1867JoeKuoD5Init,
                dim1868JoeKuoD5Init,
                dim1869JoeKuoD5Init,
                dim1870JoeKuoD5Init,
                dim1871JoeKuoD5Init,
                dim1872JoeKuoD5Init,
                dim1873JoeKuoD5Init,
                dim1874JoeKuoD5Init,
                dim1875JoeKuoD5Init,
                dim1876JoeKuoD5Init,
                dim1877JoeKuoD5Init,
                dim1878JoeKuoD5Init,
                dim1879JoeKuoD5Init,
                dim1880JoeKuoD5Init,
                dim1881JoeKuoD5Init,
                dim1882JoeKuoD5Init,
                dim1883JoeKuoD5Init,
                dim1884JoeKuoD5Init,
                dim1885JoeKuoD5Init,
                dim1886JoeKuoD5Init,
                dim1887JoeKuoD5Init,
                dim1888JoeKuoD5Init,
                dim1889JoeKuoD5Init,
                dim1890JoeKuoD5Init,
                dim1891JoeKuoD5Init,
                dim1892JoeKuoD5Init,
                dim1893JoeKuoD5Init,
                dim1894JoeKuoD5Init,
                dim1895JoeKuoD5Init,
                dim1896JoeKuoD5Init,
                dim1897JoeKuoD5Init,
                dim1898JoeKuoD5Init,
                dim1899JoeKuoD5Init,
                dim1900JoeKuoD5Init,
                dim1901JoeKuoD5Init,
                dim1902JoeKuoD5Init,
                dim1903JoeKuoD5Init,
                dim1904JoeKuoD5Init,
                dim1905JoeKuoD5Init,
                dim1906JoeKuoD5Init,
                dim1907JoeKuoD5Init,
                dim1908JoeKuoD5Init,
                dim1909JoeKuoD5Init,
                dim1910JoeKuoD5Init,
                dim1911JoeKuoD5Init,
                dim1912JoeKuoD5Init,
                dim1913JoeKuoD5Init,
                dim1914JoeKuoD5Init,
                dim1915JoeKuoD5Init,
                dim1916JoeKuoD5Init,
                dim1917JoeKuoD5Init,
                dim1918JoeKuoD5Init,
                dim1919JoeKuoD5Init,
                dim1920JoeKuoD5Init,
                dim1921JoeKuoD5Init,
                dim1922JoeKuoD5Init,
                dim1923JoeKuoD5Init,
                dim1924JoeKuoD5Init,
                dim1925JoeKuoD5Init,
                dim1926JoeKuoD5Init,
                dim1927JoeKuoD5Init,
                dim1928JoeKuoD5Init,
                dim1929JoeKuoD5Init,
                dim1930JoeKuoD5Init,
                dim1931JoeKuoD5Init,
                dim1932JoeKuoD5Init,
                dim1933JoeKuoD5Init,
                dim1934JoeKuoD5Init,
                dim1935JoeKuoD5Init,
                dim1936JoeKuoD5Init,
                dim1937JoeKuoD5Init,
                dim1938JoeKuoD5Init,
                dim1939JoeKuoD5Init,
                dim1940JoeKuoD5Init,
                dim1941JoeKuoD5Init,
                dim1942JoeKuoD5Init,
                dim1943JoeKuoD5Init,
                dim1944JoeKuoD5Init,
                dim1945JoeKuoD5Init,
                dim1946JoeKuoD5Init,
                dim1947JoeKuoD5Init,
                dim1948JoeKuoD5Init,
                dim1949JoeKuoD5Init,
                dim1950JoeKuoD5Init,
                dim1951JoeKuoD5Init,
                dim1952JoeKuoD5Init,
                dim1953JoeKuoD5Init,
                dim1954JoeKuoD5Init,
                dim1955JoeKuoD5Init,
                dim1956JoeKuoD5Init,
                dim1957JoeKuoD5Init,
                dim1958JoeKuoD5Init,
                dim1959JoeKuoD5Init,
                dim1960JoeKuoD5Init,
                dim1961JoeKuoD5Init,
                dim1962JoeKuoD5Init,
                dim1963JoeKuoD5Init,
                dim1964JoeKuoD5Init,
                dim1965JoeKuoD5Init,
                dim1966JoeKuoD5Init,
                dim1967JoeKuoD5Init,
                dim1968JoeKuoD5Init,
                dim1969JoeKuoD5Init,
                dim1970JoeKuoD5Init,
                dim1971JoeKuoD5Init,
                dim1972JoeKuoD5Init,
                dim1973JoeKuoD5Init,
                dim1974JoeKuoD5Init,
                dim1975JoeKuoD5Init,
                dim1976JoeKuoD5Init,
                dim1977JoeKuoD5Init,
                dim1978JoeKuoD5Init,
                dim1979JoeKuoD5Init,
                dim1980JoeKuoD5Init,
                dim1981JoeKuoD5Init,
                dim1982JoeKuoD5Init,
                dim1983JoeKuoD5Init,
                dim1984JoeKuoD5Init,
                dim1985JoeKuoD5Init,
                dim1986JoeKuoD5Init,
                dim1987JoeKuoD5Init,
                dim1988JoeKuoD5Init,
                dim1989JoeKuoD5Init,
                dim1990JoeKuoD5Init,
                dim1991JoeKuoD5Init,
                dim1992JoeKuoD5Init,
                dim1993JoeKuoD5Init,
                dim1994JoeKuoD5Init,
                dim1995JoeKuoD5Init,
                dim1996JoeKuoD5Init,
                dim1997JoeKuoD5Init,
                dim1998JoeKuoD5Init,
                dim1999JoeKuoD5Init
            };

    }

    const int SobolRsg::bits_ = 8*sizeof(unsigned long);
    // 1/(2^bits_) (written as (1/2)/(2^(bits_-1)) to avoid long overflow)
    const double SobolRsg::normalizationFactor_ =
        0.5/(1UL<<(SobolRsg::bits_-1));

    SobolRsg::SobolRsg(Size dimensionality,
                       unsigned long seed,
                       DirectionIntegers directionIntegers)
    : dimensionality_(dimensionality), sequenceCounter_(0), firstDraw_(true),
      sequence_(std::vector<Real> (dimensionality), 1.0),
      integerSequence_(dimensionality, 0),
      directionIntegers_(dimensionality,std::vector<unsigned long>(bits_)) {

        QL_REQUIRE(dimensionality>0,
                   "dimensionality must be greater than 0");
      QL_REQUIRE(dimensionality<=PPMT_MAX_DIM,
               "dimensionality " << dimensionality
                   << " exceeds the number of available "
                   << "primitive polynomials modulo two ("
                   << PPMT_MAX_DIM << ")");

        // initializes coefficient array of the k-th primitive polynomial
        // and degree of the k-th primitive polynomial
        std::vector<unsigned int> degree(dimensionality_);
        std::vector<long> ppmt(dimensionality_);

        bool useAltPolynomials = false;

        if (directionIntegers == Kuo || directionIntegers == Kuo2 || directionIntegers == Kuo3
            || directionIntegers == SobolLevitan || directionIntegers == SobolLevitanLemieux)
            useAltPolynomials = true;

        // degree 0 is not used
        ppmt[0]=0;
        degree[0]=0;
        Size k, index;
        unsigned int currentDegree=1;
        k=1;
        index=0;

        unsigned int altDegree = useAltPolynomials ? maxAltDegree : 0;

        for (; k<std::min<Size>(dimensionality_,altDegree); k++,index++)
        {
            ppmt[k] = AltPrimitivePolynomials[currentDegree-1][index];
            if (ppmt[k]==-1)
            {
                ++currentDegree;
                index=0;
                ppmt[k] = AltPrimitivePolynomials[currentDegree-1][index];
            }

            degree[k] = currentDegree;
        }



        for (; k<dimensionality_; k++,index++)
        {
            ppmt[k] = PrimitivePolynomials[currentDegree-1][index];
            if (ppmt[k]==-1)
            {
                ++currentDegree;
                index=0;
                ppmt[k] = PrimitivePolynomials[currentDegree-1][index];
            }
            degree[k] = currentDegree;

        }

        // initializes bits_ direction integers for each dimension
        // and store them into directionIntegers_[dimensionality_][bits_]
        //
        // In each dimension k with its associated primitive polynomial,
        // the first degree_[k] direction integers can be chosen freely
        // provided that only the l leftmost bits can be non-zero, and
        // that the l-th leftmost bit must be set

        // degenerate (no free direction integers) first dimension
        int j;
        for (j=0; j<bits_; j++)
            directionIntegers_[0][j] = (1UL<<(bits_-j-1));


        Size maxTabulated = 0;
        // dimensions from 2 (k=1) to maxTabulated (k=maxTabulated-1) included
        // are initialized from tabulated coefficients
        switch (directionIntegers) {
            case Unit:
                maxTabulated=dimensionality_;
                for (k=1; k<maxTabulated; k++) {
                    for (Size l=1; l<=degree[k]; l++) {
                        directionIntegers_[k][l-1] = 1UL;
                        directionIntegers_[k][l-1] <<= (bits_-l);
                    }
                }
                break;
            case Jaeckel:
                // maxTabulated=32
                maxTabulated = sizeof(initializers)/sizeof(unsigned long *)+1;
                for (k=1; k<std::min(dimensionality_, maxTabulated); k++) {
                    j = 0;
                    // 0UL marks coefficients' end for a given dimension
                    while (initializers[k-1][j] != 0UL) {
                        directionIntegers_[k][j] = initializers[k-1][j];
                        directionIntegers_[k][j] <<= (bits_-j-1);
                        j++;
                    }
                }
                break;
            case SobolLevitan:
                // maxTabulated=40
                maxTabulated = sizeof(SLinitializers)/sizeof(unsigned long *)+1;
                for (k=1; k<std::min(dimensionality_, maxTabulated); k++) {
                    j = 0;
                    // 0UL marks coefficients' end for a given dimension
                    while (SLinitializers[k-1][j] != 0UL) {
                        directionIntegers_[k][j] = SLinitializers[k-1][j];
                        directionIntegers_[k][j] <<= (bits_-j-1);
                        j++;
                    }
                }
                break;
            case SobolLevitanLemieux:
                // maxTabulated=360
                maxTabulated = sizeof(Linitializers)/sizeof(unsigned long *)+1;
                for (k=1; k<std::min(dimensionality_, maxTabulated); k++) {
                    j = 0;
                    // 0UL marks coefficients' end for a given dimension
                    while (Linitializers[k-1][j] != 0UL) {
                        directionIntegers_[k][j] = Linitializers[k-1][j];
                        directionIntegers_[k][j] <<= (bits_-j-1);
                        j++;
                    }
                }
                break;
            case JoeKuoD5:
                // maxTabulated=1898
                maxTabulated = sizeof(JoeKuoD5initializers)/sizeof(unsigned long *)+1;
                for (k=1; k<std::min(dimensionality_, maxTabulated); k++) {
                    j = 0;
                    // 0UL marks coefficients' end for a given dimension
                    while (JoeKuoD5initializers[k-1][j] != 0UL) {
                        directionIntegers_[k][j] = JoeKuoD5initializers[k-1][j];
                        directionIntegers_[k][j] <<= (bits_-j-1);
                        j++;
                    }
                }
                break;
            case JoeKuoD6:
                // maxTabulated=1799
                maxTabulated = sizeof(JoeKuoD6initializers)/sizeof(unsigned long *)+1;
                for (k=1; k<std::min(dimensionality_, maxTabulated); k++) {
                    j = 0;
                    // 0UL marks coefficients' end for a given dimension
                    while (JoeKuoD6initializers[k-1][j] != 0UL) {
                        directionIntegers_[k][j] = JoeKuoD6initializers[k-1][j];
                        directionIntegers_[k][j] <<= (bits_-j-1);
                        j++;
                    }
                }
                break;
            case JoeKuoD7:
                // maxTabulated=1898
                maxTabulated = sizeof(JoeKuoD7initializers)/sizeof(unsigned long *)+1;
                for (k=1; k<std::min(dimensionality_, maxTabulated); k++) {
                    j = 0;
                    // 0UL marks coefficients' end for a given dimension
                    while (JoeKuoD7initializers[k-1][j] != 0UL) {
                        directionIntegers_[k][j] = JoeKuoD7initializers[k-1][j];
                        directionIntegers_[k][j] <<= (bits_-j-1);
                        j++;
                    }
                }
                break;



            case Kuo:
                // maxTabulated=4925
                maxTabulated = sizeof(Kuoinitializers)/sizeof(unsigned long *)+1;
                for (k=1; k<std::min(dimensionality_, maxTabulated); k++) {
                    j = 0;
                    // 0UL marks coefficients' end for a given dimension
                    while (Kuoinitializers[k-1][j] != 0UL) {
                        directionIntegers_[k][j] = Kuoinitializers[k-1][j];
                        directionIntegers_[k][j] <<= (bits_-j-1);
                        j++;
                    }
                }
                break;
            case Kuo2:
                // maxTabulated=3946
                maxTabulated = sizeof(Kuo2initializers)/sizeof(unsigned long *)+1;
                for (k=1; k<std::min(dimensionality_, maxTabulated); k++) {
                    j = 0;
                    // 0UL marks coefficients' end for a given dimension
                    while (Kuo2initializers[k-1][j] != 0UL) {
                        directionIntegers_[k][j] = Kuo2initializers[k-1][j];
                        directionIntegers_[k][j] <<= (bits_-j-1);
                        j++;
                    }
                }
                break;

            case Kuo3:
                // maxTabulated=4585
                maxTabulated = sizeof(Kuo3initializers)/sizeof(unsigned long *)+1;
                for (k=1; k<std::min(dimensionality_, maxTabulated); k++) {
                    j = 0;
                    // 0UL marks coefficients' end for a given dimension
                    while (Kuo3initializers[k-1][j] != 0UL) {
                        directionIntegers_[k][j] = Kuo3initializers[k-1][j];
                        directionIntegers_[k][j] <<= (bits_-j-1);
                        j++;
                    }
                }
                break;

            default:
                break;
        }

        // random initialization for higher dimensions
        if (dimensionality_>maxTabulated) {
            MersenneTwisterUniformRng uniformRng(seed);
            for (k=maxTabulated; k<dimensionality_; k++) {
                for (Size l=1; l<=degree[k]; l++) {
                    do {
                        // u is in (0,1)
                        Real u = uniformRng.next().value;
                        // the direction integer has at most the
                        // rightmost l bits non-zero
                        directionIntegers_[k][l-1] =
                            (unsigned long)(u*(1UL<<l));
                    } while (!(directionIntegers_[k][l-1] & 1UL));
                    // iterate until the direction integer is odd
                    // that is it has the rightmost bit set

                    // shifting bits_-l bits to the left
                    // we are guaranteed that the l-th leftmost bit
                    // is set, and only the first l leftmost bit
                    // can be non-zero
                    directionIntegers_[k][l-1] <<= (bits_-l);
                }
            }
        }

        // computation of directionIntegers_[k][l] for l>=degree_[k]
        // by recurrence relation
        for (k=1; k<dimensionality_; k++) {
            unsigned int gk = degree[k];
            for (int l=gk; l<bits_; l++) {
                // eq. 8.19 "Monte Carlo Methods in Finance" by P. Jckel
                unsigned long n = (directionIntegers_[k][l-gk]>>gk);
                // a[k][j] are the coefficients of the monomials in ppmt[k]
                // The highest order coefficient a[k][0] is not actually
                // used in the recurrence relation, and the lowest order
                // coefficient a[k][gk] is always set: this is the reason
                // why the highest and lowest coefficient of
                // the polynomial ppmt[k] are not included in its encoding,
                // provided that its degree is known.
                // That is: a[k][j] = ppmt[k] >> (gk-j-1)
                for (Size j=1; j<gk; j++) {
                    // XORed with a selection of (unshifted) direction
                    // integers controlled by which of the a[k][j] are set
                    if ((ppmt[k] >> (gk-j-1)) & 1UL)
                        n ^= directionIntegers_[k][l-j];
                }
                // a[k][gk] is always set, so directionIntegers_[k][l-gk]
                // will always enter
                n ^= directionIntegers_[k][l-gk];
                directionIntegers_[k][l]=n;
            }
        }

        // in case one needs to check the directionIntegers used
        /* bool printDirectionIntegers = false;
           if (printDirectionIntegers) {
               std::ofstream outStream("directionIntegers.txt");
               for (k=0; k<std::min(32UL,dimensionality_); k++) {
                   outStream << std::endl << k+1       << "\t"
                                          << degree[k] << "\t"
                                          << ppmt[k]   << "\t";
                   for (j=0; j<10; j++) {
                       outStream << io::power_of_two(
                           directionIntegers_[k][j]) << "\t";
                   }
               }
               outStream.close();
           }
        */

        // initialize the Sobol integer/double vectors
        // first draw
        for (k=0; k<dimensionality_; k++) {
            integerSequence_[k]=directionIntegers_[k][0];
        }
    }

    void SobolRsg::skipTo(unsigned long skip) {
        unsigned long N = skip+1;
        unsigned int ops = (unsigned int)(std::log((double)N)/M_LN2)+1;

        // Convert to Gray code
        unsigned long G = N ^ (N>>1);
        for (Size k=0; k<dimensionality_; k++) {
            integerSequence_[k] = 0;
            for (Size index=0; index<ops; index++) {
                if (G>>index & 1)
                    integerSequence_[k] ^= directionIntegers_[k][index];
            }
        }

        sequenceCounter_ = skip;
    }



    const std::vector<unsigned long>& SobolRsg::nextInt32Sequence() const
    {
        if (firstDraw_) {
            // it was precomputed in the constructor
            firstDraw_ = false;
            return integerSequence_;
        }
        // increment the counter
        sequenceCounter_++;
        // did we overflow?
     QL_REQUIRE(sequenceCounter_ != 0, "period exceeded");

        // instead of using the counter n as new unique generating integer
        // for the n-th draw use the Gray code G(n) as proposed
        // by Antonov and Saleev
        unsigned long n = sequenceCounter_;
        // Find rightmost zero bit of n
        int j = 0;
        while (n & 1) { n >>= 1; j++; }
        for (Size k=0; k<dimensionality_; k++) {
            // XOR the appropriate direction number into each component of
            // the integer sequence to obtain a new Sobol integer for that
            // component
            integerSequence_[k] ^= directionIntegers_[k][j];
        }
        return integerSequence_;
    }

}
